<!doctype html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <title>MemoteReportApp</title>
  <base href="/">

  <meta name="viewport" content="width=device-width, initial-scale=1">
  <link rel="icon" type="image/x-icon" href="data:image/vnd.microsoft.icon;base64, ">
  <script type="text/javascript" async src="https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.5/latest.js?config=TeX-MML-AM_CHTML">
</script>
<link rel="stylesheet" type="text/css" href="https://cdn.jsdelivr.net/npm/taucharts@2/dist/taucharts.min.css">
<script type="text/x-mathjax-config"> 
  MathJax.Hub.Config({ 
      "HTML-CSS": { scale: 50, linebreaks: { automatic: true } }, 
      SVG: { linebreaks: { automatic:true } }, 
      displayAlign: "left" });
</script>
</head>
<body style="background-color: rgba(177, 212, 237, 0.1)">
  <app-root></app-root>
  <script>
    window.data = $results;
    window.reportType = '$report_type';
  </script>
<script type="text/javascript">
!function(e){function r(r){for(var n,i,f=r[0],l=r[1],a=r[2],c=0,s=[];c<f.length;c++)i=f[c],o[i]&&s.push(o[i][0]),o[i]=0;for(n in l)Object.prototype.hasOwnProperty.call(l,n)&&(e[n]=l[n]);for(p&&p(r);s.length;)s.shift()();return u.push.apply(u,a||[]),t()}function t(){for(var e,r=0;r<u.length;r++){for(var t=u[r],n=!0,f=1;f<t.length;f++){var l=t[f];0!==o[l]&&(n=!1)}n&&(u.splice(r--,1),e=i(i.s=t[0]))}return e}var n={},o={runtime:0},u=[];function i(r){if(n[r])return n[r].exports;var t=n[r]={i:r,l:!1,exports:{}};return e[r].call(t.exports,t,t.exports,i),t.l=!0,t.exports}i.m=e,i.c=n,i.d=function(e,r,t){i.o(e,r)||Object.defineProperty(e,r,{enumerable:!0,get:t})},i.r=function(e){"undefined"!=typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(e,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(e,"__esModule",{value:!0})},i.t=function(e,r){if(1&r&&(e=i(e)),8&r)return e;if(4&r&&"object"==typeof e&&e&&e.__esModule)return e;var t=Object.create(null);if(i.r(t),Object.defineProperty(t,"default",{enumerable:!0,value:e}),2&r&&"string"!=typeof e)for(var n in e)i.d(t,n,function(r){return e[r]}.bind(null,n));return t},i.n=function(e){var r=e&&e.__esModule?function(){return e.default}:function(){return e};return i.d(r,"a",r),r},i.o=function(e,r){return Object.prototype.hasOwnProperty.call(e,r)},i.p="";var f=window.webpackJsonp=window.webpackJsonp||[],l=f.push.bind(f);f.push=r,f=f.slice();for(var a=0;a<f.length;a++)r(f[a]);var p=l;t()}([]);
</script><script type="text/javascript">
(window.webpackJsonp=window.webpackJsonp||[]).push([["polyfills"],{"./node_modules/@angular-devkit/build-angular/src/angular-cli-files/models/jit-polyfills.js":function(e,o,t){"use strict";t.r(o);t("./node_modules/core-js/es7/reflect.js")},"./node_modules/core-js/es7/reflect.js":function(e,o,t){t("./node_modules/core-js/modules/es7.reflect.define-metadata.js"),t("./node_modules/core-js/modules/es7.reflect.delete-metadata.js"),t("./node_modules/core-js/modules/es7.reflect.get-metadata.js"),t("./node_modules/core-js/modules/es7.reflect.get-metadata-keys.js"),t("./node_modules/core-js/modules/es7.reflect.get-own-metadata.js"),t("./node_modules/core-js/modules/es7.reflect.get-own-metadata-keys.js"),t("./node_modules/core-js/modules/es7.reflect.has-metadata.js"),t("./node_modules/core-js/modules/es7.reflect.has-own-metadata.js"),t("./node_modules/core-js/modules/es7.reflect.metadata.js"),e.exports=t("./node_modules/core-js/modules/_core.js").Reflect},"./node_modules/core-js/modules/_a-function.js":function(e,o){e.exports=function(e){if("function"!=typeof e)throw TypeError(e+" is not a function!");return e}},"./node_modules/core-js/modules/_an-instance.js":function(e,o){e.exports=function(e,o,t,n){if(!(e instanceof o)||void 0!==n&&n in e)throw TypeError(t+": incorrect invocation!");return e}},"./node_modules/core-js/modules/_an-object.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_is-object.js");e.exports=function(e){if(!n(e))throw TypeError(e+" is not an object!");return e}},"./node_modules/core-js/modules/_array-from-iterable.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_for-of.js");e.exports=function(e,o){var t=[];return n(e,!1,t.push,t,o),t}},"./node_modules/core-js/modules/_array-includes.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_to-iobject.js"),r=t("./node_modules/core-js/modules/_to-length.js"),s=t("./node_modules/core-js/modules/_to-absolute-index.js");e.exports=function(e){return function(o,t,i){var a,u=n(o),c=r(u.length),l=s(i,c);if(e&&t!=t){for(;c>l;)if((a=u[l++])!=a)return!0}else for(;c>l;l++)if((e||l in u)&&u[l]===t)return e||l||0;return!e&&-1}}},"./node_modules/core-js/modules/_array-methods.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_ctx.js"),r=t("./node_modules/core-js/modules/_iobject.js"),s=t("./node_modules/core-js/modules/_to-object.js"),i=t("./node_modules/core-js/modules/_to-length.js"),a=t("./node_modules/core-js/modules/_array-species-create.js");e.exports=function(e,o){var t=1==e,u=2==e,c=3==e,l=4==e,d=6==e,f=5==e||d,m=o||a;return function(o,a,p){for(var _,h,j=s(o),v=r(j),g=n(a,p,3),y=i(v.length),b=0,k=t?m(o,y):u?m(o,0):void 0;y>b;b++)if((f||b in v)&&(h=g(_=v[b],b,j),e))if(t)k[b]=h;else if(h)switch(e){case 3:return!0;case 5:return _;case 6:return b;case 2:k.push(_)}else if(l)return!1;return d?-1:c||l?l:k}}},"./node_modules/core-js/modules/_array-species-constructor.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_is-object.js"),r=t("./node_modules/core-js/modules/_is-array.js"),s=t("./node_modules/core-js/modules/_wks.js")("species");e.exports=function(e){var o;return r(e)&&("function"!=typeof(o=e.constructor)||o!==Array&&!r(o.prototype)||(o=void 0),n(o)&&null===(o=o[s])&&(o=void 0)),void 0===o?Array:o}},"./node_modules/core-js/modules/_array-species-create.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_array-species-constructor.js");e.exports=function(e,o){return new(n(e))(o)}},"./node_modules/core-js/modules/_classof.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_cof.js"),r=t("./node_modules/core-js/modules/_wks.js")("toStringTag"),s="Arguments"==n(function(){return arguments}());e.exports=function(e){var o,t,i;return void 0===e?"Undefined":null===e?"Null":"string"==typeof(t=function(e,o){try{return e[o]}catch(e){}}(o=Object(e),r))?t:s?n(o):"Object"==(i=n(o))&&"function"==typeof o.callee?"Arguments":i}},"./node_modules/core-js/modules/_cof.js":function(e,o){var t={}.toString;e.exports=function(e){return t.call(e).slice(8,-1)}},"./node_modules/core-js/modules/_collection-strong.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_object-dp.js").f,r=t("./node_modules/core-js/modules/_object-create.js"),s=t("./node_modules/core-js/modules/_redefine-all.js"),i=t("./node_modules/core-js/modules/_ctx.js"),a=t("./node_modules/core-js/modules/_an-instance.js"),u=t("./node_modules/core-js/modules/_for-of.js"),c=t("./node_modules/core-js/modules/_iter-define.js"),l=t("./node_modules/core-js/modules/_iter-step.js"),d=t("./node_modules/core-js/modules/_set-species.js"),f=t("./node_modules/core-js/modules/_descriptors.js"),m=t("./node_modules/core-js/modules/_meta.js").fastKey,p=t("./node_modules/core-js/modules/_validate-collection.js"),_=f?"_s":"size",h=function(e,o){var t,n=m(o);if("F"!==n)return e._i[n];for(t=e._f;t;t=t.n)if(t.k==o)return t};e.exports={getConstructor:function(e,o,t,c){var l=e(function(e,n){a(e,l,o,"_i"),e._t=o,e._i=r(null),e._f=void 0,e._l=void 0,e[_]=0,void 0!=n&&u(n,t,e[c],e)});return s(l.prototype,{clear:function(){for(var e=p(this,o),t=e._i,n=e._f;n;n=n.n)n.r=!0,n.p&&(n.p=n.p.n=void 0),delete t[n.i];e._f=e._l=void 0,e[_]=0},delete:function(e){var t=p(this,o),n=h(t,e);if(n){var r=n.n,s=n.p;delete t._i[n.i],n.r=!0,s&&(s.n=r),r&&(r.p=s),t._f==n&&(t._f=r),t._l==n&&(t._l=s),t[_]--}return!!n},forEach:function(e){p(this,o);for(var t,n=i(e,arguments.length>1?arguments[1]:void 0,3);t=t?t.n:this._f;)for(n(t.v,t.k,this);t&&t.r;)t=t.p},has:function(e){return!!h(p(this,o),e)}}),f&&n(l.prototype,"size",{get:function(){return p(this,o)[_]}}),l},def:function(e,o,t){var n,r,s=h(e,o);return s?s.v=t:(e._l=s={i:r=m(o,!0),k:o,v:t,p:n=e._l,n:void 0,r:!1},e._f||(e._f=s),n&&(n.n=s),e[_]++,"F"!==r&&(e._i[r]=s)),e},getEntry:h,setStrong:function(e,o,t){c(e,o,function(e,t){this._t=p(e,o),this._k=t,this._l=void 0},function(){for(var e=this._k,o=this._l;o&&o.r;)o=o.p;return this._t&&(this._l=o=o?o.n:this._t._f)?l(0,"keys"==e?o.k:"values"==e?o.v:[o.k,o.v]):(this._t=void 0,l(1))},t?"entries":"values",!t,!0),d(o)}}},"./node_modules/core-js/modules/_collection-weak.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_redefine-all.js"),r=t("./node_modules/core-js/modules/_meta.js").getWeak,s=t("./node_modules/core-js/modules/_an-object.js"),i=t("./node_modules/core-js/modules/_is-object.js"),a=t("./node_modules/core-js/modules/_an-instance.js"),u=t("./node_modules/core-js/modules/_for-of.js"),c=t("./node_modules/core-js/modules/_array-methods.js"),l=t("./node_modules/core-js/modules/_has.js"),d=t("./node_modules/core-js/modules/_validate-collection.js"),f=c(5),m=c(6),p=0,_=function(e){return e._l||(e._l=new h)},h=function(){this.a=[]},j=function(e,o){return f(e.a,function(e){return e[0]===o})};h.prototype={get:function(e){var o=j(this,e);if(o)return o[1]},has:function(e){return!!j(this,e)},set:function(e,o){var t=j(this,e);t?t[1]=o:this.a.push([e,o])},delete:function(e){var o=m(this.a,function(o){return o[0]===e});return~o&&this.a.splice(o,1),!!~o}},e.exports={getConstructor:function(e,o,t,s){var c=e(function(e,n){a(e,c,o,"_i"),e._t=o,e._i=p++,e._l=void 0,void 0!=n&&u(n,t,e[s],e)});return n(c.prototype,{delete:function(e){if(!i(e))return!1;var t=r(e);return!0===t?_(d(this,o)).delete(e):t&&l(t,this._i)&&delete t[this._i]},has:function(e){if(!i(e))return!1;var t=r(e);return!0===t?_(d(this,o)).has(e):t&&l(t,this._i)}}),c},def:function(e,o,t){var n=r(s(o),!0);return!0===n?_(e).set(o,t):n[e._i]=t,e},ufstore:_}},"./node_modules/core-js/modules/_collection.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_global.js"),r=t("./node_modules/core-js/modules/_export.js"),s=t("./node_modules/core-js/modules/_redefine.js"),i=t("./node_modules/core-js/modules/_redefine-all.js"),a=t("./node_modules/core-js/modules/_meta.js"),u=t("./node_modules/core-js/modules/_for-of.js"),c=t("./node_modules/core-js/modules/_an-instance.js"),l=t("./node_modules/core-js/modules/_is-object.js"),d=t("./node_modules/core-js/modules/_fails.js"),f=t("./node_modules/core-js/modules/_iter-detect.js"),m=t("./node_modules/core-js/modules/_set-to-string-tag.js"),p=t("./node_modules/core-js/modules/_inherit-if-required.js");e.exports=function(e,o,t,_,h,j){var v=n[e],g=v,y=h?"set":"add",b=g&&g.prototype,k={},w=function(e){var o=b[e];s(b,e,"delete"==e?function(e){return!(j&&!l(e))&&o.call(this,0===e?0:e)}:"has"==e?function(e){return!(j&&!l(e))&&o.call(this,0===e?0:e)}:"get"==e?function(e){return j&&!l(e)?void 0:o.call(this,0===e?0:e)}:"add"==e?function(e){return o.call(this,0===e?0:e),this}:function(e,t){return o.call(this,0===e?0:e,t),this})};if("function"==typeof g&&(j||b.forEach&&!d(function(){(new g).entries().next()}))){var T=new g,E=T[y](j?{}:-0,1)!=T,x=d(function(){T.has(1)}),O=f(function(e){new g(e)}),S=!j&&d(function(){for(var e=new g,o=5;o--;)e[y](o,o);return!e.has(-0)});O||((g=o(function(o,t){c(o,g,e);var n=p(new v,o,g);return void 0!=t&&u(t,h,n[y],n),n})).prototype=b,b.constructor=g),(x||S)&&(w("delete"),w("has"),h&&w("get")),(S||E)&&w(y),j&&b.clear&&delete b.clear}else g=_.getConstructor(o,e,h,y),i(g.prototype,t),a.NEED=!0;return m(g,e),k[e]=g,r(r.G+r.W+r.F*(g!=v),k),j||_.setStrong(g,e,h),g}},"./node_modules/core-js/modules/_core.js":function(e,o){var t=e.exports={version:"2.5.1"};"number"==typeof __e&&(__e=t)},"./node_modules/core-js/modules/_ctx.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_a-function.js");e.exports=function(e,o,t){if(n(e),void 0===o)return e;switch(t){case 1:return function(t){return e.call(o,t)};case 2:return function(t,n){return e.call(o,t,n)};case 3:return function(t,n,r){return e.call(o,t,n,r)}}return function(){return e.apply(o,arguments)}}},"./node_modules/core-js/modules/_defined.js":function(e,o){e.exports=function(e){if(void 0==e)throw TypeError("Can't call method on  "+e);return e}},"./node_modules/core-js/modules/_descriptors.js":function(e,o,t){e.exports=!t("./node_modules/core-js/modules/_fails.js")(function(){return 7!=Object.defineProperty({},"a",{get:function(){return 7}}).a})},"./node_modules/core-js/modules/_dom-create.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_is-object.js"),r=t("./node_modules/core-js/modules/_global.js").document,s=n(r)&&n(r.createElement);e.exports=function(e){return s?r.createElement(e):{}}},"./node_modules/core-js/modules/_enum-bug-keys.js":function(e,o){e.exports="constructor,hasOwnProperty,isPrototypeOf,propertyIsEnumerable,toLocaleString,toString,valueOf".split(",")},"./node_modules/core-js/modules/_export.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_global.js"),r=t("./node_modules/core-js/modules/_core.js"),s=t("./node_modules/core-js/modules/_hide.js"),i=t("./node_modules/core-js/modules/_redefine.js"),a=t("./node_modules/core-js/modules/_ctx.js"),u=function(e,o,t){var c,l,d,f,m=e&u.F,p=e&u.G,_=e&u.S,h=e&u.P,j=e&u.B,v=p?n:_?n[o]||(n[o]={}):(n[o]||{}).prototype,g=p?r:r[o]||(r[o]={}),y=g.prototype||(g.prototype={});p&&(t=o);for(c in t)d=((l=!m&&v&&void 0!==v[c])?v:t)[c],f=j&&l?a(d,n):h&&"function"==typeof d?a(Function.call,d):d,v&&i(v,c,d,e&u.U),g[c]!=d&&s(g,c,f),h&&y[c]!=d&&(y[c]=d)};n.core=r,u.F=1,u.G=2,u.S=4,u.P=8,u.B=16,u.W=32,u.U=64,u.R=128,e.exports=u},"./node_modules/core-js/modules/_fails.js":function(e,o){e.exports=function(e){try{return!!e()}catch(e){return!0}}},"./node_modules/core-js/modules/_for-of.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_ctx.js"),r=t("./node_modules/core-js/modules/_iter-call.js"),s=t("./node_modules/core-js/modules/_is-array-iter.js"),i=t("./node_modules/core-js/modules/_an-object.js"),a=t("./node_modules/core-js/modules/_to-length.js"),u=t("./node_modules/core-js/modules/core.get-iterator-method.js"),c={},l={};(o=e.exports=function(e,o,t,d,f){var m,p,_,h,j=f?function(){return e}:u(e),v=n(t,d,o?2:1),g=0;if("function"!=typeof j)throw TypeError(e+" is not iterable!");if(s(j)){for(m=a(e.length);m>g;g++)if((h=o?v(i(p=e[g])[0],p[1]):v(e[g]))===c||h===l)return h}else for(_=j.call(e);!(p=_.next()).done;)if((h=r(_,v,p.value,o))===c||h===l)return h}).BREAK=c,o.RETURN=l},"./node_modules/core-js/modules/_global.js":function(e,o){var t=e.exports="undefined"!=typeof window&&window.Math==Math?window:"undefined"!=typeof self&&self.Math==Math?self:Function("return this")();"number"==typeof __g&&(__g=t)},"./node_modules/core-js/modules/_has.js":function(e,o){var t={}.hasOwnProperty;e.exports=function(e,o){return t.call(e,o)}},"./node_modules/core-js/modules/_hide.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_object-dp.js"),r=t("./node_modules/core-js/modules/_property-desc.js");e.exports=t("./node_modules/core-js/modules/_descriptors.js")?function(e,o,t){return n.f(e,o,r(1,t))}:function(e,o,t){return e[o]=t,e}},"./node_modules/core-js/modules/_html.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_global.js").document;e.exports=n&&n.documentElement},"./node_modules/core-js/modules/_ie8-dom-define.js":function(e,o,t){e.exports=!t("./node_modules/core-js/modules/_descriptors.js")&&!t("./node_modules/core-js/modules/_fails.js")(function(){return 7!=Object.defineProperty(t("./node_modules/core-js/modules/_dom-create.js")("div"),"a",{get:function(){return 7}}).a})},"./node_modules/core-js/modules/_inherit-if-required.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_is-object.js"),r=t("./node_modules/core-js/modules/_set-proto.js").set;e.exports=function(e,o,t){var s,i=o.constructor;return i!==t&&"function"==typeof i&&(s=i.prototype)!==t.prototype&&n(s)&&r&&r(e,s),e}},"./node_modules/core-js/modules/_iobject.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_cof.js");e.exports=Object("z").propertyIsEnumerable(0)?Object:function(e){return"String"==n(e)?e.split(""):Object(e)}},"./node_modules/core-js/modules/_is-array-iter.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_iterators.js"),r=t("./node_modules/core-js/modules/_wks.js")("iterator"),s=Array.prototype;e.exports=function(e){return void 0!==e&&(n.Array===e||s[r]===e)}},"./node_modules/core-js/modules/_is-array.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_cof.js");e.exports=Array.isArray||function(e){return"Array"==n(e)}},"./node_modules/core-js/modules/_is-object.js":function(e,o){e.exports=function(e){return"object"==typeof e?null!==e:"function"==typeof e}},"./node_modules/core-js/modules/_iter-call.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_an-object.js");e.exports=function(e,o,t,r){try{return r?o(n(t)[0],t[1]):o(t)}catch(o){var s=e.return;throw void 0!==s&&n(s.call(e)),o}}},"./node_modules/core-js/modules/_iter-create.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_object-create.js"),r=t("./node_modules/core-js/modules/_property-desc.js"),s=t("./node_modules/core-js/modules/_set-to-string-tag.js"),i={};t("./node_modules/core-js/modules/_hide.js")(i,t("./node_modules/core-js/modules/_wks.js")("iterator"),function(){return this}),e.exports=function(e,o,t){e.prototype=n(i,{next:r(1,t)}),s(e,o+" Iterator")}},"./node_modules/core-js/modules/_iter-define.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_library.js"),r=t("./node_modules/core-js/modules/_export.js"),s=t("./node_modules/core-js/modules/_redefine.js"),i=t("./node_modules/core-js/modules/_hide.js"),a=t("./node_modules/core-js/modules/_has.js"),u=t("./node_modules/core-js/modules/_iterators.js"),c=t("./node_modules/core-js/modules/_iter-create.js"),l=t("./node_modules/core-js/modules/_set-to-string-tag.js"),d=t("./node_modules/core-js/modules/_object-gpo.js"),f=t("./node_modules/core-js/modules/_wks.js")("iterator"),m=!([].keys&&"next"in[].keys()),p=function(){return this};e.exports=function(e,o,t,_,h,j,v){c(t,o,_);var g,y,b,k=function(e){if(!m&&e in x)return x[e];switch(e){case"keys":case"values":return function(){return new t(this,e)}}return function(){return new t(this,e)}},w=o+" Iterator",T="values"==h,E=!1,x=e.prototype,O=x[f]||x["@@iterator"]||h&&x[h],S=O||k(h),P=h?T?k("entries"):S:void 0,D="Array"==o?x.entries||O:O;if(D&&(b=d(D.call(new e)))!==Object.prototype&&b.next&&(l(b,w,!0),n||a(b,f)||i(b,f,p)),T&&O&&"values"!==O.name&&(E=!0,S=function(){return O.call(this)}),n&&!v||!m&&!E&&x[f]||i(x,f,S),u[o]=S,u[w]=p,h)if(g={values:T?S:k("values"),keys:j?S:k("keys"),entries:P},v)for(y in g)y in x||s(x,y,g[y]);else r(r.P+r.F*(m||E),o,g);return g}},"./node_modules/core-js/modules/_iter-detect.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_wks.js")("iterator"),r=!1;try{var s=[7][n]();s.return=function(){r=!0},Array.from(s,function(){throw 2})}catch(e){}e.exports=function(e,o){if(!o&&!r)return!1;var t=!1;try{var s=[7],i=s[n]();i.next=function(){return{done:t=!0}},s[n]=function(){return i},e(s)}catch(e){}return t}},"./node_modules/core-js/modules/_iter-step.js":function(e,o){e.exports=function(e,o){return{value:o,done:!!e}}},"./node_modules/core-js/modules/_iterators.js":function(e,o){e.exports={}},"./node_modules/core-js/modules/_library.js":function(e,o){e.exports=!1},"./node_modules/core-js/modules/_meta.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_uid.js")("meta"),r=t("./node_modules/core-js/modules/_is-object.js"),s=t("./node_modules/core-js/modules/_has.js"),i=t("./node_modules/core-js/modules/_object-dp.js").f,a=0,u=Object.isExtensible||function(){return!0},c=!t("./node_modules/core-js/modules/_fails.js")(function(){return u(Object.preventExtensions({}))}),l=function(e){i(e,n,{value:{i:"O"+ ++a,w:{}}})},d=e.exports={KEY:n,NEED:!1,fastKey:function(e,o){if(!r(e))return"symbol"==typeof e?e:("string"==typeof e?"S":"P")+e;if(!s(e,n)){if(!u(e))return"F";if(!o)return"E";l(e)}return e[n].i},getWeak:function(e,o){if(!s(e,n)){if(!u(e))return!0;if(!o)return!1;l(e)}return e[n].w},onFreeze:function(e){return c&&d.NEED&&u(e)&&!s(e,n)&&l(e),e}}},"./node_modules/core-js/modules/_metadata.js":function(e,o,t){var n=t("./node_modules/core-js/modules/es6.map.js"),r=t("./node_modules/core-js/modules/_export.js"),s=t("./node_modules/core-js/modules/_shared.js")("metadata"),i=s.store||(s.store=new(t("./node_modules/core-js/modules/es6.weak-map.js"))),a=function(e,o,t){var r=i.get(e);if(!r){if(!t)return;i.set(e,r=new n)}var s=r.get(o);if(!s){if(!t)return;r.set(o,s=new n)}return s};e.exports={store:i,map:a,has:function(e,o,t){var n=a(o,t,!1);return void 0!==n&&n.has(e)},get:function(e,o,t){var n=a(o,t,!1);return void 0===n?void 0:n.get(e)},set:function(e,o,t,n){a(t,n,!0).set(e,o)},keys:function(e,o){var t=a(e,o,!1),n=[];return t&&t.forEach(function(e,o){n.push(o)}),n},key:function(e){return void 0===e||"symbol"==typeof e?e:String(e)},exp:function(e){r(r.S,"Reflect",e)}}},"./node_modules/core-js/modules/_object-assign.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_object-keys.js"),r=t("./node_modules/core-js/modules/_object-gops.js"),s=t("./node_modules/core-js/modules/_object-pie.js"),i=t("./node_modules/core-js/modules/_to-object.js"),a=t("./node_modules/core-js/modules/_iobject.js"),u=Object.assign;e.exports=!u||t("./node_modules/core-js/modules/_fails.js")(function(){var e={},o={},t=Symbol(),n="abcdefghijklmnopqrst";return e[t]=7,n.split("").forEach(function(e){o[e]=e}),7!=u({},e)[t]||Object.keys(u({},o)).join("")!=n})?function(e,o){for(var t=i(e),u=arguments.length,c=1,l=r.f,d=s.f;u>c;)for(var f,m=a(arguments[c++]),p=l?n(m).concat(l(m)):n(m),_=p.length,h=0;_>h;)d.call(m,f=p[h++])&&(t[f]=m[f]);return t}:u},"./node_modules/core-js/modules/_object-create.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_an-object.js"),r=t("./node_modules/core-js/modules/_object-dps.js"),s=t("./node_modules/core-js/modules/_enum-bug-keys.js"),i=t("./node_modules/core-js/modules/_shared-key.js")("IE_PROTO"),a=function(){},u=function(){var e,o=t("./node_modules/core-js/modules/_dom-create.js")("iframe"),n=s.length;for(o.style.display="none",t("./node_modules/core-js/modules/_html.js").appendChild(o),o.src="javascript:",(e=o.contentWindow.document).open(),e.write("<script>document.F=Object<\/script>"),e.close(),u=e.F;n--;)delete u.prototype[s[n]];return u()};e.exports=Object.create||function(e,o){var t;return null!==e?(a.prototype=n(e),t=new a,a.prototype=null,t[i]=e):t=u(),void 0===o?t:r(t,o)}},"./node_modules/core-js/modules/_object-dp.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_an-object.js"),r=t("./node_modules/core-js/modules/_ie8-dom-define.js"),s=t("./node_modules/core-js/modules/_to-primitive.js"),i=Object.defineProperty;o.f=t("./node_modules/core-js/modules/_descriptors.js")?Object.defineProperty:function(e,o,t){if(n(e),o=s(o,!0),n(t),r)try{return i(e,o,t)}catch(e){}if("get"in t||"set"in t)throw TypeError("Accessors not supported!");return"value"in t&&(e[o]=t.value),e}},"./node_modules/core-js/modules/_object-dps.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_object-dp.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=t("./node_modules/core-js/modules/_object-keys.js");e.exports=t("./node_modules/core-js/modules/_descriptors.js")?Object.defineProperties:function(e,o){r(e);for(var t,i=s(o),a=i.length,u=0;a>u;)n.f(e,t=i[u++],o[t]);return e}},"./node_modules/core-js/modules/_object-gopd.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_object-pie.js"),r=t("./node_modules/core-js/modules/_property-desc.js"),s=t("./node_modules/core-js/modules/_to-iobject.js"),i=t("./node_modules/core-js/modules/_to-primitive.js"),a=t("./node_modules/core-js/modules/_has.js"),u=t("./node_modules/core-js/modules/_ie8-dom-define.js"),c=Object.getOwnPropertyDescriptor;o.f=t("./node_modules/core-js/modules/_descriptors.js")?c:function(e,o){if(e=s(e),o=i(o,!0),u)try{return c(e,o)}catch(e){}if(a(e,o))return r(!n.f.call(e,o),e[o])}},"./node_modules/core-js/modules/_object-gops.js":function(e,o){o.f=Object.getOwnPropertySymbols},"./node_modules/core-js/modules/_object-gpo.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_has.js"),r=t("./node_modules/core-js/modules/_to-object.js"),s=t("./node_modules/core-js/modules/_shared-key.js")("IE_PROTO"),i=Object.prototype;e.exports=Object.getPrototypeOf||function(e){return e=r(e),n(e,s)?e[s]:"function"==typeof e.constructor&&e instanceof e.constructor?e.constructor.prototype:e instanceof Object?i:null}},"./node_modules/core-js/modules/_object-keys-internal.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_has.js"),r=t("./node_modules/core-js/modules/_to-iobject.js"),s=t("./node_modules/core-js/modules/_array-includes.js")(!1),i=t("./node_modules/core-js/modules/_shared-key.js")("IE_PROTO");e.exports=function(e,o){var t,a=r(e),u=0,c=[];for(t in a)t!=i&&n(a,t)&&c.push(t);for(;o.length>u;)n(a,t=o[u++])&&(~s(c,t)||c.push(t));return c}},"./node_modules/core-js/modules/_object-keys.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_object-keys-internal.js"),r=t("./node_modules/core-js/modules/_enum-bug-keys.js");e.exports=Object.keys||function(e){return n(e,r)}},"./node_modules/core-js/modules/_object-pie.js":function(e,o){o.f={}.propertyIsEnumerable},"./node_modules/core-js/modules/_property-desc.js":function(e,o){e.exports=function(e,o){return{enumerable:!(1&e),configurable:!(2&e),writable:!(4&e),value:o}}},"./node_modules/core-js/modules/_redefine-all.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_redefine.js");e.exports=function(e,o,t){for(var r in o)n(e,r,o[r],t);return e}},"./node_modules/core-js/modules/_redefine.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_global.js"),r=t("./node_modules/core-js/modules/_hide.js"),s=t("./node_modules/core-js/modules/_has.js"),i=t("./node_modules/core-js/modules/_uid.js")("src"),a=Function.toString,u=(""+a).split("toString");t("./node_modules/core-js/modules/_core.js").inspectSource=function(e){return a.call(e)},(e.exports=function(e,o,t,a){var c="function"==typeof t;c&&(s(t,"name")||r(t,"name",o)),e[o]!==t&&(c&&(s(t,i)||r(t,i,e[o]?""+e[o]:u.join(String(o)))),e===n?e[o]=t:a?e[o]?e[o]=t:r(e,o,t):(delete e[o],r(e,o,t)))})(Function.prototype,"toString",function(){return"function"==typeof this&&this[i]||a.call(this)})},"./node_modules/core-js/modules/_set-proto.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_is-object.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=function(e,o){if(r(e),!n(o)&&null!==o)throw TypeError(o+": can't set as prototype!")};e.exports={set:Object.setPrototypeOf||("__proto__"in{}?function(e,o,n){try{(n=t("./node_modules/core-js/modules/_ctx.js")(Function.call,t("./node_modules/core-js/modules/_object-gopd.js").f(Object.prototype,"__proto__").set,2))(e,[]),o=!(e instanceof Array)}catch(e){o=!0}return function(e,t){return s(e,t),o?e.__proto__=t:n(e,t),e}}({},!1):void 0),check:s}},"./node_modules/core-js/modules/_set-species.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_global.js"),r=t("./node_modules/core-js/modules/_object-dp.js"),s=t("./node_modules/core-js/modules/_descriptors.js"),i=t("./node_modules/core-js/modules/_wks.js")("species");e.exports=function(e){var o=n[e];s&&o&&!o[i]&&r.f(o,i,{configurable:!0,get:function(){return this}})}},"./node_modules/core-js/modules/_set-to-string-tag.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_object-dp.js").f,r=t("./node_modules/core-js/modules/_has.js"),s=t("./node_modules/core-js/modules/_wks.js")("toStringTag");e.exports=function(e,o,t){e&&!r(e=t?e:e.prototype,s)&&n(e,s,{configurable:!0,value:o})}},"./node_modules/core-js/modules/_shared-key.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_shared.js")("keys"),r=t("./node_modules/core-js/modules/_uid.js");e.exports=function(e){return n[e]||(n[e]=r(e))}},"./node_modules/core-js/modules/_shared.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_global.js"),r="__core-js_shared__",s=n[r]||(n[r]={});e.exports=function(e){return s[e]||(s[e]={})}},"./node_modules/core-js/modules/_to-absolute-index.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_to-integer.js"),r=Math.max,s=Math.min;e.exports=function(e,o){return(e=n(e))<0?r(e+o,0):s(e,o)}},"./node_modules/core-js/modules/_to-integer.js":function(e,o){var t=Math.ceil,n=Math.floor;e.exports=function(e){return isNaN(e=+e)?0:(e>0?n:t)(e)}},"./node_modules/core-js/modules/_to-iobject.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_iobject.js"),r=t("./node_modules/core-js/modules/_defined.js");e.exports=function(e){return n(r(e))}},"./node_modules/core-js/modules/_to-length.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_to-integer.js"),r=Math.min;e.exports=function(e){return e>0?r(n(e),9007199254740991):0}},"./node_modules/core-js/modules/_to-object.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_defined.js");e.exports=function(e){return Object(n(e))}},"./node_modules/core-js/modules/_to-primitive.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_is-object.js");e.exports=function(e,o){if(!n(e))return e;var t,r;if(o&&"function"==typeof(t=e.toString)&&!n(r=t.call(e)))return r;if("function"==typeof(t=e.valueOf)&&!n(r=t.call(e)))return r;if(!o&&"function"==typeof(t=e.toString)&&!n(r=t.call(e)))return r;throw TypeError("Can't convert object to primitive value")}},"./node_modules/core-js/modules/_uid.js":function(e,o){var t=0,n=Math.random();e.exports=function(e){return"Symbol(".concat(void 0===e?"":e,")_",(++t+n).toString(36))}},"./node_modules/core-js/modules/_validate-collection.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_is-object.js");e.exports=function(e,o){if(!n(e)||e._t!==o)throw TypeError("Incompatible receiver, "+o+" required!");return e}},"./node_modules/core-js/modules/_wks.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_shared.js")("wks"),r=t("./node_modules/core-js/modules/_uid.js"),s=t("./node_modules/core-js/modules/_global.js").Symbol,i="function"==typeof s;(e.exports=function(e){return n[e]||(n[e]=i&&s[e]||(i?s:r)("Symbol."+e))}).store=n},"./node_modules/core-js/modules/core.get-iterator-method.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_classof.js"),r=t("./node_modules/core-js/modules/_wks.js")("iterator"),s=t("./node_modules/core-js/modules/_iterators.js");e.exports=t("./node_modules/core-js/modules/_core.js").getIteratorMethod=function(e){if(void 0!=e)return e[r]||e["@@iterator"]||s[n(e)]}},"./node_modules/core-js/modules/es6.map.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_collection-strong.js"),r=t("./node_modules/core-js/modules/_validate-collection.js");e.exports=t("./node_modules/core-js/modules/_collection.js")("Map",function(e){return function(){return e(this,arguments.length>0?arguments[0]:void 0)}},{get:function(e){var o=n.getEntry(r(this,"Map"),e);return o&&o.v},set:function(e,o){return n.def(r(this,"Map"),0===e?0:e,o)}},n,!0)},"./node_modules/core-js/modules/es6.set.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_collection-strong.js"),r=t("./node_modules/core-js/modules/_validate-collection.js");e.exports=t("./node_modules/core-js/modules/_collection.js")("Set",function(e){return function(){return e(this,arguments.length>0?arguments[0]:void 0)}},{add:function(e){return n.def(r(this,"Set"),e=0===e?0:e,e)}},n)},"./node_modules/core-js/modules/es6.weak-map.js":function(e,o,t){"use strict";var n,r=t("./node_modules/core-js/modules/_array-methods.js")(0),s=t("./node_modules/core-js/modules/_redefine.js"),i=t("./node_modules/core-js/modules/_meta.js"),a=t("./node_modules/core-js/modules/_object-assign.js"),u=t("./node_modules/core-js/modules/_collection-weak.js"),c=t("./node_modules/core-js/modules/_is-object.js"),l=t("./node_modules/core-js/modules/_fails.js"),d=t("./node_modules/core-js/modules/_validate-collection.js"),f=i.getWeak,m=Object.isExtensible,p=u.ufstore,_={},h=function(e){return function(){return e(this,arguments.length>0?arguments[0]:void 0)}},j={get:function(e){if(c(e)){var o=f(e);return!0===o?p(d(this,"WeakMap")).get(e):o?o[this._i]:void 0}},set:function(e,o){return u.def(d(this,"WeakMap"),e,o)}},v=e.exports=t("./node_modules/core-js/modules/_collection.js")("WeakMap",h,j,u,!0,!0);l(function(){return 7!=(new v).set((Object.freeze||Object)(_),7).get(_)})&&(a((n=u.getConstructor(h,"WeakMap")).prototype,j),i.NEED=!0,r(["delete","has","get","set"],function(e){var o=v.prototype,t=o[e];s(o,e,function(o,r){if(c(o)&&!m(o)){this._f||(this._f=new n);var s=this._f[e](o,r);return"set"==e?this:s}return t.call(this,o,r)})}))},"./node_modules/core-js/modules/es7.reflect.define-metadata.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_metadata.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=n.key,i=n.set;n.exp({defineMetadata:function(e,o,t,n){i(e,o,r(t),s(n))}})},"./node_modules/core-js/modules/es7.reflect.delete-metadata.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_metadata.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=n.key,i=n.map,a=n.store;n.exp({deleteMetadata:function(e,o){var t=arguments.length<3?void 0:s(arguments[2]),n=i(r(o),t,!1);if(void 0===n||!n.delete(e))return!1;if(n.size)return!0;var u=a.get(o);return u.delete(t),!!u.size||a.delete(o)}})},"./node_modules/core-js/modules/es7.reflect.get-metadata-keys.js":function(e,o,t){var n=t("./node_modules/core-js/modules/es6.set.js"),r=t("./node_modules/core-js/modules/_array-from-iterable.js"),s=t("./node_modules/core-js/modules/_metadata.js"),i=t("./node_modules/core-js/modules/_an-object.js"),a=t("./node_modules/core-js/modules/_object-gpo.js"),u=s.keys,c=s.key,l=function(e,o){var t=u(e,o),s=a(e);if(null===s)return t;var i=l(s,o);return i.length?t.length?r(new n(t.concat(i))):i:t};s.exp({getMetadataKeys:function(e){return l(i(e),arguments.length<2?void 0:c(arguments[1]))}})},"./node_modules/core-js/modules/es7.reflect.get-metadata.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_metadata.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=t("./node_modules/core-js/modules/_object-gpo.js"),i=n.has,a=n.get,u=n.key,c=function(e,o,t){if(i(e,o,t))return a(e,o,t);var n=s(o);return null!==n?c(e,n,t):void 0};n.exp({getMetadata:function(e,o){return c(e,r(o),arguments.length<3?void 0:u(arguments[2]))}})},"./node_modules/core-js/modules/es7.reflect.get-own-metadata-keys.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_metadata.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=n.keys,i=n.key;n.exp({getOwnMetadataKeys:function(e){return s(r(e),arguments.length<2?void 0:i(arguments[1]))}})},"./node_modules/core-js/modules/es7.reflect.get-own-metadata.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_metadata.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=n.get,i=n.key;n.exp({getOwnMetadata:function(e,o){return s(e,r(o),arguments.length<3?void 0:i(arguments[2]))}})},"./node_modules/core-js/modules/es7.reflect.has-metadata.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_metadata.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=t("./node_modules/core-js/modules/_object-gpo.js"),i=n.has,a=n.key,u=function(e,o,t){if(i(e,o,t))return!0;var n=s(o);return null!==n&&u(e,n,t)};n.exp({hasMetadata:function(e,o){return u(e,r(o),arguments.length<3?void 0:a(arguments[2]))}})},"./node_modules/core-js/modules/es7.reflect.has-own-metadata.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_metadata.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=n.has,i=n.key;n.exp({hasOwnMetadata:function(e,o){return s(e,r(o),arguments.length<3?void 0:i(arguments[2]))}})},"./node_modules/core-js/modules/es7.reflect.metadata.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_metadata.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=t("./node_modules/core-js/modules/_a-function.js"),i=n.key,a=n.set;n.exp({metadata:function(e,o){return function(t,n){a(e,o,(void 0!==n?r:s)(t),i(n))}}})},"./node_modules/zone.js/dist/zone.js":function(e,o,t){!function(){"use strict";!function(e){var o=e.performance;function t(e){o&&o.mark&&o.mark(e)}function n(e,t){o&&o.measure&&o.measure(e,t)}t("Zone");var r=!0===e.__zone_symbol__forceDuplicateZoneCheck;if(e.Zone){if(r||"function"!=typeof e.Zone.__symbol__)throw new Error("Zone already loaded.");return e.Zone}var s,i=function(){function o(e,o){this._parent=e,this._name=o?o.name||"unnamed":"<root>",this._properties=o&&o.properties||{},this._zoneDelegate=new u(this,this._parent&&this._parent._zoneDelegate,o)}return o.assertZonePatched=function(){if(e.Promise!==O.ZoneAwarePromise)throw new Error("Zone.js has detected that ZoneAwarePromise `(window|global).Promise` has been overwritten.\nMost likely cause is that a Promise polyfill has been loaded after Zone.js (Polyfilling Promise api is not necessary when zone.js is loaded. If you must load one, do so before loading zone.js.)")},Object.defineProperty(o,"root",{get:function(){for(var e=o.current;e.parent;)e=e.parent;return e},enumerable:!0,configurable:!0}),Object.defineProperty(o,"current",{get:function(){return P.zone},enumerable:!0,configurable:!0}),Object.defineProperty(o,"currentTask",{get:function(){return D},enumerable:!0,configurable:!0}),o.__load_patch=function(s,i){if(O.hasOwnProperty(s)){if(r)throw Error("Already loaded patch: "+s)}else if(!e["__Zone_disable_"+s]){var a="Zone:"+s;t(a),O[s]=i(e,o,S),n(a,a)}},Object.defineProperty(o.prototype,"parent",{get:function(){return this._parent},enumerable:!0,configurable:!0}),Object.defineProperty(o.prototype,"name",{get:function(){return this._name},enumerable:!0,configurable:!0}),o.prototype.get=function(e){var o=this.getZoneWith(e);if(o)return o._properties[e]},o.prototype.getZoneWith=function(e){for(var o=this;o;){if(o._properties.hasOwnProperty(e))return o;o=o._parent}return null},o.prototype.fork=function(e){if(!e)throw new Error("ZoneSpec required!");return this._zoneDelegate.fork(this,e)},o.prototype.wrap=function(e,o){if("function"!=typeof e)throw new Error("Expecting function got: "+e);var t=this._zoneDelegate.intercept(this,e,o),n=this;return function(){return n.runGuarded(t,this,arguments,o)}},o.prototype.run=function(e,o,t,n){P={parent:P,zone:this};try{return this._zoneDelegate.invoke(this,e,o,t,n)}finally{P=P.parent}},o.prototype.runGuarded=function(e,o,t,n){void 0===o&&(o=null),P={parent:P,zone:this};try{try{return this._zoneDelegate.invoke(this,e,o,t,n)}catch(e){if(this._zoneDelegate.handleError(this,e))throw e}}finally{P=P.parent}},o.prototype.runTask=function(e,o,t){if(e.zone!=this)throw new Error("A task can only be run in the zone of creation! (Creation: "+(e.zone||j).name+"; Execution: "+this.name+")");if(e.state!==v||e.type!==x&&e.type!==E){var n=e.state!=b;n&&e._transitionTo(b,y),e.runCount++;var r=D;D=e,P={parent:P,zone:this};try{e.type==E&&e.data&&!e.data.isPeriodic&&(e.cancelFn=void 0);try{return this._zoneDelegate.invokeTask(this,e,o,t)}catch(e){if(this._zoneDelegate.handleError(this,e))throw e}}finally{e.state!==v&&e.state!==w&&(e.type==x||e.data&&e.data.isPeriodic?n&&e._transitionTo(y,b):(e.runCount=0,this._updateTaskCount(e,-1),n&&e._transitionTo(v,b,v))),P=P.parent,D=r}}},o.prototype.scheduleTask=function(e){if(e.zone&&e.zone!==this)for(var o=this;o;){if(o===e.zone)throw Error("can not reschedule task to "+this.name+" which is descendants of the original zone "+e.zone.name);o=o.parent}e._transitionTo(g,v);var t=[];e._zoneDelegates=t,e._zone=this;try{e=this._zoneDelegate.scheduleTask(this,e)}catch(o){throw e._transitionTo(w,g,v),this._zoneDelegate.handleError(this,o),o}return e._zoneDelegates===t&&this._updateTaskCount(e,1),e.state==g&&e._transitionTo(y,g),e},o.prototype.scheduleMicroTask=function(e,o,t,n){return this.scheduleTask(new c(T,e,o,t,n,void 0))},o.prototype.scheduleMacroTask=function(e,o,t,n,r){return this.scheduleTask(new c(E,e,o,t,n,r))},o.prototype.scheduleEventTask=function(e,o,t,n,r){return this.scheduleTask(new c(x,e,o,t,n,r))},o.prototype.cancelTask=function(e){if(e.zone!=this)throw new Error("A task can only be cancelled in the zone of creation! (Creation: "+(e.zone||j).name+"; Execution: "+this.name+")");e._transitionTo(k,y,b);try{this._zoneDelegate.cancelTask(this,e)}catch(o){throw e._transitionTo(w,k),this._zoneDelegate.handleError(this,o),o}return this._updateTaskCount(e,-1),e._transitionTo(v,k),e.runCount=0,e},o.prototype._updateTaskCount=function(e,o){var t=e._zoneDelegates;-1==o&&(e._zoneDelegates=null);for(var n=0;n<t.length;n++)t[n]._updateTaskCount(e.type,o)},o.__symbol__=C,o}(),a={name:"",onHasTask:function(e,o,t,n){return e.hasTask(t,n)},onScheduleTask:function(e,o,t,n){return e.scheduleTask(t,n)},onInvokeTask:function(e,o,t,n,r,s){return e.invokeTask(t,n,r,s)},onCancelTask:function(e,o,t,n){return e.cancelTask(t,n)}},u=function(){function e(e,o,t){this._taskCounts={microTask:0,macroTask:0,eventTask:0},this.zone=e,this._parentDelegate=o,this._forkZS=t&&(t&&t.onFork?t:o._forkZS),this._forkDlgt=t&&(t.onFork?o:o._forkDlgt),this._forkCurrZone=t&&(t.onFork?this.zone:o.zone),this._interceptZS=t&&(t.onIntercept?t:o._interceptZS),this._interceptDlgt=t&&(t.onIntercept?o:o._interceptDlgt),this._interceptCurrZone=t&&(t.onIntercept?this.zone:o.zone),this._invokeZS=t&&(t.onInvoke?t:o._invokeZS),this._invokeDlgt=t&&(t.onInvoke?o:o._invokeDlgt),this._invokeCurrZone=t&&(t.onInvoke?this.zone:o.zone),this._handleErrorZS=t&&(t.onHandleError?t:o._handleErrorZS),this._handleErrorDlgt=t&&(t.onHandleError?o:o._handleErrorDlgt),this._handleErrorCurrZone=t&&(t.onHandleError?this.zone:o.zone),this._scheduleTaskZS=t&&(t.onScheduleTask?t:o._scheduleTaskZS),this._scheduleTaskDlgt=t&&(t.onScheduleTask?o:o._scheduleTaskDlgt),this._scheduleTaskCurrZone=t&&(t.onScheduleTask?this.zone:o.zone),this._invokeTaskZS=t&&(t.onInvokeTask?t:o._invokeTaskZS),this._invokeTaskDlgt=t&&(t.onInvokeTask?o:o._invokeTaskDlgt),this._invokeTaskCurrZone=t&&(t.onInvokeTask?this.zone:o.zone),this._cancelTaskZS=t&&(t.onCancelTask?t:o._cancelTaskZS),this._cancelTaskDlgt=t&&(t.onCancelTask?o:o._cancelTaskDlgt),this._cancelTaskCurrZone=t&&(t.onCancelTask?this.zone:o.zone),this._hasTaskZS=null,this._hasTaskDlgt=null,this._hasTaskDlgtOwner=null,this._hasTaskCurrZone=null;var n=t&&t.onHasTask,r=o&&o._hasTaskZS;(n||r)&&(this._hasTaskZS=n?t:a,this._hasTaskDlgt=o,this._hasTaskDlgtOwner=this,this._hasTaskCurrZone=e,t.onScheduleTask||(this._scheduleTaskZS=a,this._scheduleTaskDlgt=o,this._scheduleTaskCurrZone=this.zone),t.onInvokeTask||(this._invokeTaskZS=a,this._invokeTaskDlgt=o,this._invokeTaskCurrZone=this.zone),t.onCancelTask||(this._cancelTaskZS=a,this._cancelTaskDlgt=o,this._cancelTaskCurrZone=this.zone))}return e.prototype.fork=function(e,o){return this._forkZS?this._forkZS.onFork(this._forkDlgt,this.zone,e,o):new i(e,o)},e.prototype.intercept=function(e,o,t){return this._interceptZS?this._interceptZS.onIntercept(this._interceptDlgt,this._interceptCurrZone,e,o,t):o},e.prototype.invoke=function(e,o,t,n,r){return this._invokeZS?this._invokeZS.onInvoke(this._invokeDlgt,this._invokeCurrZone,e,o,t,n,r):o.apply(t,n)},e.prototype.handleError=function(e,o){return!this._handleErrorZS||this._handleErrorZS.onHandleError(this._handleErrorDlgt,this._handleErrorCurrZone,e,o)},e.prototype.scheduleTask=function(e,o){var t=o;if(this._scheduleTaskZS)this._hasTaskZS&&t._zoneDelegates.push(this._hasTaskDlgtOwner),(t=this._scheduleTaskZS.onScheduleTask(this._scheduleTaskDlgt,this._scheduleTaskCurrZone,e,o))||(t=o);else if(o.scheduleFn)o.scheduleFn(o);else{if(o.type!=T)throw new Error("Task is missing scheduleFn.");_(o)}return t},e.prototype.invokeTask=function(e,o,t,n){return this._invokeTaskZS?this._invokeTaskZS.onInvokeTask(this._invokeTaskDlgt,this._invokeTaskCurrZone,e,o,t,n):o.callback.apply(t,n)},e.prototype.cancelTask=function(e,o){var t;if(this._cancelTaskZS)t=this._cancelTaskZS.onCancelTask(this._cancelTaskDlgt,this._cancelTaskCurrZone,e,o);else{if(!o.cancelFn)throw Error("Task is not cancelable");t=o.cancelFn(o)}return t},e.prototype.hasTask=function(e,o){try{this._hasTaskZS&&this._hasTaskZS.onHasTask(this._hasTaskDlgt,this._hasTaskCurrZone,e,o)}catch(o){this.handleError(e,o)}},e.prototype._updateTaskCount=function(e,o){var t=this._taskCounts,n=t[e],r=t[e]=n+o;if(r<0)throw new Error("More tasks executed then were scheduled.");if(0==n||0==r){var s={microTask:t.microTask>0,macroTask:t.macroTask>0,eventTask:t.eventTask>0,change:e};this.hasTask(this.zone,s)}},e}(),c=function(){function o(t,n,r,s,i,a){this._zone=null,this.runCount=0,this._zoneDelegates=null,this._state="notScheduled",this.type=t,this.source=n,this.data=s,this.scheduleFn=i,this.cancelFn=a,this.callback=r;var u=this;t===x&&s&&s.useG?this.invoke=o.invokeTask:this.invoke=function(){return o.invokeTask.call(e,u,this,arguments)}}return o.invokeTask=function(e,o,t){e||(e=this),z++;try{return e.runCount++,e.zone.runTask(e,o,t)}finally{1==z&&h(),z--}},Object.defineProperty(o.prototype,"zone",{get:function(){return this._zone},enumerable:!0,configurable:!0}),Object.defineProperty(o.prototype,"state",{get:function(){return this._state},enumerable:!0,configurable:!0}),o.prototype.cancelScheduleRequest=function(){this._transitionTo(v,g)},o.prototype._transitionTo=function(e,o,t){if(this._state!==o&&this._state!==t)throw new Error(this.type+" '"+this.source+"': can not transition to '"+e+"', expecting state '"+o+"'"+(t?" or '"+t+"'":"")+", was '"+this._state+"'.");this._state=e,e==v&&(this._zoneDelegates=null)},o.prototype.toString=function(){return this.data&&void 0!==this.data.handleId?this.data.handleId.toString():Object.prototype.toString.call(this)},o.prototype.toJSON=function(){return{type:this.type,state:this.state,source:this.source,zone:this.zone.name,runCount:this.runCount}},o}(),l=C("setTimeout"),d=C("Promise"),f=C("then"),m=[],p=!1;function _(o){if(0===z&&0===m.length)if(s||e[d]&&(s=e[d].resolve(0)),s){var t=s[f];t||(t=s.then),t.call(s,h)}else e[l](h,0);o&&m.push(o)}function h(){if(!p){for(p=!0;m.length;){var e=m;m=[];for(var o=0;o<e.length;o++){var t=e[o];try{t.zone.runTask(t,null,null)}catch(e){S.onUnhandledError(e)}}}S.microtaskDrainDone(),p=!1}}var j={name:"NO ZONE"},v="notScheduled",g="scheduling",y="scheduled",b="running",k="canceling",w="unknown",T="microTask",E="macroTask",x="eventTask",O={},S={symbol:C,currentZoneFrame:function(){return P},onUnhandledError:Z,microtaskDrainDone:Z,scheduleMicroTask:_,showUncaughtError:function(){return!i[C("ignoreConsoleErrorUncaughtError")]},patchEventTarget:function(){return[]},patchOnProperties:Z,patchMethod:function(){return Z},bindArguments:function(){return[]},patchThen:function(){return Z},setNativePromise:function(e){e&&"function"==typeof e.resolve&&(s=e.resolve(0))}},P={parent:null,zone:new i(null,null)},D=null,z=0;function Z(){}function C(e){return"__zone_symbol__"+e}n("Zone","Zone"),e.Zone=i}("undefined"!=typeof window&&window||"undefined"!=typeof self&&self||global);var e=function(e){var o="function"==typeof Symbol&&e[Symbol.iterator],t=0;return o?o.call(e):{next:function(){return e&&t>=e.length&&(e=void 0),{value:e&&e[t++],done:!e}}}};Zone.__load_patch("ZoneAwarePromise",function(o,t,n){var r=Object.getOwnPropertyDescriptor,s=Object.defineProperty,i=n.symbol,a=[],u=i("Promise"),c=i("then"),l="__creationTrace__";n.onUnhandledError=function(e){if(n.showUncaughtError()){var o=e&&e.rejection;o?console.error("Unhandled Promise rejection:",o instanceof Error?o.message:o,"; Zone:",e.zone.name,"; Task:",e.task&&e.task.source,"; Value:",o,o instanceof Error?o.stack:void 0):console.error(e)}},n.microtaskDrainDone=function(){for(;a.length;)for(var e=function(){var e=a.shift();try{e.zone.runGuarded(function(){throw e})}catch(e){f(e)}};a.length;)e()};var d=i("unhandledPromiseRejectionHandler");function f(e){n.onUnhandledError(e);try{var o=t[d];o&&"function"==typeof o&&o.call(this,e)}catch(e){}}function m(e){return e&&e.then}function p(e){return e}function _(e){return M.reject(e)}var h=i("state"),j=i("value"),v=i("finally"),g=i("parentPromiseValue"),y=i("parentPromiseState"),b="Promise.then",k=null,w=!0,T=!1,E=0;function x(e,o){return function(t){try{D(e,o,t)}catch(o){D(e,!1,o)}}}var O=function(){var e=!1;return function(o){return function(){e||(e=!0,o.apply(null,arguments))}}},S="Promise resolved with itself",P=i("currentTaskTrace");function D(e,o,r){var i=O();if(e===r)throw new TypeError(S);if(e[h]===k){var u=null;try{"object"!=typeof r&&"function"!=typeof r||(u=r&&r.then)}catch(o){return i(function(){D(e,!1,o)})(),e}if(o!==T&&r instanceof M&&r.hasOwnProperty(h)&&r.hasOwnProperty(j)&&r[h]!==k)Z(r),D(e,r[h],r[j]);else if(o!==T&&"function"==typeof u)try{u.call(r,i(x(e,o)),i(x(e,!1)))}catch(o){i(function(){D(e,!1,o)})()}else{e[h]=o;var c=e[j];if(e[j]=r,e[v]===v&&o===w&&(e[h]=e[y],e[j]=e[g]),o===T&&r instanceof Error){var d=t.currentTask&&t.currentTask.data&&t.currentTask.data[l];d&&s(r,P,{configurable:!0,enumerable:!1,writable:!0,value:d})}for(var f=0;f<c.length;)C(e,c[f++],c[f++],c[f++],c[f++]);if(0==c.length&&o==T){e[h]=E;try{throw new Error("Uncaught (in promise): "+function(e){if(e&&e.toString===Object.prototype.toString){var o=e.constructor&&e.constructor.name;return(o||"")+": "+JSON.stringify(e)}return e?e.toString():Object.prototype.toString.call(e)}(r)+(r&&r.stack?"\n"+r.stack:""))}catch(o){var m=o;m.rejection=r,m.promise=e,m.zone=t.current,m.task=t.currentTask,a.push(m),n.scheduleMicroTask()}}}}return e}var z=i("rejectionHandledHandler");function Z(e){if(e[h]===E){try{var o=t[z];o&&"function"==typeof o&&o.call(this,{rejection:e[j],promise:e})}catch(e){}e[h]=T;for(var n=0;n<a.length;n++)e===a[n].promise&&a.splice(n,1)}}function C(e,o,t,n,r){Z(e);var s=e[h],i=s?"function"==typeof n?n:p:"function"==typeof r?r:_;o.scheduleMicroTask(b,function(){try{var n=e[j],r=t&&v===t[v];r&&(t[g]=n,t[y]=s);var a=o.run(i,void 0,r&&i!==_&&i!==p?[]:[n]);D(t,!0,a)}catch(e){D(t,!1,e)}},t)}var M=function(){function o(e){if(!(this instanceof o))throw new Error("Must be an instanceof Promise.");this[h]=k,this[j]=[];try{e&&e(x(this,w),x(this,T))}catch(e){D(this,!1,e)}}return o.toString=function(){return"function ZoneAwarePromise() { [native code] }"},o.resolve=function(e){return D(new this(null),w,e)},o.reject=function(e){return D(new this(null),T,e)},o.race=function(o){var t,n,r,s,i=new this(function(e,o){r=e,s=o});function a(e){i&&(i=r(e))}function u(e){i&&(i=s(e))}try{for(var c=e(o),l=c.next();!l.done;l=c.next()){var d=l.value;m(d)||(d=this.resolve(d)),d.then(a,u)}}catch(e){t={error:e}}finally{try{l&&!l.done&&(n=c.return)&&n.call(c)}finally{if(t)throw t.error}}return i},o.all=function(o){var t,n,r,s,i=new this(function(e,o){r=e,s=o}),a=2,u=0,c=[],l=function(e){m(e)||(e=d.resolve(e));var o=u;e.then(function(e){c[o]=e,0==--a&&r(c)},s),a++,u++},d=this;try{for(var f=e(o),p=f.next();!p.done;p=f.next()){var _=p.value;l(_)}}catch(e){t={error:e}}finally{try{p&&!p.done&&(n=f.return)&&n.call(f)}finally{if(t)throw t.error}}return 0==(a-=2)&&r(c),i},o.prototype.then=function(e,o){var n=new this.constructor(null),r=t.current;return this[h]==k?this[j].push(r,n,e,o):C(this,r,n,e,o),n},o.prototype.catch=function(e){return this.then(null,e)},o.prototype.finally=function(e){var o=new this.constructor(null);o[v]=v;var n=t.current;return this[h]==k?this[j].push(n,o,e,e):C(this,n,o,e,e),o},o}();M.resolve=M.resolve,M.reject=M.reject,M.race=M.race,M.all=M.all;var I=o[u]=o.Promise,L=t.__symbol__("ZoneAwarePromise"),R=r(o,"Promise");R&&!R.configurable||(R&&delete R.writable,R&&delete R.value,R||(R={configurable:!0,enumerable:!0}),R.get=function(){return o[L]?o[L]:o[u]},R.set=function(e){e===M?o[L]=e:(o[u]=e,e.prototype[c]||A(e),n.setNativePromise(e))},s(o,"Promise",R)),o.Promise=M;var F=i("thenPatched");function A(e){var o=e.prototype,t=r(o,"then");if(!t||!1!==t.writable&&t.configurable){var n=o.then;o[c]=n,e.prototype.then=function(e,o){var t=this,r=new M(function(e,o){n.call(t,e,o)});return r.then(e,o)},e[F]=!0}}return n.patchThen=A,I&&A(I),Promise[t.__symbol__("uncaughtPromiseErrors")]=a,M}),Zone.__load_patch("fetch",function(e,o,t){var n=e.fetch,r=e.Promise,s=t.symbol("thenPatched"),i=t.symbol("fetchTaskScheduling"),a=t.symbol("fetchTaskAborting");if("function"==typeof n){var u=e.AbortController,c="function"==typeof u,l=null;c&&(e.AbortController=function(){var e=new u,o=e.signal;return o.abortController=e,e},l=t.patchMethod(u.prototype,"abort",function(e){return function(o,t){return o.task?o.task.zone.cancelTask(o.task):e.apply(o,t)}}));var d=function(){};e.fetch=function(){var e=this,u=Array.prototype.slice.call(arguments),f=u.length>1?u[1]:null,m=f&&f.signal;return new Promise(function(f,p){var _=o.current.scheduleMacroTask("fetch",d,u,function(){var a,c=o.current;try{c[i]=!0,a=n.apply(e,u)}catch(e){return void p(e)}finally{c[i]=!1}if(!(a instanceof r)){var l=a.constructor;l[s]||t.patchThen(l)}a.then(function(e){"notScheduled"!==_.state&&_.invoke(),f(e)},function(e){"notScheduled"!==_.state&&_.invoke(),p(e)})},function(){if(c)if(m&&m.abortController&&!m.aborted&&"function"==typeof m.abortController.abort&&l)try{o.current[a]=!0,l.call(m.abortController)}finally{o.current[a]=!1}else p("cancel fetch need a AbortController.signal");else p("No AbortController supported, can not cancel fetch")});m&&m.abortController&&(m.abortController.task=_)})}}});var o=Object.getOwnPropertyDescriptor,t=Object.defineProperty,n=Object.getPrototypeOf,r=Object.create,s=Array.prototype.slice,i="addEventListener",a="removeEventListener",u=Zone.__symbol__(i),c=Zone.__symbol__(a),l="true",d="false",f="__zone_symbol__";function m(e,o){return Zone.current.wrap(e,o)}function p(e,o,t,n,r){return Zone.current.scheduleMacroTask(e,o,t,n,r)}var _=Zone.__symbol__,h="undefined"!=typeof window,j=h?window:void 0,v=h&&j||"object"==typeof self&&self||global,g="removeAttribute",y=[null];function b(e,o){for(var t=e.length-1;t>=0;t--)"function"==typeof e[t]&&(e[t]=m(e[t],o+"_"+t));return e}function k(e){return!e||!1!==e.writable&&!("function"==typeof e.get&&void 0===e.set)}var w="undefined"!=typeof WorkerGlobalScope&&self instanceof WorkerGlobalScope,T=!("nw"in v)&&void 0!==v.process&&"[object process]"==={}.toString.call(v.process),E=!T&&!w&&!(!h||!j.HTMLElement),x=void 0!==v.process&&"[object process]"==={}.toString.call(v.process)&&!w&&!(!h||!j.HTMLElement),O={},S=function(e){if(e=e||v.event){var o=O[e.type];o||(o=O[e.type]=_("ON_PROPERTY"+e.type));var t,n=this||e.target||v,r=n[o];if(E&&n===j&&"error"===e.type){var s=e;!0===(t=r&&r.call(this,s.message,s.filename,s.lineno,s.colno,s.error))&&e.preventDefault()}else void 0==(t=r&&r.apply(this,arguments))||t||e.preventDefault();return t}};function P(e,n,r){var s=o(e,n);if(!s&&r){var i=o(r,n);i&&(s={enumerable:!0,configurable:!0})}if(s&&s.configurable){var a=_("on"+n+"patched");if(!e.hasOwnProperty(a)||!e[a]){delete s.writable,delete s.value;var u=s.get,c=s.set,l=n.substr(2),d=O[l];d||(d=O[l]=_("ON_PROPERTY"+l)),s.set=function(o){var t=this;if(t||e!==v||(t=v),t){var n=t[d];n&&t.removeEventListener(l,S),c&&c.apply(t,y),"function"==typeof o?(t[d]=o,t.addEventListener(l,S,!1)):t[d]=null}},s.get=function(){var o=this;if(o||e!==v||(o=v),!o)return null;var t=o[d];if(t)return t;if(u){var r=u&&u.call(this);if(r)return s.set.call(this,r),"function"==typeof o[g]&&o.removeAttribute(n),r}return null},t(e,n,s),e[a]=!0}}}function D(e,o,t){if(o)for(var n=0;n<o.length;n++)P(e,"on"+o[n],t);else{var r=[];for(var s in e)"on"==s.substr(0,2)&&r.push(s);for(var i=0;i<r.length;i++)P(e,r[i],t)}}var z=_("originalInstance");function Z(e){var o=v[e];if(o){v[_(e)]=o,v[e]=function(){var t=b(arguments,e);switch(t.length){case 0:this[z]=new o;break;case 1:this[z]=new o(t[0]);break;case 2:this[z]=new o(t[0],t[1]);break;case 3:this[z]=new o(t[0],t[1],t[2]);break;case 4:this[z]=new o(t[0],t[1],t[2],t[3]);break;default:throw new Error("Arg list too long.")}},I(v[e],o);var n,r=new o(function(){});for(n in r)"XMLHttpRequest"===e&&"responseBlob"===n||function(o){"function"==typeof r[o]?v[e].prototype[o]=function(){return this[z][o].apply(this[z],arguments)}:t(v[e].prototype,o,{set:function(t){"function"==typeof t?(this[z][o]=m(t,e+"."+o),I(this[z][o],t)):this[z][o]=t},get:function(){return this[z][o]}})}(n);for(n in o)"prototype"!==n&&o.hasOwnProperty(n)&&(v[e][n]=o[n])}}var C=!1;function M(e,t,r){for(var s=e;s&&!s.hasOwnProperty(t);)s=n(s);!s&&e[t]&&(s=e);var i=_(t),a=null;if(s&&!(a=s[i])){a=s[i]=s[t];var u=s&&o(s,t);if(k(u)){var c=r(a,i,t);s[t]=function(){return c(this,arguments)},I(s[t],a),C&&function(e,o){"function"==typeof Object.getOwnPropertySymbols&&Object.getOwnPropertySymbols(e).forEach(function(t){var n=Object.getOwnPropertyDescriptor(e,t);Object.defineProperty(o,t,{get:function(){return e[t]},set:function(o){(!n||n.writable&&"function"==typeof n.set)&&(e[t]=o)},enumerable:!n||n.enumerable,configurable:!n||n.configurable})})}(a,s[t])}}return a}function I(e,o){e[_("OriginalDelegate")]=o}var L=!1,R=!1;Zone.__load_patch("toString",function(e){var o=Function.prototype.toString,t=_("OriginalDelegate"),n=_("Promise"),r=_("Error"),s=function(){if("function"==typeof this){var s=this[t];if(s)return"function"==typeof s?o.apply(this[t],arguments):Object.prototype.toString.call(s);if(this===Promise){var i=e[n];if(i)return o.apply(i,arguments)}if(this===Error){var a=e[r];if(a)return o.apply(a,arguments)}}return o.apply(this,arguments)};s[t]=o,Function.prototype.toString=s;var i=Object.prototype.toString;Object.prototype.toString=function(){return this instanceof Promise?"[object Promise]":i.apply(this,arguments)}});var F=!1;if("undefined"!=typeof window)try{var A=Object.defineProperty({},"passive",{get:function(){F=!0}});window.addEventListener("test",A,A),window.removeEventListener("test",A,A)}catch(e){F=!1}var H={useG:!0},N={},B={},q=/^__zone_symbol__(\w+)(true|false)$/,W="__zone_symbol__propagationStopped";function U(e,o,t){var r=t&&t.add||i,s=t&&t.rm||a,u=t&&t.listeners||"eventListeners",c=t&&t.rmAll||"removeAllListeners",m=_(r),p="."+r+":",h="prependListener",j="."+h+":",v=function(e,o,t){if(!e.isRemoved){var n=e.callback;"object"==typeof n&&n.handleEvent&&(e.callback=function(e){return n.handleEvent(e)},e.originalDelegate=n),e.invoke(e,o,[t]);var r=e.options;if(r&&"object"==typeof r&&r.once){var i=e.originalDelegate?e.originalDelegate:e.callback;o[s].call(o,t.type,i,r)}}},g=function(o){if(o=o||e.event){var t=this||o.target||e,n=t[N[o.type][d]];if(n)if(1===n.length)v(n[0],t,o);else for(var r=n.slice(),s=0;s<r.length&&(!o||!0!==o[W]);s++)v(r[s],t,o)}},y=function(o){if(o=o||e.event){var t=this||o.target||e,n=t[N[o.type][l]];if(n)if(1===n.length)v(n[0],t,o);else for(var r=n.slice(),s=0;s<r.length&&(!o||!0!==o[W]);s++)v(r[s],t,o)}};function b(o,t){if(!o)return!1;var i=!0;t&&void 0!==t.useG&&(i=t.useG);var a=t&&t.vh,v=!0;t&&void 0!==t.chkDup&&(v=t.chkDup);var b=!1;t&&void 0!==t.rt&&(b=t.rt);for(var k=o;k&&!k.hasOwnProperty(r);)k=n(k);if(!k&&o[r]&&(k=o),!k)return!1;if(k[m])return!1;var w,E=t&&t.eventNameToString,x={},O=k[m]=k[r],S=k[_(s)]=k[s],P=k[_(u)]=k[u],D=k[_(c)]=k[c];function z(e){F||"boolean"==typeof x.options||void 0===x.options||null===x.options||(e.options=!!x.options.capture,x.options=e.options)}t&&t.prepend&&(w=k[_(t.prepend)]=k[t.prepend]);var Z=i?function(e){if(!x.isExisting)return z(e),O.call(x.target,x.eventName,x.capture?y:g,x.options)}:function(e){return z(e),O.call(x.target,x.eventName,e.invoke,x.options)},C=i?function(e){if(!e.isRemoved){var o=N[e.eventName],t=void 0;o&&(t=o[e.capture?l:d]);var n=t&&e.target[t];if(n)for(var r=0;r<n.length;r++){var s=n[r];if(s===e){n.splice(r,1),e.isRemoved=!0,0===n.length&&(e.allRemoved=!0,e.target[t]=null);break}}}if(e.allRemoved)return S.call(e.target,e.eventName,e.capture?y:g,e.options)}:function(e){return S.call(e.target,e.eventName,e.invoke,e.options)},M=t&&t.diff?t.diff:function(e,o){var t=typeof o;return"function"===t&&e.callback===o||"object"===t&&e.originalDelegate===o},L=Zone[Zone.__symbol__("BLACK_LISTED_EVENTS")],R=function(o,t,n,r,s,u){return void 0===s&&(s=!1),void 0===u&&(u=!1),function(){var c=this||e,m=arguments[0],p=arguments[1];if(!p)return o.apply(this,arguments);if(T&&"uncaughtException"===m)return o.apply(this,arguments);var _=!1;if("function"!=typeof p){if(!p.handleEvent)return o.apply(this,arguments);_=!0}if(!a||a(o,p,c,arguments)){var h,j=arguments[2];if(L)for(var g=0;g<L.length;g++)if(m===L[g])return o.apply(this,arguments);var y=!1;void 0===j?h=!1:!0===j?h=!0:!1===j?h=!1:(h=!!j&&!!j.capture,y=!!j&&!!j.once);var b,k=Zone.current,w=N[m];if(w)b=w[h?l:d];else{var O=(E?E(m):m)+d,S=(E?E(m):m)+l,P=f+O,D=f+S;N[m]={},N[m][d]=P,N[m][l]=D,b=h?D:P}var z,Z=c[b],C=!1;if(Z){if(C=!0,v)for(var g=0;g<Z.length;g++)if(M(Z[g],p))return}else Z=c[b]=[];var I=c.constructor.name,R=B[I];R&&(z=R[m]),z||(z=I+t+(E?E(m):m)),x.options=j,y&&(x.options.once=!1),x.target=c,x.capture=h,x.eventName=m,x.isExisting=C;var A=i?H:void 0;A&&(A.taskData=x);var q=k.scheduleEventTask(z,p,A,n,r);return x.target=null,A&&(A.taskData=null),y&&(j.once=!0),(F||"boolean"!=typeof q.options)&&(q.options=j),q.target=c,q.capture=h,q.eventName=m,_&&(q.originalDelegate=p),u?Z.unshift(q):Z.push(q),s?c:void 0}}};return k[r]=R(O,p,Z,C,b),w&&(k[h]=R(w,j,function(e){return w.call(x.target,x.eventName,e.invoke,x.options)},C,b,!0)),k[s]=function(){var o,t=this||e,n=arguments[0],r=arguments[2];o=void 0!==r&&(!0===r||!1!==r&&!!r&&!!r.capture);var s=arguments[1];if(!s)return S.apply(this,arguments);if(!a||a(S,s,t,arguments)){var i,u=N[n];u&&(i=u[o?l:d]);var c=i&&t[i];if(c)for(var f=0;f<c.length;f++){var m=c[f];if(M(m,s))return c.splice(f,1),m.isRemoved=!0,0===c.length&&(m.allRemoved=!0,t[i]=null),m.zone.cancelTask(m),b?t:void 0}return S.apply(this,arguments)}},k[u]=function(){for(var o=this||e,t=arguments[0],n=[],r=G(o,E?E(t):t),s=0;s<r.length;s++){var i=r[s],a=i.originalDelegate?i.originalDelegate:i.callback;n.push(a)}return n},k[c]=function(){var o=this||e,t=arguments[0];if(t){var n=N[t];if(n){var r=n[d],i=n[l],a=o[r],u=o[i];if(a)for(var f=a.slice(),m=0;m<f.length;m++){var p=f[m],_=p.originalDelegate?p.originalDelegate:p.callback;this[s].call(this,t,_,p.options)}if(u)for(var f=u.slice(),m=0;m<f.length;m++){var p=f[m],_=p.originalDelegate?p.originalDelegate:p.callback;this[s].call(this,t,_,p.options)}}}else{for(var h=Object.keys(o),m=0;m<h.length;m++){var j=h[m],v=q.exec(j),g=v&&v[1];g&&"removeListener"!==g&&this[c].call(this,g)}this[c].call(this,"removeListener")}if(b)return this},I(k[r],O),I(k[s],S),D&&I(k[c],D),P&&I(k[u],P),!0}for(var k=[],w=0;w<o.length;w++)k[w]=b(o[w],t);return k}function G(e,o){var t=[];for(var n in e){var r=q.exec(n),s=r&&r[1];if(s&&(!o||s===o)){var i=e[n];if(i)for(var a=0;a<i.length;a++)t.push(i[a])}}return t}var X=_("zoneTask");function K(e,o,t,n){var r=null,s=null;t+=n;var i={};function a(o){var t=o.data;return t.args[0]=function(){try{o.invoke.apply(this,arguments)}finally{o.data&&o.data.isPeriodic||("number"==typeof t.handleId?delete i[t.handleId]:t.handleId&&(t.handleId[X]=null))}},t.handleId=r.apply(e,t.args),o}function u(e){return s(e.data.handleId)}r=M(e,o+=n,function(t){return function(r,s){if("function"==typeof s[0]){var c={isPeriodic:"Interval"===n,delay:"Timeout"===n||"Interval"===n?s[1]||0:void 0,args:s},l=p(o,s[0],c,a,u);if(!l)return l;var d=l.data.handleId;return"number"==typeof d?i[d]=l:d&&(d[X]=l),d&&d.ref&&d.unref&&"function"==typeof d.ref&&"function"==typeof d.unref&&(l.ref=d.ref.bind(d),l.unref=d.unref.bind(d)),"number"==typeof d||d?d:l}return t.apply(e,s)}}),s=M(e,t,function(o){return function(t,n){var r,s=n[0];"number"==typeof s?r=i[s]:(r=s&&s[X])||(r=s),r&&"string"==typeof r.type?"notScheduled"!==r.state&&(r.cancelFn&&r.data.isPeriodic||0===r.runCount)&&("number"==typeof s?delete i[s]:s&&(s[X]=null),r.zone.cancelTask(r)):o.apply(e,n)}})}var V=Object[_("defineProperty")]=Object.defineProperty,J=Object[_("getOwnPropertyDescriptor")]=Object.getOwnPropertyDescriptor,Y=Object.create,Q=_("unconfigurables");function $(e,o){return e&&e[Q]&&e[Q][o]}function ee(e,o,t){return Object.isFrozen(t)||(t.configurable=!0),t.configurable||(e[Q]||Object.isFrozen(e)||V(e,Q,{writable:!0,value:{}}),e[Q]&&(e[Q][o]=!0)),t}function oe(e,o,t,n){try{return V(e,o,t)}catch(s){if(!t.configurable)throw s;void 0===n?delete t.configurable:t.configurable=n;try{return V(e,o,t)}catch(n){var r=null;try{r=JSON.stringify(t)}catch(e){r=t.toString()}console.log("Attempting to configure '"+o+"' with descriptor '"+r+"' on object '"+e+"' and got error, giving up: "+n)}}}var te=["absolutedeviceorientation","afterinput","afterprint","appinstalled","beforeinstallprompt","beforeprint","beforeunload","devicelight","devicemotion","deviceorientation","deviceorientationabsolute","deviceproximity","hashchange","languagechange","message","mozbeforepaint","offline","online","paint","pageshow","pagehide","popstate","rejectionhandled","storage","unhandledrejection","unload","userproximity","vrdisplyconnected","vrdisplaydisconnected","vrdisplaypresentchange"],ne=["encrypted","waitingforkey","msneedkey","mozinterruptbegin","mozinterruptend"],re=["load"],se=["blur","error","focus","load","resize","scroll","messageerror"],ie=["bounce","finish","start"],ae=["loadstart","progress","abort","error","load","progress","timeout","loadend","readystatechange"],ue=["upgradeneeded","complete","abort","success","error","blocked","versionchange","close"],ce=["close","error","open","message"],le=["error","message"],de=["abort","animationcancel","animationend","animationiteration","auxclick","beforeinput","blur","cancel","canplay","canplaythrough","change","compositionstart","compositionupdate","compositionend","cuechange","click","close","contextmenu","curechange","dblclick","drag","dragend","dragenter","dragexit","dragleave","dragover","drop","durationchange","emptied","ended","error","focus","focusin","focusout","gotpointercapture","input","invalid","keydown","keypress","keyup","load","loadstart","loadeddata","loadedmetadata","lostpointercapture","mousedown","mouseenter","mouseleave","mousemove","mouseout","mouseover","mouseup","mousewheel","orientationchange","pause","play","playing","pointercancel","pointerdown","pointerenter","pointerleave","pointerlockchange","mozpointerlockchange","webkitpointerlockerchange","pointerlockerror","mozpointerlockerror","webkitpointerlockerror","pointermove","pointout","pointerover","pointerup","progress","ratechange","reset","resize","scroll","seeked","seeking","select","selectionchange","selectstart","show","sort","stalled","submit","suspend","timeupdate","volumechange","touchcancel","touchmove","touchstart","touchend","transitioncancel","transitionend","waiting","wheel"].concat(["webglcontextrestored","webglcontextlost","webglcontextcreationerror"],["autocomplete","autocompleteerror"],["toggle"],["afterscriptexecute","beforescriptexecute","DOMContentLoaded","freeze","fullscreenchange","mozfullscreenchange","webkitfullscreenchange","msfullscreenchange","fullscreenerror","mozfullscreenerror","webkitfullscreenerror","msfullscreenerror","readystatechange","visibilitychange","resume"],te,["beforecopy","beforecut","beforepaste","copy","cut","paste","dragstart","loadend","animationstart","search","transitionrun","transitionstart","webkitanimationend","webkitanimationiteration","webkitanimationstart","webkittransitionend"],["activate","afterupdate","ariarequest","beforeactivate","beforedeactivate","beforeeditfocus","beforeupdate","cellchange","controlselect","dataavailable","datasetchanged","datasetcomplete","errorupdate","filterchange","layoutcomplete","losecapture","move","moveend","movestart","propertychange","resizeend","resizestart","rowenter","rowexit","rowsdelete","rowsinserted","command","compassneedscalibration","deactivate","help","mscontentzoom","msmanipulationstatechanged","msgesturechange","msgesturedoubletap","msgestureend","msgesturehold","msgesturestart","msgesturetap","msgotpointercapture","msinertiastart","mslostpointercapture","mspointercancel","mspointerdown","mspointerenter","mspointerhover","mspointerleave","mspointermove","mspointerout","mspointerover","mspointerup","pointerout","mssitemodejumplistitemremoved","msthumbnailclick","stop","storagecommit"]);function fe(e,o,t,n){if(e){var r=function(e,o,t){if(!t||0===t.length)return o;var n=t.filter(function(o){return o.target===e});if(!n||0===n.length)return o;var r=n[0].ignoreProperties;return o.filter(function(e){return-1===r.indexOf(e)})}(e,o,t);D(e,r,n)}}function me(e,u){if(!T||x){var c="undefined"!=typeof WebSocket;if(function(){if((E||x)&&!o(HTMLElement.prototype,"onclick")&&"undefined"!=typeof Element){var e=o(Element.prototype,"onclick");if(e&&!e.configurable)return!1}var n="onreadystatechange",r=XMLHttpRequest.prototype,s=o(r,n);if(s){t(r,n,{enumerable:!0,configurable:!0,get:function(){return!0}});var i=new XMLHttpRequest,a=!!i.onreadystatechange;return t(r,n,s||{}),a}var u=_("fake");t(r,n,{enumerable:!0,configurable:!0,get:function(){return this[u]},set:function(e){this[u]=e}});var i=new XMLHttpRequest,c=function(){};i.onreadystatechange=c;var a=i[u]===c;return i.onreadystatechange=null,a}()){var l=u.__Zone_ignore_on_properties;if(E){var d=window,f=function(){try{var e=j.navigator.userAgent;if(-1!==e.indexOf("MSIE ")||-1!==e.indexOf("Trident/"))return!0}catch(e){}return!1}?[{target:d,ignoreProperties:["error"]}]:[];fe(d,de.concat(["messageerror"]),l?l.concat(f):l,n(d)),fe(Document.prototype,de,l),void 0!==d.SVGElement&&fe(d.SVGElement.prototype,de,l),fe(Element.prototype,de,l),fe(HTMLElement.prototype,de,l),fe(HTMLMediaElement.prototype,ne,l),fe(HTMLFrameSetElement.prototype,te.concat(se),l),fe(HTMLBodyElement.prototype,te.concat(se),l),fe(HTMLFrameElement.prototype,re,l),fe(HTMLIFrameElement.prototype,re,l);var p=d.HTMLMarqueeElement;p&&fe(p.prototype,ie,l);var h=d.Worker;h&&fe(h.prototype,le,l)}fe(XMLHttpRequest.prototype,ae,l);var v=u.XMLHttpRequestEventTarget;v&&fe(v&&v.prototype,ae,l),"undefined"!=typeof IDBIndex&&(fe(IDBIndex.prototype,ue,l),fe(IDBRequest.prototype,ue,l),fe(IDBOpenDBRequest.prototype,ue,l),fe(IDBDatabase.prototype,ue,l),fe(IDBTransaction.prototype,ue,l),fe(IDBCursor.prototype,ue,l)),c&&fe(WebSocket.prototype,ce,l)}else!function(){for(var e=function(e){var o=de[e],t="on"+o;self.addEventListener(o,function(e){var o,n,r=e.target;for(n=r?r.constructor.name+"."+t:"unknown."+t;r;)r[t]&&!r[t][pe]&&((o=m(r[t],n))[pe]=r[t],r[t]=o),r=r.parentElement},!0)},o=0;o<de.length;o++)e(o)}(),Z("XMLHttpRequest"),c&&function(e,t){var n=t.WebSocket;t.EventTarget||U(t,[n.prototype]),t.WebSocket=function(e,t){var u,c,l=arguments.length>1?new n(e,t):new n(e),d=o(l,"onmessage");return d&&!1===d.configurable?(u=r(l),c=l,[i,a,"send","close"].forEach(function(e){u[e]=function(){var o=s.call(arguments);if(e===i||e===a){var t=o.length>0?o[0]:void 0;if(t){var n=Zone.__symbol__("ON_PROPERTY"+t);l[n]=u[n]}}return l[e].apply(l,o)}})):u=l,D(u,["close","error","message","open"],c),u};var u=t.WebSocket;for(var c in n)u[c]=n[c]}(0,u)}}var pe=_("unbound");function _e(e,o){var t="Anchor,Area,Audio,BR,Base,BaseFont,Body,Button,Canvas,Content,DList,Directory,Div,Embed,FieldSet,Font,Form,Frame,FrameSet,HR,Head,Heading,Html,IFrame,Image,Input,Keygen,LI,Label,Legend,Link,Map,Marquee,Media,Menu,Meta,Meter,Mod,OList,Object,OptGroup,Option,Output,Paragraph,Pre,Progress,Quote,Script,Select,Source,Span,Style,TableCaption,TableCell,TableCol,Table,TableRow,TableSection,TextArea,Title,Track,UList,Unknown,Video",n="ApplicationCache,EventSource,FileReader,InputMethodContext,MediaController,MessagePort,Node,Performance,SVGElementInstance,SharedWorker,TextTrack,TextTrackCue,TextTrackList,WebKitNamedFlow,Window,Worker,WorkerGlobalScope,XMLHttpRequest,XMLHttpRequestEventTarget,XMLHttpRequestUpload,IDBRequest,IDBOpenDBRequest,IDBDatabase,IDBTransaction,IDBCursor,DBIndex,WebSocket".split(","),r=[],s=e.wtf,i=t.split(",");s?r=i.map(function(e){return"HTML"+e+"Element"}).concat(n):e.EventTarget?r.push("EventTarget"):r=n;for(var a=e.__Zone_disable_IE_check||!1,u=e.__Zone_enable_cross_context_check||!1,c=function(){if(L)return R;L=!0;try{var e=j.navigator.userAgent;return-1===e.indexOf("MSIE ")&&-1===e.indexOf("Trident/")&&-1===e.indexOf("Edge/")||(R=!0),R}catch(e){}}(),m="[object FunctionWrapper]",p="function __BROWSERTOOLS_CONSOLE_SAFEFUNC() { [native code] }",_=0;_<de.length;_++){var h=de[_],v=h+d,g=h+l,y=f+v,b=f+g;N[h]={},N[h][d]=y,N[h][l]=b}for(var _=0;_<t.length;_++)for(var k=i[_],w=B[k]={},T=0;T<de.length;T++){var h=de[T];w[h]=k+".addEventListener:"+h}for(var E=[],_=0;_<r.length;_++){var x=e[r[_]];E.push(x&&x.prototype)}return U(e,E,{vh:function(e,o,t,n){if(!a&&c)if(u)try{var r=o.toString();if(r===m||r==p)return e.apply(t,n),!1}catch(o){return e.apply(t,n),!1}else{var r=o.toString();if(r===m||r==p)return e.apply(t,n),!1}else if(u)try{o.toString()}catch(o){return e.apply(t,n),!1}return!0}}),o.patchEventTarget=U,!0}function he(e,o){var t,n;t=o,(n=e.Event)&&n.prototype&&t.patchMethod(n.prototype,"stopImmediatePropagation",function(e){return function(o,t){o[W]=!0,e&&e.apply(o,t)}})}function je(e,t,n,r){var s=Zone.__symbol__(n);if(!e[s]){var i=e[s]=e[n];e[n]=function(s,a,u){return a&&a.prototype&&r.forEach(function(e){var r,s,i,u,c=t+"."+n+"::"+e,l=a.prototype;if(l.hasOwnProperty(e)){var d=o(l,e);d&&d.value?(d.value=m(d.value,c),r=a.prototype,s=e,u=(i=d).configurable,i=ee(r,s,i),oe(r,s,i,u)):l[e]=m(l[e],c)}else l[e]&&(l[e]=m(l[e],c))}),i.call(e,s,a,u)},I(e[n],i)}}Zone.__load_patch("util",function(e,o,t){t.patchOnProperties=D,t.patchMethod=M,t.bindArguments=b}),Zone.__load_patch("timers",function(e){K(e,"set","clear","Timeout"),K(e,"set","clear","Interval"),K(e,"set","clear","Immediate")}),Zone.__load_patch("requestAnimationFrame",function(e){K(e,"request","cancel","AnimationFrame"),K(e,"mozRequest","mozCancel","AnimationFrame"),K(e,"webkitRequest","webkitCancel","AnimationFrame")}),Zone.__load_patch("blocking",function(e,o){for(var t=["alert","prompt","confirm"],n=0;n<t.length;n++){var r=t[n];M(e,r,function(t,n,r){return function(n,s){return o.current.run(t,e,s,r)}})}}),Zone.__load_patch("EventTarget",function(e,o,t){var n=o.__symbol__("BLACK_LISTED_EVENTS");e[n]&&(o[n]=e[n]),he(e,t),_e(e,t);var r=e.XMLHttpRequestEventTarget;r&&r.prototype&&t.patchEventTarget(e,[r.prototype]),Z("MutationObserver"),Z("WebKitMutationObserver"),Z("IntersectionObserver"),Z("FileReader")}),Zone.__load_patch("on_property",function(e,o,t){me(0,e),Object.defineProperty=function(e,o,t){if($(e,o))throw new TypeError("Cannot assign to read only property '"+o+"' of "+e);var n=t.configurable;return"prototype"!==o&&(t=ee(e,o,t)),oe(e,o,t,n)},Object.defineProperties=function(e,o){return Object.keys(o).forEach(function(t){Object.defineProperty(e,t,o[t])}),e},Object.create=function(e,o){return"object"!=typeof o||Object.isFrozen(o)||Object.keys(o).forEach(function(t){o[t]=ee(e,t,o[t])}),Y(e,o)},Object.getOwnPropertyDescriptor=function(e,o){var t=J(e,o);return t&&$(e,o)&&(t.configurable=!1),t}}),Zone.__load_patch("customElements",function(e,o,t){!function(e){(E||x)&&"registerElement"in e.document&&je(document,"Document","registerElement",["createdCallback","attachedCallback","detachedCallback","attributeChangedCallback"])}(e),function(e){(E||x)&&"customElements"in e&&je(e.customElements,"customElements","define",["connectedCallback","disconnectedCallback","adoptedCallback","attributeChangedCallback"])}(e)}),Zone.__load_patch("canvas",function(e){var o=e.HTMLCanvasElement;void 0!==o&&o.prototype&&o.prototype.toBlob&&function(e,o,t){var n=null;function r(e){var o=e.data;return o.args[o.cbIdx]=function(){e.invoke.apply(this,arguments)},n.apply(o.target,o.args),e}n=M(e,o,function(e){return function(o,n){var s=t(o,n);return s.cbIdx>=0&&"function"==typeof n[s.cbIdx]?p(s.name,n[s.cbIdx],s,r):e.apply(o,n)}})}(o.prototype,"toBlob",function(e,o){return{name:"HTMLCanvasElement.toBlob",target:e,cbIdx:0,args:o}})}),Zone.__load_patch("XHR",function(e,o){!function(e){var l=XMLHttpRequest.prototype,d=l[u],f=l[c];if(!d){var m=e.XMLHttpRequestEventTarget;if(m){var h=m.prototype;d=h[u],f=h[c]}}var j="readystatechange",v="scheduled";function g(e){var o=e.data,n=o.target;n[s]=!1,n[a]=!1;var i=n[r];d||(d=n[u],f=n[c]),i&&f.call(n,j,i);var l=n[r]=function(){if(n.readyState===n.DONE)if(!o.aborted&&n[s]&&e.state===v){var t=n.__zone_symbol__loadfalse;if(t&&t.length>0){var r=e.invoke;e.invoke=function(){for(var t=n.__zone_symbol__loadfalse,s=0;s<t.length;s++)t[s]===e&&t.splice(s,1);o.aborted||e.state!==v||r.call(e)},t.push(e)}else e.invoke()}else o.aborted||!1!==n[s]||(n[a]=!0)};d.call(n,j,l);var m=n[t];return m||(n[t]=e),E.apply(n,o.args),n[s]=!0,e}function y(){}function b(e){var o=e.data;return o.aborted=!0,x.apply(o.target,o.args)}var k=M(l,"open",function(){return function(e,o){return e[n]=0==o[2],e[i]=o[1],k.apply(e,o)}}),w=_("fetchTaskAborting"),T=_("fetchTaskScheduling"),E=M(l,"send",function(){return function(e,t){if(!0===o.current[T])return E.apply(e,t);if(e[n])return E.apply(e,t);var r={target:e,url:e[i],isPeriodic:!1,args:t,aborted:!1},s=p("XMLHttpRequest.send",y,r,g,b);e&&!0===e[a]&&!r.aborted&&s.state===v&&s.invoke()}}),x=M(l,"abort",function(){return function(e,n){var r=e[t];if(r&&"string"==typeof r.type){if(null==r.cancelFn||r.data&&r.data.aborted)return;r.zone.cancelTask(r)}else if(!0===o.current[w])return x.apply(e,n)}})}(e);var t=_("xhrTask"),n=_("xhrSync"),r=_("xhrListener"),s=_("xhrScheduled"),i=_("xhrURL"),a=_("xhrErrorBeforeScheduled")}),Zone.__load_patch("geolocation",function(e){e.navigator&&e.navigator.geolocation&&function(e,t){for(var n=e.constructor.name,r=function(r){var s=t[r],i=e[s];if(i){var a=o(e,s);if(!k(a))return"continue";e[s]=function(e){var o=function(){return e.apply(this,b(arguments,n+"."+s))};return I(o,e),o}(i)}},s=0;s<t.length;s++)r(s)}(e.navigator.geolocation,["getCurrentPosition","watchPosition"])}),Zone.__load_patch("PromiseRejectionEvent",function(e,o){function t(o){return function(t){var n=G(e,o);n.forEach(function(n){var r=e.PromiseRejectionEvent;if(r){var s=new r(o,{promise:t.promise,reason:t.rejection});n.invoke(s)}})}}e.PromiseRejectionEvent&&(o[_("unhandledPromiseRejectionHandler")]=t("unhandledrejection"),o[_("rejectionHandledHandler")]=t("rejectionhandled"))})}()},"./src/polyfills.ts":function(e,o,t){"use strict";t.r(o);t("./node_modules/zone.js/dist/zone.js")},1:function(e,o,t){t("./src/polyfills.ts"),e.exports=t("./node_modules/@angular-devkit/build-angular/src/angular-cli-files/models/jit-polyfills.js")}},[[1,"runtime"]]]);
</script><script type="text/javascript">
(window.webpackJsonp=window.webpackJsonp||[]).push([["styles"],{"./node_modules/@angular-devkit/build-angular/src/angular-cli-files/plugins/raw-css-loader.js!./node_modules/postcss-loader/src/index.js?!./node_modules/sass-loader/lib/loader.js?!./src/styles.scss":function(b,t){b.exports=[[b.i,'/**\n * Applies styles for users in high contrast mode. Note that this only applies\n * to Microsoft browsers. Chrome can be included by checking for the `html[hc]`\n * attribute, however Chrome handles high contrast differently.\n * @param target Which kind of high contrast setting to target. Defaults to `active`, can be\n *    `white-on-black` or `black-on-white`.\n */\n/* Theme for the ripple elements.*/\n/* stylelint-disable material/no-prefixes */\n/* stylelint-enable */\n.mat-badge-content {\n  font-weight: 600;\n  font-size: 12px;\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-badge-small .mat-badge-content {\n  font-size: 6px; }\n.mat-badge-large .mat-badge-content {\n  font-size: 24px; }\n.mat-h1, .mat-headline, .mat-typography h1 {\n  font: 400 24px/32px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 16px; }\n.mat-h2, .mat-title, .mat-typography h2 {\n  font: 500 20px/32px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 16px; }\n.mat-h3, .mat-subheading-2, .mat-typography h3 {\n  font: 400 16px/28px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 16px; }\n.mat-h4, .mat-subheading-1, .mat-typography h4 {\n  font: 400 15px/24px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 16px; }\n.mat-h5, .mat-typography h5 {\n  font: 400 11.62px/20px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 12px; }\n.mat-h6, .mat-typography h6 {\n  font: 400 9.38px/20px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 12px; }\n.mat-body-strong, .mat-body-2 {\n  font: 500 14px/24px Roboto, "Helvetica Neue", sans-serif; }\n.mat-body, .mat-body-1, .mat-typography {\n  font: 400 14px/20px Roboto, "Helvetica Neue", sans-serif; }\n.mat-body p, .mat-body-1 p, .mat-typography p {\n    margin: 0 0 12px; }\n.mat-small, .mat-caption {\n  font: 400 12px/20px Roboto, "Helvetica Neue", sans-serif; }\n.mat-display-4, .mat-typography .mat-display-4 {\n  font: 300 112px/112px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 56px;\n  letter-spacing: -0.05em; }\n.mat-display-3, .mat-typography .mat-display-3 {\n  font: 400 56px/56px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 64px;\n  letter-spacing: -0.02em; }\n.mat-display-2, .mat-typography .mat-display-2 {\n  font: 400 45px/48px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 64px;\n  letter-spacing: -0.005em; }\n.mat-display-1, .mat-typography .mat-display-1 {\n  font: 400 34px/40px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 64px; }\n.mat-bottom-sheet-container {\n  font: 400 14px/20px Roboto, "Helvetica Neue", sans-serif; }\n.mat-button, .mat-raised-button, .mat-icon-button, .mat-stroked-button,\n.mat-flat-button, .mat-fab, .mat-mini-fab {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 14px;\n  font-weight: 500; }\n.mat-button-toggle {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-card {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-card-title {\n  font-size: 24px;\n  font-weight: 500; }\n.mat-card-header .mat-card-title {\n  font-size: 20px; }\n.mat-card-subtitle,\n.mat-card-content {\n  font-size: 14px; }\n.mat-checkbox {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-checkbox-layout .mat-checkbox-label {\n  line-height: 24px; }\n.mat-chip {\n  font-size: 14px;\n  font-weight: 500; }\n.mat-chip .mat-chip-trailing-icon.mat-icon,\n  .mat-chip .mat-chip-remove.mat-icon {\n    font-size: 18px; }\n.mat-table {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-header-cell {\n  font-size: 12px;\n  font-weight: 500; }\n.mat-cell, .mat-footer-cell {\n  font-size: 14px; }\n.mat-calendar {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-calendar-body {\n  font-size: 13px; }\n.mat-calendar-body-label,\n.mat-calendar-period-button {\n  font-size: 14px;\n  font-weight: 500; }\n.mat-calendar-table-header th {\n  font-size: 11px;\n  font-weight: 400; }\n.mat-dialog-title {\n  font: 500 20px/32px Roboto, "Helvetica Neue", sans-serif; }\n.mat-expansion-panel-header {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 15px;\n  font-weight: 400; }\n.mat-expansion-panel-content {\n  font: 400 14px/20px Roboto, "Helvetica Neue", sans-serif; }\n.mat-form-field {\n  font-size: inherit;\n  font-weight: 400;\n  line-height: 1.125;\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-form-field-wrapper {\n  padding-bottom: 1.34375em; }\n.mat-form-field-prefix .mat-icon,\n.mat-form-field-suffix .mat-icon {\n  font-size: 150%;\n  line-height: 1.125; }\n.mat-form-field-prefix .mat-icon-button,\n.mat-form-field-suffix .mat-icon-button {\n  height: 1.5em;\n  width: 1.5em; }\n.mat-form-field-prefix .mat-icon-button .mat-icon,\n  .mat-form-field-suffix .mat-icon-button .mat-icon {\n    height: 1.125em;\n    line-height: 1.125; }\n.mat-form-field-infix {\n  padding: 0.5em 0;\n  border-top: 0.84375em solid transparent; }\n.mat-form-field-can-float.mat-form-field-should-float .mat-form-field-label,\n.mat-form-field-can-float .mat-input-server:focus + .mat-form-field-label-wrapper .mat-form-field-label {\n  -webkit-transform: translateY(-1.34375em) scale(0.75);\n          transform: translateY(-1.34375em) scale(0.75);\n  width: 133.33333333%; }\n.mat-form-field-can-float .mat-input-server[label]:not(:label-shown) + .mat-form-field-label-wrapper\n.mat-form-field-label {\n  -webkit-transform: translateY(-1.34374em) scale(0.75);\n          transform: translateY(-1.34374em) scale(0.75);\n  width: 133.33334333%; }\n.mat-form-field-label-wrapper {\n  top: -0.84375em;\n  padding-top: 0.84375em; }\n.mat-form-field-label {\n  top: 1.34375em; }\n.mat-form-field-underline {\n  bottom: 1.34375em; }\n.mat-form-field-subscript-wrapper {\n  font-size: 75%;\n  margin-top: 0.66666667em;\n  top: calc(100% - 1.79166667em); }\n.mat-form-field-appearance-legacy .mat-form-field-wrapper {\n  padding-bottom: 1.25em; }\n.mat-form-field-appearance-legacy .mat-form-field-infix {\n  padding: 0.4375em 0; }\n.mat-form-field-appearance-legacy.mat-form-field-can-float.mat-form-field-should-float .mat-form-field-label,\n.mat-form-field-appearance-legacy.mat-form-field-can-float .mat-input-server:focus + .mat-form-field-label-wrapper .mat-form-field-label {\n  -webkit-transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.001px);\n          transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.001px);\n  -ms-transform: translateY(-1.28125em) scale(0.75);\n  width: 133.33333333%; }\n.mat-form-field-appearance-legacy.mat-form-field-can-float .mat-form-field-autofill-control:-webkit-autofill + .mat-form-field-label-wrapper\n.mat-form-field-label {\n  -webkit-transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.00101px);\n          transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.00101px);\n  -ms-transform: translateY(-1.28124em) scale(0.75);\n  width: 133.33334333%; }\n.mat-form-field-appearance-legacy.mat-form-field-can-float .mat-input-server[label]:not(:label-shown) + .mat-form-field-label-wrapper\n.mat-form-field-label {\n  -webkit-transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.00102px);\n          transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.00102px);\n  -ms-transform: translateY(-1.28123em) scale(0.75);\n  width: 133.33335333%; }\n.mat-form-field-appearance-legacy .mat-form-field-label {\n  top: 1.28125em; }\n.mat-form-field-appearance-legacy .mat-form-field-underline {\n  bottom: 1.25em; }\n.mat-form-field-appearance-legacy .mat-form-field-subscript-wrapper {\n  margin-top: 0.54166667em;\n  top: calc(100% - 1.66666667em); }\n@media print {\n  .mat-form-field-appearance-legacy.mat-form-field-can-float.mat-form-field-should-float .mat-form-field-label,\n  .mat-form-field-appearance-legacy.mat-form-field-can-float .mat-input-server:focus + .mat-form-field-label-wrapper .mat-form-field-label {\n    -webkit-transform: translateY(-1.28122em) scale(0.75);\n            transform: translateY(-1.28122em) scale(0.75); }\n  .mat-form-field-appearance-legacy.mat-form-field-can-float .mat-form-field-autofill-control:-webkit-autofill + .mat-form-field-label-wrapper\n.mat-form-field-label {\n    -webkit-transform: translateY(-1.28121em) scale(0.75);\n            transform: translateY(-1.28121em) scale(0.75); }\n  .mat-form-field-appearance-legacy.mat-form-field-can-float .mat-input-server[label]:not(:label-shown) + .mat-form-field-label-wrapper\n.mat-form-field-label {\n    -webkit-transform: translateY(-1.2812em) scale(0.75);\n            transform: translateY(-1.2812em) scale(0.75); } }\n.mat-form-field-appearance-fill .mat-form-field-infix {\n  padding: 0.25em 0 0.75em 0; }\n.mat-form-field-appearance-fill .mat-form-field-label {\n  top: 1.09375em;\n  margin-top: -0.5em; }\n.mat-form-field-appearance-fill.mat-form-field-can-float.mat-form-field-should-float .mat-form-field-label,\n.mat-form-field-appearance-fill.mat-form-field-can-float .mat-input-server:focus + .mat-form-field-label-wrapper .mat-form-field-label {\n  -webkit-transform: translateY(-0.59375em) scale(0.75);\n          transform: translateY(-0.59375em) scale(0.75);\n  width: 133.33333333%; }\n.mat-form-field-appearance-fill.mat-form-field-can-float .mat-input-server[label]:not(:label-shown) + .mat-form-field-label-wrapper\n.mat-form-field-label {\n  -webkit-transform: translateY(-0.59374em) scale(0.75);\n          transform: translateY(-0.59374em) scale(0.75);\n  width: 133.33334333%; }\n.mat-form-field-appearance-outline .mat-form-field-infix {\n  padding: 1em 0 1em 0; }\n.mat-form-field-appearance-outline .mat-form-field-label {\n  top: 1.84375em;\n  margin-top: -0.25em; }\n.mat-form-field-appearance-outline.mat-form-field-can-float.mat-form-field-should-float .mat-form-field-label,\n.mat-form-field-appearance-outline.mat-form-field-can-float .mat-input-server:focus + .mat-form-field-label-wrapper .mat-form-field-label {\n  -webkit-transform: translateY(-1.59375em) scale(0.75);\n          transform: translateY(-1.59375em) scale(0.75);\n  width: 133.33333333%; }\n.mat-form-field-appearance-outline.mat-form-field-can-float .mat-input-server[label]:not(:label-shown) + .mat-form-field-label-wrapper\n.mat-form-field-label {\n  -webkit-transform: translateY(-1.59374em) scale(0.75);\n          transform: translateY(-1.59374em) scale(0.75);\n  width: 133.33334333%; }\n.mat-grid-tile-header,\n.mat-grid-tile-footer {\n  font-size: 14px; }\n.mat-grid-tile-header .mat-line,\n  .mat-grid-tile-footer .mat-line {\n    white-space: nowrap;\n    overflow: hidden;\n    text-overflow: ellipsis;\n    display: block;\n    box-sizing: border-box; }\n.mat-grid-tile-header .mat-line:nth-child(n+2),\n    .mat-grid-tile-footer .mat-line:nth-child(n+2) {\n      font-size: 12px; }\ninput.mat-input-element {\n  margin-top: -0.0625em; }\n.mat-menu-item {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 14px;\n  font-weight: 400; }\n.mat-paginator,\n.mat-paginator-page-size .mat-select-trigger {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 12px; }\n.mat-radio-button {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-select {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-select-trigger {\n  height: 1.125em; }\n.mat-slide-toggle-content {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-slider-thumb-label-text {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 12px;\n  font-weight: 500; }\n.mat-stepper-vertical, .mat-stepper-horizontal {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-step-label {\n  font-size: 14px;\n  font-weight: 400; }\n.mat-step-sub-label-error {\n  font-weight: normal; }\n.mat-step-label-error {\n  font-size: 14px; }\n.mat-step-label-selected {\n  font-size: 14px;\n  font-weight: 500; }\n.mat-tab-group {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-tab-label, .mat-tab-link {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 14px;\n  font-weight: 500; }\n.mat-toolbar,\n.mat-toolbar h1,\n.mat-toolbar h2,\n.mat-toolbar h3,\n.mat-toolbar h4,\n.mat-toolbar h5,\n.mat-toolbar h6 {\n  font: 500 20px/32px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0; }\n.mat-tooltip {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 10px;\n  padding-top: 6px;\n  padding-bottom: 6px; }\n.mat-tooltip-handset {\n  font-size: 14px;\n  padding-top: 8px;\n  padding-bottom: 8px; }\n.mat-list-item {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-list-option {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-list-base .mat-list-item {\n  font-size: 16px; }\n.mat-list-base .mat-list-item .mat-line {\n    white-space: nowrap;\n    overflow: hidden;\n    text-overflow: ellipsis;\n    display: block;\n    box-sizing: border-box; }\n.mat-list-base .mat-list-item .mat-line:nth-child(n+2) {\n      font-size: 14px; }\n.mat-list-base .mat-list-option {\n  font-size: 16px; }\n.mat-list-base .mat-list-option .mat-line {\n    white-space: nowrap;\n    overflow: hidden;\n    text-overflow: ellipsis;\n    display: block;\n    box-sizing: border-box; }\n.mat-list-base .mat-list-option .mat-line:nth-child(n+2) {\n      font-size: 14px; }\n.mat-list-base .mat-subheader {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 14px;\n  font-weight: 500; }\n.mat-list-base[dense] .mat-list-item {\n  font-size: 12px; }\n.mat-list-base[dense] .mat-list-item .mat-line {\n    white-space: nowrap;\n    overflow: hidden;\n    text-overflow: ellipsis;\n    display: block;\n    box-sizing: border-box; }\n.mat-list-base[dense] .mat-list-item .mat-line:nth-child(n+2) {\n      font-size: 12px; }\n.mat-list-base[dense] .mat-list-option {\n  font-size: 12px; }\n.mat-list-base[dense] .mat-list-option .mat-line {\n    white-space: nowrap;\n    overflow: hidden;\n    text-overflow: ellipsis;\n    display: block;\n    box-sizing: border-box; }\n.mat-list-base[dense] .mat-list-option .mat-line:nth-child(n+2) {\n      font-size: 12px; }\n.mat-list-base[dense] .mat-subheader {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 12px;\n  font-weight: 500; }\n.mat-option {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 16px; }\n.mat-optgroup-label {\n  font: 500 14px/24px Roboto, "Helvetica Neue", sans-serif; }\n.mat-simple-snackbar {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 14px; }\n.mat-simple-snackbar-action {\n  line-height: 1;\n  font-family: inherit;\n  font-size: inherit;\n  font-weight: 500; }\n.mat-tree {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-tree-node,\n.mat-nested-tree-node {\n  font-weight: 400;\n  font-size: 14px; }\n.mat-ripple {\n  overflow: hidden;\n  position: relative; }\n.mat-ripple.mat-ripple-unbounded {\n  overflow: visible; }\n.mat-ripple-element {\n  position: absolute;\n  border-radius: 50%;\n  pointer-events: none;\n  transition: opacity, -webkit-transform 0ms cubic-bezier(0, 0, 0.2, 1);\n  transition: opacity, transform 0ms cubic-bezier(0, 0, 0.2, 1);\n  transition: opacity, transform 0ms cubic-bezier(0, 0, 0.2, 1), -webkit-transform 0ms cubic-bezier(0, 0, 0.2, 1);\n  -webkit-transform: scale(0);\n          transform: scale(0); }\n@media (-ms-high-contrast: active) {\n    .mat-ripple-element {\n      display: none; } }\n.cdk-visually-hidden {\n  border: 0;\n  clip: rect(0 0 0 0);\n  height: 1px;\n  margin: -1px;\n  overflow: hidden;\n  padding: 0;\n  position: absolute;\n  width: 1px;\n  outline: 0;\n  -webkit-appearance: none;\n  -moz-appearance: none; }\n.cdk-overlay-container, .cdk-global-overlay-wrapper {\n  pointer-events: none;\n  top: 0;\n  left: 0;\n  height: 100%;\n  width: 100%; }\n.cdk-overlay-container {\n  position: fixed;\n  z-index: 1000; }\n.cdk-overlay-container:empty {\n    display: none; }\n.cdk-global-overlay-wrapper {\n  display: flex;\n  position: absolute;\n  z-index: 1000; }\n.cdk-overlay-pane {\n  position: absolute;\n  pointer-events: auto;\n  box-sizing: border-box;\n  z-index: 1000;\n  display: flex;\n  max-width: 100%;\n  max-height: 100%; }\n.cdk-overlay-backdrop {\n  position: absolute;\n  top: 0;\n  bottom: 0;\n  left: 0;\n  right: 0;\n  z-index: 1000;\n  pointer-events: auto;\n  -webkit-tap-highlight-color: transparent;\n  transition: opacity 400ms cubic-bezier(0.25, 0.8, 0.25, 1);\n  opacity: 0; }\n.cdk-overlay-backdrop.cdk-overlay-backdrop-showing {\n    opacity: 1; }\n@media screen and (-ms-high-contrast: active) {\n      .cdk-overlay-backdrop.cdk-overlay-backdrop-showing {\n        opacity: 0.6; } }\n.cdk-overlay-dark-backdrop {\n  background: rgba(0, 0, 0, 0.32); }\n.cdk-overlay-transparent-backdrop, .cdk-overlay-transparent-backdrop.cdk-overlay-backdrop-showing {\n  opacity: 0; }\n.cdk-overlay-connected-position-bounding-box {\n  position: absolute;\n  z-index: 1000;\n  display: flex;\n  flex-direction: column;\n  min-width: 1px;\n  min-height: 1px; }\n.cdk-global-scrollblock {\n  position: fixed;\n  width: 100%;\n  overflow-y: scroll; }\n@-webkit-keyframes cdk-text-field-autofill-start {\n  /*!*/ }\n@keyframes cdk-text-field-autofill-start {\n  /*!*/ }\n@-webkit-keyframes cdk-text-field-autofill-end {\n  /*!*/ }\n@keyframes cdk-text-field-autofill-end {\n  /*!*/ }\n.cdk-text-field-autofill-monitored:-webkit-autofill {\n  -webkit-animation-name: cdk-text-field-autofill-start;\n          animation-name: cdk-text-field-autofill-start; }\n.cdk-text-field-autofill-monitored:not(:-webkit-autofill) {\n  -webkit-animation-name: cdk-text-field-autofill-end;\n          animation-name: cdk-text-field-autofill-end; }\ntextarea.cdk-textarea-autosize {\n  resize: none; }\ntextarea.cdk-textarea-autosize-measuring {\n  height: auto !important;\n  overflow: hidden !important;\n  padding: 2px 0 !important;\n  box-sizing: content-box !important; }\n.mat-ripple-element {\n  background-color: rgba(0, 0, 0, 0.1); }\n.mat-option {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-option:hover:not(.mat-option-disabled), .mat-option:focus:not(.mat-option-disabled) {\n    background: rgba(0, 0, 0, 0.04); }\n.mat-option.mat-selected:not(.mat-option-multiple):not(.mat-option-disabled) {\n    background: rgba(0, 0, 0, 0.04); }\n.mat-option.mat-active {\n    background: rgba(0, 0, 0, 0.04);\n    color: rgba(0, 0, 0, 0.87); }\n.mat-option.mat-option-disabled {\n    color: rgba(0, 0, 0, 0.38); }\n.mat-primary .mat-option.mat-selected:not(.mat-option-disabled) {\n  color: #2a7bb8; }\n.mat-accent .mat-option.mat-selected:not(.mat-option-disabled) {\n  color: #81c6fa; }\n.mat-warn .mat-option.mat-selected:not(.mat-option-disabled) {\n  color: #f44336; }\n.mat-optgroup-label {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-optgroup-disabled .mat-optgroup-label {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-pseudo-checkbox {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-pseudo-checkbox::after {\n    color: #fafafa; }\n.mat-pseudo-checkbox-checked,\n.mat-pseudo-checkbox-indeterminate,\n.mat-accent .mat-pseudo-checkbox-checked,\n.mat-accent .mat-pseudo-checkbox-indeterminate {\n  background: #81c6fa; }\n.mat-primary .mat-pseudo-checkbox-checked,\n.mat-primary .mat-pseudo-checkbox-indeterminate {\n  background: #2a7bb8; }\n.mat-warn .mat-pseudo-checkbox-checked,\n.mat-warn .mat-pseudo-checkbox-indeterminate {\n  background: #f44336; }\n.mat-pseudo-checkbox-checked.mat-pseudo-checkbox-disabled,\n.mat-pseudo-checkbox-indeterminate.mat-pseudo-checkbox-disabled {\n  background: #b0b0b0; }\n.mat-elevation-z0 {\n  box-shadow: 0px 0px 0px 0px rgba(0, 0, 0, 0.2), 0px 0px 0px 0px rgba(0, 0, 0, 0.14), 0px 0px 0px 0px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z1 {\n  box-shadow: 0px 2px 1px -1px rgba(0, 0, 0, 0.2), 0px 1px 1px 0px rgba(0, 0, 0, 0.14), 0px 1px 3px 0px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z2 {\n  box-shadow: 0px 3px 1px -2px rgba(0, 0, 0, 0.2), 0px 2px 2px 0px rgba(0, 0, 0, 0.14), 0px 1px 5px 0px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z3 {\n  box-shadow: 0px 3px 3px -2px rgba(0, 0, 0, 0.2), 0px 3px 4px 0px rgba(0, 0, 0, 0.14), 0px 1px 8px 0px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z4 {\n  box-shadow: 0px 2px 4px -1px rgba(0, 0, 0, 0.2), 0px 4px 5px 0px rgba(0, 0, 0, 0.14), 0px 1px 10px 0px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z5 {\n  box-shadow: 0px 3px 5px -1px rgba(0, 0, 0, 0.2), 0px 5px 8px 0px rgba(0, 0, 0, 0.14), 0px 1px 14px 0px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z6 {\n  box-shadow: 0px 3px 5px -1px rgba(0, 0, 0, 0.2), 0px 6px 10px 0px rgba(0, 0, 0, 0.14), 0px 1px 18px 0px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z7 {\n  box-shadow: 0px 4px 5px -2px rgba(0, 0, 0, 0.2), 0px 7px 10px 1px rgba(0, 0, 0, 0.14), 0px 2px 16px 1px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z8 {\n  box-shadow: 0px 5px 5px -3px rgba(0, 0, 0, 0.2), 0px 8px 10px 1px rgba(0, 0, 0, 0.14), 0px 3px 14px 2px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z9 {\n  box-shadow: 0px 5px 6px -3px rgba(0, 0, 0, 0.2), 0px 9px 12px 1px rgba(0, 0, 0, 0.14), 0px 3px 16px 2px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z10 {\n  box-shadow: 0px 6px 6px -3px rgba(0, 0, 0, 0.2), 0px 10px 14px 1px rgba(0, 0, 0, 0.14), 0px 4px 18px 3px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z11 {\n  box-shadow: 0px 6px 7px -4px rgba(0, 0, 0, 0.2), 0px 11px 15px 1px rgba(0, 0, 0, 0.14), 0px 4px 20px 3px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z12 {\n  box-shadow: 0px 7px 8px -4px rgba(0, 0, 0, 0.2), 0px 12px 17px 2px rgba(0, 0, 0, 0.14), 0px 5px 22px 4px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z13 {\n  box-shadow: 0px 7px 8px -4px rgba(0, 0, 0, 0.2), 0px 13px 19px 2px rgba(0, 0, 0, 0.14), 0px 5px 24px 4px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z14 {\n  box-shadow: 0px 7px 9px -4px rgba(0, 0, 0, 0.2), 0px 14px 21px 2px rgba(0, 0, 0, 0.14), 0px 5px 26px 4px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z15 {\n  box-shadow: 0px 8px 9px -5px rgba(0, 0, 0, 0.2), 0px 15px 22px 2px rgba(0, 0, 0, 0.14), 0px 6px 28px 5px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z16 {\n  box-shadow: 0px 8px 10px -5px rgba(0, 0, 0, 0.2), 0px 16px 24px 2px rgba(0, 0, 0, 0.14), 0px 6px 30px 5px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z17 {\n  box-shadow: 0px 8px 11px -5px rgba(0, 0, 0, 0.2), 0px 17px 26px 2px rgba(0, 0, 0, 0.14), 0px 6px 32px 5px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z18 {\n  box-shadow: 0px 9px 11px -5px rgba(0, 0, 0, 0.2), 0px 18px 28px 2px rgba(0, 0, 0, 0.14), 0px 7px 34px 6px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z19 {\n  box-shadow: 0px 9px 12px -6px rgba(0, 0, 0, 0.2), 0px 19px 29px 2px rgba(0, 0, 0, 0.14), 0px 7px 36px 6px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z20 {\n  box-shadow: 0px 10px 13px -6px rgba(0, 0, 0, 0.2), 0px 20px 31px 3px rgba(0, 0, 0, 0.14), 0px 8px 38px 7px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z21 {\n  box-shadow: 0px 10px 13px -6px rgba(0, 0, 0, 0.2), 0px 21px 33px 3px rgba(0, 0, 0, 0.14), 0px 8px 40px 7px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z22 {\n  box-shadow: 0px 10px 14px -6px rgba(0, 0, 0, 0.2), 0px 22px 35px 3px rgba(0, 0, 0, 0.14), 0px 8px 42px 7px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z23 {\n  box-shadow: 0px 11px 14px -7px rgba(0, 0, 0, 0.2), 0px 23px 36px 3px rgba(0, 0, 0, 0.14), 0px 9px 44px 8px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z24 {\n  box-shadow: 0px 11px 15px -7px rgba(0, 0, 0, 0.2), 0px 24px 38px 3px rgba(0, 0, 0, 0.14), 0px 9px 46px 8px rgba(0, 0, 0, 0.12); }\n.mat-app-background {\n  background-color: #fafafa;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-theme-loaded-marker {\n  display: none; }\n.mat-autocomplete-panel {\n  background: white;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-autocomplete-panel:not([class*=\'mat-elevation-z\']) {\n    box-shadow: 0px 2px 4px -1px rgba(0, 0, 0, 0.2), 0px 4px 5px 0px rgba(0, 0, 0, 0.14), 0px 1px 10px 0px rgba(0, 0, 0, 0.12); }\n.mat-autocomplete-panel .mat-option.mat-selected:not(.mat-active):not(:hover) {\n    background: white; }\n.mat-autocomplete-panel .mat-option.mat-selected:not(.mat-active):not(:hover):not(.mat-option-disabled) {\n      color: rgba(0, 0, 0, 0.87); }\n.mat-badge-content {\n  color: white;\n  background: #2a7bb8; }\n@media (-ms-high-contrast: active) {\n    .mat-badge-content {\n      outline: solid 1px;\n      border-radius: 0; } }\n.mat-badge-accent .mat-badge-content {\n  background: #81c6fa;\n  color: black; }\n.mat-badge-warn .mat-badge-content {\n  color: white;\n  background: #f44336; }\n.mat-badge {\n  position: relative; }\n.mat-badge-hidden .mat-badge-content {\n  display: none; }\n.mat-badge-disabled .mat-badge-content {\n  background: #b9b9b9;\n  color: rgba(0, 0, 0, 0.38); }\n.mat-badge-content {\n  position: absolute;\n  text-align: center;\n  display: inline-block;\n  border-radius: 50%;\n  transition: -webkit-transform 200ms ease-in-out;\n  transition: transform 200ms ease-in-out;\n  transition: transform 200ms ease-in-out, -webkit-transform 200ms ease-in-out;\n  -webkit-transform: scale(0.6);\n          transform: scale(0.6);\n  overflow: hidden;\n  white-space: nowrap;\n  text-overflow: ellipsis;\n  pointer-events: none; }\n.mat-badge-content.mat-badge-active {\n  -webkit-transform: none;\n          transform: none; }\n.mat-badge-small .mat-badge-content {\n  width: 16px;\n  height: 16px;\n  line-height: 16px; }\n.mat-badge-small.mat-badge-above .mat-badge-content {\n  top: -8px; }\n.mat-badge-small.mat-badge-below .mat-badge-content {\n  bottom: -8px; }\n.mat-badge-small.mat-badge-before .mat-badge-content {\n  left: -16px; }\n[dir=\'rtl\'] .mat-badge-small.mat-badge-before .mat-badge-content {\n  left: auto;\n  right: -16px; }\n.mat-badge-small.mat-badge-after .mat-badge-content {\n  right: -16px; }\n[dir=\'rtl\'] .mat-badge-small.mat-badge-after .mat-badge-content {\n  right: auto;\n  left: -16px; }\n.mat-badge-small.mat-badge-overlap.mat-badge-before .mat-badge-content {\n  left: -8px; }\n[dir=\'rtl\'] .mat-badge-small.mat-badge-overlap.mat-badge-before .mat-badge-content {\n  left: auto;\n  right: -8px; }\n.mat-badge-small.mat-badge-overlap.mat-badge-after .mat-badge-content {\n  right: -8px; }\n[dir=\'rtl\'] .mat-badge-small.mat-badge-overlap.mat-badge-after .mat-badge-content {\n  right: auto;\n  left: -8px; }\n.mat-badge-medium .mat-badge-content {\n  width: 22px;\n  height: 22px;\n  line-height: 22px; }\n.mat-badge-medium.mat-badge-above .mat-badge-content {\n  top: -11px; }\n.mat-badge-medium.mat-badge-below .mat-badge-content {\n  bottom: -11px; }\n.mat-badge-medium.mat-badge-before .mat-badge-content {\n  left: -22px; }\n[dir=\'rtl\'] .mat-badge-medium.mat-badge-before .mat-badge-content {\n  left: auto;\n  right: -22px; }\n.mat-badge-medium.mat-badge-after .mat-badge-content {\n  right: -22px; }\n[dir=\'rtl\'] .mat-badge-medium.mat-badge-after .mat-badge-content {\n  right: auto;\n  left: -22px; }\n.mat-badge-medium.mat-badge-overlap.mat-badge-before .mat-badge-content {\n  left: -11px; }\n[dir=\'rtl\'] .mat-badge-medium.mat-badge-overlap.mat-badge-before .mat-badge-content {\n  left: auto;\n  right: -11px; }\n.mat-badge-medium.mat-badge-overlap.mat-badge-after .mat-badge-content {\n  right: -11px; }\n[dir=\'rtl\'] .mat-badge-medium.mat-badge-overlap.mat-badge-after .mat-badge-content {\n  right: auto;\n  left: -11px; }\n.mat-badge-large .mat-badge-content {\n  width: 28px;\n  height: 28px;\n  line-height: 28px; }\n.mat-badge-large.mat-badge-above .mat-badge-content {\n  top: -14px; }\n.mat-badge-large.mat-badge-below .mat-badge-content {\n  bottom: -14px; }\n.mat-badge-large.mat-badge-before .mat-badge-content {\n  left: -28px; }\n[dir=\'rtl\'] .mat-badge-large.mat-badge-before .mat-badge-content {\n  left: auto;\n  right: -28px; }\n.mat-badge-large.mat-badge-after .mat-badge-content {\n  right: -28px; }\n[dir=\'rtl\'] .mat-badge-large.mat-badge-after .mat-badge-content {\n  right: auto;\n  left: -28px; }\n.mat-badge-large.mat-badge-overlap.mat-badge-before .mat-badge-content {\n  left: -14px; }\n[dir=\'rtl\'] .mat-badge-large.mat-badge-overlap.mat-badge-before .mat-badge-content {\n  left: auto;\n  right: -14px; }\n.mat-badge-large.mat-badge-overlap.mat-badge-after .mat-badge-content {\n  right: -14px; }\n[dir=\'rtl\'] .mat-badge-large.mat-badge-overlap.mat-badge-after .mat-badge-content {\n  right: auto;\n  left: -14px; }\n.mat-bottom-sheet-container {\n  box-shadow: 0px 8px 10px -5px rgba(0, 0, 0, 0.2), 0px 16px 24px 2px rgba(0, 0, 0, 0.14), 0px 6px 30px 5px rgba(0, 0, 0, 0.12);\n  background: white;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-button, .mat-icon-button, .mat-stroked-button {\n  color: inherit;\n  background: transparent; }\n.mat-button.mat-primary, .mat-icon-button.mat-primary, .mat-stroked-button.mat-primary {\n    color: #2a7bb8; }\n.mat-button.mat-accent, .mat-icon-button.mat-accent, .mat-stroked-button.mat-accent {\n    color: #81c6fa; }\n.mat-button.mat-warn, .mat-icon-button.mat-warn, .mat-stroked-button.mat-warn {\n    color: #f44336; }\n.mat-button.mat-primary[disabled], .mat-button.mat-accent[disabled], .mat-button.mat-warn[disabled], .mat-button[disabled][disabled], .mat-icon-button.mat-primary[disabled], .mat-icon-button.mat-accent[disabled], .mat-icon-button.mat-warn[disabled], .mat-icon-button[disabled][disabled], .mat-stroked-button.mat-primary[disabled], .mat-stroked-button.mat-accent[disabled], .mat-stroked-button.mat-warn[disabled], .mat-stroked-button[disabled][disabled] {\n    color: rgba(0, 0, 0, 0.26); }\n.mat-button.mat-primary .mat-button-focus-overlay, .mat-icon-button.mat-primary .mat-button-focus-overlay, .mat-stroked-button.mat-primary .mat-button-focus-overlay {\n    background-color: #2a7bb8; }\n.mat-button.mat-accent .mat-button-focus-overlay, .mat-icon-button.mat-accent .mat-button-focus-overlay, .mat-stroked-button.mat-accent .mat-button-focus-overlay {\n    background-color: #81c6fa; }\n.mat-button.mat-warn .mat-button-focus-overlay, .mat-icon-button.mat-warn .mat-button-focus-overlay, .mat-stroked-button.mat-warn .mat-button-focus-overlay {\n    background-color: #f44336; }\n.mat-button[disabled] .mat-button-focus-overlay, .mat-icon-button[disabled] .mat-button-focus-overlay, .mat-stroked-button[disabled] .mat-button-focus-overlay {\n    background-color: transparent; }\n.mat-button .mat-ripple-element, .mat-icon-button .mat-ripple-element, .mat-stroked-button .mat-ripple-element {\n    opacity: 0.1;\n    background-color: currentColor; }\n.mat-button-focus-overlay {\n  background: black; }\n.mat-stroked-button:not([disabled]) {\n  border-color: rgba(0, 0, 0, 0.12); }\n.mat-flat-button, .mat-raised-button, .mat-fab, .mat-mini-fab {\n  color: rgba(0, 0, 0, 0.87);\n  background-color: white; }\n.mat-flat-button.mat-primary, .mat-raised-button.mat-primary, .mat-fab.mat-primary, .mat-mini-fab.mat-primary {\n    color: white; }\n.mat-flat-button.mat-accent, .mat-raised-button.mat-accent, .mat-fab.mat-accent, .mat-mini-fab.mat-accent {\n    color: black; }\n.mat-flat-button.mat-warn, .mat-raised-button.mat-warn, .mat-fab.mat-warn, .mat-mini-fab.mat-warn {\n    color: white; }\n.mat-flat-button.mat-primary[disabled], .mat-flat-button.mat-accent[disabled], .mat-flat-button.mat-warn[disabled], .mat-flat-button[disabled][disabled], .mat-raised-button.mat-primary[disabled], .mat-raised-button.mat-accent[disabled], .mat-raised-button.mat-warn[disabled], .mat-raised-button[disabled][disabled], .mat-fab.mat-primary[disabled], .mat-fab.mat-accent[disabled], .mat-fab.mat-warn[disabled], .mat-fab[disabled][disabled], .mat-mini-fab.mat-primary[disabled], .mat-mini-fab.mat-accent[disabled], .mat-mini-fab.mat-warn[disabled], .mat-mini-fab[disabled][disabled] {\n    color: rgba(0, 0, 0, 0.26); }\n.mat-flat-button.mat-primary, .mat-raised-button.mat-primary, .mat-fab.mat-primary, .mat-mini-fab.mat-primary {\n    background-color: #2a7bb8; }\n.mat-flat-button.mat-accent, .mat-raised-button.mat-accent, .mat-fab.mat-accent, .mat-mini-fab.mat-accent {\n    background-color: #81c6fa; }\n.mat-flat-button.mat-warn, .mat-raised-button.mat-warn, .mat-fab.mat-warn, .mat-mini-fab.mat-warn {\n    background-color: #f44336; }\n.mat-flat-button.mat-primary[disabled], .mat-flat-button.mat-accent[disabled], .mat-flat-button.mat-warn[disabled], .mat-flat-button[disabled][disabled], .mat-raised-button.mat-primary[disabled], .mat-raised-button.mat-accent[disabled], .mat-raised-button.mat-warn[disabled], .mat-raised-button[disabled][disabled], .mat-fab.mat-primary[disabled], .mat-fab.mat-accent[disabled], .mat-fab.mat-warn[disabled], .mat-fab[disabled][disabled], .mat-mini-fab.mat-primary[disabled], .mat-mini-fab.mat-accent[disabled], .mat-mini-fab.mat-warn[disabled], .mat-mini-fab[disabled][disabled] {\n    background-color: rgba(0, 0, 0, 0.12); }\n.mat-flat-button.mat-primary .mat-ripple-element, .mat-raised-button.mat-primary .mat-ripple-element, .mat-fab.mat-primary .mat-ripple-element, .mat-mini-fab.mat-primary .mat-ripple-element {\n    background-color: rgba(255, 255, 255, 0.1); }\n.mat-flat-button.mat-accent .mat-ripple-element, .mat-raised-button.mat-accent .mat-ripple-element, .mat-fab.mat-accent .mat-ripple-element, .mat-mini-fab.mat-accent .mat-ripple-element {\n    background-color: rgba(0, 0, 0, 0.1); }\n.mat-flat-button.mat-warn .mat-ripple-element, .mat-raised-button.mat-warn .mat-ripple-element, .mat-fab.mat-warn .mat-ripple-element, .mat-mini-fab.mat-warn .mat-ripple-element {\n    background-color: rgba(255, 255, 255, 0.1); }\n.mat-stroked-button:not([class*=\'mat-elevation-z\']), .mat-flat-button:not([class*=\'mat-elevation-z\']) {\n  box-shadow: 0px 0px 0px 0px rgba(0, 0, 0, 0.2), 0px 0px 0px 0px rgba(0, 0, 0, 0.14), 0px 0px 0px 0px rgba(0, 0, 0, 0.12); }\n.mat-raised-button:not([class*=\'mat-elevation-z\']) {\n  box-shadow: 0px 3px 1px -2px rgba(0, 0, 0, 0.2), 0px 2px 2px 0px rgba(0, 0, 0, 0.14), 0px 1px 5px 0px rgba(0, 0, 0, 0.12); }\n.mat-raised-button:not([disabled]):active:not([class*=\'mat-elevation-z\']) {\n  box-shadow: 0px 5px 5px -3px rgba(0, 0, 0, 0.2), 0px 8px 10px 1px rgba(0, 0, 0, 0.14), 0px 3px 14px 2px rgba(0, 0, 0, 0.12); }\n.mat-raised-button[disabled]:not([class*=\'mat-elevation-z\']) {\n  box-shadow: 0px 0px 0px 0px rgba(0, 0, 0, 0.2), 0px 0px 0px 0px rgba(0, 0, 0, 0.14), 0px 0px 0px 0px rgba(0, 0, 0, 0.12); }\n.mat-fab:not([class*=\'mat-elevation-z\']), .mat-mini-fab:not([class*=\'mat-elevation-z\']) {\n  box-shadow: 0px 3px 5px -1px rgba(0, 0, 0, 0.2), 0px 6px 10px 0px rgba(0, 0, 0, 0.14), 0px 1px 18px 0px rgba(0, 0, 0, 0.12); }\n.mat-fab:not([disabled]):active:not([class*=\'mat-elevation-z\']), .mat-mini-fab:not([disabled]):active:not([class*=\'mat-elevation-z\']) {\n  box-shadow: 0px 7px 8px -4px rgba(0, 0, 0, 0.2), 0px 12px 17px 2px rgba(0, 0, 0, 0.14), 0px 5px 22px 4px rgba(0, 0, 0, 0.12); }\n.mat-fab[disabled]:not([class*=\'mat-elevation-z\']), .mat-mini-fab[disabled]:not([class*=\'mat-elevation-z\']) {\n  box-shadow: 0px 0px 0px 0px rgba(0, 0, 0, 0.2), 0px 0px 0px 0px rgba(0, 0, 0, 0.14), 0px 0px 0px 0px rgba(0, 0, 0, 0.12); }\n.mat-button-toggle-standalone,\n.mat-button-toggle-group {\n  box-shadow: 0px 3px 1px -2px rgba(0, 0, 0, 0.2), 0px 2px 2px 0px rgba(0, 0, 0, 0.14), 0px 1px 5px 0px rgba(0, 0, 0, 0.12); }\n.mat-button-toggle-standalone.mat-button-toggle-appearance-standard,\n.mat-button-toggle-group-appearance-standard {\n  box-shadow: none; }\n.mat-button-toggle {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-button-toggle .mat-button-toggle-focus-overlay {\n    background-color: rgba(0, 0, 0, 0.12); }\n.mat-button-toggle-appearance-standard {\n  color: rgba(0, 0, 0, 0.87);\n  background: white; }\n.mat-button-toggle-appearance-standard .mat-button-toggle-focus-overlay {\n    background-color: black; }\n.mat-button-toggle-group-appearance-standard .mat-button-toggle + .mat-button-toggle {\n  border-left: solid 1px rgba(0, 0, 0, 0.12); }\n[dir=\'rtl\'] .mat-button-toggle-group-appearance-standard .mat-button-toggle + .mat-button-toggle {\n  border-left: none;\n  border-right: solid 1px rgba(0, 0, 0, 0.12); }\n.mat-button-toggle-group-appearance-standard.mat-button-toggle-vertical .mat-button-toggle + .mat-button-toggle {\n  border-left: none;\n  border-right: none;\n  border-top: solid 1px rgba(0, 0, 0, 0.12); }\n.mat-button-toggle-checked {\n  background-color: #e0e0e0;\n  color: rgba(0, 0, 0, 0.54); }\n.mat-button-toggle-checked.mat-button-toggle-appearance-standard {\n    color: rgba(0, 0, 0, 0.87); }\n.mat-button-toggle-disabled {\n  color: rgba(0, 0, 0, 0.26);\n  background-color: #eeeeee; }\n.mat-button-toggle-disabled.mat-button-toggle-appearance-standard {\n    background: white; }\n.mat-button-toggle-disabled.mat-button-toggle-checked {\n    background-color: #bdbdbd; }\n.mat-button-toggle-standalone.mat-button-toggle-appearance-standard,\n.mat-button-toggle-group-appearance-standard {\n  border: solid 1px rgba(0, 0, 0, 0.12); }\n.mat-card {\n  background: white;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-card:not([class*=\'mat-elevation-z\']) {\n    box-shadow: 0px 2px 1px -1px rgba(0, 0, 0, 0.2), 0px 1px 1px 0px rgba(0, 0, 0, 0.14), 0px 1px 3px 0px rgba(0, 0, 0, 0.12); }\n.mat-card.mat-card-flat:not([class*=\'mat-elevation-z\']) {\n    box-shadow: 0px 0px 0px 0px rgba(0, 0, 0, 0.2), 0px 0px 0px 0px rgba(0, 0, 0, 0.14), 0px 0px 0px 0px rgba(0, 0, 0, 0.12); }\n.mat-card-subtitle {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-checkbox-frame {\n  border-color: rgba(0, 0, 0, 0.54); }\n.mat-checkbox-checkmark {\n  fill: #fafafa; }\n.mat-checkbox-checkmark-path {\n  stroke: #fafafa !important; }\n@media (-ms-high-contrast: black-on-white) {\n    .mat-checkbox-checkmark-path {\n      stroke: #000 !important; } }\n.mat-checkbox-mixedmark {\n  background-color: #fafafa; }\n.mat-checkbox-indeterminate.mat-primary .mat-checkbox-background, .mat-checkbox-checked.mat-primary .mat-checkbox-background {\n  background-color: #2a7bb8; }\n.mat-checkbox-indeterminate.mat-accent .mat-checkbox-background, .mat-checkbox-checked.mat-accent .mat-checkbox-background {\n  background-color: #81c6fa; }\n.mat-checkbox-indeterminate.mat-warn .mat-checkbox-background, .mat-checkbox-checked.mat-warn .mat-checkbox-background {\n  background-color: #f44336; }\n.mat-checkbox-disabled.mat-checkbox-checked .mat-checkbox-background, .mat-checkbox-disabled.mat-checkbox-indeterminate .mat-checkbox-background {\n  background-color: #b0b0b0; }\n.mat-checkbox-disabled:not(.mat-checkbox-checked) .mat-checkbox-frame {\n  border-color: #b0b0b0; }\n.mat-checkbox-disabled .mat-checkbox-label {\n  color: rgba(0, 0, 0, 0.54); }\n@media (-ms-high-contrast: active) {\n  .mat-checkbox-disabled {\n    opacity: 0.5; } }\n@media (-ms-high-contrast: active) {\n  .mat-checkbox-background {\n    background: none; } }\n.mat-checkbox:not(.mat-checkbox-disabled).mat-primary .mat-checkbox-ripple .mat-ripple-element {\n  background-color: #2a7bb8; }\n.mat-checkbox:not(.mat-checkbox-disabled).mat-accent .mat-checkbox-ripple .mat-ripple-element {\n  background-color: #81c6fa; }\n.mat-checkbox:not(.mat-checkbox-disabled).mat-warn .mat-checkbox-ripple .mat-ripple-element {\n  background-color: #f44336; }\n.mat-chip.mat-standard-chip {\n  background-color: #e0e0e0;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-chip.mat-standard-chip .mat-chip-remove {\n    color: rgba(0, 0, 0, 0.87);\n    opacity: 0.4; }\n.mat-chip.mat-standard-chip:not(.mat-chip-disabled):active {\n    box-shadow: 0px 3px 3px -2px rgba(0, 0, 0, 0.2), 0px 3px 4px 0px rgba(0, 0, 0, 0.14), 0px 1px 8px 0px rgba(0, 0, 0, 0.12); }\n.mat-chip.mat-standard-chip:not(.mat-chip-disabled) .mat-chip-remove:hover {\n    opacity: 0.54; }\n.mat-chip.mat-standard-chip.mat-chip-disabled {\n    opacity: 0.4; }\n.mat-chip.mat-standard-chip::after {\n    background: black; }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-primary {\n  background-color: #2a7bb8;\n  color: white; }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-primary .mat-chip-remove {\n    color: white;\n    opacity: 0.4; }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-primary .mat-ripple-element {\n    background: rgba(255, 255, 255, 0.1); }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-warn {\n  background-color: #f44336;\n  color: white; }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-warn .mat-chip-remove {\n    color: white;\n    opacity: 0.4; }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-warn .mat-ripple-element {\n    background: rgba(255, 255, 255, 0.1); }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-accent {\n  background-color: #81c6fa;\n  color: black; }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-accent .mat-chip-remove {\n    color: black;\n    opacity: 0.4; }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-accent .mat-ripple-element {\n    background: rgba(0, 0, 0, 0.1); }\n.mat-table {\n  background: white; }\n.mat-table thead, .mat-table tbody, .mat-table tfoot,\nmat-header-row, mat-row, mat-footer-row,\n[mat-header-row], [mat-row], [mat-footer-row],\n.mat-table-sticky {\n  background: inherit; }\nmat-row, mat-header-row, mat-footer-row,\nth.mat-header-cell, td.mat-cell, td.mat-footer-cell {\n  border-bottom-color: rgba(0, 0, 0, 0.12); }\n.mat-header-cell {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-cell, .mat-footer-cell {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-calendar-arrow {\n  border-top-color: rgba(0, 0, 0, 0.54); }\n.mat-datepicker-toggle,\n.mat-datepicker-content .mat-calendar-next-button,\n.mat-datepicker-content .mat-calendar-previous-button {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-calendar-table-header {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-calendar-table-header-divider::after {\n  background: rgba(0, 0, 0, 0.12); }\n.mat-calendar-body-label {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-calendar-body-cell-content {\n  color: rgba(0, 0, 0, 0.87);\n  border-color: transparent; }\n.mat-calendar-body-disabled > .mat-calendar-body-cell-content:not(.mat-calendar-body-selected) {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-calendar-body-cell:not(.mat-calendar-body-disabled):hover > .mat-calendar-body-cell-content:not(.mat-calendar-body-selected),\n.cdk-keyboard-focused .mat-calendar-body-active > .mat-calendar-body-cell-content:not(.mat-calendar-body-selected),\n.cdk-program-focused .mat-calendar-body-active > .mat-calendar-body-cell-content:not(.mat-calendar-body-selected) {\n  background-color: rgba(0, 0, 0, 0.04); }\n.mat-calendar-body-today:not(.mat-calendar-body-selected) {\n  border-color: rgba(0, 0, 0, 0.38); }\n.mat-calendar-body-disabled > .mat-calendar-body-today:not(.mat-calendar-body-selected) {\n  border-color: rgba(0, 0, 0, 0.18); }\n.mat-calendar-body-selected {\n  background-color: #2a7bb8;\n  color: white; }\n.mat-calendar-body-disabled > .mat-calendar-body-selected {\n  background-color: rgba(42, 123, 184, 0.4); }\n.mat-calendar-body-today.mat-calendar-body-selected {\n  box-shadow: inset 0 0 0 1px white; }\n.mat-datepicker-content {\n  box-shadow: 0px 2px 4px -1px rgba(0, 0, 0, 0.2), 0px 4px 5px 0px rgba(0, 0, 0, 0.14), 0px 1px 10px 0px rgba(0, 0, 0, 0.12);\n  background-color: white;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-datepicker-content.mat-accent .mat-calendar-body-selected {\n    background-color: #81c6fa;\n    color: black; }\n.mat-datepicker-content.mat-accent .mat-calendar-body-disabled > .mat-calendar-body-selected {\n    background-color: rgba(129, 198, 250, 0.4); }\n.mat-datepicker-content.mat-accent .mat-calendar-body-today.mat-calendar-body-selected {\n    box-shadow: inset 0 0 0 1px black; }\n.mat-datepicker-content.mat-warn .mat-calendar-body-selected {\n    background-color: #f44336;\n    color: white; }\n.mat-datepicker-content.mat-warn .mat-calendar-body-disabled > .mat-calendar-body-selected {\n    background-color: rgba(244, 67, 54, 0.4); }\n.mat-datepicker-content.mat-warn .mat-calendar-body-today.mat-calendar-body-selected {\n    box-shadow: inset 0 0 0 1px white; }\n.mat-datepicker-content-touch {\n  box-shadow: 0px 0px 0px 0px rgba(0, 0, 0, 0.2), 0px 0px 0px 0px rgba(0, 0, 0, 0.14), 0px 0px 0px 0px rgba(0, 0, 0, 0.12); }\n.mat-datepicker-toggle-active {\n  color: #2a7bb8; }\n.mat-datepicker-toggle-active.mat-accent {\n    color: #81c6fa; }\n.mat-datepicker-toggle-active.mat-warn {\n    color: #f44336; }\n.mat-dialog-container {\n  box-shadow: 0px 11px 15px -7px rgba(0, 0, 0, 0.2), 0px 24px 38px 3px rgba(0, 0, 0, 0.14), 0px 9px 46px 8px rgba(0, 0, 0, 0.12);\n  background: white;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-divider {\n  border-top-color: rgba(0, 0, 0, 0.12); }\n.mat-divider-vertical {\n  border-right-color: rgba(0, 0, 0, 0.12); }\n.mat-expansion-panel {\n  background: white;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-expansion-panel:not([class*=\'mat-elevation-z\']) {\n    box-shadow: 0px 3px 1px -2px rgba(0, 0, 0, 0.2), 0px 2px 2px 0px rgba(0, 0, 0, 0.14), 0px 1px 5px 0px rgba(0, 0, 0, 0.12); }\n.mat-action-row {\n  border-top-color: rgba(0, 0, 0, 0.12); }\n.mat-expansion-panel:not(.mat-expanded) .mat-expansion-panel-header:not([aria-disabled=\'true\']).cdk-keyboard-focused, .mat-expansion-panel:not(.mat-expanded) .mat-expansion-panel-header:not([aria-disabled=\'true\']).cdk-program-focused, .mat-expansion-panel:not(.mat-expanded) .mat-expansion-panel-header:not([aria-disabled=\'true\']):hover {\n  background: rgba(0, 0, 0, 0.04); }\n@media (hover: none) {\n  .mat-expansion-panel:not(.mat-expanded):not([aria-disabled=\'true\'])\n.mat-expansion-panel-header:hover {\n    background: white; } }\n.mat-expansion-panel-header-title {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-expansion-panel-header-description,\n.mat-expansion-indicator::after {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-expansion-panel-header[aria-disabled=\'true\'] {\n  color: rgba(0, 0, 0, 0.26); }\n.mat-expansion-panel-header[aria-disabled=\'true\'] .mat-expansion-panel-header-title,\n  .mat-expansion-panel-header[aria-disabled=\'true\'] .mat-expansion-panel-header-description {\n    color: inherit; }\n.mat-form-field-label {\n  color: rgba(0, 0, 0, 0.6); }\n.mat-hint {\n  color: rgba(0, 0, 0, 0.6); }\n.mat-form-field.mat-focused .mat-form-field-label {\n  color: #2a7bb8; }\n.mat-form-field.mat-focused .mat-form-field-label.mat-accent {\n    color: #81c6fa; }\n.mat-form-field.mat-focused .mat-form-field-label.mat-warn {\n    color: #f44336; }\n.mat-focused .mat-form-field-required-marker {\n  color: #81c6fa; }\n.mat-form-field-ripple {\n  background-color: rgba(0, 0, 0, 0.87); }\n.mat-form-field.mat-focused .mat-form-field-ripple {\n  background-color: #2a7bb8; }\n.mat-form-field.mat-focused .mat-form-field-ripple.mat-accent {\n    background-color: #81c6fa; }\n.mat-form-field.mat-focused .mat-form-field-ripple.mat-warn {\n    background-color: #f44336; }\n.mat-form-field-type-mat-native-select.mat-focused:not(.mat-form-field-invalid) .mat-form-field-infix::after {\n  color: #2a7bb8; }\n.mat-form-field-type-mat-native-select.mat-focused:not(.mat-form-field-invalid).mat-accent .mat-form-field-infix::after {\n  color: #81c6fa; }\n.mat-form-field-type-mat-native-select.mat-focused:not(.mat-form-field-invalid).mat-warn .mat-form-field-infix::after {\n  color: #f44336; }\n.mat-form-field.mat-form-field-invalid .mat-form-field-label {\n  color: #f44336; }\n.mat-form-field.mat-form-field-invalid .mat-form-field-label.mat-accent,\n  .mat-form-field.mat-form-field-invalid .mat-form-field-label .mat-form-field-required-marker {\n    color: #f44336; }\n.mat-form-field.mat-form-field-invalid .mat-form-field-ripple,\n.mat-form-field.mat-form-field-invalid .mat-form-field-ripple.mat-accent {\n  background-color: #f44336; }\n.mat-error {\n  color: #f44336; }\n.mat-form-field-appearance-legacy .mat-form-field-label {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-form-field-appearance-legacy .mat-hint {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-form-field-appearance-legacy .mat-form-field-underline {\n  background-color: rgba(0, 0, 0, 0.42); }\n.mat-form-field-appearance-legacy.mat-form-field-disabled .mat-form-field-underline {\n  background-image: linear-gradient(to right, rgba(0, 0, 0, 0.42) 0%, rgba(0, 0, 0, 0.42) 33%, transparent 0%);\n  background-size: 4px 100%;\n  background-repeat: repeat-x; }\n.mat-form-field-appearance-standard .mat-form-field-underline {\n  background-color: rgba(0, 0, 0, 0.42); }\n.mat-form-field-appearance-standard.mat-form-field-disabled .mat-form-field-underline {\n  background-image: linear-gradient(to right, rgba(0, 0, 0, 0.42) 0%, rgba(0, 0, 0, 0.42) 33%, transparent 0%);\n  background-size: 4px 100%;\n  background-repeat: repeat-x; }\n.mat-form-field-appearance-fill .mat-form-field-flex {\n  background-color: rgba(0, 0, 0, 0.04); }\n.mat-form-field-appearance-fill.mat-form-field-disabled .mat-form-field-flex {\n  background-color: rgba(0, 0, 0, 0.02); }\n.mat-form-field-appearance-fill .mat-form-field-underline::before {\n  background-color: rgba(0, 0, 0, 0.42); }\n.mat-form-field-appearance-fill.mat-form-field-disabled .mat-form-field-label {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-form-field-appearance-fill.mat-form-field-disabled .mat-form-field-underline::before {\n  background-color: transparent; }\n.mat-form-field-appearance-outline .mat-form-field-outline {\n  color: rgba(0, 0, 0, 0.12); }\n.mat-form-field-appearance-outline .mat-form-field-outline-thick {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-form-field-appearance-outline.mat-focused .mat-form-field-outline-thick {\n  color: #2a7bb8; }\n.mat-form-field-appearance-outline.mat-focused.mat-accent .mat-form-field-outline-thick {\n  color: #81c6fa; }\n.mat-form-field-appearance-outline.mat-focused.mat-warn .mat-form-field-outline-thick {\n  color: #f44336; }\n.mat-form-field-appearance-outline.mat-form-field-invalid.mat-form-field-invalid .mat-form-field-outline-thick {\n  color: #f44336; }\n.mat-form-field-appearance-outline.mat-form-field-disabled .mat-form-field-label {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-form-field-appearance-outline.mat-form-field-disabled .mat-form-field-outline {\n  color: rgba(0, 0, 0, 0.06); }\n.mat-icon.mat-primary {\n  color: #2a7bb8; }\n.mat-icon.mat-accent {\n  color: #81c6fa; }\n.mat-icon.mat-warn {\n  color: #f44336; }\n.mat-form-field-type-mat-native-select .mat-form-field-infix::after {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-input-element:disabled,\n.mat-form-field-type-mat-native-select.mat-form-field-disabled .mat-form-field-infix::after {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-input-element {\n  caret-color: #2a7bb8; }\n.mat-input-element::-webkit-input-placeholder {\n    color: rgba(0, 0, 0, 0.42); }\n.mat-input-element:-ms-input-placeholder {\n    color: rgba(0, 0, 0, 0.42); }\n.mat-input-element::-ms-input-placeholder {\n    color: rgba(0, 0, 0, 0.42); }\n.mat-input-element::placeholder {\n    color: rgba(0, 0, 0, 0.42); }\n.mat-input-element::-moz-placeholder {\n    color: rgba(0, 0, 0, 0.42); }\n.mat-input-element::-webkit-input-placeholder {\n    color: rgba(0, 0, 0, 0.42); }\n.mat-input-element:-ms-input-placeholder {\n    color: rgba(0, 0, 0, 0.42); }\n.mat-accent .mat-input-element {\n  caret-color: #81c6fa; }\n.mat-warn .mat-input-element,\n.mat-form-field-invalid .mat-input-element {\n  caret-color: #f44336; }\n.mat-form-field-type-mat-native-select.mat-form-field-invalid .mat-form-field-infix::after {\n  color: #f44336; }\n.mat-list-base .mat-list-item {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-list-base .mat-list-option {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-list-base .mat-subheader {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-list-item-disabled {\n  background-color: #eeeeee; }\n.mat-list-option:hover, .mat-list-option:focus,\n.mat-nav-list .mat-list-item:hover,\n.mat-nav-list .mat-list-item:focus {\n  background: rgba(0, 0, 0, 0.04); }\n.mat-menu-panel {\n  background: white; }\n.mat-menu-panel:not([class*=\'mat-elevation-z\']) {\n    box-shadow: 0px 2px 4px -1px rgba(0, 0, 0, 0.2), 0px 4px 5px 0px rgba(0, 0, 0, 0.14), 0px 1px 10px 0px rgba(0, 0, 0, 0.12); }\n.mat-menu-item {\n  background: transparent;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-menu-item[disabled], .mat-menu-item[disabled]::after {\n    color: rgba(0, 0, 0, 0.38); }\n.mat-menu-item .mat-icon-no-color,\n.mat-menu-item-submenu-trigger::after {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-menu-item:hover:not([disabled]),\n.mat-menu-item.cdk-program-focused:not([disabled]),\n.mat-menu-item.cdk-keyboard-focused:not([disabled]),\n.mat-menu-item-highlighted:not([disabled]) {\n  background: rgba(0, 0, 0, 0.04); }\n.mat-paginator {\n  background: white; }\n.mat-paginator,\n.mat-paginator-page-size .mat-select-trigger {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-paginator-decrement,\n.mat-paginator-increment {\n  border-top: 2px solid rgba(0, 0, 0, 0.54);\n  border-right: 2px solid rgba(0, 0, 0, 0.54); }\n.mat-paginator-first,\n.mat-paginator-last {\n  border-top: 2px solid rgba(0, 0, 0, 0.54); }\n.mat-icon-button[disabled] .mat-paginator-decrement,\n.mat-icon-button[disabled] .mat-paginator-increment,\n.mat-icon-button[disabled] .mat-paginator-first,\n.mat-icon-button[disabled] .mat-paginator-last {\n  border-color: rgba(0, 0, 0, 0.38); }\n.mat-progress-bar-background {\n  fill: #b1d4ed; }\n.mat-progress-bar-buffer {\n  background-color: #b1d4ed; }\n.mat-progress-bar-fill::after {\n  background-color: #2a7bb8; }\n.mat-progress-bar.mat-accent .mat-progress-bar-background {\n  fill: #b1d4ed; }\n.mat-progress-bar.mat-accent .mat-progress-bar-buffer {\n  background-color: #b1d4ed; }\n.mat-progress-bar.mat-accent .mat-progress-bar-fill::after {\n  background-color: #81c6fa; }\n.mat-progress-bar.mat-warn .mat-progress-bar-background {\n  fill: #ffcdd2; }\n.mat-progress-bar.mat-warn .mat-progress-bar-buffer {\n  background-color: #ffcdd2; }\n.mat-progress-bar.mat-warn .mat-progress-bar-fill::after {\n  background-color: #f44336; }\n.mat-progress-spinner circle, .mat-spinner circle {\n  stroke: #2a7bb8; }\n.mat-progress-spinner.mat-accent circle, .mat-spinner.mat-accent circle {\n  stroke: #81c6fa; }\n.mat-progress-spinner.mat-warn circle, .mat-spinner.mat-warn circle {\n  stroke: #f44336; }\n.mat-radio-outer-circle {\n  border-color: rgba(0, 0, 0, 0.54); }\n.mat-radio-button.mat-primary.mat-radio-checked .mat-radio-outer-circle {\n  border-color: #2a7bb8; }\n.mat-radio-button.mat-primary .mat-radio-inner-circle,\n.mat-radio-button.mat-primary .mat-radio-ripple .mat-ripple-element:not(.mat-radio-persistent-ripple),\n.mat-radio-button.mat-primary.mat-radio-checked .mat-radio-persistent-ripple,\n.mat-radio-button.mat-primary:active .mat-radio-persistent-ripple {\n  background-color: #2a7bb8; }\n.mat-radio-button.mat-accent.mat-radio-checked .mat-radio-outer-circle {\n  border-color: #81c6fa; }\n.mat-radio-button.mat-accent .mat-radio-inner-circle,\n.mat-radio-button.mat-accent .mat-radio-ripple .mat-ripple-element:not(.mat-radio-persistent-ripple),\n.mat-radio-button.mat-accent.mat-radio-checked .mat-radio-persistent-ripple,\n.mat-radio-button.mat-accent:active .mat-radio-persistent-ripple {\n  background-color: #81c6fa; }\n.mat-radio-button.mat-warn.mat-radio-checked .mat-radio-outer-circle {\n  border-color: #f44336; }\n.mat-radio-button.mat-warn .mat-radio-inner-circle,\n.mat-radio-button.mat-warn .mat-radio-ripple .mat-ripple-element:not(.mat-radio-persistent-ripple),\n.mat-radio-button.mat-warn.mat-radio-checked .mat-radio-persistent-ripple,\n.mat-radio-button.mat-warn:active .mat-radio-persistent-ripple {\n  background-color: #f44336; }\n.mat-radio-button.mat-radio-disabled.mat-radio-checked .mat-radio-outer-circle,\n.mat-radio-button.mat-radio-disabled .mat-radio-outer-circle {\n  border-color: rgba(0, 0, 0, 0.38); }\n.mat-radio-button.mat-radio-disabled .mat-radio-ripple .mat-ripple-element,\n.mat-radio-button.mat-radio-disabled .mat-radio-inner-circle {\n  background-color: rgba(0, 0, 0, 0.38); }\n.mat-radio-button.mat-radio-disabled .mat-radio-label-content {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-radio-button .mat-ripple-element {\n  background-color: black; }\n.mat-select-value {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-select-placeholder {\n  color: rgba(0, 0, 0, 0.42); }\n.mat-select-disabled .mat-select-value {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-select-arrow {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-select-panel {\n  background: white; }\n.mat-select-panel:not([class*=\'mat-elevation-z\']) {\n    box-shadow: 0px 2px 4px -1px rgba(0, 0, 0, 0.2), 0px 4px 5px 0px rgba(0, 0, 0, 0.14), 0px 1px 10px 0px rgba(0, 0, 0, 0.12); }\n.mat-select-panel .mat-option.mat-selected:not(.mat-option-multiple) {\n    background: rgba(0, 0, 0, 0.12); }\n.mat-form-field.mat-focused.mat-primary .mat-select-arrow {\n  color: #2a7bb8; }\n.mat-form-field.mat-focused.mat-accent .mat-select-arrow {\n  color: #81c6fa; }\n.mat-form-field.mat-focused.mat-warn .mat-select-arrow {\n  color: #f44336; }\n.mat-form-field .mat-select.mat-select-invalid .mat-select-arrow {\n  color: #f44336; }\n.mat-form-field .mat-select.mat-select-disabled .mat-select-arrow {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-drawer-container {\n  background-color: #fafafa;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-drawer {\n  background-color: white;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-drawer.mat-drawer-push {\n    background-color: white; }\n.mat-drawer:not(.mat-drawer-side) {\n    box-shadow: 0px 8px 10px -5px rgba(0, 0, 0, 0.2), 0px 16px 24px 2px rgba(0, 0, 0, 0.14), 0px 6px 30px 5px rgba(0, 0, 0, 0.12); }\n.mat-drawer-side {\n  border-right: solid 1px rgba(0, 0, 0, 0.12); }\n.mat-drawer-side.mat-drawer-end {\n    border-left: solid 1px rgba(0, 0, 0, 0.12);\n    border-right: none; }\n[dir=\'rtl\'] .mat-drawer-side {\n  border-left: solid 1px rgba(0, 0, 0, 0.12);\n  border-right: none; }\n[dir=\'rtl\'] .mat-drawer-side.mat-drawer-end {\n    border-left: none;\n    border-right: solid 1px rgba(0, 0, 0, 0.12); }\n.mat-drawer-backdrop.mat-drawer-shown {\n  background-color: rgba(0, 0, 0, 0.6); }\n.mat-slide-toggle.mat-checked .mat-slide-toggle-thumb {\n  background-color: #81c6fa; }\n.mat-slide-toggle.mat-checked .mat-slide-toggle-bar {\n  background-color: rgba(129, 198, 250, 0.54); }\n.mat-slide-toggle.mat-checked .mat-ripple-element {\n  background-color: #81c6fa; }\n.mat-slide-toggle.mat-primary.mat-checked .mat-slide-toggle-thumb {\n  background-color: #2a7bb8; }\n.mat-slide-toggle.mat-primary.mat-checked .mat-slide-toggle-bar {\n  background-color: rgba(42, 123, 184, 0.54); }\n.mat-slide-toggle.mat-primary.mat-checked .mat-ripple-element {\n  background-color: #2a7bb8; }\n.mat-slide-toggle.mat-warn.mat-checked .mat-slide-toggle-thumb {\n  background-color: #f44336; }\n.mat-slide-toggle.mat-warn.mat-checked .mat-slide-toggle-bar {\n  background-color: rgba(244, 67, 54, 0.54); }\n.mat-slide-toggle.mat-warn.mat-checked .mat-ripple-element {\n  background-color: #f44336; }\n.mat-slide-toggle:not(.mat-checked) .mat-ripple-element {\n  background-color: black; }\n.mat-slide-toggle-thumb {\n  box-shadow: 0px 2px 1px -1px rgba(0, 0, 0, 0.2), 0px 1px 1px 0px rgba(0, 0, 0, 0.14), 0px 1px 3px 0px rgba(0, 0, 0, 0.12);\n  background-color: #fafafa; }\n.mat-slide-toggle-bar {\n  background-color: rgba(0, 0, 0, 0.38); }\n.mat-slider-track-background {\n  background-color: rgba(0, 0, 0, 0.26); }\n.mat-primary .mat-slider-track-fill,\n.mat-primary .mat-slider-thumb,\n.mat-primary .mat-slider-thumb-label {\n  background-color: #2a7bb8; }\n.mat-primary .mat-slider-thumb-label-text {\n  color: white; }\n.mat-accent .mat-slider-track-fill,\n.mat-accent .mat-slider-thumb,\n.mat-accent .mat-slider-thumb-label {\n  background-color: #81c6fa; }\n.mat-accent .mat-slider-thumb-label-text {\n  color: black; }\n.mat-warn .mat-slider-track-fill,\n.mat-warn .mat-slider-thumb,\n.mat-warn .mat-slider-thumb-label {\n  background-color: #f44336; }\n.mat-warn .mat-slider-thumb-label-text {\n  color: white; }\n.mat-slider-focus-ring {\n  background-color: rgba(129, 198, 250, 0.2); }\n.mat-slider:hover .mat-slider-track-background,\n.cdk-focused .mat-slider-track-background {\n  background-color: rgba(0, 0, 0, 0.38); }\n.mat-slider-disabled .mat-slider-track-background,\n.mat-slider-disabled .mat-slider-track-fill,\n.mat-slider-disabled .mat-slider-thumb {\n  background-color: rgba(0, 0, 0, 0.26); }\n.mat-slider-disabled:hover .mat-slider-track-background {\n  background-color: rgba(0, 0, 0, 0.26); }\n.mat-slider-min-value .mat-slider-focus-ring {\n  background-color: rgba(0, 0, 0, 0.12); }\n.mat-slider-min-value.mat-slider-thumb-label-showing .mat-slider-thumb,\n.mat-slider-min-value.mat-slider-thumb-label-showing .mat-slider-thumb-label {\n  background-color: rgba(0, 0, 0, 0.87); }\n.mat-slider-min-value.mat-slider-thumb-label-showing.cdk-focused .mat-slider-thumb,\n.mat-slider-min-value.mat-slider-thumb-label-showing.cdk-focused .mat-slider-thumb-label {\n  background-color: rgba(0, 0, 0, 0.26); }\n.mat-slider-min-value:not(.mat-slider-thumb-label-showing) .mat-slider-thumb {\n  border-color: rgba(0, 0, 0, 0.26);\n  background-color: transparent; }\n.mat-slider-min-value:not(.mat-slider-thumb-label-showing):hover .mat-slider-thumb, .mat-slider-min-value:not(.mat-slider-thumb-label-showing).cdk-focused .mat-slider-thumb {\n  border-color: rgba(0, 0, 0, 0.38); }\n.mat-slider-min-value:not(.mat-slider-thumb-label-showing):hover.mat-slider-disabled .mat-slider-thumb, .mat-slider-min-value:not(.mat-slider-thumb-label-showing).cdk-focused.mat-slider-disabled .mat-slider-thumb {\n  border-color: rgba(0, 0, 0, 0.26); }\n.mat-slider-has-ticks .mat-slider-wrapper::after {\n  border-color: rgba(0, 0, 0, 0.7); }\n.mat-slider-horizontal .mat-slider-ticks {\n  background-image: repeating-linear-gradient(to right, rgba(0, 0, 0, 0.7), rgba(0, 0, 0, 0.7) 2px, transparent 0, transparent);\n  background-image: -moz-repeating-linear-gradient(0.0001deg, rgba(0, 0, 0, 0.7), rgba(0, 0, 0, 0.7) 2px, transparent 0, transparent); }\n.mat-slider-vertical .mat-slider-ticks {\n  background-image: repeating-linear-gradient(to bottom, rgba(0, 0, 0, 0.7), rgba(0, 0, 0, 0.7) 2px, transparent 0, transparent); }\n.mat-step-header.cdk-keyboard-focused, .mat-step-header.cdk-program-focused, .mat-step-header:hover {\n  background-color: rgba(0, 0, 0, 0.04); }\n@media (hover: none) {\n  .mat-step-header:hover {\n    background: none; } }\n.mat-step-header .mat-step-label,\n.mat-step-header .mat-step-optional {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-step-header .mat-step-icon {\n  background-color: rgba(0, 0, 0, 0.54);\n  color: white; }\n.mat-step-header .mat-step-icon-selected,\n.mat-step-header .mat-step-icon-state-done,\n.mat-step-header .mat-step-icon-state-edit {\n  background-color: #2a7bb8;\n  color: white; }\n.mat-step-header .mat-step-icon-state-error {\n  background-color: transparent;\n  color: #f44336; }\n.mat-step-header .mat-step-label.mat-step-label-active {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-step-header .mat-step-label.mat-step-label-error {\n  color: #f44336; }\n.mat-stepper-horizontal, .mat-stepper-vertical {\n  background-color: white; }\n.mat-stepper-vertical-line::before {\n  border-left-color: rgba(0, 0, 0, 0.12); }\n.mat-horizontal-stepper-header::before,\n.mat-horizontal-stepper-header::after,\n.mat-stepper-horizontal-line {\n  border-top-color: rgba(0, 0, 0, 0.12); }\n.mat-sort-header-arrow {\n  color: #757575; }\n.mat-tab-nav-bar,\n.mat-tab-header {\n  border-bottom: 1px solid rgba(0, 0, 0, 0.12); }\n.mat-tab-group-inverted-header .mat-tab-nav-bar,\n.mat-tab-group-inverted-header .mat-tab-header {\n  border-top: 1px solid rgba(0, 0, 0, 0.12);\n  border-bottom: none; }\n.mat-tab-label, .mat-tab-link {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-tab-label.mat-tab-disabled, .mat-tab-link.mat-tab-disabled {\n    color: rgba(0, 0, 0, 0.38); }\n.mat-tab-header-pagination-chevron {\n  border-color: rgba(0, 0, 0, 0.87); }\n.mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron {\n  border-color: rgba(0, 0, 0, 0.38); }\n.mat-tab-group[class*=\'mat-background-\'] .mat-tab-header,\n.mat-tab-nav-bar[class*=\'mat-background-\'] {\n  border-bottom: none;\n  border-top: none; }\n.mat-tab-group.mat-primary .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-group.mat-primary .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-primary .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-primary .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-primary .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-primary .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-primary .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-primary .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled) {\n  background-color: rgba(177, 212, 237, 0.3); }\n.mat-tab-group.mat-primary .mat-ink-bar, .mat-tab-nav-bar.mat-primary .mat-ink-bar {\n  background-color: #2a7bb8; }\n.mat-tab-group.mat-primary.mat-background-primary .mat-ink-bar, .mat-tab-nav-bar.mat-primary.mat-background-primary .mat-ink-bar {\n  background-color: white; }\n.mat-tab-group.mat-accent .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-group.mat-accent .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-accent .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-accent .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-accent .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-accent .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-accent .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-accent .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled) {\n  background-color: rgba(177, 212, 237, 0.3); }\n.mat-tab-group.mat-accent .mat-ink-bar, .mat-tab-nav-bar.mat-accent .mat-ink-bar {\n  background-color: #81c6fa; }\n.mat-tab-group.mat-accent.mat-background-accent .mat-ink-bar, .mat-tab-nav-bar.mat-accent.mat-background-accent .mat-ink-bar {\n  background-color: black; }\n.mat-tab-group.mat-warn .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-group.mat-warn .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-warn .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-warn .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-warn .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-warn .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-warn .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-warn .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled) {\n  background-color: rgba(255, 205, 210, 0.3); }\n.mat-tab-group.mat-warn .mat-ink-bar, .mat-tab-nav-bar.mat-warn .mat-ink-bar {\n  background-color: #f44336; }\n.mat-tab-group.mat-warn.mat-background-warn .mat-ink-bar, .mat-tab-nav-bar.mat-warn.mat-background-warn .mat-ink-bar {\n  background-color: white; }\n.mat-tab-group.mat-background-primary .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-group.mat-background-primary .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-background-primary .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-background-primary .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-background-primary .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-background-primary .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-background-primary .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-background-primary .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled) {\n  background-color: rgba(177, 212, 237, 0.3); }\n.mat-tab-group.mat-background-primary .mat-tab-header, .mat-tab-group.mat-background-primary .mat-tab-links, .mat-tab-nav-bar.mat-background-primary .mat-tab-header, .mat-tab-nav-bar.mat-background-primary .mat-tab-links {\n  background-color: #2a7bb8; }\n.mat-tab-group.mat-background-primary .mat-tab-label, .mat-tab-group.mat-background-primary .mat-tab-link, .mat-tab-nav-bar.mat-background-primary .mat-tab-label, .mat-tab-nav-bar.mat-background-primary .mat-tab-link {\n  color: white; }\n.mat-tab-group.mat-background-primary .mat-tab-label.mat-tab-disabled, .mat-tab-group.mat-background-primary .mat-tab-link.mat-tab-disabled, .mat-tab-nav-bar.mat-background-primary .mat-tab-label.mat-tab-disabled, .mat-tab-nav-bar.mat-background-primary .mat-tab-link.mat-tab-disabled {\n    color: rgba(255, 255, 255, 0.4); }\n.mat-tab-group.mat-background-primary .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-primary .mat-tab-header-pagination-chevron {\n  border-color: white; }\n.mat-tab-group.mat-background-primary .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-primary .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron {\n  border-color: rgba(255, 255, 255, 0.4); }\n.mat-tab-group.mat-background-primary .mat-ripple-element, .mat-tab-nav-bar.mat-background-primary .mat-ripple-element {\n  background-color: rgba(255, 255, 255, 0.12); }\n.mat-tab-group.mat-background-accent .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-group.mat-background-accent .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-background-accent .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-background-accent .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-background-accent .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-background-accent .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-background-accent .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-background-accent .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled) {\n  background-color: rgba(177, 212, 237, 0.3); }\n.mat-tab-group.mat-background-accent .mat-tab-header, .mat-tab-group.mat-background-accent .mat-tab-links, .mat-tab-nav-bar.mat-background-accent .mat-tab-header, .mat-tab-nav-bar.mat-background-accent .mat-tab-links {\n  background-color: #81c6fa; }\n.mat-tab-group.mat-background-accent .mat-tab-label, .mat-tab-group.mat-background-accent .mat-tab-link, .mat-tab-nav-bar.mat-background-accent .mat-tab-label, .mat-tab-nav-bar.mat-background-accent .mat-tab-link {\n  color: black; }\n.mat-tab-group.mat-background-accent .mat-tab-label.mat-tab-disabled, .mat-tab-group.mat-background-accent .mat-tab-link.mat-tab-disabled, .mat-tab-nav-bar.mat-background-accent .mat-tab-label.mat-tab-disabled, .mat-tab-nav-bar.mat-background-accent .mat-tab-link.mat-tab-disabled {\n    color: rgba(0, 0, 0, 0.4); }\n.mat-tab-group.mat-background-accent .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-accent .mat-tab-header-pagination-chevron {\n  border-color: black; }\n.mat-tab-group.mat-background-accent .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-accent .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron {\n  border-color: rgba(0, 0, 0, 0.4); }\n.mat-tab-group.mat-background-accent .mat-ripple-element, .mat-tab-nav-bar.mat-background-accent .mat-ripple-element {\n  background-color: rgba(0, 0, 0, 0.12); }\n.mat-tab-group.mat-background-warn .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-group.mat-background-warn .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-background-warn .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-background-warn .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-background-warn .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-background-warn .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-background-warn .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-background-warn .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled) {\n  background-color: rgba(255, 205, 210, 0.3); }\n.mat-tab-group.mat-background-warn .mat-tab-header, .mat-tab-group.mat-background-warn .mat-tab-links, .mat-tab-nav-bar.mat-background-warn .mat-tab-header, .mat-tab-nav-bar.mat-background-warn .mat-tab-links {\n  background-color: #f44336; }\n.mat-tab-group.mat-background-warn .mat-tab-label, .mat-tab-group.mat-background-warn .mat-tab-link, .mat-tab-nav-bar.mat-background-warn .mat-tab-label, .mat-tab-nav-bar.mat-background-warn .mat-tab-link {\n  color: white; }\n.mat-tab-group.mat-background-warn .mat-tab-label.mat-tab-disabled, .mat-tab-group.mat-background-warn .mat-tab-link.mat-tab-disabled, .mat-tab-nav-bar.mat-background-warn .mat-tab-label.mat-tab-disabled, .mat-tab-nav-bar.mat-background-warn .mat-tab-link.mat-tab-disabled {\n    color: rgba(255, 255, 255, 0.4); }\n.mat-tab-group.mat-background-warn .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-warn .mat-tab-header-pagination-chevron {\n  border-color: white; }\n.mat-tab-group.mat-background-warn .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-warn .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron {\n  border-color: rgba(255, 255, 255, 0.4); }\n.mat-tab-group.mat-background-warn .mat-ripple-element, .mat-tab-nav-bar.mat-background-warn .mat-ripple-element {\n  background-color: rgba(255, 255, 255, 0.12); }\n.mat-toolbar {\n  background: whitesmoke;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-toolbar.mat-primary {\n    background: #2a7bb8;\n    color: white; }\n.mat-toolbar.mat-accent {\n    background: #81c6fa;\n    color: black; }\n.mat-toolbar.mat-warn {\n    background: #f44336;\n    color: white; }\n.mat-toolbar .mat-form-field-underline,\n  .mat-toolbar .mat-form-field-ripple,\n  .mat-toolbar .mat-focused .mat-form-field-ripple {\n    background-color: currentColor; }\n.mat-toolbar .mat-form-field-label,\n  .mat-toolbar .mat-focused .mat-form-field-label,\n  .mat-toolbar .mat-select-value,\n  .mat-toolbar .mat-select-arrow,\n  .mat-toolbar .mat-form-field.mat-focused .mat-select-arrow {\n    color: inherit; }\n.mat-toolbar .mat-input-element {\n    caret-color: currentColor; }\n.mat-tooltip {\n  background: rgba(97, 97, 97, 0.9); }\n.mat-tree {\n  background: white; }\n.mat-tree-node,\n.mat-nested-tree-node {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-snack-bar-container {\n  color: rgba(255, 255, 255, 0.7);\n  background: #323232;\n  box-shadow: 0px 3px 5px -1px rgba(0, 0, 0, 0.2), 0px 6px 10px 0px rgba(0, 0, 0, 0.14), 0px 1px 18px 0px rgba(0, 0, 0, 0.12); }\n.mat-simple-snackbar-action {\n  color: #81c6fa; }\n.fixed-header {\n  position: fixed;\n  top: 0;\n  left: 0;\n  z-index: 2;\n  width: 100%  !important; }\nbody {\n  padding-top: 3%; }\ntextarea {\n  color: Gray;\n  border: none;\n  background-color: #f0f6fc;\n  overflow-wrap: break-word;\n  width: 100%; }\n.test-title {\n  color: black;\n  font-weight: 500; }\n.MatButtonEmulation {\n  font: caption;\n  display: inline-block;\n  margin: 0.05vh;\n  white-space: normal;\n  word-wrap: break-word;\n  border-radius: 2px;\n  text-align: center;\n  color: white;\n  background: #2a7bb8;\n  width: 70px;\n  padding: 0.5vw;\n  height: auto; }\n.badge {\n  position: relative;\n  text-align: center;\n  background: #ff4081;\n  width: 1vmax;\n  height: 1vmax;\n  line-height: 1vmax;\n  border-radius: 50%; }\n.vert-align {\n  display: inline-block;\n  vertical-align: middle;\n  line-height: normal;\n  color: white;\n  font-size: 0.6vmax; }\n\n/*# sourceMappingURL=data:application/json;base64,  */',"",""]]},"./node_modules/style-loader/lib/addStyles.js":function(b,t,g){var I,c,G,C={},W=(I=function(){return window&&document&&document.all&&!window.atob},function(){return void 0===c&&(c=I.apply(this,arguments)),c}),m=(G={},function(b,t){if("function"==typeof b)return b();if(void 0===G[b]){var g=function(b,t){return t?t.querySelector(b):document.querySelector(b)}.call(this,b,t);if(window.HTMLIFrameElement&&g instanceof window.HTMLIFrameElement)try{g=g.contentDocument.head}catch(b){g=null}G[b]=g}return G[b]}),Z=null,l=0,d=[],a=g("./node_modules/style-loader/lib/urls.js");function X(b,t){for(var g=0;g<b.length;g++){var I=b[g],c=C[I.id];if(c){c.refs++;for(var G=0;G<c.parts.length;G++)c.parts[G](I.parts[G]);for(;G<I.parts.length;G++)c.parts.push(L(I.parts[G],t))}else{var W=[];for(G=0;G<I.parts.length;G++)W.push(L(I.parts[G],t));C[I.id]={id:I.id,refs:1,parts:W}}}}function i(b,t){for(var g=[],I={},c=0;c<b.length;c++){var G=b[c],C=t.base?G[0]+t.base:G[0],W={css:G[1],media:G[2],sourceMap:G[3]};I[C]?I[C].parts.push(W):g.push(I[C]={id:C,parts:[W]})}return g}function Y(b,t){var g=m(b.insertInto);if(!g)throw new Error("Couldn't find a style target. This probably means that the value for the 'insertInto' parameter is invalid.");var I=d[d.length-1];if("top"===b.insertAt)I?I.nextSibling?g.insertBefore(t,I.nextSibling):g.appendChild(t):g.insertBefore(t,g.firstChild),d.push(t);else if("bottom"===b.insertAt)g.appendChild(t);else{if("object"!=typeof b.insertAt||!b.insertAt.before)throw new Error("[Style Loader]\n\n Invalid value for parameter 'insertAt' ('options.insertAt') found.\n Must be 'top', 'bottom', or Object.\n (https://github.com/webpack-contrib/style-loader#insertat)\n");var c=m(b.insertAt.before,g);g.insertBefore(t,c)}}function A(b){if(null===b.parentNode)return!1;b.parentNode.removeChild(b);var t=d.indexOf(b);t>=0&&d.splice(t,1)}function B(b){var t=document.createElement("style");if(void 0===b.attrs.type&&(b.attrs.type="text/css"),void 0===b.attrs.nonce){var I=function(){0;return g.nc}();I&&(b.attrs.nonce=I)}return u(t,b.attrs),Y(b,t),t}function u(b,t){Object.keys(t).forEach(function(g){b.setAttribute(g,t[g])})}function L(b,t){var g,I,c,G,C,W;if(t.transform&&b.css){if(!(G="function"==typeof t.transform?t.transform(b.css):t.transform.default(b.css)))return function(){};b.css=G}if(t.singleton){var m=l++;g=Z||(Z=B(t)),I=h.bind(null,g,m,!1),c=h.bind(null,g,m,!0)}else b.sourceMap&&"function"==typeof URL&&"function"==typeof URL.createObjectURL&&"function"==typeof URL.revokeObjectURL&&"function"==typeof Blob&&"function"==typeof btoa?(C=t,W=document.createElement("link"),void 0===C.attrs.type&&(C.attrs.type="text/css"),C.attrs.rel="stylesheet",u(W,C.attrs),Y(C,W),I=function(b,t,g){var I=g.css,c=g.sourceMap,G=void 0===t.convertToAbsoluteUrls&&c;(t.convertToAbsoluteUrls||G)&&(I=a(I));c&&(I+="\n/*# sourceMappingURL=data:application/json;base64,"+btoa(unescape(encodeURIComponent(JSON.stringify(c))))+" */");var C=new Blob([I],{type:"text/css"}),W=b.href;b.href=URL.createObjectURL(C),W&&URL.revokeObjectURL(W)}.bind(null,g=W,t),c=function(){A(g),g.href&&URL.revokeObjectURL(g.href)}):(g=B(t),I=function(b,t){var g=t.css,I=t.media;I&&b.setAttribute("media",I);if(b.styleSheet)b.styleSheet.cssText=g;else{for(;b.firstChild;)b.removeChild(b.firstChild);b.appendChild(document.createTextNode(g))}}.bind(null,g),c=function(){A(g)});return I(b),function(t){if(t){if(t.css===b.css&&t.media===b.media&&t.sourceMap===b.sourceMap)return;I(b=t)}else c()}}b.exports=function(b,t){if("undefined"!=typeof DEBUG&&DEBUG&&"object"!=typeof document)throw new Error("The style-loader cannot be used in a non-browser environment");(t=t||{}).attrs="object"==typeof t.attrs?t.attrs:{},t.singleton||"boolean"==typeof t.singleton||(t.singleton=W()),t.insertInto||(t.insertInto="head"),t.insertAt||(t.insertAt="bottom");var g=i(b,t);return X(g,t),function(b){for(var I=[],c=0;c<g.length;c++){var G=g[c];(W=C[G.id]).refs--,I.push(W)}b&&X(i(b,t),t);for(c=0;c<I.length;c++){var W;if(0===(W=I[c]).refs){for(var m=0;m<W.parts.length;m++)W.parts[m]();delete C[W.id]}}}};var n,F=(n=[],function(b,t){return n[b]=t,n.filter(Boolean).join("\n")});function h(b,t,g,I){var c=g?"":I.css;if(b.styleSheet)b.styleSheet.cssText=F(t,c);else{var G=document.createTextNode(c),C=b.childNodes;C[t]&&b.removeChild(C[t]),C.length?b.insertBefore(G,C[t]):b.appendChild(G)}}},"./node_modules/style-loader/lib/urls.js":function(b,t){b.exports=function(b){var t="undefined"!=typeof window&&window.location;if(!t)throw new Error("fixUrls requires window.location");if(!b||"string"!=typeof b)return b;var g=t.protocol+"//"+t.host,I=g+t.pathname.replace(/\/[^\/]*$/,"/");return b.replace(/url\s*\(((?:[^)(]|\((?:[^)(]+|\([^)(]*\))*\))*)\)/gi,function(b,t){var c,G=t.trim().replace(/^"(.*)"$/,function(b,t){return t}).replace(/^'(.*)'$/,function(b,t){return t});return/^(#|data:|http:\/\/|https:\/\/|file:\/\/\/|\s*$)/i.test(G)?b:(c=0===G.indexOf("//")?G:0===G.indexOf("/")?g+G:I+G.replace(/^\.\//,""),"url("+JSON.stringify(c)+")")})}},"./src/styles.scss":function(b,t,g){var I=g("./node_modules/@angular-devkit/build-angular/src/angular-cli-files/plugins/raw-css-loader.js!./node_modules/postcss-loader/src/index.js?!./node_modules/sass-loader/lib/loader.js?!./src/styles.scss");"string"==typeof I&&(I=[[b.i,I,""]]);var c={hmr:!0};c.transform=void 0,c.insertInto=void 0;g("./node_modules/style-loader/lib/addStyles.js")(I,c);I.locals&&(b.exports=I.locals)},2:function(b,t,g){b.exports=g("./src/styles.scss")}},[[2,"runtime"]]]);
</script><script type="text/javascript">
(window.webpackJsonp=window.webpackJsonp||[]).push([["vendor"],{"./node_modules/@angular/animations/fesm5/animations.js":function(t,e,n){"use strict";n.r(e),n.d(e,"AnimationBuilder",function(){return r}),n.d(e,"AnimationFactory",function(){return i}),n.d(e,"AUTO_STYLE",function(){return o}),n.d(e,"animate",function(){return s}),n.d(e,"animateChild",function(){return m}),n.d(e,"animation",function(){return h}),n.d(e,"group",function(){return u}),n.d(e,"keyframes",function(){return p}),n.d(e,"query",function(){return y}),n.d(e,"sequence",function(){return c}),n.d(e,"stagger",function(){return _}),n.d(e,"state",function(){return d}),n.d(e,"style",function(){return l}),n.d(e,"transition",function(){return f}),n.d(e,"trigger",function(){return a}),n.d(e,"useAnimation",function(){return g}),n.d(e,"NoopAnimationPlayer",function(){return b}),n.d(e,"ɵPRE_STYLE",function(){return w}),n.d(e,"ɵAnimationGroupPlayer",function(){return x});var r=function(){return function(){}}(),i=function(){return function(){}}(),o="*";function a(t,e){return{type:7,name:t,definitions:e,options:{}}}function s(t,e){return void 0===e&&(e=null),{type:4,styles:e,timings:t}}function u(t,e){return void 0===e&&(e=null),{type:3,steps:t,options:e}}function c(t,e){return void 0===e&&(e=null),{type:2,steps:t,options:e}}function l(t){return{type:6,styles:t,offset:null}}function d(t,e,n){return{type:0,name:t,styles:e,options:n}}function p(t){return{type:5,steps:t}}function f(t,e,n){return void 0===n&&(n=null),{type:1,expr:t,animation:e,options:n}}function h(t,e){return void 0===e&&(e=null),{type:8,animation:t,options:e}}function m(t){return void 0===t&&(t=null),{type:9,options:t}}function g(t,e){return void 0===e&&(e=null),{type:10,animation:t,options:e}}function y(t,e,n){return void 0===n&&(n=null),{type:11,selector:t,animation:e,options:n}}function _(t,e){return{type:12,timings:t,animation:e}}function v(t){Promise.resolve(null).then(t)}var b=function(){function t(t,e){void 0===t&&(t=0),void 0===e&&(e=0),this._onDoneFns=[],this._onStartFns=[],this._onDestroyFns=[],this._started=!1,this._destroyed=!1,this._finished=!1,this.parentPlayer=null,this.totalTime=t+e}return t.prototype._onFinish=function(){this._finished||(this._finished=!0,this._onDoneFns.forEach(function(t){return t()}),this._onDoneFns=[])},t.prototype.onStart=function(t){this._onStartFns.push(t)},t.prototype.onDone=function(t){this._onDoneFns.push(t)},t.prototype.onDestroy=function(t){this._onDestroyFns.push(t)},t.prototype.hasStarted=function(){return this._started},t.prototype.init=function(){},t.prototype.play=function(){this.hasStarted()||(this._onStart(),this.triggerMicrotask()),this._started=!0},t.prototype.triggerMicrotask=function(){var t=this;v(function(){return t._onFinish()})},t.prototype._onStart=function(){this._onStartFns.forEach(function(t){return t()}),this._onStartFns=[]},t.prototype.pause=function(){},t.prototype.restart=function(){},t.prototype.finish=function(){this._onFinish()},t.prototype.destroy=function(){this._destroyed||(this._destroyed=!0,this.hasStarted()||this._onStart(),this.finish(),this._onDestroyFns.forEach(function(t){return t()}),this._onDestroyFns=[])},t.prototype.reset=function(){},t.prototype.setPosition=function(t){},t.prototype.getPosition=function(){return 0},t.prototype.triggerCallback=function(t){var e="start"==t?this._onStartFns:this._onDoneFns;e.forEach(function(t){return t()}),e.length=0},t}(),x=function(){function t(t){var e=this;this._onDoneFns=[],this._onStartFns=[],this._finished=!1,this._started=!1,this._destroyed=!1,this._onDestroyFns=[],this.parentPlayer=null,this.totalTime=0,this.players=t;var n=0,r=0,i=0,o=this.players.length;0==o?v(function(){return e._onFinish()}):this.players.forEach(function(t){t.onDone(function(){++n==o&&e._onFinish()}),t.onDestroy(function(){++r==o&&e._onDestroy()}),t.onStart(function(){++i==o&&e._onStart()})}),this.totalTime=this.players.reduce(function(t,e){return Math.max(t,e.totalTime)},0)}return t.prototype._onFinish=function(){this._finished||(this._finished=!0,this._onDoneFns.forEach(function(t){return t()}),this._onDoneFns=[])},t.prototype.init=function(){this.players.forEach(function(t){return t.init()})},t.prototype.onStart=function(t){this._onStartFns.push(t)},t.prototype._onStart=function(){this.hasStarted()||(this._started=!0,this._onStartFns.forEach(function(t){return t()}),this._onStartFns=[])},t.prototype.onDone=function(t){this._onDoneFns.push(t)},t.prototype.onDestroy=function(t){this._onDestroyFns.push(t)},t.prototype.hasStarted=function(){return this._started},t.prototype.play=function(){this.parentPlayer||this.init(),this._onStart(),this.players.forEach(function(t){return t.play()})},t.prototype.pause=function(){this.players.forEach(function(t){return t.pause()})},t.prototype.restart=function(){this.players.forEach(function(t){return t.restart()})},t.prototype.finish=function(){this._onFinish(),this.players.forEach(function(t){return t.finish()})},t.prototype.destroy=function(){this._onDestroy()},t.prototype._onDestroy=function(){this._destroyed||(this._destroyed=!0,this._onFinish(),this.players.forEach(function(t){return t.destroy()}),this._onDestroyFns.forEach(function(t){return t()}),this._onDestroyFns=[])},t.prototype.reset=function(){this.players.forEach(function(t){return t.reset()}),this._destroyed=!1,this._finished=!1,this._started=!1},t.prototype.setPosition=function(t){var e=t*this.totalTime;this.players.forEach(function(t){var n=t.totalTime?Math.min(1,e/t.totalTime):1;t.setPosition(n)})},t.prototype.getPosition=function(){var t=0;return this.players.forEach(function(e){var n=e.getPosition();t=Math.min(n,t)}),t},t.prototype.beforeDestroy=function(){this.players.forEach(function(t){t.beforeDestroy&&t.beforeDestroy()})},t.prototype.triggerCallback=function(t){var e="start"==t?this._onStartFns:this._onDoneFns;e.forEach(function(t){return t()}),e.length=0},t}(),w="!"},"./node_modules/@angular/animations/fesm5/browser.js":function(t,e,n){"use strict";n.r(e),n.d(e,"AnimationDriver",function(){return M}),n.d(e,"ɵAnimation",function(){return Et}),n.d(e,"ɵAnimationStyleNormalizer",function(){return St}),n.d(e,"ɵNoopAnimationStyleNormalizer",function(){return Mt}),n.d(e,"ɵWebAnimationsStyleNormalizer",function(){return Tt}),n.d(e,"ɵAnimationDriver",function(){return M}),n.d(e,"ɵNoopAnimationDriver",function(){return S}),n.d(e,"ɵAnimationEngine",function(){return oe}),n.d(e,"ɵCssKeyframesDriver",function(){return _e}),n.d(e,"ɵCssKeyframesPlayer",function(){return ge}),n.d(e,"ɵcontainsElement",function(){return O}),n.d(e,"ɵinvokeQuery",function(){return C}),n.d(e,"ɵmatchesElement",function(){return j}),n.d(e,"ɵvalidateStyleProperty",function(){return w}),n.d(e,"ɵWebAnimationsDriver",function(){return be}),n.d(e,"ɵsupportsWebAnimations",function(){return xe}),n.d(e,"ɵWebAnimationsPlayer",function(){return ve}),n.d(e,"ɵallowPreviousPlayerStylesMerge",function(){return Q});var r=n("./node_modules/@angular/animations/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/animations/fesm5/animations.js"),o=n("./node_modules/@angular/core/fesm5/core.js");function a(){return"undefined"!=typeof process}function s(t){switch(t.length){case 0:return new i.NoopAnimationPlayer;case 1:return t[0];default:return new i["ɵAnimationGroupPlayer"](t)}}function u(t,e,n,r,o,a){void 0===o&&(o={}),void 0===a&&(a={});var s=[],u=[],c=-1,l=null;if(r.forEach(function(t){var n=t.offset,r=n==c,d=r&&l||{};Object.keys(t).forEach(function(n){var r=n,u=t[n];if("offset"!==n)switch(r=e.normalizePropertyName(r,s),u){case i["ɵPRE_STYLE"]:u=o[n];break;case i.AUTO_STYLE:u=a[n];break;default:u=e.normalizeStyleValue(n,r,u,s)}d[r]=u}),r||u.push(d),l=d,c=n}),s.length){throw new Error("Unable to animate due to the following errors:\n - "+s.join("\n - "))}return u}function c(t,e,n,r){switch(e){case"start":t.onStart(function(){return r(n&&l(n,"start",t))});break;case"done":t.onDone(function(){return r(n&&l(n,"done",t))});break;case"destroy":t.onDestroy(function(){return r(n&&l(n,"destroy",t))})}}function l(t,e,n){var r=n.totalTime,i=!!n.disabled,o=d(t.element,t.triggerName,t.fromState,t.toState,e||t.phaseName,void 0==r?t.totalTime:r,i),a=t._data;return null!=a&&(o._data=a),o}function d(t,e,n,r,i,o,a){return void 0===i&&(i=""),void 0===o&&(o=0),{element:t,triggerName:e,fromState:n,toState:r,phaseName:i,totalTime:o,disabled:!!a}}function p(t,e,n){var r;return t instanceof Map?(r=t.get(e))||t.set(e,r=n):(r=t[e])||(r=t[e]=n),r}function f(t){var e=t.indexOf(":");return[t.substring(1,e),t.substr(e+1)]}var h=function(t,e){return!1},m=function(t,e){return!1},g=function(t,e,n){return[]},y=a();if(y||"undefined"!=typeof Element){if(h=function(t,e){return t.contains(e)},y||Element.prototype.matches)m=function(t,e){return t.matches(e)};else{var _=Element.prototype,v=_.matchesSelector||_.mozMatchesSelector||_.msMatchesSelector||_.oMatchesSelector||_.webkitMatchesSelector;v&&(m=function(t,e){return v.apply(t,[e])})}g=function(t,e,n){var i=[];if(n)i.push.apply(i,Object(r.__spread)(t.querySelectorAll(e)));else{var o=t.querySelector(e);o&&i.push(o)}return i}}var b=null,x=!1;function w(t){b||(b=function(){if("undefined"!=typeof document)return document.body;return null}()||{},x=!!b.style&&"WebkitAppearance"in b.style);var e=!0;b.style&&"ebkit"!=t.substring(1,6)&&(!(e=t in b.style)&&x&&(e="Webkit"+t.charAt(0).toUpperCase()+t.substr(1)in b.style));return e}var j=m,O=h,C=g;function E(t){var e={};return Object.keys(t).forEach(function(n){var r=n.replace(/([a-z])([A-Z])/g,"$1-$2");e[r]=t[n]}),e}var S=function(){function t(){}return t.prototype.validateStyleProperty=function(t){return w(t)},t.prototype.matchesElement=function(t,e){return j(t,e)},t.prototype.containsElement=function(t,e){return O(t,e)},t.prototype.query=function(t,e,n){return C(t,e,n)},t.prototype.computeStyle=function(t,e,n){return n||""},t.prototype.animate=function(t,e,n,r,o,a,s){return void 0===a&&(a=[]),new i.NoopAnimationPlayer(n,r)},t=Object(r.__decorate)([Object(o.Injectable)()],t)}(),M=function(){function t(){}return t.NOOP=new S,t}(),T=1e3,k="ng-trigger",A=".ng-trigger",D=".ng-animating";function I(t){if("number"==typeof t)return t;var e=t.match(/^(-?[\.\d]+)(m?s)/);return!e||e.length<2?0:P(parseFloat(e[1]),e[2])}function P(t,e){switch(e){case"s":return t*T;default:return t}}function R(t,e,n){return t.hasOwnProperty("duration")?t:function(t,e,n){var r,i=0,o="";if("string"==typeof t){var a=t.match(/^(-?[\.\d]+)(m?s)(?:\s+(-?[\.\d]+)(m?s))?(?:\s+([-a-z]+(?:\(.+?\))?))?$/i);if(null===a)return e.push('The provided timing value "'+t+'" is invalid.'),{duration:0,delay:0,easing:""};r=P(parseFloat(a[1]),a[2]);var s=a[3];null!=s&&(i=P(parseFloat(s),a[4]));var u=a[5];u&&(o=u)}else r=t;if(!n){var c=!1,l=e.length;r<0&&(e.push("Duration values below 0 are not allowed for this animation step."),c=!0),i<0&&(e.push("Delay values below 0 are not allowed for this animation step."),c=!0),c&&e.splice(l,0,'The provided timing value "'+t+'" is invalid.')}return{duration:r,delay:i,easing:o}}(t,e,n)}function N(t,e){return void 0===e&&(e={}),Object.keys(t).forEach(function(n){e[n]=t[n]}),e}function F(t){var e={};return Array.isArray(t)?t.forEach(function(t){return L(t,!1,e)}):L(t,!1,e),e}function L(t,e,n){if(void 0===n&&(n={}),e)for(var r in t)n[r]=t[r];else N(t,n);return n}function B(t,e,n){return n?e+":"+n+";":""}function V(t){for(var e="",n=0;n<t.style.length;n++){e+=B(0,r=t.style.item(n),t.style.getPropertyValue(r))}for(var r in t.style){if(t.style.hasOwnProperty(r)&&!r.startsWith("_"))e+=B(0,r.replace(/([a-z])([A-Z])/g,"$1-$2").toLowerCase(),t.style[r])}t.setAttribute("style",e)}function z(t,e){t.style&&(Object.keys(e).forEach(function(n){var r=X(n);t.style[r]=e[n]}),a()&&V(t))}function H(t,e){t.style&&(Object.keys(e).forEach(function(e){var n=X(e);t.style[n]=""}),a()&&V(t))}function U(t){return Array.isArray(t)?1==t.length?t[0]:Object(i.sequence)(t):t}var W=new RegExp("{{\\s*(.+?)\\s*}}","g");function q(t){var e=[];if("string"==typeof t){for(var n=t.toString(),r=void 0;r=W.exec(n);)e.push(r[1]);W.lastIndex=0}return e}function G(t,e,n){var r=t.toString(),i=r.replace(W,function(t,r){var i=e[r];return e.hasOwnProperty(r)||(n.push("Please provide a value for the animation param "+r),i=""),i.toString()});return i==r?t:i}function Y(t){for(var e=[],n=t.next();!n.done;)e.push(n.value),n=t.next();return e}var K=/-+([a-z0-9])/g;function X(t){return t.replace(K,function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return t[1].toUpperCase()})}function Q(t,e){return 0===t||0===e}function Z(t,e,n){var r=Object.keys(n);if(r.length&&e.length){var i=e[0],o=[];if(r.forEach(function(t){i.hasOwnProperty(t)||o.push(t),i[t]=n[t]}),o.length)for(var a=function(){var n=e[s];o.forEach(function(e){n[e]=J(t,e)})},s=1;s<e.length;s++)a()}return e}function $(t,e,n){switch(e.type){case 7:return t.visitTrigger(e,n);case 0:return t.visitState(e,n);case 1:return t.visitTransition(e,n);case 2:return t.visitSequence(e,n);case 3:return t.visitGroup(e,n);case 4:return t.visitAnimate(e,n);case 5:return t.visitKeyframes(e,n);case 6:return t.visitStyle(e,n);case 8:return t.visitReference(e,n);case 9:return t.visitAnimateChild(e,n);case 10:return t.visitAnimateRef(e,n);case 11:return t.visitQuery(e,n);case 12:return t.visitStagger(e,n);default:throw new Error("Unable to resolve animation metadata node #"+e.type)}}function J(t,e){return window.getComputedStyle(t)[e]}var tt="*";function et(t,e){var n=[];return"string"==typeof t?t.split(/\s*,\s*/).forEach(function(t){return function(t,e,n){if(":"==t[0]){var r=function(t,e){switch(t){case":enter":return"void => *";case":leave":return"* => void";case":increment":return function(t,e){return parseFloat(e)>parseFloat(t)};case":decrement":return function(t,e){return parseFloat(e)<parseFloat(t)};default:return e.push('The transition alias value "'+t+'" is not supported'),"* => *"}}(t,n);if("function"==typeof r)return void e.push(r);t=r}var i=t.match(/^(\*|[-\w]+)\s*(<?[=-]>)\s*(\*|[-\w]+)$/);if(null==i||i.length<4)return n.push('The provided transition expression "'+t+'" is not supported'),e;var o=i[1],a=i[2],s=i[3];e.push(it(o,s));var u=o==tt&&s==tt;"<"!=a[0]||u||e.push(it(s,o))}(t,n,e)}):n.push(t),n}var nt=new Set(["true","1"]),rt=new Set(["false","0"]);function it(t,e){var n=nt.has(t)||rt.has(t),r=nt.has(e)||rt.has(e);return function(i,o){var a=t==tt||t==i,s=e==tt||e==o;return!a&&n&&"boolean"==typeof i&&(a=i?nt.has(t):rt.has(t)),!s&&r&&"boolean"==typeof o&&(s=o?nt.has(e):rt.has(e)),a&&s}}var ot=":self",at=new RegExp("s*"+ot+"s*,?","g");function st(t,e,n){return new ut(t).build(e,n)}var ut=function(){function t(t){this._driver=t}return t.prototype.build=function(t,e){var n=new ct(e);return this._resetContextStyleTimingState(n),$(this,U(t),n)},t.prototype._resetContextStyleTimingState=function(t){t.currentQuerySelector="",t.collectedStyles={},t.collectedStyles[""]={},t.currentTime=0},t.prototype.visitTrigger=function(t,e){var n=this,r=e.queryCount=0,i=e.depCount=0,o=[],a=[];return"@"==t.name.charAt(0)&&e.errors.push("animation triggers cannot be prefixed with an `@` sign (e.g. trigger('@foo', [...]))"),t.definitions.forEach(function(t){if(n._resetContextStyleTimingState(e),0==t.type){var s=t,u=s.name;u.toString().split(/\s*,\s*/).forEach(function(t){s.name=t,o.push(n.visitState(s,e))}),s.name=u}else if(1==t.type){var c=n.visitTransition(t,e);r+=c.queryCount,i+=c.depCount,a.push(c)}else e.errors.push("only state() and transition() definitions can sit inside of a trigger()")}),{type:7,name:t.name,states:o,transitions:a,queryCount:r,depCount:i,options:null}},t.prototype.visitState=function(t,e){var n=this.visitStyle(t.styles,e),r=t.options&&t.options.params||null;if(n.containsDynamicStyles){var i=new Set,o=r||{};if(n.styles.forEach(function(t){if(lt(t)){var e=t;Object.keys(e).forEach(function(t){q(e[t]).forEach(function(t){o.hasOwnProperty(t)||i.add(t)})})}}),i.size){var a=Y(i.values());e.errors.push('state("'+t.name+'", ...) must define default values for all the following style substitutions: '+a.join(", "))}}return{type:0,name:t.name,style:n,options:r?{params:r}:null}},t.prototype.visitTransition=function(t,e){e.queryCount=0,e.depCount=0;var n=$(this,U(t.animation),e);return{type:1,matchers:et(t.expr,e.errors),animation:n,queryCount:e.queryCount,depCount:e.depCount,options:dt(t.options)}},t.prototype.visitSequence=function(t,e){var n=this;return{type:2,steps:t.steps.map(function(t){return $(n,t,e)}),options:dt(t.options)}},t.prototype.visitGroup=function(t,e){var n=this,r=e.currentTime,i=0,o=t.steps.map(function(t){e.currentTime=r;var o=$(n,t,e);return i=Math.max(i,e.currentTime),o});return e.currentTime=i,{type:3,steps:o,options:dt(t.options)}},t.prototype.visitAnimate=function(t,e){var n,r=function(t,e){var n=null;if(t.hasOwnProperty("duration"))n=t;else if("number"==typeof t){var r=R(t,e).duration;return pt(r,0,"")}var i=t;if(i.split(/\s+/).some(function(t){return"{"==t.charAt(0)&&"{"==t.charAt(1)})){var o=pt(0,0,"");return o.dynamic=!0,o.strValue=i,o}return pt((n=n||R(i,e)).duration,n.delay,n.easing)}(t.timings,e.errors);e.currentAnimateTimings=r;var o=t.styles?t.styles:Object(i.style)({});if(5==o.type)n=this.visitKeyframes(o,e);else{var a=t.styles,s=!1;if(!a){s=!0;var u={};r.easing&&(u.easing=r.easing),a=Object(i.style)(u)}e.currentTime+=r.duration+r.delay;var c=this.visitStyle(a,e);c.isEmptyStep=s,n=c}return e.currentAnimateTimings=null,{type:4,timings:r,style:n,options:null}},t.prototype.visitStyle=function(t,e){var n=this._makeStyleAst(t,e);return this._validateStyleAst(n,e),n},t.prototype._makeStyleAst=function(t,e){var n=[];Array.isArray(t.styles)?t.styles.forEach(function(t){"string"==typeof t?t==i.AUTO_STYLE?n.push(t):e.errors.push("The provided style string value "+t+" is not allowed."):n.push(t)}):n.push(t.styles);var r=!1,o=null;return n.forEach(function(t){if(lt(t)){var e=t,n=e.easing;if(n&&(o=n,delete e.easing),!r)for(var i in e){if(e[i].toString().indexOf("{{")>=0){r=!0;break}}}}),{type:6,styles:n,easing:o,offset:t.offset,containsDynamicStyles:r,options:null}},t.prototype._validateStyleAst=function(t,e){var n=this,r=e.currentAnimateTimings,i=e.currentTime,o=e.currentTime;r&&o>0&&(o-=r.duration+r.delay),t.styles.forEach(function(t){"string"!=typeof t&&Object.keys(t).forEach(function(r){if(n._driver.validateStyleProperty(r)){var a,s,u,c,l,d=e.collectedStyles[e.currentQuerySelector],p=d[r],f=!0;p&&(o!=i&&o>=p.startTime&&i<=p.endTime&&(e.errors.push('The CSS property "'+r+'" that exists between the times of "'+p.startTime+'ms" and "'+p.endTime+'ms" is also being animated in a parallel animation between the times of "'+o+'ms" and "'+i+'ms"'),f=!1),o=p.startTime),f&&(d[r]={startTime:o,endTime:i}),e.options&&(a=t[r],s=e.options,u=e.errors,c=s.params||{},(l=q(a)).length&&l.forEach(function(t){c.hasOwnProperty(t)||u.push("Unable to resolve the local animation param "+t+" in the given list of values")}))}else e.errors.push('The provided animation property "'+r+'" is not a supported CSS property for animations')})})},t.prototype.visitKeyframes=function(t,e){var n=this,r={type:5,styles:[],options:null};if(!e.currentAnimateTimings)return e.errors.push("keyframes() must be placed inside of a call to animate()"),r;var i=0,o=[],a=!1,s=!1,u=0,c=t.steps.map(function(t){var r=n._makeStyleAst(t,e),c=null!=r.offset?r.offset:function(t){if("string"==typeof t)return null;var e=null;if(Array.isArray(t))t.forEach(function(t){if(lt(t)&&t.hasOwnProperty("offset")){var n=t;e=parseFloat(n.offset),delete n.offset}});else if(lt(t)&&t.hasOwnProperty("offset")){var n=t;e=parseFloat(n.offset),delete n.offset}return e}(r.styles),l=0;return null!=c&&(i++,l=r.offset=c),s=s||l<0||l>1,a=a||l<u,u=l,o.push(l),r});s&&e.errors.push("Please ensure that all keyframe offsets are between 0 and 1"),a&&e.errors.push("Please ensure that all keyframe offsets are in order");var l=t.steps.length,d=0;i>0&&i<l?e.errors.push("Not all style() steps within the declared keyframes() contain offsets"):0==i&&(d=1/(l-1));var p=l-1,f=e.currentTime,h=e.currentAnimateTimings,m=h.duration;return c.forEach(function(t,i){var a=d>0?i==p?1:d*i:o[i],s=a*m;e.currentTime=f+h.delay+s,h.duration=s,n._validateStyleAst(t,e),t.offset=a,r.styles.push(t)}),r},t.prototype.visitReference=function(t,e){return{type:8,animation:$(this,U(t.animation),e),options:dt(t.options)}},t.prototype.visitAnimateChild=function(t,e){return e.depCount++,{type:9,options:dt(t.options)}},t.prototype.visitAnimateRef=function(t,e){return{type:10,animation:this.visitReference(t.animation,e),options:dt(t.options)}},t.prototype.visitQuery=function(t,e){var n=e.currentQuerySelector,i=t.options||{};e.queryCount++,e.currentQuery=t;var o=Object(r.__read)(function(t){var e=!!t.split(/\s*,\s*/).find(function(t){return t==ot});e&&(t=t.replace(at,""));return[t=t.replace(/@\*/g,A).replace(/@\w+/g,function(t){return A+"-"+t.substr(1)}).replace(/:animating/g,D),e]}(t.selector),2),a=o[0],s=o[1];e.currentQuerySelector=n.length?n+" "+a:a,p(e.collectedStyles,e.currentQuerySelector,{});var u=$(this,U(t.animation),e);return e.currentQuery=null,e.currentQuerySelector=n,{type:11,selector:a,limit:i.limit||0,optional:!!i.optional,includeSelf:s,animation:u,originalSelector:t.selector,options:dt(t.options)}},t.prototype.visitStagger=function(t,e){e.currentQuery||e.errors.push("stagger() can only be used inside of query()");var n="full"===t.timings?{duration:0,delay:0,easing:"full"}:R(t.timings,e.errors,!0);return{type:12,animation:$(this,U(t.animation),e),timings:n,options:null}},t}();var ct=function(){return function(t){this.errors=t,this.queryCount=0,this.depCount=0,this.currentTransition=null,this.currentQuery=null,this.currentQuerySelector=null,this.currentAnimateTimings=null,this.currentTime=0,this.collectedStyles={},this.options=null}}();function lt(t){return!Array.isArray(t)&&"object"==typeof t}function dt(t){var e;return t?(t=N(t)).params&&(t.params=(e=t.params)?N(e):null):t={},t}function pt(t,e,n){return{duration:t,delay:e,easing:n}}function ft(t,e,n,r,i,o,a,s){return void 0===a&&(a=null),void 0===s&&(s=!1),{type:1,element:t,keyframes:e,preStyleProps:n,postStyleProps:r,duration:i,delay:o,totalTime:i+o,easing:a,subTimeline:s}}var ht=function(){function t(){this._map=new Map}return t.prototype.consume=function(t){var e=this._map.get(t);return e?this._map.delete(t):e=[],e},t.prototype.append=function(t,e){var n=this._map.get(t);n||this._map.set(t,n=[]),n.push.apply(n,Object(r.__spread)(e))},t.prototype.has=function(t){return this._map.has(t)},t.prototype.clear=function(){this._map.clear()},t}(),mt=new RegExp(":enter","g"),gt=new RegExp(":leave","g");function yt(t,e,n,r,i,o,a,s,u,c){return void 0===o&&(o={}),void 0===a&&(a={}),void 0===c&&(c=[]),(new _t).buildKeyframes(t,e,n,r,i,o,a,s,u,c)}var _t=function(){function t(){}return t.prototype.buildKeyframes=function(t,e,n,r,i,o,a,s,u,c){void 0===c&&(c=[]),u=u||new ht;var l=new bt(t,e,u,r,i,c,[]);l.options=s,l.currentTimeline.setStyles([o],null,l.errors,s),$(this,n,l);var d=l.timelines.filter(function(t){return t.containsAnimation()});if(d.length&&Object.keys(a).length){var p=d[d.length-1];p.allowOnlyTimelineStyles()||p.setStyles([a],null,l.errors,s)}return d.length?d.map(function(t){return t.buildKeyframes()}):[ft(e,[],[],[],0,0,"",!1)]},t.prototype.visitTrigger=function(t,e){},t.prototype.visitState=function(t,e){},t.prototype.visitTransition=function(t,e){},t.prototype.visitAnimateChild=function(t,e){var n=e.subInstructions.consume(e.element);if(n){var r=e.createSubContext(t.options),i=e.currentTimeline.currentTime,o=this._visitSubInstructions(n,r,r.options);i!=o&&e.transformIntoNewTimeline(o)}e.previousNode=t},t.prototype.visitAnimateRef=function(t,e){var n=e.createSubContext(t.options);n.transformIntoNewTimeline(),this.visitReference(t.animation,n),e.transformIntoNewTimeline(n.currentTimeline.currentTime),e.previousNode=t},t.prototype._visitSubInstructions=function(t,e,n){var r=e.currentTimeline.currentTime,i=null!=n.duration?I(n.duration):null,o=null!=n.delay?I(n.delay):null;return 0!==i&&t.forEach(function(t){var n=e.appendInstructionToTimeline(t,i,o);r=Math.max(r,n.duration+n.delay)}),r},t.prototype.visitReference=function(t,e){e.updateOptions(t.options,!0),$(this,t.animation,e),e.previousNode=t},t.prototype.visitSequence=function(t,e){var n=this,r=e.subContextCount,i=e,o=t.options;if(o&&(o.params||o.delay)&&((i=e.createSubContext(o)).transformIntoNewTimeline(),null!=o.delay)){6==i.previousNode.type&&(i.currentTimeline.snapshotCurrentStyles(),i.previousNode=vt);var a=I(o.delay);i.delayNextStep(a)}t.steps.length&&(t.steps.forEach(function(t){return $(n,t,i)}),i.currentTimeline.applyStylesToKeyframe(),i.subContextCount>r&&i.transformIntoNewTimeline()),e.previousNode=t},t.prototype.visitGroup=function(t,e){var n=this,r=[],i=e.currentTimeline.currentTime,o=t.options&&t.options.delay?I(t.options.delay):0;t.steps.forEach(function(a){var s=e.createSubContext(t.options);o&&s.delayNextStep(o),$(n,a,s),i=Math.max(i,s.currentTimeline.currentTime),r.push(s.currentTimeline)}),r.forEach(function(t){return e.currentTimeline.mergeTimelineCollectedStyles(t)}),e.transformIntoNewTimeline(i),e.previousNode=t},t.prototype._visitTiming=function(t,e){if(t.dynamic){var n=t.strValue;return R(e.params?G(n,e.params,e.errors):n,e.errors)}return{duration:t.duration,delay:t.delay,easing:t.easing}},t.prototype.visitAnimate=function(t,e){var n=e.currentAnimateTimings=this._visitTiming(t.timings,e),r=e.currentTimeline;n.delay&&(e.incrementTime(n.delay),r.snapshotCurrentStyles());var i=t.style;5==i.type?this.visitKeyframes(i,e):(e.incrementTime(n.duration),this.visitStyle(i,e),r.applyStylesToKeyframe()),e.currentAnimateTimings=null,e.previousNode=t},t.prototype.visitStyle=function(t,e){var n=e.currentTimeline,r=e.currentAnimateTimings;!r&&n.getCurrentStyleProperties().length&&n.forwardFrame();var i=r&&r.easing||t.easing;t.isEmptyStep?n.applyEmptyStep(i):n.setStyles(t.styles,i,e.errors,e.options),e.previousNode=t},t.prototype.visitKeyframes=function(t,e){var n=e.currentAnimateTimings,r=e.currentTimeline.duration,i=n.duration,o=e.createSubContext().currentTimeline;o.easing=n.easing,t.styles.forEach(function(t){var n=t.offset||0;o.forwardTime(n*i),o.setStyles(t.styles,t.easing,e.errors,e.options),o.applyStylesToKeyframe()}),e.currentTimeline.mergeTimelineCollectedStyles(o),e.transformIntoNewTimeline(r+i),e.previousNode=t},t.prototype.visitQuery=function(t,e){var n=this,r=e.currentTimeline.currentTime,i=t.options||{},o=i.delay?I(i.delay):0;o&&(6===e.previousNode.type||0==r&&e.currentTimeline.getCurrentStyleProperties().length)&&(e.currentTimeline.snapshotCurrentStyles(),e.previousNode=vt);var a=r,s=e.invokeQuery(t.selector,t.originalSelector,t.limit,t.includeSelf,!!i.optional,e.errors);e.currentQueryTotal=s.length;var u=null;s.forEach(function(r,i){e.currentQueryIndex=i;var s=e.createSubContext(t.options,r);o&&s.delayNextStep(o),r===e.element&&(u=s.currentTimeline),$(n,t.animation,s),s.currentTimeline.applyStylesToKeyframe();var c=s.currentTimeline.currentTime;a=Math.max(a,c)}),e.currentQueryIndex=0,e.currentQueryTotal=0,e.transformIntoNewTimeline(a),u&&(e.currentTimeline.mergeTimelineCollectedStyles(u),e.currentTimeline.snapshotCurrentStyles()),e.previousNode=t},t.prototype.visitStagger=function(t,e){var n=e.parentContext,r=e.currentTimeline,i=t.timings,o=Math.abs(i.duration),a=o*(e.currentQueryTotal-1),s=o*e.currentQueryIndex;switch(i.duration<0?"reverse":i.easing){case"reverse":s=a-s;break;case"full":s=n.currentStaggerTime}var u=e.currentTimeline;s&&u.delayNextStep(s);var c=u.currentTime;$(this,t.animation,e),e.previousNode=t,n.currentStaggerTime=r.currentTime-c+(r.startTime-n.currentTimeline.startTime)},t}(),vt={},bt=function(){function t(t,e,n,r,i,o,a,s){this._driver=t,this.element=e,this.subInstructions=n,this._enterClassName=r,this._leaveClassName=i,this.errors=o,this.timelines=a,this.parentContext=null,this.currentAnimateTimings=null,this.previousNode=vt,this.subContextCount=0,this.options={},this.currentQueryIndex=0,this.currentQueryTotal=0,this.currentStaggerTime=0,this.currentTimeline=s||new xt(this._driver,e,0),a.push(this.currentTimeline)}return Object.defineProperty(t.prototype,"params",{get:function(){return this.options.params},enumerable:!0,configurable:!0}),t.prototype.updateOptions=function(t,e){var n=this;if(t){var r=t,i=this.options;null!=r.duration&&(i.duration=I(r.duration)),null!=r.delay&&(i.delay=I(r.delay));var o=r.params;if(o){var a=i.params;a||(a=this.options.params={}),Object.keys(o).forEach(function(t){e&&a.hasOwnProperty(t)||(a[t]=G(o[t],a,n.errors))})}}},t.prototype._copyOptions=function(){var t={};if(this.options){var e=this.options.params;if(e){var n=t.params={};Object.keys(e).forEach(function(t){n[t]=e[t]})}}return t},t.prototype.createSubContext=function(e,n,r){void 0===e&&(e=null);var i=n||this.element,o=new t(this._driver,i,this.subInstructions,this._enterClassName,this._leaveClassName,this.errors,this.timelines,this.currentTimeline.fork(i,r||0));return o.previousNode=this.previousNode,o.currentAnimateTimings=this.currentAnimateTimings,o.options=this._copyOptions(),o.updateOptions(e),o.currentQueryIndex=this.currentQueryIndex,o.currentQueryTotal=this.currentQueryTotal,o.parentContext=this,this.subContextCount++,o},t.prototype.transformIntoNewTimeline=function(t){return this.previousNode=vt,this.currentTimeline=this.currentTimeline.fork(this.element,t),this.timelines.push(this.currentTimeline),this.currentTimeline},t.prototype.appendInstructionToTimeline=function(t,e,n){var r={duration:null!=e?e:t.duration,delay:this.currentTimeline.currentTime+(null!=n?n:0)+t.delay,easing:""},i=new wt(this._driver,t.element,t.keyframes,t.preStyleProps,t.postStyleProps,r,t.stretchStartingKeyframe);return this.timelines.push(i),r},t.prototype.incrementTime=function(t){this.currentTimeline.forwardTime(this.currentTimeline.duration+t)},t.prototype.delayNextStep=function(t){t>0&&this.currentTimeline.delayNextStep(t)},t.prototype.invokeQuery=function(t,e,n,i,o,a){var s=[];if(i&&s.push(this.element),t.length>0){t=(t=t.replace(mt,"."+this._enterClassName)).replace(gt,"."+this._leaveClassName);var u=1!=n,c=this._driver.query(this.element,t,u);0!==n&&(c=n<0?c.slice(c.length+n,c.length):c.slice(0,n)),s.push.apply(s,Object(r.__spread)(c))}return o||0!=s.length||a.push('`query("'+e+'")` returned zero elements. (Use `query("'+e+'", { optional: true })` if you wish to allow this.)'),s},t}(),xt=function(){function t(t,e,n,r){this._driver=t,this.element=e,this.startTime=n,this._elementTimelineStylesLookup=r,this.duration=0,this._previousKeyframe={},this._currentKeyframe={},this._keyframes=new Map,this._styleSummary={},this._pendingStyles={},this._backFill={},this._currentEmptyStepKeyframe=null,this._elementTimelineStylesLookup||(this._elementTimelineStylesLookup=new Map),this._localTimelineStyles=Object.create(this._backFill,{}),this._globalTimelineStyles=this._elementTimelineStylesLookup.get(e),this._globalTimelineStyles||(this._globalTimelineStyles=this._localTimelineStyles,this._elementTimelineStylesLookup.set(e,this._localTimelineStyles)),this._loadKeyframe()}return t.prototype.containsAnimation=function(){switch(this._keyframes.size){case 0:return!1;case 1:return this.getCurrentStyleProperties().length>0;default:return!0}},t.prototype.getCurrentStyleProperties=function(){return Object.keys(this._currentKeyframe)},Object.defineProperty(t.prototype,"currentTime",{get:function(){return this.startTime+this.duration},enumerable:!0,configurable:!0}),t.prototype.delayNextStep=function(t){var e=1==this._keyframes.size&&Object.keys(this._pendingStyles).length;this.duration||e?(this.forwardTime(this.currentTime+t),e&&this.snapshotCurrentStyles()):this.startTime+=t},t.prototype.fork=function(e,n){return this.applyStylesToKeyframe(),new t(this._driver,e,n||this.currentTime,this._elementTimelineStylesLookup)},t.prototype._loadKeyframe=function(){this._currentKeyframe&&(this._previousKeyframe=this._currentKeyframe),this._currentKeyframe=this._keyframes.get(this.duration),this._currentKeyframe||(this._currentKeyframe=Object.create(this._backFill,{}),this._keyframes.set(this.duration,this._currentKeyframe))},t.prototype.forwardFrame=function(){this.duration+=1,this._loadKeyframe()},t.prototype.forwardTime=function(t){this.applyStylesToKeyframe(),this.duration=t,this._loadKeyframe()},t.prototype._updateStyle=function(t,e){this._localTimelineStyles[t]=e,this._globalTimelineStyles[t]=e,this._styleSummary[t]={time:this.currentTime,value:e}},t.prototype.allowOnlyTimelineStyles=function(){return this._currentEmptyStepKeyframe!==this._currentKeyframe},t.prototype.applyEmptyStep=function(t){var e=this;t&&(this._previousKeyframe.easing=t),Object.keys(this._globalTimelineStyles).forEach(function(t){e._backFill[t]=e._globalTimelineStyles[t]||i.AUTO_STYLE,e._currentKeyframe[t]=i.AUTO_STYLE}),this._currentEmptyStepKeyframe=this._currentKeyframe},t.prototype.setStyles=function(t,e,n,r){var o=this;e&&(this._previousKeyframe.easing=e);var a,s,u,c,l=r&&r.params||{},d=(a=t,s=this._globalTimelineStyles,c={},a.forEach(function(t){"*"===t?(u=u||Object.keys(s)).forEach(function(t){c[t]=i.AUTO_STYLE}):L(t,!1,c)}),c);Object.keys(d).forEach(function(t){var e=G(d[t],l,n);o._pendingStyles[t]=e,o._localTimelineStyles.hasOwnProperty(t)||(o._backFill[t]=o._globalTimelineStyles.hasOwnProperty(t)?o._globalTimelineStyles[t]:i.AUTO_STYLE),o._updateStyle(t,e)})},t.prototype.applyStylesToKeyframe=function(){var t=this,e=this._pendingStyles,n=Object.keys(e);0!=n.length&&(this._pendingStyles={},n.forEach(function(n){var r=e[n];t._currentKeyframe[n]=r}),Object.keys(this._localTimelineStyles).forEach(function(e){t._currentKeyframe.hasOwnProperty(e)||(t._currentKeyframe[e]=t._localTimelineStyles[e])}))},t.prototype.snapshotCurrentStyles=function(){var t=this;Object.keys(this._localTimelineStyles).forEach(function(e){var n=t._localTimelineStyles[e];t._pendingStyles[e]=n,t._updateStyle(e,n)})},t.prototype.getFinalKeyframe=function(){return this._keyframes.get(this.duration)},Object.defineProperty(t.prototype,"properties",{get:function(){var t=[];for(var e in this._currentKeyframe)t.push(e);return t},enumerable:!0,configurable:!0}),t.prototype.mergeTimelineCollectedStyles=function(t){var e=this;Object.keys(t._styleSummary).forEach(function(n){var r=e._styleSummary[n],i=t._styleSummary[n];(!r||i.time>r.time)&&e._updateStyle(n,i.value)})},t.prototype.buildKeyframes=function(){var t=this;this.applyStylesToKeyframe();var e=new Set,n=new Set,r=1===this._keyframes.size&&0===this.duration,o=[];this._keyframes.forEach(function(a,s){var u=L(a,!0);Object.keys(u).forEach(function(t){var r=u[t];r==i["ɵPRE_STYLE"]?e.add(t):r==i.AUTO_STYLE&&n.add(t)}),r||(u.offset=s/t.duration),o.push(u)});var a=e.size?Y(e.values()):[],s=n.size?Y(n.values()):[];if(r){var u=o[0],c=N(u);u.offset=0,c.offset=1,o=[u,c]}return ft(this.element,o,a,s,this.duration,this.startTime,this.easing,!1)},t}(),wt=function(t){function e(e,n,r,i,o,a,s){void 0===s&&(s=!1);var u=t.call(this,e,n,a.delay)||this;return u.element=n,u.keyframes=r,u.preStyleProps=i,u.postStyleProps=o,u._stretchStartingKeyframe=s,u.timings={duration:a.duration,delay:a.delay,easing:a.easing},u}return Object(r.__extends)(e,t),e.prototype.containsAnimation=function(){return this.keyframes.length>1},e.prototype.buildKeyframes=function(){var t=this.keyframes,e=this.timings,n=e.delay,r=e.duration,i=e.easing;if(this._stretchStartingKeyframe&&n){var o=[],a=r+n,s=n/a,u=L(t[0],!1);u.offset=0,o.push(u);var c=L(t[0],!1);c.offset=jt(s),o.push(c);for(var l=t.length-1,d=1;d<=l;d++){var p=L(t[d],!1),f=n+p.offset*r;p.offset=jt(f/a),o.push(p)}r=a,n=0,i="",t=o}return ft(this.element,t,this.preStyleProps,this.postStyleProps,r,n,i,!0)},e}(xt);function jt(t,e){void 0===e&&(e=3);var n=Math.pow(10,e-1);return Math.round(t*n)/n}var Ot,Ct,Et=function(){function t(t,e){this._driver=t;var n=[],r=st(t,e,n);if(n.length){var i="animation validation failed:\n"+n.join("\n");throw new Error(i)}this._animationAst=r}return t.prototype.buildTimelines=function(t,e,n,r,i){var o=Array.isArray(e)?F(e):e,a=Array.isArray(n)?F(n):n,s=[];i=i||new ht;var u=yt(this._driver,t,this._animationAst,"ng-enter","ng-leave",o,a,r,i,s);if(s.length){var c="animation building failed:\n"+s.join("\n");throw new Error(c)}return u},t}(),St=function(){return function(){}}(),Mt=function(){function t(){}return t.prototype.normalizePropertyName=function(t,e){return t},t.prototype.normalizeStyleValue=function(t,e,n,r){return n},t}(),Tt=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.normalizePropertyName=function(t,e){return X(t)},e.prototype.normalizeStyleValue=function(t,e,n,r){var i="",o=n.toString().trim();if(kt[e]&&0!==n&&"0"!==n)if("number"==typeof n)i="px";else{var a=n.match(/^[+-]?[\d\.]+([a-z]*)$/);a&&0==a[1].length&&r.push("Please provide a CSS unit value for "+t+":"+n)}return o+i},e}(St),kt=(Ot="width,height,minWidth,minHeight,maxWidth,maxHeight,left,top,bottom,right,fontSize,outlineWidth,outlineOffset,paddingTop,paddingLeft,paddingBottom,paddingRight,marginTop,marginLeft,marginBottom,marginRight,borderRadius,borderWidth,borderTopWidth,borderLeftWidth,borderRightWidth,borderBottomWidth,textIndent,perspective".split(","),Ct={},Ot.forEach(function(t){return Ct[t]=!0}),Ct);function At(t,e,n,r,i,o,a,s,u,c,l,d,p){return{type:0,element:t,triggerName:e,isRemovalTransition:i,fromState:n,fromStyles:o,toState:r,toStyles:a,timelines:s,queriedElements:u,preStyleProps:c,postStyleProps:l,totalTime:d,errors:p}}var Dt={},It=function(){function t(t,e,n){this._triggerName=t,this.ast=e,this._stateStyles=n}return t.prototype.match=function(t,e,n,r){return i=this.ast.matchers,o=t,a=e,s=n,u=r,i.some(function(t){return t(o,a,s,u)});var i,o,a,s,u},t.prototype.buildStyles=function(t,e,n){var r=this._stateStyles["*"],i=this._stateStyles[t],o=r?r.buildStyles(e,n):{};return i?i.buildStyles(e,n):o},t.prototype.build=function(t,e,n,i,o,a,s,u,c,l){var d=[],f=this.ast.options&&this.ast.options.params||Dt,h=s&&s.params||Dt,m=this.buildStyles(n,h,d),g=u&&u.params||Dt,y=this.buildStyles(i,g,d),_=new Set,v=new Map,b=new Map,x="void"===i,w={params:Object(r.__assign)({},f,g)},j=l?[]:yt(t,e,this.ast.animation,o,a,m,y,w,c,d),O=0;if(j.forEach(function(t){O=Math.max(t.duration+t.delay,O)}),d.length)return At(e,this._triggerName,n,i,x,m,y,[],[],v,b,O,d);j.forEach(function(t){var n=t.element,r=p(v,n,{});t.preStyleProps.forEach(function(t){return r[t]=!0});var i=p(b,n,{});t.postStyleProps.forEach(function(t){return i[t]=!0}),n!==e&&_.add(n)});var C=Y(_.values());return At(e,this._triggerName,n,i,x,m,y,j,C,v,b,O)},t}();var Pt=function(){function t(t,e){this.styles=t,this.defaultParams=e}return t.prototype.buildStyles=function(t,e){var n={},r=N(this.defaultParams);return Object.keys(t).forEach(function(e){var n=t[e];null!=n&&(r[e]=n)}),this.styles.styles.forEach(function(t){if("string"!=typeof t){var i=t;Object.keys(i).forEach(function(t){var o=i[t];o.length>1&&(o=G(o,r,e)),n[t]=o})}}),n},t}();var Rt=function(){function t(t,e){var n,r,i=this;this.name=t,this.ast=e,this.transitionFactories=[],this.states={},e.states.forEach(function(t){var e=t.options&&t.options.params||{};i.states[t.name]=new Pt(t.style,e)}),Nt(this.states,"true","1"),Nt(this.states,"false","0"),e.transitions.forEach(function(e){i.transitionFactories.push(new It(t,e,i.states))}),this.fallbackTransition=(n=t,r=this.states,new It(n,{type:1,animation:{type:2,steps:[],options:null},matchers:[function(t,e){return!0}],options:null,queryCount:0,depCount:0},r))}return Object.defineProperty(t.prototype,"containsQueries",{get:function(){return this.ast.queryCount>0},enumerable:!0,configurable:!0}),t.prototype.matchTransition=function(t,e,n,r){return this.transitionFactories.find(function(i){return i.match(t,e,n,r)})||null},t.prototype.matchStyles=function(t,e,n){return this.fallbackTransition.buildStyles(t,e,n)},t}();function Nt(t,e,n){t.hasOwnProperty(e)?t.hasOwnProperty(n)||(t[n]=t[e]):t.hasOwnProperty(n)&&(t[e]=t[n])}var Ft=new ht,Lt=function(){function t(t,e,n){this.bodyNode=t,this._driver=e,this._normalizer=n,this._animations={},this._playersById={},this.players=[]}return t.prototype.register=function(t,e){var n=[],r=st(this._driver,e,n);if(n.length)throw new Error("Unable to build the animation due to the following errors: "+n.join("\n"));this._animations[t]=r},t.prototype._buildPlayer=function(t,e,n){var r=t.element,i=u(this._driver,this._normalizer,0,t.keyframes,e,n);return this._driver.animate(r,i,t.duration,t.delay,t.easing,[],!0)},t.prototype.create=function(t,e,n){var r=this;void 0===n&&(n={});var o,a=[],u=this._animations[t],c=new Map;if(u?(o=yt(this._driver,e,u,"ng-enter","ng-leave",{},{},n,Ft,a)).forEach(function(t){var e=p(c,t.element,{});t.postStyleProps.forEach(function(t){return e[t]=null})}):(a.push("The requested animation doesn't exist or has already been destroyed"),o=[]),a.length)throw new Error("Unable to create the animation due to the following errors: "+a.join("\n"));c.forEach(function(t,e){Object.keys(t).forEach(function(n){t[n]=r._driver.computeStyle(e,n,i.AUTO_STYLE)})});var l=s(o.map(function(t){var e=c.get(t.element);return r._buildPlayer(t,{},e)}));return this._playersById[t]=l,l.onDestroy(function(){return r.destroy(t)}),this.players.push(l),l},t.prototype.destroy=function(t){var e=this._getPlayer(t);e.destroy(),delete this._playersById[t];var n=this.players.indexOf(e);n>=0&&this.players.splice(n,1)},t.prototype._getPlayer=function(t){var e=this._playersById[t];if(!e)throw new Error("Unable to find the timeline player referenced by "+t);return e},t.prototype.listen=function(t,e,n,r){var i=d(e,"","","");return c(this._getPlayer(t),n,i,r),function(){}},t.prototype.command=function(t,e,n,r){if("register"!=n)if("create"!=n){var i=this._getPlayer(t);switch(n){case"play":i.play();break;case"pause":i.pause();break;case"reset":i.reset();break;case"restart":i.restart();break;case"finish":i.finish();break;case"init":i.init();break;case"setPosition":i.setPosition(parseFloat(r[0]));break;case"destroy":this.destroy(t)}}else{var o=r[0]||{};this.create(t,e,o)}else this.register(t,r[0])},t}(),Bt="ng-animate-disabled",Vt=".ng-animate-disabled",zt=[],Ht={namespaceId:"",setForRemoval:!1,setForMove:!1,hasAnimation:!1,removedBeforeQueried:!1},Ut={namespaceId:"",setForMove:!1,setForRemoval:!1,hasAnimation:!1,removedBeforeQueried:!0},Wt="__ng_removed",qt=function(){function t(t,e){void 0===e&&(e=""),this.namespaceId=e;var n,r=t&&t.hasOwnProperty("value"),i=r?t.value:t;if(this.value=null!=(n=i)?n:null,r){var o=N(t);delete o.value,this.options=o}else this.options={};this.options.params||(this.options.params={})}return Object.defineProperty(t.prototype,"params",{get:function(){return this.options.params},enumerable:!0,configurable:!0}),t.prototype.absorbOptions=function(t){var e=t.params;if(e){var n=this.options.params;Object.keys(e).forEach(function(t){null==n[t]&&(n[t]=e[t])})}},t}(),Gt=new qt("void"),Yt=function(){function t(t,e,n){this.id=t,this.hostElement=e,this._engine=n,this.players=[],this._triggers={},this._queue=[],this._elementListeners=new Map,this._hostClassName="ng-tns-"+t,ee(e,this._hostClassName)}return t.prototype.listen=function(t,e,n,r){var i,o=this;if(!this._triggers.hasOwnProperty(e))throw new Error('Unable to listen on the animation trigger event "'+n+'" because the animation trigger "'+e+"\" doesn't exist!");if(null==n||0==n.length)throw new Error('Unable to listen on the animation trigger "'+e+'" because the provided event is undefined!');if("start"!=(i=n)&&"done"!=i)throw new Error('The provided animation trigger event "'+n+'" for the animation trigger "'+e+'" is not supported!');var a=p(this._elementListeners,t,[]),s={name:e,phase:n,callback:r};a.push(s);var u=p(this._engine.statesByElement,t,{});return u.hasOwnProperty(e)||(ee(t,k),ee(t,k+"-"+e),u[e]=Gt),function(){o._engine.afterFlush(function(){var t=a.indexOf(s);t>=0&&a.splice(t,1),o._triggers[e]||delete u[e]})}},t.prototype.register=function(t,e){return!this._triggers[t]&&(this._triggers[t]=e,!0)},t.prototype._getTrigger=function(t){var e=this._triggers[t];if(!e)throw new Error('The provided animation trigger "'+t+'" has not been registered!');return e},t.prototype.trigger=function(t,e,n,r){var i=this;void 0===r&&(r=!0);var o=this._getTrigger(e),a=new Xt(this.id,e,t),s=this._engine.statesByElement.get(t);s||(ee(t,k),ee(t,k+"-"+e),this._engine.statesByElement.set(t,s={}));var u=s[e],c=new qt(n,this.id);if(!(n&&n.hasOwnProperty("value"))&&u&&c.absorbOptions(u.options),s[e]=c,u||(u=Gt),"void"===c.value||u.value!==c.value){var l=p(this._engine.playersByElement,t,[]);l.forEach(function(t){t.namespaceId==i.id&&t.triggerName==e&&t.queued&&t.destroy()});var d=o.matchTransition(u.value,c.value,t,c.params),f=!1;if(!d){if(!r)return;d=o.fallbackTransition,f=!0}return this._engine.totalQueuedPlayers++,this._queue.push({element:t,triggerName:e,transition:d,fromState:u,toState:c,player:a,isFallbackTransition:f}),f||(ee(t,"ng-animate-queued"),a.onStart(function(){ne(t,"ng-animate-queued")})),a.onDone(function(){var e=i.players.indexOf(a);e>=0&&i.players.splice(e,1);var n=i._engine.playersByElement.get(t);if(n){var r=n.indexOf(a);r>=0&&n.splice(r,1)}}),this.players.push(a),l.push(a),a}if(!function(t,e){var n=Object.keys(t),r=Object.keys(e);if(n.length!=r.length)return!1;for(var i=0;i<n.length;i++){var o=n[i];if(!e.hasOwnProperty(o)||t[o]!==e[o])return!1}return!0}(u.params,c.params)){var h=[],m=o.matchStyles(u.value,u.params,h),g=o.matchStyles(c.value,c.params,h);h.length?this._engine.reportError(h):this._engine.afterFlush(function(){H(t,m),z(t,g)})}},t.prototype.deregister=function(t){var e=this;delete this._triggers[t],this._engine.statesByElement.forEach(function(e,n){delete e[t]}),this._elementListeners.forEach(function(n,r){e._elementListeners.set(r,n.filter(function(e){return e.name!=t}))})},t.prototype.clearElementCache=function(t){this._engine.statesByElement.delete(t),this._elementListeners.delete(t);var e=this._engine.playersByElement.get(t);e&&(e.forEach(function(t){return t.destroy()}),this._engine.playersByElement.delete(t))},t.prototype._signalRemovalForInnerTriggers=function(t,e,n){var r=this;void 0===n&&(n=!1),this._engine.driver.query(t,A,!0).forEach(function(t){if(!t[Wt]){var n=r._engine.fetchNamespacesByElement(t);n.size?n.forEach(function(n){return n.triggerLeaveAnimation(t,e,!1,!0)}):r.clearElementCache(t)}})},t.prototype.triggerLeaveAnimation=function(t,e,n,r){var i=this,o=this._engine.statesByElement.get(t);if(o){var a=[];if(Object.keys(o).forEach(function(e){if(i._triggers[e]){var n=i.trigger(t,e,"void",r);n&&a.push(n)}}),a.length)return this._engine.markElementAsRemoved(this.id,t,!0,e),n&&s(a).onDone(function(){return i._engine.processLeaveNode(t)}),!0}return!1},t.prototype.prepareLeaveAnimationListeners=function(t){var e=this,n=this._elementListeners.get(t);if(n){var r=new Set;n.forEach(function(n){var i=n.name;if(!r.has(i)){r.add(i);var o=e._triggers[i].fallbackTransition,a=e._engine.statesByElement.get(t)[i]||Gt,s=new qt("void"),u=new Xt(e.id,i,t);e._engine.totalQueuedPlayers++,e._queue.push({element:t,triggerName:i,transition:o,fromState:a,toState:s,player:u,isFallbackTransition:!0})}})}},t.prototype.removeNode=function(t,e){var n=this,r=this._engine;if(t.childElementCount&&this._signalRemovalForInnerTriggers(t,e,!0),!this.triggerLeaveAnimation(t,e,!0)){var i=!1;if(r.totalAnimations){var o=r.players.length?r.playersByQueriedElement.get(t):[];if(o&&o.length)i=!0;else for(var a=t;a=a.parentNode;){if(r.statesByElement.get(a)){i=!0;break}}}this.prepareLeaveAnimationListeners(t),i?r.markElementAsRemoved(this.id,t,!1,e):(r.afterFlush(function(){return n.clearElementCache(t)}),r.destroyInnerAnimations(t),r._onRemovalComplete(t,e))}},t.prototype.insertNode=function(t,e){ee(t,this._hostClassName)},t.prototype.drainQueuedTransitions=function(t){var e=this,n=[];return this._queue.forEach(function(r){var i=r.player;if(!i.destroyed){var o=r.element,a=e._elementListeners.get(o);a&&a.forEach(function(e){if(e.name==r.triggerName){var n=d(o,r.triggerName,r.fromState.value,r.toState.value);n._data=t,c(r.player,e.phase,n,e.callback)}}),i.markedForDestroy?e._engine.afterFlush(function(){i.destroy()}):n.push(r)}}),this._queue=[],n.sort(function(t,n){var r=t.transition.ast.depCount,i=n.transition.ast.depCount;return 0==r||0==i?r-i:e._engine.driver.containsElement(t.element,n.element)?1:-1})},t.prototype.destroy=function(t){this.players.forEach(function(t){return t.destroy()}),this._signalRemovalForInnerTriggers(this.hostElement,t)},t.prototype.elementContainsData=function(t){var e=!1;return this._elementListeners.has(t)&&(e=!0),e=!!this._queue.find(function(e){return e.element===t})||e},t}(),Kt=function(){function t(t,e,n){this.bodyNode=t,this.driver=e,this._normalizer=n,this.players=[],this.newHostElements=new Map,this.playersByElement=new Map,this.playersByQueriedElement=new Map,this.statesByElement=new Map,this.disabledNodes=new Set,this.totalAnimations=0,this.totalQueuedPlayers=0,this._namespaceLookup={},this._namespaceList=[],this._flushFns=[],this._whenQuietFns=[],this.namespacesByHostElement=new Map,this.collectedEnterElements=[],this.collectedLeaveElements=[],this.onRemovalComplete=function(t,e){}}return t.prototype._onRemovalComplete=function(t,e){this.onRemovalComplete(t,e)},Object.defineProperty(t.prototype,"queuedPlayers",{get:function(){var t=[];return this._namespaceList.forEach(function(e){e.players.forEach(function(e){e.queued&&t.push(e)})}),t},enumerable:!0,configurable:!0}),t.prototype.createNamespace=function(t,e){var n=new Yt(t,e,this);return e.parentNode?this._balanceNamespaceList(n,e):(this.newHostElements.set(e,n),this.collectEnterElement(e)),this._namespaceLookup[t]=n},t.prototype._balanceNamespaceList=function(t,e){var n=this._namespaceList.length-1;if(n>=0){for(var r=!1,i=n;i>=0;i--){var o=this._namespaceList[i];if(this.driver.containsElement(o.hostElement,e)){this._namespaceList.splice(i+1,0,t),r=!0;break}}r||this._namespaceList.splice(0,0,t)}else this._namespaceList.push(t);return this.namespacesByHostElement.set(e,t),t},t.prototype.register=function(t,e){var n=this._namespaceLookup[t];return n||(n=this.createNamespace(t,e)),n},t.prototype.registerTrigger=function(t,e,n){var r=this._namespaceLookup[t];r&&r.register(e,n)&&this.totalAnimations++},t.prototype.destroy=function(t,e){var n=this;if(t){var r=this._fetchNamespace(t);this.afterFlush(function(){n.namespacesByHostElement.delete(r.hostElement),delete n._namespaceLookup[t];var e=n._namespaceList.indexOf(r);e>=0&&n._namespaceList.splice(e,1)}),this.afterFlushAnimationsDone(function(){return r.destroy(e)})}},t.prototype._fetchNamespace=function(t){return this._namespaceLookup[t]},t.prototype.fetchNamespacesByElement=function(t){var e=new Set,n=this.statesByElement.get(t);if(n)for(var r=Object.keys(n),i=0;i<r.length;i++){var o=n[r[i]].namespaceId;if(o){var a=this._fetchNamespace(o);a&&e.add(a)}}return e},t.prototype.trigger=function(t,e,n,r){if(Qt(e)){var i=this._fetchNamespace(t);if(i)return i.trigger(e,n,r),!0}return!1},t.prototype.insertNode=function(t,e,n,r){if(Qt(e)){var i=e[Wt];if(i&&i.setForRemoval){i.setForRemoval=!1,i.setForMove=!0;var o=this.collectedLeaveElements.indexOf(e);o>=0&&this.collectedLeaveElements.splice(o,1)}if(t){var a=this._fetchNamespace(t);a&&a.insertNode(e,n)}r&&this.collectEnterElement(e)}},t.prototype.collectEnterElement=function(t){this.collectedEnterElements.push(t)},t.prototype.markElementAsDisabled=function(t,e){e?this.disabledNodes.has(t)||(this.disabledNodes.add(t),ee(t,Bt)):this.disabledNodes.has(t)&&(this.disabledNodes.delete(t),ne(t,Bt))},t.prototype.removeNode=function(t,e,n){if(Qt(e)){var r=t?this._fetchNamespace(t):null;r?r.removeNode(e,n):this.markElementAsRemoved(t,e,!1,n)}else this._onRemovalComplete(e,n)},t.prototype.markElementAsRemoved=function(t,e,n,r){this.collectedLeaveElements.push(e),e[Wt]={namespaceId:t,setForRemoval:r,hasAnimation:n,removedBeforeQueried:!1}},t.prototype.listen=function(t,e,n,r,i){return Qt(e)?this._fetchNamespace(t).listen(e,n,r,i):function(){}},t.prototype._buildInstruction=function(t,e,n,r,i){return t.transition.build(this.driver,t.element,t.fromState.value,t.toState.value,n,r,t.fromState.options,t.toState.options,e,i)},t.prototype.destroyInnerAnimations=function(t){var e=this,n=this.driver.query(t,A,!0);n.forEach(function(t){return e.destroyActiveAnimationsForElement(t)}),0!=this.playersByQueriedElement.size&&(n=this.driver.query(t,D,!0)).forEach(function(t){return e.finishActiveQueriedAnimationOnElement(t)})},t.prototype.destroyActiveAnimationsForElement=function(t){var e=this.playersByElement.get(t);e&&e.forEach(function(t){t.queued?t.markedForDestroy=!0:t.destroy()})},t.prototype.finishActiveQueriedAnimationOnElement=function(t){var e=this.playersByQueriedElement.get(t);e&&e.forEach(function(t){return t.finish()})},t.prototype.whenRenderingDone=function(){var t=this;return new Promise(function(e){if(t.players.length)return s(t.players).onDone(function(){return e()});e()})},t.prototype.processLeaveNode=function(t){var e=this,n=t[Wt];if(n&&n.setForRemoval){if(t[Wt]=Ht,n.namespaceId){this.destroyInnerAnimations(t);var r=this._fetchNamespace(n.namespaceId);r&&r.clearElementCache(t)}this._onRemovalComplete(t,n.setForRemoval)}this.driver.matchesElement(t,Vt)&&this.markElementAsDisabled(t,!1),this.driver.query(t,Vt,!0).forEach(function(t){e.markElementAsDisabled(t,!1)})},t.prototype.flush=function(t){var e=this;void 0===t&&(t=-1);var n=[];if(this.newHostElements.size&&(this.newHostElements.forEach(function(t,n){return e._balanceNamespaceList(t,n)}),this.newHostElements.clear()),this.totalAnimations&&this.collectedEnterElements.length)for(var r=0;r<this.collectedEnterElements.length;r++){ee(this.collectedEnterElements[r],"ng-star-inserted")}if(this._namespaceList.length&&(this.totalQueuedPlayers||this.collectedLeaveElements.length)){var i=[];try{n=this._flushAnimations(i,t)}finally{for(r=0;r<i.length;r++)i[r]()}}else for(r=0;r<this.collectedLeaveElements.length;r++){var o=this.collectedLeaveElements[r];this.processLeaveNode(o)}if(this.totalQueuedPlayers=0,this.collectedEnterElements.length=0,this.collectedLeaveElements.length=0,this._flushFns.forEach(function(t){return t()}),this._flushFns=[],this._whenQuietFns.length){var a=this._whenQuietFns;this._whenQuietFns=[],n.length?s(n).onDone(function(){a.forEach(function(t){return t()})}):a.forEach(function(t){return t()})}},t.prototype.reportError=function(t){throw new Error("Unable to process animations due to the following failed trigger transitions\n "+t.join("\n"))},t.prototype._flushAnimations=function(t,e){var n=this,o=new ht,a=[],u=new Map,c=[],l=new Map,d=new Map,f=new Map,h=new Set;this.disabledNodes.forEach(function(t){h.add(t);for(var e=n.driver.query(t,".ng-animate-queued",!0),r=0;r<e.length;r++)h.add(e[r])});var m=this.bodyNode,g=Array.from(this.statesByElement.keys()),y=Jt(g,this.collectedEnterElements),_=new Map,v=0;y.forEach(function(t,e){var n="ng-enter"+v++;_.set(e,n),t.forEach(function(t){return ee(t,n)})});for(var b=[],x=new Set,w=new Set,j=0;j<this.collectedLeaveElements.length;j++){(U=(V=this.collectedLeaveElements[j])[Wt])&&U.setForRemoval&&(b.push(V),x.add(V),U.hasAnimation?this.driver.query(V,".ng-star-inserted",!0).forEach(function(t){return x.add(t)}):w.add(V))}var O=new Map,C=Jt(g,Array.from(x));C.forEach(function(t,e){var n="ng-leave"+v++;O.set(e,n),t.forEach(function(t){return ee(t,n)})}),t.push(function(){y.forEach(function(t,e){var n=_.get(e);t.forEach(function(t){return ne(t,n)})}),C.forEach(function(t,e){var n=O.get(e);t.forEach(function(t){return ne(t,n)})}),b.forEach(function(t){n.processLeaveNode(t)})});for(var E=[],S=[],M=this._namespaceList.length-1;M>=0;M--){this._namespaceList[M].drainQueuedTransitions(e).forEach(function(t){var e=t.player,r=t.element;if(E.push(e),n.collectedEnterElements.length){var i=r[Wt];if(i&&i.setForMove)return void e.destroy()}var s=!m||!n.driver.containsElement(m,r),u=O.get(r),h=_.get(r),g=n._buildInstruction(t,o,h,u,s);if(g.errors&&g.errors.length)S.push(g);else{if(s)return e.onStart(function(){return H(r,g.fromStyles)}),e.onDestroy(function(){return z(r,g.toStyles)}),void a.push(e);if(t.isFallbackTransition)return e.onStart(function(){return H(r,g.fromStyles)}),e.onDestroy(function(){return z(r,g.toStyles)}),void a.push(e);g.timelines.forEach(function(t){return t.stretchStartingKeyframe=!0}),o.append(r,g.timelines);var y={instruction:g,player:e,element:r};c.push(y),g.queriedElements.forEach(function(t){return p(l,t,[]).push(e)}),g.preStyleProps.forEach(function(t,e){var n=Object.keys(t);if(n.length){var r=d.get(e);r||d.set(e,r=new Set),n.forEach(function(t){return r.add(t)})}}),g.postStyleProps.forEach(function(t,e){var n=Object.keys(t),r=f.get(e);r||f.set(e,r=new Set),n.forEach(function(t){return r.add(t)})})}})}if(S.length){var T=[];S.forEach(function(t){T.push("@"+t.triggerName+" has failed due to:\n"),t.errors.forEach(function(t){return T.push("- "+t+"\n")})}),E.forEach(function(t){return t.destroy()}),this.reportError(T)}var k=new Map,A=new Map;c.forEach(function(t){var e=t.element;o.has(e)&&(A.set(e,e),n._beforeAnimationBuild(t.player.namespaceId,t.instruction,k))}),a.forEach(function(t){var e=t.element;n._getPreviousPlayers(e,!1,t.namespaceId,t.triggerName,null).forEach(function(t){p(k,e,[]).push(t),t.destroy()})});var I=b.filter(function(t){return ie(t,d,f)}),P=new Map;$t(P,this.driver,w,f,i.AUTO_STYLE).forEach(function(t){ie(t,d,f)&&I.push(t)});var R=new Map;y.forEach(function(t,e){$t(R,n.driver,new Set(t),d,i["ɵPRE_STYLE"])}),I.forEach(function(t){var e=P.get(t),n=R.get(t);P.set(t,Object(r.__assign)({},e,n))});var N=[],F=[],L={};c.forEach(function(t){var e=t.element,r=t.player,i=t.instruction;if(o.has(e)){if(h.has(e))return r.onDestroy(function(){return z(e,i.toStyles)}),r.disabled=!0,r.overrideTotalTime(i.totalTime),void a.push(r);var c=L;if(A.size>1){for(var l=e,d=[];l=l.parentNode;){var p=A.get(l);if(p){c=p;break}d.push(l)}d.forEach(function(t){return A.set(t,c)})}var f=n._buildAnimation(r.namespaceId,i,k,u,R,P);if(r.setRealPlayer(f),c===L)N.push(r);else{var m=n.playersByElement.get(c);m&&m.length&&(r.parentPlayer=s(m)),a.push(r)}}else H(e,i.fromStyles),r.onDestroy(function(){return z(e,i.toStyles)}),F.push(r),h.has(e)&&a.push(r)}),F.forEach(function(t){var e=u.get(t.element);if(e&&e.length){var n=s(e);t.setRealPlayer(n)}}),a.forEach(function(t){t.parentPlayer?t.syncPlayerEvents(t.parentPlayer):t.destroy()});for(var B=0;B<b.length;B++){var V,U=(V=b[B])[Wt];if(ne(V,"ng-leave"),!U||!U.hasAnimation){var W=[];if(l.size){var q=l.get(V);q&&q.length&&W.push.apply(W,Object(r.__spread)(q));for(var G=this.driver.query(V,D,!0),Y=0;Y<G.length;Y++){var K=l.get(G[Y]);K&&K.length&&W.push.apply(W,Object(r.__spread)(K))}}var X=W.filter(function(t){return!t.destroyed});X.length?re(this,V,X):this.processLeaveNode(V)}}return b.length=0,N.forEach(function(t){n.players.push(t),t.onDone(function(){t.destroy();var e=n.players.indexOf(t);n.players.splice(e,1)}),t.play()}),N},t.prototype.elementContainsData=function(t,e){var n=!1,r=e[Wt];return r&&r.setForRemoval&&(n=!0),this.playersByElement.has(e)&&(n=!0),this.playersByQueriedElement.has(e)&&(n=!0),this.statesByElement.has(e)&&(n=!0),this._fetchNamespace(t).elementContainsData(e)||n},t.prototype.afterFlush=function(t){this._flushFns.push(t)},t.prototype.afterFlushAnimationsDone=function(t){this._whenQuietFns.push(t)},t.prototype._getPreviousPlayers=function(t,e,n,r,i){var o=[];if(e){var a=this.playersByQueriedElement.get(t);a&&(o=a)}else{var s=this.playersByElement.get(t);if(s){var u=!i||"void"==i;s.forEach(function(t){t.queued||(u||t.triggerName==r)&&o.push(t)})}}return(n||r)&&(o=o.filter(function(t){return(!n||n==t.namespaceId)&&(!r||r==t.triggerName)})),o},t.prototype._beforeAnimationBuild=function(t,e,n){var i,o,a=e.triggerName,s=e.element,u=e.isRemovalTransition?void 0:t,c=e.isRemovalTransition?void 0:a,l=function(t){var r=t.element,i=r!==s,o=p(n,r,[]);d._getPreviousPlayers(r,i,u,c,e.toState).forEach(function(t){var e=t.getRealPlayer();e.beforeDestroy&&e.beforeDestroy(),t.destroy(),o.push(t)})},d=this;try{for(var f=Object(r.__values)(e.timelines),h=f.next();!h.done;h=f.next()){l(h.value)}}catch(t){i={error:t}}finally{try{h&&!h.done&&(o=f.return)&&o.call(f)}finally{if(i)throw i.error}}H(s,e.fromStyles)},t.prototype._buildAnimation=function(t,e,n,r,o,a){var c=this,l=e.triggerName,d=e.element,f=[],h=new Set,m=new Set,g=e.timelines.map(function(e){var s=e.element;h.add(s);var p=s[Wt];if(p&&p.removedBeforeQueried)return new i.NoopAnimationPlayer(e.duration,e.delay);var g,y,_=s!==d,v=(g=(n.get(s)||zt).map(function(t){return t.getRealPlayer()}),y=[],function t(e,n){for(var r=0;r<e.length;r++){var o=e[r];o instanceof i["ɵAnimationGroupPlayer"]?t(o.players,n):n.push(o)}}(g,y),y).filter(function(t){var e=t;return!!e.element&&e.element===s}),b=o.get(s),x=a.get(s),w=u(c.driver,c._normalizer,0,e.keyframes,b,x),j=c._buildPlayer(e,w,v);if(e.subTimeline&&r&&m.add(s),_){var O=new Xt(t,l,s);O.setRealPlayer(j),f.push(O)}return j});f.forEach(function(t){p(c.playersByQueriedElement,t.element,[]).push(t),t.onDone(function(){return function(t,e,n){var r;if(t instanceof Map){if(r=t.get(e)){if(r.length){var i=r.indexOf(n);r.splice(i,1)}0==r.length&&t.delete(e)}}else if(r=t[e]){if(r.length){var i=r.indexOf(n);r.splice(i,1)}0==r.length&&delete t[e]}return r}(c.playersByQueriedElement,t.element,t)})}),h.forEach(function(t){return ee(t,"ng-animating")});var y=s(g);return y.onDestroy(function(){h.forEach(function(t){return ne(t,"ng-animating")}),z(d,e.toStyles)}),m.forEach(function(t){p(r,t,[]).push(y)}),y},t.prototype._buildPlayer=function(t,e,n){return e.length>0?this.driver.animate(t.element,e,t.duration,t.delay,t.easing,n):new i.NoopAnimationPlayer(t.duration,t.delay)},t}(),Xt=function(){function t(t,e,n){this.namespaceId=t,this.triggerName=e,this.element=n,this._player=new i.NoopAnimationPlayer,this._containsRealPlayer=!1,this._queuedCallbacks={},this.destroyed=!1,this.markedForDestroy=!1,this.disabled=!1,this.queued=!0,this.totalTime=0}return t.prototype.setRealPlayer=function(t){var e=this;this._containsRealPlayer||(this._player=t,Object.keys(this._queuedCallbacks).forEach(function(n){e._queuedCallbacks[n].forEach(function(e){return c(t,n,void 0,e)})}),this._queuedCallbacks={},this._containsRealPlayer=!0,this.overrideTotalTime(t.totalTime),this.queued=!1)},t.prototype.getRealPlayer=function(){return this._player},t.prototype.overrideTotalTime=function(t){this.totalTime=t},t.prototype.syncPlayerEvents=function(t){var e=this,n=this._player;n.triggerCallback&&t.onStart(function(){return n.triggerCallback("start")}),t.onDone(function(){return e.finish()}),t.onDestroy(function(){return e.destroy()})},t.prototype._queueEvent=function(t,e){p(this._queuedCallbacks,t,[]).push(e)},t.prototype.onDone=function(t){this.queued&&this._queueEvent("done",t),this._player.onDone(t)},t.prototype.onStart=function(t){this.queued&&this._queueEvent("start",t),this._player.onStart(t)},t.prototype.onDestroy=function(t){this.queued&&this._queueEvent("destroy",t),this._player.onDestroy(t)},t.prototype.init=function(){this._player.init()},t.prototype.hasStarted=function(){return!this.queued&&this._player.hasStarted()},t.prototype.play=function(){!this.queued&&this._player.play()},t.prototype.pause=function(){!this.queued&&this._player.pause()},t.prototype.restart=function(){!this.queued&&this._player.restart()},t.prototype.finish=function(){this._player.finish()},t.prototype.destroy=function(){this.destroyed=!0,this._player.destroy()},t.prototype.reset=function(){!this.queued&&this._player.reset()},t.prototype.setPosition=function(t){this.queued||this._player.setPosition(t)},t.prototype.getPosition=function(){return this.queued?0:this._player.getPosition()},t.prototype.triggerCallback=function(t){var e=this._player;e.triggerCallback&&e.triggerCallback(t)},t}();function Qt(t){return t&&1===t.nodeType}function Zt(t,e){var n=t.style.display;return t.style.display=null!=e?e:"none",n}function $t(t,e,n,r,i){var o=[];n.forEach(function(t){return o.push(Zt(t))});var a=[];r.forEach(function(n,r){var o={};n.forEach(function(t){var n=o[t]=e.computeStyle(r,t,i);n&&0!=n.length||(r[Wt]=Ut,a.push(r))}),t.set(r,o)});var s=0;return n.forEach(function(t){return Zt(t,o[s++])}),a}function Jt(t,e){var n=new Map;if(t.forEach(function(t){return n.set(t,[])}),0==e.length)return n;var r=1,i=new Set(e),o=new Map;return e.forEach(function(t){var e=function t(e){if(!e)return r;var a=o.get(e);if(a)return a;var s=e.parentNode;return a=n.has(s)?s:i.has(s)?r:t(s),o.set(e,a),a}(t);e!==r&&n.get(e).push(t)}),n}var te="$$classes";function ee(t,e){if(t.classList)t.classList.add(e);else{var n=t[te];n||(n=t[te]={}),n[e]=!0}}function ne(t,e){if(t.classList)t.classList.remove(e);else{var n=t[te];n&&delete n[e]}}function re(t,e,n){s(n).onDone(function(){return t.processLeaveNode(e)})}function ie(t,e,n){var r=n.get(t);if(!r)return!1;var i=e.get(t);return i?r.forEach(function(t){return i.add(t)}):e.set(t,r),n.delete(t),!0}var oe=function(){function t(t,e,n){var r=this;this.bodyNode=t,this._driver=e,this._triggerCache={},this.onRemovalComplete=function(t,e){},this._transitionEngine=new Kt(t,e,n),this._timelineEngine=new Lt(t,e,n),this._transitionEngine.onRemovalComplete=function(t,e){return r.onRemovalComplete(t,e)}}return t.prototype.registerTrigger=function(t,e,n,r,i){var o=t+"-"+r,a=this._triggerCache[o];if(!a){var s=[],u=st(this._driver,i,s);if(s.length)throw new Error('The animation trigger "'+r+'" has failed to build due to the following errors:\n - '+s.join("\n - "));a=new Rt(r,u),this._triggerCache[o]=a}this._transitionEngine.registerTrigger(e,r,a)},t.prototype.register=function(t,e){this._transitionEngine.register(t,e)},t.prototype.destroy=function(t,e){this._transitionEngine.destroy(t,e)},t.prototype.onInsert=function(t,e,n,r){this._transitionEngine.insertNode(t,e,n,r)},t.prototype.onRemove=function(t,e,n){this._transitionEngine.removeNode(t,e,n)},t.prototype.disableAnimations=function(t,e){this._transitionEngine.markElementAsDisabled(t,e)},t.prototype.process=function(t,e,n,i){if("@"==n.charAt(0)){var o=Object(r.__read)(f(n),2),a=o[0],s=o[1],u=i;this._timelineEngine.command(a,e,s,u)}else this._transitionEngine.trigger(t,e,n,i)},t.prototype.listen=function(t,e,n,i,o){if("@"==n.charAt(0)){var a=Object(r.__read)(f(n),2),s=a[0],u=a[1];return this._timelineEngine.listen(s,e,u,o)}return this._transitionEngine.listen(t,e,n,i,o)},t.prototype.flush=function(t){void 0===t&&(t=-1),this._transitionEngine.flush(t)},Object.defineProperty(t.prototype,"players",{get:function(){return this._transitionEngine.players.concat(this._timelineEngine.players)},enumerable:!0,configurable:!0}),t.prototype.whenRenderingDone=function(){return this._transitionEngine.whenRenderingDone()},t}(),ae="animation",se="animationend",ue=function(){function t(t,e,n,r,i,o,a){var s=this;this._element=t,this._name=e,this._duration=n,this._delay=r,this._easing=i,this._fillMode=o,this._onDoneFn=a,this._finished=!1,this._destroyed=!1,this._startTime=0,this._position=0,this._eventFn=function(t){return s._handleCallback(t)}}return t.prototype.apply=function(){!function(t,e){var n=he(t,"").trim(),r=0;n.length&&(r=function(t,e){for(var n=0,r=0;r<t.length;r++){var i=t.charAt(r);i===e&&n++}return n}(n,",")+1,e=n+", "+e);fe(t,"",e)}(this._element,this._duration+"ms "+this._easing+" "+this._delay+"ms 1 normal "+this._fillMode+" "+this._name),pe(this._element,this._eventFn,!1),this._startTime=Date.now()},t.prototype.pause=function(){ce(this._element,this._name,"paused")},t.prototype.resume=function(){ce(this._element,this._name,"running")},t.prototype.setPosition=function(t){var e=le(this._element,this._name);this._position=t*this._duration,fe(this._element,"Delay","-"+this._position+"ms",e)},t.prototype.getPosition=function(){return this._position},t.prototype._handleCallback=function(t){var e=t._ngTestManualTimestamp||Date.now(),n=1e3*parseFloat(t.elapsedTime.toFixed(3));t.animationName==this._name&&Math.max(e-this._startTime,0)>=this._delay&&n>=this._duration&&this.finish()},t.prototype.finish=function(){this._finished||(this._finished=!0,this._onDoneFn(),pe(this._element,this._eventFn,!0))},t.prototype.destroy=function(){this._destroyed||(this._destroyed=!0,this.finish(),function(t,e){var n=he(t,"").split(","),r=de(n,e);if(r>=0){n.splice(r,1);var i=n.join(",");fe(t,"",i)}}(this._element,this._name))},t}();function ce(t,e,n){fe(t,"PlayState",n,le(t,e))}function le(t,e){var n=he(t,"");return n.indexOf(",")>0?de(n.split(","),e):de([n],e)}function de(t,e){for(var n=0;n<t.length;n++)if(t[n].indexOf(e)>=0)return n;return-1}function pe(t,e,n){n?t.removeEventListener(se,e):t.addEventListener(se,e)}function fe(t,e,n,r){var i=ae+e;if(null!=r){var o=t.style[i];if(o.length){var a=o.split(",");a[r]=n,n=a.join(",")}}t.style[i]=n}function he(t,e){return t.style[ae+e]}var me="linear",ge=function(){function t(t,e,n,r,i,o,a){this.element=t,this.keyframes=e,this.animationName=n,this._duration=r,this._delay=i,this._finalStyles=a,this._onDoneFns=[],this._onStartFns=[],this._onDestroyFns=[],this._started=!1,this.currentSnapshot={},this._state=0,this.easing=o||me,this.totalTime=r+i,this._buildStyler()}return t.prototype.onStart=function(t){this._onStartFns.push(t)},t.prototype.onDone=function(t){this._onDoneFns.push(t)},t.prototype.onDestroy=function(t){this._onDestroyFns.push(t)},t.prototype.destroy=function(){this.init(),this._state>=4||(this._state=4,this._styler.destroy(),this._flushStartFns(),this._flushDoneFns(),this._onDestroyFns.forEach(function(t){return t()}),this._onDestroyFns=[])},t.prototype._flushDoneFns=function(){this._onDoneFns.forEach(function(t){return t()}),this._onDoneFns=[]},t.prototype._flushStartFns=function(){this._onStartFns.forEach(function(t){return t()}),this._onStartFns=[]},t.prototype.finish=function(){this.init(),this._state>=3||(this._state=3,this._styler.finish(),this._flushStartFns(),this._flushDoneFns())},t.prototype.setPosition=function(t){this._styler.setPosition(t)},t.prototype.getPosition=function(){return this._styler.getPosition()},t.prototype.hasStarted=function(){return this._state>=2},t.prototype.init=function(){if(!(this._state>=1)){this._state=1;this.element;this._styler.apply(),this._delay&&this._styler.pause()}},t.prototype.play=function(){this.init(),this.hasStarted()||(this._flushStartFns(),this._state=2),this._styler.resume()},t.prototype.pause=function(){this.init(),this._styler.pause()},t.prototype.restart=function(){this.reset(),this.play()},t.prototype.reset=function(){this._styler.destroy(),this._buildStyler(),this._styler.apply()},t.prototype._buildStyler=function(){var t=this;this._styler=new ue(this.element,this.animationName,this._duration,this._delay,this.easing,"forwards",function(){return t.finish()})},t.prototype.triggerCallback=function(t){var e="start"==t?this._onStartFns:this._onDoneFns;e.forEach(function(t){return t()}),e.length=0},t.prototype.beforeDestroy=function(){var t=this;this.init();var e={};if(this.hasStarted()){var n=this._state>=3;Object.keys(this._finalStyles).forEach(function(r){"offset"!=r&&(e[r]=n?t._finalStyles[r]:J(t.element,r))})}this.currentSnapshot=e},t}(),ye=function(t){function e(e,n){var r=t.call(this)||this;return r.element=e,r._startingStyles={},r.__initialized=!1,r._styles=E(n),r}return Object(r.__extends)(e,t),e.prototype.init=function(){var e=this;!this.__initialized&&this._startingStyles&&(this.__initialized=!0,Object.keys(this._styles).forEach(function(t){e._startingStyles[t]=e.element.style[t]}),t.prototype.init.call(this))},e.prototype.play=function(){var e=this;this._startingStyles&&(this.init(),Object.keys(this._styles).forEach(function(t){return e.element.style.setProperty(t,e._styles[t])}),t.prototype.play.call(this))},e.prototype.destroy=function(){var e=this;this._startingStyles&&(Object.keys(this._startingStyles).forEach(function(t){var n=e._startingStyles[t];n?e.element.style.setProperty(t,n):e.element.style.removeProperty(t)}),this._startingStyles=null,t.prototype.destroy.call(this))},e}(i.NoopAnimationPlayer),_e=function(){function t(){this._count=0,this._head=document.querySelector("head"),this._warningIssued=!1}return t.prototype.validateStyleProperty=function(t){return w(t)},t.prototype.matchesElement=function(t,e){return j(t,e)},t.prototype.containsElement=function(t,e){return O(t,e)},t.prototype.query=function(t,e,n){return C(t,e,n)},t.prototype.computeStyle=function(t,e,n){return window.getComputedStyle(t)[e]},t.prototype.buildKeyframeElement=function(t,e,n){var r="@keyframes "+e+" {\n",i="";(n=n.map(function(t){return E(t)})).forEach(function(t){i=" ";var e=parseFloat(t.offset);r+=""+i+100*e+"% {\n",i+=" ",Object.keys(t).forEach(function(e){var n=t[e];switch(e){case"offset":return;case"easing":return void(n&&(r+=i+"animation-timing-function: "+n+";\n"));default:return void(r+=""+i+e+": "+n+";\n")}}),r+=i+"}\n"}),r+="}\n";var o=document.createElement("style");return o.innerHTML=r,o},t.prototype.animate=function(t,e,n,r,i,o,a){void 0===o&&(o=[]),a&&this._notifyFaultyScrubber();var s=o.filter(function(t){return t instanceof ge}),u={};Q(n,r)&&s.forEach(function(t){var e=t.currentSnapshot;Object.keys(e).forEach(function(t){return u[t]=e[t]})});var c=function(t){var e={};if(t){var n=Array.isArray(t)?t:[t];n.forEach(function(t){Object.keys(t).forEach(function(n){"offset"!=n&&"easing"!=n&&(e[n]=t[n])})})}return e}(e=Z(t,e,u));if(0==n)return new ye(t,c);var l="gen_css_kf_"+this._count++,d=this.buildKeyframeElement(t,l,e);document.querySelector("head").appendChild(d);var p=new ge(t,e,l,n,r,i,c);return p.onDestroy(function(){var t;(t=d).parentNode.removeChild(t)}),p},t.prototype._notifyFaultyScrubber=function(){this._warningIssued||(console.warn("@angular/animations: please load the web-animations.js polyfill to allow programmatic access...\n","  visit http://bit.ly/IWukam to learn more about using the web-animation-js polyfill."),this._warningIssued=!0)},t}();var ve=function(){function t(t,e,n){this.element=t,this.keyframes=e,this.options=n,this._onDoneFns=[],this._onStartFns=[],this._onDestroyFns=[],this._initialized=!1,this._finished=!1,this._started=!1,this._destroyed=!1,this.time=0,this.parentPlayer=null,this.currentSnapshot={},this._duration=n.duration,this._delay=n.delay||0,this.time=this._duration+this._delay}return t.prototype._onFinish=function(){this._finished||(this._finished=!0,this._onDoneFns.forEach(function(t){return t()}),this._onDoneFns=[])},t.prototype.init=function(){this._buildPlayer(),this._preparePlayerBeforeStart()},t.prototype._buildPlayer=function(){var t=this;if(!this._initialized){this._initialized=!0;var e=this.keyframes;this.domPlayer=this._triggerWebAnimation(this.element,e,this.options),this._finalKeyframe=e.length?e[e.length-1]:{},this.domPlayer.addEventListener("finish",function(){return t._onFinish()})}},t.prototype._preparePlayerBeforeStart=function(){this._delay?this._resetDomPlayerState():this.domPlayer.pause()},t.prototype._triggerWebAnimation=function(t,e,n){return t.animate(e,n)},t.prototype.onStart=function(t){this._onStartFns.push(t)},t.prototype.onDone=function(t){this._onDoneFns.push(t)},t.prototype.onDestroy=function(t){this._onDestroyFns.push(t)},t.prototype.play=function(){this._buildPlayer(),this.hasStarted()||(this._onStartFns.forEach(function(t){return t()}),this._onStartFns=[],this._started=!0),this.domPlayer.play()},t.prototype.pause=function(){this.init(),this.domPlayer.pause()},t.prototype.finish=function(){this.init(),this._onFinish(),this.domPlayer.finish()},t.prototype.reset=function(){this._resetDomPlayerState(),this._destroyed=!1,this._finished=!1,this._started=!1},t.prototype._resetDomPlayerState=function(){this.domPlayer&&this.domPlayer.cancel()},t.prototype.restart=function(){this.reset(),this.play()},t.prototype.hasStarted=function(){return this._started},t.prototype.destroy=function(){this._destroyed||(this._destroyed=!0,this._resetDomPlayerState(),this._onFinish(),this._onDestroyFns.forEach(function(t){return t()}),this._onDestroyFns=[])},t.prototype.setPosition=function(t){this.domPlayer.currentTime=t*this.time},t.prototype.getPosition=function(){return this.domPlayer.currentTime/this.time},Object.defineProperty(t.prototype,"totalTime",{get:function(){return this._delay+this._duration},enumerable:!0,configurable:!0}),t.prototype.beforeDestroy=function(){var t=this,e={};this.hasStarted()&&Object.keys(this._finalKeyframe).forEach(function(n){"offset"!=n&&(e[n]=t._finished?t._finalKeyframe[n]:J(t.element,n))}),this.currentSnapshot=e},t.prototype.triggerCallback=function(t){var e="start"==t?this._onStartFns:this._onDoneFns;e.forEach(function(t){return t()}),e.length=0},t}(),be=function(){function t(){this._isNativeImpl=/\{\s*\[native\s+code\]\s*\}/.test(we().toString()),this._cssKeyframesDriver=new _e}return t.prototype.validateStyleProperty=function(t){return w(t)},t.prototype.matchesElement=function(t,e){return j(t,e)},t.prototype.containsElement=function(t,e){return O(t,e)},t.prototype.query=function(t,e,n){return C(t,e,n)},t.prototype.computeStyle=function(t,e,n){return window.getComputedStyle(t)[e]},t.prototype.overrideWebAnimationsSupport=function(t){this._isNativeImpl=t},t.prototype.animate=function(t,e,n,r,i,o,a){if(void 0===o&&(o=[]),!a&&!this._isNativeImpl)return this._cssKeyframesDriver.animate(t,e,n,r,i,o);var s={duration:n,delay:r,fill:0==r?"both":"forwards"};i&&(s.easing=i);var u={},c=o.filter(function(t){return t instanceof ve});return Q(n,r)&&c.forEach(function(t){var e=t.currentSnapshot;Object.keys(e).forEach(function(t){return u[t]=e[t]})}),e=Z(t,e=e.map(function(t){return L(t,!1)}),u),new ve(t,e,s)},t}();function xe(){return"function"==typeof we()}function we(){return"undefined"!=typeof window&&void 0!==window.document&&Element.prototype.animate||{}}},"./node_modules/@angular/animations/node_modules/tslib/tslib.es6.js":function(t,e,n){"use strict";n.r(e),n.d(e,"__extends",function(){return i}),n.d(e,"__assign",function(){return o}),n.d(e,"__rest",function(){return a}),n.d(e,"__decorate",function(){return s}),n.d(e,"__param",function(){return u}),n.d(e,"__metadata",function(){return c}),n.d(e,"__awaiter",function(){return l}),n.d(e,"__generator",function(){return d}),n.d(e,"__exportStar",function(){return p}),n.d(e,"__values",function(){return f}),n.d(e,"__read",function(){return h}),n.d(e,"__spread",function(){return m}),n.d(e,"__await",function(){return g}),n.d(e,"__asyncGenerator",function(){return y}),n.d(e,"__asyncDelegator",function(){return _}),n.d(e,"__asyncValues",function(){return v}),n.d(e,"__makeTemplateObject",function(){return b}),n.d(e,"__importStar",function(){return x}),n.d(e,"__importDefault",function(){return w});var r=function(t,e){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)};function i(t,e){function n(){this.constructor=t}r(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}var o=function(){return(o=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++){e=arguments[n];for(var i in e)Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i])}return t}).apply(this,arguments)};function a(t,e){var n={};for(var r in t)Object.prototype.hasOwnProperty.call(t,r)&&e.indexOf(r)<0&&(n[r]=t[r]);if(null!=t&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(t);i<r.length;i++)e.indexOf(r[i])<0&&(n[r[i]]=t[r[i]])}return n}function s(t,e,n,r){var i,o=arguments.length,a=o<3?e:null===r?r=Object.getOwnPropertyDescriptor(e,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(t,e,n,r);else for(var s=t.length-1;s>=0;s--)(i=t[s])&&(a=(o<3?i(a):o>3?i(e,n,a):i(e,n))||a);return o>3&&a&&Object.defineProperty(e,n,a),a}function u(t,e){return function(n,r){e(n,r,t)}}function c(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)}function l(t,e,n,r){return new(n||(n=Promise))(function(i,o){function a(t){try{u(r.next(t))}catch(t){o(t)}}function s(t){try{u(r.throw(t))}catch(t){o(t)}}function u(t){t.done?i(t.value):new n(function(e){e(t.value)}).then(a,s)}u((r=r.apply(t,e||[])).next())})}function d(t,e){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=e.call(t,a)}catch(t){o=[6,t],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(t,e){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}function f(t){var e="function"==typeof Symbol&&t[Symbol.iterator],n=0;return e?e.call(t):{next:function(){return t&&n>=t.length&&(t=void 0),{value:t&&t[n++],done:!t}}}}function h(t,e){var n="function"==typeof Symbol&&t[Symbol.iterator];if(!n)return t;var r,i,o=n.call(t),a=[];try{for(;(void 0===e||e-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(t){i={error:t}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var t=[],e=0;e<arguments.length;e++)t=t.concat(h(arguments[e]));return t}function g(t){return this instanceof g?(this.v=t,this):new g(t)}function y(t,e,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(t,e||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(t){i[t]&&(r[t]=function(e){return new Promise(function(n,r){o.push([t,e,n,r])>1||s(t,e)})})}function s(t,e){try{(n=i[t](e)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(t){l(o[0][3],t)}var n}function u(t){s("next",t)}function c(t){s("throw",t)}function l(t,e){t(e),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(t){var e,n;return e={},r("next"),r("throw",function(t){throw t}),r("return"),e[Symbol.iterator]=function(){return this},e;function r(r,i){e[r]=t[r]?function(e){return(n=!n)?{value:g(t[r](e)),done:"return"===r}:i?i(e):e}:i}}function v(t){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var e,n=t[Symbol.asyncIterator];return n?n.call(t):(t=f(t),e={},r("next"),r("throw"),r("return"),e[Symbol.asyncIterator]=function(){return this},e);function r(n){e[n]=t[n]&&function(e){return new Promise(function(r,i){var o,a,s,u;e=t[n](e),o=r,a=i,s=e.done,u=e.value,Promise.resolve(u).then(function(t){o({value:t,done:s})},a)})}}}function b(t,e){return Object.defineProperty?Object.defineProperty(t,"raw",{value:e}):t.raw=e,t}function x(t){if(t&&t.__esModule)return t;var e={};if(null!=t)for(var n in t)Object.hasOwnProperty.call(t,n)&&(e[n]=t[n]);return e.default=t,e}function w(t){return t&&t.__esModule?t:{default:t}}},"./node_modules/@angular/cdk/esm5/a11y.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"ARIA_DESCRIBER_PROVIDER_FACTORY",function(){return x}),n.d(e,"MESSAGES_CONTAINER_ID",function(){return h}),n.d(e,"CDK_DESCRIBEDBY_ID_PREFIX",function(){return m}),n.d(e,"CDK_DESCRIBEDBY_HOST_ATTRIBUTE",function(){return g}),n.d(e,"AriaDescriber",function(){return b}),n.d(e,"ARIA_DESCRIBER_PROVIDER",function(){return w}),n.d(e,"ActiveDescendantKeyManager",function(){return O}),n.d(e,"FocusKeyManager",function(){return C}),n.d(e,"ListKeyManager",function(){return j}),n.d(e,"FocusTrap",function(){return T}),n.d(e,"FocusTrapFactory",function(){return k}),n.d(e,"CdkTrapFocus",function(){return A}),n.d(e,"InteractivityChecker",function(){return E}),n.d(e,"LIVE_ANNOUNCER_PROVIDER_FACTORY",function(){return N}),n.d(e,"LiveAnnouncer",function(){return P}),n.d(e,"CdkAriaLive",function(){return R}),n.d(e,"LIVE_ANNOUNCER_PROVIDER",function(){return F}),n.d(e,"LIVE_ANNOUNCER_ELEMENT_TOKEN_FACTORY",function(){return I}),n.d(e,"LIVE_ANNOUNCER_ELEMENT_TOKEN",function(){return D}),n.d(e,"FOCUS_MONITOR_PROVIDER_FACTORY",function(){return H}),n.d(e,"TOUCH_BUFFER_MS",function(){return L}),n.d(e,"FocusMonitor",function(){return V}),n.d(e,"CdkMonitorFocus",function(){return z}),n.d(e,"FOCUS_MONITOR_PROVIDER",function(){return U}),n.d(e,"isFakeMousedownFromScreenReader",function(){return W}),n.d(e,"A11yModule",function(){return q});var r=n("./node_modules/@angular/common/fesm5/common.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/rxjs/_esm5/index.js"),a=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),s=n("./node_modules/rxjs/_esm5/operators/index.js"),u=n("./node_modules/tslib/tslib.es6.js"),c=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),l=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),d=n("./node_modules/@angular/cdk/esm5/observers.es5.js"),p=" ";function f(t,e){return(t.getAttribute(e)||"").match(/\S+/g)||[]}var h="cdk-describedby-message-container",m="cdk-describedby-message",g="cdk-describedby-host",y=0,_=new Map,v=null,b=function(){function t(t){this._document=t}return t.prototype.describe=function(t,e){this._canBeDescribed(t,e)&&(_.has(e)||this._createMessageElement(e),this._isElementDescribedByMessage(t,e)||this._addMessageReference(t,e))},t.prototype.removeDescription=function(t,e){if(this._canBeDescribed(t,e)){this._isElementDescribedByMessage(t,e)&&this._removeMessageReference(t,e);var n=_.get(e);n&&0===n.referenceCount&&this._deleteMessageElement(e),v&&0===v.childNodes.length&&this._deleteMessagesContainer()}},t.prototype.ngOnDestroy=function(){for(var t=this._document.querySelectorAll("["+g+"]"),e=0;e<t.length;e++)this._removeCdkDescribedByReferenceIds(t[e]),t[e].removeAttribute(g);v&&this._deleteMessagesContainer(),_.clear()},t.prototype._createMessageElement=function(t){var e=this._document.createElement("div");e.setAttribute("id",m+"-"+y++),e.appendChild(this._document.createTextNode(t)),this._createMessagesContainer(),v.appendChild(e),_.set(t,{messageElement:e,referenceCount:0})},t.prototype._deleteMessageElement=function(t){var e=_.get(t),n=e&&e.messageElement;v&&n&&v.removeChild(n),_.delete(t)},t.prototype._createMessagesContainer=function(){if(!v){var t=this._document.getElementById(h);t&&t.parentNode.removeChild(t),(v=this._document.createElement("div")).id=h,v.setAttribute("aria-hidden","true"),v.style.display="none",this._document.body.appendChild(v)}},t.prototype._deleteMessagesContainer=function(){v&&v.parentNode&&(v.parentNode.removeChild(v),v=null)},t.prototype._removeCdkDescribedByReferenceIds=function(t){var e=f(t,"aria-describedby").filter(function(t){return 0!=t.indexOf(m)});t.setAttribute("aria-describedby",e.join(" "))},t.prototype._addMessageReference=function(t,e){var n,r,i,o,a=_.get(e);n=t,r="aria-describedby",i=a.messageElement.id,(o=f(n,r)).some(function(t){return t.trim()==i.trim()})||(o.push(i.trim()),n.setAttribute(r,o.join(p))),t.setAttribute(g,""),a.referenceCount++},t.prototype._removeMessageReference=function(t,e){var n,r,i,o,a=_.get(e);a.referenceCount--,n=t,r="aria-describedby",i=a.messageElement.id,o=f(n,r).filter(function(t){return t!=i.trim()}),n.setAttribute(r,o.join(p)),t.removeAttribute(g)},t.prototype._isElementDescribedByMessage=function(t,e){var n=f(t,"aria-describedby"),r=_.get(e),i=r&&r.messageElement.id;return!!i&&-1!=n.indexOf(i)},t.prototype._canBeDescribed=function(t,e){return t.nodeType===this._document.ELEMENT_NODE&&null!=e&&!!(""+e).trim()},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:i.Inject,args:[r.DOCUMENT]}]}]},t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t(Object(i.inject)(r.DOCUMENT))},token:t,providedIn:"root"}),t}();function x(t,e){return t||new b(e)}var w={provide:b,deps:[[new i.Optional,new i.SkipSelf,b],r.DOCUMENT],useFactory:x},j=function(){function t(t){var e=this;this._items=t,this._activeItemIndex=-1,this._activeItem=null,this._wrap=!1,this._letterKeyStream=new o.Subject,this._typeaheadSubscription=o.Subscription.EMPTY,this._vertical=!0,this._allowedModifierKeys=[],this._skipPredicateFn=function(t){return t.disabled},this._pressedLetters=[],this.tabOut=new o.Subject,this.change=new o.Subject,t instanceof i.QueryList&&t.changes.subscribe(function(t){if(e._activeItem){var n=t.toArray().indexOf(e._activeItem);n>-1&&n!==e._activeItemIndex&&(e._activeItemIndex=n)}})}return t.prototype.skipPredicate=function(t){return this._skipPredicateFn=t,this},t.prototype.withWrap=function(t){return void 0===t&&(t=!0),this._wrap=t,this},t.prototype.withVerticalOrientation=function(t){return void 0===t&&(t=!0),this._vertical=t,this},t.prototype.withHorizontalOrientation=function(t){return this._horizontal=t,this},t.prototype.withAllowedModifierKeys=function(t){return this._allowedModifierKeys=t,this},t.prototype.withTypeAhead=function(t){var e=this;if(void 0===t&&(t=200),this._items.length&&this._items.some(function(t){return"function"!=typeof t.getLabel}))throw Error("ListKeyManager items in typeahead mode must implement the `getLabel` method.");return this._typeaheadSubscription.unsubscribe(),this._typeaheadSubscription=this._letterKeyStream.pipe(Object(s.tap)(function(t){return e._pressedLetters.push(t)}),Object(s.debounceTime)(t),Object(s.filter)(function(){return e._pressedLetters.length>0}),Object(s.map)(function(){return e._pressedLetters.join("")})).subscribe(function(t){for(var n=e._getItemsArray(),r=1;r<n.length+1;r++){var i=(e._activeItemIndex+r)%n.length,o=n[i];if(!e._skipPredicateFn(o)&&0===o.getLabel().toUpperCase().trim().indexOf(t)){e.setActiveItem(i);break}}e._pressedLetters=[]}),this},t.prototype.setActiveItem=function(t){var e=this._activeItemIndex;this.updateActiveItem(t),this._activeItemIndex!==e&&this.change.next(this._activeItemIndex)},t.prototype.onKeydown=function(t){var e=this,n=t.keyCode,r=["altKey","ctrlKey","metaKey","shiftKey"].every(function(n){return!t[n]||e._allowedModifierKeys.indexOf(n)>-1});switch(n){case a.TAB:return void this.tabOut.next();case a.DOWN_ARROW:if(this._vertical&&r){this.setNextItemActive();break}return;case a.UP_ARROW:if(this._vertical&&r){this.setPreviousItemActive();break}return;case a.RIGHT_ARROW:if(this._horizontal&&r){"rtl"===this._horizontal?this.setPreviousItemActive():this.setNextItemActive();break}return;case a.LEFT_ARROW:if(this._horizontal&&r){"rtl"===this._horizontal?this.setNextItemActive():this.setPreviousItemActive();break}return;default:return void((r||Object(a.hasModifierKey)(t,"shiftKey"))&&(t.key&&1===t.key.length?this._letterKeyStream.next(t.key.toLocaleUpperCase()):(n>=a.A&&n<=a.Z||n>=a.ZERO&&n<=a.NINE)&&this._letterKeyStream.next(String.fromCharCode(n))))}this._pressedLetters=[],t.preventDefault()},Object.defineProperty(t.prototype,"activeItemIndex",{get:function(){return this._activeItemIndex},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"activeItem",{get:function(){return this._activeItem},enumerable:!0,configurable:!0}),t.prototype.setFirstItemActive=function(){this._setActiveItemByIndex(0,1)},t.prototype.setLastItemActive=function(){this._setActiveItemByIndex(this._items.length-1,-1)},t.prototype.setNextItemActive=function(){this._activeItemIndex<0?this.setFirstItemActive():this._setActiveItemByDelta(1)},t.prototype.setPreviousItemActive=function(){this._activeItemIndex<0&&this._wrap?this.setLastItemActive():this._setActiveItemByDelta(-1)},t.prototype.updateActiveItem=function(t){var e=this._getItemsArray(),n="number"==typeof t?t:e.indexOf(t),r=e[n];this._activeItem=null==r?null:r,this._activeItemIndex=n},t.prototype.updateActiveItemIndex=function(t){this.updateActiveItem(t)},t.prototype._setActiveItemByDelta=function(t){this._wrap?this._setActiveInWrapMode(t):this._setActiveInDefaultMode(t)},t.prototype._setActiveInWrapMode=function(t){for(var e=this._getItemsArray(),n=1;n<=e.length;n++){var r=(this._activeItemIndex+t*n+e.length)%e.length,i=e[r];if(!this._skipPredicateFn(i))return void this.setActiveItem(r)}},t.prototype._setActiveInDefaultMode=function(t){this._setActiveItemByIndex(this._activeItemIndex+t,t)},t.prototype._setActiveItemByIndex=function(t,e){var n=this._getItemsArray();if(n[t]){for(;this._skipPredicateFn(n[t]);)if(!n[t+=e])return;this.setActiveItem(t)}},t.prototype._getItemsArray=function(){return this._items instanceof i.QueryList?this._items.toArray():this._items},t}(),O=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(u.__extends)(e,t),e.prototype.setActiveItem=function(e){this.activeItem&&this.activeItem.setInactiveStyles(),t.prototype.setActiveItem.call(this,e),this.activeItem&&this.activeItem.setActiveStyles()},e}(j),C=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._origin="program",e}return Object(u.__extends)(e,t),e.prototype.setFocusOrigin=function(t){return this._origin=t,this},e.prototype.setActiveItem=function(e){t.prototype.setActiveItem.call(this,e),this.activeItem&&this.activeItem.focus(this._origin)},e}(j),E=function(){function t(t){this._platform=t}return t.prototype.isDisabled=function(t){return t.hasAttribute("disabled")},t.prototype.isVisible=function(t){return!!((e=t).offsetWidth||e.offsetHeight||"function"==typeof e.getClientRects&&e.getClientRects().length)&&"visible"===getComputedStyle(t).visibility;var e},t.prototype.isTabbable=function(t){if(!this._platform.isBrowser)return!1;var e,n=function(t){try{return t.frameElement}catch(t){return null}}((e=t).ownerDocument&&e.ownerDocument.defaultView||window);if(n){var r=n&&n.nodeName.toLowerCase();if(-1===M(n))return!1;if((this._platform.BLINK||this._platform.WEBKIT)&&"object"===r)return!1;if((this._platform.BLINK||this._platform.WEBKIT)&&!this.isVisible(n))return!1}var i,o,a,s=t.nodeName.toLowerCase(),u=M(t);if(t.hasAttribute("contenteditable"))return-1!==u;if("iframe"===s)return!1;if("audio"===s){if(!t.hasAttribute("controls"))return!1;if(this._platform.BLINK)return!0}if("video"===s){if(!t.hasAttribute("controls")&&this._platform.TRIDENT)return!1;if(this._platform.BLINK||this._platform.FIREFOX)return!0}return("object"!==s||!this._platform.BLINK&&!this._platform.WEBKIT)&&((!this._platform.WEBKIT||!this._platform.IOS||(o=(i=t).nodeName.toLowerCase(),"text"===(a="input"===o&&i.type)||"password"===a||"select"===o||"textarea"===o))&&t.tabIndex>=0)},t.prototype.isFocusable=function(t){return function(t){if(e=t,n=e,"input"==n.nodeName.toLowerCase()&&"hidden"==e.type)return!1;var e,n;return o=t,a=o.nodeName.toLowerCase(),"input"===a||"select"===a||"button"===a||"textarea"===a||(r=t,i=r,"a"==i.nodeName.toLowerCase()&&r.hasAttribute("href"))||t.hasAttribute("contenteditable")||S(t);var r,i;var o,a}(t)&&!this.isDisabled(t)&&this.isVisible(t)},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:c.Platform}]},t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t(Object(i.inject)(c.Platform))},token:t,providedIn:"root"}),t}();function S(t){if(!t.hasAttribute("tabindex")||void 0===t.tabIndex)return!1;var e=t.getAttribute("tabindex");return"-32768"!=e&&!(!e||isNaN(parseInt(e,10)))}function M(t){if(!S(t))return null;var e=parseInt(t.getAttribute("tabindex")||"",10);return isNaN(e)?-1:e}var T=function(){function t(t,e,n,r,i){void 0===i&&(i=!1);var o=this;this._element=t,this._checker=e,this._ngZone=n,this._document=r,this._hasAttached=!1,this._startAnchorListener=function(){return o.focusLastTabbableElement()},this._endAnchorListener=function(){return o.focusFirstTabbableElement()},this._enabled=!0,i||this.attachAnchors()}return Object.defineProperty(t.prototype,"enabled",{get:function(){return this._enabled},set:function(t){this._enabled=t,this._startAnchor&&this._endAnchor&&(this._toggleAnchorTabIndex(t,this._startAnchor),this._toggleAnchorTabIndex(t,this._endAnchor))},enumerable:!0,configurable:!0}),t.prototype.destroy=function(){var t=this._startAnchor,e=this._endAnchor;t&&(t.removeEventListener("focus",this._startAnchorListener),t.parentNode&&t.parentNode.removeChild(t)),e&&(e.removeEventListener("focus",this._endAnchorListener),e.parentNode&&e.parentNode.removeChild(e)),this._startAnchor=this._endAnchor=null},t.prototype.attachAnchors=function(){var t=this;return!!this._hasAttached||(this._ngZone.runOutsideAngular(function(){t._startAnchor||(t._startAnchor=t._createAnchor(),t._startAnchor.addEventListener("focus",t._startAnchorListener)),t._endAnchor||(t._endAnchor=t._createAnchor(),t._endAnchor.addEventListener("focus",t._endAnchorListener))}),this._element.parentNode&&(this._element.parentNode.insertBefore(this._startAnchor,this._element),this._element.parentNode.insertBefore(this._endAnchor,this._element.nextSibling),this._hasAttached=!0),this._hasAttached)},t.prototype.focusInitialElementWhenReady=function(){var t=this;return new Promise(function(e){t._executeOnStable(function(){return e(t.focusInitialElement())})})},t.prototype.focusFirstTabbableElementWhenReady=function(){var t=this;return new Promise(function(e){t._executeOnStable(function(){return e(t.focusFirstTabbableElement())})})},t.prototype.focusLastTabbableElementWhenReady=function(){var t=this;return new Promise(function(e){t._executeOnStable(function(){return e(t.focusLastTabbableElement())})})},t.prototype._getRegionBoundary=function(t){for(var e=this._element.querySelectorAll("[cdk-focus-region-"+t+"], [cdkFocusRegion"+t+"], [cdk-focus-"+t+"]"),n=0;n<e.length;n++)e[n].hasAttribute("cdk-focus-"+t)?console.warn("Found use of deprecated attribute 'cdk-focus-"+t+"', use 'cdkFocusRegion"+t+"' instead. The deprecated attribute will be removed in 8.0.0.",e[n]):e[n].hasAttribute("cdk-focus-region-"+t)&&console.warn("Found use of deprecated attribute 'cdk-focus-region-"+t+"', use 'cdkFocusRegion"+t+"' instead. The deprecated attribute will be removed in 8.0.0.",e[n]);return"start"==t?e.length?e[0]:this._getFirstTabbableElement(this._element):e.length?e[e.length-1]:this._getLastTabbableElement(this._element)},t.prototype.focusInitialElement=function(){var t=this._element.querySelector("[cdk-focus-initial], [cdkFocusInitial]");return t?(t.hasAttribute("cdk-focus-initial")&&console.warn("Found use of deprecated attribute 'cdk-focus-initial', use 'cdkFocusInitial' instead. The deprecated attribute will be removed in 8.0.0",t),Object(i.isDevMode)()&&!this._checker.isFocusable(t)&&console.warn("Element matching '[cdkFocusInitial]' is not focusable.",t),t.focus(),!0):this.focusFirstTabbableElement()},t.prototype.focusFirstTabbableElement=function(){var t=this._getRegionBoundary("start");return t&&t.focus(),!!t},t.prototype.focusLastTabbableElement=function(){var t=this._getRegionBoundary("end");return t&&t.focus(),!!t},t.prototype.hasAttached=function(){return this._hasAttached},t.prototype._getFirstTabbableElement=function(t){if(this._checker.isFocusable(t)&&this._checker.isTabbable(t))return t;for(var e=t.children||t.childNodes,n=0;n<e.length;n++){var r=e[n].nodeType===this._document.ELEMENT_NODE?this._getFirstTabbableElement(e[n]):null;if(r)return r}return null},t.prototype._getLastTabbableElement=function(t){if(this._checker.isFocusable(t)&&this._checker.isTabbable(t))return t;for(var e=t.children||t.childNodes,n=e.length-1;n>=0;n--){var r=e[n].nodeType===this._document.ELEMENT_NODE?this._getLastTabbableElement(e[n]):null;if(r)return r}return null},t.prototype._createAnchor=function(){var t=this._document.createElement("div");return this._toggleAnchorTabIndex(this._enabled,t),t.classList.add("cdk-visually-hidden"),t.classList.add("cdk-focus-trap-anchor"),t},t.prototype._toggleAnchorTabIndex=function(t,e){t?e.setAttribute("tabindex","0"):e.removeAttribute("tabindex")},t.prototype._executeOnStable=function(t){this._ngZone.isStable?t():this._ngZone.onStable.asObservable().pipe(Object(s.take)(1)).subscribe(t)},t}(),k=function(){function t(t,e,n){this._checker=t,this._ngZone=e,this._document=n}return t.prototype.create=function(t,e){return void 0===e&&(e=!1),new T(t,this._checker,this._ngZone,this._document,e)},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:E},{type:i.NgZone},{type:void 0,decorators:[{type:i.Inject,args:[r.DOCUMENT]}]}]},t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t(Object(i.inject)(E),Object(i.inject)(i.NgZone),Object(i.inject)(r.DOCUMENT))},token:t,providedIn:"root"}),t}(),A=function(){function t(t,e,n){this._elementRef=t,this._focusTrapFactory=e,this._previouslyFocusedElement=null,this._document=n,this.focusTrap=this._focusTrapFactory.create(this._elementRef.nativeElement,!0)}return Object.defineProperty(t.prototype,"enabled",{get:function(){return this.focusTrap.enabled},set:function(t){this.focusTrap.enabled=Object(l.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"autoCapture",{get:function(){return this._autoCapture},set:function(t){this._autoCapture=Object(l.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){this.focusTrap.destroy(),this._previouslyFocusedElement&&(this._previouslyFocusedElement.focus(),this._previouslyFocusedElement=null)},t.prototype.ngAfterContentInit=function(){this.focusTrap.attachAnchors(),this.autoCapture&&(this._previouslyFocusedElement=this._document.activeElement,this.focusTrap.focusInitialElementWhenReady())},t.prototype.ngDoCheck=function(){this.focusTrap.hasAttached()||this.focusTrap.attachAnchors()},t.decorators=[{type:i.Directive,args:[{selector:"[cdkTrapFocus]",exportAs:"cdkTrapFocus"}]}],t.ctorParameters=function(){return[{type:i.ElementRef},{type:k},{type:void 0,decorators:[{type:i.Inject,args:[r.DOCUMENT]}]}]},t.propDecorators={enabled:[{type:i.Input,args:["cdkTrapFocus"]}],autoCapture:[{type:i.Input,args:["cdkTrapFocusAutoCapture"]}]},t}(),D=new i.InjectionToken("liveAnnouncerElement",{providedIn:"root",factory:I});function I(){return null}var P=function(){function t(t,e,n){this._ngZone=e,this._document=n,this._liveElement=t||this._createLiveElement()}return t.prototype.announce=function(t){for(var e,n,r=this,i=[],o=1;o<arguments.length;o++)i[o-1]=arguments[o];return 1===i.length&&"number"==typeof i[0]?n=i[0]:(e=i[0],n=i[1]),this.clear(),clearTimeout(this._previousTimeout),this._liveElement.setAttribute("aria-live",e||"polite"),this._ngZone.runOutsideAngular(function(){return new Promise(function(e){clearTimeout(r._previousTimeout),r._previousTimeout=setTimeout(function(){r._liveElement.textContent=t,e(),"number"==typeof n&&(r._previousTimeout=setTimeout(function(){return r.clear()},n))},100)})})},t.prototype.clear=function(){this._liveElement&&(this._liveElement.textContent="")},t.prototype.ngOnDestroy=function(){clearTimeout(this._previousTimeout),this._liveElement&&this._liveElement.parentNode&&(this._liveElement.parentNode.removeChild(this._liveElement),this._liveElement=null)},t.prototype._createLiveElement=function(){for(var t="cdk-live-announcer-element",e=this._document.getElementsByClassName(t),n=this._document.createElement("div"),r=0;r<e.length;r++)e[r].parentNode.removeChild(e[r]);return n.classList.add(t),n.classList.add("cdk-visually-hidden"),n.setAttribute("aria-atomic","true"),n.setAttribute("aria-live","polite"),this._document.body.appendChild(n),n},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:i.Optional},{type:i.Inject,args:[D]}]},{type:i.NgZone},{type:void 0,decorators:[{type:i.Inject,args:[r.DOCUMENT]}]}]},t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t(Object(i.inject)(D,8),Object(i.inject)(i.NgZone),Object(i.inject)(r.DOCUMENT))},token:t,providedIn:"root"}),t}(),R=function(){function t(t,e,n,r){this._elementRef=t,this._liveAnnouncer=e,this._contentObserver=n,this._ngZone=r,this._politeness="off"}return Object.defineProperty(t.prototype,"politeness",{get:function(){return this._politeness},set:function(t){var e=this;this._politeness="polite"===t||"assertive"===t?t:"off","off"===this._politeness?this._subscription&&(this._subscription.unsubscribe(),this._subscription=null):this._subscription||(this._subscription=this._ngZone.runOutsideAngular(function(){return e._contentObserver.observe(e._elementRef).subscribe(function(){var t=e._elementRef.nativeElement.textContent;t!==e._previousAnnouncedText&&(e._liveAnnouncer.announce(t,e._politeness),e._previousAnnouncedText=t)})}))},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){this._subscription&&this._subscription.unsubscribe()},t.decorators=[{type:i.Directive,args:[{selector:"[cdkAriaLive]",exportAs:"cdkAriaLive"}]}],t.ctorParameters=function(){return[{type:i.ElementRef},{type:P},{type:d.ContentObserver},{type:i.NgZone}]},t.propDecorators={politeness:[{type:i.Input,args:["cdkAriaLive"]}]},t}();function N(t,e,n,r){return t||new P(e,r,n)}var F={provide:P,deps:[[new i.Optional,new i.SkipSelf,P],[new i.Optional,new i.Inject(D)],r.DOCUMENT,i.NgZone],useFactory:N},L=650,B=Object(c.normalizePassiveListenerOptions)({passive:!0,capture:!0}),V=function(){function t(t,e){var n=this;this._ngZone=t,this._platform=e,this._origin=null,this._windowFocused=!1,this._elementInfo=new Map,this._monitoredElementCount=0,this._documentKeydownListener=function(){n._lastTouchTarget=null,n._setOriginForCurrentEventQueue("keyboard")},this._documentMousedownListener=function(){n._lastTouchTarget||n._setOriginForCurrentEventQueue("mouse")},this._documentTouchstartListener=function(t){null!=n._touchTimeoutId&&clearTimeout(n._touchTimeoutId),n._lastTouchTarget=t.target,n._touchTimeoutId=setTimeout(function(){return n._lastTouchTarget=null},L)},this._windowFocusListener=function(){n._windowFocused=!0,n._windowFocusTimeoutId=setTimeout(function(){return n._windowFocused=!1})}}return t.prototype.monitor=function(t,e){var n=this;if(void 0===e&&(e=!1),!this._platform.isBrowser)return Object(o.of)(null);var r=Object(l.coerceElement)(t);if(this._elementInfo.has(r)){var i=this._elementInfo.get(r);return i.checkChildren=e,i.subject.asObservable()}var a={unlisten:function(){},checkChildren:e,subject:new o.Subject};this._elementInfo.set(r,a),this._incrementMonitoredElementCount();var s=function(t){return n._onFocus(t,r)},u=function(t){return n._onBlur(t,r)};return this._ngZone.runOutsideAngular(function(){r.addEventListener("focus",s,!0),r.addEventListener("blur",u,!0)}),a.unlisten=function(){r.removeEventListener("focus",s,!0),r.removeEventListener("blur",u,!0)},a.subject.asObservable()},t.prototype.stopMonitoring=function(t){var e=Object(l.coerceElement)(t),n=this._elementInfo.get(e);n&&(n.unlisten(),n.subject.complete(),this._setClasses(e),this._elementInfo.delete(e),this._decrementMonitoredElementCount())},t.prototype.focusVia=function(t,e,n){var r=Object(l.coerceElement)(t);this._setOriginForCurrentEventQueue(e),"function"==typeof r.focus&&r.focus(n)},t.prototype.ngOnDestroy=function(){var t=this;this._elementInfo.forEach(function(e,n){return t.stopMonitoring(n)})},t.prototype._toggleClass=function(t,e,n){n?t.classList.add(e):t.classList.remove(e)},t.prototype._setClasses=function(t,e){this._elementInfo.get(t)&&(this._toggleClass(t,"cdk-focused",!!e),this._toggleClass(t,"cdk-touch-focused","touch"===e),this._toggleClass(t,"cdk-keyboard-focused","keyboard"===e),this._toggleClass(t,"cdk-mouse-focused","mouse"===e),this._toggleClass(t,"cdk-program-focused","program"===e))},t.prototype._setOriginForCurrentEventQueue=function(t){var e=this;this._ngZone.runOutsideAngular(function(){e._origin=t,e._originTimeoutId=setTimeout(function(){return e._origin=null},1)})},t.prototype._wasCausedByTouch=function(t){var e=t.target;return this._lastTouchTarget instanceof Node&&e instanceof Node&&(e===this._lastTouchTarget||e.contains(this._lastTouchTarget))},t.prototype._onFocus=function(t,e){var n=this._elementInfo.get(e);if(n&&(n.checkChildren||e===t.target)){var r=this._origin;r||(r=this._windowFocused&&this._lastFocusOrigin?this._lastFocusOrigin:this._wasCausedByTouch(t)?"touch":"program"),this._setClasses(e,r),this._emitOrigin(n.subject,r),this._lastFocusOrigin=r}},t.prototype._onBlur=function(t,e){var n=this._elementInfo.get(e);!n||n.checkChildren&&t.relatedTarget instanceof Node&&e.contains(t.relatedTarget)||(this._setClasses(e),this._emitOrigin(n.subject,null))},t.prototype._emitOrigin=function(t,e){this._ngZone.run(function(){return t.next(e)})},t.prototype._incrementMonitoredElementCount=function(){var t=this;1==++this._monitoredElementCount&&this._platform.isBrowser&&this._ngZone.runOutsideAngular(function(){document.addEventListener("keydown",t._documentKeydownListener,B),document.addEventListener("mousedown",t._documentMousedownListener,B),document.addEventListener("touchstart",t._documentTouchstartListener,B),window.addEventListener("focus",t._windowFocusListener)})},t.prototype._decrementMonitoredElementCount=function(){--this._monitoredElementCount||(document.removeEventListener("keydown",this._documentKeydownListener,B),document.removeEventListener("mousedown",this._documentMousedownListener,B),document.removeEventListener("touchstart",this._documentTouchstartListener,B),window.removeEventListener("focus",this._windowFocusListener),clearTimeout(this._windowFocusTimeoutId),clearTimeout(this._touchTimeoutId),clearTimeout(this._originTimeoutId))},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:i.NgZone},{type:c.Platform}]},t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t(Object(i.inject)(i.NgZone),Object(i.inject)(c.Platform))},token:t,providedIn:"root"}),t}(),z=function(){function t(t,e){var n=this;this._elementRef=t,this._focusMonitor=e,this.cdkFocusChange=new i.EventEmitter,this._monitorSubscription=this._focusMonitor.monitor(this._elementRef,this._elementRef.nativeElement.hasAttribute("cdkMonitorSubtreeFocus")).subscribe(function(t){return n.cdkFocusChange.emit(t)})}return t.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef),this._monitorSubscription.unsubscribe()},t.decorators=[{type:i.Directive,args:[{selector:"[cdkMonitorElementFocus], [cdkMonitorSubtreeFocus]"}]}],t.ctorParameters=function(){return[{type:i.ElementRef},{type:V}]},t.propDecorators={cdkFocusChange:[{type:i.Output}]},t}();function H(t,e,n){return t||new V(e,n)}var U={provide:V,deps:[[new i.Optional,new i.SkipSelf,V],i.NgZone,c.Platform],useFactory:H};function W(t){return 0===t.buttons}var q=function(){function t(){}return t.decorators=[{type:i.NgModule,args:[{imports:[r.CommonModule,c.PlatformModule,d.ObserversModule],declarations:[R,A,z],exports:[R,A,z]}]}],t}()},"./node_modules/@angular/cdk/esm5/accordion.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"CdkAccordionItem",function(){return l}),n.d(e,"CdkAccordion",function(){return u}),n.d(e,"CdkAccordionModule",function(){return d});var r=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/rxjs/_esm5/index.js"),a=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),s=0,u=function(){function t(){this._stateChanges=new o.Subject,this._openCloseAllActions=new o.Subject,this.id="cdk-accordion-"+s++,this._multi=!1}return Object.defineProperty(t.prototype,"multi",{get:function(){return this._multi},set:function(t){this._multi=Object(r.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),t.prototype.openAll=function(){this._openCloseAll(!0)},t.prototype.closeAll=function(){this._openCloseAll(!1)},t.prototype.ngOnChanges=function(t){this._stateChanges.next(t)},t.prototype.ngOnDestroy=function(){this._stateChanges.complete()},t.prototype._openCloseAll=function(t){this.multi&&this._openCloseAllActions.next(t)},t.decorators=[{type:i.Directive,args:[{selector:"cdk-accordion, [cdkAccordion]",exportAs:"cdkAccordion"}]}],t.propDecorators={multi:[{type:i.Input}]},t}(),c=0,l=function(){function t(t,e,n){var r=this;this.accordion=t,this._changeDetectorRef=e,this._expansionDispatcher=n,this._openCloseAllSubscription=o.Subscription.EMPTY,this.closed=new i.EventEmitter,this.opened=new i.EventEmitter,this.destroyed=new i.EventEmitter,this.expandedChange=new i.EventEmitter,this.id="cdk-accordion-child-"+c++,this._expanded=!1,this._disabled=!1,this._removeUniqueSelectionListener=function(){},this._removeUniqueSelectionListener=n.listen(function(t,e){r.accordion&&!r.accordion.multi&&r.accordion.id===e&&r.id!==t&&(r.expanded=!1)}),this.accordion&&(this._openCloseAllSubscription=this._subscribeToOpenCloseAllActions())}return Object.defineProperty(t.prototype,"expanded",{get:function(){return this._expanded},set:function(t){if(t=Object(r.coerceBooleanProperty)(t),this._expanded!==t){if(this._expanded=t,this.expandedChange.emit(t),t){this.opened.emit();var e=this.accordion?this.accordion.id:this.id;this._expansionDispatcher.notify(this.id,e)}else this.closed.emit();this._changeDetectorRef.markForCheck()}},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this._disabled},set:function(t){this._disabled=Object(r.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){this.opened.complete(),this.closed.complete(),this.destroyed.emit(),this.destroyed.complete(),this._removeUniqueSelectionListener(),this._openCloseAllSubscription.unsubscribe()},t.prototype.toggle=function(){this.disabled||(this.expanded=!this.expanded)},t.prototype.close=function(){this.disabled||(this.expanded=!1)},t.prototype.open=function(){this.disabled||(this.expanded=!0)},t.prototype._subscribeToOpenCloseAllActions=function(){var t=this;return this.accordion._openCloseAllActions.subscribe(function(e){t.disabled||(t.expanded=e)})},t.decorators=[{type:i.Directive,args:[{selector:"cdk-accordion-item, [cdkAccordionItem]",exportAs:"cdkAccordionItem",providers:[{provide:u,useValue:void 0}]}]}],t.ctorParameters=function(){return[{type:u,decorators:[{type:i.Optional},{type:i.SkipSelf}]},{type:i.ChangeDetectorRef},{type:a.UniqueSelectionDispatcher}]},t.propDecorators={closed:[{type:i.Output}],opened:[{type:i.Output}],destroyed:[{type:i.Output}],expandedChange:[{type:i.Output}],expanded:[{type:i.Input}],disabled:[{type:i.Input}]},t}(),d=function(){function t(){}return t.decorators=[{type:i.NgModule,args:[{exports:[u,l],declarations:[u,l]}]}],t}()},"./node_modules/@angular/cdk/esm5/bidi.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"Directionality",function(){return s}),n.d(e,"DIR_DOCUMENT",function(){return o}),n.d(e,"Dir",function(){return u}),n.d(e,"BidiModule",function(){return c}),n.d(e,"ɵa",function(){return a});var r=n("./node_modules/@angular/common/fesm5/common.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=new i.InjectionToken("cdk-dir-doc",{providedIn:"root",factory:a});function a(){return Object(i.inject)(r.DOCUMENT)}var s=function(){function t(t){if(this.value="ltr",this.change=new i.EventEmitter,t){var e=t.body?t.body.dir:null,n=t.documentElement?t.documentElement.dir:null,r=e||n;this.value="ltr"===r||"rtl"===r?r:"ltr"}}return t.prototype.ngOnDestroy=function(){this.change.complete()},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:i.Optional},{type:i.Inject,args:[o]}]}]},t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t(Object(i.inject)(o,8))},token:t,providedIn:"root"}),t}(),u=function(){function t(){this._dir="ltr",this._isInitialized=!1,this.change=new i.EventEmitter}return Object.defineProperty(t.prototype,"dir",{get:function(){return this._dir},set:function(t){var e=this._dir;this._rawDir=t,this._dir="ltr"===t||"rtl"===t?t:"ltr",e!==this._dir&&this._isInitialized&&this.change.emit(this._dir)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{get:function(){return this.dir},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this._isInitialized=!0},t.prototype.ngOnDestroy=function(){this.change.complete()},t.decorators=[{type:i.Directive,args:[{selector:"[dir]",providers:[{provide:s,useExisting:t}],host:{"[attr.dir]":"_rawDir"},exportAs:"dir"}]}],t.propDecorators={change:[{type:i.Output,args:["dirChange"]}],dir:[{type:i.Input}]},t}(),c=function(){function t(){}return t.decorators=[{type:i.NgModule,args:[{exports:[u],declarations:[u]}]}],t}()},"./node_modules/@angular/cdk/esm5/coercion.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"coerceBooleanProperty",function(){return i}),n.d(e,"coerceNumberProperty",function(){return o}),n.d(e,"_isNumberValue",function(){return a}),n.d(e,"coerceArray",function(){return s}),n.d(e,"coerceCssPixelValue",function(){return u}),n.d(e,"coerceElement",function(){return c});var r=n("./node_modules/@angular/core/fesm5/core.js");function i(t){return null!=t&&""+t!="false"}function o(t,e){return void 0===e&&(e=0),a(t)?Number(t):e}function a(t){return!isNaN(parseFloat(t))&&!isNaN(Number(t))}function s(t){return Array.isArray(t)?t:[t]}function u(t){return null==t?"":"string"==typeof t?t:t+"px"}function c(t){return t instanceof r.ElementRef?t.nativeElement:t}},"./node_modules/@angular/cdk/esm5/collections.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"UniqueSelectionDispatcher",function(){return l}),n.d(e,"ArrayDataSource",function(){return s}),n.d(e,"DataSource",function(){return a}),n.d(e,"getMultipleValuesInSingleSelectionError",function(){return c}),n.d(e,"SelectionModel",function(){return u});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/index.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=function(){return function(){}}(),s=function(t){function e(e){var n=t.call(this)||this;return n._data=e,n}return Object(r.__extends)(e,t),e.prototype.connect=function(){return this._data instanceof i.Observable?this._data:Object(i.of)(this._data)},e.prototype.disconnect=function(){},e}(a),u=function(){function t(t,e,n){void 0===t&&(t=!1),void 0===n&&(n=!0);var r=this;this._multiple=t,this._emitChanges=n,this._selection=new Set,this._deselectedToEmit=[],this._selectedToEmit=[],this.changed=new i.Subject,this.onChange=this.changed,e&&e.length&&(t?e.forEach(function(t){return r._markSelected(t)}):this._markSelected(e[0]),this._selectedToEmit.length=0)}return Object.defineProperty(t.prototype,"selected",{get:function(){return this._selected||(this._selected=Array.from(this._selection.values())),this._selected},enumerable:!0,configurable:!0}),t.prototype.select=function(){for(var t=this,e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];this._verifyValueAssignment(e),e.forEach(function(e){return t._markSelected(e)}),this._emitChangeEvent()},t.prototype.deselect=function(){for(var t=this,e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];this._verifyValueAssignment(e),e.forEach(function(e){return t._unmarkSelected(e)}),this._emitChangeEvent()},t.prototype.toggle=function(t){this.isSelected(t)?this.deselect(t):this.select(t)},t.prototype.clear=function(){this._unmarkAll(),this._emitChangeEvent()},t.prototype.isSelected=function(t){return this._selection.has(t)},t.prototype.isEmpty=function(){return 0===this._selection.size},t.prototype.hasValue=function(){return!this.isEmpty()},t.prototype.sort=function(t){this._multiple&&this.selected&&this._selected.sort(t)},t.prototype.isMultipleSelection=function(){return this._multiple},t.prototype._emitChangeEvent=function(){this._selected=null,(this._selectedToEmit.length||this._deselectedToEmit.length)&&(this.changed.next({source:this,added:this._selectedToEmit,removed:this._deselectedToEmit}),this._deselectedToEmit=[],this._selectedToEmit=[])},t.prototype._markSelected=function(t){this.isSelected(t)||(this._multiple||this._unmarkAll(),this._selection.add(t),this._emitChanges&&this._selectedToEmit.push(t))},t.prototype._unmarkSelected=function(t){this.isSelected(t)&&(this._selection.delete(t),this._emitChanges&&this._deselectedToEmit.push(t))},t.prototype._unmarkAll=function(){var t=this;this.isEmpty()||this._selection.forEach(function(e){return t._unmarkSelected(e)})},t.prototype._verifyValueAssignment=function(t){if(t.length>1&&!this._multiple)throw c()},t}();function c(){return Error("Cannot pass multiple values into SelectionModel with single-value mode.")}var l=function(){function t(){this._listeners=[]}return t.prototype.notify=function(t,e){for(var n=0,r=this._listeners;n<r.length;n++){(0,r[n])(t,e)}},t.prototype.listen=function(t){var e=this;return this._listeners.push(t),function(){e._listeners=e._listeners.filter(function(e){return t!==e})}},t.prototype.ngOnDestroy=function(){this._listeners=[]},t.decorators=[{type:o.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(o.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}()},"./node_modules/@angular/cdk/esm5/keycodes.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MAC_ENTER",function(){return r}),n.d(e,"BACKSPACE",function(){return i}),n.d(e,"TAB",function(){return o}),n.d(e,"NUM_CENTER",function(){return a}),n.d(e,"ENTER",function(){return s}),n.d(e,"SHIFT",function(){return u}),n.d(e,"CONTROL",function(){return c}),n.d(e,"ALT",function(){return l}),n.d(e,"PAUSE",function(){return d}),n.d(e,"CAPS_LOCK",function(){return p}),n.d(e,"ESCAPE",function(){return f}),n.d(e,"SPACE",function(){return h}),n.d(e,"PAGE_UP",function(){return m}),n.d(e,"PAGE_DOWN",function(){return g}),n.d(e,"END",function(){return y}),n.d(e,"HOME",function(){return _}),n.d(e,"LEFT_ARROW",function(){return v}),n.d(e,"UP_ARROW",function(){return b}),n.d(e,"RIGHT_ARROW",function(){return x}),n.d(e,"DOWN_ARROW",function(){return w}),n.d(e,"PLUS_SIGN",function(){return j}),n.d(e,"PRINT_SCREEN",function(){return O}),n.d(e,"INSERT",function(){return C}),n.d(e,"DELETE",function(){return E}),n.d(e,"ZERO",function(){return S}),n.d(e,"ONE",function(){return M}),n.d(e,"TWO",function(){return T}),n.d(e,"THREE",function(){return k}),n.d(e,"FOUR",function(){return A}),n.d(e,"FIVE",function(){return D}),n.d(e,"SIX",function(){return I}),n.d(e,"SEVEN",function(){return P}),n.d(e,"EIGHT",function(){return R}),n.d(e,"NINE",function(){return N}),n.d(e,"FF_SEMICOLON",function(){return F}),n.d(e,"FF_EQUALS",function(){return L}),n.d(e,"QUESTION_MARK",function(){return B}),n.d(e,"AT_SIGN",function(){return V}),n.d(e,"A",function(){return z}),n.d(e,"B",function(){return H}),n.d(e,"C",function(){return U}),n.d(e,"D",function(){return W}),n.d(e,"E",function(){return q}),n.d(e,"F",function(){return G}),n.d(e,"G",function(){return Y}),n.d(e,"H",function(){return K}),n.d(e,"I",function(){return X}),n.d(e,"J",function(){return Q}),n.d(e,"K",function(){return Z}),n.d(e,"L",function(){return $}),n.d(e,"M",function(){return J}),n.d(e,"N",function(){return tt}),n.d(e,"O",function(){return et}),n.d(e,"P",function(){return nt}),n.d(e,"Q",function(){return rt}),n.d(e,"R",function(){return it}),n.d(e,"S",function(){return ot}),n.d(e,"T",function(){return at}),n.d(e,"U",function(){return st}),n.d(e,"V",function(){return ut}),n.d(e,"W",function(){return ct}),n.d(e,"X",function(){return lt}),n.d(e,"Y",function(){return dt}),n.d(e,"Z",function(){return pt}),n.d(e,"META",function(){return ft}),n.d(e,"MAC_WK_CMD_LEFT",function(){return ht}),n.d(e,"MAC_WK_CMD_RIGHT",function(){return mt}),n.d(e,"CONTEXT_MENU",function(){return gt}),n.d(e,"NUMPAD_ZERO",function(){return yt}),n.d(e,"NUMPAD_ONE",function(){return _t}),n.d(e,"NUMPAD_TWO",function(){return vt}),n.d(e,"NUMPAD_THREE",function(){return bt}),n.d(e,"NUMPAD_FOUR",function(){return xt}),n.d(e,"NUMPAD_FIVE",function(){return wt}),n.d(e,"NUMPAD_SIX",function(){return jt}),n.d(e,"NUMPAD_SEVEN",function(){return Ot}),n.d(e,"NUMPAD_EIGHT",function(){return Ct}),n.d(e,"NUMPAD_NINE",function(){return Et}),n.d(e,"NUMPAD_MULTIPLY",function(){return St}),n.d(e,"NUMPAD_PLUS",function(){return Mt}),n.d(e,"NUMPAD_MINUS",function(){return Tt}),n.d(e,"NUMPAD_PERIOD",function(){return kt}),n.d(e,"NUMPAD_DIVIDE",function(){return At}),n.d(e,"F1",function(){return Dt}),n.d(e,"F2",function(){return It}),n.d(e,"F3",function(){return Pt}),n.d(e,"F4",function(){return Rt}),n.d(e,"F5",function(){return Nt}),n.d(e,"F6",function(){return Ft}),n.d(e,"F7",function(){return Lt}),n.d(e,"F8",function(){return Bt}),n.d(e,"F9",function(){return Vt}),n.d(e,"F10",function(){return zt}),n.d(e,"F11",function(){return Ht}),n.d(e,"F12",function(){return Ut}),n.d(e,"NUM_LOCK",function(){return Wt}),n.d(e,"SCROLL_LOCK",function(){return qt}),n.d(e,"FIRST_MEDIA",function(){return Gt}),n.d(e,"FF_MINUS",function(){return Yt}),n.d(e,"MUTE",function(){return Kt}),n.d(e,"VOLUME_DOWN",function(){return Xt}),n.d(e,"VOLUME_UP",function(){return Qt}),n.d(e,"FF_MUTE",function(){return Zt}),n.d(e,"FF_VOLUME_DOWN",function(){return $t}),n.d(e,"LAST_MEDIA",function(){return Jt}),n.d(e,"FF_VOLUME_UP",function(){return te}),n.d(e,"SEMICOLON",function(){return ee}),n.d(e,"EQUALS",function(){return ne}),n.d(e,"COMMA",function(){return re}),n.d(e,"DASH",function(){return ie}),n.d(e,"SLASH",function(){return oe}),n.d(e,"APOSTROPHE",function(){return ae}),n.d(e,"TILDE",function(){return se}),n.d(e,"OPEN_SQUARE_BRACKET",function(){return ue}),n.d(e,"BACKSLASH",function(){return ce}),n.d(e,"CLOSE_SQUARE_BRACKET",function(){return le}),n.d(e,"SINGLE_QUOTE",function(){return de}),n.d(e,"MAC_META",function(){return pe}),n.d(e,"hasModifierKey",function(){return fe});var r=3,i=8,o=9,a=12,s=13,u=16,c=17,l=18,d=19,p=20,f=27,h=32,m=33,g=34,y=35,_=36,v=37,b=38,x=39,w=40,j=43,O=44,C=45,E=46,S=48,M=49,T=50,k=51,A=52,D=53,I=54,P=55,R=56,N=57,F=59,L=61,B=63,V=64,z=65,H=66,U=67,W=68,q=69,G=70,Y=71,K=72,X=73,Q=74,Z=75,$=76,J=77,tt=78,et=79,nt=80,rt=81,it=82,ot=83,at=84,st=85,ut=86,ct=87,lt=88,dt=89,pt=90,ft=91,ht=91,mt=93,gt=93,yt=96,_t=97,vt=98,bt=99,xt=100,wt=101,jt=102,Ot=103,Ct=104,Et=105,St=106,Mt=107,Tt=109,kt=110,At=111,Dt=112,It=113,Pt=114,Rt=115,Nt=116,Ft=117,Lt=118,Bt=119,Vt=120,zt=121,Ht=122,Ut=123,Wt=144,qt=145,Gt=166,Yt=173,Kt=173,Xt=174,Qt=175,Zt=181,$t=182,Jt=183,te=183,ee=186,ne=187,re=188,ie=189,oe=191,ae=192,se=192,ue=219,ce=220,le=221,de=222,pe=224;function fe(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];return e.length?e.some(function(e){return t[e]}):t.altKey||t.shiftKey||t.ctrlKey||t.metaKey}},"./node_modules/@angular/cdk/esm5/layout.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"LayoutModule",function(){return c}),n.d(e,"BreakpointObserver",function(){return f}),n.d(e,"Breakpoints",function(){return m}),n.d(e,"MediaMatcher",function(){return d});var r,i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),a=n("./node_modules/rxjs/_esm5/index.js"),s=n("./node_modules/rxjs/_esm5/operators/index.js"),u=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),c=function(){function t(){}return t.decorators=[{type:i.NgModule,args:[{}]}],t}(),l=new Set,d=function(){function t(t){this.platform=t,this._matchMedia=this.platform.isBrowser&&window.matchMedia?window.matchMedia.bind(window):p}return t.prototype.matchMedia=function(t){return this.platform.WEBKIT&&function(t){if(l.has(t))return;try{r||((r=document.createElement("style")).setAttribute("type","text/css"),document.head.appendChild(r)),r.sheet&&(r.sheet.insertRule("@media "+t+" {.fx-query-test{ }}",0),l.add(t))}catch(t){console.error(t)}}(t),this._matchMedia(t)},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:o.Platform}]},t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t(Object(i.inject)(o.Platform))},token:t,providedIn:"root"}),t}();function p(t){return{matches:"all"===t||""===t,media:t,addListener:function(){},removeListener:function(){}}}var f=function(){function t(t,e){this.mediaMatcher=t,this.zone=e,this._queries=new Map,this._destroySubject=new a.Subject}return t.prototype.ngOnDestroy=function(){this._destroySubject.next(),this._destroySubject.complete()},t.prototype.isMatched=function(t){var e=this;return h(Object(u.coerceArray)(t)).some(function(t){return e._registerQuery(t).mql.matches})},t.prototype.observe=function(t){var e=this,n=h(Object(u.coerceArray)(t)).map(function(t){return e._registerQuery(t).observable});return Object(a.combineLatest)(n).pipe(Object(s.debounceTime)(0,a.asapScheduler),Object(s.map)(function(t){var e={matches:!1,breakpoints:{}};return t.forEach(function(t){e.matches=e.matches||t.matches,e.breakpoints[t.query]=t.matches}),e}))},t.prototype._registerQuery=function(t){var e=this;if(this._queries.has(t))return this._queries.get(t);var n,r=this.mediaMatcher.matchMedia(t),i={observable:Object(a.fromEventPattern)(function(t){n=function(n){return e.zone.run(function(){return t(n)})},r.addListener(n)},function(){return r.removeListener(n)}).pipe(Object(s.startWith)(r),Object(s.map)(function(e){return{query:t,matches:e.matches}}),Object(s.takeUntil)(this._destroySubject)),mql:r};return this._queries.set(t,i),i},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:d},{type:i.NgZone}]},t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t(Object(i.inject)(d),Object(i.inject)(i.NgZone))},token:t,providedIn:"root"}),t}();function h(t){return t.map(function(t){return t.split(",")}).reduce(function(t,e){return t.concat(e)}).map(function(t){return t.trim()})}var m={XSmall:"(max-width: 599.99px)",Small:"(min-width: 600px) and (max-width: 959.99px)",Medium:"(min-width: 960px) and (max-width: 1279.99px)",Large:"(min-width: 1280px) and (max-width: 1919.99px)",XLarge:"(min-width: 1920px)",Handset:"(max-width: 599.99px) and (orientation: portrait), (max-width: 959.99px) and (orientation: landscape)",Tablet:"(min-width: 600px) and (max-width: 839.99px) and (orientation: portrait), (min-width: 960px) and (max-width: 1279.99px) and (orientation: landscape)",Web:"(min-width: 840px) and (orientation: portrait), (min-width: 1280px) and (orientation: landscape)",HandsetPortrait:"(max-width: 599.99px) and (orientation: portrait)",TabletPortrait:"(min-width: 600px) and (max-width: 839.99px) and (orientation: portrait)",WebPortrait:"(min-width: 840px) and (orientation: portrait)",HandsetLandscape:"(max-width: 959.99px) and (orientation: landscape)",TabletLandscape:"(min-width: 960px) and (max-width: 1279.99px) and (orientation: landscape)",WebLandscape:"(min-width: 1280px) and (orientation: landscape)"}},"./node_modules/@angular/cdk/esm5/observers.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MutationObserverFactory",function(){return s}),n.d(e,"ContentObserver",function(){return u}),n.d(e,"CdkObserveContent",function(){return c}),n.d(e,"ObserversModule",function(){return l});var r=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/rxjs/_esm5/index.js"),a=n("./node_modules/rxjs/_esm5/operators/index.js"),s=function(){function t(){}return t.prototype.create=function(t){return"undefined"==typeof MutationObserver?null:new MutationObserver(t)},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(),u=function(){function t(t){this._mutationObserverFactory=t,this._observedElements=new Map}return t.prototype.ngOnDestroy=function(){var t=this;this._observedElements.forEach(function(e,n){return t._cleanupObserver(n)})},t.prototype.observe=function(t){var e=this,n=Object(r.coerceElement)(t);return o.Observable.create(function(t){var r=e._observeElement(n).subscribe(t);return function(){r.unsubscribe(),e._unobserveElement(n)}})},t.prototype._observeElement=function(t){if(this._observedElements.has(t))this._observedElements.get(t).count++;else{var e=new o.Subject,n=this._mutationObserverFactory.create(function(t){return e.next(t)});n&&n.observe(t,{characterData:!0,childList:!0,subtree:!0}),this._observedElements.set(t,{observer:n,stream:e,count:1})}return this._observedElements.get(t).stream},t.prototype._unobserveElement=function(t){this._observedElements.has(t)&&(this._observedElements.get(t).count--,this._observedElements.get(t).count||this._cleanupObserver(t))},t.prototype._cleanupObserver=function(t){if(this._observedElements.has(t)){var e=this._observedElements.get(t),n=e.observer,r=e.stream;n&&n.disconnect(),r.complete(),this._observedElements.delete(t)}},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:s}]},t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t(Object(i.inject)(s))},token:t,providedIn:"root"}),t}(),c=function(){function t(t,e,n){this._contentObserver=t,this._elementRef=e,this._ngZone=n,this.event=new i.EventEmitter,this._disabled=!1,this._currentSubscription=null}return Object.defineProperty(t.prototype,"disabled",{get:function(){return this._disabled},set:function(t){this._disabled=Object(r.coerceBooleanProperty)(t),this._disabled?this._unsubscribe():this._subscribe()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"debounce",{get:function(){return this._debounce},set:function(t){this._debounce=Object(r.coerceNumberProperty)(t),this._subscribe()},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this._currentSubscription||this.disabled||this._subscribe()},t.prototype.ngOnDestroy=function(){this._unsubscribe()},t.prototype._subscribe=function(){var t=this;this._unsubscribe();var e=this._contentObserver.observe(this._elementRef);this._ngZone.runOutsideAngular(function(){t._currentSubscription=(t.debounce?e.pipe(Object(a.debounceTime)(t.debounce)):e).subscribe(t.event)})},t.prototype._unsubscribe=function(){this._currentSubscription&&this._currentSubscription.unsubscribe()},t.decorators=[{type:i.Directive,args:[{selector:"[cdkObserveContent]",exportAs:"cdkObserveContent"}]}],t.ctorParameters=function(){return[{type:u},{type:i.ElementRef},{type:i.NgZone}]},t.propDecorators={event:[{type:i.Output,args:["cdkObserveContent"]}],disabled:[{type:i.Input,args:["cdkObserveContentDisabled"]}],debounce:[{type:i.Input}]},t}(),l=function(){function t(){}return t.decorators=[{type:i.NgModule,args:[{exports:[c],declarations:[c],providers:[s]}]}],t}()},"./node_modules/@angular/cdk/esm5/overlay.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"Overlay",function(){return U}),n.d(e,"OverlayContainer",function(){return A}),n.d(e,"CdkOverlayOrigin",function(){return G}),n.d(e,"CdkConnectedOverlay",function(){return Y}),n.d(e,"FullscreenOverlayContainer",function(){return $}),n.d(e,"OverlayRef",function(){return P}),n.d(e,"OverlayKeyboardDispatcher",function(){return M}),n.d(e,"OverlayPositionBuilder",function(){return z}),n.d(e,"GlobalPositionStrategy",function(){return V}),n.d(e,"ConnectedPositionStrategy",function(){return L}),n.d(e,"FlexibleConnectedPositionStrategy",function(){return N}),n.d(e,"OverlayConfig",function(){return w}),n.d(e,"validateVerticalPosition",function(){return E}),n.d(e,"validateHorizontalPosition",function(){return S}),n.d(e,"ConnectionPositionPair",function(){return j}),n.d(e,"ScrollingVisibility",function(){return O}),n.d(e,"ConnectedOverlayPositionChange",function(){return C}),n.d(e,"ScrollStrategyOptions",function(){return x}),n.d(e,"RepositionScrollStrategy",function(){return b}),n.d(e,"CloseScrollStrategy",function(){return g}),n.d(e,"NoopScrollStrategy",function(){return y}),n.d(e,"BlockScrollStrategy",function(){return h}),n.d(e,"OverlayModule",function(){return Q}),n.d(e,"OVERLAY_PROVIDERS",function(){return Z}),n.d(e,"ɵg",function(){return k}),n.d(e,"ɵf",function(){return T}),n.d(e,"ɵb",function(){return I}),n.d(e,"ɵa",function(){return D}),n.d(e,"ɵc",function(){return q}),n.d(e,"ɵe",function(){return X}),n.d(e,"ɵd",function(){return K});var r=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),i=n("./node_modules/@angular/cdk/esm5/scrolling.es5.js");n.d(e,"ViewportRuler",function(){return i.ViewportRuler}),n.d(e,"VIEWPORT_RULER_PROVIDER",function(){return i.VIEWPORT_RULER_PROVIDER}),n.d(e,"CdkScrollable",function(){return i.CdkScrollable}),n.d(e,"ScrollDispatcher",function(){return i.ScrollDispatcher});var o=n("./node_modules/@angular/common/fesm5/common.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),s=n("./node_modules/tslib/tslib.es6.js"),u=n("./node_modules/rxjs/_esm5/index.js"),c=n("./node_modules/rxjs/_esm5/operators/index.js"),l=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),d=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),p=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),f=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),h=function(){function t(t,e){this._viewportRuler=t,this._previousHTMLStyles={top:"",left:""},this._isEnabled=!1,this._document=e}return t.prototype.attach=function(){},t.prototype.enable=function(){if(this._canBeEnabled()){var t=this._document.documentElement;this._previousScrollPosition=this._viewportRuler.getViewportScrollPosition(),this._previousHTMLStyles.left=t.style.left||"",this._previousHTMLStyles.top=t.style.top||"",t.style.left=Object(r.coerceCssPixelValue)(-this._previousScrollPosition.left),t.style.top=Object(r.coerceCssPixelValue)(-this._previousScrollPosition.top),t.classList.add("cdk-global-scrollblock"),this._isEnabled=!0}},t.prototype.disable=function(){if(this._isEnabled){var t=this._document.documentElement,e=this._document.body,n=t.style,r=e.style,i=n.scrollBehavior||"",o=r.scrollBehavior||"";this._isEnabled=!1,n.left=this._previousHTMLStyles.left,n.top=this._previousHTMLStyles.top,t.classList.remove("cdk-global-scrollblock"),n.scrollBehavior=r.scrollBehavior="auto",window.scroll(this._previousScrollPosition.left,this._previousScrollPosition.top),n.scrollBehavior=i,r.scrollBehavior=o}},t.prototype._canBeEnabled=function(){if(this._document.documentElement.classList.contains("cdk-global-scrollblock")||this._isEnabled)return!1;var t=this._document.body,e=this._viewportRuler.getViewportSize();return t.scrollHeight>e.height||t.scrollWidth>e.width},t}();function m(){return Error("Scroll strategy has already been attached.")}var g=function(){function t(t,e,n,r){var i=this;this._scrollDispatcher=t,this._ngZone=e,this._viewportRuler=n,this._config=r,this._scrollSubscription=null,this._detach=function(){i.disable(),i._overlayRef.hasAttached()&&i._ngZone.run(function(){return i._overlayRef.detach()})}}return t.prototype.attach=function(t){if(this._overlayRef)throw m();this._overlayRef=t},t.prototype.enable=function(){var t=this;if(!this._scrollSubscription){var e=this._scrollDispatcher.scrolled(0);this._config&&this._config.threshold&&this._config.threshold>1?(this._initialScrollPosition=this._viewportRuler.getViewportScrollPosition().top,this._scrollSubscription=e.subscribe(function(){var e=t._viewportRuler.getViewportScrollPosition().top;Math.abs(e-t._initialScrollPosition)>t._config.threshold?t._detach():t._overlayRef.updatePosition()})):this._scrollSubscription=e.subscribe(this._detach)}},t.prototype.disable=function(){this._scrollSubscription&&(this._scrollSubscription.unsubscribe(),this._scrollSubscription=null)},t}(),y=function(){function t(){}return t.prototype.enable=function(){},t.prototype.disable=function(){},t.prototype.attach=function(){},t}();function _(t,e){return e.some(function(e){var n=t.bottom<e.top,r=t.top>e.bottom,i=t.right<e.left,o=t.left>e.right;return n||r||i||o})}function v(t,e){return e.some(function(e){var n=t.top<e.top,r=t.bottom>e.bottom,i=t.left<e.left,o=t.right>e.right;return n||r||i||o})}var b=function(){function t(t,e,n,r){this._scrollDispatcher=t,this._viewportRuler=e,this._ngZone=n,this._config=r,this._scrollSubscription=null}return t.prototype.attach=function(t){if(this._overlayRef)throw m();this._overlayRef=t},t.prototype.enable=function(){var t=this;if(!this._scrollSubscription){var e=this._config?this._config.scrollThrottle:0;this._scrollSubscription=this._scrollDispatcher.scrolled(e).subscribe(function(){if(t._overlayRef.updatePosition(),t._config&&t._config.autoClose){var e=t._overlayRef.overlayElement.getBoundingClientRect(),n=t._viewportRuler.getViewportSize(),r=n.width,i=n.height;_(e,[{width:r,height:i,bottom:i,right:r,top:0,left:0}])&&(t.disable(),t._ngZone.run(function(){return t._overlayRef.detach()}))}})}},t.prototype.disable=function(){this._scrollSubscription&&(this._scrollSubscription.unsubscribe(),this._scrollSubscription=null)},t}(),x=function(){function t(t,e,n,r){var i=this;this._scrollDispatcher=t,this._viewportRuler=e,this._ngZone=n,this.noop=function(){return new y},this.close=function(t){return new g(i._scrollDispatcher,i._ngZone,i._viewportRuler,t)},this.block=function(){return new h(i._viewportRuler,i._document)},this.reposition=function(t){return new b(i._scrollDispatcher,i._viewportRuler,i._ngZone,t)},this._document=r}return t.decorators=[{type:a.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:i.ScrollDispatcher},{type:i.ViewportRuler},{type:a.NgZone},{type:void 0,decorators:[{type:a.Inject,args:[o.DOCUMENT]}]}]},t.ngInjectableDef=Object(a.defineInjectable)({factory:function(){return new t(Object(a.inject)(i.ScrollDispatcher),Object(a.inject)(i.ViewportRuler),Object(a.inject)(a.NgZone),Object(a.inject)(o.DOCUMENT))},token:t,providedIn:"root"}),t}(),w=function(){return function(t){var e=this;this.scrollStrategy=new y,this.panelClass="",this.hasBackdrop=!1,this.backdropClass="cdk-overlay-dark-backdrop",this.disposeOnNavigation=!1,t&&Object.keys(t).forEach(function(n){var r=n;void 0!==t[r]&&(e[r]=t[r])})}}(),j=function(){return function(t,e,n,r,i){this.offsetX=n,this.offsetY=r,this.panelClass=i,this.originX=t.originX,this.originY=t.originY,this.overlayX=e.overlayX,this.overlayY=e.overlayY}}(),O=function(){return function(){}}(),C=function(){function t(t,e){this.connectionPair=t,this.scrollableViewProperties=e}return t.ctorParameters=function(){return[{type:j},{type:O,decorators:[{type:a.Optional}]}]},t}();function E(t,e){if("top"!==e&&"bottom"!==e&&"center"!==e)throw Error("ConnectedPosition: Invalid "+t+' "'+e+'". Expected "top", "bottom" or "center".')}function S(t,e){if("start"!==e&&"end"!==e&&"center"!==e)throw Error("ConnectedPosition: Invalid "+t+' "'+e+'". Expected "start", "end" or "center".')}var M=function(){function t(t){var e=this;this._attachedOverlays=[],this._keydownListener=function(t){for(var n=e._attachedOverlays,r=n.length-1;r>-1;r--)if(n[r]._keydownEventSubscriptions>0){n[r]._keydownEvents.next(t);break}},this._document=t}return t.prototype.ngOnDestroy=function(){this._detach()},t.prototype.add=function(t){this.remove(t),this._isAttached||(this._document.body.addEventListener("keydown",this._keydownListener,!0),this._isAttached=!0),this._attachedOverlays.push(t)},t.prototype.remove=function(t){var e=this._attachedOverlays.indexOf(t);e>-1&&this._attachedOverlays.splice(e,1),0===this._attachedOverlays.length&&this._detach()},t.prototype._detach=function(){this._isAttached&&(this._document.body.removeEventListener("keydown",this._keydownListener,!0),this._isAttached=!1)},t.decorators=[{type:a.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:a.Inject,args:[o.DOCUMENT]}]}]},t.ngInjectableDef=Object(a.defineInjectable)({factory:function(){return new t(Object(a.inject)(o.DOCUMENT))},token:t,providedIn:"root"}),t}();function T(t,e){return t||new M(e)}var k={provide:M,deps:[[new a.Optional,new a.SkipSelf,M],o.DOCUMENT],useFactory:T},A=function(){function t(t){this._document=t}return t.prototype.ngOnDestroy=function(){this._containerElement&&this._containerElement.parentNode&&this._containerElement.parentNode.removeChild(this._containerElement)},t.prototype.getContainerElement=function(){return this._containerElement||this._createContainer(),this._containerElement},t.prototype._createContainer=function(){var t=this._document.createElement("div");t.classList.add("cdk-overlay-container"),this._document.body.appendChild(t),this._containerElement=t},t.decorators=[{type:a.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:a.Inject,args:[o.DOCUMENT]}]}]},t.ngInjectableDef=Object(a.defineInjectable)({factory:function(){return new t(Object(a.inject)(o.DOCUMENT))},token:t,providedIn:"root"}),t}();function D(t,e){return t||new A(e)}var I={provide:A,deps:[[new a.Optional,new a.SkipSelf,A],o.DOCUMENT],useFactory:D},P=function(){function t(t,e,n,r,i,o,a,s){var c=this;this._portalOutlet=t,this._host=e,this._pane=n,this._config=r,this._ngZone=i,this._keyboardDispatcher=o,this._document=a,this._location=s,this._backdropElement=null,this._backdropClick=new u.Subject,this._attachments=new u.Subject,this._detachments=new u.Subject,this._locationChanges=u.Subscription.EMPTY,this._keydownEventsObservable=u.Observable.create(function(t){var e=c._keydownEvents.subscribe(t);return c._keydownEventSubscriptions++,function(){e.unsubscribe(),c._keydownEventSubscriptions--}}),this._keydownEvents=new u.Subject,this._keydownEventSubscriptions=0,r.scrollStrategy&&r.scrollStrategy.attach(this),this._positionStrategy=r.positionStrategy}return Object.defineProperty(t.prototype,"overlayElement",{get:function(){return this._pane},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"backdropElement",{get:function(){return this._backdropElement},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"hostElement",{get:function(){return this._host},enumerable:!0,configurable:!0}),t.prototype.attach=function(t){var e=this,n=this._portalOutlet.attach(t);return this._positionStrategy&&this._positionStrategy.attach(this),!this._host.parentElement&&this._previousHostParent&&this._previousHostParent.appendChild(this._host),this._updateStackingOrder(),this._updateElementSize(),this._updateElementDirection(),this._config.scrollStrategy&&this._config.scrollStrategy.enable(),this._ngZone.onStable.asObservable().pipe(Object(c.take)(1)).subscribe(function(){e.hasAttached()&&e.updatePosition()}),this._togglePointerEvents(!0),this._config.hasBackdrop&&this._attachBackdrop(),this._config.panelClass&&this._toggleClasses(this._pane,this._config.panelClass,!0),this._attachments.next(),this._keyboardDispatcher.add(this),this._config.disposeOnNavigation&&this._location&&(this._locationChanges=this._location.subscribe(function(){return e.dispose()})),n},t.prototype.detach=function(){if(this.hasAttached()){this.detachBackdrop(),this._togglePointerEvents(!1),this._positionStrategy&&this._positionStrategy.detach&&this._positionStrategy.detach(),this._config.scrollStrategy&&this._config.scrollStrategy.disable();var t=this._portalOutlet.detach();return this._detachments.next(),this._keyboardDispatcher.remove(this),this._detachContentWhenStable(),this._locationChanges.unsubscribe(),t}},t.prototype.dispose=function(){var t=this.hasAttached();this._positionStrategy&&this._positionStrategy.dispose(),this._config.scrollStrategy&&this._config.scrollStrategy.disable(),this.detachBackdrop(),this._locationChanges.unsubscribe(),this._keyboardDispatcher.remove(this),this._portalOutlet.dispose(),this._attachments.complete(),this._backdropClick.complete(),this._keydownEvents.complete(),this._host&&this._host.parentNode&&(this._host.parentNode.removeChild(this._host),this._host=null),this._previousHostParent=this._pane=null,t&&this._detachments.next(),this._detachments.complete()},t.prototype.hasAttached=function(){return this._portalOutlet.hasAttached()},t.prototype.backdropClick=function(){return this._backdropClick.asObservable()},t.prototype.attachments=function(){return this._attachments.asObservable()},t.prototype.detachments=function(){return this._detachments.asObservable()},t.prototype.keydownEvents=function(){return this._keydownEventsObservable},t.prototype.getConfig=function(){return this._config},t.prototype.updatePosition=function(){this._positionStrategy&&this._positionStrategy.apply()},t.prototype.updatePositionStrategy=function(t){t!==this._positionStrategy&&(this._positionStrategy&&this._positionStrategy.dispose(),this._positionStrategy=t,this.hasAttached()&&(t.attach(this),this.updatePosition()))},t.prototype.updateSize=function(t){this._config=Object(s.__assign)({},this._config,t),this._updateElementSize()},t.prototype.setDirection=function(t){this._config=Object(s.__assign)({},this._config,{direction:t}),this._updateElementDirection()},t.prototype.getDirection=function(){var t=this._config.direction;return t?"string"==typeof t?t:t.value:"ltr"},t.prototype._updateElementDirection=function(){this._host.setAttribute("dir",this.getDirection())},t.prototype._updateElementSize=function(){var t=this._pane.style;t.width=Object(r.coerceCssPixelValue)(this._config.width),t.height=Object(r.coerceCssPixelValue)(this._config.height),t.minWidth=Object(r.coerceCssPixelValue)(this._config.minWidth),t.minHeight=Object(r.coerceCssPixelValue)(this._config.minHeight),t.maxWidth=Object(r.coerceCssPixelValue)(this._config.maxWidth),t.maxHeight=Object(r.coerceCssPixelValue)(this._config.maxHeight)},t.prototype._togglePointerEvents=function(t){this._pane.style.pointerEvents=t?"auto":"none"},t.prototype._attachBackdrop=function(){var t=this,e="cdk-overlay-backdrop-showing";this._backdropElement=this._document.createElement("div"),this._backdropElement.classList.add("cdk-overlay-backdrop"),this._config.backdropClass&&this._toggleClasses(this._backdropElement,this._config.backdropClass,!0),this._host.parentElement.insertBefore(this._backdropElement,this._host),this._backdropElement.addEventListener("click",function(e){return t._backdropClick.next(e)}),"undefined"!=typeof requestAnimationFrame?this._ngZone.runOutsideAngular(function(){requestAnimationFrame(function(){t._backdropElement&&t._backdropElement.classList.add(e)})}):this._backdropElement.classList.add(e)},t.prototype._updateStackingOrder=function(){this._host.nextSibling&&this._host.parentNode.appendChild(this._host)},t.prototype.detachBackdrop=function(){var t=this,e=this._backdropElement;if(e){var n,r=function(){e&&e.parentNode&&e.parentNode.removeChild(e),t._backdropElement==e&&(t._backdropElement=null),t._config.backdropClass&&t._toggleClasses(e,t._config.backdropClass,!1),clearTimeout(n)};e.classList.remove("cdk-overlay-backdrop-showing"),this._ngZone.runOutsideAngular(function(){e.addEventListener("transitionend",r)}),e.style.pointerEvents="none",n=this._ngZone.runOutsideAngular(function(){return setTimeout(r,500)})}},t.prototype._toggleClasses=function(t,e,n){var i=t.classList;Object(r.coerceArray)(e).forEach(function(t){n?i.add(t):i.remove(t)})},t.prototype._detachContentWhenStable=function(){var t=this;this._ngZone.runOutsideAngular(function(){var e=t._ngZone.onStable.asObservable().pipe(Object(c.takeUntil)(Object(u.merge)(t._attachments,t._detachments))).subscribe(function(){t._pane&&t._host&&0!==t._pane.children.length||(t._pane&&t._config.panelClass&&t._toggleClasses(t._pane,t._config.panelClass,!1),t._host&&t._host.parentElement&&(t._previousHostParent=t._host.parentElement,t._previousHostParent.removeChild(t._host)),e.unsubscribe())})})},t}(),R="cdk-overlay-connected-position-bounding-box",N=function(){function t(t,e,n,r,i){var o=this;this._viewportRuler=e,this._document=n,this._platform=r,this._overlayContainer=i,this._lastBoundingBoxSize={width:0,height:0},this._isPushed=!1,this._canPush=!0,this._growAfterOpen=!1,this._hasFlexibleDimensions=!0,this._positionLocked=!1,this._viewportMargin=0,this.scrollables=[],this._preferredPositions=[],this._positionChanges=new u.Subject,this._resizeSubscription=u.Subscription.EMPTY,this._offsetX=0,this._offsetY=0,this._positionChangeSubscriptions=0,this._appliedPanelClasses=[],this.positionChanges=u.Observable.create(function(t){var e=o._positionChanges.subscribe(t);return o._positionChangeSubscriptions++,function(){e.unsubscribe(),o._positionChangeSubscriptions--}}),this.setOrigin(t)}return Object.defineProperty(t.prototype,"positions",{get:function(){return this._preferredPositions},enumerable:!0,configurable:!0}),t.prototype.attach=function(t){var e=this;if(this._overlayRef&&t!==this._overlayRef)throw Error("This position strategy is already attached to an overlay");this._validatePositions(),t.hostElement.classList.add(R),this._overlayRef=t,this._boundingBox=t.hostElement,this._pane=t.overlayElement,this._isDisposed=!1,this._isInitialRender=!0,this._lastPosition=null,this._resizeSubscription.unsubscribe(),this._resizeSubscription=this._viewportRuler.change().subscribe(function(){e._isInitialRender=!0,e.apply()})},t.prototype.apply=function(){if(!(this._isDisposed||this._platform&&!this._platform.isBrowser))if(!this._isInitialRender&&this._positionLocked&&this._lastPosition)this.reapplyLastPosition();else{this._clearPanelClasses(),this._resetOverlayElementStyles(),this._resetBoundingBoxStyles(),this._viewportRect=this._getNarrowedViewportRect(),this._originRect=this._origin.getBoundingClientRect(),this._overlayRect=this._pane.getBoundingClientRect();for(var t,e=this._originRect,n=this._overlayRect,r=this._viewportRect,i=[],o=0,a=this._preferredPositions;o<a.length;o++){var s=a[o],u=this._getOriginPoint(e,s),c=this._getOverlayPoint(u,n,s),l=this._getOverlayFit(c,n,r,s);if(l.isCompletelyWithinViewport)return this._isPushed=!1,void this._applyPosition(s,u);this._canFitWithFlexibleDimensions(l,c,r)?i.push({position:s,origin:u,overlayRect:n,boundingBoxRect:this._calculateBoundingBoxRect(u,s)}):(!t||t.overlayFit.visibleArea<l.visibleArea)&&(t={overlayFit:l,overlayPoint:c,originPoint:u,position:s,overlayRect:n})}if(i.length){for(var d=null,p=-1,f=0,h=i;f<h.length;f++){var m=h[f],g=m.boundingBoxRect.width*m.boundingBoxRect.height*(m.position.weight||1);g>p&&(p=g,d=m)}return this._isPushed=!1,void this._applyPosition(d.position,d.origin)}if(this._canPush)return this._isPushed=!0,void this._applyPosition(t.position,t.originPoint);this._applyPosition(t.position,t.originPoint)}},t.prototype.detach=function(){this._clearPanelClasses(),this._lastPosition=null,this._previousPushAmount=null,this._resizeSubscription.unsubscribe()},t.prototype.dispose=function(){this._isDisposed||(this._boundingBox&&F(this._boundingBox.style,{top:"",left:"",right:"",bottom:"",height:"",width:"",alignItems:"",justifyContent:""}),this._pane&&this._resetOverlayElementStyles(),this._overlayRef&&this._overlayRef.hostElement.classList.remove(R),this.detach(),this._positionChanges.complete(),this._overlayRef=this._boundingBox=null,this._isDisposed=!0)},t.prototype.reapplyLastPosition=function(){if(!this._isDisposed&&(!this._platform||this._platform.isBrowser)){this._originRect=this._origin.getBoundingClientRect(),this._overlayRect=this._pane.getBoundingClientRect(),this._viewportRect=this._getNarrowedViewportRect();var t=this._lastPosition||this._preferredPositions[0],e=this._getOriginPoint(this._originRect,t);this._applyPosition(t,e)}},t.prototype.withScrollableContainers=function(t){this.scrollables=t},t.prototype.withPositions=function(t){return this._preferredPositions=t,-1===t.indexOf(this._lastPosition)&&(this._lastPosition=null),this._validatePositions(),this},t.prototype.withViewportMargin=function(t){return this._viewportMargin=t,this},t.prototype.withFlexibleDimensions=function(t){return void 0===t&&(t=!0),this._hasFlexibleDimensions=t,this},t.prototype.withGrowAfterOpen=function(t){return void 0===t&&(t=!0),this._growAfterOpen=t,this},t.prototype.withPush=function(t){return void 0===t&&(t=!0),this._canPush=t,this},t.prototype.withLockedPosition=function(t){return void 0===t&&(t=!0),this._positionLocked=t,this},t.prototype.setOrigin=function(t){return this._origin=Object(r.coerceElement)(t),this},t.prototype.withDefaultOffsetX=function(t){return this._offsetX=t,this},t.prototype.withDefaultOffsetY=function(t){return this._offsetY=t,this},t.prototype.withTransformOriginOn=function(t){return this._transformOriginSelector=t,this},t.prototype._getOriginPoint=function(t,e){var n;if("center"==e.originX)n=t.left+t.width/2;else{var r=this._isRtl()?t.right:t.left,i=this._isRtl()?t.left:t.right;n="start"==e.originX?r:i}return{x:n,y:"center"==e.originY?t.top+t.height/2:"top"==e.originY?t.top:t.bottom}},t.prototype._getOverlayPoint=function(t,e,n){var r,i;return r="center"==n.overlayX?-e.width/2:"start"===n.overlayX?this._isRtl()?-e.width:0:this._isRtl()?0:-e.width,i="center"==n.overlayY?-e.height/2:"top"==n.overlayY?0:-e.height,{x:t.x+r,y:t.y+i}},t.prototype._getOverlayFit=function(t,e,n,r){var i=t.x,o=t.y,a=this._getOffset(r,"x"),s=this._getOffset(r,"y");a&&(i+=a),s&&(o+=s);var u=0-i,c=i+e.width-n.width,l=0-o,d=o+e.height-n.height,p=this._subtractOverflows(e.width,u,c),f=this._subtractOverflows(e.height,l,d),h=p*f;return{visibleArea:h,isCompletelyWithinViewport:e.width*e.height===h,fitsInViewportVertically:f===e.height,fitsInViewportHorizontally:p==e.width}},t.prototype._canFitWithFlexibleDimensions=function(t,e,n){if(this._hasFlexibleDimensions){var r=n.bottom-e.y,i=n.right-e.x,o=this._overlayRef.getConfig().minHeight,a=this._overlayRef.getConfig().minWidth,s=t.fitsInViewportVertically||null!=o&&o<=r,u=t.fitsInViewportHorizontally||null!=a&&a<=i;return s&&u}},t.prototype._pushOverlayOnScreen=function(t,e,n){if(this._previousPushAmount&&this._positionLocked)return{x:t.x+this._previousPushAmount.x,y:t.y+this._previousPushAmount.y};var r=this._viewportRect,i=Math.max(t.x+e.width-r.right,0),o=Math.max(t.y+e.height-r.bottom,0),a=Math.max(r.top-n.top-t.y,0),s=Math.max(r.left-n.left-t.x,0),u=0,c=0;return u=e.width<r.width?s||-i:t.x<this._viewportMargin?r.left-n.left-t.x:0,c=e.height<r.height?a||-o:t.y<this._viewportMargin?r.top-n.top-t.y:0,this._previousPushAmount={x:u,y:c},{x:t.x+u,y:t.y+c}},t.prototype._applyPosition=function(t,e){if(this._setTransformOrigin(t),this._setOverlayElementStyles(e,t),this._setBoundingBoxStyles(e,t),t.panelClass&&this._addPanelClasses(t.panelClass),this._lastPosition=t,this._positionChangeSubscriptions>0){var n=this._getScrollVisibility(),r=new C(t,n);this._positionChanges.next(r)}this._isInitialRender=!1},t.prototype._setTransformOrigin=function(t){if(this._transformOriginSelector){var e,n=this._boundingBox.querySelectorAll(this._transformOriginSelector),r=t.overlayY;e="center"===t.overlayX?"center":this._isRtl()?"start"===t.overlayX?"right":"left":"start"===t.overlayX?"left":"right";for(var i=0;i<n.length;i++)n[i].style.transformOrigin=e+" "+r}},t.prototype._calculateBoundingBoxRect=function(t,e){var n,r,i,o=this._viewportRect,a=this._isRtl();if("top"===e.overlayY)r=t.y,n=o.bottom-t.y;else if("bottom"===e.overlayY)i=o.height-t.y+2*this._viewportMargin,n=o.height-i+this._viewportMargin;else{var s=Math.min(o.bottom-t.y+o.top,t.y),u=this._lastBoundingBoxSize.height;n=2*s,r=t.y-s,n>u&&!this._isInitialRender&&!this._growAfterOpen&&(r=t.y-u/2)}var c,l,d,p="start"===e.overlayX&&!a||"end"===e.overlayX&&a;if("end"===e.overlayX&&!a||"start"===e.overlayX&&a)d=o.right-t.x+this._viewportMargin,c=t.x-o.left;else if(p)l=t.x,c=o.right-t.x;else{s=Math.min(o.right-t.x+o.left,t.x);var f=this._lastBoundingBoxSize.width;c=2*s,l=t.x-s,c>f&&!this._isInitialRender&&!this._growAfterOpen&&(l=t.x-f/2)}return{top:r,left:l,bottom:i,right:d,width:c,height:n}},t.prototype._setBoundingBoxStyles=function(t,e){var n=this._calculateBoundingBoxRect(t,e);this._isInitialRender||this._growAfterOpen||(n.height=Math.min(n.height,this._lastBoundingBoxSize.height),n.width=Math.min(n.width,this._lastBoundingBoxSize.width));var i={};if(this._hasExactPosition())i.top=i.left="0",i.bottom=i.right="",i.width=i.height="100%";else{var o=this._overlayRef.getConfig().maxHeight,a=this._overlayRef.getConfig().maxWidth;i.height=Object(r.coerceCssPixelValue)(n.height),i.top=Object(r.coerceCssPixelValue)(n.top),i.bottom=Object(r.coerceCssPixelValue)(n.bottom),i.width=Object(r.coerceCssPixelValue)(n.width),i.left=Object(r.coerceCssPixelValue)(n.left),i.right=Object(r.coerceCssPixelValue)(n.right),"center"===e.overlayX?i.alignItems="center":i.alignItems="end"===e.overlayX?"flex-end":"flex-start","center"===e.overlayY?i.justifyContent="center":i.justifyContent="bottom"===e.overlayY?"flex-end":"flex-start",o&&(i.maxHeight=Object(r.coerceCssPixelValue)(o)),a&&(i.maxWidth=Object(r.coerceCssPixelValue)(a))}this._lastBoundingBoxSize=n,F(this._boundingBox.style,i)},t.prototype._resetBoundingBoxStyles=function(){F(this._boundingBox.style,{top:"0",left:"0",right:"0",bottom:"0",height:"",width:"",alignItems:"",justifyContent:""})},t.prototype._resetOverlayElementStyles=function(){F(this._pane.style,{top:"",left:"",bottom:"",right:"",position:""})},t.prototype._setOverlayElementStyles=function(t,e){var n={};if(this._hasExactPosition()){var r=this._viewportRuler.getViewportScrollPosition();F(n,this._getExactOverlayY(e,t,r)),F(n,this._getExactOverlayX(e,t,r))}else n.position="static";var i="",o=this._getOffset(e,"x"),a=this._getOffset(e,"y");o&&(i+="translateX("+o+"px) "),a&&(i+="translateY("+a+"px)"),n.transform=i.trim(),this._hasFlexibleDimensions&&this._overlayRef.getConfig().maxHeight&&(n.maxHeight=""),this._hasFlexibleDimensions&&this._overlayRef.getConfig().maxWidth&&(n.maxWidth=""),F(this._pane.style,n)},t.prototype._getExactOverlayY=function(t,e,n){var i={top:null,bottom:null},o=this._getOverlayPoint(e,this._overlayRect,t);this._isPushed&&(o=this._pushOverlayOnScreen(o,this._overlayRect,n));var a=this._overlayContainer?this._overlayContainer.getContainerElement().getBoundingClientRect().top:0;if(o.y-=a,"bottom"===t.overlayY){var s=this._document.documentElement.clientHeight;i.bottom=s-(o.y+this._overlayRect.height)+"px"}else i.top=Object(r.coerceCssPixelValue)(o.y);return i},t.prototype._getExactOverlayX=function(t,e,n){var i={left:null,right:null},o=this._getOverlayPoint(e,this._overlayRect,t);if(this._isPushed&&(o=this._pushOverlayOnScreen(o,this._overlayRect,n)),"right"===(this._isRtl()?"end"===t.overlayX?"left":"right":"end"===t.overlayX?"right":"left")){var a=this._document.documentElement.clientWidth;i.right=a-(o.x+this._overlayRect.width)+"px"}else i.left=Object(r.coerceCssPixelValue)(o.x);return i},t.prototype._getScrollVisibility=function(){var t=this._origin.getBoundingClientRect(),e=this._pane.getBoundingClientRect(),n=this.scrollables.map(function(t){return t.getElementRef().nativeElement.getBoundingClientRect()});return{isOriginClipped:v(t,n),isOriginOutsideView:_(t,n),isOverlayClipped:v(e,n),isOverlayOutsideView:_(e,n)}},t.prototype._subtractOverflows=function(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];return e.reduce(function(t,e){return t-Math.max(e,0)},t)},t.prototype._getNarrowedViewportRect=function(){var t=this._document.documentElement.clientWidth,e=this._document.documentElement.clientHeight,n=this._viewportRuler.getViewportScrollPosition();return{top:n.top+this._viewportMargin,left:n.left+this._viewportMargin,right:n.left+t-this._viewportMargin,bottom:n.top+e-this._viewportMargin,width:t-2*this._viewportMargin,height:e-2*this._viewportMargin}},t.prototype._isRtl=function(){return"rtl"===this._overlayRef.getDirection()},t.prototype._hasExactPosition=function(){return!this._hasFlexibleDimensions||this._isPushed},t.prototype._getOffset=function(t,e){return"x"===e?null==t.offsetX?this._offsetX:t.offsetX:null==t.offsetY?this._offsetY:t.offsetY},t.prototype._validatePositions=function(){if(!this._preferredPositions.length)throw Error("FlexibleConnectedPositionStrategy: At least one position is required.");this._preferredPositions.forEach(function(t){S("originX",t.originX),E("originY",t.originY),S("overlayX",t.overlayX),E("overlayY",t.overlayY)})},t.prototype._addPanelClasses=function(t){var e=this;this._pane&&Object(r.coerceArray)(t).forEach(function(t){-1===e._appliedPanelClasses.indexOf(t)&&(e._appliedPanelClasses.push(t),e._pane.classList.add(t))})},t.prototype._clearPanelClasses=function(){var t=this;this._pane&&(this._appliedPanelClasses.forEach(function(e){return t._pane.classList.remove(e)}),this._appliedPanelClasses=[])},t}();function F(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);return t}var L=function(){function t(t,e,n,r,i,o){this._preferredPositions=[],this._positionStrategy=new N(n,r,i,o).withFlexibleDimensions(!1).withPush(!1).withViewportMargin(0),this.withFallbackPosition(t,e)}return Object.defineProperty(t.prototype,"_isRtl",{get:function(){return"rtl"===this._overlayRef.getDirection()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"onPositionChange",{get:function(){return this._positionStrategy.positionChanges},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"positions",{get:function(){return this._preferredPositions},enumerable:!0,configurable:!0}),t.prototype.attach=function(t){this._overlayRef=t,this._positionStrategy.attach(t),this._direction&&(t.setDirection(this._direction),this._direction=null)},t.prototype.dispose=function(){this._positionStrategy.dispose()},t.prototype.detach=function(){this._positionStrategy.detach()},t.prototype.apply=function(){this._positionStrategy.apply()},t.prototype.recalculateLastPosition=function(){this._positionStrategy.reapplyLastPosition()},t.prototype.withScrollableContainers=function(t){this._positionStrategy.withScrollableContainers(t)},t.prototype.withFallbackPosition=function(t,e,n,r){var i=new j(t,e,n,r);return this._preferredPositions.push(i),this._positionStrategy.withPositions(this._preferredPositions),this},t.prototype.withDirection=function(t){return this._overlayRef?this._overlayRef.setDirection(t):this._direction=t,this},t.prototype.withOffsetX=function(t){return this._positionStrategy.withDefaultOffsetX(t),this},t.prototype.withOffsetY=function(t){return this._positionStrategy.withDefaultOffsetY(t),this},t.prototype.withLockedPosition=function(t){return this._positionStrategy.withLockedPosition(t),this},t.prototype.withPositions=function(t){return this._preferredPositions=t.slice(),this._positionStrategy.withPositions(this._preferredPositions),this},t.prototype.setOrigin=function(t){return this._positionStrategy.setOrigin(t),this},t}(),B="cdk-global-overlay-wrapper",V=function(){function t(){this._cssPosition="static",this._topOffset="",this._bottomOffset="",this._leftOffset="",this._rightOffset="",this._alignItems="",this._justifyContent="",this._width="",this._height=""}return t.prototype.attach=function(t){var e=t.getConfig();this._overlayRef=t,this._width&&!e.width&&t.updateSize({width:this._width}),this._height&&!e.height&&t.updateSize({height:this._height}),t.hostElement.classList.add(B),this._isDisposed=!1},t.prototype.top=function(t){return void 0===t&&(t=""),this._bottomOffset="",this._topOffset=t,this._alignItems="flex-start",this},t.prototype.left=function(t){return void 0===t&&(t=""),this._rightOffset="",this._leftOffset=t,this._justifyContent="flex-start",this},t.prototype.bottom=function(t){return void 0===t&&(t=""),this._topOffset="",this._bottomOffset=t,this._alignItems="flex-end",this},t.prototype.right=function(t){return void 0===t&&(t=""),this._leftOffset="",this._rightOffset=t,this._justifyContent="flex-end",this},t.prototype.width=function(t){return void 0===t&&(t=""),this._overlayRef?this._overlayRef.updateSize({width:t}):this._width=t,this},t.prototype.height=function(t){return void 0===t&&(t=""),this._overlayRef?this._overlayRef.updateSize({height:t}):this._height=t,this},t.prototype.centerHorizontally=function(t){return void 0===t&&(t=""),this.left(t),this._justifyContent="center",this},t.prototype.centerVertically=function(t){return void 0===t&&(t=""),this.top(t),this._alignItems="center",this},t.prototype.apply=function(){if(this._overlayRef&&this._overlayRef.hasAttached()){var t=this._overlayRef.overlayElement.style,e=this._overlayRef.hostElement.style,n=this._overlayRef.getConfig();t.position=this._cssPosition,t.marginLeft="100%"===n.width?"0":this._leftOffset,t.marginTop="100%"===n.height?"0":this._topOffset,t.marginBottom=this._bottomOffset,t.marginRight=this._rightOffset,"100%"===n.width?e.justifyContent="flex-start":"center"===this._justifyContent?e.justifyContent="center":"rtl"===this._overlayRef.getConfig().direction?"flex-start"===this._justifyContent?e.justifyContent="flex-end":"flex-end"===this._justifyContent&&(e.justifyContent="flex-start"):e.justifyContent=this._justifyContent,e.alignItems="100%"===n.height?"flex-start":this._alignItems}},t.prototype.dispose=function(){if(!this._isDisposed&&this._overlayRef){var t=this._overlayRef.overlayElement.style,e=this._overlayRef.hostElement,n=e.style;e.classList.remove(B),n.justifyContent=n.alignItems=t.marginTop=t.marginBottom=t.marginLeft=t.marginRight=t.position="",this._overlayRef=null,this._isDisposed=!0}},t}(),z=function(){function t(t,e,n,r){this._viewportRuler=t,this._document=e,this._platform=n,this._overlayContainer=r}return t.prototype.global=function(){return new V},t.prototype.connectedTo=function(t,e,n){return new L(e,n,t,this._viewportRuler,this._document)},t.prototype.flexibleConnectedTo=function(t){return new N(t,this._viewportRuler,this._document,this._platform,this._overlayContainer)},t.decorators=[{type:a.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:i.ViewportRuler},{type:void 0,decorators:[{type:a.Inject,args:[o.DOCUMENT]}]},{type:l.Platform,decorators:[{type:a.Optional}]},{type:A,decorators:[{type:a.Optional}]}]},t.ngInjectableDef=Object(a.defineInjectable)({factory:function(){return new t(Object(a.inject)(i.ViewportRuler),Object(a.inject)(o.DOCUMENT),Object(a.inject)(l.Platform,8),Object(a.inject)(A,8))},token:t,providedIn:"root"}),t}(),H=0,U=function(){function t(t,e,n,r,i,o,a,s,u,c){this.scrollStrategies=t,this._overlayContainer=e,this._componentFactoryResolver=n,this._positionBuilder=r,this._keyboardDispatcher=i,this._injector=o,this._ngZone=a,this._document=s,this._directionality=u,this._location=c}return t.prototype.create=function(t){var e=this._createHostElement(),n=this._createPaneElement(e),r=this._createPortalOutlet(n),i=new w(t);return i.direction=i.direction||this._directionality.value,new P(r,e,n,i,this._ngZone,this._keyboardDispatcher,this._document,this._location)},t.prototype.position=function(){return this._positionBuilder},t.prototype._createPaneElement=function(t){var e=this._document.createElement("div");return e.id="cdk-overlay-"+H++,e.classList.add("cdk-overlay-pane"),t.appendChild(e),e},t.prototype._createHostElement=function(){var t=this._document.createElement("div");return this._overlayContainer.getContainerElement().appendChild(t),t},t.prototype._createPortalOutlet=function(t){return this._appRef||(this._appRef=this._injector.get(a.ApplicationRef)),new p.DomPortalOutlet(t,this._componentFactoryResolver,this._appRef,this._injector)},t.decorators=[{type:a.Injectable}],t.ctorParameters=function(){return[{type:x},{type:A},{type:a.ComponentFactoryResolver},{type:z},{type:M},{type:a.Injector},{type:a.NgZone},{type:void 0,decorators:[{type:a.Inject,args:[o.DOCUMENT]}]},{type:d.Directionality},{type:o.Location,decorators:[{type:a.Optional}]}]},t}(),W=[{originX:"start",originY:"bottom",overlayX:"start",overlayY:"top"},{originX:"start",originY:"top",overlayX:"start",overlayY:"bottom"},{originX:"end",originY:"top",overlayX:"end",overlayY:"bottom"},{originX:"end",originY:"bottom",overlayX:"end",overlayY:"top"}],q=new a.InjectionToken("cdk-connected-overlay-scroll-strategy"),G=function(){function t(t){this.elementRef=t}return t.decorators=[{type:a.Directive,args:[{selector:"[cdk-overlay-origin], [overlay-origin], [cdkOverlayOrigin]",exportAs:"cdkOverlayOrigin"}]}],t.ctorParameters=function(){return[{type:a.ElementRef}]},t}(),Y=function(){function t(t,e,n,r,i){this._overlay=t,this._dir=i,this._hasBackdrop=!1,this._lockPosition=!1,this._growAfterOpen=!1,this._flexibleDimensions=!1,this._push=!1,this._backdropSubscription=u.Subscription.EMPTY,this.viewportMargin=0,this.open=!1,this.backdropClick=new a.EventEmitter,this.positionChange=new a.EventEmitter,this.attach=new a.EventEmitter,this.detach=new a.EventEmitter,this.overlayKeydown=new a.EventEmitter,this._templatePortal=new p.TemplatePortal(e,n),this._scrollStrategyFactory=r,this.scrollStrategy=this._scrollStrategyFactory()}return Object.defineProperty(t.prototype,"offsetX",{get:function(){return this._offsetX},set:function(t){this._offsetX=t,this._position&&this._updatePositionStrategy(this._position)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"offsetY",{get:function(){return this._offsetY},set:function(t){this._offsetY=t,this._position&&this._updatePositionStrategy(this._position)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"hasBackdrop",{get:function(){return this._hasBackdrop},set:function(t){this._hasBackdrop=Object(r.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"lockPosition",{get:function(){return this._lockPosition},set:function(t){this._lockPosition=Object(r.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"flexibleDimensions",{get:function(){return this._flexibleDimensions},set:function(t){this._flexibleDimensions=Object(r.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"growAfterOpen",{get:function(){return this._growAfterOpen},set:function(t){this._growAfterOpen=Object(r.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"push",{get:function(){return this._push},set:function(t){this._push=Object(r.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"overlayRef",{get:function(){return this._overlayRef},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"dir",{get:function(){return this._dir?this._dir.value:"ltr"},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){this._overlayRef&&this._overlayRef.dispose(),this._backdropSubscription.unsubscribe()},t.prototype.ngOnChanges=function(t){this._position&&(this._updatePositionStrategy(this._position),t.origin&&this.open&&this._position.apply()),t.open&&(this.open?this._attachOverlay():this._detachOverlay())},t.prototype._createOverlay=function(){var t=this;this.positions&&this.positions.length||(this.positions=W),this._overlayRef=this._overlay.create(this._buildConfig()),this._overlayRef.keydownEvents().subscribe(function(e){t.overlayKeydown.next(e),e.keyCode===f.ESCAPE&&t._detachOverlay()})},t.prototype._buildConfig=function(){var t=this._position=this._createPositionStrategy(),e=new w({direction:this._dir,positionStrategy:t,scrollStrategy:this.scrollStrategy,hasBackdrop:this.hasBackdrop});return(this.width||0===this.width)&&(e.width=this.width),(this.height||0===this.height)&&(e.height=this.height),(this.minWidth||0===this.minWidth)&&(e.minWidth=this.minWidth),(this.minHeight||0===this.minHeight)&&(e.minHeight=this.minHeight),this.backdropClass&&(e.backdropClass=this.backdropClass),this.panelClass&&(e.panelClass=this.panelClass),e},t.prototype._updatePositionStrategy=function(t){var e=this,n=this.positions.map(function(t){return{originX:t.originX,originY:t.originY,overlayX:t.overlayX,overlayY:t.overlayY,offsetX:t.offsetX||e.offsetX,offsetY:t.offsetY||e.offsetY}});return t.setOrigin(this.origin.elementRef).withPositions(n).withFlexibleDimensions(this.flexibleDimensions).withPush(this.push).withGrowAfterOpen(this.growAfterOpen).withViewportMargin(this.viewportMargin).withLockedPosition(this.lockPosition)},t.prototype._createPositionStrategy=function(){var t=this,e=this._overlay.position().flexibleConnectedTo(this.origin.elementRef);return this._updatePositionStrategy(e),e.positionChanges.subscribe(function(e){return t.positionChange.emit(e)}),e},t.prototype._attachOverlay=function(){var t=this;this._overlayRef?(this._overlayRef.updateSize({width:this.width,minWidth:this.minWidth,height:this.height,minHeight:this.minHeight}),this._overlayRef.getConfig().hasBackdrop=this.hasBackdrop):this._createOverlay(),this._overlayRef.hasAttached()||(this._overlayRef.attach(this._templatePortal),this.attach.emit()),this.hasBackdrop?this._backdropSubscription=this._overlayRef.backdropClick().subscribe(function(e){t.backdropClick.emit(e)}):this._backdropSubscription.unsubscribe()},t.prototype._detachOverlay=function(){this._overlayRef&&(this._overlayRef.detach(),this.detach.emit()),this._backdropSubscription.unsubscribe()},t.decorators=[{type:a.Directive,args:[{selector:"[cdk-connected-overlay], [connected-overlay], [cdkConnectedOverlay]",exportAs:"cdkConnectedOverlay"}]}],t.ctorParameters=function(){return[{type:U},{type:a.TemplateRef},{type:a.ViewContainerRef},{type:void 0,decorators:[{type:a.Inject,args:[q]}]},{type:d.Directionality,decorators:[{type:a.Optional}]}]},t.propDecorators={origin:[{type:a.Input,args:["cdkConnectedOverlayOrigin"]}],positions:[{type:a.Input,args:["cdkConnectedOverlayPositions"]}],offsetX:[{type:a.Input,args:["cdkConnectedOverlayOffsetX"]}],offsetY:[{type:a.Input,args:["cdkConnectedOverlayOffsetY"]}],width:[{type:a.Input,args:["cdkConnectedOverlayWidth"]}],height:[{type:a.Input,args:["cdkConnectedOverlayHeight"]}],minWidth:[{type:a.Input,args:["cdkConnectedOverlayMinWidth"]}],minHeight:[{type:a.Input,args:["cdkConnectedOverlayMinHeight"]}],backdropClass:[{type:a.Input,args:["cdkConnectedOverlayBackdropClass"]}],panelClass:[{type:a.Input,args:["cdkConnectedOverlayPanelClass"]}],viewportMargin:[{type:a.Input,args:["cdkConnectedOverlayViewportMargin"]}],scrollStrategy:[{type:a.Input,args:["cdkConnectedOverlayScrollStrategy"]}],open:[{type:a.Input,args:["cdkConnectedOverlayOpen"]}],hasBackdrop:[{type:a.Input,args:["cdkConnectedOverlayHasBackdrop"]}],lockPosition:[{type:a.Input,args:["cdkConnectedOverlayLockPosition"]}],flexibleDimensions:[{type:a.Input,args:["cdkConnectedOverlayFlexibleDimensions"]}],growAfterOpen:[{type:a.Input,args:["cdkConnectedOverlayGrowAfterOpen"]}],push:[{type:a.Input,args:["cdkConnectedOverlayPush"]}],backdropClick:[{type:a.Output}],positionChange:[{type:a.Output}],attach:[{type:a.Output}],detach:[{type:a.Output}],overlayKeydown:[{type:a.Output}]},t}();function K(t){return function(){return t.scrollStrategies.reposition()}}var X={provide:q,deps:[U],useFactory:K},Q=function(){function t(){}return t.decorators=[{type:a.NgModule,args:[{imports:[d.BidiModule,p.PortalModule,i.ScrollingModule],exports:[Y,G,i.ScrollingModule],declarations:[Y,G],providers:[U,X]}]}],t}(),Z=[U,z,k,i.VIEWPORT_RULER_PROVIDER,I,X],$=function(t){function e(e){return t.call(this,e)||this}return Object(s.__extends)(e,t),e.prototype.ngOnDestroy=function(){t.prototype.ngOnDestroy.call(this),this._fullScreenEventName&&this._fullScreenListener&&this._document.removeEventListener(this._fullScreenEventName,this._fullScreenListener)},e.prototype._createContainer=function(){var e=this;t.prototype._createContainer.call(this),this._adjustParentForFullscreenChange(),this._addFullscreenChangeListener(function(){return e._adjustParentForFullscreenChange()})},e.prototype._adjustParentForFullscreenChange=function(){this._containerElement&&(this.getFullscreenElement()||this._document.body).appendChild(this._containerElement)},e.prototype._addFullscreenChangeListener=function(t){var e=this._getEventName();e&&(this._fullScreenListener&&this._document.removeEventListener(e,this._fullScreenListener),this._document.addEventListener(e,t),this._fullScreenListener=t)},e.prototype._getEventName=function(){return this._fullScreenEventName||(this._document.fullscreenEnabled?this._fullScreenEventName="fullscreenchange":this._document.webkitFullscreenEnabled?this._fullScreenEventName="webkitfullscreenchange":this._document.mozFullScreenEnabled?this._fullScreenEventName="mozfullscreenchange":this._document.msFullscreenEnabled&&(this._fullScreenEventName="MSFullscreenChange")),this._fullScreenEventName},e.prototype.getFullscreenElement=function(){return this._document.fullscreenElement||this._document.webkitFullscreenElement||this._document.mozFullScreenElement||this._document.msFullscreenElement||null},e.decorators=[{type:a.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:a.Inject,args:[o.DOCUMENT]}]}]},e.ngInjectableDef=Object(a.defineInjectable)({factory:function(){return new e(Object(a.inject)(o.DOCUMENT))},token:e,providedIn:"root"}),e}(A)},"./node_modules/@angular/cdk/esm5/platform.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"Platform",function(){return u}),n.d(e,"PlatformModule",function(){return c}),n.d(e,"getSupportedInputTypes",function(){return d}),n.d(e,"supportsPassiveEventListeners",function(){return p}),n.d(e,"normalizePassiveListenerOptions",function(){return f}),n.d(e,"supportsScrollBehavior",function(){return g}),n.d(e,"getRtlScrollAxisType",function(){return y}),n.d(e,"RtlScrollAxisType",function(){return m});var r,i,o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/common/fesm5/common.js"),s="undefined"!=typeof Intl&&Intl.v8BreakIterator,u=function(){function t(t){this._platformId=t,this.isBrowser=this._platformId?Object(a.isPlatformBrowser)(this._platformId):"object"==typeof document&&!!document,this.EDGE=this.isBrowser&&/(edge)/i.test(navigator.userAgent),this.TRIDENT=this.isBrowser&&/(msie|trident)/i.test(navigator.userAgent),this.BLINK=this.isBrowser&&!(!window.chrome&&!s)&&"undefined"!=typeof CSS&&!this.EDGE&&!this.TRIDENT,this.WEBKIT=this.isBrowser&&/AppleWebKit/i.test(navigator.userAgent)&&!this.BLINK&&!this.EDGE&&!this.TRIDENT,this.IOS=this.isBrowser&&/iPad|iPhone|iPod/.test(navigator.userAgent)&&!("MSStream"in window),this.FIREFOX=this.isBrowser&&/(firefox|minefield)/i.test(navigator.userAgent),this.ANDROID=this.isBrowser&&/android/i.test(navigator.userAgent)&&!this.TRIDENT,this.SAFARI=this.isBrowser&&/safari/i.test(navigator.userAgent)&&this.WEBKIT}return t.decorators=[{type:o.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:Object,decorators:[{type:o.Optional},{type:o.Inject,args:[o.PLATFORM_ID]}]}]},t.ngInjectableDef=Object(o.defineInjectable)({factory:function(){return new t(Object(o.inject)(o.PLATFORM_ID,8))},token:t,providedIn:"root"}),t}(),c=function(){function t(){}return t.decorators=[{type:o.NgModule,args:[{}]}],t}(),l=["color","button","checkbox","date","datetime-local","email","file","hidden","image","month","number","password","radio","range","reset","search","submit","tel","text","time","url","week"];function d(){if(r)return r;if("object"!=typeof document||!document)return r=new Set(l);var t=document.createElement("input");return r=new Set(l.filter(function(e){return t.setAttribute("type",e),t.type===e}))}function p(){if(null==i&&"undefined"!=typeof window)try{window.addEventListener("test",null,Object.defineProperty({},"passive",{get:function(){return i=!0}}))}finally{i=i||!1}return i}function f(t){return p()?t:!!t.capture}var h,m={NORMAL:0,NEGATED:1,INVERTED:2};function g(){return!!("object"==typeof document&&"scrollBehavior"in document.documentElement.style)}function y(){if("object"!=typeof document||!document)return m.NORMAL;if(!h){var t=document.createElement("div"),e=t.style;t.dir="rtl",e.height="1px",e.width="1px",e.overflow="auto",e.visibility="hidden",e.pointerEvents="none",e.position="absolute";var n=document.createElement("div"),r=n.style;r.width="2px",r.height="1px",t.appendChild(n),document.body.appendChild(t),h=m.NORMAL,0===t.scrollLeft&&(t.scrollLeft=1,h=0===t.scrollLeft?m.NEGATED:m.INVERTED),t.parentNode.removeChild(t)}return h}m[m.NORMAL]="NORMAL",m[m.NEGATED]="NEGATED",m[m.INVERTED]="INVERTED"},"./node_modules/@angular/cdk/esm5/portal.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"DomPortalHost",function(){return l}),n.d(e,"PortalHostDirective",function(){return p}),n.d(e,"TemplatePortalDirective",function(){return d}),n.d(e,"BasePortalHost",function(){return c}),n.d(e,"Portal",function(){return a}),n.d(e,"ComponentPortal",function(){return s}),n.d(e,"TemplatePortal",function(){return u}),n.d(e,"BasePortalOutlet",function(){return c}),n.d(e,"DomPortalOutlet",function(){return l}),n.d(e,"CdkPortal",function(){return d}),n.d(e,"CdkPortalOutlet",function(){return p}),n.d(e,"PortalModule",function(){return f}),n.d(e,"PortalInjector",function(){return h});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js");function o(){throw Error("Host already has a portal attached")}var a=function(){function t(){}return t.prototype.attach=function(t){return null==t&&function(){throw Error("Attempting to attach a portal to a null PortalOutlet")}(),t.hasAttached()&&o(),this._attachedHost=t,t.attach(this)},t.prototype.detach=function(){var t=this._attachedHost;null==t?function(){throw Error("Attempting to detach a portal that is not attached to a host")}():(this._attachedHost=null,t.detach())},Object.defineProperty(t.prototype,"isAttached",{get:function(){return null!=this._attachedHost},enumerable:!0,configurable:!0}),t.prototype.setAttachedHost=function(t){this._attachedHost=t},t}(),s=function(t){function e(e,n,r,i){var o=t.call(this)||this;return o.component=e,o.viewContainerRef=n,o.injector=r,o.componentFactoryResolver=i,o}return Object(r.__extends)(e,t),e}(a),u=function(t){function e(e,n,r){var i=t.call(this)||this;return i.templateRef=e,i.viewContainerRef=n,i.context=r,i}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"origin",{get:function(){return this.templateRef.elementRef},enumerable:!0,configurable:!0}),e.prototype.attach=function(e,n){return void 0===n&&(n=this.context),this.context=n,t.prototype.attach.call(this,e)},e.prototype.detach=function(){return this.context=void 0,t.prototype.detach.call(this)},e}(a),c=function(){function t(){this._isDisposed=!1}return t.prototype.hasAttached=function(){return!!this._attachedPortal},t.prototype.attach=function(t){return t||function(){throw Error("Must provide a portal to attach")}(),this.hasAttached()&&o(),this._isDisposed&&function(){throw Error("This PortalOutlet has already been disposed")}(),t instanceof s?(this._attachedPortal=t,this.attachComponentPortal(t)):t instanceof u?(this._attachedPortal=t,this.attachTemplatePortal(t)):void function(){throw Error("Attempting to attach an unknown Portal type. BasePortalOutlet accepts either a ComponentPortal or a TemplatePortal.")}()},t.prototype.detach=function(){this._attachedPortal&&(this._attachedPortal.setAttachedHost(null),this._attachedPortal=null),this._invokeDisposeFn()},t.prototype.dispose=function(){this.hasAttached()&&this.detach(),this._invokeDisposeFn(),this._isDisposed=!0},t.prototype.setDisposeFn=function(t){this._disposeFn=t},t.prototype._invokeDisposeFn=function(){this._disposeFn&&(this._disposeFn(),this._disposeFn=null)},t}(),l=function(t){function e(e,n,r,i){var o=t.call(this)||this;return o.outletElement=e,o._componentFactoryResolver=n,o._appRef=r,o._defaultInjector=i,o}return Object(r.__extends)(e,t),e.prototype.attachComponentPortal=function(t){var e,n=this,r=(t.componentFactoryResolver||this._componentFactoryResolver).resolveComponentFactory(t.component);return t.viewContainerRef?(e=t.viewContainerRef.createComponent(r,t.viewContainerRef.length,t.injector||t.viewContainerRef.injector),this.setDisposeFn(function(){return e.destroy()})):(e=r.create(t.injector||this._defaultInjector),this._appRef.attachView(e.hostView),this.setDisposeFn(function(){n._appRef.detachView(e.hostView),e.destroy()})),this.outletElement.appendChild(this._getComponentRootNode(e)),e},e.prototype.attachTemplatePortal=function(t){var e=this,n=t.viewContainerRef,r=n.createEmbeddedView(t.templateRef,t.context);return r.detectChanges(),r.rootNodes.forEach(function(t){return e.outletElement.appendChild(t)}),this.setDisposeFn(function(){var t=n.indexOf(r);-1!==t&&n.remove(t)}),r},e.prototype.dispose=function(){t.prototype.dispose.call(this),null!=this.outletElement.parentNode&&this.outletElement.parentNode.removeChild(this.outletElement)},e.prototype._getComponentRootNode=function(t){return t.hostView.rootNodes[0]},e}(c),d=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"[cdk-portal], [cdkPortal], [portal]",exportAs:"cdkPortal"}]}],e.ctorParameters=function(){return[{type:i.TemplateRef},{type:i.ViewContainerRef}]},e}(u),p=function(t){function e(e,n){var r=t.call(this)||this;return r._componentFactoryResolver=e,r._viewContainerRef=n,r._isInitialized=!1,r.attached=new i.EventEmitter,r}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"portal",{get:function(){return this._attachedPortal},set:function(e){(!this.hasAttached()||e||this._isInitialized)&&(this.hasAttached()&&t.prototype.detach.call(this),e&&t.prototype.attach.call(this,e),this._attachedPortal=e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"attachedRef",{get:function(){return this._attachedRef},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){this._isInitialized=!0},e.prototype.ngOnDestroy=function(){t.prototype.dispose.call(this),this._attachedPortal=null,this._attachedRef=null},e.prototype.attachComponentPortal=function(e){e.setAttachedHost(this);var n=null!=e.viewContainerRef?e.viewContainerRef:this._viewContainerRef,r=(e.componentFactoryResolver||this._componentFactoryResolver).resolveComponentFactory(e.component),i=n.createComponent(r,n.length,e.injector||n.injector);return t.prototype.setDisposeFn.call(this,function(){return i.destroy()}),this._attachedPortal=e,this._attachedRef=i,this.attached.emit(i),i},e.prototype.attachTemplatePortal=function(e){var n=this;e.setAttachedHost(this);var r=this._viewContainerRef.createEmbeddedView(e.templateRef,e.context);return t.prototype.setDisposeFn.call(this,function(){return n._viewContainerRef.clear()}),this._attachedPortal=e,this._attachedRef=r,this.attached.emit(r),r},e.decorators=[{type:i.Directive,args:[{selector:"[cdkPortalOutlet], [cdkPortalHost], [portalHost]",exportAs:"cdkPortalOutlet, cdkPortalHost",inputs:["portal: cdkPortalOutlet"]}]}],e.ctorParameters=function(){return[{type:i.ComponentFactoryResolver},{type:i.ViewContainerRef}]},e.propDecorators={attached:[{type:i.Output}]},e}(c),f=function(){function t(){}return t.decorators=[{type:i.NgModule,args:[{exports:[d,p],declarations:[d,p]}]}],t}(),h=function(){function t(t,e){this._parentInjector=t,this._customTokens=e}return t.prototype.get=function(t,e){var n=this._customTokens.get(t);return void 0!==n?n:this._parentInjector.get(t,e)},t}()},"./node_modules/@angular/cdk/esm5/scrolling.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"_fixedSizeVirtualScrollStrategyFactory",function(){return f}),n.d(e,"FixedSizeVirtualScrollStrategy",function(){return p}),n.d(e,"CdkFixedSizeVirtualScroll",function(){return h}),n.d(e,"SCROLL_DISPATCHER_PROVIDER_FACTORY",function(){return y}),n.d(e,"DEFAULT_SCROLL_TIME",function(){return m}),n.d(e,"ScrollDispatcher",function(){return g}),n.d(e,"SCROLL_DISPATCHER_PROVIDER",function(){return _}),n.d(e,"CdkScrollable",function(){return v}),n.d(e,"ScrollingModule",function(){return j}),n.d(e,"ScrollDispatchModule",function(){return O}),n.d(e,"VIEWPORT_RULER_PROVIDER_FACTORY",function(){return S}),n.d(e,"DEFAULT_RESIZE_TIME",function(){return C}),n.d(e,"ViewportRuler",function(){return E}),n.d(e,"VIEWPORT_RULER_PROVIDER",function(){return M}),n.d(e,"CdkVirtualForOf",function(){return w}),n.d(e,"VIRTUAL_SCROLL_STRATEGY",function(){return d}),n.d(e,"CdkVirtualScrollViewport",function(){return b});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),o=n("./node_modules/rxjs/_esm5/index.js"),a=n("./node_modules/rxjs/_esm5/operators/index.js"),s=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),u=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),c=n("./node_modules/tslib/tslib.es6.js"),l=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),d=new r.InjectionToken("VIRTUAL_SCROLL_STRATEGY"),p=function(){function t(t,e,n){this._scrolledIndexChange=new o.Subject,this.scrolledIndexChange=this._scrolledIndexChange.pipe(Object(a.distinctUntilChanged)()),this._viewport=null,this._itemSize=t,this._minBufferPx=e,this._maxBufferPx=n}return t.prototype.attach=function(t){this._viewport=t,this._updateTotalContentSize(),this._updateRenderedRange()},t.prototype.detach=function(){this._scrolledIndexChange.complete(),this._viewport=null},t.prototype.updateItemAndBufferSize=function(t,e,n){if(n<e)throw Error("CDK virtual scroll: maxBufferPx must be greater than or equal to minBufferPx");this._itemSize=t,this._minBufferPx=e,this._maxBufferPx=n,this._updateTotalContentSize(),this._updateRenderedRange()},t.prototype.onContentScrolled=function(){this._updateRenderedRange()},t.prototype.onDataLengthChanged=function(){this._updateTotalContentSize(),this._updateRenderedRange()},t.prototype.onContentRendered=function(){},t.prototype.onRenderedOffsetChanged=function(){},t.prototype.scrollToIndex=function(t,e){this._viewport&&this._viewport.scrollToOffset(t*this._itemSize,e)},t.prototype._updateTotalContentSize=function(){this._viewport&&this._viewport.setTotalContentSize(this._viewport.getDataLength()*this._itemSize)},t.prototype._updateRenderedRange=function(){if(this._viewport){var t=this._viewport.measureScrollOffset(),e=t/this._itemSize,n=this._viewport.getRenderedRange(),r={start:n.start,end:n.end},i=this._viewport.getViewportSize(),o=this._viewport.getDataLength(),a=t-r.start*this._itemSize;if(a<this._minBufferPx&&0!=r.start){var s=Math.ceil((this._maxBufferPx-a)/this._itemSize);r.start=Math.max(0,r.start-s),r.end=Math.min(o,Math.ceil(e+(i+this._minBufferPx)/this._itemSize))}else{var u=r.end*this._itemSize-(t+i);if(u<this._minBufferPx&&r.end!=o){var c=Math.ceil((this._maxBufferPx-u)/this._itemSize);c>0&&(r.end=Math.min(o,r.end+c),r.start=Math.max(0,Math.floor(e-this._minBufferPx/this._itemSize)))}}this._viewport.setRenderedRange(r),this._viewport.setRenderedContentOffset(this._itemSize*r.start),this._scrolledIndexChange.next(Math.floor(e))}},t}();function f(t){return t._scrollStrategy}var h=function(){function t(){this._itemSize=20,this._minBufferPx=100,this._maxBufferPx=200,this._scrollStrategy=new p(this.itemSize,this.minBufferPx,this.maxBufferPx)}return Object.defineProperty(t.prototype,"itemSize",{get:function(){return this._itemSize},set:function(t){this._itemSize=Object(i.coerceNumberProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"minBufferPx",{get:function(){return this._minBufferPx},set:function(t){this._minBufferPx=Object(i.coerceNumberProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"maxBufferPx",{get:function(){return this._maxBufferPx},set:function(t){this._maxBufferPx=Object(i.coerceNumberProperty)(t)},enumerable:!0,configurable:!0}),t.prototype.ngOnChanges=function(){this._scrollStrategy.updateItemAndBufferSize(this.itemSize,this.minBufferPx,this.maxBufferPx)},t.decorators=[{type:r.Directive,args:[{selector:"cdk-virtual-scroll-viewport[itemSize]",providers:[{provide:d,useFactory:f,deps:[Object(r.forwardRef)(function(){return t})]}]}]}],t.propDecorators={itemSize:[{type:r.Input}],minBufferPx:[{type:r.Input}],maxBufferPx:[{type:r.Input}]},t}(),m=20,g=function(){function t(t,e){this._ngZone=t,this._platform=e,this._scrolled=new o.Subject,this._globalSubscription=null,this._scrolledCount=0,this.scrollContainers=new Map}return t.prototype.register=function(t){var e=this;this.scrollContainers.has(t)||this.scrollContainers.set(t,t.elementScrolled().subscribe(function(){return e._scrolled.next(t)}))},t.prototype.deregister=function(t){var e=this.scrollContainers.get(t);e&&(e.unsubscribe(),this.scrollContainers.delete(t))},t.prototype.scrolled=function(t){var e=this;return void 0===t&&(t=m),this._platform.isBrowser?o.Observable.create(function(n){e._globalSubscription||e._addGlobalListener();var r=t>0?e._scrolled.pipe(Object(a.auditTime)(t)).subscribe(n):e._scrolled.subscribe(n);return e._scrolledCount++,function(){r.unsubscribe(),e._scrolledCount--,e._scrolledCount||e._removeGlobalListener()}}):Object(o.of)()},t.prototype.ngOnDestroy=function(){var t=this;this._removeGlobalListener(),this.scrollContainers.forEach(function(e,n){return t.deregister(n)}),this._scrolled.complete()},t.prototype.ancestorScrolled=function(t,e){var n=this.getAncestorScrollContainers(t);return this.scrolled(e).pipe(Object(a.filter)(function(t){return!t||n.indexOf(t)>-1}))},t.prototype.getAncestorScrollContainers=function(t){var e=this,n=[];return this.scrollContainers.forEach(function(r,i){e._scrollableContainsElement(i,t)&&n.push(i)}),n},t.prototype._scrollableContainsElement=function(t,e){var n=e.nativeElement,r=t.getElementRef().nativeElement;do{if(n==r)return!0}while(n=n.parentElement);return!1},t.prototype._addGlobalListener=function(){var t=this;this._globalSubscription=this._ngZone.runOutsideAngular(function(){return Object(o.fromEvent)(window.document,"scroll").subscribe(function(){return t._scrolled.next()})})},t.prototype._removeGlobalListener=function(){this._globalSubscription&&(this._globalSubscription.unsubscribe(),this._globalSubscription=null)},t.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:r.NgZone},{type:s.Platform}]},t.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new t(Object(r.inject)(r.NgZone),Object(r.inject)(s.Platform))},token:t,providedIn:"root"}),t}();function y(t,e,n){return t||new g(e,n)}var _={provide:g,deps:[[new r.Optional,new r.SkipSelf,g],r.NgZone,s.Platform],useFactory:y},v=function(){function t(t,e,n,r){var i=this;this.elementRef=t,this.scrollDispatcher=e,this.ngZone=n,this.dir=r,this._destroyed=new o.Subject,this._elementScrolled=o.Observable.create(function(t){return i.ngZone.runOutsideAngular(function(){return Object(o.fromEvent)(i.elementRef.nativeElement,"scroll").pipe(Object(a.takeUntil)(i._destroyed)).subscribe(t)})})}return t.prototype.ngOnInit=function(){this.scrollDispatcher.register(this)},t.prototype.ngOnDestroy=function(){this.scrollDispatcher.deregister(this),this._destroyed.next(),this._destroyed.complete()},t.prototype.elementScrolled=function(){return this._elementScrolled},t.prototype.getElementRef=function(){return this.elementRef},t.prototype.scrollTo=function(t){var e=this.elementRef.nativeElement,n=this.dir&&"rtl"==this.dir.value;t.left=null==t.left?n?t.end:t.start:t.left,t.right=null==t.right?n?t.start:t.end:t.right,null!=t.bottom&&(t.top=e.scrollHeight-e.clientHeight-t.bottom),n&&Object(s.getRtlScrollAxisType)()!=s.RtlScrollAxisType.NORMAL?(null!=t.left&&(t.right=e.scrollWidth-e.clientWidth-t.left),Object(s.getRtlScrollAxisType)()==s.RtlScrollAxisType.INVERTED?t.left=t.right:Object(s.getRtlScrollAxisType)()==s.RtlScrollAxisType.NEGATED&&(t.left=t.right?-t.right:t.right)):null!=t.right&&(t.left=e.scrollWidth-e.clientWidth-t.right),this._applyScrollToOptions(t)},t.prototype._applyScrollToOptions=function(t){var e=this.elementRef.nativeElement;Object(s.supportsScrollBehavior)()?e.scrollTo(t):(null!=t.top&&(e.scrollTop=t.top),null!=t.left&&(e.scrollLeft=t.left))},t.prototype.measureScrollOffset=function(t){var e=this.elementRef.nativeElement;if("top"==t)return e.scrollTop;if("bottom"==t)return e.scrollHeight-e.clientHeight-e.scrollTop;var n=this.dir&&"rtl"==this.dir.value;return"start"==t?t=n?"right":"left":"end"==t&&(t=n?"left":"right"),n&&Object(s.getRtlScrollAxisType)()==s.RtlScrollAxisType.INVERTED?"left"==t?e.scrollWidth-e.clientWidth-e.scrollLeft:e.scrollLeft:n&&Object(s.getRtlScrollAxisType)()==s.RtlScrollAxisType.NEGATED?"left"==t?e.scrollLeft+e.scrollWidth-e.clientWidth:-e.scrollLeft:"left"==t?e.scrollLeft:e.scrollWidth-e.clientWidth-e.scrollLeft},t.decorators=[{type:r.Directive,args:[{selector:"[cdk-scrollable], [cdkScrollable]"}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:g},{type:r.NgZone},{type:u.Directionality,decorators:[{type:r.Optional}]}]},t}();var b=function(t){function e(e,n,r,i,a,s){var u=t.call(this,e,s,r,a)||this;if(u.elementRef=e,u._changeDetectorRef=n,u._scrollStrategy=i,u._detachedSubject=new o.Subject,u._renderedRangeSubject=new o.Subject,u.orientation="vertical",u.scrolledIndexChange=o.Observable.create(function(t){return u._scrollStrategy.scrolledIndexChange.subscribe(function(e){return Promise.resolve().then(function(){return u.ngZone.run(function(){return t.next(e)})})})}),u.renderedRangeStream=u._renderedRangeSubject.asObservable(),u._totalContentSizeTransform="",u._totalContentSize=0,u._renderedRange={start:0,end:0},u._dataLength=0,u._viewportSize=0,u._renderedContentOffset=0,u._renderedContentOffsetNeedsRewrite=!1,u._isChangeDetectionPending=!1,u._runAfterChangeDetection=[],!i)throw Error('Error: cdk-virtual-scroll-viewport requires the "itemSize" property to be set.');return u}return Object(c.__extends)(e,t),e.prototype.ngOnInit=function(){var e=this;t.prototype.ngOnInit.call(this),this.ngZone.runOutsideAngular(function(){return Promise.resolve().then(function(){e._measureViewportSize(),e._scrollStrategy.attach(e),e.elementScrolled().pipe(Object(a.startWith)(null),Object(a.auditTime)(0,o.animationFrameScheduler)).subscribe(function(){return e._scrollStrategy.onContentScrolled()}),e._markChangeDetectionNeeded()})})},e.prototype.ngOnDestroy=function(){this.detach(),this._scrollStrategy.detach(),this._renderedRangeSubject.complete(),this._detachedSubject.complete(),t.prototype.ngOnDestroy.call(this)},e.prototype.attach=function(t){var e=this;if(this._forOf)throw Error("CdkVirtualScrollViewport is already attached.");this.ngZone.runOutsideAngular(function(){e._forOf=t,e._forOf.dataStream.pipe(Object(a.takeUntil)(e._detachedSubject)).subscribe(function(t){var n=t.length;n!==e._dataLength&&(e._dataLength=n,e._scrollStrategy.onDataLengthChanged()),e._doChangeDetection()})})},e.prototype.detach=function(){this._forOf=null,this._detachedSubject.next()},e.prototype.getDataLength=function(){return this._dataLength},e.prototype.getViewportSize=function(){return this._viewportSize},e.prototype.getRenderedRange=function(){return this._renderedRange},e.prototype.setTotalContentSize=function(t){if(this._totalContentSize!==t){this._totalContentSize=t;var e="horizontal"==this.orientation?"X":"Y";this._totalContentSizeTransform="scale"+e+"("+this._totalContentSize+")",this._markChangeDetectionNeeded()}},e.prototype.setRenderedRange=function(t){var e,n,r=this;e=this._renderedRange,n=t,(e.start!=n.start||e.end!=n.end)&&(this._renderedRangeSubject.next(this._renderedRange=t),this._markChangeDetectionNeeded(function(){return r._scrollStrategy.onContentRendered()}))},e.prototype.getOffsetToRenderedContentStart=function(){return this._renderedContentOffsetNeedsRewrite?null:this._renderedContentOffset},e.prototype.setRenderedContentOffset=function(t,e){var n=this;void 0===e&&(e="to-start");var r=this.dir&&"rtl"==this.dir.value,i="horizontal"==this.orientation,o=i?"X":"Y",a="translate"+o+"("+Number((i&&r?-1:1)*t)+"px)";this._renderedContentOffset=t,"to-end"===e&&(a+=" translate"+o+"(-100%)",this._renderedContentOffsetNeedsRewrite=!0),this._renderedContentTransform!=a&&(this._renderedContentTransform=a,this._markChangeDetectionNeeded(function(){n._renderedContentOffsetNeedsRewrite?(n._renderedContentOffset-=n.measureRenderedContentSize(),n._renderedContentOffsetNeedsRewrite=!1,n.setRenderedContentOffset(n._renderedContentOffset)):n._scrollStrategy.onRenderedOffsetChanged()}))},e.prototype.scrollToOffset=function(t,e){void 0===e&&(e="auto");var n={behavior:e};"horizontal"===this.orientation?n.start=t:n.top=t,this.scrollTo(n)},e.prototype.scrollToIndex=function(t,e){void 0===e&&(e="auto"),this._scrollStrategy.scrollToIndex(t,e)},e.prototype.measureScrollOffset=function(e){return t.prototype.measureScrollOffset.call(this,e||("horizontal"===this.orientation?"start":"top"))},e.prototype.measureRenderedContentSize=function(){var t=this._contentWrapper.nativeElement;return"horizontal"===this.orientation?t.offsetWidth:t.offsetHeight},e.prototype.measureRangeSize=function(t){return this._forOf?this._forOf.measureRangeSize(t,this.orientation):0},e.prototype.checkViewportSize=function(){this._measureViewportSize(),this._scrollStrategy.onDataLengthChanged()},e.prototype._measureViewportSize=function(){var t=this.elementRef.nativeElement;this._viewportSize="horizontal"===this.orientation?t.clientWidth:t.clientHeight},e.prototype._markChangeDetectionNeeded=function(t){var e=this;t&&this._runAfterChangeDetection.push(t),this._isChangeDetectionPending||(this._isChangeDetectionPending=!0,this.ngZone.runOutsideAngular(function(){return Promise.resolve().then(function(){e._doChangeDetection()})}))},e.prototype._doChangeDetection=function(){var t=this;this._isChangeDetectionPending=!1,this.ngZone.run(function(){return t._changeDetectorRef.markForCheck()}),this._contentWrapper.nativeElement.style.transform=this._renderedContentTransform;var e=this._runAfterChangeDetection;this._runAfterChangeDetection=[];for(var n=0,r=e;n<r.length;n++){(0,r[n])()}},e.decorators=[{type:r.Component,args:[{selector:"cdk-virtual-scroll-viewport",template:'<div #contentWrapper class="cdk-virtual-scroll-content-wrapper"><ng-content></ng-content></div><div class="cdk-virtual-scroll-spacer" [style.transform]="_totalContentSizeTransform"></div>',styles:["cdk-virtual-scroll-viewport{display:block;position:relative;overflow:auto;contain:strict;transform:translateZ(0);will-change:scroll-position;-webkit-overflow-scrolling:touch}.cdk-virtual-scroll-content-wrapper{position:absolute;top:0;left:0;contain:content}[dir=rtl] .cdk-virtual-scroll-content-wrapper{right:0;left:auto}.cdk-virtual-scroll-orientation-horizontal .cdk-virtual-scroll-content-wrapper{min-height:100%}.cdk-virtual-scroll-orientation-horizontal .cdk-virtual-scroll-content-wrapper>dl:not([cdkVirtualFor]),.cdk-virtual-scroll-orientation-horizontal .cdk-virtual-scroll-content-wrapper>ol:not([cdkVirtualFor]),.cdk-virtual-scroll-orientation-horizontal .cdk-virtual-scroll-content-wrapper>table:not([cdkVirtualFor]),.cdk-virtual-scroll-orientation-horizontal .cdk-virtual-scroll-content-wrapper>ul:not([cdkVirtualFor]){padding-left:0;padding-right:0;margin-left:0;margin-right:0;border-left-width:0;border-right-width:0;outline:0}.cdk-virtual-scroll-orientation-vertical .cdk-virtual-scroll-content-wrapper{min-width:100%}.cdk-virtual-scroll-orientation-vertical .cdk-virtual-scroll-content-wrapper>dl:not([cdkVirtualFor]),.cdk-virtual-scroll-orientation-vertical .cdk-virtual-scroll-content-wrapper>ol:not([cdkVirtualFor]),.cdk-virtual-scroll-orientation-vertical .cdk-virtual-scroll-content-wrapper>table:not([cdkVirtualFor]),.cdk-virtual-scroll-orientation-vertical .cdk-virtual-scroll-content-wrapper>ul:not([cdkVirtualFor]){padding-top:0;padding-bottom:0;margin-top:0;margin-bottom:0;border-top-width:0;border-bottom-width:0;outline:0}.cdk-virtual-scroll-spacer{position:absolute;top:0;left:0;height:1px;width:1px;transform-origin:0 0}[dir=rtl] .cdk-virtual-scroll-spacer{right:0;left:auto;transform-origin:100% 0}"],host:{class:"cdk-virtual-scroll-viewport","[class.cdk-virtual-scroll-orientation-horizontal]":'orientation === "horizontal"',"[class.cdk-virtual-scroll-orientation-vertical]":'orientation !== "horizontal"'},encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,providers:[{provide:v,useExisting:e}]}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:r.ChangeDetectorRef},{type:r.NgZone},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[d]}]},{type:u.Directionality,decorators:[{type:r.Optional}]},{type:g}]},e.propDecorators={orientation:[{type:r.Input}],scrolledIndexChange:[{type:r.Output}],_contentWrapper:[{type:r.ViewChild,args:["contentWrapper"]}]},e}(v);function x(t,e){var n=e;if(!n.getBoundingClientRect)return 0;var r=n.getBoundingClientRect();return"horizontal"==t?r.width:r.height}var w=function(){function t(t,e,n,r,i){var s=this;this._viewContainerRef=t,this._template=e,this._differs=n,this._viewport=r,this.viewChange=new o.Subject,this._dataSourceChanges=new o.Subject,this.cdkVirtualForTemplateCacheSize=20,this.dataStream=this._dataSourceChanges.pipe(Object(a.startWith)(null),Object(a.pairwise)(),Object(a.switchMap)(function(t){var e=t[0],n=t[1];return s._changeDataSource(e,n)}),Object(a.shareReplay)(1)),this._differ=null,this._templateCache=[],this._needsUpdate=!1,this._destroyed=new o.Subject,this.dataStream.subscribe(function(t){s._data=t,s._onRenderedDataChange()}),this._viewport.renderedRangeStream.pipe(Object(a.takeUntil)(this._destroyed)).subscribe(function(t){s._renderedRange=t,i.run(function(){return s.viewChange.next(s._renderedRange)}),s._onRenderedDataChange()}),this._viewport.attach(this)}return Object.defineProperty(t.prototype,"cdkVirtualForOf",{get:function(){return this._cdkVirtualForOf},set:function(t){this._cdkVirtualForOf=t;var e=t instanceof l.DataSource?t:new l.ArrayDataSource(t instanceof o.Observable?t:Array.prototype.slice.call(t||[]));this._dataSourceChanges.next(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"cdkVirtualForTrackBy",{get:function(){return this._cdkVirtualForTrackBy},set:function(t){var e=this;this._needsUpdate=!0,this._cdkVirtualForTrackBy=t?function(n,r){return t(n+(e._renderedRange?e._renderedRange.start:0),r)}:void 0},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"cdkVirtualForTemplate",{set:function(t){t&&(this._needsUpdate=!0,this._template=t)},enumerable:!0,configurable:!0}),t.prototype.measureRangeSize=function(t,e){if(t.start>=t.end)return 0;if(t.start<this._renderedRange.start||t.end>this._renderedRange.end)throw Error("Error: attempted to measure an item that isn't rendered.");for(var n=t.start-this._renderedRange.start,r=0,i=t.end-t.start;i--;)for(var o=this._viewContainerRef.get(i+n),a=o?o.rootNodes.length:0;a--;)r+=x(e,o.rootNodes[a]);return r},t.prototype.ngDoCheck=function(){if(this._differ&&this._needsUpdate){var t=this._differ.diff(this._renderedItems);t?this._applyChanges(t):this._updateContext(),this._needsUpdate=!1}},t.prototype.ngOnDestroy=function(){this._viewport.detach(),this._dataSourceChanges.complete(),this.viewChange.complete(),this._destroyed.next(),this._destroyed.complete();for(var t=0,e=this._templateCache;t<e.length;t++){e[t].destroy()}},t.prototype._onRenderedDataChange=function(){this._renderedRange&&(this._renderedItems=this._data.slice(this._renderedRange.start,this._renderedRange.end),this._differ||(this._differ=this._differs.find(this._renderedItems).create(this.cdkVirtualForTrackBy)),this._needsUpdate=!0)},t.prototype._changeDataSource=function(t,e){return t&&t.disconnect(this),this._needsUpdate=!0,e.connect(this)},t.prototype._updateContext=function(){for(var t=this._data.length,e=this._viewContainerRef.length;e--;){var n=this._viewContainerRef.get(e);n.context.index=this._renderedRange.start+e,n.context.count=t,this._updateComputedContextProperties(n.context),n.detectChanges()}},t.prototype._applyChanges=function(t){var e=this;t.forEachOperation(function(t,n,r){if(null==t.previousIndex){var i=e._getViewForNewItem();e._viewContainerRef.insert(i,r),i.context.$implicit=t.item}else if(null==r)e._cacheView(e._viewContainerRef.detach(n));else{i=e._viewContainerRef.get(n);e._viewContainerRef.move(i,r),i.context.$implicit=t.item}}),t.forEachIdentityChange(function(t){e._viewContainerRef.get(t.currentIndex).context.$implicit=t.item});for(var n=this._data.length,r=this._viewContainerRef.length;r--;){var i=this._viewContainerRef.get(r);i.context.index=this._renderedRange.start+r,i.context.count=n,this._updateComputedContextProperties(i.context)}},t.prototype._cacheView=function(t){if(this._templateCache.length<this.cdkVirtualForTemplateCacheSize)this._templateCache.push(t);else{var e=this._viewContainerRef.indexOf(t);-1===e?t.destroy():this._viewContainerRef.remove(e)}},t.prototype._getViewForNewItem=function(){return this._templateCache.pop()||this._viewContainerRef.createEmbeddedView(this._template,{$implicit:null,cdkVirtualForOf:this._cdkVirtualForOf,index:-1,count:-1,first:!1,last:!1,odd:!1,even:!1})},t.prototype._updateComputedContextProperties=function(t){t.first=0===t.index,t.last=t.index===t.count-1,t.even=t.index%2==0,t.odd=!t.even},t.decorators=[{type:r.Directive,args:[{selector:"[cdkVirtualFor][cdkVirtualForOf]"}]}],t.ctorParameters=function(){return[{type:r.ViewContainerRef},{type:r.TemplateRef},{type:r.IterableDiffers},{type:b,decorators:[{type:r.SkipSelf}]},{type:r.NgZone}]},t.propDecorators={cdkVirtualForOf:[{type:r.Input}],cdkVirtualForTrackBy:[{type:r.Input}],cdkVirtualForTemplate:[{type:r.Input}],cdkVirtualForTemplateCacheSize:[{type:r.Input}]},t}(),j=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[u.BidiModule,s.PlatformModule],exports:[u.BidiModule,h,v,w,b],declarations:[h,v,w,b]}]}],t}(),O=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[j],exports:[j]}]}],t}(),C=20,E=function(){function t(t,e){var n=this;this._platform=t,e.runOutsideAngular(function(){n._change=t.isBrowser?Object(o.merge)(Object(o.fromEvent)(window,"resize"),Object(o.fromEvent)(window,"orientationchange")):Object(o.of)(),n._invalidateCache=n.change().subscribe(function(){return n._updateViewportSize()})})}return t.prototype.ngOnDestroy=function(){this._invalidateCache.unsubscribe()},t.prototype.getViewportSize=function(){this._viewportSize||this._updateViewportSize();var t={width:this._viewportSize.width,height:this._viewportSize.height};return this._platform.isBrowser||(this._viewportSize=null),t},t.prototype.getViewportRect=function(){var t=this.getViewportScrollPosition(),e=this.getViewportSize(),n=e.width,r=e.height;return{top:t.top,left:t.left,bottom:t.top+r,right:t.left+n,height:r,width:n}},t.prototype.getViewportScrollPosition=function(){if(!this._platform.isBrowser)return{top:0,left:0};var t=document.documentElement,e=t.getBoundingClientRect();return{top:-e.top||document.body.scrollTop||window.scrollY||t.scrollTop||0,left:-e.left||document.body.scrollLeft||window.scrollX||t.scrollLeft||0}},t.prototype.change=function(t){return void 0===t&&(t=C),t>0?this._change.pipe(Object(a.auditTime)(t)):this._change},t.prototype._updateViewportSize=function(){this._viewportSize=this._platform.isBrowser?{width:window.innerWidth,height:window.innerHeight}:{width:0,height:0}},t.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:s.Platform},{type:r.NgZone}]},t.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new t(Object(r.inject)(s.Platform),Object(r.inject)(r.NgZone))},token:t,providedIn:"root"}),t}();function S(t,e,n){return t||new E(e,n)}var M={provide:E,deps:[[new r.Optional,new r.SkipSelf,E],s.Platform,r.NgZone],useFactory:S}},"./node_modules/@angular/cdk/esm5/stepper.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"StepperSelectionEvent",function(){return h}),n.d(e,"STEP_STATE",function(){return m}),n.d(e,"STEPPER_GLOBAL_OPTIONS",function(){return g}),n.d(e,"MAT_STEPPER_GLOBAL_OPTIONS",function(){return y}),n.d(e,"CdkStep",function(){return _}),n.d(e,"CdkStepper",function(){return v}),n.d(e,"CdkStepLabel",function(){return d}),n.d(e,"CdkStepperNext",function(){return b}),n.d(e,"CdkStepperPrevious",function(){return x}),n.d(e,"CdkStepperModule",function(){return w}),n.d(e,"CdkStepHeader",function(){return p});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),o=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),a=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),s=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),u=n("./node_modules/@angular/common/fesm5/common.js"),c=(n("./node_modules/@angular/forms/fesm5/forms.js"),n("./node_modules/rxjs/_esm5/index.js")),l=n("./node_modules/rxjs/_esm5/operators/index.js"),d=function(){function t(t){this.template=t}return t.decorators=[{type:r.Directive,args:[{selector:"[cdkStepLabel]"}]}],t.ctorParameters=function(){return[{type:r.TemplateRef}]},t}(),p=function(){function t(t){this._elementRef=t}return t.prototype.focus=function(){this._elementRef.nativeElement.focus()},t.decorators=[{type:r.Directive,args:[{selector:"[cdkStepHeader]",host:{role:"tab"}}]}],t.ctorParameters=function(){return[{type:r.ElementRef}]},t}(),f=0,h=function(){return function(){}}(),m={NUMBER:"number",EDIT:"edit",DONE:"done",ERROR:"error"},g=new r.InjectionToken("STEPPER_GLOBAL_OPTIONS"),y=g,_=function(){function t(t,e){this._stepper=t,this.interacted=!1,this._editable=!0,this._optional=!1,this._customCompleted=null,this._customError=null,this._stepperOptions=e||{},this._displayDefaultIndicatorType=!1!==this._stepperOptions.displayDefaultIndicatorType,this._showError=!!this._stepperOptions.showError}return Object.defineProperty(t.prototype,"editable",{get:function(){return this._editable},set:function(t){this._editable=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"optional",{get:function(){return this._optional},set:function(t){this._optional=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"completed",{get:function(){return null==this._customCompleted?this._getDefaultCompleted():this._customCompleted},set:function(t){this._customCompleted=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),t.prototype._getDefaultCompleted=function(){return this.stepControl?this.stepControl.valid&&this.interacted:this.interacted},Object.defineProperty(t.prototype,"hasError",{get:function(){return null==this._customError?this._getDefaultError():this._customError},set:function(t){this._customError=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),t.prototype._getDefaultError=function(){return this.stepControl&&this.stepControl.invalid&&this.interacted},t.prototype.select=function(){this._stepper.selected=this},t.prototype.reset=function(){this.interacted=!1,null!=this._customCompleted&&(this._customCompleted=!1),null!=this._customError&&(this._customError=!1),this.stepControl&&this.stepControl.reset()},t.prototype.ngOnChanges=function(){this._stepper._stateChanged()},t.decorators=[{type:r.Component,args:[{selector:"cdk-step",exportAs:"cdkStep",template:"<ng-template><ng-content></ng-content></ng-template>",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:v,decorators:[{type:r.Inject,args:[Object(r.forwardRef)(function(){return v})]}]},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[g]}]}]},t.propDecorators={stepLabel:[{type:r.ContentChild,args:[d]}],content:[{type:r.ViewChild,args:[r.TemplateRef]}],stepControl:[{type:r.Input}],label:[{type:r.Input}],errorMessage:[{type:r.Input}],ariaLabel:[{type:r.Input,args:["aria-label"]}],ariaLabelledby:[{type:r.Input,args:["aria-labelledby"]}],state:[{type:r.Input}],editable:[{type:r.Input}],optional:[{type:r.Input}],completed:[{type:r.Input}],hasError:[{type:r.Input}]},t}(),v=function(){function t(t,e,n,i){this._dir=t,this._changeDetectorRef=e,this._elementRef=n,this._destroyed=new c.Subject,this._linear=!1,this._selectedIndex=0,this.selectionChange=new r.EventEmitter,this._orientation="horizontal",this._groupId=f++,this._document=i}return Object.defineProperty(t.prototype,"linear",{get:function(){return this._linear},set:function(t){this._linear=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selectedIndex",{get:function(){return this._selectedIndex},set:function(t){if(this._steps){if(t<0||t>this._steps.length-1)throw Error("cdkStepper: Cannot assign out-of-bounds value to `selectedIndex`.");this._selectedIndex!=t&&!this._anyControlsInvalidOrPending(t)&&(t>=this._selectedIndex||this._steps.toArray()[t].editable)&&this._updateSelectedItemIndex(t)}else this._selectedIndex=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selected",{get:function(){return this._steps?this._steps.toArray()[this.selectedIndex]:void 0},set:function(t){this.selectedIndex=this._steps?this._steps.toArray().indexOf(t):-1},enumerable:!0,configurable:!0}),t.prototype.ngAfterViewInit=function(){var t=this;this._keyManager=new i.FocusKeyManager(this._stepHeader).withWrap().withVerticalOrientation("vertical"===this._orientation),(this._dir?this._dir.change:Object(c.of)()).pipe(Object(l.startWith)(this._layoutDirection()),Object(l.takeUntil)(this._destroyed)).subscribe(function(e){return t._keyManager.withHorizontalOrientation(e)}),this._keyManager.updateActiveItemIndex(this._selectedIndex),this._steps.changes.pipe(Object(l.takeUntil)(this._destroyed)).subscribe(function(){t.selected||(t._selectedIndex=Math.max(t._selectedIndex-1,0))})},t.prototype.ngOnDestroy=function(){this._destroyed.next(),this._destroyed.complete()},t.prototype.next=function(){this.selectedIndex=Math.min(this._selectedIndex+1,this._steps.length-1)},t.prototype.previous=function(){this.selectedIndex=Math.max(this._selectedIndex-1,0)},t.prototype.reset=function(){this._updateSelectedItemIndex(0),this._steps.forEach(function(t){return t.reset()}),this._stateChanged()},t.prototype._getStepLabelId=function(t){return"cdk-step-label-"+this._groupId+"-"+t},t.prototype._getStepContentId=function(t){return"cdk-step-content-"+this._groupId+"-"+t},t.prototype._stateChanged=function(){this._changeDetectorRef.markForCheck()},t.prototype._getAnimationDirection=function(t){var e=t-this._selectedIndex;return e<0?"rtl"===this._layoutDirection()?"next":"previous":e>0?"rtl"===this._layoutDirection()?"previous":"next":"current"},t.prototype._getIndicatorType=function(t,e){void 0===e&&(e=m.NUMBER);var n=this._steps.toArray()[t],r=this._isCurrentStep(t);return n._displayDefaultIndicatorType?this._getDefaultIndicatorLogic(n,r):this._getGuidelineLogic(n,r,e)},t.prototype._getDefaultIndicatorLogic=function(t,e){return t._showError&&t.hasError&&!e?m.ERROR:!t.completed||e?m.NUMBER:t.editable?m.EDIT:m.DONE},t.prototype._getGuidelineLogic=function(t,e,n){return void 0===n&&(n=m.NUMBER),t._showError&&t.hasError&&!e?m.ERROR:t.completed&&!e?m.DONE:t.completed&&e?n:t.editable&&e?m.EDIT:n},t.prototype._isCurrentStep=function(t){return this._selectedIndex===t},t.prototype._getFocusIndex=function(){return this._keyManager?this._keyManager.activeItemIndex:this._selectedIndex},t.prototype._updateSelectedItemIndex=function(t){var e=this._steps.toArray();this.selectionChange.emit({selectedIndex:t,previouslySelectedIndex:this._selectedIndex,selectedStep:e[t],previouslySelectedStep:e[this._selectedIndex]}),this._containsFocus()?this._keyManager.setActiveItem(t):this._keyManager.updateActiveItemIndex(t),this._selectedIndex=t,this._stateChanged()},t.prototype._onKeydown=function(t){var e=Object(s.hasModifierKey)(t),n=t.keyCode,r=this._keyManager;null==r.activeItemIndex||e||n!==s.SPACE&&n!==s.ENTER?n===s.HOME?(r.setFirstItemActive(),t.preventDefault()):n===s.END?(r.setLastItemActive(),t.preventDefault()):r.onKeydown(t):(this.selectedIndex=r.activeItemIndex,t.preventDefault())},t.prototype._anyControlsInvalidOrPending=function(t){var e=this._steps.toArray();return e[this._selectedIndex].interacted=!0,!!(this._linear&&t>=0)&&e.slice(0,t).some(function(t){var e=t.stepControl;return(e?e.invalid||e.pending||!t.interacted:!t.completed)&&!t.optional})},t.prototype._layoutDirection=function(){return this._dir&&"rtl"===this._dir.value?"rtl":"ltr"},t.prototype._containsFocus=function(){if(!this._document||!this._elementRef)return!1;var t=this._elementRef.nativeElement,e=this._document.activeElement;return t===e||t.contains(e)},t.decorators=[{type:r.Directive,args:[{selector:"[cdkStepper]",exportAs:"cdkStepper"}]}],t.ctorParameters=function(){return[{type:o.Directionality,decorators:[{type:r.Optional}]},{type:r.ChangeDetectorRef},{type:r.ElementRef},{type:void 0,decorators:[{type:r.Inject,args:[u.DOCUMENT]}]}]},t.propDecorators={_steps:[{type:r.ContentChildren,args:[_]}],_stepHeader:[{type:r.ContentChildren,args:[p]}],linear:[{type:r.Input}],selectedIndex:[{type:r.Input}],selected:[{type:r.Input}],selectionChange:[{type:r.Output}]},t}(),b=function(){function t(t){this._stepper=t,this.type="submit"}return t.decorators=[{type:r.Directive,args:[{selector:"button[cdkStepperNext]",host:{"(click)":"_stepper.next()","[type]":"type"}}]}],t.ctorParameters=function(){return[{type:v}]},t.propDecorators={type:[{type:r.Input}]},t}(),x=function(){function t(t){this._stepper=t,this.type="button"}return t.decorators=[{type:r.Directive,args:[{selector:"button[cdkStepperPrevious]",host:{"(click)":"_stepper.previous()","[type]":"type"}}]}],t.ctorParameters=function(){return[{type:v}]},t.propDecorators={type:[{type:r.Input}]},t}(),w=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[o.BidiModule,u.CommonModule],exports:[_,v,p,d,b,x],declarations:[_,v,p,d,b,x]}]}],t}()},"./node_modules/@angular/cdk/esm5/table.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"DataRowOutlet",function(){return L}),n.d(e,"HeaderRowOutlet",function(){return B}),n.d(e,"FooterRowOutlet",function(){return V}),n.d(e,"CDK_TABLE_TEMPLATE",function(){return z}),n.d(e,"CdkTable",function(){return H}),n.d(e,"CdkCellDef",function(){return f}),n.d(e,"CdkHeaderCellDef",function(){return h}),n.d(e,"CdkFooterCellDef",function(){return m}),n.d(e,"CdkColumnDefBase",function(){return g}),n.d(e,"_CdkColumnDefBase",function(){return y}),n.d(e,"CdkColumnDef",function(){return _}),n.d(e,"BaseCdkCell",function(){return v}),n.d(e,"CdkHeaderCell",function(){return b}),n.d(e,"CdkFooterCell",function(){return x}),n.d(e,"CdkCell",function(){return w}),n.d(e,"CDK_ROW_TEMPLATE",function(){return j}),n.d(e,"BaseRowDef",function(){return O}),n.d(e,"CdkHeaderRowDefBase",function(){return C}),n.d(e,"_CdkHeaderRowDefBase",function(){return E}),n.d(e,"CdkHeaderRowDef",function(){return S}),n.d(e,"CdkFooterRowDefBase",function(){return M}),n.d(e,"_CdkFooterRowDefBase",function(){return T}),n.d(e,"CdkFooterRowDef",function(){return k}),n.d(e,"CdkRowDef",function(){return A}),n.d(e,"CdkCellOutlet",function(){return D}),n.d(e,"CdkHeaderRow",function(){return I}),n.d(e,"CdkFooterRow",function(){return P}),n.d(e,"CdkRow",function(){return R}),n.d(e,"CdkTableModule",function(){return q}),n.d(e,"STICKY_DIRECTIONS",function(){return N}),n.d(e,"StickyStyler",function(){return F}),n.d(e,"mixinHasStickyInput",function(){return p});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/cdk/esm5/collections.es5.js");n.d(e,"DataSource",function(){return a.DataSource});var s=n("./node_modules/@angular/common/fesm5/common.js"),u=n("./node_modules/rxjs/_esm5/index.js"),c=n("./node_modules/rxjs/_esm5/operators/index.js"),l=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),d=n("./node_modules/@angular/cdk/esm5/platform.es5.js");function p(t){return function(t){function e(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];var r=t.apply(this,e)||this;return r._sticky=!1,r._hasStickyChanged=!1,r}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"sticky",{get:function(){return this._sticky},set:function(t){var e=this._sticky;this._sticky=Object(i.coerceBooleanProperty)(t),this._hasStickyChanged=e!==this._sticky},enumerable:!0,configurable:!0}),e.prototype.hasStickyChanged=function(){var t=this._hasStickyChanged;return this._hasStickyChanged=!1,t},e.prototype.resetStickyChanged=function(){this._hasStickyChanged=!1},e}(t)}var f=function(){function t(t){this.template=t}return t.decorators=[{type:o.Directive,args:[{selector:"[cdkCellDef]"}]}],t.ctorParameters=function(){return[{type:o.TemplateRef}]},t}(),h=function(){function t(t){this.template=t}return t.decorators=[{type:o.Directive,args:[{selector:"[cdkHeaderCellDef]"}]}],t.ctorParameters=function(){return[{type:o.TemplateRef}]},t}(),m=function(){function t(t){this.template=t}return t.decorators=[{type:o.Directive,args:[{selector:"[cdkFooterCellDef]"}]}],t.ctorParameters=function(){return[{type:o.TemplateRef}]},t}(),g=function(){return function(){}}(),y=p(g),_=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._stickyEnd=!1,e}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"name",{get:function(){return this._name},set:function(t){t&&(this._name=t,this.cssClassFriendlyName=t.replace(/[^a-z0-9_-]/gi,"-"))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"stickyEnd",{get:function(){return this._stickyEnd},set:function(t){var e=this._stickyEnd;this._stickyEnd=Object(i.coerceBooleanProperty)(t),this._hasStickyChanged=e!==this._stickyEnd},enumerable:!0,configurable:!0}),e.decorators=[{type:o.Directive,args:[{selector:"[cdkColumnDef]",inputs:["sticky"],providers:[{provide:"MAT_SORT_HEADER_COLUMN_DEF",useExisting:e}]}]}],e.propDecorators={name:[{type:o.Input,args:["cdkColumnDef"]}],stickyEnd:[{type:o.Input,args:["stickyEnd"]}],cell:[{type:o.ContentChild,args:[f]}],headerCell:[{type:o.ContentChild,args:[h]}],footerCell:[{type:o.ContentChild,args:[m]}]},e}(y),v=function(){return function(t,e){var n="cdk-column-"+t.cssClassFriendlyName;e.nativeElement.classList.add(n)}}(),b=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(r.__extends)(e,t),e.decorators=[{type:o.Directive,args:[{selector:"cdk-header-cell, th[cdk-header-cell]",host:{class:"cdk-header-cell",role:"columnheader"}}]}],e.ctorParameters=function(){return[{type:_},{type:o.ElementRef}]},e}(v),x=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(r.__extends)(e,t),e.decorators=[{type:o.Directive,args:[{selector:"cdk-footer-cell, td[cdk-footer-cell]",host:{class:"cdk-footer-cell",role:"gridcell"}}]}],e.ctorParameters=function(){return[{type:_},{type:o.ElementRef}]},e}(v),w=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(r.__extends)(e,t),e.decorators=[{type:o.Directive,args:[{selector:"cdk-cell, td[cdk-cell]",host:{class:"cdk-cell",role:"gridcell"}}]}],e.ctorParameters=function(){return[{type:_},{type:o.ElementRef}]},e}(v),j="<ng-container cdkCellOutlet></ng-container>",O=function(){function t(t,e){this.template=t,this._differs=e}return t.prototype.ngOnChanges=function(t){if(!this._columnsDiffer){var e=t.columns&&t.columns.currentValue||[];this._columnsDiffer=this._differs.find(e).create(),this._columnsDiffer.diff(e)}},t.prototype.getColumnsDiff=function(){return this._columnsDiffer.diff(this.columns)},t.prototype.extractCellTemplate=function(t){return this instanceof S?t.headerCell.template:this instanceof k?t.footerCell.template:t.cell.template},t}(),C=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e}(O),E=p(C),S=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(r.__extends)(e,t),e.prototype.ngOnChanges=function(e){t.prototype.ngOnChanges.call(this,e)},e.decorators=[{type:o.Directive,args:[{selector:"[cdkHeaderRowDef]",inputs:["columns: cdkHeaderRowDef","sticky: cdkHeaderRowDefSticky"]}]}],e.ctorParameters=function(){return[{type:o.TemplateRef},{type:o.IterableDiffers}]},e}(E),M=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e}(O),T=p(M),k=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(r.__extends)(e,t),e.prototype.ngOnChanges=function(e){t.prototype.ngOnChanges.call(this,e)},e.decorators=[{type:o.Directive,args:[{selector:"[cdkFooterRowDef]",inputs:["columns: cdkFooterRowDef","sticky: cdkFooterRowDefSticky"]}]}],e.ctorParameters=function(){return[{type:o.TemplateRef},{type:o.IterableDiffers}]},e}(T),A=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(r.__extends)(e,t),e.decorators=[{type:o.Directive,args:[{selector:"[cdkRowDef]",inputs:["columns: cdkRowDefColumns","when: cdkRowDefWhen"]}]}],e.ctorParameters=function(){return[{type:o.TemplateRef},{type:o.IterableDiffers}]},e}(O),D=function(){function t(e){this._viewContainer=e,t.mostRecentCellOutlet=this}return t.prototype.ngOnDestroy=function(){t.mostRecentCellOutlet===this&&(t.mostRecentCellOutlet=null)},t.mostRecentCellOutlet=null,t.decorators=[{type:o.Directive,args:[{selector:"[cdkCellOutlet]"}]}],t.ctorParameters=function(){return[{type:o.ViewContainerRef}]},t}(),I=function(){function t(){}return t.decorators=[{type:o.Component,args:[{selector:"cdk-header-row, tr[cdk-header-row]",template:j,host:{class:"cdk-header-row",role:"row"},changeDetection:o.ChangeDetectionStrategy.OnPush,encapsulation:o.ViewEncapsulation.None}]}],t}(),P=function(){function t(){}return t.decorators=[{type:o.Component,args:[{selector:"cdk-footer-row, tr[cdk-footer-row]",template:j,host:{class:"cdk-footer-row",role:"row"},changeDetection:o.ChangeDetectionStrategy.OnPush,encapsulation:o.ViewEncapsulation.None}]}],t}(),R=function(){function t(){}return t.decorators=[{type:o.Component,args:[{selector:"cdk-row, tr[cdk-row]",template:j,host:{class:"cdk-row",role:"row"},changeDetection:o.ChangeDetectionStrategy.OnPush,encapsulation:o.ViewEncapsulation.None}]}],t}();var N=["top","bottom","left","right"],F=function(){function t(t,e,n,r){void 0===r&&(r=!0),this.isNativeHtmlTable=t,this.stickCellCss=e,this.direction=n,this._isBrowser=r}return t.prototype.clearStickyPositioning=function(t,e){for(var n=0,r=t;n<r.length;n++){var i=r[n];if(i.nodeType===i.ELEMENT_NODE){this._removeStickyStyle(i,e);for(var o=0;o<i.children.length;o++){var a=i.children[o];this._removeStickyStyle(a,e)}}}},t.prototype.updateStickyColumns=function(t,e,n){var r=e.some(function(t){return t})||n.some(function(t){return t});if(t.length&&r&&this._isBrowser)for(var i=t[0],o=i.children.length,a=this._getCellWidths(i),s=this._getStickyStartColumnPositions(a,e),u=this._getStickyEndColumnPositions(a,n),c="rtl"===this.direction,l=0,d=t;l<d.length;l++)for(var p=d[l],f=0;f<o;f++){var h=p.children[f];e[f]&&this._addStickyStyle(h,c?"right":"left",s[f]),n[f]&&this._addStickyStyle(h,c?"left":"right",u[f])}},t.prototype.stickRows=function(t,e,n){if(this._isBrowser)for(var r="bottom"===n?t.reverse():t,i=0,o=0;o<r.length;o++)if(e[o]){var a=r[o];if(this.isNativeHtmlTable)for(var s=0;s<a.children.length;s++){var u=a.children[s];this._addStickyStyle(u,n,i)}else this._addStickyStyle(a,n,i);if(o===r.length-1)return;i+=a.getBoundingClientRect().height}},t.prototype.updateStickyFooterContainer=function(t,e){if(this.isNativeHtmlTable){var n=t.querySelector("tfoot");e.some(function(t){return!t})?this._removeStickyStyle(n,["bottom"]):this._addStickyStyle(n,"bottom",0)}},t.prototype._removeStickyStyle=function(t,e){for(var n=0,r=e;n<r.length;n++){var i=r[n];t.style[i]=""}t.style.zIndex=this._getCalculatedZIndex(t),N.some(function(e){return!!t.style[e]})||(t.style.position="",t.classList.remove(this.stickCellCss))},t.prototype._addStickyStyle=function(t,e,n){t.classList.add(this.stickCellCss),t.style[e]=n+"px",t.style.cssText+="position: -webkit-sticky; position: sticky; ",t.style.zIndex=this._getCalculatedZIndex(t)},t.prototype._getCalculatedZIndex=function(t){for(var e={top:100,bottom:10,left:1,right:1},n=0,r=0,i=N;r<i.length;r++){var o=i[r];t.style[o]&&(n+=e[o])}return n?""+n:""},t.prototype._getCellWidths=function(t){for(var e=[],n=t.children,r=0;r<n.length;r++){var i=n[r];e.push(i.getBoundingClientRect().width)}return e},t.prototype._getStickyStartColumnPositions=function(t,e){for(var n=[],r=0,i=0;i<t.length;i++)e[i]&&(n[i]=r,r+=t[i]);return n},t.prototype._getStickyEndColumnPositions=function(t,e){for(var n=[],r=0,i=t.length;i>0;i--)e[i]&&(n[i]=r,r+=t[i]);return n},t}(),L=function(){function t(t,e){this.viewContainer=t,this.elementRef=e}return t.decorators=[{type:o.Directive,args:[{selector:"[rowOutlet]"}]}],t.ctorParameters=function(){return[{type:o.ViewContainerRef},{type:o.ElementRef}]},t}(),B=function(){function t(t,e){this.viewContainer=t,this.elementRef=e}return t.decorators=[{type:o.Directive,args:[{selector:"[headerRowOutlet]"}]}],t.ctorParameters=function(){return[{type:o.ViewContainerRef},{type:o.ElementRef}]},t}(),V=function(){function t(t,e){this.viewContainer=t,this.elementRef=e}return t.decorators=[{type:o.Directive,args:[{selector:"[footerRowOutlet]"}]}],t.ctorParameters=function(){return[{type:o.ViewContainerRef},{type:o.ElementRef}]},t}(),z="\n  <ng-container headerRowOutlet></ng-container>\n  <ng-container rowOutlet></ng-container>\n  <ng-container footerRowOutlet></ng-container>",H=(function(t){function e(){return null!==t&&t.apply(this,arguments)||this}Object(r.__extends)(e,t)}(o.EmbeddedViewRef),function(){function t(t,e,n,r,i,o,a){this._differs=t,this._changeDetectorRef=e,this._elementRef=n,this._dir=i,this._platform=a,this._onDestroy=new u.Subject,this._columnDefsByName=new Map,this._customColumnDefs=new Set,this._customRowDefs=new Set,this._customHeaderRowDefs=new Set,this._customFooterRowDefs=new Set,this._headerRowDefChanged=!0,this._footerRowDefChanged=!0,this._cachedRenderRowsMap=new Map,this.stickyCssClass="cdk-table-sticky",this._multiTemplateDataRows=!1,this.viewChange=new u.BehaviorSubject({start:0,end:Number.MAX_VALUE}),r||this._elementRef.nativeElement.setAttribute("role","grid"),this._document=o,this._isNativeHtmlTable="TABLE"===this._elementRef.nativeElement.nodeName}return Object.defineProperty(t.prototype,"trackBy",{get:function(){return this._trackByFn},set:function(t){Object(o.isDevMode)()&&null!=t&&"function"!=typeof t&&console&&console.warn&&console.warn("trackBy must be a function, but received "+JSON.stringify(t)+"."),this._trackByFn=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"dataSource",{get:function(){return this._dataSource},set:function(t){this._dataSource!==t&&this._switchDataSource(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"multiTemplateDataRows",{get:function(){return this._multiTemplateDataRows},set:function(t){this._multiTemplateDataRows=Object(i.coerceBooleanProperty)(t),this._rowOutlet.viewContainer.length&&this._forceRenderDataRows()},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){var t=this;this._setupStickyStyler(),this._isNativeHtmlTable&&this._applyNativeTableSections(),this._dataDiffer=this._differs.find([]).create(function(e,n){return t.trackBy?t.trackBy(n.dataIndex,n.data):n})},t.prototype.ngAfterContentChecked=function(){if(this._cacheRowDefs(),this._cacheColumnDefs(),!this._headerRowDefs.length&&!this._footerRowDefs.length&&!this._rowDefs.length)throw Error("Missing definitions for header, footer, and row; cannot determine which columns should be rendered.");this._renderUpdatedColumns(),this._headerRowDefChanged&&(this._forceRenderHeaderRows(),this._headerRowDefChanged=!1),this._footerRowDefChanged&&(this._forceRenderFooterRows(),this._footerRowDefChanged=!1),this.dataSource&&this._rowDefs.length>0&&!this._renderChangeSubscription&&this._observeRenderChanges(),this._checkStickyStates()},t.prototype.ngOnDestroy=function(){this._rowOutlet.viewContainer.clear(),this._headerRowOutlet.viewContainer.clear(),this._footerRowOutlet.viewContainer.clear(),this._cachedRenderRowsMap.clear(),this._onDestroy.next(),this._onDestroy.complete(),this.dataSource instanceof a.DataSource&&this.dataSource.disconnect(this)},t.prototype.renderRows=function(){var t=this;this._renderRows=this._getAllRenderRows();var e=this._dataDiffer.diff(this._renderRows);if(e){var n=this._rowOutlet.viewContainer;e.forEachOperation(function(e,r,i){if(null==e.previousIndex)t._insertRow(e.item,i);else if(null==i)n.remove(r);else{var o=n.get(r);n.move(o,i)}}),this._updateRowIndexContext(),e.forEachIdentityChange(function(t){n.get(t.currentIndex).context.$implicit=t.item.data}),this.updateStickyColumnStyles()}},t.prototype.setHeaderRowDef=function(t){this._customHeaderRowDefs=new Set([t]),this._headerRowDefChanged=!0},t.prototype.setFooterRowDef=function(t){this._customFooterRowDefs=new Set([t]),this._footerRowDefChanged=!0},t.prototype.addColumnDef=function(t){this._customColumnDefs.add(t)},t.prototype.removeColumnDef=function(t){this._customColumnDefs.delete(t)},t.prototype.addRowDef=function(t){this._customRowDefs.add(t)},t.prototype.removeRowDef=function(t){this._customRowDefs.delete(t)},t.prototype.addHeaderRowDef=function(t){this._customHeaderRowDefs.add(t),this._headerRowDefChanged=!0},t.prototype.removeHeaderRowDef=function(t){this._customHeaderRowDefs.delete(t),this._headerRowDefChanged=!0},t.prototype.addFooterRowDef=function(t){this._customFooterRowDefs.add(t),this._footerRowDefChanged=!0},t.prototype.removeFooterRowDef=function(t){this._customFooterRowDefs.delete(t),this._footerRowDefChanged=!0},t.prototype.updateStickyHeaderRowStyles=function(){var t=this._getRenderedRows(this._headerRowOutlet);this._stickyStyler.clearStickyPositioning(t,["top"]);var e=this._headerRowDefs.map(function(t){return t.sticky});this._stickyStyler.stickRows(t,e,"top"),this._headerRowDefs.forEach(function(t){return t.resetStickyChanged()})},t.prototype.updateStickyFooterRowStyles=function(){var t=this._getRenderedRows(this._footerRowOutlet);this._stickyStyler.clearStickyPositioning(t,["bottom"]);var e=this._footerRowDefs.map(function(t){return t.sticky});this._stickyStyler.stickRows(t,e,"bottom"),this._stickyStyler.updateStickyFooterContainer(this._elementRef.nativeElement,e),this._footerRowDefs.forEach(function(t){return t.resetStickyChanged()})},t.prototype.updateStickyColumnStyles=function(){var t=this,e=this._getRenderedRows(this._headerRowOutlet),n=this._getRenderedRows(this._rowOutlet),r=this._getRenderedRows(this._footerRowOutlet);this._stickyStyler.clearStickyPositioning(e.concat(n,r),["left","right"]),e.forEach(function(e,n){t._addStickyColumnStyles([e],t._headerRowDefs[n])}),this._rowDefs.forEach(function(e){for(var r=[],i=0;i<n.length;i++)t._renderRows[i].rowDef===e&&r.push(n[i]);t._addStickyColumnStyles(r,e)}),r.forEach(function(e,n){t._addStickyColumnStyles([e],t._footerRowDefs[n])}),Array.from(this._columnDefsByName.values()).forEach(function(t){return t.resetStickyChanged()})},t.prototype._getAllRenderRows=function(){var t=[],e=this._cachedRenderRowsMap;this._cachedRenderRowsMap=new Map;for(var n=0;n<this._data.length;n++){var r=this._data[n],i=this._getRenderRowsForData(r,n,e.get(r));this._cachedRenderRowsMap.has(r)||this._cachedRenderRowsMap.set(r,new WeakMap);for(var o=0;o<i.length;o++){var a=i[o],s=this._cachedRenderRowsMap.get(a.data);s.has(a.rowDef)?s.get(a.rowDef).push(a):s.set(a.rowDef,[a]),t.push(a)}}return t},t.prototype._getRenderRowsForData=function(t,e,n){return this._getRowDefs(t,e).map(function(r){var i=n&&n.has(r)?n.get(r):[];if(i.length){var o=i.shift();return o.dataIndex=e,o}return{data:t,rowDef:r,dataIndex:e}})},t.prototype._cacheColumnDefs=function(){var t=this;this._columnDefsByName.clear(),U(this._contentColumnDefs,this._customColumnDefs).forEach(function(e){if(t._columnDefsByName.has(e.name))throw n=e.name,Error('Duplicate column definition name provided: "'+n+'".');var n;t._columnDefsByName.set(e.name,e)})},t.prototype._cacheRowDefs=function(){this._headerRowDefs=U(this._contentHeaderRowDefs,this._customHeaderRowDefs),this._footerRowDefs=U(this._contentFooterRowDefs,this._customFooterRowDefs),this._rowDefs=U(this._contentRowDefs,this._customRowDefs);var t=this._rowDefs.filter(function(t){return!t.when});if(!this.multiTemplateDataRows&&t.length>1)throw Error("There can only be one default row without a when predicate function.");this._defaultRowDef=t[0]},t.prototype._renderUpdatedColumns=function(){var t=function(t,e){return t||!!e.getColumnsDiff()};this._rowDefs.reduce(t,!1)&&this._forceRenderDataRows(),this._headerRowDefs.reduce(t,!1)&&this._forceRenderHeaderRows(),this._footerRowDefs.reduce(t,!1)&&this._forceRenderFooterRows()},t.prototype._switchDataSource=function(t){this._data=[],this.dataSource instanceof a.DataSource&&this.dataSource.disconnect(this),this._renderChangeSubscription&&(this._renderChangeSubscription.unsubscribe(),this._renderChangeSubscription=null),t||(this._dataDiffer&&this._dataDiffer.diff([]),this._rowOutlet.viewContainer.clear()),this._dataSource=t},t.prototype._observeRenderChanges=function(){var t=this;if(this.dataSource){var e;if(this.dataSource.connect instanceof Function?e=this.dataSource.connect(this):this.dataSource instanceof u.Observable?e=this.dataSource:Array.isArray(this.dataSource)&&(e=Object(u.of)(this.dataSource)),void 0===e)throw Error("Provided data source did not match an array, Observable, or DataSource");this._renderChangeSubscription=e.pipe(Object(c.takeUntil)(this._onDestroy)).subscribe(function(e){t._data=e||[],t.renderRows()})}},t.prototype._forceRenderHeaderRows=function(){var t=this;this._headerRowOutlet.viewContainer.length>0&&this._headerRowOutlet.viewContainer.clear(),this._headerRowDefs.forEach(function(e,n){return t._renderRow(t._headerRowOutlet,e,n)}),this.updateStickyHeaderRowStyles(),this.updateStickyColumnStyles()},t.prototype._forceRenderFooterRows=function(){var t=this;this._footerRowOutlet.viewContainer.length>0&&this._footerRowOutlet.viewContainer.clear(),this._footerRowDefs.forEach(function(e,n){return t._renderRow(t._footerRowOutlet,e,n)}),this.updateStickyFooterRowStyles(),this.updateStickyColumnStyles()},t.prototype._addStickyColumnStyles=function(t,e){var n=this,r=Array.from(e.columns||[]).map(function(t){return n._columnDefsByName.get(t)}),i=r.map(function(t){return t.sticky}),o=r.map(function(t){return t.stickyEnd});this._stickyStyler.updateStickyColumns(t,i,o)},t.prototype._getRenderedRows=function(t){for(var e=[],n=0;n<t.viewContainer.length;n++){var r=t.viewContainer.get(n);e.push(r.rootNodes[0])}return e},t.prototype._getRowDefs=function(t,e){if(1==this._rowDefs.length)return[this._rowDefs[0]];var n,r=[];if(this.multiTemplateDataRows)r=this._rowDefs.filter(function(n){return!n.when||n.when(e,t)});else{var i=this._rowDefs.find(function(n){return n.when&&n.when(e,t)})||this._defaultRowDef;i&&r.push(i)}if(!r.length)throw n=t,Error("Could not find a matching row definition for theprovided row data: "+JSON.stringify(n));return r},t.prototype._insertRow=function(t,e){var n=t.rowDef,r={$implicit:t.data};this._renderRow(this._rowOutlet,n,e,r)},t.prototype._renderRow=function(t,e,n,r){void 0===r&&(r={}),t.viewContainer.createEmbeddedView(e.template,r,n);for(var i=0,o=this._getCellTemplates(e);i<o.length;i++){var a=o[i];D.mostRecentCellOutlet&&D.mostRecentCellOutlet._viewContainer.createEmbeddedView(a,r)}this._changeDetectorRef.markForCheck()},t.prototype._updateRowIndexContext=function(){for(var t=this._rowOutlet.viewContainer,e=0,n=t.length;e<n;e++){var r=t.get(e).context;r.count=n,r.first=0===e,r.last=e===n-1,r.even=e%2==0,r.odd=!r.even,this.multiTemplateDataRows?(r.dataIndex=this._renderRows[e].dataIndex,r.renderIndex=e):r.index=this._renderRows[e].dataIndex}},t.prototype._getCellTemplates=function(t){var e=this;return t&&t.columns?Array.from(t.columns,function(n){var r=e._columnDefsByName.get(n);if(!r)throw Error('Could not find column with id "'+n+'".');return t.extractCellTemplate(r)}):[]},t.prototype._applyNativeTableSections=function(){for(var t=0,e=[{tag:"thead",outlet:this._headerRowOutlet},{tag:"tbody",outlet:this._rowOutlet},{tag:"tfoot",outlet:this._footerRowOutlet}];t<e.length;t++){var n=e[t],r=(this._document||document).createElement(n.tag);r.appendChild(n.outlet.elementRef.nativeElement),this._elementRef.nativeElement.appendChild(r)}},t.prototype._forceRenderDataRows=function(){this._dataDiffer.diff([]),this._rowOutlet.viewContainer.clear(),this.renderRows(),this.updateStickyColumnStyles()},t.prototype._checkStickyStates=function(){var t=function(t,e){return t||e.hasStickyChanged()};this._headerRowDefs.reduce(t,!1)&&this.updateStickyHeaderRowStyles(),this._footerRowDefs.reduce(t,!1)&&this.updateStickyFooterRowStyles(),Array.from(this._columnDefsByName.values()).reduce(t,!1)&&this.updateStickyColumnStyles()},t.prototype._setupStickyStyler=function(){var t=this,e=this._dir?this._dir.value:"ltr";this._stickyStyler=new F(this._isNativeHtmlTable,this.stickyCssClass,e,!this._platform||this._platform.isBrowser),(this._dir?this._dir.change:Object(u.of)()).pipe(Object(c.takeUntil)(this._onDestroy)).subscribe(function(e){t._stickyStyler.direction=e,t.updateStickyColumnStyles()})},t.decorators=[{type:o.Component,args:[{selector:"cdk-table, table[cdk-table]",exportAs:"cdkTable",template:z,host:{class:"cdk-table"},encapsulation:o.ViewEncapsulation.None,changeDetection:o.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:o.IterableDiffers},{type:o.ChangeDetectorRef},{type:o.ElementRef},{type:String,decorators:[{type:o.Attribute,args:["role"]}]},{type:l.Directionality,decorators:[{type:o.Optional}]},{type:void 0,decorators:[{type:o.Inject,args:[s.DOCUMENT]}]},{type:d.Platform}]},t.propDecorators={trackBy:[{type:o.Input}],dataSource:[{type:o.Input}],multiTemplateDataRows:[{type:o.Input}],_rowOutlet:[{type:o.ViewChild,args:[L]}],_headerRowOutlet:[{type:o.ViewChild,args:[B]}],_footerRowOutlet:[{type:o.ViewChild,args:[V]}],_contentColumnDefs:[{type:o.ContentChildren,args:[_]}],_contentRowDefs:[{type:o.ContentChildren,args:[A]}],_contentHeaderRowDefs:[{type:o.ContentChildren,args:[S]}],_contentFooterRowDefs:[{type:o.ContentChildren,args:[k]}]},t}());function U(t,e){return t.toArray().concat(Array.from(e))}var W=[H,A,f,D,h,m,_,w,R,b,x,I,S,P,k,L,B,V],q=function(){function t(){}return t.decorators=[{type:o.NgModule,args:[{imports:[s.CommonModule],exports:W,declarations:W}]}],t}()},"./node_modules/@angular/cdk/esm5/text-field.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"AutofillMonitor",function(){return c}),n.d(e,"CdkAutofill",function(){return l}),n.d(e,"CdkTextareaAutosize",function(){return d}),n.d(e,"TextFieldModule",function(){return p});var r=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),a=n("./node_modules/rxjs/_esm5/index.js"),s=n("./node_modules/rxjs/_esm5/operators/index.js"),u=Object(r.normalizePassiveListenerOptions)({passive:!0}),c=function(){function t(t,e){this._platform=t,this._ngZone=e,this._monitoredElements=new Map}return t.prototype.monitor=function(t){var e=this;if(!this._platform.isBrowser)return a.EMPTY;var n=Object(o.coerceElement)(t),r=this._monitoredElements.get(n);if(r)return r.subject.asObservable();var i=new a.Subject,s="cdk-text-field-autofilled",c=function(t){"cdk-text-field-autofill-start"!==t.animationName||n.classList.contains(s)?"cdk-text-field-autofill-end"===t.animationName&&n.classList.contains(s)&&(n.classList.remove(s),e._ngZone.run(function(){return i.next({target:t.target,isAutofilled:!1})})):(n.classList.add(s),e._ngZone.run(function(){return i.next({target:t.target,isAutofilled:!0})}))};return this._ngZone.runOutsideAngular(function(){n.addEventListener("animationstart",c,u),n.classList.add("cdk-text-field-autofill-monitored")}),this._monitoredElements.set(n,{subject:i,unlisten:function(){n.removeEventListener("animationstart",c,u)}}),i.asObservable()},t.prototype.stopMonitoring=function(t){var e=Object(o.coerceElement)(t),n=this._monitoredElements.get(e);n&&(n.unlisten(),n.subject.complete(),e.classList.remove("cdk-text-field-autofill-monitored"),e.classList.remove("cdk-text-field-autofilled"),this._monitoredElements.delete(e))},t.prototype.ngOnDestroy=function(){var t=this;this._monitoredElements.forEach(function(e,n){return t.stopMonitoring(n)})},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:r.Platform},{type:i.NgZone}]},t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t(Object(i.inject)(r.Platform),Object(i.inject)(i.NgZone))},token:t,providedIn:"root"}),t}(),l=function(){function t(t,e){this._elementRef=t,this._autofillMonitor=e,this.cdkAutofill=new i.EventEmitter}return t.prototype.ngOnInit=function(){var t=this;this._autofillMonitor.monitor(this._elementRef).subscribe(function(e){return t.cdkAutofill.emit(e)})},t.prototype.ngOnDestroy=function(){this._autofillMonitor.stopMonitoring(this._elementRef)},t.decorators=[{type:i.Directive,args:[{selector:"[cdkAutofill]"}]}],t.ctorParameters=function(){return[{type:i.ElementRef},{type:c}]},t.propDecorators={cdkAutofill:[{type:i.Output}]},t}(),d=function(){function t(t,e,n){this._elementRef=t,this._platform=e,this._ngZone=n,this._destroyed=new a.Subject,this._enabled=!0,this._previousMinRows=-1,this._textareaElement=this._elementRef.nativeElement}return Object.defineProperty(t.prototype,"minRows",{get:function(){return this._minRows},set:function(t){this._minRows=t,this._setMinHeight()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"maxRows",{get:function(){return this._maxRows},set:function(t){this._maxRows=t,this._setMaxHeight()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"enabled",{get:function(){return this._enabled},set:function(t){t=Object(o.coerceBooleanProperty)(t),this._enabled!==t&&((this._enabled=t)?this.resizeToFitContent(!0):this.reset())},enumerable:!0,configurable:!0}),t.prototype._setMinHeight=function(){var t=this.minRows&&this._cachedLineHeight?this.minRows*this._cachedLineHeight+"px":null;t&&(this._textareaElement.style.minHeight=t)},t.prototype._setMaxHeight=function(){var t=this.maxRows&&this._cachedLineHeight?this.maxRows*this._cachedLineHeight+"px":null;t&&(this._textareaElement.style.maxHeight=t)},t.prototype.ngAfterViewInit=function(){var t=this;this._platform.isBrowser&&(this._initialHeight=this._textareaElement.style.height,this.resizeToFitContent(),this._ngZone.runOutsideAngular(function(){Object(a.fromEvent)(window,"resize").pipe(Object(s.auditTime)(16),Object(s.takeUntil)(t._destroyed)).subscribe(function(){return t.resizeToFitContent(!0)})}))},t.prototype.ngOnDestroy=function(){this._destroyed.next(),this._destroyed.complete()},t.prototype._cacheTextareaLineHeight=function(){if(!this._cachedLineHeight){var t=this._textareaElement.cloneNode(!1);t.rows=1,t.style.position="absolute",t.style.visibility="hidden",t.style.border="none",t.style.padding="0",t.style.height="",t.style.minHeight="",t.style.maxHeight="",t.style.overflow="hidden",this._textareaElement.parentNode.appendChild(t),this._cachedLineHeight=t.clientHeight,this._textareaElement.parentNode.removeChild(t),this._setMinHeight(),this._setMaxHeight()}},t.prototype.ngDoCheck=function(){this._platform.isBrowser&&this.resizeToFitContent()},t.prototype.resizeToFitContent=function(t){var e=this;if(void 0===t&&(t=!1),this._enabled&&(this._cacheTextareaLineHeight(),this._cachedLineHeight)){var n=this._elementRef.nativeElement,r=n.value;if(t||this._minRows!==this._previousMinRows||r!==this._previousValue){var i=n.placeholder;n.classList.add("cdk-textarea-autosize-measuring"),n.placeholder="";var o=n.scrollHeight-4;n.style.height=o+"px",n.classList.remove("cdk-textarea-autosize-measuring"),n.placeholder=i,this._ngZone.runOutsideAngular(function(){"undefined"!=typeof requestAnimationFrame?requestAnimationFrame(function(){return e._scrollToCaretPosition(n)}):setTimeout(function(){return e._scrollToCaretPosition(n)})}),this._previousValue=r,this._previousMinRows=this._minRows}}},t.prototype.reset=function(){void 0!==this._initialHeight&&(this._textareaElement.style.height=this._initialHeight)},t.prototype._noopInputHandler=function(){},t.prototype._scrollToCaretPosition=function(t){var e=t.selectionStart,n=t.selectionEnd;this._destroyed.isStopped||document.activeElement!==t||t.setSelectionRange(e,n)},t.decorators=[{type:i.Directive,args:[{selector:"textarea[cdkTextareaAutosize]",exportAs:"cdkTextareaAutosize",host:{class:"cdk-textarea-autosize",rows:"1","(input)":"_noopInputHandler()"}}]}],t.ctorParameters=function(){return[{type:i.ElementRef},{type:r.Platform},{type:i.NgZone}]},t.propDecorators={minRows:[{type:i.Input,args:["cdkAutosizeMinRows"]}],maxRows:[{type:i.Input,args:["cdkAutosizeMaxRows"]}],enabled:[{type:i.Input,args:["cdkTextareaAutosize"]}]},t}(),p=function(){function t(){}return t.decorators=[{type:i.NgModule,args:[{declarations:[l,d],imports:[r.PlatformModule],exports:[l,d]}]}],t}()},"./node_modules/@angular/cdk/esm5/tree.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"BaseTreeControl",function(){return p}),n.d(e,"FlatTreeControl",function(){return f}),n.d(e,"NestedTreeControl",function(){return h}),n.d(e,"CdkNestedTreeNode",function(){return C}),n.d(e,"CdkTreeNodeOutletContext",function(){return m}),n.d(e,"CdkTreeNodeDef",function(){return g}),n.d(e,"CdkTreeNodePadding",function(){return S}),n.d(e,"CdkTreeNodeOutlet",function(){return y}),n.d(e,"CdkTree",function(){return j}),n.d(e,"CdkTreeNode",function(){return O}),n.d(e,"getTreeNoValidDataSourceError",function(){return _}),n.d(e,"getTreeMultipleDefaultNodeDefsError",function(){return v}),n.d(e,"getTreeMissingMatchingNodeDefError",function(){return b}),n.d(e,"getTreeControlMissingError",function(){return x}),n.d(e,"getTreeControlFunctionsMissingError",function(){return w}),n.d(e,"CdkTreeModule",function(){return k}),n.d(e,"CdkTreeNodeToggle",function(){return M});var r=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/rxjs/_esm5/index.js"),a=n("./node_modules/rxjs/_esm5/operators/index.js"),s=n("./node_modules/@angular/core/fesm5/core.js"),u=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),c=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),l=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),d=n("./node_modules/@angular/common/fesm5/common.js"),p=function(){function t(){this.expansionModel=new r.SelectionModel(!0)}return t.prototype.toggle=function(t){this.expansionModel.toggle(t)},t.prototype.expand=function(t){this.expansionModel.select(t)},t.prototype.collapse=function(t){this.expansionModel.deselect(t)},t.prototype.isExpanded=function(t){return this.expansionModel.isSelected(t)},t.prototype.toggleDescendants=function(t){this.expansionModel.isSelected(t)?this.collapseDescendants(t):this.expandDescendants(t)},t.prototype.collapseAll=function(){this.expansionModel.clear()},t.prototype.expandDescendants=function(t){var e,n=[t];n.push.apply(n,this.getDescendants(t)),(e=this.expansionModel).select.apply(e,n)},t.prototype.collapseDescendants=function(t){var e,n=[t];n.push.apply(n,this.getDescendants(t)),(e=this.expansionModel).deselect.apply(e,n)},t}(),f=function(t){function e(e,n){var r=t.call(this)||this;return r.getLevel=e,r.isExpandable=n,r}return Object(i.__extends)(e,t),e.prototype.getDescendants=function(t){for(var e=[],n=this.dataNodes.indexOf(t)+1;n<this.dataNodes.length&&this.getLevel(t)<this.getLevel(this.dataNodes[n]);n++)e.push(this.dataNodes[n]);return e},e.prototype.expandAll=function(){var t;(t=this.expansionModel).select.apply(t,this.dataNodes)},e}(p),h=function(t){function e(e){var n=t.call(this)||this;return n.getChildren=e,n}return Object(i.__extends)(e,t),e.prototype.expandAll=function(){var t,e=this;this.expansionModel.clear();var n=this.dataNodes.reduce(function(t,n){return t.concat(e.getDescendants(n),[n])},[]);(t=this.expansionModel).select.apply(t,n)},e.prototype.getDescendants=function(t){var e=[];return this._getDescendants(e,t),e.splice(1)},e.prototype._getDescendants=function(t,e){var n=this;t.push(e);var r=this.getChildren(e);Array.isArray(r)?r.forEach(function(e){return n._getDescendants(t,e)}):r instanceof o.Observable&&r.pipe(Object(a.take)(1),Object(a.filter)(Boolean)).subscribe(function(e){e.forEach(function(e){return n._getDescendants(t,e)})})},e}(p),m=function(){return function(t){this.$implicit=t}}(),g=function(){function t(t){this.template=t}return t.decorators=[{type:s.Directive,args:[{selector:"[cdkTreeNodeDef]",inputs:["when: cdkTreeNodeDefWhen"]}]}],t.ctorParameters=function(){return[{type:s.TemplateRef}]},t}(),y=function(){function t(t){this.viewContainer=t}return t.decorators=[{type:s.Directive,args:[{selector:"[cdkTreeNodeOutlet]"}]}],t.ctorParameters=function(){return[{type:s.ViewContainerRef}]},t}();function _(){return Error("A valid data source must be provided.")}function v(){return Error("There can only be one default row without a when predicate function.")}function b(){return Error("Could not find a matching node definition for the provided node data.")}function x(){return Error("Could not find a tree control for the tree.")}function w(){return Error("Could not find functions for nested/flat tree in tree control.")}var j=function(){function t(t,e){this._differs=t,this._changeDetectorRef=e,this._onDestroy=new o.Subject,this._levels=new Map,this.viewChange=new o.BehaviorSubject({start:0,end:Number.MAX_VALUE})}return Object.defineProperty(t.prototype,"dataSource",{get:function(){return this._dataSource},set:function(t){this._dataSource!==t&&this._switchDataSource(t)},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){if(this._dataDiffer=this._differs.find([]).create(this.trackBy),!this.treeControl)throw x()},t.prototype.ngOnDestroy=function(){this._nodeOutlet.viewContainer.clear(),this._onDestroy.next(),this._onDestroy.complete(),this._dataSource&&"function"==typeof this._dataSource.disconnect&&this.dataSource.disconnect(this),this._dataSubscription&&(this._dataSubscription.unsubscribe(),this._dataSubscription=null)},t.prototype.ngAfterContentChecked=function(){var t=this._nodeDefs.filter(function(t){return!t.when});if(t.length>1)throw v();this._defaultNodeDef=t[0],this.dataSource&&this._nodeDefs&&!this._dataSubscription&&this._observeRenderChanges()},t.prototype._switchDataSource=function(t){this._dataSource&&"function"==typeof this._dataSource.disconnect&&this.dataSource.disconnect(this),this._dataSubscription&&(this._dataSubscription.unsubscribe(),this._dataSubscription=null),t||this._nodeOutlet.viewContainer.clear(),this._dataSource=t,this._nodeDefs&&this._observeRenderChanges()},t.prototype._observeRenderChanges=function(){var t,e=this;if("function"==typeof this._dataSource.connect?t=this._dataSource.connect(this):this._dataSource instanceof o.Observable?t=this._dataSource:Array.isArray(this._dataSource)&&(t=Object(o.of)(this._dataSource)),!t)throw _();this._dataSubscription=t.pipe(Object(a.takeUntil)(this._onDestroy)).subscribe(function(t){return e.renderNodeChanges(t)})},t.prototype.renderNodeChanges=function(t,e,n,r){var i=this;void 0===e&&(e=this._dataDiffer),void 0===n&&(n=this._nodeOutlet.viewContainer);var o=e.diff(t);o&&(o.forEachOperation(function(e,o,a){if(null==e.previousIndex)i.insertNode(t[a],a,n,r);else if(null==a)n.remove(o),i._levels.delete(e.item);else{var s=n.get(o);n.move(s,a)}}),this._changeDetectorRef.detectChanges())},t.prototype._getNodeDef=function(t,e){if(1===this._nodeDefs.length)return this._nodeDefs.first;var n=this._nodeDefs.find(function(n){return n.when&&n.when(e,t)})||this._defaultNodeDef;if(!n)throw b();return n},t.prototype.insertNode=function(t,e,n,r){var i=this._getNodeDef(t,e),o=new m(t);this.treeControl.getLevel?o.level=this.treeControl.getLevel(t):void 0!==r&&this._levels.has(r)?o.level=this._levels.get(r)+1:o.level=0,this._levels.set(t,o.level),(n||this._nodeOutlet.viewContainer).createEmbeddedView(i.template,o,e),O.mostRecentTreeNode&&(O.mostRecentTreeNode.data=t)},t.decorators=[{type:s.Component,args:[{selector:"cdk-tree",exportAs:"cdkTree",template:"<ng-container cdkTreeNodeOutlet></ng-container>",host:{class:"cdk-tree",role:"tree"},encapsulation:s.ViewEncapsulation.None,changeDetection:s.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:s.IterableDiffers},{type:s.ChangeDetectorRef}]},t.propDecorators={dataSource:[{type:s.Input}],treeControl:[{type:s.Input}],trackBy:[{type:s.Input}],_nodeOutlet:[{type:s.ViewChild,args:[y]}],_nodeDefs:[{type:s.ContentChildren,args:[g]}]},t}(),O=function(){function t(e,n){this._elementRef=e,this._tree=n,this._destroyed=new o.Subject,this.role="treeitem",t.mostRecentTreeNode=this}return Object.defineProperty(t.prototype,"data",{get:function(){return this._data},set:function(t){this._data=t,this._setRoleFromData()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"isExpanded",{get:function(){return this._tree.treeControl.isExpanded(this._data)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"level",{get:function(){return this._tree.treeControl.getLevel?this._tree.treeControl.getLevel(this._data):0},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){t.mostRecentTreeNode===this&&(t.mostRecentTreeNode=null),this._destroyed.next(),this._destroyed.complete()},t.prototype.focus=function(){this._elementRef.nativeElement.focus()},t.prototype._setRoleFromData=function(){var t=this;if(this._tree.treeControl.isExpandable)this.role=this._tree.treeControl.isExpandable(this._data)?"group":"treeitem";else{if(!this._tree.treeControl.getChildren)throw w();var e=this._tree.treeControl.getChildren(this._data);Array.isArray(e)?this._setRoleFromChildren(e):e instanceof o.Observable&&e.pipe(Object(a.takeUntil)(this._destroyed)).subscribe(function(e){return t._setRoleFromChildren(e)})}},t.prototype._setRoleFromChildren=function(t){this.role=t&&t.length?"group":"treeitem"},t.mostRecentTreeNode=null,t.decorators=[{type:s.Directive,args:[{selector:"cdk-tree-node",exportAs:"cdkTreeNode",host:{"[attr.aria-expanded]":"isExpanded","[attr.aria-level]":'role === "treeitem" ? level : null',"[attr.role]":"role",class:"cdk-tree-node"}}]}],t.ctorParameters=function(){return[{type:s.ElementRef},{type:j}]},t.propDecorators={role:[{type:s.Input}]},t}(),C=function(t){function e(e,n,r){var i=t.call(this,e,n)||this;return i._elementRef=e,i._tree=n,i._differs=r,i}return Object(i.__extends)(e,t),e.prototype.ngAfterContentInit=function(){var t=this;if(this._dataDiffer=this._differs.find([]).create(this._tree.trackBy),!this._tree.treeControl.getChildren)throw w();var e=this._tree.treeControl.getChildren(this.data);Array.isArray(e)?this.updateChildrenNodes(e):e instanceof o.Observable&&e.pipe(Object(a.takeUntil)(this._destroyed)).subscribe(function(e){return t.updateChildrenNodes(e)}),this.nodeOutlet.changes.pipe(Object(a.takeUntil)(this._destroyed)).subscribe(function(){return t.updateChildrenNodes()})},e.prototype.ngOnDestroy=function(){this._clear(),t.prototype.ngOnDestroy.call(this)},e.prototype.updateChildrenNodes=function(t){if(t&&(this._children=t),this.nodeOutlet.length&&this._children){var e=this.nodeOutlet.first.viewContainer;this._tree.renderNodeChanges(this._children,this._dataDiffer,e,this._data)}else this._dataDiffer.diff([])},e.prototype._clear=function(){this.nodeOutlet&&this.nodeOutlet.first&&(this.nodeOutlet.first.viewContainer.clear(),this._dataDiffer.diff([]))},e.decorators=[{type:s.Directive,args:[{selector:"cdk-nested-tree-node",exportAs:"cdkNestedTreeNode",host:{"[attr.aria-expanded]":"isExpanded","[attr.role]":"role",class:"cdk-tree-node cdk-nested-tree-node"},providers:[{provide:O,useExisting:e}]}]}],e.ctorParameters=function(){return[{type:s.ElementRef},{type:j},{type:s.IterableDiffers}]},e.propDecorators={nodeOutlet:[{type:s.ContentChildren,args:[y]}]},e}(O),E=/([A-Za-z%]+)$/,S=function(){function t(t,e,n,r,i){var s=this;this._treeNode=t,this._tree=e,this._renderer=n,this._element=r,this._dir=i,this._destroyed=new o.Subject,this.indentUnits="px",this._indent=40,this._setPadding(),this._dir&&this._dir.change.pipe(Object(a.takeUntil)(this._destroyed)).subscribe(function(){return s._setPadding()})}return Object.defineProperty(t.prototype,"level",{get:function(){return this._level},set:function(t){this._level=Object(c.coerceNumberProperty)(t),this._setPadding()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"indent",{get:function(){return this._indent},set:function(t){var e=t,n="px";if("string"==typeof t){var r=t.split(E);e=r[0],n=r[1]||n}this.indentUnits=n,this._indent=Object(c.coerceNumberProperty)(e),this._setPadding()},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){this._destroyed.next(),this._destroyed.complete()},t.prototype._paddingIndent=function(){var t=this._treeNode.data&&this._tree.treeControl.getLevel?this._tree.treeControl.getLevel(this._treeNode.data):null,e=this._level||t;return e?""+e*this._indent+this.indentUnits:null},t.prototype._setPadding=function(){var t=this._paddingIndent(),e=this._dir&&"rtl"===this._dir.value?"paddingRight":"paddingLeft";this._renderer.setStyle(this._element.nativeElement,e,t)},t.decorators=[{type:s.Directive,args:[{selector:"[cdkTreeNodePadding]"}]}],t.ctorParameters=function(){return[{type:O},{type:j},{type:s.Renderer2},{type:s.ElementRef},{type:u.Directionality,decorators:[{type:s.Optional}]}]},t.propDecorators={level:[{type:s.Input,args:["cdkTreeNodePadding"]}],indent:[{type:s.Input,args:["cdkTreeNodePaddingIndent"]}]},t}(),M=function(){function t(t,e){this._tree=t,this._treeNode=e,this._recursive=!1}return Object.defineProperty(t.prototype,"recursive",{get:function(){return this._recursive},set:function(t){this._recursive=Object(c.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),t.prototype._toggle=function(t){this.recursive?this._tree.treeControl.toggleDescendants(this._treeNode.data):this._tree.treeControl.toggle(this._treeNode.data),t.stopPropagation()},t.decorators=[{type:s.Directive,args:[{selector:"[cdkTreeNodeToggle]",host:{"(click)":"_toggle($event)"}}]}],t.ctorParameters=function(){return[{type:j},{type:O}]},t.propDecorators={recursive:[{type:s.Input,args:["cdkTreeNodeToggleRecursive"]}]},t}(),T=[C,g,S,M,j,O,y],k=function(){function t(){}return t.decorators=[{type:s.NgModule,args:[{imports:[d.CommonModule],exports:T,declarations:T,providers:[l.FocusMonitor,g]}]}],t}()},"./node_modules/@angular/common/fesm5/common.js":function(t,e,n){"use strict";n.r(e),n.d(e,"ɵangular_packages_common_common_e",function(){return ae}),n.d(e,"ɵangular_packages_common_common_d",function(){return Z}),n.d(e,"ɵangular_packages_common_common_a",function(){return Lt}),n.d(e,"ɵangular_packages_common_common_b",function(){return Ht}),n.d(e,"ɵangular_packages_common_common_g",function(){return ke}),n.d(e,"ɵangular_packages_common_common_f",function(){return $e}),n.d(e,"ɵregisterLocaleData",function(){return m}),n.d(e,"formatDate",function(){return lt}),n.d(e,"formatCurrency",function(){return It}),n.d(e,"formatNumber",function(){return Rt}),n.d(e,"formatPercent",function(){return Pt}),n.d(e,"NgLocaleLocalization",function(){return zt}),n.d(e,"NgLocalization",function(){return Bt}),n.d(e,"registerLocaleData",function(){return m}),n.d(e,"Plural",function(){return _}),n.d(e,"NumberFormatStyle",function(){return g}),n.d(e,"FormStyle",function(){return b}),n.d(e,"TranslationWidth",function(){return w}),n.d(e,"FormatWidth",function(){return O}),n.d(e,"NumberSymbol",function(){return E}),n.d(e,"WeekDay",function(){return M}),n.d(e,"getNumberOfCurrencyDigits",function(){return tt}),n.d(e,"getCurrencySymbol",function(){return $}),n.d(e,"getLocaleDayPeriods",function(){return D}),n.d(e,"getLocaleDayNames",function(){return I}),n.d(e,"getLocaleMonthNames",function(){return P}),n.d(e,"getLocaleId",function(){return A}),n.d(e,"getLocaleEraNames",function(){return R}),n.d(e,"getLocaleWeekEndRange",function(){return F}),n.d(e,"getLocaleFirstDayOfWeek",function(){return N}),n.d(e,"getLocaleDateFormat",function(){return L}),n.d(e,"getLocaleDateTimeFormat",function(){return V}),n.d(e,"getLocaleExtraDayPeriodRules",function(){return Y}),n.d(e,"getLocaleExtraDayPeriods",function(){return K}),n.d(e,"getLocalePluralCase",function(){return q}),n.d(e,"getLocaleTimeFormat",function(){return B}),n.d(e,"getLocaleNumberSymbol",function(){return z}),n.d(e,"getLocaleNumberFormat",function(){return H}),n.d(e,"getLocaleCurrencyName",function(){return W}),n.d(e,"getLocaleCurrencySymbol",function(){return U}),n.d(e,"ɵparseCookieValue",function(){return Ut}),n.d(e,"CommonModule",function(){return Je}),n.d(e,"DeprecatedI18NPipesModule",function(){return en}),n.d(e,"NgClass",function(){return Wt}),n.d(e,"NgForOf",function(){return Yt}),n.d(e,"NgForOfContext",function(){return Gt}),n.d(e,"NgIf",function(){return Xt}),n.d(e,"NgIfContext",function(){return Qt}),n.d(e,"NgPlural",function(){return ne}),n.d(e,"NgPluralCase",function(){return re}),n.d(e,"NgStyle",function(){return ie}),n.d(e,"NgSwitch",function(){return Jt}),n.d(e,"NgSwitchCase",function(){return te}),n.d(e,"NgSwitchDefault",function(){return ee}),n.d(e,"NgTemplateOutlet",function(){return oe}),n.d(e,"NgComponentOutlet",function(){return qt}),n.d(e,"DOCUMENT",function(){return nn}),n.d(e,"AsyncPipe",function(){return Pe}),n.d(e,"DatePipe",function(){return Be}),n.d(e,"I18nPluralPipe",function(){return ze}),n.d(e,"I18nSelectPipe",function(){return He}),n.d(e,"JsonPipe",function(){return Ue}),n.d(e,"LowerCasePipe",function(){return Re}),n.d(e,"CurrencyPipe",function(){return Ke}),n.d(e,"DecimalPipe",function(){return Ge}),n.d(e,"PercentPipe",function(){return Ye}),n.d(e,"SlicePipe",function(){return Ze}),n.d(e,"UpperCasePipe",function(){return Le}),n.d(e,"TitleCasePipe",function(){return Fe}),n.d(e,"KeyValuePipe",function(){return We}),n.d(e,"DeprecatedDatePipe",function(){return Oe}),n.d(e,"DeprecatedCurrencyPipe",function(){return Te}),n.d(e,"DeprecatedDecimalPipe",function(){return Se}),n.d(e,"DeprecatedPercentPipe",function(){return Me}),n.d(e,"ɵPLATFORM_BROWSER_ID",function(){return rn}),n.d(e,"ɵPLATFORM_SERVER_ID",function(){return on}),n.d(e,"ɵPLATFORM_WORKER_APP_ID",function(){return an}),n.d(e,"ɵPLATFORM_WORKER_UI_ID",function(){return sn}),n.d(e,"isPlatformBrowser",function(){return un}),n.d(e,"isPlatformServer",function(){return cn}),n.d(e,"isPlatformWorkerApp",function(){return ln}),n.d(e,"isPlatformWorkerUi",function(){return dn}),n.d(e,"VERSION",function(){return pn}),n.d(e,"ViewportScroller",function(){return fn}),n.d(e,"ɵNullViewportScroller",function(){return mn}),n.d(e,"PlatformLocation",function(){return o}),n.d(e,"LOCATION_INITIALIZED",function(){return a}),n.d(e,"LocationStrategy",function(){return s}),n.d(e,"APP_BASE_HREF",function(){return u}),n.d(e,"HashLocationStrategy",function(){return d}),n.d(e,"PathLocationStrategy",function(){return p}),n.d(e,"Location",function(){return c});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/common/node_modules/tslib/tslib.es6.js"),o=function(){return function(){}}(),a=new r.InjectionToken("Location Initialized"),s=function(){return function(){}}(),u=new r.InjectionToken("appBaseHref"),c=function(){function t(t){var n=this;this._subject=new r.EventEmitter,this._platformStrategy=t;var i=this._platformStrategy.getBaseHref();this._baseHref=e.stripTrailingSlash(l(i)),this._platformStrategy.onPopState(function(t){n._subject.emit({url:n.path(!0),pop:!0,state:t.state,type:t.type})})}var e;return e=t,t.prototype.path=function(t){return void 0===t&&(t=!1),this.normalize(this._platformStrategy.path(t))},t.prototype.isCurrentPathEqualTo=function(t,n){return void 0===n&&(n=""),this.path()==this.normalize(t+e.normalizeQueryParams(n))},t.prototype.normalize=function(t){return e.stripTrailingSlash((n=this._baseHref,r=l(t),n&&r.startsWith(n)?r.substring(n.length):r));var n,r},t.prototype.prepareExternalUrl=function(t){return t&&"/"!==t[0]&&(t="/"+t),this._platformStrategy.prepareExternalUrl(t)},t.prototype.go=function(t,e,n){void 0===e&&(e=""),void 0===n&&(n=null),this._platformStrategy.pushState(n,"",t,e)},t.prototype.replaceState=function(t,e,n){void 0===e&&(e=""),void 0===n&&(n=null),this._platformStrategy.replaceState(n,"",t,e)},t.prototype.forward=function(){this._platformStrategy.forward()},t.prototype.back=function(){this._platformStrategy.back()},t.prototype.subscribe=function(t,e,n){return this._subject.subscribe({next:t,error:e,complete:n})},t.normalizeQueryParams=function(t){return t&&"?"!==t[0]?"?"+t:t},t.joinWithSlash=function(t,e){if(0==t.length)return e;if(0==e.length)return t;var n=0;return t.endsWith("/")&&n++,e.startsWith("/")&&n++,2==n?t+e.substring(1):1==n?t+e:t+"/"+e},t.stripTrailingSlash=function(t){var e=t.match(/#|\?|$/),n=e&&e.index||t.length,r=n-("/"===t[n-1]?1:0);return t.slice(0,r)+t.slice(n)},t=e=Object(i.__decorate)([Object(r.Injectable)(),Object(i.__metadata)("design:paramtypes",[s])],t)}();function l(t){return t.replace(/\/index.html$/,"")}var d=function(t){function e(e,n){var r=t.call(this)||this;return r._platformLocation=e,r._baseHref="",null!=n&&(r._baseHref=n),r}return Object(i.__extends)(e,t),e.prototype.onPopState=function(t){this._platformLocation.onPopState(t),this._platformLocation.onHashChange(t)},e.prototype.getBaseHref=function(){return this._baseHref},e.prototype.path=function(t){void 0===t&&(t=!1);var e=this._platformLocation.hash;return null==e&&(e="#"),e.length>0?e.substring(1):e},e.prototype.prepareExternalUrl=function(t){var e=c.joinWithSlash(this._baseHref,t);return e.length>0?"#"+e:e},e.prototype.pushState=function(t,e,n,r){var i=this.prepareExternalUrl(n+c.normalizeQueryParams(r));0==i.length&&(i=this._platformLocation.pathname),this._platformLocation.pushState(t,e,i)},e.prototype.replaceState=function(t,e,n,r){var i=this.prepareExternalUrl(n+c.normalizeQueryParams(r));0==i.length&&(i=this._platformLocation.pathname),this._platformLocation.replaceState(t,e,i)},e.prototype.forward=function(){this._platformLocation.forward()},e.prototype.back=function(){this._platformLocation.back()},e=Object(i.__decorate)([Object(r.Injectable)(),Object(i.__param)(1,Object(r.Optional)()),Object(i.__param)(1,Object(r.Inject)(u)),Object(i.__metadata)("design:paramtypes",[o,String])],e)}(s),p=function(t){function e(e,n){var r=t.call(this)||this;if(r._platformLocation=e,null==n&&(n=r._platformLocation.getBaseHrefFromDOM()),null==n)throw new Error("No base href set. Please provide a value for the APP_BASE_HREF token or add a base element to the document.");return r._baseHref=n,r}return Object(i.__extends)(e,t),e.prototype.onPopState=function(t){this._platformLocation.onPopState(t),this._platformLocation.onHashChange(t)},e.prototype.getBaseHref=function(){return this._baseHref},e.prototype.prepareExternalUrl=function(t){return c.joinWithSlash(this._baseHref,t)},e.prototype.path=function(t){void 0===t&&(t=!1);var e=this._platformLocation.pathname+c.normalizeQueryParams(this._platformLocation.search),n=this._platformLocation.hash;return n&&t?""+e+n:e},e.prototype.pushState=function(t,e,n,r){var i=this.prepareExternalUrl(n+c.normalizeQueryParams(r));this._platformLocation.pushState(t,e,i)},e.prototype.replaceState=function(t,e,n,r){var i=this.prepareExternalUrl(n+c.normalizeQueryParams(r));this._platformLocation.replaceState(t,e,i)},e.prototype.forward=function(){this._platformLocation.forward()},e.prototype.back=function(){this._platformLocation.back()},e=Object(i.__decorate)([Object(r.Injectable)(),Object(i.__param)(1,Object(r.Optional)()),Object(i.__param)(1,Object(r.Inject)(u)),Object(i.__metadata)("design:paramtypes",[o,String])],e)}(s);var f=["en",[["a","p"],["AM","PM"],void 0],[["AM","PM"],void 0,void 0],[["S","M","T","W","T","F","S"],["Sun","Mon","Tue","Wed","Thu","Fri","Sat"],["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"],["Su","Mo","Tu","We","Th","Fr","Sa"]],void 0,[["J","F","M","A","M","J","J","A","S","O","N","D"],["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"],["January","February","March","April","May","June","July","August","September","October","November","December"]],void 0,[["B","A"],["BC","AD"],["Before Christ","Anno Domini"]],0,[6,0],["M/d/yy","MMM d, y","MMMM d, y","EEEE, MMMM d, y"],["h:mm a","h:mm:ss a","h:mm:ss a z","h:mm:ss a zzzz"],["{1}, {0}",void 0,"{1} 'at' {0}",void 0],[".",",",";","%","+","-","E","×","‰","∞","NaN",":"],["#,##0.###","#,##0%","¤#,##0.00","#E0"],"$","US Dollar",{},function(t){var e=Math.floor(Math.abs(t)),n=t.toString().replace(/^[^.]*\.?/,"").length;return 1===e&&0===n?1:5}],h={};function m(t,e,n){"string"!=typeof e&&(n=e,e=t[0]),e=e.toLowerCase().replace(/_/g,"-"),h[e]=t,n&&(h[e][19]=n)}var g,y,_,v,b,x,w,j,O,C,E,S,M,T,k={ADP:[void 0,void 0,0],AFN:[void 0,void 0,0],ALL:[void 0,void 0,0],AMD:[void 0,void 0,0],AOA:[void 0,"Kz"],ARS:[void 0,"$"],AUD:["A$","$"],BAM:[void 0,"KM"],BBD:[void 0,"$"],BDT:[void 0,"৳"],BHD:[void 0,void 0,3],BIF:[void 0,void 0,0],BMD:[void 0,"$"],BND:[void 0,"$"],BOB:[void 0,"Bs"],BRL:["R$"],BSD:[void 0,"$"],BWP:[void 0,"P"],BYN:[void 0,"р.",2],BYR:[void 0,void 0,0],BZD:[void 0,"$"],CAD:["CA$","$",2],CHF:[void 0,void 0,2],CLF:[void 0,void 0,4],CLP:[void 0,"$",0],CNY:["CN¥","¥"],COP:[void 0,"$",0],CRC:[void 0,"₡",2],CUC:[void 0,"$"],CUP:[void 0,"$"],CZK:[void 0,"Kč",2],DJF:[void 0,void 0,0],DKK:[void 0,"kr",2],DOP:[void 0,"$"],EGP:[void 0,"E£"],ESP:[void 0,"₧",0],EUR:["€"],FJD:[void 0,"$"],FKP:[void 0,"£"],GBP:["£"],GEL:[void 0,"₾"],GIP:[void 0,"£"],GNF:[void 0,"FG",0],GTQ:[void 0,"Q"],GYD:[void 0,"$",0],HKD:["HK$","$"],HNL:[void 0,"L"],HRK:[void 0,"kn"],HUF:[void 0,"Ft",2],IDR:[void 0,"Rp",0],ILS:["₪"],INR:["₹"],IQD:[void 0,void 0,0],IRR:[void 0,void 0,0],ISK:[void 0,"kr",0],ITL:[void 0,void 0,0],JMD:[void 0,"$"],JOD:[void 0,void 0,3],JPY:["¥",void 0,0],KHR:[void 0,"៛"],KMF:[void 0,"CF",0],KPW:[void 0,"₩",0],KRW:["₩",void 0,0],KWD:[void 0,void 0,3],KYD:[void 0,"$"],KZT:[void 0,"₸"],LAK:[void 0,"₭",0],LBP:[void 0,"L£",0],LKR:[void 0,"Rs"],LRD:[void 0,"$"],LTL:[void 0,"Lt"],LUF:[void 0,void 0,0],LVL:[void 0,"Ls"],LYD:[void 0,void 0,3],MGA:[void 0,"Ar",0],MGF:[void 0,void 0,0],MMK:[void 0,"K",0],MNT:[void 0,"₮",0],MRO:[void 0,void 0,0],MUR:[void 0,"Rs",0],MXN:["MX$","$"],MYR:[void 0,"RM"],NAD:[void 0,"$"],NGN:[void 0,"₦"],NIO:[void 0,"C$"],NOK:[void 0,"kr",2],NPR:[void 0,"Rs"],NZD:["NZ$","$"],OMR:[void 0,void 0,3],PHP:[void 0,"₱"],PKR:[void 0,"Rs",0],PLN:[void 0,"zł"],PYG:[void 0,"₲",0],RON:[void 0,"lei"],RSD:[void 0,void 0,0],RUB:[void 0,"₽"],RUR:[void 0,"р."],RWF:[void 0,"RF",0],SBD:[void 0,"$"],SEK:[void 0,"kr",2],SGD:[void 0,"$"],SHP:[void 0,"£"],SLL:[void 0,void 0,0],SOS:[void 0,void 0,0],SRD:[void 0,"$"],SSP:[void 0,"£"],STD:[void 0,void 0,0],STN:[void 0,"Db"],SYP:[void 0,"£",0],THB:[void 0,"฿"],TMM:[void 0,void 0,0],TND:[void 0,void 0,3],TOP:[void 0,"T$"],TRL:[void 0,void 0,0],TRY:[void 0,"₺"],TTD:[void 0,"$"],TWD:["NT$","$",2],TZS:[void 0,void 0,0],UAH:[void 0,"₴"],UGX:[void 0,void 0,0],USD:["$"],UYI:[void 0,void 0,0],UYU:[void 0,"$"],UZS:[void 0,void 0,0],VEF:[void 0,"Bs"],VND:["₫",void 0,0],VUV:[void 0,void 0,0],XAF:["FCFA",void 0,0],XCD:["EC$","$"],XOF:["CFA",void 0,0],XPF:["CFPF",void 0,0],YER:[void 0,void 0,0],ZAR:[void 0,"R"],ZMK:[void 0,void 0,0],ZMW:[void 0,"ZK"],ZWD:[void 0,void 0,0]};function A(t){return Z(t)[0]}function D(t,e,n){var r=Z(t);return X(X([r[1],r[2]],e),n)}function I(t,e,n){var r=Z(t);return X(X([r[3],r[4]],e),n)}function P(t,e,n){var r=Z(t);return X(X([r[5],r[6]],e),n)}function R(t,e){return X(Z(t)[7],e)}function N(t){return Z(t)[8]}function F(t){return Z(t)[9]}function L(t,e){return X(Z(t)[10],e)}function B(t,e){return X(Z(t)[11],e)}function V(t,e){return X(Z(t)[12],e)}function z(t,e){var n=Z(t),r=n[13][e];if(void 0===r){if(e===E.CurrencyDecimal)return n[13][E.Decimal];if(e===E.CurrencyGroup)return n[13][E.Group]}return r}function H(t,e){return Z(t)[14][e]}function U(t){return Z(t)[15]||null}function W(t){return Z(t)[16]||null}function q(t){return Z(t)[18]}function G(t){if(!t[19])throw new Error('Missing extra locale data for the locale "'+t[0]+'". Use "registerLocaleData" to load new data. See the "I18n guide" on angular.io to know more.')}function Y(t){var e=Z(t);return G(e),(e[19][2]||[]).map(function(t){return"string"==typeof t?Q(t):[Q(t[0]),Q(t[1])]})}function K(t,e,n){var r=Z(t);return G(r),X(X([r[19][0],r[19][1]],e)||[],n)||[]}function X(t,e){for(var n=e;n>-1;n--)if(void 0!==t[n])return t[n];throw new Error("Locale data API: locale data undefined")}function Q(t){var e=Object(i.__read)(t.split(":"),2);return{hours:+e[0],minutes:+e[1]}}function Z(t){var e=t.toLowerCase().replace(/_/g,"-"),n=h[e];if(n)return n;var r=e.split("-")[0];if(n=h[r])return n;if("en"===r)return f;throw new Error('Missing locale data for the locale "'+t+'".')}function $(t,e,n){void 0===n&&(n="en");var r,i=(r=n,Z(r)[17])[t]||k[t]||[],o=i[1];return"narrow"===e&&"string"==typeof o?o:i[0]||t}(y=g||(g={}))[y.Decimal=0]="Decimal",y[y.Percent=1]="Percent",y[y.Currency=2]="Currency",y[y.Scientific=3]="Scientific",(v=_||(_={}))[v.Zero=0]="Zero",v[v.One=1]="One",v[v.Two=2]="Two",v[v.Few=3]="Few",v[v.Many=4]="Many",v[v.Other=5]="Other",(x=b||(b={}))[x.Format=0]="Format",x[x.Standalone=1]="Standalone",(j=w||(w={}))[j.Narrow=0]="Narrow",j[j.Abbreviated=1]="Abbreviated",j[j.Wide=2]="Wide",j[j.Short=3]="Short",(C=O||(O={}))[C.Short=0]="Short",C[C.Medium=1]="Medium",C[C.Long=2]="Long",C[C.Full=3]="Full",(S=E||(E={}))[S.Decimal=0]="Decimal",S[S.Group=1]="Group",S[S.List=2]="List",S[S.PercentSign=3]="PercentSign",S[S.PlusSign=4]="PlusSign",S[S.MinusSign=5]="MinusSign",S[S.Exponential=6]="Exponential",S[S.SuperscriptingExponent=7]="SuperscriptingExponent",S[S.PerMille=8]="PerMille",S[S[1/0]=9]="Infinity",S[S.NaN=10]="NaN",S[S.TimeSeparator=11]="TimeSeparator",S[S.CurrencyDecimal=12]="CurrencyDecimal",S[S.CurrencyGroup=13]="CurrencyGroup",(T=M||(M={}))[T.Sunday=0]="Sunday",T[T.Monday=1]="Monday",T[T.Tuesday=2]="Tuesday",T[T.Wednesday=3]="Wednesday",T[T.Thursday=4]="Thursday",T[T.Friday=5]="Friday",T[T.Saturday=6]="Saturday";var J=2;function tt(t){var e,n=k[t];return n&&(e=n[2]),"number"==typeof e?e:J}var et,nt,rt,it,ot,at,st=/^(\d{4})-?(\d\d)-?(\d\d)(?:T(\d\d)(?::?(\d\d)(?::?(\d\d)(?:\.(\d+))?)?)?(Z|([+-])(\d\d):?(\d\d))?)?$/,ut={},ct=/((?:[^GyMLwWdEabBhHmsSzZO']+)|(?:'(?:[^']|'')*')|(?:G{1,5}|y{1,4}|M{1,5}|L{1,5}|w{1,2}|W{1}|d{1,2}|E{1,6}|a{1,5}|b{1,5}|B{1,5}|h{1,2}|H{1,2}|m{1,2}|s{1,2}|S{1,3}|z{1,4}|Z{1,5}|O{1,4}))([\s\S]*)/;function lt(t,e,n,r){var o=function(t){if(wt(t))return t;if("number"==typeof t&&!isNaN(t))return new Date(t);if("string"==typeof t){t=t.trim();var e=parseFloat(t);if(!isNaN(t-e))return new Date(e);if(/^(\d{4}-\d{1,2}-\d{1,2})$/.test(t)){var n=Object(i.__read)(t.split("-").map(function(t){return+t}),3),r=n[0],o=n[1],a=n[2];return new Date(r,o-1,a)}var s=void 0;if(s=t.match(st))return xt(s)}var u=new Date(t);if(!wt(u))throw new Error('Unable to convert "'+t+'" into a date');return u}(t);e=function t(e,n){var r=A(e);ut[r]=ut[r]||{};if(ut[r][n])return ut[r][n];var i="";switch(n){case"shortDate":i=L(e,O.Short);break;case"mediumDate":i=L(e,O.Medium);break;case"longDate":i=L(e,O.Long);break;case"fullDate":i=L(e,O.Full);break;case"shortTime":i=B(e,O.Short);break;case"mediumTime":i=B(e,O.Medium);break;case"longTime":i=B(e,O.Long);break;case"fullTime":i=B(e,O.Full);break;case"short":var o=t(e,"shortTime"),a=t(e,"shortDate");i=dt(V(e,O.Short),[o,a]);break;case"medium":var s=t(e,"mediumTime"),u=t(e,"mediumDate");i=dt(V(e,O.Medium),[s,u]);break;case"long":var c=t(e,"longTime"),l=t(e,"longDate");i=dt(V(e,O.Long),[c,l]);break;case"full":var d=t(e,"fullTime"),p=t(e,"fullDate");i=dt(V(e,O.Full),[d,p])}i&&(ut[r][n]=i);return i}(n,e)||e;for(var a,s=[];e;){if(!(a=ct.exec(e))){s.push(e);break}var u=(s=s.concat(a.slice(1))).pop();if(!u)break;e=u}var c,l,d,p,f,h,m,g=o.getTimezoneOffset();r&&(g=bt(r,g),c=o,l=r,d=!0?-1:1,p=c.getTimezoneOffset(),f=bt(l,p),h=c,m=d*(f-p),(h=new Date(h.getTime())).setMinutes(h.getMinutes()+m),o=h);var y="";return s.forEach(function(t){var e=function(t){if(vt[t])return vt[t];var e;switch(t){case"G":case"GG":case"GGG":e=ht(ot.Eras,w.Abbreviated);break;case"GGGG":e=ht(ot.Eras,w.Wide);break;case"GGGGG":e=ht(ot.Eras,w.Narrow);break;case"y":e=ft(rt.FullYear,1,0,!1,!0);break;case"yy":e=ft(rt.FullYear,2,0,!0,!0);break;case"yyy":e=ft(rt.FullYear,3,0,!1,!0);break;case"yyyy":e=ft(rt.FullYear,4,0,!1,!0);break;case"M":case"L":e=ft(rt.Month,1,1);break;case"MM":case"LL":e=ft(rt.Month,2,1);break;case"MMM":e=ht(ot.Months,w.Abbreviated);break;case"MMMM":e=ht(ot.Months,w.Wide);break;case"MMMMM":e=ht(ot.Months,w.Narrow);break;case"LLL":e=ht(ot.Months,w.Abbreviated,b.Standalone);break;case"LLLL":e=ht(ot.Months,w.Wide,b.Standalone);break;case"LLLLL":e=ht(ot.Months,w.Narrow,b.Standalone);break;case"w":e=_t(1);break;case"ww":e=_t(2);break;case"W":e=_t(1,!0);break;case"d":e=ft(rt.Date,1);break;case"dd":e=ft(rt.Date,2);break;case"E":case"EE":case"EEE":e=ht(ot.Days,w.Abbreviated);break;case"EEEE":e=ht(ot.Days,w.Wide);break;case"EEEEE":e=ht(ot.Days,w.Narrow);break;case"EEEEEE":e=ht(ot.Days,w.Short);break;case"a":case"aa":case"aaa":e=ht(ot.DayPeriods,w.Abbreviated);break;case"aaaa":e=ht(ot.DayPeriods,w.Wide);break;case"aaaaa":e=ht(ot.DayPeriods,w.Narrow);break;case"b":case"bb":case"bbb":e=ht(ot.DayPeriods,w.Abbreviated,b.Standalone,!0);break;case"bbbb":e=ht(ot.DayPeriods,w.Wide,b.Standalone,!0);break;case"bbbbb":e=ht(ot.DayPeriods,w.Narrow,b.Standalone,!0);break;case"B":case"BB":case"BBB":e=ht(ot.DayPeriods,w.Abbreviated,b.Format,!0);break;case"BBBB":e=ht(ot.DayPeriods,w.Wide,b.Format,!0);break;case"BBBBB":e=ht(ot.DayPeriods,w.Narrow,b.Format,!0);break;case"h":e=ft(rt.Hours,1,-12);break;case"hh":e=ft(rt.Hours,2,-12);break;case"H":e=ft(rt.Hours,1);break;case"HH":e=ft(rt.Hours,2);break;case"m":e=ft(rt.Minutes,1);break;case"mm":e=ft(rt.Minutes,2);break;case"s":e=ft(rt.Seconds,1);break;case"ss":e=ft(rt.Seconds,2);break;case"S":e=ft(rt.FractionalSeconds,1);break;case"SS":e=ft(rt.FractionalSeconds,2);break;case"SSS":e=ft(rt.FractionalSeconds,3);break;case"Z":case"ZZ":case"ZZZ":e=mt(et.Short);break;case"ZZZZZ":e=mt(et.Extended);break;case"O":case"OO":case"OOO":case"z":case"zz":case"zzz":e=mt(et.ShortGMT);break;case"OOOO":case"ZZZZ":case"zzzz":e=mt(et.Long);break;default:return null}return vt[t]=e,e}(t);y+=e?e(o,n,g):"''"===t?"'":t.replace(/(^'|'$)/g,"").replace(/''/g,"'")}),y}function dt(t,e){return e&&(t=t.replace(/\{([^}]+)}/g,function(t,n){return null!=e&&n in e?e[n]:t})),t}function pt(t,e,n,r,i){void 0===n&&(n="-");var o="";(t<0||i&&t<=0)&&(i?t=1-t:(t=-t,o=n));for(var a=String(t);a.length<e;)a="0"+a;return r&&(a=a.substr(a.length-e)),o+a}function ft(t,e,n,r,i){return void 0===n&&(n=0),void 0===r&&(r=!1),void 0===i&&(i=!1),function(o,a){var s,u=function(t,e){switch(t){case rt.FullYear:return e.getFullYear();case rt.Month:return e.getMonth();case rt.Date:return e.getDate();case rt.Hours:return e.getHours();case rt.Minutes:return e.getMinutes();case rt.Seconds:return e.getSeconds();case rt.FractionalSeconds:return e.getMilliseconds();case rt.Day:return e.getDay();default:throw new Error('Unknown DateType value "'+t+'".')}}(t,o);if((n>0||u>-n)&&(u+=n),t===rt.Hours)0===u&&-12===n&&(u=12);else if(t===rt.FractionalSeconds)return s=e,pt(u,3).substr(0,s);var c=z(a,E.MinusSign);return pt(u,e,c,r,i)}}function ht(t,e,n,r){return void 0===n&&(n=b.Format),void 0===r&&(r=!1),function(i,o){return function(t,e,n,r,i,o){switch(n){case ot.Months:return P(e,i,r)[t.getMonth()];case ot.Days:return I(e,i,r)[t.getDay()];case ot.DayPeriods:var a=t.getHours(),s=t.getMinutes();if(o){var u,c=Y(e),l=K(e,i,r);if(c.forEach(function(t,e){if(Array.isArray(t)){var n=t[0],r=n.hours,i=n.minutes,o=t[1],c=o.hours,d=o.minutes;a>=r&&s>=i&&(a<c||a===c&&s<d)&&(u=l[e])}else{var p=t.hours,f=t.minutes;p===a&&f===s&&(u=l[e])}}),u)return u}return D(e,i,r)[a<12?0:1];case ot.Eras:return R(e,r)[t.getFullYear()<=0?0:1];default:var d=n;throw new Error("unexpected translation type "+d)}}(i,o,t,e,n,r)}}function mt(t){return function(e,n,r){var i=-1*r,o=z(n,E.MinusSign),a=i>0?Math.floor(i/60):Math.ceil(i/60);switch(t){case et.Short:return(i>=0?"+":"")+pt(a,2,o)+pt(Math.abs(i%60),2,o);case et.ShortGMT:return"GMT"+(i>=0?"+":"")+pt(a,1,o);case et.Long:return"GMT"+(i>=0?"+":"")+pt(a,2,o)+":"+pt(Math.abs(i%60),2,o);case et.Extended:return 0===r?"Z":(i>=0?"+":"")+pt(a,2,o)+":"+pt(Math.abs(i%60),2,o);default:throw new Error('Unknown zone width "'+t+'"')}}}(nt=et||(et={}))[nt.Short=0]="Short",nt[nt.ShortGMT=1]="ShortGMT",nt[nt.Long=2]="Long",nt[nt.Extended=3]="Extended",(it=rt||(rt={}))[it.FullYear=0]="FullYear",it[it.Month=1]="Month",it[it.Date=2]="Date",it[it.Hours=3]="Hours",it[it.Minutes=4]="Minutes",it[it.Seconds=5]="Seconds",it[it.FractionalSeconds=6]="FractionalSeconds",it[it.Day=7]="Day",(at=ot||(ot={}))[at.DayPeriods=0]="DayPeriods",at[at.Days=1]="Days",at[at.Months=2]="Months",at[at.Eras=3]="Eras";var gt=0,yt=4;function _t(t,e){return void 0===e&&(e=!1),function(n,r){var i,o,a,s;if(e){var u=new Date(n.getFullYear(),n.getMonth(),1).getDay()-1,c=n.getDate();i=1+Math.floor((c+u)/7)}else{var l=(a=n.getFullYear(),s=new Date(a,gt,1).getDay(),new Date(a,0,1+(s<=yt?yt:yt+7)-s)),d=(o=n,new Date(o.getFullYear(),o.getMonth(),o.getDate()+(yt-o.getDay()))).getTime()-l.getTime();i=1+Math.round(d/6048e5)}return pt(i,t,z(r,E.MinusSign))}}var vt={};function bt(t,e){t=t.replace(/:/g,"");var n=Date.parse("Jan 01, 1970 00:00:00 "+t)/6e4;return isNaN(n)?e:n}function xt(t){var e=new Date(0),n=0,r=0,i=t[8]?e.setUTCFullYear:e.setFullYear,o=t[8]?e.setUTCHours:e.setHours;t[9]&&(n=Number(t[9]+t[10]),r=Number(t[9]+t[11])),i.call(e,Number(t[1]),Number(t[2])-1,Number(t[3]));var a=Number(t[4]||0)-n,s=Number(t[5]||0)-r,u=Number(t[6]||0),c=Math.round(1e3*parseFloat("0."+(t[7]||0)));return o.call(e,a,s,u,c),e}function wt(t){return t instanceof Date&&!isNaN(t.valueOf())}var jt=/^(\d+)?\.((\d+)(-(\d+))?)?$/,Ot=22,Ct=".",Et="0",St=";",Mt=",",Tt="#",kt="¤",At="%";function Dt(t,e,n,r,i,o,a){void 0===a&&(a=!1);var s="",u=!1;if(isFinite(t)){var c=function(t){var e,n,r,i,o,a=Math.abs(t)+"",s=0;(n=a.indexOf(Ct))>-1&&(a=a.replace(Ct,""));(r=a.search(/e/i))>0?(n<0&&(n=r),n+=+a.slice(r+1),a=a.substring(0,r)):n<0&&(n=a.length);for(r=0;a.charAt(r)===Et;r++);if(r===(o=a.length))e=[0],n=1;else{for(o--;a.charAt(o)===Et;)o--;for(n-=r,e=[],i=0;r<=o;r++,i++)e[i]=Number(a.charAt(r))}n>Ot&&(e=e.splice(0,Ot-1),s=n-1,n=1);return{digits:e,exponent:s,integerLen:n}}(t);a&&(c=function(t){if(0===t.digits[0])return t;var e=t.digits.length-t.integerLen;t.exponent?t.exponent+=2:(0===e?t.digits.push(0,0):1===e&&t.digits.push(0),t.integerLen+=2);return t}(c));var l=e.minInt,d=e.minFrac,p=e.maxFrac;if(o){var f=o.match(jt);if(null===f)throw new Error(o+" is not a valid digit info");var h=f[1],m=f[3],g=f[5];null!=h&&(l=Ft(h)),null!=m&&(d=Ft(m)),null!=g?p=Ft(g):null!=m&&d>p&&(p=d)}!function(t,e,n){if(e>n)throw new Error("The minimum number of digits after fraction ("+e+") is higher than the maximum ("+n+").");var r=t.digits,i=r.length-t.integerLen,o=Math.min(Math.max(e,i),n),a=o+t.integerLen,s=r[a];if(a>0){r.splice(Math.max(t.integerLen,a));for(var u=a;u<r.length;u++)r[u]=0}else{i=Math.max(0,i),t.integerLen=1,r.length=Math.max(1,a=o+1),r[0]=0;for(var c=1;c<a;c++)r[c]=0}if(s>=5)if(a-1<0){for(var l=0;l>a;l--)r.unshift(0),t.integerLen++;r.unshift(1),t.integerLen++}else r[a-1]++;for(;i<Math.max(0,o);i++)r.push(0);var d=0!==o,p=e+t.integerLen,f=r.reduceRight(function(t,e,n,r){return e+=t,r[n]=e<10?e:e-10,d&&(0===r[n]&&n>=p?r.pop():d=!1),e>=10?1:0},0);f&&(r.unshift(f),t.integerLen++)}(c,d,p);var y=c.digits,_=c.integerLen,v=c.exponent,b=[];for(u=y.every(function(t){return!t});_<l;_++)y.unshift(0);for(;_<0;_++)y.unshift(0);_>0?b=y.splice(_,y.length):(b=y,y=[0]);var x=[];for(y.length>=e.lgSize&&x.unshift(y.splice(-e.lgSize,y.length).join(""));y.length>e.gSize;)x.unshift(y.splice(-e.gSize,y.length).join(""));y.length&&x.unshift(y.join("")),s=x.join(z(n,r)),b.length&&(s+=z(n,i)+b.join("")),v&&(s+=z(n,E.Exponential)+"+"+v)}else s=z(n,E.Infinity);return s=t<0&&!u?e.negPre+s+e.negSuf:e.posPre+s+e.posSuf}function It(t,e,n,r,i){var o=Nt(H(e,g.Currency),z(e,E.MinusSign));return o.minFrac=tt(r),o.maxFrac=o.minFrac,Dt(t,o,e,E.CurrencyGroup,E.CurrencyDecimal,i).replace(kt,n).replace(kt,"")}function Pt(t,e,n){return Dt(t,Nt(H(e,g.Percent),z(e,E.MinusSign)),e,E.Group,E.Decimal,n,!0).replace(new RegExp(At,"g"),z(e,E.PercentSign))}function Rt(t,e,n){return Dt(t,Nt(H(e,g.Decimal),z(e,E.MinusSign)),e,E.Group,E.Decimal,n)}function Nt(t,e){void 0===e&&(e="-");var n={minInt:1,minFrac:0,maxFrac:0,posPre:"",posSuf:"",negPre:"",negSuf:"",gSize:0,lgSize:0},r=t.split(St),i=r[0],o=r[1],a=-1!==i.indexOf(Ct)?i.split(Ct):[i.substring(0,i.lastIndexOf(Et)+1),i.substring(i.lastIndexOf(Et)+1)],s=a[0],u=a[1]||"";n.posPre=s.substr(0,s.indexOf(Tt));for(var c=0;c<u.length;c++){var l=u.charAt(c);l===Et?n.minFrac=n.maxFrac=c+1:l===Tt?n.maxFrac=c+1:n.posSuf+=l}var d=s.split(Mt);if(n.gSize=d[1]?d[1].length:0,n.lgSize=d[2]||d[1]?(d[2]||d[1]).length:0,o){var p=i.length-n.posPre.length-n.posSuf.length,f=o.indexOf(Tt);n.negPre=o.substr(0,f).replace(/'/g,""),n.negSuf=o.substr(f+p).replace(/'/g,"")}else n.negPre=e+n.posPre,n.negSuf=n.posSuf;return n}function Ft(t){var e=parseInt(t);if(isNaN(e))throw new Error("Invalid integer literal when parsing "+t);return e}var Lt=new r.InjectionToken("UseV4Plurals"),Bt=function(){return function(){}}();function Vt(t,e,n,r){var i="="+t;if(e.indexOf(i)>-1)return i;if(i=n.getPluralCategory(t,r),e.indexOf(i)>-1)return i;if(e.indexOf("other")>-1)return"other";throw new Error('No plural message found for value "'+t+'"')}var zt=function(t){function e(e,n){var r=t.call(this)||this;return r.locale=e,r.deprecatedPluralFn=n,r}return Object(i.__extends)(e,t),e.prototype.getPluralCategory=function(t,e){switch(this.deprecatedPluralFn?this.deprecatedPluralFn(e||this.locale,t):q(e||this.locale)(t)){case _.Zero:return"zero";case _.One:return"one";case _.Two:return"two";case _.Few:return"few";case _.Many:return"many";default:return"other"}},e=Object(i.__decorate)([Object(r.Injectable)(),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__param)(1,Object(r.Optional)()),Object(i.__param)(1,Object(r.Inject)(Lt)),Object(i.__metadata)("design:paramtypes",[String,Object])],e)}(Bt);function Ht(t,e){"string"==typeof e&&(e=parseInt(e,10));var n=e,r=n.toString().replace(/^[^.]*\.?/,""),i=Math.floor(Math.abs(n)),o=r.length,a=parseInt(r,10),s=parseInt(n.toString().replace(/^[^.]*\.?|0+$/g,""),10)||0;switch(t.split("-")[0].toLowerCase()){case"af":case"asa":case"az":case"bem":case"bez":case"bg":case"brx":case"ce":case"cgg":case"chr":case"ckb":case"ee":case"el":case"eo":case"es":case"eu":case"fo":case"fur":case"gsw":case"ha":case"haw":case"hu":case"jgo":case"jmc":case"ka":case"kk":case"kkj":case"kl":case"ks":case"ksb":case"ky":case"lb":case"lg":case"mas":case"mgo":case"ml":case"mn":case"nb":case"nd":case"ne":case"nn":case"nnh":case"nyn":case"om":case"or":case"os":case"ps":case"rm":case"rof":case"rwk":case"saq":case"seh":case"sn":case"so":case"sq":case"ta":case"te":case"teo":case"tk":case"tr":case"ug":case"uz":case"vo":case"vun":case"wae":case"xog":return 1===n?_.One:_.Other;case"ak":case"ln":case"mg":case"pa":case"ti":return n===Math.floor(n)&&n>=0&&n<=1?_.One:_.Other;case"am":case"as":case"bn":case"fa":case"gu":case"hi":case"kn":case"mr":case"zu":return 0===i||1===n?_.One:_.Other;case"ar":return 0===n?_.Zero:1===n?_.One:2===n?_.Two:n%100===Math.floor(n%100)&&n%100>=3&&n%100<=10?_.Few:n%100===Math.floor(n%100)&&n%100>=11&&n%100<=99?_.Many:_.Other;case"ast":case"ca":case"de":case"en":case"et":case"fi":case"fy":case"gl":case"it":case"nl":case"sv":case"sw":case"ur":case"yi":return 1===i&&0===o?_.One:_.Other;case"be":return n%10==1&&n%100!=11?_.One:n%10===Math.floor(n%10)&&n%10>=2&&n%10<=4&&!(n%100>=12&&n%100<=14)?_.Few:n%10==0||n%10===Math.floor(n%10)&&n%10>=5&&n%10<=9||n%100===Math.floor(n%100)&&n%100>=11&&n%100<=14?_.Many:_.Other;case"br":return n%10==1&&n%100!=11&&n%100!=71&&n%100!=91?_.One:n%10==2&&n%100!=12&&n%100!=72&&n%100!=92?_.Two:n%10===Math.floor(n%10)&&(n%10>=3&&n%10<=4||n%10==9)&&!(n%100>=10&&n%100<=19||n%100>=70&&n%100<=79||n%100>=90&&n%100<=99)?_.Few:0!==n&&n%1e6==0?_.Many:_.Other;case"bs":case"hr":case"sr":return 0===o&&i%10==1&&i%100!=11||a%10==1&&a%100!=11?_.One:0===o&&i%10===Math.floor(i%10)&&i%10>=2&&i%10<=4&&!(i%100>=12&&i%100<=14)||a%10===Math.floor(a%10)&&a%10>=2&&a%10<=4&&!(a%100>=12&&a%100<=14)?_.Few:_.Other;case"cs":case"sk":return 1===i&&0===o?_.One:i===Math.floor(i)&&i>=2&&i<=4&&0===o?_.Few:0!==o?_.Many:_.Other;case"cy":return 0===n?_.Zero:1===n?_.One:2===n?_.Two:3===n?_.Few:6===n?_.Many:_.Other;case"da":return 1===n||0!==s&&(0===i||1===i)?_.One:_.Other;case"dsb":case"hsb":return 0===o&&i%100==1||a%100==1?_.One:0===o&&i%100==2||a%100==2?_.Two:0===o&&i%100===Math.floor(i%100)&&i%100>=3&&i%100<=4||a%100===Math.floor(a%100)&&a%100>=3&&a%100<=4?_.Few:_.Other;case"ff":case"fr":case"hy":case"kab":return 0===i||1===i?_.One:_.Other;case"fil":return 0===o&&(1===i||2===i||3===i)||0===o&&i%10!=4&&i%10!=6&&i%10!=9||0!==o&&a%10!=4&&a%10!=6&&a%10!=9?_.One:_.Other;case"ga":return 1===n?_.One:2===n?_.Two:n===Math.floor(n)&&n>=3&&n<=6?_.Few:n===Math.floor(n)&&n>=7&&n<=10?_.Many:_.Other;case"gd":return 1===n||11===n?_.One:2===n||12===n?_.Two:n===Math.floor(n)&&(n>=3&&n<=10||n>=13&&n<=19)?_.Few:_.Other;case"gv":return 0===o&&i%10==1?_.One:0===o&&i%10==2?_.Two:0!==o||i%100!=0&&i%100!=20&&i%100!=40&&i%100!=60&&i%100!=80?0!==o?_.Many:_.Other:_.Few;case"he":return 1===i&&0===o?_.One:2===i&&0===o?_.Two:0!==o||n>=0&&n<=10||n%10!=0?_.Other:_.Many;case"is":return 0===s&&i%10==1&&i%100!=11||0!==s?_.One:_.Other;case"ksh":return 0===n?_.Zero:1===n?_.One:_.Other;case"kw":case"naq":case"se":case"smn":return 1===n?_.One:2===n?_.Two:_.Other;case"lag":return 0===n?_.Zero:0!==i&&1!==i||0===n?_.Other:_.One;case"lt":return n%10!=1||n%100>=11&&n%100<=19?n%10===Math.floor(n%10)&&n%10>=2&&n%10<=9&&!(n%100>=11&&n%100<=19)?_.Few:0!==a?_.Many:_.Other:_.One;case"lv":case"prg":return n%10==0||n%100===Math.floor(n%100)&&n%100>=11&&n%100<=19||2===o&&a%100===Math.floor(a%100)&&a%100>=11&&a%100<=19?_.Zero:n%10==1&&n%100!=11||2===o&&a%10==1&&a%100!=11||2!==o&&a%10==1?_.One:_.Other;case"mk":return 0===o&&i%10==1||a%10==1?_.One:_.Other;case"mt":return 1===n?_.One:0===n||n%100===Math.floor(n%100)&&n%100>=2&&n%100<=10?_.Few:n%100===Math.floor(n%100)&&n%100>=11&&n%100<=19?_.Many:_.Other;case"pl":return 1===i&&0===o?_.One:0===o&&i%10===Math.floor(i%10)&&i%10>=2&&i%10<=4&&!(i%100>=12&&i%100<=14)?_.Few:0===o&&1!==i&&i%10===Math.floor(i%10)&&i%10>=0&&i%10<=1||0===o&&i%10===Math.floor(i%10)&&i%10>=5&&i%10<=9||0===o&&i%100===Math.floor(i%100)&&i%100>=12&&i%100<=14?_.Many:_.Other;case"pt":return n===Math.floor(n)&&n>=0&&n<=2&&2!==n?_.One:_.Other;case"ro":return 1===i&&0===o?_.One:0!==o||0===n||1!==n&&n%100===Math.floor(n%100)&&n%100>=1&&n%100<=19?_.Few:_.Other;case"ru":case"uk":return 0===o&&i%10==1&&i%100!=11?_.One:0===o&&i%10===Math.floor(i%10)&&i%10>=2&&i%10<=4&&!(i%100>=12&&i%100<=14)?_.Few:0===o&&i%10==0||0===o&&i%10===Math.floor(i%10)&&i%10>=5&&i%10<=9||0===o&&i%100===Math.floor(i%100)&&i%100>=11&&i%100<=14?_.Many:_.Other;case"shi":return 0===i||1===n?_.One:n===Math.floor(n)&&n>=2&&n<=10?_.Few:_.Other;case"si":return 0===n||1===n||0===i&&1===a?_.One:_.Other;case"sl":return 0===o&&i%100==1?_.One:0===o&&i%100==2?_.Two:0===o&&i%100===Math.floor(i%100)&&i%100>=3&&i%100<=4||0!==o?_.Few:_.Other;case"tzm":return n===Math.floor(n)&&n>=0&&n<=1||n===Math.floor(n)&&n>=11&&n<=99?_.One:_.Other;default:return _.Other}}function Ut(t,e){var n,r;e=encodeURIComponent(e);try{for(var o=Object(i.__values)(t.split(";")),a=o.next();!a.done;a=o.next()){var s=a.value,u=s.indexOf("="),c=Object(i.__read)(-1==u?[s,""]:[s.slice(0,u),s.slice(u+1)],2),l=c[0],d=c[1];if(l.trim()===e)return decodeURIComponent(d)}}catch(t){n={error:t}}finally{try{a&&!a.done&&(r=o.return)&&r.call(o)}finally{if(n)throw n.error}}return null}var Wt=function(){function t(t,e,n,r){this._iterableDiffers=t,this._keyValueDiffers=e,this._ngEl=n,this._renderer=r,this._initialClasses=[]}return Object.defineProperty(t.prototype,"klass",{set:function(t){this._removeClasses(this._initialClasses),this._initialClasses="string"==typeof t?t.split(/\s+/):[],this._applyClasses(this._initialClasses),this._applyClasses(this._rawClass)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngClass",{set:function(t){this._removeClasses(this._rawClass),this._applyClasses(this._initialClasses),this._iterableDiffer=null,this._keyValueDiffer=null,this._rawClass="string"==typeof t?t.split(/\s+/):t,this._rawClass&&(Object(r["ɵisListLikeIterable"])(this._rawClass)?this._iterableDiffer=this._iterableDiffers.find(this._rawClass).create():this._keyValueDiffer=this._keyValueDiffers.find(this._rawClass).create())},enumerable:!0,configurable:!0}),t.prototype.ngDoCheck=function(){if(this._iterableDiffer){var t=this._iterableDiffer.diff(this._rawClass);t&&this._applyIterableChanges(t)}else if(this._keyValueDiffer){var e=this._keyValueDiffer.diff(this._rawClass);e&&this._applyKeyValueChanges(e)}},t.prototype._applyKeyValueChanges=function(t){var e=this;t.forEachAddedItem(function(t){return e._toggleClass(t.key,t.currentValue)}),t.forEachChangedItem(function(t){return e._toggleClass(t.key,t.currentValue)}),t.forEachRemovedItem(function(t){t.previousValue&&e._toggleClass(t.key,!1)})},t.prototype._applyIterableChanges=function(t){var e=this;t.forEachAddedItem(function(t){if("string"!=typeof t.item)throw new Error("NgClass can only toggle CSS classes expressed as strings, got "+Object(r["ɵstringify"])(t.item));e._toggleClass(t.item,!0)}),t.forEachRemovedItem(function(t){return e._toggleClass(t.item,!1)})},t.prototype._applyClasses=function(t){var e=this;t&&(Array.isArray(t)||t instanceof Set?t.forEach(function(t){return e._toggleClass(t,!0)}):Object.keys(t).forEach(function(n){return e._toggleClass(n,!!t[n])}))},t.prototype._removeClasses=function(t){var e=this;t&&(Array.isArray(t)||t instanceof Set?t.forEach(function(t){return e._toggleClass(t,!1)}):Object.keys(t).forEach(function(t){return e._toggleClass(t,!1)}))},t.prototype._toggleClass=function(t,e){var n=this;(t=t.trim())&&t.split(/\s+/g).forEach(function(t){e?n._renderer.addClass(n._ngEl.nativeElement,t):n._renderer.removeClass(n._ngEl.nativeElement,t)})},Object(i.__decorate)([Object(r.Input)("class"),Object(i.__metadata)("design:type",String),Object(i.__metadata)("design:paramtypes",[String])],t.prototype,"klass",null),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object),Object(i.__metadata)("design:paramtypes",[Object])],t.prototype,"ngClass",null),t=Object(i.__decorate)([Object(r.Directive)({selector:"[ngClass]"}),Object(i.__metadata)("design:paramtypes",[r.IterableDiffers,r.KeyValueDiffers,r.ElementRef,r.Renderer2])],t)}(),qt=function(){function t(t){this._viewContainerRef=t,this._componentRef=null,this._moduleRef=null}return t.prototype.ngOnChanges=function(t){if(this._viewContainerRef.clear(),this._componentRef=null,this.ngComponentOutlet){var e=this.ngComponentOutletInjector||this._viewContainerRef.parentInjector;if(t.ngComponentOutletNgModuleFactory)if(this._moduleRef&&this._moduleRef.destroy(),this.ngComponentOutletNgModuleFactory){var n=e.get(r.NgModuleRef);this._moduleRef=this.ngComponentOutletNgModuleFactory.create(n.injector)}else this._moduleRef=null;var i=(this._moduleRef?this._moduleRef.componentFactoryResolver:e.get(r.ComponentFactoryResolver)).resolveComponentFactory(this.ngComponentOutlet);this._componentRef=this._viewContainerRef.createComponent(i,this._viewContainerRef.length,e,this.ngComponentOutletContent)}},t.prototype.ngOnDestroy=function(){this._moduleRef&&this._moduleRef.destroy()},Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",r.Type)],t.prototype,"ngComponentOutlet",void 0),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",r.Injector)],t.prototype,"ngComponentOutletInjector",void 0),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Array)],t.prototype,"ngComponentOutletContent",void 0),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",r.NgModuleFactory)],t.prototype,"ngComponentOutletNgModuleFactory",void 0),t=Object(i.__decorate)([Object(r.Directive)({selector:"[ngComponentOutlet]"}),Object(i.__metadata)("design:paramtypes",[r.ViewContainerRef])],t)}(),Gt=function(){function t(t,e,n,r){this.$implicit=t,this.ngForOf=e,this.index=n,this.count=r}return Object.defineProperty(t.prototype,"first",{get:function(){return 0===this.index},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"last",{get:function(){return this.index===this.count-1},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"even",{get:function(){return this.index%2==0},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"odd",{get:function(){return!this.even},enumerable:!0,configurable:!0}),t}(),Yt=function(){function t(t,e,n){this._viewContainer=t,this._template=e,this._differs=n,this._ngForOfDirty=!0,this._differ=null}return Object.defineProperty(t.prototype,"ngForOf",{set:function(t){this._ngForOf=t,this._ngForOfDirty=!0},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngForTrackBy",{get:function(){return this._trackByFn},set:function(t){Object(r.isDevMode)()&&null!=t&&"function"!=typeof t&&console&&console.warn&&console.warn("trackBy must be a function, but received "+JSON.stringify(t)+". See https://angular.io/docs/ts/latest/api/common/index/NgFor-directive.html#!#change-propagation for more information."),this._trackByFn=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngForTemplate",{set:function(t){t&&(this._template=t)},enumerable:!0,configurable:!0}),t.prototype.ngDoCheck=function(){if(this._ngForOfDirty){this._ngForOfDirty=!1;var t=this._ngForOf;if(!this._differ&&t)try{this._differ=this._differs.find(t).create(this.ngForTrackBy)}catch(n){throw new Error("Cannot find a differ supporting object '"+t+"' of type '"+((e=t).name||typeof e)+"'. NgFor only supports binding to Iterables such as Arrays.")}}var e;if(this._differ){var n=this._differ.diff(this._ngForOf);n&&this._applyChanges(n)}},t.prototype._applyChanges=function(t){var e=this,n=[];t.forEachOperation(function(t,r,i){if(null==t.previousIndex){var o=e._viewContainer.createEmbeddedView(e._template,new Gt(null,e._ngForOf,-1,-1),i),a=new Kt(t,o);n.push(a)}else if(null==i)e._viewContainer.remove(r);else{o=e._viewContainer.get(r);e._viewContainer.move(o,i);a=new Kt(t,o);n.push(a)}});for(var r=0;r<n.length;r++)this._perViewChange(n[r].view,n[r].record);r=0;for(var i=this._viewContainer.length;r<i;r++){var o=this._viewContainer.get(r);o.context.index=r,o.context.count=i,o.context.ngForOf=this._ngForOf}t.forEachIdentityChange(function(t){e._viewContainer.get(t.currentIndex).context.$implicit=t.item})},t.prototype._perViewChange=function(t,e){t.context.$implicit=e.item},t.ngTemplateContextGuard=function(t,e){return!0},Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object),Object(i.__metadata)("design:paramtypes",[Object])],t.prototype,"ngForOf",null),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Function),Object(i.__metadata)("design:paramtypes",[Function])],t.prototype,"ngForTrackBy",null),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",r.TemplateRef),Object(i.__metadata)("design:paramtypes",[r.TemplateRef])],t.prototype,"ngForTemplate",null),t=Object(i.__decorate)([Object(r.Directive)({selector:"[ngFor][ngForOf]"}),Object(i.__metadata)("design:paramtypes",[r.ViewContainerRef,r.TemplateRef,r.IterableDiffers])],t)}(),Kt=function(){return function(t,e){this.record=t,this.view=e}}();var Xt=function(){function t(t,e){this._viewContainer=t,this._context=new Qt,this._thenTemplateRef=null,this._elseTemplateRef=null,this._thenViewRef=null,this._elseViewRef=null,this._thenTemplateRef=e}return Object.defineProperty(t.prototype,"ngIf",{set:function(t){this._context.$implicit=this._context.ngIf=t,this._updateView()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngIfThen",{set:function(t){Zt("ngIfThen",t),this._thenTemplateRef=t,this._thenViewRef=null,this._updateView()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngIfElse",{set:function(t){Zt("ngIfElse",t),this._elseTemplateRef=t,this._elseViewRef=null,this._updateView()},enumerable:!0,configurable:!0}),t.prototype._updateView=function(){this._context.$implicit?this._thenViewRef||(this._viewContainer.clear(),this._elseViewRef=null,this._thenTemplateRef&&(this._thenViewRef=this._viewContainer.createEmbeddedView(this._thenTemplateRef,this._context))):this._elseViewRef||(this._viewContainer.clear(),this._thenViewRef=null,this._elseTemplateRef&&(this._elseViewRef=this._viewContainer.createEmbeddedView(this._elseTemplateRef,this._context)))},t.ngTemplateGuard_ngIf=function(t,e){return!0},Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object),Object(i.__metadata)("design:paramtypes",[Object])],t.prototype,"ngIf",null),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object),Object(i.__metadata)("design:paramtypes",[Object])],t.prototype,"ngIfThen",null),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object),Object(i.__metadata)("design:paramtypes",[Object])],t.prototype,"ngIfElse",null),t=Object(i.__decorate)([Object(r.Directive)({selector:"[ngIf]"}),Object(i.__metadata)("design:paramtypes",[r.ViewContainerRef,r.TemplateRef])],t)}(),Qt=function(){return function(){this.$implicit=null,this.ngIf=null}}();function Zt(t,e){if(!!(e&&!e.createEmbeddedView))throw new Error(t+" must be a TemplateRef, but received '"+Object(r["ɵstringify"])(e)+"'.")}var $t=function(){function t(t,e){this._viewContainerRef=t,this._templateRef=e,this._created=!1}return t.prototype.create=function(){this._created=!0,this._viewContainerRef.createEmbeddedView(this._templateRef)},t.prototype.destroy=function(){this._created=!1,this._viewContainerRef.clear()},t.prototype.enforceState=function(t){t&&!this._created?this.create():!t&&this._created&&this.destroy()},t}(),Jt=function(){function t(){this._defaultUsed=!1,this._caseCount=0,this._lastCaseCheckIndex=0,this._lastCasesMatched=!1}return Object.defineProperty(t.prototype,"ngSwitch",{set:function(t){this._ngSwitch=t,0===this._caseCount&&this._updateDefaultCases(!0)},enumerable:!0,configurable:!0}),t.prototype._addCase=function(){return this._caseCount++},t.prototype._addDefault=function(t){this._defaultViews||(this._defaultViews=[]),this._defaultViews.push(t)},t.prototype._matchCase=function(t){var e=t==this._ngSwitch;return this._lastCasesMatched=this._lastCasesMatched||e,this._lastCaseCheckIndex++,this._lastCaseCheckIndex===this._caseCount&&(this._updateDefaultCases(!this._lastCasesMatched),this._lastCaseCheckIndex=0,this._lastCasesMatched=!1),e},t.prototype._updateDefaultCases=function(t){if(this._defaultViews&&t!==this._defaultUsed){this._defaultUsed=t;for(var e=0;e<this._defaultViews.length;e++){this._defaultViews[e].enforceState(t)}}},Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object),Object(i.__metadata)("design:paramtypes",[Object])],t.prototype,"ngSwitch",null),t=Object(i.__decorate)([Object(r.Directive)({selector:"[ngSwitch]"})],t)}(),te=function(){function t(t,e,n){this.ngSwitch=n,n._addCase(),this._view=new $t(t,e)}return t.prototype.ngDoCheck=function(){this._view.enforceState(this.ngSwitch._matchCase(this.ngSwitchCase))},Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object)],t.prototype,"ngSwitchCase",void 0),t=Object(i.__decorate)([Object(r.Directive)({selector:"[ngSwitchCase]"}),Object(i.__param)(2,Object(r.Host)()),Object(i.__metadata)("design:paramtypes",[r.ViewContainerRef,r.TemplateRef,Jt])],t)}(),ee=function(){function t(t,e,n){n._addDefault(new $t(t,e))}return t=Object(i.__decorate)([Object(r.Directive)({selector:"[ngSwitchDefault]"}),Object(i.__param)(2,Object(r.Host)()),Object(i.__metadata)("design:paramtypes",[r.ViewContainerRef,r.TemplateRef,Jt])],t)}(),ne=function(){function t(t){this._localization=t,this._caseViews={}}return Object.defineProperty(t.prototype,"ngPlural",{set:function(t){this._switchValue=t,this._updateView()},enumerable:!0,configurable:!0}),t.prototype.addCase=function(t,e){this._caseViews[t]=e},t.prototype._updateView=function(){this._clearViews();var t=Object.keys(this._caseViews),e=Vt(this._switchValue,t,this._localization);this._activateView(this._caseViews[e])},t.prototype._clearViews=function(){this._activeView&&this._activeView.destroy()},t.prototype._activateView=function(t){t&&(this._activeView=t,this._activeView.create())},Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Number),Object(i.__metadata)("design:paramtypes",[Number])],t.prototype,"ngPlural",null),t=Object(i.__decorate)([Object(r.Directive)({selector:"[ngPlural]"}),Object(i.__metadata)("design:paramtypes",[Bt])],t)}(),re=function(){function t(t,e,n,r){this.value=t;var i=!isNaN(Number(t));r.addCase(i?"="+t:t,new $t(n,e))}return t=Object(i.__decorate)([Object(r.Directive)({selector:"[ngPluralCase]"}),Object(i.__param)(0,Object(r.Attribute)("ngPluralCase")),Object(i.__param)(3,Object(r.Host)()),Object(i.__metadata)("design:paramtypes",[String,r.TemplateRef,r.ViewContainerRef,ne])],t)}(),ie=function(){function t(t,e,n){this._differs=t,this._ngEl=e,this._renderer=n}return Object.defineProperty(t.prototype,"ngStyle",{set:function(t){this._ngStyle=t,!this._differ&&t&&(this._differ=this._differs.find(t).create())},enumerable:!0,configurable:!0}),t.prototype.ngDoCheck=function(){if(this._differ){var t=this._differ.diff(this._ngStyle);t&&this._applyChanges(t)}},t.prototype._applyChanges=function(t){var e=this;t.forEachRemovedItem(function(t){return e._setStyle(t.key,null)}),t.forEachAddedItem(function(t){return e._setStyle(t.key,t.currentValue)}),t.forEachChangedItem(function(t){return e._setStyle(t.key,t.currentValue)})},t.prototype._setStyle=function(t,e){var n=Object(i.__read)(t.split("."),2),r=n[0],o=n[1];null!=(e=null!=e&&o?""+e+o:e)?this._renderer.setStyle(this._ngEl.nativeElement,r,e):this._renderer.removeStyle(this._ngEl.nativeElement,r)},Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object),Object(i.__metadata)("design:paramtypes",[Object])],t.prototype,"ngStyle",null),t=Object(i.__decorate)([Object(r.Directive)({selector:"[ngStyle]"}),Object(i.__metadata)("design:paramtypes",[r.KeyValueDiffers,r.ElementRef,r.Renderer2])],t)}(),oe=function(){function t(t){this._viewContainerRef=t}return t.prototype.ngOnChanges=function(t){this._shouldRecreateView(t)?(this._viewRef&&this._viewContainerRef.remove(this._viewContainerRef.indexOf(this._viewRef)),this.ngTemplateOutlet&&(this._viewRef=this._viewContainerRef.createEmbeddedView(this.ngTemplateOutlet,this.ngTemplateOutletContext))):this._viewRef&&this.ngTemplateOutletContext&&this._updateExistingContext(this.ngTemplateOutletContext)},t.prototype._shouldRecreateView=function(t){var e=t.ngTemplateOutletContext;return!!t.ngTemplateOutlet||e&&this._hasContextShapeChanged(e)},t.prototype._hasContextShapeChanged=function(t){var e,n,r=Object.keys(t.previousValue||{}),o=Object.keys(t.currentValue||{});if(r.length===o.length){try{for(var a=Object(i.__values)(o),s=a.next();!s.done;s=a.next()){var u=s.value;if(-1===r.indexOf(u))return!0}}catch(t){e={error:t}}finally{try{s&&!s.done&&(n=a.return)&&n.call(a)}finally{if(e)throw e.error}}return!1}return!0},t.prototype._updateExistingContext=function(t){var e,n;try{for(var r=Object(i.__values)(Object.keys(t)),o=r.next();!o.done;o=r.next()){var a=o.value;this._viewRef.context[a]=this.ngTemplateOutletContext[a]}}catch(t){e={error:t}}finally{try{o&&!o.done&&(n=r.return)&&n.call(r)}finally{if(e)throw e.error}}},Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object)],t.prototype,"ngTemplateOutletContext",void 0),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",r.TemplateRef)],t.prototype,"ngTemplateOutlet",void 0),t=Object(i.__decorate)([Object(r.Directive)({selector:"[ngTemplateOutlet]"}),Object(i.__metadata)("design:paramtypes",[r.ViewContainerRef])],t)}(),ae=[Wt,qt,Yt,Xt,oe,ie,Jt,te,ee,ne,re];function se(t,e){return Error("InvalidPipeArgument: '"+e+"' for pipe '"+Object(r["ɵstringify"])(t)+"'")}var ue,ce=function(){function t(){}return t.format=function(t,e,n,r){void 0===r&&(r={});var i=r.minimumIntegerDigits,o=r.minimumFractionDigits,a=r.maximumFractionDigits,s=r.currency,u=r.currencyAsSymbol,c=void 0!==u&&u,l={minimumIntegerDigits:i,minimumFractionDigits:o,maximumFractionDigits:a,style:g[n].toLowerCase()};return n==g.Currency&&(l.currency="string"==typeof s?s:void 0,l.currencyDisplay=c?"symbol":"code"),new Intl.NumberFormat(e,l).format(t)},t}(),le=/((?:[^yMLdHhmsazZEwGjJ']+)|(?:'(?:[^']|'')*')|(?:E+|y+|M+|L+|d+|H+|h+|J+|j+|m+|s+|a|z|Z|G+|w+))(.*)/,de={yMMMdjms:xe(be([_e("year",1),ve("month",3),_e("day",1),_e("hour",1),_e("minute",1),_e("second",1)])),yMdjm:xe(be([_e("year",1),_e("month",1),_e("day",1),_e("hour",1),_e("minute",1)])),yMMMMEEEEd:xe(be([_e("year",1),ve("month",4),ve("weekday",4),_e("day",1)])),yMMMMd:xe(be([_e("year",1),ve("month",4),_e("day",1)])),yMMMd:xe(be([_e("year",1),ve("month",3),_e("day",1)])),yMd:xe(be([_e("year",1),_e("month",1),_e("day",1)])),jms:xe(be([_e("hour",1),_e("second",1),_e("minute",1)])),jm:xe(be([_e("hour",1),_e("minute",1)]))},pe={yyyy:xe(_e("year",4)),yy:xe(_e("year",2)),y:xe(_e("year",1)),MMMM:xe(ve("month",4)),MMM:xe(ve("month",3)),MM:xe(_e("month",2)),M:xe(_e("month",1)),LLLL:xe(ve("month",4)),L:xe(ve("month",1)),dd:xe(_e("day",2)),d:xe(_e("day",1)),HH:fe(he(xe(ye(_e("hour",2),!1)))),H:he(xe(ye(_e("hour",1),!1))),hh:fe(he(xe(ye(_e("hour",2),!0)))),h:he(xe(ye(_e("hour",1),!0))),jj:xe(_e("hour",2)),j:xe(_e("hour",1)),mm:fe(xe(_e("minute",2))),m:xe(_e("minute",1)),ss:fe(xe(_e("second",2))),s:xe(_e("second",1)),sss:xe(_e("second",3)),EEEE:xe(ve("weekday",4)),EEE:xe(ve("weekday",3)),EE:xe(ve("weekday",2)),E:xe(ve("weekday",1)),a:(ue=xe(ye(_e("hour",1),!0)),function(t,e){return ue(t,e).split(" ")[1]}),Z:ge("short"),z:ge("long"),ww:xe({}),w:xe({}),G:xe(ve("era",1)),GG:xe(ve("era",2)),GGG:xe(ve("era",3)),GGGG:xe(ve("era",4))};function fe(t){return function(e,n){var r=t(e,n);return 1==r.length?"0"+r:r}}function he(t){return function(e,n){return t(e,n).split(" ")[0]}}function me(t,e,n){return new Intl.DateTimeFormat(e,n).format(t).replace(/[\u200e\u200f]/g,"")}function ge(t){var e={hour:"2-digit",hour12:!1,timeZoneName:t};return function(t,n){var r=me(t,n,e);return r?r.substring(3):""}}function ye(t,e){return t.hour12=e,t}function _e(t,e){var n={};return n[t]=2===e?"2-digit":"numeric",n}function ve(t,e){var n={};return n[t]=e<4?e>1?"short":"narrow":"long",n}function be(t){return t.reduce(function(t,e){return Object(i.__assign)({},t,e)},{})}function xe(t){return function(e,n){return me(e,n,t)}}var we=new Map;var je=function(){function t(){}return t.format=function(t,e,n){return function(t,e,n){var r=de[t];if(r)return r(e,n);var i=t,o=we.get(i);if(!o){o=[];var a=void 0;le.exec(t);for(var s=t;s;)(a=le.exec(s))?s=(o=o.concat(a.slice(1))).pop():(o.push(s),s=null);we.set(i,o)}return o.reduce(function(t,r){var i,o=pe[r];return t+(o?o(e,n):"''"===(i=r)?"'":i.replace(/(^'|'$)/g,"").replace(/''/g,"'"))},"")}(n,t,e)},t}(),Oe=function(){function t(t){this._locale=t}var e;return e=t,t.prototype.transform=function(t,n){if(void 0===n&&(n="mediumDate"),null==t||""===t||t!=t)return null;var r;if("string"==typeof t&&(t=t.trim()),Ce(t))r=t;else if(isNaN(t-parseFloat(t)))if("string"==typeof t&&/^(\d{4}-\d{1,2}-\d{1,2})$/.test(t)){var o=Object(i.__read)(t.split("-").map(function(t){return parseInt(t,10)}),3),a=o[0],s=o[1],u=o[2];r=new Date(a,s-1,u)}else r=new Date(t);else r=new Date(parseFloat(t));if(!Ce(r)){var c=void 0;if("string"!=typeof t||!(c=t.match(st)))throw se(e,t);r=xt(c)}return je.format(r,this._locale,e._ALIASES[n]||n)},t._ALIASES={medium:"yMMMdjms",short:"yMdjm",fullDate:"yMMMMEEEEd",longDate:"yMMMMd",mediumDate:"yMMMd",shortDate:"yMd",mediumTime:"jms",shortTime:"jm"},t=e=Object(i.__decorate)([Object(r.Pipe)({name:"date",pure:!0}),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__metadata)("design:paramtypes",[String])],t)}();function Ce(t){return t instanceof Date&&!isNaN(t.valueOf())}function Ee(t,e,n,r,i,o,a){if(void 0===o&&(o=null),void 0===a&&(a=!1),null==n)return null;if("number"!=typeof(n="string"!=typeof n||isNaN(+n-parseFloat(n))?n:+n))throw se(t,n);var s,u,c;if(r!==g.Currency&&(s=1,u=0,c=3),i){var l=i.match(jt);if(null===l)throw new Error(i+" is not a valid digit info for number pipes");null!=l[1]&&(s=Ft(l[1])),null!=l[3]&&(u=Ft(l[3])),null!=l[5]&&(c=Ft(l[5]))}return ce.format(n,e,r,{minimumIntegerDigits:s,minimumFractionDigits:u,maximumFractionDigits:c,currency:o,currencyAsSymbol:a})}var Se=function(){function t(t){this._locale=t}var e;return e=t,t.prototype.transform=function(t,n){return Ee(e,this._locale,t,g.Decimal,n)},t=e=Object(i.__decorate)([Object(r.Pipe)({name:"number"}),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__metadata)("design:paramtypes",[String])],t)}(),Me=function(){function t(t){this._locale=t}var e;return e=t,t.prototype.transform=function(t,n){return Ee(e,this._locale,t,g.Percent,n)},t=e=Object(i.__decorate)([Object(r.Pipe)({name:"percent"}),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__metadata)("design:paramtypes",[String])],t)}(),Te=function(){function t(t){this._locale=t}var e;return e=t,t.prototype.transform=function(t,n,r,i){return void 0===n&&(n="USD"),void 0===r&&(r=!1),Ee(e,this._locale,t,g.Currency,i,n,r)},t=e=Object(i.__decorate)([Object(r.Pipe)({name:"currency"}),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__metadata)("design:paramtypes",[String])],t)}(),ke=[Se,Me,Te,Oe],Ae=function(){function t(){}return t.prototype.createSubscription=function(t,e){return t.subscribe({next:e,error:function(t){throw t}})},t.prototype.dispose=function(t){t.unsubscribe()},t.prototype.onDestroy=function(t){t.unsubscribe()},t}(),De=new(function(){function t(){}return t.prototype.createSubscription=function(t,e){return t.then(e,function(t){throw t})},t.prototype.dispose=function(t){},t.prototype.onDestroy=function(t){},t}()),Ie=new Ae,Pe=function(){function t(t){this._ref=t,this._latestValue=null,this._latestReturnedValue=null,this._subscription=null,this._obj=null,this._strategy=null}var e;return e=t,t.prototype.ngOnDestroy=function(){this._subscription&&this._dispose()},t.prototype.transform=function(t){return this._obj?t!==this._obj?(this._dispose(),this.transform(t)):this._latestValue===this._latestReturnedValue?this._latestReturnedValue:(this._latestReturnedValue=this._latestValue,r.WrappedValue.wrap(this._latestValue)):(t&&this._subscribe(t),this._latestReturnedValue=this._latestValue,this._latestValue)},t.prototype._subscribe=function(t){var e=this;this._obj=t,this._strategy=this._selectStrategy(t),this._subscription=this._strategy.createSubscription(t,function(n){return e._updateLatestValue(t,n)})},t.prototype._selectStrategy=function(t){if(Object(r["ɵisPromise"])(t))return De;if(Object(r["ɵisObservable"])(t))return Ie;throw se(e,t)},t.prototype._dispose=function(){this._strategy.dispose(this._subscription),this._latestValue=null,this._latestReturnedValue=null,this._subscription=null,this._obj=null},t.prototype._updateLatestValue=function(t,e){t===this._obj&&(this._latestValue=e,this._ref.markForCheck())},t=e=Object(i.__decorate)([Object(r.Pipe)({name:"async",pure:!1}),Object(i.__metadata)("design:paramtypes",[r.ChangeDetectorRef])],t)}(),Re=function(){function t(){}var e;return e=t,t.prototype.transform=function(t){if(!t)return t;if("string"!=typeof t)throw se(e,t);return t.toLowerCase()},t=e=Object(i.__decorate)([Object(r.Pipe)({name:"lowercase"})],t)}(),Ne=/(?:[A-Za-z\xAA\xB5\xBA\xC0-\xD6\xD8-\xF6\xF8-\u02C1\u02C6-\u02D1\u02E0-\u02E4\u02EC\u02EE\u0370-\u0374\u0376\u0377\u037A-\u037D\u037F\u0386\u0388-\u038A\u038C\u038E-\u03A1\u03A3-\u03F5\u03F7-\u0481\u048A-\u052F\u0531-\u0556\u0559\u0561-\u0587\u05D0-\u05EA\u05F0-\u05F2\u0620-\u064A\u066E\u066F\u0671-\u06D3\u06D5\u06E5\u06E6\u06EE\u06EF\u06FA-\u06FC\u06FF\u0710\u0712-\u072F\u074D-\u07A5\u07B1\u07CA-\u07EA\u07F4\u07F5\u07FA\u0800-\u0815\u081A\u0824\u0828\u0840-\u0858\u0860-\u086A\u08A0-\u08B4\u08B6-\u08BD\u0904-\u0939\u093D\u0950\u0958-\u0961\u0971-\u0980\u0985-\u098C\u098F\u0990\u0993-\u09A8\u09AA-\u09B0\u09B2\u09B6-\u09B9\u09BD\u09CE\u09DC\u09DD\u09DF-\u09E1\u09F0\u09F1\u09FC\u0A05-\u0A0A\u0A0F\u0A10\u0A13-\u0A28\u0A2A-\u0A30\u0A32\u0A33\u0A35\u0A36\u0A38\u0A39\u0A59-\u0A5C\u0A5E\u0A72-\u0A74\u0A85-\u0A8D\u0A8F-\u0A91\u0A93-\u0AA8\u0AAA-\u0AB0\u0AB2\u0AB3\u0AB5-\u0AB9\u0ABD\u0AD0\u0AE0\u0AE1\u0AF9\u0B05-\u0B0C\u0B0F\u0B10\u0B13-\u0B28\u0B2A-\u0B30\u0B32\u0B33\u0B35-\u0B39\u0B3D\u0B5C\u0B5D\u0B5F-\u0B61\u0B71\u0B83\u0B85-\u0B8A\u0B8E-\u0B90\u0B92-\u0B95\u0B99\u0B9A\u0B9C\u0B9E\u0B9F\u0BA3\u0BA4\u0BA8-\u0BAA\u0BAE-\u0BB9\u0BD0\u0C05-\u0C0C\u0C0E-\u0C10\u0C12-\u0C28\u0C2A-\u0C39\u0C3D\u0C58-\u0C5A\u0C60\u0C61\u0C80\u0C85-\u0C8C\u0C8E-\u0C90\u0C92-\u0CA8\u0CAA-\u0CB3\u0CB5-\u0CB9\u0CBD\u0CDE\u0CE0\u0CE1\u0CF1\u0CF2\u0D05-\u0D0C\u0D0E-\u0D10\u0D12-\u0D3A\u0D3D\u0D4E\u0D54-\u0D56\u0D5F-\u0D61\u0D7A-\u0D7F\u0D85-\u0D96\u0D9A-\u0DB1\u0DB3-\u0DBB\u0DBD\u0DC0-\u0DC6\u0E01-\u0E30\u0E32\u0E33\u0E40-\u0E46\u0E81\u0E82\u0E84\u0E87\u0E88\u0E8A\u0E8D\u0E94-\u0E97\u0E99-\u0E9F\u0EA1-\u0EA3\u0EA5\u0EA7\u0EAA\u0EAB\u0EAD-\u0EB0\u0EB2\u0EB3\u0EBD\u0EC0-\u0EC4\u0EC6\u0EDC-\u0EDF\u0F00\u0F40-\u0F47\u0F49-\u0F6C\u0F88-\u0F8C\u1000-\u102A\u103F\u1050-\u1055\u105A-\u105D\u1061\u1065\u1066\u106E-\u1070\u1075-\u1081\u108E\u10A0-\u10C5\u10C7\u10CD\u10D0-\u10FA\u10FC-\u1248\u124A-\u124D\u1250-\u1256\u1258\u125A-\u125D\u1260-\u1288\u128A-\u128D\u1290-\u12B0\u12B2-\u12B5\u12B8-\u12BE\u12C0\u12C2-\u12C5\u12C8-\u12D6\u12D8-\u1310\u1312-\u1315\u1318-\u135A\u1380-\u138F\u13A0-\u13F5\u13F8-\u13FD\u1401-\u166C\u166F-\u167F\u1681-\u169A\u16A0-\u16EA\u16F1-\u16F8\u1700-\u170C\u170E-\u1711\u1720-\u1731\u1740-\u1751\u1760-\u176C\u176E-\u1770\u1780-\u17B3\u17D7\u17DC\u1820-\u1877\u1880-\u1884\u1887-\u18A8\u18AA\u18B0-\u18F5\u1900-\u191E\u1950-\u196D\u1970-\u1974\u1980-\u19AB\u19B0-\u19C9\u1A00-\u1A16\u1A20-\u1A54\u1AA7\u1B05-\u1B33\u1B45-\u1B4B\u1B83-\u1BA0\u1BAE\u1BAF\u1BBA-\u1BE5\u1C00-\u1C23\u1C4D-\u1C4F\u1C5A-\u1C7D\u1C80-\u1C88\u1CE9-\u1CEC\u1CEE-\u1CF1\u1CF5\u1CF6\u1D00-\u1DBF\u1E00-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC\u1FBE\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC\u1FF2-\u1FF4\u1FF6-\u1FFC\u2071\u207F\u2090-\u209C\u2102\u2107\u210A-\u2113\u2115\u2119-\u211D\u2124\u2126\u2128\u212A-\u212D\u212F-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2183\u2184\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CEE\u2CF2\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D80-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u2E2F\u3005\u3006\u3031-\u3035\u303B\u303C\u3041-\u3096\u309D-\u309F\u30A1-\u30FA\u30FC-\u30FF\u3105-\u312E\u3131-\u318E\u31A0-\u31BA\u31F0-\u31FF\u3400-\u4DB5\u4E00-\u9FEA\uA000-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA61F\uA62A\uA62B\uA640-\uA66E\uA67F-\uA69D\uA6A0-\uA6E5\uA717-\uA71F\uA722-\uA788\uA78B-\uA7AE\uA7B0-\uA7B7\uA7F7-\uA801\uA803-\uA805\uA807-\uA80A\uA80C-\uA822\uA840-\uA873\uA882-\uA8B3\uA8F2-\uA8F7\uA8FB\uA8FD\uA90A-\uA925\uA930-\uA946\uA960-\uA97C\uA984-\uA9B2\uA9CF\uA9E0-\uA9E4\uA9E6-\uA9EF\uA9FA-\uA9FE\uAA00-\uAA28\uAA40-\uAA42\uAA44-\uAA4B\uAA60-\uAA76\uAA7A\uAA7E-\uAAAF\uAAB1\uAAB5\uAAB6\uAAB9-\uAABD\uAAC0\uAAC2\uAADB-\uAADD\uAAE0-\uAAEA\uAAF2-\uAAF4\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uAB30-\uAB5A\uAB5C-\uAB65\uAB70-\uABE2\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFA6D\uFA70-\uFAD9\uFB00-\uFB06\uFB13-\uFB17\uFB1D\uFB1F-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40\uFB41\uFB43\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE70-\uFE74\uFE76-\uFEFC\uFF21-\uFF3A\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC]|\uD800[\uDC00-\uDC0B\uDC0D-\uDC26\uDC28-\uDC3A\uDC3C\uDC3D\uDC3F-\uDC4D\uDC50-\uDC5D\uDC80-\uDCFA\uDE80-\uDE9C\uDEA0-\uDED0\uDF00-\uDF1F\uDF2D-\uDF40\uDF42-\uDF49\uDF50-\uDF75\uDF80-\uDF9D\uDFA0-\uDFC3\uDFC8-\uDFCF]|\uD801[\uDC00-\uDC9D\uDCB0-\uDCD3\uDCD8-\uDCFB\uDD00-\uDD27\uDD30-\uDD63\uDE00-\uDF36\uDF40-\uDF55\uDF60-\uDF67]|\uD802[\uDC00-\uDC05\uDC08\uDC0A-\uDC35\uDC37\uDC38\uDC3C\uDC3F-\uDC55\uDC60-\uDC76\uDC80-\uDC9E\uDCE0-\uDCF2\uDCF4\uDCF5\uDD00-\uDD15\uDD20-\uDD39\uDD80-\uDDB7\uDDBE\uDDBF\uDE00\uDE10-\uDE13\uDE15-\uDE17\uDE19-\uDE33\uDE60-\uDE7C\uDE80-\uDE9C\uDEC0-\uDEC7\uDEC9-\uDEE4\uDF00-\uDF35\uDF40-\uDF55\uDF60-\uDF72\uDF80-\uDF91]|\uD803[\uDC00-\uDC48\uDC80-\uDCB2\uDCC0-\uDCF2]|\uD804[\uDC03-\uDC37\uDC83-\uDCAF\uDCD0-\uDCE8\uDD03-\uDD26\uDD50-\uDD72\uDD76\uDD83-\uDDB2\uDDC1-\uDDC4\uDDDA\uDDDC\uDE00-\uDE11\uDE13-\uDE2B\uDE80-\uDE86\uDE88\uDE8A-\uDE8D\uDE8F-\uDE9D\uDE9F-\uDEA8\uDEB0-\uDEDE\uDF05-\uDF0C\uDF0F\uDF10\uDF13-\uDF28\uDF2A-\uDF30\uDF32\uDF33\uDF35-\uDF39\uDF3D\uDF50\uDF5D-\uDF61]|\uD805[\uDC00-\uDC34\uDC47-\uDC4A\uDC80-\uDCAF\uDCC4\uDCC5\uDCC7\uDD80-\uDDAE\uDDD8-\uDDDB\uDE00-\uDE2F\uDE44\uDE80-\uDEAA\uDF00-\uDF19]|\uD806[\uDCA0-\uDCDF\uDCFF\uDE00\uDE0B-\uDE32\uDE3A\uDE50\uDE5C-\uDE83\uDE86-\uDE89\uDEC0-\uDEF8]|\uD807[\uDC00-\uDC08\uDC0A-\uDC2E\uDC40\uDC72-\uDC8F\uDD00-\uDD06\uDD08\uDD09\uDD0B-\uDD30\uDD46]|\uD808[\uDC00-\uDF99]|\uD809[\uDC80-\uDD43]|[\uD80C\uD81C-\uD820\uD840-\uD868\uD86A-\uD86C\uD86F-\uD872\uD874-\uD879][\uDC00-\uDFFF]|\uD80D[\uDC00-\uDC2E]|\uD811[\uDC00-\uDE46]|\uD81A[\uDC00-\uDE38\uDE40-\uDE5E\uDED0-\uDEED\uDF00-\uDF2F\uDF40-\uDF43\uDF63-\uDF77\uDF7D-\uDF8F]|\uD81B[\uDF00-\uDF44\uDF50\uDF93-\uDF9F\uDFE0\uDFE1]|\uD821[\uDC00-\uDFEC]|\uD822[\uDC00-\uDEF2]|\uD82C[\uDC00-\uDD1E\uDD70-\uDEFB]|\uD82F[\uDC00-\uDC6A\uDC70-\uDC7C\uDC80-\uDC88\uDC90-\uDC99]|\uD835[\uDC00-\uDC54\uDC56-\uDC9C\uDC9E\uDC9F\uDCA2\uDCA5\uDCA6\uDCA9-\uDCAC\uDCAE-\uDCB9\uDCBB\uDCBD-\uDCC3\uDCC5-\uDD05\uDD07-\uDD0A\uDD0D-\uDD14\uDD16-\uDD1C\uDD1E-\uDD39\uDD3B-\uDD3E\uDD40-\uDD44\uDD46\uDD4A-\uDD50\uDD52-\uDEA5\uDEA8-\uDEC0\uDEC2-\uDEDA\uDEDC-\uDEFA\uDEFC-\uDF14\uDF16-\uDF34\uDF36-\uDF4E\uDF50-\uDF6E\uDF70-\uDF88\uDF8A-\uDFA8\uDFAA-\uDFC2\uDFC4-\uDFCB]|\uD83A[\uDC00-\uDCC4\uDD00-\uDD43]|\uD83B[\uDE00-\uDE03\uDE05-\uDE1F\uDE21\uDE22\uDE24\uDE27\uDE29-\uDE32\uDE34-\uDE37\uDE39\uDE3B\uDE42\uDE47\uDE49\uDE4B\uDE4D-\uDE4F\uDE51\uDE52\uDE54\uDE57\uDE59\uDE5B\uDE5D\uDE5F\uDE61\uDE62\uDE64\uDE67-\uDE6A\uDE6C-\uDE72\uDE74-\uDE77\uDE79-\uDE7C\uDE7E\uDE80-\uDE89\uDE8B-\uDE9B\uDEA1-\uDEA3\uDEA5-\uDEA9\uDEAB-\uDEBB]|\uD869[\uDC00-\uDED6\uDF00-\uDFFF]|\uD86D[\uDC00-\uDF34\uDF40-\uDFFF]|\uD86E[\uDC00-\uDC1D\uDC20-\uDFFF]|\uD873[\uDC00-\uDEA1\uDEB0-\uDFFF]|\uD87A[\uDC00-\uDFE0]|\uD87E[\uDC00-\uDE1D])\S*/g,Fe=function(){function t(){}var e;return e=t,t.prototype.transform=function(t){if(!t)return t;if("string"!=typeof t)throw se(e,t);return t.replace(Ne,function(t){return t[0].toUpperCase()+t.substr(1).toLowerCase()})},t=e=Object(i.__decorate)([Object(r.Pipe)({name:"titlecase"})],t)}(),Le=function(){function t(){}var e;return e=t,t.prototype.transform=function(t){if(!t)return t;if("string"!=typeof t)throw se(e,t);return t.toUpperCase()},t=e=Object(i.__decorate)([Object(r.Pipe)({name:"uppercase"})],t)}(),Be=function(){function t(t){this.locale=t}var e;return e=t,t.prototype.transform=function(t,n,r,i){if(void 0===n&&(n="mediumDate"),null==t||""===t||t!=t)return null;try{return lt(t,n,i||this.locale,r)}catch(t){throw se(e,t.message)}},t=e=Object(i.__decorate)([Object(r.Pipe)({name:"date",pure:!0}),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__metadata)("design:paramtypes",[String])],t)}(),Ve=/#/g,ze=function(){function t(t){this._localization=t}var e;return e=t,t.prototype.transform=function(t,n,r){if(null==t)return"";if("object"!=typeof n||null===n)throw se(e,n);return n[Vt(t,Object.keys(n),this._localization,r)].replace(Ve,t.toString())},t=e=Object(i.__decorate)([Object(r.Pipe)({name:"i18nPlural",pure:!0}),Object(i.__metadata)("design:paramtypes",[Bt])],t)}(),He=function(){function t(){}var e;return e=t,t.prototype.transform=function(t,n){if(null==t)return"";if("object"!=typeof n||"string"!=typeof t)throw se(e,n);return n.hasOwnProperty(t)?n[t]:n.hasOwnProperty("other")?n.other:""},t=e=Object(i.__decorate)([Object(r.Pipe)({name:"i18nSelect",pure:!0})],t)}(),Ue=function(){function t(){}return t.prototype.transform=function(t){return JSON.stringify(t,null,2)},t=Object(i.__decorate)([Object(r.Pipe)({name:"json",pure:!1})],t)}();var We=function(){function t(t){this.differs=t,this.keyValues=[]}return t.prototype.transform=function(t,e){var n=this;if(void 0===e&&(e=qe),!t||!(t instanceof Map)&&"object"!=typeof t)return null;this.differ||(this.differ=this.differs.find(t).create());var r=this.differ.diff(t);return r&&(this.keyValues=[],r.forEachItem(function(t){var e,r;n.keyValues.push((e=t.key,r=t.currentValue,{key:e,value:r}))}),this.keyValues.sort(e)),this.keyValues},t=Object(i.__decorate)([Object(r.Pipe)({name:"keyvalue",pure:!1}),Object(i.__metadata)("design:paramtypes",[r.KeyValueDiffers])],t)}();function qe(t,e){var n=t.key,r=e.key;if(n===r)return 0;if(void 0===n)return 1;if(void 0===r)return-1;if(null===n)return 1;if(null===r)return-1;if("string"==typeof n&&"string"==typeof r)return n<r?-1:1;if("number"==typeof n&&"number"==typeof r)return n-r;if("boolean"==typeof n&&"boolean"==typeof r)return n<r?-1:1;var i=String(n),o=String(r);return i==o?0:i<o?-1:1}var Ge=function(){function t(t){this._locale=t}var e;return e=t,t.prototype.transform=function(t,n,r){if(Xe(t))return null;r=r||this._locale;try{return Rt(Qe(t),r,n)}catch(t){throw se(e,t.message)}},t=e=Object(i.__decorate)([Object(r.Pipe)({name:"number"}),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__metadata)("design:paramtypes",[String])],t)}(),Ye=function(){function t(t){this._locale=t}var e;return e=t,t.prototype.transform=function(t,n,r){if(Xe(t))return null;r=r||this._locale;try{return Pt(Qe(t),r,n)}catch(t){throw se(e,t.message)}},t=e=Object(i.__decorate)([Object(r.Pipe)({name:"percent"}),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__metadata)("design:paramtypes",[String])],t)}(),Ke=function(){function t(t){this._locale=t}var e;return e=t,t.prototype.transform=function(t,n,r,i,o){if(void 0===r&&(r="symbol"),Xe(t))return null;o=o||this._locale,"boolean"==typeof r&&(console&&console.warn&&console.warn('Warning: the currency pipe has been changed in Angular v5. The symbolDisplay option (third parameter) is now a string instead of a boolean. The accepted values are "code", "symbol" or "symbol-narrow".'),r=r?"symbol":"code");var a=n||"USD";"code"!==r&&(a="symbol"===r||"symbol-narrow"===r?$(a,"symbol"===r?"wide":"narrow",o):r);try{return It(Qe(t),o,a,n,i)}catch(t){throw se(e,t.message)}},t=e=Object(i.__decorate)([Object(r.Pipe)({name:"currency"}),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__metadata)("design:paramtypes",[String])],t)}();function Xe(t){return null==t||""===t||t!=t}function Qe(t){if("string"==typeof t&&!isNaN(Number(t)-parseFloat(t)))return Number(t);if("number"!=typeof t)throw new Error(t+" is not a number");return t}var Ze=function(){function t(){}var e;return e=t,t.prototype.transform=function(t,n,r){if(null==t)return t;if(!this.supports(t))throw se(e,t);return t.slice(n,r)},t.prototype.supports=function(t){return"string"==typeof t||Array.isArray(t)},t=e=Object(i.__decorate)([Object(r.Pipe)({name:"slice",pure:!1})],t)}(),$e=[Pe,Le,Re,Ue,Ze,Ge,Ye,Fe,Ke,Be,ze,He,We],Je=function(){function t(){}return t=Object(i.__decorate)([Object(r.NgModule)({declarations:[ae,$e],exports:[ae,$e],providers:[{provide:Bt,useClass:zt}]})],t)}(),tn=Ht,en=function(){function t(){}return t=Object(i.__decorate)([Object(r.NgModule)({declarations:[ke],exports:[ke],providers:[{provide:Lt,useValue:tn}]})],t)}(),nn=new r.InjectionToken("DocumentToken"),rn="browser",on="server",an="browserWorkerApp",sn="browserWorkerUi";function un(t){return t===rn}function cn(t){return t===on}function ln(t){return t===an}function dn(t){return t===sn}var pn=new r.Version("7.2.0"),fn=function(){function t(){}return t.ngInjectableDef=Object(r.defineInjectable)({providedIn:"root",factory:function(){return new hn(Object(r.inject)(nn),window)}}),t}(),hn=function(){function t(t,e){this.document=t,this.window=e,this.offset=function(){return[0,0]}}return t.prototype.setOffset=function(t){Array.isArray(t)?this.offset=function(){return t}:this.offset=t},t.prototype.getScrollPosition=function(){return this.supportScrollRestoration()?[this.window.scrollX,this.window.scrollY]:[0,0]},t.prototype.scrollToPosition=function(t){this.supportScrollRestoration()&&this.window.scrollTo(t[0],t[1])},t.prototype.scrollToAnchor=function(t){if(this.supportScrollRestoration()){var e=this.document.querySelector("#"+t);if(e)return void this.scrollToElement(e);var n=this.document.querySelector("[name='"+t+"']");if(n)return void this.scrollToElement(n)}},t.prototype.setHistoryScrollRestoration=function(t){if(this.supportScrollRestoration()){var e=this.window.history;e&&e.scrollRestoration&&(e.scrollRestoration=t)}},t.prototype.scrollToElement=function(t){var e=t.getBoundingClientRect(),n=e.left+this.window.pageXOffset,r=e.top+this.window.pageYOffset,i=this.offset();this.window.scrollTo(n-i[0],r-i[1])},t.prototype.supportScrollRestoration=function(){try{return!!this.window&&!!this.window.scrollTo}catch(t){return!1}},t}(),mn=function(){function t(){}return t.prototype.setOffset=function(t){},t.prototype.getScrollPosition=function(){return[0,0]},t.prototype.scrollToPosition=function(t){},t.prototype.scrollToAnchor=function(t){},t.prototype.setHistoryScrollRestoration=function(t){},t}()},"./node_modules/@angular/common/fesm5/http.js":function(t,e,n){"use strict";n.r(e),n.d(e,"ɵangular_packages_common_http_http_a",function(){return M}),n.d(e,"ɵangular_packages_common_http_http_b",function(){return k}),n.d(e,"ɵangular_packages_common_http_http_c",function(){return U}),n.d(e,"ɵangular_packages_common_http_http_d",function(){return R}),n.d(e,"ɵangular_packages_common_http_http_g",function(){return V}),n.d(e,"ɵangular_packages_common_http_http_h",function(){return z}),n.d(e,"ɵangular_packages_common_http_http_e",function(){return F}),n.d(e,"ɵangular_packages_common_http_http_f",function(){return L}),n.d(e,"HttpBackend",function(){return c}),n.d(e,"HttpHandler",function(){return u}),n.d(e,"HttpClient",function(){return C}),n.d(e,"HttpHeaders",function(){return l}),n.d(e,"HTTP_INTERCEPTORS",function(){return S}),n.d(e,"JsonpClientBackend",function(){return A}),n.d(e,"JsonpInterceptor",function(){return D}),n.d(e,"HttpClientJsonpModule",function(){return G}),n.d(e,"HttpClientModule",function(){return q}),n.d(e,"HttpClientXsrfModule",function(){return W}),n.d(e,"ɵHttpInterceptingHandler",function(){return H}),n.d(e,"HttpParams",function(){return f}),n.d(e,"HttpUrlEncodingCodec",function(){return d}),n.d(e,"HttpRequest",function(){return v}),n.d(e,"HttpErrorResponse",function(){return j}),n.d(e,"HttpEventType",function(){return y}),n.d(e,"HttpHeaderResponse",function(){return x}),n.d(e,"HttpResponse",function(){return w}),n.d(e,"HttpResponseBase",function(){return b}),n.d(e,"HttpXhrBackend",function(){return N}),n.d(e,"XhrFactory",function(){return P}),n.d(e,"HttpXsrfTokenExtractor",function(){return B});var r=n("./node_modules/@angular/common/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/rxjs/_esm5/index.js"),a=n("./node_modules/rxjs/_esm5/operators/index.js"),s=n("./node_modules/@angular/common/fesm5/common.js"),u=function(){return function(){}}(),c=function(){return function(){}}(),l=function(){function t(t){var e=this;this.normalizedNames=new Map,this.lazyUpdate=null,t?this.lazyInit="string"==typeof t?function(){e.headers=new Map,t.split("\n").forEach(function(t){var n=t.indexOf(":");if(n>0){var r=t.slice(0,n),i=r.toLowerCase(),o=t.slice(n+1).trim();e.maybeSetNormalizedName(r,i),e.headers.has(i)?e.headers.get(i).push(o):e.headers.set(i,[o])}})}:function(){e.headers=new Map,Object.keys(t).forEach(function(n){var r=t[n],i=n.toLowerCase();"string"==typeof r&&(r=[r]),r.length>0&&(e.headers.set(i,r),e.maybeSetNormalizedName(n,i))})}:this.headers=new Map}return t.prototype.has=function(t){return this.init(),this.headers.has(t.toLowerCase())},t.prototype.get=function(t){this.init();var e=this.headers.get(t.toLowerCase());return e&&e.length>0?e[0]:null},t.prototype.keys=function(){return this.init(),Array.from(this.normalizedNames.values())},t.prototype.getAll=function(t){return this.init(),this.headers.get(t.toLowerCase())||null},t.prototype.append=function(t,e){return this.clone({name:t,value:e,op:"a"})},t.prototype.set=function(t,e){return this.clone({name:t,value:e,op:"s"})},t.prototype.delete=function(t,e){return this.clone({name:t,value:e,op:"d"})},t.prototype.maybeSetNormalizedName=function(t,e){this.normalizedNames.has(e)||this.normalizedNames.set(e,t)},t.prototype.init=function(){var e=this;this.lazyInit&&(this.lazyInit instanceof t?this.copyFrom(this.lazyInit):this.lazyInit(),this.lazyInit=null,this.lazyUpdate&&(this.lazyUpdate.forEach(function(t){return e.applyUpdate(t)}),this.lazyUpdate=null))},t.prototype.copyFrom=function(t){var e=this;t.init(),Array.from(t.headers.keys()).forEach(function(n){e.headers.set(n,t.headers.get(n)),e.normalizedNames.set(n,t.normalizedNames.get(n))})},t.prototype.clone=function(e){var n=new t;return n.lazyInit=this.lazyInit&&this.lazyInit instanceof t?this.lazyInit:this,n.lazyUpdate=(this.lazyUpdate||[]).concat([e]),n},t.prototype.applyUpdate=function(t){var e=t.name.toLowerCase();switch(t.op){case"a":case"s":var n=t.value;if("string"==typeof n&&(n=[n]),0===n.length)return;this.maybeSetNormalizedName(t.name,e);var i=("a"===t.op?this.headers.get(e):void 0)||[];i.push.apply(i,Object(r.__spread)(n)),this.headers.set(e,i);break;case"d":var o=t.value;if(o){var a=this.headers.get(e);if(!a)return;0===(a=a.filter(function(t){return-1===o.indexOf(t)})).length?(this.headers.delete(e),this.normalizedNames.delete(e)):this.headers.set(e,a)}else this.headers.delete(e),this.normalizedNames.delete(e)}},t.prototype.forEach=function(t){var e=this;this.init(),Array.from(this.normalizedNames.keys()).forEach(function(n){return t(e.normalizedNames.get(n),e.headers.get(n))})},t}(),d=function(){function t(){}return t.prototype.encodeKey=function(t){return p(t)},t.prototype.encodeValue=function(t){return p(t)},t.prototype.decodeKey=function(t){return decodeURIComponent(t)},t.prototype.decodeValue=function(t){return decodeURIComponent(t)},t}();function p(t){return encodeURIComponent(t).replace(/%40/gi,"@").replace(/%3A/gi,":").replace(/%24/gi,"$").replace(/%2C/gi,",").replace(/%3B/gi,";").replace(/%2B/gi,"+").replace(/%3D/gi,"=").replace(/%3F/gi,"?").replace(/%2F/gi,"/")}var f=function(){function t(t){void 0===t&&(t={});var e,n,i,o=this;if(this.updates=null,this.cloneFrom=null,this.encoder=t.encoder||new d,t.fromString){if(t.fromObject)throw new Error("Cannot specify both fromString and fromObject.");this.map=(e=t.fromString,n=this.encoder,i=new Map,e.length>0&&e.split("&").forEach(function(t){var e=t.indexOf("="),o=Object(r.__read)(-1==e?[n.decodeKey(t),""]:[n.decodeKey(t.slice(0,e)),n.decodeValue(t.slice(e+1))],2),a=o[0],s=o[1],u=i.get(a)||[];u.push(s),i.set(a,u)}),i)}else t.fromObject?(this.map=new Map,Object.keys(t.fromObject).forEach(function(e){var n=t.fromObject[e];o.map.set(e,Array.isArray(n)?n:[n])})):this.map=null}return t.prototype.has=function(t){return this.init(),this.map.has(t)},t.prototype.get=function(t){this.init();var e=this.map.get(t);return e?e[0]:null},t.prototype.getAll=function(t){return this.init(),this.map.get(t)||null},t.prototype.keys=function(){return this.init(),Array.from(this.map.keys())},t.prototype.append=function(t,e){return this.clone({param:t,value:e,op:"a"})},t.prototype.set=function(t,e){return this.clone({param:t,value:e,op:"s"})},t.prototype.delete=function(t,e){return this.clone({param:t,value:e,op:"d"})},t.prototype.toString=function(){var t=this;return this.init(),this.keys().map(function(e){var n=t.encoder.encodeKey(e);return t.map.get(e).map(function(e){return n+"="+t.encoder.encodeValue(e)}).join("&")}).join("&")},t.prototype.clone=function(e){var n=new t({encoder:this.encoder});return n.cloneFrom=this.cloneFrom||this,n.updates=(this.updates||[]).concat([e]),n},t.prototype.init=function(){var t=this;null===this.map&&(this.map=new Map),null!==this.cloneFrom&&(this.cloneFrom.init(),this.cloneFrom.keys().forEach(function(e){return t.map.set(e,t.cloneFrom.map.get(e))}),this.updates.forEach(function(e){switch(e.op){case"a":case"s":var n=("a"===e.op?t.map.get(e.param):void 0)||[];n.push(e.value),t.map.set(e.param,n);break;case"d":if(void 0===e.value){t.map.delete(e.param);break}var r=t.map.get(e.param)||[],i=r.indexOf(e.value);-1!==i&&r.splice(i,1),r.length>0?t.map.set(e.param,r):t.map.delete(e.param)}}),this.cloneFrom=null)},t}();function h(t){return"undefined"!=typeof ArrayBuffer&&t instanceof ArrayBuffer}function m(t){return"undefined"!=typeof Blob&&t instanceof Blob}function g(t){return"undefined"!=typeof FormData&&t instanceof FormData}var y,_,v=function(){function t(t,e,n,r){var i;if(this.url=e,this.body=null,this.reportProgress=!1,this.withCredentials=!1,this.responseType="json",this.method=t.toUpperCase(),function(t){switch(t){case"DELETE":case"GET":case"HEAD":case"OPTIONS":case"JSONP":return!1;default:return!0}}(this.method)||r?(this.body=void 0!==n?n:null,i=r):i=n,i&&(this.reportProgress=!!i.reportProgress,this.withCredentials=!!i.withCredentials,i.responseType&&(this.responseType=i.responseType),i.headers&&(this.headers=i.headers),i.params&&(this.params=i.params)),this.headers||(this.headers=new l),this.params){var o=this.params.toString();if(0===o.length)this.urlWithParams=e;else{var a=e.indexOf("?"),s=-1===a?"?":a<e.length-1?"&":"";this.urlWithParams=e+s+o}}else this.params=new f,this.urlWithParams=e}return t.prototype.serializeBody=function(){return null===this.body?null:h(this.body)||m(this.body)||g(this.body)||"string"==typeof this.body?this.body:this.body instanceof f?this.body.toString():"object"==typeof this.body||"boolean"==typeof this.body||Array.isArray(this.body)?JSON.stringify(this.body):this.body.toString()},t.prototype.detectContentTypeHeader=function(){return null===this.body?null:g(this.body)?null:m(this.body)?this.body.type||null:h(this.body)?null:"string"==typeof this.body?"text/plain":this.body instanceof f?"application/x-www-form-urlencoded;charset=UTF-8":"object"==typeof this.body||"number"==typeof this.body||Array.isArray(this.body)?"application/json":null},t.prototype.clone=function(e){void 0===e&&(e={});var n=e.method||this.method,r=e.url||this.url,i=e.responseType||this.responseType,o=void 0!==e.body?e.body:this.body,a=void 0!==e.withCredentials?e.withCredentials:this.withCredentials,s=void 0!==e.reportProgress?e.reportProgress:this.reportProgress,u=e.headers||this.headers,c=e.params||this.params;return void 0!==e.setHeaders&&(u=Object.keys(e.setHeaders).reduce(function(t,n){return t.set(n,e.setHeaders[n])},u)),e.setParams&&(c=Object.keys(e.setParams).reduce(function(t,n){return t.set(n,e.setParams[n])},c)),new t(n,r,o,{params:c,headers:u,reportProgress:s,responseType:i,withCredentials:a})},t}();(_=y||(y={}))[_.Sent=0]="Sent",_[_.UploadProgress=1]="UploadProgress",_[_.ResponseHeader=2]="ResponseHeader",_[_.DownloadProgress=3]="DownloadProgress",_[_.Response=4]="Response",_[_.User=5]="User";var b=function(){return function(t,e,n){void 0===e&&(e=200),void 0===n&&(n="OK"),this.headers=t.headers||new l,this.status=void 0!==t.status?t.status:e,this.statusText=t.statusText||n,this.url=t.url||null,this.ok=this.status>=200&&this.status<300}}(),x=function(t){function e(e){void 0===e&&(e={});var n=t.call(this,e)||this;return n.type=y.ResponseHeader,n}return Object(r.__extends)(e,t),e.prototype.clone=function(t){return void 0===t&&(t={}),new e({headers:t.headers||this.headers,status:void 0!==t.status?t.status:this.status,statusText:t.statusText||this.statusText,url:t.url||this.url||void 0})},e}(b),w=function(t){function e(e){void 0===e&&(e={});var n=t.call(this,e)||this;return n.type=y.Response,n.body=void 0!==e.body?e.body:null,n}return Object(r.__extends)(e,t),e.prototype.clone=function(t){return void 0===t&&(t={}),new e({body:void 0!==t.body?t.body:this.body,headers:t.headers||this.headers,status:void 0!==t.status?t.status:this.status,statusText:t.statusText||this.statusText,url:t.url||this.url||void 0})},e}(b),j=function(t){function e(e){var n=t.call(this,e,0,"Unknown Error")||this;return n.name="HttpErrorResponse",n.ok=!1,n.status>=200&&n.status<300?n.message="Http failure during parsing for "+(e.url||"(unknown url)"):n.message="Http failure response for "+(e.url||"(unknown url)")+": "+e.status+" "+e.statusText,n.error=e.error||null,n}return Object(r.__extends)(e,t),e}(b);function O(t,e){return{body:e,headers:t.headers,observe:t.observe,params:t.params,reportProgress:t.reportProgress,responseType:t.responseType,withCredentials:t.withCredentials}}var C=function(){function t(t){this.handler=t}return t.prototype.request=function(t,e,n){var r,i=this;if(void 0===n&&(n={}),t instanceof v)r=t;else{var s=void 0;s=n.headers instanceof l?n.headers:new l(n.headers);var u=void 0;n.params&&(u=n.params instanceof f?n.params:new f({fromObject:n.params})),r=new v(t,e,void 0!==n.body?n.body:null,{headers:s,params:u,reportProgress:n.reportProgress,responseType:n.responseType||"json",withCredentials:n.withCredentials})}var c=Object(o.of)(r).pipe(Object(a.concatMap)(function(t){return i.handler.handle(t)}));if(t instanceof v||"events"===n.observe)return c;var d=c.pipe(Object(a.filter)(function(t){return t instanceof w}));switch(n.observe||"body"){case"body":switch(r.responseType){case"arraybuffer":return d.pipe(Object(a.map)(function(t){if(null!==t.body&&!(t.body instanceof ArrayBuffer))throw new Error("Response is not an ArrayBuffer.");return t.body}));case"blob":return d.pipe(Object(a.map)(function(t){if(null!==t.body&&!(t.body instanceof Blob))throw new Error("Response is not a Blob.");return t.body}));case"text":return d.pipe(Object(a.map)(function(t){if(null!==t.body&&"string"!=typeof t.body)throw new Error("Response is not a string.");return t.body}));case"json":default:return d.pipe(Object(a.map)(function(t){return t.body}))}case"response":return d;default:throw new Error("Unreachable: unhandled observe type "+n.observe+"}")}},t.prototype.delete=function(t,e){return void 0===e&&(e={}),this.request("DELETE",t,e)},t.prototype.get=function(t,e){return void 0===e&&(e={}),this.request("GET",t,e)},t.prototype.head=function(t,e){return void 0===e&&(e={}),this.request("HEAD",t,e)},t.prototype.jsonp=function(t,e){return this.request("JSONP",t,{params:(new f).append(e,"JSONP_CALLBACK"),observe:"body",responseType:"json"})},t.prototype.options=function(t,e){return void 0===e&&(e={}),this.request("OPTIONS",t,e)},t.prototype.patch=function(t,e,n){return void 0===n&&(n={}),this.request("PATCH",t,O(n,e))},t.prototype.post=function(t,e,n){return void 0===n&&(n={}),this.request("POST",t,O(n,e))},t.prototype.put=function(t,e,n){return void 0===n&&(n={}),this.request("PUT",t,O(n,e))},t=Object(r.__decorate)([Object(i.Injectable)(),Object(r.__metadata)("design:paramtypes",[u])],t)}(),E=function(){function t(t,e){this.next=t,this.interceptor=e}return t.prototype.handle=function(t){return this.interceptor.intercept(t,this.next)},t}(),S=new i.InjectionToken("HTTP_INTERCEPTORS"),M=function(){function t(){}return t.prototype.intercept=function(t,e){return e.handle(t)},t=Object(r.__decorate)([Object(i.Injectable)()],t)}(),T=0,k=function(){return function(){}}(),A=function(){function t(t,e){this.callbackMap=t,this.document=e}return t.prototype.nextCallback=function(){return"ng_jsonp_callback_"+T++},t.prototype.handle=function(t){var e=this;if("JSONP"!==t.method)throw new Error("JSONP requests must use JSONP request method.");if("json"!==t.responseType)throw new Error("JSONP requests must use Json response type.");return new o.Observable(function(n){var r=e.nextCallback(),i=t.urlWithParams.replace(/=JSONP_CALLBACK(&|$)/,"="+r+"$1"),o=e.document.createElement("script");o.src=i;var a=null,s=!1,u=!1;e.callbackMap[r]=function(t){delete e.callbackMap[r],u||(a=t,s=!0)};var c=function(){o.parentNode&&o.parentNode.removeChild(o),delete e.callbackMap[r]},l=function(t){u||(c(),s?(n.next(new w({body:a,status:200,statusText:"OK",url:i})),n.complete()):n.error(new j({url:i,status:0,statusText:"JSONP Error",error:new Error("JSONP injected script did not invoke callback.")})))},d=function(t){u||(c(),n.error(new j({error:t,status:0,statusText:"JSONP Error",url:i})))};return o.addEventListener("load",l),o.addEventListener("error",d),e.document.body.appendChild(o),n.next({type:y.Sent}),function(){u=!0,o.removeEventListener("load",l),o.removeEventListener("error",d),c()}})},t=Object(r.__decorate)([Object(i.Injectable)(),Object(r.__param)(1,Object(i.Inject)(s.DOCUMENT)),Object(r.__metadata)("design:paramtypes",[k,Object])],t)}(),D=function(){function t(t){this.jsonp=t}return t.prototype.intercept=function(t,e){return"JSONP"===t.method?this.jsonp.handle(t):e.handle(t)},t=Object(r.__decorate)([Object(i.Injectable)(),Object(r.__metadata)("design:paramtypes",[A])],t)}(),I=/^\)\]\}',?\n/;var P=function(){return function(){}}(),R=function(){function t(){}return t.prototype.build=function(){return new XMLHttpRequest},t=Object(r.__decorate)([Object(i.Injectable)(),Object(r.__metadata)("design:paramtypes",[])],t)}(),N=function(){function t(t){this.xhrFactory=t}return t.prototype.handle=function(t){var e=this;if("JSONP"===t.method)throw new Error("Attempted to construct Jsonp request without JsonpClientModule installed.");return new o.Observable(function(n){var r=e.xhrFactory.build();if(r.open(t.method,t.urlWithParams),t.withCredentials&&(r.withCredentials=!0),t.headers.forEach(function(t,e){return r.setRequestHeader(t,e.join(","))}),t.headers.has("Accept")||r.setRequestHeader("Accept","application/json, text/plain, */*"),!t.headers.has("Content-Type")){var i=t.detectContentTypeHeader();null!==i&&r.setRequestHeader("Content-Type",i)}if(t.responseType){var o=t.responseType.toLowerCase();r.responseType="json"!==o?o:"text"}var a=t.serializeBody(),s=null,u=function(){if(null!==s)return s;var e,n=1223===r.status?204:r.status,i=r.statusText||"OK",o=new l(r.getAllResponseHeaders()),a=("responseURL"in(e=r)&&e.responseURL?e.responseURL:/^X-Request-URL:/m.test(e.getAllResponseHeaders())?e.getResponseHeader("X-Request-URL"):null)||t.url;return s=new x({headers:o,status:n,statusText:i,url:a})},c=function(){var e=u(),i=e.headers,o=e.status,a=e.statusText,s=e.url,c=null;204!==o&&(c=void 0===r.response?r.responseText:r.response),0===o&&(o=c?200:0);var l=o>=200&&o<300;if("json"===t.responseType&&"string"==typeof c){var d=c;c=c.replace(I,"");try{c=""!==c?JSON.parse(c):null}catch(t){c=d,l&&(l=!1,c={error:t,text:c})}}l?(n.next(new w({body:c,headers:i,status:o,statusText:a,url:s||void 0})),n.complete()):n.error(new j({error:c,headers:i,status:o,statusText:a,url:s||void 0}))},d=function(t){var e=u().url,i=new j({error:t,status:r.status||0,statusText:r.statusText||"Unknown Error",url:e||void 0});n.error(i)},p=!1,f=function(e){p||(n.next(u()),p=!0);var i={type:y.DownloadProgress,loaded:e.loaded};e.lengthComputable&&(i.total=e.total),"text"===t.responseType&&r.responseText&&(i.partialText=r.responseText),n.next(i)},h=function(t){var e={type:y.UploadProgress,loaded:t.loaded};t.lengthComputable&&(e.total=t.total),n.next(e)};return r.addEventListener("load",c),r.addEventListener("error",d),t.reportProgress&&(r.addEventListener("progress",f),null!==a&&r.upload&&r.upload.addEventListener("progress",h)),r.send(a),n.next({type:y.Sent}),function(){r.removeEventListener("error",d),r.removeEventListener("load",c),t.reportProgress&&(r.removeEventListener("progress",f),null!==a&&r.upload&&r.upload.removeEventListener("progress",h)),r.abort()}})},t=Object(r.__decorate)([Object(i.Injectable)(),Object(r.__metadata)("design:paramtypes",[P])],t)}(),F=new i.InjectionToken("XSRF_COOKIE_NAME"),L=new i.InjectionToken("XSRF_HEADER_NAME"),B=function(){return function(){}}(),V=function(){function t(t,e,n){this.doc=t,this.platform=e,this.cookieName=n,this.lastCookieString="",this.lastToken=null,this.parseCount=0}return t.prototype.getToken=function(){if("server"===this.platform)return null;var t=this.doc.cookie||"";return t!==this.lastCookieString&&(this.parseCount++,this.lastToken=Object(s["ɵparseCookieValue"])(t,this.cookieName),this.lastCookieString=t),this.lastToken},t=Object(r.__decorate)([Object(i.Injectable)(),Object(r.__param)(0,Object(i.Inject)(s.DOCUMENT)),Object(r.__param)(1,Object(i.Inject)(i.PLATFORM_ID)),Object(r.__param)(2,Object(i.Inject)(F)),Object(r.__metadata)("design:paramtypes",[Object,String,String])],t)}(),z=function(){function t(t,e){this.tokenService=t,this.headerName=e}return t.prototype.intercept=function(t,e){var n=t.url.toLowerCase();if("GET"===t.method||"HEAD"===t.method||n.startsWith("http://")||n.startsWith("https://"))return e.handle(t);var r=this.tokenService.getToken();return null===r||t.headers.has(this.headerName)||(t=t.clone({headers:t.headers.set(this.headerName,r)})),e.handle(t)},t=Object(r.__decorate)([Object(i.Injectable)(),Object(r.__param)(1,Object(i.Inject)(L)),Object(r.__metadata)("design:paramtypes",[B,String])],t)}(),H=function(){function t(t,e){this.backend=t,this.injector=e,this.chain=null}return t.prototype.handle=function(t){if(null===this.chain){var e=this.injector.get(S,[]);this.chain=e.reduceRight(function(t,e){return new E(t,e)},this.backend)}return this.chain.handle(t)},t=Object(r.__decorate)([Object(i.Injectable)(),Object(r.__metadata)("design:paramtypes",[c,i.Injector])],t)}();function U(){return"object"==typeof window?window:{}}var W=function(){function t(){}var e;return e=t,t.disable=function(){return{ngModule:e,providers:[{provide:z,useClass:M}]}},t.withOptions=function(t){return void 0===t&&(t={}),{ngModule:e,providers:[t.cookieName?{provide:F,useValue:t.cookieName}:[],t.headerName?{provide:L,useValue:t.headerName}:[]]}},t=e=Object(r.__decorate)([Object(i.NgModule)({providers:[z,{provide:S,useExisting:z,multi:!0},{provide:B,useClass:V},{provide:F,useValue:"XSRF-TOKEN"},{provide:L,useValue:"X-XSRF-TOKEN"}]})],t)}(),q=function(){function t(){}return t=Object(r.__decorate)([Object(i.NgModule)({imports:[W.withOptions({cookieName:"XSRF-TOKEN",headerName:"X-XSRF-TOKEN"})],providers:[C,{provide:u,useClass:H},N,{provide:c,useExisting:N},R,{provide:P,useExisting:R}]})],t)}(),G=function(){function t(){}return t=Object(r.__decorate)([Object(i.NgModule)({providers:[A,{provide:k,useFactory:U},{provide:S,useClass:D,multi:!0}]})],t)}()},"./node_modules/@angular/common/node_modules/tslib/tslib.es6.js":function(t,e,n){"use strict";n.r(e),n.d(e,"__extends",function(){return i}),n.d(e,"__assign",function(){return o}),n.d(e,"__rest",function(){return a}),n.d(e,"__decorate",function(){return s}),n.d(e,"__param",function(){return u}),n.d(e,"__metadata",function(){return c}),n.d(e,"__awaiter",function(){return l}),n.d(e,"__generator",function(){return d}),n.d(e,"__exportStar",function(){return p}),n.d(e,"__values",function(){return f}),n.d(e,"__read",function(){return h}),n.d(e,"__spread",function(){return m}),n.d(e,"__await",function(){return g}),n.d(e,"__asyncGenerator",function(){return y}),n.d(e,"__asyncDelegator",function(){return _}),n.d(e,"__asyncValues",function(){return v}),n.d(e,"__makeTemplateObject",function(){return b}),n.d(e,"__importStar",function(){return x}),n.d(e,"__importDefault",function(){return w});var r=function(t,e){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)};function i(t,e){function n(){this.constructor=t}r(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}var o=function(){return(o=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++){e=arguments[n];for(var i in e)Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i])}return t}).apply(this,arguments)};function a(t,e){var n={};for(var r in t)Object.prototype.hasOwnProperty.call(t,r)&&e.indexOf(r)<0&&(n[r]=t[r]);if(null!=t&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(t);i<r.length;i++)e.indexOf(r[i])<0&&(n[r[i]]=t[r[i]])}return n}function s(t,e,n,r){var i,o=arguments.length,a=o<3?e:null===r?r=Object.getOwnPropertyDescriptor(e,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(t,e,n,r);else for(var s=t.length-1;s>=0;s--)(i=t[s])&&(a=(o<3?i(a):o>3?i(e,n,a):i(e,n))||a);return o>3&&a&&Object.defineProperty(e,n,a),a}function u(t,e){return function(n,r){e(n,r,t)}}function c(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)}function l(t,e,n,r){return new(n||(n=Promise))(function(i,o){function a(t){try{u(r.next(t))}catch(t){o(t)}}function s(t){try{u(r.throw(t))}catch(t){o(t)}}function u(t){t.done?i(t.value):new n(function(e){e(t.value)}).then(a,s)}u((r=r.apply(t,e||[])).next())})}function d(t,e){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=e.call(t,a)}catch(t){o=[6,t],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(t,e){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}function f(t){var e="function"==typeof Symbol&&t[Symbol.iterator],n=0;return e?e.call(t):{next:function(){return t&&n>=t.length&&(t=void 0),{value:t&&t[n++],done:!t}}}}function h(t,e){var n="function"==typeof Symbol&&t[Symbol.iterator];if(!n)return t;var r,i,o=n.call(t),a=[];try{for(;(void 0===e||e-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(t){i={error:t}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var t=[],e=0;e<arguments.length;e++)t=t.concat(h(arguments[e]));return t}function g(t){return this instanceof g?(this.v=t,this):new g(t)}function y(t,e,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(t,e||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(t){i[t]&&(r[t]=function(e){return new Promise(function(n,r){o.push([t,e,n,r])>1||s(t,e)})})}function s(t,e){try{(n=i[t](e)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(t){l(o[0][3],t)}var n}function u(t){s("next",t)}function c(t){s("throw",t)}function l(t,e){t(e),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(t){var e,n;return e={},r("next"),r("throw",function(t){throw t}),r("return"),e[Symbol.iterator]=function(){return this},e;function r(r,i){e[r]=t[r]?function(e){return(n=!n)?{value:g(t[r](e)),done:"return"===r}:i?i(e):e}:i}}function v(t){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var e,n=t[Symbol.asyncIterator];return n?n.call(t):(t=f(t),e={},r("next"),r("throw"),r("return"),e[Symbol.asyncIterator]=function(){return this},e);function r(n){e[n]=t[n]&&function(e){return new Promise(function(r,i){var o,a,s,u;e=t[n](e),o=r,a=i,s=e.done,u=e.value,Promise.resolve(u).then(function(t){o({value:t,done:s})},a)})}}}function b(t,e){return Object.defineProperty?Object.defineProperty(t,"raw",{value:e}):t.raw=e,t}function x(t){if(t&&t.__esModule)return t;var e={};if(null!=t)for(var n in t)Object.hasOwnProperty.call(t,n)&&(e[n]=t[n]);return e.default=t,e}function w(t){return t&&t.__esModule?t:{default:t}}},"./node_modules/@angular/compiler/fesm5/compiler.js":function(t,e,n){"use strict";n.r(e),n.d(e,"core",function(){return ot}),n.d(e,"CompilerConfig",function(){return Sd}),n.d(e,"preserveWhitespacesDefault",function(){return Md}),n.d(e,"isLoweredSymbol",function(){return Cp}),n.d(e,"createLoweredSymbol",function(){return Ep}),n.d(e,"Identifiers",function(){return gn}),n.d(e,"JitCompiler",function(){return rm}),n.d(e,"ConstantPool",function(){return dn}),n.d(e,"DirectiveResolver",function(){return Dd}),n.d(e,"PipeResolver",function(){return tf}),n.d(e,"NgModuleResolver",function(){return Qp}),n.d(e,"DEFAULT_INTERPOLATION_CONFIG",function(){return Hi}),n.d(e,"InterpolationConfig",function(){return zi}),n.d(e,"NgModuleCompiler",function(){return Xp}),n.d(e,"ArrayType",function(){return ft}),n.d(e,"AssertNotNull",function(){return Ut}),n.d(e,"BinaryOperator",function(){return ct}),n.d(e,"BinaryOperatorExpr",function(){return Yt}),n.d(e,"BuiltinMethod",function(){return Ct}),n.d(e,"BuiltinType",function(){return dt}),n.d(e,"BuiltinTypeName",function(){return at}),n.d(e,"BuiltinVar",function(){return wt}),n.d(e,"CastExpr",function(){return Wt}),n.d(e,"ClassField",function(){return ue}),n.d(e,"ClassMethod",function(){return ce}),n.d(e,"ClassStmt",function(){return de}),n.d(e,"CommaExpr",function(){return Jt}),n.d(e,"CommentStmt",function(){return fe}),n.d(e,"ConditionalExpr",function(){return zt}),n.d(e,"DeclareFunctionStmt",function(){return ie}),n.d(e,"DeclareVarStmt",function(){return re}),n.d(e,"Expression",function(){return Ot}),n.d(e,"ExpressionStatement",function(){return oe}),n.d(e,"ExpressionType",function(){return pt}),n.d(e,"ExternalExpr",function(){return Bt}),n.d(e,"ExternalReference",function(){return Vt}),n.d(e,"FunctionExpr",function(){return Gt}),n.d(e,"IfStmt",function(){return pe}),n.d(e,"InstantiateExpr",function(){return Ft}),n.d(e,"InvokeFunctionExpr",function(){return Nt}),n.d(e,"InvokeMethodExpr",function(){return Rt}),n.d(e,"JSDocCommentStmt",function(){return he}),n.d(e,"LiteralArrayExpr",function(){return Qt}),n.d(e,"LiteralExpr",function(){return Lt}),n.d(e,"LiteralMapExpr",function(){return $t}),n.d(e,"MapType",function(){return ht}),n.d(e,"NotExpr",function(){return Ht}),n.d(e,"ReadKeyExpr",function(){return Xt}),n.d(e,"ReadPropExpr",function(){return Kt}),n.d(e,"ReadVarExpr",function(){return St}),n.d(e,"ReturnStatement",function(){return ae}),n.d(e,"ThrowStmt",function(){return ge}),n.d(e,"TryCatchStmt",function(){return me}),n.d(e,"Type",function(){return ut}),n.d(e,"WrappedNodeExpr",function(){return Tt}),n.d(e,"WriteKeyExpr",function(){return At}),n.d(e,"WritePropExpr",function(){return Dt}),n.d(e,"WriteVarExpr",function(){return kt}),n.d(e,"StmtModifier",function(){return It}),n.d(e,"Statement",function(){return ne}),n.d(e,"TypeofExpr",function(){return Mt}),n.d(e,"collectExternalReferences",function(){return xe}),n.d(e,"EmitterVisitorContext",function(){return Ji}),n.d(e,"ViewCompiler",function(){return Af}),n.d(e,"getParseErrors",function(){return Ze}),n.d(e,"isSyntaxError",function(){return Qe}),n.d(e,"syntaxError",function(){return Ye}),n.d(e,"Version",function(){return on}),n.d(e,"TmplAstBoundAttribute",function(){return mc}),n.d(e,"TmplAstBoundEvent",function(){return gc}),n.d(e,"TmplAstBoundText",function(){return fc}),n.d(e,"TmplAstContent",function(){return vc}),n.d(e,"TmplAstElement",function(){return yc}),n.d(e,"TmplAstReference",function(){return xc}),n.d(e,"TmplAstTemplate",function(){return _c}),n.d(e,"TmplAstText",function(){return pc}),n.d(e,"TmplAstTextAttribute",function(){return hc}),n.d(e,"TmplAstVariable",function(){return bc}),n.d(e,"jitExpression",function(){return ao}),n.d(e,"R3ResolvedDependencyType",function(){return wi}),n.d(e,"compileInjector",function(){return uo}),n.d(e,"compileNgModule",function(){return so}),n.d(e,"compilePipeFromMetadata",function(){return lo}),n.d(e,"makeBindingParser",function(){return Ol}),n.d(e,"parseTemplate",function(){return jl}),n.d(e,"compileBaseDefFromMetadata",function(){return Il}),n.d(e,"compileComponentFromMetadata",function(){return Pl}),n.d(e,"compileDirectiveFromMetadata",function(){return Dl}),n.d(e,"parseHostBindings",function(){return ql}),n.d(e,"publishFacade",function(){return od}),n.d(e,"VERSION",function(){return cd}),n.d(e,"TextAst",function(){return ld}),n.d(e,"BoundTextAst",function(){return dd}),n.d(e,"AttrAst",function(){return pd}),n.d(e,"BoundElementPropertyAst",function(){return hd}),n.d(e,"BoundEventAst",function(){return md}),n.d(e,"ReferenceAst",function(){return gd}),n.d(e,"VariableAst",function(){return yd}),n.d(e,"ElementAst",function(){return _d}),n.d(e,"EmbeddedTemplateAst",function(){return vd}),n.d(e,"BoundDirectivePropertyAst",function(){return bd}),n.d(e,"DirectiveAst",function(){return xd}),n.d(e,"ProviderAst",function(){return wd}),n.d(e,"ProviderAstType",function(){return sd}),n.d(e,"NgContentAst",function(){return jd}),n.d(e,"NullTemplateVisitor",function(){return Od}),n.d(e,"RecursiveTemplateAstVisitor",function(){return Cd}),n.d(e,"templateVisitAll",function(){return Ed}),n.d(e,"sanitizeIdentifier",function(){return wn}),n.d(e,"identifierName",function(){return En}),n.d(e,"identifierModuleUrl",function(){return Sn}),n.d(e,"viewClassName",function(){return Mn}),n.d(e,"rendererTypeName",function(){return Tn}),n.d(e,"hostViewClassName",function(){return kn}),n.d(e,"componentFactoryName",function(){return An}),n.d(e,"CompileSummaryKind",function(){return jn}),n.d(e,"tokenName",function(){return Dn}),n.d(e,"tokenReference",function(){return In}),n.d(e,"CompileStylesheetMetadata",function(){return Pn}),n.d(e,"CompileTemplateMetadata",function(){return Rn}),n.d(e,"CompileDirectiveMetadata",function(){return Nn}),n.d(e,"CompilePipeMetadata",function(){return Fn}),n.d(e,"CompileShallowModuleMetadata",function(){return Ln}),n.d(e,"CompileNgModuleMetadata",function(){return Bn}),n.d(e,"TransitiveCompileNgModuleMetadata",function(){return Vn}),n.d(e,"ProviderMeta",function(){return Hn}),n.d(e,"flatten",function(){return Un}),n.d(e,"templateSourceUrl",function(){return qn}),n.d(e,"sharedStylesheetJitUrl",function(){return Gn}),n.d(e,"ngModuleJitUrl",function(){return Yn}),n.d(e,"templateJitUrl",function(){return Kn}),n.d(e,"createAotUrlResolver",function(){return Gh}),n.d(e,"createAotCompiler",function(){return Yh}),n.d(e,"AotCompiler",function(){return sh}),n.d(e,"analyzeNgModules",function(){return dh}),n.d(e,"analyzeAndValidateNgModules",function(){return ph}),n.d(e,"analyzeFile",function(){return hh}),n.d(e,"analyzeFileForInjectables",function(){return mh}),n.d(e,"mergeAnalyzedFiles",function(){return gh}),n.d(e,"GeneratedFile",function(){return Gf}),n.d(e,"toTypeScript",function(){return Yf}),n.d(e,"formattedError",function(){return bh}),n.d(e,"isFormattedError",function(){return xh}),n.d(e,"StaticReflector",function(){return Th}),n.d(e,"StaticSymbol",function(){return vn}),n.d(e,"StaticSymbolCache",function(){return bn}),n.d(e,"ResolvedStaticSymbol",function(){return $f}),n.d(e,"StaticSymbolResolver",function(){return Jf}),n.d(e,"unescapeIdentifier",function(){return th}),n.d(e,"unwrapResolvedMetadata",function(){return eh}),n.d(e,"AotSummaryResolver",function(){return qh}),n.d(e,"AstPath",function(){return du}),n.d(e,"SummaryResolver",function(){return Kh}),n.d(e,"JitSummaryResolver",function(){return Xh}),n.d(e,"CompileReflector",function(){return sm}),n.d(e,"createUrlResolverWithoutPackagePrefix",function(){return um}),n.d(e,"createOfflineCompileUrlResolver",function(){return cm}),n.d(e,"UrlResolver",function(){return lm}),n.d(e,"getUrlScheme",function(){return dm}),n.d(e,"ResourceLoader",function(){return ym}),n.d(e,"ElementSchemaRegistry",function(){return tc}),n.d(e,"Extractor",function(){return _m}),n.d(e,"I18NHtmlParser",function(){return up}),n.d(e,"MessageBundle",function(){return Wf}),n.d(e,"Serializer",function(){return Er}),n.d(e,"Xliff",function(){return Xd}),n.d(e,"Xliff2",function(){return Jd}),n.d(e,"Xmb",function(){return Fr}),n.d(e,"Xtb",function(){return rp}),n.d(e,"DirectiveNormalizer",function(){return Td}),n.d(e,"ParserError",function(){return po}),n.d(e,"ParseSpan",function(){return fo}),n.d(e,"AST",function(){return ho}),n.d(e,"Quote",function(){return mo}),n.d(e,"EmptyExpr",function(){return go}),n.d(e,"ImplicitReceiver",function(){return yo}),n.d(e,"Chain",function(){return _o}),n.d(e,"Conditional",function(){return vo}),n.d(e,"PropertyRead",function(){return bo}),n.d(e,"PropertyWrite",function(){return xo}),n.d(e,"SafePropertyRead",function(){return wo}),n.d(e,"KeyedRead",function(){return jo}),n.d(e,"KeyedWrite",function(){return Oo}),n.d(e,"BindingPipe",function(){return Co}),n.d(e,"LiteralPrimitive",function(){return Eo}),n.d(e,"LiteralArray",function(){return So}),n.d(e,"LiteralMap",function(){return Mo}),n.d(e,"Interpolation",function(){return To}),n.d(e,"Binary",function(){return ko}),n.d(e,"PrefixNot",function(){return Ao}),n.d(e,"NonNullAssert",function(){return Do}),n.d(e,"MethodCall",function(){return Io}),n.d(e,"SafeMethodCall",function(){return Po}),n.d(e,"FunctionCall",function(){return Ro}),n.d(e,"ASTWithSource",function(){return No}),n.d(e,"TemplateBinding",function(){return Fo}),n.d(e,"NullAstVisitor",function(){return Lo}),n.d(e,"RecursiveAstVisitor",function(){return Bo}),n.d(e,"AstTransformer",function(){return Vo}),n.d(e,"AstMemoryEfficientTransformer",function(){return zo}),n.d(e,"visitAstChildren",function(){return Ho}),n.d(e,"ParsedProperty",function(){return qo}),n.d(e,"ParsedPropertyType",function(){return Uo}),n.d(e,"ParsedEvent",function(){return Go}),n.d(e,"ParsedVariable",function(){return Yo}),n.d(e,"BoundElementProperty",function(){return Ko}),n.d(e,"TokenType",function(){return zs}),n.d(e,"Lexer",function(){return Ks}),n.d(e,"Token",function(){return Xs}),n.d(e,"EOF",function(){return $s}),n.d(e,"isIdentifier",function(){return eu}),n.d(e,"isQuote",function(){return ru}),n.d(e,"SplitInterpolation",function(){return ou}),n.d(e,"TemplateBindingParseResult",function(){return au}),n.d(e,"Parser",function(){return uu}),n.d(e,"_ParseAST",function(){return cu}),n.d(e,"ERROR_COMPONENT_TYPE",function(){return Sp}),n.d(e,"CompileMetadataResolver",function(){return Mp}),n.d(e,"Text",function(){return pu}),n.d(e,"Expansion",function(){return fu}),n.d(e,"ExpansionCase",function(){return hu}),n.d(e,"Attribute",function(){return mu}),n.d(e,"Element",function(){return gu}),n.d(e,"Comment",function(){return yu}),n.d(e,"visitAll",function(){return _u}),n.d(e,"RecursiveVisitor",function(){return xu}),n.d(e,"findNode",function(){return wu}),n.d(e,"HtmlParser",function(){return Hu}),n.d(e,"ParseTreeResult",function(){return Lu}),n.d(e,"TreeError",function(){return Fu}),n.d(e,"HtmlTagDefinition",function(){return g}),n.d(e,"getHtmlTagDefinition",function(){return y}),n.d(e,"TagContentType",function(){return r}),n.d(e,"splitNsName",function(){return a}),n.d(e,"isNgContainer",function(){return s}),n.d(e,"isNgContent",function(){return u}),n.d(e,"isNgTemplate",function(){return c}),n.d(e,"getNsPrefix",function(){return l}),n.d(e,"mergeNsAndName",function(){return d}),n.d(e,"NAMED_ENTITIES",function(){return p}),n.d(e,"NGSP_UNICODE",function(){return f}),n.d(e,"debugOutputAstAsTypeScript",function(){return Zp}),n.d(e,"TypeScriptEmitter",function(){return $p}),n.d(e,"ParseLocation",function(){return ns}),n.d(e,"ParseSourceFile",function(){return rs}),n.d(e,"ParseSourceSpan",function(){return is}),n.d(e,"ParseErrorLevel",function(){return ts}),n.d(e,"ParseError",function(){return os}),n.d(e,"typeSourceSpan",function(){return as}),n.d(e,"DomElementSchemaRegistry",function(){return sc}),n.d(e,"CssSelector",function(){return j}),n.d(e,"SelectorMatcher",function(){return O}),n.d(e,"SelectorListContext",function(){return C}),n.d(e,"SelectorContext",function(){return E}),n.d(e,"HOST_ATTR",function(){return Ps}),n.d(e,"CONTENT_ATTR",function(){return Rs}),n.d(e,"StylesCompileDependency",function(){return Ns}),n.d(e,"CompiledStylesheet",function(){return Fs}),n.d(e,"StyleCompiler",function(){return Ls}),n.d(e,"TemplateParseError",function(){return df}),n.d(e,"TemplateParseResult",function(){return pf}),n.d(e,"TemplateParser",function(){return ff}),n.d(e,"splitClasses",function(){return yf}),n.d(e,"createElementCssSelector",function(){return vf}),n.d(e,"removeSummaryDuplicates",function(){return jf}),n.d(e,"compileInjectable",function(){return Fi}),n.d(e,"R3TargetBinder",function(){return vm}),n.d(e,"R3BoundTarget",function(){return jm});var r,i,o=n("./node_modules/@angular/compiler/node_modules/tslib/tslib.es6.js");function a(t){if(":"!=t[0])return[null,t];var e=t.indexOf(":",1);if(-1==e)throw new Error('Unsupported format "'+t+'" expecting ":namespace:name"');return[t.slice(1,e),t.slice(e+1)]}function s(t){return"ng-container"===a(t)[1]}function u(t){return"ng-content"===a(t)[1]}function c(t){return"ng-template"===a(t)[1]}function l(t){return null===t?null:a(t)[0]}function d(t,e){return t?":"+t+":"+e:e}(i=r||(r={}))[i.RAW_TEXT=0]="RAW_TEXT",i[i.ESCAPABLE_RAW_TEXT=1]="ESCAPABLE_RAW_TEXT",i[i.PARSABLE_DATA=2]="PARSABLE_DATA";var p={Aacute:"Á",aacute:"á",Acirc:"Â",acirc:"â",acute:"´",AElig:"Æ",aelig:"æ",Agrave:"À",agrave:"à",alefsym:"ℵ",Alpha:"Α",alpha:"α",amp:"&",and:"∧",ang:"∠",apos:"'",Aring:"Å",aring:"å",asymp:"≈",Atilde:"Ã",atilde:"ã",Auml:"Ä",auml:"ä",bdquo:"„",Beta:"Β",beta:"β",brvbar:"¦",bull:"•",cap:"∩",Ccedil:"Ç",ccedil:"ç",cedil:"¸",cent:"¢",Chi:"Χ",chi:"χ",circ:"ˆ",clubs:"♣",cong:"≅",copy:"©",crarr:"↵",cup:"∪",curren:"¤",dagger:"†",Dagger:"‡",darr:"↓",dArr:"⇓",deg:"°",Delta:"Δ",delta:"δ",diams:"♦",divide:"÷",Eacute:"É",eacute:"é",Ecirc:"Ê",ecirc:"ê",Egrave:"È",egrave:"è",empty:"∅",emsp:" ",ensp:" ",Epsilon:"Ε",epsilon:"ε",equiv:"≡",Eta:"Η",eta:"η",ETH:"Ð",eth:"ð",Euml:"Ë",euml:"ë",euro:"€",exist:"∃",fnof:"ƒ",forall:"∀",frac12:"½",frac14:"¼",frac34:"¾",frasl:"⁄",Gamma:"Γ",gamma:"γ",ge:"≥",gt:">",harr:"↔",hArr:"⇔",hearts:"♥",hellip:"…",Iacute:"Í",iacute:"í",Icirc:"Î",icirc:"î",iexcl:"¡",Igrave:"Ì",igrave:"ì",image:"ℑ",infin:"∞",int:"∫",Iota:"Ι",iota:"ι",iquest:"¿",isin:"∈",Iuml:"Ï",iuml:"ï",Kappa:"Κ",kappa:"κ",Lambda:"Λ",lambda:"λ",lang:"⟨",laquo:"«",larr:"←",lArr:"⇐",lceil:"⌈",ldquo:"“",le:"≤",lfloor:"⌊",lowast:"∗",loz:"◊",lrm:"‎",lsaquo:"‹",lsquo:"‘",lt:"<",macr:"¯",mdash:"—",micro:"µ",middot:"·",minus:"−",Mu:"Μ",mu:"μ",nabla:"∇",nbsp:" ",ndash:"–",ne:"≠",ni:"∋",not:"¬",notin:"∉",nsub:"⊄",Ntilde:"Ñ",ntilde:"ñ",Nu:"Ν",nu:"ν",Oacute:"Ó",oacute:"ó",Ocirc:"Ô",ocirc:"ô",OElig:"Œ",oelig:"œ",Ograve:"Ò",ograve:"ò",oline:"‾",Omega:"Ω",omega:"ω",Omicron:"Ο",omicron:"ο",oplus:"⊕",or:"∨",ordf:"ª",ordm:"º",Oslash:"Ø",oslash:"ø",Otilde:"Õ",otilde:"õ",otimes:"⊗",Ouml:"Ö",ouml:"ö",para:"¶",permil:"‰",perp:"⊥",Phi:"Φ",phi:"φ",Pi:"Π",pi:"π",piv:"ϖ",plusmn:"±",pound:"£",prime:"′",Prime:"″",prod:"∏",prop:"∝",Psi:"Ψ",psi:"ψ",quot:'"',radic:"√",rang:"⟩",raquo:"»",rarr:"→",rArr:"⇒",rceil:"⌉",rdquo:"”",real:"ℜ",reg:"®",rfloor:"⌋",Rho:"Ρ",rho:"ρ",rlm:"‏",rsaquo:"›",rsquo:"’",sbquo:"‚",Scaron:"Š",scaron:"š",sdot:"⋅",sect:"§",shy:"­",Sigma:"Σ",sigma:"σ",sigmaf:"ς",sim:"∼",spades:"♠",sub:"⊂",sube:"⊆",sum:"∑",sup:"⊃",sup1:"¹",sup2:"²",sup3:"³",supe:"⊇",szlig:"ß",Tau:"Τ",tau:"τ",there4:"∴",Theta:"Θ",theta:"θ",thetasym:"ϑ",thinsp:" ",THORN:"Þ",thorn:"þ",tilde:"˜",times:"×",trade:"™",Uacute:"Ú",uacute:"ú",uarr:"↑",uArr:"⇑",Ucirc:"Û",ucirc:"û",Ugrave:"Ù",ugrave:"ù",uml:"¨",upsih:"ϒ",Upsilon:"Υ",upsilon:"υ",Uuml:"Ü",uuml:"ü",weierp:"℘",Xi:"Ξ",xi:"ξ",Yacute:"Ý",yacute:"ý",yen:"¥",yuml:"ÿ",Yuml:"Ÿ",Zeta:"Ζ",zeta:"ζ",zwj:"‍",zwnj:"‌"},f="";p.ngsp=f;var h,m,g=function(){function t(t){var e=void 0===t?{}:t,n=e.closedByChildren,i=e.requiredParents,o=e.implicitNamespacePrefix,a=e.contentType,s=void 0===a?r.PARSABLE_DATA:a,u=e.closedByParent,c=void 0!==u&&u,l=e.isVoid,d=void 0!==l&&l,p=e.ignoreFirstLf,f=void 0!==p&&p,h=this;this.closedByChildren={},this.closedByParent=!1,this.canSelfClose=!1,n&&n.length>0&&n.forEach(function(t){return h.closedByChildren[t]=!0}),this.isVoid=d,this.closedByParent=c||d,i&&i.length>0&&(this.requiredParents={},this.parentToAdd=i[0],i.forEach(function(t){return h.requiredParents[t]=!0})),this.implicitNamespacePrefix=o||null,this.contentType=s,this.ignoreFirstLf=f}return t.prototype.requireExtraParent=function(t){if(!this.requiredParents)return!1;if(!t)return!0;var e=t.toLowerCase();return!("template"===e||"ng-template"===t)&&1!=this.requiredParents[e]},t.prototype.isClosedByChild=function(t){return this.isVoid||t.toLowerCase()in this.closedByChildren},t}();function y(t){return m||(h=new g,m={base:new g({isVoid:!0}),meta:new g({isVoid:!0}),area:new g({isVoid:!0}),embed:new g({isVoid:!0}),link:new g({isVoid:!0}),img:new g({isVoid:!0}),input:new g({isVoid:!0}),param:new g({isVoid:!0}),hr:new g({isVoid:!0}),br:new g({isVoid:!0}),source:new g({isVoid:!0}),track:new g({isVoid:!0}),wbr:new g({isVoid:!0}),p:new g({closedByChildren:["address","article","aside","blockquote","div","dl","fieldset","footer","form","h1","h2","h3","h4","h5","h6","header","hgroup","hr","main","nav","ol","p","pre","section","table","ul"],closedByParent:!0}),thead:new g({closedByChildren:["tbody","tfoot"]}),tbody:new g({closedByChildren:["tbody","tfoot"],closedByParent:!0}),tfoot:new g({closedByChildren:["tbody"],closedByParent:!0}),tr:new g({closedByChildren:["tr"],requiredParents:["tbody","tfoot","thead"],closedByParent:!0}),td:new g({closedByChildren:["td","th"],closedByParent:!0}),th:new g({closedByChildren:["td","th"],closedByParent:!0}),col:new g({requiredParents:["colgroup"],isVoid:!0}),svg:new g({implicitNamespacePrefix:"svg"}),math:new g({implicitNamespacePrefix:"math"}),li:new g({closedByChildren:["li"],closedByParent:!0}),dt:new g({closedByChildren:["dt","dd"]}),dd:new g({closedByChildren:["dt","dd"],closedByParent:!0}),rb:new g({closedByChildren:["rb","rt","rtc","rp"],closedByParent:!0}),rt:new g({closedByChildren:["rb","rt","rtc","rp"],closedByParent:!0}),rtc:new g({closedByChildren:["rb","rtc","rp"],closedByParent:!0}),rp:new g({closedByChildren:["rb","rt","rtc","rp"],closedByParent:!0}),optgroup:new g({closedByChildren:["optgroup"],closedByParent:!0}),option:new g({closedByChildren:["option","optgroup"],closedByParent:!0}),pre:new g({ignoreFirstLf:!0}),listing:new g({ignoreFirstLf:!0}),style:new g({contentType:r.RAW_TEXT}),script:new g({contentType:r.RAW_TEXT}),title:new g({contentType:r.ESCAPABLE_RAW_TEXT}),textarea:new g({contentType:r.ESCAPABLE_RAW_TEXT,ignoreFirstLf:!0})}),m[t.toLowerCase()]||h}var _,v,b,x,w=new RegExp("(\\:not\\()|([-\\w]+)|(?:\\.([-\\w]+))|(?:\\[([-.\\w*]+)(?:=([\"']?)([^\\]\"']*)\\5)?\\])|(\\))|(\\s*,\\s*)","g"),j=function(){function t(){this.element=null,this.classNames=[],this.attrs=[],this.notSelectors=[]}return t.parse=function(e){var n,r=[],i=function(t,e){e.notSelectors.length>0&&!e.element&&0==e.classNames.length&&0==e.attrs.length&&(e.element="*"),t.push(e)},o=new t,a=o,s=!1;for(w.lastIndex=0;n=w.exec(e);){if(n[1]){if(s)throw new Error("Nesting :not is not allowed in a selector");s=!0,a=new t,o.notSelectors.push(a)}if(n[2]&&a.setElement(n[2]),n[3]&&a.addClassName(n[3]),n[4]&&a.addAttribute(n[4],n[6]),n[7]&&(s=!1,a=o),n[8]){if(s)throw new Error("Multiple selectors in :not are not supported");i(r,o),o=a=new t}}return i(r,o),r},t.prototype.isElementSelector=function(){return this.hasElementSelector()&&0==this.classNames.length&&0==this.attrs.length&&0===this.notSelectors.length},t.prototype.hasElementSelector=function(){return!!this.element},t.prototype.setElement=function(t){void 0===t&&(t=null),this.element=t},t.prototype.getMatchingElementTemplate=function(){for(var t=this.element||"div",e=this.classNames.length>0?' class="'+this.classNames.join(" ")+'"':"",n="",r=0;r<this.attrs.length;r+=2){n+=" "+this.attrs[r]+(""!==this.attrs[r+1]?'="'+this.attrs[r+1]+'"':"")}return y(t).isVoid?"<"+t+e+n+"/>":"<"+t+e+n+"></"+t+">"},t.prototype.getAttrs=function(){var t=[];return this.classNames.length>0&&t.push("class",this.classNames.join(" ")),t.concat(this.attrs)},t.prototype.addAttribute=function(t,e){void 0===e&&(e=""),this.attrs.push(t,e&&e.toLowerCase()||"")},t.prototype.addClassName=function(t){this.classNames.push(t.toLowerCase())},t.prototype.toString=function(){var t=this.element||"";if(this.classNames&&this.classNames.forEach(function(e){return t+="."+e}),this.attrs)for(var e=0;e<this.attrs.length;e+=2){var n=this.attrs[e],r=this.attrs[e+1];t+="["+n+(r?"="+r:"")+"]"}return this.notSelectors.forEach(function(e){return t+=":not("+e+")"}),t},t}(),O=function(){function t(){this._elementMap=new Map,this._elementPartialMap=new Map,this._classMap=new Map,this._classPartialMap=new Map,this._attrValueMap=new Map,this._attrValuePartialMap=new Map,this._listContexts=[]}return t.createNotMatcher=function(e){var n=new t;return n.addSelectables(e,null),n},t.prototype.addSelectables=function(t,e){var n=null;t.length>1&&(n=new C(t),this._listContexts.push(n));for(var r=0;r<t.length;r++)this._addSelectable(t[r],e,n)},t.prototype._addSelectable=function(t,e,n){var r=this,i=t.element,o=t.classNames,a=t.attrs,s=new E(t,e,n);i&&((c=0===a.length&&0===o.length)?this._addTerminal(r._elementMap,i,s):r=this._addPartial(r._elementPartialMap,i));if(o)for(var u=0;u<o.length;u++){var c=0===a.length&&u===o.length-1,l=o[u];c?this._addTerminal(r._classMap,l,s):r=this._addPartial(r._classPartialMap,l)}if(a)for(u=0;u<a.length;u+=2){c=u===a.length-2;var d=a[u],p=a[u+1];if(c){var f=r._attrValueMap,h=f.get(d);h||(h=new Map,f.set(d,h)),this._addTerminal(h,p,s)}else{var m=r._attrValuePartialMap,g=m.get(d);g||(g=new Map,m.set(d,g)),r=this._addPartial(g,p)}}},t.prototype._addTerminal=function(t,e,n){var r=t.get(e);r||(r=[],t.set(e,r)),r.push(n)},t.prototype._addPartial=function(e,n){var r=e.get(n);return r||(r=new t,e.set(n,r)),r},t.prototype.match=function(t,e){for(var n=!1,r=t.element,i=t.classNames,o=t.attrs,a=0;a<this._listContexts.length;a++)this._listContexts[a].alreadyMatched=!1;if(n=this._matchTerminal(this._elementMap,r,t,e)||n,n=this._matchPartial(this._elementPartialMap,r,t,e)||n,i)for(a=0;a<i.length;a++){var s=i[a];n=this._matchTerminal(this._classMap,s,t,e)||n,n=this._matchPartial(this._classPartialMap,s,t,e)||n}if(o)for(a=0;a<o.length;a+=2){var u=o[a],c=o[a+1],l=this._attrValueMap.get(u);c&&(n=this._matchTerminal(l,"",t,e)||n),n=this._matchTerminal(l,c,t,e)||n;var d=this._attrValuePartialMap.get(u);c&&(n=this._matchPartial(d,"",t,e)||n),n=this._matchPartial(d,c,t,e)||n}return n},t.prototype._matchTerminal=function(t,e,n,r){if(!t||"string"!=typeof e)return!1;var i=t.get(e)||[],o=t.get("*");if(o&&(i=i.concat(o)),0===i.length)return!1;for(var a=!1,s=0;s<i.length;s++)a=i[s].finalize(n,r)||a;return a},t.prototype._matchPartial=function(t,e,n,r){if(!t||"string"!=typeof e)return!1;var i=t.get(e);return!!i&&i.match(n,r)},t}(),C=function(){return function(t){this.selectors=t,this.alreadyMatched=!1}}(),E=function(){function t(t,e,n){this.selector=t,this.cbContext=e,this.listContext=n,this.notSelectors=t.notSelectors}return t.prototype.finalize=function(t,e){var n=!0;!(this.notSelectors.length>0)||this.listContext&&this.listContext.alreadyMatched||(n=!O.createNotMatcher(this.notSelectors).match(t,null));return!n||!e||this.listContext&&this.listContext.alreadyMatched||(this.listContext&&(this.listContext.alreadyMatched=!0),e(this.selector,this.cbContext)),n},t}(),S=tt("Inject",function(t){return{token:t}}),M=tt("InjectionToken",function(t){return{_desc:t,ngInjectableDef:void 0}}),T=tt("Attribute",function(t){return{attributeName:t}}),k=tt("ContentChildren",function(t,e){return void 0===e&&(e={}),Object(o.__assign)({selector:t,first:!1,isViewQuery:!1,descendants:!1},e)}),A=tt("ContentChild",function(t,e){return void 0===e&&(e={}),Object(o.__assign)({selector:t,first:!0,isViewQuery:!1,descendants:!0},e)}),D=tt("ViewChildren",function(t,e){return void 0===e&&(e={}),Object(o.__assign)({selector:t,first:!1,isViewQuery:!0,descendants:!0},e)}),I=tt("ViewChild",function(t,e){return Object(o.__assign)({selector:t,first:!0,isViewQuery:!0,descendants:!0},e)}),P=tt("Directive",function(t){return void 0===t&&(t={}),t});(v=_||(_={}))[v.Emulated=0]="Emulated",v[v.Native=1]="Native",v[v.None=2]="None",v[v.ShadowDom=3]="ShadowDom",(x=b||(b={}))[x.OnPush=0]="OnPush",x[x.Default=1]="Default";var R,N,F,L,B=tt("Component",function(t){return void 0===t&&(t={}),Object(o.__assign)({changeDetection:b.Default},t)}),V=tt("Pipe",function(t){return Object(o.__assign)({pure:!0},t)}),z=tt("Input",function(t){return{bindingPropertyName:t}}),H=tt("Output",function(t){return{bindingPropertyName:t}}),U=tt("HostBinding",function(t){return{hostPropertyName:t}}),W=tt("HostListener",function(t,e){return{eventName:t,args:e}}),q=tt("NgModule",function(t){return t}),G=tt("Injectable",function(t){return void 0===t&&(t={}),t}),Y={name:"custom-elements"},K={name:"no-errors-schema"},X=tt("Optional"),Q=tt("Self"),Z=tt("SkipSelf"),$=tt("Host"),J=Function;function tt(t,e){var n=function(){for(var n=[],r=0;r<arguments.length;r++)n[r]=arguments[r];var i=e?e.apply(void 0,Object(o.__spread)(n)):{};return Object(o.__assign)({ngMetadataName:t},i)};return n.isTypeOf=function(e){return e&&e.ngMetadataName===t},n.ngMetadataName=t,n}function et(t){var e,n,r,i=(n=(e=t).classNames&&e.classNames.length?Object(o.__spread)([8],e.classNames):[],r=e.element&&"*"!==e.element?e.element:"",Object(o.__spread)([r],e.attrs,n)),a=t.notSelectors&&t.notSelectors.length?t.notSelectors.map(function(t){return n=(e=t).classNames&&e.classNames.length?Object(o.__spread)([8],e.classNames):[],e.element?Object(o.__spread)([5,e.element],e.attrs,n):e.attrs.length?Object(o.__spread)([3],e.attrs,n):e.classNames&&e.classNames.length?Object(o.__spread)([9],e.classNames):[];var e,n}):[];return i.concat.apply(i,Object(o.__spread)(a))}function nt(t){return t?j.parse(t).map(et):[]}(N=R||(R={}))[N.NONE=0]="NONE",N[N.HTML=1]="HTML",N[N.STYLE=2]="STYLE",N[N.SCRIPT=3]="SCRIPT",N[N.URL=4]="URL",N[N.RESOURCE_URL=5]="RESOURCE_URL",(L=F||(F={}))[L.Error=0]="Error",L[L.Warning=1]="Warning",L[L.Ignore=2]="Ignore";var rt,it,ot=Object.freeze({createInject:S,createInjectionToken:M,createAttribute:T,createContentChildren:k,createContentChild:A,createViewChildren:D,createViewChild:I,createDirective:P,get ViewEncapsulation(){return _},get ChangeDetectionStrategy(){return b},createComponent:B,createPipe:V,createInput:z,createOutput:H,createHostBinding:U,createHostListener:W,createNgModule:q,createInjectable:G,CUSTOM_ELEMENTS_SCHEMA:Y,NO_ERRORS_SCHEMA:K,createOptional:X,createSelf:Q,createSkipSelf:Z,createHost:$,Type:J,get SecurityContext(){return R},get MissingTranslationStrategy(){return F},parseSelectorToR3Selector:nt});(it=rt||(rt={}))[it.Const=0]="Const";var at,st,ut=function(){function t(t){void 0===t&&(t=null),this.modifiers=t,t||(this.modifiers=[])}return t.prototype.hasModifier=function(t){return-1!==this.modifiers.indexOf(t)},t}();(st=at||(at={}))[st.Dynamic=0]="Dynamic",st[st.Bool=1]="Bool",st[st.String=2]="String",st[st.Int=3]="Int",st[st.Number=4]="Number",st[st.Function=5]="Function",st[st.Inferred=6]="Inferred",st[st.None=7]="None";var ct,lt,dt=function(t){function e(e,n){void 0===n&&(n=null);var r=t.call(this,n)||this;return r.name=e,r}return Object(o.__extends)(e,t),e.prototype.visitType=function(t,e){return t.visitBuiltinType(this,e)},e}(ut),pt=function(t){function e(e,n,r){void 0===n&&(n=null),void 0===r&&(r=null);var i=t.call(this,n)||this;return i.value=e,i.typeParams=r,i}return Object(o.__extends)(e,t),e.prototype.visitType=function(t,e){return t.visitExpressionType(this,e)},e}(ut),ft=function(t){function e(e,n){void 0===n&&(n=null);var r=t.call(this,n)||this;return r.of=e,r}return Object(o.__extends)(e,t),e.prototype.visitType=function(t,e){return t.visitArrayType(this,e)},e}(ut),ht=function(t){function e(e,n){void 0===n&&(n=null);var r=t.call(this,n)||this;return r.valueType=e||null,r}return Object(o.__extends)(e,t),e.prototype.visitType=function(t,e){return t.visitMapType(this,e)},e}(ut),mt=new dt(at.Dynamic),gt=new dt(at.Inferred),yt=new dt(at.Bool),_t=(new dt(at.Int),new dt(at.Number)),vt=(new dt(at.String),new dt(at.Function),new dt(at.None));function bt(t,e){return null==t||null==e?t==e:t.isEquivalent(e)}function xt(t,e){var n=t.length;if(n!==e.length)return!1;for(var r=0;r<n;r++)if(!t[r].isEquivalent(e[r]))return!1;return!0}(lt=ct||(ct={}))[lt.Equals=0]="Equals",lt[lt.NotEquals=1]="NotEquals",lt[lt.Identical=2]="Identical",lt[lt.NotIdentical=3]="NotIdentical",lt[lt.Minus=4]="Minus",lt[lt.Plus=5]="Plus",lt[lt.Divide=6]="Divide",lt[lt.Multiply=7]="Multiply",lt[lt.Modulo=8]="Modulo",lt[lt.And=9]="And",lt[lt.Or=10]="Or",lt[lt.BitwiseAnd=11]="BitwiseAnd",lt[lt.Lower=12]="Lower",lt[lt.LowerEquals=13]="LowerEquals",lt[lt.Bigger=14]="Bigger",lt[lt.BiggerEquals=15]="BiggerEquals";var wt,jt,Ot=function(){function t(t,e){this.type=t||null,this.sourceSpan=e||null}return t.prototype.prop=function(t,e){return new Kt(this,t,null,e)},t.prototype.key=function(t,e,n){return new Xt(this,t,e,n)},t.prototype.callMethod=function(t,e,n){return new Rt(this,t,e,null,n)},t.prototype.callFn=function(t,e){return new Nt(this,t,null,e)},t.prototype.instantiate=function(t,e,n){return new Ft(this,t,e,n)},t.prototype.conditional=function(t,e,n){return void 0===e&&(e=null),new zt(this,t,e,null,n)},t.prototype.equals=function(t,e){return new Yt(ct.Equals,this,t,null,e)},t.prototype.notEquals=function(t,e){return new Yt(ct.NotEquals,this,t,null,e)},t.prototype.identical=function(t,e){return new Yt(ct.Identical,this,t,null,e)},t.prototype.notIdentical=function(t,e){return new Yt(ct.NotIdentical,this,t,null,e)},t.prototype.minus=function(t,e){return new Yt(ct.Minus,this,t,null,e)},t.prototype.plus=function(t,e){return new Yt(ct.Plus,this,t,null,e)},t.prototype.divide=function(t,e){return new Yt(ct.Divide,this,t,null,e)},t.prototype.multiply=function(t,e){return new Yt(ct.Multiply,this,t,null,e)},t.prototype.modulo=function(t,e){return new Yt(ct.Modulo,this,t,null,e)},t.prototype.and=function(t,e){return new Yt(ct.And,this,t,null,e)},t.prototype.bitwiseAnd=function(t,e,n){return void 0===n&&(n=!0),new Yt(ct.BitwiseAnd,this,t,null,e,n)},t.prototype.or=function(t,e){return new Yt(ct.Or,this,t,null,e)},t.prototype.lower=function(t,e){return new Yt(ct.Lower,this,t,null,e)},t.prototype.lowerEquals=function(t,e){return new Yt(ct.LowerEquals,this,t,null,e)},t.prototype.bigger=function(t,e){return new Yt(ct.Bigger,this,t,null,e)},t.prototype.biggerEquals=function(t,e){return new Yt(ct.BiggerEquals,this,t,null,e)},t.prototype.isBlank=function(t){return this.equals(ee,t)},t.prototype.cast=function(t,e){return new Wt(this,t,e)},t.prototype.toStmt=function(){return new oe(this,null)},t}();(jt=wt||(wt={}))[jt.This=0]="This",jt[jt.Super=1]="Super",jt[jt.CatchError=2]="CatchError",jt[jt.CatchStack=3]="CatchStack";var Ct,Et,St=function(t){function e(e,n,r){var i=t.call(this,n,r)||this;return"string"==typeof e?(i.name=e,i.builtin=null):(i.name=null,i.builtin=e),i}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.name===t.name&&this.builtin===t.builtin},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitReadVarExpr(this,e)},e.prototype.set=function(t){if(!this.name)throw new Error("Built in variable "+this.builtin+" can not be assigned to.");return new kt(this.name,t,null,this.sourceSpan)},e}(Ot),Mt=function(t){function e(e,n,r){var i=t.call(this,n,r)||this;return i.expr=e,i}return Object(o.__extends)(e,t),e.prototype.visitExpression=function(t,e){return t.visitTypeofExpr(this,e)},e.prototype.isEquivalent=function(t){return t instanceof e&&t.expr.isEquivalent(this.expr)},e.prototype.isConstant=function(){return this.expr.isConstant()},e}(Ot),Tt=function(t){function e(e,n,r){var i=t.call(this,n,r)||this;return i.node=e,i}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.node===t.node},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitWrappedNodeExpr(this,e)},e}(Ot),kt=function(t){function e(e,n,r,i){var o=t.call(this,r||n.type,i)||this;return o.name=e,o.value=n,o}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.name===t.name&&this.value.isEquivalent(t.value)},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitWriteVarExpr(this,e)},e.prototype.toDeclStmt=function(t,e){return new re(this.name,this.value,t,e,this.sourceSpan)},e.prototype.toConstDecl=function(){return this.toDeclStmt(gt,[It.Final])},e}(Ot),At=function(t){function e(e,n,r,i,o){var a=t.call(this,i||r.type,o)||this;return a.receiver=e,a.index=n,a.value=r,a}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.receiver.isEquivalent(t.receiver)&&this.index.isEquivalent(t.index)&&this.value.isEquivalent(t.value)},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitWriteKeyExpr(this,e)},e}(Ot),Dt=function(t){function e(e,n,r,i,o){var a=t.call(this,i||r.type,o)||this;return a.receiver=e,a.name=n,a.value=r,a}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.receiver.isEquivalent(t.receiver)&&this.name===t.name&&this.value.isEquivalent(t.value)},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitWritePropExpr(this,e)},e}(Ot);(Et=Ct||(Ct={}))[Et.ConcatArray=0]="ConcatArray",Et[Et.SubscribeObservable=1]="SubscribeObservable",Et[Et.Bind=2]="Bind";var It,Pt,Rt=function(t){function e(e,n,r,i,o){var a=t.call(this,i,o)||this;return a.receiver=e,a.args=r,"string"==typeof n?(a.name=n,a.builtin=null):(a.name=null,a.builtin=n),a}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.receiver.isEquivalent(t.receiver)&&this.name===t.name&&this.builtin===t.builtin&&xt(this.args,t.args)},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitInvokeMethodExpr(this,e)},e}(Ot),Nt=function(t){function e(e,n,r,i,o){void 0===o&&(o=!1);var a=t.call(this,r,i)||this;return a.fn=e,a.args=n,a.pure=o,a}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.fn.isEquivalent(t.fn)&&xt(this.args,t.args)&&this.pure===t.pure},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitInvokeFunctionExpr(this,e)},e}(Ot),Ft=function(t){function e(e,n,r,i){var o=t.call(this,r,i)||this;return o.classExpr=e,o.args=n,o}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.classExpr.isEquivalent(t.classExpr)&&xt(this.args,t.args)},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitInstantiateExpr(this,e)},e}(Ot),Lt=function(t){function e(e,n,r){var i=t.call(this,n,r)||this;return i.value=e,i}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.value===t.value},e.prototype.isConstant=function(){return!0},e.prototype.visitExpression=function(t,e){return t.visitLiteralExpr(this,e)},e}(Ot),Bt=function(t){function e(e,n,r,i){void 0===r&&(r=null);var o=t.call(this,n,i)||this;return o.value=e,o.typeParams=r,o}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.value.name===t.value.name&&this.value.moduleName===t.value.moduleName&&this.value.runtime===t.value.runtime},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitExternalExpr(this,e)},e}(Ot),Vt=function(){return function(t,e,n){this.moduleName=t,this.name=e,this.runtime=n}}(),zt=function(t){function e(e,n,r,i,o){void 0===r&&(r=null);var a=t.call(this,i||n.type,o)||this;return a.condition=e,a.falseCase=r,a.trueCase=n,a}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.condition.isEquivalent(t.condition)&&this.trueCase.isEquivalent(t.trueCase)&&bt(this.falseCase,t.falseCase)},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitConditionalExpr(this,e)},e}(Ot),Ht=function(t){function e(e,n){var r=t.call(this,yt,n)||this;return r.condition=e,r}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.condition.isEquivalent(t.condition)},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitNotExpr(this,e)},e}(Ot),Ut=function(t){function e(e,n){var r=t.call(this,e.type,n)||this;return r.condition=e,r}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.condition.isEquivalent(t.condition)},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitAssertNotNullExpr(this,e)},e}(Ot),Wt=function(t){function e(e,n,r){var i=t.call(this,n,r)||this;return i.value=e,i}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.value.isEquivalent(t.value)},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitCastExpr(this,e)},e}(Ot),qt=function(){function t(t,e){void 0===e&&(e=null),this.name=t,this.type=e}return t.prototype.isEquivalent=function(t){return this.name===t.name},t}(),Gt=function(t){function e(e,n,r,i,o){var a=t.call(this,r,i)||this;return a.params=e,a.statements=n,a.name=o,a}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&xt(this.params,t.params)&&xt(this.statements,t.statements)},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitFunctionExpr(this,e)},e.prototype.toDeclStmt=function(t,e){return void 0===e&&(e=null),new ie(t,this.params,this.statements,this.type,e,this.sourceSpan)},e}(Ot),Yt=function(t){function e(e,n,r,i,o,a){void 0===a&&(a=!0);var s=t.call(this,i||n.type,o)||this;return s.operator=e,s.rhs=r,s.parens=a,s.lhs=n,s}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.operator===t.operator&&this.lhs.isEquivalent(t.lhs)&&this.rhs.isEquivalent(t.rhs)},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitBinaryOperatorExpr(this,e)},e}(Ot),Kt=function(t){function e(e,n,r,i){var o=t.call(this,r,i)||this;return o.receiver=e,o.name=n,o}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.receiver.isEquivalent(t.receiver)&&this.name===t.name},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitReadPropExpr(this,e)},e.prototype.set=function(t){return new Dt(this.receiver,this.name,t,null,this.sourceSpan)},e}(Ot),Xt=function(t){function e(e,n,r,i){var o=t.call(this,r,i)||this;return o.receiver=e,o.index=n,o}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.receiver.isEquivalent(t.receiver)&&this.index.isEquivalent(t.index)},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitReadKeyExpr(this,e)},e.prototype.set=function(t){return new At(this.receiver,this.index,t,null,this.sourceSpan)},e}(Ot),Qt=function(t){function e(e,n,r){var i=t.call(this,n,r)||this;return i.entries=e,i}return Object(o.__extends)(e,t),e.prototype.isConstant=function(){return this.entries.every(function(t){return t.isConstant()})},e.prototype.isEquivalent=function(t){return t instanceof e&&xt(this.entries,t.entries)},e.prototype.visitExpression=function(t,e){return t.visitLiteralArrayExpr(this,e)},e}(Ot),Zt=function(){function t(t,e,n){this.key=t,this.value=e,this.quoted=n}return t.prototype.isEquivalent=function(t){return this.key===t.key&&this.value.isEquivalent(t.value)},t}(),$t=function(t){function e(e,n,r){var i=t.call(this,n,r)||this;return i.entries=e,i.valueType=null,n&&(i.valueType=n.valueType),i}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&xt(this.entries,t.entries)},e.prototype.isConstant=function(){return this.entries.every(function(t){return t.value.isConstant()})},e.prototype.visitExpression=function(t,e){return t.visitLiteralMapExpr(this,e)},e}(Ot),Jt=function(t){function e(e,n){var r=t.call(this,e[e.length-1].type,n)||this;return r.parts=e,r}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&xt(this.parts,t.parts)},e.prototype.isConstant=function(){return!1},e.prototype.visitExpression=function(t,e){return t.visitCommaExpr(this,e)},e}(Ot),te=(new St(wt.This,null,null),new St(wt.Super,null,null),new St(wt.CatchError,null,null),new St(wt.CatchStack,null,null),new Lt(null,null,null)),ee=new Lt(null,gt,null);(Pt=It||(It={}))[Pt.Final=0]="Final",Pt[Pt.Private=1]="Private",Pt[Pt.Exported=2]="Exported",Pt[Pt.Static=3]="Static";var ne=function(){function t(t,e){this.modifiers=t||[],this.sourceSpan=e||null}return t.prototype.hasModifier=function(t){return-1!==this.modifiers.indexOf(t)},t}(),re=function(t){function e(e,n,r,i,o){void 0===i&&(i=null);var a=t.call(this,i,o)||this;return a.name=e,a.value=n,a.type=r||n&&n.type||null,a}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.name===t.name&&(this.value?!!t.value&&this.value.isEquivalent(t.value):!t.value)},e.prototype.visitStatement=function(t,e){return t.visitDeclareVarStmt(this,e)},e}(ne),ie=function(t){function e(e,n,r,i,o,a){void 0===o&&(o=null);var s=t.call(this,o,a)||this;return s.name=e,s.params=n,s.statements=r,s.type=i||null,s}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&xt(this.params,t.params)&&xt(this.statements,t.statements)},e.prototype.visitStatement=function(t,e){return t.visitDeclareFunctionStmt(this,e)},e}(ne),oe=function(t){function e(e,n){var r=t.call(this,null,n)||this;return r.expr=e,r}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.expr.isEquivalent(t.expr)},e.prototype.visitStatement=function(t,e){return t.visitExpressionStmt(this,e)},e}(ne),ae=function(t){function e(e,n){var r=t.call(this,null,n)||this;return r.value=e,r}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.value.isEquivalent(t.value)},e.prototype.visitStatement=function(t,e){return t.visitReturnStmt(this,e)},e}(ne),se=function(){function t(t,e){this.modifiers=e,e||(this.modifiers=[]),this.type=t||null}return t.prototype.hasModifier=function(t){return-1!==this.modifiers.indexOf(t)},t}(),ue=function(t){function e(e,n,r,i){void 0===r&&(r=null);var o=t.call(this,n,r)||this;return o.name=e,o.initializer=i,o}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return this.name===t.name},e}(se),ce=function(t){function e(e,n,r,i,o){void 0===o&&(o=null);var a=t.call(this,i,o)||this;return a.name=e,a.params=n,a.body=r,a}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return this.name===t.name&&xt(this.body,t.body)},e}(se),le=function(t){function e(e,n,r,i){void 0===i&&(i=null);var o=t.call(this,r,i)||this;return o.name=e,o.body=n,o}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return this.name===t.name&&xt(this.body,t.body)},e}(se),de=function(t){function e(e,n,r,i,o,a,s,u){void 0===s&&(s=null);var c=t.call(this,s,u)||this;return c.name=e,c.parent=n,c.fields=r,c.getters=i,c.constructorMethod=o,c.methods=a,c}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.name===t.name&&bt(this.parent,t.parent)&&xt(this.fields,t.fields)&&xt(this.getters,t.getters)&&this.constructorMethod.isEquivalent(t.constructorMethod)&&xt(this.methods,t.methods)},e.prototype.visitStatement=function(t,e){return t.visitDeclareClassStmt(this,e)},e}(ne),pe=function(t){function e(e,n,r,i){void 0===r&&(r=[]);var o=t.call(this,null,i)||this;return o.condition=e,o.trueCase=n,o.falseCase=r,o}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.condition.isEquivalent(t.condition)&&xt(this.trueCase,t.trueCase)&&xt(this.falseCase,t.falseCase)},e.prototype.visitStatement=function(t,e){return t.visitIfStmt(this,e)},e}(ne),fe=function(t){function e(e,n,r){void 0===n&&(n=!1);var i=t.call(this,null,r)||this;return i.comment=e,i.multiline=n,i}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e},e.prototype.visitStatement=function(t,e){return t.visitCommentStmt(this,e)},e}(ne),he=function(t){function e(e,n){void 0===e&&(e=[]);var r=t.call(this,null,n)||this;return r.tags=e,r}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.toString()===t.toString()},e.prototype.visitStatement=function(t,e){return t.visitJSDocCommentStmt(this,e)},e.prototype.toString=function(){return function(t){var e,n;if(0===t.length)return"";var r="*\n";try{for(var i=Object(o.__values)(t),a=i.next();!a.done;a=i.next()){var s=a.value;r+=" *",r+=Re(s).replace(/\n/g,"\n * "),r+="\n"}}catch(t){e={error:t}}finally{try{a&&!a.done&&(n=i.return)&&n.call(i)}finally{if(e)throw e.error}}return r+=" "}(this.tags)},e}(ne),me=function(t){function e(e,n,r){var i=t.call(this,null,r)||this;return i.bodyStmts=e,i.catchStmts=n,i}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&xt(this.bodyStmts,t.bodyStmts)&&xt(this.catchStmts,t.catchStmts)},e.prototype.visitStatement=function(t,e){return t.visitTryCatchStmt(this,e)},e}(ne),ge=function(t){function e(e,n){var r=t.call(this,null,n)||this;return r.error=e,r}return Object(o.__extends)(e,t),e.prototype.isEquivalent=function(t){return t instanceof me&&this.error.isEquivalent(t.error)},e.prototype.visitStatement=function(t,e){return t.visitThrowStmt(this,e)},e}(ne),ye=function(){function t(){}return t.prototype.transformExpr=function(t,e){return t},t.prototype.transformStmt=function(t,e){return t},t.prototype.visitReadVarExpr=function(t,e){return this.transformExpr(t,e)},t.prototype.visitWrappedNodeExpr=function(t,e){return this.transformExpr(t,e)},t.prototype.visitTypeofExpr=function(t,e){return this.transformExpr(new Mt(t.expr.visitExpression(this,e),t.type,t.sourceSpan),e)},t.prototype.visitWriteVarExpr=function(t,e){return this.transformExpr(new kt(t.name,t.value.visitExpression(this,e),t.type,t.sourceSpan),e)},t.prototype.visitWriteKeyExpr=function(t,e){return this.transformExpr(new At(t.receiver.visitExpression(this,e),t.index.visitExpression(this,e),t.value.visitExpression(this,e),t.type,t.sourceSpan),e)},t.prototype.visitWritePropExpr=function(t,e){return this.transformExpr(new Dt(t.receiver.visitExpression(this,e),t.name,t.value.visitExpression(this,e),t.type,t.sourceSpan),e)},t.prototype.visitInvokeMethodExpr=function(t,e){var n=t.builtin||t.name;return this.transformExpr(new Rt(t.receiver.visitExpression(this,e),n,this.visitAllExpressions(t.args,e),t.type,t.sourceSpan),e)},t.prototype.visitInvokeFunctionExpr=function(t,e){return this.transformExpr(new Nt(t.fn.visitExpression(this,e),this.visitAllExpressions(t.args,e),t.type,t.sourceSpan),e)},t.prototype.visitInstantiateExpr=function(t,e){return this.transformExpr(new Ft(t.classExpr.visitExpression(this,e),this.visitAllExpressions(t.args,e),t.type,t.sourceSpan),e)},t.prototype.visitLiteralExpr=function(t,e){return this.transformExpr(t,e)},t.prototype.visitExternalExpr=function(t,e){return this.transformExpr(t,e)},t.prototype.visitConditionalExpr=function(t,e){return this.transformExpr(new zt(t.condition.visitExpression(this,e),t.trueCase.visitExpression(this,e),t.falseCase.visitExpression(this,e),t.type,t.sourceSpan),e)},t.prototype.visitNotExpr=function(t,e){return this.transformExpr(new Ht(t.condition.visitExpression(this,e),t.sourceSpan),e)},t.prototype.visitAssertNotNullExpr=function(t,e){return this.transformExpr(new Ut(t.condition.visitExpression(this,e),t.sourceSpan),e)},t.prototype.visitCastExpr=function(t,e){return this.transformExpr(new Wt(t.value.visitExpression(this,e),t.type,t.sourceSpan),e)},t.prototype.visitFunctionExpr=function(t,e){return this.transformExpr(new Gt(t.params,this.visitAllStatements(t.statements,e),t.type,t.sourceSpan),e)},t.prototype.visitBinaryOperatorExpr=function(t,e){return this.transformExpr(new Yt(t.operator,t.lhs.visitExpression(this,e),t.rhs.visitExpression(this,e),t.type,t.sourceSpan),e)},t.prototype.visitReadPropExpr=function(t,e){return this.transformExpr(new Kt(t.receiver.visitExpression(this,e),t.name,t.type,t.sourceSpan),e)},t.prototype.visitReadKeyExpr=function(t,e){return this.transformExpr(new Xt(t.receiver.visitExpression(this,e),t.index.visitExpression(this,e),t.type,t.sourceSpan),e)},t.prototype.visitLiteralArrayExpr=function(t,e){return this.transformExpr(new Qt(this.visitAllExpressions(t.entries,e),t.type,t.sourceSpan),e)},t.prototype.visitLiteralMapExpr=function(t,e){var n=this,r=t.entries.map(function(t){return new Zt(t.key,t.value.visitExpression(n,e),t.quoted)}),i=new ht(t.valueType,null);return this.transformExpr(new $t(r,i,t.sourceSpan),e)},t.prototype.visitCommaExpr=function(t,e){return this.transformExpr(new Jt(this.visitAllExpressions(t.parts,e),t.sourceSpan),e)},t.prototype.visitAllExpressions=function(t,e){var n=this;return t.map(function(t){return t.visitExpression(n,e)})},t.prototype.visitDeclareVarStmt=function(t,e){var n=t.value&&t.value.visitExpression(this,e);return this.transformStmt(new re(t.name,n,t.type,t.modifiers,t.sourceSpan),e)},t.prototype.visitDeclareFunctionStmt=function(t,e){return this.transformStmt(new ie(t.name,t.params,this.visitAllStatements(t.statements,e),t.type,t.modifiers,t.sourceSpan),e)},t.prototype.visitExpressionStmt=function(t,e){return this.transformStmt(new oe(t.expr.visitExpression(this,e),t.sourceSpan),e)},t.prototype.visitReturnStmt=function(t,e){return this.transformStmt(new ae(t.value.visitExpression(this,e),t.sourceSpan),e)},t.prototype.visitDeclareClassStmt=function(t,e){var n=this,r=t.parent.visitExpression(this,e),i=t.getters.map(function(t){return new le(t.name,n.visitAllStatements(t.body,e),t.type,t.modifiers)}),o=t.constructorMethod&&new ce(t.constructorMethod.name,t.constructorMethod.params,this.visitAllStatements(t.constructorMethod.body,e),t.constructorMethod.type,t.constructorMethod.modifiers),a=t.methods.map(function(t){return new ce(t.name,t.params,n.visitAllStatements(t.body,e),t.type,t.modifiers)});return this.transformStmt(new de(t.name,r,t.fields,i,o,a,t.modifiers,t.sourceSpan),e)},t.prototype.visitIfStmt=function(t,e){return this.transformStmt(new pe(t.condition.visitExpression(this,e),this.visitAllStatements(t.trueCase,e),this.visitAllStatements(t.falseCase,e),t.sourceSpan),e)},t.prototype.visitTryCatchStmt=function(t,e){return this.transformStmt(new me(this.visitAllStatements(t.bodyStmts,e),this.visitAllStatements(t.catchStmts,e),t.sourceSpan),e)},t.prototype.visitThrowStmt=function(t,e){return this.transformStmt(new ge(t.error.visitExpression(this,e),t.sourceSpan),e)},t.prototype.visitCommentStmt=function(t,e){return this.transformStmt(t,e)},t.prototype.visitJSDocCommentStmt=function(t,e){return this.transformStmt(t,e)},t.prototype.visitAllStatements=function(t,e){var n=this;return t.map(function(t){return t.visitStatement(n,e)})},t}(),_e=function(){function t(){}return t.prototype.visitType=function(t,e){return t},t.prototype.visitExpression=function(t,e){return t.type&&t.type.visitType(this,e),t},t.prototype.visitBuiltinType=function(t,e){return this.visitType(t,e)},t.prototype.visitExpressionType=function(t,e){var n=this;return t.value.visitExpression(this,e),null!==t.typeParams&&t.typeParams.forEach(function(t){return n.visitType(t,e)}),this.visitType(t,e)},t.prototype.visitArrayType=function(t,e){return this.visitType(t,e)},t.prototype.visitMapType=function(t,e){return this.visitType(t,e)},t.prototype.visitWrappedNodeExpr=function(t,e){return t},t.prototype.visitTypeofExpr=function(t,e){return this.visitExpression(t,e)},t.prototype.visitReadVarExpr=function(t,e){return this.visitExpression(t,e)},t.prototype.visitWriteVarExpr=function(t,e){return t.value.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitWriteKeyExpr=function(t,e){return t.receiver.visitExpression(this,e),t.index.visitExpression(this,e),t.value.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitWritePropExpr=function(t,e){return t.receiver.visitExpression(this,e),t.value.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitInvokeMethodExpr=function(t,e){return t.receiver.visitExpression(this,e),this.visitAllExpressions(t.args,e),this.visitExpression(t,e)},t.prototype.visitInvokeFunctionExpr=function(t,e){return t.fn.visitExpression(this,e),this.visitAllExpressions(t.args,e),this.visitExpression(t,e)},t.prototype.visitInstantiateExpr=function(t,e){return t.classExpr.visitExpression(this,e),this.visitAllExpressions(t.args,e),this.visitExpression(t,e)},t.prototype.visitLiteralExpr=function(t,e){return this.visitExpression(t,e)},t.prototype.visitExternalExpr=function(t,e){var n=this;return t.typeParams&&t.typeParams.forEach(function(t){return t.visitType(n,e)}),this.visitExpression(t,e)},t.prototype.visitConditionalExpr=function(t,e){return t.condition.visitExpression(this,e),t.trueCase.visitExpression(this,e),t.falseCase.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitNotExpr=function(t,e){return t.condition.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitAssertNotNullExpr=function(t,e){return t.condition.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitCastExpr=function(t,e){return t.value.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitFunctionExpr=function(t,e){return this.visitAllStatements(t.statements,e),this.visitExpression(t,e)},t.prototype.visitBinaryOperatorExpr=function(t,e){return t.lhs.visitExpression(this,e),t.rhs.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitReadPropExpr=function(t,e){return t.receiver.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitReadKeyExpr=function(t,e){return t.receiver.visitExpression(this,e),t.index.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitLiteralArrayExpr=function(t,e){return this.visitAllExpressions(t.entries,e),this.visitExpression(t,e)},t.prototype.visitLiteralMapExpr=function(t,e){var n=this;return t.entries.forEach(function(t){return t.value.visitExpression(n,e)}),this.visitExpression(t,e)},t.prototype.visitCommaExpr=function(t,e){return this.visitAllExpressions(t.parts,e),this.visitExpression(t,e)},t.prototype.visitAllExpressions=function(t,e){var n=this;t.forEach(function(t){return t.visitExpression(n,e)})},t.prototype.visitDeclareVarStmt=function(t,e){return t.value&&t.value.visitExpression(this,e),t.type&&t.type.visitType(this,e),t},t.prototype.visitDeclareFunctionStmt=function(t,e){return this.visitAllStatements(t.statements,e),t.type&&t.type.visitType(this,e),t},t.prototype.visitExpressionStmt=function(t,e){return t.expr.visitExpression(this,e),t},t.prototype.visitReturnStmt=function(t,e){return t.value.visitExpression(this,e),t},t.prototype.visitDeclareClassStmt=function(t,e){var n=this;return t.parent.visitExpression(this,e),t.getters.forEach(function(t){return n.visitAllStatements(t.body,e)}),t.constructorMethod&&this.visitAllStatements(t.constructorMethod.body,e),t.methods.forEach(function(t){return n.visitAllStatements(t.body,e)}),t},t.prototype.visitIfStmt=function(t,e){return t.condition.visitExpression(this,e),this.visitAllStatements(t.trueCase,e),this.visitAllStatements(t.falseCase,e),t},t.prototype.visitTryCatchStmt=function(t,e){return this.visitAllStatements(t.bodyStmts,e),this.visitAllStatements(t.catchStmts,e),t},t.prototype.visitThrowStmt=function(t,e){return t.error.visitExpression(this,e),t},t.prototype.visitCommentStmt=function(t,e){return t},t.prototype.visitJSDocCommentStmt=function(t,e){return t},t.prototype.visitAllStatements=function(t,e){var n=this;t.forEach(function(t){return t.visitStatement(n,e)})},t}();function ve(t){var e=new be;return e.visitAllStatements(t,null),e.varNames}var be=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.varNames=new Set,e}return Object(o.__extends)(e,t),e.prototype.visitDeclareFunctionStmt=function(t,e){return t},e.prototype.visitDeclareClassStmt=function(t,e){return t},e.prototype.visitReadVarExpr=function(t,e){return t.name&&this.varNames.add(t.name),null},e}(_e);function xe(t){var e=new we;return e.visitAllStatements(t,null),e.externalReferences}var we=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.externalReferences=[],e}return Object(o.__extends)(e,t),e.prototype.visitExternalExpr=function(e,n){return this.externalReferences.push(e.value),t.prototype.visitExternalExpr.call(this,e,n)},e}(_e);function je(t,e){if(!e)return t;var n=new Ce(e);return t.visitStatement(n,null)}function Oe(t,e){if(!e)return t;var n=new Ce(e);return t.visitExpression(n,null)}var Ce=function(t){function e(e){var n=t.call(this)||this;return n.sourceSpan=e,n}return Object(o.__extends)(e,t),e.prototype._clone=function(t){var e=Object.create(t.constructor.prototype);for(var n in t)e[n]=t[n];return e},e.prototype.transformExpr=function(t,e){return t.sourceSpan||((t=this._clone(t)).sourceSpan=this.sourceSpan),t},e.prototype.transformStmt=function(t,e){return t.sourceSpan||((t=this._clone(t)).sourceSpan=this.sourceSpan),t},e}(ye);function Ee(t,e,n){return new St(t,e,n)}function Se(t,e,n){return void 0===e&&(e=null),new Bt(t,null,e,n)}function Me(t,e,n){return void 0===e&&(e=null),void 0===n&&(n=null),null!=t?Te(Se(t,e,null),n):null}function Te(t,e,n){return void 0===e&&(e=null),void 0===n&&(n=null),new pt(t,e,n)}function ke(t,e,n){return new Qt(t,e,n)}function Ae(t,e){return void 0===e&&(e=null),new $t(t.map(function(t){return new Zt(t.key,t.value,t.quoted)}),e,null)}function De(t,e,n,r,i){return new Gt(t,e,n,r,i)}function Ie(t,e,n){return new pe(t,e,n)}function Pe(t,e,n){return new Lt(t,e,n)}function Re(t){var e="";if(t.tagName&&(e+=" @"+t.tagName),t.text){if(t.text.match(/\/\*|\*\//))throw new Error('JSDoc text cannot contain "/*" and "*/"');e+=" "+t.text.replace(/@/g,"\\@")}return e}var Ne=/-+([a-z0-9])/g;function Fe(t,e){return Le(t,":",e)}function Le(t,e,n){var r=t.indexOf(e);return-1==r?n:[t.slice(0,r).trim(),t.slice(r+1).trim()]}function Be(t,e,n){return Array.isArray(t)?e.visitArray(t,n):"object"==typeof(r=t)&&null!==r&&Object.getPrototypeOf(r)===Je?e.visitStringMap(t,n):null==t||"string"==typeof t||"number"==typeof t||"boolean"==typeof t?e.visitPrimitive(t,n):e.visitOther(t,n);var r}function Ve(t){return null!==t&&void 0!==t}function ze(t){return void 0===t?null:t}var He=function(){function t(){}return t.prototype.visitArray=function(t,e){var n=this;return t.map(function(t){return Be(t,n,e)})},t.prototype.visitStringMap=function(t,e){var n=this,r={};return Object.keys(t).forEach(function(i){r[i]=Be(t[i],n,e)}),r},t.prototype.visitPrimitive=function(t,e){return t},t.prototype.visitOther=function(t,e){return t},t}(),Ue=function(t){if(rn(t))throw new Error("Illegal state: value cannot be a promise");return t},We=function(t,e){return rn(t)?t.then(e):e(t)},qe=function(t){return t.some(rn)?Promise.all(t):t};function Ge(t){throw new Error("Internal Error: "+t)}function Ye(t,e){var n=Error(t);return n[Ke]=!0,e&&(n[Xe]=e),n}var Ke="ngSyntaxError",Xe="ngParseErrors";function Qe(t){return t[Ke]}function Ze(t){return t[Xe]||[]}function $e(t){return t.replace(/([.*+?^=!:${}()|[\]\/\\])/g,"\\$1")}var Je=Object.getPrototypeOf({});function tn(t){for(var e="",n=0;n<t.length;n++){var r=t.charCodeAt(n);if(r>=55296&&r<=56319&&t.length>n+1){var i=t.charCodeAt(n+1);i>=56320&&i<=57343&&(n++,r=(r-55296<<10)+i-56320+65536)}r<=127?e+=String.fromCharCode(r):r<=2047?e+=String.fromCharCode(r>>6&31|192,63&r|128):r<=65535?e+=String.fromCharCode(r>>12|224,r>>6&63|128,63&r|128):r<=2097151&&(e+=String.fromCharCode(r>>18&7|240,r>>12&63|128,r>>6&63|128,63&r|128))}return e}function en(t){if("string"==typeof t)return t;if(t instanceof Array)return"["+t.map(en).join(", ")+"]";if(null==t)return""+t;if(t.overriddenName)return""+t.overriddenName;if(t.name)return""+t.name;var e=t.toString();if(null==e)return""+e;var n=e.indexOf("\n");return-1===n?e:e.substring(0,n)}function nn(t){return"function"==typeof t&&t.hasOwnProperty("__forward_ref__")?t():t}function rn(t){return!!t&&"function"==typeof t.then}var on=function(){return function(t){this.full=t;var e=t.split(".");this.major=e[0],this.minor=e[1],this.patch=e.slice(2).join(".")}}(),an="undefined"!=typeof window&&window,sn="undefined"!=typeof self&&"undefined"!=typeof WorkerGlobalScope&&self instanceof WorkerGlobalScope&&self,un="undefined"!=typeof global&&global||an||sn,cn={},ln=function(t){function e(e){var n=t.call(this,e.type)||this;return n.resolved=e,n.original=e,n}return Object(o.__extends)(e,t),e.prototype.visitExpression=function(t,e){return e===cn?this.original.visitExpression(t,e):this.resolved.visitExpression(t,e)},e.prototype.isEquivalent=function(t){return t instanceof e&&this.resolved.isEquivalent(t.resolved)},e.prototype.isConstant=function(){return!0},e.prototype.fixup=function(t){this.resolved=t,this.shared=!0},e}(Ot),dn=function(){function t(){this.statements=[],this.literals=new Map,this.literalFactories=new Map,this.injectorDefinitions=new Map,this.directiveDefinitions=new Map,this.componentDefinitions=new Map,this.pipeDefinitions=new Map,this.nextNameIndex=0}return t.prototype.getConstLiteral=function(t,e){if(t instanceof Lt||t instanceof ln)return t;var n=this.keyOf(t),r=this.literals.get(n),i=!1;if(r||(r=new ln(t),this.literals.set(n,r),i=!0),!i&&!r.shared||i&&e){var o=this.freshName();this.statements.push(Ee(o).set(t).toDeclStmt(gt,[It.Final])),r.fixup(Ee(o))}return r},t.prototype.getDefinition=function(t,e,n,r){void 0===r&&(r=!1);var i=this.definitionsOf(e),o=i.get(t),a=!1;if(!o){var s=this.propertyNameOf(e);o=new ln(n.importExpr(t).prop(s)),i.set(t,o),a=!0}if(!a&&!o.shared||a&&r){var u=this.freshName();this.statements.push(Ee(u).set(o.resolved).toDeclStmt(gt,[It.Final])),o.fixup(Ee(u))}return o},t.prototype.getLiteralFactory=function(t){if(t instanceof Qt){var e=t.entries.map(function(t){return t.isConstant()?t:Pe(null)}),n=this.keyOf(ke(e));return this._getLiteralFactory(n,t.entries,function(t){return ke(t)})}var r=Ae(t.entries.map(function(t){return{key:t.key,value:t.value.isConstant()?t.value:Pe(null),quoted:t.quoted}}));n=this.keyOf(r);return this._getLiteralFactory(n,t.entries.map(function(t){return t.value}),function(e){return Ae(e.map(function(e,n){return{key:t.entries[n].key,value:e,quoted:t.entries[n].quoted}}))})},t.prototype._getLiteralFactory=function(t,e,n){var r=this,i=this.literalFactories.get(t),o=e.filter(function(t){return!t.isConstant()});if(!i){var a=e.map(function(t,e){return t.isConstant()?r.getConstLiteral(t,!0):Ee("a"+e)}),s=De(a.filter(hn).map(function(t){return new qt(t.name,mt)}),[new ae(n(a))],gt),u=this.freshName();this.statements.push(Ee(u).set(s).toDeclStmt(gt,[It.Final])),i=Ee(u),this.literalFactories.set(t,i)}return{literalFactory:i,literalFactoryArguments:o}},t.prototype.uniqueName=function(t){return""+t+this.nextNameIndex++},t.prototype.definitionsOf=function(t){switch(t){case 2:return this.componentDefinitions;case 1:return this.directiveDefinitions;case 0:return this.injectorDefinitions;case 3:return this.pipeDefinitions}return Ge("Unknown definition kind "+t),this.componentDefinitions},t.prototype.propertyNameOf=function(t){switch(t){case 2:return"ngComponentDef";case 1:return"ngDirectiveDef";case 0:return"ngInjectorDef";case 3:return"ngPipeDef"}return Ge("Unknown definition kind "+t),"<unknown>"},t.prototype.freshName=function(){return this.uniqueName("_c")},t.prototype.keyOf=function(t){return t.visitExpression(new pn,cn)},t}(),pn=function(){function t(){this.visitWrappedNodeExpr=fn,this.visitWriteVarExpr=fn,this.visitWriteKeyExpr=fn,this.visitWritePropExpr=fn,this.visitInvokeMethodExpr=fn,this.visitInvokeFunctionExpr=fn,this.visitInstantiateExpr=fn,this.visitConditionalExpr=fn,this.visitNotExpr=fn,this.visitAssertNotNullExpr=fn,this.visitCastExpr=fn,this.visitFunctionExpr=fn,this.visitBinaryOperatorExpr=fn,this.visitReadPropExpr=fn,this.visitReadKeyExpr=fn,this.visitCommaExpr=fn}return t.prototype.visitLiteralExpr=function(t){return""+("string"==typeof t.value?'"'+t.value+'"':t.value)},t.prototype.visitLiteralArrayExpr=function(t,e){var n=this;return"["+t.entries.map(function(t){return t.visitExpression(n,e)}).join(",")+"]"},t.prototype.visitLiteralMapExpr=function(t,e){var n=this;return"{"+t.entries.map(function(t){return(i=(r=t).quoted?'"':"")+r.key+i+":"+t.value.visitExpression(n,e);var r,i}).join(",")},t.prototype.visitExternalExpr=function(t){return t.value.moduleName?"EX:"+t.value.moduleName+":"+t.value.name:"EX:"+t.value.runtime.name},t.prototype.visitReadVarExpr=function(t){return"VAR:"+t.name},t.prototype.visitTypeofExpr=function(t,e){return"TYPEOF:"+t.expr.visitExpression(this,e)},t}();function fn(t){throw new Error("Invalid state: Visitor "+this.constructor.name+" doesn't handle "+t.constructor.name)}function hn(t){return t instanceof St}var mn="@angular/core",gn=function(){function t(){}return t.ANALYZE_FOR_ENTRY_COMPONENTS={name:"ANALYZE_FOR_ENTRY_COMPONENTS",moduleName:mn},t.ElementRef={name:"ElementRef",moduleName:mn},t.NgModuleRef={name:"NgModuleRef",moduleName:mn},t.ViewContainerRef={name:"ViewContainerRef",moduleName:mn},t.ChangeDetectorRef={name:"ChangeDetectorRef",moduleName:mn},t.QueryList={name:"QueryList",moduleName:mn},t.TemplateRef={name:"TemplateRef",moduleName:mn},t.Renderer2={name:"Renderer2",moduleName:mn},t.CodegenComponentFactoryResolver={name:"ɵCodegenComponentFactoryResolver",moduleName:mn},t.ComponentFactoryResolver={name:"ComponentFactoryResolver",moduleName:mn},t.ComponentFactory={name:"ComponentFactory",moduleName:mn},t.ComponentRef={name:"ComponentRef",moduleName:mn},t.NgModuleFactory={name:"NgModuleFactory",moduleName:mn},t.createModuleFactory={name:"ɵcmf",moduleName:mn},t.moduleDef={name:"ɵmod",moduleName:mn},t.moduleProviderDef={name:"ɵmpd",moduleName:mn},t.RegisterModuleFactoryFn={name:"ɵregisterModuleFactory",moduleName:mn},t.inject={name:"inject",moduleName:mn},t.INJECTOR={name:"INJECTOR",moduleName:mn},t.Injector={name:"Injector",moduleName:mn},t.defineInjectable={name:"defineInjectable",moduleName:mn},t.InjectableDef={name:"ɵInjectableDef",moduleName:mn},t.ViewEncapsulation={name:"ViewEncapsulation",moduleName:mn},t.ChangeDetectionStrategy={name:"ChangeDetectionStrategy",moduleName:mn},t.SecurityContext={name:"SecurityContext",moduleName:mn},t.LOCALE_ID={name:"LOCALE_ID",moduleName:mn},t.TRANSLATIONS_FORMAT={name:"TRANSLATIONS_FORMAT",moduleName:mn},t.inlineInterpolate={name:"ɵinlineInterpolate",moduleName:mn},t.interpolate={name:"ɵinterpolate",moduleName:mn},t.EMPTY_ARRAY={name:"ɵEMPTY_ARRAY",moduleName:mn},t.EMPTY_MAP={name:"ɵEMPTY_MAP",moduleName:mn},t.Renderer={name:"Renderer",moduleName:mn},t.viewDef={name:"ɵvid",moduleName:mn},t.elementDef={name:"ɵeld",moduleName:mn},t.anchorDef={name:"ɵand",moduleName:mn},t.textDef={name:"ɵted",moduleName:mn},t.directiveDef={name:"ɵdid",moduleName:mn},t.providerDef={name:"ɵprd",moduleName:mn},t.queryDef={name:"ɵqud",moduleName:mn},t.pureArrayDef={name:"ɵpad",moduleName:mn},t.pureObjectDef={name:"ɵpod",moduleName:mn},t.purePipeDef={name:"ɵppd",moduleName:mn},t.pipeDef={name:"ɵpid",moduleName:mn},t.nodeValue={name:"ɵnov",moduleName:mn},t.ngContentDef={name:"ɵncd",moduleName:mn},t.unwrapValue={name:"ɵunv",moduleName:mn},t.createRendererType2={name:"ɵcrt",moduleName:mn},t.RendererType2={name:"RendererType2",moduleName:mn},t.ViewDefinition={name:"ɵViewDefinition",moduleName:mn},t.createComponentFactory={name:"ɵccf",moduleName:mn},t.setClassMetadata={name:"ɵsetClassMetadata",moduleName:mn},t}();function yn(t){return{identifier:{reference:t}}}function _n(t,e){return yn(t.resolveExternalReference(e))}var vn=function(){function t(t,e,n){this.filePath=t,this.name=e,this.members=n}return t.prototype.assertNoMembers=function(){if(this.members.length)throw new Error("Illegal state: symbol without members expected, but got "+JSON.stringify(this)+".")},t}(),bn=function(){function t(){this.cache=new Map}return t.prototype.get=function(t,e,n){var r='"'+t+'".'+e+((n=n||[]).length?"."+n.join("."):""),i=this.cache.get(r);return i||(i=new vn(t,e,n),this.cache.set(r,i)),i},t}(),xn=/^(?:(?:\[([^\]]+)\])|(?:\(([^\)]+)\)))|(\@[-\w]+)$/;function wn(t){return t.replace(/\W/g,"_")}var jn,On,Cn=0;function En(t){if(!t||!t.reference)return null;var e=t.reference;if(e instanceof vn)return e.name;if(e.__anonymousType)return e.__anonymousType;var n=en(e);return n.indexOf("(")>=0?(n="anonymous_"+Cn++,e.__anonymousType=n):n=wn(n),n}function Sn(t){var e=t.reference;return e instanceof vn?e.filePath:"./"+en(e)}function Mn(t,e){return"View_"+En({reference:t})+"_"+e}function Tn(t){return"RenderType_"+En({reference:t})}function kn(t){return"HostView_"+En({reference:t})}function An(t){return En({reference:t})+"NgFactory"}function Dn(t){return null!=t.value?wn(t.value):En(t.identifier)}function In(t){return null!=t.identifier?t.identifier.reference:t.value}(On=jn||(jn={}))[On.Pipe=0]="Pipe",On[On.Directive=1]="Directive",On[On.NgModule=2]="NgModule",On[On.Injectable=3]="Injectable";var Pn=function(){return function(t){var e=void 0===t?{}:t,n=e.moduleUrl,r=e.styles,i=e.styleUrls;this.moduleUrl=n||null,this.styles=zn(r),this.styleUrls=zn(i)}}(),Rn=function(){function t(t){var e=t.encapsulation,n=t.template,r=t.templateUrl,i=t.htmlAst,o=t.styles,a=t.styleUrls,s=t.externalStylesheets,u=t.animations,c=t.ngContentSelectors,l=t.interpolation,d=t.isInline,p=t.preserveWhitespaces;if(this.encapsulation=e,this.template=n,this.templateUrl=r,this.htmlAst=i,this.styles=zn(o),this.styleUrls=zn(a),this.externalStylesheets=zn(s),this.animations=u?Un(u):[],this.ngContentSelectors=c||[],l&&2!=l.length)throw new Error("'interpolation' should have a start and an end symbol.");this.interpolation=l,this.isInline=d,this.preserveWhitespaces=p}return t.prototype.toSummary=function(){return{ngContentSelectors:this.ngContentSelectors,encapsulation:this.encapsulation,styles:this.styles,animations:this.animations}},t}(),Nn=function(){function t(t){var e=t.isHost,n=t.type,r=t.isComponent,i=t.selector,o=t.exportAs,a=t.changeDetection,s=t.inputs,u=t.outputs,c=t.hostListeners,l=t.hostProperties,d=t.hostAttributes,p=t.providers,f=t.viewProviders,h=t.queries,m=t.guards,g=t.viewQueries,y=t.entryComponents,_=t.template,v=t.componentViewType,b=t.rendererType,x=t.componentFactory;this.isHost=!!e,this.type=n,this.isComponent=r,this.selector=i,this.exportAs=o,this.changeDetection=a,this.inputs=s,this.outputs=u,this.hostListeners=c,this.hostProperties=l,this.hostAttributes=d,this.providers=zn(p),this.viewProviders=zn(f),this.queries=zn(h),this.guards=m,this.viewQueries=zn(g),this.entryComponents=zn(y),this.template=_,this.componentViewType=v,this.rendererType=b,this.componentFactory=x}return t.create=function(e){var n=e.isHost,r=e.type,i=e.isComponent,o=e.selector,a=e.exportAs,s=e.changeDetection,u=e.inputs,c=e.outputs,l=e.host,d=e.providers,p=e.viewProviders,f=e.queries,h=e.guards,m=e.viewQueries,g=e.entryComponents,y=e.template,_=e.componentViewType,v=e.rendererType,b=e.componentFactory,x={},w={},j={};null!=l&&Object.keys(l).forEach(function(t){var e=l[t],n=t.match(xn);null===n?j[t]=e:null!=n[1]?w[n[1]]=e:null!=n[2]&&(x[n[2]]=e)});var O={};null!=u&&u.forEach(function(t){var e=Fe(t,[t,t]);O[e[0]]=e[1]});var C={};return null!=c&&c.forEach(function(t){var e=Fe(t,[t,t]);C[e[0]]=e[1]}),new t({isHost:n,type:r,isComponent:!!i,selector:o,exportAs:a,changeDetection:s,inputs:O,outputs:C,hostListeners:x,hostProperties:w,hostAttributes:j,providers:d,viewProviders:p,queries:f,guards:h,viewQueries:m,entryComponents:g,template:y,componentViewType:_,rendererType:v,componentFactory:b})},t.prototype.toSummary=function(){return{summaryKind:jn.Directive,type:this.type,isComponent:this.isComponent,selector:this.selector,exportAs:this.exportAs,inputs:this.inputs,outputs:this.outputs,hostListeners:this.hostListeners,hostProperties:this.hostProperties,hostAttributes:this.hostAttributes,providers:this.providers,viewProviders:this.viewProviders,queries:this.queries,guards:this.guards,viewQueries:this.viewQueries,entryComponents:this.entryComponents,changeDetection:this.changeDetection,template:this.template&&this.template.toSummary(),componentViewType:this.componentViewType,rendererType:this.rendererType,componentFactory:this.componentFactory}},t}(),Fn=function(){function t(t){var e=t.type,n=t.name,r=t.pure;this.type=e,this.name=n,this.pure=!!r}return t.prototype.toSummary=function(){return{summaryKind:jn.Pipe,type:this.type,name:this.name,pure:this.pure}},t}(),Ln=function(){return function(){}}(),Bn=function(){function t(t){var e=t.type,n=t.providers,r=t.declaredDirectives,i=t.exportedDirectives,o=t.declaredPipes,a=t.exportedPipes,s=t.entryComponents,u=t.bootstrapComponents,c=t.importedModules,l=t.exportedModules,d=t.schemas,p=t.transitiveModule,f=t.id;this.type=e||null,this.declaredDirectives=zn(r),this.exportedDirectives=zn(i),this.declaredPipes=zn(o),this.exportedPipes=zn(a),this.providers=zn(n),this.entryComponents=zn(s),this.bootstrapComponents=zn(u),this.importedModules=zn(c),this.exportedModules=zn(l),this.schemas=zn(d),this.id=f||null,this.transitiveModule=p||null}return t.prototype.toSummary=function(){var t=this.transitiveModule;return{summaryKind:jn.NgModule,type:this.type,entryComponents:t.entryComponents,providers:t.providers,modules:t.modules,exportedDirectives:t.exportedDirectives,exportedPipes:t.exportedPipes}},t}(),Vn=function(){function t(){this.directivesSet=new Set,this.directives=[],this.exportedDirectivesSet=new Set,this.exportedDirectives=[],this.pipesSet=new Set,this.pipes=[],this.exportedPipesSet=new Set,this.exportedPipes=[],this.modulesSet=new Set,this.modules=[],this.entryComponentsSet=new Set,this.entryComponents=[],this.providers=[]}return t.prototype.addProvider=function(t,e){this.providers.push({provider:t,module:e})},t.prototype.addDirective=function(t){this.directivesSet.has(t.reference)||(this.directivesSet.add(t.reference),this.directives.push(t))},t.prototype.addExportedDirective=function(t){this.exportedDirectivesSet.has(t.reference)||(this.exportedDirectivesSet.add(t.reference),this.exportedDirectives.push(t))},t.prototype.addPipe=function(t){this.pipesSet.has(t.reference)||(this.pipesSet.add(t.reference),this.pipes.push(t))},t.prototype.addExportedPipe=function(t){this.exportedPipesSet.has(t.reference)||(this.exportedPipesSet.add(t.reference),this.exportedPipes.push(t))},t.prototype.addModule=function(t){this.modulesSet.has(t.reference)||(this.modulesSet.add(t.reference),this.modules.push(t))},t.prototype.addEntryComponent=function(t){this.entryComponentsSet.has(t.componentType)||(this.entryComponentsSet.add(t.componentType),this.entryComponents.push(t))},t}();function zn(t){return t||[]}var Hn=function(){return function(t,e){var n=e.useClass,r=e.useValue,i=e.useExisting,o=e.useFactory,a=e.deps,s=e.multi;this.token=t,this.useClass=n||null,this.useValue=r,this.useExisting=i,this.useFactory=o||null,this.dependencies=a||null,this.multi=!!s}}();function Un(t){return t.reduce(function(t,e){var n=Array.isArray(e)?Un(e):e;return t.concat(n)},[])}function Wn(t){return t.replace(/(\w+:\/\/[\w:-]+)?(\/+)?/,"ng:///")}function qn(t,e,n){var r;return r=n.isInline?e.type.reference instanceof vn?e.type.reference.filePath+"."+e.type.reference.name+".html":En(t)+"/"+En(e.type)+".html":n.templateUrl,e.type.reference instanceof vn?r:Wn(r)}function Gn(t,e){var n=t.moduleUrl.split(/\/\\/g);return Wn("css/"+e+n[n.length-1]+".ngstyle.js")}function Yn(t){return Wn(En(t.type)+"/module.ngfactory.js")}function Kn(t,e){return Wn(En(t)+"/"+En(e.type)+".ngfactory.js")}var Xn="@angular/core",Qn=function(){function t(){}return t.NEW_METHOD="factory",t.TRANSFORM_METHOD="transform",t.PATCH_DEPS="patchedDeps",t.namespaceHTML={name:"ɵnamespaceHTML",moduleName:Xn},t.namespaceMathML={name:"ɵnamespaceMathML",moduleName:Xn},t.namespaceSVG={name:"ɵnamespaceSVG",moduleName:Xn},t.element={name:"ɵelement",moduleName:Xn},t.elementStart={name:"ɵelementStart",moduleName:Xn},t.elementEnd={name:"ɵelementEnd",moduleName:Xn},t.elementProperty={name:"ɵelementProperty",moduleName:Xn},t.componentHostSyntheticProperty={name:"ɵcomponentHostSyntheticProperty",moduleName:Xn},t.elementAttribute={name:"ɵelementAttribute",moduleName:Xn},t.elementClassProp={name:"ɵelementClassProp",moduleName:Xn},t.elementContainerStart={name:"ɵelementContainerStart",moduleName:Xn},t.elementContainerEnd={name:"ɵelementContainerEnd",moduleName:Xn},t.elementStyling={name:"ɵelementStyling",moduleName:Xn},t.elementHostAttrs={name:"ɵelementHostAttrs",moduleName:Xn},t.elementStylingMap={name:"ɵelementStylingMap",moduleName:Xn},t.elementStyleProp={name:"ɵelementStyleProp",moduleName:Xn},t.elementStylingApply={name:"ɵelementStylingApply",moduleName:Xn},t.containerCreate={name:"ɵcontainer",moduleName:Xn},t.nextContext={name:"ɵnextContext",moduleName:Xn},t.templateCreate={name:"ɵtemplate",moduleName:Xn},t.text={name:"ɵtext",moduleName:Xn},t.textBinding={name:"ɵtextBinding",moduleName:Xn},t.bind={name:"ɵbind",moduleName:Xn},t.enableBindings={name:"ɵenableBindings",moduleName:Xn},t.disableBindings={name:"ɵdisableBindings",moduleName:Xn},t.allocHostVars={name:"ɵallocHostVars",moduleName:Xn},t.getCurrentView={name:"ɵgetCurrentView",moduleName:Xn},t.restoreView={name:"ɵrestoreView",moduleName:Xn},t.interpolation1={name:"ɵinterpolation1",moduleName:Xn},t.interpolation2={name:"ɵinterpolation2",moduleName:Xn},t.interpolation3={name:"ɵinterpolation3",moduleName:Xn},t.interpolation4={name:"ɵinterpolation4",moduleName:Xn},t.interpolation5={name:"ɵinterpolation5",moduleName:Xn},t.interpolation6={name:"ɵinterpolation6",moduleName:Xn},t.interpolation7={name:"ɵinterpolation7",moduleName:Xn},t.interpolation8={name:"ɵinterpolation8",moduleName:Xn},t.interpolationV={name:"ɵinterpolationV",moduleName:Xn},t.pureFunction0={name:"ɵpureFunction0",moduleName:Xn},t.pureFunction1={name:"ɵpureFunction1",moduleName:Xn},t.pureFunction2={name:"ɵpureFunction2",moduleName:Xn},t.pureFunction3={name:"ɵpureFunction3",moduleName:Xn},t.pureFunction4={name:"ɵpureFunction4",moduleName:Xn},t.pureFunction5={name:"ɵpureFunction5",moduleName:Xn},t.pureFunction6={name:"ɵpureFunction6",moduleName:Xn},t.pureFunction7={name:"ɵpureFunction7",moduleName:Xn},t.pureFunction8={name:"ɵpureFunction8",moduleName:Xn},t.pureFunctionV={name:"ɵpureFunctionV",moduleName:Xn},t.pipeBind1={name:"ɵpipeBind1",moduleName:Xn},t.pipeBind2={name:"ɵpipeBind2",moduleName:Xn},t.pipeBind3={name:"ɵpipeBind3",moduleName:Xn},t.pipeBind4={name:"ɵpipeBind4",moduleName:Xn},t.pipeBindV={name:"ɵpipeBindV",moduleName:Xn},t.i18n={name:"ɵi18n",moduleName:Xn},t.i18nAttributes={name:"ɵi18nAttributes",moduleName:Xn},t.i18nExp={name:"ɵi18nExp",moduleName:Xn},t.i18nStart={name:"ɵi18nStart",moduleName:Xn},t.i18nEnd={name:"ɵi18nEnd",moduleName:Xn},t.i18nApply={name:"ɵi18nApply",moduleName:Xn},t.i18nPostprocess={name:"ɵi18nPostprocess",moduleName:Xn},t.load={name:"ɵload",moduleName:Xn},t.loadQueryList={name:"ɵloadQueryList",moduleName:Xn},t.pipe={name:"ɵpipe",moduleName:Xn},t.projection={name:"ɵprojection",moduleName:Xn},t.projectionDef={name:"ɵprojectionDef",moduleName:Xn},t.reference={name:"ɵreference",moduleName:Xn},t.inject={name:"inject",moduleName:Xn},t.injectAttribute={name:"ɵinjectAttribute",moduleName:Xn},t.directiveInject={name:"ɵdirectiveInject",moduleName:Xn},t.templateRefExtractor={name:"ɵtemplateRefExtractor",moduleName:Xn},t.defineBase={name:"ɵdefineBase",moduleName:Xn},t.BaseDef={name:"ɵBaseDef",moduleName:Xn},t.defineComponent={name:"ɵdefineComponent",moduleName:Xn},t.ComponentDefWithMeta={name:"ɵComponentDefWithMeta",moduleName:Xn},t.defineDirective={name:"ɵdefineDirective",moduleName:Xn},t.DirectiveDefWithMeta={name:"ɵDirectiveDefWithMeta",moduleName:Xn},t.InjectorDef={name:"ɵInjectorDef",moduleName:Xn},t.defineInjector={name:"defineInjector",moduleName:Xn},t.NgModuleDefWithMeta={name:"ɵNgModuleDefWithMeta",moduleName:Xn},t.defineNgModule={name:"ɵdefineNgModule",moduleName:Xn},t.PipeDefWithMeta={name:"ɵPipeDefWithMeta",moduleName:Xn},t.definePipe={name:"ɵdefinePipe",moduleName:Xn},t.query={name:"ɵquery",moduleName:Xn},t.queryRefresh={name:"ɵqueryRefresh",moduleName:Xn},t.registerContentQuery={name:"ɵregisterContentQuery",moduleName:Xn},t.NgOnChangesFeature={name:"ɵNgOnChangesFeature",moduleName:Xn},t.InheritDefinitionFeature={name:"ɵInheritDefinitionFeature",moduleName:Xn},t.ProvidersFeature={name:"ɵProvidersFeature",moduleName:Xn},t.listener={name:"ɵlistener",moduleName:Xn},t.getFactoryOf={name:"ɵgetFactoryOf",moduleName:Xn},t.getInheritedFactory={name:"ɵgetInheritedFactory",moduleName:Xn},t.sanitizeHtml={name:"ɵsanitizeHtml",moduleName:Xn},t.sanitizeStyle={name:"ɵsanitizeStyle",moduleName:Xn},t.defaultStyleSanitizer={name:"ɵdefaultStyleSanitizer",moduleName:Xn},t.sanitizeResourceUrl={name:"ɵsanitizeResourceUrl",moduleName:Xn},t.sanitizeScript={name:"ɵsanitizeScript",moduleName:Xn},t.sanitizeUrl={name:"ɵsanitizeUrl",moduleName:Xn},t}(),Zn=function(){return function(t,e,n,r,i,o){this.nodes=t,this.placeholders=e,this.placeholderToMessage=n,this.meaning=r,this.description=i,this.id=o,t.length?this.sources=[{filePath:t[0].sourceSpan.start.file.url,startLine:t[0].sourceSpan.start.line+1,startCol:t[0].sourceSpan.start.col+1,endLine:t[t.length-1].sourceSpan.end.line+1,endCol:t[0].sourceSpan.start.col+1}]:this.sources=[]}}(),$n=function(){function t(t,e){this.value=t,this.sourceSpan=e}return t.prototype.visit=function(t,e){return t.visitText(this,e)},t}(),Jn=function(){function t(t,e){this.children=t,this.sourceSpan=e}return t.prototype.visit=function(t,e){return t.visitContainer(this,e)},t}(),tr=function(){function t(t,e,n,r){this.expression=t,this.type=e,this.cases=n,this.sourceSpan=r}return t.prototype.visit=function(t,e){return t.visitIcu(this,e)},t}(),er=function(){function t(t,e,n,r,i,o,a){this.tag=t,this.attrs=e,this.startName=n,this.closeName=r,this.children=i,this.isVoid=o,this.sourceSpan=a}return t.prototype.visit=function(t,e){return t.visitTagPlaceholder(this,e)},t}(),nr=function(){function t(t,e,n){this.value=t,this.name=e,this.sourceSpan=n}return t.prototype.visit=function(t,e){return t.visitPlaceholder(this,e)},t}(),rr=function(){function t(t,e,n){this.value=t,this.name=e,this.sourceSpan=n}return t.prototype.visit=function(t,e){return t.visitIcuPlaceholder(this,e)},t}(),ir=function(){function t(){}return t.prototype.visitText=function(t,e){return new $n(t.value,t.sourceSpan)},t.prototype.visitContainer=function(t,e){var n=this,r=t.children.map(function(t){return t.visit(n,e)});return new Jn(r,t.sourceSpan)},t.prototype.visitIcu=function(t,e){var n=this,r={};Object.keys(t.cases).forEach(function(i){return r[i]=t.cases[i].visit(n,e)});var i=new tr(t.expression,t.type,r,t.sourceSpan);return i.expressionPlaceholder=t.expressionPlaceholder,i},t.prototype.visitTagPlaceholder=function(t,e){var n=this,r=t.children.map(function(t){return t.visit(n,e)});return new er(t.tag,t.attrs,t.startName,t.closeName,r,t.isVoid,t.sourceSpan)},t.prototype.visitPlaceholder=function(t,e){return new nr(t.value,t.name,t.sourceSpan)},t.prototype.visitIcuPlaceholder=function(t,e){return new rr(t.value,t.name,t.sourceSpan)},t}(),or=function(){function t(){}return t.prototype.visitText=function(t,e){},t.prototype.visitContainer=function(t,e){var n=this;t.children.forEach(function(t){return t.visit(n)})},t.prototype.visitIcu=function(t,e){var n=this;Object.keys(t.cases).forEach(function(e){t.cases[e].visit(n)})},t.prototype.visitTagPlaceholder=function(t,e){var n=this;t.children.forEach(function(t){return t.visit(n)})},t.prototype.visitPlaceholder=function(t,e){},t.prototype.visitIcuPlaceholder=function(t,e){},t}();function ar(t){return t.id||function(t){var e,n,r=tn(t),i=function(t,e){for(var n=Array(t.length+3>>>2),r=0;r<n.length;r++)n[r]=wr(t,4*r,e);return n}(r,lr.Big),a=8*r.length,s=new Array(80),u=Object(o.__read)([1732584193,4023233417,2562383102,271733878,3285377520],5),c=u[0],l=u[1],d=u[2],p=u[3],f=u[4];i[a>>5]|=128<<24-a%32,i[15+(a+64>>9<<4)]=a;for(var h=0;h<i.length;h+=16){for(var m=Object(o.__read)([c,l,d,p,f],5),g=m[0],y=m[1],_=m[2],v=m[3],b=m[4],x=0;x<80;x++){s[x]=x<16?i[h+x]:br(s[x-3]^s[x-8]^s[x-14]^s[x-16],1);var w=Object(o.__read)(fr(x,l,d,p),2),j=w[0],O=w[1],C=[br(c,5),j,f,O,s[x]].reduce(yr);e=Object(o.__read)([p,d,br(l,30),c,C],5),f=e[0],p=e[1],d=e[2],l=e[3],c=e[4]}n=Object(o.__read)([yr(c,g),yr(l,y),yr(d,_),yr(p,v),yr(f,b)],5),c=n[0],l=n[1],d=n[2],p=n[3],f=n[4]}return function(t){for(var e="",n=0;n<t.length;n++){var r=xr(t,n);e+=(r>>>4).toString(16)+(15&r).toString(16)}return e.toLowerCase()}(jr([c,l,d,p,f]))}((e=t.nodes,e.map(function(t){return t.visit(cr,null)})).join("")+"["+t.meaning+"]");var e}function sr(t){if(t.id)return t.id;var e=new pr;return function(t,e){var n,r=Object(o.__read)(hr(t),2),i=r[0],a=r[1];if(e){var s=Object(o.__read)(hr(e),2),u=s[0],c=s[1];n=Object(o.__read)((x=[i,a],w=1,j=Object(o.__read)(x,2),O=j[0],C=j[1],l=[O<<w|C>>>32-w,C<<w|O>>>32-w],d=[u,c],p=Object(o.__read)(l,2),f=p[0],h=p[1],m=Object(o.__read)(d,2),g=m[0],y=m[1],_=Object(o.__read)(_r(h,y),2),v=_[0],b=_[1],[yr(yr(f,g),v),b]),2),i=n[0],a=n[1]}var l,d,p,f,h,m,g,y,_,v,b;var x,w,j,O,C;return function(t){for(var e="",n="1",r=t.length-1;r>=0;r--)e=Or(e,Cr(xr(t,r),n)),n=Cr(256,n);return e.split("").reverse().join("")}(jr([2147483647&i,a]))}(t.nodes.map(function(t){return t.visit(e,null)}).join(""),t.meaning)}var ur=function(){function t(){}return t.prototype.visitText=function(t,e){return t.value},t.prototype.visitContainer=function(t,e){var n=this;return"["+t.children.map(function(t){return t.visit(n)}).join(", ")+"]"},t.prototype.visitIcu=function(t,e){var n=this,r=Object.keys(t.cases).map(function(e){return e+" {"+t.cases[e].visit(n)+"}"});return"{"+t.expression+", "+t.type+", "+r.join(", ")+"}"},t.prototype.visitTagPlaceholder=function(t,e){var n=this;return t.isVoid?'<ph tag name="'+t.startName+'"/>':'<ph tag name="'+t.startName+'">'+t.children.map(function(t){return t.visit(n)}).join(", ")+'</ph name="'+t.closeName+'">'},t.prototype.visitPlaceholder=function(t,e){return t.value?'<ph name="'+t.name+'">'+t.value+"</ph>":'<ph name="'+t.name+'"/>'},t.prototype.visitIcuPlaceholder=function(t,e){return'<ph icu name="'+t.name+'">'+t.value.visit(this)+"</ph>"},t}(),cr=new ur;var lr,dr,pr=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(o.__extends)(e,t),e.prototype.visitIcu=function(t,e){var n=this,r=Object.keys(t.cases).map(function(e){return e+" {"+t.cases[e].visit(n)+"}"});return"{"+t.type+", "+r.join(", ")+"}"},e}(ur);function fr(t,e,n,r){return t<20?[e&n|~e&r,1518500249]:t<40?[e^n^r,1859775393]:t<60?[e&n|e&r|n&r,2400959708]:[e^n^r,3395469782]}function hr(t){var e=tn(t),n=Object(o.__read)([mr(e,0),mr(e,102072)],2),r=n[0],i=n[1];return 0!=r||0!=i&&1!=i||(r^=319790063,i^=-1801410264),[r,i]}function mr(t,e){var n,r,i=Object(o.__read)([2654435769,2654435769],2),a=i[0],s=i[1],u=t.length;for(r=0;r+12<=u;r+=12)a=yr(a,wr(t,r,lr.Little)),s=yr(s,wr(t,r+4,lr.Little)),e=yr(e,wr(t,r+8,lr.Little)),a=(n=Object(o.__read)(gr([a,s,e]),3))[0],s=n[1],e=n[2];return gr([a=yr(a,wr(t,r,lr.Little)),s=yr(s,wr(t,r+4,lr.Little)),e=yr(e=yr(e,u),wr(t,r+8,lr.Little)<<8)])[2]}function gr(t){var e=Object(o.__read)(t,3),n=e[0],r=e[1],i=e[2];return n=vr(n=vr(n,r),i),n^=i>>>13,r=vr(r=vr(r,i),n),r^=n<<8,i=vr(i=vr(i,n),r),i^=r>>>13,n=vr(n=vr(n,r),i),n^=i>>>12,r=vr(r=vr(r,i),n),r^=n<<16,i=vr(i=vr(i,n),r),i^=r>>>5,n=vr(n=vr(n,r),i),n^=i>>>3,r=vr(r=vr(r,i),n),r^=n<<10,i=vr(i=vr(i,n),r),[n,r,i^=r>>>15]}function yr(t,e){return _r(t,e)[1]}function _r(t,e){var n=(65535&t)+(65535&e),r=(t>>>16)+(e>>>16)+(n>>>16);return[r>>>16,r<<16|65535&n]}function vr(t,e){var n=(65535&t)-(65535&e);return(t>>16)-(e>>16)+(n>>16)<<16|65535&n}function br(t,e){return t<<e|t>>>32-e}function xr(t,e){return e>=t.length?0:255&t.charCodeAt(e)}function wr(t,e,n){var r=0;if(n===lr.Big)for(var i=0;i<4;i++)r+=xr(t,e+i)<<24-8*i;else for(i=0;i<4;i++)r+=xr(t,e+i)<<8*i;return r}function jr(t){return t.reduce(function(t,e){return t+function(t){for(var e="",n=0;n<4;n++)e+=String.fromCharCode(t>>>8*(3-n)&255);return e}(e)},"")}function Or(t,e){for(var n="",r=Math.max(t.length,e.length),i=0,o=0;i<r||o;i++){var a=o+ +(t[i]||0)+ +(e[i]||0);a>=10?(o=1,n+=a-10):(o=0,n+=a)}return n}function Cr(t,e){for(var n="",r=e;0!==t;t>>>=1)1&t&&(n=Or(n,r)),r=Or(r,r);return n}(dr=lr||(lr={}))[dr.Little=0]="Little",dr[dr.Big=1]="Big";var Er=function(){function t(){}return t.prototype.createNameMapper=function(t){return null},t}(),Sr=function(t){function e(e,n){var r=t.call(this)||this;return r.mapName=n,r.internalToPublic={},r.publicToNextId={},r.publicToInternal={},e.nodes.forEach(function(t){return t.visit(r)}),r}return Object(o.__extends)(e,t),e.prototype.toPublicName=function(t){return this.internalToPublic.hasOwnProperty(t)?this.internalToPublic[t]:null},e.prototype.toInternalName=function(t){return this.publicToInternal.hasOwnProperty(t)?this.publicToInternal[t]:null},e.prototype.visitText=function(t,e){return null},e.prototype.visitTagPlaceholder=function(e,n){this.visitPlaceholderName(e.startName),t.prototype.visitTagPlaceholder.call(this,e,n),this.visitPlaceholderName(e.closeName)},e.prototype.visitPlaceholder=function(t,e){this.visitPlaceholderName(t.name)},e.prototype.visitIcuPlaceholder=function(t,e){this.visitPlaceholderName(t.name)},e.prototype.visitPlaceholderName=function(t){if(t&&!this.internalToPublic.hasOwnProperty(t)){var e=this.mapName(t);if(this.publicToInternal.hasOwnProperty(e)){var n=this.publicToNextId[e];this.publicToNextId[e]=n+1,e=e+"_"+n}else this.publicToNextId[e]=1;this.internalToPublic[t]=e,this.publicToInternal[e]=t}},e}(or),Mr=new(function(){function t(){}return t.prototype.visitTag=function(t){var e=this,n=this._serializeAttributes(t.attrs);if(0==t.children.length)return"<"+t.name+n+"/>";var r=t.children.map(function(t){return t.visit(e)});return"<"+t.name+n+">"+r.join("")+"</"+t.name+">"},t.prototype.visitText=function(t){return t.value},t.prototype.visitDeclaration=function(t){return"<?xml"+this._serializeAttributes(t.attrs)+" ?>"},t.prototype._serializeAttributes=function(t){var e=Object.keys(t).map(function(e){return e+'="'+t[e]+'"'}).join(" ");return e.length>0?" "+e:""},t.prototype.visitDoctype=function(t){return"<!DOCTYPE "+t.rootTag+" [\n"+t.dtd+"\n]>"},t}());function Tr(t){return t.map(function(t){return t.visit(Mr)}).join("")}var kr=function(){function t(t){var e=this;this.attrs={},Object.keys(t).forEach(function(n){e.attrs[n]=Nr(t[n])})}return t.prototype.visit=function(t){return t.visitDeclaration(this)},t}(),Ar=function(){function t(t,e){this.rootTag=t,this.dtd=e}return t.prototype.visit=function(t){return t.visitDoctype(this)},t}(),Dr=function(){function t(t,e,n){void 0===e&&(e={}),void 0===n&&(n=[]);var r=this;this.name=t,this.children=n,this.attrs={},Object.keys(e).forEach(function(t){r.attrs[t]=Nr(e[t])})}return t.prototype.visit=function(t){return t.visitTag(this)},t}(),Ir=function(){function t(t){this.value=Nr(t)}return t.prototype.visit=function(t){return t.visitText(this)},t}(),Pr=function(t){function e(e){return void 0===e&&(e=0),t.call(this,"\n"+new Array(e+1).join(" "))||this}return Object(o.__extends)(e,t),e}(Ir),Rr=[[/&/g,"&amp;"],[/"/g,"&quot;"],[/'/g,"&apos;"],[/</g,"&lt;"],[/>/g,"&gt;"]];function Nr(t){return Rr.reduce(function(t,e){return t.replace(e[0],e[1])},t)}var Fr=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(o.__extends)(e,t),e.prototype.write=function(t,e){var n=new Vr,r=new Lr,i=new Dr("messagebundle");return t.forEach(function(t){var e={id:t.id};t.description&&(e.desc=t.description),t.meaning&&(e.meaning=t.meaning);var n=[];t.sources.forEach(function(t){n.push(new Dr("source",{},[new Ir(t.filePath+":"+t.startLine+(t.endLine!==t.startLine?","+t.endLine:""))]))}),i.children.push(new Pr(2),new Dr("msg",e,Object(o.__spread)(n,r.serialize(t.nodes))))}),i.children.push(new Pr),Tr([new kr({version:"1.0",encoding:"UTF-8"}),new Pr,new Ar("messagebundle",'<!ELEMENT messagebundle (msg)*>\n<!ATTLIST messagebundle class CDATA #IMPLIED>\n\n<!ELEMENT msg (#PCDATA|ph|source)*>\n<!ATTLIST msg id CDATA #IMPLIED>\n<!ATTLIST msg seq CDATA #IMPLIED>\n<!ATTLIST msg name CDATA #IMPLIED>\n<!ATTLIST msg desc CDATA #IMPLIED>\n<!ATTLIST msg meaning CDATA #IMPLIED>\n<!ATTLIST msg obsolete (obsolete) #IMPLIED>\n<!ATTLIST msg xml:space (default|preserve) "default">\n<!ATTLIST msg is_hidden CDATA #IMPLIED>\n\n<!ELEMENT source (#PCDATA)>\n\n<!ELEMENT ph (#PCDATA|ex)*>\n<!ATTLIST ph name CDATA #REQUIRED>\n\n<!ELEMENT ex (#PCDATA)>'),new Pr,n.addDefaultExamples(i),new Pr])},e.prototype.load=function(t,e){throw new Error("Unsupported")},e.prototype.digest=function(t){return Br(t)},e.prototype.createNameMapper=function(t){return new Sr(t,zr)},e}(Er),Lr=function(){function t(){}return t.prototype.visitText=function(t,e){return[new Ir(t.value)]},t.prototype.visitContainer=function(t,e){var n=this,r=[];return t.children.forEach(function(t){return r.push.apply(r,Object(o.__spread)(t.visit(n)))}),r},t.prototype.visitIcu=function(t,e){var n=this,r=[new Ir("{"+t.expressionPlaceholder+", "+t.type+", ")];return Object.keys(t.cases).forEach(function(e){r.push.apply(r,Object(o.__spread)([new Ir(e+" {")],t.cases[e].visit(n),[new Ir("} ")]))}),r.push(new Ir("}")),r},t.prototype.visitTagPlaceholder=function(t,e){var n=new Ir("<"+t.tag+">"),r=new Dr("ex",{},[n]),i=new Dr("ph",{name:t.startName},[r,n]);if(t.isVoid)return[i];var a=new Ir("</"+t.tag+">"),s=new Dr("ex",{},[a]),u=new Dr("ph",{name:t.closeName},[s,a]);return Object(o.__spread)([i],this.serialize(t.children),[u])},t.prototype.visitPlaceholder=function(t,e){var n=new Ir("{{"+t.value+"}}"),r=new Dr("ex",{},[n]);return[new Dr("ph",{name:t.name},[r,n])]},t.prototype.visitIcuPlaceholder=function(t,e){var n=t.value.expression,r=t.value.type,i=Object.keys(t.value.cases).map(function(t){return t+" {...}"}).join(" "),o=new Ir("{"+n+", "+r+", "+i+"}"),a=new Dr("ex",{},[o]);return[new Dr("ph",{name:t.name},[a,o])]},t.prototype.serialize=function(t){var e=this;return[].concat.apply([],Object(o.__spread)(t.map(function(t){return t.visit(e)})))},t}();function Br(t){return sr(t)}var Vr=function(){function t(){}return t.prototype.addDefaultExamples=function(t){return t.visit(this),t},t.prototype.visitTag=function(t){var e=this;if("ph"===t.name){if(!t.children||0==t.children.length){var n=new Ir(t.attrs.name||"...");t.children=[new Dr("ex",{},[n])]}}else t.children&&t.children.forEach(function(t){return t.visit(e)})},t.prototype.visitText=function(t){},t.prototype.visitDeclaration=function(t){},t.prototype.visitDoctype=function(t){},t}();function zr(t){return t.toUpperCase().replace(/[^A-Z0-9_]/g,"_")}function Hr(t,e){return void 0===e&&(e=!1),Ae(Object.keys(t).map(function(n){return{key:n,quoted:e,value:t[n]}}))}var Ur="MSG_",Wr=new RegExp("^"+Ur),qr="I18N_",Gr="goog.getMsg",Yr="|",Kr="@@",Xr="i18n",Qr="i18n-",Zr="�";function $r(t,e,n){var r=[Pe(e)];n&&Object.keys(n).length&&r.push(Hr(n,!0));var i=Ee(Gr).callFn(r);return t.set(i).toDeclStmt(gt,[It.Final])}function Jr(t){return t===Xr||t.startsWith(Qr)}function ti(t){return t instanceof Zn}function ei(t){return ti(t)&&1===t.nodes.length&&t.nodes[0]instanceof tr}function ni(t,e){return void 0===e&&(e=null),{id:"string"==typeof e?e:t.id||"",meaning:t.meaning||"",description:t.description||""}}function ri(t){return t.nodes[0]}function ii(t,e){return void 0===e&&(e=0),""+Zr+t+(e>0?":"+e:"")+Zr}function oi(t){var e={};return t.forEach(function(t,n){e[n]=Pe(t.length>1?"["+t.join("|")+"]":t[0])}),e}function ai(t,e){for(var n=[],r=2;r<arguments.length;r++)n[r-2]=arguments[r];var i=t.get(e)||[];i.push.apply(i,Object(o.__spread)(n)),t.set(e,i)}function si(t,e,n){void 0===e&&(e=0),void 0===n&&(n=0);var r=e,i=new Map,o=t instanceof Zn?t.nodes.find(function(t){return t instanceof Jn}):t;return o&&o.children.filter(function(t){return t instanceof nr}).forEach(function(t,e){var o=ii(r+e,n);ai(i,t.name,o)}),i}function ui(t){var e,n=zr(t).split("_");if(1===n.length)return t.toLowerCase();/^\d+$/.test(n[n.length-1])&&(e=n.pop());var r=n.shift().toLowerCase();return n.length&&(r+=n.map(function(t){return t.charAt(0).toUpperCase()+t.slice(1).toLowerCase()}).join("")),e?r+"_"+e:r}function ci(t){return(""+Ur+t).toUpperCase()}function li(t,e,n,r,i){void 0===r&&(r={});var o,a,s=[],u=(a=[],(o=n).description&&a.push({tagName:"desc",text:o.description}),o.meaning&&a.push({tagName:"meaning",text:o.meaning}),0==a.length?null:new he(a));if(u&&s.push(u),i){s.push($r(t,e,r));var c=Ee(t.name);t.name=t.name.replace(Wr,qr),s.push(t.set(i(c)).toDeclStmt(gt,[It.Final]))}else s.push($r(t,e,r));return s}var di="_t",pi="ctx",fi="rf";function hi(t,e){var n=null;return function(){return n||(t.push(new re(di,void 0,mt)),n=Ee(e)),n}}function mi(t){if(this)throw new Error("Builder "+this.constructor.name+" doesn't support "+t+" yet");throw new Error("Feature "+t+" is not supported yet")}function gi(t){throw new Error("Invalid state: Visitor "+this.constructor.name+" doesn't handle "+void 0)}function yi(t){return Array.isArray(t)?ke(t.map(yi)):Pe(t,gt)}function _i(t,e){return Object.getOwnPropertyNames(t).length>0?(n=t,r=e,Ae(Object.getOwnPropertyNames(n).map(function(t){var e,i,a,s,u=n[t];return Array.isArray(u)?(e=Object(o.__read)(u,2),s=e[0],a=e[1]):(i=Object(o.__read)(Fe(t,[t,u]),2),a=i[0],s=i[1]),{key:a,quoted:!1,value:r&&s!==a?ke([yi(s),yi(a)]):yi(s)}}))):null;var n,r}function vi(t){for(;(e=t[t.length-1])instanceof Lt&&null===e.value;)t.pop();var e;return t}var bi,xi,wi,ji,Oi=function(){function t(){this.values=[]}return t.prototype.set=function(t,e){e&&this.values.push({key:t,value:e,quoted:!1})},t.prototype.toLiteralMap=function(){return Ae(this.values)},t}();function Ci(t){var e={};return t.attributes.forEach(function(t){Jr(t.name)||(e[t.name]=t.value)}),t.inputs.forEach(function(t){e[t.name]=""}),t.outputs.forEach(function(t){e[t.name]=""}),e}function Ei(t){var e=Ee("t"),n=[],r=Ti(t)?e:new Yt(ct.Or,e,t.type),i=null;if(null!==t.deps)i=new Ft(r,Si(t.deps,t.injectFn));else{var a=Ee("ɵ"+t.name+"_BaseFactory"),s=Se(Qn.getInheritedFactory),u=a.set(s.callFn([t.type])).toDeclStmt(gt,[It.Exported,It.Final]);n.push(u),i=a.callFn([r])}var c,l=i,d=[],p=null;function f(t){var n=Ee("r");return d.push(n.set(te).toDeclStmt()),d.push(Ie(e,[n.set(l).toStmt()],[n.set(t).toStmt()])),n}if(Ti(t)&&t.delegateType===bi.Factory){var h=Ee("ɵ"+t.name+"_BaseFactory"),m=Se(Qn.getFactoryOf);if(t.delegate.isEquivalent(t.type))throw new Error("Illegal state: compiling factory that delegates to itself");var g=h.set(m.callFn([t.delegate])).toDeclStmt(gt,[It.Exported,It.Final]);n.push(g),p=f(h.callFn([]))}else if(Ti(t)){var y=Si(t.delegateDeps,t.injectFn);p=f(new(t.delegateType===bi.Class?Ft:Nt)(t.delegate,y))}else c=t,p=void 0!==c.expression?f(t.expression):i;return{factory:De([new qt("t",mt)],Object(o.__spread)(d,[new ae(p)]),gt,void 0,t.name+"_Factory"),statements:n}}function Si(t,e){return t.map(function(t){return function(t,e){switch(t.resolved){case wi.Token:var n=0|(t.self?2:0)|(t.skipSelf?4:0)|(t.host?1:0)|(t.optional?8:0),r=[t.token];return(0!==n||t.optional)&&r.push(Pe(n)),Se(e).callFn(r);case wi.Attribute:return Se(Qn.injectAttribute).callFn([t.token]);default:return mi("Unknown R3ResolvedDependencyType: "+wi[t.resolved])}}(t,e)})}function Mi(t,e,n){n.resolveExternalReference(gn.Injector);var r,i,a=[];try{for(var s=Object(o.__values)(t.diDeps),u=s.next();!u.done;u=s.next()){var c=u.value;if(c.token){var l=In(c.token),d=c.isAttribute?wi.Attribute:wi.Token,p=l instanceof vn?e.importExpr(l):Pe(l);a.push({token:p,resolved:d,host:!!c.isHost,optional:!!c.isOptional,self:!!c.isSelf,skipSelf:!!c.isSkipSelf})}else mi("dependency without a token")}}catch(t){r={error:t}}finally{try{u&&!u.done&&(i=s.return)&&i.call(s)}finally{if(r)throw r.error}}return a}function Ti(t){return void 0!==t.delegateType}function ki(t){return Ae(Object.keys(t).map(function(e){return{key:e,value:t[e],quoted:!1}}))}function Ai(t,e){if(Array.isArray(t))return ke(t.map(function(t){return Ai(t,e)}));if(t instanceof vn)return e.importExpr(t);if(null==t)return Pe(t);throw new Error("Internal error: Unsupported or unknown metadata: "+t)}function Di(t,e){var n=null;if(e>0){n=[];for(var r=0;r<e;r++)n.push(mt)}return Te(t,null,n)}(xi=bi||(bi={}))[xi.Class=0]="Class",xi[xi.Function=1]="Function",xi[xi.Factory=2]="Factory",(ji=wi||(wi={}))[ji.Token=0]="Token",ji[ji.Attribute=1]="Attribute";var Ii="@";function Pi(t){return""+Ii+t}function Ri(t,e){return""+Ii+t+"."+e}function Ni(t,e){return"animation_"+t+"_"+e}function Fi(t){var e=null,n={name:t.name,type:t.type,deps:t.ctorDeps,injectFn:gn.inject};if(void 0!==t.useClass){var r=t.useClass.isEquivalent(t.type),i=t.userDeps||r&&t.ctorDeps||void 0;e=Ei(void 0!==i?Object(o.__assign)({},n,{delegate:t.useClass,delegateDeps:i,delegateType:bi.Class}):Object(o.__assign)({},n,{delegate:t.useClass,delegateType:bi.Factory}))}else e=void 0!==t.useFactory?Ei(Object(o.__assign)({},n,{delegate:t.useFactory,delegateDeps:t.userDeps||[],delegateType:bi.Function})):void 0!==t.useValue?Ei(Object(o.__assign)({},n,{expression:t.useValue})):void 0!==t.useExisting?Ei(Object(o.__assign)({},n,{expression:Se(gn.inject).callFn([t.useExisting])})):Ei(n);var a=t.type,s=t.providedIn;return{expression:Se(gn.defineInjectable).callFn([ki({token:a,factory:e.factory,providedIn:s})]),type:new pt(Se(gn.InjectableDef,[Di(t.type,t.typeArgumentCount)])),statements:e.statements}}function Li(t,e){if(null!=e){if(!Array.isArray(e))throw new Error("Expected '"+t+"' to be an array of strings.");for(var n=0;n<e.length;n+=1)if("string"!=typeof e[n])throw new Error("Expected '"+t+"' to be an array of strings.")}}var Bi=[/^\s*$/,/[<>]/,/^[{}]$/,/&(#|[a-z])/i,/^\/\//];function Vi(t,e){if(!(null==e||Array.isArray(e)&&2==e.length))throw new Error("Expected '"+t+"' to be an array, [start, end].");if(null!=e){var n=e[0],r=e[1];Bi.forEach(function(t){if(t.test(n)||t.test(r))throw new Error("['"+n+"', '"+r+"'] contains unusable interpolation symbol.")})}}var zi=function(){function t(t,e){this.start=t,this.end=e}return t.fromArray=function(e){return e?(Vi("interpolation",e),new t(e[0],e[1])):Hi},t}(),Hi=new zi("{{","}}"),Ui=function(){function t(t){void 0===t&&(t=null),this.file=t,this.sourcesContent=new Map,this.lines=[],this.lastCol0=0,this.hasMappings=!1}return t.prototype.addSource=function(t,e){return void 0===e&&(e=null),this.sourcesContent.has(t)||this.sourcesContent.set(t,e),this},t.prototype.addLine=function(){return this.lines.push([]),this.lastCol0=0,this},t.prototype.addMapping=function(t,e,n,r){if(!this.currentLine)throw new Error("A line must be added before mappings can be added");if(null!=e&&!this.sourcesContent.has(e))throw new Error('Unknown source file "'+e+'"');if(null==t)throw new Error("The column in the generated code must be provided");if(t<this.lastCol0)throw new Error("Mapping should be added in output order");if(e&&(null==n||null==r))throw new Error("The source location must be provided when a source url is provided");return this.hasMappings=!0,this.lastCol0=t,this.currentLine.push({col0:t,sourceUrl:e,sourceLine0:n,sourceCol0:r}),this},Object.defineProperty(t.prototype,"currentLine",{get:function(){return this.lines.slice(-1)[0]},enumerable:!0,configurable:!0}),t.prototype.toJSON=function(){var t=this;if(!this.hasMappings)return null;var e=new Map,n=[],r=[];Array.from(this.sourcesContent.keys()).forEach(function(i,o){e.set(i,o),n.push(i),r.push(t.sourcesContent.get(i)||null)});var i="",o=0,a=0,s=0,u=0;return this.lines.forEach(function(t){o=0,i+=t.map(function(t){var n=Wi(t.col0-o);return o=t.col0,null!=t.sourceUrl&&(n+=Wi(e.get(t.sourceUrl)-a),a=e.get(t.sourceUrl),n+=Wi(t.sourceLine0-s),s=t.sourceLine0,n+=Wi(t.sourceCol0-u),u=t.sourceCol0),n}).join(","),i+=";"}),i=i.slice(0,-1),{file:this.file||"",version:3,sourceRoot:"",sources:n,sourcesContent:r,mappings:i}},t.prototype.toJsComment=function(){return this.hasMappings?"//# sourceMappingURL=data:application/json;base64,"+function(t){var e="";t=tn(t);for(var n=0;n<t.length;){var r=t.charCodeAt(n++),i=t.charCodeAt(n++),o=t.charCodeAt(n++);e+=Gi(r>>2),e+=Gi((3&r)<<4|(isNaN(i)?0:i>>4)),e+=isNaN(i)?"=":Gi((15&i)<<2|o>>6),e+=isNaN(i)||isNaN(o)?"=":Gi(63&o)}return e}(JSON.stringify(this,null,0)):""},t}();function Wi(t){t=t<0?1+(-t<<1):t<<1;var e="";do{var n=31&t;(t>>=5)>0&&(n|=32),e+=Gi(n)}while(t>0);return e}var qi="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";function Gi(t){if(t<0||t>=64)throw new Error("Can only encode value in the range [0, 63]");return qi[t]}var Yi=/'|\\|\n|\r|\$/g,Ki=/^[$A-Z_][0-9A-Z_$]*$/i,Xi="  ",Qi=Ee("error",null,null),Zi=Ee("stack",null,null),$i=function(){return function(t){this.indent=t,this.partsLength=0,this.parts=[],this.srcSpans=[]}}(),Ji=function(){function t(t){this._indent=t,this._classes=[],this._preambleLineCount=0,this._lines=[new $i(t)]}return t.createRoot=function(){return new t(0)},Object.defineProperty(t.prototype,"_currentLine",{get:function(){return this._lines[this._lines.length-1]},enumerable:!0,configurable:!0}),t.prototype.println=function(t,e){void 0===e&&(e=""),this.print(t||null,e,!0)},t.prototype.lineIsEmpty=function(){return 0===this._currentLine.parts.length},t.prototype.lineLength=function(){return this._currentLine.indent*Xi.length+this._currentLine.partsLength},t.prototype.print=function(t,e,n){void 0===n&&(n=!1),e.length>0&&(this._currentLine.parts.push(e),this._currentLine.partsLength+=e.length,this._currentLine.srcSpans.push(t&&t.sourceSpan||null)),n&&this._lines.push(new $i(this._indent))},t.prototype.removeEmptyLastLine=function(){this.lineIsEmpty()&&this._lines.pop()},t.prototype.incIndent=function(){this._indent++,this.lineIsEmpty()&&(this._currentLine.indent=this._indent)},t.prototype.decIndent=function(){this._indent--,this.lineIsEmpty()&&(this._currentLine.indent=this._indent)},t.prototype.pushClass=function(t){this._classes.push(t)},t.prototype.popClass=function(){return this._classes.pop()},Object.defineProperty(t.prototype,"currentClass",{get:function(){return this._classes.length>0?this._classes[this._classes.length-1]:null},enumerable:!0,configurable:!0}),t.prototype.toSource=function(){return this.sourceLines.map(function(t){return t.parts.length>0?no(t.indent)+t.parts.join(""):""}).join("\n")},t.prototype.toSourceMapGenerator=function(t,e){void 0===e&&(e=0);for(var n=new Ui(t),r=!1,i=function(){r||(n.addSource(t," ").addMapping(0,t,0,0),r=!0)},o=0;o<e;o++)n.addLine(),i();return this.sourceLines.forEach(function(t,e){n.addLine();for(var o=t.srcSpans,a=t.parts,s=t.indent*Xi.length,u=0;u<o.length&&!o[u];)s+=a[u].length,u++;for(u<o.length&&0===e&&0===s?r=!0:i();u<o.length;){var c=o[u],l=c.start.file,d=c.start.line,p=c.start.col;for(n.addSource(l.url,l.content).addMapping(s,l.url,d,p),s+=a[u].length,u++;u<o.length&&(c===o[u]||!o[u]);)s+=a[u].length,u++}}),n},t.prototype.setPreambleLineCount=function(t){return this._preambleLineCount=t},t.prototype.spanOf=function(t,e){var n=this._lines[t-this._preambleLineCount];if(n)for(var r=e-no(n.indent).length,i=0;i<n.parts.length;i++){var o=n.parts[i];if(o.length>r)return n.srcSpans[i];r-=o.length}return null},Object.defineProperty(t.prototype,"sourceLines",{get:function(){return this._lines.length&&0===this._lines[this._lines.length-1].parts.length?this._lines.slice(0,-1):this._lines},enumerable:!0,configurable:!0}),t}(),to=function(){function t(t){this._escapeDollarInStrings=t}return t.prototype.visitExpressionStmt=function(t,e){return t.expr.visitExpression(this,e),e.println(t,";"),null},t.prototype.visitReturnStmt=function(t,e){return e.print(t,"return "),t.value.visitExpression(this,e),e.println(t,";"),null},t.prototype.visitIfStmt=function(t,e){e.print(t,"if ("),t.condition.visitExpression(this,e),e.print(t,") {");var n=null!=t.falseCase&&t.falseCase.length>0;return t.trueCase.length<=1&&!n?(e.print(t," "),this.visitAllStatements(t.trueCase,e),e.removeEmptyLastLine(),e.print(t," ")):(e.println(),e.incIndent(),this.visitAllStatements(t.trueCase,e),e.decIndent(),n&&(e.println(t,"} else {"),e.incIndent(),this.visitAllStatements(t.falseCase,e),e.decIndent())),e.println(t,"}"),null},t.prototype.visitThrowStmt=function(t,e){return e.print(t,"throw "),t.error.visitExpression(this,e),e.println(t,";"),null},t.prototype.visitCommentStmt=function(t,e){return t.multiline?e.println(t,"/* "+t.comment+" */"):t.comment.split("\n").forEach(function(n){e.println(t,"// "+n)}),null},t.prototype.visitJSDocCommentStmt=function(t,e){return e.println(t,"/*"+t.toString()+"*/"),null},t.prototype.visitWriteVarExpr=function(t,e){var n=e.lineIsEmpty();return n||e.print(t,"("),e.print(t,t.name+" = "),t.value.visitExpression(this,e),n||e.print(t,")"),null},t.prototype.visitWriteKeyExpr=function(t,e){var n=e.lineIsEmpty();return n||e.print(t,"("),t.receiver.visitExpression(this,e),e.print(t,"["),t.index.visitExpression(this,e),e.print(t,"] = "),t.value.visitExpression(this,e),n||e.print(t,")"),null},t.prototype.visitWritePropExpr=function(t,e){var n=e.lineIsEmpty();return n||e.print(t,"("),t.receiver.visitExpression(this,e),e.print(t,"."+t.name+" = "),t.value.visitExpression(this,e),n||e.print(t,")"),null},t.prototype.visitInvokeMethodExpr=function(t,e){t.receiver.visitExpression(this,e);var n=t.name;return null!=t.builtin&&null==(n=this.getBuiltinMethodName(t.builtin))?null:(e.print(t,"."+n+"("),this.visitAllExpressions(t.args,e,","),e.print(t,")"),null)},t.prototype.visitInvokeFunctionExpr=function(t,e){return t.fn.visitExpression(this,e),e.print(t,"("),this.visitAllExpressions(t.args,e,","),e.print(t,")"),null},t.prototype.visitWrappedNodeExpr=function(t,e){throw new Error("Abstract emitter cannot visit WrappedNodeExpr.")},t.prototype.visitTypeofExpr=function(t,e){e.print(t,"typeof "),t.expr.visitExpression(this,e)},t.prototype.visitReadVarExpr=function(t,e){var n=t.name;if(null!=t.builtin)switch(t.builtin){case wt.Super:n="super";break;case wt.This:n="this";break;case wt.CatchError:n=Qi.name;break;case wt.CatchStack:n=Zi.name;break;default:throw new Error("Unknown builtin variable "+t.builtin)}return e.print(t,n),null},t.prototype.visitInstantiateExpr=function(t,e){return e.print(t,"new "),t.classExpr.visitExpression(this,e),e.print(t,"("),this.visitAllExpressions(t.args,e,","),e.print(t,")"),null},t.prototype.visitLiteralExpr=function(t,e){var n=t.value;return"string"==typeof n?e.print(t,eo(n,this._escapeDollarInStrings)):e.print(t,""+n),null},t.prototype.visitConditionalExpr=function(t,e){return e.print(t,"("),t.condition.visitExpression(this,e),e.print(t,"? "),t.trueCase.visitExpression(this,e),e.print(t,": "),t.falseCase.visitExpression(this,e),e.print(t,")"),null},t.prototype.visitNotExpr=function(t,e){return e.print(t,"!"),t.condition.visitExpression(this,e),null},t.prototype.visitAssertNotNullExpr=function(t,e){return t.condition.visitExpression(this,e),null},t.prototype.visitBinaryOperatorExpr=function(t,e){var n;switch(t.operator){case ct.Equals:n="==";break;case ct.Identical:n="===";break;case ct.NotEquals:n="!=";break;case ct.NotIdentical:n="!==";break;case ct.And:n="&&";break;case ct.BitwiseAnd:n="&";break;case ct.Or:n="||";break;case ct.Plus:n="+";break;case ct.Minus:n="-";break;case ct.Divide:n="/";break;case ct.Multiply:n="*";break;case ct.Modulo:n="%";break;case ct.Lower:n="<";break;case ct.LowerEquals:n="<=";break;case ct.Bigger:n=">";break;case ct.BiggerEquals:n=">=";break;default:throw new Error("Unknown operator "+t.operator)}return t.parens&&e.print(t,"("),t.lhs.visitExpression(this,e),e.print(t," "+n+" "),t.rhs.visitExpression(this,e),t.parens&&e.print(t,")"),null},t.prototype.visitReadPropExpr=function(t,e){return t.receiver.visitExpression(this,e),e.print(t,"."),e.print(t,t.name),null},t.prototype.visitReadKeyExpr=function(t,e){return t.receiver.visitExpression(this,e),e.print(t,"["),t.index.visitExpression(this,e),e.print(t,"]"),null},t.prototype.visitLiteralArrayExpr=function(t,e){return e.print(t,"["),this.visitAllExpressions(t.entries,e,","),e.print(t,"]"),null},t.prototype.visitLiteralMapExpr=function(t,e){var n=this;return e.print(t,"{"),this.visitAllObjects(function(r){e.print(t,eo(r.key,n._escapeDollarInStrings,r.quoted)+":"),r.value.visitExpression(n,e)},t.entries,e,","),e.print(t,"}"),null},t.prototype.visitCommaExpr=function(t,e){return e.print(t,"("),this.visitAllExpressions(t.parts,e,","),e.print(t,")"),null},t.prototype.visitAllExpressions=function(t,e,n){var r=this;this.visitAllObjects(function(t){return t.visitExpression(r,e)},t,e,n)},t.prototype.visitAllObjects=function(t,e,n,r){for(var i=!1,o=0;o<e.length;o++)o>0&&(n.lineLength()>80?(n.print(null,r,!0),i||(n.incIndent(),n.incIndent(),i=!0)):n.print(null,r,!1)),t(e[o]);i&&(n.decIndent(),n.decIndent())},t.prototype.visitAllStatements=function(t,e){var n=this;t.forEach(function(t){return t.visitStatement(n,e)})},t}();function eo(t,e,n){if(void 0===n&&(n=!0),null==t)return null;var r=t.replace(Yi,function(){for(var t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];return"$"==t[0]?e?"\\$":"$":"\n"==t[0]?"\\n":"\r"==t[0]?"\\r":"\\"+t[0]});return n||!Ki.test(r)?"'"+r+"'":r}function no(t){for(var e="",n=0;n<t;n++)e+=Xi;return e}function ro(t,e,n,r){var i=new io(n),a=Ji.createRoot();return i.visitAllStatements(e,a),i.createReturnStmt(a),function(t,e,n,r){var i=e.toSource()+"\n//# sourceURL="+t,a=[],s=[];for(var u in n)a.push(u),s.push(n[u]);if(r){var c=(new(Function.bind.apply(Function,Object(o.__spread)([void 0],a.concat("return null;"))))).toString(),l=c.slice(0,c.indexOf("return null;")).split("\n").length-1;i+="\n"+e.toSourceMapGenerator(t,l).toJsComment()}return(new(Function.bind.apply(Function,Object(o.__spread)([void 0],a.concat(i))))).apply(void 0,Object(o.__spread)(s))}(t,a,i.getArgs(),r)}var io=function(t){function e(e){var n=t.call(this)||this;return n.reflector=e,n._evalArgNames=[],n._evalArgValues=[],n._evalExportedVars=[],n}return Object(o.__extends)(e,t),e.prototype.createReturnStmt=function(t){new ae(new $t(this._evalExportedVars.map(function(t){return new Zt(t,Ee(t),!1)}))).visitStatement(this,t)},e.prototype.getArgs=function(){for(var t={},e=0;e<this._evalArgNames.length;e++)t[this._evalArgNames[e]]=this._evalArgValues[e];return t},e.prototype.visitExternalExpr=function(t,e){return this._emitReferenceToExternal(t,this.reflector.resolveExternalReference(t.value),e),null},e.prototype.visitWrappedNodeExpr=function(t,e){return this._emitReferenceToExternal(t,t.node,e),null},e.prototype.visitDeclareVarStmt=function(e,n){return e.hasModifier(It.Exported)&&this._evalExportedVars.push(e.name),t.prototype.visitDeclareVarStmt.call(this,e,n)},e.prototype.visitDeclareFunctionStmt=function(e,n){return e.hasModifier(It.Exported)&&this._evalExportedVars.push(e.name),t.prototype.visitDeclareFunctionStmt.call(this,e,n)},e.prototype.visitDeclareClassStmt=function(e,n){return e.hasModifier(It.Exported)&&this._evalExportedVars.push(e.name),t.prototype.visitDeclareClassStmt.call(this,e,n)},e.prototype._emitReferenceToExternal=function(t,e,n){var r=this._evalArgValues.indexOf(e);if(-1===r){r=this._evalArgValues.length,this._evalArgValues.push(e);var i=En({reference:e})||"val";this._evalArgNames.push("jit_"+i+"_"+r)}n.print(t,this._evalArgNames[r])},e}(function(t){function e(){return t.call(this,!1)||this}return Object(o.__extends)(e,t),e.prototype.visitDeclareClassStmt=function(t,e){var n=this;return e.pushClass(t),this._visitClassConstructor(t,e),null!=t.parent&&(e.print(t,t.name+".prototype = Object.create("),t.parent.visitExpression(this,e),e.println(t,".prototype);")),t.getters.forEach(function(r){return n._visitClassGetter(t,r,e)}),t.methods.forEach(function(r){return n._visitClassMethod(t,r,e)}),e.popClass(),null},e.prototype._visitClassConstructor=function(t,e){e.print(t,"function "+t.name+"("),null!=t.constructorMethod&&this._visitParams(t.constructorMethod.params,e),e.println(t,") {"),e.incIndent(),null!=t.constructorMethod&&t.constructorMethod.body.length>0&&(e.println(t,"var self = this;"),this.visitAllStatements(t.constructorMethod.body,e)),e.decIndent(),e.println(t,"}")},e.prototype._visitClassGetter=function(t,e,n){n.println(t,"Object.defineProperty("+t.name+".prototype, '"+e.name+"', { get: function() {"),n.incIndent(),e.body.length>0&&(n.println(t,"var self = this;"),this.visitAllStatements(e.body,n)),n.decIndent(),n.println(t,"}});")},e.prototype._visitClassMethod=function(t,e,n){n.print(t,t.name+".prototype."+e.name+" = function("),this._visitParams(e.params,n),n.println(t,") {"),n.incIndent(),e.body.length>0&&(n.println(t,"var self = this;"),this.visitAllStatements(e.body,n)),n.decIndent(),n.println(t,"};")},e.prototype.visitWrappedNodeExpr=function(t,e){throw new Error("Cannot emit a WrappedNodeExpr in Javascript.")},e.prototype.visitReadVarExpr=function(e,n){if(e.builtin===wt.This)n.print(e,"self");else{if(e.builtin===wt.Super)throw new Error("'super' needs to be handled at a parent ast node, not at the variable level!");t.prototype.visitReadVarExpr.call(this,e,n)}return null},e.prototype.visitDeclareVarStmt=function(t,e){return e.print(t,"var "+t.name),t.value&&(e.print(t," = "),t.value.visitExpression(this,e)),e.println(t,";"),null},e.prototype.visitCastExpr=function(t,e){return t.value.visitExpression(this,e),null},e.prototype.visitInvokeFunctionExpr=function(e,n){var r=e.fn;return r instanceof St&&r.builtin===wt.Super?(n.currentClass.parent.visitExpression(this,n),n.print(e,".call(this"),e.args.length>0&&(n.print(e,", "),this.visitAllExpressions(e.args,n,",")),n.print(e,")")):t.prototype.visitInvokeFunctionExpr.call(this,e,n),null},e.prototype.visitFunctionExpr=function(t,e){return e.print(t,"function"+(t.name?" "+t.name:"")+"("),this._visitParams(t.params,e),e.println(t,") {"),e.incIndent(),this.visitAllStatements(t.statements,e),e.decIndent(),e.print(t,"}"),null},e.prototype.visitDeclareFunctionStmt=function(t,e){return e.print(t,"function "+t.name+"("),this._visitParams(t.params,e),e.println(t,") {"),e.incIndent(),this.visitAllStatements(t.statements,e),e.decIndent(),e.println(t,"}"),null},e.prototype.visitTryCatchStmt=function(t,e){e.println(t,"try {"),e.incIndent(),this.visitAllStatements(t.bodyStmts,e),e.decIndent(),e.println(t,"} catch ("+Qi.name+") {"),e.incIndent();var n=[Zi.set(Qi.prop("stack")).toDeclStmt(null,[It.Final])].concat(t.catchStmts);return this.visitAllStatements(n,e),e.decIndent(),e.println(t,"}"),null},e.prototype._visitParams=function(t,e){this.visitAllObjects(function(t){return e.print(null,t.name)},t,e,",")},e.prototype.getBuiltinMethodName=function(t){var e;switch(t){case Ct.ConcatArray:e="concat";break;case Ct.SubscribeObservable:e="subscribe";break;case Ct.Bind:e="bind";break;default:throw new Error("Unknown builtin method: "+t)}return e},e}(to)),oo=function(){function t(t){this.context=t}return t.prototype.resolveExternalReference=function(t){if("@angular/core"!==t.moduleName)throw new Error("Cannot resolve external reference to "+t.moduleName+", only references to @angular/core are supported.");if(!this.context.hasOwnProperty(t.name))throw new Error("No value provided for @angular/core symbol '"+t.name+"'.");return this.context[t.name]},t.prototype.parameters=function(t){throw new Error("Not implemented.")},t.prototype.annotations=function(t){throw new Error("Not implemented.")},t.prototype.shallowAnnotations=function(t){throw new Error("Not implemented.")},t.prototype.tryAnnotations=function(t){throw new Error("Not implemented.")},t.prototype.propMetadata=function(t){throw new Error("Not implemented.")},t.prototype.hasLifecycleHook=function(t,e){throw new Error("Not implemented.")},t.prototype.guards=function(t){throw new Error("Not implemented.")},t.prototype.componentModuleUrl=function(t,e){throw new Error("Not implemented.")},t}();function ao(t,e,n,r){return ro(n,Object(o.__spread)(r,[new re("$def",t,void 0,[It.Exported])]),new oo(e),!1).$def}function so(t){var e=t.type,n=t.bootstrap,r=t.declarations,i=t.imports,o=t.exports;return{expression:Se(Qn.defineNgModule).callFn([ki({type:e,bootstrap:ke(n.map(function(t){return t.value})),declarations:ke(r.map(function(t){return t.value})),imports:ke(i.map(function(t){return t.value})),exports:ke(o.map(function(t){return t.value}))})]),type:new pt(Se(Qn.NgModuleDefWithMeta,[new pt(e),co(r),co(i),co(o)])),additionalStatements:[]}}function uo(t){var e=Ei({name:t.name,type:t.type,deps:t.deps,injectFn:Qn.inject});return{expression:Se(Qn.defineInjector).callFn([ki({factory:e.factory,providers:t.providers,imports:t.imports})]),type:new pt(Se(Qn.InjectorDef,[new pt(t.type)])),statements:e.statements}}function co(t){var e=t.map(function(t){return e=t.type,new Mt(e);var e});return t.length>0?Te(ke(e)):vt}function lo(t){var e=[];e.push({key:"name",value:Pe(t.pipeName),quoted:!1}),e.push({key:"type",value:t.type,quoted:!1});var n=Ei({name:t.name,type:t.type,deps:t.deps,injectFn:Qn.directiveInject});return e.push({key:"factory",value:n.factory,quoted:!1}),e.push({key:"pure",value:Pe(t.pure),quoted:!1}),{expression:Se(Qn.definePipe).callFn([Ae(e)]),type:new pt(Se(Qn.PipeDefWithMeta,[new pt(t.type),new pt(new Lt(t.pipeName))])),statements:n.statements}}var po=function(){return function(t,e,n,r){this.input=e,this.errLocation=n,this.ctxLocation=r,this.message="Parser Error: "+t+" "+n+" ["+e+"] in "+r}}(),fo=function(){return function(t,e){this.start=t,this.end=e}}(),ho=function(){function t(t){this.span=t}return t.prototype.visit=function(t,e){return void 0===e&&(e=null),null},t.prototype.toString=function(){return"AST"},t}(),mo=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.prefix=n,o.uninterpretedExpression=r,o.location=i,o}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitQuote(this,e)},e.prototype.toString=function(){return"Quote"},e}(ho),go=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){void 0===e&&(e=null)},e}(ho),yo=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitImplicitReceiver(this,e)},e}(ho),_o=function(t){function e(e,n){var r=t.call(this,e)||this;return r.expressions=n,r}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitChain(this,e)},e}(ho),vo=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.condition=n,o.trueExp=r,o.falseExp=i,o}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitConditional(this,e)},e}(ho),bo=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.receiver=n,i.name=r,i}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitPropertyRead(this,e)},e}(ho),xo=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.receiver=n,o.name=r,o.value=i,o}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitPropertyWrite(this,e)},e}(ho),wo=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.receiver=n,i.name=r,i}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitSafePropertyRead(this,e)},e}(ho),jo=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.obj=n,i.key=r,i}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitKeyedRead(this,e)},e}(ho),Oo=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.obj=n,o.key=r,o.value=i,o}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitKeyedWrite(this,e)},e}(ho),Co=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.exp=n,o.name=r,o.args=i,o}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitPipe(this,e)},e}(ho),Eo=function(t){function e(e,n){var r=t.call(this,e)||this;return r.value=n,r}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitLiteralPrimitive(this,e)},e}(ho),So=function(t){function e(e,n){var r=t.call(this,e)||this;return r.expressions=n,r}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitLiteralArray(this,e)},e}(ho),Mo=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.keys=n,i.values=r,i}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitLiteralMap(this,e)},e}(ho),To=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.strings=n,i.expressions=r,i}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitInterpolation(this,e)},e}(ho),ko=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.operation=n,o.left=r,o.right=i,o}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitBinary(this,e)},e}(ho),Ao=function(t){function e(e,n){var r=t.call(this,e)||this;return r.expression=n,r}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitPrefixNot(this,e)},e}(ho),Do=function(t){function e(e,n){var r=t.call(this,e)||this;return r.expression=n,r}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitNonNullAssert(this,e)},e}(ho),Io=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.receiver=n,o.name=r,o.args=i,o}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitMethodCall(this,e)},e}(ho),Po=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.receiver=n,o.name=r,o.args=i,o}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitSafeMethodCall(this,e)},e}(ho),Ro=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.target=n,i.args=r,i}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitFunctionCall(this,e)},e}(ho),No=function(t){function e(e,n,r,i){var o=t.call(this,new fo(0,null==n?0:n.length))||this;return o.ast=e,o.source=n,o.location=r,o.errors=i,o}return Object(o.__extends)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),this.ast.visit(t,e)},e.prototype.toString=function(){return this.source+" in "+this.location},e}(ho),Fo=function(){return function(t,e,n,r,i){this.span=t,this.key=e,this.keyIsVar=n,this.name=r,this.expression=i}}(),Lo=function(){function t(){}return t.prototype.visitBinary=function(t,e){},t.prototype.visitChain=function(t,e){},t.prototype.visitConditional=function(t,e){},t.prototype.visitFunctionCall=function(t,e){},t.prototype.visitImplicitReceiver=function(t,e){},t.prototype.visitInterpolation=function(t,e){},t.prototype.visitKeyedRead=function(t,e){},t.prototype.visitKeyedWrite=function(t,e){},t.prototype.visitLiteralArray=function(t,e){},t.prototype.visitLiteralMap=function(t,e){},t.prototype.visitLiteralPrimitive=function(t,e){},t.prototype.visitMethodCall=function(t,e){},t.prototype.visitPipe=function(t,e){},t.prototype.visitPrefixNot=function(t,e){},t.prototype.visitNonNullAssert=function(t,e){},t.prototype.visitPropertyRead=function(t,e){},t.prototype.visitPropertyWrite=function(t,e){},t.prototype.visitQuote=function(t,e){},t.prototype.visitSafeMethodCall=function(t,e){},t.prototype.visitSafePropertyRead=function(t,e){},t}(),Bo=function(){function t(){}return t.prototype.visitBinary=function(t,e){return t.left.visit(this),t.right.visit(this),null},t.prototype.visitChain=function(t,e){return this.visitAll(t.expressions,e)},t.prototype.visitConditional=function(t,e){return t.condition.visit(this),t.trueExp.visit(this),t.falseExp.visit(this),null},t.prototype.visitPipe=function(t,e){return t.exp.visit(this),this.visitAll(t.args,e),null},t.prototype.visitFunctionCall=function(t,e){return t.target.visit(this),this.visitAll(t.args,e),null},t.prototype.visitImplicitReceiver=function(t,e){return null},t.prototype.visitInterpolation=function(t,e){return this.visitAll(t.expressions,e)},t.prototype.visitKeyedRead=function(t,e){return t.obj.visit(this),t.key.visit(this),null},t.prototype.visitKeyedWrite=function(t,e){return t.obj.visit(this),t.key.visit(this),t.value.visit(this),null},t.prototype.visitLiteralArray=function(t,e){return this.visitAll(t.expressions,e)},t.prototype.visitLiteralMap=function(t,e){return this.visitAll(t.values,e)},t.prototype.visitLiteralPrimitive=function(t,e){return null},t.prototype.visitMethodCall=function(t,e){return t.receiver.visit(this),this.visitAll(t.args,e)},t.prototype.visitPrefixNot=function(t,e){return t.expression.visit(this),null},t.prototype.visitNonNullAssert=function(t,e){return t.expression.visit(this),null},t.prototype.visitPropertyRead=function(t,e){return t.receiver.visit(this),null},t.prototype.visitPropertyWrite=function(t,e){return t.receiver.visit(this),t.value.visit(this),null},t.prototype.visitSafePropertyRead=function(t,e){return t.receiver.visit(this),null},t.prototype.visitSafeMethodCall=function(t,e){return t.receiver.visit(this),this.visitAll(t.args,e)},t.prototype.visitAll=function(t,e){var n=this;return t.forEach(function(t){return t.visit(n,e)}),null},t.prototype.visitQuote=function(t,e){return null},t}(),Vo=function(){function t(){}return t.prototype.visitImplicitReceiver=function(t,e){return t},t.prototype.visitInterpolation=function(t,e){return new To(t.span,t.strings,this.visitAll(t.expressions))},t.prototype.visitLiteralPrimitive=function(t,e){return new Eo(t.span,t.value)},t.prototype.visitPropertyRead=function(t,e){return new bo(t.span,t.receiver.visit(this),t.name)},t.prototype.visitPropertyWrite=function(t,e){return new xo(t.span,t.receiver.visit(this),t.name,t.value.visit(this))},t.prototype.visitSafePropertyRead=function(t,e){return new wo(t.span,t.receiver.visit(this),t.name)},t.prototype.visitMethodCall=function(t,e){return new Io(t.span,t.receiver.visit(this),t.name,this.visitAll(t.args))},t.prototype.visitSafeMethodCall=function(t,e){return new Po(t.span,t.receiver.visit(this),t.name,this.visitAll(t.args))},t.prototype.visitFunctionCall=function(t,e){return new Ro(t.span,t.target.visit(this),this.visitAll(t.args))},t.prototype.visitLiteralArray=function(t,e){return new So(t.span,this.visitAll(t.expressions))},t.prototype.visitLiteralMap=function(t,e){return new Mo(t.span,t.keys,this.visitAll(t.values))},t.prototype.visitBinary=function(t,e){return new ko(t.span,t.operation,t.left.visit(this),t.right.visit(this))},t.prototype.visitPrefixNot=function(t,e){return new Ao(t.span,t.expression.visit(this))},t.prototype.visitNonNullAssert=function(t,e){return new Do(t.span,t.expression.visit(this))},t.prototype.visitConditional=function(t,e){return new vo(t.span,t.condition.visit(this),t.trueExp.visit(this),t.falseExp.visit(this))},t.prototype.visitPipe=function(t,e){return new Co(t.span,t.exp.visit(this),t.name,this.visitAll(t.args))},t.prototype.visitKeyedRead=function(t,e){return new jo(t.span,t.obj.visit(this),t.key.visit(this))},t.prototype.visitKeyedWrite=function(t,e){return new Oo(t.span,t.obj.visit(this),t.key.visit(this),t.value.visit(this))},t.prototype.visitAll=function(t){for(var e=new Array(t.length),n=0;n<t.length;++n)e[n]=t[n].visit(this);return e},t.prototype.visitChain=function(t,e){return new _o(t.span,this.visitAll(t.expressions))},t.prototype.visitQuote=function(t,e){return new mo(t.span,t.prefix,t.uninterpretedExpression,t.location)},t}(),zo=function(){function t(){}return t.prototype.visitImplicitReceiver=function(t,e){return t},t.prototype.visitInterpolation=function(t,e){var n=this.visitAll(t.expressions);return n!==t.expressions?new To(t.span,t.strings,n):t},t.prototype.visitLiteralPrimitive=function(t,e){return t},t.prototype.visitPropertyRead=function(t,e){var n=t.receiver.visit(this);return n!==t.receiver?new bo(t.span,n,t.name):t},t.prototype.visitPropertyWrite=function(t,e){var n=t.receiver.visit(this),r=t.value.visit(this);return n!==t.receiver||r!==t.value?new xo(t.span,n,t.name,r):t},t.prototype.visitSafePropertyRead=function(t,e){var n=t.receiver.visit(this);return n!==t.receiver?new wo(t.span,n,t.name):t},t.prototype.visitMethodCall=function(t,e){var n=t.receiver.visit(this);return n!==t.receiver?new Io(t.span,n,t.name,this.visitAll(t.args)):t},t.prototype.visitSafeMethodCall=function(t,e){var n=t.receiver.visit(this),r=this.visitAll(t.args);return n!==t.receiver||r!==t.args?new Po(t.span,n,t.name,r):t},t.prototype.visitFunctionCall=function(t,e){var n=t.target&&t.target.visit(this),r=this.visitAll(t.args);return n!==t.target||r!==t.args?new Ro(t.span,n,r):t},t.prototype.visitLiteralArray=function(t,e){var n=this.visitAll(t.expressions);return n!==t.expressions?new So(t.span,n):t},t.prototype.visitLiteralMap=function(t,e){var n=this.visitAll(t.values);return n!==t.values?new Mo(t.span,t.keys,n):t},t.prototype.visitBinary=function(t,e){var n=t.left.visit(this),r=t.right.visit(this);return n!==t.left||r!==t.right?new ko(t.span,t.operation,n,r):t},t.prototype.visitPrefixNot=function(t,e){var n=t.expression.visit(this);return n!==t.expression?new Ao(t.span,n):t},t.prototype.visitNonNullAssert=function(t,e){var n=t.expression.visit(this);return n!==t.expression?new Do(t.span,n):t},t.prototype.visitConditional=function(t,e){var n=t.condition.visit(this),r=t.trueExp.visit(this),i=t.falseExp.visit(this);return n!==t.condition||r!==t.trueExp||i!=i?new vo(t.span,n,r,i):t},t.prototype.visitPipe=function(t,e){var n=t.exp.visit(this),r=this.visitAll(t.args);return n!==t.exp||r!==t.args?new Co(t.span,n,t.name,r):t},t.prototype.visitKeyedRead=function(t,e){var n=t.obj.visit(this),r=t.key.visit(this);return n!==t.obj||r!==t.key?new jo(t.span,n,r):t},t.prototype.visitKeyedWrite=function(t,e){var n=t.obj.visit(this),r=t.key.visit(this),i=t.value.visit(this);return n!==t.obj||r!==t.key||i!==t.value?new Oo(t.span,n,r,i):t},t.prototype.visitAll=function(t){for(var e=new Array(t.length),n=!1,r=0;r<t.length;++r){var i=t[r],o=i.visit(this);e[r]=o,n=n||o!==i}return n?e:t},t.prototype.visitChain=function(t,e){var n=this.visitAll(t.expressions);return n!==t.expressions?new _o(t.span,n):t},t.prototype.visitQuote=function(t,e){return t},t}();function Ho(t,e,n){function r(t){e.visit&&e.visit(t,n)||t.visit(e,n)}function i(t){t.forEach(r)}t.visit({visitBinary:function(t){r(t.left),r(t.right)},visitChain:function(t){i(t.expressions)},visitConditional:function(t){r(t.condition),r(t.trueExp),r(t.falseExp)},visitFunctionCall:function(t){t.target&&r(t.target),i(t.args)},visitImplicitReceiver:function(t){},visitInterpolation:function(t){i(t.expressions)},visitKeyedRead:function(t){r(t.obj),r(t.key)},visitKeyedWrite:function(t){r(t.obj),r(t.key),r(t.obj)},visitLiteralArray:function(t){i(t.expressions)},visitLiteralMap:function(t){},visitLiteralPrimitive:function(t){},visitMethodCall:function(t){r(t.receiver),i(t.args)},visitPipe:function(t){r(t.exp),i(t.args)},visitPrefixNot:function(t){r(t.expression)},visitNonNullAssert:function(t){r(t.expression)},visitPropertyRead:function(t){r(t.receiver)},visitPropertyWrite:function(t){r(t.receiver),r(t.value)},visitQuote:function(t){},visitSafeMethodCall:function(t){r(t.receiver),i(t.args)},visitSafePropertyRead:function(t){r(t.receiver)}})}var Uo,Wo,qo=function(){return function(t,e,n,r){this.name=t,this.expression=e,this.type=n,this.sourceSpan=r,this.isLiteral=this.type===Uo.LITERAL_ATTR,this.isAnimation=this.type===Uo.ANIMATION}}();(Wo=Uo||(Uo={}))[Wo.DEFAULT=0]="DEFAULT",Wo[Wo.LITERAL_ATTR=1]="LITERAL_ATTR",Wo[Wo.ANIMATION=2]="ANIMATION";var Go=function(){return function(t,e,n,r,i){this.name=t,this.targetOrPhase=e,this.type=n,this.handler=r,this.sourceSpan=i}}(),Yo=function(){return function(t,e,n){this.name=t,this.value=e,this.sourceSpan=n}}(),Ko=function(){return function(t,e,n,r,i,o){this.name=t,this.type=e,this.securityContext=n,this.value=r,this.unit=i,this.sourceSpan=o}}(),Xo=function(){function t(){}return t.event=Ee("$event"),t}(),Qo=function(){return function(t,e){this.stmts=t,this.allowDefault=e,this.render3Stmts=t.map(function(t){if(t instanceof re&&t.name==e.name&&t.value instanceof Yt){var n=t.value.lhs;return new ae(n.value)}return t})}}();function Zo(t,e,n,r,i){t||(t=new da);var o=$o({createLiteralArrayConverter:function(t){return function(t){return ke(t)}},createLiteralMapConverter:function(t){return function(e){return Ae(t.map(function(t,n){return{key:t.key,value:e[n],quoted:t.quoted}}))}},createPipeConverter:function(t){throw new Error("Illegal State: Actions are not allowed to contain pipes. Pipe: "+t)}},n),a=new la(t,e,r,i),s=[];!function t(e,n){Array.isArray(e)?e.forEach(function(e){return t(e,n)}):n.push(e)}(o.visit(a,ea.Statement),s),function(t,e,n){for(var r=t-1;r>=0;r--)n.unshift(aa(e,r))}(a.temporaryCount,r,s);var u=s.length-1,c=null;if(u>=0){var l=function(t){if(t instanceof oe)return t.expr;if(t instanceof ae)return t.value;return null}(s[u]);l&&(c=Ee("pd_"+r),s[u]=c.set(l.cast(mt).notIdentical(Pe(!1))).toDeclStmt(null,[It.Final]))}return new Qo(s,c)}function $o(t,e){return n=e,r=new ca(t),n.visit(r);var n,r}var Jo,ta,ea,na,ra=function(){return function(t,e){this.stmts=t,this.currValExpr=e}}();function ia(t,e,n,r,i,o){t||(t=new da);var a=Ee("currVal_"+r),s=[],u=new la(t,e,r,o),c=n.visit(u,ea.Expression);if(u.temporaryCount)for(var l=0;l<u.temporaryCount;l++)s.push(aa(r,l));else if(i==Jo.TrySimple)return new ra([],c);return s.push(a.set(c).toDeclStmt(mt,[It.Final])),new ra(s,a)}function oa(t,e){return"tmp_"+t+"_"+e}function aa(t,e){return new re(oa(t,e),te)}function sa(t,e){if(t!==ea.Expression)throw new Error("Expected an expression, but saw "+e)}function ua(t,e){return t===ea.Statement?e.toStmt():e}(ta=Jo||(Jo={}))[ta.General=0]="General",ta[ta.TrySimple=1]="TrySimple",(na=ea||(ea={}))[na.Statement=0]="Statement",na[na.Expression=1]="Expression";var ca=function(t){function e(e){var n=t.call(this)||this;return n._converterFactory=e,n}return Object(o.__extends)(e,t),e.prototype.visitPipe=function(t,e){var n=this,r=Object(o.__spread)([t.exp],t.args).map(function(t){return t.visit(n,e)});return new ha(t.span,r,this._converterFactory.createPipeConverter(t.name,r.length))},e.prototype.visitLiteralArray=function(t,e){var n=this,r=t.expressions.map(function(t){return t.visit(n,e)});return new ha(t.span,r,this._converterFactory.createLiteralArrayConverter(t.expressions.length))},e.prototype.visitLiteralMap=function(t,e){var n=this,r=t.values.map(function(t){return t.visit(n,e)});return new ha(t.span,r,this._converterFactory.createLiteralMapConverter(t.keys))},e}(Vo),la=function(){function t(t,e,n,r){this._localResolver=t,this._implicitReceiver=e,this.bindingId=n,this.interpolationFunction=r,this._nodeMap=new Map,this._resultMap=new Map,this._currentTemporary=0,this.temporaryCount=0}return t.prototype.visitBinary=function(t,e){var n;switch(t.operation){case"+":n=ct.Plus;break;case"-":n=ct.Minus;break;case"*":n=ct.Multiply;break;case"/":n=ct.Divide;break;case"%":n=ct.Modulo;break;case"&&":n=ct.And;break;case"||":n=ct.Or;break;case"==":n=ct.Equals;break;case"!=":n=ct.NotEquals;break;case"===":n=ct.Identical;break;case"!==":n=ct.NotIdentical;break;case"<":n=ct.Lower;break;case">":n=ct.Bigger;break;case"<=":n=ct.LowerEquals;break;case">=":n=ct.BiggerEquals;break;default:throw new Error("Unsupported operation "+t.operation)}return ua(e,new Yt(n,this._visit(t.left,ea.Expression),this._visit(t.right,ea.Expression)))},t.prototype.visitChain=function(t,e){return function(t,e){if(t!==ea.Statement)throw new Error("Expected a statement, but saw "+e)}(e,t),this.visitAll(t.expressions,e)},t.prototype.visitConditional=function(t,e){return ua(e,this._visit(t.condition,ea.Expression).conditional(this._visit(t.trueExp,ea.Expression),this._visit(t.falseExp,ea.Expression)))},t.prototype.visitPipe=function(t,e){throw new Error("Illegal state: Pipes should have been converted into functions. Pipe: "+t.name)},t.prototype.visitFunctionCall=function(t,e){var n=this.visitAll(t.args,ea.Expression);return ua(e,t instanceof ha?t.converter(n):this._visit(t.target,ea.Expression).callFn(n))},t.prototype.visitImplicitReceiver=function(t,e){return sa(e,t),this._implicitReceiver},t.prototype.visitInterpolation=function(t,e){sa(e,t);for(var n=[Pe(t.expressions.length)],r=0;r<t.strings.length-1;r++)n.push(Pe(t.strings[r])),n.push(this._visit(t.expressions[r],ea.Expression));return n.push(Pe(t.strings[t.strings.length-1])),this.interpolationFunction?this.interpolationFunction(n):t.expressions.length<=9?Se(gn.inlineInterpolate).callFn(n):Se(gn.interpolate).callFn([n[0],ke(n.slice(1))])},t.prototype.visitKeyedRead=function(t,e){var n=this.leftMostSafeNode(t);return n?this.convertSafeAccess(t,n,e):ua(e,this._visit(t.obj,ea.Expression).key(this._visit(t.key,ea.Expression)))},t.prototype.visitKeyedWrite=function(t,e){var n=this._visit(t.obj,ea.Expression),r=this._visit(t.key,ea.Expression),i=this._visit(t.value,ea.Expression);return ua(e,n.key(r).set(i))},t.prototype.visitLiteralArray=function(t,e){throw new Error("Illegal State: literal arrays should have been converted into functions")},t.prototype.visitLiteralMap=function(t,e){throw new Error("Illegal State: literal maps should have been converted into functions")},t.prototype.visitLiteralPrimitive=function(t,e){var n=null===t.value||void 0===t.value||!0===t.value||!0===t.value?gt:void 0;return ua(e,Pe(t.value,n))},t.prototype._getLocal=function(t){return this._localResolver.getLocal(t)},t.prototype.visitMethodCall=function(t,e){if(t.receiver instanceof yo&&"$any"==t.name){if(1!=(r=this.visitAll(t.args,ea.Expression)).length)throw new Error("Invalid call to $any, expected 1 argument but received "+(r.length||"none"));return r[0].cast(mt)}var n=this.leftMostSafeNode(t);if(n)return this.convertSafeAccess(t,n,e);var r=this.visitAll(t.args,ea.Expression),i=null,o=this._visit(t.receiver,ea.Expression);if(o===this._implicitReceiver){var a=this._getLocal(t.name);a&&(i=a.callFn(r))}return null==i&&(i=o.callMethod(t.name,r)),ua(e,i)},t.prototype.visitPrefixNot=function(t,e){return ua(e,(n=this._visit(t.expression,ea.Expression),new Ht(n,r)));var n,r},t.prototype.visitNonNullAssert=function(t,e){return ua(e,(n=this._visit(t.expression,ea.Expression),new Ut(n,r)));var n,r},t.prototype.visitPropertyRead=function(t,e){var n=this.leftMostSafeNode(t);if(n)return this.convertSafeAccess(t,n,e);var r=null,i=this._visit(t.receiver,ea.Expression);return i===this._implicitReceiver&&(r=this._getLocal(t.name)),null==r&&(r=i.prop(t.name)),ua(e,r)},t.prototype.visitPropertyWrite=function(t,e){var n=this._visit(t.receiver,ea.Expression),r=null;if(n===this._implicitReceiver){var i=this._getLocal(t.name);if(i){if(!(i instanceof Kt))throw new Error("Cannot assign to a reference or variable!");r=i}}return null===r&&(r=n.prop(t.name)),ua(e,r.set(this._visit(t.value,ea.Expression)))},t.prototype.visitSafePropertyRead=function(t,e){return this.convertSafeAccess(t,this.leftMostSafeNode(t),e)},t.prototype.visitSafeMethodCall=function(t,e){return this.convertSafeAccess(t,this.leftMostSafeNode(t),e)},t.prototype.visitAll=function(t,e){var n=this;return t.map(function(t){return n._visit(t,e)})},t.prototype.visitQuote=function(t,e){throw new Error("Quotes are not supported for evaluation!\n        Statement: "+t.uninterpretedExpression+" located at "+t.location)},t.prototype._visit=function(t,e){var n=this._resultMap.get(t);return n||(this._nodeMap.get(t)||t).visit(this,e)},t.prototype.convertSafeAccess=function(t,e,n){var r=this._visit(e.receiver,ea.Expression),i=void 0;this.needsTemporary(e.receiver)&&(r=(i=this.allocateTemporary()).set(r),this._resultMap.set(e.receiver,i));var o=r.isBlank();e instanceof Po?this._nodeMap.set(e,new Io(e.span,e.receiver,e.name,e.args)):this._nodeMap.set(e,new bo(e.span,e.receiver,e.name));var a=this._visit(t,ea.Expression);return this._nodeMap.delete(e),i&&this.releaseTemporary(i),ua(n,o.conditional(Pe(null),a))},t.prototype.leftMostSafeNode=function(t){var e=this,n=function(t,n){return(e._nodeMap.get(n)||n).visit(t)};return t.visit({visitBinary:function(t){return null},visitChain:function(t){return null},visitConditional:function(t){return null},visitFunctionCall:function(t){return null},visitImplicitReceiver:function(t){return null},visitInterpolation:function(t){return null},visitKeyedRead:function(t){return n(this,t.obj)},visitKeyedWrite:function(t){return null},visitLiteralArray:function(t){return null},visitLiteralMap:function(t){return null},visitLiteralPrimitive:function(t){return null},visitMethodCall:function(t){return n(this,t.receiver)},visitPipe:function(t){return null},visitPrefixNot:function(t){return null},visitNonNullAssert:function(t){return null},visitPropertyRead:function(t){return n(this,t.receiver)},visitPropertyWrite:function(t){return null},visitQuote:function(t){return null},visitSafeMethodCall:function(t){return n(this,t.receiver)||t},visitSafePropertyRead:function(t){return n(this,t.receiver)||t}})},t.prototype.needsTemporary=function(t){var e=this,n=function(t,n){return n&&(e._nodeMap.get(n)||n).visit(t)};return t.visit({visitBinary:function(t){return n(this,t.left)||n(this,t.right)},visitChain:function(t){return!1},visitConditional:function(t){return n(this,t.condition)||n(this,t.trueExp)||n(this,t.falseExp)},visitFunctionCall:function(t){return!0},visitImplicitReceiver:function(t){return!1},visitInterpolation:function(t){return e=this,t.expressions.some(function(t){return n(e,t)});var e},visitKeyedRead:function(t){return!1},visitKeyedWrite:function(t){return!1},visitLiteralArray:function(t){return!0},visitLiteralMap:function(t){return!0},visitLiteralPrimitive:function(t){return!1},visitMethodCall:function(t){return!0},visitPipe:function(t){return!0},visitPrefixNot:function(t){return n(this,t.expression)},visitNonNullAssert:function(t){return n(this,t.expression)},visitPropertyRead:function(t){return!1},visitPropertyWrite:function(t){return!1},visitQuote:function(t){return!1},visitSafeMethodCall:function(t){return!0},visitSafePropertyRead:function(t){return!1}})},t.prototype.allocateTemporary=function(){var t=this._currentTemporary++;return this.temporaryCount=Math.max(this._currentTemporary,this.temporaryCount),new St(oa(this.bindingId,t))},t.prototype.releaseTemporary=function(t){if(this._currentTemporary--,t.name!=oa(this.bindingId,this._currentTemporary))throw new Error("Temporary "+t.name+" released out of order")},t}();var da=function(){function t(){}return t.prototype.getLocal=function(t){return t===Xo.event.name?Xo.event:null},t}();var pa,fa,ha=function(t){function e(e,n,r){var i=t.call(this,e,null,n)||this;return i.args=n,i.converter=r,i}return Object(o.__extends)(e,t),e}(Ro);(fa=pa||(pa={}))[fa.OnInit=0]="OnInit",fa[fa.OnDestroy=1]="OnDestroy",fa[fa.DoCheck=2]="DoCheck",fa[fa.OnChanges=3]="OnChanges",fa[fa.AfterContentInit=4]="AfterContentInit",fa[fa.AfterContentChecked=5]="AfterContentChecked",fa[fa.AfterViewInit=6]="AfterViewInit",fa[fa.AfterViewChecked=7]="AfterViewChecked";var ma=[pa.OnInit,pa.OnDestroy,pa.DoCheck,pa.OnChanges,pa.AfterContentInit,pa.AfterContentChecked,pa.AfterViewInit,pa.AfterViewChecked];function ga(t,e,n){return t.hasLifecycleHook(n,function(t){switch(t){case pa.OnInit:return"ngOnInit";case pa.OnDestroy:return"ngOnDestroy";case pa.DoCheck:return"ngDoCheck";case pa.OnChanges:return"ngOnChanges";case pa.AfterContentInit:return"ngAfterContentInit";case pa.AfterContentChecked:return"ngAfterContentChecked";case pa.AfterViewInit:return"ngAfterViewInit";case pa.AfterViewChecked:return"ngAfterViewChecked";default:var e=t;throw new Error("unexpected "+e)}}(e))}var ya=0,_a=9,va=10,ba=11,xa=12,wa=13,ja=32,Oa=34,Ca=36,Ea=39,Sa=43,Ma=45,Ta=47,ka=59,Aa=61,Da=62,Ia=48,Pa=57,Ra=65,Na=69,Fa=70,La=90,Ba=95,Va=97,za=101,Ha=102,Ua=110,Wa=114,qa=116,Ga=118,Ya=122,Ka=123,Xa=160,Qa=96;function Za(t){return t>=_a&&t<=ja||t==Xa}function $a(t){return Ia<=t&&t<=Pa}function Ja(t){return t>=Va&&t<=Ya||t>=Ra&&t<=La}var ts,es,ns=function(){function t(t,e,n,r){this.file=t,this.offset=e,this.line=n,this.col=r}return t.prototype.toString=function(){return null!=this.offset?this.file.url+"@"+this.line+":"+this.col:this.file.url},t.prototype.moveBy=function(e){for(var n=this.file.content,r=n.length,i=this.offset,o=this.line,a=this.col;i>0&&e<0;){if(i--,e++,(u=n.charCodeAt(i))==va){o--;var s=n.substr(0,i-1).lastIndexOf(String.fromCharCode(va));a=s>0?i-s:i}else a--}for(;i<r&&e>0;){var u=n.charCodeAt(i);i++,e--,u==va?(o++,a=0):a++}return new t(this.file,i,o,a)},t.prototype.getContext=function(t,e){var n=this.file.content,r=this.offset;if(null!=r){r>n.length-1&&(r=n.length-1);for(var i=r,o=0,a=0;o<t&&r>0&&(o++,"\n"!=n[--r]||++a!=e););for(o=0,a=0;o<t&&i<n.length-1&&(o++,"\n"!=n[++i]||++a!=e););return{before:n.substring(r,this.offset),after:n.substring(this.offset,i+1)}}return null},t}(),rs=function(){return function(t,e){this.content=t,this.url=e}}(),is=function(){function t(t,e,n){void 0===n&&(n=null),this.start=t,this.end=e,this.details=n}return t.prototype.toString=function(){return this.start.file.content.substring(this.start.offset,this.end.offset)},t}();(es=ts||(ts={}))[es.WARNING=0]="WARNING",es[es.ERROR=1]="ERROR";var os=function(){function t(t,e,n){void 0===n&&(n=ts.ERROR),this.span=t,this.msg=e,this.level=n}return t.prototype.contextualMessage=function(){var t=this.span.start.getContext(100,3);return t?this.msg+' ("'+t.before+"["+ts[this.level]+" ->]"+t.after+'")':this.msg},t.prototype.toString=function(){var t=this.span.details?", "+this.span.details:"";return this.contextualMessage()+": "+this.span.start+t},t}();function as(t,e){var n=Sn(e),r=null!=n?"in "+t+" "+En(e)+" in "+n:"in "+t+" "+En(e),i=new rs("",r);return new is(new ns(i,-1,-1,-1),new ns(i,-1,-1,-1))}var ss=function(){function t(){this.strictStyling=!0}return t.prototype.shimCssText=function(t,e,n){void 0===n&&(n="");var r=t.match(Es)||[];t=t.replace(Cs,""),t=this._insertDirectives(t);var i=this._scopeCssText(t,e,n);return Object(o.__spread)([i],r).join("\n")},t.prototype._insertDirectives=function(t){return t=this._insertPolyfillDirectivesInCssText(t),this._insertPolyfillRulesInCssText(t)},t.prototype._insertPolyfillDirectivesInCssText=function(t){return t.replace(cs,function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return t[2]+"{"})},t.prototype._insertPolyfillRulesInCssText=function(t){return t.replace(ls,function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];var n=t[0].replace(t[1],"").replace(t[2],"");return t[4]+n})},t.prototype._scopeCssText=function(t,e,n){var r=this._extractUnscopedRulesFromCssText(t);return t=this._insertPolyfillHostInCssText(t),t=this._convertColonHost(t),t=this._convertColonHostContext(t),t=this._convertShadowDOMSelectors(t),e&&(t=this._scopeSelectors(t,e,n)),(t=t+"\n"+r).trim()},t.prototype._extractUnscopedRulesFromCssText=function(t){var e,n="";for(ds.lastIndex=0;null!==(e=ds.exec(t));){n+=e[0].replace(e[2],"").replace(e[1],e[4])+"\n\n"}return n},t.prototype._convertColonHost=function(t){return this._convertColonRule(t,ms,this._colonHostPartReplacer)},t.prototype._convertColonHostContext=function(t){return this._convertColonRule(t,gs,this._colonHostContextPartReplacer)},t.prototype._convertColonRule=function(t,e,n){return t.replace(e,function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];if(t[2]){for(var r=t[2].split(","),i=[],o=0;o<r.length;o++){var a=r[o].trim();if(!a)break;i.push(n(ys,a,t[3]))}return i.join(",")}return ys+t[3]})},t.prototype._colonHostContextPartReplacer=function(t,e,n){return e.indexOf(ps)>-1?this._colonHostPartReplacer(t,e,n):t+e+n+", "+e+" "+t+n},t.prototype._colonHostPartReplacer=function(t,e,n){return t+e.replace(ps,"")+n},t.prototype._convertShadowDOMSelectors=function(t){return vs.reduce(function(t,e){return t.replace(e," ")},t)},t.prototype._scopeSelectors=function(t,e,n){var r,i,o,a=this;return r=function(t){var r=t.selector,i=t.content;return"@"!=t.selector[0]?r=a._scopeSelector(t.selector,e,n,a.strictStyling):(t.selector.startsWith("@media")||t.selector.startsWith("@supports")||t.selector.startsWith("@page")||t.selector.startsWith("@document"))&&(i=a._scopeSelectors(t.content,e,n)),new Ds(r,i)},i=function(t){for(var e=t.split(Ms),n=[],r=[],i=0,o=[],a=0;a<e.length;a++){var s=e[a];s==ks&&i--,i>0?o.push(s):(o.length>0&&(r.push(o.join("")),n.push(As),o=[]),n.push(s)),s==Ts&&i++}o.length>0&&(r.push(o.join("")),n.push(As));return new Is(n.join(""),r)}(t),o=0,i.escapedString.replace(Ss,function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];var n=t[2],a="",s=t[4],u="";s&&s.startsWith("{"+As)&&(a=i.blocks[o++],s=s.substring(As.length+1),u="{");var c=r(new Ds(n,a));return""+t[1]+c.selector+t[3]+u+c.content+s})},t.prototype._scopeSelector=function(t,e,n,r){var i=this;return t.split(",").map(function(t){return t.trim().split(bs)}).map(function(t){var a,s=Object(o.__read)(t),u=s[0],c=s.slice(1);return Object(o.__spread)([(a=u,i._selectorNeedsScoping(a,e)?r?i._applyStrictSelectorScope(a,e,n):i._applySelectorScope(a,e,n):a)],c).join(" ")}).join(", ")},t.prototype._selectorNeedsScoping=function(t,e){return!this._makeScopeMatcher(e).test(t)},t.prototype._makeScopeMatcher=function(t){return t=t.replace(/\[/g,"\\[").replace(/\]/g,"\\]"),new RegExp("^("+t+")"+xs,"m")},t.prototype._applySelectorScope=function(t,e,n){return this._applySimpleSelectorScope(t,e,n)},t.prototype._applySimpleSelectorScope=function(t,e,n){if(ws.lastIndex=0,ws.test(t)){var r=this.strictStyling?"["+n+"]":e;return t.replace(_s,function(t,e){return e.replace(/([^:]*)(:*)(.*)/,function(t,e,n,i){return e+r+n+i})}).replace(ws,r+" ")}return e+" "+t},t.prototype._applyStrictSelectorScope=function(t,e,n){for(var r,i=this,o="["+(e=e.replace(/\[is=([^\]]*)\]/g,function(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];return e[0]}))+"]",a=function(t){var r=t.trim();if(!r)return"";if(t.indexOf(ys)>-1)r=i._applySimpleSelectorScope(t,e,n);else{var a=t.replace(ws,"");if(a.length>0){var s=a.match(/([^:]*)(:*)(.*)/);s&&(r=s[1]+o+s[2]+s[3])}}return r},s=new us(t),u="",c=0,l=/( |>|\+|~(?!=))\s*/g,d=!((t=s.content()).indexOf(ys)>-1);null!==(r=l.exec(t));){var p=r[1],f=t.slice(c,r.index).trim();u+=((d=d||f.indexOf(ys)>-1)?a(f):f)+" "+p+" ",c=l.lastIndex}var h=t.substring(c);return u+=(d=d||h.indexOf(ys)>-1)?a(h):h,s.restore(u)},t.prototype._insertPolyfillHostInCssText=function(t){return t.replace(Os,fs).replace(js,ps)},t}(),us=function(){function t(t){var e=this;this.placeholders=[],this.index=0,t=t.replace(/(\[[^\]]*\])/g,function(t,n){var r="__ph-"+e.index+"__";return e.placeholders.push(n),e.index++,r}),this._content=t.replace(/(:nth-[-\w]+)(\([^)]+\))/g,function(t,n,r){var i="__ph-"+e.index+"__";return e.placeholders.push(r),e.index++,n+i})}return t.prototype.restore=function(t){var e=this;return t.replace(/__ph-(\d+)__/g,function(t,n){return e.placeholders[+n]})},t.prototype.content=function(){return this._content},t}(),cs=/polyfill-next-selector[^}]*content:[\s]*?(['"])(.*?)\1[;\s]*}([^{]*?){/gim,ls=/(polyfill-rule)[^}]*(content:[\s]*(['"])(.*?)\3)[;\s]*[^}]*}/gim,ds=/(polyfill-unscoped-rule)[^}]*(content:[\s]*(['"])(.*?)\3)[;\s]*[^}]*}/gim,ps="-shadowcsshost",fs="-shadowcsscontext",hs=")(?:\\(((?:\\([^)(]*\\)|[^)(]*)+?)\\))?([^,{]*)",ms=new RegExp("("+ps+hs,"gim"),gs=new RegExp("("+fs+hs,"gim"),ys=ps+"-no-combinator",_s=/-shadowcsshost-no-combinator([^\s]*)/,vs=[/::shadow/g,/::content/g,/\/shadow-deep\//g,/\/shadow\//g],bs=/(?:>>>)|(?:\/deep\/)|(?:::ng-deep)/g,xs="([>\\s~+[.,{:][\\s\\S]*)?$",ws=/-shadowcsshost/gim,js=/:host/gim,Os=/:host-context/gim,Cs=/\/\*\s*[\s\S]*?\*\//g;var Es=/\/\*\s*#\s*source(Mapping)?URL=[\s\S]+?\*\//g;var Ss=/(\s*)([^;\{\}]+?)(\s*)((?:{%BLOCK%}?\s*;?)|(?:\s*;))/g,Ms=/([{}])/g,Ts="{",ks="}",As="%BLOCK%",Ds=function(){return function(t,e){this.selector=t,this.content=e}}();var Is=function(){return function(t,e){this.escapedString=t,this.blocks=e}}();var Ps="_nghost-%COMP%",Rs="_ngcontent-%COMP%",Ns=function(){return function(t,e,n){this.name=t,this.moduleUrl=e,this.setValue=n}}(),Fs=function(){return function(t,e,n,r,i){this.outputCtx=t,this.stylesVar=e,this.dependencies=n,this.isShimmed=r,this.meta=i}}(),Ls=function(){function t(t){this._urlResolver=t,this._shadowCss=new ss}return t.prototype.compileComponent=function(t,e){var n=e.template;return this._compileStyles(t,e,new Pn({styles:n.styles,styleUrls:n.styleUrls,moduleUrl:Sn(e.type)}),this.needsStyleShim(e),!0)},t.prototype.compileStyles=function(t,e,n,r){return void 0===r&&(r=this.needsStyleShim(e)),this._compileStyles(t,e,n,r,!1)},t.prototype.needsStyleShim=function(t){return t.template.encapsulation===_.Emulated},t.prototype._compileStyles=function(t,e,n,r,i){var o=this,a=n.styles.map(function(t){return Pe(o._shimIfNeeded(t,r))}),s=[];n.styleUrls.forEach(function(e){var n=a.length;a.push(null),s.push(new Ns(Bs(null),e,function(e){return a[n]=t.importExpr(e)}))});var u=Bs(i?e:null),c=Ee(u).set(ke(a,new ft(mt,[rt.Const]))).toDeclStmt(null,i?[It.Final]:[It.Final,It.Exported]);return t.statements.push(c),new Fs(t,u,s,r,n)},t.prototype._shimIfNeeded=function(t,e){return e?this._shadowCss.shimCssText(t,Rs,Ps):t},t}();function Bs(t){var e="styles";return t&&(e+="_"+En(t.type)),e}function Vs(t){var e=t.charCodeAt(0);if(e==t.charCodeAt(t.length-1)&&(39==e||34==e)){var n=t.substring(1,t.length-1);-1==n.indexOf("'")&&-1==n.indexOf('"')&&(t=n)}return t}var zs,Hs,Us=function(){function t(t,e){this._elementIndexExpr=t,this._directiveExpr=e,this._hasInitialValues=!1,this._hasBindings=!1,this._classMapInput=null,this._styleMapInput=null,this._singleStyleInputs=null,this._singleClassInputs=null,this._lastStylingInput=null,this._stylesIndex=new Map,this._classesIndex=new Map,this._initialStyleValues=[],this._initialClassValues=[],this._useDefaultSanitizer=!1}return t.prototype.hasBindingsOrInitialValues=function(){return this._hasBindings||this._hasInitialValues},t.prototype.registerBoundInput=function(t){var e,n=t.name,r=null;switch(t.type){case 0:"style"==n?r=this.registerStyleInput(null,t.value,"",t.sourceSpan):("className"==(e=t.name)||"class"==e)&&(r=this.registerClassInput(null,t.value,t.sourceSpan));break;case 3:r=this.registerStyleInput(t.name,t.value,t.unit,t.sourceSpan);break;case 2:r=this.registerClassInput(t.name,t.value,t.sourceSpan)}return!!r},t.prototype.registerStyleInput=function(t,e,n,r){var i,o={name:t,unit:n,value:e,sourceSpan:r};return t?((this._singleStyleInputs=this._singleStyleInputs||[]).push(o),this._useDefaultSanitizer=this._useDefaultSanitizer||("background-image"===(i=t)||"background"===i||"border-image"===i||"filter"===i||"list-style"===i||"list-style-image"===i),Ws(this._stylesIndex,t)):(this._useDefaultSanitizer=!0,this._styleMapInput=o),this._lastStylingInput=o,this._hasBindings=!0,o},t.prototype.registerClassInput=function(t,e,n){var r={name:t,value:e,sourceSpan:n};return t?((this._singleClassInputs=this._singleClassInputs||[]).push(r),Ws(this._classesIndex,t)):this._classMapInput=r,this._lastStylingInput=r,this._hasBindings=!0,r},t.prototype.registerStyleAttr=function(t){this._initialStyleValues=function(t){for(var e=[],n=0,r=0,i=0,o=0,a=0,s=null,u=!1;n<t.length;)switch(t.charCodeAt(n++)){case 40:r++;break;case 41:r--;break;case 39:u=u||o>0,0===i?i=39:39===i&&92!==t.charCodeAt(n-1)&&(i=0);break;case 34:u=u||o>0,0===i?i=34:34===i&&92!==t.charCodeAt(n-1)&&(i=0);break;case 58:s||0!==r||0!==i||(s=t.substring(a,n-1).trim().replace(/[a-z][A-Z]/g,function(t){return t.charAt(0)+"-"+t.charAt(1)}).toLowerCase(),o=n);break;case 59:if(s&&o>0&&0===r&&0===i){var c=t.substring(o,n-1).trim();e.push(s,u?Vs(c):c),a=n,o=0,s=null,u=!1}}s&&o&&(c=t.substr(o).trim(),e.push(s,u?Vs(c):c));return e}(t),this._hasInitialValues=!0},t.prototype.registerClassAttr=function(t){this._initialClassValues=t.trim().split(/\s+/g),this._hasInitialValues=!0},t.prototype.populateInitialStylingAttrs=function(t){if(this._initialClassValues.length){t.push(Pe(1));for(var e=0;e<this._initialClassValues.length;e++)t.push(Pe(this._initialClassValues[e]))}if(this._initialStyleValues.length){t.push(Pe(2));for(e=0;e<this._initialStyleValues.length;e+=2)t.push(Pe(this._initialStyleValues[e]),Pe(this._initialStyleValues[e+1]))}},t.prototype.buildDirectiveHostAttrsInstruction=function(t,e){var n=this;return this._hasInitialValues&&this._directiveExpr?{sourceSpan:t,reference:Qn.elementHostAttrs,buildParams:function(){var t=[];return n.populateInitialStylingAttrs(t),[n._directiveExpr,qs(e,t)]}}:null},t.prototype.buildElementStylingInstruction=function(t,e){var n=this;return this._hasBindings?{sourceSpan:t,reference:Qn.elementStyling,buildParams:function(){var t=n._singleStyleInputs?n._singleStyleInputs.map(function(t){return Pe(t.name)}):[],r=n._singleClassInputs?n._singleClassInputs.map(function(t){return Pe(t.name)}):[],i=0;n._directiveExpr?i=4:n._useDefaultSanitizer?i=3:t.length?i=2:r.length&&(i=1);var o=[];return Gs(o,r.length>0,qs(e,r),1,i),Gs(o,t.length>0,qs(e,t),2,i),Gs(o,n._useDefaultSanitizer,Se(Qn.defaultStyleSanitizer),3,i),n._directiveExpr&&o.push(n._directiveExpr),o}}:null},t.prototype.buildElementStylingMapInstruction=function(t){var e=this;if(this._classMapInput||this._styleMapInput){var n=this._classMapInput||this._styleMapInput,r=this._classMapInput?this._classMapInput.value.visit(t):null,i=this._styleMapInput?this._styleMapInput.value.visit(t):null;return{sourceSpan:n.sourceSpan,reference:Qn.elementStylingMap,buildParams:function(t){var n=[e._elementIndexExpr];return r?n.push(t(r)):e._styleMapInput&&n.push(te),i?n.push(t(i)):e._directiveExpr&&n.push(te),e._directiveExpr&&n.push(e._directiveExpr),n}}}return null},t.prototype._buildSingleInputs=function(t,e,n,r,i){var o=this;return e.map(function(e){var a=n.get(e.name),s=e.value.visit(i);return{sourceSpan:e.sourceSpan,reference:t,buildParams:function(t){var n=[o._elementIndexExpr,Pe(a),t(s)];return r&&(e.unit?n.push(Pe(e.unit)):o._directiveExpr&&n.push(te)),o._directiveExpr&&n.push(o._directiveExpr),n}}})},t.prototype._buildClassInputs=function(t){return this._singleClassInputs?this._buildSingleInputs(Qn.elementClassProp,this._singleClassInputs,this._classesIndex,!1,t):[]},t.prototype._buildStyleInputs=function(t){return this._singleStyleInputs?this._buildSingleInputs(Qn.elementStyleProp,this._singleStyleInputs,this._stylesIndex,!0,t):[]},t.prototype._buildApplyFn=function(){var t=this;return{sourceSpan:this._lastStylingInput?this._lastStylingInput.sourceSpan:null,reference:Qn.elementStylingApply,buildParams:function(){var e=[t._elementIndexExpr];return t._directiveExpr&&e.push(t._directiveExpr),e}}},t.prototype.buildUpdateLevelInstructions=function(t){var e=[];if(this._hasBindings){var n=this.buildElementStylingMapInstruction(t);n&&e.push(n),e.push.apply(e,Object(o.__spread)(this._buildStyleInputs(t))),e.push.apply(e,Object(o.__spread)(this._buildClassInputs(t))),e.push(this._buildApplyFn())}return e},t}();function Ws(t,e){t.has(e)||t.set(e,t.size)}function qs(t,e){return e.length?t.getConstLiteral(ke(e),!0):te}function Gs(t,e,n,r,i){e?t.push(n):r<i&&t.push(te)}(Hs=zs||(zs={}))[Hs.Character=0]="Character",Hs[Hs.Identifier=1]="Identifier",Hs[Hs.Keyword=2]="Keyword",Hs[Hs.String=3]="String",Hs[Hs.Operator=4]="Operator",Hs[Hs.Number=5]="Number",Hs[Hs.Error=6]="Error";var Ys=["var","let","as","null","undefined","true","false","if","else","this"],Ks=function(){function t(){}return t.prototype.tokenize=function(t){for(var e=new Js(t),n=[],r=e.scanToken();null!=r;)n.push(r),r=e.scanToken();return n},t}(),Xs=function(){function t(t,e,n,r){this.index=t,this.type=e,this.numValue=n,this.strValue=r}return t.prototype.isCharacter=function(t){return this.type==zs.Character&&this.numValue==t},t.prototype.isNumber=function(){return this.type==zs.Number},t.prototype.isString=function(){return this.type==zs.String},t.prototype.isOperator=function(t){return this.type==zs.Operator&&this.strValue==t},t.prototype.isIdentifier=function(){return this.type==zs.Identifier},t.prototype.isKeyword=function(){return this.type==zs.Keyword},t.prototype.isKeywordLet=function(){return this.type==zs.Keyword&&"let"==this.strValue},t.prototype.isKeywordAs=function(){return this.type==zs.Keyword&&"as"==this.strValue},t.prototype.isKeywordNull=function(){return this.type==zs.Keyword&&"null"==this.strValue},t.prototype.isKeywordUndefined=function(){return this.type==zs.Keyword&&"undefined"==this.strValue},t.prototype.isKeywordTrue=function(){return this.type==zs.Keyword&&"true"==this.strValue},t.prototype.isKeywordFalse=function(){return this.type==zs.Keyword&&"false"==this.strValue},t.prototype.isKeywordThis=function(){return this.type==zs.Keyword&&"this"==this.strValue},t.prototype.isError=function(){return this.type==zs.Error},t.prototype.toNumber=function(){return this.type==zs.Number?this.numValue:-1},t.prototype.toString=function(){switch(this.type){case zs.Character:case zs.Identifier:case zs.Keyword:case zs.Operator:case zs.String:case zs.Error:return this.strValue;case zs.Number:return this.numValue.toString();default:return null}},t}();function Qs(t,e){return new Xs(t,zs.Character,e,String.fromCharCode(e))}function Zs(t,e){return new Xs(t,zs.Operator,0,e)}var $s=new Xs(-1,zs.Character,0,""),Js=function(){function t(t){this.input=t,this.peek=0,this.index=-1,this.length=t.length,this.advance()}return t.prototype.advance=function(){this.peek=++this.index>=this.length?ya:this.input.charCodeAt(this.index)},t.prototype.scanToken=function(){for(var t=this.input,e=this.length,n=this.peek,r=this.index;n<=ja;){if(++r>=e){n=ya;break}n=t.charCodeAt(r)}if(this.peek=n,this.index=r,r>=e)return null;if(tu(n))return this.scanIdentifier();if($a(n))return this.scanNumber(r);var i=r;switch(n){case 46:return this.advance(),$a(this.peek)?this.scanNumber(i):Qs(i,46);case 40:case 41:case Ka:case 125:case 91:case 93:case 44:case 58:case ka:return this.scanCharacter(i,n);case Ea:case Oa:return this.scanString();case 35:case Sa:case Ma:case 42:case Ta:case 37:case 94:return this.scanOperator(i,String.fromCharCode(n));case 63:return this.scanComplexOperator(i,"?",46,".");case 60:case Da:return this.scanComplexOperator(i,String.fromCharCode(n),Aa,"=");case 33:case Aa:return this.scanComplexOperator(i,String.fromCharCode(n),Aa,"=",Aa,"=");case 38:return this.scanComplexOperator(i,"&",38,"&");case 124:return this.scanComplexOperator(i,"|",124,"|");case Xa:for(;Za(this.peek);)this.advance();return this.scanToken()}return this.advance(),this.error("Unexpected character ["+String.fromCharCode(n)+"]",0)},t.prototype.scanCharacter=function(t,e){return this.advance(),Qs(t,e)},t.prototype.scanOperator=function(t,e){return this.advance(),Zs(t,e)},t.prototype.scanComplexOperator=function(t,e,n,r,i,o){this.advance();var a=e;return this.peek==n&&(this.advance(),a+=r),null!=i&&this.peek==i&&(this.advance(),a+=o),Zs(t,a)},t.prototype.scanIdentifier=function(){var t=this.index;for(this.advance();nu(this.peek);)this.advance();var e,n,r=this.input.substring(t,this.index);return Ys.indexOf(r)>-1?(n=r,new Xs(t,zs.Keyword,0,n)):(e=r,new Xs(t,zs.Identifier,0,e))},t.prototype.scanNumber=function(t){var e,n,r=this.index===t;for(this.advance();;){if($a(this.peek));else if(46==this.peek)r=!1;else{if((n=this.peek)!=za&&n!=Na)break;if(this.advance(),((e=this.peek)==Ma||e==Sa)&&this.advance(),!$a(this.peek))return this.error("Invalid exponent",-1);r=!1}this.advance()}var i,o=this.input.substring(t,this.index),a=r?function(t){var e=parseInt(t);if(isNaN(e))throw new Error("Invalid integer literal when parsing "+t);return e}(o):parseFloat(o);return i=a,new Xs(t,zs.Number,i,"")},t.prototype.scanString=function(){var t=this.index,e=this.peek;this.advance();for(var n="",r=this.index,i=this.input;this.peek!=e;)if(92==this.peek){n+=i.substring(r,this.index),this.advance();var o=void 0;if(this.peek=this.peek,117==this.peek){var a=i.substring(this.index+1,this.index+5);if(!/^[0-9a-f]+$/i.test(a))return this.error("Invalid unicode escape [\\u"+a+"]",0);o=parseInt(a,16);for(var s=0;s<5;s++)this.advance()}else o=iu(this.peek),this.advance();n+=String.fromCharCode(o),r=this.index}else{if(this.peek==ya)return this.error("Unterminated quote",0);this.advance()}var u,c=i.substring(r,this.index);return this.advance(),u=n+c,new Xs(t,zs.String,0,u)},t.prototype.error=function(t,e){var n,r,i=this.index+e;return n=i,r="Lexer Error: "+t+" at column "+i+" in expression ["+this.input+"]",new Xs(n,zs.Error,0,r)},t}();function tu(t){return Va<=t&&t<=Ya||Ra<=t&&t<=La||t==Ba||t==Ca}function eu(t){if(0==t.length)return!1;var e=new Js(t);if(!tu(e.peek))return!1;for(e.advance();e.peek!==ya;){if(!nu(e.peek))return!1;e.advance()}return!0}function nu(t){return Ja(t)||$a(t)||t==Ba||t==Ca}function ru(t){return t===Ea||t===Oa||t===Qa}function iu(t){switch(t){case Ua:return va;case Ha:return xa;case Wa:return wa;case qa:return _a;case Ga:return ba;default:return t}}var ou=function(){return function(t,e,n){this.strings=t,this.expressions=e,this.offsets=n}}(),au=function(){return function(t,e,n){this.templateBindings=t,this.warnings=e,this.errors=n}}();function su(t){var e=$e(t.start)+"([\\s\\S]*?)"+$e(t.end);return new RegExp(e,"g")}var uu=function(){function t(t){this._lexer=t,this.errors=[]}return t.prototype.parseAction=function(t,e,n){void 0===n&&(n=Hi),this._checkNoInterpolation(t,e,n);var r=this._stripComments(t),i=this._lexer.tokenize(this._stripComments(t)),o=new cu(t,e,i,r.length,!0,this.errors,t.length-r.length).parseChain();return new No(o,t,e,this.errors)},t.prototype.parseBinding=function(t,e,n){void 0===n&&(n=Hi);var r=this._parseBindingAst(t,e,n);return new No(r,t,e,this.errors)},t.prototype.parseSimpleBinding=function(t,e,n){void 0===n&&(n=Hi);var r=this._parseBindingAst(t,e,n),i=lu.check(r);return i.length>0&&this._reportError("Host binding expression cannot contain "+i.join(" "),t,e),new No(r,t,e,this.errors)},t.prototype._reportError=function(t,e,n,r){this.errors.push(new po(t,e,n,r))},t.prototype._parseBindingAst=function(t,e,n){var r=this._parseQuote(t,e);if(null!=r)return r;this._checkNoInterpolation(t,e,n);var i=this._stripComments(t),o=this._lexer.tokenize(i);return new cu(t,e,o,i.length,!1,this.errors,t.length-i.length).parseChain()},t.prototype._parseQuote=function(t,e){if(null==t)return null;var n=t.indexOf(":");if(-1==n)return null;var r=t.substring(0,n).trim();if(!eu(r))return null;var i=t.substring(n+1);return new mo(new fo(0,t.length),r,i,e)},t.prototype.parseTemplateBindings=function(t,e,n){var r=this._lexer.tokenize(e);return new cu(e,n,r,e.length,!1,this.errors,0).parseTemplateBindings(t)},t.prototype.parseInterpolation=function(t,e,n){void 0===n&&(n=Hi);var r=this.splitInterpolation(t,e,n);if(null==r)return null;for(var i=[],o=0;o<r.expressions.length;++o){var a=r.expressions[o],s=this._stripComments(a),u=this._lexer.tokenize(s),c=new cu(t,e,u,s.length,!1,this.errors,r.offsets[o]+(a.length-s.length)).parseChain();i.push(c)}return new No(new To(new fo(0,null==t?0:t.length),r.strings,i),t,e,this.errors)},t.prototype.splitInterpolation=function(t,e,n){void 0===n&&(n=Hi);var r=su(n),i=t.split(r);if(i.length<=1)return null;for(var o=[],a=[],s=[],u=0,c=0;c<i.length;c++){var l=i[c];c%2==0?(o.push(l),u+=l.length):l.trim().length>0?(u+=n.start.length,a.push(l),s.push(u),u+=l.length+n.end.length):(this._reportError("Blank expressions are not allowed in interpolated strings",t,"at column "+this._findInterpolationErrorColumn(i,c,n)+" in",e),a.push("$implict"),s.push(u))}return new ou(o,a,s)},t.prototype.wrapLiteralPrimitive=function(t,e){return new No(new Eo(new fo(0,null==t?0:t.length),t),t,e,this.errors)},t.prototype._stripComments=function(t){var e=this._commentStart(t);return null!=e?t.substring(0,e).trim():t},t.prototype._commentStart=function(t){for(var e=null,n=0;n<t.length-1;n++){var r=t.charCodeAt(n),i=t.charCodeAt(n+1);if(r===Ta&&i==Ta&&null==e)return n;e===r?e=null:null==e&&ru(r)&&(e=r)}return null},t.prototype._checkNoInterpolation=function(t,e,n){var r=su(n),i=t.split(r);i.length>1&&this._reportError("Got interpolation ("+n.start+n.end+") where expression was expected",t,"at column "+this._findInterpolationErrorColumn(i,1,n)+" in",e)},t.prototype._findInterpolationErrorColumn=function(t,e,n){for(var r="",i=0;i<e;i++)r+=i%2==0?t[i]:""+n.start+t[i]+n.end;return r.length},t}(),cu=function(){function t(t,e,n,r,i,o,a){this.input=t,this.location=e,this.tokens=n,this.inputLength=r,this.parseAction=i,this.errors=o,this.offset=a,this.rparensExpected=0,this.rbracketsExpected=0,this.rbracesExpected=0,this.index=0}return t.prototype.peek=function(t){var e=this.index+t;return e<this.tokens.length?this.tokens[e]:$s},Object.defineProperty(t.prototype,"next",{get:function(){return this.peek(0)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"inputIndex",{get:function(){return this.index<this.tokens.length?this.next.index+this.offset:this.inputLength+this.offset},enumerable:!0,configurable:!0}),t.prototype.span=function(t){return new fo(t,this.inputIndex)},t.prototype.advance=function(){this.index++},t.prototype.optionalCharacter=function(t){return!!this.next.isCharacter(t)&&(this.advance(),!0)},t.prototype.peekKeywordLet=function(){return this.next.isKeywordLet()},t.prototype.peekKeywordAs=function(){return this.next.isKeywordAs()},t.prototype.expectCharacter=function(t){this.optionalCharacter(t)||this.error("Missing expected "+String.fromCharCode(t))},t.prototype.optionalOperator=function(t){return!!this.next.isOperator(t)&&(this.advance(),!0)},t.prototype.expectOperator=function(t){this.optionalOperator(t)||this.error("Missing expected operator "+t)},t.prototype.expectIdentifierOrKeyword=function(){var t=this.next;return t.isIdentifier()||t.isKeyword()?(this.advance(),t.toString()):(this.error("Unexpected token "+t+", expected identifier or keyword"),"")},t.prototype.expectIdentifierOrKeywordOrString=function(){var t=this.next;return t.isIdentifier()||t.isKeyword()||t.isString()?(this.advance(),t.toString()):(this.error("Unexpected token "+t+", expected identifier, keyword, or string"),"")},t.prototype.parseChain=function(){for(var t=[],e=this.inputIndex;this.index<this.tokens.length;){var n=this.parsePipe();if(t.push(n),this.optionalCharacter(ka))for(this.parseAction||this.error("Binding expression cannot contain chained expression");this.optionalCharacter(ka););else this.index<this.tokens.length&&this.error("Unexpected token '"+this.next+"'")}return 0==t.length?new go(this.span(e)):1==t.length?t[0]:new _o(this.span(e),t)},t.prototype.parsePipe=function(){var t=this.parseExpression();if(this.optionalOperator("|")){this.parseAction&&this.error("Cannot have a pipe in an action expression");do{for(var e=this.expectIdentifierOrKeyword(),n=[];this.optionalCharacter(58);)n.push(this.parseExpression());t=new Co(this.span(t.span.start),t,e,n)}while(this.optionalOperator("|"))}return t},t.prototype.parseExpression=function(){return this.parseConditional()},t.prototype.parseConditional=function(){var t=this.inputIndex,e=this.parseLogicalOr();if(this.optionalOperator("?")){var n=this.parsePipe(),r=void 0;if(this.optionalCharacter(58))r=this.parsePipe();else{var i=this.inputIndex,o=this.input.substring(t,i);this.error("Conditional expression "+o+" requires all 3 expressions"),r=new go(this.span(t))}return new vo(this.span(t),e,n,r)}return e},t.prototype.parseLogicalOr=function(){for(var t=this.parseLogicalAnd();this.optionalOperator("||");){var e=this.parseLogicalAnd();t=new ko(this.span(t.span.start),"||",t,e)}return t},t.prototype.parseLogicalAnd=function(){for(var t=this.parseEquality();this.optionalOperator("&&");){var e=this.parseEquality();t=new ko(this.span(t.span.start),"&&",t,e)}return t},t.prototype.parseEquality=function(){for(var t=this.parseRelational();this.next.type==zs.Operator;){var e=this.next.strValue;switch(e){case"==":case"===":case"!=":case"!==":this.advance();var n=this.parseRelational();t=new ko(this.span(t.span.start),e,t,n);continue}break}return t},t.prototype.parseRelational=function(){for(var t=this.parseAdditive();this.next.type==zs.Operator;){var e=this.next.strValue;switch(e){case"<":case">":case"<=":case">=":this.advance();var n=this.parseAdditive();t=new ko(this.span(t.span.start),e,t,n);continue}break}return t},t.prototype.parseAdditive=function(){for(var t=this.parseMultiplicative();this.next.type==zs.Operator;){var e=this.next.strValue;switch(e){case"+":case"-":this.advance();var n=this.parseMultiplicative();t=new ko(this.span(t.span.start),e,t,n);continue}break}return t},t.prototype.parseMultiplicative=function(){for(var t=this.parsePrefix();this.next.type==zs.Operator;){var e=this.next.strValue;switch(e){case"*":case"%":case"/":this.advance();var n=this.parsePrefix();t=new ko(this.span(t.span.start),e,t,n);continue}break}return t},t.prototype.parsePrefix=function(){if(this.next.type==zs.Operator){var t=this.inputIndex,e=this.next.strValue,n=void 0;switch(e){case"+":return this.advance(),n=this.parsePrefix(),new ko(this.span(t),"-",n,new Eo(new fo(t,t),0));case"-":return this.advance(),n=this.parsePrefix(),new ko(this.span(t),e,new Eo(new fo(t,t),0),n);case"!":return this.advance(),n=this.parsePrefix(),new Ao(this.span(t),n)}}return this.parseCallChain()},t.prototype.parseCallChain=function(){for(var t=this.parsePrimary();;)if(this.optionalCharacter(46))t=this.parseAccessMemberOrMethodCall(t,!1);else if(this.optionalOperator("?."))t=this.parseAccessMemberOrMethodCall(t,!0);else if(this.optionalCharacter(91)){this.rbracketsExpected++;var e=this.parsePipe();if(this.rbracketsExpected--,this.expectCharacter(93),this.optionalOperator("=")){var n=this.parseConditional();t=new Oo(this.span(t.span.start),t,e,n)}else t=new jo(this.span(t.span.start),t,e)}else if(this.optionalCharacter(40)){this.rparensExpected++;var r=this.parseCallArguments();this.rparensExpected--,this.expectCharacter(41),t=new Ro(this.span(t.span.start),t,r)}else{if(!this.optionalOperator("!"))return t;t=new Do(this.span(t.span.start),t)}},t.prototype.parsePrimary=function(){var t=this.inputIndex;if(this.optionalCharacter(40)){this.rparensExpected++;var e=this.parsePipe();return this.rparensExpected--,this.expectCharacter(41),e}if(this.next.isKeywordNull())return this.advance(),new Eo(this.span(t),null);if(this.next.isKeywordUndefined())return this.advance(),new Eo(this.span(t),void 0);if(this.next.isKeywordTrue())return this.advance(),new Eo(this.span(t),!0);if(this.next.isKeywordFalse())return this.advance(),new Eo(this.span(t),!1);if(this.next.isKeywordThis())return this.advance(),new yo(this.span(t));if(this.optionalCharacter(91)){this.rbracketsExpected++;var n=this.parseExpressionList(93);return this.rbracketsExpected--,this.expectCharacter(93),new So(this.span(t),n)}if(this.next.isCharacter(Ka))return this.parseLiteralMap();if(this.next.isIdentifier())return this.parseAccessMemberOrMethodCall(new yo(this.span(t)),!1);if(this.next.isNumber()){var r=this.next.toNumber();return this.advance(),new Eo(this.span(t),r)}if(this.next.isString()){var i=this.next.toString();return this.advance(),new Eo(this.span(t),i)}return this.index>=this.tokens.length?(this.error("Unexpected end of expression: "+this.input),new go(this.span(t))):(this.error("Unexpected token "+this.next),new go(this.span(t)))},t.prototype.parseExpressionList=function(t){var e=[];if(!this.next.isCharacter(t))do{e.push(this.parsePipe())}while(this.optionalCharacter(44));return e},t.prototype.parseLiteralMap=function(){var t=[],e=[],n=this.inputIndex;if(this.expectCharacter(Ka),!this.optionalCharacter(125)){this.rbracesExpected++;do{var r=this.next.isString(),i=this.expectIdentifierOrKeywordOrString();t.push({key:i,quoted:r}),this.expectCharacter(58),e.push(this.parsePipe())}while(this.optionalCharacter(44));this.rbracesExpected--,this.expectCharacter(125)}return new Mo(this.span(n),t,e)},t.prototype.parseAccessMemberOrMethodCall=function(t,e){void 0===e&&(e=!1);var n=t.span.start,r=this.expectIdentifierOrKeyword();if(this.optionalCharacter(40)){this.rparensExpected++;var i=this.parseCallArguments();this.expectCharacter(41),this.rparensExpected--;var o=this.span(n);return e?new Po(o,t,r,i):new Io(o,t,r,i)}if(e)return this.optionalOperator("=")?(this.error("The '?.' operator cannot be used in the assignment"),new go(this.span(n))):new wo(this.span(n),t,r);if(this.optionalOperator("=")){if(!this.parseAction)return this.error("Bindings cannot contain assignments"),new go(this.span(n));var a=this.parseConditional();return new xo(this.span(n),t,r,a)}return new bo(this.span(n),t,r)},t.prototype.parseCallArguments=function(){if(this.next.isCharacter(41))return[];var t=[];do{t.push(this.parsePipe())}while(this.optionalCharacter(44));return t},t.prototype.expectTemplateBindingKey=function(){var t="",e=!1;do{t+=this.expectIdentifierOrKeywordOrString(),(e=this.optionalOperator("-"))&&(t+="-")}while(e);return t.toString()},t.prototype.parseTemplateBindings=function(t){var e=!0,n=[];do{var r=this.inputIndex,i=void 0,o=void 0,a=!1;e?(i=o=t,e=!1):((a=this.peekKeywordLet())&&this.advance(),i=this.expectTemplateBindingKey(),o=a?i:t+i[0].toUpperCase()+i.substring(1),this.optionalCharacter(58));var s=null,u=null;if(a)s=this.optionalOperator("=")?this.expectTemplateBindingKey():"$implicit";else if(this.peekKeywordAs())this.advance(),s=i,o=this.expectTemplateBindingKey(),a=!0;else if(this.next!==$s&&!this.peekKeywordLet()){var c=this.inputIndex,l=this.parsePipe(),d=this.input.substring(c-this.offset,this.inputIndex-this.offset);u=new No(l,d,this.location,this.errors)}if(n.push(new Fo(this.span(r),o,a,s,u)),this.peekKeywordAs()&&!a){var p=this.inputIndex;this.advance();var f=this.expectTemplateBindingKey();n.push(new Fo(this.span(p),f,!0,o,null))}this.optionalCharacter(ka)||this.optionalCharacter(44)}while(this.index<this.tokens.length);return new au(n,[],this.errors)},t.prototype.error=function(t,e){void 0===e&&(e=null),this.errors.push(new po(t,this.input,this.locationText(e),this.location)),this.skip()},t.prototype.locationText=function(t){return void 0===t&&(t=null),null==t&&(t=this.index),t<this.tokens.length?"at column "+(this.tokens[t].index+1)+" in":"at the end of the expression"},t.prototype.skip=function(){for(var t=this.next;this.index<this.tokens.length&&!t.isCharacter(ka)&&(this.rparensExpected<=0||!t.isCharacter(41))&&(this.rbracesExpected<=0||!t.isCharacter(125))&&(this.rbracketsExpected<=0||!t.isCharacter(93));)this.next.isError()&&this.errors.push(new po(this.next.toString(),this.input,this.locationText(),this.location)),this.advance(),t=this.next},t}(),lu=function(){function t(){this.errors=[]}return t.check=function(e){var n=new t;return e.visit(n),n.errors},t.prototype.visitImplicitReceiver=function(t,e){},t.prototype.visitInterpolation=function(t,e){},t.prototype.visitLiteralPrimitive=function(t,e){},t.prototype.visitPropertyRead=function(t,e){},t.prototype.visitPropertyWrite=function(t,e){},t.prototype.visitSafePropertyRead=function(t,e){},t.prototype.visitMethodCall=function(t,e){},t.prototype.visitSafeMethodCall=function(t,e){},t.prototype.visitFunctionCall=function(t,e){},t.prototype.visitLiteralArray=function(t,e){this.visitAll(t.expressions)},t.prototype.visitLiteralMap=function(t,e){this.visitAll(t.values)},t.prototype.visitBinary=function(t,e){},t.prototype.visitPrefixNot=function(t,e){},t.prototype.visitNonNullAssert=function(t,e){},t.prototype.visitConditional=function(t,e){},t.prototype.visitPipe=function(t,e){this.errors.push("pipes")},t.prototype.visitKeyedRead=function(t,e){},t.prototype.visitKeyedWrite=function(t,e){},t.prototype.visitAll=function(t){var e=this;return t.map(function(t){return t.visit(e)})},t.prototype.visitChain=function(t,e){},t.prototype.visitQuote=function(t,e){},t}(),du=function(){function t(t,e){void 0===e&&(e=-1),this.path=t,this.position=e}return Object.defineProperty(t.prototype,"empty",{get:function(){return!this.path||!this.path.length},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"head",{get:function(){return this.path[0]},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"tail",{get:function(){return this.path[this.path.length-1]},enumerable:!0,configurable:!0}),t.prototype.parentOf=function(t){return t&&this.path[this.path.indexOf(t)-1]},t.prototype.childOf=function(t){return this.path[this.path.indexOf(t)+1]},t.prototype.first=function(t){for(var e=this.path.length-1;e>=0;e--){var n=this.path[e];if(n instanceof t)return n}},t.prototype.push=function(t){this.path.push(t)},t.prototype.pop=function(){return this.path.pop()},t}(),pu=function(){function t(t,e,n){this.value=t,this.sourceSpan=e,this.i18n=n}return t.prototype.visit=function(t,e){return t.visitText(this,e)},t}(),fu=function(){function t(t,e,n,r,i,o){this.switchValue=t,this.type=e,this.cases=n,this.sourceSpan=r,this.switchValueSourceSpan=i,this.i18n=o}return t.prototype.visit=function(t,e){return t.visitExpansion(this,e)},t}(),hu=function(){function t(t,e,n,r,i){this.value=t,this.expression=e,this.sourceSpan=n,this.valueSourceSpan=r,this.expSourceSpan=i}return t.prototype.visit=function(t,e){return t.visitExpansionCase(this,e)},t}(),mu=function(){function t(t,e,n,r,i){this.name=t,this.value=e,this.sourceSpan=n,this.valueSpan=r,this.i18n=i}return t.prototype.visit=function(t,e){return t.visitAttribute(this,e)},t}(),gu=function(){function t(t,e,n,r,i,o,a){void 0===i&&(i=null),void 0===o&&(o=null),this.name=t,this.attrs=e,this.children=n,this.sourceSpan=r,this.startSourceSpan=i,this.endSourceSpan=o,this.i18n=a}return t.prototype.visit=function(t,e){return t.visitElement(this,e)},t}(),yu=function(){function t(t,e){this.value=t,this.sourceSpan=e}return t.prototype.visit=function(t,e){return t.visitComment(this,e)},t}();function _u(t,e,n){void 0===n&&(n=null);var r=[],i=t.visit?function(e){return t.visit(e,n)||e.visit(t,n)}:function(e){return e.visit(t,n)};return e.forEach(function(t){var e=i(t);e&&r.push(e)}),r}var vu,bu,xu=function(){function t(){}return t.prototype.visitElement=function(t,e){this.visitChildren(e,function(e){e(t.attrs),e(t.children)})},t.prototype.visitAttribute=function(t,e){},t.prototype.visitText=function(t,e){},t.prototype.visitComment=function(t,e){},t.prototype.visitExpansion=function(t,e){return this.visitChildren(e,function(e){e(t.cases)})},t.prototype.visitExpansionCase=function(t,e){},t.prototype.visitChildren=function(t,e){var n=[],r=this;return e(function(e){e&&n.push(_u(r,e,t))}),[].concat.apply([],n)},t}();function wu(t,e){var n=[];return _u(new(function(t){function r(){return null!==t&&t.apply(this,arguments)||this}return Object(o.__extends)(r,t),r.prototype.visit=function(t,r){var i=function t(e){var n=e.sourceSpan.start.offset,r=e.sourceSpan.end.offset;return e instanceof gu&&(e.endSourceSpan?r=e.endSourceSpan.end.offset:e.children&&e.children.length&&(r=t(e.children[e.children.length-1]).end)),{start:n,end:r}}(t);if(!(i.start<=e&&e<i.end))return!0;n.push(t)},r}(xu)),t),new du(n,e)}(bu=vu||(vu={}))[bu.TAG_OPEN_START=0]="TAG_OPEN_START",bu[bu.TAG_OPEN_END=1]="TAG_OPEN_END",bu[bu.TAG_OPEN_END_VOID=2]="TAG_OPEN_END_VOID",bu[bu.TAG_CLOSE=3]="TAG_CLOSE",bu[bu.TEXT=4]="TEXT",bu[bu.ESCAPABLE_RAW_TEXT=5]="ESCAPABLE_RAW_TEXT",bu[bu.RAW_TEXT=6]="RAW_TEXT",bu[bu.COMMENT_START=7]="COMMENT_START",bu[bu.COMMENT_END=8]="COMMENT_END",bu[bu.CDATA_START=9]="CDATA_START",bu[bu.CDATA_END=10]="CDATA_END",bu[bu.ATTR_NAME=11]="ATTR_NAME",bu[bu.ATTR_VALUE=12]="ATTR_VALUE",bu[bu.DOC_TYPE=13]="DOC_TYPE",bu[bu.EXPANSION_FORM_START=14]="EXPANSION_FORM_START",bu[bu.EXPANSION_CASE_VALUE=15]="EXPANSION_CASE_VALUE",bu[bu.EXPANSION_CASE_EXP_START=16]="EXPANSION_CASE_EXP_START",bu[bu.EXPANSION_CASE_EXP_END=17]="EXPANSION_CASE_EXP_END",bu[bu.EXPANSION_FORM_END=18]="EXPANSION_FORM_END",bu[bu.EOF=19]="EOF";var ju=function(){return function(t,e,n){this.type=t,this.parts=e,this.sourceSpan=n}}(),Ou=function(t){function e(e,n,r){var i=t.call(this,r,e)||this;return i.tokenType=n,i}return Object(o.__extends)(e,t),e}(os),Cu=function(){return function(t,e){this.tokens=t,this.errors=e}}();var Eu=/\r\n?/g;function Su(t){return'Unexpected character "'+(t===ya?"EOF":String.fromCharCode(t))+'"'}function Mu(t){return'Unknown entity "'+t+'" - use the "&#<decimal>;" or  "&#x<hex>;" syntax'}var Tu=function(){return function(t){this.error=t}}(),ku=function(){function t(t,e,n,r){void 0===r&&(r=Hi),this._file=t,this._getTagDefinition=e,this._tokenizeIcu=n,this._interpolationConfig=r,this._peek=-1,this._nextPeek=-1,this._index=-1,this._line=0,this._column=-1,this._expansionCaseStack=[],this._inInterpolation=!1,this.tokens=[],this.errors=[],this._input=t.content,this._length=t.content.length,this._advance()}return t.prototype._processCarriageReturns=function(t){return t.replace(Eu,"\n")},t.prototype.tokenize=function(){for(;this._peek!==ya;){var t=this._getLocation();try{this._attemptCharCode(60)?this._attemptCharCode(33)?this._attemptCharCode(91)?this._consumeCdata(t):this._attemptCharCode(Ma)?this._consumeComment(t):this._consumeDocType(t):this._attemptCharCode(Ta)?this._consumeTagClose(t):this._consumeTagOpen(t):this._tokenizeIcu&&this._tokenizeExpansionForm()||this._consumeText()}catch(t){if(!(t instanceof Tu))throw t;this.errors.push(t.error)}}return this._beginToken(vu.EOF),this._endToken([]),new Cu(function(t){for(var e=[],n=void 0,r=0;r<t.length;r++){var i=t[r];n&&n.type==vu.TEXT&&i.type==vu.TEXT?(n.parts[0]+=i.parts[0],n.sourceSpan.end=i.sourceSpan.end):(n=i,e.push(n))}return e}(this.tokens),this.errors)},t.prototype._tokenizeExpansionForm=function(){if(Ru(this._input,this._index,this._interpolationConfig))return this._consumeExpansionFormStart(),!0;if(((t=this._peek)===Aa||Ja(t)||$a(t))&&this._isInExpansionForm())return this._consumeExpansionCaseStart(),!0;var t;if(125===this._peek){if(this._isInExpansionCase())return this._consumeExpansionCaseEnd(),!0;if(this._isInExpansionForm())return this._consumeExpansionFormEnd(),!0}return!1},t.prototype._getLocation=function(){return new ns(this._file,this._index,this._line,this._column)},t.prototype._getSpan=function(t,e){return void 0===t&&(t=this._getLocation()),void 0===e&&(e=this._getLocation()),new is(t,e)},t.prototype._beginToken=function(t,e){void 0===e&&(e=this._getLocation()),this._currentTokenStart=e,this._currentTokenType=t},t.prototype._endToken=function(t,e){void 0===e&&(e=this._getLocation());var n=new ju(this._currentTokenType,t,new is(this._currentTokenStart,e));return this.tokens.push(n),this._currentTokenStart=null,this._currentTokenType=null,n},t.prototype._createError=function(t,e){this._isInExpansionForm()&&(t+=' (Do you have an unescaped "{" in your template? Use "{{ \'{\' }}") to escape it.)');var n=new Ou(t,this._currentTokenType,e);return this._currentTokenStart=null,this._currentTokenType=null,new Tu(n)},t.prototype._advance=function(){if(this._index>=this._length)throw this._createError(Su(ya),this._getSpan());this._peek===va?(this._line++,this._column=0):this._peek!==va&&this._peek!==wa&&this._column++,this._index++,this._peek=this._index>=this._length?ya:this._input.charCodeAt(this._index),this._nextPeek=this._index+1>=this._length?ya:this._input.charCodeAt(this._index+1)},t.prototype._attemptCharCode=function(t){return this._peek===t&&(this._advance(),!0)},t.prototype._attemptCharCodeCaseInsensitive=function(t){return e=this._peek,n=t,Nu(e)==Nu(n)&&(this._advance(),!0);var e,n},t.prototype._requireCharCode=function(t){var e=this._getLocation();if(!this._attemptCharCode(t))throw this._createError(Su(this._peek),this._getSpan(e,e))},t.prototype._attemptStr=function(t){var e=t.length;if(this._index+e>this._length)return!1;for(var n=this._savePosition(),r=0;r<e;r++)if(!this._attemptCharCode(t.charCodeAt(r)))return this._restorePosition(n),!1;return!0},t.prototype._attemptStrCaseInsensitive=function(t){for(var e=0;e<t.length;e++)if(!this._attemptCharCodeCaseInsensitive(t.charCodeAt(e)))return!1;return!0},t.prototype._requireStr=function(t){var e=this._getLocation();if(!this._attemptStr(t))throw this._createError(Su(this._peek),this._getSpan(e))},t.prototype._attemptCharCodeUntilFn=function(t){for(;!t(this._peek);)this._advance()},t.prototype._requireCharCodeUntilFn=function(t,e){var n=this._getLocation();if(this._attemptCharCodeUntilFn(t),this._index-n.offset<e)throw this._createError(Su(this._peek),this._getSpan(n,n))},t.prototype._attemptUntilChar=function(t){for(;this._peek!==t;)this._advance()},t.prototype._readChar=function(t){if(t&&38===this._peek)return this._decodeEntity();var e=this._index;return this._advance(),this._input[e]},t.prototype._decodeEntity=function(){var t=this._getLocation();if(this._advance(),!this._attemptCharCode(35)){var e=this._savePosition();if(this._attemptCharCodeUntilFn(Pu),this._peek!=ka)return this._restorePosition(e),"&";this._advance();var n=this._input.substring(t.offset+1,this._index-1),r=p[n];if(!r)throw this._createError(Mu(n),this._getSpan(t));return r}var i=this._attemptCharCode(120)||this._attemptCharCode(88),o=this._getLocation().offset;if(this._attemptCharCodeUntilFn(Iu),this._peek!=ka)throw this._createError(Su(this._peek),this._getSpan());this._advance();var a=this._input.substring(o,this._index-1);try{var s=parseInt(a,i?16:10);return String.fromCharCode(s)}catch(e){var u=this._input.substring(t.offset+1,this._index-1);throw this._createError(Mu(u),this._getSpan(t))}},t.prototype._consumeRawText=function(t,e,n){var r,i=this._getLocation();this._beginToken(t?vu.ESCAPABLE_RAW_TEXT:vu.RAW_TEXT,i);for(var o=[];r=this._getLocation(),!this._attemptCharCode(e)||!n();)for(this._index>r.offset&&o.push(this._input.substring(r.offset,this._index));this._peek!==e;)o.push(this._readChar(t));return this._endToken([this._processCarriageReturns(o.join(""))],r)},t.prototype._consumeComment=function(t){var e=this;this._beginToken(vu.COMMENT_START,t),this._requireCharCode(Ma),this._endToken([]);var n=this._consumeRawText(!1,Ma,function(){return e._attemptStr("->")});this._beginToken(vu.COMMENT_END,n.sourceSpan.end),this._endToken([])},t.prototype._consumeCdata=function(t){var e=this;this._beginToken(vu.CDATA_START,t),this._requireStr("CDATA["),this._endToken([]);var n=this._consumeRawText(!1,93,function(){return e._attemptStr("]>")});this._beginToken(vu.CDATA_END,n.sourceSpan.end),this._endToken([])},t.prototype._consumeDocType=function(t){this._beginToken(vu.DOC_TYPE,t),this._attemptUntilChar(Da),this._advance(),this._endToken([this._input.substring(t.offset+2,this._index-1)])},t.prototype._consumePrefixAndName=function(){for(var t,e,n=this._index,r=null;58!==this._peek&&!(((t=this._peek)<Va||Ya<t)&&(t<Ra||La<t)&&(t<Ia||t>Pa));)this._advance();return 58===this._peek?(this._advance(),r=this._input.substring(n,this._index-1),e=this._index):e=n,this._requireCharCodeUntilFn(Du,this._index===e?1:0),[r,this._input.substring(e,this._index)]},t.prototype._consumeTagOpen=function(t){var e,n,i=this._savePosition();try{if(!Ja(this._peek))throw this._createError(Su(this._peek),this._getSpan());var o=this._index;for(this._consumeTagOpenStart(t),n=(e=this._input.substring(o,this._index)).toLowerCase(),this._attemptCharCodeUntilFn(Au);this._peek!==Ta&&this._peek!==Da;)this._consumeAttributeName(),this._attemptCharCodeUntilFn(Au),this._attemptCharCode(Aa)&&(this._attemptCharCodeUntilFn(Au),this._consumeAttributeValue()),this._attemptCharCodeUntilFn(Au);this._consumeTagOpenEnd()}catch(e){if(e instanceof Tu)return this._restorePosition(i),this._beginToken(vu.TEXT,t),void this._endToken(["<"]);throw e}var a=this._getTagDefinition(e).contentType;a===r.RAW_TEXT?this._consumeRawTextWithTagClose(n,!1):a===r.ESCAPABLE_RAW_TEXT&&this._consumeRawTextWithTagClose(n,!0)},t.prototype._consumeRawTextWithTagClose=function(t,e){var n=this,r=this._consumeRawText(e,60,function(){return!!n._attemptCharCode(Ta)&&(n._attemptCharCodeUntilFn(Au),!!n._attemptStrCaseInsensitive(t)&&(n._attemptCharCodeUntilFn(Au),n._attemptCharCode(Da)))});this._beginToken(vu.TAG_CLOSE,r.sourceSpan.end),this._endToken([null,t])},t.prototype._consumeTagOpenStart=function(t){this._beginToken(vu.TAG_OPEN_START,t);var e=this._consumePrefixAndName();this._endToken(e)},t.prototype._consumeAttributeName=function(){this._beginToken(vu.ATTR_NAME);var t=this._consumePrefixAndName();this._endToken(t)},t.prototype._consumeAttributeValue=function(){var t;if(this._beginToken(vu.ATTR_VALUE),this._peek===Ea||this._peek===Oa){var e=this._peek;this._advance();for(var n=[];this._peek!==e;)n.push(this._readChar(!0));t=n.join(""),this._advance()}else{var r=this._index;this._requireCharCodeUntilFn(Du,1),t=this._input.substring(r,this._index)}this._endToken([this._processCarriageReturns(t)])},t.prototype._consumeTagOpenEnd=function(){var t=this._attemptCharCode(Ta)?vu.TAG_OPEN_END_VOID:vu.TAG_OPEN_END;this._beginToken(t),this._requireCharCode(Da),this._endToken([])},t.prototype._consumeTagClose=function(t){this._beginToken(vu.TAG_CLOSE,t),this._attemptCharCodeUntilFn(Au);var e=this._consumePrefixAndName();this._attemptCharCodeUntilFn(Au),this._requireCharCode(Da),this._endToken(e)},t.prototype._consumeExpansionFormStart=function(){this._beginToken(vu.EXPANSION_FORM_START,this._getLocation()),this._requireCharCode(Ka),this._endToken([]),this._expansionCaseStack.push(vu.EXPANSION_FORM_START),this._beginToken(vu.RAW_TEXT,this._getLocation());var t=this._readUntil(44);this._endToken([t],this._getLocation()),this._requireCharCode(44),this._attemptCharCodeUntilFn(Au),this._beginToken(vu.RAW_TEXT,this._getLocation());var e=this._readUntil(44);this._endToken([e],this._getLocation()),this._requireCharCode(44),this._attemptCharCodeUntilFn(Au)},t.prototype._consumeExpansionCaseStart=function(){this._beginToken(vu.EXPANSION_CASE_VALUE,this._getLocation());var t=this._readUntil(Ka).trim();this._endToken([t],this._getLocation()),this._attemptCharCodeUntilFn(Au),this._beginToken(vu.EXPANSION_CASE_EXP_START,this._getLocation()),this._requireCharCode(Ka),this._endToken([],this._getLocation()),this._attemptCharCodeUntilFn(Au),this._expansionCaseStack.push(vu.EXPANSION_CASE_EXP_START)},t.prototype._consumeExpansionCaseEnd=function(){this._beginToken(vu.EXPANSION_CASE_EXP_END,this._getLocation()),this._requireCharCode(125),this._endToken([],this._getLocation()),this._attemptCharCodeUntilFn(Au),this._expansionCaseStack.pop()},t.prototype._consumeExpansionFormEnd=function(){this._beginToken(vu.EXPANSION_FORM_END,this._getLocation()),this._requireCharCode(125),this._endToken([]),this._expansionCaseStack.pop()},t.prototype._consumeText=function(){var t=this._getLocation();this._beginToken(vu.TEXT,t);var e=[];do{this._interpolationConfig&&this._attemptStr(this._interpolationConfig.start)?(e.push(this._interpolationConfig.start),this._inInterpolation=!0):this._interpolationConfig&&this._inInterpolation&&this._attemptStr(this._interpolationConfig.end)?(e.push(this._interpolationConfig.end),this._inInterpolation=!1):e.push(this._readChar(!0))}while(!this._isTextEnd());this._endToken([this._processCarriageReturns(e.join(""))])},t.prototype._isTextEnd=function(){if(60===this._peek||this._peek===ya)return!0;if(this._tokenizeIcu&&!this._inInterpolation){if(Ru(this._input,this._index,this._interpolationConfig))return!0;if(125===this._peek&&this._isInExpansionCase())return!0}return!1},t.prototype._savePosition=function(){return[this._peek,this._index,this._column,this._line,this.tokens.length]},t.prototype._readUntil=function(t){var e=this._index;return this._attemptUntilChar(t),this._input.substring(e,this._index)},t.prototype._restorePosition=function(t){this._peek=t[0],this._index=t[1],this._column=t[2],this._line=t[3];var e=t[4];e<this.tokens.length&&(this.tokens=this.tokens.slice(0,e))},t.prototype._isInExpansionCase=function(){return this._expansionCaseStack.length>0&&this._expansionCaseStack[this._expansionCaseStack.length-1]===vu.EXPANSION_CASE_EXP_START},t.prototype._isInExpansionForm=function(){return this._expansionCaseStack.length>0&&this._expansionCaseStack[this._expansionCaseStack.length-1]===vu.EXPANSION_FORM_START},t}();function Au(t){return!Za(t)||t===ya}function Du(t){return Za(t)||t===Da||t===Ta||t===Ea||t===Oa||t===Aa}function Iu(t){return t==ka||t==ya||!((e=t)>=Va&&e<=Ha||e>=Ra&&e<=Fa||$a(e));var e}function Pu(t){return t==ka||t==ya||!Ja(t)}function Ru(t,e,n){var r=!!n&&t.indexOf(n.start,e)==e;return t.charCodeAt(e)==Ka&&!r}function Nu(t){return t>=Va&&t<=Ya?t-Va+Ra:t}var Fu=function(t){function e(e,n,r){var i=t.call(this,n,r)||this;return i.elementName=e,i}return Object(o.__extends)(e,t),e.create=function(t,n,r){return new e(t,n,r)},e}(os),Lu=function(){return function(t,e){this.rootNodes=t,this.errors=e}}(),Bu=function(){function t(t){this.getTagDefinition=t}return t.prototype.parse=function(t,e,n,r){void 0===n&&(n=!1),void 0===r&&(r=Hi);var i,o,a,s,u,c=(i=t,o=e,a=this.getTagDefinition,u=r,void 0===(s=n)&&(s=!1),void 0===u&&(u=Hi),new ku(new rs(i,o),a,s,u).tokenize()),l=new Vu(c.tokens,this.getTagDefinition).build();return new Lu(l.rootNodes,c.errors.concat(l.errors))},t}(),Vu=function(){function t(t,e){this.tokens=t,this.getTagDefinition=e,this._index=-1,this._rootNodes=[],this._errors=[],this._elementStack=[],this._advance()}return t.prototype.build=function(){for(;this._peek.type!==vu.EOF;)this._peek.type===vu.TAG_OPEN_START?this._consumeStartTag(this._advance()):this._peek.type===vu.TAG_CLOSE?this._consumeEndTag(this._advance()):this._peek.type===vu.CDATA_START?(this._closeVoidElement(),this._consumeCdata(this._advance())):this._peek.type===vu.COMMENT_START?(this._closeVoidElement(),this._consumeComment(this._advance())):this._peek.type===vu.TEXT||this._peek.type===vu.RAW_TEXT||this._peek.type===vu.ESCAPABLE_RAW_TEXT?(this._closeVoidElement(),this._consumeText(this._advance())):this._peek.type===vu.EXPANSION_FORM_START?this._consumeExpansion(this._advance()):this._advance();return new Lu(this._rootNodes,this._errors)},t.prototype._advance=function(){var t=this._peek;return this._index<this.tokens.length-1&&this._index++,this._peek=this.tokens[this._index],t},t.prototype._advanceIf=function(t){return this._peek.type===t?this._advance():null},t.prototype._consumeCdata=function(t){this._consumeText(this._advance()),this._advanceIf(vu.CDATA_END)},t.prototype._consumeComment=function(t){var e=this._advanceIf(vu.RAW_TEXT);this._advanceIf(vu.COMMENT_END);var n=null!=e?e.parts[0].trim():null;this._addToParent(new yu(n,t.sourceSpan))},t.prototype._consumeExpansion=function(t){for(var e=this._advance(),n=this._advance(),r=[];this._peek.type===vu.EXPANSION_CASE_VALUE;){var i=this._parseExpansionCase();if(!i)return;r.push(i)}if(this._peek.type===vu.EXPANSION_FORM_END){var o=new is(t.sourceSpan.start,this._peek.sourceSpan.end);this._addToParent(new fu(e.parts[0],n.parts[0],r,o,e.sourceSpan)),this._advance()}else this._errors.push(Fu.create(null,this._peek.sourceSpan,"Invalid ICU message. Missing '}'."))},t.prototype._parseExpansionCase=function(){var e=this._advance();if(this._peek.type!==vu.EXPANSION_CASE_EXP_START)return this._errors.push(Fu.create(null,this._peek.sourceSpan,"Invalid ICU message. Missing '{'.")),null;var n=this._advance(),r=this._collectExpansionExpTokens(n);if(!r)return null;var i=this._advance();r.push(new ju(vu.EOF,[],i.sourceSpan));var o=new t(r,this.getTagDefinition).build();if(o.errors.length>0)return this._errors=this._errors.concat(o.errors),null;var a=new is(e.sourceSpan.start,i.sourceSpan.end),s=new is(n.sourceSpan.start,i.sourceSpan.end);return new hu(e.parts[0],o.rootNodes,a,e.sourceSpan,s)},t.prototype._collectExpansionExpTokens=function(t){for(var e=[],n=[vu.EXPANSION_CASE_EXP_START];;){if(this._peek.type!==vu.EXPANSION_FORM_START&&this._peek.type!==vu.EXPANSION_CASE_EXP_START||n.push(this._peek.type),this._peek.type===vu.EXPANSION_CASE_EXP_END){if(!zu(n,vu.EXPANSION_CASE_EXP_START))return this._errors.push(Fu.create(null,t.sourceSpan,"Invalid ICU message. Missing '}'.")),null;if(n.pop(),0==n.length)return e}if(this._peek.type===vu.EXPANSION_FORM_END){if(!zu(n,vu.EXPANSION_FORM_START))return this._errors.push(Fu.create(null,t.sourceSpan,"Invalid ICU message. Missing '}'.")),null;n.pop()}if(this._peek.type===vu.EOF)return this._errors.push(Fu.create(null,t.sourceSpan,"Invalid ICU message. Missing '}'.")),null;e.push(this._advance())}},t.prototype._consumeText=function(t){var e=t.parts[0];if(e.length>0&&"\n"==e[0]){var n=this._getParentElement();null!=n&&0==n.children.length&&this.getTagDefinition(n.name).ignoreFirstLf&&(e=e.substring(1))}e.length>0&&this._addToParent(new pu(e,t.sourceSpan))},t.prototype._closeVoidElement=function(){var t=this._getParentElement();t&&this.getTagDefinition(t.name).isVoid&&this._elementStack.pop()},t.prototype._consumeStartTag=function(t){for(var e=t.parts[0],n=t.parts[1],r=[];this._peek.type===vu.ATTR_NAME;)r.push(this._consumeAttr(this._advance()));var i=this._getElementFullName(e,n,this._getParentElement()),o=!1;if(this._peek.type===vu.TAG_OPEN_END_VOID){this._advance(),o=!0;var a=this.getTagDefinition(i);a.canSelfClose||null!==l(i)||a.isVoid||this._errors.push(Fu.create(i,t.sourceSpan,'Only void and foreign elements can be self closed "'+t.parts[1]+'"'))}else this._peek.type===vu.TAG_OPEN_END&&(this._advance(),o=!1);var s=this._peek.sourceSpan.start,u=new is(t.sourceSpan.start,s),c=new gu(i,r,[],u,u,void 0);this._pushElement(c),o&&(this._popElement(i),c.endSourceSpan=u)},t.prototype._pushElement=function(t){var e=this._getParentElement();e&&this.getTagDefinition(e.name).isClosedByChild(t.name)&&this._elementStack.pop();var n=this.getTagDefinition(t.name),r=this._getParentElementSkippingContainers(),i=r.parent,o=r.container;if(i&&n.requireExtraParent(i.name)){var a=new gu(n.parentToAdd,[],[],t.sourceSpan,t.startSourceSpan,t.endSourceSpan);this._insertBeforeContainer(i,o,a)}this._addToParent(t),this._elementStack.push(t)},t.prototype._consumeEndTag=function(t){var e=this._getElementFullName(t.parts[0],t.parts[1],this._getParentElement());if(this._getParentElement()&&(this._getParentElement().endSourceSpan=t.sourceSpan),this.getTagDefinition(e).isVoid)this._errors.push(Fu.create(e,t.sourceSpan,'Void elements do not have end tags "'+t.parts[1]+'"'));else if(!this._popElement(e)){var n='Unexpected closing tag "'+e+'". It may happen when the tag has already been closed by another tag. For more info see https://www.w3.org/TR/html5/syntax.html#closing-elements-that-have-implied-end-tags';this._errors.push(Fu.create(e,t.sourceSpan,n))}},t.prototype._popElement=function(t){for(var e=this._elementStack.length-1;e>=0;e--){var n=this._elementStack[e];if(n.name==t)return this._elementStack.splice(e,this._elementStack.length-e),!0;if(!this.getTagDefinition(n.name).closedByParent)return!1}return!1},t.prototype._consumeAttr=function(t){var e=d(t.parts[0],t.parts[1]),n=t.sourceSpan.end,r="",i=void 0;if(this._peek.type===vu.ATTR_VALUE){var o=this._advance();r=o.parts[0],n=o.sourceSpan.end,i=o.sourceSpan}return new mu(e,r,new is(t.sourceSpan.start,n),i)},t.prototype._getParentElement=function(){return this._elementStack.length>0?this._elementStack[this._elementStack.length-1]:null},t.prototype._getParentElementSkippingContainers=function(){for(var t=null,e=this._elementStack.length-1;e>=0;e--){if(!s(this._elementStack[e].name))return{parent:this._elementStack[e],container:t};t=this._elementStack[e]}return{parent:null,container:t}},t.prototype._addToParent=function(t){var e=this._getParentElement();null!=e?e.children.push(t):this._rootNodes.push(t)},t.prototype._insertBeforeContainer=function(t,e,n){if(e){if(t){var r=t.children.indexOf(e);t.children[r]=n}else this._rootNodes.push(n);n.children.push(e),this._elementStack.splice(this._elementStack.indexOf(e),0,n)}else this._addToParent(n),this._elementStack.push(n)},t.prototype._getElementFullName=function(t,e,n){return null==t&&null==(t=this.getTagDefinition(e).implicitNamespacePrefix)&&null!=n&&(t=l(n.name)),d(t,e)},t}();function zu(t,e){return t.length>0&&t[t.length-1]===e}var Hu=function(t){function e(){return t.call(this,y)||this}return Object(o.__extends)(e,t),e.prototype.parse=function(e,n,r,i){return void 0===r&&(r=!1),void 0===i&&(i=Hi),t.prototype.parse.call(this,e,n,r,i)},e}(Bu),Uu="ngPreserveWhitespaces",Wu=new Set(["pre","template","textarea","script","style"]),qu=" \f\n\r\t\v ᠎ - \u2028\u2029  　\ufeff",Gu=new RegExp("[^"+qu+"]"),Yu=new RegExp("["+qu+"]{2,}","g");function Ku(t){return t.replace(new RegExp(f,"g")," ")}var Xu,Qu=function(){function t(){}return t.prototype.visitElement=function(t,e){return Wu.has(t.name)||t.attrs.some(function(t){return t.name===Uu})?new gu(t.name,_u(this,t.attrs),t.children,t.sourceSpan,t.startSourceSpan,t.endSourceSpan,t.i18n):new gu(t.name,t.attrs,_u(this,t.children),t.sourceSpan,t.startSourceSpan,t.endSourceSpan,t.i18n)},t.prototype.visitAttribute=function(t,e){return t.name!==Uu?t:null},t.prototype.visitText=function(t,e){return t.value.match(Gu)?new pu(Ku(t.value).replace(Yu," "),t.sourceSpan,t.i18n):null},t.prototype.visitComment=function(t,e){return t},t.prototype.visitExpansion=function(t,e){return t},t.prototype.visitExpansionCase=function(t,e){return t},t}();function Zu(t){return new Lu(_u(new Qu,t.rootNodes),t.errors)}function $u(){return Xu||(Xu={},Ju(R.HTML,["iframe|srcdoc","*|innerHTML","*|outerHTML"]),Ju(R.STYLE,["*|style"]),Ju(R.URL,["*|formAction","area|href","area|ping","audio|src","a|href","a|ping","blockquote|cite","body|background","del|cite","form|action","img|src","img|srcset","input|src","ins|cite","q|cite","source|src","source|srcset","track|src","video|poster","video|src"]),Ju(R.RESOURCE_URL,["applet|code","applet|codebase","base|href","embed|src","frame|src","head|profile","html|manifest","iframe|src","link|href","media|src","object|codebase","object|data","script|src"])),Xu}function Ju(t,e){var n,r;try{for(var i=Object(o.__values)(e),a=i.next();!a.done;a=i.next()){var s=a.value;Xu[s.toLowerCase()]=t}}catch(t){n={error:t}}finally{try{a&&!a.done&&(r=i.return)&&r.call(i)}finally{if(n)throw n.error}}}var tc=function(){return function(){}}(),ec="boolean",nc="number",rc="string",ic="object",oc=["[Element]|textContent,%classList,className,id,innerHTML,*beforecopy,*beforecut,*beforepaste,*copy,*cut,*paste,*search,*selectstart,*webkitfullscreenchange,*webkitfullscreenerror,*wheel,outerHTML,#scrollLeft,#scrollTop,slot,*message,*mozfullscreenchange,*mozfullscreenerror,*mozpointerlockchange,*mozpointerlockerror,*webglcontextcreationerror,*webglcontextlost,*webglcontextrestored","[HTMLElement]^[Element]|accessKey,contentEditable,dir,!draggable,!hidden,innerText,lang,*abort,*auxclick,*blur,*cancel,*canplay,*canplaythrough,*change,*click,*close,*contextmenu,*cuechange,*dblclick,*drag,*dragend,*dragenter,*dragleave,*dragover,*dragstart,*drop,*durationchange,*emptied,*ended,*error,*focus,*gotpointercapture,*input,*invalid,*keydown,*keypress,*keyup,*load,*loadeddata,*loadedmetadata,*loadstart,*lostpointercapture,*mousedown,*mouseenter,*mouseleave,*mousemove,*mouseout,*mouseover,*mouseup,*mousewheel,*pause,*play,*playing,*pointercancel,*pointerdown,*pointerenter,*pointerleave,*pointermove,*pointerout,*pointerover,*pointerup,*progress,*ratechange,*reset,*resize,*scroll,*seeked,*seeking,*select,*show,*stalled,*submit,*suspend,*timeupdate,*toggle,*volumechange,*waiting,outerText,!spellcheck,%style,#tabIndex,title,!translate","abbr,address,article,aside,b,bdi,bdo,cite,code,dd,dfn,dt,em,figcaption,figure,footer,header,i,kbd,main,mark,nav,noscript,rb,rp,rt,rtc,ruby,s,samp,section,small,strong,sub,sup,u,var,wbr^[HTMLElement]|accessKey,contentEditable,dir,!draggable,!hidden,innerText,lang,*abort,*auxclick,*blur,*cancel,*canplay,*canplaythrough,*change,*click,*close,*contextmenu,*cuechange,*dblclick,*drag,*dragend,*dragenter,*dragleave,*dragover,*dragstart,*drop,*durationchange,*emptied,*ended,*error,*focus,*gotpointercapture,*input,*invalid,*keydown,*keypress,*keyup,*load,*loadeddata,*loadedmetadata,*loadstart,*lostpointercapture,*mousedown,*mouseenter,*mouseleave,*mousemove,*mouseout,*mouseover,*mouseup,*mousewheel,*pause,*play,*playing,*pointercancel,*pointerdown,*pointerenter,*pointerleave,*pointermove,*pointerout,*pointerover,*pointerup,*progress,*ratechange,*reset,*resize,*scroll,*seeked,*seeking,*select,*show,*stalled,*submit,*suspend,*timeupdate,*toggle,*volumechange,*waiting,outerText,!spellcheck,%style,#tabIndex,title,!translate","media^[HTMLElement]|!autoplay,!controls,%controlsList,%crossOrigin,#currentTime,!defaultMuted,#defaultPlaybackRate,!disableRemotePlayback,!loop,!muted,*encrypted,*waitingforkey,#playbackRate,preload,src,%srcObject,#volume",":svg:^[HTMLElement]|*abort,*auxclick,*blur,*cancel,*canplay,*canplaythrough,*change,*click,*close,*contextmenu,*cuechange,*dblclick,*drag,*dragend,*dragenter,*dragleave,*dragover,*dragstart,*drop,*durationchange,*emptied,*ended,*error,*focus,*gotpointercapture,*input,*invalid,*keydown,*keypress,*keyup,*load,*loadeddata,*loadedmetadata,*loadstart,*lostpointercapture,*mousedown,*mouseenter,*mouseleave,*mousemove,*mouseout,*mouseover,*mouseup,*mousewheel,*pause,*play,*playing,*pointercancel,*pointerdown,*pointerenter,*pointerleave,*pointermove,*pointerout,*pointerover,*pointerup,*progress,*ratechange,*reset,*resize,*scroll,*seeked,*seeking,*select,*show,*stalled,*submit,*suspend,*timeupdate,*toggle,*volumechange,*waiting,%style,#tabIndex",":svg:graphics^:svg:|",":svg:animation^:svg:|*begin,*end,*repeat",":svg:geometry^:svg:|",":svg:componentTransferFunction^:svg:|",":svg:gradient^:svg:|",":svg:textContent^:svg:graphics|",":svg:textPositioning^:svg:textContent|","a^[HTMLElement]|charset,coords,download,hash,host,hostname,href,hreflang,name,password,pathname,ping,port,protocol,referrerPolicy,rel,rev,search,shape,target,text,type,username","area^[HTMLElement]|alt,coords,download,hash,host,hostname,href,!noHref,password,pathname,ping,port,protocol,referrerPolicy,rel,search,shape,target,username","audio^media|","br^[HTMLElement]|clear","base^[HTMLElement]|href,target","body^[HTMLElement]|aLink,background,bgColor,link,*beforeunload,*blur,*error,*focus,*hashchange,*languagechange,*load,*message,*offline,*online,*pagehide,*pageshow,*popstate,*rejectionhandled,*resize,*scroll,*storage,*unhandledrejection,*unload,text,vLink","button^[HTMLElement]|!autofocus,!disabled,formAction,formEnctype,formMethod,!formNoValidate,formTarget,name,type,value","canvas^[HTMLElement]|#height,#width","content^[HTMLElement]|select","dl^[HTMLElement]|!compact","datalist^[HTMLElement]|","details^[HTMLElement]|!open","dialog^[HTMLElement]|!open,returnValue","dir^[HTMLElement]|!compact","div^[HTMLElement]|align","embed^[HTMLElement]|align,height,name,src,type,width","fieldset^[HTMLElement]|!disabled,name","font^[HTMLElement]|color,face,size","form^[HTMLElement]|acceptCharset,action,autocomplete,encoding,enctype,method,name,!noValidate,target","frame^[HTMLElement]|frameBorder,longDesc,marginHeight,marginWidth,name,!noResize,scrolling,src","frameset^[HTMLElement]|cols,*beforeunload,*blur,*error,*focus,*hashchange,*languagechange,*load,*message,*offline,*online,*pagehide,*pageshow,*popstate,*rejectionhandled,*resize,*scroll,*storage,*unhandledrejection,*unload,rows","hr^[HTMLElement]|align,color,!noShade,size,width","head^[HTMLElement]|","h1,h2,h3,h4,h5,h6^[HTMLElement]|align","html^[HTMLElement]|version","iframe^[HTMLElement]|align,!allowFullscreen,frameBorder,height,longDesc,marginHeight,marginWidth,name,referrerPolicy,%sandbox,scrolling,src,srcdoc,width","img^[HTMLElement]|align,alt,border,%crossOrigin,#height,#hspace,!isMap,longDesc,lowsrc,name,referrerPolicy,sizes,src,srcset,useMap,#vspace,#width","input^[HTMLElement]|accept,align,alt,autocapitalize,autocomplete,!autofocus,!checked,!defaultChecked,defaultValue,dirName,!disabled,%files,formAction,formEnctype,formMethod,!formNoValidate,formTarget,#height,!incremental,!indeterminate,max,#maxLength,min,#minLength,!multiple,name,pattern,placeholder,!readOnly,!required,selectionDirection,#selectionEnd,#selectionStart,#size,src,step,type,useMap,value,%valueAsDate,#valueAsNumber,#width","li^[HTMLElement]|type,#value","label^[HTMLElement]|htmlFor","legend^[HTMLElement]|align","link^[HTMLElement]|as,charset,%crossOrigin,!disabled,href,hreflang,integrity,media,referrerPolicy,rel,%relList,rev,%sizes,target,type","map^[HTMLElement]|name","marquee^[HTMLElement]|behavior,bgColor,direction,height,#hspace,#loop,#scrollAmount,#scrollDelay,!trueSpeed,#vspace,width","menu^[HTMLElement]|!compact","meta^[HTMLElement]|content,httpEquiv,name,scheme","meter^[HTMLElement]|#high,#low,#max,#min,#optimum,#value","ins,del^[HTMLElement]|cite,dateTime","ol^[HTMLElement]|!compact,!reversed,#start,type","object^[HTMLElement]|align,archive,border,code,codeBase,codeType,data,!declare,height,#hspace,name,standby,type,useMap,#vspace,width","optgroup^[HTMLElement]|!disabled,label","option^[HTMLElement]|!defaultSelected,!disabled,label,!selected,text,value","output^[HTMLElement]|defaultValue,%htmlFor,name,value","p^[HTMLElement]|align","param^[HTMLElement]|name,type,value,valueType","picture^[HTMLElement]|","pre^[HTMLElement]|#width","progress^[HTMLElement]|#max,#value","q,blockquote,cite^[HTMLElement]|","script^[HTMLElement]|!async,charset,%crossOrigin,!defer,event,htmlFor,integrity,src,text,type","select^[HTMLElement]|!autofocus,!disabled,#length,!multiple,name,!required,#selectedIndex,#size,value","shadow^[HTMLElement]|","slot^[HTMLElement]|name","source^[HTMLElement]|media,sizes,src,srcset,type","span^[HTMLElement]|","style^[HTMLElement]|!disabled,media,type","caption^[HTMLElement]|align","th,td^[HTMLElement]|abbr,align,axis,bgColor,ch,chOff,#colSpan,headers,height,!noWrap,#rowSpan,scope,vAlign,width","col,colgroup^[HTMLElement]|align,ch,chOff,#span,vAlign,width","table^[HTMLElement]|align,bgColor,border,%caption,cellPadding,cellSpacing,frame,rules,summary,%tFoot,%tHead,width","tr^[HTMLElement]|align,bgColor,ch,chOff,vAlign","tfoot,thead,tbody^[HTMLElement]|align,ch,chOff,vAlign","template^[HTMLElement]|","textarea^[HTMLElement]|autocapitalize,!autofocus,#cols,defaultValue,dirName,!disabled,#maxLength,#minLength,name,placeholder,!readOnly,!required,#rows,selectionDirection,#selectionEnd,#selectionStart,value,wrap","title^[HTMLElement]|text","track^[HTMLElement]|!default,kind,label,src,srclang","ul^[HTMLElement]|!compact,type","unknown^[HTMLElement]|","video^media|#height,poster,#width",":svg:a^:svg:graphics|",":svg:animate^:svg:animation|",":svg:animateMotion^:svg:animation|",":svg:animateTransform^:svg:animation|",":svg:circle^:svg:geometry|",":svg:clipPath^:svg:graphics|",":svg:defs^:svg:graphics|",":svg:desc^:svg:|",":svg:discard^:svg:|",":svg:ellipse^:svg:geometry|",":svg:feBlend^:svg:|",":svg:feColorMatrix^:svg:|",":svg:feComponentTransfer^:svg:|",":svg:feComposite^:svg:|",":svg:feConvolveMatrix^:svg:|",":svg:feDiffuseLighting^:svg:|",":svg:feDisplacementMap^:svg:|",":svg:feDistantLight^:svg:|",":svg:feDropShadow^:svg:|",":svg:feFlood^:svg:|",":svg:feFuncA^:svg:componentTransferFunction|",":svg:feFuncB^:svg:componentTransferFunction|",":svg:feFuncG^:svg:componentTransferFunction|",":svg:feFuncR^:svg:componentTransferFunction|",":svg:feGaussianBlur^:svg:|",":svg:feImage^:svg:|",":svg:feMerge^:svg:|",":svg:feMergeNode^:svg:|",":svg:feMorphology^:svg:|",":svg:feOffset^:svg:|",":svg:fePointLight^:svg:|",":svg:feSpecularLighting^:svg:|",":svg:feSpotLight^:svg:|",":svg:feTile^:svg:|",":svg:feTurbulence^:svg:|",":svg:filter^:svg:|",":svg:foreignObject^:svg:graphics|",":svg:g^:svg:graphics|",":svg:image^:svg:graphics|",":svg:line^:svg:geometry|",":svg:linearGradient^:svg:gradient|",":svg:mpath^:svg:|",":svg:marker^:svg:|",":svg:mask^:svg:|",":svg:metadata^:svg:|",":svg:path^:svg:geometry|",":svg:pattern^:svg:|",":svg:polygon^:svg:geometry|",":svg:polyline^:svg:geometry|",":svg:radialGradient^:svg:gradient|",":svg:rect^:svg:geometry|",":svg:svg^:svg:graphics|#currentScale,#zoomAndPan",":svg:script^:svg:|type",":svg:set^:svg:animation|",":svg:stop^:svg:|",":svg:style^:svg:|!disabled,media,title,type",":svg:switch^:svg:graphics|",":svg:symbol^:svg:|",":svg:tspan^:svg:textPositioning|",":svg:text^:svg:textPositioning|",":svg:textPath^:svg:textContent|",":svg:title^:svg:|",":svg:use^:svg:graphics|",":svg:view^:svg:|#zoomAndPan","data^[HTMLElement]|value","keygen^[HTMLElement]|!autofocus,challenge,!disabled,form,keytype,name","menuitem^[HTMLElement]|type,label,icon,!disabled,!checked,radiogroup,!default","summary^[HTMLElement]|","time^[HTMLElement]|dateTime",":svg:cursor^:svg:|"],ac={class:"className",for:"htmlFor",formaction:"formAction",innerHtml:"innerHTML",readonly:"readOnly",tabindex:"tabIndex"},sc=function(t){function e(){var e=t.call(this)||this;return e._schema={},oc.forEach(function(t){var n={},r=Object(o.__read)(t.split("|"),2),i=r[0],a=r[1].split(","),s=Object(o.__read)(i.split("^"),2),u=s[0],c=s[1];u.split(",").forEach(function(t){return e._schema[t.toLowerCase()]=n});var l=c&&e._schema[c.toLowerCase()];l&&Object.keys(l).forEach(function(t){n[t]=l[t]}),a.forEach(function(t){if(t.length>0)switch(t[0]){case"*":break;case"!":n[t.substring(1)]=ec;break;case"#":n[t.substring(1)]=nc;break;case"%":n[t.substring(1)]=ic;break;default:n[t]=rc}})}),e}return Object(o.__extends)(e,t),e.prototype.hasProperty=function(t,e,n){if(n.some(function(t){return t.name===K.name}))return!0;if(t.indexOf("-")>-1){if(s(t)||u(t))return!1;if(n.some(function(t){return t.name===Y.name}))return!0}return!!(this._schema[t.toLowerCase()]||this._schema.unknown)[e]},e.prototype.hasElement=function(t,e){if(e.some(function(t){return t.name===K.name}))return!0;if(t.indexOf("-")>-1){if(s(t)||u(t))return!0;if(e.some(function(t){return t.name===Y.name}))return!0}return!!this._schema[t.toLowerCase()]},e.prototype.securityContext=function(t,e,n){n&&(e=this.getMappedPropName(e)),t=t.toLowerCase(),e=e.toLowerCase();var r=$u()[t+"|"+e];return r||((r=$u()["*|"+e])||R.NONE)},e.prototype.getMappedPropName=function(t){return ac[t]||t},e.prototype.getDefaultComponentElementName=function(){return"ng-component"},e.prototype.validateProperty=function(t){return t.toLowerCase().startsWith("on")?{error:!0,msg:"Binding to event property '"+t+"' is disallowed for security reasons, please use ("+t.slice(2)+")=...\nIf '"+t+"' is a directive input, make sure the directive is imported by the current module."}:{error:!1}},e.prototype.validateAttribute=function(t){return t.toLowerCase().startsWith("on")?{error:!0,msg:"Binding to event attribute '"+t+"' is disallowed for security reasons, please use ("+t.slice(2)+")=..."}:{error:!1}},e.prototype.allKnownElementNames=function(){return Object.keys(this._schema)},e.prototype.normalizeAnimationStyleProperty=function(t){return t.replace(Ne,function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return t[1].toUpperCase()})},e.prototype.normalizeAnimationStyleValue=function(t,e,n){var r="",i=n.toString().trim(),o=null;if(function(t){switch(t){case"width":case"height":case"minWidth":case"minHeight":case"maxWidth":case"maxHeight":case"left":case"top":case"bottom":case"right":case"fontSize":case"outlineWidth":case"outlineOffset":case"paddingTop":case"paddingLeft":case"paddingBottom":case"paddingRight":case"marginTop":case"marginLeft":case"marginBottom":case"marginRight":case"borderRadius":case"borderWidth":case"borderTopWidth":case"borderLeftWidth":case"borderRightWidth":case"borderBottomWidth":case"textIndent":return!0;default:return!1}}(t)&&0!==n&&"0"!==n)if("number"==typeof n)r="px";else{var a=n.match(/^[+-]?[\d\.]+([a-z]*)$/);a&&0==a[1].length&&(o="Please provide a CSS unit value for "+e+":"+n)}return{error:o,value:i+r}},e}(tc);var uc=function(){function t(t,e,n,r,i){if(this._exprParser=t,this._interpolationConfig=e,this._schemaRegistry=n,this.errors=i,this.pipesByName=null,this._usedPipes=new Map,r){var o=new Map;r.forEach(function(t){return o.set(t.name,t)}),this.pipesByName=o}}return Object.defineProperty(t.prototype,"interpolationConfig",{get:function(){return this._interpolationConfig},enumerable:!0,configurable:!0}),t.prototype.getUsedPipes=function(){return Array.from(this._usedPipes.values())},t.prototype.createBoundHostProperties=function(t,e){var n=this;if(t.hostProperties){var r=[];return Object.keys(t.hostProperties).forEach(function(i){var o=t.hostProperties[i];"string"==typeof o?n.parsePropertyBinding(i,o,!0,e,[],r):n._reportError('Value of the host property binding "'+i+'" needs to be a string representing an expression but got "'+o+'" ('+typeof o+")",e)}),r}return null},t.prototype.createDirectiveHostPropertyAsts=function(t,e,n){var r=this,i=this.createBoundHostProperties(t,n);return i&&i.map(function(t){return r.createBoundElementProperty(e,t)})},t.prototype.createDirectiveHostEventAsts=function(t,e){var n=this;if(t.hostListeners){var r=[];return Object.keys(t.hostListeners).forEach(function(i){var o=t.hostListeners[i];"string"==typeof o?n.parseEvent(i,o,e,[],r):n._reportError('Value of the host listener "'+i+'" needs to be a string representing an expression but got "'+o+'" ('+typeof o+")",e)}),r}return null},t.prototype.parseInterpolation=function(t,e){var n=e.start.toString();try{var r=this._exprParser.parseInterpolation(t,n,this._interpolationConfig);return r&&this._reportExpressionParserErrors(r.errors,e),this._checkPipes(r,e),r}catch(t){return this._reportError(""+t,e),this._exprParser.wrapLiteralPrimitive("ERROR",n)}},t.prototype.parseInlineTemplateBinding=function(t,e,n,r,i,o){for(var a=this._parseTemplateBindings(t,e,n),s=0;s<a.length;s++){var u=a[s];u.keyIsVar?o.push(new Yo(u.key,u.name,n)):u.expression?this._parsePropertyAst(u.key,u.expression,n,r,i):(r.push([u.key,""]),this.parseLiteralAttr(u.key,null,n,r,i))}},t.prototype._parseTemplateBindings=function(t,e,n){var r=this,i=n.start.toString();try{var o=this._exprParser.parseTemplateBindings(t,e,i);return this._reportExpressionParserErrors(o.errors,n),o.templateBindings.forEach(function(t){t.expression&&r._checkPipes(t.expression,n)}),o.warnings.forEach(function(t){r._reportError(t,n,ts.WARNING)}),o.templateBindings}catch(t){return this._reportError(""+t,n),[]}},t.prototype.parseLiteralAttr=function(t,e,n,r,i){lc(t)?(t=t.substring(1),e&&this._reportError('Assigning animation triggers via @prop="exp" attributes with an expression is invalid. Use property bindings (e.g. [@prop]="exp") or use an attribute without a value (e.g. @prop) instead.',n,ts.ERROR),this._parseAnimation(t,e,n,r,i)):i.push(new qo(t,this._exprParser.wrapLiteralPrimitive(e,""),Uo.LITERAL_ATTR,n))},t.prototype.parsePropertyBinding=function(t,e,n,r,i,o){var a=!1;t.startsWith("animate-")?(a=!0,t=t.substring("animate-".length)):lc(t)&&(a=!0,t=t.substring(1)),a?this._parseAnimation(t,e,r,i,o):this._parsePropertyAst(t,this._parseBinding(e,n,r),r,i,o)},t.prototype.parsePropertyInterpolation=function(t,e,n,r,i){var o=this.parseInterpolation(e,n);return!!o&&(this._parsePropertyAst(t,o,n,r,i),!0)},t.prototype._parsePropertyAst=function(t,e,n,r,i){r.push([t,e.source]),i.push(new qo(t,e,Uo.DEFAULT,n))},t.prototype._parseAnimation=function(t,e,n,r,i){var o=this._parseBinding(e||"undefined",!1,n);r.push([t,o.source]),i.push(new qo(t,o,Uo.ANIMATION,n))},t.prototype._parseBinding=function(t,e,n){var r=(n&&n.start||"(unknown)").toString();try{var i=e?this._exprParser.parseSimpleBinding(t,r,this._interpolationConfig):this._exprParser.parseBinding(t,r,this._interpolationConfig);return i&&this._reportExpressionParserErrors(i.errors,n),this._checkPipes(i,n),i}catch(t){return this._reportError(""+t,n),this._exprParser.wrapLiteralPrimitive("ERROR",r)}},t.prototype.createBoundElementProperty=function(t,e){if(e.isAnimation)return new Ko(e.name,4,R.NONE,e.expression,null,e.sourceSpan);var n=null,r=void 0,i=null,o=e.name.split("."),a=void 0;if(o.length>1)if("attr"==o[0]){i=o[1],this._validatePropertyOrAttributeName(i,e.sourceSpan,!0),a=dc(this._schemaRegistry,t,i,!0);var s=i.indexOf(":");if(s>-1)i=d(i.substring(0,s),i.substring(s+1));r=1}else"class"==o[0]?(i=o[1],r=2,a=[R.NONE]):"style"==o[0]&&(n=o.length>2?o[2]:null,i=o[1],r=3,a=[R.STYLE]);return null===i&&(i=this._schemaRegistry.getMappedPropName(e.name),a=dc(this._schemaRegistry,t,i,!1),r=0,this._validatePropertyOrAttributeName(i,e.sourceSpan,!1)),new Ko(i,r,a[0],e.expression,n,e.sourceSpan)},t.prototype.parseEvent=function(t,e,n,r,i){lc(t)?(t=t.substr(1),this._parseAnimationEvent(t,e,n,i)):this._parseRegularEvent(t,e,n,r,i)},t.prototype._parseAnimationEvent=function(t,e,n,r){var i=Le(t,".",[t,""]),o=i[0],a=i[1].toLowerCase();if(a)switch(a){case"start":case"done":var s=this._parseAction(e,n);r.push(new Go(o,a,1,s,n));break;default:this._reportError('The provided animation output phase value "'+a+'" for "@'+o+'" is not supported (use start or done)',n)}else this._reportError("The animation trigger output event (@"+o+") is missing its phase value name (start or done are currently supported)",n)},t.prototype._parseRegularEvent=function(t,e,n,r,i){var a=Object(o.__read)(Fe(t,[null,t]),2),s=a[0],u=a[1],c=this._parseAction(e,n);r.push([t,c.source]),i.push(new Go(u,s,0,c,n))},t.prototype._parseAction=function(t,e){var n=(e&&e.start||"(unknown").toString();try{var r=this._exprParser.parseAction(t,n,this._interpolationConfig);return r&&this._reportExpressionParserErrors(r.errors,e),!r||r.ast instanceof go?(this._reportError("Empty expressions are not allowed",e),this._exprParser.wrapLiteralPrimitive("ERROR",n)):(this._checkPipes(r,e),r)}catch(t){return this._reportError(""+t,e),this._exprParser.wrapLiteralPrimitive("ERROR",n)}},t.prototype._reportError=function(t,e,n){void 0===n&&(n=ts.ERROR),this.errors.push(new os(e,t,n))},t.prototype._reportExpressionParserErrors=function(t,e){var n,r;try{for(var i=Object(o.__values)(t),a=i.next();!a.done;a=i.next()){var s=a.value;this._reportError(s.message,e)}}catch(t){n={error:t}}finally{try{a&&!a.done&&(r=i.return)&&r.call(i)}finally{if(n)throw n.error}}},t.prototype._checkPipes=function(t,e){var n=this;if(t&&this.pipesByName){var r=new cc;t.visit(r),r.pipes.forEach(function(t,r){var i=n.pipesByName.get(r);i?n._usedPipes.set(r,i):n._reportError("The pipe '"+r+"' could not be found",new is(e.start.moveBy(t.span.start),e.start.moveBy(t.span.end)))})}},t.prototype._validatePropertyOrAttributeName=function(t,e,n){var r=n?this._schemaRegistry.validateAttribute(t):this._schemaRegistry.validateProperty(t);r.error&&this._reportError(r.msg,e,ts.ERROR)},t}(),cc=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.pipes=new Map,e}return Object(o.__extends)(e,t),e.prototype.visitPipe=function(t,e){return this.pipes.set(t.name,t),t.exp.visit(this),this.visitAll(t.args,e),null},e}(Bo);function lc(t){return"@"==t[0]}function dc(t,e,n,r){var i=[];return j.parse(e).forEach(function(e){var a=e.element?[e.element]:t.allKnownElementNames(),s=new Set(e.notSelectors.filter(function(t){return t.isElementSelector()}).map(function(t){return t.element})),u=a.filter(function(t){return!s.has(t)});i.push.apply(i,Object(o.__spread)(u.map(function(e){return t.securityContext(e,n,r)})))}),0===i.length?[R.NONE]:Array.from(new Set(i)).sort()}var pc=function(){function t(t,e){this.value=t,this.sourceSpan=e}return t.prototype.visit=function(t){return t.visitText(this)},t}(),fc=function(){function t(t,e,n){this.value=t,this.sourceSpan=e,this.i18n=n}return t.prototype.visit=function(t){return t.visitBoundText(this)},t}(),hc=function(){function t(t,e,n,r,i){this.name=t,this.value=e,this.sourceSpan=n,this.valueSpan=r,this.i18n=i}return t.prototype.visit=function(t){return t.visitTextAttribute(this)},t}(),mc=function(){function t(t,e,n,r,i,o,a){this.name=t,this.type=e,this.securityContext=n,this.value=r,this.unit=i,this.sourceSpan=o,this.i18n=a}return t.fromBoundElementProperty=function(e,n){return new t(e.name,e.type,e.securityContext,e.value,e.unit,e.sourceSpan,n)},t.prototype.visit=function(t){return t.visitBoundAttribute(this)},t}(),gc=function(){function t(t,e,n,r,i,o){this.name=t,this.type=e,this.handler=n,this.target=r,this.phase=i,this.sourceSpan=o}return t.fromParsedEvent=function(e){var n=0===e.type?e.targetOrPhase:null,r=1===e.type?e.targetOrPhase:null;return new t(e.name,e.type,e.handler,n,r,e.sourceSpan)},t.prototype.visit=function(t){return t.visitBoundEvent(this)},t}(),yc=function(){function t(t,e,n,r,i,o,a,s,u,c){this.name=t,this.attributes=e,this.inputs=n,this.outputs=r,this.children=i,this.references=o,this.sourceSpan=a,this.startSourceSpan=s,this.endSourceSpan=u,this.i18n=c}return t.prototype.visit=function(t){return t.visitElement(this)},t}(),_c=function(){function t(t,e,n,r,i,o,a,s,u,c,l){this.tagName=t,this.attributes=e,this.inputs=n,this.outputs=r,this.children=i,this.references=o,this.variables=a,this.sourceSpan=s,this.startSourceSpan=u,this.endSourceSpan=c,this.i18n=l}return t.prototype.visit=function(t){return t.visitTemplate(this)},t}(),vc=function(){function t(t,e,n,r){this.selector=t,this.attributes=e,this.sourceSpan=n,this.i18n=r}return t.prototype.visit=function(t){return t.visitContent(this)},t}(),bc=function(){function t(t,e,n){this.name=t,this.value=e,this.sourceSpan=n}return t.prototype.visit=function(t){return t.visitVariable(this)},t}(),xc=function(){function t(t,e,n){this.name=t,this.value=e,this.sourceSpan=n}return t.prototype.visit=function(t){return t.visitReference(this)},t}(),wc=function(){function t(t,e,n,r){this.vars=t,this.placeholders=e,this.sourceSpan=n,this.i18n=r}return t.prototype.visit=function(t){return t.visitIcu(this)},t}();function jc(t,e){var n,r,i,a,s=[];if(t.visit)try{for(var u=Object(o.__values)(e),c=u.next();!c.done;c=u.next())var l=c.value,d=t.visit(l)||l.visit(t)}catch(t){n={error:t}}finally{try{c&&!c.done&&(r=u.return)&&r.call(u)}finally{if(n)throw n.error}}else try{for(var p=Object(o.__values)(e),f=p.next();!f.done;f=p.next()){(d=(l=f.value).visit(t))&&s.push(d)}}catch(t){i={error:t}}finally{try{f&&!f.done&&(a=p.return)&&a.call(p)}finally{if(i)throw i.error}}return s}var Oc=function(){return function(t,e){this.style=t,this.styleUrls=e}}();function Cc(t){if(null==t||0===t.length||"/"==t[0])return!1;var e=t.match(kc);return null===e||"package"==e[1]||"asset"==e[1]}var Ec,Sc,Mc=/@import\s+(?:url\()?\s*(?:(?:['"]([^'"]*))|([^;\)\s]*))[^;]*;?/g,Tc=/\/\*(?!#\s*(?:sourceURL|sourceMappingURL)=)[\s\S]+?\*\//g,kc=/^([^:/?#]+):/,Ac="select",Dc="link",Ic="rel",Pc="href",Rc="stylesheet",Nc="style",Fc="script",Lc="ngNonBindable",Bc="ngProjectAs";function Vc(t){var e=null,n=null,r=null,i=!1,o="";t.attrs.forEach(function(t){var a=t.name.toLowerCase();a==Ac?e=t.value:a==Pc?n=t.value:a==Ic?r=t.value:t.name==Lc?i=!0:t.name==Bc&&t.value.length>0&&(o=t.value)}),e=function(t){if(null===t||0===t.length)return"*";return t}(e);var a=t.name.toLowerCase(),s=Ec.OTHER;return u(a)?s=Ec.NG_CONTENT:a==Nc?s=Ec.STYLE:a==Fc?s=Ec.SCRIPT:a==Dc&&r==Rc&&(s=Ec.STYLESHEET),new zc(s,e,n,i,o)}(Sc=Ec||(Ec={}))[Sc.NG_CONTENT=0]="NG_CONTENT",Sc[Sc.STYLE=1]="STYLE",Sc[Sc.STYLESHEET=2]="STYLESHEET",Sc[Sc.SCRIPT=3]="SCRIPT",Sc[Sc.OTHER=4]="OTHER";var zc=function(){return function(t,e,n,r,i){this.type=t,this.selectAttr=e,this.hrefAttr=n,this.nonBindable=r,this.projectAs=i}}();var Hc=/^(?:(?:(?:(bind-)|(let-)|(ref-|#)|(on-)|(bindon-)|(@))(.+))|\[\(([^\)]+)\)\]|\[([^\]]+)\]|\(([^\)]+)\))$/;function Uc(t,e){var n=new Gc(e),r=_u(n,t),i=e.errors.concat(n.errors),o=i.filter(function(t){return t.level===ts.ERROR});if(o.length>0)throw Ye("Template parse errors:\n"+o.join("\n"),o);return{nodes:r,errors:i}}var Wc,qc,Gc=function(){function t(t){this.bindingParser=t,this.errors=[]}return t.prototype.visitElement=function(t){var e,n,r=this,i=Vc(t);if(i.type===Ec.SCRIPT||i.type===Ec.STYLE)return null;if(i.type===Ec.STYLESHEET&&Cc(i.hrefAttr))return null;var a=c(t.name),s=[],u=[],l=[],d=[],p=[],f={},h=[],m=[],g=!1;try{for(var y=Object(o.__values)(t.attrs),_=y.next();!_.done;_=y.next()){var v=_.value,b=!1,x=Kc(v.name),w=!1;if(v.i18n&&(f[v.name]=v.i18n),x.startsWith("*")){g&&this.reportError("Can't have multiple template bindings on one element. Use only one attribute prefixed with *",v.sourceSpan),w=!0,g=!0;var j=v.value,O=x.substring("*".length),C=[];this.bindingParser.parseInlineTemplateBinding(O,j,v.sourceSpan,[],h,C),m.push.apply(m,Object(o.__spread)(C.map(function(t){return new bc(t.name,t.value,t.sourceSpan)})))}else b=this.parseAttribute(a,v,[],s,u,l,d);b||w||p.push(this.visitAttribute(v))}}catch(t){e={error:t}}finally{try{_&&!_.done&&(n=y.return)&&n.call(y)}finally{if(e)throw e.error}}var E,S=_u(i.nonBindable?Yc:this,t.children);if(i.type===Ec.NG_CONTENT){t.children&&!t.children.every(Qc)&&this.reportError("<ng-content> element cannot have content.",t.sourceSpan);var M=i.selectAttr,T=t.attrs.map(function(t){return r.visitAttribute(t)});E=new vc(M,T,t.sourceSpan,t.i18n)}else if(a){T=this.extractAttributes(t.name,s,f);E=new _c(t.name,p,T.bound,u,S,d,l,t.sourceSpan,t.startSourceSpan,t.endSourceSpan,t.i18n)}else{T=this.extractAttributes(t.name,s,f);E=new yc(t.name,p,T.bound,u,S,d,t.sourceSpan,t.startSourceSpan,t.endSourceSpan,t.i18n)}if(g){T=this.extractAttributes("ng-template",h,f);E=new _c(E.name,T.literal,T.bound,[],[E],[],m,t.sourceSpan,t.startSourceSpan,t.endSourceSpan,t.i18n)}return E},t.prototype.visitAttribute=function(t){return new hc(t.name,t.value,t.sourceSpan,t.valueSpan,t.i18n)},t.prototype.visitText=function(t){return this._visitTextWithInterpolation(t.value,t.sourceSpan,t.i18n)},t.prototype.visitExpansion=function(t){var e=this,n=t.i18n;if(!n)return null;var r={},i={};return Object.keys(n.placeholders).forEach(function(o){var a=n.placeholders[o];if(o.startsWith("VAR_")){var s=e.bindingParser.interpolationConfig,u=""+s.start+a+s.end;r[o]=e._visitTextWithInterpolation(u,t.sourceSpan)}else i[o]=e._visitTextWithInterpolation(a,t.sourceSpan)}),new wc(r,i,t.sourceSpan,n)},t.prototype.visitExpansionCase=function(t){return null},t.prototype.visitComment=function(t){return null},t.prototype.extractAttributes=function(t,e,n){var r=this,i=[],o=[];return e.forEach(function(e){var a=n[e.name];if(e.isLiteral)o.push(new hc(e.name,e.expression.source||"",e.sourceSpan,void 0,a));else{var s=r.bindingParser.createBoundElementProperty(t,e);i.push(mc.fromBoundElementProperty(s,a))}}),{bound:i,literal:o}},t.prototype.parseAttribute=function(t,e,n,r,i,o,a){var s=Kc(e.name),u=e.value,c=e.sourceSpan,l=s.match(Hc),d=!1;if(l){if(d=!0,null!=l[1])this.bindingParser.parsePropertyBinding(l[7],u,!1,c,n,r);else if(l[2])if(t){var p=l[7];this.parseVariable(p,u,c,o)}else this.reportError('"let-" is only supported on ng-template elements.',c);else if(l[3]){p=l[7];this.parseReference(p,u,c,a)}else if(l[4]){var f=[];this.bindingParser.parseEvent(l[7],u,c,n,f),Xc(f,i)}else if(l[5])this.bindingParser.parsePropertyBinding(l[7],u,!1,c,n,r),this.parseAssignmentEvent(l[7],u,c,n,i);else if(l[6])this.bindingParser.parseLiteralAttr(s,u,c,n,r);else if(l[8])this.bindingParser.parsePropertyBinding(l[8],u,!1,c,n,r),this.parseAssignmentEvent(l[8],u,c,n,i);else if(l[9])this.bindingParser.parsePropertyBinding(l[9],u,!1,c,n,r);else if(l[10]){f=[];this.bindingParser.parseEvent(l[10],u,c,n,f),Xc(f,i)}}else d=this.bindingParser.parsePropertyInterpolation(s,u,c,n,r);return d},t.prototype._visitTextWithInterpolation=function(t,e,n){var r=Ku(t),i=this.bindingParser.parseInterpolation(r,e);return i?new fc(i,e,n):new pc(r,e)},t.prototype.parseVariable=function(t,e,n,r){t.indexOf("-")>-1&&this.reportError('"-" is not allowed in variable names',n),r.push(new bc(t,e,n))},t.prototype.parseReference=function(t,e,n,r){t.indexOf("-")>-1&&this.reportError('"-" is not allowed in reference names',n),r.push(new xc(t,e,n))},t.prototype.parseAssignmentEvent=function(t,e,n,r,i){var o=[];this.bindingParser.parseEvent(t+"Change",e+"=$event",n,r,o),Xc(o,i)},t.prototype.reportError=function(t,e,n){void 0===n&&(n=ts.ERROR),this.errors.push(new os(e,t,n))},t}(),Yc=new(function(){function t(){}return t.prototype.visitElement=function(t){var e=Vc(t);if(e.type===Ec.SCRIPT||e.type===Ec.STYLE||e.type===Ec.STYLESHEET)return null;var n=_u(this,t.children,null);return new yc(t.name,_u(this,t.attrs),[],[],n,[],t.sourceSpan,t.startSourceSpan,t.endSourceSpan)},t.prototype.visitComment=function(t){return null},t.prototype.visitAttribute=function(t){return new hc(t.name,t.value,t.sourceSpan,void 0,t.i18n)},t.prototype.visitText=function(t){return new pc(t.value,t.sourceSpan)},t.prototype.visitExpansion=function(t){return null},t.prototype.visitExpansionCase=function(t){return null},t}());function Kc(t){return/^data-/i.test(t)?t.substring(5):t}function Xc(t,e){e.push.apply(e,Object(o.__spread)(t.map(function(t){return gc.fromParsedEvent(t)})))}function Qc(t){return t instanceof pu&&0==t.value.trim().length}(qc=Wc||(Wc={}))[qc.ELEMENT=0]="ELEMENT",qc[qc.TEMPLATE=1]="TEMPLATE";var Zc=function(){function t(t,e,n,r,i,o){void 0===n&&(n=0),void 0===r&&(r=null),this.index=t,this.ref=e,this.level=n,this.templateIndex=r,this.meta=i,this.registry=o,this.bindings=new Set,this.placeholders=new Map,this._unresolvedCtxCount=0,this._registry=o||{getUniqueId:function(t){void 0===t&&(t=0);var e=t;return function(){return e++}}(),icus:new Map},this.id=this._registry.getUniqueId()}return t.prototype.appendTag=function(t,e,n,r){if(!e.isVoid||!r){var i=e.isVoid||!r?e.startName:e.closeName,o={type:t,index:n,ctx:this.id,isVoid:e.isVoid,closed:r};ai(this.placeholders,i,o)}},Object.defineProperty(t.prototype,"icus",{get:function(){return this._registry.icus},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"isRoot",{get:function(){return 0===this.level},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"isResolved",{get:function(){return 0===this._unresolvedCtxCount},enumerable:!0,configurable:!0}),t.prototype.getSerializedPlaceholders=function(){var t=new Map;return this.placeholders.forEach(function(e,n){return t.set(n,e.map(el))}),t},t.prototype.appendBinding=function(t){this.bindings.add(t)},t.prototype.appendIcu=function(t,e){ai(this._registry.icus,t,e)},t.prototype.appendBoundText=function(t){var e=this;si(t,this.bindings.size,this.id).forEach(function(t,n){return ai.apply(void 0,Object(o.__spread)([e.placeholders,n],t))})},t.prototype.appendTemplate=function(t,e){this.appendTag(Wc.TEMPLATE,t,e,!1),this.appendTag(Wc.TEMPLATE,t,e,!0),this._unresolvedCtxCount++},t.prototype.appendElement=function(t,e,n){this.appendTag(Wc.ELEMENT,t,e,n)},t.prototype.forkChildContext=function(e,n,r){return new t(e,this.ref,this.level+1,n,r,this._registry)},t.prototype.reconcileChildContext=function(t){var e=this;["start","close"].forEach(function(n){var r=t.meta[n+"Name"],i=(e.placeholders.get(r)||[]).find(tl(e.id,t.templateIndex));i&&(i.ctx=t.id)}),t.placeholders.forEach(function(n,r){var i=e.placeholders.get(r);if(i){var a=function(t,e){for(var n=0;n<t.length;n++)if(e(t[n]))return n;return-1}(i,tl(t.id,t.templateIndex));if(a>=0){var s=r.startsWith("CLOSE");if(r.endsWith("NG-TEMPLATE"))i.splice.apply(i,Object(o.__spread)([a+(s?0:1),0],n));else n[s?n.length-1:0].tmpl=i[a],i.splice.apply(i,Object(o.__spread)([a,1],n))}else i.push.apply(i,Object(o.__spread)(n));e.placeholders.set(r,i)}else e.placeholders.set(r,n)}),this._unresolvedCtxCount--},t}();function $c(t,e,n,r){return ii(""+(r?"/":"")+t+e,n)}function Jc(t,e,n){var r=e.index,i=e.ctx;return e.isVoid?$c(t,r,i)+$c(t,r,i,!0):$c(t,r,i,n)}function tl(t,e){return function(n){return"object"==typeof n&&n.type===Wc.TEMPLATE&&n.index===e&&n.ctx===t}}function el(t){var e=function(t,e){return Jc("#",t,e)},n=function(t,e){return Jc("*",t,e)};switch(t.type){case Wc.ELEMENT:return t.closed?e(t,!0)+(t.tmpl?n(t.tmpl,!0):""):t.tmpl?n(t.tmpl)+e(t)+(t.isVoid?n(t.tmpl,!0):""):e(t);case Wc.TEMPLATE:return n(t,t.closed);default:return t}}var nl={A:"LINK",B:"BOLD_TEXT",BR:"LINE_BREAK",EM:"EMPHASISED_TEXT",H1:"HEADING_LEVEL1",H2:"HEADING_LEVEL2",H3:"HEADING_LEVEL3",H4:"HEADING_LEVEL4",H5:"HEADING_LEVEL5",H6:"HEADING_LEVEL6",HR:"HORIZONTAL_RULE",I:"ITALIC_TEXT",LI:"LIST_ITEM",LINK:"MEDIA_LINK",OL:"ORDERED_LIST",P:"PARAGRAPH",Q:"QUOTATION",S:"STRIKETHROUGH_TEXT",SMALL:"SMALL_TEXT",SUB:"SUBSTRIPT",SUP:"SUPERSCRIPT",TBODY:"TABLE_BODY",TD:"TABLE_CELL",TFOOT:"TABLE_FOOTER",TH:"TABLE_HEADER_CELL",THEAD:"TABLE_HEADER",TR:"TABLE_ROW",TT:"MONOSPACED_TEXT",U:"UNDERLINED_TEXT",UL:"UNORDERED_LIST"},rl=function(){function t(){this._placeHolderNameCounts={},this._signatureToName={}}return t.prototype.getStartTagPlaceholderName=function(t,e,n){var r=this._hashTag(t,e,n);if(this._signatureToName[r])return this._signatureToName[r];var i=t.toUpperCase(),o=nl[i]||"TAG_"+i,a=this._generateUniqueName(n?o:"START_"+o);return this._signatureToName[r]=a,a},t.prototype.getCloseTagPlaceholderName=function(t){var e=this._hashClosingTag(t);if(this._signatureToName[e])return this._signatureToName[e];var n=t.toUpperCase(),r=nl[n]||"TAG_"+n,i=this._generateUniqueName("CLOSE_"+r);return this._signatureToName[e]=i,i},t.prototype.getPlaceholderName=function(t,e){var n=t.toUpperCase(),r="PH: "+n+"="+e;if(this._signatureToName[r])return this._signatureToName[r];var i=this._generateUniqueName(n);return this._signatureToName[r]=i,i},t.prototype.getUniquePlaceholder=function(t){return this._generateUniqueName(t.toUpperCase())},t.prototype._hashTag=function(t,e,n){return"<"+t+Object.keys(e).sort().map(function(t){return" "+t+"="+e[t]}).join("")+(n?"/>":"></"+t+">")},t.prototype._hashClosingTag=function(t){return this._hashTag("/"+t,{},!1)},t.prototype._generateUniqueName=function(t){if(!this._placeHolderNameCounts.hasOwnProperty(t))return this._placeHolderNameCounts[t]=1,t;var e=this._placeHolderNameCounts[t];return this._placeHolderNameCounts[t]=e+1,t+"_"+e},t}(),il=new uu(new Ks);function ol(t){var e=new al(il,t);return function(t,n,r,i,o){return e.toI18nMessage(t,n,r,i,o)}}var al=function(){function t(t,e){this._expressionParser=t,this._interpolationConfig=e}return t.prototype.toI18nMessage=function(t,e,n,r,i){this._isIcu=1==t.length&&t[0]instanceof fu,this._icuDepth=0,this._placeholderRegistry=new rl,this._placeholderToContent={},this._placeholderToMessage={},this._visitNodeFn=i;var o=_u(this,t,{});return new Zn(o,this._placeholderToContent,this._placeholderToMessage,e,n,r)},t.prototype._visitNode=function(t,e){return this._visitNodeFn&&this._visitNodeFn(t,e),e},t.prototype.visitElement=function(t,e){var n=_u(this,t.children),r={};t.attrs.forEach(function(t){r[t.name]=t.value});var i=y(t.name).isVoid,o=this._placeholderRegistry.getStartTagPlaceholderName(t.name,r,i);this._placeholderToContent[o]=t.sourceSpan.toString();var a="";i||(a=this._placeholderRegistry.getCloseTagPlaceholderName(t.name),this._placeholderToContent[a]="</"+t.name+">");var s=new er(t.name,r,o,a,n,i,t.sourceSpan);return this._visitNode(t,s)},t.prototype.visitAttribute=function(t,e){var n=this._visitTextWithInterpolation(t.value,t.sourceSpan);return this._visitNode(t,n)},t.prototype.visitText=function(t,e){var n=this._visitTextWithInterpolation(t.value,t.sourceSpan);return this._visitNode(t,n)},t.prototype.visitComment=function(t,e){return null},t.prototype.visitExpansion=function(e,n){var r=this;this._icuDepth++;var i={},o=new tr(e.switchValue,e.type,i,e.sourceSpan);if(e.cases.forEach(function(t){i[t.value]=new Jn(t.expression.map(function(t){return t.visit(r,{})}),t.expSourceSpan)}),this._icuDepth--,this._isIcu||this._icuDepth>0){var a=this._placeholderRegistry.getUniquePlaceholder("VAR_"+e.type);return o.expressionPlaceholder=a,this._placeholderToContent[a]=e.switchValue,this._visitNode(e,o)}var s=this._placeholderRegistry.getPlaceholderName("ICU",e.sourceSpan.toString()),u=new t(this._expressionParser,this._interpolationConfig);this._placeholderToMessage[s]=u.toI18nMessage([e],"","","");var c=new rr(o,s,e.sourceSpan);return this._visitNode(e,c)},t.prototype.visitExpansionCase=function(t,e){throw new Error("Unreachable code")},t.prototype._visitTextWithInterpolation=function(t,e){var n=this._expressionParser.splitInterpolation(t,e.start.toString(),this._interpolationConfig);if(!n)return new $n(t,e);for(var r=[],i=new Jn(r,e),o=this._interpolationConfig,a=o.start,s=o.end,u=0;u<n.strings.length-1;u++){var c=n.expressions[u],l=c.split(sl)[2]||"INTERPOLATION",d=this._placeholderRegistry.getPlaceholderName(l,c);n.strings[u].length&&r.push(new $n(n.strings[u],e)),r.push(new nr(c,d,e)),this._placeholderToContent[d]=a+c+s}var p=n.strings.length-1;return n.strings[p].length&&r.push(new $n(n.strings[p],e)),i},t}(),sl=/\/\/[\s\S]*i18n[\s\S]*\([\s\S]*ph[\s\S]*=[\s\S]*("|')([\s\S]*?)\1[\s\S]*\)/g;function ul(t,e){t.i18n=e}var cl=function(){function t(t,e){void 0===t&&(t=Hi),void 0===e&&(e=!1),this.interpolationConfig=t,this.keepI18nAttrs=e,this._createI18nMessage=ol(t)}return t.prototype._generateI18nMessage=function(t,e,n){void 0===e&&(e="");var r="string"==typeof e?function(t){var e,n,r,i,a;if(t){var s=t.indexOf(Kr),u=t.indexOf(Yr),c=void 0;c=(e=Object(o.__read)(s>-1?[t.slice(0,s),t.slice(s+2)]:[t,""],2))[0],r=e[1],i=(n=Object(o.__read)(u>-1?[c.slice(0,u),c.slice(u+1)]:["",c],2))[0],a=n[1]}return{id:r,meaning:i,description:a}}(e):ni(e),i=this._createI18nMessage(t,r.meaning||"",r.description||"",r.id||"",n);return i.id||(i.id="string"!=typeof e&&e.id||sr(i)),i},t.prototype.visitElement=function(t,e){var n,r,i,a;if(t.attrs.some(function(t){return Jr(t.name)})){var s=[],u={};try{for(var c=Object(o.__values)(t.attrs),l=c.next();!l.done;l=c.next()){if((m=l.value).name===Xr){var d=t.i18n||m.value,p=this._generateI18nMessage(t.children,d,ul);p.nodes.length&&(t.i18n=p)}else if(m.name.startsWith(Qr)){u[m.name.slice(Qr.length)]=m.value}else s.push(m)}}catch(t){n={error:t}}finally{try{l&&!l.done&&(r=c.return)&&r.call(c)}finally{if(n)throw n.error}}if(Object.keys(u).length)try{for(var f=Object(o.__values)(s),h=f.next();!h.done;h=f.next()){var m,g=u[(m=h.value).name];void 0!==g&&m.value&&(m.i18n=this._generateI18nMessage([m],m.i18n||g))}}catch(t){i={error:t}}finally{try{h&&!h.done&&(a=f.return)&&a.call(f)}finally{if(i)throw i.error}}this.keepI18nAttrs||(t.attrs=s)}return _u(this,t.children),t},t.prototype.visitExpansion=function(t,e){var n,r=t.i18n;if(r instanceof rr){var i=r.name;ri(n=this._generateI18nMessage([t],r)).name=i}else n=this._generateI18nMessage([t],r);return t.i18n=n,t},t.prototype.visitText=function(t,e){return t},t.prototype.visitAttribute=function(t,e){return t},t.prototype.visitComment=function(t,e){return t},t.prototype.visitExpansionCase=function(t,e){return t},t}(),ll=function(t){return"{$"+ui(t)+"}"},dl=new(function(){function t(){}return t.prototype.visitText=function(t,e){return t.value},t.prototype.visitContainer=function(t,e){var n=this;return t.children.map(function(t){return t.visit(n)}).join("")},t.prototype.visitIcu=function(t,e){var n=this,r=Object.keys(t.cases).map(function(e){return e+" {"+t.cases[e].visit(n)+"}"});return"{"+t.expressionPlaceholder+", "+t.type+", "+r.join(" ")+"}"},t.prototype.visitTagPlaceholder=function(t,e){var n=this;return t.isVoid?ll(t.startName):""+ll(t.startName)+t.children.map(function(t){return t.visit(n)}).join("")+ll(t.closeName)},t.prototype.visitPlaceholder=function(t,e){return ll(t.name)},t.prototype.visitIcuPlaceholder=function(t,e){return ll(t.name)},t}());function pl(t,e){return Ie(Ee(fi).bitwiseAnd(Pe(t),null,!1),e)}var fl=function(){function t(t,e,n,r,i,o,a,s,u,c,l,d,p,f,h){void 0===n&&(n=0);var m=this;this.constantPool=t,this.level=n,this.contextName=r,this.i18nContext=i,this.templateIndex=o,this.templateName=a,this.viewQueries=s,this.directiveMatcher=u,this.directives=c,this.pipeTypeByName=l,this.pipes=d,this._namespace=p,this.relativeContextFilePath=f,this.i18nUseExternalIds=h,this._dataIndex=0,this._bindingContext=0,this._prefixCode=[],this._creationCodeFns=[],this._updateCodeFns=[],this._tempVariables=[],this._nestedTemplateFns=[],this._unsupported=mi,this.i18n=null,this._pureFunctionSlots=0,this._bindingSlots=0,this._hasNgContent=!1,this._ngContentSelectors=[],this._ngContentSelectorsOffset=0,this.visitReference=gi,this.visitVariable=gi,this.visitTextAttribute=gi,this.visitBoundAttribute=gi,this.visitBoundEvent=gi,this._dataIndex=s.length,this._bindingScope=e.nestedScope(n),this.fileBasedI18nSuffix=f.replace(/[^A-Za-z0-9]/g,"_")+"_",this._valueConverter=new hl(t,function(){return m.allocateDataSlot()},function(t){return m.allocatePureFunctionSlots(t)},function(t,e,n,r){var i=l.get(t);i&&m.pipes.add(i),m._bindingScope.set(m.level,e,r),m.creationInstruction(null,Qn.pipe,[Pe(n),Pe(t)])})}return t.prototype.registerContextVariables=function(t){var e=this._bindingScope.freshReferenceName(),n=this.level,r=Ee(t.name+e);this._bindingScope.set(n,t.name,r,1,function(e,i){var o;if(e.bindingLevel===n)o=Ee(pi);else{var a=e.getSharedContextName(n);o=a||_l(i)}return[r.set(o.prop(t.value||"$implicit")).toConstDecl()]})},t.prototype.buildTemplateFunction=function(t,e,n,r){var i=this;void 0===n&&(n=0),this._ngContentSelectorsOffset=n,this._namespace!==Qn.namespaceHTML&&this.creationInstruction(null,this._namespace),e.forEach(function(t){return i.registerContextVariables(t)});var a,s=this.i18nContext||ti(r)&&!ei(r)&&!(1===(a=t).length&&a[0]instanceof yc&&t[0].i18n===r),u=El(t);if(s&&this.i18nStart(null,r,u),jc(this,t),this._pureFunctionSlots+=this._bindingSlots,this._valueConverter.updatePipeSlotOffsets(this._bindingSlots),this._nestedTemplateFns.forEach(function(t){return t()}),0===this.level&&this._hasNgContent){var c=[];if(this._ngContentSelectors.length){var l=this._ngContentSelectors.map(function(t){return nt(t)}),d=this.constantPool.getConstLiteral(yi(l),!0),p=this.constantPool.getConstLiteral(yi(this._ngContentSelectors),!0);c.push(d,p)}this.creationInstruction(null,Qn.projectionDef,c,!0)}s&&this.i18nEnd(null,u);var f=this._creationCodeFns.map(function(t){return t()}),h=this._updateCodeFns.map(function(t){return t()}),m=this._bindingScope.viewSnapshotStatements(),g=this._bindingScope.variableDeclarations().concat(this._tempVariables),y=f.length>0?[pl(1,m.concat(f))]:[],_=h.length>0?[pl(2,g.concat(h))]:[];return De([new qt(fi,_t),new qt(pi,null)],Object(o.__spread)(this._prefixCode,y,_),gt,null,this.templateName)},t.prototype.getLocal=function(t){return this._bindingScope.get(t)},t.prototype.i18nTranslate=function(t,e,n,r){var i;void 0===e&&(e={});var a=n||this.i18nAllocateRef(t.id),s={};e&&Object.keys(e).length&&Object.keys(e).forEach(function(t){return s[ui(t)]=e[t]});var u=ni(t),c=li(a,t.nodes.map(function(t){return t.visit(dl,null)}).join(""),u,s,r);return(i=this.constantPool.statements).push.apply(i,Object(o.__spread)(c)),a},t.prototype.i18nAppendBindings=function(t){var e=this;if(this.i18n&&t.length){var n=Ee(pi);t.forEach(function(t){var r=e.convertExpressionBinding(n,t);e.i18n.appendBinding(r)})}},t.prototype.i18nBindProps=function(t){var e=this,n={};return Object.keys(t).forEach(function(r){var i=t[r];if(i instanceof pc)n[r]=Pe(i.value);else{var o=i.value.visit(e._valueConverter);if(e.allocateBindingSlots(o),o instanceof To){var a=o.strings,s=o.expressions,u=e.i18n,c=u.id,l=function(t,e,n){if(void 0===e&&(e=0),void 0===n&&(n=0),!t.length)return"";for(var r="",i=t.length-1,o=0;o<i;o++)r+=""+t[o]+ii(e+o,n);return r+=t[i]}(a,u.bindings.size,c);e.i18nAppendBindings(s),n[r]=Pe(l)}}}),n},t.prototype.i18nAllocateRef=function(t){var e,n=this.fileBasedI18nSuffix.toUpperCase();if(this.i18nUseExternalIds){e=""+(r=ci("EXTERNAL_"))+t+"$$"+this.constantPool.uniqueName(n)}else{var r=ci(n);e=this.constantPool.uniqueName(r)}return Ee(e)},t.prototype.i18nUpdateRef=function(t){var e=t.icus,n=t.meta,r=t.isRoot,i=t.isResolved;if(r&&i&&!ei(n)){var o=t.getSerializedPlaceholders(),a={},s=o.size?oi(o):{};e.size&&e.forEach(function(t,e){if(1===t.length)s[e]=t[0];else{var n=ii("I18N_EXP_"+e);s[e]=Pe(n),a[e]=ke(t)}});var u=void 0;(Array.from(o.values()).some(function(t){return t.length>1})||Object.keys(a).length)&&(u=function(t){var e=[t];return Object.keys(a).length&&e.push(Hr(a,!0)),yl(null,Qn.i18nPostprocess,e)}),this.i18nTranslate(n,s,t.ref,u)}},t.prototype.i18nStart=function(t,e,n){void 0===t&&(t=null);var r=this.allocateDataSlot();if(this.i18nContext)this.i18n=this.i18nContext.forkChildContext(r,this.templateIndex,e);else{var i=this.i18nAllocateRef(e.id);this.i18n=new Zc(r,i,0,this.templateIndex,e)}var o=this.i18n,a=o.id,s=o.ref,u=[Pe(r),s];a>0&&u.push(Pe(a)),this.creationInstruction(t,n?Qn.i18n:Qn.i18nStart,u)},t.prototype.i18nEnd=function(t,e){var n=this;if(void 0===t&&(t=null),!this.i18n)throw new Error("i18nEnd is executed with no i18n context present");this.i18nContext?(this.i18nContext.reconcileChildContext(this.i18n),this.i18nUpdateRef(this.i18nContext)):this.i18nUpdateRef(this.i18n);var r=this.i18n,i=r.index,o=r.bindings;o.size&&(o.forEach(function(e){return n.updateInstruction(t,Qn.i18nExp,[e])}),this.updateInstruction(t,Qn.i18nApply,[Pe(i)])),e||this.creationInstruction(t,Qn.i18nEnd),this.i18n=null},t.prototype.visitContent=function(t){this._hasNgContent=!0;var e=this.allocateDataSlot(),n="*"===t.selector?0:this._ngContentSelectors.push(t.selector)+this._ngContentSelectorsOffset,r=[Pe(e)],i=[];t.attributes.forEach(function(t){var e=t.name,n=t.value;"select"!==e.toLowerCase()&&i.push(e,n)}),i.length>0?r.push(Pe(n),yi(i)):0!==n&&r.push(Pe(n)),this.creationInstruction(t.sourceSpan,Qn.projection,r)},t.prototype.getNamespaceInstruction=function(t){switch(t){case"math":return Qn.namespaceMathML;case"svg":return Qn.namespaceSVG;default:return Qn.namespaceHTML}},t.prototype.addNamespaceInstruction=function(t,e){this._namespace=t,this.creationInstruction(e.sourceSpan,t)},t.prototype.visitElement=function(t){var e,n,r=this,i=this.allocateDataSlot(),u=new Us(Pe(i),null),c=!1,l=ti(t.i18n)&&!ei(t.i18n);if(l&&this.i18n)throw new Error("Could not mark an element as translatable inside of a translatable section");var d=[],p=[],f=Object(o.__read)(a(t.name),2),h=f[0],m=f[1],g=s(t.name);try{for(var y=Object(o.__values)(t.attributes),_=y.next();!_.done;_=y.next()){var v=_.value,b=v.name,x=v.value;"ngNonBindable"===b?c=!0:"style"===b?u.registerStyleAttr(x):"class"===b?u.registerClassAttr(x):v.i18n?d.push(v):p.push(v)}}catch(t){e={error:t}}finally{try{_&&!_.done&&(n=y.return)&&n.call(y)}finally{if(e)throw e.error}}this.matchDirectives(t.name,t);var w=[Pe(i)];g||w.push(Pe(m));var j=[],O=[];t.inputs.forEach(function(t){u.registerBoundInput(t)||(0===t.type&&t.i18n?d.push(t):O.push(t))}),p.forEach(function(t){return j.push(Pe(t.name),Pe(t.value))}),j.push.apply(j,Object(o.__spread)(this.prepareSyntheticAndSelectOnlyAttrs(O,t.outputs,u))),w.push(this.toAttrsParam(j)),w.push(this.prepareRefsParameter(t.references));var C=this._namespace,E=this.getNamespaceInstruction(h);E!==C&&this.addNamespaceInstruction(E,t);var S=Ee(pi);this.i18n&&this.i18n.appendElement(t.i18n,i);var M=!u.hasBindingsOrInitialValues()&&!g&&0===t.outputs.length&&0===d.length&&!(!l&&r.i18n?!El(t.children):t.children.length>0),T=!M&&!u.hasBindingsOrInitialValues()&&El(t.children);if(M)this.creationInstruction(t.sourceSpan,Qn.element,vi(w));else{if(this.creationInstruction(t.sourceSpan,g?Qn.elementContainerStart:Qn.elementStart,vi(w)),c&&this.creationInstruction(t.sourceSpan,Qn.disableBindings),l&&this.i18nStart(t.sourceSpan,t.i18n,T),d.length){var k=!1,A=[];if(d.forEach(function(e){var n=e.i18n;if(e instanceof hc)A.push(Pe(e.name),r.i18nTranslate(n));else{var i=e.value.visit(r._valueConverter);if(r.allocateBindingSlots(i),i instanceof To){var o=oi(si(n));A.push(Pe(e.name),r.i18nTranslate(n,o)),i.expressions.forEach(function(e){k=!0;var n=r.convertExpressionBinding(S,e);r.updateInstruction(t.sourceSpan,Qn.i18nExp,[n])})}}}),A.length){var D=Pe(this.allocateDataSlot()),I=this.constantPool.getConstLiteral(ke(A),!0);this.creationInstruction(t.sourceSpan,Qn.i18nAttributes,[D,I]),k&&this.updateInstruction(t.sourceSpan,Qn.i18nApply,[D])}}this.processStylingInstruction(S,u.buildElementStylingInstruction(t.sourceSpan,this.constantPool),!0),t.outputs.forEach(function(e){r.creationInstruction(e.sourceSpan,Qn.listener,r.prepareListenerParameter(t.name,e,i))})}if(u.buildUpdateLevelInstructions(this._valueConverter).forEach(function(t){r.processStylingInstruction(S,t,!1)}),O.forEach(function(t){var e=function(t){switch(t){case 0:case 4:return Qn.elementProperty;case 2:return Qn.elementClassProp;case 1:return Qn.elementAttribute;default:return}}(t.type);if(4===t.type){var n=t.value.visit(r._valueConverter);if(n.name||n.value){var a=Pi(t.name);r.allocateBindingSlots(n),r.updateInstruction(t.sourceSpan,Qn.elementProperty,function(){return[Pe(i),Pe(a),r.convertPropertyBinding(S,n)]})}}else if(e){var s=[],u=function(t,e){switch(e){case R.HTML:return Se(Qn.sanitizeHtml);case R.SCRIPT:return Se(Qn.sanitizeScript);case R.STYLE:return 1===t.type?Se(Qn.sanitizeStyle):null;case R.URL:return Se(Qn.sanitizeUrl);case R.RESOURCE_URL:return Se(Qn.sanitizeResourceUrl);default:return null}}(t,t.securityContext);u&&s.push(u);var c=t.value.visit(r._valueConverter);r.allocateBindingSlots(c),r.updateInstruction(t.sourceSpan,e,function(){return Object(o.__spread)([Pe(i),Pe(t.name),r.convertPropertyBinding(S,c)],s)})}else r._unsupported("binding type "+t.type)}),jc(this,t.children),!l&&this.i18n&&this.i18n.appendElement(t.i18n,i,!0),!M){var P=t.endSourceSpan||t.sourceSpan;l&&this.i18nEnd(P,T),c&&this.creationInstruction(P,Qn.enableBindings),this.creationInstruction(P,g?Qn.elementContainerEnd:Qn.elementEnd)}},t.prototype.visitTemplate=function(e){var n=this,r=this.allocateDataSlot();this.i18n&&this.i18n.appendTemplate(e.i18n,r);var i=wn(e.tagName||""),a=(i?this.contextName+"_"+i:"")+"_"+r,s=a+"_Template",u=[Pe(r),Ee(s),Pe(e.tagName)];this.matchDirectives("ng-template",e);var c=[];e.attributes.forEach(function(t){c.push(yi(t.name),yi(t.value))}),c.push.apply(c,Object(o.__spread)(this.prepareSyntheticAndSelectOnlyAttrs(e.inputs,e.outputs))),u.push(this.toAttrsParam(c)),e.references&&e.references.length&&(u.push(this.prepareRefsParameter(e.references)),u.push(Se(Qn.templateRefExtractor)));var l=Ee(pi);e.inputs.forEach(function(t){var i=t.value.visit(n._valueConverter);n.allocateBindingSlots(i),n.updateInstruction(e.sourceSpan,Qn.elementProperty,function(){return[Pe(r),Pe(t.name),n.convertPropertyBinding(l,i)]})});var d=new t(this.constantPool,this._bindingScope,this.level+1,a,this.i18n,r,s,[],this.directiveMatcher,this.directives,this.pipeTypeByName,this.pipes,this._namespace,this.fileBasedI18nSuffix,this.i18nUseExternalIds);this._nestedTemplateFns.push(function(){var t,r=d.buildTemplateFunction(e.children,e.variables,n._ngContentSelectors.length+n._ngContentSelectorsOffset,e.i18n);n.constantPool.statements.push(r.toDeclStmt(s,null)),d._hasNgContent&&(n._hasNgContent=!0,(t=n._ngContentSelectors).push.apply(t,Object(o.__spread)(d._ngContentSelectors)))}),this.creationInstruction(e.sourceSpan,Qn.templateCreate,function(){return u.splice(2,0,Pe(d.getConstCount()),Pe(d.getVarCount())),vi(u)}),e.outputs.forEach(function(t){n.creationInstruction(t.sourceSpan,Qn.listener,n.prepareListenerParameter("ng_template",t,r))})},t.prototype.visitBoundText=function(t){var e=this;if(this.i18n){var n=t.value.visit(this._valueConverter);return this.allocateBindingSlots(n),void(n instanceof To&&(this.i18n.appendBoundText(t.i18n),this.i18nAppendBindings(n.expressions)))}var r=this.allocateDataSlot();this.creationInstruction(t.sourceSpan,Qn.text,[Pe(r)]);var i=t.value.visit(this._valueConverter);this.allocateBindingSlots(i),this.updateInstruction(t.sourceSpan,Qn.textBinding,function(){return[Pe(r),e.convertPropertyBinding(Ee(pi),i)]})},t.prototype.visitText=function(t){this.i18n||this.creationInstruction(t.sourceSpan,Qn.text,[Pe(this.allocateDataSlot()),Pe(t.value)])},t.prototype.visitIcu=function(t){var e=!1;this.i18n||(e=!0,this.i18nStart(null,t.i18n,!0));var n=this.i18n,r=this.i18nBindProps(t.vars),i=this.i18nBindProps(t.placeholders),o=t.i18n,a=function(t){return yl(null,Qn.i18nPostprocess,[t,Hr(r,!0)])};if(ei(n.meta))this.i18nTranslate(o,i,n.ref,a);else{var s=this.i18nTranslate(o,i,void 0,a);n.appendIcu(ri(o).name,s)}return e&&this.i18nEnd(null,!0),null},t.prototype.allocateDataSlot=function(){return this._dataIndex++},t.prototype.getConstCount=function(){return this._dataIndex},t.prototype.getVarCount=function(){return this._pureFunctionSlots},t.prototype.bindingContext=function(){return""+this._bindingContext++},t.prototype.instructionFn=function(t,e,n,r,i){void 0===i&&(i=!1),t[i?"unshift":"push"](function(){var t=Array.isArray(r)?r:r();return yl(e,n,t).toStmt()})},t.prototype.processStylingInstruction=function(t,e,n){var r=this;if(e){var i=function(){return e.buildParams(function(e){return r.convertPropertyBinding(t,e,!0)})};n?this.creationInstruction(e.sourceSpan,e.reference,i):this.updateInstruction(e.sourceSpan,e.reference,i)}},t.prototype.creationInstruction=function(t,e,n,r){this.instructionFn(this._creationCodeFns,t,e,n||[],r)},t.prototype.updateInstruction=function(t,e,n){this.instructionFn(this._updateCodeFns,t,e,n||[])},t.prototype.allocatePureFunctionSlots=function(t){var e=this._pureFunctionSlots;return this._pureFunctionSlots+=t,e},t.prototype.allocateBindingSlots=function(t){this._bindingSlots+=t instanceof To?t.expressions.length:1},t.prototype.convertExpressionBinding=function(t,e){var n=ia(this,t,e,this.bindingContext(),Jo.TrySimple).currValExpr;return Se(Qn.bind).callFn([n])},t.prototype.convertPropertyBinding=function(t,e,n){var r,i=e instanceof To?wl:function(){return Ge("Unexpected interpolation")},a=ia(this,t,e,this.bindingContext(),Jo.TrySimple,i);(r=this._tempVariables).push.apply(r,Object(o.__spread)(a.stmts));var s=a.currValExpr;return e instanceof To||n?s:Se(Qn.bind).callFn([s])},t.prototype.matchDirectives=function(t,e){var n,r,i,o=this;if(this.directiveMatcher){var a=(n=t,r=Ci(e),(i=new j).setElement(n),Object.getOwnPropertyNames(r).forEach(function(t){var e=r[t];i.addAttribute(t,e),"class"===t.toLowerCase()&&e.trim().split(/\s+/).forEach(function(t){return i.addClassName(t)})}),i);this.directiveMatcher.match(a,function(t,e){o.directives.add(e)})}},t.prototype.prepareSyntheticAndSelectOnlyAttrs=function(t,e,n){var r=[],i=[],o=new Set;function a(t,e){"string"==typeof t?o.has(t)||(r.push(Pe(t)),void 0!==e&&r.push(e),o.add(t)):r.push(Pe(t))}if(t.length){var s=yi("");t.forEach(function(t){if(4===t.type){var e=t.value;if(e instanceof No){var n=e.ast;n instanceof Eo&&void 0===n.value&&a(Pi(t.name),s)}}else i.push(t)})}return n&&n.populateInitialStylingAttrs(r),(i.length||e.length)&&(a(3),i.forEach(function(t){return a(t.name)}),e.forEach(function(t){var e,n;a(1===t.type?(e=t.name,n=e.indexOf("."),(e=n>0?e.substring(0,n):e).charAt(0)!==Ii&&(e=Ii+e),e):t.name)})),r},t.prototype.toAttrsParam=function(t){return t.length>0?this.constantPool.getConstLiteral(ke(t),!0):ee},t.prototype.prepareRefsParameter=function(t){var e=this;if(!t||0===t.length)return ee;var n=Un(t.map(function(t){var n=e.allocateDataSlot(),r=e._bindingScope.freshReferenceName(),i=e.level,o=Ee(r);return e._bindingScope.set(i,t.name,o,0,function(t,e){var r=e>0?[_l(e).toStmt()]:[],i=o.set(Se(Qn.reference).callFn([Pe(n)]));return r.concat(i.toConstDecl())},!0),[t.name,t.value]}));return this.constantPool.getConstLiteral(yi(n),!0)},t.prototype.prepareListenerParameter=function(t,e,n){var r,i=this,a=e.name;1===e.type?(r=Ni(a,e.phase),a=Ri(a,e.phase)):r=wn(a);var s=wn(t),u=this.templateName+"_"+s+"_"+r+"_"+n+"_listener";return function(){var t=i._bindingScope.nestedScope(i._bindingScope.bindingLevel),n=Zo(t,Ee(pi),e.handler,"b",function(){return Ge("Unexpected interpolation")}),r=Object(o.__spread)(t.restoreViewStatement(),t.variableDeclarations(),n.render3Stmts),s=De([new qt("$event",mt)],r,gt,null,u);return[Pe(a),s]}},t}(),hl=function(t){function e(e,n,r,i){var o=t.call(this)||this;return o.constantPool=e,o.allocateSlot=n,o.allocatePureFunctionSlots=r,o.definePipe=i,o._pipeBindExprs=[],o}return Object(o.__extends)(e,t),e.prototype.visitPipe=function(t,e){var n,r,i=this.allocateSlot(),a="PIPE:"+i,s=this.allocatePureFunctionSlots(2+t.args.length),u=new bo(t.span,new yo(t.span),a),c=(n=t.args,{identifier:(r=ml[n.length])||Qn.pipeBindV,isVarLength:!r}),l=c.identifier,d=c.isVarLength;this.definePipe(t.name,a,i,Se(l));var p=Object(o.__spread)([t.exp],t.args),f=d?this.visitAll([new So(t.span,p)]):this.visitAll(p),h=new Ro(t.span,u,Object(o.__spread)([new Eo(t.span,i),new Eo(t.span,s)],f));return this._pipeBindExprs.push(h),h},e.prototype.updatePipeSlotOffsets=function(t){this._pipeBindExprs.forEach(function(e){e.args[1].value+=t})},e.prototype.visitLiteralArray=function(t,e){var n=this;return new ha(t.span,this.visitAll(t.expressions),function(t){var e=ke(t);return t.every(function(t){return t.isConstant()})?n.constantPool.getConstLiteral(e,!0):vl(n.constantPool,e,n.allocatePureFunctionSlots)})},e.prototype.visitLiteralMap=function(t,e){var n=this;return new ha(t.span,this.visitAll(t.values),function(e){var r=Ae(e.map(function(e,n){return{key:t.keys[n].key,value:e,quoted:t.keys[n].quoted}}));return e.every(function(t){return t.isConstant()})?n.constantPool.getConstLiteral(r,!0):vl(n.constantPool,r,n.allocatePureFunctionSlots)})},e}(zo),ml=[Qn.pipeBind1,Qn.pipeBind2,Qn.pipeBind3,Qn.pipeBind4];var gl=[Qn.pureFunction0,Qn.pureFunction1,Qn.pureFunction2,Qn.pureFunction3,Qn.pureFunction4,Qn.pureFunction5,Qn.pureFunction6,Qn.pureFunction7,Qn.pureFunction8];function yl(t,e,n){return Se(e,null,t).callFn(n,t)}function _l(t){return Se(Qn.nextContext).callFn(t>1?[Pe(t)]:[])}function vl(t,e,n){var r=t.getLiteralFactory(e),i=r.literalFactory,a=r.literalFactoryArguments,s=n(1+a.length);a.length>0||Ge("Expected arguments to a literal factory function");var u,c={identifier:(u=gl[a.length])||Qn.pureFunctionV,isVarLength:!u},l=c.identifier,d=c.isVarLength,p=[Pe(s),i];return d?p.push(ke(a)):p.push.apply(p,Object(o.__spread)(a)),Se(l).callFn(p)}var bl="$$shared_ctx$$",xl=function(){function t(t,e){void 0===t&&(t=0),void 0===e&&(e=null),this.bindingLevel=t,this.parent=e,this.map=new Map,this.referenceNameIndex=0,this.restoreViewVariable=null}return Object.defineProperty(t,"ROOT_SCOPE",{get:function(){return t._ROOT_SCOPE||(t._ROOT_SCOPE=(new t).set(0,"$event",Ee("$event"))),t._ROOT_SCOPE},enumerable:!0,configurable:!0}),t.prototype.get=function(t){for(var e=this;e;){var n=e.map.get(t);if(null!=n)return e!==this&&(n={retrievalLevel:n.retrievalLevel,lhs:n.lhs,declareLocalCallback:n.declareLocalCallback,declare:!1,priority:n.priority,localRef:n.localRef},this.map.set(t,n),this.maybeGenerateSharedContextVar(n),this.maybeRestoreView(n.retrievalLevel,n.localRef)),n.declareLocalCallback&&!n.declare&&(n.declare=!0),n.lhs;e=e.parent}return 0===this.bindingLevel?null:this.getComponentProperty(t)},t.prototype.set=function(t,e,n,r,i,o){return void 0===r&&(r=0),!this.map.has(e)||Ge("The name "+e+" is already defined in scope to be "+this.map.get(e)),this.map.set(e,{retrievalLevel:t,lhs:n,declare:!1,declareLocalCallback:i,priority:r,localRef:o||!1}),this},t.prototype.getLocal=function(t){return this.get(t)},t.prototype.nestedScope=function(e){var n=new t(e,this);return e>0&&n.generateSharedContextVar(0),n},t.prototype.getSharedContextName=function(t){var e=this.map.get(bl+t);return e&&e.declare?e.lhs:null},t.prototype.maybeGenerateSharedContextVar=function(t){if(1===t.priority){var e=this.map.get(bl+t.retrievalLevel);e?e.declare=!0:this.generateSharedContextVar(t.retrievalLevel)}},t.prototype.generateSharedContextVar=function(t){var e=Ee(pi+this.freshReferenceName());this.map.set(bl+t,{retrievalLevel:t,lhs:e,declareLocalCallback:function(t,n){return[e.set(_l(n)).toConstDecl()]},declare:!1,priority:2,localRef:!1})},t.prototype.getComponentProperty=function(t){var e=this.map.get(bl+0);return e.declare=!0,this.maybeRestoreView(0,!1),e.lhs.prop(t)},t.prototype.maybeRestoreView=function(t,e){this.isListenerScope()&&(t<this.bindingLevel||e)&&(this.parent.restoreViewVariable||(this.parent.restoreViewVariable=Ee(this.parent.freshReferenceName())),this.restoreViewVariable=this.parent.restoreViewVariable)},t.prototype.restoreViewStatement=function(){return this.restoreViewVariable?[yl(null,Qn.restoreView,[this.restoreViewVariable]).toStmt()]:[]},t.prototype.viewSnapshotStatements=function(){var t=yl(null,Qn.getCurrentView,[]);return this.restoreViewVariable?[this.restoreViewVariable.set(t).toConstDecl()]:[]},t.prototype.isListenerScope=function(){return this.parent&&this.parent.bindingLevel===this.bindingLevel},t.prototype.variableDeclarations=function(){var t=this,e=0;return Array.from(this.map.values()).filter(function(t){return t.declare}).sort(function(t,e){return e.retrievalLevel-t.retrievalLevel||e.priority-t.priority}).reduce(function(n,r){var i=t.bindingLevel-r.retrievalLevel,o=r.declareLocalCallback(t,i-e);return e=i,n.concat(o)},[])},t.prototype.freshReferenceName=function(){for(var t=this;t.parent;)t=t.parent;return"_r"+t.referenceNameIndex++},t}();function wl(t){switch((t=t.slice(1)).length){case 3:return Se(Qn.interpolation1).callFn(t);case 5:return Se(Qn.interpolation2).callFn(t);case 7:return Se(Qn.interpolation3).callFn(t);case 9:return Se(Qn.interpolation4).callFn(t);case 11:return Se(Qn.interpolation5).callFn(t);case 13:return Se(Qn.interpolation6).callFn(t);case 15:return Se(Qn.interpolation7).callFn(t);case 17:return Se(Qn.interpolation8).callFn(t)}return t.length>=19&&t.length%2==1||Ge("Invalid interpolation argument length "+t.length),Se(Qn.interpolationV).callFn([ke(t)])}function jl(t,e,n){void 0===n&&(n={});var r=n.interpolationConfig,i=n.preserveWhitespaces,o=Ol(r),a=(new Hu).parse(t,e,!0,r);if(a.errors&&a.errors.length>0)return{errors:a.errors,nodes:[]};var s=a.rootNodes;s=_u(new cl(r,!i),s),i||(s=_u(new Qu,s),s=_u(new cl(r,!1),s));var u=Uc(s,o),c=u.nodes,l=u.errors;return l&&l.length>0?{errors:l,nodes:[]}:{nodes:c}}function Ol(t){return void 0===t&&(t=Hi),new uc(new uu(new Ks),t,new sc,null,[])}function Cl(t){return t instanceof pc||t instanceof fc||t instanceof wc}function El(t){return t.every(Cl)}var Sl=[],Ml=/attr\.([^\]]+)/;function Tl(t){return t.substring(0,5).toLowerCase()}function kl(t,e,n){var r=new Oi;r.set("type",t.type),r.set("selectors",yi(nt(t.selector)));var i=Ei({name:t.name,type:t.type,deps:t.deps,injectFn:Qn.directiveInject});r.set("factory",i.factory),r.set("contentQueries",function(t,e){if(t.queries.length){var n=t.queries.map(function(t){var n=Ll(t,e,null);return Se(Qn.registerContentQuery).callFn([n,Ee("dirIndex")]).toStmt()}),r=t.name,i=[new qt("dirIndex",_t)];return De(i,n,gt,null,r?r+"_ContentQueries":null)}return null}(t,e)),r.set("contentQueriesRefresh",function(t){if(t.queries.length>0){var e=[],n=t.name,r=[new qt("dirIndex",_t),new qt("queryStartIndex",_t)],i=Ee("instance"),o=hi(e,di);return e.push(i.set(Se(Qn.load).callFn([Ee("dirIndex")])).toDeclStmt(gt,[It.Final])),t.queries.forEach(function(t,n){var r=Ee("queryStartIndex"),a=Se(Qn.loadQueryList).callFn([n>0?r.plus(Pe(n)):r]),s=o().set(a),u=Se(Qn.queryRefresh).callFn([s]),c=i.prop(t.propertyName).set(t.first?o().prop("first"):o()),l=u.and(c);e.push(l.toStmt())}),De(r,e,gt,null,n?n+"_ContentQueriesRefresh":null)}return null}(t));for(var a=Object.keys(t.host.properties).filter(function(t){var e=Tl(t);return"style"!==e&&"class"!==e}).length,s=Ee("elIndex"),u=Ee(pi),c=new Us(s,u),l={},d=Object.getOwnPropertyNames(t.host.attributes),p=0;p<d.length;p++){var f=d[p],h=t.host.attributes[f];switch(f){case"style":c.registerStyleAttr(h);break;case"class":c.registerClassAttr(h);break;default:l[f]=h}}return r.set("attributes",function(t){var e,n,r=[];try{for(var i=Object(o.__values)(Object.getOwnPropertyNames(t)),a=i.next();!a.done;a=i.next()){var s=a.value,u=t[s];r.push(Pe(s),Pe(u))}}catch(t){e={error:t}}finally{try{a&&!a.done&&(n=i.return)&&n.call(i)}finally{if(e)throw e.error}}if(r.length>0)return ke(r);return null}(l)),r.set("hostBindings",function(t,e,n,r,i,a,s){var u,c,l=[],d=[],p=s,f=t.typeSourceSpan,h=(g=t,{hostAttributes:g.host.attributes,hostListeners:g.host.listeners,hostProperties:g.host.properties}),m=i.createDirectiveHostEventAsts(h,f);var g;if(m&&m.length){var y=(_=n,v=t,m.map(function(t){var e=Zo(null,_,t.handler,"b",function(){return Ge("Unexpected interpolation")}),n=t.name&&wn(t.name),r=n;1===t.type&&(r=Ni(n,t.targetOrPhase),n=Ri(n,t.targetOrPhase));var i=v.name,a=i&&n?i+"_"+r+"_HostBindingHandler":null,s=De([new qt("$event",mt)],Object(o.__spread)(e.render3Stmts),gt,null,a);return Se(Qn.listener).callFn([Pe(n),s]).toStmt()}));l.push.apply(l,Object(o.__spread)(y))}var _,v;var b=i.createBoundHostProperties(h,f),x=function(t,e){return ia(null,t,e,"b",Jo.TrySimple,function(){return Ge("Unexpected interpolation")})};if(b){var w=new hl(a,function(){return Ge("Unexpected node")},function(t){var e=p;return p+=t,e},function(){return Ge("Unexpected pipe")});try{for(var j=Object(o.__values)(b),O=j.next();!O.done;O=j.next()){var C=O.value,E=C.name,S=Tl(E);if("style"===S){var M=Gl(E),T=M.propertyName,k=M.unit;r.registerStyleInput(T,C.expression,k,C.sourceSpan)}else if("class"===S)r.registerClassInput(Gl(E).propertyName,C.expression,C.sourceSpan);else{var A=C.expression.visit(w),D=x(n,A),I=Ul(C),P=I.bindingName,R=I.instruction,N=I.extraParams,F=[e,Pe(P),Se(Qn.bind).callFn([D.currValExpr])];d.push.apply(d,Object(o.__spread)(D.stmts)),d.push(Se(R).callFn(F.concat(N)).toStmt())}}}catch(t){u={error:t}}finally{try{O&&!O.done&&(c=j.return)&&c.call(j)}finally{if(u)throw u.error}}if(r.hasBindingsOrInitialValues()){var L=r.buildDirectiveHostAttrsInstruction(null,a);L&&l.push(Hl(L,n,x));var B=r.buildElementStylingInstruction(null,a);B&&l.push(Hl(B,n,x)),r.buildUpdateLevelInstructions(w).forEach(function(t){d.push(Hl(t,n,x))})}}p&&l.unshift(Se(Qn.allocHostVars).callFn([Pe(p)]).toStmt());if(l.length>0||d.length>0){var V=t.name?t.name+"_HostBindings":null,z=[];return l.length>0&&z.push(pl(1,l)),d.length>0&&z.push(pl(2,d)),De([new qt(fi,_t),new qt(pi,null),new qt(e.name,_t)],z,gt,null,V)}return null}(t,s,u,c,n,e,a)),r.set("inputs",_i(t.inputs,!0)),r.set("outputs",_i(t.outputs)),null!==t.exportAs&&r.set("exportAs",Pe(t.exportAs)),{definitionMap:r,statements:i.statements}}function Al(t,e){var n=[],r=e.providers,i=e.viewProviders;if(r||i){var o=[r||new Qt([])];i&&o.push(i),n.push(Se(Qn.ProvidersFeature).callFn(o))}e.usesInheritance&&n.push(Se(Qn.InheritDefinitionFeature)),e.lifecycle.usesOnChanges&&n.push(Se(Qn.NgOnChangesFeature)),n.length&&t.set("features",ke(n))}function Dl(t,e,n){var r=kl(t,e,n),i=r.definitionMap,o=r.statements;Al(i,t);var a=Se(Qn.defineDirective).callFn([i.toLiteralMap()]);(t.selector||"").replace(/\n/g,"");return{expression:a,type:zl(t,Qn.DirectiveDefWithMeta),statements:o}}function Il(t){var e=new Oi;if(t.inputs){var n=t.inputs,r=Object.keys(n).map(function(t){var e=n[t];return{key:t,value:Array.isArray(e)?ke(e.map(function(t){return Pe(t)})):Pe(e),quoted:!1}});e.set("inputs",Ae(r))}if(t.outputs){var i=t.outputs,o=Object.keys(i).map(function(t){return{key:t,value:Pe(i[t]),quoted:!1}});e.set("outputs",Ae(o))}return{expression:Se(Qn.defineBase).callFn([e.toLiteralMap()]),type:new pt(Se(Qn.BaseDef))}}function Pl(t,e,n){var r,i,a=kl(t,e,n),s=a.definitionMap,u=a.statements;Al(s,t);var c=t.selector&&j.parse(t.selector),l=c&&c[0];if(l){var d=l.getAttrs();d.length&&s.set("attrs",e.getConstLiteral(ke(d.map(function(t){return Pe(null!=t?t:void 0)})),!0))}var p=null;if(t.directives.length>0){var f=new O;try{for(var h=Object(o.__values)(t.directives),m=h.next();!m.done;m=h.next()){var g=m.value,y=g.selector,v=g.expression;f.addSelectables(j.parse(y),v)}}catch(t){r={error:t}}finally{try{m&&!m.done&&(i=h.return)&&i.call(h)}finally{if(r)throw r.error}}p=f}t.viewQueries.length&&s.set("viewQuery",function(t,e){for(var n=[],r=[],i=hi(r,di),o=0;o<t.viewQueries.length;o++){var a=t.viewQueries[o],s=Ll(a,e,o);n.push(s.toStmt());var u=i(),c=Se(Qn.load).callFn([Pe(o)]),l=Se(Qn.queryRefresh).callFn([u.set(c)]),d=Ee(pi).prop(a.propertyName).set(a.first?u.prop("first"):u);r.push(l.and(d).toStmt())}var p=t.name?t.name+"_Query":null;return De([new qt(fi,_t),new qt(pi,null)],[pl(1,n),pl(2,r)],gt,null,p)}(t,e));var x,w,C,E,S=t.name,M=S?S+"_Template":null,T=new Set,k=new Set,A=t.changeDetection,D=t.template,I=new fl(e,xl.ROOT_SCOPE,0,S,null,null,M,t.viewQueries,p,T,t.pipes,k,Qn.namespaceHTML,t.relativeContextFilePath,t.i18nUseExternalIds),P=I.buildTemplateFunction(D.nodes,[]);if(s.set("consts",Pe(I.getConstCount())),s.set("vars",Pe(I.getVarCount())),s.set("template",P),T.size){var R=ke(Array.from(T));t.wrapDirectivesAndPipesInClosure&&(R=De([],[new ae(R)])),s.set("directives",R)}if(k.size){var N=ke(Array.from(k));t.wrapDirectivesAndPipesInClosure&&(N=De([],[new ae(N)])),s.set("pipes",N)}if(null===t.encapsulation&&(t.encapsulation=_.Emulated),t.styles&&t.styles.length){var F=(t.encapsulation==_.Emulated?(x=t.styles,w=Rs,C=Ps,E=new ss,x.map(function(t){return E.shimCssText(t,w,C)})):t.styles).map(function(t){return Pe(t)});s.set("styles",ke(F))}else t.encapsulation===_.Emulated&&(t.encapsulation=_.None);t.encapsulation!==_.Emulated&&s.set("encapsulation",Pe(t.encapsulation)),null!==t.animations&&s.set("data",Ae([{key:"animation",value:t.animations,quoted:!1}])),null!=A&&A!==b.Default&&s.set("changeDetection",Pe(A));(t.selector||"").replace(/\n/g,"");return{expression:Se(Qn.defineComponent).callFn([s.toLiteralMap()]),type:zl(t,Qn.ComponentDefWithMeta),statements:u}}function Rl(t,e,n,r,i,a,s){var u=En(e.type);u||Ge("Cannot resolver the name of "+e.type);var c,l,d,p=t.constantPool.propertyNameOf(2),f=e.toSummary(),h=Pl(Object(o.__assign)({},Nl(e,t,r),{selector:e.selector,template:{nodes:n.nodes},directives:[],pipes:(c=s,l=t,d=Array.from(c).map(function(t){var e=Object(o.__read)(t,2),n=e[0],r=e[1];return[n,l.importExpr(r)]}),new Map(d)),viewQueries:Fl(e.viewQueries,t),wrapDirectivesAndPipesInClosure:!1,styles:f.template&&f.template.styles||Sl,encapsulation:f.template&&f.template.encapsulation||_.Emulated,interpolation:Hi,animations:null,viewProviders:e.viewProviders.length>0?new Tt(e.viewProviders):null,relativeContextFilePath:"",i18nUseExternalIds:!0}),t.constantPool,i);t.statements.push(new de(u,null,[new ue(p,gt,[It.Static],h.expression)],[],new ce(null,[],[]),[]))}function Nl(t,e,n){var r=t.toSummary(),i=En(t.type);return i||Ge("Cannot resolver the name of "+t.type),{name:i,type:e.importExpr(t.type.reference),typeArgumentCount:0,typeSourceSpan:as(t.isComponent?"Component":"Directive",t.type),selector:t.selector,deps:Mi(t.type,e,n),queries:Fl(t.queries,e),lifecycle:{usesOnChanges:t.type.lifecycleHooks.some(function(t){return t==pa.OnChanges})},host:{attributes:t.hostAttributes,listeners:r.hostListeners,properties:r.hostProperties},inputs:t.inputs,outputs:t.outputs,usesInheritance:!1,exportAs:null,providers:t.providers.length>0?new Tt(t.providers):null}}function Fl(t,e){return t.map(function(t){var n=null;return t.read&&t.read.identifier&&(n=e.importExpr(t.read.identifier.reference)),{propertyName:t.propertyName,first:t.first,predicate:function(t,e){if(t.length>1||1==t.length&&t[0].value){var n=t.map(function(t){return t.value});return n.some(function(t){return!t})&&Ge("Found a type among the string selectors expected"),e.constantPool.getConstLiteral(ke(n.map(function(t){return Pe(t)})))}if(1==t.length){var r=t[0];if(r.identifier)return e.importExpr(r.identifier.reference)}return Ge("Unexpected query form"),te}(t.selectors,e),descendants:t.descendants,read:n}})}function Ll(t,e,n){var r=function(t,e){if(Array.isArray(t.predicate)){var n=[];return t.predicate.forEach(function(t){var e=t.split(",").map(function(t){return Pe(t.trim())});n.push.apply(n,Object(o.__spread)(e))}),e.getConstLiteral(ke(n),!0)}return t.predicate}(t,e),i=[Pe(n,gt),r,Pe(t.descendants)];return t.read&&i.push(t.read),Se(Qn.query).callFn(i)}function Bl(t){return Te(Pe(t))}function Vl(t){return Te(Ae(Object.keys(t).map(function(e){return{key:e,value:Pe(Array.isArray(t[e])?t[e][0]:t[e]),quoted:!0}})))}function zl(t,e){var n,r=(t.selector||"").replace(/\n/g,"");return Te(Se(e,[Di(t.type,t.typeArgumentCount),Bl(r),null!==t.exportAs?Bl(t.exportAs):vt,Vl(t.inputs),Vl(t.outputs),(n=t.queries.map(function(t){return t.propertyName}),n.length>0?Te(ke(n.map(function(t){return Pe(t)}))):vt)]))}function Hl(t,e,n){var r=t.buildParams(function(t){return n(e,t).currValExpr});return Se(t.reference,null,t.sourceSpan).callFn(r,t.sourceSpan).toStmt()}function Ul(t){var e,n=t.name,r=[],i=n.match(Ml);return i?(n=i[1],e=Qn.elementAttribute):(t.isAnimation?(n=Pi(n),e=Qn.componentHostSyntheticProperty):e=Qn.elementProperty,r.push(Pe(null),Pe(!0))),{bindingName:n,instruction:e,extraParams:r}}var Wl=/^(?:\[([^\]]+)\])|(?:\(([^\)]+)\))$/;function ql(t){var e={},n={},r={};return Object.keys(t).forEach(function(i){var o=t[i],a=i.match(Wl);null===a?e[i]=o:null!=a[1]?r[a[1]]=o:null!=a[2]&&(n[a[2]]=o)}),{attributes:e,listeners:n,properties:r}}function Gl(t){var e="",n="",r=t.indexOf(".");if(r>0){var i=t.lastIndexOf(".");i!==r?(e=t.substring(i+1,t.length),n=t.substring(r+1,i)):n=t.substring(r+1,t.length)}return{propertyName:n,unit:e}}var Yl=function(){function t(){this.R3ResolvedDependencyType=wi,this.elementSchemaRegistry=new sc}return t.prototype.compilePipe=function(t,e,n){var r=lo({name:n.name,type:new Tt(n.type),deps:rd(n.deps),pipeName:n.pipeName,pure:n.pure});return ao(r.expression,t,e,r.statements)},t.prototype.compileInjectable=function(t,e,n){var r,i=Fi({name:n.name,type:new Tt(n.type),typeArgumentCount:n.typeArgumentCount,providedIn:(r=n.providedIn,null==r||"string"==typeof r?new Lt(r):new Tt(r)),useClass:ed(n,Kl),useFactory:ed(n,Xl),useValue:ed(n,Ql),useExisting:ed(n,Zl),ctorDeps:rd(n.ctorDeps),userDeps:rd(n.userDeps)||void 0});return ao(i.expression,t,e,i.statements)},t.prototype.compileInjector=function(t,e,n){var r=uo({name:n.name,type:new Tt(n.type),deps:rd(n.deps),providers:new Tt(n.providers),imports:new Tt(n.imports)});return ao(r.expression,t,e,r.statements)},t.prototype.compileNgModule=function(t,e,n){return ao(so({type:new Tt(n.type),bootstrap:n.bootstrap.map($l),declarations:n.declarations.map($l),imports:n.imports.map($l),exports:n.exports.map($l),emitInline:!0}).expression,t,e,[])},t.prototype.compileDirective=function(t,e,n){var r=new dn,i=Ol(),a=Dl(td(n),r,i),s=Object(o.__spread)(r.statements,a.statements);return ao(a.expression,t,e,s)},t.prototype.compileComponent=function(t,e,n){var r=new dn,i=n.interpolation?zi.fromArray(n.interpolation):Hi,a=jl(n.template,e,{preserveWhitespaces:n.preserveWhitespaces||!1,interpolationConfig:i});if(void 0!==a.errors){var s=a.errors.map(function(t){return t.toString()}).join(", ");throw new Error("Errors during JIT compilation of template for "+n.name+": "+s)}var u=Pl(Object(o.__assign)({},n,td(n),{selector:n.selector||this.elementSchemaRegistry.getDefaultComponentElementName(),template:a,viewQueries:n.viewQueries.map(Jl),wrapDirectivesAndPipesInClosure:!1,styles:n.styles||[],encapsulation:n.encapsulation,interpolation:i,changeDetection:n.changeDetection,animations:null!=n.animations?new Tt(n.animations):null,viewProviders:null!=n.viewProviders?new Tt(n.viewProviders):null,relativeContextFilePath:"",i18nUseExternalIds:!0}),r,Ol(i)),c=Object(o.__spread)(r.statements,u.statements);return ao(u.expression,t,e,c)},t}(),Kl=Object.keys({useClass:null})[0],Xl=Object.keys({useFactory:null})[0],Ql=Object.keys({useValue:null})[0],Zl=Object.keys({useExisting:null})[0],$l=function(t){var e=new Tt(t);return{value:e,type:e}};function Jl(t){return Object(o.__assign)({},t,{predicate:Array.isArray(t.predicate)?t.predicate:new Tt(t.predicate),read:t.read?new Tt(t.read):null})}function td(t){var e=id(t.inputs||[]),n=id(t.outputs||[]),r=t.propMetadata,i={},a={},s=function(t){r.hasOwnProperty(t)&&r[t].forEach(function(e){"Input"===e.ngMetadataName?i[t]=e.bindingPropertyName?[e.bindingPropertyName,t]:t:"Output"===e.ngMetadataName&&(a[t]=e.bindingPropertyName||t)})};for(var u in r)s(u);return Object(o.__assign)({},t,{typeSourceSpan:null,type:new Tt(t.type),deps:rd(t.deps),host:function(t,e){var n=ql(t||{}),r=n.attributes,i=n.listeners,o=n.properties,a=function(t){e.hasOwnProperty(t)&&e[t].forEach(function(e){"HostBinding"===e.ngMetadataName?o[e.hostPropertyName||t]=t:"HostListener"===e.ngMetadataName&&(i[e.eventName||t]=t+"("+(e.args||[]).join(",")+")")})};for(var s in e)a(s);return{attributes:r,listeners:i,properties:o}}(t.host,t.propMetadata),inputs:Object(o.__assign)({},e,i),outputs:Object(o.__assign)({},n,a),queries:t.queries.map(Jl),providers:null!=t.providers?new Tt(t.providers):null})}function ed(t,e){return t.hasOwnProperty(e)?new Tt(t[e]):void 0}function nd(t){return{token:null===t.token?new Lt(null):t.resolved===wi.Attribute?new Lt(t.token):new Tt(t.token),resolved:t.resolved,host:t.host,optional:t.optional,self:t.self,skipSelf:t.skipSelf}}function rd(t){return null==t?null:t.map(nd)}function id(t){return t.reduce(function(t,e){var n=Object(o.__read)(e.split(",").map(function(t){return t.trim()}),2),r=n[0],i=n[1];return t[r]=i||r,t},{})}function od(t){(t.ng||(t.ng={})).ɵcompilerFacade=new Yl}var ad,sd,ud,cd=new on("7.2.0"),ld=function(){function t(t,e,n){this.value=t,this.ngContentIndex=e,this.sourceSpan=n}return t.prototype.visit=function(t,e){return t.visitText(this,e)},t}(),dd=function(){function t(t,e,n){this.value=t,this.ngContentIndex=e,this.sourceSpan=n}return t.prototype.visit=function(t,e){return t.visitBoundText(this,e)},t}(),pd=function(){function t(t,e,n){this.name=t,this.value=e,this.sourceSpan=n}return t.prototype.visit=function(t,e){return t.visitAttr(this,e)},t}(),fd=((ad={})[4]=4,ad[1]=1,ad[2]=2,ad[0]=0,ad[3]=3,ad),hd=function(){function t(t,e,n,r,i,o){this.name=t,this.type=e,this.securityContext=n,this.value=r,this.unit=i,this.sourceSpan=o,this.isAnimation=4===this.type}return t.fromBoundProperty=function(e){var n=fd[e.type];return new t(e.name,n,e.securityContext,e.value,e.unit,e.sourceSpan)},t.prototype.visit=function(t,e){return t.visitElementProperty(this,e)},t}(),md=function(){function t(e,n,r,i,o){this.name=e,this.target=n,this.phase=r,this.handler=i,this.sourceSpan=o,this.fullName=t.calcFullName(this.name,this.target,this.phase),this.isAnimation=!!this.phase}return t.calcFullName=function(t,e,n){return e?e+":"+t:n?"@"+t+"."+n:t},t.fromParsedEvent=function(e){var n=0===e.type?e.targetOrPhase:null,r=1===e.type?e.targetOrPhase:null;return new t(e.name,n,r,e.handler,e.sourceSpan)},t.prototype.visit=function(t,e){return t.visitEvent(this,e)},t}(),gd=function(){function t(t,e,n,r){this.name=t,this.value=e,this.originalValue=n,this.sourceSpan=r}return t.prototype.visit=function(t,e){return t.visitReference(this,e)},t}(),yd=function(){function t(t,e,n){this.name=t,this.value=e,this.sourceSpan=n}return t.fromParsedVariable=function(e){return new t(e.name,e.value,e.sourceSpan)},t.prototype.visit=function(t,e){return t.visitVariable(this,e)},t}(),_d=function(){function t(t,e,n,r,i,o,a,s,u,c,l,d,p){this.name=t,this.attrs=e,this.inputs=n,this.outputs=r,this.references=i,this.directives=o,this.providers=a,this.hasViewContainer=s,this.queryMatches=u,this.children=c,this.ngContentIndex=l,this.sourceSpan=d,this.endSourceSpan=p}return t.prototype.visit=function(t,e){return t.visitElement(this,e)},t}(),vd=function(){function t(t,e,n,r,i,o,a,s,u,c,l){this.attrs=t,this.outputs=e,this.references=n,this.variables=r,this.directives=i,this.providers=o,this.hasViewContainer=a,this.queryMatches=s,this.children=u,this.ngContentIndex=c,this.sourceSpan=l}return t.prototype.visit=function(t,e){return t.visitEmbeddedTemplate(this,e)},t}(),bd=function(){function t(t,e,n,r){this.directiveName=t,this.templateName=e,this.value=n,this.sourceSpan=r}return t.prototype.visit=function(t,e){return t.visitDirectiveProperty(this,e)},t}(),xd=function(){function t(t,e,n,r,i,o){this.directive=t,this.inputs=e,this.hostProperties=n,this.hostEvents=r,this.contentQueryStartId=i,this.sourceSpan=o}return t.prototype.visit=function(t,e){return t.visitDirective(this,e)},t}(),wd=function(){function t(t,e,n,r,i,o,a,s){this.token=t,this.multiProvider=e,this.eager=n,this.providers=r,this.providerType=i,this.lifecycleHooks=o,this.sourceSpan=a,this.isModule=s}return t.prototype.visit=function(t,e){return null},t}();(ud=sd||(sd={}))[ud.PublicService=0]="PublicService",ud[ud.PrivateService=1]="PrivateService",ud[ud.Component=2]="Component",ud[ud.Directive=3]="Directive",ud[ud.Builtin=4]="Builtin";var jd=function(){function t(t,e,n){this.index=t,this.ngContentIndex=e,this.sourceSpan=n}return t.prototype.visit=function(t,e){return t.visitNgContent(this,e)},t}(),Od=function(){function t(){}return t.prototype.visitNgContent=function(t,e){},t.prototype.visitEmbeddedTemplate=function(t,e){},t.prototype.visitElement=function(t,e){},t.prototype.visitReference=function(t,e){},t.prototype.visitVariable=function(t,e){},t.prototype.visitEvent=function(t,e){},t.prototype.visitElementProperty=function(t,e){},t.prototype.visitAttr=function(t,e){},t.prototype.visitBoundText=function(t,e){},t.prototype.visitText=function(t,e){},t.prototype.visitDirective=function(t,e){},t.prototype.visitDirectiveProperty=function(t,e){},t}(),Cd=function(t){function e(){return t.call(this)||this}return Object(o.__extends)(e,t),e.prototype.visitEmbeddedTemplate=function(t,e){return this.visitChildren(e,function(e){e(t.attrs),e(t.references),e(t.variables),e(t.directives),e(t.providers),e(t.children)})},e.prototype.visitElement=function(t,e){return this.visitChildren(e,function(e){e(t.attrs),e(t.inputs),e(t.outputs),e(t.references),e(t.directives),e(t.providers),e(t.children)})},e.prototype.visitDirective=function(t,e){return this.visitChildren(e,function(e){e(t.inputs),e(t.hostProperties),e(t.hostEvents)})},e.prototype.visitChildren=function(t,e){var n=[],r=this;return e(function(e){e&&e.length&&n.push(Ed(r,e,t))}),[].concat.apply([],n)},e}(Od);function Ed(t,e,n){void 0===n&&(n=null);var r=[],i=t.visit?function(e){return t.visit(e,n)||e.visit(t,n)}:function(e){return e.visit(t,n)};return e.forEach(function(t){var e=i(t);e&&r.push(e)}),r}var Sd=function(){return function(t){var e=void 0===t?{}:t,n=e.defaultEncapsulation,r=void 0===n?_.Emulated:n,i=e.useJit,o=void 0===i||i,a=e.jitDevMode,s=void 0!==a&&a,u=e.missingTranslation,c=void 0===u?null:u,l=e.preserveWhitespaces,d=e.strictInjectionParameters;this.defaultEncapsulation=r,this.useJit=!!o,this.jitDevMode=!!s,this.missingTranslation=c,this.preserveWhitespaces=Md(ze(l)),this.strictInjectionParameters=!0===d}}();function Md(t,e){return void 0===e&&(e=!1),null===t?e:t}var Td=function(){function t(t,e,n,r){this._resourceLoader=t,this._urlResolver=e,this._htmlParser=n,this._config=r,this._resourceLoaderCache=new Map}return t.prototype.clearCache=function(){this._resourceLoaderCache.clear()},t.prototype.clearCacheFor=function(t){var e=this;if(t.isComponent){var n=t.template;this._resourceLoaderCache.delete(n.templateUrl),n.externalStylesheets.forEach(function(t){e._resourceLoaderCache.delete(t.moduleUrl)})}},t.prototype._fetch=function(t){var e=this._resourceLoaderCache.get(t);return e||(e=this._resourceLoader.get(t),this._resourceLoaderCache.set(t,e)),e},t.prototype.normalizeTemplate=function(t){var e=this;if(Ve(t.template)){if(Ve(t.templateUrl))throw Ye("'"+en(t.componentType)+"' component cannot define both template and templateUrl");if("string"!=typeof t.template)throw Ye("The template specified for component "+en(t.componentType)+" is not a string")}else{if(!Ve(t.templateUrl))throw Ye("No template specified for component "+en(t.componentType));if("string"!=typeof t.templateUrl)throw Ye("The templateUrl specified for component "+en(t.componentType)+" is not a string")}if(Ve(t.preserveWhitespaces)&&"boolean"!=typeof t.preserveWhitespaces)throw Ye("The preserveWhitespaces option for component "+en(t.componentType)+" must be a boolean");return We(this._preParseTemplate(t),function(n){return e._normalizeTemplateMetadata(t,n)})},t.prototype._preParseTemplate=function(t){var e,n,r=this;return null!=t.template?(e=t.template,n=t.moduleUrl):(n=this._urlResolver.resolve(t.moduleUrl,t.templateUrl),e=this._fetch(n)),We(e,function(e){return r._preparseLoadedTemplate(t,e,n)})},t.prototype._preparseLoadedTemplate=function(t,e,n){var r=!!t.template,i=zi.fromArray(t.interpolation),o=this._htmlParser.parse(e,qn({reference:t.ngModuleType},{type:{reference:t.componentType}},{isInline:r,templateUrl:n}),!0,i);if(o.errors.length>0)throw Ye("Template parse errors:\n"+o.errors.join("\n"));var a=this._normalizeStylesheet(new Pn({styles:t.styles,moduleUrl:t.moduleUrl})),s=new kd;_u(s,o.rootNodes);var u=this._normalizeStylesheet(new Pn({styles:s.styles,styleUrls:s.styleUrls,moduleUrl:n}));return{template:e,templateUrl:n,isInline:r,htmlAst:o,styles:a.styles.concat(u.styles),inlineStyleUrls:a.styleUrls.concat(u.styleUrls),styleUrls:this._normalizeStylesheet(new Pn({styleUrls:t.styleUrls,moduleUrl:t.moduleUrl})).styleUrls,ngContentSelectors:s.ngContentSelectors}},t.prototype._normalizeTemplateMetadata=function(t,e){var n=this;return We(this._loadMissingExternalStylesheets(e.styleUrls.concat(e.inlineStyleUrls)),function(r){return n._normalizeLoadedTemplateMetadata(t,e,r)})},t.prototype._normalizeLoadedTemplateMetadata=function(t,e,n){var r=this,i=Object(o.__spread)(e.styles);this._inlineStyles(e.inlineStyleUrls,n,i);var a=e.styleUrls,s=a.map(function(t){var e=n.get(t),i=Object(o.__spread)(e.styles);return r._inlineStyles(e.styleUrls,n,i),new Pn({moduleUrl:t,styles:i})}),u=t.encapsulation;return null==u&&(u=this._config.defaultEncapsulation),u===_.Emulated&&0===i.length&&0===a.length&&(u=_.None),new Rn({encapsulation:u,template:e.template,templateUrl:e.templateUrl,htmlAst:e.htmlAst,styles:i,styleUrls:a,ngContentSelectors:e.ngContentSelectors,animations:t.animations,interpolation:t.interpolation,isInline:e.isInline,externalStylesheets:s,preserveWhitespaces:Md(t.preserveWhitespaces,this._config.preserveWhitespaces)})},t.prototype._inlineStyles=function(t,e,n){var r=this;t.forEach(function(t){var i=e.get(t);i.styles.forEach(function(t){return n.push(t)}),r._inlineStyles(i.styleUrls,e,n)})},t.prototype._loadMissingExternalStylesheets=function(t,e){var n=this;return void 0===e&&(e=new Map),We(qe(t.filter(function(t){return!e.has(t)}).map(function(t){return We(n._fetch(t),function(r){var i=n._normalizeStylesheet(new Pn({styles:[r],moduleUrl:t}));return e.set(t,i),n._loadMissingExternalStylesheets(i.styleUrls,e)})})),function(t){return e})},t.prototype._normalizeStylesheet=function(t){var e=this,n=t.moduleUrl,r=t.styleUrls.filter(Cc).map(function(t){return e._urlResolver.resolve(n,t)}),i=t.styles.map(function(t){var i,a,s,u,c=(i=e._urlResolver,a=n,s=[],u=t.replace(Tc,"").replace(Mc,function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];var n=t[1]||t[2];return Cc(n)?(s.push(i.resolve(a,n)),""):t[0]}),new Oc(u,s));return r.push.apply(r,Object(o.__spread)(c.styleUrls)),c.style});return new Pn({styles:i,styleUrls:r,moduleUrl:n})},t}(),kd=function(){function t(){this.ngContentSelectors=[],this.styles=[],this.styleUrls=[],this.ngNonBindableStackCount=0}return t.prototype.visitElement=function(t,e){var n=Vc(t);switch(n.type){case Ec.NG_CONTENT:0===this.ngNonBindableStackCount&&this.ngContentSelectors.push(n.selectAttr);break;case Ec.STYLE:var r="";t.children.forEach(function(t){t instanceof pu&&(r+=t.value)}),this.styles.push(r);break;case Ec.STYLESHEET:this.styleUrls.push(n.hrefAttr)}return n.nonBindable&&this.ngNonBindableStackCount++,_u(this,t.children),n.nonBindable&&this.ngNonBindableStackCount--,null},t.prototype.visitExpansion=function(t,e){_u(this,t.cases)},t.prototype.visitExpansionCase=function(t,e){_u(this,t.expression)},t.prototype.visitComment=function(t,e){return null},t.prototype.visitAttribute=function(t,e){return null},t.prototype.visitText=function(t,e){return null},t}(),Ad=[I,D,A,k],Dd=function(){function t(t){this._reflector=t}return t.prototype.isDirective=function(t){var e=this._reflector.annotations(nn(t));return e&&e.some(Id)},t.prototype.resolve=function(t,e){void 0===e&&(e=!0);var n=this._reflector.annotations(nn(t));if(n){var r=Pd(n,Id);if(r){var i=this._reflector.propMetadata(t),o=this._reflector.guards(t);return this._mergeWithPropertyMetadata(r,i,o,t)}}if(e)throw new Error("No Directive annotation found on "+en(t));return null},t.prototype._mergeWithPropertyMetadata=function(t,e,n,r){var i=[],o=[],a={},s={};return Object.keys(e).forEach(function(t){var n=Pd(e[t],function(t){return z.isTypeOf(t)});n&&(n.bindingPropertyName?i.push(t+": "+n.bindingPropertyName):i.push(t));var r=Pd(e[t],function(t){return H.isTypeOf(t)});r&&(r.bindingPropertyName?o.push(t+": "+r.bindingPropertyName):o.push(t)),e[t].filter(function(t){return U.isTypeOf(t)}).forEach(function(e){if(e.hostPropertyName){var n=e.hostPropertyName[0];if("("===n)throw new Error("@HostBinding can not bind to events. Use @HostListener instead.");if("["===n)throw new Error("@HostBinding parameter should be a property name, 'class.<name>', or 'attr.<name>'.");a["["+e.hostPropertyName+"]"]=t}else a["["+t+"]"]=t}),e[t].filter(function(t){return W.isTypeOf(t)}).forEach(function(e){var n=e.args||[];a["("+e.eventName+")"]=t+"("+n.join(",")+")"});var u=Pd(e[t],function(t){return Ad.some(function(e){return e.isTypeOf(t)})});u&&(s[t]=u)}),this._merge(t,i,o,a,s,n,r)},t.prototype._extractPublicName=function(t){return Fe(t,[null,t])[1].trim()},t.prototype._dedupeBindings=function(t){for(var e=new Set,n=new Set,r=[],i=t.length-1;i>=0;i--){var o=t[i],a=this._extractPublicName(o);n.add(a),e.has(a)||(e.add(a),r.push(o))}return r.reverse()},t.prototype._merge=function(t,e,n,r,i,a,s){var u=this._dedupeBindings(t.inputs?t.inputs.concat(e):e),c=this._dedupeBindings(t.outputs?t.outputs.concat(n):n),l=t.host?Object(o.__assign)({},t.host,r):r,d=t.queries?Object(o.__assign)({},t.queries,i):i;if(B.isTypeOf(t)){var p=t;return B({selector:p.selector,inputs:u,outputs:c,host:l,exportAs:p.exportAs,moduleId:p.moduleId,queries:d,changeDetection:p.changeDetection,providers:p.providers,viewProviders:p.viewProviders,entryComponents:p.entryComponents,template:p.template,templateUrl:p.templateUrl,styles:p.styles,styleUrls:p.styleUrls,encapsulation:p.encapsulation,animations:p.animations,interpolation:p.interpolation,preserveWhitespaces:t.preserveWhitespaces})}return P({selector:t.selector,inputs:u,outputs:c,host:l,exportAs:t.exportAs,queries:d,providers:t.providers,guards:a})},t}();function Id(t){return P.isTypeOf(t)||B.isTypeOf(t)}function Pd(t,e){for(var n=t.length-1;n>=0;n--)if(e(t[n]))return t[n];return null}var Rd=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(o.__extends)(e,t),e}(os),Nd="i18n",Fd=/^i18n:?/,Ld="|",Bd="@@",Vd=!1;var zd,Hd,Ud=function(){return function(t,e){this.messages=t,this.errors=e}}();(Hd=zd||(zd={}))[Hd.Extract=0]="Extract",Hd[Hd.Merge=1]="Merge";var Wd=function(){function t(t,e){this._implicitTags=t,this._implicitAttrs=e}return t.prototype.extract=function(t,e){var n=this;return this._init(zd.Extract,e),t.forEach(function(t){return t.visit(n,null)}),this._inI18nBlock&&this._reportError(t[t.length-1],"Unclosed block"),new Ud(this._messages,this._errors)},t.prototype.merge=function(t,e,n){this._init(zd.Merge,n),this._translations=e;var r=new gu("wrapper",[],t,void 0,void 0,void 0).visit(this,null);return this._inI18nBlock&&this._reportError(t[t.length-1],"Unclosed block"),new Lu(r.children,this._errors)},t.prototype.visitExpansionCase=function(t,e){var n=_u(this,t.expression,e);if(this._mode===zd.Merge)return new hu(t.value,n,t.sourceSpan,t.valueSourceSpan,t.expSourceSpan)},t.prototype.visitExpansion=function(t,e){this._mayBeAddBlockChildren(t);var n=this._inIcu;this._inIcu||(this._isInTranslatableSection&&this._addMessage([t]),this._inIcu=!0);var r=_u(this,t.cases,e);return this._mode===zd.Merge&&(t=new fu(t.switchValue,t.type,r,t.sourceSpan,t.switchValueSourceSpan)),this._inIcu=n,t},t.prototype.visitComment=function(t,e){var n,r=!!((n=t)instanceof yu&&n.value&&n.value.startsWith("i18n"));if(r&&this._isInTranslatableSection)this._reportError(t,"Could not start a block inside a translatable section");else{var i,o=!!((i=t)instanceof yu&&i.value&&"/i18n"===i.value);if(!o||this._inI18nBlock){if(!this._inI18nNode&&!this._inIcu)if(this._inI18nBlock){if(o){if(this._depth==this._blockStartDepth){this._closeTranslatableSection(t,this._blockChildren),this._inI18nBlock=!1;var a=this._addMessage(this._blockChildren,this._blockMeaningAndDesc);return _u(this,this._translateMessage(t,a))}return void this._reportError(t,"I18N blocks should not cross element boundaries")}}else if(r){if(!Vd&&console&&console.warn){Vd=!0;var s=t.sourceSpan.details?", "+t.sourceSpan.details:"";console.warn("I18n comments are deprecated, use an <ng-container> element instead ("+t.sourceSpan.start+s+")")}this._inI18nBlock=!0,this._blockStartDepth=this._depth,this._blockChildren=[],this._blockMeaningAndDesc=t.value.replace(Fd,"").trim(),this._openTranslatableSection(t)}}else this._reportError(t,"Trying to close an unopened block")}},t.prototype.visitText=function(t,e){return this._isInTranslatableSection&&this._mayBeAddBlockChildren(t),t},t.prototype.visitElement=function(t,e){var n=this;this._mayBeAddBlockChildren(t),this._depth++;var r=this._inI18nNode,i=this._inImplicitNode,o=[],a=void 0,s=t.attrs.find(function(t){return t.name===Nd})||null,u=s?s.value:"",c=this._implicitTags.some(function(e){return t.name===e})&&!this._inIcu&&!this._isInTranslatableSection,l=!i&&c;if(this._inImplicitNode=i||c,this._isInTranslatableSection||this._inIcu)(s||l)&&this._reportError(t,"Could not mark an element as translatable inside a translatable section"),this._mode==zd.Extract&&_u(this,t.children);else{if(s||l){this._inI18nNode=!0;var d=this._addMessage(t.children,u);a=this._translateMessage(t,d)}if(this._mode==zd.Extract){var p=s||l;p&&this._openTranslatableSection(t),_u(this,t.children),p&&this._closeTranslatableSection(t,t.children)}}this._mode===zd.Merge&&(a||t.children).forEach(function(t){var r=t.visit(n,e);r&&!n._isInTranslatableSection&&(o=o.concat(r))});if(this._visitAttributesOf(t),this._depth--,this._inI18nNode=r,this._inImplicitNode=i,this._mode===zd.Merge){var f=this._translateAttributes(t);return new gu(t.name,f,o,t.sourceSpan,t.startSourceSpan,t.endSourceSpan)}return null},t.prototype.visitAttribute=function(t,e){throw new Error("unreachable code")},t.prototype._init=function(t,e){this._mode=t,this._inI18nBlock=!1,this._inI18nNode=!1,this._depth=0,this._inIcu=!1,this._msgCountAtSectionStart=void 0,this._errors=[],this._messages=[],this._inImplicitNode=!1,this._createI18nMessage=ol(e)},t.prototype._visitAttributesOf=function(t){var e=this,n={},r=this._implicitAttrs[t.name]||[];t.attrs.filter(function(t){return t.name.startsWith("i18n-")}).forEach(function(t){return n[t.name.slice("i18n-".length)]=t.value}),t.attrs.forEach(function(t){t.name in n?e._addMessage([t],n[t.name]):r.some(function(e){return t.name===e})&&e._addMessage([t])})},t.prototype._addMessage=function(t,e){if(0==t.length||1==t.length&&t[0]instanceof mu&&!t[0].value)return null;var n=qd(e),r=n.meaning,i=n.description,o=n.id,a=this._createI18nMessage(t,r,i,o);return this._messages.push(a),a},t.prototype._translateMessage=function(t,e){if(e&&this._mode===zd.Merge){var n=this._translations.get(e);if(n)return n;this._reportError(t,'Translation unavailable for message id="'+this._translations.digest(e)+'"')}return[]},t.prototype._translateAttributes=function(t){var e=this,n=t.attrs,r={};n.forEach(function(t){t.name.startsWith("i18n-")&&(r[t.name.slice("i18n-".length)]=qd(t.value))});var i=[];return n.forEach(function(n){if(n.name!==Nd&&!n.name.startsWith("i18n-"))if(n.value&&""!=n.value&&r.hasOwnProperty(n.name)){var o=r[n.name],a=o.meaning,s=o.description,u=o.id,c=e._createI18nMessage([n],a,s,u),l=e._translations.get(c);if(l)if(0==l.length)i.push(new mu(n.name,"",n.sourceSpan));else if(l[0]instanceof pu){var d=l[0].value;i.push(new mu(n.name,d,n.sourceSpan))}else e._reportError(t,'Unexpected translation for attribute "'+n.name+'" (id="'+(u||e._translations.digest(c))+'")');else e._reportError(t,'Translation unavailable for attribute "'+n.name+'" (id="'+(u||e._translations.digest(c))+'")')}else i.push(n)}),i},t.prototype._mayBeAddBlockChildren=function(t){this._inI18nBlock&&!this._inIcu&&this._depth==this._blockStartDepth&&this._blockChildren.push(t)},t.prototype._openTranslatableSection=function(t){this._isInTranslatableSection?this._reportError(t,"Unexpected section start"):this._msgCountAtSectionStart=this._messages.length},Object.defineProperty(t.prototype,"_isInTranslatableSection",{get:function(){return void 0!==this._msgCountAtSectionStart},enumerable:!0,configurable:!0}),t.prototype._closeTranslatableSection=function(t,e){if(this._isInTranslatableSection){var n=this._msgCountAtSectionStart;if(1==e.reduce(function(t,e){return t+(e instanceof yu?0:1)},0))for(var r=this._messages.length-1;r>=n;r--){var i=this._messages[r].nodes;if(!(1==i.length&&i[0]instanceof $n)){this._messages.splice(r,1);break}}this._msgCountAtSectionStart=void 0}else this._reportError(t,"Unexpected section end")},t.prototype._reportError=function(t,e){this._errors.push(new Rd(t.sourceSpan,e))},t}();function qd(t){if(!t)return{meaning:"",description:"",id:""};var e=t.indexOf(Bd),n=t.indexOf(Ld),r=Object(o.__read)(e>-1?[t.slice(0,e),t.slice(e+2)]:[t,""],2),i=r[0],a=r[1],s=Object(o.__read)(n>-1?[i.slice(0,n),i.slice(n+1)]:["",i],2);return{meaning:s[0],description:s[1],id:a}}var Gd=new(function(){function t(){this.closedByParent=!1,this.contentType=r.PARSABLE_DATA,this.isVoid=!1,this.ignoreFirstLf=!1,this.canSelfClose=!0}return t.prototype.requireExtraParent=function(t){return!1},t.prototype.isClosedByChild=function(t){return!1},t}());function Yd(t){return Gd}var Kd=function(t){function e(){return t.call(this,Yd)||this}return Object(o.__extends)(e,t),e.prototype.parse=function(e,n,r){return void 0===r&&(r=!1),t.prototype.parse.call(this,e,n,r)},e}(Bu),Xd=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(o.__extends)(e,t),e.prototype.write=function(t,e){var n=new Qd,r=[];t.forEach(function(t){var e,i=[];t.sources.forEach(function(t){var e=new Dr("context-group",{purpose:"location"});e.children.push(new Pr(10),new Dr("context",{"context-type":"sourcefile"},[new Ir(t.filePath)]),new Pr(10),new Dr("context",{"context-type":"linenumber"},[new Ir(""+t.startLine)]),new Pr(8)),i.push(new Pr(8),e)});var a=new Dr("trans-unit",{id:t.id,datatype:"html"});(e=a.children).push.apply(e,Object(o.__spread)([new Pr(8),new Dr("source",{},n.serialize(t.nodes))],i)),t.description&&a.children.push(new Pr(8),new Dr("note",{priority:"1",from:"description"},[new Ir(t.description)])),t.meaning&&a.children.push(new Pr(8),new Dr("note",{priority:"1",from:"meaning"},[new Ir(t.meaning)])),a.children.push(new Pr(6)),r.push(new Pr(6),a)});var i=new Dr("body",{},Object(o.__spread)(r,[new Pr(4)])),a=new Dr("file",{"source-language":e||"en",datatype:"plaintext",original:"ng2.template"},[new Pr(4),i,new Pr(2)]),s=new Dr("xliff",{version:"1.2",xmlns:"urn:oasis:names:tc:xliff:document:1.2"},[new Pr(2),a,new Pr]);return Tr([new kr({version:"1.0",encoding:"UTF-8"}),new Pr,s,new Pr])},e.prototype.load=function(t,e){var n=(new Zd).parse(t,e),r=n.locale,i=n.msgIdToHtml,a=n.errors,s={},u=new $d;if(Object.keys(i).forEach(function(t){var n=u.convert(i[t],e),r=n.i18nNodes,c=n.errors;a.push.apply(a,Object(o.__spread)(c)),s[t]=r}),a.length)throw new Error("xliff parse errors:\n"+a.join("\n"));return{locale:r,i18nNodesByMsgId:s}},e.prototype.digest=function(t){return ar(t)},e}(Er),Qd=function(){function t(){}return t.prototype.visitText=function(t,e){return[new Ir(t.value)]},t.prototype.visitContainer=function(t,e){var n=this,r=[];return t.children.forEach(function(t){return r.push.apply(r,Object(o.__spread)(t.visit(n)))}),r},t.prototype.visitIcu=function(t,e){var n=this,r=[new Ir("{"+t.expressionPlaceholder+", "+t.type+", ")];return Object.keys(t.cases).forEach(function(e){r.push.apply(r,Object(o.__spread)([new Ir(e+" {")],t.cases[e].visit(n),[new Ir("} ")]))}),r.push(new Ir("}")),r},t.prototype.visitTagPlaceholder=function(t,e){var n=function(t){switch(t.toLowerCase()){case"br":return"lb";case"img":return"image";default:return"x-"+t}}(t.tag);if(t.isVoid)return[new Dr("x",{id:t.startName,ctype:n,"equiv-text":"<"+t.tag+"/>"})];var r=new Dr("x",{id:t.startName,ctype:n,"equiv-text":"<"+t.tag+">"}),i=new Dr("x",{id:t.closeName,ctype:n,"equiv-text":"</"+t.tag+">"});return Object(o.__spread)([r],this.serialize(t.children),[i])},t.prototype.visitPlaceholder=function(t,e){return[new Dr("x",{id:t.name,"equiv-text":"{{"+t.value+"}}"})]},t.prototype.visitIcuPlaceholder=function(t,e){var n="{"+t.value.expression+", "+t.value.type+", "+Object.keys(t.value.cases).map(function(t){return t+" {...}"}).join(" ")+"}";return[new Dr("x",{id:t.name,"equiv-text":n})]},t.prototype.serialize=function(t){var e=this;return[].concat.apply([],Object(o.__spread)(t.map(function(t){return t.visit(e)})))},t}(),Zd=function(){function t(){this._locale=null}return t.prototype.parse=function(t,e){this._unitMlString=null,this._msgIdToHtml={};var n=(new Kd).parse(t,e,!1);return this._errors=n.errors,_u(this,n.rootNodes,null),{msgIdToHtml:this._msgIdToHtml,errors:this._errors,locale:this._locale}},t.prototype.visitElement=function(t,e){switch(t.name){case"trans-unit":this._unitMlString=null;var n=t.attrs.find(function(t){return"id"===t.name});if(n){var r=n.value;this._msgIdToHtml.hasOwnProperty(r)?this._addError(t,"Duplicated translations for msg "+r):(_u(this,t.children,null),"string"==typeof this._unitMlString?this._msgIdToHtml[r]=this._unitMlString:this._addError(t,"Message "+r+" misses a translation"))}else this._addError(t,'<trans-unit> misses the "id" attribute');break;case"source":case"seg-source":break;case"target":var i=t.startSourceSpan.end.offset,o=t.endSourceSpan.start.offset,a=t.startSourceSpan.start.file.content.slice(i,o);this._unitMlString=a;break;case"file":var s=t.attrs.find(function(t){return"target-language"===t.name});s&&(this._locale=s.value),_u(this,t.children,null);break;default:_u(this,t.children,null)}},t.prototype.visitAttribute=function(t,e){},t.prototype.visitText=function(t,e){},t.prototype.visitComment=function(t,e){},t.prototype.visitExpansion=function(t,e){},t.prototype.visitExpansionCase=function(t,e){},t.prototype._addError=function(t,e){this._errors.push(new Rd(t.sourceSpan,e))},t}(),$d=function(){function t(){}return t.prototype.convert=function(t,e){var n=(new Kd).parse(t,e,!0);return this._errors=n.errors,{i18nNodes:this._errors.length>0||0==n.rootNodes.length?[]:[].concat.apply([],Object(o.__spread)(_u(this,n.rootNodes))),errors:this._errors}},t.prototype.visitText=function(t,e){return new $n(t.value,t.sourceSpan)},t.prototype.visitElement=function(t,e){if("x"===t.name){var n=t.attrs.find(function(t){return"id"===t.name});return n?new nr("",n.value,t.sourceSpan):(this._addError(t,'<x> misses the "id" attribute'),null)}return"mrk"===t.name?[].concat.apply([],Object(o.__spread)(_u(this,t.children))):(this._addError(t,"Unexpected tag"),null)},t.prototype.visitExpansion=function(t,e){var n={};return _u(this,t.cases).forEach(function(e){n[e.value]=new Jn(e.nodes,t.sourceSpan)}),new tr(t.switchValue,t.type,n,t.sourceSpan)},t.prototype.visitExpansionCase=function(t,e){return{value:t.value,nodes:_u(this,t.expression)}},t.prototype.visitComment=function(t,e){},t.prototype.visitAttribute=function(t,e){},t.prototype._addError=function(t,e){this._errors.push(new Rd(t.sourceSpan,e))},t}();var Jd=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(o.__extends)(e,t),e.prototype.write=function(t,e){var n=new tp,r=[];t.forEach(function(t){var e=new Dr("unit",{id:t.id}),i=new Dr("notes");(t.description||t.meaning)&&(t.description&&i.children.push(new Pr(8),new Dr("note",{category:"description"},[new Ir(t.description)])),t.meaning&&i.children.push(new Pr(8),new Dr("note",{category:"meaning"},[new Ir(t.meaning)]))),t.sources.forEach(function(t){i.children.push(new Pr(8),new Dr("note",{category:"location"},[new Ir(t.filePath+":"+t.startLine+(t.endLine!==t.startLine?","+t.endLine:""))]))}),i.children.push(new Pr(6)),e.children.push(new Pr(6),i);var o=new Dr("segment");o.children.push(new Pr(8),new Dr("source",{},n.serialize(t.nodes)),new Pr(6)),e.children.push(new Pr(6),o,new Pr(4)),r.push(new Pr(4),e)});var i=new Dr("file",{original:"ng.template",id:"ngi18n"},Object(o.__spread)(r,[new Pr(2)])),a=new Dr("xliff",{version:"2.0",xmlns:"urn:oasis:names:tc:xliff:document:2.0",srcLang:e||"en"},[new Pr(2),i,new Pr]);return Tr([new kr({version:"1.0",encoding:"UTF-8"}),new Pr,a,new Pr])},e.prototype.load=function(t,e){var n=(new ep).parse(t,e),r=n.locale,i=n.msgIdToHtml,a=n.errors,s={},u=new np;if(Object.keys(i).forEach(function(t){var n=u.convert(i[t],e),r=n.i18nNodes,c=n.errors;a.push.apply(a,Object(o.__spread)(c)),s[t]=r}),a.length)throw new Error("xliff2 parse errors:\n"+a.join("\n"));return{locale:r,i18nNodesByMsgId:s}},e.prototype.digest=function(t){return sr(t)},e}(Er),tp=function(){function t(){}return t.prototype.visitText=function(t,e){return[new Ir(t.value)]},t.prototype.visitContainer=function(t,e){var n=this,r=[];return t.children.forEach(function(t){return r.push.apply(r,Object(o.__spread)(t.visit(n)))}),r},t.prototype.visitIcu=function(t,e){var n=this,r=[new Ir("{"+t.expressionPlaceholder+", "+t.type+", ")];return Object.keys(t.cases).forEach(function(e){r.push.apply(r,Object(o.__spread)([new Ir(e+" {")],t.cases[e].visit(n),[new Ir("} ")]))}),r.push(new Ir("}")),r},t.prototype.visitTagPlaceholder=function(t,e){var n=this,r=function(t){switch(t.toLowerCase()){case"br":case"b":case"i":case"u":return"fmt";case"img":return"image";case"a":return"link";default:return"other"}}(t.tag);if(t.isVoid)return[new Dr("ph",{id:(this._nextPlaceholderId++).toString(),equiv:t.startName,type:r,disp:"<"+t.tag+"/>"})];var i=new Dr("pc",{id:(this._nextPlaceholderId++).toString(),equivStart:t.startName,equivEnd:t.closeName,type:r,dispStart:"<"+t.tag+">",dispEnd:"</"+t.tag+">"}),a=[].concat.apply([],Object(o.__spread)(t.children.map(function(t){return t.visit(n)})));return a.length?a.forEach(function(t){return i.children.push(t)}):i.children.push(new Ir("")),[i]},t.prototype.visitPlaceholder=function(t,e){var n=(this._nextPlaceholderId++).toString();return[new Dr("ph",{id:n,equiv:t.name,disp:"{{"+t.value+"}}"})]},t.prototype.visitIcuPlaceholder=function(t,e){var n=Object.keys(t.value.cases).map(function(t){return t+" {...}"}).join(" "),r=(this._nextPlaceholderId++).toString();return[new Dr("ph",{id:r,equiv:t.name,disp:"{"+t.value.expression+", "+t.value.type+", "+n+"}"})]},t.prototype.serialize=function(t){var e=this;return this._nextPlaceholderId=0,[].concat.apply([],Object(o.__spread)(t.map(function(t){return t.visit(e)})))},t}(),ep=function(){function t(){this._locale=null}return t.prototype.parse=function(t,e){this._unitMlString=null,this._msgIdToHtml={};var n=(new Kd).parse(t,e,!1);return this._errors=n.errors,_u(this,n.rootNodes,null),{msgIdToHtml:this._msgIdToHtml,errors:this._errors,locale:this._locale}},t.prototype.visitElement=function(t,e){switch(t.name){case"unit":this._unitMlString=null;var n=t.attrs.find(function(t){return"id"===t.name});if(n){var r=n.value;this._msgIdToHtml.hasOwnProperty(r)?this._addError(t,"Duplicated translations for msg "+r):(_u(this,t.children,null),"string"==typeof this._unitMlString?this._msgIdToHtml[r]=this._unitMlString:this._addError(t,"Message "+r+" misses a translation"))}else this._addError(t,'<unit> misses the "id" attribute');break;case"source":break;case"target":var i=t.startSourceSpan.end.offset,o=t.endSourceSpan.start.offset,a=t.startSourceSpan.start.file.content.slice(i,o);this._unitMlString=a;break;case"xliff":var s=t.attrs.find(function(t){return"trgLang"===t.name});s&&(this._locale=s.value);var u=t.attrs.find(function(t){return"version"===t.name});if(u){var c=u.value;"2.0"!==c?this._addError(t,"The XLIFF file version "+c+" is not compatible with XLIFF 2.0 serializer"):_u(this,t.children,null)}break;default:_u(this,t.children,null)}},t.prototype.visitAttribute=function(t,e){},t.prototype.visitText=function(t,e){},t.prototype.visitComment=function(t,e){},t.prototype.visitExpansion=function(t,e){},t.prototype.visitExpansionCase=function(t,e){},t.prototype._addError=function(t,e){this._errors.push(new Rd(t.sourceSpan,e))},t}(),np=function(){function t(){}return t.prototype.convert=function(t,e){var n=(new Kd).parse(t,e,!0);return this._errors=n.errors,{i18nNodes:this._errors.length>0||0==n.rootNodes.length?[]:[].concat.apply([],Object(o.__spread)(_u(this,n.rootNodes))),errors:this._errors}},t.prototype.visitText=function(t,e){return new $n(t.value,t.sourceSpan)},t.prototype.visitElement=function(t,e){var n=this;switch(t.name){case"ph":var r=t.attrs.find(function(t){return"equiv"===t.name});if(r)return[new nr("",r.value,t.sourceSpan)];this._addError(t,'<ph> misses the "equiv" attribute');break;case"pc":var i=t.attrs.find(function(t){return"equivStart"===t.name}),a=t.attrs.find(function(t){return"equivEnd"===t.name});if(i){if(a){var s=i.value,u=a.value,c=[];return c.concat.apply(c,Object(o.__spread)([new nr("",s,t.sourceSpan)],t.children.map(function(t){return t.visit(n,null)}),[new nr("",u,t.sourceSpan)]))}this._addError(t,'<ph> misses the "equivEnd" attribute')}else this._addError(t,'<ph> misses the "equivStart" attribute');break;case"mrk":return[].concat.apply([],Object(o.__spread)(_u(this,t.children)));default:this._addError(t,"Unexpected tag")}return null},t.prototype.visitExpansion=function(t,e){var n={};return _u(this,t.cases).forEach(function(e){n[e.value]=new Jn(e.nodes,t.sourceSpan)}),new tr(t.switchValue,t.type,n,t.sourceSpan)},t.prototype.visitExpansionCase=function(t,e){return{value:t.value,nodes:[].concat.apply([],Object(o.__spread)(_u(this,t.expression)))}},t.prototype.visitComment=function(t,e){},t.prototype.visitAttribute=function(t,e){},t.prototype._addError=function(t,e){this._errors.push(new Rd(t.sourceSpan,e))},t}();var rp=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(o.__extends)(e,t),e.prototype.write=function(t,e){throw new Error("Unsupported")},e.prototype.load=function(t,e){var n=(new ip).parse(t,e),r=n.locale,i=n.msgIdToHtml,o=n.errors,a={},s=new op;if(Object.keys(i).forEach(function(t){var n,r,o;n=a,r=t,o=function(){var n=s.convert(i[t],e),r=n.i18nNodes,o=n.errors;if(o.length)throw new Error("xtb parse errors:\n"+o.join("\n"));return r},Object.defineProperty(n,r,{configurable:!0,enumerable:!0,get:function(){var t=o();return Object.defineProperty(n,r,{enumerable:!0,value:t}),t},set:function(t){throw new Error("Could not overwrite an XTB translation")}})}),o.length)throw new Error("xtb parse errors:\n"+o.join("\n"));return{locale:r,i18nNodesByMsgId:a}},e.prototype.digest=function(t){return Br(t)},e.prototype.createNameMapper=function(t){return new Sr(t,zr)},e}(Er);var ip=function(){function t(){this._locale=null}return t.prototype.parse=function(t,e){this._bundleDepth=0,this._msgIdToHtml={};var n=(new Kd).parse(t,e,!1);return this._errors=n.errors,_u(this,n.rootNodes),{msgIdToHtml:this._msgIdToHtml,errors:this._errors,locale:this._locale}},t.prototype.visitElement=function(t,e){switch(t.name){case"translationbundle":this._bundleDepth++,this._bundleDepth>1&&this._addError(t,"<translationbundle> elements can not be nested");var n=t.attrs.find(function(t){return"lang"===t.name});n&&(this._locale=n.value),_u(this,t.children,null),this._bundleDepth--;break;case"translation":var r=t.attrs.find(function(t){return"id"===t.name});if(r){var i=r.value;if(this._msgIdToHtml.hasOwnProperty(i))this._addError(t,"Duplicated translations for msg "+i);else{var o=t.startSourceSpan.end.offset,a=t.endSourceSpan.start.offset,s=t.startSourceSpan.start.file.content.slice(o,a);this._msgIdToHtml[i]=s}}else this._addError(t,'<translation> misses the "id" attribute');break;default:this._addError(t,"Unexpected tag")}},t.prototype.visitAttribute=function(t,e){},t.prototype.visitText=function(t,e){},t.prototype.visitComment=function(t,e){},t.prototype.visitExpansion=function(t,e){},t.prototype.visitExpansionCase=function(t,e){},t.prototype._addError=function(t,e){this._errors.push(new Rd(t.sourceSpan,e))},t}(),op=function(){function t(){}return t.prototype.convert=function(t,e){var n=(new Kd).parse(t,e,!0);return this._errors=n.errors,{i18nNodes:this._errors.length>0||0==n.rootNodes.length?[]:_u(this,n.rootNodes),errors:this._errors}},t.prototype.visitText=function(t,e){return new $n(t.value,t.sourceSpan)},t.prototype.visitExpansion=function(t,e){var n={};return _u(this,t.cases).forEach(function(e){n[e.value]=new Jn(e.nodes,t.sourceSpan)}),new tr(t.switchValue,t.type,n,t.sourceSpan)},t.prototype.visitExpansionCase=function(t,e){return{value:t.value,nodes:_u(this,t.expression)}},t.prototype.visitElement=function(t,e){if("ph"===t.name){var n=t.attrs.find(function(t){return"name"===t.name});if(n)return new nr("",n.value,t.sourceSpan);this._addError(t,'<ph> misses the "name" attribute')}else this._addError(t,"Unexpected tag");return null},t.prototype.visitComment=function(t,e){},t.prototype.visitAttribute=function(t,e){},t.prototype._addError=function(t,e){this._errors.push(new Rd(t.sourceSpan,e))},t}(),ap=function(){function t(t,e,n,r,i,o){void 0===t&&(t={}),void 0===i&&(i=F.Warning),this._i18nNodesByMsgId=t,this.digest=n,this.mapperFactory=r,this._i18nToHtml=new sp(t,e,n,r,i,o)}return t.load=function(e,n,r,i,o){var a=r.load(e,n),s=a.locale;return new t(a.i18nNodesByMsgId,s,function(t){return r.digest(t)},function(t){return r.createNameMapper(t)},i,o)},t.prototype.get=function(t){var e=this._i18nToHtml.convert(t);if(e.errors.length)throw new Error(e.errors.join("\n"));return e.nodes},t.prototype.has=function(t){return this.digest(t)in this._i18nNodesByMsgId},t}(),sp=function(){function t(t,e,n,r,i,o){void 0===t&&(t={}),this._i18nNodesByMsgId=t,this._locale=e,this._digest=n,this._mapperFactory=r,this._missingTranslationStrategy=i,this._console=o,this._contextStack=[],this._errors=[]}return t.prototype.convert=function(t){this._contextStack.length=0,this._errors.length=0;var e=this._convertToText(t),n=t.nodes[0].sourceSpan.start.file.url,r=(new Hu).parse(e,n,!0);return{nodes:r.rootNodes,errors:Object(o.__spread)(this._errors,r.errors)}},t.prototype.visitText=function(t,e){return Nr(t.value)},t.prototype.visitContainer=function(t,e){var n=this;return t.children.map(function(t){return t.visit(n)}).join("")},t.prototype.visitIcu=function(t,e){var n=this,r=Object.keys(t.cases).map(function(e){return e+" {"+t.cases[e].visit(n)+"}"});return"{"+(this._srcMsg.placeholders.hasOwnProperty(t.expression)?this._srcMsg.placeholders[t.expression]:t.expression)+", "+t.type+", "+r.join(" ")+"}"},t.prototype.visitPlaceholder=function(t,e){var n=this._mapper(t.name);return this._srcMsg.placeholders.hasOwnProperty(n)?this._srcMsg.placeholders[n]:this._srcMsg.placeholderToMessage.hasOwnProperty(n)?this._convertToText(this._srcMsg.placeholderToMessage[n]):(this._addError(t,'Unknown placeholder "'+t.name+'"'),"")},t.prototype.visitTagPlaceholder=function(t,e){var n=this,r=""+t.tag,i=Object.keys(t.attrs).map(function(e){return e+'="'+t.attrs[e]+'"'}).join(" ");return t.isVoid?"<"+r+" "+i+"/>":"<"+r+" "+i+">"+t.children.map(function(t){return t.visit(n)}).join("")+"</"+r+">"},t.prototype.visitIcuPlaceholder=function(t,e){return this._convertToText(this._srcMsg.placeholderToMessage[t.name])},t.prototype._convertToText=function(t){var e,n=this,r=this._digest(t),i=this._mapperFactory?this._mapperFactory(t):null;if(this._contextStack.push({msg:this._srcMsg,mapper:this._mapper}),this._srcMsg=t,this._i18nNodesByMsgId.hasOwnProperty(r))e=this._i18nNodesByMsgId[r],this._mapper=function(t){return i?i.toInternalName(t):t};else{if(this._missingTranslationStrategy===F.Error){var o=this._locale?' for locale "'+this._locale+'"':"";this._addError(t.nodes[0],'Missing translation for message "'+r+'"'+o)}else if(this._console&&this._missingTranslationStrategy===F.Warning){o=this._locale?' for locale "'+this._locale+'"':"";this._console.warn('Missing translation for message "'+r+'"'+o)}e=t.nodes,this._mapper=function(t){return t}}var a=e.map(function(t){return t.visit(n)}).join(""),s=this._contextStack.pop();return this._srcMsg=s.msg,this._mapper=s.mapper,a},t.prototype._addError=function(t,e){this._errors.push(new Rd(t.sourceSpan,e))},t}(),up=function(){function t(t,e,n,r,i){if(void 0===r&&(r=F.Warning),this._htmlParser=t,e){var o=function(t){switch(t=(t||"xlf").toLowerCase()){case"xmb":return new Fr;case"xtb":return new rp;case"xliff2":case"xlf2":return new Jd;case"xliff":case"xlf":default:return new Xd}}(n);this._translationBundle=ap.load(e,"i18n",o,r,i)}else this._translationBundle=new ap({},null,ar,void 0,r,i)}return t.prototype.parse=function(t,e,n,r){void 0===n&&(n=!1),void 0===r&&(r=Hi);var i,o,a,s=this._htmlParser.parse(t,e,n,r);return s.errors.length?new Lu(s.rootNodes,s.errors):(i=s.rootNodes,o=this._translationBundle,a=r,new Wd([],{}).merge(i,o,a))},t}();function cp(t,e,n){return void 0===n&&(n=null),Be(e,new lp(t),n)}var lp=function(){function t(t){this.ctx=t}return t.prototype.visitArray=function(t,e){var n=this;return ke(t.map(function(t){return Be(t,n,null)}),e)},t.prototype.visitStringMap=function(t,e){var n=this,r=[],i=new Set(t&&t.$quoted$);return Object.keys(t).forEach(function(e){r.push(new Zt(e,Be(t[e],n,null),i.has(e)))}),new $t(r,e)},t.prototype.visitPrimitive=function(t,e){return Pe(t,e)},t.prototype.visitOther=function(t,e){return t instanceof Ot?t:this.ctx.importExpr(t)},t}();function dp(t,e){return{key:t,value:e,quoted:!1}}var pp=function(){function t(t,e){this.reflector=t,this.alwaysGenerateDef=e,this.tokenInjector=t.resolveExternalReference(gn.Injector)}return t.prototype.depsArray=function(t,e){var n=this;return t.map(function(t){var r,i=t,o=[i],a=0;if(Array.isArray(t))for(var s=0;s<t.length;s++){var u=t[s];u&&("Optional"===u.ngMetadataName?a|=8:"SkipSelf"===u.ngMetadataName?a|=4:"Self"===u.ngMetadataName?a|=2:i="Inject"===u.ngMetadataName?u.token:u)}return r="string"==typeof i?Pe(i):i===n.tokenInjector?Se(gn.INJECTOR):e.importExpr(i),o=0!==a?[r,Pe(a)]:[r],Se(gn.inject).callFn(o)})},t.prototype.factoryFor=function(t,e){var n;if(t.useExisting)n=Se(gn.inject).callFn([e.importExpr(t.useExisting)]);else if(t.useFactory){var r=t.deps||[];if(!(r.length>0))return e.importExpr(t.useFactory);n=e.importExpr(t.useFactory).callFn(this.depsArray(r,e))}else if(t.useValue)n=cp(e,t.useValue);else{var i=t.useClass||t.symbol,o=this.depsArray(this.reflector.parameters(i),e);n=new Ft(e.importExpr(i),o)}return De([],[new ae(n)],void 0,void 0,t.symbol.name+"_Factory")},t.prototype.injectableDef=function(t,e){var n=te;void 0!==t.providedIn&&(n=null===t.providedIn?te:"string"==typeof t.providedIn?Pe(t.providedIn):e.importExpr(t.providedIn));var r=[dp("factory",this.factoryFor(t,e)),dp("token",e.importExpr(t.type.reference)),dp("providedIn",n)];return Se(gn.defineInjectable).callFn([Ae(r)])},t.prototype.compile=function(t,e){if(this.alwaysGenerateDef||void 0!==t.providedIn){var n=En(t.type),r=new de(n,null,[new ue("ngInjectableDef",gt,[It.Static],this.injectableDef(t,e))],[],new ce(null,[],[]),[]);e.statements.push(r)}},t}(),fp=/(\.ts|\.d\.ts|\.js|\.jsx|\.tsx)$/,hp=/\.ngfactory\.|\.ngsummary\./,mp=/\.ngsummary\./,gp=/NgSummary$/;function yp(t,e){void 0===e&&(e=!1);var n=vp(t,e);return n[0]+".ngfactory"+bp(n[1])}function _p(t){return t.replace(hp,".")}function vp(t,e){if(void 0===e&&(e=!1),t.endsWith(".d.ts"))return[t.slice(0,-5),e?".ts":".d.ts"];var n=t.lastIndexOf(".");return-1!==n?[t.substring(0,n),t.substring(n)]:[t,""]}function bp(t){return".tsx"===t?".ts":t}function xp(t){return t.replace(fp,"")+".ngsummary.json"}function wp(t,e){void 0===e&&(e=!1);var n=vp(_p(t),e);return n[0]+".ngsummary"+n[1]}function jp(t){return t+"NgSummary"}var Op=/\u0275\d+/;function Cp(t){return Op.test(t)}function Ep(t){return"ɵ"+t}var Sp="ngComponentType",Mp=function(){function t(t,e,n,r,i,o,a,s,u,c,l,d){this._config=t,this._htmlParser=e,this._ngModuleResolver=n,this._directiveResolver=r,this._pipeResolver=i,this._summaryResolver=o,this._schemaRegistry=a,this._directiveNormalizer=s,this._console=u,this._staticSymbolCache=c,this._reflector=l,this._errorCollector=d,this._nonNormalizedDirectiveCache=new Map,this._directiveCache=new Map,this._summaryCache=new Map,this._pipeCache=new Map,this._ngModuleCache=new Map,this._ngModuleOfTypes=new Map,this._shallowModuleCache=new Map}return t.prototype.getReflector=function(){return this._reflector},t.prototype.clearCacheFor=function(t){var e=this._directiveCache.get(t);this._directiveCache.delete(t),this._nonNormalizedDirectiveCache.delete(t),this._summaryCache.delete(t),this._pipeCache.delete(t),this._ngModuleOfTypes.delete(t),this._ngModuleCache.clear(),e&&this._directiveNormalizer.clearCacheFor(e)},t.prototype.clearCache=function(){this._directiveCache.clear(),this._nonNormalizedDirectiveCache.clear(),this._summaryCache.clear(),this._pipeCache.clear(),this._ngModuleCache.clear(),this._ngModuleOfTypes.clear(),this._directiveNormalizer.clearCache()},t.prototype._createProxyClass=function(t,e){var n=null,r=function(){if(!n)throw new Error("Illegal state: Class "+e+" for type "+en(t)+" is not compiled yet!");return n.apply(this,arguments)};return r.setDelegate=function(t){n=t,r.prototype=t.prototype},r.overriddenName=e,r},t.prototype.getGeneratedClass=function(t,e){return t instanceof vn?this._staticSymbolCache.get(yp(t.filePath),e):this._createProxyClass(t,e)},t.prototype.getComponentViewClass=function(t){return this.getGeneratedClass(t,Mn(t,0))},t.prototype.getHostComponentViewClass=function(t){return this.getGeneratedClass(t,kn(t))},t.prototype.getHostComponentType=function(t){var e=En({reference:t})+"_Host";return t instanceof vn?this._staticSymbolCache.get(t.filePath,e):this._createProxyClass(t,e)},t.prototype.getRendererType=function(t){return t instanceof vn?this._staticSymbolCache.get(yp(t.filePath),Tn(t)):{}},t.prototype.getComponentFactory=function(t,e,n,r){if(e instanceof vn)return this._staticSymbolCache.get(yp(e.filePath),An(e));var i=this.getHostComponentViewClass(e);return this._reflector.resolveExternalReference(gn.createComponentFactory)(t,e,i,n,r,[])},t.prototype.initComponentFactory=function(t,e){var n;t instanceof vn||(n=t.ngContentSelectors).push.apply(n,Object(o.__spread)(e))},t.prototype._loadSummary=function(t,e){var n=this._summaryCache.get(t);if(!n){var r=this._summaryResolver.resolveSummary(t);n=r?r.type:null,this._summaryCache.set(t,n||null)}return n&&n.summaryKind===e?n:null},t.prototype.getHostComponentMetadata=function(t,e){var n=this.getHostComponentType(t.type.reference);e||(e=this.getHostComponentViewClass(n));var r=j.parse(t.selector)[0].getMatchingElementTemplate(),i=this._htmlParser.parse(r,"");return Nn.create({isHost:!0,type:{reference:n,diDeps:[],lifecycleHooks:[]},template:new Rn({encapsulation:_.None,template:r,templateUrl:"",htmlAst:i,styles:[],styleUrls:[],ngContentSelectors:[],animations:[],isInline:!0,externalStylesheets:[],interpolation:null,preserveWhitespaces:!1}),exportAs:null,changeDetection:b.Default,inputs:[],outputs:[],host:{},isComponent:!0,selector:"*",providers:[],viewProviders:[],queries:[],guards:{},viewQueries:[],componentViewType:e,rendererType:{id:"__Host__",encapsulation:_.None,styles:[],data:{}},entryComponents:[],componentFactory:null})},t.prototype.loadDirectiveMetadata=function(t,e,n){var r=this;if(this._directiveCache.has(e))return null;e=nn(e);var i,o,a=this.getNonNormalizedDirectiveMetadata(e),s=a.annotation,u=a.metadata,c=function(t){var n=new Nn({isHost:!1,type:u.type,isComponent:u.isComponent,selector:u.selector,exportAs:u.exportAs,changeDetection:u.changeDetection,inputs:u.inputs,outputs:u.outputs,hostListeners:u.hostListeners,hostProperties:u.hostProperties,hostAttributes:u.hostAttributes,providers:u.providers,viewProviders:u.viewProviders,queries:u.queries,guards:u.guards,viewQueries:u.viewQueries,entryComponents:u.entryComponents,componentViewType:u.componentViewType,rendererType:u.rendererType,componentFactory:u.componentFactory,template:t});return t&&r.initComponentFactory(u.componentFactory,t.ngContentSelectors),r._directiveCache.set(e,n),r._summaryCache.set(e,n.toSummary()),null};if(u.isComponent){var l=u.template,d=this._directiveNormalizer.normalizeTemplate({ngModuleType:t,componentType:e,moduleUrl:this._reflector.componentModuleUrl(e,s),encapsulation:l.encapsulation,template:l.template,templateUrl:l.templateUrl,styles:l.styles,styleUrls:l.styleUrls,animations:l.animations,interpolation:l.interpolation,preserveWhitespaces:l.preserveWhitespaces});return rn(d)&&n?(this._reportError((i=e,(o=Error("Can't compile synchronously as "+en(i)+" is still being loaded!"))[Sp]=i,o),e),null):We(d,c)}return c(null),null},t.prototype.getNonNormalizedDirectiveMetadata=function(t){var e=this;if(!(t=nn(t)))return null;var n=this._nonNormalizedDirectiveCache.get(t);if(n)return n;var r=this._directiveResolver.resolve(t,!1);if(!r)return null;var i=void 0;if(B.isTypeOf(r)){Li("styles",(a=r).styles),Li("styleUrls",a.styleUrls),Vi("interpolation",a.interpolation);var o=a.animations;i=new Rn({encapsulation:ze(a.encapsulation),template:ze(a.template),templateUrl:ze(a.templateUrl),htmlAst:null,styles:a.styles||[],styleUrls:a.styleUrls||[],animations:o||[],interpolation:ze(a.interpolation),isInline:!!a.template,externalStylesheets:[],ngContentSelectors:[],preserveWhitespaces:ze(r.preserveWhitespaces)})}var a,s=null,u=[],c=[],l=r.selector;B.isTypeOf(r)?(s=(a=r).changeDetection,a.viewProviders&&(u=this._getProvidersMetadata(a.viewProviders,c,'viewProviders for "'+Dp(t)+'"',[],t)),a.entryComponents&&(c=Tp(a.entryComponents).map(function(t){return e._getEntryComponentMetadata(t)}).concat(c)),l||(l=this._schemaRegistry.getDefaultComponentElementName())):l||(this._reportError(Ye("Directive "+Dp(t)+" has no selector, please add it!"),t),l="error");var d=[];null!=r.providers&&(d=this._getProvidersMetadata(r.providers,c,'providers for "'+Dp(t)+'"',[],t));var p=[],f=[];null!=r.queries&&(p=this._getQueriesMetadata(r.queries,!1,t),f=this._getQueriesMetadata(r.queries,!0,t));var h=Nn.create({isHost:!1,selector:l,exportAs:ze(r.exportAs),isComponent:!!i,type:this._getTypeMetadata(t),template:i,changeDetection:s,inputs:r.inputs||[],outputs:r.outputs||[],host:r.host||{},providers:d||[],viewProviders:u||[],queries:p||[],guards:r.guards||{},viewQueries:f||[],entryComponents:c,componentViewType:i?this.getComponentViewClass(t):null,rendererType:i?this.getRendererType(t):null,componentFactory:null});return i&&(h.componentFactory=this.getComponentFactory(l,t,h.inputs,h.outputs)),n={metadata:h,annotation:r},this._nonNormalizedDirectiveCache.set(t,n),n},t.prototype.getDirectiveMetadata=function(t){var e=this._directiveCache.get(t);return e||this._reportError(Ye("Illegal state: getDirectiveMetadata can only be called after loadNgModuleDirectiveAndPipeMetadata for a module that declares it. Directive "+Dp(t)+"."),t),e},t.prototype.getDirectiveSummary=function(t){var e=this._loadSummary(t,jn.Directive);return e||this._reportError(Ye("Illegal state: Could not load the summary for directive "+Dp(t)+"."),t),e},t.prototype.isDirective=function(t){return!!this._loadSummary(t,jn.Directive)||this._directiveResolver.isDirective(t)},t.prototype.isPipe=function(t){return!!this._loadSummary(t,jn.Pipe)||this._pipeResolver.isPipe(t)},t.prototype.isNgModule=function(t){return!!this._loadSummary(t,jn.NgModule)||this._ngModuleResolver.isNgModule(t)},t.prototype.getNgModuleSummary=function(t,e){void 0===e&&(e=null);var n=this._loadSummary(t,jn.NgModule);if(!n){var r=this.getNgModuleMetadata(t,!1,e);(n=r?r.toSummary():null)&&this._summaryCache.set(t,n)}return n},t.prototype.loadNgModuleDirectiveAndPipeMetadata=function(t,e,n){var r=this;void 0===n&&(n=!0);var i=this.getNgModuleMetadata(t,n),o=[];return i&&(i.declaredDirectives.forEach(function(n){var i=r.loadDirectiveMetadata(t,n.reference,e);i&&o.push(i)}),i.declaredPipes.forEach(function(t){return r._loadPipeMetadata(t.reference)})),Promise.all(o)},t.prototype.getShallowModuleMetadata=function(t){var e=this._shallowModuleCache.get(t);if(e)return e;var n=Pd(this._reflector.shallowAnnotations(t),q.isTypeOf);return e={type:this._getTypeMetadata(t),rawExports:n.exports,rawImports:n.imports,rawProviders:n.providers},this._shallowModuleCache.set(t,e),e},t.prototype.getNgModuleMetadata=function(t,e,n){var r=this;void 0===e&&(e=!0),void 0===n&&(n=null),t=nn(t);var i=this._ngModuleCache.get(t);if(i)return i;var a=this._ngModuleResolver.resolve(t,e);if(!a)return null;var s=[],u=[],c=[],l=[],d=[],p=[],f=[],h=[],m=[];a.imports&&Tp(a.imports).forEach(function(e){var i=void 0;if(kp(e))i=e;else if(e&&e.ngModule){var a=e;i=a.ngModule,a.providers&&p.push.apply(p,Object(o.__spread)(r._getProvidersMetadata(a.providers,f,"provider for the NgModule '"+Dp(i)+"'",[],e)))}if(i){if(!r._checkSelfImport(t,i))if(n||(n=new Set),n.has(i))r._reportError(Ye(r._getTypeDescriptor(i)+" '"+Dp(e)+"' is imported recursively by the module '"+Dp(t)+"'."),t);else{n.add(i);var s=r.getNgModuleSummary(i,n);n.delete(i),s?l.push(s):r._reportError(Ye("Unexpected "+r._getTypeDescriptor(e)+" '"+Dp(e)+"' imported by the module '"+Dp(t)+"'. Please add a @NgModule annotation."),t)}}else r._reportError(Ye("Unexpected value '"+Dp(e)+"' imported by the module '"+Dp(t)+"'"),t)}),a.exports&&Tp(a.exports).forEach(function(e){if(kp(e))if(n||(n=new Set),n.has(e))r._reportError(Ye(r._getTypeDescriptor(e)+" '"+en(e)+"' is exported recursively by the module '"+Dp(t)+"'"),t);else{n.add(e);var i=r.getNgModuleSummary(e,n);n.delete(e),i?d.push(i):u.push(r._getIdentifierMetadata(e))}else r._reportError(Ye("Unexpected value '"+Dp(e)+"' exported by the module '"+Dp(t)+"'"),t)});var g=this._getTransitiveNgModuleMetadata(l,d);a.declarations&&Tp(a.declarations).forEach(function(e){if(kp(e)){var n=r._getIdentifierMetadata(e);if(r.isDirective(e))g.addDirective(n),s.push(n),r._addTypeToModule(e,t);else{if(!r.isPipe(e))return void r._reportError(Ye("Unexpected "+r._getTypeDescriptor(e)+" '"+Dp(e)+"' declared by the module '"+Dp(t)+"'. Please add a @Pipe/@Directive/@Component annotation."),t);g.addPipe(n),g.pipes.push(n),c.push(n),r._addTypeToModule(e,t)}}else r._reportError(Ye("Unexpected value '"+Dp(e)+"' declared by the module '"+Dp(t)+"'"),t)});var y=[],_=[];return u.forEach(function(e){if(g.directivesSet.has(e.reference))y.push(e),g.addExportedDirective(e);else{if(!g.pipesSet.has(e.reference))return void r._reportError(Ye("Can't export "+r._getTypeDescriptor(e.reference)+" "+Dp(e.reference)+" from "+Dp(t)+" as it was neither declared nor imported!"),t);_.push(e),g.addExportedPipe(e)}}),a.providers&&p.push.apply(p,Object(o.__spread)(this._getProvidersMetadata(a.providers,f,"provider for the NgModule '"+Dp(t)+"'",[],t))),a.entryComponents&&f.push.apply(f,Object(o.__spread)(Tp(a.entryComponents).map(function(t){return r._getEntryComponentMetadata(t)}))),a.bootstrap&&Tp(a.bootstrap).forEach(function(e){kp(e)?h.push(r._getIdentifierMetadata(e)):r._reportError(Ye("Unexpected value '"+Dp(e)+"' used in the bootstrap property of module '"+Dp(t)+"'"),t)}),f.push.apply(f,Object(o.__spread)(h.map(function(t){return r._getEntryComponentMetadata(t.reference)}))),a.schemas&&m.push.apply(m,Object(o.__spread)(Tp(a.schemas))),i=new Bn({type:this._getTypeMetadata(t),providers:p,entryComponents:f,bootstrapComponents:h,schemas:m,declaredDirectives:s,exportedDirectives:y,declaredPipes:c,exportedPipes:_,importedModules:l,exportedModules:d,transitiveModule:g,id:a.id||null}),f.forEach(function(t){return g.addEntryComponent(t)}),p.forEach(function(t){return g.addProvider(t,i.type)}),g.addModule(i.type),this._ngModuleCache.set(t,i),i},t.prototype._checkSelfImport=function(t,e){return t===e&&(this._reportError(Ye("'"+Dp(t)+"' module can't import itself"),t),!0)},t.prototype._getTypeDescriptor=function(t){if(kp(t)){if(this.isDirective(t))return"directive";if(this.isPipe(t))return"pipe";if(this.isNgModule(t))return"module"}return t.provide?"provider":"value"},t.prototype._addTypeToModule=function(t,e){var n=this._ngModuleOfTypes.get(t);n&&n!==e?this._reportError(Ye("Type "+Dp(t)+" is part of the declarations of 2 modules: "+Dp(n)+" and "+Dp(e)+"! Please consider moving "+Dp(t)+" to a higher module that imports "+Dp(n)+" and "+Dp(e)+". You can also create a new NgModule that exports and includes "+Dp(t)+" then import that NgModule in "+Dp(n)+" and "+Dp(e)+"."),e):this._ngModuleOfTypes.set(t,e)},t.prototype._getTransitiveNgModuleMetadata=function(t,e){var n=new Vn,r=new Map;return t.concat(e).forEach(function(t){t.modules.forEach(function(t){return n.addModule(t)}),t.entryComponents.forEach(function(t){return n.addEntryComponent(t)});var e=new Set;t.providers.forEach(function(t){var i=In(t.provider.token),o=r.get(i);o||(o=new Set,r.set(i,o));var a=t.module.reference;!e.has(i)&&o.has(a)||(o.add(a),e.add(i),n.addProvider(t.provider,t.module))})}),e.forEach(function(t){t.exportedDirectives.forEach(function(t){return n.addExportedDirective(t)}),t.exportedPipes.forEach(function(t){return n.addExportedPipe(t)})}),t.forEach(function(t){t.exportedDirectives.forEach(function(t){return n.addDirective(t)}),t.exportedPipes.forEach(function(t){return n.addPipe(t)})}),n},t.prototype._getIdentifierMetadata=function(t){return{reference:t=nn(t)}},t.prototype.isInjectable=function(t){return this._reflector.tryAnnotations(t).some(function(t){return G.isTypeOf(t)})},t.prototype.getInjectableSummary=function(t){return{summaryKind:jn.Injectable,type:this._getTypeMetadata(t,null,!1)}},t.prototype.getInjectableMetadata=function(t,e,n){void 0===e&&(e=null),void 0===n&&(n=!0);var r=this._loadSummary(t,jn.Injectable),i=r?r.type:this._getTypeMetadata(t,e,n),o=this._reflector.annotations(t).filter(function(t){return G.isTypeOf(t)});if(0===o.length)return null;var a=o[o.length-1];return{symbol:t,type:i,providedIn:a.providedIn,useValue:a.useValue,useClass:a.useClass,useExisting:a.useExisting,useFactory:a.useFactory,deps:a.deps}},t.prototype._getTypeMetadata=function(t,e,n){void 0===e&&(e=null),void 0===n&&(n=!0);var r,i,o=this._getIdentifierMetadata(t);return{reference:o.reference,diDeps:this._getDependenciesMetadata(o.reference,e,n),lifecycleHooks:(r=this._reflector,i=o.reference,ma.filter(function(t){return ga(r,t,i)}))}},t.prototype._getFactoryMetadata=function(t,e){return void 0===e&&(e=null),{reference:t=nn(t),diDeps:this._getDependenciesMetadata(t,e)}},t.prototype.getPipeMetadata=function(t){var e=this._pipeCache.get(t);return e||this._reportError(Ye("Illegal state: getPipeMetadata can only be called after loadNgModuleDirectiveAndPipeMetadata for a module that declares it. Pipe "+Dp(t)+"."),t),e||null},t.prototype.getPipeSummary=function(t){var e=this._loadSummary(t,jn.Pipe);return e||this._reportError(Ye("Illegal state: Could not load the summary for pipe "+Dp(t)+"."),t),e},t.prototype.getOrLoadPipeMetadata=function(t){var e=this._pipeCache.get(t);return e||(e=this._loadPipeMetadata(t)),e},t.prototype._loadPipeMetadata=function(t){t=nn(t);var e=this._pipeResolver.resolve(t),n=new Fn({type:this._getTypeMetadata(t),name:e.name,pure:!!e.pure});return this._pipeCache.set(t,n),this._summaryCache.set(t,n.toSummary()),n},t.prototype._getDependenciesMetadata=function(t,e,n){var r=this;void 0===n&&(n=!0);var i=!1,o=(e||this._reflector.parameters(t)||[]).map(function(t){var e=!1,n=!1,o=!1,a=!1,s=!1,u=null;return Array.isArray(t)?t.forEach(function(t){$.isTypeOf(t)?n=!0:Q.isTypeOf(t)?o=!0:Z.isTypeOf(t)?a=!0:X.isTypeOf(t)?s=!0:T.isTypeOf(t)?(e=!0,u=t.attributeName):S.isTypeOf(t)?u=t.token:M.isTypeOf(t)||t instanceof vn?u=t:kp(t)&&null==u&&(u=t)}):u=t,null==u?(i=!0,null):{isAttribute:e,isHost:n,isSelf:o,isSkipSelf:a,isOptional:s,token:r._getTokenMetadata(u)}});if(i){var a=o.map(function(t){return t?Dp(t.token):"?"}).join(", "),s="Can't resolve all parameters for "+Dp(t)+": ("+a+").";n||this._config.strictInjectionParameters?this._reportError(Ye(s),t):this._console.warn("Warning: "+s+" This will become an error in Angular v6.x")}return o},t.prototype._getTokenMetadata=function(t){return"string"==typeof(t=nn(t))?{value:t}:{identifier:{reference:t}}},t.prototype._getProvidersMetadata=function(t,e,n,r,i){var a=this;return void 0===r&&(r=[]),t.forEach(function(s,u){if(Array.isArray(s))a._getProvidersMetadata(s,e,n,r);else{var c=void 0;if((s=nn(s))&&"object"==typeof s&&s.hasOwnProperty("provide"))a._validateProvider(s),c=new Hn(s.provide,s);else{if(!kp(s)){if(void 0===s)return void a._reportError(Ye("Encountered undefined provider! Usually this means you have a circular dependencies. This might be caused by using 'barrel' index.ts files."));var l=t.reduce(function(t,e,n){return n<u?t.push(""+Dp(e)):n==u?t.push("?"+Dp(e)+"?"):n==u+1&&t.push("..."),t},[]).join(", ");return void a._reportError(Ye("Invalid "+(n||"provider")+" - only instances of Provider and Type are allowed, got: ["+l+"]"),i)}c=new Hn(s,{useClass:s})}c.token===a._reflector.resolveExternalReference(gn.ANALYZE_FOR_ENTRY_COMPONENTS)?e.push.apply(e,Object(o.__spread)(a._getEntryComponentsFromProvider(c,i))):r.push(a.getProviderMetadata(c))}}),r},t.prototype._validateProvider=function(t){t.hasOwnProperty("useClass")&&null==t.useClass&&this._reportError(Ye("Invalid provider for "+Dp(t.provide)+". useClass cannot be "+t.useClass+".\n           Usually it happens when:\n           1. There's a circular dependency (might be caused by using index.ts (barrel) files).\n           2. Class was used before it was declared. Use forwardRef in this case."))},t.prototype._getEntryComponentsFromProvider=function(t,e){var n,r,i=this,o=[],a=[];return t.useFactory||t.useExisting||t.useClass?(this._reportError(Ye("The ANALYZE_FOR_ENTRY_COMPONENTS token only supports useValue!"),e),[]):t.multi?(n=t.useValue,r=a,Be(n,new Ap,r),a.forEach(function(t){var e=i._getEntryComponentMetadata(t.reference,!1);e&&o.push(e)}),o):(this._reportError(Ye("The ANALYZE_FOR_ENTRY_COMPONENTS token only supports 'multi = true'!"),e),[])},t.prototype._getEntryComponentMetadata=function(t,e){void 0===e&&(e=!0);var n=this.getNonNormalizedDirectiveMetadata(t);if(n&&n.metadata.isComponent)return{componentType:t,componentFactory:n.metadata.componentFactory};var r=this._loadSummary(t,jn.Directive);if(r&&r.isComponent)return{componentType:t,componentFactory:r.componentFactory};if(e)throw Ye(t.name+" cannot be used as an entry component.");return null},t.prototype._getInjectableTypeMetadata=function(t,e){void 0===e&&(e=null);var n=this._loadSummary(t,jn.Injectable);return n?n.type:this._getTypeMetadata(t,e)},t.prototype.getProviderMetadata=function(t){var e=void 0,n=null,r=null,i=this._getTokenMetadata(t.token);return t.useClass?(e=(n=this._getInjectableTypeMetadata(t.useClass,t.dependencies)).diDeps,t.token===t.useClass&&(i={identifier:n})):t.useFactory&&(e=(r=this._getFactoryMetadata(t.useFactory,t.dependencies)).diDeps),{token:i,useClass:n,useValue:t.useValue,useFactory:r,useExisting:t.useExisting?this._getTokenMetadata(t.useExisting):void 0,deps:e,multi:t.multi}},t.prototype._getQueriesMetadata=function(t,e,n){var r=this,i=[];return Object.keys(t).forEach(function(o){var a=t[o];a.isViewQuery===e&&i.push(r._getQueryMetadata(a,o,n))}),i},t.prototype._queryVarBindings=function(t){return t.split(/\s*,\s*/)},t.prototype._getQueryMetadata=function(t,e,n){var r,i=this;return"string"==typeof t.selector?r=this._queryVarBindings(t.selector).map(function(t){return i._getTokenMetadata(t)}):t.selector?r=[this._getTokenMetadata(t.selector)]:(this._reportError(Ye("Can't construct a query for the property \""+e+'" of "'+Dp(n)+"\" since the query selector wasn't defined."),n),r=[]),{selectors:r,first:t.first,descendants:t.descendants,propertyName:e,read:t.read?this._getTokenMetadata(t.read):null}},t.prototype._reportError=function(t,e,n){if(!this._errorCollector)throw t;this._errorCollector(t,e),n&&this._errorCollector(t,n)},t}();function Tp(t){return(e=function t(e,n){if(void 0===n&&(n=[]),e)for(var r=0;r<e.length;r++){var i=nn(e[r]);Array.isArray(i)?t(i,n):n.push(i)}return n}(t))?Array.from(new Set(e)):[];var e}function kp(t){return t instanceof vn||t instanceof J}var Ap=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(o.__extends)(e,t),e.prototype.visitOther=function(t,e){e.push({reference:t})},e}(He);function Dp(t){return t instanceof vn?t.name+" in "+t.filePath:en(t)}var Ip=function(t){function e(e,n){return t.call(this,n,e)||this}return Object(o.__extends)(e,t),e}(os),Pp=function(){return function(t,e){var n=this;this.reflector=t,this.component=e,this.errors=[],this.viewQueries=(r=e,i=1,o=new Map,r.viewQueries&&r.viewQueries.forEach(function(t){return Vp(o,{meta:t,queryId:i++})}),o),this.viewProviders=new Map,e.viewProviders.forEach(function(t){null==n.viewProviders.get(In(t.token))&&n.viewProviders.set(In(t.token),!0)});var r,i,o}}(),Rp=function(){function t(t,e,n,r,i,o,a,s,u){var c=this;this.viewContext=t,this._parent=e,this._isViewRoot=n,this._directiveAsts=r,this._sourceSpan=u,this._transformedProviders=new Map,this._seenProviders=new Map,this._queriedTokens=new Map,this.transformedHasViewContainer=!1,this._attrs={},i.forEach(function(t){return c._attrs[t.name]=t.value});var l,d,p,f,h,m,g,y=r.map(function(t){return t.directive});if(this._allProviders=(l=y,d=u,p=t.errors,f=new Map,l.forEach(function(t){Bp([{token:{identifier:t.type},useClass:t.type}],t.isComponent?sd.Component:sd.Directive,!0,d,p,f,!1)}),l.filter(function(t){return t.isComponent}).concat(l.filter(function(t){return!t.isComponent})).forEach(function(t){Bp(t.providers,sd.PublicService,!1,d,p,f,!1),Bp(t.viewProviders,sd.PrivateService,!1,d,p,f,!1)}),f),this._contentQueries=(h=y,m=s,g=new Map,h.forEach(function(t,e){t.queries&&t.queries.forEach(function(t){return Vp(g,{meta:t,queryId:m++})})}),g),Array.from(this._allProviders.values()).forEach(function(t){c._addQueryReadsTo(t.token,t.token,c._queriedTokens)}),a){var _=_n(this.viewContext.reflector,gn.TemplateRef);this._addQueryReadsTo(_,_,this._queriedTokens)}o.forEach(function(t){var e=t.value||_n(c.viewContext.reflector,gn.ElementRef);c._addQueryReadsTo({value:t.name},e,c._queriedTokens)}),this._queriedTokens.get(this.viewContext.reflector.resolveExternalReference(gn.ViewContainerRef))&&(this.transformedHasViewContainer=!0),Array.from(this._allProviders.values()).forEach(function(t){(t.eager||c._queriedTokens.get(In(t.token)))&&c._getOrCreateLocalProvider(t.providerType,t.token,!0)})}return t.prototype.afterElement=function(){var t=this;Array.from(this._allProviders.values()).forEach(function(e){t._getOrCreateLocalProvider(e.providerType,e.token,!1)})},Object.defineProperty(t.prototype,"transformProviders",{get:function(){var t=[],e=[];return this._transformedProviders.forEach(function(n){n.eager?e.push(n):t.push(n)}),t.concat(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"transformedDirectiveAsts",{get:function(){var t=this.transformProviders.map(function(t){return t.token.identifier}),e=this._directiveAsts.slice();return e.sort(function(e,n){return t.indexOf(e.directive.type)-t.indexOf(n.directive.type)}),e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"queryMatches",{get:function(){var t=[];return this._queriedTokens.forEach(function(e){t.push.apply(t,Object(o.__spread)(e))}),t},enumerable:!0,configurable:!0}),t.prototype._addQueryReadsTo=function(t,e,n){this._getQueriesFor(t).forEach(function(t){var r=t.meta.read||e,i=In(r),o=n.get(i);o||(o=[],n.set(i,o)),o.push({queryId:t.queryId,value:r})})},t.prototype._getQueriesFor=function(t){for(var e,n=[],r=this,i=0;null!==r;)(e=r._contentQueries.get(In(t)))&&n.push.apply(n,Object(o.__spread)(e.filter(function(t){return t.meta.descendants||i<=1}))),r._directiveAsts.length>0&&i++,r=r._parent;return(e=this.viewContext.viewQueries.get(In(t)))&&n.push.apply(n,Object(o.__spread)(e)),n},t.prototype._getOrCreateLocalProvider=function(t,e,n){var r=this,i=this._allProviders.get(In(e));if(!i||(t===sd.Directive||t===sd.PublicService)&&i.providerType===sd.PrivateService||(t===sd.PrivateService||t===sd.PublicService)&&i.providerType===sd.Builtin)return null;var o=this._transformedProviders.get(In(e));if(o)return o;if(null!=this._seenProviders.get(In(e)))return this.viewContext.errors.push(new Ip("Cannot instantiate cyclic dependency! "+Dn(e),this._sourceSpan)),null;this._seenProviders.set(In(e),!0);var a=i.providers.map(function(t){var e=t.useValue,o=t.useExisting,a=void 0;if(null!=t.useExisting){var s=r._getDependency(i.providerType,{token:t.useExisting},n);null!=s.token?o=s.token:(o=null,e=s.value)}else if(t.useFactory){a=(t.deps||t.useFactory.diDeps).map(function(t){return r._getDependency(i.providerType,t,n)})}else if(t.useClass){a=(t.deps||t.useClass.diDeps).map(function(t){return r._getDependency(i.providerType,t,n)})}return Fp(t,{useExisting:o,useValue:e,deps:a})});return o=Lp(i,{eager:n,providers:a}),this._transformedProviders.set(In(e),o),o},t.prototype._getLocalDependency=function(t,e,n){if(void 0===n&&(n=!1),e.isAttribute){var r=this._attrs[e.token.value];return{isValue:!0,value:null==r?null:r}}if(null!=e.token){if(t===sd.Directive||t===sd.Component){if(In(e.token)===this.viewContext.reflector.resolveExternalReference(gn.Renderer)||In(e.token)===this.viewContext.reflector.resolveExternalReference(gn.ElementRef)||In(e.token)===this.viewContext.reflector.resolveExternalReference(gn.ChangeDetectorRef)||In(e.token)===this.viewContext.reflector.resolveExternalReference(gn.TemplateRef))return e;In(e.token)===this.viewContext.reflector.resolveExternalReference(gn.ViewContainerRef)&&(this.transformedHasViewContainer=!0)}if(In(e.token)===this.viewContext.reflector.resolveExternalReference(gn.Injector))return e;if(null!=this._getOrCreateLocalProvider(t,e.token,n))return e}return null},t.prototype._getDependency=function(t,e,n){void 0===n&&(n=!1);var r=this,i=n,o=null;if(e.isSkipSelf||(o=this._getLocalDependency(t,e,n)),e.isSelf)!o&&e.isOptional&&(o={isValue:!0,value:null});else{for(;!o&&r._parent;){var a=r;r=r._parent,a._isViewRoot&&(i=!1),o=r._getLocalDependency(sd.PublicService,e,i)}o||(o=!e.isHost||this.viewContext.component.isHost||this.viewContext.component.type.reference===In(e.token)||null!=this.viewContext.viewProviders.get(In(e.token))?e:e.isOptional?{isValue:!0,value:null}:null)}return o||this.viewContext.errors.push(new Ip("No provider for "+Dn(e.token),this._sourceSpan)),o},t}(),Np=function(){function t(t,e,n,r){var i=this;this.reflector=t,this._transformedProviders=new Map,this._seenProviders=new Map,this._errors=[],this._allProviders=new Map,e.transitiveModule.modules.forEach(function(t){Bp([{token:{identifier:t},useClass:t}],sd.PublicService,!0,r,i._errors,i._allProviders,!0)}),Bp(e.transitiveModule.providers.map(function(t){return t.provider}).concat(n),sd.PublicService,!1,r,this._errors,this._allProviders,!1)}return t.prototype.parse=function(){var t=this;if(Array.from(this._allProviders.values()).forEach(function(e){t._getOrCreateLocalProvider(e.token,e.eager)}),this._errors.length>0){var e=this._errors.join("\n");throw new Error("Provider parse errors:\n"+e)}var n=[],r=[];return this._transformedProviders.forEach(function(t){t.eager?r.push(t):n.push(t)}),n.concat(r)},t.prototype._getOrCreateLocalProvider=function(t,e){var n=this,r=this._allProviders.get(In(t));if(!r)return null;var i=this._transformedProviders.get(In(t));if(i)return i;if(null!=this._seenProviders.get(In(t)))return this._errors.push(new Ip("Cannot instantiate cyclic dependency! "+Dn(t),r.sourceSpan)),null;this._seenProviders.set(In(t),!0);var o=r.providers.map(function(t){var i=t.useValue,o=t.useExisting,a=void 0;if(null!=t.useExisting){var s=n._getDependency({token:t.useExisting},e,r.sourceSpan);null!=s.token?o=s.token:(o=null,i=s.value)}else if(t.useFactory){a=(t.deps||t.useFactory.diDeps).map(function(t){return n._getDependency(t,e,r.sourceSpan)})}else if(t.useClass){a=(t.deps||t.useClass.diDeps).map(function(t){return n._getDependency(t,e,r.sourceSpan)})}return Fp(t,{useExisting:o,useValue:i,deps:a})});return i=Lp(r,{eager:e,providers:o}),this._transformedProviders.set(In(t),i),i},t.prototype._getDependency=function(t,e,n){return void 0===e&&(e=!1),t.isSkipSelf||null==t.token||In(t.token)===this.reflector.resolveExternalReference(gn.Injector)||In(t.token)===this.reflector.resolveExternalReference(gn.ComponentFactoryResolver)||this._getOrCreateLocalProvider(t.token,e),t},t}();function Fp(t,e){var n=e.useExisting,r=e.useValue,i=e.deps;return{token:t.token,useClass:t.useClass,useExisting:n,useFactory:t.useFactory,useValue:r,deps:i,multi:t.multi}}function Lp(t,e){var n=e.eager,r=e.providers;return new wd(t.token,t.multiProvider,t.eager||n,r,t.providerType,t.lifecycleHooks,t.sourceSpan,t.isModule)}function Bp(t,e,n,r,i,o,a){t.forEach(function(t){var s=o.get(In(t.token));if(null!=s&&!!s.multiProvider!=!!t.multi&&i.push(new Ip("Mixing multi and non multi provider is not possible for token "+Dn(s.token),r)),s)t.multi||(s.providers.length=0),s.providers.push(t);else{var u=t.token.identifier&&t.token.identifier.lifecycleHooks?t.token.identifier.lifecycleHooks:[],c=!(t.useClass||t.useExisting||t.useFactory);s=new wd(t.token,!!t.multi,n||c,[t],e,u,r,a),o.set(In(t.token),s)}})}function Vp(t,e){e.meta.selectors.forEach(function(n){var r=t.get(In(n));r||(r=[],t.set(In(n),r)),r.push(e)})}function zp(t,e){var n=0;e.eager||(n|=4096),e.providerType===sd.PrivateService&&(n|=8192),e.isModule&&(n|=1073741824),e.lifecycleHooks.forEach(function(t){t!==pa.OnDestroy&&e.providerType!==sd.Directive&&e.providerType!==sd.Component||(n|=qp(t))});var r=e.multiProvider?function(t,e,n){var r=[],i=[],o=n.map(function(e,n){var r;if(e.useClass){var i=a(n,e.deps||e.useClass.diDeps);r=t.importExpr(e.useClass.reference).instantiate(i)}else if(e.useFactory){var i=a(n,e.deps||e.useFactory.diDeps);r=t.importExpr(e.useFactory.reference).callFn(i)}else if(e.useExisting){var i=a(n,[{token:e.useExisting}]);r=i[0]}else r=cp(t,e.useValue);return r});return{providerExpr:De(i,[new ae(ke(o))],gt),flags:1024|e,depsExpr:ke(r)};function a(e,n){return n.map(function(n,o){var a="p"+e+"_"+o;return i.push(new qt(a,mt)),r.push(Wp(t,n)),Ee(a)})}}(t,n,e.providers):Hp(t,n,e.providerType,e.providers[0]);return{providerExpr:r.providerExpr,flags:r.flags,depsExpr:r.depsExpr,tokenExpr:Up(t,e.token)}}function Hp(t,e,n,r){var i,o;return n===sd.Directive||n===sd.Component?(i=t.importExpr(r.useClass.reference),e|=16384,o=r.deps||r.useClass.diDeps):r.useClass?(i=t.importExpr(r.useClass.reference),e|=512,o=r.deps||r.useClass.diDeps):r.useFactory?(i=t.importExpr(r.useFactory.reference),e|=1024,o=r.deps||r.useFactory.diDeps):r.useExisting?(i=te,e|=2048,o=[{token:r.useExisting}]):(i=cp(t,r.useValue),e|=256,o=[]),{providerExpr:i,flags:e,depsExpr:ke(o.map(function(e){return Wp(t,e)}))}}function Up(t,e){return e.identifier?t.importExpr(e.identifier.reference):Pe(e.value)}function Wp(t,e){var n=e.isValue?cp(t,e.value):Up(t,e.token),r=0;return e.isSkipSelf&&(r|=1),e.isOptional&&(r|=2),e.isSelf&&(r|=4),e.isValue&&(r|=8),0===r?n:ke([Pe(r),n])}function qp(t){var e=0;switch(t){case pa.AfterContentChecked:e=2097152;break;case pa.AfterContentInit:e=1048576;break;case pa.AfterViewChecked:e=8388608;break;case pa.AfterViewInit:e=4194304;break;case pa.DoCheck:e=262144;break;case pa.OnChanges:e=524288;break;case pa.OnDestroy:e=131072;break;case pa.OnInit:e=65536}return e}function Gp(t,e,n,r){var i=r.map(function(t){return e.importExpr(t.componentFactory)}),o=_n(t,gn.ComponentFactoryResolver),a={diDeps:[{isValue:!0,value:ke(i)},{token:o,isSkipSelf:!0,isOptional:!0},{token:_n(t,gn.NgModuleRef)}],lifecycleHooks:[],reference:t.resolveExternalReference(gn.CodegenComponentFactoryResolver)},s=Hp(e,n,sd.PrivateService,{token:o,multi:!1,useClass:a});return{providerExpr:s.providerExpr,flags:s.flags,depsExpr:s.depsExpr,tokenExpr:Up(e,o)}}var Yp=function(){return function(t){this.ngModuleFactoryVar=t}}(),Kp=Ee("_l"),Xp=function(){function t(t){this.reflector=t}return t.prototype.compile=function(t,e,n){var r=as("NgModule",e.type),i=e.transitiveModule.entryComponents,o=e.bootstrapComponents,a=new Np(this.reflector,e,n,r),s=[Gp(this.reflector,t,0,i)].concat(a.parse().map(function(e){return zp(t,e)})).map(function(t){var e=t.providerExpr,n=t.depsExpr,r=t.flags,i=t.tokenExpr;return Se(gn.moduleProviderDef).callFn([Pe(r),i,e,n])}),u=Se(gn.moduleDef).callFn([ke(s)]),c=De([new qt(Kp.name)],[new ae(u)],gt),l=En(e.type)+"NgFactory";if(this._createNgModuleFactory(t,e.type.reference,Se(gn.createModuleFactory).callFn([t.importExpr(e.type.reference),ke(o.map(function(e){return t.importExpr(e.reference)})),c])),e.id){var d="string"==typeof e.id?Pe(e.id):t.importExpr(e.id),p=Se(gn.RegisterModuleFactoryFn).callFn([d,Ee(l)]).toStmt();t.statements.push(p)}return new Yp(l)},t.prototype.createStub=function(t,e){this._createNgModuleFactory(t,e,te)},t.prototype._createNgModuleFactory=function(t,e,n){var r=Ee(En({reference:e})+"NgFactory").set(n).toDeclStmt(Me(gn.NgModuleFactory,[Te(t.importExpr(e))],[rt.Const]),[It.Final,It.Exported]);t.statements.push(r)},t}(),Qp=function(){function t(t){this._reflector=t}return t.prototype.isNgModule=function(t){return this._reflector.annotations(t).some(q.isTypeOf)},t.prototype.resolve=function(t,e){void 0===e&&(e=!0);var n=Pd(this._reflector.annotations(t),q.isTypeOf);if(n)return n;if(e)throw new Error("No NgModule metadata found for '"+en(t)+"'.");return null},t}();function Zp(t){var e=new Jp,n=Ji.createRoot();return(Array.isArray(t)?t:[t]).forEach(function(t){if(t instanceof ne)t.visitStatement(e,n);else if(t instanceof Ot)t.visitExpression(e,n);else{if(!(t instanceof ut))throw new Error("Don't know how to print debug info for "+t);t.visitType(e,n)}}),n.toSource()}var $p=function(){function t(){}return t.prototype.emitStatementsAndContext=function(t,e,n,r,i,a){void 0===n&&(n=""),void 0===r&&(r=!0);var s=new Jp(i,a),u=Ji.createRoot();s.visitAllStatements(e,u);var c=n?n.split("\n"):[];s.reexports.forEach(function(t,e){var n=t.map(function(t){return t.name+" as "+t.as}).join(",");c.push("export {"+n+"} from '"+e+"';")}),s.importsWithPrefixes.forEach(function(t,e){c.push("import * as "+t+" from '"+e+"';")});var l=r?u.toSourceMapGenerator(t,c.length).toJsComment():"",d=Object(o.__spread)(c,[u.toSource(),l]);return l&&d.push(""),u.setPreambleLineCount(c.length),{sourceText:d.join("\n"),context:u}},t.prototype.emitStatements=function(t,e,n){return void 0===n&&(n=""),this.emitStatementsAndContext(t,e,n).sourceText},t}(),Jp=function(t){function e(e,n){var r=t.call(this,!1)||this;return r.referenceFilter=e,r.importFilter=n,r.typeExpression=0,r.importsWithPrefixes=new Map,r.reexports=new Map,r}return Object(o.__extends)(e,t),e.prototype.visitType=function(t,e,n){void 0===n&&(n="any"),t?(this.typeExpression++,t.visitType(this,e),this.typeExpression--):e.print(null,n)},e.prototype.visitLiteralExpr=function(e,n){var r=e.value;return null==r&&e.type!=gt?(n.print(e,"("+r+" as any)"),null):t.prototype.visitLiteralExpr.call(this,e,n)},e.prototype.visitLiteralArrayExpr=function(e,n){0===e.entries.length&&n.print(e,"(");var r=t.prototype.visitLiteralArrayExpr.call(this,e,n);return 0===e.entries.length&&n.print(e," as any[])"),r},e.prototype.visitExternalExpr=function(t,e){return this._visitIdentifier(t.value,t.typeParams,e),null},e.prototype.visitAssertNotNullExpr=function(e,n){var r=t.prototype.visitAssertNotNullExpr.call(this,e,n);return n.print(e,"!"),r},e.prototype.visitDeclareVarStmt=function(t,e){if(t.hasModifier(It.Exported)&&t.value instanceof Bt&&!t.type){var n=t.value.value,r=n.name,i=n.moduleName;if(i){var o=this.reexports.get(i);return o||(o=[],this.reexports.set(i,o)),o.push({name:r,as:t.name}),null}}return t.hasModifier(It.Exported)&&e.print(t,"export "),t.hasModifier(It.Final)?e.print(t,"const"):e.print(t,"var"),e.print(t," "+t.name),this._printColonType(t.type,e),t.value&&(e.print(t," = "),t.value.visitExpression(this,e)),e.println(t,";"),null},e.prototype.visitWrappedNodeExpr=function(t,e){throw new Error("Cannot visit a WrappedNodeExpr when outputting Typescript.")},e.prototype.visitCastExpr=function(t,e){return e.print(t,"(<"),t.type.visitType(this,e),e.print(t,">"),t.value.visitExpression(this,e),e.print(t,")"),null},e.prototype.visitInstantiateExpr=function(t,e){return e.print(t,"new "),this.typeExpression++,t.classExpr.visitExpression(this,e),this.typeExpression--,e.print(t,"("),this.visitAllExpressions(t.args,e,","),e.print(t,")"),null},e.prototype.visitDeclareClassStmt=function(t,e){var n=this;return e.pushClass(t),t.hasModifier(It.Exported)&&e.print(t,"export "),e.print(t,"class "+t.name),null!=t.parent&&(e.print(t," extends "),this.typeExpression++,t.parent.visitExpression(this,e),this.typeExpression--),e.println(t," {"),e.incIndent(),t.fields.forEach(function(t){return n._visitClassField(t,e)}),null!=t.constructorMethod&&this._visitClassConstructor(t,e),t.getters.forEach(function(t){return n._visitClassGetter(t,e)}),t.methods.forEach(function(t){return n._visitClassMethod(t,e)}),e.decIndent(),e.println(t,"}"),e.popClass(),null},e.prototype._visitClassField=function(t,e){t.hasModifier(It.Private)&&e.print(null,"/*private*/ "),t.hasModifier(It.Static)&&e.print(null,"static "),e.print(null,t.name),this._printColonType(t.type,e),t.initializer&&(e.print(null," = "),t.initializer.visitExpression(this,e)),e.println(null,";")},e.prototype._visitClassGetter=function(t,e){t.hasModifier(It.Private)&&e.print(null,"private "),e.print(null,"get "+t.name+"()"),this._printColonType(t.type,e),e.println(null," {"),e.incIndent(),this.visitAllStatements(t.body,e),e.decIndent(),e.println(null,"}")},e.prototype._visitClassConstructor=function(t,e){e.print(t,"constructor("),this._visitParams(t.constructorMethod.params,e),e.println(t,") {"),e.incIndent(),this.visitAllStatements(t.constructorMethod.body,e),e.decIndent(),e.println(t,"}")},e.prototype._visitClassMethod=function(t,e){t.hasModifier(It.Private)&&e.print(null,"private "),e.print(null,t.name+"("),this._visitParams(t.params,e),e.print(null,")"),this._printColonType(t.type,e,"void"),e.println(null," {"),e.incIndent(),this.visitAllStatements(t.body,e),e.decIndent(),e.println(null,"}")},e.prototype.visitFunctionExpr=function(t,e){return t.name&&(e.print(t,"function "),e.print(t,t.name)),e.print(t,"("),this._visitParams(t.params,e),e.print(t,")"),this._printColonType(t.type,e,"void"),t.name||e.print(t," => "),e.println(t,"{"),e.incIndent(),this.visitAllStatements(t.statements,e),e.decIndent(),e.print(t,"}"),null},e.prototype.visitDeclareFunctionStmt=function(t,e){return t.hasModifier(It.Exported)&&e.print(t,"export "),e.print(t,"function "+t.name+"("),this._visitParams(t.params,e),e.print(t,")"),this._printColonType(t.type,e,"void"),e.println(t," {"),e.incIndent(),this.visitAllStatements(t.statements,e),e.decIndent(),e.println(t,"}"),null},e.prototype.visitTryCatchStmt=function(t,e){e.println(t,"try {"),e.incIndent(),this.visitAllStatements(t.bodyStmts,e),e.decIndent(),e.println(t,"} catch ("+Qi.name+") {"),e.incIndent();var n=[Zi.set(Qi.prop("stack",null)).toDeclStmt(null,[It.Final])].concat(t.catchStmts);return this.visitAllStatements(n,e),e.decIndent(),e.println(t,"}"),null},e.prototype.visitBuiltinType=function(t,e){var n;switch(t.name){case at.Bool:n="boolean";break;case at.Dynamic:n="any";break;case at.Function:n="Function";break;case at.Number:case at.Int:n="number";break;case at.String:n="string";break;case at.None:n="never";break;default:throw new Error("Unsupported builtin type "+t.name)}return e.print(null,n),null},e.prototype.visitExpressionType=function(t,e){var n=this;return t.value.visitExpression(this,e),null!==t.typeParams&&(e.print(null,"<"),this.visitAllObjects(function(t){return n.visitType(t,e)},t.typeParams,e,","),e.print(null,">")),null},e.prototype.visitArrayType=function(t,e){return this.visitType(t.of,e),e.print(null,"[]"),null},e.prototype.visitMapType=function(t,e){return e.print(null,"{[key: string]:"),this.visitType(t.valueType,e),e.print(null,"}"),null},e.prototype.getBuiltinMethodName=function(t){var e;switch(t){case Ct.ConcatArray:e="concat";break;case Ct.SubscribeObservable:e="subscribe";break;case Ct.Bind:e="bind";break;default:throw new Error("Unknown builtin method: "+t)}return e},e.prototype._visitParams=function(t,e){var n=this;this.visitAllObjects(function(t){e.print(null,t.name),n._printColonType(t.type,e)},t,e,",")},e.prototype._visitIdentifier=function(t,e,n){var r=this,i=t.name,o=t.moduleName;if(this.referenceFilter&&this.referenceFilter(t))n.print(null,"(null as any)");else{if(o&&(!this.importFilter||!this.importFilter(t))){var a=this.importsWithPrefixes.get(o);null==a&&(a="i"+this.importsWithPrefixes.size,this.importsWithPrefixes.set(o,a)),n.print(null,a+".")}if(n.print(null,i),this.typeExpression>0)(e||[]).length>0&&(n.print(null,"<"),this.visitAllObjects(function(t){return t.visitType(r,n)},e,n,","),n.print(null,">"))}},e.prototype._printColonType=function(t,e,n){t!==gt&&(e.print(null,":"),this.visitType(t,e,n))},e}(to),tf=function(){function t(t){this._reflector=t}return t.prototype.isPipe=function(t){var e=this._reflector.annotations(nn(t));return e&&e.some(V.isTypeOf)},t.prototype.resolve=function(t,e){void 0===e&&(e=!0);var n=this._reflector.annotations(nn(t));if(n){var r=Pd(n,V.isTypeOf);if(r)return r}if(e)throw new Error("No Pipe decorator found on "+en(t));return null},t}(),ef=["zero","one","two","few","many","other"];function nf(t){var e=new af;return new rf(_u(e,t),e.isExpanded,e.errors)}var rf=function(){return function(t,e,n){this.nodes=t,this.expanded=e,this.errors=n}}(),of=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(o.__extends)(e,t),e}(os),af=function(){function t(){this.isExpanded=!1,this.errors=[]}return t.prototype.visitElement=function(t,e){return new gu(t.name,t.attrs,_u(this,t.children),t.sourceSpan,t.startSourceSpan,t.endSourceSpan)},t.prototype.visitAttribute=function(t,e){return t},t.prototype.visitText=function(t,e){return t},t.prototype.visitComment=function(t,e){return t},t.prototype.visitExpansion=function(t,e){return this.isExpanded=!0,"plural"==t.type?(s=t,u=this.errors,c=s.cases.map(function(t){-1!=ef.indexOf(t.value)||t.value.match(/^=\d+$/)||u.push(new of(t.valueSourceSpan,'Plural cases should be "=<number>" or one of '+ef.join(", ")));var e=nf(t.expression);return u.push.apply(u,Object(o.__spread)(e.errors)),new gu("ng-template",[new mu("ngPluralCase",""+t.value,t.valueSourceSpan)],e.nodes,t.sourceSpan,t.sourceSpan,t.sourceSpan)}),l=new mu("[ngPlural]",s.switchValue,s.switchValueSourceSpan),new gu("ng-container",[l],c,s.sourceSpan,s.sourceSpan,s.sourceSpan)):(n=t,r=this.errors,i=n.cases.map(function(t){var e=nf(t.expression);return r.push.apply(r,Object(o.__spread)(e.errors)),"other"===t.value?new gu("ng-template",[new mu("ngSwitchDefault","",t.valueSourceSpan)],e.nodes,t.sourceSpan,t.sourceSpan,t.sourceSpan):new gu("ng-template",[new mu("ngSwitchCase",""+t.value,t.valueSourceSpan)],e.nodes,t.sourceSpan,t.sourceSpan,t.sourceSpan)}),a=new mu("[ngSwitch]",n.switchValue,n.switchValueSourceSpan),new gu("ng-container",[a],i,n.sourceSpan,n.sourceSpan,n.sourceSpan));var n,r,i,a,s,u,c,l},t.prototype.visitExpansionCase=function(t,e){throw new Error("Should not be reached")},t}();var sf,uf=/^(?:(?:(?:(bind-)|(let-)|(ref-|#)|(on-)|(bindon-)|(@))(.+))|\[\(([^\)]+)\)\]|\[([^\]]+)\]|\(([^\)]+)\))$/,cf="class";function lf(){return sf||(sf=j.parse("*")[0]),sf}var df=function(t){function e(e,n,r){return t.call(this,n,e,r)||this}return Object(o.__extends)(e,t),e}(os),pf=function(){return function(t,e,n){this.templateAst=t,this.usedPipes=e,this.errors=n}}(),ff=function(){function t(t,e,n,r,i,o,a){this._config=t,this._reflector=e,this._exprParser=n,this._schemaRegistry=r,this._htmlParser=i,this._console=o,this.transforms=a}return Object.defineProperty(t.prototype,"expressionParser",{get:function(){return this._exprParser},enumerable:!0,configurable:!0}),t.prototype.parse=function(t,e,n,r,i,o,a){var s=this.tryParse(t,e,n,r,i,o,a),u=s.errors.filter(function(t){return t.level===ts.WARNING}),c=s.errors.filter(function(t){return t.level===ts.ERROR});if(u.length>0&&this._console.warn("Template parse warnings:\n"+u.join("\n")),c.length>0)throw Ye("Template parse errors:\n"+c.join("\n"),c);return{template:s.templateAst,pipes:s.usedPipes}},t.prototype.tryParse=function(t,e,n,r,i,o,a){var s="string"==typeof e?this._htmlParser.parse(e,o,!0,this.getInterpolationConfig(t)):e;return a||(s=Zu(s)),this.tryParseHtml(this.expandHtml(s),t,n,r,i)},t.prototype.tryParseHtml=function(t,e,n,r,i){var a,s=t.errors,u=[];if(t.rootNodes.length>0){var c=jf(n),l=jf(r),d=new Pp(this._reflector,e),p=void 0;e.template&&e.template.interpolation&&(p={start:e.template.interpolation[0],end:e.template.interpolation[1]});var f=new uc(this._exprParser,p,this._schemaRegistry,l,s),h=new hf(this._reflector,this._config,d,c,f,this._schemaRegistry,i,s);a=_u(h,t.rootNodes,bf),s.push.apply(s,Object(o.__spread)(d.errors)),u.push.apply(u,Object(o.__spread)(f.getUsedPipes()))}else a=[];return this._assertNoReferenceDuplicationOnTemplate(a,s),s.length>0?new pf(a,u,s):(this.transforms&&this.transforms.forEach(function(t){a=Ed(t,a)}),new pf(a,u,s))},t.prototype.expandHtml=function(t,e){void 0===e&&(e=!1);var n=t.errors;if(0==n.length||e){var r=nf(t.rootNodes);n.push.apply(n,Object(o.__spread)(r.errors)),t=new Lu(r.nodes,n)}return t},t.prototype.getInterpolationConfig=function(t){if(t.template)return zi.fromArray(t.template.interpolation)},t.prototype._assertNoReferenceDuplicationOnTemplate=function(t,e){var n=[];t.filter(function(t){return!!t.references}).forEach(function(t){return t.references.forEach(function(t){var r=t.name;if(n.indexOf(r)<0)n.push(r);else{var i=new df('Reference "#'+r+'" is defined several times',t.sourceSpan,ts.ERROR);e.push(i)}})})},t}(),hf=function(){function t(t,e,n,r,i,o,a,s){var u=this;this.reflector=t,this.config=e,this.providerViewContext=n,this._bindingParser=i,this._schemaRegistry=o,this._schemas=a,this._targetErrors=s,this.selectorMatcher=new O,this.directivesIndex=new Map,this.ngContentCount=0,this.contentQueryStartId=n.component.viewQueries.length+1,r.forEach(function(t,e){var n=j.parse(t.selector);u.selectorMatcher.addSelectables(n,t),u.directivesIndex.set(t,e)})}return t.prototype.visitExpansion=function(t,e){return null},t.prototype.visitExpansionCase=function(t,e){return null},t.prototype.visitText=function(t,e){var n=e.findNgContentIndex(lf()),r=Ku(t.value),i=this._bindingParser.parseInterpolation(r,t.sourceSpan);return i?new dd(i,n,t.sourceSpan):new ld(r,n,t.sourceSpan)},t.prototype.visitAttribute=function(t,e){return new pd(t.name,t.value,t.sourceSpan)},t.prototype.visitComment=function(t,e){return null},t.prototype.visitElement=function(t,e){var n=this,r=this.contentQueryStartId,i=t.name,a=Vc(t);if(a.type===Ec.SCRIPT||a.type===Ec.STYLE)return null;if(a.type===Ec.STYLESHEET&&Cc(a.hrefAttr))return null;var s=[],u=[],l=[],d=[],p=[],f=[],h=[],m=[],g=!1,y=[],_=c(t.name);t.attrs.forEach(function(t){var e,r,i=n._parseAttr(_,t,s,u,p,l,d);d.push.apply(d,Object(o.__spread)([].map(function(t){return yd.fromParsedVariable(t)})));var a=n._normalizeAttributeName(t.name);a.startsWith("*")&&(e=t.value,r=a.substring("*".length));var c=null!=e;if(c){g&&n._reportError("Can't have multiple template bindings on one element. Use only one attribute prefixed with *",t.sourceSpan),g=!0;var v=[];n._bindingParser.parseInlineTemplateBinding(r,e,t.sourceSpan,h,f,v),m.push.apply(m,Object(o.__spread)(v.map(function(t){return yd.fromParsedVariable(t)})))}i||c||(y.push(n.visitAttribute(t,null)),s.push([t.name,t.value]))});var v=vf(i,s),b=this._parseDirectives(this.selectorMatcher,v),x=b.directives,w=b.matchElement,O=[],C=new Set,E=this._createDirectiveAsts(_,t.name,x,u,l,t.sourceSpan,O,C),S=this._createElementPropertyAsts(t.name,u,C),M=e.isTemplateElement||g,T=new Rp(this.providerViewContext,e.providerContext,M,E,y,O,_,r,t.sourceSpan),k=_u(a.nonBindable?xf:this,t.children,_f.create(_,E,_?e.providerContext:T));T.afterElement();var A,D=""!=a.projectAs?j.parse(a.projectAs)[0]:v,I=e.findNgContentIndex(D);if(a.type===Ec.NG_CONTENT)t.children&&!t.children.every(wf)&&this._reportError("<ng-content> element cannot have content.",t.sourceSpan),A=new jd(this.ngContentCount++,g?null:I,t.sourceSpan);else if(_)this._assertAllEventsPublishedByDirectives(E,p),this._assertNoComponentsNorElementBindingsOnTemplate(E,S,t.sourceSpan),A=new vd(y,p,O,d,T.transformedDirectiveAsts,T.transformProviders,T.transformedHasViewContainer,T.queryMatches,k,g?null:I,t.sourceSpan);else{this._assertElementExists(w,t),this._assertOnlyOneComponent(E,t.sourceSpan);var P=g?null:e.findNgContentIndex(D);A=new _d(i,y,S,p,O,T.transformedDirectiveAsts,T.transformProviders,T.transformedHasViewContainer,T.queryMatches,k,g?null:P,t.sourceSpan,t.endSourceSpan||null)}if(g){var R=this.contentQueryStartId,N=vf("ng-template",h),F=this._parseDirectives(this.selectorMatcher,N).directives,L=new Set,B=this._createDirectiveAsts(!0,i,F,f,[],t.sourceSpan,[],L),V=this._createElementPropertyAsts(i,f,L);this._assertNoComponentsNorElementBindingsOnTemplate(B,V,t.sourceSpan);var z=new Rp(this.providerViewContext,e.providerContext,e.isTemplateElement,B,[],[],!0,R,t.sourceSpan);z.afterElement(),A=new vd([],[],[],m,z.transformedDirectiveAsts,z.transformProviders,z.transformedHasViewContainer,z.queryMatches,[A],I,t.sourceSpan)}return A},t.prototype._parseAttr=function(t,e,n,r,i,a,s){var u=this._normalizeAttributeName(e.name),c=e.value,l=e.sourceSpan,d=[],p=u.match(uf),f=!1;if(null!==p)if(f=!0,null!=p[1])this._bindingParser.parsePropertyBinding(p[7],c,!1,l,n,r);else if(p[2])if(t){var h=p[7];this._parseVariable(h,c,l,s)}else this._reportError('"let-" is only supported on ng-template elements.',l);else if(p[3]){h=p[7];this._parseReference(h,c,l,a)}else p[4]?this._bindingParser.parseEvent(p[7],c,l,n,d):p[5]?(this._bindingParser.parsePropertyBinding(p[7],c,!1,l,n,r),this._parseAssignmentEvent(p[7],c,l,n,d)):p[6]?this._bindingParser.parseLiteralAttr(u,c,l,n,r):p[8]?(this._bindingParser.parsePropertyBinding(p[8],c,!1,l,n,r),this._parseAssignmentEvent(p[8],c,l,n,d)):p[9]?this._bindingParser.parsePropertyBinding(p[9],c,!1,l,n,r):p[10]&&this._bindingParser.parseEvent(p[10],c,l,n,d);else f=this._bindingParser.parsePropertyInterpolation(u,c,l,n,r);return f||this._bindingParser.parseLiteralAttr(u,c,l,n,r),i.push.apply(i,Object(o.__spread)(d.map(function(t){return md.fromParsedEvent(t)}))),f},t.prototype._normalizeAttributeName=function(t){return/^data-/i.test(t)?t.substring(5):t},t.prototype._parseVariable=function(t,e,n,r){t.indexOf("-")>-1&&this._reportError('"-" is not allowed in variable names',n),r.push(new yd(t,e,n))},t.prototype._parseReference=function(t,e,n,r){t.indexOf("-")>-1&&this._reportError('"-" is not allowed in reference names',n),r.push(new gf(t,e,n))},t.prototype._parseAssignmentEvent=function(t,e,n,r,i){this._bindingParser.parseEvent(t+"Change",e+"=$event",n,r,i)},t.prototype._parseDirectives=function(t,e){var n=this,r=new Array(this.directivesIndex.size),i=!1;return t.match(e,function(t,e){r[n.directivesIndex.get(e)]=e,i=i||t.hasElementSelector()}),{directives:r.filter(function(t){return!!t}),matchElement:i}},t.prototype._createDirectiveAsts=function(t,e,n,r,i,o,a,s){var u=this,c=new Set,l=null,d=n.map(function(t){var n=new is(o.start,o.end,"Directive "+En(t.type));t.isComponent&&(l=t);var d=[],p=u._bindingParser.createDirectiveHostPropertyAsts(t,e,n).map(function(t){return hd.fromBoundProperty(t)});p=u._checkPropertiesInSchema(e,p);var f=u._bindingParser.createDirectiveHostEventAsts(t,n);u._createDirectivePropertyAsts(t.inputs,r,d,s),i.forEach(function(e){(0===e.value.length&&t.isComponent||e.isReferenceToDirective(t))&&(a.push(new gd(e.name,yn(t.type.reference),e.value,e.sourceSpan)),c.add(e.name))});var h=f.map(function(t){return md.fromParsedEvent(t)}),m=u.contentQueryStartId;return u.contentQueryStartId+=t.queries.length,new xd(t,d,p,h,m,n)});return i.forEach(function(e){if(e.value.length>0)c.has(e.name)||u._reportError('There is no directive with "exportAs" set to "'+e.value+'"',e.sourceSpan);else if(!l){var n=null;t&&(n=_n(u.reflector,gn.TemplateRef)),a.push(new gd(e.name,n,e.value,e.sourceSpan))}}),d},t.prototype._createDirectivePropertyAsts=function(t,e,n,r){if(t){var i=new Map;e.forEach(function(t){var e=i.get(t.name);e&&!e.isLiteral||i.set(t.name,t)}),Object.keys(t).forEach(function(e){var o=t[e],a=i.get(o);a&&(r.add(a.name),Of(a.expression)||n.push(new bd(e,a.name,a.expression,a.sourceSpan)))})}},t.prototype._createElementPropertyAsts=function(t,e,n){var r=this,i=[];return e.forEach(function(e){if(!e.isLiteral&&!n.has(e.name)){var o=r._bindingParser.createBoundElementProperty(t,e);i.push(hd.fromBoundProperty(o))}}),this._checkPropertiesInSchema(t,i)},t.prototype._findComponentDirectives=function(t){return t.filter(function(t){return t.directive.isComponent})},t.prototype._findComponentDirectiveNames=function(t){return this._findComponentDirectives(t).map(function(t){return En(t.directive.type)})},t.prototype._assertOnlyOneComponent=function(t,e){var n=this._findComponentDirectiveNames(t);n.length>1&&this._reportError("More than one component matched on this element.\nMake sure that only one component's selector can match a given element.\nConflicting components: "+n.join(","),e)},t.prototype._assertElementExists=function(t,e){var n=e.name.replace(/^:xhtml:/,"");if(!t&&!this._schemaRegistry.hasElement(n,this._schemas)){var r="'"+n+"' is not a known element:\n";r+="1. If '"+n+"' is an Angular component, then verify that it is part of this module.\n",n.indexOf("-")>-1?r+="2. If '"+n+"' is a Web Component then add 'CUSTOM_ELEMENTS_SCHEMA' to the '@NgModule.schemas' of this component to suppress this message.":r+="2. To allow any element add 'NO_ERRORS_SCHEMA' to the '@NgModule.schemas' of this component.",this._reportError(r,e.sourceSpan)}},t.prototype._assertNoComponentsNorElementBindingsOnTemplate=function(t,e,n){var r=this,i=this._findComponentDirectiveNames(t);i.length>0&&this._reportError("Components on an embedded template: "+i.join(","),n),e.forEach(function(t){r._reportError("Property binding "+t.name+' not used by any directive on an embedded template. Make sure that the property name is spelled correctly and all directives are listed in the "@NgModule.declarations".',n)})},t.prototype._assertAllEventsPublishedByDirectives=function(t,e){var n=this,r=new Set;t.forEach(function(t){Object.keys(t.directive.outputs).forEach(function(e){var n=t.directive.outputs[e];r.add(n)})}),e.forEach(function(t){null==t.target&&r.has(t.name)||n._reportError("Event binding "+t.fullName+' not emitted by any directive on an embedded template. Make sure that the event name is spelled correctly and all directives are listed in the "@NgModule.declarations".',t.sourceSpan)})},t.prototype._checkPropertiesInSchema=function(t,e){var n=this;return e.filter(function(e){if(0===e.type&&!n._schemaRegistry.hasProperty(t,e.name,n._schemas)){var r="Can't bind to '"+e.name+"' since it isn't a known property of '"+t+"'.";t.startsWith("ng-")?r+="\n1. If '"+e.name+"' is an Angular directive, then add 'CommonModule' to the '@NgModule.imports' of this component.\n2. To allow any property add 'NO_ERRORS_SCHEMA' to the '@NgModule.schemas' of this component.":t.indexOf("-")>-1&&(r+="\n1. If '"+t+"' is an Angular component and it has '"+e.name+"' input, then verify that it is part of this module.\n2. If '"+t+"' is a Web Component then add 'CUSTOM_ELEMENTS_SCHEMA' to the '@NgModule.schemas' of this component to suppress this message.\n3. To allow any property add 'NO_ERRORS_SCHEMA' to the '@NgModule.schemas' of this component."),n._reportError(r,e.sourceSpan)}return!Of(e.value)})},t.prototype._reportError=function(t,e,n){void 0===n&&(n=ts.ERROR),this._targetErrors.push(new os(e,t,n))},t}(),mf=function(){function t(){}return t.prototype.visitElement=function(t,e){var n=Vc(t);if(n.type===Ec.SCRIPT||n.type===Ec.STYLE||n.type===Ec.STYLESHEET)return null;var r=t.attrs.map(function(t){return[t.name,t.value]}),i=vf(t.name,r),o=e.findNgContentIndex(i),a=_u(this,t.children,bf);return new _d(t.name,_u(this,t.attrs),[],[],[],[],[],!1,[],a,o,t.sourceSpan,t.endSourceSpan)},t.prototype.visitComment=function(t,e){return null},t.prototype.visitAttribute=function(t,e){return new pd(t.name,t.value,t.sourceSpan)},t.prototype.visitText=function(t,e){var n=e.findNgContentIndex(lf());return new ld(t.value,n,t.sourceSpan)},t.prototype.visitExpansion=function(t,e){return t},t.prototype.visitExpansionCase=function(t,e){return t},t}(),gf=function(){function t(t,e,n){this.name=t,this.value=e,this.sourceSpan=n}return t.prototype.isReferenceToDirective=function(t){return-1!==(e=t.exportAs,e?e.split(",").map(function(t){return t.trim()}):[]).indexOf(this.value);var e},t}();function yf(t){return t.trim().split(/\s+/g)}var _f=function(){function t(t,e,n,r){this.isTemplateElement=t,this._ngContentIndexMatcher=e,this._wildcardNgContentIndex=n,this.providerContext=r}return t.create=function(e,n,r){var i=new O,o=null,a=n.find(function(t){return t.directive.isComponent});if(a)for(var s=a.directive.template.ngContentSelectors,u=0;u<s.length;u++){"*"===s[u]?o=u:i.addSelectables(j.parse(s[u]),u)}return new t(e,i,o,r)},t.prototype.findNgContentIndex=function(t){var e=[];return this._ngContentIndexMatcher.match(t,function(t,n){e.push(n)}),e.sort(),null!=this._wildcardNgContentIndex&&e.push(this._wildcardNgContentIndex),e.length>0?e[0]:null},t}();function vf(t,e){var n=new j,r=a(t)[1];n.setElement(r);for(var i=0;i<e.length;i++){var o=e[i][0],s=a(o)[1],u=e[i][1];if(n.addAttribute(s,u),o.toLowerCase()==cf)yf(u).forEach(function(t){return n.addClassName(t)})}return n}var bf=new _f(!0,new O,null,null),xf=new mf;function wf(t){return t instanceof pu&&0==t.value.trim().length}function jf(t){var e=new Map;return t.forEach(function(t){e.get(t.type.reference)||e.set(t.type.reference,t)}),Array.from(e.values())}function Of(t){return t instanceof No&&(t=t.ast),t instanceof go}var Cf=function(){function t(t,e){this.options=t,this.reflector=e}return t.prototype.compileComponent=function(t,e,n,r,i,o){var a=this,s=new Map;r.forEach(function(t){return s.set(t.name,t.type.reference)});var u=0,c=function(t,n){var r=u++;return new Sf(a.options,a.reflector,i,t,e.type.reference,e.isHost,r,s,n,o,c)},l=c(null,[]);return l.visitAll([],n),l.build(t)},t}(),Ef=new(function(){function t(){}return t.prototype.getLocal=function(t){return t===Xo.event.name?Ee("_any"):null},t}()),Sf=function(){function t(t,e,n,r,i,o,a,s,u,c,l){this.options=t,this.reflector=e,this.externalReferenceVars=n,this.parent=r,this.component=i,this.isHostComponent=o,this.embeddedViewIndex=a,this.pipes=s,this.guards=u,this.ctx=c,this.viewBuilderFactory=l,this.refOutputVars=new Map,this.variables=[],this.children=[],this.updates=[],this.actions=[]}return t.prototype.getOutputVar=function(t){var e;if(!(e=t===this.component&&this.isHostComponent?"_any":t instanceof vn?this.externalReferenceVars.get(t):"_any"))throw new Error("Illegal State: referring to a type without a variable "+JSON.stringify(t));return e},t.prototype.getTypeGuardExpressions=function(t){var e,n,r,i,a=Object(o.__spread)(this.guards);try{for(var s=Object(o.__values)(t.directives),u=s.next();!u.done;u=s.next()){var c=u.value;try{for(var l=Object(o.__values)(c.inputs),d=l.next();!d.done;d=l.next()){var p=d.value,f=c.directive.guards[p.directiveName];if(f){var h="UseIf"===f;a.push({guard:f,useIf:h,expression:{context:this.component,value:p.value}})}}}catch(t){r={error:t}}finally{try{d&&!d.done&&(i=l.return)&&i.call(l)}finally{if(r)throw r.error}}}}catch(t){e={error:t}}finally{try{u&&!u.done&&(n=s.return)&&n.call(s)}finally{if(e)throw e.error}}return a},t.prototype.visitAll=function(t,e){this.variables=t,Ed(this,e)},t.prototype.build=function(t,e){var n,r,i=this;void 0===e&&(e=[]),this.children.forEach(function(n){return n.build(t,e)});var a=[Ee("_any").set(te).toDeclStmt(mt)],s=0;if(this.updates.forEach(function(t){var e=i.preprocessUpdateExpression(t),n=e.sourceSpan,r=e.context,u=e.value,c=""+s++,l=ia(r===i.component?i:Ef,Ee(i.getOutputVar(r)),u,c,Jo.General),d=l.stmts,p=l.currValExpr;d.push(new oe(p)),a.push.apply(a,Object(o.__spread)(d.map(function(t){return je(t,n)})))}),this.actions.forEach(function(t){var e=t.sourceSpan,n=t.context,r=t.value,u=""+s++,c=Zo(n===i.component?i:Ef,Ee(i.getOutputVar(n)),r,u).stmts;a.push.apply(a,Object(o.__spread)(c.map(function(t){return je(t,e)})))}),this.guards.length){var u=void 0;try{for(var c=Object(o.__values)(this.guards),l=c.next();!l.done;l=c.next()){var d=l.value,p=this.preprocessUpdateExpression(d.expression),f=p.context,h=p.value,m=""+s++,g=ia(f===this.component?this:Ef,Ee(this.getOutputVar(f)),h,m,Jo.TrySimple),y=g.stmts,_=g.currValExpr;if(0==y.length){var v=d.useIf?_:this.ctx.importExpr(d.guard).callFn([_]);u=u?u.and(v):v}}}catch(t){n={error:t}}finally{try{l&&!l.done&&(r=c.return)&&r.call(c)}finally{if(n)throw n.error}}u&&(a=[new pe(u,a)])}var b="_View_"+t+"_"+this.embeddedViewIndex,x=new ie(b,[],a);return e.push(x),e},t.prototype.visitBoundText=function(t,e){var n=this;t.value.ast.expressions.forEach(function(e){return n.updates.push({context:n.component,value:e,sourceSpan:t.sourceSpan})})},t.prototype.visitEmbeddedTemplate=function(t,e){if(this.visitElementOrTemplate(t),this.options.fullTemplateTypeCheck){var n=this.getTypeGuardExpressions(t),r=this.viewBuilderFactory(this,n);this.children.push(r),r.visitAll(t.variables,t.children)}},t.prototype.visitElement=function(t,e){var n=this;this.visitElementOrTemplate(t),t.inputs.forEach(function(t){n.updates.push({context:n.component,value:t.value,sourceSpan:t.sourceSpan})}),Ed(this,t.children)},t.prototype.visitElementOrTemplate=function(t){var e=this;t.directives.forEach(function(t){e.visitDirective(t)}),t.references.forEach(function(t){var n=null;n=t.value&&t.value.identifier&&e.options.fullTemplateTypeCheck?t.value.identifier.reference:at.Dynamic,e.refOutputVars.set(t.name,n)}),t.outputs.forEach(function(t){e.actions.push({context:e.component,value:t.handler,sourceSpan:t.sourceSpan})})},t.prototype.visitDirective=function(t){var e=this,n=t.directive.type.reference;t.inputs.forEach(function(t){return e.updates.push({context:e.component,value:t.value,sourceSpan:t.sourceSpan})}),this.options.fullTemplateTypeCheck&&(t.hostProperties.forEach(function(t){return e.updates.push({context:n,value:t.value,sourceSpan:t.sourceSpan})}),t.hostEvents.forEach(function(t){return e.actions.push({context:n,value:t.handler,sourceSpan:t.sourceSpan})}))},t.prototype.getLocal=function(t){if(t==Xo.event.name)return Ee(this.getOutputVar(at.Dynamic));for(var e=this;e;e=e.parent){var n=void 0;if(null==(n=e.refOutputVars.get(t)))e.variables.find(function(e){return e.name===t})&&(n=at.Dynamic);if(null!=n)return Ee(this.getOutputVar(n))}return null},t.prototype.pipeOutputVar=function(t){var e=this.pipes.get(t);if(!e)throw new Error("Illegal State: Could not find pipe "+t+" in template of "+this.component);return this.getOutputVar(e)},t.prototype.preprocessUpdateExpression=function(t){var e=this;return{sourceSpan:t.sourceSpan,context:t.context,value:$o({createLiteralArrayConverter:function(t){return function(t){var n=ke(t);return e.options.fullTemplateTypeCheck?n:n.cast(mt)}},createLiteralMapConverter:function(t){return function(n){var r=Ae(t.map(function(t,e){return{key:t.key,value:n[e],quoted:t.quoted}}));return e.options.fullTemplateTypeCheck?r:r.cast(mt)}},createPipeConverter:function(t,n){return function(n){return(e.options.fullTemplateTypeCheck?Ee(e.pipeOutputVar(t)):Ee(e.getOutputVar(at.Dynamic))).callMethod("transform",n)}}},t.value)}},t.prototype.visitNgContent=function(t,e){},t.prototype.visitText=function(t,e){},t.prototype.visitDirectiveProperty=function(t,e){},t.prototype.visitReference=function(t,e){},t.prototype.visitVariable=function(t,e){},t.prototype.visitEvent=function(t,e){},t.prototype.visitElementProperty=function(t,e){},t.prototype.visitAttr=function(t,e){},t}(),Mf="class",Tf="style",kf=function(){return function(t,e){this.viewClassVar=t,this.rendererTypeVar=e}}(),Af=function(){function t(t){this._reflector=t}return t.prototype.compileComponent=function(t,e,n,r,i){var a,s=this,u=0,c=function t(e,n){void 0===n&&(n=new Map);e.forEach(function(e){var r=new Set,i=new Set,o=void 0;e instanceof _d?(t(e.children,n),e.children.forEach(function(t){var e=n.get(t);e.staticQueryIds.forEach(function(t){return r.add(t)}),e.dynamicQueryIds.forEach(function(t){return i.add(t)})}),o=e.queryMatches):e instanceof vd&&(t(e.children,n),e.children.forEach(function(t){var e=n.get(t);e.staticQueryIds.forEach(function(t){return i.add(t)}),e.dynamicQueryIds.forEach(function(t){return i.add(t)})}),o=e.queryMatches),o&&o.forEach(function(t){return r.add(t.queryId)}),i.forEach(function(t){return r.delete(t)}),n.set(e,{staticQueryIds:r,dynamicQueryIds:i})});return n}(n),l=void 0;if(!e.isHost){var d=e.template,p=[];d.animations&&d.animations.length&&p.push(new Zt("animation",cp(t,d.animations),!0));var f=Ee(Tn(e.type.reference));l=f.name,t.statements.push(f.set(Se(gn.createRendererType2).callFn([new $t([new Zt("encapsulation",Pe(d.encapsulation),!1),new Zt("styles",r,!1),new Zt("data",new $t(p),!1)])])).toDeclStmt(Me(gn.RendererType2),[It.Final,It.Exported]))}var h=function(n){var r=u++;return new Lf(s._reflector,t,n,e,r,i,c,h)},m=h(null);return m.visitAll([],n),(a=t.statements).push.apply(a,Object(o.__spread)(m.build())),new kf(m.viewName,l)},t}(),Df=Ee("_l"),If=Ee("_v"),Pf=Ee("_ck"),Rf=Ee("_co"),Nf=Ee("en"),Ff=Ee("ad"),Lf=function(){function t(t,e,n,r,i,o,a,s){this.reflector=t,this.outputCtx=e,this.parent=n,this.component=r,this.embeddedViewIndex=i,this.usedPipes=o,this.staticQueryIds=a,this.viewBuilderFactory=s,this.nodes=[],this.purePipeNodeIndices=Object.create(null),this.refNodeIndices=Object.create(null),this.variables=[],this.children=[],this.compType=this.embeddedViewIndex>0?mt:Te(e.importExpr(this.component.type.reference)),this.viewName=Mn(this.component.type.reference,this.embeddedViewIndex)}return t.prototype.visitAll=function(t,e){var n,r,i,o=this;if(this.variables=t,this.parent||this.usedPipes.forEach(function(t){t.pure&&(o.purePipeNodeIndices[t.name]=o._createPipe(null,t))}),!this.parent){var a=(n=this.staticQueryIds,r=new Set,i=new Set,Array.from(n.values()).forEach(function(t){t.staticQueryIds.forEach(function(t){return r.add(t)}),t.dynamicQueryIds.forEach(function(t){return i.add(t)})}),i.forEach(function(t){return r.delete(t)}),{staticQueryIds:r,dynamicQueryIds:i});this.component.viewQueries.forEach(function(t,e){var n=e+1,r=t.first?0:1,i=134217728|Hf(a,n,t.first);o.nodes.push(function(){return{sourceSpan:null,nodeFlags:i,nodeDef:Se(gn.queryDef).callFn([Pe(i),Pe(n),new $t([new Zt(t.propertyName,Pe(r),!1)])])}})})}Ed(this,e),this.parent&&(0===e.length||function t(e){var n=e[e.length-1];if(n instanceof vd)return n.hasViewContainer;if(n instanceof _d)return s(n.name)&&n.children.length?t(n.children):n.hasViewContainer;return n instanceof jd}(e))&&this.nodes.push(function(){return{sourceSpan:null,nodeFlags:1,nodeDef:Se(gn.anchorDef).callFn([Pe(0),te,te,Pe(0)])}})},t.prototype.build=function(t){void 0===t&&(t=[]),this.children.forEach(function(e){return e.build(t)});var e=this._createNodeExpressions(),n=e.updateRendererStmts,r=e.updateDirectivesStmts,i=e.nodeDefExprs,o=this._createUpdateFn(n),a=this._createUpdateFn(r),s=0;this.parent||this.component.changeDetection!==b.OnPush||(s|=2);var u=new ie(this.viewName,[new qt(Df.name)],[new ae(Se(gn.viewDef).callFn([Pe(s),ke(i),a,o]))],Me(gn.ViewDefinition),0===this.embeddedViewIndex?[It.Exported]:[]);return t.push(u),t},t.prototype._createUpdateFn=function(t){var e;if(t.length>0){var n=[];!this.component.isHost&&ve(t).has(Rf.name)&&n.push(Rf.set(If.prop("component")).toDeclStmt(this.compType)),e=De([new qt(Pf.name,gt),new qt(If.name,gt)],Object(o.__spread)(n,t),gt)}else e=te;return e},t.prototype.visitNgContent=function(t,e){this.nodes.push(function(){return{sourceSpan:t.sourceSpan,nodeFlags:8,nodeDef:Se(gn.ngContentDef).callFn([Pe(t.ngContentIndex),Pe(t.index)])}})},t.prototype.visitText=function(t,e){this.nodes.push(function(){return{sourceSpan:t.sourceSpan,nodeFlags:2,nodeDef:Se(gn.textDef).callFn([Pe(-1),Pe(t.ngContentIndex),ke([Pe(t.value)])])}})},t.prototype.visitBoundText=function(t,e){var n=this,r=this.nodes.length;this.nodes.push(null);var i=t.value.ast,o=i.expressions.map(function(e,i){return n._preprocessUpdateExpression({nodeIndex:r,bindingIndex:i,sourceSpan:t.sourceSpan,context:Rf,value:e})}),a=r;this.nodes[r]=function(){return{sourceSpan:t.sourceSpan,nodeFlags:2,nodeDef:Se(gn.textDef).callFn([Pe(a),Pe(t.ngContentIndex),ke(i.strings.map(function(t){return Pe(t)}))]),updateRenderer:o}}},t.prototype.visitEmbeddedTemplate=function(t,e){var n=this,r=this.nodes.length;this.nodes.push(null);var i=this._visitElementOrTemplate(r,t),o=i.flags,a=i.queryMatchesExpr,s=i.hostEvents,u=this.viewBuilderFactory(this);this.children.push(u),u.visitAll(t.variables,t.children);var c=this.nodes.length-r-1;this.nodes[r]=function(){return{sourceSpan:t.sourceSpan,nodeFlags:1|o,nodeDef:Se(gn.anchorDef).callFn([Pe(o),a,Pe(t.ngContentIndex),Pe(c),n._createElementHandleEventFn(r,s),Ee(u.viewName)])}}},t.prototype.visitElement=function(t,e){var n=this,r=this.nodes.length;this.nodes.push(null);var i=s(t.name)?null:t.name,a=this._visitElementOrTemplate(r,t),u=a.flags,c=a.usedEvents,l=a.queryMatchesExpr,d=a.hostBindings,p=a.hostEvents,f=[],h=[],m=[];if(i){var g=t.inputs.map(function(t){return{context:Rf,inputAst:t,dirAst:null}}).concat(d);g.length&&(h=g.map(function(t,e){return n._preprocessUpdateExpression({context:t.context,nodeIndex:r,bindingIndex:e,sourceSpan:t.inputAst.sourceSpan,value:t.inputAst.value})}),f=g.map(function(t){return function(t,e){var n=t.type;switch(n){case 1:return ke([Pe(1),Pe(t.name),Pe(t.securityContext)]);case 0:return ke([Pe(8),Pe(t.name),Pe(t.securityContext)]);case 4:var r=8|(e&&e.directive.isComponent?32:16);return ke([Pe(r),Pe("@"+t.name),Pe(t.securityContext)]);case 2:return ke([Pe(2),Pe(t.name),te]);case 3:return ke([Pe(4),Pe(t.name),Pe(t.unit)]);default:var i=n;throw new Error("unexpected "+i)}}(t.inputAst,t.dirAst)})),m=c.map(function(t){var e=Object(o.__read)(t,2),n=e[0],r=e[1];return ke([Pe(n),Pe(r)])})}Ed(this,t.children);var y=this.nodes.length-r-1,_=t.directives.find(function(t){return t.directive.isComponent}),v=te,b=te;_&&(b=this.outputCtx.importExpr(_.directive.componentViewType),v=this.outputCtx.importExpr(_.directive.rendererType));var x=r;this.nodes[r]=function(){return{sourceSpan:t.sourceSpan,nodeFlags:1|u,nodeDef:Se(gn.elementDef).callFn([Pe(x),Pe(u),l,Pe(t.ngContentIndex),Pe(y),Pe(i),i?(e=t,o=Object.create(null),e.attrs.forEach(function(t){o[t.name]=t.value}),e.directives.forEach(function(t){Object.keys(t.directive.hostAttributes).forEach(function(e){var n,r,i=t.directive.hostAttributes[e],a=o[e];o[e]=null!=a?(r=i,(n=e)==Mf||n==Tf?a+" "+r:r):i})}),ke(Object.keys(o).sort().map(function(t){return ke([Pe(t),Pe(o[t])])}))):te,f.length?ke(f):te,m.length?ke(m):te,n._createElementHandleEventFn(r,p),b,v]),updateRenderer:h};var e,o}},t.prototype._visitElementOrTemplate=function(t,e){var n=this,r=0;e.hasViewContainer&&(r|=16777216);var i=new Map;e.outputs.forEach(function(t){var e=zf(t,null),n=e.name,r=e.target;i.set(Uf(r,n),[r,n])}),e.directives.forEach(function(t){t.hostEvents.forEach(function(e){var n=zf(e,t),r=n.name,o=n.target;i.set(Uf(o,r),[o,r])})});var a=[],s=[];this._visitComponentFactoryResolverProvider(e.directives),e.providers.forEach(function(r,u){var c=void 0,l=void 0;if(e.directives.forEach(function(t,e){t.directive.type.reference===In(r.token)&&(c=t,l=e)}),c){var d=n._visitDirective(r,c,l,t,e.references,e.queryMatches,i,n.staticQueryIds.get(e)),p=d.hostBindings,f=d.hostEvents;a.push.apply(a,Object(o.__spread)(p)),s.push.apply(s,Object(o.__spread)(f))}else n._visitProvider(r,e.queryMatches)});var u=[];return e.queryMatches.forEach(function(t){var e=void 0;In(t.value)===n.reflector.resolveExternalReference(gn.ElementRef)?e=0:In(t.value)===n.reflector.resolveExternalReference(gn.ViewContainerRef)?e=3:In(t.value)===n.reflector.resolveExternalReference(gn.TemplateRef)&&(e=2),null!=e&&u.push(ke([Pe(t.queryId),Pe(e)]))}),e.references.forEach(function(e){var r=void 0;e.value?In(e.value)===n.reflector.resolveExternalReference(gn.TemplateRef)&&(r=2):r=1,null!=r&&(n.refNodeIndices[e.name]=t,u.push(ke([Pe(e.name),Pe(r)])))}),e.outputs.forEach(function(t){s.push({context:Rf,eventAst:t,dirAst:null})}),{flags:r,usedEvents:Array.from(i.values()),queryMatchesExpr:u.length?ke(u):te,hostBindings:a,hostEvents:s}},t.prototype._visitDirective=function(t,e,n,r,i,o,a,s){var u=this,c=this.nodes.length;this.nodes.push(null),e.directive.queries.forEach(function(t,n){var r=e.contentQueryStartId+n,i=67108864|Hf(s,r,t.first),o=t.first?0:1;u.nodes.push(function(){return{sourceSpan:e.sourceSpan,nodeFlags:i,nodeDef:Se(gn.queryDef).callFn([Pe(i),Pe(r),new $t([new Zt(t.propertyName,Pe(o),!1)])])}})});var l=this.nodes.length-c-1,d=this._visitProviderOrDirective(t,o),p=d.flags,f=d.queryMatchExprs,h=d.providerExpr,m=d.depsExpr;i.forEach(function(e){e.value&&In(e.value)===In(t.token)&&(u.refNodeIndices[e.name]=c,f.push(ke([Pe(e.name),Pe(4)])))}),e.directive.isComponent&&(p|=32768);var g=e.inputs.map(function(t,e){var n=ke([Pe(e),Pe(t.directiveName)]);return new Zt(t.directiveName,n,!1)}),y=[],_=e.directive;Object.keys(_.outputs).forEach(function(t){var e=_.outputs[t];a.has(e)&&y.push(new Zt(t,Pe(e),!1))});var v=[];(e.inputs.length||(327680&p)>0)&&(v=e.inputs.map(function(t,e){return u._preprocessUpdateExpression({nodeIndex:c,bindingIndex:e,sourceSpan:t.sourceSpan,context:Rf,value:t.value})}));var b=Se(gn.nodeValue).callFn([If,Pe(c)]),x=e.hostProperties.map(function(t){return{context:b,dirAst:e,inputAst:t}}),w=e.hostEvents.map(function(t){return{context:b,eventAst:t,dirAst:e}}),j=c;return this.nodes[c]=function(){return{sourceSpan:e.sourceSpan,nodeFlags:16384|p,nodeDef:Se(gn.directiveDef).callFn([Pe(j),Pe(p),f.length?ke(f):te,Pe(l),h,m,g.length?new $t(g):te,y.length?new $t(y):te]),updateDirectives:v,directive:e.directive.type}},{hostBindings:x,hostEvents:w}},t.prototype._visitProvider=function(t,e){this._addProviderNode(this._visitProviderOrDirective(t,e))},t.prototype._visitComponentFactoryResolverProvider=function(t){var e=t.find(function(t){return t.directive.isComponent});if(e&&e.directive.entryComponents.length){var n=Gp(this.reflector,this.outputCtx,8192,e.directive.entryComponents),r=n.providerExpr,i=n.depsExpr,o=n.flags,a=n.tokenExpr;this._addProviderNode({providerExpr:r,depsExpr:i,flags:o,tokenExpr:a,queryMatchExprs:[],sourceSpan:e.sourceSpan})}},t.prototype._addProviderNode=function(t){this.nodes.length;this.nodes.push(function(){return{sourceSpan:t.sourceSpan,nodeFlags:t.flags,nodeDef:Se(gn.providerDef).callFn([Pe(t.flags),t.queryMatchExprs.length?ke(t.queryMatchExprs):te,t.tokenExpr,t.providerExpr,t.depsExpr])}})},t.prototype._visitProviderOrDirective=function(t,e){var n=[];e.forEach(function(e){In(e.value)===In(t.token)&&n.push(ke([Pe(e.queryId),Pe(4)]))});var r=zp(this.outputCtx,t),i=r.providerExpr,o=r.depsExpr,a=r.flags,s=r.tokenExpr;return{flags:0|a,queryMatchExprs:n,providerExpr:i,depsExpr:o,tokenExpr:s,sourceSpan:t.sourceSpan}},t.prototype.getLocal=function(t){if(t==Xo.event.name)return Xo.event;for(var e=If,n=this;n;n=n.parent,e=e.prop("parent").cast(mt)){var r=n.refNodeIndices[t];if(null!=r)return Se(gn.nodeValue).callFn([e,Pe(r)]);var i=n.variables.find(function(e){return e.name===t});if(i){var o=i.value||"$implicit";return e.prop("context").prop(o)}}return null},t.prototype._createLiteralArrayConverter=function(t,e){if(0===e){var n=Se(gn.EMPTY_ARRAY);return function(){return n}}var r=this.nodes.length;return this.nodes.push(function(){return{sourceSpan:t,nodeFlags:32,nodeDef:Se(gn.pureArrayDef).callFn([Pe(r),Pe(e)])}}),function(t){return Bf(r,t)}},t.prototype._createLiteralMapConverter=function(t,e){if(0===e.length){var n=Se(gn.EMPTY_MAP);return function(){return n}}var r=Ae(e.map(function(t,e){return Object(o.__assign)({},t,{value:Pe(e)})})),i=this.nodes.length;return this.nodes.push(function(){return{sourceSpan:t,nodeFlags:64,nodeDef:Se(gn.pureObjectDef).callFn([Pe(i),r])}}),function(t){return Bf(i,t)}},t.prototype._createPipeConverter=function(t,e,n){var r=this.usedPipes.find(function(t){return t.name===e});if(r.pure){var i=this.nodes.length;this.nodes.push(function(){return{sourceSpan:t.sourceSpan,nodeFlags:128,nodeDef:Se(gn.purePipeDef).callFn([Pe(i),Pe(n)])}});for(var o=If,a=this;a.parent;)a=a.parent,o=o.prop("parent").cast(mt);var s=a.purePipeNodeIndices[e],u=Se(gn.nodeValue).callFn([o,Pe(s)]);return function(e){return Vf(t.nodeIndex,t.bindingIndex,Bf(i,[u].concat(e)))}}var c=this._createPipe(t.sourceSpan,r),l=Se(gn.nodeValue).callFn([If,Pe(c)]);return function(e){return Vf(t.nodeIndex,t.bindingIndex,l.callMethod("transform",e))}},t.prototype._createPipe=function(t,e){var n=this,r=this.nodes.length,i=0;e.type.lifecycleHooks.forEach(function(t){t===pa.OnDestroy&&(i|=qp(t))});var o=e.type.diDeps.map(function(t){return Wp(n.outputCtx,t)});return this.nodes.push(function(){return{sourceSpan:t,nodeFlags:16,nodeDef:Se(gn.pipeDef).callFn([Pe(i),n.outputCtx.importExpr(e.type.reference),ke(o)])}}),r},t.prototype._preprocessUpdateExpression=function(t){var e=this;return{nodeIndex:t.nodeIndex,bindingIndex:t.bindingIndex,sourceSpan:t.sourceSpan,context:t.context,value:$o({createLiteralArrayConverter:function(n){return e._createLiteralArrayConverter(t.sourceSpan,n)},createLiteralMapConverter:function(n){return e._createLiteralMapConverter(t.sourceSpan,n)},createPipeConverter:function(n,r){return e._createPipeConverter(t,n,r)}},t.value)}},t.prototype._createNodeExpressions=function(){var t=this,e=0,n=[],r=[],i=this.nodes.map(function(t,e){var i=t(),s=i.nodeDef,u=i.nodeFlags,c=i.updateDirectives,l=i.updateRenderer,d=i.sourceSpan;return l&&n.push.apply(n,Object(o.__spread)(a(e,d,l,!1))),c&&r.push.apply(r,Object(o.__spread)(a(e,d,c,(327680&u)>0))),Oe(3&u?new Jt([Df.callFn([]).callFn([]),s]):s,d)});return{updateRendererStmts:n,updateDirectivesStmts:r,nodeDefExprs:i};function a(n,r,i,a){var s=[],u=i.map(function(n){var r=n.sourceSpan,i=n.context,a=n.value,u=""+e++,c=ia(i===Rf?t:null,i,a,u,Jo.General),l=c.stmts,d=c.currValExpr;return s.push.apply(s,Object(o.__spread)(l.map(function(t){return je(t,r)}))),Oe(d,r)});return(i.length||a)&&s.push(je(Bf(n,u).toStmt(),r)),s}},t.prototype._createElementHandleEventFn=function(t,e){var n,r=this,i=[],a=0;if(e.forEach(function(t){var e=t.context,n=t.eventAst,o=t.dirAst,s=""+a++,u=Zo(e===Rf?r:null,e,n.handler,s),c=u.stmts,l=u.allowDefault,d=c;l&&d.push(Ff.set(l.and(Ff)).toStmt());var p=zf(n,o),f=Uf(p.target,p.name);i.push(je(new pe(Pe(f).identical(Nf),d),n.sourceSpan))}),i.length>0){var s=[Ff.set(Pe(!0)).toDeclStmt(yt)];!this.component.isHost&&ve(i).has(Rf.name)&&s.push(Rf.set(If.prop("component")).toDeclStmt(this.compType)),n=De([new qt(If.name,gt),new qt(Nf.name,gt),new qt(Xo.event.name,gt)],Object(o.__spread)(s,i,[new ae(Ff)]),gt)}else n=te;return n},t.prototype.visitDirective=function(t,e){},t.prototype.visitDirectiveProperty=function(t,e){},t.prototype.visitReference=function(t,e){},t.prototype.visitVariable=function(t,e){},t.prototype.visitEvent=function(t,e){},t.prototype.visitElementProperty=function(t,e){},t.prototype.visitAttr=function(t,e){},t}();function Bf(t,e){return e.length>10?Pf.callFn([If,Pe(t),Pe(1),ke(e)]):Pf.callFn(Object(o.__spread)([If,Pe(t),Pe(0)],e))}function Vf(t,e,n){return Se(gn.unwrapValue).callFn([If,Pe(t),Pe(e),n])}function zf(t,e){return t.isAnimation?{name:"@"+t.name+"."+t.phase,target:e&&e.directive.isComponent?"component":null}:t}function Hf(t,e,n){var r=0;return!n||!t.staticQueryIds.has(e)&&t.dynamicQueryIds.has(e)?r|=536870912:r|=268435456,r}function Uf(t,e){return t?t+":"+e:e}var Wf=function(){function t(t,e,n,r){void 0===r&&(r=null),this._htmlParser=t,this._implicitTags=e,this._implicitAttrs=n,this._locale=r,this._messages=[]}return t.prototype.updateFromTemplate=function(t,e,n){var r,i=this._htmlParser.parse(t,e,!0,n);if(i.errors.length)return i.errors;var a,s,u,c,l=(a=i.rootNodes,s=n,u=this._implicitTags,c=this._implicitAttrs,new Wd(u,c).extract(a,s));return l.errors.length?l.errors:((r=this._messages).push.apply(r,Object(o.__spread)(l.messages)),[])},t.prototype.getMessages=function(){return this._messages},t.prototype.write=function(t,e){var n={},r=new qf;this._messages.forEach(function(e){var r,i=t.digest(e);n.hasOwnProperty(i)?(r=n[i].sources).push.apply(r,Object(o.__spread)(e.sources)):n[i]=e});var i=Object.keys(n).map(function(i){var o=t.createNameMapper(n[i]),a=n[i],s=o?r.convert(a.nodes,o):a.nodes,u=new Zn(s,{},{},a.meaning,a.description,i);return u.sources=a.sources,e&&u.sources.forEach(function(t){return t.filePath=e(t.filePath)}),u});return t.write(i,this._locale)},t}(),qf=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(o.__extends)(e,t),e.prototype.convert=function(t,e){var n=this;return e?t.map(function(t){return t.visit(n,e)}):t},e.prototype.visitTagPlaceholder=function(t,e){var n=this,r=e.toPublicName(t.startName),i=t.closeName?e.toPublicName(t.closeName):t.closeName,o=t.children.map(function(t){return t.visit(n,e)});return new er(t.tag,t.attrs,r,i,o,t.isVoid,t.sourceSpan)},e.prototype.visitPlaceholder=function(t,e){return new nr(t.value,e.toPublicName(t.name),t.sourceSpan)},e.prototype.visitIcuPlaceholder=function(t,e){return new rr(t.value,e.toPublicName(t.name),t.sourceSpan)},e}(ir),Gf=function(){function t(t,e,n){this.srcFileUrl=t,this.genFileUrl=e,"string"==typeof n?(this.source=n,this.stmts=null):(this.source=null,this.stmts=n)}return t.prototype.isEquivalent=function(t){return this.genFileUrl===t.genFileUrl&&(this.source?this.source===t.source:null!=t.stmts&&xt(this.stmts,t.stmts))},t}();function Yf(t,e){if(void 0===e&&(e=""),!t.stmts)throw new Error("Illegal state: No stmts present on GeneratedFile "+t.genFileUrl);return(new $p).emitStatements(t.genFileUrl,t.stmts,e)}function Kf(t,e){var n,r,i,a,s=[];try{for(var u=Object(o.__values)(t.transitiveModule.providers),c=u.next();!c.done;c=u.next()){var l=c.value,d=l.provider,p=l.module;if(In(d.token)===e.ROUTES){var f=Xf(d.useValue);try{for(var h=Object(o.__values)(f),m=h.next();!m.done;m=h.next()){var g=m.value;s.push(Qf(g,e,p.reference))}}catch(t){i={error:t}}finally{try{m&&!m.done&&(a=h.return)&&a.call(h)}finally{if(i)throw i.error}}}}}catch(t){n={error:t}}finally{try{c&&!c.done&&(r=u.return)&&r.call(u)}finally{if(n)throw n.error}}return s}function Xf(t,e){var n,r;if(void 0===e&&(e=[]),"string"==typeof t)e.push(t);else if(Array.isArray(t))try{for(var i=Object(o.__values)(t),a=i.next();!a.done;a=i.next()){Xf(a.value,e)}}catch(t){n={error:t}}finally{try{a&&!a.done&&(r=i.return)&&r.call(i)}finally{if(n)throw n.error}}else t.loadChildren?Xf(t.loadChildren,e):t.children&&Xf(t.children,e);return e}function Qf(t,e,n){var r=Object(o.__read)(t.split("#"),2),i=r[0],a=r[1],s=e.resolveExternalReference({moduleName:i,name:a},n?n.filePath:void 0);return{route:t,module:n||s,referencedModule:s}}var Zf=/^(?!.*\.d\.ts$).*\.ts$/,$f=function(){return function(t,e){this.symbol=t,this.metadata=e}}(),Jf=function(){function t(t,e,n,r){this.host=t,this.staticSymbolCache=e,this.summaryResolver=n,this.errorRecorder=r,this.metadataCache=new Map,this.resolvedSymbols=new Map,this.resolvedFilePaths=new Set,this.importAs=new Map,this.symbolResourcePaths=new Map,this.symbolFromFile=new Map,this.knownFileNameToModuleNames=new Map}return t.prototype.resolveSymbol=function(t){if(t.members.length>0)return this._resolveSymbolMembers(t);var e=this._resolveSymbolFromSummary(t);if(e)return e;var n=this.resolvedSymbols.get(t);return n||(this._createSymbolsOf(t.filePath),this.resolvedSymbols.get(t))},t.prototype.getImportAs=function(t,e){if(void 0===e&&(e=!0),t.members.length){var n=this.getStaticSymbol(t.filePath,t.name);return(i=this.getImportAs(n,e))?this.getStaticSymbol(i.filePath,i.name,t.members):null}var r=t.filePath.replace(mp,".");if(r!==t.filePath){var i,o=t.name.replace(gp,"");n=this.getStaticSymbol(r,o,t.members);return(i=this.getImportAs(n,e))?this.getStaticSymbol(wp(i.filePath),jp(i.name),n.members):null}var a=e&&this.summaryResolver.getImportAs(t)||null;return a||(a=this.importAs.get(t)),a},t.prototype.getResourcePath=function(t){return this.symbolResourcePaths.get(t)||t.filePath},t.prototype.getTypeArity=function(t){if(e=t.filePath,hp.test(e))return null;for(var e,n=eh(this.resolveSymbol(t));n&&n.metadata instanceof vn;)n=eh(this.resolveSymbol(n.metadata));return n&&n.metadata&&n.metadata.arity||null},t.prototype.getKnownModuleName=function(t){return this.knownFileNameToModuleNames.get(t)||null},t.prototype.recordImportAs=function(t,e){t.assertNoMembers(),e.assertNoMembers(),this.importAs.set(t,e)},t.prototype.recordModuleNameForFileName=function(t,e){this.knownFileNameToModuleNames.set(t,e)},t.prototype.invalidateFile=function(t){var e,n;this.metadataCache.delete(t),this.resolvedFilePaths.delete(t);var r=this.symbolFromFile.get(t);if(r){this.symbolFromFile.delete(t);try{for(var i=Object(o.__values)(r),a=i.next();!a.done;a=i.next()){var s=a.value;this.resolvedSymbols.delete(s),this.importAs.delete(s),this.symbolResourcePaths.delete(s)}}catch(t){e={error:t}}finally{try{a&&!a.done&&(n=i.return)&&n.call(i)}finally{if(e)throw e.error}}}},t.prototype.ignoreErrorsFor=function(t){var e=this.errorRecorder;this.errorRecorder=function(){};try{return t()}finally{this.errorRecorder=e}},t.prototype._resolveSymbolMembers=function(t){var e=t.members,n=this.resolveSymbol(this.getStaticSymbol(t.filePath,t.name));if(!n)return null;var r=eh(n.metadata);if(r instanceof vn)return new $f(t,this.getStaticSymbol(r.filePath,r.name,e));if(!r||"class"!==r.__symbolic){for(var i=r,o=0;o<e.length&&i;o++)i=i[e[o]];return new $f(t,i)}return r.statics&&1===e.length?new $f(t,r.statics[e[0]]):null},t.prototype._resolveSymbolFromSummary=function(t){var e=this.summaryResolver.resolveSummary(t);return e?new $f(t,e.metadata):null},t.prototype.getStaticSymbol=function(t,e,n){return this.staticSymbolCache.get(t,e,n)},t.prototype.hasDecorators=function(t){var e=this.getModuleMetadata(t);return!!e.metadata&&Object.keys(e.metadata).some(function(t){var n=e.metadata[t];return n&&"class"===n.__symbolic&&n.decorators})},t.prototype.getSymbolsOf=function(t){var e=this.summaryResolver.getSymbolsOf(t);if(e)return e;this._createSymbolsOf(t);var n=[];return this.resolvedSymbols.forEach(function(e){e.symbol.filePath===t&&n.push(e.symbol)}),n},t.prototype._createSymbolsOf=function(t){var e,n,r=this;if(!this.resolvedFilePaths.has(t)){this.resolvedFilePaths.add(t);var i=[],a=this.getModuleMetadata(t);if(a.importAs&&this.knownFileNameToModuleNames.set(t,a.importAs),a.exports){var s=function(e){if(e.export)e.export.forEach(function(n){var o,a=o=th(o="string"==typeof n?n:n.as);"string"!=typeof n&&(a=th(n.name));var s=r.resolveModule(e.from,t);if(s){var u=r.getStaticSymbol(s,a),c=r.getStaticSymbol(t,o);i.push(r.createExport(c,u))}});else{var n=u.resolveModule(e.from,t);if(n)u.getSymbolsOf(n).forEach(function(e){var n=r.getStaticSymbol(t,e.name);i.push(r.createExport(n,e))})}},u=this;try{for(var c=Object(o.__values)(a.exports),l=c.next();!l.done;l=c.next()){s(l.value)}}catch(t){e={error:t}}finally{try{l&&!l.done&&(n=c.return)&&n.call(c)}finally{if(e)throw e.error}}}if(a.metadata){var d=new Set(Object.keys(a.metadata).map(th)),p=a.origins||{};Object.keys(a.metadata).forEach(function(e){var n=a.metadata[e],o=th(e),s=r.getStaticSymbol(t,o),u=p.hasOwnProperty(e)&&p[e];if(u){var c=r.resolveModule(u,t);c?r.symbolResourcePaths.set(s,c):r.reportError(new Error("Couldn't resolve original symbol for "+u+" from "+r.host.getOutputName(t)))}i.push(r.createResolvedSymbol(s,t,d,n))})}i.forEach(function(t){return r.resolvedSymbols.set(t.symbol,t)}),this.symbolFromFile.set(t,i.map(function(t){return t.symbol}))}},t.prototype.createResolvedSymbol=function(t,e,n,r){var i,a=this,s=Zf.test(t.filePath);if(this.summaryResolver.isLibraryFile(t.filePath)&&!s&&r&&"class"===r.__symbolic){var u={__symbolic:"class",arity:r.arity};return new $f(t,u)}var c=function(){return i||(i=a.host.getOutputName(e.replace(/((\.ts)|(\.d\.ts)|)$/,".ts").replace(/^.*node_modules[/\\]/,""))),i},l=this,d=Be(r,new(function(r){function i(){return null!==r&&r.apply(this,arguments)||this}return Object(o.__extends)(i,r),i.prototype.visitStringMap=function(i,a){var s=i.__symbolic;if("function"===s){var u=a.length;a.push.apply(a,Object(o.__spread)(i.parameters||[]));var d=r.prototype.visitStringMap.call(this,i,a);return a.length=u,d}if("reference"!==s)return"error"===s?Object(o.__assign)({},i,{fileName:c()}):r.prototype.visitStringMap.call(this,i,a);var p=i.module,f=i.name?th(i.name):i.name;if(!f)return null;var h=void 0;return p?(h=l.resolveModule(p,t.filePath))?{__symbolic:"resolved",symbol:l.getStaticSymbol(h,f),line:i.line,character:i.character,fileName:c()}:{__symbolic:"error",message:"Could not resolve "+p+" relative to "+l.host.getMetadataFor(t.filePath)+".",line:i.line,character:i.character,fileName:c()}:a.indexOf(f)>=0?{__symbolic:"reference",name:f}:n.has(f)?l.getStaticSymbol(e,f):void 0},i}(He)),[]),p=eh(d);return p instanceof vn?this.createExport(t,p):new $f(t,d)},t.prototype.createExport=function(t,e){return t.assertNoMembers(),e.assertNoMembers(),this.summaryResolver.isLibraryFile(t.filePath)&&this.summaryResolver.isLibraryFile(e.filePath)&&this.importAs.set(e,this.getImportAs(t)||t),new $f(t,e)},t.prototype.reportError=function(t,e,n){if(!this.errorRecorder)throw t;this.errorRecorder(t,e&&e.filePath||n)},t.prototype.getModuleMetadata=function(t){var e=this.metadataCache.get(t);if(!e){var n=this.host.getMetadataFor(t);if(n){var r=-1;n.forEach(function(t){t&&t.version>r&&(r=t.version,e=t)})}if(e||(e={__symbolic:"module",version:4,module:t,metadata:{}}),4!=e.version){var i=2==e.version?"Unsupported metadata version "+e.version+" for module "+t+". This module should be compiled with a newer version of ngc":"Metadata version mismatch for module "+this.host.getOutputName(t)+", found version "+e.version+", expected 4";this.reportError(new Error(i))}this.metadataCache.set(t,e)}return e},t.prototype.getSymbolByModule=function(t,e,n){var r=this.resolveModule(t,n);return r?this.getStaticSymbol(r,e):(this.reportError(new Error("Could not resolve module "+t+(n?" relative to "+this.host.getOutputName(n):""))),this.getStaticSymbol("ERROR:"+t,e))},t.prototype.resolveModule=function(t,e){try{return this.host.moduleNameToFileName(t,e)}catch(n){console.error("Could not resolve module '"+t+"' relative to file "+e),this.reportError(n,void 0,e)}return null},t}();function th(t){return t.startsWith("___")?t.substr(1):t}function eh(t){return t&&"resolved"===t.__symbolic?t.symbol:t}function nh(t,e,n){var r=jp(e.name);t.statements.push(De([],[new ae(n)],new ft(mt)).toDeclStmt(r,[It.Final,It.Exported]))}var rh=function(t){function e(e,n,r){var i=t.call(this)||this;return i.symbolResolver=e,i.summaryResolver=n,i.srcFileName=r,i.symbols=[],i.indexBySymbol=new Map,i.reexportedBy=new Map,i.processedSummaryBySymbol=new Map,i.processedSummaries=[],i.unprocessedSymbolSummariesBySymbol=new Map,i.moduleName=e.getKnownModuleName(r),i}return Object(o.__extends)(e,t),e.prototype.addSummary=function(t){var e,n,r=this,i=this.unprocessedSymbolSummariesBySymbol.get(t.symbol),o=this.processedSummaryBySymbol.get(t.symbol);if(i||(i={symbol:t.symbol,metadata:void 0},this.unprocessedSymbolSummariesBySymbol.set(t.symbol,i),o={symbol:this.processValue(t.symbol,0)},this.processedSummaries.push(o),this.processedSummaryBySymbol.set(t.symbol,o)),!i.metadata&&t.metadata){var a=t.metadata||{};if("class"===a.__symbolic){var s={};Object.keys(a).forEach(function(t){"decorators"!==t&&(s[t]=a[t])}),a=s}else ah(a)&&(ah(n=a)&&eh(n.expression)instanceof vn||ah(e=a)&&e.expression&&"select"===e.expression.__symbolic&&eh(e.expression.expression)instanceof vn||(a={__symbolic:"error",message:"Complex function calls are not supported."}));if(i.metadata=a,o.metadata=this.processValue(a,1),a instanceof vn&&this.summaryResolver.isLibraryFile(a.filePath)){var u=this.symbols[this.indexBySymbol.get(a)];Cp(u.name)||this.reexportedBy.set(u,t.symbol)}}if(!i.type&&t.type&&(i.type=t.type,o.type=this.processValue(t.type,0),t.type.summaryKind===jn.NgModule)){var c=t.type;c.exportedDirectives.concat(c.exportedPipes).forEach(function(t){var e=t.reference;if(r.summaryResolver.isLibraryFile(e.filePath)&&!r.unprocessedSymbolSummariesBySymbol.has(e)){var n=r.summaryResolver.resolveSummary(e);n&&r.addSummary(n)}})}},e.prototype.serialize=function(){var t=this,e=[];return{json:JSON.stringify({moduleName:this.moduleName,summaries:this.processedSummaries,symbols:this.symbols.map(function(n,r){n.assertNoMembers();var i=void 0;if(t.summaryResolver.isLibraryFile(n.filePath)){var o=t.reexportedBy.get(n);if(o)i=t.indexBySymbol.get(o);else{var a=t.unprocessedSymbolSummariesBySymbol.get(n);a&&a.metadata&&"interface"===a.metadata.__symbolic||(i=n.name+"_"+r,e.push({symbol:n,exportAs:i}))}}return{__symbol:r,name:n.name,filePath:t.summaryResolver.toSummaryFileName(n.filePath,t.srcFileName),importAs:i}})}),exportAs:e}},e.prototype.processValue=function(t,e){return Be(t,this,e)},e.prototype.visitOther=function(t,e){if(t instanceof vn){var n=this.symbolResolver.getStaticSymbol(t.filePath,t.name);return{__symbol:this.visitStaticSymbol(n,e),members:t.members}}},e.prototype.visitStringMap=function(e,n){return"resolved"===e.__symbolic?Be(e.symbol,this,n):("error"===e.__symbolic&&(delete e.line,delete e.character),t.prototype.visitStringMap.call(this,e,n))},e.prototype.visitStaticSymbol=function(t,e){var n=this.indexBySymbol.get(t),r=null;if(1&e&&this.summaryResolver.isLibraryFile(t.filePath)){if(this.unprocessedSymbolSummariesBySymbol.has(t))return n;(r=this.loadSummary(t))&&r.metadata instanceof vn&&(n=this.visitStaticSymbol(r.metadata,e),r=null)}else if(null!=n)return n;return null==n&&(n=this.symbols.length,this.symbols.push(t)),this.indexBySymbol.set(t,n),r&&this.addSummary(r),n},e.prototype.loadSummary=function(t){var e=this.summaryResolver.resolveSummary(t);if(!e){var n=this.symbolResolver.resolveSymbol(t);n&&(e={symbol:n.symbol,metadata:n.metadata})}return e},e}(He),ih=function(){function t(t,e,n){this.outputCtx=t,this.symbolResolver=e,this.summaryResolver=n,this.data=[]}return t.prototype.addSourceType=function(t,e){this.data.push({summary:t,metadata:e,isLibrary:!1})},t.prototype.addLibType=function(t){this.data.push({summary:t,metadata:null,isLibrary:!0})},t.prototype.serialize=function(t){var e,n,r,i,a,s,u=this,c=new Map;try{for(var l=Object(o.__values)(t),d=l.next();!d.done;d=l.next()){var p=d.value,f=p.symbol,h=p.exportAs;c.set(f,h)}}catch(t){e={error:t}}finally{try{d&&!d.done&&(n=l.return)&&n.call(l)}finally{if(e)throw e.error}}var m=new Set;try{for(var g=Object(o.__values)(this.data),y=g.next();!y.done;y=g.next()){var _=y.value,v=_.summary,b=_.metadata,x=_.isLibrary;if(v.summaryKind===jn.NgModule){m.add(v.type.reference);var w=v;try{for(var j=Object(o.__values)(w.modules),O=j.next();!O.done;O=j.next()){var C=O.value;m.add(C.reference)}}catch(t){a={error:t}}finally{try{O&&!O.done&&(s=j.return)&&s.call(j)}finally{if(a)throw a.error}}}if(!x){jp(v.type.reference.name);nh(this.outputCtx,v.type.reference,this.serializeSummaryWithDeps(v,b))}}}catch(t){r={error:t}}finally{try{y&&!y.done&&(i=g.return)&&i.call(g)}finally{if(r)throw r.error}}m.forEach(function(t){if(u.summaryResolver.isLibraryFile(t.filePath)){var e=jp(c.get(t)||t.name);u.outputCtx.statements.push(Ee(e).set(u.serializeSummaryRef(t)).toDeclStmt(null,[It.Exported]))}})},t.prototype.serializeSummaryWithDeps=function(t,e){var n=this,r=[this.serializeSummary(t)],i=[];if(e instanceof Bn)r.push.apply(r,Object(o.__spread)(e.declaredDirectives.concat(e.declaredPipes).map(function(t){return t.reference}).concat(e.transitiveModule.modules.map(function(t){return t.reference}).filter(function(t){return t!==e.type.reference})).map(function(t){return n.serializeSummaryRef(t)}))),i=e.providers;else if(t.summaryKind===jn.Directive){var a=t;i=a.providers.concat(a.viewProviders)}return r.push.apply(r,Object(o.__spread)(i.filter(function(t){return!!t.useClass}).map(function(t){return n.serializeSummary({summaryKind:jn.Injectable,type:t.useClass})}))),ke(r)},t.prototype.serializeSummaryRef=function(t){var e=this.symbolResolver.getStaticSymbol(wp(t.filePath),jp(t.name));return this.outputCtx.importExpr(e)},t.prototype.serializeSummary=function(t){var e=this.outputCtx;return Be(t,new(function(){function t(){}return t.prototype.visitArray=function(t,e){var n=this;return ke(t.map(function(t){return Be(t,n,e)}))},t.prototype.visitStringMap=function(t,e){var n=this;return new $t(Object.keys(t).map(function(r){return new Zt(r,Be(t[r],n,e),!1)}))},t.prototype.visitPrimitive=function(t,e){return Pe(t)},t.prototype.visitOther=function(t,n){if(t instanceof vn)return e.importExpr(t);throw new Error("Illegal State: Encountered value "+t)},t}()),null)},t}(),oh=function(t){function e(e,n){var r=t.call(this)||this;return r.symbolCache=e,r.summaryResolver=n,r}return Object(o.__extends)(e,t),e.prototype.deserialize=function(t,e){var n=this,r=JSON.parse(e),i=[];this.symbols=r.symbols.map(function(e){return n.symbolCache.get(n.summaryResolver.fromSummaryFileName(e.filePath,t),e.name)}),r.symbols.forEach(function(e,r){var o=n.symbols[r],a=e.importAs;"number"==typeof a?i.push({symbol:o,importAs:n.symbols[a]}):"string"==typeof a&&i.push({symbol:o,importAs:n.symbolCache.get(yp(t),a)})});var o=Be(r.summaries,this,null);return{moduleName:r.moduleName,summaries:o,importAs:i}},e.prototype.visitStringMap=function(e,n){if("__symbol"in e){var r=this.symbols[e.__symbol],i=e.members;return i.length?this.symbolCache.get(r.filePath,r.name,i):r}return t.prototype.visitStringMap.call(this,e,n)},e}(He);function ah(t){return t&&"call"===t.__symbolic}var sh=function(){function t(t,e,n,r,i,o,a,s,u,c,l,d,p,f){this._config=t,this._options=e,this._host=n,this.reflector=r,this._metadataResolver=i,this._templateParser=o,this._styleCompiler=a,this._viewCompiler=s,this._typeCheckCompiler=u,this._ngModuleCompiler=c,this._injectableCompiler=l,this._outputEmitter=d,this._summaryResolver=p,this._symbolResolver=f,this._templateAstCache=new Map,this._analyzedFiles=new Map,this._analyzedFilesForInjectables=new Map}return t.prototype.clearCache=function(){this._metadataResolver.clearCache()},t.prototype.analyzeModulesSync=function(t){var e=this,n=ph(t,this._host,this._symbolResolver,this._metadataResolver);return n.ngModules.forEach(function(t){return e._metadataResolver.loadNgModuleDirectiveAndPipeMetadata(t.type.reference,!0)}),n},t.prototype.analyzeModulesAsync=function(t){var e=this,n=ph(t,this._host,this._symbolResolver,this._metadataResolver);return Promise.all(n.ngModules.map(function(t){return e._metadataResolver.loadNgModuleDirectiveAndPipeMetadata(t.type.reference,!1)})).then(function(){return n})},t.prototype._analyzeFile=function(t){var e=this._analyzedFiles.get(t);return e||(e=hh(this._host,this._symbolResolver,this._metadataResolver,t),this._analyzedFiles.set(t,e)),e},t.prototype._analyzeFileForInjectables=function(t){var e=this._analyzedFilesForInjectables.get(t);return e||(e=mh(this._host,this._symbolResolver,this._metadataResolver,t),this._analyzedFilesForInjectables.set(t,e)),e},t.prototype.findGeneratedFileNames=function(t){var e=this,n=[],r=this._analyzeFile(t);(this._options.allowEmptyCodegenFiles||r.directives.length||r.pipes.length||r.injectables.length||r.ngModules.length||r.exportsNonSourceFiles)&&(n.push(yp(r.fileName,!0)),this._options.enableSummariesForJit&&n.push(wp(r.fileName,!0)));var i=bp(vp(r.fileName,!0)[1]);return r.directives.forEach(function(t){var o=e._metadataResolver.getNonNormalizedDirectiveMetadata(t).metadata;o.isComponent&&o.template.styleUrls.forEach(function(t){var a=e._host.resourceNameToFileName(t,r.fileName);if(!a)throw Ye("Couldn't resolve resource "+t+" relative to "+r.fileName);var s=(o.template.encapsulation||e._config.defaultEncapsulation)===_.Emulated;n.push(lh(a,s,i)),e._options.allowEmptyCodegenFiles&&n.push(lh(a,!s,i))})}),n},t.prototype.emitBasicStub=function(t,e){var n=this._createOutputContext(t);if(t.endsWith(".ngfactory.ts")){if(!e)throw new Error("Assertion error: require the original file for .ngfactory.ts stubs. File: "+t);var r=this._analyzeFile(e);this._createNgFactoryStub(n,r,1)}else if(t.endsWith(".ngsummary.ts")){if(this._options.enableSummariesForJit){if(!e)throw new Error("Assertion error: require the original file for .ngsummary.ts stubs. File: "+t);r=this._analyzeFile(e);uh(n),r.ngModules.forEach(function(t){var e,r;e=n,r=t.type.reference,nh(e,r,te)})}}else t.endsWith(".ngstyle.ts")&&uh(n);return this._codegenSourceModule("unknown",n)},t.prototype.emitTypeCheckStub=function(t,e){var n=this._analyzeFile(e),r=this._createOutputContext(t);return t.endsWith(".ngfactory.ts")&&this._createNgFactoryStub(r,n,2),r.statements.length>0?this._codegenSourceModule(n.fileName,r):null},t.prototype.loadFilesAsync=function(t,e){var n=this,r=t.map(function(t){return n._analyzeFile(t)}),i=[];r.forEach(function(t){return t.ngModules.forEach(function(t){return i.push(n._metadataResolver.loadNgModuleDirectiveAndPipeMetadata(t.type.reference,!1))})});var o=e.map(function(t){return n._analyzeFileForInjectables(t)});return Promise.all(i).then(function(t){return{analyzedModules:yh(r),analyzedInjectables:o}})},t.prototype.loadFilesSync=function(t,e){var n=this,r=t.map(function(t){return n._analyzeFile(t)});r.forEach(function(t){return t.ngModules.forEach(function(t){return n._metadataResolver.loadNgModuleDirectiveAndPipeMetadata(t.type.reference,!0)})});var i=e.map(function(t){return n._analyzeFileForInjectables(t)});return{analyzedModules:yh(r),analyzedInjectables:i}},t.prototype._createNgFactoryStub=function(t,e,n){var r=this,i=0;e.ngModules.forEach(function(e,a){r._ngModuleCompiler.createStub(t,e.type.reference);var s=Object(o.__spread)(e.transitiveModule.directives.map(function(t){return t.reference}),e.transitiveModule.pipes.map(function(t){return t.reference}),e.importedModules.map(function(t){return t.type.reference}),e.exportedModules.map(function(t){return t.type.reference}),r._externalIdentifierReferences([gn.TemplateRef,gn.ElementRef])),u=new Map;s.forEach(function(t,e){u.set(t,"_decl"+a+"_"+e)}),u.forEach(function(e,n){t.statements.push(Ee(e).set(te.cast(mt)).toDeclStmt(Te(t.importExpr(n,null,!1))))}),2&n&&e.declaredDirectives.forEach(function(n){var o=r._metadataResolver.getDirectiveMetadata(n.reference);o.isComponent&&(i++,r._createTypeCheckBlock(t,o.type.reference.name+"_Host_"+i,e,r._metadataResolver.getHostComponentMetadata(o),[o.type],u),r._createTypeCheckBlock(t,o.type.reference.name+"_"+i,e,o,e.transitiveModule.directives,u))})}),0===t.statements.length&&uh(t)},t.prototype._externalIdentifierReferences=function(t){var e,n,r=[];try{for(var i=Object(o.__values)(t),a=i.next();!a.done;a=i.next()){var s=a.value,u=_n(this.reflector,s);u.identifier&&r.push(u.identifier.reference)}}catch(t){e={error:t}}finally{try{a&&!a.done&&(n=i.return)&&n.call(i)}finally{if(e)throw e.error}}return r},t.prototype._createTypeCheckBlock=function(t,e,n,r,i,a){var s,u=this._parseTemplate(r,n,i),c=u.template,l=u.pipes;(s=t.statements).push.apply(s,Object(o.__spread)(this._typeCheckCompiler.compileComponent(e,r,c,l,a,t)))},t.prototype.emitMessageBundle=function(t,e){var n=this,r=[],i=new Hu,a=new Wf(i,[],{},e);if(t.files.forEach(function(t){var e=[];t.directives.forEach(function(t){var r=n._metadataResolver.getDirectiveMetadata(t);r&&r.isComponent&&e.push(r)}),e.forEach(function(t){var e=t.template.template,n=t.template.templateUrl,i=zi.fromArray(t.template.interpolation);r.push.apply(r,Object(o.__spread)(a.updateFromTemplate(e,n,i)))})}),r.length)throw new Error(r.map(function(t){return t.toString()}).join("\n"));return a},t.prototype.emitAllPartialModules=function(t,e){var n=this,r=t.ngModuleByPipeOrDirective,i=t.files,a=new Map,s=function(t){return a.has(t)||a.set(t,n._createOutputContext(t)),a.get(t)};return i.forEach(function(t){return n._compilePartialModule(t.fileName,r,t.directives,t.pipes,t.ngModules,t.injectables,s(t.fileName))}),e.forEach(function(t){return n._compileShallowModules(t.fileName,t.shallowModules,s(t.fileName))}),Array.from(a.values()).map(function(t){return{fileName:t.genFilePath,statements:Object(o.__spread)(t.constantPool.statements,t.statements)}})},t.prototype._compileShallowModules=function(t,e,n){var r=this;e.forEach(function(t){return e=n,i=t,a=r._injectableCompiler,s=En(i.type),u=i.rawImports?[i.rawImports]:[],c=i.rawExports?[i.rawExports]:[],l=Hr({factory:a.factoryFor({type:i.type,symbol:i.type.reference},e),providers:Ai(i.rawProviders,e),imports:Ai(Object(o.__spread)(u,c),e)}),d=Se(Qn.defineInjector).callFn([l]),void e.statements.push(new de(s,null,[new ue("ngInjectorDef",gt,[It.Static],d)],[],new ce(null,[],[]),[]));var e,i,a,s,u,c,l,d})},t.prototype._compilePartialModule=function(t,e,n,r,i,o,a){var s=this,u=new sc,c=new uc(this._templateParser.expressionParser,Hi,u,[],[]);n.forEach(function(t){var n=s._metadataResolver.getDirectiveMetadata(t);if(n.isComponent){var r=e.get(t);r||Ge("Cannot determine the module for component '"+En(n.type)+"'");var i=n.template.htmlAst;n.template.preserveWhitespaces||(i=Zu(i));var o=Uc(i.rootNodes,c),u=new Map;r.transitiveModule.directives.map(function(t){return s._metadataResolver.getDirectiveSummary(t.reference)}).forEach(function(t){t.selector&&u.set(t.selector,t.type.reference)});var l=new Map;r.transitiveModule.pipes.map(function(t){return s._metadataResolver.getPipeSummary(t.reference)}).forEach(function(t){l.set(t.name,t.type.reference)}),Rl(a,n,o,s.reflector,c,0,l)}else!function(t,e,n,r){var i=En(e.type);i||Ge("Cannot resolver the name of "+e.type);var o=t.constantPool.propertyNameOf(1),a=Dl(Nl(e,t,n),t.constantPool,r);t.statements.push(new de(i,null,[new ue(o,gt,[It.Static],a.expression)],[],new ce(null,[],[]),[]))}(a,n,s.reflector,c)}),r.forEach(function(t){var e=s._metadataResolver.getPipeMetadata(t);e&&function(t,e,n){var r=En(e.type);if(!r)return Ge("Cannot resolve the name of "+e.type);var i=lo({name:r,pipeName:e.name,type:t.importExpr(e.type.reference),deps:Mi(e.type,t,n),pure:e.pure}),o=t.constantPool.propertyNameOf(3);t.statements.push(new de(r,null,[new ue(o,gt,[It.Static],i.expression)],[],new ce(null,[],[]),[]))}(a,e,s.reflector)}),o.forEach(function(t){return s._injectableCompiler.compile(t,a)})},t.prototype.emitAllPartialModules2=function(t){var e=this;return t.reduce(function(t,n){return t.push.apply(t,Object(o.__spread)(e._emitPartialModule2(n.fileName,n.injectables))),t},[])},t.prototype._emitPartialModule2=function(t,e){var n=this,r=this._createOutputContext(t);return e.forEach(function(t){return n._injectableCompiler.compile(t,r)}),r.statements&&r.statements.length>0?[{fileName:t,statements:Object(o.__spread)(r.constantPool.statements,r.statements)}]:[]},t.prototype.emitAllImpls=function(t){var e=this,n=t.ngModuleByPipeOrDirective;return Un(t.files.map(function(t){return e._compileImplFile(t.fileName,n,t.directives,t.pipes,t.ngModules,t.injectables)}))},t.prototype._compileImplFile=function(t,e,n,r,i,a){var s=this,u=bp(vp(t,!0)[1]),c=[],l=this._createOutputContext(yp(t,!0));if(c.push.apply(c,Object(o.__spread)(this._createSummary(t,n,r,i,a,l))),i.forEach(function(t){return s._compileModule(l,t)}),n.forEach(function(n){var r=s._metadataResolver.getDirectiveMetadata(n);if(r.isComponent){var i=e.get(n);if(!i)throw new Error("Internal Error: cannot determine the module for component "+En(r.type)+"!");var o=s._styleCompiler.compileComponent(l,r);r.template.externalStylesheets.forEach(function(e){var n=s._styleCompiler.needsStyleShim(r);c.push(s._codegenStyles(t,r,e,n,u)),s._options.allowEmptyCodegenFiles&&c.push(s._codegenStyles(t,r,e,!n,u))});s._compileComponent(l,r,i,i.transitiveModule.directives,o,u);s._compileComponentFactory(l,r,i,u)}}),l.statements.length>0||this._options.allowEmptyCodegenFiles){var d=this._codegenSourceModule(t,l);c.unshift(d)}return c},t.prototype._createSummary=function(t,e,n,r,i,a){var s=this,u=this._symbolResolver.getSymbolsOf(t).map(function(t){return s._symbolResolver.resolveSymbol(t)}),c=Object(o.__spread)(r.map(function(t){return{summary:s._metadataResolver.getNgModuleSummary(t.type.reference),metadata:s._metadataResolver.getNgModuleMetadata(t.type.reference)}}),e.map(function(t){return{summary:s._metadataResolver.getDirectiveSummary(t),metadata:s._metadataResolver.getDirectiveMetadata(t)}}),n.map(function(t){return{summary:s._metadataResolver.getPipeSummary(t),metadata:s._metadataResolver.getPipeMetadata(t)}}),i.map(function(t){return{summary:s._metadataResolver.getInjectableSummary(t.symbol),metadata:s._metadataResolver.getInjectableSummary(t.symbol).type}})),l=this._options.enableSummariesForJit?this._createOutputContext(wp(t,!0)):null,d=function(t,e,n,r,i,o){var a=new rh(r,n,t);i.forEach(function(t){return a.addSummary({symbol:t.symbol,metadata:t.metadata})}),o.forEach(function(t){var e=t.summary;t.metadata,a.addSummary({symbol:e.type.reference,metadata:void 0,type:e})});var s=a.serialize(),u=s.json,c=s.exportAs;if(e){var l=new ih(e,r,n);o.forEach(function(t){var e=t.summary,n=t.metadata;l.addSourceType(e,n)}),a.unprocessedSymbolSummariesBySymbol.forEach(function(t){n.isLibraryFile(t.symbol.filePath)&&t.type&&l.addLibType(t.type)}),l.serialize(c)}return{json:u,exportAs:c}}(t,l,this._summaryResolver,this._symbolResolver,u,c),p=d.json;d.exportAs.forEach(function(t){a.statements.push(Ee(t.exportAs).set(a.importExpr(t.symbol)).toDeclStmt(null,[It.Exported]))});var f=[new Gf(t,xp(t),p)];return l&&f.push(this._codegenSourceModule(t,l)),f},t.prototype._compileModule=function(t,e){var n=[];if(this._options.locale){var r=this._options.locale.replace(/_/g,"-");n.push({token:_n(this.reflector,gn.LOCALE_ID),useValue:r})}this._options.i18nFormat&&n.push({token:_n(this.reflector,gn.TRANSLATIONS_FORMAT),useValue:this._options.i18nFormat}),this._ngModuleCompiler.compile(t,e,n)},t.prototype._compileComponentFactory=function(t,e,n,r){var i=this._metadataResolver.getHostComponentMetadata(e),o=this._compileComponent(t,i,n,[e.type],null,r).viewClassVar,a=An(e.type.reference),s=[];for(var u in e.inputs){var c=e.inputs[u];s.push(new Zt(u,Pe(c),!1))}var l=[];for(var u in e.outputs){c=e.outputs[u];l.push(new Zt(u,Pe(c),!1))}t.statements.push(Ee(a).set(Se(gn.createComponentFactory).callFn([Pe(e.selector),t.importExpr(e.type.reference),Ee(o),new $t(s),new $t(l),ke(e.template.ngContentSelectors.map(function(t){return Pe(t)}))])).toDeclStmt(Me(gn.ComponentFactory,[Te(t.importExpr(e.type.reference))],[rt.Const]),[It.Final,It.Exported]))},t.prototype._compileComponent=function(t,e,n,r,i,o){var a=this._parseTemplate(e,n,r),s=a.template,u=a.pipes,c=i?Ee(i.stylesVar):ke([]),l=this._viewCompiler.compileComponent(t,e,s,c,u);return i&&ch(this._symbolResolver,i,this._styleCompiler.needsStyleShim(e),o),l},t.prototype._parseTemplate=function(t,e,n){var r=this;if(this._templateAstCache.has(t.type.reference))return this._templateAstCache.get(t.type.reference);var i=t.template.preserveWhitespaces,o=n.map(function(t){return r._metadataResolver.getDirectiveSummary(t.reference)}),a=e.transitiveModule.pipes.map(function(t){return r._metadataResolver.getPipeSummary(t.reference)}),s=this._templateParser.parse(t,t.template.htmlAst,o,a,e.schemas,qn(e.type,t,t.template),i);return this._templateAstCache.set(t.type.reference,s),s},t.prototype._createOutputContext=function(t){var e=this;return{statements:[],genFilePath:t,importExpr:function(n,r,i){if(void 0===r&&(r=null),void 0===i&&(i=!0),!(n instanceof vn))throw new Error("Internal error: unknown identifier "+JSON.stringify(n));var o=e._symbolResolver.getTypeArity(n)||0,a=e._symbolResolver.getImportAs(n,i)||n,s=a.filePath,u=a.name,c=a.members,l=e._fileNameToModuleName(s,t),d=l===e._fileNameToModuleName(t,t)?null:l,p=r||[],f=o-p.length,h=p.concat(new Array(f).fill(mt));return c.reduce(function(t,e){return t.prop(e)},Se(new Vt(d,u,null),h))},constantPool:new dn}},t.prototype._fileNameToModuleName=function(t,e){return this._summaryResolver.getKnownModuleName(t)||this._symbolResolver.getKnownModuleName(t)||this._host.fileNameToModuleName(t,e)},t.prototype._codegenStyles=function(t,e,n,r,i){var o=this._createOutputContext(lh(n.moduleUrl,r,i)),a=this._styleCompiler.compileStyles(o,e,n,r);return ch(this._symbolResolver,a,r,i),this._codegenSourceModule(t,o)},t.prototype._codegenSourceModule=function(t,e){return new Gf(t,e.genFilePath,e.statements)},t.prototype.listLazyRoutes=function(t,e){var n,r,i,a,s=this;if(t)return function t(e,n,r){void 0===n&&(n=new Set);void 0===r&&(r=[]);var i,a;if(n.has(e)||!e.name)return r;n.add(e);var u=Kf(s._metadataResolver.getNgModuleMetadata(e,!0),s.reflector);try{for(var c=Object(o.__values)(u),l=c.next();!l.done;l=c.next()){var d=l.value;r.push(d),t(d.referencedModule,n,r)}}catch(t){i={error:t}}finally{try{l&&!l.done&&(a=c.return)&&a.call(c)}finally{if(i)throw i.error}}return r}(Qf(t,this.reflector).referencedModule);if(e){var u=[];try{for(var c=Object(o.__values)(e.ngModules),l=c.next();!l.done;l=c.next()){var d=Kf(l.value,this.reflector);try{for(var p=Object(o.__values)(d),f=p.next();!f.done;f=p.next()){var h=f.value;u.push(h)}}catch(t){i={error:t}}finally{try{f&&!f.done&&(a=p.return)&&a.call(p)}finally{if(i)throw i.error}}}}catch(t){n={error:t}}finally{try{l&&!l.done&&(r=c.return)&&r.call(c)}finally{if(n)throw n.error}}return u}throw new Error("Either route or analyzedModules has to be specified!")},t}();function uh(t){t.statements.push(Se(gn.ComponentFactory).toStmt())}function ch(t,e,n,r){e.dependencies.forEach(function(e){e.setValue(t.getStaticSymbol(lh(e.moduleUrl,n,r),e.name))})}function lh(t,e,n){return t+(e?".shim":"")+".ngstyle"+n}function dh(t,e,n,r){var i,o,a,s,u,c,l;return gh((i=t,o=e,a=n,s=r,u=new Set,c=[],l=function(t){if(u.has(t)||!o.isSourceFile(t))return!1;u.add(t);var e=hh(o,a,s,t);c.push(e),e.ngModules.forEach(function(t){t.transitiveModule.modules.forEach(function(t){return l(t.reference.filePath)})})},i.forEach(function(t){return l(t)}),c))}function ph(t,e,n,r){return fh(dh(t,e,n,r))}function fh(t){if(t.symbolsMissingModule&&t.symbolsMissingModule.length)throw Ye(t.symbolsMissingModule.map(function(t){return"Cannot determine the module for class "+t.name+" in "+t.filePath+"! Add "+t.name+" to the NgModule to fix it."}).join("\n"));return t}function hh(t,e,n,r){var i=[],o=[],a=[],s=[],u=e.hasDecorators(r),c=!1;return r.endsWith(".d.ts")&&!u||e.getSymbolsOf(r).forEach(function(r){var u=e.resolveSymbol(r).metadata;if(u&&"error"!==u.__symbolic){var l,d,p,f,h=!1;if("class"===u.__symbolic)if(n.isDirective(r))h=!0,i.push(r);else if(n.isPipe(r))h=!0,o.push(r);else if(n.isNgModule(r)){var m=n.getNgModuleMetadata(r,!1);m&&(h=!0,s.push(m))}else if(n.isInjectable(r)){h=!0;var g=n.getInjectableMetadata(r,null,!1);g&&a.push(g)}h||(c=c||(l=t,d=u,p=!1,f=function(){function t(){}return t.prototype.visitArray=function(t,e){var n=this;t.forEach(function(t){return Be(t,n,e)})},t.prototype.visitStringMap=function(t,e){var n=this;Object.keys(t).forEach(function(r){return Be(t[r],n,e)})},t.prototype.visitPrimitive=function(t,e){},t.prototype.visitOther=function(t,e){t instanceof vn&&!l.isSourceFile(t.filePath)&&(p=!0)},t}(),Be(d,new f,null),p))}}),{fileName:r,directives:i,pipes:o,ngModules:s,injectables:a,exportsNonSourceFiles:c}}function mh(t,e,n,r){var i=[],o=[];return e.hasDecorators(r)&&e.getSymbolsOf(r).forEach(function(t){var r=e.resolveSymbol(t).metadata;if(r&&"error"!==r.__symbolic&&"class"===r.__symbolic)if(n.isInjectable(t)){var a=n.getInjectableMetadata(t,null,!1);a&&i.push(a)}else if(n.isNgModule(t)){var s=n.getShallowModuleMetadata(t);s&&o.push(s)}}),{fileName:r,injectables:i,shallowModules:o}}function gh(t){var e=[],n=new Map,r=new Set;t.forEach(function(t){t.ngModules.forEach(function(t){e.push(t),t.declaredDirectives.forEach(function(e){return n.set(e.reference,t)}),t.declaredPipes.forEach(function(e){return n.set(e.reference,t)})}),t.directives.forEach(function(t){return r.add(t)}),t.pipes.forEach(function(t){return r.add(t)})});var i=[];return r.forEach(function(t){n.has(t)||i.push(t)}),{ngModules:e,ngModuleByPipeOrDirective:n,symbolsMissingModule:i,files:t}}function yh(t){return fh(gh(t))}var _h="ngFormattedMessage";function vh(t,e){if(void 0===e&&(e=0),!t)return"";var n=t.position?t.position.fileName+"("+(t.position.line+1)+","+(t.position.column+1)+")":"",r=n&&0!==e?" at "+n:"",i=""+(n&&0===e?n+": ":"")+t.message+r;return""+function t(e){if(e<=0)return"";if(e<6)return[""," ","  ","   ","    ","     "][e];var n=t(Math.floor(e/2));return n+n+(e%2==1?" ":"")}(e)+i+(t.next&&"\n"+vh(t.next,e+2)||"")}function bh(t){var e=Ye(vh(t)+".");return e[_h]=!0,e.chain=t,e.position=t.position,e}function xh(t){return!!t[_h]}var wh="@angular/core",jh=/^\$.*\$$/,Oh={__symbolic:"ignore"},Ch="useValue",Eh="provide",Sh=new Set([Ch,"useFactory","data","id","loadChildren"]);function Mh(t){return t&&"ignore"==t.__symbolic}var Th=function(){function t(t,e,n,r,i){void 0===n&&(n=[]),void 0===r&&(r=[]);var o=this;this.summaryResolver=t,this.symbolResolver=e,this.errorRecorder=i,this.annotationCache=new Map,this.shallowAnnotationCache=new Map,this.propertyCache=new Map,this.parameterCache=new Map,this.methodCache=new Map,this.staticCache=new Map,this.conversionMap=new Map,this.resolvedExternalReferences=new Map,this.annotationForParentClassWithSummaryKind=new Map,this.initializeConversionMap(),n.forEach(function(t){return o._registerDecoratorOrConstructor(o.getStaticSymbol(t.filePath,t.name),t.ctor)}),r.forEach(function(t){return o._registerFunction(o.getStaticSymbol(t.filePath,t.name),t.fn)}),this.annotationForParentClassWithSummaryKind.set(jn.Directive,[P,B]),this.annotationForParentClassWithSummaryKind.set(jn.Pipe,[V]),this.annotationForParentClassWithSummaryKind.set(jn.NgModule,[q]),this.annotationForParentClassWithSummaryKind.set(jn.Injectable,[G,V,P,B,q])}return t.prototype.componentModuleUrl=function(t){var e=this.findSymbolDeclaration(t);return this.symbolResolver.getResourcePath(e)},t.prototype.resolveExternalReference=function(t,e){var n=void 0;if(!e){n=t.moduleName+":"+t.name;var r=this.resolvedExternalReferences.get(n);if(r)return r}var i=this.symbolResolver.getSymbolByModule(t.moduleName,t.name,e),o=this.findSymbolDeclaration(i);return e||(this.symbolResolver.recordModuleNameForFileName(i.filePath,t.moduleName),this.symbolResolver.recordImportAs(o,i)),n&&this.resolvedExternalReferences.set(n,o),o},t.prototype.findDeclaration=function(t,e,n){return this.findSymbolDeclaration(this.symbolResolver.getSymbolByModule(t,e,n))},t.prototype.tryFindDeclaration=function(t,e,n){var r=this;return this.symbolResolver.ignoreErrorsFor(function(){return r.findDeclaration(t,e,n)})},t.prototype.findSymbolDeclaration=function(t){var e=this.symbolResolver.resolveSymbol(t);if(e){var n=e.metadata;if(n&&"resolved"===n.__symbolic&&(n=n.symbol),n instanceof vn)return this.findSymbolDeclaration(e.metadata)}return t},t.prototype.tryAnnotations=function(t){var e=this.errorRecorder;this.errorRecorder=function(t,e){};try{return this.annotations(t)}finally{this.errorRecorder=e}},t.prototype.annotations=function(t){var e=this;return this._annotations(t,function(t,n){return e.simplify(t,n)},this.annotationCache)},t.prototype.shallowAnnotations=function(t){var e=this;return this._annotations(t,function(t,n){return e.simplify(t,n,!0)},this.shallowAnnotationCache)},t.prototype._annotations=function(t,e,n){var r=n.get(t);if(!r){r=[];var i=this.getTypeMetadata(t),a=this.findParentType(t,i);if(a){var s=this.annotations(a);r.push.apply(r,Object(o.__spread)(s))}var u=[];if(i.decorators&&(u=e(t,i.decorators))&&r.push.apply(r,Object(o.__spread)(u)),a&&!this.summaryResolver.isLibraryFile(t.filePath)&&this.summaryResolver.isLibraryFile(a.filePath)){var c=this.summaryResolver.resolveSummary(a);if(c&&c.type){var l=this.annotationForParentClassWithSummaryKind.get(c.type.summaryKind);l.some(function(t){return u.some(function(e){return t.isTypeOf(e)})})||this.reportError(Wh(Ah("Class "+t.name+" in "+t.filePath+" extends from a "+jn[c.type.summaryKind]+" in another compilation unit without duplicating the decorator",void 0,"Please add a "+l.map(function(t){return t.ngMetadataName}).join(" or ")+" decorator to the class"),t),t)}}n.set(t,r.filter(function(t){return!!t}))}return r},t.prototype.propMetadata=function(t){var e=this,n=this.propertyCache.get(t);if(!n){var r=this.getTypeMetadata(t);n={};var i=this.findParentType(t,r);if(i){var a=this.propMetadata(i);Object.keys(a).forEach(function(t){n[t]=a[t]})}var s=r.members||{};Object.keys(s).forEach(function(r){var i=s[r].find(function(t){return"property"==t.__symbolic||"method"==t.__symbolic}),a=[];n[r]&&a.push.apply(a,Object(o.__spread)(n[r])),n[r]=a,i&&i.decorators&&a.push.apply(a,Object(o.__spread)(e.simplify(t,i.decorators)))}),this.propertyCache.set(t,n)}return n},t.prototype.parameters=function(t){var e=this;if(!(t instanceof vn))return this.reportError(new Error("parameters received "+JSON.stringify(t)+" which is not a StaticSymbol"),t),[];try{var n=this.parameterCache.get(t);if(!n){var r=this.getTypeMetadata(t),i=this.findParentType(t,r),a=r?r.members:null,s=a?a.__ctor__:null;if(s){var u=s.find(function(t){return"constructor"==t.__symbolic}),c=u.parameters||[],l=this.simplify(t,u.parameterDecorators||[]);n=[],c.forEach(function(r,i){var a=[],s=e.trySimplify(t,r);s&&a.push(s);var u=l?l[i]:null;u&&a.push.apply(a,Object(o.__spread)(u)),n.push(a)})}else i&&(n=this.parameters(i));n||(n=[]),this.parameterCache.set(t,n)}return n}catch(e){throw console.error("Failed on type "+JSON.stringify(t)+" with error "+e),e}},t.prototype._methodNames=function(t){var e=this.methodCache.get(t);if(!e){var n=this.getTypeMetadata(t);e={};var r=this.findParentType(t,n);if(r){var i=this._methodNames(r);Object.keys(i).forEach(function(t){e[t]=i[t]})}var o=n.members||{};Object.keys(o).forEach(function(t){var n=o[t].some(function(t){return"method"==t.__symbolic});e[t]=e[t]||n}),this.methodCache.set(t,e)}return e},t.prototype._staticMembers=function(t){var e=this.staticCache.get(t);if(!e){var n=this.getTypeMetadata(t).statics||{};e=Object.keys(n),this.staticCache.set(t,e)}return e},t.prototype.findParentType=function(t,e){var n=this.trySimplify(t,e.extends);if(n instanceof vn)return n},t.prototype.hasLifecycleHook=function(t,e){t instanceof vn||this.reportError(new Error("hasLifecycleHook received "+JSON.stringify(t)+" which is not a StaticSymbol"),t);try{return!!this._methodNames(t)[e]}catch(e){throw console.error("Failed on type "+JSON.stringify(t)+" with error "+e),e}},t.prototype.guards=function(t){var e,n;if(!(t instanceof vn))return this.reportError(new Error("guards received "+JSON.stringify(t)+" which is not a StaticSymbol"),t),{};var r=this._staticMembers(t),i={};try{for(var a=Object(o.__values)(r),s=a.next();!s.done;s=a.next()){var u=s.value;if(u.endsWith("TypeGuard")){var c=u.substr(0,u.length-"TypeGuard".length),l=void 0;c.endsWith("UseIf")?(c=u.substr(0,c.length-"UseIf".length),l="UseIf"):l=this.getStaticSymbol(t.filePath,t.name,[u]),i[c]=l}}}catch(t){e={error:t}}finally{try{s&&!s.done&&(n=a.return)&&n.call(a)}finally{if(e)throw e.error}}return i},t.prototype._registerDecoratorOrConstructor=function(t,e){this.conversionMap.set(t,function(t,n){return new(e.bind.apply(e,Object(o.__spread)([void 0],n)))})},t.prototype._registerFunction=function(t,e){this.conversionMap.set(t,function(t,n){return e.apply(void 0,n)})},t.prototype.initializeConversionMap=function(){this._registerDecoratorOrConstructor(this.findDeclaration(wh,"Injectable"),G),this.injectionToken=this.findDeclaration(wh,"InjectionToken"),this.opaqueToken=this.findDeclaration(wh,"OpaqueToken"),this.ROUTES=this.tryFindDeclaration("@angular/router","ROUTES"),this.ANALYZE_FOR_ENTRY_COMPONENTS=this.findDeclaration(wh,"ANALYZE_FOR_ENTRY_COMPONENTS"),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"Host"),$),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"Self"),Q),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"SkipSelf"),Z),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"Inject"),S),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"Optional"),X),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"Attribute"),T),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"ContentChild"),A),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"ContentChildren"),k),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"ViewChild"),I),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"ViewChildren"),D),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"Input"),z),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"Output"),H),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"Pipe"),V),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"HostBinding"),U),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"HostListener"),W),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"Directive"),P),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"Component"),B),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"NgModule"),q),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"Host"),$),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"Self"),Q),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"SkipSelf"),Z),this._registerDecoratorOrConstructor(this.findDeclaration(wh,"Optional"),X)},t.prototype.getStaticSymbol=function(t,e,n){return this.symbolResolver.getStaticSymbol(t,e,n)},t.prototype.trySimplify=function(t,e){var n=this.errorRecorder;this.errorRecorder=function(t,e){};var r=this.simplify(t,e);return this.errorRecorder=n,r},t.prototype.simplify=function(t,e,n){void 0===n&&(n=!1);var r,i=this,a=zh.empty,s=new Map;try{r=function t(e,n,r,u){function c(t){var e=i.symbolResolver.resolveSymbol(t);return e?e.metadata:null}function l(n){return t(e,n,r,0)}function d(n,o){if(n===e)return t(n,o,r+1,u);try{return t(n,o,r+1,u)}catch(t){if(!Dh(t))throw t;var a=t.chain?"references '"+t.symbol.name+"'":function(t){if(t.summary)return t.summary;switch(t.message){case Ih:if(t.context&&t.context.className)return"references non-exported class "+t.context.className;break;case Ph:return"is not initialized";case Rh:return"is a destructured variable";case Nh:return"could not be resolved";case Fh:return t.context&&t.context.name?"calls '"+t.context.name+"'":"calls a function";case Lh:return t.context&&t.context.name?"references local variable "+t.context.name:"references a local variable"}return"contains the error"}(t),s={message:"'"+n.name+"' "+a,position:t.position,next:t.chain};i.error({message:t.message,advise:t.advise,context:t.context,chain:s,symbol:n},e)}}function p(n){var f,h,m,g;if(Vh(n))return n;if(n instanceof Array){var y=[];try{for(var _=Object(o.__values)(n),v=_.next();!v.done;v=_.next()){var b=v.value;if(b&&"spread"===b.__symbolic){var x=l(b.expression);if(Array.isArray(x)){try{for(var w=Object(o.__values)(x),j=w.next();!j.done;j=w.next()){var O=j.value;y.push(O)}}catch(t){m={error:t}}finally{try{j&&!j.done&&(g=w.return)&&g.call(w)}finally{if(m)throw m.error}}continue}}var C=p(b);Mh(C)||y.push(C)}}catch(t){f={error:t}}finally{try{v&&!v.done&&(h=_.return)&&h.call(_)}finally{if(f)throw f.error}}return y}if(n instanceof vn)return n===i.injectionToken||i.conversionMap.has(n)||u>0&&!n.members.length?n:null!=(R=c(E=n))?d(E,R):E;if(n){if(n.__symbolic){var E=void 0;switch(n.__symbolic){case"binop":var S=p(n.left);if(Mh(S))return S;var M=p(n.right);if(Mh(M))return M;switch(n.operator){case"&&":return S&&M;case"||":return S||M;case"|":return S|M;case"^":return S^M;case"&":return S&M;case"==":return S==M;case"!=":return S!=M;case"===":return S===M;case"!==":return S!==M;case"<":return S<M;case">":return S>M;case"<=":return S<=M;case">=":return S>=M;case"<<":return S<<M;case">>":return S>>M;case"+":return S+M;case"-":return S-M;case"*":return S*M;case"/":return S/M;case"%":return S%M}return null;case"if":return p(p(n.condition)?n.thenExpression:n.elseExpression);case"pre":var T=p(n.operand);if(Mh(T))return T;switch(n.operator){case"+":return T;case"-":return-T;case"!":return!T;case"~":return~T}return null;case"index":var k=l(n.expression),A=l(n.index);return k&&Vh(A)?k[A]:null;case"select":var D=n.member,I=e,P=p(n.expression);if(P instanceof vn){var R,N=P.members.concat(D);return null!=(R=c(I=i.getStaticSymbol(P.filePath,P.name,N)))?d(I,R):I}return P&&Vh(D)?d(I,P[D]):null;case"reference":var F=n.name,L=a.resolve(F);if(L!=zh.missing)return L;break;case"resolved":try{return p(n.symbol)}catch(t){throw Dh(t)&&null!=n.fileName&&null!=n.line&&null!=n.character&&(t.position={fileName:n.fileName,line:n.line,column:n.character}),t}case"class":case"function":return e;case"new":case"call":if((E=t(e,n.expression,r+1,0))instanceof vn){if(E===i.injectionToken||E===i.opaqueToken)return e;var B=n.arguments||[],V=i.conversionMap.get(E);if(V){var z=B.map(function(t){return d(e,t)}).map(function(t){return Mh(t)?void 0:t});return V(e,z)}return function(t,n,u,c){if(n&&"function"==n.__symbolic){s.get(t)&&i.error({message:"Recursion is not supported",summary:"called '"+t.name+"' recursively",value:n},t);try{var l=n.value;if(l&&(0!=r||"error"!=l.__symbolic)){var f=n.parameters,h=n.defaults;u=u.map(function(t){return d(e,t)}).map(function(t){return Mh(t)?void 0:t}),h&&h.length>u.length&&u.push.apply(u,Object(o.__spread)(h.slice(u.length).map(function(t){return p(t)}))),s.set(t,!0);for(var m=zh.build(),g=0;g<f.length;g++)m.define(f[g],u[g]);var y,_=a;try{a=m.done(),y=d(t,l)}finally{a=_}return y}}finally{s.delete(t)}}if(0===r)return Oh;var v=void 0;if(c&&"resolved"==c.__symbolic){var b=c.line,x=c.character,w=c.fileName;null!=w&&null!=b&&null!=x&&(v={fileName:w,line:b,column:x})}i.error({message:Fh,context:t,value:n,position:v},e)}(E,c(E),B,n.expression)}return Oh;case"error":var H=n.message;return null!=n.line?i.error({message:H,context:n.context,value:n,position:{fileName:n.fileName,line:n.line,column:n.character}},e):i.error({message:H,context:n.context},e),Oh;case"ignore":return n}return null}return function(t,e){if(!t)return{};var n={};return Object.keys(t).forEach(function(r){var i=e(t[r],r);Mh(i)||(jh.test(r)?Object.defineProperty(n,r,{enumerable:!1,configurable:!0,value:i}):n[r]=i)}),n}(n,function(o,a){if(Sh.has(a)){if(a===Ch&&Eh in n){var s=p(n.provide);if(s===i.ROUTES||s==i.ANALYZE_FOR_ENTRY_COMPONENTS)return p(o)}return t(e,o,r,u+1)}return p(o)})}return Oh}return p(n)}(t,e,0,n?1:0)}catch(e){if(!this.errorRecorder)throw Wh(e,t);this.reportError(e,t)}if(!Mh(r))return r},t.prototype.getTypeMetadata=function(t){var e=this.symbolResolver.resolveSymbol(t);return e&&e.metadata?e.metadata:{__symbolic:"class"}},t.prototype.reportError=function(t,e,n){if(!this.errorRecorder)throw t;this.errorRecorder(Wh(t,e),e&&e.filePath||n)},t.prototype.error=function(t,e){var n=t.message,r=t.summary,i=t.advise,o=t.position,a=t.context,s=(t.value,t.symbol),u=t.chain;this.reportError(Ah(n,r,i,o,s,a,u),e)},t}(),kh="ngMetadataError";function Ah(t,e,n,r,i,o,a){var s=Ye(t);return s[kh]=!0,n&&(s.advise=n),r&&(s.position=r),e&&(s.summary=e),o&&(s.context=o),a&&(s.chain=a),i&&(s.symbol=i),s}function Dh(t){return!!t[kh]}var Ih="Reference to non-exported class",Ph="Variable not initialized",Rh="Destructuring not supported",Nh="Could not resolve type",Fh="Function call not supported",Lh="Reference to a local symbol",Bh="Lambda not supported";function Vh(t){return null===t||"function"!=typeof t&&"object"!=typeof t}var zh=function(){function t(){}return t.build=function(){var e=new Map;return{define:function(t,n){return e.set(t,n),this},done:function(){return e.size>0?new Hh(e):t.empty}}},t.missing={},t.empty={resolve:function(e){return t.missing}},t}(),Hh=function(t){function e(e){var n=t.call(this)||this;return n.bindings=e,n}return Object(o.__extends)(e,t),e.prototype.resolve=function(t){return this.bindings.has(t)?this.bindings.get(t):zh.missing},e}(zh);function Uh(t,e){return{message:""+function(t,e){switch(t){case Ih:if(e&&e.className)return"References to a non-exported class are not supported in decorators but "+e.className+" was referenced.";break;case Ph:return"Only initialized variables and constants can be referenced in decorators because the value of this variable is needed by the template compiler";case Rh:return"Referencing an exported destructured variable or constant is not supported in decorators and this value is needed by the template compiler";case Nh:if(e&&e.typeName)return"Could not resolve type "+e.typeName;break;case Fh:return e&&e.name?"Function calls are not supported in decorators but '"+e.name+"' was called":"Function calls are not supported in decorators";case Lh:if(e&&e.name)return"Reference to a local (non-exported) symbols are not supported in decorators but '"+e.name+"' was referenced";break;case Bh:return"Function expressions are not supported in decorators"}return t}(t.message,t.context)+(t.symbol?" in '"+t.symbol.name+"'":""),position:t.position,next:t.next?Uh(t.next,e):e?{message:e}:void 0}}function Wh(t,e){if(Dh(t)){var n=t.position;return bh(Uh({message:"Error during template compile of '"+e.name+"'",position:n,next:{message:t.message,next:t.chain,context:t.context,symbol:t.symbol}},t.advise||function(t,e){switch(t){case Ih:if(e&&e.className)return"Consider exporting '"+e.className+"'";break;case Rh:return"Consider simplifying to avoid destructuring";case Lh:if(e&&e.name)return"Consider exporting '"+e.name+"'";break;case Bh:return"Consider changing the function expression into an exported function"}}(t.message,t.context)))}return t}var qh=function(){function t(t,e){this.host=t,this.staticSymbolCache=e,this.summaryCache=new Map,this.loadedFilePaths=new Map,this.importAs=new Map,this.knownFileNameToModuleNames=new Map}return t.prototype.isLibraryFile=function(t){return!this.host.isSourceFile(_p(t))},t.prototype.toSummaryFileName=function(t,e){return this.host.toSummaryFileName(t,e)},t.prototype.fromSummaryFileName=function(t,e){return this.host.fromSummaryFileName(t,e)},t.prototype.resolveSummary=function(t){var e=t.members.length?this.staticSymbolCache.get(t.filePath,t.name):t,n=this.summaryCache.get(e);return n||(this._loadSummaryFile(t.filePath),n=this.summaryCache.get(t)),e===t&&n||null},t.prototype.getSymbolsOf=function(t){return this._loadSummaryFile(t)?Array.from(this.summaryCache.keys()).filter(function(e){return e.filePath===t}):null},t.prototype.getImportAs=function(t){return t.assertNoMembers(),this.importAs.get(t)},t.prototype.getKnownModuleName=function(t){return this.knownFileNameToModuleNames.get(t)||null},t.prototype.addSummary=function(t){this.summaryCache.set(t.symbol,t)},t.prototype._loadSummaryFile=function(t){var e=this,n=this.loadedFilePaths.get(t);if(null!=n)return n;var r,i,o,a=null;if(this.isLibraryFile(t)){var s=xp(t);try{a=this.host.loadSummary(s)}catch(t){throw console.error("Error loading summary file "+s),t}}if(n=null!=a,this.loadedFilePaths.set(t,n),a){var u=(r=this.staticSymbolCache,i=t,o=a,new oh(r,this).deserialize(i,o)),c=u.moduleName,l=u.summaries,d=u.importAs;l.forEach(function(t){return e.summaryCache.set(t.symbol,t)}),c&&this.knownFileNameToModuleNames.set(t,c),d.forEach(function(t){e.importAs.set(t.symbol,t.importAs)})}return n},t}();function Gh(t){return{resolve:function(e,n){var r=t.resourceNameToFileName(n,e);if(!r)throw Ye("Couldn't resolve resource "+n+" from "+e);return r}}}function Yh(t,e,n){var r,i=e.translations||"",o=Gh(t),a=new bn,s=new qh(t,a),u=new Jf(t,a,s),c=new Th(s,u,[],[],n);r=e.enableIvy?new Hu:new up(new Hu,i,e.i18nFormat,e.missingTranslation,console);var l=new Sd({defaultEncapsulation:_.Emulated,useJit:!1,missingTranslation:e.missingTranslation,preserveWhitespaces:e.preserveWhitespaces,strictInjectionParameters:e.strictInjectionParameters}),d=new Td({get:function(e){return t.loadResource(e)}},o,r,l),p=new uu(new Ks),f=new sc,h=new ff(l,c,p,f,r,console,[]),m=new Mp(l,r,new Qp(c),new Dd(c),new tf(c),s,f,d,console,a,c,n),g=new Af(c),y=new Cf(e,c);return{compiler:new sh(l,e,t,c,m,h,new Ls(o),g,y,new Xp(c),new pp(c,!!e.enableIvy),new $p,s,u),reflector:c}}var Kh=function(){return function(){}}(),Xh=function(){function t(){this._summaries=new Map}return t.prototype.isLibraryFile=function(){return!1},t.prototype.toSummaryFileName=function(t){return t},t.prototype.fromSummaryFileName=function(t){return t},t.prototype.resolveSummary=function(t){return this._summaries.get(t)||null},t.prototype.getSymbolsOf=function(){return[]},t.prototype.getImportAs=function(t){return t},t.prototype.getKnownModuleName=function(t){return null},t.prototype.addSummary=function(t){this._summaries.set(t.symbol,t)},t}();function Qh(t,e,n,r,i){for(var o=r.createChildWihtLocalVars(),a=0;a<t.length;a++)o.vars.set(t[a],e[a]);var s=i.visitAllStatements(n,o);return s?s.value:null}var Zh=function(){function t(t,e,n,r){this.parent=t,this.instance=e,this.className=n,this.vars=r,this.exports=[]}return t.prototype.createChildWihtLocalVars=function(){return new t(this,this.instance,this.className,new Map)},t}(),$h=function(){return function(t){this.value=t}}();var Jh=function(){function t(t){this.reflector=t}return t.prototype.debugAst=function(t){return Zp(t)},t.prototype.visitDeclareVarStmt=function(t,e){var n=t.value?t.value.visitExpression(this,e):void 0;return e.vars.set(t.name,n),t.hasModifier(It.Exported)&&e.exports.push(t.name),null},t.prototype.visitWriteVarExpr=function(t,e){for(var n=t.value.visitExpression(this,e),r=e;null!=r;){if(r.vars.has(t.name))return r.vars.set(t.name,n),n;r=r.parent}throw new Error("Not declared variable "+t.name)},t.prototype.visitWrappedNodeExpr=function(t,e){throw new Error("Cannot interpret a WrappedNodeExpr.")},t.prototype.visitTypeofExpr=function(t,e){throw new Error("Cannot interpret a TypeofExpr")},t.prototype.visitReadVarExpr=function(t,e){var n=t.name;if(null!=t.builtin)switch(t.builtin){case wt.Super:return e.instance.__proto__;case wt.This:return e.instance;case wt.CatchError:n=em;break;case wt.CatchStack:n=nm;break;default:throw new Error("Unknown builtin variable "+t.builtin)}for(var r=e;null!=r;){if(r.vars.has(n))return r.vars.get(n);r=r.parent}throw new Error("Not declared variable "+n)},t.prototype.visitWriteKeyExpr=function(t,e){var n=t.receiver.visitExpression(this,e),r=t.index.visitExpression(this,e),i=t.value.visitExpression(this,e);return n[r]=i,i},t.prototype.visitWritePropExpr=function(t,e){var n=t.receiver.visitExpression(this,e),r=t.value.visitExpression(this,e);return n[t.name]=r,r},t.prototype.visitInvokeMethodExpr=function(t,e){var n,r=t.receiver.visitExpression(this,e),i=this.visitAllExpressions(t.args,e);if(null!=t.builtin)switch(t.builtin){case Ct.ConcatArray:n=r.concat.apply(r,Object(o.__spread)(i));break;case Ct.SubscribeObservable:n=r.subscribe({next:i[0]});break;case Ct.Bind:n=r.bind.apply(r,Object(o.__spread)(i));break;default:throw new Error("Unknown builtin method "+t.builtin)}else n=r[t.name].apply(r,i);return n},t.prototype.visitInvokeFunctionExpr=function(t,e){var n=this.visitAllExpressions(t.args,e),r=t.fn;return r instanceof St&&r.builtin===wt.Super?(e.instance.constructor.prototype.constructor.apply(e.instance,n),null):t.fn.visitExpression(this,e).apply(null,n)},t.prototype.visitReturnStmt=function(t,e){return new $h(t.value.visitExpression(this,e))},t.prototype.visitDeclareClassStmt=function(t,e){var n=function(t,e,n){var r={};t.getters.forEach(function(i){r[i.name]={configurable:!1,get:function(){var r=new Zh(e,this,t.name,e.vars);return Qh([],[],i.body,r,n)}}}),t.methods.forEach(function(i){var o=i.params.map(function(t){return t.name});r[i.name]={writable:!1,configurable:!1,value:function(){for(var r=[],a=0;a<arguments.length;a++)r[a]=arguments[a];var s=new Zh(e,this,t.name,e.vars);return Qh(o,r,i.body,s,n)}}});var i=t.constructorMethod.params.map(function(t){return t.name}),o=function(){for(var r=this,o=[],a=0;a<arguments.length;a++)o[a]=arguments[a];var s=new Zh(e,this,t.name,e.vars);t.fields.forEach(function(t){r[t.name]=void 0}),Qh(i,o,t.constructorMethod.body,s,n)},a=t.parent?t.parent.visitExpression(n,e):Object;return o.prototype=Object.create(a.prototype,r),o}(t,e,this);return e.vars.set(t.name,n),t.hasModifier(It.Exported)&&e.exports.push(t.name),null},t.prototype.visitExpressionStmt=function(t,e){return t.expr.visitExpression(this,e)},t.prototype.visitIfStmt=function(t,e){return t.condition.visitExpression(this,e)?this.visitAllStatements(t.trueCase,e):null!=t.falseCase?this.visitAllStatements(t.falseCase,e):null},t.prototype.visitTryCatchStmt=function(t,e){try{return this.visitAllStatements(t.bodyStmts,e)}catch(r){var n=e.createChildWihtLocalVars();return n.vars.set(em,r),n.vars.set(nm,r.stack),this.visitAllStatements(t.catchStmts,n)}},t.prototype.visitThrowStmt=function(t,e){throw t.error.visitExpression(this,e)},t.prototype.visitCommentStmt=function(t,e){return null},t.prototype.visitJSDocCommentStmt=function(t,e){return null},t.prototype.visitInstantiateExpr=function(t,e){var n=this.visitAllExpressions(t.args,e),r=t.classExpr.visitExpression(this,e);return new(r.bind.apply(r,Object(o.__spread)([void 0],n)))},t.prototype.visitLiteralExpr=function(t,e){return t.value},t.prototype.visitExternalExpr=function(t,e){return this.reflector.resolveExternalReference(t.value)},t.prototype.visitConditionalExpr=function(t,e){return t.condition.visitExpression(this,e)?t.trueCase.visitExpression(this,e):null!=t.falseCase?t.falseCase.visitExpression(this,e):null},t.prototype.visitNotExpr=function(t,e){return!t.condition.visitExpression(this,e)},t.prototype.visitAssertNotNullExpr=function(t,e){return t.condition.visitExpression(this,e)},t.prototype.visitCastExpr=function(t,e){return t.value.visitExpression(this,e)},t.prototype.visitFunctionExpr=function(t,e){return tm(t.params.map(function(t){return t.name}),t.statements,e,this)},t.prototype.visitDeclareFunctionStmt=function(t,e){var n=t.params.map(function(t){return t.name});return e.vars.set(t.name,tm(n,t.statements,e,this)),t.hasModifier(It.Exported)&&e.exports.push(t.name),null},t.prototype.visitBinaryOperatorExpr=function(t,e){var n=this,r=function(){return t.lhs.visitExpression(n,e)},i=function(){return t.rhs.visitExpression(n,e)};switch(t.operator){case ct.Equals:return r()==i();case ct.Identical:return r()===i();case ct.NotEquals:return r()!=i();case ct.NotIdentical:return r()!==i();case ct.And:return r()&&i();case ct.Or:return r()||i();case ct.Plus:return r()+i();case ct.Minus:return r()-i();case ct.Divide:return r()/i();case ct.Multiply:return r()*i();case ct.Modulo:return r()%i();case ct.Lower:return r()<i();case ct.LowerEquals:return r()<=i();case ct.Bigger:return r()>i();case ct.BiggerEquals:return r()>=i();default:throw new Error("Unknown operator "+t.operator)}},t.prototype.visitReadPropExpr=function(t,e){return t.receiver.visitExpression(this,e)[t.name]},t.prototype.visitReadKeyExpr=function(t,e){return t.receiver.visitExpression(this,e)[t.index.visitExpression(this,e)]},t.prototype.visitLiteralArrayExpr=function(t,e){return this.visitAllExpressions(t.entries,e)},t.prototype.visitLiteralMapExpr=function(t,e){var n=this,r={};return t.entries.forEach(function(t){return r[t.key]=t.value.visitExpression(n,e)}),r},t.prototype.visitCommaExpr=function(t,e){var n=this.visitAllExpressions(t.parts,e);return n[n.length-1]},t.prototype.visitAllExpressions=function(t,e){var n=this;return t.map(function(t){return t.visitExpression(n,e)})},t.prototype.visitAllStatements=function(t,e){for(var n=0;n<t.length;n++){var r=t[n].visitStatement(this,e);if(r instanceof $h)return r}return null},t}();function tm(t,e,n,r){return function(){for(var i=[],o=0;o<arguments.length;o++)i[o]=arguments[o];return Qh(t,i,e,n,r)}}var em="error",nm="stack",rm=function(){function t(t,e,n,r,i,o,a,s,u,c){this._metadataResolver=t,this._templateParser=e,this._styleCompiler=n,this._viewCompiler=r,this._ngModuleCompiler=i,this._summaryResolver=o,this._reflector=a,this._compilerConfig=s,this._console=u,this.getExtraNgModuleProviders=c,this._compiledTemplateCache=new Map,this._compiledHostTemplateCache=new Map,this._compiledDirectiveWrapperCache=new Map,this._compiledNgModuleCache=new Map,this._sharedStylesheetCount=0,this._addedAotSummaries=new Set}return t.prototype.compileModuleSync=function(t){return Ue(this._compileModuleAndComponents(t,!0))},t.prototype.compileModuleAsync=function(t){return Promise.resolve(this._compileModuleAndComponents(t,!1))},t.prototype.compileModuleAndAllComponentsSync=function(t){return Ue(this._compileModuleAndAllComponents(t,!0))},t.prototype.compileModuleAndAllComponentsAsync=function(t){return Promise.resolve(this._compileModuleAndAllComponents(t,!1))},t.prototype.getComponentFactory=function(t){return this._metadataResolver.getDirectiveSummary(t).componentFactory},t.prototype.loadAotSummaries=function(t){this.clearCache(),this._addAotSummaries(t)},t.prototype._addAotSummaries=function(t){if(!this._addedAotSummaries.has(t)){this._addedAotSummaries.add(t);for(var e=t(),n=0;n<e.length;n++){var r=e[n];if("function"==typeof r)this._addAotSummaries(r);else{var i=r;this._summaryResolver.addSummary({symbol:i.type.reference,metadata:null,type:i})}}}},t.prototype.hasAotSummary=function(t){return!!this._summaryResolver.resolveSummary(t)},t.prototype._filterJitIdentifiers=function(t){var e=this;return t.map(function(t){return t.reference}).filter(function(t){return!e.hasAotSummary(t)})},t.prototype._compileModuleAndComponents=function(t,e){var n=this;return We(this._loadModules(t,e),function(){return n._compileComponents(t,null),n._compileModule(t)})},t.prototype._compileModuleAndAllComponents=function(t,e){var n=this;return We(this._loadModules(t,e),function(){var e=[];return n._compileComponents(t,e),{ngModuleFactory:n._compileModule(t),componentFactories:e}})},t.prototype._loadModules=function(t,e){var n=this,r=[],i=this._metadataResolver.getNgModuleMetadata(t);return this._filterJitIdentifiers(i.transitiveModule.modules).forEach(function(t){var i=n._metadataResolver.getNgModuleMetadata(t);n._filterJitIdentifiers(i.declaredDirectives).forEach(function(t){var o=n._metadataResolver.loadDirectiveMetadata(i.type.reference,t,e);o&&r.push(o)}),n._filterJitIdentifiers(i.declaredPipes).forEach(function(t){return n._metadataResolver.getOrLoadPipeMetadata(t)})}),qe(r)},t.prototype._compileModule=function(t){var e=this._compiledNgModuleCache.get(t);if(!e){var n=this._metadataResolver.getNgModuleMetadata(t),r=this.getExtraNgModuleProviders(n.type.reference),i=am(),o=this._ngModuleCompiler.compile(i,n,r);e=this._interpretOrJit(Yn(n),i.statements)[o.ngModuleFactoryVar],this._compiledNgModuleCache.set(n.type.reference,e)}return e},t.prototype._compileComponents=function(t,e){var n=this,r=this._metadataResolver.getNgModuleMetadata(t),i=new Map,o=new Set,a=this._filterJitIdentifiers(r.transitiveModule.modules);a.forEach(function(t){var r=n._metadataResolver.getNgModuleMetadata(t);n._filterJitIdentifiers(r.declaredDirectives).forEach(function(t){i.set(t,r);var a=n._metadataResolver.getDirectiveMetadata(t);if(a.isComponent&&(o.add(n._createCompiledTemplate(a,r)),e)){var s=n._createCompiledHostTemplate(a.type.reference,r);o.add(s),e.push(a.componentFactory)}})}),a.forEach(function(t){var e=n._metadataResolver.getNgModuleMetadata(t);n._filterJitIdentifiers(e.declaredDirectives).forEach(function(t){var e=n._metadataResolver.getDirectiveMetadata(t);e.isComponent&&e.entryComponents.forEach(function(t){var e=i.get(t.componentType);o.add(n._createCompiledHostTemplate(t.componentType,e))})}),e.entryComponents.forEach(function(t){if(!n.hasAotSummary(t.componentType)){var e=i.get(t.componentType);o.add(n._createCompiledHostTemplate(t.componentType,e))}})}),o.forEach(function(t){return n._compileTemplate(t)})},t.prototype.clearCacheFor=function(t){this._compiledNgModuleCache.delete(t),this._metadataResolver.clearCacheFor(t),this._compiledHostTemplateCache.delete(t),this._compiledTemplateCache.get(t)&&this._compiledTemplateCache.delete(t)},t.prototype.clearCache=function(){this._metadataResolver.clearCache(),this._compiledTemplateCache.clear(),this._compiledHostTemplateCache.clear(),this._compiledNgModuleCache.clear()},t.prototype._createCompiledHostTemplate=function(t,e){if(!e)throw new Error("Component "+en(t)+" is not part of any NgModule or the module has not been imported into your module.");var n=this._compiledHostTemplateCache.get(t);if(!n){var r=this._metadataResolver.getDirectiveMetadata(t);om(r);var i=this._metadataResolver.getHostComponentMetadata(r,r.componentFactory.viewDefFactory);n=new im(!0,r.type,i,e,[r.type]),this._compiledHostTemplateCache.set(t,n)}return n},t.prototype._createCompiledTemplate=function(t,e){var n=this._compiledTemplateCache.get(t.type.reference);return n||(om(t),n=new im(!1,t.type,t,e,e.transitiveModule.directives),this._compiledTemplateCache.set(t.type.reference,n)),n},t.prototype._compileTemplate=function(t){var e=this;if(!t.isCompiled){var n=t.compMeta,r=new Map,i=am(),o=this._styleCompiler.compileComponent(i,n);n.template.externalStylesheets.forEach(function(t){var i=e._styleCompiler.compileStyles(am(),n,t);r.set(t.moduleUrl,i)}),this._resolveStylesCompileResult(o,r);t.ngModule.transitiveModule.pipes.map(function(t){return e._metadataResolver.getPipeSummary(t.reference)});var a=this._parseTemplate(n,t.ngModule,t.directives),s=a.template,u=a.pipes,c=this._viewCompiler.compileComponent(i,n,s,Ee(o.stylesVar),u),l=this._interpretOrJit(Kn(t.ngModule.type,t.compMeta),i.statements),d=l[c.viewClassVar],p=l[c.rendererTypeVar];t.compiled(d,p)}},t.prototype._parseTemplate=function(t,e,n){var r=this,i=t.template.preserveWhitespaces,o=n.map(function(t){return r._metadataResolver.getDirectiveSummary(t.reference)}),a=e.transitiveModule.pipes.map(function(t){return r._metadataResolver.getPipeSummary(t.reference)});return this._templateParser.parse(t,t.template.htmlAst,o,a,e.schemas,qn(e.type,t,t.template),i)},t.prototype._resolveStylesCompileResult=function(t,e){var n=this;t.dependencies.forEach(function(t,r){var i=e.get(t.moduleUrl),o=n._resolveAndEvalStylesCompileResult(i,e);t.setValue(o)})},t.prototype._resolveAndEvalStylesCompileResult=function(t,e){return this._resolveStylesCompileResult(t,e),this._interpretOrJit(Gn(t.meta,this._sharedStylesheetCount++),t.outputCtx.statements)[t.stylesVar]},t.prototype._interpretOrJit=function(t,e){return this._compilerConfig.useJit?ro(t,e,this._reflector,this._compilerConfig.jitDevMode):function(t,e){var n=new Zh(null,null,null,new Map);new Jh(e).visitAllStatements(t,n);var r={};return n.exports.forEach(function(t){r[t]=n.vars.get(t)}),r}(e,this._reflector)},t}(),im=function(){function t(t,e,n,r,i){this.isHost=t,this.compType=e,this.compMeta=n,this.ngModule=r,this.directives=i,this._viewClass=null,this.isCompiled=!1}return t.prototype.compiled=function(t,e){this._viewClass=t,this.compMeta.componentViewType.setDelegate(t);for(var n in e)this.compMeta.rendererType[n]=e[n];this.isCompiled=!0},t}();function om(t){if(!t.isComponent)throw new Error("Could not compile '"+En(t.type)+"' because it is not a component.")}function am(){return{statements:[],genFilePath:"",importExpr:function(t){return Se({name:En(t),moduleName:null,runtime:t})},constantPool:new dn}}var sm=function(){return function(){}}();function um(){return new lm}function cm(){return new lm(".")}var lm=function(){function t(t){void 0===t&&(t=null),this._packagePrefix=t}return t.prototype.resolve=function(t,e){var n=e;null!=t&&t.length>0&&(n=function(t,e){var n=mm(encodeURI(e)),r=mm(t);if(null!=n[pm.Scheme])return gm(n);n[pm.Scheme]=r[pm.Scheme];for(var i=pm.Scheme;i<=pm.Port;i++)null==n[i]&&(n[i]=r[i]);if("/"==n[pm.Path][0])return gm(n);var o=r[pm.Path];null==o&&(o="/");var a=o.lastIndexOf("/");return o=o.substring(0,a+1)+n[pm.Path],n[pm.Path]=o,gm(n)}(t,n));var r=mm(n),i=this._packagePrefix;if(null!=i&&null!=r&&"package"==r[pm.Scheme]){var o=r[pm.Path];return(i=i.replace(/\/+$/,""))+"/"+(o=o.replace(/^\/+/,""))}return n},t}();function dm(t){var e=mm(t);return e&&e[pm.Scheme]||""}var pm,fm,hm=new RegExp("^(?:([^:/?#.]+):)?(?://(?:([^/?#]*)@)?([\\w\\d\\-\\u0100-\\uffff.%]*)(?::([0-9]+))?)?([^?#]+)?(?:\\?([^#]*))?(?:#(.*))?$");function mm(t){return t.match(hm)}function gm(t){var e,n,r,i,o,a,s,u,c=t[pm.Path];return c=null==c?"":function(t){if("/"==t)return"/";for(var e="/"==t[0]?"/":"",n="/"===t[t.length-1]?"/":"",r=t.split("/"),i=[],o=0,a=0;a<r.length;a++){var s=r[a];switch(s){case"":case".":break;case"..":i.length>0?i.pop():o++;break;default:i.push(s)}}if(""==e){for(;o-- >0;)i.unshift("..");0===i.length&&i.push(".")}return e+i.join("/")+n}(c),t[pm.Path]=c,e=t[pm.Scheme],n=t[pm.UserInfo],r=t[pm.Domain],i=t[pm.Port],o=c,a=t[pm.QueryData],s=t[pm.Fragment],u=[],null!=e&&u.push(e+":"),null!=r&&(u.push("//"),null!=n&&u.push(n+"@"),u.push(r),null!=i&&u.push(":"+i)),null!=o&&u.push(o),null!=a&&u.push("?"+a),null!=s&&u.push("#"+s),u.join("")}(fm=pm||(pm={}))[fm.Scheme=1]="Scheme",fm[fm.UserInfo=2]="UserInfo",fm[fm.Domain=3]="Domain",fm[fm.Port=4]="Port",fm[fm.Path=5]="Path",fm[fm.QueryData=6]="QueryData",fm[fm.Fragment=7]="Fragment";var ym=function(){function t(){}return t.prototype.get=function(t){return""},t}(),_m=function(){function t(t,e,n,r){this.host=t,this.staticSymbolResolver=e,this.messageBundle=n,this.metadataResolver=r}return t.prototype.extract=function(t){var e=this,n=ph(t,this.host,this.staticSymbolResolver,this.metadataResolver),r=n.files,i=n.ngModules;return Promise.all(i.map(function(t){return e.metadataResolver.loadNgModuleDirectiveAndPipeMetadata(t.type.reference,!1)})).then(function(){var t=[];if(r.forEach(function(n){var r=[];n.directives.forEach(function(t){var n=e.metadataResolver.getDirectiveMetadata(t);n&&n.isComponent&&r.push(n)}),r.forEach(function(n){var r=n.template.template,i=n.template.templateUrl,a=zi.fromArray(n.template.interpolation);t.push.apply(t,Object(o.__spread)(e.messageBundle.updateFromTemplate(r,i,a)))})}),t.length)throw new Error(t.map(function(t){return t.toString()}).join("\n"));return e.messageBundle})},t.create=function(e,n){var r=new Hu,i=Gh(e),o=new bn,a=new qh(e,o),s=new Jf(e,o,a),u=new Th(a,s),c=new Sd({defaultEncapsulation:_.Emulated,useJit:!1}),l=new Td({get:function(t){return e.loadResource(t)}},i,r,c),d=new sc,p=new Mp(c,r,new Qp(u),new Dd(u),new tf(u),a,d,l,console,o,u),f=new Wf(r,[],{},n);return{extractor:new t(e,s,f,p),staticReflector:u}},t}(),vm=function(){function t(t){this.directiveMatcher=t}return t.prototype.bind=function(t){if(!t.template)throw new Error("Binding without a template not yet supported");var e=bm.apply(t.template),n=xm.apply(t.template,this.directiveMatcher),r=n.directives,i=n.bindings,o=n.references,a=wm.apply(t.template,e),s=a.expressions,u=a.symbols,c=a.nestingLevel;return new jm(t,r,i,o,s,u,c)},t}(),bm=function(){function t(t){this.parentScope=t,this.namedEntities=new Map,this.childScopes=new Map}return t.apply=function(e){var n=new t;return n.ingest(e),n},t.prototype.ingest=function(t){var e=this;t instanceof _c?(t.variables.forEach(function(t){return e.visitVariable(t)}),t.children.forEach(function(t){return t.visit(e)})):t.forEach(function(t){return t.visit(e)})},t.prototype.visitElement=function(t){var e=this;t.references.forEach(function(t){return e.visitReference(t)}),t.children.forEach(function(t){return t.visit(e)})},t.prototype.visitTemplate=function(e){var n=this;e.references.forEach(function(t){return n.visitReference(t)});var r=new t(this);r.ingest(e),this.childScopes.set(e,r)},t.prototype.visitVariable=function(t){this.maybeDeclare(t)},t.prototype.visitReference=function(t){this.maybeDeclare(t)},t.prototype.visitContent=function(t){},t.prototype.visitBoundAttribute=function(t){},t.prototype.visitBoundEvent=function(t){},t.prototype.visitBoundText=function(t){},t.prototype.visitText=function(t){},t.prototype.visitTextAttribute=function(t){},t.prototype.visitIcu=function(t){},t.prototype.maybeDeclare=function(t){this.namedEntities.has(t.name)||this.namedEntities.set(t.name,t)},t.prototype.lookup=function(t){return this.namedEntities.has(t)?this.namedEntities.get(t):void 0!==this.parentScope?this.parentScope.lookup(t):null},t.prototype.getChildScope=function(t){var e=this.childScopes.get(t);if(void 0===e)throw new Error("Assertion error: child scope for "+t+" not found");return e},t}(),xm=function(){function t(t,e,n,r){this.matcher=t,this.directives=e,this.bindings=n,this.references=r}return t.apply=function(e,n){var r=new Map,i=new Map,o=new Map;return new t(n,r,i,o).ingest(e),{directives:r,bindings:i,references:o}},t.prototype.ingest=function(t){var e=this;t.forEach(function(t){return t.visit(e)})},t.prototype.visitElement=function(t){this.visitElementOrTemplate(t.name,t)},t.prototype.visitTemplate=function(t){this.visitElementOrTemplate("ng-template",t)},t.prototype.visitElementOrTemplate=function(t,e){var n=this,r=new j;r.setElement(t);var i=Ci(e);Object.getOwnPropertyNames(i).forEach(function(t){var e=i[t];(r.addAttribute(t,e),"class"===t.toLowerCase())&&e.trim().split(/\s+/g).forEach(function(t){return r.addClassName(t)})});var a=[];this.matcher.match(r,function(t,e){return a.push(e)}),a.length>0&&this.directives.set(e,a),e.references.forEach(function(t){var r=null;if(""===t.value.trim())r=a.find(function(t){return t.isComponent})||null;else if(null===(r=a.find(function(e){return e.exportAs===t.value})||null))throw new Error("Assertion error: failed to find directive with exportAs: "+t.value);null!==r?n.references.set(t,{directive:r,node:e}):n.references.set(t,e)}),Object(o.__spread)(e.attributes,e.inputs).forEach(function(t){var r=a.find(function(e){return e.inputs.hasOwnProperty(t.name)});void 0!==r?n.bindings.set(t,r):n.bindings.set(t,e)}),e.outputs.forEach(function(t){var r=a.find(function(e){return e.outputs.hasOwnProperty(t.name)});void 0!==r?n.bindings.set(t,r):n.bindings.set(t,e)}),e.children.forEach(function(t){return t.visit(n)})},t.prototype.visitContent=function(t){},t.prototype.visitVariable=function(t){},t.prototype.visitReference=function(t){},t.prototype.visitTextAttribute=function(t){},t.prototype.visitBoundAttribute=function(t){},t.prototype.visitBoundEvent=function(t){},t.prototype.visitBoundAttributeOrEvent=function(t){},t.prototype.visitText=function(t){},t.prototype.visitBoundText=function(t){},t.prototype.visitIcu=function(t){},t}(),wm=function(t){function e(e,n,r,i,o,a){var s=t.call(this)||this;return s.bindings=e,s.symbols=n,s.nestingLevel=r,s.scope=i,s.template=o,s.level=a,s.visitNode=function(t){return t.visit(s)},s}return Object(o.__extends)(e,t),e.apply=function(t,n){var r=new Map,i=new Map,o=new Map;return new e(r,i,o,n,t instanceof _c?t:null,0).ingest(t),{expressions:r,symbols:i,nestingLevel:o}},e.prototype.ingest=function(t){t instanceof _c?(t.inputs.forEach(this.visitNode),t.outputs.forEach(this.visitNode),t.variables.forEach(this.visitNode),t.children.forEach(this.visitNode),this.nestingLevel.set(t,this.level)):t.forEach(this.visitNode)},e.prototype.visitElement=function(t){t.inputs.forEach(this.visitNode),t.outputs.forEach(this.visitNode),t.children.forEach(this.visitNode)},e.prototype.visitTemplate=function(t){t.inputs.forEach(this.visitNode),t.outputs.forEach(this.visitNode),t.references.forEach(this.visitNode);var n=this.scope.getChildScope(t);new e(this.bindings,this.symbols,this.nestingLevel,n,t,this.level+1).ingest(t)},e.prototype.visitVariable=function(t){null!==this.template&&this.symbols.set(t,this.template)},e.prototype.visitReference=function(t){null!==this.template&&this.symbols.set(t,this.template)},e.prototype.visitText=function(t){},e.prototype.visitContent=function(t){},e.prototype.visitTextAttribute=function(t){},e.prototype.visitIcu=function(t){},e.prototype.visitBoundAttribute=function(t){t.value.visit(this)},e.prototype.visitBoundEvent=function(t){t.handler.visit(this)},e.prototype.visitBoundText=function(t){t.value.visit(this)},e.prototype.visitPropertyRead=function(e,n){return this.maybeMap(n,e,e.name),t.prototype.visitPropertyRead.call(this,e,n)},e.prototype.visitSafePropertyRead=function(e,n){return this.maybeMap(n,e,e.name),t.prototype.visitSafePropertyRead.call(this,e,n)},e.prototype.visitPropertyWrite=function(e,n){return this.maybeMap(n,e,e.name),t.prototype.visitPropertyWrite.call(this,e,n)},e.prototype.visitMethodCall=function(e,n){return this.maybeMap(n,e,e.name),t.prototype.visitMethodCall.call(this,e,n)},e.prototype.visitSafeMethodCall=function(e,n){return this.maybeMap(n,e,e.name),t.prototype.visitSafeMethodCall.call(this,e,n)},e.prototype.maybeMap=function(t,e,n){if(e.receiver instanceof yo){var r=this.scope.lookup(n);null!==r&&this.bindings.set(e,r)}},e}(Bo),jm=function(){function t(t,e,n,r,i,o,a){this.target=t,this.directives=e,this.bindings=n,this.references=r,this.exprTargets=i,this.symbols=o,this.nestingLevel=a}return t.prototype.getDirectivesOfNode=function(t){return this.directives.get(t)||null},t.prototype.getReferenceTarget=function(t){return this.references.get(t)||null},t.prototype.getConsumerOfBinding=function(t){return this.bindings.get(t)||null},t.prototype.getExpressionTarget=function(t){return this.exprTargets.get(t)||null},t.prototype.getTemplateOfSymbol=function(t){return this.symbols.get(t)||null},t.prototype.getNestingLevel=function(t){return this.nestingLevel.get(t)||0},t.prototype.getUsedDirectives=function(){var t=new Set;return this.directives.forEach(function(e){return e.forEach(function(e){return t.add(e)})}),Array.from(t.values())},t}();od(un)},"./node_modules/@angular/compiler/node_modules/tslib/tslib.es6.js":function(t,e,n){"use strict";n.r(e),n.d(e,"__extends",function(){return i}),n.d(e,"__assign",function(){return o}),n.d(e,"__rest",function(){return a}),n.d(e,"__decorate",function(){return s}),n.d(e,"__param",function(){return u}),n.d(e,"__metadata",function(){return c}),n.d(e,"__awaiter",function(){return l}),n.d(e,"__generator",function(){return d}),n.d(e,"__exportStar",function(){return p}),n.d(e,"__values",function(){return f}),n.d(e,"__read",function(){return h}),n.d(e,"__spread",function(){return m}),n.d(e,"__await",function(){return g}),n.d(e,"__asyncGenerator",function(){return y}),n.d(e,"__asyncDelegator",function(){return _}),n.d(e,"__asyncValues",function(){return v}),n.d(e,"__makeTemplateObject",function(){return b}),n.d(e,"__importStar",function(){return x}),n.d(e,"__importDefault",function(){return w});var r=function(t,e){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)};function i(t,e){function n(){this.constructor=t}r(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}var o=function(){return(o=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++){e=arguments[n];for(var i in e)Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i])}return t}).apply(this,arguments)};function a(t,e){var n={};for(var r in t)Object.prototype.hasOwnProperty.call(t,r)&&e.indexOf(r)<0&&(n[r]=t[r]);if(null!=t&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(t);i<r.length;i++)e.indexOf(r[i])<0&&(n[r[i]]=t[r[i]])}return n}function s(t,e,n,r){var i,o=arguments.length,a=o<3?e:null===r?r=Object.getOwnPropertyDescriptor(e,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(t,e,n,r);else for(var s=t.length-1;s>=0;s--)(i=t[s])&&(a=(o<3?i(a):o>3?i(e,n,a):i(e,n))||a);return o>3&&a&&Object.defineProperty(e,n,a),a}function u(t,e){return function(n,r){e(n,r,t)}}function c(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)}function l(t,e,n,r){return new(n||(n=Promise))(function(i,o){function a(t){try{u(r.next(t))}catch(t){o(t)}}function s(t){try{u(r.throw(t))}catch(t){o(t)}}function u(t){t.done?i(t.value):new n(function(e){e(t.value)}).then(a,s)}u((r=r.apply(t,e||[])).next())})}function d(t,e){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=e.call(t,a)}catch(t){o=[6,t],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(t,e){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}function f(t){var e="function"==typeof Symbol&&t[Symbol.iterator],n=0;return e?e.call(t):{next:function(){return t&&n>=t.length&&(t=void 0),{value:t&&t[n++],done:!t}}}}function h(t,e){var n="function"==typeof Symbol&&t[Symbol.iterator];if(!n)return t;var r,i,o=n.call(t),a=[];try{for(;(void 0===e||e-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(t){i={error:t}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var t=[],e=0;e<arguments.length;e++)t=t.concat(h(arguments[e]));return t}function g(t){return this instanceof g?(this.v=t,this):new g(t)}function y(t,e,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(t,e||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(t){i[t]&&(r[t]=function(e){return new Promise(function(n,r){o.push([t,e,n,r])>1||s(t,e)})})}function s(t,e){try{(n=i[t](e)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(t){l(o[0][3],t)}var n}function u(t){s("next",t)}function c(t){s("throw",t)}function l(t,e){t(e),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(t){var e,n;return e={},r("next"),r("throw",function(t){throw t}),r("return"),e[Symbol.iterator]=function(){return this},e;function r(r,i){e[r]=t[r]?function(e){return(n=!n)?{value:g(t[r](e)),done:"return"===r}:i?i(e):e}:i}}function v(t){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var e,n=t[Symbol.asyncIterator];return n?n.call(t):(t=f(t),e={},r("next"),r("throw"),r("return"),e[Symbol.asyncIterator]=function(){return this},e);function r(n){e[n]=t[n]&&function(e){return new Promise(function(r,i){var o,a,s,u;e=t[n](e),o=r,a=i,s=e.done,u=e.value,Promise.resolve(u).then(function(t){o({value:t,done:s})},a)})}}}function b(t,e){return Object.defineProperty?Object.defineProperty(t,"raw",{value:e}):t.raw=e,t}function x(t){if(t&&t.__esModule)return t;var e={};if(null!=t)for(var n in t)Object.hasOwnProperty.call(t,n)&&(e[n]=t[n]);return e.default=t,e}function w(t){return t&&t.__esModule?t:{default:t}}},"./node_modules/@angular/core/fesm5/core.js":function(t,e,n){"use strict";n.r(e),n.d(e,"ɵangular_packages_core_core_s",function(){return Im}),n.d(e,"ɵangular_packages_core_core_p",function(){return km}),n.d(e,"ɵangular_packages_core_core_q",function(){return Am}),n.d(e,"ɵangular_packages_core_core_r",function(){return Dm}),n.d(e,"ɵangular_packages_core_core_g",function(){return Af}),n.d(e,"ɵangular_packages_core_core_n",function(){return um}),n.d(e,"ɵangular_packages_core_core_o",function(){return mm}),n.d(e,"ɵangular_packages_core_core_m",function(){return Zh}),n.d(e,"ɵangular_packages_core_core_l",function(){return Qh}),n.d(e,"ɵangular_packages_core_core_c",function(){return Ge}),n.d(e,"ɵangular_packages_core_core_d",function(){return Cf}),n.d(e,"ɵangular_packages_core_core_e",function(){return ff}),n.d(e,"ɵangular_packages_core_core_f",function(){return vf}),n.d(e,"ɵangular_packages_core_core_k",function(){return Id}),n.d(e,"ɵangular_packages_core_core_t",function(){return oh}),n.d(e,"ɵangular_packages_core_core_v",function(){return eh}),n.d(e,"ɵangular_packages_core_core_u",function(){return th}),n.d(e,"ɵangular_packages_core_core_y",function(){return ih}),n.d(e,"ɵangular_packages_core_core_w",function(){return nh}),n.d(e,"ɵangular_packages_core_core_x",function(){return rh}),n.d(e,"ɵangular_packages_core_core_bb",function(){return Gn}),n.d(e,"ɵangular_packages_core_core_bi",function(){return l}),n.d(e,"ɵangular_packages_core_core_bc",function(){return gn}),n.d(e,"ɵangular_packages_core_core_bd",function(){return _n}),n.d(e,"ɵangular_packages_core_core_be",function(){return Rn}),n.d(e,"ɵangular_packages_core_core_bh",function(){return Oi}),n.d(e,"ɵangular_packages_core_core_bl",function(){return ce}),n.d(e,"ɵangular_packages_core_core_h",function(){return Fu}),n.d(e,"ɵangular_packages_core_core_i",function(){return Lu}),n.d(e,"ɵangular_packages_core_core_j",function(){return Bu}),n.d(e,"ɵangular_packages_core_core_a",function(){return E}),n.d(e,"ɵangular_packages_core_core_b",function(){return S}),n.d(e,"ɵangular_packages_core_core_bj",function(){return a}),n.d(e,"ɵangular_packages_core_core_z",function(){return Oy}),n.d(e,"ɵangular_packages_core_core_ba",function(){return Wm}),n.d(e,"createPlatform",function(){return Sh}),n.d(e,"assertPlatform",function(){return Th}),n.d(e,"destroyPlatform",function(){return kh}),n.d(e,"getPlatform",function(){return Ah}),n.d(e,"PlatformRef",function(){return Dh}),n.d(e,"ApplicationRef",function(){return Ph}),n.d(e,"createPlatformFactory",function(){return Mh}),n.d(e,"NgProbeToken",function(){return Eh}),n.d(e,"enableProdMode",function(){return hc}),n.d(e,"isDevMode",function(){return fc}),n.d(e,"APP_ID",function(){return kf}),n.d(e,"PACKAGE_ROOT_URL",function(){return Ff}),n.d(e,"PLATFORM_INITIALIZER",function(){return Pf}),n.d(e,"PLATFORM_ID",function(){return Rf}),n.d(e,"APP_BOOTSTRAP_LISTENER",function(){return Nf}),n.d(e,"APP_INITIALIZER",function(){return Mf}),n.d(e,"ApplicationInitStatus",function(){return Tf}),n.d(e,"DebugElement",function(){return sm}),n.d(e,"DebugNode",function(){return am}),n.d(e,"asNativeElements",function(){return $h}),n.d(e,"getDebugNode",function(){return im}),n.d(e,"Testability",function(){return _h}),n.d(e,"TestabilityRegistry",function(){return vh}),n.d(e,"setTestabilityGetter",function(){return bh}),n.d(e,"TRANSLATIONS",function(){return Mm}),n.d(e,"TRANSLATIONS_FORMAT",function(){return Tm}),n.d(e,"LOCALE_ID",function(){return Sm}),n.d(e,"MissingTranslationStrategy",function(){return vm}),n.d(e,"ApplicationModule",function(){return Pm}),n.d(e,"wtfCreateScope",function(){return sh}),n.d(e,"wtfLeave",function(){return uh}),n.d(e,"wtfStartTimeRange",function(){return ch}),n.d(e,"wtfEndTimeRange",function(){return lh}),n.d(e,"Type",function(){return Fd}),n.d(e,"EventEmitter",function(){return Yl}),n.d(e,"ErrorHandler",function(){return of}),n.d(e,"Sanitizer",function(){return ec}),n.d(e,"SecurityContext",function(){return Qu}),n.d(e,"ANALYZE_FOR_ENTRY_COMPONENTS",function(){return D}),n.d(e,"Attribute",function(){return I}),n.d(e,"ContentChild",function(){return N}),n.d(e,"ContentChildren",function(){return R}),n.d(e,"Query",function(){return P}),n.d(e,"ViewChild",function(){return L}),n.d(e,"ViewChildren",function(){return F}),n.d(e,"Component",function(){return jp}),n.d(e,"Directive",function(){return wp}),n.d(e,"HostBinding",function(){return Mp}),n.d(e,"HostListener",function(){return Tp}),n.d(e,"Input",function(){return Ep}),n.d(e,"Output",function(){return Sp}),n.d(e,"Pipe",function(){return Op}),n.d(e,"CUSTOM_ELEMENTS_SCHEMA",function(){return Bp}),n.d(e,"NO_ERRORS_SCHEMA",function(){return Vp}),n.d(e,"NgModule",function(){return zp}),n.d(e,"ViewEncapsulation",function(){return J}),n.d(e,"Version",function(){return nc}),n.d(e,"VERSION",function(){return rc}),n.d(e,"defineInjectable",function(){return g}),n.d(e,"defineInjector",function(){return y}),n.d(e,"forwardRef",function(){return Q}),n.d(e,"resolveForwardRef",function(){return Z}),n.d(e,"Injectable",function(){return Xp}),n.d(e,"INJECTOR",function(){return Ns}),n.d(e,"Injector",function(){return Ls}),n.d(e,"inject",function(){return Ye}),n.d(e,"InjectFlags",function(){return Pe}),n.d(e,"ReflectiveInjector",function(){return Of}),n.d(e,"createInjector",function(){return su}),n.d(e,"ResolvedReflectiveFactory",function(){return gf}),n.d(e,"ReflectiveKey",function(){return lf}),n.d(e,"InjectionToken",function(){return b}),n.d(e,"Inject",function(){return Ne}),n.d(e,"Optional",function(){return Fe}),n.d(e,"Self",function(){return Le}),n.d(e,"SkipSelf",function(){return Be}),n.d(e,"Host",function(){return Ve}),n.d(e,"NgZone",function(){return dh}),n.d(e,"ɵNoopNgZone",function(){return yh}),n.d(e,"RenderComponentType",function(){return qu}),n.d(e,"Renderer",function(){return Yu}),n.d(e,"Renderer2",function(){return $u}),n.d(e,"RendererFactory2",function(){return Xu}),n.d(e,"RendererStyleFlags2",function(){return Vu}),n.d(e,"RootRenderer",function(){return Ku}),n.d(e,"COMPILER_OPTIONS",function(){return $f}),n.d(e,"Compiler",function(){return Zf}),n.d(e,"CompilerFactory",function(){return Jf}),n.d(e,"ModuleWithComponentFactories",function(){return Bf}),n.d(e,"ComponentFactory",function(){return ju}),n.d(e,"ComponentRef",function(){return wu}),n.d(e,"ComponentFactoryResolver",function(){return ku}),n.d(e,"ElementRef",function(){return Hu}),n.d(e,"NgModuleFactory",function(){return Pu}),n.d(e,"NgModuleRef",function(){return Iu}),n.d(e,"NgModuleFactoryLoader",function(){return Td}),n.d(e,"getModuleFactory",function(){return Rd}),n.d(e,"QueryList",function(){return Nh}),n.d(e,"SystemJsNgModuleLoader",function(){return Bh}),n.d(e,"SystemJsNgModuleLoaderConfig",function(){return Fh}),n.d(e,"TemplateRef",function(){return Kl}),n.d(e,"ViewContainerRef",function(){return zh}),n.d(e,"EmbeddedViewRef",function(){return Kh}),n.d(e,"ViewRef",function(){return Yh}),n.d(e,"ChangeDetectionStrategy",function(){return M}),n.d(e,"ChangeDetectorRef",function(){return Wh}),n.d(e,"DefaultIterableDiffer",function(){return lm}),n.d(e,"IterableDiffers",function(){return _m}),n.d(e,"KeyValueDiffers",function(){return xm}),n.d(e,"SimpleChange",function(){return kr}),n.d(e,"WrappedValue",function(){return Tr}),n.d(e,"platformCore",function(){return Em}),n.d(e,"ɵALLOW_MULTIPLE_PLATFORMS",function(){return Ch}),n.d(e,"ɵAPP_ID_RANDOM_PROVIDER",function(){return Df}),n.d(e,"ɵdefaultIterableDiffers",function(){return Om}),n.d(e,"ɵdefaultKeyValueDiffers",function(){return Cm}),n.d(e,"ɵdevModeEqual",function(){return Mr}),n.d(e,"ɵisListLikeIterable",function(){return Ar}),n.d(e,"ɵChangeDetectorStatus",function(){return k}),n.d(e,"ɵisDefaultChangeDetectionStrategy",function(){return B}),n.d(e,"ɵConsole",function(){return Lf}),n.d(e,"ɵgetInjectableDef",function(){return _}),n.d(e,"ɵinject",function(){return Ye}),n.d(e,"ɵsetCurrentInjector",function(){return We}),n.d(e,"ɵAPP_ROOT",function(){return eu}),n.d(e,"ɵivyEnabled",function(){return Nm}),n.d(e,"ɵComponentFactory",function(){return ju}),n.d(e,"ɵCodegenComponentFactoryResolver",function(){return Au}),n.d(e,"ɵresolveComponentResources",function(){return $}),n.d(e,"ɵReflectionCapabilities",function(){return Hd}),n.d(e,"ɵRenderDebugInfo",function(){return Gu}),n.d(e,"ɵ_sanitizeHtml",function(){return Lc}),n.d(e,"ɵ_sanitizeStyle",function(){return wd}),n.d(e,"ɵ_sanitizeUrl",function(){return _c}),n.d(e,"ɵglobal",function(){return H}),n.d(e,"ɵlooseIdentical",function(){return Y}),n.d(e,"ɵstringify",function(){return K}),n.d(e,"ɵmakeDecorator",function(){return O}),n.d(e,"ɵisObservable",function(){return Sf}),n.d(e,"ɵisPromise",function(){return Ef}),n.d(e,"ɵclearOverrides",function(){return av}),n.d(e,"ɵinitServicesIfNeeded",function(){return b_}),n.d(e,"ɵoverrideComponentView",function(){return ov}),n.d(e,"ɵoverrideProvider",function(){return iv}),n.d(e,"ɵNOT_FOUND_CHECK_ONLY_ELEMENT_INJECTOR",function(){return Dy}),n.d(e,"ɵdefineBase",function(){return ft}),n.d(e,"ɵdefineComponent",function(){return ut}),n.d(e,"ɵdefineDirective",function(){return ht}),n.d(e,"ɵdefinePipe",function(){return mt}),n.d(e,"ɵdefineNgModule",function(){return dt}),n.d(e,"ɵdetectChanges",function(){return Va}),n.d(e,"ɵrenderComponent",function(){return bs}),n.d(e,"ɵRender3ComponentFactory",function(){return cc}),n.d(e,"ɵRender3ComponentRef",function(){return lc}),n.d(e,"ɵdirectiveInject",function(){return as}),n.d(e,"ɵinjectAttribute",function(){return ss}),n.d(e,"ɵgetFactoryOf",function(){return er}),n.d(e,"ɵgetInheritedFactory",function(){return nr}),n.d(e,"ɵtemplateRefExtractor",function(){return dd}),n.d(e,"ɵProvidersFeature",function(){return xu}),n.d(e,"ɵInheritDefinitionFeature",function(){return Ss}),n.d(e,"ɵNgOnChangesFeature",function(){return ks}),n.d(e,"ɵLifecycleHooksFeature",function(){return Os}),n.d(e,"ɵRender3NgModuleRef",function(){return El}),n.d(e,"ɵmarkDirty",function(){return Wa}),n.d(e,"ɵNgModuleFactory",function(){return Sl}),n.d(e,"ɵNO_CHANGE",function(){return Pr}),n.d(e,"ɵcontainer",function(){return Ea}),n.d(e,"ɵnextContext",function(){return Po}),n.d(e,"ɵelementStart",function(){return Wo}),n.d(e,"ɵnamespaceHTML",function(){return Vo}),n.d(e,"ɵnamespaceMathML",function(){return Bo}),n.d(e,"ɵnamespaceSVG",function(){return Lo}),n.d(e,"ɵelement",function(){return zo}),n.d(e,"ɵlistener",function(){return $o}),n.d(e,"ɵtext",function(){return ha}),n.d(e,"ɵembeddedViewStart",function(){return ka}),n.d(e,"ɵquery",function(){return cd}),n.d(e,"ɵregisterContentQuery",function(){return us}),n.d(e,"ɵprojection",function(){return Ra}),n.d(e,"ɵbind",function(){return qa}),n.d(e,"ɵinterpolation1",function(){return Ka}),n.d(e,"ɵinterpolation2",function(){return Xa}),n.d(e,"ɵinterpolation3",function(){return Qa}),n.d(e,"ɵinterpolation4",function(){return Za}),n.d(e,"ɵinterpolation5",function(){return $a}),n.d(e,"ɵinterpolation6",function(){return Ja}),n.d(e,"ɵinterpolation7",function(){return ts}),n.d(e,"ɵinterpolation8",function(){return es}),n.d(e,"ɵinterpolationV",function(){return Ya}),n.d(e,"ɵpipeBind1",function(){return Vl}),n.d(e,"ɵpipeBind2",function(){return zl}),n.d(e,"ɵpipeBind3",function(){return Hl}),n.d(e,"ɵpipeBind4",function(){return Ul}),n.d(e,"ɵpipeBindV",function(){return Wl}),n.d(e,"ɵpureFunction0",function(){return Tl}),n.d(e,"ɵpureFunction1",function(){return kl}),n.d(e,"ɵpureFunction2",function(){return Al}),n.d(e,"ɵpureFunction3",function(){return Dl}),n.d(e,"ɵpureFunction4",function(){return Il}),n.d(e,"ɵpureFunction5",function(){return Pl}),n.d(e,"ɵpureFunction6",function(){return Rl}),n.d(e,"ɵpureFunction7",function(){return Nl}),n.d(e,"ɵpureFunction8",function(){return Fl}),n.d(e,"ɵpureFunctionV",function(){return Ll}),n.d(e,"ɵgetCurrentView",function(){return ds}),n.d(e,"ɵgetHostElement",function(){return br}),n.d(e,"ɵrestoreView",function(){return yn}),n.d(e,"ɵcontainerRefreshStart",function(){return Ma}),n.d(e,"ɵcontainerRefreshEnd",function(){return Ta}),n.d(e,"ɵqueryRefresh",function(){return ld}),n.d(e,"ɵloadQueryList",function(){return is}),n.d(e,"ɵelementEnd",function(){return Jo}),n.d(e,"ɵelementProperty",function(){return ea}),n.d(e,"ɵcomponentHostSyntheticProperty",function(){return na}),n.d(e,"ɵprojectionDef",function(){return Ia}),n.d(e,"ɵreference",function(){return rs}),n.d(e,"ɵenableBindings",function(){return hn}),n.d(e,"ɵdisableBindings",function(){return mn}),n.d(e,"ɵallocHostVars",function(){return Ga}),n.d(e,"ɵelementAttribute",function(){return ta}),n.d(e,"ɵelementContainerStart",function(){return Ho}),n.d(e,"ɵelementContainerEnd",function(){return Uo}),n.d(e,"ɵelementStyling",function(){return ua}),n.d(e,"ɵelementHostAttrs",function(){return ca}),n.d(e,"ɵelementStylingMap",function(){return fa}),n.d(e,"ɵelementStyleProp",function(){return da}),n.d(e,"ɵelementStylingApply",function(){return la}),n.d(e,"ɵelementClassProp",function(){return pa}),n.d(e,"ɵtextBinding",function(){return ma}),n.d(e,"ɵtemplate",function(){return Ca}),n.d(e,"ɵembeddedViewEnd",function(){return Aa}),n.d(e,"ɵstore",function(){return ns}),n.d(e,"ɵload",function(){return os}),n.d(e,"ɵpipe",function(){return Bl}),n.d(e,"ɵwhenRendered",function(){return Cs}),n.d(e,"ɵi18n",function(){return pl}),n.d(e,"ɵi18nAttributes",function(){return fl}),n.d(e,"ɵi18nExp",function(){return _l}),n.d(e,"ɵi18nStart",function(){return al}),n.d(e,"ɵi18nEnd",function(){return cl}),n.d(e,"ɵi18nApply",function(){return vl}),n.d(e,"ɵi18nPostprocess",function(){return ul}),n.d(e,"ɵsetClassMetadata",function(){return Ml}),n.d(e,"ɵcompileComponent",function(){return hp}),n.d(e,"ɵcompileDirective",function(){return mp}),n.d(e,"ɵcompileNgModule",function(){return Jd}),n.d(e,"ɵcompileNgModuleDefs",function(){return tp}),n.d(e,"ɵpatchComponentDefWithScope",function(){return cp}),n.d(e,"ɵresetCompiledComponents",function(){return ap}),n.d(e,"ɵcompilePipe",function(){return xp}),n.d(e,"ɵsanitizeHtml",function(){return jd}),n.d(e,"ɵsanitizeStyle",function(){return Od}),n.d(e,"ɵsanitizeUrl",function(){return Cd}),n.d(e,"ɵsanitizeResourceUrl",function(){return Ed}),n.d(e,"ɵbypassSanitizationTrustHtml",function(){return hd}),n.d(e,"ɵbypassSanitizationTrustStyle",function(){return md}),n.d(e,"ɵbypassSanitizationTrustScript",function(){return gd}),n.d(e,"ɵbypassSanitizationTrustUrl",function(){return yd}),n.d(e,"ɵbypassSanitizationTrustResourceUrl",function(){return _d}),n.d(e,"ɵgetLContext",function(){return rr}),n.d(e,"ɵbindPlayerFactory",function(){return ji}),n.d(e,"ɵaddPlayer",function(){return hs}),n.d(e,"ɵgetPlayers",function(){return ms}),n.d(e,"ɵcompileNgModuleFactory__POST_R3__",function(){return Oh}),n.d(e,"ɵSWITCH_COMPILE_COMPONENT__POST_R3__",function(){return kp}),n.d(e,"ɵSWITCH_COMPILE_DIRECTIVE__POST_R3__",function(){return Ap}),n.d(e,"ɵSWITCH_COMPILE_PIPE__POST_R3__",function(){return Dp}),n.d(e,"ɵSWITCH_COMPILE_NGMODULE__POST_R3__",function(){return Hp}),n.d(e,"ɵgetDebugNode__POST_R3__",function(){return rm}),n.d(e,"ɵSWITCH_COMPILE_INJECTABLE__POST_R3__",function(){return Qp}),n.d(e,"ɵSWITCH_IVY_ENABLED__POST_R3__",function(){return Rm}),n.d(e,"ɵSWITCH_CHANGE_DETECTOR_REF_FACTORY__POST_R3__",function(){return qh}),n.d(e,"ɵCompiler_compileModuleSync__POST_R3__",function(){return Uf}),n.d(e,"ɵCompiler_compileModuleAsync__POST_R3__",function(){return qf}),n.d(e,"ɵCompiler_compileModuleAndAllComponentsSync__POST_R3__",function(){return Yf}),n.d(e,"ɵCompiler_compileModuleAndAllComponentsAsync__POST_R3__",function(){return Xf}),n.d(e,"ɵSWITCH_ELEMENT_REF_FACTORY__POST_R3__",function(){return Uu}),n.d(e,"ɵSWITCH_TEMPLATE_REF_FACTORY__POST_R3__",function(){return Xl}),n.d(e,"ɵSWITCH_VIEW_CONTAINER_REF_FACTORY__POST_R3__",function(){return Hh}),n.d(e,"ɵSWITCH_RENDERER2_FACTORY__POST_R3__",function(){return Ju}),n.d(e,"ɵgetModuleFactory__POST_R3__",function(){return Pd}),n.d(e,"ɵpublishGlobalUtil",function(){return vs}),n.d(e,"ɵpublishDefaultGlobalUtils",function(){return _s}),n.d(e,"ɵSWITCH_INJECTOR_FACTORY__POST_R3__",function(){return Bs}),n.d(e,"ɵregisterModuleFactory",function(){return Ad}),n.d(e,"ɵEMPTY_ARRAY",function(){return Ag}),n.d(e,"ɵEMPTY_MAP",function(){return Dg}),n.d(e,"ɵand",function(){return Ig}),n.d(e,"ɵccf",function(){return ty}),n.d(e,"ɵcmf",function(){return sv}),n.d(e,"ɵcrt",function(){return eg}),n.d(e,"ɵdid",function(){return xy}),n.d(e,"ɵeld",function(){return Pg}),n.d(e,"ɵelementEventFullName",function(){return pg}),n.d(e,"ɵgetComponentViewDefinitionFactory",function(){return ey}),n.d(e,"ɵinlineInterpolate",function(){return Tg}),n.d(e,"ɵinterpolate",function(){return Mg}),n.d(e,"ɵmod",function(){return Wg}),n.d(e,"ɵmpd",function(){return Ug}),n.d(e,"ɵncd",function(){return Wy}),n.d(e,"ɵnov",function(){return cy}),n.d(e,"ɵpid",function(){return wy}),n.d(e,"ɵprd",function(){return jy}),n.d(e,"ɵpad",function(){return Gy}),n.d(e,"ɵpod",function(){return Yy}),n.d(e,"ɵppd",function(){return qy}),n.d(e,"ɵqud",function(){return By}),n.d(e,"ɵted",function(){return Xy}),n.d(e,"ɵunv",function(){return $m}),n.d(e,"ɵvid",function(){return $y});var r=n("./node_modules/@angular/core/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/index.js"),o=n("./node_modules/rxjs/_esm5/operators/index.js");function a(t){for(var e in t)if(t[e]===a)return e;throw Error("Could not find renamed property on target object.")}function s(t,e){for(var n in e)e.hasOwnProperty(n)&&!t.hasOwnProperty(n)&&(t[n]=e[n])}var u=a({ngComponentDef:a}),c=a({ngDirectiveDef:a}),l=a({ngInjectableDef:a}),d=a({ngInjectorDef:a}),p=a({ngPipeDef:a}),f=a({ngModuleDef:a}),h=a({ngBaseDef:a}),m=a({__NG_ELEMENT_ID__:a});function g(t){return{providedIn:t.providedIn||null,factory:t.factory,value:void 0}}function y(t){return{factory:t.factory,providers:t.providers||[],imports:t.imports||[]}}function _(t){return t&&t.hasOwnProperty(l)?t[l]:null}function v(t){return t&&t.hasOwnProperty(d)?t[d]:null}var b=function(){function t(t,e){this._desc=t,this.ngMetadataName="InjectionToken",this.ngInjectableDef=void 0!==e?g({providedIn:e.providedIn||"root",factory:e.factory}):void 0}return t.prototype.toString=function(){return"InjectionToken "+this._desc},t}(),x="__annotations__",w="__parameters__",j="__prop__metadata__";function O(t,e,n,i,o){var a=C(e);function s(){for(var t,e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];if(this instanceof s)return a.call.apply(a,Object(r.__spread)([this],e)),this;var u=new((t=s).bind.apply(t,Object(r.__spread)([void 0],e)));return function(t){return o&&o.apply(void 0,Object(r.__spread)([t],e)),(t.hasOwnProperty(x)?t[x]:Object.defineProperty(t,x,{value:[]})[x]).push(u),i&&i(t),t}}return n&&(s.prototype=Object.create(n.prototype)),s.prototype.ngMetadataName=t,s.annotationCls=s,s}function C(t){return function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];if(t){var i=t.apply(void 0,Object(r.__spread)(e));for(var o in i)this[o]=i[o]}}}function E(t,e,n){var i=C(e);function o(){for(var t,e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];if(this instanceof o)return i.apply(this,e),this;var a=new((t=o).bind.apply(t,Object(r.__spread)([void 0],e)));return s.annotation=a,s;function s(t,e,n){for(var r=t.hasOwnProperty(w)?t[w]:Object.defineProperty(t,w,{value:[]})[w];r.length<=n;)r.push(null);return(r[n]=r[n]||[]).push(a),t}}return n&&(o.prototype=Object.create(n.prototype)),o.prototype.ngMetadataName=t,o.annotationCls=o,o}function S(t,e,n,i){var o=C(e);function a(){for(var t,e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];if(this instanceof a)return o.apply(this,e),this;var s=new((t=a).bind.apply(t,Object(r.__spread)([void 0],e)));return function(t,n){var o=t.constructor,a=o.hasOwnProperty(j)?o[j]:Object.defineProperty(o,j,{value:{}})[j];a[n]=a.hasOwnProperty(n)&&a[n]||[],a[n].unshift(s),i&&i.apply(void 0,Object(r.__spread)([t,n],e))}}return n&&(a.prototype=Object.create(n.prototype)),a.prototype.ngMetadataName=t,a.annotationCls=a,a}var M,T,k,A,D=new b("AnalyzeForEntryComponents"),I=E("Attribute",function(t){return{attributeName:t}}),P=function(){return function(){}}(),R=S("ContentChildren",function(t,e){return void 0===e&&(e={}),Object(r.__assign)({selector:t,first:!1,isViewQuery:!1,descendants:!1},e)},P),N=S("ContentChild",function(t,e){return void 0===e&&(e={}),Object(r.__assign)({selector:t,first:!0,isViewQuery:!1,descendants:!0},e)},P),F=S("ViewChildren",function(t,e){return void 0===e&&(e={}),Object(r.__assign)({selector:t,first:!1,isViewQuery:!0,descendants:!0},e)},P),L=S("ViewChild",function(t,e){return Object(r.__assign)({selector:t,first:!0,isViewQuery:!0,descendants:!0},e)},P);function B(t){return null==t||t===M.Default}(T=M||(M={}))[T.OnPush=0]="OnPush",T[T.Default=1]="Default",(A=k||(k={}))[A.CheckOnce=0]="CheckOnce",A[A.Checked=1]="Checked",A[A.CheckAlways=2]="CheckAlways",A[A.Detached=3]="Detached",A[A.Errored=4]="Errored",A[A.Destroyed=5]="Destroyed";var V="undefined"!=typeof window&&window,z="undefined"!=typeof self&&"undefined"!=typeof WorkerGlobalScope&&self instanceof WorkerGlobalScope&&self,H="undefined"!=typeof global&&global||V||z,U=Promise.resolve(0),W=null;function q(){if(!W){var t=H.Symbol;if(t&&t.iterator)W=t.iterator;else for(var e=Object.getOwnPropertyNames(Map.prototype),n=0;n<e.length;++n){var r=e[n];"entries"!==r&&"size"!==r&&Map.prototype[r]===Map.prototype.entries&&(W=r)}}return W}function G(t){"undefined"==typeof Zone?U.then(function(){t&&t.apply(null,null)}):Zone.current.scheduleMicroTask("scheduleMicrotask",t)}function Y(t,e){return t===e||"number"==typeof t&&"number"==typeof e&&isNaN(t)&&isNaN(e)}function K(t){if("string"==typeof t)return t;if(t instanceof Array)return"["+t.map(K).join(", ")+"]";if(null==t)return""+t;if(t.overriddenName)return""+t.overriddenName;if(t.name)return""+t.name;var e=t.toString();if(null==e)return""+e;var n=e.indexOf("\n");return-1===n?e:e.substring(0,n)}var X=a({__forward_ref__:a});function Q(t){return t.__forward_ref__=Q,t.toString=function(){return K(this())},t}function Z(t){var e=t;return"function"==typeof e&&e.hasOwnProperty(X)&&e.__forward_ref__===Q?e():t}function $(t){var e=[],n=new Map;function r(r){var i=n.get(r);if(!i){var o=t(r);n.set(r,i=o.then(it)),e.push(i)}return i}return nt.forEach(function(t){t.templateUrl&&r(t.templateUrl).then(function(e){t.template=e,t.templateUrl=void 0});var e=t.styleUrls,n=t.styles||(t.styles=[]),i=t.styles.length;e&&e.forEach(function(o,a){n.push(""),r(o).then(function(r){n[i+a]=r,e.splice(e.indexOf(o),1),0==e.length&&(t.styleUrls=void 0)})})}),nt.clear(),Promise.all(e).then(function(){return null})}var J,tt,et,nt=new Set;function rt(t){return t.templateUrl||t.styleUrls&&t.styleUrls.length}function it(t){return"string"==typeof t?t:t.text()}(tt=J||(J={}))[tt.Emulated=0]="Emulated",tt[tt.Native=1]="Native",tt[tt.None=2]="None",tt[tt.ShadowDom=3]="ShadowDom",("undefined"==typeof ngDevMode||ngDevMode)&&(et={firstTemplatePass:0,tNode:0,tView:0,rendererCreateTextNode:0,rendererSetText:0,rendererCreateElement:0,rendererAddEventListener:0,rendererSetAttribute:0,rendererRemoveAttribute:0,rendererSetProperty:0,rendererSetClassName:0,rendererAddClass:0,rendererRemoveClass:0,rendererSetStyle:0,rendererRemoveStyle:0,rendererDestroy:0,rendererDestroyNode:0,rendererMoveNode:0,rendererRemoveNode:0,rendererCreateComment:0},"undefined"!=typeof window&&(window.ngDevMode=et),"undefined"!=typeof global&&(global.ngDevMode=et),"undefined"!=typeof self&&(self.ngDevMode=et));var ot={},at=[];"undefined"!=typeof ngDevMode&&ngDevMode&&(Object.freeze(ot),Object.freeze(at));var st=0;function ut(t){var e=t.type,n=e.prototype,r={},i={type:e,providersResolver:null,consts:t.consts,vars:t.vars,factory:t.factory,template:t.template||null,hostBindings:t.hostBindings||null,contentQueries:t.contentQueries||null,contentQueriesRefresh:t.contentQueriesRefresh||null,attributes:t.attributes||null,declaredInputs:r,inputs:null,outputs:null,exportAs:t.exportAs||null,onInit:n.ngOnInit||null,doCheck:n.ngDoCheck||null,afterContentInit:n.ngAfterContentInit||null,afterContentChecked:n.ngAfterContentChecked||null,afterViewInit:n.ngAfterViewInit||null,afterViewChecked:n.ngAfterViewChecked||null,onDestroy:n.ngOnDestroy||null,onPush:t.changeDetection===M.OnPush,directiveDefs:null,pipeDefs:null,selectors:t.selectors,viewQuery:t.viewQuery||null,features:t.features||null,data:t.data||{},encapsulation:t.encapsulation||J.Emulated,id:"c",styles:t.styles||at,_:null};return i._=""+{toString:function(){var e=t.directives,n=t.features,o=t.pipes;i.id+=st++,i.inputs=pt(t.inputs,r),i.outputs=pt(t.outputs),n&&n.forEach(function(t){return t(i)}),i.directiveDefs=e?function(){return("function"==typeof e?e():e).map(ct)}:null,i.pipeDefs=o?function(){return("function"==typeof o?o():o).map(lt)}:null}},i}function ct(t){var e=gt(t)||yt(t);if(ngDevMode&&!e)throw new Error("'"+t.name+"' is neither 'ComponentType' or 'DirectiveType'.");return e}function lt(t){var e=_t(t);if(ngDevMode&&!e)throw new Error("'"+t.name+"' is not a 'PipeType'.");return e}function dt(t){return{type:t.type,bootstrap:t.bootstrap||at,declarations:t.declarations||at,imports:t.imports||at,exports:t.exports||at,transitiveCompileScopes:null}}function pt(t,e){if(null==t)return ot;var n={};for(var r in t)if(t.hasOwnProperty(r)){var i=t[r],o=i;Array.isArray(i)&&(o=i[1],i=i[0]),n[i]=r,e&&(e[i]=o)}return n}function ft(t){var e={};return{inputs:pt(t.inputs,e),declaredInputs:e,outputs:pt(t.outputs)}}var ht=ut;function mt(t){return{name:t.name,factory:t.factory,pure:!1!==t.pure,onDestroy:t.type.prototype.ngOnDestroy||null}}function gt(t){return t[u]||null}function yt(t){return t[c]||null}function _t(t){return t[p]||null}function vt(t,e){var n=t[f]||null;if(!n&&!0===e)throw new Error("Type "+K(t)+" does not have 'ngModuleDef' property.");return n}function bt(t,e,n){t!=e&&Et(n)}function xt(t,e,n){t==e&&Et(n)}function wt(t,e,n){t>=e&&Et(n)}function jt(t,e,n){t<=e&&Et(n)}function Ot(t,e){null==t&&Et(e)}function Ct(t,e){void 0===e&&(e="Type passed in is not ComponentType, it does not have 'ngComponentDef' property."),gt(t)||Et(e)}function Et(t){throw new Error("ASSERTION ERROR: "+t)}function St(t){bt(t instanceof Node,!0,"The provided value must be an instance of a DOM Node")}function Mt(t){bt(t,!0,"previousOrParentTNode should be a parent")}function Tt(t){Ot(t.parent,"previousOrParentTNode should have a parent")}function kt(t,e){wt(e,t?t.length:0,"index expected to be a valid data index")}var At=0,Dt=1,It=2,Pt=3,Rt=4,Nt=5,Ft=6,Lt=7,Bt=8,Vt=9,zt=10,Ht=11,Ut=12,Wt=13,qt=14,Gt=15,Yt=16,Kt=17,Xt=18,Qt=0,Zt=1,$t=6,Jt=7,te=8,ee="__ngContext__",ne=8,re=8,ie=9,oe=-1,ae=function(){return function(t,e,n){this.factory=t,this.resolving=!1,this.canSeeViewProviders=e,this.injectImpl=n}}(),se=ae.prototype;function ue(t){return"function"==typeof t?t.name||t:"string"==typeof t?t:null==t?"":"object"==typeof t&&"function"==typeof t.type?t.type.name||t.type:""+t}function ce(t,e){return ngDevMode&&kt(t,e+Xt),t[e+Xt]}function le(t){for(;Array.isArray(t);)t=t[Nt];return t}function de(t,e){return le(e[t+Xt])}function pe(t,e){return le(e[t.index])}function fe(t,e){return ngDevMode&&jt(t,-1,"wrong index for TNode"),ngDevMode&&wt(t,e[At].data.length,"wrong index for TNode"),e[At].data[t+Xt]}function he(t,e){var n=e[t];return n.length>=Xt?n:n[Nt]}function me(t){return 0!=(4&t.flags)}function ge(t){return 1==(1&t.flags)}function ye(t){return null!==t.template}function _e(t){return Array.isArray(t)&&t.length===te}function ve(t){return 0!=(128&t[Dt])}function be(t){ngDevMode&&Ot(t,"component");for(var e=Array.isArray(t)?t:je(t);e&&!(128&e[Dt]);)e=e[It];return e}function xe(t){var e=be(t);return ngDevMode&&Ot(e[Vt],"RootView has no context. Perhaps it is disconnected?"),e[Vt]}function we(t){return ngDevMode&&Ot(t,"Target expected"),t[ee]}function je(t){var e=we(t);return e?Array.isArray(e)?e:e.lView:null}function Oe(t){return t!==oe}function Ce(t){return 32767&t}function Ee(t){return t>>16}function Se(t,e){for(var n=Ee(t),r=e;n>0;)r=r[Kt],n--;return r}var Me,Te,ke=("undefined"!=typeof requestAnimationFrame&&requestAnimationFrame||setTimeout).bind(H);function Ae(t,e){for(var n=0;n<t.length;n++)e.push(t[n])}function De(t){for(var e=t[Ft];e&&2===e.type;)ngDevMode&&Ot(t[Kt],"lView[DECLARATION_VIEW]"),e=(t=t[Kt])[Ft];return t}function Ie(){var t=H.ng;if(!t||!t.ɵcompilerFacade)throw new Error("Angular JIT compilation failed: '@angular/compiler' not loaded!\n  - JIT compilation is discouraged for production use-cases! Consider AOT mode instead.\n  - Did you bootstrap using '@angular/platform-browser-dynamic' or '@angular/platform-server'?\n  - Alternatively provide the compiler with 'import \"@angular/compiler\";' before bootstrapping.");return t.ɵcompilerFacade}(Te=Me||(Me={}))[Te.Token=0]="Token",Te[Te.Attribute=1]="Attribute";var Pe,Re,Ne=E("Inject",function(t){return{token:t}}),Fe=E("Optional"),Le=E("Self"),Be=E("SkipSelf"),Ve=E("Host");(Re=Pe||(Pe={}))[Re.Default=0]="Default",Re[Re.Host=1]="Host",Re[Re.Self=2]="Self",Re[Re.SkipSelf=4]="SkipSelf",Re[Re.Optional=8]="Optional";var ze,He,Ue=void 0;function We(t){var e=Ue;return Ue=t,e}function qe(t){var e=ze;return ze=t,e}function Ge(t,e){if(void 0===e&&(e=Pe.Default),void 0===Ue)throw new Error("inject() must be called from an injection context");return null===Ue?Ke(t,void 0,e):Ue.get(t,e&Pe.Optional?null:void 0,e)}function Ye(t,e){return void 0===e&&(e=Pe.Default),(ze||Ge)(t,e)}function Ke(t,e,n){var r=_(t);if(r&&"root"==r.providedIn)return void 0===r.value?r.value=r.factory():r.value;if(n&Pe.Optional)return null;if(void 0!==e)return e;throw new Error("Injector: NOT_FOUND ["+K(t)+"]")}function Xe(t){for(var e=[],n=0;n<t.length;n++){var r=t[n];if(Array.isArray(r)){if(0===r.length)throw new Error("Arguments array must have arguments.");for(var i=void 0,o=Pe.Default,a=0;a<r.length;a++){var s=r[a];s instanceof Fe||"Optional"===s.ngMetadataName?o|=Pe.Optional:s instanceof Be||"SkipSelf"===s.ngMetadataName?o|=Pe.SkipSelf:s instanceof Le||"Self"===s.ngMetadataName?o|=Pe.Self:i=s instanceof Ne?s.token:s}e.push(Ye(i,o))}else e.push(Ye(r))}return e}function Qe(t,e){Ot(t,"should be called with a TNode"),bt(t.type,e,"should be a "+$e(e))}function Ze(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];Ot(t,"should be called with a TNode"),bt(e.some(function(e){return t.type===e}),!0,"Should be one of "+e.map($e).join(", ")+" but got "+$e(t.type))}function $e(t){return 1==t?"Projection":0==t?"Container":2==t?"View":3==t?"Element":4==t?"ElementContainer":"<unknown>"}function Je(t,e,n,r){ngDevMode&&bt(r.firstTemplatePass,!0,"Should only be called on first template pass"),e&&(r.initHooks||(r.initHooks=[])).push(t,e),n&&((r.initHooks||(r.initHooks=[])).push(t,n),(r.checkHooks||(r.checkHooks=[])).push(t,n))}function tn(t,e){if(t.firstTemplatePass)for(var n=e.directiveStart,r=e.directiveEnd;n<r;n++){var i=t.data[n];en(i,t,n),nn(i,t,n),rn(i,t,n)}}function en(t,e,n){t.afterContentInit&&(e.contentHooks||(e.contentHooks=[])).push(n,t.afterContentInit),t.afterContentChecked&&((e.contentHooks||(e.contentHooks=[])).push(n,t.afterContentChecked),(e.contentCheckHooks||(e.contentCheckHooks=[])).push(n,t.afterContentChecked))}function nn(t,e,n){t.afterViewInit&&(e.viewHooks||(e.viewHooks=[])).push(n,t.afterViewInit),t.afterViewChecked&&((e.viewHooks||(e.viewHooks=[])).push(n,t.afterViewChecked),(e.viewCheckHooks||(e.viewCheckHooks=[])).push(n,t.afterViewChecked))}function rn(t,e,n){null!=t.onDestroy&&(e.destroyHooks||(e.destroyHooks=[])).push(n,t.onDestroy)}function on(t,e,n){!n&&32&t[Dt]&&(an(t,e.initHooks,e.checkHooks,n),t[Dt]&=-33)}function an(t,e,n,r){if(!r){var i=2&t[Dt]?e:n;i&&sn(t,i)}}function sn(t,e){for(var n=0;n<e.length;n+=2)e[n+1].call(t[e[n]])}var un,cn,ln,dn,pn=null;function fn(t){pn=t}function hn(){un=!0}function mn(){un=!1}function gn(){return dn}function yn(t){On=t}function _n(){return cn}function vn(t){cn=t}function bn(t,e){cn=t,dn=e}function xn(){return ln}function wn(t){ln=t}function jn(t){return void 0===t&&(t=dn),1==(1&t[Dt])}var On=null;var Cn=!1;function En(){return Cn}function Sn(t){Cn=t}var Mn=!0;function Tn(){return Mn}function kn(t){Mn=t}var An=-1;function Dn(){return An}function In(t){An=t}function Pn(t,e){var n=dn;if(t){var r=t[At];Mn=r.firstTemplatePass,An=r.bindingStartIndex}return cn=e,ln=!0,dn=On=t,n}function Rn(t){return void 0===t&&(t=1),(On=function(t,e){for(;t>0;)ngDevMode&&Ot(e[Kt],"Declaration view should be defined if nesting level is greater than 0."),e=e[Kt],t--;return e}(t,On))[Vt]}function Nn(t){var e=dn[At];jn(dn)?dn[Dt]&=-2:(an(dn,e.viewHooks,e.viewCheckHooks,Cn),dn[Dt]&=-11,dn[Dt]|=32,dn[Lt]=e.bindingStartIndex),Pn(t,null)}var Fn=!0;function Ln(t){var e=Fn;return Fn=t,e}var Bn=255,Vn=0;function zn(t,e){var n=Un(t,e);if(-1!==n)return n;var r=e[At];r.firstTemplatePass&&(t.injectorIndex=e.length,Hn(r.data,t),Hn(e,null),Hn(r.blueprint,null),ngDevMode&&bt(0===t.flags||1===t.flags,!0,"expected tNode.flags to not be initialized"));var i=Wn(t,e),o=Ce(i),a=Se(i,e),s=t.injectorIndex;if(Oe(i))for(var u=a[At].data,c=0;c<8;c++)e[s+c]=a[o+c]|u[o+c];return e[s+re]=i,s}function Hn(t,e){t.push(0,0,0,0,0,0,0,0,e)}function Un(t,e){return-1===t.injectorIndex||t.parent&&t.parent.injectorIndex===t.injectorIndex||null==e[t.injectorIndex+re]?-1:t.injectorIndex}function Wn(t,e){if(t.parent&&-1!==t.parent.injectorIndex)return t.parent.injectorIndex;for(var n=e[Ft],r=1;n&&-1===n.injectorIndex;)n=(e=e[Kt])?e[Ft]:null,r++;return n?n.injectorIndex|r<<16:-1}function qn(t,e,n){!function(t,e,n){ngDevMode&&bt(e.firstTemplatePass,!0,"expected firstTemplatePass to be true");var r="string"!=typeof n?n[m]:n.charCodeAt(0)||0;null==r&&(r=n[m]=Vn++);var i=r&Bn,o=1<<i,a=128&i,s=64&i,u=32&i,c=e.data;a?s?u?c[t+7]|=o:c[t+6]|=o:u?c[t+5]|=o:c[t+4]|=o:s?u?c[t+3]|=o:c[t+2]|=o:u?c[t+1]|=o:c[t]|=o}(t,e[At],n)}function Gn(t,e){ngDevMode&&Ze(t,0,3,4),ngDevMode&&Ot(t,"expecting tNode");var n=t.attrs;if(n)for(var r=0;r<n.length;r+=2){var i=n[r];if(3===i)break;if(i==e)return n[r+1]}return null}function Yn(t,e,n,r,i){if(void 0===r&&(r=Pe.Default),t){var o=function(t){if(ngDevMode&&Ot(t,"token must be defined"),"string"==typeof t)return t.charCodeAt(0)||0;var e=t[m];return"number"==typeof e?e&Bn:e}(n);if("function"==typeof o){var a=_n(),s=gn();bn(t,e);try{var u=o();if(null!=u||r&Pe.Optional)return u;throw new Error("No provider for "+ue(n)+"!")}finally{bn(a,s)}}else if("number"==typeof o){var c=null,l=Un(t,e),d=oe,p=r&Pe.Host?De(e)[Ft]:null;for((-1===l||r&Pe.SkipSelf)&&(d=-1===l?Wn(t,e):e[l+re],Jn(r,!1)?(c=e[At],l=Ce(d),e=Se(d,e)):l=-1);-1!==l;){d=e[l+re];var f=e[At];if($n(o,l,f.data)){var h=Xn(l,e,n,c,r,p);if(h!==Kn)return h}Jn(r,e[At].data[l+ne]===p)&&$n(o,l,e)?(c=f,l=Ce(d),e=Se(d,e)):l=-1}}}if(r&Pe.Optional&&void 0===i&&(i=null),0==(r&(Pe.Self|Pe.Host))){var g=e[zt];return g?g.get(n,i,r&Pe.Optional):Ke(n,i,r&Pe.Optional)}if(r&Pe.Optional)return i;throw new Error("NodeInjector: NOT_FOUND ["+ue(n)+"]")}var Kn={};function Xn(t,e,n,r,i,o){var a=e[At],s=a.data[t+ne],u=Qn(s,e,n,null==r?ge(s)&&Fn:r!=a&&3===s.type,i&Pe.Host&&o===s);return null!==u?Zn(a.data,e,u,s):Kn}function Qn(t,e,n,r,i){for(var o=e[At],a=t.providerIndexes,s=o.data,u=65535&a,c=t.directiveStart,l=t.directiveEnd,d=a>>16,p=i?u+d:l,f=r?u:u+d;f<p;f++){var h=s[f];if(f<c&&n===h||f>=c&&h.type===n)return f}if(i){var m=s[c];if(m&&ye(m)&&m.type===n)return c}return null}function Zn(t,e,n,r){var i,o=e[n];if(null!=(i=o)&&"object"==typeof i&&Object.getPrototypeOf(i)==se){var a=o;if(a.resolving)throw new Error("Circular dep for "+ue(t[n]));var s=Ln(a.canSeeViewProviders);a.resolving=!0;var u=void 0;a.injectImpl&&(u=qe(a.injectImpl));var c=_n(),l=gn();bn(r,e);try{o=e[n]=a.factory(null,t,e,r)}finally{a.injectImpl&&qe(u),Ln(s),a.resolving=!1,bn(c,l)}}return o}function $n(t,e,n){var r=1<<t,i=64&t,o=32&t;return!!((128&t?i?o?n[e+7]:n[e+6]:o?n[e+5]:n[e+4]:i?o?n[e+3]:n[e+2]:o?n[e+1]:n[e])&r)}function Jn(t,e){return!(t&Pe.Self||t&Pe.Host&&e)}var tr=function(){function t(t,e){this._tNode=t,this._lView=e}return t.prototype.get=function(t,e){return Yn(this._tNode,this._lView,t,void 0,e)},t}();function er(t){var e=t,n=gt(e)||yt(e)||_t(e)||_(e)||v(e);return n&&void 0!==n.factory?n.factory:null}function nr(t){var e=er(Object.getPrototypeOf(t.prototype).constructor);return null!==e?e:function(t){return new t}}function rr(t){var e,n,r=we(t);if(r){if(Array.isArray(r)){var i=r,o=void 0,a=void 0,s=void 0;if((n=t)&&n.constructor&&n.constructor.ngComponentDef){if(-1==(o=cr(i,t)))throw new Error("The provided component was not found in the application");a=t}else if(e=t,e&&e.constructor&&e.constructor.ngDirectiveDef){if(-1==(o=function(t,e){var n=t[At].firstChild;for(;n;){for(var r=n.directiveStart,i=n.directiveEnd,o=r;o<i;o++)if(t[o]===e)return n.index;n=ur(n)}return-1}(i,t)))throw new Error("The provided directive was not found in the application");s=lr(o,i,!1)}else if(-1==(o=sr(i,t)))return null;var u=we(m=le(i[o])),c=u&&!Array.isArray(u)?u:ir(i,o,m);if(a&&void 0===c.component&&(c.component=a,ar(c.component,c)),s&&void 0===c.directives){c.directives=s;for(var l=0;l<s.length;l++)ar(s[l],c)}ar(c.native,c),r=c}}else{var d=t;ngDevMode&&St(d);for(var p=d;p=p.parentNode;){var f=we(p);if(f){i=void 0;if(!(i=Array.isArray(f)?f:f.lView))return null;var h=sr(i,d);if(h>=0){var m;ar(m=le(i[h]),c=ir(i,h,m)),r=c;break}}}}return r||null}function ir(t,e,n){return{lView:t,nodeIndex:e,native:n,component:void 0,directives:void 0,localRefs:void 0}}function or(t){var e,n=we(t);if(Array.isArray(n)){var r=cr(n,t);(i=ir(n,r,(e=he(r,n))[Nt])).component=t,ar(t,i),ar(i.native,i)}else{var i;e=he((i=n).nodeIndex,i.lView)}return e}function ar(t,e){t[ee]=e}function sr(t,e){for(var n=t[At].firstChild;n;){if(pe(n,t)===e)return n.index;n=ur(n)}return-1}function ur(t){if(t.child)return t.child;if(t.next)return t.next;for(;t.parent&&!t.parent.next;)t=t.parent;return t.parent&&t.parent.next}function cr(t,e){var n=t[At].components;if(n)for(var r=0;r<n.length;r++){var i=n[r];if(he(i,t)[Vt]===e)return i}else if(he(Xt,t)[Vt]===e)return Xt;return-1}function lr(t,e,n){var r=e[At].data[t],i=r.directiveStart;if(0==i)return at;var o=r.directiveEnd;return!n&&1&r.flags&&i++,e.slice(i,o)}function dr(t){var e,n,r,i,o=xr(t);return void 0===o.component&&(o.component=(e=o.nodeIndex,n=o.lView,r=n[At].data[e],i=r.directiveStart,1&r.flags?n[i]:null)),o.component}function pr(t){return xr(t).lView[Vt]}function fr(t){for(var e=_r(t).lView;e[It]&&null===e[Nt];)e=e[It];return 128&e[Dt]?null:e[Vt]}function hr(t){return function(t){var e;Array.isArray(t)?(ngDevMode&&Ot(t,"lView"),e=t):(ngDevMode&&Ot(t,"component"),e=je(t));for(;e&&!(128&e[Dt]);)e=e[It];return e}(Array.isArray(t)?t:_r(t).lView)[Vt]}function mr(t){return Object(r.__spread)(hr(t).components)}function gr(t){var e=_r(t),n=e.lView[At].data[e.nodeIndex];return new tr(n,e.lView)}function yr(t){var e=_r(t);return void 0===e.directives&&(e.directives=lr(e.nodeIndex,e.lView,!1)),e.directives||[]}function _r(t,e){void 0===e&&(e=!0);var n=rr(t);if(!n&&e)throw new Error(ngDevMode?"Unable to find context associated with "+ue(t):"Invalid ng target");return n}function vr(t){var e=_r(t);return void 0===e.localRefs&&(e.localRefs=function(t,e){var n=t[At].data[e];if(n&&n.localNames){for(var r={},i=0;i<n.localNames.length;i+=2){var o=n.localNames[i],a=n.localNames[i+1];r[o]=-1===a?pe(n,t):t[a]}return r}return null}(e.lView,e.nodeIndex)),e.localRefs||{}}function br(t){return rr(t).native}function xr(t){if(!(t instanceof Node))throw new Error("Expecting instance of DOM Node");return _r(t)}function wr(t){return"boolean"==typeof t.useCapture}function jr(t){var e=xr(t).lView,n=e[At],r=e[Bt],i=n.cleanup,o=[];if(i&&r)for(var a=0;a<i.length;){var s=i[a++],u=i[a++];if("string"==typeof s){var c=s,l=le(e[u]),d=r[i[a++]],p=i[a++],f="boolean"==typeof p?p:!(p>=0)&&null;t==l&&o.push({element:t,name:c,callback:d,useCapture:f})}}return o.sort(Or),o}function Or(t,e){return t.name==e.name?0:t.name<e.name?-1:1}function Cr(t){return"ng-reflect-"+(t=t.replace(/[$@]/g,"_").replace(Er,function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return"-"+t[1].toLowerCase()}))}var Er=/([A-Z])/g;function Sr(t){try{return null!=t?t.toString().slice(0,30):t}catch(t){return"[ERROR] Exception while trying to serialize the value"}}function Mr(t,e){var n=Ar(t),r=Ar(e);return n&&r?function(t,e,n){var r=t[q()](),i=e[q()]();for(;;){var o=r.next(),a=i.next();if(o.done&&a.done)return!0;if(o.done||a.done)return!1;if(!n(o.value,a.value))return!1}}(t,e,Mr):!(n||!(t&&("object"==typeof t||"function"==typeof t))||r||!(e&&("object"==typeof e||"function"==typeof e)))||Y(t,e)}var Tr=function(){function t(t){this.wrapped=t}return t.wrap=function(e){return new t(e)},t.unwrap=function(e){return t.isWrapped(e)?e.wrapped:e},t.isWrapped=function(e){return e instanceof t},t}(),kr=function(){function t(t,e,n){this.previousValue=t,this.currentValue=e,this.firstChange=n}return t.prototype.isFirstChange=function(){return this.firstChange},t}();function Ar(t){return!!Dr(t)&&(Array.isArray(t)||!(t instanceof Map)&&q()in t)}function Dr(t){return null!==t&&("function"==typeof t||"object"==typeof t)}function Ir(t){throw new Error("Multiple components match node with tagname "+t.tagName)}var Pr={};function Rr(t,e,n){return t[e]=n}function Nr(t,e){return ngDevMode&&kt(t,t[e]),ngDevMode&&xt(t[e],Pr,"Stored value should never be NO_CHANGE."),t[e]}function Fr(t,e,n){if(ngDevMode&&xt(n,Pr,"Incoming value should never be NO_CHANGE."),ngDevMode&&wt(e,t.length,"Slot should have been initialized to NO_CHANGE"),t[e]===Pr)t[e]=n;else{if(r=t[e],i=n,r!=r&&i!=i||r===i)return!1;ngDevMode&&En()&&(Mr(t[e],n)||function(t,e,n){var r="ExpressionChangedAfterItHasBeenCheckedError: Expression has changed after it was checked. Previous value: '"+e+"'. Current value: '"+n+"'.";throw t&&(r+=" It seems like the view has been created after its parent and its children have been dirty checked. Has it been created in a change detection hook ?"),new Error(r)}(jn(t),t[e],n)),t[e]=n}var r,i;return!0}function Lr(t,e,n,r){var i=Fr(t,e,n);return Fr(t,e+1,r)||i}function Br(t,e,n,r,i){var o=Lr(t,e,n,r);return Fr(t,e+2,i)||o}function Vr(t,e,n,r,i,o){var a=Lr(t,e,n,r);return Lr(t,e+2,i,o)||a}var zr,Hr,Ur="ngProjectAs";function Wr(t){return!!t.listen}(Hr=zr||(zr={}))[Hr.Important=1]="Important",Hr[Hr.DashCase=2]="DashCase";var qr={createRenderer:function(t,e){return document}};function Gr(t,e){return null==t.parent?(r=(n=e)[Ft])&&2!==r.type?pe(r,n[It]):null:pe(Yr(t),e);var n,r}function Yr(t){for(var e=t.parent;e&&5===e.type;)e=e.parent;return e}function Kr(t,e){if(-1===t.index){var n=e[Gt];return n>-1?e[It][n]:null}return e[It][t.parent.index]}function Xr(t,e){var n=Kr(t,e);return n?n[Jt]:null}var Qr=[];function Zr(t,e,n,r,i){for(var o=t[At].node,a=-1,s=t,u=o.child;u;){var c=null;if(3===u.type){$r(e,n,r,pe(u,s),i);var l=s[u.index];_e(l)&&$r(e,n,r,l[$t],i)}else if(0===u.type){var d=s[u.index];$r(e,n,r,d[$t],i),r&&(d[Jt]=r),d[Zt].length&&(c=(s=d[Zt][0])[At].node,i=d[$t])}else if(1===u.type){var p=De(s),f=p[Ft].projection[u.projection];Qr[++a]=u,Qr[++a]=s,f&&(c=(s=p[It])[At].data[f.index])}else c=u.child;if(null===c)for(null===u.next&&2&u.flags&&(s=Qr[a--],u=Qr[a--]),c=u.next;!c;){if(null===(u=u.parent||s[At].node)||u===o)return null;0===u.type&&(i=(s=s[It])[u.index][$t]),c=2===u.type&&s[Pt]?(s=s[Pt])[At].node:u.next}u=c}}function $r(t,e,n,r,i){0===t?Wr(e)?e.insertBefore(n,r,i):n.insertBefore(r,i,!0):1===t?Wr(e)?e.removeChild(n,r):n.removeChild(r):2===t&&(ngDevMode&&ngDevMode.rendererDestroyNode++,e.destroyNode(r))}function Jr(t,e){return Wr(e)?e.createText(ue(t)):e.createTextNode(ue(t))}function ti(t,e,n){var r=Xr(t[At].node,t);if(ngDevMode&&Qe(t[At].node,2),r){var i=t[Ut];Zr(t,e?0:1,i,r,n)}}function ei(t,e,n,r,i){var o=e[Zt];r>0&&(o[r-1][Pt]=t),r<o.length?(t[Pt]=o[r],o.splice(r,0,t)):(o.push(t),t[Pt]=null),i>-1&&(t[Gt]=i,t[It]=n),t[Rt]&&t[Rt].insertView(r),t[Dt]|=16}function ni(t,e,n){var r=t[Zt],i=r[e];return e>0&&(r[e-1][Pt]=i[Pt]),r.splice(e,1),n||ti(i,!1),i[Rt]&&i[Rt].removeView(),i[Gt]=-1,i[It]=null,i[Dt]&=-17,i}function ri(t,e,n){var r=t[Zt][n];ni(t,n,!!e.detached),oi(r)}function ii(t){var e=t[At].childIndex;return-1===e?null:t[e]}function oi(t){var e=t[Ut];Wr(e)&&e.destroyNode&&Zr(t,2,e,null),function(t){if(-1===t[At].childIndex)return si(t);for(var e=ii(t);e;){var n=null;if(e.length>=Xt){var r=e;r[At].childIndex>-1&&(n=ii(r))}else{var i=e;i[Zt].length&&(n=i[Zt][0])}if(null==n){for(;e&&!e[Pt]&&e!==t;)si(e),e=ai(e,t);si(e||t),n=e&&e[Pt]}e=n}}(t),t[Dt]|=64}function ai(t,e){var n;return t.length>=Xt&&(n=t[Ft])&&2===n.type?Kr(n,t):t[It]===e?null:t[It]}function si(t){if(t.length>=Xt){var e=t;null!=(s=(o=e)[At])&&null!=(a=s.destroyHooks)&&sn(o,a),(i=(r=e)[At]&&r[At].pipeDestroyHooks)&&sn(r,i),function(t){var e=t[At].cleanup;if(null!=e){for(var n=t[Bt],r=0;r<e.length-1;r+=2)if("string"==typeof e[r]){var i=e[r+1],o=n[e[r+2]],a=le(t[i]),s=e[r+3];"boolean"==typeof s?a.removeEventListener(e[r],o,s):s>=0?n[s]():n[-s].unsubscribe(),r+=2}else if("number"==typeof e[r]){var u=n[e[r]];u()}else{var c=n[e[r+1]];e[r].call(c)}t[Bt]=null}}(e);var n=e[Ft];n&&3===n.type&&Wr(e[Ut])&&(ngDevMode&&ngDevMode.rendererDestroy++,e[Ut].destroy())}var r,i,o,a,s}function ui(t,e){if(ci(t,e)){if(ve(e))return di(e[Ut],pe(t,e));var n=e[Ft],r=t.parent;return null!=r&&4===r.type&&(t=fi(r)),null==t.parent&&2===n.type?Xr(n,e):Gr(t,e)}return null}function ci(t,e){var n,r,i=t,o=t.parent;return t.parent&&(4===t.parent.type?o=(i=fi(t)).parent:5===t.parent.type&&(o=(i=Yr(i)).parent)),null===o&&(o=e[Ft]),o&&2===o.type?null!=(r=Kr(o,e))&&null!=r[Jt]:null==(n=i).parent||!(3!==n.parent.type||1&n.parent.flags)}function li(t,e,n,r){Wr(t)?t.insertBefore(e,n,r):e.insertBefore(n,r,!0)}function di(t,e){return Wr(t)?t.parentNode(e):e.parentNode}function pi(t,e,n){if(void 0===t&&(t=null),null!==t&&ci(e,n)){var r=n[Ut],i=Gr(e,n),o=e.parent||n[Ft];if(2===o.type){var a=Kr(o,n),s=a[Zt],u=s.indexOf(n);li(r,a[Jt],t,hi(u,s,a[$t]))}else if(4===o.type){li(r,ui(e,n),t,i)}else if(5===o.type){li(r,i,t,pe(e.parent,n))}else Wr(r)?r.appendChild(i,t):i.appendChild(t);return!0}return!1}function fi(t){for(;null!=t.parent&&4===t.parent.type;)t=t.parent;return t}function hi(t,e,n){if(t+1<e.length){var r=e[t+1],i=r[Ft];return i.child?pe(i.child,r):n}return n}function mi(t,e,n){if(null!==e&&ci(t,n)){var r=Gr(t,n),i=n[Ut];return Wr(i)?i.removeChild(r,e):r.removeChild(e),!0}return!1}function gi(t,e,n,r){var i=pe(t,r);pi(i,e,n),ar(i,r);var o=ui(e,n),a=r[t.index];if(0===t.type){a[Jt]=o;for(var s=a[Zt],u=0;u<s.length;u++)ti(s[u],!0,a[$t])}else{if(4===t.type)for(var c=t.child;c;)gi(c,e,n,r),c=c.next;_e(a)&&(a[Jt]=o,pi(a[$t],e,n))}}var yi="ng-template";function _i(t,e,n){ngDevMode&&Ot(e[0],"Selector should have a tag name");for(var r,i,o,a,s,u,c,l,d=4,p=t.attrs,f=p?p.indexOf(3):-1,h=!1,m=0;m<e.length;m++){var g=e[m];if("number"!=typeof g){if(!h)if(4&d){if(d=2|1&d,""!==g&&(l=n,(c=g)!==(u=t).tagName||!(l||3===u.type||4===u.type||0===u.type&&c===yi))||""===g&&1===e.length){if(vi(d))return!1;h=!0}}else{var y=bi(8&d?"class":g,p);if(-1===y){if(vi(d))return!1;h=!0;continue}var _=8&d?g:e[++m];if(""!==_){var v=void 0,b=p[y];if(f>-1&&y>f?v="":(ngDevMode&&xt(b,0,"We do not match directives on namespaced attributes"),v=p[y+1]),8&d&&(i=_,void 0,void 0,void 0,o=(r=v).length,a=r.indexOf(i),s=a+i.length,-1===a||a>0&&" "!==r[a-1]||s<o&&" "!==r[s])||2&d&&_!==v){if(vi(d))return!1;h=!0}}}}else{if(!h&&!vi(d)&&!vi(g))return!1;if(h&&vi(g))continue;h=!1,d=g|1&d}}return vi(d)||h}function vi(t){return 0==(1&t)}function bi(t,e){if(null===e)return-1;for(var n=!1,r=0;r<e.length;){var i=e[r];if(i===t)return r;0===i?r+=4:(3===i&&(n=!0),r+=n?1:2)}return-1}function xi(t,e,n){void 0===n&&(n=!1);for(var r=0;r<e.length;r++)if(_i(t,e[r],n))return!0;return!1}function wi(t,e,n){for(var r=function(t){var e=t.attrs;if(null!=e){var n=e.indexOf(Ur);if(0==(1&n))return e[n+1]}return null}(t),i=0;i<e.length;i++)if(r===n[i]||null===r&&xi(t,e[i],!0))return i+1;return 0}function ji(t,e){return new Oi(t,e)}var Oi=function(){return function(t,e){this.fn=t,this.value=e}}(),Ci=function(){function t(){this._players=[]}return t.prototype.flushPlayers=function(){for(var t=0;t<this._players.length;t++){var e=this._players[t];e.parent||0!==e.state||e.play()}this._players.length=0},t.prototype.queuePlayer=function(t){this._players.push(t)},t}(),Ei="@";function Si(t,e,n,r){return[0,[null,-1,!1,e||null],n||[null],r||[null],[0,0],t||null,null,null,null]}function Mi(t,e){for(var n=t,r=e[n],i=e;Array.isArray(r);)i=r,r=r[Nt];if(s=i,Array.isArray(s)&&"number"==typeof s[0]&&Array.isArray(s[2]))return i;var o,a,s,u=fe(t-Xt,e).stylingTemplate;return i!==e&&(n=Nt),i[n]=u?(o=r,(a=u.slice())[5]=o,a[0]|=32,a):Si(r)}function Ti(t){return t[0]===Ei}function ki(t,e,n,r,i,o){return o=o||n,i?t[i]=r:t.push(r),!!r&&(r.addEventListener(200,function(){var e=t.indexOf(r),n=t[0];e&&(e<n?t[e]=null:t.splice(e,1)),r.destroy()}),(e.playerHandler||(e.playerHandler=new Ci)).queuePlayer(r,o),!0)}function Ai(t){return t[8]}function Di(t){return t[8]=[5,null,null,null,null]}function Ii(){throw new Error("Only elements that exist in an Angular application can be used for animations")}function Pi(t){return!!(8&t.flags)}function Ri(t){for(var e=Si(),n=e[3]=[null],r=e[2]=[null],i=-1,o=0;o<t.length;o++){var a=t[o];if("number"==typeof a)i=a;else if(2===i)r.push(a,t[++o]);else if(1===i)n.push(a,!0);else if(3===i)break}return e}function Ni(t,e,n){for(var r=1;r<t.length;){if(t[r]===e){var i=t[r+1];return void(null!=i&&0!=i||(t[r+1]=n))}r+=2}t.push(e,n)}function Fi(t,e,n,r){for(var i=1;i<n.length;i+=2){var o=n[i+1];o&&(r?Ui(t,n[i+0],!0,e,null):Hi(t,n[i+0],o,e,null))}}function Li(t,e,n,r,i,o){if(!(32&t[0])){var a=function(t,e,n){var r,i=t[1],o=t[4].length,a=wo(i,e);if(-1===a)r=i.length/4,i.push(e,o,!1,n||null);else{var s=a+1;if(i[s]>=0)return-1;r=a/4;var u=a+1;i[u]=o;var c=a+3;i[c]=n||null}return r}(t,e,i);if(-1!==a){var s=t[4],u=s[1],c=s[0],l=4*c,d=9+l,p=d+4*u,f=p+l,h=s.length;s.push(r?r.length:0,n?n.length:0);var m=0,g=[];if(r&&r.length)for(var y=0;y<r.length;y++){var _=r[y];-1==(x=Bi(t,_,9,d))&&(x=d+m,m+=4,g.push(_)),s.push(x)}var v=[];if(n&&n.length)for(var b=0;b<n.length;b++){var x,w=n[b];-1==(x=Bi(t,w,d,p))?(x=p+m,m+=4,v.push(w)):x+=4*g.length,s.push(x)}var j=2;if(g.length)for(;j<h;){var O=s[j+0],C=s[j+1];if(C)for(var E=j+2+O,S=E;S<E+C;S++)s[S]+=4*g.length;j+=2+(O+C)}for(var M=v.length+g.length,T=9;T<t.length;T+=4){var k=T>=p,A=T>=(k?f:d),D=ao(t,T),I=Xi(D),P=Qi(D);oo(t,T,Yi(D,I,P+=k?A?4*g.length:0:4*M+4*(A?g.length:0)))}for(var R=0;R<4*g.length;R++)t.splice(f,0,null),t.splice(d,0,null),d++,p++,f+=2;for(var N=0;N<4*v.length;N++)t.splice(p,0,null),t.push(null),p++,f++;for(var F=t[3],L=t[2],B=0;B<M;B++){var V=B>=g.length,z=V?B-g.length:B,H=V?v[z]:g[z],U=void 0,W=void 0;V?(U=f+4*(u+z),W=d+4*(u+z)):(U=p+4*(c+z),W=9+4*(c+z));var q=V?F:L,G=jo(q,H);-1===G?(G=q.length+1,q.push(H,!V&&null)):G+=1;var Y=yo(t,H,V,i||null);oo(t,W,Yi(Y,G,U)),$i(t,W,H),Ji(t,W,null),no(t,W,0,a),oo(t,U,Yi(Y,G,W)),$i(t,U,H),Ji(t,U,null),no(t,U,0,a)}s[1]=u+v.length,s[0]=c+g.length,oo(t,0,Yi(0,0,p)|(o?16:0))}}}function Bi(t,e,n,r){for(var i=n;i<r;i+=4)if(uo(t,i)===e)return i;return-1}function Vi(t,e,n,r,i){var o,a,s,u,c,l,d,p,f,h=xo(t,i||null),m=(a=h,s=e,u=r,c=(o=t)[1][4*a+1],l=o[4],d=c+2+(u?l[c+0]:0)+s,l[d]),g=so(t,m),y=ao(t,m),_=bo(t,m),v=n instanceof Oi?n.value:n;if(_o(y,g,v)&&function(t,e,n,r){if(t)return e?r<=n:n===r;return!0}(g,v,_,h)){var b=2==(2&y),x=t[5],w=n instanceof Oi?new vo(n,x,b?1:2):null,j=w?n.value:n,O=ro(t,m),C=!1,E=w?O:0;if(to(t,w,O)){var S=eo(t,w,O);E=w?S:0,C=!0}if((C||_!==h)&&no(t,m,E,h),_!==h){var M=uo(t,m),T=Oo(t,h);p=t,f=m,!(!T||!T(M))?p[f]|=4:p[f]&=-5}Ji(t,m,j);var k=Qi(y),A=so(t,k);if(!A||_o(y,A,j)){var D=!1,I=!0;!go(j,b)&&go(A,b)&&(D=!0,I=!1),Wi(t,k,D),Wi(t,m,I),Co(t,h,!0),lo(t,!0)}C&&po(t,!0)}}function zi(t,e,n,r,i,o,a){var s,u=0,c=xo(t,a||null);if(qi(t,0)&&(s=c,t[1][4*s+2])){for(var l=8&t[0],d=t[5],p=Zi(t),f=co(t),h=!1,m=9;m<t.length;m+=4)if(qi(t,m)){var g=ao(t,m),y=bo(t,m);if(c!==y){h=!0;continue}var _=uo(t,m),v=so(t,m),b=4&g?Oo(t,y):null,x=io(t,m),w=!!(2&g),j=!w||!f,O=v;if(m<p&&!go(O,w))O=so(t,Qi(g));w||go(O,w)||!j||(O=Ki(t,g)),(!r||O)&&(w?Ui(d,_,!!O,e,i,x):Hi(d,_,O,e,b,o,x)),Wi(t,m,!1)}if(l){var C=Array.isArray(n)?xe(n):n,E=Ai(t),S=E[0];for(m=1;m<S;m+=2){var M=E[m],T=m+1,k=E[T];if(M){var A=M.buildPlayer(k,r);if(void 0!==A){if(null!=A)ki(E,C,d,A,T)&&u++;k&&k.destroy()}}else k&&k.destroy()}po(t,!1)}Co(t,c,!1),lo(t,h)}return u}function Hi(t,e,n,r,i,o,a){n=i&&n?i(e,n):n,o||a?(o&&o.setValue(e,n),a&&a.setValue(e,n)):n?(n=n.toString(),ngDevMode&&ngDevMode.rendererSetStyle++,Wr(r)?r.setStyle(t,e,n,zr.DashCase):t.style.setProperty(e,n)):(ngDevMode&&ngDevMode.rendererRemoveStyle++,Wr(r)?r.removeStyle(t,e,zr.DashCase):t.style.removeProperty(e))}function Ui(t,e,n,r,i,o){i||o?(i&&i.setValue(e,n),o&&o.setValue(e,n)):n?(ngDevMode&&ngDevMode.rendererAddClass++,Wr(r)?r.addClass(t,e):t.classList.add(e)):(ngDevMode&&ngDevMode.rendererRemoveClass++,Wr(r)?r.removeClass(t,e):t.classList.remove(e))}function Wi(t,e,n){var r=e>=9?e+0:e;n?t[r]|=1:t[r]&=-2}function qi(t,e){return 1==(1&t[e>=9?e+0:e])}function Gi(t,e){return 2==(2&t[e>=9?e+0:e])}function Yi(t,e,n){return 63&t|e<<6|n<<20}function Ki(t,e){var n=Xi(e);return(2&e?t[3]:t[2])[n]}function Xi(t){return t>>6&16383}function Qi(t){var e=t>>20&16383;return e>=9?e:-1}function Zi(t){return Qi(t[0])}function $i(t,e,n){t[e+1]=n}function Ji(t,e,n){t[e+2]=n}function to(t,e,n){var r=t[8];if(e){if(!r||0===n)return!0}else if(!r)return!1;return r[n]!==e}function eo(t,e,n){var r=t[8]||Di(t);return n>0?r[n]=e:(n=r[0],r.splice(n,0,e,null),r[0]+=2),n}function no(t,e,n,r){var i=n<<16|r;t[e+3]=i}function ro(t,e){return t[e+3]>>16&65535}function io(t,e){var n=ro(t,e);if(n){var r=t[8];if(r)return r[n]}return null}function oo(t,e,n){t[0===e?e:e+0]=n}function ao(t,e){return t[0===e?e:e+0]}function so(t,e){return t[e+2]}function uo(t,e){return t[e+1]}function co(t){return 16&t[0]}function lo(t,e){Wi(t,0,e)}function po(t,e){e?t[0]|=8:t[0]&=-9}function fo(t,e,n){for(var r=(n||0)+1;r<t.length;r+=4){if(t[r]==e)return r-1}return-1}function ho(t,e,n){var r=so(t,e),i=uo(t,e),o=ao(t,e),a=ro(t,e),s=o,u=ao(t,n),c=Qi(s);c>=0&&oo(t,c,Yi(l=ao(t,c),Xi(l),n));var l,d=Qi(u);d>=0&&oo(t,d,Yi(l=ao(t,d),Xi(l),e));Ji(t,e,so(t,n)),$i(t,e,uo(t,n)),oo(t,e,ao(t,n));no(t,e,ro(t,n),0),Ji(t,n,r),$i(t,n,i),oo(t,n,o),no(t,n,a,0)}function mo(t,e,n,r,i,o,a,s){var u=e<t.length;t.splice(e,0,1|i|(n?2:0),r,o,0),no(t,e,s,a),u&&function(t,e){for(var n=e;n<t.length;n+=4){var r=Qi(ao(t,n));if(r>0){var i=Xi(ao(t,r));oo(t,r,Yi((qi(t,r)?1:0)|(Gi(t,r)?2:0)|(4==(4&t[(o=r)>=9?o+0:o])?4:0),i,n))}}var o}(t,e+4)}function go(t,e){return e?!!t:null!==t}function yo(t,e,n,r){var i,o=r&&r(e)?4:0;return n?(o|=2,i=jo(t[3],e)):i=jo(t[2],e),Yi(o,i=i>0?i+1:0,0)}function _o(t,e,n){return!(2&t)&&(e&&n)&&4&t?e.toString()!==n.toString():e!==n}var vo=function(){function t(t,e,n){this._element=e,this._type=n,this._values={},this._dirty=!1,this._factory=t}return t.prototype.setValue=function(t,e){this._values[t]!==e&&(this._values[t]=e,this._dirty=!0)},t.prototype.buildPlayer=function(t,e){if(this._dirty){var n=this._factory.fn(this._element,this._type,this._values,e,t||null);return this._values={},this._dirty=!1,n}},t}();function bo(t,e){return 65535&t[e+3]}function xo(t,e){var n=wo(t[1],e);return ngDevMode&&xt(n,-1,"The provided directive "+e+" has not been allocated to the element's style/class bindings"),n>0?n/4:0}function wo(t,e){for(var n=0;n<t.length;n+=4)if(t[n]===e)return n;return-1}function jo(t,e){for(var n=1;n<t.length;n+=2)if(t[n]===e)return n;return-1}function Oo(t,e){var n=t[1];return n[4*e+3]||n[3]||null}function Co(t,e,n){t[1][4*e+2]=n}function Eo(t){var e=t[6];if(null==e){e="";for(var n=t[3],r=1;r<n.length;r+=2){n[r+1]&&(e+=(e.length?" ":"")+n[r])}t[6]=e}return e}var So=Promise.resolve(null);function Mo(t){var e=t[At];if(e.firstTemplatePass=!1,kn(!1),!jn(t)){var n=En();on(t,e,n),function(t){for(var e=ii(t);null!==e;e=e[Pt])if(e.length<Xt&&-1===e[Qt])for(var n=e,r=0;r<n[Zt].length;r++){var i=n[Zt][r];ngDevMode&&Ot(i[At],"TView must be allocated"),Io(i,i[At],i[Vt])}}(t),function(t){if(null!=t.contentQueries)for(var e=0;e<t.contentQueries.length;e+=2){var n=t.contentQueries[e],r=t.data[n];r.contentQueriesRefresh(n-Xt,t.contentQueries[e+1])}}(e),an(t,e.contentHooks,e.contentCheckHooks,n),function(t,e){if(t.expandoInstructions){var n=e[Lt]=t.expandoStartIndex;In(n);for(var r=-1,i=-1,o=0;o<t.expandoInstructions.length;o++){var a=t.expandoInstructions[o];if("number"==typeof a){if(a<=0){i=-a;var s=t.expandoInstructions[++o];r=n+=ie+s}else n+=a;In(n)}else null!==a&&(e[Lt]=n,a(2,le(e[r]),i)),r++}}}(e,t)}!function(t){if(null!=t)for(var e=0;e<t.length;e++)Da(t[e])}(e.components)}function To(t,e,n,r,i,o,a,s){var u=e.blueprint.slice();return u[Dt]=51|r,u[It]=u[Kt]=t,u[Vt]=n,u[Ht]=i||t&&t[Ht],ngDevMode&&Ot(u[Ht],"RendererFactory is required"),u[Ut]=o||t&&t[Ut],ngDevMode&&Ot(u[Ut],"Renderer is required"),u[Wt]=a||t&&t[Wt]||null,u[zt]=s||t&&t[zt]||null,u}function ko(t,e,n,r,i){var o=gn(),a=o[At],s=t+Xt;ngDevMode&&wt(s,o.length,"Slot should have been initialized with null"),o[s]=n;var u=a.data[s];null==u&&(u=a.data[s]=oa(o,e,s,r,i,null));var c=_n(),l=xn();return c&&(!l||null!=c.child||null===u.parent&&2!==c.type?l||(c.next=u):c.child=u),null==a.firstChild&&(a.firstChild=u),vn(u),wn(!0),u}function Ao(t,e){return null==e[At].node&&(e[At].node=oa(e,2,t,null,null,null)),e[Ft]=e[At].node}function Do(t){var e=t[At];e.firstTemplatePass&&(e.expandoStartIndex++,e.blueprint.push(null),e.data.push(null),t.push(null))}function Io(t,e,n){var r,i=xn(),o=_n();if(wn(!0),vn(null),128&t[Dt])Ba(xe(t));else try{wn(!0),vn(null),r=Pn(t,t[Ft]),Vo(),e.template(No(t),n),t[At].firstTemplatePass=!1,kn(!1),Mo(t)}finally{Nn(r),wn(i),vn(o)}}function Po(t){return void 0===t&&(t=1),Rn(t)}function Ro(t,e,n){var r=t[Ht],i=Pn(t,t[Ft]),o=!En();try{o&&r.begin&&r.begin(),jn(t)&&(n&&(Vo(),n(1,e)),Mo(t),t[Dt]&=-2),n&&n(2,e),Mo(t)}finally{o&&r.end&&r.end(),Nn(i)}}function No(t){return jn(t)?1:2}var Fo=null;function Lo(){Fo="http://www.w3.org/2000/svg"}function Bo(){Fo="http://www.w3.org/1998/MathML/"}function Vo(){Fo=null}function zo(t,e,n,r){Wo(t,e,n,r),Jo()}function Ho(t,e,n){var r=gn(),i=r[At],o=r[Ut];ngDevMode&&bt(r[Lt],i.bindingStartIndex,"element containers should be created before any bindings"),ngDevMode&&ngDevMode.rendererCreateComment++;var a=o.createComment(ngDevMode?"ng-container":"");ngDevMode&&kt(r,t-1),pi(a,ko(t,4,a,"ng-container",e||null),r),Go(i,r,n),ar(a,r)}function Uo(){var t=_n(),e=gn(),n=e[At];xn()?wn(!1):(ngDevMode&&Tt(_n()),vn(t=t.parent)),ngDevMode&&Qe(t,4);var r=e[Rt];r&&(e[Rt]=r.addNode(t)),tn(n,t)}function Wo(t,e,n,r){var i=gn(),o=i[At];ngDevMode&&bt(i[Lt],o.bindingStartIndex,"elements should be created before any bindings "),ngDevMode&&ngDevMode.rendererCreateElement++;var a=qo(e);ngDevMode&&kt(i,t-1);var s,u,c,l=ko(t,3,a,e,n||null);if(n&&(o.firstTemplatePass&&!l.stylingTemplate&&function(t){for(var e=0;e<t.length;e++){var n=t[e];if(1==n||2==n)return!0}return!1}(n)&&(l.stylingTemplate=Ri(n)),Xo(a,n)),pi(a,l,i),Go(o,i,r),0===He&&ar(a,i),He++,o.firstTemplatePass){var d=ls(l);d&&d.hasOwnProperty("class")&&(l.flags|=8)}l.stylingTemplate&&0==(8&l.flags)&&(s=a,u=l.stylingTemplate,c=i[Ut],Fi(s,c,u[3],!0),Fi(s,c,u[2],!1))}function qo(t,e){var n=e||gn()[Ut];return Wr(n)?n.createElement(t,Fo):null===Fo?n.createElement(t):n.createElementNS(Fo,t)}function Go(t,e,n,r){if(void 0===r&&(r=pe),un){var i=_n();Tn()&&(ngDevMode&&ngDevMode.firstTemplatePass++,function(t,e,n,r,i){ngDevMode&&bt(Tn(),!0,"should run on first template pass only");var o=i?{"":-1}:null;if(n){xa(r,t.data.length,n.length);for(var a=0;a<n.length;a++){var s=n[a];s.providersResolver&&s.providersResolver(s)}ga(t,r,n.length);for(var a=0;a<n.length;a++){var s=n[a],u=t.data.length;wa(t,e,s,s.factory),ba(t.data.length-1,s,o),Je(u,s.onInit,s.doCheck,t)}}o&&function(t,e,n){if(e)for(var r=t.localNames=[],i=0;i<e.length;i+=2){var o=n[e[i+1]];if(null==o)throw new Error("Export of name '"+e[i+1]+"' not found!");r.push(e[i],o)}}(r,i,o)}(t,e,function(t,e,n){ngDevMode&&bt(Tn(),!0,"should run on first template pass only");var r=t.directiveRegistry,i=null;if(r)for(var o=0;o<r.length;o++){var a=r[o];xi(n,a.selectors,!1)&&(i||(i=[]),qn(zn(_n(),e),e,a.type),ye(a)?(1&n.flags&&Ir(n),n.flags=1,i.unshift(a)):i.push(a))}return i}(t,e,i),i,n||null)),function(t,e,n){var r=n.directiveStart,i=n.directiveEnd;!Tn()&&r<i&&zn(n,e);for(var o=r;o<i;o++){var a=t.data[o];ye(a)&&ja(e,n,a);var s=Zn(t.data,e,o,n);ya(e,s,a,o)}}(t,e,i),function(t,e,n){for(var r=n.directiveStart,i=n.directiveEnd,o=t.expandoInstructions,a=Tn(),s=r;s<i;s++){var u=t.data[s],c=e[s];if(u.hostBindings){var l=o.length;fn(u),u.hostBindings(1,c,n.index-Xt),fn(null),l===o.length&&a&&o.push(u.hostBindings)}else a&&o.push(null)}}(t,e,i),function(t,e,n){var r=e.localNames;if(r)for(var i=e.index+1,o=0;o<r.length;o+=2){var a=r[o+1],s=-1===a?n(e,t):t[a];t[i++]=s}}(e,i,r)}}function Yo(t,e,n,r,i,o){return t.ngPrivateData||(t.ngPrivateData=Ko(-1,t,e,n,r,i,o))}function Ko(t,e,n,r,i,o,a){ngDevMode&&ngDevMode.tView++;var s,u,c=Xt+n,l=c+r,d=(s=c,(u=new Array(l).fill(null,0,s).fill(Pr,s))[Gt]=-1,u[Lt]=s,u);return d[At]={id:t,blueprint:d,template:e,viewQuery:a,node:null,data:d.slice(),childIndex:-1,bindingStartIndex:c,expandoStartIndex:l,expandoInstructions:null,firstTemplatePass:!0,initHooks:null,checkHooks:null,contentHooks:null,contentCheckHooks:null,viewHooks:null,viewCheckHooks:null,destroyHooks:null,pipeDestroyHooks:null,cleanup:null,contentQueries:null,components:null,directiveRegistry:"function"==typeof i?i():i,pipeRegistry:"function"==typeof o?o():o,firstChild:null}}function Xo(t,e){for(var n=gn()[Ut],r=Wr(n),i=0;i<e.length;){var o=e[i++];if("number"==typeof o){if(0!==o)break;var a=e[i++],s=e[i++],u=e[i++];ngDevMode&&ngDevMode.rendererSetAttribute++,r?n.setAttribute(t,s,u,a):t.setAttributeNS(a,s,u)}else{u=e[i++];o!==Ur&&(ngDevMode&&ngDevMode.rendererSetAttribute++,Ti(o)?r&&n.setProperty(t,o,u):r?n.setAttribute(t,o,u):t.setAttribute(o,u))}}}function Qo(t,e){return new Error("Renderer: "+t+" ["+ue(e)+"]")}function Zo(t,e){var n=t.createRenderer(null,null),r="string"==typeof e?Wr(n)?n.selectRootElement(e):n.querySelector(e):e;if(ngDevMode&&!r)throw Qo("string"==typeof e?"Host node with selector not found:":"Host node is required:",e);return r}function $o(t,e,n){void 0===n&&(n=!1);var r,i=gn(),o=_n(),a=i[At],s=a.firstTemplatePass&&(a.cleanup||(a.cleanup=[]));if(ngDevMode&&Ze(o,3,0,4),3===o.type){var u=pe(o,i);ngDevMode&&ngDevMode.rendererAddEventListener++;var c=i[Ut],l=(y=ps(i)).length,d=n;if(Wr(c)){var p=c.listen(u,t,e);y.push(e,p),d=l+1}else{var f=(r=e,function(t){!1===r(t)&&(t.preventDefault(),t.returnValue=!1)});u.addEventListener(t,f,n),y.push(f)}s&&s.push(t,o.index,l,d)}void 0===o.outputs&&(o.outputs=sa(o,1));var h,m=o.outputs;if(m&&(h=m[t])){var g=h.length;if(g)for(var y=ps(i),_=0;_<g;_+=2){ngDevMode&&kt(i,h[_]);var v=i[h[_]][h[_+1]].subscribe(e),b=y.length;y.push(e,v),s&&s.push(t,o.index,b,-(b+1))}}}function Jo(){var t=_n();xn()?wn(!1):(ngDevMode&&Tt(_n()),vn(t=t.parent)),ngDevMode&&Qe(t,3);var e=gn(),n=e[Rt];if(n&&(e[Rt]=n.addNode(t)),tn(gn()[At],t),He--,Pi(t)){var r=Mi(t.index,e);aa(e,t.inputs.class,Eo(r))}}function ta(t,e,n,r){if(n!==Pr){var i=gn(),o=i[Ut],a=de(t,i);if(null==n)ngDevMode&&ngDevMode.rendererRemoveAttribute++,Wr(o)?o.removeAttribute(a,e):a.removeAttribute(e);else{ngDevMode&&ngDevMode.rendererSetAttribute++;var s=null==r?ue(n):r(n);Wr(o)?o.setAttribute(a,e,s):a.setAttribute(e,s)}}}function ea(t,e,n,r,i){ia(t,e,n,r,i)}function na(t,e,n,r,i){ia(t,e,n,r,i,ra)}function ra(t,e){return e[t.index][Ut]}function ia(t,e,n,r,i,o){if(n!==Pr){var a,s,u,c=gn(),l=de(t,c),d=fe(t,c);if(!i&&(a=ls(d))&&(s=a[e]))aa(c,s,n),ge(d)&&(4&(u=he(t+Xt,c))[Dt]||(u[Dt]|=8)),ngDevMode&&(3!==d.type&&0!==d.type||function(t,e,n,r,i){for(var o,a=0;a<r.length;a+=2){var s=t[Ut],u=Cr(r[a+1]),c=Sr(i);if(3===n)Wr(s)?s.setAttribute(e,u,c):e.setAttribute(u,c);else if(void 0!==i){var l="bindings="+JSON.stringify(((o={})[u]=c,o),null,2);Wr(s)?s.setValue(e,l):e.textContent=l}}}(c,l,d.type,s,n));else if(3===d.type){var p=o?o(d,c):c[Ut];n=null!=r?r(n):n,ngDevMode&&ngDevMode.rendererSetProperty++,Wr(p)?p.setProperty(l,e,n):Ti(e)||(l.setProperty?l.setProperty(e,n):l[e]=n)}}}function oa(t,e,n,r,i,o){var a=_n();ngDevMode&&ngDevMode.tNode++;var s=xn()?a:a&&a.parent,u=s&&t&&s!==t[Ft]?s:null;return{type:e,index:n,injectorIndex:u?u.injectorIndex:-1,directiveStart:-1,directiveEnd:-1,flags:0,providerIndexes:0,tagName:r,attrs:i,localNames:null,initialInputs:void 0,inputs:void 0,outputs:void 0,tViews:o,next:null,child:null,parent:u,detached:null,stylingTemplate:null,projection:null}}function aa(t,e,n){for(var r=0;r<e.length;r+=2)ngDevMode&&kt(t,e[r]),t[e[r]][e[r+1]]=n}function sa(t,e){var n=gn()[At],r=null,i=t.directiveStart,o=t.directiveEnd;if(o>i)for(var a=0===e,s=n.data,u=i;u<o;u++){var c=s[u],l=a?c.inputs:c.outputs;for(var d in l)if(l.hasOwnProperty(d)){r=r||{};var p=l[d];r.hasOwnProperty(d)?r[d].push(u,p):r[d]=[u,p]}}return r}function ua(t,e,n,r){var i=_n();i.stylingTemplate||(i.stylingTemplate=Si()),Li(i.stylingTemplate,r||null,t,e,n,Pi(i))}function ca(t,e){var n=_n();n.stylingTemplate||(n.stylingTemplate=Ri(e)),function(t,e,n){var r=t[1];if(-1==wo(r,n)){r.push(n,-1,!1,null);for(var i=null,o=null,a=-1,s=0;s<e.length;s++){var u=e[s];"number"==typeof u?a=u:1==a?Ni(i=i||t[3],u,!0):2==a&&Ni(o=o||t[2],u,e[++s])}}}(n.stylingTemplate,e,t)}function la(t,e){var n=gn(),r=0!=(2&n[Dt]);zi(Mi(t+Xt,n),n[Ut],n,r,null,null,e)>0&&La(xe(n),2)}function da(t,e,n,r,i){var o=null;null!==n&&(o=r?ue(n)+r:n),Vi(Mi(t+Xt,gn()),e,o,!1,i)}function pa(t,e,n,r){var i=n instanceof Oi?n:!!n;Vi(Mi(t+Xt,gn()),e,i,!0,r)}function fa(t,e,n,r){if(void 0!=r)return function(t,e,n,r){throw new Error("unimplemented. Should not be needed by ViewEngine compatibility")}();var i=gn(),o=fe(t,i),a=Mi(t+Xt,i);if(Pi(o)&&e!==Pr){var s=Eo(a),u=(s.length?s+" ":"")+e;aa(i,o.inputs.class,u)}else!function(t,e,n,r){n=n||null;var i=xo(t,r||null),o=t[5],a=e instanceof Oi?new vo(e,o,1):null,s=n instanceof Oi?new vo(n,o,2):null,u=a?e.value:e,c=s?n.value:n,l=co(t)||u===Pr||u===t[6],d=c===Pr||c===t[7];if(!l||!d){t[6]=u,t[7]=c;var p=at,f=!1,h=!1,m=a?1:0;to(t,a,1)&&(eo(t,a,1),h=!0);var g=s?3:0;to(t,s,3)&&(eo(t,s,3),h=!0),l||("string"==typeof u?(p=u.split(/\s+/),f=!0):p=u?Object.keys(u):at);for(var y=u||ot,_=c?Object.keys(c):at,v=c||ot,b=_.length,x=!1,w=Zi(t),j=0,O=_.length+p.length;w<t.length&&j<O;){if(!(F=j>=b)&&!d||F&&!l){var C=F?j-b:j,E=F?p[C]:_[C],S=F?!!f||y[E]:v[E],M=F?m:g;if((L=uo(t,w))===E){var T=so(t,w),k=ao(t,w);if(no(t,w,M,i),_o(k,T,S)){Ji(t,w,S),h=h||!!M;var A=Ki(t,k);(null!=T||_o(k,A,S))&&(Wi(t,w,!0),x=!0)}}else{var D=fo(t,E,w);if(D>0){var I=so(t,D),P=ao(t,D);ho(t,w,D),_o(P,I,S)&&(A=Ki(t,P),Ji(t,w,S),(null!=I||_o(P,A,S))&&(Wi(t,w,!0),h=h||!!M,x=!0))}else{var R=yo(t,E,F,Oo(t,i));h=h||!!M,mo(t,w,F,E,R,S,i,M),x=!0}}}w+=4,j++}for(;w<t.length;)(!(F=2==(2&(k=ao(t,w))))&&!d||F&&!l)&&go(T=so(t,w),F)&&(Wi(t,w,!0),Ji(t,w,null),no(t,w,M=F?m:g,i),x=!0),w+=4;for(var N=Oo(t,i);j<O;){var F;if(!(F=j>=b)&&!d||F&&!l){C=F?j-b:j;var L=F?p[C]:_[C],B=(T=F?!!f||y[L]:v[L],k=1|yo(t,L,F,N),M=F?m:g,t.length);t.push(k,L,T,0),no(t,B,M,i),x=!0}j++}x&&(lo(t,!0),Co(t,i,!0)),h&&po(t,!0)}}(a,e,n)}function ha(t,e){var n=gn();ngDevMode&&bt(n[Lt],n[At].bindingStartIndex,"text nodes should be created before any bindings"),ngDevMode&&ngDevMode.rendererCreateTextNode++;var r=Jr(e,n[Ut]),i=ko(t,3,r,null,null);wn(!1),pi(r,i,n)}function ma(t,e){if(e!==Pr){var n=gn();ngDevMode&&kt(n,t+Xt);var r=de(t,n);ngDevMode&&Ot(r,"native element should exist"),ngDevMode&&ngDevMode.rendererSetText++;var i=n[Ut];Wr(i)?i.setValue(r,ue(e)):r.textContent=ue(e)}}function ga(t,e,n){ngDevMode&&bt(t.firstTemplatePass,!0,"Expando block should only be generated on first template pass.");var r=-(e.index-Xt),i=65535&e.providerIndexes,o=t.data.length-i;(t.expandoInstructions||(t.expandoInstructions=[])).push(r,o,n)}function ya(t,e,n,r){var i=_n();(_a(t,i,e,n),ngDevMode&&Ot(i,"previousOrParentTNode"),i&&i.attrs&&function(t,e,n,r){var i=r.initialInputs;(void 0===i||t>=i.length)&&(i=function(t,e,n){var r=n.initialInputs||(n.initialInputs=[]);r[t]=null;var i=n.attrs,o=0;for(;o<i.length;){var a=i[o];if(3===a)break;if(0!==a){var s=e[a],u=i[o+1];if(void 0!==s){var c=r[t]||(r[t]=[]);c.push(s,u)}o+=2}else o+=4}return r}(t,n,r));var o=i[t];if(o)for(var a=0;a<o.length;a+=2)e[o[a]]=o[a+1]}(r,e,n.inputs,i),n.contentQueries&&n.contentQueries(r),ye(n))&&(he(i.index,t)[Vt]=e)}function _a(t,e,n,r){var i=pe(e,t);ngDevMode&&bt(t[Lt],t[At].bindingStartIndex,"directives should be created before any bindings"),ngDevMode&&Mt(xn()),ar(n,t),i&&ar(i,t),null!=r.attributes&&3==e.type&&Xo(i,r.attributes)}function va(t){ngDevMode&&bt(Tn(),!0,"Should only be called in first template pass.");var e=gn()[At];(e.components||(e.components=[])).push(t.index)}function ba(t,e,n){n&&(e.exportAs&&(n[e.exportAs]=t),e.template&&(n[""]=t))}function xa(t,e,n){ngDevMode&&bt(Tn(),!0,"expected firstTemplatePass to be true");var r=t.flags;ngDevMode&&bt(0===r||1===r,!0,"expected node flags to not be initialized"),ngDevMode&&xt(n,t.directiveEnd-t.directiveStart,"Reached the max number of directives"),t.flags=1&r,t.directiveStart=e,t.directiveEnd=e+n,t.providerIndexes=e}function wa(t,e,n,r){t.data.push(n);var i=new ae(r,ye(n),null);t.blueprint.push(i),e.push(i)}function ja(t,e,n){var r=pe(e,t),i=Yo(n.template,n.consts,n.vars,n.directiveDefs,n.pipeDefs,n.viewQuery),o=t[Ht],a=Na(t,e.index,To(t,i,null,n.onPush?8:4,o,t[Ht].createRenderer(r,n)));a[Ft]=e,a[Nt]=t[e.index],t[e.index]=a,Tn()&&va(e)}function Oa(t,e,n,r,i){return[i?-1:0,[],n,null,null,t,r,ui(e,n)]}function Ca(t,e,n,r,i,o,a,s){var u=gn(),c=u[At],l=Sa(t,i||null,o||null);Tn()&&(l.tViews=Ko(-1,e,n,r,c.directiveRegistry,c.pipeRegistry,null)),Go(c,u,a,s);var d=u[Rt],p=_n();ar(pe(p,u),u),d&&(u[Rt]=d.addNode(p)),tn(c,l),wn(!1)}function Ea(t){var e=Sa(t,null,null);Tn()&&(e.tViews=[]),wn(!1)}function Sa(t,e,n){var r=gn();ngDevMode&&bt(r[Lt],r[At].bindingStartIndex,"container nodes should be created before any bindings");var i=t+Xt,o=r[Ut].createComment(ngDevMode?"container":"");ngDevMode&&ngDevMode.rendererCreateComment++;var a=ko(t,0,o,e,n),s=r[i]=Oa(r[i],a,r,o);pi(o,a,r),Na(r,t+Xt,s);var u=r[Rt];return u&&(s[Rt]=u.container()),ngDevMode&&Qe(_n(),0),a}function Ma(t){var e=gn(),n=e[At],r=ce(n.data,t);vn(r),ngDevMode&&Qe(r,0),wn(!0),e[t+Xt][Qt]=0,on(e,n,En())}function Ta(){var t=_n();xn()?wn(!1):(ngDevMode&&Qe(t,2),ngDevMode&&Tt(t),vn(t=t.parent)),ngDevMode&&Qe(t,0);for(var e=gn()[t.index],n=e[Qt];n<e[Zt].length;)ri(e,t,n)}function ka(t,e,n){var r=gn(),i=_n(),o=2===i.type?i.parent:i,a=r[o.index];ngDevMode&&Qe(o,0);var s=function(t,e,n,r){for(var i=t[Zt],o=n;o<i.length;o++){var a=i[o][At].id;if(a===r)return i[o];if(!(a<r))break;ri(t,e,o)}return null}(a,o,a[Qt],t);return s?(wn(!0),Pn(s,s[At].node)):(s=To(r,function(t,e,n,r){var i=gn()[At];ngDevMode&&Qe(r,0);var o=r.tViews;ngDevMode&&Ot(o,"TView expected"),ngDevMode&&bt(Array.isArray(o),!0,"TViews should be in an array"),(t>=o.length||null==o[t])&&(o[t]=Ko(t,null,e,n,i.directiveRegistry,i.pipeRegistry,null));return o[t]}(t,e,n,o),null,4),a[Rt]&&(s[Rt]=a[Rt].createView()),Ao(t,s),Pn(s,s[At].node)),a&&(jn(s)&&ei(s,a,r,a[Qt],-1),a[Qt]++),jn(s)?3:2}function Aa(){var t=gn(),e=t[Ft];jn(t)&&(Mo(t),t[Dt]&=-2),Mo(t),Nn(t[It]),vn(e),wn(!1)}function Da(t){var e=gn();ngDevMode&&kt(e,t);var n=he(t,e);ngDevMode&&Qe(e[At].data[t],3),16==(16&n[Dt])&&12&n[Dt]&&(!function(t){for(var e=t[At],n=t.length;n<e.blueprint.length;n++)t[n]=e.blueprint[n]}(n),Ua(n,n[Vt]))}function Ia(t,e){var n=De(gn())[Ft];if(!n.projection)for(var r=t?t.length+1:1,i=n.projection=new Array(r).fill(null),o=i.slice(),a=n.child;null!==a;){var s=t?wi(a,t,e):0,u=a.next;o[s]?o[s].next=a:(i[s]=a,a.next=null),o[s]=a,a=u}}var Pa=[];function Ra(t,e,n){void 0===e&&(e=0);var r=gn(),i=ko(t,1,null,null,n||null);null===i.projection&&(i.projection=e),wn(!1);for(var o=De(r),a=o[Ft].projection[e],s=o[It],u=-1;a;){if(1===a.type){var c=De(s),l=c[Ft].projection[a.projection];if(l){Pa[++u]=a,Pa[++u]=s,a=l,s=c[It];continue}}else a.flags|=2,gi(a,i,r,s);null===a.next&&s!==o[It]&&(s=Pa[u--],a=Pa[u--]),a=a.next}}function Na(t,e,n){var r=t[At],i=Tn();return t[qt]?t[qt][Pt]=n:i&&(r.childIndex=e),t[qt]=n,n}function Fa(t){for(;t&&!(128&t[Dt]);)t[Dt]|=8,t=t[It];t[Dt]|=8,ngDevMode&&Ot(t[Vt],"rootContext should be defined"),La(t[Vt],1)}function La(t,e){var n,r=0===t.flags;(t.flags|=e,r&&t.clean==So)&&(t.clean=new Promise(function(t){return n=t}),t.scheduler(function(){if(1&t.flags&&(t.flags&=-2,Ba(t)),2&t.flags){t.flags&=-3;var e=t.playerHandler;e&&e.flushPlayers()}t.clean=So,n(null)}))}function Ba(t){for(var e=0;e<t.components.length;e++){var n=t.components[e];Ro(je(n),n)}}function Va(t){za(or(t),t)}function za(t,e){var n=t[Ht];n.begin&&n.begin(),jn(t)&&Ua(t,e),Ua(t,e),n.end&&n.end()}function Ha(t){Ba(t[Vt])}function Ua(t,e){var n,r,i,o,a=t[At],s=Pn(t,t[Ft]),u=a.template,c=a.viewQuery;try{Vo(),o=e,(i=c)&&jn(t)&&i(1,o),u(No(t),e),Mo(t),r=e,(n=c)&&!jn(t)&&n(2,r)}finally{Nn(s)}}function Wa(t){ngDevMode&&Ot(t,"component"),Fa(or(t))}function qa(t){var e=gn();return Fr(e,e[Lt]++,t)?t:Pr}function Ga(t){if(Tn()){var e=gn(),n=e[At];!function(t,e,n){ngDevMode&&bt(Tn(),!0,"Should only be called in first template pass.");var r=t.expandoInstructions,i=r.length;i>=2&&r[i-2]===e.hostBindings?r[i-1]=r[i-1]+n:r.push(e.hostBindings,n)}(n,pn,t),function(t,e,n){ngDevMode&&bt(Tn(),!0,"Should only be called in first template pass.");for(var r=0;r<n;r++)e.push(Pr),t.blueprint.push(Pr),t.data.push(null)}(n,e,t)}}function Ya(t){ngDevMode&&wt(2,t.length,"should have at least 3 values"),ngDevMode&&bt(t.length%2,1,"should have an odd number of values");for(var e=!1,n=gn(),r=n[Lt],i=1;i<t.length;i+=2)Fr(n,r++,t[i])&&(e=!0);if(n[Lt]=r,!e)return Pr;var o=t[0];for(i=1;i<t.length;i+=2)o+=ue(t[i])+t[i+1];return o}function Ka(t,e,n){var r=gn(),i=Fr(r,r[Lt],e);return r[Lt]+=1,i?t+ue(e)+n:Pr}function Xa(t,e,n,r,i){var o=gn(),a=Lr(o,o[Lt],e,r);return o[Lt]+=2,a?t+ue(e)+n+ue(r)+i:Pr}function Qa(t,e,n,r,i,o,a){var s=gn(),u=Br(s,s[Lt],e,r,o);return s[Lt]+=3,u?t+ue(e)+n+ue(r)+i+ue(o)+a:Pr}function Za(t,e,n,r,i,o,a,s,u){var c=gn(),l=Vr(c,c[Lt],e,r,o,s);return c[Lt]+=4,l?t+ue(e)+n+ue(r)+i+ue(o)+a+ue(s)+u:Pr}function $a(t,e,n,r,i,o,a,s,u,c,l){var d=gn(),p=d[Lt],f=Vr(d,p,e,r,o,s);return f=Fr(d,p+4,c)||f,d[Lt]+=5,f?t+ue(e)+n+ue(r)+i+ue(o)+a+ue(s)+u+ue(c)+l:Pr}function Ja(t,e,n,r,i,o,a,s,u,c,l,d,p){var f=gn(),h=f[Lt],m=Vr(f,h,e,r,o,s);return m=Lr(f,h+4,c,d)||m,f[Lt]+=6,m?t+ue(e)+n+ue(r)+i+ue(o)+a+ue(s)+u+ue(c)+l+ue(d)+p:Pr}function ts(t,e,n,r,i,o,a,s,u,c,l,d,p,f,h){var m=gn(),g=m[Lt],y=Vr(m,g,e,r,o,s);return y=Br(m,g+4,c,d,f)||y,m[Lt]+=7,y?t+ue(e)+n+ue(r)+i+ue(o)+a+ue(s)+u+ue(c)+l+ue(d)+p+ue(f)+h:Pr}function es(t,e,n,r,i,o,a,s,u,c,l,d,p,f,h,m,g){var y=gn(),_=y[Lt],v=Vr(y,_,e,r,o,s);return v=Vr(y,_+4,c,d,f,m)||v,y[Lt]+=8,v?t+ue(e)+n+ue(r)+i+ue(o)+a+ue(s)+u+ue(c)+l+ue(d)+p+ue(f)+h+ue(m)+g:Pr}function ns(t,e){var n=gn(),r=n[At],i=t+Xt;i>=r.data.length&&(r.data[i]=null),n[i]=e}function rs(t){return ce(On,t)}function is(t){var e=gn();return ngDevMode&&Ot(e[Yt],"Content QueryList array should be defined if reading a query."),ngDevMode&&kt(e[Yt],t),e[Yt][t]}function os(t){return ce(gn(),t)}function as(t,e){return void 0===e&&(e=Pe.Default),t=Z(t),Yn(_n(),gn(),t,e)}function ss(t){return Gn(_n(),t)}function us(t,e){var n=gn(),r=n[At],i=(n[Yt]||(n[Yt]=[])).push(t);if(Tn()){var o=r.contentQueries||(r.contentQueries=[]);e!==(r.contentQueries.length?r.contentQueries[r.contentQueries.length-2]:-1)&&o.push(e,i-1)}}var cs=So;function ls(t){return t?(void 0===t.inputs&&(t.inputs=sa(t,0)),t.inputs):null}function ds(){return gn()}function ps(t){return t[Bt]||(t[Bt]=[])}function fs(t){return t[At].cleanup||(t[At].cleanup=[])}function hs(t,e){var n=rr(t);if(n){var r=n.native,i=n.lView,o=function(t,e){if(!(e=e||rr(t)))return ngDevMode&&Ii(),null;var n=e.lView,r=Mi(e.nodeIndex,n);return Ai(r)||Di(r)}(r,n),a=hr(i);ki(o,a,r,e,0,t),La(a,2)}else ngDevMode&&Ii()}function ms(t){var e=rr(t);if(!e)return ngDevMode&&Ii(),[];var n=Mi(e.nodeIndex,e.lView),r=n?Ai(n):null;return r?function(t){for(var e=[],n=t[0],r=2;r<n;r+=2){var i=t[r];i&&e.push(i)}for(r=n;r<t.length;r++)e.push(t[r]);return e}(r):[]}var gs="ng",ys=!1;function _s(){ys||(ys=!0,vs("getComponent",dr),vs("getContext",pr),vs("getListeners",jr),vs("getViewComponent",fr),vs("getHostElement",br),vs("getInjector",gr),vs("getRootComponents",mr),vs("getDirectives",yr),vs("getPlayers",ms),vs("markDirty",Wa))}function vs(t,e){var n=H;if(ngDevMode&&Ot(e,"function not defined"),n){var r=n[gs];r||(r=n[gs]={}),r[t]=e}}function bs(t,e){void 0===e&&(e={}),ngDevMode&&_s(),ngDevMode&&Ct(t);var n=e.rendererFactory||qr,r=e.sanitizer||null,i=gt(t);i.type!=t&&(i.type=t);var o,a=i.selectors[0][0],s=Zo(n,e.host||a),u=i.onPush?136:132,c=js(e.scheduler,e.playerHandler),l=n.createRenderer(s,i),d=To(null,Ko(-1,null,1,0,null,null,null),c,u,n,l,void 0,e.injector||null),p=Pn(d,null);try{n.begin&&n.begin(),o=ws(xs(s,i,d,n,l,r),i,d,c,e.hostFeatures||null),Mo(d),d[Dt]&=-2,Mo(d)}finally{Nn(p),n.end&&n.end()}return o}function xs(t,e,n,r,i,o){ln=!1,cn=null,He=0,un=!0;var a=n[At],s=To(n,Yo(e.template,e.consts,e.vars,e.directiveDefs,e.pipeDefs,e.viewQuery),null,e.onPush?8:4,r,i,o),u=ko(0,3,t,null,null);return a.firstTemplatePass&&(qn(zn(u,n),n,e.type),u.flags=1,xa(u,n.length,1),va(u)),s[Nt]=n[Xt],s[Ft]=u,n[Xt]=s}function ws(t,e,n,r,i){var o=n[At],a=function(t,e,n){var r=_n();t.firstTemplatePass&&(n.providersResolver&&n.providersResolver(n),ga(t,r,1),wa(t,e,n,n.factory));var i=Zn(t.data,e,e.length-1,r);return _a(e,r,i,n),i}(o,n,e);if(r.components.push(a),t[Vt]=a,i&&i.forEach(function(t){return t(a,e)}),o.firstTemplatePass&&e.hostBindings){var s=_n();fn(e),e.hostBindings(1,a,s.index-Xt),fn(null)}return a}function js(t,e){return{components:[],scheduler:t||ke,clean:cs,playerHandler:e||null,flags:0}}function Os(t,e){var n=je(t)[At],r=n.data.length-1;Je(r,e.onInit,e.doCheck,n),tn(n,{directiveStart:r,directiveEnd:r+1})}function Cs(t){return(e=t,n=be(e)[Vt],ngDevMode&&Ot(n,"rootContext"),n).clean;var e,n}function Es(t){return"function"==typeof t.template}function Ss(t){for(var e,n=(e=t.type,Object.getPrototypeOf(e.prototype).constructor),i=function(){var e,i,o=void 0;if(Es(t))o=n.ngComponentDef||n.ngDirectiveDef;else{if(n.ngComponentDef)throw new Error("Directives cannot inherit Components");o=n.ngDirectiveDef}var a=n.ngBaseDef;if(a||o){var u=t;u.inputs=Ms(t.inputs),u.declaredInputs=Ms(t.declaredInputs),u.outputs=Ms(t.outputs)}if(a&&(s(t.inputs,a.inputs),s(t.declaredInputs,a.declaredInputs),s(t.outputs,a.outputs)),o){var c=t.hostBindings,l=o.hostBindings;if(l&&(t.hostBindings=c?function(t,e,n){l(t,e,n),c(t,e,n)}:l),Es(t)&&Es(o)){var d=t.viewQuery,p=o.viewQuery;p&&(t.viewQuery=d?function(t,e){p(t,e),d(t,e)}:p)}var f=t.contentQueries,h=o.contentQueries;h&&(t.contentQueries=f?function(t){h(t),f(t)}:h);var m=t.contentQueriesRefresh,g=o.contentQueriesRefresh;g&&(t.contentQueriesRefresh=m?function(t,e){g(t,e),m(t,e)}:g),s(t.inputs,o.inputs),s(t.declaredInputs,o.declaredInputs),s(t.outputs,o.outputs),t.afterContentChecked=t.afterContentChecked||o.afterContentChecked,t.afterContentInit=t.afterContentInit||o.afterContentInit,t.afterViewChecked=t.afterViewChecked||o.afterViewChecked,t.afterViewInit=t.afterViewInit||o.afterViewInit,t.doCheck=t.doCheck||o.doCheck,t.onDestroy=t.onDestroy||o.onDestroy,t.onInit=t.onInit||o.onInit;var y=o.features;if(y)try{for(var _=Object(r.__values)(y),v=_.next();!v.done;v=_.next()){var b=v.value;b&&b.ngInherit&&b(t)}}catch(t){e={error:t}}finally{try{v&&!v.done&&(i=_.return)&&i.call(_)}finally{if(e)throw e.error}}return"break"}var x=n.prototype;x&&(t.afterContentChecked=t.afterContentChecked||x.afterContentChecked,t.afterContentInit=t.afterContentInit||x.afterContentInit,t.afterViewChecked=t.afterViewChecked||x.afterViewChecked,t.afterViewInit=t.afterViewInit||x.afterViewInit,t.doCheck=t.doCheck||x.doCheck,t.onDestroy=t.onDestroy||x.onDestroy,t.onInit=t.onInit||x.onInit),n=Object.getPrototypeOf(n)};n;){if("break"===i())break}}function Ms(t){return t===ot?{}:t===at?[]:t}var Ts="__ngOnChanges_";function ks(t){var e=t.declaredInputs,n=t.inputs,r=t.type.prototype,i=function(t){if(e.hasOwnProperty(t)){for(var i=n[t],o=e[t],a=Ts+i,s=void 0,u=r;!s&&u&&Object.getPrototypeOf(u)!==Object.getPrototypeOf(Object.prototype);)s=Object.getOwnPropertyDescriptor(u,i),u=Object.getPrototypeOf(u);var c=s&&s.get,l=s&&s.set;Object.defineProperty(r,i,{get:c||(l?void 0:function(){return this[a]}),set:function(t){var e=this[Ts];e||(e={},Object.defineProperty(this,Ts,{value:e,writable:!0}));var n=!this.hasOwnProperty(a),r=e[o];r?r.currentValue=t:e[o]=new kr(this[a],t,n),n?Object.defineProperty(this,a,{value:t,writable:!0}):this[a]=t,l&&l.call(this,t)},configurable:!!ngDevMode})}};for(var o in e)i(o);null!=t.onInit&&(t.onInit=As(t.onInit)),t.doCheck=As(t.doCheck)}function As(t){return function(){var e=this[Ts];null!=e&&(this.ngOnChanges(e),this[Ts]=null),t&&t.apply(this)}}function Ds(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e]}ks.ngInherit=!0;var Is="__source",Ps=new Object,Rs=Ps,Ns=new b("INJECTOR"),Fs=function(){function t(){}return t.prototype.get=function(t,e){if(void 0===e&&(e=Ps),e===Ps)throw new Error("NullInjectorError: No provider for "+K(t)+"!");return e},t}(),Ls=function(){function t(){}return t.create=function(t,e){return Array.isArray(t)?new Qs(t,e):new Qs(t.providers,t.parent,t.name||null)},t.THROW_IF_NOT_FOUND=Ps,t.NULL=new Fs,t.ngInjectableDef=g({providedIn:"any",factory:function(){return Ye(Ns)}}),t.__NG_ELEMENT_ID__=function(){return Vs()},t}(),Bs=function(){return t=_n(),new tr(t,gn());var t},Vs=Ds,zs=function(t){return t},Hs=[],Us=zs,Ws=function(){return Array.prototype.slice.call(arguments)},qs=a({provide:String,useValue:a}),Gs="ngTempTokenPath",Ys=Ls.NULL,Ks=/\n/gm,Xs="ɵ",Qs=function(){function t(t,e,n){void 0===e&&(e=Ys),void 0===n&&(n=null),this.parent=e,this.source=n;var r=this._records=new Map;r.set(Ls,{token:Ls,fn:zs,deps:Hs,value:this,useNew:!1}),r.set(Ns,{token:Ns,fn:zs,deps:Hs,value:this,useNew:!1}),function t(e,n){if(n)if((n=Z(n))instanceof Array)for(var r=0;r<n.length;r++)t(e,n[r]);else{if("function"==typeof n)throw tu("Function/Class not supported",n);if(!n||"object"!=typeof n||!n.provide)throw tu("Unexpected provider",n);var i=Z(n.provide),o=function(t){var e=function(t){var e=Hs,n=t.deps;if(n&&n.length){e=[];for(var r=0;r<n.length;r++){var i=6,o=Z(n[r]);if(o instanceof Array)for(var a=0,s=o;a<s.length;a++){var u=s[a];u instanceof Fe||u==Fe?i|=1:u instanceof Be||u==Be?i&=-3:u instanceof Le||u==Le?i&=-5:o=u instanceof Ne?u.token:Z(u)}e.push({token:o,options:i})}}else if(t.useExisting){var o=Z(t.useExisting);e=[{token:o,options:6}]}else if(!(n||qs in t))throw tu("'deps' required",t);return e}(t),n=zs,r=Hs,i=!1,o=Z(t.provide);if(qs in t)r=t.useValue;else if(t.useFactory)n=t.useFactory;else if(t.useExisting);else if(t.useClass)i=!0,n=Z(t.useClass);else{if("function"!=typeof o)throw tu("StaticProvider does not have [useValue|useFactory|useExisting|useClass] or [provide] is not newable",t);i=!0,n=o}return{deps:e,fn:n,useNew:i,value:r}}(n);if(!0===n.multi){var a=e.get(i);if(a){if(a.fn!==Ws)throw Zs(i)}else e.set(i,a={token:n.provide,deps:[],useNew:!1,fn:Ws,value:Hs});i=n,a.deps.push({token:i,options:6})}var s=e.get(i);if(s&&s.fn==Ws)throw Zs(i);e.set(i,o)}}(r,t)}return t.prototype.get=function(t,e,n){void 0===n&&(n=Pe.Default);var r=this._records.get(t);try{return $s(t,r,this._records,this.parent,e,n)}catch(e){var i=e[Gs];throw t[Is]&&i.unshift(t[Is]),e.message=Js("\n"+e.message,i,this.source),e.ngTokenPath=i,e[Gs]=null,e}},t.prototype.toString=function(){var t=[];return this._records.forEach(function(e,n){return t.push(K(n))}),"StaticInjector["+t.join(", ")+"]"},t}();function Zs(t){return tu("Cannot mix multi providers and regular providers",t)}function $s(t,e,n,i,o,a){try{return function(t,e,n,i,o,a){var s,u;if(!e||a&Pe.SkipSelf)a&Pe.Self||(u=i.get(t,o,Pe.Default));else{if((u=e.value)==Us)throw Error(Xs+"Circular dependency");if(u===Hs){e.value=Us;var c=void 0,l=e.useNew,d=e.fn,p=e.deps,f=Hs;if(p.length){f=[];for(var h=0;h<p.length;h++){var m=p[h],g=m.options,y=2&g?n.get(m.token):void 0;f.push($s(m.token,y,n,y||4&g?i:Ys,1&g?null:Ls.THROW_IF_NOT_FOUND,Pe.Default))}}e.value=u=l?new((s=d).bind.apply(s,Object(r.__spread)([void 0],f))):d.apply(c,f)}}return u}(t,e,n,i,o,a)}catch(n){throw n instanceof Error||(n=new Error(n)),(n[Gs]=n[Gs]||[]).unshift(t),e&&e.value==Us&&(e.value=Hs),n}}function Js(t,e,n){void 0===n&&(n=null),t=t&&"\n"===t.charAt(0)&&t.charAt(1)==Xs?t.substr(2):t;var r=K(e);if(e instanceof Array)r=e.map(K).join(" -> ");else if("object"==typeof e){var i=[];for(var o in e)if(e.hasOwnProperty(o)){var a=e[o];i.push(o+":"+("string"==typeof a?JSON.stringify(a):K(a)))}r="{"+i.join(", ")+"}"}return"StaticInjectorError"+(n?"("+n+")":"")+"["+r+"]: "+t.replace(Ks,"\n  ")}function tu(t,e){return new Error(Js(t,e))}var eu=new b("The presence of this token marks an injector as being the root injector."),nu={},ru={},iu=[],ou=void 0;function au(){return void 0===ou&&(ou=new Fs),ou}function su(t,e,n){return void 0===e&&(e=null),void 0===n&&(n=null),e=e||au(),new uu(t,n,e)}var uu=function(){function t(t,e,n){var r=this;this.parent=n,this.records=new Map,this.injectorDefTypes=new Set,this.onDestroy=new Set,this.destroyed=!1;var i=[];pu([t],function(t){return r.processInjectorType(t,[],i)}),e&&pu(e,function(n){return r.processProvider(n,t,e)}),this.records.set(Ns,du(void 0,this)),this.isRootInjector=this.records.has(eu),this.injectorDefTypes.forEach(function(t){return r.get(t)})}return t.prototype.destroy=function(){this.assertNotDestroyed(),this.destroyed=!0;try{this.onDestroy.forEach(function(t){return t.ngOnDestroy()})}finally{this.records.clear(),this.onDestroy.clear(),this.injectorDefTypes.clear()}},t.prototype.get=function(t,e,n){void 0===e&&(e=Rs),void 0===n&&(n=Pe.Default),this.assertNotDestroyed();var r,i=We(this);try{if(!(n&Pe.SkipSelf)){var o=this.records.get(t);if(void 0===o){var a=("function"==typeof(r=t)||"object"==typeof r&&r instanceof b)&&_(t);a&&this.injectableDefInScope(a)&&(o=du(cu(t),nu),this.records.set(t,o))}if(void 0!==o)return this.hydrate(t,o)}return(n&Pe.Self?au():this.parent).get(t,e)}finally{We(i)}},t.prototype.assertNotDestroyed=function(){if(this.destroyed)throw new Error("Injector has already been destroyed.")},t.prototype.processInjectorType=function(t,e,n){var r=this;if(t=Z(t)){var i=v(t),o=null==i&&t.ngModule||void 0,a=void 0===o?t:o;if(ngDevMode&&-1!==e.indexOf(a)){var s=K(a);throw new Error("Circular dependency in DI detected for type "+s+". Dependency path: "+e.map(function(t){return K(t)}).join(" > ")+" > "+s+".")}var u=-1!==n.indexOf(a),c=void 0!==o&&t.providers||iu;if(void 0!==o&&(i=v(o)),null!=i){if(this.injectorDefTypes.add(a),this.records.set(a,du(i.factory,nu)),null!=i.imports&&!u){ngDevMode&&e.push(a),n.push(a);try{pu(i.imports,function(t){return r.processInjectorType(t,e,n)})}finally{ngDevMode&&e.pop()}}var l=i.providers;if(null!=l&&!u){var d=t;pu(l,function(t){return r.processProvider(t,d,l)})}var p=t.ngModule;pu(c,function(t){return r.processProvider(t,p,c)})}}},t.prototype.processProvider=function(t,e,n){var r,i,o=hu(t=Z(t))?t:Z(t&&t.provide),a=(i=lu(r=t,e,n),fu(r)?du(void 0,r.useValue):du(i,nu));if(hu(t)||!0!==t.multi){var s=this.records.get(o);if(s&&void 0!==s.multi)throw new Error("Mixed multi-provider for "+K(o))}else{var u=this.records.get(o);if(u){if(void 0===u.multi)throw new Error("Mixed multi-provider for "+o+".")}else(u=du(void 0,nu,!0)).factory=function(){return Xe(u.multi)},this.records.set(o,u);o=t,u.multi.push(t)}this.records.set(o,a)},t.prototype.hydrate=function(t,e){if(e.value===ru)throw new Error("Cannot instantiate cyclic dependency! "+K(t));var n;return e.value===nu&&(e.value=ru,e.value=e.factory()),"object"==typeof e.value&&e.value&&("object"==typeof(n=e.value)&&null!=n&&n.ngOnDestroy&&"function"==typeof n.ngOnDestroy)&&this.onDestroy.add(e.value),e.value},t.prototype.injectableDefInScope=function(t){return!!t.providedIn&&("string"==typeof t.providedIn?"any"===t.providedIn||"root"===t.providedIn&&this.isRootInjector:this.injectorDefTypes.has(t.providedIn))},t}();function cu(t){var e=_(t);if(null===e){var n=v(t);if(null!==n)return n.factory;if(t instanceof b)throw new Error("Token "+K(t)+" is missing an ngInjectableDef definition.");if(t instanceof Function){var r=t.length;if(r>0){var i=new Array(r).fill("?");throw new Error("Can't resolve all parameters for "+K(t)+": ("+i.join(", ")+").")}return function(){return new t}}throw new Error("unreachable")}return e.factory}function lu(t,e,n){var i,o,a=void 0;if(hu(t))return cu(Z(t));if(fu(t))a=function(){return Z(t.useValue)};else if(o=t,o&&o.useExisting)a=function(){return Ye(Z(t.useExisting))};else if(i=t,i&&i.useFactory)a=function(){return t.useFactory.apply(t,Object(r.__spread)(Xe(t.deps||[])))};else{var s=Z(t&&(t.useClass||t.provide));if(!s){var u="";if(e&&n)u=" - only instances of Provider and Type are allowed, got: ["+n.map(function(e){return e==t?"?"+t+"?":"..."}).join(", ")+"]";throw new Error("Invalid provider for the NgModule '"+K(e)+"'"+u)}if(!t.deps)return cu(s);a=function(){return new(s.bind.apply(s,Object(r.__spread)([void 0],Xe(t.deps))))}}return a}function du(t,e,n){return void 0===n&&(n=!1),{factory:t,value:e,multi:n?[]:void 0}}function pu(t,e){t.forEach(function(t){return Array.isArray(t)?pu(t,e):e(t)})}function fu(t){return t&&"object"==typeof t&&qs in t}function hu(t){return"function"==typeof t}function mu(t,e,n,r,i){if(t=Z(t),Array.isArray(t))for(var o=0;o<t.length;o++)mu(t[o],e,n,r,i);else{var a=gn(),s=hu(t)?t:Z(t.provide),u=lu(t),c=_n(),l=65535&c.providerIndexes,d=c.directiveStart,p=c.providerIndexes>>16;if(hu(t)||!t.multi){var f=new ae(u,i,as),h=yu(s,e,i?l:l+p,d);-1==h?(qn(zn(c,a),a,s),e.push(s),c.directiveStart++,c.directiveEnd++,i&&(c.providerIndexes+=65536),n.push(f),a.push(f)):(n[h]=f,a[h]=f)}else{var m=yu(s,e,l+p,d),g=yu(s,e,l,l+p),y=m>=0&&n[m],_=g>=0&&n[g];if(i&&!_||!i&&!y){qn(zn(c,a),a,s);f=(v=i?vu:_u,b=n.length,w=r,j=u,(O=new ae(v,x=i,as)).multi=[],O.index=b,O.componentProviders=0,gu(O,j,w&&!x),O);!i&&_&&(n[g].providerFactory=f),e.push(s),c.directiveStart++,c.directiveEnd++,i&&(c.providerIndexes+=65536),n.push(f),a.push(f)}else gu(n[i?g:m],u,!i&&r);!i&&r&&_&&n[g].componentProviders++}}var v,b,x,w,j,O}function gu(t,e,n){t.multi.push(e),n&&t.componentProviders++}function yu(t,e,n,r){for(var i=n;i<r;i++)if(e[i]===t)return i;return-1}function _u(t,e,n,r){return bu(this.multi,[])}function vu(t,e,n,r){var i,o=this.multi;if(this.providerFactory){var a=this.providerFactory.componentProviders,s=Zn(e,n,this.providerFactory.index,r);bu(o,i=s.slice(0,a));for(var u=a;u<s.length;u++)i.push(s[u])}else bu(o,i=[]);return i}function bu(t,e){for(var n=0;n<t.length;n++){var r=t[n];e.push(r())}return e}function xu(t,e){return void 0===e&&(e=[]),function(n){n.providersResolver=function(n){return function(t,e,n){var r=gn()[At];if(r.firstTemplatePass){var i=ye(t);mu(n,r.data,r.blueprint,i,!0),mu(e,r.data,r.blueprint,i,!1)}}(n,t,e)}}}var wu=function(){return function(){}}(),ju=function(){return function(){}}();function Ou(t){var e=Error("No component factory found for "+K(t)+". Did you add it to @NgModule.entryComponents?");return e[Mu]=t,e}var Cu,Eu,Su,Mu="ngComponent",Tu=function(){function t(){}return t.prototype.resolveComponentFactory=function(t){throw Ou(t)},t}(),ku=function(){function t(){}return t.NULL=new Tu,t}(),Au=function(){function t(t,e,n){this._parent=e,this._ngModule=n,this._factories=new Map;for(var r=0;r<t.length;r++){var i=t[r];this._factories.set(i.componentType,i)}}return t.prototype.resolveComponentFactory=function(t){var e=this._factories.get(t);if(!e&&this._parent&&(e=this._parent.resolveComponentFactory(t)),!e)throw Ou(t);return new Du(e,this._ngModule)},t}(),Du=function(t){function e(e,n){var r=t.call(this)||this;return r.factory=e,r.ngModule=n,r.selector=e.selector,r.componentType=e.componentType,r.ngContentSelectors=e.ngContentSelectors,r.inputs=e.inputs,r.outputs=e.outputs,r}return Object(r.__extends)(e,t),e.prototype.create=function(t,e,n,r){return this.factory.create(t,e,n,r||this.ngModule)},e}(ju),Iu=function(){return function(){}}(),Pu=function(){return function(){}}(),Ru=function(){function t(t,e,n){this._context=e,this._componentIndex=n,this._appRef=null,this._viewContainerRef=null,this._tViewNode=null,this._lView=t}return Object.defineProperty(t.prototype,"rootNodes",{get:function(){if(null==this._lView[Nt]){var t=this._lView[Ft];return function t(e,n,r){var i=n.child;for(;i;)r.push(pe(i,e)),4===i.type&&t(e,i,r),i=i.next;return r}(this._lView,t,[])}return[]},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"context",{get:function(){return this._context?this._context:this._lookUpContext()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"destroyed",{get:function(){return 64==(64&this._lView[Dt])},enumerable:!0,configurable:!0}),t.prototype.destroy=function(){if(this._appRef)this._appRef.detachView(this);else if(this._viewContainerRef){var t=this._viewContainerRef.indexOf(this);t>-1&&this._viewContainerRef.detach(t),this._viewContainerRef=null}oi(this._lView)},t.prototype.onDestroy=function(t){var e,n;e=this._lView,n=t,ps(e).push(n),e[At].firstTemplatePass&&fs(e).push(e[Bt].length-1,null)},t.prototype.markForCheck=function(){Fa(this._lView)},t.prototype.detach=function(){this._lView[Dt]&=-17},t.prototype.reattach=function(){this._lView[Dt]|=16},t.prototype.detectChanges=function(){za(this._lView,this.context)},t.prototype.checkNoChanges=function(){!function(t){Sn(!0);try{Va(t)}finally{Sn(!1)}}(this.context)},t.prototype.attachToViewContainerRef=function(t){if(this._appRef)throw new Error("This view is already attached directly to the ApplicationRef!");this._viewContainerRef=t},t.prototype.detachFromAppRef=function(){this._appRef=null},t.prototype.attachToAppRef=function(t){if(this._viewContainerRef)throw new Error("This view is already attached to a ViewContainer!");this._appRef=t},t.prototype._lookUpContext=function(){return this._context=this._lView[It][this._componentIndex]},t}(),Nu=function(t){function e(e){var n=t.call(this,e,null,-1)||this;return n._view=e,n}return Object(r.__extends)(e,t),e.prototype.detectChanges=function(){Ha(this._view)},e.prototype.checkNoChanges=function(){!function(t){Sn(!0);try{Ha(t)}finally{Sn(!1)}}(this._view)},Object.defineProperty(e.prototype,"context",{get:function(){return null},enumerable:!0,configurable:!0}),e}(Ru);function Fu(t,e,n){return Cu||(Cu=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e}(t)),new Cu(pe(e,n))}function Lu(t,e,n,i){if(Eu||(Eu=function(t){function e(e,n,r,i,o,a){var s=t.call(this)||this;return s._declarationParentView=e,s.elementRef=n,s._tView=r,s._renderer=i,s._queries=o,s._injectorIndex=a,s}return Object(r.__extends)(e,t),e.prototype.createEmbeddedView=function(t,e,n,r,i){var o=function(t,e,n,r,i,o){var a=xn(),s=_n();wn(!0),vn(null);var u=To(n,t,e,4);return u[Kt]=n,i&&(u[Rt]=i.createView()),Ao(-1,u),t.firstTemplatePass&&(t.node.injectorIndex=o),wn(a),vn(s),u}(this._tView,t,this._declarationParentView,this._renderer,this._queries,this._injectorIndex);e&&ei(o,e,r,i,n.index),Io(o,this._tView,t);var a=new Ru(o,t,-1);return a._tViewNode=o[Ft],a},e}(t)),0===n.type){var o=i[n.index];return ngDevMode&&Ot(n.tViews,"TView must be allocated"),new Eu(i,Fu(e,n,i),n.tViews,gn()[Ut],o[Rt],n.injectorIndex)}return null}function Bu(t,e,n){if(ge(t)){var r=t.directiveStart,i=he(t.index,e);return new Ru(i,n,r)}if(3===t.type){var o=De(e);return new Ru(o,o[Vt],-1)}return null}var Vu,zu,Hu=function(){function t(t){this.nativeElement=t}return t.__NG_ELEMENT_ID__=function(){return Wu(t)},t}(),Uu=function(t){return Fu(t,_n(),gn())},Wu=Ds,qu=function(){return function(t,e,n,r,i,o){this.id=t,this.templateUrl=e,this.slotCount=n,this.encapsulation=r,this.styles=i,this.animations=o}}(),Gu=function(){return function(){}}(),Yu=function(){return function(){}}(),Ku=(new b("Renderer2Interceptor"),function(){return function(){}}()),Xu=function(){return function(){}}();(zu=Vu||(Vu={}))[zu.Important=1]="Important",zu[zu.DashCase=2]="DashCase";var Qu,Zu,$u=function(){function t(){}return t.__NG_ELEMENT_ID__=function(){return tc()},t}(),Ju=function(){return function(t){var e=t[Ut];if(Wr(e))return e;throw new Error("Cannot inject Renderer2 when the application uses Renderer3!")}(gn())},tc=Ds;(Zu=Qu||(Qu={}))[Zu.NONE=0]="NONE",Zu[Zu.HTML=1]="HTML",Zu[Zu.STYLE=2]="STYLE",Zu[Zu.SCRIPT=3]="SCRIPT",Zu[Zu.URL=4]="URL",Zu[Zu.RESOURCE_URL=5]="RESOURCE_URL";var ec=function(){return function(){}}(),nc=function(){return function(t){this.full=t,this.major=t.split(".")[0],this.minor=t.split(".")[1],this.patch=t.split(".").slice(2).join(".")}}(),rc=new nc("7.2.0"),ic=function(t){function e(e){var n=t.call(this)||this;return n.ngModule=e,n}return Object(r.__extends)(e,t),e.prototype.resolveComponentFactory=function(t){ngDevMode&&Ct(t);var e=gt(t);return new cc(e,this.ngModule)},e}(ku);function oc(t){var e=[];for(var n in t)if(t.hasOwnProperty(n)){var r=t[n];e.push({propName:r,templateName:n})}return e}var ac=new b("ROOT_CONTEXT_TOKEN",{providedIn:"root",factory:function(){return js(Ye(sc))}}),sc=new b("SCHEDULER_TOKEN",{providedIn:"root",factory:function(){return ke}}),uc={};var cc=function(t){function e(e,n){var r=t.call(this)||this;return r.componentDef=e,r.ngModule=n,r.componentType=e.type,r.selector=e.selectors[0][0],r.ngContentSelectors=[],r}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"inputs",{get:function(){return oc(this.componentDef.inputs)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"outputs",{get:function(){return oc(this.componentDef.outputs)},enumerable:!0,configurable:!0}),e.prototype.create=function(t,e,n,r){var i,o,a=void 0===n,s=(r=r||this.ngModule)?(i=t,o=r.injector,{get:function(t,e){var n=i.get(t,uc);return n!==uc||e===uc?n:o.get(t,e)}}):t,u=s.get(Xu,qr),c=s.get(ec,null),l=a?qo(this.selector,u.createRenderer(null,this.componentDef)):Zo(u,n),d=this.componentDef.onPush?136:132,p=a?js():s.get(ac),f=u.createRenderer(l,this.componentDef);n&&l&&(ngDevMode&&ngDevMode.rendererSetAttribute++,Wr(f)?f.setAttribute(l,"ng-version",rc.full):l.setAttribute("ng-version",rc.full));var h,m,g=To(null,Ko(-1,null,1,0,null,null,null),p,d,u,f,c,s),y=Pn(g,null);try{u.begin&&u.begin();var _=xs(l,this.componentDef,g,u,f);if(m=fe(0,g),e)for(var v=0,b=g[At],x=m.projection=[],w=0;w<e.length;w++){for(var j=e[w],O=null,C=null,E=0;E<j.length;E++){b.firstTemplatePass&&(b.expandoStartIndex++,b.blueprint.splice(++v+Xt,0,null),b.data.splice(v+Xt,0,null),g.splice(v+Xt,0,null));var S=ko(v,3,j[E],null,null);C?C.next=S:O=S,C=S}x.push(O)}h=ws(_,this.componentDef,g,p,[Os]),Na(g,Xt,_),Mo(g)}finally{Nn(y),u.end&&u.end()}var M=new lc(this.componentType,h,Fu(Hu,m,g),g,m);return a&&(M.hostView._tViewNode.child=m),M},e}(ju),lc=(new ic,function(t){function e(e,n,r,i,o){var a=t.call(this)||this;return a.location=r,a._rootLView=i,a._tNode=o,a.destroyCbs=[],a.instance=n,a.hostView=a.changeDetectorRef=new Nu(i),a.hostView._tViewNode=Ao(-1,i),a.componentType=e,a}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"injector",{get:function(){return new tr(this._tNode,this._rootLView)},enumerable:!0,configurable:!0}),e.prototype.destroy=function(){ngDevMode&&Ot(this.destroyCbs,"NgModule already destroyed"),this.destroyCbs.forEach(function(t){return t()}),this.destroyCbs=null,this.hostView.destroy()},e.prototype.onDestroy=function(t){ngDevMode&&Ot(this.destroyCbs,"NgModule already destroyed"),this.destroyCbs.push(t)},e}(wu)),dc=!0,pc=!1;function fc(){return pc=!0,dc}function hc(){if(pc)throw new Error("Cannot enable prod mode after platform setup.");dc=!1}var mc=function(){function t(t){if(this.defaultDoc=t,this.inertDocument=this.defaultDoc.implementation.createHTMLDocument("sanitization-inert"),this.inertBodyElement=this.inertDocument.body,null==this.inertBodyElement){var e=this.inertDocument.createElement("html");this.inertDocument.appendChild(e),this.inertBodyElement=this.inertDocument.createElement("body"),e.appendChild(this.inertBodyElement)}this.inertBodyElement.innerHTML='<svg><g onload="this.parentNode.remove()"></g></svg>',!this.inertBodyElement.querySelector||this.inertBodyElement.querySelector("svg")?(this.inertBodyElement.innerHTML='<svg><p><style><img src="</style><img src=x onerror=alert(1)//">',this.inertBodyElement.querySelector&&this.inertBodyElement.querySelector("svg img")&&function(){try{return!!window.DOMParser}catch(t){return!1}}()?this.getInertBodyElement=this.getInertBodyElement_DOMParser:this.getInertBodyElement=this.getInertBodyElement_InertDocument):this.getInertBodyElement=this.getInertBodyElement_XHR}return t.prototype.getInertBodyElement_XHR=function(t){t="<body><remove></remove>"+t+"</body>";try{t=encodeURI(t)}catch(t){return null}var e=new XMLHttpRequest;e.responseType="document",e.open("GET","data:text/html;charset=utf-8,"+t,!1),e.send(void 0);var n=e.response.body;return n.removeChild(n.firstChild),n},t.prototype.getInertBodyElement_DOMParser=function(t){t="<body><remove></remove>"+t+"</body>";try{var e=(new window.DOMParser).parseFromString(t,"text/html").body;return e.removeChild(e.firstChild),e}catch(t){return null}},t.prototype.getInertBodyElement_InertDocument=function(t){var e=this.inertDocument.createElement("template");return"content"in e?(e.innerHTML=t,e):(this.inertBodyElement.innerHTML=t,this.defaultDoc.documentMode&&this.stripCustomNsAttrs(this.inertBodyElement),this.inertBodyElement)},t.prototype.stripCustomNsAttrs=function(t){for(var e=t.attributes,n=e.length-1;0<n;n--){var r=e.item(n).name;"xmlns:ns1"!==r&&0!==r.indexOf("ns1:")||t.removeAttribute(r)}for(var i=t.firstChild;i;)i.nodeType===Node.ELEMENT_NODE&&this.stripCustomNsAttrs(i),i=i.nextSibling},t}();var gc=/^(?:(?:https?|mailto|ftp|tel|file):|[^&:/?#]*(?:[/?#]|$))/gi,yc=/^data:(?:image\/(?:bmp|gif|jpeg|jpg|png|tiff|webp)|video\/(?:mpeg|mp4|ogg|webm)|audio\/(?:mp3|oga|ogg|opus));base64,[a-z0-9+\/]+=*$/i;function _c(t){return(t=String(t)).match(gc)||t.match(yc)?t:(fc()&&console.warn("WARNING: sanitizing unsafe URL value "+t+" (see http://g.co/ng/security#xss)"),"unsafe:"+t)}function vc(t){return(t=String(t)).split(",").map(function(t){return _c(t.trim())}).join(", ")}function bc(t){var e,n,i={};try{for(var o=Object(r.__values)(t.split(",")),a=o.next();!a.done;a=o.next()){i[a.value]=!0}}catch(t){e={error:t}}finally{try{a&&!a.done&&(n=o.return)&&n.call(o)}finally{if(e)throw e.error}}return i}function xc(){for(var t,e,n=[],i=0;i<arguments.length;i++)n[i]=arguments[i];var o={};try{for(var a=Object(r.__values)(n),s=a.next();!s.done;s=a.next()){var u=s.value;for(var c in u)u.hasOwnProperty(c)&&(o[c]=!0)}}catch(e){t={error:e}}finally{try{s&&!s.done&&(e=a.return)&&e.call(a)}finally{if(t)throw t.error}}return o}var wc,jc=bc("area,br,col,hr,img,wbr"),Oc=bc("colgroup,dd,dt,li,p,tbody,td,tfoot,th,thead,tr"),Cc=bc("rp,rt"),Ec=xc(Cc,Oc),Sc=xc(Oc,bc("address,article,aside,blockquote,caption,center,del,details,dialog,dir,div,dl,figure,figcaption,footer,h1,h2,h3,h4,h5,h6,header,hgroup,hr,ins,main,map,menu,nav,ol,pre,section,summary,table,ul")),Mc=xc(Cc,bc("a,abbr,acronym,audio,b,bdi,bdo,big,br,cite,code,del,dfn,em,font,i,img,ins,kbd,label,map,mark,picture,q,ruby,rp,rt,s,samp,small,source,span,strike,strong,sub,sup,time,track,tt,u,var,video")),Tc=xc(jc,Sc,Mc,Ec),kc=bc("background,cite,href,itemtype,longdesc,poster,src,xlink:href"),Ac=bc("srcset"),Dc=bc("abbr,accesskey,align,alt,autoplay,axis,bgcolor,border,cellpadding,cellspacing,class,clear,color,cols,colspan,compact,controls,coords,datetime,default,dir,download,face,headers,height,hidden,hreflang,hspace,ismap,itemscope,itemprop,kind,label,lang,language,loop,media,muted,nohref,nowrap,open,preload,rel,rev,role,rows,rowspan,rules,scope,scrolling,shape,size,sizes,span,srclang,start,summary,tabindex,target,title,translate,type,usemap,valign,value,vspace,width"),Ic=xc(kc,Ac,Dc),Pc=function(){function t(){this.sanitizedSomething=!1,this.buf=[]}return t.prototype.sanitizeChildren=function(t){for(var e=t.firstChild,n=!0;e;)if(e.nodeType===Node.ELEMENT_NODE?n=this.startElement(e):e.nodeType===Node.TEXT_NODE?this.chars(e.nodeValue):this.sanitizedSomething=!0,n&&e.firstChild)e=e.firstChild;else for(;e;){e.nodeType===Node.ELEMENT_NODE&&this.endElement(e);var r=this.checkClobberedElement(e,e.nextSibling);if(r){e=r;break}e=this.checkClobberedElement(e,e.parentNode)}return this.buf.join("")},t.prototype.startElement=function(t){var e=t.nodeName.toLowerCase();if(!Tc.hasOwnProperty(e))return this.sanitizedSomething=!0,!1;this.buf.push("<"),this.buf.push(e);for(var n=t.attributes,r=0;r<n.length;r++){var i=n.item(r),o=i.name,a=o.toLowerCase();if(Ic.hasOwnProperty(a)){var s=i.value;kc[a]&&(s=_c(s)),Ac[a]&&(s=vc(s)),this.buf.push(" ",o,'="',Fc(s),'"')}else this.sanitizedSomething=!0}return this.buf.push(">"),!0},t.prototype.endElement=function(t){var e=t.nodeName.toLowerCase();Tc.hasOwnProperty(e)&&!jc.hasOwnProperty(e)&&(this.buf.push("</"),this.buf.push(e),this.buf.push(">"))},t.prototype.chars=function(t){this.buf.push(Fc(t))},t.prototype.checkClobberedElement=function(t,e){if(e&&(t.compareDocumentPosition(e)&Node.DOCUMENT_POSITION_CONTAINED_BY)===Node.DOCUMENT_POSITION_CONTAINED_BY)throw new Error("Failed to sanitize html because the element is clobbered: "+t.outerHTML);return e},t}(),Rc=/[\uD800-\uDBFF][\uDC00-\uDFFF]/g,Nc=/([^\#-~ |!])/g;function Fc(t){return t.replace(/&/g,"&amp;").replace(Rc,function(t){return"&#"+(1024*(t.charCodeAt(0)-55296)+(t.charCodeAt(1)-56320)+65536)+";"}).replace(Nc,function(t){return"&#"+t.charCodeAt(0)+";"}).replace(/</g,"&lt;").replace(/>/g,"&gt;")}function Lc(t,e){var n=null;try{wc=wc||new mc(t);var r=e?String(e):"";n=wc.getInertBodyElement(r);var i=5,o=r;do{if(0===i)throw new Error("Failed to sanitize html because the input is unstable");i--,r=o,o=n.innerHTML,n=wc.getInertBodyElement(r)}while(r!==o);var a=new Pc,s=a.sanitizeChildren(Bc(n)||n);return fc()&&a.sanitizedSomething&&console.warn("WARNING: sanitizing HTML stripped some content (see http://g.co/ng/security#xss)."),s}finally{if(n)for(var u=Bc(n)||n;u.firstChild;)u.removeChild(u.firstChild)}}function Bc(t){return"content"in t&&((e=t).nodeType===Node.ELEMENT_NODE&&"TEMPLATE"===e.nodeName)?t.content:null;var e}var Vc={marker:"element"},zc={marker:"comment"},Hc="�",Uc=/^\s*(�\d+:?\d*�)\s*,\s*(select|plural)\s*,/,Wc=/�\/?\*(\d+:\d+)�/gi,qc=/�(\/?[#*]\d+):?\d*�/gi,Gc=/�(\d+):?\d*�/gi,Yc=/({\s*�\d+:?\d*�\s*,\s*\S{6}\s*,[\s\S]*})/gi,Kc=/\[(�.+?�?)\]/g,Xc=/({\s*)(VAR_(PLURAL|SELECT)(_\d+)?)(\s*,)/g,Qc=/�I18N_EXP_(ICU(_\d+)?)�/g;function Zc(t){if(!t)return[];var e,n=0,r=[],i=[],o=/[{}]/g;for(o.lastIndex=0;e=o.exec(t);){var a=e.index;if("}"==e[0]){if(r.pop(),0==r.length){var s=t.substring(n,a);Uc.test(s)?i.push($c(s)):s&&i.push(s),n=a+1}}else{if(0==r.length){var u=t.substring(n,a);i.push(u),n=a+1}r.push("{")}}var c=t.substring(n);return""!=c&&i.push(c),i}function $c(t){for(var e=[],n=[],r=1,i=0,o=Zc(t=t.replace(Uc,function(t,e,n){return r="select"===n?0:1,i=parseInt(e.substr(1),10),""})),a=0;a<o.length;){var s=o[a++].trim();1===r&&(s=s.replace(/\s*(?:=)?(\w+)\s*/,"$1")),s.length&&e.push(s);var u=Zc(o[a++]);u.length&&n.push(u)}return jt(e.indexOf("other"),-1,'Missing key "other" in ICU statement.'),{type:r,mainBinding:i,cases:e,values:n}}function Jc(t){for(var e,n,r="",i=0,o=!1;null!==(e=Wc.exec(t));)o?e[0]===Hc+"/*"+n+Hc&&(i=e.index,o=!1):(r+=t.substring(i,e.index+e[0].length),n=e[1],o=!0);return ngDevMode&&bt(o,!1,'Tag mismatch: unable to find the end of the sub-template in the translation "'+t+'"'),r+=t.substr(i)}function tl(t,e,n,r){void 0===r&&(r=null);for(var i=[null,null],o=t.split(Gc),a=0,s=0;s<o.length;s++){var u=o[s];if(1&s){var c=parseInt(u,10);i.push(-1-c),a|=il(c)}else""!==u&&i.push(u)}return i.push(e<<2|(n?1:0)),n&&i.push(n,r),i[0]=a,i[1]=i.length-2,i}function el(t,e){var n;void 0===e&&(e=0),e|=il(t.mainBinding);for(var r=0;r<t.values.length;r++)for(var i=t.values[r],o=0;o<i.length;o++){var a=i[o];if("string"==typeof a)for(;n=Gc.exec(a);)e|=il(parseInt(n[1],10));else e=el(a,e)}return e}var nl=[],rl=-1;function il(t){return 1<<Math.min(t,31)}var ol=[];function al(t,e,n){var r=gn()[At];ngDevMode&&Ot(r,"tView should be defined"),nl[++rl]=t,r.firstTemplatePass&&null===r.data[t+Xt]&&function(t,e,n,r){var i=gn(),o=t.blueprint.length-Xt,a=_n(),s=xn()?_n():a&&a.parent,u=s&&s!==i[Ft]?s.index-Xt:e,c=0;ol[c]=u;var l=[];e>0&&a!==s&&l.push(a.index<<3|0);for(var d=[],p=[],f=function(t,e){if("number"!=typeof e)return Jc(t);var n=t.indexOf(":"+e+Hc)+2+e.toString().length,r=t.search(new RegExp(Hc+"\\/\\*\\d+:"+e+Hc));return Jc(t.substring(n,r))}(n,r).split(qc),h=0;h<f.length;h++){var m=f[h];if(1&h)if("/"===m.charAt(0)){if("#"===m.charAt(1)){var g=parseInt(m.substr(2),10);u=ol[--c],l.push(g<<3|5)}}else{var g=parseInt(m.substr(1),10);l.push(g<<3|0,u<<17|1),"#"===m.charAt(0)&&(ol[++c]=u=g)}else for(var y=m.split(Yc),_=0;_<y.length;_++)if(m=y[_],1&_){Do(i);var v=t.blueprint.length-1-Xt;l.push(zc,ngDevMode?"ICU "+v:"",u<<17|1);var b=$c(m.substr(1,m.length-2)),x=el(b);wl(p,b,v,v);var w=p.length-1;d.push(il(b.mainBinding),3,-1-b.mainBinding,v<<2|2,w,x,2,v<<2|3,w)}else if(""!==m){var j=m.match(Gc);Do(i),l.push(j?"":m,u<<17|1),j&&Ae(tl(m,t.blueprint.length-1-Xt),d)}}var O={vars:t.blueprint.length-Xt-o,expandoStartIndex:o,create:l,update:d,icus:p.length?p:null};t.data[e+Xt]=O}(r,t,e,n)}function sl(t,e,n){ngDevMode&&ngDevMode.rendererMoveNode++;var r=gn();n||(n=e),n===e&&t!==e.child?(t.next=e.child,e.child=t):n!==e&&t!==n.next?(t.next=n.next,n.next=t):t.next=null,e!==r[Ft]&&(t.parent=e),pi(pe(t,r),t,r);var i=r[t.index];return 0!==t.type&&_e(i)&&pi(i[$t],t,r),t}function ul(t,e){var n={},r=t.replace(Kc,function(t,e){if(n[e]||(n[e]=e.split("|")),!n[e].length)throw new Error("i18n postprocess: unmatched placeholder - "+e);return n[e].shift()});if(Object.keys(n).some(function(t){return!!n[t].length}))throw new Error("i18n postprocess: unmatched values - "+JSON.stringify(n));return Object.keys(e).length?r=(r=r.replace(Xc,function(t,n,r,i,o,a){return e.hasOwnProperty(r)?""+n+e[r]+a:t})).replace(Qc,function(t,n){if(e.hasOwnProperty(n)){var r=e[n];if(!r.length)throw new Error("i18n postprocess: unmatched ICU - "+t+" with key: "+n);return r.shift()}return t}):r}function cl(){var t=gn()[At];ngDevMode&&Ot(t,"tView should be defined"),function(t){var e=gn();ngDevMode&&bt(e[Lt],e[At].bindingStartIndex,"i18nEnd should be called before any binding");var n=nl[rl--],r=t.data[n+Xt];ngDevMode&&Ot(r,"You should call i18nStart before i18nEnd");for(var i=_n(),o=ll(n,r.create,r.expandoStartIndex,e),a=n+1;a<=i.index-Xt;a++)-1===o.indexOf(a)&&dl(a,e)}(t)}function ll(t,e,n,r){for(var i=gn()[Ut],o=null,a=null,s=[],u=0;u<e.length;u++){var c=e[u];if("string"==typeof c){var l=Jr(c,i);ngDevMode&&ngDevMode.rendererCreateTextNode++,a=o,o=ko(n++,3,l,null,null),wn(!1)}else if("number"==typeof c)switch(7&c){case 1:var d=c>>>17,p=void 0;p=d===t?r[Ft]:fe(d,r),ngDevMode&&Ot(o,"You need to create or select a node before you can insert it into the DOM"),a=sl(o,p,a),p.next=null;break;case 0:var f=c>>>3;s.push(f),a=o,(o=fe(f,r))&&(vn(o),3===o.type&&wn(!0));break;case 5:a=o=fe(c>>>3,r),vn(o),wn(!1);break;case 4:ta(c>>>3,e[++u],e[++u]);break;default:throw new Error('Unable to determine the type of mutate operation for "'+c+'"')}else switch(c){case zc:var h=e[++u];ngDevMode&&bt(typeof h,"string",'Expected "'+h+'" to be a comment node value');var m=i.createComment(h);ngDevMode&&ngDevMode.rendererCreateComment++,a=o,o=ko(n++,5,m,null,null),ar(m,r),o.activeCaseIndex=null,wn(!1);break;case Vc:var g=e[++u];ngDevMode&&bt(typeof g,"string",'Expected "'+g+'" to be an element node tag name');var y=i.createElement(g);ngDevMode&&ngDevMode.rendererCreateElement++,a=o,o=ko(n++,3,y,g,null);break;default:throw new Error('Unable to determine the type of mutate operation for "'+c+'"')}}return wn(!1),s}function dl(t,e){var n=fe(t,e);mi(n,de(t,e)||null,e),n.detached=!0,ngDevMode&&ngDevMode.rendererRemoveNode++;var r=os(t);if(_e(r)){var i=r;0!==n.type&&mi(n,i[$t]||null,e),i[Jt]=null}}function pl(t,e,n){al(t,e,n),cl()}function fl(t,e){var n=gn()[At];ngDevMode&&Ot(n,"tView should be defined"),ngDevMode&&bt(n.firstTemplatePass,!0,"You should only call i18nEnd on first template pass"),n.firstTemplatePass&&null===n.data[t+Xt]&&function(t,e,n){for(var r=_n().index-Xt,i=[],o=0;o<n.length;o+=2)for(var a=n[o],s=n[o+1],u=s.split(Yc),c=0;c<u.length;c++){var l=u[c];if(1&c);else if(""!==l){var d=!!l.match(Gc);d?Ae(tl(l,r,a),i):ta(r,a,l)}}t.data[e+Xt]=i}(n,t,e)}var hl,ml,gl=0,yl=0;function _l(t){t!==Pr&&(gl|=1<<yl),yl++}function vl(t){if(yl){var e=gn(),n=e[At];ngDevMode&&Ot(n,"tView should be defined");var r=n.data[t+Xt],i=void 0,o=null;Array.isArray(r)?i=r:(i=r.update,o=r.icus),!function t(e,n,r,i,o,a){void 0===a&&(a=!1);for(var s=!1,u=0;u<e.length;u++){var c=e[u],l=e[++u];if(a||c&i)for(var d="",p=u+1;p<=u+l;p++){var f=e[p];if("string"==typeof f)d+=f;else if("number"==typeof f)if(f<0)d+=ue(o[r-f]);else{var h=f>>>2;switch(3&f){case 1:ta(h,e[++p],d,e[++p]);break;case 0:ma(h,d);break;case 2:var m=e[++p],g=n[m],y=fe(h,o);if(null!==y.activeCaseIndex)for(var _=g.remove[y.activeCaseIndex],v=0;v<_.length;v++){var b=_[v];switch(7&b){case 3:dl(b>>>3,o);break;case 6:var x=fe(_[v+1]>>>3,o).activeCaseIndex;null!==x&&Ae(n[b>>>3].remove[x],_)}}var w=xl(g,d);y.activeCaseIndex=-1!==w?w:null,ll(-1,g.create[w],g.expandoStartIndex,o),s=!0;break;case 3:g=n[m=e[++p]],y=fe(h,o),t(g.update[y.activeCaseIndex],n,r,i,o,s)}}}u+=l}}(i,o,e[Lt]-yl-1,gl,e),gl=0,yl=0}}function bl(t,e){switch(function(t,e){"string"==typeof e&&(e=parseInt(e,10));var n=e,r=n.toString().replace(/^[^.]*\.?/,""),i=Math.floor(Math.abs(n)),o=r.length,a=parseInt(r,10),s=parseInt(n.toString().replace(/^[^.]*\.?|0+$/g,""),10)||0;switch(t.split("-")[0].toLowerCase()){case"af":case"asa":case"az":case"bem":case"bez":case"bg":case"brx":case"ce":case"cgg":case"chr":case"ckb":case"ee":case"el":case"eo":case"es":case"eu":case"fo":case"fur":case"gsw":case"ha":case"haw":case"hu":case"jgo":case"jmc":case"ka":case"kk":case"kkj":case"kl":case"ks":case"ksb":case"ky":case"lb":case"lg":case"mas":case"mgo":case"ml":case"mn":case"nb":case"nd":case"ne":case"nn":case"nnh":case"nyn":case"om":case"or":case"os":case"ps":case"rm":case"rof":case"rwk":case"saq":case"seh":case"sn":case"so":case"sq":case"ta":case"te":case"teo":case"tk":case"tr":case"ug":case"uz":case"vo":case"vun":case"wae":case"xog":return 1===n?hl.One:hl.Other;case"ak":case"ln":case"mg":case"pa":case"ti":return n===Math.floor(n)&&n>=0&&n<=1?hl.One:hl.Other;case"am":case"as":case"bn":case"fa":case"gu":case"hi":case"kn":case"mr":case"zu":return 0===i||1===n?hl.One:hl.Other;case"ar":return 0===n?hl.Zero:1===n?hl.One:2===n?hl.Two:n%100===Math.floor(n%100)&&n%100>=3&&n%100<=10?hl.Few:n%100===Math.floor(n%100)&&n%100>=11&&n%100<=99?hl.Many:hl.Other;case"ast":case"ca":case"de":case"en":case"et":case"fi":case"fy":case"gl":case"it":case"nl":case"sv":case"sw":case"ur":case"yi":return 1===i&&0===o?hl.One:hl.Other;case"be":return n%10==1&&n%100!=11?hl.One:n%10===Math.floor(n%10)&&n%10>=2&&n%10<=4&&!(n%100>=12&&n%100<=14)?hl.Few:n%10==0||n%10===Math.floor(n%10)&&n%10>=5&&n%10<=9||n%100===Math.floor(n%100)&&n%100>=11&&n%100<=14?hl.Many:hl.Other;case"br":return n%10==1&&n%100!=11&&n%100!=71&&n%100!=91?hl.One:n%10==2&&n%100!=12&&n%100!=72&&n%100!=92?hl.Two:n%10===Math.floor(n%10)&&(n%10>=3&&n%10<=4||n%10==9)&&!(n%100>=10&&n%100<=19||n%100>=70&&n%100<=79||n%100>=90&&n%100<=99)?hl.Few:0!==n&&n%1e6==0?hl.Many:hl.Other;case"bs":case"hr":case"sr":return 0===o&&i%10==1&&i%100!=11||a%10==1&&a%100!=11?hl.One:0===o&&i%10===Math.floor(i%10)&&i%10>=2&&i%10<=4&&!(i%100>=12&&i%100<=14)||a%10===Math.floor(a%10)&&a%10>=2&&a%10<=4&&!(a%100>=12&&a%100<=14)?hl.Few:hl.Other;case"cs":case"sk":return 1===i&&0===o?hl.One:i===Math.floor(i)&&i>=2&&i<=4&&0===o?hl.Few:0!==o?hl.Many:hl.Other;case"cy":return 0===n?hl.Zero:1===n?hl.One:2===n?hl.Two:3===n?hl.Few:6===n?hl.Many:hl.Other;case"da":return 1===n||0!==s&&(0===i||1===i)?hl.One:hl.Other;case"dsb":case"hsb":return 0===o&&i%100==1||a%100==1?hl.One:0===o&&i%100==2||a%100==2?hl.Two:0===o&&i%100===Math.floor(i%100)&&i%100>=3&&i%100<=4||a%100===Math.floor(a%100)&&a%100>=3&&a%100<=4?hl.Few:hl.Other;case"ff":case"fr":case"hy":case"kab":return 0===i||1===i?hl.One:hl.Other;case"fil":return 0===o&&(1===i||2===i||3===i)||0===o&&i%10!=4&&i%10!=6&&i%10!=9||0!==o&&a%10!=4&&a%10!=6&&a%10!=9?hl.One:hl.Other;case"ga":return 1===n?hl.One:2===n?hl.Two:n===Math.floor(n)&&n>=3&&n<=6?hl.Few:n===Math.floor(n)&&n>=7&&n<=10?hl.Many:hl.Other;case"gd":return 1===n||11===n?hl.One:2===n||12===n?hl.Two:n===Math.floor(n)&&(n>=3&&n<=10||n>=13&&n<=19)?hl.Few:hl.Other;case"gv":return 0===o&&i%10==1?hl.One:0===o&&i%10==2?hl.Two:0!==o||i%100!=0&&i%100!=20&&i%100!=40&&i%100!=60&&i%100!=80?0!==o?hl.Many:hl.Other:hl.Few;case"he":return 1===i&&0===o?hl.One:2===i&&0===o?hl.Two:0!==o||n>=0&&n<=10||n%10!=0?hl.Other:hl.Many;case"is":return 0===s&&i%10==1&&i%100!=11||0!==s?hl.One:hl.Other;case"ksh":return 0===n?hl.Zero:1===n?hl.One:hl.Other;case"kw":case"naq":case"se":case"smn":return 1===n?hl.One:2===n?hl.Two:hl.Other;case"lag":return 0===n?hl.Zero:0!==i&&1!==i||0===n?hl.Other:hl.One;case"lt":return n%10!=1||n%100>=11&&n%100<=19?n%10===Math.floor(n%10)&&n%10>=2&&n%10<=9&&!(n%100>=11&&n%100<=19)?hl.Few:0!==a?hl.Many:hl.Other:hl.One;case"lv":case"prg":return n%10==0||n%100===Math.floor(n%100)&&n%100>=11&&n%100<=19||2===o&&a%100===Math.floor(a%100)&&a%100>=11&&a%100<=19?hl.Zero:n%10==1&&n%100!=11||2===o&&a%10==1&&a%100!=11||2!==o&&a%10==1?hl.One:hl.Other;case"mk":return 0===o&&i%10==1||a%10==1?hl.One:hl.Other;case"mt":return 1===n?hl.One:0===n||n%100===Math.floor(n%100)&&n%100>=2&&n%100<=10?hl.Few:n%100===Math.floor(n%100)&&n%100>=11&&n%100<=19?hl.Many:hl.Other;case"pl":return 1===i&&0===o?hl.One:0===o&&i%10===Math.floor(i%10)&&i%10>=2&&i%10<=4&&!(i%100>=12&&i%100<=14)?hl.Few:0===o&&1!==i&&i%10===Math.floor(i%10)&&i%10>=0&&i%10<=1||0===o&&i%10===Math.floor(i%10)&&i%10>=5&&i%10<=9||0===o&&i%100===Math.floor(i%100)&&i%100>=12&&i%100<=14?hl.Many:hl.Other;case"pt":return n===Math.floor(n)&&n>=0&&n<=2&&2!==n?hl.One:hl.Other;case"ro":return 1===i&&0===o?hl.One:0!==o||0===n||1!==n&&n%100===Math.floor(n%100)&&n%100>=1&&n%100<=19?hl.Few:hl.Other;case"ru":case"uk":return 0===o&&i%10==1&&i%100!=11?hl.One:0===o&&i%10===Math.floor(i%10)&&i%10>=2&&i%10<=4&&!(i%100>=12&&i%100<=14)?hl.Few:0===o&&i%10==0||0===o&&i%10===Math.floor(i%10)&&i%10>=5&&i%10<=9||0===o&&i%100===Math.floor(i%100)&&i%100>=11&&i%100<=14?hl.Many:hl.Other;case"shi":return 0===i||1===n?hl.One:n===Math.floor(n)&&n>=2&&n<=10?hl.Few:hl.Other;case"si":return 0===n||1===n||0===i&&1===a?hl.One:hl.Other;case"sl":return 0===o&&i%100==1?hl.One:0===o&&i%100==2?hl.Two:0===o&&i%100===Math.floor(i%100)&&i%100>=3&&i%100<=4||0!==o?hl.Few:hl.Other;case"tzm":return n===Math.floor(n)&&n>=0&&n<=1||n===Math.floor(n)&&n>=11&&n<=99?hl.One:hl.Other;default:return hl.Other}}(e,t)){case hl.Zero:return"zero";case hl.One:return"one";case hl.Two:return"two";case hl.Few:return"few";case hl.Many:return"many";default:return"other"}}function xl(t,e){var n=t.cases.indexOf(e);if(-1===n)switch(t.type){case 1:var r=bl(e,"en-US");-1===(n=t.cases.indexOf(r))&&"other"!==r&&(n=t.cases.indexOf("other"));break;case 0:n=t.cases.indexOf("other")}return n}function wl(t,e,n,i){for(var o=[],a=[],s=[],u=[],c=[],l=0;l<e.values.length;l++){for(var d=e.values[l],p=[],f=0;f<d.length;f++){var h=d[f];if("string"!=typeof h){var m=p.push(h)-1;d[f]="\x3c!--�"+m+"�--\x3e"}}var g=jl(d.join(""),n,p,t,i);o.push(g.create),a.push(g.remove),s.push(g.update),u.push(g.vars),c.push(g.childIcus)}var y={type:e.type,vars:u,expandoStartIndex:i+1,childIcus:c,cases:e.cases,create:o,remove:a,update:s};t.push(y);var _=gn(),v=Math.max.apply(Math,Object(r.__spread)(u));for(l=0;l<v;l++)Do(_)}function jl(t,e,n,i,o){var a=new mc(document).getInertBodyElement(t);if(!a)throw new Error("Unable to generate inert body element");var s={vars:0,childIcus:[],create:[],remove:[],update:[]};return function t(e,n,i,o,a,s){if(e){for(var u=[];e;){var c=e.nextSibling,l=s+ ++n.vars;switch(e.nodeType){case Node.ELEMENT_NODE:var d=e,p=d.tagName.toLowerCase();if(Tc.hasOwnProperty(p)){n.create.push(Vc,p,i<<17|1);for(var f=d.attributes,h=0;h<f.length;h++){var m=f.item(h),g=m.name.toLowerCase(),y=!!m.value.match(Gc);y?Ic.hasOwnProperty(g)?kc[g]?Ae(tl(m.value,l,m.name,_c),n.update):Ac[g]?Ae(tl(m.value,l,m.name,vc),n.update):Ae(tl(m.value,l,m.name),n.update):ngDevMode&&console.warn("WARNING: ignoring unsafe attribute value "+g+" on element "+p+" (see http://g.co/ng/security#xss)"):n.create.push(l<<3|4,m.name,m.value)}t(e.firstChild,n,l,o,a,s),n.remove.push(l<<3|3)}else n.vars--;break;case Node.TEXT_NODE:var _=e.textContent||"",v=_.match(Gc);n.create.push(v?"":_,i<<17|1),n.remove.push(l<<3|3),v&&Ae(tl(_,l),n.update);break;case Node.COMMENT_NODE:var b=Ol.exec(e.textContent||"");if(b){var x=parseInt(b[1],10),w=ngDevMode?"nested ICU "+x:"";n.create.push(zc,w,i<<17|1);var j=o[x];u.push([j,l])}else n.vars--;break;default:n.vars--}e=c}for(var h=0;h<u.length;h++){var j=u[h][0],O=u[h][1];wl(a,j,O,s+n.vars);var C=a.length-1;n.vars+=Math.max.apply(Math,Object(r.__spread)(a[C].vars)),n.childIcus.push(C);var E=el(j);n.update.push(il(j.mainBinding),3,-1-j.mainBinding,O<<2|2,C,E,2,O<<2|3,C),n.remove.push(C<<3|6,O<<3|3)}}}((Bc(a)||a).firstChild,s,e,n,i,o),s}(ml=hl||(hl={}))[ml.Zero=0]="Zero",ml[ml.One=1]="One",ml[ml.Two=2]="Two",ml[ml.Few=3]="Few",ml[ml.Many=4]="Many",ml[ml.Other=5]="Other";var Ol=/�(\d+)�/;var Cl={provide:ku,useClass:ic,deps:[Iu]},El=function(t){function e(e,n){var r=t.call(this)||this;r._parent=n,r._bootstrapComponents=[],r.injector=r,r.destroyCbs=[];var i=vt(e);ngDevMode&&Ot(i,"NgModule '"+K(e)+"' is not a subtype of 'NgModuleType'."),r._bootstrapComponents=i.bootstrap;var o=[{provide:Iu,useValue:r},Cl];return r._r3Injector=su(e,n,o),r.instance=r.get(e),r}return Object(r.__extends)(e,t),e.prototype.get=function(t,e,n){return void 0===e&&(e=Ls.THROW_IF_NOT_FOUND),void 0===n&&(n=Pe.Default),t===Ls||t===Iu||t===Ns?this:this._r3Injector.get(t,e,n)},Object.defineProperty(e.prototype,"componentFactoryResolver",{get:function(){return this.get(ku)},enumerable:!0,configurable:!0}),e.prototype.destroy=function(){ngDevMode&&Ot(this.destroyCbs,"NgModule already destroyed"),this.destroyCbs.forEach(function(t){return t()}),this.destroyCbs=null},e.prototype.onDestroy=function(t){ngDevMode&&Ot(this.destroyCbs,"NgModule already destroyed"),this.destroyCbs.push(t)},e}(Iu),Sl=function(t){function e(e){var n=t.call(this)||this;return n.moduleType=e,n}return Object(r.__extends)(e,t),e.prototype.create=function(t){return new El(this.moduleType,t)},e}(Pu);function Ml(t,e,n,i){var o,a=t;null!==e&&(void 0!==a.decorators?(o=a.decorators).push.apply(o,Object(r.__spread)(e)):a.decorators=e),null!==n&&(a.ctorParameters=n),null!==i&&(void 0!==a.propDecorators?a.propDecorators=Object(r.__assign)({},a.propDecorators,i):a.propDecorators=i)}function Tl(t,e,n){var r=Dn()+t,i=gn();return jn()?Rr(i,r,n?e.call(n):e()):Nr(i,r)}function kl(t,e,n,r){var i=gn(),o=Dn()+t;return Fr(i,o,n)?Rr(i,o+1,r?e.call(r,n):e(n)):Nr(i,o+1)}function Al(t,e,n,r,i){var o=Dn()+t,a=gn();return Lr(a,o,n,r)?Rr(a,o+2,i?e.call(i,n,r):e(n,r)):Nr(a,o+2)}function Dl(t,e,n,r,i,o){var a=Dn()+t,s=gn();return Br(s,a,n,r,i)?Rr(s,a+3,o?e.call(o,n,r,i):e(n,r,i)):Nr(s,a+3)}function Il(t,e,n,r,i,o,a){var s=Dn()+t,u=gn();return Vr(u,s,n,r,i,o)?Rr(u,s+4,a?e.call(a,n,r,i,o):e(n,r,i,o)):Nr(u,s+4)}function Pl(t,e,n,r,i,o,a,s){var u=Dn()+t,c=gn(),l=Vr(c,u,n,r,i,o);return Fr(c,u+4,a)||l?Rr(c,u+5,s?e.call(s,n,r,i,o,a):e(n,r,i,o,a)):Nr(c,u+5)}function Rl(t,e,n,r,i,o,a,s,u){var c=Dn()+t,l=gn(),d=Vr(l,c,n,r,i,o);return Lr(l,c+4,a,s)||d?Rr(l,c+6,u?e.call(u,n,r,i,o,a,s):e(n,r,i,o,a,s)):Nr(l,c+6)}function Nl(t,e,n,r,i,o,a,s,u,c){var l=Dn()+t,d=gn(),p=Vr(d,l,n,r,i,o);return Br(d,l+4,a,s,u)||p?Rr(d,l+7,c?e.call(c,n,r,i,o,a,s,u):e(n,r,i,o,a,s,u)):Nr(d,l+7)}function Fl(t,e,n,r,i,o,a,s,u,c,l){var d=Dn()+t,p=gn(),f=Vr(p,d,n,r,i,o);return Vr(p,d+4,a,s,u,c)||f?Rr(p,d+8,l?e.call(l,n,r,i,o,a,s,u,c):e(n,r,i,o,a,s,u,c)):Nr(p,d+8)}function Ll(t,e,n,r){for(var i=Dn()+t,o=!1,a=gn(),s=0;s<n.length;s++)Fr(a,i++,n[s])&&(o=!0);return o?Rr(a,i,e.apply(r,n)):Nr(a,i)}function Bl(t,e){var n,r=gn()[At],i=t+Xt;r.firstTemplatePass?(n=function(t,e){if(e)for(var n=e.length-1;n>=0;n--){var r=e[n];if(t===r.name)return r}throw new Error("The pipe '"+t+"' could not be found!")}(e,r.pipeRegistry),r.data[i]=n,n.onDestroy&&(r.pipeDestroyHooks||(r.pipeDestroyHooks=[])).push(i,n.onDestroy)):n=r.data[i];var o=n.factory(null);return ns(t,o),o}function Vl(t,e,n){var r=os(t);return Gl(ql(t)?kl(e,r.transform,n,r):r.transform(n))}function zl(t,e,n,r){var i=os(t);return Gl(ql(t)?Al(e,i.transform,n,r,i):i.transform(n,r))}function Hl(t,e,n,r,i){var o=os(t);return Gl(ql(t)?Dl(e,o.transform,n,r,i,o):o.transform(n,r,i))}function Ul(t,e,n,r,i,o){var a=os(t);return Gl(ql(t)?Il(e,a.transform,n,r,i,o,a):a.transform(n,r,i,o))}function Wl(t,e,n){var r=os(t);return Gl(ql(t)?Ll(e,r.transform,n,r):r.transform.apply(r,n))}function ql(t){return gn()[At].data[t+Xt].pure}function Gl(t){return Tr.isWrapped(t)&&(t=Tr.unwrap(t),gn()[Dn()]=Pr),t}var Yl=function(t){function e(e){void 0===e&&(e=!1);var n=t.call(this)||this;return n.__isAsync=e,n}return Object(r.__extends)(e,t),e.prototype.emit=function(e){t.prototype.next.call(this,e)},e.prototype.subscribe=function(e,n,r){var o,a=function(t){return null},s=function(){return null};e&&"object"==typeof e?(o=this.__isAsync?function(t){setTimeout(function(){return e.next(t)})}:function(t){e.next(t)},e.error&&(a=this.__isAsync?function(t){setTimeout(function(){return e.error(t)})}:function(t){e.error(t)}),e.complete&&(s=this.__isAsync?function(){setTimeout(function(){return e.complete()})}:function(){e.complete()})):(o=this.__isAsync?function(t){setTimeout(function(){return e(t)})}:function(t){e(t)},n&&(a=this.__isAsync?function(t){setTimeout(function(){return n(t)})}:function(t){n(t)}),r&&(s=this.__isAsync?function(){setTimeout(function(){return r()})}:function(){r()}));var u=t.prototype.subscribe.call(this,o,a,s);return e instanceof i.Subscription&&e.add(u),u},e}(i.Subject),Kl=function(){function t(){}return t.__NG_ELEMENT_ID__=function(){return Ql(t,Hu)},t}(),Xl=function(t,e){return Lu(t,e,_n(),gn())},Ql=Ds,Zl=function(){function t(t,e,n){this.parent=t,this.shallow=e,this.deep=n}return t.prototype.track=function(t,e,n,r){n?this.deep=sd(this.deep,t,e,null!=r?r:null):this.shallow=sd(this.shallow,t,e,null!=r?r:null)},t.prototype.clone=function(){return new t(this,null,this.deep)},t.prototype.container=function(){var e=$l(this.shallow),n=$l(this.deep);return e||n?new t(this,e,n):null},t.prototype.createView=function(){var e=Jl(this.shallow),n=Jl(this.deep);return e||n?new t(this,e,n):null},t.prototype.insertView=function(t){td(t,this.shallow),td(t,this.deep)},t.prototype.addNode=function(t){return od(this.deep,t),me(t)?(od(this.shallow,t),t.parent&&me(t.parent)&&od(this.parent.shallow,t),this.parent):((null===(e=t).parent||me(e.parent))&&od(this.shallow,t),this);var e},t.prototype.removeView=function(){ed(this.shallow),ed(this.deep)},t}();function $l(t){for(var e=null;t;){var n=[];t.values.push(n),e={next:e,list:t.list,predicate:t.predicate,values:n,containerValues:null},t=t.next}return e}function Jl(t){for(var e=null;t;){e={next:e,list:t.list,predicate:t.predicate,values:[],containerValues:t.values},t=t.next}return e}function td(t,e){for(;e;)ngDevMode&&Ot(e.containerValues,"View queries need to have a pointer to container values."),e.containerValues.splice(t,0,e.values),e=e.next}function ed(t){for(;t;){ngDevMode&&Ot(t.containerValues,"View queries need to have a pointer to container values.");var e=t.containerValues,n=e.indexOf(t.values),r=e.splice(n,1);ngDevMode&&bt(r.length,1,"removed.length"),r[0].length&&t.list.setDirty(),t=t.next}}function nd(t,e){var n=t.localNames;if(n)for(var r=0;r<n.length;r+=2)if(n[r]===e)return n[r+1];return null}function rd(t,e,n){var r=t[m];if("function"==typeof r)return r();var i=Qn(e,n,t,!1,!1);return null!==i?Zn(n[At].data,n,i,e):null}function id(t,e,n,r){return n?rd(n,t,e):r>-1?Zn(e[At].data,e,r,t):(o=e,3===(i=t).type||4===i.type?Fu(Hu,i,o):0===i.type?Lu(Kl,Hu,i,o):null);var i,o}function od(t,e){for(var n,r,i,o,a,s=gn();t;){var u=t.predicate,c=u.type;if(c){var l=null;if(c===Kl)n=c,r=e,i=s,o=u.read,void 0,a=n[m](),l=o?a?rd(o,r,i):null:a;else null!==(f=Qn(e,s,c,!1,!1))&&(l=id(e,s,u.read,f));null!==l&&ad(t,l)}else for(var d=u.selector,p=0;p<d.length;p++){var f;if(null!==(f=nd(e,d[p])))null!==(l=id(e,s,u.read,f))&&ad(t,l)}t=t.next}}function ad(t,e){t.values.push(e),t.list.setDirty()}function sd(t,e,n,r){return{next:t,list:e,predicate:(i=n,o=r,a=Array.isArray(i),{type:a?null:i,selector:a?i:null,read:o}),values:e._valuesTree,containerValues:null};var i,o,a}var ud=function(){function t(){this.dirty=!0,this.changes=new Yl,this._values=[],this._valuesTree=[]}return Object.defineProperty(t.prototype,"length",{get:function(){return this._values.length},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"first",{get:function(){var t=this._values;return t.length?t[0]:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"last",{get:function(){var t=this._values;return t.length?t[t.length-1]:null},enumerable:!0,configurable:!0}),t.prototype.map=function(t){return this._values.map(t)},t.prototype.filter=function(t){return this._values.filter(t)},t.prototype.find=function(t){return this._values.find(t)},t.prototype.reduce=function(t,e){return this._values.reduce(t,e)},t.prototype.forEach=function(t){this._values.forEach(t)},t.prototype.some=function(t){return this._values.some(t)},t.prototype.toArray=function(){return this._values.slice(0)},t.prototype[q()]=function(){return this._values[q()]()},t.prototype.toString=function(){return this._values.toString()},t.prototype.reset=function(t){this._values=function(t){for(var e=[],n=0;n<t.length;){var r=t[n];Array.isArray(r)?r.length>0?(t=r.concat(t.slice(n+1)),n=0):n++:(e.push(r),n++)}return e}(t),this.dirty=!1},t.prototype.notifyOnChanges=function(){this.changes.emit(this)},t.prototype.setDirty=function(){this.dirty=!0},t.prototype.destroy=function(){this.changes.complete(),this.changes.unsubscribe()},t}();function cd(t,e,n,r){ngDevMode&&Mt(xn());var i,o,a,s,u,c,l,d=new ud;return(i=Zl,o=gn(),a=o[Rt],cn&&cn!==o[Ft]&&!me(cn)&&(a&&(a=o[Rt]=a.clone()),cn.flags|=4),a||(o[Rt]=new i(null,null,null))).track(d,e,n,r),s=gn(),u=d,c=d.destroy,(l=ps(s)).push(u),s[At].firstTemplatePass&&fs(s).push(c,l.length-1),null!=t&&ns(t,d),d}function ld(t){var e=t;return!!t.dirty&&(t.reset(e._valuesTree),t.notifyOnChanges(),!0)}function dd(t,e){return Lu(Kl,Hu,t,e)}var pd="__SANITIZER_TRUSTED_BRAND__";function fd(t,e){return t instanceof String&&t[pd]===e}function hd(t){return vd(t,"Html")}function md(t){return vd(t,"Style")}function gd(t){return vd(t,"Script")}function yd(t){return vd(t,"Url")}function _d(t){return vd(t,"ResourceUrl")}function vd(t,e){var n=new String(t);return n[pd]=e,n}var bd=new RegExp("^([-,.\"'%_!# a-zA-Z0-9]+|(?:(?:matrix|translate|scale|rotate|skew|perspective)(?:X|Y|3d)?|(?:rgb|hsl)a?|(?:repeating-)?(?:linear|radial)-gradient|(?:calc|attr))\\([-0-9.%, #a-zA-Z]+\\))$","g"),xd=/^url\(([^)]+)\)$/;function wd(t){if(!(t=String(t).trim()))return"";var e=t.match(xd);return e&&_c(e[1])===e[1]||t.match(bd)&&function(t){for(var e=!0,n=!0,r=0;r<t.length;r++){var i=t.charAt(r);"'"===i&&n?e=!e:'"'===i&&e&&(n=!n)}return e&&n}(t)?t:(fc()&&console.warn("WARNING: sanitizing unsafe style value "+t+" (see http://g.co/ng/security#xss)."),"unsafe")}function jd(t){var e=Sd();return e?e.sanitize(Qu.HTML,t)||"":fd(t,"Html")?t.toString():Lc(document,ue(t))}function Od(t){var e=Sd();return e?e.sanitize(Qu.STYLE,t)||"":fd(t,"Style")?t.toString():wd(ue(t))}function Cd(t){var e=Sd();return e?e.sanitize(Qu.URL,t)||"":fd(t,"Url")?t.toString():_c(ue(t))}function Ed(t){var e=Sd();if(e)return e.sanitize(Qu.RESOURCE_URL,t)||"";if(fd(t,"ResourceUrl"))return t.toString();throw new Error("unsafe value used in a resource URL context (see http://g.co/ng/security#xss)")}function Sd(){var t=gn();return t&&t[Wt]}var Md={"ɵdefineBase":ft,"ɵdefineComponent":ut,"ɵdefineDirective":ht,defineInjectable:g,defineInjector:y,"ɵdefineNgModule":dt,"ɵdefinePipe":mt,"ɵdirectiveInject":as,"ɵgetFactoryOf":er,"ɵgetInheritedFactory":nr,inject:Ye,"ɵinjectAttribute":ss,"ɵtemplateRefExtractor":dd,"ɵNgOnChangesFeature":ks,"ɵProvidersFeature":xu,"ɵInheritDefinitionFeature":Ss,"ɵelementAttribute":ta,"ɵbind":qa,"ɵcontainer":Ea,"ɵnextContext":Po,"ɵcontainerRefreshStart":Ma,"ɵcontainerRefreshEnd":Ta,"ɵloadQueryList":is,"ɵnamespaceHTML":Vo,"ɵnamespaceMathML":Bo,"ɵnamespaceSVG":Lo,"ɵenableBindings":hn,"ɵdisableBindings":mn,"ɵallocHostVars":Ga,"ɵelementStart":Wo,"ɵelementEnd":Jo,"ɵelement":zo,"ɵelementContainerStart":Ho,"ɵelementContainerEnd":Uo,"ɵpureFunction0":Tl,"ɵpureFunction1":kl,"ɵpureFunction2":Al,"ɵpureFunction3":Dl,"ɵpureFunction4":Il,"ɵpureFunction5":Pl,"ɵpureFunction6":Rl,"ɵpureFunction7":Nl,"ɵpureFunction8":Fl,"ɵpureFunctionV":Ll,"ɵgetCurrentView":ds,"ɵrestoreView":yn,"ɵinterpolation1":Ka,"ɵinterpolation2":Xa,"ɵinterpolation3":Qa,"ɵinterpolation4":Za,"ɵinterpolation5":$a,"ɵinterpolation6":Ja,"ɵinterpolation7":ts,"ɵinterpolation8":es,"ɵinterpolationV":Ya,"ɵelementClassProp":pa,"ɵlistener":$o,"ɵload":os,"ɵprojection":Ra,"ɵelementProperty":ea,"ɵcomponentHostSyntheticProperty":na,"ɵpipeBind1":Vl,"ɵpipeBind2":zl,"ɵpipeBind3":Hl,"ɵpipeBind4":Ul,"ɵpipeBindV":Wl,"ɵprojectionDef":Ia,"ɵpipe":Bl,"ɵquery":cd,"ɵqueryRefresh":ld,"ɵregisterContentQuery":us,"ɵreference":rs,"ɵelementStyling":ua,"ɵelementHostAttrs":ca,"ɵelementStylingMap":fa,"ɵelementStyleProp":da,"ɵelementStylingApply":la,"ɵtemplate":Ca,"ɵtext":ha,"ɵtextBinding":ma,"ɵembeddedViewStart":ka,"ɵembeddedViewEnd":Aa,"ɵi18n":pl,"ɵi18nAttributes":fl,"ɵi18nExp":_l,"ɵi18nStart":al,"ɵi18nEnd":cl,"ɵi18nApply":vl,"ɵi18nPostprocess":ul,"ɵsanitizeHtml":jd,"ɵsanitizeStyle":Od,"ɵdefaultStyleSanitizer":function(t,e){return void 0===e?"background-image"===t||"background"===t||"border-image"===t||"filter"===t||"filter"===t||"list-style"===t||"list-style-image"===t:Od(e)},"ɵsanitizeResourceUrl":Ed,"ɵsanitizeScript":function(t){var e=Sd();if(e)return e.sanitize(Qu.SCRIPT,t)||"";if(fd(t,"Script"))return t.toString();throw new Error("unsafe value used in a script context")},"ɵsanitizeUrl":Cd},Td=function(){return function(){}}(),kd=new Map;function Ad(t,e){var n=kd.get(t);Dd(t,n&&n.moduleType),kd.set(t,e)}function Dd(t,e){if(e)throw new Error("Duplicate module registered for "+t+" - "+K(e)+" vs "+K(e.name))}function Id(t){var e=kd.get(t);if(!e)throw Nd(t);return e}function Pd(t){var e=kd.get(t);if(!e)throw Nd(t);return new Sl(e)}var Rd=Id;function Nd(t){return new Error("No module with ID "+t+" loaded")}var Fd=Function;function Ld(t){return"function"==typeof t}var Bd=/^function\s+\S+\(\)\s*{[\s\S]+\.apply\(this,\s*arguments\)/,Vd=/^class\s+[A-Za-z\d$_]*\s*extends\s+[^{]+{/,zd=/^class\s+[A-Za-z\d$_]*\s*extends\s+[^{]+{[\s\S]*constructor\s*\(/,Hd=function(){function t(t){this._reflect=t||H.Reflect}return t.prototype.isReflectionEnabled=function(){return!0},t.prototype.factory=function(t){return function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return new(t.bind.apply(t,Object(r.__spread)([void 0],e)))}},t.prototype._zipTypesAndAnnotations=function(t,e){var n;n=void 0===t?new Array(e.length):new Array(t.length);for(var r=0;r<n.length;r++)void 0===t?n[r]=[]:t[r]!=Object?n[r]=[t[r]]:n[r]=[],e&&null!=e[r]&&(n[r]=n[r].concat(e[r]));return n},t.prototype._ownParameters=function(t,e){var n=t.toString();if(Bd.exec(n)||Vd.exec(n)&&!zd.exec(n))return null;if(t.parameters&&t.parameters!==e.parameters)return t.parameters;var r=t.ctorParameters;if(r&&r!==e.ctorParameters){var i="function"==typeof r?r():r,o=i.map(function(t){return t&&t.type}),a=i.map(function(t){return t&&Ud(t.decorators)});return this._zipTypesAndAnnotations(o,a)}var s=t.hasOwnProperty(w)&&t[w],u=this._reflect&&this._reflect.getOwnMetadata&&this._reflect.getOwnMetadata("design:paramtypes",t);return u||s?this._zipTypesAndAnnotations(u,s):new Array(t.length).fill(void 0)},t.prototype.parameters=function(t){if(!Ld(t))return[];var e=Wd(t),n=this._ownParameters(t,e);return n||e===Object||(n=this.parameters(e)),n||[]},t.prototype._ownAnnotations=function(t,e){if(t.annotations&&t.annotations!==e.annotations){var n=t.annotations;return"function"==typeof n&&n.annotations&&(n=n.annotations),n}return t.decorators&&t.decorators!==e.decorators?Ud(t.decorators):t.hasOwnProperty(x)?t[x]:null},t.prototype.annotations=function(t){if(!Ld(t))return[];var e=Wd(t),n=this._ownAnnotations(t,e)||[];return(e!==Object?this.annotations(e):[]).concat(n)},t.prototype._ownPropMetadata=function(t,e){if(t.propMetadata&&t.propMetadata!==e.propMetadata){var n=t.propMetadata;return"function"==typeof n&&n.propMetadata&&(n=n.propMetadata),n}if(t.propDecorators&&t.propDecorators!==e.propDecorators){var r=t.propDecorators,i={};return Object.keys(r).forEach(function(t){i[t]=Ud(r[t])}),i}return t.hasOwnProperty(j)?t[j]:null},t.prototype.propMetadata=function(t){if(!Ld(t))return{};var e=Wd(t),n={};if(e!==Object){var i=this.propMetadata(e);Object.keys(i).forEach(function(t){n[t]=i[t]})}var o=this._ownPropMetadata(t,e);return o&&Object.keys(o).forEach(function(t){var e=[];n.hasOwnProperty(t)&&e.push.apply(e,Object(r.__spread)(n[t])),e.push.apply(e,Object(r.__spread)(o[t])),n[t]=e}),n},t.prototype.hasLifecycleHook=function(t,e){return t instanceof Fd&&e in t.prototype},t.prototype.guards=function(t){return{}},t.prototype.getter=function(t){return new Function("o","return o."+t+";")},t.prototype.setter=function(t){return new Function("o","v","return o."+t+" = v;")},t.prototype.method=function(t){return new Function("o","args","if (!o."+t+") throw new Error('\""+t+"\" is undefined');\n        return o."+t+".apply(o, args);")},t.prototype.importUri=function(t){return"object"==typeof t&&t.filePath?t.filePath:"./"+K(t)},t.prototype.resourceUri=function(t){return"./"+K(t)},t.prototype.resolveIdentifier=function(t,e,n,r){return r},t.prototype.resolveEnum=function(t,e){return t[e]},t}();function Ud(t){return t?t.map(function(t){var e=t.type.annotationCls,n=t.args?t.args:[];return new(e.bind.apply(e,Object(r.__spread)([void 0],n)))}):[]}function Wd(t){var e=t.prototype?Object.getPrototypeOf(t.prototype):null;return(e?e.constructor:null)||Object}var qd=null;function Gd(){return qd=qd||new Hd}function Yd(t){return Kd(Gd().parameters(t))}function Kd(t){var e=Ie();return t.map(function(t){return function(t,e){var n={token:null,host:!1,optional:!1,resolved:t.R3ResolvedDependencyType.Token,self:!1,skipSelf:!1};function r(e){n.resolved=t.R3ResolvedDependencyType.Token,n.token=e}if(Array.isArray(e)){if(0===e.length)throw new Error("Dependency array must have arguments.");for(var i=0;i<e.length;i++){var o=e[i];if(void 0!==o)if(o instanceof Fe||"Optional"===o.__proto__.ngMetadataName)n.optional=!0;else if(o instanceof Be||"SkipSelf"===o.__proto__.ngMetadataName)n.skipSelf=!0;else if(o instanceof Le||"Self"===o.__proto__.ngMetadataName)n.self=!0;else if(o instanceof Ve||"Host"===o.__proto__.ngMetadataName)n.host=!0;else if(o instanceof Ne)n.token=o.token;else if(o instanceof I){if(void 0===o.attributeName)throw new Error("Attribute name must be defined.");n.token=o.attributeName,n.resolved=t.R3ResolvedDependencyType.Attribute}else r(o)}}else r(e);return n}(e,t)})}var Xd=[],Qd=[];var Zd=!1;function $d(t){return Array.isArray(t)?t.every($d):!!Z(t)}function Jd(t,e){var n,r;void 0===e&&(e={}),tp(t,e),n=t,r=e,Qd.push({moduleType:n,ngModule:r})}function tp(t,e){ngDevMode&&Ot(t,"Required value moduleType"),ngDevMode&&Ot(e,"Required value ngModule");var n,r,i=dp(e.declarations||Xd),o=null;Object.defineProperty(t,f,{configurable:!0,get:function(){return null===o&&(o=Ie().compileNgModule(Md,"ng://"+t.name+"/ngModuleDef.js",{type:t,bootstrap:dp(e.bootstrap||Xd,Z),declarations:i.map(Z),imports:dp(e.imports||Xd,Z).map(pp),exports:dp(e.exports||Xd,Z).map(pp),emitInline:!0})),o}}),e.id&&(n=e.id,r=t,Dd(n,kd.get(n)),kd.set(n,r));var a=null;Object.defineProperty(t,d,{get:function(){if(null===a){ngDevMode&&ep(t);var n={name:t.name,type:t,deps:Yd(t),providers:e.providers||Xd,imports:[(e.imports||Xd).map(Z),(e.exports||Xd).map(Z)]};a=Ie().compileInjector(Md,"ng://"+t.name+"/ngInjectorDef.js",n)}return a},configurable:!!ngDevMode})}function ep(t){if(!op.get(t)){op.set(t,!0);var e=vt(t=Z(t),!0),n=[];e.declarations.forEach(function(e){gt(e=Z(e))||yt(e)||_t(e)||n.push("Unexpected value '"+ue(e)+"' declared by the module '"+ue(t)+"'. Please add a @Pipe/@Directive/@Component annotation.")});var i=Object(r.__spread)(e.declarations.map(Z),dp(e.imports.map(sp),Z));e.exports.forEach(function(e){var r=(gt(e=Z(e))?"component":yt(e)&&"directive")||_t(e)&&"pipe";r&&-1===i.lastIndexOf(e)&&n.push("Can't export "+r+" "+ue(e)+" from "+ue(t)+" as it was neither declared nor imported!")}),e.declarations.forEach(function(e){e=Z(e);var r=ip.get(e);if(r&&r!==t){var i=[r,t].map(ue).sort();n.push("Type "+ue(e)+" is part of the declarations of 2 modules: "+i[0]+" and "+i[1]+"! Please consider moving "+ue(e)+" to a higher module that imports "+i[0]+" and "+i[1]+". You can also create a new NgModule that exports and includes "+ue(e)+" then import that NgModule in "+i[0]+" and "+i[1]+".")}else ip.set(e,t)}),e.declarations.forEach(function(t){if(gt(t=Z(t))){var e=rp(t,"Component");e&&e.entryComponents&&e.entryComponents.forEach(a)}});var o=rp(t,"NgModule");if(o&&(o.imports&&dp(o.imports,np).forEach(ep),o.bootstrap&&o.bootstrap.forEach(a),o.entryComponents&&o.entryComponents.forEach(a)),n.length)throw new Error(n.join("\n"))}function a(t){t=Z(t),ip.get(t)||n.push("Component "+ue(t)+" is not part of any NgModule or the module has not been imported into your module.")}}function np(t){return(t=Z(t)).ngModule||t}function rp(t,e){var n=null;return r(t.__annotations__),r(t.decorators),n;function r(t){t&&t.forEach(i)}function i(t){if(!n)if(Object.getPrototypeOf(t).ngMetadataName==e)n=t;else if(t.type){Object.getPrototypeOf(t.type).ngMetadataName==e&&(n=t.args[0])}}}var ip=new Map,op=new Map;function ap(){ip=new Map,op=new Map,Qd.length=0}function sp(t){var e=vt(t=Z(t),!0);return Object(r.__spread)(dp(e.exports.map(function(t){return vt(t)?(ep(t),sp(t)):t})))}function up(t,e){var n=dp(e.declarations||Xd),r=lp(t);n.forEach(function(e){e.hasOwnProperty(u)?cp(gt(e),r):e.hasOwnProperty(c)||e.hasOwnProperty(p)||(e.ngSelectorScope=t)})}function cp(t,e){t.directiveDefs=function(){return Array.from(e.compilation.directives).map(function(t){return yt(t)||gt(t)}).filter(function(t){return!!t})},t.pipeDefs=function(){return Array.from(e.compilation.pipes).map(function(t){return _t(t)})}}function lp(t){if(!fp(t))throw new Error(t.name+" does not have an ngModuleDef");var e=vt(t);if(null!==e.transitiveCompileScopes)return e.transitiveCompileScopes;var n={compilation:{directives:new Set,pipes:new Set},exported:{directives:new Set,pipes:new Set}};return e.declarations.forEach(function(t){_t(t)?n.compilation.pipes.add(t):n.compilation.directives.add(t)}),e.imports.forEach(function(t){var e=t;if(!fp(e))throw new Error("Importing "+e.name+" which does not have an ngModuleDef");var r=lp(e);r.exported.directives.forEach(function(t){return n.compilation.directives.add(t)}),r.exported.pipes.forEach(function(t){return n.compilation.pipes.add(t)})}),e.exports.forEach(function(t){var e=t;if(fp(e)){var r=lp(e);r.exported.directives.forEach(function(t){n.compilation.directives.add(t),n.exported.directives.add(t)}),r.exported.pipes.forEach(function(t){n.compilation.pipes.add(t),n.exported.pipes.add(t)})}else _t(e)?n.exported.pipes.add(e):n.exported.directives.add(e)}),e.transitiveCompileScopes=n,n}function dp(t,e){var n=[];return t.forEach(function(t){Array.isArray(t)?n.push.apply(n,Object(r.__spread)(dp(t,e))):n.push(e?e(t):t)}),n}function pp(t){return void 0!==t.ngModule?t.ngModule:t}function fp(t){return!!vt(t)}function hp(t,e){var n,i=null;rt(n=e)&&nt.add(n),Object.defineProperty(t,u,{get:function(){var n=Ie();if(null===i){if(rt(e)){var o=["Component '"+ue(t)+"' is not resolved:"];throw e.templateUrl&&o.push(" - templateUrl: "+ue(e.templateUrl)),e.styleUrls&&e.styleUrls.length&&o.push(" - styleUrls: "+JSON.stringify(e.styleUrls)),o.push("Did you run and wait for 'resolveComponentResources()'?"),new Error(o.join("\n"))}var a=Object(r.__assign)({},gp(t,e),{template:e.template||"",preserveWhitespaces:e.preserveWhitespaces||!1,styles:e.styles||at,animations:e.animations,viewQueries:_p(t,Gd().propMetadata(t),bp),directives:[],changeDetection:e.changeDetection,pipes:new Map,encapsulation:e.encapsulation||J.Emulated,interpolation:e.interpolation,viewProviders:e.viewProviders||null});if(i=n.compileComponent(Md,"ng://"+ue(t)+"/template.html",a),function(){if(!Zd){Zd=!0;try{for(var t=Qd.length-1;t>=0;t--){var e=Qd[t],n=e.moduleType,r=e.ngModule;r.declarations&&r.declarations.every($d)&&(Qd.splice(t,1),up(n,r))}}finally{Zd=!1}}}(),void 0!==t.ngSelectorScope){var s=lp(t.ngSelectorScope);cp(i,s)}}return i},configurable:!!ngDevMode})}function mp(t,e){var n=null;Object.defineProperty(t,c,{get:function(){if(null===n){var r=gp(t,e);n=Ie().compileDirective(Md,"ng://"+(t&&t.name)+"/ngDirectiveDef.js",r)}return n},configurable:!!ngDevMode})}function gp(t,e){var n,r=Gd().propMetadata(t);return{name:t.name,type:t,typeArgumentCount:0,selector:e.selector,deps:Yd(t),host:e.host||ot,propMetadata:r,inputs:e.inputs||at,outputs:e.outputs||at,queries:_p(t,r,vp),lifecycle:{usesOnChanges:void 0!==t.prototype.ngOnChanges},typeSourceSpan:null,usesInheritance:(n=t,!(Object.getPrototypeOf(n.prototype)===Object.prototype)),exportAs:e.exportAs||null,providers:e.providers||null}}function yp(t,e){return{propertyName:t,predicate:(n=e.selector,"string"==typeof n?n.split(",").map(function(t){return t.trim()}):Z(n)),descendants:e.descendants,first:e.first,read:e.read?e.read:null};var n}function _p(t,e,n){var r=[],i=function(i){e.hasOwnProperty(i)&&e[i].forEach(function(e){if(n(e)){if(!e.selector)throw new Error("Can't construct a query for the property \""+i+'" of "'+ue(t)+"\" since the query selector wasn't defined.");r.push(yp(i,e))}})};for(var o in e)i(o);return r}function vp(t){var e=t.ngMetadataName;return"ContentChild"===e||"ContentChildren"===e}function bp(t){var e=t.ngMetadataName;return"ViewChild"===e||"ViewChildren"===e}function xp(t,e){var n=null;Object.defineProperty(t,p,{get:function(){return null===n&&(n=Ie().compilePipe(Md,"ng://"+ue(t)+"/ngPipeDef.js",{type:t,name:t.name,deps:Yd(t),pipeName:e.name,pure:void 0===e.pure||e.pure})),n},configurable:!!ngDevMode})}var wp=O("Directive",function(t){return void 0===t&&(t={}),t},void 0,void 0,function(t,e){return Pp(t,e)}),jp=O("Component",function(t){return void 0===t&&(t={}),Object(r.__assign)({changeDetection:M.Default},t)},wp,void 0,function(t,e){return Ip(t,e)}),Op=O("Pipe",function(t){return Object(r.__assign)({pure:!0},t)},void 0,void 0,function(t,e){return Rp(t,e)}),Cp=function(t){return function(e,n){for(var r=[],i=2;i<arguments.length;i++)r[i-2]=arguments[i];var o,a,u,c=e.constructor;c.hasOwnProperty(h)||(o=e.constructor,a=o.ngBaseDef,u=o.ngBaseDef={inputs:{},outputs:{},declaredInputs:{}},a&&(s(u.inputs,a.inputs),s(u.outputs,a.outputs),s(u.declaredInputs,a.declaredInputs)));var l=c.ngBaseDef;t(l)[n]=r[0]}},Ep=S("Input",function(t){return{bindingPropertyName:t}},void 0,Cp(function(t){return t.inputs||{}})),Sp=S("Output",function(t){return{bindingPropertyName:t}},void 0,Cp(function(t){return t.outputs||{}})),Mp=S("HostBinding",function(t){return{hostPropertyName:t}}),Tp=S("HostListener",function(t,e){return{eventName:t,args:e}}),kp=hp,Ap=mp,Dp=xp,Ip=Ds,Pp=Ds,Rp=Ds,Np=a({provide:String,useValue:a}),Fp=[];function Lp(t,e){if(!e){var n=(c=new Hd).parameters(t);return function(){return new(t.bind.apply(t,Object(r.__spread)([void 0],Xe(n))))}}if(Np in e){var i=e;return function(){return i.useValue}}if(e.useExisting){var o=e;return function(){return Ye(o.useExisting)}}if(e.useFactory){var a=e;return function(){return a.useFactory.apply(a,Object(r.__spread)(Xe(a.deps||Fp)))}}if(e.useClass){var s=e,u=e.deps;if(!u){var c=new Hd;u=c.parameters(t)}return function(){var t;return new((t=s.useClass).bind.apply(t,Object(r.__spread)([void 0],Xe(u))))}}var l=e.deps;if(!l){c=new Hd;l=c.parameters(t)}return function(){return new(t.bind.apply(t,Object(r.__spread)([void 0],Xe(l))))}}var Bp={name:"custom-elements"},Vp={name:"no-errors-schema"},zp=O("NgModule",function(t){return t},void 0,void 0,function(t,e){return Up(t,e)});var Hp=Jd,Up=function(t,e){var n=e&&e.imports||[];e&&e.exports&&(n=Object(r.__spread)(n,[e.exports])),t.ngInjectorDef=y({factory:Lp(t,{useClass:t}),providers:e&&e.providers,imports:n})};var Wp=a({provide:String,useValue:a});function qp(t){return void 0!==t.useClass}function Gp(t){return Wp in t}function Yp(t){return void 0!==t.useFactory}function Kp(t){return void 0!==t.useExisting}var Xp=O("Injectable",void 0,void 0,void 0,function(t,e){return Zp(t,e)});var Qp=function(t,e){var n=null;t.hasOwnProperty(l)||Object.defineProperty(t,l,{get:function(){if(null===n){var r=e||{providedIn:null},i=qp(r)||Yp(r)||Gp(r)||Kp(r),o={name:t.name,type:t,typeArgumentCount:0,providedIn:r.providedIn,ctorDeps:Yd(t),userDeps:void 0};if((qp(r)||Yp(r))&&void 0!==r.deps&&(o.userDeps=Kd(r.deps)),i)if(qp(r))o.useClass=r.useClass;else if(Gp(r))o.useValue=r.useValue;else if(Yp(r))o.useFactory=r.useFactory;else{if(!Kp(r))throw new Error("Unreachable state.");o.useExisting=r.useExisting}else o.useClass=t;n=Ie().compileInjectable(Md,"ng://"+t.name+"/ngInjectableDef.js",o)}return n}})},Zp=function(t,e){e&&void 0!==e.providedIn&&!_(t)&&(t.ngInjectableDef=g({providedIn:e.providedIn,factory:Lp(t,e)}))},$p="ngDebugContext",Jp="ngOriginalError",tf="ngErrorLogger";function ef(t){return t[$p]}function nf(t){return t[Jp]}function rf(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];t.error.apply(t,Object(r.__spread)(e))}var of=function(){function t(){this._console=console}return t.prototype.handleError=function(t){var e=this._findOriginalError(t),n=this._findContext(t),r=t[tf]||rf;r(this._console,"ERROR",t),e&&r(this._console,"ORIGINAL ERROR",e),n&&r(this._console,"ERROR CONTEXT",n)},t.prototype._findContext=function(t){return t?ef(t)?ef(t):this._findContext(nf(t)):null},t.prototype._findOriginalError=function(t){for(var e=nf(t);e&&nf(e);)e=nf(e);return e},t}();function af(t){return t.length>1?" ("+function(t){for(var e=[],n=0;n<t.length;++n){if(e.indexOf(t[n])>-1)return e.push(t[n]),e;e.push(t[n])}return e}(t.slice().reverse()).map(function(t){return K(t.token)}).join(" -> ")+")":""}function sf(t,e,n,r){var i,o,a,s=[e],u=n(s),c=r?(o=u+" caused by: "+((i=r)instanceof Error?i.message:i),(a=Error(o))[Jp]=i,a):Error(u);return c.addKey=uf,c.keys=s,c.injectors=[t],c.constructResolvingMessage=n,c[Jp]=r,c}function uf(t,e){this.injectors.push(t),this.keys.push(e),this.message=this.constructResolvingMessage(this.keys)}function cf(t,e){for(var n=[],r=0,i=e.length;r<i;r++){var o=e[r];o&&0!=o.length?n.push(o.map(K).join(" ")):n.push("?")}return Error("Cannot resolve all parameters for '"+K(t)+"'("+n.join(", ")+"). Make sure that all the parameters are decorated with Inject or have valid type annotations and that '"+K(t)+"' is decorated with Injectable.")}var lf=function(){function t(t,e){if(this.token=t,this.id=e,!t)throw new Error("Token must be defined!");this.displayName=K(this.token)}return t.get=function(t){return df.get(Z(t))},Object.defineProperty(t,"numberOfKeys",{get:function(){return df.numberOfKeys},enumerable:!0,configurable:!0}),t}(),df=new(function(){function t(){this._allKeys=new Map}return t.prototype.get=function(t){if(t instanceof lf)return t;if(this._allKeys.has(t))return this._allKeys.get(t);var e=new lf(t,lf.numberOfKeys);return this._allKeys.set(t,e),e},Object.defineProperty(t.prototype,"numberOfKeys",{get:function(){return this._allKeys.size},enumerable:!0,configurable:!0}),t}()),pf=new(function(){function t(t){this.reflectionCapabilities=t}return t.prototype.updateCapabilities=function(t){this.reflectionCapabilities=t},t.prototype.factory=function(t){return this.reflectionCapabilities.factory(t)},t.prototype.parameters=function(t){return this.reflectionCapabilities.parameters(t)},t.prototype.annotations=function(t){return this.reflectionCapabilities.annotations(t)},t.prototype.propMetadata=function(t){return this.reflectionCapabilities.propMetadata(t)},t.prototype.hasLifecycleHook=function(t,e){return this.reflectionCapabilities.hasLifecycleHook(t,e)},t.prototype.getter=function(t){return this.reflectionCapabilities.getter(t)},t.prototype.setter=function(t){return this.reflectionCapabilities.setter(t)},t.prototype.method=function(t){return this.reflectionCapabilities.method(t)},t.prototype.importUri=function(t){return this.reflectionCapabilities.importUri(t)},t.prototype.resourceUri=function(t){return this.reflectionCapabilities.resourceUri(t)},t.prototype.resolveIdentifier=function(t,e,n,r){return this.reflectionCapabilities.resolveIdentifier(t,e,n,r)},t.prototype.resolveEnum=function(t,e){return this.reflectionCapabilities.resolveEnum(t,e)},t}())(new Hd),ff=function(){function t(t,e,n){this.key=t,this.optional=e,this.visibility=n}return t.fromKey=function(e){return new t(e,!1,null)},t}(),hf=[],mf=function(){return function(t,e,n){this.key=t,this.resolvedFactories=e,this.multiProvider=n,this.resolvedFactory=this.resolvedFactories[0]}}(),gf=function(){return function(t,e){this.factory=t,this.dependencies=e}}();function yf(t){var e,n;if(t.useClass){var r=Z(t.useClass);e=pf.factory(r),n=bf(r)}else t.useExisting?(e=function(t){return t},n=[ff.fromKey(lf.get(t.useExisting))]):t.useFactory?(e=t.useFactory,n=function(t,e){if(e){var n=e.map(function(t){return[t]});return e.map(function(e){return xf(t,e,n)})}return bf(t)}(t.useFactory,t.deps)):(e=function(){return t.useValue},n=hf);return new gf(e,n)}function _f(t){return new mf(lf.get(t.provide),[yf(t)],t.multi||!1)}function vf(t){var e=function(t,e){for(var n=0;n<t.length;n++){var r=t[n],i=e.get(r.key.id);if(i){if(r.multiProvider!==i.multiProvider)throw Error("Cannot mix multi providers and regular providers, got: "+i+" "+r);if(r.multiProvider)for(var o=0;o<r.resolvedFactories.length;o++)i.resolvedFactories.push(r.resolvedFactories[o]);else e.set(r.key.id,r)}else{var a=void 0;a=r.multiProvider?new mf(r.key,r.resolvedFactories.slice(),r.multiProvider):r,e.set(r.key.id,a)}}return e}(function t(e,n){e.forEach(function(e){if(e instanceof Fd)n.push({provide:e,useClass:e});else if(e&&"object"==typeof e&&void 0!==e.provide)n.push(e);else{if(!(e instanceof Array))throw Error("Invalid provider - only instances of Provider and Type are allowed, got: "+e);t(e,n)}});return n}(t,[]).map(_f),new Map);return Array.from(e.values())}function bf(t){var e=pf.parameters(t);if(!e)return[];if(e.some(function(t){return null==t}))throw cf(t,e);return e.map(function(n){return xf(t,n,e)})}function xf(t,e,n){var r=null,i=!1;if(!Array.isArray(e))return wf(e instanceof Ne?e.token:e,i,null);for(var o=null,a=0;a<e.length;++a){var s=e[a];s instanceof Fd?r=s:s instanceof Ne?r=s.token:s instanceof Fe?i=!0:s instanceof Le||s instanceof Be?o=s:s instanceof b&&(r=s)}if(null!=(r=Z(r)))return wf(r,i,o);throw cf(t,n)}function wf(t,e,n){return new ff(lf.get(t),e,n)}var jf=new Object,Of=function(){function t(){}return t.resolve=function(t){return vf(t)},t.resolveAndCreate=function(e,n){var r=t.resolve(e);return t.fromResolvedProviders(r,n)},t.fromResolvedProviders=function(t,e){return new Cf(t,e)},t}(),Cf=function(){function t(t,e){this._constructionCounter=0,this._providers=t,this.parent=e||null;var n=t.length;this.keyIds=new Array(n),this.objs=new Array(n);for(var r=0;r<n;r++)this.keyIds[r]=t[r].key.id,this.objs[r]=jf}return t.prototype.get=function(t,e){return void 0===e&&(e=Rs),this._getByKey(lf.get(t),null,e)},t.prototype.resolveAndCreateChild=function(t){var e=Of.resolve(t);return this.createChildFromResolved(e)},t.prototype.createChildFromResolved=function(e){var n=new t(e);return n.parent=this,n},t.prototype.resolveAndInstantiate=function(t){return this.instantiateResolved(Of.resolve([t])[0])},t.prototype.instantiateResolved=function(t){return this._instantiateProvider(t)},t.prototype.getProviderAtIndex=function(t){if(t<0||t>=this._providers.length)throw Error("Index "+t+" is out-of-bounds.");return this._providers[t]},t.prototype._new=function(t){if(this._constructionCounter++>this._getMaxNumberOfObjects())throw e=this,n=t.key,sf(e,n,function(t){return"Cannot instantiate cyclic dependency!"+af(t)});var e,n;return this._instantiateProvider(t)},t.prototype._getMaxNumberOfObjects=function(){return this.objs.length},t.prototype._instantiateProvider=function(t){if(t.multiProvider){for(var e=new Array(t.resolvedFactories.length),n=0;n<t.resolvedFactories.length;++n)e[n]=this._instantiate(t,t.resolvedFactories[n]);return e}return this._instantiate(t,t.resolvedFactories[0])},t.prototype._instantiate=function(t,e){var n,i,o,a,s,u=this,c=e.factory;try{n=e.dependencies.map(function(t){return u._getByReflectiveDependency(t)})}catch(e){throw e.addKey&&e.addKey(this,t.key),e}try{i=c.apply(void 0,Object(r.__spread)(n))}catch(e){throw o=this,a=e,e.stack,s=t.key,sf(o,s,function(t){var e=K(t[0].token);return a.message+": Error during instantiation of "+e+"!"+af(t)+"."},a)}return i},t.prototype._getByReflectiveDependency=function(t){return this._getByKey(t.key,t.visibility,t.optional?null:Rs)},t.prototype._getByKey=function(e,n,r){return e===t.INJECTOR_KEY?this:n instanceof Le?this._getByKeySelf(e,r):this._getByKeyDefault(e,r,n)},t.prototype._getObjByKeyId=function(t){for(var e=0;e<this.keyIds.length;e++)if(this.keyIds[e]===t)return this.objs[e]===jf&&(this.objs[e]=this._new(this._providers[e])),this.objs[e];return jf},t.prototype._throwOrNull=function(t,e){if(e!==Rs)return e;throw sf(this,t,function(t){return"No provider for "+K(t[0].token)+"!"+af(t)})},t.prototype._getByKeySelf=function(t,e){var n=this._getObjByKeyId(t.id);return n!==jf?n:this._throwOrNull(t,e)},t.prototype._getByKeyDefault=function(e,n,r){var i;for(i=r instanceof Be?this.parent:this;i instanceof t;){var o=i,a=o._getObjByKeyId(e.id);if(a!==jf)return a;i=o.parent}return null!==i?i.get(e.token,n):this._throwOrNull(e,n)},Object.defineProperty(t.prototype,"displayName",{get:function(){return"ReflectiveInjector(providers: ["+function(t,e){for(var n=new Array(t._providers.length),r=0;r<t._providers.length;++r)n[r]=e(t.getProviderAtIndex(r));return n}(this,function(t){return' "'+t.key.displayName+'" '}).join(", ")+"])"},enumerable:!0,configurable:!0}),t.prototype.toString=function(){return this.displayName},t.INJECTOR_KEY=lf.get(Ls),t}();function Ef(t){return!!t&&"function"==typeof t.then}function Sf(t){return!!t&&"function"==typeof t.subscribe}var Mf=new b("Application Initializer"),Tf=function(){function t(t){var e=this;this.appInits=t,this.initialized=!1,this.done=!1,this.donePromise=new Promise(function(t,n){e.resolve=t,e.reject=n})}return t.prototype.runInitializers=function(){var t=this;if(!this.initialized){var e=[],n=function(){t.done=!0,t.resolve()};if(this.appInits)for(var r=0;r<this.appInits.length;r++){var i=this.appInits[r]();Ef(i)&&e.push(i)}Promise.all(e).then(function(){n()}).catch(function(e){t.reject(e)}),0===e.length&&n(),this.initialized=!0}},t=Object(r.__decorate)([Xp(),Object(r.__param)(0,Ne(Mf)),Object(r.__param)(0,Fe()),Object(r.__metadata)("design:paramtypes",[Array])],t)}(),kf=new b("AppId");function Af(){return""+If()+If()+If()}var Df={provide:kf,useFactory:Af,deps:[]};function If(){return String.fromCharCode(97+Math.floor(25*Math.random()))}var Pf=new b("Platform Initializer"),Rf=new b("Platform ID"),Nf=new b("appBootstrapListener"),Ff=new b("Application Packages Root URL"),Lf=function(){function t(){}return t.prototype.log=function(t){console.log(t)},t.prototype.warn=function(t){console.warn(t)},t=Object(r.__decorate)([Xp()],t)}(),Bf=function(){return function(t,e){this.ngModuleFactory=t,this.componentFactories=e}}();function Vf(){throw new Error("Runtime compiler is not loaded")}var zf,Hf,Uf=function(t){return new Sl(t)},Wf=Vf,qf=function(t){return Promise.resolve(Uf(t))},Gf=Vf,Yf=function(t){return new Bf(Uf(t),[])},Kf=Vf,Xf=function(t){return Promise.resolve(Yf(t))},Qf=Vf,Zf=function(){function t(){this.compileModuleSync=Wf,this.compileModuleAsync=Gf,this.compileModuleAndAllComponentsSync=Kf,this.compileModuleAndAllComponentsAsync=Qf}return t.prototype.clearCache=function(){},t.prototype.clearCacheFor=function(t){},t.prototype.getModuleId=function(t){},t=Object(r.__decorate)([Xp()],t)}(),$f=new b("compilerOptions"),Jf=function(){return function(){}}();function th(){var t=H.wtf;return!(!t||!(zf=t.trace))&&(Hf=zf.events,!0)}function eh(t,e){return void 0===e&&(e=null),Hf.createScope(t,e)}function nh(t,e){return zf.leaveScope(t,e),e}function rh(t,e){return zf.beginTimeRange(t,e)}function ih(t){zf.endTimeRange(t)}var oh=th();function ah(t,e){return null}var sh=oh?eh:function(t,e){return ah},uh=oh?nh:function(t,e){return e},ch=oh?rh:function(t,e){return null},lh=oh?ih:function(t){return null},dh=function(){function t(t){var e=t.enableLongStackTrace,n=void 0!==e&&e;if(this.hasPendingMicrotasks=!1,this.hasPendingMacrotasks=!1,this.isStable=!0,this.onUnstable=new Yl(!1),this.onMicrotaskEmpty=new Yl(!1),this.onStable=new Yl(!1),this.onError=new Yl(!1),"undefined"==typeof Zone)throw new Error("In this configuration Angular requires Zone.js");Zone.assertZonePatched();var r;this._nesting=0,this._outer=this._inner=Zone.current,Zone.wtfZoneSpec&&(this._inner=this._inner.fork(Zone.wtfZoneSpec)),Zone.TaskTrackingZoneSpec&&(this._inner=this._inner.fork(new Zone.TaskTrackingZoneSpec)),n&&Zone.longStackTraceZoneSpec&&(this._inner=this._inner.fork(Zone.longStackTraceZoneSpec)),(r=this)._inner=r._inner.fork({name:"angular",properties:{isAngularZone:!0},onInvokeTask:function(t,e,n,i,o,a){try{return mh(r),t.invokeTask(n,i,o,a)}finally{gh(r)}},onInvoke:function(t,e,n,i,o,a,s){try{return mh(r),t.invoke(n,i,o,a,s)}finally{gh(r)}},onHasTask:function(t,e,n,i){t.hasTask(n,i),e===n&&("microTask"==i.change?(r.hasPendingMicrotasks=i.microTask,hh(r)):"macroTask"==i.change&&(r.hasPendingMacrotasks=i.macroTask))},onHandleError:function(t,e,n,i){return t.handleError(n,i),r.runOutsideAngular(function(){return r.onError.emit(i)}),!1}})}return t.isInAngularZone=function(){return!0===Zone.current.get("isAngularZone")},t.assertInAngularZone=function(){if(!t.isInAngularZone())throw new Error("Expected to be in Angular Zone, but it is not!")},t.assertNotInAngularZone=function(){if(t.isInAngularZone())throw new Error("Expected to not be in Angular Zone, but it is!")},t.prototype.run=function(t,e,n){return this._inner.run(t,e,n)},t.prototype.runTask=function(t,e,n,r){var i=this._inner,o=i.scheduleEventTask("NgZoneEvent: "+r,t,fh,ph,ph);try{return i.runTask(o,e,n)}finally{i.cancelTask(o)}},t.prototype.runGuarded=function(t,e,n){return this._inner.runGuarded(t,e,n)},t.prototype.runOutsideAngular=function(t){return this._outer.run(t)},t}();function ph(){}var fh={};function hh(t){if(0==t._nesting&&!t.hasPendingMicrotasks&&!t.isStable)try{t._nesting++,t.onMicrotaskEmpty.emit(null)}finally{if(t._nesting--,!t.hasPendingMicrotasks)try{t.runOutsideAngular(function(){return t.onStable.emit(null)})}finally{t.isStable=!0}}}function mh(t){t._nesting++,t.isStable&&(t.isStable=!1,t.onUnstable.emit(null))}function gh(t){t._nesting--,hh(t)}var yh=function(){function t(){this.hasPendingMicrotasks=!1,this.hasPendingMacrotasks=!1,this.isStable=!0,this.onUnstable=new Yl,this.onMicrotaskEmpty=new Yl,this.onStable=new Yl,this.onError=new Yl}return t.prototype.run=function(t){return t()},t.prototype.runGuarded=function(t){return t()},t.prototype.runOutsideAngular=function(t){return t()},t.prototype.runTask=function(t){return t()},t}(),_h=function(){function t(t){var e=this;this._ngZone=t,this._pendingCount=0,this._isZoneStable=!0,this._didWork=!1,this._callbacks=[],this.taskTrackingZone=null,this._watchAngularEvents(),t.run(function(){e.taskTrackingZone="undefined"==typeof Zone?null:Zone.current.get("TaskTrackingZone")})}return t.prototype._watchAngularEvents=function(){var t=this;this._ngZone.onUnstable.subscribe({next:function(){t._didWork=!0,t._isZoneStable=!1}}),this._ngZone.runOutsideAngular(function(){t._ngZone.onStable.subscribe({next:function(){dh.assertNotInAngularZone(),G(function(){t._isZoneStable=!0,t._runCallbacksIfReady()})}})})},t.prototype.increasePendingRequestCount=function(){return this._pendingCount+=1,this._didWork=!0,this._pendingCount},t.prototype.decreasePendingRequestCount=function(){if(this._pendingCount-=1,this._pendingCount<0)throw new Error("pending async requests below zero");return this._runCallbacksIfReady(),this._pendingCount},t.prototype.isStable=function(){return this._isZoneStable&&0===this._pendingCount&&!this._ngZone.hasPendingMacrotasks},t.prototype._runCallbacksIfReady=function(){var t=this;if(this.isStable())G(function(){for(;0!==t._callbacks.length;){var e=t._callbacks.pop();clearTimeout(e.timeoutId),e.doneCb(t._didWork)}t._didWork=!1});else{var e=this.getPendingTasks();this._callbacks=this._callbacks.filter(function(t){return!t.updateCb||!t.updateCb(e)||(clearTimeout(t.timeoutId),!1)}),this._didWork=!0}},t.prototype.getPendingTasks=function(){return this.taskTrackingZone?this.taskTrackingZone.macroTasks.map(function(t){return{source:t.source,creationLocation:t.creationLocation,data:t.data}}):[]},t.prototype.addCallback=function(t,e,n){var r=this,i=-1;e&&e>0&&(i=setTimeout(function(){r._callbacks=r._callbacks.filter(function(t){return t.timeoutId!==i}),t(r._didWork,r.getPendingTasks())},e)),this._callbacks.push({doneCb:t,timeoutId:i,updateCb:n})},t.prototype.whenStable=function(t,e,n){if(n&&!this.taskTrackingZone)throw new Error('Task tracking zone is required when passing an update callback to whenStable(). Is "zone.js/dist/task-tracking.js" loaded?');this.addCallback(t,e,n),this._runCallbacksIfReady()},t.prototype.getPendingRequestCount=function(){return this._pendingCount},t.prototype.findProviders=function(t,e,n){return[]},t=Object(r.__decorate)([Xp(),Object(r.__metadata)("design:paramtypes",[dh])],t)}(),vh=function(){function t(){this._applications=new Map,wh.addToWindow(this)}return t.prototype.registerApplication=function(t,e){this._applications.set(t,e)},t.prototype.unregisterApplication=function(t){this._applications.delete(t)},t.prototype.unregisterAllApplications=function(){this._applications.clear()},t.prototype.getTestability=function(t){return this._applications.get(t)||null},t.prototype.getAllTestabilities=function(){return Array.from(this._applications.values())},t.prototype.getAllRootElements=function(){return Array.from(this._applications.keys())},t.prototype.findTestabilityInTree=function(t,e){return void 0===e&&(e=!0),wh.findTestabilityInTree(this,t,e)},t=Object(r.__decorate)([Xp(),Object(r.__metadata)("design:paramtypes",[])],t)}();function bh(t){wh=t}var xh,wh=new(function(){function t(){}return t.prototype.addToWindow=function(t){},t.prototype.findTestabilityInTree=function(t,e,n){return null},t}()),jh=function(t,e,n){return t.get(Jf).createCompiler([e]).compileModuleAsync(n)};function Oh(t,e,n){var r;return ngDevMode&&(void 0===r&&(r="Type passed in is not NgModuleType, it does not have 'ngModuleDef' property."),vt(n)||Et(r)),Promise.resolve(new Sl(n))}var Ch=new b("AllowMultipleToken"),Eh=function(){return function(t,e){this.name=t,this.token=e}}();function Sh(t){if(xh&&!xh.destroyed&&!xh.injector.get(Ch,!1))throw new Error("There can be only one platform. Destroy the previous one to create a new one.");xh=t.get(Dh);var e=t.get(Pf,null);return e&&e.forEach(function(t){return t()}),xh}function Mh(t,e,n){void 0===n&&(n=[]);var r="Platform: "+e,i=new b(r);return function(e){void 0===e&&(e=[]);var o=Ah();if(!o||o.injector.get(Ch,!1))if(t)t(n.concat(e).concat({provide:i,useValue:!0}));else{var a=n.concat(e).concat({provide:i,useValue:!0});Sh(Ls.create({providers:a,name:r}))}return Th(i)}}function Th(t){var e=Ah();if(!e)throw new Error("No platform exists!");if(!e.injector.get(t,null))throw new Error("A platform with a different configuration has been created. Please destroy it first.");return e}function kh(){xh&&!xh.destroyed&&xh.destroy()}function Ah(){return xh&&!xh.destroyed?xh:null}var Dh=function(){function t(t){this._injector=t,this._modules=[],this._destroyListeners=[],this._destroyed=!1}return t.prototype.bootstrapModuleFactory=function(t,e){var n=this,r=function(t){var e;e="noop"===t?new yh:("zone.js"===t?void 0:t)||new dh({enableLongStackTrace:fc()});return e}(e?e.ngZone:void 0),i=[{provide:dh,useValue:r}];return r.run(function(){var e=Ls.create({providers:i,parent:n.injector,name:t.moduleType.name}),o=t.create(e),a=o.injector.get(of,null);if(!a)throw new Error("No ErrorHandler. Is platform module (BrowserModule) included?");return o.onDestroy(function(){return Rh(n._modules,o)}),r.runOutsideAngular(function(){return r.onError.subscribe({next:function(t){a.handleError(t)}})}),function(t,e,n){try{var r=n();return Ef(r)?r.catch(function(n){throw e.runOutsideAngular(function(){return t.handleError(n)}),n}):r}catch(n){throw e.runOutsideAngular(function(){return t.handleError(n)}),n}}(a,r,function(){var t=o.injector.get(Tf);return t.runInitializers(),t.donePromise.then(function(){return n._moduleDoBootstrap(o),o})})})},t.prototype.bootstrapModule=function(t,e){var n=this;void 0===e&&(e=[]);var r=Ih({},e);return jh(this.injector,r,t).then(function(t){return n.bootstrapModuleFactory(t,r)})},t.prototype._moduleDoBootstrap=function(t){var e=t.injector.get(Ph);if(t._bootstrapComponents.length>0)t._bootstrapComponents.forEach(function(t){return e.bootstrap(t)});else{if(!t.instance.ngDoBootstrap)throw new Error("The module "+K(t.instance.constructor)+' was bootstrapped, but it does not declare "@NgModule.bootstrap" components nor a "ngDoBootstrap" method. Please define one of these.');t.instance.ngDoBootstrap(e)}this._modules.push(t)},t.prototype.onDestroy=function(t){this._destroyListeners.push(t)},Object.defineProperty(t.prototype,"injector",{get:function(){return this._injector},enumerable:!0,configurable:!0}),t.prototype.destroy=function(){if(this._destroyed)throw new Error("The platform has already been destroyed!");this._modules.slice().forEach(function(t){return t.destroy()}),this._destroyListeners.forEach(function(t){return t()}),this._destroyed=!0},Object.defineProperty(t.prototype,"destroyed",{get:function(){return this._destroyed},enumerable:!0,configurable:!0}),t=Object(r.__decorate)([Xp(),Object(r.__metadata)("design:paramtypes",[Ls])],t)}();function Ih(t,e){return t=Array.isArray(e)?e.reduce(Ih,t):Object(r.__assign)({},t,e)}var Ph=function(){function t(t,e,n,r,a,s){var u=this;this._zone=t,this._console=e,this._injector=n,this._exceptionHandler=r,this._componentFactoryResolver=a,this._initStatus=s,this._bootstrapListeners=[],this._views=[],this._runningTick=!1,this._enforceNoNewChanges=!1,this._stable=!0,this.componentTypes=[],this.components=[],this._enforceNoNewChanges=fc(),this._zone.onMicrotaskEmpty.subscribe({next:function(){u._zone.run(function(){u.tick()})}});var c=new i.Observable(function(t){u._stable=u._zone.isStable&&!u._zone.hasPendingMacrotasks&&!u._zone.hasPendingMicrotasks,u._zone.runOutsideAngular(function(){t.next(u._stable),t.complete()})}),l=new i.Observable(function(t){var e;u._zone.runOutsideAngular(function(){e=u._zone.onStable.subscribe(function(){dh.assertNotInAngularZone(),G(function(){u._stable||u._zone.hasPendingMacrotasks||u._zone.hasPendingMicrotasks||(u._stable=!0,t.next(!0))})})});var n=u._zone.onUnstable.subscribe(function(){dh.assertInAngularZone(),u._stable&&(u._stable=!1,u._zone.runOutsideAngular(function(){t.next(!1)}))});return function(){e.unsubscribe(),n.unsubscribe()}});this.isStable=Object(i.merge)(c,l.pipe(Object(o.share)()))}var e;return e=t,t.prototype.bootstrap=function(t,e){var n,r=this;if(!this._initStatus.done)throw new Error("Cannot bootstrap as there are still asynchronous initializers running. Bootstrap components in the `ngDoBootstrap` method of the root module.");n=t instanceof ju?t:this._componentFactoryResolver.resolveComponentFactory(t),this.componentTypes.push(n.componentType);var i=n instanceof Du?null:this._injector.get(Iu),o=e||n.selector,a=n.create(Ls.NULL,[],o,i);a.onDestroy(function(){r._unloadComponent(a)});var s=a.injector.get(_h,null);return s&&a.injector.get(vh).registerApplication(a.location.nativeElement,s),this._loadComponent(a),fc()&&this._console.log("Angular is running in the development mode. Call enableProdMode() to enable the production mode."),a},t.prototype.tick=function(){var t=this;if(this._runningTick)throw new Error("ApplicationRef.tick is called recursively");var n=e._tickScope();try{this._runningTick=!0,this._views.forEach(function(t){return t.detectChanges()}),this._enforceNoNewChanges&&this._views.forEach(function(t){return t.checkNoChanges()})}catch(e){this._zone.runOutsideAngular(function(){return t._exceptionHandler.handleError(e)})}finally{this._runningTick=!1,uh(n)}},t.prototype.attachView=function(t){var e=t;this._views.push(e),e.attachToAppRef(this)},t.prototype.detachView=function(t){var e=t;Rh(this._views,e),e.detachFromAppRef()},t.prototype._loadComponent=function(t){this.attachView(t.hostView),this.tick(),this.components.push(t),this._injector.get(Nf,[]).concat(this._bootstrapListeners).forEach(function(e){return e(t)})},t.prototype._unloadComponent=function(t){this.detachView(t.hostView),Rh(this.components,t)},t.prototype.ngOnDestroy=function(){this._views.slice().forEach(function(t){return t.destroy()})},Object.defineProperty(t.prototype,"viewCount",{get:function(){return this._views.length},enumerable:!0,configurable:!0}),t._tickScope=sh("ApplicationRef#tick()"),t=e=Object(r.__decorate)([Xp(),Object(r.__metadata)("design:paramtypes",[dh,Lf,Ls,of,ku,Tf])],t)}();function Rh(t,e){var n=t.indexOf(e);n>-1&&t.splice(n,1)}var Nh=function(){function t(){this.dirty=!0,this._results=[],this.changes=new Yl,this.length=0}return t.prototype.map=function(t){return this._results.map(t)},t.prototype.filter=function(t){return this._results.filter(t)},t.prototype.find=function(t){return this._results.find(t)},t.prototype.reduce=function(t,e){return this._results.reduce(t,e)},t.prototype.forEach=function(t){this._results.forEach(t)},t.prototype.some=function(t){return this._results.some(t)},t.prototype.toArray=function(){return this._results.slice()},t.prototype[q()]=function(){return this._results[q()]()},t.prototype.toString=function(){return this._results.toString()},t.prototype.reset=function(t){this._results=function t(e){return e.reduce(function(e,n){var r=Array.isArray(n)?t(n):n;return e.concat(r)},[])}(t),this.dirty=!1,this.length=this._results.length,this.last=this._results[this.length-1],this.first=this._results[0]},t.prototype.notifyOnChanges=function(){this.changes.emit(this)},t.prototype.setDirty=function(){this.dirty=!0},t.prototype.destroy=function(){this.changes.complete(),this.changes.unsubscribe()},t}();var Fh=function(){return function(){}}(),Lh={factoryPathPrefix:"",factoryPathSuffix:".ngfactory"},Bh=function(){function t(t,e){this._compiler=t,this._config=e||Lh}return t.prototype.load=function(t){return this._compiler instanceof Zf?this.loadFactory(t):this.loadAndCompile(t)},t.prototype.loadAndCompile=function(t){var e=this,i=Object(r.__read)(t.split("#"),2),o=i[0],a=i[1];return void 0===a&&(a="default"),n("./src/$$_lazy_route_resource lazy recursive")(o).then(function(t){return t[a]}).then(function(t){return Vh(t,o,a)}).then(function(t){return e._compiler.compileModuleAsync(t)})},t.prototype.loadFactory=function(t){var e=Object(r.__read)(t.split("#"),2),i=e[0],o=e[1],a="NgFactory";return void 0===o&&(o="default",a=""),n("./src/$$_lazy_route_resource lazy recursive")(this._config.factoryPathPrefix+i+this._config.factoryPathSuffix).then(function(t){return t[o+a]}).then(function(t){return Vh(t,i,o)})},t=Object(r.__decorate)([Xp(),Object(r.__param)(1,Fe()),Object(r.__metadata)("design:paramtypes",[Zf,Fh])],t)}();function Vh(t,e,n){if(!t)throw new Error("Cannot find '"+n+"' in '"+e+"'");return t}var zh=function(){function t(){}return t.__NG_ELEMENT_ID__=function(){return Uh(t,Hu)},t}(),Hh=function(t,e){return function(t,e,n,i){var o;Su||(Su=function(t){function n(e,n,r){var i=t.call(this)||this;return i._lContainer=e,i._hostTNode=n,i._hostView=r,i._viewRefs=[],i}return Object(r.__extends)(n,t),Object.defineProperty(n.prototype,"element",{get:function(){return Fu(e,this._hostTNode,this._hostView)},enumerable:!0,configurable:!0}),Object.defineProperty(n.prototype,"injector",{get:function(){return new tr(this._hostTNode,this._hostView)},enumerable:!0,configurable:!0}),Object.defineProperty(n.prototype,"parentInjector",{get:function(){var t=Wn(this._hostTNode,this._hostView),e=Se(t,this._hostView),n=function(t,e,n){if(n.parent&&-1!==n.parent.injectorIndex){for(var r=n.parent.injectorIndex,i=n.parent;null!=i.parent&&r==i.injectorIndex;)i=i.parent;return i}for(var o=Ee(t),a=e,s=e[Ft];o>1;)s=(a=a[Kt])[Ft],o--;return s}(t,this._hostView,this._hostTNode);return Oe(t)&&null!=n?new tr(n,e):new tr(null,this._hostView)},enumerable:!0,configurable:!0}),n.prototype.clear=function(){for(;this._lContainer[Zt].length;)this.remove(0)},n.prototype.get=function(t){return this._viewRefs[t]||null},Object.defineProperty(n.prototype,"length",{get:function(){return this._lContainer[Zt].length},enumerable:!0,configurable:!0}),n.prototype.createEmbeddedView=function(t,e,n){var r=this._adjustIndex(n),i=t.createEmbeddedView(e||{},this._lContainer,this._hostTNode,this._hostView,r);return i.attachToViewContainerRef(this),this._viewRefs.splice(r,0,i),i},n.prototype.createComponent=function(t,e,n,r,i){var o=n||this.parentInjector;!i&&null==t.ngModule&&o&&(i=o.get(Iu,null));var a=t.create(o,r,void 0,i);return this.insert(a.hostView,e),a},n.prototype.insert=function(t,e){if(t.destroyed)throw new Error("Cannot insert a destroyed View in a ViewContainer!");var n=t._lView,r=this._adjustIndex(e);ei(n,this._lContainer,this._hostView,r,this._hostTNode.index);var i=hi(r,this._lContainer[Zt],this._lContainer[$t]);return ti(n,!0,i),t.attachToViewContainerRef(this),this._viewRefs.splice(r,0,t),t},n.prototype.move=function(t,e){if(t.destroyed)throw new Error("Cannot move a destroyed View in a ViewContainer!");var n=this.indexOf(t);return this.detach(n),this.insert(t,this._adjustIndex(e)),t},n.prototype.indexOf=function(t){return this._viewRefs.indexOf(t)},n.prototype.remove=function(t){var e=this._adjustIndex(t,-1);ri(this._lContainer,this._hostTNode,e),this._viewRefs.splice(e,1)},n.prototype.detach=function(t){var e=this._adjustIndex(t,-1),n=ni(this._lContainer,e,!!this._hostTNode.detached),r=null!=this._viewRefs.splice(e,1)[0];return r?new Ru(n,n[Vt],n[Gt]):null},n.prototype._adjustIndex=function(t,e){return void 0===e&&(e=0),null==t?this._lContainer[Zt].length+e:(ngDevMode&&(jt(t,-1,"index must be positive"),wt(t,this._lContainer[Zt].length+1+e,"index")),t)},n}(t)),ngDevMode&&Ze(n,0,3,4);var a,s,u=i[n.index];if(_e(u))(o=u)[Qt]=-1;else{var c=i[Ut].createComment(ngDevMode?"container":"");if(ngDevMode&&ngDevMode.rendererCreateComment++,ve(i)){var l=i[Ut],d=pe(n,i),p=di(l,d);li(l,p,c,(s=d,Wr(a=l)?a.nextSibling(s):s.nextSibling))}else pi(c,n,i);i[n.index]=o=Oa(u,n,i,c,!0),Na(i,n.index,o)}return new Su(o,n,i)}(t,e,_n(),gn())},Uh=Ds,Wh=function(){function t(){}return t.__NG_ELEMENT_ID__=function(){return Gh()},t}(),qh=function(){return Bu(_n(),gn(),null)},Gh=function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e]},Yh=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e}(Wh),Kh=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e}(Yh),Xh=function(){return function(t,e){this.name=t,this.callback=e}}(),Qh=function(){function t(t,e,n){this.listeners=[],this.parent=null,this._debugContext=n,this.nativeNode=t,e&&e instanceof Zh&&e.addChild(this)}return Object.defineProperty(t.prototype,"injector",{get:function(){return this._debugContext.injector},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"componentInstance",{get:function(){return this._debugContext.component},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"context",{get:function(){return this._debugContext.context},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"references",{get:function(){return this._debugContext.references},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"providerTokens",{get:function(){return this._debugContext.providerTokens},enumerable:!0,configurable:!0}),t}(),Zh=function(t){function e(e,n,r){var i=t.call(this,e,n,r)||this;return i.properties={},i.attributes={},i.classes={},i.styles={},i.childNodes=[],i.nativeElement=e,i}return Object(r.__extends)(e,t),e.prototype.addChild=function(t){t&&(this.childNodes.push(t),t.parent=this)},e.prototype.removeChild=function(t){var e=this.childNodes.indexOf(t);-1!==e&&(t.parent=null,this.childNodes.splice(e,1))},e.prototype.insertChildrenAfter=function(t,e){var n,i=this,o=this.childNodes.indexOf(t);-1!==o&&((n=this.childNodes).splice.apply(n,Object(r.__spread)([o+1,0],e)),e.forEach(function(e){e.parent&&e.parent.removeChild(e),t.parent=i}))},e.prototype.insertBefore=function(t,e){var n=this.childNodes.indexOf(t);-1===n?this.addChild(e):(e.parent&&e.parent.removeChild(e),e.parent=this,this.childNodes.splice(n,0,e))},e.prototype.query=function(t){return this.queryAll(t)[0]||null},e.prototype.queryAll=function(t){var e=[];return function t(e,n,r){e.childNodes.forEach(function(e){e instanceof Zh&&(n(e)&&r.push(e),t(e,n,r))})}(this,t,e),e},e.prototype.queryAllNodes=function(t){var e=[];return function t(e,n,r){e instanceof Zh&&e.childNodes.forEach(function(e){n(e)&&r.push(e),e instanceof Zh&&t(e,n,r)})}(this,t,e),e},Object.defineProperty(e.prototype,"children",{get:function(){return this.childNodes.filter(function(t){return t instanceof e})},enumerable:!0,configurable:!0}),e.prototype.triggerEventHandler=function(t,e){this.listeners.forEach(function(n){n.name==t&&n.callback(e)})},e}(Qh);function $h(t){return t.map(function(t){return t.nativeElement})}var Jh=function(){function t(t){this.nativeNode=t}return Object.defineProperty(t.prototype,"parent",{get:function(){var t=this.nativeNode.parentNode;return t?new tm(t):null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"injector",{get:function(){return gr(this.nativeNode)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"componentInstance",{get:function(){var t=this.nativeNode;return t&&dr(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"context",{get:function(){return pr(this.nativeNode)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"listeners",{get:function(){return jr(this.nativeNode).filter(wr)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"references",{get:function(){return vr(this.nativeNode)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"providerTokens",{get:function(){return function(t){var e=_r(t,!1);if(!e)return[];for(var n,r=e.lView[At],i=r.data[e.nodeIndex],o=[],a=65535&i.providerIndexes,s=i.directiveEnd,u=a;u<s;u++){var c=r.data[u];void 0!==(n=c).type&&void 0!==n.template&&void 0!==n.declaredInputs&&(c=c.type),o.push(c)}return o}(this.nativeNode)},enumerable:!0,configurable:!0}),t}(),tm=function(t){function e(e){return ngDevMode&&St(e),t.call(this,e)||this}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"nativeElement",{get:function(){return this.nativeNode.nodeType==Node.ELEMENT_NODE?this.nativeNode:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"name",{get:function(){return this.nativeElement.nodeName},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"properties",{get:function(){var t=_r(this.nativeNode);t.lView[At].data[t.nodeIndex];return{}},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"attributes",{get:function(){var t={},e=this.nativeElement;if(e)for(var n=e.attributes,r=0;r<n.length;r++){var i=n[r];t[i.name]=i.value}return t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"classes",{get:function(){var t={},e=this.nativeElement;if(e){var n=xr(e),r=n.lView[n.nodeIndex];if(Mi(n.nodeIndex,n.lView)){for(var i=9;i<r.length;i+=4)if(Gi(r,i)){var o=uo(r,i),a=so(r,i);"boolean"==typeof a&&(t[o]=a)}}else{var s=e.classList;for(i=0;i<s.length;i++)t[s[i]]=!0}}return t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"styles",{get:function(){var t={},e=this.nativeElement;if(e){var n=xr(e),r=n.lView[n.nodeIndex];if(Mi(n.nodeIndex,n.lView)){for(var i=9;i<r.length;i+=4)if(!Gi(r,i)){var o=uo(r,i),a=so(r,i);null!==a&&(t[o]=a)}}else{var s=e.style;for(i=0;i<s.length;i++){var u=s.item(i);t[u]=s.getPropertyValue(u)}}}return t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"childNodes",{get:function(){for(var t=this.nativeNode.childNodes,e=[],n=0;n<t.length;n++){var r=t[n];e.push(rm(r))}return e},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"children",{get:function(){var t=this.nativeElement;if(!t)return[];for(var e=t.children,n=[],r=0;r<e.length;r++){var i=e[r];n.push(rm(i))}return n},enumerable:!0,configurable:!0}),e.prototype.query=function(t){return this.queryAll(t)[0]||null},e.prototype.queryAll=function(t){var e=[];return em(this,t,e,!0),e},e.prototype.queryAllNodes=function(t){var e=[];return em(this,t,e,!1),e},e.prototype.triggerEventHandler=function(t,e){this.listeners.forEach(function(n){n.name===t&&n.callback(e)})},e}(Jh);function em(t,e,n,r){t instanceof tm&&t.childNodes.forEach(function(t){e(t)&&n.push(t),t instanceof tm&&(r&&!t.nativeElement||em(t,e,n,r))})}var nm=new Map;function rm(t){return t instanceof Node?t.nodeType==Node.ELEMENT_NODE?new tm(t):new Jh(t):null}var im=function(t){return nm.get(t)||null};function om(t){nm.set(t.nativeNode,t)}var am=Qh,sm=Zh,um=function(){function t(){}return t.prototype.supports=function(t){return Ar(t)},t.prototype.create=function(t){return new lm(t)},t}(),cm=function(t,e){return e},lm=function(){function t(t){this.length=0,this._linkedRecords=null,this._unlinkedRecords=null,this._previousItHead=null,this._itHead=null,this._itTail=null,this._additionsHead=null,this._additionsTail=null,this._movesHead=null,this._movesTail=null,this._removalsHead=null,this._removalsTail=null,this._identityChangesHead=null,this._identityChangesTail=null,this._trackByFn=t||cm}return t.prototype.forEachItem=function(t){var e;for(e=this._itHead;null!==e;e=e._next)t(e)},t.prototype.forEachOperation=function(t){for(var e=this._itHead,n=this._removalsHead,r=0,i=null;e||n;){var o=!n||e&&e.currentIndex<hm(n,r,i)?e:n,a=hm(o,r,i),s=o.currentIndex;if(o===n)r--,n=n._nextRemoved;else if(e=e._next,null==o.previousIndex)r++;else{i||(i=[]);var u=a-r,c=s-r;if(u!=c){for(var l=0;l<u;l++){var d=l<i.length?i[l]:i[l]=0,p=d+l;c<=p&&p<u&&(i[l]=d+1)}i[o.previousIndex]=c-u}}a!==s&&t(o,a,s)}},t.prototype.forEachPreviousItem=function(t){var e;for(e=this._previousItHead;null!==e;e=e._nextPrevious)t(e)},t.prototype.forEachAddedItem=function(t){var e;for(e=this._additionsHead;null!==e;e=e._nextAdded)t(e)},t.prototype.forEachMovedItem=function(t){var e;for(e=this._movesHead;null!==e;e=e._nextMoved)t(e)},t.prototype.forEachRemovedItem=function(t){var e;for(e=this._removalsHead;null!==e;e=e._nextRemoved)t(e)},t.prototype.forEachIdentityChange=function(t){var e;for(e=this._identityChangesHead;null!==e;e=e._nextIdentityChange)t(e)},t.prototype.diff=function(t){if(null==t&&(t=[]),!Ar(t))throw new Error("Error trying to diff '"+K(t)+"'. Only arrays and iterables are allowed");return this.check(t)?this:null},t.prototype.onDestroy=function(){},t.prototype.check=function(t){var e=this;this._reset();var n,r,i,o=this._itHead,a=!1;if(Array.isArray(t)){this.length=t.length;for(var s=0;s<this.length;s++)r=t[s],i=this._trackByFn(s,r),null!==o&&Y(o.trackById,i)?(a&&(o=this._verifyReinsertion(o,r,i,s)),Y(o.item,r)||this._addIdentityChange(o,r)):(o=this._mismatch(o,r,i,s),a=!0),o=o._next}else n=0,function(t,e){if(Array.isArray(t))for(var n=0;n<t.length;n++)e(t[n]);else for(var r=t[q()](),i=void 0;!(i=r.next()).done;)e(i.value)}(t,function(t){i=e._trackByFn(n,t),null!==o&&Y(o.trackById,i)?(a&&(o=e._verifyReinsertion(o,t,i,n)),Y(o.item,t)||e._addIdentityChange(o,t)):(o=e._mismatch(o,t,i,n),a=!0),o=o._next,n++}),this.length=n;return this._truncate(o),this.collection=t,this.isDirty},Object.defineProperty(t.prototype,"isDirty",{get:function(){return null!==this._additionsHead||null!==this._movesHead||null!==this._removalsHead||null!==this._identityChangesHead},enumerable:!0,configurable:!0}),t.prototype._reset=function(){if(this.isDirty){var t=void 0,e=void 0;for(t=this._previousItHead=this._itHead;null!==t;t=t._next)t._nextPrevious=t._next;for(t=this._additionsHead;null!==t;t=t._nextAdded)t.previousIndex=t.currentIndex;for(this._additionsHead=this._additionsTail=null,t=this._movesHead;null!==t;t=e)t.previousIndex=t.currentIndex,e=t._nextMoved;this._movesHead=this._movesTail=null,this._removalsHead=this._removalsTail=null,this._identityChangesHead=this._identityChangesTail=null}},t.prototype._mismatch=function(t,e,n,r){var i;return null===t?i=this._itTail:(i=t._prev,this._remove(t)),null!==(t=null===this._linkedRecords?null:this._linkedRecords.get(n,r))?(Y(t.item,e)||this._addIdentityChange(t,e),this._moveAfter(t,i,r)):null!==(t=null===this._unlinkedRecords?null:this._unlinkedRecords.get(n,null))?(Y(t.item,e)||this._addIdentityChange(t,e),this._reinsertAfter(t,i,r)):t=this._addAfter(new dm(e,n),i,r),t},t.prototype._verifyReinsertion=function(t,e,n,r){var i=null===this._unlinkedRecords?null:this._unlinkedRecords.get(n,null);return null!==i?t=this._reinsertAfter(i,t._prev,r):t.currentIndex!=r&&(t.currentIndex=r,this._addToMoves(t,r)),t},t.prototype._truncate=function(t){for(;null!==t;){var e=t._next;this._addToRemovals(this._unlink(t)),t=e}null!==this._unlinkedRecords&&this._unlinkedRecords.clear(),null!==this._additionsTail&&(this._additionsTail._nextAdded=null),null!==this._movesTail&&(this._movesTail._nextMoved=null),null!==this._itTail&&(this._itTail._next=null),null!==this._removalsTail&&(this._removalsTail._nextRemoved=null),null!==this._identityChangesTail&&(this._identityChangesTail._nextIdentityChange=null)},t.prototype._reinsertAfter=function(t,e,n){null!==this._unlinkedRecords&&this._unlinkedRecords.remove(t);var r=t._prevRemoved,i=t._nextRemoved;return null===r?this._removalsHead=i:r._nextRemoved=i,null===i?this._removalsTail=r:i._prevRemoved=r,this._insertAfter(t,e,n),this._addToMoves(t,n),t},t.prototype._moveAfter=function(t,e,n){return this._unlink(t),this._insertAfter(t,e,n),this._addToMoves(t,n),t},t.prototype._addAfter=function(t,e,n){return this._insertAfter(t,e,n),null===this._additionsTail?this._additionsTail=this._additionsHead=t:this._additionsTail=this._additionsTail._nextAdded=t,t},t.prototype._insertAfter=function(t,e,n){var r=null===e?this._itHead:e._next;return t._next=r,t._prev=e,null===r?this._itTail=t:r._prev=t,null===e?this._itHead=t:e._next=t,null===this._linkedRecords&&(this._linkedRecords=new fm),this._linkedRecords.put(t),t.currentIndex=n,t},t.prototype._remove=function(t){return this._addToRemovals(this._unlink(t))},t.prototype._unlink=function(t){null!==this._linkedRecords&&this._linkedRecords.remove(t);var e=t._prev,n=t._next;return null===e?this._itHead=n:e._next=n,null===n?this._itTail=e:n._prev=e,t},t.prototype._addToMoves=function(t,e){return t.previousIndex===e?t:(null===this._movesTail?this._movesTail=this._movesHead=t:this._movesTail=this._movesTail._nextMoved=t,t)},t.prototype._addToRemovals=function(t){return null===this._unlinkedRecords&&(this._unlinkedRecords=new fm),this._unlinkedRecords.put(t),t.currentIndex=null,t._nextRemoved=null,null===this._removalsTail?(this._removalsTail=this._removalsHead=t,t._prevRemoved=null):(t._prevRemoved=this._removalsTail,this._removalsTail=this._removalsTail._nextRemoved=t),t},t.prototype._addIdentityChange=function(t,e){return t.item=e,null===this._identityChangesTail?this._identityChangesTail=this._identityChangesHead=t:this._identityChangesTail=this._identityChangesTail._nextIdentityChange=t,t},t}(),dm=function(){return function(t,e){this.item=t,this.trackById=e,this.currentIndex=null,this.previousIndex=null,this._nextPrevious=null,this._prev=null,this._next=null,this._prevDup=null,this._nextDup=null,this._prevRemoved=null,this._nextRemoved=null,this._nextAdded=null,this._nextMoved=null,this._nextIdentityChange=null}}(),pm=function(){function t(){this._head=null,this._tail=null}return t.prototype.add=function(t){null===this._head?(this._head=this._tail=t,t._nextDup=null,t._prevDup=null):(this._tail._nextDup=t,t._prevDup=this._tail,t._nextDup=null,this._tail=t)},t.prototype.get=function(t,e){var n;for(n=this._head;null!==n;n=n._nextDup)if((null===e||e<=n.currentIndex)&&Y(n.trackById,t))return n;return null},t.prototype.remove=function(t){var e=t._prevDup,n=t._nextDup;return null===e?this._head=n:e._nextDup=n,null===n?this._tail=e:n._prevDup=e,null===this._head},t}(),fm=function(){function t(){this.map=new Map}return t.prototype.put=function(t){var e=t.trackById,n=this.map.get(e);n||(n=new pm,this.map.set(e,n)),n.add(t)},t.prototype.get=function(t,e){var n=t,r=this.map.get(n);return r?r.get(t,e):null},t.prototype.remove=function(t){var e=t.trackById;return this.map.get(e).remove(t)&&this.map.delete(e),t},Object.defineProperty(t.prototype,"isEmpty",{get:function(){return 0===this.map.size},enumerable:!0,configurable:!0}),t.prototype.clear=function(){this.map.clear()},t}();function hm(t,e,n){var r=t.previousIndex;if(null===r)return r;var i=0;return n&&r<n.length&&(i=n[r]),r+e+i}var mm=function(){function t(){}return t.prototype.supports=function(t){return t instanceof Map||Dr(t)},t.prototype.create=function(){return new gm},t}(),gm=function(){function t(){this._records=new Map,this._mapHead=null,this._appendAfter=null,this._previousMapHead=null,this._changesHead=null,this._changesTail=null,this._additionsHead=null,this._additionsTail=null,this._removalsHead=null,this._removalsTail=null}return Object.defineProperty(t.prototype,"isDirty",{get:function(){return null!==this._additionsHead||null!==this._changesHead||null!==this._removalsHead},enumerable:!0,configurable:!0}),t.prototype.forEachItem=function(t){var e;for(e=this._mapHead;null!==e;e=e._next)t(e)},t.prototype.forEachPreviousItem=function(t){var e;for(e=this._previousMapHead;null!==e;e=e._nextPrevious)t(e)},t.prototype.forEachChangedItem=function(t){var e;for(e=this._changesHead;null!==e;e=e._nextChanged)t(e)},t.prototype.forEachAddedItem=function(t){var e;for(e=this._additionsHead;null!==e;e=e._nextAdded)t(e)},t.prototype.forEachRemovedItem=function(t){var e;for(e=this._removalsHead;null!==e;e=e._nextRemoved)t(e)},t.prototype.diff=function(t){if(t){if(!(t instanceof Map||Dr(t)))throw new Error("Error trying to diff '"+K(t)+"'. Only maps and objects are allowed")}else t=new Map;return this.check(t)?this:null},t.prototype.onDestroy=function(){},t.prototype.check=function(t){var e=this;this._reset();var n=this._mapHead;if(this._appendAfter=null,this._forEach(t,function(t,r){if(n&&n.key===r)e._maybeAddToChanges(n,t),e._appendAfter=n,n=n._next;else{var i=e._getOrCreateRecordForKey(r,t);n=e._insertBeforeOrAppend(n,i)}}),n){n._prev&&(n._prev._next=null),this._removalsHead=n;for(var r=n;null!==r;r=r._nextRemoved)r===this._mapHead&&(this._mapHead=null),this._records.delete(r.key),r._nextRemoved=r._next,r.previousValue=r.currentValue,r.currentValue=null,r._prev=null,r._next=null}return this._changesTail&&(this._changesTail._nextChanged=null),this._additionsTail&&(this._additionsTail._nextAdded=null),this.isDirty},t.prototype._insertBeforeOrAppend=function(t,e){if(t){var n=t._prev;return e._next=t,e._prev=n,t._prev=e,n&&(n._next=e),t===this._mapHead&&(this._mapHead=e),this._appendAfter=t,t}return this._appendAfter?(this._appendAfter._next=e,e._prev=this._appendAfter):this._mapHead=e,this._appendAfter=e,null},t.prototype._getOrCreateRecordForKey=function(t,e){if(this._records.has(t)){var n=this._records.get(t);this._maybeAddToChanges(n,e);var r=n._prev,i=n._next;return r&&(r._next=i),i&&(i._prev=r),n._next=null,n._prev=null,n}var o=new ym(t);return this._records.set(t,o),o.currentValue=e,this._addToAdditions(o),o},t.prototype._reset=function(){if(this.isDirty){var t=void 0;for(this._previousMapHead=this._mapHead,t=this._previousMapHead;null!==t;t=t._next)t._nextPrevious=t._next;for(t=this._changesHead;null!==t;t=t._nextChanged)t.previousValue=t.currentValue;for(t=this._additionsHead;null!=t;t=t._nextAdded)t.previousValue=t.currentValue;this._changesHead=this._changesTail=null,this._additionsHead=this._additionsTail=null,this._removalsHead=null}},t.prototype._maybeAddToChanges=function(t,e){Y(e,t.currentValue)||(t.previousValue=t.currentValue,t.currentValue=e,this._addToChanges(t))},t.prototype._addToAdditions=function(t){null===this._additionsHead?this._additionsHead=this._additionsTail=t:(this._additionsTail._nextAdded=t,this._additionsTail=t)},t.prototype._addToChanges=function(t){null===this._changesHead?this._changesHead=this._changesTail=t:(this._changesTail._nextChanged=t,this._changesTail=t)},t.prototype._forEach=function(t,e){t instanceof Map?t.forEach(e):Object.keys(t).forEach(function(n){return e(t[n],n)})},t}(),ym=function(){return function(t){this.key=t,this.previousValue=null,this.currentValue=null,this._nextPrevious=null,this._next=null,this._prev=null,this._nextAdded=null,this._nextRemoved=null,this._nextChanged=null}}(),_m=function(){function t(t){this.factories=t}return t.create=function(e,n){if(null!=n){var r=n.factories.slice();e=e.concat(r)}return new t(e)},t.extend=function(e){return{provide:t,useFactory:function(n){if(!n)throw new Error("Cannot extend IterableDiffers without a parent injector");return t.create(e,n)},deps:[[t,new Be,new Fe]]}},t.prototype.find=function(t){var e,n=this.factories.find(function(e){return e.supports(t)});if(null!=n)return n;throw new Error("Cannot find a differ supporting object '"+t+"' of type '"+((e=t).name||typeof e)+"'")},t.ngInjectableDef=g({providedIn:"root",factory:function(){return new t([new um])}}),t}();var vm,bm,xm=function(){function t(t){this.factories=t}return t.create=function(e,n){if(n){var r=n.factories.slice();e=e.concat(r)}return new t(e)},t.extend=function(e){return{provide:t,useFactory:function(n){if(!n)throw new Error("Cannot extend KeyValueDiffers without a parent injector");return t.create(e,n)},deps:[[t,new Be,new Fe]]}},t.prototype.find=function(t){var e=this.factories.find(function(e){return e.supports(t)});if(e)return e;throw new Error("Cannot find a differ supporting object '"+t+"'")},t.ngInjectableDef=g({providedIn:"root",factory:function(){return new t([new mm])}}),t}(),wm=[new mm],jm=[new um],Om=new _m(jm),Cm=new xm(wm),Em=Mh(null,"core",[{provide:Rf,useValue:"unknown"},{provide:Dh,deps:[Ls]},{provide:vh,deps:[]},{provide:Lf,deps:[]}]),Sm=new b("LocaleId"),Mm=new b("Translations"),Tm=new b("TranslationsFormat");function km(){return Om}function Am(){return Cm}function Dm(t){return t||"en-US"}(bm=vm||(vm={}))[bm.Error=0]="Error",bm[bm.Warning=1]="Warning",bm[bm.Ignore=2]="Ignore";var Im=[{provide:Ph,useClass:Ph,deps:[dh,Lf,Ls,of,ku,Tf]},{provide:Tf,useClass:Tf,deps:[[new Fe,Mf]]},{provide:Zf,useClass:Zf,deps:[]},Df,{provide:_m,useFactory:km,deps:[]},{provide:xm,useFactory:Am,deps:[]},{provide:Sm,useFactory:Dm,deps:[[new Ne(Sm),new Fe,new Be]]}],Pm=function(){function t(t){}return t=Object(r.__decorate)([zp({providers:Im}),Object(r.__metadata)("design:paramtypes",[Ph])],t)}(),Rm=!0,Nm=!1;function Fm(t,e,n){var r=t.state,i=1792&r;return i===e?(t.state=-1793&r|n,t.initIndex=-1,!0):i===n}function Lm(t,e,n){return(1792&t.state)===e&&t.initIndex<=n&&(t.initIndex=n+1,!0)}function Bm(t,e){return t.nodes[e]}function Vm(t,e){return t.nodes[e]}function zm(t,e){return t.nodes[e]}function Hm(t,e){return t.nodes[e]}function Um(t,e){return t.nodes[e]}var Wm=function(){return function(){}}(),qm={setCurrentNode:void 0,createRootView:void 0,createEmbeddedView:void 0,createComponentView:void 0,createNgModuleRef:void 0,overrideProvider:void 0,overrideComponentView:void 0,clearOverrides:void 0,checkAndUpdateView:void 0,checkNoChangesView:void 0,destroyView:void 0,resolveDep:void 0,createDebugContext:void 0,handleEvent:void 0,updateDirectives:void 0,updateRenderer:void 0,dirtyParentQueries:void 0};function Gm(t,e,n,r){var i,o,a="ExpressionChangedAfterItHasBeenCheckedError: Expression has changed after it was checked. Previous value: '"+e+"'. Current value: '"+n+"'.";return r&&(a+=" It seems like the view has been created after its parent and its children have been dirty checked. Has it been created in a change detection hook ?"),i=t,Ym(o=new Error(a),i),o}function Ym(t,e){t[$p]=e,t[tf]=e.logError.bind(e)}function Km(t){return new Error("ViewDestroyedError: Attempt to use a destroyed view: "+t)}var Xm=function(){},Qm=new Map;function Zm(t){var e=Qm.get(t);return e||(e=K(t)+"_"+Qm.size,Qm.set(t,e)),e}function $m(t,e,n,r){if(Tr.isWrapped(r)){r=Tr.unwrap(r);var i=t.def.nodes[e].bindingIndex+n,o=Tr.unwrap(t.oldValues[i]);t.oldValues[i]=new Tr(o)}return r}var Jm="$$undefined",tg="$$empty";function eg(t){return{id:Jm,styles:t.styles,encapsulation:t.encapsulation,data:t.data}}var ng=0;function rg(t,e,n,r){var i=t.oldValues;return!(!(2&t.state)&&Y(i[e.bindingIndex+n],r))}function ig(t,e,n,r){return!!rg(t,e,n,r)&&(t.oldValues[e.bindingIndex+n]=r,!0)}function og(t,e,n,r){var i=t.oldValues[e.bindingIndex+n];if(1&t.state||!Mr(i,r)){var o=e.bindings[n].name;throw Gm(qm.createDebugContext(t,e.nodeIndex),o+": "+i,o+": "+r,0!=(1&t.state))}}function ag(t){for(var e=t;e;)2&e.def.flags&&(e.state|=8),e=e.viewContainerParent||e.parent}function sg(t,e){for(var n=t;n&&n!==e;)n.state|=64,n=n.viewContainerParent||n.parent}function ug(t,e,n,r){try{return ag(33554432&t.def.nodes[e].flags?Vm(t,e).componentView:t),qm.handleEvent(t,e,n,r)}catch(e){t.root.errorHandler.handleError(e)}}function cg(t){return t.parent?Vm(t.parent,t.parentNodeDef.nodeIndex):null}function lg(t){return t.parent?t.parentNodeDef.parent:null}function dg(t,e){switch(201347067&e.flags){case 1:return Vm(t,e.nodeIndex).renderElement;case 2:return Bm(t,e.nodeIndex).renderText}}function pg(t,e){return t?t+":"+e:e}function fg(t){return!!t.parent&&!!(32768&t.parentNodeDef.flags)}function hg(t){return 1<<t%32}function mg(t){var e={},n=0,i={};return t&&t.forEach(function(t){var o=Object(r.__read)(t,2),a=o[0],s=o[1];"number"==typeof a?(e[a]=s,n|=hg(a)):i[a]=s}),{matchedQueries:e,references:i,matchedQueryIds:n}}function gg(t,e){return t.map(function(t){var n,i,o;return Array.isArray(t)?(o=(n=Object(r.__read)(t,2))[0],i=n[1]):(o=0,i=t),i&&("function"==typeof i||"object"==typeof i)&&e&&Object.defineProperty(i,Is,{value:e,configurable:!0}),{flags:o,token:i,tokenKey:Zm(i)}})}function yg(t,e,n){var r=n.renderParent;return r?0==(1&r.flags)||0==(33554432&r.flags)||r.element.componentRendererType&&r.element.componentRendererType.encapsulation===J.Native?Vm(t,n.renderParent.nodeIndex).renderElement:void 0:e}var _g=new WeakMap;function vg(t){var e=_g.get(t);return e||((e=t(function(){return Xm})).factory=t,_g.set(t,e)),e}function bg(t,e,n,r,i){3===e&&(n=t.renderer.parentNode(dg(t,t.def.lastRenderRootNode))),xg(t,e,0,t.def.nodes.length-1,n,r,i)}function xg(t,e,n,r,i,o,a){for(var s=n;s<=r;s++){var u=t.def.nodes[s];11&u.flags&&jg(t,u,e,i,o,a),s+=u.childCount}}function wg(t,e,n,r,i,o){for(var a=t;a&&!fg(a);)a=a.parent;for(var s=a.parent,u=lg(a),c=u.nodeIndex+1,l=u.nodeIndex+u.childCount,d=c;d<=l;d++){var p=s.def.nodes[d];p.ngContentIndex===e&&jg(s,p,n,r,i,o),d+=p.childCount}if(!s.parent){var f=t.root.projectableNodes[e];if(f)for(d=0;d<f.length;d++)Og(t,f[d],n,r,i,o)}}function jg(t,e,n,r,i,o){if(8&e.flags)wg(t,e.ngContent.index,n,r,i,o);else{var a=dg(t,e);if(3===n&&33554432&e.flags&&48&e.bindingFlags){if(16&e.bindingFlags&&Og(t,a,n,r,i,o),32&e.bindingFlags)Og(Vm(t,e.nodeIndex).componentView,a,n,r,i,o)}else Og(t,a,n,r,i,o);if(16777216&e.flags)for(var s=Vm(t,e.nodeIndex).viewContainer._embeddedViews,u=0;u<s.length;u++)bg(s[u],n,r,i,o);1&e.flags&&!e.element.name&&xg(t,n,e.nodeIndex+1,e.nodeIndex+e.childCount,r,i,o)}}function Og(t,e,n,r,i,o){var a=t.renderer;switch(n){case 1:a.appendChild(r,e);break;case 2:a.insertBefore(r,e,i);break;case 3:a.removeChild(r,e);break;case 0:o.push(e)}}var Cg=/^:([^:]+):(.+)$/;function Eg(t){if(":"===t[0]){var e=t.match(Cg);return[e[1],e[2]]}return["",t]}function Sg(t){for(var e=0,n=0;n<t.length;n++)e|=t[n].flags;return e}function Mg(t,e){for(var n="",r=0;r<2*t;r+=2)n=n+e[r]+kg(e[r+1]);return n+e[2*t]}function Tg(t,e,n,r,i,o,a,s,u,c,l,d,p,f,h,m,g,y,_,v){switch(t){case 1:return e+kg(n)+r;case 2:return e+kg(n)+r+kg(i)+o;case 3:return e+kg(n)+r+kg(i)+o+kg(a)+s;case 4:return e+kg(n)+r+kg(i)+o+kg(a)+s+kg(u)+c;case 5:return e+kg(n)+r+kg(i)+o+kg(a)+s+kg(u)+c+kg(l)+d;case 6:return e+kg(n)+r+kg(i)+o+kg(a)+s+kg(u)+c+kg(l)+d+kg(p)+f;case 7:return e+kg(n)+r+kg(i)+o+kg(a)+s+kg(u)+c+kg(l)+d+kg(p)+f+kg(h)+m;case 8:return e+kg(n)+r+kg(i)+o+kg(a)+s+kg(u)+c+kg(l)+d+kg(p)+f+kg(h)+m+kg(g)+y;case 9:return e+kg(n)+r+kg(i)+o+kg(a)+s+kg(u)+c+kg(l)+d+kg(p)+f+kg(h)+m+kg(g)+y+kg(_)+v;default:throw new Error("Does not support more than 9 expressions")}}function kg(t){return null!=t?t.toString():""}var Ag=[],Dg={};function Ig(t,e,n,r,i,o){t|=1;var a=mg(e),s=a.matchedQueries,u=a.references;return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,flags:t,checkIndex:-1,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:s,matchedQueryIds:a.matchedQueryIds,references:u,ngContentIndex:n,childCount:r,bindings:[],bindingFlags:0,outputs:[],element:{ns:null,name:null,attrs:null,template:o?vg(o):null,componentProvider:null,componentView:null,componentRendererType:null,publicProviders:null,allProviders:null,handleEvent:i||Xm},provider:null,text:null,query:null,ngContent:null}}function Pg(t,e,n,i,o,a,s,u,c,l,d,p){var f;void 0===s&&(s=[]),l||(l=Xm);var h=mg(n),m=h.matchedQueries,g=h.references,y=h.matchedQueryIds,_=null,v=null;a&&(_=(f=Object(r.__read)(Eg(a),2))[0],v=f[1]),u=u||[];for(var b=new Array(u.length),x=0;x<u.length;x++){var w=Object(r.__read)(u[x],3),j=w[0],O=w[1],C=w[2],E=Object(r.__read)(Eg(O),2),S=E[0],M=E[1],T=void 0,k=void 0;switch(15&j){case 4:k=C;break;case 1:case 8:T=C}b[x]={flags:j,ns:S,name:M,nonMinifiedName:M,securityContext:T,suffix:k}}c=c||[];var A=new Array(c.length);for(x=0;x<c.length;x++){var D=Object(r.__read)(c[x],2),I=D[0],P=D[1];A[x]={type:0,target:I,eventName:P,propName:null}}var R=(s=s||[]).map(function(t){var e=Object(r.__read)(t,2),n=e[0],i=e[1],o=Object(r.__read)(Eg(n),2);return[o[0],o[1],i]});return p=function(t){if(t&&t.id===Jm){var e=null!=t.encapsulation&&t.encapsulation!==J.None||t.styles.length||Object.keys(t.data).length;t.id=e?"c"+ng++:tg}return t&&t.id===tg&&(t=null),t||null}(p),d&&(e|=33554432),{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:t,flags:e|=1,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:m,matchedQueryIds:y,references:g,ngContentIndex:i,childCount:o,bindings:b,bindingFlags:Sg(b),outputs:A,element:{ns:_,name:v,attrs:R,template:null,componentProvider:null,componentView:d||null,componentRendererType:p,publicProviders:null,allProviders:null,handleEvent:l||Xm},provider:null,text:null,query:null,ngContent:null}}function Rg(t,e,n){var i,o=n.element,a=t.root.selectorOrNode,s=t.renderer;if(t.parent||!a){i=o.name?s.createElement(o.name,o.ns):s.createComment("");var u=yg(t,e,n);u&&s.appendChild(u,i)}else{var c=!!o.componentRendererType&&o.componentRendererType.encapsulation===J.ShadowDom;i=s.selectRootElement(a,c)}if(o.attrs)for(var l=0;l<o.attrs.length;l++){var d=Object(r.__read)(o.attrs[l],3),p=d[0],f=d[1],h=d[2];s.setAttribute(i,f,h,p)}return i}function Ng(t,e,n,r){for(var i=0;i<n.outputs.length;i++){var o=n.outputs[i],a=Fg(t,n.nodeIndex,pg(o.target,o.eventName)),s=o.target,u=t;"component"===o.target&&(s=null,u=e);var c=u.renderer.listen(s||r,o.eventName,a);t.disposables[n.outputIndex+i]=c}}function Fg(t,e,n){return function(r){return ug(t,e,n,r)}}function Lg(t,e,n,r){if(!ig(t,e,n,r))return!1;var i,o,a,s,u,c,l,d,p,f,h=e.bindings[n],m=Vm(t,e.nodeIndex),g=m.renderElement,y=h.name;switch(15&h.flags){case 1:!function(t,e,n,r,i,o){var a=e.securityContext,s=a?t.root.sanitizer.sanitize(a,o):o;s=null!=s?s.toString():null;var u=t.renderer;null!=o?u.setAttribute(n,i,s,r):u.removeAttribute(n,i,r)}(t,h,g,h.ns,y,r);break;case 2:l=g,d=y,p=r,f=t.renderer,p?f.addClass(l,d):f.removeClass(l,d);break;case 4:!function(t,e,n,r,i){var o=t.root.sanitizer.sanitize(Qu.STYLE,i);if(null!=o){o=o.toString();var a=e.suffix;null!=a&&(o+=a)}else o=null;var s=t.renderer;null!=o?s.setStyle(n,r,o):s.removeStyle(n,r)}(t,h,g,y,r);break;case 8:var _=33554432&e.flags&&32&h.flags?m.componentView:t;i=_,o=g,a=y,s=r,u=h.securityContext,c=u?i.root.sanitizer.sanitize(u,s):s,i.renderer.setProperty(o,a,c)}return!0}var Bg=new Object,Vg=Zm(Ls),zg=Zm(Ns),Hg=Zm(Iu);function Ug(t,e,n,r){return n=Z(n),{index:-1,deps:gg(r,K(e)),flags:t,token:e,value:n}}function Wg(t){for(var e={},n=[],r=!1,i=0;i<t.length;i++){var o=t[i];o.token===eu&&!0===o.value&&(r=!0),1073741824&o.flags&&n.push(o.token),o.index=i,e[Zm(o.token)]=o}return{factory:null,providersByKey:e,providers:t,modules:n,isRoot:r}}function qg(t,e,n){void 0===n&&(n=Ls.THROW_IF_NOT_FOUND);var r,i,o,a,s=We(t);try{if(8&e.flags)return e.token;if(2&e.flags&&(n=null),1&e.flags)return t._parent.get(e.token,n);var u=e.tokenKey;switch(u){case Vg:case zg:case Hg:return t}var c=t._def.providersByKey[u],l=void 0;if(c){var d=t._providers[c.index];return void 0===d&&(d=t._providers[c.index]=Gg(t,c)),d===Bg?void 0:d}if((l=_(e.token))&&(r=t,null!=(i=l).providedIn&&(o=r,a=i.providedIn,o._def.modules.indexOf(a)>-1||"root"===i.providedIn&&r._def.isRoot))){var p=t._providers.length;return t._def.providersByKey[e.tokenKey]={flags:5120,value:l.factory,deps:[],index:p,token:e.token},t._providers[p]=Bg,t._providers[p]=Gg(t,t._def.providersByKey[e.tokenKey])}return 4&e.flags?n:t._parent.get(e.token,n)}finally{We(s)}}function Gg(t,e){var n;switch(201347067&e.flags){case 512:n=function(t,e,n){var i=n.length;switch(i){case 0:return new e;case 1:return new e(qg(t,n[0]));case 2:return new e(qg(t,n[0]),qg(t,n[1]));case 3:return new e(qg(t,n[0]),qg(t,n[1]),qg(t,n[2]));default:for(var o=new Array(i),a=0;a<i;a++)o[a]=qg(t,n[a]);return new(e.bind.apply(e,Object(r.__spread)([void 0],o)))}}(t,e.value,e.deps);break;case 1024:n=function(t,e,n){var i=n.length;switch(i){case 0:return e();case 1:return e(qg(t,n[0]));case 2:return e(qg(t,n[0]),qg(t,n[1]));case 3:return e(qg(t,n[0]),qg(t,n[1]),qg(t,n[2]));default:for(var o=Array(i),a=0;a<i;a++)o[a]=qg(t,n[a]);return e.apply(void 0,Object(r.__spread)(o))}}(t,e.value,e.deps);break;case 2048:n=qg(t,e.deps[0]);break;case 256:n=e.value}return n===Bg||null==n||"object"!=typeof n||131072&e.flags||"function"!=typeof n.ngOnDestroy||(e.flags|=131072),void 0===n?Bg:n}function Yg(t,e,n,r){var i=e.viewContainer._embeddedViews;null!==n&&void 0!==n||(n=i.length),r.viewContainerParent=t,Zg(i,n,r),function(t,e){var n=cg(e);if(!n||n===t||16&e.state)return;e.state|=16;var r=n.template._projectedViews;r||(r=n.template._projectedViews=[]);r.push(e),function(t,e){if(4&e.flags)return;t.nodeFlags|=4,e.flags|=4;var n=e.parent;for(;n;)n.childFlags|=4,n=n.parent}(e.parent.def,e.parentNodeDef)}(e,r),qm.dirtyParentQueries(r),Xg(e,n>0?i[n-1]:null,r)}function Kg(t,e){var n=t.viewContainer._embeddedViews;if((null==e||e>=n.length)&&(e=n.length-1),e<0)return null;var r=n[e];return r.viewContainerParent=null,$g(n,e),qm.dirtyParentQueries(r),Qg(r),r}function Xg(t,e,n){var r=e?dg(e,e.def.lastRenderRootNode):t.renderElement;bg(n,2,n.renderer.parentNode(r),n.renderer.nextSibling(r),void 0)}function Qg(t){bg(t,3,null,null,void 0)}function Zg(t,e,n){e>=t.length?t.push(n):t.splice(e,0,n)}function $g(t,e){e>=t.length-1?t.pop():t.splice(e,1)}var Jg=new Object;function ty(t,e,n,r,i,o){return new ny(t,e,n,r,i,o)}function ey(t){return t.viewDefFactory}var ny=function(t){function e(e,n,r,i,o,a){var s=t.call(this)||this;return s.selector=e,s.componentType=n,s._inputs=i,s._outputs=o,s.ngContentSelectors=a,s.viewDefFactory=r,s}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"inputs",{get:function(){var t=[],e=this._inputs;for(var n in e){var r=e[n];t.push({propName:n,templateName:r})}return t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"outputs",{get:function(){var t=[];for(var e in this._outputs){var n=this._outputs[e];t.push({propName:e,templateName:n})}return t},enumerable:!0,configurable:!0}),e.prototype.create=function(t,e,n,r){if(!r)throw new Error("ngModule should be provided");var i=vg(this.viewDefFactory),o=i.nodes[0].element.componentProvider.nodeIndex,a=qm.createRootView(t,e||[],n,i,r,Jg),s=zm(a,o).instance;return n&&a.renderer.setAttribute(Vm(a,0).renderElement,"ng-version",rc.full),new ry(a,new oy(a),s)},e}(ju),ry=function(t){function e(e,n,r){var i=t.call(this)||this;return i._view=e,i._viewRef=n,i._component=r,i._elDef=i._view.def.nodes[0],i.hostView=n,i.changeDetectorRef=n,i.instance=r,i}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"location",{get:function(){return new Hu(Vm(this._view,this._elDef.nodeIndex).renderElement)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"injector",{get:function(){return new uy(this._view,this._elDef)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"componentType",{get:function(){return this._component.constructor},enumerable:!0,configurable:!0}),e.prototype.destroy=function(){this._viewRef.destroy()},e.prototype.onDestroy=function(t){this._viewRef.onDestroy(t)},e}(wu);var iy=function(){function t(t,e,n){this._view=t,this._elDef=e,this._data=n,this._embeddedViews=[]}return Object.defineProperty(t.prototype,"element",{get:function(){return new Hu(this._data.renderElement)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"injector",{get:function(){return new uy(this._view,this._elDef)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"parentInjector",{get:function(){for(var t=this._view,e=this._elDef.parent;!e&&t;)e=lg(t),t=t.parent;return t?new uy(t,e):new uy(this._view,null)},enumerable:!0,configurable:!0}),t.prototype.clear=function(){for(var t=this._embeddedViews.length-1;t>=0;t--){var e=Kg(this._data,t);qm.destroyView(e)}},t.prototype.get=function(t){var e=this._embeddedViews[t];if(e){var n=new oy(e);return n.attachToViewContainerRef(this),n}return null},Object.defineProperty(t.prototype,"length",{get:function(){return this._embeddedViews.length},enumerable:!0,configurable:!0}),t.prototype.createEmbeddedView=function(t,e,n){var r=t.createEmbeddedView(e||{});return this.insert(r,n),r},t.prototype.createComponent=function(t,e,n,r,i){var o=n||this.parentInjector;i||t instanceof Du||(i=o.get(Iu));var a=t.create(o,r,void 0,i);return this.insert(a.hostView,e),a},t.prototype.insert=function(t,e){if(t.destroyed)throw new Error("Cannot insert a destroyed View in a ViewContainer!");var n=t,r=n._view;return Yg(this._view,this._data,e,r),n.attachToViewContainerRef(this),t},t.prototype.move=function(t,e){if(t.destroyed)throw new Error("Cannot move a destroyed View in a ViewContainer!");var n,r,i,o,a,s=this._embeddedViews.indexOf(t._view);return n=this._data,r=s,i=e,o=n.viewContainer._embeddedViews,a=o[r],$g(o,r),null==i&&(i=o.length),Zg(o,i,a),qm.dirtyParentQueries(a),Qg(a),Xg(n,i>0?o[i-1]:null,a),t},t.prototype.indexOf=function(t){return this._embeddedViews.indexOf(t._view)},t.prototype.remove=function(t){var e=Kg(this._data,t);e&&qm.destroyView(e)},t.prototype.detach=function(t){var e=Kg(this._data,t);return e?new oy(e):null},t}();var oy=function(){function t(t){this._view=t,this._viewContainerRef=null,this._appRef=null}return Object.defineProperty(t.prototype,"rootNodes",{get:function(){return bg(this._view,0,void 0,void 0,t=[]),t;var t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"context",{get:function(){return this._view.context},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"destroyed",{get:function(){return 0!=(128&this._view.state)},enumerable:!0,configurable:!0}),t.prototype.markForCheck=function(){ag(this._view)},t.prototype.detach=function(){this._view.state&=-5},t.prototype.detectChanges=function(){var t=this._view.root.rendererFactory;t.begin&&t.begin();try{qm.checkAndUpdateView(this._view)}finally{t.end&&t.end()}},t.prototype.checkNoChanges=function(){qm.checkNoChangesView(this._view)},t.prototype.reattach=function(){this._view.state|=4},t.prototype.onDestroy=function(t){this._view.disposables||(this._view.disposables=[]),this._view.disposables.push(t)},t.prototype.destroy=function(){this._appRef?this._appRef.detachView(this):this._viewContainerRef&&this._viewContainerRef.detach(this._viewContainerRef.indexOf(this)),qm.destroyView(this._view)},t.prototype.detachFromAppRef=function(){this._appRef=null,Qg(this._view),qm.dirtyParentQueries(this._view)},t.prototype.attachToAppRef=function(t){if(this._viewContainerRef)throw new Error("This view is already attached to a ViewContainer!");this._appRef=t},t.prototype.attachToViewContainerRef=function(t){if(this._appRef)throw new Error("This view is already attached directly to the ApplicationRef!");this._viewContainerRef=t},t}();var ay=function(t){function e(e,n){var r=t.call(this)||this;return r._parentView=e,r._def=n,r}return Object(r.__extends)(e,t),e.prototype.createEmbeddedView=function(t){return new oy(qm.createEmbeddedView(this._parentView,this._def,this._def.element.template,t))},Object.defineProperty(e.prototype,"elementRef",{get:function(){return new Hu(Vm(this._parentView,this._def.nodeIndex).renderElement)},enumerable:!0,configurable:!0}),e}(Kl);function sy(t,e){return new uy(t,e)}var uy=function(){function t(t,e){this.view=t,this.elDef=e}return t.prototype.get=function(t,e){void 0===e&&(e=Ls.THROW_IF_NOT_FOUND);var n=!!this.elDef&&0!=(33554432&this.elDef.flags);return qm.resolveDep(this.view,this.elDef,n,{flags:0,token:t,tokenKey:Zm(t)},e)},t}();function cy(t,e){var n=t.def.nodes[e];if(1&n.flags){var r=Vm(t,n.nodeIndex);return n.element.template?r.template:r.renderElement}if(2&n.flags)return Bm(t,n.nodeIndex).renderText;if(20240&n.flags)return zm(t,n.nodeIndex).instance;throw new Error("Illegal state: read nodeValue for node index "+e)}var ly=function(){function t(t){this.delegate=t}return t.prototype.selectRootElement=function(t){return this.delegate.selectRootElement(t)},t.prototype.createElement=function(t,e){var n=Object(r.__read)(Eg(e),2),i=n[0],o=n[1],a=this.delegate.createElement(o,i);return t&&this.delegate.appendChild(t,a),a},t.prototype.createViewRoot=function(t){return t},t.prototype.createTemplateAnchor=function(t){var e=this.delegate.createComment("");return t&&this.delegate.appendChild(t,e),e},t.prototype.createText=function(t,e){var n=this.delegate.createText(e);return t&&this.delegate.appendChild(t,n),n},t.prototype.projectNodes=function(t,e){for(var n=0;n<e.length;n++)this.delegate.appendChild(t,e[n])},t.prototype.attachViewAfter=function(t,e){for(var n=this.delegate.parentNode(t),r=this.delegate.nextSibling(t),i=0;i<e.length;i++)this.delegate.insertBefore(n,e[i],r)},t.prototype.detachView=function(t){for(var e=0;e<t.length;e++){var n=t[e],r=this.delegate.parentNode(n);this.delegate.removeChild(r,n)}},t.prototype.destroyView=function(t,e){for(var n=0;n<e.length;n++)this.delegate.destroyNode(e[n])},t.prototype.listen=function(t,e,n){return this.delegate.listen(t,e,n)},t.prototype.listenGlobal=function(t,e,n){return this.delegate.listen(t,e,n)},t.prototype.setElementProperty=function(t,e,n){this.delegate.setProperty(t,e,n)},t.prototype.setElementAttribute=function(t,e,n){var i=Object(r.__read)(Eg(e),2),o=i[0],a=i[1];null!=n?this.delegate.setAttribute(t,a,n,o):this.delegate.removeAttribute(t,a,o)},t.prototype.setBindingDebugInfo=function(t,e,n){},t.prototype.setElementClass=function(t,e,n){n?this.delegate.addClass(t,e):this.delegate.removeClass(t,e)},t.prototype.setElementStyle=function(t,e,n){null!=n?this.delegate.setStyle(t,e,n):this.delegate.removeStyle(t,e)},t.prototype.invokeElementMethod=function(t,e,n){t[e].apply(t,n)},t.prototype.setText=function(t,e){this.delegate.setValue(t,e)},t.prototype.animate=function(){throw new Error("Renderer.animate is no longer supported!")},t}();function dy(t,e,n,r){return new py(t,e,n,r)}var py=function(){function t(t,e,n,r){this._moduleType=t,this._parent=e,this._bootstrapComponents=n,this._def=r,this._destroyListeners=[],this._destroyed=!1,this.injector=this,function(t){for(var e=t._def,n=t._providers=new Array(e.providers.length),r=0;r<e.providers.length;r++){var i=e.providers[r];4096&i.flags||void 0===n[r]&&(n[r]=Gg(t,i))}}(this)}return t.prototype.get=function(t,e,n){void 0===e&&(e=Ls.THROW_IF_NOT_FOUND),void 0===n&&(n=Pe.Default);var r=0;return n&Pe.SkipSelf?r|=1:n&Pe.Self&&(r|=4),qg(this,{token:t,tokenKey:Zm(t),flags:r},e)},Object.defineProperty(t.prototype,"instance",{get:function(){return this.get(this._moduleType)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"componentFactoryResolver",{get:function(){return this.get(ku)},enumerable:!0,configurable:!0}),t.prototype.destroy=function(){if(this._destroyed)throw new Error("The ng module "+K(this.instance.constructor)+" has already been destroyed.");this._destroyed=!0,function(t,e){for(var n=t._def,r=new Set,i=0;i<n.providers.length;i++)if(131072&n.providers[i].flags){var o=t._providers[i];if(o&&o!==Bg){var a=o.ngOnDestroy;"function"!=typeof a||r.has(o)||(a.apply(o),r.add(o))}}}(this),this._destroyListeners.forEach(function(t){return t()})},t.prototype.onDestroy=function(t){this._destroyListeners.push(t)},t}(),fy=Zm(Yu),hy=Zm($u),my=Zm(Hu),gy=Zm(zh),yy=Zm(Kl),_y=Zm(Wh),vy=Zm(Ls),by=Zm(Ns);function xy(t,e,n,i,o,a,s,u){var c=[];if(s)for(var l in s){var d=Object(r.__read)(s[l],2),p=d[0],f=d[1];c[p]={flags:8,name:l,nonMinifiedName:f,ns:null,securityContext:null,suffix:null}}var h=[];if(u)for(var m in u)h.push({type:1,propName:m,target:null,eventName:u[m]});return Oy(t,e|=16384,n,i,o,o,a,c,h)}function wy(t,e,n){return Oy(-1,t|=16,null,0,e,e,n)}function jy(t,e,n,r,i){return Oy(-1,t,e,0,n,r,i)}function Oy(t,e,n,r,i,o,a,s,u){var c=mg(n),l=c.matchedQueries,d=c.references,p=c.matchedQueryIds;u||(u=[]),s||(s=[]),o=Z(o);var f=gg(a,K(i));return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:t,flags:e,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:l,matchedQueryIds:p,references:d,ngContentIndex:-1,childCount:r,bindings:s,bindingFlags:Sg(s),outputs:u,element:null,provider:{token:i,value:o,deps:f},text:null,query:null,ngContent:null}}function Cy(t,e){for(var n=t;n.parent&&!fg(n);)n=n.parent;return Ty(n.parent,lg(n),!0,e.provider.value,e.provider.deps)}function Ey(t,e){var n=(32768&e.flags)>0,r=Ty(t,e.parent,n,e.provider.value,e.provider.deps);if(e.outputs.length)for(var i=0;i<e.outputs.length;i++){var o=e.outputs[i],a=r[o.propName];if(!Sf(a))throw new Error("@Output "+o.propName+" not initialized in '"+r.constructor.name+"'.");var s=a.subscribe(Sy(t,e.parent.nodeIndex,o.eventName));t.disposables[e.outputIndex+i]=s.unsubscribe.bind(s)}return r}function Sy(t,e,n){return function(r){return ug(t,e,n,r)}}function My(t,e){var n=(8192&e.flags)>0,i=e.provider;switch(201347067&e.flags){case 512:return Ty(t,e.parent,n,i.value,i.deps);case 1024:return function(t,e,n,i,o){var a=o.length;switch(a){case 0:return i();case 1:return i(Iy(t,e,n,o[0]));case 2:return i(Iy(t,e,n,o[0]),Iy(t,e,n,o[1]));case 3:return i(Iy(t,e,n,o[0]),Iy(t,e,n,o[1]),Iy(t,e,n,o[2]));default:for(var s=Array(a),u=0;u<a;u++)s[u]=Iy(t,e,n,o[u]);return i.apply(void 0,Object(r.__spread)(s))}}(t,e.parent,n,i.value,i.deps);case 2048:return Iy(t,e.parent,n,i.deps[0]);case 256:return i.value}}function Ty(t,e,n,i,o){var a=o.length;switch(a){case 0:return new i;case 1:return new i(Iy(t,e,n,o[0]));case 2:return new i(Iy(t,e,n,o[0]),Iy(t,e,n,o[1]));case 3:return new i(Iy(t,e,n,o[0]),Iy(t,e,n,o[1]),Iy(t,e,n,o[2]));default:for(var s=new Array(a),u=0;u<a;u++)s[u]=Iy(t,e,n,o[u]);return new(i.bind.apply(i,Object(r.__spread)([void 0],s)))}}var ky,Ay,Dy={};function Iy(t,e,n,r,i){if(void 0===i&&(i=Ls.THROW_IF_NOT_FOUND),8&r.flags)return r.token;var o=t;2&r.flags&&(i=null);var a=r.tokenKey;a===_y&&(n=!(!e||!e.element.componentView)),e&&1&r.flags&&(n=!1,e=e.parent);for(var s=t;s;){if(e)switch(a){case fy:var u=Py(s,e,n);return new ly(u.renderer);case hy:return(u=Py(s,e,n)).renderer;case my:return new Hu(Vm(s,e.nodeIndex).renderElement);case gy:return Vm(s,e.nodeIndex).viewContainer;case yy:if(e.element.template)return Vm(s,e.nodeIndex).template;break;case _y:var c=Py(s,e,n);return new oy(c);case vy:case by:return sy(s,e);default:var l=(n?e.element.allProviders:e.element.publicProviders)[a];if(l){var d=zm(s,l.nodeIndex);return d||(d={instance:My(s,l)},s.nodes[l.nodeIndex]=d),d.instance}}n=fg(s),e=lg(s),s=s.parent,4&r.flags&&(s=null)}var p=o.root.injector.get(r.token,Dy);return p!==Dy||i===Dy?p:o.root.ngModule.injector.get(r.token,i)}function Py(t,e,n){var r;if(n)r=Vm(t,e.nodeIndex).componentView;else for(r=t;r.parent&&!fg(r);)r=r.parent;return r}function Ry(t,e,n,r,i,o){if(32768&n.flags){var a=Vm(t,n.parent.nodeIndex).componentView;2&a.def.flags&&(a.state|=8)}var s=n.bindings[r].name;if(e.instance[s]=i,524288&n.flags){o=o||{};var u=Tr.unwrap(t.oldValues[n.bindingIndex+r]);o[n.bindings[r].nonMinifiedName]=new kr(u,i,0!=(2&t.state))}return t.oldValues[n.bindingIndex+r]=i,o}function Ny(t,e){if(t.def.nodeFlags&e)for(var n=t.def.nodes,r=0,i=0;i<n.length;i++){var o=n[i],a=o.parent;for(!a&&o.flags&e&&Ly(t,i,o.flags&e,r++),0==(o.childFlags&e)&&(i+=o.childCount);a&&1&a.flags&&i===a.nodeIndex+a.childCount;)a.directChildFlags&e&&(r=Fy(t,a,e,r)),a=a.parent}}function Fy(t,e,n,r){for(var i=e.nodeIndex+1;i<=e.nodeIndex+e.childCount;i++){var o=t.def.nodes[i];o.flags&n&&Ly(t,i,o.flags&n,r++),i+=o.childCount}return r}function Ly(t,e,n,r){var i=zm(t,e);if(i){var o=i.instance;o&&(qm.setCurrentNode(t,e),1048576&n&&Lm(t,512,r)&&o.ngAfterContentInit(),2097152&n&&o.ngAfterContentChecked(),4194304&n&&Lm(t,768,r)&&o.ngAfterViewInit(),8388608&n&&o.ngAfterViewChecked(),131072&n&&o.ngOnDestroy())}}function By(t,e,n){var r=[];for(var i in n){var o=n[i];r.push({propName:i,bindingType:o})}return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:-1,flags:t,childFlags:0,directChildFlags:0,childMatchedQueries:0,ngContentIndex:-1,matchedQueries:{},matchedQueryIds:0,references:{},childCount:0,bindings:[],bindingFlags:0,outputs:[],element:null,provider:null,text:null,query:{id:e,filterId:hg(e),bindings:r},ngContent:null}}function Vy(t){for(var e,n=t.def.nodeMatchedQueries;t.parent&&((e=t).parent&&!(32768&e.parentNodeDef.flags));){var r=t.parentNodeDef;t=t.parent;for(var i=r.nodeIndex+r.childCount,o=0;o<=i;o++){67108864&(a=t.def.nodes[o]).flags&&536870912&a.flags&&(a.query.filterId&n)===a.query.filterId&&Um(t,o).setDirty(),!(1&a.flags&&o+a.childCount<r.nodeIndex)&&67108864&a.childFlags&&536870912&a.childFlags||(o+=a.childCount)}}if(134217728&t.def.nodeFlags)for(o=0;o<t.def.nodes.length;o++){var a;134217728&(a=t.def.nodes[o]).flags&&536870912&a.flags&&Um(t,o).setDirty(),o+=a.childCount}}function zy(t,e){var n=Um(t,e.nodeIndex);if(n.dirty){var r,i=void 0;if(67108864&e.flags){var o=e.parent.parent;i=Hy(t,o.nodeIndex,o.nodeIndex+o.childCount,e.query,[]),r=zm(t,e.parent.nodeIndex).instance}else 134217728&e.flags&&(i=Hy(t,0,t.def.nodes.length-1,e.query,[]),r=t.component);n.reset(i);for(var a=e.query.bindings,s=!1,u=0;u<a.length;u++){var c=a[u],l=void 0;switch(c.bindingType){case 0:l=n.first;break;case 1:l=n,s=!0}r[c.propName]=l}s&&n.notifyOnChanges()}}function Hy(t,e,n,r,i){for(var o=e;o<=n;o++){var a=t.def.nodes[o],s=a.matchedQueries[r.id];if(null!=s&&i.push(Uy(t,a,s)),1&a.flags&&a.element.template&&(a.element.template.nodeMatchedQueries&r.filterId)===r.filterId){var u=Vm(t,o);if((a.childMatchedQueries&r.filterId)===r.filterId&&(Hy(t,o+1,o+a.childCount,r,i),o+=a.childCount),16777216&a.flags)for(var c=u.viewContainer._embeddedViews,l=0;l<c.length;l++){var d=c[l],p=cg(d);p&&p===u&&Hy(d,0,d.def.nodes.length-1,r,i)}var f=u.template._projectedViews;if(f)for(l=0;l<f.length;l++){var h=f[l];Hy(h,0,h.def.nodes.length-1,r,i)}}(a.childMatchedQueries&r.filterId)!==r.filterId&&(o+=a.childCount)}return i}function Uy(t,e,n){if(null!=n)switch(n){case 1:return Vm(t,e.nodeIndex).renderElement;case 0:return new Hu(Vm(t,e.nodeIndex).renderElement);case 2:return Vm(t,e.nodeIndex).template;case 3:return Vm(t,e.nodeIndex).viewContainer;case 4:return zm(t,e.nodeIndex).instance}}function Wy(t,e){return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:-1,flags:8,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:{},matchedQueryIds:0,references:{},ngContentIndex:t,childCount:0,bindings:[],bindingFlags:0,outputs:[],element:null,provider:null,text:null,query:null,ngContent:{index:e}}}function qy(t,e){return Ky(128,t,new Array(e+1))}function Gy(t,e){return Ky(32,t,new Array(e))}function Yy(t,e){for(var n=Object.keys(e),r=n.length,i=new Array(r),o=0;o<r;o++){var a=n[o];i[e[a]]=a}return Ky(64,t,i)}function Ky(t,e,n){for(var r=new Array(n.length),i=0;i<n.length;i++){var o=n[i];r[i]={flags:8,name:o,ns:null,nonMinifiedName:o,securityContext:null,suffix:null}}return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:e,flags:t,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:{},matchedQueryIds:0,references:{},ngContentIndex:-1,childCount:0,bindings:r,bindingFlags:Sg(r),outputs:[],element:null,provider:null,text:null,query:null,ngContent:null}}function Xy(t,e,n){for(var r=new Array(n.length-1),i=1;i<n.length;i++)r[i-1]={flags:8,name:null,ns:null,nonMinifiedName:null,securityContext:null,suffix:n[i]};return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:t,flags:2,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:{},matchedQueryIds:0,references:{},ngContentIndex:e,childCount:0,bindings:r,bindingFlags:8,outputs:[],element:null,provider:null,text:{prefix:n[0]},query:null,ngContent:null}}function Qy(t,e,n){var r,i=t.renderer;r=i.createText(n.text.prefix);var o=yg(t,e,n);return o&&i.appendChild(o,r),{renderText:r}}function Zy(t,e){return(null!=t?t.toString():"")+e.suffix}function $y(t,e,n,r){for(var i=0,o=0,a=0,s=0,u=0,c=null,l=null,d=!1,p=!1,f=null,h=0;h<e.length;h++){var m=e[h];if(m.nodeIndex=h,m.parent=c,m.bindingIndex=i,m.outputIndex=o,m.renderParent=l,a|=m.flags,u|=m.matchedQueryIds,m.element){var g=m.element;g.publicProviders=c?c.element.publicProviders:Object.create(null),g.allProviders=g.publicProviders,d=!1,p=!1,m.element.template&&(u|=m.element.template.nodeMatchedQueries)}if(t_(c,m,e.length),i+=m.bindings.length,o+=m.outputs.length,!l&&3&m.flags&&(f=m),20224&m.flags){d||(d=!0,c.element.publicProviders=Object.create(c.element.publicProviders),c.element.allProviders=c.element.publicProviders);var y=0!=(8192&m.flags),_=0!=(32768&m.flags);!y||_?c.element.publicProviders[Zm(m.provider.token)]=m:(p||(p=!0,c.element.allProviders=Object.create(c.element.publicProviders)),c.element.allProviders[Zm(m.provider.token)]=m),_&&(c.element.componentProvider=m)}if(c?(c.childFlags|=m.flags,c.directChildFlags|=m.flags,c.childMatchedQueries|=m.matchedQueryIds,m.element&&m.element.template&&(c.childMatchedQueries|=m.element.template.nodeMatchedQueries)):s|=m.flags,m.childCount>0)c=m,Jy(m)||(l=m);else for(;c&&h===c.nodeIndex+c.childCount;){var v=c.parent;v&&(v.childFlags|=c.childFlags,v.childMatchedQueries|=c.childMatchedQueries),l=(c=v)&&Jy(c)?c.renderParent:c}}return{factory:null,nodeFlags:a,rootNodeFlags:s,nodeMatchedQueries:u,flags:t,nodes:e,updateDirectives:n||Xm,updateRenderer:r||Xm,handleEvent:function(t,n,r,i){return e[n].element.handleEvent(t,r,i)},bindingCount:i,outputCount:o,lastRenderRootNode:f}}function Jy(t){return 0!=(1&t.flags)&&null===t.element.name}function t_(t,e,n){var r=e.element&&e.element.template;if(r){if(!r.lastRenderRootNode)throw new Error("Illegal State: Embedded templates without nodes are not allowed!");if(r.lastRenderRootNode&&16777216&r.lastRenderRootNode.flags)throw new Error("Illegal State: Last root node of a template can't have embedded views, at index "+e.nodeIndex+"!")}if(20224&e.flags&&0==(1&(t?t.flags:0)))throw new Error("Illegal State: StaticProvider/Directive nodes need to be children of elements or anchors, at index "+e.nodeIndex+"!");if(e.query){if(67108864&e.flags&&(!t||0==(16384&t.flags)))throw new Error("Illegal State: Content Query nodes need to be children of directives, at index "+e.nodeIndex+"!");if(134217728&e.flags&&t)throw new Error("Illegal State: View Query nodes have to be top level nodes, at index "+e.nodeIndex+"!")}if(e.childCount){var i=t?t.nodeIndex+t.childCount:n-1;if(e.nodeIndex<=i&&e.nodeIndex+e.childCount>i)throw new Error("Illegal State: childCount of node leads outside of parent, at index "+e.nodeIndex+"!")}}function e_(t,e,n,r){var i=i_(t.root,t.renderer,t,e,n);return o_(i,t.component,r),a_(i),i}function n_(t,e,n){var r=i_(t,t.renderer,null,null,e);return o_(r,n,n),a_(r),r}function r_(t,e,n,r){var i,o=e.element.componentRendererType;return i=o?t.root.rendererFactory.createRenderer(r,o):t.root.renderer,i_(t.root,i,t,e.element.componentProvider,n)}function i_(t,e,n,r,i){var o=new Array(i.nodes.length),a=i.outputCount?new Array(i.outputCount):null;return{def:i,parent:n,viewContainerParent:null,parentNodeDef:r,context:null,component:null,nodes:o,state:13,root:t,renderer:e,oldValues:new Array(i.bindingCount),disposables:a,initIndex:-1}}function o_(t,e,n){t.component=e,t.context=n}function a_(t){var e;if(fg(t)){var n=t.parentNodeDef;e=Vm(t.parent,n.parent.nodeIndex).renderElement}for(var r,i,o,a,s,u=t.def,c=t.nodes,l=0;l<u.nodes.length;l++){var d=u.nodes[l];qm.setCurrentNode(t,l);var p=void 0;switch(201347067&d.flags){case 1:var f=Rg(t,e,d),h=void 0;if(33554432&d.flags){var m=vg(d.element.componentView);h=qm.createComponentView(t,d,m,f)}Ng(t,h,d,f),p={renderElement:f,componentView:h,viewContainer:null,template:d.element.template?(a=t,s=d,new ay(a,s)):void 0},16777216&d.flags&&(p.viewContainer=new iy(t,d,p));break;case 2:p=Qy(t,e,d);break;case 512:case 1024:case 2048:case 256:if(!((p=c[l])||4096&d.flags))p={instance:My(t,d)};break;case 16:p={instance:Cy(t,d)};break;case 16384:if(!(p=c[l]))p={instance:Ey(t,d)};if(32768&d.flags)o_(Vm(t,d.parent.nodeIndex).componentView,p.instance,p.instance);break;case 32:case 64:case 128:p={value:void 0};break;case 67108864:case 134217728:p=new Nh;break;case 8:void 0,(o=yg(r=t,e,i=d))&&wg(r,i.ngContent.index,1,o,null,void 0),p=void 0}c[l]=p}h_(t,ky.CreateViewNodes),__(t,201326592,268435456,0)}function s_(t){l_(t),qm.updateDirectives(t,1),m_(t,ky.CheckNoChanges),qm.updateRenderer(t,1),h_(t,ky.CheckNoChanges),t.state&=-97}function u_(t){1&t.state?(t.state&=-2,t.state|=2):t.state&=-3,Fm(t,0,256),l_(t),qm.updateDirectives(t,0),m_(t,ky.CheckAndUpdate),__(t,67108864,536870912,0);var e=Fm(t,256,512);Ny(t,2097152|(e?1048576:0)),qm.updateRenderer(t,0),h_(t,ky.CheckAndUpdate),__(t,134217728,536870912,0),Ny(t,8388608|((e=Fm(t,512,768))?4194304:0)),2&t.def.flags&&(t.state&=-9),t.state&=-97,Fm(t,768,1024)}function c_(t,e,n,i,o,a,s,u,c,l,d,p,f){return 0===n?function(t,e,n,r,i,o,a,s,u,c,l,d){switch(201347067&e.flags){case 1:return T=t,A=n,D=r,I=i,P=o,R=a,N=s,F=u,L=c,B=l,V=d,z=(k=e).bindings.length,H=!1,z>0&&Lg(T,k,0,A)&&(H=!0),z>1&&Lg(T,k,1,D)&&(H=!0),z>2&&Lg(T,k,2,I)&&(H=!0),z>3&&Lg(T,k,3,P)&&(H=!0),z>4&&Lg(T,k,4,R)&&(H=!0),z>5&&Lg(T,k,5,N)&&(H=!0),z>6&&Lg(T,k,6,F)&&(H=!0),z>7&&Lg(T,k,7,L)&&(H=!0),z>8&&Lg(T,k,8,B)&&(H=!0),z>9&&Lg(T,k,9,V)&&(H=!0),H;case 2:return function(t,e,n,r,i,o,a,s,u,c,l,d){var p=!1,f=e.bindings,h=f.length;if(h>0&&ig(t,e,0,n)&&(p=!0),h>1&&ig(t,e,1,r)&&(p=!0),h>2&&ig(t,e,2,i)&&(p=!0),h>3&&ig(t,e,3,o)&&(p=!0),h>4&&ig(t,e,4,a)&&(p=!0),h>5&&ig(t,e,5,s)&&(p=!0),h>6&&ig(t,e,6,u)&&(p=!0),h>7&&ig(t,e,7,c)&&(p=!0),h>8&&ig(t,e,8,l)&&(p=!0),h>9&&ig(t,e,9,d)&&(p=!0),p){var m=e.text.prefix;h>0&&(m+=Zy(n,f[0])),h>1&&(m+=Zy(r,f[1])),h>2&&(m+=Zy(i,f[2])),h>3&&(m+=Zy(o,f[3])),h>4&&(m+=Zy(a,f[4])),h>5&&(m+=Zy(s,f[5])),h>6&&(m+=Zy(u,f[6])),h>7&&(m+=Zy(c,f[7])),h>8&&(m+=Zy(l,f[8])),h>9&&(m+=Zy(d,f[9]));var g=Bm(t,e.nodeIndex).renderText;t.renderer.setValue(g,m)}return p}(t,e,n,r,i,o,a,s,u,c,l,d);case 16384:return h=n,m=r,g=i,y=o,_=a,v=s,b=u,x=c,w=l,j=d,O=zm(p=t,(f=e).nodeIndex),C=O.instance,E=!1,S=void 0,(M=f.bindings.length)>0&&rg(p,f,0,h)&&(E=!0,S=Ry(p,O,f,0,h,S)),M>1&&rg(p,f,1,m)&&(E=!0,S=Ry(p,O,f,1,m,S)),M>2&&rg(p,f,2,g)&&(E=!0,S=Ry(p,O,f,2,g,S)),M>3&&rg(p,f,3,y)&&(E=!0,S=Ry(p,O,f,3,y,S)),M>4&&rg(p,f,4,_)&&(E=!0,S=Ry(p,O,f,4,_,S)),M>5&&rg(p,f,5,v)&&(E=!0,S=Ry(p,O,f,5,v,S)),M>6&&rg(p,f,6,b)&&(E=!0,S=Ry(p,O,f,6,b,S)),M>7&&rg(p,f,7,x)&&(E=!0,S=Ry(p,O,f,7,x,S)),M>8&&rg(p,f,8,w)&&(E=!0,S=Ry(p,O,f,8,w,S)),M>9&&rg(p,f,9,j)&&(E=!0,S=Ry(p,O,f,9,j,S)),S&&C.ngOnChanges(S),65536&f.flags&&Lm(p,256,f.nodeIndex)&&C.ngOnInit(),262144&f.flags&&C.ngDoCheck(),E;case 32:case 64:case 128:return function(t,e,n,r,i,o,a,s,u,c,l,d){var p=e.bindings,f=!1,h=p.length;if(h>0&&ig(t,e,0,n)&&(f=!0),h>1&&ig(t,e,1,r)&&(f=!0),h>2&&ig(t,e,2,i)&&(f=!0),h>3&&ig(t,e,3,o)&&(f=!0),h>4&&ig(t,e,4,a)&&(f=!0),h>5&&ig(t,e,5,s)&&(f=!0),h>6&&ig(t,e,6,u)&&(f=!0),h>7&&ig(t,e,7,c)&&(f=!0),h>8&&ig(t,e,8,l)&&(f=!0),h>9&&ig(t,e,9,d)&&(f=!0),f){var m=Hm(t,e.nodeIndex),g=void 0;switch(201347067&e.flags){case 32:g=new Array(p.length),h>0&&(g[0]=n),h>1&&(g[1]=r),h>2&&(g[2]=i),h>3&&(g[3]=o),h>4&&(g[4]=a),h>5&&(g[5]=s),h>6&&(g[6]=u),h>7&&(g[7]=c),h>8&&(g[8]=l),h>9&&(g[9]=d);break;case 64:g={},h>0&&(g[p[0].name]=n),h>1&&(g[p[1].name]=r),h>2&&(g[p[2].name]=i),h>3&&(g[p[3].name]=o),h>4&&(g[p[4].name]=a),h>5&&(g[p[5].name]=s),h>6&&(g[p[6].name]=u),h>7&&(g[p[7].name]=c),h>8&&(g[p[8].name]=l),h>9&&(g[p[9].name]=d);break;case 128:var y=n;switch(h){case 1:g=y.transform(n);break;case 2:g=y.transform(r);break;case 3:g=y.transform(r,i);break;case 4:g=y.transform(r,i,o);break;case 5:g=y.transform(r,i,o,a);break;case 6:g=y.transform(r,i,o,a,s);break;case 7:g=y.transform(r,i,o,a,s,u);break;case 8:g=y.transform(r,i,o,a,s,u,c);break;case 9:g=y.transform(r,i,o,a,s,u,c,l);break;case 10:g=y.transform(r,i,o,a,s,u,c,l,d)}}m.value=g}return f}(t,e,n,r,i,o,a,s,u,c,l,d);default:throw"unreachable"}var p,f,h,m,g,y,_,v,b,x,w,j,O,C,E,S,M;var T,k,A,D,I,P,R,N,F,L,B,V,z,H}(t,e,i,o,a,s,u,c,l,d,p,f):function(t,e,n){switch(201347067&e.flags){case 1:return function(t,e,n){for(var r=!1,i=0;i<n.length;i++)Lg(t,e,i,n[i])&&(r=!0);return r}(t,e,n);case 2:return function(t,e,n){for(var r=e.bindings,i=!1,o=0;o<n.length;o++)ig(t,e,o,n[o])&&(i=!0);if(i){var a="";for(o=0;o<n.length;o++)a+=Zy(n[o],r[o]);a=e.text.prefix+a;var s=Bm(t,e.nodeIndex).renderText;t.renderer.setValue(s,a)}return i}(t,e,n);case 16384:return function(t,e,n){for(var r=zm(t,e.nodeIndex),i=r.instance,o=!1,a=void 0,s=0;s<n.length;s++)rg(t,e,s,n[s])&&(o=!0,a=Ry(t,r,e,s,n[s],a));return a&&i.ngOnChanges(a),65536&e.flags&&Lm(t,256,e.nodeIndex)&&i.ngOnInit(),262144&e.flags&&i.ngDoCheck(),o}(t,e,n);case 32:case 64:case 128:return function(t,e,n){for(var i=e.bindings,o=!1,a=0;a<n.length;a++)ig(t,e,a,n[a])&&(o=!0);if(o){var s=Hm(t,e.nodeIndex),u=void 0;switch(201347067&e.flags){case 32:u=n;break;case 64:for(u={},a=0;a<n.length;a++)u[i[a].name]=n[a];break;case 128:var c=n[0],l=n.slice(1);u=c.transform.apply(c,Object(r.__spread)(l))}s.value=u}return o}(t,e,n);default:throw"unreachable"}}(t,e,i)}function l_(t){var e=t.def;if(4&e.nodeFlags)for(var n=0;n<e.nodes.length;n++){var r=e.nodes[n];if(4&r.flags){var i=Vm(t,n).template._projectedViews;if(i)for(var o=0;o<i.length;o++){var a=i[o];a.state|=32,sg(a,t)}}else 0==(4&r.childFlags)&&(n+=r.childCount)}}function d_(t,e,n,r,i,o,a,s,u,c,l,d,p){return 0===n?function(t,e,n,r,i,o,a,s,u,c,l,d){var p=e.bindings.length;p>0&&og(t,e,0,n);p>1&&og(t,e,1,r);p>2&&og(t,e,2,i);p>3&&og(t,e,3,o);p>4&&og(t,e,4,a);p>5&&og(t,e,5,s);p>6&&og(t,e,6,u);p>7&&og(t,e,7,c);p>8&&og(t,e,8,l);p>9&&og(t,e,9,d)}(t,e,r,i,o,a,s,u,c,l,d,p):function(t,e,n){for(var r=0;r<n.length;r++)og(t,e,r,n[r])}(t,e,r),!1}function p_(t,e){if(Um(t,e.nodeIndex).dirty)throw Gm(qm.createDebugContext(t,e.nodeIndex),"Query "+e.query.id+" not dirty","Query "+e.query.id+" dirty",0!=(1&t.state))}function f_(t){if(!(128&t.state)){if(m_(t,ky.Destroy),h_(t,ky.Destroy),Ny(t,131072),t.disposables)for(var e=0;e<t.disposables.length;e++)t.disposables[e]();!function(t){if(16&t.state){var e=cg(t);if(e){var n=e.template._projectedViews;n&&($g(n,n.indexOf(t)),qm.dirtyParentQueries(t))}}}(t),t.renderer.destroyNode&&function(t){for(var e=t.def.nodes.length,n=0;n<e;n++){var r=t.def.nodes[n];1&r.flags?t.renderer.destroyNode(Vm(t,n).renderElement):2&r.flags?t.renderer.destroyNode(Bm(t,n).renderText):(67108864&r.flags||134217728&r.flags)&&Um(t,n).destroy()}}(t),fg(t)&&t.renderer.destroy(),t.state|=128}}function h_(t,e){var n=t.def;if(33554432&n.nodeFlags)for(var r=0;r<n.nodes.length;r++){var i=n.nodes[r];33554432&i.flags?g_(Vm(t,r).componentView,e):0==(33554432&i.childFlags)&&(r+=i.childCount)}}function m_(t,e){var n=t.def;if(16777216&n.nodeFlags)for(var r=0;r<n.nodes.length;r++){var i=n.nodes[r];if(16777216&i.flags)for(var o=Vm(t,r).viewContainer._embeddedViews,a=0;a<o.length;a++)g_(o[a],e);else 0==(16777216&i.childFlags)&&(r+=i.childCount)}}function g_(t,e){var n=t.state;switch(e){case ky.CheckNoChanges:0==(128&n)&&(12==(12&n)?s_(t):64&n&&y_(t,ky.CheckNoChangesProjectedViews));break;case ky.CheckNoChangesProjectedViews:0==(128&n)&&(32&n?s_(t):64&n&&y_(t,e));break;case ky.CheckAndUpdate:0==(128&n)&&(12==(12&n)?u_(t):64&n&&y_(t,ky.CheckAndUpdateProjectedViews));break;case ky.CheckAndUpdateProjectedViews:0==(128&n)&&(32&n?u_(t):64&n&&y_(t,e));break;case ky.Destroy:f_(t);break;case ky.CreateViewNodes:a_(t)}}function y_(t,e){m_(t,e),h_(t,e)}function __(t,e,n,r){if(t.def.nodeFlags&e&&t.def.nodeFlags&n)for(var i=t.def.nodes.length,o=0;o<i;o++){var a=t.def.nodes[o];if(a.flags&e&&a.flags&n)switch(qm.setCurrentNode(t,a.nodeIndex),r){case 0:zy(t,a);break;case 1:p_(t,a)}a.childFlags&e&&a.childFlags&n||(o+=a.childCount)}}(Ay=ky||(ky={}))[Ay.CreateViewNodes=0]="CreateViewNodes",Ay[Ay.CheckNoChanges=1]="CheckNoChanges",Ay[Ay.CheckNoChangesProjectedViews=2]="CheckNoChangesProjectedViews",Ay[Ay.CheckAndUpdate=3]="CheckAndUpdate",Ay[Ay.CheckAndUpdateProjectedViews=4]="CheckAndUpdateProjectedViews",Ay[Ay.Destroy=5]="Destroy";var v_=!1;function b_(){if(!v_){v_=!0;var t=fc()?{setCurrentNode:W_,createRootView:w_,createEmbeddedView:O_,createComponentView:C_,createNgModuleRef:E_,overrideProvider:R_,overrideComponentView:N_,clearOverrides:F_,checkAndUpdateView:z_,checkNoChangesView:H_,destroyView:U_,createDebugContext:function(t,e){return new $_(t,e)},handleEvent:q_,updateDirectives:G_,updateRenderer:Y_}:{setCurrentNode:function(){},createRootView:x_,createEmbeddedView:e_,createComponentView:r_,createNgModuleRef:dy,overrideProvider:Xm,overrideComponentView:Xm,clearOverrides:Xm,checkAndUpdateView:u_,checkNoChangesView:s_,destroyView:f_,createDebugContext:function(t,e){return new $_(t,e)},handleEvent:function(t,e,n,r){return t.def.handleEvent(t,e,n,r)},updateDirectives:function(t,e){return t.def.updateDirectives(0===e?B_:V_,t)},updateRenderer:function(t,e){return t.def.updateRenderer(0===e?B_:V_,t)}};qm.setCurrentNode=t.setCurrentNode,qm.createRootView=t.createRootView,qm.createEmbeddedView=t.createEmbeddedView,qm.createComponentView=t.createComponentView,qm.createNgModuleRef=t.createNgModuleRef,qm.overrideProvider=t.overrideProvider,qm.overrideComponentView=t.overrideComponentView,qm.clearOverrides=t.clearOverrides,qm.checkAndUpdateView=t.checkAndUpdateView,qm.checkNoChangesView=t.checkNoChangesView,qm.destroyView=t.destroyView,qm.resolveDep=Iy,qm.createDebugContext=t.createDebugContext,qm.handleEvent=t.handleEvent,qm.updateDirectives=t.updateDirectives,qm.updateRenderer=t.updateRenderer,qm.dirtyParentQueries=Vy}}function x_(t,e,n,r,i,o){return n_(j_(t,i,i.injector.get(Xu),e,n),r,o)}function w_(t,e,n,r,i,o){var a=i.injector.get(Xu),s=j_(t,i,new nv(a),e,n),u=L_(r);return tv(S_.create,n_,null,[s,u,o])}function j_(t,e,n,r,i){var o=e.injector.get(ec),a=e.injector.get(of);return{ngModule:e,injector:t,projectableNodes:r,selectorOrNode:i,sanitizer:o,rendererFactory:n,renderer:n.createRenderer(null,null),errorHandler:a}}function O_(t,e,n,r){var i=L_(n);return tv(S_.create,e_,null,[t,e,i,r])}function C_(t,e,n,r){var i=P_.get(e.element.componentProvider.provider.token);return n=i||L_(n),tv(S_.create,r_,null,[t,e,n,r])}function E_(t,e,n,r){return dy(t,e,n,function(t){var e=function(t){var e=!1,n=!1;if(0===D_.size)return{hasOverrides:e,hasDeprecatedOverrides:n};return t.providers.forEach(function(t){var r=D_.get(t.token);3840&t.flags&&r&&(e=!0,n=n||r.deprecatedBehavior)}),t.modules.forEach(function(t){I_.forEach(function(r,i){_(i).providedIn===t&&(e=!0,n=n||r.deprecatedBehavior)})}),{hasOverrides:e,hasDeprecatedOverrides:n}}(t),n=e.hasOverrides,r=e.hasDeprecatedOverrides;if(!n)return t;return function(t){for(var e=0;e<t.providers.length;e++){var n=t.providers[e];r&&(n.flags|=4096);var i=D_.get(n.token);i&&(n.flags=-3841&n.flags|i.flags,n.deps=gg(i.deps),n.value=i.value)}if(I_.size>0){var o=new Set(t.modules);I_.forEach(function(e,n){if(o.has(_(n).providedIn)){var i={token:n,flags:e.flags|(r?4096:0),deps:gg(e.deps),value:e.value,index:t.providers.length};t.providers.push(i),t.providersByKey[Zm(n)]=i}})}}(t=t.factory(function(){return Xm})),t}(r))}var S_,M_,T_,k_,A_,D_=new Map,I_=new Map,P_=new Map;function R_(t){var e;D_.set(t.token,t),"function"==typeof t.token&&(e=_(t.token))&&"function"==typeof e.providedIn&&I_.set(t.token,t)}function N_(t,e){var n=vg(vg(ey(e)).nodes[0].element.componentView);P_.set(t,n)}function F_(){D_.clear(),I_.clear(),P_.clear()}function L_(t){if(0===D_.size)return t;var e=function(t){for(var e=[],n=null,r=0;r<t.nodes.length;r++){var i=t.nodes[r];1&i.flags&&(n=i),n&&3840&i.flags&&D_.has(i.provider.token)&&(e.push(n.nodeIndex),n=null)}return e}(t);if(0===e.length)return t;t=t.factory(function(){return Xm});for(var n=0;n<e.length;n++)r(t,e[n]);return t;function r(t,e){for(var n=e+1;n<t.nodes.length;n++){var r=t.nodes[n];if(1&r.flags)return;if(3840&r.flags){var i=r.provider,o=D_.get(i.token);o&&(r.flags=-3841&r.flags|o.flags,i.deps=gg(o.deps),i.value=o.value)}}}}function B_(t,e,n,r,i,o,a,s,u,c,l,d,p){var f=t.def.nodes[e];return c_(t,f,n,r,i,o,a,s,u,c,l,d,p),224&f.flags?Hm(t,e).value:void 0}function V_(t,e,n,r,i,o,a,s,u,c,l,d,p){var f=t.def.nodes[e];return d_(t,f,n,r,i,o,a,s,u,c,l,d,p),224&f.flags?Hm(t,e).value:void 0}function z_(t){return tv(S_.detectChanges,u_,null,[t])}function H_(t){return tv(S_.checkNoChanges,s_,null,[t])}function U_(t){return tv(S_.destroy,f_,null,[t])}function W_(t,e){k_=t,A_=e}function q_(t,e,n,r){return W_(t,e),tv(S_.handleEvent,t.def.handleEvent,null,[t,e,n,r])}function G_(t,e){if(128&t.state)throw Km(S_[T_]);return W_(t,Q_(t,0)),t.def.updateDirectives(function(t,n,r){for(var i=[],o=3;o<arguments.length;o++)i[o-3]=arguments[o];var a=t.def.nodes[n];0===e?K_(t,a,r,i):X_(t,a,r,i);16384&a.flags&&W_(t,Q_(t,n));return 224&a.flags?Hm(t,a.nodeIndex).value:void 0},t)}function Y_(t,e){if(128&t.state)throw Km(S_[T_]);return W_(t,Z_(t,0)),t.def.updateRenderer(function(t,n,r){for(var i=[],o=3;o<arguments.length;o++)i[o-3]=arguments[o];var a=t.def.nodes[n];0===e?K_(t,a,r,i):X_(t,a,r,i);3&a.flags&&W_(t,Z_(t,n));return 224&a.flags?Hm(t,a.nodeIndex).value:void 0},t)}function K_(t,e,n,i){if(c_.apply(void 0,Object(r.__spread)([t,e,n],i))){var o=1===n?i[0]:i;if(16384&e.flags){for(var a={},s=0;s<e.bindings.length;s++){var u=e.bindings[s],c=o[s];8&u.flags&&(a[Cr(u.nonMinifiedName)]=Sr(c))}var l=e.parent,d=Vm(t,l.nodeIndex).renderElement;if(l.element.name)for(var p in a){null!=(c=a[p])?t.renderer.setAttribute(d,p,c):t.renderer.removeAttribute(d,p)}else t.renderer.setValue(d,"bindings="+JSON.stringify(a,null,2))}}}function X_(t,e,n,i){d_.apply(void 0,Object(r.__spread)([t,e,n],i))}function Q_(t,e){for(var n=e;n<t.def.nodes.length;n++){var r=t.def.nodes[n];if(16384&r.flags&&r.bindings&&r.bindings.length)return n}return null}function Z_(t,e){for(var n=e;n<t.def.nodes.length;n++){var r=t.def.nodes[n];if(3&r.flags&&r.bindings&&r.bindings.length)return n}return null}(M_=S_||(S_={}))[M_.create=0]="create",M_[M_.detectChanges=1]="detectChanges",M_[M_.checkNoChanges=2]="checkNoChanges",M_[M_.destroy=3]="destroy",M_[M_.handleEvent=4]="handleEvent";var $_=function(){function t(t,e){this.view=t,this.nodeIndex=e,null==e&&(this.nodeIndex=e=0),this.nodeDef=t.def.nodes[e];for(var n=this.nodeDef,r=t;n&&0==(1&n.flags);)n=n.parent;if(!n)for(;!n&&r;)n=lg(r),r=r.parent;this.elDef=n,this.elView=r}return Object.defineProperty(t.prototype,"elOrCompView",{get:function(){return Vm(this.elView,this.elDef.nodeIndex).componentView||this.view},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"injector",{get:function(){return sy(this.elView,this.elDef)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"component",{get:function(){return this.elOrCompView.component},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"context",{get:function(){return this.elOrCompView.context},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"providerTokens",{get:function(){var t=[];if(this.elDef)for(var e=this.elDef.nodeIndex+1;e<=this.elDef.nodeIndex+this.elDef.childCount;e++){var n=this.elView.def.nodes[e];20224&n.flags&&t.push(n.provider.token),e+=n.childCount}return t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"references",{get:function(){var t={};if(this.elDef){J_(this.elView,this.elDef,t);for(var e=this.elDef.nodeIndex+1;e<=this.elDef.nodeIndex+this.elDef.childCount;e++){var n=this.elView.def.nodes[e];20224&n.flags&&J_(this.elView,n,t),e+=n.childCount}}return t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"componentRenderElement",{get:function(){var t=function(t){for(;t&&!fg(t);)t=t.parent;if(t.parent)return Vm(t.parent,lg(t).nodeIndex);return null}(this.elOrCompView);return t?t.renderElement:void 0},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"renderNode",{get:function(){return 2&this.nodeDef.flags?dg(this.view,this.nodeDef):dg(this.elView,this.elDef)},enumerable:!0,configurable:!0}),t.prototype.logError=function(t){for(var e,n,i=[],o=1;o<arguments.length;o++)i[o-1]=arguments[o];2&this.nodeDef.flags?(e=this.view.def,n=this.nodeDef.nodeIndex):(e=this.elView.def,n=this.elDef.nodeIndex);var a=function(t,e){for(var n=-1,r=0;r<=e;r++){var i=t.nodes[r];3&i.flags&&n++}return n}(e,n),s=-1;e.factory(function(){var e;return++s===a?(e=t.error).bind.apply(e,Object(r.__spread)([t],i)):Xm}),s<a&&(t.error("Illegal state: the ViewDefinitionFactory did not call the logger!"),t.error.apply(t,Object(r.__spread)(i)))},t}();function J_(t,e,n){for(var r in e.references)n[r]=Uy(t,e,e.references[r])}function tv(t,e,n,r){var i,o,a=T_,s=k_,u=A_;try{T_=t;var c=e.apply(n,r);return k_=s,A_=u,T_=a,c}catch(t){if(ef(t)||!k_)throw t;throw i=t,o=ev(),i instanceof Error||(i=new Error(i.toString())),Ym(i,o),i}}function ev(){return k_?new $_(k_,A_):null}var nv=function(){function t(t){this.delegate=t}return t.prototype.createRenderer=function(t,e){return new rv(this.delegate.createRenderer(t,e))},t.prototype.begin=function(){this.delegate.begin&&this.delegate.begin()},t.prototype.end=function(){this.delegate.end&&this.delegate.end()},t.prototype.whenRenderingDone=function(){return this.delegate.whenRenderingDone?this.delegate.whenRenderingDone():Promise.resolve(null)},t}(),rv=function(){function t(t){this.delegate=t,this.debugContextFactory=ev,this.data=this.delegate.data}return t.prototype.createDebugContext=function(t){return this.debugContextFactory(t)},t.prototype.destroyNode=function(t){var e;e=im(t),nm.delete(e.nativeNode),this.delegate.destroyNode&&this.delegate.destroyNode(t)},t.prototype.destroy=function(){this.delegate.destroy()},t.prototype.createElement=function(t,e){var n=this.delegate.createElement(t,e),r=this.createDebugContext(n);if(r){var i=new Zh(n,null,r);i.name=t,om(i)}return n},t.prototype.createComment=function(t){var e=this.delegate.createComment(t),n=this.createDebugContext(e);return n&&om(new Qh(e,null,n)),e},t.prototype.createText=function(t){var e=this.delegate.createText(t),n=this.createDebugContext(e);return n&&om(new Qh(e,null,n)),e},t.prototype.appendChild=function(t,e){var n=im(t),r=im(e);n&&r&&n instanceof Zh&&n.addChild(r),this.delegate.appendChild(t,e)},t.prototype.insertBefore=function(t,e,n){var r=im(t),i=im(e),o=im(n);r&&i&&r instanceof Zh&&r.insertBefore(o,i),this.delegate.insertBefore(t,e,n)},t.prototype.removeChild=function(t,e){var n=im(t),r=im(e);n&&r&&n instanceof Zh&&n.removeChild(r),this.delegate.removeChild(t,e)},t.prototype.selectRootElement=function(t,e){var n=this.delegate.selectRootElement(t,e),r=ev();return r&&om(new Zh(n,null,r)),n},t.prototype.setAttribute=function(t,e,n,r){var i=im(t);if(i&&i instanceof Zh){var o=r?r+":"+e:e;i.attributes[o]=n}this.delegate.setAttribute(t,e,n,r)},t.prototype.removeAttribute=function(t,e,n){var r=im(t);if(r&&r instanceof Zh){var i=n?n+":"+e:e;r.attributes[i]=null}this.delegate.removeAttribute(t,e,n)},t.prototype.addClass=function(t,e){var n=im(t);n&&n instanceof Zh&&(n.classes[e]=!0),this.delegate.addClass(t,e)},t.prototype.removeClass=function(t,e){var n=im(t);n&&n instanceof Zh&&(n.classes[e]=!1),this.delegate.removeClass(t,e)},t.prototype.setStyle=function(t,e,n,r){var i=im(t);i&&i instanceof Zh&&(i.styles[e]=n),this.delegate.setStyle(t,e,n,r)},t.prototype.removeStyle=function(t,e,n){var r=im(t);r&&r instanceof Zh&&(r.styles[e]=null),this.delegate.removeStyle(t,e,n)},t.prototype.setProperty=function(t,e,n){var r=im(t);r&&r instanceof Zh&&(r.properties[e]=n),this.delegate.setProperty(t,e,n)},t.prototype.listen=function(t,e,n){if("string"!=typeof t){var r=im(t);r&&r.listeners.push(new Xh(e,n))}return this.delegate.listen(t,e,n)},t.prototype.parentNode=function(t){return this.delegate.parentNode(t)},t.prototype.nextSibling=function(t){return this.delegate.nextSibling(t)},t.prototype.setValue=function(t,e){return this.delegate.setValue(t,e)},t}();function iv(t){return b_(),qm.overrideProvider(t)}function ov(t,e){return b_(),qm.overrideComponentView(t,e)}function av(){return b_(),qm.clearOverrides()}function sv(t,e,n){return new uv(t,e,n)}var uv=function(t){function e(e,n,r){var i=t.call(this)||this;return i.moduleType=e,i._bootstrapComponents=n,i._ngModuleDefFactory=r,i}return Object(r.__extends)(e,t),e.prototype.create=function(t){b_();var e=function(t){var e=Array.from(t.providers),n=Array.from(t.modules),r={};for(var i in t.providersByKey)r[i]=t.providersByKey[i];return{factory:t.factory,isRoot:t.isRoot,providers:e,modules:n,providersByKey:r}}(vg(this._ngModuleDefFactory));return qm.createNgModuleRef(this.moduleType,t||Ls.NULL,this._bootstrapComponents,e)},e}(Pu)},"./node_modules/@angular/core/node_modules/tslib/tslib.es6.js":function(t,e,n){"use strict";n.r(e),n.d(e,"__extends",function(){return i}),n.d(e,"__assign",function(){return o}),n.d(e,"__rest",function(){return a}),n.d(e,"__decorate",function(){return s}),n.d(e,"__param",function(){return u}),n.d(e,"__metadata",function(){return c}),n.d(e,"__awaiter",function(){return l}),n.d(e,"__generator",function(){return d}),n.d(e,"__exportStar",function(){return p}),n.d(e,"__values",function(){return f}),n.d(e,"__read",function(){return h}),n.d(e,"__spread",function(){return m}),n.d(e,"__await",function(){return g}),n.d(e,"__asyncGenerator",function(){return y}),n.d(e,"__asyncDelegator",function(){return _}),n.d(e,"__asyncValues",function(){return v}),n.d(e,"__makeTemplateObject",function(){return b}),n.d(e,"__importStar",function(){return x}),n.d(e,"__importDefault",function(){return w});var r=function(t,e){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)};function i(t,e){function n(){this.constructor=t}r(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}var o=function(){return(o=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++){e=arguments[n];for(var i in e)Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i])}return t}).apply(this,arguments)};function a(t,e){var n={};for(var r in t)Object.prototype.hasOwnProperty.call(t,r)&&e.indexOf(r)<0&&(n[r]=t[r]);if(null!=t&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(t);i<r.length;i++)e.indexOf(r[i])<0&&(n[r[i]]=t[r[i]])}return n}function s(t,e,n,r){var i,o=arguments.length,a=o<3?e:null===r?r=Object.getOwnPropertyDescriptor(e,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(t,e,n,r);else for(var s=t.length-1;s>=0;s--)(i=t[s])&&(a=(o<3?i(a):o>3?i(e,n,a):i(e,n))||a);return o>3&&a&&Object.defineProperty(e,n,a),a}function u(t,e){return function(n,r){e(n,r,t)}}function c(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)}function l(t,e,n,r){return new(n||(n=Promise))(function(i,o){function a(t){try{u(r.next(t))}catch(t){o(t)}}function s(t){try{u(r.throw(t))}catch(t){o(t)}}function u(t){t.done?i(t.value):new n(function(e){e(t.value)}).then(a,s)}u((r=r.apply(t,e||[])).next())})}function d(t,e){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=e.call(t,a)}catch(t){o=[6,t],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(t,e){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}function f(t){var e="function"==typeof Symbol&&t[Symbol.iterator],n=0;return e?e.call(t):{next:function(){return t&&n>=t.length&&(t=void 0),{value:t&&t[n++],done:!t}}}}function h(t,e){var n="function"==typeof Symbol&&t[Symbol.iterator];if(!n)return t;var r,i,o=n.call(t),a=[];try{for(;(void 0===e||e-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(t){i={error:t}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var t=[],e=0;e<arguments.length;e++)t=t.concat(h(arguments[e]));return t}function g(t){return this instanceof g?(this.v=t,this):new g(t)}function y(t,e,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(t,e||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(t){i[t]&&(r[t]=function(e){return new Promise(function(n,r){o.push([t,e,n,r])>1||s(t,e)})})}function s(t,e){try{(n=i[t](e)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(t){l(o[0][3],t)}var n}function u(t){s("next",t)}function c(t){s("throw",t)}function l(t,e){t(e),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(t){var e,n;return e={},r("next"),r("throw",function(t){throw t}),r("return"),e[Symbol.iterator]=function(){return this},e;function r(r,i){e[r]=t[r]?function(e){return(n=!n)?{value:g(t[r](e)),done:"return"===r}:i?i(e):e}:i}}function v(t){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var e,n=t[Symbol.asyncIterator];return n?n.call(t):(t=f(t),e={},r("next"),r("throw"),r("return"),e[Symbol.asyncIterator]=function(){return this},e);function r(n){e[n]=t[n]&&function(e){return new Promise(function(r,i){var o,a,s,u;e=t[n](e),o=r,a=i,s=e.done,u=e.value,Promise.resolve(u).then(function(t){o({value:t,done:s})},a)})}}}function b(t,e){return Object.defineProperty?Object.defineProperty(t,"raw",{value:e}):t.raw=e,t}function x(t){if(t&&t.__esModule)return t;var e={};if(null!=t)for(var n in t)Object.hasOwnProperty.call(t,n)&&(e[n]=t[n]);return e.default=t,e}function w(t){return t&&t.__esModule?t:{default:t}}},"./node_modules/@angular/flex-layout/esm5/core.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"removeStyles",function(){return u}),n.d(e,"BROWSER_PROVIDER",function(){return c}),n.d(e,"CLASS_NAME",function(){return l}),n.d(e,"CoreModule",function(){return d}),n.d(e,"MediaChange",function(){return p}),n.d(e,"StylesheetMap",function(){return f}),n.d(e,"DEFAULT_CONFIG",function(){return h}),n.d(e,"LAYOUT_CONFIG",function(){return m}),n.d(e,"SERVER_TOKEN",function(){return g}),n.d(e,"BREAKPOINT",function(){return y}),n.d(e,"BaseDirective2",function(){return x}),n.d(e,"sortDescendingPriority",function(){return R}),n.d(e,"sortAscendingPriority",function(){return N}),n.d(e,"DEFAULT_BREAKPOINTS",function(){return w}),n.d(e,"ScreenTypes",function(){return T}),n.d(e,"ORIENTATION_BREAKPOINTS",function(){return k}),n.d(e,"BreakPointRegistry",function(){return L}),n.d(e,"BREAKPOINTS",function(){return F}),n.d(e,"MatchMedia",function(){return B}),n.d(e,"MockMatchMedia",function(){return z}),n.d(e,"MockMediaQueryList",function(){return H}),n.d(e,"MockMatchMediaProvider",function(){return U}),n.d(e,"ServerMediaQueryList",function(){return W}),n.d(e,"ServerMatchMedia",function(){return q}),n.d(e,"MediaObserver",function(){return Z}),n.d(e,"StyleUtils",function(){return J}),n.d(e,"StyleBuilder",function(){return tt}),n.d(e,"validateBasis",function(){return et}),n.d(e,"MediaMarshaller",function(){return rt}),n.d(e,"BREAKPOINT_PRINT",function(){return K}),n.d(e,"PrintHook",function(){return X});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/common/fesm5/common.js"),o=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/rxjs/_esm5/index.js"),s=n("./node_modules/rxjs/_esm5/operators/index.js");function u(t,e){return function(){if(Object(i.isPlatformBrowser)(e)){var n=/\bflex-layout-.+?\b/g;Array.from(t.querySelectorAll("[class*="+l+"]")).forEach(function(t){t.classList.contains(l+"ssr")&&t.parentNode?t.parentNode.removeChild(t):t.className.replace(n,"")})}}}var c={provide:r.APP_BOOTSTRAP_LISTENER,useFactory:u,deps:[i.DOCUMENT,r.PLATFORM_ID],multi:!0},l="flex-layout-",d=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{providers:[c]}]}],t}(),p=function(){function t(t,e,n,r){void 0===t&&(t=!1),void 0===e&&(e="all"),void 0===n&&(n=""),void 0===r&&(r=""),this.matches=t,this.mediaQuery=e,this.mqAlias=n,this.suffix=r,this.property=""}return t.prototype.clone=function(){return new t(this.matches,this.mediaQuery,this.mqAlias,this.suffix)},t}(),f=function(){function t(){this.stylesheet=new Map}return t.prototype.addStyleToElement=function(t,e,n){var r=this.stylesheet.get(t);r?r.set(e,n):this.stylesheet.set(t,new Map([[e,n]]))},t.prototype.clearStyles=function(){this.stylesheet.clear()},t.prototype.getStyleForElement=function(t,e){var n=this.stylesheet.get(t),r="";if(n){var i=n.get(e);"number"!=typeof i&&"string"!=typeof i||(r=i+"")}return r},t.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(),h={addFlexToParent:!0,addOrientationBps:!1,disableDefaultBps:!1,disableVendorPrefixes:!1,serverLoaded:!1,useColumnBasisZero:!0,printWithBreakpoints:[]},m=new r.InjectionToken("Flex Layout token, config options for the library",{providedIn:"root",factory:function(){return h}}),g=new r.InjectionToken("FlexLayoutServerLoaded",{providedIn:"root",factory:function(){return!1}}),y=new r.InjectionToken("Flex Layout token, collect all breakpoints into one provider",{providedIn:"root",factory:function(){return null}}),_="inline",v=["row","column","row-reverse","column-reverse"];function b(t){var e=function(t){var e=(t=t?t.toLowerCase():"").split(" "),n=e[0],r=e[1],i=e[2];v.find(function(t){return t===n})||(n=v[0]);r===_&&(r=i!==_?i:"",i=_);return[n,function(t){if(t)switch(t.toLowerCase()){case"reverse":case"wrap-reverse":case"reverse-wrap":t="wrap-reverse";break;case"no":case"none":case"nowrap":t="nowrap";break;default:t="wrap"}return t}(r),!!i]}(t);return function(t,e,n){void 0===e&&(e=null);void 0===n&&(n=!1);return{display:n?"inline-flex":"flex","box-sizing":"border-box","flex-direction":t,"flex-wrap":e||null}}(e[0],e[1],e[2])}var x=function(){function t(t,e,n,r){this.elementRef=t,this.styleBuilder=e,this.styler=n,this.marshal=r,this.DIRECTIVE_KEY="",this.inputs=[],this.mru={},this.destroySubject=new a.Subject,this.styleCache=new Map}return Object.defineProperty(t.prototype,"parentElement",{get:function(){return this.elementRef.nativeElement.parentElement},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"nativeElement",{get:function(){return this.elementRef.nativeElement},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"activatedValue",{get:function(){return this.marshal.getValue(this.nativeElement,this.DIRECTIVE_KEY)},set:function(t){this.marshal.setValue(this.nativeElement,this.DIRECTIVE_KEY,t,this.marshal.activatedAlias)},enumerable:!0,configurable:!0}),t.prototype.ngOnChanges=function(t){var e=this;Object.keys(t).forEach(function(n){if(-1!==e.inputs.indexOf(n)){var r=n.split(".").slice(1).join("."),i=t[n].currentValue;e.setValue(i,r)}})},t.prototype.ngOnDestroy=function(){this.destroySubject.next(),this.destroySubject.complete(),this.marshal.releaseElement(this.nativeElement)},t.prototype.init=function(t){void 0===t&&(t=[]),this.marshal.init(this.elementRef.nativeElement,this.DIRECTIVE_KEY,this.updateWithValue.bind(this),this.clearStyles.bind(this),t)},t.prototype.addStyles=function(t,e){var n=this.styleBuilder,r=n.shouldCache,i=this.styleCache.get(t);i&&r||(i=n.buildStyles(t,e),r&&this.styleCache.set(t,i)),this.mru=Object(o.__assign)({},i),this.applyStyleToElement(i),n.sideEffect(t,i,e)},t.prototype.clearStyles=function(){var t=this;Object.keys(this.mru).forEach(function(e){t.mru[e]=""}),this.applyStyleToElement(this.mru),this.mru={}},t.prototype.triggerUpdate=function(){var t=this.marshal.getValue(this.nativeElement,this.DIRECTIVE_KEY);void 0!==t&&this.marshal.updateElement(this.nativeElement,this.DIRECTIVE_KEY,t)},t.prototype.getFlexFlowDirection=function(t,e){if(void 0===e&&(e=!1),t){var n=this.styler.getFlowDirection(t),r=n[0];if(!n[1]&&e){var i=b(r),o=[t];this.styler.applyStyleToElements(i,o)}return r.trim()}return"row"},t.prototype.applyStyleToElement=function(t,e,n){void 0===n&&(n=this.nativeElement),this.styler.applyStyleToElement(n,t,e)},t.prototype.setValue=function(t,e){this.marshal.setValue(this.nativeElement,this.DIRECTIVE_KEY,t,e)},t.prototype.updateWithValue=function(t){this.addStyles(t)},t}(),w=[{alias:"xs",mediaQuery:"screen and (min-width: 0px) and (max-width: 599px)",priority:1e3},{alias:"sm",mediaQuery:"screen and (min-width: 600px) and (max-width: 959px)",priority:900},{alias:"md",mediaQuery:"screen and (min-width: 960px) and (max-width: 1279px)",priority:800},{alias:"lg",mediaQuery:"screen and (min-width: 1280px) and (max-width: 1919px)",priority:700},{alias:"xl",mediaQuery:"screen and (min-width: 1920px) and (max-width: 5000px)",priority:600},{alias:"lt-sm",overlapping:!0,mediaQuery:"screen and (max-width: 599px)",priority:950},{alias:"lt-md",overlapping:!0,mediaQuery:"screen and (max-width: 959px)",priority:850},{alias:"lt-lg",overlapping:!0,mediaQuery:"screen and (max-width: 1279px)",priority:750},{alias:"lt-xl",overlapping:!0,priority:650,mediaQuery:"screen and (max-width: 1919px)"},{alias:"gt-xs",overlapping:!0,mediaQuery:"screen and (min-width: 600px)",priority:-950},{alias:"gt-sm",overlapping:!0,mediaQuery:"screen and (min-width: 960px)",priority:-850},{alias:"gt-md",overlapping:!0,mediaQuery:"screen and (min-width: 1280px)",priority:-750},{alias:"gt-lg",overlapping:!0,mediaQuery:"screen and (min-width: 1920px)",priority:-650}],j="(orientation: portrait) and (max-width: 599px)",O="(orientation: landscape) and (max-width: 959px)",C="(orientation: landscape) and (min-width: 960px) and (max-width: 1279px)",E="(orientation: portrait) and (min-width: 600px) and (max-width: 839px)",S="(orientation: portrait) and (min-width: 840px)",M="(orientation: landscape) and (min-width: 1280px)",T={HANDSET:j+", "+O,TABLET:E+" , "+C,WEB:S+", "+M+" ",HANDSET_PORTRAIT:""+j,TABLET_PORTRAIT:E+" ",WEB_PORTRAIT:""+S,HANDSET_LANDSCAPE:O+"]",TABLET_LANDSCAPE:""+C,WEB_LANDSCAPE:""+M},k=[{alias:"handset",priority:2e3,mediaQuery:T.HANDSET},{alias:"handset.landscape",priority:2e3,mediaQuery:T.HANDSET_LANDSCAPE},{alias:"handset.portrait",priority:2e3,mediaQuery:T.HANDSET_PORTRAIT},{alias:"tablet",priority:2100,mediaQuery:T.TABLET},{alias:"tablet.landscape",priority:2100,mediaQuery:T.TABLET},{alias:"tablet.portrait",priority:2100,mediaQuery:T.TABLET_PORTRAIT},{alias:"web",priority:2200,mediaQuery:T.WEB,overlapping:!0},{alias:"web.landscape",priority:2200,mediaQuery:T.WEB_LANDSCAPE,overlapping:!0},{alias:"web.portrait",priority:2200,mediaQuery:T.WEB_PORTRAIT,overlapping:!0}];function A(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];if(null==t)throw TypeError("Cannot convert undefined or null to object");for(var r=0,i=e;r<i.length;r++){var o=i[r];if(null!=o)for(var a in o)o.hasOwnProperty(a)&&(t[a]=o[a])}return t}var D=/(\.|-|_)/g;function I(t){var e=t.length>0?t.charAt(0):"",n=t.length>1?t.slice(1):"";return e.toUpperCase()+n}function P(t){return t.forEach(function(t){t.suffix||(t.suffix=t.alias.replace(D,"|").split("|").map(I).join(""),t.overlapping=!!t.overlapping)}),t}function R(t,e){var n=t?t.priority||0:0;return(e?e.priority||0:0)-n}function N(t,e){return(t.priority||0)-(e.priority||0)}var F=new r.InjectionToken("Token (@angular/flex-layout) Breakpoints",{providedIn:"root",factory:function(){var t=Object(r.inject)(y),e=Object(r.inject)(m),n=[].concat.apply([],(t||[]).map(function(t){return Array.isArray(t)?t:[t]}));return function(t,e){void 0===e&&(e=[]);var n={};return t.forEach(function(t){n[t.alias]=t}),e.forEach(function(t){n[t.alias]?A(n[t.alias],t):n[t.alias]=t}),P(Object.keys(n).map(function(t){return n[t]}))}((e.disableDefaultBps?[]:w).concat(e.addOrientationBps?k:[]),n)}}),L=function(){function t(t){this.findByMap=new Map,this.items=t.slice().sort(N)}return t.prototype.findByAlias=function(t){return t?this.findWithPredicate(t,function(e){return e.alias==t}):null},t.prototype.findByQuery=function(t){return this.findWithPredicate(t,function(e){return e.mediaQuery==t})},Object.defineProperty(t.prototype,"overlappings",{get:function(){return this.items.filter(function(t){return 1==t.overlapping})},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"aliases",{get:function(){return this.items.map(function(t){return t.alias})},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"suffixes",{get:function(){return this.items.map(function(t){return t.suffix?t.suffix:""})},enumerable:!0,configurable:!0}),t.prototype.findWithPredicate=function(t,e){var n=this.findByMap.get(t);return n||(n=this.items.find(e)||null,this.findByMap.set(t,n)),n||null},t.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:Array,decorators:[{type:r.Inject,args:[F]}]}]},t.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new t(Object(r.inject)(F))},token:t,providedIn:"root"}),t}(),B=function(){function t(t,e,n){this._zone=t,this._platformId=e,this._document=n,this._source=new a.BehaviorSubject(new p(!0)),this._registry=new Map,this._observable$=this._source.asObservable()}return t.prototype.isActive=function(t){var e=this._registry.get(t);return!!e&&e.matches},t.prototype.observe=function(t,e){var n=this;if(void 0===e&&(e=!1),t){var r=this._observable$.pipe(Object(s.filter)(function(n){return!e||t.indexOf(n.mediaQuery)>-1})),i=new a.Observable(function(e){var r=n.registerQuery(t);if(r.length){var i=r.pop();r.forEach(function(t){e.next(t)}),n._source.next(i)}e.complete()});return Object(a.merge)(i,r)}return this._observable$},t.prototype.registerQuery=function(t){var e=this,n=Array.isArray(t)?t:[t],r=[];return function(t,e){var n=t.filter(function(t){return!V[t]});if(n.length>0){var r=n.join(", ");try{var i=e.createElement("style");if(i.setAttribute("type","text/css"),!i.styleSheet){var o="\n/*\n  @angular/flex-layout - workaround for possible browser quirk with mediaQuery listeners\n  see http://bit.ly/2sd4HMP\n*/\n@media "+r+" {.fx-query-test{ }}\n";i.appendChild(e.createTextNode(o))}e.head.appendChild(i),n.forEach(function(t){return V[t]=i})}catch(t){console.error(t)}}}(n,this._document),n.forEach(function(t){var n=e._registry.get(t);n||((n=e.buildMQL(t)).addListener(function(n){e._zone.run(function(){return e._source.next(new p(n.matches,t))})}),e._registry.set(t,n)),n.matches&&r.push(new p(!0,t))}),r},t.prototype.buildMQL=function(t){return e=t,Object(i.isPlatformBrowser)(this._platformId)&&window.matchMedia("all").addListener?window.matchMedia(e):{matches:"all"===e||""===e,media:e,addListener:function(){},removeListener:function(){}};var e},t.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:r.NgZone},{type:Object,decorators:[{type:r.Inject,args:[r.PLATFORM_ID]}]},{type:void 0,decorators:[{type:r.Inject,args:[i.DOCUMENT]}]}]},t.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new t(Object(r.inject)(r.NgZone),Object(r.inject)(r.PLATFORM_ID),Object(r.inject)(i.DOCUMENT))},token:t,providedIn:"root"}),t}(),V={};var z=function(t){function e(e,n,r,i){var o=t.call(this,e,n,r)||this;return o._breakpoints=i,o.autoRegisterQueries=!0,o.useOverlaps=!1,o._registry=new Map,o._actives=[],o}return Object(o.__extends)(e,t),e.prototype.clearAll=function(){this._registry.forEach(function(t){t.destroy()}),this._registry.clear(),this.useOverlaps=!1},e.prototype.activate=function(t,e){return void 0===e&&(e=!1),e=e||this.useOverlaps,t=this._validateQuery(t),!e&&this.isActive(t)||(this._deactivateAll(),this._registerMediaQuery(t),this._activateWithOverlaps(t,e)),this.hasActivated},e.prototype._validateQuery=function(t){var e=this._breakpoints.findByAlias(t);return e&&e.mediaQuery||t},e.prototype._activateWithOverlaps=function(t,e){if(e){var n=this._breakpoints.findByQuery(t),r=n?n.alias:"unknown";switch(r){case"lg":this._activateByAlias("lt-xl");break;case"md":this._activateByAlias("lt-xl, lt-lg");break;case"sm":this._activateByAlias("lt-xl, lt-lg, lt-md");break;case"xs":this._activateByAlias("lt-xl, lt-lg, lt-md, lt-sm")}switch(r){case"xl":this._activateByAlias("gt-lg, gt-md, gt-sm, gt-xs");break;case"lg":this._activateByAlias("gt-md, gt-sm, gt-xs");break;case"md":this._activateByAlias("gt-sm, gt-xs");break;case"sm":this._activateByAlias("gt-xs")}}return this._activateByQuery(t)},e.prototype._activateByAlias=function(t){var e=this;t.split(",").forEach(function(t){return n=t.trim(),r=e._breakpoints.findByAlias(n),void e._activateByQuery(r?r.mediaQuery:n);var n,r})},e.prototype._activateByQuery=function(t){var e=this._registry.get(t),n=this._actives.reduce(function(t,n){return t||!!e&&n.media===e.media},!1);return e&&!n&&this._actives.push(e.activate()),this.hasActivated},e.prototype._deactivateAll=function(){return this._actives.forEach(function(t){return t.deactivate()}),this._actives=[],this},e.prototype._registerMediaQuery=function(t){!this._registry.has(t)&&this.autoRegisterQueries&&this.registerQuery(t)},e.prototype.buildMQL=function(t){return new H(t)},Object.defineProperty(e.prototype,"hasActivated",{get:function(){return this._actives.length>0},enumerable:!0,configurable:!0}),e.decorators=[{type:r.Injectable}],e.ctorParameters=function(){return[{type:r.NgZone},{type:Object,decorators:[{type:r.Inject,args:[r.PLATFORM_ID]}]},{type:void 0,decorators:[{type:r.Inject,args:[i.DOCUMENT]}]},{type:L}]},e}(B),H=function(){function t(t){this._mediaQuery=t,this._isActive=!1,this._listeners=[],this.onchange=null}return Object.defineProperty(t.prototype,"matches",{get:function(){return this._isActive},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"media",{get:function(){return this._mediaQuery},enumerable:!0,configurable:!0}),t.prototype.destroy=function(){this.deactivate(),this._listeners=[]},t.prototype.activate=function(){var t=this;return this._isActive||(this._isActive=!0,this._listeners.forEach(function(e){e.call(null,t)})),this},t.prototype.deactivate=function(){var t=this;return this._isActive&&(this._isActive=!1,this._listeners.forEach(function(e){e.call(null,t)})),this},t.prototype.addListener=function(t){(-1===this._listeners.indexOf(t)&&this._listeners.push(t),this._isActive)&&t.call(null,this)},t.prototype.removeListener=function(t){},t.prototype.addEventListener=function(t,e,n){},t.prototype.removeEventListener=function(t,e,n){},t.prototype.dispatchEvent=function(t){return!1},t}(),U={provide:B,useClass:z},W=function(){function t(t){this._mediaQuery=t,this._isActive=!1,this._listeners=[],this.onchange=null}return Object.defineProperty(t.prototype,"matches",{get:function(){return this._isActive},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"media",{get:function(){return this._mediaQuery},enumerable:!0,configurable:!0}),t.prototype.destroy=function(){this.deactivate(),this._listeners=[]},t.prototype.activate=function(){var t=this;return this._isActive||(this._isActive=!0,this._listeners.forEach(function(e){e.call(null,t)})),this},t.prototype.deactivate=function(){var t=this;return this._isActive&&(this._isActive=!1,this._listeners.forEach(function(e){e.call(null,t)})),this},t.prototype.addListener=function(t){(-1===this._listeners.indexOf(t)&&this._listeners.push(t),this._isActive)&&t.call(null,this)},t.prototype.removeListener=function(t){},t.prototype.addEventListener=function(t,e,n){},t.prototype.removeEventListener=function(t,e,n){},t.prototype.dispatchEvent=function(t){return!1},t}(),q=function(t){function e(e,n,r){var i=t.call(this,e,n,r)||this;return i._zone=e,i._platformId=n,i._document=r,i._registry=new Map,i}return Object(o.__extends)(e,t),e.prototype.activateBreakpoint=function(t){var e=this._registry.get(t.mediaQuery);e&&e.activate()},e.prototype.deactivateBreakpoint=function(t){var e=this._registry.get(t.mediaQuery);e&&e.deactivate()},e.prototype.buildMQL=function(t){return new W(t)},e.decorators=[{type:r.Injectable}],e.ctorParameters=function(){return[{type:r.NgZone},{type:Object,decorators:[{type:r.Inject,args:[r.PLATFORM_ID]}]},{type:void 0,decorators:[{type:r.Inject,args:[i.DOCUMENT]}]}]},e}(B);function G(t,e){return A(t||{},e?{mqAlias:e.alias,suffix:e.suffix}:{})}var Y="print",K={alias:Y,mediaQuery:Y,priority:1e3},X=function(){function t(t,e){this.breakpoints=t,this.layoutConfig=e,this.isPrinting=!1,this.queue=new Q,this.deactivations=[]}return t.prototype.withPrintQuery=function(t){return t.concat([Y])},t.prototype.isPrintEvent=function(t){return t.mediaQuery.startsWith(Y)},Object.defineProperty(t.prototype,"printAlias",{get:function(){return this.layoutConfig.printWithBreakpoints||[]},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"printBreakPoints",{get:function(){var t=this;return this.printAlias.map(function(e){return t.breakpoints.findByAlias(e)}).filter(function(t){return null!==t})},enumerable:!0,configurable:!0}),t.prototype.getEventBreakpoints=function(t){var e=t.mediaQuery,n=this.breakpoints.findByQuery(e);return(n?this.printBreakPoints.concat([n]):this.printBreakPoints).sort(R)},t.prototype.updateEvent=function(t){var e=this.breakpoints.findByQuery(t.mediaQuery);return this.isPrintEvent(t)&&(e=this.getEventBreakpoints(t)[0],t.mediaQuery=e?e.mediaQuery:""),G(t,e)},t.prototype.interceptEvents=function(t){var e=this;return function(n){e.isPrintEvent(n)?n.matches&&!e.isPrinting?(e.startPrinting(t,e.getEventBreakpoints(n)),t.updateStyles()):!n.matches&&e.isPrinting&&(e.stopPrinting(t),t.updateStyles()):e.collectActivations(n)}},t.prototype.blockPropagation=function(){var t=this;return function(e){return!(t.isPrinting||t.isPrintEvent(e))}},t.prototype.startPrinting=function(t,e){this.isPrinting=!0,t.activatedBreakpoints=this.queue.addPrintBreakpoints(e)},t.prototype.stopPrinting=function(t){t.activatedBreakpoints=this.deactivations,this.deactivations=[],this.queue.clear(),this.isPrinting=!1},t.prototype.collectActivations=function(t){if(!this.isPrinting)if(t.matches)this.deactivations=[];else{var e=this.breakpoints.findByQuery(t.mediaQuery);e&&(this.deactivations.push(e),this.deactivations.sort(R))}},t.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:L},{type:void 0,decorators:[{type:r.Inject,args:[m]}]}]},t.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new t(Object(r.inject)(L),Object(r.inject)(m))},token:t,providedIn:"root"}),t}(),Q=function(){function t(){this.printBreakpoints=[]}return t.prototype.addPrintBreakpoints=function(t){var e=this;return t.push(K),t.sort(R),t.forEach(function(t){return e.addBreakpoint(t)}),this.printBreakpoints},t.prototype.addBreakpoint=function(t){var e;t&&(void 0===this.printBreakpoints.find(function(e){return e.mediaQuery===t.mediaQuery})&&(this.printBreakpoints=(e=t)&&e.mediaQuery.startsWith(Y)?[t].concat(this.printBreakpoints):this.printBreakpoints.concat([t])))},t.prototype.clear=function(){this.printBreakpoints=[]},t}();var Z=function(){function t(t,e,n){this.breakpoints=t,this.mediaWatcher=e,this.hook=n,this.filterOverlaps=!0,this.media$=this.watchActivations()}return t.prototype.isActive=function(t){return this.mediaWatcher.isActive(this.toMediaQuery(t))},t.prototype.watchActivations=function(){var t=this.breakpoints.items.map(function(t){return t.mediaQuery});return this.buildObservable(t)},t.prototype.buildObservable=function(t){var e=this,n=this.breakpoints;return this.mediaWatcher.observe(this.hook.withPrintQuery(t)).pipe(Object(s.filter)(function(t){return t.matches}),Object(s.filter)(function(t){var r=n.findByQuery(t.mediaQuery);return!r||!(e.filterOverlaps&&r.overlapping)}),Object(s.map)(function(t){return e.hook.isPrintEvent(t)?e.hook.updateEvent(t):G(t,n.findByQuery(t.mediaQuery))}),Object(s.filter)(function(t){return!t.mediaQuery.startsWith("print")}),Object(s.filter)(function(t){return""!==t.mediaQuery}))},t.prototype.toMediaQuery=function(t){var e=this.breakpoints,n=e.findByAlias(t)||e.findByQuery(t);return n?n.mediaQuery:t},t.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:L},{type:B},{type:X}]},t.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new t(Object(r.inject)(L),Object(r.inject)(B),Object(r.inject)(X))},token:t,providedIn:"root"}),t}();function $(t){for(var e in t){var n=t[e]||"";switch(e){case"display":t.display="flex"===n?["-webkit-flex","flex"]:"inline-flex"===n?["-webkit-inline-flex","inline-flex"]:n;break;case"align-items":case"align-self":case"align-content":case"flex":case"flex-basis":case"flex-flow":case"flex-grow":case"flex-shrink":case"flex-wrap":case"justify-content":t["-webkit-"+e]=n;break;case"flex-direction":n=n||"row",t["-webkit-flex-direction"]=n,t["flex-direction"]=n;break;case"order":t.order=t["-webkit-"+e]=isNaN(+n)?"0":n}}return t}var J=function(){function t(t,e,n,r){this._serverStylesheet=t,this._serverModuleLoaded=e,this._platformId=n,this.layoutConfig=r}return t.prototype.applyStyleToElement=function(t,e,n){void 0===n&&(n=null);var r={};"string"==typeof e&&(r[e]=n,e=r),r=this.layoutConfig.disableVendorPrefixes?e:$(e),this._applyMultiValueStyleToElement(r,t)},t.prototype.applyStyleToElements=function(t,e){var n=this;void 0===e&&(e=[]);var r=this.layoutConfig.disableVendorPrefixes?t:$(t);e.forEach(function(t){n._applyMultiValueStyleToElement(r,t)})},t.prototype.getFlowDirection=function(t){var e=this.lookupStyle(t,"flex-direction"),n=this.lookupInlineStyle(t,"flex-direction")||Object(i.isPlatformServer)(this._platformId)&&this._serverModuleLoaded?e:"";return[e||"row",n]},t.prototype.lookupAttributeValue=function(t,e){return t.getAttribute(e)||""},t.prototype.lookupInlineStyle=function(t,e){return Object(i.isPlatformBrowser)(this._platformId)?t.style.getPropertyValue(e):this._getServerStyle(t,e)},t.prototype.lookupStyle=function(t,e,n){void 0===n&&(n=!1);var r="";t&&((r=this.lookupInlineStyle(t,e))||(Object(i.isPlatformBrowser)(this._platformId)?n||(r=getComputedStyle(t).getPropertyValue(e)):this._serverModuleLoaded&&(r=this._serverStylesheet.getStyleForElement(t,e))));return r.trim()},t.prototype._applyMultiValueStyleToElement=function(t,e){var n=this;Object.keys(t).sort().forEach(function(r){var o=t[r],a=Array.isArray(o)?o:[o];a.sort();for(var s=0,u=a;s<u.length;s++){var c=u[s];c=c?c+"":"",Object(i.isPlatformBrowser)(n._platformId)||!n._serverModuleLoaded?Object(i.isPlatformBrowser)(n._platformId)?e.style.setProperty(r,c):n._setServerStyle(e,r,c):n._serverStylesheet.addStyleToElement(e,r,c)}})},t.prototype._setServerStyle=function(t,e,n){e=e.replace(/([a-z])([A-Z])/g,"$1-$2").toLowerCase();var r=this._readStyleAttribute(t);r[e]=n||"",this._writeStyleAttribute(t,r)},t.prototype._getServerStyle=function(t,e){return this._readStyleAttribute(t)[e]||""},t.prototype._readStyleAttribute=function(t){var e={},n=t.getAttribute("style");if(n)for(var r=n.split(/;+/g),i=0;i<r.length;i++){var o=r[i].trim();if(o.length>0){var a=o.indexOf(":");if(-1===a)throw new Error("Invalid CSS style: "+o);e[o.substr(0,a).trim()]=o.substr(a+1).trim()}}return e},t.prototype._writeStyleAttribute=function(t,e){var n="";for(var r in e){e[r]&&(n+=r+":"+e[r]+";")}t.setAttribute("style",n)},t.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:f,decorators:[{type:r.Optional}]},{type:Boolean,decorators:[{type:r.Optional},{type:r.Inject,args:[g]}]},{type:Object,decorators:[{type:r.Inject,args:[r.PLATFORM_ID]}]},{type:void 0,decorators:[{type:r.Inject,args:[m]}]}]},t.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new t(Object(r.inject)(f,8),Object(r.inject)(g,8),Object(r.inject)(r.PLATFORM_ID),Object(r.inject)(m))},token:t,providedIn:"root"}),t}(),tt=function(){function t(){this.shouldCache=!0}return t.prototype.sideEffect=function(t,e,n){},t}();function et(t,e,n){void 0===e&&(e="1"),void 0===n&&(n="1");var r=[e,n,t],i=t.indexOf("calc");if(i>0)r[2]=nt(t.substring(i).trim()),2==(o=t.substr(0,i).trim().split(" ")).length&&(r[0]=o[0],r[1]=o[1]);else if(0==i)r[2]=nt(t.trim());else{var o;r=3===(o=t.split(" ")).length?o:[e,n,t]}return r}function nt(t){return t.replace(/[\s]/g,"").replace(/[\/\*\+\-]/g," $& ")}var rt=function(){function t(t,e,n){this.matchMedia=t,this.breakpoints=e,this.hook=n,this.activatedBreakpoints=[],this.elementMap=new Map,this.elementKeyMap=new WeakMap,this.watcherMap=new WeakMap,this.updateMap=new WeakMap,this.clearMap=new WeakMap,this.subject=new a.Subject,this.observeActivations()}return Object.defineProperty(t.prototype,"activatedAlias",{get:function(){return this.activatedBreakpoints[0]?this.activatedBreakpoints[0].alias:""},enumerable:!0,configurable:!0}),t.prototype.onMediaChange=function(t){var e=this.findByQuery(t.mediaQuery);e&&((t=G(t,e)).matches&&-1===this.activatedBreakpoints.indexOf(e)?(this.activatedBreakpoints.push(e),this.activatedBreakpoints.sort(R),this.updateStyles()):t.matches||-1===this.activatedBreakpoints.indexOf(e)||(this.activatedBreakpoints.splice(this.activatedBreakpoints.indexOf(e),1),this.activatedBreakpoints.sort(R),this.updateStyles()))},t.prototype.init=function(t,e,n,r,i){void 0===i&&(i=[]),it(this.updateMap,t,e,n),it(this.clearMap,t,e,r),this.buildElementKeyMap(t,e),this.watchExtraTriggers(t,e,i)},t.prototype.getValue=function(t,e,n){var r=this.elementMap.get(t);if(r){var i=void 0!==n?r.get(n):this.getActivatedValues(r,e);if(i)return i.get(e)}},t.prototype.hasValue=function(t,e){var n=this.elementMap.get(t);if(n){var r=this.getActivatedValues(n,e);if(r)return void 0!==r.get(e)||!1}return!1},t.prototype.setValue=function(t,e,n,r){var i=this.elementMap.get(t);if(i){var o=(i.get(r)||new Map).set(e,n);i.set(r,o),this.elementMap.set(t,i)}else i=(new Map).set(r,(new Map).set(e,n)),this.elementMap.set(t,i);var a=this.getValue(t,e);void 0!==a&&this.updateElement(t,e,a)},t.prototype.trackValue=function(t,e){return this.subject.asObservable().pipe(Object(s.filter)(function(n){return n.element===t&&n.key===e}))},t.prototype.updateStyles=function(){var t=this;this.elementMap.forEach(function(e,n){var r=new Set(t.elementKeyMap.get(n)),i=t.getActivatedValues(e);i&&i.forEach(function(e,i){t.updateElement(n,i,e),r.delete(i)}),r.forEach(function(r){if(i=t.getActivatedValues(e,r)){var o=i.get(r);t.updateElement(n,r,o)}else t.clearElement(n,r)})})},t.prototype.clearElement=function(t,e){var n=this.clearMap.get(t);if(n){var r=n.get(e);r&&(r(),this.subject.next({element:t,key:e,value:""}))}},t.prototype.updateElement=function(t,e,n){var r=this.updateMap.get(t);if(r){var i=r.get(e);i&&(i(n),this.subject.next({element:t,key:e,value:n}))}},t.prototype.releaseElement=function(t){var e=this.watcherMap.get(t);e&&(e.forEach(function(t){return t.unsubscribe()}),this.watcherMap.delete(t));var n=this.elementMap.get(t);n&&(n.forEach(function(t,e){return n.delete(e)}),this.elementMap.delete(t))},t.prototype.buildElementKeyMap=function(t,e){var n=this.elementKeyMap.get(t);n||(n=new Set,this.elementKeyMap.set(t,n)),n.add(e)},t.prototype.watchExtraTriggers=function(t,e,n){var r=this;if(n&&n.length){var i=this.watcherMap.get(t);if(i||(i=new Map,this.watcherMap.set(t,i)),!i.get(e)){var o=a.merge.apply(void 0,n).subscribe(function(){var n=r.getValue(t,e);r.updateElement(t,e,n)});i.set(e,o)}}},t.prototype.findByQuery=function(t){return this.breakpoints.findByQuery(t)},t.prototype.getActivatedValues=function(t,e){for(var n=0;n<this.activatedBreakpoints.length;n++){var r=this.activatedBreakpoints[n],i=t.get(r.alias);if(i&&(void 0===e||i.has(e)))return i}var o=t.get("");return void 0===e||o&&o.has(e)?o:void 0},t.prototype.observeActivations=function(){var t=this.breakpoints.items.map(function(t){return t.mediaQuery});this.matchMedia.observe(this.hook.withPrintQuery(t)).pipe(Object(s.tap)(this.hook.interceptEvents(this)),Object(s.filter)(this.hook.blockPropagation())).subscribe(this.onMediaChange.bind(this))},t.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:B},{type:L},{type:X}]},t.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new t(Object(r.inject)(B),Object(r.inject)(L),Object(r.inject)(X))},token:t,providedIn:"root"}),t}();function it(t,e,n,r){if(void 0!==r){var i=t.get(e);i||(i=new Map,t.set(e,i)),i.set(n,r)}}},"./node_modules/@angular/flex-layout/esm5/extended.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"ExtendedModule",function(){return I}),n.d(e,"ClassDirective",function(){return m}),n.d(e,"DefaultClassDirective",function(){return y}),n.d(e,"ImgSrcStyleBuilder",function(){return l}),n.d(e,"ImgSrcDirective",function(){return d}),n.d(e,"DefaultImgSrcDirective",function(){return h}),n.d(e,"ShowHideStyleBuilder",function(){return _}),n.d(e,"ShowHideDirective",function(){return v}),n.d(e,"DefaultShowHideDirective",function(){return w}),n.d(e,"StyleDirective",function(){return M}),n.d(e,"DefaultStyleDirective",function(){return k});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/common/fesm5/common.js"),a=n("./node_modules/@angular/flex-layout/esm5/core.es5.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/rxjs/_esm5/operators/index.js"),c=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),l=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t){return{content:t?"url("+t+")":""}},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(a.StyleBuilder),d=function(t){function e(e,n,r,i,a,s){var u=t.call(this,e,n,r,i)||this;return u.elementRef=e,u.styleBuilder=n,u.styler=r,u.marshal=i,u.platformId=a,u.serverModuleLoaded=s,u.DIRECTIVE_KEY="img-src",u.defaultSrc="",u.styleCache=p,u.init(),u.setValue("",u.nativeElement.getAttribute("src")||""),Object(o.isPlatformServer)(u.platformId)&&u.serverModuleLoaded&&u.nativeElement.setAttribute("src",""),u}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"src",{set:function(t){this.defaultSrc=t,this.setValue("",this.defaultSrc)},enumerable:!0,configurable:!0}),e.prototype.updateWithValue=function(){var t=this.activatedValue||this.defaultSrc;Object(o.isPlatformServer)(this.platformId)&&this.serverModuleLoaded?this.addStyles(t):this.nativeElement.setAttribute("src",String(t))},e.ctorParameters=function(){return[{type:i.ElementRef},{type:l},{type:a.StyleUtils},{type:a.MediaMarshaller},{type:Object,decorators:[{type:i.Inject,args:[i.PLATFORM_ID]}]},{type:Boolean,decorators:[{type:i.Inject,args:[a.SERVER_TOKEN]}]}]},e.propDecorators={src:[{type:i.Input,args:["src"]}]},e}(a.BaseDirective2),p=new Map,f=["src.xs","src.sm","src.md","src.lg","src.xl","src.lt-sm","src.lt-md","src.lt-lg","src.lt-xl","src.gt-xs","src.gt-sm","src.gt-md","src.gt-lg"],h=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=f,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  img[src.xs],    img[src.sm],    img[src.md],    img[src.lg],   img[src.xl],\n  img[src.lt-sm], img[src.lt-md], img[src.lt-lg], img[src.lt-xl],\n  img[src.gt-xs], img[src.gt-sm], img[src.gt-md], img[src.gt-lg]\n",inputs:f}]}],e}(d),m=function(t){function e(e,n,r,i,a,s,u){var c=t.call(this,e,null,n,r)||this;return c.elementRef=e,c.styler=n,c.marshal=r,c.iterableDiffers=i,c.keyValueDiffers=a,c.renderer=s,c.ngClassInstance=u,c.DIRECTIVE_KEY="ngClass",c.ngClassInstance||(c.ngClassInstance=new o.NgClass(c.iterableDiffers,c.keyValueDiffers,c.elementRef,c.renderer)),c.init(),c}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"klass",{set:function(t){this.ngClassInstance.klass=t,this.setValue(t,"")},enumerable:!0,configurable:!0}),e.prototype.updateWithValue=function(t){this.ngClassInstance.ngClass=t,this.ngClassInstance.ngDoCheck()},e.prototype.ngDoCheck=function(){this.ngClassInstance.ngDoCheck()},e.ctorParameters=function(){return[{type:i.ElementRef},{type:a.StyleUtils},{type:a.MediaMarshaller},{type:i.IterableDiffers},{type:i.KeyValueDiffers},{type:i.Renderer2},{type:o.NgClass,decorators:[{type:i.Optional},{type:i.Self}]}]},e.propDecorators={klass:[{type:i.Input,args:["class"]}]},e}(a.BaseDirective2),g=["ngClass","ngClass.xs","ngClass.sm","ngClass.md","ngClass.lg","ngClass.xl","ngClass.lt-sm","ngClass.lt-md","ngClass.lt-lg","ngClass.lt-xl","ngClass.gt-xs","ngClass.gt-sm","ngClass.gt-md","ngClass.gt-lg"],y=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=g,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [ngClass], [ngClass.xs], [ngClass.sm], [ngClass.md], [ngClass.lg], [ngClass.xl],\n  [ngClass.lt-sm], [ngClass.lt-md], [ngClass.lt-lg], [ngClass.lt-xl],\n  [ngClass.gt-xs], [ngClass.gt-sm], [ngClass.gt-md], [ngClass.gt-lg]\n",inputs:g}]}],e}(m),_=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t,e){return{display:"true"===t?e.display:"none"}},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(a.StyleBuilder),v=function(t){function e(e,n,r,i,o,a,s){var u=t.call(this,e,n,r,i)||this;return u.elementRef=e,u.styleBuilder=n,u.styler=r,u.marshal=i,u.layoutConfig=o,u.platformId=a,u.serverModuleLoaded=s,u.DIRECTIVE_KEY="show-hide",u.display="",u.hasLayout=!1,u.hasFlexChild=!1,u}return Object(r.__extends)(e,t),e.prototype.ngAfterViewInit=function(){this.trackExtraTriggers();for(var t=Array.from(this.nativeElement.children),e=0;e<t.length;e++)if(this.marshal.hasValue(t[e],"flex")){this.hasFlexChild=!0;break}b.has(this.nativeElement)?this.display=b.get(this.nativeElement):(this.display=this.getDisplayStyle(),b.set(this.nativeElement,this.display)),this.init();var n=this.marshal.getValue(this.nativeElement,this.DIRECTIVE_KEY,"");void 0===n||""===n?this.setValue(!0,""):this.triggerUpdate()},e.prototype.ngOnChanges=function(t){var e=this;Object.keys(t).forEach(function(n){if(-1!==e.inputs.indexOf(n)){var r=n.split("."),i=r.slice(1).join("."),o=t[n].currentValue,a=""===o||0!==o&&Object(s.coerceBooleanProperty)(o);"fxHide"===r[0]&&(a=!a),e.setValue(a,i)}})},e.prototype.trackExtraTriggers=function(){var t=this;this.hasLayout=this.marshal.hasValue(this.nativeElement,"layout"),["layout","layout-align"].forEach(function(e){t.marshal.trackValue(t.nativeElement,e).pipe(Object(u.takeUntil)(t.destroySubject)).subscribe(t.triggerUpdate.bind(t))})},e.prototype.getDisplayStyle=function(){return this.hasLayout||this.hasFlexChild&&this.layoutConfig.addFlexToParent?"flex":this.styler.lookupStyle(this.nativeElement,"display",!0)},e.prototype.updateWithValue=function(t){void 0===t&&(t=!0),""!==t&&(this.addStyles(t?"true":"false",{display:this.display}),Object(o.isPlatformServer)(this.platformId)&&this.serverModuleLoaded&&this.nativeElement.style.setProperty("display",""))},e.ctorParameters=function(){return[{type:i.ElementRef},{type:_},{type:a.StyleUtils},{type:a.MediaMarshaller},{type:void 0,decorators:[{type:i.Inject,args:[a.LAYOUT_CONFIG]}]},{type:Object,decorators:[{type:i.Inject,args:[i.PLATFORM_ID]}]},{type:Boolean,decorators:[{type:i.Optional},{type:i.Inject,args:[a.SERVER_TOKEN]}]}]},e}(a.BaseDirective2),b=new WeakMap,x=["fxShow","fxShow.print","fxShow.xs","fxShow.sm","fxShow.md","fxShow.lg","fxShow.xl","fxShow.lt-sm","fxShow.lt-md","fxShow.lt-lg","fxShow.lt-xl","fxShow.gt-xs","fxShow.gt-sm","fxShow.gt-md","fxShow.gt-lg","fxHide","fxHide.print","fxHide.xs","fxHide.sm","fxHide.md","fxHide.lg","fxHide.xl","fxHide.lt-sm","fxHide.lt-md","fxHide.lt-lg","fxHide.lt-xl","fxHide.gt-xs","fxHide.gt-sm","fxHide.gt-md","fxHide.gt-lg"],w=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=x,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [fxShow], [fxShow.print],\n  [fxShow.xs], [fxShow.sm], [fxShow.md], [fxShow.lg], [fxShow.xl],\n  [fxShow.lt-sm], [fxShow.lt-md], [fxShow.lt-lg], [fxShow.lt-xl],\n  [fxShow.gt-xs], [fxShow.gt-sm], [fxShow.gt-md], [fxShow.gt-lg],\n  [fxHide], [fxHide.print],\n  [fxHide.xs], [fxHide.sm], [fxHide.md], [fxHide.lg], [fxHide.xl],\n  [fxHide.lt-sm], [fxHide.lt-md], [fxHide.lt-lg], [fxHide.lt-xl],\n  [fxHide.gt-xs], [fxHide.gt-sm], [fxHide.gt-md], [fxHide.gt-lg]\n",inputs:x}]}],e}(v),j=function(){return function(t,e,n){void 0===n&&(n=!0),this.key=t,this.value=e,this.key=n?t.replace(/['"]/g,"").trim():t.trim(),this.value=n?e.replace(/['"]/g,"").trim():e.trim(),this.value=this.value.replace(/;/,"")}}();function O(t){var e=typeof t;return"object"===e?t.constructor===Array?"array":t.constructor===Set?"set":"object":e}function C(t,e){var n,r=[];return"set"===O(t)?t.forEach(function(t){return r.push(t)}):Object.keys(t).forEach(function(e){r.push(e+":"+t[e])}),n=e,r.map(E).filter(function(t){return!!t}).map(function(t){return n&&(t.value=n(t.value)),t}).reduce(S,{})}function E(t){var e=t.split(":"),n=e[0],r=e.slice(1);return new j(n,r.join(":"))}function S(t,e){return e.key&&(t[e.key]=e.value),t}var M=function(t){function e(e,n,r,i,a,s,u,c,l){var d=t.call(this,e,null,n,r)||this;d.elementRef=e,d.styler=n,d.marshal=r,d.keyValueDiffers=i,d.renderer=a,d.sanitizer=s,d.ngStyleInstance=u,d.DIRECTIVE_KEY="ngStyle",d.ngStyleInstance||(d.ngStyleInstance=new o.NgStyle(d.keyValueDiffers,d.elementRef,d.renderer)),d.init();var p=d.nativeElement.getAttribute("style")||"";return d.fallbackStyles=d.buildStyleMap(p),d.isServer=c&&Object(o.isPlatformServer)(l),d}return Object(r.__extends)(e,t),e.prototype.updateWithValue=function(t){var e=this.buildStyleMap(t);this.ngStyleInstance.ngStyle=Object(r.__assign)({},this.fallbackStyles,e),this.isServer&&this.applyStyleToElement(e),this.ngStyleInstance.ngDoCheck()},e.prototype.clearStyles=function(){this.ngStyleInstance.ngStyle=this.fallbackStyles,this.ngStyleInstance.ngDoCheck()},e.prototype.buildStyleMap=function(t){var e,n=this,r=function(t){return n.sanitizer.sanitize(i.SecurityContext.STYLE,t)||""};if(t)switch(O(t)){case"string":return A((void 0===e&&(e=";"),String(t).trim().split(e).map(function(t){return t.trim()}).filter(function(t){return""!==t})),r);case"array":return A(t,r);case"set":default:return C(t,r)}return{}},e.prototype.ngDoCheck=function(){this.ngStyleInstance.ngDoCheck()},e.ctorParameters=function(){return[{type:i.ElementRef},{type:a.StyleUtils},{type:a.MediaMarshaller},{type:i.KeyValueDiffers},{type:i.Renderer2},{type:c.DomSanitizer},{type:o.NgStyle,decorators:[{type:i.Optional},{type:i.Self}]},{type:Boolean,decorators:[{type:i.Optional},{type:i.Inject,args:[a.SERVER_TOKEN]}]},{type:Object,decorators:[{type:i.Inject,args:[i.PLATFORM_ID]}]}]},e}(a.BaseDirective2),T=["ngStyle","ngStyle.xs","ngStyle.sm","ngStyle.md","ngStyle.lg","ngStyle.xl","ngStyle.lt-sm","ngStyle.lt-md","ngStyle.lt-lg","ngStyle.lt-xl","ngStyle.gt-xs","ngStyle.gt-sm","ngStyle.gt-md","ngStyle.gt-lg"],k=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=T,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [ngStyle],\n  [ngStyle.xs], [ngStyle.sm], [ngStyle.md], [ngStyle.lg], [ngStyle.xl],\n  [ngStyle.lt-sm], [ngStyle.lt-md], [ngStyle.lt-lg], [ngStyle.lt-xl],\n  [ngStyle.gt-xs], [ngStyle.gt-sm], [ngStyle.gt-md], [ngStyle.gt-lg]\n",inputs:T}]}],e}(M);function A(t,e){return t.map(E).filter(function(t){return!!t}).map(function(t){return e&&(t.value=e(t.value)),t}).reduce(S,{})}var D=[w,y,k,h],I=function(){function t(){}return t.decorators=[{type:i.NgModule,args:[{imports:[a.CoreModule],declarations:D.slice(),exports:D.slice()}]}],t}()},"./node_modules/@angular/flex-layout/esm5/flex-layout.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"VERSION",function(){return l}),n.d(e,"FlexLayoutModule",function(){return d});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/common/fesm5/common.js"),a=n("./node_modules/@angular/flex-layout/esm5/core.es5.js");n.d(e,"removeStyles",function(){return a.removeStyles}),n.d(e,"BROWSER_PROVIDER",function(){return a.BROWSER_PROVIDER}),n.d(e,"CLASS_NAME",function(){return a.CLASS_NAME}),n.d(e,"CoreModule",function(){return a.CoreModule}),n.d(e,"MediaChange",function(){return a.MediaChange}),n.d(e,"StylesheetMap",function(){return a.StylesheetMap}),n.d(e,"DEFAULT_CONFIG",function(){return a.DEFAULT_CONFIG}),n.d(e,"LAYOUT_CONFIG",function(){return a.LAYOUT_CONFIG}),n.d(e,"SERVER_TOKEN",function(){return a.SERVER_TOKEN}),n.d(e,"BREAKPOINT",function(){return a.BREAKPOINT}),n.d(e,"BaseDirective2",function(){return a.BaseDirective2}),n.d(e,"sortDescendingPriority",function(){return a.sortDescendingPriority}),n.d(e,"sortAscendingPriority",function(){return a.sortAscendingPriority}),n.d(e,"DEFAULT_BREAKPOINTS",function(){return a.DEFAULT_BREAKPOINTS}),n.d(e,"ScreenTypes",function(){return a.ScreenTypes}),n.d(e,"ORIENTATION_BREAKPOINTS",function(){return a.ORIENTATION_BREAKPOINTS}),n.d(e,"BreakPointRegistry",function(){return a.BreakPointRegistry}),n.d(e,"BREAKPOINTS",function(){return a.BREAKPOINTS}),n.d(e,"MatchMedia",function(){return a.MatchMedia}),n.d(e,"MockMatchMedia",function(){return a.MockMatchMedia}),n.d(e,"MockMediaQueryList",function(){return a.MockMediaQueryList}),n.d(e,"MockMatchMediaProvider",function(){return a.MockMatchMediaProvider}),n.d(e,"ServerMediaQueryList",function(){return a.ServerMediaQueryList}),n.d(e,"ServerMatchMedia",function(){return a.ServerMatchMedia}),n.d(e,"MediaObserver",function(){return a.MediaObserver}),n.d(e,"StyleUtils",function(){return a.StyleUtils}),n.d(e,"StyleBuilder",function(){return a.StyleBuilder}),n.d(e,"validateBasis",function(){return a.validateBasis}),n.d(e,"MediaMarshaller",function(){return a.MediaMarshaller}),n.d(e,"BREAKPOINT_PRINT",function(){return a.BREAKPOINT_PRINT}),n.d(e,"PrintHook",function(){return a.PrintHook});var s=n("./node_modules/@angular/flex-layout/esm5/extended.es5.js");n.d(e,"ExtendedModule",function(){return s.ExtendedModule}),n.d(e,"ClassDirective",function(){return s.ClassDirective}),n.d(e,"DefaultClassDirective",function(){return s.DefaultClassDirective}),n.d(e,"ImgSrcStyleBuilder",function(){return s.ImgSrcStyleBuilder}),n.d(e,"ImgSrcDirective",function(){return s.ImgSrcDirective}),n.d(e,"DefaultImgSrcDirective",function(){return s.DefaultImgSrcDirective}),n.d(e,"ShowHideStyleBuilder",function(){return s.ShowHideStyleBuilder}),n.d(e,"ShowHideDirective",function(){return s.ShowHideDirective}),n.d(e,"DefaultShowHideDirective",function(){return s.DefaultShowHideDirective}),n.d(e,"StyleDirective",function(){return s.StyleDirective}),n.d(e,"DefaultStyleDirective",function(){return s.DefaultStyleDirective});var u=n("./node_modules/@angular/flex-layout/esm5/flex.es5.js");n.d(e,"FlexModule",function(){return u.FlexModule}),n.d(e,"FlexStyleBuilder",function(){return u.FlexStyleBuilder}),n.d(e,"FlexDirective",function(){return u.FlexDirective}),n.d(e,"DefaultFlexDirective",function(){return u.DefaultFlexDirective}),n.d(e,"FlexAlignStyleBuilder",function(){return u.FlexAlignStyleBuilder}),n.d(e,"FlexAlignDirective",function(){return u.FlexAlignDirective}),n.d(e,"DefaultFlexAlignDirective",function(){return u.DefaultFlexAlignDirective}),n.d(e,"FlexFillStyleBuilder",function(){return u.FlexFillStyleBuilder}),n.d(e,"FlexFillDirective",function(){return u.FlexFillDirective}),n.d(e,"FlexOffsetStyleBuilder",function(){return u.FlexOffsetStyleBuilder}),n.d(e,"FlexOffsetDirective",function(){return u.FlexOffsetDirective}),n.d(e,"DefaultFlexOffsetDirective",function(){return u.DefaultFlexOffsetDirective}),n.d(e,"FlexOrderStyleBuilder",function(){return u.FlexOrderStyleBuilder}),n.d(e,"FlexOrderDirective",function(){return u.FlexOrderDirective}),n.d(e,"DefaultFlexOrderDirective",function(){return u.DefaultFlexOrderDirective}),n.d(e,"LayoutStyleBuilder",function(){return u.LayoutStyleBuilder}),n.d(e,"LayoutDirective",function(){return u.LayoutDirective}),n.d(e,"DefaultLayoutDirective",function(){return u.DefaultLayoutDirective}),n.d(e,"LayoutAlignStyleBuilder",function(){return u.LayoutAlignStyleBuilder}),n.d(e,"LayoutAlignDirective",function(){return u.LayoutAlignDirective}),n.d(e,"DefaultLayoutAlignDirective",function(){return u.DefaultLayoutAlignDirective}),n.d(e,"LayoutGapStyleBuilder",function(){return u.LayoutGapStyleBuilder}),n.d(e,"LayoutGapDirective",function(){return u.LayoutGapDirective}),n.d(e,"DefaultLayoutGapDirective",function(){return u.DefaultLayoutGapDirective});var c=n("./node_modules/@angular/flex-layout/esm5/grid.es5.js");n.d(e,"ɵf2",function(){return c["ɵf2"]}),n.d(e,"ɵe2",function(){return c["ɵe2"]}),n.d(e,"ɵd2",function(){return c["ɵd2"]}),n.d(e,"ɵi2",function(){return c["ɵi2"]}),n.d(e,"ɵh2",function(){return c["ɵh2"]}),n.d(e,"ɵg2",function(){return c["ɵg2"]}),n.d(e,"ɵl2",function(){return c["ɵl2"]}),n.d(e,"ɵk2",function(){return c["ɵk2"]}),n.d(e,"ɵj2",function(){return c["ɵj2"]}),n.d(e,"ɵo2",function(){return c["ɵo2"]}),n.d(e,"ɵn2",function(){return c["ɵn2"]}),n.d(e,"ɵm2",function(){return c["ɵm2"]}),n.d(e,"ɵr2",function(){return c["ɵr2"]}),n.d(e,"ɵq2",function(){return c["ɵq2"]}),n.d(e,"ɵp2",function(){return c["ɵp2"]}),n.d(e,"ɵu2",function(){return c["ɵu2"]}),n.d(e,"ɵt2",function(){return c["ɵt2"]}),n.d(e,"ɵs2",function(){return c["ɵs2"]}),n.d(e,"ɵx2",function(){return c["ɵx2"]}),n.d(e,"ɵw2",function(){return c["ɵw2"]}),n.d(e,"ɵv2",function(){return c["ɵv2"]}),n.d(e,"ɵba2",function(){return c["ɵba2"]}),n.d(e,"ɵz2",function(){return c["ɵz2"]}),n.d(e,"ɵy2",function(){return c["ɵy2"]}),n.d(e,"ɵc2",function(){return c["ɵc2"]}),n.d(e,"ɵb2",function(){return c["ɵb2"]}),n.d(e,"ɵa2",function(){return c["ɵa2"]}),n.d(e,"ɵbd2",function(){return c["ɵbd2"]}),n.d(e,"ɵbc2",function(){return c["ɵbc2"]}),n.d(e,"ɵbb2",function(){return c["ɵbb2"]}),n.d(e,"ɵbg2",function(){return c["ɵbg2"]}),n.d(e,"ɵbf2",function(){return c["ɵbf2"]}),n.d(e,"ɵbe2",function(){return c["ɵbe2"]}),n.d(e,"GridModule",function(){return c.GridModule});var l=new r.Version("7.0.0-beta.23"),d=function(){function t(t,e){Object(o.isPlatformServer)(e)&&!t&&console.warn("Warning: Flex Layout loaded on the server without FlexLayoutServerModule")}return t.withConfig=function(e,n){return void 0===n&&(n=[]),{ngModule:t,providers:e.serverLoaded?[{provide:a.LAYOUT_CONFIG,useValue:Object(i.__assign)({},a.DEFAULT_CONFIG,e)},{provide:a.BREAKPOINT,useValue:n,multi:!0},{provide:a.SERVER_TOKEN,useValue:!0}]:[{provide:a.LAYOUT_CONFIG,useValue:Object(i.__assign)({},a.DEFAULT_CONFIG,e)},{provide:a.BREAKPOINT,useValue:n,multi:!0}]}},t.decorators=[{type:r.NgModule,args:[{imports:[u.FlexModule,s.ExtendedModule,c.GridModule],exports:[u.FlexModule,s.ExtendedModule,c.GridModule]}]}],t.ctorParameters=function(){return[{type:Boolean,decorators:[{type:r.Optional},{type:r.Inject,args:[a.SERVER_TOKEN]}]},{type:Object,decorators:[{type:r.Inject,args:[r.PLATFORM_ID]}]}]},t}()},"./node_modules/@angular/flex-layout/esm5/flex.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"FlexModule",function(){return ht}),n.d(e,"FlexStyleBuilder",function(){return T}),n.d(e,"FlexDirective",function(){return A}),n.d(e,"DefaultFlexDirective",function(){return D}),n.d(e,"FlexAlignStyleBuilder",function(){return Q}),n.d(e,"FlexAlignDirective",function(){return $}),n.d(e,"DefaultFlexAlignDirective",function(){return tt}),n.d(e,"FlexFillStyleBuilder",function(){return nt}),n.d(e,"FlexFillDirective",function(){return rt}),n.d(e,"FlexOffsetStyleBuilder",function(){return H}),n.d(e,"FlexOffsetDirective",function(){return W}),n.d(e,"DefaultFlexOffsetDirective",function(){return q}),n.d(e,"FlexOrderStyleBuilder",function(){return F}),n.d(e,"FlexOrderDirective",function(){return B}),n.d(e,"DefaultFlexOrderDirective",function(){return z}),n.d(e,"LayoutStyleBuilder",function(){return f}),n.d(e,"LayoutDirective",function(){return m}),n.d(e,"DefaultLayoutDirective",function(){return g}),n.d(e,"LayoutAlignStyleBuilder",function(){return ot}),n.d(e,"LayoutAlignDirective",function(){return st}),n.d(e,"DefaultLayoutAlignDirective",function(){return ut}),n.d(e,"LayoutGapStyleBuilder",function(){return v}),n.d(e,"LayoutGapDirective",function(){return x}),n.d(e,"DefaultLayoutGapDirective",function(){return w});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/flex-layout/esm5/core.es5.js"),a=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),s=n("./node_modules/rxjs/_esm5/index.js"),u=n("./node_modules/rxjs/_esm5/operators/index.js"),c="inline",l=["row","column","row-reverse","column-reverse"];function d(t){var e=(t=t?t.toLowerCase():"").split(" "),n=e[0],r=e[1],i=e[2];return l.find(function(t){return t===n})||(n=l[0]),r===c&&(r=i!==c?i:"",i=c),[n,function(t){if(t)switch(t.toLowerCase()){case"reverse":case"wrap-reverse":case"reverse-wrap":t="wrap-reverse";break;case"no":case"none":case"nowrap":t="nowrap";break;default:t="wrap"}return t}(r),!!i]}function p(t){return d(t)[0].indexOf("row")>-1}var f=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t){return e=d(t),n=e[0],r=e[1],i=e[2],void 0===r&&(r=null),void 0===i&&(i=!1),{display:i?"inline-flex":"flex","box-sizing":"border-box","flex-direction":n,"flex-wrap":r||null};var e,n,r,i},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),h=["fxLayout","fxLayout.xs","fxLayout.sm","fxLayout.md","fxLayout.lg","fxLayout.xl","fxLayout.lt-sm","fxLayout.lt-md","fxLayout.lt-lg","fxLayout.lt-xl","fxLayout.gt-xs","fxLayout.gt-sm","fxLayout.gt-md","fxLayout.gt-lg"],m=function(t){function e(e,n,r,i){var o=t.call(this,e,r,n,i)||this;return o.elRef=e,o.styleUtils=n,o.styleBuilder=r,o.marshal=i,o.DIRECTIVE_KEY="layout",o.styleCache=y,o.init(),o}return Object(r.__extends)(e,t),e.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:f,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller}]},e}(o.BaseDirective2),g=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=h,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [fxLayout], [fxLayout.xs], [fxLayout.sm], [fxLayout.md],\n  [fxLayout.lg], [fxLayout.xl], [fxLayout.lt-sm], [fxLayout.lt-md],\n  [fxLayout.lt-lg], [fxLayout.lt-xl], [fxLayout.gt-xs], [fxLayout.gt-sm],\n  [fxLayout.gt-md], [fxLayout.gt-lg]\n",inputs:h}]}],e}(m),y=new Map,_={"margin-left":null,"margin-right":null,"margin-top":null,"margin-bottom":null},v=function(t){function e(e){var n=t.call(this)||this;return n._styler=e,n}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t,e){return t.endsWith(S)?function(t,e){var n="0px",r="0px";"rtl"===e?r="-"+t:n="-"+t;return{margin:"0px "+n+" -"+t+" "+r}}(t=t.slice(0,t.indexOf(S)),e.directionality):{}},e.prototype.sideEffect=function(t,e,n){var i=n.items;if(t.endsWith(S)){var o=function(t,e){var n="0px",r="0px";"rtl"===e?r=t:n=t;return{padding:"0px "+n+" "+t+" "+r}}(t=t.slice(0,t.indexOf(S)),n.directionality);this._styler.applyStyleToElements(o,n.items)}else{var a=i.pop(),s=function(t,e){var n,i=Object(r.__assign)({},_);switch(e.layout){case"column":n="margin-bottom";break;case"column-reverse":n="margin-top";break;case"row":n="rtl"===e.directionality?"margin-left":"margin-right";break;case"row-reverse":n="rtl"===e.directionality?"margin-right":"margin-left";break;default:n="rtl"===e.directionality?"margin-left":"margin-right"}return i[n]=t,i}(t,n);this._styler.applyStyleToElements(s,i),this._styler.applyStyleToElements(_,[a])}},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:o.StyleUtils}]},e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e(Object(i.inject)(o.StyleUtils))},token:e,providedIn:"root"}),e}(o.StyleBuilder),b=["fxLayoutGap","fxLayoutGap.xs","fxLayoutGap.sm","fxLayoutGap.md","fxLayoutGap.lg","fxLayoutGap.xl","fxLayoutGap.lt-sm","fxLayoutGap.lt-md","fxLayoutGap.lt-lg","fxLayoutGap.lt-xl","fxLayoutGap.gt-xs","fxLayoutGap.gt-sm","fxLayoutGap.gt-md","fxLayoutGap.gt-lg"],x=function(t){function e(e,n,r,i,o,a){var c=t.call(this,e,o,i,a)||this;c.elRef=e,c.zone=n,c.directionality=r,c.styleUtils=i,c.styleBuilder=o,c.marshal=a,c.layout="row",c.DIRECTIVE_KEY="layout-gap",c.observerSubject=new s.Subject;var l=[c.directionality.change,c.observerSubject.asObservable()];return c.init(l),c.marshal.trackValue(c.nativeElement,"layout").pipe(Object(u.takeUntil)(c.destroySubject)).subscribe(c.onLayoutChange.bind(c)),c}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"childrenNodes",{get:function(){for(var t=this.nativeElement.children,e=[],n=t.length;n--;)e[n]=t[n];return e},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){this.buildChildObservable(),this.triggerUpdate()},e.prototype.ngOnDestroy=function(){t.prototype.ngOnDestroy.call(this),this.observer&&this.observer.disconnect()},e.prototype.onLayoutChange=function(t){var e=this,n=t.value.split(" ");this.layout=n[0],l.find(function(t){return t===e.layout})||(this.layout="row"),this.triggerUpdate()},e.prototype.updateWithValue=function(t){var e=this,n=this.childrenNodes.filter(function(t){return 1===t.nodeType&&e.willDisplay(t)}).sort(function(t,n){var r=+e.styler.lookupStyle(t,"order"),i=+e.styler.lookupStyle(n,"order");return isNaN(r)||isNaN(i)||r===i?0:r>i?1:-1});if(n.length>0){var r=this.directionality.value,i=this.layout;"row"===i&&"rtl"===r?this.styleCache=j:"row"===i&&"rtl"!==r?this.styleCache=C:"column"===i&&"rtl"===r?this.styleCache=O:"column"===i&&"rtl"!==r&&(this.styleCache=E),this.addStyles(t,{directionality:r,items:n,layout:i})}},e.prototype.willDisplay=function(t){var e=this.marshal.getValue(t,"show-hide");return!0===e||void 0===e&&"none"!==this.styleUtils.lookupStyle(t,"display")},e.prototype.buildChildObservable=function(){var t=this;this.zone.runOutsideAngular(function(){"undefined"!=typeof MutationObserver&&(t.observer=new MutationObserver(function(e){e.some(function(t){return t.addedNodes&&t.addedNodes.length>0||t.removedNodes&&t.removedNodes.length>0})&&t.observerSubject.next()}),t.observer.observe(t.nativeElement,{childList:!0}))})},e.ctorParameters=function(){return[{type:i.ElementRef},{type:i.NgZone},{type:a.Directionality},{type:o.StyleUtils},{type:v,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller}]},e}(o.BaseDirective2),w=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=b,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [fxLayoutGap], [fxLayoutGap.xs], [fxLayoutGap.sm], [fxLayoutGap.md],\n  [fxLayoutGap.lg], [fxLayoutGap.xl], [fxLayoutGap.lt-sm], [fxLayoutGap.lt-md],\n  [fxLayoutGap.lt-lg], [fxLayoutGap.lt-xl], [fxLayoutGap.gt-xs], [fxLayoutGap.gt-sm],\n  [fxLayoutGap.gt-md], [fxLayoutGap.gt-lg]\n",inputs:b}]}],e}(x),j=new Map,O=new Map,C=new Map,E=new Map,S=" grid";function M(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];if(null==t)throw TypeError("Cannot convert undefined or null to object");for(var r=0,i=e;r<i.length;r++){var o=i[r];if(null!=o)for(var a in o)o.hasOwnProperty(a)&&(t[a]=o[a])}return t}var T=function(t){function e(e){var n=t.call(this)||this;return n.layoutConfig=e,n}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t,e){var n=t.split(" "),r=n[0],i=n[1],o=n.slice(2).join(" "),a=e.direction.indexOf("column")>-1?"column":"row",s=p(a)?"max-width":"max-height",u=p(a)?"min-width":"min-height",c=String(o).indexOf("calc")>-1,l=c||"auto"===o,d=String(o).indexOf("%")>-1&&!c,f=String(o).indexOf("px")>-1||String(o).indexOf("rem")>-1||String(o).indexOf("em")>-1||String(o).indexOf("vw")>-1||String(o).indexOf("vh")>-1,h=c||f;r="0"==r?0:r,i="0"==i?0:i;var m=!r&&!i,g={},y={"max-width":null,"max-height":null,"min-width":null,"min-height":null};switch(o||""){case"":var _=!1!==this.layoutConfig.useColumnBasisZero;o="row"===a?"0%":_?"0.000000001px":"auto";break;case"initial":case"nogrow":r=0,o="auto";break;case"grow":o="100%";break;case"noshrink":i=0,o="auto";break;case"auto":break;case"none":r=0,i=0,o="auto";break;default:h||d||isNaN(o)||(o+="%"),"0%"===o&&(h=!0),"0px"===o&&(o="0%"),g=M(y,c?{"flex-grow":r,"flex-shrink":i,"flex-basis":h?o:"100%"}:{flex:r+" "+i+" "+(h?o:"100%")})}return g.flex||g["flex-grow"]||(g=M(y,c?{"flex-grow":r,"flex-shrink":i,"flex-basis":o}:{flex:r+" "+i+" "+o})),"0%"!==o&&"0px"!==o&&"0.000000001px"!==o&&"auto"!==o&&(g[u]=m||h&&r?o:null,g[s]=m||!l&&i?o:null),g[u]||g[s]?e.hasWrap&&(g[c?"flex-basis":"flex"]=g[s]?c?g[s]:r+" "+i+" "+g[s]:c?g[u]:r+" "+i+" "+g[u]):g=M(y,c?{"flex-grow":r,"flex-shrink":i,"flex-basis":o}:{flex:r+" "+i+" "+o}),M(g,{"box-sizing":"border-box"})},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:i.Inject,args:[o.LAYOUT_CONFIG]}]}]},e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e(Object(i.inject)(o.LAYOUT_CONFIG))},token:e,providedIn:"root"}),e}(o.StyleBuilder),k=["fxFlex","fxFlex.xs","fxFlex.sm","fxFlex.md","fxFlex.lg","fxFlex.xl","fxFlex.lt-sm","fxFlex.lt-md","fxFlex.lt-lg","fxFlex.lt-xl","fxFlex.gt-xs","fxFlex.gt-sm","fxFlex.gt-md","fxFlex.gt-lg"],A=function(t){function e(e,n,r,i,o){var a=t.call(this,e,i,n,o)||this;return a.elRef=e,a.styleUtils=n,a.layoutConfig=r,a.styleBuilder=i,a.marshal=o,a.DIRECTIVE_KEY="flex",a.direction="",a.wrap=!1,a.flexGrow="1",a.flexShrink="1",a.init(),a.parentElement&&a.marshal.trackValue(a.parentElement,"layout").pipe(Object(u.takeUntil)(a.destroySubject)).subscribe(a.onLayoutChange.bind(a)),a}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"shrink",{get:function(){return this.flexShrink},set:function(t){this.flexShrink=t||"1",this.triggerReflow()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"grow",{get:function(){return this.flexGrow},set:function(t){this.flexGrow=t||"1",this.triggerReflow()},enumerable:!0,configurable:!0}),e.prototype.onLayoutChange=function(t){var e=t.value.split(" ");this.direction=e[0],this.wrap=void 0!==e[1]&&"wrap"===e[1],this.triggerUpdate()},e.prototype.updateWithValue=function(t){var e=!1!==this.layoutConfig.addFlexToParent;this.direction||(this.direction=this.getFlexFlowDirection(this.parentElement,e));var n=this.direction,r=n.startsWith("row"),i=this.wrap;r&&i?this.styleCache=R:r&&!i?this.styleCache=I:!r&&i?this.styleCache=N:r||i||(this.styleCache=P);var a=String(t).replace(";",""),s=Object(o.validateBasis)(a,this.flexGrow,this.flexShrink);this.addStyles(s.join(" "),{direction:n,hasWrap:i})},e.prototype.triggerReflow=function(){var t=this.activatedValue;if(void 0!==t){var e=Object(o.validateBasis)(t,this.flexGrow,this.flexShrink);this.marshal.updateElement(this.nativeElement,this.DIRECTIVE_KEY,e.join(" "))}},e.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:void 0,decorators:[{type:i.Inject,args:[o.LAYOUT_CONFIG]}]},{type:T},{type:o.MediaMarshaller}]},e.propDecorators={shrink:[{type:i.Input,args:["fxShrink"]}],grow:[{type:i.Input,args:["fxGrow"]}]},e}(o.BaseDirective2),D=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=k,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{inputs:k,selector:"\n  [fxFlex], [fxFlex.xs], [fxFlex.sm], [fxFlex.md],\n  [fxFlex.lg], [fxFlex.xl], [fxFlex.lt-sm], [fxFlex.lt-md],\n  [fxFlex.lt-lg], [fxFlex.lt-xl], [fxFlex.gt-xs], [fxFlex.gt-sm],\n  [fxFlex.gt-md], [fxFlex.gt-lg]\n"}]}],e}(A),I=new Map,P=new Map,R=new Map,N=new Map,F=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t){return{order:t&&parseInt(t,10)||""}},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),L=["fxFlexOrder","fxFlexOrder.xs","fxFlexOrder.sm","fxFlexOrder.md","fxFlexOrder.lg","fxFlexOrder.xl","fxFlexOrder.lt-sm","fxFlexOrder.lt-md","fxFlexOrder.lt-lg","fxFlexOrder.lt-xl","fxFlexOrder.gt-xs","fxFlexOrder.gt-sm","fxFlexOrder.gt-md","fxFlexOrder.gt-lg"],B=function(t){function e(e,n,r,i){var o=t.call(this,e,r,n,i)||this;return o.elRef=e,o.styleUtils=n,o.styleBuilder=r,o.marshal=i,o.DIRECTIVE_KEY="flex-order",o.styleCache=V,o.init(),o}return Object(r.__extends)(e,t),e.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:F,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller}]},e}(o.BaseDirective2),V=new Map,z=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=L,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [fxFlexOrder], [fxFlexOrder.xs], [fxFlexOrder.sm], [fxFlexOrder.md],\n  [fxFlexOrder.lg], [fxFlexOrder.xl], [fxFlexOrder.lt-sm], [fxFlexOrder.lt-md],\n  [fxFlexOrder.lt-lg], [fxFlexOrder.lt-xl], [fxFlexOrder.gt-xs], [fxFlexOrder.gt-sm],\n  [fxFlexOrder.gt-md], [fxFlexOrder.gt-lg]\n",inputs:L}]}],e}(B),H=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t,e){var n;""===t&&(t="0");var r=String(t).indexOf("%")>-1;String(t).indexOf("px")>-1||r||isNaN(+t)||(t+="%");var i=e.isRtl?"margin-right":"margin-left";return p(e.layout)?((n={})[i]=""+t,n):{"margin-top":""+t}},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),U=["fxFlexOffset","fxFlexOffset.xs","fxFlexOffset.sm","fxFlexOffset.md","fxFlexOffset.lg","fxFlexOffset.xl","fxFlexOffset.lt-sm","fxFlexOffset.lt-md","fxFlexOffset.lt-lg","fxFlexOffset.lt-xl","fxFlexOffset.gt-xs","fxFlexOffset.gt-sm","fxFlexOffset.gt-md","fxFlexOffset.gt-lg"],W=function(t){function e(e,n,r,i,o){var a=t.call(this,e,r,o,i)||this;return a.elRef=e,a.directionality=n,a.styleBuilder=r,a.marshal=i,a.styler=o,a.DIRECTIVE_KEY="flex-offset",a.init([a.directionality.change]),a.parentElement&&a.marshal.trackValue(a.parentElement,"layout-gap").pipe(Object(u.takeUntil)(a.destroySubject)).subscribe(a.triggerUpdate.bind(a)),a}return Object(r.__extends)(e,t),e.prototype.updateWithValue=function(t){void 0===t&&(t="");var e=this.getFlexFlowDirection(this.parentElement,!0),n="rtl"===this.directionality.value;"row"===e&&n?this.styleCache=G:"row"!==e||n?"column"===e&&n?this.styleCache=Y:"column"!==e||n||(this.styleCache=X):this.styleCache=K,this.addStyles(t+"",{layout:e,isRtl:n})},e.ctorParameters=function(){return[{type:i.ElementRef},{type:a.Directionality},{type:H,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller},{type:o.StyleUtils}]},e}(o.BaseDirective2),q=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=U,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [fxFlexOffset], [fxFlexOffset.xs], [fxFlexOffset.sm], [fxFlexOffset.md],\n  [fxFlexOffset.lg], [fxFlexOffset.xl], [fxFlexOffset.lt-sm], [fxFlexOffset.lt-md],\n  [fxFlexOffset.lt-lg], [fxFlexOffset.lt-xl], [fxFlexOffset.gt-xs], [fxFlexOffset.gt-sm],\n  [fxFlexOffset.gt-md], [fxFlexOffset.gt-lg]\n",inputs:U}]}],e}(W),G=new Map,Y=new Map,K=new Map,X=new Map,Q=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t){var e={};switch(t=t||"stretch"){case"start":e["align-self"]="flex-start";break;case"end":e["align-self"]="flex-end";break;default:e["align-self"]=t}return e},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),Z=["fxFlexAlign","fxFlexAlign.xs","fxFlexAlign.sm","fxFlexAlign.md","fxFlexAlign.lg","fxFlexAlign.xl","fxFlexAlign.lt-sm","fxFlexAlign.lt-md","fxFlexAlign.lt-lg","fxFlexAlign.lt-xl","fxFlexAlign.gt-xs","fxFlexAlign.gt-sm","fxFlexAlign.gt-md","fxFlexAlign.gt-lg"],$=function(t){function e(e,n,r,i){var o=t.call(this,e,r,n,i)||this;return o.elRef=e,o.styleUtils=n,o.styleBuilder=r,o.marshal=i,o.DIRECTIVE_KEY="flex-align",o.styleCache=J,o.init(),o}return Object(r.__extends)(e,t),e.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:Q,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller}]},e}(o.BaseDirective2),J=new Map,tt=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=Z,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [fxFlexAlign], [fxFlexAlign.xs], [fxFlexAlign.sm], [fxFlexAlign.md],\n  [fxFlexAlign.lg], [fxFlexAlign.xl], [fxFlexAlign.lt-sm], [fxFlexAlign.lt-md],\n  [fxFlexAlign.lt-lg], [fxFlexAlign.lt-xl], [fxFlexAlign.gt-xs], [fxFlexAlign.gt-sm],\n  [fxFlexAlign.gt-md], [fxFlexAlign.gt-lg]\n",inputs:Z}]}],e}($),et={margin:0,width:"100%",height:"100%","min-width":"100%","min-height":"100%"},nt=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t){return et},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),rt=function(t){function e(e,n,r,i){var o=t.call(this,e,r,n,i)||this;return o.elRef=e,o.styleUtils=n,o.styleBuilder=r,o.marshal=i,o.styleCache=it,o.addStyles(""),o}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"[fxFill], [fxFlexFill]"}]}],e.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:nt},{type:o.MediaMarshaller}]},e}(o.BaseDirective2),it=new Map,ot=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t,e){var n={},r=t.split(" "),i=r[0],o=r[1];switch(i){case"center":n["justify-content"]="center";break;case"space-around":n["justify-content"]="space-around";break;case"space-between":n["justify-content"]="space-between";break;case"space-evenly":n["justify-content"]="space-evenly";break;case"end":case"flex-end":n["justify-content"]="flex-end";break;case"start":case"flex-start":default:n["justify-content"]="flex-start"}switch(o){case"start":case"flex-start":n["align-items"]=n["align-content"]="flex-start";break;case"center":n["align-items"]=n["align-content"]="center";break;case"end":case"flex-end":n["align-items"]=n["align-content"]="flex-end";break;case"space-between":n["align-content"]="space-between",n["align-items"]="stretch";break;case"space-around":n["align-content"]="space-around",n["align-items"]="stretch";break;case"baseline":n["align-content"]="stretch",n["align-items"]="baseline";break;case"stretch":default:n["align-items"]=n["align-content"]="stretch"}return M(n,{display:"flex","flex-direction":e.layout,"box-sizing":"border-box","max-width":"stretch"===o?p(e.layout)?null:"100%":null,"max-height":"stretch"===o&&p(e.layout)?"100%":null})},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),at=["fxLayoutAlign","fxLayoutAlign.xs","fxLayoutAlign.sm","fxLayoutAlign.md","fxLayoutAlign.lg","fxLayoutAlign.xl","fxLayoutAlign.lt-sm","fxLayoutAlign.lt-md","fxLayoutAlign.lt-lg","fxLayoutAlign.lt-xl","fxLayoutAlign.gt-xs","fxLayoutAlign.gt-sm","fxLayoutAlign.gt-md","fxLayoutAlign.gt-lg"],st=function(t){function e(e,n,r,i){var o=t.call(this,e,r,n,i)||this;return o.elRef=e,o.styleUtils=n,o.styleBuilder=r,o.marshal=i,o.DIRECTIVE_KEY="layout-align",o.layout="row",o.init(),o.marshal.trackValue(o.nativeElement,"layout").pipe(Object(u.takeUntil)(o.destroySubject)).subscribe(o.onLayoutChange.bind(o)),o}return Object(r.__extends)(e,t),e.prototype.updateWithValue=function(t){var e=this.layout||"row";"row"===e?this.styleCache=ct:"row-reverse"===e?this.styleCache=dt:"column"===e?this.styleCache=lt:"column-reverse"===e&&(this.styleCache=pt),this.addStyles(t,{layout:e})},e.prototype.onLayoutChange=function(t){var e=this,n=t.value;this.layout=n.split(" ")[0],l.find(function(t){return t===e.layout})||(this.layout="row"),this.triggerUpdate()},e.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:ot,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller}]},e}(o.BaseDirective2),ut=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=at,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [fxLayoutAlign], [fxLayoutAlign.xs], [fxLayoutAlign.sm], [fxLayoutAlign.md],\n  [fxLayoutAlign.lg], [fxLayoutAlign.xl], [fxLayoutAlign.lt-sm], [fxLayoutAlign.lt-md],\n  [fxLayoutAlign.lt-lg], [fxLayoutAlign.lt-xl], [fxLayoutAlign.gt-xs], [fxLayoutAlign.gt-sm],\n  [fxLayoutAlign.gt-md], [fxLayoutAlign.gt-lg]\n",inputs:at}]}],e}(st),ct=new Map,lt=new Map,dt=new Map,pt=new Map,ft=[g,w,ut,z,q,rt,tt,D],ht=function(){function t(){}return t.decorators=[{type:i.NgModule,args:[{imports:[o.CoreModule,a.BidiModule],declarations:ft.slice(),exports:ft.slice()}]}],t}()},"./node_modules/@angular/flex-layout/esm5/grid.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"GridModule",function(){return wt}),n.d(e,"ɵf2",function(){return x}),n.d(e,"ɵe2",function(){return y}),n.d(e,"ɵd2",function(){return g}),n.d(e,"ɵi2",function(){return T}),n.d(e,"ɵh2",function(){return C}),n.d(e,"ɵg2",function(){return O}),n.d(e,"ɵl2",function(){return P}),n.d(e,"ɵk2",function(){return A}),n.d(e,"ɵj2",function(){return k}),n.d(e,"ɵo2",function(){return V}),n.d(e,"ɵn2",function(){return N}),n.d(e,"ɵm2",function(){return R}),n.d(e,"ɵr2",function(){return G}),n.d(e,"ɵq2",function(){return H}),n.d(e,"ɵp2",function(){return z}),n.d(e,"ɵu2",function(){return Z}),n.d(e,"ɵt2",function(){return K}),n.d(e,"ɵs2",function(){return Y}),n.d(e,"ɵx2",function(){return rt}),n.d(e,"ɵw2",function(){return J}),n.d(e,"ɵv2",function(){return $}),n.d(e,"ɵba2",function(){return ct}),n.d(e,"ɵz2",function(){return ot}),n.d(e,"ɵy2",function(){return it}),n.d(e,"ɵc2",function(){return f}),n.d(e,"ɵb2",function(){return l}),n.d(e,"ɵa2",function(){return c}),n.d(e,"ɵbd2",function(){return ht}),n.d(e,"ɵbc2",function(){return dt}),n.d(e,"ɵbb2",function(){return lt}),n.d(e,"ɵbg2",function(){return bt}),n.d(e,"ɵbf2",function(){return gt}),n.d(e,"ɵbe2",function(){return mt});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/flex-layout/esm5/core.es5.js"),a=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),s="stretch",u="stretch",c=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t){return function(t){void 0===t&&(t="");var e={},n=t.split(" "),r=n[0],i=n[1];switch(r){case"end":e["justify-self"]="end";break;case"center":e["justify-self"]="center";break;case"stretch":e["justify-self"]="stretch";break;case"start":e["justify-self"]="start";break;default:e["justify-self"]=s}switch(i){case"end":e["align-self"]="end";break;case"center":e["align-self"]="center";break;case"stretch":e["align-self"]="stretch";break;case"start":e["align-self"]="start";break;default:e["align-self"]=u}return e}(t||s)},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),l=function(t){function e(e,n,r,i){var o=t.call(this,e,n,r,i)||this;return o.elementRef=e,o.styleBuilder=n,o.styler=r,o.marshal=i,o.DIRECTIVE_KEY="grid-align",o.styleCache=d,o.init(),o}return Object(r.__extends)(e,t),e.ctorParameters=function(){return[{type:i.ElementRef},{type:c,decorators:[{type:i.Optional}]},{type:o.StyleUtils},{type:o.MediaMarshaller}]},e}(o.BaseDirective2),d=new Map,p=["gdGridAlign","gdGridAlign.xs","gdGridAlign.sm","gdGridAlign.md","gdGridAlign.lg","gdGridAlign.xl","gdGridAlign.lt-sm","gdGridAlign.lt-md","gdGridAlign.lt-lg","gdGridAlign.lt-xl","gdGridAlign.gt-xs","gdGridAlign.gt-sm","gdGridAlign.gt-md","gdGridAlign.gt-lg"],f=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=p,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [gdGridAlign],\n  [gdGridAlign.xs], [gdGridAlign.sm], [gdGridAlign.md], [gdGridAlign.lg],[gdGridAlign.xl],\n  [gdGridAlign.lt-sm], [gdGridAlign.lt-md], [gdGridAlign.lt-lg], [gdGridAlign.lt-xl],\n  [gdGridAlign.gt-xs], [gdGridAlign.gt-sm], [gdGridAlign.gt-md], [gdGridAlign.gt-lg]\n",inputs:p}]}],e}(l);var h="start",m="stretch",g=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t,e){return function(t,e){var n={},r=t.split(" "),i=r[0],o=r[1];switch(i){case"center":n["align-content"]="center";break;case"space-around":n["align-content"]="space-around";break;case"space-between":n["align-content"]="space-between";break;case"space-evenly":n["align-content"]="space-evenly";break;case"end":n["align-content"]="end";break;case"start":n["align-content"]="start";break;case"stretch":n["align-content"]="stretch";break;default:n["align-content"]=h}switch(o){case"start":n["align-items"]="start";break;case"center":n["align-items"]="center";break;case"end":n["align-items"]="end";break;case"stretch":n["align-items"]="stretch";break;default:n["align-items"]=m}return n.display=e?"inline-grid":"grid",n}(t||h+" "+m,e.inline)},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),y=function(t){function e(e,n,r,i){var o=t.call(this,e,n,r,i)||this;return o.elementRef=e,o.styleBuilder=n,o.styler=r,o.marshal=i,o.DIRECTIVE_KEY="grid-align-columns",o._inline=!1,o.init(),o}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"inline",{get:function(){return this._inline},set:function(t){this._inline=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e.prototype.updateWithValue=function(t){this.styleCache=this.inline?v:_,this.addStyles(t,{inline:this.inline})},e.ctorParameters=function(){return[{type:i.ElementRef},{type:g,decorators:[{type:i.Optional}]},{type:o.StyleUtils},{type:o.MediaMarshaller}]},e.propDecorators={inline:[{type:i.Input,args:["gdInline"]}]},e}(o.BaseDirective2),_=new Map,v=new Map,b=["gdAlignColumns","gdAlignColumns.xs","gdAlignColumns.sm","gdAlignColumns.md","gdAlignColumns.lg","gdAlignColumns.xl","gdAlignColumns.lt-sm","gdAlignColumns.lt-md","gdAlignColumns.lt-lg","gdAlignColumns.lt-xl","gdAlignColumns.gt-xs","gdAlignColumns.gt-sm","gdAlignColumns.gt-md","gdAlignColumns.gt-lg"],x=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=b,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [gdAlignColumns],\n  [gdAlignColumns.xs], [gdAlignColumns.sm], [gdAlignColumns.md],\n  [gdAlignColumns.lg], [gdAlignColumns.xl], [gdAlignColumns.lt-sm],\n  [gdAlignColumns.lt-md], [gdAlignColumns.lt-lg], [gdAlignColumns.lt-xl],\n  [gdAlignColumns.gt-xs], [gdAlignColumns.gt-sm], [gdAlignColumns.gt-md],\n  [gdAlignColumns.gt-lg]\n",inputs:b}]}],e}(y);var w="start",j="stretch",O=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t,e){return function(t,e){var n={},r=t.split(" "),i=r[0],o=r[1];switch(i){case"center":case"space-around":case"space-between":case"space-evenly":case"end":case"start":case"stretch":n["justify-content"]=i;break;default:n["justify-content"]=w}switch(o){case"start":case"center":case"end":case"stretch":n["justify-items"]=o;break;default:n["justify-items"]=j}return n.display=e?"inline-grid":"grid",n}(t||w+" "+j,e.inline)},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),C=function(t){function e(e,n,r,i){var o=t.call(this,e,n,r,i)||this;return o.elementRef=e,o.styleBuilder=n,o.styler=r,o.marshal=i,o.DIRECTIVE_KEY="grid-align-rows",o._inline=!1,o.init(),o}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"inline",{get:function(){return this._inline},set:function(t){this._inline=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e.prototype.updateWithValue=function(t){this.styleCache=this.inline?S:E,this.addStyles(t,{inline:this.inline})},e.ctorParameters=function(){return[{type:i.ElementRef},{type:O,decorators:[{type:i.Optional}]},{type:o.StyleUtils},{type:o.MediaMarshaller}]},e.propDecorators={inline:[{type:i.Input,args:["gdInline"]}]},e}(o.BaseDirective2),E=new Map,S=new Map,M=["gdAlignRows","gdAlignRows.xs","gdAlignRows.sm","gdAlignRows.md","gdAlignRows.lg","gdAlignRows.xl","gdAlignRows.lt-sm","gdAlignRows.lt-md","gdAlignRows.lt-lg","gdAlignRows.lt-xl","gdAlignRows.gt-xs","gdAlignRows.gt-sm","gdAlignRows.gt-md","gdAlignRows.gt-lg"],T=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=M,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [gdAlignRows],\n  [gdAlignRows.xs], [gdAlignRows.sm], [gdAlignRows.md],\n  [gdAlignRows.lg], [gdAlignRows.xl], [gdAlignRows.lt-sm],\n  [gdAlignRows.lt-md], [gdAlignRows.lt-lg], [gdAlignRows.lt-xl],\n  [gdAlignRows.gt-xs], [gdAlignRows.gt-sm], [gdAlignRows.gt-md],\n  [gdAlignRows.gt-lg]\n",inputs:M}]}],e}(C);var k=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t){return{"grid-area":t||"auto"}},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),A=function(t){function e(e,n,r,i){var o=t.call(this,e,r,n,i)||this;return o.elRef=e,o.styleUtils=n,o.styleBuilder=r,o.marshal=i,o.DIRECTIVE_KEY="grid-area",o.styleCache=D,o.init(),o}return Object(r.__extends)(e,t),e.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:k,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller}]},e}(o.BaseDirective2),D=new Map,I=["gdArea","gdArea.xs","gdArea.sm","gdArea.md","gdArea.lg","gdArea.xl","gdArea.lt-sm","gdArea.lt-md","gdArea.lt-lg","gdArea.lt-xl","gdArea.gt-xs","gdArea.gt-sm","gdArea.gt-md","gdArea.gt-lg"],P=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=I,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [gdArea],\n  [gdArea.xs], [gdArea.sm], [gdArea.md], [gdArea.lg], [gdArea.xl],\n  [gdArea.lt-sm], [gdArea.lt-md], [gdArea.lt-lg], [gdArea.lt-xl],\n  [gdArea.gt-xs], [gdArea.gt-sm], [gdArea.gt-md], [gdArea.gt-lg]\n",inputs:I}]}],e}(A),R=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t,e){var n=(t||"none").split("|").map(function(t){return'"'+t.trim()+'"'});return{display:e.inline?"inline-grid":"grid","grid-template-areas":n.join(" ")}},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),N=function(t){function e(e,n,r,i){var o=t.call(this,e,r,n,i)||this;return o.elRef=e,o.styleUtils=n,o.styleBuilder=r,o.marshal=i,o.DIRECTIVE_KEY="grid-areas",o._inline=!1,o.init(),o}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"inline",{get:function(){return this._inline},set:function(t){this._inline=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e.prototype.updateWithValue=function(t){this.styleCache=this.inline?L:F,this.addStyles(t,{inline:this.inline})},e.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:R,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller}]},e.propDecorators={inline:[{type:i.Input,args:["gdInline"]}]},e}(o.BaseDirective2),F=new Map,L=new Map,B=["gdAreas","gdAreas.xs","gdAreas.sm","gdAreas.md","gdAreas.lg","gdAreas.xl","gdAreas.lt-sm","gdAreas.lt-md","gdAreas.lt-lg","gdAreas.lt-xl","gdAreas.gt-xs","gdAreas.gt-sm","gdAreas.gt-md","gdAreas.gt-lg"],V=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=B,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [gdAreas],\n  [gdAreas.xs], [gdAreas.sm], [gdAreas.md], [gdAreas.lg], [gdAreas.xl],\n  [gdAreas.lt-sm], [gdAreas.lt-md], [gdAreas.lt-lg], [gdAreas.lt-xl],\n  [gdAreas.gt-xs], [gdAreas.gt-sm], [gdAreas.gt-md], [gdAreas.gt-lg]\n",inputs:B}]}],e}(N),z=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t,e){var n=(t||"initial").split(" "),r=n[0],i=n[1];return"column"!==r&&"row"!==r&&"dense"!==r&&(r="row"),i="dense"===i&&"dense"!==r?" dense":"",{display:e.inline?"inline-grid":"grid","grid-auto-flow":r+i}},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),H=function(t){function e(e,n,r,i){var o=t.call(this,e,n,r,i)||this;return o.elementRef=e,o.styleBuilder=n,o.styler=r,o.marshal=i,o._inline=!1,o.DIRECTIVE_KEY="grid-auto",o.init(),o}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"inline",{get:function(){return this._inline},set:function(t){this._inline=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e.prototype.updateWithValue=function(t){this.styleCache=this.inline?W:U,this.addStyles(t,{inline:this.inline})},e.ctorParameters=function(){return[{type:i.ElementRef},{type:z,decorators:[{type:i.Optional}]},{type:o.StyleUtils},{type:o.MediaMarshaller}]},e.propDecorators={inline:[{type:i.Input,args:["gdInline"]}]},e}(o.BaseDirective2),U=new Map,W=new Map,q=["gdAuto","gdAuto.xs","gdAuto.sm","gdAuto.md","gdAuto.lg","gdAuto.xl","gdAuto.lt-sm","gdAuto.lt-md","gdAuto.lt-lg","gdAuto.lt-xl","gdAuto.gt-xs","gdAuto.gt-sm","gdAuto.gt-md","gdAuto.gt-lg"],G=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=q,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [gdAuto],\n  [gdAuto.xs], [gdAuto.sm], [gdAuto.md], [gdAuto.lg], [gdAuto.xl],\n  [gdAuto.lt-sm], [gdAuto.lt-md], [gdAuto.lt-lg], [gdAuto.lt-xl],\n  [gdAuto.gt-xs], [gdAuto.gt-sm], [gdAuto.gt-md], [gdAuto.gt-lg]\n",inputs:q}]}],e}(H),Y=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t){return{"grid-column":t||"auto"}},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),K=function(t){function e(e,n,r,i){var o=t.call(this,e,n,r,i)||this;return o.elementRef=e,o.styleBuilder=n,o.styler=r,o.marshal=i,o.DIRECTIVE_KEY="grid-column",o.styleCache=X,o.init(),o}return Object(r.__extends)(e,t),e.ctorParameters=function(){return[{type:i.ElementRef},{type:Y,decorators:[{type:i.Optional}]},{type:o.StyleUtils},{type:o.MediaMarshaller}]},e}(o.BaseDirective2),X=new Map,Q=["gdColumn","gdColumn.xs","gdColumn.sm","gdColumn.md","gdColumn.lg","gdColumn.xl","gdColumn.lt-sm","gdColumn.lt-md","gdColumn.lt-lg","gdColumn.lt-xl","gdColumn.gt-xs","gdColumn.gt-sm","gdColumn.gt-md","gdColumn.gt-lg"],Z=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=Q,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [gdColumn],\n  [gdColumn.xs], [gdColumn.sm], [gdColumn.md], [gdColumn.lg], [gdColumn.xl],\n  [gdColumn.lt-sm], [gdColumn.lt-md], [gdColumn.lt-lg], [gdColumn.lt-xl],\n  [gdColumn.gt-xs], [gdColumn.gt-sm], [gdColumn.gt-md], [gdColumn.gt-lg]\n",inputs:Q}]}],e}(K),$=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t,e){var n=!1;(t=t||"none").endsWith("!")&&(t=t.substring(0,t.indexOf("!")),n=!0);var r={display:e.inline?"inline-grid":"grid","grid-auto-columns":"","grid-template-columns":""};return r[n?"grid-auto-columns":"grid-template-columns"]=t,r},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),J=function(t){function e(e,n,r,i){var o=t.call(this,e,n,r,i)||this;return o.elementRef=e,o.styleBuilder=n,o.styler=r,o.marshal=i,o.DIRECTIVE_KEY="grid-columns",o._inline=!1,o.init(),o}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"inline",{get:function(){return this._inline},set:function(t){this._inline=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e.prototype.updateWithValue=function(t){this.styleCache=this.inline?et:tt,this.addStyles(t,{inline:this.inline})},e.ctorParameters=function(){return[{type:i.ElementRef},{type:$,decorators:[{type:i.Optional}]},{type:o.StyleUtils},{type:o.MediaMarshaller}]},e.propDecorators={inline:[{type:i.Input,args:["gdInline"]}]},e}(o.BaseDirective2),tt=new Map,et=new Map,nt=["gdColumns","gdColumns.xs","gdColumns.sm","gdColumns.md","gdColumns.lg","gdColumns.xl","gdColumns.lt-sm","gdColumns.lt-md","gdColumns.lt-lg","gdColumns.lt-xl","gdColumns.gt-xs","gdColumns.gt-sm","gdColumns.gt-md","gdColumns.gt-lg"],rt=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=nt,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [gdColumns],\n  [gdColumns.xs], [gdColumns.sm], [gdColumns.md], [gdColumns.lg], [gdColumns.xl],\n  [gdColumns.lt-sm], [gdColumns.lt-md], [gdColumns.lt-lg], [gdColumns.lt-xl],\n  [gdColumns.gt-xs], [gdColumns.gt-sm], [gdColumns.gt-md], [gdColumns.gt-lg]\n",inputs:nt}]}],e}(J),it=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t,e){return{display:e.inline?"inline-grid":"grid","grid-gap":t||"0"}},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),ot=function(t){function e(e,n,r,i){var o=t.call(this,e,r,n,i)||this;return o.elRef=e,o.styleUtils=n,o.styleBuilder=r,o.marshal=i,o.DIRECTIVE_KEY="grid-gap",o._inline=!1,o.init(),o}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"inline",{get:function(){return this._inline},set:function(t){this._inline=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e.prototype.updateWithValue=function(t){this.styleCache=this.inline?st:at,this.addStyles(t,{inline:this.inline})},e.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:it,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller}]},e.propDecorators={inline:[{type:i.Input,args:["gdInline"]}]},e}(o.BaseDirective2),at=new Map,st=new Map,ut=["gdGap","gdGap.xs","gdGap.sm","gdGap.md","gdGap.lg","gdGap.xl","gdGap.lt-sm","gdGap.lt-md","gdGap.lt-lg","gdGap.lt-xl","gdGap.gt-xs","gdGap.gt-sm","gdGap.gt-md","gdGap.gt-lg"],ct=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=ut,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [gdGap],\n  [gdGap.xs], [gdGap.sm], [gdGap.md], [gdGap.lg], [gdGap.xl],\n  [gdGap.lt-sm], [gdGap.lt-md], [gdGap.lt-lg], [gdGap.lt-xl],\n  [gdGap.gt-xs], [gdGap.gt-sm], [gdGap.gt-md], [gdGap.gt-lg]\n",inputs:ut}]}],e}(ot),lt=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t){return{"grid-row":t||"auto"}},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),dt=function(t){function e(e,n,r,i){var o=t.call(this,e,n,r,i)||this;return o.elementRef=e,o.styleBuilder=n,o.styler=r,o.marshal=i,o.DIRECTIVE_KEY="grid-row",o.styleCache=pt,o.init(),o}return Object(r.__extends)(e,t),e.ctorParameters=function(){return[{type:i.ElementRef},{type:lt,decorators:[{type:i.Optional}]},{type:o.StyleUtils},{type:o.MediaMarshaller}]},e}(o.BaseDirective2),pt=new Map,ft=["gdRow","gdRow.xs","gdRow.sm","gdRow.md","gdRow.lg","gdRow.xl","gdRow.lt-sm","gdRow.lt-md","gdRow.lt-lg","gdRow.lt-xl","gdRow.gt-xs","gdRow.gt-sm","gdRow.gt-md","gdRow.gt-lg"],ht=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=ft,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [gdRow],\n  [gdRow.xs], [gdRow.sm], [gdRow.md], [gdRow.lg], [gdRow.xl],\n  [gdRow.lt-sm], [gdRow.lt-md], [gdRow.lt-lg], [gdRow.lt-xl],\n  [gdRow.gt-xs], [gdRow.gt-sm], [gdRow.gt-md], [gdRow.gt-lg]\n",inputs:ft}]}],e}(dt),mt=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.buildStyles=function(t,e){var n=!1;(t=t||"none").endsWith("!")&&(t=t.substring(0,t.indexOf("!")),n=!0);var r={display:e.inline?"inline-grid":"grid","grid-auto-rows":"","grid-template-rows":""};return r[n?"grid-auto-rows":"grid-template-rows"]=t,r},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(o.StyleBuilder),gt=function(t){function e(e,n,r,i){var o=t.call(this,e,n,r,i)||this;return o.elementRef=e,o.styleBuilder=n,o.styler=r,o.marshal=i,o.DIRECTIVE_KEY="grid-rows",o._inline=!1,o.init(),o}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"inline",{get:function(){return this._inline},set:function(t){this._inline=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e.prototype.updateWithValue=function(t){this.styleCache=this.inline?_t:yt,this.addStyles(t,{inline:this.inline})},e.ctorParameters=function(){return[{type:i.ElementRef},{type:mt,decorators:[{type:i.Optional}]},{type:o.StyleUtils},{type:o.MediaMarshaller}]},e.propDecorators={inline:[{type:i.Input,args:["gdInline"]}]},e}(o.BaseDirective2),yt=new Map,_t=new Map,vt=["gdRows","gdRows.xs","gdRows.sm","gdRows.md","gdRows.lg","gdRows.xl","gdRows.lt-sm","gdRows.lt-md","gdRows.lt-lg","gdRows.lt-xl","gdRows.gt-xs","gdRows.gt-sm","gdRows.gt-md","gdRows.gt-lg"],bt=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.inputs=vt,e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"\n  [gdRows],\n  [gdRows.xs], [gdRows.sm], [gdRows.md], [gdRows.lg], [gdRows.xl],\n  [gdRows.lt-sm], [gdRows.lt-md], [gdRows.lt-lg], [gdRows.lt-xl],\n  [gdRows.gt-xs], [gdRows.gt-sm], [gdRows.gt-md], [gdRows.gt-lg]\n",inputs:vt}]}],e}(gt),xt=[f,x,T,P,V,G,Z,rt,ct,ht,bt],wt=function(){function t(){}return t.decorators=[{type:i.NgModule,args:[{imports:[o.CoreModule],declarations:xt.slice(),exports:xt.slice()}]}],t}()},"./node_modules/@angular/forms/fesm5/forms.js":function(t,e,n){"use strict";n.r(e),n.d(e,"ɵangular_packages_forms_forms_bc",function(){return le}),n.d(e,"ɵangular_packages_forms_forms_bb",function(){return ce}),n.d(e,"ɵangular_packages_forms_forms_z",function(){return se}),n.d(e,"ɵangular_packages_forms_forms_ba",function(){return ue}),n.d(e,"ɵangular_packages_forms_forms_a",function(){return v}),n.d(e,"ɵangular_packages_forms_forms_b",function(){return x}),n.d(e,"ɵangular_packages_forms_forms_c",function(){return lt}),n.d(e,"ɵangular_packages_forms_forms_d",function(){return dt}),n.d(e,"ɵangular_packages_forms_forms_e",function(){return jt}),n.d(e,"ɵangular_packages_forms_forms_f",function(){return St}),n.d(e,"ɵangular_packages_forms_forms_g",function(){return At}),n.d(e,"ɵangular_packages_forms_forms_h",function(){return Tt}),n.d(e,"ɵangular_packages_forms_forms_bh",function(){return ae}),n.d(e,"ɵangular_packages_forms_forms_bd",function(){return E}),n.d(e,"ɵangular_packages_forms_forms_be",function(){return S}),n.d(e,"ɵangular_packages_forms_forms_i",function(){return k}),n.d(e,"ɵangular_packages_forms_forms_j",function(){return A}),n.d(e,"ɵangular_packages_forms_forms_bf",function(){return I}),n.d(e,"ɵangular_packages_forms_forms_bg",function(){return P}),n.d(e,"ɵangular_packages_forms_forms_k",function(){return Pt}),n.d(e,"ɵangular_packages_forms_forms_l",function(){return Rt}),n.d(e,"ɵangular_packages_forms_forms_m",function(){return Wt}),n.d(e,"ɵangular_packages_forms_forms_n",function(){return Ft}),n.d(e,"ɵangular_packages_forms_forms_p",function(){return zt}),n.d(e,"ɵangular_packages_forms_forms_o",function(){return Bt}),n.d(e,"ɵangular_packages_forms_forms_q",function(){return z}),n.d(e,"ɵangular_packages_forms_forms_s",function(){return K}),n.d(e,"ɵangular_packages_forms_forms_r",function(){return q}),n.d(e,"ɵangular_packages_forms_forms_u",function(){return Yt}),n.d(e,"ɵangular_packages_forms_forms_v",function(){return Qt}),n.d(e,"ɵangular_packages_forms_forms_x",function(){return te}),n.d(e,"ɵangular_packages_forms_forms_w",function(){return $t}),n.d(e,"ɵangular_packages_forms_forms_y",function(){return ne}),n.d(e,"ɵangular_packages_forms_forms_t",function(){return Gt}),n.d(e,"AbstractControlDirective",function(){return u}),n.d(e,"AbstractFormGroupDirective",function(){return ct}),n.d(e,"CheckboxControlValueAccessor",function(){return b}),n.d(e,"ControlContainer",function(){return c}),n.d(e,"NG_VALUE_ACCESSOR",function(){return _}),n.d(e,"COMPOSITION_BUFFER_MODE",function(){return w}),n.d(e,"DefaultValueAccessor",function(){return j}),n.d(e,"NgControl",function(){return T}),n.d(e,"NgControlStatus",function(){return pt}),n.d(e,"NgControlStatusGroup",function(){return ft}),n.d(e,"NgForm",function(){return Ct}),n.d(e,"NgFormSelectorWarning",function(){return Mt}),n.d(e,"NgModel",function(){return It}),n.d(e,"NgModelGroup",function(){return kt}),n.d(e,"RadioControlValueAccessor",function(){return D}),n.d(e,"FormControlDirective",function(){return Nt}),n.d(e,"FormControlName",function(){return qt}),n.d(e,"FormGroupDirective",function(){return Lt}),n.d(e,"FormArrayName",function(){return Ht}),n.d(e,"FormGroupName",function(){return Vt}),n.d(e,"NgSelectOption",function(){return W}),n.d(e,"SelectControlValueAccessor",function(){return U}),n.d(e,"SelectMultipleControlValueAccessor",function(){return Y}),n.d(e,"CheckboxRequiredValidator",function(){return Xt}),n.d(e,"EmailValidator",function(){return Zt}),n.d(e,"MaxLengthValidator",function(){return ee}),n.d(e,"MinLengthValidator",function(){return Jt}),n.d(e,"PatternValidator",function(){return re}),n.d(e,"RequiredValidator",function(){return Kt}),n.d(e,"FormBuilder",function(){return ie}),n.d(e,"AbstractControl",function(){return vt}),n.d(e,"FormArray",function(){return wt}),n.d(e,"FormControl",function(){return bt}),n.d(e,"FormGroup",function(){return xt}),n.d(e,"NG_ASYNC_VALIDATORS",function(){return p}),n.d(e,"NG_VALIDATORS",function(){return d}),n.d(e,"Validators",function(){return h}),n.d(e,"VERSION",function(){return oe}),n.d(e,"FormsModule",function(){return de}),n.d(e,"ReactiveFormsModule",function(){return pe});var r=n("./node_modules/@angular/forms/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/rxjs/_esm5/index.js"),a=n("./node_modules/rxjs/_esm5/operators/index.js"),s=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),u=function(){function t(){}return Object.defineProperty(t.prototype,"value",{get:function(){return this.control?this.control.value:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"valid",{get:function(){return this.control?this.control.valid:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"invalid",{get:function(){return this.control?this.control.invalid:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"pending",{get:function(){return this.control?this.control.pending:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this.control?this.control.disabled:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"enabled",{get:function(){return this.control?this.control.enabled:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"errors",{get:function(){return this.control?this.control.errors:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"pristine",{get:function(){return this.control?this.control.pristine:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"dirty",{get:function(){return this.control?this.control.dirty:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"touched",{get:function(){return this.control?this.control.touched:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"status",{get:function(){return this.control?this.control.status:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"untouched",{get:function(){return this.control?this.control.untouched:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"statusChanges",{get:function(){return this.control?this.control.statusChanges:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"valueChanges",{get:function(){return this.control?this.control.valueChanges:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"path",{get:function(){return null},enumerable:!0,configurable:!0}),t.prototype.reset=function(t){void 0===t&&(t=void 0),this.control&&this.control.reset(t)},t.prototype.hasError=function(t,e){return!!this.control&&this.control.hasError(t,e)},t.prototype.getError=function(t,e){return this.control?this.control.getError(t,e):null},t}(),c=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"formDirective",{get:function(){return null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"path",{get:function(){return null},enumerable:!0,configurable:!0}),e}(u);function l(t){return null==t||0===t.length}var d=new i.InjectionToken("NgValidators"),p=new i.InjectionToken("NgAsyncValidators"),f=/^(?=.{1,254}$)(?=.{1,64}@)[-!#$%&'*+/0-9=?A-Z^_`a-z{|}~]+(\.[-!#$%&'*+/0-9=?A-Z^_`a-z{|}~]+)*@[A-Za-z0-9]([A-Za-z0-9-]{0,61}[A-Za-z0-9])?(\.[A-Za-z0-9]([A-Za-z0-9-]{0,61}[A-Za-z0-9])?)*$/,h=function(){function t(){}return t.min=function(t){return function(e){if(l(e.value)||l(t))return null;var n=parseFloat(e.value);return!isNaN(n)&&n<t?{min:{min:t,actual:e.value}}:null}},t.max=function(t){return function(e){if(l(e.value)||l(t))return null;var n=parseFloat(e.value);return!isNaN(n)&&n>t?{max:{max:t,actual:e.value}}:null}},t.required=function(t){return l(t.value)?{required:!0}:null},t.requiredTrue=function(t){return!0===t.value?null:{required:!0}},t.email=function(t){return l(t.value)?null:f.test(t.value)?null:{email:!0}},t.minLength=function(t){return function(e){if(l(e.value))return null;var n=e.value?e.value.length:0;return n<t?{minlength:{requiredLength:t,actualLength:n}}:null}},t.maxLength=function(t){return function(e){var n=e.value?e.value.length:0;return n>t?{maxlength:{requiredLength:t,actualLength:n}}:null}},t.pattern=function(e){return e?("string"==typeof e?(r="","^"!==e.charAt(0)&&(r+="^"),r+=e,"$"!==e.charAt(e.length-1)&&(r+="$"),n=new RegExp(r)):(r=e.toString(),n=e),function(t){if(l(t.value))return null;var e=t.value;return n.test(e)?null:{pattern:{requiredPattern:r,actualValue:e}}}):t.nullValidator;var n,r},t.nullValidator=function(t){return null},t.compose=function(t){if(!t)return null;var e=t.filter(m);return 0==e.length?null:function(t){return y((n=t,e.map(function(t){return t(n)})));var n}},t.composeAsync=function(t){if(!t)return null;var e=t.filter(m);return 0==e.length?null:function(t){var n,r,i=(n=t,r=e,r.map(function(t){return t(n)})).map(g);return Object(o.forkJoin)(i).pipe(Object(a.map)(y))}},t}();function m(t){return null!=t}function g(t){var e=Object(i["ɵisPromise"])(t)?Object(o.from)(t):t;if(!Object(i["ɵisObservable"])(e))throw new Error("Expected validator to return Promise or Observable.");return e}function y(t){var e=t.reduce(function(t,e){return null!=e?Object(r.__assign)({},t,e):t},{});return 0===Object.keys(e).length?null:e}var _=new i.InjectionToken("NgValueAccessor"),v={provide:_,useExisting:Object(i.forwardRef)(function(){return b}),multi:!0},b=function(){function t(t,e){this._renderer=t,this._elementRef=e,this.onChange=function(t){},this.onTouched=function(){}}return t.prototype.writeValue=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"checked",t)},t.prototype.registerOnChange=function(t){this.onChange=t},t.prototype.registerOnTouched=function(t){this.onTouched=t},t.prototype.setDisabledState=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",t)},t=Object(r.__decorate)([Object(i.Directive)({selector:"input[type=checkbox][formControlName],input[type=checkbox][formControl],input[type=checkbox][ngModel]",host:{"(change)":"onChange($event.target.checked)","(blur)":"onTouched()"},providers:[v]}),Object(r.__metadata)("design:paramtypes",[i.Renderer2,i.ElementRef])],t)}(),x={provide:_,useExisting:Object(i.forwardRef)(function(){return j}),multi:!0};var w=new i.InjectionToken("CompositionEventMode"),j=function(){function t(t,e,n){var r;this._renderer=t,this._elementRef=e,this._compositionMode=n,this.onChange=function(t){},this.onTouched=function(){},this._composing=!1,null==this._compositionMode&&(this._compositionMode=(r=Object(s["ɵgetDOM"])()?Object(s["ɵgetDOM"])().getUserAgent():"",!/android (\d+)/.test(r.toLowerCase())))}return t.prototype.writeValue=function(t){var e=null==t?"":t;this._renderer.setProperty(this._elementRef.nativeElement,"value",e)},t.prototype.registerOnChange=function(t){this.onChange=t},t.prototype.registerOnTouched=function(t){this.onTouched=t},t.prototype.setDisabledState=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",t)},t.prototype._handleInput=function(t){(!this._compositionMode||this._compositionMode&&!this._composing)&&this.onChange(t)},t.prototype._compositionStart=function(){this._composing=!0},t.prototype._compositionEnd=function(t){this._composing=!1,this._compositionMode&&this.onChange(t)},t=Object(r.__decorate)([Object(i.Directive)({selector:"input:not([type=checkbox])[formControlName],textarea[formControlName],input:not([type=checkbox])[formControl],textarea[formControl],input:not([type=checkbox])[ngModel],textarea[ngModel],[ngDefaultControl]",host:{"(input)":"$any(this)._handleInput($event.target.value)","(blur)":"onTouched()","(compositionstart)":"$any(this)._compositionStart()","(compositionend)":"$any(this)._compositionEnd($event.target.value)"},providers:[x]}),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Inject)(w)),Object(r.__metadata)("design:paramtypes",[i.Renderer2,i.ElementRef,Boolean])],t)}();function O(t){return t.validate?function(e){return t.validate(e)}:t}function C(t){return t.validate?function(e){return t.validate(e)}:t}var E={provide:_,useExisting:Object(i.forwardRef)(function(){return S}),multi:!0},S=function(){function t(t,e){this._renderer=t,this._elementRef=e,this.onChange=function(t){},this.onTouched=function(){}}return t.prototype.writeValue=function(t){var e=null==t?"":t;this._renderer.setProperty(this._elementRef.nativeElement,"value",e)},t.prototype.registerOnChange=function(t){this.onChange=function(e){t(""==e?null:parseFloat(e))}},t.prototype.registerOnTouched=function(t){this.onTouched=t},t.prototype.setDisabledState=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",t)},t=Object(r.__decorate)([Object(i.Directive)({selector:"input[type=number][formControlName],input[type=number][formControl],input[type=number][ngModel]",host:{"(change)":"onChange($event.target.value)","(input)":"onChange($event.target.value)","(blur)":"onTouched()"},providers:[E]}),Object(r.__metadata)("design:paramtypes",[i.Renderer2,i.ElementRef])],t)}();function M(){throw new Error("unimplemented")}var T=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._parent=null,e.name=null,e.valueAccessor=null,e._rawValidators=[],e._rawAsyncValidators=[],e}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"validator",{get:function(){return M()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"asyncValidator",{get:function(){return M()},enumerable:!0,configurable:!0}),e}(u),k={provide:_,useExisting:Object(i.forwardRef)(function(){return D}),multi:!0},A=function(){function t(){this._accessors=[]}return t.prototype.add=function(t,e){this._accessors.push([t,e])},t.prototype.remove=function(t){for(var e=this._accessors.length-1;e>=0;--e)if(this._accessors[e][1]===t)return void this._accessors.splice(e,1)},t.prototype.select=function(t){var e=this;this._accessors.forEach(function(n){e._isSameGroup(n,t)&&n[1]!==t&&n[1].fireUncheck(t.value)})},t.prototype._isSameGroup=function(t,e){return!!t[0].control&&(t[0]._parent===e._control._parent&&t[1].name===e.name)},t=Object(r.__decorate)([Object(i.Injectable)()],t)}(),D=function(){function t(t,e,n,r){this._renderer=t,this._elementRef=e,this._registry=n,this._injector=r,this.onChange=function(){},this.onTouched=function(){}}return t.prototype.ngOnInit=function(){this._control=this._injector.get(T),this._checkName(),this._registry.add(this._control,this)},t.prototype.ngOnDestroy=function(){this._registry.remove(this)},t.prototype.writeValue=function(t){this._state=t===this.value,this._renderer.setProperty(this._elementRef.nativeElement,"checked",this._state)},t.prototype.registerOnChange=function(t){var e=this;this._fn=t,this.onChange=function(){t(e.value),e._registry.select(e)}},t.prototype.fireUncheck=function(t){this.writeValue(t)},t.prototype.registerOnTouched=function(t){this.onTouched=t},t.prototype.setDisabledState=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",t)},t.prototype._checkName=function(){this.name&&this.formControlName&&this.name!==this.formControlName&&this._throwNameError(),!this.name&&this.formControlName&&(this.name=this.formControlName)},t.prototype._throwNameError=function(){throw new Error('\n      If you define both a name and a formControlName attribute on your radio button, their values\n      must match. Ex: <input type="radio" formControlName="food" name="food">\n    ')},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",String)],t.prototype,"name",void 0),Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",String)],t.prototype,"formControlName",void 0),Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",Object)],t.prototype,"value",void 0),t=Object(r.__decorate)([Object(i.Directive)({selector:"input[type=radio][formControlName],input[type=radio][formControl],input[type=radio][ngModel]",host:{"(change)":"onChange()","(blur)":"onTouched()"},providers:[k]}),Object(r.__metadata)("design:paramtypes",[i.Renderer2,i.ElementRef,A,i.Injector])],t)}(),I={provide:_,useExisting:Object(i.forwardRef)(function(){return P}),multi:!0},P=function(){function t(t,e){this._renderer=t,this._elementRef=e,this.onChange=function(t){},this.onTouched=function(){}}return t.prototype.writeValue=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"value",parseFloat(t))},t.prototype.registerOnChange=function(t){this.onChange=function(e){t(""==e?null:parseFloat(e))}},t.prototype.registerOnTouched=function(t){this.onTouched=t},t.prototype.setDisabledState=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",t)},t=Object(r.__decorate)([Object(i.Directive)({selector:"input[type=range][formControlName],input[type=range][formControl],input[type=range][ngModel]",host:{"(change)":"onChange($event.target.value)","(input)":"onChange($event.target.value)","(blur)":"onTouched()"},providers:[I]}),Object(r.__metadata)("design:paramtypes",[i.Renderer2,i.ElementRef])],t)}(),R='\n    <div [formGroup]="myGroup">\n      <input formControlName="firstName">\n    </div>\n\n    In your class:\n\n    this.myGroup = new FormGroup({\n       firstName: new FormControl()\n    });',N='\n    <div [formGroup]="myGroup">\n       <div formGroupName="person">\n          <input formControlName="firstName">\n       </div>\n    </div>\n\n    In your class:\n\n    this.myGroup = new FormGroup({\n       person: new FormGroup({ firstName: new FormControl() })\n    });',F='\n    <div [formGroup]="myGroup">\n      <div formArrayName="cities">\n        <div *ngFor="let city of cityArray.controls; index as i">\n          <input [formControlName]="i">\n        </div>\n      </div>\n    </div>\n\n    In your class:\n\n    this.cityArray = new FormArray([new FormControl(\'SF\')]);\n    this.myGroup = new FormGroup({\n      cities: this.cityArray\n    });',L='\n    <form>\n       <div ngModelGroup="person">\n          <input [(ngModel)]="person.name" name="firstName">\n       </div>\n    </form>',B='\n    <div [formGroup]="myGroup">\n       <input formControlName="firstName">\n       <input [(ngModel)]="showMoreControls" [ngModelOptions]="{standalone: true}">\n    </div>\n  ',V=function(){function t(){}return t.controlParentException=function(){throw new Error("formControlName must be used with a parent formGroup directive.  You'll want to add a formGroup\n       directive and pass it an existing FormGroup instance (you can create one in your class).\n\n      Example:\n\n      "+R)},t.ngModelGroupException=function(){throw new Error('formControlName cannot be used with an ngModelGroup parent. It is only compatible with parents\n       that also have a "form" prefix: formGroupName, formArrayName, or formGroup.\n\n       Option 1:  Update the parent to be formGroupName (reactive form strategy)\n\n        '+N+"\n\n        Option 2: Use ngModel instead of formControlName (template-driven strategy)\n\n        "+L)},t.missingFormException=function(){throw new Error("formGroup expects a FormGroup instance. Please pass one in.\n\n       Example:\n\n       "+R)},t.groupParentException=function(){throw new Error("formGroupName must be used with a parent formGroup directive.  You'll want to add a formGroup\n      directive and pass it an existing FormGroup instance (you can create one in your class).\n\n      Example:\n\n      "+N)},t.arrayParentException=function(){throw new Error("formArrayName must be used with a parent formGroup directive.  You'll want to add a formGroup\n       directive and pass it an existing FormGroup instance (you can create one in your class).\n\n        Example:\n\n        "+F)},t.disabledAttrWarning=function(){console.warn("\n      It looks like you're using the disabled attribute with a reactive form directive. If you set disabled to true\n      when you set up this control in your component class, the disabled attribute will actually be set in the DOM for\n      you. We recommend using this approach to avoid 'changed after checked' errors.\n       \n      Example: \n      form = new FormGroup({\n        first: new FormControl({value: 'Nancy', disabled: true}, Validators.required),\n        last: new FormControl('Drew', Validators.required)\n      });\n    ")},t.ngModelWarning=function(t){console.warn("\n    It looks like you're using ngModel on the same form field as "+t+". \n    Support for using the ngModel input property and ngModelChange event with \n    reactive form directives has been deprecated in Angular v6 and will be removed \n    in Angular v7.\n    \n    For more information on this, see our API docs here:\n    https://angular.io/api/forms/"+("formControl"===t?"FormControlDirective":"FormControlName")+"#use-with-ngmodel\n    ")},t}(),z={provide:_,useExisting:Object(i.forwardRef)(function(){return U}),multi:!0};function H(t,e){return null==t?""+e:(e&&"object"==typeof e&&(e="Object"),(t+": "+e).slice(0,50))}var U=function(){function t(t,e){this._renderer=t,this._elementRef=e,this._optionMap=new Map,this._idCounter=0,this.onChange=function(t){},this.onTouched=function(){},this._compareWith=i["ɵlooseIdentical"]}return Object.defineProperty(t.prototype,"compareWith",{set:function(t){if("function"!=typeof t)throw new Error("compareWith must be a function, but received "+JSON.stringify(t));this._compareWith=t},enumerable:!0,configurable:!0}),t.prototype.writeValue=function(t){this.value=t;var e=this._getOptionId(t);null==e&&this._renderer.setProperty(this._elementRef.nativeElement,"selectedIndex",-1);var n=H(e,t);this._renderer.setProperty(this._elementRef.nativeElement,"value",n)},t.prototype.registerOnChange=function(t){var e=this;this.onChange=function(n){e.value=e._getOptionValue(n),t(e.value)}},t.prototype.registerOnTouched=function(t){this.onTouched=t},t.prototype.setDisabledState=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",t)},t.prototype._registerOption=function(){return(this._idCounter++).toString()},t.prototype._getOptionId=function(t){var e,n;try{for(var i=Object(r.__values)(Array.from(this._optionMap.keys())),o=i.next();!o.done;o=i.next()){var a=o.value;if(this._compareWith(this._optionMap.get(a),t))return a}}catch(t){e={error:t}}finally{try{o&&!o.done&&(n=i.return)&&n.call(i)}finally{if(e)throw e.error}}return null},t.prototype._getOptionValue=function(t){var e=t.split(":")[0];return this._optionMap.has(e)?this._optionMap.get(e):t},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",Function),Object(r.__metadata)("design:paramtypes",[Function])],t.prototype,"compareWith",null),t=Object(r.__decorate)([Object(i.Directive)({selector:"select:not([multiple])[formControlName],select:not([multiple])[formControl],select:not([multiple])[ngModel]",host:{"(change)":"onChange($event.target.value)","(blur)":"onTouched()"},providers:[z]}),Object(r.__metadata)("design:paramtypes",[i.Renderer2,i.ElementRef])],t)}(),W=function(){function t(t,e,n){this._element=t,this._renderer=e,this._select=n,this._select&&(this.id=this._select._registerOption())}return Object.defineProperty(t.prototype,"ngValue",{set:function(t){null!=this._select&&(this._select._optionMap.set(this.id,t),this._setElementValue(H(this.id,t)),this._select.writeValue(this._select.value))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{set:function(t){this._setElementValue(t),this._select&&this._select.writeValue(this._select.value)},enumerable:!0,configurable:!0}),t.prototype._setElementValue=function(t){this._renderer.setProperty(this._element.nativeElement,"value",t)},t.prototype.ngOnDestroy=function(){this._select&&(this._select._optionMap.delete(this.id),this._select.writeValue(this._select.value))},Object(r.__decorate)([Object(i.Input)("ngValue"),Object(r.__metadata)("design:type",Object),Object(r.__metadata)("design:paramtypes",[Object])],t.prototype,"ngValue",null),Object(r.__decorate)([Object(i.Input)("value"),Object(r.__metadata)("design:type",Object),Object(r.__metadata)("design:paramtypes",[Object])],t.prototype,"value",null),t=Object(r.__decorate)([Object(i.Directive)({selector:"option"}),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Host)()),Object(r.__metadata)("design:paramtypes",[i.ElementRef,i.Renderer2,U])],t)}(),q={provide:_,useExisting:Object(i.forwardRef)(function(){return Y}),multi:!0};function G(t,e){return null==t?""+e:("string"==typeof e&&(e="'"+e+"'"),e&&"object"==typeof e&&(e="Object"),(t+": "+e).slice(0,50))}var Y=function(){function t(t,e){this._renderer=t,this._elementRef=e,this._optionMap=new Map,this._idCounter=0,this.onChange=function(t){},this.onTouched=function(){},this._compareWith=i["ɵlooseIdentical"]}return Object.defineProperty(t.prototype,"compareWith",{set:function(t){if("function"!=typeof t)throw new Error("compareWith must be a function, but received "+JSON.stringify(t));this._compareWith=t},enumerable:!0,configurable:!0}),t.prototype.writeValue=function(t){var e,n=this;if(this.value=t,Array.isArray(t)){var r=t.map(function(t){return n._getOptionId(t)});e=function(t,e){t._setSelected(r.indexOf(e.toString())>-1)}}else e=function(t,e){t._setSelected(!1)};this._optionMap.forEach(e)},t.prototype.registerOnChange=function(t){var e=this;this.onChange=function(n){var r=[];if(n.hasOwnProperty("selectedOptions"))for(var i=n.selectedOptions,o=0;o<i.length;o++){var a=i.item(o),s=e._getOptionValue(a.value);r.push(s)}else for(i=n.options,o=0;o<i.length;o++){if((a=i.item(o)).selected){s=e._getOptionValue(a.value);r.push(s)}}e.value=r,t(r)}},t.prototype.registerOnTouched=function(t){this.onTouched=t},t.prototype.setDisabledState=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",t)},t.prototype._registerOption=function(t){var e=(this._idCounter++).toString();return this._optionMap.set(e,t),e},t.prototype._getOptionId=function(t){var e,n;try{for(var i=Object(r.__values)(Array.from(this._optionMap.keys())),o=i.next();!o.done;o=i.next()){var a=o.value;if(this._compareWith(this._optionMap.get(a)._value,t))return a}}catch(t){e={error:t}}finally{try{o&&!o.done&&(n=i.return)&&n.call(i)}finally{if(e)throw e.error}}return null},t.prototype._getOptionValue=function(t){var e=t.split(":")[0];return this._optionMap.has(e)?this._optionMap.get(e)._value:t},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",Function),Object(r.__metadata)("design:paramtypes",[Function])],t.prototype,"compareWith",null),t=Object(r.__decorate)([Object(i.Directive)({selector:"select[multiple][formControlName],select[multiple][formControl],select[multiple][ngModel]",host:{"(change)":"onChange($event.target)","(blur)":"onTouched()"},providers:[q]}),Object(r.__metadata)("design:paramtypes",[i.Renderer2,i.ElementRef])],t)}(),K=function(){function t(t,e,n){this._element=t,this._renderer=e,this._select=n,this._select&&(this.id=this._select._registerOption(this))}return Object.defineProperty(t.prototype,"ngValue",{set:function(t){null!=this._select&&(this._value=t,this._setElementValue(G(this.id,t)),this._select.writeValue(this._select.value))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{set:function(t){this._select?(this._value=t,this._setElementValue(G(this.id,t)),this._select.writeValue(this._select.value)):this._setElementValue(t)},enumerable:!0,configurable:!0}),t.prototype._setElementValue=function(t){this._renderer.setProperty(this._element.nativeElement,"value",t)},t.prototype._setSelected=function(t){this._renderer.setProperty(this._element.nativeElement,"selected",t)},t.prototype.ngOnDestroy=function(){this._select&&(this._select._optionMap.delete(this.id),this._select.writeValue(this._select.value))},Object(r.__decorate)([Object(i.Input)("ngValue"),Object(r.__metadata)("design:type",Object),Object(r.__metadata)("design:paramtypes",[Object])],t.prototype,"ngValue",null),Object(r.__decorate)([Object(i.Input)("value"),Object(r.__metadata)("design:type",Object),Object(r.__metadata)("design:paramtypes",[Object])],t.prototype,"value",null),t=Object(r.__decorate)([Object(i.Directive)({selector:"option"}),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Host)()),Object(r.__metadata)("design:paramtypes",[i.ElementRef,i.Renderer2,Y])],t)}();function X(t,e){return Object(r.__spread)(e.path,[t])}function Q(t,e){var n,r,i,o,a;t||tt(e,"Cannot find control with"),e.valueAccessor||tt(e,"No value accessor for form control with"),t.validator=h.compose([t.validator,e.validator]),t.asyncValidator=h.composeAsync([t.asyncValidator,e.asyncValidator]),e.valueAccessor.writeValue(t.value),n=t,(r=e).valueAccessor.registerOnChange(function(t){n._pendingValue=t,n._pendingChange=!0,n._pendingDirty=!0,"change"===n.updateOn&&Z(n,r)}),i=e,t.registerOnChange(function(t,e){i.valueAccessor.writeValue(t),e&&i.viewToModelUpdate(t)}),o=t,(a=e).valueAccessor.registerOnTouched(function(){o._pendingTouched=!0,"blur"===o.updateOn&&o._pendingChange&&Z(o,a),"submit"!==o.updateOn&&o.markAsTouched()}),e.valueAccessor.setDisabledState&&t.registerOnDisabledChange(function(t){e.valueAccessor.setDisabledState(t)}),e._rawValidators.forEach(function(e){e.registerOnValidatorChange&&e.registerOnValidatorChange(function(){return t.updateValueAndValidity()})}),e._rawAsyncValidators.forEach(function(e){e.registerOnValidatorChange&&e.registerOnValidatorChange(function(){return t.updateValueAndValidity()})})}function Z(t,e){t._pendingDirty&&t.markAsDirty(),t.setValue(t._pendingValue,{emitModelToViewChange:!1}),e.viewToModelUpdate(t._pendingValue),t._pendingChange=!1}function $(t,e){null==t&&tt(e,"Cannot find control with"),t.validator=h.compose([t.validator,e.validator]),t.asyncValidator=h.composeAsync([t.asyncValidator,e.asyncValidator])}function J(t){return tt(t,"There is no FormControl instance attached to form control element with")}function tt(t,e){var n;throw n=t.path.length>1?"path: '"+t.path.join(" -> ")+"'":t.path[0]?"name: '"+t.path+"'":"unspecified name attribute",new Error(e+" "+n)}function et(t){return null!=t?h.compose(t.map(O)):null}function nt(t){return null!=t?h.composeAsync(t.map(C)):null}function rt(t,e){if(!t.hasOwnProperty("model"))return!1;var n=t.model;return!!n.isFirstChange()||!Object(i["ɵlooseIdentical"])(e,n.currentValue)}var it=[b,P,S,U,Y,D];function ot(t,e){t._syncPendingControls(),e.forEach(function(t){var e=t.control;"submit"===e.updateOn&&e._pendingChange&&(t.viewToModelUpdate(e._pendingValue),e._pendingChange=!1)})}function at(t,e){if(!e)return null;Array.isArray(e)||tt(t,"Value accessor was not provided as an array for form control with");var n=void 0,r=void 0,i=void 0;return e.forEach(function(e){var o;e.constructor===j?n=e:(o=e,it.some(function(t){return o.constructor===t})?(r&&tt(t,"More than one built-in value accessor matches form control with"),r=e):(i&&tt(t,"More than one custom value accessor matches form control with"),i=e))}),i||(r||(n||(tt(t,"No valid value accessor for form control with"),null)))}function st(t,e){var n=t.indexOf(e);n>-1&&t.splice(n,1)}function ut(t,e,n,r){Object(i.isDevMode)()&&"never"!==r&&((null!==r&&"once"!==r||e._ngModelWarningSentOnce)&&("always"!==r||n._ngModelWarningSent)||(V.ngModelWarning(t),e._ngModelWarningSentOnce=!0,n._ngModelWarningSent=!0))}var ct=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.ngOnInit=function(){this._checkParentType(),this.formDirective.addFormGroup(this)},e.prototype.ngOnDestroy=function(){this.formDirective&&this.formDirective.removeFormGroup(this)},Object.defineProperty(e.prototype,"control",{get:function(){return this.formDirective.getFormGroup(this)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"path",{get:function(){return X(this.name,this._parent)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"formDirective",{get:function(){return this._parent?this._parent.formDirective:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"validator",{get:function(){return et(this._validators)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"asyncValidator",{get:function(){return nt(this._asyncValidators)},enumerable:!0,configurable:!0}),e.prototype._checkParentType=function(){},e}(c),lt=function(){function t(t){this._cd=t}return Object.defineProperty(t.prototype,"ngClassUntouched",{get:function(){return!!this._cd.control&&this._cd.control.untouched},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngClassTouched",{get:function(){return!!this._cd.control&&this._cd.control.touched},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngClassPristine",{get:function(){return!!this._cd.control&&this._cd.control.pristine},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngClassDirty",{get:function(){return!!this._cd.control&&this._cd.control.dirty},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngClassValid",{get:function(){return!!this._cd.control&&this._cd.control.valid},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngClassInvalid",{get:function(){return!!this._cd.control&&this._cd.control.invalid},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngClassPending",{get:function(){return!!this._cd.control&&this._cd.control.pending},enumerable:!0,configurable:!0}),t}(),dt={"[class.ng-untouched]":"ngClassUntouched","[class.ng-touched]":"ngClassTouched","[class.ng-pristine]":"ngClassPristine","[class.ng-dirty]":"ngClassDirty","[class.ng-valid]":"ngClassValid","[class.ng-invalid]":"ngClassInvalid","[class.ng-pending]":"ngClassPending"},pt=function(t){function e(e){return t.call(this,e)||this}return Object(r.__extends)(e,t),e=Object(r.__decorate)([Object(i.Directive)({selector:"[formControlName],[ngModel],[formControl]",host:dt}),Object(r.__param)(0,Object(i.Self)()),Object(r.__metadata)("design:paramtypes",[T])],e)}(lt),ft=function(t){function e(e){return t.call(this,e)||this}return Object(r.__extends)(e,t),e=Object(r.__decorate)([Object(i.Directive)({selector:"[formGroupName],[formArrayName],[ngModelGroup],[formGroup],form:not([ngNoForm]),[ngForm]",host:dt}),Object(r.__param)(0,Object(i.Self)()),Object(r.__metadata)("design:paramtypes",[c])],e)}(lt),ht="PENDING",mt="DISABLED";function gt(t){var e=_t(t)?t.validators:t;return Array.isArray(e)?et(e):e||null}function yt(t,e){var n=_t(e)?e.asyncValidators:t;return Array.isArray(n)?nt(n):n||null}function _t(t){return null!=t&&!Array.isArray(t)&&"object"==typeof t}var vt=function(){function t(t,e){this.validator=t,this.asyncValidator=e,this._onCollectionChange=function(){},this.pristine=!0,this.touched=!1,this._onDisabledChange=[]}return Object.defineProperty(t.prototype,"parent",{get:function(){return this._parent},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"valid",{get:function(){return"VALID"===this.status},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"invalid",{get:function(){return"INVALID"===this.status},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"pending",{get:function(){return this.status==ht},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this.status===mt},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"enabled",{get:function(){return this.status!==mt},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"dirty",{get:function(){return!this.pristine},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"untouched",{get:function(){return!this.touched},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"updateOn",{get:function(){return this._updateOn?this._updateOn:this.parent?this.parent.updateOn:"change"},enumerable:!0,configurable:!0}),t.prototype.setValidators=function(t){this.validator=gt(t)},t.prototype.setAsyncValidators=function(t){this.asyncValidator=yt(t)},t.prototype.clearValidators=function(){this.validator=null},t.prototype.clearAsyncValidators=function(){this.asyncValidator=null},t.prototype.markAsTouched=function(t){void 0===t&&(t={}),this.touched=!0,this._parent&&!t.onlySelf&&this._parent.markAsTouched(t)},t.prototype.markAsUntouched=function(t){void 0===t&&(t={}),this.touched=!1,this._pendingTouched=!1,this._forEachChild(function(t){t.markAsUntouched({onlySelf:!0})}),this._parent&&!t.onlySelf&&this._parent._updateTouched(t)},t.prototype.markAsDirty=function(t){void 0===t&&(t={}),this.pristine=!1,this._parent&&!t.onlySelf&&this._parent.markAsDirty(t)},t.prototype.markAsPristine=function(t){void 0===t&&(t={}),this.pristine=!0,this._pendingDirty=!1,this._forEachChild(function(t){t.markAsPristine({onlySelf:!0})}),this._parent&&!t.onlySelf&&this._parent._updatePristine(t)},t.prototype.markAsPending=function(t){void 0===t&&(t={}),this.status=ht,!1!==t.emitEvent&&this.statusChanges.emit(this.status),this._parent&&!t.onlySelf&&this._parent.markAsPending(t)},t.prototype.disable=function(t){void 0===t&&(t={}),this.status=mt,this.errors=null,this._forEachChild(function(e){e.disable(Object(r.__assign)({},t,{onlySelf:!0}))}),this._updateValue(),!1!==t.emitEvent&&(this.valueChanges.emit(this.value),this.statusChanges.emit(this.status)),this._updateAncestors(t),this._onDisabledChange.forEach(function(t){return t(!0)})},t.prototype.enable=function(t){void 0===t&&(t={}),this.status="VALID",this._forEachChild(function(e){e.enable(Object(r.__assign)({},t,{onlySelf:!0}))}),this.updateValueAndValidity({onlySelf:!0,emitEvent:t.emitEvent}),this._updateAncestors(t),this._onDisabledChange.forEach(function(t){return t(!1)})},t.prototype._updateAncestors=function(t){this._parent&&!t.onlySelf&&(this._parent.updateValueAndValidity(t),this._parent._updatePristine(),this._parent._updateTouched())},t.prototype.setParent=function(t){this._parent=t},t.prototype.updateValueAndValidity=function(t){void 0===t&&(t={}),this._setInitialStatus(),this._updateValue(),this.enabled&&(this._cancelExistingSubscription(),this.errors=this._runValidator(),this.status=this._calculateStatus(),"VALID"!==this.status&&this.status!==ht||this._runAsyncValidator(t.emitEvent)),!1!==t.emitEvent&&(this.valueChanges.emit(this.value),this.statusChanges.emit(this.status)),this._parent&&!t.onlySelf&&this._parent.updateValueAndValidity(t)},t.prototype._updateTreeValidity=function(t){void 0===t&&(t={emitEvent:!0}),this._forEachChild(function(e){return e._updateTreeValidity(t)}),this.updateValueAndValidity({onlySelf:!0,emitEvent:t.emitEvent})},t.prototype._setInitialStatus=function(){this.status=this._allControlsDisabled()?mt:"VALID"},t.prototype._runValidator=function(){return this.validator?this.validator(this):null},t.prototype._runAsyncValidator=function(t){var e=this;if(this.asyncValidator){this.status=ht;var n=g(this.asyncValidator(this));this._asyncValidationSubscription=n.subscribe(function(n){return e.setErrors(n,{emitEvent:t})})}},t.prototype._cancelExistingSubscription=function(){this._asyncValidationSubscription&&this._asyncValidationSubscription.unsubscribe()},t.prototype.setErrors=function(t,e){void 0===e&&(e={}),this.errors=t,this._updateControlsErrors(!1!==e.emitEvent)},t.prototype.get=function(t){return e=this,r=".",null==(n=t)?null:(n instanceof Array||(n=n.split(r)),n instanceof Array&&0===n.length?null:n.reduce(function(t,e){return t instanceof xt?t.controls.hasOwnProperty(e)?t.controls[e]:null:t instanceof wt?t.at(e)||null:null},e));var e,n,r},t.prototype.getError=function(t,e){var n=e?this.get(e):this;return n&&n.errors?n.errors[t]:null},t.prototype.hasError=function(t,e){return!!this.getError(t,e)},Object.defineProperty(t.prototype,"root",{get:function(){for(var t=this;t._parent;)t=t._parent;return t},enumerable:!0,configurable:!0}),t.prototype._updateControlsErrors=function(t){this.status=this._calculateStatus(),t&&this.statusChanges.emit(this.status),this._parent&&this._parent._updateControlsErrors(t)},t.prototype._initObservables=function(){this.valueChanges=new i.EventEmitter,this.statusChanges=new i.EventEmitter},t.prototype._calculateStatus=function(){return this._allControlsDisabled()?mt:this.errors?"INVALID":this._anyControlsHaveStatus(ht)?ht:this._anyControlsHaveStatus("INVALID")?"INVALID":"VALID"},t.prototype._anyControlsHaveStatus=function(t){return this._anyControls(function(e){return e.status===t})},t.prototype._anyControlsDirty=function(){return this._anyControls(function(t){return t.dirty})},t.prototype._anyControlsTouched=function(){return this._anyControls(function(t){return t.touched})},t.prototype._updatePristine=function(t){void 0===t&&(t={}),this.pristine=!this._anyControlsDirty(),this._parent&&!t.onlySelf&&this._parent._updatePristine(t)},t.prototype._updateTouched=function(t){void 0===t&&(t={}),this.touched=this._anyControlsTouched(),this._parent&&!t.onlySelf&&this._parent._updateTouched(t)},t.prototype._isBoxedValue=function(t){return"object"==typeof t&&null!==t&&2===Object.keys(t).length&&"value"in t&&"disabled"in t},t.prototype._registerOnCollectionChange=function(t){this._onCollectionChange=t},t.prototype._setUpdateStrategy=function(t){_t(t)&&null!=t.updateOn&&(this._updateOn=t.updateOn)},t}(),bt=function(t){function e(e,n,r){void 0===e&&(e=null);var i=t.call(this,gt(n),yt(r,n))||this;return i._onChange=[],i._applyFormState(e),i._setUpdateStrategy(n),i.updateValueAndValidity({onlySelf:!0,emitEvent:!1}),i._initObservables(),i}return Object(r.__extends)(e,t),e.prototype.setValue=function(t,e){var n=this;void 0===e&&(e={}),this.value=this._pendingValue=t,this._onChange.length&&!1!==e.emitModelToViewChange&&this._onChange.forEach(function(t){return t(n.value,!1!==e.emitViewToModelChange)}),this.updateValueAndValidity(e)},e.prototype.patchValue=function(t,e){void 0===e&&(e={}),this.setValue(t,e)},e.prototype.reset=function(t,e){void 0===t&&(t=null),void 0===e&&(e={}),this._applyFormState(t),this.markAsPristine(e),this.markAsUntouched(e),this.setValue(this.value,e),this._pendingChange=!1},e.prototype._updateValue=function(){},e.prototype._anyControls=function(t){return!1},e.prototype._allControlsDisabled=function(){return this.disabled},e.prototype.registerOnChange=function(t){this._onChange.push(t)},e.prototype._clearChangeFns=function(){this._onChange=[],this._onDisabledChange=[],this._onCollectionChange=function(){}},e.prototype.registerOnDisabledChange=function(t){this._onDisabledChange.push(t)},e.prototype._forEachChild=function(t){},e.prototype._syncPendingControls=function(){return!("submit"!==this.updateOn||(this._pendingDirty&&this.markAsDirty(),this._pendingTouched&&this.markAsTouched(),!this._pendingChange))&&(this.setValue(this._pendingValue,{onlySelf:!0,emitModelToViewChange:!1}),!0)},e.prototype._applyFormState=function(t){this._isBoxedValue(t)?(this.value=this._pendingValue=t.value,t.disabled?this.disable({onlySelf:!0,emitEvent:!1}):this.enable({onlySelf:!0,emitEvent:!1})):this.value=this._pendingValue=t},e}(vt),xt=function(t){function e(e,n,r){var i=t.call(this,gt(n),yt(r,n))||this;return i.controls=e,i._initObservables(),i._setUpdateStrategy(n),i._setUpControls(),i.updateValueAndValidity({onlySelf:!0,emitEvent:!1}),i}return Object(r.__extends)(e,t),e.prototype.registerControl=function(t,e){return this.controls[t]?this.controls[t]:(this.controls[t]=e,e.setParent(this),e._registerOnCollectionChange(this._onCollectionChange),e)},e.prototype.addControl=function(t,e){this.registerControl(t,e),this.updateValueAndValidity(),this._onCollectionChange()},e.prototype.removeControl=function(t){this.controls[t]&&this.controls[t]._registerOnCollectionChange(function(){}),delete this.controls[t],this.updateValueAndValidity(),this._onCollectionChange()},e.prototype.setControl=function(t,e){this.controls[t]&&this.controls[t]._registerOnCollectionChange(function(){}),delete this.controls[t],e&&this.registerControl(t,e),this.updateValueAndValidity(),this._onCollectionChange()},e.prototype.contains=function(t){return this.controls.hasOwnProperty(t)&&this.controls[t].enabled},e.prototype.setValue=function(t,e){var n=this;void 0===e&&(e={}),this._checkAllValuesPresent(t),Object.keys(t).forEach(function(r){n._throwIfControlMissing(r),n.controls[r].setValue(t[r],{onlySelf:!0,emitEvent:e.emitEvent})}),this.updateValueAndValidity(e)},e.prototype.patchValue=function(t,e){var n=this;void 0===e&&(e={}),Object.keys(t).forEach(function(r){n.controls[r]&&n.controls[r].patchValue(t[r],{onlySelf:!0,emitEvent:e.emitEvent})}),this.updateValueAndValidity(e)},e.prototype.reset=function(t,e){void 0===t&&(t={}),void 0===e&&(e={}),this._forEachChild(function(n,r){n.reset(t[r],{onlySelf:!0,emitEvent:e.emitEvent})}),this.updateValueAndValidity(e),this._updatePristine(e),this._updateTouched(e)},e.prototype.getRawValue=function(){return this._reduceChildren({},function(t,e,n){return t[n]=e instanceof bt?e.value:e.getRawValue(),t})},e.prototype._syncPendingControls=function(){var t=this._reduceChildren(!1,function(t,e){return!!e._syncPendingControls()||t});return t&&this.updateValueAndValidity({onlySelf:!0}),t},e.prototype._throwIfControlMissing=function(t){if(!Object.keys(this.controls).length)throw new Error("\n        There are no form controls registered with this group yet.  If you're using ngModel,\n        you may want to check next tick (e.g. use setTimeout).\n      ");if(!this.controls[t])throw new Error("Cannot find form control with name: "+t+".")},e.prototype._forEachChild=function(t){var e=this;Object.keys(this.controls).forEach(function(n){return t(e.controls[n],n)})},e.prototype._setUpControls=function(){var t=this;this._forEachChild(function(e){e.setParent(t),e._registerOnCollectionChange(t._onCollectionChange)})},e.prototype._updateValue=function(){this.value=this._reduceValue()},e.prototype._anyControls=function(t){var e=this,n=!1;return this._forEachChild(function(r,i){n=n||e.contains(i)&&t(r)}),n},e.prototype._reduceValue=function(){var t=this;return this._reduceChildren({},function(e,n,r){return(n.enabled||t.disabled)&&(e[r]=n.value),e})},e.prototype._reduceChildren=function(t,e){var n=t;return this._forEachChild(function(t,r){n=e(n,t,r)}),n},e.prototype._allControlsDisabled=function(){var t,e;try{for(var n=Object(r.__values)(Object.keys(this.controls)),i=n.next();!i.done;i=n.next()){var o=i.value;if(this.controls[o].enabled)return!1}}catch(e){t={error:e}}finally{try{i&&!i.done&&(e=n.return)&&e.call(n)}finally{if(t)throw t.error}}return Object.keys(this.controls).length>0||this.disabled},e.prototype._checkAllValuesPresent=function(t){this._forEachChild(function(e,n){if(void 0===t[n])throw new Error("Must supply a value for form control with name: '"+n+"'.")})},e}(vt),wt=function(t){function e(e,n,r){var i=t.call(this,gt(n),yt(r,n))||this;return i.controls=e,i._initObservables(),i._setUpdateStrategy(n),i._setUpControls(),i.updateValueAndValidity({onlySelf:!0,emitEvent:!1}),i}return Object(r.__extends)(e,t),e.prototype.at=function(t){return this.controls[t]},e.prototype.push=function(t){this.controls.push(t),this._registerControl(t),this.updateValueAndValidity(),this._onCollectionChange()},e.prototype.insert=function(t,e){this.controls.splice(t,0,e),this._registerControl(e),this.updateValueAndValidity()},e.prototype.removeAt=function(t){this.controls[t]&&this.controls[t]._registerOnCollectionChange(function(){}),this.controls.splice(t,1),this.updateValueAndValidity()},e.prototype.setControl=function(t,e){this.controls[t]&&this.controls[t]._registerOnCollectionChange(function(){}),this.controls.splice(t,1),e&&(this.controls.splice(t,0,e),this._registerControl(e)),this.updateValueAndValidity(),this._onCollectionChange()},Object.defineProperty(e.prototype,"length",{get:function(){return this.controls.length},enumerable:!0,configurable:!0}),e.prototype.setValue=function(t,e){var n=this;void 0===e&&(e={}),this._checkAllValuesPresent(t),t.forEach(function(t,r){n._throwIfControlMissing(r),n.at(r).setValue(t,{onlySelf:!0,emitEvent:e.emitEvent})}),this.updateValueAndValidity(e)},e.prototype.patchValue=function(t,e){var n=this;void 0===e&&(e={}),t.forEach(function(t,r){n.at(r)&&n.at(r).patchValue(t,{onlySelf:!0,emitEvent:e.emitEvent})}),this.updateValueAndValidity(e)},e.prototype.reset=function(t,e){void 0===t&&(t=[]),void 0===e&&(e={}),this._forEachChild(function(n,r){n.reset(t[r],{onlySelf:!0,emitEvent:e.emitEvent})}),this.updateValueAndValidity(e),this._updatePristine(e),this._updateTouched(e)},e.prototype.getRawValue=function(){return this.controls.map(function(t){return t instanceof bt?t.value:t.getRawValue()})},e.prototype._syncPendingControls=function(){var t=this.controls.reduce(function(t,e){return!!e._syncPendingControls()||t},!1);return t&&this.updateValueAndValidity({onlySelf:!0}),t},e.prototype._throwIfControlMissing=function(t){if(!this.controls.length)throw new Error("\n        There are no form controls registered with this array yet.  If you're using ngModel,\n        you may want to check next tick (e.g. use setTimeout).\n      ");if(!this.at(t))throw new Error("Cannot find form control at index "+t)},e.prototype._forEachChild=function(t){this.controls.forEach(function(e,n){t(e,n)})},e.prototype._updateValue=function(){var t=this;this.value=this.controls.filter(function(e){return e.enabled||t.disabled}).map(function(t){return t.value})},e.prototype._anyControls=function(t){return this.controls.some(function(e){return e.enabled&&t(e)})},e.prototype._setUpControls=function(){var t=this;this._forEachChild(function(e){return t._registerControl(e)})},e.prototype._checkAllValuesPresent=function(t){this._forEachChild(function(e,n){if(void 0===t[n])throw new Error("Must supply a value for form control at index: "+n+".")})},e.prototype._allControlsDisabled=function(){var t,e;try{for(var n=Object(r.__values)(this.controls),i=n.next();!i.done;i=n.next()){if(i.value.enabled)return!1}}catch(e){t={error:e}}finally{try{i&&!i.done&&(e=n.return)&&e.call(n)}finally{if(t)throw t.error}}return this.controls.length>0||this.disabled},e.prototype._registerControl=function(t){t.setParent(this),t._registerOnCollectionChange(this._onCollectionChange)},e}(vt),jt={provide:c,useExisting:Object(i.forwardRef)(function(){return Ct})},Ot=Promise.resolve(null),Ct=function(t){function e(e,n){var r=t.call(this)||this;return r.submitted=!1,r._directives=[],r.ngSubmit=new i.EventEmitter,r.form=new xt({},et(e),nt(n)),r}return Object(r.__extends)(e,t),e.prototype.ngAfterViewInit=function(){this._setUpdateStrategy()},Object.defineProperty(e.prototype,"formDirective",{get:function(){return this},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"control",{get:function(){return this.form},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"path",{get:function(){return[]},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"controls",{get:function(){return this.form.controls},enumerable:!0,configurable:!0}),e.prototype.addControl=function(t){var e=this;Ot.then(function(){var n=e._findContainer(t.path);t.control=n.registerControl(t.name,t.control),Q(t.control,t),t.control.updateValueAndValidity({emitEvent:!1}),e._directives.push(t)})},e.prototype.getControl=function(t){return this.form.get(t.path)},e.prototype.removeControl=function(t){var e=this;Ot.then(function(){var n=e._findContainer(t.path);n&&n.removeControl(t.name),st(e._directives,t)})},e.prototype.addFormGroup=function(t){var e=this;Ot.then(function(){var n=e._findContainer(t.path),r=new xt({});$(r,t),n.registerControl(t.name,r),r.updateValueAndValidity({emitEvent:!1})})},e.prototype.removeFormGroup=function(t){var e=this;Ot.then(function(){var n=e._findContainer(t.path);n&&n.removeControl(t.name)})},e.prototype.getFormGroup=function(t){return this.form.get(t.path)},e.prototype.updateModel=function(t,e){var n=this;Ot.then(function(){n.form.get(t.path).setValue(e)})},e.prototype.setValue=function(t){this.control.setValue(t)},e.prototype.onSubmit=function(t){return this.submitted=!0,ot(this.form,this._directives),this.ngSubmit.emit(t),!1},e.prototype.onReset=function(){this.resetForm()},e.prototype.resetForm=function(t){void 0===t&&(t=void 0),this.form.reset(t),this.submitted=!1},e.prototype._setUpdateStrategy=function(){this.options&&null!=this.options.updateOn&&(this.form._updateOn=this.options.updateOn)},e.prototype._findContainer=function(t){return t.pop(),t.length?this.form.get(t):this.form},Object(r.__decorate)([Object(i.Input)("ngFormOptions"),Object(r.__metadata)("design:type",Object)],e.prototype,"options",void 0),e=Object(r.__decorate)([Object(i.Directive)({selector:"form:not([ngNoForm]):not([formGroup]),ngForm,ng-form,[ngForm]",providers:[jt],host:{"(submit)":"onSubmit($event)","(reset)":"onReset()"},outputs:["ngSubmit"],exportAs:"ngForm"}),Object(r.__param)(0,Object(i.Optional)()),Object(r.__param)(0,Object(i.Self)()),Object(r.__param)(0,Object(i.Inject)(d)),Object(r.__param)(1,Object(i.Optional)()),Object(r.__param)(1,Object(i.Self)()),Object(r.__param)(1,Object(i.Inject)(p)),Object(r.__metadata)("design:paramtypes",[Array,Array])],e)}(c),Et=function(){function t(){}return t.modelParentException=function(){throw new Error('\n      ngModel cannot be used to register form controls with a parent formGroup directive.  Try using\n      formGroup\'s partner directive "formControlName" instead.  Example:\n\n      '+R+"\n\n      Or, if you'd like to avoid registering this form control, indicate that it's standalone in ngModelOptions:\n\n      Example:\n\n      "+B)},t.formGroupNameException=function(){throw new Error("\n      ngModel cannot be used to register form controls with a parent formGroupName or formArrayName directive.\n\n      Option 1: Use formControlName instead of ngModel (reactive strategy):\n\n      "+N+"\n\n      Option 2:  Update ngModel's parent be ngModelGroup (template-driven strategy):\n\n      "+L)},t.missingNameException=function(){throw new Error('If ngModel is used within a form tag, either the name attribute must be set or the form\n      control must be defined as \'standalone\' in ngModelOptions.\n\n      Example 1: <input [(ngModel)]="person.firstName" name="first">\n      Example 2: <input [(ngModel)]="person.firstName" [ngModelOptions]="{standalone: true}">')},t.modelGroupParentException=function(){throw new Error("\n      ngModelGroup cannot be used with a parent formGroup directive.\n\n      Option 1: Use formGroupName instead of ngModelGroup (reactive strategy):\n\n      "+N+"\n\n      Option 2:  Use a regular form tag instead of the formGroup directive (template-driven strategy):\n\n      "+L)},t.ngFormWarning=function(){console.warn("\n    It looks like you're using 'ngForm'.\n\n    Support for using the 'ngForm' element selector has been deprecated in Angular v6 and will be removed\n    in Angular v9.\n\n    Use 'ng-form' instead.\n\n    Before:\n    <ngForm #myForm=\"ngForm\">\n\n    After:\n    <ng-form #myForm=\"ngForm\">\n    ")},t}(),St=new i.InjectionToken("NgFormSelectorWarning"),Mt=function(){function t(t){(t&&"once"!==t||e._ngFormWarning)&&"always"!==t||(Et.ngFormWarning(),e._ngFormWarning=!0)}var e;return e=t,t._ngFormWarning=!1,t=e=Object(r.__decorate)([Object(i.Directive)({selector:"ngForm"}),Object(r.__param)(0,Object(i.Optional)()),Object(r.__param)(0,Object(i.Inject)(St)),Object(r.__metadata)("design:paramtypes",[Object])],t)}(),Tt={provide:c,useExisting:Object(i.forwardRef)(function(){return kt})},kt=function(t){function e(e,n,r){var i=t.call(this)||this;return i._parent=e,i._validators=n,i._asyncValidators=r,i}var n;return Object(r.__extends)(e,t),n=e,e.prototype._checkParentType=function(){this._parent instanceof n||this._parent instanceof Ct||Et.modelGroupParentException()},Object(r.__decorate)([Object(i.Input)("ngModelGroup"),Object(r.__metadata)("design:type",String)],e.prototype,"name",void 0),e=n=Object(r.__decorate)([Object(i.Directive)({selector:"[ngModelGroup]",providers:[Tt],exportAs:"ngModelGroup"}),Object(r.__param)(0,Object(i.Host)()),Object(r.__param)(0,Object(i.SkipSelf)()),Object(r.__param)(1,Object(i.Optional)()),Object(r.__param)(1,Object(i.Self)()),Object(r.__param)(1,Object(i.Inject)(d)),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Self)()),Object(r.__param)(2,Object(i.Inject)(p)),Object(r.__metadata)("design:paramtypes",[c,Array,Array])],e)}(ct),At={provide:T,useExisting:Object(i.forwardRef)(function(){return It})},Dt=Promise.resolve(null),It=function(t){function e(e,n,r,o){var a=t.call(this)||this;return a.control=new bt,a._registered=!1,a.update=new i.EventEmitter,a._parent=e,a._rawValidators=n||[],a._rawAsyncValidators=r||[],a.valueAccessor=at(a,o),a}return Object(r.__extends)(e,t),e.prototype.ngOnChanges=function(t){this._checkForErrors(),this._registered||this._setUpControl(),"isDisabled"in t&&this._updateDisabled(t),rt(t,this.viewModel)&&(this._updateValue(this.model),this.viewModel=this.model)},e.prototype.ngOnDestroy=function(){this.formDirective&&this.formDirective.removeControl(this)},Object.defineProperty(e.prototype,"path",{get:function(){return this._parent?X(this.name,this._parent):[this.name]},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"formDirective",{get:function(){return this._parent?this._parent.formDirective:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"validator",{get:function(){return et(this._rawValidators)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"asyncValidator",{get:function(){return nt(this._rawAsyncValidators)},enumerable:!0,configurable:!0}),e.prototype.viewToModelUpdate=function(t){this.viewModel=t,this.update.emit(t)},e.prototype._setUpControl=function(){this._setUpdateStrategy(),this._isStandalone()?this._setUpStandalone():this.formDirective.addControl(this),this._registered=!0},e.prototype._setUpdateStrategy=function(){this.options&&null!=this.options.updateOn&&(this.control._updateOn=this.options.updateOn)},e.prototype._isStandalone=function(){return!this._parent||!(!this.options||!this.options.standalone)},e.prototype._setUpStandalone=function(){Q(this.control,this),this.control.updateValueAndValidity({emitEvent:!1})},e.prototype._checkForErrors=function(){this._isStandalone()||this._checkParentType(),this._checkName()},e.prototype._checkParentType=function(){!(this._parent instanceof kt)&&this._parent instanceof ct?Et.formGroupNameException():this._parent instanceof kt||this._parent instanceof Ct||Et.modelParentException()},e.prototype._checkName=function(){this.options&&this.options.name&&(this.name=this.options.name),this._isStandalone()||this.name||Et.missingNameException()},e.prototype._updateValue=function(t){var e=this;Dt.then(function(){e.control.setValue(t,{emitViewToModelChange:!1})})},e.prototype._updateDisabled=function(t){var e=this,n=t.isDisabled.currentValue,r=""===n||n&&"false"!==n;Dt.then(function(){r&&!e.control.disabled?e.control.disable():!r&&e.control.disabled&&e.control.enable()})},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",String)],e.prototype,"name",void 0),Object(r.__decorate)([Object(i.Input)("disabled"),Object(r.__metadata)("design:type",Boolean)],e.prototype,"isDisabled",void 0),Object(r.__decorate)([Object(i.Input)("ngModel"),Object(r.__metadata)("design:type",Object)],e.prototype,"model",void 0),Object(r.__decorate)([Object(i.Input)("ngModelOptions"),Object(r.__metadata)("design:type",Object)],e.prototype,"options",void 0),Object(r.__decorate)([Object(i.Output)("ngModelChange"),Object(r.__metadata)("design:type",Object)],e.prototype,"update",void 0),e=Object(r.__decorate)([Object(i.Directive)({selector:"[ngModel]:not([formControlName]):not([formControl])",providers:[At],exportAs:"ngModel"}),Object(r.__param)(0,Object(i.Optional)()),Object(r.__param)(0,Object(i.Host)()),Object(r.__param)(1,Object(i.Optional)()),Object(r.__param)(1,Object(i.Self)()),Object(r.__param)(1,Object(i.Inject)(d)),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Self)()),Object(r.__param)(2,Object(i.Inject)(p)),Object(r.__param)(3,Object(i.Optional)()),Object(r.__param)(3,Object(i.Self)()),Object(r.__param)(3,Object(i.Inject)(_)),Object(r.__metadata)("design:paramtypes",[c,Array,Array,Array])],e)}(T),Pt=new i.InjectionToken("NgModelWithFormControlWarning"),Rt={provide:T,useExisting:Object(i.forwardRef)(function(){return Nt})},Nt=function(t){function e(e,n,r,o){var a=t.call(this)||this;return a._ngModelWarningConfig=o,a.update=new i.EventEmitter,a._ngModelWarningSent=!1,a._rawValidators=e||[],a._rawAsyncValidators=n||[],a.valueAccessor=at(a,r),a}var n;return Object(r.__extends)(e,t),n=e,Object.defineProperty(e.prototype,"isDisabled",{set:function(t){V.disabledAttrWarning()},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){this._isControlChanged(t)&&(Q(this.form,this),this.control.disabled&&this.valueAccessor.setDisabledState&&this.valueAccessor.setDisabledState(!0),this.form.updateValueAndValidity({emitEvent:!1})),rt(t,this.viewModel)&&(ut("formControl",n,this,this._ngModelWarningConfig),this.form.setValue(this.model),this.viewModel=this.model)},Object.defineProperty(e.prototype,"path",{get:function(){return[]},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"validator",{get:function(){return et(this._rawValidators)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"asyncValidator",{get:function(){return nt(this._rawAsyncValidators)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"control",{get:function(){return this.form},enumerable:!0,configurable:!0}),e.prototype.viewToModelUpdate=function(t){this.viewModel=t,this.update.emit(t)},e.prototype._isControlChanged=function(t){return t.hasOwnProperty("form")},e._ngModelWarningSentOnce=!1,Object(r.__decorate)([Object(i.Input)("formControl"),Object(r.__metadata)("design:type",bt)],e.prototype,"form",void 0),Object(r.__decorate)([Object(i.Input)("disabled"),Object(r.__metadata)("design:type",Boolean),Object(r.__metadata)("design:paramtypes",[Boolean])],e.prototype,"isDisabled",null),Object(r.__decorate)([Object(i.Input)("ngModel"),Object(r.__metadata)("design:type",Object)],e.prototype,"model",void 0),Object(r.__decorate)([Object(i.Output)("ngModelChange"),Object(r.__metadata)("design:type",Object)],e.prototype,"update",void 0),e=n=Object(r.__decorate)([Object(i.Directive)({selector:"[formControl]",providers:[Rt],exportAs:"ngForm"}),Object(r.__param)(0,Object(i.Optional)()),Object(r.__param)(0,Object(i.Self)()),Object(r.__param)(0,Object(i.Inject)(d)),Object(r.__param)(1,Object(i.Optional)()),Object(r.__param)(1,Object(i.Self)()),Object(r.__param)(1,Object(i.Inject)(p)),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Self)()),Object(r.__param)(2,Object(i.Inject)(_)),Object(r.__param)(3,Object(i.Optional)()),Object(r.__param)(3,Object(i.Inject)(Pt)),Object(r.__metadata)("design:paramtypes",[Array,Array,Array,Object])],e)}(T),Ft={provide:c,useExisting:Object(i.forwardRef)(function(){return Lt})},Lt=function(t){function e(e,n){var r=t.call(this)||this;return r._validators=e,r._asyncValidators=n,r.submitted=!1,r.directives=[],r.form=null,r.ngSubmit=new i.EventEmitter,r}return Object(r.__extends)(e,t),e.prototype.ngOnChanges=function(t){this._checkFormPresent(),t.hasOwnProperty("form")&&(this._updateValidators(),this._updateDomValue(),this._updateRegistrations())},Object.defineProperty(e.prototype,"formDirective",{get:function(){return this},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"control",{get:function(){return this.form},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"path",{get:function(){return[]},enumerable:!0,configurable:!0}),e.prototype.addControl=function(t){var e=this.form.get(t.path);return Q(e,t),e.updateValueAndValidity({emitEvent:!1}),this.directives.push(t),e},e.prototype.getControl=function(t){return this.form.get(t.path)},e.prototype.removeControl=function(t){st(this.directives,t)},e.prototype.addFormGroup=function(t){var e=this.form.get(t.path);$(e,t),e.updateValueAndValidity({emitEvent:!1})},e.prototype.removeFormGroup=function(t){},e.prototype.getFormGroup=function(t){return this.form.get(t.path)},e.prototype.addFormArray=function(t){var e=this.form.get(t.path);$(e,t),e.updateValueAndValidity({emitEvent:!1})},e.prototype.removeFormArray=function(t){},e.prototype.getFormArray=function(t){return this.form.get(t.path)},e.prototype.updateModel=function(t,e){this.form.get(t.path).setValue(e)},e.prototype.onSubmit=function(t){return this.submitted=!0,ot(this.form,this.directives),this.ngSubmit.emit(t),!1},e.prototype.onReset=function(){this.resetForm()},e.prototype.resetForm=function(t){void 0===t&&(t=void 0),this.form.reset(t),this.submitted=!1},e.prototype._updateDomValue=function(){var t=this;this.directives.forEach(function(e){var n,r,i=t.form.get(e.path);e.control!==i&&(n=e.control,(r=e).valueAccessor.registerOnChange(function(){return J(r)}),r.valueAccessor.registerOnTouched(function(){return J(r)}),r._rawValidators.forEach(function(t){t.registerOnValidatorChange&&t.registerOnValidatorChange(null)}),r._rawAsyncValidators.forEach(function(t){t.registerOnValidatorChange&&t.registerOnValidatorChange(null)}),n&&n._clearChangeFns(),i&&Q(i,e),e.control=i)}),this.form._updateTreeValidity({emitEvent:!1})},e.prototype._updateRegistrations=function(){var t=this;this.form._registerOnCollectionChange(function(){return t._updateDomValue()}),this._oldForm&&this._oldForm._registerOnCollectionChange(function(){}),this._oldForm=this.form},e.prototype._updateValidators=function(){var t=et(this._validators);this.form.validator=h.compose([this.form.validator,t]);var e=nt(this._asyncValidators);this.form.asyncValidator=h.composeAsync([this.form.asyncValidator,e])},e.prototype._checkFormPresent=function(){this.form||V.missingFormException()},Object(r.__decorate)([Object(i.Input)("formGroup"),Object(r.__metadata)("design:type",xt)],e.prototype,"form",void 0),Object(r.__decorate)([Object(i.Output)(),Object(r.__metadata)("design:type",Object)],e.prototype,"ngSubmit",void 0),e=Object(r.__decorate)([Object(i.Directive)({selector:"[formGroup]",providers:[Ft],host:{"(submit)":"onSubmit($event)","(reset)":"onReset()"},exportAs:"ngForm"}),Object(r.__param)(0,Object(i.Optional)()),Object(r.__param)(0,Object(i.Self)()),Object(r.__param)(0,Object(i.Inject)(d)),Object(r.__param)(1,Object(i.Optional)()),Object(r.__param)(1,Object(i.Self)()),Object(r.__param)(1,Object(i.Inject)(p)),Object(r.__metadata)("design:paramtypes",[Array,Array])],e)}(c),Bt={provide:c,useExisting:Object(i.forwardRef)(function(){return Vt})},Vt=function(t){function e(e,n,r){var i=t.call(this)||this;return i._parent=e,i._validators=n,i._asyncValidators=r,i}return Object(r.__extends)(e,t),e.prototype._checkParentType=function(){Ut(this._parent)&&V.groupParentException()},Object(r.__decorate)([Object(i.Input)("formGroupName"),Object(r.__metadata)("design:type",String)],e.prototype,"name",void 0),e=Object(r.__decorate)([Object(i.Directive)({selector:"[formGroupName]",providers:[Bt]}),Object(r.__param)(0,Object(i.Optional)()),Object(r.__param)(0,Object(i.Host)()),Object(r.__param)(0,Object(i.SkipSelf)()),Object(r.__param)(1,Object(i.Optional)()),Object(r.__param)(1,Object(i.Self)()),Object(r.__param)(1,Object(i.Inject)(d)),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Self)()),Object(r.__param)(2,Object(i.Inject)(p)),Object(r.__metadata)("design:paramtypes",[c,Array,Array])],e)}(ct),zt={provide:c,useExisting:Object(i.forwardRef)(function(){return Ht})},Ht=function(t){function e(e,n,r){var i=t.call(this)||this;return i._parent=e,i._validators=n,i._asyncValidators=r,i}return Object(r.__extends)(e,t),e.prototype.ngOnInit=function(){this._checkParentType(),this.formDirective.addFormArray(this)},e.prototype.ngOnDestroy=function(){this.formDirective&&this.formDirective.removeFormArray(this)},Object.defineProperty(e.prototype,"control",{get:function(){return this.formDirective.getFormArray(this)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"formDirective",{get:function(){return this._parent?this._parent.formDirective:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"path",{get:function(){return X(this.name,this._parent)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"validator",{get:function(){return et(this._validators)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"asyncValidator",{get:function(){return nt(this._asyncValidators)},enumerable:!0,configurable:!0}),e.prototype._checkParentType=function(){Ut(this._parent)&&V.arrayParentException()},Object(r.__decorate)([Object(i.Input)("formArrayName"),Object(r.__metadata)("design:type",String)],e.prototype,"name",void 0),e=Object(r.__decorate)([Object(i.Directive)({selector:"[formArrayName]",providers:[zt]}),Object(r.__param)(0,Object(i.Optional)()),Object(r.__param)(0,Object(i.Host)()),Object(r.__param)(0,Object(i.SkipSelf)()),Object(r.__param)(1,Object(i.Optional)()),Object(r.__param)(1,Object(i.Self)()),Object(r.__param)(1,Object(i.Inject)(d)),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Self)()),Object(r.__param)(2,Object(i.Inject)(p)),Object(r.__metadata)("design:paramtypes",[c,Array,Array])],e)}(c);function Ut(t){return!(t instanceof Vt||t instanceof Lt||t instanceof Ht)}var Wt={provide:T,useExisting:Object(i.forwardRef)(function(){return qt})},qt=function(t){function e(e,n,r,o,a){var s=t.call(this)||this;return s._ngModelWarningConfig=a,s._added=!1,s.update=new i.EventEmitter,s._ngModelWarningSent=!1,s._parent=e,s._rawValidators=n||[],s._rawAsyncValidators=r||[],s.valueAccessor=at(s,o),s}var n;return Object(r.__extends)(e,t),n=e,Object.defineProperty(e.prototype,"isDisabled",{set:function(t){V.disabledAttrWarning()},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){this._added||this._setUpControl(),rt(t,this.viewModel)&&(ut("formControlName",n,this,this._ngModelWarningConfig),this.viewModel=this.model,this.formDirective.updateModel(this,this.model))},e.prototype.ngOnDestroy=function(){this.formDirective&&this.formDirective.removeControl(this)},e.prototype.viewToModelUpdate=function(t){this.viewModel=t,this.update.emit(t)},Object.defineProperty(e.prototype,"path",{get:function(){return X(this.name,this._parent)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"formDirective",{get:function(){return this._parent?this._parent.formDirective:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"validator",{get:function(){return et(this._rawValidators)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"asyncValidator",{get:function(){return nt(this._rawAsyncValidators)},enumerable:!0,configurable:!0}),e.prototype._checkParentType=function(){!(this._parent instanceof Vt)&&this._parent instanceof ct?V.ngModelGroupException():this._parent instanceof Vt||this._parent instanceof Lt||this._parent instanceof Ht||V.controlParentException()},e.prototype._setUpControl=function(){this._checkParentType(),this.control=this.formDirective.addControl(this),this.control.disabled&&this.valueAccessor.setDisabledState&&this.valueAccessor.setDisabledState(!0),this._added=!0},e._ngModelWarningSentOnce=!1,Object(r.__decorate)([Object(i.Input)("formControlName"),Object(r.__metadata)("design:type",String)],e.prototype,"name",void 0),Object(r.__decorate)([Object(i.Input)("disabled"),Object(r.__metadata)("design:type",Boolean),Object(r.__metadata)("design:paramtypes",[Boolean])],e.prototype,"isDisabled",null),Object(r.__decorate)([Object(i.Input)("ngModel"),Object(r.__metadata)("design:type",Object)],e.prototype,"model",void 0),Object(r.__decorate)([Object(i.Output)("ngModelChange"),Object(r.__metadata)("design:type",Object)],e.prototype,"update",void 0),e=n=Object(r.__decorate)([Object(i.Directive)({selector:"[formControlName]",providers:[Wt]}),Object(r.__param)(0,Object(i.Optional)()),Object(r.__param)(0,Object(i.Host)()),Object(r.__param)(0,Object(i.SkipSelf)()),Object(r.__param)(1,Object(i.Optional)()),Object(r.__param)(1,Object(i.Self)()),Object(r.__param)(1,Object(i.Inject)(d)),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Self)()),Object(r.__param)(2,Object(i.Inject)(p)),Object(r.__param)(3,Object(i.Optional)()),Object(r.__param)(3,Object(i.Self)()),Object(r.__param)(3,Object(i.Inject)(_)),Object(r.__param)(4,Object(i.Optional)()),Object(r.__param)(4,Object(i.Inject)(Pt)),Object(r.__metadata)("design:paramtypes",[c,Array,Array,Array,Object])],e)}(T),Gt={provide:d,useExisting:Object(i.forwardRef)(function(){return Kt}),multi:!0},Yt={provide:d,useExisting:Object(i.forwardRef)(function(){return Xt}),multi:!0},Kt=function(){function t(){}return Object.defineProperty(t.prototype,"required",{get:function(){return this._required},set:function(t){this._required=null!=t&&!1!==t&&""+t!="false",this._onChange&&this._onChange()},enumerable:!0,configurable:!0}),t.prototype.validate=function(t){return this.required?h.required(t):null},t.prototype.registerOnValidatorChange=function(t){this._onChange=t},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",Object),Object(r.__metadata)("design:paramtypes",[Object])],t.prototype,"required",null),t=Object(r.__decorate)([Object(i.Directive)({selector:":not([type=checkbox])[required][formControlName],:not([type=checkbox])[required][formControl],:not([type=checkbox])[required][ngModel]",providers:[Gt],host:{"[attr.required]":'required ? "" : null'}})],t)}(),Xt=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.validate=function(t){return this.required?h.requiredTrue(t):null},e=Object(r.__decorate)([Object(i.Directive)({selector:"input[type=checkbox][required][formControlName],input[type=checkbox][required][formControl],input[type=checkbox][required][ngModel]",providers:[Yt],host:{"[attr.required]":'required ? "" : null'}})],e)}(Kt),Qt={provide:d,useExisting:Object(i.forwardRef)(function(){return Zt}),multi:!0},Zt=function(){function t(){}return Object.defineProperty(t.prototype,"email",{set:function(t){this._enabled=""===t||!0===t||"true"===t,this._onChange&&this._onChange()},enumerable:!0,configurable:!0}),t.prototype.validate=function(t){return this._enabled?h.email(t):null},t.prototype.registerOnValidatorChange=function(t){this._onChange=t},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",Object),Object(r.__metadata)("design:paramtypes",[Object])],t.prototype,"email",null),t=Object(r.__decorate)([Object(i.Directive)({selector:"[email][formControlName],[email][formControl],[email][ngModel]",providers:[Qt]})],t)}(),$t={provide:d,useExisting:Object(i.forwardRef)(function(){return Jt}),multi:!0},Jt=function(){function t(){}return t.prototype.ngOnChanges=function(t){"minlength"in t&&(this._createValidator(),this._onChange&&this._onChange())},t.prototype.validate=function(t){return null==this.minlength?null:this._validator(t)},t.prototype.registerOnValidatorChange=function(t){this._onChange=t},t.prototype._createValidator=function(){this._validator=h.minLength(parseInt(this.minlength,10))},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",String)],t.prototype,"minlength",void 0),t=Object(r.__decorate)([Object(i.Directive)({selector:"[minlength][formControlName],[minlength][formControl],[minlength][ngModel]",providers:[$t],host:{"[attr.minlength]":"minlength ? minlength : null"}})],t)}(),te={provide:d,useExisting:Object(i.forwardRef)(function(){return ee}),multi:!0},ee=function(){function t(){}return t.prototype.ngOnChanges=function(t){"maxlength"in t&&(this._createValidator(),this._onChange&&this._onChange())},t.prototype.validate=function(t){return null!=this.maxlength?this._validator(t):null},t.prototype.registerOnValidatorChange=function(t){this._onChange=t},t.prototype._createValidator=function(){this._validator=h.maxLength(parseInt(this.maxlength,10))},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",String)],t.prototype,"maxlength",void 0),t=Object(r.__decorate)([Object(i.Directive)({selector:"[maxlength][formControlName],[maxlength][formControl],[maxlength][ngModel]",providers:[te],host:{"[attr.maxlength]":"maxlength ? maxlength : null"}})],t)}(),ne={provide:d,useExisting:Object(i.forwardRef)(function(){return re}),multi:!0},re=function(){function t(){}return t.prototype.ngOnChanges=function(t){"pattern"in t&&(this._createValidator(),this._onChange&&this._onChange())},t.prototype.validate=function(t){return this._validator(t)},t.prototype.registerOnValidatorChange=function(t){this._onChange=t},t.prototype._createValidator=function(){this._validator=h.pattern(this.pattern)},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",Object)],t.prototype,"pattern",void 0),t=Object(r.__decorate)([Object(i.Directive)({selector:"[pattern][formControlName],[pattern][formControl],[pattern][ngModel]",providers:[ne],host:{"[attr.pattern]":"pattern ? pattern : null"}})],t)}();var ie=function(){function t(){}return t.prototype.group=function(t,e){void 0===e&&(e=null);var n,r=this._reduceControls(t),i=null,o=null,a=void 0;return null!=e&&(void 0!==(n=e).asyncValidators||void 0!==n.validators||void 0!==n.updateOn?(i=null!=e.validators?e.validators:null,o=null!=e.asyncValidators?e.asyncValidators:null,a=null!=e.updateOn?e.updateOn:void 0):(i=null!=e.validator?e.validator:null,o=null!=e.asyncValidator?e.asyncValidator:null)),new xt(r,{asyncValidators:o,updateOn:a,validators:i})},t.prototype.control=function(t,e,n){return new bt(t,e,n)},t.prototype.array=function(t,e,n){var r=this,i=t.map(function(t){return r._createControl(t)});return new wt(i,e,n)},t.prototype._reduceControls=function(t){var e=this,n={};return Object.keys(t).forEach(function(r){n[r]=e._createControl(t[r])}),n},t.prototype._createControl=function(t){if(t instanceof bt||t instanceof xt||t instanceof wt)return t;if(Array.isArray(t)){var e=t[0],n=t.length>1?t[1]:null,r=t.length>2?t[2]:null;return this.control(e,n,r)}return this.control(t)},t=Object(r.__decorate)([Object(i.Injectable)()],t)}(),oe=new i.Version("7.2.0"),ae=function(){function t(){}return t=Object(r.__decorate)([Object(i.Directive)({selector:"form:not([ngNoForm]):not([ngNativeValidate])",host:{novalidate:""}})],t)}(),se=[ae,W,K,j,S,P,b,U,Y,D,pt,ft,Kt,Jt,ee,re,Xt,Zt],ue=[It,kt,Ct,Mt],ce=[Nt,Lt,qt,Vt,Ht],le=function(){function t(){}return t=Object(r.__decorate)([Object(i.NgModule)({declarations:se,exports:se})],t)}(),de=function(){function t(){}var e;return e=t,t.withConfig=function(t){return{ngModule:e,providers:[{provide:St,useValue:t.warnOnDeprecatedNgFormSelector}]}},t=e=Object(r.__decorate)([Object(i.NgModule)({declarations:ue,providers:[A],exports:[le,ue]})],t)}(),pe=function(){function t(){}var e;return e=t,t.withConfig=function(t){return{ngModule:e,providers:[{provide:Pt,useValue:t.warnOnNgModelWithFormControl}]}},t=e=Object(r.__decorate)([Object(i.NgModule)({declarations:[ce],providers:[ie,A],exports:[le,ce]})],t)}()},"./node_modules/@angular/forms/node_modules/tslib/tslib.es6.js":function(t,e,n){"use strict";n.r(e),n.d(e,"__extends",function(){return i}),n.d(e,"__assign",function(){return o}),n.d(e,"__rest",function(){return a}),n.d(e,"__decorate",function(){return s}),n.d(e,"__param",function(){return u}),n.d(e,"__metadata",function(){return c}),n.d(e,"__awaiter",function(){return l}),n.d(e,"__generator",function(){return d}),n.d(e,"__exportStar",function(){return p}),n.d(e,"__values",function(){return f}),n.d(e,"__read",function(){return h}),n.d(e,"__spread",function(){return m}),n.d(e,"__await",function(){return g}),n.d(e,"__asyncGenerator",function(){return y}),n.d(e,"__asyncDelegator",function(){return _}),n.d(e,"__asyncValues",function(){return v}),n.d(e,"__makeTemplateObject",function(){return b}),n.d(e,"__importStar",function(){return x}),n.d(e,"__importDefault",function(){return w});var r=function(t,e){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)};function i(t,e){function n(){this.constructor=t}r(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}var o=function(){return(o=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++){e=arguments[n];for(var i in e)Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i])}return t}).apply(this,arguments)};function a(t,e){var n={};for(var r in t)Object.prototype.hasOwnProperty.call(t,r)&&e.indexOf(r)<0&&(n[r]=t[r]);if(null!=t&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(t);i<r.length;i++)e.indexOf(r[i])<0&&(n[r[i]]=t[r[i]])}return n}function s(t,e,n,r){var i,o=arguments.length,a=o<3?e:null===r?r=Object.getOwnPropertyDescriptor(e,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(t,e,n,r);else for(var s=t.length-1;s>=0;s--)(i=t[s])&&(a=(o<3?i(a):o>3?i(e,n,a):i(e,n))||a);return o>3&&a&&Object.defineProperty(e,n,a),a}function u(t,e){return function(n,r){e(n,r,t)}}function c(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)}function l(t,e,n,r){return new(n||(n=Promise))(function(i,o){function a(t){try{u(r.next(t))}catch(t){o(t)}}function s(t){try{u(r.throw(t))}catch(t){o(t)}}function u(t){t.done?i(t.value):new n(function(e){e(t.value)}).then(a,s)}u((r=r.apply(t,e||[])).next())})}function d(t,e){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=e.call(t,a)}catch(t){o=[6,t],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(t,e){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}function f(t){var e="function"==typeof Symbol&&t[Symbol.iterator],n=0;return e?e.call(t):{next:function(){return t&&n>=t.length&&(t=void 0),{value:t&&t[n++],done:!t}}}}function h(t,e){var n="function"==typeof Symbol&&t[Symbol.iterator];if(!n)return t;var r,i,o=n.call(t),a=[];try{for(;(void 0===e||e-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(t){i={error:t}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var t=[],e=0;e<arguments.length;e++)t=t.concat(h(arguments[e]));return t}function g(t){return this instanceof g?(this.v=t,this):new g(t)}function y(t,e,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(t,e||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(t){i[t]&&(r[t]=function(e){return new Promise(function(n,r){o.push([t,e,n,r])>1||s(t,e)})})}function s(t,e){try{(n=i[t](e)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(t){l(o[0][3],t)}var n}function u(t){s("next",t)}function c(t){s("throw",t)}function l(t,e){t(e),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(t){var e,n;return e={},r("next"),r("throw",function(t){throw t}),r("return"),e[Symbol.iterator]=function(){return this},e;function r(r,i){e[r]=t[r]?function(e){return(n=!n)?{value:g(t[r](e)),done:"return"===r}:i?i(e):e}:i}}function v(t){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var e,n=t[Symbol.asyncIterator];return n?n.call(t):(t=f(t),e={},r("next"),r("throw"),r("return"),e[Symbol.asyncIterator]=function(){return this},e);function r(n){e[n]=t[n]&&function(e){return new Promise(function(r,i){var o,a,s,u;e=t[n](e),o=r,a=i,s=e.done,u=e.value,Promise.resolve(u).then(function(t){o({value:t,done:s})},a)})}}}function b(t,e){return Object.defineProperty?Object.defineProperty(t,"raw",{value:e}):t.raw=e,t}function x(t){if(t&&t.__esModule)return t;var e={};if(null!=t)for(var n in t)Object.hasOwnProperty.call(t,n)&&(e[n]=t[n]);return e.default=t,e}function w(t){return t&&t.__esModule?t:{default:t}}},"./node_modules/@angular/material/esm5/autocomplete.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MAT_AUTOCOMPLETE_DEFAULT_OPTIONS_FACTORY",function(){return j}),n.d(e,"MatAutocompleteSelectedEvent",function(){return v}),n.d(e,"MatAutocompleteBase",function(){return b}),n.d(e,"_MatAutocompleteMixinBase",function(){return x}),n.d(e,"MAT_AUTOCOMPLETE_DEFAULT_OPTIONS",function(){return w}),n.d(e,"MatAutocomplete",function(){return O}),n.d(e,"MatAutocompleteModule",function(){return P}),n.d(e,"MAT_AUTOCOMPLETE_SCROLL_STRATEGY_FACTORY",function(){return T}),n.d(e,"getMatAutocompleteMissingPanelError",function(){return D}),n.d(e,"AUTOCOMPLETE_OPTION_HEIGHT",function(){return E}),n.d(e,"AUTOCOMPLETE_PANEL_HEIGHT",function(){return S}),n.d(e,"MAT_AUTOCOMPLETE_SCROLL_STRATEGY",function(){return M}),n.d(e,"MAT_AUTOCOMPLETE_SCROLL_STRATEGY_FACTORY_PROVIDER",function(){return k}),n.d(e,"MAT_AUTOCOMPLETE_VALUE_ACCESSOR",function(){return A}),n.d(e,"MatAutocompleteTrigger",function(){return I}),n.d(e,"ɵa28",function(){return C});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),o=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),s=n("./node_modules/@angular/material/esm5/core.es5.js"),u=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),c=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),l=n("./node_modules/@angular/cdk/esm5/overlay.es5.js"),d=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),p=n("./node_modules/@angular/common/fesm5/common.js"),f=n("./node_modules/rxjs/_esm5/operators/index.js"),h=n("./node_modules/@angular/cdk/esm5/scrolling.es5.js"),m=n("./node_modules/@angular/forms/fesm5/forms.js"),g=n("./node_modules/@angular/material/esm5/form-field.es5.js"),y=n("./node_modules/rxjs/_esm5/index.js"),_=0,v=function(){return function(t,e){this.source=t,this.option=e}}(),b=function(){return function(){}}(),x=Object(s.mixinDisableRipple)(b),w=new a.InjectionToken("mat-autocomplete-default-options",{providedIn:"root",factory:j});function j(){return{autoActiveFirstOption:!1}}var O=function(t){function e(e,n,r){var i=t.call(this)||this;return i._changeDetectorRef=e,i._elementRef=n,i.showPanel=!1,i._isOpen=!1,i.displayWith=null,i.optionSelected=new a.EventEmitter,i.opened=new a.EventEmitter,i.closed=new a.EventEmitter,i._classList={},i.id="mat-autocomplete-"+_++,i._autoActiveFirstOption=!!r.autoActiveFirstOption,i}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"isOpen",{get:function(){return this._isOpen&&this.showPanel},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"autoActiveFirstOption",{get:function(){return this._autoActiveFirstOption},set:function(t){this._autoActiveFirstOption=Object(o.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"classList",{set:function(t){var e=this;t&&t.length&&(t.split(" ").forEach(function(t){return e._classList[t.trim()]=!0}),this._elementRef.nativeElement.className="")},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){this._keyManager=new i.ActiveDescendantKeyManager(this.options).withWrap(),this._setVisibility()},e.prototype._setScrollTop=function(t){this.panel&&(this.panel.nativeElement.scrollTop=t)},e.prototype._getScrollTop=function(){return this.panel?this.panel.nativeElement.scrollTop:0},e.prototype._setVisibility=function(){this.showPanel=!!this.options.length,this._classList["mat-autocomplete-visible"]=this.showPanel,this._classList["mat-autocomplete-hidden"]=!this.showPanel,this._changeDetectorRef.markForCheck()},e.prototype._emitSelectEvent=function(t){var e=new v(this,t);this.optionSelected.emit(e)},e.decorators=[{type:a.Component,args:[{selector:"mat-autocomplete",template:'<ng-template><div class="mat-autocomplete-panel" role="listbox" [id]="id" [ngClass]="_classList" #panel><ng-content></ng-content></div></ng-template>',styles:[".mat-autocomplete-panel{min-width:112px;max-width:280px;overflow:auto;-webkit-overflow-scrolling:touch;visibility:hidden;max-width:none;max-height:256px;position:relative;width:100%;border-bottom-left-radius:4px;border-bottom-right-radius:4px}.mat-autocomplete-panel.mat-autocomplete-visible{visibility:visible}.mat-autocomplete-panel.mat-autocomplete-hidden{visibility:hidden}.mat-autocomplete-panel-above .mat-autocomplete-panel{border-radius:0;border-top-left-radius:4px;border-top-right-radius:4px}.mat-autocomplete-panel .mat-divider-horizontal{margin-top:-1px}@media (-ms-high-contrast:active){.mat-autocomplete-panel{outline:solid 1px}}"],encapsulation:a.ViewEncapsulation.None,changeDetection:a.ChangeDetectionStrategy.OnPush,exportAs:"matAutocomplete",inputs:["disableRipple"],host:{class:"mat-autocomplete"},providers:[{provide:s.MAT_OPTION_PARENT_COMPONENT,useExisting:e}]}]}],e.ctorParameters=function(){return[{type:a.ChangeDetectorRef},{type:a.ElementRef},{type:void 0,decorators:[{type:a.Inject,args:[w]}]}]},e.propDecorators={template:[{type:a.ViewChild,args:[a.TemplateRef]}],panel:[{type:a.ViewChild,args:["panel"]}],options:[{type:a.ContentChildren,args:[s.MatOption,{descendants:!0}]}],optionGroups:[{type:a.ContentChildren,args:[s.MatOptgroup]}],displayWith:[{type:a.Input}],autoActiveFirstOption:[{type:a.Input}],panelWidth:[{type:a.Input}],optionSelected:[{type:a.Output}],opened:[{type:a.Output}],closed:[{type:a.Output}],classList:[{type:a.Input,args:["class"]}]},e}(x),C=function(){function t(t){this.elementRef=t}return t.decorators=[{type:a.Directive,args:[{selector:"[matAutocompleteOrigin]",exportAs:"matAutocompleteOrigin"}]}],t.ctorParameters=function(){return[{type:a.ElementRef}]},t}(),E=48,S=256,M=new a.InjectionToken("mat-autocomplete-scroll-strategy");function T(t){return function(){return t.scrollStrategies.reposition()}}var k={provide:M,deps:[l.Overlay],useFactory:T},A={provide:m.NG_VALUE_ACCESSOR,useExisting:Object(a.forwardRef)(function(){return I}),multi:!0};function D(){return Error("Attempting to open an undefined instance of `mat-autocomplete`. Make sure that the id passed to the `matAutocomplete` is correct and that you're attempting to open it after the ngAfterContentInit hook.")}var I=function(){function t(t,e,n,r,i,o,a,s,u,c){var l=this;this._element=t,this._overlay=e,this._viewContainerRef=n,this._zone=r,this._changeDetectorRef=i,this._dir=a,this._formField=s,this._document=u,this._viewportRuler=c,this._componentDestroyed=!1,this._autocompleteDisabled=!1,this._manuallyFloatingLabel=!1,this._viewportSubscription=y.Subscription.EMPTY,this._canOpenOnNextFocus=!0,this._closeKeyEventStream=new y.Subject,this._windowBlurHandler=function(){l._canOpenOnNextFocus=document.activeElement!==l._element.nativeElement||l.panelOpen},this._onChange=function(){},this._onTouched=function(){},this.autocompleteAttribute="off",this._overlayAttached=!1,this.optionSelections=Object(y.defer)(function(){return l.autocomplete&&l.autocomplete.options?y.merge.apply(void 0,l.autocomplete.options.map(function(t){return t.onSelectionChange})):l._zone.onStable.asObservable().pipe(Object(f.take)(1),Object(f.switchMap)(function(){return l.optionSelections}))}),"undefined"!=typeof window&&r.runOutsideAngular(function(){window.addEventListener("blur",l._windowBlurHandler)}),this._scrollStrategy=o}return Object.defineProperty(t.prototype,"autocompleteDisabled",{get:function(){return this._autocompleteDisabled},set:function(t){this._autocompleteDisabled=Object(o.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){"undefined"!=typeof window&&window.removeEventListener("blur",this._windowBlurHandler),this._viewportSubscription.unsubscribe(),this._componentDestroyed=!0,this._destroyPanel(),this._closeKeyEventStream.complete()},Object.defineProperty(t.prototype,"panelOpen",{get:function(){return this._overlayAttached&&this.autocomplete.showPanel},enumerable:!0,configurable:!0}),t.prototype.openPanel=function(){this._attachOverlay(),this._floatLabel()},t.prototype.closePanel=function(){this._resetLabel(),this._overlayAttached&&(this.panelOpen&&this.autocomplete.closed.emit(),this.autocomplete._isOpen=this._overlayAttached=!1,this._overlayRef&&this._overlayRef.hasAttached()&&(this._overlayRef.detach(),this._closingActionsSubscription.unsubscribe()),this._componentDestroyed||this._changeDetectorRef.detectChanges())},t.prototype.updatePosition=function(){this._overlayAttached&&this._overlayRef.updatePosition()},Object.defineProperty(t.prototype,"panelClosingActions",{get:function(){var t=this;return Object(y.merge)(this.optionSelections,this.autocomplete._keyManager.tabOut.pipe(Object(f.filter)(function(){return t._overlayAttached})),this._closeKeyEventStream,this._getOutsideClickStream(),this._overlayRef?this._overlayRef.detachments().pipe(Object(f.filter)(function(){return t._overlayAttached})):Object(y.of)()).pipe(Object(f.map)(function(t){return t instanceof s.MatOptionSelectionChange?t:null}))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"activeOption",{get:function(){return this.autocomplete&&this.autocomplete._keyManager?this.autocomplete._keyManager.activeItem:null},enumerable:!0,configurable:!0}),t.prototype._getOutsideClickStream=function(){var t=this;return this._document?Object(y.merge)(Object(y.fromEvent)(this._document,"click"),Object(y.fromEvent)(this._document,"touchend")).pipe(Object(f.filter)(function(e){var n=e.target,r=t._formField?t._formField._elementRef.nativeElement:null;return t._overlayAttached&&n!==t._element.nativeElement&&(!r||!r.contains(n))&&!!t._overlayRef&&!t._overlayRef.overlayElement.contains(n)})):Object(y.of)(null)},t.prototype.writeValue=function(t){var e=this;Promise.resolve(null).then(function(){return e._setTriggerValue(t)})},t.prototype.registerOnChange=function(t){this._onChange=t},t.prototype.registerOnTouched=function(t){this._onTouched=t},t.prototype.setDisabledState=function(t){this._element.nativeElement.disabled=t},t.prototype._handleKeydown=function(t){var e=t.keyCode;if(e===c.ESCAPE&&t.preventDefault(),this.activeOption&&e===c.ENTER&&this.panelOpen)this.activeOption._selectViaInteraction(),this._resetActiveItem(),t.preventDefault();else if(this.autocomplete){var n=this.autocomplete._keyManager.activeItem,r=e===c.UP_ARROW||e===c.DOWN_ARROW;this.panelOpen||e===c.TAB?this.autocomplete._keyManager.onKeydown(t):r&&this._canOpen()&&this.openPanel(),(r||this.autocomplete._keyManager.activeItem!==n)&&this._scrollToOption()}},t.prototype._handleInput=function(t){var e=t.target,n=e.value;"number"===e.type&&(n=""==n?null:parseFloat(n)),this._previousValue!==n&&(this._previousValue=n,this._onChange(n),this._canOpen()&&document.activeElement===t.target&&this.openPanel())},t.prototype._handleFocus=function(){this._canOpenOnNextFocus?this._canOpen()&&(this._previousValue=this._element.nativeElement.value,this._attachOverlay(),this._floatLabel(!0)):this._canOpenOnNextFocus=!0},t.prototype._floatLabel=function(t){void 0===t&&(t=!1),this._formField&&"auto"===this._formField.floatLabel&&(t?this._formField._animateAndLockLabel():this._formField.floatLabel="always",this._manuallyFloatingLabel=!0)},t.prototype._resetLabel=function(){this._manuallyFloatingLabel&&(this._formField.floatLabel="auto",this._manuallyFloatingLabel=!1)},t.prototype._scrollToOption=function(){var t=this.autocomplete._keyManager.activeItemIndex||0,e=Object(s._countGroupLabelsBeforeOption)(t,this.autocomplete.options,this.autocomplete.optionGroups),n=Object(s._getOptionScrollPosition)(t+e,E,this.autocomplete._getScrollTop(),S);this.autocomplete._setScrollTop(n)},t.prototype._subscribeToClosingActions=function(){var t=this,e=this._zone.onStable.asObservable().pipe(Object(f.take)(1)),n=this.autocomplete.options.changes.pipe(Object(f.tap)(function(){return t._positionStrategy.reapplyLastPosition()}),Object(f.delay)(0));return Object(y.merge)(e,n).pipe(Object(f.switchMap)(function(){return t._resetActiveItem(),t.autocomplete._setVisibility(),t.panelOpen&&t._overlayRef.updatePosition(),t.panelClosingActions}),Object(f.take)(1)).subscribe(function(e){return t._setValueAndClose(e)})},t.prototype._destroyPanel=function(){this._overlayRef&&(this.closePanel(),this._overlayRef.dispose(),this._overlayRef=null)},t.prototype._setTriggerValue=function(t){var e=this.autocomplete&&this.autocomplete.displayWith?this.autocomplete.displayWith(t):t,n=null!=e?e:"";this._formField?this._formField._control.value=n:this._element.nativeElement.value=n,this._previousValue=n},t.prototype._setValueAndClose=function(t){t&&t.source&&(this._clearPreviousSelectedOption(t.source),this._setTriggerValue(t.source.value),this._onChange(t.source.value),this._element.nativeElement.focus(),this.autocomplete._emitSelectEvent(t.source)),this.closePanel()},t.prototype._clearPreviousSelectedOption=function(t){this.autocomplete.options.forEach(function(e){e!=t&&e.selected&&e.deselect()})},t.prototype._attachOverlay=function(){var t=this;if(!this.autocomplete)throw D();this._overlayRef?this._overlayRef.updateSize({width:this._getPanelWidth()}):(this._portal=new d.TemplatePortal(this.autocomplete.template,this._viewContainerRef),this._overlayRef=this._overlay.create(this._getOverlayConfig()),this._overlayRef.keydownEvents().subscribe(function(e){(e.keyCode===c.ESCAPE||e.keyCode===c.UP_ARROW&&e.altKey)&&(t._resetActiveItem(),t._closeKeyEventStream.next())}),this._viewportRuler&&(this._viewportSubscription=this._viewportRuler.change().subscribe(function(){t.panelOpen&&t._overlayRef&&t._overlayRef.updateSize({width:t._getPanelWidth()})}))),this._overlayRef&&!this._overlayRef.hasAttached()&&(this._overlayRef.attach(this._portal),this._closingActionsSubscription=this._subscribeToClosingActions());var e=this.panelOpen;this.autocomplete._setVisibility(),this.autocomplete._isOpen=this._overlayAttached=!0,this.panelOpen&&e!==this.panelOpen&&this.autocomplete.opened.emit()},t.prototype._getOverlayConfig=function(){return new l.OverlayConfig({positionStrategy:this._getOverlayPosition(),scrollStrategy:this._scrollStrategy(),width:this._getPanelWidth(),direction:this._dir})},t.prototype._getOverlayPosition=function(){return this._positionStrategy=this._overlay.position().flexibleConnectedTo(this._getConnectedElement()).withFlexibleDimensions(!1).withPush(!1).withPositions([{originX:"start",originY:"bottom",overlayX:"start",overlayY:"top"},{originX:"start",originY:"top",overlayX:"start",overlayY:"bottom",panelClass:"mat-autocomplete-panel-above"}]),this._positionStrategy},t.prototype._getConnectedElement=function(){return this.connectedTo?this.connectedTo.elementRef:this._formField?this._formField.getConnectedOverlayOrigin():this._element},t.prototype._getPanelWidth=function(){return this.autocomplete.panelWidth||this._getHostWidth()},t.prototype._getHostWidth=function(){return this._getConnectedElement().nativeElement.getBoundingClientRect().width},t.prototype._resetActiveItem=function(){this.autocomplete._keyManager.setActiveItem(this.autocomplete.autoActiveFirstOption?0:-1)},t.prototype._canOpen=function(){var t=this._element.nativeElement;return!t.readOnly&&!t.disabled&&!this._autocompleteDisabled},t.decorators=[{type:a.Directive,args:[{selector:"input[matAutocomplete], textarea[matAutocomplete]",host:{"[attr.autocomplete]":"autocompleteAttribute","[attr.role]":'autocompleteDisabled ? null : "combobox"',"[attr.aria-autocomplete]":'autocompleteDisabled ? null : "list"',"[attr.aria-activedescendant]":"(panelOpen && activeOption) ? activeOption.id : null","[attr.aria-expanded]":"autocompleteDisabled ? null : panelOpen.toString()","[attr.aria-owns]":"(autocompleteDisabled || !panelOpen) ? null : autocomplete?.id","(focusin)":"_handleFocus()","(blur)":"_onTouched()","(input)":"_handleInput($event)","(keydown)":"_handleKeydown($event)"},exportAs:"matAutocompleteTrigger",providers:[A]}]}],t.ctorParameters=function(){return[{type:a.ElementRef},{type:l.Overlay},{type:a.ViewContainerRef},{type:a.NgZone},{type:a.ChangeDetectorRef},{type:void 0,decorators:[{type:a.Inject,args:[M]}]},{type:u.Directionality,decorators:[{type:a.Optional}]},{type:g.MatFormField,decorators:[{type:a.Optional},{type:a.Host}]},{type:void 0,decorators:[{type:a.Optional},{type:a.Inject,args:[p.DOCUMENT]}]},{type:h.ViewportRuler}]},t.propDecorators={autocomplete:[{type:a.Input,args:["matAutocomplete"]}],connectedTo:[{type:a.Input,args:["matAutocompleteConnectedTo"]}],autocompleteAttribute:[{type:a.Input,args:["autocomplete"]}],autocompleteDisabled:[{type:a.Input,args:["matAutocompleteDisabled"]}]},t}(),P=function(){function t(){}return t.decorators=[{type:a.NgModule,args:[{imports:[s.MatOptionModule,l.OverlayModule,s.MatCommonModule,p.CommonModule],exports:[O,s.MatOptionModule,I,C,s.MatCommonModule],declarations:[O,I,C],providers:[k]}]}],t}()},"./node_modules/@angular/material/esm5/badge.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatBadgeModule",function(){return f}),n.d(e,"MatBadgeBase",function(){return l}),n.d(e,"_MatBadgeMixinBase",function(){return d}),n.d(e,"MatBadge",function(){return p});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),o=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),a=n("./node_modules/@angular/common/fesm5/common.js"),s=n("./node_modules/@angular/core/fesm5/core.js"),u=n("./node_modules/@angular/material/esm5/core.es5.js"),c=0,l=function(){return function(){}}(),d=Object(u.mixinDisabled)(l),p=function(t){function e(e,n,r,i,o){var a=t.call(this)||this;return a._document=e,a._ngZone=n,a._elementRef=r,a._ariaDescriber=i,a._renderer=o,a._hasContent=!1,a._color="primary",a._overlap=!0,a.position="above after",a.size="medium",a._id=c++,a}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"color",{get:function(){return this._color},set:function(t){this._setColor(t),this._color=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"overlap",{get:function(){return this._overlap},set:function(t){this._overlap=Object(o.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"content",{get:function(){return this._content},set:function(t){this._content=t,this._hasContent=null!=t&&(""+t).trim().length>0,this._updateTextContent()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"description",{get:function(){return this._description},set:function(t){if(t!==this._description){var e=this._badgeElement;this._updateHostAriaDescription(t,this._description),this._description=t,e&&(t?e.setAttribute("aria-label",t):e.removeAttribute("aria-label"))}},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hidden",{get:function(){return this._hidden},set:function(t){this._hidden=Object(o.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e.prototype.isAbove=function(){return-1===this.position.indexOf("below")},e.prototype.isAfter=function(){return-1===this.position.indexOf("before")},e.prototype.ngOnDestroy=function(){var t=this._badgeElement;t&&(this.description&&this._ariaDescriber.removeDescription(t,this.description),this._renderer&&this._renderer.destroyNode&&this._renderer.destroyNode(t))},e.prototype._updateTextContent=function(){return this._badgeElement?this._badgeElement.textContent=this.content:this._badgeElement=this._createBadgeElement(),this._badgeElement},e.prototype._createBadgeElement=function(){var t=(this._renderer||this._document).createElement("span");return t.setAttribute("id","mat-badge-content-"+this._id),t.classList.add("mat-badge-content"),t.textContent=this.content,this.description&&t.setAttribute("aria-label",this.description),this._elementRef.nativeElement.appendChild(t),"function"==typeof requestAnimationFrame?this._ngZone.runOutsideAngular(function(){requestAnimationFrame(function(){t.classList.add("mat-badge-active")})}):t.classList.add("mat-badge-active"),t},e.prototype._updateHostAriaDescription=function(t,e){var n=this._updateTextContent();e&&this._ariaDescriber.removeDescription(n,e),t&&this._ariaDescriber.describe(n,t)},e.prototype._setColor=function(t){t!==this._color&&(this._color&&this._elementRef.nativeElement.classList.remove("mat-badge-"+this._color),t&&this._elementRef.nativeElement.classList.add("mat-badge-"+t))},e.decorators=[{type:s.Directive,args:[{selector:"[matBadge]",inputs:["disabled: matBadgeDisabled"],host:{class:"mat-badge","[class.mat-badge-overlap]":"overlap","[class.mat-badge-above]":"isAbove()","[class.mat-badge-below]":"!isAbove()","[class.mat-badge-before]":"!isAfter()","[class.mat-badge-after]":"isAfter()","[class.mat-badge-small]":'size === "small"',"[class.mat-badge-medium]":'size === "medium"',"[class.mat-badge-large]":'size === "large"',"[class.mat-badge-hidden]":"hidden || !_hasContent","[class.mat-badge-disabled]":"disabled"}}]}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:s.Optional},{type:s.Inject,args:[a.DOCUMENT]}]},{type:s.NgZone},{type:s.ElementRef},{type:i.AriaDescriber},{type:s.Renderer2}]},e.propDecorators={color:[{type:s.Input,args:["matBadgeColor"]}],overlap:[{type:s.Input,args:["matBadgeOverlap"]}],position:[{type:s.Input,args:["matBadgePosition"]}],content:[{type:s.Input,args:["matBadge"]}],description:[{type:s.Input,args:["matBadgeDescription"]}],size:[{type:s.Input,args:["matBadgeSize"]}],hidden:[{type:s.Input,args:["matBadgeHidden"]}]},e}(d),f=function(){function t(){}return t.decorators=[{type:s.NgModule,args:[{imports:[i.A11yModule,u.MatCommonModule],exports:[p],declarations:[p]}]}],t}()},"./node_modules/@angular/material/esm5/bottom-sheet.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatBottomSheetModule",function(){return b}),n.d(e,"MAT_BOTTOM_SHEET_DEFAULT_OPTIONS",function(){return w}),n.d(e,"MatBottomSheet",function(){return j}),n.d(e,"MAT_BOTTOM_SHEET_DATA",function(){return g}),n.d(e,"MatBottomSheetConfig",function(){return y}),n.d(e,"MatBottomSheetContainer",function(){return v}),n.d(e,"matBottomSheetAnimations",function(){return _}),n.d(e,"MatBottomSheetRef",function(){return x});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/animations/fesm5/animations.js"),o=n("./node_modules/@angular/material/esm5/core.es5.js"),a=n("./node_modules/tslib/tslib.es6.js"),s=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),u=n("./node_modules/@angular/cdk/esm5/layout.es5.js"),c=n("./node_modules/@angular/common/fesm5/common.js"),l=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),d=n("./node_modules/@angular/cdk/esm5/overlay.es5.js"),p=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),f=n("./node_modules/rxjs/_esm5/index.js"),h=n("./node_modules/rxjs/_esm5/operators/index.js"),m=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),g=new r.InjectionToken("MatBottomSheetData"),y=function(){return function(){this.data=null,this.hasBackdrop=!0,this.disableClose=!1,this.ariaLabel=null,this.closeOnNavigation=!0,this.autoFocus=!1,this.restoreFocus=!0}}(),_={bottomSheetState:Object(i.trigger)("state",[Object(i.state)("void, hidden",Object(i.style)({transform:"translateY(100%)"})),Object(i.state)("visible",Object(i.style)({transform:"translateY(0%)"})),Object(i.transition)("visible => void, visible => hidden",Object(i.animate)(o.AnimationDurations.COMPLEX+" "+o.AnimationCurves.ACCELERATION_CURVE)),Object(i.transition)("void => visible",Object(i.animate)(o.AnimationDurations.EXITING+" "+o.AnimationCurves.DECELERATION_CURVE))])},v=function(t){function e(e,n,i,o,a,s){var c=t.call(this)||this;return c._elementRef=e,c._changeDetectorRef=n,c._focusTrapFactory=i,c.bottomSheetConfig=s,c._animationState="void",c._animationStateChanged=new r.EventEmitter,c._elementFocusedBeforeOpened=null,c._document=a,c._breakpointSubscription=o.observe([u.Breakpoints.Medium,u.Breakpoints.Large,u.Breakpoints.XLarge]).subscribe(function(){c._toggleClass("mat-bottom-sheet-container-medium",o.isMatched(u.Breakpoints.Medium)),c._toggleClass("mat-bottom-sheet-container-large",o.isMatched(u.Breakpoints.Large)),c._toggleClass("mat-bottom-sheet-container-xlarge",o.isMatched(u.Breakpoints.XLarge))}),c}return Object(a.__extends)(e,t),e.prototype.attachComponentPortal=function(t){return this._validatePortalAttached(),this._setPanelClass(),this._savePreviouslyFocusedElement(),this._portalOutlet.attachComponentPortal(t)},e.prototype.attachTemplatePortal=function(t){return this._validatePortalAttached(),this._setPanelClass(),this._savePreviouslyFocusedElement(),this._portalOutlet.attachTemplatePortal(t)},e.prototype.enter=function(){this._destroyed||(this._animationState="visible",this._changeDetectorRef.detectChanges())},e.prototype.exit=function(){this._destroyed||(this._animationState="hidden",this._changeDetectorRef.markForCheck())},e.prototype.ngOnDestroy=function(){this._breakpointSubscription.unsubscribe(),this._destroyed=!0},e.prototype._onAnimationDone=function(t){"hidden"===t.toState?this._restoreFocus():"visible"===t.toState&&this.bottomSheetConfig.autoFocus&&this._trapFocus(),this._animationStateChanged.emit(t)},e.prototype._onAnimationStart=function(t){this._animationStateChanged.emit(t)},e.prototype._toggleClass=function(t,e){var n=this._elementRef.nativeElement.classList;e?n.add(t):n.remove(t)},e.prototype._validatePortalAttached=function(){if(this._portalOutlet.hasAttached())throw Error("Attempting to attach bottom sheet content after content is already attached")},e.prototype._setPanelClass=function(){var t=this._elementRef.nativeElement,e=this.bottomSheetConfig.panelClass;Array.isArray(e)?e.forEach(function(e){return t.classList.add(e)}):e&&t.classList.add(e)},e.prototype._trapFocus=function(){this._focusTrap||(this._focusTrap=this._focusTrapFactory.create(this._elementRef.nativeElement)),this._focusTrap.focusInitialElementWhenReady()},e.prototype._restoreFocus=function(){var t=this._elementFocusedBeforeOpened;this.bottomSheetConfig.restoreFocus&&t&&"function"==typeof t.focus&&t.focus(),this._focusTrap&&this._focusTrap.destroy()},e.prototype._savePreviouslyFocusedElement=function(){var t=this;this._elementFocusedBeforeOpened=this._document.activeElement,this._elementRef.nativeElement.focus&&Promise.resolve().then(function(){return t._elementRef.nativeElement.focus()})},e.decorators=[{type:r.Component,args:[{selector:"mat-bottom-sheet-container",template:"<ng-template cdkPortalOutlet></ng-template>",styles:[".mat-bottom-sheet-container{padding:8px 16px;min-width:100vw;box-sizing:border-box;display:block;outline:0;max-height:80vh;overflow:auto}@media (-ms-high-contrast:active){.mat-bottom-sheet-container{outline:1px solid}}.mat-bottom-sheet-container-large,.mat-bottom-sheet-container-medium,.mat-bottom-sheet-container-xlarge{border-top-left-radius:4px;border-top-right-radius:4px}.mat-bottom-sheet-container-medium{min-width:384px;max-width:calc(100vw - 128px)}.mat-bottom-sheet-container-large{min-width:512px;max-width:calc(100vw - 256px)}.mat-bottom-sheet-container-xlarge{min-width:576px;max-width:calc(100vw - 384px)}"],changeDetection:r.ChangeDetectionStrategy.OnPush,encapsulation:r.ViewEncapsulation.None,animations:[_.bottomSheetState],host:{class:"mat-bottom-sheet-container",tabindex:"-1",role:"dialog","aria-modal":"true","[attr.aria-label]":"bottomSheetConfig?.ariaLabel","[@state]":"_animationState","(@state.start)":"_onAnimationStart($event)","(@state.done)":"_onAnimationDone($event)"}}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:r.ChangeDetectorRef},{type:l.FocusTrapFactory},{type:u.BreakpointObserver},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[c.DOCUMENT]}]},{type:y}]},e.propDecorators={_portalOutlet:[{type:r.ViewChild,args:[s.CdkPortalOutlet]}]},e}(s.BasePortalOutlet),b=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[c.CommonModule,d.OverlayModule,o.MatCommonModule,s.PortalModule],exports:[v,o.MatCommonModule],declarations:[v],entryComponents:[v]}]}],t}(),x=function(){function t(t,e,n){var r=this;this._overlayRef=e,this._afterDismissed=new f.Subject,this._afterOpened=new f.Subject,this.containerInstance=t,t._animationStateChanged.pipe(Object(h.filter)(function(t){return"done"===t.phaseName&&"visible"===t.toState}),Object(h.take)(1)).subscribe(function(){r._afterOpened.next(),r._afterOpened.complete()}),t._animationStateChanged.pipe(Object(h.filter)(function(t){return"done"===t.phaseName&&"hidden"===t.toState}),Object(h.take)(1)).subscribe(function(){r._overlayRef.dispose(),r._afterDismissed.next(r._result),r._afterDismissed.complete()}),t.bottomSheetConfig.disableClose||Object(f.merge)(e.backdropClick(),e.keydownEvents().pipe(Object(h.filter)(function(t){return t.keyCode===p.ESCAPE}))).subscribe(function(){return r.dismiss()})}return t.prototype.dismiss=function(t){var e=this;this._afterDismissed.closed||(this.containerInstance._animationStateChanged.pipe(Object(h.filter)(function(t){return"start"===t.phaseName}),Object(h.take)(1)).subscribe(function(){return e._overlayRef.detachBackdrop()}),this._result=t,this.containerInstance.exit())},t.prototype.afterDismissed=function(){return this._afterDismissed.asObservable()},t.prototype.afterOpened=function(){return this._afterOpened.asObservable()},t.prototype.backdropClick=function(){return this._overlayRef.backdropClick()},t.prototype.keydownEvents=function(){return this._overlayRef.keydownEvents()},t}(),w=new r.InjectionToken("mat-bottom-sheet-default-options"),j=function(){function t(t,e,n,r,i){this._overlay=t,this._injector=e,this._parentBottomSheet=n,this._location=r,this._defaultOptions=i,this._bottomSheetRefAtThisLevel=null}return Object.defineProperty(t.prototype,"_openedBottomSheetRef",{get:function(){var t=this._parentBottomSheet;return t?t._openedBottomSheetRef:this._bottomSheetRefAtThisLevel},set:function(t){this._parentBottomSheet?this._parentBottomSheet._openedBottomSheetRef=t:this._bottomSheetRefAtThisLevel=t},enumerable:!0,configurable:!0}),t.prototype.open=function(t,e){var n,i,o=this,u=(n=this._defaultOptions||new y,i=e,Object(a.__assign)({},n,i)),c=this._createOverlay(u),l=this._attachContainer(c,u),d=new x(l,c,this._location);if(t instanceof r.TemplateRef)l.attachTemplatePortal(new s.TemplatePortal(t,null,{$implicit:u.data,bottomSheetRef:d}));else{var p=new s.ComponentPortal(t,void 0,this._createInjector(u,d)),f=l.attachComponentPortal(p);d.instance=f.instance}return d.afterDismissed().subscribe(function(){o._openedBottomSheetRef==d&&(o._openedBottomSheetRef=null)}),this._openedBottomSheetRef?(this._openedBottomSheetRef.afterDismissed().subscribe(function(){return d.containerInstance.enter()}),this._openedBottomSheetRef.dismiss()):d.containerInstance.enter(),this._openedBottomSheetRef=d,d},t.prototype.dismiss=function(){this._openedBottomSheetRef&&this._openedBottomSheetRef.dismiss()},t.prototype.ngOnDestroy=function(){this._bottomSheetRefAtThisLevel&&this._bottomSheetRefAtThisLevel.dismiss()},t.prototype._attachContainer=function(t,e){var n=e&&e.viewContainerRef&&e.viewContainerRef.injector,r=new s.PortalInjector(n||this._injector,new WeakMap([[y,e]])),i=new s.ComponentPortal(v,e.viewContainerRef,r);return t.attach(i).instance},t.prototype._createOverlay=function(t){var e=new d.OverlayConfig({direction:t.direction,hasBackdrop:t.hasBackdrop,disposeOnNavigation:t.closeOnNavigation,maxWidth:"100%",scrollStrategy:this._overlay.scrollStrategies.block(),positionStrategy:this._overlay.position().global().centerHorizontally().bottom("0")});return t.backdropClass&&(e.backdropClass=t.backdropClass),this._overlay.create(e)},t.prototype._createInjector=function(t,e){var n=t&&t.viewContainerRef&&t.viewContainerRef.injector,r=new WeakMap([[x,e],[g,t.data]]);return!t.direction||n&&n.get(m.Directionality,null)||r.set(m.Directionality,{value:t.direction,change:Object(f.of)()}),new s.PortalInjector(n||this._injector,r)},t.decorators=[{type:r.Injectable,args:[{providedIn:b}]}],t.ctorParameters=function(){return[{type:d.Overlay},{type:r.Injector},{type:t,decorators:[{type:r.Optional},{type:r.SkipSelf}]},{type:c.Location,decorators:[{type:r.Optional}]},{type:y,decorators:[{type:r.Optional},{type:r.Inject,args:[w]}]}]},t.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new t(Object(r.inject)(d.Overlay),Object(r.inject)(r.INJECTOR),Object(r.inject)(t,12),Object(r.inject)(c.Location,8),Object(r.inject)(w,8))},token:t,providedIn:b}),t}()},"./node_modules/@angular/material/esm5/button-toggle.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MAT_BUTTON_TOGGLE_DEFAULT_OPTIONS",function(){return l}),n.d(e,"MAT_BUTTON_TOGGLE_GROUP_VALUE_ACCESSOR",function(){return d}),n.d(e,"MatButtonToggleGroupMultiple",function(){return p}),n.d(e,"MatButtonToggleChange",function(){return h}),n.d(e,"MatButtonToggleGroup",function(){return m}),n.d(e,"MatButtonToggleBase",function(){return g}),n.d(e,"_MatButtonToggleMixinBase",function(){return y}),n.d(e,"MatButtonToggle",function(){return _}),n.d(e,"MatButtonToggleModule",function(){return v});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),o=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),a=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),s=n("./node_modules/@angular/core/fesm5/core.js"),u=n("./node_modules/@angular/forms/fesm5/forms.js"),c=n("./node_modules/@angular/material/esm5/core.es5.js"),l=new s.InjectionToken("MAT_BUTTON_TOGGLE_DEFAULT_OPTIONS"),d={provide:u.NG_VALUE_ACCESSOR,useExisting:Object(s.forwardRef)(function(){return m}),multi:!0},p=function(){return function(){}}(),f=0,h=function(){return function(t,e){this.source=t,this.value=e}}(),m=function(){function t(t,e){this._changeDetector=t,this._vertical=!1,this._multiple=!1,this._disabled=!1,this._controlValueAccessorChangeFn=function(){},this._onTouched=function(){},this._name="mat-button-toggle-group-"+f++,this.valueChange=new s.EventEmitter,this.change=new s.EventEmitter,this.appearance=e&&e.appearance?e.appearance:"standard"}return Object.defineProperty(t.prototype,"name",{get:function(){return this._name},set:function(t){var e=this;this._name=t,this._buttonToggles&&this._buttonToggles.forEach(function(t){return t.name=e._name})},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"vertical",{get:function(){return this._vertical},set:function(t){this._vertical=Object(o.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{get:function(){var t=this._selectionModel?this._selectionModel.selected:[];return this.multiple?t.map(function(t){return t.value}):t[0]?t[0].value:void 0},set:function(t){this._setSelectionByValue(t),this.valueChange.emit(this.value)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selected",{get:function(){var t=this._selectionModel.selected;return this.multiple?t:t[0]||null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"multiple",{get:function(){return this._multiple},set:function(t){this._multiple=Object(o.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this._disabled},set:function(t){this._disabled=Object(o.coerceBooleanProperty)(t),this._buttonToggles&&this._buttonToggles.forEach(function(t){return t._markForCheck()})},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){this._selectionModel=new a.SelectionModel(this.multiple,void 0,!1)},t.prototype.ngAfterContentInit=function(){var t;(t=this._selectionModel).select.apply(t,this._buttonToggles.filter(function(t){return t.checked}))},t.prototype.writeValue=function(t){this.value=t,this._changeDetector.markForCheck()},t.prototype.registerOnChange=function(t){this._controlValueAccessorChangeFn=t},t.prototype.registerOnTouched=function(t){this._onTouched=t},t.prototype.setDisabledState=function(t){this.disabled=t},t.prototype._emitChangeEvent=function(){var t=this.selected,e=Array.isArray(t)?t[t.length-1]:t,n=new h(e,this.value);this._controlValueAccessorChangeFn(n.value),this.change.emit(n)},t.prototype._syncButtonToggle=function(t,e,n){void 0===n&&(n=!1),this.multiple||!this.selected||t.checked||(this.selected.checked=!1),e?this._selectionModel.select(t):this._selectionModel.deselect(t),n&&this._emitChangeEvent(),this.valueChange.emit(this.value)},t.prototype._isSelected=function(t){return this._selectionModel.isSelected(t)},t.prototype._isPrechecked=function(t){return void 0!==this._rawValue&&(this.multiple&&Array.isArray(this._rawValue)?this._rawValue.some(function(e){return null!=t.value&&e===t.value}):t.value===this._rawValue)},t.prototype._setSelectionByValue=function(t){var e=this;if(this._rawValue=t,this._buttonToggles)if(this.multiple&&t){if(!Array.isArray(t))throw Error("Value must be an array in multiple-selection mode.");this._clearSelection(),t.forEach(function(t){return e._selectValue(t)})}else this._clearSelection(),this._selectValue(t)},t.prototype._clearSelection=function(){this._selectionModel.clear(),this._buttonToggles.forEach(function(t){return t.checked=!1})},t.prototype._selectValue=function(t){var e=this._buttonToggles.find(function(e){return null!=e.value&&e.value===t});e&&(e.checked=!0,this._selectionModel.select(e))},t.decorators=[{type:s.Directive,args:[{selector:"mat-button-toggle-group",providers:[d,{provide:p,useExisting:t}],host:{role:"group",class:"mat-button-toggle-group","[attr.aria-disabled]":"disabled","[class.mat-button-toggle-vertical]":"vertical","[class.mat-button-toggle-group-appearance-standard]":'appearance === "standard"'},exportAs:"matButtonToggleGroup"}]}],t.ctorParameters=function(){return[{type:s.ChangeDetectorRef},{type:void 0,decorators:[{type:s.Optional},{type:s.Inject,args:[l]}]}]},t.propDecorators={_buttonToggles:[{type:s.ContentChildren,args:[Object(s.forwardRef)(function(){return _})]}],appearance:[{type:s.Input}],name:[{type:s.Input}],vertical:[{type:s.Input}],value:[{type:s.Input}],valueChange:[{type:s.Output}],multiple:[{type:s.Input}],disabled:[{type:s.Input}],change:[{type:s.Output}]},t}(),g=function(){return function(){}}(),y=Object(c.mixinDisableRipple)(g),_=function(t){function e(e,n,r,i,o,a){var u=t.call(this)||this;u._changeDetectorRef=n,u._elementRef=r,u._focusMonitor=i,u._isSingleSelector=!1,u._checked=!1,u.ariaLabelledby=null,u._disabled=!1,u.change=new s.EventEmitter;var c=Number(o);return u.tabIndex=c||0===c?c:null,u.buttonToggleGroup=e,u.appearance=a&&a.appearance?a.appearance:"standard",u}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"buttonId",{get:function(){return this.id+"-button"},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"appearance",{get:function(){return this.buttonToggleGroup?this.buttonToggleGroup.appearance:this._appearance},set:function(t){this._appearance=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"checked",{get:function(){return this.buttonToggleGroup?this.buttonToggleGroup._isSelected(this):this._checked},set:function(t){var e=Object(o.coerceBooleanProperty)(t);e!==this._checked&&(this._checked=e,this.buttonToggleGroup&&this.buttonToggleGroup._syncButtonToggle(this,this._checked),this._changeDetectorRef.markForCheck())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return this._disabled||this.buttonToggleGroup&&this.buttonToggleGroup.disabled},set:function(t){this._disabled=Object(o.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){this._isSingleSelector=this.buttonToggleGroup&&!this.buttonToggleGroup.multiple,this._type=this._isSingleSelector?"radio":"checkbox",this.id=this.id||"mat-button-toggle-"+f++,this._isSingleSelector&&(this.name=this.buttonToggleGroup.name),this.buttonToggleGroup&&this.buttonToggleGroup._isPrechecked(this)&&(this.checked=!0),this._focusMonitor.monitor(this._elementRef,!0)},e.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef)},e.prototype.focus=function(){this._buttonElement.nativeElement.focus()},e.prototype._onButtonClick=function(){var t=!!this._isSingleSelector||!this._checked;t!==this._checked&&(this._checked=t,this.buttonToggleGroup&&(this.buttonToggleGroup._syncButtonToggle(this,this._checked,!0),this.buttonToggleGroup._onTouched())),this.change.emit(new h(this,this.value))},e.prototype._markForCheck=function(){this._changeDetectorRef.markForCheck()},e.decorators=[{type:s.Component,args:[{selector:"mat-button-toggle",template:'<button #button class="mat-button-toggle-button" type="button" [id]="buttonId" [attr.tabindex]="disabled ? -1 : tabIndex" [attr.aria-pressed]="checked" [disabled]="disabled || null" [attr.name]="name || null" [attr.aria-label]="ariaLabel" [attr.aria-labelledby]="ariaLabelledby" (click)="_onButtonClick()"><div class="mat-button-toggle-label-content"><ng-content></ng-content></div></button><div class="mat-button-toggle-focus-overlay"></div><div class="mat-button-toggle-ripple" matRipple [matRippleTrigger]="button" [matRippleDisabled]="this.disableRipple || this.disabled"></div>',styles:[".mat-button-toggle-group,.mat-button-toggle-standalone{position:relative;display:inline-flex;flex-direction:row;white-space:nowrap;overflow:hidden;border-radius:2px;-webkit-tap-highlight-color:transparent}@media (-ms-high-contrast:active){.mat-button-toggle-group,.mat-button-toggle-standalone{outline:solid 1px}}.mat-button-toggle-group-appearance-standard,.mat-button-toggle-standalone.mat-button-toggle-appearance-standard{border-radius:4px}@media (-ms-high-contrast:active){.mat-button-toggle-group-appearance-standard,.mat-button-toggle-standalone.mat-button-toggle-appearance-standard{outline:0}}.mat-button-toggle-vertical{flex-direction:column}.mat-button-toggle-vertical .mat-button-toggle-label-content{display:block}.mat-button-toggle{white-space:nowrap;position:relative}.mat-button-toggle .mat-icon svg{vertical-align:top}.mat-button-toggle.cdk-keyboard-focused .mat-button-toggle-focus-overlay{opacity:1}@media (-ms-high-contrast:active){.mat-button-toggle.cdk-keyboard-focused .mat-button-toggle-focus-overlay{opacity:.5}}.mat-button-toggle-appearance-standard:not(.mat-button-toggle-disabled):hover .mat-button-toggle-focus-overlay{opacity:.04}.mat-button-toggle-appearance-standard.cdk-keyboard-focused:not(.mat-button-toggle-disabled) .mat-button-toggle-focus-overlay{opacity:.12}@media (-ms-high-contrast:active){.mat-button-toggle-appearance-standard.cdk-keyboard-focused:not(.mat-button-toggle-disabled) .mat-button-toggle-focus-overlay{opacity:.5}}@media (hover:none){.mat-button-toggle-appearance-standard:not(.mat-button-toggle-disabled):hover .mat-button-toggle-focus-overlay{display:none}}.mat-button-toggle-label-content{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;display:inline-block;line-height:36px;padding:0 16px;position:relative}.mat-button-toggle-appearance-standard .mat-button-toggle-label-content{line-height:48px;padding:0 12px}.mat-button-toggle-label-content>*{vertical-align:middle}.mat-button-toggle-focus-overlay{border-radius:inherit;pointer-events:none;opacity:0;top:0;left:0;right:0;bottom:0;position:absolute}.mat-button-toggle-checked .mat-button-toggle-focus-overlay{border-bottom:solid 36px}@media (-ms-high-contrast:active){.mat-button-toggle-checked .mat-button-toggle-focus-overlay{opacity:.5;height:0}}@media (-ms-high-contrast:active){.mat-button-toggle-checked.mat-button-toggle-appearance-standard .mat-button-toggle-focus-overlay{border-bottom:solid 48px}}.mat-button-toggle .mat-button-toggle-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-button-toggle-button{border:0;background:0 0;color:inherit;padding:0;margin:0;font:inherit;outline:0;width:100%;cursor:pointer}.mat-button-toggle-disabled .mat-button-toggle-button{cursor:default}.mat-button-toggle-button::-moz-focus-inner{border:0}"],encapsulation:s.ViewEncapsulation.None,exportAs:"matButtonToggle",changeDetection:s.ChangeDetectionStrategy.OnPush,inputs:["disableRipple"],host:{"[class.mat-button-toggle-standalone]":"!buttonToggleGroup","[class.mat-button-toggle-checked]":"checked","[class.mat-button-toggle-disabled]":"disabled","[class.mat-button-toggle-appearance-standard]":'appearance === "standard"',class:"mat-button-toggle","[attr.tabindex]":"-1","[attr.id]":"id","(focus)":"focus()"}}]}],e.ctorParameters=function(){return[{type:m,decorators:[{type:s.Optional}]},{type:s.ChangeDetectorRef},{type:s.ElementRef},{type:i.FocusMonitor},{type:String,decorators:[{type:s.Attribute,args:["tabindex"]}]},{type:void 0,decorators:[{type:s.Optional},{type:s.Inject,args:[l]}]}]},e.propDecorators={ariaLabel:[{type:s.Input,args:["aria-label"]}],ariaLabelledby:[{type:s.Input,args:["aria-labelledby"]}],_buttonElement:[{type:s.ViewChild,args:["button"]}],id:[{type:s.Input}],name:[{type:s.Input}],value:[{type:s.Input}],tabIndex:[{type:s.Input}],appearance:[{type:s.Input}],checked:[{type:s.Input}],disabled:[{type:s.Input}],change:[{type:s.Output}]},e}(y),v=function(){function t(){}return t.decorators=[{type:s.NgModule,args:[{imports:[c.MatCommonModule,c.MatRippleModule],exports:[c.MatCommonModule,m,_],declarations:[m,_]}]}],t}()},"./node_modules/@angular/material/esm5/button.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatButtonModule",function(){return g}),n.d(e,"MatButtonBase",function(){return p}),n.d(e,"_MatButtonMixinBase",function(){return f}),n.d(e,"MatButton",function(){return h}),n.d(e,"MatAnchor",function(){return m});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),o=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),s=n("./node_modules/@angular/material/esm5/core.es5.js"),u=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),c=n("./node_modules/@angular/common/fesm5/common.js"),l="accent",d=["mat-button","mat-flat-button","mat-icon-button","mat-raised-button","mat-stroked-button","mat-mini-fab","mat-fab"],p=function(){return function(t){this._elementRef=t}}(),f=Object(s.mixinColor)(Object(s.mixinDisabled)(Object(s.mixinDisableRipple)(p))),h=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;o._focusMonitor=r,o._animationMode=i,o.isRoundButton=o._hasHostAttributes("mat-fab","mat-mini-fab"),o.isIconButton=o._hasHostAttributes("mat-icon-button");for(var a=0,s=d;a<s.length;a++){var u=s[a];o._hasHostAttributes(u)&&e.nativeElement.classList.add(u)}return o._focusMonitor.monitor(o._elementRef,!0),o.isRoundButton&&(o.color=l),o}return Object(r.__extends)(e,t),e.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef)},e.prototype.focus=function(){this._getHostElement().focus()},e.prototype._getHostElement=function(){return this._elementRef.nativeElement},e.prototype._isRippleDisabled=function(){return this.disableRipple||this.disabled},e.prototype._hasHostAttributes=function(){for(var t=this,e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return e.some(function(e){return t._getHostElement().hasAttribute(e)})},e.decorators=[{type:a.Component,args:[{selector:"button[mat-button], button[mat-raised-button], button[mat-icon-button],\n             button[mat-fab], button[mat-mini-fab], button[mat-stroked-button],\n             button[mat-flat-button]",exportAs:"matButton",host:{"[disabled]":"disabled || null","[class._mat-animation-noopable]":'_animationMode === "NoopAnimations"'},template:'<span class="mat-button-wrapper"><ng-content></ng-content></span><div matRipple class="mat-button-ripple" [class.mat-button-ripple-round]="isRoundButton || isIconButton" [matRippleDisabled]="_isRippleDisabled()" [matRippleCentered]="isIconButton" [matRippleTrigger]="_getHostElement()"></div><div class="mat-button-focus-overlay"></div>',styles:[".mat-button .mat-button-focus-overlay,.mat-icon-button .mat-button-focus-overlay{opacity:0}.mat-button:hover .mat-button-focus-overlay,.mat-stroked-button:hover .mat-button-focus-overlay{opacity:.04}@media (hover:none){.mat-button:hover .mat-button-focus-overlay,.mat-stroked-button:hover .mat-button-focus-overlay{opacity:0}}.mat-button,.mat-flat-button,.mat-icon-button,.mat-stroked-button{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:64px;line-height:36px;padding:0 16px;border-radius:4px;overflow:visible}.mat-button::-moz-focus-inner,.mat-flat-button::-moz-focus-inner,.mat-icon-button::-moz-focus-inner,.mat-stroked-button::-moz-focus-inner{border:0}.mat-button[disabled],.mat-flat-button[disabled],.mat-icon-button[disabled],.mat-stroked-button[disabled]{cursor:default}.mat-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-button.cdk-program-focused .mat-button-focus-overlay,.mat-flat-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-flat-button.cdk-program-focused .mat-button-focus-overlay,.mat-icon-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-icon-button.cdk-program-focused .mat-button-focus-overlay,.mat-stroked-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-stroked-button.cdk-program-focused .mat-button-focus-overlay{opacity:.12}.mat-button::-moz-focus-inner,.mat-flat-button::-moz-focus-inner,.mat-icon-button::-moz-focus-inner,.mat-stroked-button::-moz-focus-inner{border:0}.mat-raised-button{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:64px;line-height:36px;padding:0 16px;border-radius:4px;overflow:visible;transform:translate3d(0,0,0);transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1)}.mat-raised-button::-moz-focus-inner{border:0}.mat-raised-button[disabled]{cursor:default}.mat-raised-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-raised-button.cdk-program-focused .mat-button-focus-overlay{opacity:.12}.mat-raised-button::-moz-focus-inner{border:0}._mat-animation-noopable.mat-raised-button{transition:none;animation:none}.mat-stroked-button{border:1px solid currentColor;padding:0 15px;line-height:34px}.mat-stroked-button .mat-button-focus-overlay,.mat-stroked-button .mat-button-ripple.mat-ripple{top:-1px;left:-1px;right:-1px;bottom:-1px}.mat-fab{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:64px;line-height:36px;padding:0 16px;border-radius:4px;overflow:visible;transform:translate3d(0,0,0);transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1);min-width:0;border-radius:50%;width:56px;height:56px;padding:0;flex-shrink:0}.mat-fab::-moz-focus-inner{border:0}.mat-fab[disabled]{cursor:default}.mat-fab.cdk-keyboard-focused .mat-button-focus-overlay,.mat-fab.cdk-program-focused .mat-button-focus-overlay{opacity:.12}.mat-fab::-moz-focus-inner{border:0}._mat-animation-noopable.mat-fab{transition:none;animation:none}.mat-fab .mat-button-wrapper{padding:16px 0;display:inline-block;line-height:24px}.mat-mini-fab{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:64px;line-height:36px;padding:0 16px;border-radius:4px;overflow:visible;transform:translate3d(0,0,0);transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1);min-width:0;border-radius:50%;width:40px;height:40px;padding:0;flex-shrink:0}.mat-mini-fab::-moz-focus-inner{border:0}.mat-mini-fab[disabled]{cursor:default}.mat-mini-fab.cdk-keyboard-focused .mat-button-focus-overlay,.mat-mini-fab.cdk-program-focused .mat-button-focus-overlay{opacity:.12}.mat-mini-fab::-moz-focus-inner{border:0}._mat-animation-noopable.mat-mini-fab{transition:none;animation:none}.mat-mini-fab .mat-button-wrapper{padding:8px 0;display:inline-block;line-height:24px}.mat-icon-button{padding:0;min-width:0;width:40px;height:40px;flex-shrink:0;line-height:40px;border-radius:50%}.mat-icon-button .mat-icon,.mat-icon-button i{line-height:24px}.mat-button-focus-overlay,.mat-button-ripple.mat-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none;border-radius:inherit}.mat-button-focus-overlay{opacity:0;transition:opacity .2s cubic-bezier(.35,0,.25,1),background-color .2s cubic-bezier(.35,0,.25,1)}._mat-animation-noopable .mat-button-focus-overlay{transition:none}@media (-ms-high-contrast:active){.mat-button-focus-overlay{background-color:rgba(255,255,255,.5)}}.mat-button-ripple-round{border-radius:50%;z-index:1}.mat-button .mat-button-wrapper>*,.mat-fab .mat-button-wrapper>*,.mat-flat-button .mat-button-wrapper>*,.mat-icon-button .mat-button-wrapper>*,.mat-mini-fab .mat-button-wrapper>*,.mat-raised-button .mat-button-wrapper>*,.mat-stroked-button .mat-button-wrapper>*{vertical-align:middle}.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-prefix .mat-icon-button,.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-suffix .mat-icon-button{display:block;font-size:inherit;width:2.5em;height:2.5em}@media (-ms-high-contrast:active){.mat-button,.mat-fab,.mat-flat-button,.mat-icon-button,.mat-mini-fab,.mat-raised-button{outline:solid 1px}}"],inputs:["disabled","disableRipple","color"],encapsulation:a.ViewEncapsulation.None,changeDetection:a.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:a.ElementRef},{type:o.Platform},{type:i.FocusMonitor},{type:String,decorators:[{type:a.Optional},{type:a.Inject,args:[u.ANIMATION_MODULE_TYPE]}]}]},e.propDecorators={ripple:[{type:a.ViewChild,args:[s.MatRipple]}]},e}(f),m=function(t){function e(e,n,r,i){return t.call(this,r,e,n,i)||this}return Object(r.__extends)(e,t),e.prototype._haltDisabledEvents=function(t){this.disabled&&(t.preventDefault(),t.stopImmediatePropagation())},e.decorators=[{type:a.Component,args:[{selector:"a[mat-button], a[mat-raised-button], a[mat-icon-button], a[mat-fab],\n             a[mat-mini-fab], a[mat-stroked-button], a[mat-flat-button]",exportAs:"matButton, matAnchor",host:{"[attr.tabindex]":"disabled ? -1 : (tabIndex || 0)","[attr.disabled]":"disabled || null","[attr.aria-disabled]":"disabled.toString()","(click)":"_haltDisabledEvents($event)","[class._mat-animation-noopable]":'_animationMode === "NoopAnimations"'},inputs:["disabled","disableRipple","color"],template:'<span class="mat-button-wrapper"><ng-content></ng-content></span><div matRipple class="mat-button-ripple" [class.mat-button-ripple-round]="isRoundButton || isIconButton" [matRippleDisabled]="_isRippleDisabled()" [matRippleCentered]="isIconButton" [matRippleTrigger]="_getHostElement()"></div><div class="mat-button-focus-overlay"></div>',styles:[".mat-button .mat-button-focus-overlay,.mat-icon-button .mat-button-focus-overlay{opacity:0}.mat-button:hover .mat-button-focus-overlay,.mat-stroked-button:hover .mat-button-focus-overlay{opacity:.04}@media (hover:none){.mat-button:hover .mat-button-focus-overlay,.mat-stroked-button:hover .mat-button-focus-overlay{opacity:0}}.mat-button,.mat-flat-button,.mat-icon-button,.mat-stroked-button{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:64px;line-height:36px;padding:0 16px;border-radius:4px;overflow:visible}.mat-button::-moz-focus-inner,.mat-flat-button::-moz-focus-inner,.mat-icon-button::-moz-focus-inner,.mat-stroked-button::-moz-focus-inner{border:0}.mat-button[disabled],.mat-flat-button[disabled],.mat-icon-button[disabled],.mat-stroked-button[disabled]{cursor:default}.mat-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-button.cdk-program-focused .mat-button-focus-overlay,.mat-flat-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-flat-button.cdk-program-focused .mat-button-focus-overlay,.mat-icon-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-icon-button.cdk-program-focused .mat-button-focus-overlay,.mat-stroked-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-stroked-button.cdk-program-focused .mat-button-focus-overlay{opacity:.12}.mat-button::-moz-focus-inner,.mat-flat-button::-moz-focus-inner,.mat-icon-button::-moz-focus-inner,.mat-stroked-button::-moz-focus-inner{border:0}.mat-raised-button{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:64px;line-height:36px;padding:0 16px;border-radius:4px;overflow:visible;transform:translate3d(0,0,0);transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1)}.mat-raised-button::-moz-focus-inner{border:0}.mat-raised-button[disabled]{cursor:default}.mat-raised-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-raised-button.cdk-program-focused .mat-button-focus-overlay{opacity:.12}.mat-raised-button::-moz-focus-inner{border:0}._mat-animation-noopable.mat-raised-button{transition:none;animation:none}.mat-stroked-button{border:1px solid currentColor;padding:0 15px;line-height:34px}.mat-stroked-button .mat-button-focus-overlay,.mat-stroked-button .mat-button-ripple.mat-ripple{top:-1px;left:-1px;right:-1px;bottom:-1px}.mat-fab{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:64px;line-height:36px;padding:0 16px;border-radius:4px;overflow:visible;transform:translate3d(0,0,0);transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1);min-width:0;border-radius:50%;width:56px;height:56px;padding:0;flex-shrink:0}.mat-fab::-moz-focus-inner{border:0}.mat-fab[disabled]{cursor:default}.mat-fab.cdk-keyboard-focused .mat-button-focus-overlay,.mat-fab.cdk-program-focused .mat-button-focus-overlay{opacity:.12}.mat-fab::-moz-focus-inner{border:0}._mat-animation-noopable.mat-fab{transition:none;animation:none}.mat-fab .mat-button-wrapper{padding:16px 0;display:inline-block;line-height:24px}.mat-mini-fab{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:64px;line-height:36px;padding:0 16px;border-radius:4px;overflow:visible;transform:translate3d(0,0,0);transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1);min-width:0;border-radius:50%;width:40px;height:40px;padding:0;flex-shrink:0}.mat-mini-fab::-moz-focus-inner{border:0}.mat-mini-fab[disabled]{cursor:default}.mat-mini-fab.cdk-keyboard-focused .mat-button-focus-overlay,.mat-mini-fab.cdk-program-focused .mat-button-focus-overlay{opacity:.12}.mat-mini-fab::-moz-focus-inner{border:0}._mat-animation-noopable.mat-mini-fab{transition:none;animation:none}.mat-mini-fab .mat-button-wrapper{padding:8px 0;display:inline-block;line-height:24px}.mat-icon-button{padding:0;min-width:0;width:40px;height:40px;flex-shrink:0;line-height:40px;border-radius:50%}.mat-icon-button .mat-icon,.mat-icon-button i{line-height:24px}.mat-button-focus-overlay,.mat-button-ripple.mat-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none;border-radius:inherit}.mat-button-focus-overlay{opacity:0;transition:opacity .2s cubic-bezier(.35,0,.25,1),background-color .2s cubic-bezier(.35,0,.25,1)}._mat-animation-noopable .mat-button-focus-overlay{transition:none}@media (-ms-high-contrast:active){.mat-button-focus-overlay{background-color:rgba(255,255,255,.5)}}.mat-button-ripple-round{border-radius:50%;z-index:1}.mat-button .mat-button-wrapper>*,.mat-fab .mat-button-wrapper>*,.mat-flat-button .mat-button-wrapper>*,.mat-icon-button .mat-button-wrapper>*,.mat-mini-fab .mat-button-wrapper>*,.mat-raised-button .mat-button-wrapper>*,.mat-stroked-button .mat-button-wrapper>*{vertical-align:middle}.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-prefix .mat-icon-button,.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-suffix .mat-icon-button{display:block;font-size:inherit;width:2.5em;height:2.5em}@media (-ms-high-contrast:active){.mat-button,.mat-fab,.mat-flat-button,.mat-icon-button,.mat-mini-fab,.mat-raised-button{outline:solid 1px}}"],encapsulation:a.ViewEncapsulation.None,changeDetection:a.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:o.Platform},{type:i.FocusMonitor},{type:a.ElementRef},{type:String,decorators:[{type:a.Optional},{type:a.Inject,args:[u.ANIMATION_MODULE_TYPE]}]}]},e.propDecorators={tabIndex:[{type:a.Input}]},e}(h),g=function(){function t(){}return t.decorators=[{type:a.NgModule,args:[{imports:[c.CommonModule,s.MatRippleModule,s.MatCommonModule],exports:[h,m,s.MatCommonModule],declarations:[h,m]}]}],t}()},"./node_modules/@angular/material/esm5/card.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatCardContent",function(){return o}),n.d(e,"MatCardTitle",function(){return a}),n.d(e,"MatCardSubtitle",function(){return s}),n.d(e,"MatCardActions",function(){return u}),n.d(e,"MatCardFooter",function(){return c}),n.d(e,"MatCardImage",function(){return l}),n.d(e,"MatCardSmImage",function(){return d}),n.d(e,"MatCardMdImage",function(){return p}),n.d(e,"MatCardLgImage",function(){return f}),n.d(e,"MatCardXlImage",function(){return h}),n.d(e,"MatCardAvatar",function(){return m}),n.d(e,"MatCard",function(){return g}),n.d(e,"MatCardHeader",function(){return y}),n.d(e,"MatCardTitleGroup",function(){return _}),n.d(e,"MatCardModule",function(){return v});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/material/esm5/core.es5.js"),o=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"mat-card-content",host:{class:"mat-card-content"}}]}],t}(),a=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"mat-card-title, [mat-card-title], [matCardTitle]",host:{class:"mat-card-title"}}]}],t}(),s=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"mat-card-subtitle, [mat-card-subtitle], [matCardSubtitle]",host:{class:"mat-card-subtitle"}}]}],t}(),u=function(){function t(){this.align="start"}return t.decorators=[{type:r.Directive,args:[{selector:"mat-card-actions",exportAs:"matCardActions",host:{class:"mat-card-actions","[class.mat-card-actions-align-end]":'align === "end"'}}]}],t.propDecorators={align:[{type:r.Input}]},t}(),c=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"mat-card-footer",host:{class:"mat-card-footer"}}]}],t}(),l=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"[mat-card-image], [matCardImage]",host:{class:"mat-card-image"}}]}],t}(),d=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"[mat-card-sm-image], [matCardImageSmall]",host:{class:"mat-card-sm-image"}}]}],t}(),p=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"[mat-card-md-image], [matCardImageMedium]",host:{class:"mat-card-md-image"}}]}],t}(),f=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"[mat-card-lg-image], [matCardImageLarge]",host:{class:"mat-card-lg-image"}}]}],t}(),h=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"[mat-card-xl-image], [matCardImageXLarge]",host:{class:"mat-card-xl-image"}}]}],t}(),m=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"[mat-card-avatar], [matCardAvatar]",host:{class:"mat-card-avatar"}}]}],t}(),g=function(){function t(){}return t.decorators=[{type:r.Component,args:[{selector:"mat-card",exportAs:"matCard",template:'<ng-content></ng-content><ng-content select="mat-card-footer"></ng-content>',styles:[".mat-card{transition:box-shadow 280ms cubic-bezier(.4,0,.2,1);display:block;position:relative;padding:16px;border-radius:4px}.mat-card .mat-divider-horizontal{position:absolute;left:0;width:100%}[dir=rtl] .mat-card .mat-divider-horizontal{left:auto;right:0}.mat-card .mat-divider-horizontal.mat-divider-inset{position:static;margin:0}[dir=rtl] .mat-card .mat-divider-horizontal.mat-divider-inset{margin-right:0}@media (-ms-high-contrast:active){.mat-card{outline:solid 1px}}.mat-card-actions,.mat-card-content,.mat-card-subtitle{display:block;margin-bottom:16px}.mat-card-title{display:block;margin-bottom:8px}.mat-card-actions{margin-left:-8px;margin-right:-8px;padding:8px 0}.mat-card-actions-align-end{display:flex;justify-content:flex-end}.mat-card-image{width:calc(100% + 32px);margin:0 -16px 16px -16px}.mat-card-footer{display:block;margin:0 -16px -16px -16px}.mat-card-actions .mat-button,.mat-card-actions .mat-raised-button{margin:0 8px}.mat-card-header{display:flex;flex-direction:row}.mat-card-header .mat-card-title{margin-bottom:12px}.mat-card-header-text{margin:0 16px}.mat-card-avatar{height:40px;width:40px;border-radius:50%;flex-shrink:0;object-fit:cover}.mat-card-title-group{display:flex;justify-content:space-between}.mat-card-sm-image{width:80px;height:80px}.mat-card-md-image{width:112px;height:112px}.mat-card-lg-image{width:152px;height:152px}.mat-card-xl-image{width:240px;height:240px;margin:-8px}.mat-card-title-group>.mat-card-xl-image{margin:-8px 0 8px 0}@media (max-width:599px){.mat-card-title-group{margin:0}.mat-card-xl-image{margin-left:0;margin-right:0}}.mat-card-content>:first-child,.mat-card>:first-child{margin-top:0}.mat-card-content>:last-child:not(.mat-card-footer),.mat-card>:last-child:not(.mat-card-footer){margin-bottom:0}.mat-card-image:first-child{margin-top:-16px;border-top-left-radius:inherit;border-top-right-radius:inherit}.mat-card>.mat-card-actions:last-child{margin-bottom:-8px;padding-bottom:0}.mat-card-actions .mat-button:first-child,.mat-card-actions .mat-raised-button:first-child{margin-left:0;margin-right:0}.mat-card-subtitle:not(:first-child),.mat-card-title:not(:first-child){margin-top:-4px}.mat-card-header .mat-card-subtitle:not(:first-child){margin-top:-8px}.mat-card>.mat-card-xl-image:first-child{margin-top:-8px}.mat-card>.mat-card-xl-image:last-child{margin-bottom:-8px}"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,host:{class:"mat-card"}}]}],t}(),y=function(){function t(){}return t.decorators=[{type:r.Component,args:[{selector:"mat-card-header",template:'<ng-content select="[mat-card-avatar], [matCardAvatar]"></ng-content><div class="mat-card-header-text"><ng-content select="mat-card-title, mat-card-subtitle, [mat-card-title], [mat-card-subtitle], [matCardTitle], [matCardSubtitle]"></ng-content></div><ng-content></ng-content>',encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,host:{class:"mat-card-header"}}]}],t}(),_=function(){function t(){}return t.decorators=[{type:r.Component,args:[{selector:"mat-card-title-group",template:'<div><ng-content select="mat-card-title, mat-card-subtitle, [mat-card-title], [mat-card-subtitle], [matCardTitle], [matCardSubtitle]"></ng-content></div><ng-content select="img"></ng-content><ng-content></ng-content>',encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,host:{class:"mat-card-title-group"}}]}],t}(),v=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[i.MatCommonModule],exports:[g,y,_,o,a,s,u,c,d,p,f,l,h,m,i.MatCommonModule],declarations:[g,y,_,o,a,s,u,c,d,p,f,l,h,m]}]}],t}()},"./node_modules/@angular/material/esm5/checkbox.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MAT_CHECKBOX_CONTROL_VALUE_ACCESSOR",function(){return m}),n.d(e,"TransitionCheckState",function(){return g}),n.d(e,"MatCheckboxChange",function(){return y}),n.d(e,"MatCheckboxBase",function(){return _}),n.d(e,"_MatCheckboxMixinBase",function(){return v}),n.d(e,"MatCheckbox",function(){return b}),n.d(e,"MAT_CHECKBOX_CLICK_ACTION",function(){return p}),n.d(e,"MatCheckboxModule",function(){return j}),n.d(e,"MAT_CHECKBOX_REQUIRED_VALIDATOR",function(){return x}),n.d(e,"MatCheckboxRequiredValidator",function(){return w});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),a=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),s=n("./node_modules/@angular/forms/fesm5/forms.js"),u=n("./node_modules/@angular/material/esm5/core.es5.js"),c=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),l=n("./node_modules/@angular/cdk/esm5/observers.es5.js"),d=n("./node_modules/@angular/common/fesm5/common.js"),p=new r.InjectionToken("mat-checkbox-click-action"),f=0,h=0,m={provide:s.NG_VALUE_ACCESSOR,useExisting:Object(r.forwardRef)(function(){return b}),multi:!0},g={Init:0,Checked:1,Unchecked:2,Indeterminate:3};g[g.Init]="Init",g[g.Checked]="Checked",g[g.Unchecked]="Unchecked",g[g.Indeterminate]="Indeterminate";var y=function(){return function(){}}(),_=function(){return function(t){this._elementRef=t}}(),v=Object(u.mixinTabIndex)(Object(u.mixinColor)(Object(u.mixinDisableRipple)(Object(u.mixinDisabled)(_)),"accent")),b=function(t){function e(e,n,i,o,a,s,u){var c=t.call(this,e)||this;return c._changeDetectorRef=n,c._focusMonitor=i,c._ngZone=o,c._clickAction=s,c._animationMode=u,c.ariaLabel="",c.ariaLabelledby=null,c._uniqueId="mat-checkbox-"+ ++f,c.id=c._uniqueId,c.labelPosition="after",c.name=null,c.change=new r.EventEmitter,c.indeterminateChange=new r.EventEmitter,c._onTouched=function(){},c._currentAnimationClass="",c._currentCheckState=g.Init,c._controlValueAccessorChangeFn=function(){},c._checked=!1,c._disabled=!1,c._indeterminate=!1,c.tabIndex=parseInt(a)||0,c._focusMonitor.monitor(e,!0).subscribe(function(t){t||Promise.resolve().then(function(){return c._onTouched()})}),c}return Object(i.__extends)(e,t),Object.defineProperty(e.prototype,"inputId",{get:function(){return(this.id||this._uniqueId)+"-input"},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"required",{get:function(){return this._required},set:function(t){this._required=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e.prototype.ngAfterViewChecked=function(){this._calculateRippleRadius()},e.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef)},Object.defineProperty(e.prototype,"checked",{get:function(){return this._checked},set:function(t){t!=this.checked&&(this._checked=t,this._changeDetectorRef.markForCheck())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return this._disabled},set:function(t){var e=Object(a.coerceBooleanProperty)(t);e!==this.disabled&&(this._disabled=e,this._changeDetectorRef.markForCheck())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"indeterminate",{get:function(){return this._indeterminate},set:function(t){var e=t!=this._indeterminate;this._indeterminate=t,e&&(this._indeterminate?this._transitionCheckState(g.Indeterminate):this._transitionCheckState(this.checked?g.Checked:g.Unchecked),this.indeterminateChange.emit(this._indeterminate))},enumerable:!0,configurable:!0}),e.prototype._isRippleDisabled=function(){return this.disableRipple||this.disabled},e.prototype._onLabelTextChange=function(){this._changeDetectorRef.detectChanges()},e.prototype.writeValue=function(t){this.checked=!!t},e.prototype.registerOnChange=function(t){this._controlValueAccessorChangeFn=t},e.prototype.registerOnTouched=function(t){this._onTouched=t},e.prototype.setDisabledState=function(t){this.disabled=t},e.prototype._getAriaChecked=function(){return this.checked?"true":this.indeterminate?"mixed":"false"},e.prototype._transitionCheckState=function(t){var e=this._currentCheckState,n=this._elementRef.nativeElement;if(e!==t&&(this._currentAnimationClass.length>0&&n.classList.remove(this._currentAnimationClass),this._currentAnimationClass=this._getAnimationClassForCheckStateTransition(e,t),this._currentCheckState=t,this._currentAnimationClass.length>0)){n.classList.add(this._currentAnimationClass);var r=this._currentAnimationClass;this._ngZone.runOutsideAngular(function(){setTimeout(function(){n.classList.remove(r)},1e3)})}},e.prototype._emitChangeEvent=function(){var t=new y;t.source=this,t.checked=this.checked,this._controlValueAccessorChangeFn(this.checked),this.change.emit(t)},e.prototype.toggle=function(){this.checked=!this.checked},e.prototype._onInputClick=function(t){var e=this;t.stopPropagation(),this.disabled||"noop"===this._clickAction?this.disabled||"noop"!==this._clickAction||(this._inputElement.nativeElement.checked=this.checked,this._inputElement.nativeElement.indeterminate=this.indeterminate):(this.indeterminate&&"check"!==this._clickAction&&Promise.resolve().then(function(){e._indeterminate=!1,e.indeterminateChange.emit(e._indeterminate)}),this.toggle(),this._transitionCheckState(this._checked?g.Checked:g.Unchecked),this._emitChangeEvent())},e.prototype.focus=function(){this._focusMonitor.focusVia(this._inputElement,"keyboard")},e.prototype._onInteractionEvent=function(t){t.stopPropagation()},e.prototype._getAnimationClassForCheckStateTransition=function(t,e){if("NoopAnimations"===this._animationMode)return"";var n="";switch(t){case g.Init:if(e===g.Checked)n="unchecked-checked";else{if(e!=g.Indeterminate)return"";n="unchecked-indeterminate"}break;case g.Unchecked:n=e===g.Checked?"unchecked-checked":"unchecked-indeterminate";break;case g.Checked:n=e===g.Unchecked?"checked-unchecked":"checked-indeterminate";break;case g.Indeterminate:n=e===g.Checked?"indeterminate-checked":"indeterminate-unchecked"}return"mat-checkbox-anim-"+n},e.prototype._calculateRippleRadius=function(){if(!h){var t=this._elementRef.nativeElement.querySelector(".mat-checkbox-ripple").clientWidth||0;h=t/2}this.ripple.radius=h},e.decorators=[{type:r.Component,args:[{selector:"mat-checkbox",template:'<label [attr.for]="inputId" class="mat-checkbox-layout" #label><div class="mat-checkbox-inner-container" [class.mat-checkbox-inner-container-no-side-margin]="!checkboxLabel.textContent || !checkboxLabel.textContent.trim()"><input #input class="mat-checkbox-input cdk-visually-hidden" type="checkbox" [id]="inputId" [required]="required" [checked]="checked" [attr.value]="value" [disabled]="disabled" [attr.name]="name" [tabIndex]="tabIndex" [indeterminate]="indeterminate" [attr.aria-label]="ariaLabel || null" [attr.aria-labelledby]="ariaLabelledby" [attr.aria-checked]="_getAriaChecked()" (change)="_onInteractionEvent($event)" (click)="_onInputClick($event)"><div matRipple class="mat-checkbox-ripple" [matRippleTrigger]="label" [matRippleDisabled]="_isRippleDisabled()" [matRippleRadius]="20" [matRippleCentered]="true" [matRippleAnimation]="{enterDuration: 150}"><div class="mat-ripple-element mat-checkbox-persistent-ripple"></div></div><div class="mat-checkbox-frame"></div><div class="mat-checkbox-background"><svg version="1.1" focusable="false" class="mat-checkbox-checkmark" viewBox="0 0 24 24" xml:space="preserve"><path class="mat-checkbox-checkmark-path" fill="none" stroke="white" d="M4.1,12.7 9,17.6 20.3,6.3"/></svg><div class="mat-checkbox-mixedmark"></div></div></div><span class="mat-checkbox-label" #checkboxLabel (cdkObserveContent)="_onLabelTextChange()"><span style="display:none">&nbsp;</span><ng-content></ng-content></span></label>',styles:["@keyframes mat-checkbox-fade-in-background{0%{opacity:0}50%{opacity:1}}@keyframes mat-checkbox-fade-out-background{0%,50%{opacity:1}100%{opacity:0}}@keyframes mat-checkbox-unchecked-checked-checkmark-path{0%,50%{stroke-dashoffset:22.91026}50%{animation-timing-function:cubic-bezier(0,0,.2,.1)}100%{stroke-dashoffset:0}}@keyframes mat-checkbox-unchecked-indeterminate-mixedmark{0%,68.2%{transform:scaleX(0)}68.2%{animation-timing-function:cubic-bezier(0,0,0,1)}100%{transform:scaleX(1)}}@keyframes mat-checkbox-checked-unchecked-checkmark-path{from{animation-timing-function:cubic-bezier(.4,0,1,1);stroke-dashoffset:0}to{stroke-dashoffset:-22.91026}}@keyframes mat-checkbox-checked-indeterminate-checkmark{from{animation-timing-function:cubic-bezier(0,0,.2,.1);opacity:1;transform:rotate(0)}to{opacity:0;transform:rotate(45deg)}}@keyframes mat-checkbox-indeterminate-checked-checkmark{from{animation-timing-function:cubic-bezier(.14,0,0,1);opacity:0;transform:rotate(45deg)}to{opacity:1;transform:rotate(360deg)}}@keyframes mat-checkbox-checked-indeterminate-mixedmark{from{animation-timing-function:cubic-bezier(0,0,.2,.1);opacity:0;transform:rotate(-45deg)}to{opacity:1;transform:rotate(0)}}@keyframes mat-checkbox-indeterminate-checked-mixedmark{from{animation-timing-function:cubic-bezier(.14,0,0,1);opacity:1;transform:rotate(0)}to{opacity:0;transform:rotate(315deg)}}@keyframes mat-checkbox-indeterminate-unchecked-mixedmark{0%{animation-timing-function:linear;opacity:1;transform:scaleX(1)}100%,32.8%{opacity:0;transform:scaleX(0)}}.mat-checkbox-background,.mat-checkbox-frame{top:0;left:0;right:0;bottom:0;position:absolute;border-radius:2px;box-sizing:border-box;pointer-events:none}.mat-checkbox{transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1);cursor:pointer;-webkit-tap-highlight-color:transparent}._mat-animation-noopable.mat-checkbox{transition:none;animation:none}.mat-checkbox .mat-ripple-element:not(.mat-checkbox-persistent-ripple){opacity:.16}.mat-checkbox-layout{cursor:inherit;align-items:baseline;vertical-align:middle;display:inline-flex;white-space:nowrap}.mat-checkbox-inner-container{display:inline-block;height:16px;line-height:0;margin:auto;margin-right:8px;order:0;position:relative;vertical-align:middle;white-space:nowrap;width:16px;flex-shrink:0}[dir=rtl] .mat-checkbox-inner-container{margin-left:8px;margin-right:auto}.mat-checkbox-inner-container-no-side-margin{margin-left:0;margin-right:0}.mat-checkbox-frame{background-color:transparent;transition:border-color 90ms cubic-bezier(0,0,.2,.1);border-width:2px;border-style:solid}._mat-animation-noopable .mat-checkbox-frame{transition:none}@media (-ms-high-contrast:active){.mat-checkbox.cdk-keyboard-focused .mat-checkbox-frame{border-style:dotted}}.mat-checkbox-background{align-items:center;display:inline-flex;justify-content:center;transition:background-color 90ms cubic-bezier(0,0,.2,.1),opacity 90ms cubic-bezier(0,0,.2,.1)}._mat-animation-noopable .mat-checkbox-background{transition:none}.mat-checkbox-persistent-ripple{width:100%;height:100%;transform:none}.mat-checkbox-inner-container:hover .mat-checkbox-persistent-ripple{opacity:.04}.mat-checkbox.cdk-keyboard-focused .mat-checkbox-persistent-ripple{opacity:.12}.mat-checkbox-persistent-ripple,.mat-checkbox.mat-disabled .mat-checkbox-inner-container:hover .mat-checkbox-persistent-ripple{opacity:0}.mat-checkbox-checkmark{top:0;left:0;right:0;bottom:0;position:absolute;width:100%}.mat-checkbox-checkmark-path{stroke-dashoffset:22.91026;stroke-dasharray:22.91026;stroke-width:2.13333px}.mat-checkbox-mixedmark{width:calc(100% - 6px);height:2px;opacity:0;transform:scaleX(0) rotate(0);border-radius:2px}@media (-ms-high-contrast:active){.mat-checkbox-mixedmark{height:0;border-top:solid 2px;margin-top:2px}}.mat-checkbox-label-before .mat-checkbox-inner-container{order:1;margin-left:8px;margin-right:auto}[dir=rtl] .mat-checkbox-label-before .mat-checkbox-inner-container{margin-left:auto;margin-right:8px}.mat-checkbox-checked .mat-checkbox-checkmark{opacity:1}.mat-checkbox-checked .mat-checkbox-checkmark-path{stroke-dashoffset:0}.mat-checkbox-checked .mat-checkbox-mixedmark{transform:scaleX(1) rotate(-45deg)}.mat-checkbox-indeterminate .mat-checkbox-checkmark{opacity:0;transform:rotate(45deg)}.mat-checkbox-indeterminate .mat-checkbox-checkmark-path{stroke-dashoffset:0}.mat-checkbox-indeterminate .mat-checkbox-mixedmark{opacity:1;transform:scaleX(1) rotate(0)}.mat-checkbox-unchecked .mat-checkbox-background{background-color:transparent}.mat-checkbox-disabled{cursor:default}.mat-checkbox-anim-unchecked-checked .mat-checkbox-background{animation:180ms linear 0s mat-checkbox-fade-in-background}.mat-checkbox-anim-unchecked-checked .mat-checkbox-checkmark-path{animation:180ms linear 0s mat-checkbox-unchecked-checked-checkmark-path}.mat-checkbox-anim-unchecked-indeterminate .mat-checkbox-background{animation:180ms linear 0s mat-checkbox-fade-in-background}.mat-checkbox-anim-unchecked-indeterminate .mat-checkbox-mixedmark{animation:90ms linear 0s mat-checkbox-unchecked-indeterminate-mixedmark}.mat-checkbox-anim-checked-unchecked .mat-checkbox-background{animation:180ms linear 0s mat-checkbox-fade-out-background}.mat-checkbox-anim-checked-unchecked .mat-checkbox-checkmark-path{animation:90ms linear 0s mat-checkbox-checked-unchecked-checkmark-path}.mat-checkbox-anim-checked-indeterminate .mat-checkbox-checkmark{animation:90ms linear 0s mat-checkbox-checked-indeterminate-checkmark}.mat-checkbox-anim-checked-indeterminate .mat-checkbox-mixedmark{animation:90ms linear 0s mat-checkbox-checked-indeterminate-mixedmark}.mat-checkbox-anim-indeterminate-checked .mat-checkbox-checkmark{animation:.5s linear 0s mat-checkbox-indeterminate-checked-checkmark}.mat-checkbox-anim-indeterminate-checked .mat-checkbox-mixedmark{animation:.5s linear 0s mat-checkbox-indeterminate-checked-mixedmark}.mat-checkbox-anim-indeterminate-unchecked .mat-checkbox-background{animation:180ms linear 0s mat-checkbox-fade-out-background}.mat-checkbox-anim-indeterminate-unchecked .mat-checkbox-mixedmark{animation:.3s linear 0s mat-checkbox-indeterminate-unchecked-mixedmark}.mat-checkbox-input{bottom:0;left:50%}.mat-checkbox .mat-checkbox-ripple{position:absolute;left:calc(50% - 20px);top:calc(50% - 20px);height:40px;width:40px;z-index:1;pointer-events:none}"],exportAs:"matCheckbox",host:{class:"mat-checkbox","[id]":"id","[attr.tabindex]":"null","[class.mat-checkbox-indeterminate]":"indeterminate","[class.mat-checkbox-checked]":"checked","[class.mat-checkbox-disabled]":"disabled","[class.mat-checkbox-label-before]":'labelPosition == "before"',"[class._mat-animation-noopable]":"_animationMode === 'NoopAnimations'"},providers:[m],inputs:["disableRipple","color","tabIndex"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:r.ChangeDetectorRef},{type:o.FocusMonitor},{type:r.NgZone},{type:String,decorators:[{type:r.Attribute,args:["tabindex"]}]},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[p]}]},{type:String,decorators:[{type:r.Optional},{type:r.Inject,args:[c.ANIMATION_MODULE_TYPE]}]}]},e.propDecorators={ariaLabel:[{type:r.Input,args:["aria-label"]}],ariaLabelledby:[{type:r.Input,args:["aria-labelledby"]}],id:[{type:r.Input}],required:[{type:r.Input}],labelPosition:[{type:r.Input}],name:[{type:r.Input}],change:[{type:r.Output}],indeterminateChange:[{type:r.Output}],value:[{type:r.Input}],_inputElement:[{type:r.ViewChild,args:["input"]}],ripple:[{type:r.ViewChild,args:[u.MatRipple]}],checked:[{type:r.Input}],disabled:[{type:r.Input}],indeterminate:[{type:r.Input}]},e}(v),x={provide:s.NG_VALIDATORS,useExisting:Object(r.forwardRef)(function(){return w}),multi:!0},w=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(i.__extends)(e,t),e.decorators=[{type:r.Directive,args:[{selector:"mat-checkbox[required][formControlName],\n             mat-checkbox[required][formControl], mat-checkbox[required][ngModel]",providers:[x],host:{"[attr.required]":'required ? "" : null'}}]}],e}(s.CheckboxRequiredValidator),j=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[d.CommonModule,u.MatRippleModule,u.MatCommonModule,l.ObserversModule],exports:[b,w,u.MatCommonModule],declarations:[b,w]}]}],t}()},"./node_modules/@angular/material/esm5/chips.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatChipsModule",function(){return P}),n.d(e,"MatChipListBase",function(){return C}),n.d(e,"_MatChipListMixinBase",function(){return E}),n.d(e,"MatChipListChange",function(){return M}),n.d(e,"MatChipList",function(){return T}),n.d(e,"MatChipSelectionChange",function(){return g}),n.d(e,"MatChipBase",function(){return y}),n.d(e,"_MatChipMixinBase",function(){return _}),n.d(e,"MatChipAvatar",function(){return b}),n.d(e,"MatChipTrailingIcon",function(){return x}),n.d(e,"MatChip",function(){return w}),n.d(e,"MatChipRemove",function(){return j}),n.d(e,"MatChipInput",function(){return A}),n.d(e,"MAT_CHIPS_DEFAULT_OPTIONS",function(){return O});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),o=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),a=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),s=n("./node_modules/@angular/core/fesm5/core.js"),u=n("./node_modules/@angular/material/esm5/core.es5.js"),c=n("./node_modules/rxjs/_esm5/index.js"),l=n("./node_modules/rxjs/_esm5/operators/index.js"),d=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),p=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),f=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),h=n("./node_modules/@angular/forms/fesm5/forms.js"),m=n("./node_modules/@angular/material/esm5/form-field.es5.js"),g=function(){return function(t,e,n){void 0===n&&(n=!1),this.source=t,this.selected=e,this.isUserInput=n}}(),y=function(){return function(t){this._elementRef=t}}(),_=Object(u.mixinColor)(Object(u.mixinDisableRipple)(Object(u.mixinDisabled)(y)),"primary"),v=["mat-basic-chip"],b=function(){function t(){}return t.decorators=[{type:s.Directive,args:[{selector:"mat-chip-avatar, [matChipAvatar]",host:{class:"mat-chip-avatar"}}]}],t}(),x=function(){function t(){}return t.decorators=[{type:s.Directive,args:[{selector:"mat-chip-trailing-icon, [matChipTrailingIcon]",host:{class:"mat-chip-trailing-icon"}}]}],t}(),w=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o._elementRef=e,o._ngZone=n,o._ripplesGloballyDisabled=!1,o.rippleConfig={},o._hasFocus=!1,o.chipListSelectable=!0,o._selected=!1,o._selectable=!0,o._removable=!0,o._onFocus=new c.Subject,o._onBlur=new c.Subject,o.selectionChange=new s.EventEmitter,o.destroyed=new s.EventEmitter,o.removed=new s.EventEmitter,o._addHostClassName(),o._chipRipple=new u.RippleRenderer(o,n,e,r),o._chipRipple.setupTriggerEvents(e.nativeElement),i&&(o._ripplesGloballyDisabled=!!i.disabled,o.rippleConfig={animation:i.animation,terminateOnPointerUp:i.terminateOnPointerUp}),o}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"rippleDisabled",{get:function(){return this.disabled||this.disableRipple||this._ripplesGloballyDisabled},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selected",{get:function(){return this._selected},set:function(t){var e=Object(i.coerceBooleanProperty)(t);e!==this._selected&&(this._selected=e,this._dispatchSelectionChange())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return void 0!=this._value?this._value:this._elementRef.nativeElement.textContent},set:function(t){this._value=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selectable",{get:function(){return this._selectable&&this.chipListSelectable},set:function(t){this._selectable=Object(i.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"removable",{get:function(){return this._removable},set:function(t){this._removable=Object(i.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ariaSelected",{get:function(){return this.selectable?this.selected.toString():null},enumerable:!0,configurable:!0}),e.prototype._addHostClassName=function(){for(var t=0,e=v;t<e.length;t++){var n=e[t];if(this._elementRef.nativeElement.hasAttribute(n)||this._elementRef.nativeElement.tagName.toLowerCase()===n)return void this._elementRef.nativeElement.classList.add(n)}this._elementRef.nativeElement.classList.add("mat-standard-chip")},e.prototype.ngOnDestroy=function(){this.destroyed.emit({chip:this}),this._chipRipple._removeTriggerEvents()},e.prototype.select=function(){this._selected||(this._selected=!0,this._dispatchSelectionChange())},e.prototype.deselect=function(){this._selected&&(this._selected=!1,this._dispatchSelectionChange())},e.prototype.selectViaInteraction=function(){this._selected||(this._selected=!0,this._dispatchSelectionChange(!0))},e.prototype.toggleSelected=function(t){return void 0===t&&(t=!1),this._selected=!this.selected,this._dispatchSelectionChange(t),this.selected},e.prototype.focus=function(){this._hasFocus||(this._elementRef.nativeElement.focus(),this._onFocus.next({chip:this})),this._hasFocus=!0},e.prototype.remove=function(){this.removable&&this.removed.emit({chip:this})},e.prototype._handleClick=function(t){this.disabled?t.preventDefault():t.stopPropagation()},e.prototype._handleKeydown=function(t){if(!this.disabled)switch(t.keyCode){case o.DELETE:case o.BACKSPACE:this.remove(),t.preventDefault();break;case o.SPACE:this.selectable&&this.toggleSelected(!0),t.preventDefault()}},e.prototype._blur=function(){var t=this;this._ngZone.onStable.asObservable().pipe(Object(l.take)(1)).subscribe(function(){t._ngZone.run(function(){t._hasFocus=!1,t._onBlur.next({chip:t})})})},e.prototype._dispatchSelectionChange=function(t){void 0===t&&(t=!1),this.selectionChange.emit({source:this,isUserInput:t,selected:this._selected})},e.decorators=[{type:s.Directive,args:[{selector:"mat-basic-chip, [mat-basic-chip], mat-chip, [mat-chip]",inputs:["color","disabled","disableRipple"],exportAs:"matChip",host:{class:"mat-chip","[attr.tabindex]":"disabled ? null : -1",role:"option","[class.mat-chip-selected]":"selected","[class.mat-chip-with-avatar]":"avatar","[class.mat-chip-with-trailing-icon]":"trailingIcon || removeIcon","[class.mat-chip-disabled]":"disabled","[attr.disabled]":"disabled || null","[attr.aria-disabled]":"disabled.toString()","[attr.aria-selected]":"ariaSelected","(click)":"_handleClick($event)","(keydown)":"_handleKeydown($event)","(focus)":"focus()","(blur)":"_blur()"}}]}],e.ctorParameters=function(){return[{type:s.ElementRef},{type:s.NgZone},{type:a.Platform},{type:void 0,decorators:[{type:s.Optional},{type:s.Inject,args:[u.MAT_RIPPLE_GLOBAL_OPTIONS]}]}]},e.propDecorators={avatar:[{type:s.ContentChild,args:[b]}],trailingIcon:[{type:s.ContentChild,args:[x]}],removeIcon:[{type:s.ContentChild,args:[Object(s.forwardRef)(function(){return j})]}],selected:[{type:s.Input}],value:[{type:s.Input}],selectable:[{type:s.Input}],removable:[{type:s.Input}],selectionChange:[{type:s.Output}],destroyed:[{type:s.Output}],removed:[{type:s.Output}]},e}(_),j=function(){function t(t){this._parentChip=t}return t.prototype._handleClick=function(t){this._parentChip.removable&&this._parentChip.remove(),t.stopPropagation()},t.decorators=[{type:s.Directive,args:[{selector:"[matChipRemove]",host:{class:"mat-chip-remove mat-chip-trailing-icon","(click)":"_handleClick($event)"}}]}],t.ctorParameters=function(){return[{type:w}]},t}(),O=new s.InjectionToken("mat-chips-default-options"),C=function(){return function(t,e,n,r){this._defaultErrorStateMatcher=t,this._parentForm=e,this._parentFormGroup=n,this.ngControl=r}}(),E=Object(u.mixinErrorState)(C),S=0,M=function(){return function(t,e){this.source=t,this.value=e}}(),T=function(t){function e(e,n,r,i,o,a,u){var l=t.call(this,a,i,o,u)||this;return l._elementRef=e,l._changeDetectorRef=n,l._dir=r,l.ngControl=u,l.controlType="mat-chip-list",l._lastDestroyedChipIndex=null,l._destroyed=new c.Subject,l._uid="mat-chip-list-"+S++,l._tabIndex=0,l._userTabIndex=null,l._onTouched=function(){},l._onChange=function(){},l._multiple=!1,l._compareWith=function(t,e){return t===e},l._required=!1,l._disabled=!1,l.ariaOrientation="horizontal",l._selectable=!0,l.change=new s.EventEmitter,l.valueChange=new s.EventEmitter,l.ngControl&&(l.ngControl.valueAccessor=l),l}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"selected",{get:function(){return this.multiple?this._selectionModel.selected:this._selectionModel.selected[0]},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"role",{get:function(){return this.empty?null:"listbox"},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"multiple",{get:function(){return this._multiple},set:function(t){this._multiple=Object(i.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"compareWith",{get:function(){return this._compareWith},set:function(t){this._compareWith=t,this._selectionModel&&this._initializeSelection()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return this._value},set:function(t){this.writeValue(t),this._value=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"id",{get:function(){return this._chipInput?this._chipInput.id:this._uid},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"required",{get:function(){return this._required},set:function(t){this._required=Object(i.coerceBooleanProperty)(t),this.stateChanges.next()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"placeholder",{get:function(){return this._chipInput?this._chipInput.placeholder:this._placeholder},set:function(t){this._placeholder=t,this.stateChanges.next()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"focused",{get:function(){return this._chipInput&&this._chipInput.focused||this._hasFocusedChip()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"empty",{get:function(){return(!this._chipInput||this._chipInput.empty)&&0===this.chips.length},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"shouldLabelFloat",{get:function(){return!this.empty||this.focused},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return this.ngControl?!!this.ngControl.disabled:this._disabled},set:function(t){var e=this;this._disabled=Object(i.coerceBooleanProperty)(t),this.chips&&this.chips.forEach(function(t){return t.disabled=e._disabled})},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selectable",{get:function(){return this._selectable},set:function(t){var e=this;this._selectable=Object(i.coerceBooleanProperty)(t),this.chips&&this.chips.forEach(function(t){return t.chipListSelectable=e._selectable})},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"tabIndex",{set:function(t){this._userTabIndex=t,this._tabIndex=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"chipSelectionChanges",{get:function(){return c.merge.apply(void 0,this.chips.map(function(t){return t.selectionChange}))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"chipFocusChanges",{get:function(){return c.merge.apply(void 0,this.chips.map(function(t){return t._onFocus}))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"chipBlurChanges",{get:function(){return c.merge.apply(void 0,this.chips.map(function(t){return t._onBlur}))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"chipRemoveChanges",{get:function(){return c.merge.apply(void 0,this.chips.map(function(t){return t.destroyed}))},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){var t=this;this._keyManager=new d.FocusKeyManager(this.chips).withWrap().withVerticalOrientation().withHorizontalOrientation(this._dir?this._dir.value:"ltr"),this._dir&&this._dir.change.pipe(Object(l.takeUntil)(this._destroyed)).subscribe(function(e){return t._keyManager.withHorizontalOrientation(e)}),this._keyManager.tabOut.pipe(Object(l.takeUntil)(this._destroyed)).subscribe(function(){t._tabIndex=-1,setTimeout(function(){return t._tabIndex=t._userTabIndex||0})}),this.chips.changes.pipe(Object(l.startWith)(null),Object(l.takeUntil)(this._destroyed)).subscribe(function(){t._resetChips(),t._initializeSelection(),t._updateTabIndex(),t._updateFocusForDestroyedChips(),t.stateChanges.next()})},e.prototype.ngOnInit=function(){this._selectionModel=new f.SelectionModel(this.multiple,void 0,!1),this.stateChanges.next()},e.prototype.ngDoCheck=function(){this.ngControl&&this.updateErrorState()},e.prototype.ngOnDestroy=function(){this._destroyed.next(),this._destroyed.complete(),this.stateChanges.complete(),this._dropSubscriptions()},e.prototype.registerInput=function(t){this._chipInput=t},e.prototype.setDescribedByIds=function(t){this._ariaDescribedby=t.join(" ")},e.prototype.writeValue=function(t){this.chips&&this._setSelectionByValue(t,!1)},e.prototype.registerOnChange=function(t){this._onChange=t},e.prototype.registerOnTouched=function(t){this._onTouched=t},e.prototype.setDisabledState=function(t){this.disabled=t,this.stateChanges.next()},e.prototype.onContainerClick=function(t){this._originatesFromChip(t)||this.focus()},e.prototype.focus=function(){this.disabled||this._chipInput&&this._chipInput.focused||(this.chips.length>0?(this._keyManager.setFirstItemActive(),this.stateChanges.next()):(this._focusInput(),this.stateChanges.next()))},e.prototype._focusInput=function(){this._chipInput&&this._chipInput.focus()},e.prototype._keydown=function(t){var e=t.target;t.keyCode===o.BACKSPACE&&this._isInputEmpty(e)?(this._keyManager.setLastItemActive(),t.preventDefault()):e&&e.classList.contains("mat-chip")&&(t.keyCode===o.HOME?(this._keyManager.setFirstItemActive(),t.preventDefault()):t.keyCode===o.END?(this._keyManager.setLastItemActive(),t.preventDefault()):this._keyManager.onKeydown(t),this.stateChanges.next())},e.prototype._updateTabIndex=function(){this._tabIndex=this._userTabIndex||(0===this.chips.length?-1:0)},e.prototype._updateFocusForDestroyedChips=function(){if(null!=this._lastDestroyedChipIndex&&this.chips.length){var t=Math.min(this._lastDestroyedChipIndex,this.chips.length-1);this._keyManager.setActiveItem(t)}this._lastDestroyedChipIndex=null},e.prototype._isValidIndex=function(t){return t>=0&&t<this.chips.length},e.prototype._isInputEmpty=function(t){return!(!t||"input"!==t.nodeName.toLowerCase())&&!t.value},e.prototype._setSelectionByValue=function(t,e){var n=this;if(void 0===e&&(e=!0),this._clearSelection(),this.chips.forEach(function(t){return t.deselect()}),Array.isArray(t))t.forEach(function(t){return n._selectValue(t,e)}),this._sortValues();else{var r=this._selectValue(t,e);r&&e&&this._keyManager.setActiveItem(r)}},e.prototype._selectValue=function(t,e){var n=this;void 0===e&&(e=!0);var r=this.chips.find(function(e){return null!=e.value&&n._compareWith(e.value,t)});return r&&(e?r.selectViaInteraction():r.select(),this._selectionModel.select(r)),r},e.prototype._initializeSelection=function(){var t=this;Promise.resolve().then(function(){(t.ngControl||t._value)&&(t._setSelectionByValue(t.ngControl?t.ngControl.value:t._value,!1),t.stateChanges.next())})},e.prototype._clearSelection=function(t){this._selectionModel.clear(),this.chips.forEach(function(e){e!==t&&e.deselect()}),this.stateChanges.next()},e.prototype._sortValues=function(){var t=this;this._multiple&&(this._selectionModel.clear(),this.chips.forEach(function(e){e.selected&&t._selectionModel.select(e)}),this.stateChanges.next())},e.prototype._propagateChanges=function(t){var e=null;e=Array.isArray(this.selected)?this.selected.map(function(t){return t.value}):this.selected?this.selected.value:t,this._value=e,this.change.emit(new M(this,e)),this.valueChange.emit(e),this._onChange(e),this._changeDetectorRef.markForCheck()},e.prototype._blur=function(){var t=this;this._hasFocusedChip()||this._keyManager.setActiveItem(-1),this.disabled||(this._chipInput?setTimeout(function(){t.focused||t._markAsTouched()}):this._markAsTouched())},e.prototype._markAsTouched=function(){this._onTouched(),this._changeDetectorRef.markForCheck(),this.stateChanges.next()},e.prototype._resetChips=function(){this._dropSubscriptions(),this._listenToChipsFocus(),this._listenToChipsSelection(),this._listenToChipsRemoved()},e.prototype._dropSubscriptions=function(){this._chipFocusSubscription&&(this._chipFocusSubscription.unsubscribe(),this._chipFocusSubscription=null),this._chipBlurSubscription&&(this._chipBlurSubscription.unsubscribe(),this._chipBlurSubscription=null),this._chipSelectionSubscription&&(this._chipSelectionSubscription.unsubscribe(),this._chipSelectionSubscription=null),this._chipRemoveSubscription&&(this._chipRemoveSubscription.unsubscribe(),this._chipRemoveSubscription=null)},e.prototype._listenToChipsSelection=function(){var t=this;this._chipSelectionSubscription=this.chipSelectionChanges.subscribe(function(e){e.source.selected?t._selectionModel.select(e.source):t._selectionModel.deselect(e.source),t.multiple||t.chips.forEach(function(e){!t._selectionModel.isSelected(e)&&e.selected&&e.deselect()}),e.isUserInput&&t._propagateChanges()})},e.prototype._listenToChipsFocus=function(){var t=this;this._chipFocusSubscription=this.chipFocusChanges.subscribe(function(e){var n=t.chips.toArray().indexOf(e.chip);t._isValidIndex(n)&&t._keyManager.updateActiveItemIndex(n),t.stateChanges.next()}),this._chipBlurSubscription=this.chipBlurChanges.subscribe(function(){t._blur(),t.stateChanges.next()})},e.prototype._listenToChipsRemoved=function(){var t=this;this._chipRemoveSubscription=this.chipRemoveChanges.subscribe(function(e){var n=e.chip,r=t.chips.toArray().indexOf(e.chip);t._isValidIndex(r)&&n._hasFocus&&(t._lastDestroyedChipIndex=r)})},e.prototype._originatesFromChip=function(t){for(var e=t.target;e&&e!==this._elementRef.nativeElement;){if(e.classList.contains("mat-chip"))return!0;e=e.parentElement}return!1},e.prototype._hasFocusedChip=function(){return this.chips.some(function(t){return t._hasFocus})},e.decorators=[{type:s.Component,args:[{selector:"mat-chip-list",template:'<div class="mat-chip-list-wrapper"><ng-content></ng-content></div>',exportAs:"matChipList",host:{"[attr.tabindex]":"disabled ? null : _tabIndex","[attr.aria-describedby]":"_ariaDescribedby || null","[attr.aria-required]":"required.toString()","[attr.aria-disabled]":"disabled.toString()","[attr.aria-invalid]":"errorState","[attr.aria-multiselectable]":"multiple","[attr.role]":"role","[class.mat-chip-list-disabled]":"disabled","[class.mat-chip-list-invalid]":"errorState","[class.mat-chip-list-required]":"required","[attr.aria-orientation]":"ariaOrientation",class:"mat-chip-list","(focus)":"focus()","(blur)":"_blur()","(keydown)":"_keydown($event)","[id]":"_uid"},providers:[{provide:m.MatFormFieldControl,useExisting:e}],styles:[".mat-chip{position:relative;overflow:hidden;box-sizing:border-box;-webkit-tap-highlight-color:transparent;transform:translateZ(0)}.mat-standard-chip{transition:box-shadow 280ms cubic-bezier(.4,0,.2,1);display:inline-flex;padding:7px 12px;border-radius:16px;align-items:center;cursor:default;min-height:32px;height:1px}.mat-standard-chip .mat-chip-remove.mat-icon{width:18px;height:18px}.mat-standard-chip::after{top:0;left:0;right:0;bottom:0;position:absolute;border-radius:inherit;opacity:0;content:'';pointer-events:none;transition:opacity .2s cubic-bezier(.35,0,.25,1)}.mat-standard-chip:hover::after{opacity:.12}.mat-standard-chip:focus{outline:0}.mat-standard-chip:focus::after{opacity:.16}@media (-ms-high-contrast:active){.mat-standard-chip{outline:solid 1px}.mat-standard-chip:focus{outline:dotted 2px}}.mat-standard-chip.mat-chip-disabled::after{opacity:0}.mat-standard-chip.mat-chip-disabled .mat-chip-remove,.mat-standard-chip.mat-chip-disabled .mat-chip-trailing-icon{cursor:default}.mat-standard-chip.mat-chip-with-avatar,.mat-standard-chip.mat-chip-with-trailing-icon.mat-chip-with-avatar{padding-top:0;padding-bottom:0}.mat-standard-chip.mat-chip-with-trailing-icon.mat-chip-with-avatar{padding-right:8px;padding-left:0}[dir=rtl] .mat-standard-chip.mat-chip-with-trailing-icon.mat-chip-with-avatar{padding-left:8px;padding-right:0}.mat-standard-chip.mat-chip-with-trailing-icon{padding-top:7px;padding-bottom:7px;padding-right:8px;padding-left:12px}[dir=rtl] .mat-standard-chip.mat-chip-with-trailing-icon{padding-left:8px;padding-right:12px}.mat-standard-chip.mat-chip-with-avatar{padding-left:0;padding-right:12px}[dir=rtl] .mat-standard-chip.mat-chip-with-avatar{padding-right:0;padding-left:12px}.mat-standard-chip .mat-chip-avatar{width:24px;height:24px;margin-right:8px;margin-left:4px}[dir=rtl] .mat-standard-chip .mat-chip-avatar{margin-left:8px;margin-right:4px}.mat-standard-chip .mat-chip-remove,.mat-standard-chip .mat-chip-trailing-icon{width:18px;height:18px;cursor:pointer}.mat-standard-chip .mat-chip-remove,.mat-standard-chip .mat-chip-trailing-icon{margin-left:8px;margin-right:0}[dir=rtl] .mat-standard-chip .mat-chip-remove,[dir=rtl] .mat-standard-chip .mat-chip-trailing-icon{margin-right:8px;margin-left:0}.mat-chip-list-wrapper{display:flex;flex-direction:row;flex-wrap:wrap;align-items:center;margin:-4px}.mat-chip-list-wrapper .mat-standard-chip,.mat-chip-list-wrapper input.mat-input-element{margin:4px}.mat-chip-list-stacked .mat-chip-list-wrapper{flex-direction:column;align-items:flex-start}.mat-chip-list-stacked .mat-chip-list-wrapper .mat-standard-chip{width:100%}.mat-chip-avatar{border-radius:50%;justify-content:center;align-items:center;display:flex;overflow:hidden;object-fit:cover}input.mat-chip-input{width:150px;margin:4px;flex:1 0 150px}"],encapsulation:s.ViewEncapsulation.None,changeDetection:s.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:s.ElementRef},{type:s.ChangeDetectorRef},{type:p.Directionality,decorators:[{type:s.Optional}]},{type:h.NgForm,decorators:[{type:s.Optional}]},{type:h.FormGroupDirective,decorators:[{type:s.Optional}]},{type:u.ErrorStateMatcher},{type:h.NgControl,decorators:[{type:s.Optional},{type:s.Self}]}]},e.propDecorators={errorStateMatcher:[{type:s.Input}],multiple:[{type:s.Input}],compareWith:[{type:s.Input}],value:[{type:s.Input}],required:[{type:s.Input}],placeholder:[{type:s.Input}],disabled:[{type:s.Input}],ariaOrientation:[{type:s.Input,args:["aria-orientation"]}],selectable:[{type:s.Input}],tabIndex:[{type:s.Input}],change:[{type:s.Output}],valueChange:[{type:s.Output}],chips:[{type:s.ContentChildren,args:[w]}]},e}(E),k=0,A=function(){function t(t,e){this._elementRef=t,this._defaultOptions=e,this.focused=!1,this._addOnBlur=!1,this.separatorKeyCodes=this._defaultOptions.separatorKeyCodes,this.chipEnd=new s.EventEmitter,this.placeholder="",this.id="mat-chip-list-input-"+k++,this._disabled=!1,this._inputElement=this._elementRef.nativeElement}return Object.defineProperty(t.prototype,"chipList",{set:function(t){t&&(this._chipList=t,this._chipList.registerInput(this))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"addOnBlur",{get:function(){return this._addOnBlur},set:function(t){this._addOnBlur=Object(i.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this._disabled||this._chipList&&this._chipList.disabled},set:function(t){this._disabled=Object(i.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"empty",{get:function(){return!this._inputElement.value},enumerable:!0,configurable:!0}),t.prototype.ngOnChanges=function(){this._chipList.stateChanges.next()},t.prototype._keydown=function(t){this._emitChipEnd(t)},t.prototype._blur=function(){this.addOnBlur&&this._emitChipEnd(),this.focused=!1,this._chipList.focused||this._chipList._blur(),this._chipList.stateChanges.next()},t.prototype._focus=function(){this.focused=!0,this._chipList.stateChanges.next()},t.prototype._emitChipEnd=function(t){!this._inputElement.value&&t&&this._chipList._keydown(t),t&&!this._isSeparatorKey(t)||(this.chipEnd.emit({input:this._inputElement,value:this._inputElement.value}),t&&t.preventDefault())},t.prototype._onInput=function(){this._chipList.stateChanges.next()},t.prototype.focus=function(){this._inputElement.focus()},t.prototype._isSeparatorKey=function(t){if(Object(o.hasModifierKey)(t))return!1;var e=this.separatorKeyCodes,n=t.keyCode;return Array.isArray(e)?e.indexOf(n)>-1:e.has(n)},t.decorators=[{type:s.Directive,args:[{selector:"input[matChipInputFor]",exportAs:"matChipInput, matChipInputFor",host:{class:"mat-chip-input mat-input-element","(keydown)":"_keydown($event)","(blur)":"_blur()","(focus)":"_focus()","(input)":"_onInput()","[id]":"id","[attr.disabled]":"disabled || null","[attr.placeholder]":"placeholder || null","[attr.aria-invalid]":"_chipList && _chipList.ngControl ? _chipList.ngControl.invalid : null"}}]}],t.ctorParameters=function(){return[{type:s.ElementRef},{type:void 0,decorators:[{type:s.Inject,args:[O]}]}]},t.propDecorators={chipList:[{type:s.Input,args:["matChipInputFor"]}],addOnBlur:[{type:s.Input,args:["matChipInputAddOnBlur"]}],separatorKeyCodes:[{type:s.Input,args:["matChipInputSeparatorKeyCodes"]}],chipEnd:[{type:s.Output,args:["matChipInputTokenEnd"]}],placeholder:[{type:s.Input}],id:[{type:s.Input}],disabled:[{type:s.Input}]},t}(),D=[T,w,A,j,b,x],I={separatorKeyCodes:[o.ENTER]},P=function(){function t(){}return t.decorators=[{type:s.NgModule,args:[{exports:D,declarations:D,providers:[u.ErrorStateMatcher,{provide:O,useValue:I}]}]}],t}()},"./node_modules/@angular/material/esm5/core.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"AnimationCurves",function(){return m}),n.d(e,"AnimationDurations",function(){return g}),n.d(e,"MatCommonModule",function(){return v}),n.d(e,"MATERIAL_SANITY_CHECKS",function(){return y}),n.d(e,"mixinDisabled",function(){return b}),n.d(e,"mixinColor",function(){return x}),n.d(e,"mixinDisableRipple",function(){return w}),n.d(e,"mixinTabIndex",function(){return j}),n.d(e,"mixinErrorState",function(){return O}),n.d(e,"mixinInitialized",function(){return C}),n.d(e,"NativeDateModule",function(){return B}),n.d(e,"MatNativeDateModule",function(){return z}),n.d(e,"MAT_DATE_LOCALE_FACTORY",function(){return S}),n.d(e,"MAT_DATE_LOCALE",function(){return E}),n.d(e,"MAT_DATE_LOCALE_PROVIDER",function(){return M}),n.d(e,"DateAdapter",function(){return T}),n.d(e,"MAT_DATE_FORMATS",function(){return k}),n.d(e,"NativeDateAdapter",function(){return F}),n.d(e,"MAT_NATIVE_DATE_FORMATS",function(){return L}),n.d(e,"ShowOnDirtyErrorStateMatcher",function(){return H}),n.d(e,"ErrorStateMatcher",function(){return U}),n.d(e,"MAT_HAMMER_OPTIONS",function(){return W}),n.d(e,"GestureConfig",function(){return Y}),n.d(e,"setLines",function(){return X}),n.d(e,"MatLine",function(){return K}),n.d(e,"MatLineSetter",function(){return Z}),n.d(e,"MatLineModule",function(){return $}),n.d(e,"MatOptionModule",function(){return bt}),n.d(e,"_countGroupLabelsBeforeOption",function(){return _t}),n.d(e,"_getOptionScrollPosition",function(){return vt}),n.d(e,"MatOptionSelectionChange",function(){return mt}),n.d(e,"MAT_OPTION_PARENT_COMPONENT",function(){return gt}),n.d(e,"MatOption",function(){return yt}),n.d(e,"MatOptgroupBase",function(){return lt}),n.d(e,"_MatOptgroupMixinBase",function(){return dt}),n.d(e,"MatOptgroup",function(){return ft}),n.d(e,"MAT_LABEL_GLOBAL_OPTIONS",function(){return xt}),n.d(e,"MatRippleModule",function(){return st}),n.d(e,"MAT_RIPPLE_GLOBAL_OPTIONS",function(){return ot}),n.d(e,"MatRipple",function(){return at}),n.d(e,"RippleState",function(){return J}),n.d(e,"RippleRef",function(){return tt}),n.d(e,"defaultRippleAnimationConfig",function(){return et}),n.d(e,"RippleRenderer",function(){return it}),n.d(e,"MatPseudoCheckboxModule",function(){return ct}),n.d(e,"MatPseudoCheckbox",function(){return ut}),n.d(e,"JAN",function(){return wt}),n.d(e,"FEB",function(){return jt}),n.d(e,"MAR",function(){return Ot}),n.d(e,"APR",function(){return Ct}),n.d(e,"MAY",function(){return Et}),n.d(e,"JUN",function(){return St}),n.d(e,"JUL",function(){return Mt}),n.d(e,"AUG",function(){return Tt}),n.d(e,"SEP",function(){return kt}),n.d(e,"OCT",function(){return At}),n.d(e,"NOV",function(){return Dt}),n.d(e,"DEC",function(){return It}),n.d(e,"ɵa1",function(){return _});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),o=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),a=n("./node_modules/tslib/tslib.es6.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/rxjs/_esm5/index.js"),c=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),l=n("./node_modules/rxjs/_esm5/operators/index.js"),d=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),p=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),f=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),h=n("./node_modules/@angular/common/fesm5/common.js"),m=function(){function t(){}return t.STANDARD_CURVE="cubic-bezier(0.4,0.0,0.2,1)",t.DECELERATION_CURVE="cubic-bezier(0.0,0.0,0.2,1)",t.ACCELERATION_CURVE="cubic-bezier(0.4,0.0,1,1)",t.SHARP_CURVE="cubic-bezier(0.4,0.0,0.6,1)",t}(),g=function(){function t(){}return t.COMPLEX="375ms",t.ENTERING="225ms",t.EXITING="195ms",t}(),y=new r.InjectionToken("mat-sanity-checks",{providedIn:"root",factory:_});function _(){return!0}var v=function(){function t(t,e){this._sanityChecksEnabled=t,this._hammerLoader=e,this._hasDoneGlobalChecks=!1,this._hasCheckedHammer=!1,this._document="object"==typeof document&&document?document:null,this._window="object"==typeof window&&window?window:null,this._areChecksEnabled()&&!this._hasDoneGlobalChecks&&(this._checkDoctypeIsDefined(),this._checkThemeIsPresent(),this._hasDoneGlobalChecks=!0)}return t.prototype._areChecksEnabled=function(){return this._sanityChecksEnabled&&Object(r.isDevMode)()&&!this._isTestEnv()},t.prototype._isTestEnv=function(){var t=this._window;return t&&(t.__karma__||t.jasmine)},t.prototype._checkDoctypeIsDefined=function(){this._document&&!this._document.doctype&&console.warn("Current document does not have a doctype. This may cause some Angular Material components not to behave as expected.")},t.prototype._checkThemeIsPresent=function(){if(this._document&&this._document.body&&"function"==typeof getComputedStyle){var t=this._document.createElement("div");t.classList.add("mat-theme-loaded-marker"),this._document.body.appendChild(t);var e=getComputedStyle(t);e&&"none"!==e.display&&console.warn("Could not find Angular Material core theme. Most Material components may not work as expected. For more info refer to the theming guide: https://material.angular.io/guide/theming"),this._document.body.removeChild(t)}},t.prototype._checkHammerIsAvailable=function(){!this._hasCheckedHammer&&this._window&&(!this._areChecksEnabled()||this._window.Hammer||this._hammerLoader||console.warn("Could not find HammerJS. Certain Angular Material components may not work correctly."),this._hasCheckedHammer=!0)},t.decorators=[{type:r.NgModule,args:[{imports:[o.BidiModule],exports:[o.BidiModule]}]}],t.ctorParameters=function(){return[{type:Boolean,decorators:[{type:r.Optional},{type:r.Inject,args:[y]}]},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[i.HAMMER_LOADER]}]}]},t}();function b(t){return function(t){function e(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];var r=t.apply(this,e)||this;return r._disabled=!1,r}return Object(a.__extends)(e,t),Object.defineProperty(e.prototype,"disabled",{get:function(){return this._disabled},set:function(t){this._disabled=Object(s.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e}(t)}function x(t,e){return function(t){function n(){for(var n=[],r=0;r<arguments.length;r++)n[r]=arguments[r];var i=t.apply(this,n)||this;return i.color=e,i}return Object(a.__extends)(n,t),Object.defineProperty(n.prototype,"color",{get:function(){return this._color},set:function(t){var n=t||e;n!==this._color&&(this._color&&this._elementRef.nativeElement.classList.remove("mat-"+this._color),n&&this._elementRef.nativeElement.classList.add("mat-"+n),this._color=n)},enumerable:!0,configurable:!0}),n}(t)}function w(t){return function(t){function e(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];var r=t.apply(this,e)||this;return r._disableRipple=!1,r}return Object(a.__extends)(e,t),Object.defineProperty(e.prototype,"disableRipple",{get:function(){return this._disableRipple},set:function(t){this._disableRipple=Object(s.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e}(t)}function j(t,e){return void 0===e&&(e=0),function(t){function n(){for(var n=[],r=0;r<arguments.length;r++)n[r]=arguments[r];var i=t.apply(this,n)||this;return i._tabIndex=e,i}return Object(a.__extends)(n,t),Object.defineProperty(n.prototype,"tabIndex",{get:function(){return this.disabled?-1:this._tabIndex},set:function(t){this._tabIndex=null!=t?t:e},enumerable:!0,configurable:!0}),n}(t)}function O(t){return function(t){function e(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];var r=t.apply(this,e)||this;return r.errorState=!1,r.stateChanges=new u.Subject,r}return Object(a.__extends)(e,t),e.prototype.updateErrorState=function(){var t=this.errorState,e=this._parentFormGroup||this._parentForm,n=this.errorStateMatcher||this._defaultErrorStateMatcher,r=this.ngControl?this.ngControl.control:null,i=n.isErrorState(r,e);i!==t&&(this.errorState=i,this.stateChanges.next())},e}(t)}function C(t){return function(t){function e(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];var r=t.apply(this,e)||this;return r._isInitialized=!1,r._pendingSubscribers=[],r.initialized=new u.Observable(function(t){r._isInitialized?r._notifySubscriber(t):r._pendingSubscribers.push(t)}),r}return Object(a.__extends)(e,t),e.prototype._markInitialized=function(){if(this._isInitialized)throw Error("This directive has already been marked as initialized and should not be called twice.");this._isInitialized=!0,this._pendingSubscribers.forEach(this._notifySubscriber),this._pendingSubscribers=null},e.prototype._notifySubscriber=function(t){t.next(),t.complete()},e}(t)}var E=new r.InjectionToken("MAT_DATE_LOCALE",{providedIn:"root",factory:S});function S(){return Object(r.inject)(r.LOCALE_ID)}var M={provide:E,useExisting:r.LOCALE_ID},T=function(){function t(){this._localeChanges=new u.Subject}return Object.defineProperty(t.prototype,"localeChanges",{get:function(){return this._localeChanges},enumerable:!0,configurable:!0}),t.prototype.deserialize=function(t){return null==t||this.isDateInstance(t)&&this.isValid(t)?t:this.invalid()},t.prototype.setLocale=function(t){this.locale=t,this._localeChanges.next()},t.prototype.compareDate=function(t,e){return this.getYear(t)-this.getYear(e)||this.getMonth(t)-this.getMonth(e)||this.getDate(t)-this.getDate(e)},t.prototype.sameDate=function(t,e){if(t&&e){var n=this.isValid(t),r=this.isValid(e);return n&&r?!this.compareDate(t,e):n==r}return t==e},t.prototype.clampDate=function(t,e,n){return e&&this.compareDate(t,e)<0?e:n&&this.compareDate(t,n)>0?n:t},t}(),k=new r.InjectionToken("mat-date-formats"),A="undefined"!=typeof Intl,D={long:["January","February","March","April","May","June","July","August","September","October","November","December"],short:["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"],narrow:["J","F","M","A","M","J","J","A","S","O","N","D"]},I=N(31,function(t){return String(t+1)}),P={long:["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"],short:["Sun","Mon","Tue","Wed","Thu","Fri","Sat"],narrow:["S","M","T","W","T","F","S"]},R=/^\d{4}-\d{2}-\d{2}(?:T\d{2}:\d{2}:\d{2}(?:\.\d+)?(?:Z|(?:(?:\+|-)\d{2}:\d{2}))?)?$/;function N(t,e){for(var n=Array(t),r=0;r<t;r++)n[r]=e(r);return n}var F=function(t){function e(e,n){var r=t.call(this)||this;return r.useUtcForDisplay=!0,t.prototype.setLocale.call(r,e),r.useUtcForDisplay=!n.TRIDENT,r._clampDate=n.TRIDENT||n.EDGE,r}return Object(a.__extends)(e,t),e.prototype.getYear=function(t){return t.getFullYear()},e.prototype.getMonth=function(t){return t.getMonth()},e.prototype.getDate=function(t){return t.getDate()},e.prototype.getDayOfWeek=function(t){return t.getDay()},e.prototype.getMonthNames=function(t){var e=this;if(A){var n=new Intl.DateTimeFormat(this.locale,{month:t,timeZone:"utc"});return N(12,function(t){return e._stripDirectionalityCharacters(e._format(n,new Date(2017,t,1)))})}return D[t]},e.prototype.getDateNames=function(){var t=this;if(A){var e=new Intl.DateTimeFormat(this.locale,{day:"numeric",timeZone:"utc"});return N(31,function(n){return t._stripDirectionalityCharacters(t._format(e,new Date(2017,0,n+1)))})}return I},e.prototype.getDayOfWeekNames=function(t){var e=this;if(A){var n=new Intl.DateTimeFormat(this.locale,{weekday:t,timeZone:"utc"});return N(7,function(t){return e._stripDirectionalityCharacters(e._format(n,new Date(2017,0,t+1)))})}return P[t]},e.prototype.getYearName=function(t){if(A){var e=new Intl.DateTimeFormat(this.locale,{year:"numeric",timeZone:"utc"});return this._stripDirectionalityCharacters(this._format(e,t))}return String(this.getYear(t))},e.prototype.getFirstDayOfWeek=function(){return 0},e.prototype.getNumDaysInMonth=function(t){return this.getDate(this._createDateWithOverflow(this.getYear(t),this.getMonth(t)+1,0))},e.prototype.clone=function(t){return new Date(t.getTime())},e.prototype.createDate=function(t,e,n){if(e<0||e>11)throw Error('Invalid month index "'+e+'". Month index has to be between 0 and 11.');if(n<1)throw Error('Invalid date "'+n+'". Date has to be greater than 0.');var r=this._createDateWithOverflow(t,e,n);if(r.getMonth()!=e)throw Error('Invalid date "'+n+'" for month with index "'+e+'".');return r},e.prototype.today=function(){return new Date},e.prototype.parse=function(t){return"number"==typeof t?new Date(t):t?new Date(Date.parse(t)):null},e.prototype.format=function(t,e){if(!this.isValid(t))throw Error("NativeDateAdapter: Cannot format invalid date.");if(A){this._clampDate&&(t.getFullYear()<1||t.getFullYear()>9999)&&(t=this.clone(t)).setFullYear(Math.max(1,Math.min(9999,t.getFullYear()))),e=Object(a.__assign)({},e,{timeZone:"utc"});var n=new Intl.DateTimeFormat(this.locale,e);return this._stripDirectionalityCharacters(this._format(n,t))}return this._stripDirectionalityCharacters(t.toDateString())},e.prototype.addCalendarYears=function(t,e){return this.addCalendarMonths(t,12*e)},e.prototype.addCalendarMonths=function(t,e){var n=this._createDateWithOverflow(this.getYear(t),this.getMonth(t)+e,this.getDate(t));return this.getMonth(n)!=((this.getMonth(t)+e)%12+12)%12&&(n=this._createDateWithOverflow(this.getYear(n),this.getMonth(n),0)),n},e.prototype.addCalendarDays=function(t,e){return this._createDateWithOverflow(this.getYear(t),this.getMonth(t),this.getDate(t)+e)},e.prototype.toIso8601=function(t){return[t.getUTCFullYear(),this._2digit(t.getUTCMonth()+1),this._2digit(t.getUTCDate())].join("-")},e.prototype.deserialize=function(e){if("string"==typeof e){if(!e)return null;if(R.test(e)){var n=new Date(e);if(this.isValid(n))return n}}return t.prototype.deserialize.call(this,e)},e.prototype.isDateInstance=function(t){return t instanceof Date},e.prototype.isValid=function(t){return!isNaN(t.getTime())},e.prototype.invalid=function(){return new Date(NaN)},e.prototype._createDateWithOverflow=function(t,e,n){var r=new Date(t,e,n);return t>=0&&t<100&&r.setFullYear(this.getYear(r)-1900),r},e.prototype._2digit=function(t){return("00"+t).slice(-2)},e.prototype._stripDirectionalityCharacters=function(t){return t.replace(/[\u200e\u200f]/g,"")},e.prototype._format=function(t,e){var n=new Date(Date.UTC(e.getFullYear(),e.getMonth(),e.getDate(),e.getHours(),e.getMinutes(),e.getSeconds(),e.getMilliseconds()));return t.format(n)},e.decorators=[{type:r.Injectable}],e.ctorParameters=function(){return[{type:String,decorators:[{type:r.Optional},{type:r.Inject,args:[E]}]},{type:c.Platform}]},e}(T),L={parse:{dateInput:null},display:{dateInput:{year:"numeric",month:"numeric",day:"numeric"},monthYearLabel:{year:"numeric",month:"short"},dateA11yLabel:{year:"numeric",month:"long",day:"numeric"},monthYearA11yLabel:{year:"numeric",month:"long"}}},B=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[c.PlatformModule],providers:[{provide:T,useClass:F}]}]}],t}(),V=L,z=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[B],providers:[{provide:k,useValue:V}]}]}],t}(),H=function(){function t(){}return t.prototype.isErrorState=function(t,e){return!!(t&&t.invalid&&(t.dirty||e&&e.submitted))},t.decorators=[{type:r.Injectable}],t}(),U=function(){function t(){}return t.prototype.isErrorState=function(t,e){return!!(t&&t.invalid&&(t.touched||e&&e.submitted))},t.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(),W=new r.InjectionToken("MAT_HAMMER_OPTIONS"),q=["longpress","slide","slidestart","slideend","slideright","slideleft"],G={on:function(){},off:function(){}},Y=function(t){function e(e,n){var r=t.call(this)||this;return r._hammerOptions=e,r.events=q,n&&n._checkHammerIsAvailable(),r}return Object(a.__extends)(e,t),e.prototype.buildHammer=function(t){var e="undefined"!=typeof window?window.Hammer:null;if(!e)return G;var n=new e(t,this._hammerOptions||void 0),r=new e.Pan,i=new e.Swipe,o=new e.Press,a=this._createRecognizer(r,{event:"slide",threshold:0},i),s=this._createRecognizer(o,{event:"longpress",time:500});return r.recognizeWith(i),s.recognizeWith(a),n.add([i,o,r,a,s]),n},e.prototype._createRecognizer=function(t,e){for(var n=[],r=2;r<arguments.length;r++)n[r-2]=arguments[r];var i=new t.constructor(e);return n.push(t),n.forEach(function(t){return i.recognizeWith(t)}),i},e.decorators=[{type:r.Injectable}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[W]}]},{type:v,decorators:[{type:r.Optional}]}]},e}(i.HammerGestureConfig),K=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"[mat-line], [matLine]",host:{class:"mat-line"}}]}],t}();function X(t,e){t.changes.pipe(Object(l.startWith)(t)).subscribe(function(t){var n=t.length;Q(e,"mat-2-line",!1),Q(e,"mat-3-line",!1),Q(e,"mat-multi-line",!1),2===n||3===n?Q(e,"mat-"+n+"-line",!0):n>3&&Q(e,"mat-multi-line",!0)})}function Q(t,e,n){var r=t.nativeElement.classList;n?r.add(e):r.remove(e)}var Z=function(){return function(t,e){X(t,e)}}(),$=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[v],exports:[K,v],declarations:[K]}]}],t}(),J={FADING_IN:0,VISIBLE:1,FADING_OUT:2,HIDDEN:3};J[J.FADING_IN]="FADING_IN",J[J.VISIBLE]="VISIBLE",J[J.FADING_OUT]="FADING_OUT",J[J.HIDDEN]="HIDDEN";var tt=function(){function t(t,e,n){this._renderer=t,this.element=e,this.config=n,this.state=J.HIDDEN}return t.prototype.fadeOut=function(){this._renderer.fadeOutRipple(this)},t}(),et={enterDuration:450,exitDuration:400},nt=800,rt=Object(c.normalizePassiveListenerOptions)({passive:!0}),it=function(){function t(t,e,n,r){var i=this;this._target=t,this._ngZone=e,this._isPointerDown=!1,this._triggerEvents=new Map,this._activeRipples=new Set,this.onMousedown=function(t){var e=Object(d.isFakeMousedownFromScreenReader)(t),n=i._lastTouchStartEvent&&Date.now()<i._lastTouchStartEvent+nt;i._target.rippleDisabled||e||n||(i._isPointerDown=!0,i.fadeInRipple(t.clientX,t.clientY,i._target.rippleConfig))},this.onTouchStart=function(t){if(!i._target.rippleDisabled){i._lastTouchStartEvent=Date.now(),i._isPointerDown=!0;for(var e=t.changedTouches,n=0;n<e.length;n++)i.fadeInRipple(e[n].clientX,e[n].clientY,i._target.rippleConfig)}},this.onPointerUp=function(){i._isPointerDown&&(i._isPointerDown=!1,i._activeRipples.forEach(function(t){var e=t.state===J.VISIBLE||t.config.terminateOnPointerUp&&t.state===J.FADING_IN;!t.config.persistent&&e&&t.fadeOut()}))},r.isBrowser&&(this._containerElement=n.nativeElement,this._triggerEvents.set("mousedown",this.onMousedown).set("mouseup",this.onPointerUp).set("mouseleave",this.onPointerUp).set("touchstart",this.onTouchStart).set("touchend",this.onPointerUp).set("touchcancel",this.onPointerUp))}return t.prototype.fadeInRipple=function(t,e,n){var r=this;void 0===n&&(n={});var i=this._containerRect=this._containerRect||this._containerElement.getBoundingClientRect(),o=Object(a.__assign)({},et,n.animation);n.centered&&(t=i.left+i.width/2,e=i.top+i.height/2);var s,u,c,l,d,p,f=n.radius||(s=t,u=e,c=i,l=Math.max(Math.abs(s-c.left),Math.abs(s-c.right)),d=Math.max(Math.abs(u-c.top),Math.abs(u-c.bottom)),Math.sqrt(l*l+d*d)),h=t-i.left,m=e-i.top,g=o.enterDuration,y=document.createElement("div");y.classList.add("mat-ripple-element"),y.style.left=h-f+"px",y.style.top=m-f+"px",y.style.height=2*f+"px",y.style.width=2*f+"px",y.style.backgroundColor=n.color||null,y.style.transitionDuration=g+"ms",this._containerElement.appendChild(y),p=y,window.getComputedStyle(p).getPropertyValue("opacity"),y.style.transform="scale(1)";var _=new tt(this,y,n);return _.state=J.FADING_IN,this._activeRipples.add(_),n.persistent||(this._mostRecentTransientRipple=_),this.runTimeoutOutsideZone(function(){var t=_===r._mostRecentTransientRipple;_.state=J.VISIBLE,n.persistent||t&&r._isPointerDown||_.fadeOut()},g),_},t.prototype.fadeOutRipple=function(t){var e=this._activeRipples.delete(t);if(t===this._mostRecentTransientRipple&&(this._mostRecentTransientRipple=null),this._activeRipples.size||(this._containerRect=null),e){var n=t.element,r=Object(a.__assign)({},et,t.config.animation);n.style.transitionDuration=r.exitDuration+"ms",n.style.opacity="0",t.state=J.FADING_OUT,this.runTimeoutOutsideZone(function(){t.state=J.HIDDEN,n.parentNode.removeChild(n)},r.exitDuration)}},t.prototype.fadeOutAll=function(){this._activeRipples.forEach(function(t){return t.fadeOut()})},t.prototype.setupTriggerEvents=function(t){var e=this;t&&t!==this._triggerElement&&(this._removeTriggerEvents(),this._ngZone.runOutsideAngular(function(){e._triggerEvents.forEach(function(e,n){t.addEventListener(n,e,rt)})}),this._triggerElement=t)},t.prototype.runTimeoutOutsideZone=function(t,e){void 0===e&&(e=0),this._ngZone.runOutsideAngular(function(){return setTimeout(t,e)})},t.prototype._removeTriggerEvents=function(){var t=this;this._triggerElement&&this._triggerEvents.forEach(function(e,n){t._triggerElement.removeEventListener(n,e,rt)})},t}();var ot=new r.InjectionToken("mat-ripple-global-options"),at=function(){function t(t,e,n,r,i){this._elementRef=t,this.radius=0,this._disabled=!1,this._isInitialized=!1,this._globalOptions=r||{},this._rippleRenderer=new it(this,e,t,n),"NoopAnimations"===i&&(this._globalOptions.animation={enterDuration:0,exitDuration:0})}return Object.defineProperty(t.prototype,"disabled",{get:function(){return this._disabled},set:function(t){this._disabled=t,this._setupTriggerEventsIfEnabled()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"trigger",{get:function(){return this._trigger||this._elementRef.nativeElement},set:function(t){this._trigger=t,this._setupTriggerEventsIfEnabled()},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){this._isInitialized=!0,this._setupTriggerEventsIfEnabled()},t.prototype.ngOnDestroy=function(){this._rippleRenderer._removeTriggerEvents()},t.prototype.fadeOutAll=function(){this._rippleRenderer.fadeOutAll()},Object.defineProperty(t.prototype,"rippleConfig",{get:function(){return{centered:this.centered,radius:this.radius,color:this.color,animation:Object(a.__assign)({},this._globalOptions.animation,this.animation),terminateOnPointerUp:this._globalOptions.terminateOnPointerUp}},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"rippleDisabled",{get:function(){return this.disabled||!!this._globalOptions.disabled},enumerable:!0,configurable:!0}),t.prototype._setupTriggerEventsIfEnabled=function(){!this.disabled&&this._isInitialized&&this._rippleRenderer.setupTriggerEvents(this.trigger)},t.prototype.launch=function(t,e,n){return void 0===e&&(e=0),"number"==typeof t?this._rippleRenderer.fadeInRipple(t,e,Object(a.__assign)({},this.rippleConfig,n)):this._rippleRenderer.fadeInRipple(0,0,Object(a.__assign)({},this.rippleConfig,t))},t.decorators=[{type:r.Directive,args:[{selector:"[mat-ripple], [matRipple]",exportAs:"matRipple",host:{class:"mat-ripple","[class.mat-ripple-unbounded]":"unbounded"}}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:r.NgZone},{type:c.Platform},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[ot]}]},{type:String,decorators:[{type:r.Optional},{type:r.Inject,args:[p.ANIMATION_MODULE_TYPE]}]}]},t.propDecorators={color:[{type:r.Input,args:["matRippleColor"]}],unbounded:[{type:r.Input,args:["matRippleUnbounded"]}],centered:[{type:r.Input,args:["matRippleCentered"]}],radius:[{type:r.Input,args:["matRippleRadius"]}],animation:[{type:r.Input,args:["matRippleAnimation"]}],disabled:[{type:r.Input,args:["matRippleDisabled"]}],trigger:[{type:r.Input,args:["matRippleTrigger"]}]},t}(),st=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[v,c.PlatformModule],exports:[at,v],declarations:[at]}]}],t}(),ut=function(){function t(t){this._animationMode=t,this.state="unchecked",this.disabled=!1}return t.decorators=[{type:r.Component,args:[{encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,selector:"mat-pseudo-checkbox",styles:[".mat-pseudo-checkbox{width:16px;height:16px;border:2px solid;border-radius:2px;cursor:pointer;display:inline-block;vertical-align:middle;box-sizing:border-box;position:relative;flex-shrink:0;transition:border-color 90ms cubic-bezier(0,0,.2,.1),background-color 90ms cubic-bezier(0,0,.2,.1)}.mat-pseudo-checkbox::after{position:absolute;opacity:0;content:'';border-bottom:2px solid currentColor;transition:opacity 90ms cubic-bezier(0,0,.2,.1)}.mat-pseudo-checkbox.mat-pseudo-checkbox-checked,.mat-pseudo-checkbox.mat-pseudo-checkbox-indeterminate{border-color:transparent}._mat-animation-noopable.mat-pseudo-checkbox{transition:none;animation:none}._mat-animation-noopable.mat-pseudo-checkbox::after{transition:none}.mat-pseudo-checkbox-disabled{cursor:default}.mat-pseudo-checkbox-indeterminate::after{top:5px;left:1px;width:10px;opacity:1;border-radius:2px}.mat-pseudo-checkbox-checked::after{top:2.4px;left:1px;width:8px;height:3px;border-left:2px solid currentColor;transform:rotate(-45deg);opacity:1}"],template:"",host:{class:"mat-pseudo-checkbox","[class.mat-pseudo-checkbox-indeterminate]":'state === "indeterminate"',"[class.mat-pseudo-checkbox-checked]":'state === "checked"',"[class.mat-pseudo-checkbox-disabled]":"disabled","[class._mat-animation-noopable]":'_animationMode === "NoopAnimations"'}}]}],t.ctorParameters=function(){return[{type:String,decorators:[{type:r.Optional},{type:r.Inject,args:[p.ANIMATION_MODULE_TYPE]}]}]},t.propDecorators={state:[{type:r.Input}],disabled:[{type:r.Input}]},t}(),ct=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{exports:[ut],declarations:[ut]}]}],t}(),lt=function(){return function(){}}(),dt=b(lt),pt=0,ft=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._labelId="mat-optgroup-label-"+pt++,e}return Object(a.__extends)(e,t),e.decorators=[{type:r.Component,args:[{selector:"mat-optgroup",exportAs:"matOptgroup",template:'<label class="mat-optgroup-label" [id]="_labelId">{{ label }}<ng-content></ng-content></label><ng-content select="mat-option, ng-container"></ng-content>',encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,inputs:["disabled"],styles:[".mat-optgroup-label{white-space:nowrap;overflow:hidden;text-overflow:ellipsis;display:block;line-height:48px;height:48px;padding:0 16px;text-align:left;text-decoration:none;max-width:100%;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:default}.mat-optgroup-label[disabled]{cursor:default}[dir=rtl] .mat-optgroup-label{text-align:right}.mat-optgroup-label .mat-icon{margin-right:16px;vertical-align:middle}.mat-optgroup-label .mat-icon svg{vertical-align:top}[dir=rtl] .mat-optgroup-label .mat-icon{margin-left:16px;margin-right:0}"],host:{class:"mat-optgroup",role:"group","[class.mat-optgroup-disabled]":"disabled","[attr.aria-disabled]":"disabled.toString()","[attr.aria-labelledby]":"_labelId"}}]}],e.propDecorators={label:[{type:r.Input}]},e}(dt),ht=0,mt=function(){return function(t,e){void 0===e&&(e=!1),this.source=t,this.isUserInput=e}}(),gt=new r.InjectionToken("MAT_OPTION_PARENT_COMPONENT"),yt=function(){function t(t,e,n,i){this._element=t,this._changeDetectorRef=e,this._parent=n,this.group=i,this._selected=!1,this._active=!1,this._disabled=!1,this._mostRecentViewValue="",this.id="mat-option-"+ht++,this.onSelectionChange=new r.EventEmitter,this._stateChanges=new u.Subject}return Object.defineProperty(t.prototype,"multiple",{get:function(){return this._parent&&this._parent.multiple},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selected",{get:function(){return this._selected},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this.group&&this.group.disabled||this._disabled},set:function(t){this._disabled=Object(s.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disableRipple",{get:function(){return this._parent&&this._parent.disableRipple},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"active",{get:function(){return this._active},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"viewValue",{get:function(){return(this._getHostElement().textContent||"").trim()},enumerable:!0,configurable:!0}),t.prototype.select=function(){this._selected||(this._selected=!0,this._changeDetectorRef.markForCheck(),this._emitSelectionChangeEvent())},t.prototype.deselect=function(){this._selected&&(this._selected=!1,this._changeDetectorRef.markForCheck(),this._emitSelectionChangeEvent())},t.prototype.focus=function(){var t=this._getHostElement();"function"==typeof t.focus&&t.focus()},t.prototype.setActiveStyles=function(){this._active||(this._active=!0,this._changeDetectorRef.markForCheck())},t.prototype.setInactiveStyles=function(){this._active&&(this._active=!1,this._changeDetectorRef.markForCheck())},t.prototype.getLabel=function(){return this.viewValue},t.prototype._handleKeydown=function(t){t.keyCode!==f.ENTER&&t.keyCode!==f.SPACE||Object(f.hasModifierKey)(t)||(this._selectViaInteraction(),t.preventDefault())},t.prototype._selectViaInteraction=function(){this.disabled||(this._selected=!this.multiple||!this._selected,this._changeDetectorRef.markForCheck(),this._emitSelectionChangeEvent(!0))},t.prototype._getTabIndex=function(){return this.disabled?"-1":"0"},t.prototype._getHostElement=function(){return this._element.nativeElement},t.prototype.ngAfterViewChecked=function(){if(this._selected){var t=this.viewValue;t!==this._mostRecentViewValue&&(this._mostRecentViewValue=t,this._stateChanges.next())}},t.prototype.ngOnDestroy=function(){this._stateChanges.complete()},t.prototype._emitSelectionChangeEvent=function(t){void 0===t&&(t=!1),this.onSelectionChange.emit(new mt(this,t))},t.decorators=[{type:r.Component,args:[{selector:"mat-option",exportAs:"matOption",host:{role:"option","[attr.tabindex]":"_getTabIndex()","[class.mat-selected]":"selected","[class.mat-option-multiple]":"multiple","[class.mat-active]":"active","[id]":"id","[attr.aria-selected]":"selected.toString()","[attr.aria-disabled]":"disabled.toString()","[class.mat-option-disabled]":"disabled","(click)":"_selectViaInteraction()","(keydown)":"_handleKeydown($event)",class:"mat-option"},styles:[".mat-option{white-space:nowrap;overflow:hidden;text-overflow:ellipsis;display:block;line-height:48px;height:48px;padding:0 16px;text-align:left;text-decoration:none;max-width:100%;position:relative;cursor:pointer;outline:0;display:flex;flex-direction:row;max-width:100%;box-sizing:border-box;align-items:center;-webkit-tap-highlight-color:transparent}.mat-option[disabled]{cursor:default}[dir=rtl] .mat-option{text-align:right}.mat-option .mat-icon{margin-right:16px;vertical-align:middle}.mat-option .mat-icon svg{vertical-align:top}[dir=rtl] .mat-option .mat-icon{margin-left:16px;margin-right:0}.mat-option[aria-disabled=true]{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:default}.mat-optgroup .mat-option:not(.mat-option-multiple){padding-left:32px}[dir=rtl] .mat-optgroup .mat-option:not(.mat-option-multiple){padding-left:16px;padding-right:32px}@media (-ms-high-contrast:active){.mat-option{margin:0 1px}.mat-option.mat-active{border:solid 1px currentColor;margin:0}}.mat-option-text{display:inline-block;flex-grow:1;overflow:hidden;text-overflow:ellipsis}.mat-option .mat-option-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}@media (-ms-high-contrast:active){.mat-option .mat-option-ripple{opacity:.5}}.mat-option-pseudo-checkbox{margin-right:8px}[dir=rtl] .mat-option-pseudo-checkbox{margin-left:8px;margin-right:0}"],template:'<mat-pseudo-checkbox *ngIf="multiple" class="mat-option-pseudo-checkbox" [state]="selected ? \'checked\' : \'\'" [disabled]="disabled"></mat-pseudo-checkbox><span class="mat-option-text"><ng-content></ng-content></span><div class="mat-option-ripple" mat-ripple [matRippleTrigger]="_getHostElement()" [matRippleDisabled]="disabled || disableRipple"></div>',encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:r.ChangeDetectorRef},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[gt]}]},{type:ft,decorators:[{type:r.Optional}]}]},t.propDecorators={value:[{type:r.Input}],id:[{type:r.Input}],disabled:[{type:r.Input}],onSelectionChange:[{type:r.Output}]},t}();function _t(t,e,n){if(n.length){for(var r=e.toArray(),i=n.toArray(),o=0,a=0;a<t+1;a++)r[a].group&&r[a].group===i[o]&&o++;return o}return 0}function vt(t,e,n,r){var i=t*e;return i<n?i:i+e>n+r?Math.max(0,i-r+e):n}var bt=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[st,h.CommonModule,ct],exports:[yt,ft],declarations:[yt,ft]}]}],t}(),xt=new r.InjectionToken("mat-label-global-options"),wt=0,jt=1,Ot=2,Ct=3,Et=4,St=5,Mt=6,Tt=7,kt=8,At=9,Dt=10,It=11},"./node_modules/@angular/material/esm5/datepicker.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatDatepickerModule",function(){return q}),n.d(e,"MatCalendarHeader",function(){return M}),n.d(e,"MatCalendar",function(){return T}),n.d(e,"MatCalendarCell",function(){return j}),n.d(e,"MatCalendarBody",function(){return O}),n.d(e,"MAT_DATEPICKER_SCROLL_STRATEGY_FACTORY",function(){return I}),n.d(e,"MAT_DATEPICKER_SCROLL_STRATEGY",function(){return D}),n.d(e,"MAT_DATEPICKER_SCROLL_STRATEGY_FACTORY_PROVIDER",function(){return P}),n.d(e,"MatDatepickerContentBase",function(){return R}),n.d(e,"_MatDatepickerContentMixinBase",function(){return N}),n.d(e,"MatDatepickerContent",function(){return F}),n.d(e,"MatDatepicker",function(){return L}),n.d(e,"matDatepickerAnimations",function(){return k}),n.d(e,"MAT_DATEPICKER_VALUE_ACCESSOR",function(){return B}),n.d(e,"MAT_DATEPICKER_VALIDATORS",function(){return V}),n.d(e,"MatDatepickerInputEvent",function(){return z}),n.d(e,"MatDatepickerInput",function(){return H}),n.d(e,"MatDatepickerIntl",function(){return w}),n.d(e,"MatDatepickerToggleIcon",function(){return U}),n.d(e,"MatDatepickerToggle",function(){return W}),n.d(e,"MatMonthView",function(){return C}),n.d(e,"MatYearView",function(){return S}),n.d(e,"ɵa34",function(){return E});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/rxjs/_esm5/index.js"),o=n("./node_modules/rxjs/_esm5/operators/index.js"),a=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),s=n("./node_modules/@angular/material/esm5/core.es5.js"),u=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),c=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),l=n("./node_modules/@angular/animations/fesm5/animations.js"),d=n("./node_modules/tslib/tslib.es6.js"),p=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),f=n("./node_modules/@angular/cdk/esm5/overlay.es5.js"),h=n("./node_modules/@angular/common/fesm5/common.js"),m=n("./node_modules/@angular/material/esm5/dialog.es5.js"),g=n("./node_modules/@angular/forms/fesm5/forms.js"),y=n("./node_modules/@angular/material/esm5/form-field.es5.js"),_=n("./node_modules/@angular/material/esm5/input.es5.js"),v=n("./node_modules/@angular/material/esm5/button.es5.js"),b=n("./node_modules/@angular/cdk/esm5/a11y.es5.js");function x(t){return Error("MatDatepicker: No provider found for "+t+". You must import one of the following modules at your application root: MatNativeDateModule, MatMomentDateModule, or provide a custom implementation.")}var w=function(){function t(){this.changes=new i.Subject,this.calendarLabel="Calendar",this.openCalendarLabel="Open calendar",this.prevMonthLabel="Previous month",this.nextMonthLabel="Next month",this.prevYearLabel="Previous year",this.nextYearLabel="Next year",this.prevMultiYearLabel="Previous 20 years",this.nextMultiYearLabel="Next 20 years",this.switchToMonthViewLabel="Choose date",this.switchToMultiYearViewLabel="Choose month and year"}return t.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(),j=function(){return function(t,e,n,r,i){this.value=t,this.displayValue=e,this.ariaLabel=n,this.enabled=r,this.cssClasses=i}}(),O=function(){function t(t,e){this._elementRef=t,this._ngZone=e,this.numCols=7,this.activeCell=0,this.cellAspectRatio=1,this.selectedValueChange=new r.EventEmitter}return t.prototype._cellClicked=function(t){t.enabled&&this.selectedValueChange.emit(t.value)},t.prototype.ngOnChanges=function(t){var e=t.numCols,n=this.rows,r=this.numCols;(t.rows||e)&&(this._firstRowOffset=n&&n.length&&n[0].length?r-n[0].length:0),(t.cellAspectRatio||e||!this._cellPadding)&&(this._cellPadding=50*this.cellAspectRatio/r+"%"),!e&&this._cellWidth||(this._cellWidth=100/r+"%")},t.prototype._isActiveCell=function(t,e){var n=t*this.numCols+e;return t&&(n-=this._firstRowOffset),n==this.activeCell},t.prototype._focusActiveCell=function(){var t=this;this._ngZone.runOutsideAngular(function(){t._ngZone.onStable.asObservable().pipe(Object(o.take)(1)).subscribe(function(){var e=t._elementRef.nativeElement.querySelector(".mat-calendar-body-active");e&&e.focus()})})},t.decorators=[{type:r.Component,args:[{selector:"[mat-calendar-body]",template:'<tr *ngIf="_firstRowOffset < labelMinRequiredCells" aria-hidden="true"><td class="mat-calendar-body-label" [attr.colspan]="numCols" [style.paddingTop]="_cellPadding" [style.paddingBottom]="_cellPadding">{{label}}</td></tr><tr *ngFor="let row of rows; let rowIndex = index" role="row"><td *ngIf="rowIndex === 0 && _firstRowOffset" aria-hidden="true" class="mat-calendar-body-label" [attr.colspan]="_firstRowOffset" [style.paddingTop]="_cellPadding" [style.paddingBottom]="_cellPadding">{{_firstRowOffset >= labelMinRequiredCells ? label : \'\'}}</td><td *ngFor="let item of row; let colIndex = index" role="gridcell" class="mat-calendar-body-cell" [ngClass]="item.cssClasses" [tabindex]="_isActiveCell(rowIndex, colIndex) ? 0 : -1" [class.mat-calendar-body-disabled]="!item.enabled" [class.mat-calendar-body-active]="_isActiveCell(rowIndex, colIndex)" [attr.aria-label]="item.ariaLabel" [attr.aria-disabled]="!item.enabled || null" [attr.aria-selected]="selectedValue === item.value" (click)="_cellClicked(item)" [style.width]="_cellWidth" [style.paddingTop]="_cellPadding" [style.paddingBottom]="_cellPadding"><div class="mat-calendar-body-cell-content" [class.mat-calendar-body-selected]="selectedValue === item.value" [class.mat-calendar-body-today]="todayValue === item.value">{{item.displayValue}}</div></td></tr>',styles:[".mat-calendar-body{min-width:224px}.mat-calendar-body-label{height:0;line-height:0;text-align:left;padding-left:4.71429%;padding-right:4.71429%}.mat-calendar-body-cell{position:relative;height:0;line-height:0;text-align:center;outline:0;cursor:pointer}.mat-calendar-body-disabled{cursor:default}.mat-calendar-body-cell-content{position:absolute;top:5%;left:5%;display:flex;align-items:center;justify-content:center;box-sizing:border-box;width:90%;height:90%;line-height:1;border-width:1px;border-style:solid;border-radius:999px}@media (-ms-high-contrast:active){.mat-calendar-body-cell-content{border:none}}@media (-ms-high-contrast:active){.mat-calendar-body-selected,.mat-datepicker-popup:not(:empty){outline:solid 1px}.mat-calendar-body-today{outline:dotted 1px}}[dir=rtl] .mat-calendar-body-label{text-align:right}"],host:{class:"mat-calendar-body",role:"grid","aria-readonly":"true"},exportAs:"matCalendarBody",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:r.NgZone}]},t.propDecorators={label:[{type:r.Input}],rows:[{type:r.Input}],todayValue:[{type:r.Input}],selectedValue:[{type:r.Input}],labelMinRequiredCells:[{type:r.Input}],numCols:[{type:r.Input}],activeCell:[{type:r.Input}],cellAspectRatio:[{type:r.Input}],selectedValueChange:[{type:r.Output}]},t}(),C=function(){function t(t,e,n,i){if(this._changeDetectorRef=t,this._dateFormats=e,this._dateAdapter=n,this._dir=i,this.selectedChange=new r.EventEmitter,this._userSelection=new r.EventEmitter,this.activeDateChange=new r.EventEmitter,!this._dateAdapter)throw x("DateAdapter");if(!this._dateFormats)throw x("MAT_DATE_FORMATS");var o=this._dateAdapter.getFirstDayOfWeek(),a=this._dateAdapter.getDayOfWeekNames("narrow"),s=this._dateAdapter.getDayOfWeekNames("long").map(function(t,e){return{long:t,narrow:a[e]}});this._weekdays=s.slice(o).concat(s.slice(0,o)),this._activeDate=this._dateAdapter.today()}return Object.defineProperty(t.prototype,"activeDate",{get:function(){return this._activeDate},set:function(t){var e=this._activeDate,n=this._getValidDateOrNull(this._dateAdapter.deserialize(t))||this._dateAdapter.today();this._activeDate=this._dateAdapter.clampDate(n,this.minDate,this.maxDate),this._hasSameMonthAndYear(e,this._activeDate)||this._init()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selected",{get:function(){return this._selected},set:function(t){this._selected=this._getValidDateOrNull(this._dateAdapter.deserialize(t)),this._selectedDate=this._getDateInCurrentMonth(this._selected)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"minDate",{get:function(){return this._minDate},set:function(t){this._minDate=this._getValidDateOrNull(this._dateAdapter.deserialize(t))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"maxDate",{get:function(){return this._maxDate},set:function(t){this._maxDate=this._getValidDateOrNull(this._dateAdapter.deserialize(t))},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this._init()},t.prototype._dateSelected=function(t){if(this._selectedDate!=t){var e=this._dateAdapter.getYear(this.activeDate),n=this._dateAdapter.getMonth(this.activeDate),r=this._dateAdapter.createDate(e,n,t);this.selectedChange.emit(r)}this._userSelection.emit()},t.prototype._handleCalendarBodyKeydown=function(t){var e=this._activeDate,n=this._isRtl();switch(t.keyCode){case a.LEFT_ARROW:this.activeDate=this._dateAdapter.addCalendarDays(this._activeDate,n?1:-1);break;case a.RIGHT_ARROW:this.activeDate=this._dateAdapter.addCalendarDays(this._activeDate,n?-1:1);break;case a.UP_ARROW:this.activeDate=this._dateAdapter.addCalendarDays(this._activeDate,-7);break;case a.DOWN_ARROW:this.activeDate=this._dateAdapter.addCalendarDays(this._activeDate,7);break;case a.HOME:this.activeDate=this._dateAdapter.addCalendarDays(this._activeDate,1-this._dateAdapter.getDate(this._activeDate));break;case a.END:this.activeDate=this._dateAdapter.addCalendarDays(this._activeDate,this._dateAdapter.getNumDaysInMonth(this._activeDate)-this._dateAdapter.getDate(this._activeDate));break;case a.PAGE_UP:this.activeDate=t.altKey?this._dateAdapter.addCalendarYears(this._activeDate,-1):this._dateAdapter.addCalendarMonths(this._activeDate,-1);break;case a.PAGE_DOWN:this.activeDate=t.altKey?this._dateAdapter.addCalendarYears(this._activeDate,1):this._dateAdapter.addCalendarMonths(this._activeDate,1);break;case a.ENTER:case a.SPACE:return void(this.dateFilter&&!this.dateFilter(this._activeDate)||(this._dateSelected(this._dateAdapter.getDate(this._activeDate)),this._userSelection.emit(),t.preventDefault()));default:return}this._dateAdapter.compareDate(e,this.activeDate)&&this.activeDateChange.emit(this.activeDate),this._focusActiveCell(),t.preventDefault()},t.prototype._init=function(){this._selectedDate=this._getDateInCurrentMonth(this.selected),this._todayDate=this._getDateInCurrentMonth(this._dateAdapter.today()),this._monthLabel=this._dateAdapter.getMonthNames("short")[this._dateAdapter.getMonth(this.activeDate)].toLocaleUpperCase();var t=this._dateAdapter.createDate(this._dateAdapter.getYear(this.activeDate),this._dateAdapter.getMonth(this.activeDate),1);this._firstWeekOffset=(7+this._dateAdapter.getDayOfWeek(t)-this._dateAdapter.getFirstDayOfWeek())%7,this._createWeekCells(),this._changeDetectorRef.markForCheck()},t.prototype._focusActiveCell=function(){this._matCalendarBody._focusActiveCell()},t.prototype._createWeekCells=function(){var t=this._dateAdapter.getNumDaysInMonth(this.activeDate),e=this._dateAdapter.getDateNames();this._weeks=[[]];for(var n=0,r=this._firstWeekOffset;n<t;n++,r++){7==r&&(this._weeks.push([]),r=0);var i=this._dateAdapter.createDate(this._dateAdapter.getYear(this.activeDate),this._dateAdapter.getMonth(this.activeDate),n+1),o=this._shouldEnableDate(i),a=this._dateAdapter.format(i,this._dateFormats.display.dateA11yLabel),s=this.dateClass?this.dateClass(i):void 0;this._weeks[this._weeks.length-1].push(new j(n+1,e[n],a,o,s))}},t.prototype._shouldEnableDate=function(t){return!!t&&(!this.dateFilter||this.dateFilter(t))&&(!this.minDate||this._dateAdapter.compareDate(t,this.minDate)>=0)&&(!this.maxDate||this._dateAdapter.compareDate(t,this.maxDate)<=0)},t.prototype._getDateInCurrentMonth=function(t){return t&&this._hasSameMonthAndYear(t,this.activeDate)?this._dateAdapter.getDate(t):null},t.prototype._hasSameMonthAndYear=function(t,e){return!(!t||!e||this._dateAdapter.getMonth(t)!=this._dateAdapter.getMonth(e)||this._dateAdapter.getYear(t)!=this._dateAdapter.getYear(e))},t.prototype._getValidDateOrNull=function(t){return this._dateAdapter.isDateInstance(t)&&this._dateAdapter.isValid(t)?t:null},t.prototype._isRtl=function(){return this._dir&&"rtl"===this._dir.value},t.decorators=[{type:r.Component,args:[{selector:"mat-month-view",template:'<table class="mat-calendar-table"><thead class="mat-calendar-table-header"><tr><th *ngFor="let day of _weekdays" [attr.aria-label]="day.long">{{day.narrow}}</th></tr><tr><th class="mat-calendar-table-header-divider" colspan="7" aria-hidden="true"></th></tr></thead><tbody mat-calendar-body [label]="_monthLabel" [rows]="_weeks" [todayValue]="_todayDate" [selectedValue]="_selectedDate" [labelMinRequiredCells]="3" [activeCell]="_dateAdapter.getDate(activeDate) - 1" (selectedValueChange)="_dateSelected($event)" (keydown)="_handleCalendarBodyKeydown($event)"></tbody></table>',exportAs:"matMonthView",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:r.ChangeDetectorRef},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[s.MAT_DATE_FORMATS]}]},{type:s.DateAdapter,decorators:[{type:r.Optional}]},{type:u.Directionality,decorators:[{type:r.Optional}]}]},t.propDecorators={activeDate:[{type:r.Input}],selected:[{type:r.Input}],minDate:[{type:r.Input}],maxDate:[{type:r.Input}],dateFilter:[{type:r.Input}],dateClass:[{type:r.Input}],selectedChange:[{type:r.Output}],_userSelection:[{type:r.Output}],activeDateChange:[{type:r.Output}],_matCalendarBody:[{type:r.ViewChild,args:[O]}]},t}(),E=function(){function t(t,e,n){if(this._changeDetectorRef=t,this._dateAdapter=e,this._dir=n,this.selectedChange=new r.EventEmitter,this.yearSelected=new r.EventEmitter,this.activeDateChange=new r.EventEmitter,!this._dateAdapter)throw x("DateAdapter");this._activeDate=this._dateAdapter.today()}return Object.defineProperty(t.prototype,"activeDate",{get:function(){return this._activeDate},set:function(t){var e=this._activeDate,n=this._getValidDateOrNull(this._dateAdapter.deserialize(t))||this._dateAdapter.today();this._activeDate=this._dateAdapter.clampDate(n,this.minDate,this.maxDate),Math.floor(this._dateAdapter.getYear(e)/24)!=Math.floor(this._dateAdapter.getYear(this._activeDate)/24)&&this._init()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selected",{get:function(){return this._selected},set:function(t){this._selected=this._getValidDateOrNull(this._dateAdapter.deserialize(t)),this._selectedYear=this._selected&&this._dateAdapter.getYear(this._selected)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"minDate",{get:function(){return this._minDate},set:function(t){this._minDate=this._getValidDateOrNull(this._dateAdapter.deserialize(t))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"maxDate",{get:function(){return this._maxDate},set:function(t){this._maxDate=this._getValidDateOrNull(this._dateAdapter.deserialize(t))},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this._init()},t.prototype._init=function(){var t=this;this._todayYear=this._dateAdapter.getYear(this._dateAdapter.today());var e=this._dateAdapter.getYear(this._activeDate),n=e%24;this._years=[];for(var r=0,i=[];r<24;r++)i.push(e-n+r),4==i.length&&(this._years.push(i.map(function(e){return t._createCellForYear(e)})),i=[]);this._changeDetectorRef.markForCheck()},t.prototype._yearSelected=function(t){this.yearSelected.emit(this._dateAdapter.createDate(t,0,1));var e=this._dateAdapter.getMonth(this.activeDate),n=this._dateAdapter.getNumDaysInMonth(this._dateAdapter.createDate(t,e,1));this.selectedChange.emit(this._dateAdapter.createDate(t,e,Math.min(this._dateAdapter.getDate(this.activeDate),n)))},t.prototype._handleCalendarBodyKeydown=function(t){var e=this._activeDate,n=this._isRtl();switch(t.keyCode){case a.LEFT_ARROW:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,n?1:-1);break;case a.RIGHT_ARROW:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,n?-1:1);break;case a.UP_ARROW:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,-4);break;case a.DOWN_ARROW:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,4);break;case a.HOME:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,-this._dateAdapter.getYear(this._activeDate)%24);break;case a.END:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,24-this._dateAdapter.getYear(this._activeDate)%24-1);break;case a.PAGE_UP:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,t.altKey?-240:-24);break;case a.PAGE_DOWN:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,t.altKey?240:24);break;case a.ENTER:case a.SPACE:this._yearSelected(this._dateAdapter.getYear(this._activeDate));break;default:return}this._dateAdapter.compareDate(e,this.activeDate)&&this.activeDateChange.emit(this.activeDate),this._focusActiveCell(),t.preventDefault()},t.prototype._getActiveCell=function(){return this._dateAdapter.getYear(this.activeDate)%24},t.prototype._focusActiveCell=function(){this._matCalendarBody._focusActiveCell()},t.prototype._createCellForYear=function(t){var e=this._dateAdapter.getYearName(this._dateAdapter.createDate(t,0,1));return new j(t,e,e,this._shouldEnableYear(t))},t.prototype._shouldEnableYear=function(t){if(void 0===t||null===t||this.maxDate&&t>this._dateAdapter.getYear(this.maxDate)||this.minDate&&t<this._dateAdapter.getYear(this.minDate))return!1;if(!this.dateFilter)return!0;for(var e=this._dateAdapter.createDate(t,0,1);this._dateAdapter.getYear(e)==t;e=this._dateAdapter.addCalendarDays(e,1))if(this.dateFilter(e))return!0;return!1},t.prototype._getValidDateOrNull=function(t){return this._dateAdapter.isDateInstance(t)&&this._dateAdapter.isValid(t)?t:null},t.prototype._isRtl=function(){return this._dir&&"rtl"===this._dir.value},t.decorators=[{type:r.Component,args:[{selector:"mat-multi-year-view",template:'<table class="mat-calendar-table"><thead class="mat-calendar-table-header"><tr><th class="mat-calendar-table-header-divider" colspan="4"></th></tr></thead><tbody mat-calendar-body [rows]="_years" [todayValue]="_todayYear" [selectedValue]="_selectedYear" [numCols]="4" [cellAspectRatio]="4 / 7" [activeCell]="_getActiveCell()" (selectedValueChange)="_yearSelected($event)" (keydown)="_handleCalendarBodyKeydown($event)"></tbody></table>',exportAs:"matMultiYearView",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:r.ChangeDetectorRef},{type:s.DateAdapter,decorators:[{type:r.Optional}]},{type:u.Directionality,decorators:[{type:r.Optional}]}]},t.propDecorators={activeDate:[{type:r.Input}],selected:[{type:r.Input}],minDate:[{type:r.Input}],maxDate:[{type:r.Input}],dateFilter:[{type:r.Input}],selectedChange:[{type:r.Output}],yearSelected:[{type:r.Output}],activeDateChange:[{type:r.Output}],_matCalendarBody:[{type:r.ViewChild,args:[O]}]},t}(),S=function(){function t(t,e,n,i){if(this._changeDetectorRef=t,this._dateFormats=e,this._dateAdapter=n,this._dir=i,this.selectedChange=new r.EventEmitter,this.monthSelected=new r.EventEmitter,this.activeDateChange=new r.EventEmitter,!this._dateAdapter)throw x("DateAdapter");if(!this._dateFormats)throw x("MAT_DATE_FORMATS");this._activeDate=this._dateAdapter.today()}return Object.defineProperty(t.prototype,"activeDate",{get:function(){return this._activeDate},set:function(t){var e=this._activeDate,n=this._getValidDateOrNull(this._dateAdapter.deserialize(t))||this._dateAdapter.today();this._activeDate=this._dateAdapter.clampDate(n,this.minDate,this.maxDate),this._dateAdapter.getYear(e)!==this._dateAdapter.getYear(this._activeDate)&&this._init()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selected",{get:function(){return this._selected},set:function(t){this._selected=this._getValidDateOrNull(this._dateAdapter.deserialize(t)),this._selectedMonth=this._getMonthInCurrentYear(this._selected)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"minDate",{get:function(){return this._minDate},set:function(t){this._minDate=this._getValidDateOrNull(this._dateAdapter.deserialize(t))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"maxDate",{get:function(){return this._maxDate},set:function(t){this._maxDate=this._getValidDateOrNull(this._dateAdapter.deserialize(t))},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this._init()},t.prototype._monthSelected=function(t){var e=this._dateAdapter.createDate(this._dateAdapter.getYear(this.activeDate),t,1);this.monthSelected.emit(e);var n=this._dateAdapter.getNumDaysInMonth(e);this.selectedChange.emit(this._dateAdapter.createDate(this._dateAdapter.getYear(this.activeDate),t,Math.min(this._dateAdapter.getDate(this.activeDate),n)))},t.prototype._handleCalendarBodyKeydown=function(t){var e=this._activeDate,n=this._isRtl();switch(t.keyCode){case a.LEFT_ARROW:this.activeDate=this._dateAdapter.addCalendarMonths(this._activeDate,n?1:-1);break;case a.RIGHT_ARROW:this.activeDate=this._dateAdapter.addCalendarMonths(this._activeDate,n?-1:1);break;case a.UP_ARROW:this.activeDate=this._dateAdapter.addCalendarMonths(this._activeDate,-4);break;case a.DOWN_ARROW:this.activeDate=this._dateAdapter.addCalendarMonths(this._activeDate,4);break;case a.HOME:this.activeDate=this._dateAdapter.addCalendarMonths(this._activeDate,-this._dateAdapter.getMonth(this._activeDate));break;case a.END:this.activeDate=this._dateAdapter.addCalendarMonths(this._activeDate,11-this._dateAdapter.getMonth(this._activeDate));break;case a.PAGE_UP:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,t.altKey?-10:-1);break;case a.PAGE_DOWN:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,t.altKey?10:1);break;case a.ENTER:case a.SPACE:this._monthSelected(this._dateAdapter.getMonth(this._activeDate));break;default:return}this._dateAdapter.compareDate(e,this.activeDate)&&this.activeDateChange.emit(this.activeDate),this._focusActiveCell(),t.preventDefault()},t.prototype._init=function(){var t=this;this._selectedMonth=this._getMonthInCurrentYear(this.selected),this._todayMonth=this._getMonthInCurrentYear(this._dateAdapter.today()),this._yearLabel=this._dateAdapter.getYearName(this.activeDate);var e=this._dateAdapter.getMonthNames("short");this._months=[[0,1,2,3],[4,5,6,7],[8,9,10,11]].map(function(n){return n.map(function(n){return t._createCellForMonth(n,e[n])})}),this._changeDetectorRef.markForCheck()},t.prototype._focusActiveCell=function(){this._matCalendarBody._focusActiveCell()},t.prototype._getMonthInCurrentYear=function(t){return t&&this._dateAdapter.getYear(t)==this._dateAdapter.getYear(this.activeDate)?this._dateAdapter.getMonth(t):null},t.prototype._createCellForMonth=function(t,e){var n=this._dateAdapter.format(this._dateAdapter.createDate(this._dateAdapter.getYear(this.activeDate),t,1),this._dateFormats.display.monthYearA11yLabel);return new j(t,e.toLocaleUpperCase(),n,this._shouldEnableMonth(t))},t.prototype._shouldEnableMonth=function(t){var e=this._dateAdapter.getYear(this.activeDate);if(void 0===t||null===t||this._isYearAndMonthAfterMaxDate(e,t)||this._isYearAndMonthBeforeMinDate(e,t))return!1;if(!this.dateFilter)return!0;for(var n=this._dateAdapter.createDate(e,t,1);this._dateAdapter.getMonth(n)==t;n=this._dateAdapter.addCalendarDays(n,1))if(this.dateFilter(n))return!0;return!1},t.prototype._isYearAndMonthAfterMaxDate=function(t,e){if(this.maxDate){var n=this._dateAdapter.getYear(this.maxDate),r=this._dateAdapter.getMonth(this.maxDate);return t>n||t===n&&e>r}return!1},t.prototype._isYearAndMonthBeforeMinDate=function(t,e){if(this.minDate){var n=this._dateAdapter.getYear(this.minDate),r=this._dateAdapter.getMonth(this.minDate);return t<n||t===n&&e<r}return!1},t.prototype._getValidDateOrNull=function(t){return this._dateAdapter.isDateInstance(t)&&this._dateAdapter.isValid(t)?t:null},t.prototype._isRtl=function(){return this._dir&&"rtl"===this._dir.value},t.decorators=[{type:r.Component,args:[{selector:"mat-year-view",template:'<table class="mat-calendar-table"><thead class="mat-calendar-table-header"><tr><th class="mat-calendar-table-header-divider" colspan="4"></th></tr></thead><tbody mat-calendar-body [label]="_yearLabel" [rows]="_months" [todayValue]="_todayMonth" [selectedValue]="_selectedMonth" [labelMinRequiredCells]="2" [numCols]="4" [cellAspectRatio]="4 / 7" [activeCell]="_dateAdapter.getMonth(activeDate)" (selectedValueChange)="_monthSelected($event)" (keydown)="_handleCalendarBodyKeydown($event)"></tbody></table>',exportAs:"matYearView",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:r.ChangeDetectorRef},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[s.MAT_DATE_FORMATS]}]},{type:s.DateAdapter,decorators:[{type:r.Optional}]},{type:u.Directionality,decorators:[{type:r.Optional}]}]},t.propDecorators={activeDate:[{type:r.Input}],selected:[{type:r.Input}],minDate:[{type:r.Input}],maxDate:[{type:r.Input}],dateFilter:[{type:r.Input}],selectedChange:[{type:r.Output}],monthSelected:[{type:r.Output}],activeDateChange:[{type:r.Output}],_matCalendarBody:[{type:r.ViewChild,args:[O]}]},t}(),M=function(){function t(t,e,n,r,i){this._intl=t,this.calendar=e,this._dateAdapter=n,this._dateFormats=r,this.calendar.stateChanges.subscribe(function(){return i.markForCheck()})}return Object.defineProperty(t.prototype,"periodButtonText",{get:function(){if("month"==this.calendar.currentView)return this._dateAdapter.format(this.calendar.activeDate,this._dateFormats.display.monthYearLabel).toLocaleUpperCase();if("year"==this.calendar.currentView)return this._dateAdapter.getYearName(this.calendar.activeDate);var t=this._dateAdapter.getYear(this.calendar.activeDate);return this._dateAdapter.getYearName(this._dateAdapter.createDate(t-t%24,0,1))+" – "+this._dateAdapter.getYearName(this._dateAdapter.createDate(t+24-1-t%24,0,1))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"periodButtonLabel",{get:function(){return"month"==this.calendar.currentView?this._intl.switchToMultiYearViewLabel:this._intl.switchToMonthViewLabel},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"prevButtonLabel",{get:function(){return{month:this._intl.prevMonthLabel,year:this._intl.prevYearLabel,"multi-year":this._intl.prevMultiYearLabel}[this.calendar.currentView]},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"nextButtonLabel",{get:function(){return{month:this._intl.nextMonthLabel,year:this._intl.nextYearLabel,"multi-year":this._intl.nextMultiYearLabel}[this.calendar.currentView]},enumerable:!0,configurable:!0}),t.prototype.currentPeriodClicked=function(){this.calendar.currentView="month"==this.calendar.currentView?"multi-year":"month"},t.prototype.previousClicked=function(){this.calendar.activeDate="month"==this.calendar.currentView?this._dateAdapter.addCalendarMonths(this.calendar.activeDate,-1):this._dateAdapter.addCalendarYears(this.calendar.activeDate,"year"==this.calendar.currentView?-1:-24)},t.prototype.nextClicked=function(){this.calendar.activeDate="month"==this.calendar.currentView?this._dateAdapter.addCalendarMonths(this.calendar.activeDate,1):this._dateAdapter.addCalendarYears(this.calendar.activeDate,"year"==this.calendar.currentView?1:24)},t.prototype.previousEnabled=function(){return!this.calendar.minDate||(!this.calendar.minDate||!this._isSameView(this.calendar.activeDate,this.calendar.minDate))},t.prototype.nextEnabled=function(){return!this.calendar.maxDate||!this._isSameView(this.calendar.activeDate,this.calendar.maxDate)},t.prototype._isSameView=function(t,e){return"month"==this.calendar.currentView?this._dateAdapter.getYear(t)==this._dateAdapter.getYear(e)&&this._dateAdapter.getMonth(t)==this._dateAdapter.getMonth(e):"year"==this.calendar.currentView?this._dateAdapter.getYear(t)==this._dateAdapter.getYear(e):Math.floor(this._dateAdapter.getYear(t)/24)==Math.floor(this._dateAdapter.getYear(e)/24)},t.decorators=[{type:r.Component,args:[{selector:"mat-calendar-header",template:'<div class="mat-calendar-header"><div class="mat-calendar-controls"><button mat-button type="button" class="mat-calendar-period-button" (click)="currentPeriodClicked()" [attr.aria-label]="periodButtonLabel" cdkAriaLive="polite">{{periodButtonText}}<div class="mat-calendar-arrow" [class.mat-calendar-invert]="calendar.currentView != \'month\'"></div></button><div class="mat-calendar-spacer"></div><ng-content></ng-content><button mat-icon-button type="button" class="mat-calendar-previous-button" [disabled]="!previousEnabled()" (click)="previousClicked()" [attr.aria-label]="prevButtonLabel"></button> <button mat-icon-button type="button" class="mat-calendar-next-button" [disabled]="!nextEnabled()" (click)="nextClicked()" [attr.aria-label]="nextButtonLabel"></button></div></div>',exportAs:"matCalendarHeader",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:w},{type:T,decorators:[{type:r.Inject,args:[Object(r.forwardRef)(function(){return T})]}]},{type:s.DateAdapter,decorators:[{type:r.Optional}]},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[s.MAT_DATE_FORMATS]}]},{type:r.ChangeDetectorRef}]},t}(),T=function(){function t(t,e,n,o){var a=this;if(this._dateAdapter=e,this._dateFormats=n,this._changeDetectorRef=o,this._moveFocusOnNextTick=!1,this.startView="month",this.selectedChange=new r.EventEmitter,this.yearSelected=new r.EventEmitter,this.monthSelected=new r.EventEmitter,this._userSelection=new r.EventEmitter,this.stateChanges=new i.Subject,!this._dateAdapter)throw x("DateAdapter");if(!this._dateFormats)throw x("MAT_DATE_FORMATS");this._intlChanges=t.changes.subscribe(function(){o.markForCheck(),a.stateChanges.next()})}return Object.defineProperty(t.prototype,"startAt",{get:function(){return this._startAt},set:function(t){this._startAt=this._getValidDateOrNull(this._dateAdapter.deserialize(t))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selected",{get:function(){return this._selected},set:function(t){this._selected=this._getValidDateOrNull(this._dateAdapter.deserialize(t))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"minDate",{get:function(){return this._minDate},set:function(t){this._minDate=this._getValidDateOrNull(this._dateAdapter.deserialize(t))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"maxDate",{get:function(){return this._maxDate},set:function(t){this._maxDate=this._getValidDateOrNull(this._dateAdapter.deserialize(t))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"activeDate",{get:function(){return this._clampedActiveDate},set:function(t){this._clampedActiveDate=this._dateAdapter.clampDate(t,this.minDate,this.maxDate),this.stateChanges.next()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"currentView",{get:function(){return this._currentView},set:function(t){this._currentView=t,this._moveFocusOnNextTick=!0},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this._calendarHeaderPortal=new c.ComponentPortal(this.headerComponent||M),this.activeDate=this.startAt||this._dateAdapter.today(),this._currentView=this.startView},t.prototype.ngAfterViewChecked=function(){this._moveFocusOnNextTick&&(this._moveFocusOnNextTick=!1,this.focusActiveCell())},t.prototype.ngOnDestroy=function(){this._intlChanges.unsubscribe(),this.stateChanges.complete()},t.prototype.ngOnChanges=function(t){var e=t.minDate||t.maxDate||t.dateFilter;if(e&&!e.firstChange){var n=this._getCurrentViewComponent();n&&(this._changeDetectorRef.detectChanges(),n._init())}this.stateChanges.next()},t.prototype.focusActiveCell=function(){this._getCurrentViewComponent()._focusActiveCell()},t.prototype.updateTodaysDate=function(){("month"==this.currentView?this.monthView:"year"==this.currentView?this.yearView:this.multiYearView).ngAfterContentInit()},t.prototype._dateSelected=function(t){this._dateAdapter.sameDate(t,this.selected)||this.selectedChange.emit(t)},t.prototype._yearSelectedInMultiYearView=function(t){this.yearSelected.emit(t)},t.prototype._monthSelectedInYearView=function(t){this.monthSelected.emit(t)},t.prototype._userSelected=function(){this._userSelection.emit()},t.prototype._goToDateInView=function(t,e){this.activeDate=t,this.currentView=e},t.prototype._getValidDateOrNull=function(t){return this._dateAdapter.isDateInstance(t)&&this._dateAdapter.isValid(t)?t:null},t.prototype._getCurrentViewComponent=function(){return this.monthView||this.yearView||this.multiYearView},t.decorators=[{type:r.Component,args:[{selector:"mat-calendar",template:'<ng-template [cdkPortalOutlet]="_calendarHeaderPortal"></ng-template><div class="mat-calendar-content" [ngSwitch]="currentView" cdkMonitorSubtreeFocus tabindex="-1"><mat-month-view *ngSwitchCase="\'month\'" [(activeDate)]="activeDate" [selected]="selected" [dateFilter]="dateFilter" [maxDate]="maxDate" [minDate]="minDate" [dateClass]="dateClass" (selectedChange)="_dateSelected($event)" (_userSelection)="_userSelected()"></mat-month-view><mat-year-view *ngSwitchCase="\'year\'" [(activeDate)]="activeDate" [selected]="selected" [dateFilter]="dateFilter" [maxDate]="maxDate" [minDate]="minDate" (monthSelected)="_monthSelectedInYearView($event)" (selectedChange)="_goToDateInView($event, \'month\')"></mat-year-view><mat-multi-year-view *ngSwitchCase="\'multi-year\'" [(activeDate)]="activeDate" [selected]="selected" [dateFilter]="dateFilter" [maxDate]="maxDate" [minDate]="minDate" (yearSelected)="_yearSelectedInMultiYearView($event)" (selectedChange)="_goToDateInView($event, \'year\')"></mat-multi-year-view></div>',styles:[".mat-calendar{display:block}.mat-calendar-header{padding:8px 8px 0 8px}.mat-calendar-content{padding:0 8px 8px 8px;outline:0}.mat-calendar-controls{display:flex;margin:5% calc(33% / 7 - 16px)}.mat-calendar-spacer{flex:1 1 auto}.mat-calendar-period-button{min-width:0}.mat-calendar-arrow{display:inline-block;width:0;height:0;border-left:5px solid transparent;border-right:5px solid transparent;border-top-width:5px;border-top-style:solid;margin:0 0 0 5px;vertical-align:middle}.mat-calendar-arrow.mat-calendar-invert{transform:rotate(180deg)}[dir=rtl] .mat-calendar-arrow{margin:0 5px 0 0}.mat-calendar-next-button,.mat-calendar-previous-button{position:relative}.mat-calendar-next-button::after,.mat-calendar-previous-button::after{top:0;left:0;right:0;bottom:0;position:absolute;content:'';margin:15.5px;border:0 solid currentColor;border-top-width:2px}[dir=rtl] .mat-calendar-next-button,[dir=rtl] .mat-calendar-previous-button{transform:rotate(180deg)}.mat-calendar-previous-button::after{border-left-width:2px;transform:translateX(2px) rotate(-45deg)}.mat-calendar-next-button::after{border-right-width:2px;transform:translateX(-2px) rotate(45deg)}.mat-calendar-table{border-spacing:0;border-collapse:collapse;width:100%}.mat-calendar-table-header th{text-align:center;padding:0 0 8px 0}.mat-calendar-table-header-divider{position:relative;height:1px}.mat-calendar-table-header-divider::after{content:'';position:absolute;top:0;left:-8px;right:-8px;height:1px}"],host:{class:"mat-calendar"},exportAs:"matCalendar",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:w},{type:s.DateAdapter,decorators:[{type:r.Optional}]},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[s.MAT_DATE_FORMATS]}]},{type:r.ChangeDetectorRef}]},t.propDecorators={headerComponent:[{type:r.Input}],startAt:[{type:r.Input}],startView:[{type:r.Input}],selected:[{type:r.Input}],minDate:[{type:r.Input}],maxDate:[{type:r.Input}],dateFilter:[{type:r.Input}],dateClass:[{type:r.Input}],selectedChange:[{type:r.Output}],yearSelected:[{type:r.Output}],monthSelected:[{type:r.Output}],_userSelection:[{type:r.Output}],monthView:[{type:r.ViewChild,args:[C]}],yearView:[{type:r.ViewChild,args:[S]}],multiYearView:[{type:r.ViewChild,args:[E]}]},t}(),k={transformPanel:Object(l.trigger)("transformPanel",[Object(l.state)("void",Object(l.style)({opacity:0,transform:"scale(1, 0.8)"})),Object(l.transition)("void => enter",Object(l.animate)("120ms cubic-bezier(0, 0, 0.2, 1)",Object(l.style)({opacity:1,transform:"scale(1, 1)"}))),Object(l.transition)("* => void",Object(l.animate)("100ms linear",Object(l.style)({opacity:0})))]),fadeInCalendar:Object(l.trigger)("fadeInCalendar",[Object(l.state)("void",Object(l.style)({opacity:0})),Object(l.state)("enter",Object(l.style)({opacity:1})),Object(l.transition)("void => *",Object(l.animate)("120ms 100ms cubic-bezier(0.55, 0, 0.55, 0.2)"))])},A=0,D=new r.InjectionToken("mat-datepicker-scroll-strategy");function I(t){return function(){return t.scrollStrategies.reposition()}}var P={provide:D,deps:[f.Overlay],useFactory:I},R=function(){return function(t){this._elementRef=t}}(),N=Object(s.mixinColor)(R),F=function(t){function e(e){return t.call(this,e)||this}return Object(d.__extends)(e,t),e.prototype.ngAfterViewInit=function(){this._calendar.focusActiveCell()},e.decorators=[{type:r.Component,args:[{selector:"mat-datepicker-content",template:'<mat-calendar cdkTrapFocus [id]="datepicker.id" [ngClass]="datepicker.panelClass" [startAt]="datepicker.startAt" [startView]="datepicker.startView" [minDate]="datepicker._minDate" [maxDate]="datepicker._maxDate" [dateFilter]="datepicker._dateFilter" [headerComponent]="datepicker.calendarHeaderComponent" [selected]="datepicker._selected" [dateClass]="datepicker.dateClass" [@fadeInCalendar]="\'enter\'" (selectedChange)="datepicker.select($event)" (yearSelected)="datepicker._selectYear($event)" (monthSelected)="datepicker._selectMonth($event)" (_userSelection)="datepicker.close()"></mat-calendar>',styles:[".mat-datepicker-content{display:block;border-radius:4px}.mat-datepicker-content .mat-calendar{width:296px;height:354px}.mat-datepicker-content-touch{display:block;max-height:80vh;overflow:auto;margin:-24px}.mat-datepicker-content-touch .mat-calendar{min-width:250px;min-height:312px;max-width:750px;max-height:788px}@media all and (orientation:landscape){.mat-datepicker-content-touch .mat-calendar{width:64vh;height:80vh}}@media all and (orientation:portrait){.mat-datepicker-content-touch .mat-calendar{width:80vw;height:100vw}}"],host:{class:"mat-datepicker-content","[@transformPanel]":'"enter"',"[class.mat-datepicker-content-touch]":"datepicker.touchUi"},animations:[k.transformPanel,k.fadeInCalendar],exportAs:"matDatepickerContent",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,inputs:["color"]}]}],e.ctorParameters=function(){return[{type:r.ElementRef}]},e.propDecorators={_calendar:[{type:r.ViewChild,args:[T]}]},e}(N),L=function(){function t(t,e,n,o,a,s,u,c){if(this._dialog=t,this._overlay=e,this._ngZone=n,this._viewContainerRef=o,this._dateAdapter=s,this._dir=u,this._document=c,this.startView="month",this._touchUi=!1,this.yearSelected=new r.EventEmitter,this.monthSelected=new r.EventEmitter,this.openedStream=new r.EventEmitter,this.closedStream=new r.EventEmitter,this._opened=!1,this.id="mat-datepicker-"+A++,this._validSelected=null,this._focusedElementBeforeOpen=null,this._inputSubscription=i.Subscription.EMPTY,this._disabledChange=new i.Subject,this._selectedChanged=new i.Subject,!this._dateAdapter)throw x("DateAdapter");this._scrollStrategy=a}return Object.defineProperty(t.prototype,"startAt",{get:function(){return this._startAt||(this._datepickerInput?this._datepickerInput.value:null)},set:function(t){this._startAt=this._getValidDateOrNull(this._dateAdapter.deserialize(t))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"color",{get:function(){return this._color||(this._datepickerInput?this._datepickerInput._getThemePalette():void 0)},set:function(t){this._color=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"touchUi",{get:function(){return this._touchUi},set:function(t){this._touchUi=Object(p.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return void 0===this._disabled&&this._datepickerInput?this._datepickerInput.disabled:!!this._disabled},set:function(t){var e=Object(p.coerceBooleanProperty)(t);e!==this._disabled&&(this._disabled=e,this._disabledChange.next(e))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"opened",{get:function(){return this._opened},set:function(t){t?this.open():this.close()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_selected",{get:function(){return this._validSelected},set:function(t){this._validSelected=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_minDate",{get:function(){return this._datepickerInput&&this._datepickerInput.min},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_maxDate",{get:function(){return this._datepickerInput&&this._datepickerInput.max},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_dateFilter",{get:function(){return this._datepickerInput&&this._datepickerInput._dateFilter},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){this.close(),this._inputSubscription.unsubscribe(),this._disabledChange.complete(),this._popupRef&&(this._popupRef.dispose(),this._popupComponentRef=null)},t.prototype.select=function(t){var e=this._selected;this._selected=t,this._dateAdapter.sameDate(e,this._selected)||this._selectedChanged.next(t)},t.prototype._selectYear=function(t){this.yearSelected.emit(t)},t.prototype._selectMonth=function(t){this.monthSelected.emit(t)},t.prototype._registerInput=function(t){var e=this;if(this._datepickerInput)throw Error("A MatDatepicker can only be associated with a single input.");this._datepickerInput=t,this._inputSubscription=this._datepickerInput._valueChange.subscribe(function(t){return e._selected=t})},t.prototype.open=function(){if(!this._opened&&!this.disabled){if(!this._datepickerInput)throw Error("Attempted to open an MatDatepicker with no associated input.");this._document&&(this._focusedElementBeforeOpen=this._document.activeElement),this.touchUi?this._openAsDialog():this._openAsPopup(),this._opened=!0,this.openedStream.emit()}},t.prototype.close=function(){var t=this;if(this._opened){this._popupRef&&this._popupRef.hasAttached()&&this._popupRef.detach(),this._dialogRef&&(this._dialogRef.close(),this._dialogRef=null),this._calendarPortal&&this._calendarPortal.isAttached&&this._calendarPortal.detach();var e=function(){t._opened&&(t._opened=!1,t.closedStream.emit(),t._focusedElementBeforeOpen=null)};this._focusedElementBeforeOpen&&"function"==typeof this._focusedElementBeforeOpen.focus?(this._focusedElementBeforeOpen.focus(),setTimeout(e)):e()}},t.prototype._openAsDialog=function(){var t=this;this._dialogRef&&this._dialogRef.close(),this._dialogRef=this._dialog.open(F,{direction:this._dir?this._dir.value:"ltr",viewContainerRef:this._viewContainerRef,panelClass:"mat-datepicker-dialog"}),this._dialogRef.afterClosed().subscribe(function(){return t.close()}),this._dialogRef.componentInstance.datepicker=this,this._setColor()},t.prototype._openAsPopup=function(){var t=this;this._calendarPortal||(this._calendarPortal=new c.ComponentPortal(F,this._viewContainerRef)),this._popupRef||this._createPopup(),this._popupRef.hasAttached()||(this._popupComponentRef=this._popupRef.attach(this._calendarPortal),this._popupComponentRef.instance.datepicker=this,this._setColor(),this._ngZone.onStable.asObservable().pipe(Object(o.take)(1)).subscribe(function(){t._popupRef.updatePosition()}))},t.prototype._createPopup=function(){var t=this,e=new f.OverlayConfig({positionStrategy:this._createPopupPositionStrategy(),hasBackdrop:!0,backdropClass:"mat-overlay-transparent-backdrop",direction:this._dir,scrollStrategy:this._scrollStrategy(),panelClass:"mat-datepicker-popup"});this._popupRef=this._overlay.create(e),this._popupRef.overlayElement.setAttribute("role","dialog"),Object(i.merge)(this._popupRef.backdropClick(),this._popupRef.detachments(),this._popupRef.keydownEvents().pipe(Object(o.filter)(function(e){return e.keyCode===a.ESCAPE||t._datepickerInput&&e.altKey&&e.keyCode===a.UP_ARROW}))).subscribe(function(){return t.close()})},t.prototype._createPopupPositionStrategy=function(){return this._overlay.position().flexibleConnectedTo(this._datepickerInput.getConnectedOverlayOrigin()).withTransformOriginOn(".mat-datepicker-content").withFlexibleDimensions(!1).withViewportMargin(8).withLockedPosition().withPositions([{originX:"start",originY:"bottom",overlayX:"start",overlayY:"top"},{originX:"start",originY:"top",overlayX:"start",overlayY:"bottom"},{originX:"end",originY:"bottom",overlayX:"end",overlayY:"top"},{originX:"end",originY:"top",overlayX:"end",overlayY:"bottom"}])},t.prototype._getValidDateOrNull=function(t){return this._dateAdapter.isDateInstance(t)&&this._dateAdapter.isValid(t)?t:null},t.prototype._setColor=function(){var t=this.color;this._popupComponentRef&&(this._popupComponentRef.instance.color=t),this._dialogRef&&(this._dialogRef.componentInstance.color=t)},t.decorators=[{type:r.Component,args:[{selector:"mat-datepicker",template:"",exportAs:"matDatepicker",changeDetection:r.ChangeDetectionStrategy.OnPush,encapsulation:r.ViewEncapsulation.None}]}],t.ctorParameters=function(){return[{type:m.MatDialog},{type:f.Overlay},{type:r.NgZone},{type:r.ViewContainerRef},{type:void 0,decorators:[{type:r.Inject,args:[D]}]},{type:s.DateAdapter,decorators:[{type:r.Optional}]},{type:u.Directionality,decorators:[{type:r.Optional}]},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[h.DOCUMENT]}]}]},t.propDecorators={calendarHeaderComponent:[{type:r.Input}],startAt:[{type:r.Input}],startView:[{type:r.Input}],color:[{type:r.Input}],touchUi:[{type:r.Input}],disabled:[{type:r.Input}],yearSelected:[{type:r.Output}],monthSelected:[{type:r.Output}],panelClass:[{type:r.Input}],dateClass:[{type:r.Input}],openedStream:[{type:r.Output,args:["opened"]}],closedStream:[{type:r.Output,args:["closed"]}],opened:[{type:r.Input}]},t}(),B={provide:g.NG_VALUE_ACCESSOR,useExisting:Object(r.forwardRef)(function(){return H}),multi:!0},V={provide:g.NG_VALIDATORS,useExisting:Object(r.forwardRef)(function(){return H}),multi:!0},z=function(){return function(t,e){this.target=t,this.targetElement=e,this.value=this.target.value}}(),H=function(){function t(t,e,n,o){var a=this;if(this._elementRef=t,this._dateAdapter=e,this._dateFormats=n,this._formField=o,this.dateChange=new r.EventEmitter,this.dateInput=new r.EventEmitter,this._valueChange=new r.EventEmitter,this._disabledChange=new r.EventEmitter,this._onTouched=function(){},this._cvaOnChange=function(){},this._validatorOnChange=function(){},this._datepickerSubscription=i.Subscription.EMPTY,this._localeSubscription=i.Subscription.EMPTY,this._parseValidator=function(){return a._lastValueValid?null:{matDatepickerParse:{text:a._elementRef.nativeElement.value}}},this._minValidator=function(t){var e=a._getValidDateOrNull(a._dateAdapter.deserialize(t.value));return!a.min||!e||a._dateAdapter.compareDate(a.min,e)<=0?null:{matDatepickerMin:{min:a.min,actual:e}}},this._maxValidator=function(t){var e=a._getValidDateOrNull(a._dateAdapter.deserialize(t.value));return!a.max||!e||a._dateAdapter.compareDate(a.max,e)>=0?null:{matDatepickerMax:{max:a.max,actual:e}}},this._filterValidator=function(t){var e=a._getValidDateOrNull(a._dateAdapter.deserialize(t.value));return a._dateFilter&&e&&!a._dateFilter(e)?{matDatepickerFilter:!0}:null},this._validator=g.Validators.compose([this._parseValidator,this._minValidator,this._maxValidator,this._filterValidator]),this._lastValueValid=!1,!this._dateAdapter)throw x("DateAdapter");if(!this._dateFormats)throw x("MAT_DATE_FORMATS");this._localeSubscription=e.localeChanges.subscribe(function(){a.value=a.value})}return Object.defineProperty(t.prototype,"matDatepicker",{set:function(t){var e=this;t&&(this._datepicker=t,this._datepicker._registerInput(this),this._datepickerSubscription.unsubscribe(),this._datepickerSubscription=this._datepicker._selectedChanged.subscribe(function(t){e.value=t,e._cvaOnChange(t),e._onTouched(),e.dateInput.emit(new z(e,e._elementRef.nativeElement)),e.dateChange.emit(new z(e,e._elementRef.nativeElement))}))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"matDatepickerFilter",{set:function(t){this._dateFilter=t,this._validatorOnChange()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{get:function(){return this._value},set:function(t){t=this._dateAdapter.deserialize(t),this._lastValueValid=!t||this._dateAdapter.isValid(t),t=this._getValidDateOrNull(t);var e=this.value;this._value=t,this._formatValue(t),this._dateAdapter.sameDate(e,t)||this._valueChange.emit(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"min",{get:function(){return this._min},set:function(t){this._min=this._getValidDateOrNull(this._dateAdapter.deserialize(t)),this._validatorOnChange()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"max",{get:function(){return this._max},set:function(t){this._max=this._getValidDateOrNull(this._dateAdapter.deserialize(t)),this._validatorOnChange()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return!!this._disabled},set:function(t){var e=Object(p.coerceBooleanProperty)(t),n=this._elementRef.nativeElement;this._disabled!==e&&(this._disabled=e,this._disabledChange.emit(e)),e&&n.blur&&n.blur()},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){this._datepickerSubscription.unsubscribe(),this._localeSubscription.unsubscribe(),this._valueChange.complete(),this._disabledChange.complete()},t.prototype.registerOnValidatorChange=function(t){this._validatorOnChange=t},t.prototype.validate=function(t){return this._validator?this._validator(t):null},t.prototype.getPopupConnectionElementRef=function(){return this.getConnectedOverlayOrigin()},t.prototype.getConnectedOverlayOrigin=function(){return this._formField?this._formField.getConnectedOverlayOrigin():this._elementRef},t.prototype.writeValue=function(t){this.value=t},t.prototype.registerOnChange=function(t){this._cvaOnChange=t},t.prototype.registerOnTouched=function(t){this._onTouched=t},t.prototype.setDisabledState=function(t){this.disabled=t},t.prototype._onKeydown=function(t){var e=t.altKey&&t.keyCode===a.DOWN_ARROW;this._datepicker&&e&&!this._elementRef.nativeElement.readOnly&&(this._datepicker.open(),t.preventDefault())},t.prototype._onInput=function(t){var e=this._dateAdapter.parse(t,this._dateFormats.parse.dateInput);this._lastValueValid=!e||this._dateAdapter.isValid(e),e=this._getValidDateOrNull(e),this._dateAdapter.sameDate(e,this._value)||(this._value=e,this._cvaOnChange(e),this._valueChange.emit(e),this.dateInput.emit(new z(this,this._elementRef.nativeElement)))},t.prototype._onChange=function(){this.dateChange.emit(new z(this,this._elementRef.nativeElement))},t.prototype._getThemePalette=function(){return this._formField?this._formField.color:void 0},t.prototype._onBlur=function(){this.value&&this._formatValue(this.value),this._onTouched()},t.prototype._formatValue=function(t){this._elementRef.nativeElement.value=t?this._dateAdapter.format(t,this._dateFormats.display.dateInput):""},t.prototype._getValidDateOrNull=function(t){return this._dateAdapter.isDateInstance(t)&&this._dateAdapter.isValid(t)?t:null},t.decorators=[{type:r.Directive,args:[{selector:"input[matDatepicker]",providers:[B,V,{provide:_.MAT_INPUT_VALUE_ACCESSOR,useExisting:t}],host:{"[attr.aria-haspopup]":"true","[attr.aria-owns]":"(_datepicker?.opened && _datepicker.id) || null","[attr.min]":"min ? _dateAdapter.toIso8601(min) : null","[attr.max]":"max ? _dateAdapter.toIso8601(max) : null","[disabled]":"disabled","(input)":"_onInput($event.target.value)","(change)":"_onChange()","(blur)":"_onBlur()","(keydown)":"_onKeydown($event)"},exportAs:"matDatepickerInput"}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:s.DateAdapter,decorators:[{type:r.Optional}]},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[s.MAT_DATE_FORMATS]}]},{type:y.MatFormField,decorators:[{type:r.Optional}]}]},t.propDecorators={matDatepicker:[{type:r.Input}],matDatepickerFilter:[{type:r.Input}],value:[{type:r.Input}],min:[{type:r.Input}],max:[{type:r.Input}],disabled:[{type:r.Input}],dateChange:[{type:r.Output}],dateInput:[{type:r.Output}]},t}(),U=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"[matDatepickerToggleIcon]"}]}],t}(),W=function(){function t(t,e,n){this._intl=t,this._changeDetectorRef=e,this._stateChanges=i.Subscription.EMPTY;var r=Number(n);this.tabIndex=r||0===r?r:null}return Object.defineProperty(t.prototype,"disabled",{get:function(){return void 0===this._disabled?this.datepicker.disabled:!!this._disabled},set:function(t){this._disabled=Object(p.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),t.prototype.ngOnChanges=function(t){t.datepicker&&this._watchStateChanges()},t.prototype.ngOnDestroy=function(){this._stateChanges.unsubscribe()},t.prototype.ngAfterContentInit=function(){this._watchStateChanges()},t.prototype._open=function(t){this.datepicker&&!this.disabled&&(this.datepicker.open(),t.stopPropagation())},t.prototype._watchStateChanges=function(){var t=this,e=this.datepicker?this.datepicker._disabledChange:Object(i.of)(),n=this.datepicker&&this.datepicker._datepickerInput?this.datepicker._datepickerInput._disabledChange:Object(i.of)(),r=this.datepicker?Object(i.merge)(this.datepicker.openedStream,this.datepicker.closedStream):Object(i.of)();this._stateChanges.unsubscribe(),this._stateChanges=Object(i.merge)(this._intl.changes,e,n,r).subscribe(function(){return t._changeDetectorRef.markForCheck()})},t.decorators=[{type:r.Component,args:[{selector:"mat-datepicker-toggle",template:'<button #button mat-icon-button type="button" aria-haspopup="true" [attr.aria-label]="_intl.openCalendarLabel" [attr.tabindex]="disabled ? -1 : tabIndex" [disabled]="disabled" [disableRipple]="disableRipple" (click)="_open($event)"><svg *ngIf="!_customIcon" class="mat-datepicker-toggle-default-icon" viewBox="0 0 24 24" width="24px" height="24px" fill="currentColor" focusable="false"><path d="M19 3h-1V1h-2v2H8V1H6v2H5c-1.11 0-1.99.9-1.99 2L3 19c0 1.1.89 2 2 2h14c1.1 0 2-.9 2-2V5c0-1.1-.9-2-2-2zm0 16H5V8h14v11zM7 10h5v5H7z"/></svg><ng-content select="[matDatepickerToggleIcon]"></ng-content></button>',styles:[".mat-form-field-appearance-legacy .mat-form-field-prefix .mat-datepicker-toggle-default-icon,.mat-form-field-appearance-legacy .mat-form-field-suffix .mat-datepicker-toggle-default-icon{width:1em}.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-prefix .mat-datepicker-toggle-default-icon,.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-suffix .mat-datepicker-toggle-default-icon{display:block;width:1.5em;height:1.5em}.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-prefix .mat-icon-button .mat-datepicker-toggle-default-icon,.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-suffix .mat-icon-button .mat-datepicker-toggle-default-icon{margin:auto}"],host:{class:"mat-datepicker-toggle","[attr.tabindex]":"-1","[class.mat-datepicker-toggle-active]":"datepicker && datepicker.opened","[class.mat-accent]":'datepicker && datepicker.color === "accent"',"[class.mat-warn]":'datepicker && datepicker.color === "warn"',"(focus)":"_button.focus()"},exportAs:"matDatepickerToggle",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:w},{type:r.ChangeDetectorRef},{type:String,decorators:[{type:r.Attribute,args:["tabindex"]}]}]},t.propDecorators={datepicker:[{type:r.Input,args:["for"]}],tabIndex:[{type:r.Input}],disabled:[{type:r.Input}],disableRipple:[{type:r.Input}],_customIcon:[{type:r.ContentChild,args:[U]}],_button:[{type:r.ViewChild,args:["button"]}]},t}(),q=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[h.CommonModule,v.MatButtonModule,m.MatDialogModule,f.OverlayModule,b.A11yModule,c.PortalModule],exports:[T,O,L,F,H,W,U,C,S,E,M],declarations:[T,O,L,F,H,W,U,C,S,E,M],providers:[w,P],entryComponents:[F,M]}]}],t}()},"./node_modules/@angular/material/esm5/dialog.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatDialogModule",function(){return R}),n.d(e,"MAT_DIALOG_SCROLL_STRATEGY_FACTORY",function(){return C}),n.d(e,"MAT_DIALOG_SCROLL_STRATEGY_PROVIDER_FACTORY",function(){return E}),n.d(e,"MAT_DIALOG_DATA",function(){return w}),n.d(e,"MAT_DIALOG_DEFAULT_OPTIONS",function(){return j}),n.d(e,"MAT_DIALOG_SCROLL_STRATEGY",function(){return O}),n.d(e,"MAT_DIALOG_SCROLL_STRATEGY_PROVIDER",function(){return S}),n.d(e,"MatDialog",function(){return M}),n.d(e,"throwMatDialogContentAlreadyAttachedError",function(){return _}),n.d(e,"MatDialogContainer",function(){return v}),n.d(e,"MatDialogClose",function(){return k}),n.d(e,"MatDialogTitle",function(){return A}),n.d(e,"MatDialogContent",function(){return D}),n.d(e,"MatDialogActions",function(){return I}),n.d(e,"MatDialogConfig",function(){return m}),n.d(e,"MatDialogRef",function(){return x}),n.d(e,"matDialogAnimations",function(){return y});var r=n("./node_modules/@angular/animations/fesm5/animations.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/common/fesm5/common.js"),s=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),u=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),c=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),l=n("./node_modules/rxjs/_esm5/index.js"),d=n("./node_modules/rxjs/_esm5/operators/index.js"),p=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),f=n("./node_modules/@angular/cdk/esm5/overlay.es5.js"),h=n("./node_modules/@angular/material/esm5/core.es5.js"),m=function(){return function(){this.role="dialog",this.panelClass="",this.hasBackdrop=!0,this.backdropClass="",this.disableClose=!1,this.width="",this.height="",this.maxWidth="80vw",this.data=null,this.ariaDescribedBy=null,this.ariaLabel=null,this.autoFocus=!0,this.restoreFocus=!0,this.closeOnNavigation=!0}}(),g=[Object(r.state)("void, exit",Object(r.style)({opacity:0,transform:"scale(0.7)"})),Object(r.state)("enter",Object(r.style)({transform:"none"})),Object(r.transition)("* => enter",Object(r.animate)("150ms cubic-bezier(0, 0, 0.2, 1)",Object(r.style)({transform:"none",opacity:1}))),Object(r.transition)("* => void, * => exit",Object(r.animate)("75ms cubic-bezier(0.4, 0.0, 0.2, 1)",Object(r.style)({opacity:0})))],y={dialogContainer:Object(r.trigger)("dialogContainer",g),slideDialog:Object(r.trigger)("slideDialog",g)};function _(){throw Error("Attempting to attach dialog content after content is already attached")}var v=function(t){function e(e,n,r,i,a){var s=t.call(this)||this;return s._elementRef=e,s._focusTrapFactory=n,s._changeDetectorRef=r,s._document=i,s._config=a,s._elementFocusedBeforeDialogWasOpened=null,s._state="enter",s._animationStateChanged=new o.EventEmitter,s._ariaLabelledBy=null,s}return Object(i.__extends)(e,t),e.prototype.attachComponentPortal=function(t){return this._portalOutlet.hasAttached()&&_(),this._savePreviouslyFocusedElement(),this._portalOutlet.attachComponentPortal(t)},e.prototype.attachTemplatePortal=function(t){return this._portalOutlet.hasAttached()&&_(),this._savePreviouslyFocusedElement(),this._portalOutlet.attachTemplatePortal(t)},e.prototype._trapFocus=function(){this._focusTrap||(this._focusTrap=this._focusTrapFactory.create(this._elementRef.nativeElement)),this._config.autoFocus&&this._focusTrap.focusInitialElementWhenReady()},e.prototype._restoreFocus=function(){var t=this._elementFocusedBeforeDialogWasOpened;this._config.restoreFocus&&t&&"function"==typeof t.focus&&t.focus(),this._focusTrap&&this._focusTrap.destroy()},e.prototype._savePreviouslyFocusedElement=function(){var t=this;this._document&&(this._elementFocusedBeforeDialogWasOpened=this._document.activeElement,this._elementRef.nativeElement.focus&&Promise.resolve().then(function(){return t._elementRef.nativeElement.focus()}))},e.prototype._onAnimationDone=function(t){"enter"===t.toState?this._trapFocus():"exit"===t.toState&&this._restoreFocus(),this._animationStateChanged.emit(t)},e.prototype._onAnimationStart=function(t){this._animationStateChanged.emit(t)},e.prototype._startExitAnimation=function(){this._state="exit",this._changeDetectorRef.markForCheck()},e.decorators=[{type:o.Component,args:[{selector:"mat-dialog-container",template:"<ng-template cdkPortalOutlet></ng-template>",styles:[".mat-dialog-container{display:block;padding:24px;border-radius:4px;box-sizing:border-box;overflow:auto;outline:0;width:100%;height:100%;min-height:inherit;max-height:inherit}@media (-ms-high-contrast:active){.mat-dialog-container{outline:solid 1px}}.mat-dialog-content{display:block;margin:0 -24px;padding:0 24px;max-height:65vh;overflow:auto;-webkit-overflow-scrolling:touch}.mat-dialog-title{margin:0 0 20px;display:block}.mat-dialog-actions{padding:8px 0;display:flex;flex-wrap:wrap;min-height:52px;align-items:center;margin-bottom:-24px}.mat-dialog-actions[align=end]{justify-content:flex-end}.mat-dialog-actions[align=center]{justify-content:center}.mat-dialog-actions .mat-button+.mat-button,.mat-dialog-actions .mat-button+.mat-raised-button,.mat-dialog-actions .mat-raised-button+.mat-button,.mat-dialog-actions .mat-raised-button+.mat-raised-button{margin-left:8px}[dir=rtl] .mat-dialog-actions .mat-button+.mat-button,[dir=rtl] .mat-dialog-actions .mat-button+.mat-raised-button,[dir=rtl] .mat-dialog-actions .mat-raised-button+.mat-button,[dir=rtl] .mat-dialog-actions .mat-raised-button+.mat-raised-button{margin-left:0;margin-right:8px}"],encapsulation:o.ViewEncapsulation.None,changeDetection:o.ChangeDetectionStrategy.Default,animations:[y.dialogContainer],host:{class:"mat-dialog-container",tabindex:"-1","aria-modal":"true","[attr.id]":"_id","[attr.role]":"_config.role","[attr.aria-labelledby]":"_config.ariaLabel ? null : _ariaLabelledBy","[attr.aria-label]":"_config.ariaLabel","[attr.aria-describedby]":"_config.ariaDescribedBy || null","[@dialogContainer]":"_state","(@dialogContainer.start)":"_onAnimationStart($event)","(@dialogContainer.done)":"_onAnimationDone($event)"}}]}],e.ctorParameters=function(){return[{type:o.ElementRef},{type:u.FocusTrapFactory},{type:o.ChangeDetectorRef},{type:void 0,decorators:[{type:o.Optional},{type:o.Inject,args:[a.DOCUMENT]}]},{type:m}]},e.propDecorators={_portalOutlet:[{type:o.ViewChild,args:[s.CdkPortalOutlet]}]},e}(s.BasePortalOutlet),b=0,x=function(){function t(t,e,n,r){void 0===r&&(r="mat-dialog-"+b++);var i=this;this._overlayRef=t,this._containerInstance=e,this.id=r,this.disableClose=this._containerInstance._config.disableClose,this._afterOpened=new l.Subject,this._afterClosed=new l.Subject,this._beforeClosed=new l.Subject,e._id=r,e._animationStateChanged.pipe(Object(d.filter)(function(t){return"done"===t.phaseName&&"enter"===t.toState}),Object(d.take)(1)).subscribe(function(){i._afterOpened.next(),i._afterOpened.complete()}),e._animationStateChanged.pipe(Object(d.filter)(function(t){return"done"===t.phaseName&&"exit"===t.toState}),Object(d.take)(1)).subscribe(function(){return i._overlayRef.dispose()}),t.detachments().subscribe(function(){i._beforeClosed.next(i._result),i._beforeClosed.complete(),i._afterClosed.next(i._result),i._afterClosed.complete(),i.componentInstance=null,i._overlayRef.dispose()}),t.keydownEvents().pipe(Object(d.filter)(function(t){return t.keyCode===c.ESCAPE&&!i.disableClose})).subscribe(function(){return i.close()})}return t.prototype.close=function(t){var e=this;this._result=t,this._containerInstance._animationStateChanged.pipe(Object(d.filter)(function(t){return"start"===t.phaseName}),Object(d.take)(1)).subscribe(function(){e._beforeClosed.next(t),e._beforeClosed.complete(),e._overlayRef.detachBackdrop()}),this._containerInstance._startExitAnimation()},t.prototype.afterOpened=function(){return this._afterOpened.asObservable()},t.prototype.afterClosed=function(){return this._afterClosed.asObservable()},t.prototype.beforeClosed=function(){return this._beforeClosed.asObservable()},t.prototype.backdropClick=function(){return this._overlayRef.backdropClick()},t.prototype.keydownEvents=function(){return this._overlayRef.keydownEvents()},t.prototype.updatePosition=function(t){var e=this._getPositionStrategy();return t&&(t.left||t.right)?t.left?e.left(t.left):e.right(t.right):e.centerHorizontally(),t&&(t.top||t.bottom)?t.top?e.top(t.top):e.bottom(t.bottom):e.centerVertically(),this._overlayRef.updatePosition(),this},t.prototype.updateSize=function(t,e){return void 0===t&&(t=""),void 0===e&&(e=""),this._getPositionStrategy().width(t).height(e),this._overlayRef.updatePosition(),this},t.prototype.afterOpen=function(){return this.afterOpened()},t.prototype.beforeClose=function(){return this.beforeClosed()},t.prototype._getPositionStrategy=function(){return this._overlayRef.getConfig().positionStrategy},t}(),w=new o.InjectionToken("MatDialogData"),j=new o.InjectionToken("mat-dialog-default-options"),O=new o.InjectionToken("mat-dialog-scroll-strategy");function C(t){return function(){return t.scrollStrategies.block()}}function E(t){return function(){return t.scrollStrategies.block()}}var S={provide:O,deps:[f.Overlay],useFactory:E},M=function(){function t(t,e,n,r,i,o,a){var s=this;this._overlay=t,this._injector=e,this._location=n,this._defaultOptions=r,this._parentDialog=o,this._overlayContainer=a,this._openDialogsAtThisLevel=[],this._afterAllClosedAtThisLevel=new l.Subject,this._afterOpenedAtThisLevel=new l.Subject,this._ariaHiddenElements=new Map,this.afterAllClosed=Object(l.defer)(function(){return s.openDialogs.length?s._afterAllClosed:s._afterAllClosed.pipe(Object(d.startWith)(void 0))}),this._scrollStrategy=i}return Object.defineProperty(t.prototype,"openDialogs",{get:function(){return this._parentDialog?this._parentDialog.openDialogs:this._openDialogsAtThisLevel},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"afterOpened",{get:function(){return this._parentDialog?this._parentDialog.afterOpened:this._afterOpenedAtThisLevel},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"afterOpen",{get:function(){return this.afterOpened},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_afterAllClosed",{get:function(){var t=this._parentDialog;return t?t._afterAllClosed:this._afterAllClosedAtThisLevel},enumerable:!0,configurable:!0}),t.prototype.open=function(t,e){var n,r,o=this;if(n=e,r=this._defaultOptions||new m,(e=Object(i.__assign)({},r,n)).id&&this.getDialogById(e.id))throw Error('Dialog with id "'+e.id+'" exists already. The dialog id must be unique.');var a=this._createOverlay(e),s=this._attachDialogContainer(a,e),u=this._attachDialogContent(t,s,a,e);return this.openDialogs.length||this._hideNonDialogContentFromAssistiveTechnology(),this.openDialogs.push(u),u.afterClosed().subscribe(function(){return o._removeOpenDialog(u)}),this.afterOpened.next(u),u},t.prototype.closeAll=function(){this._closeDialogs(this.openDialogs)},t.prototype.getDialogById=function(t){return this.openDialogs.find(function(e){return e.id===t})},t.prototype.ngOnDestroy=function(){this._closeDialogs(this._openDialogsAtThisLevel),this._afterAllClosedAtThisLevel.complete(),this._afterOpenedAtThisLevel.complete()},t.prototype._createOverlay=function(t){var e=this._getOverlayConfig(t);return this._overlay.create(e)},t.prototype._getOverlayConfig=function(t){var e=new f.OverlayConfig({positionStrategy:this._overlay.position().global(),scrollStrategy:t.scrollStrategy||this._scrollStrategy(),panelClass:t.panelClass,hasBackdrop:t.hasBackdrop,direction:t.direction,minWidth:t.minWidth,minHeight:t.minHeight,maxWidth:t.maxWidth,maxHeight:t.maxHeight,disposeOnNavigation:t.closeOnNavigation});return t.backdropClass&&(e.backdropClass=t.backdropClass),e},t.prototype._attachDialogContainer=function(t,e){var n=e&&e.viewContainerRef&&e.viewContainerRef.injector,r=new s.PortalInjector(n||this._injector,new WeakMap([[m,e]])),i=new s.ComponentPortal(v,e.viewContainerRef,r);return t.attach(i).instance},t.prototype._attachDialogContent=function(t,e,n,r){var i=new x(n,e,this._location,r.id);if(r.hasBackdrop&&n.backdropClick().subscribe(function(){i.disableClose||i.close()}),t instanceof o.TemplateRef)e.attachTemplatePortal(new s.TemplatePortal(t,null,{$implicit:r.data,dialogRef:i}));else{var a=this._createInjector(r,i,e),u=e.attachComponentPortal(new s.ComponentPortal(t,void 0,a));i.componentInstance=u.instance}return i.updateSize(r.width,r.height).updatePosition(r.position),i},t.prototype._createInjector=function(t,e,n){var r=t&&t.viewContainerRef&&t.viewContainerRef.injector,i=new WeakMap([[v,n],[w,t.data],[x,e]]);return!t.direction||r&&r.get(p.Directionality,null)||i.set(p.Directionality,{value:t.direction,change:Object(l.of)()}),new s.PortalInjector(r||this._injector,i)},t.prototype._removeOpenDialog=function(t){var e=this.openDialogs.indexOf(t);e>-1&&(this.openDialogs.splice(e,1),this.openDialogs.length||(this._ariaHiddenElements.forEach(function(t,e){t?e.setAttribute("aria-hidden",t):e.removeAttribute("aria-hidden")}),this._ariaHiddenElements.clear(),this._afterAllClosed.next()))},t.prototype._hideNonDialogContentFromAssistiveTechnology=function(){var t=this._overlayContainer.getContainerElement();if(t.parentElement)for(var e=t.parentElement.children,n=e.length-1;n>-1;n--){var r=e[n];r===t||"SCRIPT"===r.nodeName||"STYLE"===r.nodeName||r.hasAttribute("aria-live")||(this._ariaHiddenElements.set(r,r.getAttribute("aria-hidden")),r.setAttribute("aria-hidden","true"))}},t.prototype._closeDialogs=function(t){for(var e=t.length;e--;)t[e].close()},t.decorators=[{type:o.Injectable}],t.ctorParameters=function(){return[{type:f.Overlay},{type:o.Injector},{type:a.Location,decorators:[{type:o.Optional}]},{type:m,decorators:[{type:o.Optional},{type:o.Inject,args:[j]}]},{type:void 0,decorators:[{type:o.Inject,args:[O]}]},{type:t,decorators:[{type:o.Optional},{type:o.SkipSelf}]},{type:f.OverlayContainer}]},t}();var T=0,k=function(){function t(t,e,n){this.dialogRef=t,this._elementRef=e,this._dialog=n,this.ariaLabel="Close dialog"}return t.prototype.ngOnInit=function(){this.dialogRef||(this.dialogRef=P(this._elementRef,this._dialog.openDialogs))},t.prototype.ngOnChanges=function(t){var e=t._matDialogClose||t._matDialogCloseResult;e&&(this.dialogResult=e.currentValue)},t.decorators=[{type:o.Directive,args:[{selector:"button[mat-dialog-close], button[matDialogClose]",exportAs:"matDialogClose",host:{"(click)":"dialogRef.close(dialogResult)","[attr.aria-label]":"ariaLabel",type:"button"}}]}],t.ctorParameters=function(){return[{type:x,decorators:[{type:o.Optional}]},{type:o.ElementRef},{type:M}]},t.propDecorators={ariaLabel:[{type:o.Input,args:["aria-label"]}],dialogResult:[{type:o.Input,args:["mat-dialog-close"]}],_matDialogClose:[{type:o.Input,args:["matDialogClose"]}]},t}(),A=function(){function t(t,e,n){this._dialogRef=t,this._elementRef=e,this._dialog=n,this.id="mat-dialog-title-"+T++}return t.prototype.ngOnInit=function(){var t=this;this._dialogRef||(this._dialogRef=P(this._elementRef,this._dialog.openDialogs)),this._dialogRef&&Promise.resolve().then(function(){var e=t._dialogRef._containerInstance;e&&!e._ariaLabelledBy&&(e._ariaLabelledBy=t.id)})},t.decorators=[{type:o.Directive,args:[{selector:"[mat-dialog-title], [matDialogTitle]",exportAs:"matDialogTitle",host:{class:"mat-dialog-title","[id]":"id"}}]}],t.ctorParameters=function(){return[{type:x,decorators:[{type:o.Optional}]},{type:o.ElementRef},{type:M}]},t.propDecorators={id:[{type:o.Input}]},t}(),D=function(){function t(){}return t.decorators=[{type:o.Directive,args:[{selector:"[mat-dialog-content], mat-dialog-content, [matDialogContent]",host:{class:"mat-dialog-content"}}]}],t}(),I=function(){function t(){}return t.decorators=[{type:o.Directive,args:[{selector:"[mat-dialog-actions], mat-dialog-actions, [matDialogActions]",host:{class:"mat-dialog-actions"}}]}],t}();function P(t,e){for(var n=t.nativeElement.parentElement;n&&!n.classList.contains("mat-dialog-container");)n=n.parentElement;return n?e.find(function(t){return t.id===n.id}):null}var R=function(){function t(){}return t.decorators=[{type:o.NgModule,args:[{imports:[a.CommonModule,f.OverlayModule,s.PortalModule,h.MatCommonModule],exports:[v,k,A,D,I,h.MatCommonModule],declarations:[v,k,A,I,D],providers:[M,S],entryComponents:[v]}]}],t}()},"./node_modules/@angular/material/esm5/divider.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatDivider",function(){return s}),n.d(e,"MatDividerModule",function(){return u});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),o=n("./node_modules/@angular/common/fesm5/common.js"),a=n("./node_modules/@angular/material/esm5/core.es5.js"),s=function(){function t(){this._vertical=!1,this._inset=!1}return Object.defineProperty(t.prototype,"vertical",{get:function(){return this._vertical},set:function(t){this._vertical=Object(i.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"inset",{get:function(){return this._inset},set:function(t){this._inset=Object(i.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),t.decorators=[{type:r.Component,args:[{selector:"mat-divider",host:{role:"separator","[attr.aria-orientation]":'vertical ? "vertical" : "horizontal"',"[class.mat-divider-vertical]":"vertical","[class.mat-divider-horizontal]":"!vertical","[class.mat-divider-inset]":"inset",class:"mat-divider"},template:"",styles:[".mat-divider{display:block;margin:0;border-top-width:1px;border-top-style:solid}.mat-divider.mat-divider-vertical{border-top:0;border-right-width:1px;border-right-style:solid}.mat-divider.mat-divider-inset{margin-left:80px}[dir=rtl] .mat-divider.mat-divider-inset{margin-left:auto;margin-right:80px}"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],t.propDecorators={vertical:[{type:r.Input}],inset:[{type:r.Input}]},t}(),u=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[a.MatCommonModule,o.CommonModule],exports:[s,a.MatCommonModule],declarations:[s]}]}],t}()},"./node_modules/@angular/material/esm5/expansion.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatExpansionModule",function(){return M}),n.d(e,"MatAccordion",function(){return S}),n.d(e,"MAT_ACCORDION",function(){return g}),n.d(e,"MAT_EXPANSION_PANEL_DEFAULT_OPTIONS",function(){return x}),n.d(e,"MatExpansionPanel",function(){return w}),n.d(e,"MatExpansionPanelActionRow",function(){return j}),n.d(e,"MatExpansionPanelHeader",function(){return O}),n.d(e,"MatExpansionPanelDescription",function(){return C}),n.d(e,"MatExpansionPanelTitle",function(){return E}),n.d(e,"MatExpansionPanelContent",function(){return v}),n.d(e,"EXPANSION_PANEL_ANIMATION_TIMING",function(){return y}),n.d(e,"matExpansionAnimations",function(){return _});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/animations/fesm5/animations.js"),o=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/cdk/esm5/accordion.es5.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),c=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),l=n("./node_modules/@angular/common/fesm5/common.js"),d=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),p=n("./node_modules/rxjs/_esm5/index.js"),f=n("./node_modules/rxjs/_esm5/operators/index.js"),h=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),m=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),g=new r.InjectionToken("MAT_ACCORDION"),y="225ms cubic-bezier(0.4,0.0,0.2,1)",_={indicatorRotate:Object(i.trigger)("indicatorRotate",[Object(i.state)("collapsed, void",Object(i.style)({transform:"rotate(0deg)"})),Object(i.state)("expanded",Object(i.style)({transform:"rotate(180deg)"})),Object(i.transition)("expanded <=> collapsed, void => collapsed",Object(i.animate)(y))]),expansionHeaderHeight:Object(i.trigger)("expansionHeight",[Object(i.state)("collapsed, void",Object(i.style)({height:"{{collapsedHeight}}"}),{params:{collapsedHeight:"48px"}}),Object(i.state)("expanded",Object(i.style)({height:"{{expandedHeight}}"}),{params:{expandedHeight:"64px"}}),Object(i.transition)("expanded <=> collapsed, void => collapsed",Object(i.group)([Object(i.query)("@indicatorRotate",Object(i.animateChild)(),{optional:!0}),Object(i.animate)(y)]))]),bodyExpansion:Object(i.trigger)("bodyExpansion",[Object(i.state)("collapsed, void",Object(i.style)({height:"0px",visibility:"hidden"})),Object(i.state)("expanded",Object(i.style)({height:"*",visibility:"visible"})),Object(i.transition)("expanded <=> collapsed, void => collapsed",Object(i.animate)(y))])},v=function(){function t(t){this._template=t}return t.decorators=[{type:r.Directive,args:[{selector:"ng-template[matExpansionPanelContent]"}]}],t.ctorParameters=function(){return[{type:r.TemplateRef}]},t}(),b=0,x=new r.InjectionToken("MAT_EXPANSION_PANEL_DEFAULT_OPTIONS"),w=function(t){function e(e,n,i,o,a,s,u){var c=t.call(this,e,n,i)||this;return c._viewContainerRef=o,c._animationMode=s,c._hideToggle=!1,c.afterExpand=new r.EventEmitter,c.afterCollapse=new r.EventEmitter,c._inputChanges=new p.Subject,c._headerId="mat-expansion-panel-header-"+b++,c._bodyAnimationDone=new p.Subject,c.accordion=e,c._document=a,c._bodyAnimationDone.pipe(Object(f.distinctUntilChanged)(function(t,e){return t.fromState===e.fromState&&t.toState===e.toState})).subscribe(function(t){"void"!==t.fromState&&("expanded"===t.toState?c.afterExpand.emit():"collapsed"===t.toState&&c.afterCollapse.emit())}),u&&(c.hideToggle=u.hideToggle),c}return Object(o.__extends)(e,t),Object.defineProperty(e.prototype,"hideToggle",{get:function(){return this._hideToggle||this.accordion&&this.accordion.hideToggle},set:function(t){this._hideToggle=Object(s.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e.prototype._hasSpacing=function(){return!!this.accordion&&"default"===(this.expanded?this.accordion.displayMode:this._getExpandedState())},e.prototype._getExpandedState=function(){return this.expanded?"expanded":"collapsed"},e.prototype.ngAfterContentInit=function(){var t=this;this._lazyContent&&this.opened.pipe(Object(f.startWith)(null),Object(f.filter)(function(){return t.expanded&&!t._portal}),Object(f.take)(1)).subscribe(function(){t._portal=new c.TemplatePortal(t._lazyContent._template,t._viewContainerRef)})},e.prototype.ngOnChanges=function(t){this._inputChanges.next(t)},e.prototype.ngOnDestroy=function(){t.prototype.ngOnDestroy.call(this),this._bodyAnimationDone.complete(),this._inputChanges.complete()},e.prototype._containsFocus=function(){if(this._body&&this._document){var t=this._document.activeElement,e=this._body.nativeElement;return t===e||e.contains(t)}return!1},e.decorators=[{type:r.Component,args:[{styles:[".mat-expansion-panel{box-sizing:content-box;display:block;margin:0;border-radius:4px;overflow:hidden;transition:margin 225ms cubic-bezier(.4,0,.2,1),box-shadow 280ms cubic-bezier(.4,0,.2,1)}.mat-accordion .mat-expansion-panel:not(.mat-expanded),.mat-accordion .mat-expansion-panel:not(.mat-expansion-panel-spacing){border-radius:0}.mat-accordion .mat-expansion-panel:first-of-type{border-top-right-radius:4px;border-top-left-radius:4px}.mat-accordion .mat-expansion-panel:last-of-type{border-bottom-right-radius:4px;border-bottom-left-radius:4px}@media (-ms-high-contrast:active){.mat-expansion-panel{outline:solid 1px}}.mat-expansion-panel._mat-animation-noopable,.mat-expansion-panel.ng-animate-disabled,.ng-animate-disabled .mat-expansion-panel{transition:none}.mat-expansion-panel-content{display:flex;flex-direction:column;overflow:visible}.mat-expansion-panel-body{padding:0 24px 16px}.mat-expansion-panel-spacing{margin:16px 0}.mat-accordion>.mat-expansion-panel-spacing:first-child,.mat-accordion>:first-child:not(.mat-expansion-panel) .mat-expansion-panel-spacing{margin-top:0}.mat-accordion>.mat-expansion-panel-spacing:last-child,.mat-accordion>:last-child:not(.mat-expansion-panel) .mat-expansion-panel-spacing{margin-bottom:0}.mat-action-row{border-top-style:solid;border-top-width:1px;display:flex;flex-direction:row;justify-content:flex-end;padding:16px 8px 16px 24px}.mat-action-row button.mat-button{margin-left:8px}[dir=rtl] .mat-action-row button.mat-button{margin-left:0;margin-right:8px}"],selector:"mat-expansion-panel",exportAs:"matExpansionPanel",template:'<ng-content select="mat-expansion-panel-header"></ng-content><div class="mat-expansion-panel-content" role="region" [@bodyExpansion]="_getExpandedState()" (@bodyExpansion.done)="_bodyAnimationDone.next($event)" [attr.aria-labelledby]="_headerId" [id]="id" #body><div class="mat-expansion-panel-body"><ng-content></ng-content><ng-template [cdkPortalOutlet]="_portal"></ng-template></div><ng-content select="mat-action-row"></ng-content></div>',encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,inputs:["disabled","expanded"],outputs:["opened","closed","expandedChange"],animations:[_.bodyExpansion],providers:[{provide:g,useValue:void 0}],host:{class:"mat-expansion-panel","[class.mat-expanded]":"expanded","[class._mat-animation-noopable]":'_animationMode === "NoopAnimations"',"[class.mat-expansion-panel-spacing]":"_hasSpacing()"}}]}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.Optional},{type:r.SkipSelf},{type:r.Inject,args:[g]}]},{type:r.ChangeDetectorRef},{type:u.UniqueSelectionDispatcher},{type:r.ViewContainerRef},{type:void 0,decorators:[{type:r.Inject,args:[l.DOCUMENT]}]},{type:String,decorators:[{type:r.Optional},{type:r.Inject,args:[d.ANIMATION_MODULE_TYPE]}]},{type:void 0,decorators:[{type:r.Inject,args:[x]},{type:r.Optional}]}]},e.propDecorators={hideToggle:[{type:r.Input}],afterExpand:[{type:r.Output}],afterCollapse:[{type:r.Output}],_lazyContent:[{type:r.ContentChild,args:[v]}],_body:[{type:r.ViewChild,args:["body"]}]},e}(a.CdkAccordionItem),j=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"mat-action-row",host:{class:"mat-action-row"}}]}],t}(),O=function(){function t(t,e,n,r,i){var o=this;this.panel=t,this._element=e,this._focusMonitor=n,this._changeDetectorRef=r,this._parentChangeSubscription=p.Subscription.EMPTY;var a=t.accordion?t.accordion._stateChanges.pipe(Object(f.filter)(function(t){return!!t.hideToggle})):p.EMPTY;this._parentChangeSubscription=Object(p.merge)(t.opened,t.closed,a,t._inputChanges.pipe(Object(f.filter)(function(t){return!(!t.hideToggle&&!t.disabled)}))).subscribe(function(){return o._changeDetectorRef.markForCheck()}),t.closed.pipe(Object(f.filter)(function(){return t._containsFocus()})).subscribe(function(){return n.focusVia(e,"program")}),n.monitor(e).subscribe(function(e){e&&t.accordion&&t.accordion._handleHeaderFocus(o)}),i&&(this.expandedHeight=i.expandedHeight,this.collapsedHeight=i.collapsedHeight)}return Object.defineProperty(t.prototype,"disabled",{get:function(){return this.panel.disabled},enumerable:!0,configurable:!0}),t.prototype._toggle=function(){this.panel.toggle()},t.prototype._isExpanded=function(){return this.panel.expanded},t.prototype._getExpandedState=function(){return this.panel._getExpandedState()},t.prototype._getPanelId=function(){return this.panel.id},t.prototype._showToggle=function(){return!this.panel.hideToggle&&!this.panel.disabled},t.prototype._keydown=function(t){switch(t.keyCode){case m.SPACE:case m.ENTER:Object(m.hasModifierKey)(t)||(t.preventDefault(),this._toggle());break;default:return void(this.panel.accordion&&this.panel.accordion._handleHeaderKeydown(t))}},t.prototype.focus=function(t){void 0===t&&(t="program"),this._focusMonitor.focusVia(this._element,t)},t.prototype.ngOnDestroy=function(){this._parentChangeSubscription.unsubscribe(),this._focusMonitor.stopMonitoring(this._element)},t.decorators=[{type:r.Component,args:[{selector:"mat-expansion-panel-header",styles:[".mat-expansion-panel-header{display:flex;flex-direction:row;align-items:center;padding:0 24px;border-radius:inherit}.mat-expansion-panel-header:focus,.mat-expansion-panel-header:hover{outline:0}.mat-expansion-panel-header.mat-expanded:focus,.mat-expansion-panel-header.mat-expanded:hover{background:inherit}.mat-expansion-panel-header:not([aria-disabled=true]){cursor:pointer}.mat-content{display:flex;flex:1;flex-direction:row;overflow:hidden}.mat-expansion-panel-header-description,.mat-expansion-panel-header-title{display:flex;flex-grow:1;margin-right:16px}[dir=rtl] .mat-expansion-panel-header-description,[dir=rtl] .mat-expansion-panel-header-title{margin-right:0;margin-left:16px}.mat-expansion-panel-header-description{flex-grow:2}.mat-expansion-indicator::after{border-style:solid;border-width:0 2px 2px 0;content:'';display:inline-block;padding:3px;transform:rotate(45deg);vertical-align:middle}"],template:'<span class="mat-content"><ng-content select="mat-panel-title"></ng-content><ng-content select="mat-panel-description"></ng-content><ng-content></ng-content></span><span [@indicatorRotate]="_getExpandedState()" *ngIf="_showToggle()" class="mat-expansion-indicator"></span>',encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,animations:[_.indicatorRotate,_.expansionHeaderHeight],host:{class:"mat-expansion-panel-header",role:"button","[attr.id]":"panel._headerId","[attr.tabindex]":"disabled ? -1 : 0","[attr.aria-controls]":"_getPanelId()","[attr.aria-expanded]":"_isExpanded()","[attr.aria-disabled]":"panel.disabled","[class.mat-expanded]":"_isExpanded()","(click)":"_toggle()","(keydown)":"_keydown($event)","[@expansionHeight]":"{\n        value: _getExpandedState(),\n        params: {\n          collapsedHeight: collapsedHeight,\n          expandedHeight: expandedHeight\n        }\n    }"}}]}],t.ctorParameters=function(){return[{type:w,decorators:[{type:r.Host}]},{type:r.ElementRef},{type:h.FocusMonitor},{type:r.ChangeDetectorRef},{type:void 0,decorators:[{type:r.Inject,args:[x]},{type:r.Optional}]}]},t.propDecorators={expandedHeight:[{type:r.Input}],collapsedHeight:[{type:r.Input}]},t}(),C=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"mat-panel-description",host:{class:"mat-expansion-panel-header-description"}}]}],t}(),E=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"mat-panel-title",host:{class:"mat-expansion-panel-header-title"}}]}],t}(),S=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._hideToggle=!1,e.displayMode="default",e}return Object(o.__extends)(e,t),Object.defineProperty(e.prototype,"hideToggle",{get:function(){return this._hideToggle},set:function(t){this._hideToggle=Object(s.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){this._keyManager=new h.FocusKeyManager(this._headers).withWrap()},e.prototype._handleHeaderKeydown=function(t){var e=t.keyCode,n=this._keyManager;e===m.HOME?(n.setFirstItemActive(),t.preventDefault()):e===m.END?(n.setLastItemActive(),t.preventDefault()):this._keyManager.onKeydown(t)},e.prototype._handleHeaderFocus=function(t){this._keyManager.updateActiveItem(t)},e.decorators=[{type:r.Directive,args:[{selector:"mat-accordion",exportAs:"matAccordion",inputs:["multi"],providers:[{provide:g,useExisting:e}],host:{class:"mat-accordion"}}]}],e.propDecorators={_headers:[{type:r.ContentChildren,args:[O,{descendants:!0}]}],hideToggle:[{type:r.Input}],displayMode:[{type:r.Input}]},e}(a.CdkAccordion),M=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[l.CommonModule,a.CdkAccordionModule,c.PortalModule],exports:[S,w,j,O,E,C,v],declarations:[S,w,j,O,E,C,v]}]}],t}()},"./node_modules/@angular/material/esm5/form-field.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatFormFieldModule",function(){return I}),n.d(e,"MatError",function(){return g}),n.d(e,"MatFormFieldBase",function(){return T}),n.d(e,"_MatFormFieldMixinBase",function(){return k}),n.d(e,"MAT_FORM_FIELD_DEFAULT_OPTIONS",function(){return A}),n.d(e,"MatFormField",function(){return D}),n.d(e,"MatFormFieldControl",function(){return _}),n.d(e,"getMatFormFieldPlaceholderConflictError",function(){return v}),n.d(e,"getMatFormFieldDuplicatedHintError",function(){return b}),n.d(e,"getMatFormFieldMissingControlError",function(){return x}),n.d(e,"MatHint",function(){return j}),n.d(e,"MatPlaceholder",function(){return C}),n.d(e,"MatPrefix",function(){return E}),n.d(e,"MatSuffix",function(){return S}),n.d(e,"MatLabel",function(){return O}),n.d(e,"matFormFieldAnimations",function(){return y});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/animations/fesm5/animations.js"),o=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/@angular/material/esm5/core.es5.js"),c=n("./node_modules/rxjs/_esm5/index.js"),l=n("./node_modules/rxjs/_esm5/operators/index.js"),d=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),p=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),f=n("./node_modules/@angular/common/fesm5/common.js"),h=n("./node_modules/@angular/cdk/esm5/observers.es5.js"),m=0,g=function(){function t(){this.id="mat-error-"+m++}return t.decorators=[{type:r.Directive,args:[{selector:"mat-error",host:{class:"mat-error",role:"alert","[attr.id]":"id"}}]}],t.propDecorators={id:[{type:r.Input}]},t}(),y={transitionMessages:Object(i.trigger)("transitionMessages",[Object(i.state)("enter",Object(i.style)({opacity:1,transform:"translateY(0%)"})),Object(i.transition)("void => enter",[Object(i.style)({opacity:0,transform:"translateY(-100%)"}),Object(i.animate)("300ms cubic-bezier(0.55, 0, 0.55, 0.2)")])])},_=function(){return function(){}}();function v(){return Error("Placeholder attribute and child element were both specified.")}function b(t){return Error("A hint was already declared for 'align=\""+t+"\"'.")}function x(){return Error("mat-form-field must contain a MatFormFieldControl.")}var w=0,j=function(){function t(){this.align="start",this.id="mat-hint-"+w++}return t.decorators=[{type:r.Directive,args:[{selector:"mat-hint",host:{class:"mat-hint","[class.mat-right]":'align == "end"',"[attr.id]":"id","[attr.align]":"null"}}]}],t.propDecorators={align:[{type:r.Input}],id:[{type:r.Input}]},t}(),O=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"mat-label"}]}],t}(),C=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"mat-placeholder"}]}],t}(),E=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"[matPrefix]"}]}],t}(),S=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"[matSuffix]"}]}],t}(),M=0,T=function(){return function(t){this._elementRef=t}}(),k=Object(u.mixinColor)(T,"primary"),A=new r.InjectionToken("MAT_FORM_FIELD_DEFAULT_OPTIONS"),D=function(t){function e(e,n,r,i,o,a,s,u){var l=t.call(this,e)||this;return l._elementRef=e,l._changeDetectorRef=n,l._dir=i,l._defaults=o,l._platform=a,l._ngZone=s,l._outlineGapCalculationNeededImmediately=!1,l._outlineGapCalculationNeededOnStable=!1,l._destroyed=new c.Subject,l._showAlwaysAnimate=!1,l._subscriptAnimationState="",l._hintLabel="",l._hintLabelId="mat-hint-"+M++,l._labelId="mat-form-field-label-"+M++,l._labelOptions=r||{},l.floatLabel=l._labelOptions.float||"auto",l._animationsEnabled="NoopAnimations"!==u,l.appearance=o&&o.appearance?o.appearance:"legacy",l}return Object(o.__extends)(e,t),Object.defineProperty(e.prototype,"appearance",{get:function(){return this._appearance},set:function(t){var e=this._appearance;this._appearance=t||this._defaults&&this._defaults.appearance||"legacy","outline"===this._appearance&&e!==t&&this._updateOutlineGapOnStable()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hideRequiredMarker",{get:function(){return this._hideRequiredMarker},set:function(t){this._hideRequiredMarker=Object(s.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_shouldAlwaysFloat",{get:function(){return"always"===this.floatLabel&&!this._showAlwaysAnimate},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_canLabelFloat",{get:function(){return"never"!==this.floatLabel},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hintLabel",{get:function(){return this._hintLabel},set:function(t){this._hintLabel=t,this._processHints()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"floatLabel",{get:function(){return"legacy"!==this.appearance&&"never"===this._floatLabel?"auto":this._floatLabel},set:function(t){t!==this._floatLabel&&(this._floatLabel=t||this._labelOptions.float||"auto",this._changeDetectorRef.markForCheck())},enumerable:!0,configurable:!0}),e.prototype.getConnectedOverlayOrigin=function(){return this._connectionContainerRef||this._elementRef},e.prototype.ngAfterContentInit=function(){var t=this;this._validateControlChild();var e=this._control;e.controlType&&this._elementRef.nativeElement.classList.add("mat-form-field-type-"+e.controlType),e.stateChanges.pipe(Object(l.startWith)(null)).subscribe(function(){t._validatePlaceholders(),t._syncDescribedByIds(),t._changeDetectorRef.markForCheck()}),e.ngControl&&e.ngControl.valueChanges&&e.ngControl.valueChanges.pipe(Object(l.takeUntil)(this._destroyed)).subscribe(function(){return t._changeDetectorRef.markForCheck()}),this._ngZone&&this._ngZone.onStable.asObservable().pipe(Object(l.takeUntil)(this._destroyed)).subscribe(function(){t._outlineGapCalculationNeededOnStable&&t.updateOutlineGap()}),Object(c.merge)(this._prefixChildren.changes,this._suffixChildren.changes).subscribe(function(){t._updateOutlineGapOnStable(),t._changeDetectorRef.markForCheck()}),this._hintChildren.changes.pipe(Object(l.startWith)(null)).subscribe(function(){t._processHints(),t._changeDetectorRef.markForCheck()}),this._errorChildren.changes.pipe(Object(l.startWith)(null)).subscribe(function(){t._syncDescribedByIds(),t._changeDetectorRef.markForCheck()}),this._dir&&this._dir.change.pipe(Object(l.takeUntil)(this._destroyed)).subscribe(function(){return t.updateOutlineGap()})},e.prototype.ngAfterContentChecked=function(){this._validateControlChild(),this._outlineGapCalculationNeededImmediately&&this.updateOutlineGap()},e.prototype.ngAfterViewInit=function(){this._subscriptAnimationState="enter",this._changeDetectorRef.detectChanges()},e.prototype.ngOnDestroy=function(){this._destroyed.next(),this._destroyed.complete()},e.prototype._shouldForward=function(t){var e=this._control?this._control.ngControl:null;return e&&e[t]},e.prototype._hasPlaceholder=function(){return!!(this._control&&this._control.placeholder||this._placeholderChild)},e.prototype._hasLabel=function(){return!!this._labelChild},e.prototype._shouldLabelFloat=function(){return this._canLabelFloat&&(this._control.shouldLabelFloat||this._shouldAlwaysFloat)},e.prototype._hideControlPlaceholder=function(){return"legacy"===this.appearance&&!this._hasLabel()||this._hasLabel()&&!this._shouldLabelFloat()},e.prototype._hasFloatingLabel=function(){return this._hasLabel()||"legacy"===this.appearance&&this._hasPlaceholder()},e.prototype._getDisplayedMessages=function(){return this._errorChildren&&this._errorChildren.length>0&&this._control.errorState?"error":"hint"},e.prototype._animateAndLockLabel=function(){var t=this;this._hasFloatingLabel()&&this._canLabelFloat&&(this._animationsEnabled&&(this._showAlwaysAnimate=!0,Object(c.fromEvent)(this._label.nativeElement,"transitionend").pipe(Object(l.take)(1)).subscribe(function(){t._showAlwaysAnimate=!1})),this.floatLabel="always",this._changeDetectorRef.markForCheck())},e.prototype._validatePlaceholders=function(){if(this._control.placeholder&&this._placeholderChild)throw v()},e.prototype._processHints=function(){this._validateHints(),this._syncDescribedByIds()},e.prototype._validateHints=function(){var t,e,n=this;this._hintChildren&&this._hintChildren.forEach(function(r){if("start"===r.align){if(t||n.hintLabel)throw b("start");t=r}else if("end"===r.align){if(e)throw b("end");e=r}})},e.prototype._syncDescribedByIds=function(){if(this._control){var t=[];if("hint"===this._getDisplayedMessages()){var e=this._hintChildren?this._hintChildren.find(function(t){return"start"===t.align}):null,n=this._hintChildren?this._hintChildren.find(function(t){return"end"===t.align}):null;e?t.push(e.id):this._hintLabel&&t.push(this._hintLabelId),n&&t.push(n.id)}else this._errorChildren&&(t=this._errorChildren.map(function(t){return t.id}));this._control.setDescribedByIds(t)}},e.prototype._validateControlChild=function(){if(!this._control)throw x()},e.prototype.updateOutlineGap=function(){var t=this._label?this._label.nativeElement:null;if("outline"===this.appearance&&t&&t.children.length&&t.textContent.trim()&&(!this._platform||this._platform.isBrowser))if(document.documentElement.contains(this._elementRef.nativeElement)){var e=0,n=0,r=this._connectionContainerRef.nativeElement,i=r.querySelectorAll(".mat-form-field-outline-start"),o=r.querySelectorAll(".mat-form-field-outline-gap");if(this._label&&this._label.nativeElement.children.length){var a=r.getBoundingClientRect();if(0===a.width&&0===a.height)return this._outlineGapCalculationNeededOnStable=!0,void(this._outlineGapCalculationNeededImmediately=!1);for(var s=this._getStartEnd(a),u=this._getStartEnd(t.children[0].getBoundingClientRect()),c=0,l=0,d=t.children;l<d.length;l++){c+=d[l].offsetWidth}e=u-s-5,n=c>0?.75*c+10:0}for(var p=0;p<i.length;p++)i.item(p).style.width=e+"px";for(p=0;p<o.length;p++)o.item(p).style.width=n+"px";this._outlineGapCalculationNeededOnStable=this._outlineGapCalculationNeededImmediately=!1}else this._outlineGapCalculationNeededImmediately=!0},e.prototype._getStartEnd=function(t){return this._dir&&"rtl"===this._dir.value?t.right:t.left},e.prototype._updateOutlineGapOnStable=function(){var t=this;this._ngZone?this._outlineGapCalculationNeededOnStable=!0:Promise.resolve().then(function(){return t.updateOutlineGap()})},e.decorators=[{type:r.Component,args:[{selector:"mat-form-field",exportAs:"matFormField",template:'<div class="mat-form-field-wrapper"><div class="mat-form-field-flex" #connectionContainer (click)="_control.onContainerClick && _control.onContainerClick($event)"><ng-container *ngIf="appearance == \'outline\'"><div class="mat-form-field-outline"><div class="mat-form-field-outline-start"></div><div class="mat-form-field-outline-gap"></div><div class="mat-form-field-outline-end"></div></div><div class="mat-form-field-outline mat-form-field-outline-thick"><div class="mat-form-field-outline-start"></div><div class="mat-form-field-outline-gap"></div><div class="mat-form-field-outline-end"></div></div></ng-container><div class="mat-form-field-prefix" *ngIf="_prefixChildren.length"><ng-content select="[matPrefix]"></ng-content></div><div class="mat-form-field-infix" #inputContainer><ng-content></ng-content><span class="mat-form-field-label-wrapper"><label class="mat-form-field-label" (cdkObserveContent)="updateOutlineGap()" [cdkObserveContentDisabled]="appearance != \'outline\'" [id]="_labelId" [attr.for]="_control.id" [attr.aria-owns]="_control.id" [class.mat-empty]="_control.empty && !_shouldAlwaysFloat" [class.mat-form-field-empty]="_control.empty && !_shouldAlwaysFloat" [class.mat-accent]="color == \'accent\'" [class.mat-warn]="color == \'warn\'" #label *ngIf="_hasFloatingLabel()" [ngSwitch]="_hasLabel()"><ng-container *ngSwitchCase="false"><ng-content select="mat-placeholder"></ng-content>{{_control.placeholder}}</ng-container><ng-content select="mat-label" *ngSwitchCase="true"></ng-content><span class="mat-placeholder-required mat-form-field-required-marker" aria-hidden="true" *ngIf="!hideRequiredMarker && _control.required && !_control.disabled">&nbsp;*</span></label></span></div><div class="mat-form-field-suffix" *ngIf="_suffixChildren.length"><ng-content select="[matSuffix]"></ng-content></div></div><div class="mat-form-field-underline" #underline *ngIf="appearance != \'outline\'"><span class="mat-form-field-ripple" [class.mat-accent]="color == \'accent\'" [class.mat-warn]="color == \'warn\'"></span></div><div class="mat-form-field-subscript-wrapper" [ngSwitch]="_getDisplayedMessages()"><div *ngSwitchCase="\'error\'" [@transitionMessages]="_subscriptAnimationState"><ng-content select="mat-error"></ng-content></div><div class="mat-form-field-hint-wrapper" *ngSwitchCase="\'hint\'" [@transitionMessages]="_subscriptAnimationState"><div *ngIf="hintLabel" [id]="_hintLabelId" class="mat-hint">{{hintLabel}}</div><ng-content select="mat-hint:not([align=\'end\'])"></ng-content><div class="mat-form-field-hint-spacer"></div><ng-content select="mat-hint[align=\'end\']"></ng-content></div></div></div>',styles:[".mat-form-field{display:inline-block;position:relative;text-align:left}[dir=rtl] .mat-form-field{text-align:right}.mat-form-field-wrapper{position:relative}.mat-form-field-flex{display:inline-flex;align-items:baseline;box-sizing:border-box;width:100%}.mat-form-field-prefix,.mat-form-field-suffix{white-space:nowrap;flex:none;position:relative}.mat-form-field-infix{display:block;position:relative;flex:auto;min-width:0;width:180px}@media (-ms-high-contrast:active){.mat-form-field-infix{border-image:linear-gradient(transparent,transparent)}}.mat-form-field-label-wrapper{position:absolute;left:0;box-sizing:content-box;width:100%;height:100%;overflow:hidden;pointer-events:none}[dir=rtl] .mat-form-field-label-wrapper{left:auto;right:0}.mat-form-field-label{position:absolute;left:0;font:inherit;pointer-events:none;width:100%;white-space:nowrap;text-overflow:ellipsis;overflow:hidden;transform-origin:0 0;transition:transform .4s cubic-bezier(.25,.8,.25,1),color .4s cubic-bezier(.25,.8,.25,1),width .4s cubic-bezier(.25,.8,.25,1);display:none}[dir=rtl] .mat-form-field-label{transform-origin:100% 0;left:auto;right:0}.mat-form-field-can-float.mat-form-field-should-float .mat-form-field-label,.mat-form-field-empty.mat-form-field-label{display:block}.mat-form-field-autofill-control:-webkit-autofill+.mat-form-field-label-wrapper .mat-form-field-label{display:none}.mat-form-field-can-float .mat-form-field-autofill-control:-webkit-autofill+.mat-form-field-label-wrapper .mat-form-field-label{display:block;transition:none}.mat-input-server:focus+.mat-form-field-label-wrapper .mat-form-field-label,.mat-input-server[placeholder]:not(:placeholder-shown)+.mat-form-field-label-wrapper .mat-form-field-label{display:none}.mat-form-field-can-float .mat-input-server:focus+.mat-form-field-label-wrapper .mat-form-field-label,.mat-form-field-can-float .mat-input-server[placeholder]:not(:placeholder-shown)+.mat-form-field-label-wrapper .mat-form-field-label{display:block}.mat-form-field-label:not(.mat-form-field-empty){transition:none}.mat-form-field-underline{position:absolute;width:100%;pointer-events:none;transform:scaleY(1.0001)}.mat-form-field-ripple{position:absolute;left:0;width:100%;transform-origin:50%;transform:scaleX(.5);opacity:0;transition:background-color .3s cubic-bezier(.55,0,.55,.2)}.mat-form-field.mat-focused .mat-form-field-ripple,.mat-form-field.mat-form-field-invalid .mat-form-field-ripple{opacity:1;transform:scaleX(1);transition:transform .3s cubic-bezier(.25,.8,.25,1),opacity .1s cubic-bezier(.25,.8,.25,1),background-color .3s cubic-bezier(.25,.8,.25,1)}.mat-form-field-subscript-wrapper{position:absolute;box-sizing:border-box;width:100%;overflow:hidden}.mat-form-field-label-wrapper .mat-icon,.mat-form-field-subscript-wrapper .mat-icon{width:1em;height:1em;font-size:inherit;vertical-align:baseline}.mat-form-field-hint-wrapper{display:flex}.mat-form-field-hint-spacer{flex:1 0 1em}.mat-error{display:block}.mat-form-field-control-wrapper{position:relative}.mat-form-field._mat-animation-noopable .mat-form-field-label,.mat-form-field._mat-animation-noopable .mat-form-field-ripple{transition:none} .mat-form-field-appearance-fill .mat-form-field-flex{border-radius:4px 4px 0 0;padding:.75em .75em 0 .75em}@media (-ms-high-contrast:active){.mat-form-field-appearance-fill .mat-form-field-flex{outline:solid 1px}}.mat-form-field-appearance-fill .mat-form-field-underline::before{content:'';display:block;position:absolute;bottom:0;height:1px;width:100%}.mat-form-field-appearance-fill .mat-form-field-ripple{bottom:0;height:2px}@media (-ms-high-contrast:active){.mat-form-field-appearance-fill .mat-form-field-ripple{height:0;border-top:solid 2px}}.mat-form-field-appearance-fill:not(.mat-form-field-disabled) .mat-form-field-flex:hover~.mat-form-field-underline .mat-form-field-ripple{opacity:1;transform:none;transition:opacity .6s cubic-bezier(.25,.8,.25,1)}.mat-form-field-appearance-fill._mat-animation-noopable:not(.mat-form-field-disabled) .mat-form-field-flex:hover~.mat-form-field-underline .mat-form-field-ripple{transition:none}.mat-form-field-appearance-fill .mat-form-field-subscript-wrapper{padding:0 1em} .mat-input-element{font:inherit;background:0 0;color:currentColor;border:none;outline:0;padding:0;margin:0;width:100%;max-width:100%;vertical-align:bottom;text-align:inherit}.mat-input-element:-moz-ui-invalid{box-shadow:none}.mat-input-element::-ms-clear,.mat-input-element::-ms-reveal{display:none}.mat-input-element,.mat-input-element::-webkit-search-cancel-button,.mat-input-element::-webkit-search-decoration,.mat-input-element::-webkit-search-results-button,.mat-input-element::-webkit-search-results-decoration{-webkit-appearance:none}.mat-input-element::-webkit-caps-lock-indicator,.mat-input-element::-webkit-contacts-auto-fill-button,.mat-input-element::-webkit-credentials-auto-fill-button{visibility:hidden}.mat-input-element[type=date]::after,.mat-input-element[type=datetime-local]::after,.mat-input-element[type=datetime]::after,.mat-input-element[type=month]::after,.mat-input-element[type=time]::after,.mat-input-element[type=week]::after{content:' ';white-space:pre;width:1px}.mat-input-element::-webkit-calendar-picker-indicator,.mat-input-element::-webkit-clear-button,.mat-input-element::-webkit-inner-spin-button{font-size:.75em}.mat-input-element::placeholder{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;transition:color .4s .133s cubic-bezier(.25,.8,.25,1)}.mat-input-element::-moz-placeholder{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;transition:color .4s .133s cubic-bezier(.25,.8,.25,1)}.mat-input-element::-webkit-input-placeholder{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;transition:color .4s .133s cubic-bezier(.25,.8,.25,1)}.mat-input-element:-ms-input-placeholder{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;transition:color .4s .133s cubic-bezier(.25,.8,.25,1)}.mat-form-field-hide-placeholder .mat-input-element::placeholder{color:transparent!important;-webkit-text-fill-color:transparent;transition:none}.mat-form-field-hide-placeholder .mat-input-element::-moz-placeholder{color:transparent!important;-webkit-text-fill-color:transparent;transition:none}.mat-form-field-hide-placeholder .mat-input-element::-webkit-input-placeholder{color:transparent!important;-webkit-text-fill-color:transparent;transition:none}.mat-form-field-hide-placeholder .mat-input-element:-ms-input-placeholder{color:transparent!important;-webkit-text-fill-color:transparent;transition:none}textarea.mat-input-element{resize:vertical;overflow:auto}textarea.mat-input-element.cdk-textarea-autosize{resize:none}textarea.mat-input-element{padding:2px 0;margin:-2px 0}select.mat-input-element{-moz-appearance:none;-webkit-appearance:none;position:relative;background-color:transparent;display:inline-flex;box-sizing:border-box;padding-top:1em;top:-1em;margin-bottom:-1em}select.mat-input-element::-ms-expand{display:none}select.mat-input-element::-moz-focus-inner{border:0}select.mat-input-element:not(:disabled){cursor:pointer}select.mat-input-element::-ms-value{color:inherit;background:0 0}.mat-form-field-type-mat-native-select .mat-form-field-infix::after{content:'';width:0;height:0;border-left:5px solid transparent;border-right:5px solid transparent;border-top:5px solid;position:absolute;top:50%;right:0;margin-top:-2.5px}[dir=rtl] .mat-form-field-type-mat-native-select .mat-form-field-infix::after{right:auto;left:0}.mat-form-field-type-mat-native-select .mat-input-element{padding-right:15px}[dir=rtl] .mat-form-field-type-mat-native-select .mat-input-element{padding-right:0;padding-left:15px}.mat-form-field-type-mat-native-select .mat-form-field-label-wrapper{max-width:calc(100% - 10px)}.mat-form-field-type-mat-native-select.mat-form-field-appearance-outline .mat-form-field-infix::after{margin-top:-5px}.mat-form-field-type-mat-native-select.mat-form-field-appearance-fill .mat-form-field-infix::after{margin-top:-10px} .mat-form-field-appearance-legacy .mat-form-field-label{transform:perspective(100px);-ms-transform:none}.mat-form-field-appearance-legacy .mat-form-field-prefix .mat-icon,.mat-form-field-appearance-legacy .mat-form-field-suffix .mat-icon{width:1em}.mat-form-field-appearance-legacy .mat-form-field-prefix .mat-icon-button,.mat-form-field-appearance-legacy .mat-form-field-suffix .mat-icon-button{font:inherit;vertical-align:baseline}.mat-form-field-appearance-legacy .mat-form-field-prefix .mat-icon-button .mat-icon,.mat-form-field-appearance-legacy .mat-form-field-suffix .mat-icon-button .mat-icon{font-size:inherit}.mat-form-field-appearance-legacy .mat-form-field-underline{height:1px}@media (-ms-high-contrast:active){.mat-form-field-appearance-legacy .mat-form-field-underline{height:0;border-top:solid 1px}}.mat-form-field-appearance-legacy .mat-form-field-ripple{top:0;height:2px;overflow:hidden}@media (-ms-high-contrast:active){.mat-form-field-appearance-legacy .mat-form-field-ripple{height:0;border-top:solid 2px}}.mat-form-field-appearance-legacy.mat-form-field-disabled .mat-form-field-underline{background-position:0;background-color:transparent}@media (-ms-high-contrast:active){.mat-form-field-appearance-legacy.mat-form-field-disabled .mat-form-field-underline{border-top-style:dotted;border-top-width:2px}}.mat-form-field-appearance-legacy.mat-form-field-invalid:not(.mat-focused) .mat-form-field-ripple{height:1px} .mat-form-field-appearance-outline .mat-form-field-wrapper{margin:.25em 0}.mat-form-field-appearance-outline .mat-form-field-flex{padding:0 .75em 0 .75em;margin-top:-.25em;position:relative}.mat-form-field-appearance-outline .mat-form-field-prefix,.mat-form-field-appearance-outline .mat-form-field-suffix{top:.25em}.mat-form-field-appearance-outline .mat-form-field-outline{display:flex;position:absolute;top:.25em;left:0;right:0;bottom:0;pointer-events:none}.mat-form-field-appearance-outline .mat-form-field-outline-end,.mat-form-field-appearance-outline .mat-form-field-outline-start{border:1px solid currentColor;min-width:5px}.mat-form-field-appearance-outline .mat-form-field-outline-start{border-radius:5px 0 0 5px;border-right-style:none}[dir=rtl] .mat-form-field-appearance-outline .mat-form-field-outline-start{border-right-style:solid;border-left-style:none;border-radius:0 5px 5px 0}.mat-form-field-appearance-outline .mat-form-field-outline-end{border-radius:0 5px 5px 0;border-left-style:none;flex-grow:1}[dir=rtl] .mat-form-field-appearance-outline .mat-form-field-outline-end{border-left-style:solid;border-right-style:none;border-radius:5px 0 0 5px}.mat-form-field-appearance-outline .mat-form-field-outline-gap{border-radius:.000001px;border:1px solid currentColor;border-left-style:none;border-right-style:none}.mat-form-field-appearance-outline.mat-form-field-can-float.mat-form-field-should-float .mat-form-field-outline-gap{border-top-color:transparent}.mat-form-field-appearance-outline .mat-form-field-outline-thick{opacity:0}.mat-form-field-appearance-outline .mat-form-field-outline-thick .mat-form-field-outline-end,.mat-form-field-appearance-outline .mat-form-field-outline-thick .mat-form-field-outline-gap,.mat-form-field-appearance-outline .mat-form-field-outline-thick .mat-form-field-outline-start{border-width:2px;transition:border-color .3s cubic-bezier(.25,.8,.25,1)}.mat-form-field-appearance-outline.mat-focused .mat-form-field-outline,.mat-form-field-appearance-outline.mat-form-field-invalid .mat-form-field-outline{opacity:0;transition:opacity .1s cubic-bezier(.25,.8,.25,1)}.mat-form-field-appearance-outline.mat-focused .mat-form-field-outline-thick,.mat-form-field-appearance-outline.mat-form-field-invalid .mat-form-field-outline-thick{opacity:1}.mat-form-field-appearance-outline:not(.mat-form-field-disabled) .mat-form-field-flex:hover .mat-form-field-outline{opacity:0;transition:opacity .6s cubic-bezier(.25,.8,.25,1)}.mat-form-field-appearance-outline:not(.mat-form-field-disabled) .mat-form-field-flex:hover .mat-form-field-outline-thick{opacity:1}.mat-form-field-appearance-outline .mat-form-field-subscript-wrapper{padding:0 1em}.mat-form-field-appearance-outline._mat-animation-noopable .mat-form-field-outline,.mat-form-field-appearance-outline._mat-animation-noopable .mat-form-field-outline-end,.mat-form-field-appearance-outline._mat-animation-noopable .mat-form-field-outline-gap,.mat-form-field-appearance-outline._mat-animation-noopable .mat-form-field-outline-start,.mat-form-field-appearance-outline._mat-animation-noopable:not(.mat-form-field-disabled) .mat-form-field-flex:hover~.mat-form-field-outline{transition:none} .mat-form-field-appearance-standard .mat-form-field-flex{padding-top:.75em}.mat-form-field-appearance-standard .mat-form-field-underline{height:1px}@media (-ms-high-contrast:active){.mat-form-field-appearance-standard .mat-form-field-underline{height:0;border-top:solid 1px}}.mat-form-field-appearance-standard .mat-form-field-ripple{bottom:0;height:2px}@media (-ms-high-contrast:active){.mat-form-field-appearance-standard .mat-form-field-ripple{height:0;border-top:2px}}.mat-form-field-appearance-standard.mat-form-field-disabled .mat-form-field-underline{background-position:0;background-color:transparent}@media (-ms-high-contrast:active){.mat-form-field-appearance-standard.mat-form-field-disabled .mat-form-field-underline{border-top-style:dotted;border-top-width:2px}}.mat-form-field-appearance-standard:not(.mat-form-field-disabled) .mat-form-field-flex:hover~.mat-form-field-underline .mat-form-field-ripple{opacity:1;transform:none;transition:opacity .6s cubic-bezier(.25,.8,.25,1)}.mat-form-field-appearance-standard._mat-animation-noopable:not(.mat-form-field-disabled) .mat-form-field-flex:hover~.mat-form-field-underline .mat-form-field-ripple{transition:none}"],animations:[y.transitionMessages],host:{class:"mat-form-field","[class.mat-form-field-appearance-standard]":'appearance == "standard"',"[class.mat-form-field-appearance-fill]":'appearance == "fill"',"[class.mat-form-field-appearance-outline]":'appearance == "outline"',"[class.mat-form-field-appearance-legacy]":'appearance == "legacy"',"[class.mat-form-field-invalid]":"_control.errorState","[class.mat-form-field-can-float]":"_canLabelFloat","[class.mat-form-field-should-float]":"_shouldLabelFloat()","[class.mat-form-field-has-label]":"_hasFloatingLabel()","[class.mat-form-field-hide-placeholder]":"_hideControlPlaceholder()","[class.mat-form-field-disabled]":"_control.disabled","[class.mat-form-field-autofilled]":"_control.autofilled","[class.mat-focused]":"_control.focused","[class.mat-accent]":'color == "accent"',"[class.mat-warn]":'color == "warn"',"[class.ng-untouched]":'_shouldForward("untouched")',"[class.ng-touched]":'_shouldForward("touched")',"[class.ng-pristine]":'_shouldForward("pristine")',"[class.ng-dirty]":'_shouldForward("dirty")',"[class.ng-valid]":'_shouldForward("valid")',"[class.ng-invalid]":'_shouldForward("invalid")',"[class.ng-pending]":'_shouldForward("pending")',"[class._mat-animation-noopable]":"!_animationsEnabled"},inputs:["color"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:r.ChangeDetectorRef},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[u.MAT_LABEL_GLOBAL_OPTIONS]}]},{type:a.Directionality,decorators:[{type:r.Optional}]},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[A]}]},{type:d.Platform},{type:r.NgZone},{type:String,decorators:[{type:r.Optional},{type:r.Inject,args:[p.ANIMATION_MODULE_TYPE]}]}]},e.propDecorators={appearance:[{type:r.Input}],hideRequiredMarker:[{type:r.Input}],hintLabel:[{type:r.Input}],floatLabel:[{type:r.Input}],underlineRef:[{type:r.ViewChild,args:["underline"]}],_connectionContainerRef:[{type:r.ViewChild,args:["connectionContainer"]}],_inputContainerRef:[{type:r.ViewChild,args:["inputContainer"]}],_label:[{type:r.ViewChild,args:["label"]}],_control:[{type:r.ContentChild,args:[_]}],_placeholderChild:[{type:r.ContentChild,args:[C]}],_labelChild:[{type:r.ContentChild,args:[O]}],_errorChildren:[{type:r.ContentChildren,args:[g]}],_hintChildren:[{type:r.ContentChildren,args:[j]}],_prefixChildren:[{type:r.ContentChildren,args:[E]}],_suffixChildren:[{type:r.ContentChildren,args:[S]}]},e}(k),I=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{declarations:[g,D,j,O,C,E,S],imports:[f.CommonModule,h.ObserversModule],exports:[g,D,j,O,C,E,S]}]}],t}()},"./node_modules/@angular/material/esm5/grid-list.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatGridListModule",function(){return O}),n.d(e,"MatGridList",function(){return j}),n.d(e,"MatGridTile",function(){return c}),n.d(e,"MatGridTileText",function(){return l}),n.d(e,"MatGridAvatarCssMatStyler",function(){return d}),n.d(e,"MatGridTileHeaderCssMatStyler",function(){return p}),n.d(e,"MatGridTileFooterCssMatStyler",function(){return f}),n.d(e,"ɵa15",function(){return u});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/material/esm5/core.es5.js"),o=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),a=n("./node_modules/tslib/tslib.es6.js"),s=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),u=new r.InjectionToken("MAT_GRID_LIST"),c=function(){function t(t,e){this._element=t,this._gridList=e,this._rowspan=1,this._colspan=1}return Object.defineProperty(t.prototype,"rowspan",{get:function(){return this._rowspan},set:function(t){this._rowspan=Math.round(Object(o.coerceNumberProperty)(t))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"colspan",{get:function(){return this._colspan},set:function(t){this._colspan=Math.round(Object(o.coerceNumberProperty)(t))},enumerable:!0,configurable:!0}),t.prototype._setStyle=function(t,e){this._element.nativeElement.style[t]=e},t.decorators=[{type:r.Component,args:[{selector:"mat-grid-tile",exportAs:"matGridTile",host:{class:"mat-grid-tile"},template:'<figure class="mat-figure"><ng-content></ng-content></figure>',styles:[".mat-grid-list{display:block;position:relative}.mat-grid-tile{display:block;position:absolute;overflow:hidden}.mat-grid-tile .mat-figure{top:0;left:0;right:0;bottom:0;position:absolute;display:flex;align-items:center;justify-content:center;height:100%;padding:0;margin:0}.mat-grid-tile .mat-grid-tile-footer,.mat-grid-tile .mat-grid-tile-header{display:flex;align-items:center;height:48px;color:#fff;background:rgba(0,0,0,.38);overflow:hidden;padding:0 16px;position:absolute;left:0;right:0}.mat-grid-tile .mat-grid-tile-footer>*,.mat-grid-tile .mat-grid-tile-header>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-grid-tile .mat-grid-tile-footer.mat-2-line,.mat-grid-tile .mat-grid-tile-header.mat-2-line{height:68px}.mat-grid-tile .mat-grid-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden}.mat-grid-tile .mat-grid-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-grid-tile .mat-grid-list-text:empty{display:none}.mat-grid-tile .mat-grid-tile-header{top:0}.mat-grid-tile .mat-grid-tile-footer{bottom:0}.mat-grid-tile .mat-grid-avatar{padding-right:16px}[dir=rtl] .mat-grid-tile .mat-grid-avatar{padding-right:0;padding-left:16px}.mat-grid-tile .mat-grid-avatar:empty{display:none}"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[u]}]}]},t.propDecorators={rowspan:[{type:r.Input}],colspan:[{type:r.Input}]},t}(),l=function(){function t(t){this._element=t}return t.prototype.ngAfterContentInit=function(){Object(i.setLines)(this._lines,this._element)},t.decorators=[{type:r.Component,args:[{selector:"mat-grid-tile-header, mat-grid-tile-footer",template:'<ng-content select="[mat-grid-avatar], [matGridAvatar]"></ng-content><div class="mat-grid-list-text"><ng-content select="[mat-line], [matLine]"></ng-content></div><ng-content></ng-content>',changeDetection:r.ChangeDetectionStrategy.OnPush,encapsulation:r.ViewEncapsulation.None}]}],t.ctorParameters=function(){return[{type:r.ElementRef}]},t.propDecorators={_lines:[{type:r.ContentChildren,args:[i.MatLine]}]},t}(),d=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"[mat-grid-avatar], [matGridAvatar]",host:{class:"mat-grid-avatar"}}]}],t}(),p=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"mat-grid-tile-header",host:{class:"mat-grid-tile-header"}}]}],t}(),f=function(){function t(){}return t.decorators=[{type:r.Directive,args:[{selector:"mat-grid-tile-footer",host:{class:"mat-grid-tile-footer"}}]}],t}(),h=function(){function t(){this.columnIndex=0,this.rowIndex=0}return Object.defineProperty(t.prototype,"rowCount",{get:function(){return this.rowIndex+1},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"rowspan",{get:function(){var t=Math.max.apply(Math,this.tracker);return t>1?this.rowCount+t-1:this.rowCount},enumerable:!0,configurable:!0}),t.prototype.update=function(t,e){var n=this;this.columnIndex=0,this.rowIndex=0,this.tracker=new Array(t),this.tracker.fill(0,0,this.tracker.length),this.positions=e.map(function(t){return n._trackTile(t)})},t.prototype._trackTile=function(t){var e=this._findMatchingGap(t.colspan);return this._markTilePosition(e,t),this.columnIndex=e+t.colspan,new m(this.rowIndex,e)},t.prototype._findMatchingGap=function(t){if(t>this.tracker.length)throw Error("mat-grid-list: tile with colspan "+t+' is wider than grid with cols="'+this.tracker.length+'".');var e=-1,n=-1;do{this.columnIndex+t>this.tracker.length?(this._nextRow(),e=this.tracker.indexOf(0,this.columnIndex),n=this._findGapEndIndex(e)):-1!=(e=this.tracker.indexOf(0,this.columnIndex))?(n=this._findGapEndIndex(e),this.columnIndex=e+1):(this._nextRow(),e=this.tracker.indexOf(0,this.columnIndex),n=this._findGapEndIndex(e))}while(n-e<t||0==n);return Math.max(e,0)},t.prototype._nextRow=function(){this.columnIndex=0,this.rowIndex++;for(var t=0;t<this.tracker.length;t++)this.tracker[t]=Math.max(0,this.tracker[t]-1)},t.prototype._findGapEndIndex=function(t){for(var e=t+1;e<this.tracker.length;e++)if(0!=this.tracker[e])return e;return this.tracker.length},t.prototype._markTilePosition=function(t,e){for(var n=0;n<e.colspan;n++)this.tracker[t+n]=e.rowspan},t}(),m=function(){return function(t,e){this.row=t,this.col=e}}(),g=/^-?\d+((\.\d+)?[A-Za-z%$]?)+$/,y=function(){function t(){this._rows=0,this._rowspan=0}return t.prototype.init=function(t,e,n,r){this._gutterSize=w(t),this._rows=e.rowCount,this._rowspan=e.rowspan,this._cols=n,this._direction=r},t.prototype.getBaseTileSize=function(t,e){return"("+t+"% - ("+this._gutterSize+" * "+e+"))"},t.prototype.getTilePosition=function(t,e){return 0===e?"0":x("("+t+" + "+this._gutterSize+") * "+e)},t.prototype.getTileSize=function(t,e){return"("+t+" * "+e+") + ("+(e-1)+" * "+this._gutterSize+")"},t.prototype.setStyle=function(t,e,n){var r=100/this._cols,i=(this._cols-1)/this._cols;this.setColStyles(t,n,r,i),this.setRowStyles(t,e,r,i)},t.prototype.setColStyles=function(t,e,n,r){var i=this.getBaseTileSize(n,r),o="rtl"===this._direction?"right":"left";t._setStyle(o,this.getTilePosition(i,e)),t._setStyle("width",x(this.getTileSize(i,t.colspan)))},t.prototype.getGutterSpan=function(){return this._gutterSize+" * ("+this._rowspan+" - 1)"},t.prototype.getTileSpan=function(t){return this._rowspan+" * "+this.getTileSize(t,1)},t.prototype.getComputedHeight=function(){return null},t}(),_=function(t){function e(e){var n=t.call(this)||this;return n.fixedRowHeight=e,n}return Object(a.__extends)(e,t),e.prototype.init=function(e,n,r,i){if(t.prototype.init.call(this,e,n,r,i),this.fixedRowHeight=w(this.fixedRowHeight),!g.test(this.fixedRowHeight))throw Error('Invalid value "'+this.fixedRowHeight+'" set as rowHeight.')},e.prototype.setRowStyles=function(t,e){t._setStyle("top",this.getTilePosition(this.fixedRowHeight,e)),t._setStyle("height",x(this.getTileSize(this.fixedRowHeight,t.rowspan)))},e.prototype.getComputedHeight=function(){return["height",x(this.getTileSpan(this.fixedRowHeight)+" + "+this.getGutterSpan())]},e.prototype.reset=function(t){t._setListStyle(["height",null]),t._tiles&&t._tiles.forEach(function(t){t._setStyle("top",null),t._setStyle("height",null)})},e}(y),v=function(t){function e(e){var n=t.call(this)||this;return n._parseRatio(e),n}return Object(a.__extends)(e,t),e.prototype.setRowStyles=function(t,e,n,r){var i=n/this.rowHeightRatio;this.baseTileHeight=this.getBaseTileSize(i,r),t._setStyle("marginTop",this.getTilePosition(this.baseTileHeight,e)),t._setStyle("paddingTop",x(this.getTileSize(this.baseTileHeight,t.rowspan)))},e.prototype.getComputedHeight=function(){return["paddingBottom",x(this.getTileSpan(this.baseTileHeight)+" + "+this.getGutterSpan())]},e.prototype.reset=function(t){t._setListStyle(["paddingBottom",null]),t._tiles.forEach(function(t){t._setStyle("marginTop",null),t._setStyle("paddingTop",null)})},e.prototype._parseRatio=function(t){var e=t.split(":");if(2!==e.length)throw Error('mat-grid-list: invalid ratio given for row-height: "'+t+'"');this.rowHeightRatio=parseFloat(e[0])/parseFloat(e[1])},e}(y),b=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(a.__extends)(e,t),e.prototype.setRowStyles=function(t,e){var n=100/this._rowspan,r=(this._rows-1)/this._rows,i=this.getBaseTileSize(n,r);t._setStyle("top",this.getTilePosition(i,e)),t._setStyle("height",x(this.getTileSize(i,t.rowspan)))},e.prototype.reset=function(t){t._tiles&&t._tiles.forEach(function(t){t._setStyle("top",null),t._setStyle("height",null)})},e}(y);function x(t){return"calc("+t+")"}function w(t){return t.match(/([A-Za-z%]+)$/)?t:t+"px"}var j=function(){function t(t,e){this._element=t,this._dir=e,this._gutter="1px"}return Object.defineProperty(t.prototype,"cols",{get:function(){return this._cols},set:function(t){this._cols=Math.max(1,Math.round(Object(o.coerceNumberProperty)(t)))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"gutterSize",{get:function(){return this._gutter},set:function(t){this._gutter=""+(null==t?"":t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"rowHeight",{get:function(){return this._rowHeight},set:function(t){var e=""+(null==t?"":t);e!==this._rowHeight&&(this._rowHeight=e,this._setTileStyler(this._rowHeight))},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){this._checkCols(),this._checkRowHeight()},t.prototype.ngAfterContentChecked=function(){this._layoutTiles()},t.prototype._checkCols=function(){if(!this.cols)throw Error('mat-grid-list: must pass in number of columns. Example: <mat-grid-list cols="3">')},t.prototype._checkRowHeight=function(){this._rowHeight||this._setTileStyler("1:1")},t.prototype._setTileStyler=function(t){this._tileStyler&&this._tileStyler.reset(this),"fit"===t?this._tileStyler=new b:t&&t.indexOf(":")>-1?this._tileStyler=new v(t):this._tileStyler=new _(t)},t.prototype._layoutTiles=function(){var t=this;this._tileCoordinator||(this._tileCoordinator=new h);var e=this._tileCoordinator,n=this._tiles.filter(function(e){return!e._gridList||e._gridList===t}),r=this._dir?this._dir.value:"ltr";this._tileCoordinator.update(this.cols,n),this._tileStyler.init(this.gutterSize,e,this.cols,r),n.forEach(function(n,r){var i=e.positions[r];t._tileStyler.setStyle(n,i.row,i.col)}),this._setListStyle(this._tileStyler.getComputedHeight())},t.prototype._setListStyle=function(t){t&&(this._element.nativeElement.style[t[0]]=t[1])},t.decorators=[{type:r.Component,args:[{selector:"mat-grid-list",exportAs:"matGridList",template:"<div><ng-content></ng-content></div>",styles:[".mat-grid-list{display:block;position:relative}.mat-grid-tile{display:block;position:absolute;overflow:hidden}.mat-grid-tile .mat-figure{top:0;left:0;right:0;bottom:0;position:absolute;display:flex;align-items:center;justify-content:center;height:100%;padding:0;margin:0}.mat-grid-tile .mat-grid-tile-footer,.mat-grid-tile .mat-grid-tile-header{display:flex;align-items:center;height:48px;color:#fff;background:rgba(0,0,0,.38);overflow:hidden;padding:0 16px;position:absolute;left:0;right:0}.mat-grid-tile .mat-grid-tile-footer>*,.mat-grid-tile .mat-grid-tile-header>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-grid-tile .mat-grid-tile-footer.mat-2-line,.mat-grid-tile .mat-grid-tile-header.mat-2-line{height:68px}.mat-grid-tile .mat-grid-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden}.mat-grid-tile .mat-grid-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-grid-tile .mat-grid-list-text:empty{display:none}.mat-grid-tile .mat-grid-tile-header{top:0}.mat-grid-tile .mat-grid-tile-footer{bottom:0}.mat-grid-tile .mat-grid-avatar{padding-right:16px}[dir=rtl] .mat-grid-tile .mat-grid-avatar{padding-right:0;padding-left:16px}.mat-grid-tile .mat-grid-avatar:empty{display:none}"],host:{class:"mat-grid-list"},providers:[{provide:u,useExisting:t}],changeDetection:r.ChangeDetectionStrategy.OnPush,encapsulation:r.ViewEncapsulation.None}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:s.Directionality,decorators:[{type:r.Optional}]}]},t.propDecorators={_tiles:[{type:r.ContentChildren,args:[c,{descendants:!0}]}],cols:[{type:r.Input}],gutterSize:[{type:r.Input}],rowHeight:[{type:r.Input}]},t}(),O=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[i.MatLineModule,i.MatCommonModule],exports:[j,c,l,i.MatLineModule,i.MatCommonModule,p,f,d],declarations:[j,c,l,p,f,d]}]}],t}()},"./node_modules/@angular/material/esm5/icon.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatIconModule",function(){return k}),n.d(e,"MAT_ICON_LOCATION_FACTORY",function(){return C}),n.d(e,"MatIconBase",function(){return w}),n.d(e,"_MatIconMixinBase",function(){return j}),n.d(e,"MAT_ICON_LOCATION",function(){return O}),n.d(e,"MatIcon",function(){return T}),n.d(e,"getMatIconNameNotFoundError",function(){return p}),n.d(e,"getMatIconNoHttpProviderError",function(){return f}),n.d(e,"getMatIconFailedToSanitizeUrlError",function(){return h}),n.d(e,"getMatIconFailedToSanitizeLiteralError",function(){return m}),n.d(e,"ICON_REGISTRY_PROVIDER_FACTORY",function(){return _}),n.d(e,"MatIconRegistry",function(){return y}),n.d(e,"ICON_REGISTRY_PROVIDER",function(){return v});var r=n("./node_modules/@angular/common/fesm5/common.js"),i=n("./node_modules/@angular/common/fesm5/http.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),s=n("./node_modules/rxjs/_esm5/index.js"),u=n("./node_modules/rxjs/_esm5/operators/index.js"),c=n("./node_modules/tslib/tslib.es6.js"),l=n("./node_modules/@angular/material/esm5/core.es5.js"),d=n("./node_modules/@angular/cdk/esm5/coercion.es5.js");function p(t){return Error('Unable to find icon with the name "'+t+'"')}function f(){return Error("Could not find HttpClient provider for use with Angular Material icons. Please include the HttpClientModule from @angular/common/http in your app imports.")}function h(t){return Error("The URL provided to MatIconRegistry was not trusted as a resource URL via Angular's DomSanitizer. Attempted URL was \""+t+'".')}function m(t){return Error("The literal provided to MatIconRegistry was not trusted as safe HTML by Angular's DomSanitizer. Attempted literal was \""+t+'".')}var g=function(){return function(t){t.nodeName?this.svgElement=t:this.url=t}}(),y=function(){function t(t,e,n){this._httpClient=t,this._sanitizer=e,this._svgIconConfigs=new Map,this._iconSetConfigs=new Map,this._cachedIconsByUrl=new Map,this._inProgressUrlFetches=new Map,this._fontCssClassesByAlias=new Map,this._defaultFontSetClass="material-icons",this._document=n}return t.prototype.addSvgIcon=function(t,e){return this.addSvgIconInNamespace("",t,e)},t.prototype.addSvgIconLiteral=function(t,e){return this.addSvgIconLiteralInNamespace("",t,e)},t.prototype.addSvgIconInNamespace=function(t,e,n){return this._addSvgIconConfig(t,e,new g(n))},t.prototype.addSvgIconLiteralInNamespace=function(t,e,n){var r=this._sanitizer.sanitize(o.SecurityContext.HTML,n);if(!r)throw m(n);var i=this._createSvgElementForSingleIcon(r);return this._addSvgIconConfig(t,e,new g(i))},t.prototype.addSvgIconSet=function(t){return this.addSvgIconSetInNamespace("",t)},t.prototype.addSvgIconSetLiteral=function(t){return this.addSvgIconSetLiteralInNamespace("",t)},t.prototype.addSvgIconSetInNamespace=function(t,e){return this._addSvgIconSetConfig(t,new g(e))},t.prototype.addSvgIconSetLiteralInNamespace=function(t,e){var n=this._sanitizer.sanitize(o.SecurityContext.HTML,e);if(!n)throw m(e);var r=this._svgElementFromString(n);return this._addSvgIconSetConfig(t,new g(r))},t.prototype.registerFontClassAlias=function(t,e){return void 0===e&&(e=t),this._fontCssClassesByAlias.set(t,e),this},t.prototype.classNameForFontAlias=function(t){return this._fontCssClassesByAlias.get(t)||t},t.prototype.setDefaultFontSetClass=function(t){return this._defaultFontSetClass=t,this},t.prototype.getDefaultFontSetClass=function(){return this._defaultFontSetClass},t.prototype.getSvgIconFromUrl=function(t){var e=this,n=this._sanitizer.sanitize(o.SecurityContext.RESOURCE_URL,t);if(!n)throw h(t);var r=this._cachedIconsByUrl.get(n);return r?Object(s.of)(b(r)):this._loadSvgIconFromConfig(new g(t)).pipe(Object(u.tap)(function(t){return e._cachedIconsByUrl.set(n,t)}),Object(u.map)(function(t){return b(t)}))},t.prototype.getNamedSvgIcon=function(t,e){void 0===e&&(e="");var n=x(e,t),r=this._svgIconConfigs.get(n);if(r)return this._getSvgFromConfig(r);var i=this._iconSetConfigs.get(e);return i?this._getSvgFromIconSetConfigs(t,i):Object(s.throwError)(p(n))},t.prototype._getSvgFromConfig=function(t){return t.svgElement?Object(s.of)(b(t.svgElement)):this._loadSvgIconFromConfig(t).pipe(Object(u.tap)(function(e){return t.svgElement=e}),Object(u.map)(function(t){return b(t)}))},t.prototype._getSvgFromIconSetConfigs=function(t,e){var n=this,r=this._extractIconWithNameFromAnySet(t,e);if(r)return Object(s.of)(r);var i=e.filter(function(t){return!t.svgElement}).map(function(t){return n._loadSvgIconSetFromConfig(t).pipe(Object(u.catchError)(function(e){var r=n._sanitizer.sanitize(o.SecurityContext.RESOURCE_URL,t.url);return console.error("Loading icon set URL: "+r+" failed: "+e.message),Object(s.of)(null)}))});return Object(s.forkJoin)(i).pipe(Object(u.map)(function(){var r=n._extractIconWithNameFromAnySet(t,e);if(!r)throw p(t);return r}))},t.prototype._extractIconWithNameFromAnySet=function(t,e){for(var n=e.length-1;n>=0;n--){var r=e[n];if(r.svgElement){var i=this._extractSvgIconFromSet(r.svgElement,t);if(i)return i}}return null},t.prototype._loadSvgIconFromConfig=function(t){var e=this;return this._fetchUrl(t.url).pipe(Object(u.map)(function(t){return e._createSvgElementForSingleIcon(t)}))},t.prototype._loadSvgIconSetFromConfig=function(t){var e=this;return t.svgElement?Object(s.of)(t.svgElement):this._fetchUrl(t.url).pipe(Object(u.map)(function(n){return t.svgElement||(t.svgElement=e._svgElementFromString(n)),t.svgElement}))},t.prototype._createSvgElementForSingleIcon=function(t){var e=this._svgElementFromString(t);return this._setSvgAttributes(e),e},t.prototype._extractSvgIconFromSet=function(t,e){var n=t.querySelector("#"+e);if(!n)return null;var r=n.cloneNode(!0);if(r.removeAttribute("id"),"svg"===r.nodeName.toLowerCase())return this._setSvgAttributes(r);if("symbol"===r.nodeName.toLowerCase())return this._setSvgAttributes(this._toSvgElement(r));var i=this._svgElementFromString("<svg></svg>");return i.appendChild(r),this._setSvgAttributes(i)},t.prototype._svgElementFromString=function(t){var e=this._document.createElement("DIV");e.innerHTML=t;var n=e.querySelector("svg");if(!n)throw Error("<svg> tag not found");return n},t.prototype._toSvgElement=function(t){for(var e=this._svgElementFromString("<svg></svg>"),n=0;n<t.childNodes.length;n++)t.childNodes[n].nodeType===this._document.ELEMENT_NODE&&e.appendChild(t.childNodes[n].cloneNode(!0));return e},t.prototype._setSvgAttributes=function(t){return t.setAttribute("fit",""),t.setAttribute("height","100%"),t.setAttribute("width","100%"),t.setAttribute("preserveAspectRatio","xMidYMid meet"),t.setAttribute("focusable","false"),t},t.prototype._fetchUrl=function(t){var e=this;if(!this._httpClient)throw f();if(null==t)throw Error('Cannot fetch icon from URL "'+t+'".');var n=this._sanitizer.sanitize(o.SecurityContext.RESOURCE_URL,t);if(!n)throw h(t);var r=this._inProgressUrlFetches.get(n);if(r)return r;var i=this._httpClient.get(n,{responseType:"text"}).pipe(Object(u.finalize)(function(){return e._inProgressUrlFetches.delete(n)}),Object(u.share)());return this._inProgressUrlFetches.set(n,i),i},t.prototype._addSvgIconConfig=function(t,e,n){return this._svgIconConfigs.set(x(t,e),n),this},t.prototype._addSvgIconSetConfig=function(t,e){var n=this._iconSetConfigs.get(t);return n?n.push(e):this._iconSetConfigs.set(t,[e]),this},t.decorators=[{type:o.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:i.HttpClient,decorators:[{type:o.Optional}]},{type:a.DomSanitizer},{type:void 0,decorators:[{type:o.Optional},{type:o.Inject,args:[r.DOCUMENT]}]}]},t.ngInjectableDef=Object(o.defineInjectable)({factory:function(){return new t(Object(o.inject)(i.HttpClient,8),Object(o.inject)(a.DomSanitizer),Object(o.inject)(r.DOCUMENT,8))},token:t,providedIn:"root"}),t}();function _(t,e,n,r){return t||new y(e,n,r)}var v={provide:y,deps:[[new o.Optional,new o.SkipSelf,y],[new o.Optional,i.HttpClient],a.DomSanitizer,[new o.Optional,r.DOCUMENT]],useFactory:_};function b(t){return t.cloneNode(!0)}function x(t,e){return t+":"+e}var w=function(){return function(t){this._elementRef=t}}(),j=Object(l.mixinColor)(w),O=new o.InjectionToken("mat-icon-location",{providedIn:"root",factory:C});function C(){var t=Object(o.inject)(r.DOCUMENT),e=t?t.location:null;return{getPathname:function(){return e?e.pathname+e.search:""}}}var E=["clip-path","color-profile","src","cursor","fill","filter","marker","marker-start","marker-mid","marker-end","mask","stroke"],S=E.map(function(t){return"["+t+"]"}).join(", "),M=/^url\(['"]?#(.*?)['"]?\)$/,T=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o._iconRegistry=n,o._location=i,o._inline=!1,r||e.nativeElement.setAttribute("aria-hidden","true"),o}return Object(c.__extends)(e,t),Object.defineProperty(e.prototype,"inline",{get:function(){return this._inline},set:function(t){this._inline=Object(d.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"fontSet",{get:function(){return this._fontSet},set:function(t){this._fontSet=this._cleanupFontValue(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"fontIcon",{get:function(){return this._fontIcon},set:function(t){this._fontIcon=this._cleanupFontValue(t)},enumerable:!0,configurable:!0}),e.prototype._splitIconName=function(t){if(!t)return["",""];var e=t.split(":");switch(e.length){case 1:return["",e[0]];case 2:return e;default:throw Error('Invalid icon name: "'+t+'"')}},e.prototype.ngOnChanges=function(t){var e=this;if(t.svgIcon)if(this.svgIcon){var n=this._splitIconName(this.svgIcon),r=n[0],i=n[1];this._iconRegistry.getNamedSvgIcon(i,r).pipe(Object(u.take)(1)).subscribe(function(t){return e._setSvgElement(t)},function(t){return console.log("Error retrieving icon: "+t.message)})}else this._clearSvgElement();this._usingFontIcon()&&this._updateFontIconClasses()},e.prototype.ngOnInit=function(){this._usingFontIcon()&&this._updateFontIconClasses()},e.prototype.ngAfterViewChecked=function(){var t=this._elementsWithExternalReferences;if(t&&this._location&&t.size){var e=this._location.getPathname();e!==this._previousPath&&(this._previousPath=e,this._prependPathToReferences(e))}},e.prototype.ngOnDestroy=function(){this._elementsWithExternalReferences&&this._elementsWithExternalReferences.clear()},e.prototype._usingFontIcon=function(){return!this.svgIcon},e.prototype._setSvgElement=function(t){this._clearSvgElement();for(var e=t.querySelectorAll("style"),n=0;n<e.length;n++)e[n].textContent+=" ";if(this._location){var r=this._location.getPathname();this._previousPath=r,this._cacheChildrenWithExternalReferences(t),this._prependPathToReferences(r)}this._elementRef.nativeElement.appendChild(t)},e.prototype._clearSvgElement=function(){var t=this._elementRef.nativeElement,e=t.childNodes.length;for(this._elementsWithExternalReferences&&this._elementsWithExternalReferences.clear();e--;){var n=t.childNodes[e];1===n.nodeType&&"svg"!==n.nodeName.toLowerCase()||t.removeChild(n)}},e.prototype._updateFontIconClasses=function(){if(this._usingFontIcon()){var t=this._elementRef.nativeElement,e=this.fontSet?this._iconRegistry.classNameForFontAlias(this.fontSet):this._iconRegistry.getDefaultFontSetClass();e!=this._previousFontSetClass&&(this._previousFontSetClass&&t.classList.remove(this._previousFontSetClass),e&&t.classList.add(e),this._previousFontSetClass=e),this.fontIcon!=this._previousFontIconClass&&(this._previousFontIconClass&&t.classList.remove(this._previousFontIconClass),this.fontIcon&&t.classList.add(this.fontIcon),this._previousFontIconClass=this.fontIcon)}},e.prototype._cleanupFontValue=function(t){return"string"==typeof t?t.trim().split(" ")[0]:t},e.prototype._prependPathToReferences=function(t){var e=this._elementsWithExternalReferences;e&&e.forEach(function(e,n){e.forEach(function(e){n.setAttribute(e.name,"url('"+t+"#"+e.value+"')")})})},e.prototype._cacheChildrenWithExternalReferences=function(t){for(var e=t.querySelectorAll(S),n=this._elementsWithExternalReferences=this._elementsWithExternalReferences||new Map,r=function(t){E.forEach(function(r){var i=e[t],o=i.getAttribute(r),a=o?o.match(M):null;if(a){var s=n.get(i);s||(s=[],n.set(i,s)),s.push({name:r,value:a[1]})}})},i=0;i<e.length;i++)r(i)},e.decorators=[{type:o.Component,args:[{template:"<ng-content></ng-content>",selector:"mat-icon",exportAs:"matIcon",styles:[".mat-icon{background-repeat:no-repeat;display:inline-block;fill:currentColor;height:24px;width:24px}.mat-icon.mat-icon-inline{font-size:inherit;height:inherit;line-height:inherit;width:inherit}[dir=rtl] .mat-icon-rtl-mirror{transform:scale(-1,1)}.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-prefix .mat-icon,.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-suffix .mat-icon{display:block}.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-prefix .mat-icon-button .mat-icon,.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-suffix .mat-icon-button .mat-icon{margin:auto}"],inputs:["color"],host:{role:"img",class:"mat-icon","[class.mat-icon-inline]":"inline","[class.mat-icon-no-color]":'color !== "primary" && color !== "accent" && color !== "warn"'},encapsulation:o.ViewEncapsulation.None,changeDetection:o.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:o.ElementRef},{type:y},{type:String,decorators:[{type:o.Attribute,args:["aria-hidden"]}]},{type:void 0,decorators:[{type:o.Optional},{type:o.Inject,args:[O]}]}]},e.propDecorators={inline:[{type:o.Input}],svgIcon:[{type:o.Input}],fontSet:[{type:o.Input}],fontIcon:[{type:o.Input}]},e}(j),k=function(){function t(){}return t.decorators=[{type:o.NgModule,args:[{imports:[l.MatCommonModule],exports:[T,l.MatCommonModule],declarations:[T]}]}],t}()},"./node_modules/@angular/material/esm5/input.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatTextareaAutosize",function(){return f}),n.d(e,"MatInputBase",function(){return _}),n.d(e,"_MatInputMixinBase",function(){return v}),n.d(e,"MatInput",function(){return b}),n.d(e,"getMatInputUnsupportedTypeError",function(){return h}),n.d(e,"MatInputModule",function(){return x}),n.d(e,"MAT_INPUT_VALUE_ACCESSOR",function(){return m});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/text-field.es5.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),s=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),u=n("./node_modules/@angular/forms/fesm5/forms.js"),c=n("./node_modules/@angular/material/esm5/core.es5.js"),l=n("./node_modules/@angular/material/esm5/form-field.es5.js"),d=n("./node_modules/rxjs/_esm5/index.js"),p=n("./node_modules/@angular/common/fesm5/common.js"),f=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"matAutosizeMinRows",{get:function(){return this.minRows},set:function(t){this.minRows=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"matAutosizeMaxRows",{get:function(){return this.maxRows},set:function(t){this.maxRows=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"matAutosize",{get:function(){return this.enabled},set:function(t){this.enabled=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"matTextareaAutosize",{get:function(){return this.enabled},set:function(t){this.enabled=t},enumerable:!0,configurable:!0}),e.decorators=[{type:o.Directive,args:[{selector:"textarea[mat-autosize], textarea[matTextareaAutosize]",exportAs:"matTextareaAutosize",inputs:["cdkAutosizeMinRows","cdkAutosizeMaxRows"],host:{class:"cdk-textarea-autosize mat-autosize",rows:"1","(input)":"_noopInputHandler()"}}]}],e.propDecorators={matAutosizeMinRows:[{type:o.Input}],matAutosizeMaxRows:[{type:o.Input}],matAutosize:[{type:o.Input,args:["mat-autosize"]}],matTextareaAutosize:[{type:o.Input}]},e}(i.CdkTextareaAutosize);function h(t){return Error('Input type "'+t+"\" isn't supported by matInput.")}var m=new o.InjectionToken("MAT_INPUT_VALUE_ACCESSOR"),g=["button","checkbox","file","hidden","image","radio","range","reset","submit"],y=0,_=function(){return function(t,e,n,r){this._defaultErrorStateMatcher=t,this._parentForm=e,this._parentFormGroup=n,this.ngControl=r}}(),v=Object(c.mixinErrorState)(_),b=function(t){function e(e,n,r,i,o,a,u,c,l){var p=t.call(this,a,i,o,r)||this;p._elementRef=e,p._platform=n,p.ngControl=r,p._autofillMonitor=c,p._uid="mat-input-"+y++,p._isServer=!1,p._isNativeSelect=!1,p.focused=!1,p.stateChanges=new d.Subject,p.controlType="mat-input",p.autofilled=!1,p._disabled=!1,p._required=!1,p._type="text",p._readonly=!1,p._neverEmptyInputTypes=["date","datetime","datetime-local","month","time","week"].filter(function(t){return Object(s.getSupportedInputTypes)().has(t)});var f=p._elementRef.nativeElement;return p._inputValueAccessor=u||f,p._previousNativeValue=p.value,p.id=p.id,n.IOS&&l.runOutsideAngular(function(){e.nativeElement.addEventListener("keyup",function(t){var e=t.target;e.value||e.selectionStart||e.selectionEnd||(e.setSelectionRange(1,1),e.setSelectionRange(0,0))})}),p._isServer=!p._platform.isBrowser,p._isNativeSelect="select"===f.nodeName.toLowerCase(),p._isNativeSelect&&(p.controlType=f.multiple?"mat-native-select-multiple":"mat-native-select"),p}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"disabled",{get:function(){return this.ngControl&&null!==this.ngControl.disabled?this.ngControl.disabled:this._disabled},set:function(t){this._disabled=Object(a.coerceBooleanProperty)(t),this.focused&&(this.focused=!1,this.stateChanges.next())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"id",{get:function(){return this._id},set:function(t){this._id=t||this._uid},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"required",{get:function(){return this._required},set:function(t){this._required=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"type",{get:function(){return this._type},set:function(t){this._type=t||"text",this._validateType(),!this._isTextarea()&&Object(s.getSupportedInputTypes)().has(this._type)&&(this._elementRef.nativeElement.type=this._type)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return this._inputValueAccessor.value},set:function(t){t!==this.value&&(this._inputValueAccessor.value=t,this.stateChanges.next())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"readonly",{get:function(){return this._readonly},set:function(t){this._readonly=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){var t=this;this._platform.isBrowser&&this._autofillMonitor.monitor(this._elementRef.nativeElement).subscribe(function(e){t.autofilled=e.isAutofilled,t.stateChanges.next()})},e.prototype.ngOnChanges=function(){this.stateChanges.next()},e.prototype.ngOnDestroy=function(){this.stateChanges.complete(),this._platform.isBrowser&&this._autofillMonitor.stopMonitoring(this._elementRef.nativeElement)},e.prototype.ngDoCheck=function(){this.ngControl&&this.updateErrorState(),this._dirtyCheckNativeValue()},e.prototype.focus=function(){this._elementRef.nativeElement.focus()},e.prototype._focusChanged=function(t){t===this.focused||this.readonly&&t||(this.focused=t,this.stateChanges.next())},e.prototype._onInput=function(){},e.prototype._dirtyCheckNativeValue=function(){var t=this._elementRef.nativeElement.value;this._previousNativeValue!==t&&(this._previousNativeValue=t,this.stateChanges.next())},e.prototype._validateType=function(){if(g.indexOf(this._type)>-1)throw h(this._type)},e.prototype._isNeverEmpty=function(){return this._neverEmptyInputTypes.indexOf(this._type)>-1},e.prototype._isBadInput=function(){var t=this._elementRef.nativeElement.validity;return t&&t.badInput},e.prototype._isTextarea=function(){return"textarea"===this._elementRef.nativeElement.nodeName.toLowerCase()},Object.defineProperty(e.prototype,"empty",{get:function(){return!(this._isNeverEmpty()||this._elementRef.nativeElement.value||this._isBadInput()||this.autofilled)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"shouldLabelFloat",{get:function(){if(this._isNativeSelect){var t=this._elementRef.nativeElement,e=t.options[0];return this.focused||t.multiple||!this.empty||!!(t.selectedIndex>-1&&e&&e.label)}return this.focused||!this.empty},enumerable:!0,configurable:!0}),e.prototype.setDescribedByIds=function(t){this._ariaDescribedby=t.join(" ")},e.prototype.onContainerClick=function(){this.focused||this.focus()},e.decorators=[{type:o.Directive,args:[{selector:"input[matInput], textarea[matInput], select[matNativeControl],\n      input[matNativeControl], textarea[matNativeControl]",exportAs:"matInput",host:{class:"mat-input-element mat-form-field-autofill-control","[class.mat-input-server]":"_isServer","[attr.id]":"id","[attr.placeholder]":"placeholder","[disabled]":"disabled","[required]":"required","[attr.readonly]":"readonly && !_isNativeSelect || null","[attr.aria-describedby]":"_ariaDescribedby || null","[attr.aria-invalid]":"errorState","[attr.aria-required]":"required.toString()","(blur)":"_focusChanged(false)","(focus)":"_focusChanged(true)","(input)":"_onInput()"},providers:[{provide:l.MatFormFieldControl,useExisting:e}]}]}],e.ctorParameters=function(){return[{type:o.ElementRef},{type:s.Platform},{type:u.NgControl,decorators:[{type:o.Optional},{type:o.Self}]},{type:u.NgForm,decorators:[{type:o.Optional}]},{type:u.FormGroupDirective,decorators:[{type:o.Optional}]},{type:c.ErrorStateMatcher},{type:void 0,decorators:[{type:o.Optional},{type:o.Self},{type:o.Inject,args:[m]}]},{type:i.AutofillMonitor},{type:o.NgZone}]},e.propDecorators={disabled:[{type:o.Input}],id:[{type:o.Input}],placeholder:[{type:o.Input}],required:[{type:o.Input}],type:[{type:o.Input}],errorStateMatcher:[{type:o.Input}],value:[{type:o.Input}],readonly:[{type:o.Input}]},e}(v),x=function(){function t(){}return t.decorators=[{type:o.NgModule,args:[{declarations:[b,f],imports:[p.CommonModule,i.TextFieldModule,l.MatFormFieldModule],exports:[i.TextFieldModule,l.MatFormFieldModule,b,f],providers:[c.ErrorStateMatcher]}]}],t}()},"./node_modules/@angular/material/esm5/list.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatListModule",function(){return D}),n.d(e,"MatListBase",function(){return h}),n.d(e,"_MatListMixinBase",function(){return m}),n.d(e,"MatListItemBase",function(){return g}),n.d(e,"_MatListItemMixinBase",function(){return y}),n.d(e,"MatNavList",function(){return _}),n.d(e,"MatList",function(){return v}),n.d(e,"MatListAvatarCssMatStyler",function(){return b}),n.d(e,"MatListIconCssMatStyler",function(){return x}),n.d(e,"MatListSubheaderCssMatStyler",function(){return w}),n.d(e,"MatListItem",function(){return j}),n.d(e,"MatSelectionListBase",function(){return O}),n.d(e,"_MatSelectionListMixinBase",function(){return C}),n.d(e,"MatListOptionBase",function(){return E}),n.d(e,"_MatListOptionMixinBase",function(){return S}),n.d(e,"MAT_SELECTION_LIST_VALUE_ACCESSOR",function(){return M}),n.d(e,"MatSelectionListChange",function(){return T}),n.d(e,"MatListOption",function(){return k}),n.d(e,"MatSelectionList",function(){return A});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/material/esm5/core.es5.js"),a=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),c=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),l=n("./node_modules/@angular/forms/fesm5/forms.js"),d=n("./node_modules/rxjs/_esm5/index.js"),p=n("./node_modules/@angular/common/fesm5/common.js"),f=n("./node_modules/@angular/material/esm5/divider.es5.js"),h=function(){return function(){}}(),m=Object(o.mixinDisableRipple)(h),g=function(){return function(){}}(),y=Object(o.mixinDisableRipple)(g),_=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.decorators=[{type:i.Component,args:[{selector:"mat-nav-list",exportAs:"matNavList",host:{role:"navigation",class:"mat-nav-list mat-list-base"},template:"<ng-content></ng-content>",styles:[".mat-subheader{display:flex;box-sizing:border-box;padding:16px;align-items:center}.mat-list-base .mat-subheader{margin:0}.mat-list-base{padding-top:8px;display:block;-webkit-tap-highlight-color:transparent}.mat-list-base .mat-subheader{height:48px;line-height:16px}.mat-list-base .mat-subheader:first-child{margin-top:-8px}.mat-list-base .mat-list-item,.mat-list-base .mat-list-option{display:block;height:48px;-webkit-tap-highlight-color:transparent;width:100%;padding:0}.mat-list-base .mat-list-item .mat-list-item-content,.mat-list-base .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list-base .mat-list-item .mat-list-item-content-reverse,.mat-list-base .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list-base .mat-list-item .mat-list-item-ripple,.mat-list-base .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list-base .mat-list-item.mat-list-item-with-avatar,.mat-list-base .mat-list-option.mat-list-item-with-avatar{height:56px}.mat-list-base .mat-list-item.mat-2-line,.mat-list-base .mat-list-option.mat-2-line{height:72px}.mat-list-base .mat-list-item.mat-3-line,.mat-list-base .mat-list-option.mat-3-line{height:88px}.mat-list-base .mat-list-item.mat-multi-line,.mat-list-base .mat-list-option.mat-multi-line{height:auto}.mat-list-base .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list-base .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list-base .mat-list-item .mat-list-text,.mat-list-base .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0}.mat-list-base .mat-list-item .mat-list-text>*,.mat-list-base .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list-base .mat-list-item .mat-list-text:empty,.mat-list-base .mat-list-option .mat-list-text:empty{display:none}.mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:0;padding-left:16px}[dir=rtl] .mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:16px;padding-left:0}.mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-left:0;padding-right:16px}[dir=rtl] .mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:0;padding-left:16px}.mat-list-base .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:16px;padding-left:16px}.mat-list-base .mat-list-item .mat-list-avatar,.mat-list-base .mat-list-option .mat-list-avatar{flex-shrink:0;width:40px;height:40px;border-radius:50%;object-fit:cover}.mat-list-base .mat-list-item .mat-list-avatar~.mat-divider-inset,.mat-list-base .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:72px;width:calc(100% - 72px)}[dir=rtl] .mat-list-base .mat-list-item .mat-list-avatar~.mat-divider-inset,[dir=rtl] .mat-list-base .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:auto;margin-right:72px}.mat-list-base .mat-list-item .mat-list-icon,.mat-list-base .mat-list-option .mat-list-icon{flex-shrink:0;width:24px;height:24px;font-size:24px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-list-base .mat-list-item .mat-list-icon~.mat-divider-inset,.mat-list-base .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:64px;width:calc(100% - 64px)}[dir=rtl] .mat-list-base .mat-list-item .mat-list-icon~.mat-divider-inset,[dir=rtl] .mat-list-base .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:auto;margin-right:64px}.mat-list-base .mat-list-item .mat-divider,.mat-list-base .mat-list-option .mat-divider{position:absolute;bottom:0;left:0;width:100%;margin:0}[dir=rtl] .mat-list-base .mat-list-item .mat-divider,[dir=rtl] .mat-list-base .mat-list-option .mat-divider{margin-left:auto;margin-right:0}.mat-list-base .mat-list-item .mat-divider.mat-divider-inset,.mat-list-base .mat-list-option .mat-divider.mat-divider-inset{position:absolute}.mat-list-base[dense]{padding-top:4px;display:block}.mat-list-base[dense] .mat-subheader{height:40px;line-height:8px}.mat-list-base[dense] .mat-subheader:first-child{margin-top:-4px}.mat-list-base[dense] .mat-list-item,.mat-list-base[dense] .mat-list-option{display:block;height:40px;-webkit-tap-highlight-color:transparent;width:100%;padding:0}.mat-list-base[dense] .mat-list-item .mat-list-item-content,.mat-list-base[dense] .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list-base[dense] .mat-list-item .mat-list-item-content-reverse,.mat-list-base[dense] .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list-base[dense] .mat-list-item .mat-list-item-ripple,.mat-list-base[dense] .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar{height:48px}.mat-list-base[dense] .mat-list-item.mat-2-line,.mat-list-base[dense] .mat-list-option.mat-2-line{height:60px}.mat-list-base[dense] .mat-list-item.mat-3-line,.mat-list-base[dense] .mat-list-option.mat-3-line{height:76px}.mat-list-base[dense] .mat-list-item.mat-multi-line,.mat-list-base[dense] .mat-list-option.mat-multi-line{height:auto}.mat-list-base[dense] .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list-base[dense] .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list-base[dense] .mat-list-item .mat-list-text,.mat-list-base[dense] .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0}.mat-list-base[dense] .mat-list-item .mat-list-text>*,.mat-list-base[dense] .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list-base[dense] .mat-list-item .mat-list-text:empty,.mat-list-base[dense] .mat-list-option .mat-list-text:empty{display:none}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:0;padding-left:16px}[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:16px;padding-left:0}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-left:0;padding-right:16px}[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:0;padding-left:16px}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:16px;padding-left:16px}.mat-list-base[dense] .mat-list-item .mat-list-avatar,.mat-list-base[dense] .mat-list-option .mat-list-avatar{flex-shrink:0;width:36px;height:36px;border-radius:50%;object-fit:cover}.mat-list-base[dense] .mat-list-item .mat-list-avatar~.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:68px;width:calc(100% - 68px)}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-list-avatar~.mat-divider-inset,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:auto;margin-right:68px}.mat-list-base[dense] .mat-list-item .mat-list-icon,.mat-list-base[dense] .mat-list-option .mat-list-icon{flex-shrink:0;width:20px;height:20px;font-size:20px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-list-base[dense] .mat-list-item .mat-list-icon~.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:60px;width:calc(100% - 60px)}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-list-icon~.mat-divider-inset,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:auto;margin-right:60px}.mat-list-base[dense] .mat-list-item .mat-divider,.mat-list-base[dense] .mat-list-option .mat-divider{position:absolute;bottom:0;left:0;width:100%;margin:0}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-divider,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-divider{margin-left:auto;margin-right:0}.mat-list-base[dense] .mat-list-item .mat-divider.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-divider.mat-divider-inset{position:absolute}.mat-nav-list a{text-decoration:none;color:inherit}.mat-nav-list .mat-list-item{cursor:pointer;outline:0}mat-action-list button{background:0 0;color:inherit;border:none;font:inherit;outline:inherit}mat-action-list .mat-list-item{cursor:pointer;outline:inherit}.mat-list-option:not(.mat-list-item-disabled){cursor:pointer;outline:0}@media (-ms-high-contrast:active){.mat-selection-list:focus{outline-style:dotted}.mat-list-option:focus,.mat-list-option:hover,.mat-nav-list .mat-list-item:focus,.mat-nav-list .mat-list-item:hover,mat-action-list .mat-list-item:focus,mat-action-list .mat-list-item:hover{outline:dotted 1px}}@media (hover:none){.mat-list-option:not(.mat-list-item-disabled):hover,.mat-nav-list .mat-list-item:not(.mat-list-item-disabled):hover{background:0 0}}"],inputs:["disableRipple"],encapsulation:i.ViewEncapsulation.None,changeDetection:i.ChangeDetectionStrategy.OnPush}]}],e}(m),v=function(t){function e(e){var n=t.call(this)||this;return n._elementRef=e,n}return Object(r.__extends)(e,t),e.prototype._getListType=function(){var t=this._elementRef;if(t){var e=t.nativeElement.nodeName.toLowerCase();if("mat-list"===e)return"list";if("mat-action-list"===e)return"action-list"}return null},e.decorators=[{type:i.Component,args:[{selector:"mat-list, mat-action-list",exportAs:"matList",template:"<ng-content></ng-content>",host:{class:"mat-list mat-list-base"},styles:[".mat-subheader{display:flex;box-sizing:border-box;padding:16px;align-items:center}.mat-list-base .mat-subheader{margin:0}.mat-list-base{padding-top:8px;display:block;-webkit-tap-highlight-color:transparent}.mat-list-base .mat-subheader{height:48px;line-height:16px}.mat-list-base .mat-subheader:first-child{margin-top:-8px}.mat-list-base .mat-list-item,.mat-list-base .mat-list-option{display:block;height:48px;-webkit-tap-highlight-color:transparent;width:100%;padding:0}.mat-list-base .mat-list-item .mat-list-item-content,.mat-list-base .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list-base .mat-list-item .mat-list-item-content-reverse,.mat-list-base .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list-base .mat-list-item .mat-list-item-ripple,.mat-list-base .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list-base .mat-list-item.mat-list-item-with-avatar,.mat-list-base .mat-list-option.mat-list-item-with-avatar{height:56px}.mat-list-base .mat-list-item.mat-2-line,.mat-list-base .mat-list-option.mat-2-line{height:72px}.mat-list-base .mat-list-item.mat-3-line,.mat-list-base .mat-list-option.mat-3-line{height:88px}.mat-list-base .mat-list-item.mat-multi-line,.mat-list-base .mat-list-option.mat-multi-line{height:auto}.mat-list-base .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list-base .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list-base .mat-list-item .mat-list-text,.mat-list-base .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0}.mat-list-base .mat-list-item .mat-list-text>*,.mat-list-base .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list-base .mat-list-item .mat-list-text:empty,.mat-list-base .mat-list-option .mat-list-text:empty{display:none}.mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:0;padding-left:16px}[dir=rtl] .mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:16px;padding-left:0}.mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-left:0;padding-right:16px}[dir=rtl] .mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:0;padding-left:16px}.mat-list-base .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:16px;padding-left:16px}.mat-list-base .mat-list-item .mat-list-avatar,.mat-list-base .mat-list-option .mat-list-avatar{flex-shrink:0;width:40px;height:40px;border-radius:50%;object-fit:cover}.mat-list-base .mat-list-item .mat-list-avatar~.mat-divider-inset,.mat-list-base .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:72px;width:calc(100% - 72px)}[dir=rtl] .mat-list-base .mat-list-item .mat-list-avatar~.mat-divider-inset,[dir=rtl] .mat-list-base .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:auto;margin-right:72px}.mat-list-base .mat-list-item .mat-list-icon,.mat-list-base .mat-list-option .mat-list-icon{flex-shrink:0;width:24px;height:24px;font-size:24px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-list-base .mat-list-item .mat-list-icon~.mat-divider-inset,.mat-list-base .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:64px;width:calc(100% - 64px)}[dir=rtl] .mat-list-base .mat-list-item .mat-list-icon~.mat-divider-inset,[dir=rtl] .mat-list-base .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:auto;margin-right:64px}.mat-list-base .mat-list-item .mat-divider,.mat-list-base .mat-list-option .mat-divider{position:absolute;bottom:0;left:0;width:100%;margin:0}[dir=rtl] .mat-list-base .mat-list-item .mat-divider,[dir=rtl] .mat-list-base .mat-list-option .mat-divider{margin-left:auto;margin-right:0}.mat-list-base .mat-list-item .mat-divider.mat-divider-inset,.mat-list-base .mat-list-option .mat-divider.mat-divider-inset{position:absolute}.mat-list-base[dense]{padding-top:4px;display:block}.mat-list-base[dense] .mat-subheader{height:40px;line-height:8px}.mat-list-base[dense] .mat-subheader:first-child{margin-top:-4px}.mat-list-base[dense] .mat-list-item,.mat-list-base[dense] .mat-list-option{display:block;height:40px;-webkit-tap-highlight-color:transparent;width:100%;padding:0}.mat-list-base[dense] .mat-list-item .mat-list-item-content,.mat-list-base[dense] .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list-base[dense] .mat-list-item .mat-list-item-content-reverse,.mat-list-base[dense] .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list-base[dense] .mat-list-item .mat-list-item-ripple,.mat-list-base[dense] .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar{height:48px}.mat-list-base[dense] .mat-list-item.mat-2-line,.mat-list-base[dense] .mat-list-option.mat-2-line{height:60px}.mat-list-base[dense] .mat-list-item.mat-3-line,.mat-list-base[dense] .mat-list-option.mat-3-line{height:76px}.mat-list-base[dense] .mat-list-item.mat-multi-line,.mat-list-base[dense] .mat-list-option.mat-multi-line{height:auto}.mat-list-base[dense] .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list-base[dense] .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list-base[dense] .mat-list-item .mat-list-text,.mat-list-base[dense] .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0}.mat-list-base[dense] .mat-list-item .mat-list-text>*,.mat-list-base[dense] .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list-base[dense] .mat-list-item .mat-list-text:empty,.mat-list-base[dense] .mat-list-option .mat-list-text:empty{display:none}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:0;padding-left:16px}[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:16px;padding-left:0}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-left:0;padding-right:16px}[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:0;padding-left:16px}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:16px;padding-left:16px}.mat-list-base[dense] .mat-list-item .mat-list-avatar,.mat-list-base[dense] .mat-list-option .mat-list-avatar{flex-shrink:0;width:36px;height:36px;border-radius:50%;object-fit:cover}.mat-list-base[dense] .mat-list-item .mat-list-avatar~.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:68px;width:calc(100% - 68px)}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-list-avatar~.mat-divider-inset,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:auto;margin-right:68px}.mat-list-base[dense] .mat-list-item .mat-list-icon,.mat-list-base[dense] .mat-list-option .mat-list-icon{flex-shrink:0;width:20px;height:20px;font-size:20px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-list-base[dense] .mat-list-item .mat-list-icon~.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:60px;width:calc(100% - 60px)}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-list-icon~.mat-divider-inset,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:auto;margin-right:60px}.mat-list-base[dense] .mat-list-item .mat-divider,.mat-list-base[dense] .mat-list-option .mat-divider{position:absolute;bottom:0;left:0;width:100%;margin:0}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-divider,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-divider{margin-left:auto;margin-right:0}.mat-list-base[dense] .mat-list-item .mat-divider.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-divider.mat-divider-inset{position:absolute}.mat-nav-list a{text-decoration:none;color:inherit}.mat-nav-list .mat-list-item{cursor:pointer;outline:0}mat-action-list button{background:0 0;color:inherit;border:none;font:inherit;outline:inherit}mat-action-list .mat-list-item{cursor:pointer;outline:inherit}.mat-list-option:not(.mat-list-item-disabled){cursor:pointer;outline:0}@media (-ms-high-contrast:active){.mat-selection-list:focus{outline-style:dotted}.mat-list-option:focus,.mat-list-option:hover,.mat-nav-list .mat-list-item:focus,.mat-nav-list .mat-list-item:hover,mat-action-list .mat-list-item:focus,mat-action-list .mat-list-item:hover{outline:dotted 1px}}@media (hover:none){.mat-list-option:not(.mat-list-item-disabled):hover,.mat-nav-list .mat-list-item:not(.mat-list-item-disabled):hover{background:0 0}}"],inputs:["disableRipple"],encapsulation:i.ViewEncapsulation.None,changeDetection:i.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:i.ElementRef}]},e}(m),b=function(){function t(){}return t.decorators=[{type:i.Directive,args:[{selector:"[mat-list-avatar], [matListAvatar]",host:{class:"mat-list-avatar"}}]}],t}(),x=function(){function t(){}return t.decorators=[{type:i.Directive,args:[{selector:"[mat-list-icon], [matListIcon]",host:{class:"mat-list-icon"}}]}],t}(),w=function(){function t(){}return t.decorators=[{type:i.Directive,args:[{selector:"[mat-subheader], [matSubheader]",host:{class:"mat-subheader"}}]}],t}(),j=function(t){function e(e,n,r){var i=t.call(this)||this;i._element=e,i._isInteractiveList=!1,i._isInteractiveList=!!(n||r&&"action-list"===r._getListType()),i._list=n||r;var o=i._getHostElement();return"button"!==o.nodeName.toLowerCase()||o.hasAttribute("type")||o.setAttribute("type","button"),i}return Object(r.__extends)(e,t),e.prototype.ngAfterContentInit=function(){Object(o.setLines)(this._lines,this._element)},e.prototype._isRippleDisabled=function(){return!this._isInteractiveList||this.disableRipple||!(!this._list||!this._list.disableRipple)},e.prototype._getHostElement=function(){return this._element.nativeElement},e.decorators=[{type:i.Component,args:[{selector:"mat-list-item, a[mat-list-item], button[mat-list-item]",exportAs:"matListItem",host:{class:"mat-list-item","[class.mat-list-item-avatar]":"_avatar || _icon","[class.mat-list-item-with-avatar]":"_avatar || _icon"},inputs:["disableRipple"],template:'<div class="mat-list-item-content"><div class="mat-list-item-ripple" mat-ripple [matRippleTrigger]="_getHostElement()" [matRippleDisabled]="_isRippleDisabled()"></div><ng-content select="[mat-list-avatar], [mat-list-icon], [matListAvatar], [matListIcon]"></ng-content><div class="mat-list-text"><ng-content select="[mat-line], [matLine]"></ng-content></div><ng-content></ng-content></div>',encapsulation:i.ViewEncapsulation.None,changeDetection:i.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:i.ElementRef},{type:_,decorators:[{type:i.Optional}]},{type:v,decorators:[{type:i.Optional}]}]},e.propDecorators={_lines:[{type:i.ContentChildren,args:[o.MatLine]}],_avatar:[{type:i.ContentChild,args:[b]}],_icon:[{type:i.ContentChild,args:[x]}]},e}(y),O=function(){return function(){}}(),C=Object(o.mixinDisableRipple)(O),E=function(){return function(){}}(),S=Object(o.mixinDisableRipple)(E),M={provide:l.NG_VALUE_ACCESSOR,useExisting:Object(i.forwardRef)(function(){return A}),multi:!0},T=function(){return function(t,e){this.source=t,this.option=e}}(),k=function(t){function e(e,n,r){var i=t.call(this)||this;return i._element=e,i._changeDetector=n,i.selectionList=r,i._selected=!1,i._disabled=!1,i._hasFocus=!1,i.checkboxPosition="after",i}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"disabled",{get:function(){return this._disabled||this.selectionList&&this.selectionList.disabled},set:function(t){var e=Object(s.coerceBooleanProperty)(t);e!==this._disabled&&(this._disabled=e,this._changeDetector.markForCheck())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selected",{get:function(){return this.selectionList.selectedOptions.isSelected(this)},set:function(t){var e=Object(s.coerceBooleanProperty)(t);e!==this._selected&&(this._setSelected(e),this.selectionList._reportValueChange())},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){var t=this,e=this._selected;Promise.resolve().then(function(){(t._selected||e)&&(t.selected=!0,t._changeDetector.markForCheck())})},e.prototype.ngAfterContentInit=function(){Object(o.setLines)(this._lines,this._element)},e.prototype.ngOnDestroy=function(){var t=this;this.selected&&Promise.resolve().then(function(){return t.selected=!1});var e=this._hasFocus,n=this.selectionList._removeOptionFromList(this);e&&n&&n.focus()},e.prototype.toggle=function(){this.selected=!this.selected},e.prototype.focus=function(){this._element.nativeElement.focus()},e.prototype.getLabel=function(){return this._text?this._text.nativeElement.textContent||"":""},e.prototype._isRippleDisabled=function(){return this.disabled||this.disableRipple||this.selectionList.disableRipple},e.prototype._handleClick=function(){this.disabled||(this.toggle(),this.selectionList._emitChangeEvent(this))},e.prototype._handleFocus=function(){this.selectionList._setFocusedOption(this),this._hasFocus=!0},e.prototype._handleBlur=function(){this.selectionList._onTouched(),this._hasFocus=!1},e.prototype._getHostElement=function(){return this._element.nativeElement},e.prototype._setSelected=function(t){return t!==this._selected&&(this._selected=t,t?this.selectionList.selectedOptions.select(this):this.selectionList.selectedOptions.deselect(this),this._changeDetector.markForCheck(),!0)},e.prototype._markForCheck=function(){this._changeDetector.markForCheck()},e.decorators=[{type:i.Component,args:[{selector:"mat-list-option",exportAs:"matListOption",inputs:["disableRipple"],host:{role:"option",class:"mat-list-item mat-list-option","(focus)":"_handleFocus()","(blur)":"_handleBlur()","(click)":"_handleClick()",tabindex:"-1","[class.mat-list-item-disabled]":"disabled","[class.mat-list-item-with-avatar]":"_avatar || _icon","[attr.aria-selected]":"selected.toString()","[attr.aria-disabled]":"disabled.toString()"},template:'<div class="mat-list-item-content" [class.mat-list-item-content-reverse]="checkboxPosition == \'after\'"><div mat-ripple class="mat-list-item-ripple" [matRippleTrigger]="_getHostElement()" [matRippleDisabled]="_isRippleDisabled()"></div><mat-pseudo-checkbox [state]="selected ? \'checked\' : \'unchecked\'" [disabled]="disabled"></mat-pseudo-checkbox><div class="mat-list-text" #text><ng-content></ng-content></div><ng-content select="[mat-list-avatar], [mat-list-icon], [matListAvatar], [matListIcon]"></ng-content></div>',encapsulation:i.ViewEncapsulation.None,changeDetection:i.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:i.ElementRef},{type:i.ChangeDetectorRef},{type:A,decorators:[{type:i.Inject,args:[Object(i.forwardRef)(function(){return A})]}]}]},e.propDecorators={_avatar:[{type:i.ContentChild,args:[b]}],_icon:[{type:i.ContentChild,args:[x]}],_lines:[{type:i.ContentChildren,args:[o.MatLine]}],_text:[{type:i.ViewChild,args:["text"]}],checkboxPosition:[{type:i.Input}],value:[{type:i.Input}],disabled:[{type:i.Input}],selected:[{type:i.Input}]},e}(S),A=function(t){function e(e,n){var r=t.call(this)||this;return r._element=e,r.selectionChange=new i.EventEmitter,r.tabIndex=0,r._disabled=!1,r.selectedOptions=new u.SelectionModel(!0),r._onChange=function(t){},r._modelChanges=d.Subscription.EMPTY,r._onTouched=function(){},r.tabIndex=parseInt(n)||0,r}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"disabled",{get:function(){return this._disabled},set:function(t){this._disabled=Object(s.coerceBooleanProperty)(t),this.options&&this.options.forEach(function(t){return t._markForCheck()})},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){this._keyManager=new a.FocusKeyManager(this.options).withWrap().withTypeAhead().skipPredicate(function(){return!1}).withAllowedModifierKeys(["shiftKey"]),this._tempValues&&(this._setOptionsFromValues(this._tempValues),this._tempValues=null),this._modelChanges=this.selectedOptions.onChange.subscribe(function(t){if(t.added)for(var e=0,n=t.added;e<n.length;e++){n[e].selected=!0}if(t.removed)for(var r=0,i=t.removed;r<i.length;r++){i[r].selected=!1}})},e.prototype.ngOnDestroy=function(){this._modelChanges.unsubscribe()},e.prototype.focus=function(){this._element.nativeElement.focus()},e.prototype.selectAll=function(){this._setAllOptionsSelected(!0)},e.prototype.deselectAll=function(){this._setAllOptionsSelected(!1)},e.prototype._setFocusedOption=function(t){this._keyManager.updateActiveItemIndex(this._getOptionIndex(t))},e.prototype._removeOptionFromList=function(t){var e=this._getOptionIndex(t);return e>-1&&this._keyManager.activeItemIndex===e&&(e>0?this._keyManager.updateActiveItemIndex(e-1):0===e&&this.options.length>1&&this._keyManager.updateActiveItemIndex(Math.min(e+1,this.options.length-1))),this._keyManager.activeItem},e.prototype._keydown=function(t){var e=t.keyCode,n=this._keyManager,r=n.activeItemIndex,i=Object(c.hasModifierKey)(t);switch(e){case c.SPACE:case c.ENTER:i||(this._toggleFocusedOption(),t.preventDefault());break;case c.HOME:case c.END:i||(e===c.HOME?n.setFirstItemActive():n.setLastItemActive(),t.preventDefault());break;case c.A:Object(c.hasModifierKey)(t,"ctrlKey")&&(this.options.find(function(t){return!t.selected})?this.selectAll():this.deselectAll(),t.preventDefault());break;default:n.onKeydown(t)}e!==c.UP_ARROW&&e!==c.DOWN_ARROW||!t.shiftKey||n.activeItemIndex===r||this._toggleFocusedOption()},e.prototype._reportValueChange=function(){this.options&&this._onChange(this._getSelectedOptionValues())},e.prototype._emitChangeEvent=function(t){this.selectionChange.emit(new T(this,t))},e.prototype.writeValue=function(t){this.options?this._setOptionsFromValues(t||[]):this._tempValues=t},e.prototype.setDisabledState=function(t){this.disabled=t},e.prototype.registerOnChange=function(t){this._onChange=t},e.prototype.registerOnTouched=function(t){this._onTouched=t},e.prototype._setOptionsFromValues=function(t){var e=this;this.options.forEach(function(t){return t._setSelected(!1)}),t.forEach(function(t){var n=e.options.find(function(n){return!n.selected&&(e.compareWith?e.compareWith(n.value,t):n.value===t)});n&&n._setSelected(!0)})},e.prototype._getSelectedOptionValues=function(){return this.options.filter(function(t){return t.selected}).map(function(t){return t.value})},e.prototype._toggleFocusedOption=function(){var t=this._keyManager.activeItemIndex;if(null!=t&&this._isValidIndex(t)){var e=this.options.toArray()[t];e&&!e.disabled&&(e.toggle(),this._emitChangeEvent(e))}},e.prototype._setAllOptionsSelected=function(t){var e=!1;this.options.forEach(function(n){n._setSelected(t)&&(e=!0)}),e&&this._reportValueChange()},e.prototype._isValidIndex=function(t){return t>=0&&t<this.options.length},e.prototype._getOptionIndex=function(t){return this.options.toArray().indexOf(t)},e.decorators=[{type:i.Component,args:[{selector:"mat-selection-list",exportAs:"matSelectionList",inputs:["disabled","disableRipple","tabIndex"],host:{role:"listbox","[tabIndex]":"tabIndex",class:"mat-selection-list mat-list-base","(focus)":"focus()","(blur)":"_onTouched()","(keydown)":"_keydown($event)","aria-multiselectable":"true","[attr.aria-disabled]":"disabled.toString()"},template:"<ng-content></ng-content>",styles:[".mat-subheader{display:flex;box-sizing:border-box;padding:16px;align-items:center}.mat-list-base .mat-subheader{margin:0}.mat-list-base{padding-top:8px;display:block;-webkit-tap-highlight-color:transparent}.mat-list-base .mat-subheader{height:48px;line-height:16px}.mat-list-base .mat-subheader:first-child{margin-top:-8px}.mat-list-base .mat-list-item,.mat-list-base .mat-list-option{display:block;height:48px;-webkit-tap-highlight-color:transparent;width:100%;padding:0}.mat-list-base .mat-list-item .mat-list-item-content,.mat-list-base .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list-base .mat-list-item .mat-list-item-content-reverse,.mat-list-base .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list-base .mat-list-item .mat-list-item-ripple,.mat-list-base .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list-base .mat-list-item.mat-list-item-with-avatar,.mat-list-base .mat-list-option.mat-list-item-with-avatar{height:56px}.mat-list-base .mat-list-item.mat-2-line,.mat-list-base .mat-list-option.mat-2-line{height:72px}.mat-list-base .mat-list-item.mat-3-line,.mat-list-base .mat-list-option.mat-3-line{height:88px}.mat-list-base .mat-list-item.mat-multi-line,.mat-list-base .mat-list-option.mat-multi-line{height:auto}.mat-list-base .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list-base .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list-base .mat-list-item .mat-list-text,.mat-list-base .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0}.mat-list-base .mat-list-item .mat-list-text>*,.mat-list-base .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list-base .mat-list-item .mat-list-text:empty,.mat-list-base .mat-list-option .mat-list-text:empty{display:none}.mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:0;padding-left:16px}[dir=rtl] .mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:16px;padding-left:0}.mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-left:0;padding-right:16px}[dir=rtl] .mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:0;padding-left:16px}.mat-list-base .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:16px;padding-left:16px}.mat-list-base .mat-list-item .mat-list-avatar,.mat-list-base .mat-list-option .mat-list-avatar{flex-shrink:0;width:40px;height:40px;border-radius:50%;object-fit:cover}.mat-list-base .mat-list-item .mat-list-avatar~.mat-divider-inset,.mat-list-base .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:72px;width:calc(100% - 72px)}[dir=rtl] .mat-list-base .mat-list-item .mat-list-avatar~.mat-divider-inset,[dir=rtl] .mat-list-base .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:auto;margin-right:72px}.mat-list-base .mat-list-item .mat-list-icon,.mat-list-base .mat-list-option .mat-list-icon{flex-shrink:0;width:24px;height:24px;font-size:24px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-list-base .mat-list-item .mat-list-icon~.mat-divider-inset,.mat-list-base .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:64px;width:calc(100% - 64px)}[dir=rtl] .mat-list-base .mat-list-item .mat-list-icon~.mat-divider-inset,[dir=rtl] .mat-list-base .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:auto;margin-right:64px}.mat-list-base .mat-list-item .mat-divider,.mat-list-base .mat-list-option .mat-divider{position:absolute;bottom:0;left:0;width:100%;margin:0}[dir=rtl] .mat-list-base .mat-list-item .mat-divider,[dir=rtl] .mat-list-base .mat-list-option .mat-divider{margin-left:auto;margin-right:0}.mat-list-base .mat-list-item .mat-divider.mat-divider-inset,.mat-list-base .mat-list-option .mat-divider.mat-divider-inset{position:absolute}.mat-list-base[dense]{padding-top:4px;display:block}.mat-list-base[dense] .mat-subheader{height:40px;line-height:8px}.mat-list-base[dense] .mat-subheader:first-child{margin-top:-4px}.mat-list-base[dense] .mat-list-item,.mat-list-base[dense] .mat-list-option{display:block;height:40px;-webkit-tap-highlight-color:transparent;width:100%;padding:0}.mat-list-base[dense] .mat-list-item .mat-list-item-content,.mat-list-base[dense] .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list-base[dense] .mat-list-item .mat-list-item-content-reverse,.mat-list-base[dense] .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list-base[dense] .mat-list-item .mat-list-item-ripple,.mat-list-base[dense] .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar{height:48px}.mat-list-base[dense] .mat-list-item.mat-2-line,.mat-list-base[dense] .mat-list-option.mat-2-line{height:60px}.mat-list-base[dense] .mat-list-item.mat-3-line,.mat-list-base[dense] .mat-list-option.mat-3-line{height:76px}.mat-list-base[dense] .mat-list-item.mat-multi-line,.mat-list-base[dense] .mat-list-option.mat-multi-line{height:auto}.mat-list-base[dense] .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list-base[dense] .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list-base[dense] .mat-list-item .mat-list-text,.mat-list-base[dense] .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0}.mat-list-base[dense] .mat-list-item .mat-list-text>*,.mat-list-base[dense] .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list-base[dense] .mat-list-item .mat-list-text:empty,.mat-list-base[dense] .mat-list-option .mat-list-text:empty{display:none}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:0;padding-left:16px}[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:16px;padding-left:0}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-left:0;padding-right:16px}[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:0;padding-left:16px}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:16px;padding-left:16px}.mat-list-base[dense] .mat-list-item .mat-list-avatar,.mat-list-base[dense] .mat-list-option .mat-list-avatar{flex-shrink:0;width:36px;height:36px;border-radius:50%;object-fit:cover}.mat-list-base[dense] .mat-list-item .mat-list-avatar~.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:68px;width:calc(100% - 68px)}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-list-avatar~.mat-divider-inset,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:auto;margin-right:68px}.mat-list-base[dense] .mat-list-item .mat-list-icon,.mat-list-base[dense] .mat-list-option .mat-list-icon{flex-shrink:0;width:20px;height:20px;font-size:20px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-list-base[dense] .mat-list-item .mat-list-icon~.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:60px;width:calc(100% - 60px)}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-list-icon~.mat-divider-inset,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:auto;margin-right:60px}.mat-list-base[dense] .mat-list-item .mat-divider,.mat-list-base[dense] .mat-list-option .mat-divider{position:absolute;bottom:0;left:0;width:100%;margin:0}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-divider,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-divider{margin-left:auto;margin-right:0}.mat-list-base[dense] .mat-list-item .mat-divider.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-divider.mat-divider-inset{position:absolute}.mat-nav-list a{text-decoration:none;color:inherit}.mat-nav-list .mat-list-item{cursor:pointer;outline:0}mat-action-list button{background:0 0;color:inherit;border:none;font:inherit;outline:inherit}mat-action-list .mat-list-item{cursor:pointer;outline:inherit}.mat-list-option:not(.mat-list-item-disabled){cursor:pointer;outline:0}@media (-ms-high-contrast:active){.mat-selection-list:focus{outline-style:dotted}.mat-list-option:focus,.mat-list-option:hover,.mat-nav-list .mat-list-item:focus,.mat-nav-list .mat-list-item:hover,mat-action-list .mat-list-item:focus,mat-action-list .mat-list-item:hover{outline:dotted 1px}}@media (hover:none){.mat-list-option:not(.mat-list-item-disabled):hover,.mat-nav-list .mat-list-item:not(.mat-list-item-disabled):hover{background:0 0}}"],encapsulation:i.ViewEncapsulation.None,providers:[M],changeDetection:i.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:i.ElementRef},{type:String,decorators:[{type:i.Attribute,args:["tabindex"]}]}]},e.propDecorators={options:[{type:i.ContentChildren,args:[k]}],selectionChange:[{type:i.Output}],tabIndex:[{type:i.Input}],compareWith:[{type:i.Input}],disabled:[{type:i.Input}]},e}(C),D=function(){function t(){}return t.decorators=[{type:i.NgModule,args:[{imports:[o.MatLineModule,o.MatRippleModule,o.MatCommonModule,o.MatPseudoCheckboxModule,p.CommonModule],exports:[v,_,j,b,o.MatLineModule,o.MatCommonModule,x,w,o.MatPseudoCheckboxModule,A,k,f.MatDividerModule],declarations:[v,_,j,b,x,w,A,k]}]}],t}()},"./node_modules/@angular/material/esm5/material.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"VERSION",function(){return B});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/material/esm5/autocomplete.es5.js");n.d(e,"ɵa28",function(){return i["ɵa28"]}),n.d(e,"MAT_AUTOCOMPLETE_DEFAULT_OPTIONS_FACTORY",function(){return i.MAT_AUTOCOMPLETE_DEFAULT_OPTIONS_FACTORY}),n.d(e,"MatAutocompleteSelectedEvent",function(){return i.MatAutocompleteSelectedEvent}),n.d(e,"MatAutocompleteBase",function(){return i.MatAutocompleteBase}),n.d(e,"_MatAutocompleteMixinBase",function(){return i._MatAutocompleteMixinBase}),n.d(e,"MAT_AUTOCOMPLETE_DEFAULT_OPTIONS",function(){return i.MAT_AUTOCOMPLETE_DEFAULT_OPTIONS}),n.d(e,"MatAutocomplete",function(){return i.MatAutocomplete}),n.d(e,"MatAutocompleteModule",function(){return i.MatAutocompleteModule}),n.d(e,"MAT_AUTOCOMPLETE_SCROLL_STRATEGY_FACTORY",function(){return i.MAT_AUTOCOMPLETE_SCROLL_STRATEGY_FACTORY}),n.d(e,"getMatAutocompleteMissingPanelError",function(){return i.getMatAutocompleteMissingPanelError}),n.d(e,"AUTOCOMPLETE_OPTION_HEIGHT",function(){return i.AUTOCOMPLETE_OPTION_HEIGHT}),n.d(e,"AUTOCOMPLETE_PANEL_HEIGHT",function(){return i.AUTOCOMPLETE_PANEL_HEIGHT}),n.d(e,"MAT_AUTOCOMPLETE_SCROLL_STRATEGY",function(){return i.MAT_AUTOCOMPLETE_SCROLL_STRATEGY}),n.d(e,"MAT_AUTOCOMPLETE_SCROLL_STRATEGY_FACTORY_PROVIDER",function(){return i.MAT_AUTOCOMPLETE_SCROLL_STRATEGY_FACTORY_PROVIDER}),n.d(e,"MAT_AUTOCOMPLETE_VALUE_ACCESSOR",function(){return i.MAT_AUTOCOMPLETE_VALUE_ACCESSOR}),n.d(e,"MatAutocompleteTrigger",function(){return i.MatAutocompleteTrigger});var o=n("./node_modules/@angular/material/esm5/badge.es5.js");n.d(e,"MatBadgeModule",function(){return o.MatBadgeModule}),n.d(e,"MatBadgeBase",function(){return o.MatBadgeBase}),n.d(e,"_MatBadgeMixinBase",function(){return o._MatBadgeMixinBase}),n.d(e,"MatBadge",function(){return o.MatBadge});var a=n("./node_modules/@angular/material/esm5/bottom-sheet.es5.js");n.d(e,"MatBottomSheetModule",function(){return a.MatBottomSheetModule}),n.d(e,"MAT_BOTTOM_SHEET_DEFAULT_OPTIONS",function(){return a.MAT_BOTTOM_SHEET_DEFAULT_OPTIONS}),n.d(e,"MatBottomSheet",function(){return a.MatBottomSheet}),n.d(e,"MAT_BOTTOM_SHEET_DATA",function(){return a.MAT_BOTTOM_SHEET_DATA}),n.d(e,"MatBottomSheetConfig",function(){return a.MatBottomSheetConfig}),n.d(e,"MatBottomSheetContainer",function(){return a.MatBottomSheetContainer}),n.d(e,"matBottomSheetAnimations",function(){return a.matBottomSheetAnimations}),n.d(e,"MatBottomSheetRef",function(){return a.MatBottomSheetRef});var s=n("./node_modules/@angular/material/esm5/button.es5.js");n.d(e,"MatButtonModule",function(){return s.MatButtonModule}),n.d(e,"MatButtonBase",function(){return s.MatButtonBase}),n.d(e,"_MatButtonMixinBase",function(){return s._MatButtonMixinBase}),n.d(e,"MatButton",function(){return s.MatButton}),n.d(e,"MatAnchor",function(){return s.MatAnchor});var u=n("./node_modules/@angular/material/esm5/button-toggle.es5.js");n.d(e,"MAT_BUTTON_TOGGLE_DEFAULT_OPTIONS",function(){return u.MAT_BUTTON_TOGGLE_DEFAULT_OPTIONS}),n.d(e,"MAT_BUTTON_TOGGLE_GROUP_VALUE_ACCESSOR",function(){return u.MAT_BUTTON_TOGGLE_GROUP_VALUE_ACCESSOR}),n.d(e,"MatButtonToggleGroupMultiple",function(){return u.MatButtonToggleGroupMultiple}),n.d(e,"MatButtonToggleChange",function(){return u.MatButtonToggleChange}),n.d(e,"MatButtonToggleGroup",function(){return u.MatButtonToggleGroup}),n.d(e,"MatButtonToggleBase",function(){return u.MatButtonToggleBase}),n.d(e,"_MatButtonToggleMixinBase",function(){return u._MatButtonToggleMixinBase}),n.d(e,"MatButtonToggle",function(){return u.MatButtonToggle}),n.d(e,"MatButtonToggleModule",function(){return u.MatButtonToggleModule});var c=n("./node_modules/@angular/material/esm5/card.es5.js");n.d(e,"MatCardContent",function(){return c.MatCardContent}),n.d(e,"MatCardTitle",function(){return c.MatCardTitle}),n.d(e,"MatCardSubtitle",function(){return c.MatCardSubtitle}),n.d(e,"MatCardActions",function(){return c.MatCardActions}),n.d(e,"MatCardFooter",function(){return c.MatCardFooter}),n.d(e,"MatCardImage",function(){return c.MatCardImage}),n.d(e,"MatCardSmImage",function(){return c.MatCardSmImage}),n.d(e,"MatCardMdImage",function(){return c.MatCardMdImage}),n.d(e,"MatCardLgImage",function(){return c.MatCardLgImage}),n.d(e,"MatCardXlImage",function(){return c.MatCardXlImage}),n.d(e,"MatCardAvatar",function(){return c.MatCardAvatar}),n.d(e,"MatCard",function(){return c.MatCard}),n.d(e,"MatCardHeader",function(){return c.MatCardHeader}),n.d(e,"MatCardTitleGroup",function(){return c.MatCardTitleGroup}),n.d(e,"MatCardModule",function(){return c.MatCardModule});var l=n("./node_modules/@angular/material/esm5/checkbox.es5.js");n.d(e,"MAT_CHECKBOX_CONTROL_VALUE_ACCESSOR",function(){return l.MAT_CHECKBOX_CONTROL_VALUE_ACCESSOR}),n.d(e,"TransitionCheckState",function(){return l.TransitionCheckState}),n.d(e,"MatCheckboxChange",function(){return l.MatCheckboxChange}),n.d(e,"MatCheckboxBase",function(){return l.MatCheckboxBase}),n.d(e,"_MatCheckboxMixinBase",function(){return l._MatCheckboxMixinBase}),n.d(e,"MatCheckbox",function(){return l.MatCheckbox}),n.d(e,"MAT_CHECKBOX_CLICK_ACTION",function(){return l.MAT_CHECKBOX_CLICK_ACTION}),n.d(e,"MatCheckboxModule",function(){return l.MatCheckboxModule}),n.d(e,"MAT_CHECKBOX_REQUIRED_VALIDATOR",function(){return l.MAT_CHECKBOX_REQUIRED_VALIDATOR}),n.d(e,"MatCheckboxRequiredValidator",function(){return l.MatCheckboxRequiredValidator});var d=n("./node_modules/@angular/material/esm5/chips.es5.js");n.d(e,"MatChipsModule",function(){return d.MatChipsModule}),n.d(e,"MatChipListBase",function(){return d.MatChipListBase}),n.d(e,"_MatChipListMixinBase",function(){return d._MatChipListMixinBase}),n.d(e,"MatChipListChange",function(){return d.MatChipListChange}),n.d(e,"MatChipList",function(){return d.MatChipList}),n.d(e,"MatChipSelectionChange",function(){return d.MatChipSelectionChange}),n.d(e,"MatChipBase",function(){return d.MatChipBase}),n.d(e,"_MatChipMixinBase",function(){return d._MatChipMixinBase}),n.d(e,"MatChipAvatar",function(){return d.MatChipAvatar}),n.d(e,"MatChipTrailingIcon",function(){return d.MatChipTrailingIcon}),n.d(e,"MatChip",function(){return d.MatChip}),n.d(e,"MatChipRemove",function(){return d.MatChipRemove}),n.d(e,"MatChipInput",function(){return d.MatChipInput}),n.d(e,"MAT_CHIPS_DEFAULT_OPTIONS",function(){return d.MAT_CHIPS_DEFAULT_OPTIONS});var p=n("./node_modules/@angular/material/esm5/core.es5.js");n.d(e,"ɵa1",function(){return p["ɵa1"]}),n.d(e,"AnimationCurves",function(){return p.AnimationCurves}),n.d(e,"AnimationDurations",function(){return p.AnimationDurations}),n.d(e,"MatCommonModule",function(){return p.MatCommonModule}),n.d(e,"MATERIAL_SANITY_CHECKS",function(){return p.MATERIAL_SANITY_CHECKS}),n.d(e,"mixinDisabled",function(){return p.mixinDisabled}),n.d(e,"mixinColor",function(){return p.mixinColor}),n.d(e,"mixinDisableRipple",function(){return p.mixinDisableRipple}),n.d(e,"mixinTabIndex",function(){return p.mixinTabIndex}),n.d(e,"mixinErrorState",function(){return p.mixinErrorState}),n.d(e,"mixinInitialized",function(){return p.mixinInitialized}),n.d(e,"NativeDateModule",function(){return p.NativeDateModule}),n.d(e,"MatNativeDateModule",function(){return p.MatNativeDateModule}),n.d(e,"MAT_DATE_LOCALE_FACTORY",function(){return p.MAT_DATE_LOCALE_FACTORY}),n.d(e,"MAT_DATE_LOCALE",function(){return p.MAT_DATE_LOCALE}),n.d(e,"MAT_DATE_LOCALE_PROVIDER",function(){return p.MAT_DATE_LOCALE_PROVIDER}),n.d(e,"DateAdapter",function(){return p.DateAdapter}),n.d(e,"MAT_DATE_FORMATS",function(){return p.MAT_DATE_FORMATS}),n.d(e,"NativeDateAdapter",function(){return p.NativeDateAdapter}),n.d(e,"MAT_NATIVE_DATE_FORMATS",function(){return p.MAT_NATIVE_DATE_FORMATS}),n.d(e,"ShowOnDirtyErrorStateMatcher",function(){return p.ShowOnDirtyErrorStateMatcher}),n.d(e,"ErrorStateMatcher",function(){return p.ErrorStateMatcher}),n.d(e,"MAT_HAMMER_OPTIONS",function(){return p.MAT_HAMMER_OPTIONS}),n.d(e,"GestureConfig",function(){return p.GestureConfig}),n.d(e,"setLines",function(){return p.setLines}),n.d(e,"MatLine",function(){return p.MatLine}),n.d(e,"MatLineSetter",function(){return p.MatLineSetter}),n.d(e,"MatLineModule",function(){return p.MatLineModule}),n.d(e,"MatOptionModule",function(){return p.MatOptionModule}),n.d(e,"_countGroupLabelsBeforeOption",function(){return p._countGroupLabelsBeforeOption}),n.d(e,"_getOptionScrollPosition",function(){return p._getOptionScrollPosition}),n.d(e,"MatOptionSelectionChange",function(){return p.MatOptionSelectionChange}),n.d(e,"MAT_OPTION_PARENT_COMPONENT",function(){return p.MAT_OPTION_PARENT_COMPONENT}),n.d(e,"MatOption",function(){return p.MatOption}),n.d(e,"MatOptgroupBase",function(){return p.MatOptgroupBase}),n.d(e,"_MatOptgroupMixinBase",function(){return p._MatOptgroupMixinBase}),n.d(e,"MatOptgroup",function(){return p.MatOptgroup}),n.d(e,"MAT_LABEL_GLOBAL_OPTIONS",function(){return p.MAT_LABEL_GLOBAL_OPTIONS}),n.d(e,"MatRippleModule",function(){return p.MatRippleModule}),n.d(e,"MAT_RIPPLE_GLOBAL_OPTIONS",function(){return p.MAT_RIPPLE_GLOBAL_OPTIONS}),n.d(e,"MatRipple",function(){return p.MatRipple}),n.d(e,"RippleState",function(){return p.RippleState}),n.d(e,"RippleRef",function(){return p.RippleRef}),n.d(e,"defaultRippleAnimationConfig",function(){return p.defaultRippleAnimationConfig}),n.d(e,"RippleRenderer",function(){return p.RippleRenderer}),n.d(e,"MatPseudoCheckboxModule",function(){return p.MatPseudoCheckboxModule}),n.d(e,"MatPseudoCheckbox",function(){return p.MatPseudoCheckbox}),n.d(e,"JAN",function(){return p.JAN}),n.d(e,"FEB",function(){return p.FEB}),n.d(e,"MAR",function(){return p.MAR}),n.d(e,"APR",function(){return p.APR}),n.d(e,"MAY",function(){return p.MAY}),n.d(e,"JUN",function(){return p.JUN}),n.d(e,"JUL",function(){return p.JUL}),n.d(e,"AUG",function(){return p.AUG}),n.d(e,"SEP",function(){return p.SEP}),n.d(e,"OCT",function(){return p.OCT}),n.d(e,"NOV",function(){return p.NOV}),n.d(e,"DEC",function(){return p.DEC});var f=n("./node_modules/@angular/material/esm5/datepicker.es5.js");n.d(e,"ɵa34",function(){return f["ɵa34"]}),n.d(e,"MatDatepickerModule",function(){return f.MatDatepickerModule}),n.d(e,"MatCalendarHeader",function(){return f.MatCalendarHeader}),n.d(e,"MatCalendar",function(){return f.MatCalendar}),n.d(e,"MatCalendarCell",function(){return f.MatCalendarCell}),n.d(e,"MatCalendarBody",function(){return f.MatCalendarBody}),n.d(e,"MAT_DATEPICKER_SCROLL_STRATEGY_FACTORY",function(){return f.MAT_DATEPICKER_SCROLL_STRATEGY_FACTORY}),n.d(e,"MAT_DATEPICKER_SCROLL_STRATEGY",function(){return f.MAT_DATEPICKER_SCROLL_STRATEGY}),n.d(e,"MAT_DATEPICKER_SCROLL_STRATEGY_FACTORY_PROVIDER",function(){return f.MAT_DATEPICKER_SCROLL_STRATEGY_FACTORY_PROVIDER}),n.d(e,"MatDatepickerContentBase",function(){return f.MatDatepickerContentBase}),n.d(e,"_MatDatepickerContentMixinBase",function(){return f._MatDatepickerContentMixinBase}),n.d(e,"MatDatepickerContent",function(){return f.MatDatepickerContent}),n.d(e,"MatDatepicker",function(){return f.MatDatepicker}),n.d(e,"matDatepickerAnimations",function(){return f.matDatepickerAnimations}),n.d(e,"MAT_DATEPICKER_VALUE_ACCESSOR",function(){return f.MAT_DATEPICKER_VALUE_ACCESSOR}),n.d(e,"MAT_DATEPICKER_VALIDATORS",function(){return f.MAT_DATEPICKER_VALIDATORS}),n.d(e,"MatDatepickerInputEvent",function(){return f.MatDatepickerInputEvent}),n.d(e,"MatDatepickerInput",function(){return f.MatDatepickerInput}),n.d(e,"MatDatepickerIntl",function(){return f.MatDatepickerIntl}),n.d(e,"MatDatepickerToggleIcon",function(){return f.MatDatepickerToggleIcon}),n.d(e,"MatDatepickerToggle",function(){return f.MatDatepickerToggle}),n.d(e,"MatMonthView",function(){return f.MatMonthView}),n.d(e,"MatYearView",function(){return f.MatYearView});var h=n("./node_modules/@angular/material/esm5/dialog.es5.js");n.d(e,"MatDialogModule",function(){return h.MatDialogModule}),n.d(e,"MAT_DIALOG_SCROLL_STRATEGY_FACTORY",function(){return h.MAT_DIALOG_SCROLL_STRATEGY_FACTORY}),n.d(e,"MAT_DIALOG_SCROLL_STRATEGY_PROVIDER_FACTORY",function(){return h.MAT_DIALOG_SCROLL_STRATEGY_PROVIDER_FACTORY}),n.d(e,"MAT_DIALOG_DATA",function(){return h.MAT_DIALOG_DATA}),n.d(e,"MAT_DIALOG_DEFAULT_OPTIONS",function(){return h.MAT_DIALOG_DEFAULT_OPTIONS}),n.d(e,"MAT_DIALOG_SCROLL_STRATEGY",function(){return h.MAT_DIALOG_SCROLL_STRATEGY}),n.d(e,"MAT_DIALOG_SCROLL_STRATEGY_PROVIDER",function(){return h.MAT_DIALOG_SCROLL_STRATEGY_PROVIDER}),n.d(e,"MatDialog",function(){return h.MatDialog}),n.d(e,"throwMatDialogContentAlreadyAttachedError",function(){return h.throwMatDialogContentAlreadyAttachedError}),n.d(e,"MatDialogContainer",function(){return h.MatDialogContainer}),n.d(e,"MatDialogClose",function(){return h.MatDialogClose}),n.d(e,"MatDialogTitle",function(){return h.MatDialogTitle}),n.d(e,"MatDialogContent",function(){return h.MatDialogContent}),n.d(e,"MatDialogActions",function(){return h.MatDialogActions}),n.d(e,"MatDialogConfig",function(){return h.MatDialogConfig}),n.d(e,"MatDialogRef",function(){return h.MatDialogRef}),n.d(e,"matDialogAnimations",function(){return h.matDialogAnimations});var m=n("./node_modules/@angular/material/esm5/divider.es5.js");n.d(e,"MatDivider",function(){return m.MatDivider}),n.d(e,"MatDividerModule",function(){return m.MatDividerModule});var g=n("./node_modules/@angular/material/esm5/expansion.es5.js");n.d(e,"MatExpansionModule",function(){return g.MatExpansionModule}),n.d(e,"MatAccordion",function(){return g.MatAccordion}),n.d(e,"MAT_ACCORDION",function(){return g.MAT_ACCORDION}),n.d(e,"MAT_EXPANSION_PANEL_DEFAULT_OPTIONS",function(){return g.MAT_EXPANSION_PANEL_DEFAULT_OPTIONS}),n.d(e,"MatExpansionPanel",function(){return g.MatExpansionPanel}),n.d(e,"MatExpansionPanelActionRow",function(){return g.MatExpansionPanelActionRow}),n.d(e,"MatExpansionPanelHeader",function(){return g.MatExpansionPanelHeader}),n.d(e,"MatExpansionPanelDescription",function(){return g.MatExpansionPanelDescription}),n.d(e,"MatExpansionPanelTitle",function(){return g.MatExpansionPanelTitle}),n.d(e,"MatExpansionPanelContent",function(){return g.MatExpansionPanelContent}),n.d(e,"EXPANSION_PANEL_ANIMATION_TIMING",function(){return g.EXPANSION_PANEL_ANIMATION_TIMING}),n.d(e,"matExpansionAnimations",function(){return g.matExpansionAnimations});var y=n("./node_modules/@angular/material/esm5/form-field.es5.js");n.d(e,"MatFormFieldModule",function(){return y.MatFormFieldModule}),n.d(e,"MatError",function(){return y.MatError}),n.d(e,"MatFormFieldBase",function(){return y.MatFormFieldBase}),n.d(e,"_MatFormFieldMixinBase",function(){return y._MatFormFieldMixinBase}),n.d(e,"MAT_FORM_FIELD_DEFAULT_OPTIONS",function(){return y.MAT_FORM_FIELD_DEFAULT_OPTIONS}),n.d(e,"MatFormField",function(){return y.MatFormField}),n.d(e,"MatFormFieldControl",function(){return y.MatFormFieldControl}),n.d(e,"getMatFormFieldPlaceholderConflictError",function(){return y.getMatFormFieldPlaceholderConflictError}),n.d(e,"getMatFormFieldDuplicatedHintError",function(){return y.getMatFormFieldDuplicatedHintError}),n.d(e,"getMatFormFieldMissingControlError",function(){return y.getMatFormFieldMissingControlError}),n.d(e,"MatHint",function(){return y.MatHint}),n.d(e,"MatPlaceholder",function(){return y.MatPlaceholder}),n.d(e,"MatPrefix",function(){return y.MatPrefix}),n.d(e,"MatSuffix",function(){return y.MatSuffix}),n.d(e,"MatLabel",function(){return y.MatLabel}),n.d(e,"matFormFieldAnimations",function(){return y.matFormFieldAnimations});var _=n("./node_modules/@angular/material/esm5/grid-list.es5.js");n.d(e,"ɵa15",function(){return _["ɵa15"]}),n.d(e,"MatGridListModule",function(){return _.MatGridListModule}),n.d(e,"MatGridList",function(){return _.MatGridList}),n.d(e,"MatGridTile",function(){return _.MatGridTile}),n.d(e,"MatGridTileText",function(){return _.MatGridTileText}),n.d(e,"MatGridAvatarCssMatStyler",function(){return _.MatGridAvatarCssMatStyler}),n.d(e,"MatGridTileHeaderCssMatStyler",function(){return _.MatGridTileHeaderCssMatStyler}),n.d(e,"MatGridTileFooterCssMatStyler",function(){return _.MatGridTileFooterCssMatStyler});var v=n("./node_modules/@angular/material/esm5/icon.es5.js");n.d(e,"MatIconModule",function(){return v.MatIconModule}),n.d(e,"MAT_ICON_LOCATION_FACTORY",function(){return v.MAT_ICON_LOCATION_FACTORY}),n.d(e,"MatIconBase",function(){return v.MatIconBase}),n.d(e,"_MatIconMixinBase",function(){return v._MatIconMixinBase}),n.d(e,"MAT_ICON_LOCATION",function(){return v.MAT_ICON_LOCATION}),n.d(e,"MatIcon",function(){return v.MatIcon}),n.d(e,"getMatIconNameNotFoundError",function(){return v.getMatIconNameNotFoundError}),n.d(e,"getMatIconNoHttpProviderError",function(){return v.getMatIconNoHttpProviderError}),n.d(e,"getMatIconFailedToSanitizeUrlError",function(){return v.getMatIconFailedToSanitizeUrlError}),n.d(e,"getMatIconFailedToSanitizeLiteralError",function(){return v.getMatIconFailedToSanitizeLiteralError}),n.d(e,"ICON_REGISTRY_PROVIDER_FACTORY",function(){return v.ICON_REGISTRY_PROVIDER_FACTORY}),n.d(e,"MatIconRegistry",function(){return v.MatIconRegistry}),n.d(e,"ICON_REGISTRY_PROVIDER",function(){return v.ICON_REGISTRY_PROVIDER});var b=n("./node_modules/@angular/material/esm5/input.es5.js");n.d(e,"MatTextareaAutosize",function(){return b.MatTextareaAutosize}),n.d(e,"MatInputBase",function(){return b.MatInputBase}),n.d(e,"_MatInputMixinBase",function(){return b._MatInputMixinBase}),n.d(e,"MatInput",function(){return b.MatInput}),n.d(e,"getMatInputUnsupportedTypeError",function(){return b.getMatInputUnsupportedTypeError}),n.d(e,"MatInputModule",function(){return b.MatInputModule}),n.d(e,"MAT_INPUT_VALUE_ACCESSOR",function(){return b.MAT_INPUT_VALUE_ACCESSOR});var x=n("./node_modules/@angular/material/esm5/list.es5.js");n.d(e,"MatListModule",function(){return x.MatListModule}),n.d(e,"MatListBase",function(){return x.MatListBase}),n.d(e,"_MatListMixinBase",function(){return x._MatListMixinBase}),n.d(e,"MatListItemBase",function(){return x.MatListItemBase}),n.d(e,"_MatListItemMixinBase",function(){return x._MatListItemMixinBase}),n.d(e,"MatNavList",function(){return x.MatNavList}),n.d(e,"MatList",function(){return x.MatList}),n.d(e,"MatListAvatarCssMatStyler",function(){return x.MatListAvatarCssMatStyler}),n.d(e,"MatListIconCssMatStyler",function(){return x.MatListIconCssMatStyler}),n.d(e,"MatListSubheaderCssMatStyler",function(){return x.MatListSubheaderCssMatStyler}),n.d(e,"MatListItem",function(){return x.MatListItem}),n.d(e,"MatSelectionListBase",function(){return x.MatSelectionListBase}),n.d(e,"_MatSelectionListMixinBase",function(){return x._MatSelectionListMixinBase}),n.d(e,"MatListOptionBase",function(){return x.MatListOptionBase}),n.d(e,"_MatListOptionMixinBase",function(){return x._MatListOptionMixinBase}),n.d(e,"MAT_SELECTION_LIST_VALUE_ACCESSOR",function(){return x.MAT_SELECTION_LIST_VALUE_ACCESSOR}),n.d(e,"MatSelectionListChange",function(){return x.MatSelectionListChange}),n.d(e,"MatListOption",function(){return x.MatListOption}),n.d(e,"MatSelectionList",function(){return x.MatSelectionList});var w=n("./node_modules/@angular/material/esm5/menu.es5.js");n.d(e,"ɵa23",function(){return w["ɵa23"]}),n.d(e,"ɵb23",function(){return w["ɵb23"]}),n.d(e,"ɵc23",function(){return w["ɵc23"]}),n.d(e,"ɵf23",function(){return w["ɵf23"]}),n.d(e,"ɵd23",function(){return w["ɵd23"]}),n.d(e,"ɵe23",function(){return w["ɵe23"]}),n.d(e,"MAT_MENU_SCROLL_STRATEGY",function(){return w.MAT_MENU_SCROLL_STRATEGY}),n.d(e,"MatMenuModule",function(){return w.MatMenuModule}),n.d(e,"MatMenu",function(){return w.MatMenu}),n.d(e,"MAT_MENU_DEFAULT_OPTIONS",function(){return w.MAT_MENU_DEFAULT_OPTIONS}),n.d(e,"MatMenuItem",function(){return w.MatMenuItem}),n.d(e,"MatMenuTrigger",function(){return w.MatMenuTrigger}),n.d(e,"matMenuAnimations",function(){return w.matMenuAnimations}),n.d(e,"fadeInItems",function(){return w.fadeInItems}),n.d(e,"transformMenu",function(){return w.transformMenu}),n.d(e,"MatMenuContent",function(){return w.MatMenuContent});var j=n("./node_modules/@angular/material/esm5/paginator.es5.js");n.d(e,"MatPaginatorModule",function(){return j.MatPaginatorModule}),n.d(e,"PageEvent",function(){return j.PageEvent}),n.d(e,"MatPaginatorBase",function(){return j.MatPaginatorBase}),n.d(e,"_MatPaginatorBase",function(){return j._MatPaginatorBase}),n.d(e,"MatPaginator",function(){return j.MatPaginator}),n.d(e,"MAT_PAGINATOR_INTL_PROVIDER_FACTORY",function(){return j.MAT_PAGINATOR_INTL_PROVIDER_FACTORY}),n.d(e,"MatPaginatorIntl",function(){return j.MatPaginatorIntl}),n.d(e,"MAT_PAGINATOR_INTL_PROVIDER",function(){return j.MAT_PAGINATOR_INTL_PROVIDER});var O=n("./node_modules/@angular/material/esm5/progress-bar.es5.js");n.d(e,"MatProgressBarModule",function(){return O.MatProgressBarModule}),n.d(e,"MAT_PROGRESS_BAR_LOCATION_FACTORY",function(){return O.MAT_PROGRESS_BAR_LOCATION_FACTORY}),n.d(e,"MatProgressBarBase",function(){return O.MatProgressBarBase}),n.d(e,"_MatProgressBarMixinBase",function(){return O._MatProgressBarMixinBase}),n.d(e,"MAT_PROGRESS_BAR_LOCATION",function(){return O.MAT_PROGRESS_BAR_LOCATION}),n.d(e,"MatProgressBar",function(){return O.MatProgressBar});var C=n("./node_modules/@angular/material/esm5/progress-spinner.es5.js");n.d(e,"MatProgressSpinnerModule",function(){return C.MatProgressSpinnerModule}),n.d(e,"MAT_PROGRESS_SPINNER_DEFAULT_OPTIONS_FACTORY",function(){return C.MAT_PROGRESS_SPINNER_DEFAULT_OPTIONS_FACTORY}),n.d(e,"MatProgressSpinnerBase",function(){return C.MatProgressSpinnerBase}),n.d(e,"_MatProgressSpinnerMixinBase",function(){return C._MatProgressSpinnerMixinBase}),n.d(e,"MAT_PROGRESS_SPINNER_DEFAULT_OPTIONS",function(){return C.MAT_PROGRESS_SPINNER_DEFAULT_OPTIONS}),n.d(e,"MatProgressSpinner",function(){return C.MatProgressSpinner}),n.d(e,"MatSpinner",function(){return C.MatSpinner});var E=n("./node_modules/@angular/material/esm5/radio.es5.js");n.d(e,"MatRadioModule",function(){return E.MatRadioModule}),n.d(e,"MAT_RADIO_GROUP_CONTROL_VALUE_ACCESSOR",function(){return E.MAT_RADIO_GROUP_CONTROL_VALUE_ACCESSOR}),n.d(e,"MatRadioChange",function(){return E.MatRadioChange}),n.d(e,"MatRadioGroup",function(){return E.MatRadioGroup}),n.d(e,"MatRadioButtonBase",function(){return E.MatRadioButtonBase}),n.d(e,"_MatRadioButtonMixinBase",function(){return E._MatRadioButtonMixinBase}),n.d(e,"MatRadioButton",function(){return E.MatRadioButton});var S=n("./node_modules/@angular/material/esm5/select.es5.js");n.d(e,"MatSelectModule",function(){return S.MatSelectModule}),n.d(e,"MAT_SELECT_SCROLL_STRATEGY_PROVIDER_FACTORY",function(){return S.MAT_SELECT_SCROLL_STRATEGY_PROVIDER_FACTORY}),n.d(e,"SELECT_PANEL_MAX_HEIGHT",function(){return S.SELECT_PANEL_MAX_HEIGHT}),n.d(e,"SELECT_PANEL_PADDING_X",function(){return S.SELECT_PANEL_PADDING_X}),n.d(e,"SELECT_PANEL_INDENT_PADDING_X",function(){return S.SELECT_PANEL_INDENT_PADDING_X}),n.d(e,"SELECT_ITEM_HEIGHT_EM",function(){return S.SELECT_ITEM_HEIGHT_EM}),n.d(e,"SELECT_MULTIPLE_PANEL_PADDING_X",function(){return S.SELECT_MULTIPLE_PANEL_PADDING_X}),n.d(e,"SELECT_PANEL_VIEWPORT_PADDING",function(){return S.SELECT_PANEL_VIEWPORT_PADDING}),n.d(e,"MAT_SELECT_SCROLL_STRATEGY",function(){return S.MAT_SELECT_SCROLL_STRATEGY}),n.d(e,"MAT_SELECT_SCROLL_STRATEGY_PROVIDER",function(){return S.MAT_SELECT_SCROLL_STRATEGY_PROVIDER}),n.d(e,"MatSelectChange",function(){return S.MatSelectChange}),n.d(e,"MatSelectBase",function(){return S.MatSelectBase}),n.d(e,"_MatSelectMixinBase",function(){return S._MatSelectMixinBase}),n.d(e,"MatSelectTrigger",function(){return S.MatSelectTrigger}),n.d(e,"MatSelect",function(){return S.MatSelect}),n.d(e,"matSelectAnimations",function(){return S.matSelectAnimations}),n.d(e,"transformPanel",function(){return S.transformPanel}),n.d(e,"fadeInContent",function(){return S.fadeInContent});var M=n("./node_modules/@angular/material/esm5/sidenav.es5.js");n.d(e,"MatSidenavModule",function(){return M.MatSidenavModule}),n.d(e,"throwMatDuplicatedDrawerError",function(){return M.throwMatDuplicatedDrawerError}),n.d(e,"MAT_DRAWER_DEFAULT_AUTOSIZE_FACTORY",function(){return M.MAT_DRAWER_DEFAULT_AUTOSIZE_FACTORY}),n.d(e,"MAT_DRAWER_DEFAULT_AUTOSIZE",function(){return M.MAT_DRAWER_DEFAULT_AUTOSIZE}),n.d(e,"MatDrawerContent",function(){return M.MatDrawerContent}),n.d(e,"MatDrawer",function(){return M.MatDrawer}),n.d(e,"MatDrawerContainer",function(){return M.MatDrawerContainer}),n.d(e,"MatSidenavContent",function(){return M.MatSidenavContent}),n.d(e,"MatSidenav",function(){return M.MatSidenav}),n.d(e,"MatSidenavContainer",function(){return M.MatSidenavContainer}),n.d(e,"matDrawerAnimations",function(){return M.matDrawerAnimations});var T=n("./node_modules/@angular/material/esm5/slide-toggle.es5.js");n.d(e,"MatSlideToggleModule",function(){return T.MatSlideToggleModule}),n.d(e,"MAT_SLIDE_TOGGLE_VALUE_ACCESSOR",function(){return T.MAT_SLIDE_TOGGLE_VALUE_ACCESSOR}),n.d(e,"MatSlideToggleChange",function(){return T.MatSlideToggleChange}),n.d(e,"MatSlideToggleBase",function(){return T.MatSlideToggleBase}),n.d(e,"_MatSlideToggleMixinBase",function(){return T._MatSlideToggleMixinBase}),n.d(e,"MatSlideToggle",function(){return T.MatSlideToggle}),n.d(e,"MAT_SLIDE_TOGGLE_DEFAULT_OPTIONS",function(){return T.MAT_SLIDE_TOGGLE_DEFAULT_OPTIONS});var k=n("./node_modules/@angular/material/esm5/slider.es5.js");n.d(e,"MatSliderModule",function(){return k.MatSliderModule}),n.d(e,"MAT_SLIDER_VALUE_ACCESSOR",function(){return k.MAT_SLIDER_VALUE_ACCESSOR}),n.d(e,"MatSliderChange",function(){return k.MatSliderChange}),n.d(e,"MatSliderBase",function(){return k.MatSliderBase}),n.d(e,"_MatSliderMixinBase",function(){return k._MatSliderMixinBase}),n.d(e,"MatSlider",function(){return k.MatSlider});var A=n("./node_modules/@angular/material/esm5/snack-bar.es5.js");n.d(e,"MatSnackBarModule",function(){return A.MatSnackBarModule}),n.d(e,"MAT_SNACK_BAR_DEFAULT_OPTIONS_FACTORY",function(){return A.MAT_SNACK_BAR_DEFAULT_OPTIONS_FACTORY}),n.d(e,"MAT_SNACK_BAR_DEFAULT_OPTIONS",function(){return A.MAT_SNACK_BAR_DEFAULT_OPTIONS}),n.d(e,"MatSnackBar",function(){return A.MatSnackBar}),n.d(e,"MatSnackBarContainer",function(){return A.MatSnackBarContainer}),n.d(e,"MAT_SNACK_BAR_DATA",function(){return A.MAT_SNACK_BAR_DATA}),n.d(e,"MatSnackBarConfig",function(){return A.MatSnackBarConfig}),n.d(e,"MatSnackBarRef",function(){return A.MatSnackBarRef}),n.d(e,"SimpleSnackBar",function(){return A.SimpleSnackBar}),n.d(e,"matSnackBarAnimations",function(){return A.matSnackBarAnimations});var D=n("./node_modules/@angular/material/esm5/sort.es5.js");n.d(e,"MatSortModule",function(){return D.MatSortModule}),n.d(e,"MatSortHeaderBase",function(){return D.MatSortHeaderBase}),n.d(e,"_MatSortHeaderMixinBase",function(){return D._MatSortHeaderMixinBase}),n.d(e,"MatSortHeader",function(){return D.MatSortHeader}),n.d(e,"MAT_SORT_HEADER_INTL_PROVIDER_FACTORY",function(){return D.MAT_SORT_HEADER_INTL_PROVIDER_FACTORY}),n.d(e,"MatSortHeaderIntl",function(){return D.MatSortHeaderIntl}),n.d(e,"MAT_SORT_HEADER_INTL_PROVIDER",function(){return D.MAT_SORT_HEADER_INTL_PROVIDER}),n.d(e,"MatSortBase",function(){return D.MatSortBase}),n.d(e,"_MatSortMixinBase",function(){return D._MatSortMixinBase}),n.d(e,"MatSort",function(){return D.MatSort}),n.d(e,"matSortAnimations",function(){return D.matSortAnimations});var I=n("./node_modules/@angular/material/esm5/stepper.es5.js");n.d(e,"MatStepperModule",function(){return I.MatStepperModule}),n.d(e,"MatStepLabel",function(){return I.MatStepLabel}),n.d(e,"MatStep",function(){return I.MatStep}),n.d(e,"MatStepper",function(){return I.MatStepper}),n.d(e,"MatHorizontalStepper",function(){return I.MatHorizontalStepper}),n.d(e,"MatVerticalStepper",function(){return I.MatVerticalStepper}),n.d(e,"MatStepperNext",function(){return I.MatStepperNext}),n.d(e,"MatStepperPrevious",function(){return I.MatStepperPrevious}),n.d(e,"MatStepHeader",function(){return I.MatStepHeader}),n.d(e,"MAT_STEPPER_INTL_PROVIDER_FACTORY",function(){return I.MAT_STEPPER_INTL_PROVIDER_FACTORY}),n.d(e,"MatStepperIntl",function(){return I.MatStepperIntl}),n.d(e,"MAT_STEPPER_INTL_PROVIDER",function(){return I.MAT_STEPPER_INTL_PROVIDER}),n.d(e,"matStepperAnimations",function(){return I.matStepperAnimations}),n.d(e,"MatStepperIcon",function(){return I.MatStepperIcon});var P=n("./node_modules/@angular/material/esm5/table.es5.js");n.d(e,"MatTableModule",function(){return P.MatTableModule}),n.d(e,"MatCellDef",function(){return P.MatCellDef}),n.d(e,"MatHeaderCellDef",function(){return P.MatHeaderCellDef}),n.d(e,"MatFooterCellDef",function(){return P.MatFooterCellDef}),n.d(e,"MatColumnDef",function(){return P.MatColumnDef}),n.d(e,"MatHeaderCell",function(){return P.MatHeaderCell}),n.d(e,"MatFooterCell",function(){return P.MatFooterCell}),n.d(e,"MatCell",function(){return P.MatCell}),n.d(e,"MatTable",function(){return P.MatTable}),n.d(e,"MatHeaderRowDef",function(){return P.MatHeaderRowDef}),n.d(e,"MatFooterRowDef",function(){return P.MatFooterRowDef}),n.d(e,"MatRowDef",function(){return P.MatRowDef}),n.d(e,"MatHeaderRow",function(){return P.MatHeaderRow}),n.d(e,"MatFooterRow",function(){return P.MatFooterRow}),n.d(e,"MatRow",function(){return P.MatRow}),n.d(e,"MatTableDataSource",function(){return P.MatTableDataSource});var R=n("./node_modules/@angular/material/esm5/tabs.es5.js");n.d(e,"ɵa17",function(){return R["ɵa17"]}),n.d(e,"ɵf17",function(){return R["ɵf17"]}),n.d(e,"ɵg17",function(){return R["ɵg17"]}),n.d(e,"ɵb17",function(){return R["ɵb17"]}),n.d(e,"ɵc17",function(){return R["ɵc17"]}),n.d(e,"ɵd17",function(){return R["ɵd17"]}),n.d(e,"ɵe17",function(){return R["ɵe17"]}),n.d(e,"ɵj17",function(){return R["ɵj17"]}),n.d(e,"ɵh17",function(){return R["ɵh17"]}),n.d(e,"ɵk17",function(){return R["ɵk17"]}),n.d(e,"ɵi17",function(){return R["ɵi17"]}),n.d(e,"MatInkBar",function(){return R.MatInkBar}),n.d(e,"_MAT_INK_BAR_POSITIONER",function(){return R._MAT_INK_BAR_POSITIONER}),n.d(e,"MatTabBody",function(){return R.MatTabBody}),n.d(e,"MatTabBodyPortal",function(){return R.MatTabBodyPortal}),n.d(e,"MatTabHeader",function(){return R.MatTabHeader}),n.d(e,"MatTabLabelWrapper",function(){return R.MatTabLabelWrapper}),n.d(e,"MatTab",function(){return R.MatTab}),n.d(e,"MatTabLabel",function(){return R.MatTabLabel}),n.d(e,"MatTabNav",function(){return R.MatTabNav}),n.d(e,"MatTabLink",function(){return R.MatTabLink}),n.d(e,"MatTabContent",function(){return R.MatTabContent}),n.d(e,"MatTabsModule",function(){return R.MatTabsModule}),n.d(e,"MatTabChangeEvent",function(){return R.MatTabChangeEvent}),n.d(e,"MAT_TABS_CONFIG",function(){return R.MAT_TABS_CONFIG}),n.d(e,"MatTabGroupBase",function(){return R.MatTabGroupBase}),n.d(e,"_MatTabGroupMixinBase",function(){return R._MatTabGroupMixinBase}),n.d(e,"MatTabGroup",function(){return R.MatTabGroup}),n.d(e,"matTabsAnimations",function(){return R.matTabsAnimations});var N=n("./node_modules/@angular/material/esm5/toolbar.es5.js");n.d(e,"MatToolbarModule",function(){return N.MatToolbarModule}),n.d(e,"throwToolbarMixedModesError",function(){return N.throwToolbarMixedModesError}),n.d(e,"MatToolbarBase",function(){return N.MatToolbarBase}),n.d(e,"_MatToolbarMixinBase",function(){return N._MatToolbarMixinBase}),n.d(e,"MatToolbarRow",function(){return N.MatToolbarRow}),n.d(e,"MatToolbar",function(){return N.MatToolbar});var F=n("./node_modules/@angular/material/esm5/tooltip.es5.js");n.d(e,"MatTooltipModule",function(){return F.MatTooltipModule}),n.d(e,"getMatTooltipInvalidPositionError",function(){return F.getMatTooltipInvalidPositionError}),n.d(e,"MAT_TOOLTIP_SCROLL_STRATEGY_FACTORY",function(){return F.MAT_TOOLTIP_SCROLL_STRATEGY_FACTORY}),n.d(e,"MAT_TOOLTIP_DEFAULT_OPTIONS_FACTORY",function(){return F.MAT_TOOLTIP_DEFAULT_OPTIONS_FACTORY}),n.d(e,"SCROLL_THROTTLE_MS",function(){return F.SCROLL_THROTTLE_MS}),n.d(e,"TOOLTIP_PANEL_CLASS",function(){return F.TOOLTIP_PANEL_CLASS}),n.d(e,"MAT_TOOLTIP_SCROLL_STRATEGY",function(){return F.MAT_TOOLTIP_SCROLL_STRATEGY}),n.d(e,"MAT_TOOLTIP_SCROLL_STRATEGY_FACTORY_PROVIDER",function(){return F.MAT_TOOLTIP_SCROLL_STRATEGY_FACTORY_PROVIDER}),n.d(e,"MAT_TOOLTIP_DEFAULT_OPTIONS",function(){return F.MAT_TOOLTIP_DEFAULT_OPTIONS}),n.d(e,"MatTooltip",function(){return F.MatTooltip}),n.d(e,"TooltipComponent",function(){return F.TooltipComponent}),n.d(e,"matTooltipAnimations",function(){return F.matTooltipAnimations});var L=n("./node_modules/@angular/material/esm5/tree.es5.js");n.d(e,"_MatTreeNodeMixinBase",function(){return L._MatTreeNodeMixinBase}),n.d(e,"_MatNestedTreeNodeMixinBase",function(){return L._MatNestedTreeNodeMixinBase}),n.d(e,"MatTreeNode",function(){return L.MatTreeNode}),n.d(e,"MatTreeNodeDef",function(){return L.MatTreeNodeDef}),n.d(e,"MatNestedTreeNode",function(){return L.MatNestedTreeNode}),n.d(e,"MatTreeNodePadding",function(){return L.MatTreeNodePadding}),n.d(e,"MatTree",function(){return L.MatTree}),n.d(e,"MatTreeModule",function(){return L.MatTreeModule}),n.d(e,"MatTreeNodeToggle",function(){return L.MatTreeNodeToggle}),n.d(e,"MatTreeNodeOutlet",function(){return L.MatTreeNodeOutlet}),n.d(e,"MatTreeFlattener",function(){return L.MatTreeFlattener}),n.d(e,"MatTreeFlatDataSource",function(){return L.MatTreeFlatDataSource}),n.d(e,"MatTreeNestedDataSource",function(){return L.MatTreeNestedDataSource});var B=new r.Version("7.2.1")},"./node_modules/@angular/material/esm5/menu.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MAT_MENU_SCROLL_STRATEGY",function(){return M}),n.d(e,"MatMenuModule",function(){return I}),n.d(e,"MatMenu",function(){return S}),n.d(e,"MAT_MENU_DEFAULT_OPTIONS",function(){return C}),n.d(e,"MatMenuItem",function(){return O}),n.d(e,"MatMenuTrigger",function(){return D}),n.d(e,"matMenuAnimations",function(){return _}),n.d(e,"fadeInItems",function(){return v}),n.d(e,"transformMenu",function(){return b}),n.d(e,"MatMenuContent",function(){return y}),n.d(e,"ɵa23",function(){return E}),n.d(e,"ɵb23",function(){return w}),n.d(e,"ɵc23",function(){return j}),n.d(e,"ɵf23",function(){return x}),n.d(e,"ɵd23",function(){return T}),n.d(e,"ɵe23",function(){return k});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),o=n("./node_modules/@angular/common/fesm5/common.js"),a=n("./node_modules/rxjs/_esm5/index.js"),s=n("./node_modules/@angular/animations/fesm5/animations.js"),u=n("./node_modules/tslib/tslib.es6.js"),c=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),l=n("./node_modules/@angular/material/esm5/core.es5.js"),d=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),p=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),f=n("./node_modules/rxjs/_esm5/operators/index.js"),h=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),m=n("./node_modules/@angular/cdk/esm5/overlay.es5.js"),g=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),y=function(){function t(t,e,n,r,i,o){this._template=t,this._componentFactoryResolver=e,this._appRef=n,this._injector=r,this._viewContainerRef=i,this._document=o,this._attached=new a.Subject}return t.prototype.attach=function(t){void 0===t&&(t={}),this._portal||(this._portal=new i.TemplatePortal(this._template,this._viewContainerRef)),this.detach(),this._outlet||(this._outlet=new i.DomPortalOutlet(this._document.createElement("div"),this._componentFactoryResolver,this._appRef,this._injector));var e=this._template.elementRef.nativeElement;e.parentNode.insertBefore(this._outlet.outletElement,e),this._portal.attach(this._outlet,t),this._attached.next()},t.prototype.detach=function(){this._portal.isAttached&&this._portal.detach()},t.prototype.ngOnDestroy=function(){this._outlet&&this._outlet.dispose()},t.decorators=[{type:r.Directive,args:[{selector:"ng-template[matMenuContent]"}]}],t.ctorParameters=function(){return[{type:r.TemplateRef},{type:r.ComponentFactoryResolver},{type:r.ApplicationRef},{type:r.Injector},{type:r.ViewContainerRef},{type:void 0,decorators:[{type:r.Inject,args:[o.DOCUMENT]}]}]},t}(),_={transformMenu:Object(s.trigger)("transformMenu",[Object(s.state)("void",Object(s.style)({opacity:0,transform:"scale(0.8)"})),Object(s.transition)("void => enter",Object(s.group)([Object(s.query)(".mat-menu-content",Object(s.animate)("100ms linear",Object(s.style)({opacity:1}))),Object(s.animate)("120ms cubic-bezier(0, 0, 0.2, 1)",Object(s.style)({transform:"scale(1)"}))])),Object(s.transition)("* => void",Object(s.animate)("100ms 25ms linear",Object(s.style)({opacity:0})))]),fadeInItems:Object(s.trigger)("fadeInItems",[Object(s.state)("showing",Object(s.style)({opacity:1})),Object(s.transition)("void => *",[Object(s.style)({opacity:0}),Object(s.animate)("400ms 100ms cubic-bezier(0.55, 0, 0.55, 0.2)")])])},v=_.fadeInItems,b=_.transformMenu;var x=new r.InjectionToken("MAT_MENU_PANEL"),w=function(){return function(){}}(),j=Object(l.mixinDisableRipple)(Object(l.mixinDisabled)(w)),O=function(t){function e(e,n,r,i){var o=t.call(this)||this;return o._elementRef=e,o._focusMonitor=r,o._parentMenu=i,o.role="menuitem",o._hovered=new a.Subject,o._highlighted=!1,o._triggersSubmenu=!1,r&&r.monitor(o._elementRef,!1),i&&i.addItem&&i.addItem(o),o._document=n,o}return Object(u.__extends)(e,t),e.prototype.focus=function(t){void 0===t&&(t="program"),this._focusMonitor?this._focusMonitor.focusVia(this._getHostElement(),t):this._getHostElement().focus()},e.prototype.ngOnDestroy=function(){this._focusMonitor&&this._focusMonitor.stopMonitoring(this._elementRef),this._parentMenu&&this._parentMenu.removeItem&&this._parentMenu.removeItem(this),this._hovered.complete()},e.prototype._getTabIndex=function(){return this.disabled?"-1":"0"},e.prototype._getHostElement=function(){return this._elementRef.nativeElement},e.prototype._checkDisabled=function(t){this.disabled&&(t.preventDefault(),t.stopPropagation())},e.prototype._handleMouseEnter=function(){this._hovered.next(this)},e.prototype.getLabel=function(){var t=this._elementRef.nativeElement,e=this._document?this._document.TEXT_NODE:3,n="";if(t.childNodes)for(var r=t.childNodes.length,i=0;i<r;i++)t.childNodes[i].nodeType===e&&(n+=t.childNodes[i].textContent);return n.trim()},e.decorators=[{type:r.Component,args:[{selector:"[mat-menu-item]",exportAs:"matMenuItem",inputs:["disabled","disableRipple"],host:{"[attr.role]":"role",class:"mat-menu-item","[class.mat-menu-item-highlighted]":"_highlighted","[class.mat-menu-item-submenu-trigger]":"_triggersSubmenu","[attr.tabindex]":"_getTabIndex()","[attr.aria-disabled]":"disabled.toString()","[attr.disabled]":"disabled || null","(click)":"_checkDisabled($event)","(mouseenter)":"_handleMouseEnter()"},changeDetection:r.ChangeDetectionStrategy.OnPush,encapsulation:r.ViewEncapsulation.None,template:'<ng-content></ng-content><div class="mat-menu-ripple" matRipple [matRippleDisabled]="disableRipple || disabled" [matRippleTrigger]="_getHostElement()"></div>'}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:void 0,decorators:[{type:r.Inject,args:[o.DOCUMENT]}]},{type:c.FocusMonitor},{type:void 0,decorators:[{type:r.Inject,args:[x]},{type:r.Optional}]}]},e.propDecorators={role:[{type:r.Input}]},e}(j),C=new r.InjectionToken("mat-menu-default-options",{providedIn:"root",factory:E});function E(){return{overlapTrigger:!1,xPosition:"after",yPosition:"below",backdropClass:"cdk-overlay-transparent-backdrop"}}var S=function(){function t(t,e,n){this._elementRef=t,this._ngZone=e,this._defaultOptions=n,this._xPosition=this._defaultOptions.xPosition,this._yPosition=this._defaultOptions.yPosition,this._items=[],this._itemChanges=new a.Subject,this._tabSubscription=a.Subscription.EMPTY,this._classList={},this._panelAnimationState="void",this._animationDone=new a.Subject,this.backdropClass=this._defaultOptions.backdropClass,this._overlapTrigger=this._defaultOptions.overlapTrigger,this._hasBackdrop=this._defaultOptions.hasBackdrop,this.closed=new r.EventEmitter,this.close=this.closed}return Object.defineProperty(t.prototype,"xPosition",{get:function(){return this._xPosition},set:function(t){"before"!==t&&"after"!==t&&function(){throw Error('xPosition value must be either \'before\' or after\'.\n      Example: <mat-menu xPosition="before" #menu="matMenu"></mat-menu>')}(),this._xPosition=t,this.setPositionClasses()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"yPosition",{get:function(){return this._yPosition},set:function(t){"above"!==t&&"below"!==t&&function(){throw Error('yPosition value must be either \'above\' or below\'.\n      Example: <mat-menu yPosition="above" #menu="matMenu"></mat-menu>')}(),this._yPosition=t,this.setPositionClasses()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"overlapTrigger",{get:function(){return this._overlapTrigger},set:function(t){this._overlapTrigger=Object(d.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"hasBackdrop",{get:function(){return this._hasBackdrop},set:function(t){this._hasBackdrop=Object(d.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"panelClass",{set:function(t){t&&t.length&&(this._classList=t.split(" ").reduce(function(t,e){return t[e]=!0,t},{}),this._elementRef.nativeElement.className="")},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"classList",{get:function(){return this.panelClass},set:function(t){this.panelClass=t},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){this.setPositionClasses()},t.prototype.ngAfterContentInit=function(){var t=this;this._keyManager=new c.FocusKeyManager(this._items).withWrap().withTypeAhead(),this._tabSubscription=this._keyManager.tabOut.subscribe(function(){return t.closed.emit("tab")})},t.prototype.ngOnDestroy=function(){this._tabSubscription.unsubscribe(),this.closed.complete()},t.prototype._hovered=function(){return this._itemChanges.pipe(Object(f.startWith)(this._items),Object(f.switchMap)(function(t){return a.merge.apply(void 0,t.map(function(t){return t._hovered}))}))},t.prototype._handleKeydown=function(t){var e=t.keyCode;switch(e){case p.ESCAPE:this.closed.emit("keydown");break;case p.LEFT_ARROW:this.parentMenu&&"ltr"===this.direction&&this.closed.emit("keydown");break;case p.RIGHT_ARROW:this.parentMenu&&"rtl"===this.direction&&this.closed.emit("keydown");break;default:e!==p.UP_ARROW&&e!==p.DOWN_ARROW||this._keyManager.setFocusOrigin("keyboard"),this._keyManager.onKeydown(t)}},t.prototype.focusFirstItem=function(t){var e=this;void 0===t&&(t="program"),this.lazyContent?this._ngZone.onStable.asObservable().pipe(Object(f.take)(1)).subscribe(function(){return e._keyManager.setFocusOrigin(t).setFirstItemActive()}):this._keyManager.setFocusOrigin(t).setFirstItemActive()},t.prototype.resetActiveItem=function(){this._keyManager.setActiveItem(-1)},t.prototype.setElevation=function(t){var e="mat-elevation-z"+(4+t),n=Object.keys(this._classList).find(function(t){return t.startsWith("mat-elevation-z")});n&&n!==this._previousElevation||(this._previousElevation&&(this._classList[this._previousElevation]=!1),this._classList[e]=!0,this._previousElevation=e)},t.prototype.addItem=function(t){-1===this._items.indexOf(t)&&(this._items.push(t),this._itemChanges.next(this._items))},t.prototype.removeItem=function(t){var e=this._items.indexOf(t);this._items.indexOf(t)>-1&&(this._items.splice(e,1),this._itemChanges.next(this._items))},t.prototype.setPositionClasses=function(t,e){void 0===t&&(t=this.xPosition),void 0===e&&(e=this.yPosition);var n=this._classList;n["mat-menu-before"]="before"===t,n["mat-menu-after"]="after"===t,n["mat-menu-above"]="above"===e,n["mat-menu-below"]="below"===e},t.prototype._startAnimation=function(){this._panelAnimationState="enter"},t.prototype._resetAnimation=function(){this._panelAnimationState="void"},t.prototype._onAnimationDone=function(t){this._animationDone.next(t),this._isAnimating=!1},t.prototype._onAnimationStart=function(t){this._isAnimating=!0,"enter"===t.toState&&0===this._keyManager.activeItemIndex&&(t.element.scrollTop=0)},t.decorators=[{type:r.Component,args:[{selector:"mat-menu",template:'<ng-template><div class="mat-menu-panel" [ngClass]="_classList" (keydown)="_handleKeydown($event)" (click)="closed.emit(\'click\')" [@transformMenu]="_panelAnimationState" (@transformMenu.start)="_onAnimationStart($event)" (@transformMenu.done)="_onAnimationDone($event)" tabindex="-1" role="menu"><div class="mat-menu-content"><ng-content></ng-content></div></div></ng-template>',styles:[".mat-menu-panel{min-width:112px;max-width:280px;overflow:auto;-webkit-overflow-scrolling:touch;max-height:calc(100vh - 48px);border-radius:4px;outline:0}.mat-menu-panel.ng-animating{pointer-events:none}@media (-ms-high-contrast:active){.mat-menu-panel{outline:solid 1px}}.mat-menu-content:not(:empty){padding-top:8px;padding-bottom:8px}.mat-menu-item{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;white-space:nowrap;overflow:hidden;text-overflow:ellipsis;display:block;line-height:48px;height:48px;padding:0 16px;text-align:left;text-decoration:none;max-width:100%;position:relative}.mat-menu-item::-moz-focus-inner{border:0}.mat-menu-item[disabled]{cursor:default}[dir=rtl] .mat-menu-item{text-align:right}.mat-menu-item .mat-icon{margin-right:16px;vertical-align:middle}.mat-menu-item .mat-icon svg{vertical-align:top}[dir=rtl] .mat-menu-item .mat-icon{margin-left:16px;margin-right:0}@media (-ms-high-contrast:active){.mat-menu-item-highlighted,.mat-menu-item.cdk-keyboard-focused,.mat-menu-item.cdk-program-focused{outline:dotted 1px}}.mat-menu-item-submenu-trigger{padding-right:32px}.mat-menu-item-submenu-trigger::after{width:0;height:0;border-style:solid;border-width:5px 0 5px 5px;border-color:transparent transparent transparent currentColor;content:'';display:inline-block;position:absolute;top:50%;right:16px;transform:translateY(-50%)}[dir=rtl] .mat-menu-item-submenu-trigger{padding-right:16px;padding-left:32px}[dir=rtl] .mat-menu-item-submenu-trigger::after{right:auto;left:16px;transform:rotateY(180deg) translateY(-50%)}button.mat-menu-item{width:100%}.mat-menu-item .mat-menu-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}"],changeDetection:r.ChangeDetectionStrategy.OnPush,encapsulation:r.ViewEncapsulation.None,exportAs:"matMenu",animations:[_.transformMenu,_.fadeInItems],providers:[{provide:x,useExisting:t}]}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:r.NgZone},{type:void 0,decorators:[{type:r.Inject,args:[C]}]}]},t.propDecorators={backdropClass:[{type:r.Input}],xPosition:[{type:r.Input}],yPosition:[{type:r.Input}],templateRef:[{type:r.ViewChild,args:[r.TemplateRef]}],items:[{type:r.ContentChildren,args:[O]}],lazyContent:[{type:r.ContentChild,args:[y]}],overlapTrigger:[{type:r.Input}],hasBackdrop:[{type:r.Input}],panelClass:[{type:r.Input,args:["class"]}],classList:[{type:r.Input}],closed:[{type:r.Output}],close:[{type:r.Output}]},t}(),M=new r.InjectionToken("mat-menu-scroll-strategy");function T(t){return function(){return t.scrollStrategies.reposition()}}var k={provide:M,deps:[m.Overlay],useFactory:T},A=Object(g.normalizePassiveListenerOptions)({passive:!0}),D=function(){function t(t,e,n,i,o,s,u,c){var l=this;this._overlay=t,this._element=e,this._viewContainerRef=n,this._parentMenu=o,this._menuItemInstance=s,this._dir=u,this._focusMonitor=c,this._overlayRef=null,this._menuOpen=!1,this._closeSubscription=a.Subscription.EMPTY,this._hoverSubscription=a.Subscription.EMPTY,this._menuCloseSubscription=a.Subscription.EMPTY,this._handleTouchStart=function(){return l._openedBy="touch"},this._openedBy=null,this.menuOpened=new r.EventEmitter,this.onMenuOpen=this.menuOpened,this.menuClosed=new r.EventEmitter,this.onMenuClose=this.menuClosed,e.nativeElement.addEventListener("touchstart",this._handleTouchStart,A),s&&(s._triggersSubmenu=this.triggersSubmenu()),this._scrollStrategy=i}return Object.defineProperty(t.prototype,"_deprecatedMatMenuTriggerFor",{get:function(){return this.menu},set:function(t){this.menu=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"menu",{get:function(){return this._menu},set:function(t){var e=this;t!==this._menu&&(this._menu=t,this._menuCloseSubscription.unsubscribe(),t&&(this._menuCloseSubscription=t.close.asObservable().subscribe(function(t){e._destroyMenu(),"click"!==t&&"tab"!==t||!e._parentMenu||e._parentMenu.closed.emit(t)})))},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this._checkMenu(),this._handleHover()},t.prototype.ngOnDestroy=function(){this._overlayRef&&(this._overlayRef.dispose(),this._overlayRef=null),this._element.nativeElement.removeEventListener("touchstart",this._handleTouchStart,A),this._cleanUpSubscriptions()},Object.defineProperty(t.prototype,"menuOpen",{get:function(){return this._menuOpen},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"dir",{get:function(){return this._dir&&"rtl"===this._dir.value?"rtl":"ltr"},enumerable:!0,configurable:!0}),t.prototype.triggersSubmenu=function(){return!(!this._menuItemInstance||!this._parentMenu)},t.prototype.toggleMenu=function(){return this._menuOpen?this.closeMenu():this.openMenu()},t.prototype.openMenu=function(){var t=this;if(!this._menuOpen){this._checkMenu();var e=this._createOverlay(),n=e.getConfig();this._setPosition(n.positionStrategy),n.hasBackdrop=null==this.menu.hasBackdrop?!this.triggersSubmenu():this.menu.hasBackdrop,e.attach(this._getPortal()),this.menu.lazyContent&&this.menu.lazyContent.attach(this.menuData),this._closeSubscription=this._menuClosingActions().subscribe(function(){return t.closeMenu()}),this._initMenu(),this.menu instanceof S&&this.menu._startAnimation()}},t.prototype.closeMenu=function(){this.menu.close.emit()},t.prototype.focus=function(t){void 0===t&&(t="program"),this._focusMonitor?this._focusMonitor.focusVia(this._element,t):this._element.nativeElement.focus()},t.prototype._destroyMenu=function(){var t=this;if(this._overlayRef&&this.menuOpen){var e=this.menu;this._closeSubscription.unsubscribe(),this._overlayRef.detach(),e instanceof S?(e._resetAnimation(),e.lazyContent?e._animationDone.pipe(Object(f.filter)(function(t){return"void"===t.toState}),Object(f.take)(1),Object(f.takeUntil)(e.lazyContent._attached)).subscribe(function(){return e.lazyContent.detach()},void 0,function(){t._resetMenu()}):this._resetMenu()):(this._resetMenu(),e.lazyContent&&e.lazyContent.detach())}},t.prototype._initMenu=function(){this.menu.parentMenu=this.triggersSubmenu()?this._parentMenu:void 0,this.menu.direction=this.dir,this._setMenuElevation(),this._setIsMenuOpen(!0),this.menu.focusFirstItem(this._openedBy||"program")},t.prototype._setMenuElevation=function(){if(this.menu.setElevation){for(var t=0,e=this.menu.parentMenu;e;)t++,e=e.parentMenu;this.menu.setElevation(t)}},t.prototype._resetMenu=function(){this._setIsMenuOpen(!1),this._openedBy?this.triggersSubmenu()||this.focus(this._openedBy):this.focus(),this._openedBy=null},t.prototype._setIsMenuOpen=function(t){this._menuOpen=t,this._menuOpen?this.menuOpened.emit():this.menuClosed.emit(),this.triggersSubmenu()&&(this._menuItemInstance._highlighted=t)},t.prototype._checkMenu=function(){this.menu||function(){throw Error('matMenuTriggerFor: must pass in an mat-menu instance.\n\n    Example:\n      <mat-menu #menu="matMenu"></mat-menu>\n      <button [matMenuTriggerFor]="menu"></button>')}()},t.prototype._createOverlay=function(){if(!this._overlayRef){var t=this._getOverlayConfig();this._subscribeToPositions(t.positionStrategy),this._overlayRef=this._overlay.create(t),this._overlayRef.keydownEvents().subscribe()}return this._overlayRef},t.prototype._getOverlayConfig=function(){return new m.OverlayConfig({positionStrategy:this._overlay.position().flexibleConnectedTo(this._element).withLockedPosition().withTransformOriginOn(".mat-menu-panel"),backdropClass:this.menu.backdropClass||"cdk-overlay-transparent-backdrop",scrollStrategy:this._scrollStrategy(),direction:this._dir})},t.prototype._subscribeToPositions=function(t){var e=this;this.menu.setPositionClasses&&t.positionChanges.subscribe(function(t){var n="start"===t.connectionPair.overlayX?"after":"before",r="top"===t.connectionPair.overlayY?"below":"above";e.menu.setPositionClasses(n,r)})},t.prototype._setPosition=function(t){var e="before"===this.menu.xPosition?["end","start"]:["start","end"],n=e[0],r=e[1],i="above"===this.menu.yPosition?["bottom","top"]:["top","bottom"],o=i[0],a=i[1],s=[o,a],u=s[0],c=s[1],l=[n,r],d=l[0],p=l[1],f=0;this.triggersSubmenu()?(p=n="before"===this.menu.xPosition?"start":"end",r=d="end"===n?"start":"end",f="bottom"===o?8:-8):this.menu.overlapTrigger||(u="top"===o?"bottom":"top",c="top"===a?"bottom":"top"),t.withPositions([{originX:n,originY:u,overlayX:d,overlayY:o,offsetY:f},{originX:r,originY:u,overlayX:p,overlayY:o,offsetY:f},{originX:n,originY:c,overlayX:d,overlayY:a,offsetY:-f},{originX:r,originY:c,overlayX:p,overlayY:a,offsetY:-f}])},t.prototype._cleanUpSubscriptions=function(){this._closeSubscription.unsubscribe(),this._hoverSubscription.unsubscribe()},t.prototype._menuClosingActions=function(){var t=this,e=this._overlayRef.backdropClick(),n=this._overlayRef.detachments(),r=this._parentMenu?this._parentMenu.closed:Object(a.of)(),i=this._parentMenu?this._parentMenu._hovered().pipe(Object(f.filter)(function(e){return e!==t._menuItemInstance}),Object(f.filter)(function(){return t._menuOpen})):Object(a.of)();return Object(a.merge)(e,r,i,n)},t.prototype._handleMousedown=function(t){Object(c.isFakeMousedownFromScreenReader)(t)||(this._openedBy=0===t.button?"mouse":null,this.triggersSubmenu()&&t.preventDefault())},t.prototype._handleKeydown=function(t){var e=t.keyCode;this.triggersSubmenu()&&(e===p.RIGHT_ARROW&&"ltr"===this.dir||e===p.LEFT_ARROW&&"rtl"===this.dir)&&this.openMenu()},t.prototype._handleClick=function(t){this.triggersSubmenu()?(t.stopPropagation(),this.openMenu()):this.toggleMenu()},t.prototype._handleHover=function(){var t=this;this.triggersSubmenu()&&(this._hoverSubscription=this._parentMenu._hovered().pipe(Object(f.filter)(function(e){return e===t._menuItemInstance&&!e.disabled}),Object(f.delay)(0,a.asapScheduler)).subscribe(function(){t._openedBy="mouse",t.menu instanceof S&&t.menu._isAnimating?t.menu._animationDone.pipe(Object(f.take)(1),Object(f.delay)(0,a.asapScheduler),Object(f.takeUntil)(t._parentMenu._hovered())).subscribe(function(){return t.openMenu()}):t.openMenu()}))},t.prototype._getPortal=function(){return this._portal&&this._portal.templateRef===this.menu.templateRef||(this._portal=new i.TemplatePortal(this.menu.templateRef,this._viewContainerRef)),this._portal},t.decorators=[{type:r.Directive,args:[{selector:"[mat-menu-trigger-for], [matMenuTriggerFor]",host:{"aria-haspopup":"true","[attr.aria-expanded]":"menuOpen || null","(mousedown)":"_handleMousedown($event)","(keydown)":"_handleKeydown($event)","(click)":"_handleClick($event)"},exportAs:"matMenuTrigger"}]}],t.ctorParameters=function(){return[{type:m.Overlay},{type:r.ElementRef},{type:r.ViewContainerRef},{type:void 0,decorators:[{type:r.Inject,args:[M]}]},{type:S,decorators:[{type:r.Optional}]},{type:O,decorators:[{type:r.Optional},{type:r.Self}]},{type:h.Directionality,decorators:[{type:r.Optional}]},{type:c.FocusMonitor}]},t.propDecorators={_deprecatedMatMenuTriggerFor:[{type:r.Input,args:["mat-menu-trigger-for"]}],menu:[{type:r.Input,args:["matMenuTriggerFor"]}],menuData:[{type:r.Input,args:["matMenuTriggerData"]}],menuOpened:[{type:r.Output}],onMenuOpen:[{type:r.Output}],menuClosed:[{type:r.Output}],onMenuClose:[{type:r.Output}]},t}(),I=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[o.CommonModule,l.MatCommonModule,l.MatRippleModule,m.OverlayModule],exports:[S,O,D,y,l.MatCommonModule],declarations:[S,O,D,y],providers:[k]}]}],t}()},"./node_modules/@angular/material/esm5/paginator.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatPaginatorModule",function(){return v}),n.d(e,"PageEvent",function(){return m}),n.d(e,"MatPaginatorBase",function(){return g}),n.d(e,"_MatPaginatorBase",function(){return y}),n.d(e,"MatPaginator",function(){return _}),n.d(e,"MAT_PAGINATOR_INTL_PROVIDER_FACTORY",function(){return f}),n.d(e,"MatPaginatorIntl",function(){return p}),n.d(e,"MAT_PAGINATOR_INTL_PROVIDER",function(){return h});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/rxjs/_esm5/index.js"),o=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),s=n("./node_modules/@angular/material/esm5/core.es5.js"),u=n("./node_modules/@angular/common/fesm5/common.js"),c=n("./node_modules/@angular/material/esm5/button.es5.js"),l=n("./node_modules/@angular/material/esm5/select.es5.js"),d=n("./node_modules/@angular/material/esm5/tooltip.es5.js"),p=function(){function t(){this.changes=new i.Subject,this.itemsPerPageLabel="Items per page:",this.nextPageLabel="Next page",this.previousPageLabel="Previous page",this.firstPageLabel="First page",this.lastPageLabel="Last page",this.getRangeLabel=function(t,e,n){if(0==n||0==e)return"0 of "+n;var r=t*e;return r+1+" - "+(r<(n=Math.max(n,0))?Math.min(r+e,n):r+e)+" of "+n}}return t.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}();function f(t){return t||new p}var h={provide:p,deps:[[new r.Optional,new r.SkipSelf,p]],useFactory:f},m=function(){return function(){}}(),g=function(){return function(){}}(),y=Object(s.mixinDisabled)(Object(s.mixinInitialized)(g)),_=function(t){function e(e,n){var i=t.call(this)||this;return i._intl=e,i._changeDetectorRef=n,i._pageIndex=0,i._length=0,i._pageSizeOptions=[],i._hidePageSize=!1,i._showFirstLastButtons=!1,i.page=new r.EventEmitter,i._intlChanges=e.changes.subscribe(function(){return i._changeDetectorRef.markForCheck()}),i}return Object(o.__extends)(e,t),Object.defineProperty(e.prototype,"pageIndex",{get:function(){return this._pageIndex},set:function(t){this._pageIndex=Math.max(Object(a.coerceNumberProperty)(t),0),this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"length",{get:function(){return this._length},set:function(t){this._length=Object(a.coerceNumberProperty)(t),this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"pageSize",{get:function(){return this._pageSize},set:function(t){this._pageSize=Math.max(Object(a.coerceNumberProperty)(t),0),this._updateDisplayedPageSizeOptions()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"pageSizeOptions",{get:function(){return this._pageSizeOptions},set:function(t){this._pageSizeOptions=(t||[]).map(function(t){return Object(a.coerceNumberProperty)(t)}),this._updateDisplayedPageSizeOptions()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hidePageSize",{get:function(){return this._hidePageSize},set:function(t){this._hidePageSize=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"showFirstLastButtons",{get:function(){return this._showFirstLastButtons},set:function(t){this._showFirstLastButtons=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){this._initialized=!0,this._updateDisplayedPageSizeOptions(),this._markInitialized()},e.prototype.ngOnDestroy=function(){this._intlChanges.unsubscribe()},e.prototype.nextPage=function(){if(this.hasNextPage()){var t=this.pageIndex;this.pageIndex++,this._emitPageEvent(t)}},e.prototype.previousPage=function(){if(this.hasPreviousPage()){var t=this.pageIndex;this.pageIndex--,this._emitPageEvent(t)}},e.prototype.firstPage=function(){if(this.hasPreviousPage()){var t=this.pageIndex;this.pageIndex=0,this._emitPageEvent(t)}},e.prototype.lastPage=function(){if(this.hasNextPage()){var t=this.pageIndex;this.pageIndex=this.getNumberOfPages()-1,this._emitPageEvent(t)}},e.prototype.hasPreviousPage=function(){return this.pageIndex>=1&&0!=this.pageSize},e.prototype.hasNextPage=function(){var t=this.getNumberOfPages()-1;return this.pageIndex<t&&0!=this.pageSize},e.prototype.getNumberOfPages=function(){return this.pageSize?Math.ceil(this.length/this.pageSize):0},e.prototype._changePageSize=function(t){var e=this.pageIndex*this.pageSize,n=this.pageIndex;this.pageIndex=Math.floor(e/t)||0,this.pageSize=t,this._emitPageEvent(n)},e.prototype._nextButtonsDisabled=function(){return this.disabled||!this.hasNextPage()},e.prototype._previousButtonsDisabled=function(){return this.disabled||!this.hasPreviousPage()},e.prototype._updateDisplayedPageSizeOptions=function(){this._initialized&&(this.pageSize||(this._pageSize=0!=this.pageSizeOptions.length?this.pageSizeOptions[0]:50),this._displayedPageSizeOptions=this.pageSizeOptions.slice(),-1===this._displayedPageSizeOptions.indexOf(this.pageSize)&&this._displayedPageSizeOptions.push(this.pageSize),this._displayedPageSizeOptions.sort(function(t,e){return t-e}),this._changeDetectorRef.markForCheck())},e.prototype._emitPageEvent=function(t){this.page.emit({previousPageIndex:t,pageIndex:this.pageIndex,pageSize:this.pageSize,length:this.length})},e.decorators=[{type:r.Component,args:[{selector:"mat-paginator",exportAs:"matPaginator",template:'<div class="mat-paginator-outer-container"><div class="mat-paginator-container"><div class="mat-paginator-page-size" *ngIf="!hidePageSize"><div class="mat-paginator-page-size-label">{{_intl.itemsPerPageLabel}}</div><mat-form-field *ngIf="_displayedPageSizeOptions.length > 1" [color]="color" class="mat-paginator-page-size-select"><mat-select [value]="pageSize" [disabled]="disabled" [aria-label]="_intl.itemsPerPageLabel" (selectionChange)="_changePageSize($event.value)"><mat-option *ngFor="let pageSizeOption of _displayedPageSizeOptions" [value]="pageSizeOption">{{pageSizeOption}}</mat-option></mat-select></mat-form-field><div *ngIf="_displayedPageSizeOptions.length <= 1">{{pageSize}}</div></div><div class="mat-paginator-range-actions"><div class="mat-paginator-range-label">{{_intl.getRangeLabel(pageIndex, pageSize, length)}}</div><button mat-icon-button type="button" class="mat-paginator-navigation-first" (click)="firstPage()" [attr.aria-label]="_intl.firstPageLabel" [matTooltip]="_intl.firstPageLabel" [matTooltipDisabled]="_previousButtonsDisabled()" [matTooltipPosition]="\'above\'" [disabled]="_previousButtonsDisabled()" *ngIf="showFirstLastButtons"><svg class="mat-paginator-icon" viewBox="0 0 24 24" focusable="false"><path d="M18.41 16.59L13.82 12l4.59-4.59L17 6l-6 6 6 6zM6 6h2v12H6z"/></svg></button> <button mat-icon-button type="button" class="mat-paginator-navigation-previous" (click)="previousPage()" [attr.aria-label]="_intl.previousPageLabel" [matTooltip]="_intl.previousPageLabel" [matTooltipDisabled]="_previousButtonsDisabled()" [matTooltipPosition]="\'above\'" [disabled]="_previousButtonsDisabled()"><svg class="mat-paginator-icon" viewBox="0 0 24 24" focusable="false"><path d="M15.41 7.41L14 6l-6 6 6 6 1.41-1.41L10.83 12z"/></svg></button> <button mat-icon-button type="button" class="mat-paginator-navigation-next" (click)="nextPage()" [attr.aria-label]="_intl.nextPageLabel" [matTooltip]="_intl.nextPageLabel" [matTooltipDisabled]="_nextButtonsDisabled()" [matTooltipPosition]="\'above\'" [disabled]="_nextButtonsDisabled()"><svg class="mat-paginator-icon" viewBox="0 0 24 24" focusable="false"><path d="M10 6L8.59 7.41 13.17 12l-4.58 4.59L10 18l6-6z"/></svg></button> <button mat-icon-button type="button" class="mat-paginator-navigation-last" (click)="lastPage()" [attr.aria-label]="_intl.lastPageLabel" [matTooltip]="_intl.lastPageLabel" [matTooltipDisabled]="_nextButtonsDisabled()" [matTooltipPosition]="\'above\'" [disabled]="_nextButtonsDisabled()" *ngIf="showFirstLastButtons"><svg class="mat-paginator-icon" viewBox="0 0 24 24" focusable="false"><path d="M5.59 7.41L10.18 12l-4.59 4.59L7 18l6-6-6-6zM16 6h2v12h-2z"/></svg></button></div></div></div>',styles:[".mat-paginator{display:block}.mat-paginator-outer-container{display:flex}.mat-paginator-container{display:flex;align-items:center;justify-content:flex-end;min-height:56px;padding:0 8px;flex-wrap:wrap-reverse;width:100%}.mat-paginator-page-size{display:flex;align-items:baseline;margin-right:8px}[dir=rtl] .mat-paginator-page-size{margin-right:0;margin-left:8px}.mat-paginator-page-size-label{margin:0 4px}.mat-paginator-page-size-select{margin:6px 4px 0 4px;width:56px}.mat-paginator-page-size-select.mat-form-field-appearance-outline{width:64px}.mat-paginator-page-size-select.mat-form-field-appearance-fill{width:64px}.mat-paginator-range-label{margin:0 32px 0 24px}.mat-paginator-range-actions{display:flex;align-items:center}.mat-paginator-icon{width:28px;fill:currentColor}[dir=rtl] .mat-paginator-icon{transform:rotate(180deg)}"],inputs:["disabled"],host:{class:"mat-paginator"},changeDetection:r.ChangeDetectionStrategy.OnPush,encapsulation:r.ViewEncapsulation.None}]}],e.ctorParameters=function(){return[{type:p},{type:r.ChangeDetectorRef}]},e.propDecorators={color:[{type:r.Input}],pageIndex:[{type:r.Input}],length:[{type:r.Input}],pageSize:[{type:r.Input}],pageSizeOptions:[{type:r.Input}],hidePageSize:[{type:r.Input}],showFirstLastButtons:[{type:r.Input}],page:[{type:r.Output}]},e}(y),v=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[u.CommonModule,c.MatButtonModule,l.MatSelectModule,d.MatTooltipModule],exports:[_],declarations:[_],providers:[h]}]}],t}()},"./node_modules/@angular/material/esm5/progress-bar.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatProgressBarModule",function(){return y}),n.d(e,"MAT_PROGRESS_BAR_LOCATION_FACTORY",function(){return f}),n.d(e,"MatProgressBarBase",function(){return l}),n.d(e,"_MatProgressBarMixinBase",function(){return d}),n.d(e,"MAT_PROGRESS_BAR_LOCATION",function(){return p}),n.d(e,"MatProgressBar",function(){return m});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/rxjs/_esm5/index.js"),a=n("./node_modules/rxjs/_esm5/operators/index.js"),s=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),u=n("./node_modules/@angular/material/esm5/core.es5.js"),c=n("./node_modules/@angular/common/fesm5/common.js"),l=function(){return function(t){this._elementRef=t}}(),d=Object(u.mixinColor)(l,"primary"),p=new i.InjectionToken("mat-progress-bar-location",{providedIn:"root",factory:f});function f(){var t=Object(i.inject)(c.DOCUMENT),e=t?t.location:null;return{getPathname:function(){return e?e.pathname+e.search:""}}}var h=0,m=function(t){function e(e,n,r,a){var s=t.call(this,e)||this;s._elementRef=e,s._ngZone=n,s._animationMode=r,s._isNoopAnimation=!1,s._value=0,s._bufferValue=0,s.animationEnd=new i.EventEmitter,s._animationEndSubscription=o.Subscription.EMPTY,s.mode="determinate",s.progressbarId="mat-progress-bar-"+h++;var u=a?a.getPathname().split("#")[0]:"";return s._rectangleFillValue="url('"+u+"#"+s.progressbarId+"')",s._isNoopAnimation="NoopAnimations"===r,s}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"value",{get:function(){return this._value},set:function(t){this._value=g(t||0),this._isNoopAnimation&&this.emitAnimationEnd()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"bufferValue",{get:function(){return this._bufferValue},set:function(t){this._bufferValue=g(t||0)},enumerable:!0,configurable:!0}),e.prototype._primaryTransform=function(){return{transform:"scaleX("+this.value/100+")"}},e.prototype._bufferTransform=function(){if("buffer"===this.mode)return{transform:"scaleX("+this.bufferValue/100+")"}},e.prototype.ngAfterViewInit=function(){var t=this;this._isNoopAnimation||this._ngZone.runOutsideAngular(function(){t._animationEndSubscription=Object(o.fromEvent)(t._primaryValueBar.nativeElement,"transitionend").pipe(Object(a.filter)(function(e){return e.target===t._primaryValueBar.nativeElement})).subscribe(function(e){return t._ngZone.run(function(){return t.emitAnimationEnd()})})})},e.prototype.ngOnDestroy=function(){this._animationEndSubscription.unsubscribe()},e.prototype.emitAnimationEnd=function(){"determinate"!==this.mode&&"buffer"!==this.mode||this.animationEnd.next({value:this.value})},e.decorators=[{type:i.Component,args:[{selector:"mat-progress-bar",exportAs:"matProgressBar",host:{role:"progressbar","aria-valuemin":"0","aria-valuemax":"100","[attr.aria-valuenow]":"value","[attr.mode]":"mode",class:"mat-progress-bar","[class._mat-animation-noopable]":"_isNoopAnimation"},inputs:["color"],template:'<svg width="100%" height="4" focusable="false" class="mat-progress-bar-background mat-progress-bar-element"><defs><pattern [id]="progressbarId" x="4" y="0" width="8" height="4" patternUnits="userSpaceOnUse"><circle cx="2" cy="2" r="2"/></pattern></defs><rect [attr.fill]="_rectangleFillValue" width="100%" height="100%"/></svg><div class="mat-progress-bar-buffer mat-progress-bar-element" [ngStyle]="_bufferTransform()"></div><div class="mat-progress-bar-primary mat-progress-bar-fill mat-progress-bar-element" [ngStyle]="_primaryTransform()" #primaryValueBar></div><div class="mat-progress-bar-secondary mat-progress-bar-fill mat-progress-bar-element"></div>',styles:[".mat-progress-bar{display:block;height:4px;overflow:hidden;position:relative;transition:opacity 250ms linear;width:100%}._mat-animation-noopable.mat-progress-bar{transition:none;animation:none}.mat-progress-bar .mat-progress-bar-element,.mat-progress-bar .mat-progress-bar-fill::after{height:100%;position:absolute;width:100%}.mat-progress-bar .mat-progress-bar-background{width:calc(100% + 10px)}@media (-ms-high-contrast:active){.mat-progress-bar .mat-progress-bar-background{display:none}}.mat-progress-bar .mat-progress-bar-buffer{transform-origin:top left;transition:transform 250ms ease}@media (-ms-high-contrast:active){.mat-progress-bar .mat-progress-bar-buffer{border-top:solid 5px;opacity:.5}}.mat-progress-bar .mat-progress-bar-secondary{display:none}.mat-progress-bar .mat-progress-bar-fill{animation:none;transform-origin:top left;transition:transform 250ms ease}@media (-ms-high-contrast:active){.mat-progress-bar .mat-progress-bar-fill{border-top:solid 4px}}.mat-progress-bar .mat-progress-bar-fill::after{animation:none;content:'';display:inline-block;left:0}.mat-progress-bar[dir=rtl],[dir=rtl] .mat-progress-bar{transform:rotateY(180deg)}.mat-progress-bar[mode=query]{transform:rotateZ(180deg)}.mat-progress-bar[mode=query][dir=rtl],[dir=rtl] .mat-progress-bar[mode=query]{transform:rotateZ(180deg) rotateY(180deg)}.mat-progress-bar[mode=indeterminate] .mat-progress-bar-fill,.mat-progress-bar[mode=query] .mat-progress-bar-fill{transition:none}.mat-progress-bar[mode=indeterminate] .mat-progress-bar-primary,.mat-progress-bar[mode=query] .mat-progress-bar-primary{-webkit-backface-visibility:hidden;backface-visibility:hidden;animation:mat-progress-bar-primary-indeterminate-translate 2s infinite linear;left:-145.166611%}.mat-progress-bar[mode=indeterminate] .mat-progress-bar-primary.mat-progress-bar-fill::after,.mat-progress-bar[mode=query] .mat-progress-bar-primary.mat-progress-bar-fill::after{-webkit-backface-visibility:hidden;backface-visibility:hidden;animation:mat-progress-bar-primary-indeterminate-scale 2s infinite linear}.mat-progress-bar[mode=indeterminate] .mat-progress-bar-secondary,.mat-progress-bar[mode=query] .mat-progress-bar-secondary{-webkit-backface-visibility:hidden;backface-visibility:hidden;animation:mat-progress-bar-secondary-indeterminate-translate 2s infinite linear;left:-54.888891%;display:block}.mat-progress-bar[mode=indeterminate] .mat-progress-bar-secondary.mat-progress-bar-fill::after,.mat-progress-bar[mode=query] .mat-progress-bar-secondary.mat-progress-bar-fill::after{-webkit-backface-visibility:hidden;backface-visibility:hidden;animation:mat-progress-bar-secondary-indeterminate-scale 2s infinite linear}.mat-progress-bar[mode=buffer] .mat-progress-bar-background{-webkit-backface-visibility:hidden;backface-visibility:hidden;animation:mat-progress-bar-background-scroll 250ms infinite linear;display:block}.mat-progress-bar._mat-animation-noopable .mat-progress-bar-background,.mat-progress-bar._mat-animation-noopable .mat-progress-bar-buffer,.mat-progress-bar._mat-animation-noopable .mat-progress-bar-fill,.mat-progress-bar._mat-animation-noopable .mat-progress-bar-fill::after,.mat-progress-bar._mat-animation-noopable .mat-progress-bar-primary,.mat-progress-bar._mat-animation-noopable .mat-progress-bar-primary.mat-progress-bar-fill::after,.mat-progress-bar._mat-animation-noopable .mat-progress-bar-secondary,.mat-progress-bar._mat-animation-noopable .mat-progress-bar-secondary.mat-progress-bar-fill::after{animation:none;transition:none}@keyframes mat-progress-bar-primary-indeterminate-translate{0%{transform:translateX(0)}20%{animation-timing-function:cubic-bezier(.5,0,.70173,.49582);transform:translateX(0)}59.15%{animation-timing-function:cubic-bezier(.30244,.38135,.55,.95635);transform:translateX(83.67142%)}100%{transform:translateX(200.61106%)}}@keyframes mat-progress-bar-primary-indeterminate-scale{0%{transform:scaleX(.08)}36.65%{animation-timing-function:cubic-bezier(.33473,.12482,.78584,1);transform:scaleX(.08)}69.15%{animation-timing-function:cubic-bezier(.06,.11,.6,1);transform:scaleX(.66148)}100%{transform:scaleX(.08)}}@keyframes mat-progress-bar-secondary-indeterminate-translate{0%{animation-timing-function:cubic-bezier(.15,0,.51506,.40969);transform:translateX(0)}25%{animation-timing-function:cubic-bezier(.31033,.28406,.8,.73371);transform:translateX(37.65191%)}48.35%{animation-timing-function:cubic-bezier(.4,.62704,.6,.90203);transform:translateX(84.38617%)}100%{transform:translateX(160.27778%)}}@keyframes mat-progress-bar-secondary-indeterminate-scale{0%{animation-timing-function:cubic-bezier(.15,0,.51506,.40969);transform:scaleX(.08)}19.15%{animation-timing-function:cubic-bezier(.31033,.28406,.8,.73371);transform:scaleX(.4571)}44.15%{animation-timing-function:cubic-bezier(.4,.62704,.6,.90203);transform:scaleX(.72796)}100%{transform:scaleX(.08)}}@keyframes mat-progress-bar-background-scroll{to{transform:translateX(-8px)}}"],changeDetection:i.ChangeDetectionStrategy.OnPush,encapsulation:i.ViewEncapsulation.None}]}],e.ctorParameters=function(){return[{type:i.ElementRef},{type:i.NgZone},{type:String,decorators:[{type:i.Optional},{type:i.Inject,args:[s.ANIMATION_MODULE_TYPE]}]},{type:void 0,decorators:[{type:i.Optional},{type:i.Inject,args:[p]}]}]},e.propDecorators={value:[{type:i.Input}],bufferValue:[{type:i.Input}],_primaryValueBar:[{type:i.ViewChild,args:["primaryValueBar"]}],animationEnd:[{type:i.Output}],mode:[{type:i.Input}]},e}(d);function g(t,e,n){return void 0===e&&(e=0),void 0===n&&(n=100),Math.max(e,Math.min(n,t))}var y=function(){function t(){}return t.decorators=[{type:i.NgModule,args:[{imports:[c.CommonModule,u.MatCommonModule],exports:[m,u.MatCommonModule],declarations:[m]}]}],t}()},"./node_modules/@angular/material/esm5/progress-spinner.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatProgressSpinnerModule",function(){return y}),n.d(e,"MAT_PROGRESS_SPINNER_DEFAULT_OPTIONS_FACTORY",function(){return h}),n.d(e,"MatProgressSpinnerBase",function(){return d}),n.d(e,"_MatProgressSpinnerMixinBase",function(){return p}),n.d(e,"MAT_PROGRESS_SPINNER_DEFAULT_OPTIONS",function(){return f}),n.d(e,"MatProgressSpinner",function(){return m}),n.d(e,"MatSpinner",function(){return g});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),o=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),a=n("./node_modules/@angular/common/fesm5/common.js"),s=n("./node_modules/@angular/core/fesm5/core.js"),u=n("./node_modules/@angular/material/esm5/core.es5.js"),c=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),l=100,d=function(){return function(t){this._elementRef=t}}(),p=Object(u.mixinColor)(d,"primary"),f=new s.InjectionToken("mat-progress-spinner-default-options",{providedIn:"root",factory:h});function h(){return{diameter:l}}var m=function(t){function e(e,n,r,i,o){var a=t.call(this,e)||this;a._elementRef=e,a._document=r,a.animationMode=i,a.defaults=o,a._value=0,a._fallbackAnimation=!1,a._noopAnimations="NoopAnimations"===a.animationMode&&!!a.defaults&&!a.defaults._forceAnimations,a._diameter=l,a.mode="determinate",a._fallbackAnimation=n.EDGE||n.TRIDENT,o&&(o.diameter&&(a.diameter=o.diameter),o.strokeWidth&&(a.strokeWidth=o.strokeWidth));var s="mat-progress-spinner-indeterminate"+(a._fallbackAnimation?"-fallback":"")+"-animation";return e.nativeElement.classList.add(s),a}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"diameter",{get:function(){return this._diameter},set:function(t){this._diameter=Object(i.coerceNumberProperty)(t),this._fallbackAnimation||e.diameters.has(this._diameter)||this._attachStyleNode()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"strokeWidth",{get:function(){return this._strokeWidth||this.diameter/10},set:function(t){this._strokeWidth=Object(i.coerceNumberProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return"determinate"===this.mode?this._value:0},set:function(t){this._value=Math.max(0,Math.min(100,Object(i.coerceNumberProperty)(t)))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_circleRadius",{get:function(){return(this.diameter-10)/2},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_viewBox",{get:function(){var t=2*this._circleRadius+this.strokeWidth;return"0 0 "+t+" "+t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_strokeCircumference",{get:function(){return 2*Math.PI*this._circleRadius},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_strokeDashOffset",{get:function(){return"determinate"===this.mode?this._strokeCircumference*(100-this._value)/100:this._fallbackAnimation&&"indeterminate"===this.mode?.2*this._strokeCircumference:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_circleStrokeWidth",{get:function(){return this.strokeWidth/this.diameter*100},enumerable:!0,configurable:!0}),e.prototype._attachStyleNode=function(){var t=e.styleTag;t||(t=this._document.createElement("style"),this._document.head.appendChild(t),e.styleTag=t),t&&t.sheet&&t.sheet.insertRule(this._getAnimationText(),0),e.diameters.add(this.diameter)},e.prototype._getAnimationText=function(){return"\n @keyframes mat-progress-spinner-stroke-rotate-DIAMETER {\n    0%      { stroke-dashoffset: START_VALUE;  transform: rotate(0); }\n    12.5%   { stroke-dashoffset: END_VALUE;    transform: rotate(0); }\n    12.5001%  { stroke-dashoffset: END_VALUE;    transform: rotateX(180deg) rotate(72.5deg); }\n    25%     { stroke-dashoffset: START_VALUE;  transform: rotateX(180deg) rotate(72.5deg); }\n\n    25.0001%   { stroke-dashoffset: START_VALUE;  transform: rotate(270deg); }\n    37.5%   { stroke-dashoffset: END_VALUE;    transform: rotate(270deg); }\n    37.5001%  { stroke-dashoffset: END_VALUE;    transform: rotateX(180deg) rotate(161.5deg); }\n    50%     { stroke-dashoffset: START_VALUE;  transform: rotateX(180deg) rotate(161.5deg); }\n\n    50.0001%  { stroke-dashoffset: START_VALUE;  transform: rotate(180deg); }\n    62.5%   { stroke-dashoffset: END_VALUE;    transform: rotate(180deg); }\n    62.5001%  { stroke-dashoffset: END_VALUE;    transform: rotateX(180deg) rotate(251.5deg); }\n    75%     { stroke-dashoffset: START_VALUE;  transform: rotateX(180deg) rotate(251.5deg); }\n\n    75.0001%  { stroke-dashoffset: START_VALUE;  transform: rotate(90deg); }\n    87.5%   { stroke-dashoffset: END_VALUE;    transform: rotate(90deg); }\n    87.5001%  { stroke-dashoffset: END_VALUE;    transform: rotateX(180deg) rotate(341.5deg); }\n    100%    { stroke-dashoffset: START_VALUE;  transform: rotateX(180deg) rotate(341.5deg); }\n  }\n".replace(/START_VALUE/g,""+.95*this._strokeCircumference).replace(/END_VALUE/g,""+.2*this._strokeCircumference).replace(/DIAMETER/g,""+this.diameter)},e.diameters=new Set([l]),e.styleTag=null,e.decorators=[{type:s.Component,args:[{selector:"mat-progress-spinner",exportAs:"matProgressSpinner",host:{role:"progressbar",class:"mat-progress-spinner","[class._mat-animation-noopable]":"_noopAnimations","[style.width.px]":"diameter","[style.height.px]":"diameter","[attr.aria-valuemin]":'mode === "determinate" ? 0 : null',"[attr.aria-valuemax]":'mode === "determinate" ? 100 : null',"[attr.aria-valuenow]":"value","[attr.mode]":"mode"},inputs:["color"],template:'<svg [style.width.px]="diameter" [style.height.px]="diameter" [attr.viewBox]="_viewBox" preserveAspectRatio="xMidYMid meet" focusable="false" [ngSwitch]="mode === \'indeterminate\'"><circle *ngSwitchCase="true" cx="50%" cy="50%" [attr.r]="_circleRadius" [style.animation-name]="\'mat-progress-spinner-stroke-rotate-\' + diameter" [style.stroke-dashoffset.px]="_strokeDashOffset" [style.stroke-dasharray.px]="_strokeCircumference" [style.stroke-width.%]="_circleStrokeWidth"></circle><circle *ngSwitchCase="false" cx="50%" cy="50%" [attr.r]="_circleRadius" [style.stroke-dashoffset.px]="_strokeDashOffset" [style.stroke-dasharray.px]="_strokeCircumference" [style.stroke-width.%]="_circleStrokeWidth"></circle></svg>',styles:[".mat-progress-spinner{display:block;position:relative}.mat-progress-spinner svg{position:absolute;transform:rotate(-90deg);top:0;left:0;transform-origin:center;overflow:visible}.mat-progress-spinner circle{fill:transparent;transform-origin:center;transition:stroke-dashoffset 225ms linear}._mat-animation-noopable.mat-progress-spinner circle{transition:none;animation:none}.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate]{animation:mat-progress-spinner-linear-rotate 2s linear infinite}._mat-animation-noopable.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate]{transition:none;animation:none}.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate] circle{transition-property:stroke;animation-duration:4s;animation-timing-function:cubic-bezier(.35,0,.25,1);animation-iteration-count:infinite}._mat-animation-noopable.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate] circle{transition:none;animation:none}.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate]{animation:mat-progress-spinner-stroke-rotate-fallback 10s cubic-bezier(.87,.03,.33,1) infinite}._mat-animation-noopable.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate]{transition:none;animation:none}.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate] circle{transition-property:stroke}._mat-animation-noopable.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate] circle{transition:none;animation:none}@keyframes mat-progress-spinner-linear-rotate{0%{transform:rotate(0)}100%{transform:rotate(360deg)}}@keyframes mat-progress-spinner-stroke-rotate-100{0%{stroke-dashoffset:268.60617px;transform:rotate(0)}12.5%{stroke-dashoffset:56.54867px;transform:rotate(0)}12.5001%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(72.5deg)}25%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(72.5deg)}25.0001%{stroke-dashoffset:268.60617px;transform:rotate(270deg)}37.5%{stroke-dashoffset:56.54867px;transform:rotate(270deg)}37.5001%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(161.5deg)}50%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(161.5deg)}50.0001%{stroke-dashoffset:268.60617px;transform:rotate(180deg)}62.5%{stroke-dashoffset:56.54867px;transform:rotate(180deg)}62.5001%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(251.5deg)}75%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(251.5deg)}75.0001%{stroke-dashoffset:268.60617px;transform:rotate(90deg)}87.5%{stroke-dashoffset:56.54867px;transform:rotate(90deg)}87.5001%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(341.5deg)}100%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(341.5deg)}}@keyframes mat-progress-spinner-stroke-rotate-fallback{0%{transform:rotate(0)}25%{transform:rotate(1170deg)}50%{transform:rotate(2340deg)}75%{transform:rotate(3510deg)}100%{transform:rotate(4680deg)}}"],changeDetection:s.ChangeDetectionStrategy.OnPush,encapsulation:s.ViewEncapsulation.None}]}],e.ctorParameters=function(){return[{type:s.ElementRef},{type:o.Platform},{type:void 0,decorators:[{type:s.Optional},{type:s.Inject,args:[a.DOCUMENT]}]},{type:String,decorators:[{type:s.Optional},{type:s.Inject,args:[c.ANIMATION_MODULE_TYPE]}]},{type:void 0,decorators:[{type:s.Inject,args:[f]}]}]},e.propDecorators={diameter:[{type:s.Input}],strokeWidth:[{type:s.Input}],mode:[{type:s.Input}],value:[{type:s.Input}]},e}(p),g=function(t){function e(e,n,r,i,o){var a=t.call(this,e,n,r,i,o)||this;return a.mode="indeterminate",a}return Object(r.__extends)(e,t),e.decorators=[{type:s.Component,args:[{selector:"mat-spinner",host:{role:"progressbar",mode:"indeterminate",class:"mat-spinner mat-progress-spinner","[class._mat-animation-noopable]":"_noopAnimations","[style.width.px]":"diameter","[style.height.px]":"diameter"},inputs:["color"],template:'<svg [style.width.px]="diameter" [style.height.px]="diameter" [attr.viewBox]="_viewBox" preserveAspectRatio="xMidYMid meet" focusable="false" [ngSwitch]="mode === \'indeterminate\'"><circle *ngSwitchCase="true" cx="50%" cy="50%" [attr.r]="_circleRadius" [style.animation-name]="\'mat-progress-spinner-stroke-rotate-\' + diameter" [style.stroke-dashoffset.px]="_strokeDashOffset" [style.stroke-dasharray.px]="_strokeCircumference" [style.stroke-width.%]="_circleStrokeWidth"></circle><circle *ngSwitchCase="false" cx="50%" cy="50%" [attr.r]="_circleRadius" [style.stroke-dashoffset.px]="_strokeDashOffset" [style.stroke-dasharray.px]="_strokeCircumference" [style.stroke-width.%]="_circleStrokeWidth"></circle></svg>',styles:[".mat-progress-spinner{display:block;position:relative}.mat-progress-spinner svg{position:absolute;transform:rotate(-90deg);top:0;left:0;transform-origin:center;overflow:visible}.mat-progress-spinner circle{fill:transparent;transform-origin:center;transition:stroke-dashoffset 225ms linear}._mat-animation-noopable.mat-progress-spinner circle{transition:none;animation:none}.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate]{animation:mat-progress-spinner-linear-rotate 2s linear infinite}._mat-animation-noopable.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate]{transition:none;animation:none}.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate] circle{transition-property:stroke;animation-duration:4s;animation-timing-function:cubic-bezier(.35,0,.25,1);animation-iteration-count:infinite}._mat-animation-noopable.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate] circle{transition:none;animation:none}.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate]{animation:mat-progress-spinner-stroke-rotate-fallback 10s cubic-bezier(.87,.03,.33,1) infinite}._mat-animation-noopable.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate]{transition:none;animation:none}.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate] circle{transition-property:stroke}._mat-animation-noopable.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate] circle{transition:none;animation:none}@keyframes mat-progress-spinner-linear-rotate{0%{transform:rotate(0)}100%{transform:rotate(360deg)}}@keyframes mat-progress-spinner-stroke-rotate-100{0%{stroke-dashoffset:268.60617px;transform:rotate(0)}12.5%{stroke-dashoffset:56.54867px;transform:rotate(0)}12.5001%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(72.5deg)}25%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(72.5deg)}25.0001%{stroke-dashoffset:268.60617px;transform:rotate(270deg)}37.5%{stroke-dashoffset:56.54867px;transform:rotate(270deg)}37.5001%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(161.5deg)}50%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(161.5deg)}50.0001%{stroke-dashoffset:268.60617px;transform:rotate(180deg)}62.5%{stroke-dashoffset:56.54867px;transform:rotate(180deg)}62.5001%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(251.5deg)}75%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(251.5deg)}75.0001%{stroke-dashoffset:268.60617px;transform:rotate(90deg)}87.5%{stroke-dashoffset:56.54867px;transform:rotate(90deg)}87.5001%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(341.5deg)}100%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(341.5deg)}}@keyframes mat-progress-spinner-stroke-rotate-fallback{0%{transform:rotate(0)}25%{transform:rotate(1170deg)}50%{transform:rotate(2340deg)}75%{transform:rotate(3510deg)}100%{transform:rotate(4680deg)}}"],changeDetection:s.ChangeDetectionStrategy.OnPush,encapsulation:s.ViewEncapsulation.None}]}],e.ctorParameters=function(){return[{type:s.ElementRef},{type:o.Platform},{type:void 0,decorators:[{type:s.Optional},{type:s.Inject,args:[a.DOCUMENT]}]},{type:String,decorators:[{type:s.Optional},{type:s.Inject,args:[c.ANIMATION_MODULE_TYPE]}]},{type:void 0,decorators:[{type:s.Inject,args:[f]}]}]},e}(m),y=function(){function t(){}return t.decorators=[{type:s.NgModule,args:[{imports:[u.MatCommonModule,a.CommonModule],exports:[m,g,u.MatCommonModule],declarations:[m,g]}]}],t}()},"./node_modules/@angular/material/esm5/radio.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatRadioModule",function(){return v}),n.d(e,"MAT_RADIO_GROUP_CONTROL_VALUE_ACCESSOR",function(){return f}),n.d(e,"MatRadioChange",function(){return h}),n.d(e,"MatRadioGroup",function(){return m}),n.d(e,"MatRadioButtonBase",function(){return g}),n.d(e,"_MatRadioButtonMixinBase",function(){return y}),n.d(e,"MatRadioButton",function(){return _});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),o=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),a=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),s=n("./node_modules/@angular/core/fesm5/core.js"),u=n("./node_modules/@angular/forms/fesm5/forms.js"),c=n("./node_modules/@angular/material/esm5/core.es5.js"),l=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),d=n("./node_modules/@angular/common/fesm5/common.js"),p=0,f={provide:u.NG_VALUE_ACCESSOR,useExisting:Object(s.forwardRef)(function(){return m}),multi:!0},h=function(){return function(t,e){this.source=t,this.value=e}}(),m=function(){function t(t){this._changeDetector=t,this._value=null,this._name="mat-radio-group-"+p++,this._selected=null,this._isInitialized=!1,this._labelPosition="after",this._disabled=!1,this._required=!1,this._controlValueAccessorChangeFn=function(){},this.onTouched=function(){},this.change=new s.EventEmitter}return Object.defineProperty(t.prototype,"name",{get:function(){return this._name},set:function(t){this._name=t,this._updateRadioButtonNames()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"labelPosition",{get:function(){return this._labelPosition},set:function(t){this._labelPosition="before"===t?"before":"after",this._markRadiosForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{get:function(){return this._value},set:function(t){this._value!==t&&(this._value=t,this._updateSelectedRadioFromValue(),this._checkSelectedRadioButton())},enumerable:!0,configurable:!0}),t.prototype._checkSelectedRadioButton=function(){this._selected&&!this._selected.checked&&(this._selected.checked=!0)},Object.defineProperty(t.prototype,"selected",{get:function(){return this._selected},set:function(t){this._selected=t,this.value=t?t.value:null,this._checkSelectedRadioButton()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this._disabled},set:function(t){this._disabled=Object(o.coerceBooleanProperty)(t),this._markRadiosForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"required",{get:function(){return this._required},set:function(t){this._required=Object(o.coerceBooleanProperty)(t),this._markRadiosForCheck()},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this._isInitialized=!0},t.prototype._touch=function(){this.onTouched&&this.onTouched()},t.prototype._updateRadioButtonNames=function(){var t=this;this._radios&&this._radios.forEach(function(e){e.name=t.name})},t.prototype._updateSelectedRadioFromValue=function(){var t=this,e=null!==this._selected&&this._selected.value===this._value;this._radios&&!e&&(this._selected=null,this._radios.forEach(function(e){e.checked=t.value===e.value,e.checked&&(t._selected=e)}))},t.prototype._emitChangeEvent=function(){this._isInitialized&&this.change.emit(new h(this._selected,this._value))},t.prototype._markRadiosForCheck=function(){this._radios&&this._radios.forEach(function(t){return t._markForCheck()})},t.prototype.writeValue=function(t){this.value=t,this._changeDetector.markForCheck()},t.prototype.registerOnChange=function(t){this._controlValueAccessorChangeFn=t},t.prototype.registerOnTouched=function(t){this.onTouched=t},t.prototype.setDisabledState=function(t){this.disabled=t,this._changeDetector.markForCheck()},t.decorators=[{type:s.Directive,args:[{selector:"mat-radio-group",exportAs:"matRadioGroup",providers:[f],host:{role:"radiogroup",class:"mat-radio-group"}}]}],t.ctorParameters=function(){return[{type:s.ChangeDetectorRef}]},t.propDecorators={change:[{type:s.Output}],_radios:[{type:s.ContentChildren,args:[Object(s.forwardRef)(function(){return _}),{descendants:!0}]}],name:[{type:s.Input}],labelPosition:[{type:s.Input}],value:[{type:s.Input}],selected:[{type:s.Input}],disabled:[{type:s.Input}],required:[{type:s.Input}]},t}(),g=function(){return function(t){this._elementRef=t}}(),y=Object(c.mixinColor)(Object(c.mixinDisableRipple)(Object(c.mixinTabIndex)(g)),"accent"),_=function(t){function e(e,n,r,i,o,a){var u=t.call(this,n)||this;return u._changeDetector=r,u._focusMonitor=i,u._radioDispatcher=o,u._animationMode=a,u._uniqueId="mat-radio-"+ ++p,u.id=u._uniqueId,u.change=new s.EventEmitter,u._checked=!1,u._value=null,u._removeUniqueSelectionListener=function(){},u.radioGroup=e,u._removeUniqueSelectionListener=o.listen(function(t,e){t!==u.id&&e===u.name&&(u.checked=!1)}),u}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"checked",{get:function(){return this._checked},set:function(t){var e=Object(o.coerceBooleanProperty)(t);this._checked!==e&&(this._checked=e,e&&this.radioGroup&&this.radioGroup.value!==this.value?this.radioGroup.selected=this:!e&&this.radioGroup&&this.radioGroup.value===this.value&&(this.radioGroup.selected=null),e&&this._radioDispatcher.notify(this.id,this.name),this._changeDetector.markForCheck())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return this._value},set:function(t){this._value!==t&&(this._value=t,null!==this.radioGroup&&(this.checked||(this.checked=this.radioGroup.value===t),this.checked&&(this.radioGroup.selected=this)))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"labelPosition",{get:function(){return this._labelPosition||this.radioGroup&&this.radioGroup.labelPosition||"after"},set:function(t){this._labelPosition=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return this._disabled||null!==this.radioGroup&&this.radioGroup.disabled},set:function(t){var e=Object(o.coerceBooleanProperty)(t);this._disabled!==e&&(this._disabled=e,this._changeDetector.markForCheck())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"required",{get:function(){return this._required||this.radioGroup&&this.radioGroup.required},set:function(t){this._required=Object(o.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"inputId",{get:function(){return(this.id||this._uniqueId)+"-input"},enumerable:!0,configurable:!0}),e.prototype.focus=function(){this._focusMonitor.focusVia(this._inputElement,"keyboard")},e.prototype._markForCheck=function(){this._changeDetector.markForCheck()},e.prototype.ngOnInit=function(){this.radioGroup&&(this.checked=this.radioGroup.value===this._value,this.name=this.radioGroup.name)},e.prototype.ngAfterViewInit=function(){var t=this;this._focusMonitor.monitor(this._elementRef,!0).subscribe(function(e){!e&&t.radioGroup&&t.radioGroup._touch()})},e.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef),this._removeUniqueSelectionListener()},e.prototype._emitChangeEvent=function(){this.change.emit(new h(this,this._value))},e.prototype._isRippleDisabled=function(){return this.disableRipple||this.disabled},e.prototype._onInputClick=function(t){t.stopPropagation()},e.prototype._onInputChange=function(t){t.stopPropagation();var e=this.radioGroup&&this.value!==this.radioGroup.value;this.checked=!0,this._emitChangeEvent(),this.radioGroup&&(this.radioGroup._controlValueAccessorChangeFn(this.value),this.radioGroup._touch(),e&&this.radioGroup._emitChangeEvent())},e.decorators=[{type:s.Component,args:[{selector:"mat-radio-button",template:'<label [attr.for]="inputId" class="mat-radio-label" #label><div class="mat-radio-container"><div class="mat-radio-outer-circle"></div><div class="mat-radio-inner-circle"></div><div mat-ripple class="mat-radio-ripple" [matRippleTrigger]="label" [matRippleDisabled]="_isRippleDisabled()" [matRippleCentered]="true" [matRippleRadius]="20" [matRippleAnimation]="{enterDuration: 150}"><div class="mat-ripple-element mat-radio-persistent-ripple"></div></div><input #input class="mat-radio-input cdk-visually-hidden" type="radio" [id]="inputId" [checked]="checked" [disabled]="disabled" [tabIndex]="tabIndex" [attr.name]="name" [required]="required" [attr.aria-label]="ariaLabel" [attr.aria-labelledby]="ariaLabelledby" [attr.aria-describedby]="ariaDescribedby" (change)="_onInputChange($event)" (click)="_onInputClick($event)"></div><div class="mat-radio-label-content" [class.mat-radio-label-before]="labelPosition == \'before\'"><span style="display:none">&nbsp;</span><ng-content></ng-content></div></label>',styles:[".mat-radio-button{display:inline-block;-webkit-tap-highlight-color:transparent;outline:0}.mat-radio-label{cursor:pointer;display:inline-flex;align-items:center;white-space:nowrap;vertical-align:middle}.mat-radio-container{box-sizing:border-box;display:inline-block;position:relative;width:20px;height:20px;flex-shrink:0}.mat-radio-outer-circle{box-sizing:border-box;height:20px;left:0;position:absolute;top:0;transition:border-color ease 280ms;width:20px;border-width:2px;border-style:solid;border-radius:50%}._mat-animation-noopable .mat-radio-outer-circle{transition:none}.mat-radio-inner-circle{border-radius:50%;box-sizing:border-box;height:20px;left:0;position:absolute;top:0;transition:transform ease 280ms,background-color ease 280ms;width:20px;transform:scale(.001)}._mat-animation-noopable .mat-radio-inner-circle{transition:none}.mat-radio-checked .mat-radio-inner-circle{transform:scale(.5)}@media (-ms-high-contrast:active){.mat-radio-checked .mat-radio-inner-circle{border:solid 10px}}.mat-radio-label-content{display:inline-block;order:0;line-height:inherit;padding-left:8px;padding-right:0}[dir=rtl] .mat-radio-label-content{padding-right:8px;padding-left:0}.mat-radio-label-content.mat-radio-label-before{order:-1;padding-left:0;padding-right:8px}[dir=rtl] .mat-radio-label-content.mat-radio-label-before{padding-right:0;padding-left:8px}.mat-radio-disabled,.mat-radio-disabled .mat-radio-label{cursor:default}.mat-radio-button .mat-radio-ripple{position:absolute;left:calc(50% - 20px);top:calc(50% - 20px);height:40px;width:40px;z-index:1;pointer-events:none}.mat-radio-button .mat-radio-ripple .mat-ripple-element:not(.mat-radio-persistent-ripple){opacity:.16}.mat-radio-persistent-ripple{width:100%;height:100%;transform:none}.mat-radio-container:hover .mat-radio-persistent-ripple{opacity:.04}.mat-radio-button:not(.mat-radio-disabled).cdk-focused .mat-radio-persistent-ripple{opacity:.12}.mat-radio-disabled .mat-radio-container:hover .mat-radio-persistent-ripple,.mat-radio-persistent-ripple{opacity:0}@media (hover:none){.mat-radio-container:hover .mat-radio-persistent-ripple{display:none}}.mat-radio-input{bottom:0;left:50%}"],inputs:["color","disableRipple","tabIndex"],encapsulation:s.ViewEncapsulation.None,exportAs:"matRadioButton",host:{class:"mat-radio-button","[class.mat-radio-checked]":"checked","[class.mat-radio-disabled]":"disabled","[class._mat-animation-noopable]":'_animationMode === "NoopAnimations"',"[attr.tabindex]":"-1","[attr.id]":"id","(focus)":"_inputElement.nativeElement.focus()"},changeDetection:s.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:m,decorators:[{type:s.Optional}]},{type:s.ElementRef},{type:s.ChangeDetectorRef},{type:i.FocusMonitor},{type:a.UniqueSelectionDispatcher},{type:String,decorators:[{type:s.Optional},{type:s.Inject,args:[l.ANIMATION_MODULE_TYPE]}]}]},e.propDecorators={id:[{type:s.Input}],name:[{type:s.Input}],ariaLabel:[{type:s.Input,args:["aria-label"]}],ariaLabelledby:[{type:s.Input,args:["aria-labelledby"]}],ariaDescribedby:[{type:s.Input,args:["aria-describedby"]}],checked:[{type:s.Input}],value:[{type:s.Input}],labelPosition:[{type:s.Input}],disabled:[{type:s.Input}],required:[{type:s.Input}],change:[{type:s.Output}],_inputElement:[{type:s.ViewChild,args:["input"]}]},e}(y),v=function(){function t(){}return t.decorators=[{type:s.NgModule,args:[{imports:[d.CommonModule,c.MatRippleModule,c.MatCommonModule],exports:[m,_,c.MatCommonModule],declarations:[m,_]}]}],t}()},"./node_modules/@angular/material/esm5/select.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatSelectModule",function(){return F}),n.d(e,"MAT_SELECT_SCROLL_STRATEGY_PROVIDER_FACTORY",function(){return k}),n.d(e,"SELECT_PANEL_MAX_HEIGHT",function(){return j}),n.d(e,"SELECT_PANEL_PADDING_X",function(){return O}),n.d(e,"SELECT_PANEL_INDENT_PADDING_X",function(){return C}),n.d(e,"SELECT_ITEM_HEIGHT_EM",function(){return E}),n.d(e,"SELECT_MULTIPLE_PANEL_PADDING_X",function(){return S}),n.d(e,"SELECT_PANEL_VIEWPORT_PADDING",function(){return M}),n.d(e,"MAT_SELECT_SCROLL_STRATEGY",function(){return T}),n.d(e,"MAT_SELECT_SCROLL_STRATEGY_PROVIDER",function(){return A}),n.d(e,"MatSelectChange",function(){return D}),n.d(e,"MatSelectBase",function(){return I}),n.d(e,"_MatSelectMixinBase",function(){return P}),n.d(e,"MatSelectTrigger",function(){return R}),n.d(e,"MatSelect",function(){return N}),n.d(e,"matSelectAnimations",function(){return v}),n.d(e,"transformPanel",function(){return b}),n.d(e,"fadeInContent",function(){return x});var r=n("./node_modules/@angular/animations/fesm5/animations.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),a=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),c=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),l=n("./node_modules/@angular/cdk/esm5/overlay.es5.js"),d=n("./node_modules/@angular/cdk/esm5/scrolling.es5.js"),p=n("./node_modules/@angular/core/fesm5/core.js"),f=n("./node_modules/@angular/forms/fesm5/forms.js"),h=n("./node_modules/@angular/material/esm5/core.es5.js"),m=n("./node_modules/@angular/material/esm5/form-field.es5.js"),g=n("./node_modules/rxjs/_esm5/index.js"),y=n("./node_modules/rxjs/_esm5/operators/index.js"),_=n("./node_modules/@angular/common/fesm5/common.js"),v={transformPanel:Object(r.trigger)("transformPanel",[Object(r.state)("void",Object(r.style)({transform:"scaleY(0.8)",minWidth:"100%",opacity:0})),Object(r.state)("showing",Object(r.style)({opacity:1,minWidth:"calc(100% + 32px)",transform:"scaleY(1)"})),Object(r.state)("showing-multiple",Object(r.style)({opacity:1,minWidth:"calc(100% + 64px)",transform:"scaleY(1)"})),Object(r.transition)("void => *",Object(r.animate)("120ms cubic-bezier(0, 0, 0.2, 1)")),Object(r.transition)("* => void",Object(r.animate)("100ms 25ms linear",Object(r.style)({opacity:0})))]),fadeInContent:Object(r.trigger)("fadeInContent",[Object(r.state)("showing",Object(r.style)({opacity:1})),Object(r.transition)("void => showing",[Object(r.style)({opacity:0}),Object(r.animate)("150ms 100ms cubic-bezier(0.55, 0, 0.55, 0.2)")])])},b=v.transformPanel,x=v.fadeInContent;var w=0,j=256,O=16,C=2*O,E=3,S=0,M=8,T=new p.InjectionToken("mat-select-scroll-strategy");function k(t){return function(){return t.scrollStrategies.reposition()}}var A={provide:T,deps:[l.Overlay],useFactory:k},D=function(){return function(t,e){this.source=t,this.value=e}}(),I=function(){return function(t,e,n,r,i){this._elementRef=t,this._defaultErrorStateMatcher=e,this._parentForm=n,this._parentFormGroup=r,this.ngControl=i}}(),P=Object(h.mixinDisableRipple)(Object(h.mixinTabIndex)(Object(h.mixinDisabled)(Object(h.mixinErrorState)(I)))),R=function(){function t(){}return t.decorators=[{type:p.Directive,args:[{selector:"mat-select-trigger"}]}],t}(),N=function(t){function e(e,n,r,i,o,a,s,u,c,l,d,f,h){var m=t.call(this,o,i,s,u,l)||this;return m._viewportRuler=e,m._changeDetectorRef=n,m._ngZone=r,m._dir=a,m._parentFormField=c,m.ngControl=l,m._liveAnnouncer=h,m._panelOpen=!1,m._required=!1,m._scrollTop=0,m._multiple=!1,m._compareWith=function(t,e){return t===e},m._uid="mat-select-"+w++,m._destroy=new g.Subject,m._triggerFontSize=0,m._onChange=function(){},m._onTouched=function(){},m._optionIds="",m._transformOrigin="top",m._panelDoneAnimatingStream=new g.Subject,m._offsetY=0,m._positions=[{originX:"start",originY:"top",overlayX:"start",overlayY:"top"},{originX:"start",originY:"bottom",overlayX:"start",overlayY:"bottom"}],m._disableOptionCentering=!1,m._focused=!1,m.controlType="mat-select",m.ariaLabel="",m.optionSelectionChanges=Object(g.defer)(function(){return m.options?g.merge.apply(void 0,m.options.map(function(t){return t.onSelectionChange})):m._ngZone.onStable.asObservable().pipe(Object(y.take)(1),Object(y.switchMap)(function(){return m.optionSelectionChanges}))}),m.openedChange=new p.EventEmitter,m._openedStream=m.openedChange.pipe(Object(y.filter)(function(t){return t}),Object(y.map)(function(){})),m._closedStream=m.openedChange.pipe(Object(y.filter)(function(t){return!t}),Object(y.map)(function(){})),m.selectionChange=new p.EventEmitter,m.valueChange=new p.EventEmitter,m.ngControl&&(m.ngControl.valueAccessor=m),m._scrollStrategyFactory=f,m._scrollStrategy=m._scrollStrategyFactory(),m.tabIndex=parseInt(d)||0,m.id=m.id,m}return Object(i.__extends)(e,t),Object.defineProperty(e.prototype,"focused",{get:function(){return this._focused||this._panelOpen},set:function(t){this._focused=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"placeholder",{get:function(){return this._placeholder},set:function(t){this._placeholder=t,this.stateChanges.next()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"required",{get:function(){return this._required},set:function(t){this._required=Object(s.coerceBooleanProperty)(t),this.stateChanges.next()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"multiple",{get:function(){return this._multiple},set:function(t){if(this._selectionModel)throw Error("Cannot change `multiple` mode of select after initialization.");this._multiple=Object(s.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disableOptionCentering",{get:function(){return this._disableOptionCentering},set:function(t){this._disableOptionCentering=Object(s.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"compareWith",{get:function(){return this._compareWith},set:function(t){if("function"!=typeof t)throw Error("`compareWith` must be a function.");this._compareWith=t,this._selectionModel&&this._initializeSelection()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return this._value},set:function(t){t!==this._value&&(this.writeValue(t),this._value=t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"id",{get:function(){return this._id},set:function(t){this._id=t||this._uid,this.stateChanges.next()},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){var t=this;this._selectionModel=new u.SelectionModel(this.multiple),this.stateChanges.next(),this._panelDoneAnimatingStream.pipe(Object(y.distinctUntilChanged)(),Object(y.takeUntil)(this._destroy)).subscribe(function(){t.panelOpen?(t._scrollTop=0,t.openedChange.emit(!0)):(t.openedChange.emit(!1),t.overlayDir.offsetX=0,t._changeDetectorRef.markForCheck())})},e.prototype.ngAfterContentInit=function(){var t=this;this._initKeyManager(),this._selectionModel.onChange.pipe(Object(y.takeUntil)(this._destroy)).subscribe(function(t){t.added.forEach(function(t){return t.select()}),t.removed.forEach(function(t){return t.deselect()})}),this.options.changes.pipe(Object(y.startWith)(null),Object(y.takeUntil)(this._destroy)).subscribe(function(){t._resetOptions(),t._initializeSelection()})},e.prototype.ngDoCheck=function(){this.ngControl&&this.updateErrorState()},e.prototype.ngOnChanges=function(t){t.disabled&&this.stateChanges.next()},e.prototype.ngOnDestroy=function(){this._destroy.next(),this._destroy.complete(),this.stateChanges.complete()},e.prototype.toggle=function(){this.panelOpen?this.close():this.open()},e.prototype.open=function(){var t=this;!this.disabled&&this.options&&this.options.length&&!this._panelOpen&&(this._triggerRect=this.trigger.nativeElement.getBoundingClientRect(),this._triggerFontSize=parseInt(getComputedStyle(this.trigger.nativeElement).fontSize||"0"),this._panelOpen=!0,this._keyManager.withHorizontalOrientation(null),this._calculateOverlayPosition(),this._highlightCorrectOption(),this._changeDetectorRef.markForCheck(),this._ngZone.onStable.asObservable().pipe(Object(y.take)(1)).subscribe(function(){t._triggerFontSize&&t.overlayDir.overlayRef&&t.overlayDir.overlayRef.overlayElement&&(t.overlayDir.overlayRef.overlayElement.style.fontSize=t._triggerFontSize+"px")}))},e.prototype.close=function(){this._panelOpen&&(this._panelOpen=!1,this._keyManager.withHorizontalOrientation(this._isRtl()?"rtl":"ltr"),this._changeDetectorRef.markForCheck(),this._onTouched())},e.prototype.writeValue=function(t){this.options&&this._setSelectionByValue(t)},e.prototype.registerOnChange=function(t){this._onChange=t},e.prototype.registerOnTouched=function(t){this._onTouched=t},e.prototype.setDisabledState=function(t){this.disabled=t,this._changeDetectorRef.markForCheck(),this.stateChanges.next()},Object.defineProperty(e.prototype,"panelOpen",{get:function(){return this._panelOpen},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selected",{get:function(){return this.multiple?this._selectionModel.selected:this._selectionModel.selected[0]},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"triggerValue",{get:function(){if(this.empty)return"";if(this._multiple){var t=this._selectionModel.selected.map(function(t){return t.viewValue});return this._isRtl()&&t.reverse(),t.join(", ")}return this._selectionModel.selected[0].viewValue},enumerable:!0,configurable:!0}),e.prototype._isRtl=function(){return!!this._dir&&"rtl"===this._dir.value},e.prototype._handleKeydown=function(t){this.disabled||(this.panelOpen?this._handleOpenKeydown(t):this._handleClosedKeydown(t))},e.prototype._handleClosedKeydown=function(t){var e=t.keyCode,n=e===c.DOWN_ARROW||e===c.UP_ARROW||e===c.LEFT_ARROW||e===c.RIGHT_ARROW,r=e===c.ENTER||e===c.SPACE,i=this._keyManager;if(r&&!Object(c.hasModifierKey)(t)||(this.multiple||t.altKey)&&n)t.preventDefault(),this.open();else if(!this.multiple){var o=this.selected;e===c.HOME||e===c.END?(e===c.HOME?i.setFirstItemActive():i.setLastItemActive(),t.preventDefault()):i.onKeydown(t),this._liveAnnouncer&&o!==this.selected&&this._liveAnnouncer.announce(this.selected.viewValue)}},e.prototype._handleOpenKeydown=function(t){var e=t.keyCode,n=e===c.DOWN_ARROW||e===c.UP_ARROW,r=this._keyManager;if(e===c.HOME||e===c.END)t.preventDefault(),e===c.HOME?r.setFirstItemActive():r.setLastItemActive();else if(n&&t.altKey)t.preventDefault(),this.close();else if(e!==c.ENTER&&e!==c.SPACE||!r.activeItem||Object(c.hasModifierKey)(t))if(this._multiple&&e===c.A&&t.ctrlKey){t.preventDefault();var i=this.options.some(function(t){return!t.disabled&&!t.selected});this.options.forEach(function(t){t.disabled||(i?t.select():t.deselect())})}else{var o=r.activeItemIndex;r.onKeydown(t),this._multiple&&n&&t.shiftKey&&r.activeItem&&r.activeItemIndex!==o&&r.activeItem._selectViaInteraction()}else t.preventDefault(),r.activeItem._selectViaInteraction()},e.prototype._onFocus=function(){this.disabled||(this._focused=!0,this.stateChanges.next())},e.prototype._onBlur=function(){this._focused=!1,this.disabled||this.panelOpen||(this._onTouched(),this._changeDetectorRef.markForCheck(),this.stateChanges.next())},e.prototype._onAttached=function(){var t=this;this.overlayDir.positionChange.pipe(Object(y.take)(1)).subscribe(function(){t._setPseudoCheckboxPaddingSize(),t._changeDetectorRef.detectChanges(),t._calculateOverlayOffsetX(),t.panel.nativeElement.scrollTop=t._scrollTop})},e.prototype._getPanelTheme=function(){return this._parentFormField?"mat-"+this._parentFormField.color:""},e.prototype._setPseudoCheckboxPaddingSize=function(){if(!S&&this.multiple){var t=this.panel.nativeElement.querySelector(".mat-pseudo-checkbox");t&&(S=1.5*O+t.offsetWidth)}},Object.defineProperty(e.prototype,"empty",{get:function(){return!this._selectionModel||this._selectionModel.isEmpty()},enumerable:!0,configurable:!0}),e.prototype._initializeSelection=function(){var t=this;Promise.resolve().then(function(){t._setSelectionByValue(t.ngControl?t.ngControl.value:t._value),t.stateChanges.next()})},e.prototype._setSelectionByValue=function(t){var e=this;if(this.multiple&&t){if(!Array.isArray(t))throw Error("Value must be an array in multiple-selection mode.");this._selectionModel.clear(),t.forEach(function(t){return e._selectValue(t)}),this._sortValues()}else{this._selectionModel.clear();var n=this._selectValue(t);n&&this._keyManager.setActiveItem(n)}this._changeDetectorRef.markForCheck()},e.prototype._selectValue=function(t){var e=this,n=this.options.find(function(n){try{return null!=n.value&&e._compareWith(n.value,t)}catch(t){return Object(p.isDevMode)()&&console.warn(t),!1}});return n&&this._selectionModel.select(n),n},e.prototype._initKeyManager=function(){var t=this;this._keyManager=new o.ActiveDescendantKeyManager(this.options).withTypeAhead().withVerticalOrientation().withHorizontalOrientation(this._isRtl()?"rtl":"ltr").withAllowedModifierKeys(["shiftKey"]),this._keyManager.tabOut.pipe(Object(y.takeUntil)(this._destroy)).subscribe(function(){t.focus(),t.close()}),this._keyManager.change.pipe(Object(y.takeUntil)(this._destroy)).subscribe(function(){t._panelOpen&&t.panel?t._scrollActiveOptionIntoView():t._panelOpen||t.multiple||!t._keyManager.activeItem||t._keyManager.activeItem._selectViaInteraction()})},e.prototype._resetOptions=function(){var t=this,e=Object(g.merge)(this.options.changes,this._destroy);this.optionSelectionChanges.pipe(Object(y.takeUntil)(e)).subscribe(function(e){t._onSelect(e.source,e.isUserInput),e.isUserInput&&!t.multiple&&t._panelOpen&&(t.close(),t.focus())}),g.merge.apply(void 0,this.options.map(function(t){return t._stateChanges})).pipe(Object(y.takeUntil)(e)).subscribe(function(){t._changeDetectorRef.markForCheck(),t.stateChanges.next()}),this._setOptionIds()},e.prototype._onSelect=function(t,e){var n=this._selectionModel.isSelected(t);null!=t.value||this._multiple?(t.selected?this._selectionModel.select(t):this._selectionModel.deselect(t),e&&this._keyManager.setActiveItem(t),this.multiple&&(this._sortValues(),e&&this.focus())):(t.deselect(),this._selectionModel.clear(),this._propagateChanges(t.value)),n!==this._selectionModel.isSelected(t)&&this._propagateChanges(),this.stateChanges.next()},e.prototype._sortValues=function(){var t=this;if(this.multiple){var e=this.options.toArray();this._selectionModel.sort(function(n,r){return t.sortComparator?t.sortComparator(n,r,e):e.indexOf(n)-e.indexOf(r)}),this.stateChanges.next()}},e.prototype._propagateChanges=function(t){var e=null;e=this.multiple?this.selected.map(function(t){return t.value}):this.selected?this.selected.value:t,this._value=e,this.valueChange.emit(e),this._onChange(e),this.selectionChange.emit(new D(this,e)),this._changeDetectorRef.markForCheck()},e.prototype._setOptionIds=function(){this._optionIds=this.options.map(function(t){return t.id}).join(" ")},e.prototype._highlightCorrectOption=function(){this._keyManager&&(this.empty?this._keyManager.setFirstItemActive():this._keyManager.setActiveItem(this._selectionModel.selected[0]))},e.prototype._scrollActiveOptionIntoView=function(){var t=this._keyManager.activeItemIndex||0,e=Object(h._countGroupLabelsBeforeOption)(t,this.options,this.optionGroups);this.panel.nativeElement.scrollTop=Object(h._getOptionScrollPosition)(t+e,this._getItemHeight(),this.panel.nativeElement.scrollTop,j)},e.prototype.focus=function(){this._elementRef.nativeElement.focus()},e.prototype._getOptionIndex=function(t){return this.options.reduce(function(e,n,r){return void 0===e?t===n?r:void 0:e},void 0)},e.prototype._calculateOverlayPosition=function(){var t=this._getItemHeight(),e=this._getItemCount(),n=Math.min(e*t,j),r=e*t-n,i=this.empty?0:this._getOptionIndex(this._selectionModel.selected[0]);i+=Object(h._countGroupLabelsBeforeOption)(i,this.options,this.optionGroups);var o=n/2;this._scrollTop=this._calculateOverlayScroll(i,o,r),this._offsetY=this._calculateOverlayOffsetY(i,o,r),this._checkOverlayWithinViewport(r)},e.prototype._calculateOverlayScroll=function(t,e,n){var r=this._getItemHeight(),i=r*t-e+r/2;return Math.min(Math.max(0,i),n)},e.prototype._getAriaLabel=function(){return this.ariaLabelledby?null:this.ariaLabel||this.placeholder},e.prototype._getAriaLabelledby=function(){return this.ariaLabelledby?this.ariaLabelledby:this._parentFormField&&this._parentFormField._hasFloatingLabel()&&!this._getAriaLabel()?this._parentFormField._labelId||null:null},e.prototype._getAriaActiveDescendant=function(){return this.panelOpen&&this._keyManager&&this._keyManager.activeItem?this._keyManager.activeItem.id:null},e.prototype._calculateOverlayOffsetX=function(){var t,e=this.overlayDir.overlayRef.overlayElement.getBoundingClientRect(),n=this._viewportRuler.getViewportSize(),r=this._isRtl(),i=this.multiple?S+O:2*O;if(this.multiple)t=S;else{var o=this._selectionModel.selected[0]||this.options.first;t=o&&o.group?C:O}r||(t*=-1);var a=0-(e.left+t-(r?i:0)),s=e.right+t-n.width+(r?0:i);a>0?t+=a+M:s>0&&(t-=s+M),this.overlayDir.offsetX=Math.round(t),this.overlayDir.overlayRef.updatePosition()},e.prototype._calculateOverlayOffsetY=function(t,e,n){var r,i=this._getItemHeight(),o=(i-this._triggerRect.height)/2,a=Math.floor(j/i);if(this._disableOptionCentering)return 0;if(0===this._scrollTop)r=t*i;else if(this._scrollTop===n){r=(t-(this._getItemCount()-a))*i+(i-(this._getItemCount()*i-j)%i)}else r=e-i/2;return Math.round(-1*r-o)},e.prototype._checkOverlayWithinViewport=function(t){var e=this._getItemHeight(),n=this._viewportRuler.getViewportSize(),r=this._triggerRect.top-M,i=n.height-this._triggerRect.bottom-M,o=Math.abs(this._offsetY),a=Math.min(this._getItemCount()*e,j)-o-this._triggerRect.height;a>i?this._adjustPanelUp(a,i):o>r?this._adjustPanelDown(o,r,t):this._transformOrigin=this._getOriginBasedOnOption()},e.prototype._adjustPanelUp=function(t,e){var n=Math.round(t-e);this._scrollTop-=n,this._offsetY-=n,this._transformOrigin=this._getOriginBasedOnOption(),this._scrollTop<=0&&(this._scrollTop=0,this._offsetY=0,this._transformOrigin="50% bottom 0px")},e.prototype._adjustPanelDown=function(t,e,n){var r=Math.round(t-e);if(this._scrollTop+=r,this._offsetY+=r,this._transformOrigin=this._getOriginBasedOnOption(),this._scrollTop>=n)return this._scrollTop=n,this._offsetY=0,void(this._transformOrigin="50% top 0px")},e.prototype._getOriginBasedOnOption=function(){var t=this._getItemHeight(),e=(t-this._triggerRect.height)/2;return"50% "+(Math.abs(this._offsetY)-e+t/2)+"px 0px"},e.prototype._getItemCount=function(){return this.options.length+this.optionGroups.length},e.prototype._getItemHeight=function(){return this._triggerFontSize*E},e.prototype.setDescribedByIds=function(t){this._ariaDescribedby=t.join(" ")},e.prototype.onContainerClick=function(){this.focus(),this.open()},Object.defineProperty(e.prototype,"shouldLabelFloat",{get:function(){return this._panelOpen||!this.empty},enumerable:!0,configurable:!0}),e.decorators=[{type:p.Component,args:[{selector:"mat-select",exportAs:"matSelect",template:'<div cdk-overlay-origin class="mat-select-trigger" aria-hidden="true" (click)="toggle()" #origin="cdkOverlayOrigin" #trigger><div class="mat-select-value" [ngSwitch]="empty"><span class="mat-select-placeholder" *ngSwitchCase="true">{{placeholder || \' \'}}</span> <span class="mat-select-value-text" *ngSwitchCase="false" [ngSwitch]="!!customTrigger"><span *ngSwitchDefault>{{triggerValue || \' \'}}</span><ng-content select="mat-select-trigger" *ngSwitchCase="true"></ng-content></span></div><div class="mat-select-arrow-wrapper"><div class="mat-select-arrow"></div></div></div><ng-template cdk-connected-overlay cdkConnectedOverlayLockPosition cdkConnectedOverlayHasBackdrop cdkConnectedOverlayBackdropClass="cdk-overlay-transparent-backdrop" [cdkConnectedOverlayScrollStrategy]="_scrollStrategy" [cdkConnectedOverlayOrigin]="origin" [cdkConnectedOverlayOpen]="panelOpen" [cdkConnectedOverlayPositions]="_positions" [cdkConnectedOverlayMinWidth]="_triggerRect?.width" [cdkConnectedOverlayOffsetY]="_offsetY" (backdropClick)="close()" (attach)="_onAttached()" (detach)="close()"><div #panel class="mat-select-panel {{ _getPanelTheme() }}" [ngClass]="panelClass" [@transformPanel]="multiple ? \'showing-multiple\' : \'showing\'" (@transformPanel.done)="_panelDoneAnimatingStream.next($event.toState)" [style.transformOrigin]="_transformOrigin" [style.font-size.px]="_triggerFontSize" (keydown)="_handleKeydown($event)"><ng-content></ng-content></div></ng-template>',styles:[".mat-select{display:inline-block;width:100%;outline:0}.mat-select-trigger{display:inline-table;cursor:pointer;position:relative;box-sizing:border-box}.mat-select-disabled .mat-select-trigger{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:default}.mat-select-value{display:table-cell;max-width:0;width:100%;overflow:hidden;text-overflow:ellipsis;white-space:nowrap}.mat-select-value-text{white-space:nowrap;overflow:hidden;text-overflow:ellipsis}.mat-select-arrow-wrapper{display:table-cell;vertical-align:middle}.mat-form-field-appearance-fill .mat-select-arrow-wrapper{transform:translateY(-50%)}.mat-form-field-appearance-outline .mat-select-arrow-wrapper{transform:translateY(-25%)}.mat-form-field-appearance-standard.mat-form-field-has-label .mat-select:not(.mat-select-empty) .mat-select-arrow-wrapper{transform:translateY(-50%)}.mat-form-field-appearance-standard .mat-select.mat-select-empty .mat-select-arrow-wrapper{transition:transform .4s cubic-bezier(.25,.8,.25,1)}._mat-animation-noopable.mat-form-field-appearance-standard .mat-select.mat-select-empty .mat-select-arrow-wrapper{transition:none}.mat-select-arrow{width:0;height:0;border-left:5px solid transparent;border-right:5px solid transparent;border-top:5px solid;margin:0 4px}.mat-select-panel{min-width:112px;max-width:280px;overflow:auto;-webkit-overflow-scrolling:touch;padding-top:0;padding-bottom:0;max-height:256px;min-width:100%;border-radius:4px}@media (-ms-high-contrast:active){.mat-select-panel{outline:solid 1px}}.mat-select-panel .mat-optgroup-label,.mat-select-panel .mat-option{font-size:inherit;line-height:3em;height:3em}.mat-form-field-type-mat-select:not(.mat-form-field-disabled) .mat-form-field-flex{cursor:pointer}.mat-form-field-type-mat-select .mat-form-field-label{width:calc(100% - 18px)}.mat-select-placeholder{transition:color .4s .133s cubic-bezier(.25,.8,.25,1)}._mat-animation-noopable .mat-select-placeholder{transition:none}.mat-form-field-hide-placeholder .mat-select-placeholder{color:transparent;-webkit-text-fill-color:transparent;transition:none;display:block}"],inputs:["disabled","disableRipple","tabIndex"],encapsulation:p.ViewEncapsulation.None,changeDetection:p.ChangeDetectionStrategy.OnPush,host:{role:"listbox","[attr.id]":"id","[attr.tabindex]":"tabIndex","[attr.aria-label]":"_getAriaLabel()","[attr.aria-labelledby]":"_getAriaLabelledby()","[attr.aria-required]":"required.toString()","[attr.aria-disabled]":"disabled.toString()","[attr.aria-invalid]":"errorState","[attr.aria-owns]":"panelOpen ? _optionIds : null","[attr.aria-multiselectable]":"multiple","[attr.aria-describedby]":"_ariaDescribedby || null","[attr.aria-activedescendant]":"_getAriaActiveDescendant()","[class.mat-select-disabled]":"disabled","[class.mat-select-invalid]":"errorState","[class.mat-select-required]":"required","[class.mat-select-empty]":"empty",class:"mat-select","(keydown)":"_handleKeydown($event)","(focus)":"_onFocus()","(blur)":"_onBlur()"},animations:[v.transformPanel],providers:[{provide:m.MatFormFieldControl,useExisting:e},{provide:h.MAT_OPTION_PARENT_COMPONENT,useExisting:e}]}]}],e.ctorParameters=function(){return[{type:d.ViewportRuler},{type:p.ChangeDetectorRef},{type:p.NgZone},{type:h.ErrorStateMatcher},{type:p.ElementRef},{type:a.Directionality,decorators:[{type:p.Optional}]},{type:f.NgForm,decorators:[{type:p.Optional}]},{type:f.FormGroupDirective,decorators:[{type:p.Optional}]},{type:m.MatFormField,decorators:[{type:p.Optional}]},{type:f.NgControl,decorators:[{type:p.Self},{type:p.Optional}]},{type:String,decorators:[{type:p.Attribute,args:["tabindex"]}]},{type:void 0,decorators:[{type:p.Inject,args:[T]}]},{type:o.LiveAnnouncer}]},e.propDecorators={trigger:[{type:p.ViewChild,args:["trigger"]}],panel:[{type:p.ViewChild,args:["panel"]}],overlayDir:[{type:p.ViewChild,args:[l.CdkConnectedOverlay]}],options:[{type:p.ContentChildren,args:[h.MatOption,{descendants:!0}]}],optionGroups:[{type:p.ContentChildren,args:[h.MatOptgroup]}],panelClass:[{type:p.Input}],customTrigger:[{type:p.ContentChild,args:[R]}],placeholder:[{type:p.Input}],required:[{type:p.Input}],multiple:[{type:p.Input}],disableOptionCentering:[{type:p.Input}],compareWith:[{type:p.Input}],value:[{type:p.Input}],ariaLabel:[{type:p.Input,args:["aria-label"]}],ariaLabelledby:[{type:p.Input,args:["aria-labelledby"]}],errorStateMatcher:[{type:p.Input}],sortComparator:[{type:p.Input}],id:[{type:p.Input}],openedChange:[{type:p.Output}],_openedStream:[{type:p.Output,args:["opened"]}],_closedStream:[{type:p.Output,args:["closed"]}],selectionChange:[{type:p.Output}],valueChange:[{type:p.Output}]},e}(P),F=function(){function t(){}return t.decorators=[{type:p.NgModule,args:[{imports:[_.CommonModule,l.OverlayModule,h.MatOptionModule,h.MatCommonModule],exports:[m.MatFormFieldModule,N,R,h.MatOptionModule,h.MatCommonModule],declarations:[N,R],providers:[A]}]}],t}()},"./node_modules/@angular/material/esm5/sidenav.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatSidenavModule",function(){return S}),n.d(e,"throwMatDuplicatedDrawerError",function(){return _}),n.d(e,"MAT_DRAWER_DEFAULT_AUTOSIZE_FACTORY",function(){return b}),n.d(e,"MAT_DRAWER_DEFAULT_AUTOSIZE",function(){return v}),n.d(e,"MatDrawerContent",function(){return x}),n.d(e,"MatDrawer",function(){return w}),n.d(e,"MatDrawerContainer",function(){return j}),n.d(e,"MatSidenavContent",function(){return O}),n.d(e,"MatSidenav",function(){return C}),n.d(e,"MatSidenavContainer",function(){return E}),n.d(e,"matDrawerAnimations",function(){return y});var r=n("./node_modules/@angular/animations/fesm5/animations.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),a=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),c=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),l=n("./node_modules/@angular/cdk/esm5/scrolling.es5.js"),d=n("./node_modules/@angular/common/fesm5/common.js"),p=n("./node_modules/@angular/core/fesm5/core.js"),f=n("./node_modules/rxjs/_esm5/index.js"),h=n("./node_modules/rxjs/_esm5/operators/index.js"),m=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),g=n("./node_modules/@angular/material/esm5/core.es5.js"),y={transformDrawer:Object(r.trigger)("transform",[Object(r.state)("open, open-instant",Object(r.style)({transform:"none",visibility:"visible"})),Object(r.state)("void",Object(r.style)({"box-shadow":"none",visibility:"hidden"})),Object(r.transition)("void => open-instant",Object(r.animate)("0ms")),Object(r.transition)("void <=> open, open-instant => void",Object(r.animate)("400ms cubic-bezier(0.25, 0.8, 0.25, 1)"))])};function _(t){throw Error("A drawer was already declared for 'position=\""+t+"\"'")}var v=new p.InjectionToken("MAT_DRAWER_DEFAULT_AUTOSIZE",{providedIn:"root",factory:b});function b(){return!1}var x=function(t){function e(e,n,r,i,o){var a=t.call(this,r,i,o)||this;return a._changeDetectorRef=e,a._container=n,a}return Object(i.__extends)(e,t),e.prototype.ngAfterContentInit=function(){var t=this;this._container._contentMarginChanges.subscribe(function(){t._changeDetectorRef.markForCheck()})},e.decorators=[{type:p.Component,args:[{selector:"mat-drawer-content",template:"<ng-content></ng-content>",host:{class:"mat-drawer-content","[style.margin-left.px]":"_container._contentMargins.left","[style.margin-right.px]":"_container._contentMargins.right"},changeDetection:p.ChangeDetectionStrategy.OnPush,encapsulation:p.ViewEncapsulation.None}]}],e.ctorParameters=function(){return[{type:p.ChangeDetectorRef},{type:j,decorators:[{type:p.Inject,args:[Object(p.forwardRef)(function(){return j})]}]},{type:p.ElementRef},{type:l.ScrollDispatcher},{type:p.NgZone}]},e}(l.CdkScrollable),w=function(){function t(t,e,n,r,i,o){var a=this;this._elementRef=t,this._focusTrapFactory=e,this._focusMonitor=n,this._platform=r,this._ngZone=i,this._doc=o,this._elementFocusedBeforeDrawerWasOpened=null,this._enableAnimations=!1,this._position="start",this._mode="over",this._disableClose=!1,this._autoFocus=!0,this._animationStarted=new f.Subject,this._animationEnd=new f.Subject,this._animationState="void",this.openedChange=new p.EventEmitter(!0),this._destroyed=new f.Subject,this.onPositionChanged=new p.EventEmitter,this._modeChanged=new f.Subject,this._opened=!1,this.openedChange.subscribe(function(t){t?(a._doc&&(a._elementFocusedBeforeDrawerWasOpened=a._doc.activeElement),a._isFocusTrapEnabled&&a._focusTrap&&a._trapFocus()):a._restoreFocus()}),this._ngZone.runOutsideAngular(function(){Object(f.fromEvent)(a._elementRef.nativeElement,"keydown").pipe(Object(h.filter)(function(t){return t.keyCode===u.ESCAPE&&!a.disableClose}),Object(h.takeUntil)(a._destroyed)).subscribe(function(t){return a._ngZone.run(function(){a.close(),t.stopPropagation()})})}),this._animationEnd.pipe(Object(h.distinctUntilChanged)(function(t,e){return t.fromState===e.fromState&&t.toState===e.toState})).subscribe(function(t){var e=t.fromState,n=t.toState;(0===n.indexOf("open")&&"void"===e||"void"===n&&0===e.indexOf("open"))&&a.openedChange.emit(a._opened)})}return Object.defineProperty(t.prototype,"position",{get:function(){return this._position},set:function(t){(t="end"===t?"end":"start")!=this._position&&(this._position=t,this.onPositionChanged.emit())},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"mode",{get:function(){return this._mode},set:function(t){this._mode=t,this._modeChanged.next()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disableClose",{get:function(){return this._disableClose},set:function(t){this._disableClose=Object(s.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"autoFocus",{get:function(){return this._autoFocus},set:function(t){this._autoFocus=Object(s.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_openedStream",{get:function(){return this.openedChange.pipe(Object(h.filter)(function(t){return t}),Object(h.map)(function(){}))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"openedStart",{get:function(){return this._animationStarted.pipe(Object(h.filter)(function(t){return t.fromState!==t.toState&&0===t.toState.indexOf("open")}),Object(h.map)(function(){}))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_closedStream",{get:function(){return this.openedChange.pipe(Object(h.filter)(function(t){return!t}),Object(h.map)(function(){}))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"closedStart",{get:function(){return this._animationStarted.pipe(Object(h.filter)(function(t){return t.fromState!==t.toState&&"void"===t.toState}),Object(h.map)(function(){}))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_isFocusTrapEnabled",{get:function(){return this.opened&&"side"!==this.mode},enumerable:!0,configurable:!0}),t.prototype._trapFocus=function(){var t=this;this.autoFocus&&this._focusTrap.focusInitialElementWhenReady().then(function(e){e||"function"!=typeof t._elementRef.nativeElement.focus||t._elementRef.nativeElement.focus()})},t.prototype._restoreFocus=function(){if(this.autoFocus){var t=this._doc&&this._doc.activeElement;t&&this._elementRef.nativeElement.contains(t)&&(this._elementFocusedBeforeDrawerWasOpened instanceof HTMLElement?this._focusMonitor.focusVia(this._elementFocusedBeforeDrawerWasOpened,this._openedVia):this._elementRef.nativeElement.blur()),this._elementFocusedBeforeDrawerWasOpened=null,this._openedVia=null}},t.prototype.ngAfterContentInit=function(){this._focusTrap=this._focusTrapFactory.create(this._elementRef.nativeElement),this._focusTrap.enabled=this._isFocusTrapEnabled},t.prototype.ngAfterContentChecked=function(){this._platform.isBrowser&&(this._enableAnimations=!0)},t.prototype.ngOnDestroy=function(){this._focusTrap&&this._focusTrap.destroy(),this._animationStarted.complete(),this._animationEnd.complete(),this._destroyed.next(),this._destroyed.complete()},Object.defineProperty(t.prototype,"opened",{get:function(){return this._opened},set:function(t){this.toggle(Object(s.coerceBooleanProperty)(t))},enumerable:!0,configurable:!0}),t.prototype.open=function(t){return this.toggle(!0,t)},t.prototype.close=function(){return this.toggle(!1)},t.prototype.toggle=function(t,e){var n=this;return void 0===t&&(t=!this.opened),void 0===e&&(e="program"),this._opened=t,t?(this._animationState=this._enableAnimations?"open":"open-instant",this._openedVia=e):(this._animationState="void",this._restoreFocus()),this._focusTrap&&(this._focusTrap.enabled=this._isFocusTrapEnabled),new Promise(function(t){n.openedChange.pipe(Object(h.take)(1)).subscribe(function(e){return t(e?"open":"close")})})},Object.defineProperty(t.prototype,"_width",{get:function(){return this._elementRef.nativeElement?this._elementRef.nativeElement.offsetWidth||0:0},enumerable:!0,configurable:!0}),t.decorators=[{type:p.Component,args:[{selector:"mat-drawer",exportAs:"matDrawer",template:'<div class="mat-drawer-inner-container"><ng-content></ng-content></div>',animations:[y.transformDrawer],host:{class:"mat-drawer","[@transform]":"_animationState","(@transform.start)":"_animationStarted.next($event)","(@transform.done)":"_animationEnd.next($event)","[attr.align]":"null","[class.mat-drawer-end]":'position === "end"',"[class.mat-drawer-over]":'mode === "over"',"[class.mat-drawer-push]":'mode === "push"',"[class.mat-drawer-side]":'mode === "side"',tabIndex:"-1"},changeDetection:p.ChangeDetectionStrategy.OnPush,encapsulation:p.ViewEncapsulation.None}]}],t.ctorParameters=function(){return[{type:p.ElementRef},{type:o.FocusTrapFactory},{type:o.FocusMonitor},{type:c.Platform},{type:p.NgZone},{type:void 0,decorators:[{type:p.Optional},{type:p.Inject,args:[d.DOCUMENT]}]}]},t.propDecorators={position:[{type:p.Input}],mode:[{type:p.Input}],disableClose:[{type:p.Input}],autoFocus:[{type:p.Input}],openedChange:[{type:p.Output}],_openedStream:[{type:p.Output,args:["opened"]}],openedStart:[{type:p.Output}],_closedStream:[{type:p.Output,args:["closed"]}],closedStart:[{type:p.Output}],onPositionChanged:[{type:p.Output,args:["positionChanged"]}],opened:[{type:p.Input}]},t}(),j=function(){function t(t,e,n,r,i,o,a){void 0===i&&(i=!1);var s=this;this._dir=t,this._element=e,this._ngZone=n,this._changeDetectorRef=r,this._animationMode=o,this.backdropClick=new p.EventEmitter,this._destroyed=new f.Subject,this._doCheckSubject=new f.Subject,this._contentMargins={left:null,right:null},this._contentMarginChanges=new f.Subject,t&&t.change.pipe(Object(h.takeUntil)(this._destroyed)).subscribe(function(){s._validateDrawers(),s._updateContentMargins()}),a&&a.change().pipe(Object(h.takeUntil)(this._destroyed)).subscribe(function(){return s._updateContentMargins()}),this._autosize=i}return Object.defineProperty(t.prototype,"start",{get:function(){return this._start},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"end",{get:function(){return this._end},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"autosize",{get:function(){return this._autosize},set:function(t){this._autosize=Object(s.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"hasBackdrop",{get:function(){return null==this._backdropOverride?!this._start||"side"!==this._start.mode||!this._end||"side"!==this._end.mode:this._backdropOverride},set:function(t){this._backdropOverride=null==t?null:Object(s.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"scrollable",{get:function(){return this._userContent||this._content},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){var t=this;this._drawers.changes.pipe(Object(h.startWith)(null)).subscribe(function(){t._validateDrawers(),t._drawers.forEach(function(e){t._watchDrawerToggle(e),t._watchDrawerPosition(e),t._watchDrawerMode(e)}),(!t._drawers.length||t._isDrawerOpen(t._start)||t._isDrawerOpen(t._end))&&t._updateContentMargins(),t._changeDetectorRef.markForCheck()}),this._doCheckSubject.pipe(Object(h.debounceTime)(10),Object(h.takeUntil)(this._destroyed)).subscribe(function(){return t._updateContentMargins()})},t.prototype.ngOnDestroy=function(){this._doCheckSubject.complete(),this._destroyed.next(),this._destroyed.complete()},t.prototype.open=function(){this._drawers.forEach(function(t){return t.open()})},t.prototype.close=function(){this._drawers.forEach(function(t){return t.close()})},t.prototype.ngDoCheck=function(){var t=this;this._autosize&&this._isPushed()&&this._ngZone.runOutsideAngular(function(){return t._doCheckSubject.next()})},t.prototype._watchDrawerToggle=function(t){var e=this;t._animationStarted.pipe(Object(h.filter)(function(t){return t.fromState!==t.toState}),Object(h.takeUntil)(this._drawers.changes)).subscribe(function(t){"open-instant"!==t.toState&&"NoopAnimations"!==e._animationMode&&e._element.nativeElement.classList.add("mat-drawer-transition"),e._updateContentMargins(),e._changeDetectorRef.markForCheck()}),"side"!==t.mode&&t.openedChange.pipe(Object(h.takeUntil)(this._drawers.changes)).subscribe(function(){return e._setContainerClass(t.opened)})},t.prototype._watchDrawerPosition=function(t){var e=this;t&&t.onPositionChanged.pipe(Object(h.takeUntil)(this._drawers.changes)).subscribe(function(){e._ngZone.onMicrotaskEmpty.asObservable().pipe(Object(h.take)(1)).subscribe(function(){e._validateDrawers()})})},t.prototype._watchDrawerMode=function(t){var e=this;t&&t._modeChanged.pipe(Object(h.takeUntil)(Object(f.merge)(this._drawers.changes,this._destroyed))).subscribe(function(){e._updateContentMargins(),e._changeDetectorRef.markForCheck()})},t.prototype._setContainerClass=function(t){t?this._element.nativeElement.classList.add("mat-drawer-opened"):this._element.nativeElement.classList.remove("mat-drawer-opened")},t.prototype._validateDrawers=function(){var t=this;this._start=this._end=null,this._drawers.forEach(function(e){"end"==e.position?(null!=t._end&&_("end"),t._end=e):(null!=t._start&&_("start"),t._start=e)}),this._right=this._left=null,this._dir&&"rtl"===this._dir.value?(this._left=this._end,this._right=this._start):(this._left=this._start,this._right=this._end)},t.prototype._isPushed=function(){return this._isDrawerOpen(this._start)&&"over"!=this._start.mode||this._isDrawerOpen(this._end)&&"over"!=this._end.mode},t.prototype._onBackdropClicked=function(){this.backdropClick.emit(),this._closeModalDrawer()},t.prototype._closeModalDrawer=function(){var t=this;[this._start,this._end].filter(function(e){return e&&!e.disableClose&&t._canHaveBackdrop(e)}).forEach(function(t){return t.close()})},t.prototype._isShowingBackdrop=function(){return this._isDrawerOpen(this._start)&&this._canHaveBackdrop(this._start)||this._isDrawerOpen(this._end)&&this._canHaveBackdrop(this._end)},t.prototype._canHaveBackdrop=function(t){return"side"!==t.mode||!!this._backdropOverride},t.prototype._isDrawerOpen=function(t){return null!=t&&t.opened},t.prototype._updateContentMargins=function(){var t=this,e=0,n=0;if(this._left&&this._left.opened)if("side"==this._left.mode)e+=this._left._width;else if("push"==this._left.mode){e+=r=this._left._width,n-=r}if(this._right&&this._right.opened)if("side"==this._right.mode)n+=this._right._width;else if("push"==this._right.mode){var r;n+=r=this._right._width,e-=r}e=e||null,n=n||null,e===this._contentMargins.left&&n===this._contentMargins.right||(this._contentMargins={left:e,right:n},this._ngZone.run(function(){return t._contentMarginChanges.next(t._contentMargins)}))},t.decorators=[{type:p.Component,args:[{selector:"mat-drawer-container",exportAs:"matDrawerContainer",template:'<div class="mat-drawer-backdrop" (click)="_onBackdropClicked()" *ngIf="hasBackdrop" [class.mat-drawer-shown]="_isShowingBackdrop()"></div><ng-content select="mat-drawer"></ng-content><ng-content select="mat-drawer-content"></ng-content><mat-drawer-content *ngIf="!_content"><ng-content></ng-content></mat-drawer-content>',styles:[".mat-drawer-container{position:relative;z-index:1;box-sizing:border-box;-webkit-overflow-scrolling:touch;display:block;overflow:hidden}.mat-drawer-container[fullscreen]{top:0;left:0;right:0;bottom:0;position:absolute}.mat-drawer-container[fullscreen].mat-drawer-opened{overflow:hidden}.mat-drawer-container.mat-drawer-container-explicit-backdrop .mat-drawer-side{z-index:3}.mat-drawer-container.ng-animate-disabled .mat-drawer-backdrop,.mat-drawer-container.ng-animate-disabled .mat-drawer-content,.ng-animate-disabled .mat-drawer-container .mat-drawer-backdrop,.ng-animate-disabled .mat-drawer-container .mat-drawer-content{transition:none}.mat-drawer-backdrop{top:0;left:0;right:0;bottom:0;position:absolute;display:block;z-index:3;visibility:hidden}.mat-drawer-backdrop.mat-drawer-shown{visibility:visible}.mat-drawer-transition .mat-drawer-backdrop{transition-duration:.4s;transition-timing-function:cubic-bezier(.25,.8,.25,1);transition-property:background-color,visibility}@media (-ms-high-contrast:active){.mat-drawer-backdrop{opacity:.5}}.mat-drawer-content{position:relative;z-index:1;display:block;height:100%;overflow:auto}.mat-drawer-transition .mat-drawer-content{transition-duration:.4s;transition-timing-function:cubic-bezier(.25,.8,.25,1);transition-property:transform,margin-left,margin-right}.mat-drawer{position:relative;z-index:4;display:block;position:absolute;top:0;bottom:0;z-index:3;outline:0;box-sizing:border-box;overflow-y:auto;transform:translate3d(-100%,0,0)}@media (-ms-high-contrast:active){.mat-drawer,[dir=rtl] .mat-drawer.mat-drawer-end{border-right:solid 1px currentColor}}@media (-ms-high-contrast:active){.mat-drawer.mat-drawer-end,[dir=rtl] .mat-drawer{border-left:solid 1px currentColor;border-right:none}}.mat-drawer.mat-drawer-side{z-index:2}.mat-drawer.mat-drawer-end{right:0;transform:translate3d(100%,0,0)}[dir=rtl] .mat-drawer{transform:translate3d(100%,0,0)}[dir=rtl] .mat-drawer.mat-drawer-end{left:0;right:auto;transform:translate3d(-100%,0,0)}.mat-drawer-inner-container{width:100%;height:100%;overflow:auto;-webkit-overflow-scrolling:touch}.mat-sidenav-fixed{position:fixed}"],host:{class:"mat-drawer-container","[class.mat-drawer-container-explicit-backdrop]":"_backdropOverride"},changeDetection:p.ChangeDetectionStrategy.OnPush,encapsulation:p.ViewEncapsulation.None}]}],t.ctorParameters=function(){return[{type:a.Directionality,decorators:[{type:p.Optional}]},{type:p.ElementRef},{type:p.NgZone},{type:p.ChangeDetectorRef},{type:void 0,decorators:[{type:p.Inject,args:[v]}]},{type:String,decorators:[{type:p.Optional},{type:p.Inject,args:[m.ANIMATION_MODULE_TYPE]}]},{type:l.ViewportRuler,decorators:[{type:p.Optional}]}]},t.propDecorators={_drawers:[{type:p.ContentChildren,args:[w]}],_content:[{type:p.ContentChild,args:[x]}],_userContent:[{type:p.ViewChild,args:[x]}],autosize:[{type:p.Input}],hasBackdrop:[{type:p.Input}],backdropClick:[{type:p.Output}]},t}(),O=function(t){function e(e,n,r,i,o){return t.call(this,e,n,r,i,o)||this}return Object(i.__extends)(e,t),e.decorators=[{type:p.Component,args:[{selector:"mat-sidenav-content",template:"<ng-content></ng-content>",host:{class:"mat-drawer-content mat-sidenav-content","[style.margin-left.px]":"_container._contentMargins.left","[style.margin-right.px]":"_container._contentMargins.right"},changeDetection:p.ChangeDetectionStrategy.OnPush,encapsulation:p.ViewEncapsulation.None}]}],e.ctorParameters=function(){return[{type:p.ChangeDetectorRef},{type:E,decorators:[{type:p.Inject,args:[Object(p.forwardRef)(function(){return E})]}]},{type:p.ElementRef},{type:l.ScrollDispatcher},{type:p.NgZone}]},e}(x),C=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._fixedInViewport=!1,e._fixedTopGap=0,e._fixedBottomGap=0,e}return Object(i.__extends)(e,t),Object.defineProperty(e.prototype,"fixedInViewport",{get:function(){return this._fixedInViewport},set:function(t){this._fixedInViewport=Object(s.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"fixedTopGap",{get:function(){return this._fixedTopGap},set:function(t){this._fixedTopGap=Object(s.coerceNumberProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"fixedBottomGap",{get:function(){return this._fixedBottomGap},set:function(t){this._fixedBottomGap=Object(s.coerceNumberProperty)(t)},enumerable:!0,configurable:!0}),e.decorators=[{type:p.Component,args:[{selector:"mat-sidenav",exportAs:"matSidenav",template:'<div class="mat-drawer-inner-container"><ng-content></ng-content></div>',animations:[y.transformDrawer],host:{class:"mat-drawer mat-sidenav",tabIndex:"-1","[@transform]":"_animationState","(@transform.start)":"_animationStarted.next($event)","(@transform.done)":"_animationEnd.next($event)","[attr.align]":"null","[class.mat-drawer-end]":'position === "end"',"[class.mat-drawer-over]":'mode === "over"',"[class.mat-drawer-push]":'mode === "push"',"[class.mat-drawer-side]":'mode === "side"',"[class.mat-sidenav-fixed]":"fixedInViewport","[style.top.px]":"fixedInViewport ? fixedTopGap : null","[style.bottom.px]":"fixedInViewport ? fixedBottomGap : null"},changeDetection:p.ChangeDetectionStrategy.OnPush,encapsulation:p.ViewEncapsulation.None}]}],e.propDecorators={fixedInViewport:[{type:p.Input}],fixedTopGap:[{type:p.Input}],fixedBottomGap:[{type:p.Input}]},e}(w),E=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(i.__extends)(e,t),e.decorators=[{type:p.Component,args:[{selector:"mat-sidenav-container",exportAs:"matSidenavContainer",template:'<div class="mat-drawer-backdrop" (click)="_onBackdropClicked()" *ngIf="hasBackdrop" [class.mat-drawer-shown]="_isShowingBackdrop()"></div><ng-content select="mat-sidenav"></ng-content><ng-content select="mat-sidenav-content"></ng-content><mat-sidenav-content *ngIf="!_content" cdkScrollable><ng-content></ng-content></mat-sidenav-content>',styles:[".mat-drawer-container{position:relative;z-index:1;box-sizing:border-box;-webkit-overflow-scrolling:touch;display:block;overflow:hidden}.mat-drawer-container[fullscreen]{top:0;left:0;right:0;bottom:0;position:absolute}.mat-drawer-container[fullscreen].mat-drawer-opened{overflow:hidden}.mat-drawer-container.mat-drawer-container-explicit-backdrop .mat-drawer-side{z-index:3}.mat-drawer-container.ng-animate-disabled .mat-drawer-backdrop,.mat-drawer-container.ng-animate-disabled .mat-drawer-content,.ng-animate-disabled .mat-drawer-container .mat-drawer-backdrop,.ng-animate-disabled .mat-drawer-container .mat-drawer-content{transition:none}.mat-drawer-backdrop{top:0;left:0;right:0;bottom:0;position:absolute;display:block;z-index:3;visibility:hidden}.mat-drawer-backdrop.mat-drawer-shown{visibility:visible}.mat-drawer-transition .mat-drawer-backdrop{transition-duration:.4s;transition-timing-function:cubic-bezier(.25,.8,.25,1);transition-property:background-color,visibility}@media (-ms-high-contrast:active){.mat-drawer-backdrop{opacity:.5}}.mat-drawer-content{position:relative;z-index:1;display:block;height:100%;overflow:auto}.mat-drawer-transition .mat-drawer-content{transition-duration:.4s;transition-timing-function:cubic-bezier(.25,.8,.25,1);transition-property:transform,margin-left,margin-right}.mat-drawer{position:relative;z-index:4;display:block;position:absolute;top:0;bottom:0;z-index:3;outline:0;box-sizing:border-box;overflow-y:auto;transform:translate3d(-100%,0,0)}@media (-ms-high-contrast:active){.mat-drawer,[dir=rtl] .mat-drawer.mat-drawer-end{border-right:solid 1px currentColor}}@media (-ms-high-contrast:active){.mat-drawer.mat-drawer-end,[dir=rtl] .mat-drawer{border-left:solid 1px currentColor;border-right:none}}.mat-drawer.mat-drawer-side{z-index:2}.mat-drawer.mat-drawer-end{right:0;transform:translate3d(100%,0,0)}[dir=rtl] .mat-drawer{transform:translate3d(100%,0,0)}[dir=rtl] .mat-drawer.mat-drawer-end{left:0;right:auto;transform:translate3d(-100%,0,0)}.mat-drawer-inner-container{width:100%;height:100%;overflow:auto;-webkit-overflow-scrolling:touch}.mat-sidenav-fixed{position:fixed}"],host:{class:"mat-drawer-container mat-sidenav-container","[class.mat-drawer-container-explicit-backdrop]":"_backdropOverride"},changeDetection:p.ChangeDetectionStrategy.OnPush,encapsulation:p.ViewEncapsulation.None}]}],e.propDecorators={_drawers:[{type:p.ContentChildren,args:[C]}],_content:[{type:p.ContentChild,args:[O]}]},e}(j),S=function(){function t(){}return t.decorators=[{type:p.NgModule,args:[{imports:[d.CommonModule,g.MatCommonModule,l.ScrollingModule,c.PlatformModule],exports:[g.MatCommonModule,w,j,x,C,E,O],declarations:[w,j,x,C,E,O]}]}],t}()},"./node_modules/@angular/material/esm5/slide-toggle.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatSlideToggleModule",function(){return x}),n.d(e,"MAT_SLIDE_TOGGLE_VALUE_ACCESSOR",function(){return g}),n.d(e,"MatSlideToggleChange",function(){return y}),n.d(e,"MatSlideToggleBase",function(){return _}),n.d(e,"_MatSlideToggleMixinBase",function(){return v}),n.d(e,"MatSlideToggle",function(){return b}),n.d(e,"MAT_SLIDE_TOGGLE_DEFAULT_OPTIONS",function(){return h});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),a=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),c=n("./node_modules/@angular/forms/fesm5/forms.js"),l=n("./node_modules/@angular/material/esm5/core.es5.js"),d=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),p=n("./node_modules/@angular/cdk/esm5/observers.es5.js"),f=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),h=new r.InjectionToken("mat-slide-toggle-default-options",{providedIn:"root",factory:function(){return{disableToggleValue:!1,disableDragValue:!1}}}),m=0,g={provide:c.NG_VALUE_ACCESSOR,useExisting:Object(r.forwardRef)(function(){return b}),multi:!0},y=function(){return function(t,e){this.source=t,this.checked=e}}(),_=function(){return function(t){this._elementRef=t}}(),v=Object(l.mixinTabIndex)(Object(l.mixinColor)(Object(l.mixinDisableRipple)(Object(l.mixinDisabled)(_)),"accent")),b=function(t){function e(e,n,i,o,a,s,u,c,l){var d=t.call(this,e)||this;return d._focusMonitor=i,d._changeDetectorRef=o,d._ngZone=s,d.defaults=u,d._animationMode=c,d._dir=l,d.onChange=function(t){},d.onTouched=function(){},d._uniqueId="mat-slide-toggle-"+ ++m,d._required=!1,d._checked=!1,d._dragging=!1,d.name=null,d.id=d._uniqueId,d.labelPosition="after",d.ariaLabel=null,d.ariaLabelledby=null,d.change=new r.EventEmitter,d.toggleChange=new r.EventEmitter,d.dragChange=new r.EventEmitter,d.tabIndex=parseInt(a)||0,d}return Object(i.__extends)(e,t),Object.defineProperty(e.prototype,"required",{get:function(){return this._required},set:function(t){this._required=Object(s.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"checked",{get:function(){return this._checked},set:function(t){this._checked=Object(s.coerceBooleanProperty)(t),this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"inputId",{get:function(){return(this.id||this._uniqueId)+"-input"},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){var t=this;this._focusMonitor.monitor(this._elementRef,!0).subscribe(function(e){e||Promise.resolve().then(function(){return t.onTouched()})})},e.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef)},e.prototype._onChangeEvent=function(t){t.stopPropagation(),this._dragging||this.toggleChange.emit(),this._dragging||this.defaults.disableToggleValue?this._inputElement.nativeElement.checked=this.checked:(this.checked=this._inputElement.nativeElement.checked,this._emitChangeEvent())},e.prototype._onInputClick=function(t){t.stopPropagation()},e.prototype.writeValue=function(t){this.checked=!!t},e.prototype.registerOnChange=function(t){this.onChange=t},e.prototype.registerOnTouched=function(t){this.onTouched=t},e.prototype.setDisabledState=function(t){this.disabled=t,this._changeDetectorRef.markForCheck()},e.prototype.focus=function(){this._focusMonitor.focusVia(this._inputElement,"keyboard")},e.prototype.toggle=function(){this.checked=!this.checked,this.onChange(this.checked)},e.prototype._emitChangeEvent=function(){this.onChange(this.checked),this.change.emit(new y(this,this.checked))},e.prototype._getDragPercentage=function(t){var e=t/this._thumbBarWidth*100;return this._previousChecked&&(e+=100),Math.max(0,Math.min(e,100))},e.prototype._onDragStart=function(){if(!this.disabled&&!this._dragging){var t=this._thumbEl.nativeElement;this._thumbBarWidth=this._thumbBarEl.nativeElement.clientWidth-t.clientWidth,t.classList.add("mat-dragging"),this._previousChecked=this.checked,this._dragging=!0}},e.prototype._onDrag=function(t){if(this._dragging){var e=this._dir&&"rtl"===this._dir.value?-1:1;this._dragPercentage=this._getDragPercentage(t.deltaX*e);var n=this._dragPercentage/100*this._thumbBarWidth*e;this._thumbEl.nativeElement.style.transform="translate3d("+n+"px, 0, 0)"}},e.prototype._onDragEnd=function(){var t=this;if(this._dragging){var e=this._dragPercentage>50;e!==this.checked&&(this.dragChange.emit(),this.defaults.disableDragValue||(this.checked=e,this._emitChangeEvent())),this._ngZone.runOutsideAngular(function(){return setTimeout(function(){t._dragging&&(t._dragging=!1,t._thumbEl.nativeElement.classList.remove("mat-dragging"),t._thumbEl.nativeElement.style.transform="")})})}},e.prototype._onLabelTextChange=function(){this._changeDetectorRef.detectChanges()},e.decorators=[{type:r.Component,args:[{selector:"mat-slide-toggle",exportAs:"matSlideToggle",host:{class:"mat-slide-toggle","[id]":"id","[attr.tabindex]":"-1","[class.mat-checked]":"checked","[class.mat-disabled]":"disabled","[class.mat-slide-toggle-label-before]":'labelPosition == "before"',"[class._mat-animation-noopable]":'_animationMode === "NoopAnimations"',"(focus)":"_inputElement.nativeElement.focus()"},template:'<label [attr.for]="inputId" class="mat-slide-toggle-label" #label><div #toggleBar class="mat-slide-toggle-bar" [class.mat-slide-toggle-bar-no-side-margin]="!labelContent.textContent || !labelContent.textContent.trim()"><input #input class="mat-slide-toggle-input cdk-visually-hidden" type="checkbox" [id]="inputId" [required]="required" [tabIndex]="tabIndex" [checked]="checked" [disabled]="disabled" [attr.name]="name" [attr.aria-label]="ariaLabel" [attr.aria-labelledby]="ariaLabelledby" (change)="_onChangeEvent($event)" (click)="_onInputClick($event)"><div class="mat-slide-toggle-thumb-container" #thumbContainer (slidestart)="_onDragStart()" (slide)="_onDrag($event)" (slideend)="_onDragEnd()"><div class="mat-slide-toggle-thumb"></div><div class="mat-slide-toggle-ripple" mat-ripple [matRippleTrigger]="label" [matRippleDisabled]="disableRipple || disabled" [matRippleCentered]="true" [matRippleRadius]="20" [matRippleAnimation]="{enterDuration: 150}"><div class="mat-ripple-element mat-slide-toggle-persistent-ripple"></div></div></div></div><span class="mat-slide-toggle-content" #labelContent (cdkObserveContent)="_onLabelTextChange()"><span style="display:none">&nbsp;</span><ng-content></ng-content></span></label>',styles:[".mat-slide-toggle{display:inline-block;height:24px;max-width:100%;line-height:24px;white-space:nowrap;outline:0;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;-webkit-tap-highlight-color:transparent}.mat-slide-toggle.mat-checked .mat-slide-toggle-thumb-container{transform:translate3d(16px,0,0)}[dir=rtl] .mat-slide-toggle.mat-checked .mat-slide-toggle-thumb-container{transform:translate3d(-16px,0,0)}.mat-slide-toggle.mat-disabled{opacity:.38}.mat-slide-toggle.mat-disabled .mat-slide-toggle-label,.mat-slide-toggle.mat-disabled .mat-slide-toggle-thumb-container{cursor:default}.mat-slide-toggle-label{display:flex;flex:1;flex-direction:row;align-items:center;height:inherit;cursor:pointer}.mat-slide-toggle-content{white-space:nowrap;overflow:hidden;text-overflow:ellipsis}.mat-slide-toggle-label-before .mat-slide-toggle-label{order:1}.mat-slide-toggle-label-before .mat-slide-toggle-bar{order:2}.mat-slide-toggle-bar,[dir=rtl] .mat-slide-toggle-label-before .mat-slide-toggle-bar{margin-right:8px;margin-left:0}.mat-slide-toggle-label-before .mat-slide-toggle-bar,[dir=rtl] .mat-slide-toggle-bar{margin-left:8px;margin-right:0}.mat-slide-toggle-bar-no-side-margin{margin-left:0;margin-right:0}.mat-slide-toggle-thumb-container{position:absolute;z-index:1;width:20px;height:20px;top:-3px;left:0;transform:translate3d(0,0,0);transition:all 80ms linear;transition-property:transform;cursor:-webkit-grab;cursor:grab}.mat-slide-toggle-thumb-container.mat-dragging,.mat-slide-toggle-thumb-container:active{cursor:-webkit-grabbing;cursor:grabbing;transition-duration:0s}._mat-animation-noopable .mat-slide-toggle-thumb-container{transition:none}[dir=rtl] .mat-slide-toggle-thumb-container{left:auto;right:0}.mat-slide-toggle-thumb{height:20px;width:20px;border-radius:50%}.mat-slide-toggle-bar{position:relative;width:36px;height:14px;flex-shrink:0;border-radius:8px}.mat-slide-toggle-input{bottom:0;left:10px}[dir=rtl] .mat-slide-toggle-input{left:auto;right:10px}.mat-slide-toggle-bar,.mat-slide-toggle-thumb{transition:all 80ms linear;transition-property:background-color;transition-delay:50ms}._mat-animation-noopable .mat-slide-toggle-bar,._mat-animation-noopable .mat-slide-toggle-thumb{transition:none}.mat-slide-toggle .mat-slide-toggle-ripple{position:absolute;top:calc(50% - 20px);left:calc(50% - 20px);height:40px;width:40px;z-index:1;pointer-events:none}.mat-slide-toggle .mat-slide-toggle-ripple .mat-ripple-element:not(.mat-slide-toggle-persistent-ripple){opacity:.12}.mat-slide-toggle-persistent-ripple{width:100%;height:100%;transform:none}.mat-slide-toggle-bar:hover .mat-slide-toggle-persistent-ripple{opacity:.04}.mat-slide-toggle:not(.mat-disabled).cdk-focused .mat-slide-toggle-persistent-ripple{opacity:.12}.mat-slide-toggle-persistent-ripple,.mat-slide-toggle.mat-disabled .mat-slide-toggle-bar:hover .mat-slide-toggle-persistent-ripple{opacity:0}@media (hover:none){.mat-slide-toggle-bar:hover .mat-slide-toggle-persistent-ripple{display:none}}@media (-ms-high-contrast:active){.mat-slide-toggle-thumb{background:#fff;border:1px solid #000}.mat-slide-toggle.mat-checked .mat-slide-toggle-thumb{background:#000;border:1px solid #fff}.mat-slide-toggle-bar{background:#fff}.mat-slide-toggle.cdk-keyboard-focused .mat-slide-toggle-bar{outline:1px dotted;outline-offset:5px}}@media (-ms-high-contrast:black-on-white){.mat-slide-toggle-bar{border:1px solid #000}}"],providers:[g],inputs:["disabled","disableRipple","color","tabIndex"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:u.Platform},{type:o.FocusMonitor},{type:r.ChangeDetectorRef},{type:String,decorators:[{type:r.Attribute,args:["tabindex"]}]},{type:r.NgZone},{type:void 0,decorators:[{type:r.Inject,args:[h]}]},{type:String,decorators:[{type:r.Optional},{type:r.Inject,args:[d.ANIMATION_MODULE_TYPE]}]},{type:a.Directionality,decorators:[{type:r.Optional}]}]},e.propDecorators={_thumbEl:[{type:r.ViewChild,args:["thumbContainer"]}],_thumbBarEl:[{type:r.ViewChild,args:["toggleBar"]}],name:[{type:r.Input}],id:[{type:r.Input}],labelPosition:[{type:r.Input}],ariaLabel:[{type:r.Input,args:["aria-label"]}],ariaLabelledby:[{type:r.Input,args:["aria-labelledby"]}],required:[{type:r.Input}],checked:[{type:r.Input}],change:[{type:r.Output}],toggleChange:[{type:r.Output}],dragChange:[{type:r.Output}],_inputElement:[{type:r.ViewChild,args:["input"]}]},e}(v),x=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[l.MatRippleModule,l.MatCommonModule,p.ObserversModule],exports:[b,l.MatCommonModule],declarations:[b],providers:[{provide:f.HAMMER_GESTURE_CONFIG,useClass:l.GestureConfig}]}]}],t}()},"./node_modules/@angular/material/esm5/slider.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatSliderModule",function(){return b}),n.d(e,"MAT_SLIDER_VALUE_ACCESSOR",function(){return m}),n.d(e,"MatSliderChange",function(){return g}),n.d(e,"MatSliderBase",function(){return y}),n.d(e,"_MatSliderMixinBase",function(){return _}),n.d(e,"MatSlider",function(){return v});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),o=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),a=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),s=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),u=n("./node_modules/@angular/core/fesm5/core.js"),c=n("./node_modules/@angular/forms/fesm5/forms.js"),l=n("./node_modules/@angular/material/esm5/core.es5.js"),d=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),p=n("./node_modules/rxjs/_esm5/index.js"),f=n("./node_modules/@angular/common/fesm5/common.js"),h=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),m={provide:c.NG_VALUE_ACCESSOR,useExisting:Object(u.forwardRef)(function(){return v}),multi:!0},g=function(){return function(){}}(),y=function(){return function(t){this._elementRef=t}}(),_=Object(l.mixinTabIndex)(Object(l.mixinColor)(Object(l.mixinDisabled)(y),"accent")),v=function(t){function e(e,n,r,i,o,a){var s=t.call(this,e)||this;return s._focusMonitor=n,s._changeDetectorRef=r,s._dir=i,s._animationMode=a,s._invert=!1,s._max=100,s._min=0,s._step=1,s._thumbLabel=!1,s._tickInterval=0,s._value=null,s._vertical=!1,s.change=new u.EventEmitter,s.input=new u.EventEmitter,s.valueChange=new u.EventEmitter,s.onTouched=function(){},s._percent=0,s._isSliding=!1,s._isActive=!1,s._tickIntervalPercent=0,s._sliderDimensions=null,s._controlValueAccessorChangeFn=function(){},s._dirChangeSubscription=p.Subscription.EMPTY,s.tabIndex=parseInt(o)||0,s}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"invert",{get:function(){return this._invert},set:function(t){this._invert=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"max",{get:function(){return this._max},set:function(t){this._max=Object(a.coerceNumberProperty)(t,this._max),this._percent=this._calculatePercentage(this._value),this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"min",{get:function(){return this._min},set:function(t){this._min=Object(a.coerceNumberProperty)(t,this._min),null===this._value&&(this.value=this._min),this._percent=this._calculatePercentage(this._value),this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"step",{get:function(){return this._step},set:function(t){this._step=Object(a.coerceNumberProperty)(t,this._step),this._step%1!=0&&(this._roundToDecimal=this._step.toString().split(".").pop().length),this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"thumbLabel",{get:function(){return this._thumbLabel},set:function(t){this._thumbLabel=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"tickInterval",{get:function(){return this._tickInterval},set:function(t){this._tickInterval="auto"===t?"auto":"number"==typeof t||"string"==typeof t?Object(a.coerceNumberProperty)(t,this._tickInterval):0},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return null===this._value&&(this.value=this._min),this._value},set:function(t){if(t!==this._value){var e=Object(a.coerceNumberProperty)(t);this._roundToDecimal&&(e=parseFloat(e.toFixed(this._roundToDecimal))),this._value=e,this._percent=this._calculatePercentage(this._value),this._changeDetectorRef.markForCheck()}},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"vertical",{get:function(){return this._vertical},set:function(t){this._vertical=Object(a.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"displayValue",{get:function(){return this.displayWith?this.displayWith(this.value):this._roundToDecimal&&this.value&&this.value%1!=0?this.value.toFixed(this._roundToDecimal):this.value||0},enumerable:!0,configurable:!0}),e.prototype.focus=function(){this._focusHostElement()},e.prototype.blur=function(){this._blurHostElement()},Object.defineProperty(e.prototype,"percent",{get:function(){return this._clamp(this._percent)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_invertAxis",{get:function(){return this.vertical?!this.invert:this.invert},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_isMinValue",{get:function(){return 0===this.percent},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_thumbGap",{get:function(){return this.disabled?7:this._isMinValue&&!this.thumbLabel?this._isActive?10:7:0},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_trackBackgroundStyles",{get:function(){var t=this.vertical?"Y":"X",e=this.vertical?"1, "+(1-this.percent)+", 1":1-this.percent+", 1, 1";return{transform:"translate"+t+"("+(this._shouldInvertMouseCoords()?"-":"")+this._thumbGap+"px) scale3d("+e+")"}},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_trackFillStyles",{get:function(){var t=this.vertical?"Y":"X",e=this.vertical?"1, "+this.percent+", 1":this.percent+", 1, 1";return{transform:"translate"+t+"("+(this._shouldInvertMouseCoords()?"":"-")+this._thumbGap+"px) scale3d("+e+")"}},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_ticksContainerStyles",{get:function(){return{transform:"translate"+(this.vertical?"Y":"X")+"("+(this.vertical||"rtl"!=this._getDirection()?"-":"")+this._tickIntervalPercent/2*100+"%)"}},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_ticksStyles",{get:function(){var t=100*this._tickIntervalPercent,e={backgroundSize:this.vertical?"2px "+t+"%":t+"% 2px",transform:"translateZ(0) translate"+(this.vertical?"Y":"X")+"("+(this.vertical||"rtl"!=this._getDirection()?"":"-")+t/2+"%)"+(this.vertical||"rtl"!=this._getDirection()?"":" rotate(180deg)")};this._isMinValue&&this._thumbGap&&(e["padding"+(this.vertical?this._invertAxis?"Bottom":"Top":this._invertAxis?"Right":"Left")]=this._thumbGap+"px");return e},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_thumbContainerStyles",{get:function(){return{transform:"translate"+(this.vertical?"Y":"X")+"(-"+100*(("rtl"!=this._getDirection()||this.vertical?this._invertAxis:!this._invertAxis)?this.percent:1-this.percent)+"%)"}},enumerable:!0,configurable:!0}),e.prototype._shouldInvertMouseCoords=function(){return"rtl"!=this._getDirection()||this.vertical?this._invertAxis:!this._invertAxis},e.prototype._getDirection=function(){return this._dir&&"rtl"==this._dir.value?"rtl":"ltr"},e.prototype.ngOnInit=function(){var t=this;this._focusMonitor.monitor(this._elementRef,!0).subscribe(function(e){t._isActive=!!e&&"keyboard"!==e,t._changeDetectorRef.detectChanges()}),this._dir&&(this._dirChangeSubscription=this._dir.change.subscribe(function(){t._changeDetectorRef.markForCheck()}))},e.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef),this._dirChangeSubscription.unsubscribe()},e.prototype._onMouseenter=function(){this.disabled||(this._sliderDimensions=this._getSliderDimensions(),this._updateTickIntervalPercent())},e.prototype._onMousedown=function(t){if(!this.disabled&&0===t.button){var e=this.value;this._isSliding=!1,this._focusHostElement(),this._updateValueFromPosition({x:t.clientX,y:t.clientY}),e!=this.value&&(this._emitInputEvent(),this._emitChangeEvent())}},e.prototype._onSlide=function(t){if(!this.disabled){this._isSliding||this._onSlideStart(null),t.preventDefault();var e=this.value;this._updateValueFromPosition({x:t.center.x,y:t.center.y}),e!=this.value&&this._emitInputEvent()}},e.prototype._onSlideStart=function(t){this.disabled||this._isSliding||(this._onMouseenter(),this._isSliding=!0,this._focusHostElement(),this._valueOnSlideStart=this.value,t&&(this._updateValueFromPosition({x:t.center.x,y:t.center.y}),t.preventDefault()))},e.prototype._onSlideEnd=function(){this._isSliding=!1,this._valueOnSlideStart==this.value||this.disabled||this._emitChangeEvent(),this._valueOnSlideStart=null},e.prototype._onFocus=function(){this._sliderDimensions=this._getSliderDimensions(),this._updateTickIntervalPercent()},e.prototype._onBlur=function(){this.onTouched()},e.prototype._onKeydown=function(t){if(!this.disabled&&!Object(s.hasModifierKey)(t)){var e=this.value;switch(t.keyCode){case s.PAGE_UP:this._increment(10);break;case s.PAGE_DOWN:this._increment(-10);break;case s.END:this.value=this.max;break;case s.HOME:this.value=this.min;break;case s.LEFT_ARROW:this._increment("rtl"==this._getDirection()?1:-1);break;case s.UP_ARROW:this._increment(1);break;case s.RIGHT_ARROW:this._increment("rtl"==this._getDirection()?-1:1);break;case s.DOWN_ARROW:this._increment(-1);break;default:return}e!=this.value&&(this._emitInputEvent(),this._emitChangeEvent()),this._isSliding=!0,t.preventDefault()}},e.prototype._onKeyup=function(){this._isSliding=!1},e.prototype._increment=function(t){this.value=this._clamp((this.value||0)+this.step*t,this.min,this.max)},e.prototype._updateValueFromPosition=function(t){if(this._sliderDimensions){var e=this.vertical?this._sliderDimensions.top:this._sliderDimensions.left,n=this.vertical?this._sliderDimensions.height:this._sliderDimensions.width,r=this.vertical?t.y:t.x,i=this._clamp((r-e)/n);if(this._shouldInvertMouseCoords()&&(i=1-i),0===i)this.value=this.min;else if(1===i)this.value=this.max;else{var o=this._calculateValue(i),a=Math.round((o-this.min)/this.step)*this.step+this.min;this.value=this._clamp(a,this.min,this.max)}}},e.prototype._emitChangeEvent=function(){this._controlValueAccessorChangeFn(this.value),this.valueChange.emit(this.value),this.change.emit(this._createChangeEvent())},e.prototype._emitInputEvent=function(){this.input.emit(this._createChangeEvent())},e.prototype._updateTickIntervalPercent=function(){if(this.tickInterval&&this._sliderDimensions)if("auto"==this.tickInterval){var t=this.vertical?this._sliderDimensions.height:this._sliderDimensions.width,e=t*this.step/(this.max-this.min),n=Math.ceil(30/e)*this.step;this._tickIntervalPercent=n/t}else this._tickIntervalPercent=this.tickInterval*this.step/(this.max-this.min)},e.prototype._createChangeEvent=function(t){void 0===t&&(t=this.value);var e=new g;return e.source=this,e.value=t,e},e.prototype._calculatePercentage=function(t){return((t||0)-this.min)/(this.max-this.min)},e.prototype._calculateValue=function(t){return this.min+t*(this.max-this.min)},e.prototype._clamp=function(t,e,n){return void 0===e&&(e=0),void 0===n&&(n=1),Math.max(e,Math.min(t,n))},e.prototype._getSliderDimensions=function(){return this._sliderWrapper?this._sliderWrapper.nativeElement.getBoundingClientRect():null},e.prototype._focusHostElement=function(){this._elementRef.nativeElement.focus()},e.prototype._blurHostElement=function(){this._elementRef.nativeElement.blur()},e.prototype.writeValue=function(t){this.value=t},e.prototype.registerOnChange=function(t){this._controlValueAccessorChangeFn=t},e.prototype.registerOnTouched=function(t){this.onTouched=t},e.prototype.setDisabledState=function(t){this.disabled=t},e.decorators=[{type:u.Component,args:[{selector:"mat-slider",exportAs:"matSlider",providers:[m],host:{"(focus)":"_onFocus()","(blur)":"_onBlur()","(mousedown)":"_onMousedown($event)","(keydown)":"_onKeydown($event)","(keyup)":"_onKeyup()","(mouseenter)":"_onMouseenter()","(slide)":"_onSlide($event)","(slideend)":"_onSlideEnd()","(slidestart)":"_onSlideStart($event)",class:"mat-slider",role:"slider","[tabIndex]":"tabIndex","[attr.aria-disabled]":"disabled","[attr.aria-valuemax]":"max","[attr.aria-valuemin]":"min","[attr.aria-valuenow]":"value","[attr.aria-orientation]":'vertical ? "vertical" : "horizontal"',"[class.mat-slider-disabled]":"disabled","[class.mat-slider-has-ticks]":"tickInterval","[class.mat-slider-horizontal]":"!vertical","[class.mat-slider-axis-inverted]":"_invertAxis","[class.mat-slider-sliding]":"_isSliding","[class.mat-slider-thumb-label-showing]":"thumbLabel","[class.mat-slider-vertical]":"vertical","[class.mat-slider-min-value]":"_isMinValue","[class.mat-slider-hide-last-tick]":"disabled || _isMinValue && _thumbGap && _invertAxis","[class._mat-animation-noopable]":'_animationMode === "NoopAnimations"'},template:'<div class="mat-slider-wrapper" #sliderWrapper><div class="mat-slider-track-wrapper"><div class="mat-slider-track-background" [ngStyle]="_trackBackgroundStyles"></div><div class="mat-slider-track-fill" [ngStyle]="_trackFillStyles"></div></div><div class="mat-slider-ticks-container" [ngStyle]="_ticksContainerStyles"><div class="mat-slider-ticks" [ngStyle]="_ticksStyles"></div></div><div class="mat-slider-thumb-container" [ngStyle]="_thumbContainerStyles"><div class="mat-slider-focus-ring"></div><div class="mat-slider-thumb"></div><div class="mat-slider-thumb-label"><span class="mat-slider-thumb-label-text">{{displayValue}}</span></div></div></div>',styles:[".mat-slider{display:inline-block;position:relative;box-sizing:border-box;padding:8px;outline:0;vertical-align:middle}.mat-slider-wrapper{position:absolute}.mat-slider-track-wrapper{position:absolute;top:0;left:0;overflow:hidden}.mat-slider-track-fill{position:absolute;transform-origin:0 0;transition:transform .4s cubic-bezier(.25,.8,.25,1),background-color .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-track-background{position:absolute;transform-origin:100% 100%;transition:transform .4s cubic-bezier(.25,.8,.25,1),background-color .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-ticks-container{position:absolute;left:0;top:0;overflow:hidden}.mat-slider-ticks{background-repeat:repeat;background-clip:content-box;box-sizing:border-box;opacity:0;transition:opacity .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-thumb-container{position:absolute;z-index:1;transition:transform .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-focus-ring{position:absolute;width:30px;height:30px;border-radius:50%;transform:scale(0);opacity:0;transition:transform .4s cubic-bezier(.25,.8,.25,1),background-color .4s cubic-bezier(.25,.8,.25,1),opacity .4s cubic-bezier(.25,.8,.25,1)}.cdk-keyboard-focused .mat-slider-focus-ring,.cdk-program-focused .mat-slider-focus-ring{transform:scale(1);opacity:1}.mat-slider:not(.mat-slider-disabled) .mat-slider-thumb,.mat-slider:not(.mat-slider-disabled) .mat-slider-thumb-label{cursor:-webkit-grab;cursor:grab}.mat-slider-sliding:not(.mat-slider-disabled) .mat-slider-thumb,.mat-slider-sliding:not(.mat-slider-disabled) .mat-slider-thumb-label,.mat-slider:not(.mat-slider-disabled) .mat-slider-thumb-label:active,.mat-slider:not(.mat-slider-disabled) .mat-slider-thumb:active{cursor:-webkit-grabbing;cursor:grabbing}.mat-slider-thumb{position:absolute;right:-10px;bottom:-10px;box-sizing:border-box;width:20px;height:20px;border:3px solid transparent;border-radius:50%;transform:scale(.7);transition:transform .4s cubic-bezier(.25,.8,.25,1),background-color .4s cubic-bezier(.25,.8,.25,1),border-color .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-thumb-label{display:none;align-items:center;justify-content:center;position:absolute;width:28px;height:28px;border-radius:50%;transition:transform .4s cubic-bezier(.25,.8,.25,1),border-radius .4s cubic-bezier(.25,.8,.25,1),background-color .4s cubic-bezier(.25,.8,.25,1)}@media (-ms-high-contrast:active){.mat-slider-thumb-label{outline:solid 1px}}.mat-slider-thumb-label-text{z-index:1;opacity:0;transition:opacity .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-sliding .mat-slider-thumb-container,.mat-slider-sliding .mat-slider-track-background,.mat-slider-sliding .mat-slider-track-fill{transition-duration:0s}.mat-slider-has-ticks .mat-slider-wrapper::after{content:'';position:absolute;border-width:0;border-style:solid;opacity:0;transition:opacity .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-has-ticks.cdk-focused:not(.mat-slider-hide-last-tick) .mat-slider-wrapper::after,.mat-slider-has-ticks:hover:not(.mat-slider-hide-last-tick) .mat-slider-wrapper::after{opacity:1}.mat-slider-has-ticks.cdk-focused:not(.mat-slider-disabled) .mat-slider-ticks,.mat-slider-has-ticks:hover:not(.mat-slider-disabled) .mat-slider-ticks{opacity:1}.mat-slider-thumb-label-showing .mat-slider-focus-ring{transform:scale(0);opacity:0}.mat-slider-thumb-label-showing .mat-slider-thumb-label{display:flex}.mat-slider-axis-inverted .mat-slider-track-fill{transform-origin:100% 100%}.mat-slider-axis-inverted .mat-slider-track-background{transform-origin:0 0}.mat-slider:not(.mat-slider-disabled).cdk-focused.mat-slider-thumb-label-showing .mat-slider-thumb{transform:scale(0)}.mat-slider:not(.mat-slider-disabled).cdk-focused .mat-slider-thumb-label{border-radius:50% 50% 0}.mat-slider:not(.mat-slider-disabled).cdk-focused .mat-slider-thumb-label-text{opacity:1}.mat-slider:not(.mat-slider-disabled).cdk-mouse-focused .mat-slider-thumb,.mat-slider:not(.mat-slider-disabled).cdk-program-focused .mat-slider-thumb,.mat-slider:not(.mat-slider-disabled).cdk-touch-focused .mat-slider-thumb{border-width:2px;transform:scale(1)}.mat-slider-disabled .mat-slider-focus-ring{transform:scale(0);opacity:0}.mat-slider-disabled .mat-slider-thumb{border-width:4px;transform:scale(.5)}.mat-slider-disabled .mat-slider-thumb-label{display:none}.mat-slider-horizontal{height:48px;min-width:128px}.mat-slider-horizontal .mat-slider-wrapper{height:2px;top:23px;left:8px;right:8px}.mat-slider-horizontal .mat-slider-wrapper::after{height:2px;border-left-width:2px;right:0;top:0}.mat-slider-horizontal .mat-slider-track-wrapper{height:2px;width:100%}.mat-slider-horizontal .mat-slider-track-fill{height:2px;width:100%;transform:scaleX(0)}.mat-slider-horizontal .mat-slider-track-background{height:2px;width:100%;transform:scaleX(1)}.mat-slider-horizontal .mat-slider-ticks-container{height:2px;width:100%}@media (-ms-high-contrast:active){.mat-slider-horizontal .mat-slider-ticks-container{height:0;outline:solid 2px;top:1px}}.mat-slider-horizontal .mat-slider-ticks{height:2px;width:100%}.mat-slider-horizontal .mat-slider-thumb-container{width:100%;height:0;top:50%}.mat-slider-horizontal .mat-slider-focus-ring{top:-15px;right:-15px}.mat-slider-horizontal .mat-slider-thumb-label{right:-14px;top:-40px;transform:translateY(26px) scale(.01) rotate(45deg)}.mat-slider-horizontal .mat-slider-thumb-label-text{transform:rotate(-45deg)}.mat-slider-horizontal.cdk-focused .mat-slider-thumb-label{transform:rotate(45deg)}@media (-ms-high-contrast:active){.mat-slider-horizontal.cdk-focused .mat-slider-thumb-label,.mat-slider-horizontal.cdk-focused .mat-slider-thumb-label-text{transform:none}}.mat-slider-vertical{width:48px;min-height:128px}.mat-slider-vertical .mat-slider-wrapper{width:2px;top:8px;bottom:8px;left:23px}.mat-slider-vertical .mat-slider-wrapper::after{width:2px;border-top-width:2px;bottom:0;left:0}.mat-slider-vertical .mat-slider-track-wrapper{height:100%;width:2px}.mat-slider-vertical .mat-slider-track-fill{height:100%;width:2px;transform:scaleY(0)}.mat-slider-vertical .mat-slider-track-background{height:100%;width:2px;transform:scaleY(1)}.mat-slider-vertical .mat-slider-ticks-container{width:2px;height:100%}@media (-ms-high-contrast:active){.mat-slider-vertical .mat-slider-ticks-container{width:0;outline:solid 2px;left:1px}}.mat-slider-vertical .mat-slider-focus-ring{bottom:-15px;left:-15px}.mat-slider-vertical .mat-slider-ticks{width:2px;height:100%}.mat-slider-vertical .mat-slider-thumb-container{height:100%;width:0;left:50%}.mat-slider-vertical .mat-slider-thumb{-webkit-backface-visibility:hidden;backface-visibility:hidden}.mat-slider-vertical .mat-slider-thumb-label{bottom:-14px;left:-40px;transform:translateX(26px) scale(.01) rotate(-45deg)}.mat-slider-vertical .mat-slider-thumb-label-text{transform:rotate(45deg)}.mat-slider-vertical.cdk-focused .mat-slider-thumb-label{transform:rotate(-45deg)}[dir=rtl] .mat-slider-wrapper::after{left:0;right:auto}[dir=rtl] .mat-slider-horizontal .mat-slider-track-fill{transform-origin:100% 100%}[dir=rtl] .mat-slider-horizontal .mat-slider-track-background{transform-origin:0 0}[dir=rtl] .mat-slider-horizontal.mat-slider-axis-inverted .mat-slider-track-fill{transform-origin:0 0}[dir=rtl] .mat-slider-horizontal.mat-slider-axis-inverted .mat-slider-track-background{transform-origin:100% 100%}.mat-slider._mat-animation-noopable .mat-slider-focus-ring,.mat-slider._mat-animation-noopable .mat-slider-has-ticks .mat-slider-wrapper::after,.mat-slider._mat-animation-noopable .mat-slider-thumb,.mat-slider._mat-animation-noopable .mat-slider-thumb-container,.mat-slider._mat-animation-noopable .mat-slider-thumb-label,.mat-slider._mat-animation-noopable .mat-slider-thumb-label-text,.mat-slider._mat-animation-noopable .mat-slider-ticks,.mat-slider._mat-animation-noopable .mat-slider-track-background,.mat-slider._mat-animation-noopable .mat-slider-track-fill{transition:none}"],inputs:["disabled","color","tabIndex"],encapsulation:u.ViewEncapsulation.None,changeDetection:u.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:u.ElementRef},{type:i.FocusMonitor},{type:u.ChangeDetectorRef},{type:o.Directionality,decorators:[{type:u.Optional}]},{type:String,decorators:[{type:u.Attribute,args:["tabindex"]}]},{type:String,decorators:[{type:u.Optional},{type:u.Inject,args:[d.ANIMATION_MODULE_TYPE]}]}]},e.propDecorators={invert:[{type:u.Input}],max:[{type:u.Input}],min:[{type:u.Input}],step:[{type:u.Input}],thumbLabel:[{type:u.Input}],tickInterval:[{type:u.Input}],value:[{type:u.Input}],displayWith:[{type:u.Input}],vertical:[{type:u.Input}],change:[{type:u.Output}],input:[{type:u.Output}],valueChange:[{type:u.Output}],_sliderWrapper:[{type:u.ViewChild,args:["sliderWrapper"]}]},e}(_),b=function(){function t(){}return t.decorators=[{type:u.NgModule,args:[{imports:[f.CommonModule,l.MatCommonModule],exports:[v,l.MatCommonModule],declarations:[v],providers:[{provide:h.HAMMER_GESTURE_CONFIG,useClass:l.GestureConfig}]}]}],t}()},"./node_modules/@angular/material/esm5/snack-bar.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatSnackBarModule",function(){return x}),n.d(e,"MAT_SNACK_BAR_DEFAULT_OPTIONS_FACTORY",function(){return j}),n.d(e,"MAT_SNACK_BAR_DEFAULT_OPTIONS",function(){return w}),n.d(e,"MatSnackBar",function(){return O}),n.d(e,"MatSnackBarContainer",function(){return b}),n.d(e,"MAT_SNACK_BAR_DATA",function(){return g}),n.d(e,"MatSnackBarConfig",function(){return y}),n.d(e,"MatSnackBarRef",function(){return m}),n.d(e,"SimpleSnackBar",function(){return _}),n.d(e,"matSnackBarAnimations",function(){return v});var r=n("./node_modules/rxjs/_esm5/index.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/animations/fesm5/animations.js"),a=n("./node_modules/tslib/tslib.es6.js"),s=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),u=n("./node_modules/rxjs/_esm5/operators/index.js"),c=n("./node_modules/@angular/cdk/esm5/overlay.es5.js"),l=n("./node_modules/@angular/common/fesm5/common.js"),d=n("./node_modules/@angular/material/esm5/core.es5.js"),p=n("./node_modules/@angular/material/esm5/button.es5.js"),f=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),h=n("./node_modules/@angular/cdk/esm5/layout.es5.js"),m=function(){function t(t,e){var n=this;this._overlayRef=e,this._afterDismissed=new r.Subject,this._afterOpened=new r.Subject,this._onAction=new r.Subject,this._dismissedByAction=!1,this.containerInstance=t,this.onAction().subscribe(function(){return n.dismiss()}),t._onExit.subscribe(function(){return n._finishDismiss()})}return t.prototype.dismiss=function(){this._afterDismissed.closed||this.containerInstance.exit(),clearTimeout(this._durationTimeoutId)},t.prototype.dismissWithAction=function(){this._onAction.closed||(this._dismissedByAction=!0,this._onAction.next(),this._onAction.complete())},t.prototype.closeWithAction=function(){this.dismissWithAction()},t.prototype._dismissAfter=function(t){var e=this;this._durationTimeoutId=setTimeout(function(){return e.dismiss()},t)},t.prototype._open=function(){this._afterOpened.closed||(this._afterOpened.next(),this._afterOpened.complete())},t.prototype._finishDismiss=function(){this._overlayRef.dispose(),this._onAction.closed||this._onAction.complete(),this._afterDismissed.next({dismissedByAction:this._dismissedByAction}),this._afterDismissed.complete(),this._dismissedByAction=!1},t.prototype.afterDismissed=function(){return this._afterDismissed.asObservable()},t.prototype.afterOpened=function(){return this.containerInstance._onEnter},t.prototype.onAction=function(){return this._onAction.asObservable()},t}(),g=new i.InjectionToken("MatSnackBarData"),y=function(){return function(){this.politeness="assertive",this.announcementMessage="",this.duration=0,this.data=null,this.horizontalPosition="center",this.verticalPosition="bottom"}}(),_=function(){function t(t,e){this.snackBarRef=t,this.data=e}return t.prototype.action=function(){this.snackBarRef.dismissWithAction()},Object.defineProperty(t.prototype,"hasAction",{get:function(){return!!this.data.action},enumerable:!0,configurable:!0}),t.decorators=[{type:i.Component,args:[{selector:"simple-snack-bar",template:'<span>{{data.message}}</span><div class="mat-simple-snackbar-action" *ngIf="hasAction"><button mat-button (click)="action()">{{data.action}}</button></div>',styles:[".mat-simple-snackbar{display:flex;justify-content:space-between;align-items:center;height:100%;line-height:20px;opacity:1}.mat-simple-snackbar-action{flex-shrink:0;margin:-8px -8px -8px 8px}.mat-simple-snackbar-action button{max-height:36px;min-width:0}[dir=rtl] .mat-simple-snackbar-action{margin-left:-8px;margin-right:8px}"],encapsulation:i.ViewEncapsulation.None,changeDetection:i.ChangeDetectionStrategy.OnPush,host:{class:"mat-simple-snackbar"}}]}],t.ctorParameters=function(){return[{type:m},{type:void 0,decorators:[{type:i.Inject,args:[g]}]}]},t}(),v={snackBarState:Object(o.trigger)("state",[Object(o.state)("void, hidden",Object(o.style)({transform:"scale(0.8)",opacity:0})),Object(o.state)("visible",Object(o.style)({transform:"scale(1)",opacity:1})),Object(o.transition)("* => visible",Object(o.animate)("150ms cubic-bezier(0, 0, 0.2, 1)")),Object(o.transition)("* => void, * => hidden",Object(o.animate)("75ms cubic-bezier(0.4, 0.0, 1, 1)",Object(o.style)({opacity:0})))])},b=function(t){function e(e,n,i,o){var a=t.call(this)||this;return a._ngZone=e,a._elementRef=n,a._changeDetectorRef=i,a.snackBarConfig=o,a._destroyed=!1,a._onExit=new r.Subject,a._onEnter=new r.Subject,a._animationState="void","assertive"!==o.politeness||o.announcementMessage?"off"===o.politeness?a._role=null:a._role="status":a._role="alert",a}return Object(a.__extends)(e,t),e.prototype.attachComponentPortal=function(t){return this._assertNotAttached(),this._applySnackBarClasses(),this._portalOutlet.attachComponentPortal(t)},e.prototype.attachTemplatePortal=function(t){return this._assertNotAttached(),this._applySnackBarClasses(),this._portalOutlet.attachTemplatePortal(t)},e.prototype.onAnimationEnd=function(t){var e=t.fromState,n=t.toState;if(("void"===n&&"void"!==e||"hidden"===n)&&this._completeExit(),"visible"===n){var r=this._onEnter;this._ngZone.run(function(){r.next(),r.complete()})}},e.prototype.enter=function(){this._destroyed||(this._animationState="visible",this._changeDetectorRef.detectChanges())},e.prototype.exit=function(){return this._animationState="hidden",this._onExit},e.prototype.ngOnDestroy=function(){this._destroyed=!0,this._completeExit()},e.prototype._completeExit=function(){var t=this;this._ngZone.onMicrotaskEmpty.asObservable().pipe(Object(u.take)(1)).subscribe(function(){t._onExit.next(),t._onExit.complete()})},e.prototype._applySnackBarClasses=function(){var t=this._elementRef.nativeElement,e=this.snackBarConfig.panelClass;e&&(Array.isArray(e)?e.forEach(function(e){return t.classList.add(e)}):t.classList.add(e)),"center"===this.snackBarConfig.horizontalPosition&&t.classList.add("mat-snack-bar-center"),"top"===this.snackBarConfig.verticalPosition&&t.classList.add("mat-snack-bar-top")},e.prototype._assertNotAttached=function(){if(this._portalOutlet.hasAttached())throw Error("Attempting to attach snack bar content after content is already attached")},e.decorators=[{type:i.Component,args:[{selector:"snack-bar-container",template:"<ng-template cdkPortalOutlet></ng-template>",styles:[".mat-snack-bar-container{border-radius:4px;box-sizing:border-box;display:block;margin:24px;max-width:33vw;min-width:344px;padding:14px 16px;min-height:48px;transform-origin:center}@media (-ms-high-contrast:active){.mat-snack-bar-container{border:solid 1px}}.mat-snack-bar-handset{width:100%}.mat-snack-bar-handset .mat-snack-bar-container{margin:8px;max-width:100%;min-width:0;width:100%}"],changeDetection:i.ChangeDetectionStrategy.OnPush,encapsulation:i.ViewEncapsulation.None,animations:[v.snackBarState],host:{"[attr.role]":"_role",class:"mat-snack-bar-container","[@state]":"_animationState","(@state.done)":"onAnimationEnd($event)"}}]}],e.ctorParameters=function(){return[{type:i.NgZone},{type:i.ElementRef},{type:i.ChangeDetectorRef},{type:y}]},e.propDecorators={_portalOutlet:[{type:i.ViewChild,args:[s.CdkPortalOutlet]}]},e}(s.BasePortalOutlet),x=function(){function t(){}return t.decorators=[{type:i.NgModule,args:[{imports:[c.OverlayModule,s.PortalModule,l.CommonModule,p.MatButtonModule,d.MatCommonModule],exports:[b,d.MatCommonModule],declarations:[b,_],entryComponents:[b,_]}]}],t}(),w=new i.InjectionToken("mat-snack-bar-default-options",{providedIn:"root",factory:j});function j(){return new y}var O=function(){function t(t,e,n,r,i,o){this._overlay=t,this._live=e,this._injector=n,this._breakpointObserver=r,this._parentSnackBar=i,this._defaultConfig=o,this._snackBarRefAtThisLevel=null}return Object.defineProperty(t.prototype,"_openedSnackBarRef",{get:function(){var t=this._parentSnackBar;return t?t._openedSnackBarRef:this._snackBarRefAtThisLevel},set:function(t){this._parentSnackBar?this._parentSnackBar._openedSnackBarRef=t:this._snackBarRefAtThisLevel=t},enumerable:!0,configurable:!0}),t.prototype.openFromComponent=function(t,e){return this._attach(t,e)},t.prototype.openFromTemplate=function(t,e){return this._attach(t,e)},t.prototype.open=function(t,e,n){void 0===e&&(e="");var r=Object(a.__assign)({},this._defaultConfig,n);return r.data={message:t,action:e},r.announcementMessage||(r.announcementMessage=t),this.openFromComponent(_,r)},t.prototype.dismiss=function(){this._openedSnackBarRef&&this._openedSnackBarRef.dismiss()},t.prototype.ngOnDestroy=function(){this._snackBarRefAtThisLevel&&this._snackBarRefAtThisLevel.dismiss()},t.prototype._attachSnackBarContainer=function(t,e){var n=e&&e.viewContainerRef&&e.viewContainerRef.injector,r=new s.PortalInjector(n||this._injector,new WeakMap([[y,e]])),i=new s.ComponentPortal(b,e.viewContainerRef,r),o=t.attach(i);return o.instance.snackBarConfig=e,o.instance},t.prototype._attach=function(t,e){var n=Object(a.__assign)({},new y,this._defaultConfig,e),r=this._createOverlay(n),o=this._attachSnackBarContainer(r,n),c=new m(o,r);if(t instanceof i.TemplateRef){var l=new s.TemplatePortal(t,null,{$implicit:n.data,snackBarRef:c});c.instance=o.attachTemplatePortal(l)}else{var d=this._createInjector(n,c),p=(l=new s.ComponentPortal(t,void 0,d),o.attachComponentPortal(l));c.instance=p.instance}return this._breakpointObserver.observe(h.Breakpoints.Handset).pipe(Object(u.takeUntil)(r.detachments().pipe(Object(u.take)(1)))).subscribe(function(t){t.matches?r.overlayElement.classList.add("mat-snack-bar-handset"):r.overlayElement.classList.remove("mat-snack-bar-handset")}),this._animateSnackBar(c,n),this._openedSnackBarRef=c,this._openedSnackBarRef},t.prototype._animateSnackBar=function(t,e){var n=this;t.afterDismissed().subscribe(function(){n._openedSnackBarRef==t&&(n._openedSnackBarRef=null),e.announcementMessage&&n._live.clear()}),this._openedSnackBarRef?(this._openedSnackBarRef.afterDismissed().subscribe(function(){t.containerInstance.enter()}),this._openedSnackBarRef.dismiss()):t.containerInstance.enter(),e.duration&&e.duration>0&&t.afterOpened().subscribe(function(){return t._dismissAfter(e.duration)}),e.announcementMessage&&this._live.announce(e.announcementMessage,e.politeness)},t.prototype._createOverlay=function(t){var e=new c.OverlayConfig;e.direction=t.direction;var n=this._overlay.position().global(),r="rtl"===t.direction,i="left"===t.horizontalPosition||"start"===t.horizontalPosition&&!r||"end"===t.horizontalPosition&&r,o=!i&&"center"!==t.horizontalPosition;return i?n.left("0"):o?n.right("0"):n.centerHorizontally(),"top"===t.verticalPosition?n.top("0"):n.bottom("0"),e.positionStrategy=n,this._overlay.create(e)},t.prototype._createInjector=function(t,e){var n=t&&t.viewContainerRef&&t.viewContainerRef.injector;return new s.PortalInjector(n||this._injector,new WeakMap([[m,e],[g,t.data]]))},t.decorators=[{type:i.Injectable,args:[{providedIn:x}]}],t.ctorParameters=function(){return[{type:c.Overlay},{type:f.LiveAnnouncer},{type:i.Injector},{type:h.BreakpointObserver},{type:t,decorators:[{type:i.Optional},{type:i.SkipSelf}]},{type:y,decorators:[{type:i.Inject,args:[w]}]}]},t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t(Object(i.inject)(c.Overlay),Object(i.inject)(f.LiveAnnouncer),Object(i.inject)(i.INJECTOR),Object(i.inject)(h.BreakpointObserver),Object(i.inject)(t,12),Object(i.inject)(w))},token:t,providedIn:x}),t}()},"./node_modules/@angular/material/esm5/sort.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatSortModule",function(){return x}),n.d(e,"MatSortHeaderBase",function(){return _}),n.d(e,"_MatSortHeaderMixinBase",function(){return v}),n.d(e,"MatSortHeader",function(){return b}),n.d(e,"MAT_SORT_HEADER_INTL_PROVIDER_FACTORY",function(){return g}),n.d(e,"MatSortHeaderIntl",function(){return m}),n.d(e,"MAT_SORT_HEADER_INTL_PROVIDER",function(){return y}),n.d(e,"MatSortBase",function(){return l}),n.d(e,"_MatSortMixinBase",function(){return d}),n.d(e,"MatSort",function(){return p}),n.d(e,"matSortAnimations",function(){return h});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/material/esm5/core.es5.js"),s=n("./node_modules/rxjs/_esm5/index.js"),u=n("./node_modules/@angular/animations/fesm5/animations.js"),c=n("./node_modules/@angular/common/fesm5/common.js");var l=function(){return function(){}}(),d=Object(a.mixinInitialized)(Object(a.mixinDisabled)(l)),p=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.sortables=new Map,e._stateChanges=new s.Subject,e.start="asc",e._direction="",e.sortChange=new o.EventEmitter,e}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"direction",{get:function(){return this._direction},set:function(t){if(Object(o.isDevMode)()&&t&&"asc"!==t&&"desc"!==t)throw Error(t+" is not a valid sort direction ('asc' or 'desc').");this._direction=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disableClear",{get:function(){return this._disableClear},set:function(t){this._disableClear=Object(i.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e.prototype.register=function(t){if(!t.id)throw Error("MatSortHeader must be provided with a unique id.");if(this.sortables.has(t.id))throw e=t.id,Error("Cannot have two MatSortables with the same id ("+e+").");var e;this.sortables.set(t.id,t)},e.prototype.deregister=function(t){this.sortables.delete(t.id)},e.prototype.sort=function(t){this.active!=t.id?(this.active=t.id,this.direction=t.start?t.start:this.start):this.direction=this.getNextSortDirection(t),this.sortChange.emit({active:this.active,direction:this.direction})},e.prototype.getNextSortDirection=function(t){if(!t)return"";var e=null!=t.disableClear?t.disableClear:this.disableClear,n=function(t,e){var n=["asc","desc"];"desc"==t&&n.reverse();e||n.push("");return n}(t.start||this.start,e),r=n.indexOf(this.direction)+1;return r>=n.length&&(r=0),n[r]},e.prototype.ngOnInit=function(){this._markInitialized()},e.prototype.ngOnChanges=function(){this._stateChanges.next()},e.prototype.ngOnDestroy=function(){this._stateChanges.complete()},e.decorators=[{type:o.Directive,args:[{selector:"[matSort]",exportAs:"matSort",inputs:["disabled: matSortDisabled"]}]}],e.propDecorators={active:[{type:o.Input,args:["matSortActive"]}],start:[{type:o.Input,args:["matSortStart"]}],direction:[{type:o.Input,args:["matSortDirection"]}],disableClear:[{type:o.Input,args:["matSortDisableClear"]}],sortChange:[{type:o.Output,args:["matSortChange"]}]},e}(d);var f=a.AnimationDurations.ENTERING+" "+a.AnimationCurves.STANDARD_CURVE,h={indicator:Object(u.trigger)("indicator",[Object(u.state)("active-asc, asc",Object(u.style)({transform:"translateY(0px)"})),Object(u.state)("active-desc, desc",Object(u.style)({transform:"translateY(10px)"})),Object(u.transition)("active-asc <=> active-desc",Object(u.animate)(f))]),leftPointer:Object(u.trigger)("leftPointer",[Object(u.state)("active-asc, asc",Object(u.style)({transform:"rotate(-45deg)"})),Object(u.state)("active-desc, desc",Object(u.style)({transform:"rotate(45deg)"})),Object(u.transition)("active-asc <=> active-desc",Object(u.animate)(f))]),rightPointer:Object(u.trigger)("rightPointer",[Object(u.state)("active-asc, asc",Object(u.style)({transform:"rotate(45deg)"})),Object(u.state)("active-desc, desc",Object(u.style)({transform:"rotate(-45deg)"})),Object(u.transition)("active-asc <=> active-desc",Object(u.animate)(f))]),arrowOpacity:Object(u.trigger)("arrowOpacity",[Object(u.state)("desc-to-active, asc-to-active, active",Object(u.style)({opacity:1})),Object(u.state)("desc-to-hint, asc-to-hint, hint",Object(u.style)({opacity:.54})),Object(u.state)("hint-to-desc, active-to-desc, desc, hint-to-asc, active-to-asc, asc, void",Object(u.style)({opacity:0})),Object(u.transition)("* => asc, * => desc, * => active, * => hint, * => void",Object(u.animate)("0ms")),Object(u.transition)("* <=> *",Object(u.animate)(f))]),arrowPosition:Object(u.trigger)("arrowPosition",[Object(u.transition)("* => desc-to-hint, * => desc-to-active",Object(u.animate)(f,Object(u.keyframes)([Object(u.style)({transform:"translateY(-25%)"}),Object(u.style)({transform:"translateY(0)"})]))),Object(u.transition)("* => hint-to-desc, * => active-to-desc",Object(u.animate)(f,Object(u.keyframes)([Object(u.style)({transform:"translateY(0)"}),Object(u.style)({transform:"translateY(25%)"})]))),Object(u.transition)("* => asc-to-hint, * => asc-to-active",Object(u.animate)(f,Object(u.keyframes)([Object(u.style)({transform:"translateY(25%)"}),Object(u.style)({transform:"translateY(0)"})]))),Object(u.transition)("* => hint-to-asc, * => active-to-asc",Object(u.animate)(f,Object(u.keyframes)([Object(u.style)({transform:"translateY(0)"}),Object(u.style)({transform:"translateY(-25%)"})]))),Object(u.state)("desc-to-hint, asc-to-hint, hint, desc-to-active, asc-to-active, active",Object(u.style)({transform:"translateY(0)"})),Object(u.state)("hint-to-desc, active-to-desc, desc",Object(u.style)({transform:"translateY(-25%)"})),Object(u.state)("hint-to-asc, active-to-asc, asc",Object(u.style)({transform:"translateY(25%)"}))]),allowChildren:Object(u.trigger)("allowChildren",[Object(u.transition)("* <=> *",[Object(u.query)("@*",Object(u.animateChild)(),{optional:!0})])])},m=function(){function t(){this.changes=new s.Subject,this.sortButtonLabel=function(t){return"Change sorting for "+t}}return t.decorators=[{type:o.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(o.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}();function g(t){return t||new m}var y={provide:m,deps:[[new o.Optional,new o.SkipSelf,m]],useFactory:g},_=function(){return function(){}}(),v=Object(a.mixinDisabled)(_),b=function(t){function e(e,n,r,i){var o=t.call(this)||this;if(o._intl=e,o._sort=r,o._columnDef=i,o._showIndicatorHint=!1,o._arrowDirection="",o._disableViewStateAnimation=!1,o.arrowPosition="after",!r)throw Error("MatSortHeader must be placed within a parent element with the MatSort directive.");return o._rerenderSubscription=Object(s.merge)(r.sortChange,r._stateChanges,e.changes).subscribe(function(){o._isSorted()&&o._updateArrowDirection(),!o._isSorted()&&o._viewState&&"active"===o._viewState.toState&&(o._disableViewStateAnimation=!1,o._setAnimationTransitionState({fromState:"active",toState:o._arrowDirection})),n.markForCheck()}),o}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"disableClear",{get:function(){return this._disableClear},set:function(t){this._disableClear=Object(i.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){!this.id&&this._columnDef&&(this.id=this._columnDef.name),this._updateArrowDirection(),this._setAnimationTransitionState({toState:this._isSorted()?"active":this._arrowDirection}),this._sort.register(this)},e.prototype.ngOnDestroy=function(){this._sort.deregister(this),this._rerenderSubscription.unsubscribe()},e.prototype._setIndicatorHintVisible=function(t){this._isDisabled()&&t||(this._showIndicatorHint=t,this._isSorted()||(this._updateArrowDirection(),this._showIndicatorHint?this._setAnimationTransitionState({fromState:this._arrowDirection,toState:"hint"}):this._setAnimationTransitionState({fromState:"hint",toState:this._arrowDirection})))},e.prototype._setAnimationTransitionState=function(t){this._viewState=t,this._disableViewStateAnimation&&(this._viewState={toState:t.toState})},e.prototype._handleClick=function(){if(!this._isDisabled()){this._sort.sort(this),"hint"!==this._viewState.toState&&"active"!==this._viewState.toState||(this._disableViewStateAnimation=!0);var t=this._isSorted()?{fromState:this._arrowDirection,toState:"active"}:{fromState:"active",toState:this._arrowDirection};this._setAnimationTransitionState(t),this._showIndicatorHint=!1}},e.prototype._isSorted=function(){return this._sort.active==this.id&&("asc"===this._sort.direction||"desc"===this._sort.direction)},e.prototype._getArrowDirectionState=function(){return(this._isSorted()?"active-":"")+this._arrowDirection},e.prototype._getArrowViewState=function(){var t=this._viewState.fromState;return(t?t+"-to-":"")+this._viewState.toState},e.prototype._updateArrowDirection=function(){this._arrowDirection=this._isSorted()?this._sort.direction:this.start||this._sort.start},e.prototype._isDisabled=function(){return this._sort.disabled||this.disabled},e.prototype._getAriaSortAttribute=function(){return this._isSorted()?"asc"==this._sort.direction?"ascending":"descending":null},e.decorators=[{type:o.Component,args:[{selector:"[mat-sort-header]",exportAs:"matSortHeader",template:'<div class="mat-sort-header-container" [class.mat-sort-header-sorted]="_isSorted()" [class.mat-sort-header-position-before]="arrowPosition == \'before\'"><button class="mat-sort-header-button" type="button" [attr.disabled]="_isDisabled() || null" [attr.aria-label]="_intl.sortButtonLabel(id)" (focus)="_setIndicatorHintVisible(true)" (blur)="_setIndicatorHintVisible(false)"><ng-content></ng-content></button><div class="mat-sort-header-arrow" [@arrowOpacity]="_getArrowViewState()" [@arrowPosition]="_getArrowViewState()" [@allowChildren]="_getArrowDirectionState()" (@arrowPosition.start)="_disableViewStateAnimation = true" (@arrowPosition.done)="_disableViewStateAnimation = false"><div class="mat-sort-header-stem"></div><div class="mat-sort-header-indicator" [@indicator]="_getArrowDirectionState()"><div class="mat-sort-header-pointer-left" [@leftPointer]="_getArrowDirectionState()"></div><div class="mat-sort-header-pointer-right" [@rightPointer]="_getArrowDirectionState()"></div><div class="mat-sort-header-pointer-middle"></div></div></div></div>',styles:[".mat-sort-header-container{display:flex;cursor:pointer;align-items:center}.mat-sort-header-disabled .mat-sort-header-container{cursor:default}.mat-sort-header-position-before{flex-direction:row-reverse}.mat-sort-header-button{border:none;background:0 0;display:flex;align-items:center;padding:0;cursor:inherit;outline:0;font:inherit;color:currentColor}.mat-sort-header-arrow{height:12px;width:12px;min-width:12px;position:relative;display:flex;opacity:0}.mat-sort-header-arrow,[dir=rtl] .mat-sort-header-position-before .mat-sort-header-arrow{margin:0 0 0 6px}.mat-sort-header-position-before .mat-sort-header-arrow,[dir=rtl] .mat-sort-header-arrow{margin:0 6px 0 0}.mat-sort-header-stem{background:currentColor;height:10px;width:2px;margin:auto;display:flex;align-items:center}@media (-ms-high-contrast:active){.mat-sort-header-stem{width:0;border-left:solid 2px}}.mat-sort-header-indicator{width:100%;height:2px;display:flex;align-items:center;position:absolute;top:0;left:0}.mat-sort-header-pointer-middle{margin:auto;height:2px;width:2px;background:currentColor;transform:rotate(45deg)}@media (-ms-high-contrast:active){.mat-sort-header-pointer-middle{width:0;height:0;border-top:solid 2px;border-left:solid 2px}}.mat-sort-header-pointer-left,.mat-sort-header-pointer-right{background:currentColor;width:6px;height:2px;position:absolute;top:0}@media (-ms-high-contrast:active){.mat-sort-header-pointer-left,.mat-sort-header-pointer-right{width:0;height:0;border-left:solid 6px;border-top:solid 2px}}.mat-sort-header-pointer-left{transform-origin:right;left:0}.mat-sort-header-pointer-right{transform-origin:left;right:0}"],host:{"(click)":"_handleClick()","(mouseenter)":"_setIndicatorHintVisible(true)","(longpress)":"_setIndicatorHintVisible(true)","(mouseleave)":"_setIndicatorHintVisible(false)","[attr.aria-sort]":"_getAriaSortAttribute()","[class.mat-sort-header-disabled]":"_isDisabled()"},encapsulation:o.ViewEncapsulation.None,changeDetection:o.ChangeDetectionStrategy.OnPush,inputs:["disabled"],animations:[h.indicator,h.leftPointer,h.rightPointer,h.arrowOpacity,h.arrowPosition,h.allowChildren]}]}],e.ctorParameters=function(){return[{type:m},{type:o.ChangeDetectorRef},{type:p,decorators:[{type:o.Optional}]},{type:void 0,decorators:[{type:o.Inject,args:["MAT_SORT_HEADER_COLUMN_DEF"]},{type:o.Optional}]}]},e.propDecorators={id:[{type:o.Input,args:["mat-sort-header"]}],arrowPosition:[{type:o.Input}],start:[{type:o.Input}],disableClear:[{type:o.Input}]},e}(v),x=function(){function t(){}return t.decorators=[{type:o.NgModule,args:[{imports:[c.CommonModule],exports:[p,b],declarations:[p,b],providers:[y]}]}],t}()},"./node_modules/@angular/material/esm5/stepper.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatStepperModule",function(){return T}),n.d(e,"MatStepLabel",function(){return g}),n.d(e,"MatStep",function(){return j}),n.d(e,"MatStepper",function(){return O}),n.d(e,"MatHorizontalStepper",function(){return C}),n.d(e,"MatVerticalStepper",function(){return E}),n.d(e,"MatStepperNext",function(){return S}),n.d(e,"MatStepperPrevious",function(){return M}),n.d(e,"MatStepHeader",function(){return b}),n.d(e,"MAT_STEPPER_INTL_PROVIDER_FACTORY",function(){return _}),n.d(e,"MatStepperIntl",function(){return y}),n.d(e,"MAT_STEPPER_INTL_PROVIDER",function(){return v}),n.d(e,"matStepperAnimations",function(){return x}),n.d(e,"MatStepperIcon",function(){return w});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/cdk/esm5/stepper.es5.js"),a=n("./node_modules/rxjs/_esm5/index.js"),s=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),u=n("./node_modules/@angular/animations/fesm5/animations.js"),c=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),l=n("./node_modules/@angular/common/fesm5/common.js"),d=n("./node_modules/@angular/material/esm5/core.es5.js"),p=n("./node_modules/rxjs/_esm5/operators/index.js"),f=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),h=n("./node_modules/@angular/material/esm5/button.es5.js"),m=n("./node_modules/@angular/material/esm5/icon.es5.js"),g=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"[matStepLabel]"}]}],e}(o.CdkStepLabel),y=function(){function t(){this.changes=new a.Subject,this.optionalLabel="Optional"}return t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}();function _(t){return t||new y}var v={provide:y,deps:[[new i.Optional,new i.SkipSelf,y]],useFactory:_},b=function(t){function e(e,n,r,i){var o=t.call(this,r)||this;return o._intl=e,o._focusMonitor=n,n.monitor(r,!0),o._intlSubscription=e.changes.subscribe(function(){return i.markForCheck()}),o}return Object(r.__extends)(e,t),e.prototype.ngOnDestroy=function(){this._intlSubscription.unsubscribe(),this._focusMonitor.stopMonitoring(this._elementRef)},e.prototype._stringLabel=function(){return this.label instanceof g?null:this.label},e.prototype._templateLabel=function(){return this.label instanceof g?this.label:null},e.prototype._getHostElement=function(){return this._elementRef.nativeElement},e.prototype._getIconContext=function(){return{index:this.index,active:this.active,optional:this.optional}},e.decorators=[{type:i.Component,args:[{selector:"mat-step-header",template:'<div class="mat-step-header-ripple" mat-ripple [matRippleTrigger]="_getHostElement()"></div><div class="mat-step-icon-state-{{state}} mat-step-icon" [class.mat-step-icon-selected]="selected" [ngSwitch]="state"><div class="mat-step-icon-content"><ng-container *ngSwitchCase="\'number\'" [ngSwitch]="!!(iconOverrides && iconOverrides.number)"><ng-container *ngSwitchCase="true" [ngTemplateOutlet]="iconOverrides.number" [ngTemplateOutletContext]="_getIconContext()"></ng-container><span *ngSwitchDefault>{{index + 1}}</span></ng-container><ng-container *ngSwitchCase="\'edit\'" [ngSwitch]="!!(iconOverrides && iconOverrides.edit)"><ng-container *ngSwitchCase="true" [ngTemplateOutlet]="iconOverrides.edit" [ngTemplateOutletContext]="_getIconContext()"></ng-container><mat-icon *ngSwitchDefault>create</mat-icon></ng-container><ng-container *ngSwitchCase="\'done\'" [ngSwitch]="!!(iconOverrides && iconOverrides.done)"><ng-container *ngSwitchCase="true" [ngTemplateOutlet]="iconOverrides.done" [ngTemplateOutletContext]="_getIconContext()"></ng-container><mat-icon *ngSwitchDefault>done</mat-icon></ng-container><ng-container *ngSwitchCase="\'error\'" [ngSwitch]="!!(iconOverrides && iconOverrides.error)"><ng-container *ngSwitchCase="true" [ngTemplateOutlet]="iconOverrides.error" [ngTemplateOutletContext]="_getIconContext()"></ng-container><mat-icon *ngSwitchDefault>warning</mat-icon></ng-container><ng-container *ngSwitchDefault [ngSwitch]="!!(iconOverrides && iconOverrides[state])"><ng-container *ngSwitchCase="true" [ngTemplateOutlet]="iconOverrides[state]" [ngTemplateOutletContext]="_getIconContext()"></ng-container><mat-icon *ngSwitchDefault>{{state}}</mat-icon></ng-container></div></div><div class="mat-step-label" [class.mat-step-label-active]="active" [class.mat-step-label-selected]="selected" [class.mat-step-label-error]="state == \'error\'"><ng-container *ngIf="_templateLabel()" [ngTemplateOutlet]="_templateLabel()!.template"></ng-container><div class="mat-step-text-label" *ngIf="_stringLabel()">{{label}}</div><div class="mat-step-optional" *ngIf="optional && state != \'error\'">{{_intl.optionalLabel}}</div><div class="mat-step-sub-label-error" *ngIf="state == \'error\'">{{errorMessage}}</div></div>',styles:[".mat-step-header{overflow:hidden;outline:0;cursor:pointer;position:relative;box-sizing:content-box;-webkit-tap-highlight-color:transparent}.mat-step-optional,.mat-step-sub-label-error{font-size:12px}.mat-step-icon{border-radius:50%;height:24px;width:24px;flex-shrink:0;position:relative}.mat-step-icon .mat-icon,.mat-step-icon-content{position:absolute;top:50%;left:50%;transform:translate(-50%,-50%)}.mat-step-icon .mat-icon{font-size:16px;height:16px;width:16px}.mat-step-icon-state-error .mat-icon{font-size:24px;height:24px;width:24px}.mat-step-label{display:inline-block;white-space:nowrap;overflow:hidden;text-overflow:ellipsis;min-width:50px;vertical-align:middle}.mat-step-text-label{text-overflow:ellipsis;overflow:hidden}.mat-step-header .mat-step-header-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}"],host:{class:"mat-step-header",role:"tab"},encapsulation:i.ViewEncapsulation.None,changeDetection:i.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:y},{type:s.FocusMonitor},{type:i.ElementRef},{type:i.ChangeDetectorRef}]},e.propDecorators={state:[{type:i.Input}],label:[{type:i.Input}],errorMessage:[{type:i.Input}],iconOverrides:[{type:i.Input}],index:[{type:i.Input}],selected:[{type:i.Input}],active:[{type:i.Input}],optional:[{type:i.Input}]},e}(o.CdkStepHeader),x={horizontalStepTransition:Object(u.trigger)("stepTransition",[Object(u.state)("previous",Object(u.style)({transform:"translate3d(-100%, 0, 0)",visibility:"hidden"})),Object(u.state)("current",Object(u.style)({transform:"none",visibility:"visible"})),Object(u.state)("next",Object(u.style)({transform:"translate3d(100%, 0, 0)",visibility:"hidden"})),Object(u.transition)("* => *",Object(u.animate)("500ms cubic-bezier(0.35, 0, 0.25, 1)"))]),verticalStepTransition:Object(u.trigger)("stepTransition",[Object(u.state)("previous",Object(u.style)({height:"0px",visibility:"hidden"})),Object(u.state)("next",Object(u.style)({height:"0px",visibility:"hidden"})),Object(u.state)("current",Object(u.style)({height:"*",visibility:"visible"})),Object(u.transition)("* <=> current",Object(u.animate)("225ms cubic-bezier(0.4, 0.0, 0.2, 1)"))])},w=function(){function t(t){this.templateRef=t}return t.decorators=[{type:i.Directive,args:[{selector:"ng-template[matStepperIcon]"}]}],t.ctorParameters=function(){return[{type:i.TemplateRef}]},t.propDecorators={name:[{type:i.Input,args:["matStepperIcon"]}]},t}(),j=function(t){function e(e,n,r){var i=t.call(this,e,r)||this;return i._errorStateMatcher=n,i}return Object(r.__extends)(e,t),e.prototype.isErrorState=function(t,e){var n=this._errorStateMatcher.isErrorState(t,e),r=!!(t&&t.invalid&&this.interacted);return n||r},e.decorators=[{type:i.Component,args:[{selector:"mat-step",template:"<ng-template><ng-content></ng-content></ng-template>",providers:[{provide:d.ErrorStateMatcher,useExisting:e}],encapsulation:i.ViewEncapsulation.None,exportAs:"matStep",changeDetection:i.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:O,decorators:[{type:i.Inject,args:[Object(i.forwardRef)(function(){return O})]}]},{type:d.ErrorStateMatcher,decorators:[{type:i.SkipSelf}]},{type:void 0,decorators:[{type:i.Optional},{type:i.Inject,args:[o.STEPPER_GLOBAL_OPTIONS]}]}]},e.propDecorators={stepLabel:[{type:i.ContentChild,args:[g]}]},e}(o.CdkStep),O=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.animationDone=new i.EventEmitter,e._iconOverrides={},e._animationDone=new a.Subject,e}return Object(r.__extends)(e,t),e.prototype.ngAfterContentInit=function(){var t=this;this._icons.forEach(function(e){var n=e.name,r=e.templateRef;return t._iconOverrides[n]=r}),this._steps.changes.pipe(Object(p.takeUntil)(this._destroyed)).subscribe(function(){return t._stateChanged()}),this._animationDone.pipe(Object(p.distinctUntilChanged)(function(t,e){return t.fromState===e.fromState&&t.toState===e.toState}),Object(p.takeUntil)(this._destroyed)).subscribe(function(e){"current"===e.toState&&t.animationDone.emit()})},e.decorators=[{type:i.Directive,args:[{selector:"[matStepper]"}]}],e.propDecorators={_stepHeader:[{type:i.ViewChildren,args:[b]}],_steps:[{type:i.ContentChildren,args:[j]}],_icons:[{type:i.ContentChildren,args:[w]}],animationDone:[{type:i.Output}]},e}(o.CdkStepper),C=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.labelPosition="end",e}return Object(r.__extends)(e,t),e.decorators=[{type:i.Component,args:[{selector:"mat-horizontal-stepper",exportAs:"matHorizontalStepper",template:'<div class="mat-horizontal-stepper-header-container"><ng-container *ngFor="let step of _steps; let i = index; let isLast = last"><mat-step-header class="mat-horizontal-stepper-header" (click)="step.select()" (keydown)="_onKeydown($event)" [tabIndex]="_getFocusIndex() === i ? 0 : -1" [id]="_getStepLabelId(i)" [attr.aria-posinset]="i + 1" [attr.aria-setsize]="_steps.length" [attr.aria-controls]="_getStepContentId(i)" [attr.aria-selected]="selectedIndex == i" [attr.aria-label]="step.ariaLabel || null" [attr.aria-labelledby]="(!step.ariaLabel && step.ariaLabelledby) ? step.ariaLabelledby : null" [index]="i" [state]="_getIndicatorType(i, step.state)" [label]="step.stepLabel || step.label" [selected]="selectedIndex === i" [active]="step.completed || selectedIndex === i || !linear" [optional]="step.optional" [errorMessage]="step.errorMessage" [iconOverrides]="_iconOverrides"></mat-step-header><div *ngIf="!isLast" class="mat-stepper-horizontal-line"></div></ng-container></div><div class="mat-horizontal-content-container"><div *ngFor="let step of _steps; let i = index" class="mat-horizontal-stepper-content" role="tabpanel" [@stepTransition]="_getAnimationDirection(i)" (@stepTransition.done)="_animationDone.next($event)" [id]="_getStepContentId(i)" [attr.aria-labelledby]="_getStepLabelId(i)" [attr.aria-expanded]="selectedIndex === i"><ng-container [ngTemplateOutlet]="step.content"></ng-container></div></div>',styles:[".mat-stepper-horizontal,.mat-stepper-vertical{display:block}.mat-horizontal-stepper-header-container{white-space:nowrap;display:flex;align-items:center}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header-container{align-items:flex-start}.mat-stepper-horizontal-line{border-top-width:1px;border-top-style:solid;flex:auto;height:0;margin:0 -16px;min-width:32px}.mat-stepper-label-position-bottom .mat-stepper-horizontal-line{margin:0;min-width:0;position:relative;top:36px}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:first-child)::before,.mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:last-child)::after,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:first-child)::after,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:last-child)::before{border-top-width:1px;border-top-style:solid;content:'';display:inline-block;height:0;position:absolute;top:36px;width:calc(50% - 20px)}.mat-horizontal-stepper-header{display:flex;height:72px;overflow:hidden;align-items:center;padding:0 24px}.mat-horizontal-stepper-header .mat-step-icon{margin-right:8px;flex:none}[dir=rtl] .mat-horizontal-stepper-header .mat-step-icon{margin-right:0;margin-left:8px}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header{box-sizing:border-box;flex-direction:column;height:auto;padding:24px}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:last-child)::after,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:first-child)::after{right:0}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:first-child)::before,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:last-child)::before{left:0}[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:first-child::after,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:last-child::before{display:none}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header .mat-step-icon,.mat-stepper-label-position-bottom .mat-horizontal-stepper-header .mat-step-icon-not-touched{margin-right:0;margin-left:0}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header .mat-step-label{padding:16px 0 0 0;text-align:center;width:100%}.mat-vertical-stepper-header{display:flex;align-items:center;padding:24px;height:24px}.mat-vertical-stepper-header .mat-step-icon{margin-right:12px}[dir=rtl] .mat-vertical-stepper-header .mat-step-icon{margin-right:0;margin-left:12px}.mat-horizontal-stepper-content[aria-expanded=false]{height:0;overflow:hidden}.mat-horizontal-content-container{overflow:hidden;padding:0 24px 24px 24px}.mat-vertical-content-container{margin-left:36px;border:0;position:relative}[dir=rtl] .mat-vertical-content-container{margin-left:0;margin-right:36px}.mat-stepper-vertical-line::before{content:'';position:absolute;top:-16px;bottom:-16px;left:0;border-left-width:1px;border-left-style:solid}[dir=rtl] .mat-stepper-vertical-line::before{left:auto;right:0}.mat-vertical-stepper-content{overflow:hidden}.mat-vertical-content{padding:0 24px 24px 24px}.mat-step:last-child .mat-vertical-content-container{border:none}"],inputs:["selectedIndex"],host:{class:"mat-stepper-horizontal","[class.mat-stepper-label-position-end]":'labelPosition == "end"',"[class.mat-stepper-label-position-bottom]":'labelPosition == "bottom"',"aria-orientation":"horizontal",role:"tablist"},animations:[x.horizontalStepTransition],providers:[{provide:O,useExisting:e}],encapsulation:i.ViewEncapsulation.None,changeDetection:i.ChangeDetectionStrategy.OnPush}]}],e.propDecorators={labelPosition:[{type:i.Input}]},e}(O),E=function(t){function e(e,n,r,i){var o=t.call(this,e,n,r,i)||this;return o._orientation="vertical",o}return Object(r.__extends)(e,t),e.decorators=[{type:i.Component,args:[{selector:"mat-vertical-stepper",exportAs:"matVerticalStepper",template:'<div class="mat-step" *ngFor="let step of _steps; let i = index; let isLast = last"><mat-step-header class="mat-vertical-stepper-header" (click)="step.select()" (keydown)="_onKeydown($event)" [tabIndex]="_getFocusIndex() == i ? 0 : -1" [id]="_getStepLabelId(i)" [attr.aria-posinset]="i + 1" [attr.aria-setsize]="_steps.length" [attr.aria-controls]="_getStepContentId(i)" [attr.aria-selected]="selectedIndex === i" [attr.aria-label]="step.ariaLabel || null" [attr.aria-labelledby]="(!step.ariaLabel && step.ariaLabelledby) ? step.ariaLabelledby : null" [index]="i" [state]="_getIndicatorType(i, step.state)" [label]="step.stepLabel || step.label" [selected]="selectedIndex === i" [active]="step.completed || selectedIndex === i || !linear" [optional]="step.optional" [errorMessage]="step.errorMessage" [iconOverrides]="_iconOverrides"></mat-step-header><div class="mat-vertical-content-container" [class.mat-stepper-vertical-line]="!isLast"><div class="mat-vertical-stepper-content" role="tabpanel" [@stepTransition]="_getAnimationDirection(i)" (@stepTransition.done)="_animationDone.next($event)" [id]="_getStepContentId(i)" [attr.aria-labelledby]="_getStepLabelId(i)" [attr.aria-expanded]="selectedIndex === i"><div class="mat-vertical-content"><ng-container [ngTemplateOutlet]="step.content"></ng-container></div></div></div></div>',styles:[".mat-stepper-horizontal,.mat-stepper-vertical{display:block}.mat-horizontal-stepper-header-container{white-space:nowrap;display:flex;align-items:center}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header-container{align-items:flex-start}.mat-stepper-horizontal-line{border-top-width:1px;border-top-style:solid;flex:auto;height:0;margin:0 -16px;min-width:32px}.mat-stepper-label-position-bottom .mat-stepper-horizontal-line{margin:0;min-width:0;position:relative;top:36px}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:first-child)::before,.mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:last-child)::after,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:first-child)::after,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:last-child)::before{border-top-width:1px;border-top-style:solid;content:'';display:inline-block;height:0;position:absolute;top:36px;width:calc(50% - 20px)}.mat-horizontal-stepper-header{display:flex;height:72px;overflow:hidden;align-items:center;padding:0 24px}.mat-horizontal-stepper-header .mat-step-icon{margin-right:8px;flex:none}[dir=rtl] .mat-horizontal-stepper-header .mat-step-icon{margin-right:0;margin-left:8px}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header{box-sizing:border-box;flex-direction:column;height:auto;padding:24px}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:last-child)::after,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:first-child)::after{right:0}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:first-child)::before,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:last-child)::before{left:0}[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:first-child::after,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:last-child::before{display:none}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header .mat-step-icon,.mat-stepper-label-position-bottom .mat-horizontal-stepper-header .mat-step-icon-not-touched{margin-right:0;margin-left:0}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header .mat-step-label{padding:16px 0 0 0;text-align:center;width:100%}.mat-vertical-stepper-header{display:flex;align-items:center;padding:24px;height:24px}.mat-vertical-stepper-header .mat-step-icon{margin-right:12px}[dir=rtl] .mat-vertical-stepper-header .mat-step-icon{margin-right:0;margin-left:12px}.mat-horizontal-stepper-content[aria-expanded=false]{height:0;overflow:hidden}.mat-horizontal-content-container{overflow:hidden;padding:0 24px 24px 24px}.mat-vertical-content-container{margin-left:36px;border:0;position:relative}[dir=rtl] .mat-vertical-content-container{margin-left:0;margin-right:36px}.mat-stepper-vertical-line::before{content:'';position:absolute;top:-16px;bottom:-16px;left:0;border-left-width:1px;border-left-style:solid}[dir=rtl] .mat-stepper-vertical-line::before{left:auto;right:0}.mat-vertical-stepper-content{overflow:hidden}.mat-vertical-content{padding:0 24px 24px 24px}.mat-step:last-child .mat-vertical-content-container{border:none}"],inputs:["selectedIndex"],host:{class:"mat-stepper-vertical","aria-orientation":"vertical",role:"tablist"},animations:[x.verticalStepTransition],providers:[{provide:O,useExisting:e}],encapsulation:i.ViewEncapsulation.None,changeDetection:i.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:c.Directionality,decorators:[{type:i.Optional}]},{type:i.ChangeDetectorRef},{type:i.ElementRef},{type:void 0,decorators:[{type:i.Inject,args:[l.DOCUMENT]}]}]},e}(O),S=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"button[matStepperNext]",host:{"(click)":"_stepper.next()","[type]":"type"},inputs:["type"],providers:[{provide:o.CdkStepper,useExisting:O}]}]}],e}(o.CdkStepperNext),M=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.decorators=[{type:i.Directive,args:[{selector:"button[matStepperPrevious]",host:{"(click)":"_stepper.previous()","[type]":"type"},inputs:["type"],providers:[{provide:o.CdkStepper,useExisting:O}]}]}],e}(o.CdkStepperPrevious),T=function(){function t(){}return t.decorators=[{type:i.NgModule,args:[{imports:[d.MatCommonModule,l.CommonModule,f.PortalModule,h.MatButtonModule,o.CdkStepperModule,m.MatIconModule,d.MatRippleModule],exports:[d.MatCommonModule,C,E,j,g,O,S,M,b,w],declarations:[C,E,j,g,O,S,M,b,w],providers:[v,d.ErrorStateMatcher]}]}],t}()},"./node_modules/@angular/material/esm5/table.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatTableModule",function(){return E}),n.d(e,"MatCellDef",function(){return p}),n.d(e,"MatHeaderCellDef",function(){return f}),n.d(e,"MatFooterCellDef",function(){return h}),n.d(e,"MatColumnDef",function(){return m}),n.d(e,"MatHeaderCell",function(){return g}),n.d(e,"MatFooterCell",function(){return y}),n.d(e,"MatCell",function(){return _}),n.d(e,"MatTable",function(){return d}),n.d(e,"MatHeaderRowDef",function(){return v}),n.d(e,"MatFooterRowDef",function(){return b}),n.d(e,"MatRowDef",function(){return x}),n.d(e,"MatHeaderRow",function(){return w}),n.d(e,"MatFooterRow",function(){return j}),n.d(e,"MatRow",function(){return O}),n.d(e,"MatTableDataSource",function(){return M});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/table.es5.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/common/fesm5/common.js"),s=n("./node_modules/@angular/material/esm5/core.es5.js"),u=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),c=n("./node_modules/rxjs/_esm5/index.js"),l=n("./node_modules/rxjs/_esm5/operators/index.js"),d=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.stickyCssClass="mat-table-sticky",e}return Object(r.__extends)(e,t),e.decorators=[{type:o.Component,args:[{selector:"mat-table, table[mat-table]",exportAs:"matTable",template:i.CDK_TABLE_TEMPLATE,styles:["mat-table{display:block}mat-header-row{min-height:56px}mat-footer-row,mat-row{min-height:48px}mat-footer-row,mat-header-row,mat-row{display:flex;border-width:0;border-bottom-width:1px;border-style:solid;align-items:center;box-sizing:border-box}mat-footer-row::after,mat-header-row::after,mat-row::after{display:inline-block;min-height:inherit;content:''}mat-cell:first-of-type,mat-footer-cell:first-of-type,mat-header-cell:first-of-type{padding-left:24px}[dir=rtl] mat-cell:first-of-type,[dir=rtl] mat-footer-cell:first-of-type,[dir=rtl] mat-header-cell:first-of-type{padding-left:0;padding-right:24px}mat-cell:last-of-type,mat-footer-cell:last-of-type,mat-header-cell:last-of-type{padding-right:24px}[dir=rtl] mat-cell:last-of-type,[dir=rtl] mat-footer-cell:last-of-type,[dir=rtl] mat-header-cell:last-of-type{padding-right:0;padding-left:24px}mat-cell,mat-footer-cell,mat-header-cell{flex:1;display:flex;align-items:center;overflow:hidden;word-wrap:break-word;min-height:inherit}table.mat-table{border-spacing:0}tr.mat-header-row{height:56px}tr.mat-footer-row,tr.mat-row{height:48px}th.mat-header-cell{text-align:left}[dir=rtl] th.mat-header-cell{text-align:right}td.mat-cell,td.mat-footer-cell,th.mat-header-cell{padding:0;border-bottom-width:1px;border-bottom-style:solid}td.mat-cell:first-of-type,td.mat-footer-cell:first-of-type,th.mat-header-cell:first-of-type{padding-left:24px}[dir=rtl] td.mat-cell:first-of-type,[dir=rtl] td.mat-footer-cell:first-of-type,[dir=rtl] th.mat-header-cell:first-of-type{padding-left:0;padding-right:24px}td.mat-cell:last-of-type,td.mat-footer-cell:last-of-type,th.mat-header-cell:last-of-type{padding-right:24px}[dir=rtl] td.mat-cell:last-of-type,[dir=rtl] td.mat-footer-cell:last-of-type,[dir=rtl] th.mat-header-cell:last-of-type{padding-right:0;padding-left:24px}"],host:{class:"mat-table"},encapsulation:o.ViewEncapsulation.None,changeDetection:o.ChangeDetectionStrategy.OnPush}]}],e}(i.CdkTable),p=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.decorators=[{type:o.Directive,args:[{selector:"[matCellDef]",providers:[{provide:i.CdkCellDef,useExisting:e}]}]}],e}(i.CdkCellDef),f=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.decorators=[{type:o.Directive,args:[{selector:"[matHeaderCellDef]",providers:[{provide:i.CdkHeaderCellDef,useExisting:e}]}]}],e}(i.CdkHeaderCellDef),h=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.decorators=[{type:o.Directive,args:[{selector:"[matFooterCellDef]",providers:[{provide:i.CdkFooterCellDef,useExisting:e}]}]}],e}(i.CdkFooterCellDef),m=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.decorators=[{type:o.Directive,args:[{selector:"[matColumnDef]",providers:[{provide:i.CdkColumnDef,useExisting:e},{provide:"MAT_SORT_HEADER_COLUMN_DEF",useExisting:e}]}]}],e.propDecorators={name:[{type:o.Input,args:["matColumnDef"]}],sticky:[{type:o.Input}],stickyEnd:[{type:o.Input}]},e}(i.CdkColumnDef),g=function(t){function e(e,n){var r=t.call(this,e,n)||this;return n.nativeElement.classList.add("mat-column-"+e.cssClassFriendlyName),r}return Object(r.__extends)(e,t),e.decorators=[{type:o.Directive,args:[{selector:"mat-header-cell, th[mat-header-cell]",host:{class:"mat-header-cell",role:"columnheader"}}]}],e.ctorParameters=function(){return[{type:i.CdkColumnDef},{type:o.ElementRef}]},e}(i.CdkHeaderCell),y=function(t){function e(e,n){var r=t.call(this,e,n)||this;return n.nativeElement.classList.add("mat-column-"+e.cssClassFriendlyName),r}return Object(r.__extends)(e,t),e.decorators=[{type:o.Directive,args:[{selector:"mat-footer-cell, td[mat-footer-cell]",host:{class:"mat-footer-cell",role:"gridcell"}}]}],e.ctorParameters=function(){return[{type:i.CdkColumnDef},{type:o.ElementRef}]},e}(i.CdkFooterCell),_=function(t){function e(e,n){var r=t.call(this,e,n)||this;return n.nativeElement.classList.add("mat-column-"+e.cssClassFriendlyName),r}return Object(r.__extends)(e,t),e.decorators=[{type:o.Directive,args:[{selector:"mat-cell, td[mat-cell]",host:{class:"mat-cell",role:"gridcell"}}]}],e.ctorParameters=function(){return[{type:i.CdkColumnDef},{type:o.ElementRef}]},e}(i.CdkCell),v=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.decorators=[{type:o.Directive,args:[{selector:"[matHeaderRowDef]",providers:[{provide:i.CdkHeaderRowDef,useExisting:e}],inputs:["columns: matHeaderRowDef","sticky: matHeaderRowDefSticky"]}]}],e}(i.CdkHeaderRowDef),b=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.decorators=[{type:o.Directive,args:[{selector:"[matFooterRowDef]",providers:[{provide:i.CdkFooterRowDef,useExisting:e}],inputs:["columns: matFooterRowDef","sticky: matFooterRowDefSticky"]}]}],e}(i.CdkFooterRowDef),x=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.decorators=[{type:o.Directive,args:[{selector:"[matRowDef]",providers:[{provide:i.CdkRowDef,useExisting:e}],inputs:["columns: matRowDefColumns","when: matRowDefWhen"]}]}],e}(i.CdkRowDef),w=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.decorators=[{type:o.Component,args:[{selector:"mat-header-row, tr[mat-header-row]",template:i.CDK_ROW_TEMPLATE,host:{class:"mat-header-row",role:"row"},changeDetection:o.ChangeDetectionStrategy.OnPush,encapsulation:o.ViewEncapsulation.None,exportAs:"matHeaderRow",providers:[{provide:i.CdkHeaderRow,useExisting:e}]}]}],e}(i.CdkHeaderRow),j=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.decorators=[{type:o.Component,args:[{selector:"mat-footer-row, tr[mat-footer-row]",template:i.CDK_ROW_TEMPLATE,host:{class:"mat-footer-row",role:"row"},changeDetection:o.ChangeDetectionStrategy.OnPush,encapsulation:o.ViewEncapsulation.None,exportAs:"matFooterRow",providers:[{provide:i.CdkFooterRow,useExisting:e}]}]}],e}(i.CdkFooterRow),O=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.decorators=[{type:o.Component,args:[{selector:"mat-row, tr[mat-row]",template:i.CDK_ROW_TEMPLATE,host:{class:"mat-row",role:"row"},changeDetection:o.ChangeDetectionStrategy.OnPush,encapsulation:o.ViewEncapsulation.None,exportAs:"matRow",providers:[{provide:i.CdkRow,useExisting:e}]}]}],e}(i.CdkRow),C=[d,f,v,m,p,x,h,b,g,_,y,w,O,j],E=function(){function t(){}return t.decorators=[{type:o.NgModule,args:[{imports:[i.CdkTableModule,a.CommonModule,s.MatCommonModule],exports:C,declarations:C}]}],t}(),S=9007199254740991,M=function(t){function e(e){void 0===e&&(e=[]);var n=t.call(this)||this;return n._renderData=new c.BehaviorSubject([]),n._filter=new c.BehaviorSubject(""),n._renderChangesSubscription=c.Subscription.EMPTY,n.sortingDataAccessor=function(t,e){var n=t[e];if(Object(u._isNumberValue)(n)){var r=Number(n);return r<S?r:n}return n},n.sortData=function(t,e){var r=e.active,i=e.direction;return r&&""!=i?t.sort(function(t,e){var o=n.sortingDataAccessor(t,r),a=n.sortingDataAccessor(e,r),s=0;return null!=o&&null!=a?o>a?s=1:o<a&&(s=-1):null!=o?s=1:null!=a&&(s=-1),s*("asc"==i?1:-1)}):t},n.filterPredicate=function(t,e){var n=Object.keys(t).reduce(function(e,n){return e+t[n]+"◬"},"").toLowerCase(),r=e.trim().toLowerCase();return-1!=n.indexOf(r)},n._data=new c.BehaviorSubject(e),n._updateChangeSubscription(),n}return Object(r.__extends)(e,t),Object.defineProperty(e.prototype,"data",{get:function(){return this._data.value},set:function(t){this._data.next(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"filter",{get:function(){return this._filter.value},set:function(t){this._filter.next(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"sort",{get:function(){return this._sort},set:function(t){this._sort=t,this._updateChangeSubscription()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"paginator",{get:function(){return this._paginator},set:function(t){this._paginator=t,this._updateChangeSubscription()},enumerable:!0,configurable:!0}),e.prototype._updateChangeSubscription=function(){var t=this,e=this._sort?Object(c.merge)(this._sort.sortChange,this._sort.initialized):Object(c.of)(null),n=this._paginator?Object(c.merge)(this._paginator.page,this._paginator.initialized):Object(c.of)(null),r=this._data,i=Object(c.combineLatest)(r,this._filter).pipe(Object(l.map)(function(e){var n=e[0];return t._filterData(n)})),o=Object(c.combineLatest)(i,e).pipe(Object(l.map)(function(e){var n=e[0];return t._orderData(n)})),a=Object(c.combineLatest)(o,n).pipe(Object(l.map)(function(e){var n=e[0];return t._pageData(n)}));this._renderChangesSubscription.unsubscribe(),this._renderChangesSubscription=a.subscribe(function(e){return t._renderData.next(e)})},e.prototype._filterData=function(t){var e=this;return this.filteredData=this.filter?t.filter(function(t){return e.filterPredicate(t,e.filter)}):t,this.paginator&&this._updatePaginator(this.filteredData.length),this.filteredData},e.prototype._orderData=function(t){return this.sort?this.sortData(t.slice(),this.sort):t},e.prototype._pageData=function(t){if(!this.paginator)return t;var e=this.paginator.pageIndex*this.paginator.pageSize;return t.slice().splice(e,this.paginator.pageSize)},e.prototype._updatePaginator=function(t){var e=this;Promise.resolve().then(function(){if(e.paginator&&(e.paginator.length=t,e.paginator.pageIndex>0)){var n=Math.ceil(e.paginator.length/e.paginator.pageSize)-1||0;e.paginator.pageIndex=Math.min(e.paginator.pageIndex,n)}})},e.prototype.connect=function(){return this._renderData},e.prototype.disconnect=function(){},e}(i.DataSource)},"./node_modules/@angular/material/esm5/tabs.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatInkBar",function(){return b}),n.d(e,"_MAT_INK_BAR_POSITIONER",function(){return _}),n.d(e,"MatTabBody",function(){return M}),n.d(e,"MatTabBodyPortal",function(){return S}),n.d(e,"MatTabHeader",function(){return P}),n.d(e,"MatTabLabelWrapper",function(){return A}),n.d(e,"MatTab",function(){return C}),n.d(e,"MatTabLabel",function(){return w}),n.d(e,"MatTabNav",function(){return U}),n.d(e,"MatTabLink",function(){return G}),n.d(e,"MatTabContent",function(){return x}),n.d(e,"MatTabsModule",function(){return Y}),n.d(e,"MatTabChangeEvent",function(){return N}),n.d(e,"MAT_TABS_CONFIG",function(){return F}),n.d(e,"MatTabGroupBase",function(){return L}),n.d(e,"_MatTabGroupMixinBase",function(){return B}),n.d(e,"MatTabGroup",function(){return V}),n.d(e,"matTabsAnimations",function(){return E}),n.d(e,"ɵa17",function(){return v}),n.d(e,"ɵf17",function(){return j}),n.d(e,"ɵg17",function(){return O}),n.d(e,"ɵb17",function(){return D}),n.d(e,"ɵc17",function(){return I}),n.d(e,"ɵd17",function(){return T}),n.d(e,"ɵe17",function(){return k}),n.d(e,"ɵj17",function(){return W}),n.d(e,"ɵh17",function(){return z}),n.d(e,"ɵk17",function(){return q}),n.d(e,"ɵi17",function(){return H});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),a=n("./node_modules/@angular/material/esm5/core.es5.js"),s=n("./node_modules/rxjs/_esm5/index.js"),u=n("./node_modules/@angular/animations/fesm5/animations.js"),c=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),l=n("./node_modules/rxjs/_esm5/operators/index.js"),d=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),p=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),f=n("./node_modules/@angular/cdk/esm5/scrolling.es5.js"),h=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),m=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),g=n("./node_modules/@angular/cdk/esm5/observers.es5.js"),y=n("./node_modules/@angular/common/fesm5/common.js"),_=new r.InjectionToken("MatInkBarPositioner",{providedIn:"root",factory:v});function v(){return function(t){return{left:t?(t.offsetLeft||0)+"px":"0",width:t?(t.offsetWidth||0)+"px":"0"}}}var b=function(){function t(t,e,n){this._elementRef=t,this._ngZone=e,this._inkBarPositioner=n}return t.prototype.alignToElement=function(t){var e=this;this.show(),"undefined"!=typeof requestAnimationFrame?this._ngZone.runOutsideAngular(function(){requestAnimationFrame(function(){return e._setStyles(t)})}):this._setStyles(t)},t.prototype.show=function(){this._elementRef.nativeElement.style.visibility="visible"},t.prototype.hide=function(){this._elementRef.nativeElement.style.visibility="hidden"},t.prototype._setStyles=function(t){var e=this._inkBarPositioner(t),n=this._elementRef.nativeElement;n.style.left=e.left,n.style.width=e.width},t.decorators=[{type:r.Directive,args:[{selector:"mat-ink-bar",host:{class:"mat-ink-bar"}}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:r.NgZone},{type:void 0,decorators:[{type:r.Inject,args:[_]}]}]},t}(),x=function(){function t(t){this.template=t}return t.decorators=[{type:r.Directive,args:[{selector:"[matTabContent]"}]}],t.ctorParameters=function(){return[{type:r.TemplateRef}]},t}(),w=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(i.__extends)(e,t),e.decorators=[{type:r.Directive,args:[{selector:"[mat-tab-label], [matTabLabel]"}]}],e}(o.CdkPortal),j=function(){return function(){}}(),O=Object(a.mixinDisabled)(j),C=function(t){function e(e){var n=t.call(this)||this;return n._viewContainerRef=e,n.textLabel="",n._contentPortal=null,n._stateChanges=new s.Subject,n.position=null,n.origin=null,n.isActive=!1,n}return Object(i.__extends)(e,t),Object.defineProperty(e.prototype,"content",{get:function(){return this._contentPortal},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){(t.hasOwnProperty("textLabel")||t.hasOwnProperty("disabled"))&&this._stateChanges.next()},e.prototype.ngOnDestroy=function(){this._stateChanges.complete()},e.prototype.ngOnInit=function(){this._contentPortal=new o.TemplatePortal(this._explicitContent||this._implicitContent,this._viewContainerRef)},e.decorators=[{type:r.Component,args:[{selector:"mat-tab",template:"<ng-template><ng-content></ng-content></ng-template>",inputs:["disabled"],changeDetection:r.ChangeDetectionStrategy.OnPush,encapsulation:r.ViewEncapsulation.None,exportAs:"matTab"}]}],e.ctorParameters=function(){return[{type:r.ViewContainerRef}]},e.propDecorators={templateLabel:[{type:r.ContentChild,args:[w]}],_explicitContent:[{type:r.ContentChild,args:[x,{read:r.TemplateRef}]}],_implicitContent:[{type:r.ViewChild,args:[r.TemplateRef]}],textLabel:[{type:r.Input,args:["label"]}],ariaLabel:[{type:r.Input,args:["aria-label"]}],ariaLabelledby:[{type:r.Input,args:["aria-labelledby"]}]},e}(O),E={translateTab:Object(u.trigger)("translateTab",[Object(u.state)("center, void, left-origin-center, right-origin-center",Object(u.style)({transform:"none"})),Object(u.state)("left",Object(u.style)({transform:"translate3d(-100%, 0, 0)",minHeight:"1px"})),Object(u.state)("right",Object(u.style)({transform:"translate3d(100%, 0, 0)",minHeight:"1px"})),Object(u.transition)("* => left, * => right, left => center, right => center",Object(u.animate)("{{animationDuration}} cubic-bezier(0.35, 0, 0.25, 1)")),Object(u.transition)("void => left-origin-center",[Object(u.style)({transform:"translate3d(-100%, 0, 0)"}),Object(u.animate)("{{animationDuration}} cubic-bezier(0.35, 0, 0.25, 1)")]),Object(u.transition)("void => right-origin-center",[Object(u.style)({transform:"translate3d(100%, 0, 0)"}),Object(u.animate)("{{animationDuration}} cubic-bezier(0.35, 0, 0.25, 1)")])])},S=function(t){function e(e,n,r){var i=t.call(this,e,n)||this;return i._host=r,i._centeringSub=s.Subscription.EMPTY,i._leavingSub=s.Subscription.EMPTY,i}return Object(i.__extends)(e,t),e.prototype.ngOnInit=function(){var e=this;t.prototype.ngOnInit.call(this),this._centeringSub=this._host._beforeCentering.pipe(Object(l.startWith)(this._host._isCenterPosition(this._host._position))).subscribe(function(t){t&&!e.hasAttached()&&e.attach(e._host._content)}),this._leavingSub=this._host._afterLeavingCenter.subscribe(function(){e.detach()})},e.prototype.ngOnDestroy=function(){t.prototype.ngOnDestroy.call(this),this._centeringSub.unsubscribe(),this._leavingSub.unsubscribe()},e.decorators=[{type:r.Directive,args:[{selector:"[matTabBodyHost]"}]}],e.ctorParameters=function(){return[{type:r.ComponentFactoryResolver},{type:r.ViewContainerRef},{type:M,decorators:[{type:r.Inject,args:[Object(r.forwardRef)(function(){return M})]}]}]},e}(o.CdkPortalOutlet),M=function(){function t(t,e,n){var i=this;this._elementRef=t,this._dir=e,this._dirChangeSubscription=s.Subscription.EMPTY,this._translateTabComplete=new s.Subject,this._onCentering=new r.EventEmitter,this._beforeCentering=new r.EventEmitter,this._afterLeavingCenter=new r.EventEmitter,this._onCentered=new r.EventEmitter(!0),this.animationDuration="500ms",this._dir&&n&&(this._dirChangeSubscription=this._dir.change.subscribe(function(t){i._computePositionAnimationState(t),n.markForCheck()})),this._translateTabComplete.pipe(Object(l.distinctUntilChanged)(function(t,e){return t.fromState===e.fromState&&t.toState===e.toState})).subscribe(function(t){i._isCenterPosition(t.toState)&&i._isCenterPosition(i._position)&&i._onCentered.emit(),i._isCenterPosition(t.fromState)&&!i._isCenterPosition(i._position)&&i._afterLeavingCenter.emit()})}return Object.defineProperty(t.prototype,"position",{set:function(t){this._positionIndex=t,this._computePositionAnimationState()},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){"center"==this._position&&null!=this.origin&&(this._position=this._computePositionFromOrigin())},t.prototype.ngOnDestroy=function(){this._dirChangeSubscription.unsubscribe(),this._translateTabComplete.complete()},t.prototype._onTranslateTabStarted=function(t){var e=this._isCenterPosition(t.toState);this._beforeCentering.emit(e),e&&this._onCentering.emit(this._elementRef.nativeElement.clientHeight)},t.prototype._getLayoutDirection=function(){return this._dir&&"rtl"===this._dir.value?"rtl":"ltr"},t.prototype._isCenterPosition=function(t){return"center"==t||"left-origin-center"==t||"right-origin-center"==t},t.prototype._computePositionAnimationState=function(t){void 0===t&&(t=this._getLayoutDirection()),this._positionIndex<0?this._position="ltr"==t?"left":"right":this._positionIndex>0?this._position="ltr"==t?"right":"left":this._position="center"},t.prototype._computePositionFromOrigin=function(){var t=this._getLayoutDirection();return"ltr"==t&&this.origin<=0||"rtl"==t&&this.origin>0?"left-origin-center":"right-origin-center"},t.decorators=[{type:r.Component,args:[{selector:"mat-tab-body",template:'<div class="mat-tab-body-content" #content [@translateTab]="{ value: _position, params: {animationDuration: animationDuration} }" (@translateTab.start)="_onTranslateTabStarted($event)" (@translateTab.done)="_translateTabComplete.next($event)"><ng-template matTabBodyHost></ng-template></div>',styles:[".mat-tab-body-content{height:100%;overflow:auto}.mat-tab-group-dynamic-height .mat-tab-body-content{overflow:hidden}"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,animations:[E.translateTab],host:{class:"mat-tab-body"}}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:c.Directionality,decorators:[{type:r.Optional}]},{type:r.ChangeDetectorRef}]},t.propDecorators={_onCentering:[{type:r.Output}],_beforeCentering:[{type:r.Output}],_afterLeavingCenter:[{type:r.Output}],_onCentered:[{type:r.Output}],_portalHost:[{type:r.ViewChild,args:[o.PortalHostDirective]}],_content:[{type:r.Input,args:["content"]}],origin:[{type:r.Input}],animationDuration:[{type:r.Input}],position:[{type:r.Input}]},t}(),T=function(){return function(){}}(),k=Object(a.mixinDisabled)(T),A=function(t){function e(e){var n=t.call(this)||this;return n.elementRef=e,n}return Object(i.__extends)(e,t),e.prototype.focus=function(){this.elementRef.nativeElement.focus()},e.prototype.getOffsetLeft=function(){return this.elementRef.nativeElement.offsetLeft},e.prototype.getOffsetWidth=function(){return this.elementRef.nativeElement.offsetWidth},e.decorators=[{type:r.Directive,args:[{selector:"[matTabLabelWrapper]",inputs:["disabled"],host:{"[class.mat-tab-disabled]":"disabled","[attr.aria-disabled]":"!!disabled"}}]}],e.ctorParameters=function(){return[{type:r.ElementRef}]},e}(k),D=function(){return function(){}}(),I=Object(a.mixinDisableRipple)(D),P=function(t){function e(e,n,i,o,a){var u=t.call(this)||this;return u._elementRef=e,u._changeDetectorRef=n,u._viewportRuler=i,u._dir=o,u._ngZone=a,u._scrollDistance=0,u._selectedIndexChanged=!1,u._destroyed=new s.Subject,u._showPaginationControls=!1,u._disableScrollAfter=!0,u._disableScrollBefore=!0,u._selectedIndex=0,u.selectFocusedIndex=new r.EventEmitter,u.indexFocused=new r.EventEmitter,u}return Object(i.__extends)(e,t),Object.defineProperty(e.prototype,"selectedIndex",{get:function(){return this._selectedIndex},set:function(t){t=Object(d.coerceNumberProperty)(t),this._selectedIndexChanged=this._selectedIndex!=t,this._selectedIndex=t,this._keyManager&&this._keyManager.updateActiveItemIndex(t)},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentChecked=function(){this._tabLabelCount!=this._labelWrappers.length&&(this.updatePagination(),this._tabLabelCount=this._labelWrappers.length,this._changeDetectorRef.markForCheck()),this._selectedIndexChanged&&(this._scrollToLabel(this._selectedIndex),this._checkScrollingControls(),this._alignInkBarToSelectedTab(),this._selectedIndexChanged=!1,this._changeDetectorRef.markForCheck()),this._scrollDistanceChanged&&(this._updateTabScrollPosition(),this._scrollDistanceChanged=!1,this._changeDetectorRef.markForCheck())},e.prototype._handleKeydown=function(t){switch(t.keyCode){case p.HOME:this._keyManager.setFirstItemActive(),t.preventDefault();break;case p.END:this._keyManager.setLastItemActive(),t.preventDefault();break;case p.ENTER:case p.SPACE:this.selectFocusedIndex.emit(this.focusIndex),t.preventDefault();break;default:this._keyManager.onKeydown(t)}},e.prototype.ngAfterContentInit=function(){var t=this,e=this._dir?this._dir.change:Object(s.of)(null),n=this._viewportRuler.change(150),r=function(){t.updatePagination(),t._alignInkBarToSelectedTab()};this._keyManager=new h.FocusKeyManager(this._labelWrappers).withHorizontalOrientation(this._getLayoutDirection()).withWrap(),this._keyManager.updateActiveItem(0),"undefined"!=typeof requestAnimationFrame?requestAnimationFrame(r):r(),Object(s.merge)(e,n).pipe(Object(l.takeUntil)(this._destroyed)).subscribe(function(){r(),t._keyManager.withHorizontalOrientation(t._getLayoutDirection())}),this._keyManager.change.pipe(Object(l.takeUntil)(this._destroyed)).subscribe(function(e){t.indexFocused.emit(e),t._setTabFocus(e)})},e.prototype.ngOnDestroy=function(){this._destroyed.next(),this._destroyed.complete()},e.prototype._onContentChanges=function(){var t=this,e=this._elementRef.nativeElement.textContent;if(e!==this._currentTextContent){this._currentTextContent=e;var n=function(){t.updatePagination(),t._alignInkBarToSelectedTab(),t._changeDetectorRef.markForCheck()};this._ngZone?this._ngZone.run(n):n()}},e.prototype.updatePagination=function(){this._checkPaginationEnabled(),this._checkScrollingControls(),this._updateTabScrollPosition()},Object.defineProperty(e.prototype,"focusIndex",{get:function(){return this._keyManager?this._keyManager.activeItemIndex:0},set:function(t){this._isValidIndex(t)&&this.focusIndex!==t&&this._keyManager&&this._keyManager.setActiveItem(t)},enumerable:!0,configurable:!0}),e.prototype._isValidIndex=function(t){if(!this._labelWrappers)return!0;var e=this._labelWrappers?this._labelWrappers.toArray()[t]:null;return!!e&&!e.disabled},e.prototype._setTabFocus=function(t){if(this._showPaginationControls&&this._scrollToLabel(t),this._labelWrappers&&this._labelWrappers.length){this._labelWrappers.toArray()[t].focus();var e=this._tabListContainer.nativeElement,n=this._getLayoutDirection();e.scrollLeft="ltr"==n?0:e.scrollWidth-e.offsetWidth}},e.prototype._getLayoutDirection=function(){return this._dir&&"rtl"===this._dir.value?"rtl":"ltr"},e.prototype._updateTabScrollPosition=function(){var t=this.scrollDistance,e="ltr"===this._getLayoutDirection()?-t:t;this._tabList.nativeElement.style.transform="translateX("+Math.round(e)+"px)",this._tabListContainer.nativeElement.scrollLeft=0},Object.defineProperty(e.prototype,"scrollDistance",{get:function(){return this._scrollDistance},set:function(t){this._scrollDistance=Math.max(0,Math.min(this._getMaxScrollDistance(),t)),this._scrollDistanceChanged=!0,this._checkScrollingControls()},enumerable:!0,configurable:!0}),e.prototype._scrollHeader=function(t){var e=this._tabListContainer.nativeElement.offsetWidth;this.scrollDistance+=("before"==t?-1:1)*e/3},e.prototype._scrollToLabel=function(t){var e=this._labelWrappers?this._labelWrappers.toArray()[t]:null;if(e){var n,r,i=this._tabListContainer.nativeElement.offsetWidth;"ltr"==this._getLayoutDirection()?r=(n=e.getOffsetLeft())+e.getOffsetWidth():n=(r=this._tabList.nativeElement.offsetWidth-e.getOffsetLeft())-e.getOffsetWidth();var o=this.scrollDistance,a=this.scrollDistance+i;n<o?this.scrollDistance-=o-n+60:r>a&&(this.scrollDistance+=r-a+60)}},e.prototype._checkPaginationEnabled=function(){var t=this._tabList.nativeElement.scrollWidth>this._elementRef.nativeElement.offsetWidth;t||(this.scrollDistance=0),t!==this._showPaginationControls&&this._changeDetectorRef.markForCheck(),this._showPaginationControls=t},e.prototype._checkScrollingControls=function(){this._disableScrollBefore=0==this.scrollDistance,this._disableScrollAfter=this.scrollDistance==this._getMaxScrollDistance(),this._changeDetectorRef.markForCheck()},e.prototype._getMaxScrollDistance=function(){return this._tabList.nativeElement.scrollWidth-this._tabListContainer.nativeElement.offsetWidth||0},e.prototype._alignInkBarToSelectedTab=function(){var t=this._labelWrappers&&this._labelWrappers.length?this._labelWrappers.toArray()[this.selectedIndex].elementRef.nativeElement:null;this._inkBar.alignToElement(t)},e.decorators=[{type:r.Component,args:[{selector:"mat-tab-header",template:'<div class="mat-tab-header-pagination mat-tab-header-pagination-before mat-elevation-z4" aria-hidden="true" mat-ripple [matRippleDisabled]="_disableScrollBefore || disableRipple" [class.mat-tab-header-pagination-disabled]="_disableScrollBefore" (click)="_scrollHeader(\'before\')"><div class="mat-tab-header-pagination-chevron"></div></div><div class="mat-tab-label-container" #tabListContainer (keydown)="_handleKeydown($event)"><div class="mat-tab-list" #tabList role="tablist" (cdkObserveContent)="_onContentChanges()"><div class="mat-tab-labels"><ng-content></ng-content></div><mat-ink-bar></mat-ink-bar></div></div><div class="mat-tab-header-pagination mat-tab-header-pagination-after mat-elevation-z4" aria-hidden="true" mat-ripple [matRippleDisabled]="_disableScrollAfter || disableRipple" [class.mat-tab-header-pagination-disabled]="_disableScrollAfter" (click)="_scrollHeader(\'after\')"><div class="mat-tab-header-pagination-chevron"></div></div>',styles:[".mat-tab-header{display:flex;overflow:hidden;position:relative;flex-shrink:0}.mat-tab-label{height:48px;padding:0 24px;cursor:pointer;box-sizing:border-box;opacity:.6;min-width:160px;text-align:center;display:inline-flex;justify-content:center;align-items:center;white-space:nowrap;position:relative}.mat-tab-label:focus{outline:0}.mat-tab-label:focus:not(.mat-tab-disabled){opacity:1}@media (-ms-high-contrast:active){.mat-tab-label:focus{outline:dotted 2px}}.mat-tab-label.mat-tab-disabled{cursor:default}@media (-ms-high-contrast:active){.mat-tab-label.mat-tab-disabled{opacity:.5}}.mat-tab-label .mat-tab-label-content{display:inline-flex;justify-content:center;align-items:center;white-space:nowrap}@media (-ms-high-contrast:active){.mat-tab-label{opacity:1}}@media (max-width:599px){.mat-tab-label{min-width:72px}}.mat-ink-bar{position:absolute;bottom:0;height:2px;transition:.5s cubic-bezier(.35,0,.25,1)}.mat-tab-group-inverted-header .mat-ink-bar{bottom:auto;top:0}@media (-ms-high-contrast:active){.mat-ink-bar{outline:solid 2px;height:0}}.mat-tab-header-pagination{position:relative;display:none;justify-content:center;align-items:center;min-width:32px;cursor:pointer;z-index:2}.mat-tab-header-pagination-controls-enabled .mat-tab-header-pagination{display:flex}.mat-tab-header-pagination-before,.mat-tab-header-rtl .mat-tab-header-pagination-after{padding-left:4px}.mat-tab-header-pagination-before .mat-tab-header-pagination-chevron,.mat-tab-header-rtl .mat-tab-header-pagination-after .mat-tab-header-pagination-chevron{transform:rotate(-135deg)}.mat-tab-header-pagination-after,.mat-tab-header-rtl .mat-tab-header-pagination-before{padding-right:4px}.mat-tab-header-pagination-after .mat-tab-header-pagination-chevron,.mat-tab-header-rtl .mat-tab-header-pagination-before .mat-tab-header-pagination-chevron{transform:rotate(45deg)}.mat-tab-header-pagination-chevron{border-style:solid;border-width:2px 2px 0 0;content:'';height:8px;width:8px}.mat-tab-header-pagination-disabled{box-shadow:none;cursor:default}.mat-tab-label-container{display:flex;flex-grow:1;overflow:hidden;z-index:1}.mat-tab-list{flex-grow:1;position:relative;transition:transform .5s cubic-bezier(.35,0,.25,1)}.mat-tab-labels{display:flex}[mat-align-tabs=center] .mat-tab-labels{justify-content:center}[mat-align-tabs=end] .mat-tab-labels{justify-content:flex-end}"],inputs:["disableRipple"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,host:{class:"mat-tab-header","[class.mat-tab-header-pagination-controls-enabled]":"_showPaginationControls","[class.mat-tab-header-rtl]":"_getLayoutDirection() == 'rtl'"}}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:r.ChangeDetectorRef},{type:f.ViewportRuler},{type:c.Directionality,decorators:[{type:r.Optional}]},{type:r.NgZone}]},e.propDecorators={_labelWrappers:[{type:r.ContentChildren,args:[A]}],_inkBar:[{type:r.ViewChild,args:[b]}],_tabListContainer:[{type:r.ViewChild,args:["tabListContainer"]}],_tabList:[{type:r.ViewChild,args:["tabList"]}],selectedIndex:[{type:r.Input}],selectFocusedIndex:[{type:r.Output}],indexFocused:[{type:r.Output}]},e}(I),R=0,N=function(){return function(){}}(),F=new r.InjectionToken("MAT_TABS_CONFIG"),L=function(){return function(t){this._elementRef=t}}(),B=Object(a.mixinColor)(Object(a.mixinDisableRipple)(L),"primary"),V=function(t){function e(e,n,i){var o=t.call(this,e)||this;return o._changeDetectorRef=n,o._indexToSelect=0,o._tabBodyWrapperHeight=0,o._tabsSubscription=s.Subscription.EMPTY,o._tabLabelSubscription=s.Subscription.EMPTY,o._dynamicHeight=!1,o._selectedIndex=null,o.headerPosition="above",o.selectedIndexChange=new r.EventEmitter,o.focusChange=new r.EventEmitter,o.animationDone=new r.EventEmitter,o.selectedTabChange=new r.EventEmitter(!0),o._groupId=R++,o.animationDuration=i&&i.animationDuration?i.animationDuration:"500ms",o}return Object(i.__extends)(e,t),Object.defineProperty(e.prototype,"dynamicHeight",{get:function(){return this._dynamicHeight},set:function(t){this._dynamicHeight=Object(d.coerceBooleanProperty)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selectedIndex",{get:function(){return this._selectedIndex},set:function(t){this._indexToSelect=Object(d.coerceNumberProperty)(t,null)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"backgroundColor",{get:function(){return this._backgroundColor},set:function(t){var e=this._elementRef.nativeElement;e.classList.remove("mat-background-"+this.backgroundColor),t&&e.classList.add("mat-background-"+t),this._backgroundColor=t},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentChecked=function(){var t=this,e=this._indexToSelect=this._clampTabIndex(this._indexToSelect);if(this._selectedIndex!=e){var n=null==this._selectedIndex;n||this.selectedTabChange.emit(this._createChangeEvent(e)),Promise.resolve().then(function(){t._tabs.forEach(function(t,n){return t.isActive=n===e}),n||t.selectedIndexChange.emit(e)})}this._tabs.forEach(function(n,r){n.position=r-e,null==t._selectedIndex||0!=n.position||n.origin||(n.origin=e-t._selectedIndex)}),this._selectedIndex!==e&&(this._selectedIndex=e,this._changeDetectorRef.markForCheck())},e.prototype.ngAfterContentInit=function(){var t=this;this._subscribeToTabLabels(),this._tabsSubscription=this._tabs.changes.subscribe(function(){if(t._clampTabIndex(t._indexToSelect)===t._selectedIndex)for(var e=t._tabs.toArray(),n=0;n<e.length;n++)if(e[n].isActive){t._indexToSelect=t._selectedIndex=n;break}t._subscribeToTabLabels(),t._changeDetectorRef.markForCheck()})},e.prototype.ngOnDestroy=function(){this._tabsSubscription.unsubscribe(),this._tabLabelSubscription.unsubscribe()},e.prototype.realignInkBar=function(){this._tabHeader&&this._tabHeader._alignInkBarToSelectedTab()},e.prototype._focusChanged=function(t){this.focusChange.emit(this._createChangeEvent(t))},e.prototype._createChangeEvent=function(t){var e=new N;return e.index=t,this._tabs&&this._tabs.length&&(e.tab=this._tabs.toArray()[t]),e},e.prototype._subscribeToTabLabels=function(){var t=this;this._tabLabelSubscription&&this._tabLabelSubscription.unsubscribe(),this._tabLabelSubscription=s.merge.apply(void 0,this._tabs.map(function(t){return t._stateChanges})).subscribe(function(){return t._changeDetectorRef.markForCheck()})},e.prototype._clampTabIndex=function(t){return Math.min(this._tabs.length-1,Math.max(t||0,0))},e.prototype._getTabLabelId=function(t){return"mat-tab-label-"+this._groupId+"-"+t},e.prototype._getTabContentId=function(t){return"mat-tab-content-"+this._groupId+"-"+t},e.prototype._setTabBodyWrapperHeight=function(t){if(this._dynamicHeight&&this._tabBodyWrapperHeight){var e=this._tabBodyWrapper.nativeElement;e.style.height=this._tabBodyWrapperHeight+"px",this._tabBodyWrapper.nativeElement.offsetHeight&&(e.style.height=t+"px")}},e.prototype._removeTabBodyWrapperHeight=function(){var t=this._tabBodyWrapper.nativeElement;this._tabBodyWrapperHeight=t.clientHeight,t.style.height="",this.animationDone.emit()},e.prototype._handleClick=function(t,e,n){t.disabled||(this.selectedIndex=e.focusIndex=n)},e.prototype._getTabIndex=function(t,e){return t.disabled?null:this.selectedIndex===e?0:-1},e.decorators=[{type:r.Component,args:[{selector:"mat-tab-group",exportAs:"matTabGroup",template:'<mat-tab-header #tabHeader [selectedIndex]="selectedIndex" [disableRipple]="disableRipple" (indexFocused)="_focusChanged($event)" (selectFocusedIndex)="selectedIndex = $event"><div class="mat-tab-label" role="tab" matTabLabelWrapper mat-ripple cdkMonitorElementFocus *ngFor="let tab of _tabs; let i = index" [id]="_getTabLabelId(i)" [attr.tabIndex]="_getTabIndex(tab, i)" [attr.aria-posinset]="i + 1" [attr.aria-setsize]="_tabs.length" [attr.aria-controls]="_getTabContentId(i)" [attr.aria-selected]="selectedIndex == i" [attr.aria-label]="tab.ariaLabel || null" [attr.aria-labelledby]="(!tab.ariaLabel && tab.ariaLabelledby) ? tab.ariaLabelledby : null" [class.mat-tab-label-active]="selectedIndex == i" [disabled]="tab.disabled" [matRippleDisabled]="tab.disabled || disableRipple" (click)="_handleClick(tab, tabHeader, i)"><div class="mat-tab-label-content"><ng-template [ngIf]="tab.templateLabel"><ng-template [cdkPortalOutlet]="tab.templateLabel"></ng-template></ng-template><ng-template [ngIf]="!tab.templateLabel">{{tab.textLabel}}</ng-template></div></div></mat-tab-header><div class="mat-tab-body-wrapper" #tabBodyWrapper><mat-tab-body role="tabpanel" *ngFor="let tab of _tabs; let i = index" [id]="_getTabContentId(i)" [attr.aria-labelledby]="_getTabLabelId(i)" [class.mat-tab-body-active]="selectedIndex == i" [content]="tab.content" [position]="tab.position" [origin]="tab.origin" [animationDuration]="animationDuration" (_onCentered)="_removeTabBodyWrapperHeight()" (_onCentering)="_setTabBodyWrapperHeight($event)"></mat-tab-body></div>',styles:[".mat-tab-group{display:flex;flex-direction:column}.mat-tab-group.mat-tab-group-inverted-header{flex-direction:column-reverse}.mat-tab-label{height:48px;padding:0 24px;cursor:pointer;box-sizing:border-box;opacity:.6;min-width:160px;text-align:center;display:inline-flex;justify-content:center;align-items:center;white-space:nowrap;position:relative}.mat-tab-label:focus{outline:0}.mat-tab-label:focus:not(.mat-tab-disabled){opacity:1}@media (-ms-high-contrast:active){.mat-tab-label:focus{outline:dotted 2px}}.mat-tab-label.mat-tab-disabled{cursor:default}@media (-ms-high-contrast:active){.mat-tab-label.mat-tab-disabled{opacity:.5}}.mat-tab-label .mat-tab-label-content{display:inline-flex;justify-content:center;align-items:center;white-space:nowrap}@media (-ms-high-contrast:active){.mat-tab-label{opacity:1}}@media (max-width:599px){.mat-tab-label{padding:0 12px}}@media (max-width:959px){.mat-tab-label{padding:0 12px}}.mat-tab-group[mat-stretch-tabs]>.mat-tab-header .mat-tab-label{flex-basis:0;flex-grow:1}.mat-tab-body-wrapper{position:relative;overflow:hidden;display:flex;transition:height .5s cubic-bezier(.35,0,.25,1)}.mat-tab-body{top:0;left:0;right:0;bottom:0;position:absolute;display:block;overflow:hidden;flex-basis:100%}.mat-tab-body.mat-tab-body-active{position:relative;overflow-x:hidden;overflow-y:auto;z-index:1;flex-grow:1}.mat-tab-group.mat-tab-group-dynamic-height .mat-tab-body.mat-tab-body-active{overflow-y:hidden}"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,inputs:["color","disableRipple"],host:{class:"mat-tab-group","[class.mat-tab-group-dynamic-height]":"dynamicHeight","[class.mat-tab-group-inverted-header]":'headerPosition === "below"'}}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:r.ChangeDetectorRef},{type:void 0,decorators:[{type:r.Inject,args:[F]},{type:r.Optional}]}]},e.propDecorators={_tabs:[{type:r.ContentChildren,args:[C]}],_tabBodyWrapper:[{type:r.ViewChild,args:["tabBodyWrapper"]}],_tabHeader:[{type:r.ViewChild,args:["tabHeader"]}],dynamicHeight:[{type:r.Input}],selectedIndex:[{type:r.Input}],headerPosition:[{type:r.Input}],animationDuration:[{type:r.Input}],backgroundColor:[{type:r.Input}],selectedIndexChange:[{type:r.Output}],focusChange:[{type:r.Output}],animationDone:[{type:r.Output}],selectedTabChange:[{type:r.Output}]},e}(B),z=function(){return function(t){this._elementRef=t}}(),H=Object(a.mixinDisableRipple)(Object(a.mixinColor)(z,"primary")),U=function(t){function e(e,n,r,i,o){var a=t.call(this,e)||this;return a._dir=n,a._ngZone=r,a._changeDetectorRef=i,a._viewportRuler=o,a._onDestroy=new s.Subject,a}return Object(i.__extends)(e,t),Object.defineProperty(e.prototype,"backgroundColor",{get:function(){return this._backgroundColor},set:function(t){var e=this._elementRef.nativeElement;e.classList.remove("mat-background-"+this.backgroundColor),t&&e.classList.add("mat-background-"+t),this._backgroundColor=t},enumerable:!0,configurable:!0}),e.prototype.updateActiveLink=function(t){this._activeLinkChanged=!!t,this._changeDetectorRef.markForCheck()},e.prototype.ngAfterContentInit=function(){var t=this;this._ngZone.runOutsideAngular(function(){var e=t._dir?t._dir.change:Object(s.of)(null);return Object(s.merge)(e,t._viewportRuler.change(10)).pipe(Object(l.takeUntil)(t._onDestroy)).subscribe(function(){return t._alignInkBar()})})},e.prototype.ngAfterContentChecked=function(){if(this._activeLinkChanged){var t=this._tabLinks.find(function(t){return t.active});this._activeLinkElement=t?t._elementRef:null,this._alignInkBar(),this._activeLinkChanged=!1}},e.prototype.ngOnDestroy=function(){this._onDestroy.next(),this._onDestroy.complete()},e.prototype._alignInkBar=function(){this._activeLinkElement?(this._inkBar.show(),this._inkBar.alignToElement(this._activeLinkElement.nativeElement)):this._inkBar.hide()},e.decorators=[{type:r.Component,args:[{selector:"[mat-tab-nav-bar]",exportAs:"matTabNavBar, matTabNav",inputs:["color","disableRipple"],template:'<div class="mat-tab-links" (cdkObserveContent)="_alignInkBar()"><ng-content></ng-content><mat-ink-bar></mat-ink-bar></div>',styles:[".mat-tab-nav-bar{overflow:hidden;position:relative;flex-shrink:0}.mat-tab-links{position:relative;display:flex}[mat-align-tabs=center] .mat-tab-links{justify-content:center}[mat-align-tabs=end] .mat-tab-links{justify-content:flex-end}.mat-tab-link{height:48px;padding:0 24px;cursor:pointer;box-sizing:border-box;opacity:.6;min-width:160px;text-align:center;display:inline-flex;justify-content:center;align-items:center;white-space:nowrap;vertical-align:top;text-decoration:none;position:relative;overflow:hidden;-webkit-tap-highlight-color:transparent}.mat-tab-link:focus{outline:0}.mat-tab-link:focus:not(.mat-tab-disabled){opacity:1}@media (-ms-high-contrast:active){.mat-tab-link:focus{outline:dotted 2px}}.mat-tab-link.mat-tab-disabled{cursor:default}@media (-ms-high-contrast:active){.mat-tab-link.mat-tab-disabled{opacity:.5}}.mat-tab-link .mat-tab-label-content{display:inline-flex;justify-content:center;align-items:center;white-space:nowrap}@media (-ms-high-contrast:active){.mat-tab-link{opacity:1}}[mat-stretch-tabs] .mat-tab-link{flex-basis:0;flex-grow:1}.mat-tab-link.mat-tab-disabled{pointer-events:none}@media (max-width:599px){.mat-tab-link{min-width:72px}}.mat-ink-bar{position:absolute;bottom:0;height:2px;transition:.5s cubic-bezier(.35,0,.25,1)}.mat-tab-group-inverted-header .mat-ink-bar{bottom:auto;top:0}@media (-ms-high-contrast:active){.mat-ink-bar{outline:solid 2px;height:0}}"],host:{class:"mat-tab-nav-bar"},encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:c.Directionality,decorators:[{type:r.Optional}]},{type:r.NgZone},{type:r.ChangeDetectorRef},{type:f.ViewportRuler}]},e.propDecorators={_inkBar:[{type:r.ViewChild,args:[b]}],_tabLinks:[{type:r.ContentChildren,args:[Object(r.forwardRef)(function(){return G}),{descendants:!0}]}],backgroundColor:[{type:r.Input}]},e}(H),W=function(){return function(){}}(),q=Object(a.mixinTabIndex)(Object(a.mixinDisableRipple)(Object(a.mixinDisabled)(W))),G=function(t){function e(e,n,r,i,o,s,u){var c=t.call(this)||this;return c._tabNavBar=e,c._elementRef=n,c._focusMonitor=u,c._isActive=!1,c._ripplesGloballyDisabled=!1,c.rippleConfig={},c._tabLinkRipple=new a.RippleRenderer(c,r,n,i),c._tabLinkRipple.setupTriggerEvents(n.nativeElement),c.tabIndex=parseInt(s)||0,o&&(c._ripplesGloballyDisabled=!!o.disabled,c.rippleConfig={terminateOnPointerUp:o.terminateOnPointerUp,animation:o.animation}),u&&u.monitor(n),c}return Object(i.__extends)(e,t),Object.defineProperty(e.prototype,"active",{get:function(){return this._isActive},set:function(t){t!==this._isActive&&(this._isActive=t,this._tabNavBar.updateActiveLink(this._elementRef))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"rippleDisabled",{get:function(){return this.disabled||this.disableRipple||this._tabNavBar.disableRipple||this._ripplesGloballyDisabled},enumerable:!0,configurable:!0}),e.prototype.ngOnDestroy=function(){this._tabLinkRipple._removeTriggerEvents(),this._focusMonitor&&this._focusMonitor.stopMonitoring(this._elementRef)},e.decorators=[{type:r.Directive,args:[{selector:"[mat-tab-link], [matTabLink]",exportAs:"matTabLink",inputs:["disabled","disableRipple","tabIndex"],host:{class:"mat-tab-link","[attr.aria-current]":"active","[attr.aria-disabled]":"disabled.toString()","[attr.tabIndex]":"tabIndex","[class.mat-tab-disabled]":"disabled","[class.mat-tab-label-active]":"active"}}]}],e.ctorParameters=function(){return[{type:U},{type:r.ElementRef},{type:r.NgZone},{type:m.Platform},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[a.MAT_RIPPLE_GLOBAL_OPTIONS]}]},{type:String,decorators:[{type:r.Attribute,args:["tabindex"]}]},{type:h.FocusMonitor}]},e.propDecorators={active:[{type:r.Input}]},e}(q),Y=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[y.CommonModule,a.MatCommonModule,o.PortalModule,a.MatRippleModule,g.ObserversModule,h.A11yModule],exports:[a.MatCommonModule,V,w,C,U,G,x],declarations:[V,w,C,b,A,U,G,M,S,P,x]}]}],t}()},"./node_modules/@angular/material/esm5/toolbar.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatToolbarModule",function(){return f}),n.d(e,"throwToolbarMixedModesError",function(){return p}),n.d(e,"MatToolbarBase",function(){return u}),n.d(e,"_MatToolbarMixinBase",function(){return c}),n.d(e,"MatToolbarRow",function(){return l}),n.d(e,"MatToolbar",function(){return d});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),o=n("./node_modules/@angular/common/fesm5/common.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),s=n("./node_modules/@angular/material/esm5/core.es5.js"),u=function(){return function(t){this._elementRef=t}}(),c=Object(s.mixinColor)(u),l=function(){function t(){}return t.decorators=[{type:a.Directive,args:[{selector:"mat-toolbar-row",exportAs:"matToolbarRow",host:{class:"mat-toolbar-row"}}]}],t}(),d=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i._platform=n,i._document=r,i}return Object(r.__extends)(e,t),e.prototype.ngAfterViewInit=function(){var t=this;Object(a.isDevMode)()&&this._platform.isBrowser&&(this._checkToolbarMixedModes(),this._toolbarRows.changes.subscribe(function(){return t._checkToolbarMixedModes()}))},e.prototype._checkToolbarMixedModes=function(){var t=this;this._toolbarRows.length&&(Array.from(this._elementRef.nativeElement.childNodes).filter(function(t){return!(t.classList&&t.classList.contains("mat-toolbar-row"))}).filter(function(e){return e.nodeType!==(t._document?t._document.COMMENT_NODE:8)}).some(function(t){return!(!t.textContent||!t.textContent.trim())})&&p())},e.decorators=[{type:a.Component,args:[{selector:"mat-toolbar",exportAs:"matToolbar",template:'<ng-content></ng-content><ng-content select="mat-toolbar-row"></ng-content>',styles:["@media (-ms-high-contrast:active){.mat-toolbar{outline:solid 1px}}.mat-toolbar-row,.mat-toolbar-single-row{display:flex;box-sizing:border-box;padding:0 16px;width:100%;flex-direction:row;align-items:center;white-space:nowrap}.mat-toolbar-multiple-rows{display:flex;box-sizing:border-box;flex-direction:column;width:100%}.mat-toolbar-multiple-rows{min-height:64px}.mat-toolbar-row,.mat-toolbar-single-row{height:64px}@media (max-width:599px){.mat-toolbar-multiple-rows{min-height:56px}.mat-toolbar-row,.mat-toolbar-single-row{height:56px}}"],inputs:["color"],host:{class:"mat-toolbar","[class.mat-toolbar-multiple-rows]":"_toolbarRows.length > 0","[class.mat-toolbar-single-row]":"_toolbarRows.length === 0"},changeDetection:a.ChangeDetectionStrategy.OnPush,encapsulation:a.ViewEncapsulation.None}]}],e.ctorParameters=function(){return[{type:a.ElementRef},{type:i.Platform},{type:void 0,decorators:[{type:a.Inject,args:[o.DOCUMENT]}]}]},e.propDecorators={_toolbarRows:[{type:a.ContentChildren,args:[l]}]},e}(c);function p(){throw Error("MatToolbar: Attempting to combine different toolbar modes. Either specify multiple `<mat-toolbar-row>` elements explicitly or just place content inside of a `<mat-toolbar>` for a single row.")}var f=function(){function t(){}return t.decorators=[{type:a.NgModule,args:[{imports:[s.MatCommonModule],exports:[d,l,s.MatCommonModule],declarations:[d,l]}]}],t}()},"./node_modules/@angular/material/esm5/tooltip.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"MatTooltipModule",function(){return A}),n.d(e,"getMatTooltipInvalidPositionError",function(){return j}),n.d(e,"MAT_TOOLTIP_SCROLL_STRATEGY_FACTORY",function(){return C}),n.d(e,"MAT_TOOLTIP_DEFAULT_OPTIONS_FACTORY",function(){return M}),n.d(e,"SCROLL_THROTTLE_MS",function(){return x}),n.d(e,"TOOLTIP_PANEL_CLASS",function(){return w}),n.d(e,"MAT_TOOLTIP_SCROLL_STRATEGY",function(){return O}),n.d(e,"MAT_TOOLTIP_SCROLL_STRATEGY_FACTORY_PROVIDER",function(){return E}),n.d(e,"MAT_TOOLTIP_DEFAULT_OPTIONS",function(){return S}),n.d(e,"MatTooltip",function(){return T}),n.d(e,"TooltipComponent",function(){return k}),n.d(e,"matTooltipAnimations",function(){return b});var r=n("./node_modules/@angular/animations/fesm5/animations.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),a=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),c=n("./node_modules/@angular/cdk/esm5/layout.es5.js"),l=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),d=n("./node_modules/@angular/cdk/esm5/overlay.es5.js"),p=n("./node_modules/@angular/cdk/esm5/scrolling.es5.js"),f=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),h=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),m=n("./node_modules/rxjs/_esm5/operators/index.js"),g=n("./node_modules/@angular/core/fesm5/core.js"),y=n("./node_modules/rxjs/_esm5/index.js"),_=n("./node_modules/@angular/common/fesm5/common.js"),v=n("./node_modules/@angular/material/esm5/core.es5.js"),b={tooltipState:Object(r.trigger)("state",[Object(r.state)("initial, void, hidden",Object(r.style)({opacity:0,transform:"scale(0)"})),Object(r.state)("visible",Object(r.style)({transform:"scale(1)"})),Object(r.transition)("* => visible",Object(r.animate)("200ms cubic-bezier(0, 0, 0.2, 1)",Object(r.keyframes)([Object(r.style)({opacity:0,transform:"scale(0)",offset:0}),Object(r.style)({opacity:.5,transform:"scale(0.99)",offset:.5}),Object(r.style)({opacity:1,transform:"scale(1)",offset:1})]))),Object(r.transition)("* => hidden",Object(r.animate)("100ms cubic-bezier(0, 0, 0.2, 1)",Object(r.style)({opacity:0})))])},x=20,w="mat-tooltip-panel";function j(t){return Error('Tooltip position "'+t+'" is invalid.')}var O=new g.InjectionToken("mat-tooltip-scroll-strategy");function C(t){return function(){return t.scrollStrategies.reposition({scrollThrottle:x})}}var E={provide:O,deps:[d.Overlay],useFactory:C},S=new g.InjectionToken("mat-tooltip-default-options",{providedIn:"root",factory:M});function M(){return{showDelay:0,hideDelay:0,touchendHideDelay:1500}}var T=function(){function t(t,e,n,r,i,o,a,s,u,c,l,d){var p=this;this._overlay=t,this._elementRef=e,this._scrollDispatcher=n,this._viewContainerRef=r,this._ngZone=i,this._ariaDescriber=a,this._focusMonitor=s,this._dir=c,this._defaultOptions=l,this._position="below",this._disabled=!1,this.showDelay=this._defaultOptions.showDelay,this.hideDelay=this._defaultOptions.hideDelay,this._message="",this._manualListeners=new Map,this._destroyed=new y.Subject,this._scrollStrategy=u;var f=e.nativeElement,h=f.style,g="undefined"==typeof window||window.Hammer||d;o.IOS||o.ANDROID?g||this._manualListeners.set("touchstart",function(){return p.show()}):this._manualListeners.set("mouseenter",function(){return p.show()}).set("mouseleave",function(){return p.hide()}),this._manualListeners.forEach(function(t,e){return f.addEventListener(e,t)}),"INPUT"!==f.nodeName&&"TEXTAREA"!==f.nodeName||(h.webkitUserSelect=h.userSelect=h.msUserSelect=""),f.draggable&&"none"===h.webkitUserDrag&&(h.webkitUserDrag=""),s.monitor(e).pipe(Object(m.takeUntil)(this._destroyed)).subscribe(function(t){t?"keyboard"===t&&i.run(function(){return p.show()}):i.run(function(){return p.hide(0)})})}return Object.defineProperty(t.prototype,"position",{get:function(){return this._position},set:function(t){t!==this._position&&(this._position=t,this._overlayRef&&(this._updatePosition(),this._tooltipInstance&&this._tooltipInstance.show(0),this._overlayRef.updatePosition()))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this._disabled},set:function(t){this._disabled=Object(s.coerceBooleanProperty)(t),this._disabled&&this.hide(0)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"message",{get:function(){return this._message},set:function(t){this._ariaDescriber.removeDescription(this._elementRef.nativeElement,this._message),this._message=null!=t?(""+t).trim():"",!this._message&&this._isTooltipVisible()?this.hide(0):(this._updateTooltipMessage(),this._ariaDescriber.describe(this._elementRef.nativeElement,this.message))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"tooltipClass",{get:function(){return this._tooltipClass},set:function(t){this._tooltipClass=t,this._tooltipInstance&&this._setTooltipClass(this._tooltipClass)},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){var t=this;this._overlayRef&&(this._overlayRef.dispose(),this._tooltipInstance=null),this._manualListeners.forEach(function(e,n){t._elementRef.nativeElement.removeEventListener(n,e)}),this._manualListeners.clear(),this._destroyed.next(),this._destroyed.complete(),this._ariaDescriber.removeDescription(this._elementRef.nativeElement,this.message),this._focusMonitor.stopMonitoring(this._elementRef)},t.prototype.show=function(t){var e=this;if(void 0===t&&(t=this.showDelay),!this.disabled&&this.message&&(!this._isTooltipVisible()||this._tooltipInstance._showTimeoutId||this._tooltipInstance._hideTimeoutId)){var n=this._createOverlay();this._detach(),this._portal=this._portal||new h.ComponentPortal(k,this._viewContainerRef),this._tooltipInstance=n.attach(this._portal).instance,this._tooltipInstance.afterHidden().pipe(Object(m.takeUntil)(this._destroyed)).subscribe(function(){return e._detach()}),this._setTooltipClass(this._tooltipClass),this._updateTooltipMessage(),this._tooltipInstance.show(t)}},t.prototype.hide=function(t){void 0===t&&(t=this.hideDelay),this._tooltipInstance&&this._tooltipInstance.hide(t)},t.prototype.toggle=function(){this._isTooltipVisible()?this.hide():this.show()},t.prototype._isTooltipVisible=function(){return!!this._tooltipInstance&&this._tooltipInstance.isVisible()},t.prototype._handleKeydown=function(t){this._isTooltipVisible()&&t.keyCode===u.ESCAPE&&(t.stopPropagation(),this.hide(0))},t.prototype._handleTouchend=function(){this.hide(this._defaultOptions.touchendHideDelay)},t.prototype._createOverlay=function(){var t=this;if(this._overlayRef)return this._overlayRef;var e=this._overlay.position().flexibleConnectedTo(this._elementRef).withTransformOriginOn(".mat-tooltip").withFlexibleDimensions(!1).withViewportMargin(8),n=this._scrollDispatcher.getAncestorScrollContainers(this._elementRef);return e.withScrollableContainers(n),e.positionChanges.pipe(Object(m.takeUntil)(this._destroyed)).subscribe(function(e){t._tooltipInstance&&e.scrollableViewProperties.isOverlayClipped&&t._tooltipInstance.isVisible()&&t._ngZone.run(function(){return t.hide(0)})}),this._overlayRef=this._overlay.create({direction:this._dir,positionStrategy:e,panelClass:w,scrollStrategy:this._scrollStrategy()}),this._updatePosition(),this._overlayRef.detachments().pipe(Object(m.takeUntil)(this._destroyed)).subscribe(function(){return t._detach()}),this._overlayRef},t.prototype._detach=function(){this._overlayRef&&this._overlayRef.hasAttached()&&this._overlayRef.detach(),this._tooltipInstance=null},t.prototype._updatePosition=function(){var t=this._overlayRef.getConfig().positionStrategy,e=this._getOrigin(),n=this._getOverlayPosition();t.withPositions([Object(i.__assign)({},e.main,n.main),Object(i.__assign)({},e.fallback,n.fallback)])},t.prototype._getOrigin=function(){var t,e=!this._dir||"ltr"==this._dir.value,n=this.position;if("above"==n||"below"==n)t={originX:"center",originY:"above"==n?"top":"bottom"};else if("before"==n||"left"==n&&e||"right"==n&&!e)t={originX:"start",originY:"center"};else{if(!("after"==n||"right"==n&&e||"left"==n&&!e))throw j(n);t={originX:"end",originY:"center"}}var r=this._invertPosition(t.originX,t.originY);return{main:t,fallback:{originX:r.x,originY:r.y}}},t.prototype._getOverlayPosition=function(){var t,e=!this._dir||"ltr"==this._dir.value,n=this.position;if("above"==n)t={overlayX:"center",overlayY:"bottom"};else if("below"==n)t={overlayX:"center",overlayY:"top"};else if("before"==n||"left"==n&&e||"right"==n&&!e)t={overlayX:"end",overlayY:"center"};else{if(!("after"==n||"right"==n&&e||"left"==n&&!e))throw j(n);t={overlayX:"start",overlayY:"center"}}var r=this._invertPosition(t.overlayX,t.overlayY);return{main:t,fallback:{overlayX:r.x,overlayY:r.y}}},t.prototype._updateTooltipMessage=function(){var t=this;this._tooltipInstance&&(this._tooltipInstance.message=this.message,this._tooltipInstance._markForCheck(),this._ngZone.onMicrotaskEmpty.asObservable().pipe(Object(m.take)(1),Object(m.takeUntil)(this._destroyed)).subscribe(function(){t._tooltipInstance&&t._overlayRef.updatePosition()}))},t.prototype._setTooltipClass=function(t){this._tooltipInstance&&(this._tooltipInstance.tooltipClass=t,this._tooltipInstance._markForCheck())},t.prototype._invertPosition=function(t,e){return"above"===this.position||"below"===this.position?"top"===e?e="bottom":"bottom"===e&&(e="top"):"end"===t?t="start":"start"===t&&(t="end"),{x:t,y:e}},t.decorators=[{type:g.Directive,args:[{selector:"[matTooltip]",exportAs:"matTooltip",host:{"(longpress)":"show()","(keydown)":"_handleKeydown($event)","(touchend)":"_handleTouchend()"}}]}],t.ctorParameters=function(){return[{type:d.Overlay},{type:g.ElementRef},{type:p.ScrollDispatcher},{type:g.ViewContainerRef},{type:g.NgZone},{type:f.Platform},{type:o.AriaDescriber},{type:o.FocusMonitor},{type:void 0,decorators:[{type:g.Inject,args:[O]}]},{type:a.Directionality,decorators:[{type:g.Optional}]},{type:void 0,decorators:[{type:g.Optional},{type:g.Inject,args:[S]}]},{type:void 0,decorators:[{type:g.Optional},{type:g.Inject,args:[l.HAMMER_LOADER]}]}]},t.propDecorators={position:[{type:g.Input,args:["matTooltipPosition"]}],disabled:[{type:g.Input,args:["matTooltipDisabled"]}],showDelay:[{type:g.Input,args:["matTooltipShowDelay"]}],hideDelay:[{type:g.Input,args:["matTooltipHideDelay"]}],message:[{type:g.Input,args:["matTooltip"]}],tooltipClass:[{type:g.Input,args:["matTooltipClass"]}]},t}(),k=function(){function t(t,e){this._changeDetectorRef=t,this._breakpointObserver=e,this._visibility="initial",this._closeOnInteraction=!1,this._onHide=new y.Subject,this._isHandset=this._breakpointObserver.observe(c.Breakpoints.Handset)}return t.prototype.show=function(t){var e=this;this._hideTimeoutId&&(clearTimeout(this._hideTimeoutId),this._hideTimeoutId=null),this._closeOnInteraction=!0,this._showTimeoutId=setTimeout(function(){e._visibility="visible",e._showTimeoutId=null,e._markForCheck()},t)},t.prototype.hide=function(t){var e=this;this._showTimeoutId&&(clearTimeout(this._showTimeoutId),this._showTimeoutId=null),this._hideTimeoutId=setTimeout(function(){e._visibility="hidden",e._hideTimeoutId=null,e._markForCheck()},t)},t.prototype.afterHidden=function(){return this._onHide.asObservable()},t.prototype.isVisible=function(){return"visible"===this._visibility},t.prototype._animationStart=function(){this._closeOnInteraction=!1},t.prototype._animationDone=function(t){var e=t.toState;"hidden"!==e||this.isVisible()||this._onHide.next(),"visible"!==e&&"hidden"!==e||(this._closeOnInteraction=!0)},t.prototype._handleBodyInteraction=function(){this._closeOnInteraction&&this.hide(0)},t.prototype._markForCheck=function(){this._changeDetectorRef.markForCheck()},t.decorators=[{type:g.Component,args:[{selector:"mat-tooltip-component",template:'<div class="mat-tooltip" [ngClass]="tooltipClass" [class.mat-tooltip-handset]="(_isHandset | async)?.matches" [@state]="_visibility" (@state.start)="_animationStart()" (@state.done)="_animationDone($event)">{{message}}</div>',styles:[".mat-tooltip-panel{pointer-events:none!important}.mat-tooltip{color:#fff;border-radius:4px;margin:14px;max-width:250px;padding-left:8px;padding-right:8px;overflow:hidden;text-overflow:ellipsis}@media (-ms-high-contrast:active){.mat-tooltip{outline:solid 1px}}.mat-tooltip-handset{margin:24px;padding-left:16px;padding-right:16px}"],encapsulation:g.ViewEncapsulation.None,changeDetection:g.ChangeDetectionStrategy.OnPush,animations:[b.tooltipState],host:{"[style.zoom]":'_visibility === "visible" ? 1 : null',"(body:click)":"this._handleBodyInteraction()","aria-hidden":"true"}}]}],t.ctorParameters=function(){return[{type:g.ChangeDetectorRef},{type:c.BreakpointObserver}]},t}(),A=function(){function t(){}return t.decorators=[{type:g.NgModule,args:[{imports:[o.A11yModule,_.CommonModule,d.OverlayModule,v.MatCommonModule],exports:[T,k,v.MatCommonModule],declarations:[T,k],entryComponents:[k],providers:[E,{provide:l.HAMMER_GESTURE_CONFIG,useClass:v.GestureConfig}]}]}],t}()},"./node_modules/@angular/material/esm5/tree.es5.js":function(t,e,n){"use strict";n.r(e),n.d(e,"_MatTreeNodeMixinBase",function(){return p}),n.d(e,"_MatNestedTreeNodeMixinBase",function(){return f}),n.d(e,"MatTreeNode",function(){return h}),n.d(e,"MatTreeNodeDef",function(){return m}),n.d(e,"MatNestedTreeNode",function(){return g}),n.d(e,"MatTreeNodePadding",function(){return y}),n.d(e,"MatTree",function(){return _}),n.d(e,"MatTreeModule",function(){return x}),n.d(e,"MatTreeNodeToggle",function(){return v}),n.d(e,"MatTreeNodeOutlet",function(){return d}),n.d(e,"MatTreeFlattener",function(){return w}),n.d(e,"MatTreeFlatDataSource",function(){return j}),n.d(e,"MatTreeNestedDataSource",function(){return O});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/cdk/esm5/tree.es5.js"),a=n("./node_modules/@angular/material/esm5/core.es5.js"),s=n("./node_modules/@angular/common/fesm5/common.js"),u=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),c=n("./node_modules/rxjs/_esm5/index.js"),l=n("./node_modules/rxjs/_esm5/operators/index.js"),d=function(){function t(t){this.viewContainer=t}return t.decorators=[{type:r.Directive,args:[{selector:"[matTreeNodeOutlet]"}]}],t.ctorParameters=function(){return[{type:r.ViewContainerRef}]},t}(),p=Object(a.mixinTabIndex)(Object(a.mixinDisabled)(o.CdkTreeNode)),f=Object(a.mixinTabIndex)(Object(a.mixinDisabled)(o.CdkNestedTreeNode)),h=function(t){function e(e,n,r){var i=t.call(this,e,n)||this;return i._elementRef=e,i._tree=n,i.role="treeitem",i.tabIndex=Number(r)||0,i}return Object(i.__extends)(e,t),e.decorators=[{type:r.Directive,args:[{selector:"mat-tree-node",exportAs:"matTreeNode",inputs:["disabled","tabIndex"],host:{"[attr.aria-expanded]":"isExpanded","[attr.aria-level]":'role === "treeitem" ? level : null',"[attr.role]":"role",class:"mat-tree-node"},providers:[{provide:o.CdkTreeNode,useExisting:e}]}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:o.CdkTree},{type:String,decorators:[{type:r.Attribute,args:["tabindex"]}]}]},e.propDecorators={role:[{type:r.Input}]},e}(p),m=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(i.__extends)(e,t),e.decorators=[{type:r.Directive,args:[{selector:"[matTreeNodeDef]",inputs:["when: matTreeNodeDefWhen"],providers:[{provide:o.CdkTreeNodeDef,useExisting:e}]}]}],e.propDecorators={data:[{type:r.Input,args:["matTreeNode"]}]},e}(o.CdkTreeNodeDef),g=function(t){function e(e,n,r,i){var o=t.call(this,e,n,r)||this;return o._elementRef=e,o._tree=n,o._differs=r,o.tabIndex=Number(i)||0,o}return Object(i.__extends)(e,t),e.prototype.ngAfterContentInit=function(){t.prototype.ngAfterContentInit.call(this)},e.prototype.ngOnDestroy=function(){t.prototype.ngOnDestroy.call(this)},e.decorators=[{type:r.Directive,args:[{selector:"mat-nested-tree-node",exportAs:"matNestedTreeNode",host:{"[attr.aria-expanded]":"isExpanded","[attr.role]":"role",class:"mat-nested-tree-node"},inputs:["disabled","tabIndex"],providers:[{provide:o.CdkNestedTreeNode,useExisting:e},{provide:o.CdkTreeNode,useExisting:e}]}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:o.CdkTree},{type:r.IterableDiffers},{type:String,decorators:[{type:r.Attribute,args:["tabindex"]}]}]},e.propDecorators={node:[{type:r.Input,args:["matNestedTreeNode"]}],nodeOutlet:[{type:r.ContentChildren,args:[d]}]},e}(f),y=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(i.__extends)(e,t),e.decorators=[{type:r.Directive,args:[{selector:"[matTreeNodePadding]",providers:[{provide:o.CdkTreeNodePadding,useExisting:e}]}]}],e.propDecorators={level:[{type:r.Input,args:["matTreeNodePadding"]}],indent:[{type:r.Input,args:["matTreeNodePaddingIndent"]}]},e}(o.CdkTreeNodePadding),_=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(i.__extends)(e,t),e.decorators=[{type:r.Component,args:[{selector:"mat-tree",exportAs:"matTree",template:"<ng-container matTreeNodeOutlet></ng-container>",host:{class:"mat-tree",role:"tree"},styles:[".mat-tree{display:block}.mat-tree-node{display:flex;align-items:center;min-height:48px;flex:1;overflow:hidden;word-wrap:break-word}.mat-nested-tree-ndoe{border-bottom-width:0}"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,providers:[{provide:o.CdkTree,useExisting:e}]}]}],e.propDecorators={_nodeOutlet:[{type:r.ViewChild,args:[d]}]},e}(o.CdkTree),v=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.recursive=!1,e}return Object(i.__extends)(e,t),e.decorators=[{type:r.Directive,args:[{selector:"[matTreeNodeToggle]",host:{"(click)":"_toggle($event)"},providers:[{provide:o.CdkTreeNodeToggle,useExisting:e}]}]}],e.propDecorators={recursive:[{type:r.Input,args:["matTreeNodeToggleRecursive"]}]},e}(o.CdkTreeNodeToggle),b=[g,m,y,v,_,h,d],x=function(){function t(){}return t.decorators=[{type:r.NgModule,args:[{imports:[o.CdkTreeModule,s.CommonModule,a.MatCommonModule],exports:b,declarations:b}]}],t}(),w=function(){function t(t,e,n,r){this.transformFunction=t,this.getLevel=e,this.isExpandable=n,this.getChildren=r}return t.prototype._flattenNode=function(t,e,n,r){var i=this,o=this.transformFunction(t,e);if(n.push(o),this.isExpandable(o)){var a=this.getChildren(t);a&&(Array.isArray(a)?this._flattenChildren(a,e,n,r):a.pipe(Object(l.take)(1)).subscribe(function(t){i._flattenChildren(t,e,n,r)}))}return n},t.prototype._flattenChildren=function(t,e,n,r){var i=this;t.forEach(function(o,a){var s=r.slice();s.push(a!=t.length-1),i._flattenNode(o,e+1,n,s)})},t.prototype.flattenNodes=function(t){var e=this,n=[];return t.forEach(function(t){return e._flattenNode(t,0,n,[])}),n},t.prototype.expandFlattenedNodes=function(t,e){var n=this,r=[],i=[];return i[0]=!0,t.forEach(function(t){for(var o=!0,a=0;a<=n.getLevel(t);a++)o=o&&i[a];o&&r.push(t),n.isExpandable(t)&&(i[n.getLevel(t)+1]=e.isExpanded(t))}),r},t}(),j=function(t){function e(e,n,r){void 0===r&&(r=[]);var i=t.call(this)||this;return i.treeControl=e,i.treeFlattener=n,i._flattenedData=new c.BehaviorSubject([]),i._expandedData=new c.BehaviorSubject([]),i._data=new c.BehaviorSubject(r),i}return Object(i.__extends)(e,t),Object.defineProperty(e.prototype,"data",{get:function(){return this._data.value},set:function(t){this._data.next(t),this._flattenedData.next(this.treeFlattener.flattenNodes(this.data)),this.treeControl.dataNodes=this._flattenedData.value},enumerable:!0,configurable:!0}),e.prototype.connect=function(t){var e=this,n=[t.viewChange,this.treeControl.expansionModel.onChange,this._flattenedData];return c.merge.apply(void 0,n).pipe(Object(l.map)(function(){return e._expandedData.next(e.treeFlattener.expandFlattenedNodes(e._flattenedData.value,e.treeControl)),e._expandedData.value}))},e.prototype.disconnect=function(){},e}(u.DataSource),O=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._data=new c.BehaviorSubject([]),e}return Object(i.__extends)(e,t),Object.defineProperty(e.prototype,"data",{get:function(){return this._data.value},set:function(t){this._data.next(t)},enumerable:!0,configurable:!0}),e.prototype.connect=function(t){var e=this;return c.merge.apply(void 0,[t.viewChange,this._data]).pipe(Object(l.map)(function(){return e.data}))},e.prototype.disconnect=function(){},e}(u.DataSource)},"./node_modules/@angular/platform-browser-dynamic/fesm5/platform-browser-dynamic.js":function(t,e,n){"use strict";n.r(e),n.d(e,"ɵangular_packages_platform_browser_dynamic_platform_browser_dynamic_a",function(){return w}),n.d(e,"RESOURCE_CACHE_PROVIDER",function(){return O}),n.d(e,"platformBrowserDynamic",function(){return C}),n.d(e,"VERSION",function(){return j}),n.d(e,"JitCompilerFactory",function(){return y}),n.d(e,"ɵCompilerImpl",function(){return m}),n.d(e,"ɵplatformCoreDynamic",function(){return v}),n.d(e,"ɵINTERNAL_BROWSER_DYNAMIC_PLATFORM_PROVIDERS",function(){return x}),n.d(e,"ɵResourceLoaderImpl",function(){return b});var r,i=n("./node_modules/@angular/compiler/fesm5/compiler.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/platform-browser-dynamic/node_modules/tslib/tslib.es6.js"),s=n("./node_modules/@angular/common/fesm5/common.js"),u=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),c=((r=new Map).set(i.Identifiers.ANALYZE_FOR_ENTRY_COMPONENTS,o.ANALYZE_FOR_ENTRY_COMPONENTS),r.set(i.Identifiers.ElementRef,o.ElementRef),r.set(i.Identifiers.NgModuleRef,o.NgModuleRef),r.set(i.Identifiers.ViewContainerRef,o.ViewContainerRef),r.set(i.Identifiers.ChangeDetectorRef,o.ChangeDetectorRef),r.set(i.Identifiers.Renderer2,o.Renderer2),r.set(i.Identifiers.QueryList,o.QueryList),r.set(i.Identifiers.TemplateRef,o.TemplateRef),r.set(i.Identifiers.CodegenComponentFactoryResolver,o["ɵCodegenComponentFactoryResolver"]),r.set(i.Identifiers.ComponentFactoryResolver,o.ComponentFactoryResolver),r.set(i.Identifiers.ComponentFactory,o.ComponentFactory),r.set(i.Identifiers.ComponentRef,o.ComponentRef),r.set(i.Identifiers.NgModuleFactory,o.NgModuleFactory),r.set(i.Identifiers.createModuleFactory,o["ɵcmf"]),r.set(i.Identifiers.moduleDef,o["ɵmod"]),r.set(i.Identifiers.moduleProviderDef,o["ɵmpd"]),r.set(i.Identifiers.RegisterModuleFactoryFn,o["ɵregisterModuleFactory"]),r.set(i.Identifiers.Injector,o.Injector),r.set(i.Identifiers.ViewEncapsulation,o.ViewEncapsulation),r.set(i.Identifiers.ChangeDetectionStrategy,o.ChangeDetectionStrategy),r.set(i.Identifiers.SecurityContext,o.SecurityContext),r.set(i.Identifiers.LOCALE_ID,o.LOCALE_ID),r.set(i.Identifiers.TRANSLATIONS_FORMAT,o.TRANSLATIONS_FORMAT),r.set(i.Identifiers.inlineInterpolate,o["ɵinlineInterpolate"]),r.set(i.Identifiers.interpolate,o["ɵinterpolate"]),r.set(i.Identifiers.EMPTY_ARRAY,o["ɵEMPTY_ARRAY"]),r.set(i.Identifiers.EMPTY_MAP,o["ɵEMPTY_MAP"]),r.set(i.Identifiers.Renderer,o.Renderer),r.set(i.Identifiers.viewDef,o["ɵvid"]),r.set(i.Identifiers.elementDef,o["ɵeld"]),r.set(i.Identifiers.anchorDef,o["ɵand"]),r.set(i.Identifiers.textDef,o["ɵted"]),r.set(i.Identifiers.directiveDef,o["ɵdid"]),r.set(i.Identifiers.providerDef,o["ɵprd"]),r.set(i.Identifiers.queryDef,o["ɵqud"]),r.set(i.Identifiers.pureArrayDef,o["ɵpad"]),r.set(i.Identifiers.pureObjectDef,o["ɵpod"]),r.set(i.Identifiers.purePipeDef,o["ɵppd"]),r.set(i.Identifiers.pipeDef,o["ɵpid"]),r.set(i.Identifiers.nodeValue,o["ɵnov"]),r.set(i.Identifiers.ngContentDef,o["ɵncd"]),r.set(i.Identifiers.unwrapValue,o["ɵunv"]),r.set(i.Identifiers.createRendererType2,o["ɵcrt"]),r.set(i.Identifiers.createComponentFactory,o["ɵccf"]),r),l=function(){function t(){this.reflectionCapabilities=new o["ɵReflectionCapabilities"]}return t.prototype.componentModuleUrl=function(t,e){var n=e.moduleId;if("string"==typeof n)return Object(i.getUrlScheme)(n)?n:"package:"+n;if(null!==n&&void 0!==n)throw Object(i.syntaxError)('moduleId should be a string in "'+Object(o["ɵstringify"])(t)+"\". See https://goo.gl/wIDDiL for more information.\nIf you're using Webpack you should inline the template and the styles, see https://goo.gl/X2J8zc.");return"./"+Object(o["ɵstringify"])(t)},t.prototype.parameters=function(t){return this.reflectionCapabilities.parameters(t)},t.prototype.tryAnnotations=function(t){return this.annotations(t)},t.prototype.annotations=function(t){return this.reflectionCapabilities.annotations(t)},t.prototype.shallowAnnotations=function(t){throw new Error("Not supported in JIT mode")},t.prototype.propMetadata=function(t){return this.reflectionCapabilities.propMetadata(t)},t.prototype.hasLifecycleHook=function(t,e){return this.reflectionCapabilities.hasLifecycleHook(t,e)},t.prototype.guards=function(t){return this.reflectionCapabilities.guards(t)},t.prototype.resolveExternalReference=function(t){return c.get(t)||t.runtime},t}();var d=new o.InjectionToken("ErrorCollector"),p={provide:o.PACKAGE_ROOT_URL,useValue:"/"},f={get:function(t){throw new Error("No ResourceLoader implementation has been provided. Can't read the url \""+t+'"')}},h=new o.InjectionToken("HtmlParser"),m=function(){function t(t,e,n,r,o,a,s,u,c,l){this._metadataResolver=e,this._delegate=new i.JitCompiler(e,n,r,o,a,s,u,c,l,this.getExtraNgModuleProviders.bind(this)),this.injector=t}return t.prototype.getExtraNgModuleProviders=function(){return[this._metadataResolver.getProviderMetadata(new i.ProviderMeta(o.Compiler,{useValue:this}))]},t.prototype.compileModuleSync=function(t){return this._delegate.compileModuleSync(t)},t.prototype.compileModuleAsync=function(t){return this._delegate.compileModuleAsync(t)},t.prototype.compileModuleAndAllComponentsSync=function(t){var e=this._delegate.compileModuleAndAllComponentsSync(t);return{ngModuleFactory:e.ngModuleFactory,componentFactories:e.componentFactories}},t.prototype.compileModuleAndAllComponentsAsync=function(t){return this._delegate.compileModuleAndAllComponentsAsync(t).then(function(t){return{ngModuleFactory:t.ngModuleFactory,componentFactories:t.componentFactories}})},t.prototype.loadAotSummaries=function(t){this._delegate.loadAotSummaries(t)},t.prototype.hasAotSummary=function(t){return this._delegate.hasAotSummary(t)},t.prototype.getComponentFactory=function(t){return this._delegate.getComponentFactory(t)},t.prototype.clearCache=function(){this._delegate.clearCache()},t.prototype.clearCacheFor=function(t){this._delegate.clearCacheFor(t)},t.prototype.getModuleId=function(t){var e=this._metadataResolver.getNgModuleMetadata(t);return e&&e.id||void 0},t}(),g=[{provide:i.CompileReflector,useValue:new l},{provide:i.ResourceLoader,useValue:f},{provide:i.JitSummaryResolver,deps:[]},{provide:i.SummaryResolver,useExisting:i.JitSummaryResolver},{provide:o["ɵConsole"],deps:[]},{provide:i.Lexer,deps:[]},{provide:i.Parser,deps:[i.Lexer]},{provide:h,useClass:i.HtmlParser,deps:[]},{provide:i.I18NHtmlParser,useFactory:function(t,e,n,r,a){var s=(e=e||"")?r.missingTranslation:o.MissingTranslationStrategy.Ignore;return new i.I18NHtmlParser(t,e,n,s,a)},deps:[h,[new o.Optional,new o.Inject(o.TRANSLATIONS)],[new o.Optional,new o.Inject(o.TRANSLATIONS_FORMAT)],[i.CompilerConfig],[o["ɵConsole"]]]},{provide:i.HtmlParser,useExisting:i.I18NHtmlParser},{provide:i.TemplateParser,deps:[i.CompilerConfig,i.CompileReflector,i.Parser,i.ElementSchemaRegistry,i.I18NHtmlParser,o["ɵConsole"]]},{provide:i.DirectiveNormalizer,deps:[i.ResourceLoader,i.UrlResolver,i.HtmlParser,i.CompilerConfig]},{provide:i.CompileMetadataResolver,deps:[i.CompilerConfig,i.HtmlParser,i.NgModuleResolver,i.DirectiveResolver,i.PipeResolver,i.SummaryResolver,i.ElementSchemaRegistry,i.DirectiveNormalizer,o["ɵConsole"],[o.Optional,i.StaticSymbolCache],i.CompileReflector,[o.Optional,d]]},p,{provide:i.StyleCompiler,deps:[i.UrlResolver]},{provide:i.ViewCompiler,deps:[i.CompileReflector]},{provide:i.NgModuleCompiler,deps:[i.CompileReflector]},{provide:i.CompilerConfig,useValue:new i.CompilerConfig},{provide:o.Compiler,useClass:m,deps:[o.Injector,i.CompileMetadataResolver,i.TemplateParser,i.StyleCompiler,i.ViewCompiler,i.NgModuleCompiler,i.SummaryResolver,i.CompileReflector,i.CompilerConfig,o["ɵConsole"]]},{provide:i.DomElementSchemaRegistry,deps:[]},{provide:i.ElementSchemaRegistry,useExisting:i.DomElementSchemaRegistry},{provide:i.UrlResolver,deps:[o.PACKAGE_ROOT_URL]},{provide:i.DirectiveResolver,deps:[i.CompileReflector]},{provide:i.PipeResolver,deps:[i.CompileReflector]},{provide:i.NgModuleResolver,deps:[i.CompileReflector]}],y=function(){function t(t){var e={useJit:!0,defaultEncapsulation:o.ViewEncapsulation.Emulated,missingTranslation:o.MissingTranslationStrategy.Warning};this._defaultOptions=Object(a.__spread)([e],t)}return t.prototype.createCompiler=function(t){void 0===t&&(t=[]);var e,n,r,s={useJit:_((e=this._defaultOptions.concat(t)).map(function(t){return t.useJit})),defaultEncapsulation:_(e.map(function(t){return t.defaultEncapsulation})),providers:(n=e.map(function(t){return t.providers}),r=[],n.forEach(function(t){return t&&r.push.apply(r,Object(a.__spread)(t))}),r),missingTranslation:_(e.map(function(t){return t.missingTranslation})),preserveWhitespaces:_(e.map(function(t){return t.preserveWhitespaces}))};return o.Injector.create([g,{provide:i.CompilerConfig,useFactory:function(){return new i.CompilerConfig({useJit:s.useJit,jitDevMode:Object(o.isDevMode)(),defaultEncapsulation:s.defaultEncapsulation,missingTranslation:s.missingTranslation,preserveWhitespaces:s.preserveWhitespaces})},deps:[]},s.providers]).get(o.Compiler)},t}();function _(t){for(var e=t.length-1;e>=0;e--)if(void 0!==t[e])return t[e]}var v=Object(o.createPlatformFactory)(o.platformCore,"coreDynamic",[{provide:o.COMPILER_OPTIONS,useValue:{},multi:!0},{provide:o.CompilerFactory,useClass:y,deps:[o.COMPILER_OPTIONS]}]),b=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(a.__extends)(e,t),e.prototype.get=function(t){var e,n,r=new Promise(function(t,r){e=t,n=r}),i=new XMLHttpRequest;return i.open("GET",t,!0),i.responseType="text",i.onload=function(){var r=i.response||i.responseText,o=1223===i.status?204:i.status;0===o&&(o=r?200:0),200<=o&&o<=300?e(r):n("Failed to load "+t)},i.onerror=function(){n("Failed to load "+t)},i.send(),r},e=Object(a.__decorate)([Object(o.Injectable)()],e)}(i.ResourceLoader),x=[u["ɵINTERNAL_BROWSER_PLATFORM_PROVIDERS"],{provide:o.COMPILER_OPTIONS,useValue:{providers:[{provide:i.ResourceLoader,useClass:b,deps:[]}]},multi:!0},{provide:o.PLATFORM_ID,useValue:s["ɵPLATFORM_BROWSER_ID"]}],w=function(t){function e(){var e=t.call(this)||this;if(e._cache=o["ɵglobal"].$templateCache,null==e._cache)throw new Error("CachedResourceLoader: Template cache was not found in $templateCache.");return e}return Object(a.__extends)(e,t),e.prototype.get=function(t){return this._cache.hasOwnProperty(t)?Promise.resolve(this._cache[t]):Promise.reject("CachedResourceLoader: Did not find cached template for "+t)},e}(i.ResourceLoader),j=new o.Version("7.2.0"),O=[{provide:i.ResourceLoader,useClass:w,deps:[]}],C=Object(o.createPlatformFactory)(v,"browserDynamic",x)},"./node_modules/@angular/platform-browser-dynamic/node_modules/tslib/tslib.es6.js":function(t,e,n){"use strict";n.r(e),n.d(e,"__extends",function(){return i}),n.d(e,"__assign",function(){return o}),n.d(e,"__rest",function(){return a}),n.d(e,"__decorate",function(){return s}),n.d(e,"__param",function(){return u}),n.d(e,"__metadata",function(){return c}),n.d(e,"__awaiter",function(){return l}),n.d(e,"__generator",function(){return d}),n.d(e,"__exportStar",function(){return p}),n.d(e,"__values",function(){return f}),n.d(e,"__read",function(){return h}),n.d(e,"__spread",function(){return m}),n.d(e,"__await",function(){return g}),n.d(e,"__asyncGenerator",function(){return y}),n.d(e,"__asyncDelegator",function(){return _}),n.d(e,"__asyncValues",function(){return v}),n.d(e,"__makeTemplateObject",function(){return b}),n.d(e,"__importStar",function(){return x}),n.d(e,"__importDefault",function(){return w});var r=function(t,e){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)};function i(t,e){function n(){this.constructor=t}r(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}var o=function(){return(o=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++){e=arguments[n];for(var i in e)Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i])}return t}).apply(this,arguments)};function a(t,e){var n={};for(var r in t)Object.prototype.hasOwnProperty.call(t,r)&&e.indexOf(r)<0&&(n[r]=t[r]);if(null!=t&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(t);i<r.length;i++)e.indexOf(r[i])<0&&(n[r[i]]=t[r[i]])}return n}function s(t,e,n,r){var i,o=arguments.length,a=o<3?e:null===r?r=Object.getOwnPropertyDescriptor(e,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(t,e,n,r);else for(var s=t.length-1;s>=0;s--)(i=t[s])&&(a=(o<3?i(a):o>3?i(e,n,a):i(e,n))||a);return o>3&&a&&Object.defineProperty(e,n,a),a}function u(t,e){return function(n,r){e(n,r,t)}}function c(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)}function l(t,e,n,r){return new(n||(n=Promise))(function(i,o){function a(t){try{u(r.next(t))}catch(t){o(t)}}function s(t){try{u(r.throw(t))}catch(t){o(t)}}function u(t){t.done?i(t.value):new n(function(e){e(t.value)}).then(a,s)}u((r=r.apply(t,e||[])).next())})}function d(t,e){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=e.call(t,a)}catch(t){o=[6,t],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(t,e){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}function f(t){var e="function"==typeof Symbol&&t[Symbol.iterator],n=0;return e?e.call(t):{next:function(){return t&&n>=t.length&&(t=void 0),{value:t&&t[n++],done:!t}}}}function h(t,e){var n="function"==typeof Symbol&&t[Symbol.iterator];if(!n)return t;var r,i,o=n.call(t),a=[];try{for(;(void 0===e||e-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(t){i={error:t}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var t=[],e=0;e<arguments.length;e++)t=t.concat(h(arguments[e]));return t}function g(t){return this instanceof g?(this.v=t,this):new g(t)}function y(t,e,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(t,e||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(t){i[t]&&(r[t]=function(e){return new Promise(function(n,r){o.push([t,e,n,r])>1||s(t,e)})})}function s(t,e){try{(n=i[t](e)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(t){l(o[0][3],t)}var n}function u(t){s("next",t)}function c(t){s("throw",t)}function l(t,e){t(e),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(t){var e,n;return e={},r("next"),r("throw",function(t){throw t}),r("return"),e[Symbol.iterator]=function(){return this},e;function r(r,i){e[r]=t[r]?function(e){return(n=!n)?{value:g(t[r](e)),done:"return"===r}:i?i(e):e}:i}}function v(t){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var e,n=t[Symbol.asyncIterator];return n?n.call(t):(t=f(t),e={},r("next"),r("throw"),r("return"),e[Symbol.asyncIterator]=function(){return this},e);function r(n){e[n]=t[n]&&function(e){return new Promise(function(r,i){var o,a,s,u;e=t[n](e),o=r,a=i,s=e.done,u=e.value,Promise.resolve(u).then(function(t){o({value:t,done:s})},a)})}}}function b(t,e){return Object.defineProperty?Object.defineProperty(t,"raw",{value:e}):t.raw=e,t}function x(t){if(t&&t.__esModule)return t;var e={};if(null!=t)for(var n in t)Object.hasOwnProperty.call(t,n)&&(e[n]=t[n]);return e.default=t,e}function w(t){return t&&t.__esModule?t:{default:t}}},"./node_modules/@angular/platform-browser/fesm5/animations.js":function(t,e,n){"use strict";n.r(e),n.d(e,"ɵangular_packages_platform_browser_animations_animations_f",function(){return h}),n.d(e,"ɵangular_packages_platform_browser_animations_animations_d",function(){return w}),n.d(e,"ɵangular_packages_platform_browser_animations_animations_e",function(){return j}),n.d(e,"ɵangular_packages_platform_browser_animations_animations_b",function(){return _}),n.d(e,"ɵangular_packages_platform_browser_animations_animations_c",function(){return v}),n.d(e,"ɵangular_packages_platform_browser_animations_animations_a",function(){return y}),n.d(e,"BrowserAnimationsModule",function(){return O}),n.d(e,"NoopAnimationsModule",function(){return C}),n.d(e,"ANIMATION_MODULE_TYPE",function(){return b}),n.d(e,"ɵBrowserAnimationBuilder",function(){return c}),n.d(e,"ɵBrowserAnimationFactory",function(){return l}),n.d(e,"ɵAnimationRenderer",function(){return m}),n.d(e,"ɵAnimationRendererFactory",function(){return f}),n.d(e,"ɵInjectableAnimationEngine",function(){return g});var r=n("./node_modules/@angular/platform-browser/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/animations/fesm5/animations.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),s=n("./node_modules/@angular/animations/fesm5/browser.js"),u=n("./node_modules/@angular/common/fesm5/common.js"),c=function(t){function e(e,n){var r=t.call(this)||this;r._nextAnimationId=0;var i={id:"0",encapsulation:o.ViewEncapsulation.None,styles:[],data:{animation:[]}};return r._renderer=e.createRenderer(n.body,i),r}return Object(r.__extends)(e,t),e.prototype.build=function(t){var e=this._nextAnimationId.toString();this._nextAnimationId++;var n=Array.isArray(t)?Object(i.sequence)(t):t;return p(this._renderer,null,e,"register",[n]),new l(e,this._renderer)},e=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(1,Object(o.Inject)(a.DOCUMENT)),Object(r.__metadata)("design:paramtypes",[o.RendererFactory2,Object])],e)}(i.AnimationBuilder),l=function(t){function e(e,n){var r=t.call(this)||this;return r._id=e,r._renderer=n,r}return Object(r.__extends)(e,t),e.prototype.create=function(t,e){return new d(this._id,t,e||{},this._renderer)},e}(i.AnimationFactory),d=function(){function t(t,e,n,r){this.id=t,this.element=e,this._renderer=r,this.parentPlayer=null,this._started=!1,this.totalTime=0,this._command("create",n)}return t.prototype._listen=function(t,e){return this._renderer.listen(this.element,"@@"+this.id+":"+t,e)},t.prototype._command=function(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];return p(this._renderer,this.element,this.id,t,e)},t.prototype.onDone=function(t){this._listen("done",t)},t.prototype.onStart=function(t){this._listen("start",t)},t.prototype.onDestroy=function(t){this._listen("destroy",t)},t.prototype.init=function(){this._command("init")},t.prototype.hasStarted=function(){return this._started},t.prototype.play=function(){this._command("play"),this._started=!0},t.prototype.pause=function(){this._command("pause")},t.prototype.restart=function(){this._command("restart")},t.prototype.finish=function(){this._command("finish")},t.prototype.destroy=function(){this._command("destroy")},t.prototype.reset=function(){this._command("reset")},t.prototype.setPosition=function(t){this._command("setPosition",t)},t.prototype.getPosition=function(){return 0},t}();function p(t,e,n,r,i){return t.setProperty(e,"@@"+n+":"+r,i)}var f=function(){function t(t,e,n){this.delegate=t,this.engine=e,this._zone=n,this._currentId=0,this._microtaskId=1,this._animationCallbacksBuffer=[],this._rendererCache=new Map,this._cdRecurDepth=0,this.promise=Promise.resolve(0),e.onRemovalComplete=function(t,e){e&&e.parentNode(t)&&e.removeChild(t.parentNode,t)}}return t.prototype.createRenderer=function(t,e){var n=this,r=this.delegate.createRenderer(t,e);if(!(t&&e&&e.data&&e.data.animation)){var i=this._rendererCache.get(r);return i||(i=new h("",r,this.engine),this._rendererCache.set(r,i)),i}var o=e.id,a=e.id+"-"+this._currentId;return this._currentId++,this.engine.register(a,t),e.data.animation.forEach(function(e){return n.engine.registerTrigger(o,a,t,e.name,e)}),new m(this,a,r,this.engine)},t.prototype.begin=function(){this._cdRecurDepth++,this.delegate.begin&&this.delegate.begin()},t.prototype._scheduleCountTask=function(){var t=this;this.promise.then(function(){t._microtaskId++})},t.prototype.scheduleListenerCallback=function(t,e,n){var i=this;t>=0&&t<this._microtaskId?this._zone.run(function(){return e(n)}):(0==this._animationCallbacksBuffer.length&&Promise.resolve(null).then(function(){i._zone.run(function(){i._animationCallbacksBuffer.forEach(function(t){var e=Object(r.__read)(t,2);(0,e[0])(e[1])}),i._animationCallbacksBuffer=[]})}),this._animationCallbacksBuffer.push([e,n]))},t.prototype.end=function(){var t=this;this._cdRecurDepth--,0==this._cdRecurDepth&&this._zone.runOutsideAngular(function(){t._scheduleCountTask(),t.engine.flush(t._microtaskId)}),this.delegate.end&&this.delegate.end()},t.prototype.whenRenderingDone=function(){return this.engine.whenRenderingDone()},t=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__metadata)("design:paramtypes",[o.RendererFactory2,s["ɵAnimationEngine"],o.NgZone])],t)}(),h=function(){function t(t,e,n){this.namespaceId=t,this.delegate=e,this.engine=n,this.destroyNode=this.delegate.destroyNode?function(t){return e.destroyNode(t)}:null}return Object.defineProperty(t.prototype,"data",{get:function(){return this.delegate.data},enumerable:!0,configurable:!0}),t.prototype.destroy=function(){this.engine.destroy(this.namespaceId,this.delegate),this.delegate.destroy()},t.prototype.createElement=function(t,e){return this.delegate.createElement(t,e)},t.prototype.createComment=function(t){return this.delegate.createComment(t)},t.prototype.createText=function(t){return this.delegate.createText(t)},t.prototype.appendChild=function(t,e){this.delegate.appendChild(t,e),this.engine.onInsert(this.namespaceId,e,t,!1)},t.prototype.insertBefore=function(t,e,n){this.delegate.insertBefore(t,e,n),this.engine.onInsert(this.namespaceId,e,t,!0)},t.prototype.removeChild=function(t,e){this.engine.onRemove(this.namespaceId,e,this.delegate)},t.prototype.selectRootElement=function(t,e){return this.delegate.selectRootElement(t,e)},t.prototype.parentNode=function(t){return this.delegate.parentNode(t)},t.prototype.nextSibling=function(t){return this.delegate.nextSibling(t)},t.prototype.setAttribute=function(t,e,n,r){this.delegate.setAttribute(t,e,n,r)},t.prototype.removeAttribute=function(t,e,n){this.delegate.removeAttribute(t,e,n)},t.prototype.addClass=function(t,e){this.delegate.addClass(t,e)},t.prototype.removeClass=function(t,e){this.delegate.removeClass(t,e)},t.prototype.setStyle=function(t,e,n,r){this.delegate.setStyle(t,e,n,r)},t.prototype.removeStyle=function(t,e,n){this.delegate.removeStyle(t,e,n)},t.prototype.setProperty=function(t,e,n){"@"==e.charAt(0)&&"@.disabled"==e?this.disableAnimations(t,!!n):this.delegate.setProperty(t,e,n)},t.prototype.setValue=function(t,e){this.delegate.setValue(t,e)},t.prototype.listen=function(t,e,n){return this.delegate.listen(t,e,n)},t.prototype.disableAnimations=function(t,e){this.engine.disableAnimations(t,e)},t}(),m=function(t){function e(e,n,r,i){var o=t.call(this,n,r,i)||this;return o.factory=e,o.namespaceId=n,o}return Object(r.__extends)(e,t),e.prototype.setProperty=function(t,e,n){"@"==e.charAt(0)?"."==e.charAt(1)&&"@.disabled"==e?(n=void 0===n||!!n,this.disableAnimations(t,n)):this.engine.process(this.namespaceId,t,e.substr(1),n):this.delegate.setProperty(t,e,n)},e.prototype.listen=function(t,e,n){var i,o,a,s,u,c=this;if("@"==e.charAt(0)){var l=function(t){switch(t){case"body":return document.body;case"document":return document;case"window":return window;default:return t}}(t),d=e.substr(1),p="";return"@"!=d.charAt(0)&&(d=(i=Object(r.__read)((a=(o=d).indexOf("."),s=o.substring(0,a),u=o.substr(a+1),[s,u]),2))[0],p=i[1]),this.engine.listen(this.namespaceId,l,d,p,function(t){var e=t._data||-1;c.factory.scheduleListenerCallback(e,n,t)})}return this.delegate.listen(t,e,n)},e}(h);var g=function(t){function e(e,n,r){return t.call(this,e.body,n,r)||this}return Object(r.__extends)(e,t),e=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(0,Object(o.Inject)(u.DOCUMENT)),Object(r.__metadata)("design:paramtypes",[Object,s.AnimationDriver,s["ɵAnimationStyleNormalizer"]])],e)}(s["ɵAnimationEngine"]);function y(){return Object(s["ɵsupportsWebAnimations"])()?new s["ɵWebAnimationsDriver"]:new s["ɵCssKeyframesDriver"]}function _(){return new s["ɵWebAnimationsStyleNormalizer"]}function v(t,e,n){return new f(t,e,n)}var b=new o.InjectionToken("AnimationModuleType"),x=[{provide:i.AnimationBuilder,useClass:c},{provide:s["ɵAnimationStyleNormalizer"],useFactory:_},{provide:s["ɵAnimationEngine"],useClass:g},{provide:o.RendererFactory2,useFactory:v,deps:[a["ɵDomRendererFactory2"],s["ɵAnimationEngine"],o.NgZone]}],w=Object(r.__spread)([{provide:s.AnimationDriver,useFactory:y},{provide:b,useValue:"BrowserAnimations"}],x),j=Object(r.__spread)([{provide:s.AnimationDriver,useClass:s["ɵNoopAnimationDriver"]},{provide:b,useValue:"NoopAnimations"}],x),O=function(){function t(){}return t=Object(r.__decorate)([Object(o.NgModule)({exports:[a.BrowserModule],providers:w})],t)}(),C=function(){function t(){}return t=Object(r.__decorate)([Object(o.NgModule)({exports:[a.BrowserModule],providers:j})],t)}()},"./node_modules/@angular/platform-browser/fesm5/platform-browser.js":function(t,e,n){"use strict";n.r(e),n.d(e,"ɵangular_packages_platform_browser_platform_browser_c",function(){return At}),n.d(e,"ɵangular_packages_platform_browser_platform_browser_b",function(){return kt}),n.d(e,"ɵangular_packages_platform_browser_platform_browser_a",function(){return Tt}),n.d(e,"ɵangular_packages_platform_browser_platform_browser_k",function(){return d}),n.d(e,"ɵangular_packages_platform_browser_platform_browser_d",function(){return It}),n.d(e,"ɵangular_packages_platform_browser_platform_browser_i",function(){return j}),n.d(e,"ɵangular_packages_platform_browser_platform_browser_h",function(){return w}),n.d(e,"ɵangular_packages_platform_browser_platform_browser_e",function(){return Rt}),n.d(e,"ɵangular_packages_platform_browser_platform_browser_f",function(){return Gt}),n.d(e,"ɵangular_packages_platform_browser_platform_browser_j",function(){return k}),n.d(e,"ɵangular_packages_platform_browser_platform_browser_g",function(){return P}),n.d(e,"BrowserModule",function(){return Dt}),n.d(e,"platformBrowser",function(){return St}),n.d(e,"Meta",function(){return Pt}),n.d(e,"Title",function(){return Nt}),n.d(e,"disableDebugTools",function(){return Ht}),n.d(e,"enableDebugTools",function(){return zt}),n.d(e,"BrowserTransferStateModule",function(){return Yt}),n.d(e,"TransferState",function(){return qt}),n.d(e,"makeStateKey",function(){return Wt}),n.d(e,"By",function(){return Kt}),n.d(e,"DOCUMENT",function(){return _}),n.d(e,"EVENT_MANAGER_PLUGINS",function(){return D}),n.d(e,"EventManager",function(){return I}),n.d(e,"HAMMER_GESTURE_CONFIG",function(){return lt}),n.d(e,"HAMMER_LOADER",function(){return dt}),n.d(e,"HammerGestureConfig",function(){return pt}),n.d(e,"DomSanitizer",function(){return yt}),n.d(e,"VERSION",function(){return Xt}),n.d(e,"ɵBROWSER_SANITIZATION_PROVIDERS",function(){return Et}),n.d(e,"ɵINTERNAL_BROWSER_PLATFORM_PROVIDERS",function(){return Ct}),n.d(e,"ɵinitDomAdapter",function(){return Mt}),n.d(e,"ɵBrowserDomAdapter",function(){return g}),n.d(e,"ɵBrowserPlatformLocation",function(){return b}),n.d(e,"ɵTRANSITION_ID",function(){return x}),n.d(e,"ɵBrowserGetTestability",function(){return O}),n.d(e,"ɵescapeHtml",function(){return Ut}),n.d(e,"ɵELEMENT_PROBE_PROVIDERS",function(){return A}),n.d(e,"ɵDomAdapter",function(){return l}),n.d(e,"ɵgetDOM",function(){return s}),n.d(e,"ɵsetRootDomAdapter",function(){return u}),n.d(e,"ɵDomRendererFactory2",function(){return q}),n.d(e,"ɵNAMESPACE_URIS",function(){return F}),n.d(e,"ɵflattenStyles",function(){return U}),n.d(e,"ɵshimContentAttribute",function(){return z}),n.d(e,"ɵshimHostAttribute",function(){return H}),n.d(e,"ɵDomEventsPlugin",function(){return ut}),n.d(e,"ɵHammerGesturesPlugin",function(){return ft}),n.d(e,"ɵKeyEventsPlugin",function(){return gt}),n.d(e,"ɵDomSharedStylesHost",function(){return N}),n.d(e,"ɵSharedStylesHost",function(){return R}),n.d(e,"ɵDomSanitizerImpl",function(){return _t});var r=n("./node_modules/@angular/platform-browser/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/common/fesm5/common.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=null;function s(){return a}function u(t){a||(a=t)}var c,l=function(){function t(){this.resourceLoaderType=null}return Object.defineProperty(t.prototype,"attrToPropMap",{get:function(){return this._attrToPropMap},set:function(t){this._attrToPropMap=t},enumerable:!0,configurable:!0}),t}(),d=function(t){function e(){var e=t.call(this)||this;e._animationPrefix=null,e._transitionEnd=null;try{var n=e.createElement("div",document);if(null!=e.getStyle(n,"animationName"))e._animationPrefix="";else for(var r=["Webkit","Moz","O","ms"],i=0;i<r.length;i++)if(null!=e.getStyle(n,r[i]+"AnimationName")){e._animationPrefix="-"+r[i].toLowerCase()+"-";break}var o={WebkitTransition:"webkitTransitionEnd",MozTransition:"transitionend",OTransition:"oTransitionEnd otransitionend",transition:"transitionend"};Object.keys(o).forEach(function(t){null!=e.getStyle(n,t)&&(e._transitionEnd=o[t])})}catch(t){e._animationPrefix=null,e._transitionEnd=null}return e}return Object(r.__extends)(e,t),e.prototype.getDistributedNodes=function(t){return t.getDistributedNodes()},e.prototype.resolveAndSetHref=function(t,e,n){t.href=null==n?e:e+"/../"+n},e.prototype.supportsDOMEvents=function(){return!0},e.prototype.supportsNativeShadowDOM=function(){return"function"==typeof document.body.createShadowRoot},e.prototype.getAnimationPrefix=function(){return this._animationPrefix?this._animationPrefix:""},e.prototype.getTransitionEnd=function(){return this._transitionEnd?this._transitionEnd:""},e.prototype.supportsAnimation=function(){return null!=this._animationPrefix&&null!=this._transitionEnd},e}(l),p={class:"className",innerHtml:"innerHTML",readonly:"readOnly",tabindex:"tabIndex"},f={"\b":"Backspace","\t":"Tab","":"Delete","":"Escape",Del:"Delete",Esc:"Escape",Left:"ArrowLeft",Right:"ArrowRight",Up:"ArrowUp",Down:"ArrowDown",Menu:"ContextMenu",Scroll:"ScrollLock",Win:"OS"},h={A:"1",B:"2",C:"3",D:"4",E:"5",F:"6",G:"7",H:"8",I:"9",J:"*",K:"+",M:"-",N:".",O:"/","`":"0","":"NumLock"};o["ɵglobal"].Node&&(c=o["ɵglobal"].Node.prototype.contains||function(t){return!!(16&this.compareDocumentPosition(t))});var m,g=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.parse=function(t){throw new Error("parse not implemented")},e.makeCurrent=function(){u(new e)},e.prototype.hasProperty=function(t,e){return e in t},e.prototype.setProperty=function(t,e,n){t[e]=n},e.prototype.getProperty=function(t,e){return t[e]},e.prototype.invoke=function(t,e,n){var i;(i=t)[e].apply(i,Object(r.__spread)(n))},e.prototype.logError=function(t){window.console&&(console.error?console.error(t):console.log(t))},e.prototype.log=function(t){window.console&&window.console.log&&window.console.log(t)},e.prototype.logGroup=function(t){window.console&&window.console.group&&window.console.group(t)},e.prototype.logGroupEnd=function(){window.console&&window.console.groupEnd&&window.console.groupEnd()},Object.defineProperty(e.prototype,"attrToPropMap",{get:function(){return p},enumerable:!0,configurable:!0}),e.prototype.contains=function(t,e){return c.call(t,e)},e.prototype.querySelector=function(t,e){return t.querySelector(e)},e.prototype.querySelectorAll=function(t,e){return t.querySelectorAll(e)},e.prototype.on=function(t,e,n){t.addEventListener(e,n,!1)},e.prototype.onAndCancel=function(t,e,n){return t.addEventListener(e,n,!1),function(){t.removeEventListener(e,n,!1)}},e.prototype.dispatchEvent=function(t,e){t.dispatchEvent(e)},e.prototype.createMouseEvent=function(t){var e=this.getDefaultDocument().createEvent("MouseEvent");return e.initEvent(t,!0,!0),e},e.prototype.createEvent=function(t){var e=this.getDefaultDocument().createEvent("Event");return e.initEvent(t,!0,!0),e},e.prototype.preventDefault=function(t){t.preventDefault(),t.returnValue=!1},e.prototype.isPrevented=function(t){return t.defaultPrevented||null!=t.returnValue&&!t.returnValue},e.prototype.getInnerHTML=function(t){return t.innerHTML},e.prototype.getTemplateContent=function(t){return"content"in t&&this.isTemplateElement(t)?t.content:null},e.prototype.getOuterHTML=function(t){return t.outerHTML},e.prototype.nodeName=function(t){return t.nodeName},e.prototype.nodeValue=function(t){return t.nodeValue},e.prototype.type=function(t){return t.type},e.prototype.content=function(t){return this.hasProperty(t,"content")?t.content:t},e.prototype.firstChild=function(t){return t.firstChild},e.prototype.nextSibling=function(t){return t.nextSibling},e.prototype.parentElement=function(t){return t.parentNode},e.prototype.childNodes=function(t){return t.childNodes},e.prototype.childNodesAsList=function(t){for(var e=t.childNodes,n=new Array(e.length),r=0;r<e.length;r++)n[r]=e[r];return n},e.prototype.clearNodes=function(t){for(;t.firstChild;)t.removeChild(t.firstChild)},e.prototype.appendChild=function(t,e){t.appendChild(e)},e.prototype.removeChild=function(t,e){t.removeChild(e)},e.prototype.replaceChild=function(t,e,n){t.replaceChild(e,n)},e.prototype.remove=function(t){return t.parentNode&&t.parentNode.removeChild(t),t},e.prototype.insertBefore=function(t,e,n){t.insertBefore(n,e)},e.prototype.insertAllBefore=function(t,e,n){n.forEach(function(n){return t.insertBefore(n,e)})},e.prototype.insertAfter=function(t,e,n){t.insertBefore(n,e.nextSibling)},e.prototype.setInnerHTML=function(t,e){t.innerHTML=e},e.prototype.getText=function(t){return t.textContent},e.prototype.setText=function(t,e){t.textContent=e},e.prototype.getValue=function(t){return t.value},e.prototype.setValue=function(t,e){t.value=e},e.prototype.getChecked=function(t){return t.checked},e.prototype.setChecked=function(t,e){t.checked=e},e.prototype.createComment=function(t){return this.getDefaultDocument().createComment(t)},e.prototype.createTemplate=function(t){var e=this.getDefaultDocument().createElement("template");return e.innerHTML=t,e},e.prototype.createElement=function(t,e){return(e=e||this.getDefaultDocument()).createElement(t)},e.prototype.createElementNS=function(t,e,n){return(n=n||this.getDefaultDocument()).createElementNS(t,e)},e.prototype.createTextNode=function(t,e){return(e=e||this.getDefaultDocument()).createTextNode(t)},e.prototype.createScriptTag=function(t,e,n){var r=(n=n||this.getDefaultDocument()).createElement("SCRIPT");return r.setAttribute(t,e),r},e.prototype.createStyleElement=function(t,e){var n=(e=e||this.getDefaultDocument()).createElement("style");return this.appendChild(n,this.createTextNode(t,e)),n},e.prototype.createShadowRoot=function(t){return t.createShadowRoot()},e.prototype.getShadowRoot=function(t){return t.shadowRoot},e.prototype.getHost=function(t){return t.host},e.prototype.clone=function(t){return t.cloneNode(!0)},e.prototype.getElementsByClassName=function(t,e){return t.getElementsByClassName(e)},e.prototype.getElementsByTagName=function(t,e){return t.getElementsByTagName(e)},e.prototype.classList=function(t){return Array.prototype.slice.call(t.classList,0)},e.prototype.addClass=function(t,e){t.classList.add(e)},e.prototype.removeClass=function(t,e){t.classList.remove(e)},e.prototype.hasClass=function(t,e){return t.classList.contains(e)},e.prototype.setStyle=function(t,e,n){t.style[e]=n},e.prototype.removeStyle=function(t,e){t.style[e]=""},e.prototype.getStyle=function(t,e){return t.style[e]},e.prototype.hasStyle=function(t,e,n){var r=this.getStyle(t,e)||"";return n?r==n:r.length>0},e.prototype.tagName=function(t){return t.tagName},e.prototype.attributeMap=function(t){for(var e=new Map,n=t.attributes,r=0;r<n.length;r++){var i=n.item(r);e.set(i.name,i.value)}return e},e.prototype.hasAttribute=function(t,e){return t.hasAttribute(e)},e.prototype.hasAttributeNS=function(t,e,n){return t.hasAttributeNS(e,n)},e.prototype.getAttribute=function(t,e){return t.getAttribute(e)},e.prototype.getAttributeNS=function(t,e,n){return t.getAttributeNS(e,n)},e.prototype.setAttribute=function(t,e,n){t.setAttribute(e,n)},e.prototype.setAttributeNS=function(t,e,n,r){t.setAttributeNS(e,n,r)},e.prototype.removeAttribute=function(t,e){t.removeAttribute(e)},e.prototype.removeAttributeNS=function(t,e,n){t.removeAttributeNS(e,n)},e.prototype.templateAwareRoot=function(t){return this.isTemplateElement(t)?this.content(t):t},e.prototype.createHtmlDocument=function(){return document.implementation.createHTMLDocument("fakeTitle")},e.prototype.getDefaultDocument=function(){return document},e.prototype.getBoundingClientRect=function(t){try{return t.getBoundingClientRect()}catch(t){return{top:0,bottom:0,left:0,right:0,width:0,height:0}}},e.prototype.getTitle=function(t){return t.title},e.prototype.setTitle=function(t,e){t.title=e||""},e.prototype.elementMatches=function(t,e){return!!this.isElementNode(t)&&(t.matches&&t.matches(e)||t.msMatchesSelector&&t.msMatchesSelector(e)||t.webkitMatchesSelector&&t.webkitMatchesSelector(e))},e.prototype.isTemplateElement=function(t){return this.isElementNode(t)&&"TEMPLATE"===t.nodeName},e.prototype.isTextNode=function(t){return t.nodeType===Node.TEXT_NODE},e.prototype.isCommentNode=function(t){return t.nodeType===Node.COMMENT_NODE},e.prototype.isElementNode=function(t){return t.nodeType===Node.ELEMENT_NODE},e.prototype.hasShadowRoot=function(t){return null!=t.shadowRoot&&t instanceof HTMLElement},e.prototype.isShadowRoot=function(t){return t instanceof DocumentFragment},e.prototype.importIntoDoc=function(t){return document.importNode(this.templateAwareRoot(t),!0)},e.prototype.adoptNode=function(t){return document.adoptNode(t)},e.prototype.getHref=function(t){return t.getAttribute("href")},e.prototype.getEventKey=function(t){var e=t.key;if(null==e){if(null==(e=t.keyIdentifier))return"Unidentified";e.startsWith("U+")&&(e=String.fromCharCode(parseInt(e.substring(2),16)),3===t.location&&h.hasOwnProperty(e)&&(e=h[e]))}return f[e]||e},e.prototype.getGlobalEventTarget=function(t,e){return"window"===e?window:"document"===e?t:"body"===e?t.body:null},e.prototype.getHistory=function(){return window.history},e.prototype.getLocation=function(){return window.location},e.prototype.getBaseHref=function(t){var e=function(){if(!y&&!(y=document.querySelector("base")))return null;return y.getAttribute("href")}();return null==e?null:function(t){m||(m=document.createElement("a"));return m.setAttribute("href",t),"/"===m.pathname.charAt(0)?m.pathname:"/"+m.pathname}(e)},e.prototype.resetBaseElement=function(){y=null},e.prototype.getUserAgent=function(){return window.navigator.userAgent},e.prototype.setData=function(t,e,n){this.setAttribute(t,"data-"+e,n)},e.prototype.getData=function(t,e){return this.getAttribute(t,"data-"+e)},e.prototype.getComputedStyle=function(t){return getComputedStyle(t)},e.prototype.supportsWebAnimation=function(){return"function"==typeof Element.prototype.animate},e.prototype.performanceNow=function(){return window.performance&&window.performance.now?window.performance.now():(new Date).getTime()},e.prototype.supportsCookies=function(){return!0},e.prototype.getCookie=function(t){return Object(i["ɵparseCookieValue"])(document.cookie,t)},e.prototype.setCookie=function(t,e){document.cookie=encodeURIComponent(t)+"="+encodeURIComponent(e)},e}(d),y=null;var _=i.DOCUMENT;function v(){return!!window.history.pushState}var b=function(t){function e(e){var n=t.call(this)||this;return n._doc=e,n._init(),n}return Object(r.__extends)(e,t),e.prototype._init=function(){this.location=s().getLocation(),this._history=s().getHistory()},e.prototype.getBaseHrefFromDOM=function(){return s().getBaseHref(this._doc)},e.prototype.onPopState=function(t){s().getGlobalEventTarget(this._doc,"window").addEventListener("popstate",t,!1)},e.prototype.onHashChange=function(t){s().getGlobalEventTarget(this._doc,"window").addEventListener("hashchange",t,!1)},Object.defineProperty(e.prototype,"pathname",{get:function(){return this.location.pathname},set:function(t){this.location.pathname=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"search",{get:function(){return this.location.search},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hash",{get:function(){return this.location.hash},enumerable:!0,configurable:!0}),e.prototype.pushState=function(t,e,n){v()?this._history.pushState(t,e,n):this.location.hash=n},e.prototype.replaceState=function(t,e,n){v()?this._history.replaceState(t,e,n):this.location.hash=n},e.prototype.forward=function(){this._history.forward()},e.prototype.back=function(){this._history.back()},e=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(0,Object(o.Inject)(_)),Object(r.__metadata)("design:paramtypes",[Object])],e)}(i.PlatformLocation),x=new o.InjectionToken("TRANSITION_ID");function w(t,e,n){return function(){n.get(o.ApplicationInitStatus).donePromise.then(function(){var n=s();Array.prototype.slice.apply(n.querySelectorAll(e,"style[ng-transition]")).filter(function(e){return n.getAttribute(e,"ng-transition")===t}).forEach(function(t){return n.remove(t)})})}}var j=[{provide:o.APP_INITIALIZER,useFactory:w,deps:[x,_,o.Injector],multi:!0}],O=function(){function t(){}return t.init=function(){Object(o.setTestabilityGetter)(new t)},t.prototype.addToWindow=function(t){o["ɵglobal"].getAngularTestability=function(e,n){void 0===n&&(n=!0);var r=t.findTestabilityInTree(e,n);if(null==r)throw new Error("Could not find testability for element.");return r},o["ɵglobal"].getAllAngularTestabilities=function(){return t.getAllTestabilities()},o["ɵglobal"].getAllAngularRootElements=function(){return t.getAllRootElements()};o["ɵglobal"].frameworkStabilizers||(o["ɵglobal"].frameworkStabilizers=[]),o["ɵglobal"].frameworkStabilizers.push(function(t){var e=o["ɵglobal"].getAllAngularTestabilities(),n=e.length,r=!1,i=function(e){r=r||e,0==--n&&t(r)};e.forEach(function(t){t.whenStable(i)})})},t.prototype.findTestabilityInTree=function(t,e,n){if(null==e)return null;var r=t.getTestability(e);return null!=r?r:n?s().isShadowRoot(e)?this.findTestabilityInTree(t,s().getHost(e),!0):this.findTestabilityInTree(t,s().parentElement(e),!0):null},t}();function C(t,e){"undefined"!=typeof COMPILED&&COMPILED||((o["ɵglobal"].ng=o["ɵglobal"].ng||{})[t]=e)}var E={ApplicationRef:o.ApplicationRef,NgZone:o.NgZone},S="probe",M="coreTokens";function T(t){return Object(o.getDebugNode)(t)}function k(t){return C(S,T),C(M,Object(r.__assign)({},E,(t||[]).reduce(function(t,e){return t[e.name]=e.token,t},{}))),function(){return T}}var A=[{provide:o.APP_INITIALIZER,useFactory:k,deps:[[o.NgProbeToken,new o.Optional]],multi:!0}],D=new o.InjectionToken("EventManagerPlugins"),I=function(){function t(t,e){var n=this;this._zone=e,this._eventNameToPlugin=new Map,t.forEach(function(t){return t.manager=n}),this._plugins=t.slice().reverse()}return t.prototype.addEventListener=function(t,e,n){return this._findPluginFor(e).addEventListener(t,e,n)},t.prototype.addGlobalEventListener=function(t,e,n){return this._findPluginFor(e).addGlobalEventListener(t,e,n)},t.prototype.getZone=function(){return this._zone},t.prototype._findPluginFor=function(t){var e=this._eventNameToPlugin.get(t);if(e)return e;for(var n=this._plugins,r=0;r<n.length;r++){var i=n[r];if(i.supports(t))return this._eventNameToPlugin.set(t,i),i}throw new Error("No event manager plugin found for event "+t)},t=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(0,Object(o.Inject)(D)),Object(r.__metadata)("design:paramtypes",[Array,o.NgZone])],t)}(),P=function(){function t(t){this._doc=t}return t.prototype.addGlobalEventListener=function(t,e,n){var r=s().getGlobalEventTarget(this._doc,t);if(!r)throw new Error("Unsupported event target "+r+" for event "+e);return this.addEventListener(r,e,n)},t}(),R=function(){function t(){this._stylesSet=new Set}return t.prototype.addStyles=function(t){var e=this,n=new Set;t.forEach(function(t){e._stylesSet.has(t)||(e._stylesSet.add(t),n.add(t))}),this.onStylesAdded(n)},t.prototype.onStylesAdded=function(t){},t.prototype.getAllStyles=function(){return Array.from(this._stylesSet)},t=Object(r.__decorate)([Object(o.Injectable)()],t)}(),N=function(t){function e(e){var n=t.call(this)||this;return n._doc=e,n._hostNodes=new Set,n._styleNodes=new Set,n._hostNodes.add(e.head),n}return Object(r.__extends)(e,t),e.prototype._addStylesToHost=function(t,e){var n=this;t.forEach(function(t){var r=n._doc.createElement("style");r.textContent=t,n._styleNodes.add(e.appendChild(r))})},e.prototype.addHost=function(t){this._addStylesToHost(this._stylesSet,t),this._hostNodes.add(t)},e.prototype.removeHost=function(t){this._hostNodes.delete(t)},e.prototype.onStylesAdded=function(t){var e=this;this._hostNodes.forEach(function(n){return e._addStylesToHost(t,n)})},e.prototype.ngOnDestroy=function(){this._styleNodes.forEach(function(t){return s().remove(t)})},e=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(0,Object(o.Inject)(_)),Object(r.__metadata)("design:paramtypes",[Object])],e)}(R),F={svg:"http://www.w3.org/2000/svg",xhtml:"http://www.w3.org/1999/xhtml",xlink:"http://www.w3.org/1999/xlink",xml:"http://www.w3.org/XML/1998/namespace",xmlns:"http://www.w3.org/2000/xmlns/"},L=/%COMP%/g,B="_nghost-%COMP%",V="_ngcontent-%COMP%";function z(t){return V.replace(L,t)}function H(t){return B.replace(L,t)}function U(t,e,n){for(var r=0;r<e.length;r++){var i=e[r];Array.isArray(i)?U(t,i,n):(i=i.replace(L,t),n.push(i))}return n}function W(t){return function(e){!1===t(e)&&(e.preventDefault(),e.returnValue=!1)}}var q=function(){function t(t,e){this.eventManager=t,this.sharedStylesHost=e,this.rendererByCompId=new Map,this.defaultRenderer=new G(t)}return t.prototype.createRenderer=function(t,e){if(!t||!e)return this.defaultRenderer;switch(e.encapsulation){case o.ViewEncapsulation.Emulated:var n=this.rendererByCompId.get(e.id);return n||(n=new Q(this.eventManager,this.sharedStylesHost,e),this.rendererByCompId.set(e.id,n)),n.applyToHost(t),n;case o.ViewEncapsulation.Native:case o.ViewEncapsulation.ShadowDom:return new Z(this.eventManager,this.sharedStylesHost,t,e);default:if(!this.rendererByCompId.has(e.id)){var r=U(e.id,e.styles,[]);this.sharedStylesHost.addStyles(r),this.rendererByCompId.set(e.id,this.defaultRenderer)}return this.defaultRenderer}},t.prototype.begin=function(){},t.prototype.end=function(){},t=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__metadata)("design:paramtypes",[I,N])],t)}(),G=function(){function t(t){this.eventManager=t,this.data=Object.create(null)}return t.prototype.destroy=function(){},t.prototype.createElement=function(t,e){return e?document.createElementNS(F[e],t):document.createElement(t)},t.prototype.createComment=function(t){return document.createComment(t)},t.prototype.createText=function(t){return document.createTextNode(t)},t.prototype.appendChild=function(t,e){t.appendChild(e)},t.prototype.insertBefore=function(t,e,n){t&&t.insertBefore(e,n)},t.prototype.removeChild=function(t,e){t&&t.removeChild(e)},t.prototype.selectRootElement=function(t,e){var n="string"==typeof t?document.querySelector(t):t;if(!n)throw new Error('The selector "'+t+'" did not match any elements');return e||(n.textContent=""),n},t.prototype.parentNode=function(t){return t.parentNode},t.prototype.nextSibling=function(t){return t.nextSibling},t.prototype.setAttribute=function(t,e,n,r){if(r){e=r+":"+e;var i=F[r];i?t.setAttributeNS(i,e,n):t.setAttribute(e,n)}else t.setAttribute(e,n)},t.prototype.removeAttribute=function(t,e,n){if(n){var r=F[n];r?t.removeAttributeNS(r,e):t.removeAttribute(n+":"+e)}else t.removeAttribute(e)},t.prototype.addClass=function(t,e){t.classList.add(e)},t.prototype.removeClass=function(t,e){t.classList.remove(e)},t.prototype.setStyle=function(t,e,n,r){r&o.RendererStyleFlags2.DashCase?t.style.setProperty(e,n,r&o.RendererStyleFlags2.Important?"important":""):t.style[e]=n},t.prototype.removeStyle=function(t,e,n){n&o.RendererStyleFlags2.DashCase?t.style.removeProperty(e):t.style[e]=""},t.prototype.setProperty=function(t,e,n){K(e,"property"),t[e]=n},t.prototype.setValue=function(t,e){t.nodeValue=e},t.prototype.listen=function(t,e,n){return K(e,"listener"),"string"==typeof t?this.eventManager.addGlobalEventListener(t,e,W(n)):this.eventManager.addEventListener(t,e,W(n))},t}(),Y="@".charCodeAt(0);function K(t,e){if(t.charCodeAt(0)===Y)throw new Error("Found the synthetic "+e+" "+t+'. Please include either "BrowserAnimationsModule" or "NoopAnimationsModule" in your application.')}var X,Q=function(t){function e(e,n,r){var i=t.call(this,e)||this;i.component=r;var o=U(r.id,r.styles,[]);return n.addStyles(o),i.contentAttr=z(r.id),i.hostAttr=H(r.id),i}return Object(r.__extends)(e,t),e.prototype.applyToHost=function(e){t.prototype.setAttribute.call(this,e,this.hostAttr,"")},e.prototype.createElement=function(e,n){var r=t.prototype.createElement.call(this,e,n);return t.prototype.setAttribute.call(this,r,this.contentAttr,""),r},e}(G),Z=function(t){function e(e,n,r,i){var a=t.call(this,e)||this;a.sharedStylesHost=n,a.hostEl=r,a.component=i,i.encapsulation===o.ViewEncapsulation.ShadowDom?a.shadowRoot=r.attachShadow({mode:"open"}):a.shadowRoot=r.createShadowRoot(),a.sharedStylesHost.addHost(a.shadowRoot);for(var s=U(i.id,i.styles,[]),u=0;u<s.length;u++){var c=document.createElement("style");c.textContent=s[u],a.shadowRoot.appendChild(c)}return a}return Object(r.__extends)(e,t),e.prototype.nodeOrShadowRoot=function(t){return t===this.hostEl?this.shadowRoot:t},e.prototype.destroy=function(){this.sharedStylesHost.removeHost(this.shadowRoot)},e.prototype.appendChild=function(e,n){return t.prototype.appendChild.call(this,this.nodeOrShadowRoot(e),n)},e.prototype.insertBefore=function(e,n,r){return t.prototype.insertBefore.call(this,this.nodeOrShadowRoot(e),n,r)},e.prototype.removeChild=function(e,n){return t.prototype.removeChild.call(this,this.nodeOrShadowRoot(e),n)},e.prototype.parentNode=function(e){return this.nodeOrShadowRoot(t.prototype.parentNode.call(this,this.nodeOrShadowRoot(e)))},e}(G),$="undefined"!=typeof Zone&&Zone.__symbol__||function(t){return"__zone_symbol__"+t},J=$("addEventListener"),tt=$("removeEventListener"),et={},nt="removeEventListener",rt="__zone_symbol__propagationStopped",it="__zone_symbol__stopImmediatePropagation",ot="undefined"!=typeof Zone&&Zone[$("BLACK_LISTED_EVENTS")];ot&&(X={},ot.forEach(function(t){X[t]=t}));var at=function(t){return!!X&&X.hasOwnProperty(t)},st=function(t){var e=et[t.type];if(e){var n=this[e];if(n){var r=[t];if(1===n.length)return(a=n[0]).zone!==Zone.current?a.zone.run(a.handler,this,r):a.handler.apply(this,r);for(var i=n.slice(),o=0;o<i.length&&!0!==t[rt];o++){var a;(a=i[o]).zone!==Zone.current?a.zone.run(a.handler,this,r):a.handler.apply(this,r)}}}},ut=function(t){function e(e,n,r){var o=t.call(this,e)||this;return o.ngZone=n,r&&Object(i.isPlatformServer)(r)||o.patchEvent(),o}return Object(r.__extends)(e,t),e.prototype.patchEvent=function(){if("undefined"!=typeof Event&&Event&&Event.prototype&&!Event.prototype[it]){var t=Event.prototype[it]=Event.prototype.stopImmediatePropagation;Event.prototype.stopImmediatePropagation=function(){this&&(this[rt]=!0),t&&t.apply(this,arguments)}}},e.prototype.supports=function(t){return!0},e.prototype.addEventListener=function(t,e,n){var r=this,i=n;if(!t[J]||o.NgZone.isInAngularZone()&&!at(e))t.addEventListener(e,i,!1);else{var a=et[e];a||(a=et[e]=$("ANGULAR"+e+"FALSE"));var s=t[a],u=s&&s.length>0;s||(s=t[a]=[]);var c=at(e)?Zone.root:Zone.current;if(0===s.length)s.push({zone:c,handler:i});else{for(var l=!1,d=0;d<s.length;d++)if(s[d].handler===i){l=!0;break}l||s.push({zone:c,handler:i})}u||t[J](e,st,!1)}return function(){return r.removeEventListener(t,e,i)}},e.prototype.removeEventListener=function(t,e,n){var r=t[tt];if(!r)return t[nt].apply(t,[e,n,!1]);var i=et[e],o=i&&t[i];if(!o)return t[nt].apply(t,[e,n,!1]);for(var a=!1,s=0;s<o.length;s++)if(o[s].handler===n){a=!0,o.splice(s,1);break}a?0===o.length&&r.apply(t,[e,st,!1]):t[nt].apply(t,[e,n,!1])},e=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(0,Object(o.Inject)(_)),Object(r.__param)(2,Object(o.Optional)()),Object(r.__param)(2,Object(o.Inject)(o.PLATFORM_ID)),Object(r.__metadata)("design:paramtypes",[Object,o.NgZone,Object])],e)}(P),ct={pan:!0,panstart:!0,panmove:!0,panend:!0,pancancel:!0,panleft:!0,panright:!0,panup:!0,pandown:!0,pinch:!0,pinchstart:!0,pinchmove:!0,pinchend:!0,pinchcancel:!0,pinchin:!0,pinchout:!0,press:!0,pressup:!0,rotate:!0,rotatestart:!0,rotatemove:!0,rotateend:!0,rotatecancel:!0,swipe:!0,swipeleft:!0,swiperight:!0,swipeup:!0,swipedown:!0,tap:!0},lt=new o.InjectionToken("HammerGestureConfig"),dt=new o.InjectionToken("HammerLoader"),pt=function(){function t(){this.events=[],this.overrides={}}return t.prototype.buildHammer=function(t){var e=new Hammer(t,this.options);e.get("pinch").set({enable:!0}),e.get("rotate").set({enable:!0});for(var n in this.overrides)e.get(n).set(this.overrides[n]);return e},t=Object(r.__decorate)([Object(o.Injectable)()],t)}(),ft=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o._config=n,o.console=r,o.loader=i,o}return Object(r.__extends)(e,t),e.prototype.supports=function(t){return!(!ct.hasOwnProperty(t.toLowerCase())&&!this.isCustomEvent(t))&&(!(!window.Hammer&&!this.loader)||(this.console.warn('The "'+t+'" event cannot be bound because Hammer.JS is not loaded and no custom loader has been specified.'),!1))},e.prototype.addEventListener=function(t,e,n){var r=this,i=this.manager.getZone();if(e=e.toLowerCase(),!window.Hammer&&this.loader){var o=!1,a=function(){o=!0};return this.loader().then(function(){if(!window.Hammer)return r.console.warn("The custom HAMMER_LOADER completed, but Hammer.JS is not present."),void(a=function(){});o||(a=r.addEventListener(t,e,n))}).catch(function(){r.console.warn('The "'+e+'" event cannot be bound because the custom Hammer.JS loader failed.'),a=function(){}}),function(){a()}}return i.runOutsideAngular(function(){var o=r._config.buildHammer(t),a=function(t){i.runGuarded(function(){n(t)})};return o.on(e,a),function(){o.off(e,a),"function"==typeof o.destroy&&o.destroy()}})},e.prototype.isCustomEvent=function(t){return this._config.events.indexOf(t)>-1},e=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(0,Object(o.Inject)(_)),Object(r.__param)(1,Object(o.Inject)(lt)),Object(r.__param)(3,Object(o.Optional)()),Object(r.__param)(3,Object(o.Inject)(dt)),Object(r.__metadata)("design:paramtypes",[Object,pt,o["ɵConsole"],Object])],e)}(P),ht=["alt","control","meta","shift"],mt={alt:function(t){return t.altKey},control:function(t){return t.ctrlKey},meta:function(t){return t.metaKey},shift:function(t){return t.shiftKey}},gt=function(t){function e(e){return t.call(this,e)||this}var n;return Object(r.__extends)(e,t),n=e,e.prototype.supports=function(t){return null!=n.parseEventName(t)},e.prototype.addEventListener=function(t,e,r){var i=n.parseEventName(e),o=n.eventCallback(i.fullKey,r,this.manager.getZone());return this.manager.getZone().runOutsideAngular(function(){return s().onAndCancel(t,i.domEventName,o)})},e.parseEventName=function(t){var e=t.toLowerCase().split("."),r=e.shift();if(0===e.length||"keydown"!==r&&"keyup"!==r)return null;var i=n._normalizeKey(e.pop()),o="";if(ht.forEach(function(t){var n=e.indexOf(t);n>-1&&(e.splice(n,1),o+=t+".")}),o+=i,0!=e.length||0===i.length)return null;var a={};return a.domEventName=r,a.fullKey=o,a},e.getEventFullKey=function(t){var e="",n=s().getEventKey(t);return" "===(n=n.toLowerCase())?n="space":"."===n&&(n="dot"),ht.forEach(function(r){r!=n&&((0,mt[r])(t)&&(e+=r+"."))}),e+=n},e.eventCallback=function(t,e,r){return function(i){n.getEventFullKey(i)===t&&r.runGuarded(function(){return e(i)})}},e._normalizeKey=function(t){switch(t){case"esc":return"escape";default:return t}},e=n=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(0,Object(o.Inject)(_)),Object(r.__metadata)("design:paramtypes",[Object])],e)}(P),yt=function(){return function(){}}(),_t=function(t){function e(e){var n=t.call(this)||this;return n._doc=e,n}return Object(r.__extends)(e,t),e.prototype.sanitize=function(t,e){if(null==e)return null;switch(t){case o.SecurityContext.NONE:return e;case o.SecurityContext.HTML:return e instanceof bt?e.changingThisBreaksApplicationSecurity:(this.checkNotSafeValue(e,"HTML"),Object(o["ɵ_sanitizeHtml"])(this._doc,String(e)));case o.SecurityContext.STYLE:return e instanceof xt?e.changingThisBreaksApplicationSecurity:(this.checkNotSafeValue(e,"Style"),Object(o["ɵ_sanitizeStyle"])(e));case o.SecurityContext.SCRIPT:if(e instanceof wt)return e.changingThisBreaksApplicationSecurity;throw this.checkNotSafeValue(e,"Script"),new Error("unsafe value used in a script context");case o.SecurityContext.URL:return e instanceof Ot||e instanceof jt?e.changingThisBreaksApplicationSecurity:(this.checkNotSafeValue(e,"URL"),Object(o["ɵ_sanitizeUrl"])(String(e)));case o.SecurityContext.RESOURCE_URL:if(e instanceof Ot)return e.changingThisBreaksApplicationSecurity;throw this.checkNotSafeValue(e,"ResourceURL"),new Error("unsafe value used in a resource URL context (see http://g.co/ng/security#xss)");default:throw new Error("Unexpected SecurityContext "+t+" (see http://g.co/ng/security#xss)")}},e.prototype.checkNotSafeValue=function(t,e){if(t instanceof vt)throw new Error("Required a safe "+e+", got a "+t.getTypeName()+" (see http://g.co/ng/security#xss)")},e.prototype.bypassSecurityTrustHtml=function(t){return new bt(t)},e.prototype.bypassSecurityTrustStyle=function(t){return new xt(t)},e.prototype.bypassSecurityTrustScript=function(t){return new wt(t)},e.prototype.bypassSecurityTrustUrl=function(t){return new jt(t)},e.prototype.bypassSecurityTrustResourceUrl=function(t){return new Ot(t)},e=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(0,Object(o.Inject)(_)),Object(r.__metadata)("design:paramtypes",[Object])],e)}(yt),vt=function(){function t(t){this.changingThisBreaksApplicationSecurity=t}return t.prototype.toString=function(){return"SafeValue must use [property]=binding: "+this.changingThisBreaksApplicationSecurity+" (see http://g.co/ng/security#xss)"},t}(),bt=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.getTypeName=function(){return"HTML"},e}(vt),xt=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.getTypeName=function(){return"Style"},e}(vt),wt=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.getTypeName=function(){return"Script"},e}(vt),jt=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.getTypeName=function(){return"URL"},e}(vt),Ot=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.__extends)(e,t),e.prototype.getTypeName=function(){return"ResourceURL"},e}(vt),Ct=[{provide:o.PLATFORM_ID,useValue:i["ɵPLATFORM_BROWSER_ID"]},{provide:o.PLATFORM_INITIALIZER,useValue:Mt,multi:!0},{provide:i.PlatformLocation,useClass:b,deps:[_]},{provide:_,useFactory:kt,deps:[]}],Et=[{provide:o.Sanitizer,useExisting:yt},{provide:yt,useClass:_t,deps:[_]}],St=Object(o.createPlatformFactory)(o.platformCore,"browser",Ct);function Mt(){g.makeCurrent(),O.init()}function Tt(){return new o.ErrorHandler}function kt(){return document}var At=[Et,{provide:o["ɵAPP_ROOT"],useValue:!0},{provide:o.ErrorHandler,useFactory:Tt,deps:[]},{provide:D,useClass:ut,multi:!0,deps:[_,o.NgZone,o.PLATFORM_ID]},{provide:D,useClass:gt,multi:!0,deps:[_]},{provide:D,useClass:ft,multi:!0,deps:[_,lt,o["ɵConsole"],[new o.Optional,dt]]},{provide:lt,useClass:pt,deps:[]},{provide:q,useClass:q,deps:[I,N]},{provide:o.RendererFactory2,useExisting:q},{provide:R,useExisting:N},{provide:N,useClass:N,deps:[_]},{provide:o.Testability,useClass:o.Testability,deps:[o.NgZone]},{provide:I,useClass:I,deps:[D,o.NgZone]},A],Dt=function(){function t(t){if(t)throw new Error("BrowserModule has already been loaded. If you need access to common directives such as NgIf and NgFor from a lazy loaded module, import CommonModule instead.")}var e;return e=t,t.withServerTransition=function(t){return{ngModule:e,providers:[{provide:o.APP_ID,useValue:t.appId},{provide:x,useExisting:o.APP_ID},j]}},t=e=Object(r.__decorate)([Object(o.NgModule)({providers:At,exports:[i.CommonModule,o.ApplicationModule]}),Object(r.__param)(0,Object(o.Optional)()),Object(r.__param)(0,Object(o.SkipSelf)()),Object(r.__param)(0,Object(o.Inject)(e)),Object(r.__metadata)("design:paramtypes",[Object])],t)}();function It(){return new Pt(Object(o.inject)(_))}var Pt=function(){function t(t){this._doc=t,this._dom=s()}return t.prototype.addTag=function(t,e){return void 0===e&&(e=!1),t?this._getOrCreateElement(t,e):null},t.prototype.addTags=function(t,e){var n=this;return void 0===e&&(e=!1),t?t.reduce(function(t,r){return r&&t.push(n._getOrCreateElement(r,e)),t},[]):[]},t.prototype.getTag=function(t){return t?this._dom.querySelector(this._doc,"meta["+t+"]")||null:null},t.prototype.getTags=function(t){if(!t)return[];var e=this._dom.querySelectorAll(this._doc,"meta["+t+"]");return e?[].slice.call(e):[]},t.prototype.updateTag=function(t,e){if(!t)return null;e=e||this._parseSelector(t);var n=this.getTag(e);return n?this._setMetaElementAttributes(t,n):this._getOrCreateElement(t,!0)},t.prototype.removeTag=function(t){this.removeTagElement(this.getTag(t))},t.prototype.removeTagElement=function(t){t&&this._dom.remove(t)},t.prototype._getOrCreateElement=function(t,e){if(void 0===e&&(e=!1),!e){var n=this._parseSelector(t),r=this.getTag(n);if(r&&this._containsAttributes(t,r))return r}var i=this._dom.createElement("meta");this._setMetaElementAttributes(t,i);var o=this._dom.getElementsByTagName(this._doc,"head")[0];return this._dom.appendChild(o,i),i},t.prototype._setMetaElementAttributes=function(t,e){var n=this;return Object.keys(t).forEach(function(r){return n._dom.setAttribute(e,r,t[r])}),e},t.prototype._parseSelector=function(t){var e=t.name?"name":"property";return e+'="'+t[e]+'"'},t.prototype._containsAttributes=function(t,e){var n=this;return Object.keys(t).every(function(r){return n._dom.getAttribute(e,r)===t[r]})},t.ngInjectableDef=Object(o.defineInjectable)({factory:It,token:t,providedIn:"root"}),t=Object(r.__decorate)([Object(o.Injectable)({providedIn:"root",useFactory:It,deps:[]}),Object(r.__param)(0,Object(o.Inject)(_)),Object(r.__metadata)("design:paramtypes",[Object])],t)}();function Rt(){return new Nt(Object(o.inject)(_))}var Nt=function(){function t(t){this._doc=t}return t.prototype.getTitle=function(){return s().getTitle(this._doc)},t.prototype.setTitle=function(t){s().setTitle(this._doc,t)},t.ngInjectableDef=Object(o.defineInjectable)({factory:Rt,token:t,providedIn:"root"}),t=Object(r.__decorate)([Object(o.Injectable)({providedIn:"root",useFactory:Rt,deps:[]}),Object(r.__param)(0,Object(o.Inject)(_)),Object(r.__metadata)("design:paramtypes",[Object])],t)}(),Ft="undefined"!=typeof window&&window||{},Lt=function(){return function(t,e){this.msPerTick=t,this.numTicks=e}}(),Bt=function(){function t(t){this.appRef=t.injector.get(o.ApplicationRef)}return t.prototype.timeChangeDetection=function(t){var e=t&&t.record,n=null!=Ft.console.profile;e&&n&&Ft.console.profile("Change Detection");for(var r=s().performanceNow(),i=0;i<5||s().performanceNow()-r<500;)this.appRef.tick(),i++;var o=s().performanceNow();e&&n&&Ft.console.profileEnd("Change Detection");var a=(o-r)/i;return Ft.console.log("ran "+i+" change detection cycles"),Ft.console.log(a.toFixed(2)+" ms per check"),new Lt(a,i)},t}(),Vt="profiler";function zt(t){return C(Vt,new Bt(t)),t}function Ht(){C(Vt,null)}function Ut(t){var e={"&":"&a;",'"':"&q;","'":"&s;","<":"&l;",">":"&g;"};return t.replace(/[&"'<>]/g,function(t){return e[t]})}function Wt(t){return t}var qt=function(){function t(){this.store={},this.onSerializeCallbacks={}}var e;return e=t,t.init=function(t){var n=new e;return n.store=t,n},t.prototype.get=function(t,e){return void 0!==this.store[t]?this.store[t]:e},t.prototype.set=function(t,e){this.store[t]=e},t.prototype.remove=function(t){delete this.store[t]},t.prototype.hasKey=function(t){return this.store.hasOwnProperty(t)},t.prototype.onSerialize=function(t,e){this.onSerializeCallbacks[t]=e},t.prototype.toJson=function(){for(var t in this.onSerializeCallbacks)if(this.onSerializeCallbacks.hasOwnProperty(t))try{this.store[t]=this.onSerializeCallbacks[t]()}catch(t){console.warn("Exception in onSerialize callback: ",t)}return JSON.stringify(this.store)},t=e=Object(r.__decorate)([Object(o.Injectable)()],t)}();function Gt(t,e){var n,r,i=t.getElementById(e+"-state"),o={};if(i&&i.textContent)try{o=JSON.parse((n=i.textContent,r={"&a;":"&","&q;":'"',"&s;":"'","&l;":"<","&g;":">"},n.replace(/&[^;]+;/g,function(t){return r[t]})))}catch(t){console.warn("Exception while restoring TransferState for app "+e,t)}return qt.init(o)}var Yt=function(){function t(){}return t=Object(r.__decorate)([Object(o.NgModule)({providers:[{provide:qt,useFactory:Gt,deps:[_,o.APP_ID]}]})],t)}(),Kt=function(){function t(){}return t.all=function(){return function(t){return!0}},t.css=function(t){return function(e){return null!=e.nativeElement&&s().elementMatches(e.nativeElement,t)}},t.directive=function(t){return function(e){return-1!==e.providerTokens.indexOf(t)}},t}(),Xt=new o.Version("7.2.0")},"./node_modules/@angular/platform-browser/node_modules/tslib/tslib.es6.js":function(t,e,n){"use strict";n.r(e),n.d(e,"__extends",function(){return i}),n.d(e,"__assign",function(){return o}),n.d(e,"__rest",function(){return a}),n.d(e,"__decorate",function(){return s}),n.d(e,"__param",function(){return u}),n.d(e,"__metadata",function(){return c}),n.d(e,"__awaiter",function(){return l}),n.d(e,"__generator",function(){return d}),n.d(e,"__exportStar",function(){return p}),n.d(e,"__values",function(){return f}),n.d(e,"__read",function(){return h}),n.d(e,"__spread",function(){return m}),n.d(e,"__await",function(){return g}),n.d(e,"__asyncGenerator",function(){return y}),n.d(e,"__asyncDelegator",function(){return _}),n.d(e,"__asyncValues",function(){return v}),n.d(e,"__makeTemplateObject",function(){return b}),n.d(e,"__importStar",function(){return x}),n.d(e,"__importDefault",function(){return w});var r=function(t,e){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)};function i(t,e){function n(){this.constructor=t}r(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}var o=function(){return(o=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++){e=arguments[n];for(var i in e)Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i])}return t}).apply(this,arguments)};function a(t,e){var n={};for(var r in t)Object.prototype.hasOwnProperty.call(t,r)&&e.indexOf(r)<0&&(n[r]=t[r]);if(null!=t&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(t);i<r.length;i++)e.indexOf(r[i])<0&&(n[r[i]]=t[r[i]])}return n}function s(t,e,n,r){var i,o=arguments.length,a=o<3?e:null===r?r=Object.getOwnPropertyDescriptor(e,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(t,e,n,r);else for(var s=t.length-1;s>=0;s--)(i=t[s])&&(a=(o<3?i(a):o>3?i(e,n,a):i(e,n))||a);return o>3&&a&&Object.defineProperty(e,n,a),a}function u(t,e){return function(n,r){e(n,r,t)}}function c(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)}function l(t,e,n,r){return new(n||(n=Promise))(function(i,o){function a(t){try{u(r.next(t))}catch(t){o(t)}}function s(t){try{u(r.throw(t))}catch(t){o(t)}}function u(t){t.done?i(t.value):new n(function(e){e(t.value)}).then(a,s)}u((r=r.apply(t,e||[])).next())})}function d(t,e){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=e.call(t,a)}catch(t){o=[6,t],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(t,e){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}function f(t){var e="function"==typeof Symbol&&t[Symbol.iterator],n=0;return e?e.call(t):{next:function(){return t&&n>=t.length&&(t=void 0),{value:t&&t[n++],done:!t}}}}function h(t,e){var n="function"==typeof Symbol&&t[Symbol.iterator];if(!n)return t;var r,i,o=n.call(t),a=[];try{for(;(void 0===e||e-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(t){i={error:t}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var t=[],e=0;e<arguments.length;e++)t=t.concat(h(arguments[e]));return t}function g(t){return this instanceof g?(this.v=t,this):new g(t)}function y(t,e,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(t,e||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(t){i[t]&&(r[t]=function(e){return new Promise(function(n,r){o.push([t,e,n,r])>1||s(t,e)})})}function s(t,e){try{(n=i[t](e)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(t){l(o[0][3],t)}var n}function u(t){s("next",t)}function c(t){s("throw",t)}function l(t,e){t(e),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(t){var e,n;return e={},r("next"),r("throw",function(t){throw t}),r("return"),e[Symbol.iterator]=function(){return this},e;function r(r,i){e[r]=t[r]?function(e){return(n=!n)?{value:g(t[r](e)),done:"return"===r}:i?i(e):e}:i}}function v(t){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var e,n=t[Symbol.asyncIterator];return n?n.call(t):(t=f(t),e={},r("next"),r("throw"),r("return"),e[Symbol.asyncIterator]=function(){return this},e);function r(n){e[n]=t[n]&&function(e){return new Promise(function(r,i){var o,a,s,u;e=t[n](e),o=r,a=i,s=e.done,u=e.value,Promise.resolve(u).then(function(t){o({value:t,done:s})},a)})}}}function b(t,e){return Object.defineProperty?Object.defineProperty(t,"raw",{value:e}):t.raw=e,t}function x(t){if(t&&t.__esModule)return t;var e={};if(null!=t)for(var n in t)Object.hasOwnProperty.call(t,n)&&(e[n]=t[n]);return e.default=t,e}function w(t){return t&&t.__esModule?t:{default:t}}},"./node_modules/d3-array/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-array/src/bisect.js");n.d(e,"bisect",function(){return r.default}),n.d(e,"bisectRight",function(){return r.bisectRight}),n.d(e,"bisectLeft",function(){return r.bisectLeft});var i=n("./node_modules/d3-array/src/ascending.js");n.d(e,"ascending",function(){return i.default});var o=n("./node_modules/d3-array/src/bisector.js");n.d(e,"bisector",function(){return o.default});var a=n("./node_modules/d3-array/src/cross.js");n.d(e,"cross",function(){return a.default});var s=n("./node_modules/d3-array/src/descending.js");n.d(e,"descending",function(){return s.default});var u=n("./node_modules/d3-array/src/deviation.js");n.d(e,"deviation",function(){return u.default});var c=n("./node_modules/d3-array/src/extent.js");n.d(e,"extent",function(){return c.default});var l=n("./node_modules/d3-array/src/histogram.js");n.d(e,"histogram",function(){return l.default});var d=n("./node_modules/d3-array/src/threshold/freedmanDiaconis.js");n.d(e,"thresholdFreedmanDiaconis",function(){return d.default});var p=n("./node_modules/d3-array/src/threshold/scott.js");n.d(e,"thresholdScott",function(){return p.default});var f=n("./node_modules/d3-array/src/threshold/sturges.js");n.d(e,"thresholdSturges",function(){return f.default});var h=n("./node_modules/d3-array/src/max.js");n.d(e,"max",function(){return h.default});var m=n("./node_modules/d3-array/src/mean.js");n.d(e,"mean",function(){return m.default});var g=n("./node_modules/d3-array/src/median.js");n.d(e,"median",function(){return g.default});var y=n("./node_modules/d3-array/src/merge.js");n.d(e,"merge",function(){return y.default});var _=n("./node_modules/d3-array/src/min.js");n.d(e,"min",function(){return _.default});var v=n("./node_modules/d3-array/src/pairs.js");n.d(e,"pairs",function(){return v.default});var b=n("./node_modules/d3-array/src/permute.js");n.d(e,"permute",function(){return b.default});var x=n("./node_modules/d3-array/src/quantile.js");n.d(e,"quantile",function(){return x.default});var w=n("./node_modules/d3-array/src/range.js");n.d(e,"range",function(){return w.default});var j=n("./node_modules/d3-array/src/scan.js");n.d(e,"scan",function(){return j.default});var O=n("./node_modules/d3-array/src/shuffle.js");n.d(e,"shuffle",function(){return O.default});var C=n("./node_modules/d3-array/src/sum.js");n.d(e,"sum",function(){return C.default});var E=n("./node_modules/d3-array/src/ticks.js");n.d(e,"ticks",function(){return E.default}),n.d(e,"tickIncrement",function(){return E.tickIncrement}),n.d(e,"tickStep",function(){return E.tickStep});var S=n("./node_modules/d3-array/src/transpose.js");n.d(e,"transpose",function(){return S.default});var M=n("./node_modules/d3-array/src/variance.js");n.d(e,"variance",function(){return M.default});var T=n("./node_modules/d3-array/src/zip.js");n.d(e,"zip",function(){return T.default})},"./node_modules/d3-array/src/array.js":function(t,e,n){"use strict";n.r(e),n.d(e,"slice",function(){return i}),n.d(e,"map",function(){return o});var r=Array.prototype,i=r.slice,o=r.map},"./node_modules/d3-array/src/ascending.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){return t<e?-1:t>e?1:t>=e?0:NaN}},"./node_modules/d3-array/src/bisect.js":function(t,e,n){"use strict";n.r(e),n.d(e,"bisectRight",function(){return a}),n.d(e,"bisectLeft",function(){return s});var r=n("./node_modules/d3-array/src/ascending.js"),i=n("./node_modules/d3-array/src/bisector.js"),o=Object(i.default)(r.default),a=o.right,s=o.left;e.default=a},"./node_modules/d3-array/src/bisector.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-array/src/ascending.js");e.default=function(t){var e;return 1===t.length&&(e=t,t=function(t,n){return Object(r.default)(e(t),n)}),{left:function(e,n,r,i){for(null==r&&(r=0),null==i&&(i=e.length);r<i;){var o=r+i>>>1;t(e[o],n)<0?r=o+1:i=o}return r},right:function(e,n,r,i){for(null==r&&(r=0),null==i&&(i=e.length);r<i;){var o=r+i>>>1;t(e[o],n)>0?i=o:r=o+1}return r}}}},"./node_modules/d3-array/src/constant.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return function(){return t}}},"./node_modules/d3-array/src/cross.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-array/src/pairs.js");e.default=function(t,e,n){var i,o,a,s,u=t.length,c=e.length,l=new Array(u*c);for(null==n&&(n=r.pair),i=a=0;i<u;++i)for(s=t[i],o=0;o<c;++o,++a)l[a]=n(s,e[o]);return l}},"./node_modules/d3-array/src/descending.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){return e<t?-1:e>t?1:e>=t?0:NaN}},"./node_modules/d3-array/src/deviation.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-array/src/variance.js");e.default=function(t,e){var n=Object(r.default)(t,e);return n?Math.sqrt(n):n}},"./node_modules/d3-array/src/extent.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){var n,r,i,o=t.length,a=-1;if(null==e){for(;++a<o;)if(null!=(n=t[a])&&n>=n)for(r=i=n;++a<o;)null!=(n=t[a])&&(r>n&&(r=n),i<n&&(i=n))}else for(;++a<o;)if(null!=(n=e(t[a],a,t))&&n>=n)for(r=i=n;++a<o;)null!=(n=e(t[a],a,t))&&(r>n&&(r=n),i<n&&(i=n));return[r,i]}},"./node_modules/d3-array/src/histogram.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-array/src/array.js"),i=n("./node_modules/d3-array/src/bisect.js"),o=n("./node_modules/d3-array/src/constant.js"),a=n("./node_modules/d3-array/src/extent.js"),s=n("./node_modules/d3-array/src/identity.js"),u=n("./node_modules/d3-array/src/range.js"),c=n("./node_modules/d3-array/src/ticks.js"),l=n("./node_modules/d3-array/src/threshold/sturges.js");e.default=function(){var t=s.default,e=a.default,n=l.default;function d(r){var o,a,s=r.length,l=new Array(s);for(o=0;o<s;++o)l[o]=t(r[o],o,r);var d=e(l),p=d[0],f=d[1],h=n(l,p,f);Array.isArray(h)||(h=Object(c.tickStep)(p,f,h),h=Object(u.default)(Math.ceil(p/h)*h,Math.floor(f/h)*h,h));for(var m=h.length;h[0]<=p;)h.shift(),--m;for(;h[m-1]>f;)h.pop(),--m;var g,y=new Array(m+1);for(o=0;o<=m;++o)(g=y[o]=[]).x0=o>0?h[o-1]:p,g.x1=o<m?h[o]:f;for(o=0;o<s;++o)p<=(a=l[o])&&a<=f&&y[Object(i.default)(h,a,0,m)].push(r[o]);return y}return d.value=function(e){return arguments.length?(t="function"==typeof e?e:Object(o.default)(e),d):t},d.domain=function(t){return arguments.length?(e="function"==typeof t?t:Object(o.default)([t[0],t[1]]),d):e},d.thresholds=function(t){return arguments.length?(n="function"==typeof t?t:Array.isArray(t)?Object(o.default)(r.slice.call(t)):Object(o.default)(t),d):n},d}},"./node_modules/d3-array/src/identity.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return t}},"./node_modules/d3-array/src/max.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){var n,r,i=t.length,o=-1;if(null==e){for(;++o<i;)if(null!=(n=t[o])&&n>=n)for(r=n;++o<i;)null!=(n=t[o])&&n>r&&(r=n)}else for(;++o<i;)if(null!=(n=e(t[o],o,t))&&n>=n)for(r=n;++o<i;)null!=(n=e(t[o],o,t))&&n>r&&(r=n);return r}},"./node_modules/d3-array/src/mean.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-array/src/number.js");e.default=function(t,e){var n,i=t.length,o=i,a=-1,s=0;if(null==e)for(;++a<i;)isNaN(n=Object(r.default)(t[a]))?--o:s+=n;else for(;++a<i;)isNaN(n=Object(r.default)(e(t[a],a,t)))?--o:s+=n;if(o)return s/o}},"./node_modules/d3-array/src/median.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-array/src/ascending.js"),i=n("./node_modules/d3-array/src/number.js"),o=n("./node_modules/d3-array/src/quantile.js");e.default=function(t,e){var n,a=t.length,s=-1,u=[];if(null==e)for(;++s<a;)isNaN(n=Object(i.default)(t[s]))||u.push(n);else for(;++s<a;)isNaN(n=Object(i.default)(e(t[s],s,t)))||u.push(n);return Object(o.default)(u.sort(r.default),.5)}},"./node_modules/d3-array/src/merge.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){for(var e,n,r,i=t.length,o=-1,a=0;++o<i;)a+=t[o].length;for(n=new Array(a);--i>=0;)for(e=(r=t[i]).length;--e>=0;)n[--a]=r[e];return n}},"./node_modules/d3-array/src/min.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){var n,r,i=t.length,o=-1;if(null==e){for(;++o<i;)if(null!=(n=t[o])&&n>=n)for(r=n;++o<i;)null!=(n=t[o])&&r>n&&(r=n)}else for(;++o<i;)if(null!=(n=e(t[o],o,t))&&n>=n)for(r=n;++o<i;)null!=(n=e(t[o],o,t))&&r>n&&(r=n);return r}},"./node_modules/d3-array/src/number.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return null===t?NaN:+t}},"./node_modules/d3-array/src/pairs.js":function(t,e,n){"use strict";function r(t,e){return[t,e]}n.r(e),n.d(e,"pair",function(){return r}),e.default=function(t,e){null==e&&(e=r);for(var n=0,i=t.length-1,o=t[0],a=new Array(i<0?0:i);n<i;)a[n]=e(o,o=t[++n]);return a}},"./node_modules/d3-array/src/permute.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){for(var n=e.length,r=new Array(n);n--;)r[n]=t[e[n]];return r}},"./node_modules/d3-array/src/quantile.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-array/src/number.js");e.default=function(t,e,n){if(null==n&&(n=r.default),i=t.length){if((e=+e)<=0||i<2)return+n(t[0],0,t);if(e>=1)return+n(t[i-1],i-1,t);var i,o=(i-1)*e,a=Math.floor(o),s=+n(t[a],a,t);return s+(+n(t[a+1],a+1,t)-s)*(o-a)}}},"./node_modules/d3-array/src/range.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e,n){t=+t,e=+e,n=(i=arguments.length)<2?(e=t,t=0,1):i<3?1:+n;for(var r=-1,i=0|Math.max(0,Math.ceil((e-t)/n)),o=new Array(i);++r<i;)o[r]=t+r*n;return o}},"./node_modules/d3-array/src/scan.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-array/src/ascending.js");e.default=function(t,e){if(n=t.length){var n,i,o=0,a=0,s=t[a];for(null==e&&(e=r.default);++o<n;)(e(i=t[o],s)<0||0!==e(s,s))&&(s=i,a=o);return 0===e(s,s)?a:void 0}}},"./node_modules/d3-array/src/shuffle.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e,n){for(var r,i,o=(null==n?t.length:n)-(e=null==e?0:+e);o;)i=Math.random()*o--|0,r=t[o+e],t[o+e]=t[i+e],t[i+e]=r;return t}},"./node_modules/d3-array/src/sum.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){var n,r=t.length,i=-1,o=0;if(null==e)for(;++i<r;)(n=+t[i])&&(o+=n);else for(;++i<r;)(n=+e(t[i],i,t))&&(o+=n);return o}},"./node_modules/d3-array/src/threshold/freedmanDiaconis.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-array/src/array.js"),i=n("./node_modules/d3-array/src/ascending.js"),o=n("./node_modules/d3-array/src/number.js"),a=n("./node_modules/d3-array/src/quantile.js");e.default=function(t,e,n){return t=r.map.call(t,o.default).sort(i.default),Math.ceil((n-e)/(2*(Object(a.default)(t,.75)-Object(a.default)(t,.25))*Math.pow(t.length,-1/3)))}},"./node_modules/d3-array/src/threshold/scott.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-array/src/deviation.js");e.default=function(t,e,n){return Math.ceil((n-e)/(3.5*Object(r.default)(t)*Math.pow(t.length,-1/3)))}},"./node_modules/d3-array/src/threshold/sturges.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return Math.ceil(Math.log(t.length)/Math.LN2)+1}},"./node_modules/d3-array/src/ticks.js":function(t,e,n){"use strict";n.r(e),n.d(e,"tickIncrement",function(){return a}),n.d(e,"tickStep",function(){return s});var r=Math.sqrt(50),i=Math.sqrt(10),o=Math.sqrt(2);function a(t,e,n){var a=(e-t)/Math.max(0,n),s=Math.floor(Math.log(a)/Math.LN10),u=a/Math.pow(10,s);return s>=0?(u>=r?10:u>=i?5:u>=o?2:1)*Math.pow(10,s):-Math.pow(10,-s)/(u>=r?10:u>=i?5:u>=o?2:1)}function s(t,e,n){var a=Math.abs(e-t)/Math.max(0,n),s=Math.pow(10,Math.floor(Math.log(a)/Math.LN10)),u=a/s;return u>=r?s*=10:u>=i?s*=5:u>=o&&(s*=2),e<t?-s:s}e.default=function(t,e,n){var r,i,o,s,u=-1;if(n=+n,(t=+t)===(e=+e)&&n>0)return[t];if((r=e<t)&&(i=t,t=e,e=i),0===(s=a(t,e,n))||!isFinite(s))return[];if(s>0)for(t=Math.ceil(t/s),e=Math.floor(e/s),o=new Array(i=Math.ceil(e-t+1));++u<i;)o[u]=(t+u)*s;else for(t=Math.floor(t*s),e=Math.ceil(e*s),o=new Array(i=Math.ceil(t-e+1));++u<i;)o[u]=(t-u)/s;return r&&o.reverse(),o}},"./node_modules/d3-array/src/transpose.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-array/src/min.js");function i(t){return t.length}e.default=function(t){if(!(a=t.length))return[];for(var e=-1,n=Object(r.default)(t,i),o=new Array(n);++e<n;)for(var a,s=-1,u=o[e]=new Array(a);++s<a;)u[s]=t[s][e];return o}},"./node_modules/d3-array/src/variance.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-array/src/number.js");e.default=function(t,e){var n,i,o=t.length,a=0,s=-1,u=0,c=0;if(null==e)for(;++s<o;)isNaN(n=Object(r.default)(t[s]))||(c+=(i=n-u)*(n-(u+=i/++a)));else for(;++s<o;)isNaN(n=Object(r.default)(e(t[s],s,t)))||(c+=(i=n-u)*(n-(u+=i/++a)));if(a>1)return c/(a-1)}},"./node_modules/d3-array/src/zip.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-array/src/transpose.js");e.default=function(){return Object(r.default)(arguments)}},"./node_modules/d3-axis/src/array.js":function(t,e,n){"use strict";n.r(e),n.d(e,"slice",function(){return r});var r=Array.prototype.slice},"./node_modules/d3-axis/src/axis.js":function(t,e,n){"use strict";n.r(e),n.d(e,"axisTop",function(){return h}),n.d(e,"axisRight",function(){return m}),n.d(e,"axisBottom",function(){return g}),n.d(e,"axisLeft",function(){return y});var r=n("./node_modules/d3-axis/src/array.js"),i=n("./node_modules/d3-axis/src/identity.js"),o=1,a=2,s=3,u=4,c=1e-6;function l(t){return"translate("+(t+.5)+",0)"}function d(t){return"translate(0,"+(t+.5)+")"}function p(){return!this.__axis}function f(t,e){var n=[],f=null,h=null,m=6,g=6,y=3,_=t===o||t===u?-1:1,v=t===u||t===a?"x":"y",b=t===o||t===s?l:d;function x(r){var l=null==f?e.ticks?e.ticks.apply(e,n):e.domain():f,d=null==h?e.tickFormat?e.tickFormat.apply(e,n):i.default:h,x=Math.max(m,0)+y,w=e.range(),j=+w[0]+.5,O=+w[w.length-1]+.5,C=(e.bandwidth?function(t){var e=Math.max(0,t.bandwidth()-1)/2;return t.round()&&(e=Math.round(e)),function(n){return+t(n)+e}}:function(t){return function(e){return+t(e)}})(e.copy()),E=r.selection?r.selection():r,S=E.selectAll(".domain").data([null]),M=E.selectAll(".tick").data(l,e).order(),T=M.exit(),k=M.enter().append("g").attr("class","tick"),A=M.select("line"),D=M.select("text");S=S.merge(S.enter().insert("path",".tick").attr("class","domain").attr("stroke","currentColor")),M=M.merge(k),A=A.merge(k.append("line").attr("stroke","currentColor").attr(v+"2",_*m)),D=D.merge(k.append("text").attr("fill","currentColor").attr(v,_*x).attr("dy",t===o?"0em":t===s?"0.71em":"0.32em")),r!==E&&(S=S.transition(r),M=M.transition(r),A=A.transition(r),D=D.transition(r),T=T.transition(r).attr("opacity",c).attr("transform",function(t){return isFinite(t=C(t))?b(t):this.getAttribute("transform")}),k.attr("opacity",c).attr("transform",function(t){var e=this.parentNode.__axis;return b(e&&isFinite(e=e(t))?e:C(t))})),T.remove(),S.attr("d",t===u||t==a?g?"M"+_*g+","+j+"H0.5V"+O+"H"+_*g:"M0.5,"+j+"V"+O:g?"M"+j+","+_*g+"V0.5H"+O+"V"+_*g:"M"+j+",0.5H"+O),M.attr("opacity",1).attr("transform",function(t){return b(C(t))}),A.attr(v+"2",_*m),D.attr(v,_*x).text(d),E.filter(p).attr("fill","none").attr("font-size",10).attr("font-family","sans-serif").attr("text-anchor",t===a?"start":t===u?"end":"middle"),E.each(function(){this.__axis=C})}return x.scale=function(t){return arguments.length?(e=t,x):e},x.ticks=function(){return n=r.slice.call(arguments),x},x.tickArguments=function(t){return arguments.length?(n=null==t?[]:r.slice.call(t),x):n.slice()},x.tickValues=function(t){return arguments.length?(f=null==t?null:r.slice.call(t),x):f&&f.slice()},x.tickFormat=function(t){return arguments.length?(h=t,x):h},x.tickSize=function(t){return arguments.length?(m=g=+t,x):m},x.tickSizeInner=function(t){return arguments.length?(m=+t,x):m},x.tickSizeOuter=function(t){return arguments.length?(g=+t,x):g},x.tickPadding=function(t){return arguments.length?(y=+t,x):y},x}function h(t){return f(o,t)}function m(t){return f(a,t)}function g(t){return f(s,t)}function y(t){return f(u,t)}},"./node_modules/d3-axis/src/identity.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return t}},"./node_modules/d3-axis/src/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-axis/src/axis.js");n.d(e,"axisTop",function(){return r.axisTop}),n.d(e,"axisRight",function(){return r.axisRight}),n.d(e,"axisBottom",function(){return r.axisBottom}),n.d(e,"axisLeft",function(){return r.axisLeft})},"./node_modules/d3-brush/src/brush.js":function(t,e,n){"use strict";n.r(e),n.d(e,"brushSelection",function(){return M}),n.d(e,"brushX",function(){return T}),n.d(e,"brushY",function(){return k});var r=n("./node_modules/d3-dispatch/index.js"),i=n("./node_modules/d3-drag/src/index.js"),o=n("./node_modules/d3-interpolate/index.js"),a=n("./node_modules/d3-selection/src/index.js"),s=n("./node_modules/d3-transition/src/index.js"),u=n("./node_modules/d3-brush/src/constant.js"),c=n("./node_modules/d3-brush/src/event.js"),l=n("./node_modules/d3-brush/src/noevent.js"),d={name:"drag"},p={name:"space"},f={name:"handle"},h={name:"center"},m={name:"x",handles:["e","w"].map(j),input:function(t,e){return t&&[[t[0],e[0][1]],[t[1],e[1][1]]]},output:function(t){return t&&[t[0][0],t[1][0]]}},g={name:"y",handles:["n","s"].map(j),input:function(t,e){return t&&[[e[0][0],t[0]],[e[1][0],t[1]]]},output:function(t){return t&&[t[0][1],t[1][1]]}},y={name:"xy",handles:["n","e","s","w","nw","ne","se","sw"].map(j),input:function(t){return t},output:function(t){return t}},_={overlay:"crosshair",selection:"move",n:"ns-resize",e:"ew-resize",s:"ns-resize",w:"ew-resize",nw:"nwse-resize",ne:"nesw-resize",se:"nwse-resize",sw:"nesw-resize"},v={e:"w",w:"e",nw:"ne",ne:"nw",se:"sw",sw:"se"},b={n:"s",s:"n",nw:"sw",ne:"se",se:"ne",sw:"nw"},x={overlay:1,selection:1,n:null,e:1,s:null,w:-1,nw:-1,ne:1,se:1,sw:-1},w={overlay:1,selection:1,n:-1,e:null,s:1,w:null,nw:-1,ne:-1,se:1,sw:1};function j(t){return{type:t}}function O(){return!a.event.button}function C(){var t=this.ownerSVGElement||this;return[[0,0],[t.width.baseVal.value,t.height.baseVal.value]]}function E(t){for(;!t.__brush;)if(!(t=t.parentNode))return;return t.__brush}function S(t){return t[0][0]===t[1][0]||t[0][1]===t[1][1]}function M(t){var e=t.__brush;return e?e.dim.output(e.selection):null}function T(){return A(m)}function k(){return A(g)}function A(t){var e,n=C,y=O,M=Object(r.dispatch)(k,"start","brush","end"),T=6;function k(e){var n=e.property("__brush",R).selectAll(".overlay").data([j("overlay")]);n.enter().append("rect").attr("class","overlay").attr("pointer-events","all").attr("cursor",_.overlay).merge(n).each(function(){var t=E(this).extent;Object(a.select)(this).attr("x",t[0][0]).attr("y",t[0][1]).attr("width",t[1][0]-t[0][0]).attr("height",t[1][1]-t[0][1])}),e.selectAll(".selection").data([j("selection")]).enter().append("rect").attr("class","selection").attr("cursor",_.selection).attr("fill","#777").attr("fill-opacity",.3).attr("stroke","#fff").attr("shape-rendering","crispEdges");var r=e.selectAll(".handle").data(t.handles,function(t){return t.type});r.exit().remove(),r.enter().append("rect").attr("class",function(t){return"handle handle--"+t.type}).attr("cursor",function(t){return _[t.type]}),e.each(A).attr("fill","none").attr("pointer-events","all").style("-webkit-tap-highlight-color","rgba(0,0,0,0)").on("mousedown.brush touchstart.brush",P)}function A(){var t=Object(a.select)(this),e=E(this).selection;e?(t.selectAll(".selection").style("display",null).attr("x",e[0][0]).attr("y",e[0][1]).attr("width",e[1][0]-e[0][0]).attr("height",e[1][1]-e[0][1]),t.selectAll(".handle").style("display",null).attr("x",function(t){return"e"===t.type[t.type.length-1]?e[1][0]-T/2:e[0][0]-T/2}).attr("y",function(t){return"s"===t.type[0]?e[1][1]-T/2:e[0][1]-T/2}).attr("width",function(t){return"n"===t.type||"s"===t.type?e[1][0]-e[0][0]+T:T}).attr("height",function(t){return"e"===t.type||"w"===t.type?e[1][1]-e[0][1]+T:T})):t.selectAll(".selection,.handle").style("display","none").attr("x",null).attr("y",null).attr("width",null).attr("height",null)}function D(t,e){return t.__brush.emitter||new I(t,e)}function I(t,e){this.that=t,this.args=e,this.state=t.__brush,this.active=0}function P(){if(a.event.touches){if(a.event.changedTouches.length<a.event.touches.length)return Object(l.default)()}else if(e)return;if(y.apply(this,arguments)){var n,r,o,u,c,j,O,C,M,T,k,I,P,R=this,N=a.event.target.__data__.type,F="selection"===(a.event.metaKey?N="overlay":N)?d:a.event.altKey?h:f,L=t===g?null:x[N],B=t===m?null:w[N],V=E(R),z=V.extent,H=V.selection,U=z[0][0],W=z[0][1],q=z[1][0],G=z[1][1],Y=L&&B&&a.event.shiftKey,K=Object(a.mouse)(R),X=K,Q=D(R,arguments).beforestart();"overlay"===N?V.selection=H=[[n=t===g?U:K[0],o=t===m?W:K[1]],[c=t===g?q:n,O=t===m?G:o]]:(n=H[0][0],o=H[0][1],c=H[1][0],O=H[1][1]),r=n,u=o,j=c,C=O;var Z=Object(a.select)(R).attr("pointer-events","none"),$=Z.selectAll(".overlay").attr("cursor",_[N]);if(a.event.touches)Z.on("touchmove.brush",tt,!0).on("touchend.brush touchcancel.brush",nt,!0);else{var J=Object(a.select)(a.event.view).on("keydown.brush",function(){switch(a.event.keyCode){case 16:Y=L&&B;break;case 18:F===f&&(L&&(c=j-M*L,n=r+M*L),B&&(O=C-T*B,o=u+T*B),F=h,et());break;case 32:F!==f&&F!==h||(L<0?c=j-M:L>0&&(n=r-M),B<0?O=C-T:B>0&&(o=u-T),F=p,$.attr("cursor",_.selection),et());break;default:return}Object(l.default)()},!0).on("keyup.brush",function(){switch(a.event.keyCode){case 16:Y&&(I=P=Y=!1,et());break;case 18:F===h&&(L<0?c=j:L>0&&(n=r),B<0?O=C:B>0&&(o=u),F=f,et());break;case 32:F===p&&(a.event.altKey?(L&&(c=j-M*L,n=r+M*L),B&&(O=C-T*B,o=u+T*B),F=h):(L<0?c=j:L>0&&(n=r),B<0?O=C:B>0&&(o=u),F=f),$.attr("cursor",_[N]),et());break;default:return}Object(l.default)()},!0).on("mousemove.brush",tt,!0).on("mouseup.brush",nt,!0);Object(i.dragDisable)(a.event.view)}Object(l.nopropagation)(),Object(s.interrupt)(R),A.call(R),Q.start()}function tt(){var t=Object(a.mouse)(R);!Y||I||P||(Math.abs(t[0]-X[0])>Math.abs(t[1]-X[1])?P=!0:I=!0),X=t,k=!0,Object(l.default)(),et()}function et(){var t;switch(M=X[0]-K[0],T=X[1]-K[1],F){case p:case d:L&&(M=Math.max(U-n,Math.min(q-c,M)),r=n+M,j=c+M),B&&(T=Math.max(W-o,Math.min(G-O,T)),u=o+T,C=O+T);break;case f:L<0?(M=Math.max(U-n,Math.min(q-n,M)),r=n+M,j=c):L>0&&(M=Math.max(U-c,Math.min(q-c,M)),r=n,j=c+M),B<0?(T=Math.max(W-o,Math.min(G-o,T)),u=o+T,C=O):B>0&&(T=Math.max(W-O,Math.min(G-O,T)),u=o,C=O+T);break;case h:L&&(r=Math.max(U,Math.min(q,n-M*L)),j=Math.max(U,Math.min(q,c+M*L))),B&&(u=Math.max(W,Math.min(G,o-T*B)),C=Math.max(W,Math.min(G,O+T*B)))}j<r&&(L*=-1,t=n,n=c,c=t,t=r,r=j,j=t,N in v&&$.attr("cursor",_[N=v[N]])),C<u&&(B*=-1,t=o,o=O,O=t,t=u,u=C,C=t,N in b&&$.attr("cursor",_[N=b[N]])),V.selection&&(H=V.selection),I&&(r=H[0][0],j=H[1][0]),P&&(u=H[0][1],C=H[1][1]),H[0][0]===r&&H[0][1]===u&&H[1][0]===j&&H[1][1]===C||(V.selection=[[r,u],[j,C]],A.call(R),Q.brush())}function nt(){if(Object(l.nopropagation)(),a.event.touches){if(a.event.touches.length)return;e&&clearTimeout(e),e=setTimeout(function(){e=null},500),Z.on("touchmove.brush touchend.brush touchcancel.brush",null)}else Object(i.dragEnable)(a.event.view,k),J.on("keydown.brush keyup.brush mousemove.brush mouseup.brush",null);Z.attr("pointer-events","all"),$.attr("cursor",_.overlay),V.selection&&(H=V.selection),S(H)&&(V.selection=null,A.call(R)),Q.end()}}function R(){var e=this.__brush||{selection:null};return e.extent=n.apply(this,arguments),e.dim=t,e}return k.move=function(e,n){e.selection?e.on("start.brush",function(){D(this,arguments).beforestart().start()}).on("interrupt.brush end.brush",function(){D(this,arguments).end()}).tween("brush",function(){var e=this,r=e.__brush,i=D(e,arguments),a=r.selection,s=t.input("function"==typeof n?n.apply(this,arguments):n,r.extent),u=Object(o.interpolate)(a,s);function c(t){r.selection=1===t&&S(s)?null:u(t),A.call(e),i.brush()}return a&&s?c:c(1)}):e.each(function(){var e=arguments,r=this.__brush,i=t.input("function"==typeof n?n.apply(this,e):n,r.extent),o=D(this,e).beforestart();Object(s.interrupt)(this),r.selection=null==i||S(i)?null:i,A.call(this),o.start().brush().end()})},I.prototype={beforestart:function(){return 1==++this.active&&(this.state.emitter=this,this.starting=!0),this},start:function(){return this.starting&&(this.starting=!1,this.emit("start")),this},brush:function(){return this.emit("brush"),this},end:function(){return 0==--this.active&&(delete this.state.emitter,this.emit("end")),this},emit:function(e){Object(a.customEvent)(new c.default(k,e,t.output(this.state.selection)),M.apply,M,[e,this.that,this.args])}},k.extent=function(t){return arguments.length?(n="function"==typeof t?t:Object(u.default)([[+t[0][0],+t[0][1]],[+t[1][0],+t[1][1]]]),k):n},k.filter=function(t){return arguments.length?(y="function"==typeof t?t:Object(u.default)(!!t),k):y},k.handleSize=function(t){return arguments.length?(T=+t,k):T},k.on=function(){var t=M.on.apply(M,arguments);return t===M?k:t},k}e.default=function(){return A(y)}},"./node_modules/d3-brush/src/constant.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return function(){return t}}},"./node_modules/d3-brush/src/event.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e,n){this.target=t,this.type=e,this.selection=n}},"./node_modules/d3-brush/src/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-brush/src/brush.js");n.d(e,"brush",function(){return r.default}),n.d(e,"brushX",function(){return r.brushX}),n.d(e,"brushY",function(){return r.brushY}),n.d(e,"brushSelection",function(){return r.brushSelection})},"./node_modules/d3-brush/src/noevent.js":function(t,e,n){"use strict";n.r(e),n.d(e,"nopropagation",function(){return i});var r=n("./node_modules/d3-selection/src/index.js");function i(){r.event.stopImmediatePropagation()}e.default=function(){r.event.preventDefault(),r.event.stopImmediatePropagation()}},"./node_modules/d3-collection/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-collection/src/nest.js");n.d(e,"nest",function(){return r.default});var i=n("./node_modules/d3-collection/src/set.js");n.d(e,"set",function(){return i.default});var o=n("./node_modules/d3-collection/src/map.js");n.d(e,"map",function(){return o.default});var a=n("./node_modules/d3-collection/src/keys.js");n.d(e,"keys",function(){return a.default});var s=n("./node_modules/d3-collection/src/values.js");n.d(e,"values",function(){return s.default});var u=n("./node_modules/d3-collection/src/entries.js");n.d(e,"entries",function(){return u.default})},"./node_modules/d3-collection/src/entries.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){var e=[];for(var n in t)e.push({key:n,value:t[n]});return e}},"./node_modules/d3-collection/src/keys.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){var e=[];for(var n in t)e.push(n);return e}},"./node_modules/d3-collection/src/map.js":function(t,e,n){"use strict";n.r(e),n.d(e,"prefix",function(){return r});var r="$";function i(){}function o(t,e){var n=new i;if(t instanceof i)t.each(function(t,e){n.set(e,t)});else if(Array.isArray(t)){var r,o=-1,a=t.length;if(null==e)for(;++o<a;)n.set(o,t[o]);else for(;++o<a;)n.set(e(r=t[o],o,t),r)}else if(t)for(var s in t)n.set(s,t[s]);return n}i.prototype=o.prototype={constructor:i,has:function(t){return r+t in this},get:function(t){return this[r+t]},set:function(t,e){return this[r+t]=e,this},remove:function(t){var e=r+t;return e in this&&delete this[e]},clear:function(){for(var t in this)t[0]===r&&delete this[t]},keys:function(){var t=[];for(var e in this)e[0]===r&&t.push(e.slice(1));return t},values:function(){var t=[];for(var e in this)e[0]===r&&t.push(this[e]);return t},entries:function(){var t=[];for(var e in this)e[0]===r&&t.push({key:e.slice(1),value:this[e]});return t},size:function(){var t=0;for(var e in this)e[0]===r&&++t;return t},empty:function(){for(var t in this)if(t[0]===r)return!1;return!0},each:function(t){for(var e in this)e[0]===r&&t(this[e],e.slice(1),this)}},e.default=o},"./node_modules/d3-collection/src/nest.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-collection/src/map.js");function i(){return{}}function o(t,e,n){t[e]=n}function a(){return Object(r.default)()}function s(t,e,n){t.set(e,n)}e.default=function(){var t,e,n,u=[],c=[];function l(n,i,o,a){if(i>=u.length)return null!=t&&n.sort(t),null!=e?e(n):n;for(var s,c,d,p=-1,f=n.length,h=u[i++],m=Object(r.default)(),g=o();++p<f;)(d=m.get(s=h(c=n[p])+""))?d.push(c):m.set(s,[c]);return m.each(function(t,e){a(g,e,l(t,i,o,a))}),g}return n={object:function(t){return l(t,0,i,o)},map:function(t){return l(t,0,a,s)},entries:function(t){return function t(n,r){if(++r>u.length)return n;var i,o=c[r-1];return null!=e&&r>=u.length?i=n.entries():(i=[],n.each(function(e,n){i.push({key:n,values:t(e,r)})})),null!=o?i.sort(function(t,e){return o(t.key,e.key)}):i}(l(t,0,a,s),0)},key:function(t){return u.push(t),n},sortKeys:function(t){return c[u.length-1]=t,n},sortValues:function(e){return t=e,n},rollup:function(t){return e=t,n}}}},"./node_modules/d3-collection/src/set.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-collection/src/map.js");function i(){}var o=r.default.prototype;function a(t,e){var n=new i;if(t instanceof i)t.each(function(t){n.add(t)});else if(t){var r=-1,o=t.length;if(null==e)for(;++r<o;)n.add(t[r]);else for(;++r<o;)n.add(e(t[r],r,t))}return n}i.prototype=a.prototype={constructor:i,has:o.has,add:function(t){return t+="",this[r.prefix+t]=t,this},remove:o.remove,clear:o.clear,values:o.keys,size:o.size,empty:o.empty,each:o.each},e.default=a},"./node_modules/d3-collection/src/values.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){var e=[];for(var n in t)e.push(t[n]);return e}},"./node_modules/d3-color/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-color/src/color.js");n.d(e,"color",function(){return r.default}),n.d(e,"rgb",function(){return r.rgb}),n.d(e,"hsl",function(){return r.hsl});var i=n("./node_modules/d3-color/src/lab.js");n.d(e,"lab",function(){return i.default}),n.d(e,"hcl",function(){return i.hcl});var o=n("./node_modules/d3-color/src/cubehelix.js");n.d(e,"cubehelix",function(){return o.default})},"./node_modules/d3-color/src/color.js":function(t,e,n){"use strict";n.r(e),n.d(e,"Color",function(){return i}),n.d(e,"darker",function(){return o}),n.d(e,"brighter",function(){return a}),n.d(e,"default",function(){return v}),n.d(e,"rgbConvert",function(){return w}),n.d(e,"rgb",function(){return j}),n.d(e,"Rgb",function(){return O}),n.d(e,"hslConvert",function(){return E}),n.d(e,"hsl",function(){return S});var r=n("./node_modules/d3-color/src/define.js");function i(){}var o=.7,a=1/o,s="\\s*([+-]?\\d+)\\s*",u="\\s*([+-]?\\d*\\.?\\d+(?:[eE][+-]?\\d+)?)\\s*",c="\\s*([+-]?\\d*\\.?\\d+(?:[eE][+-]?\\d+)?)%\\s*",l=/^#([0-9a-f]{3})$/,d=/^#([0-9a-f]{6})$/,p=new RegExp("^rgb\\("+[s,s,s]+"\\)$"),f=new RegExp("^rgb\\("+[c,c,c]+"\\)$"),h=new RegExp("^rgba\\("+[s,s,s,u]+"\\)$"),m=new RegExp("^rgba\\("+[c,c,c,u]+"\\)$"),g=new RegExp("^hsl\\("+[u,c,c]+"\\)$"),y=new RegExp("^hsla\\("+[u,c,c,u]+"\\)$"),_={aliceblue:15792383,antiquewhite:16444375,aqua:65535,aquamarine:8388564,azure:15794175,beige:16119260,bisque:16770244,black:0,blanchedalmond:16772045,blue:255,blueviolet:9055202,brown:10824234,burlywood:14596231,cadetblue:6266528,chartreuse:8388352,chocolate:13789470,coral:16744272,cornflowerblue:6591981,cornsilk:16775388,crimson:14423100,cyan:65535,darkblue:139,darkcyan:35723,darkgoldenrod:12092939,darkgray:11119017,darkgreen:25600,darkgrey:11119017,darkkhaki:12433259,darkmagenta:9109643,darkolivegreen:5597999,darkorange:16747520,darkorchid:10040012,darkred:9109504,darksalmon:15308410,darkseagreen:9419919,darkslateblue:4734347,darkslategray:3100495,darkslategrey:3100495,darkturquoise:52945,darkviolet:9699539,deeppink:16716947,deepskyblue:49151,dimgray:6908265,dimgrey:6908265,dodgerblue:2003199,firebrick:11674146,floralwhite:16775920,forestgreen:2263842,fuchsia:16711935,gainsboro:14474460,ghostwhite:16316671,gold:16766720,goldenrod:14329120,gray:8421504,green:32768,greenyellow:11403055,grey:8421504,honeydew:15794160,hotpink:16738740,indianred:13458524,indigo:4915330,ivory:16777200,khaki:15787660,lavender:15132410,lavenderblush:16773365,lawngreen:8190976,lemonchiffon:16775885,lightblue:11393254,lightcoral:15761536,lightcyan:14745599,lightgoldenrodyellow:16448210,lightgray:13882323,lightgreen:9498256,lightgrey:13882323,lightpink:16758465,lightsalmon:16752762,lightseagreen:2142890,lightskyblue:8900346,lightslategray:7833753,lightslategrey:7833753,lightsteelblue:11584734,lightyellow:16777184,lime:65280,limegreen:3329330,linen:16445670,magenta:16711935,maroon:8388608,mediumaquamarine:6737322,mediumblue:205,mediumorchid:12211667,mediumpurple:9662683,mediumseagreen:3978097,mediumslateblue:8087790,mediumspringgreen:64154,mediumturquoise:4772300,mediumvioletred:13047173,midnightblue:1644912,mintcream:16121850,mistyrose:16770273,moccasin:16770229,navajowhite:16768685,navy:128,oldlace:16643558,olive:8421376,olivedrab:7048739,orange:16753920,orangered:16729344,orchid:14315734,palegoldenrod:15657130,palegreen:10025880,paleturquoise:11529966,palevioletred:14381203,papayawhip:16773077,peachpuff:16767673,peru:13468991,pink:16761035,plum:14524637,powderblue:11591910,purple:8388736,rebeccapurple:6697881,red:16711680,rosybrown:12357519,royalblue:4286945,saddlebrown:9127187,salmon:16416882,sandybrown:16032864,seagreen:3050327,seashell:16774638,sienna:10506797,silver:12632256,skyblue:8900331,slateblue:6970061,slategray:7372944,slategrey:7372944,snow:16775930,springgreen:65407,steelblue:4620980,tan:13808780,teal:32896,thistle:14204888,tomato:16737095,turquoise:4251856,violet:15631086,wheat:16113331,white:16777215,whitesmoke:16119285,yellow:16776960,yellowgreen:10145074};function v(t){var e;return t=(t+"").trim().toLowerCase(),(e=l.exec(t))?new O((e=parseInt(e[1],16))>>8&15|e>>4&240,e>>4&15|240&e,(15&e)<<4|15&e,1):(e=d.exec(t))?b(parseInt(e[1],16)):(e=p.exec(t))?new O(e[1],e[2],e[3],1):(e=f.exec(t))?new O(255*e[1]/100,255*e[2]/100,255*e[3]/100,1):(e=h.exec(t))?x(e[1],e[2],e[3],e[4]):(e=m.exec(t))?x(255*e[1]/100,255*e[2]/100,255*e[3]/100,e[4]):(e=g.exec(t))?C(e[1],e[2]/100,e[3]/100,1):(e=y.exec(t))?C(e[1],e[2]/100,e[3]/100,e[4]):_.hasOwnProperty(t)?b(_[t]):"transparent"===t?new O(NaN,NaN,NaN,0):null}function b(t){return new O(t>>16&255,t>>8&255,255&t,1)}function x(t,e,n,r){return r<=0&&(t=e=n=NaN),new O(t,e,n,r)}function w(t){return t instanceof i||(t=v(t)),t?new O((t=t.rgb()).r,t.g,t.b,t.opacity):new O}function j(t,e,n,r){return 1===arguments.length?w(t):new O(t,e,n,null==r?1:r)}function O(t,e,n,r){this.r=+t,this.g=+e,this.b=+n,this.opacity=+r}function C(t,e,n,r){return r<=0?t=e=n=NaN:n<=0||n>=1?t=e=NaN:e<=0&&(t=NaN),new M(t,e,n,r)}function E(t){if(t instanceof M)return new M(t.h,t.s,t.l,t.opacity);if(t instanceof i||(t=v(t)),!t)return new M;if(t instanceof M)return t;var e=(t=t.rgb()).r/255,n=t.g/255,r=t.b/255,o=Math.min(e,n,r),a=Math.max(e,n,r),s=NaN,u=a-o,c=(a+o)/2;return u?(s=e===a?(n-r)/u+6*(n<r):n===a?(r-e)/u+2:(e-n)/u+4,u/=c<.5?a+o:2-a-o,s*=60):u=c>0&&c<1?0:s,new M(s,u,c,t.opacity)}function S(t,e,n,r){return 1===arguments.length?E(t):new M(t,e,n,null==r?1:r)}function M(t,e,n,r){this.h=+t,this.s=+e,this.l=+n,this.opacity=+r}function T(t,e,n){return 255*(t<60?e+(n-e)*t/60:t<180?n:t<240?e+(n-e)*(240-t)/60:e)}Object(r.default)(i,v,{displayable:function(){return this.rgb().displayable()},toString:function(){return this.rgb()+""}}),Object(r.default)(O,j,Object(r.extend)(i,{brighter:function(t){return t=null==t?a:Math.pow(a,t),new O(this.r*t,this.g*t,this.b*t,this.opacity)},darker:function(t){return t=null==t?o:Math.pow(o,t),new O(this.r*t,this.g*t,this.b*t,this.opacity)},rgb:function(){return this},displayable:function(){return 0<=this.r&&this.r<=255&&0<=this.g&&this.g<=255&&0<=this.b&&this.b<=255&&0<=this.opacity&&this.opacity<=1},toString:function(){var t=this.opacity;return(1===(t=isNaN(t)?1:Math.max(0,Math.min(1,t)))?"rgb(":"rgba(")+Math.max(0,Math.min(255,Math.round(this.r)||0))+", "+Math.max(0,Math.min(255,Math.round(this.g)||0))+", "+Math.max(0,Math.min(255,Math.round(this.b)||0))+(1===t?")":", "+t+")")}})),Object(r.default)(M,S,Object(r.extend)(i,{brighter:function(t){return t=null==t?a:Math.pow(a,t),new M(this.h,this.s,this.l*t,this.opacity)},darker:function(t){return t=null==t?o:Math.pow(o,t),new M(this.h,this.s,this.l*t,this.opacity)},rgb:function(){var t=this.h%360+360*(this.h<0),e=isNaN(t)||isNaN(this.s)?0:this.s,n=this.l,r=n+(n<.5?n:1-n)*e,i=2*n-r;return new O(T(t>=240?t-240:t+120,i,r),T(t,i,r),T(t<120?t+240:t-120,i,r),this.opacity)},displayable:function(){return(0<=this.s&&this.s<=1||isNaN(this.s))&&0<=this.l&&this.l<=1&&0<=this.opacity&&this.opacity<=1}}))},"./node_modules/d3-color/src/cubehelix.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return p}),n.d(e,"Cubehelix",function(){return f});var r=n("./node_modules/d3-color/src/define.js"),i=n("./node_modules/d3-color/src/color.js"),o=n("./node_modules/d3-color/src/math.js"),a=-.29227,s=-.90649,u=1.97294,c=u*s,l=1.78277*u,d=1.78277*a- -.14861*s;function p(t,e,n,r){return 1===arguments.length?function(t){if(t instanceof f)return new f(t.h,t.s,t.l,t.opacity);t instanceof i.Rgb||(t=Object(i.rgbConvert)(t));var e=t.r/255,n=t.g/255,r=t.b/255,p=(d*r+c*e-l*n)/(d+c-l),h=r-p,m=(u*(n-p)-a*h)/s,g=Math.sqrt(m*m+h*h)/(u*p*(1-p)),y=g?Math.atan2(m,h)*o.rad2deg-120:NaN;return new f(y<0?y+360:y,g,p,t.opacity)}(t):new f(t,e,n,null==r?1:r)}function f(t,e,n,r){this.h=+t,this.s=+e,this.l=+n,this.opacity=+r}Object(r.default)(f,p,Object(r.extend)(i.Color,{brighter:function(t){return t=null==t?i.brighter:Math.pow(i.brighter,t),new f(this.h,this.s,this.l*t,this.opacity)},darker:function(t){return t=null==t?i.darker:Math.pow(i.darker,t),new f(this.h,this.s,this.l*t,this.opacity)},rgb:function(){var t=isNaN(this.h)?0:(this.h+120)*o.deg2rad,e=+this.l,n=isNaN(this.s)?0:this.s*e*(1-e),r=Math.cos(t),c=Math.sin(t);return new i.Rgb(255*(e+n*(-.14861*r+1.78277*c)),255*(e+n*(a*r+s*c)),255*(e+n*(u*r)),this.opacity)}}))},"./node_modules/d3-color/src/define.js":function(t,e,n){"use strict";function r(t,e){var n=Object.create(t.prototype);for(var r in e)n[r]=e[r];return n}n.r(e),n.d(e,"extend",function(){return r}),e.default=function(t,e,n){t.prototype=e.prototype=n,n.constructor=t}},"./node_modules/d3-color/src/lab.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return h}),n.d(e,"Lab",function(){return m}),n.d(e,"hcl",function(){return b}),n.d(e,"Hcl",function(){return x});var r=n("./node_modules/d3-color/src/define.js"),i=n("./node_modules/d3-color/src/color.js"),o=n("./node_modules/d3-color/src/math.js"),a=.95047,s=1,u=1.08883,c=4/29,l=6/29,d=3*l*l,p=l*l*l;function f(t){if(t instanceof m)return new m(t.l,t.a,t.b,t.opacity);if(t instanceof x){var e=t.h*o.deg2rad;return new m(t.l,Math.cos(e)*t.c,Math.sin(e)*t.c,t.opacity)}t instanceof i.Rgb||(t=Object(i.rgbConvert)(t));var n=v(t.r),r=v(t.g),c=v(t.b),l=g((.4124564*n+.3575761*r+.1804375*c)/a),d=g((.2126729*n+.7151522*r+.072175*c)/s);return new m(116*d-16,500*(l-d),200*(d-g((.0193339*n+.119192*r+.9503041*c)/u)),t.opacity)}function h(t,e,n,r){return 1===arguments.length?f(t):new m(t,e,n,null==r?1:r)}function m(t,e,n,r){this.l=+t,this.a=+e,this.b=+n,this.opacity=+r}function g(t){return t>p?Math.pow(t,1/3):t/d+c}function y(t){return t>l?t*t*t:d*(t-c)}function _(t){return 255*(t<=.0031308?12.92*t:1.055*Math.pow(t,1/2.4)-.055)}function v(t){return(t/=255)<=.04045?t/12.92:Math.pow((t+.055)/1.055,2.4)}function b(t,e,n,r){return 1===arguments.length?function(t){if(t instanceof x)return new x(t.h,t.c,t.l,t.opacity);t instanceof m||(t=f(t));var e=Math.atan2(t.b,t.a)*o.rad2deg;return new x(e<0?e+360:e,Math.sqrt(t.a*t.a+t.b*t.b),t.l,t.opacity)}(t):new x(t,e,n,null==r?1:r)}function x(t,e,n,r){this.h=+t,this.c=+e,this.l=+n,this.opacity=+r}Object(r.default)(m,h,Object(r.extend)(i.Color,{brighter:function(t){return new m(this.l+18*(null==t?1:t),this.a,this.b,this.opacity)},darker:function(t){return new m(this.l-18*(null==t?1:t),this.a,this.b,this.opacity)},rgb:function(){var t=(this.l+16)/116,e=isNaN(this.a)?t:t+this.a/500,n=isNaN(this.b)?t:t-this.b/200;return t=s*y(t),e=a*y(e),n=u*y(n),new i.Rgb(_(3.2404542*e-1.5371385*t-.4985314*n),_(-.969266*e+1.8760108*t+.041556*n),_(.0556434*e-.2040259*t+1.0572252*n),this.opacity)}})),Object(r.default)(x,b,Object(r.extend)(i.Color,{brighter:function(t){return new x(this.h,this.c,this.l+18*(null==t?1:t),this.opacity)},darker:function(t){return new x(this.h,this.c,this.l-18*(null==t?1:t),this.opacity)},rgb:function(){return f(this).rgb()}}))},"./node_modules/d3-color/src/math.js":function(t,e,n){"use strict";n.r(e),n.d(e,"deg2rad",function(){return r}),n.d(e,"rad2deg",function(){return i});var r=Math.PI/180,i=180/Math.PI},"./node_modules/d3-dispatch/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-dispatch/src/dispatch.js");n.d(e,"dispatch",function(){return r.default})},"./node_modules/d3-dispatch/src/dispatch.js":function(t,e,n){"use strict";n.r(e);var r={value:function(){}};function i(){for(var t,e=0,n=arguments.length,r={};e<n;++e){if(!(t=arguments[e]+"")||t in r)throw new Error("illegal type: "+t);r[t]=[]}return new o(r)}function o(t){this._=t}function a(t,e){for(var n,r=0,i=t.length;r<i;++r)if((n=t[r]).name===e)return n.value}function s(t,e,n){for(var i=0,o=t.length;i<o;++i)if(t[i].name===e){t[i]=r,t=t.slice(0,i).concat(t.slice(i+1));break}return null!=n&&t.push({name:e,value:n}),t}o.prototype=i.prototype={constructor:o,on:function(t,e){var n,r,i=this._,o=(r=i,(t+"").trim().split(/^|\s+/).map(function(t){var e="",n=t.indexOf(".");if(n>=0&&(e=t.slice(n+1),t=t.slice(0,n)),t&&!r.hasOwnProperty(t))throw new Error("unknown type: "+t);return{type:t,name:e}})),u=-1,c=o.length;if(!(arguments.length<2)){if(null!=e&&"function"!=typeof e)throw new Error("invalid callback: "+e);for(;++u<c;)if(n=(t=o[u]).type)i[n]=s(i[n],t.name,e);else if(null==e)for(n in i)i[n]=s(i[n],t.name,null);return this}for(;++u<c;)if((n=(t=o[u]).type)&&(n=a(i[n],t.name)))return n},copy:function(){var t={},e=this._;for(var n in e)t[n]=e[n].slice();return new o(t)},call:function(t,e){if((n=arguments.length-2)>0)for(var n,r,i=new Array(n),o=0;o<n;++o)i[o]=arguments[o+2];if(!this._.hasOwnProperty(t))throw new Error("unknown type: "+t);for(o=0,n=(r=this._[t]).length;o<n;++o)r[o].value.apply(e,i)},apply:function(t,e,n){if(!this._.hasOwnProperty(t))throw new Error("unknown type: "+t);for(var r=this._[t],i=0,o=r.length;i<o;++i)r[i].value.apply(e,n)}},e.default=i},"./node_modules/d3-drag/src/constant.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return function(){return t}}},"./node_modules/d3-drag/src/drag.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-dispatch/index.js"),i=n("./node_modules/d3-selection/src/index.js"),o=n("./node_modules/d3-drag/src/nodrag.js"),a=n("./node_modules/d3-drag/src/noevent.js"),s=n("./node_modules/d3-drag/src/constant.js"),u=n("./node_modules/d3-drag/src/event.js");function c(){return!i.event.button}function l(){return this.parentNode}function d(t){return null==t?{x:i.event.x,y:i.event.y}:t}function p(){return"ontouchstart"in this}e.default=function(){var t,e,n,f,h=c,m=l,g=d,y=p,_={},v=Object(r.dispatch)("start","drag","end"),b=0,x=0;function w(t){t.on("mousedown.drag",j).filter(y).on("touchstart.drag",E).on("touchmove.drag",S).on("touchend.drag touchcancel.drag",M).style("touch-action","none").style("-webkit-tap-highlight-color","rgba(0,0,0,0)")}function j(){if(!f&&h.apply(this,arguments)){var r=T("mouse",m.apply(this,arguments),i.mouse,this,arguments);r&&(Object(i.select)(i.event.view).on("mousemove.drag",O,!0).on("mouseup.drag",C,!0),Object(o.default)(i.event.view),Object(a.nopropagation)(),n=!1,t=i.event.clientX,e=i.event.clientY,r("start"))}}function O(){if(Object(a.default)(),!n){var r=i.event.clientX-t,o=i.event.clientY-e;n=r*r+o*o>x}_.mouse("drag")}function C(){Object(i.select)(i.event.view).on("mousemove.drag mouseup.drag",null),Object(o.yesdrag)(i.event.view,n),Object(a.default)(),_.mouse("end")}function E(){if(h.apply(this,arguments)){var t,e,n=i.event.changedTouches,r=m.apply(this,arguments),o=n.length;for(t=0;t<o;++t)(e=T(n[t].identifier,r,i.touch,this,arguments))&&(Object(a.nopropagation)(),e("start"))}}function S(){var t,e,n=i.event.changedTouches,r=n.length;for(t=0;t<r;++t)(e=_[n[t].identifier])&&(Object(a.default)(),e("drag"))}function M(){var t,e,n=i.event.changedTouches,r=n.length;for(f&&clearTimeout(f),f=setTimeout(function(){f=null},500),t=0;t<r;++t)(e=_[n[t].identifier])&&(Object(a.nopropagation)(),e("end"))}function T(t,e,n,r,o){var a,s,c,l=n(e,t),d=v.copy();if(Object(i.customEvent)(new u.default(w,"beforestart",a,t,b,l[0],l[1],0,0,d),function(){return null!=(i.event.subject=a=g.apply(r,o))&&(s=a.x-l[0]||0,c=a.y-l[1]||0,!0)}))return function p(f){var h,m=l;switch(f){case"start":_[t]=p,h=b++;break;case"end":delete _[t],--b;case"drag":l=n(e,t),h=b}Object(i.customEvent)(new u.default(w,f,a,t,h,l[0]+s,l[1]+c,l[0]-m[0],l[1]-m[1],d),d.apply,d,[f,r,o])}}return w.filter=function(t){return arguments.length?(h="function"==typeof t?t:Object(s.default)(!!t),w):h},w.container=function(t){return arguments.length?(m="function"==typeof t?t:Object(s.default)(t),w):m},w.subject=function(t){return arguments.length?(g="function"==typeof t?t:Object(s.default)(t),w):g},w.touchable=function(t){return arguments.length?(y="function"==typeof t?t:Object(s.default)(!!t),w):y},w.on=function(){var t=v.on.apply(v,arguments);return t===v?w:t},w.clickDistance=function(t){return arguments.length?(x=(t=+t)*t,w):Math.sqrt(x)},w}},"./node_modules/d3-drag/src/event.js":function(t,e,n){"use strict";function r(t,e,n,r,i,o,a,s,u,c){this.target=t,this.type=e,this.subject=n,this.identifier=r,this.active=i,this.x=o,this.y=a,this.dx=s,this.dy=u,this._=c}n.r(e),n.d(e,"default",function(){return r}),r.prototype.on=function(){var t=this._.on.apply(this._,arguments);return t===this._?this:t}},"./node_modules/d3-drag/src/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-drag/src/drag.js");n.d(e,"drag",function(){return r.default});var i=n("./node_modules/d3-drag/src/nodrag.js");n.d(e,"dragDisable",function(){return i.default}),n.d(e,"dragEnable",function(){return i.yesdrag})},"./node_modules/d3-drag/src/nodrag.js":function(t,e,n){"use strict";n.r(e),n.d(e,"yesdrag",function(){return o});var r=n("./node_modules/d3-selection/src/index.js"),i=n("./node_modules/d3-drag/src/noevent.js");function o(t,e){var n=t.document.documentElement,o=Object(r.select)(t).on("dragstart.drag",null);e&&(o.on("click.drag",i.default,!0),setTimeout(function(){o.on("click.drag",null)},0)),"onselectstart"in n?o.on("selectstart.drag",null):(n.style.MozUserSelect=n.__noselect,delete n.__noselect)}e.default=function(t){var e=t.document.documentElement,n=Object(r.select)(t).on("dragstart.drag",i.default,!0);"onselectstart"in e?n.on("selectstart.drag",i.default,!0):(e.__noselect=e.style.MozUserSelect,e.style.MozUserSelect="none")}},"./node_modules/d3-drag/src/noevent.js":function(t,e,n){"use strict";n.r(e),n.d(e,"nopropagation",function(){return i});var r=n("./node_modules/d3-selection/src/index.js");function i(){r.event.stopImmediatePropagation()}e.default=function(){r.event.preventDefault(),r.event.stopImmediatePropagation()}},"./node_modules/d3-ease/src/back.js":function(t,e,n){"use strict";n.r(e),n.d(e,"backIn",function(){return r}),n.d(e,"backOut",function(){return i}),n.d(e,"backInOut",function(){return o});var r=function t(e){function n(t){return t*t*((e+1)*t-e)}return e=+e,n.overshoot=t,n}(1.70158),i=function t(e){function n(t){return--t*t*((e+1)*t+e)+1}return e=+e,n.overshoot=t,n}(1.70158),o=function t(e){function n(t){return((t*=2)<1?t*t*((e+1)*t-e):(t-=2)*t*((e+1)*t+e)+2)/2}return e=+e,n.overshoot=t,n}(1.70158)},"./node_modules/d3-ease/src/bounce.js":function(t,e,n){"use strict";n.r(e),n.d(e,"bounceIn",function(){return f}),n.d(e,"bounceOut",function(){return h}),n.d(e,"bounceInOut",function(){return m});var r=4/11,i=6/11,o=8/11,a=.75,s=9/11,u=10/11,c=.9375,l=21/22,d=63/64,p=1/r/r;function f(t){return 1-h(1-t)}function h(t){return(t=+t)<r?p*t*t:t<o?p*(t-=i)*t+a:t<u?p*(t-=s)*t+c:p*(t-=l)*t+d}function m(t){return((t*=2)<=1?1-h(1-t):h(t-1)+1)/2}},"./node_modules/d3-ease/src/circle.js":function(t,e,n){"use strict";function r(t){return 1-Math.sqrt(1-t*t)}function i(t){return Math.sqrt(1- --t*t)}function o(t){return((t*=2)<=1?1-Math.sqrt(1-t*t):Math.sqrt(1-(t-=2)*t)+1)/2}n.r(e),n.d(e,"circleIn",function(){return r}),n.d(e,"circleOut",function(){return i}),n.d(e,"circleInOut",function(){return o})},"./node_modules/d3-ease/src/cubic.js":function(t,e,n){"use strict";function r(t){return t*t*t}function i(t){return--t*t*t+1}function o(t){return((t*=2)<=1?t*t*t:(t-=2)*t*t+2)/2}n.r(e),n.d(e,"cubicIn",function(){return r}),n.d(e,"cubicOut",function(){return i}),n.d(e,"cubicInOut",function(){return o})},"./node_modules/d3-ease/src/elastic.js":function(t,e,n){"use strict";n.r(e),n.d(e,"elasticIn",function(){return i}),n.d(e,"elasticOut",function(){return o}),n.d(e,"elasticInOut",function(){return a});var r=2*Math.PI,i=function t(e,n){var i=Math.asin(1/(e=Math.max(1,e)))*(n/=r);function o(t){return e*Math.pow(2,10*--t)*Math.sin((i-t)/n)}return o.amplitude=function(e){return t(e,n*r)},o.period=function(n){return t(e,n)},o}(1,.3),o=function t(e,n){var i=Math.asin(1/(e=Math.max(1,e)))*(n/=r);function o(t){return 1-e*Math.pow(2,-10*(t=+t))*Math.sin((t+i)/n)}return o.amplitude=function(e){return t(e,n*r)},o.period=function(n){return t(e,n)},o}(1,.3),a=function t(e,n){var i=Math.asin(1/(e=Math.max(1,e)))*(n/=r);function o(t){return((t=2*t-1)<0?e*Math.pow(2,10*t)*Math.sin((i-t)/n):2-e*Math.pow(2,-10*t)*Math.sin((i+t)/n))/2}return o.amplitude=function(e){return t(e,n*r)},o.period=function(n){return t(e,n)},o}(1,.3)},"./node_modules/d3-ease/src/exp.js":function(t,e,n){"use strict";function r(t){return Math.pow(2,10*t-10)}function i(t){return 1-Math.pow(2,-10*t)}function o(t){return((t*=2)<=1?Math.pow(2,10*t-10):2-Math.pow(2,10-10*t))/2}n.r(e),n.d(e,"expIn",function(){return r}),n.d(e,"expOut",function(){return i}),n.d(e,"expInOut",function(){return o})},"./node_modules/d3-ease/src/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-ease/src/linear.js");n.d(e,"easeLinear",function(){return r.linear});var i=n("./node_modules/d3-ease/src/quad.js");n.d(e,"easeQuad",function(){return i.quadInOut}),n.d(e,"easeQuadIn",function(){return i.quadIn}),n.d(e,"easeQuadOut",function(){return i.quadOut}),n.d(e,"easeQuadInOut",function(){return i.quadInOut});var o=n("./node_modules/d3-ease/src/cubic.js");n.d(e,"easeCubic",function(){return o.cubicInOut}),n.d(e,"easeCubicIn",function(){return o.cubicIn}),n.d(e,"easeCubicOut",function(){return o.cubicOut}),n.d(e,"easeCubicInOut",function(){return o.cubicInOut});var a=n("./node_modules/d3-ease/src/poly.js");n.d(e,"easePoly",function(){return a.polyInOut}),n.d(e,"easePolyIn",function(){return a.polyIn}),n.d(e,"easePolyOut",function(){return a.polyOut}),n.d(e,"easePolyInOut",function(){return a.polyInOut});var s=n("./node_modules/d3-ease/src/sin.js");n.d(e,"easeSin",function(){return s.sinInOut}),n.d(e,"easeSinIn",function(){return s.sinIn}),n.d(e,"easeSinOut",function(){return s.sinOut}),n.d(e,"easeSinInOut",function(){return s.sinInOut});var u=n("./node_modules/d3-ease/src/exp.js");n.d(e,"easeExp",function(){return u.expInOut}),n.d(e,"easeExpIn",function(){return u.expIn}),n.d(e,"easeExpOut",function(){return u.expOut}),n.d(e,"easeExpInOut",function(){return u.expInOut});var c=n("./node_modules/d3-ease/src/circle.js");n.d(e,"easeCircle",function(){return c.circleInOut}),n.d(e,"easeCircleIn",function(){return c.circleIn}),n.d(e,"easeCircleOut",function(){return c.circleOut}),n.d(e,"easeCircleInOut",function(){return c.circleInOut});var l=n("./node_modules/d3-ease/src/bounce.js");n.d(e,"easeBounce",function(){return l.bounceOut}),n.d(e,"easeBounceIn",function(){return l.bounceIn}),n.d(e,"easeBounceOut",function(){return l.bounceOut}),n.d(e,"easeBounceInOut",function(){return l.bounceInOut});var d=n("./node_modules/d3-ease/src/back.js");n.d(e,"easeBack",function(){return d.backInOut}),n.d(e,"easeBackIn",function(){return d.backIn}),n.d(e,"easeBackOut",function(){return d.backOut}),n.d(e,"easeBackInOut",function(){return d.backInOut});var p=n("./node_modules/d3-ease/src/elastic.js");n.d(e,"easeElastic",function(){return p.elasticOut}),n.d(e,"easeElasticIn",function(){return p.elasticIn}),n.d(e,"easeElasticOut",function(){return p.elasticOut}),n.d(e,"easeElasticInOut",function(){return p.elasticInOut})},"./node_modules/d3-ease/src/linear.js":function(t,e,n){"use strict";function r(t){return+t}n.r(e),n.d(e,"linear",function(){return r})},"./node_modules/d3-ease/src/poly.js":function(t,e,n){"use strict";n.r(e),n.d(e,"polyIn",function(){return r}),n.d(e,"polyOut",function(){return i}),n.d(e,"polyInOut",function(){return o});var r=function t(e){function n(t){return Math.pow(t,e)}return e=+e,n.exponent=t,n}(3),i=function t(e){function n(t){return 1-Math.pow(1-t,e)}return e=+e,n.exponent=t,n}(3),o=function t(e){function n(t){return((t*=2)<=1?Math.pow(t,e):2-Math.pow(2-t,e))/2}return e=+e,n.exponent=t,n}(3)},"./node_modules/d3-ease/src/quad.js":function(t,e,n){"use strict";function r(t){return t*t}function i(t){return t*(2-t)}function o(t){return((t*=2)<=1?t*t:--t*(2-t)+1)/2}n.r(e),n.d(e,"quadIn",function(){return r}),n.d(e,"quadOut",function(){return i}),n.d(e,"quadInOut",function(){return o})},"./node_modules/d3-ease/src/sin.js":function(t,e,n){"use strict";n.r(e),n.d(e,"sinIn",function(){return o}),n.d(e,"sinOut",function(){return a}),n.d(e,"sinInOut",function(){return s});var r=Math.PI,i=r/2;function o(t){return 1-Math.cos(t*i)}function a(t){return Math.sin(t*i)}function s(t){return(1-Math.cos(r*t))/2}},"./node_modules/d3-interpolate/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-interpolate/src/value.js");n.d(e,"interpolate",function(){return r.default});var i=n("./node_modules/d3-interpolate/src/array.js");n.d(e,"interpolateArray",function(){return i.default});var o=n("./node_modules/d3-interpolate/src/basis.js");n.d(e,"interpolateBasis",function(){return o.default});var a=n("./node_modules/d3-interpolate/src/basisClosed.js");n.d(e,"interpolateBasisClosed",function(){return a.default});var s=n("./node_modules/d3-interpolate/src/date.js");n.d(e,"interpolateDate",function(){return s.default});var u=n("./node_modules/d3-interpolate/src/number.js");n.d(e,"interpolateNumber",function(){return u.default});var c=n("./node_modules/d3-interpolate/src/object.js");n.d(e,"interpolateObject",function(){return c.default});var l=n("./node_modules/d3-interpolate/src/round.js");n.d(e,"interpolateRound",function(){return l.default});var d=n("./node_modules/d3-interpolate/src/string.js");n.d(e,"interpolateString",function(){return d.default});var p=n("./node_modules/d3-interpolate/src/transform/index.js");n.d(e,"interpolateTransformCss",function(){return p.interpolateTransformCss}),n.d(e,"interpolateTransformSvg",function(){return p.interpolateTransformSvg});var f=n("./node_modules/d3-interpolate/src/zoom.js");n.d(e,"interpolateZoom",function(){return f.default});var h=n("./node_modules/d3-interpolate/src/rgb.js");n.d(e,"interpolateRgb",function(){return h.default}),n.d(e,"interpolateRgbBasis",function(){return h.rgbBasis}),n.d(e,"interpolateRgbBasisClosed",function(){return h.rgbBasisClosed});var m=n("./node_modules/d3-interpolate/src/hsl.js");n.d(e,"interpolateHsl",function(){return m.default}),n.d(e,"interpolateHslLong",function(){return m.hslLong});var g=n("./node_modules/d3-interpolate/src/lab.js");n.d(e,"interpolateLab",function(){return g.default});var y=n("./node_modules/d3-interpolate/src/hcl.js");n.d(e,"interpolateHcl",function(){return y.default}),n.d(e,"interpolateHclLong",function(){return y.hclLong});var _=n("./node_modules/d3-interpolate/src/cubehelix.js");n.d(e,"interpolateCubehelix",function(){return _.default}),n.d(e,"interpolateCubehelixLong",function(){return _.cubehelixLong});var v=n("./node_modules/d3-interpolate/src/quantize.js");n.d(e,"quantize",function(){return v.default})},"./node_modules/d3-interpolate/src/array.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-interpolate/src/value.js");e.default=function(t,e){var n,i=e?e.length:0,o=t?Math.min(i,t.length):0,a=new Array(o),s=new Array(i);for(n=0;n<o;++n)a[n]=Object(r.default)(t[n],e[n]);for(;n<i;++n)s[n]=e[n];return function(t){for(n=0;n<o;++n)s[n]=a[n](t);return s}}},"./node_modules/d3-interpolate/src/basis.js":function(t,e,n){"use strict";function r(t,e,n,r,i){var o=t*t,a=o*t;return((1-3*t+3*o-a)*e+(4-6*o+3*a)*n+(1+3*t+3*o-3*a)*r+a*i)/6}n.r(e),n.d(e,"basis",function(){return r}),e.default=function(t){var e=t.length-1;return function(n){var i=n<=0?n=0:n>=1?(n=1,e-1):Math.floor(n*e),o=t[i],a=t[i+1],s=i>0?t[i-1]:2*o-a,u=i<e-1?t[i+2]:2*a-o;return r((n-i/e)*e,s,o,a,u)}}},"./node_modules/d3-interpolate/src/basisClosed.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-interpolate/src/basis.js");e.default=function(t){var e=t.length;return function(n){var i=Math.floor(((n%=1)<0?++n:n)*e),o=t[(i+e-1)%e],a=t[i%e],s=t[(i+1)%e],u=t[(i+2)%e];return Object(r.basis)((n-i/e)*e,o,a,s,u)}}},"./node_modules/d3-interpolate/src/color.js":function(t,e,n){"use strict";n.r(e),n.d(e,"hue",function(){return o}),n.d(e,"gamma",function(){return a}),n.d(e,"default",function(){return s});var r=n("./node_modules/d3-interpolate/src/constant.js");function i(t,e){return function(n){return t+n*e}}function o(t,e){var n=e-t;return n?i(t,n>180||n<-180?n-360*Math.round(n/360):n):Object(r.default)(isNaN(t)?e:t)}function a(t){return 1==(t=+t)?s:function(e,n){return n-e?(i=e,o=n,a=t,i=Math.pow(i,a),o=Math.pow(o,a)-i,a=1/a,function(t){return Math.pow(i+t*o,a)}):Object(r.default)(isNaN(e)?n:e);var i,o,a}}function s(t,e){var n=e-t;return n?i(t,n):Object(r.default)(isNaN(t)?e:t)}},"./node_modules/d3-interpolate/src/constant.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return function(){return t}}},"./node_modules/d3-interpolate/src/cubehelix.js":function(t,e,n){"use strict";n.r(e),n.d(e,"cubehelixLong",function(){return a});var r=n("./node_modules/d3-color/index.js"),i=n("./node_modules/d3-interpolate/src/color.js");function o(t){return function e(n){function o(e,o){var a=t((e=Object(r.cubehelix)(e)).h,(o=Object(r.cubehelix)(o)).h),s=Object(i.default)(e.s,o.s),u=Object(i.default)(e.l,o.l),c=Object(i.default)(e.opacity,o.opacity);return function(t){return e.h=a(t),e.s=s(t),e.l=u(Math.pow(t,n)),e.opacity=c(t),e+""}}return n=+n,o.gamma=e,o}(1)}e.default=o(i.hue);var a=o(i.default)},"./node_modules/d3-interpolate/src/date.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){var n=new Date;return e-=t=+t,function(r){return n.setTime(t+e*r),n}}},"./node_modules/d3-interpolate/src/hcl.js":function(t,e,n){"use strict";n.r(e),n.d(e,"hclLong",function(){return a});var r=n("./node_modules/d3-color/index.js"),i=n("./node_modules/d3-interpolate/src/color.js");function o(t){return function(e,n){var o=t((e=Object(r.hcl)(e)).h,(n=Object(r.hcl)(n)).h),a=Object(i.default)(e.c,n.c),s=Object(i.default)(e.l,n.l),u=Object(i.default)(e.opacity,n.opacity);return function(t){return e.h=o(t),e.c=a(t),e.l=s(t),e.opacity=u(t),e+""}}}e.default=o(i.hue);var a=o(i.default)},"./node_modules/d3-interpolate/src/hsl.js":function(t,e,n){"use strict";n.r(e),n.d(e,"hslLong",function(){return a});var r=n("./node_modules/d3-color/index.js"),i=n("./node_modules/d3-interpolate/src/color.js");function o(t){return function(e,n){var o=t((e=Object(r.hsl)(e)).h,(n=Object(r.hsl)(n)).h),a=Object(i.default)(e.s,n.s),s=Object(i.default)(e.l,n.l),u=Object(i.default)(e.opacity,n.opacity);return function(t){return e.h=o(t),e.s=a(t),e.l=s(t),e.opacity=u(t),e+""}}}e.default=o(i.hue);var a=o(i.default)},"./node_modules/d3-interpolate/src/lab.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return o});var r=n("./node_modules/d3-color/index.js"),i=n("./node_modules/d3-interpolate/src/color.js");function o(t,e){var n=Object(i.default)((t=Object(r.lab)(t)).l,(e=Object(r.lab)(e)).l),o=Object(i.default)(t.a,e.a),a=Object(i.default)(t.b,e.b),s=Object(i.default)(t.opacity,e.opacity);return function(e){return t.l=n(e),t.a=o(e),t.b=a(e),t.opacity=s(e),t+""}}},"./node_modules/d3-interpolate/src/number.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){return e-=t=+t,function(n){return t+e*n}}},"./node_modules/d3-interpolate/src/object.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-interpolate/src/value.js");e.default=function(t,e){var n,i={},o={};null!==t&&"object"==typeof t||(t={}),null!==e&&"object"==typeof e||(e={});for(n in e)n in t?i[n]=Object(r.default)(t[n],e[n]):o[n]=e[n];return function(t){for(n in i)o[n]=i[n](t);return o}}},"./node_modules/d3-interpolate/src/quantize.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){for(var n=new Array(e),r=0;r<e;++r)n[r]=t(r/(e-1));return n}},"./node_modules/d3-interpolate/src/rgb.js":function(t,e,n){"use strict";n.r(e),n.d(e,"rgbBasis",function(){return u}),n.d(e,"rgbBasisClosed",function(){return c});var r=n("./node_modules/d3-color/index.js"),i=n("./node_modules/d3-interpolate/src/basis.js"),o=n("./node_modules/d3-interpolate/src/basisClosed.js"),a=n("./node_modules/d3-interpolate/src/color.js");function s(t){return function(e){var n,i,o=e.length,a=new Array(o),s=new Array(o),u=new Array(o);for(n=0;n<o;++n)i=Object(r.rgb)(e[n]),a[n]=i.r||0,s[n]=i.g||0,u[n]=i.b||0;return a=t(a),s=t(s),u=t(u),i.opacity=1,function(t){return i.r=a(t),i.g=s(t),i.b=u(t),i+""}}}e.default=function t(e){var n=Object(a.gamma)(e);function i(t,e){var i=n((t=Object(r.rgb)(t)).r,(e=Object(r.rgb)(e)).r),o=n(t.g,e.g),s=n(t.b,e.b),u=Object(a.default)(t.opacity,e.opacity);return function(e){return t.r=i(e),t.g=o(e),t.b=s(e),t.opacity=u(e),t+""}}return i.gamma=t,i}(1);var u=s(i.default),c=s(o.default)},"./node_modules/d3-interpolate/src/round.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){return e-=t=+t,function(n){return Math.round(t+e*n)}}},"./node_modules/d3-interpolate/src/string.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-interpolate/src/number.js"),i=/[-+]?(?:\d+\.?\d*|\.?\d+)(?:[eE][-+]?\d+)?/g,o=new RegExp(i.source,"g");e.default=function(t,e){var n,a,s,u,c,l=i.lastIndex=o.lastIndex=0,d=-1,p=[],f=[];for(t+="",e+="";(n=i.exec(t))&&(a=o.exec(e));)(s=a.index)>l&&(s=e.slice(l,s),p[d]?p[d]+=s:p[++d]=s),(n=n[0])===(a=a[0])?p[d]?p[d]+=a:p[++d]=a:(p[++d]=null,f.push({i:d,x:Object(r.default)(n,a)})),l=o.lastIndex;return l<e.length&&(s=e.slice(l),p[d]?p[d]+=s:p[++d]=s),p.length<2?f[0]?(c=f[0].x,function(t){return c(t)+""}):(u=e,function(){return u}):(e=f.length,function(t){for(var n,r=0;r<e;++r)p[(n=f[r]).i]=n.x(t);return p.join("")})}},"./node_modules/d3-interpolate/src/transform/decompose.js":function(t,e,n){"use strict";n.r(e),n.d(e,"identity",function(){return i});var r=180/Math.PI,i={translateX:0,translateY:0,rotate:0,skewX:0,scaleX:1,scaleY:1};e.default=function(t,e,n,i,o,a){var s,u,c;return(s=Math.sqrt(t*t+e*e))&&(t/=s,e/=s),(c=t*n+e*i)&&(n-=t*c,i-=e*c),(u=Math.sqrt(n*n+i*i))&&(n/=u,i/=u,c/=u),t*i<e*n&&(t=-t,e=-e,c=-c,s=-s),{translateX:o,translateY:a,rotate:Math.atan2(e,t)*r,skewX:Math.atan(c)*r,scaleX:s,scaleY:u}}},"./node_modules/d3-interpolate/src/transform/index.js":function(t,e,n){"use strict";n.r(e),n.d(e,"interpolateTransformCss",function(){return a}),n.d(e,"interpolateTransformSvg",function(){return s});var r=n("./node_modules/d3-interpolate/src/number.js"),i=n("./node_modules/d3-interpolate/src/transform/parse.js");function o(t,e,n,i){function o(t){return t.length?t.pop()+" ":""}return function(a,s){var u,c,l,d,p,f,h=[],m=[];return a=t(a),s=t(s),function(t,i,o,a,s,u){if(t!==o||i!==a){var c=s.push("translate(",null,e,null,n);u.push({i:c-4,x:Object(r.default)(t,o)},{i:c-2,x:Object(r.default)(i,a)})}else(o||a)&&s.push("translate("+o+e+a+n)}(a.translateX,a.translateY,s.translateX,s.translateY,h,m),u=a.rotate,c=s.rotate,l=h,u!==c?(u-c>180?c+=360:c-u>180&&(u+=360),m.push({i:l.push(o(l)+"rotate(",null,i)-2,x:Object(r.default)(u,c)})):c&&l.push(o(l)+"rotate("+c+i),d=a.skewX,p=s.skewX,f=h,d!==p?m.push({i:f.push(o(f)+"skewX(",null,i)-2,x:Object(r.default)(d,p)}):p&&f.push(o(f)+"skewX("+p+i),function(t,e,n,i,a,s){if(t!==n||e!==i){var u=a.push(o(a)+"scale(",null,",",null,")");s.push({i:u-4,x:Object(r.default)(t,n)},{i:u-2,x:Object(r.default)(e,i)})}else 1===n&&1===i||a.push(o(a)+"scale("+n+","+i+")")}(a.scaleX,a.scaleY,s.scaleX,s.scaleY,h,m),a=s=null,function(t){for(var e,n=-1,r=m.length;++n<r;)h[(e=m[n]).i]=e.x(t);return h.join("")}}}var a=o(i.parseCss,"px, ","px)","deg)"),s=o(i.parseSvg,", ",")",")")},"./node_modules/d3-interpolate/src/transform/parse.js":function(t,e,n){"use strict";n.r(e),n.d(e,"parseCss",function(){return u}),n.d(e,"parseSvg",function(){return c});var r,i,o,a,s=n("./node_modules/d3-interpolate/src/transform/decompose.js");function u(t){return"none"===t?s.identity:(r||(r=document.createElement("DIV"),i=document.documentElement,o=document.defaultView),r.style.transform=t,t=o.getComputedStyle(i.appendChild(r),null).getPropertyValue("transform"),i.removeChild(r),t=t.slice(7,-1).split(","),Object(s.default)(+t[0],+t[1],+t[2],+t[3],+t[4],+t[5]))}function c(t){return null==t?s.identity:(a||(a=document.createElementNS("http://www.w3.org/2000/svg","g")),a.setAttribute("transform",t),(t=a.transform.baseVal.consolidate())?(t=t.matrix,Object(s.default)(t.a,t.b,t.c,t.d,t.e,t.f)):s.identity)}},"./node_modules/d3-interpolate/src/value.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-color/index.js"),i=n("./node_modules/d3-interpolate/src/rgb.js"),o=n("./node_modules/d3-interpolate/src/array.js"),a=n("./node_modules/d3-interpolate/src/date.js"),s=n("./node_modules/d3-interpolate/src/number.js"),u=n("./node_modules/d3-interpolate/src/object.js"),c=n("./node_modules/d3-interpolate/src/string.js"),l=n("./node_modules/d3-interpolate/src/constant.js");e.default=function(t,e){var n,d=typeof e;return null==e||"boolean"===d?Object(l.default)(e):("number"===d?s.default:"string"===d?(n=Object(r.color)(e))?(e=n,i.default):c.default:e instanceof r.color?i.default:e instanceof Date?a.default:Array.isArray(e)?o.default:"function"!=typeof e.valueOf&&"function"!=typeof e.toString||isNaN(e)?u.default:s.default)(t,e)}},"./node_modules/d3-interpolate/src/zoom.js":function(t,e,n){"use strict";n.r(e);var r=Math.SQRT2;function i(t){return((t=Math.exp(t))+1/t)/2}e.default=function(t,e){var n,o,a=t[0],s=t[1],u=t[2],c=e[0],l=e[1],d=e[2],p=c-a,f=l-s,h=p*p+f*f;if(h<1e-12)o=Math.log(d/u)/r,n=function(t){return[a+t*p,s+t*f,u*Math.exp(r*t*o)]};else{var m=Math.sqrt(h),g=(d*d-u*u+4*h)/(2*u*2*m),y=(d*d-u*u-4*h)/(2*d*2*m),_=Math.log(Math.sqrt(g*g+1)-g),v=Math.log(Math.sqrt(y*y+1)-y);o=(v-_)/r,n=function(t){var e,n,c=t*o,l=i(_),d=u/(2*m)*(l*(n=r*c+_,((n=Math.exp(2*n))-1)/(n+1))-(e=_,((e=Math.exp(e))-1/e)/2));return[a+d*p,s+d*f,u*l/i(r*c+_)]}}return n.duration=1e3*o,n}},"./node_modules/d3-path/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-path/src/path.js");n.d(e,"path",function(){return r.default})},"./node_modules/d3-path/src/path.js":function(t,e,n){"use strict";n.r(e);var r=Math.PI,i=2*r,o=i-1e-6;function a(){this._x0=this._y0=this._x1=this._y1=null,this._=""}function s(){return new a}a.prototype=s.prototype={constructor:a,moveTo:function(t,e){this._+="M"+(this._x0=this._x1=+t)+","+(this._y0=this._y1=+e)},closePath:function(){null!==this._x1&&(this._x1=this._x0,this._y1=this._y0,this._+="Z")},lineTo:function(t,e){this._+="L"+(this._x1=+t)+","+(this._y1=+e)},quadraticCurveTo:function(t,e,n,r){this._+="Q"+ +t+","+ +e+","+(this._x1=+n)+","+(this._y1=+r)},bezierCurveTo:function(t,e,n,r,i,o){this._+="C"+ +t+","+ +e+","+ +n+","+ +r+","+(this._x1=+i)+","+(this._y1=+o)},arcTo:function(t,e,n,i,o){t=+t,e=+e,n=+n,i=+i,o=+o;var a=this._x1,s=this._y1,u=n-t,c=i-e,l=a-t,d=s-e,p=l*l+d*d;if(o<0)throw new Error("negative radius: "+o);if(null===this._x1)this._+="M"+(this._x1=t)+","+(this._y1=e);else if(p>1e-6)if(Math.abs(d*u-c*l)>1e-6&&o){var f=n-a,h=i-s,m=u*u+c*c,g=f*f+h*h,y=Math.sqrt(m),_=Math.sqrt(p),v=o*Math.tan((r-Math.acos((m+p-g)/(2*y*_)))/2),b=v/_,x=v/y;Math.abs(b-1)>1e-6&&(this._+="L"+(t+b*l)+","+(e+b*d)),this._+="A"+o+","+o+",0,0,"+ +(d*f>l*h)+","+(this._x1=t+x*u)+","+(this._y1=e+x*c)}else this._+="L"+(this._x1=t)+","+(this._y1=e);else;},arc:function(t,e,n,a,s,u){t=+t,e=+e;var c=(n=+n)*Math.cos(a),l=n*Math.sin(a),d=t+c,p=e+l,f=1^u,h=u?a-s:s-a;if(n<0)throw new Error("negative radius: "+n);null===this._x1?this._+="M"+d+","+p:(Math.abs(this._x1-d)>1e-6||Math.abs(this._y1-p)>1e-6)&&(this._+="L"+d+","+p),n&&(h<0&&(h=h%i+i),h>o?this._+="A"+n+","+n+",0,1,"+f+","+(t-c)+","+(e-l)+"A"+n+","+n+",0,1,"+f+","+(this._x1=d)+","+(this._y1=p):h>1e-6&&(this._+="A"+n+","+n+",0,"+ +(h>=r)+","+f+","+(this._x1=t+n*Math.cos(s))+","+(this._y1=e+n*Math.sin(s))))},rect:function(t,e,n,r){this._+="M"+(this._x0=this._x1=+t)+","+(this._y0=this._y1=+e)+"h"+ +n+"v"+ +r+"h"+-n+"Z"},toString:function(){return this._}},e.default=s},"./node_modules/d3-selection/src/constant.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return function(){return t}}},"./node_modules/d3-selection/src/create.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/creator.js"),i=n("./node_modules/d3-selection/src/select.js");e.default=function(t){return Object(i.default)(Object(r.default)(t).call(document.documentElement))}},"./node_modules/d3-selection/src/creator.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/namespace.js"),i=n("./node_modules/d3-selection/src/namespaces.js");e.default=function(t){var e=Object(r.default)(t);return(e.local?function(t){return function(){return this.ownerDocument.createElementNS(t.space,t.local)}}:function(t){return function(){var e=this.ownerDocument,n=this.namespaceURI;return n===i.xhtml&&e.documentElement.namespaceURI===i.xhtml?e.createElement(t):e.createElementNS(n,t)}})(e)}},"./node_modules/d3-selection/src/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/create.js");n.d(e,"create",function(){return r.default});var i=n("./node_modules/d3-selection/src/creator.js");n.d(e,"creator",function(){return i.default});var o=n("./node_modules/d3-selection/src/local.js");n.d(e,"local",function(){return o.default});var a=n("./node_modules/d3-selection/src/matcher.js");n.d(e,"matcher",function(){return a.default});var s=n("./node_modules/d3-selection/src/mouse.js");n.d(e,"mouse",function(){return s.default});var u=n("./node_modules/d3-selection/src/namespace.js");n.d(e,"namespace",function(){return u.default});var c=n("./node_modules/d3-selection/src/namespaces.js");n.d(e,"namespaces",function(){return c.default});var l=n("./node_modules/d3-selection/src/point.js");n.d(e,"clientPoint",function(){return l.default});var d=n("./node_modules/d3-selection/src/select.js");n.d(e,"select",function(){return d.default});var p=n("./node_modules/d3-selection/src/selectAll.js");n.d(e,"selectAll",function(){return p.default});var f=n("./node_modules/d3-selection/src/selection/index.js");n.d(e,"selection",function(){return f.default});var h=n("./node_modules/d3-selection/src/selector.js");n.d(e,"selector",function(){return h.default});var m=n("./node_modules/d3-selection/src/selectorAll.js");n.d(e,"selectorAll",function(){return m.default});var g=n("./node_modules/d3-selection/src/selection/style.js");n.d(e,"style",function(){return g.styleValue});var y=n("./node_modules/d3-selection/src/touch.js");n.d(e,"touch",function(){return y.default});var _=n("./node_modules/d3-selection/src/touches.js");n.d(e,"touches",function(){return _.default});var v=n("./node_modules/d3-selection/src/window.js");n.d(e,"window",function(){return v.default});var b=n("./node_modules/d3-selection/src/selection/on.js");n.d(e,"event",function(){return b.event}),n.d(e,"customEvent",function(){return b.customEvent})},"./node_modules/d3-selection/src/local.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return i});var r=0;function i(){return new o}function o(){this._="@"+(++r).toString(36)}o.prototype=i.prototype={constructor:o,get:function(t){for(var e=this._;!(e in t);)if(!(t=t.parentNode))return;return t[e]},set:function(t,e){return t[this._]=e},remove:function(t){return this._ in t&&delete t[this._]},toString:function(){return this._}}},"./node_modules/d3-selection/src/matcher.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return function(){return this.matches(t)}}},"./node_modules/d3-selection/src/mouse.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/sourceEvent.js"),i=n("./node_modules/d3-selection/src/point.js");e.default=function(t){var e=Object(r.default)();return e.changedTouches&&(e=e.changedTouches[0]),Object(i.default)(t,e)}},"./node_modules/d3-selection/src/namespace.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/namespaces.js");e.default=function(t){var e=t+="",n=e.indexOf(":");return n>=0&&"xmlns"!==(e=t.slice(0,n))&&(t=t.slice(n+1)),r.default.hasOwnProperty(e)?{space:r.default[e],local:t}:t}},"./node_modules/d3-selection/src/namespaces.js":function(t,e,n){"use strict";n.r(e),n.d(e,"xhtml",function(){return r});var r="http://www.w3.org/1999/xhtml";e.default={svg:"http://www.w3.org/2000/svg",xhtml:r,xlink:"http://www.w3.org/1999/xlink",xml:"http://www.w3.org/XML/1998/namespace",xmlns:"http://www.w3.org/2000/xmlns/"}},"./node_modules/d3-selection/src/point.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){var n=t.ownerSVGElement||t;if(n.createSVGPoint){var r=n.createSVGPoint();return r.x=e.clientX,r.y=e.clientY,[(r=r.matrixTransform(t.getScreenCTM().inverse())).x,r.y]}var i=t.getBoundingClientRect();return[e.clientX-i.left-t.clientLeft,e.clientY-i.top-t.clientTop]}},"./node_modules/d3-selection/src/select.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/selection/index.js");e.default=function(t){return"string"==typeof t?new r.Selection([[document.querySelector(t)]],[document.documentElement]):new r.Selection([[t]],r.root)}},"./node_modules/d3-selection/src/selectAll.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/selection/index.js");e.default=function(t){return"string"==typeof t?new r.Selection([document.querySelectorAll(t)],[document.documentElement]):new r.Selection([null==t?[]:t],r.root)}},"./node_modules/d3-selection/src/selection/append.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/creator.js");e.default=function(t){var e="function"==typeof t?t:Object(r.default)(t);return this.select(function(){return this.appendChild(e.apply(this,arguments))})}},"./node_modules/d3-selection/src/selection/attr.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/namespace.js");e.default=function(t,e){var n=Object(r.default)(t);if(arguments.length<2){var i=this.node();return n.local?i.getAttributeNS(n.space,n.local):i.getAttribute(n)}return this.each((null==e?n.local?function(t){return function(){this.removeAttributeNS(t.space,t.local)}}:function(t){return function(){this.removeAttribute(t)}}:"function"==typeof e?n.local?function(t,e){return function(){var n=e.apply(this,arguments);null==n?this.removeAttributeNS(t.space,t.local):this.setAttributeNS(t.space,t.local,n)}}:function(t,e){return function(){var n=e.apply(this,arguments);null==n?this.removeAttribute(t):this.setAttribute(t,n)}}:n.local?function(t,e){return function(){this.setAttributeNS(t.space,t.local,e)}}:function(t,e){return function(){this.setAttribute(t,e)}})(n,e))}},"./node_modules/d3-selection/src/selection/call.js":function(t,e,n){"use strict";n.r(e),e.default=function(){var t=arguments[0];return arguments[0]=this,t.apply(null,arguments),this}},"./node_modules/d3-selection/src/selection/classed.js":function(t,e,n){"use strict";function r(t){return t.trim().split(/^|\s+/)}function i(t){return t.classList||new o(t)}function o(t){this._node=t,this._names=r(t.getAttribute("class")||"")}function a(t,e){for(var n=i(t),r=-1,o=e.length;++r<o;)n.add(e[r])}function s(t,e){for(var n=i(t),r=-1,o=e.length;++r<o;)n.remove(e[r])}n.r(e),o.prototype={add:function(t){this._names.indexOf(t)<0&&(this._names.push(t),this._node.setAttribute("class",this._names.join(" ")))},remove:function(t){var e=this._names.indexOf(t);e>=0&&(this._names.splice(e,1),this._node.setAttribute("class",this._names.join(" ")))},contains:function(t){return this._names.indexOf(t)>=0}},e.default=function(t,e){var n=r(t+"");if(arguments.length<2){for(var o=i(this.node()),u=-1,c=n.length;++u<c;)if(!o.contains(n[u]))return!1;return!0}return this.each(("function"==typeof e?function(t,e){return function(){(e.apply(this,arguments)?a:s)(this,t)}}:e?function(t){return function(){a(this,t)}}:function(t){return function(){s(this,t)}})(n,e))}},"./node_modules/d3-selection/src/selection/clone.js":function(t,e,n){"use strict";function r(){return this.parentNode.insertBefore(this.cloneNode(!1),this.nextSibling)}function i(){return this.parentNode.insertBefore(this.cloneNode(!0),this.nextSibling)}n.r(e),e.default=function(t){return this.select(t?i:r)}},"./node_modules/d3-selection/src/selection/data.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/selection/index.js"),i=n("./node_modules/d3-selection/src/selection/enter.js"),o=n("./node_modules/d3-selection/src/constant.js"),a="$";function s(t,e,n,r,o,a){for(var s,u=0,c=e.length,l=a.length;u<l;++u)(s=e[u])?(s.__data__=a[u],r[u]=s):n[u]=new i.EnterNode(t,a[u]);for(;u<c;++u)(s=e[u])&&(o[u]=s)}function u(t,e,n,r,o,s,u){var c,l,d,p={},f=e.length,h=s.length,m=new Array(f);for(c=0;c<f;++c)(l=e[c])&&(m[c]=d=a+u.call(l,l.__data__,c,e),d in p?o[c]=l:p[d]=l);for(c=0;c<h;++c)(l=p[d=a+u.call(t,s[c],c,s)])?(r[c]=l,l.__data__=s[c],p[d]=null):n[c]=new i.EnterNode(t,s[c]);for(c=0;c<f;++c)(l=e[c])&&p[m[c]]===l&&(o[c]=l)}e.default=function(t,e){if(!t)return y=new Array(this.size()),f=-1,this.each(function(t){y[++f]=t}),y;var n=e?u:s,i=this._parents,a=this._groups;"function"!=typeof t&&(t=Object(o.default)(t));for(var c=a.length,l=new Array(c),d=new Array(c),p=new Array(c),f=0;f<c;++f){var h=i[f],m=a[f],g=m.length,y=t.call(h,h&&h.__data__,f,i),_=y.length,v=d[f]=new Array(_),b=l[f]=new Array(_);n(h,m,v,b,p[f]=new Array(g),y,e);for(var x,w,j=0,O=0;j<_;++j)if(x=v[j]){for(j>=O&&(O=j+1);!(w=b[O])&&++O<_;);x._next=w||null}}return(l=new r.Selection(l,i))._enter=d,l._exit=p,l}},"./node_modules/d3-selection/src/selection/datum.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return arguments.length?this.property("__data__",t):this.node().__data__}},"./node_modules/d3-selection/src/selection/dispatch.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/window.js");function i(t,e,n){var i=Object(r.default)(t),o=i.CustomEvent;"function"==typeof o?o=new o(e,n):(o=i.document.createEvent("Event"),n?(o.initEvent(e,n.bubbles,n.cancelable),o.detail=n.detail):o.initEvent(e,!1,!1)),t.dispatchEvent(o)}e.default=function(t,e){return this.each(("function"==typeof e?function(t,e){return function(){return i(this,t,e.apply(this,arguments))}}:function(t,e){return function(){return i(this,t,e)}})(t,e))}},"./node_modules/d3-selection/src/selection/each.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){for(var e=this._groups,n=0,r=e.length;n<r;++n)for(var i,o=e[n],a=0,s=o.length;a<s;++a)(i=o[a])&&t.call(i,i.__data__,a,o);return this}},"./node_modules/d3-selection/src/selection/empty.js":function(t,e,n){"use strict";n.r(e),e.default=function(){return!this.node()}},"./node_modules/d3-selection/src/selection/enter.js":function(t,e,n){"use strict";n.r(e),n.d(e,"EnterNode",function(){return o});var r=n("./node_modules/d3-selection/src/selection/sparse.js"),i=n("./node_modules/d3-selection/src/selection/index.js");function o(t,e){this.ownerDocument=t.ownerDocument,this.namespaceURI=t.namespaceURI,this._next=null,this._parent=t,this.__data__=e}e.default=function(){return new i.Selection(this._enter||this._groups.map(r.default),this._parents)},o.prototype={constructor:o,appendChild:function(t){return this._parent.insertBefore(t,this._next)},insertBefore:function(t,e){return this._parent.insertBefore(t,e)},querySelector:function(t){return this._parent.querySelector(t)},querySelectorAll:function(t){return this._parent.querySelectorAll(t)}}},"./node_modules/d3-selection/src/selection/exit.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/selection/sparse.js"),i=n("./node_modules/d3-selection/src/selection/index.js");e.default=function(){return new i.Selection(this._exit||this._groups.map(r.default),this._parents)}},"./node_modules/d3-selection/src/selection/filter.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/selection/index.js"),i=n("./node_modules/d3-selection/src/matcher.js");e.default=function(t){"function"!=typeof t&&(t=Object(i.default)(t));for(var e=this._groups,n=e.length,o=new Array(n),a=0;a<n;++a)for(var s,u=e[a],c=u.length,l=o[a]=[],d=0;d<c;++d)(s=u[d])&&t.call(s,s.__data__,d,u)&&l.push(s);return new r.Selection(o,this._parents)}},"./node_modules/d3-selection/src/selection/html.js":function(t,e,n){"use strict";function r(){this.innerHTML=""}n.r(e),e.default=function(t){return arguments.length?this.each(null==t?r:("function"==typeof t?function(t){return function(){var e=t.apply(this,arguments);this.innerHTML=null==e?"":e}}:function(t){return function(){this.innerHTML=t}})(t)):this.node().innerHTML}},"./node_modules/d3-selection/src/selection/index.js":function(t,e,n){"use strict";n.r(e),n.d(e,"root",function(){return P}),n.d(e,"Selection",function(){return R});var r=n("./node_modules/d3-selection/src/selection/select.js"),i=n("./node_modules/d3-selection/src/selection/selectAll.js"),o=n("./node_modules/d3-selection/src/selection/filter.js"),a=n("./node_modules/d3-selection/src/selection/data.js"),s=n("./node_modules/d3-selection/src/selection/enter.js"),u=n("./node_modules/d3-selection/src/selection/exit.js"),c=n("./node_modules/d3-selection/src/selection/join.js"),l=n("./node_modules/d3-selection/src/selection/merge.js"),d=n("./node_modules/d3-selection/src/selection/order.js"),p=n("./node_modules/d3-selection/src/selection/sort.js"),f=n("./node_modules/d3-selection/src/selection/call.js"),h=n("./node_modules/d3-selection/src/selection/nodes.js"),m=n("./node_modules/d3-selection/src/selection/node.js"),g=n("./node_modules/d3-selection/src/selection/size.js"),y=n("./node_modules/d3-selection/src/selection/empty.js"),_=n("./node_modules/d3-selection/src/selection/each.js"),v=n("./node_modules/d3-selection/src/selection/attr.js"),b=n("./node_modules/d3-selection/src/selection/style.js"),x=n("./node_modules/d3-selection/src/selection/property.js"),w=n("./node_modules/d3-selection/src/selection/classed.js"),j=n("./node_modules/d3-selection/src/selection/text.js"),O=n("./node_modules/d3-selection/src/selection/html.js"),C=n("./node_modules/d3-selection/src/selection/raise.js"),E=n("./node_modules/d3-selection/src/selection/lower.js"),S=n("./node_modules/d3-selection/src/selection/append.js"),M=n("./node_modules/d3-selection/src/selection/insert.js"),T=n("./node_modules/d3-selection/src/selection/remove.js"),k=n("./node_modules/d3-selection/src/selection/clone.js"),A=n("./node_modules/d3-selection/src/selection/datum.js"),D=n("./node_modules/d3-selection/src/selection/on.js"),I=n("./node_modules/d3-selection/src/selection/dispatch.js"),P=[null];function R(t,e){this._groups=t,this._parents=e}function N(){return new R([[document.documentElement]],P)}R.prototype=N.prototype={constructor:R,select:r.default,selectAll:i.default,filter:o.default,data:a.default,enter:s.default,exit:u.default,join:c.default,merge:l.default,order:d.default,sort:p.default,call:f.default,nodes:h.default,node:m.default,size:g.default,empty:y.default,each:_.default,attr:v.default,style:b.default,property:x.default,classed:w.default,text:j.default,html:O.default,raise:C.default,lower:E.default,append:S.default,insert:M.default,remove:T.default,clone:k.default,datum:A.default,on:D.default,dispatch:I.default},e.default=N},"./node_modules/d3-selection/src/selection/insert.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/creator.js"),i=n("./node_modules/d3-selection/src/selector.js");function o(){return null}e.default=function(t,e){var n="function"==typeof t?t:Object(r.default)(t),a=null==e?o:"function"==typeof e?e:Object(i.default)(e);return this.select(function(){return this.insertBefore(n.apply(this,arguments),a.apply(this,arguments)||null)})}},"./node_modules/d3-selection/src/selection/join.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e,n){var r=this.enter(),i=this,o=this.exit();return r="function"==typeof t?t(r):r.append(t+""),null!=e&&(i=e(i)),null==n?o.remove():n(o),r&&i?r.merge(i).order():i}},"./node_modules/d3-selection/src/selection/lower.js":function(t,e,n){"use strict";function r(){this.previousSibling&&this.parentNode.insertBefore(this,this.parentNode.firstChild)}n.r(e),e.default=function(){return this.each(r)}},"./node_modules/d3-selection/src/selection/merge.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/selection/index.js");e.default=function(t){for(var e=this._groups,n=t._groups,i=e.length,o=n.length,a=Math.min(i,o),s=new Array(i),u=0;u<a;++u)for(var c,l=e[u],d=n[u],p=l.length,f=s[u]=new Array(p),h=0;h<p;++h)(c=l[h]||d[h])&&(f[h]=c);for(;u<i;++u)s[u]=e[u];return new r.Selection(s,this._parents)}},"./node_modules/d3-selection/src/selection/node.js":function(t,e,n){"use strict";n.r(e),e.default=function(){for(var t=this._groups,e=0,n=t.length;e<n;++e)for(var r=t[e],i=0,o=r.length;i<o;++i){var a=r[i];if(a)return a}return null}},"./node_modules/d3-selection/src/selection/nodes.js":function(t,e,n){"use strict";n.r(e),e.default=function(){var t=new Array(this.size()),e=-1;return this.each(function(){t[++e]=this}),t}},"./node_modules/d3-selection/src/selection/on.js":function(t,e,n){"use strict";n.r(e),n.d(e,"event",function(){return i}),n.d(e,"customEvent",function(){return c});var r={},i=null;"undefined"!=typeof document&&("onmouseenter"in document.documentElement||(r={mouseenter:"mouseover",mouseleave:"mouseout"}));function o(t,e,n){return t=a(t,e,n),function(e){var n=e.relatedTarget;n&&(n===this||8&n.compareDocumentPosition(this))||t.call(this,e)}}function a(t,e,n){return function(r){var o=i;i=r;try{t.call(this,this.__data__,e,n)}finally{i=o}}}function s(t){return function(){var e=this.__on;if(e){for(var n,r=0,i=-1,o=e.length;r<o;++r)n=e[r],t.type&&n.type!==t.type||n.name!==t.name?e[++i]=n:this.removeEventListener(n.type,n.listener,n.capture);++i?e.length=i:delete this.__on}}}function u(t,e,n){var i=r.hasOwnProperty(t.type)?o:a;return function(r,o,a){var s,u=this.__on,c=i(e,o,a);if(u)for(var l=0,d=u.length;l<d;++l)if((s=u[l]).type===t.type&&s.name===t.name)return this.removeEventListener(s.type,s.listener,s.capture),this.addEventListener(s.type,s.listener=c,s.capture=n),void(s.value=e);this.addEventListener(t.type,c,n),s={type:t.type,name:t.name,value:e,listener:c,capture:n},u?u.push(s):this.__on=[s]}}function c(t,e,n,r){var o=i;t.sourceEvent=i,i=t;try{return e.apply(n,r)}finally{i=o}}e.default=function(t,e,n){var r,i,o=(t+"").trim().split(/^|\s+/).map(function(t){var e="",n=t.indexOf(".");return n>=0&&(e=t.slice(n+1),t=t.slice(0,n)),{type:t,name:e}}),a=o.length;if(!(arguments.length<2)){for(c=e?u:s,null==n&&(n=!1),r=0;r<a;++r)this.each(c(o[r],e,n));return this}var c=this.node().__on;if(c)for(var l,d=0,p=c.length;d<p;++d)for(r=0,l=c[d];r<a;++r)if((i=o[r]).type===l.type&&i.name===l.name)return l.value}},"./node_modules/d3-selection/src/selection/order.js":function(t,e,n){"use strict";n.r(e),e.default=function(){for(var t=this._groups,e=-1,n=t.length;++e<n;)for(var r,i=t[e],o=i.length-1,a=i[o];--o>=0;)(r=i[o])&&(a&&4^r.compareDocumentPosition(a)&&a.parentNode.insertBefore(r,a),a=r);return this}},"./node_modules/d3-selection/src/selection/property.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){return arguments.length>1?this.each((null==e?function(t){return function(){delete this[t]}}:"function"==typeof e?function(t,e){return function(){var n=e.apply(this,arguments);null==n?delete this[t]:this[t]=n}}:function(t,e){return function(){this[t]=e}})(t,e)):this.node()[t]}},"./node_modules/d3-selection/src/selection/raise.js":function(t,e,n){"use strict";function r(){this.nextSibling&&this.parentNode.appendChild(this)}n.r(e),e.default=function(){return this.each(r)}},"./node_modules/d3-selection/src/selection/remove.js":function(t,e,n){"use strict";function r(){var t=this.parentNode;t&&t.removeChild(this)}n.r(e),e.default=function(){return this.each(r)}},"./node_modules/d3-selection/src/selection/select.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/selection/index.js"),i=n("./node_modules/d3-selection/src/selector.js");e.default=function(t){"function"!=typeof t&&(t=Object(i.default)(t));for(var e=this._groups,n=e.length,o=new Array(n),a=0;a<n;++a)for(var s,u,c=e[a],l=c.length,d=o[a]=new Array(l),p=0;p<l;++p)(s=c[p])&&(u=t.call(s,s.__data__,p,c))&&("__data__"in s&&(u.__data__=s.__data__),d[p]=u);return new r.Selection(o,this._parents)}},"./node_modules/d3-selection/src/selection/selectAll.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/selection/index.js"),i=n("./node_modules/d3-selection/src/selectorAll.js");e.default=function(t){"function"!=typeof t&&(t=Object(i.default)(t));for(var e=this._groups,n=e.length,o=[],a=[],s=0;s<n;++s)for(var u,c=e[s],l=c.length,d=0;d<l;++d)(u=c[d])&&(o.push(t.call(u,u.__data__,d,c)),a.push(u));return new r.Selection(o,a)}},"./node_modules/d3-selection/src/selection/size.js":function(t,e,n){"use strict";n.r(e),e.default=function(){var t=0;return this.each(function(){++t}),t}},"./node_modules/d3-selection/src/selection/sort.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/selection/index.js");function i(t,e){return t<e?-1:t>e?1:t>=e?0:NaN}e.default=function(t){function e(e,n){return e&&n?t(e.__data__,n.__data__):!e-!n}t||(t=i);for(var n=this._groups,o=n.length,a=new Array(o),s=0;s<o;++s){for(var u,c=n[s],l=c.length,d=a[s]=new Array(l),p=0;p<l;++p)(u=c[p])&&(d[p]=u);d.sort(e)}return new r.Selection(a,this._parents).order()}},"./node_modules/d3-selection/src/selection/sparse.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return new Array(t.length)}},"./node_modules/d3-selection/src/selection/style.js":function(t,e,n){"use strict";n.r(e),n.d(e,"styleValue",function(){return i});var r=n("./node_modules/d3-selection/src/window.js");function i(t,e){return t.style.getPropertyValue(e)||Object(r.default)(t).getComputedStyle(t,null).getPropertyValue(e)}e.default=function(t,e,n){return arguments.length>1?this.each((null==e?function(t){return function(){this.style.removeProperty(t)}}:"function"==typeof e?function(t,e,n){return function(){var r=e.apply(this,arguments);null==r?this.style.removeProperty(t):this.style.setProperty(t,r,n)}}:function(t,e,n){return function(){this.style.setProperty(t,e,n)}})(t,e,null==n?"":n)):i(this.node(),t)}},"./node_modules/d3-selection/src/selection/text.js":function(t,e,n){"use strict";function r(){this.textContent=""}n.r(e),e.default=function(t){return arguments.length?this.each(null==t?r:("function"==typeof t?function(t){return function(){var e=t.apply(this,arguments);this.textContent=null==e?"":e}}:function(t){return function(){this.textContent=t}})(t)):this.node().textContent}},"./node_modules/d3-selection/src/selector.js":function(t,e,n){"use strict";function r(){}n.r(e),e.default=function(t){return null==t?r:function(){return this.querySelector(t)}}},"./node_modules/d3-selection/src/selectorAll.js":function(t,e,n){"use strict";function r(){return[]}n.r(e),e.default=function(t){return null==t?r:function(){return this.querySelectorAll(t)}}},"./node_modules/d3-selection/src/sourceEvent.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/selection/on.js");e.default=function(){for(var t,e=r.event;t=e.sourceEvent;)e=t;return e}},"./node_modules/d3-selection/src/touch.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/sourceEvent.js"),i=n("./node_modules/d3-selection/src/point.js");e.default=function(t,e,n){arguments.length<3&&(n=e,e=Object(r.default)().changedTouches);for(var o,a=0,s=e?e.length:0;a<s;++a)if((o=e[a]).identifier===n)return Object(i.default)(t,o);return null}},"./node_modules/d3-selection/src/touches.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/sourceEvent.js"),i=n("./node_modules/d3-selection/src/point.js");e.default=function(t,e){null==e&&(e=Object(r.default)().touches);for(var n=0,o=e?e.length:0,a=new Array(o);n<o;++n)a[n]=Object(i.default)(t,e[n]);return a}},"./node_modules/d3-selection/src/window.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return t.ownerDocument&&t.ownerDocument.defaultView||t.document&&t||t.defaultView}},"./node_modules/d3-timer/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-timer/src/timer.js");n.d(e,"now",function(){return r.now}),n.d(e,"timer",function(){return r.timer}),n.d(e,"timerFlush",function(){return r.timerFlush});var i=n("./node_modules/d3-timer/src/timeout.js");n.d(e,"timeout",function(){return i.default});var o=n("./node_modules/d3-timer/src/interval.js");n.d(e,"interval",function(){return o.default})},"./node_modules/d3-timer/src/interval.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-timer/src/timer.js");e.default=function(t,e,n){var i=new r.Timer,o=e;return null==e?(i.restart(t,e,n),i):(e=+e,n=null==n?Object(r.now)():+n,i.restart(function r(a){a+=o,i.restart(r,o+=e,n),t(a)},e,n),i)}},"./node_modules/d3-timer/src/timeout.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-timer/src/timer.js");e.default=function(t,e,n){var i=new r.Timer;return e=null==e?0:+e,i.restart(function(n){i.stop(),t(n+e)},e,n),i}},"./node_modules/d3-timer/src/timer.js":function(t,e,n){"use strict";n.r(e),n.d(e,"now",function(){return h}),n.d(e,"Timer",function(){return g}),n.d(e,"timer",function(){return y}),n.d(e,"timerFlush",function(){return _});var r,i,o=0,a=0,s=0,u=1e3,c=0,l=0,d=0,p="object"==typeof performance&&performance.now?performance:Date,f="object"==typeof window&&window.requestAnimationFrame?window.requestAnimationFrame.bind(window):function(t){setTimeout(t,17)};function h(){return l||(f(m),l=p.now()+d)}function m(){l=0}function g(){this._call=this._time=this._next=null}function y(t,e,n){var r=new g;return r.restart(t,e,n),r}function _(){h(),++o;for(var t,e=r;e;)(t=l-e._time)>=0&&e._call.call(null,t),e=e._next;--o}function v(){l=(c=p.now())+d,o=a=0;try{_()}finally{o=0,function(){var t,e,n=r,o=1/0;for(;n;)n._call?(o>n._time&&(o=n._time),t=n,n=n._next):(e=n._next,n._next=null,n=t?t._next=e:r=e);i=t,x(o)}(),l=0}}function b(){var t=p.now(),e=t-c;e>u&&(d-=e,c=t)}function x(t){o||(a&&(a=clearTimeout(a)),t-l>24?(t<1/0&&(a=setTimeout(v,t-p.now()-d)),s&&(s=clearInterval(s))):(s||(c=p.now(),s=setInterval(b,u)),o=1,f(v)))}g.prototype=y.prototype={constructor:g,restart:function(t,e,n){if("function"!=typeof t)throw new TypeError("callback is not a function");n=(null==n?h():+n)+(null==e?0:+e),this._next||i===this||(i?i._next=this:r=this,i=this),this._call=t,this._time=n,x()},stop:function(){this._call&&(this._call=null,this._time=1/0,x())}}},"./node_modules/d3-transition/src/active.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-transition/src/transition/index.js"),i=n("./node_modules/d3-transition/src/transition/schedule.js"),o=[null];e.default=function(t,e){var n,a,s=t.__transition;if(s){e=null==e?null:e+"";for(a in s)if((n=s[a]).state>i.SCHEDULED&&n.name===e)return new r.Transition([[t]],o,e,+a)}return null}},"./node_modules/d3-transition/src/index.js":function(t,e,n){"use strict";n.r(e);n("./node_modules/d3-transition/src/selection/index.js");var r=n("./node_modules/d3-transition/src/transition/index.js");n.d(e,"transition",function(){return r.default});var i=n("./node_modules/d3-transition/src/active.js");n.d(e,"active",function(){return i.default});var o=n("./node_modules/d3-transition/src/interrupt.js");n.d(e,"interrupt",function(){return o.default})},"./node_modules/d3-transition/src/interrupt.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-transition/src/transition/schedule.js");e.default=function(t,e){var n,i,o,a=t.__transition,s=!0;if(a){e=null==e?null:e+"";for(o in a)(n=a[o]).name===e?(i=n.state>r.STARTING&&n.state<r.ENDING,n.state=r.ENDED,n.timer.stop(),n.on.call(i?"interrupt":"cancel",t,t.__data__,n.index,n.group),delete a[o]):s=!1;s&&delete t.__transition}}},"./node_modules/d3-transition/src/selection/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/index.js"),i=n("./node_modules/d3-transition/src/selection/interrupt.js"),o=n("./node_modules/d3-transition/src/selection/transition.js");r.selection.prototype.interrupt=i.default,r.selection.prototype.transition=o.default},"./node_modules/d3-transition/src/selection/interrupt.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-transition/src/interrupt.js");e.default=function(t){return this.each(function(){Object(r.default)(this,t)})}},"./node_modules/d3-transition/src/selection/transition.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-transition/src/transition/index.js"),i=n("./node_modules/d3-transition/src/transition/schedule.js"),o=n("./node_modules/d3-ease/src/index.js"),a=n("./node_modules/d3-timer/index.js"),s={time:null,delay:0,duration:250,ease:o.easeCubicInOut};function u(t,e){for(var n;!(n=t.__transition)||!(n=n[e]);)if(!(t=t.parentNode))return s.time=Object(a.now)(),s;return n}e.default=function(t){var e,n;t instanceof r.Transition?(e=t._id,t=t._name):(e=Object(r.newId)(),(n=s).time=Object(a.now)(),t=null==t?null:t+"");for(var o=this._groups,c=o.length,l=0;l<c;++l)for(var d,p=o[l],f=p.length,h=0;h<f;++h)(d=p[h])&&Object(i.default)(d,t,e,h,p,n||u(d,e));return new r.Transition(o,this._parents,t,e)}},"./node_modules/d3-transition/src/transition/attr.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-interpolate/index.js"),i=n("./node_modules/d3-selection/src/index.js"),o=n("./node_modules/d3-transition/src/transition/tween.js"),a=n("./node_modules/d3-transition/src/transition/interpolate.js");e.default=function(t,e){var n=Object(i.namespace)(t),s="transform"===n?r.interpolateTransformSvg:a.default;return this.attrTween(t,"function"==typeof e?(n.local?function(t,e,n){var r,i,o;return function(){var a,s,u=n(this);if(null!=u)return(a=this.getAttributeNS(t.space,t.local))===(s=u+"")?null:a===r&&s===i?o:(i=s,o=e(r=a,u));this.removeAttributeNS(t.space,t.local)}}:function(t,e,n){var r,i,o;return function(){var a,s,u=n(this);if(null!=u)return(a=this.getAttribute(t))===(s=u+"")?null:a===r&&s===i?o:(i=s,o=e(r=a,u));this.removeAttribute(t)}})(n,s,Object(o.tweenValue)(this,"attr."+t,e)):null==e?(n.local?function(t){return function(){this.removeAttributeNS(t.space,t.local)}}:function(t){return function(){this.removeAttribute(t)}})(n):(n.local?function(t,e,n){var r,i,o=n+"";return function(){var a=this.getAttributeNS(t.space,t.local);return a===o?null:a===r?i:i=e(r=a,n)}}:function(t,e,n){var r,i,o=n+"";return function(){var a=this.getAttribute(t);return a===o?null:a===r?i:i=e(r=a,n)}})(n,s,e))}},"./node_modules/d3-transition/src/transition/attrTween.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/index.js");e.default=function(t,e){var n="attr."+t;if(arguments.length<2)return(n=this.tween(n))&&n._value;if(null==e)return this.tween(n,null);if("function"!=typeof e)throw new Error;var i=Object(r.namespace)(t);return this.tween(n,(i.local?function(t,e){var n,r;function i(){var i,o,a=e.apply(this,arguments);return a!==r&&(n=(r=a)&&(i=t,o=a,function(t){this.setAttributeNS(i.space,i.local,o(t))})),n}return i._value=e,i}:function(t,e){var n,r;function i(){var i,o,a=e.apply(this,arguments);return a!==r&&(n=(r=a)&&(i=t,o=a,function(t){this.setAttribute(i,o(t))})),n}return i._value=e,i})(i,e))}},"./node_modules/d3-transition/src/transition/delay.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-transition/src/transition/schedule.js");e.default=function(t){var e=this._id;return arguments.length?this.each(("function"==typeof t?function(t,e){return function(){Object(r.init)(this,t).delay=+e.apply(this,arguments)}}:function(t,e){return e=+e,function(){Object(r.init)(this,t).delay=e}})(e,t)):Object(r.get)(this.node(),e).delay}},"./node_modules/d3-transition/src/transition/duration.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-transition/src/transition/schedule.js");e.default=function(t){var e=this._id;return arguments.length?this.each(("function"==typeof t?function(t,e){return function(){Object(r.set)(this,t).duration=+e.apply(this,arguments)}}:function(t,e){return e=+e,function(){Object(r.set)(this,t).duration=e}})(e,t)):Object(r.get)(this.node(),e).duration}},"./node_modules/d3-transition/src/transition/ease.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-transition/src/transition/schedule.js");e.default=function(t){var e=this._id;return arguments.length?this.each(function(t,e){if("function"!=typeof e)throw new Error;return function(){Object(r.set)(this,t).ease=e}}(e,t)):Object(r.get)(this.node(),e).ease}},"./node_modules/d3-transition/src/transition/end.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-transition/src/transition/schedule.js");e.default=function(){var t,e,n=this,i=n._id,o=n.size();return new Promise(function(a,s){var u={value:s},c={value:function(){0==--o&&a()}};n.each(function(){var n=Object(r.set)(this,i),o=n.on;o!==t&&((e=(t=o).copy())._.cancel.push(u),e._.interrupt.push(u),e._.end.push(c)),n.on=e})})}},"./node_modules/d3-transition/src/transition/filter.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/index.js"),i=n("./node_modules/d3-transition/src/transition/index.js");e.default=function(t){"function"!=typeof t&&(t=Object(r.matcher)(t));for(var e=this._groups,n=e.length,o=new Array(n),a=0;a<n;++a)for(var s,u=e[a],c=u.length,l=o[a]=[],d=0;d<c;++d)(s=u[d])&&t.call(s,s.__data__,d,u)&&l.push(s);return new i.Transition(o,this._parents,this._name,this._id)}},"./node_modules/d3-transition/src/transition/index.js":function(t,e,n){"use strict";n.r(e),n.d(e,"Transition",function(){return j}),n.d(e,"default",function(){return O}),n.d(e,"newId",function(){return C});var r=n("./node_modules/d3-selection/src/index.js"),i=n("./node_modules/d3-transition/src/transition/attr.js"),o=n("./node_modules/d3-transition/src/transition/attrTween.js"),a=n("./node_modules/d3-transition/src/transition/delay.js"),s=n("./node_modules/d3-transition/src/transition/duration.js"),u=n("./node_modules/d3-transition/src/transition/ease.js"),c=n("./node_modules/d3-transition/src/transition/filter.js"),l=n("./node_modules/d3-transition/src/transition/merge.js"),d=n("./node_modules/d3-transition/src/transition/on.js"),p=n("./node_modules/d3-transition/src/transition/remove.js"),f=n("./node_modules/d3-transition/src/transition/select.js"),h=n("./node_modules/d3-transition/src/transition/selectAll.js"),m=n("./node_modules/d3-transition/src/transition/selection.js"),g=n("./node_modules/d3-transition/src/transition/style.js"),y=n("./node_modules/d3-transition/src/transition/styleTween.js"),_=n("./node_modules/d3-transition/src/transition/text.js"),v=n("./node_modules/d3-transition/src/transition/transition.js"),b=n("./node_modules/d3-transition/src/transition/tween.js"),x=n("./node_modules/d3-transition/src/transition/end.js"),w=0;function j(t,e,n,r){this._groups=t,this._parents=e,this._name=n,this._id=r}function O(t){return Object(r.selection)().transition(t)}function C(){return++w}var E=r.selection.prototype;j.prototype=O.prototype={constructor:j,select:f.default,selectAll:h.default,filter:c.default,merge:l.default,selection:m.default,transition:v.default,call:E.call,nodes:E.nodes,node:E.node,size:E.size,empty:E.empty,each:E.each,on:d.default,attr:i.default,attrTween:o.default,style:g.default,styleTween:y.default,text:_.default,remove:p.default,tween:b.default,delay:a.default,duration:s.default,ease:u.default,end:x.default}},"./node_modules/d3-transition/src/transition/interpolate.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-color/index.js"),i=n("./node_modules/d3-interpolate/index.js");e.default=function(t,e){var n;return("number"==typeof e?i.interpolateNumber:e instanceof r.color?i.interpolateRgb:(n=Object(r.color)(e))?(e=n,i.interpolateRgb):i.interpolateString)(t,e)}},"./node_modules/d3-transition/src/transition/merge.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-transition/src/transition/index.js");e.default=function(t){if(t._id!==this._id)throw new Error;for(var e=this._groups,n=t._groups,i=e.length,o=n.length,a=Math.min(i,o),s=new Array(i),u=0;u<a;++u)for(var c,l=e[u],d=n[u],p=l.length,f=s[u]=new Array(p),h=0;h<p;++h)(c=l[h]||d[h])&&(f[h]=c);for(;u<i;++u)s[u]=e[u];return new r.Transition(s,this._parents,this._name,this._id)}},"./node_modules/d3-transition/src/transition/on.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-transition/src/transition/schedule.js");e.default=function(t,e){var n,i,o,a,s,u,c=this._id;return arguments.length<2?Object(r.get)(this.node(),c).on.on(t):this.each((n=c,o=e,u=((i=t)+"").trim().split(/^|\s+/).every(function(t){var e=t.indexOf(".");return e>=0&&(t=t.slice(0,e)),!t||"start"===t})?r.init:r.set,function(){var t=u(this,n),e=t.on;e!==a&&(s=(a=e).copy()).on(i,o),t.on=s}))}},"./node_modules/d3-transition/src/transition/remove.js":function(t,e,n){"use strict";n.r(e),e.default=function(){return this.on("end.remove",(t=this._id,function(){var e=this.parentNode;for(var n in this.__transition)if(+n!==t)return;e&&e.removeChild(this)}));var t}},"./node_modules/d3-transition/src/transition/schedule.js":function(t,e,n){"use strict";n.r(e),n.d(e,"CREATED",function(){return s}),n.d(e,"SCHEDULED",function(){return u}),n.d(e,"STARTING",function(){return c}),n.d(e,"STARTED",function(){return l}),n.d(e,"RUNNING",function(){return d}),n.d(e,"ENDING",function(){return p}),n.d(e,"ENDED",function(){return f}),n.d(e,"init",function(){return h}),n.d(e,"set",function(){return m}),n.d(e,"get",function(){return g});var r=n("./node_modules/d3-dispatch/index.js"),i=n("./node_modules/d3-timer/index.js"),o=Object(r.dispatch)("start","end","cancel","interrupt"),a=[],s=0,u=1,c=2,l=3,d=4,p=5,f=6;function h(t,e){var n=g(t,e);if(n.state>s)throw new Error("too late; already scheduled");return n}function m(t,e){var n=g(t,e);if(n.state>l)throw new Error("too late; already running");return n}function g(t,e){var n=t.__transition;if(!n||!(n=n[e]))throw new Error("transition not found");return n}e.default=function(t,e,n,r,h,m){var g=t.__transition;if(g){if(n in g)return}else t.__transition={};!function(t,e,n){var r,o=t.__transition;function a(p){var m,g,y,_;if(n.state!==u)return h();for(m in o)if((_=o[m]).name===n.name){if(_.state===l)return Object(i.timeout)(a);_.state===d?(_.state=f,_.timer.stop(),_.on.call("interrupt",t,t.__data__,_.index,_.group),delete o[m]):+m<e&&(_.state=f,_.timer.stop(),_.on.call("cancel",t,t.__data__,_.index,_.group),delete o[m])}if(Object(i.timeout)(function(){n.state===l&&(n.state=d,n.timer.restart(s,n.delay,n.time),s(p))}),n.state=c,n.on.call("start",t,t.__data__,n.index,n.group),n.state===c){for(n.state=l,r=new Array(y=n.tween.length),m=0,g=-1;m<y;++m)(_=n.tween[m].value.call(t,t.__data__,n.index,n.group))&&(r[++g]=_);r.length=g+1}}function s(e){for(var i=e<n.duration?n.ease.call(null,e/n.duration):(n.timer.restart(h),n.state=p,1),o=-1,a=r.length;++o<a;)r[o].call(t,i);n.state===p&&(n.on.call("end",t,t.__data__,n.index,n.group),h())}function h(){n.state=f,n.timer.stop(),delete o[e];for(var r in o)return;delete t.__transition}o[e]=n,n.timer=Object(i.timer)(function(t){n.state=u,n.timer.restart(a,n.delay,n.time),n.delay<=t&&a(t-n.delay)},0,n.time)}(t,n,{name:e,index:r,group:h,on:o,tween:a,time:m.time,delay:m.delay,duration:m.duration,ease:m.ease,timer:null,state:s})}},"./node_modules/d3-transition/src/transition/select.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/index.js"),i=n("./node_modules/d3-transition/src/transition/index.js"),o=n("./node_modules/d3-transition/src/transition/schedule.js");e.default=function(t){var e=this._name,n=this._id;"function"!=typeof t&&(t=Object(r.selector)(t));for(var a=this._groups,s=a.length,u=new Array(s),c=0;c<s;++c)for(var l,d,p=a[c],f=p.length,h=u[c]=new Array(f),m=0;m<f;++m)(l=p[m])&&(d=t.call(l,l.__data__,m,p))&&("__data__"in l&&(d.__data__=l.__data__),h[m]=d,Object(o.default)(h[m],e,n,m,h,Object(o.get)(l,n)));return new i.Transition(u,this._parents,e,n)}},"./node_modules/d3-transition/src/transition/selectAll.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/index.js"),i=n("./node_modules/d3-transition/src/transition/index.js"),o=n("./node_modules/d3-transition/src/transition/schedule.js");e.default=function(t){var e=this._name,n=this._id;"function"!=typeof t&&(t=Object(r.selectorAll)(t));for(var a=this._groups,s=a.length,u=[],c=[],l=0;l<s;++l)for(var d,p=a[l],f=p.length,h=0;h<f;++h)if(d=p[h]){for(var m,g=t.call(d,d.__data__,h,p),y=Object(o.get)(d,n),_=0,v=g.length;_<v;++_)(m=g[_])&&Object(o.default)(m,e,n,_,g,y);u.push(g),c.push(d)}return new i.Transition(u,c,e,n)}},"./node_modules/d3-transition/src/transition/selection.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-selection/src/index.js").selection.prototype.constructor;e.default=function(){return new r(this._groups,this._parents)}},"./node_modules/d3-transition/src/transition/style.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-interpolate/index.js"),i=n("./node_modules/d3-selection/src/index.js"),o=n("./node_modules/d3-transition/src/transition/schedule.js"),a=n("./node_modules/d3-transition/src/transition/tween.js"),s=n("./node_modules/d3-transition/src/transition/interpolate.js");function u(t){return function(){this.style.removeProperty(t)}}e.default=function(t,e,n){var c,l,d,p,f,h,m,g,y,_,v,b,x,w,j,O,C,E,S,M,T,k,A,D,I,P="transform"==(t+="")?r.interpolateTransformCss:s.default;return null==e?this.styleTween(t,(T=t,k=P,function(){var t=Object(i.style)(this,T),e=(this.style.removeProperty(T),Object(i.style)(this,T));return t===e?null:t===A&&e===D?I:I=k(A=t,D=e)})).on("end.style."+t,u(t)):"function"==typeof e?this.styleTween(t,(j=t,O=P,C=Object(a.tweenValue)(this,"style."+t,e),function(){var t=Object(i.style)(this,j),e=C(this),n=e+"";return null==e&&(this.style.removeProperty(j),n=e=Object(i.style)(this,j)),t===n?null:t===E&&n===S?M:(S=n,M=O(E=t,e))})).each((m=this._id,w="end."+(x="style."+(g=t)),function(){var t=Object(o.set)(this,m),e=t.on,n=null==t.value[x]?b||(b=u(g)):void 0;e===y&&v===n||(_=(y=e).copy()).on(w,v=n),t.on=_})):this.styleTween(t,(c=t,l=P,d=e,h=d+"",function(){var t=Object(i.style)(this,c);return t===h?null:t===p?f:f=l(p=t,d)}),n).on("end.style."+t,null)}},"./node_modules/d3-transition/src/transition/styleTween.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e,n){var r="style."+(t+="");if(arguments.length<2)return(r=this.tween(r))&&r._value;if(null==e)return this.tween(r,null);if("function"!=typeof e)throw new Error;return this.tween(r,function(t,e,n){var r,i;function o(){var o,a,s,u=e.apply(this,arguments);return u!==i&&(r=(i=u)&&(o=t,a=u,s=n,function(t){this.style.setProperty(o,a(t),s)})),r}return o._value=e,o}(t,e,null==n?"":n))}},"./node_modules/d3-transition/src/transition/text.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-transition/src/transition/tween.js");e.default=function(t){return this.tween("text","function"==typeof t?(n=Object(r.tweenValue)(this,"text",t),function(){var t=n(this);this.textContent=null==t?"":t}):(e=null==t?"":t+"",function(){this.textContent=e}));var e,n}},"./node_modules/d3-transition/src/transition/transition.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-transition/src/transition/index.js"),i=n("./node_modules/d3-transition/src/transition/schedule.js");e.default=function(){for(var t=this._name,e=this._id,n=Object(r.newId)(),o=this._groups,a=o.length,s=0;s<a;++s)for(var u,c=o[s],l=c.length,d=0;d<l;++d)if(u=c[d]){var p=Object(i.get)(u,e);Object(i.default)(u,t,n,d,c,{time:p.time+p.delay+p.duration,delay:0,duration:p.duration,ease:p.ease})}return new r.Transition(o,this._parents,t,n)}},"./node_modules/d3-transition/src/transition/tween.js":function(t,e,n){"use strict";n.r(e),n.d(e,"tweenValue",function(){return i});var r=n("./node_modules/d3-transition/src/transition/schedule.js");function i(t,e,n){var i=t._id;return t.each(function(){var t=Object(r.set)(this,i);(t.value||(t.value={}))[e]=n.apply(this,arguments)}),function(t){return Object(r.get)(t,i).value[e]}}e.default=function(t,e){var n=this._id;if(t+="",arguments.length<2){for(var i,o=Object(r.get)(this.node(),n).tween,a=0,s=o.length;a<s;++a)if((i=o[a]).name===t)return i.value;return null}return this.each((null==e?function(t,e){var n,i;return function(){var o=Object(r.set)(this,t),a=o.tween;if(a!==n)for(var s=0,u=(i=n=a).length;s<u;++s)if(i[s].name===e){(i=i.slice()).splice(s,1);break}o.tween=i}}:function(t,e,n){var i,o;if("function"!=typeof n)throw new Error;return function(){var a=Object(r.set)(this,t),s=a.tween;if(s!==i){o=(i=s).slice();for(var u={name:e,value:n},c=0,l=o.length;c<l;++c)if(o[c].name===e){o[c]=u;break}c===l&&o.push(u)}a.tween=o}})(n,t,e))}},"./node_modules/rxjs/_esm5/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/rxjs/_esm5/internal/Observable.js");n.d(e,"Observable",function(){return r.Observable});var i=n("./node_modules/rxjs/_esm5/internal/observable/ConnectableObservable.js");n.d(e,"ConnectableObservable",function(){return i.ConnectableObservable});var o=n("./node_modules/rxjs/_esm5/internal/operators/groupBy.js");n.d(e,"GroupedObservable",function(){return o.GroupedObservable});var a=n("./node_modules/rxjs/_esm5/internal/symbol/observable.js");n.d(e,"observable",function(){return a.observable});var s=n("./node_modules/rxjs/_esm5/internal/Subject.js");n.d(e,"Subject",function(){return s.Subject});var u=n("./node_modules/rxjs/_esm5/internal/BehaviorSubject.js");n.d(e,"BehaviorSubject",function(){return u.BehaviorSubject});var c=n("./node_modules/rxjs/_esm5/internal/ReplaySubject.js");n.d(e,"ReplaySubject",function(){return c.ReplaySubject});var l=n("./node_modules/rxjs/_esm5/internal/AsyncSubject.js");n.d(e,"AsyncSubject",function(){return l.AsyncSubject});var d=n("./node_modules/rxjs/_esm5/internal/scheduler/asap.js");n.d(e,"asapScheduler",function(){return d.asap});var p=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js");n.d(e,"asyncScheduler",function(){return p.async});var f=n("./node_modules/rxjs/_esm5/internal/scheduler/queue.js");n.d(e,"queueScheduler",function(){return f.queue});var h=n("./node_modules/rxjs/_esm5/internal/scheduler/animationFrame.js");n.d(e,"animationFrameScheduler",function(){return h.animationFrame});var m=n("./node_modules/rxjs/_esm5/internal/scheduler/VirtualTimeScheduler.js");n.d(e,"VirtualTimeScheduler",function(){return m.VirtualTimeScheduler}),n.d(e,"VirtualAction",function(){return m.VirtualAction});var g=n("./node_modules/rxjs/_esm5/internal/Scheduler.js");n.d(e,"Scheduler",function(){return g.Scheduler});var y=n("./node_modules/rxjs/_esm5/internal/Subscription.js");n.d(e,"Subscription",function(){return y.Subscription});var _=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");n.d(e,"Subscriber",function(){return _.Subscriber});var v=n("./node_modules/rxjs/_esm5/internal/Notification.js");n.d(e,"Notification",function(){return v.Notification});var b=n("./node_modules/rxjs/_esm5/internal/util/pipe.js");n.d(e,"pipe",function(){return b.pipe});var x=n("./node_modules/rxjs/_esm5/internal/util/noop.js");n.d(e,"noop",function(){return x.noop});var w=n("./node_modules/rxjs/_esm5/internal/util/identity.js");n.d(e,"identity",function(){return w.identity});var j=n("./node_modules/rxjs/_esm5/internal/util/isObservable.js");n.d(e,"isObservable",function(){return j.isObservable});var O=n("./node_modules/rxjs/_esm5/internal/util/ArgumentOutOfRangeError.js");n.d(e,"ArgumentOutOfRangeError",function(){return O.ArgumentOutOfRangeError});var C=n("./node_modules/rxjs/_esm5/internal/util/EmptyError.js");n.d(e,"EmptyError",function(){return C.EmptyError});var E=n("./node_modules/rxjs/_esm5/internal/util/ObjectUnsubscribedError.js");n.d(e,"ObjectUnsubscribedError",function(){return E.ObjectUnsubscribedError});var S=n("./node_modules/rxjs/_esm5/internal/util/UnsubscriptionError.js");n.d(e,"UnsubscriptionError",function(){return S.UnsubscriptionError});var M=n("./node_modules/rxjs/_esm5/internal/util/TimeoutError.js");n.d(e,"TimeoutError",function(){return M.TimeoutError});var T=n("./node_modules/rxjs/_esm5/internal/observable/bindCallback.js");n.d(e,"bindCallback",function(){return T.bindCallback});var k=n("./node_modules/rxjs/_esm5/internal/observable/bindNodeCallback.js");n.d(e,"bindNodeCallback",function(){return k.bindNodeCallback});var A=n("./node_modules/rxjs/_esm5/internal/observable/combineLatest.js");n.d(e,"combineLatest",function(){return A.combineLatest});var D=n("./node_modules/rxjs/_esm5/internal/observable/concat.js");n.d(e,"concat",function(){return D.concat});var I=n("./node_modules/rxjs/_esm5/internal/observable/defer.js");n.d(e,"defer",function(){return I.defer});var P=n("./node_modules/rxjs/_esm5/internal/observable/empty.js");n.d(e,"empty",function(){return P.empty});var R=n("./node_modules/rxjs/_esm5/internal/observable/forkJoin.js");n.d(e,"forkJoin",function(){return R.forkJoin});var N=n("./node_modules/rxjs/_esm5/internal/observable/from.js");n.d(e,"from",function(){return N.from});var F=n("./node_modules/rxjs/_esm5/internal/observable/fromEvent.js");n.d(e,"fromEvent",function(){return F.fromEvent});var L=n("./node_modules/rxjs/_esm5/internal/observable/fromEventPattern.js");n.d(e,"fromEventPattern",function(){return L.fromEventPattern});var B=n("./node_modules/rxjs/_esm5/internal/observable/generate.js");n.d(e,"generate",function(){return B.generate});var V=n("./node_modules/rxjs/_esm5/internal/observable/iif.js");n.d(e,"iif",function(){return V.iif});var z=n("./node_modules/rxjs/_esm5/internal/observable/interval.js");n.d(e,"interval",function(){return z.interval});var H=n("./node_modules/rxjs/_esm5/internal/observable/merge.js");n.d(e,"merge",function(){return H.merge});var U=n("./node_modules/rxjs/_esm5/internal/observable/never.js");n.d(e,"never",function(){return U.never});var W=n("./node_modules/rxjs/_esm5/internal/observable/of.js");n.d(e,"of",function(){return W.of});var q=n("./node_modules/rxjs/_esm5/internal/observable/onErrorResumeNext.js");n.d(e,"onErrorResumeNext",function(){return q.onErrorResumeNext});var G=n("./node_modules/rxjs/_esm5/internal/observable/pairs.js");n.d(e,"pairs",function(){return G.pairs});var Y=n("./node_modules/rxjs/_esm5/internal/observable/race.js");n.d(e,"race",function(){return Y.race});var K=n("./node_modules/rxjs/_esm5/internal/observable/range.js");n.d(e,"range",function(){return K.range});var X=n("./node_modules/rxjs/_esm5/internal/observable/throwError.js");n.d(e,"throwError",function(){return X.throwError});var Q=n("./node_modules/rxjs/_esm5/internal/observable/timer.js");n.d(e,"timer",function(){return Q.timer});var Z=n("./node_modules/rxjs/_esm5/internal/observable/using.js");n.d(e,"using",function(){return Z.using});var $=n("./node_modules/rxjs/_esm5/internal/observable/zip.js");n.d(e,"zip",function(){return $.zip}),n.d(e,"EMPTY",function(){return P.EMPTY}),n.d(e,"NEVER",function(){return U.NEVER});var J=n("./node_modules/rxjs/_esm5/internal/config.js");n.d(e,"config",function(){return J.config})},"./node_modules/rxjs/_esm5/internal/AsyncSubject.js":function(t,e,n){"use strict";n.r(e),n.d(e,"AsyncSubject",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),a=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.value=null,e.hasNext=!1,e.hasCompleted=!1,e}return r.__extends(e,t),e.prototype._subscribe=function(e){return this.hasError?(e.error(this.thrownError),o.Subscription.EMPTY):this.hasCompleted&&this.hasNext?(e.next(this.value),e.complete(),o.Subscription.EMPTY):t.prototype._subscribe.call(this,e)},e.prototype.next=function(t){this.hasCompleted||(this.value=t,this.hasNext=!0)},e.prototype.error=function(e){this.hasCompleted||t.prototype.error.call(this,e)},e.prototype.complete=function(){this.hasCompleted=!0,this.hasNext&&t.prototype.next.call(this,this.value),t.prototype.complete.call(this)},e}(i.Subject)},"./node_modules/rxjs/_esm5/internal/BehaviorSubject.js":function(t,e,n){"use strict";n.r(e),n.d(e,"BehaviorSubject",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/util/ObjectUnsubscribedError.js"),a=function(t){function e(e){var n=t.call(this)||this;return n._value=e,n}return r.__extends(e,t),Object.defineProperty(e.prototype,"value",{get:function(){return this.getValue()},enumerable:!0,configurable:!0}),e.prototype._subscribe=function(e){var n=t.prototype._subscribe.call(this,e);return n&&!n.closed&&e.next(this._value),n},e.prototype.getValue=function(){if(this.hasError)throw this.thrownError;if(this.closed)throw new o.ObjectUnsubscribedError;return this._value},e.prototype.next=function(e){t.prototype.next.call(this,this._value=e)},e}(i.Subject)},"./node_modules/rxjs/_esm5/internal/InnerSubscriber.js":function(t,e,n){"use strict";n.r(e),n.d(e,"InnerSubscriber",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(t){function e(e,n,r){var i=t.call(this)||this;return i.parent=e,i.outerValue=n,i.outerIndex=r,i.index=0,i}return r.__extends(e,t),e.prototype._next=function(t){this.parent.notifyNext(this.outerValue,t,this.outerIndex,this.index++,this)},e.prototype._error=function(t){this.parent.notifyError(t,this),this.unsubscribe()},e.prototype._complete=function(){this.parent.notifyComplete(this),this.unsubscribe()},e}(n("./node_modules/rxjs/_esm5/internal/Subscriber.js").Subscriber)},"./node_modules/rxjs/_esm5/internal/Notification.js":function(t,e,n){"use strict";n.r(e),n.d(e,"Notification",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/observable/empty.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/of.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/throwError.js"),a=function(){function t(t,e,n){this.kind=t,this.value=e,this.error=n,this.hasValue="N"===t}return t.prototype.observe=function(t){switch(this.kind){case"N":return t.next&&t.next(this.value);case"E":return t.error&&t.error(this.error);case"C":return t.complete&&t.complete()}},t.prototype.do=function(t,e,n){switch(this.kind){case"N":return t&&t(this.value);case"E":return e&&e(this.error);case"C":return n&&n()}},t.prototype.accept=function(t,e,n){return t&&"function"==typeof t.next?this.observe(t):this.do(t,e,n)},t.prototype.toObservable=function(){switch(this.kind){case"N":return Object(i.of)(this.value);case"E":return Object(o.throwError)(this.error);case"C":return Object(r.empty)()}throw new Error("unexpected notification kind value")},t.createNext=function(e){return void 0!==e?new t("N",e):t.undefinedValueNotification},t.createError=function(e){return new t("E",void 0,e)},t.createComplete=function(){return t.completeNotification},t.completeNotification=new t("C"),t.undefinedValueNotification=new t("N",void 0),t}()},"./node_modules/rxjs/_esm5/internal/Observable.js":function(t,e,n){"use strict";n.r(e),n.d(e,"Observable",function(){return u});var r=n("./node_modules/rxjs/_esm5/internal/util/canReportError.js"),i=n("./node_modules/rxjs/_esm5/internal/util/toSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/symbol/observable.js"),a=n("./node_modules/rxjs/_esm5/internal/util/pipe.js"),s=n("./node_modules/rxjs/_esm5/internal/config.js"),u=function(){function t(t){this._isScalar=!1,t&&(this._subscribe=t)}return t.prototype.lift=function(e){var n=new t;return n.source=this,n.operator=e,n},t.prototype.subscribe=function(t,e,n){var r=this.operator,o=Object(i.toSubscriber)(t,e,n);if(r?r.call(o,this.source):o.add(this.source||s.config.useDeprecatedSynchronousErrorHandling&&!o.syncErrorThrowable?this._subscribe(o):this._trySubscribe(o)),s.config.useDeprecatedSynchronousErrorHandling&&o.syncErrorThrowable&&(o.syncErrorThrowable=!1,o.syncErrorThrown))throw o.syncErrorValue;return o},t.prototype._trySubscribe=function(t){try{return this._subscribe(t)}catch(e){s.config.useDeprecatedSynchronousErrorHandling&&(t.syncErrorThrown=!0,t.syncErrorValue=e),Object(r.canReportError)(t)?t.error(e):console.warn(e)}},t.prototype.forEach=function(t,e){var n=this;return new(e=c(e))(function(e,r){var i;i=n.subscribe(function(e){try{t(e)}catch(t){r(t),i&&i.unsubscribe()}},r,e)})},t.prototype._subscribe=function(t){var e=this.source;return e&&e.subscribe(t)},t.prototype[o.observable]=function(){return this},t.prototype.pipe=function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return 0===t.length?this:Object(a.pipeFromArray)(t)(this)},t.prototype.toPromise=function(t){var e=this;return new(t=c(t))(function(t,n){var r;e.subscribe(function(t){return r=t},function(t){return n(t)},function(){return t(r)})})},t.create=function(e){return new t(e)},t}();function c(t){if(t||(t=s.config.Promise||Promise),!t)throw new Error("no Promise impl found");return t}},"./node_modules/rxjs/_esm5/internal/Observer.js":function(t,e,n){"use strict";n.r(e),n.d(e,"empty",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/config.js"),i=n("./node_modules/rxjs/_esm5/internal/util/hostReportError.js"),o={closed:!0,next:function(t){},error:function(t){if(r.config.useDeprecatedSynchronousErrorHandling)throw t;Object(i.hostReportError)(t)},complete:function(){}}},"./node_modules/rxjs/_esm5/internal/OuterSubscriber.js":function(t,e,n){"use strict";n.r(e),n.d(e,"OuterSubscriber",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return r.__extends(e,t),e.prototype.notifyNext=function(t,e,n,r,i){this.destination.next(e)},e.prototype.notifyError=function(t,e){this.destination.error(t)},e.prototype.notifyComplete=function(t){this.destination.complete()},e}(n("./node_modules/rxjs/_esm5/internal/Subscriber.js").Subscriber)},"./node_modules/rxjs/_esm5/internal/ReplaySubject.js":function(t,e,n){"use strict";n.r(e),n.d(e,"ReplaySubject",function(){return l});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/scheduler/queue.js"),a=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),s=n("./node_modules/rxjs/_esm5/internal/operators/observeOn.js"),u=n("./node_modules/rxjs/_esm5/internal/util/ObjectUnsubscribedError.js"),c=n("./node_modules/rxjs/_esm5/internal/SubjectSubscription.js"),l=function(t){function e(e,n,r){void 0===e&&(e=Number.POSITIVE_INFINITY),void 0===n&&(n=Number.POSITIVE_INFINITY);var i=t.call(this)||this;return i.scheduler=r,i._events=[],i._infiniteTimeWindow=!1,i._bufferSize=e<1?1:e,i._windowTime=n<1?1:n,n===Number.POSITIVE_INFINITY?(i._infiniteTimeWindow=!0,i.next=i.nextInfiniteTimeWindow):i.next=i.nextTimeWindow,i}return r.__extends(e,t),e.prototype.nextInfiniteTimeWindow=function(e){var n=this._events;n.push(e),n.length>this._bufferSize&&n.shift(),t.prototype.next.call(this,e)},e.prototype.nextTimeWindow=function(e){this._events.push(new d(this._getNow(),e)),this._trimBufferThenGetEvents(),t.prototype.next.call(this,e)},e.prototype._subscribe=function(t){var e,n=this._infiniteTimeWindow,r=n?this._events:this._trimBufferThenGetEvents(),i=this.scheduler,o=r.length;if(this.closed)throw new u.ObjectUnsubscribedError;if(this.isStopped||this.hasError?e=a.Subscription.EMPTY:(this.observers.push(t),e=new c.SubjectSubscription(this,t)),i&&t.add(t=new s.ObserveOnSubscriber(t,i)),n)for(var l=0;l<o&&!t.closed;l++)t.next(r[l]);else for(l=0;l<o&&!t.closed;l++)t.next(r[l].value);return this.hasError?t.error(this.thrownError):this.isStopped&&t.complete(),e},e.prototype._getNow=function(){return(this.scheduler||o.queue).now()},e.prototype._trimBufferThenGetEvents=function(){for(var t=this._getNow(),e=this._bufferSize,n=this._windowTime,r=this._events,i=r.length,o=0;o<i&&!(t-r[o].time<n);)o++;return i>e&&(o=Math.max(o,i-e)),o>0&&r.splice(0,o),r},e}(i.Subject),d=function(){return function(t,e){this.time=t,this.value=e}}()},"./node_modules/rxjs/_esm5/internal/Scheduler.js":function(t,e,n){"use strict";n.r(e),n.d(e,"Scheduler",function(){return r});var r=function(){function t(e,n){void 0===n&&(n=t.now),this.SchedulerAction=e,this.now=n}return t.prototype.schedule=function(t,e,n){return void 0===e&&(e=0),new this.SchedulerAction(this,t).schedule(n,e)},t.now=function(){return Date.now()},t}()},"./node_modules/rxjs/_esm5/internal/Subject.js":function(t,e,n){"use strict";n.r(e),n.d(e,"SubjectSubscriber",function(){return l}),n.d(e,"Subject",function(){return d}),n.d(e,"AnonymousSubject",function(){return p});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Observable.js"),o=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),a=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),s=n("./node_modules/rxjs/_esm5/internal/util/ObjectUnsubscribedError.js"),u=n("./node_modules/rxjs/_esm5/internal/SubjectSubscription.js"),c=n("./node_modules/rxjs/_esm5/internal/symbol/rxSubscriber.js"),l=function(t){function e(e){var n=t.call(this,e)||this;return n.destination=e,n}return r.__extends(e,t),e}(o.Subscriber),d=function(t){function e(){var e=t.call(this)||this;return e.observers=[],e.closed=!1,e.isStopped=!1,e.hasError=!1,e.thrownError=null,e}return r.__extends(e,t),e.prototype[c.rxSubscriber]=function(){return new l(this)},e.prototype.lift=function(t){var e=new p(this,this);return e.operator=t,e},e.prototype.next=function(t){if(this.closed)throw new s.ObjectUnsubscribedError;if(!this.isStopped)for(var e=this.observers,n=e.length,r=e.slice(),i=0;i<n;i++)r[i].next(t)},e.prototype.error=function(t){if(this.closed)throw new s.ObjectUnsubscribedError;this.hasError=!0,this.thrownError=t,this.isStopped=!0;for(var e=this.observers,n=e.length,r=e.slice(),i=0;i<n;i++)r[i].error(t);this.observers.length=0},e.prototype.complete=function(){if(this.closed)throw new s.ObjectUnsubscribedError;this.isStopped=!0;for(var t=this.observers,e=t.length,n=t.slice(),r=0;r<e;r++)n[r].complete();this.observers.length=0},e.prototype.unsubscribe=function(){this.isStopped=!0,this.closed=!0,this.observers=null},e.prototype._trySubscribe=function(e){if(this.closed)throw new s.ObjectUnsubscribedError;return t.prototype._trySubscribe.call(this,e)},e.prototype._subscribe=function(t){if(this.closed)throw new s.ObjectUnsubscribedError;return this.hasError?(t.error(this.thrownError),a.Subscription.EMPTY):this.isStopped?(t.complete(),a.Subscription.EMPTY):(this.observers.push(t),new u.SubjectSubscription(this,t))},e.prototype.asObservable=function(){var t=new i.Observable;return t.source=this,t},e.create=function(t,e){return new p(t,e)},e}(i.Observable),p=function(t){function e(e,n){var r=t.call(this)||this;return r.destination=e,r.source=n,r}return r.__extends(e,t),e.prototype.next=function(t){var e=this.destination;e&&e.next&&e.next(t)},e.prototype.error=function(t){var e=this.destination;e&&e.error&&this.destination.error(t)},e.prototype.complete=function(){var t=this.destination;t&&t.complete&&this.destination.complete()},e.prototype._subscribe=function(t){return this.source?this.source.subscribe(t):a.Subscription.EMPTY},e}(d)},"./node_modules/rxjs/_esm5/internal/SubjectSubscription.js":function(t,e,n){"use strict";n.r(e),n.d(e,"SubjectSubscription",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(t){function e(e,n){var r=t.call(this)||this;return r.subject=e,r.subscriber=n,r.closed=!1,r}return r.__extends(e,t),e.prototype.unsubscribe=function(){if(!this.closed){this.closed=!0;var t=this.subject,e=t.observers;if(this.subject=null,e&&0!==e.length&&!t.isStopped&&!t.closed){var n=e.indexOf(this.subscriber);-1!==n&&e.splice(n,1)}}},e}(n("./node_modules/rxjs/_esm5/internal/Subscription.js").Subscription)},"./node_modules/rxjs/_esm5/internal/Subscriber.js":function(t,e,n){"use strict";n.r(e),n.d(e,"Subscriber",function(){return l}),n.d(e,"SafeSubscriber",function(){return d});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/util/isFunction.js"),o=n("./node_modules/rxjs/_esm5/internal/Observer.js"),a=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),s=n("./node_modules/rxjs/_esm5/internal/symbol/rxSubscriber.js"),u=n("./node_modules/rxjs/_esm5/internal/config.js"),c=n("./node_modules/rxjs/_esm5/internal/util/hostReportError.js"),l=function(t){function e(n,r,i){var a=t.call(this)||this;switch(a.syncErrorValue=null,a.syncErrorThrown=!1,a.syncErrorThrowable=!1,a.isStopped=!1,a._parentSubscription=null,arguments.length){case 0:a.destination=o.empty;break;case 1:if(!n){a.destination=o.empty;break}if("object"==typeof n){n instanceof e?(a.syncErrorThrowable=n.syncErrorThrowable,a.destination=n,n.add(a)):(a.syncErrorThrowable=!0,a.destination=new d(a,n));break}default:a.syncErrorThrowable=!0,a.destination=new d(a,n,r,i)}return a}return r.__extends(e,t),e.prototype[s.rxSubscriber]=function(){return this},e.create=function(t,n,r){var i=new e(t,n,r);return i.syncErrorThrowable=!1,i},e.prototype.next=function(t){this.isStopped||this._next(t)},e.prototype.error=function(t){this.isStopped||(this.isStopped=!0,this._error(t))},e.prototype.complete=function(){this.isStopped||(this.isStopped=!0,this._complete())},e.prototype.unsubscribe=function(){this.closed||(this.isStopped=!0,t.prototype.unsubscribe.call(this))},e.prototype._next=function(t){this.destination.next(t)},e.prototype._error=function(t){this.destination.error(t),this.unsubscribe()},e.prototype._complete=function(){this.destination.complete(),this.unsubscribe()},e.prototype._unsubscribeAndRecycle=function(){var t=this._parent,e=this._parents;return this._parent=null,this._parents=null,this.unsubscribe(),this.closed=!1,this.isStopped=!1,this._parent=t,this._parents=e,this._parentSubscription=null,this},e}(a.Subscription),d=function(t){function e(e,n,r,a){var s,u=t.call(this)||this;u._parentSubscriber=e;var c=u;return Object(i.isFunction)(n)?s=n:n&&(s=n.next,r=n.error,a=n.complete,n!==o.empty&&(c=Object.create(n),Object(i.isFunction)(c.unsubscribe)&&u.add(c.unsubscribe.bind(c)),c.unsubscribe=u.unsubscribe.bind(u))),u._context=c,u._next=s,u._error=r,u._complete=a,u}return r.__extends(e,t),e.prototype.next=function(t){if(!this.isStopped&&this._next){var e=this._parentSubscriber;u.config.useDeprecatedSynchronousErrorHandling&&e.syncErrorThrowable?this.__tryOrSetError(e,this._next,t)&&this.unsubscribe():this.__tryOrUnsub(this._next,t)}},e.prototype.error=function(t){if(!this.isStopped){var e=this._parentSubscriber,n=u.config.useDeprecatedSynchronousErrorHandling;if(this._error)n&&e.syncErrorThrowable?(this.__tryOrSetError(e,this._error,t),this.unsubscribe()):(this.__tryOrUnsub(this._error,t),this.unsubscribe());else if(e.syncErrorThrowable)n?(e.syncErrorValue=t,e.syncErrorThrown=!0):Object(c.hostReportError)(t),this.unsubscribe();else{if(this.unsubscribe(),n)throw t;Object(c.hostReportError)(t)}}},e.prototype.complete=function(){var t=this;if(!this.isStopped){var e=this._parentSubscriber;if(this._complete){var n=function(){return t._complete.call(t._context)};u.config.useDeprecatedSynchronousErrorHandling&&e.syncErrorThrowable?(this.__tryOrSetError(e,n),this.unsubscribe()):(this.__tryOrUnsub(n),this.unsubscribe())}else this.unsubscribe()}},e.prototype.__tryOrUnsub=function(t,e){try{t.call(this._context,e)}catch(t){if(this.unsubscribe(),u.config.useDeprecatedSynchronousErrorHandling)throw t;Object(c.hostReportError)(t)}},e.prototype.__tryOrSetError=function(t,e,n){if(!u.config.useDeprecatedSynchronousErrorHandling)throw new Error("bad call");try{e.call(this._context,n)}catch(e){return u.config.useDeprecatedSynchronousErrorHandling?(t.syncErrorValue=e,t.syncErrorThrown=!0,!0):(Object(c.hostReportError)(e),!0)}return!1},e.prototype._unsubscribe=function(){var t=this._parentSubscriber;this._context=null,this._parentSubscriber=null,t.unsubscribe()},e}(l)},"./node_modules/rxjs/_esm5/internal/Subscription.js":function(t,e,n){"use strict";n.r(e),n.d(e,"Subscription",function(){return c});var r=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),i=n("./node_modules/rxjs/_esm5/internal/util/isObject.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isFunction.js"),a=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),s=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),u=n("./node_modules/rxjs/_esm5/internal/util/UnsubscriptionError.js"),c=function(){function t(t){this.closed=!1,this._parent=null,this._parents=null,this._subscriptions=null,t&&(this._unsubscribe=t)}var e;return t.prototype.unsubscribe=function(){var t,e=!1;if(!this.closed){var n=this._parent,c=this._parents,d=this._unsubscribe,p=this._subscriptions;this.closed=!0,this._parent=null,this._parents=null,this._subscriptions=null;for(var f=-1,h=c?c.length:0;n;)n.remove(this),n=++f<h&&c[f]||null;if(Object(o.isFunction)(d))Object(a.tryCatch)(d).call(this)===s.errorObject&&(e=!0,t=t||(s.errorObject.e instanceof u.UnsubscriptionError?l(s.errorObject.e.errors):[s.errorObject.e]));if(Object(r.isArray)(p))for(f=-1,h=p.length;++f<h;){var m=p[f];if(Object(i.isObject)(m))if(Object(a.tryCatch)(m.unsubscribe).call(m)===s.errorObject){e=!0,t=t||[];var g=s.errorObject.e;g instanceof u.UnsubscriptionError?t=t.concat(l(g.errors)):t.push(g)}}if(e)throw new u.UnsubscriptionError(t)}},t.prototype.add=function(e){if(!e||e===t.EMPTY)return t.EMPTY;if(e===this)return this;var n=e;switch(typeof e){case"function":n=new t(e);case"object":if(n.closed||"function"!=typeof n.unsubscribe)return n;if(this.closed)return n.unsubscribe(),n;if("function"!=typeof n._addParent){var r=n;(n=new t)._subscriptions=[r]}break;default:throw new Error("unrecognized teardown "+e+" added to Subscription.")}return(this._subscriptions||(this._subscriptions=[])).push(n),n._addParent(this),n},t.prototype.remove=function(t){var e=this._subscriptions;if(e){var n=e.indexOf(t);-1!==n&&e.splice(n,1)}},t.prototype._addParent=function(t){var e=this._parent,n=this._parents;e&&e!==t?n?-1===n.indexOf(t)&&n.push(t):this._parents=[t]:this._parent=t},t.EMPTY=((e=new t).closed=!0,e),t}();function l(t){return t.reduce(function(t,e){return t.concat(e instanceof u.UnsubscriptionError?e.errors:e)},[])}},"./node_modules/rxjs/_esm5/internal/config.js":function(t,e,n){"use strict";n.r(e),n.d(e,"config",function(){return i});var r=!1,i={Promise:void 0,set useDeprecatedSynchronousErrorHandling(t){if(t){var e=new Error;e.stack}r=t},get useDeprecatedSynchronousErrorHandling(){return r}}},"./node_modules/rxjs/_esm5/internal/observable/ConnectableObservable.js":function(t,e,n){"use strict";n.r(e),n.d(e,"ConnectableObservable",function(){return c}),n.d(e,"connectableObservableDescriptor",function(){return d});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/Observable.js"),a=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),u=n("./node_modules/rxjs/_esm5/internal/operators/refCount.js"),c=function(t){function e(e,n){var r=t.call(this)||this;return r.source=e,r.subjectFactory=n,r._refCount=0,r._isComplete=!1,r}return r.__extends(e,t),e.prototype._subscribe=function(t){return this.getSubject().subscribe(t)},e.prototype.getSubject=function(){var t=this._subject;return t&&!t.isStopped||(this._subject=this.subjectFactory()),this._subject},e.prototype.connect=function(){var t=this._connection;return t||(this._isComplete=!1,(t=this._connection=new s.Subscription).add(this.source.subscribe(new p(this.getSubject(),this))),t.closed?(this._connection=null,t=s.Subscription.EMPTY):this._connection=t),t},e.prototype.refCount=function(){return Object(u.refCount)()(this)},e}(o.Observable),l=c.prototype,d={operator:{value:null},_refCount:{value:0,writable:!0},_subject:{value:null,writable:!0},_connection:{value:null,writable:!0},_subscribe:{value:l._subscribe},_isComplete:{value:l._isComplete,writable:!0},getSubject:{value:l.getSubject},connect:{value:l.connect},refCount:{value:l.refCount}},p=function(t){function e(e,n){var r=t.call(this,e)||this;return r.connectable=n,r}return r.__extends(e,t),e.prototype._error=function(e){this._unsubscribe(),t.prototype._error.call(this,e)},e.prototype._complete=function(){this.connectable._isComplete=!0,this._unsubscribe(),t.prototype._complete.call(this)},e.prototype._unsubscribe=function(){var t=this.connectable;if(t){this.connectable=null;var e=t._connection;t._refCount=0,t._subject=null,t._connection=null,e&&e.unsubscribe()}},e}(i.SubjectSubscriber);a.Subscriber},"./node_modules/rxjs/_esm5/internal/observable/SubscribeOnObservable.js":function(t,e,n){"use strict";n.r(e),n.d(e,"SubscribeOnObservable",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Observable.js"),o=n("./node_modules/rxjs/_esm5/internal/scheduler/asap.js"),a=n("./node_modules/rxjs/_esm5/internal/util/isNumeric.js"),s=function(t){function e(e,n,r){void 0===n&&(n=0),void 0===r&&(r=o.asap);var i=t.call(this)||this;return i.source=e,i.delayTime=n,i.scheduler=r,(!Object(a.isNumeric)(n)||n<0)&&(i.delayTime=0),r&&"function"==typeof r.schedule||(i.scheduler=o.asap),i}return r.__extends(e,t),e.create=function(t,n,r){return void 0===n&&(n=0),void 0===r&&(r=o.asap),new e(t,n,r)},e.dispatch=function(t){var e=t.source,n=t.subscriber;return this.add(e.subscribe(n))},e.prototype._subscribe=function(t){var n=this.delayTime,r=this.source;return this.scheduler.schedule(e.dispatch,n,{source:r,subscriber:t})},e}(i.Observable)},"./node_modules/rxjs/_esm5/internal/observable/bindCallback.js":function(t,e,n){"use strict";n.r(e),n.d(e,"bindCallback",function(){return c});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/AsyncSubject.js"),o=n("./node_modules/rxjs/_esm5/internal/operators/map.js"),a=n("./node_modules/rxjs/_esm5/internal/util/canReportError.js"),s=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),u=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js");function c(t,e,n){if(e){if(!Object(u.isScheduler)(e))return function(){for(var r=[],i=0;i<arguments.length;i++)r[i]=arguments[i];return c(t,n).apply(void 0,r).pipe(Object(o.map)(function(t){return Object(s.isArray)(t)?e.apply(void 0,t):e(t)}))};n=e}return function(){for(var e=[],o=0;o<arguments.length;o++)e[o]=arguments[o];var s,u=this,c={context:u,subject:s,callbackFunc:t,scheduler:n};return new r.Observable(function(r){if(n){var o={args:e,subscriber:r,params:c};return n.schedule(l,0,o)}if(!s){s=new i.AsyncSubject;try{t.apply(u,e.concat([function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];s.next(t.length<=1?t[0]:t),s.complete()}]))}catch(t){Object(a.canReportError)(s)?s.error(t):console.warn(t)}}return s.subscribe(r)})}}function l(t){var e=this,n=t.args,r=t.subscriber,o=t.params,a=o.callbackFunc,s=o.context,u=o.scheduler,c=o.subject;if(!c){c=o.subject=new i.AsyncSubject;try{a.apply(s,n.concat([function(){for(var t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];var r=t.length<=1?t[0]:t;e.add(u.schedule(d,0,{value:r,subject:c}))}]))}catch(t){c.error(t)}}this.add(c.subscribe(r))}function d(t){var e=t.value,n=t.subject;n.next(e),n.complete()}},"./node_modules/rxjs/_esm5/internal/observable/bindNodeCallback.js":function(t,e,n){"use strict";n.r(e),n.d(e,"bindNodeCallback",function(){return c});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/AsyncSubject.js"),o=n("./node_modules/rxjs/_esm5/internal/operators/map.js"),a=n("./node_modules/rxjs/_esm5/internal/util/canReportError.js"),s=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js"),u=n("./node_modules/rxjs/_esm5/internal/util/isArray.js");function c(t,e,n){if(e){if(!Object(s.isScheduler)(e))return function(){for(var r=[],i=0;i<arguments.length;i++)r[i]=arguments[i];return c(t,n).apply(void 0,r).pipe(Object(o.map)(function(t){return Object(u.isArray)(t)?e.apply(void 0,t):e(t)}))};n=e}return function(){for(var e=[],o=0;o<arguments.length;o++)e[o]=arguments[o];var s={subject:void 0,args:e,callbackFunc:t,scheduler:n,context:this};return new r.Observable(function(r){var o=s.context,u=s.subject;if(n)return n.schedule(l,0,{params:s,subscriber:r,context:o});if(!u){u=s.subject=new i.AsyncSubject;try{t.apply(o,e.concat([function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];var n=t.shift();n?u.error(n):(u.next(t.length<=1?t[0]:t),u.complete())}]))}catch(t){Object(a.canReportError)(u)?u.error(t):console.warn(t)}}return u.subscribe(r)})}}function l(t){var e=this,n=t.params,r=t.subscriber,o=t.context,a=n.callbackFunc,s=n.args,u=n.scheduler,c=n.subject;if(!c){c=n.subject=new i.AsyncSubject;try{a.apply(o,s.concat([function(){for(var t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];var r=t.shift();if(r)e.add(u.schedule(p,0,{err:r,subject:c}));else{var i=t.length<=1?t[0]:t;e.add(u.schedule(d,0,{value:i,subject:c}))}}]))}catch(t){this.add(u.schedule(p,0,{err:t,subject:c}))}}this.add(c.subscribe(r))}function d(t){var e=t.value,n=t.subject;n.next(e),n.complete()}function p(t){var e=t.err;t.subject.error(e)}},"./node_modules/rxjs/_esm5/internal/observable/combineLatest.js":function(t,e,n){"use strict";n.r(e),n.d(e,"combineLatest",function(){return l}),n.d(e,"CombineLatestOperator",function(){return d}),n.d(e,"CombineLatestSubscriber",function(){return p});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),a=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),u=n("./node_modules/rxjs/_esm5/internal/observable/fromArray.js"),c={};function l(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];var n=null,r=null;return Object(i.isScheduler)(t[t.length-1])&&(r=t.pop()),"function"==typeof t[t.length-1]&&(n=t.pop()),1===t.length&&Object(o.isArray)(t[0])&&(t=t[0]),Object(u.fromArray)(t,r).lift(new d(n))}var d=function(){function t(t){this.resultSelector=t}return t.prototype.call=function(t,e){return e.subscribe(new p(t,this.resultSelector))},t}(),p=function(t){function e(e,n){var r=t.call(this,e)||this;return r.resultSelector=n,r.active=0,r.values=[],r.observables=[],r}return r.__extends(e,t),e.prototype._next=function(t){this.values.push(c),this.observables.push(t)},e.prototype._complete=function(){var t=this.observables,e=t.length;if(0===e)this.destination.complete();else{this.active=e,this.toRespond=e;for(var n=0;n<e;n++){var r=t[n];this.add(Object(s.subscribeToResult)(this,r,r,n))}}},e.prototype.notifyComplete=function(t){0==(this.active-=1)&&this.destination.complete()},e.prototype.notifyNext=function(t,e,n,r,i){var o=this.values,a=o[n],s=this.toRespond?a===c?--this.toRespond:this.toRespond:0;o[n]=e,0===s&&(this.resultSelector?this._tryResultSelector(o):this.destination.next(o.slice()))},e.prototype._tryResultSelector=function(t){var e;try{e=this.resultSelector.apply(this,t)}catch(t){return void this.destination.error(t)}this.destination.next(e)},e}(a.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/observable/concat.js":function(t,e,n){"use strict";n.r(e),n.d(e,"concat",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/of.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/from.js"),a=n("./node_modules/rxjs/_esm5/internal/operators/concatAll.js");function s(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return 1===t.length||2===t.length&&Object(r.isScheduler)(t[1])?Object(o.from)(t[0]):Object(a.concatAll)()(i.of.apply(void 0,t))}},"./node_modules/rxjs/_esm5/internal/observable/defer.js":function(t,e,n){"use strict";n.r(e),n.d(e,"defer",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/from.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/empty.js");function a(t){return new r.Observable(function(e){var n;try{n=t()}catch(t){return void e.error(t)}return(n?Object(i.from)(n):Object(o.empty)()).subscribe(e)})}},"./node_modules/rxjs/_esm5/internal/observable/empty.js":function(t,e,n){"use strict";n.r(e),n.d(e,"EMPTY",function(){return i}),n.d(e,"empty",function(){return o}),n.d(e,"emptyScheduled",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=new r.Observable(function(t){return t.complete()});function o(t){return t?a(t):i}function a(t){return new r.Observable(function(e){return t.schedule(function(){return e.complete()})})}},"./node_modules/rxjs/_esm5/internal/observable/forkJoin.js":function(t,e,n){"use strict";n.r(e),n.d(e,"forkJoin",function(){return l});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Observable.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/empty.js"),s=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),u=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),c=n("./node_modules/rxjs/_esm5/internal/operators/map.js");function l(){for(var t,e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return"function"==typeof e[e.length-1]&&(t=e.pop()),1===e.length&&Object(o.isArray)(e[0])&&(e=e[0]),0===e.length?a.EMPTY:t?l(e).pipe(Object(c.map)(function(e){return t.apply(void 0,e)})):new i.Observable(function(t){return new d(t,e)})}var d=function(t){function e(e,n){var r=t.call(this,e)||this;r.sources=n,r.completed=0,r.haveValues=0;var i=n.length;r.values=new Array(i);for(var o=0;o<i;o++){var a=n[o],u=Object(s.subscribeToResult)(r,a,null,o);u&&r.add(u)}return r}return r.__extends(e,t),e.prototype.notifyNext=function(t,e,n,r,i){this.values[n]=e,i._hasValue||(i._hasValue=!0,this.haveValues++)},e.prototype.notifyComplete=function(t){var e=this.destination,n=this.haveValues,r=this.values,i=r.length;t._hasValue?(this.completed++,this.completed===i&&(n===i&&e.next(r),e.complete())):e.complete()},e}(u.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/observable/from.js":function(t,e,n){"use strict";n.r(e),n.d(e,"from",function(){return f});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/util/isPromise.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isArrayLike.js"),a=n("./node_modules/rxjs/_esm5/internal/util/isInteropObservable.js"),s=n("./node_modules/rxjs/_esm5/internal/util/isIterable.js"),u=n("./node_modules/rxjs/_esm5/internal/observable/fromArray.js"),c=n("./node_modules/rxjs/_esm5/internal/observable/fromPromise.js"),l=n("./node_modules/rxjs/_esm5/internal/observable/fromIterable.js"),d=n("./node_modules/rxjs/_esm5/internal/observable/fromObservable.js"),p=n("./node_modules/rxjs/_esm5/internal/util/subscribeTo.js");function f(t,e){if(!e)return t instanceof r.Observable?t:new r.Observable(Object(p.subscribeTo)(t));if(null!=t){if(Object(a.isInteropObservable)(t))return Object(d.fromObservable)(t,e);if(Object(i.isPromise)(t))return Object(c.fromPromise)(t,e);if(Object(o.isArrayLike)(t))return Object(u.fromArray)(t,e);if(Object(s.isIterable)(t)||"string"==typeof t)return Object(l.fromIterable)(t,e)}throw new TypeError((null!==t&&typeof t||t)+" is not observable")}},"./node_modules/rxjs/_esm5/internal/observable/fromArray.js":function(t,e,n){"use strict";n.r(e),n.d(e,"fromArray",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToArray.js");function a(t,e){return e?new r.Observable(function(n){var r=new i.Subscription,o=0;return r.add(e.schedule(function(){o!==t.length?(n.next(t[o++]),n.closed||r.add(this.schedule())):n.complete()})),r}):new r.Observable(Object(o.subscribeToArray)(t))}},"./node_modules/rxjs/_esm5/internal/observable/fromEvent.js":function(t,e,n){"use strict";n.r(e),n.d(e,"fromEvent",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isFunction.js"),a=n("./node_modules/rxjs/_esm5/internal/operators/map.js");Object.prototype.toString;function s(t,e,n,u){return Object(o.isFunction)(n)&&(u=n,n=void 0),u?s(t,e,n).pipe(Object(a.map)(function(t){return Object(i.isArray)(t)?u.apply(void 0,t):u(t)})):new r.Observable(function(r){!function t(e,n,r,i,o){var a;if(h=e,h&&"function"==typeof h.addEventListener&&"function"==typeof h.removeEventListener){var s=e;e.addEventListener(n,r,o),a=function(){return s.removeEventListener(n,r,o)}}else if(f=e,f&&"function"==typeof f.on&&"function"==typeof f.off){var u=e;e.on(n,r),a=function(){return u.off(n,r)}}else if(p=e,p&&"function"==typeof p.addListener&&"function"==typeof p.removeListener){var c=e;e.addListener(n,r),a=function(){return c.removeListener(n,r)}}else{if(!e||!e.length)throw new TypeError("Invalid event target");for(var l=0,d=e.length;l<d;l++)t(e[l],n,r,i,o)}var p;var f;var h;i.add(a)}(t,e,function(t){arguments.length>1?r.next(Array.prototype.slice.call(arguments)):r.next(t)},r,n)})}},"./node_modules/rxjs/_esm5/internal/observable/fromEventPattern.js":function(t,e,n){"use strict";n.r(e),n.d(e,"fromEventPattern",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isFunction.js"),a=n("./node_modules/rxjs/_esm5/internal/operators/map.js");function s(t,e,n){return n?s(t,e).pipe(Object(a.map)(function(t){return Object(i.isArray)(t)?n.apply(void 0,t):n(t)})):new r.Observable(function(n){var r,i=function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return n.next(1===t.length?t[0]:t)};try{r=t(i)}catch(t){return void n.error(t)}if(Object(o.isFunction)(e))return function(){return e(i,r)}})}},"./node_modules/rxjs/_esm5/internal/observable/fromIterable.js":function(t,e,n){"use strict";n.r(e),n.d(e,"fromIterable",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),o=n("./node_modules/rxjs/_esm5/internal/symbol/iterator.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToIterable.js");function s(t,e){if(!t)throw new Error("Iterable cannot be null");return e?new r.Observable(function(n){var r,a=new i.Subscription;return a.add(function(){r&&"function"==typeof r.return&&r.return()}),a.add(e.schedule(function(){r=t[o.iterator](),a.add(e.schedule(function(){if(!n.closed){var t,e;try{var i=r.next();t=i.value,e=i.done}catch(t){return void n.error(t)}e?n.complete():(n.next(t),this.schedule())}}))})),a}):new r.Observable(Object(a.subscribeToIterable)(t))}},"./node_modules/rxjs/_esm5/internal/observable/fromObservable.js":function(t,e,n){"use strict";n.r(e),n.d(e,"fromObservable",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),o=n("./node_modules/rxjs/_esm5/internal/symbol/observable.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToObservable.js");function s(t,e){return e?new r.Observable(function(n){var r=new i.Subscription;return r.add(e.schedule(function(){var i=t[o.observable]();r.add(i.subscribe({next:function(t){r.add(e.schedule(function(){return n.next(t)}))},error:function(t){r.add(e.schedule(function(){return n.error(t)}))},complete:function(){r.add(e.schedule(function(){return n.complete()}))}}))})),r}):new r.Observable(Object(a.subscribeToObservable)(t))}},"./node_modules/rxjs/_esm5/internal/observable/fromPromise.js":function(t,e,n){"use strict";n.r(e),n.d(e,"fromPromise",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToPromise.js");function a(t,e){return e?new r.Observable(function(n){var r=new i.Subscription;return r.add(e.schedule(function(){return t.then(function(t){r.add(e.schedule(function(){n.next(t),r.add(e.schedule(function(){return n.complete()}))}))},function(t){r.add(e.schedule(function(){return n.error(t)}))})})),r}):new r.Observable(Object(o.subscribeToPromise)(t))}},"./node_modules/rxjs/_esm5/internal/observable/generate.js":function(t,e,n){"use strict";n.r(e),n.d(e,"generate",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/util/identity.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js");function a(t,e,n,a,u){var c,l;if(1==arguments.length){var d=t;l=d.initialState,e=d.condition,n=d.iterate,c=d.resultSelector||i.identity,u=d.scheduler}else void 0===a||Object(o.isScheduler)(a)?(l=t,c=i.identity,u=a):(l=t,c=a);return new r.Observable(function(t){var r=l;if(u)return u.schedule(s,0,{subscriber:t,iterate:n,condition:e,resultSelector:c,state:r});for(;;){if(e){var i=void 0;try{i=e(r)}catch(e){return void t.error(e)}if(!i){t.complete();break}}var o=void 0;try{o=c(r)}catch(e){return void t.error(e)}if(t.next(o),t.closed)break;try{r=n(r)}catch(e){return void t.error(e)}}})}function s(t){var e=t.subscriber,n=t.condition;if(!e.closed){if(t.needIterate)try{t.state=t.iterate(t.state)}catch(t){return void e.error(t)}else t.needIterate=!0;if(n){var r=void 0;try{r=n(t.state)}catch(t){return void e.error(t)}if(!r)return void e.complete();if(e.closed)return}var i;try{i=t.resultSelector(t.state)}catch(t){return void e.error(t)}if(!e.closed&&(e.next(i),!e.closed))return this.schedule(t)}}},"./node_modules/rxjs/_esm5/internal/observable/iif.js":function(t,e,n){"use strict";n.r(e),n.d(e,"iif",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/observable/defer.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/empty.js");function o(t,e,n){return void 0===e&&(e=i.EMPTY),void 0===n&&(n=i.EMPTY),Object(r.defer)(function(){return t()?e:n})}},"./node_modules/rxjs/_esm5/internal/observable/interval.js":function(t,e,n){"use strict";n.r(e),n.d(e,"interval",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isNumeric.js");function a(t,e){return void 0===t&&(t=0),void 0===e&&(e=i.async),(!Object(o.isNumeric)(t)||t<0)&&(t=0),e&&"function"==typeof e.schedule||(e=i.async),new r.Observable(function(n){return n.add(e.schedule(s,t,{subscriber:n,counter:0,period:t})),n})}function s(t){var e=t.subscriber,n=t.counter,r=t.period;e.next(n),this.schedule({subscriber:e,counter:n+1,period:r},r)}},"./node_modules/rxjs/_esm5/internal/observable/merge.js":function(t,e,n){"use strict";n.r(e),n.d(e,"merge",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js"),o=n("./node_modules/rxjs/_esm5/internal/operators/mergeAll.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/fromArray.js");function s(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];var n=Number.POSITIVE_INFINITY,s=null,u=t[t.length-1];return Object(i.isScheduler)(u)?(s=t.pop(),t.length>1&&"number"==typeof t[t.length-1]&&(n=t.pop())):"number"==typeof u&&(n=t.pop()),null===s&&1===t.length&&t[0]instanceof r.Observable?t[0]:Object(o.mergeAll)(n)(Object(a.fromArray)(t,s))}},"./node_modules/rxjs/_esm5/internal/observable/never.js":function(t,e,n){"use strict";n.r(e),n.d(e,"NEVER",function(){return o}),n.d(e,"never",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/util/noop.js"),o=new r.Observable(i.noop);function a(){return o}},"./node_modules/rxjs/_esm5/internal/observable/of.js":function(t,e,n){"use strict";n.r(e),n.d(e,"of",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/fromArray.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/empty.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/scalar.js");function s(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];var n=t[t.length-1];switch(Object(r.isScheduler)(n)?t.pop():n=void 0,t.length){case 0:return Object(o.empty)(n);case 1:return n?Object(i.fromArray)(t,n):Object(a.scalar)(t[0]);default:return Object(i.fromArray)(t,n)}}},"./node_modules/rxjs/_esm5/internal/observable/onErrorResumeNext.js":function(t,e,n){"use strict";n.r(e),n.d(e,"onErrorResumeNext",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/from.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/empty.js");function s(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];if(0===t.length)return a.EMPTY;var n=t[0],u=t.slice(1);return 1===t.length&&Object(o.isArray)(n)?s.apply(void 0,n):new r.Observable(function(t){var e=function(){return t.add(s.apply(void 0,u).subscribe(t))};return Object(i.from)(n).subscribe({next:function(e){t.next(e)},error:e,complete:e})})}},"./node_modules/rxjs/_esm5/internal/observable/pairs.js":function(t,e,n){"use strict";n.r(e),n.d(e,"pairs",function(){return o}),n.d(e,"dispatch",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscription.js");function o(t,e){return e?new r.Observable(function(n){var r=Object.keys(t),o=new i.Subscription;return o.add(e.schedule(a,0,{keys:r,index:0,subscriber:n,subscription:o,obj:t})),o}):new r.Observable(function(e){for(var n=Object.keys(t),r=0;r<n.length&&!e.closed;r++){var i=n[r];t.hasOwnProperty(i)&&e.next([i,t[i]])}e.complete()})}function a(t){var e=t.keys,n=t.index,r=t.subscriber,i=t.subscription,o=t.obj;if(!r.closed)if(n<e.length){var a=e[n];r.next([a,o[a]]),i.add(this.schedule({keys:e,index:n+1,subscriber:r,subscription:i,obj:o}))}else r.complete()}},"./node_modules/rxjs/_esm5/internal/observable/race.js":function(t,e,n){"use strict";n.r(e),n.d(e,"race",function(){return u}),n.d(e,"RaceOperator",function(){return c}),n.d(e,"RaceSubscriber",function(){return l});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/fromArray.js"),a=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function u(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];if(1===t.length){if(!Object(i.isArray)(t[0]))return t[0];t=t[0]}return Object(o.fromArray)(t,void 0).lift(new c)}var c=function(){function t(){}return t.prototype.call=function(t,e){return e.subscribe(new l(t))},t}(),l=function(t){function e(e){var n=t.call(this,e)||this;return n.hasFirst=!1,n.observables=[],n.subscriptions=[],n}return r.__extends(e,t),e.prototype._next=function(t){this.observables.push(t)},e.prototype._complete=function(){var t=this.observables,e=t.length;if(0===e)this.destination.complete();else{for(var n=0;n<e&&!this.hasFirst;n++){var r=t[n],i=Object(s.subscribeToResult)(this,r,r,n);this.subscriptions&&this.subscriptions.push(i),this.add(i)}this.observables=null}},e.prototype.notifyNext=function(t,e,n,r,i){if(!this.hasFirst){this.hasFirst=!0;for(var o=0;o<this.subscriptions.length;o++)if(o!==n){var a=this.subscriptions[o];a.unsubscribe(),this.remove(a)}this.subscriptions=null}this.destination.next(e)},e}(a.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/observable/range.js":function(t,e,n){"use strict";n.r(e),n.d(e,"range",function(){return i}),n.d(e,"dispatch",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js");function i(t,e,n){return void 0===t&&(t=0),void 0===e&&(e=0),new r.Observable(function(r){var i=0,a=t;if(n)return n.schedule(o,0,{index:i,count:e,start:t,subscriber:r});for(;;){if(i++>=e){r.complete();break}if(r.next(a++),r.closed)break}})}function o(t){var e=t.start,n=t.index,r=t.count,i=t.subscriber;n>=r?i.complete():(i.next(e),i.closed||(t.index=n+1,t.start=e+1,this.schedule(t)))}},"./node_modules/rxjs/_esm5/internal/observable/scalar.js":function(t,e,n){"use strict";n.r(e),n.d(e,"scalar",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js");function i(t){var e=new r.Observable(function(e){e.next(t),e.complete()});return e._isScalar=!0,e.value=t,e}},"./node_modules/rxjs/_esm5/internal/observable/throwError.js":function(t,e,n){"use strict";n.r(e),n.d(e,"throwError",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js");function i(t,e){return e?new r.Observable(function(n){return e.schedule(o,0,{error:t,subscriber:n})}):new r.Observable(function(e){return e.error(t)})}function o(t){var e=t.error;t.subscriber.error(e)}},"./node_modules/rxjs/_esm5/internal/observable/timer.js":function(t,e,n){"use strict";n.r(e),n.d(e,"timer",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isNumeric.js"),a=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js");function s(t,e,n){void 0===t&&(t=0);var s=-1;return Object(o.isNumeric)(e)?s=Number(e)<1?1:Number(e):Object(a.isScheduler)(e)&&(n=e),Object(a.isScheduler)(n)||(n=i.async),new r.Observable(function(e){var r=Object(o.isNumeric)(t)?t:+t-n.now();return n.schedule(u,r,{index:0,period:s,subscriber:e})})}function u(t){var e=t.index,n=t.period,r=t.subscriber;if(r.next(e),!r.closed){if(-1===n)return r.complete();t.index=e+1,this.schedule(t,n)}}},"./node_modules/rxjs/_esm5/internal/observable/using.js":function(t,e,n){"use strict";n.r(e),n.d(e,"using",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/from.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/empty.js");function a(t,e){return new r.Observable(function(n){var r,a;try{r=t()}catch(t){return void n.error(t)}try{a=e(r)}catch(t){return void n.error(t)}var s=(a?Object(i.from)(a):o.EMPTY).subscribe(n);return function(){s.unsubscribe(),r&&r.unsubscribe()}})}},"./node_modules/rxjs/_esm5/internal/observable/zip.js":function(t,e,n){"use strict";n.r(e),n.d(e,"zip",function(){return l}),n.d(e,"ZipOperator",function(){return d}),n.d(e,"ZipSubscriber",function(){return p});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/fromArray.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),a=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),u=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),c=n("./node_modules/rxjs/_esm5/internal/symbol/iterator.js");function l(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];var n=t[t.length-1];return"function"==typeof n&&t.pop(),Object(i.fromArray)(t,void 0).lift(new d(n))}var d=function(){function t(t){this.resultSelector=t}return t.prototype.call=function(t,e){return e.subscribe(new p(t,this.resultSelector))},t}(),p=function(t){function e(e,n,r){void 0===r&&(r=Object.create(null));var i=t.call(this,e)||this;return i.iterators=[],i.active=0,i.resultSelector="function"==typeof n?n:null,i.values=r,i}return r.__extends(e,t),e.prototype._next=function(t){var e=this.iterators;Object(o.isArray)(t)?e.push(new h(t)):"function"==typeof t[c.iterator]?e.push(new f(t[c.iterator]())):e.push(new m(this.destination,this,t))},e.prototype._complete=function(){var t=this.iterators,e=t.length;if(this.unsubscribe(),0!==e){this.active=e;for(var n=0;n<e;n++){var r=t[n];if(r.stillUnsubscribed)this.destination.add(r.subscribe(r,n));else this.active--}}else this.destination.complete()},e.prototype.notifyInactive=function(){this.active--,0===this.active&&this.destination.complete()},e.prototype.checkIterators=function(){for(var t=this.iterators,e=t.length,n=this.destination,r=0;r<e;r++){if("function"==typeof(a=t[r]).hasValue&&!a.hasValue())return}var i=!1,o=[];for(r=0;r<e;r++){var a,s=(a=t[r]).next();if(a.hasCompleted()&&(i=!0),s.done)return void n.complete();o.push(s.value)}this.resultSelector?this._tryresultSelector(o):n.next(o),i&&n.complete()},e.prototype._tryresultSelector=function(t){var e;try{e=this.resultSelector.apply(this,t)}catch(t){return void this.destination.error(t)}this.destination.next(e)},e}(a.Subscriber),f=function(){function t(t){this.iterator=t,this.nextResult=t.next()}return t.prototype.hasValue=function(){return!0},t.prototype.next=function(){var t=this.nextResult;return this.nextResult=this.iterator.next(),t},t.prototype.hasCompleted=function(){var t=this.nextResult;return t&&t.done},t}(),h=function(){function t(t){this.array=t,this.index=0,this.length=0,this.length=t.length}return t.prototype[c.iterator]=function(){return this},t.prototype.next=function(t){var e=this.index++,n=this.array;return e<this.length?{value:n[e],done:!1}:{value:null,done:!0}},t.prototype.hasValue=function(){return this.array.length>this.index},t.prototype.hasCompleted=function(){return this.array.length===this.index},t}(),m=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.parent=n,i.observable=r,i.stillUnsubscribed=!0,i.buffer=[],i.isComplete=!1,i}return r.__extends(e,t),e.prototype[c.iterator]=function(){return this},e.prototype.next=function(){var t=this.buffer;return 0===t.length&&this.isComplete?{value:null,done:!0}:{value:t.shift(),done:!1}},e.prototype.hasValue=function(){return this.buffer.length>0},e.prototype.hasCompleted=function(){return 0===this.buffer.length&&this.isComplete},e.prototype.notifyComplete=function(){this.buffer.length>0?(this.isComplete=!0,this.parent.notifyInactive()):this.destination.complete()},e.prototype.notifyNext=function(t,e,n,r,i){this.buffer.push(e),this.parent.checkIterators()},e.prototype.subscribe=function(t,e){return Object(u.subscribeToResult)(this,this.observable,this,e)},e}(s.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/audit.js":function(t,e,n){"use strict";n.r(e),n.d(e,"audit",function(){return u});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),o=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),a=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function u(t){return function(e){return e.lift(new c(t))}}var c=function(){function t(t){this.durationSelector=t}return t.prototype.call=function(t,e){return e.subscribe(new l(t,this.durationSelector))},t}(),l=function(t){function e(e,n){var r=t.call(this,e)||this;return r.durationSelector=n,r.hasValue=!1,r}return r.__extends(e,t),e.prototype._next=function(t){if(this.value=t,this.hasValue=!0,!this.throttled){var e=Object(i.tryCatch)(this.durationSelector)(t);if(e===o.errorObject)this.destination.error(o.errorObject.e);else{var n=Object(s.subscribeToResult)(this,e);!n||n.closed?this.clearThrottle():this.add(this.throttled=n)}}},e.prototype.clearThrottle=function(){var t=this.value,e=this.hasValue,n=this.throttled;n&&(this.remove(n),this.throttled=null,n.unsubscribe()),e&&(this.value=null,this.hasValue=!1,this.destination.next(t))},e.prototype.notifyNext=function(t,e,n,r){this.clearThrottle()},e.prototype.notifyComplete=function(){this.clearThrottle()},e}(a.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/auditTime.js":function(t,e,n){"use strict";n.r(e),n.d(e,"auditTime",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/audit.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/timer.js");function a(t,e){return void 0===e&&(e=r.async),Object(i.audit)(function(){return Object(o.timer)(t,e)})}},"./node_modules/rxjs/_esm5/internal/operators/buffer.js":function(t,e,n){"use strict";n.r(e),n.d(e,"buffer",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function a(t){return function(e){return e.lift(new s(t))}}var s=function(){function t(t){this.closingNotifier=t}return t.prototype.call=function(t,e){return e.subscribe(new u(t,this.closingNotifier))},t}(),u=function(t){function e(e,n){var r=t.call(this,e)||this;return r.buffer=[],r.add(Object(o.subscribeToResult)(r,n)),r}return r.__extends(e,t),e.prototype._next=function(t){this.buffer.push(t)},e.prototype.notifyNext=function(t,e,n,r,i){var o=this.buffer;this.buffer=[],this.destination.next(o)},e}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/bufferCount.js":function(t,e,n){"use strict";n.r(e),n.d(e,"bufferCount",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(t,e){return void 0===e&&(e=null),function(n){return n.lift(new a(t,e))}}var a=function(){function t(t,e){this.bufferSize=t,this.startBufferEvery=e,this.subscriberClass=e&&t!==e?u:s}return t.prototype.call=function(t,e){return e.subscribe(new this.subscriberClass(t,this.bufferSize,this.startBufferEvery))},t}(),s=function(t){function e(e,n){var r=t.call(this,e)||this;return r.bufferSize=n,r.buffer=[],r}return r.__extends(e,t),e.prototype._next=function(t){var e=this.buffer;e.push(t),e.length==this.bufferSize&&(this.destination.next(e),this.buffer=[])},e.prototype._complete=function(){var e=this.buffer;e.length>0&&this.destination.next(e),t.prototype._complete.call(this)},e}(i.Subscriber),u=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.bufferSize=n,i.startBufferEvery=r,i.buffers=[],i.count=0,i}return r.__extends(e,t),e.prototype._next=function(t){var e=this.bufferSize,n=this.startBufferEvery,r=this.buffers,i=this.count;this.count++,i%n==0&&r.push([]);for(var o=r.length;o--;){var a=r[o];a.push(t),a.length===e&&(r.splice(o,1),this.destination.next(a))}},e.prototype._complete=function(){for(var e=this.buffers,n=this.destination;e.length>0;){var r=e.shift();r.length>0&&n.next(r)}t.prototype._complete.call(this)},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/bufferTime.js":function(t,e,n){"use strict";n.r(e),n.d(e,"bufferTime",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),o=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),a=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js");function s(t){var e=arguments.length,n=i.async;Object(a.isScheduler)(arguments[arguments.length-1])&&(n=arguments[arguments.length-1],e--);var r=null;e>=2&&(r=arguments[1]);var o=Number.POSITIVE_INFINITY;return e>=3&&(o=arguments[2]),function(e){return e.lift(new u(t,r,o,n))}}var u=function(){function t(t,e,n,r){this.bufferTimeSpan=t,this.bufferCreationInterval=e,this.maxBufferSize=n,this.scheduler=r}return t.prototype.call=function(t,e){return e.subscribe(new l(t,this.bufferTimeSpan,this.bufferCreationInterval,this.maxBufferSize,this.scheduler))},t}(),c=function(){return function(){this.buffer=[]}}(),l=function(t){function e(e,n,r,i,o){var a=t.call(this,e)||this;a.bufferTimeSpan=n,a.bufferCreationInterval=r,a.maxBufferSize=i,a.scheduler=o,a.contexts=[];var s=a.openContext();if(a.timespanOnly=null==r||r<0,a.timespanOnly){var u={subscriber:a,context:s,bufferTimeSpan:n};a.add(s.closeAction=o.schedule(d,n,u))}else{var c={subscriber:a,context:s},l={bufferTimeSpan:n,bufferCreationInterval:r,subscriber:a,scheduler:o};a.add(s.closeAction=o.schedule(f,n,c)),a.add(o.schedule(p,r,l))}return a}return r.__extends(e,t),e.prototype._next=function(t){for(var e,n=this.contexts,r=n.length,i=0;i<r;i++){var o=n[i],a=o.buffer;a.push(t),a.length==this.maxBufferSize&&(e=o)}e&&this.onBufferFull(e)},e.prototype._error=function(e){this.contexts.length=0,t.prototype._error.call(this,e)},e.prototype._complete=function(){for(var e=this.contexts,n=this.destination;e.length>0;){var r=e.shift();n.next(r.buffer)}t.prototype._complete.call(this)},e.prototype._unsubscribe=function(){this.contexts=null},e.prototype.onBufferFull=function(t){this.closeContext(t);var e=t.closeAction;if(e.unsubscribe(),this.remove(e),!this.closed&&this.timespanOnly){t=this.openContext();var n=this.bufferTimeSpan,r={subscriber:this,context:t,bufferTimeSpan:n};this.add(t.closeAction=this.scheduler.schedule(d,n,r))}},e.prototype.openContext=function(){var t=new c;return this.contexts.push(t),t},e.prototype.closeContext=function(t){this.destination.next(t.buffer);var e=this.contexts;(e?e.indexOf(t):-1)>=0&&e.splice(e.indexOf(t),1)},e}(o.Subscriber);function d(t){var e=t.subscriber,n=t.context;n&&e.closeContext(n),e.closed||(t.context=e.openContext(),t.context.closeAction=this.schedule(t,t.bufferTimeSpan))}function p(t){var e=t.bufferCreationInterval,n=t.bufferTimeSpan,r=t.subscriber,i=t.scheduler,o=r.openContext();r.closed||(r.add(o.closeAction=i.schedule(f,n,{subscriber:r,context:o})),this.schedule(t,e))}function f(t){var e=t.subscriber,n=t.context;e.closeContext(n)}},"./node_modules/rxjs/_esm5/internal/operators/bufferToggle.js":function(t,e,n){"use strict";n.r(e),n.d(e,"bufferToggle",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),a=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js");function s(t,e){return function(n){return n.lift(new u(t,e))}}var u=function(){function t(t,e){this.openings=t,this.closingSelector=e}return t.prototype.call=function(t,e){return e.subscribe(new c(t,this.openings,this.closingSelector))},t}(),c=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.openings=n,i.closingSelector=r,i.contexts=[],i.add(Object(o.subscribeToResult)(i,n)),i}return r.__extends(e,t),e.prototype._next=function(t){for(var e=this.contexts,n=e.length,r=0;r<n;r++)e[r].buffer.push(t)},e.prototype._error=function(e){for(var n=this.contexts;n.length>0;){var r=n.shift();r.subscription.unsubscribe(),r.buffer=null,r.subscription=null}this.contexts=null,t.prototype._error.call(this,e)},e.prototype._complete=function(){for(var e=this.contexts;e.length>0;){var n=e.shift();this.destination.next(n.buffer),n.subscription.unsubscribe(),n.buffer=null,n.subscription=null}this.contexts=null,t.prototype._complete.call(this)},e.prototype.notifyNext=function(t,e,n,r,i){t?this.closeBuffer(t):this.openBuffer(e)},e.prototype.notifyComplete=function(t){this.closeBuffer(t.context)},e.prototype.openBuffer=function(t){try{var e=this.closingSelector.call(this,t);e&&this.trySubscribe(e)}catch(t){this._error(t)}},e.prototype.closeBuffer=function(t){var e=this.contexts;if(e&&t){var n=t.buffer,r=t.subscription;this.destination.next(n),e.splice(e.indexOf(t),1),this.remove(r),r.unsubscribe()}},e.prototype.trySubscribe=function(t){var e=this.contexts,n=new i.Subscription,r={buffer:[],subscription:n};e.push(r);var a=Object(o.subscribeToResult)(this,t,r);!a||a.closed?this.closeBuffer(r):(a.context=r,this.add(a),n.add(a))},e}(a.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/bufferWhen.js":function(t,e,n){"use strict";n.r(e),n.d(e,"bufferWhen",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),o=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),a=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),s=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),u=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function c(t){return function(e){return e.lift(new l(t))}}var l=function(){function t(t){this.closingSelector=t}return t.prototype.call=function(t,e){return e.subscribe(new d(t,this.closingSelector))},t}(),d=function(t){function e(e,n){var r=t.call(this,e)||this;return r.closingSelector=n,r.subscribing=!1,r.openBuffer(),r}return r.__extends(e,t),e.prototype._next=function(t){this.buffer.push(t)},e.prototype._complete=function(){var e=this.buffer;e&&this.destination.next(e),t.prototype._complete.call(this)},e.prototype._unsubscribe=function(){this.buffer=null,this.subscribing=!1},e.prototype.notifyNext=function(t,e,n,r,i){this.openBuffer()},e.prototype.notifyComplete=function(){this.subscribing?this.complete():this.openBuffer()},e.prototype.openBuffer=function(){var t=this.closingSubscription;t&&(this.remove(t),t.unsubscribe());var e=this.buffer;this.buffer&&this.destination.next(e),this.buffer=[];var n=Object(o.tryCatch)(this.closingSelector)();n===a.errorObject?this.error(a.errorObject.e):(t=new i.Subscription,this.closingSubscription=t,this.add(t),this.subscribing=!0,t.add(Object(u.subscribeToResult)(this,n)),this.subscribing=!1)},e}(s.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/catchError.js":function(t,e,n){"use strict";n.r(e),n.d(e,"catchError",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/InnerSubscriber.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function s(t){return function(e){var n=new u(t),r=e.lift(n);return n.caught=r}}var u=function(){function t(t){this.selector=t}return t.prototype.call=function(t,e){return e.subscribe(new c(t,this.selector,this.caught))},t}(),c=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.selector=n,i.caught=r,i}return r.__extends(e,t),e.prototype.error=function(e){if(!this.isStopped){var n=void 0;try{n=this.selector(e,this.caught)}catch(e){return void t.prototype.error.call(this,e)}this._unsubscribeAndRecycle();var r=new o.InnerSubscriber(this,void 0,void 0);this.add(r),Object(a.subscribeToResult)(this,n,void 0,void 0,r)}},e}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/combineAll.js":function(t,e,n){"use strict";n.r(e),n.d(e,"combineAll",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/observable/combineLatest.js");function i(t){return function(e){return e.lift(new r.CombineLatestOperator(t))}}},"./node_modules/rxjs/_esm5/internal/operators/combineLatest.js":function(t,e,n){"use strict";n.r(e),n.d(e,"combineLatest",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/combineLatest.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/from.js");function a(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];var n=null;return"function"==typeof t[t.length-1]&&(n=t.pop()),1===t.length&&Object(r.isArray)(t[0])&&(t=t[0].slice()),function(e){return e.lift.call(Object(o.from)([e].concat(t)),new i.CombineLatestOperator(n))}}},"./node_modules/rxjs/_esm5/internal/operators/concat.js":function(t,e,n){"use strict";n.r(e),n.d(e,"concat",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/observable/concat.js");function i(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return function(e){return e.lift.call(r.concat.apply(void 0,[e].concat(t)))}}},"./node_modules/rxjs/_esm5/internal/operators/concatAll.js":function(t,e,n){"use strict";n.r(e),n.d(e,"concatAll",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/mergeAll.js");function i(){return Object(r.mergeAll)(1)}},"./node_modules/rxjs/_esm5/internal/operators/concatMap.js":function(t,e,n){"use strict";n.r(e),n.d(e,"concatMap",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/mergeMap.js");function i(t,e){return Object(r.mergeMap)(t,e,1)}},"./node_modules/rxjs/_esm5/internal/operators/concatMapTo.js":function(t,e,n){"use strict";n.r(e),n.d(e,"concatMapTo",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/concatMap.js");function i(t,e){return Object(r.concatMap)(function(){return t},e)}},"./node_modules/rxjs/_esm5/internal/operators/count.js":function(t,e,n){"use strict";n.r(e),n.d(e,"count",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(t){return function(e){return e.lift(new a(t,e))}}var a=function(){function t(t,e){this.predicate=t,this.source=e}return t.prototype.call=function(t,e){return e.subscribe(new s(t,this.predicate,this.source))},t}(),s=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.predicate=n,i.source=r,i.count=0,i.index=0,i}return r.__extends(e,t),e.prototype._next=function(t){this.predicate?this._tryPredicate(t):this.count++},e.prototype._tryPredicate=function(t){var e;try{e=this.predicate(t,this.index++,this.source)}catch(t){return void this.destination.error(t)}e&&this.count++},e.prototype._complete=function(){this.destination.next(this.count),this.destination.complete()},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/debounce.js":function(t,e,n){"use strict";n.r(e),n.d(e,"debounce",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function a(t){return function(e){return e.lift(new s(t))}}var s=function(){function t(t){this.durationSelector=t}return t.prototype.call=function(t,e){return e.subscribe(new u(t,this.durationSelector))},t}(),u=function(t){function e(e,n){var r=t.call(this,e)||this;return r.durationSelector=n,r.hasValue=!1,r.durationSubscription=null,r}return r.__extends(e,t),e.prototype._next=function(t){try{var e=this.durationSelector.call(this,t);e&&this._tryNext(t,e)}catch(t){this.destination.error(t)}},e.prototype._complete=function(){this.emitValue(),this.destination.complete()},e.prototype._tryNext=function(t,e){var n=this.durationSubscription;this.value=t,this.hasValue=!0,n&&(n.unsubscribe(),this.remove(n)),(n=Object(o.subscribeToResult)(this,e))&&!n.closed&&this.add(this.durationSubscription=n)},e.prototype.notifyNext=function(t,e,n,r,i){this.emitValue()},e.prototype.notifyComplete=function(){this.emitValue()},e.prototype.emitValue=function(){if(this.hasValue){var e=this.value,n=this.durationSubscription;n&&(this.durationSubscription=null,n.unsubscribe(),this.remove(n)),this.value=null,this.hasValue=!1,t.prototype._next.call(this,e)}},e}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/debounceTime.js":function(t,e,n){"use strict";n.r(e),n.d(e,"debounceTime",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js");function a(t,e){return void 0===e&&(e=o.async),function(n){return n.lift(new s(t,e))}}var s=function(){function t(t,e){this.dueTime=t,this.scheduler=e}return t.prototype.call=function(t,e){return e.subscribe(new u(t,this.dueTime,this.scheduler))},t}(),u=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.dueTime=n,i.scheduler=r,i.debouncedSubscription=null,i.lastValue=null,i.hasValue=!1,i}return r.__extends(e,t),e.prototype._next=function(t){this.clearDebounce(),this.lastValue=t,this.hasValue=!0,this.add(this.debouncedSubscription=this.scheduler.schedule(c,this.dueTime,this))},e.prototype._complete=function(){this.debouncedNext(),this.destination.complete()},e.prototype.debouncedNext=function(){if(this.clearDebounce(),this.hasValue){var t=this.lastValue;this.lastValue=null,this.hasValue=!1,this.destination.next(t)}},e.prototype.clearDebounce=function(){var t=this.debouncedSubscription;null!==t&&(this.remove(t),t.unsubscribe(),this.debouncedSubscription=null)},e}(i.Subscriber);function c(t){t.debouncedNext()}},"./node_modules/rxjs/_esm5/internal/operators/defaultIfEmpty.js":function(t,e,n){"use strict";n.r(e),n.d(e,"defaultIfEmpty",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(t){return void 0===t&&(t=null),function(e){return e.lift(new a(t))}}var a=function(){function t(t){this.defaultValue=t}return t.prototype.call=function(t,e){return e.subscribe(new s(t,this.defaultValue))},t}(),s=function(t){function e(e,n){var r=t.call(this,e)||this;return r.defaultValue=n,r.isEmpty=!0,r}return r.__extends(e,t),e.prototype._next=function(t){this.isEmpty=!1,this.destination.next(t)},e.prototype._complete=function(){this.isEmpty&&this.destination.next(this.defaultValue),this.destination.complete()},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/delay.js":function(t,e,n){"use strict";n.r(e),n.d(e,"delay",function(){return u});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isDate.js"),a=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/Notification.js");function u(t,e){void 0===e&&(e=i.async);var n=Object(o.isDate)(t)?+t-e.now():Math.abs(t);return function(t){return t.lift(new c(n,e))}}var c=function(){function t(t,e){this.delay=t,this.scheduler=e}return t.prototype.call=function(t,e){return e.subscribe(new l(t,this.delay,this.scheduler))},t}(),l=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.delay=n,i.scheduler=r,i.queue=[],i.active=!1,i.errored=!1,i}return r.__extends(e,t),e.dispatch=function(t){for(var e=t.source,n=e.queue,r=t.scheduler,i=t.destination;n.length>0&&n[0].time-r.now()<=0;)n.shift().notification.observe(i);if(n.length>0){var o=Math.max(0,n[0].time-r.now());this.schedule(t,o)}else this.unsubscribe(),e.active=!1},e.prototype._schedule=function(t){this.active=!0,this.destination.add(t.schedule(e.dispatch,this.delay,{source:this,destination:this.destination,scheduler:t}))},e.prototype.scheduleNotification=function(t){if(!0!==this.errored){var e=this.scheduler,n=new d(e.now()+this.delay,t);this.queue.push(n),!1===this.active&&this._schedule(e)}},e.prototype._next=function(t){this.scheduleNotification(s.Notification.createNext(t))},e.prototype._error=function(t){this.errored=!0,this.queue=[],this.destination.error(t),this.unsubscribe()},e.prototype._complete=function(){this.scheduleNotification(s.Notification.createComplete()),this.unsubscribe()},e}(a.Subscriber),d=function(){return function(t,e){this.time=t,this.notification=e}}()},"./node_modules/rxjs/_esm5/internal/operators/delayWhen.js":function(t,e,n){"use strict";n.r(e),n.d(e,"delayWhen",function(){return u});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/Observable.js"),a=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function u(t,e){return e?function(n){return new d(n,e).lift(new c(t))}:function(e){return e.lift(new c(t))}}var c=function(){function t(t){this.delayDurationSelector=t}return t.prototype.call=function(t,e){return e.subscribe(new l(t,this.delayDurationSelector))},t}(),l=function(t){function e(e,n){var r=t.call(this,e)||this;return r.delayDurationSelector=n,r.completed=!1,r.delayNotifierSubscriptions=[],r.index=0,r}return r.__extends(e,t),e.prototype.notifyNext=function(t,e,n,r,i){this.destination.next(t),this.removeSubscription(i),this.tryComplete()},e.prototype.notifyError=function(t,e){this._error(t)},e.prototype.notifyComplete=function(t){var e=this.removeSubscription(t);e&&this.destination.next(e),this.tryComplete()},e.prototype._next=function(t){var e=this.index++;try{var n=this.delayDurationSelector(t,e);n&&this.tryDelay(n,t)}catch(t){this.destination.error(t)}},e.prototype._complete=function(){this.completed=!0,this.tryComplete(),this.unsubscribe()},e.prototype.removeSubscription=function(t){t.unsubscribe();var e=this.delayNotifierSubscriptions.indexOf(t);return-1!==e&&this.delayNotifierSubscriptions.splice(e,1),t.outerValue},e.prototype.tryDelay=function(t,e){var n=Object(s.subscribeToResult)(this,t,e);n&&!n.closed&&(this.destination.add(n),this.delayNotifierSubscriptions.push(n))},e.prototype.tryComplete=function(){this.completed&&0===this.delayNotifierSubscriptions.length&&this.destination.complete()},e}(a.OuterSubscriber),d=function(t){function e(e,n){var r=t.call(this)||this;return r.source=e,r.subscriptionDelay=n,r}return r.__extends(e,t),e.prototype._subscribe=function(t){this.subscriptionDelay.subscribe(new p(t,this.source))},e}(o.Observable),p=function(t){function e(e,n){var r=t.call(this)||this;return r.parent=e,r.source=n,r.sourceSubscribed=!1,r}return r.__extends(e,t),e.prototype._next=function(t){this.subscribeToSource()},e.prototype._error=function(t){this.unsubscribe(),this.parent.error(t)},e.prototype._complete=function(){this.unsubscribe(),this.subscribeToSource()},e.prototype.subscribeToSource=function(){this.sourceSubscribed||(this.sourceSubscribed=!0,this.unsubscribe(),this.source.subscribe(this.parent))},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/dematerialize.js":function(t,e,n){"use strict";n.r(e),n.d(e,"dematerialize",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(){return function(t){return t.lift(new a)}}var a=function(){function t(){}return t.prototype.call=function(t,e){return e.subscribe(new s(t))},t}(),s=function(t){function e(e){return t.call(this,e)||this}return r.__extends(e,t),e.prototype._next=function(t){t.observe(this.destination)},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/distinct.js":function(t,e,n){"use strict";n.r(e),n.d(e,"distinct",function(){return a}),n.d(e,"DistinctSubscriber",function(){return u});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function a(t,e){return function(n){return n.lift(new s(t,e))}}var s=function(){function t(t,e){this.keySelector=t,this.flushes=e}return t.prototype.call=function(t,e){return e.subscribe(new u(t,this.keySelector,this.flushes))},t}(),u=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.keySelector=n,i.values=new Set,r&&i.add(Object(o.subscribeToResult)(i,r)),i}return r.__extends(e,t),e.prototype.notifyNext=function(t,e,n,r,i){this.values.clear()},e.prototype.notifyError=function(t,e){this._error(t)},e.prototype._next=function(t){this.keySelector?this._useKeySelector(t):this._finalizeNext(t,t)},e.prototype._useKeySelector=function(t){var e,n=this.destination;try{e=this.keySelector(t)}catch(t){return void n.error(t)}this._finalizeNext(e,t)},e.prototype._finalizeNext=function(t,e){var n=this.values;n.has(t)||(n.add(t),this.destination.next(e))},e}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/distinctUntilChanged.js":function(t,e,n){"use strict";n.r(e),n.d(e,"distinctUntilChanged",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),a=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js");function s(t,e){return function(n){return n.lift(new u(t,e))}}var u=function(){function t(t,e){this.compare=t,this.keySelector=e}return t.prototype.call=function(t,e){return e.subscribe(new c(t,this.compare,this.keySelector))},t}(),c=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.keySelector=r,i.hasKey=!1,"function"==typeof n&&(i.compare=n),i}return r.__extends(e,t),e.prototype.compare=function(t,e){return t===e},e.prototype._next=function(t){var e=t;if(this.keySelector&&(e=Object(o.tryCatch)(this.keySelector)(t))===a.errorObject)return this.destination.error(a.errorObject.e);var n=!1;if(this.hasKey){if((n=Object(o.tryCatch)(this.compare)(this.key,e))===a.errorObject)return this.destination.error(a.errorObject.e)}else this.hasKey=!0;!1===Boolean(n)&&(this.key=e,this.destination.next(t))},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/distinctUntilKeyChanged.js":function(t,e,n){"use strict";n.r(e),n.d(e,"distinctUntilKeyChanged",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/distinctUntilChanged.js");function i(t,e){return Object(r.distinctUntilChanged)(function(n,r){return e?e(n[t],r[t]):n[t]===r[t]})}},"./node_modules/rxjs/_esm5/internal/operators/elementAt.js":function(t,e,n){"use strict";n.r(e),n.d(e,"elementAt",function(){return u});var r=n("./node_modules/rxjs/_esm5/internal/util/ArgumentOutOfRangeError.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/filter.js"),o=n("./node_modules/rxjs/_esm5/internal/operators/throwIfEmpty.js"),a=n("./node_modules/rxjs/_esm5/internal/operators/defaultIfEmpty.js"),s=n("./node_modules/rxjs/_esm5/internal/operators/take.js");function u(t,e){if(t<0)throw new r.ArgumentOutOfRangeError;var n=arguments.length>=2;return function(u){return u.pipe(Object(i.filter)(function(e,n){return n===t}),Object(s.take)(1),n?Object(a.defaultIfEmpty)(e):Object(o.throwIfEmpty)(function(){return new r.ArgumentOutOfRangeError}))}}},"./node_modules/rxjs/_esm5/internal/operators/endWith.js":function(t,e,n){"use strict";n.r(e),n.d(e,"endWith",function(){return u});var r=n("./node_modules/rxjs/_esm5/internal/observable/fromArray.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/scalar.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/empty.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/concat.js"),s=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js");function u(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return function(e){var n=t[t.length-1];Object(s.isScheduler)(n)?t.pop():n=null;var u=t.length;return 1!==u||n?u>0?Object(a.concat)(e,Object(r.fromArray)(t,n)):Object(a.concat)(e,Object(o.empty)(n)):Object(a.concat)(e,Object(i.scalar)(t[0]))}}},"./node_modules/rxjs/_esm5/internal/operators/every.js":function(t,e,n){"use strict";n.r(e),n.d(e,"every",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(t,e){return function(n){return n.lift(new a(t,e,n))}}var a=function(){function t(t,e,n){this.predicate=t,this.thisArg=e,this.source=n}return t.prototype.call=function(t,e){return e.subscribe(new s(t,this.predicate,this.thisArg,this.source))},t}(),s=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.predicate=n,o.thisArg=r,o.source=i,o.index=0,o.thisArg=r||o,o}return r.__extends(e,t),e.prototype.notifyComplete=function(t){this.destination.next(t),this.destination.complete()},e.prototype._next=function(t){var e=!1;try{e=this.predicate.call(this.thisArg,t,this.index++,this.source)}catch(t){return void this.destination.error(t)}e||this.notifyComplete(!1)},e.prototype._complete=function(){this.notifyComplete(!0)},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/exhaust.js":function(t,e,n){"use strict";n.r(e),n.d(e,"exhaust",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function a(){return function(t){return t.lift(new s)}}var s=function(){function t(){}return t.prototype.call=function(t,e){return e.subscribe(new u(t))},t}(),u=function(t){function e(e){var n=t.call(this,e)||this;return n.hasCompleted=!1,n.hasSubscription=!1,n}return r.__extends(e,t),e.prototype._next=function(t){this.hasSubscription||(this.hasSubscription=!0,this.add(Object(o.subscribeToResult)(this,t)))},e.prototype._complete=function(){this.hasCompleted=!0,this.hasSubscription||this.destination.complete()},e.prototype.notifyComplete=function(t){this.remove(t),this.hasSubscription=!1,this.hasCompleted&&this.destination.complete()},e}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/exhaustMap.js":function(t,e,n){"use strict";n.r(e),n.d(e,"exhaustMap",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/InnerSubscriber.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),s=n("./node_modules/rxjs/_esm5/internal/operators/map.js"),u=n("./node_modules/rxjs/_esm5/internal/observable/from.js");function c(t,e){return e?function(n){return n.pipe(c(function(n,r){return Object(u.from)(t(n,r)).pipe(Object(s.map)(function(t,i){return e(n,t,r,i)}))}))}:function(e){return e.lift(new l(t))}}var l=function(){function t(t){this.project=t}return t.prototype.call=function(t,e){return e.subscribe(new d(t,this.project))},t}(),d=function(t){function e(e,n){var r=t.call(this,e)||this;return r.project=n,r.hasSubscription=!1,r.hasCompleted=!1,r.index=0,r}return r.__extends(e,t),e.prototype._next=function(t){this.hasSubscription||this.tryNext(t)},e.prototype.tryNext=function(t){var e,n=this.index++;try{e=this.project(t,n)}catch(t){return void this.destination.error(t)}this.hasSubscription=!0,this._innerSub(e,t,n)},e.prototype._innerSub=function(t,e,n){var r=new o.InnerSubscriber(this,void 0,void 0);this.destination.add(r),Object(a.subscribeToResult)(this,t,e,n,r)},e.prototype._complete=function(){this.hasCompleted=!0,this.hasSubscription||this.destination.complete(),this.unsubscribe()},e.prototype.notifyNext=function(t,e,n,r,i){this.destination.next(e)},e.prototype.notifyError=function(t){this.destination.error(t)},e.prototype.notifyComplete=function(t){this.destination.remove(t),this.hasSubscription=!1,this.hasCompleted&&this.destination.complete()},e}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/expand.js":function(t,e,n){"use strict";n.r(e),n.d(e,"expand",function(){return u}),n.d(e,"ExpandOperator",function(){return c}),n.d(e,"ExpandSubscriber",function(){return l});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),o=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),a=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function u(t,e,n){return void 0===e&&(e=Number.POSITIVE_INFINITY),void 0===n&&(n=void 0),e=(e||0)<1?Number.POSITIVE_INFINITY:e,function(r){return r.lift(new c(t,e,n))}}var c=function(){function t(t,e,n){this.project=t,this.concurrent=e,this.scheduler=n}return t.prototype.call=function(t,e){return e.subscribe(new l(t,this.project,this.concurrent,this.scheduler))},t}(),l=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.project=n,o.concurrent=r,o.scheduler=i,o.index=0,o.active=0,o.hasCompleted=!1,r<Number.POSITIVE_INFINITY&&(o.buffer=[]),o}return r.__extends(e,t),e.dispatch=function(t){var e=t.subscriber,n=t.result,r=t.value,i=t.index;e.subscribeToProjection(n,r,i)},e.prototype._next=function(t){var n=this.destination;if(n.closed)this._complete();else{var r=this.index++;if(this.active<this.concurrent){n.next(t);var a=Object(i.tryCatch)(this.project)(t,r);if(a===o.errorObject)n.error(o.errorObject.e);else if(this.scheduler){var s={subscriber:this,result:a,value:t,index:r};this.destination.add(this.scheduler.schedule(e.dispatch,0,s))}else this.subscribeToProjection(a,t,r)}else this.buffer.push(t)}},e.prototype.subscribeToProjection=function(t,e,n){this.active++,this.destination.add(Object(s.subscribeToResult)(this,t,e,n))},e.prototype._complete=function(){this.hasCompleted=!0,this.hasCompleted&&0===this.active&&this.destination.complete(),this.unsubscribe()},e.prototype.notifyNext=function(t,e,n,r,i){this._next(e)},e.prototype.notifyComplete=function(t){var e=this.buffer;this.destination.remove(t),this.active--,e&&e.length>0&&this._next(e.shift()),this.hasCompleted&&0===this.active&&this.destination.complete()},e}(a.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/filter.js":function(t,e,n){"use strict";n.r(e),n.d(e,"filter",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(t,e){return function(n){return n.lift(new a(t,e))}}var a=function(){function t(t,e){this.predicate=t,this.thisArg=e}return t.prototype.call=function(t,e){return e.subscribe(new s(t,this.predicate,this.thisArg))},t}(),s=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.predicate=n,i.thisArg=r,i.count=0,i}return r.__extends(e,t),e.prototype._next=function(t){var e;try{e=this.predicate.call(this.thisArg,t,this.count++)}catch(t){return void this.destination.error(t)}e&&this.destination.next(t)},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/finalize.js":function(t,e,n){"use strict";n.r(e),n.d(e,"finalize",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/Subscription.js");function a(t){return function(e){return e.lift(new s(t))}}var s=function(){function t(t){this.callback=t}return t.prototype.call=function(t,e){return e.subscribe(new u(t,this.callback))},t}(),u=function(t){function e(e,n){var r=t.call(this,e)||this;return r.add(new o.Subscription(n)),r}return r.__extends(e,t),e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/find.js":function(t,e,n){"use strict";n.r(e),n.d(e,"find",function(){return o}),n.d(e,"FindValueOperator",function(){return a}),n.d(e,"FindValueSubscriber",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(t,e){if("function"!=typeof t)throw new TypeError("predicate is not a function");return function(n){return n.lift(new a(t,n,!1,e))}}var a=function(){function t(t,e,n,r){this.predicate=t,this.source=e,this.yieldIndex=n,this.thisArg=r}return t.prototype.call=function(t,e){return e.subscribe(new s(t,this.predicate,this.source,this.yieldIndex,this.thisArg))},t}(),s=function(t){function e(e,n,r,i,o){var a=t.call(this,e)||this;return a.predicate=n,a.source=r,a.yieldIndex=i,a.thisArg=o,a.index=0,a}return r.__extends(e,t),e.prototype.notifyComplete=function(t){var e=this.destination;e.next(t),e.complete(),this.unsubscribe()},e.prototype._next=function(t){var e=this.predicate,n=this.thisArg,r=this.index++;try{e.call(n||this,t,r,this.source)&&this.notifyComplete(this.yieldIndex?r:t)}catch(t){this.destination.error(t)}},e.prototype._complete=function(){this.notifyComplete(this.yieldIndex?-1:void 0)},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/findIndex.js":function(t,e,n){"use strict";n.r(e),n.d(e,"findIndex",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/find.js");function i(t,e){return function(n){return n.lift(new r.FindValueOperator(t,n,!0,e))}}},"./node_modules/rxjs/_esm5/internal/operators/first.js":function(t,e,n){"use strict";n.r(e),n.d(e,"first",function(){return c});var r=n("./node_modules/rxjs/_esm5/internal/util/EmptyError.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/filter.js"),o=n("./node_modules/rxjs/_esm5/internal/operators/take.js"),a=n("./node_modules/rxjs/_esm5/internal/operators/defaultIfEmpty.js"),s=n("./node_modules/rxjs/_esm5/internal/operators/throwIfEmpty.js"),u=n("./node_modules/rxjs/_esm5/internal/util/identity.js");function c(t,e){var n=arguments.length>=2;return function(c){return c.pipe(t?Object(i.filter)(function(e,n){return t(e,n,c)}):u.identity,Object(o.take)(1),n?Object(a.defaultIfEmpty)(e):Object(s.throwIfEmpty)(function(){return new r.EmptyError}))}}},"./node_modules/rxjs/_esm5/internal/operators/groupBy.js":function(t,e,n){"use strict";n.r(e),n.d(e,"groupBy",function(){return u}),n.d(e,"GroupedObservable",function(){return p});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),a=n("./node_modules/rxjs/_esm5/internal/Observable.js"),s=n("./node_modules/rxjs/_esm5/internal/Subject.js");function u(t,e,n,r){return function(i){return i.lift(new c(t,e,n,r))}}var c=function(){function t(t,e,n,r){this.keySelector=t,this.elementSelector=e,this.durationSelector=n,this.subjectSelector=r}return t.prototype.call=function(t,e){return e.subscribe(new l(t,this.keySelector,this.elementSelector,this.durationSelector,this.subjectSelector))},t}(),l=function(t){function e(e,n,r,i,o){var a=t.call(this,e)||this;return a.keySelector=n,a.elementSelector=r,a.durationSelector=i,a.subjectSelector=o,a.groups=null,a.attemptedToUnsubscribe=!1,a.count=0,a}return r.__extends(e,t),e.prototype._next=function(t){var e;try{e=this.keySelector(t)}catch(t){return void this.error(t)}this._group(t,e)},e.prototype._group=function(t,e){var n=this.groups;n||(n=this.groups=new Map);var r,i=n.get(e);if(this.elementSelector)try{r=this.elementSelector(t)}catch(t){this.error(t)}else r=t;if(!i){i=this.subjectSelector?this.subjectSelector():new s.Subject,n.set(e,i);var o=new p(e,i,this);if(this.destination.next(o),this.durationSelector){var a=void 0;try{a=this.durationSelector(new p(e,i))}catch(t){return void this.error(t)}this.add(a.subscribe(new d(e,i,this)))}}i.closed||i.next(r)},e.prototype._error=function(t){var e=this.groups;e&&(e.forEach(function(e,n){e.error(t)}),e.clear()),this.destination.error(t)},e.prototype._complete=function(){var t=this.groups;t&&(t.forEach(function(t,e){t.complete()}),t.clear()),this.destination.complete()},e.prototype.removeGroup=function(t){this.groups.delete(t)},e.prototype.unsubscribe=function(){this.closed||(this.attemptedToUnsubscribe=!0,0===this.count&&t.prototype.unsubscribe.call(this))},e}(i.Subscriber),d=function(t){function e(e,n,r){var i=t.call(this,n)||this;return i.key=e,i.group=n,i.parent=r,i}return r.__extends(e,t),e.prototype._next=function(t){this.complete()},e.prototype._unsubscribe=function(){var t=this.parent,e=this.key;this.key=this.parent=null,t&&t.removeGroup(e)},e}(i.Subscriber),p=function(t){function e(e,n,r){var i=t.call(this)||this;return i.key=e,i.groupSubject=n,i.refCountSubscription=r,i}return r.__extends(e,t),e.prototype._subscribe=function(t){var e=new o.Subscription,n=this.refCountSubscription,r=this.groupSubject;return n&&!n.closed&&e.add(new f(n)),e.add(r.subscribe(t)),e},e}(a.Observable),f=function(t){function e(e){var n=t.call(this)||this;return n.parent=e,e.count++,n}return r.__extends(e,t),e.prototype.unsubscribe=function(){var e=this.parent;e.closed||this.closed||(t.prototype.unsubscribe.call(this),e.count-=1,0===e.count&&e.attemptedToUnsubscribe&&e.unsubscribe())},e}(o.Subscription)},"./node_modules/rxjs/_esm5/internal/operators/ignoreElements.js":function(t,e,n){"use strict";n.r(e),n.d(e,"ignoreElements",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(){return function(t){return t.lift(new a)}}var a=function(){function t(){}return t.prototype.call=function(t,e){return e.subscribe(new s(t))},t}(),s=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return r.__extends(e,t),e.prototype._next=function(t){},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/isEmpty.js":function(t,e,n){"use strict";n.r(e),n.d(e,"isEmpty",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(){return function(t){return t.lift(new a)}}var a=function(){function t(){}return t.prototype.call=function(t,e){return e.subscribe(new s(t))},t}(),s=function(t){function e(e){return t.call(this,e)||this}return r.__extends(e,t),e.prototype.notifyComplete=function(t){var e=this.destination;e.next(t),e.complete()},e.prototype._next=function(t){this.notifyComplete(!1)},e.prototype._complete=function(){this.notifyComplete(!0)},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/last.js":function(t,e,n){"use strict";n.r(e),n.d(e,"last",function(){return c});var r=n("./node_modules/rxjs/_esm5/internal/util/EmptyError.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/filter.js"),o=n("./node_modules/rxjs/_esm5/internal/operators/takeLast.js"),a=n("./node_modules/rxjs/_esm5/internal/operators/throwIfEmpty.js"),s=n("./node_modules/rxjs/_esm5/internal/operators/defaultIfEmpty.js"),u=n("./node_modules/rxjs/_esm5/internal/util/identity.js");function c(t,e){var n=arguments.length>=2;return function(c){return c.pipe(t?Object(i.filter)(function(e,n){return t(e,n,c)}):u.identity,Object(o.takeLast)(1),n?Object(s.defaultIfEmpty)(e):Object(a.throwIfEmpty)(function(){return new r.EmptyError}))}}},"./node_modules/rxjs/_esm5/internal/operators/map.js":function(t,e,n){"use strict";n.r(e),n.d(e,"map",function(){return o}),n.d(e,"MapOperator",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(t,e){return function(n){if("function"!=typeof t)throw new TypeError("argument is not a function. Are you looking for `mapTo()`?");return n.lift(new a(t,e))}}var a=function(){function t(t,e){this.project=t,this.thisArg=e}return t.prototype.call=function(t,e){return e.subscribe(new s(t,this.project,this.thisArg))},t}(),s=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.project=n,i.count=0,i.thisArg=r||i,i}return r.__extends(e,t),e.prototype._next=function(t){var e;try{e=this.project.call(this.thisArg,t,this.count++)}catch(t){return void this.destination.error(t)}this.destination.next(e)},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/mapTo.js":function(t,e,n){"use strict";n.r(e),n.d(e,"mapTo",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(t){return function(e){return e.lift(new a(t))}}var a=function(){function t(t){this.value=t}return t.prototype.call=function(t,e){return e.subscribe(new s(t,this.value))},t}(),s=function(t){function e(e,n){var r=t.call(this,e)||this;return r.value=n,r}return r.__extends(e,t),e.prototype._next=function(t){this.destination.next(this.value)},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/materialize.js":function(t,e,n){"use strict";n.r(e),n.d(e,"materialize",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/Notification.js");function a(){return function(t){return t.lift(new s)}}var s=function(){function t(){}return t.prototype.call=function(t,e){return e.subscribe(new u(t))},t}(),u=function(t){function e(e){return t.call(this,e)||this}return r.__extends(e,t),e.prototype._next=function(t){this.destination.next(o.Notification.createNext(t))},e.prototype._error=function(t){var e=this.destination;e.next(o.Notification.createError(t)),e.complete()},e.prototype._complete=function(){var t=this.destination;t.next(o.Notification.createComplete()),t.complete()},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/max.js":function(t,e,n){"use strict";n.r(e),n.d(e,"max",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/reduce.js");function i(t){var e="function"==typeof t?function(e,n){return t(e,n)>0?e:n}:function(t,e){return t>e?t:e};return Object(r.reduce)(e)}},"./node_modules/rxjs/_esm5/internal/operators/merge.js":function(t,e,n){"use strict";n.r(e),n.d(e,"merge",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/observable/merge.js");function i(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return function(e){return e.lift.call(r.merge.apply(void 0,[e].concat(t)))}}},"./node_modules/rxjs/_esm5/internal/operators/mergeAll.js":function(t,e,n){"use strict";n.r(e),n.d(e,"mergeAll",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/operators/mergeMap.js"),i=n("./node_modules/rxjs/_esm5/internal/util/identity.js");function o(t){return void 0===t&&(t=Number.POSITIVE_INFINITY),Object(r.mergeMap)(i.identity,t)}},"./node_modules/rxjs/_esm5/internal/operators/mergeMap.js":function(t,e,n){"use strict";n.r(e),n.d(e,"mergeMap",function(){return c}),n.d(e,"MergeMapOperator",function(){return l}),n.d(e,"MergeMapSubscriber",function(){return d});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),o=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),a=n("./node_modules/rxjs/_esm5/internal/InnerSubscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/operators/map.js"),u=n("./node_modules/rxjs/_esm5/internal/observable/from.js");function c(t,e,n){return void 0===n&&(n=Number.POSITIVE_INFINITY),"function"==typeof e?function(r){return r.pipe(c(function(n,r){return Object(u.from)(t(n,r)).pipe(Object(s.map)(function(t,i){return e(n,t,r,i)}))},n))}:("number"==typeof e&&(n=e),function(e){return e.lift(new l(t,n))})}var l=function(){function t(t,e){void 0===e&&(e=Number.POSITIVE_INFINITY),this.project=t,this.concurrent=e}return t.prototype.call=function(t,e){return e.subscribe(new d(t,this.project,this.concurrent))},t}(),d=function(t){function e(e,n,r){void 0===r&&(r=Number.POSITIVE_INFINITY);var i=t.call(this,e)||this;return i.project=n,i.concurrent=r,i.hasCompleted=!1,i.buffer=[],i.active=0,i.index=0,i}return r.__extends(e,t),e.prototype._next=function(t){this.active<this.concurrent?this._tryNext(t):this.buffer.push(t)},e.prototype._tryNext=function(t){var e,n=this.index++;try{e=this.project(t,n)}catch(t){return void this.destination.error(t)}this.active++,this._innerSub(e,t,n)},e.prototype._innerSub=function(t,e,n){var r=new a.InnerSubscriber(this,void 0,void 0);this.destination.add(r),Object(i.subscribeToResult)(this,t,e,n,r)},e.prototype._complete=function(){this.hasCompleted=!0,0===this.active&&0===this.buffer.length&&this.destination.complete(),this.unsubscribe()},e.prototype.notifyNext=function(t,e,n,r,i){this.destination.next(e)},e.prototype.notifyComplete=function(t){var e=this.buffer;this.remove(t),this.active--,e.length>0?this._next(e.shift()):0===this.active&&this.hasCompleted&&this.destination.complete()},e}(o.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/mergeMapTo.js":function(t,e,n){"use strict";n.r(e),n.d(e,"mergeMapTo",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/mergeMap.js");function i(t,e,n){return void 0===n&&(n=Number.POSITIVE_INFINITY),"function"==typeof e?Object(r.mergeMap)(function(){return t},e,n):("number"==typeof e&&(n=e),Object(r.mergeMap)(function(){return t},n))}},"./node_modules/rxjs/_esm5/internal/operators/mergeScan.js":function(t,e,n){"use strict";n.r(e),n.d(e,"mergeScan",function(){return c}),n.d(e,"MergeScanOperator",function(){return l}),n.d(e,"MergeScanSubscriber",function(){return d});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),o=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),s=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),u=n("./node_modules/rxjs/_esm5/internal/InnerSubscriber.js");function c(t,e,n){return void 0===n&&(n=Number.POSITIVE_INFINITY),function(r){return r.lift(new l(t,e,n))}}var l=function(){function t(t,e,n){this.accumulator=t,this.seed=e,this.concurrent=n}return t.prototype.call=function(t,e){return e.subscribe(new d(t,this.accumulator,this.seed,this.concurrent))},t}(),d=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.accumulator=n,o.acc=r,o.concurrent=i,o.hasValue=!1,o.hasCompleted=!1,o.buffer=[],o.active=0,o.index=0,o}return r.__extends(e,t),e.prototype._next=function(t){if(this.active<this.concurrent){var e=this.index++,n=Object(i.tryCatch)(this.accumulator)(this.acc,t),r=this.destination;n===o.errorObject?r.error(o.errorObject.e):(this.active++,this._innerSub(n,t,e))}else this.buffer.push(t)},e.prototype._innerSub=function(t,e,n){var r=new u.InnerSubscriber(this,void 0,void 0);this.destination.add(r),Object(a.subscribeToResult)(this,t,e,n,r)},e.prototype._complete=function(){this.hasCompleted=!0,0===this.active&&0===this.buffer.length&&(!1===this.hasValue&&this.destination.next(this.acc),this.destination.complete()),this.unsubscribe()},e.prototype.notifyNext=function(t,e,n,r,i){var o=this.destination;this.acc=e,this.hasValue=!0,o.next(e)},e.prototype.notifyComplete=function(t){var e=this.buffer;this.destination.remove(t),this.active--,e.length>0?this._next(e.shift()):0===this.active&&this.hasCompleted&&(!1===this.hasValue&&this.destination.next(this.acc),this.destination.complete())},e}(s.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/min.js":function(t,e,n){"use strict";n.r(e),n.d(e,"min",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/reduce.js");function i(t){var e="function"==typeof t?function(e,n){return t(e,n)<0?e:n}:function(t,e){return t<e?t:e};return Object(r.reduce)(e)}},"./node_modules/rxjs/_esm5/internal/operators/multicast.js":function(t,e,n){"use strict";n.r(e),n.d(e,"multicast",function(){return i}),n.d(e,"MulticastOperator",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/observable/ConnectableObservable.js");function i(t,e){return function(n){var i;if(i="function"==typeof t?t:function(){return t},"function"==typeof e)return n.lift(new o(i,e));var a=Object.create(n,r.connectableObservableDescriptor);return a.source=n,a.subjectFactory=i,a}}var o=function(){function t(t,e){this.subjectFactory=t,this.selector=e}return t.prototype.call=function(t,e){var n=this.selector,r=this.subjectFactory(),i=n(r).subscribe(t);return i.add(e.subscribe(r)),i},t}()},"./node_modules/rxjs/_esm5/internal/operators/observeOn.js":function(t,e,n){"use strict";n.r(e),n.d(e,"observeOn",function(){return a}),n.d(e,"ObserveOnOperator",function(){return s}),n.d(e,"ObserveOnSubscriber",function(){return u}),n.d(e,"ObserveOnMessage",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/Notification.js");function a(t,e){return void 0===e&&(e=0),function(n){return n.lift(new s(t,e))}}var s=function(){function t(t,e){void 0===e&&(e=0),this.scheduler=t,this.delay=e}return t.prototype.call=function(t,e){return e.subscribe(new u(t,this.scheduler,this.delay))},t}(),u=function(t){function e(e,n,r){void 0===r&&(r=0);var i=t.call(this,e)||this;return i.scheduler=n,i.delay=r,i}return r.__extends(e,t),e.dispatch=function(t){var e=t.notification,n=t.destination;e.observe(n),this.unsubscribe()},e.prototype.scheduleMessage=function(t){this.destination.add(this.scheduler.schedule(e.dispatch,this.delay,new c(t,this.destination)))},e.prototype._next=function(t){this.scheduleMessage(o.Notification.createNext(t))},e.prototype._error=function(t){this.scheduleMessage(o.Notification.createError(t)),this.unsubscribe()},e.prototype._complete=function(){this.scheduleMessage(o.Notification.createComplete()),this.unsubscribe()},e}(i.Subscriber),c=function(){return function(t,e){this.notification=t,this.destination=e}}()},"./node_modules/rxjs/_esm5/internal/operators/onErrorResumeNext.js":function(t,e,n){"use strict";n.r(e),n.d(e,"onErrorResumeNext",function(){return c}),n.d(e,"onErrorResumeNextStatic",function(){return l});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/from.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),a=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/InnerSubscriber.js"),u=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function c(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return 1===t.length&&Object(o.isArray)(t[0])&&(t=t[0]),function(e){return e.lift(new d(t))}}function l(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];var n;return 1===t.length&&Object(o.isArray)(t[0])&&(t=t[0]),n=t.shift(),Object(i.from)(n,null).lift(new d(t))}var d=function(){function t(t){this.nextSources=t}return t.prototype.call=function(t,e){return e.subscribe(new p(t,this.nextSources))},t}(),p=function(t){function e(e,n){var r=t.call(this,e)||this;return r.destination=e,r.nextSources=n,r}return r.__extends(e,t),e.prototype.notifyError=function(t,e){this.subscribeToNextSource()},e.prototype.notifyComplete=function(t){this.subscribeToNextSource()},e.prototype._error=function(t){this.subscribeToNextSource(),this.unsubscribe()},e.prototype._complete=function(){this.subscribeToNextSource(),this.unsubscribe()},e.prototype.subscribeToNextSource=function(){var t=this.nextSources.shift();if(t){var e=new s.InnerSubscriber(this,void 0,void 0);this.destination.add(e),Object(u.subscribeToResult)(this,t,void 0,void 0,e)}else this.destination.complete()},e}(a.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/pairwise.js":function(t,e,n){"use strict";n.r(e),n.d(e,"pairwise",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(){return function(t){return t.lift(new a)}}var a=function(){function t(){}return t.prototype.call=function(t,e){return e.subscribe(new s(t))},t}(),s=function(t){function e(e){var n=t.call(this,e)||this;return n.hasPrev=!1,n}return r.__extends(e,t),e.prototype._next=function(t){this.hasPrev?this.destination.next([this.prev,t]):this.hasPrev=!0,this.prev=t},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/partition.js":function(t,e,n){"use strict";n.r(e),n.d(e,"partition",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/util/not.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/filter.js");function o(t,e){return function(n){return[Object(i.filter)(t,e)(n),Object(i.filter)(Object(r.not)(t,e))(n)]}}},"./node_modules/rxjs/_esm5/internal/operators/pluck.js":function(t,e,n){"use strict";n.r(e),n.d(e,"pluck",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/map.js");function i(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];var n=t.length;if(0===n)throw new Error("list of properties cannot be empty.");return function(e){return Object(r.map)((i=t,o=n,function(t){for(var e=t,n=0;n<o;n++){var r=e[i[n]];if(void 0===r)return;e=r}return e}))(e);var i,o}}},"./node_modules/rxjs/_esm5/internal/operators/publish.js":function(t,e,n){"use strict";n.r(e),n.d(e,"publish",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/Subject.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/multicast.js");function o(t){return t?Object(i.multicast)(function(){return new r.Subject},t):Object(i.multicast)(new r.Subject)}},"./node_modules/rxjs/_esm5/internal/operators/publishBehavior.js":function(t,e,n){"use strict";n.r(e),n.d(e,"publishBehavior",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/BehaviorSubject.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/multicast.js");function o(t){return function(e){return Object(i.multicast)(new r.BehaviorSubject(t))(e)}}},"./node_modules/rxjs/_esm5/internal/operators/publishLast.js":function(t,e,n){"use strict";n.r(e),n.d(e,"publishLast",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/AsyncSubject.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/multicast.js");function o(){return function(t){return Object(i.multicast)(new r.AsyncSubject)(t)}}},"./node_modules/rxjs/_esm5/internal/operators/publishReplay.js":function(t,e,n){"use strict";n.r(e),n.d(e,"publishReplay",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/ReplaySubject.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/multicast.js");function o(t,e,n,o){n&&"function"!=typeof n&&(o=n);var a="function"==typeof n?n:void 0,s=new r.ReplaySubject(t,e,o);return function(t){return Object(i.multicast)(function(){return s},a)(t)}}},"./node_modules/rxjs/_esm5/internal/operators/race.js":function(t,e,n){"use strict";n.r(e),n.d(e,"race",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/race.js");function o(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return function(e){return 1===t.length&&Object(r.isArray)(t[0])&&(t=t[0]),e.lift.call(i.race.apply(void 0,[e].concat(t)))}}},"./node_modules/rxjs/_esm5/internal/operators/reduce.js":function(t,e,n){"use strict";n.r(e),n.d(e,"reduce",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/operators/scan.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/takeLast.js"),o=n("./node_modules/rxjs/_esm5/internal/operators/defaultIfEmpty.js"),a=n("./node_modules/rxjs/_esm5/internal/util/pipe.js");function s(t,e){return arguments.length>=2?function(n){return Object(a.pipe)(Object(r.scan)(t,e),Object(i.takeLast)(1),Object(o.defaultIfEmpty)(e))(n)}:function(e){return Object(a.pipe)(Object(r.scan)(function(e,n,r){return t(e,n,r+1)}),Object(i.takeLast)(1))(e)}}},"./node_modules/rxjs/_esm5/internal/operators/refCount.js":function(t,e,n){"use strict";n.r(e),n.d(e,"refCount",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(){return function(t){return t.lift(new a(t))}}var a=function(){function t(t){this.connectable=t}return t.prototype.call=function(t,e){var n=this.connectable;n._refCount++;var r=new s(t,n),i=e.subscribe(r);return r.closed||(r.connection=n.connect()),i},t}(),s=function(t){function e(e,n){var r=t.call(this,e)||this;return r.connectable=n,r}return r.__extends(e,t),e.prototype._unsubscribe=function(){var t=this.connectable;if(t){this.connectable=null;var e=t._refCount;if(e<=0)this.connection=null;else if(t._refCount=e-1,e>1)this.connection=null;else{var n=this.connection,r=t._connection;this.connection=null,!r||n&&r!==n||r.unsubscribe()}}else this.connection=null},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/repeat.js":function(t,e,n){"use strict";n.r(e),n.d(e,"repeat",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/empty.js");function a(t){return void 0===t&&(t=-1),function(e){return 0===t?Object(o.empty)():t<0?e.lift(new s(-1,e)):e.lift(new s(t-1,e))}}var s=function(){function t(t,e){this.count=t,this.source=e}return t.prototype.call=function(t,e){return e.subscribe(new u(t,this.count,this.source))},t}(),u=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.count=n,i.source=r,i}return r.__extends(e,t),e.prototype.complete=function(){if(!this.isStopped){var e=this.source,n=this.count;if(0===n)return t.prototype.complete.call(this);n>-1&&(this.count=n-1),e.subscribe(this._unsubscribeAndRecycle())}},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/repeatWhen.js":function(t,e,n){"use strict";n.r(e),n.d(e,"repeatWhen",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),a=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),s=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),u=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function c(t){return function(e){return e.lift(new l(t))}}var l=function(){function t(t){this.notifier=t}return t.prototype.call=function(t,e){return e.subscribe(new d(t,this.notifier,e))},t}(),d=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.notifier=n,i.source=r,i.sourceIsBeingSubscribedTo=!0,i}return r.__extends(e,t),e.prototype.notifyNext=function(t,e,n,r,i){this.sourceIsBeingSubscribedTo=!0,this.source.subscribe(this)},e.prototype.notifyComplete=function(e){if(!1===this.sourceIsBeingSubscribedTo)return t.prototype.complete.call(this)},e.prototype.complete=function(){if(this.sourceIsBeingSubscribedTo=!1,!this.isStopped){if(this.retries||this.subscribeToRetries(),!this.retriesSubscription||this.retriesSubscription.closed)return t.prototype.complete.call(this);this._unsubscribeAndRecycle(),this.notifications.next()}},e.prototype._unsubscribe=function(){var t=this.notifications,e=this.retriesSubscription;t&&(t.unsubscribe(),this.notifications=null),e&&(e.unsubscribe(),this.retriesSubscription=null),this.retries=null},e.prototype._unsubscribeAndRecycle=function(){var e=this._unsubscribe;return this._unsubscribe=null,t.prototype._unsubscribeAndRecycle.call(this),this._unsubscribe=e,this},e.prototype.subscribeToRetries=function(){this.notifications=new i.Subject;var e=Object(o.tryCatch)(this.notifier)(this.notifications);if(e===a.errorObject)return t.prototype.complete.call(this);this.retries=e,this.retriesSubscription=Object(u.subscribeToResult)(this,e)},e}(s.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/retry.js":function(t,e,n){"use strict";n.r(e),n.d(e,"retry",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(t){return void 0===t&&(t=-1),function(e){return e.lift(new a(t,e))}}var a=function(){function t(t,e){this.count=t,this.source=e}return t.prototype.call=function(t,e){return e.subscribe(new s(t,this.count,this.source))},t}(),s=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.count=n,i.source=r,i}return r.__extends(e,t),e.prototype.error=function(e){if(!this.isStopped){var n=this.source,r=this.count;if(0===r)return t.prototype.error.call(this,e);r>-1&&(this.count=r-1),n.subscribe(this._unsubscribeAndRecycle())}},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/retryWhen.js":function(t,e,n){"use strict";n.r(e),n.d(e,"retryWhen",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),a=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),s=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),u=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function c(t){return function(e){return e.lift(new l(t,e))}}var l=function(){function t(t,e){this.notifier=t,this.source=e}return t.prototype.call=function(t,e){return e.subscribe(new d(t,this.notifier,this.source))},t}(),d=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.notifier=n,i.source=r,i}return r.__extends(e,t),e.prototype.error=function(e){if(!this.isStopped){var n=this.errors,r=this.retries,s=this.retriesSubscription;if(r)this.errors=null,this.retriesSubscription=null;else{if(n=new i.Subject,(r=Object(o.tryCatch)(this.notifier)(n))===a.errorObject)return t.prototype.error.call(this,a.errorObject.e);s=Object(u.subscribeToResult)(this,r)}this._unsubscribeAndRecycle(),this.errors=n,this.retries=r,this.retriesSubscription=s,n.next(e)}},e.prototype._unsubscribe=function(){var t=this.errors,e=this.retriesSubscription;t&&(t.unsubscribe(),this.errors=null),e&&(e.unsubscribe(),this.retriesSubscription=null),this.retries=null},e.prototype.notifyNext=function(t,e,n,r,i){var o=this._unsubscribe;this._unsubscribe=null,this._unsubscribeAndRecycle(),this._unsubscribe=o,this.source.subscribe(this)},e}(s.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/sample.js":function(t,e,n){"use strict";n.r(e),n.d(e,"sample",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function a(t){return function(e){return e.lift(new s(t))}}var s=function(){function t(t){this.notifier=t}return t.prototype.call=function(t,e){var n=new u(t),r=e.subscribe(n);return r.add(Object(o.subscribeToResult)(n,this.notifier)),r},t}(),u=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.hasValue=!1,e}return r.__extends(e,t),e.prototype._next=function(t){this.value=t,this.hasValue=!0},e.prototype.notifyNext=function(t,e,n,r,i){this.emitValue()},e.prototype.notifyComplete=function(){this.emitValue()},e.prototype.emitValue=function(){this.hasValue&&(this.hasValue=!1,this.destination.next(this.value))},e}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/sampleTime.js":function(t,e,n){"use strict";n.r(e),n.d(e,"sampleTime",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js");function a(t,e){return void 0===e&&(e=o.async),function(n){return n.lift(new s(t,e))}}var s=function(){function t(t,e){this.period=t,this.scheduler=e}return t.prototype.call=function(t,e){return e.subscribe(new u(t,this.period,this.scheduler))},t}(),u=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.period=n,i.scheduler=r,i.hasValue=!1,i.add(r.schedule(c,n,{subscriber:i,period:n})),i}return r.__extends(e,t),e.prototype._next=function(t){this.lastValue=t,this.hasValue=!0},e.prototype.notifyNext=function(){this.hasValue&&(this.hasValue=!1,this.destination.next(this.lastValue))},e}(i.Subscriber);function c(t){var e=t.subscriber,n=t.period;e.notifyNext(),this.schedule(t,n)}},"./node_modules/rxjs/_esm5/internal/operators/scan.js":function(t,e,n){"use strict";n.r(e),n.d(e,"scan",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(t,e){var n=!1;return arguments.length>=2&&(n=!0),function(r){return r.lift(new a(t,e,n))}}var a=function(){function t(t,e,n){void 0===n&&(n=!1),this.accumulator=t,this.seed=e,this.hasSeed=n}return t.prototype.call=function(t,e){return e.subscribe(new s(t,this.accumulator,this.seed,this.hasSeed))},t}(),s=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.accumulator=n,o._seed=r,o.hasSeed=i,o.index=0,o}return r.__extends(e,t),Object.defineProperty(e.prototype,"seed",{get:function(){return this._seed},set:function(t){this.hasSeed=!0,this._seed=t},enumerable:!0,configurable:!0}),e.prototype._next=function(t){if(this.hasSeed)return this._tryNext(t);this.seed=t,this.destination.next(t)},e.prototype._tryNext=function(t){var e,n=this.index++;try{e=this.accumulator(this.seed,t,n)}catch(t){this.destination.error(t)}this.seed=e,this.destination.next(e)},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/sequenceEqual.js":function(t,e,n){"use strict";n.r(e),n.d(e,"sequenceEqual",function(){return s}),n.d(e,"SequenceEqualOperator",function(){return u}),n.d(e,"SequenceEqualSubscriber",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),a=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js");function s(t,e){return function(n){return n.lift(new u(t,e))}}var u=function(){function t(t,e){this.compareTo=t,this.comparor=e}return t.prototype.call=function(t,e){return e.subscribe(new c(t,this.compareTo,this.comparor))},t}(),c=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.compareTo=n,i.comparor=r,i._a=[],i._b=[],i._oneComplete=!1,i.destination.add(n.subscribe(new l(e,i))),i}return r.__extends(e,t),e.prototype._next=function(t){this._oneComplete&&0===this._b.length?this.emit(!1):(this._a.push(t),this.checkValues())},e.prototype._complete=function(){this._oneComplete?this.emit(0===this._a.length&&0===this._b.length):this._oneComplete=!0,this.unsubscribe()},e.prototype.checkValues=function(){for(var t=this._a,e=this._b,n=this.comparor;t.length>0&&e.length>0;){var r=t.shift(),i=e.shift(),s=!1;n?(s=Object(o.tryCatch)(n)(r,i))===a.errorObject&&this.destination.error(a.errorObject.e):s=r===i,s||this.emit(!1)}},e.prototype.emit=function(t){var e=this.destination;e.next(t),e.complete()},e.prototype.nextB=function(t){this._oneComplete&&0===this._a.length?this.emit(!1):(this._b.push(t),this.checkValues())},e.prototype.completeB=function(){this._oneComplete?this.emit(0===this._a.length&&0===this._b.length):this._oneComplete=!0},e}(i.Subscriber),l=function(t){function e(e,n){var r=t.call(this,e)||this;return r.parent=n,r}return r.__extends(e,t),e.prototype._next=function(t){this.parent.nextB(t)},e.prototype._error=function(t){this.parent.error(t),this.unsubscribe()},e.prototype._complete=function(){this.parent.completeB(),this.unsubscribe()},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/share.js":function(t,e,n){"use strict";n.r(e),n.d(e,"share",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/operators/multicast.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/refCount.js"),o=n("./node_modules/rxjs/_esm5/internal/Subject.js");function a(){return new o.Subject}function s(){return function(t){return Object(i.refCount)()(Object(r.multicast)(a)(t))}}},"./node_modules/rxjs/_esm5/internal/operators/shareReplay.js":function(t,e,n){"use strict";n.r(e),n.d(e,"shareReplay",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/ReplaySubject.js");function i(t,e,n){return void 0===t&&(t=Number.POSITIVE_INFINITY),void 0===e&&(e=Number.POSITIVE_INFINITY),function(i){return i.lift((o=t,a=e,s=n,l=0,d=!1,p=!1,function(t){l++,u&&!d||(d=!1,u=new r.ReplaySubject(o,a,s),c=t.subscribe({next:function(t){u.next(t)},error:function(t){d=!0,u.error(t)},complete:function(){p=!0,u.complete()}}));var e=u.subscribe(this);return function(){l--,e.unsubscribe(),c&&0===l&&p&&c.unsubscribe()}}));var o,a,s,u,c,l,d,p}}},"./node_modules/rxjs/_esm5/internal/operators/single.js":function(t,e,n){"use strict";n.r(e),n.d(e,"single",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/EmptyError.js");function a(t){return function(e){return e.lift(new s(t,e))}}var s=function(){function t(t,e){this.predicate=t,this.source=e}return t.prototype.call=function(t,e){return e.subscribe(new u(t,this.predicate,this.source))},t}(),u=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.predicate=n,i.source=r,i.seenValue=!1,i.index=0,i}return r.__extends(e,t),e.prototype.applySingleValue=function(t){this.seenValue?this.destination.error("Sequence contains more than one element"):(this.seenValue=!0,this.singleValue=t)},e.prototype._next=function(t){var e=this.index++;this.predicate?this.tryNext(t,e):this.applySingleValue(t)},e.prototype.tryNext=function(t,e){try{this.predicate(t,e,this.source)&&this.applySingleValue(t)}catch(t){this.destination.error(t)}},e.prototype._complete=function(){var t=this.destination;this.index>0?(t.next(this.seenValue?this.singleValue:void 0),t.complete()):t.error(new o.EmptyError)},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/skip.js":function(t,e,n){"use strict";n.r(e),n.d(e,"skip",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(t){return function(e){return e.lift(new a(t))}}var a=function(){function t(t){this.total=t}return t.prototype.call=function(t,e){return e.subscribe(new s(t,this.total))},t}(),s=function(t){function e(e,n){var r=t.call(this,e)||this;return r.total=n,r.count=0,r}return r.__extends(e,t),e.prototype._next=function(t){++this.count>this.total&&this.destination.next(t)},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/skipLast.js":function(t,e,n){"use strict";n.r(e),n.d(e,"skipLast",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/ArgumentOutOfRangeError.js");function a(t){return function(e){return e.lift(new s(t))}}var s=function(){function t(t){if(this._skipCount=t,this._skipCount<0)throw new o.ArgumentOutOfRangeError}return t.prototype.call=function(t,e){return 0===this._skipCount?e.subscribe(new i.Subscriber(t)):e.subscribe(new u(t,this._skipCount))},t}(),u=function(t){function e(e,n){var r=t.call(this,e)||this;return r._skipCount=n,r._count=0,r._ring=new Array(n),r}return r.__extends(e,t),e.prototype._next=function(t){var e=this._skipCount,n=this._count++;if(n<e)this._ring[n]=t;else{var r=n%e,i=this._ring,o=i[r];i[r]=t,this.destination.next(o)}},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/skipUntil.js":function(t,e,n){"use strict";n.r(e),n.d(e,"skipUntil",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/InnerSubscriber.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function s(t){return function(e){return e.lift(new u(t))}}var u=function(){function t(t){this.notifier=t}return t.prototype.call=function(t,e){return e.subscribe(new c(t,this.notifier))},t}(),c=function(t){function e(e,n){var r=t.call(this,e)||this;r.hasValue=!1;var i=new o.InnerSubscriber(r,void 0,void 0);return r.add(i),r.innerSubscription=i,Object(a.subscribeToResult)(r,n,void 0,void 0,i),r}return r.__extends(e,t),e.prototype._next=function(e){this.hasValue&&t.prototype._next.call(this,e)},e.prototype.notifyNext=function(t,e,n,r,i){this.hasValue=!0,this.innerSubscription&&this.innerSubscription.unsubscribe()},e.prototype.notifyComplete=function(){},e}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/skipWhile.js":function(t,e,n){"use strict";n.r(e),n.d(e,"skipWhile",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(t){return function(e){return e.lift(new a(t))}}var a=function(){function t(t){this.predicate=t}return t.prototype.call=function(t,e){return e.subscribe(new s(t,this.predicate))},t}(),s=function(t){function e(e,n){var r=t.call(this,e)||this;return r.predicate=n,r.skipping=!0,r.index=0,r}return r.__extends(e,t),e.prototype._next=function(t){var e=this.destination;this.skipping&&this.tryCallPredicate(t),this.skipping||e.next(t)},e.prototype.tryCallPredicate=function(t){try{var e=this.predicate(t,this.index++);this.skipping=Boolean(e)}catch(t){this.destination.error(t)}},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/startWith.js":function(t,e,n){"use strict";n.r(e),n.d(e,"startWith",function(){return u});var r=n("./node_modules/rxjs/_esm5/internal/observable/fromArray.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/scalar.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/empty.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/concat.js"),s=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js");function u(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return function(e){var n=t[t.length-1];Object(s.isScheduler)(n)?t.pop():n=null;var u=t.length;return 1!==u||n?u>0?Object(a.concat)(Object(r.fromArray)(t,n),e):Object(a.concat)(Object(o.empty)(n),e):Object(a.concat)(Object(i.scalar)(t[0]),e)}}},"./node_modules/rxjs/_esm5/internal/operators/subscribeOn.js":function(t,e,n){"use strict";n.r(e),n.d(e,"subscribeOn",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/observable/SubscribeOnObservable.js");function i(t,e){return void 0===e&&(e=0),function(n){return n.lift(new o(t,e))}}var o=function(){function t(t,e){this.scheduler=t,this.delay=e}return t.prototype.call=function(t,e){return new r.SubscribeOnObservable(e,this.delay,this.scheduler).subscribe(t)},t}()},"./node_modules/rxjs/_esm5/internal/operators/switchAll.js":function(t,e,n){"use strict";n.r(e),n.d(e,"switchAll",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/operators/switchMap.js"),i=n("./node_modules/rxjs/_esm5/internal/util/identity.js");function o(){return Object(r.switchMap)(i.identity)}},"./node_modules/rxjs/_esm5/internal/operators/switchMap.js":function(t,e,n){"use strict";n.r(e),n.d(e,"switchMap",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/InnerSubscriber.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),s=n("./node_modules/rxjs/_esm5/internal/operators/map.js"),u=n("./node_modules/rxjs/_esm5/internal/observable/from.js");function c(t,e){return"function"==typeof e?function(n){return n.pipe(c(function(n,r){return Object(u.from)(t(n,r)).pipe(Object(s.map)(function(t,i){return e(n,t,r,i)}))}))}:function(e){return e.lift(new l(t))}}var l=function(){function t(t){this.project=t}return t.prototype.call=function(t,e){return e.subscribe(new d(t,this.project))},t}(),d=function(t){function e(e,n){var r=t.call(this,e)||this;return r.project=n,r.index=0,r}return r.__extends(e,t),e.prototype._next=function(t){var e,n=this.index++;try{e=this.project(t,n)}catch(t){return void this.destination.error(t)}this._innerSub(e,t,n)},e.prototype._innerSub=function(t,e,n){var r=this.innerSubscription;r&&r.unsubscribe();var i=new o.InnerSubscriber(this,void 0,void 0);this.destination.add(i),this.innerSubscription=Object(a.subscribeToResult)(this,t,e,n,i)},e.prototype._complete=function(){var e=this.innerSubscription;e&&!e.closed||t.prototype._complete.call(this),this.unsubscribe()},e.prototype._unsubscribe=function(){this.innerSubscription=null},e.prototype.notifyComplete=function(e){this.destination.remove(e),this.innerSubscription=null,this.isStopped&&t.prototype._complete.call(this)},e.prototype.notifyNext=function(t,e,n,r,i){this.destination.next(e)},e}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/switchMapTo.js":function(t,e,n){"use strict";n.r(e),n.d(e,"switchMapTo",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/switchMap.js");function i(t,e){return e?Object(r.switchMap)(function(){return t},e):Object(r.switchMap)(function(){return t})}},"./node_modules/rxjs/_esm5/internal/operators/take.js":function(t,e,n){"use strict";n.r(e),n.d(e,"take",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/ArgumentOutOfRangeError.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/empty.js");function s(t){return function(e){return 0===t?Object(a.empty)():e.lift(new u(t))}}var u=function(){function t(t){if(this.total=t,this.total<0)throw new o.ArgumentOutOfRangeError}return t.prototype.call=function(t,e){return e.subscribe(new c(t,this.total))},t}(),c=function(t){function e(e,n){var r=t.call(this,e)||this;return r.total=n,r.count=0,r}return r.__extends(e,t),e.prototype._next=function(t){var e=this.total,n=++this.count;n<=e&&(this.destination.next(t),n===e&&(this.destination.complete(),this.unsubscribe()))},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/takeLast.js":function(t,e,n){"use strict";n.r(e),n.d(e,"takeLast",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/ArgumentOutOfRangeError.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/empty.js");function s(t){return function(e){return 0===t?Object(a.empty)():e.lift(new u(t))}}var u=function(){function t(t){if(this.total=t,this.total<0)throw new o.ArgumentOutOfRangeError}return t.prototype.call=function(t,e){return e.subscribe(new c(t,this.total))},t}(),c=function(t){function e(e,n){var r=t.call(this,e)||this;return r.total=n,r.ring=new Array,r.count=0,r}return r.__extends(e,t),e.prototype._next=function(t){var e=this.ring,n=this.total,r=this.count++;e.length<n?e.push(t):e[r%n]=t},e.prototype._complete=function(){var t=this.destination,e=this.count;if(e>0)for(var n=this.count>=this.total?this.total:this.count,r=this.ring,i=0;i<n;i++){var o=e++%n;t.next(r[o])}t.complete()},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/takeUntil.js":function(t,e,n){"use strict";n.r(e),n.d(e,"takeUntil",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function a(t){return function(e){return e.lift(new s(t))}}var s=function(){function t(t){this.notifier=t}return t.prototype.call=function(t,e){var n=new u(t),r=Object(o.subscribeToResult)(n,this.notifier);return r&&!n.seenValue?(n.add(r),e.subscribe(n)):n},t}(),u=function(t){function e(e){var n=t.call(this,e)||this;return n.seenValue=!1,n}return r.__extends(e,t),e.prototype.notifyNext=function(t,e,n,r,i){this.seenValue=!0,this.complete()},e.prototype.notifyComplete=function(){},e}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/takeWhile.js":function(t,e,n){"use strict";n.r(e),n.d(e,"takeWhile",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(t){return function(e){return e.lift(new a(t))}}var a=function(){function t(t){this.predicate=t}return t.prototype.call=function(t,e){return e.subscribe(new s(t,this.predicate))},t}(),s=function(t){function e(e,n){var r=t.call(this,e)||this;return r.predicate=n,r.index=0,r}return r.__extends(e,t),e.prototype._next=function(t){var e,n=this.destination;try{e=this.predicate(t,this.index++)}catch(t){return void n.error(t)}this.nextOrComplete(t,e)},e.prototype.nextOrComplete=function(t,e){var n=this.destination;Boolean(e)?n.next(t):n.complete()},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/tap.js":function(t,e,n){"use strict";n.r(e),n.d(e,"tap",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/noop.js"),a=n("./node_modules/rxjs/_esm5/internal/util/isFunction.js");function s(t,e,n){return function(r){return r.lift(new u(t,e,n))}}var u=function(){function t(t,e,n){this.nextOrObserver=t,this.error=e,this.complete=n}return t.prototype.call=function(t,e){return e.subscribe(new c(t,this.nextOrObserver,this.error,this.complete))},t}(),c=function(t){function e(e,n,r,i){var s=t.call(this,e)||this;return s._tapNext=o.noop,s._tapError=o.noop,s._tapComplete=o.noop,s._tapError=r||o.noop,s._tapComplete=i||o.noop,Object(a.isFunction)(n)?(s._context=s,s._tapNext=n):n&&(s._context=n,s._tapNext=n.next||o.noop,s._tapError=n.error||o.noop,s._tapComplete=n.complete||o.noop),s}return r.__extends(e,t),e.prototype._next=function(t){try{this._tapNext.call(this._context,t)}catch(t){return void this.destination.error(t)}this.destination.next(t)},e.prototype._error=function(t){try{this._tapError.call(this._context,t)}catch(t){return void this.destination.error(t)}this.destination.error(t)},e.prototype._complete=function(){try{this._tapComplete.call(this._context)}catch(t){return void this.destination.error(t)}return this.destination.complete()},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/throttle.js":function(t,e,n){"use strict";n.r(e),n.d(e,"defaultThrottleConfig",function(){return a}),n.d(e,"throttle",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),a={leading:!0,trailing:!1};function s(t,e){return void 0===e&&(e=a),function(n){return n.lift(new u(t,e.leading,e.trailing))}}var u=function(){function t(t,e,n){this.durationSelector=t,this.leading=e,this.trailing=n}return t.prototype.call=function(t,e){return e.subscribe(new c(t,this.durationSelector,this.leading,this.trailing))},t}(),c=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.destination=e,o.durationSelector=n,o._leading=r,o._trailing=i,o._hasValue=!1,o}return r.__extends(e,t),e.prototype._next=function(t){this._hasValue=!0,this._sendValue=t,this._throttled||(this._leading?this.send():this.throttle(t))},e.prototype.send=function(){var t=this._hasValue,e=this._sendValue;t&&(this.destination.next(e),this.throttle(e)),this._hasValue=!1,this._sendValue=null},e.prototype.throttle=function(t){var e=this.tryDurationSelector(t);e&&this.add(this._throttled=Object(o.subscribeToResult)(this,e))},e.prototype.tryDurationSelector=function(t){try{return this.durationSelector(t)}catch(t){return this.destination.error(t),null}},e.prototype.throttlingDone=function(){var t=this._throttled,e=this._trailing;t&&t.unsubscribe(),this._throttled=null,e&&this.send()},e.prototype.notifyNext=function(t,e,n,r,i){this.throttlingDone()},e.prototype.notifyComplete=function(){this.throttlingDone()},e}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/throttleTime.js":function(t,e,n){"use strict";n.r(e),n.d(e,"throttleTime",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),a=n("./node_modules/rxjs/_esm5/internal/operators/throttle.js");function s(t,e,n){return void 0===e&&(e=o.async),void 0===n&&(n=a.defaultThrottleConfig),function(r){return r.lift(new u(t,e,n.leading,n.trailing))}}var u=function(){function t(t,e,n,r){this.duration=t,this.scheduler=e,this.leading=n,this.trailing=r}return t.prototype.call=function(t,e){return e.subscribe(new c(t,this.duration,this.scheduler,this.leading,this.trailing))},t}(),c=function(t){function e(e,n,r,i,o){var a=t.call(this,e)||this;return a.duration=n,a.scheduler=r,a.leading=i,a.trailing=o,a._hasTrailingValue=!1,a._trailingValue=null,a}return r.__extends(e,t),e.prototype._next=function(t){this.throttled?this.trailing&&(this._trailingValue=t,this._hasTrailingValue=!0):(this.add(this.throttled=this.scheduler.schedule(l,this.duration,{subscriber:this})),this.leading&&this.destination.next(t))},e.prototype._complete=function(){this._hasTrailingValue?(this.destination.next(this._trailingValue),this.destination.complete()):this.destination.complete()},e.prototype.clearThrottle=function(){var t=this.throttled;t&&(this.trailing&&this._hasTrailingValue&&(this.destination.next(this._trailingValue),this._trailingValue=null,this._hasTrailingValue=!1),t.unsubscribe(),this.remove(t),this.throttled=null)},e}(i.Subscriber);function l(t){t.subscriber.clearThrottle()}},"./node_modules/rxjs/_esm5/internal/operators/throwIfEmpty.js":function(t,e,n){"use strict";n.r(e),n.d(e,"throwIfEmpty",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/operators/tap.js"),i=n("./node_modules/rxjs/_esm5/internal/util/EmptyError.js"),o=function(t){return void 0===t&&(t=a),Object(r.tap)({hasValue:!1,next:function(){this.hasValue=!0},complete:function(){if(!this.hasValue)throw t()}})};function a(){return new i.EmptyError}},"./node_modules/rxjs/_esm5/internal/operators/timeInterval.js":function(t,e,n){"use strict";n.r(e),n.d(e,"timeInterval",function(){return s}),n.d(e,"TimeInterval",function(){return u});var r=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/scan.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/defer.js"),a=n("./node_modules/rxjs/_esm5/internal/operators/map.js");function s(t){return void 0===t&&(t=r.async),function(e){return Object(o.defer)(function(){return e.pipe(Object(i.scan)(function(e,n){var r=e.current;return{value:n,current:t.now(),last:r}},{current:t.now(),value:void 0,last:void 0}),Object(a.map)(function(t){var e=t.current,n=t.last,r=t.value;return new u(r,e-n)}))})}}var u=function(){return function(t,e){this.value=t,this.interval=e}}()},"./node_modules/rxjs/_esm5/internal/operators/timeout.js":function(t,e,n){"use strict";n.r(e),n.d(e,"timeout",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),i=n("./node_modules/rxjs/_esm5/internal/util/TimeoutError.js"),o=n("./node_modules/rxjs/_esm5/internal/operators/timeoutWith.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/throwError.js");function s(t,e){return void 0===e&&(e=r.async),Object(o.timeoutWith)(t,Object(a.throwError)(new i.TimeoutError),e)}},"./node_modules/rxjs/_esm5/internal/operators/timeoutWith.js":function(t,e,n){"use strict";n.r(e),n.d(e,"timeoutWith",function(){return u});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isDate.js"),a=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function u(t,e,n){return void 0===n&&(n=i.async),function(r){var i=Object(o.isDate)(t),a=i?+t-n.now():Math.abs(t);return r.lift(new c(a,i,e,n))}}var c=function(){function t(t,e,n,r){this.waitFor=t,this.absoluteTimeout=e,this.withObservable=n,this.scheduler=r}return t.prototype.call=function(t,e){return e.subscribe(new l(t,this.absoluteTimeout,this.waitFor,this.withObservable,this.scheduler))},t}(),l=function(t){function e(e,n,r,i,o){var a=t.call(this,e)||this;return a.absoluteTimeout=n,a.waitFor=r,a.withObservable=i,a.scheduler=o,a.action=null,a.scheduleTimeout(),a}return r.__extends(e,t),e.dispatchTimeout=function(t){var e=t.withObservable;t._unsubscribeAndRecycle(),t.add(Object(s.subscribeToResult)(t,e))},e.prototype.scheduleTimeout=function(){var t=this.action;t?this.action=t.schedule(this,this.waitFor):this.add(this.action=this.scheduler.schedule(e.dispatchTimeout,this.waitFor,this))},e.prototype._next=function(e){this.absoluteTimeout||this.scheduleTimeout(),t.prototype._next.call(this,e)},e.prototype._unsubscribe=function(){this.action=null,this.scheduler=null,this.withObservable=null},e}(a.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/timestamp.js":function(t,e,n){"use strict";n.r(e),n.d(e,"timestamp",function(){return o}),n.d(e,"Timestamp",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/map.js");function o(t){return void 0===t&&(t=r.async),Object(i.map)(function(e){return new a(e,t.now())})}var a=function(){return function(t,e){this.value=t,this.timestamp=e}}()},"./node_modules/rxjs/_esm5/internal/operators/toArray.js":function(t,e,n){"use strict";n.r(e),n.d(e,"toArray",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/operators/reduce.js");function i(t,e,n){return 0===n?[e]:(t.push(e),t)}function o(){return Object(r.reduce)(i,[])}},"./node_modules/rxjs/_esm5/internal/operators/window.js":function(t,e,n){"use strict";n.r(e),n.d(e,"window",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function s(t){return function(e){return e.lift(new u(t))}}var u=function(){function t(t){this.windowBoundaries=t}return t.prototype.call=function(t,e){var n=new c(t),r=e.subscribe(n);return r.closed||n.add(Object(a.subscribeToResult)(n,this.windowBoundaries)),r},t}(),c=function(t){function e(e){var n=t.call(this,e)||this;return n.window=new i.Subject,e.next(n.window),n}return r.__extends(e,t),e.prototype.notifyNext=function(t,e,n,r,i){this.openWindow()},e.prototype.notifyError=function(t,e){this._error(t)},e.prototype.notifyComplete=function(t){this._complete()},e.prototype._next=function(t){this.window.next(t)},e.prototype._error=function(t){this.window.error(t),this.destination.error(t)},e.prototype._complete=function(){this.window.complete(),this.destination.complete()},e.prototype._unsubscribe=function(){this.window=null},e.prototype.openWindow=function(){var t=this.window;t&&t.complete();var e=this.destination,n=this.window=new i.Subject;e.next(n)},e}(o.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/windowCount.js":function(t,e,n){"use strict";n.r(e),n.d(e,"windowCount",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/Subject.js");function a(t,e){return void 0===e&&(e=0),function(n){return n.lift(new s(t,e))}}var s=function(){function t(t,e){this.windowSize=t,this.startWindowEvery=e}return t.prototype.call=function(t,e){return e.subscribe(new u(t,this.windowSize,this.startWindowEvery))},t}(),u=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.destination=e,i.windowSize=n,i.startWindowEvery=r,i.windows=[new o.Subject],i.count=0,e.next(i.windows[0]),i}return r.__extends(e,t),e.prototype._next=function(t){for(var e=this.startWindowEvery>0?this.startWindowEvery:this.windowSize,n=this.destination,r=this.windowSize,i=this.windows,a=i.length,s=0;s<a&&!this.closed;s++)i[s].next(t);var u=this.count-r+1;if(u>=0&&u%e==0&&!this.closed&&i.shift().complete(),++this.count%e==0&&!this.closed){var c=new o.Subject;i.push(c),n.next(c)}},e.prototype._error=function(t){var e=this.windows;if(e)for(;e.length>0&&!this.closed;)e.shift().error(t);this.destination.error(t)},e.prototype._complete=function(){var t=this.windows;if(t)for(;t.length>0&&!this.closed;)t.shift().complete();this.destination.complete()},e.prototype._unsubscribe=function(){this.count=0,this.windows=null},e}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/windowTime.js":function(t,e,n){"use strict";n.r(e),n.d(e,"windowTime",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),a=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/util/isNumeric.js"),u=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js");function c(t){var e=o.async,n=null,r=Number.POSITIVE_INFINITY;return Object(u.isScheduler)(arguments[3])&&(e=arguments[3]),Object(u.isScheduler)(arguments[2])?e=arguments[2]:Object(s.isNumeric)(arguments[2])&&(r=arguments[2]),Object(u.isScheduler)(arguments[1])?e=arguments[1]:Object(s.isNumeric)(arguments[1])&&(n=arguments[1]),function(i){return i.lift(new l(t,n,r,e))}}var l=function(){function t(t,e,n,r){this.windowTimeSpan=t,this.windowCreationInterval=e,this.maxWindowSize=n,this.scheduler=r}return t.prototype.call=function(t,e){return e.subscribe(new p(t,this.windowTimeSpan,this.windowCreationInterval,this.maxWindowSize,this.scheduler))},t}(),d=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._numberOfNextedValues=0,e}return r.__extends(e,t),e.prototype.next=function(e){this._numberOfNextedValues++,t.prototype.next.call(this,e)},Object.defineProperty(e.prototype,"numberOfNextedValues",{get:function(){return this._numberOfNextedValues},enumerable:!0,configurable:!0}),e}(i.Subject),p=function(t){function e(e,n,r,i,o){var a=t.call(this,e)||this;a.destination=e,a.windowTimeSpan=n,a.windowCreationInterval=r,a.maxWindowSize=i,a.scheduler=o,a.windows=[];var s=a.openWindow();if(null!==r&&r>=0){var u={subscriber:a,window:s,context:null},c={windowTimeSpan:n,windowCreationInterval:r,subscriber:a,scheduler:o};a.add(o.schedule(m,n,u)),a.add(o.schedule(h,r,c))}else{var l={subscriber:a,window:s,windowTimeSpan:n};a.add(o.schedule(f,n,l))}return a}return r.__extends(e,t),e.prototype._next=function(t){for(var e=this.windows,n=e.length,r=0;r<n;r++){var i=e[r];i.closed||(i.next(t),i.numberOfNextedValues>=this.maxWindowSize&&this.closeWindow(i))}},e.prototype._error=function(t){for(var e=this.windows;e.length>0;)e.shift().error(t);this.destination.error(t)},e.prototype._complete=function(){for(var t=this.windows;t.length>0;){var e=t.shift();e.closed||e.complete()}this.destination.complete()},e.prototype.openWindow=function(){var t=new d;return this.windows.push(t),this.destination.next(t),t},e.prototype.closeWindow=function(t){t.complete();var e=this.windows;e.splice(e.indexOf(t),1)},e}(a.Subscriber);function f(t){var e=t.subscriber,n=t.windowTimeSpan,r=t.window;r&&e.closeWindow(r),t.window=e.openWindow(),this.schedule(t,n)}function h(t){var e=t.windowTimeSpan,n=t.subscriber,r=t.scheduler,i=t.windowCreationInterval,o={action:this,subscription:null},a={subscriber:n,window:n.openWindow(),context:o};o.subscription=r.schedule(m,e,a),this.add(o.subscription),this.schedule(t,i)}function m(t){var e=t.subscriber,n=t.window,r=t.context;r&&r.action&&r.subscription&&r.action.remove(r.subscription),e.closeWindow(n)}},"./node_modules/rxjs/_esm5/internal/operators/windowToggle.js":function(t,e,n){"use strict";n.r(e),n.d(e,"windowToggle",function(){return l});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),a=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),s=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),u=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),c=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function l(t,e){return function(n){return n.lift(new d(t,e))}}var d=function(){function t(t,e){this.openings=t,this.closingSelector=e}return t.prototype.call=function(t,e){return e.subscribe(new p(t,this.openings,this.closingSelector))},t}(),p=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.openings=n,i.closingSelector=r,i.contexts=[],i.add(i.openSubscription=Object(c.subscribeToResult)(i,n,n)),i}return r.__extends(e,t),e.prototype._next=function(t){var e=this.contexts;if(e)for(var n=e.length,r=0;r<n;r++)e[r].window.next(t)},e.prototype._error=function(e){var n=this.contexts;if(this.contexts=null,n)for(var r=n.length,i=-1;++i<r;){var o=n[i];o.window.error(e),o.subscription.unsubscribe()}t.prototype._error.call(this,e)},e.prototype._complete=function(){var e=this.contexts;if(this.contexts=null,e)for(var n=e.length,r=-1;++r<n;){var i=e[r];i.window.complete(),i.subscription.unsubscribe()}t.prototype._complete.call(this)},e.prototype._unsubscribe=function(){var t=this.contexts;if(this.contexts=null,t)for(var e=t.length,n=-1;++n<e;){var r=t[n];r.window.unsubscribe(),r.subscription.unsubscribe()}},e.prototype.notifyNext=function(t,e,n,r,u){if(t===this.openings){var l=this.closingSelector,d=Object(a.tryCatch)(l)(e);if(d===s.errorObject)return this.error(s.errorObject.e);var p=new i.Subject,f=new o.Subscription,h={window:p,subscription:f};this.contexts.push(h);var m=Object(c.subscribeToResult)(this,d,h);m.closed?this.closeWindow(this.contexts.length-1):(m.context=h,f.add(m)),this.destination.next(p)}else this.closeWindow(this.contexts.indexOf(t))},e.prototype.notifyError=function(t){this.error(t)},e.prototype.notifyComplete=function(t){t!==this.openSubscription&&this.closeWindow(this.contexts.indexOf(t.context))},e.prototype.closeWindow=function(t){if(-1!==t){var e=this.contexts,n=e[t],r=n.window,i=n.subscription;e.splice(t,1),r.complete(),i.unsubscribe()}},e}(u.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/windowWhen.js":function(t,e,n){"use strict";n.r(e),n.d(e,"windowWhen",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),a=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),s=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),u=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function c(t){return function(e){return e.lift(new l(t))}}var l=function(){function t(t){this.closingSelector=t}return t.prototype.call=function(t,e){return e.subscribe(new d(t,this.closingSelector))},t}(),d=function(t){function e(e,n){var r=t.call(this,e)||this;return r.destination=e,r.closingSelector=n,r.openWindow(),r}return r.__extends(e,t),e.prototype.notifyNext=function(t,e,n,r,i){this.openWindow(i)},e.prototype.notifyError=function(t,e){this._error(t)},e.prototype.notifyComplete=function(t){this.openWindow(t)},e.prototype._next=function(t){this.window.next(t)},e.prototype._error=function(t){this.window.error(t),this.destination.error(t),this.unsubscribeClosingNotification()},e.prototype._complete=function(){this.window.complete(),this.destination.complete(),this.unsubscribeClosingNotification()},e.prototype.unsubscribeClosingNotification=function(){this.closingNotification&&this.closingNotification.unsubscribe()},e.prototype.openWindow=function(t){void 0===t&&(t=null),t&&(this.remove(t),t.unsubscribe());var e=this.window;e&&e.complete();var n=this.window=new i.Subject;this.destination.next(n);var r=Object(o.tryCatch)(this.closingSelector)();if(r===a.errorObject){var s=a.errorObject.e;this.destination.error(s),this.window.error(s)}else this.add(this.closingNotification=Object(u.subscribeToResult)(this,r))},e}(s.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/withLatestFrom.js":function(t,e,n){"use strict";n.r(e),n.d(e,"withLatestFrom",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function a(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return function(e){var n;"function"==typeof t[t.length-1]&&(n=t.pop());var r=t;return e.lift(new s(r,n))}}var s=function(){function t(t,e){this.observables=t,this.project=e}return t.prototype.call=function(t,e){return e.subscribe(new u(t,this.observables,this.project))},t}(),u=function(t){function e(e,n,r){var i=t.call(this,e)||this;i.observables=n,i.project=r,i.toRespond=[];var a=n.length;i.values=new Array(a);for(var s=0;s<a;s++)i.toRespond.push(s);for(s=0;s<a;s++){var u=n[s];i.add(Object(o.subscribeToResult)(i,u,u,s))}return i}return r.__extends(e,t),e.prototype.notifyNext=function(t,e,n,r,i){this.values[n]=e;var o=this.toRespond;if(o.length>0){var a=o.indexOf(n);-1!==a&&o.splice(a,1)}},e.prototype.notifyComplete=function(){},e.prototype._next=function(t){if(0===this.toRespond.length){var e=[t].concat(this.values);this.project?this._tryProject(e):this.destination.next(e)}},e.prototype._tryProject=function(t){var e;try{e=this.project.apply(this,t)}catch(t){return void this.destination.error(t)}this.destination.next(e)},e}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/zip.js":function(t,e,n){"use strict";n.r(e),n.d(e,"zip",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/observable/zip.js");function i(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return function(e){return e.lift.call(r.zip.apply(void 0,[e].concat(t)))}}},"./node_modules/rxjs/_esm5/internal/operators/zipAll.js":function(t,e,n){"use strict";n.r(e),n.d(e,"zipAll",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/observable/zip.js");function i(t){return function(e){return e.lift(new r.ZipOperator(t))}}},"./node_modules/rxjs/_esm5/internal/scheduler/Action.js":function(t,e,n){"use strict";n.r(e),n.d(e,"Action",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(t){function e(e,n){return t.call(this)||this}return r.__extends(e,t),e.prototype.schedule=function(t,e){return void 0===e&&(e=0),this},e}(n("./node_modules/rxjs/_esm5/internal/Subscription.js").Subscription)},"./node_modules/rxjs/_esm5/internal/scheduler/AnimationFrameAction.js":function(t,e,n){"use strict";n.r(e),n.d(e,"AnimationFrameAction",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(t){function e(e,n){var r=t.call(this,e,n)||this;return r.scheduler=e,r.work=n,r}return r.__extends(e,t),e.prototype.requestAsyncId=function(e,n,r){return void 0===r&&(r=0),null!==r&&r>0?t.prototype.requestAsyncId.call(this,e,n,r):(e.actions.push(this),e.scheduled||(e.scheduled=requestAnimationFrame(function(){return e.flush(null)})))},e.prototype.recycleAsyncId=function(e,n,r){if(void 0===r&&(r=0),null!==r&&r>0||null===r&&this.delay>0)return t.prototype.recycleAsyncId.call(this,e,n,r);0===e.actions.length&&(cancelAnimationFrame(n),e.scheduled=void 0)},e}(n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncAction.js").AsyncAction)},"./node_modules/rxjs/_esm5/internal/scheduler/AnimationFrameScheduler.js":function(t,e,n){"use strict";n.r(e),n.d(e,"AnimationFrameScheduler",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return r.__extends(e,t),e.prototype.flush=function(t){this.active=!0,this.scheduled=void 0;var e,n=this.actions,r=-1,i=n.length;t=t||n.shift();do{if(e=t.execute(t.state,t.delay))break}while(++r<i&&(t=n.shift()));if(this.active=!1,e){for(;++r<i&&(t=n.shift());)t.unsubscribe();throw e}},e}(n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncScheduler.js").AsyncScheduler)},"./node_modules/rxjs/_esm5/internal/scheduler/AsapAction.js":function(t,e,n){"use strict";n.r(e),n.d(e,"AsapAction",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/util/Immediate.js"),o=function(t){function e(e,n){var r=t.call(this,e,n)||this;return r.scheduler=e,r.work=n,r}return r.__extends(e,t),e.prototype.requestAsyncId=function(e,n,r){return void 0===r&&(r=0),null!==r&&r>0?t.prototype.requestAsyncId.call(this,e,n,r):(e.actions.push(this),e.scheduled||(e.scheduled=i.Immediate.setImmediate(e.flush.bind(e,null))))},e.prototype.recycleAsyncId=function(e,n,r){if(void 0===r&&(r=0),null!==r&&r>0||null===r&&this.delay>0)return t.prototype.recycleAsyncId.call(this,e,n,r);0===e.actions.length&&(i.Immediate.clearImmediate(n),e.scheduled=void 0)},e}(n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncAction.js").AsyncAction)},"./node_modules/rxjs/_esm5/internal/scheduler/AsapScheduler.js":function(t,e,n){"use strict";n.r(e),n.d(e,"AsapScheduler",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return r.__extends(e,t),e.prototype.flush=function(t){this.active=!0,this.scheduled=void 0;var e,n=this.actions,r=-1,i=n.length;t=t||n.shift();do{if(e=t.execute(t.state,t.delay))break}while(++r<i&&(t=n.shift()));if(this.active=!1,e){for(;++r<i&&(t=n.shift());)t.unsubscribe();throw e}},e}(n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncScheduler.js").AsyncScheduler)},"./node_modules/rxjs/_esm5/internal/scheduler/AsyncAction.js":function(t,e,n){"use strict";n.r(e),n.d(e,"AsyncAction",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(t){function e(e,n){var r=t.call(this,e,n)||this;return r.scheduler=e,r.work=n,r.pending=!1,r}return r.__extends(e,t),e.prototype.schedule=function(t,e){if(void 0===e&&(e=0),this.closed)return this;this.state=t;var n=this.id,r=this.scheduler;return null!=n&&(this.id=this.recycleAsyncId(r,n,e)),this.pending=!0,this.delay=e,this.id=this.id||this.requestAsyncId(r,this.id,e),this},e.prototype.requestAsyncId=function(t,e,n){return void 0===n&&(n=0),setInterval(t.flush.bind(t,this),n)},e.prototype.recycleAsyncId=function(t,e,n){if(void 0===n&&(n=0),null!==n&&this.delay===n&&!1===this.pending)return e;clearInterval(e)},e.prototype.execute=function(t,e){if(this.closed)return new Error("executing a cancelled action");this.pending=!1;var n=this._execute(t,e);if(n)return n;!1===this.pending&&null!=this.id&&(this.id=this.recycleAsyncId(this.scheduler,this.id,null))},e.prototype._execute=function(t,e){var n=!1,r=void 0;try{this.work(t)}catch(t){n=!0,r=!!t&&t||new Error(t)}if(n)return this.unsubscribe(),r},e.prototype._unsubscribe=function(){var t=this.id,e=this.scheduler,n=e.actions,r=n.indexOf(this);this.work=null,this.state=null,this.pending=!1,this.scheduler=null,-1!==r&&n.splice(r,1),null!=t&&(this.id=this.recycleAsyncId(e,t,null)),this.delay=null},e}(n("./node_modules/rxjs/_esm5/internal/scheduler/Action.js").Action)},"./node_modules/rxjs/_esm5/internal/scheduler/AsyncScheduler.js":function(t,e,n){"use strict";n.r(e),n.d(e,"AsyncScheduler",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Scheduler.js"),o=function(t){function e(n,r){void 0===r&&(r=i.Scheduler.now);var o=t.call(this,n,function(){return e.delegate&&e.delegate!==o?e.delegate.now():r()})||this;return o.actions=[],o.active=!1,o.scheduled=void 0,o}return r.__extends(e,t),e.prototype.schedule=function(n,r,i){return void 0===r&&(r=0),e.delegate&&e.delegate!==this?e.delegate.schedule(n,r,i):t.prototype.schedule.call(this,n,r,i)},e.prototype.flush=function(t){var e=this.actions;if(this.active)e.push(t);else{var n;this.active=!0;do{if(n=t.execute(t.state,t.delay))break}while(t=e.shift());if(this.active=!1,n){for(;t=e.shift();)t.unsubscribe();throw n}}},e}(i.Scheduler)},"./node_modules/rxjs/_esm5/internal/scheduler/QueueAction.js":function(t,e,n){"use strict";n.r(e),n.d(e,"QueueAction",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(t){function e(e,n){var r=t.call(this,e,n)||this;return r.scheduler=e,r.work=n,r}return r.__extends(e,t),e.prototype.schedule=function(e,n){return void 0===n&&(n=0),n>0?t.prototype.schedule.call(this,e,n):(this.delay=n,this.state=e,this.scheduler.flush(this),this)},e.prototype.execute=function(e,n){return n>0||this.closed?t.prototype.execute.call(this,e,n):this._execute(e,n)},e.prototype.requestAsyncId=function(e,n,r){return void 0===r&&(r=0),null!==r&&r>0||null===r&&this.delay>0?t.prototype.requestAsyncId.call(this,e,n,r):e.flush(this)},e}(n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncAction.js").AsyncAction)},"./node_modules/rxjs/_esm5/internal/scheduler/QueueScheduler.js":function(t,e,n){"use strict";n.r(e),n.d(e,"QueueScheduler",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return r.__extends(e,t),e}(n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncScheduler.js").AsyncScheduler)},"./node_modules/rxjs/_esm5/internal/scheduler/VirtualTimeScheduler.js":function(t,e,n){"use strict";n.r(e),n.d(e,"VirtualTimeScheduler",function(){return o}),n.d(e,"VirtualAction",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncAction.js"),o=function(t){function e(e,n){void 0===e&&(e=a),void 0===n&&(n=Number.POSITIVE_INFINITY);var r=t.call(this,e,function(){return r.frame})||this;return r.maxFrames=n,r.frame=0,r.index=-1,r}return r.__extends(e,t),e.prototype.flush=function(){for(var t,e,n=this.actions,r=this.maxFrames;(e=n.shift())&&(this.frame=e.delay)<=r&&!(t=e.execute(e.state,e.delay)););if(t){for(;e=n.shift();)e.unsubscribe();throw t}},e.frameTimeFactor=10,e}(n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncScheduler.js").AsyncScheduler),a=function(t){function e(e,n,r){void 0===r&&(r=e.index+=1);var i=t.call(this,e,n)||this;return i.scheduler=e,i.work=n,i.index=r,i.active=!0,i.index=e.index=r,i}return r.__extends(e,t),e.prototype.schedule=function(n,r){if(void 0===r&&(r=0),!this.id)return t.prototype.schedule.call(this,n,r);this.active=!1;var i=new e(this.scheduler,this.work);return this.add(i),i.schedule(n,r)},e.prototype.requestAsyncId=function(t,n,r){void 0===r&&(r=0),this.delay=t.frame+r;var i=t.actions;return i.push(this),i.sort(e.sortActions),!0},e.prototype.recycleAsyncId=function(t,e,n){void 0===n&&(n=0)},e.prototype._execute=function(e,n){if(!0===this.active)return t.prototype._execute.call(this,e,n)},e.sortActions=function(t,e){return t.delay===e.delay?t.index===e.index?0:t.index>e.index?1:-1:t.delay>e.delay?1:-1},e}(i.AsyncAction)},"./node_modules/rxjs/_esm5/internal/scheduler/animationFrame.js":function(t,e,n){"use strict";n.r(e),n.d(e,"animationFrame",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/scheduler/AnimationFrameAction.js"),i=new(n("./node_modules/rxjs/_esm5/internal/scheduler/AnimationFrameScheduler.js").AnimationFrameScheduler)(r.AnimationFrameAction)},"./node_modules/rxjs/_esm5/internal/scheduler/asap.js":function(t,e,n){"use strict";n.r(e),n.d(e,"asap",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/scheduler/AsapAction.js"),i=new(n("./node_modules/rxjs/_esm5/internal/scheduler/AsapScheduler.js").AsapScheduler)(r.AsapAction)},"./node_modules/rxjs/_esm5/internal/scheduler/async.js":function(t,e,n){"use strict";n.r(e),n.d(e,"async",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncAction.js"),i=new(n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncScheduler.js").AsyncScheduler)(r.AsyncAction)},"./node_modules/rxjs/_esm5/internal/scheduler/queue.js":function(t,e,n){"use strict";n.r(e),n.d(e,"queue",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/scheduler/QueueAction.js"),i=new(n("./node_modules/rxjs/_esm5/internal/scheduler/QueueScheduler.js").QueueScheduler)(r.QueueAction)},"./node_modules/rxjs/_esm5/internal/symbol/iterator.js":function(t,e,n){"use strict";function r(){return"function"==typeof Symbol&&Symbol.iterator?Symbol.iterator:"@@iterator"}n.r(e),n.d(e,"getSymbolIterator",function(){return r}),n.d(e,"iterator",function(){return i}),n.d(e,"$$iterator",function(){return o});var i=r(),o=i},"./node_modules/rxjs/_esm5/internal/symbol/observable.js":function(t,e,n){"use strict";n.r(e),n.d(e,"observable",function(){return r});var r="function"==typeof Symbol&&Symbol.observable||"@@observable"},"./node_modules/rxjs/_esm5/internal/symbol/rxSubscriber.js":function(t,e,n){"use strict";n.r(e),n.d(e,"rxSubscriber",function(){return r}),n.d(e,"$$rxSubscriber",function(){return i});var r="function"==typeof Symbol?Symbol("rxSubscriber"):"@@rxSubscriber_"+Math.random(),i=r},"./node_modules/rxjs/_esm5/internal/util/ArgumentOutOfRangeError.js":function(t,e,n){"use strict";function r(){return Error.call(this),this.message="argument out of range",this.name="ArgumentOutOfRangeError",this}n.r(e),n.d(e,"ArgumentOutOfRangeError",function(){return i}),r.prototype=Object.create(Error.prototype);var i=r},"./node_modules/rxjs/_esm5/internal/util/EmptyError.js":function(t,e,n){"use strict";function r(){return Error.call(this),this.message="no elements in sequence",this.name="EmptyError",this}n.r(e),n.d(e,"EmptyError",function(){return i}),r.prototype=Object.create(Error.prototype);var i=r},"./node_modules/rxjs/_esm5/internal/util/Immediate.js":function(t,e,n){"use strict";n.r(e),n.d(e,"Immediate",function(){return o});var r=1,i={};var o={setImmediate:function(t){var e=r++;return i[e]=t,Promise.resolve().then(function(){var t;(t=i[e])&&t()}),e},clearImmediate:function(t){delete i[t]}}},"./node_modules/rxjs/_esm5/internal/util/ObjectUnsubscribedError.js":function(t,e,n){"use strict";function r(){return Error.call(this),this.message="object unsubscribed",this.name="ObjectUnsubscribedError",this}n.r(e),n.d(e,"ObjectUnsubscribedError",function(){return i}),r.prototype=Object.create(Error.prototype);var i=r},"./node_modules/rxjs/_esm5/internal/util/TimeoutError.js":function(t,e,n){"use strict";function r(){return Error.call(this),this.message="Timeout has occurred",this.name="TimeoutError",this}n.r(e),n.d(e,"TimeoutError",function(){return i}),r.prototype=Object.create(Error.prototype);var i=r},"./node_modules/rxjs/_esm5/internal/util/UnsubscriptionError.js":function(t,e,n){"use strict";function r(t){return Error.call(this),this.message=t?t.length+" errors occurred during unsubscription:\n"+t.map(function(t,e){return e+1+") "+t.toString()}).join("\n  "):"",this.name="UnsubscriptionError",this.errors=t,this}n.r(e),n.d(e,"UnsubscriptionError",function(){return i}),r.prototype=Object.create(Error.prototype);var i=r},"./node_modules/rxjs/_esm5/internal/util/canReportError.js":function(t,e,n){"use strict";n.r(e),n.d(e,"canReportError",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function i(t){for(;t;){var e=t,n=e.closed,i=e.destination,o=e.isStopped;if(n||o)return!1;t=i&&i instanceof r.Subscriber?i:null}return!0}},"./node_modules/rxjs/_esm5/internal/util/errorObject.js":function(t,e,n){"use strict";n.r(e),n.d(e,"errorObject",function(){return r});var r={e:{}}},"./node_modules/rxjs/_esm5/internal/util/hostReportError.js":function(t,e,n){"use strict";function r(t){setTimeout(function(){throw t})}n.r(e),n.d(e,"hostReportError",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/identity.js":function(t,e,n){"use strict";function r(t){return t}n.r(e),n.d(e,"identity",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/isArray.js":function(t,e,n){"use strict";n.r(e),n.d(e,"isArray",function(){return r});var r=Array.isArray||function(t){return t&&"number"==typeof t.length}},"./node_modules/rxjs/_esm5/internal/util/isArrayLike.js":function(t,e,n){"use strict";n.r(e),n.d(e,"isArrayLike",function(){return r});var r=function(t){return t&&"number"==typeof t.length&&"function"!=typeof t}},"./node_modules/rxjs/_esm5/internal/util/isDate.js":function(t,e,n){"use strict";function r(t){return t instanceof Date&&!isNaN(+t)}n.r(e),n.d(e,"isDate",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/isFunction.js":function(t,e,n){"use strict";function r(t){return"function"==typeof t}n.r(e),n.d(e,"isFunction",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/isInteropObservable.js":function(t,e,n){"use strict";n.r(e),n.d(e,"isInteropObservable",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/symbol/observable.js");function i(t){return t&&"function"==typeof t[r.observable]}},"./node_modules/rxjs/_esm5/internal/util/isIterable.js":function(t,e,n){"use strict";n.r(e),n.d(e,"isIterable",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/symbol/iterator.js");function i(t){return t&&"function"==typeof t[r.iterator]}},"./node_modules/rxjs/_esm5/internal/util/isNumeric.js":function(t,e,n){"use strict";n.r(e),n.d(e,"isNumeric",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/util/isArray.js");function i(t){return!Object(r.isArray)(t)&&t-parseFloat(t)+1>=0}},"./node_modules/rxjs/_esm5/internal/util/isObject.js":function(t,e,n){"use strict";function r(t){return null!=t&&"object"==typeof t}n.r(e),n.d(e,"isObject",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/isObservable.js":function(t,e,n){"use strict";n.r(e),n.d(e,"isObservable",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js");function i(t){return!!t&&(t instanceof r.Observable||"function"==typeof t.lift&&"function"==typeof t.subscribe)}},"./node_modules/rxjs/_esm5/internal/util/isPromise.js":function(t,e,n){"use strict";function r(t){return t&&"function"!=typeof t.subscribe&&"function"==typeof t.then}n.r(e),n.d(e,"isPromise",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/isScheduler.js":function(t,e,n){"use strict";function r(t){return t&&"function"==typeof t.schedule}n.r(e),n.d(e,"isScheduler",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/noop.js":function(t,e,n){"use strict";function r(){}n.r(e),n.d(e,"noop",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/not.js":function(t,e,n){"use strict";function r(t,e){function n(){return!n.pred.apply(n.thisArg,arguments)}return n.pred=t,n.thisArg=e,n}n.r(e),n.d(e,"not",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/pipe.js":function(t,e,n){"use strict";n.r(e),n.d(e,"pipe",function(){return i}),n.d(e,"pipeFromArray",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/util/noop.js");function i(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return o(t)}function o(t){return t?1===t.length?t[0]:function(e){return t.reduce(function(t,e){return e(t)},e)}:r.noop}},"./node_modules/rxjs/_esm5/internal/util/subscribeTo.js":function(t,e,n){"use strict";n.r(e),n.d(e,"subscribeTo",function(){return f});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/util/subscribeToArray.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToPromise.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToIterable.js"),s=n("./node_modules/rxjs/_esm5/internal/util/subscribeToObservable.js"),u=n("./node_modules/rxjs/_esm5/internal/util/isArrayLike.js"),c=n("./node_modules/rxjs/_esm5/internal/util/isPromise.js"),l=n("./node_modules/rxjs/_esm5/internal/util/isObject.js"),d=n("./node_modules/rxjs/_esm5/internal/symbol/iterator.js"),p=n("./node_modules/rxjs/_esm5/internal/symbol/observable.js"),f=function(t){if(t instanceof r.Observable)return function(e){return t._isScalar?(e.next(t.value),void e.complete()):t.subscribe(e)};if(t&&"function"==typeof t[p.observable])return Object(s.subscribeToObservable)(t);if(Object(u.isArrayLike)(t))return Object(i.subscribeToArray)(t);if(Object(c.isPromise)(t))return Object(o.subscribeToPromise)(t);if(t&&"function"==typeof t[d.iterator])return Object(a.subscribeToIterable)(t);var e=Object(l.isObject)(t)?"an invalid object":"'"+t+"'";throw new TypeError("You provided "+e+" where a stream was expected. You can provide an Observable, Promise, Array, or Iterable.")}},"./node_modules/rxjs/_esm5/internal/util/subscribeToArray.js":function(t,e,n){"use strict";n.r(e),n.d(e,"subscribeToArray",function(){return r});var r=function(t){return function(e){for(var n=0,r=t.length;n<r&&!e.closed;n++)e.next(t[n]);e.closed||e.complete()}}},"./node_modules/rxjs/_esm5/internal/util/subscribeToIterable.js":function(t,e,n){"use strict";n.r(e),n.d(e,"subscribeToIterable",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/symbol/iterator.js"),i=function(t){return function(e){for(var n=t[r.iterator]();;){var i=n.next();if(i.done){e.complete();break}if(e.next(i.value),e.closed)break}return"function"==typeof n.return&&e.add(function(){n.return&&n.return()}),e}}},"./node_modules/rxjs/_esm5/internal/util/subscribeToObservable.js":function(t,e,n){"use strict";n.r(e),n.d(e,"subscribeToObservable",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/symbol/observable.js"),i=function(t){return function(e){var n=t[r.observable]();if("function"!=typeof n.subscribe)throw new TypeError("Provided object does not correctly implement Symbol.observable");return n.subscribe(e)}}},"./node_modules/rxjs/_esm5/internal/util/subscribeToPromise.js":function(t,e,n){"use strict";n.r(e),n.d(e,"subscribeToPromise",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/util/hostReportError.js"),i=function(t){return function(e){return t.then(function(t){e.closed||(e.next(t),e.complete())},function(t){return e.error(t)}).then(null,r.hostReportError),e}}},"./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js":function(t,e,n){"use strict";n.r(e),n.d(e,"subscribeToResult",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/InnerSubscriber.js"),i=n("./node_modules/rxjs/_esm5/internal/util/subscribeTo.js");function o(t,e,n,o,a){if(void 0===a&&(a=new r.InnerSubscriber(t,n,o)),!a.closed)return Object(i.subscribeTo)(e)(a)}},"./node_modules/rxjs/_esm5/internal/util/toSubscriber.js":function(t,e,n){"use strict";n.r(e),n.d(e,"toSubscriber",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),i=n("./node_modules/rxjs/_esm5/internal/symbol/rxSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/Observer.js");function a(t,e,n){if(t){if(t instanceof r.Subscriber)return t;if(t[i.rxSubscriber])return t[i.rxSubscriber]()}return t||e||n?new r.Subscriber(t,e,n):new r.Subscriber(o.empty)}},"./node_modules/rxjs/_esm5/internal/util/tryCatch.js":function(t,e,n){"use strict";n.r(e),n.d(e,"tryCatch",function(){return a});var r,i=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js");function o(){try{return r.apply(this,arguments)}catch(t){return i.errorObject.e=t,i.errorObject}}function a(t){return r=t,o}},"./node_modules/rxjs/_esm5/operators/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/rxjs/_esm5/internal/operators/audit.js");n.d(e,"audit",function(){return r.audit});var i=n("./node_modules/rxjs/_esm5/internal/operators/auditTime.js");n.d(e,"auditTime",function(){return i.auditTime});var o=n("./node_modules/rxjs/_esm5/internal/operators/buffer.js");n.d(e,"buffer",function(){return o.buffer});var a=n("./node_modules/rxjs/_esm5/internal/operators/bufferCount.js");n.d(e,"bufferCount",function(){return a.bufferCount});var s=n("./node_modules/rxjs/_esm5/internal/operators/bufferTime.js");n.d(e,"bufferTime",function(){return s.bufferTime});var u=n("./node_modules/rxjs/_esm5/internal/operators/bufferToggle.js");n.d(e,"bufferToggle",function(){return u.bufferToggle});var c=n("./node_modules/rxjs/_esm5/internal/operators/bufferWhen.js");n.d(e,"bufferWhen",function(){return c.bufferWhen});var l=n("./node_modules/rxjs/_esm5/internal/operators/catchError.js");n.d(e,"catchError",function(){return l.catchError});var d=n("./node_modules/rxjs/_esm5/internal/operators/combineAll.js");n.d(e,"combineAll",function(){return d.combineAll});var p=n("./node_modules/rxjs/_esm5/internal/operators/combineLatest.js");n.d(e,"combineLatest",function(){return p.combineLatest});var f=n("./node_modules/rxjs/_esm5/internal/operators/concat.js");n.d(e,"concat",function(){return f.concat});var h=n("./node_modules/rxjs/_esm5/internal/operators/concatAll.js");n.d(e,"concatAll",function(){return h.concatAll});var m=n("./node_modules/rxjs/_esm5/internal/operators/concatMap.js");n.d(e,"concatMap",function(){return m.concatMap});var g=n("./node_modules/rxjs/_esm5/internal/operators/concatMapTo.js");n.d(e,"concatMapTo",function(){return g.concatMapTo});var y=n("./node_modules/rxjs/_esm5/internal/operators/count.js");n.d(e,"count",function(){return y.count});var _=n("./node_modules/rxjs/_esm5/internal/operators/debounce.js");n.d(e,"debounce",function(){return _.debounce});var v=n("./node_modules/rxjs/_esm5/internal/operators/debounceTime.js");n.d(e,"debounceTime",function(){return v.debounceTime});var b=n("./node_modules/rxjs/_esm5/internal/operators/defaultIfEmpty.js");n.d(e,"defaultIfEmpty",function(){return b.defaultIfEmpty});var x=n("./node_modules/rxjs/_esm5/internal/operators/delay.js");n.d(e,"delay",function(){return x.delay});var w=n("./node_modules/rxjs/_esm5/internal/operators/delayWhen.js");n.d(e,"delayWhen",function(){return w.delayWhen});var j=n("./node_modules/rxjs/_esm5/internal/operators/dematerialize.js");n.d(e,"dematerialize",function(){return j.dematerialize});var O=n("./node_modules/rxjs/_esm5/internal/operators/distinct.js");n.d(e,"distinct",function(){return O.distinct});var C=n("./node_modules/rxjs/_esm5/internal/operators/distinctUntilChanged.js");n.d(e,"distinctUntilChanged",function(){return C.distinctUntilChanged});var E=n("./node_modules/rxjs/_esm5/internal/operators/distinctUntilKeyChanged.js");n.d(e,"distinctUntilKeyChanged",function(){return E.distinctUntilKeyChanged});var S=n("./node_modules/rxjs/_esm5/internal/operators/elementAt.js");n.d(e,"elementAt",function(){return S.elementAt});var M=n("./node_modules/rxjs/_esm5/internal/operators/endWith.js");n.d(e,"endWith",function(){return M.endWith});var T=n("./node_modules/rxjs/_esm5/internal/operators/every.js");n.d(e,"every",function(){return T.every});var k=n("./node_modules/rxjs/_esm5/internal/operators/exhaust.js");n.d(e,"exhaust",function(){return k.exhaust});var A=n("./node_modules/rxjs/_esm5/internal/operators/exhaustMap.js");n.d(e,"exhaustMap",function(){return A.exhaustMap});var D=n("./node_modules/rxjs/_esm5/internal/operators/expand.js");n.d(e,"expand",function(){return D.expand});var I=n("./node_modules/rxjs/_esm5/internal/operators/filter.js");n.d(e,"filter",function(){return I.filter});var P=n("./node_modules/rxjs/_esm5/internal/operators/finalize.js");n.d(e,"finalize",function(){return P.finalize});var R=n("./node_modules/rxjs/_esm5/internal/operators/find.js");n.d(e,"find",function(){return R.find});var N=n("./node_modules/rxjs/_esm5/internal/operators/findIndex.js");n.d(e,"findIndex",function(){return N.findIndex});var F=n("./node_modules/rxjs/_esm5/internal/operators/first.js");n.d(e,"first",function(){return F.first});var L=n("./node_modules/rxjs/_esm5/internal/operators/groupBy.js");n.d(e,"groupBy",function(){return L.groupBy});var B=n("./node_modules/rxjs/_esm5/internal/operators/ignoreElements.js");n.d(e,"ignoreElements",function(){return B.ignoreElements});var V=n("./node_modules/rxjs/_esm5/internal/operators/isEmpty.js");n.d(e,"isEmpty",function(){return V.isEmpty});var z=n("./node_modules/rxjs/_esm5/internal/operators/last.js");n.d(e,"last",function(){return z.last});var H=n("./node_modules/rxjs/_esm5/internal/operators/map.js");n.d(e,"map",function(){return H.map});var U=n("./node_modules/rxjs/_esm5/internal/operators/mapTo.js");n.d(e,"mapTo",function(){return U.mapTo});var W=n("./node_modules/rxjs/_esm5/internal/operators/materialize.js");n.d(e,"materialize",function(){return W.materialize});var q=n("./node_modules/rxjs/_esm5/internal/operators/max.js");n.d(e,"max",function(){return q.max});var G=n("./node_modules/rxjs/_esm5/internal/operators/merge.js");n.d(e,"merge",function(){return G.merge});var Y=n("./node_modules/rxjs/_esm5/internal/operators/mergeAll.js");n.d(e,"mergeAll",function(){return Y.mergeAll});var K=n("./node_modules/rxjs/_esm5/internal/operators/mergeMap.js");n.d(e,"mergeMap",function(){return K.mergeMap}),n.d(e,"flatMap",function(){return K.mergeMap});var X=n("./node_modules/rxjs/_esm5/internal/operators/mergeMapTo.js");n.d(e,"mergeMapTo",function(){return X.mergeMapTo});var Q=n("./node_modules/rxjs/_esm5/internal/operators/mergeScan.js");n.d(e,"mergeScan",function(){return Q.mergeScan});var Z=n("./node_modules/rxjs/_esm5/internal/operators/min.js");n.d(e,"min",function(){return Z.min});var $=n("./node_modules/rxjs/_esm5/internal/operators/multicast.js");n.d(e,"multicast",function(){return $.multicast});var J=n("./node_modules/rxjs/_esm5/internal/operators/observeOn.js");n.d(e,"observeOn",function(){return J.observeOn});var tt=n("./node_modules/rxjs/_esm5/internal/operators/onErrorResumeNext.js");n.d(e,"onErrorResumeNext",function(){return tt.onErrorResumeNext});var et=n("./node_modules/rxjs/_esm5/internal/operators/pairwise.js");n.d(e,"pairwise",function(){return et.pairwise});var nt=n("./node_modules/rxjs/_esm5/internal/operators/partition.js");n.d(e,"partition",function(){return nt.partition});var rt=n("./node_modules/rxjs/_esm5/internal/operators/pluck.js");n.d(e,"pluck",function(){return rt.pluck});var it=n("./node_modules/rxjs/_esm5/internal/operators/publish.js");n.d(e,"publish",function(){return it.publish});var ot=n("./node_modules/rxjs/_esm5/internal/operators/publishBehavior.js");n.d(e,"publishBehavior",function(){return ot.publishBehavior});var at=n("./node_modules/rxjs/_esm5/internal/operators/publishLast.js");n.d(e,"publishLast",function(){return at.publishLast});var st=n("./node_modules/rxjs/_esm5/internal/operators/publishReplay.js");n.d(e,"publishReplay",function(){return st.publishReplay});var ut=n("./node_modules/rxjs/_esm5/internal/operators/race.js");n.d(e,"race",function(){return ut.race});var ct=n("./node_modules/rxjs/_esm5/internal/operators/reduce.js");n.d(e,"reduce",function(){return ct.reduce});var lt=n("./node_modules/rxjs/_esm5/internal/operators/repeat.js");n.d(e,"repeat",function(){return lt.repeat});var dt=n("./node_modules/rxjs/_esm5/internal/operators/repeatWhen.js");n.d(e,"repeatWhen",function(){return dt.repeatWhen});var pt=n("./node_modules/rxjs/_esm5/internal/operators/retry.js");n.d(e,"retry",function(){return pt.retry});var ft=n("./node_modules/rxjs/_esm5/internal/operators/retryWhen.js");n.d(e,"retryWhen",function(){return ft.retryWhen});var ht=n("./node_modules/rxjs/_esm5/internal/operators/refCount.js");n.d(e,"refCount",function(){return ht.refCount});var mt=n("./node_modules/rxjs/_esm5/internal/operators/sample.js");n.d(e,"sample",function(){return mt.sample});var gt=n("./node_modules/rxjs/_esm5/internal/operators/sampleTime.js");n.d(e,"sampleTime",function(){return gt.sampleTime});var yt=n("./node_modules/rxjs/_esm5/internal/operators/scan.js");n.d(e,"scan",function(){return yt.scan});var _t=n("./node_modules/rxjs/_esm5/internal/operators/sequenceEqual.js");n.d(e,"sequenceEqual",function(){return _t.sequenceEqual});var vt=n("./node_modules/rxjs/_esm5/internal/operators/share.js");n.d(e,"share",function(){return vt.share});var bt=n("./node_modules/rxjs/_esm5/internal/operators/shareReplay.js");n.d(e,"shareReplay",function(){return bt.shareReplay});var xt=n("./node_modules/rxjs/_esm5/internal/operators/single.js");n.d(e,"single",function(){return xt.single});var wt=n("./node_modules/rxjs/_esm5/internal/operators/skip.js");n.d(e,"skip",function(){return wt.skip});var jt=n("./node_modules/rxjs/_esm5/internal/operators/skipLast.js");n.d(e,"skipLast",function(){return jt.skipLast});var Ot=n("./node_modules/rxjs/_esm5/internal/operators/skipUntil.js");n.d(e,"skipUntil",function(){return Ot.skipUntil});var Ct=n("./node_modules/rxjs/_esm5/internal/operators/skipWhile.js");n.d(e,"skipWhile",function(){return Ct.skipWhile});var Et=n("./node_modules/rxjs/_esm5/internal/operators/startWith.js");n.d(e,"startWith",function(){return Et.startWith});var St=n("./node_modules/rxjs/_esm5/internal/operators/subscribeOn.js");n.d(e,"subscribeOn",function(){return St.subscribeOn});var Mt=n("./node_modules/rxjs/_esm5/internal/operators/switchAll.js");n.d(e,"switchAll",function(){return Mt.switchAll});var Tt=n("./node_modules/rxjs/_esm5/internal/operators/switchMap.js");n.d(e,"switchMap",function(){return Tt.switchMap});var kt=n("./node_modules/rxjs/_esm5/internal/operators/switchMapTo.js");n.d(e,"switchMapTo",function(){return kt.switchMapTo});var At=n("./node_modules/rxjs/_esm5/internal/operators/take.js");n.d(e,"take",function(){return At.take});var Dt=n("./node_modules/rxjs/_esm5/internal/operators/takeLast.js");n.d(e,"takeLast",function(){return Dt.takeLast});var It=n("./node_modules/rxjs/_esm5/internal/operators/takeUntil.js");n.d(e,"takeUntil",function(){return It.takeUntil});var Pt=n("./node_modules/rxjs/_esm5/internal/operators/takeWhile.js");n.d(e,"takeWhile",function(){return Pt.takeWhile});var Rt=n("./node_modules/rxjs/_esm5/internal/operators/tap.js");n.d(e,"tap",function(){return Rt.tap});var Nt=n("./node_modules/rxjs/_esm5/internal/operators/throttle.js");n.d(e,"throttle",function(){return Nt.throttle});var Ft=n("./node_modules/rxjs/_esm5/internal/operators/throttleTime.js");n.d(e,"throttleTime",function(){return Ft.throttleTime});var Lt=n("./node_modules/rxjs/_esm5/internal/operators/throwIfEmpty.js");n.d(e,"throwIfEmpty",function(){return Lt.throwIfEmpty});var Bt=n("./node_modules/rxjs/_esm5/internal/operators/timeInterval.js");n.d(e,"timeInterval",function(){return Bt.timeInterval});var Vt=n("./node_modules/rxjs/_esm5/internal/operators/timeout.js");n.d(e,"timeout",function(){return Vt.timeout});var zt=n("./node_modules/rxjs/_esm5/internal/operators/timeoutWith.js");n.d(e,"timeoutWith",function(){return zt.timeoutWith});var Ht=n("./node_modules/rxjs/_esm5/internal/operators/timestamp.js");n.d(e,"timestamp",function(){return Ht.timestamp});var Ut=n("./node_modules/rxjs/_esm5/internal/operators/toArray.js");n.d(e,"toArray",function(){return Ut.toArray});var Wt=n("./node_modules/rxjs/_esm5/internal/operators/window.js");n.d(e,"window",function(){return Wt.window});var qt=n("./node_modules/rxjs/_esm5/internal/operators/windowCount.js");n.d(e,"windowCount",function(){return qt.windowCount});var Gt=n("./node_modules/rxjs/_esm5/internal/operators/windowTime.js");n.d(e,"windowTime",function(){return Gt.windowTime});var Yt=n("./node_modules/rxjs/_esm5/internal/operators/windowToggle.js");n.d(e,"windowToggle",function(){return Yt.windowToggle});var Kt=n("./node_modules/rxjs/_esm5/internal/operators/windowWhen.js");n.d(e,"windowWhen",function(){return Kt.windowWhen});var Xt=n("./node_modules/rxjs/_esm5/internal/operators/withLatestFrom.js");n.d(e,"withLatestFrom",function(){return Xt.withLatestFrom});var Qt=n("./node_modules/rxjs/_esm5/internal/operators/zip.js");n.d(e,"zip",function(){return Qt.zip});var Zt=n("./node_modules/rxjs/_esm5/internal/operators/zipAll.js");n.d(e,"zipAll",function(){return Zt.zipAll})},"./node_modules/rxjs/node_modules/tslib/tslib.es6.js":function(t,e,n){"use strict";n.r(e),n.d(e,"__extends",function(){return i}),n.d(e,"__assign",function(){return o}),n.d(e,"__rest",function(){return a}),n.d(e,"__decorate",function(){return s}),n.d(e,"__param",function(){return u}),n.d(e,"__metadata",function(){return c}),n.d(e,"__awaiter",function(){return l}),n.d(e,"__generator",function(){return d}),n.d(e,"__exportStar",function(){return p}),n.d(e,"__values",function(){return f}),n.d(e,"__read",function(){return h}),n.d(e,"__spread",function(){return m}),n.d(e,"__await",function(){return g}),n.d(e,"__asyncGenerator",function(){return y}),n.d(e,"__asyncDelegator",function(){return _}),n.d(e,"__asyncValues",function(){return v}),n.d(e,"__makeTemplateObject",function(){return b}),n.d(e,"__importStar",function(){return x}),n.d(e,"__importDefault",function(){return w});var r=function(t,e){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)};function i(t,e){function n(){this.constructor=t}r(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}var o=function(){return(o=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++){e=arguments[n];for(var i in e)Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i])}return t}).apply(this,arguments)};function a(t,e){var n={};for(var r in t)Object.prototype.hasOwnProperty.call(t,r)&&e.indexOf(r)<0&&(n[r]=t[r]);if(null!=t&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(t);i<r.length;i++)e.indexOf(r[i])<0&&(n[r[i]]=t[r[i]])}return n}function s(t,e,n,r){var i,o=arguments.length,a=o<3?e:null===r?r=Object.getOwnPropertyDescriptor(e,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(t,e,n,r);else for(var s=t.length-1;s>=0;s--)(i=t[s])&&(a=(o<3?i(a):o>3?i(e,n,a):i(e,n))||a);return o>3&&a&&Object.defineProperty(e,n,a),a}function u(t,e){return function(n,r){e(n,r,t)}}function c(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)}function l(t,e,n,r){return new(n||(n=Promise))(function(i,o){function a(t){try{u(r.next(t))}catch(t){o(t)}}function s(t){try{u(r.throw(t))}catch(t){o(t)}}function u(t){t.done?i(t.value):new n(function(e){e(t.value)}).then(a,s)}u((r=r.apply(t,e||[])).next())})}function d(t,e){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=e.call(t,a)}catch(t){o=[6,t],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(t,e){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}function f(t){var e="function"==typeof Symbol&&t[Symbol.iterator],n=0;return e?e.call(t):{next:function(){return t&&n>=t.length&&(t=void 0),{value:t&&t[n++],done:!t}}}}function h(t,e){var n="function"==typeof Symbol&&t[Symbol.iterator];if(!n)return t;var r,i,o=n.call(t),a=[];try{for(;(void 0===e||e-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(t){i={error:t}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var t=[],e=0;e<arguments.length;e++)t=t.concat(h(arguments[e]));return t}function g(t){return this instanceof g?(this.v=t,this):new g(t)}function y(t,e,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(t,e||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(t){i[t]&&(r[t]=function(e){return new Promise(function(n,r){o.push([t,e,n,r])>1||s(t,e)})})}function s(t,e){try{(n=i[t](e)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(t){l(o[0][3],t)}var n}function u(t){s("next",t)}function c(t){s("throw",t)}function l(t,e){t(e),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(t){var e,n;return e={},r("next"),r("throw",function(t){throw t}),r("return"),e[Symbol.iterator]=function(){return this},e;function r(r,i){e[r]=t[r]?function(e){return(n=!n)?{value:g(t[r](e)),done:"return"===r}:i?i(e):e}:i}}function v(t){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var e,n=t[Symbol.asyncIterator];return n?n.call(t):(t=f(t),e={},r("next"),r("throw"),r("return"),e[Symbol.asyncIterator]=function(){return this},e);function r(n){e[n]=t[n]&&function(e){return new Promise(function(r,i){var o,a,s,u;e=t[n](e),o=r,a=i,s=e.done,u=e.value,Promise.resolve(u).then(function(t){o({value:t,done:s})},a)})}}}function b(t,e){return Object.defineProperty?Object.defineProperty(t,"raw",{value:e}):t.raw=e,t}function x(t){if(t&&t.__esModule)return t;var e={};if(null!=t)for(var n in t)Object.hasOwnProperty.call(t,n)&&(e[n]=t[n]);return e.default=t,e}function w(t){return t&&t.__esModule?t:{default:t}}},"./node_modules/taucharts/dist/plugins/export-to.js":function(t,e,n){var r,i;window,t.exports=(r=n("./node_modules/taucharts/dist/taucharts.js"),i=n("./node_modules/d3-selection/src/index.js"),function(t){var e={};function n(r){if(e[r])return e[r].exports;var i=e[r]={i:r,l:!1,exports:{}};return t[r].call(i.exports,i,i.exports,n),i.l=!0,i.exports}return n.m=t,n.c=e,n.d=function(t,e,r){n.o(t,e)||Object.defineProperty(t,e,{enumerable:!0,get:r})},n.r=function(t){"undefined"!=typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(t,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(t,"__esModule",{value:!0})},n.t=function(t,e){if(1&e&&(t=n(t)),8&e)return t;if(4&e&&"object"==typeof t&&t&&t.__esModule)return t;var r=Object.create(null);if(n.r(r),Object.defineProperty(r,"default",{enumerable:!0,value:t}),2&e&&"string"!=typeof t)for(var i in t)n.d(r,i,function(e){return t[e]}.bind(null,i));return r},n.n=function(t){var e=t&&t.__esModule?function(){return t.default}:function(){return t};return n.d(e,"a",e),e},n.o=function(t,e){return Object.prototype.hasOwnProperty.call(t,e)},n.p="",n(n.s=19)}([function(t,e){t.exports=r},function(t,e){t.exports=i},,,,,function(t,e,n){t.exports=function(t,e){"use strict";var n;return t=t&&t.hasOwnProperty("default")?t.default:t,e=e&&e.hasOwnProperty("default")?e.default:e,function(n){var r;function i(){return document.createElement("canvas")}n.exports,(r=window).DOMParser=window.DOMParser;var o,a=function(n,s,u){if(null!=n||null!=s||null!=u){var c=function(n){var a={opts:n,FRAMERATE:30,MAX_VIRTUAL_PIXELS:3e4,rootEmSize:12,emSize:12,log:function(t){}};1==a.opts.log&&"undefined"!=typeof console&&(a.log=function(t){console.log(t)}),a.init=function(t){var e=0;a.UniqueId=function(){return"canvg"+ ++e},a.Definitions={},a.Styles={},a.StylesSpecificity={},a.Animations=[],a.Images=[],a.ctx=t,a.ViewPort=new function(){this.viewPorts=[],this.Clear=function(){this.viewPorts=[]},this.SetCurrent=function(t,e){this.viewPorts.push({width:t,height:e})},this.RemoveCurrent=function(){this.viewPorts.pop()},this.Current=function(){return this.viewPorts[this.viewPorts.length-1]},this.width=function(){return this.Current().width},this.height=function(){return this.Current().height},this.ComputeSize=function(t){return null!=t&&"number"==typeof t?t:"x"==t?this.width():"y"==t?this.height():Math.sqrt(Math.pow(this.width(),2)+Math.pow(this.height(),2))/Math.sqrt(2)}}},a.init(),a.ImagesLoaded=function(){for(var t=0;t<a.Images.length;t++)if(!a.Images[t].loaded)return!1;return!0},a.trim=function(t){return t.replace(/^\s+|\s+$/g,"")},a.compressSpaces=function(t){return t.replace(/(?!\u3000)\s+/gm," ")},a.ajax=function(t){var e;return(e=r.XMLHttpRequest?new r.XMLHttpRequest:new ActiveXObject("Microsoft.XMLHTTP"))?(e.open("GET",t,!1),e.send(null),e.responseText):null},a.parseXml=function(t){if("undefined"!=typeof Windows&&void 0!==Windows.Data&&void 0!==Windows.Data.Xml){var e=new Windows.Data.Xml.Dom.XmlDocument,i=new Windows.Data.Xml.Dom.XmlLoadSettings;return i.prohibitDtd=!1,e.loadXml(t,i),e}if(!r.DOMParser){t=t.replace(/<!DOCTYPE svg[^>]*>/,"");var e=new ActiveXObject("Microsoft.XMLDOM");return e.async="false",e.loadXML(t),e}try{var o=n.xmldom?new r.DOMParser(n.xmldom):new r.DOMParser;return o.parseFromString(t,"image/svg+xml")}catch(e){return(o=n.xmldom?new r.DOMParser(n.xmldom):new r.DOMParser).parseFromString(t,"text/xml")}},a.Property=function(t,e){this.name=t,this.value=e},a.Property.prototype.getValue=function(){return this.value},a.Property.prototype.hasValue=function(){return null!=this.value&&""!==this.value},a.Property.prototype.numValue=function(){if(!this.hasValue())return 0;var t=parseFloat(this.value);return(this.value+"").match(/%$/)&&(t/=100),t},a.Property.prototype.valueOrDefault=function(t){return this.hasValue()?this.value:t},a.Property.prototype.numValueOrDefault=function(t){return this.hasValue()?this.numValue():t},a.Property.prototype.addOpacity=function(e){var n=this.value;if(null!=e.value&&""!=e.value&&"string"==typeof this.value){var r=new t(this.value);r.ok&&(n="rgba("+r.r+", "+r.g+", "+r.b+", "+e.numValue()+")")}return new a.Property(this.name,n)},a.Property.prototype.getDefinition=function(){var t=this.value.match(/#([^\)'"]+)/);return t&&(t=t[1]),t||(t=this.value),a.Definitions[t]},a.Property.prototype.isUrlDefinition=function(){return 0==this.value.indexOf("url(")},a.Property.prototype.getFillStyleDefinition=function(t,e){var n=this.getDefinition();if(null!=n&&n.createGradient)return n.createGradient(a.ctx,t,e);if(null!=n&&n.createPattern){if(n.getHrefAttribute().hasValue()){var r=n.attribute("patternTransform");n=n.getHrefAttribute().getDefinition(),r.hasValue()&&(n.attribute("patternTransform",!0).value=r.value)}return n.createPattern(a.ctx,t)}return null},a.Property.prototype.getDPI=function(t){return 96},a.Property.prototype.getREM=function(t){return a.rootEmSize},a.Property.prototype.getEM=function(t){return a.emSize},a.Property.prototype.getUnits=function(){var t=this.value+"";return t.replace(/[0-9\.\-]/g,"")},a.Property.prototype.isPixels=function(){if(!this.hasValue())return!1;var t=this.value+"";return!!t.match(/px$/)||!!t.match(/^[0-9]+$/)},a.Property.prototype.toPixels=function(t,e){if(!this.hasValue())return 0;var n=this.value+"";if(n.match(/rem$/))return this.numValue()*this.getREM(t);if(n.match(/em$/))return this.numValue()*this.getEM(t);if(n.match(/ex$/))return this.numValue()*this.getEM(t)/2;if(n.match(/px$/))return this.numValue();if(n.match(/pt$/))return this.numValue()*this.getDPI(t)*(1/72);if(n.match(/pc$/))return 15*this.numValue();if(n.match(/cm$/))return this.numValue()*this.getDPI(t)/2.54;if(n.match(/mm$/))return this.numValue()*this.getDPI(t)/25.4;if(n.match(/in$/))return this.numValue()*this.getDPI(t);if(n.match(/%$/))return this.numValue()*a.ViewPort.ComputeSize(t);var r=this.numValue();return e&&r<1?r*a.ViewPort.ComputeSize(t):r},a.Property.prototype.toMilliseconds=function(){if(!this.hasValue())return 0;var t=this.value+"";return t.match(/s$/)?1e3*this.numValue():(t.match(/ms$/),this.numValue())},a.Property.prototype.toRadians=function(){if(!this.hasValue())return 0;var t=this.value+"";return t.match(/deg$/)?this.numValue()*(Math.PI/180):t.match(/grad$/)?this.numValue()*(Math.PI/200):t.match(/rad$/)?this.numValue():this.numValue()*(Math.PI/180)};var s={baseline:"alphabetic","before-edge":"top","text-before-edge":"top",middle:"middle",central:"middle","after-edge":"bottom","text-after-edge":"bottom",ideographic:"ideographic",alphabetic:"alphabetic",hanging:"hanging",mathematical:"alphabetic"};return a.Property.prototype.toTextBaseline=function(){return this.hasValue()?s[this.value]:null},a.Font=new function(){this.Styles="normal|italic|oblique|inherit",this.Variants="normal|small-caps|inherit",this.Weights="normal|bold|bolder|lighter|100|200|300|400|500|600|700|800|900|inherit",this.CreateFont=function(t,e,n,r,i,o){var s=null!=o?this.Parse(o):this.CreateFont("","","","","",a.ctx.font);return{fontFamily:i=i||s.fontFamily,fontSize:r||s.fontSize,fontStyle:t||s.fontStyle,fontWeight:n||s.fontWeight,fontVariant:e||s.fontVariant,toString:function(){return[this.fontStyle,this.fontVariant,this.fontWeight,this.fontSize,this.fontFamily].join(" ")}}};var t=this;this.Parse=function(e){for(var n={},r=a.trim(a.compressSpaces(e||"")).split(" "),i={fontSize:!1,fontStyle:!1,fontWeight:!1,fontVariant:!1},o="",s=0;s<r.length;s++)i.fontStyle||-1==t.Styles.indexOf(r[s])?i.fontVariant||-1==t.Variants.indexOf(r[s])?i.fontWeight||-1==t.Weights.indexOf(r[s])?i.fontSize?"inherit"!=r[s]&&(o+=r[s]):("inherit"!=r[s]&&(n.fontSize=r[s].split("/")[0]),i.fontStyle=i.fontVariant=i.fontWeight=i.fontSize=!0):("inherit"!=r[s]&&(n.fontWeight=r[s]),i.fontStyle=i.fontVariant=i.fontWeight=!0):("inherit"!=r[s]&&(n.fontVariant=r[s]),i.fontStyle=i.fontVariant=!0):("inherit"!=r[s]&&(n.fontStyle=r[s]),i.fontStyle=!0);return""!=o&&(n.fontFamily=o),n}},a.ToNumberArray=function(t){for(var e=a.trim(a.compressSpaces((t||"").replace(/,/g," "))).split(" "),n=0;n<e.length;n++)e[n]=parseFloat(e[n]);return e},a.Point=function(t,e){this.x=t,this.y=e},a.Point.prototype.angleTo=function(t){return Math.atan2(t.y-this.y,t.x-this.x)},a.Point.prototype.applyTransform=function(t){var e=this.x*t[0]+this.y*t[2]+t[4],n=this.x*t[1]+this.y*t[3]+t[5];this.x=e,this.y=n},a.CreatePoint=function(t){var e=a.ToNumberArray(t);return new a.Point(e[0],e[1])},a.CreatePath=function(t){for(var e=a.ToNumberArray(t),n=[],r=0;r<e.length;r+=2)n.push(new a.Point(e[r],e[r+1]));return n},a.BoundingBox=function(t,e,n,r){this.x1=Number.NaN,this.y1=Number.NaN,this.x2=Number.NaN,this.y2=Number.NaN,this.x=function(){return this.x1},this.y=function(){return this.y1},this.width=function(){return this.x2-this.x1},this.height=function(){return this.y2-this.y1},this.addPoint=function(t,e){null!=t&&((isNaN(this.x1)||isNaN(this.x2))&&(this.x1=t,this.x2=t),t<this.x1&&(this.x1=t),t>this.x2&&(this.x2=t)),null!=e&&((isNaN(this.y1)||isNaN(this.y2))&&(this.y1=e,this.y2=e),e<this.y1&&(this.y1=e),e>this.y2&&(this.y2=e))},this.addX=function(t){this.addPoint(t,null)},this.addY=function(t){this.addPoint(null,t)},this.addBoundingBox=function(t){this.addPoint(t.x1,t.y1),this.addPoint(t.x2,t.y2)},this.addQuadraticCurve=function(t,e,n,r,i,o){var a=t+2/3*(n-t),s=e+2/3*(r-e),u=a+1/3*(i-t),c=s+1/3*(o-e);this.addBezierCurve(t,e,a,u,s,c,i,o)},this.addBezierCurve=function(t,e,n,r,i,o,a,s){var u=[t,e],c=[n,r],l=[i,o],d=[a,s];this.addPoint(u[0],u[1]),this.addPoint(d[0],d[1]);for(var p=0;p<=1;p++){var f=function(t){return Math.pow(1-t,3)*u[p]+3*Math.pow(1-t,2)*t*c[p]+3*(1-t)*Math.pow(t,2)*l[p]+Math.pow(t,3)*d[p]},h=6*u[p]-12*c[p]+6*l[p],m=-3*u[p]+9*c[p]-9*l[p]+3*d[p],g=3*c[p]-3*u[p];if(0!=m){var y=Math.pow(h,2)-4*g*m;if(!(y<0)){var _=(-h+Math.sqrt(y))/(2*m);0<_&&_<1&&(0==p&&this.addX(f(_)),1==p&&this.addY(f(_)));var v=(-h-Math.sqrt(y))/(2*m);0<v&&v<1&&(0==p&&this.addX(f(v)),1==p&&this.addY(f(v)))}}else{if(0==h)continue;var b=-g/h;0<b&&b<1&&(0==p&&this.addX(f(b)),1==p&&this.addY(f(b)))}}},this.isPointInBox=function(t,e){return this.x1<=t&&t<=this.x2&&this.y1<=e&&e<=this.y2},this.addPoint(t,e),this.addPoint(n,r)},a.Transform=function(t){var e=this;this.Type={},this.Type.translate=function(t){this.p=a.CreatePoint(t),this.apply=function(t){t.translate(this.p.x||0,this.p.y||0)},this.unapply=function(t){t.translate(-1*this.p.x||0,-1*this.p.y||0)},this.applyToPoint=function(t){t.applyTransform([1,0,0,1,this.p.x||0,this.p.y||0])}},this.Type.rotate=function(t){var e=a.ToNumberArray(t);this.angle=new a.Property("angle",e[0]),this.cx=e[1]||0,this.cy=e[2]||0,this.apply=function(t){t.translate(this.cx,this.cy),t.rotate(this.angle.toRadians()),t.translate(-this.cx,-this.cy)},this.unapply=function(t){t.translate(this.cx,this.cy),t.rotate(-1*this.angle.toRadians()),t.translate(-this.cx,-this.cy)},this.applyToPoint=function(t){var e=this.angle.toRadians();t.applyTransform([1,0,0,1,this.p.x||0,this.p.y||0]),t.applyTransform([Math.cos(e),Math.sin(e),-Math.sin(e),Math.cos(e),0,0]),t.applyTransform([1,0,0,1,-this.p.x||0,-this.p.y||0])}},this.Type.scale=function(t){this.p=a.CreatePoint(t),this.apply=function(t){t.scale(this.p.x||1,this.p.y||this.p.x||1)},this.unapply=function(t){t.scale(1/this.p.x||1,1/this.p.y||this.p.x||1)},this.applyToPoint=function(t){t.applyTransform([this.p.x||0,0,0,this.p.y||0,0,0])}},this.Type.matrix=function(t){this.m=a.ToNumberArray(t),this.apply=function(t){t.transform(this.m[0],this.m[1],this.m[2],this.m[3],this.m[4],this.m[5])},this.unapply=function(t){var e=this.m[0],n=this.m[2],r=this.m[4],i=this.m[1],o=this.m[3],a=this.m[5],s=1/(e*(1*o-0*a)-n*(1*i-0*a)+r*(0*i-0*o));t.transform(s*(1*o-0*a),s*(0*a-1*i),s*(0*r-1*n),s*(1*e-0*r),s*(n*a-r*o),s*(r*i-e*a))},this.applyToPoint=function(t){t.applyTransform(this.m)}},this.Type.SkewBase=function(t){this.base=e.Type.matrix,this.base(t),this.angle=new a.Property("angle",t)},this.Type.SkewBase.prototype=new this.Type.matrix,this.Type.skewX=function(t){this.base=e.Type.SkewBase,this.base(t),this.m=[1,0,Math.tan(this.angle.toRadians()),1,0,0]},this.Type.skewX.prototype=new this.Type.SkewBase,this.Type.skewY=function(t){this.base=e.Type.SkewBase,this.base(t),this.m=[1,Math.tan(this.angle.toRadians()),0,1,0,0]},this.Type.skewY.prototype=new this.Type.SkewBase,this.transforms=[],this.apply=function(t){for(var e=0;e<this.transforms.length;e++)this.transforms[e].apply(t)},this.unapply=function(t){for(var e=this.transforms.length-1;0<=e;e--)this.transforms[e].unapply(t)},this.applyToPoint=function(t){for(var e=0;e<this.transforms.length;e++)this.transforms[e].applyToPoint(t)};for(var n=a.trim(a.compressSpaces(t)).replace(/\)([a-zA-Z])/g,") $1").replace(/\)(\s?,\s?)/g,") ").split(/\s(?=[a-z])/),r=0;r<n.length;r++)if("none"!==n[r]){var i=a.trim(n[r].split("(")[0]),o=n[r].split("(")[1].replace(")",""),s=this.Type[i];if(void 0!==s){var u=new s(o);u.type=i,this.transforms.push(u)}}},a.AspectRatio=function(t,e,n,r,i,o,s,u,c,l){var d=(e=(e=a.compressSpaces(e)).replace(/^defer\s/,"")).split(" ")[0]||"xMidYMid",p=e.split(" ")[1]||"meet",f=n/r,h=i/o,m=Math.min(f,h),g=Math.max(f,h);"meet"==p&&(r*=m,o*=m),"slice"==p&&(r*=g,o*=g),c=new a.Property("refX",c),l=new a.Property("refY",l),c.hasValue()&&l.hasValue()?t.translate(-m*c.toPixels("x"),-m*l.toPixels("y")):(d.match(/^xMid/)&&("meet"==p&&m==h||"slice"==p&&g==h)&&t.translate(n/2-r/2,0),d.match(/YMid$/)&&("meet"==p&&m==f||"slice"==p&&g==f)&&t.translate(0,i/2-o/2),d.match(/^xMax/)&&("meet"==p&&m==h||"slice"==p&&g==h)&&t.translate(n-r,0),d.match(/YMax$/)&&("meet"==p&&m==f||"slice"==p&&g==f)&&t.translate(0,i-o)),"none"==d?t.scale(f,h):"meet"==p?t.scale(m,m):"slice"==p&&t.scale(g,g),t.translate(null==s?0:-s,null==u?0:-u)},a.Element={},a.EmptyProperty=new a.Property("EMPTY",""),a.Element.ElementBase=function(t){this.attributes={},this.styles={},this.stylesSpecificity={},this.children=[],this.attribute=function(t,e){var n=this.attributes[t];return null!=n?n:(1==e&&(n=new a.Property(t,""),this.attributes[t]=n),n||a.EmptyProperty)},this.getHrefAttribute=function(){for(var t in this.attributes)if("href"==t||t.match(/:href$/))return this.attributes[t];return a.EmptyProperty},this.style=function(t,e,n){var r=this.styles[t];if(null!=r)return r;var i=this.attribute(t);if(null!=i&&i.hasValue())return this.styles[t]=i;if(1!=n){var o=this.parent;if(null!=o){var s=o.style(t);if(null!=s&&s.hasValue())return s}}return 1==e&&(r=new a.Property(t,""),this.styles[t]=r),r||a.EmptyProperty},this.render=function(t){if("none"!=this.style("display").value&&"hidden"!=this.style("visibility").value){if(t.save(),this.style("mask").hasValue()){var e=this.style("mask").getDefinition();null!=e&&e.apply(t,this)}else if(this.style("filter").hasValue()){var n=this.style("filter").getDefinition();null!=n&&n.apply(t,this)}else this.setContext(t),this.renderChildren(t),this.clearContext(t);t.restore()}},this.setContext=function(t){},this.clearContext=function(t){},this.renderChildren=function(t){for(var e=0;e<this.children.length;e++)this.children[e].render(t)},this.addChild=function(t,e){var n=t;e&&(n=a.CreateElement(t)),n.parent=this,"title"!=n.type&&this.children.push(n)},this.addStylesFromStyleDefinition=function(){for(var e in a.Styles)if("@"!=e[0]&&o(t,e)){var n=a.Styles[e],r=a.StylesSpecificity[e];if(null!=n)for(var i in n){var s=this.stylesSpecificity[i];void 0===s&&(s="000"),s<r&&(this.styles[i]=n[i],this.stylesSpecificity[i]=r)}}};var e,n=new RegExp("^[A-Z-]+$");if(null!=t&&1==t.nodeType){for(var r=0;r<t.attributes.length;r++){var i=t.attributes[r],s=(e=i.nodeName,n.test(e)?e.toLowerCase():e);this.attributes[s]=new a.Property(s,i.value)}if(this.addStylesFromStyleDefinition(),this.attribute("style").hasValue()){var u=this.attribute("style").value.split(";");for(r=0;r<u.length;r++)if(""!=a.trim(u[r])){var c=u[r].split(":"),l=a.trim(c[0]),d=a.trim(c[1]);this.styles[l]=new a.Property(l,d)}}for(this.attribute("id").hasValue()&&null==a.Definitions[this.attribute("id").value]&&(a.Definitions[this.attribute("id").value]=this),r=0;r<t.childNodes.length;r++){var p=t.childNodes[r];if(1==p.nodeType&&this.addChild(p,!0),this.captureTextNodes&&(3==p.nodeType||4==p.nodeType)){var f=p.value||p.text||p.textContent||"";""!=a.compressSpaces(f)&&this.addChild(new a.Element.tspan(p),!1)}}}},a.Element.RenderedElementBase=function(t){this.base=a.Element.ElementBase,this.base(t),this.calculateOpacity=function(){for(var t=1,e=this;null!=e;){var n=e.style("opacity",!1,!0);n.hasValue()&&(t*=n.numValue()),e=e.parent}return t},this.setContext=function(t,e){if(!e){var n;if(this.style("fill").isUrlDefinition())null!=(n=this.style("fill").getFillStyleDefinition(this,this.style("fill-opacity")))&&(t.fillStyle=n);else if(this.style("fill").hasValue()){var r;"currentColor"==(r=this.style("fill")).value&&(r.value=this.style("color").value),"inherit"!=r.value&&(t.fillStyle="none"==r.value?"rgba(0,0,0,0)":r.value)}if(this.style("fill-opacity").hasValue()&&(r=(r=new a.Property("fill",t.fillStyle)).addOpacity(this.style("fill-opacity")),t.fillStyle=r.value),this.style("stroke").isUrlDefinition())null!=(n=this.style("stroke").getFillStyleDefinition(this,this.style("stroke-opacity")))&&(t.strokeStyle=n);else if(this.style("stroke").hasValue()){var i;"currentColor"==(i=this.style("stroke")).value&&(i.value=this.style("color").value),"inherit"!=i.value&&(t.strokeStyle="none"==i.value?"rgba(0,0,0,0)":i.value)}if(this.style("stroke-opacity").hasValue()&&(i=(i=new a.Property("stroke",t.strokeStyle)).addOpacity(this.style("stroke-opacity")),t.strokeStyle=i.value),this.style("stroke-width").hasValue()){var o=this.style("stroke-width").toPixels();t.lineWidth=0==o?.001:o}if(this.style("stroke-linecap").hasValue()&&(t.lineCap=this.style("stroke-linecap").value),this.style("stroke-linejoin").hasValue()&&(t.lineJoin=this.style("stroke-linejoin").value),this.style("stroke-miterlimit").hasValue()&&(t.miterLimit=this.style("stroke-miterlimit").value),this.style("paint-order").hasValue()&&(t.paintOrder=this.style("paint-order").value),this.style("stroke-dasharray").hasValue()&&"none"!=this.style("stroke-dasharray").value){var s=a.ToNumberArray(this.style("stroke-dasharray").value);void 0!==t.setLineDash?t.setLineDash(s):void 0!==t.webkitLineDash?t.webkitLineDash=s:void 0===t.mozDash||1==s.length&&0==s[0]||(t.mozDash=s);var u=this.style("stroke-dashoffset").toPixels();void 0!==t.lineDashOffset?t.lineDashOffset=u:void 0!==t.webkitLineDashOffset?t.webkitLineDashOffset=u:void 0!==t.mozDashOffset&&(t.mozDashOffset=u)}}if(void 0!==t.font){t.font=a.Font.CreateFont(this.style("font-style").value,this.style("font-variant").value,this.style("font-weight").value,this.style("font-size").hasValue()?this.style("font-size").toPixels()+"px":"",this.style("font-family").value).toString();var c=this.style("font-size",!1,!1);c.isPixels()&&(a.emSize=c.toPixels())}if(this.style("transform",!1,!0).hasValue()&&new a.Transform(this.style("transform",!1,!0).value).apply(t),this.style("clip-path",!1,!0).hasValue()){var l=this.style("clip-path",!1,!0).getDefinition();null!=l&&l.apply(t)}t.globalAlpha=this.calculateOpacity()}},a.Element.RenderedElementBase.prototype=new a.Element.ElementBase,a.Element.PathElementBase=function(t){this.base=a.Element.RenderedElementBase,this.base(t),this.path=function(t){return null!=t&&t.beginPath(),new a.BoundingBox},this.renderChildren=function(t){this.path(t),a.Mouse.checkPath(this,t),""!=t.fillStyle&&("inherit"!=this.style("fill-rule").valueOrDefault("inherit")?t.fill(this.style("fill-rule").value):t.fill()),""!=t.strokeStyle&&t.stroke();var e=this.getMarkers();if(null!=e){if(this.style("marker-start").isUrlDefinition()&&(n=this.style("marker-start").getDefinition()).render(t,e[0][0],e[0][1]),this.style("marker-mid").isUrlDefinition())for(var n=this.style("marker-mid").getDefinition(),r=1;r<e.length-1;r++)n.render(t,e[r][0],e[r][1]);this.style("marker-end").isUrlDefinition()&&(n=this.style("marker-end").getDefinition()).render(t,e[e.length-1][0],e[e.length-1][1])}},this.getBoundingBox=function(){return this.path()},this.getMarkers=function(){return null}},a.Element.PathElementBase.prototype=new a.Element.RenderedElementBase,a.Element.svg=function(t){this.base=a.Element.RenderedElementBase,this.base(t),this.baseClearContext=this.clearContext,this.clearContext=function(t){this.baseClearContext(t),a.ViewPort.RemoveCurrent()},this.baseSetContext=this.setContext,this.setContext=function(t){if(t.strokeStyle="rgba(0,0,0,0)",t.lineCap="butt",t.lineJoin="miter",t.miterLimit=4,t.canvas.style&&void 0!==t.font&&void 0!==r.getComputedStyle){t.font=r.getComputedStyle(t.canvas).getPropertyValue("font");var e=new a.Property("fontSize",a.Font.Parse(t.font).fontSize);e.hasValue()&&(a.rootEmSize=a.emSize=e.toPixels("y"))}this.baseSetContext(t),this.attribute("x").hasValue()||(this.attribute("x",!0).value=0),this.attribute("y").hasValue()||(this.attribute("y",!0).value=0),t.translate(this.attribute("x").toPixels("x"),this.attribute("y").toPixels("y"));var n=a.ViewPort.width(),i=a.ViewPort.height();if(this.attribute("width").hasValue()||(this.attribute("width",!0).value="100%"),this.attribute("height").hasValue()||(this.attribute("height",!0).value="100%"),void 0===this.root){n=this.attribute("width").toPixels("x"),i=this.attribute("height").toPixels("y");var o=0,s=0;this.attribute("refX").hasValue()&&this.attribute("refY").hasValue()&&(o=-this.attribute("refX").toPixels("x"),s=-this.attribute("refY").toPixels("y")),"visible"!=this.attribute("overflow").valueOrDefault("hidden")&&(t.beginPath(),t.moveTo(o,s),t.lineTo(n,s),t.lineTo(n,i),t.lineTo(o,i),t.closePath(),t.clip())}if(a.ViewPort.SetCurrent(n,i),this.attribute("viewBox").hasValue()){var u=a.ToNumberArray(this.attribute("viewBox").value),c=u[0],l=u[1];n=u[2],i=u[3],a.AspectRatio(t,this.attribute("preserveAspectRatio").value,a.ViewPort.width(),n,a.ViewPort.height(),i,c,l,this.attribute("refX").value,this.attribute("refY").value),a.ViewPort.RemoveCurrent(),a.ViewPort.SetCurrent(u[2],u[3])}}},a.Element.svg.prototype=new a.Element.RenderedElementBase,a.Element.rect=function(t){this.base=a.Element.PathElementBase,this.base(t),this.path=function(t){var e=this.attribute("x").toPixels("x"),n=this.attribute("y").toPixels("y"),r=this.attribute("width").toPixels("x"),i=this.attribute("height").toPixels("y"),o=this.attribute("rx").toPixels("x"),s=this.attribute("ry").toPixels("y");if(this.attribute("rx").hasValue()&&!this.attribute("ry").hasValue()&&(s=o),this.attribute("ry").hasValue()&&!this.attribute("rx").hasValue()&&(o=s),o=Math.min(o,r/2),s=Math.min(s,i/2),null!=t){var u=(Math.sqrt(2)-1)/3*4;t.beginPath(),t.moveTo(e+o,n),t.lineTo(e+r-o,n),t.bezierCurveTo(e+r-o+u*o,n,e+r,n+s-u*s,e+r,n+s),t.lineTo(e+r,n+i-s),t.bezierCurveTo(e+r,n+i-s+u*s,e+r-o+u*o,n+i,e+r-o,n+i),t.lineTo(e+o,n+i),t.bezierCurveTo(e+o-u*o,n+i,e,n+i-s+u*s,e,n+i-s),t.lineTo(e,n+s),t.bezierCurveTo(e,n+s-u*s,e+o-u*o,n,e+o,n),t.closePath()}return new a.BoundingBox(e,n,e+r,n+i)}},a.Element.rect.prototype=new a.Element.PathElementBase,a.Element.circle=function(t){this.base=a.Element.PathElementBase,this.base(t),this.path=function(t){var e=this.attribute("cx").toPixels("x"),n=this.attribute("cy").toPixels("y"),r=this.attribute("r").toPixels();return null!=t&&(t.beginPath(),t.arc(e,n,r,0,2*Math.PI,!1),t.closePath()),new a.BoundingBox(e-r,n-r,e+r,n+r)}},a.Element.circle.prototype=new a.Element.PathElementBase,a.Element.ellipse=function(t){this.base=a.Element.PathElementBase,this.base(t),this.path=function(t){var e=(Math.sqrt(2)-1)/3*4,n=this.attribute("rx").toPixels("x"),r=this.attribute("ry").toPixels("y"),i=this.attribute("cx").toPixels("x"),o=this.attribute("cy").toPixels("y");return null!=t&&(t.beginPath(),t.moveTo(i+n,o),t.bezierCurveTo(i+n,o+e*r,i+e*n,o+r,i,o+r),t.bezierCurveTo(i-e*n,o+r,i-n,o+e*r,i-n,o),t.bezierCurveTo(i-n,o-e*r,i-e*n,o-r,i,o-r),t.bezierCurveTo(i+e*n,o-r,i+n,o-e*r,i+n,o),t.closePath()),new a.BoundingBox(i-n,o-r,i+n,o+r)}},a.Element.ellipse.prototype=new a.Element.PathElementBase,a.Element.line=function(t){this.base=a.Element.PathElementBase,this.base(t),this.getPoints=function(){return[new a.Point(this.attribute("x1").toPixels("x"),this.attribute("y1").toPixels("y")),new a.Point(this.attribute("x2").toPixels("x"),this.attribute("y2").toPixels("y"))]},this.path=function(t){var e=this.getPoints();return null!=t&&(t.beginPath(),t.moveTo(e[0].x,e[0].y),t.lineTo(e[1].x,e[1].y)),new a.BoundingBox(e[0].x,e[0].y,e[1].x,e[1].y)},this.getMarkers=function(){var t=this.getPoints(),e=t[0].angleTo(t[1]);return[[t[0],e],[t[1],e]]}},a.Element.line.prototype=new a.Element.PathElementBase,a.Element.polyline=function(t){this.base=a.Element.PathElementBase,this.base(t),this.points=a.CreatePath(this.attribute("points").value),this.path=function(t){var e=new a.BoundingBox(this.points[0].x,this.points[0].y);null!=t&&(t.beginPath(),t.moveTo(this.points[0].x,this.points[0].y));for(var n=1;n<this.points.length;n++)e.addPoint(this.points[n].x,this.points[n].y),null!=t&&t.lineTo(this.points[n].x,this.points[n].y);return e},this.getMarkers=function(){for(var t=[],e=0;e<this.points.length-1;e++)t.push([this.points[e],this.points[e].angleTo(this.points[e+1])]);return 0<t.length&&t.push([this.points[this.points.length-1],t[t.length-1][1]]),t}},a.Element.polyline.prototype=new a.Element.PathElementBase,a.Element.polygon=function(t){this.base=a.Element.polyline,this.base(t),this.basePath=this.path,this.path=function(t){var e=this.basePath(t);return null!=t&&(t.lineTo(this.points[0].x,this.points[0].y),t.closePath()),e}},a.Element.polygon.prototype=new a.Element.polyline,a.Element.path=function(t){this.base=a.Element.PathElementBase,this.base(t);var e=this.attribute("d").value;e=e.replace(/,/gm," ");for(var n=0;n<2;n++)e=e.replace(/([MmZzLlHhVvCcSsQqTtAa])([^\s])/gm,"$1 $2");for(e=(e=e.replace(/([^\s])([MmZzLlHhVvCcSsQqTtAa])/gm,"$1 $2")).replace(/([0-9])([+\-])/gm,"$1 $2"),n=0;n<2;n++)e=e.replace(/(\.[0-9]*)(\.)/gm,"$1 $2");e=e.replace(/([Aa](\s+[0-9]+){3})\s+([01])\s*([01])/gm,"$1 $3 $4 "),e=a.compressSpaces(e),e=a.trim(e),this.PathParser=new function(t){this.tokens=t.split(" "),this.reset=function(){this.i=-1,this.command="",this.previousCommand="",this.start=new a.Point(0,0),this.control=new a.Point(0,0),this.current=new a.Point(0,0),this.points=[],this.angles=[]},this.isEnd=function(){return this.i>=this.tokens.length-1},this.isCommandOrEnd=function(){return!!this.isEnd()||null!=this.tokens[this.i+1].match(/^[A-Za-z]$/)},this.isRelativeCommand=function(){switch(this.command){case"m":case"l":case"h":case"v":case"c":case"s":case"q":case"t":case"a":case"z":return!0}return!1},this.getToken=function(){return this.i++,this.tokens[this.i]},this.getScalar=function(){return parseFloat(this.getToken())},this.nextCommand=function(){this.previousCommand=this.command,this.command=this.getToken()},this.getPoint=function(){var t=new a.Point(this.getScalar(),this.getScalar());return this.makeAbsolute(t)},this.getAsControlPoint=function(){var t=this.getPoint();return this.control=t},this.getAsCurrentPoint=function(){var t=this.getPoint();return this.current=t},this.getReflectedControlPoint=function(){return"c"!=this.previousCommand.toLowerCase()&&"s"!=this.previousCommand.toLowerCase()&&"q"!=this.previousCommand.toLowerCase()&&"t"!=this.previousCommand.toLowerCase()?this.current:new a.Point(2*this.current.x-this.control.x,2*this.current.y-this.control.y)},this.makeAbsolute=function(t){return this.isRelativeCommand()&&(t.x+=this.current.x,t.y+=this.current.y),t},this.addMarker=function(t,e,n){null!=n&&0<this.angles.length&&null==this.angles[this.angles.length-1]&&(this.angles[this.angles.length-1]=this.points[this.points.length-1].angleTo(n)),this.addMarkerAngle(t,null==e?null:e.angleTo(t))},this.addMarkerAngle=function(t,e){this.points.push(t),this.angles.push(e)},this.getMarkerPoints=function(){return this.points},this.getMarkerAngles=function(){for(var t=0;t<this.angles.length;t++)if(null==this.angles[t])for(var e=t+1;e<this.angles.length;e++)if(null!=this.angles[e]){this.angles[t]=this.angles[e];break}return this.angles}}(e),this.path=function(t){var e=this.PathParser;e.reset();var n=new a.BoundingBox;for(null!=t&&t.beginPath();!e.isEnd();)switch(e.nextCommand(),e.command){case"M":case"m":var r=e.getAsCurrentPoint();for(e.addMarker(r),n.addPoint(r.x,r.y),null!=t&&t.moveTo(r.x,r.y),e.start=e.current;!e.isCommandOrEnd();)r=e.getAsCurrentPoint(),e.addMarker(r,e.start),n.addPoint(r.x,r.y),null!=t&&t.lineTo(r.x,r.y);break;case"L":case"l":for(;!e.isCommandOrEnd();){var i=e.current;r=e.getAsCurrentPoint(),e.addMarker(r,i),n.addPoint(r.x,r.y),null!=t&&t.lineTo(r.x,r.y)}break;case"H":case"h":for(;!e.isCommandOrEnd();){var o=new a.Point((e.isRelativeCommand()?e.current.x:0)+e.getScalar(),e.current.y);e.addMarker(o,e.current),e.current=o,n.addPoint(e.current.x,e.current.y),null!=t&&t.lineTo(e.current.x,e.current.y)}break;case"V":case"v":for(;!e.isCommandOrEnd();)o=new a.Point(e.current.x,(e.isRelativeCommand()?e.current.y:0)+e.getScalar()),e.addMarker(o,e.current),e.current=o,n.addPoint(e.current.x,e.current.y),null!=t&&t.lineTo(e.current.x,e.current.y);break;case"C":case"c":for(;!e.isCommandOrEnd();){var s=e.current,u=e.getPoint(),c=e.getAsControlPoint(),l=e.getAsCurrentPoint();e.addMarker(l,c,u),n.addBezierCurve(s.x,s.y,u.x,u.y,c.x,c.y,l.x,l.y),null!=t&&t.bezierCurveTo(u.x,u.y,c.x,c.y,l.x,l.y)}break;case"S":case"s":for(;!e.isCommandOrEnd();)s=e.current,u=e.getReflectedControlPoint(),c=e.getAsControlPoint(),l=e.getAsCurrentPoint(),e.addMarker(l,c,u),n.addBezierCurve(s.x,s.y,u.x,u.y,c.x,c.y,l.x,l.y),null!=t&&t.bezierCurveTo(u.x,u.y,c.x,c.y,l.x,l.y);break;case"Q":case"q":for(;!e.isCommandOrEnd();)s=e.current,c=e.getAsControlPoint(),l=e.getAsCurrentPoint(),e.addMarker(l,c,c),n.addQuadraticCurve(s.x,s.y,c.x,c.y,l.x,l.y),null!=t&&t.quadraticCurveTo(c.x,c.y,l.x,l.y);break;case"T":case"t":for(;!e.isCommandOrEnd();)s=e.current,c=e.getReflectedControlPoint(),e.control=c,l=e.getAsCurrentPoint(),e.addMarker(l,c,c),n.addQuadraticCurve(s.x,s.y,c.x,c.y,l.x,l.y),null!=t&&t.quadraticCurveTo(c.x,c.y,l.x,l.y);break;case"A":case"a":for(;!e.isCommandOrEnd();){s=e.current;var d=e.getScalar(),p=e.getScalar(),f=e.getScalar()*(Math.PI/180),h=e.getScalar(),m=e.getScalar(),g=(l=e.getAsCurrentPoint(),new a.Point(Math.cos(f)*(s.x-l.x)/2+Math.sin(f)*(s.y-l.y)/2,-Math.sin(f)*(s.x-l.x)/2+Math.cos(f)*(s.y-l.y)/2)),y=Math.pow(g.x,2)/Math.pow(d,2)+Math.pow(g.y,2)/Math.pow(p,2);1<y&&(d*=Math.sqrt(y),p*=Math.sqrt(y));var _=(h==m?-1:1)*Math.sqrt((Math.pow(d,2)*Math.pow(p,2)-Math.pow(d,2)*Math.pow(g.y,2)-Math.pow(p,2)*Math.pow(g.x,2))/(Math.pow(d,2)*Math.pow(g.y,2)+Math.pow(p,2)*Math.pow(g.x,2)));isNaN(_)&&(_=0);var v=new a.Point(_*d*g.y/p,_*-p*g.x/d),b=new a.Point((s.x+l.x)/2+Math.cos(f)*v.x-Math.sin(f)*v.y,(s.y+l.y)/2+Math.sin(f)*v.x+Math.cos(f)*v.y),x=function(t){return Math.sqrt(Math.pow(t[0],2)+Math.pow(t[1],2))},w=function(t,e){return(t[0]*e[0]+t[1]*e[1])/(x(t)*x(e))},j=function(t,e){return(t[0]*e[1]<t[1]*e[0]?-1:1)*Math.acos(w(t,e))},O=j([1,0],[(g.x-v.x)/d,(g.y-v.y)/p]),C=[(g.x-v.x)/d,(g.y-v.y)/p],E=[(-g.x-v.x)/d,(-g.y-v.y)/p],S=j(C,E);w(C,E)<=-1&&(S=Math.PI),1<=w(C,E)&&(S=0);var M=1-m?1:-1,T=O+M*(S/2),k=new a.Point(b.x+d*Math.cos(T),b.y+p*Math.sin(T));if(e.addMarkerAngle(k,T-M*Math.PI/2),e.addMarkerAngle(l,T-M*Math.PI),n.addPoint(l.x,l.y),null!=t){w=p<d?d:p;var A=p<d?1:d/p,D=p<d?p/d:1;t.translate(b.x,b.y),t.rotate(f),t.scale(A,D),t.arc(0,0,w,O,O+S,1-m),t.scale(1/A,1/D),t.rotate(-f),t.translate(-b.x,-b.y)}}break;case"Z":case"z":null!=t&&n.x1!==n.x2&&n.y1!==n.y2&&t.closePath(),e.current=e.start}return n},this.getMarkers=function(){for(var t=this.PathParser.getMarkerPoints(),e=this.PathParser.getMarkerAngles(),n=[],r=0;r<t.length;r++)n.push([t[r],e[r]]);return n}},a.Element.path.prototype=new a.Element.PathElementBase,a.Element.pattern=function(t){this.base=a.Element.ElementBase,this.base(t),this.createPattern=function(t,e){var n=this.attribute("width").toPixels("x",!0),r=this.attribute("height").toPixels("y",!0),o=new a.Element.svg;o.attributes.viewBox=new a.Property("viewBox",this.attribute("viewBox").value),o.attributes.width=new a.Property("width",n+"px"),o.attributes.height=new a.Property("height",r+"px"),o.attributes.transform=new a.Property("transform",this.attribute("patternTransform").value),o.children=this.children;var s=i();s.width=n,s.height=r;var u=s.getContext("2d");this.attribute("x").hasValue()&&this.attribute("y").hasValue()&&u.translate(this.attribute("x").toPixels("x",!0),this.attribute("y").toPixels("y",!0));for(var c=-1;c<=1;c++)for(var l=-1;l<=1;l++)u.save(),o.attributes.x=new a.Property("x",c*s.width),o.attributes.y=new a.Property("y",l*s.height),o.render(u),u.restore();return t.createPattern(s,"repeat")}},a.Element.pattern.prototype=new a.Element.ElementBase,a.Element.marker=function(t){this.base=a.Element.ElementBase,this.base(t),this.baseRender=this.render,this.render=function(t,e,n){if(e){t.translate(e.x,e.y),"auto"==this.attribute("orient").valueOrDefault("auto")&&t.rotate(n),"strokeWidth"==this.attribute("markerUnits").valueOrDefault("strokeWidth")&&t.scale(t.lineWidth,t.lineWidth),t.save();var r=new a.Element.svg;r.attributes.viewBox=new a.Property("viewBox",this.attribute("viewBox").value),r.attributes.refX=new a.Property("refX",this.attribute("refX").value),r.attributes.refY=new a.Property("refY",this.attribute("refY").value),r.attributes.width=new a.Property("width",this.attribute("markerWidth").value),r.attributes.height=new a.Property("height",this.attribute("markerHeight").value),r.attributes.fill=new a.Property("fill",this.attribute("fill").valueOrDefault("black")),r.attributes.stroke=new a.Property("stroke",this.attribute("stroke").valueOrDefault("none")),r.children=this.children,r.render(t),t.restore(),"strokeWidth"==this.attribute("markerUnits").valueOrDefault("strokeWidth")&&t.scale(1/t.lineWidth,1/t.lineWidth),"auto"==this.attribute("orient").valueOrDefault("auto")&&t.rotate(-n),t.translate(-e.x,-e.y)}}},a.Element.marker.prototype=new a.Element.ElementBase,a.Element.defs=function(t){this.base=a.Element.ElementBase,this.base(t),this.render=function(t){}},a.Element.defs.prototype=new a.Element.ElementBase,a.Element.GradientBase=function(t){this.base=a.Element.ElementBase,this.base(t),this.stops=[];for(var e=0;e<this.children.length;e++){var n=this.children[e];"stop"==n.type&&this.stops.push(n)}this.getGradient=function(){},this.gradientUnits=function(){return this.attribute("gradientUnits").valueOrDefault("objectBoundingBox")},this.attributesToInherit=["gradientUnits"],this.inheritStopContainer=function(t){for(var e=0;e<this.attributesToInherit.length;e++){var n=this.attributesToInherit[e];!this.attribute(n).hasValue()&&t.attribute(n).hasValue()&&(this.attribute(n,!0).value=t.attribute(n).value)}},this.createGradient=function(t,e,n){var r=this;this.getHrefAttribute().hasValue()&&(r=this.getHrefAttribute().getDefinition(),this.inheritStopContainer(r));var o=function(t){return n.hasValue()?new a.Property("color",t).addOpacity(n).value:t},s=this.getGradient(t,e);if(null==s)return o(r.stops[r.stops.length-1].color);for(var u=0;u<r.stops.length;u++)s.addColorStop(r.stops[u].offset,o(r.stops[u].color));if(this.attribute("gradientTransform").hasValue()){var c=a.ViewPort.viewPorts[0],l=new a.Element.rect;l.attributes.x=new a.Property("x",-a.MAX_VIRTUAL_PIXELS/3),l.attributes.y=new a.Property("y",-a.MAX_VIRTUAL_PIXELS/3),l.attributes.width=new a.Property("width",a.MAX_VIRTUAL_PIXELS),l.attributes.height=new a.Property("height",a.MAX_VIRTUAL_PIXELS);var d=new a.Element.g;d.attributes.transform=new a.Property("transform",this.attribute("gradientTransform").value),d.children=[l];var p=new a.Element.svg;p.attributes.x=new a.Property("x",0),p.attributes.y=new a.Property("y",0),p.attributes.width=new a.Property("width",c.width),p.attributes.height=new a.Property("height",c.height),p.children=[d];var f=i();f.width=c.width,f.height=c.height;var h=f.getContext("2d");return h.fillStyle=s,p.render(h),h.createPattern(f,"no-repeat")}return s}},a.Element.GradientBase.prototype=new a.Element.ElementBase,a.Element.linearGradient=function(t){this.base=a.Element.GradientBase,this.base(t),this.attributesToInherit.push("x1"),this.attributesToInherit.push("y1"),this.attributesToInherit.push("x2"),this.attributesToInherit.push("y2"),this.getGradient=function(t,e){var n="objectBoundingBox"==this.gradientUnits()?e.getBoundingBox(t):null;this.attribute("x1").hasValue()||this.attribute("y1").hasValue()||this.attribute("x2").hasValue()||this.attribute("y2").hasValue()||(this.attribute("x1",!0).value=0,this.attribute("y1",!0).value=0,this.attribute("x2",!0).value=1,this.attribute("y2",!0).value=0);var r="objectBoundingBox"==this.gradientUnits()?n.x()+n.width()*this.attribute("x1").numValue():this.attribute("x1").toPixels("x"),i="objectBoundingBox"==this.gradientUnits()?n.y()+n.height()*this.attribute("y1").numValue():this.attribute("y1").toPixels("y"),o="objectBoundingBox"==this.gradientUnits()?n.x()+n.width()*this.attribute("x2").numValue():this.attribute("x2").toPixels("x"),a="objectBoundingBox"==this.gradientUnits()?n.y()+n.height()*this.attribute("y2").numValue():this.attribute("y2").toPixels("y");return r==o&&i==a?null:t.createLinearGradient(r,i,o,a)}},a.Element.linearGradient.prototype=new a.Element.GradientBase,a.Element.radialGradient=function(t){this.base=a.Element.GradientBase,this.base(t),this.attributesToInherit.push("cx"),this.attributesToInherit.push("cy"),this.attributesToInherit.push("r"),this.attributesToInherit.push("fx"),this.attributesToInherit.push("fy"),this.getGradient=function(t,e){var n=e.getBoundingBox(t);this.attribute("cx").hasValue()||(this.attribute("cx",!0).value="50%"),this.attribute("cy").hasValue()||(this.attribute("cy",!0).value="50%"),this.attribute("r").hasValue()||(this.attribute("r",!0).value="50%");var r="objectBoundingBox"==this.gradientUnits()?n.x()+n.width()*this.attribute("cx").numValue():this.attribute("cx").toPixels("x"),i="objectBoundingBox"==this.gradientUnits()?n.y()+n.height()*this.attribute("cy").numValue():this.attribute("cy").toPixels("y"),o=r,a=i;this.attribute("fx").hasValue()&&(o="objectBoundingBox"==this.gradientUnits()?n.x()+n.width()*this.attribute("fx").numValue():this.attribute("fx").toPixels("x")),this.attribute("fy").hasValue()&&(a="objectBoundingBox"==this.gradientUnits()?n.y()+n.height()*this.attribute("fy").numValue():this.attribute("fy").toPixels("y"));var s="objectBoundingBox"==this.gradientUnits()?(n.width()+n.height())/2*this.attribute("r").numValue():this.attribute("r").toPixels();return t.createRadialGradient(o,a,0,r,i,s)}},a.Element.radialGradient.prototype=new a.Element.GradientBase,a.Element.stop=function(t){this.base=a.Element.ElementBase,this.base(t),this.offset=this.attribute("offset").numValue(),this.offset<0&&(this.offset=0),1<this.offset&&(this.offset=1);var e=this.style("stop-color",!0);""===e.value&&(e.value="#000"),this.style("stop-opacity").hasValue()&&(e=e.addOpacity(this.style("stop-opacity"))),this.color=e.value},a.Element.stop.prototype=new a.Element.ElementBase,a.Element.AnimateBase=function(t){this.base=a.Element.ElementBase,this.base(t),a.Animations.push(this),this.duration=0,this.begin=this.attribute("begin").toMilliseconds(),this.maxDuration=this.begin+this.attribute("dur").toMilliseconds(),this.getProperty=function(){var t=this.attribute("attributeType").value,e=this.attribute("attributeName").value;return"CSS"==t?this.parent.style(e,!0):this.parent.attribute(e,!0)},this.initialValue=null,this.initialUnits="",this.removed=!1,this.calcValue=function(){return""},this.update=function(t){if(null==this.initialValue&&(this.initialValue=this.getProperty().value,this.initialUnits=this.getProperty().getUnits()),this.duration>this.maxDuration){if("indefinite"==this.attribute("repeatCount").value||"indefinite"==this.attribute("repeatDur").value)this.duration=0;else if("freeze"!=this.attribute("fill").valueOrDefault("remove")||this.frozen){if("remove"==this.attribute("fill").valueOrDefault("remove")&&!this.removed)return this.removed=!0,this.getProperty().value=this.parent.animationFrozen?this.parent.animationFrozenValue:this.initialValue,!0}else this.frozen=!0,this.parent.animationFrozen=!0,this.parent.animationFrozenValue=this.getProperty().value;return!1}this.duration=this.duration+t;var e=!1;if(this.begin<this.duration){var n=this.calcValue();this.attribute("type").hasValue()&&(n=this.attribute("type").value+"("+n+")"),this.getProperty().value=n,e=!0}return e},this.from=this.attribute("from"),this.to=this.attribute("to"),this.values=this.attribute("values"),this.values.hasValue()&&(this.values.value=this.values.value.split(";")),this.progress=function(){var t={progress:(this.duration-this.begin)/(this.maxDuration-this.begin)};if(this.values.hasValue()){var e=t.progress*(this.values.value.length-1),n=Math.floor(e),r=Math.ceil(e);t.from=new a.Property("from",parseFloat(this.values.value[n])),t.to=new a.Property("to",parseFloat(this.values.value[r])),t.progress=(e-n)/(r-n)}else t.from=this.from,t.to=this.to;return t}},a.Element.AnimateBase.prototype=new a.Element.ElementBase,a.Element.animate=function(t){this.base=a.Element.AnimateBase,this.base(t),this.calcValue=function(){var t=this.progress();return t.from.numValue()+(t.to.numValue()-t.from.numValue())*t.progress+this.initialUnits}},a.Element.animate.prototype=new a.Element.AnimateBase,a.Element.animateColor=function(e){this.base=a.Element.AnimateBase,this.base(e),this.calcValue=function(){var e=this.progress(),n=new t(e.from.value),r=new t(e.to.value);if(n.ok&&r.ok){var i=n.r+(r.r-n.r)*e.progress,o=n.g+(r.g-n.g)*e.progress,a=n.b+(r.b-n.b)*e.progress;return"rgb("+parseInt(i,10)+","+parseInt(o,10)+","+parseInt(a,10)+")"}return this.attribute("from").value}},a.Element.animateColor.prototype=new a.Element.AnimateBase,a.Element.animateTransform=function(t){this.base=a.Element.AnimateBase,this.base(t),this.calcValue=function(){for(var t=this.progress(),e=a.ToNumberArray(t.from.value),n=a.ToNumberArray(t.to.value),r="",i=0;i<e.length;i++)r+=e[i]+(n[i]-e[i])*t.progress+" ";return r}},a.Element.animateTransform.prototype=new a.Element.animate,a.Element.font=function(t){this.base=a.Element.ElementBase,this.base(t),this.horizAdvX=this.attribute("horiz-adv-x").numValue(),this.isRTL=!1,this.isArabic=!1,this.fontFace=null,this.missingGlyph=null,this.glyphs=[];for(var e=0;e<this.children.length;e++){var n=this.children[e];"font-face"==n.type?(this.fontFace=n).style("font-family").hasValue()&&(a.Definitions[n.style("font-family").value]=this):"missing-glyph"==n.type?this.missingGlyph=n:"glyph"==n.type&&(""!=n.arabicForm?(this.isRTL=!0,this.isArabic=!0,void 0===this.glyphs[n.unicode]&&(this.glyphs[n.unicode]=[]),this.glyphs[n.unicode][n.arabicForm]=n):this.glyphs[n.unicode]=n)}},a.Element.font.prototype=new a.Element.ElementBase,a.Element.fontface=function(t){this.base=a.Element.ElementBase,this.base(t),this.ascent=this.attribute("ascent").value,this.descent=this.attribute("descent").value,this.unitsPerEm=this.attribute("units-per-em").numValue()},a.Element.fontface.prototype=new a.Element.ElementBase,a.Element.missingglyph=function(t){this.base=a.Element.path,this.base(t),this.horizAdvX=0},a.Element.missingglyph.prototype=new a.Element.path,a.Element.glyph=function(t){this.base=a.Element.path,this.base(t),this.horizAdvX=this.attribute("horiz-adv-x").numValue(),this.unicode=this.attribute("unicode").value,this.arabicForm=this.attribute("arabic-form").value},a.Element.glyph.prototype=new a.Element.path,a.Element.text=function(t){this.captureTextNodes=!0,this.base=a.Element.RenderedElementBase,this.base(t),this.baseSetContext=this.setContext,this.setContext=function(t){this.baseSetContext(t);var e=this.style("dominant-baseline").toTextBaseline();null==e&&(e=this.style("alignment-baseline").toTextBaseline()),null!=e&&(t.textBaseline=e)},this.initializeCoordinates=function(t){this.x=this.attribute("x").toPixels("x"),this.y=this.attribute("y").toPixels("y"),this.attribute("dx").hasValue()&&(this.x+=this.attribute("dx").toPixels("x")),this.attribute("dy").hasValue()&&(this.y+=this.attribute("dy").toPixels("y")),this.x+=this.getAnchorDelta(t,this,0)},this.getBoundingBox=function(t){this.initializeCoordinates(t);for(var e=null,n=0;n<this.children.length;n++){var r=this.getChildBoundingBox(t,this,this,n);null==e?e=r:e.addBoundingBox(r)}return e},this.renderChildren=function(t){this.initializeCoordinates(t);for(var e=0;e<this.children.length;e++)this.renderChild(t,this,this,e)},this.getAnchorDelta=function(t,e,n){var r=this.style("text-anchor").valueOrDefault("start");if("start"!=r){for(var i=0,o=n;o<e.children.length;o++){var a=e.children[o];if(n<o&&a.attribute("x").hasValue())break;i+=a.measureTextRecursive(t)}return-1*("end"==r?i:i/2)}return 0},this.adjustChildCoordinates=function(t,e,n,r){var i=n.children[r];return i.attribute("x").hasValue()?(i.x=i.attribute("x").toPixels("x")+e.getAnchorDelta(t,n,r),i.attribute("dx").hasValue()&&(i.x+=i.attribute("dx").toPixels("x"))):(i.attribute("dx").hasValue()&&(e.x+=i.attribute("dx").toPixels("x")),i.x=e.x),e.x=i.x+i.measureText(t),i.attribute("y").hasValue()?(i.y=i.attribute("y").toPixels("y"),i.attribute("dy").hasValue()&&(i.y+=i.attribute("dy").toPixels("y"))):(i.attribute("dy").hasValue()&&(e.y+=i.attribute("dy").toPixels("y")),i.y=e.y),e.y=i.y,i},this.getChildBoundingBox=function(t,e,n,r){var i=this.adjustChildCoordinates(t,e,n,r),o=i.getBoundingBox(t);for(r=0;r<i.children.length;r++){var a=e.getChildBoundingBox(t,e,i,r);o.addBoundingBox(a)}return o},this.renderChild=function(t,e,n,r){var i=this.adjustChildCoordinates(t,e,n,r);for(i.render(t),r=0;r<i.children.length;r++)e.renderChild(t,e,i,r)}},a.Element.text.prototype=new a.Element.RenderedElementBase,a.Element.TextElementBase=function(t){this.base=a.Element.RenderedElementBase,this.base(t),this.getGlyph=function(t,e,n){var r=e[n],i=null;if(t.isArabic){var o="isolated";(0==n||" "==e[n-1])&&n<e.length-2&&" "!=e[n+1]&&(o="terminal"),0<n&&" "!=e[n-1]&&n<e.length-2&&" "!=e[n+1]&&(o="medial"),0<n&&" "!=e[n-1]&&(n==e.length-1||" "==e[n+1])&&(o="initial"),void 0!==t.glyphs[r]&&null==(i=t.glyphs[r][o])&&"glyph"==t.glyphs[r].type&&(i=t.glyphs[r])}else i=t.glyphs[r];return null==i&&(i=t.missingGlyph),i},this.renderChildren=function(t){var e=this.parent.style("font-family").getDefinition();if(null==e)"stroke"==t.paintOrder?(""!=t.strokeStyle&&t.strokeText(a.compressSpaces(this.getText()),this.x,this.y),""!=t.fillStyle&&t.fillText(a.compressSpaces(this.getText()),this.x,this.y)):(""!=t.fillStyle&&t.fillText(a.compressSpaces(this.getText()),this.x,this.y),""!=t.strokeStyle&&t.strokeText(a.compressSpaces(this.getText()),this.x,this.y));else{var n=this.parent.style("font-size").numValueOrDefault(a.Font.Parse(a.ctx.font).fontSize),r=this.parent.style("font-style").valueOrDefault(a.Font.Parse(a.ctx.font).fontStyle),i=this.getText();e.isRTL&&(i=i.split("").reverse().join(""));for(var o=a.ToNumberArray(this.parent.attribute("dx").value),s=0;s<i.length;s++){var u=this.getGlyph(e,i,s),c=n/e.fontFace.unitsPerEm;t.translate(this.x,this.y),t.scale(c,-c);var l=t.lineWidth;t.lineWidth=t.lineWidth*e.fontFace.unitsPerEm/n,"italic"==r&&t.transform(1,0,.4,1,0,0),u.render(t),"italic"==r&&t.transform(1,0,-.4,1,0,0),t.lineWidth=l,t.scale(1/c,-1/c),t.translate(-this.x,-this.y),this.x+=n*(u.horizAdvX||e.horizAdvX)/e.fontFace.unitsPerEm,void 0===o[s]||isNaN(o[s])||(this.x+=o[s])}}},this.getText=function(){},this.measureTextRecursive=function(t){for(var e=this.measureText(t),n=0;n<this.children.length;n++)e+=this.children[n].measureTextRecursive(t);return e},this.measureText=function(t){var e=this.parent.style("font-family").getDefinition();if(null!=e){var n=this.parent.style("font-size").numValueOrDefault(a.Font.Parse(a.ctx.font).fontSize),r=0,i=this.getText();e.isRTL&&(i=i.split("").reverse().join(""));for(var o=a.ToNumberArray(this.parent.attribute("dx").value),s=0;s<i.length;s++)r+=(this.getGlyph(e,i,s).horizAdvX||e.horizAdvX)*n/e.fontFace.unitsPerEm,void 0===o[s]||isNaN(o[s])||(r+=o[s]);return r}var u=a.compressSpaces(this.getText());if(!t.measureText)return 10*u.length;t.save(),this.setContext(t,!0);var c=t.measureText(u).width;return t.restore(),c},this.getBoundingBox=function(t){var e=this.parent.style("font-size").numValueOrDefault(a.Font.Parse(a.ctx.font).fontSize);return new a.BoundingBox(this.x,this.y-e,this.x+this.measureText(t),this.y)}},a.Element.TextElementBase.prototype=new a.Element.RenderedElementBase,a.Element.tspan=function(t){this.captureTextNodes=!0,this.base=a.Element.TextElementBase,this.base(t),this.text=a.compressSpaces(t.value||t.text||t.textContent||""),this.getText=function(){return 0<this.children.length?"":this.text}},a.Element.tspan.prototype=new a.Element.TextElementBase,a.Element.tref=function(t){this.base=a.Element.TextElementBase,this.base(t),this.getText=function(){var t=this.getHrefAttribute().getDefinition();if(null!=t)return t.children[0].getText()}},a.Element.tref.prototype=new a.Element.TextElementBase,a.Element.a=function(t){this.base=a.Element.TextElementBase,this.base(t),this.hasText=0<t.childNodes.length;for(var e=0;e<t.childNodes.length;e++)3!=t.childNodes[e].nodeType&&(this.hasText=!1);this.text=this.hasText?t.childNodes[0].value||t.childNodes[0].data:"",this.getText=function(){return this.text},this.baseRenderChildren=this.renderChildren,this.renderChildren=function(t){if(this.hasText){this.baseRenderChildren(t);var e=new a.Property("fontSize",a.Font.Parse(a.ctx.font).fontSize);a.Mouse.checkBoundingBox(this,new a.BoundingBox(this.x,this.y-e.toPixels("y"),this.x+this.measureText(t),this.y))}else if(0<this.children.length){var n=new a.Element.g;n.children=this.children,n.parent=this,n.render(t)}},this.onclick=function(){r.open(this.getHrefAttribute().value)},this.onmousemove=function(){a.ctx.canvas.style.cursor="pointer"}},a.Element.a.prototype=new a.Element.TextElementBase,a.Element.image=function(t){this.base=a.Element.RenderedElementBase,this.base(t);var e=this.getHrefAttribute().value;if(""!=e){var n=e.match(/\.svg$/);if(a.Images.push(this),this.loaded=!1,n)this.img=a.ajax(e),this.loaded=!0;else{this.img=document.createElement("img"),1==a.opts.useCORS&&(this.img.crossOrigin="Anonymous");var r=this;this.img.onload=function(){r.loaded=!0},this.img.onerror=function(){a.log('ERROR: image "'+e+'" not found'),r.loaded=!0},this.img.src=e}this.renderChildren=function(t){var e=this.attribute("x").toPixels("x"),i=this.attribute("y").toPixels("y"),o=this.attribute("width").toPixels("x"),s=this.attribute("height").toPixels("y");0!=o&&0!=s&&(t.save(),n?t.drawSvg(this.img,e,i,o,s):(t.translate(e,i),a.AspectRatio(t,this.attribute("preserveAspectRatio").value,o,this.img.width,s,this.img.height,0,0),r.loaded&&(void 0===this.img.complete||this.img.complete)&&t.drawImage(this.img,0,0)),t.restore())},this.getBoundingBox=function(){var t=this.attribute("x").toPixels("x"),e=this.attribute("y").toPixels("y"),n=this.attribute("width").toPixels("x"),r=this.attribute("height").toPixels("y");return new a.BoundingBox(t,e,t+n,e+r)}}},a.Element.image.prototype=new a.Element.RenderedElementBase,a.Element.g=function(t){this.base=a.Element.RenderedElementBase,this.base(t),this.getBoundingBox=function(t){for(var e=new a.BoundingBox,n=0;n<this.children.length;n++)e.addBoundingBox(this.children[n].getBoundingBox(t));return e}},a.Element.g.prototype=new a.Element.RenderedElementBase,a.Element.symbol=function(t){this.base=a.Element.RenderedElementBase,this.base(t),this.render=function(t){}},a.Element.symbol.prototype=new a.Element.RenderedElementBase,a.Element.style=function(t){this.base=a.Element.ElementBase,this.base(t);for(var e="",n=0;n<t.childNodes.length;n++)e+=t.childNodes[n].data;e=e.replace(/(\/\*([^*]|[\r\n]|(\*+([^*\/]|[\r\n])))*\*+\/)|(^[\s]*\/\/.*)/gm,"");var r=(e=a.compressSpaces(e)).split("}");for(n=0;n<r.length;n++)if(""!=a.trim(r[n]))for(var i=r[n].split("{"),o=i[0].split(","),s=i[1].split(";"),u=0;u<o.length;u++){var c=a.trim(o[u]);if(""!=c){for(var l=a.Styles[c]||{},d=0;d<s.length;d++){var p=s[d].indexOf(":"),f=s[d].substr(0,p),m=s[d].substr(p+1,s[d].length-p);null!=f&&null!=m&&(l[a.trim(f)]=new a.Property(a.trim(f),a.trim(m)))}if(a.Styles[c]=l,a.StylesSpecificity[c]=h(c),"@font-face"==c)for(var g=l["font-family"].value.replace(/"/g,""),y=l.src.value.split(","),_=0;_<y.length;_++)if(0<y[_].indexOf('format("svg")'))for(var v=y[_].indexOf("url"),b=y[_].indexOf(")",v),x=y[_].substr(v+5,b-v-6),w=a.parseXml(a.ajax(x)).getElementsByTagName("font"),j=0;j<w.length;j++){var O=a.CreateElement(w[j]);a.Definitions[g]=O}}}},a.Element.style.prototype=new a.Element.ElementBase,a.Element.use=function(t){this.base=a.Element.RenderedElementBase,this.base(t),this.baseSetContext=this.setContext,this.setContext=function(t){this.baseSetContext(t),this.attribute("x").hasValue()&&t.translate(this.attribute("x").toPixels("x"),0),this.attribute("y").hasValue()&&t.translate(0,this.attribute("y").toPixels("y"))};var e=this.getHrefAttribute().getDefinition();this.path=function(t){null!=e&&e.path(t)},this.elementTransform=function(){if(null!=e&&e.style("transform",!1,!0).hasValue())return new a.Transform(e.style("transform",!1,!0).value)},this.getBoundingBox=function(t){if(null!=e)return e.getBoundingBox(t)},this.renderChildren=function(t){if(null!=e){var n=e;"symbol"==e.type&&((n=new a.Element.svg).type="svg",n.attributes.viewBox=new a.Property("viewBox",e.attribute("viewBox").value),n.attributes.preserveAspectRatio=new a.Property("preserveAspectRatio",e.attribute("preserveAspectRatio").value),n.attributes.overflow=new a.Property("overflow",e.attribute("overflow").value),n.children=e.children),"svg"==n.type&&(this.attribute("width").hasValue()&&(n.attributes.width=new a.Property("width",this.attribute("width").value)),this.attribute("height").hasValue()&&(n.attributes.height=new a.Property("height",this.attribute("height").value)));var r=n.parent;n.parent=null,n.render(t),n.parent=r}}},a.Element.use.prototype=new a.Element.RenderedElementBase,a.Element.mask=function(t){this.base=a.Element.ElementBase,this.base(t),this.apply=function(t,e){var n=this.attribute("x").toPixels("x"),r=this.attribute("y").toPixels("y"),o=this.attribute("width").toPixels("x"),s=this.attribute("height").toPixels("y");if(0==o&&0==s){for(var u=new a.BoundingBox,c=0;c<this.children.length;c++)u.addBoundingBox(this.children[c].getBoundingBox(t));n=Math.floor(u.x1),r=Math.floor(u.y1),o=Math.floor(u.width()),s=Math.floor(u.height())}var l=e.attribute("mask").value;e.attribute("mask").value="";var d=i();d.width=n+o,d.height=r+s;var p=d.getContext("2d");this.renderChildren(p);var f=i();f.width=n+o,f.height=r+s;var h=f.getContext("2d");e.render(h),h.globalCompositeOperation="destination-in",h.fillStyle=p.createPattern(d,"no-repeat"),h.fillRect(0,0,n+o,r+s),t.fillStyle=h.createPattern(f,"no-repeat"),t.fillRect(0,0,n+o,r+s),e.attribute("mask").value=l},this.render=function(t){}},a.Element.mask.prototype=new a.Element.ElementBase,a.Element.clipPath=function(t){this.base=a.Element.ElementBase,this.base(t),this.apply=function(t){var e="undefined"!=typeof CanvasRenderingContext2D,n=t.beginPath,r=t.closePath;e&&(CanvasRenderingContext2D.prototype.beginPath=function(){},CanvasRenderingContext2D.prototype.closePath=function(){}),n.call(t);for(var i=0;i<this.children.length;i++){var o=this.children[i];if(void 0!==o.path){var s=void 0!==o.elementTransform&&o.elementTransform();!s&&o.style("transform",!1,!0).hasValue()&&(s=new a.Transform(o.style("transform",!1,!0).value)),s&&s.apply(t),o.path(t),e&&(CanvasRenderingContext2D.prototype.closePath=r),s&&s.unapply(t)}}r.call(t),t.clip(),e&&(CanvasRenderingContext2D.prototype.beginPath=n,CanvasRenderingContext2D.prototype.closePath=r)},this.render=function(t){}},a.Element.clipPath.prototype=new a.Element.ElementBase,a.Element.filter=function(t){this.base=a.Element.ElementBase,this.base(t),this.apply=function(t,e){var n=e.getBoundingBox(t),r=Math.floor(n.x1),o=Math.floor(n.y1),a=Math.floor(n.width()),s=Math.floor(n.height()),u=e.style("filter").value;e.style("filter").value="";for(var c=0,l=0,d=0;d<this.children.length;d++){var p=this.children[d].extraFilterDistance||0;c=Math.max(c,p),l=Math.max(l,p)}var f=i();f.width=a+2*c,f.height=s+2*l;var h=f.getContext("2d");for(h.translate(-r+c,-o+l),e.render(h),d=0;d<this.children.length;d++)"function"==typeof this.children[d].apply&&this.children[d].apply(h,0,0,a+2*c,s+2*l);t.drawImage(f,0,0,a+2*c,s+2*l,r-c,o-l,a+2*c,s+2*l),e.style("filter",!0).value=u},this.render=function(t){}},a.Element.filter.prototype=new a.Element.ElementBase,a.Element.feMorphology=function(t){this.base=a.Element.ElementBase,this.base(t),this.apply=function(t,e,n,r,i){}},a.Element.feMorphology.prototype=new a.Element.ElementBase,a.Element.feComposite=function(t){this.base=a.Element.ElementBase,this.base(t),this.apply=function(t,e,n,r,i){}},a.Element.feComposite.prototype=new a.Element.ElementBase,a.Element.feColorMatrix=function(t){this.base=a.Element.ElementBase,this.base(t);var e=a.ToNumberArray(this.attribute("values").value);switch(this.attribute("type").valueOrDefault("matrix")){case"saturate":var n=e[0];e=[.213+.787*n,.715-.715*n,.072-.072*n,0,0,.213-.213*n,.715+.285*n,.072-.072*n,0,0,.213-.213*n,.715-.715*n,.072+.928*n,0,0,0,0,0,1,0,0,0,0,0,1];break;case"hueRotate":var r=e[0]*Math.PI/180,i=function(t,e,n){return t+Math.cos(r)*e+Math.sin(r)*n};e=[i(.213,.787,-.213),i(.715,-.715,-.715),i(.072,-.072,.928),0,0,i(.213,-.213,.143),i(.715,.285,.14),i(.072,-.072,-.283),0,0,i(.213,-.213,-.787),i(.715,-.715,.715),i(.072,.928,.072),0,0,0,0,0,1,0,0,0,0,0,1];break;case"luminanceToAlpha":e=[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,.2125,.7154,.0721,0,0,0,0,0,0,1]}function o(t,e,n,r,i,o){return t[n*r*4+4*e+o]}function s(t,e,n,r,i,o,a){t[n*r*4+4*e+o]=a}function u(t,n){var r=e[t];return r*(r<0?n-255:n)}this.apply=function(t,e,n,r,i){var a=t.getImageData(0,0,r,i);for(n=0;n<i;n++)for(e=0;e<r;e++){var c=o(a.data,e,n,r,0,0),l=o(a.data,e,n,r,0,1),d=o(a.data,e,n,r,0,2),p=o(a.data,e,n,r,0,3);s(a.data,e,n,r,0,0,u(0,c)+u(1,l)+u(2,d)+u(3,p)+u(4,1)),s(a.data,e,n,r,0,1,u(5,c)+u(6,l)+u(7,d)+u(8,p)+u(9,1)),s(a.data,e,n,r,0,2,u(10,c)+u(11,l)+u(12,d)+u(13,p)+u(14,1)),s(a.data,e,n,r,0,3,u(15,c)+u(16,l)+u(17,d)+u(18,p)+u(19,1))}t.clearRect(0,0,r,i),t.putImageData(a,0,0)}},a.Element.feColorMatrix.prototype=new a.Element.ElementBase,a.Element.feGaussianBlur=function(t){this.base=a.Element.ElementBase,this.base(t),this.blurRadius=Math.floor(this.attribute("stdDeviation").numValue()),this.extraFilterDistance=this.blurRadius,this.apply=function(t,n,r,i,o){e&&void 0!==e.canvasRGBA?(t.canvas.id=a.UniqueId(),t.canvas.style.display="none",document.body.appendChild(t.canvas),e.canvasRGBA(t.canvas,n,r,i,o,this.blurRadius),document.body.removeChild(t.canvas)):a.log("ERROR: StackBlur.js must be included for blur to work")}},a.Element.feGaussianBlur.prototype=new a.Element.ElementBase,a.Element.title=function(t){},a.Element.title.prototype=new a.Element.ElementBase,a.Element.desc=function(t){},a.Element.desc.prototype=new a.Element.ElementBase,a.Element.MISSING=function(t){a.log("ERROR: Element '"+t.nodeName+"' not yet implemented.")},a.Element.MISSING.prototype=new a.Element.ElementBase,a.CreateElement=function(t){var e=t.nodeName.replace(/^[^:]+:/,"");e=e.replace(/\-/g,"");var n=null;return(n=void 0!==a.Element[e]?new a.Element[e](t):new a.Element.MISSING(t)).type=t.nodeName,n},a.load=function(t,e){a.loadXml(t,a.ajax(e))},a.loadXml=function(t,e){a.loadXmlDoc(t,a.parseXml(e))},a.loadXmlDoc=function(t,e){a.init(t);var n=function(e){for(var n=t.canvas;n;)e.x-=n.offsetLeft,e.y-=n.offsetTop,n=n.offsetParent;return r.scrollX&&(e.x+=r.scrollX),r.scrollY&&(e.y+=r.scrollY),e};1!=a.opts.ignoreMouse&&(t.canvas.onclick=function(t){var e=n(new a.Point(null!=t?t.clientX:event.clientX,null!=t?t.clientY:event.clientY));a.Mouse.onclick(e.x,e.y)},t.canvas.onmousemove=function(t){var e=n(new a.Point(null!=t?t.clientX:event.clientX,null!=t?t.clientY:event.clientY));a.Mouse.onmousemove(e.x,e.y)});var i=a.CreateElement(e.documentElement);i.root=!0,i.addStylesFromStyleDefinition();var o=!0,s=function(){a.ViewPort.Clear(),t.canvas.parentNode?a.ViewPort.SetCurrent(t.canvas.parentNode.clientWidth,t.canvas.parentNode.clientHeight):a.ViewPort.SetCurrent(800,600),1!=a.opts.ignoreDimensions&&(i.style("width").hasValue()&&(t.canvas.width=i.style("width").toPixels("x"),t.canvas.style&&(t.canvas.style.width=t.canvas.width+"px")),i.style("height").hasValue()&&(t.canvas.height=i.style("height").toPixels("y"),t.canvas.style&&(t.canvas.style.height=t.canvas.height+"px")));var n=t.canvas.clientWidth||t.canvas.width,r=t.canvas.clientHeight||t.canvas.height;if(1==a.opts.ignoreDimensions&&i.style("width").hasValue()&&i.style("height").hasValue()&&(n=i.style("width").toPixels("x"),r=i.style("height").toPixels("y")),a.ViewPort.SetCurrent(n,r),null!=a.opts.offsetX&&(i.attribute("x",!0).value=a.opts.offsetX),null!=a.opts.offsetY&&(i.attribute("y",!0).value=a.opts.offsetY),null!=a.opts.scaleWidth||null!=a.opts.scaleHeight){var s=null,u=null,c=a.ToNumberArray(i.attribute("viewBox").value);null!=a.opts.scaleWidth&&(i.attribute("width").hasValue()?s=i.attribute("width").toPixels("x")/a.opts.scaleWidth:isNaN(c[2])||(s=c[2]/a.opts.scaleWidth)),null!=a.opts.scaleHeight&&(i.attribute("height").hasValue()?u=i.attribute("height").toPixels("y")/a.opts.scaleHeight:isNaN(c[3])||(u=c[3]/a.opts.scaleHeight)),null==s&&(s=u),null==u&&(u=s),i.attribute("width",!0).value=a.opts.scaleWidth,i.attribute("height",!0).value=a.opts.scaleHeight,i.style("transform",!0,!0).value+=" scale("+1/s+","+1/u+")"}1!=a.opts.ignoreClear&&t.clearRect(0,0,n,r),i.render(t),o&&(o=!1,"function"==typeof a.opts.renderCallback&&a.opts.renderCallback(e))},u=!0;a.ImagesLoaded()&&(u=!1,s()),a.intervalID=setInterval(function(){var t=!1;if(u&&a.ImagesLoaded()&&(t=!(u=!1)),1!=a.opts.ignoreMouse&&(t|=a.Mouse.hasEvents()),1!=a.opts.ignoreAnimation)for(var e=0;e<a.Animations.length;e++)t|=a.Animations[e].update(1e3/a.FRAMERATE);"function"==typeof a.opts.forceRedraw&&1==a.opts.forceRedraw()&&(t=!0),t&&(s(),a.Mouse.runEvents())},1e3/a.FRAMERATE)},a.stop=function(){a.intervalID&&clearInterval(a.intervalID)},a.Mouse=new function(){this.events=[],this.hasEvents=function(){return 0!=this.events.length},this.onclick=function(t,e){this.events.push({type:"onclick",x:t,y:e,run:function(t){t.onclick&&t.onclick()}})},this.onmousemove=function(t,e){this.events.push({type:"onmousemove",x:t,y:e,run:function(t){t.onmousemove&&t.onmousemove()}})},this.eventElements=[],this.checkPath=function(t,e){for(var n=0;n<this.events.length;n++){var r=this.events[n];e.isPointInPath&&e.isPointInPath(r.x,r.y)&&(this.eventElements[n]=t)}},this.checkBoundingBox=function(t,e){for(var n=0;n<this.events.length;n++){var r=this.events[n];e.isPointInBox(r.x,r.y)&&(this.eventElements[n]=t)}},this.runEvents=function(){a.ctx.canvas.style.cursor="";for(var t=0;t<this.events.length;t++)for(var e=this.events[t],n=this.eventElements[t];n;)e.run(n),n=n.parent;this.events=[],this.eventElements=[]}},a}(u||{});"string"==typeof n&&(n=document.getElementById(n)),null!=n.svg&&n.svg.stop(),n.childNodes&&1==n.childNodes.length&&"OBJECT"==n.childNodes[0].nodeName||(n.svg=c);var l=n.getContext("2d");void 0!==s.documentElement?c.loadXmlDoc(l,s):"<"==s.substr(0,1)?c.loadXml(l,s):c.load(l,s)}else for(var d=document.querySelectorAll("svg"),p=0;p<d.length;p++){var f=d[p],m=document.createElement("canvas");m.width=f.clientWidth,m.height=f.clientHeight,f.parentNode.insertBefore(m,f),f.parentNode.removeChild(f);var g=document.createElement("div");g.appendChild(f),a(m,g.innerHTML)}};"undefined"==typeof Element||(void 0!==Element.prototype.matches?o=function(t,e){return t.matches(e)}:void 0!==Element.prototype.webkitMatchesSelector?o=function(t,e){return t.webkitMatchesSelector(e)}:void 0!==Element.prototype.mozMatchesSelector?o=function(t,e){return t.mozMatchesSelector(e)}:void 0!==Element.prototype.msMatchesSelector?o=function(t,e){return t.msMatchesSelector(e)}:void 0!==Element.prototype.oMatchesSelector?o=function(t,e){return t.oMatchesSelector(e)}:("function"!=typeof jQuery&&"function"!=typeof Zepto||(o=function(t,e){return $(t).is(e)}),void 0===o&&"undefined"!=typeof Sizzle&&(o=Sizzle.matchesSelector)));var s=/(\[[^\]]+\])/g,u=/(#[^\s\+>~\.\[:]+)/g,c=/(\.[^\s\+>~\.\[:]+)/g,l=/(::[^\s\+>~\.\[:]+|:first-line|:first-letter|:before|:after)/gi,d=/(:[\w-]+\([^\)]*\))/gi,p=/(:[^\s\+>~\.\[:]+)/g,f=/([^\s\+>~\.\[:]+)/g;function h(t){var e=[0,0,0],n=function(n,r){var i=t.match(n);null!=i&&(e[r]+=i.length,t=t.replace(n," "))};return t=(t=t.replace(/:not\(([^\)]*)\)/g,"     $1 ")).replace(/{[\s\S]*/gm," "),n(s,1),n(u,0),n(c,1),n(l,2),n(d,1),n(p,1),t=(t=t.replace(/[\*\s\+>~]/g," ")).replace(/[#\.]/g," "),n(f,2),e.join("")}"undefined"!=typeof CanvasRenderingContext2D&&(CanvasRenderingContext2D.prototype.drawSvg=function(t,e,n,r,i,o){var s={ignoreMouse:!0,ignoreAnimation:!0,ignoreDimensions:!0,ignoreClear:!0,offsetX:e,offsetY:n,scaleWidth:r,scaleHeight:i};for(var u in o)o.hasOwnProperty(u)&&(s[u]=o[u]);a(this.canvas,t,s)}),n.exports=a}(n={exports:{}}),n.exports}(n(20),n(21))},function(t,e,n){(function(n){var r,i;void 0===(i="function"==typeof(r=function(){"use strict";function e(t,e,n){var r=new XMLHttpRequest;r.open("GET",t),r.responseType="blob",r.onload=function(){a(r.response,e,n)},r.onerror=function(){console.error("could not download file")},r.send()}function r(t){var e=new XMLHttpRequest;return e.open("HEAD",t,!1),e.send(),200<=e.status&&299>=e.status}function i(t){try{t.dispatchEvent(new MouseEvent("click"))}catch(n){var e=document.createEvent("MouseEvents");e.initMouseEvent("click",!0,!0,window,0,0,0,80,20,!1,!1,!1,!1,0,null),t.dispatchEvent(e)}}var o="object"==typeof window&&window.window===window?window:"object"==typeof self&&self.self===self?self:"object"==typeof n&&n.global===n?n:void 0,a=o.saveAs||("object"!=typeof window||window!==o?function(){}:"download"in HTMLAnchorElement.prototype?function(t,n,a){var s=o.URL||o.webkitURL,u=document.createElement("a");n=n||t.name||"download",u.download=n,u.rel="noopener","string"==typeof t?(u.href=t,u.origin===location.origin?i(u):r(u.href)?e(t,n,a):i(u,u.target="_blank")):(u.href=s.createObjectURL(t),setTimeout(function(){s.revokeObjectURL(u.href)},4e4),setTimeout(function(){i(u)},0))}:"msSaveOrOpenBlob"in navigator?function(t,n,o){if(n=n||t.name||"download","string"!=typeof t)navigator.msSaveOrOpenBlob((s=t,u=o,void 0===u?u={autoBom:!1}:"object"!=typeof u&&(console.warn("Deprecated: Expected third argument to be a object"),u={autoBom:!u}),u.autoBom&&/^\s*(?:text\/\S*|application\/xml|\S*\/\S*\+xml)\s*;.*charset\s*=\s*utf-8/i.test(s.type)?new Blob(["\ufeff",s],{type:s.type}):s),n);else if(r(t))e(t,n,o);else{var a=document.createElement("a");a.href=t,a.target="_blank",setTimeout(function(){i(a)})}var s,u}:function(t,n,r,i){if((i=i||open("","_blank"))&&(i.document.title=i.document.body.innerText="downloading..."),"string"==typeof t)return e(t,n,r);var a="application/octet-stream"===t.type,s=/constructor/i.test(o.HTMLElement)||o.safari,u=/CriOS\/[\d]+/.test(navigator.userAgent);if((u||a&&s)&&"object"==typeof FileReader){var c=new FileReader;c.onloadend=function(){var t=c.result;t=u?t:t.replace(/^data:[^;]*;/,"data:attachment/file;"),i?i.location.href=t:location=t,i=null},c.readAsDataURL(t)}else{var l=o.URL||o.webkitURL,d=l.createObjectURL(t);i?i.location=d:location.href=d,i=null,setTimeout(function(){l.revokeObjectURL(d)},4e4)}});o.saveAs=a.saveAs=a,t.exports=a})?r.apply(e,[]):r)||(t.exports=i)}).call(this,n(22))},function(t,e,n){(t.exports=n(23)(!1)).push([t.i,"body > * {\n    visibility: hidden;\n}\nbody {\n    overflow: hidden;\n}\nbody * {\n    visibility: hidden !important;\n}\n\n.tau-chart__print-block {\n    position: absolute;\n    top: 0;\n    left: 0;\n    visibility: visible !important;\n    display: block !important;\n    width: 100%;\n    /*height: 100%;*/\n}\n",""])},,,,,,,,,,,function(t,e,n){"use strict";n.r(e);var r,i=n(0),o=n.n(i),a=n(6),s=n.n(a),u=n(7),c=n(8),l=n.n(c),d=n(1),p="Taucharts Export Plug-in:",f=o.a.api.utils,h=o.a.api.pluginsSDK,m=h.tokens(),g=function(t,e){return t.replace(new RegExp("^"+e+"+|"+e+"+$","g"),"")},y=function(t){return(t=Math.round(t))%2?t+1:t},_=function(t){return null===t||""===t||void 0===t},v={BACKSPACE:8,COMMA:188,DELETE:46,DOWN:40,END:35,ENTER:13,ESCAPE:27,HOME:36,LEFT:37,PAGE_DOWN:34,PAGE_UP:33,PERIOD:190,RIGHT:39,SPACE:32,TAB:9,UP:38},b=function(t,e){e=e||"all";var n=document.createElement("style");return n.setAttribute("media",e),n.innerHTML=t,n},x=b(l.a,"print"),w=function(){x&&x.parentNode&&x.parentNode.removeChild(x),r&&r.parentNode&&r.parentNode.removeChild(r)};/PhantomJS/.test(navigator.userAgent)||("onafterprint"in window?window.addEventListener("afterprint",w):window.matchMedia("screen").addListener(function(t){t.matches&&w()}));var j=function(t,e,n){var r="data:"+e+";charset=UTF-8,%ef%bb%bf"+encodeURIComponent(n),i=document.createElement("a");i.setAttribute("href",r),i.setAttribute("download",t),i.setAttribute("target","_new"),document.body.appendChild(i),i.click(),document.body.removeChild(i),i=null},O=function(t){return[].slice.call(t.querySelectorAll("text.label")).forEach(function(t){t.innerHTML=[].slice.call(t.querySelectorAll("tspan")).reduce(function(t,e){var n=e.value||e.text||e.textContent||"";return t+(n=n.charAt(0).toUpperCase()+n.substr(1))},"")}),t},C=function(t,e,n){n=null==n?"":String(n);var r=(t||{})[e]||{};return(f.isObject(r.label)?r.label.text:r.label)||n};function E(t){return{onRender:function(){this._info=h.extractFieldsFormatInfo(this._chart.getSpec())},_normalizeExportFields:function(t,e){var n=this._info;return t.map(function(t){var e=t,r=n[t]||{};return"string"==typeof t&&(e={field:t,title:r.label||t}),"function"!=typeof e.value&&(e.value=function(t){var e=t[this.field];return r.isComplexField?(e||{})[r.tickLabel]:e}),e}).filter(function(t){return!e.find(function(e){return t.field===e})})},_handleError:function(t){o.a.api.globalSettings.log([p,t],"error")},_createDataUrl:function(t,e){var n=this;(function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];var n=t.slice(0,t.length-1),r=t[t.length-1];if(0!==n.length)var i=f.range(0,n.length).map(function(){return null}),o=n.map(function(t,e){var n=new XMLHttpRequest;n.onload=function(){if(n.status>=200&&n.status<300){var t=n.responseText;i[e]=t,o.splice(o.indexOf(n),1),0===o.length&&r(null,i)}else o.forEach(function(t){return t.abort()}),r(new Error(n.status+": "+n.statusText),null)},n.onerror=function(t){o.forEach(function(t){return t.abort()}),r(t.error,null)},n.open("GET",t,!0),n.send(null)});else setTimeout(function(){return r(null,[])},0)}).apply(void 0,this._cssPaths.concat([function(r,i){if(r)e(r,null);else try{var a=i.join(" ").replace(/&/g,""),u=b(a),c=document.createElement("div");t.fire("beforeExportSVGNode");var l=t.getSVG().cloneNode(!0);t.fire("afterExportSVGNode"),c.appendChild(O(l)),d.select(l).attr("version",1.1).attr("xmlns","http://www.w3.org/2000/svg"),l.insertBefore(u,l.firstChild),n._renderAdditionalInfo(l,t),n._addBackground(l,n._backgroundColor);var p=document.createElement("canvas");p.height=l.getAttribute("height"),p.width=l.getAttribute("width"),s()(p,l.parentNode.innerHTML,{renderCallback:function(t){var n=(new XMLSerializer).serializeToString(t);"<html"===n.substring(0,5).toLowerCase()&&(o.a.api.globalSettings.log("[export plugin]: canvg error","error"),o.a.api.globalSettings.log(n,"error")),e(null,p.toDataURL("image/png"))}})}catch(r){e(r,null)}}]))},_findUnit:function(t){var e=t.getSpec(),n=t.getSpec(),r=function(t){var e=n.scales[t];return e&&e.dim&&e.source&&n.sources[e.source].dims[e.dim]};return h.depthFirstSearch(e.unit,function(t){if(r(t.color))return!0;if(r(t.size)){var e=n.scales[t.size];return"measure"===n.sources[e.source].dims[e.dim].type}})},_toPng:function(t){var e=this;this._createDataUrl(t,function(t,n){if(t)e._handleError(t);else{for(var r=atob(n.substring("data:image/png;base64,".length)),i=new Uint8Array(r.length),o=0,a=r.length;o<a;++o)i[o]=r.charCodeAt(o);var s=new Blob([i.buffer],{type:"image/png"});Object(u.saveAs)(s,(e._fileName||"export")+".png")}})},_toPrint:function(t){var e=this;this._createDataUrl(t,function(t,n){if(t)e._handleError(t);else{(r=document.createElement("img")).classList.add("tau-chart__print-block");var i=r;document.body.appendChild(i),i.src=n,document.head.appendChild(x),i.onload=function(){window.print()}}})},_toJson:function(t){var e=this._exportFields,n=t.getData(),r=t.getDataDims(),i=e.length?e:Object.keys(r),o=this._normalizeExportFields(i.concat(this._appendFields),this._excludeFields),a=n.map(function(t){return o.reduce(function(e,n){return e[n.title]=n.value(t),e},{})}),s=JSON.stringify(a,null,2),u=(this._fileName||"export")+".json";j(u,"application/json",s)},_toCsv:function(t){var e=this._csvSeparator,n=this._exportFields,r=t.getData(),i=t.getDataDims(),o=n.length?n:Object.keys(i),a=this._normalizeExportFields(o.concat(this._appendFields),this._excludeFields),s=r.reduce(function(t,n){return t.concat(a.reduce(function(t,e){var r=e.value(n),i=JSON.stringify(r);if(!f.isDate(r)&&f.isObject(r))i='"'+i.replace(/"/g,'""')+'"';else{var o=g(i,'"').replace(/"/g,'""');i=Boolean(['"',",",";","\n","\r"].find(function(t){return o.indexOf(t)>=0}))?'"'+o+'"':o}return t.concat(i)},[]).join(e))},[a.map(function(t){return t.title}).join(e)]).join("\r\n"),u=(this._fileName||"export")+".csv";j(u,"text/csv",s)},_renderFillLegend:function(e,n,r,i){var o=function(t,e){var n=t[0],r=t[1],i=(r-n)/(e-1),o=f.range(e-2).map(function(t){return n+i*(t+1)});return[n].concat(o).concat(r)},a=this._unit.getScale("color"),s=C(e.guide,"color",a.dim).toUpperCase(),u="text-transform:uppercase;font-weight:600;font-size:"+t.fontSize+"px",c=a.domain().sort(function(t,e){return t-e}),l=a.brewer.length,d=t.fontSize,p=o(c,l).reverse().map(function(t,e){return'<stop offset="'+e/(l-1)*100+'%"      style="stop-color:'+a(t)+';stop-opacity:1" />'}),h=o(c,3).reverse().map(function(t,e,n){return'<text x="25" y="'+(e/2*100+.5*(0===e?d:e===n.length-1?-d:0)+d/2)+'">'+t+"</text>"}),m=['<svg version="1.1" xmlns="http://www.w3.org/2000/svg" height="120" width="100%">',"   <defs>",'       <linearGradient id="grad1" x1="0%" y1="0%" x2="0%" y2="100%">',p.join(""),"       </linearGradient>","   </defs>",'   <text x="0" y="10" style="'+u+'">'+s+"</text>",'   <g transform="translate(0,20)">','       <rect x="0" y="0" height="100" width="20" fill="url(#grad1)">',"       </rect>",h.join(""),"   </g>","   Sorry, your browser does not support inline SVG.","</svg>"].join(""),g=(new DOMParser).parseFromString(m,"application/xml").documentElement;return document.body.appendChild(g),n.append("g").attr("class","legend").attr("transform","translate("+(i+10)+","+t.paddingTop+")").node().appendChild(g),{h:120,w:0}},_renderColorLegend:function(e,n,r,i){var o=this._unit.getScale("color"),a=C(e.guide,"color",o.dim).toUpperCase(),s=this._getColorMap(r.getChartModelData({excludeFilter:["legend"]}),o,o.dim).values,u=n.append("g").attr("class","legend").attr("transform","translate("+(i+10)+","+t.paddingTop+")");return u.append("text").text(a.toUpperCase()).style({"text-transform":"uppercase","font-weight":"600","font-size":t.fontSize+"px"}),u.selectAll("g").data(s).enter().append("g").call(function(e){e.attr("transform",function(t,e){return"translate(5,"+20*(e+1)+")"}),e.append("circle").attr("r",6).attr("fill",function(t){return o.toColor(t.color)}).attr("class",function(t){return o.toClass(t.color)}),e.append("text").attr("x",12).attr("y",5).text(function(t){return f.escape(_(t.label)?"No "+a:t.label)}).style({"font-size":t.fontSize+"px"})}),{h:20*s.length,w:0}},_renderSizeLegend:function(e,n,r,i,o){var a,s,u=this._unit.getScale("size"),c=C(e.guide,"size",u.dim).toUpperCase(),l=r.getChartModelData().sort(function(t,e){return u(t[u.dim])-u(e[u.dim])}),d=l.length,p=l[0][u.dim],h=l[d-1][u.dim];if(h-p){var m=(s=h-p,Math.log(s)/Math.LN10),g=4-m<0?0:Math.round(4-m),_=Math.pow(10,g),v=(h-p)/5;a=f.unique([p,p+v,p+2*v,p+3*v,h].map(function(t){return t===h||t===p?t:Math.round(t*_)/_}))}else a=[p];var b=a.map(function(t){var n=u(t),r=n/2;return{diameter:y(n+2),radius:r,value:t,className:e.color?"color-definite":""}}.bind(this)).reverse(),x=Math.max.apply(null,b.map(function(t){return t.diameter})),w=t.fontSize,j=0,O=n.append("g").attr("class","legend").attr("transform","translate("+(i+10)+","+(t.paddingTop+o.h+20)+")");O.append("text").text(c.toUpperCase()).style({"text-transform":"uppercase","font-weight":"600","font-size":w+"px"}),O.selectAll("g").data(b).enter().append("g").call(function(t){t.attr("transform",function(){var t="translate(5,"+(j+=x)+")";return j+=10,t}),t.append("circle").attr("r",function(t){return t.radius}).attr("class",function(t){return t.className}).style({opacity:.4}),t.append("g").attr("transform",function(){return"translate("+x+","+w/2+")"}).append("text").attr("x",0).attr("y",0).text(function(t){return t.value}).style({"font-size":w+"px"})})},_renderAdditionalInfo:function(t,e){var n=this._findUnit(e);if(n){var r={h:0,w:0};t=d.select(t);var i=parseInt(t.attr("width"),10);t.attr("height"),t.attr("width",i+160);var o=e.getScaleInfo(n.color);if(o.dim&&!o.discrete){var a=this._renderFillLegend(n,t,e,i);r.h=a.h+20,r.w=a.w}if(o.dim&&o.discrete){var s=this._renderColorLegend(n,t,e,i);r.h=s.h+20,r.w=s.w}var u=e.getScaleInfo(n.size);u.dim&&!u.discrete&&this._renderSizeLegend(n,t,e,i,r)}},_addBackground:function(t,e){if(e&&"transparent"!==e){var n=document.createElementNS("http://www.w3.org/2000/svg","rect");n.setAttribute("fill",e),n.setAttribute("x",0),n.setAttribute("y",0),n.setAttribute("width",t.getAttribute("width")),n.setAttribute("height",t.getAttribute("height")),t.insertBefore(n,t.firstChild)}},onUnitDraw:function(t,e){o.a.api.isChartElement(e)&&"chart"===e.config.namespace&&(this._unit=e)},_getColorMap:function(t,e,n){return f.unique(t.map(function(t){var r=t[n];return{color:e(r),value:r,label:r}}),function(t){return t.value}).reduce(function(t,e){return t.brewer[e.value]=e.color,t.values.push(e),t},{brewer:{},values:[]})},_select:function(t,e){var n=this["_to"+(t=t||"").charAt(0).toUpperCase()+t.slice(1)];n&&n.call(this,e)},_handleMenu:function(t,e,n){t.addEventListener("click",function(t){if("a"===t.target.tagName.toLowerCase()){var r=t.target.getAttribute("data-value");this._select(r,e),n.hide()}}.bind(this)),t.addEventListener("mouseover",function(t){"a"===t.target.tagName.toLowerCase()&&t.target.focus()}.bind(this)),t.addEventListener("keydown",function(t){if(t.keyCode===v.ESCAPE&&n.hide(),t.keyCode===v.DOWN&&(t.target.parentNode.nextSibling?t.target.parentNode.nextSibling.childNodes[0].focus():t.target.parentNode.parentNode.firstChild.childNodes[0].focus()),t.keyCode===v.UP&&(t.target.parentNode.previousSibling?t.target.parentNode.previousSibling.childNodes[0].focus():t.target.parentNode.parentNode.lastChild.childNodes[0].focus()),t.keyCode===v.ENTER){var r=t.target.getAttribute("data-value");this._select(r,e)}t.preventDefault()}.bind(this));var r=null,i=function(){r=setTimeout(function(){n.hide()},100)},o=function(){clearTimeout(r)},a=function(){n.toggle(),n.hidden||t.querySelectorAll("a")[0].focus()};t.addEventListener("blur",i,!0),t.addEventListener("focus",o,!0),this._container.addEventListener("click",a),this._onDestroy(function(){t.removeEventListener("blur",i,!0),t.removeEventListener("focus",o,!0),this._container.removeEventListener("click",a),clearTimeout(r)})},init:function(e){t=t||{},t=f.defaults(t,{backgroundColor:"white",visible:!0,fontSize:13,paddingTop:30}),this._chart=e,this._info={},this._cssPaths=t.cssPaths,this._fileName=t.fileName,this._backgroundColor=t.backgroundColor||"white",this._destroyListeners=[],this._csvSeparator=t.csvSeparator||",",this._exportFields=t.exportFields||[],this._appendFields=t.appendFields||[],this._excludeFields=t.excludeFields||[],this._cssPaths||(this._cssPaths=[],o.a.api.globalSettings.log('[export plugin]: the "cssPath" parameter should be specified for correct operation',"warn"));var n=t.visible?"":"display:none";this._container=e.insertToHeader('<a class="tau-chart__export" style="'+n+'">Export</a>');var r=e.addBalloon({place:"bottom-left"});r.content(['<ul class="tau-chart__export__list">','<li class="tau-chart__export__item">','   <a data-value="print" tabindex="1">'+m.get("Print")+"</a>","</li>",'<li class="tau-chart__export__item">','   <a data-value="png" tabindex="2">'+m.get("Export to png")+"</a>","</li>",'<li class="tau-chart__export__item">','   <a data-value="csv" tabindex="3">'+m.get("Export to CSV")+"</a>","</li>",'<li class="tau-chart__export__item">','   <a data-value="json" tabindex="4">'+m.get("Export to JSON")+"</a>","</li>","</ul>"].join("")),r.attach(this._container);var i=r.getElement();i.setAttribute("tabindex","-1"),this._handleMenu(i,e,r);var a=e.on("export-to",function(t,e){var n=e.type,r=e.fileName;this._fileName=r||this._fileName,this._select(n,t)}.bind(this)),s=e.on("exportTo",function(t,e){o.a.api.globalSettings.log([p,"`exportTo` event is deprecated, use `export-to` instead."],"warn"),this._select(e,t)}.bind(this));this._onDestroy(function(){r.destroy(),e.removeHandler(a),e.removeHandler(s)})},_onDestroy:function(t){this._destroyListeners.push(t)},destroy:function(){var t;this._destroyListeners.forEach(function(t){t.call(this)},this),(t=this._container)&&t.parentElement&&t.parentElement.removeChild(t)}}}o.a.api.plugins.add("export-to",E),e.default=E},function(t,e){t.exports=function(t){this.ok=!1,this.alpha=1,"#"==t.charAt(0)&&(t=t.substr(1,6)),t=(t=t.replace(/ /g,"")).toLowerCase();var e={aliceblue:"f0f8ff",antiquewhite:"faebd7",aqua:"00ffff",aquamarine:"7fffd4",azure:"f0ffff",beige:"f5f5dc",bisque:"ffe4c4",black:"000000",blanchedalmond:"ffebcd",blue:"0000ff",blueviolet:"8a2be2",brown:"a52a2a",burlywood:"deb887",cadetblue:"5f9ea0",chartreuse:"7fff00",chocolate:"d2691e",coral:"ff7f50",cornflowerblue:"6495ed",cornsilk:"fff8dc",crimson:"dc143c",cyan:"00ffff",darkblue:"00008b",darkcyan:"008b8b",darkgoldenrod:"b8860b",darkgray:"a9a9a9",darkgreen:"006400",darkkhaki:"bdb76b",darkmagenta:"8b008b",darkolivegreen:"556b2f",darkorange:"ff8c00",darkorchid:"9932cc",darkred:"8b0000",darksalmon:"e9967a",darkseagreen:"8fbc8f",darkslateblue:"483d8b",darkslategray:"2f4f4f",darkturquoise:"00ced1",darkviolet:"9400d3",deeppink:"ff1493",deepskyblue:"00bfff",dimgray:"696969",dodgerblue:"1e90ff",feldspar:"d19275",firebrick:"b22222",floralwhite:"fffaf0",forestgreen:"228b22",fuchsia:"ff00ff",gainsboro:"dcdcdc",ghostwhite:"f8f8ff",gold:"ffd700",goldenrod:"daa520",gray:"808080",green:"008000",greenyellow:"adff2f",honeydew:"f0fff0",hotpink:"ff69b4",indianred:"cd5c5c",indigo:"4b0082",ivory:"fffff0",khaki:"f0e68c",lavender:"e6e6fa",lavenderblush:"fff0f5",lawngreen:"7cfc00",lemonchiffon:"fffacd",lightblue:"add8e6",lightcoral:"f08080",lightcyan:"e0ffff",lightgoldenrodyellow:"fafad2",lightgrey:"d3d3d3",lightgreen:"90ee90",lightpink:"ffb6c1",lightsalmon:"ffa07a",lightseagreen:"20b2aa",lightskyblue:"87cefa",lightslateblue:"8470ff",lightslategray:"778899",lightsteelblue:"b0c4de",lightyellow:"ffffe0",lime:"00ff00",limegreen:"32cd32",linen:"faf0e6",magenta:"ff00ff",maroon:"800000",mediumaquamarine:"66cdaa",mediumblue:"0000cd",mediumorchid:"ba55d3",mediumpurple:"9370d8",mediumseagreen:"3cb371",mediumslateblue:"7b68ee",mediumspringgreen:"00fa9a",mediumturquoise:"48d1cc",mediumvioletred:"c71585",midnightblue:"191970",mintcream:"f5fffa",mistyrose:"ffe4e1",moccasin:"ffe4b5",navajowhite:"ffdead",navy:"000080",oldlace:"fdf5e6",olive:"808000",olivedrab:"6b8e23",orange:"ffa500",orangered:"ff4500",orchid:"da70d6",palegoldenrod:"eee8aa",palegreen:"98fb98",paleturquoise:"afeeee",palevioletred:"d87093",papayawhip:"ffefd5",peachpuff:"ffdab9",peru:"cd853f",pink:"ffc0cb",plum:"dda0dd",powderblue:"b0e0e6",purple:"800080",rebeccapurple:"663399",red:"ff0000",rosybrown:"bc8f8f",royalblue:"4169e1",saddlebrown:"8b4513",salmon:"fa8072",sandybrown:"f4a460",seagreen:"2e8b57",seashell:"fff5ee",sienna:"a0522d",silver:"c0c0c0",skyblue:"87ceeb",slateblue:"6a5acd",slategray:"708090",snow:"fffafa",springgreen:"00ff7f",steelblue:"4682b4",tan:"d2b48c",teal:"008080",thistle:"d8bfd8",tomato:"ff6347",turquoise:"40e0d0",violet:"ee82ee",violetred:"d02090",wheat:"f5deb3",white:"ffffff",whitesmoke:"f5f5f5",yellow:"ffff00",yellowgreen:"9acd32"};t=e[t]||t;for(var n=[{re:/^rgba\((\d{1,3}),\s*(\d{1,3}),\s*(\d{1,3}),\s*((?:\d?\.)?\d)\)$/,example:["rgba(123, 234, 45, 0.8)","rgba(255,234,245,1.0)"],process:function(t){return[parseInt(t[1]),parseInt(t[2]),parseInt(t[3]),parseFloat(t[4])]}},{re:/^rgb\((\d{1,3}),\s*(\d{1,3}),\s*(\d{1,3})\)$/,example:["rgb(123, 234, 45)","rgb(255,234,245)"],process:function(t){return[parseInt(t[1]),parseInt(t[2]),parseInt(t[3])]}},{re:/^([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})$/,example:["#00ff00","336699"],process:function(t){return[parseInt(t[1],16),parseInt(t[2],16),parseInt(t[3],16)]}},{re:/^([0-9a-fA-F]{1})([0-9a-fA-F]{1})([0-9a-fA-F]{1})$/,example:["#fb0","f0f"],process:function(t){return[parseInt(t[1]+t[1],16),parseInt(t[2]+t[2],16),parseInt(t[3]+t[3],16)]}}],r=0;r<n.length;r++){var i=n[r].re,o=n[r].process,a=i.exec(t);if(a){var s=o(a);this.r=s[0],this.g=s[1],this.b=s[2],s.length>3&&(this.alpha=s[3]),this.ok=!0}}this.r=this.r<0||isNaN(this.r)?0:this.r>255?255:this.r,this.g=this.g<0||isNaN(this.g)?0:this.g>255?255:this.g,this.b=this.b<0||isNaN(this.b)?0:this.b>255?255:this.b,this.alpha=this.alpha<0?0:this.alpha>1||isNaN(this.alpha)?1:this.alpha,this.toRGB=function(){return"rgb("+this.r+", "+this.g+", "+this.b+")"},this.toRGBA=function(){return"rgba("+this.r+", "+this.g+", "+this.b+", "+this.alpha+")"},this.toHex=function(){var t=this.r.toString(16),e=this.g.toString(16),n=this.b.toString(16);return 1==t.length&&(t="0"+t),1==e.length&&(e="0"+e),1==n.length&&(n="0"+n),"#"+t+e+n},this.getHelpXML=function(){for(var t=new Array,r=0;r<n.length;r++)for(var i=n[r].example,o=0;o<i.length;o++)t[t.length]=i[o];for(var a in e)t[t.length]=a;var s=document.createElement("ul");for(s.setAttribute("id","rgbcolor-examples"),r=0;r<t.length;r++)try{var u=document.createElement("li"),c=new RGBColor(t[r]),l=document.createElement("div");l.style.cssText="margin: 3px; border: 1px solid black; background:"+c.toHex()+"; color:"+c.toHex(),l.appendChild(document.createTextNode("test"));var d=document.createTextNode(" "+t[r]+" -> "+c.toRGB()+" -> "+c.toHex());u.appendChild(l),u.appendChild(d),s.appendChild(u)}catch(t){}return s}}},function(t,e){var n=[512,512,456,512,328,456,335,512,405,328,271,456,388,335,292,512,454,405,364,328,298,271,496,456,420,388,360,335,312,292,273,512,482,454,428,405,383,364,345,328,312,298,284,271,259,496,475,456,437,420,404,388,374,360,347,335,323,312,302,292,282,273,265,512,497,482,468,454,441,428,417,405,394,383,373,364,354,345,337,328,320,312,305,298,291,284,278,271,265,259,507,496,485,475,465,456,446,437,428,420,412,404,396,388,381,374,367,360,354,347,341,335,329,323,318,312,307,302,297,292,287,282,278,273,269,265,261,512,505,497,489,482,475,468,461,454,447,441,435,428,422,417,411,405,399,394,389,383,378,373,368,364,359,354,350,345,341,337,332,328,324,320,316,312,309,305,301,298,294,291,287,284,281,278,274,271,268,265,262,259,257,507,501,496,491,485,480,475,470,465,460,456,451,446,442,437,433,428,424,420,416,412,408,404,400,396,392,388,385,381,377,374,370,367,363,360,357,354,350,347,344,341,338,335,332,329,326,323,320,318,315,312,310,307,304,302,299,297,294,292,289,287,285,282,280,278,275,273,271,269,267,265,263,261,259],r=[9,11,12,13,13,14,14,15,15,15,15,16,16,16,16,17,17,17,17,17,17,17,18,18,18,18,18,18,18,18,18,19,19,19,19,19,19,19,19,19,19,19,19,19,19,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24];function i(t,e,n,r,i){if("string"==typeof t)t=document.getElementById(t);else if("undefined"!=typeof HTMLCanvasElement&&!t instanceof HTMLCanvasElement)return;var o,a=t.getContext("2d");try{try{o=a.getImageData(e,n,r,i)}catch(t){throw new Error("unable to access local image data: "+t)}}catch(t){throw new Error("unable to access image data: "+t)}return o}function o(t,e,n,r,o,s){if(!(isNaN(s)||s<1)){s|=0;var u=i(t,e,n,r,o);u=a(u,0,0,r,o,s),t.getContext("2d").putImageData(u,e,n)}}function a(t,e,i,o,a,s){var u,l,d,p,f,h,m,g,y,_,v,b,x,w,j,O,C,E,S,M,T,k,A,D,I=t.data,P=s+s+1,R=o-1,N=a-1,F=s+1,L=F*(F+1)/2,B=new c,V=B;for(d=1;d<P;d++)if(V=V.next=new c,d==F)var z=V;V.next=B;var H=null,U=null;m=h=0;var W=n[s],q=r[s];for(l=0;l<a;l++){for(O=C=E=S=g=y=_=v=0,b=F*(M=I[h]),x=F*(T=I[h+1]),w=F*(k=I[h+2]),j=F*(A=I[h+3]),g+=L*M,y+=L*T,_+=L*k,v+=L*A,V=B,d=0;d<F;d++)V.r=M,V.g=T,V.b=k,V.a=A,V=V.next;for(d=1;d<F;d++)p=h+((R<d?R:d)<<2),g+=(V.r=M=I[p])*(D=F-d),y+=(V.g=T=I[p+1])*D,_+=(V.b=k=I[p+2])*D,v+=(V.a=A=I[p+3])*D,O+=M,C+=T,E+=k,S+=A,V=V.next;for(H=B,U=z,u=0;u<o;u++)I[h+3]=A=v*W>>q,0!=A?(A=255/A,I[h]=(g*W>>q)*A,I[h+1]=(y*W>>q)*A,I[h+2]=(_*W>>q)*A):I[h]=I[h+1]=I[h+2]=0,g-=b,y-=x,_-=w,v-=j,b-=H.r,x-=H.g,w-=H.b,j-=H.a,p=m+((p=u+s+1)<R?p:R)<<2,g+=O+=H.r=I[p],y+=C+=H.g=I[p+1],_+=E+=H.b=I[p+2],v+=S+=H.a=I[p+3],H=H.next,b+=M=U.r,x+=T=U.g,w+=k=U.b,j+=A=U.a,O-=M,C-=T,E-=k,S-=A,U=U.next,h+=4;m+=o}for(u=0;u<o;u++){for(C=E=S=O=y=_=v=g=0,b=F*(M=I[h=u<<2]),x=F*(T=I[h+1]),w=F*(k=I[h+2]),j=F*(A=I[h+3]),g+=L*M,y+=L*T,_+=L*k,v+=L*A,V=B,d=0;d<F;d++)V.r=M,V.g=T,V.b=k,V.a=A,V=V.next;for(f=o,d=1;d<=s;d++)h=f+u<<2,g+=(V.r=M=I[h])*(D=F-d),y+=(V.g=T=I[h+1])*D,_+=(V.b=k=I[h+2])*D,v+=(V.a=A=I[h+3])*D,O+=M,C+=T,E+=k,S+=A,V=V.next,d<N&&(f+=o);for(h=u,H=B,U=z,l=0;l<a;l++)I[3+(p=h<<2)]=A=v*W>>q,A>0?(A=255/A,I[p]=(g*W>>q)*A,I[p+1]=(y*W>>q)*A,I[p+2]=(_*W>>q)*A):I[p]=I[p+1]=I[p+2]=0,g-=b,y-=x,_-=w,v-=j,b-=H.r,x-=H.g,w-=H.b,j-=H.a,p=u+((p=l+F)<N?p:N)*o<<2,g+=O+=H.r=I[p],y+=C+=H.g=I[p+1],_+=E+=H.b=I[p+2],v+=S+=H.a=I[p+3],H=H.next,b+=M=U.r,x+=T=U.g,w+=k=U.b,j+=A=U.a,O-=M,C-=T,E-=k,S-=A,U=U.next,h+=o}return t}function s(t,e,n,r,o,a){if(!(isNaN(a)||a<1)){a|=0;var s=i(t,e,n,r,o);s=u(s,0,0,r,o,a),t.getContext("2d").putImageData(s,e,n)}}function u(t,e,i,o,a,s){var u,l,d,p,f,h,m,g,y,_,v,b,x,w,j,O,C,E,S,M,T=t.data,k=s+s+1,A=o-1,D=a-1,I=s+1,P=I*(I+1)/2,R=new c,N=R;for(d=1;d<k;d++)if(N=N.next=new c,d==I)var F=N;N.next=R;var L=null,B=null;m=h=0;var V=n[s],z=r[s];for(l=0;l<a;l++){for(w=j=O=g=y=_=0,v=I*(C=T[h]),b=I*(E=T[h+1]),x=I*(S=T[h+2]),g+=P*C,y+=P*E,_+=P*S,N=R,d=0;d<I;d++)N.r=C,N.g=E,N.b=S,N=N.next;for(d=1;d<I;d++)p=h+((A<d?A:d)<<2),g+=(N.r=C=T[p])*(M=I-d),y+=(N.g=E=T[p+1])*M,_+=(N.b=S=T[p+2])*M,w+=C,j+=E,O+=S,N=N.next;for(L=R,B=F,u=0;u<o;u++)T[h]=g*V>>z,T[h+1]=y*V>>z,T[h+2]=_*V>>z,g-=v,y-=b,_-=x,v-=L.r,b-=L.g,x-=L.b,p=m+((p=u+s+1)<A?p:A)<<2,g+=w+=L.r=T[p],y+=j+=L.g=T[p+1],_+=O+=L.b=T[p+2],L=L.next,v+=C=B.r,b+=E=B.g,x+=S=B.b,w-=C,j-=E,O-=S,B=B.next,h+=4;m+=o}for(u=0;u<o;u++){for(j=O=w=y=_=g=0,v=I*(C=T[h=u<<2]),b=I*(E=T[h+1]),x=I*(S=T[h+2]),g+=P*C,y+=P*E,_+=P*S,N=R,d=0;d<I;d++)N.r=C,N.g=E,N.b=S,N=N.next;for(f=o,d=1;d<=s;d++)h=f+u<<2,g+=(N.r=C=T[h])*(M=I-d),y+=(N.g=E=T[h+1])*M,_+=(N.b=S=T[h+2])*M,w+=C,j+=E,O+=S,N=N.next,d<D&&(f+=o);for(h=u,L=R,B=F,l=0;l<a;l++)T[p=h<<2]=g*V>>z,T[p+1]=y*V>>z,T[p+2]=_*V>>z,g-=v,y-=b,_-=x,v-=L.r,b-=L.g,x-=L.b,p=u+((p=l+I)<D?p:D)*o<<2,g+=w+=L.r=T[p],y+=j+=L.g=T[p+1],_+=O+=L.b=T[p+2],L=L.next,v+=C=B.r,b+=E=B.g,x+=S=B.b,w-=C,j-=E,O-=S,B=B.next,h+=o}return t}function c(){this.r=0,this.g=0,this.b=0,this.a=0,this.next=null}t.exports={image:function(t,e,n,r){if("string"==typeof t)t=document.getElementById(t);else if("undefined"!=typeof HTMLImageElement&&!t instanceof HTMLImageElement)return;var i=t.naturalWidth,a=t.naturalHeight;if("string"==typeof e)e=document.getElementById(e);else if("undefined"!=typeof HTMLCanvasElement&&!e instanceof HTMLCanvasElement)return;e.style.width=i+"px",e.style.height=a+"px",e.width=i,e.height=a;var u=e.getContext("2d");u.clearRect(0,0,i,a),u.drawImage(t,0,0),isNaN(n)||n<1||(r?o(e,0,0,i,a,n):s(e,0,0,i,a,n))},canvasRGBA:o,canvasRGB:s,imageDataRGBA:a,imageDataRGB:u}},function(t,e){var n;n=function(){return this}();try{n=n||new Function("return this")()}catch(t){"object"==typeof window&&(n=window)}t.exports=n},function(t,e,n){"use strict";t.exports=function(t){var e=[];return e.toString=function(){return this.map(function(e){var n=function(t,e){var n,r=t[1]||"",i=t[3];if(!i)return r;if(e&&"function"==typeof btoa){var o=(n=i,"/*# sourceMappingURL=data:application/json;charset=utf-8;base64,"+btoa(unescape(encodeURIComponent(JSON.stringify(n))))+" */"),a=i.sources.map(function(t){return"/*# sourceURL="+i.sourceRoot+t+" */"});return[r].concat(a).concat([o]).join("\n")}return[r].join("\n")}(e,t);return e[2]?"@media "+e[2]+"{"+n+"}":n}).join("")},e.i=function(t,n){"string"==typeof t&&(t=[[null,t,""]]);for(var r={},i=0;i<this.length;i++){var o=this[i][0];null!=o&&(r[o]=!0)}for(i=0;i<t.length;i++){var a=t[i];null!=a[0]&&r[a[0]]||(n&&!a[2]?a[2]=n:n&&(a[2]="("+a[2]+") and ("+n+")"),e.push(a))}},e}}]))},"./node_modules/taucharts/dist/plugins/legend.js":function(t,e,n){var r,i;window,t.exports=(r=n("./node_modules/taucharts/dist/taucharts.js"),i=n("./node_modules/taucharts/node_modules/d3-format/src/index.js"),function(t){var e={};function n(r){if(e[r])return e[r].exports;var i=e[r]={i:r,l:!1,exports:{}};return t[r].call(i.exports,i,i.exports,n),i.l=!0,i.exports}return n.m=t,n.c=e,n.d=function(t,e,r){n.o(t,e)||Object.defineProperty(t,e,{enumerable:!0,get:r})},n.r=function(t){"undefined"!=typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(t,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(t,"__esModule",{value:!0})},n.t=function(t,e){if(1&e&&(t=n(t)),8&e)return t;if(4&e&&"object"==typeof t&&t&&t.__esModule)return t;var r=Object.create(null);if(n.r(r),Object.defineProperty(r,"default",{enumerable:!0,value:t}),2&e&&"string"!=typeof t)for(var i in t)n.d(r,i,function(e){return t[e]}.bind(null,i));return r},n.n=function(t){var e=t&&t.__esModule?function(){return t.default}:function(){return t};return n.d(e,"a",e),e},n.o=function(t,e){return Object.prototype.hasOwnProperty.call(t,e)},n.p="",n(n.s=28)}({0:function(t,e){t.exports=r},28:function(t,e,n){"use strict";n.r(e);var r,i,o=n(0),a=n.n(o),s=n(9),u=a.a.api.utils,c=a.a.api.pluginsSDK,l=".tau-chart__legend__reset",d=".tau-chart__legend__item-color",p=".tau-chart__legend__guide--color__overlay",f=4,h=13,m=0,g=function(){return++m},y=a.a.api.utils.xml,_=function(t,e){var n=t[0],r=t[1],i=(r-n)/(e-1),o=u.range(e-2).map(function(t){return n+i*(t+1)});return[n].concat(o).concat(r)},v=function(t,e,n){if(t.length<3)return t.slice(0);if(e<3)return[t[0],t[t.length-1]];var r,i=t[0]<0?Math.abs(t[0]):0,o=function(t){return t},a="sqrt"===n?function(t){return Math.sqrt(t+i)}:o,s="sqrt"===n?function(t){return Math.pow(t,2)-i}:o,c=[(t=t.map(a))[0]],l=t[t.length-1]-t[0],d=.5*l/(e-1),p=u.range(1,e-1).map(function(t){var n=l*t/(e-1);return{min:n-d,mid:n,max:n+d,diff:Number.MAX_VALUE,closest:null}}),f=0,h=function(){if(f!==p.length){var t=r;(r=p[f++]).min=Math.max(r.min,(t&&null!==t.closest?t.closest:c[0])+d)}};return h(),t.forEach(function(t){if(!(t<r.min)){t>r.max&&h();var e=Math.abs(t-r.mid);e<r.diff&&e<d?(r.diff=e,r.closest=t):h(),0===e&&h()}}),p.forEach(function(t){null!==t.closest&&c.push(t.closest)}),c.push(t[t.length-1]),c=c.map(s)},b=function(t){return Math.log(t)/Math.LN10},x=function(t){return 0===t?0:Math.floor(b(Math.abs(t)))},w=(r=/\.0+([^\d].*)?$/,i=/(\.\d+?)0+([^\d].*)?$/,function(t){return t.replace(r,"$1").replace(i,"$1$2")}),j=s.format(".3s"),O=function(t){return w(j(t))},C=function(t,e){var n=Math.max(Math.abs(t),Math.abs(e)),r=x(n),i=t*e>0?Math.abs(e-t):n,o=x(i),a=Math.abs(r-o);return Math.abs(r)>3&&a<=3?O:function(t){var e=x(n-t),r=Math.min((o<0?Math.abs(o):0)+(e<o?1:0),20);return w(t.toFixed(r))}};function E(t){var e=!0,n=u.defaults(t||{},{formatters:{},onSelect:function(){}}),r=function(t){return null===t||""===t||void 0===t},i=function(t){return t.every(function(t){return u.isDate(t)})},o=function(t,e){return function(n){var i=n[t],o=JSON.stringify(r(i)?null:i);return e===o}},a=function(t,e,n,r){t.addEventListener(e,function(t){for(var e=t.target;e!==t.currentTarget&&null!==e;)e.matches(n)&&r(t,e),e=e.parentNode})};return{init:function(t){var e=this;this.instanceId=g(),this._chart=t,this._currentFilters={},this._legendColorByScaleId={},this._legendOrderState={};var r=this._chart.getSpec(),i=function(t){return function(e,n){var i=r.scales[n];return i.type===t&&i.dim&&e.push(n),e}};this._color=Object.keys(r.scales).reduce(i("color"),[]).filter(function(e){return t.getScaleInfo(e).discrete}),this._fill=Object.keys(r.scales).reduce(i("color"),[]).filter(function(e){return!t.getScaleInfo(e).discrete}),this._size=Object.keys(r.scales).reduce(i("size"),[]);var o=this._color.length>0,s=this._fill.length>0,u=this._size.length>0;if(this._assignStaticBrewersOrEx(),o||s||u){switch(n.position){case"left":this._container=this._chart.insertToLeftSidebar(this._containerTemplate);break;case"right":this._container=this._chart.insertToRightSidebar(this._containerTemplate);break;case"top":this._container=this._chart.insertToHeader(this._containerTemplate);break;case"bottom":this._container=this._chart.insertToFooter(this._containerTemplate);break;default:this._container=this._chart.insertToRightSidebar(this._containerTemplate)}o&&(a(this._container,"click",l,function(t,n){e._toggleLegendItem(n,"reset")}),a(this._container,"click",d,function(t,n){var r=t.ctrlKey||t.target.matches(p)?"leave-others":"focus-single";e._toggleLegendItem(n,r)}),a(this._container,"mouseover",d,function(t,n){e._highlightToggle(n,!0)}),a(this._container,"mouseout",d,function(t,n){e._highlightToggle(n,!1)}))}},destroy:function(){var t=this._currentFilters,e=this._chart;Object.keys(t).forEach(function(n){return e.removeFilter(t[n])}),this._container&&this._container.parentElement&&(this._clearPanel(),this._container.parentElement.removeChild(this._container))},onSpecReady:function(t,e){this._formatters=c.getFieldFormatters(e,n.formatters)},_getFormat:function(t){return this._formatters[t]?this._formatters[t].format:function(t){return String(t)}},onRender:function(){var t=this;if(e&&n.selectedCategories&&0!==n.selectedCategories.length){var r=this._getLegendColorByScales();return Object.keys(r).forEach(function(e){r[e].legendColorItems.forEach(function(e){var r=e.value,i=e.dim;if(-1===n.selectedCategories.indexOf(JSON.parse(r))){var a=i+r,s=o(i,r);t._currentFilters[a]=t._chart.addFilter({tag:"legend",predicate:function(t){return!s(t)}})}})}),e=!1,void this._chart.refresh()}this._clearPanel(),this._drawColorLegend(),this._drawFillLegend(),this._drawSizeLegend()},_containerTemplate:'<div class="tau-chart__legend"></div>',_template:u.template(['<div class="tau-chart__legend__wrap">',"<%=top%>",'<div class="tau-chart__legend__title"><%=name%></div>',"<%=items%>","</div>"].join("")),_itemTemplate:u.template(["<div data-scale-id='<%= scaleId %>' data-dim='<%= dim %>' data-value='<%= value %>' class=\"tau-chart__legend__item tau-chart__legend__item-color <%=classDisabled%>\">",'   <div class="tau-chart__legend__guide__wrap">','   <div class="tau-chart__legend__guide tau-chart__legend__guide--color <%=cssClass%>"','        style="background-color: <%=cssColor%>; border-color: <%=borderColor%>;">','       <div class="tau-chart__legend__guide--color__overlay">',"       </div>","   </div>","   </div>",'   <span class="tau-chart__legend__guide__label"><%=label%></span>',"</div>"].join("")),_resetTemplate:u.template(['<div class="tau-chart__legend__reset <%=classDisabled%>">','    <div role="button" class="tau-chart__button">Reset</div>',"</div>"].join("")),_clearPanel:function(){this._container&&(clearTimeout(this._scrollTimeout),this._getScrollContainer().removeEventListener("scroll",this._scrollListener),this._container.innerHTML="")},_drawFillLegend:function(){var t=this;t._fill.forEach(function(e){var n,r,o,a,s,l=t._chart.select(function(t){return t.config.color===e})[0];if(l){var d=l.config.guide||{},p=l.getScale("color"),f=p.domain().sort(function(t,e){return t-e}),m=i(f),g=m?f.map(Number):f,v=C(g[0],g[g.length-1]),b=(a=t._chart.getSpec(),(s=c.extractFieldsFormatInfo(a)[p.dim].format)||(s=function(t){return new Date(t)}),function(t){return String(s(t))}),x=m?b:v,w=p.brewer.length,j=((d.color||{}).label||{}).text||p.dim,O=function(t){return t.length*h*.618},E=p.isInteger?(g[1]-g[0])%3==0?4:(g[1]-g[0])%2==0?3:2:3,S=_(g,E),M=(m?S.map(function(t){return new Date(t)}):S).map(x);M[0]===M[M.length-1]&&(M=[M[0]]),t._container.insertAdjacentHTML("beforeend",t._template({name:u.escape(j),top:null,items:'<div class="tau-chart__legend__gradient-wrapper"></div>'}));var T=t._container.lastElementChild.querySelector(".tau-chart__legend__gradient-wrapper"),k=T.getBoundingClientRect().width,A=!1;M.reduce(function(t,e){return t+O(e)},0)>k&&(M.length>1&&O(M[0])+O(M[M.length-1])>k?A=!0:M=[M[0],M[M.length-1]]);var D=A?(o=-.382*h/2,{width:k,height:120,barX:0,barY:0,barWidth:20,barHeight:120,textAnchor:"start",textX:u.range(E).map(function(){return 25}),textY:1===M.length?60+.618*h:M.map(function(t,e){var n=(M.length-1-e)/(M.length-1);return h*(1-n)+120*n+o})}):(n=O(M[0])/2,r=O(M[M.length-1])/2,{width:k,height:28+h,barX:0,barY:0,barWidth:k,barHeight:20,textAnchor:"middle",textX:1===M.length?[k/2]:M.map(function(t,e){var i=e/(M.length-1);return n*(1-i)+(k-r)*i}),textY:u.range(E).map(function(){return 28+h})}),I=_(g,w).map(function(t,e){return y("stop",{offset:e/(w-1)*100+"%",style:"stop-color:"+p(t)+';stop-opacity:1"'})}),P="legend-gradient-"+t.instanceId,R=y.apply(void 0,["svg",{class:"tau-chart__legend__gradient",width:D.width,height:D.height},y("defs",y.apply(void 0,["linearGradient",{id:P,x1:"0%",y1:A?"100%":"0%",x2:A?"0%":"100%",y2:"0%"}].concat(I))),y("rect",{class:"tau-chart__legend__gradient__bar",x:D.barX,y:D.barY,width:D.barWidth,height:D.barHeight,fill:"url(#"+P+")"})].concat(M.map(function(t,e){return y("text",{x:D.textX[e],y:D.textY[e],"text-anchor":D.textAnchor},t)})));T.insertAdjacentHTML("beforeend",R)}})},_drawSizeLegend:function(){var t=this;t._size.forEach(function(e){var n=t._chart.select(function(t){return t.config.size===e})[0];if(n){var r=n.config.guide||{},i=n.getScale("size"),o=i.domain().sort(function(t,e){return t-e});if(!Array.isArray(o)||!o.every(isFinite))return;var a=((r.size||{}).label||{}).text||i.dim,s=o[0],c=o[o.length-1],l=[s];if(c-s){var d=b(c-s),p=Math.round(4-d),m=Math.pow(10,p),g=u.unique(t._chart.getDataSources({excludeFilter:["legend"]})[i.source].data.map(function(t){return t[i.dim]}).filter(function(t){return t>=s&&t<=c})).sort(function(t,e){return t-e}),_=v(g,f,i.funcType);l=u.unique(_.map(function(t){return Math.round(t*m)/m}))}var x=C(l[0],l[l.length-1]),w=function(t){return t.length*h*.618};l.reverse();var j=l.map(i),O=Math.max.apply(null,j),E=l.map(x);t._container.insertAdjacentHTML("beforeend",t._template({name:u.escape(a),top:null,items:'<div class="tau-chart__legend__size-wrapper"></div>'}));var S=t._container.lastElementChild.querySelector(".tau-chart__legend__size-wrapper"),M=S.getBoundingClientRect().width,T=!1;(Math.max.apply(null,E.map(w))>M/4||1===E.length)&&(T=!0);var k=T?function(){for(var t=h,e=j[0]/2,n=j[j.length-1]/2,r=[e],i=1,o=void 0,a=void 0;i<j.length;i++)a=j[i-1]/2,o=j[i]/2,r.push(r[i-1]+Math.max(1.618*h,a+t+o));var s=.618*h/2;return{width:M,height:r[r.length-1]+Math.max(n,h/2),circleX:u.range(j.length).map(function(){return O/2}),circleY:r,textAnchor:"start",textX:u.range(E.length).map(function(){return O+8}),textY:r.map(function(t){return t+s})}}():function(){for(var t=Math.max(w(E[0])/2,j[0]/2),e=Math.max(w(E[E.length-1])/2,j[j.length-1]/2),n=(M-j.reduce(function(t,e,n){return t+(0===n||n===j.length-1?e/2:e)},0)-t-e)/(f-1),r=[t],i=1,o=void 0,a=void 0;i<j.length;i++)a=j[i-1]/2,o=j[i]/2,r.push(r[i-1]+a+n+o);var s=j.map(function(t){return O-t/2});return{width:M,height:O+8+h,circleX:r,circleY:s,textAnchor:"middle",textX:r,textY:u.range(E.length).map(function(){return O+8+h})}}(),A=y.apply(void 0,["svg",{class:"tau-chart__legend__size",width:k.width,height:k.height}].concat(j.map(function(t,e){return y("circle",{class:"tau-chart__legend__size__item__circle "+(n.config.color?"color-definite":"color-default-size"),cx:k.circleX[e],cy:k.circleY[e],r:t/2})}),E.map(function(t,e){return y("text",{class:"tau-chart__legend__size__item__label",x:k.textX[e],y:k.textY[e],"text-anchor":k.textAnchor},t)})));S.insertAdjacentHTML("beforeend",A)}})},_getLegendColorByScales:function(){var t=this;return t._color.reduce(function(e,n){var o=t._chart.select(function(t){return t.config.color===n})[0];if(o){var a=o.config.guide||{},s=o.getScale("color"),c=t._chart.getDataSources({excludeFilter:["legend"]}),l=u.unique(c[s.source].data.map(function(t){return t[s.dim]})),d=t._chart.getSpec().scales[n],p=i(l);if(d.order)l=u.union(u.intersection(d.order,l),l);else if("order"===d.dimType&&p)l=l.sort(function(t,e){return new Date(t)-new Date(e)});else{var f=t._legendOrderState[n];l=l.sort(function(t,e){var n=f[t]-f[e];return n&&n/Math.abs(n)})}var h=((a.color||{}).label||{}).text||s.dim,m=(a.color||{}).tickFormatNullAlias||"No "+h,g=t._getFormat(s.dim),y=l.map(function(e){var i=JSON.stringify(r(e)?null:e),o=s.dim+i;return{scaleId:n,dim:s.dim,color:s(e),disabled:t._currentFilters.hasOwnProperty(o),label:g(e),value:i}});e[n]={legendColorItems:y,title:h,colorScale:s,noVal:m}}return e},{})},_drawColorLegend:function(){var t=this,e=this._getLegendColorByScales();Object.keys(e).forEach(function(n){var i=e[n],o=i.legendColorItems,a=i.title,s=i.colorScale,c=i.noVal;t._container.insertAdjacentHTML("beforeend",t._template({name:u.escape(a),top:t._resetTemplate({classDisabled:o.some(function(t){return t.disabled})?"":"disabled"}),items:o.map(function(e){return t._itemTemplate({scaleId:e.scaleId,dim:u.escape(e.dim),color:e.color,cssClass:s.toClass(e.color),cssColor:e.disabled?"transparent":s.toColor(e.color),borderColor:s.toColor(e.color),classDisabled:e.disabled?"disabled":"",label:u.escape(r(e.label)?c:e.label),value:u.escape(e.value)})}).join("")}))}),t._color.length>0&&(t._updateResetButtonPosition(),t._scrollTimeout=null,t._scrollListener=function(){var e=t._container.querySelector(l);e.style.display="none",t._scrollTimeout&&clearTimeout(t._scrollTimeout),t._scrollTimeout=setTimeout(function(){t._updateResetButtonPosition(),e.style.display="",t._scrollTimeout=null},250)},t._getScrollContainer().addEventListener("scroll",t._scrollListener))},_toggleLegendItem:function(t,e){var r=this,i=this._currentFilters,a=t?Array.prototype.filter.call(t.parentNode.childNodes,function(t){return t.matches(d)}):null,s=function(t){var e=t.getAttribute("data-dim"),n=t.getAttribute("data-value");return{sid:t.getAttribute("data-scale-id"),dim:e,val:n,key:e+n}},u=function(t){return t in i},c=function(t,e){var n=s(t);if(u(n.key)===e)if(e){var a=i[n.key];delete i[n.key],t.classList.remove("disabled"),r._chart.removeFilter(a)}else{t.classList.add("disabled");var c=o(n.dim,n.val);i[n.key]=r._chart.addFilter({tag:"legend",predicate:function(t){return!c(t)}})}},l=function(e){return e===t},p=!!t&&u(s(t).key),f=function(t,e){t.querySelector(".tau-chart__legend__guide").style.backgroundColor=e?"":"transparent"};if("reset"===e)a.forEach(function(t){c(t,!0),f(t,!0)});else if("leave-others"===e)a.forEach(function(t){l(t)&&c(t,p)}),f(t,p);else if("focus-single"===e){var h=!p&&a.every(function(t){return l(t)||u(s(t).key)});a.forEach(function(t){var e=l(t)||h;c(t,e)}),p&&f(t,!0)}var m=a.filter(function(t){return!u(s(t).key)}).map(function(t){return JSON.parse(s(t).val)});n.onSelect({type:e,selectedCategories:m}),this._chart.refresh()},_highlightToggle:function(t,e){if(!t.matches(".disabled")){var n=t.getAttribute("data-dim"),r=t.getAttribute("data-value"),i=e?o(n,r):function(t){return null};this._chart.select(function(t){return!0}).forEach(function(t){t.fire("highlight",i)})}},_getScrollContainer:function(){return this._container.parentNode.parentNode},_updateResetButtonPosition:function(){this._container.querySelector(l).style.top=this._getScrollContainer().scrollTop+"px"},_generateColorMap:function(t,e){var n=e.length;return t.reduce(function(t,r,i){return t[r]=e[i%n],t},{})},_assignStaticBrewersOrEx:function(){var t=this;t._color.forEach(function(e){var n=t._chart.getSpec().scales[e],r=t._chart.getDataSources({excludeFilter:["legend"]}),i=t._chart.getScaleFactory(r).createScaleInfoByName(e).domain();if(!n.brewer||Array.isArray(n.brewer)){var o=n.brewer||u.range(20).map(function(t){return"color20-"+(1+t)});n.brewer=t._generateColorMap(i,o)}t._legendOrderState[e]=i.reduce(function(t,e,n){return t[e]=n,t},{})})}}}a.a.api.plugins.add("legend",E),e.default=E},9:function(t,e){t.exports=i}}))},"./node_modules/taucharts/dist/plugins/tooltip.js":function(t,e,n){var r,i;window,t.exports=(r=n("./node_modules/taucharts/dist/taucharts.js"),i=n("./node_modules/d3-selection/src/index.js"),function(t){var e={};function n(r){if(e[r])return e[r].exports;var i=e[r]={i:r,l:!1,exports:{}};return t[r].call(i.exports,i,i.exports,n),i.l=!0,i.exports}return n.m=t,n.c=e,n.d=function(t,e,r){n.o(t,e)||Object.defineProperty(t,e,{enumerable:!0,get:r})},n.r=function(t){"undefined"!=typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(t,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(t,"__esModule",{value:!0})},n.t=function(t,e){if(1&e&&(t=n(t)),8&e)return t;if(4&e&&"object"==typeof t&&t&&t.__esModule)return t;var r=Object.create(null);if(n.r(r),Object.defineProperty(r,"default",{enumerable:!0,value:t}),2&e&&"string"!=typeof t)for(var i in t)n.d(r,i,function(e){return t[e]}.bind(null,i));return r},n.n=function(t){var e=t&&t.__esModule?function(){return t.default}:function(){return t};return n.d(e,"a",e),e},n.o=function(t,e){return Object.prototype.hasOwnProperty.call(t,e)},n.p="",n(n.s=33)}({0:function(t,e){t.exports=r},1:function(t,e){t.exports=i},2:function(t,e,n){"use strict";n.d(e,"a",function(){return r}),n.d(e,"b",function(){return i});var r="tau-chart__tooltip";function i(t,e){return{render:function(t){return this.args=t,t=Object.assign({},t,{fields:this.filterFields(t.fields)}),this.rootTemplate(t)},rootTemplate:function(t){return['<div class="'+r+"__buttons "+r+'__clickable">',this.buttonsTemplate(),"</div>",'<div class="i-role-content '+r+'__content">',this.contentTemplate(t),"</div>"].join("\n")},contentTemplate:function(t){return this.fieldsTemplate(t)},filterFields:function(t){return t},getLabel:function(e){return t.getFieldLabel(e)},getFormatter:function(e){return t.getFieldFormat(e)},fieldsTemplate:function(t){var e=this,n=t.data;return t.fields.map(function(t){return e.itemTemplate({data:n,field:t})}).join("\n")},itemTemplate:function(t){var e=t.data,n=t.field,i=this.getLabel(n),o=this.getFormatter(n)(e[n]);return['<div class="'+r+'__list__item">','  <div class="'+r+'__list__elem">'+i+"</div>",'  <div class="'+r+'__list__elem">'+o+"</div>","</div>"].join("\n")},buttonsTemplate:function(){return[this.buttonTemplate({cls:"i-role-exclude",text:"Exclude",icon:function(){return'<span class="tau-icon-close-gray"></span>'}})].join("\n")},buttonTemplate:function(t){var e=t.icon,n=t.text,i=t.cls;return['<div class="'+r+"__button "+i+'">','  <div class="'+r+'__button__wrap">',"    "+(e?e()+" ":"")+n,"  </div>","</div>"].join("\n")},didMount:function(){var e=t.getDomNode().querySelector(".i-role-exclude");e&&e.addEventListener("click",function(){t.excludeHighlightedElement(),t.setState({highlight:null,isStuck:!1})})}}}},3:function(t,e,n){"use strict";var r=n(0),i=n.n(r),o=n(1),a=n(2),s=i.a.api.utils,u=i.a.api.domUtils,c=i.a.api.pluginsSDK,l="tau-chart__tooltip",d=function(){function t(t){this.settings=s.defaults(t||{},{align:"bottom-right",clickable:!0,clsClickable:l+"__clickable",clsStuck:"stuck",clsTarget:l+"-target",escapeHtml:!0,fields:null,formatters:{},getTemplate:null,spacing:24,winBound:12,onExclude:function(){}}),this.onRender=this._getRenderHandler()}return t.prototype.init=function(t){this._chart=t,this._tooltip=this._chart.addBalloon({spacing:this.settings.spacing,winBound:this.settings.winBound,auto:!0,effectClass:"fade"}),this._initDomEvents(),this.state={highlight:null,isStuck:!1},this.setState(this.state),this._template=this._getTemplate()},t.prototype._getTemplate=function(){var t=Object(a.b)(this,this.settings);return"function"==typeof this.settings.getTemplate?this.settings.getTemplate(t,this,this.settings):t},t.prototype._renderTemplate=function(t,e){return this._template.render({data:t,fields:e})},t.prototype._initDomEvents=function(){var t=this;this._scrollHandler=function(){t.setState({highlight:null,isStuck:!1})},window.addEventListener("scroll",this._scrollHandler,!0),this.settings.clickable&&(this._outerClickHandler=function(e){var n=Array.from(document.querySelectorAll("."+t.settings.clsClickable)).concat(t.getDomNode()).map(function(t){return t.getBoundingClientRect()}),r=Math.min.apply(Math,n.map(function(t){return t.top})),i=Math.min.apply(Math,n.map(function(t){return t.left})),o=Math.max.apply(Math,n.map(function(t){return t.right})),a=Math.max.apply(Math,n.map(function(t){return t.bottom}));(e.clientX<i||e.clientX>o||e.clientY<r||e.clientY>a)&&t.setState({highlight:null,isStuck:!1})})},t.prototype.getDomNode=function(){return this._tooltip.getElement()},t.prototype.setState=function(t){var e=this,n=this.settings,r=this.state,i=this.state=Object.assign({},r,t);r.highlight=r.highlight||{data:null,cursor:null,unit:null},i.highlight=i.highlight||{data:null,cursor:null,unit:null},i.isStuck&&r.highlight.data&&(i.highlight=r.highlight),i.highlight.data!==r.highlight.data&&(i.highlight.data?(this._hideTooltip(),this._showTooltip(i.highlight.data,i.highlight.cursor),this._setTargetSvgClass(!0),requestAnimationFrame(function(){e._setTargetSvgClass(!0)})):i.isStuck||!r.highlight.data||i.highlight.data||(this._removeFocus(),this._hideTooltip(),this._setTargetSvgClass(!1))),!i.highlight.data||r.highlight.cursor&&i.highlight.cursor.x===r.highlight.cursor.x&&i.highlight.cursor.y===r.highlight.cursor.y||(this._tooltip.position(i.highlight.cursor.x,i.highlight.cursor.y),this._tooltip.updateSize());var o=this.getDomNode();this.settings.clickable&&i.isStuck!==r.isStuck&&(i.isStuck?(window.addEventListener("click",this._outerClickHandler,!0),o.classList.add(n.clsStuck),this._setTargetEventsEnabled(!1),this._accentFocus(i.highlight.data),this._tooltip.updateSize()):(window.removeEventListener("click",this._outerClickHandler,!0),o.classList.remove(n.clsStuck),requestAnimationFrame(function(){e._setTargetEventsEnabled(!0);var t=e._chart.getSVG();t&&u.dispatchMouseEvent(t,"mouseleave")})))},t.prototype._showTooltip=function(t,e){var n=this.settings,r=n.fields||"function"==typeof n.getFields&&n.getFields(this._chart)||Object.keys(t),i=this._renderTemplate(t,r);this._tooltip.content(i).position(e.x,e.y).place(n.align).show().updateSize(),this._template.didMount&&this._template.didMount()},t.prototype._hideTooltip=function(){window.removeEventListener("click",this._outerClickHandler,!0),this._template.willUnmount&&this._template.willUnmount(),this._tooltip.hide()},t.prototype.destroy=function(){window.removeEventListener("scroll",this._scrollHandler,!0),this._setTargetSvgClass(!1),this.setState({highlight:null,isStuck:!1}),this._tooltip.destroy()},t.prototype._subscribeToHover=function(){var t=this,e=["ELEMENT.LINE","ELEMENT.AREA","ELEMENT.PATH","ELEMENT.INTERVAL","ELEMENT.INTERVAL.STACKED","ELEMENT.POINT"];this._chart.select(function(t){return e.indexOf(t.config.type)>=0}).forEach(function(e){e.on("data-hover",function(e,n){var r=document.body.getBoundingClientRect();t.setState({highlight:n.data?{data:n.data,cursor:{x:n.event.clientX-r.left,y:n.event.clientY-r.top},unit:n.unit}:null})}),t.settings.clickable&&e.on("data-click",function(e,n){var r=document.body.getBoundingClientRect();t.setState(n.data?{highlight:{data:n.data,cursor:{x:n.event.clientX-r.left,y:n.event.clientY-r.top},unit:n.unit},isStuck:!0}:{highlight:null,isStuck:null})})})},t.prototype.getFieldFormat=function(t){var e=this._formatters[t]?this._formatters[t].format:function(t){return String(t)};return this.settings.escapeHtml?function(t){return s.escapeHtml(e(t))}:e},t.prototype.getFieldLabel=function(t){var e=this._formatters[t]?this._formatters[t].label:t;return this.settings.escapeHtml?s.escapeHtml(e):e},t.prototype._accentFocus=function(t){var e=function(e){return e===t};this._chart.select(function(){return!0}).forEach(function(t){t.fire("highlight",e)})},t.prototype._removeFocus=function(){var t=function(){return null};this._chart.select(function(){return!0}).forEach(function(e){e.fire("highlight",t),e.fire("highlight-data-points",t)})},t.prototype.excludeHighlightedElement=function(){var t=this.state.highlight.data;this._chart.addFilter({tag:"exclude",predicate:function(e){return e!==t}}),this.settings.onExclude(t),this._chart.refresh()},t.prototype._getRenderHandler=function(){return function(){this._formatters=c.getFieldFormatters(this._chart.getSpec(),this.settings.formatters),this._subscribeToHover(),this.setState({highlight:null,isStuck:!1})}},t.prototype._setTargetSvgClass=function(t){o.select(this._chart.getSVG()).classed(this.settings.clsTarget,t)},t.prototype._setTargetEventsEnabled=function(t){t?this._chart.enablePointerEvents():this._chart.disablePointerEvents()},t}();e.a=d},33:function(t,e,n){"use strict";n.r(e);var r=n(0),i=n.n(r),o=n(3);function a(t){return new o.a(t)}i.a.api.plugins.add("tooltip",a),e.default=a}}))},"./node_modules/taucharts/dist/taucharts.js":function(t,e,n){var r,i,o,a,s,u,c,l,d,p,f,h,m,g;window,t.exports=(r=n("./node_modules/d3-selection/src/index.js"),i=n("./node_modules/d3-array/index.js"),o=n("./node_modules/taucharts/node_modules/d3-scale/index.js"),a=n("./node_modules/taucharts/node_modules/topojson-client/index.js"),s=n("./node_modules/d3-transition/src/index.js"),u=n("./node_modules/taucharts/node_modules/d3-format/src/index.js"),c=n("./node_modules/taucharts/node_modules/d3-time-format/src/index.js"),l=n("./node_modules/d3-brush/src/index.js"),d=n("./node_modules/taucharts/node_modules/d3-color/src/index.js"),p=n("./node_modules/d3-axis/src/index.js"),f=n("./node_modules/taucharts/node_modules/d3-geo/src/index.js"),h=n("./node_modules/taucharts/node_modules/d3-quadtree/src/index.js"),m=n("./node_modules/taucharts/node_modules/d3-shape/src/index.js"),g=n("./node_modules/taucharts/node_modules/d3-time/src/index.js"),function(t){var e={};function n(r){if(e[r])return e[r].exports;var i=e[r]={i:r,l:!1,exports:{}};return t[r].call(i.exports,i,i.exports,n),i.l=!0,i.exports}return n.m=t,n.c=e,n.d=function(t,e,r){n.o(t,e)||Object.defineProperty(t,e,{enumerable:!0,get:r})},n.r=function(t){"undefined"!=typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(t,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(t,"__esModule",{value:!0})},n.t=function(t,e){if(1&e&&(t=n(t)),8&e)return t;if(4&e&&"object"==typeof t&&t&&t.__esModule)return t;var r=Object.create(null);if(n.r(r),Object.defineProperty(r,"default",{enumerable:!0,value:t}),2&e&&"string"!=typeof t)for(var i in t)n.d(r,i,function(e){return t[e]}.bind(null,i));return r},n.n=function(t){var e=t&&t.__esModule?function(){return t.default}:function(){return t};return n.d(e,"a",e),e},n.o=function(t,e){return Object.prototype.hasOwnProperty.call(t,e)},n.p="",n(n.s=20)}([function(t,e){t.exports=r},function(t,e,n){"use strict";n.r(e);var r={};n.r(r),n.d(r,"traverseJSON",function(){return yt}),n.d(r,"traverseSpec",function(){return _t}),n.d(r,"clone",function(){return Tt}),n.d(r,"isDate",function(){return kt}),n.d(r,"isObject",function(){return At}),n.d(r,"niceZeroBased",function(){return Dt}),n.d(r,"niceTimeDomain",function(){return It}),n.d(r,"generateHash",function(){return Nt}),n.d(r,"generateRatioFunction",function(){return Ft}),n.d(r,"isSpecRectCoordsOnly",function(){return Lt}),n.d(r,"throttleLastEvent",function(){return Bt}),n.d(r,"splitEvenly",function(){return Vt}),n.d(r,"extRGBColor",function(){return zt}),n.d(r,"extCSSClass",function(){return Ht}),n.d(r,"toRadian",function(){return Ut}),n.d(r,"normalizeAngle",function(){return Wt}),n.d(r,"range",function(){return qt}),n.d(r,"hasXOverflow",function(){return Gt}),n.d(r,"hasYOverflow",function(){return Yt}),n.d(r,"flatten",function(){return Kt}),n.d(r,"unique",function(){return Xt}),n.d(r,"groupBy",function(){return Qt}),n.d(r,"union",function(){return Zt}),n.d(r,"intersection",function(){return $t}),n.d(r,"defaults",function(){return Jt}),n.d(r,"omit",function(){return te}),n.d(r,"memoize",function(){return ee}),n.d(r,"createMultiSorter",function(){return ne}),n.d(r,"pick",function(){return re}),n.d(r,"escape",function(){return ie}),n.d(r,"template",function(){return oe}),n.d(r,"escapeHtml",function(){return ae}),n.d(r,"xml",function(){return le}),n.d(r,"take",function(){return de}),n.d(r,"isChartElement",function(){return fe}),n.d(r,"isFacetUnit",function(){return he});var i={};n.r(i),n.d(i,"appendTo",function(){return ye}),n.d(i,"getScrollbarSize",function(){return _e}),n.d(i,"setScrollPadding",function(){return ve}),n.d(i,"getStyle",function(){return be}),n.d(i,"getStyleAsNum",function(){return xe}),n.d(i,"getContainerSize",function(){return we}),n.d(i,"getAxisTickLabelSize",function(){return je}),n.d(i,"getLabelSize",function(){return Oe}),n.d(i,"getCharSize",function(){return Ce}),n.d(i,"selectOrAppend",function(){return Ee}),n.d(i,"selectImmediate",function(){return Se}),n.d(i,"selectAllImmediate",function(){return Me}),n.d(i,"sortChildren",function(){return Te}),n.d(i,"classes",function(){return ke}),n.d(i,"dispatchMouseEvent",function(){return Ae});var o={};n.r(o),n.d(o,"translate",function(){return De}),n.d(o,"rotate",function(){return Ie}),n.d(o,"getOrientation",function(){return Pe}),n.d(o,"parseTransformTranslate",function(){return Re}),n.d(o,"isIntersect",function(){return Ne}),n.d(o,"getDeepTransformTranslate",function(){return Fe}),n.d(o,"raiseElements",function(){return Le});var a=n(0),s=n(2),u=n(3),c={},l={};function d(t){var e=l[t];return e||(e=function(){for(var e,n,r=this,i=0,o=[];r=r.handler;){if("function"==typeof(n=r.callbacks[t])){if(!e)for(e=[this],i=0;i<arguments.length;i++)e.push(arguments[i]);o.unshift({fn:n,context:r.context,args:e})}if("function"==typeof(n=r.callbacks["*"])){if(!e)for(e=[this],i=0;i<arguments.length;i++)e.push(arguments[i]);o.unshift({fn:n,context:r.context,args:[{sender:this,type:t,args:e}]})}}o.forEach(function(t){return t.fn.apply(t.context,t.args)})},l[t]=e),e}var p,f,h=function(){function t(){this.handler=null,this.emit_destroy=d("destroy")}return t.prototype.addHandler=function(t,e){e=e||this,this.handler={callbacks:t,context:e,handler:this.handler}},t.prototype.on=function(t,e,n){var r={};return r[t]=e,this.addHandler(r,n),r},t.prototype.fire=function(t,e){d.call(this,t).call(this,e)},t.prototype.removeHandler=function(t,e){var n,r=this;for(e=e||this;n=r,r=r.handler;)if(r.callbacks===t&&r.context===e)return r.callbacks=c,void(n.handler=r.handler)},t.prototype.destroy=function(){this.emit_destroy(),this.handler=null},t}(),m=(p=function(t,e){return(p=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}p(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),g=function(){return(g=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)},y=(g({},a),function(t){function e(e){var n=t.call(this)||this;return n.screenModel=null,n._elementNameSpace=e.namespace||"default",n._elementScalesHub={},n}return m(e,t),e.prototype.regScale=function(t,e){return this._elementScalesHub[t]=e,this},e.prototype.getScale=function(t){return this._elementScalesHub[t]||null},e.prototype.fireNameSpaceEvent=function(t,e){var n=this._elementNameSpace;this.fire(t+"."+n,e)},e.prototype.subscribe=function(t,e,n){void 0===e&&(e=function(t){return t}),void 0===n&&(n=function(t){return t});var r=this,i={};[{event:"mouseover",limit:0},{event:"mouseout",limit:0},{event:"click",limit:0},{event:"mousemove",limit:"requestAnimationFrame"}].forEach(function(o){var s=o.event,u=o.limit;t.on(s,Bt(i,s,function(t){var i={data:e.call(this,t),event:n.call(this,a.event,t)};r.fire(s,i),r.fireNameSpaceEvent(s,i)},u))})},e.prototype.allocateRect=function(){return{left:0,top:0,width:0,height:0}},e.prototype.defineGrammarModel=function(t){return{}},e.prototype.getGrammarRules=function(){return[]},e.prototype.getAdjustScalesRules=function(){return[]},e.prototype.createScreenModel=function(t){return null},e.prototype.getClosestElement=function(t,e){return null},e.prototype.addInteraction=function(){},e.prototype.draw=function(){this.config.options.container=this.config.options.slot(this.config.uid),this.drawFrames(this.config.frames)},e.prototype.data=function(){return this.config.frames.reduce(function(t,e){return t.concat(e.part())},[])},e.prototype.node=function(){return this},e}(h)),_=function(t){return{day:{cast:function(e){var n=new Date(e);return new Date(n[t.setHours](0,0,0,0))},next:function(e){var n=new Date(e),r=new Date(n[t.setDate](n[t.getDate]()+1));return this.cast(r)}},week:{cast:function(e){var n=new Date(e);return n=new Date(n[t.setHours](0,0,0,0)),new Date(n[t.setDate](n[t.getDate]()-n[t.getDay]()))},next:function(e){var n=new Date(e),r=new Date(n[t.setDate](n[t.getDate]()+7));return this.cast(r)}},month:{cast:function(e){var n=new Date(e);return n=new Date(n[t.setHours](0,0,0,0)),n=new Date(n[t.setDate](1))},next:function(e){var n=new Date(e),r=new Date(n[t.setMonth](n[t.getMonth]()+1));return this.cast(r)}},quarter:{cast:function(e){var n=new Date(e);n=new Date(n[t.setHours](0,0,0,0));var r=(n=new Date(n[t.setDate](1)))[t.getMonth](),i=r-r%3;return new Date(n[t.setMonth](i))},next:function(e){var n=new Date(e),r=new Date(n[t.setMonth](n[t.getMonth]()+3));return this.cast(r)}},year:{cast:function(e){var n=new Date(e);return n=new Date(n[t.setHours](0,0,0,0)),n=new Date(n[t.setDate](1)),n=new Date(n[t.setMonth](0))},next:function(e){var n=new Date(e),r=new Date(n[t.setFullYear](n[t.getFullYear]()+1));return this.cast(r)}}}},v=_({setHours:"setHours",setDate:"setDate",getDate:"getDate",setMonth:"setMonth",getDay:"getDay",getMonth:"getMonth",setFullYear:"setFullYear",getFullYear:"getFullYear"}),b=_({setHours:"setUTCHours",setDate:"setUTCDate",getDate:"getUTCDate",setMonth:"setUTCMonth",getDay:"getUTCDay",getMonth:"getUTCMonth",setFullYear:"setUTCFullYear",getFullYear:"getUTCFullYear"}),x={add:function(t,e,n){return((void 0===n?{utc:!1}:n).utc?b:v)[t.toLowerCase()]=e,this},get:function(t,e){return((void 0===e?{utc:!1}:e).utc?b:v)[(t||"").toLowerCase()]||null},generate:function(t,e,n,r){var i=(void 0===r?{utc:!1}:r).utc,o=[],a=x.get(n,{utc:i});if(a){var s=a.cast(new Date(e)),u=a.cast(new Date(t));for(o.push(u);(u=a.next(new Date(u)))<=s;)o.push(u)}return o}},w=(f=function(t,e){return(f=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}f(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),j=function(t){function e(e,n,r){var i=t.call(this)||this;return i.name="TauChartError",i.message=e,i.errorCode=n,i.errorArgs=r,i}return w(e,t),e}(Error),O={STACKED_FIELD_NOT_NUMBER:"STACKED_FIELD_NOT_NUMBER",NO_DATA:"NO_DATA",NOT_SUPPORTED_TYPE_CHART:"NOT_SUPPORTED_TYPE_CHART",UNKNOWN_UNIT_TYPE:"UNKNOWN_UNIT_TYPE",INVALID_LOG_DOMAIN:"INVALID_LOG_DOMAIN"},C={},E={get:function(t){return C[t]},reg:function(t,e){return C[t]=e,this}},S="taucharts_synthetic_record";E.reg("identity",function(){return{}}).reg("flip",function(t){var e=t.scaleY,n=t.scaleX,r=n.domain(),i=n.discrete?r[0]:Math.max(0,Math.min.apply(Math,r)),o=n.value(i)+-.5*n.stepSize(i);return{flip:!0,scaleX:e,scaleY:n,xi:function(t){return e.value(t[e.dim])},yi:function(t){return n.value(t[n.dim])},y0:function(){return o}}}).reg("obsoleteVerticalStackOrder",function(){return{obsoleteVerticalStackOrder:!0}}).reg("positioningByColor",function(t){return(t.scaleX.discrete?function(t){var e=t.data().reduce(function(e,n){var r=n[t.scaleX.dim],i=n[t.scaleColor.dim];return e.hasOwnProperty(r)||(e[r]=[]),e[r].indexOf(i)<0&&e[r].push(i),e},{}),n=t.scaleX,r=t.scaleColor,i=r.discrete?r.domain():r.originalSeries().sort(function(t,e){return t-e}),o=i.length||1,a=i.reduce(function(t,e,n){return t[e]=n,t},{});return Object.keys(e).forEach(function(t){return e[t].sort(function(t,e){return a[t]-a[e]})}),{xi:function(r){var i,a=r[t.scaleX.dim],s=e[a]||[r[t.scaleColor.dim]],u=s.length,c=s.indexOf(r[t.scaleColor.dim]),l=(i=r,n.stepSize(i[n.dim])*(o/(1+o))/(o+1));return t.xi(r)-(u+1)*l/2+(1+c)*l}}}:function(){return{}})(t)}).reg("groupOrderByAvg",function(t){var e=t.data(),n=e.reduce(function(e,n){var r=t.group(n);return e[r]=e[r]||[],e[r].push(n),e},{}),r=Object.keys(n).map(function(e){return[e,(r=n[e],r.map(t.yi).reduce(function(t,e){return t+e},0)/r.length)];var r}).sort(function(t,e){return t[1]-e[1]}).map(function(t){return t[0]});return{order:function(t){var e=r.indexOf(t);return e<0?Number.MAX_VALUE:e}}}).reg("stack",function(t){var e=t.data(),n=t.scaleX,r=t.scaleY;if(r.discrete||r.domain().some(function(t){return"number"!=typeof t}))throw new j("Stacked field ["+r.dim+"] should be a number",O.STACKED_FIELD_NOT_NUMBER,{field:r.dim});var i=function(t){return function(e){var i=e[n.dim],o=e[r.dim],a=o>=0?t.positive:t.negative,s=a[i]||0,u=s+o;return a[i]=u,{nextStack:u,prevStack:s}}},o=i({positive:{},negative:{}}),a=i({positive:{},negative:{}}),s=function(e){return ee(e,t.id)},u=Number.MAX_VALUE,c=Number.MIN_VALUE,l=function(t){return u=t<u?t:u,c=t>c?t:c,r.value(t)},d=s(function(t){return l(o(t).nextStack)}),p=s(function(t){return l(a(t).prevStack)}),f=function(e){return t.group(e)+"/"+(e[r.dim]>=0?1:-1)},h=Qt(e,f),m=Object.keys(h).sort(t.flip||!t.flip&&t.obsoleteVerticalStackOrder?function(e,n){return t.order(e)-t.order(n)}:function(e,n){return t.order(n)-t.order(e)}).reduce(function(t,e){return t.concat(h[e])},[]);return m.forEach(function(t){d(t),p(t)}),r.fixup(function(t){var e={};return(!t.hasOwnProperty("max")||t.max<c)&&(e.max=c),(!t.hasOwnProperty("min")||t.min>u)&&(e.min=u),e}),{group:f,data:function(){return m},yi:d,y0:p}}).reg("size_distribute_evenly",function(t,e){var n=e.minLimit,r=e.maxLimit,i=e.defMin,o=e.defMax,a=t.data(),s=function(t,e){return t-e},u=t.scaleX.discrete?t.scaleX.stepSize()/2:Number.MAX_VALUE,c=a.map(function(e){return t.xi(e)}).sort(s),l=c[0],d=c.slice(1).map(function(t){var e=t-l;return l=t,e}).filter(function(t){return t>0}).sort(s).concat(Number.MAX_VALUE)[0],p=Math.min(d,u),f="number"==typeof n?n:i,h={minSize:f,maxSize:"number"==typeof r?r:Math.max(f,Math.min(o,p))};return t.scaleSize.fixup(function(t){var e={};return t.fixed?t.maxSize>h.maxSize&&(e.maxSize=h.maxSize):(e.fixed=!0,e.minSize=h.minSize,e.maxSize=h.maxSize),e}),{}}).reg("adjustStaticSizeScale",function(t,e){var n=e.minLimit,r=e.maxLimit,i=e.defMin,o=e.defMax,a={minSize:"number"==typeof n?n:i,maxSize:"number"==typeof r?r:o};return t.scaleSize.fixup(function(t){var e={};return t.fixed||(e.fixed=!0,e.minSize=a.minSize,e.maxSize=a.maxSize),e}),{}}).reg("adjustSigmaSizeScale",function(t,e){var n,r,i,o,a=e.minLimit,s=e.maxLimit,u=e.defMin,c=e.defMax,l=function(t,e){return t-e},d=t.data().map(function(e){return t.xi(e)}).sort(l),p=d[0],f=d.slice(1).map(function(t){var e=t-p;return p=t,e}).filter(function(t){return t>0}).sort(l).concat(Number.MAX_VALUE)[0],h=t.scaleX.discrete?t.scaleX.stepSize()/2:Number.MAX_VALUE,m=Math.min(f,h),g="number"==typeof a?a:u,y="number"==typeof s?s:c,_={minSize:g,maxSize:Math.max(g,Math.min(y,(n=m,r=(g+y)/2,i=y,o=g,Math.round(r+(i-r)/(1+Math.exp(-(n-o)/.5))))))};return t.scaleSize.fixup(function(t){var e={};return t.fixed?t.maxSize>_.maxSize&&(e.maxSize=_.maxSize):(e.fixed=!0,e.minSize=_.minSize,e.maxSize=_.maxSize),e}),{}}).reg("avoidScalesOverflow",function(t,e){var n,r,i=e.sizeDirection,o=function(t,e){return!t||t.discrete||"logarithmic"===t.scaleType||i.indexOf(e)<0},a=o(t.scaleX,"x"),s=o(t.scaleY,"y");if(a&&s)return{};t.scaleSize.fixup(function(t){return n=t.minSize,r=t.maxSize,t});var u=t.data().reduce(function(e,i){var o,u,c=t.size(i),l=(c>=n?c:n+c*(r-n))/2;return a||(o=t.xi(i),e.left=Math.min(e.left,o-l),e.right=Math.max(e.right,o+l)),s||(u=t.yi(i),e.top=Math.min(e.top,u-l),e.bottom=Math.max(e.bottom,u+l)),e},{top:Number.MAX_VALUE,right:-Number.MAX_VALUE,bottom:-Number.MAX_VALUE,left:Number.MAX_VALUE}),c=function(e,n,r,i){var o=e.domain(),a=Math.abs(e.value(o[1])-e.value(o[0])),s=(o[1]-o[0])/a;if(0===a)return 1;var u=Math.max(0,-n),c=Math.max(0,r-a),l=t.flip?c:u,d=t.flip?u:c,p=Number(o[0])-(i?d:l)*s,f=Number(o[1])+(i?l:d)*s;return e.fixup(function(t){var e={};if(t.fixedBorders){var i=t.fixedBorders.slice(),o=i[0],a=i[1];(o>n||a<r)&&(e.min=Math.min(t.min,p),e.max=Math.max(t.max,f),o=Math.min(n,o),a=Math.max(r,a)),e.fixedBorders=[o,a]}else e.fixed=!0,e.min=p,e.max=f,e.nice=!1,e.fixedBorders=[n,r];return e}),a/(l+a+d)},l=a?1:c(t.scaleX,u.left,u.right,!1),d=s?1:c(t.scaleY,u.top,u.bottom,!0),p=Math.min(n*l,n*d),f=Math.min(r*l,r*d);return t.scaleSize.fixup(function(){return{minSize:p,maxSize:f}}),{}}).reg("fillGaps",function(t,e){var n,r,i,o,a,s,u,c=e.isStack,l=e.xPeriod,d=e.utc,p=Qt(t.data(),t.group),f=Object.keys(p).sort(function(e,n){return t.order(e)-t.order(n)}).reduce(function(t,e){return t.concat([p[e]])},[]),h=t.scaleX.dim,m=t.scaleY.dim,g=t.scaleColor.dim,y=t.scaleSplit.dim,_=function(t){return t[m]>=0?1:-1},v=function(e,n,r){var i=Qt(n,function(t){return t[h]}),o=n[0];return e.reduce(function(e,n){return e.concat(i[n]||(s=o,u=r,l=[a=n,t.id(s),u].join(" "),(c={})[h]=a,c[m]=1e-10*u,c[y]=s[y],c[g]=s[g],c[S]=!0,c[S+"id"]=l,c));var a,s,u,c,l},[])},b=function(t,e){return t-e},w=function(){return Xt(f.reduce(function(t,e){return t.concat(e.map(function(t){return t[h]}))},[])).sort(b)},j=l?(n=w(),r=Math.max.apply(Math,n.map(function(t){return Number(t)})),i=t.scaleX.domain(),o=x.generate(i[0],i[1],l,{utc:d}).filter(function(t){return t>=i[0]&&t<=i[1]}),a=0,s=[],u=x.get(l,{utc:d}),o.forEach(function(t){var e=Number(t);if(!(e>=r)){for(var i=a;i<n.length;i++)if(Number(u.cast(n[i]))===e)return void a++;s.push(t)}}),n.concat(s).sort(b)):w(),O=f.map(function(e){return e.sort(function(e,n){return t.xi(e)-t.xi(n)})}).reduce(c?function(t,e){var n=Qt(e,function(t){return String(_(t))});return Object.keys(n).reduce(function(t,e){return t.concat(v(j,n[e],e))},t)}:function(t,e){Qt(e,function(t){return String(_(t))});var n=Math.max.apply(Math,e.map(function(t){return t[h]}));return t.concat(v(j.filter(function(t){return t<=n}),e,0))},[]);return{data:function(){return O},size:function(e){return e[S]?t.scaleSize(0):t.size(e)},id:function(e){return e[S]?e[S+"id"]:t.id(e)}}});var M=n(5);function T(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];return 2===e.length?e[0]*(1-t)+e[1]*t:3===e.length?e[0]*(1-t)*(1-t)+2*e[1]*(1-t)*t+e[2]*t*t:e[0]*(1-t)*(1-t)*(1-t)+3*e[1]*(1-t)*(1-t)*t+3*e[2]*(1-t)*t*t+e[3]*t*t*t}function k(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];var r=e.map(function(t){return t.x}),i=e.map(function(t){return t.y});return{x:T.apply(void 0,[t].concat(r)),y:T.apply(void 0,[t].concat(i))}}function A(t,e,n,r,i){var o=k(t,e,n),a=k(t,e,n,r),s=k(t,n,r,i),u=k(t,r,i);return[e,o,a,k(t,a,s),s,u,i]}function D(t){var e=t.polylineFrom,n=t.polylineTo,r=t.isCubic,i=t.toOppositeScale,o=e.length>n.length?e:n,a=1===n.length,s=e[0].id!==n[0].id,u=Boolean(s!==a);return function(t){var e=(r?z:V)({t:t,polyline:o,decreasing:a,rightToLeft:u});a===u?e.shift():e.pop();var n=B(e.map(i),e,a?1-t:t);return n.forEach(function(t){return t.positionIsBeingChanged=!0}),n}}function I(t){var e=t.polylineFrom,n=t.polylineTo,r=t.isCubic,i=e.length,o=n.length;if(o!==i){var a=o<i,s=a?e:n,u=(r?function(t){var e=t.smallerPolyline,n=t.biggerPolyline,r=t.decreasing,i=(e.length-1)/3,o=(n.length-1)/3,a=Math.floor(o/i)+1,s=o%i,u=qt(i).map(function(t){return a+Number(t<s)}),c=[e[0]],l=3;return u.forEach(function(t){if(t>2){var i=function(t,e){for(var n,r,i=[e[0]],o=0;o<t.length;o++)n=0===o?t[0]:t[o]/(1-t[o-1]),r=W(n,e),N(i,r.slice(1,4)),e=r.slice(3);return N(i,e.slice(1)),i}(qt(1,t-1).map(function(e){return e/(t-1)}),e.slice(l-3,l+1));qt(t-2).forEach(function(t){return i[3*(t+1)].id=n[c.length-1+3*t].id}),r&&i.forEach(function(t,e){e>0&&e<i.length-1&&(t.isInterpolated=!0)}),N(c,i.slice(1))}else{var o=Object.assign({},e[l-2]),a=Object.assign({},e[l-1]),s=Object.assign({},e[l]);r||(s.id=n[c.length+2].id),c.push(o,a,s)}l+=3}),c}:function(t){var e=t.smallerPolyline,n=t.biggerPolyline,r=t.decreasing,i=e.length-1,o=n.length-1,a=Math.floor(o/i)+1,s=o%i,u=qt(i).map(function(t){return a+Number(t<s)}),c=[e[0]],l=1;return u.forEach(function(t){qt(1,t).forEach(function(i){var o;i===t-1?(o=Object.assign({},e[l]),r||(o.id=n[c.length].id)):((o=L(e[l-1],e[l],i/(t-1))).id=n[c.length].id,r&&(o.isInterpolated=!0)),c.push(o)}),l++}),c})({smallerPolyline:a?n:e,biggerPolyline:s,decreasing:a}),c=s.slice(1,s.length-1),l=u.slice(1,u.length-1);return function(t){var e=B(l,c,a?1-t:t);return e.forEach(function(t){return t.positionIsBeingChanged=!0}),e}}var d=e.slice(1,e.length-1),p=n.slice(1,n.length-1);return function(t){var e=B(d,p,t);return e.forEach(function(t){return t.positionIsBeingChanged=!0}),e}}function P(t){var e=t.pointFrom,n=t.pointTo;return function(t){return[L(e,n,t)]}}function R(t){var e=t.polylineFrom,n=t.polylineTo,r=t.isCubic,i=0===n.length,o=i,a=i?e:n;return function(t){var e=(r?z:V)({t:t,polyline:a,decreasing:i,rightToLeft:o});return e.forEach(function(t,e){e>0&&(t.positionIsBeingChanged=!0)}),e}}function N(t,e){return Array.prototype.push.apply(t,e)}function F(t,e,n){return void 0===e?t:"number"==typeof e?t+n*(e-t):e}function L(t,e,n){if(t===e)return e;var r={};return Object.keys(t).forEach(function(i){return r[i]=F(t[i],e[i],n)}),void 0!==e.id&&(r.id=e.id),r}function B(t,e,n){return t.map(function(t,r){return L(t,e[r],n)})}function V(t){var e=t.t,n=t.polyline,r=t.decreasing,i=t.rightToLeft,o=Boolean(r)!==Boolean(i),a=function(t,e){var n=0;if(t>0){for(var r,i,o,a,s=[0],u=0,c=1;c<e.length;c++)o=e[c-1].x,a=e[c-1].y,r=e[c].x,i=e[c].y,u+=Math.sqrt((r-o)*(r-o)+(i-a)*(i-a)),s.push(u);var l=t*u;for(c=1;c<s.length;c++)if(l<=s[c]){n=Math.min(1,(c-1+(l-s[c-1])/(s[c]-s[c-1]))/(e.length-1));break}}var d=Math.floor((e.length-1)*n)+1,p=e.length-d,f=d,h=e.slice(0,d);if(n<1){var m=n*(e.length-1)%1,g=L(e[d-1],e[d],m);N(h,qt(p).map(function(t){return Object.assign({},g,{id:e[f+t].id,isInterpolated:!0})}))}return h}(r?1-e:e,o?n.slice(0).reverse():n);return o&&a.reverse(),a}function z(t){var e=t.t,n=t.polyline,r=t.decreasing,i=t.rightToLeft,o=Boolean(r)!==Boolean(i),a=function(t,e){var n=(e.length-1)/3+1,r=0;if(t>0){for(var i,o,a,s,u,c,l,d,p=[0],f=0,h=1;h<n;h++)a=e[3*h-3].x,s=e[3*h-3].y,u=e[3*h-2].x,c=e[3*h-2].y,l=e[3*h-1].x,d=e[3*h-1].y,i=e[3*h].x,o=e[3*h].y,f+=(U(a,s,u,c)+U(u,c,l,d)+U(l,d,i,o)+U(i,o,a,s))/2,p.push(f);var m=t*f;for(h=1;h<p.length;h++)if(m<=p[h]){r=Math.min(1,(h-1+(m-p[h-1])/(p[h]-p[h-1]))/(n-1));break}}var g=Math.floor((n-1)*r)+1,y=n-g,_=3*g,v=e.slice(0,3*(g-1)+1);if(r<1){var b=W(r*(n-1)%1,e.slice(3*(g-1),3*g+1)).slice(1,4);b.forEach(function(t){return t.isInterpolated=!0}),b[2].id=e[_].id,N(v,b),qt(1,y).forEach(function(t){N(v,[{x:b[2].x,y:b[2].y,isCubicControl:!0,isInterpolated:!0},{x:b[2].x,y:b[2].y,isCubicControl:!0,isInterpolated:!0},Object.assign({},b[2],{id:e[_+3*t].id,isInterpolated:!0})])})}return v}(r?1-e:e,o?n.slice(0).reverse():n);return o&&a.reverse(),a}function H(t,e){var n,r,i,o,a=[],s=[],u=0,c=t.length,l=e.length;for(n=0;n<c;n++)for(i=t[n],r=u;r<l;r++)if(o=e[r],i.id===o.id){u=r+1,a.push(i),s.push(o);break}if(a.length<1||s.length<1)return function(t){return t};var d=Object.keys(a[0]).filter(function(t){return"number"==typeof a[0][t]}).filter(function(t){return"id"!==t}),p={},f=function(t,e,n,r){return function(i){return r+(i-e)*(r-n)/(e-t)}};return d.forEach(function(t){for(var e,n,r=a[0][t],i=s[0][t],o=a.length-1;o>0;o--)if((e=a[o][t])!==r)return n=s[o][t],void(p[t]=f(r,e,i,n));var u,c;p[t]=(u=r,c=i,function(t){return t-u+c})}),function(t){var e=Object.assign({},t);return d.forEach(function(n){e[n]=p[n](t[n])}),e}}function U(t,e,n,r){return Math.sqrt((n-t)*(n-t)+(r-e)*(r-e))}function W(t,e){var n=e[0],r=e[1],i=e[2],o=e[3],a=A(t,n,r,i,o);return[a[1],a[2],a[4],a[5]].forEach(function(t){return t.isCubicControl=!0}),Object.keys(o).forEach(function(e){"x"!==e&&"y"!==e&&"id"!==e&&(a[3][e]=F(n[e],o[e],t))}),a}function q(t,e){if(t.length<2)return t.slice(0);if(2===t.length)return[t[0],{x:G(t[0].x,t[1].x,1/3),y:G(t[0].y,t[1].y,1/3)},{x:G(t[0].x,t[1].x,2/3),y:G(t[0].y,t[1].y,2/3)},t[1]];for(var n,r,i,o,a,s,u,c,l,d,p,f,h,m,g=new Array(3*(t.length-1)+1),y=g.length-1,_=0;_<t.length;_++)g[3*_]=t[_],_>0&&(g[3*_-2]=k(1/3,t[_-1],t[_]),g[3*_-1]=k(2/3,t[_-1],t[_]));for(var v=g.slice(0),b=0;b<3;b++){for(g[1]={x:G(g[0].x,g[3].x,1/3),y:G(g[0].y,G(g[3].y,g[2].y,1.5),2/3)},g[y-1]={x:G(g[y].x,g[y-3].x,1/3),y:G(g[y].y,G(g[y-3].y,g[y-2].y,1.5),2/3)},e&&((g[1].y-g[0].y)*(g[3].y-g[2].y)<0&&(g[1]={x:g[1].x,y:g[0].y}),(g[y-1].y-g[y].y)*(g[y-3].y-g[y-2].y)<0&&(g[y-1]={x:g[y-1].x,y:g[y].y})),_=6;_<v.length;_+=3)n=v[_-5],r=v[_-3],i=v[_-1],(r.x-n.x)*(i.x-r.x)*1e12<1?(o=G(n.x,r.x,.5),s=G(r.x,i.x,.5),a=G(n.y,r.y,.5),u=G(r.y,i.y,.5)):(d=(r.x-n.x)/(i.x-n.x),c=(r.x-n.x*(1-d)*(1-d)-i.x*d*d)/(2*(1-d)*d),l=(r.y-n.y*(1-d)*(1-d)-i.y*d*d)/(2*(1-d)*d),o=G(n.x,c,d),s=G(c,i.x,d),a=G(n.y,l,d),u=G(l,i.y,d),e&&(f=r.x-o,h=s-r.x,p=(u-r.y)/h,(r.y-n.y)*(i.y-r.y)<=0?p=0:(r.y>n.y==u>i.y&&(h=G(h*(m=(i.y-r.y)/(u-r.y)),h,1/(1+Math.abs(m))),p=(i.y-r.y)/h),r.y>n.y==a<n.y&&(f=G(f*(m=(r.y-n.y)/(r.y-a)),f,1/(1+Math.abs(m))),p=(r.y-n.y)/f)),o=r.x-f,s=r.x+h,a=r.y-p*f,u=r.y+p*h)),g[_-4]={x:o,y:a},g[_-2]={x:s,y:u};v=g.slice(0)}return v}function G(t,e,n){return t+n*(e-t)}var Y={linear:function(t){return t},step:function(t){for(var e,n,r,i,o=[],a=void 0!==t[0].id,s=void 0!==t[0].size,u=1;u<t.length;u++)e=t[u-1],n=t[u],r={x:(e.x+n.x)/2,y:e.y},i={x:(e.x+n.x)/2,y:n.y},a&&(r.id=e.id+"-"+n.id+"-1",i.id=e.id+"-"+n.id+"-2"),s&&(r.size=e.size,i.size=n.size),1===u&&o.push(e),o.push(r,i,n);return o},"step-before":function(t){for(var e,n,r,i=[],o=void 0!==t[0].id,a=void 0!==t[0].size,s=1;s<t.length;s++)e=t[s-1],n=t[s],r={x:e.x,y:n.y},o&&(r.id=e.id+"-"+n.id),a&&(r.size=n.size),1===s&&i.push(e),i.push(r,n);return i},"step-after":function(t){for(var e,n,r,i=[],o=void 0!==t[0].id,a=void 0!==t[0].size,s=1;s<t.length;s++)e=t[s-1],r={x:(n=t[s]).x,y:e.y},o&&(r.id=e.id+"-"+n.id),a&&(r.size=e.size),1===s&&i.push(e),i.push(r,n);return i}},K={smooth:function(t){return q(t,!1)},"smooth-keep-extremum":function(t){return q(t,!0)}};function X(t){return Y[t]||K[t]}function Q(t){return void 0!==K[t]?"cubic":"polyline"}var Z,$=function(){return($=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)},J=$({},a,M),tt=function(){return ee(function(t){return t.node().getComputedTextLength()},function(t){return t.node().textContent.length})},et=function(t,e,n){n=n||tt(),t.each(function(){var t=J.select(this.parentNode).data()[0],r=e(t),i=J.select(this),o=i.text().split(/\s+/),a=!1,s=o.reduce(function(t,e,o){if(a)return t;var s=o>0?[t,e].join(" "):e,u=n(i.text(s));if(u<r)t=s;else{var c=Math.floor(r/u*s.length);t=s.substr(0,c-4)+"...",a=!0}return t},"");i.text(s)})},nt=function(t,e,n,r,i,o){o=o||tt();var a=function(t,e,n,r,i,o,a){var s=a*n+o;return t.append("tspan").attr("x",r).attr("y",i).attr("dy",s+"em").text(e)};t.each(function(){var t=J.select(this.parentNode).data()[0],s=e(t),u=J.select(this),c=u.text().split(/\s+/),l=u.attr("x"),d=u.attr("y"),p=parseFloat(u.attr("dy"));u.text(null);var f=a(u,null,1.1,l,d,p,0),h=!1,m=c.length-1,g=c.reduce(function(t,e,r){if(h)return t;var i=t.length===n||r===m,a=t[t.length-1],u=""!==a?a+" "+e:e,c=o(f.text(u)),l=c>s;if(l&&i){var d=Math.floor(s/c*u.length);t[t.length-1]=u.substr(0,d-4)+"...",h=!0}return l&&!i&&t.push(e),l||(t[t.length-1]=u),t},[""]).filter(function(t){return t.length>0});d=i?-1*(g.length-1)*Math.floor(.5*r):d,g.forEach(function(t,e){return a(u,t,1.1,l,d,p,e)}),f.remove()})},rt=function(t,e,n){return e>0&&!document.hidden&&((t=t.transition(n).duration(e)).attr=it),t.onTransitionEnd=function(t){return st(this,t),this},t},it=function(t,e){var n,r,i=this,o=J.transition.prototype.attr.apply(this,arguments);if(0===arguments.length)throw new Error("Unexpected `transition().attr()` arguments.");1===arguments.length?r=t:arguments.length>1&&((n={})[t]=e,r=n);var a="__transitionAttrs__",s="__lastTransitions__",u=at();this.each(function(){var t=this,e={};for(var n in r)"function"==typeof r[n]?e[n]=r[n].apply(this,arguments):e[n]=r[n];this[a]=Object.assign(this[a]||{},e),this[a][s]||Object.defineProperty(this[a],s,{value:{}}),Object.keys(e).forEach(function(e){return t[a][s][e]=u})});var c=function(){var t=this;this[a]&&(Object.keys(r).filter(function(e){return t[a][s][e]===u}).forEach(function(e){return delete t[a][e]}),0===Object.keys(this[a]).length&&delete this[a])};return this.on("interrupt."+u,function(){return i.each(c)}),this.on("end."+u,function(){return i.each(c)}),o},ot=0,at=function(){return++ot},st=function(t,e){if(J.transition.prototype.isPrototypeOf(t)&&!t.empty()){var n=function(){return e.call(null,t)};return t.on("interrupt.d3_on_transition_end",n),t.on("end.d3_on_transition_end",n),t}e.call(null,t)},ut=function(t,e,n,r){var i=r||function(t){return t},o=function(){i(this)};return function(r){var i=r;return e&&(i=i.call(dt(Jt(e,n)))),i=(i=rt(i,t)).call(dt(n)),t>0?i.on("end.d3_animationInterceptor",function(){return i.each(o)}):i.each(o),i}},ct=function(t,e){var n=t.node();return t.selectAll(e).filter(function(){return this.parentNode===n})},lt=function(t,e,n,r,i){void 0===i&&(i="linear");var o="__pathPoints__";return function(t){var a=this;this[o]||(this[o]=n.map(function(){return[]}));var s=n.map(function(e,n){var s,u,c,l,d=Xt(t,r).map(e),p=(X(i)||X("linear"))(d),f=a[o][n];return{pointsFrom:f,pointsTo:p,interpolate:(s=f,u=p,c=Q(i),void 0===c&&(c="polyline"),function(t){return 0===t?s:1===t?u:(l||(l=("cubic"===c?function(t,e){for(var n=2;n<t.length-1;n+=3)t[n-1].isCubicControl=!0,t[n].isCubicControl=!0;for(n=2;n<e.length-1;n+=3)e[n-1].isCubicControl=!0,e[n].isCubicControl=!0;var r,i;for(t=t.filter(function(t){return!t.isInterpolated}),n=t.length-2;n>=0;n--)i=t[n+1],(r=t[n]).isCubicControl||i.isCubicControl||(t.splice(n+1,0,k(1/3,i,r),k(2/3,i,r)),t[n+1].isCubicControl=!0,t[n+2].isCubicControl=!0);var o=t.filter(function(t,e){return e%3==0}),a=e.filter(function(t,e){return e%3==0}),s=o.map(function(t){return t.id}),u=a.map(function(t){return t.id}),c=s.reduce(function(e,n){return e[n]=t.findIndex(function(t){return t.id===n}),e},{}),l=u.reduce(function(t,n){return t[n]=e.findIndex(function(t){return t.id===n}),t},{}),d=s.filter(function(t){return u.indexOf(t)>=0}),p=o.filter(function(t){return!t.positionIsBeingChanged}),f=a.filter(function(t){return!t.positionIsBeingChanged}),h=H(p,f),m=H(f,p),g=[];return d.forEach(function(n,r){var i,o,a,s=c[n],u=l[n];if(0===r&&(s>0||u>0)&&g.push(D({polylineFrom:t.slice(0,s+1),polylineTo:e.slice(0,u+1),toOppositeScale:0===u?h:m,isCubic:!0})),r>0){var p=c[d[r-1]],f=l[d[r-1]];s-p>3||u-f>3?g.push(I({polylineFrom:t.slice(p,s+1),polylineTo:e.slice(f,u+1),isCubic:!0})):g.push((i={polylineFrom:t.slice(p,s+1),polylineTo:e.slice(f,u+1)},o=i.polylineFrom,a=i.polylineTo,function(t){return B(o.slice(1,3),a.slice(1,3),t)}))}g.push(P({pointFrom:t[s],pointTo:e[u]})),r===d.length-1&&(t.length-s-1>0||e.length-u-1>0)&&g.push(D({polylineFrom:t.slice(s),polylineTo:e.slice(u),toOppositeScale:e.length-u==1?h:m,isCubic:!0}))}),0===g.length&&(e.length>0&&0===d.length||t.length>0&&0===d.length)&&g.push(R({polylineFrom:t.slice(0),polylineTo:e.slice(0),isCubic:!0})),function(t){var e=[];return g.forEach(function(n){var r=n(t);N(e,r)}),e}}:function(t,e){var n=(t=t.filter(function(t){return!t.isInterpolated})).map(function(t){return t.id}),r=e.map(function(t){return t.id}),i=n.filter(function(t){return r.indexOf(t)>=0}),o=t.filter(function(t){return!t.positionIsBeingChanged}),a=e.filter(function(t){return!t.positionIsBeingChanged}),s=H(o,a),u=H(a,o),c=[];return i.forEach(function(o,a){var l=n.indexOf(o),d=r.indexOf(o);if(0===a&&(l>0||d>0)&&c.push(D({isCubic:!1,polylineFrom:t.slice(0,l+1),polylineTo:e.slice(0,d+1),toOppositeScale:0===d?s:u})),a>0){var p=n.indexOf(i[a-1]),f=r.indexOf(i[a-1]);(l-p>1||d-f>1)&&c.push(I({isCubic:!1,polylineFrom:t.slice(p,l+1),polylineTo:e.slice(f,d+1)}))}c.push(P({pointFrom:t[l],pointTo:e[d]})),a===i.length-1&&(t.length-l-1>0||e.length-d-1>0)&&c.push(D({isCubic:!1,polylineFrom:t.slice(l),polylineTo:e.slice(d),toOppositeScale:e.length-d==1?s:u}))}),0===c.length&&(e.length>0&&0===i.length||t.length>0&&0===i.length)&&c.push(R({isCubic:!1,polylineFrom:t.slice(0),polylineTo:e.slice(0)})),function(t){var e=[];return c.forEach(function(n){var r=n(t);N(e,r)}),e}})(s,u)),l(t))})}});return function(t){if(0===t){var n=s.map(function(t){return t.pointsFrom});return e.apply(void 0,n)}if(1===t){var r=s.map(function(t){return t.pointsTo});return a[o]=r,e.apply(void 0,r)}var i=s.map(function(e){return e.interpolate(t)});return a[o]=i,e.apply(void 0,i)}}},dt=function(t){return function(e){return Object.keys(t).forEach(function(n){return e.attr(n,t[n])}),e}},pt=function(t){return function(e){return Object.keys(t).forEach(function(n){return e.classed(n,t[n])}),e}},ft=(Z=function(t,e){return(Z=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}Z(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),ht=function(t){function e(e){var n=t.call(this,e)||this;n.config=e,n.config.guide=Jt(n.config.guide||{},{animationSpeed:0,enableColorToBarPosition:!1}),n.config.guide.size=n.config.guide.size||{};var r=n.config.stack,i=n.config.guide.enableColorToBarPosition,o=[e.flip&&E.get("flip"),r&&E.get("stack"),i&&E.get("positioningByColor")];return n.decorators=(n.config.transformRules||o).concat(e.transformModel||[]),n.adjusters=(n.config.adjustRules||[]).concat(e.adjustScales||[]),n}return ft(e,t),e.prototype.defineGrammarModel=function(t){var e=this,n=this.config;this.regScale("x",t("pos",n.x,[0,n.options.width])).regScale("y",t("pos",n.y,[n.options.height,0])).regScale("y",t("pos",n.y,function(t){return["ordinal","period"].indexOf(t.type)>=0?[0,n.options.height]:[n.options.height,0]})).regScale("size",t("size",n.size,{})).regScale("color",t("color",n.color,{})).regScale("split",t("split",n.split,{})).regScale("label",t("label",n.label,{})).regScale("identity",t("identity",n.identity,{}));var r=this.getScale("x"),i=this.getScale("y"),o=this.getScale("size"),a=this.getScale("label"),s=this.getScale("color"),u=this.getScale("split"),c=this.getScale("identity"),l=i.domain(),d=i.discrete?l[0]:Math.max(0,Math.min.apply(Math,l)),p=i.value(d)+.5*i.stepSize(d),f=s.domain(),h={data:function(){return e.data()},flip:!1,scaleX:r,scaleY:i,scaleSize:o,scaleLabel:a,scaleColor:s,scaleSplit:u,scaleIdentity:c,color:function(t){return s.value(t[s.dim])},label:function(t){return a.value(t[a.dim])},group:function(t){return t[s.dim]+"(@taucharts@)"+t[u.dim]},order:function(t){var e=t.split("(@taucharts@)")[0],n=f.indexOf(e);return n<0?Number.MAX_VALUE:n},size:function(t){return o.value(t[o.dim])},id:function(t){return c.value(t[c.dim],t)},xi:function(t){return r.value(t[r.dim])},yi:function(t){return i.value(t[i.dim])},y0:function(){return p}};return h.data().forEach(function(t){return h.id(t)}),h},e.prototype.getGrammarRules=function(){return this.decorators.filter(function(t){return t})},e.prototype.getAdjustScalesRules=function(){return(this.adjusters||[]).filter(function(t){return t})},e.prototype.createScreenModel=function(t){var e=t.flip,n=function(t,e,n){return t?e:n};return{flip:e,id:t.id,x:n(e,t.yi,t.xi),y:n(e,t.xi,t.yi),x0:n(e,t.y0,t.xi),y0:n(e,t.xi,t.y0),size:t.size,group:t.group,order:t.order,label:t.label,color:function(e){return t.scaleColor.toColor(t.color(e))},class:function(e){return t.scaleColor.toClass(t.color(e))},model:t,toFibers:function(){var e=Qt(t.data(),t.group);return Object.keys(e).sort(function(e,n){return t.order(e)-t.order(n)}).reduce(function(t,n){return t.concat([e[n]])},[])}}},e.prototype.drawFrames=function(){var t=this,e=this.config.options,n=function(e){return n=t.screenModel.size(e)/2,r=Math.pow(10,4),Math.round(r*n)/r;var n,r},r=ut,i=function(e,n,i){var o=t.config.guide.animationSpeed,s=e.selectAll("."+n).data(function(t){return[t]},t.screenModel.id);s.exit().call(r(o,null,{width:0},function(t){return a.select(t).remove()})),s.call(r(o,null,i)),s.enter().append("rect").style("stroke-width",0).call(r(o,{width:0},i))},o=this.config.flip,s=o?"y":"x",u=o?"x":"y",c=o?"x0":"y0",l=o?"height":"width",d=o?"width":"height",p=function(e){var r,o,a;i(e,"lvl-top",((r={})[l]=function(t){return n(t)},r[d]=1,r[s]=function(e){return t.screenModel[s](e)-n(e)/2},r[u]=function(e){return t.screenModel[u](e)},r.fill=function(e){return t.screenModel.color(e)},r.class=function(e){return"lvl-top "+t.screenModel.class(e)},r)),i(e,"lvl-btm",((o={})[l]=function(t){return n(t)},o[d]=1,o[s]=function(e){return t.screenModel[s](e)-n(e)/2},o[u]=function(e){return t.screenModel[c](e)},o.fill=function(e){return t.screenModel.color(e)},o.class=function(e){return"lvl-btm "+t.screenModel.class(e)},o)),i(e,"lvl-link",((a={})[l]=.5,a[d]=function(e){return Math.abs(t.screenModel[u](e)-t.screenModel[c](e))},a[s]=function(e){return t.screenModel[s](e)-.25},a[u]=function(e){return Math.min(t.screenModel[u](e),t.screenModel[c](e))},a.fill=function(e){return t.screenModel.color(e)},a.class=function(e){return"lvl-link "+t.screenModel.class(e)},a))},f=function(e){e.attr("class","frame-id-"+t.config.uid).call(function(e){var n=e.selectAll(".generic").data(function(t){return t},t.screenModel.id);n.exit().remove(),n.call(p),n.enter().append("g").attr("class","generic").call(p)})},h=Qt(this.data(),t.screenModel.group),m=Object.keys(h).sort(function(e,n){return t.screenModel.order(e)-t.screenModel.order(n)}).reduce(function(t,e){return t.concat([h[e]])},[]),g=e.container.selectAll(".frame-id-"+t.config.uid).data(m);g.exit().remove(),g.call(f),g.enter().append("g").call(f)},e}(y),mt=function(){return(mt=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)},gt=mt({},s,u);function yt(t,e,n,r){var i=r(n(t),t);return(i[e]||[]).forEach(function(t){return yt(t,e,n,r)}),i}function _t(t,e,n,r){void 0===r&&(r=0),e(t,r)&&(t.units||[]).map(function(t){return _t(t,e,n,r+1)}),n(t,r)}var vt=function(){var t=[];function e(t){for(var e in t)this[e]=t[e]}function n(){this.copiedObjects=[];var t=this;this.recursiveDeepCopy=function(e){return t.deepCopy(e)},this.depth=0}function r(t,e){var r=new n;return e&&(r.maxDepth=e),r.deepCopy(t)}return e.prototype={constructor:e,canCopy:function(t){return!1},create:function(t){},populate:function(t,e,n){}},n.prototype={constructor:n,maxDepth:256,cacheResult:function(t,e){this.copiedObjects.push([t,e])},getCachedResult:function(t){for(var e=this.copiedObjects,n=e.length,r=0;r<n;r++)if(e[r][0]===t)return e[r][1]},deepCopy:function(e){if(null===e)return null;if("object"!=typeof e)return e;var n=this.getCachedResult(e);if(n)return n;for(var r=0;r<t.length;r++){var i=t[r];if(i.canCopy(e))return this.applyDeepCopier(i,e)}throw new Error("no DeepCopier is able to copy "+e)},applyDeepCopier:function(t,e){var n=t.create(e);if(this.cacheResult(e,n),this.depth++,this.depth>this.maxDepth)throw new Error("Exceeded max recursion depth in deep copy.");return t.populate(this.recursiveDeepCopy,e,n),this.depth--,n}},r.DeepCopier=e,r.deepCopiers=t,r.register=function(n){n instanceof e||(n=new e(n)),t.unshift(n)},r.register({canCopy:function(){return!0},create:function(t){return t instanceof t.constructor?"object"==typeof(e=t.constructor.prototype)?JSON.parse(JSON.stringify(e)):e:{};var e},populate:function(t,e,n){for(var r in e)e.hasOwnProperty(r)&&(n[r]=t(e[r]));return n}}),r.register({canCopy:function(t){return t instanceof Array},create:function(t){return new t.constructor},populate:function(t,e,n){for(var r=0;r<e.length;r++)n.push(t(e[r]));return n}}),r.register({canCopy:function(t){return t instanceof Date},create:function(t){return new Date(t)}}),r}(),bt=function(t){return/^(#|rgb\(|rgba\()/.test(t)},xt=/(.)^/,wt={"&":"&amp;","<":"&lt;",">":"&gt;",'"':"&quot;","'":"&#x27;","`":"&#x60;"},jt={"'":"'","\\":"\\","\r":"r","\n":"n","\u2028":"u2028","\u2029":"u2029"},Ot=/\\|'|\r|\n|\u2028|\u2029/g,Ct="(?:"+Object.keys(wt).join("|")+")",Et=RegExp(Ct),St=RegExp(Ct,"g"),Mt={evaluate:/<%([\s\S]+?)%>/g,interpolate:/<%=([\s\S]+?)%>/g,escape:/<%-([\s\S]+?)%>/g};function Tt(t){return vt(t)}function kt(t){return t instanceof Date&&!isNaN(Number(t))}function At(t){return null!=t&&"object"==typeof t}function Dt(t){var e=parseFloat(Math.min.apply(Math,t).toFixed(15)),n=parseFloat(Math.max.apply(Math,t).toFixed(15));e===n&&(n-=(n>=0?-1:1)*(n||1)/10);for(var r=[e=Math.min(0,e),n=Math.max(0,n)],i=r[1]-r[0],o=Math.pow(10,Math.floor(Math.log(i/10)/Math.LN10)),a=10/i*o,s=[[.15,10],[.35,5],[.75,2],[1,1],[2,1]],u=-1;a>s[++u][0];);o*=s[u][1],r[0]=Math.floor(r[0]/o)*o,r[1]=Math.ceil(r[1]/o)*o;var c=e-r[0],l=r[1]-n,d=o/2;if(e<0){var p=c>=d?-c:0;r[0]=r[0]-p}if(n>0){var f=l>=d?-l:0;r[1]=r[1]+f}return[parseFloat(r[0].toFixed(15)),parseFloat(r[1].toFixed(15))]}function It(t,e,n){var r=(void 0===n?{utc:!1}:n).utc,i=gt.extent(t),o=i[0],a=i[1],s=+a-+o,u=r?gt.scaleUtc:gt.scaleTime;if(0===s)return o=new Date(o.getTime()-864e5),a=new Date(a.getTime()+864e5),u().domain([o,a]).nice(e).domain();var c=u().domain([o,a]).nice(e);if(e)return c.domain();var l=u().domain([o,a]).nice(e).domain(),d=l[0],p=l[1],f=c.ticks(),h=f.length-1;return(+o-+d)/(+f[1]-+d)<.5&&(o=d),(+p-+a)/(+p-+f[h-1])<.5&&(a=p),[o,a]}var Pt=0,Rt={};function Nt(t){var e=btoa(encodeURIComponent(t)).replace(/=/g,"_");return Rt.hasOwnProperty(e)||(Rt[e]="H"+ ++Pt),Rt[e]}function Ft(t,e,n){var r=0,i=null,o=ee(function(t,e){return Xt(t.map(function(t){return e.reduce(function(e,n){return e.concat(kt(r=t[n])?r.getTime():r);var r},[])}),function(t){return JSON.stringify(t)}).reduce(function(t,e){var n=e[0];return t[n]=t[n]||0,t[n]+=1,t},{})},function(t,e){var n=i===t?r:++r;return i=t,e.join("")+"-"+n});return function(r,i,a){var s=a.length,u=n.getSpec(),c=u.sources["/"].data,l=u.unit.units[0].guide||{};l.padding=l.padding||{l:0,r:0,t:0,b:0};var d=0;"x"===t?d=l.padding.l+l.padding.r:"y"===t&&(d=l.padding.t+l.padding.b);var p,f,h=(i-s*d)/(p=o(c,e),Object.keys(p).reduce(function(t,e){return t+p[e]},0));return f=r,(o(c,e)[f]*h+d)/i}}function Lt(t){var e=!0;try{_t(t,function(t){if(0===t.type.indexOf("COORDS.")&&"COORDS.RECT"!==t.type)throw new Error("Not applicable")},function(t){return t})}catch(t){"Not applicable"===t.message&&(e=!1)}return e}function Bt(t,e,n,r){if(void 0===r&&(r=0),"requestAnimationFrame"===r){var i=!1;return function(){for(var r=[],o=0;o<arguments.length;o++)r[o]=arguments[o];i||(requestAnimationFrame(function(){i=!1}),n.apply(this,r),i=!0),t.e=e,t.ts=Date.now()}}return function(){for(var i=[],o=0;o<arguments.length;o++)i[o]=arguments[o];var a={e:e,ts:Date.now()};(t.e&&t.e===a.e?a.ts-t.ts:r)>=r&&n.apply(this,i),t.e=a.e,t.ts=a.ts}}function Vt(t,e){var n=t[0],r=t[1],i=(r-n)/(e-1),o=e>=2?qt(e-2).map(function(t){return n+i*(t+1)}):[];return[n].concat(o,[r])}function zt(t){return bt(t)?t:""}function Ht(t){return bt(t)?"":t}function Ut(t){return t/180*Math.PI}function Wt(t){return Math.abs(t)>=360&&(t%=360),t<0&&(t=360+t),t}function qt(t,e){1===arguments.length&&(e=t,t=0);for(var n=[],r=t;r<e;r++)n.push(r);return n}function Gt(t,e,n,r){return t+e*(90===n?-1:1)-r>100}function Yt(t,e,n,r){return t+e*(-90===n?-1:1)-r>20}function Kt(t){return Array.isArray(t)?[].concat.apply([],t.map(function(t){return Kt(t)})):t}function Xt(t,e){for(var n={},r=[],i=t.length,o=e||function(t){return String(t)},a=0;a<i;++a){var s=t[a],u=o(s);n.hasOwnProperty(u)||(n[u]=!0,r.push(s))}return r}function Qt(t,e){return t.reduce(function(t,n){var r=e(n);return t[r]=t[r]||[],t[r].push(n),t},{})}function Zt(t,e){return Xt(t.concat(e))}function $t(t,e){return t.filter(function(t){return-1!==e.indexOf(t)})}function Jt(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];var r=e.length;if(0===r||!t)return t;for(var i=0;i<r;i++)for(var o=e[i],a=At(o)?Object.keys(o):[],s=a.length,u=0;u<s;u++){var c=a[u];void 0===t[c]&&(t[c]=o[c])}return t}function te(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];var r=Object.assign({},t);return e.forEach(function(t){delete r[t]}),r}function ee(t,e){var n=function(r){var i=n.cache,o=String(e?e.apply(this,arguments):r);return i.hasOwnProperty(o)||(i[o]=t.apply(this,arguments)),i[o]};return n.cache={},n}function ne(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return function(e,n){var r=0;return t.every(function(t){return 0===(r=t(e,n))}),r}}function re(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];return null==t?{}:e.reduce(function(e,n){var r=t[n];return r&&(e[n]=r),e},{})}function ie(t){return t=null==t?"":String(t),Et.test(t)?t.replace(St,function(t){return wt[t]}):t}function oe(t,e,n){!e&&n&&(e=n),e=Jt({},e,Mt);var r=RegExp([(e.escape||xt).source,(e.interpolate||xt).source,(e.evaluate||xt).source].join("|")+"|$","g"),i=0,o="__p+='";t.replace(r,function(e,n,r,a,s){return o+=t.slice(i,s).replace(Ot,function(t){return"\\"+jt[t]}),i=s+e.length,n?o+="'+\n((__t=("+n+"))==null?'':utils.escape(__t))+\n'":r?o+="'+\n((__t=("+r+"))==null?'':__t)+\n'":a&&(o+="';\n"+a+"\n__p+='"),e}),o+="';\n",e.variable||(o="with(obj||{}){\n"+o+"}\n"),o="var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};\n"+o+"return __p;\n";try{var a=new Function(e.variable||"obj",o)}catch(t){throw t.source=o,t}var s=function(t){return a.call(this,t)},u=e.variable||"obj";return s.source="function("+u+"){\n"+o+"}",s}function ae(t){return String(t).replace(/&/g,"&amp;").replace(/</g,"&lt;").replace(/>/g,"&gt;").replace(/"/g,"&quot;").replace(/'/g,"&#039;")}var se="  ",ue=32,ce=["img","input","br","embed","link","meta","area","base","basefont","bgsound","col","command","frame","hr","image","isindex","keygen","menuitem","nextid","param","source","track","wbr","circle","ellipse","line","path","polygon","rect"].reduce(function(t,e){return t[e]=!0,t},{});function le(t){var e=2,n=arguments[1];("object"!=typeof arguments[1]||Array.isArray(arguments[1]))&&(e=1,n={});var r=Kt(Array.prototype.slice.call(arguments,e)),i=1===r.length&&"<"!==r[0].trim()[0],o=ce[t];if(o&&r.length>0)throw new Error('Tag "'+t+'" is void but content is assigned to it');var a="<"+t,s=Object.keys(n).map(function(t){return" "+t+'="'+n[t]+'"'}).join("");s.length>ue&&(s=Object.keys(n).map(function(t){return"\n"+se+t+'="'+n[t]+'"'}).join(""));var u=i?r[0]:"\n"+r.map(function(t){return String(t).split("\n").map(function(t){return""+se+t}).join("\n")}).join("\n")+"\n";return""+a+s+(o?"/>":">"+u+"</"+t+">")}function de(t){var e=t,n={then:function(t){return e=t(e),n},result:function(){return e}};return n}var pe=[ht];function fe(t){return pe.some(function(e){return t instanceof e})}function he(t){return(t.units||[]).some(function(t){return t.hasOwnProperty("units")})}var me=document.createElement("div"),ge=new WeakMap;function ye(t,e){var n;return t instanceof Node?n=t:(me.insertAdjacentHTML("afterbegin",t),n=me.childNodes[0]),e.appendChild(n),n}function _e(t){if(ge.has(t))return ge.get(t);var e=t.style.overflow;t.style.overflow="scroll";var n={width:t.offsetWidth-t.clientWidth,height:t.offsetHeight-t.clientHeight};return t.style.overflow=e,ge.set(t,n),n}function ve(t,e){var n="horizontal"===(e=e||"both")||"both"===e,r="vertical"===e||"both"===e,i=_e(t),o=r?i.width+"px":"0",a=n?i.height+"px":"0";t.style.overflow="hidden",t.style.padding="0 "+o+" "+a+" 0";var s=t.scrollWidth>t.clientWidth,u=t.scrollHeight>t.clientHeight,c=r&&!u?i.width+"px":"0",l=n&&!s?i.height+"px":"0";return t.style.padding="0 "+c+" "+l+" 0",t.style.overflow="",t.style.overflowX=s?"scroll":"hidden",t.style.overflowY=u?"scroll":"hidden",i}function be(t,e){return window.getComputedStyle(t).getPropertyValue(e)}function xe(t,e){return parseInt(be(t,e)||"0",10)}function we(t){var e=xe(t,"padding-left"),n=xe(t,"padding-right"),r=xe(t,"padding-bottom"),i=xe(t,"padding-top"),o=xe(t,"border-top-width")+xe(t,"border-left-width")+xe(t,"border-right-width")+xe(t,"border-bottom-width"),a=t.getBoundingClientRect();return{width:a.width-e-n-2*o,height:a.height-r-i-2*o}}function je(t){var e=document.createElement("div");e.style.position="absolute",e.style.visibility="hidden",e.style.width="100px",e.style.height="100px",e.style.border="1px solid green",e.style.top="0",document.body.appendChild(e),e.innerHTML='<svg class="tau-chart__svg">\n                <g class="tau-chart__cell cell">\n                <g class="x axis">\n                <g class="tick"><text></text></g>\n                </g>\n                </g>\n                </svg>';var n=e.querySelector(".x.axis .tick text");n.textContent=t;var r={width:0,height:0},i=n.getBoundingClientRect();r.width=i.right-i.left,r.height=i.bottom-i.top;var o=0!==t.length?r.width/t.length:0;return r.width=r.width+1.5*o,document.body.removeChild(e),r}function Oe(t,e){var n=e.fontSize,r=e.fontFamily,i=e.fontWeight,o="string"==typeof n?n:n+"px",a=t.map(function(t){for(var e=0,n=0;e<=t.length-1;e++)n+=Ce(t.charAt(e),{fontSize:o,fontFamily:r,fontWeight:i}).width;return n}).sort(function(t,e){return e-t})[0],s=t.length,u=parseInt(o);return{width:a,height:u*s+.39*u*s}}var Ce=ee(function(t,e){var n=e.fontSize,r=e.fontFamily,i=e.fontWeight,o=document.createElement("div");o.style.position="absolute",o.style.visibility="hidden",o.style.border="0px",o.style.top="0",o.style.fontSize=n,o.style.fontFamily=r,o.style.fontWeight=i,document.body.appendChild(o),o.innerHTML=" "===t?"&nbsp;":t;var a={width:0,height:0},s=o.getBoundingClientRect();return a.width=s.right-s.left,a.height=s.bottom-s.top,document.body.removeChild(o),a},function(t,e){return t+"_"+JSON.stringify(e)});function Ee(t,e){var n={".":function(t,e){return e.classed(t,!0)},"#":function(t,e){return e.attr("id",t)}},r=Object.keys(n).join("");if(e.indexOf(" ")>=0)throw new Error("Selector should not contain whitespaces.");if(r.indexOf(e[0])>=0)throw new Error("Selector must have tag at the beginning.");var i,o=t instanceof Element,s=o?a.select(t):t,u=function(t){return o?t.node():t},c=s.selectAll(e).filter(function(){return this.parentNode===s.node()}).filter(function(t,e){return 0===e});if(!c.empty())return u(c);for(var l,d=-1,p=null,f=1,h=e.length;f<=h;f++)(f==h||r.indexOf(e[f])>=0)&&(l=e.substring(d+1,f),d<0?i=s.append(l):n[p].call(null,l,i),p=e[f],d=f);return u(i)}function Se(t,e){return Me(t,e)[0]||null}function Me(t,e){for(var n=[],r=Element.prototype.matches||Element.prototype.msMatchesSelector||Element.prototype.webkitMatchesSelector,i=t.firstElementChild;Boolean(i);i=i.nextElementSibling)r.call(i,e)&&n.push(i);return n}function Te(t,e){if(t.childElementCount>0){var n,r,i=Array.prototype.filter.call(t.childNodes,function(t){return t.nodeType===Node.ELEMENT_NODE}),o=i.slice().sort(e),a=i.reduce(function(t,e,n){return t.set(e,n),t},new Map),s=o.reduce(function(t,e,i){var s=a.get(e),u=i-s;return u!==r&&(n&&t.push(n),r=u,n={from:s,to:i,elements:[]}),n.elements.push(e),i===o.length-1&&t.push(n),t},[]),u=s.slice().sort(function(t,e){return t.from-e.from}),c=u.reduce(function(t,e,n){return t.set(e,n),t},new Map),l=function(t){for(var e,n,r,i=s.map(function(t,e){return{elements:t.elements,from:c.get(t),to:e}}).sort(ne(function(t,e){return t.elements.length-e.elements.length},t?function(t,e){return e.to-t.to}:function(t,e){return t.to-e.to})),o=0;o<i.length;o++){if((n=i[o]).from>n.to)for(e=o+1;e<i.length;e++)(r=i[e]).from>=n.to&&r.from<n.from&&r.from++;if(n.from<n.to)for(e=o+1;e<i.length;e++)(r=i[e]).from>n.from&&r.from<=n.to&&r.from--}return i.filter(function(t){return t.from!==t.to})},d=l(!0),p=l(!1),f=d.length<p.length?d:p,h=u.map(function(t){return t.elements});f.forEach(function(e){var n,r=h.splice(e.from,1)[0],i=h[e.to],o=i?i[0]:null;1===e.elements.length?n=r[0]:(n=document.createDocumentFragment(),r.forEach(function(t){n.appendChild(t)})),t.insertBefore(n,o),h.splice(e.to,0,r)})}}function ke(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];var n=[];return t.filter(function(t){return Boolean(t)}).forEach(function(t){"string"==typeof t?n.push(t):"object"==typeof t&&n.push.apply(n,Object.keys(t).filter(function(e){return Boolean(t[e])}))}),Xt(n).join(" ").trim().replace(/\s{2,}/g," ")}function Ae(t,e){for(var n,r=[],i=2;i<arguments.length;i++)r[i-2]=arguments[i];var o=document.createEvent("MouseEvents"),a=[!0,!0,window,0,0,0,0,0,!1,!1,!1,!1,0,null],s=r.concat(a.slice(r.length));(n=o).initMouseEvent.apply(n,[e].concat(s)),t.dispatchEvent(o)}function De(t,e){return"translate("+t+","+e+")"}function Ie(t){return"rotate("+t+")"}function Pe(t){return["bottom","top"].indexOf(t.toLowerCase())>=0?"h":"v"}function Re(t){var e={x:0,y:0},n=t.indexOf("translate(");if(n>=0){var r=t.indexOf(")",n+10),i=t.substring(n+10,r).trim().replace(","," ").replace(/\s+/," ").split(" ");e.x=parseFloat(i[0]),i.length>1&&(e.y=parseFloat(i[1]))}return e}function Ne(t,e,n,r,i,o,a,s){var u,c,l,d,p,f;return f=((l=a-i)*(e-o)-(d=s-o)*(t-i))/(-l*(c=r-e)+(u=n-t)*d),(p=(-c*(t-i)+u*(e-o))/(-l*c+u*d))>=0&&p<=1&&f>=0&&f<=1}function Fe(t){for(var e,n,r={x:0,y:0},i=t;i&&"SVG"!==i.nodeName.toUpperCase();)(n=i.getAttribute("transform"))&&(e=Re(n),r.x+=e.x,r.y+=e.y),i=i.parentNode;return r}function Le(t,e,n){var r=t.selectAll(e).filter(n);if(!r.empty()){var i=a.select(r.node().parentNode).selectAll(e).filter(function(t){return!n(t)}).nodes(),o=i[i.length-1];if(o){var s=Array.prototype.indexOf.call(o.parentNode.childNodes,o),u=o.nextSibling;r.each(function(){Array.prototype.indexOf.call(this.parentNode.childNodes,this)>s||this.parentNode.insertBefore(this,u)})}}}var Be,Ve,ze=function(t){return kt(t)?t.getTime():t},He={cross:function(t,e,n){var r=t(),i=Xt(r.map(function(t){return t[e]}),ze),o=Xt(r.map(function(t){return t[n]}),ze),a=0===i.length?[null]:i;return(0===o.length?[null]:o).reduce(function(t,r){return t.concat(a.map(function(t){var i={};return e&&(i[e]=ze(t)),n&&(i[n]=ze(r)),i}))},[])},cross_period:function(t,e,n,r,i,o){var a=t(),s=!!o&&o.utcTime,u=Xt(a.map(function(t){return t[e]}),ze),c=Xt(a.map(function(t){return t[n]}),ze),l=0===u.length?[null]:u,d=0===c.length?[null]:c;return r&&(l=x.generate(Math.min.apply(Math,u),Math.max.apply(Math,u),r,{utc:s})),i&&(d=x.generate(Math.min.apply(Math,c),Math.max.apply(Math,c),i,{utc:s})),d.reduce(function(t,r){return t.concat(l.map(function(t){var i={};return e&&(i[e]=ze(t)),n&&(i[n]=ze(r)),i}))},[])},groupBy:function(t,e){return Xt(t().map(function(t){return t[e]}),ze).map(function(t){var n;return(n={})[e]=ze(t),n})},none:function(){return[null]}},Ue=function(){function t(t,e,n){var r=t.key,i=t.pipe,o=t.source,a=t.units;void 0===n&&(n={}),this.key=r,this.pipe=i||[],this.source=o,this.units=a,this._frame={key:r,source:o,pipe:this.pipe},this._data=e,this._pipeReducer=function(t,e){return n[e.type](t,e.args)}}return t.prototype.hash=function(){return Nt([this._frame.pipe,this._frame.key,this._frame.source].map(function(t){return JSON.stringify(t)}).join(""))},t.prototype.full=function(){return this._data},t.prototype.part=function(t){return void 0===t&&(t=function(t){return t}),this._frame.pipe.map(t).reduce(this._pipeReducer,this._data)},t}(),We=(Ve=function(t,e){return(Ve=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}Ve(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),qe=function(t){return kt(t)?t.getTime():t},Ge=function(t){var e=this;Object.keys(t).forEach(function(n){return e[n]=t[n]})},Ye=function(t,e,n){return t.map(function(t){return"string"==typeof t?n.get(t):t}).filter(function(t){return t}).reduce(function(t,e){return n=t,void 0===(r=e(t,{}))&&(r={}),Object.assign(new Ge(n),r);var n,r},e)},Ke=function(t){function e(e,n,r,i){var o=t.call(this)||this;return Jt(e.scales,{size_null:{type:"size",source:"?"},split_null:{type:"value",source:"?"},label_null:{type:"value",source:"?"},color_null:{type:"color",source:"?"},identity_null:{type:"identity",source:"?"},"size:default":{type:"size",source:"?"},"color:default":{type:"color",source:"?"},"split:default":{type:"value",source:"?"},"label:default":{type:"value",source:"?"},"identity:default":{type:"identity",source:"?"}}),e.settings=e.settings||{},o.config=e,o.sources=e.sources,o.scales=e.scales,o.unitSet=r,o.grammarRules=i,o.scalesHub=n,o.transformations=Object.assign(e.transformations||{},{where:function(t,e){var n=Object.keys(e||{}).map(function(t){return function(n){return qe(n[t])===e[t]}});return t.filter(function(t){return n.every(function(e){return e(t)})})}}),o}return We(e,t),e.traverseSpec=function(t,e,n,r,i){void 0===r&&(r=null),void 0===i&&(i=null);var o=[],a=function(t,e,n,r,i){o.push(function(){e(t,r,i)}),t.frames&&t.frames.forEach(function(r){(r.units||[]).map(function(i){return a(i,e,n,t,r)})}),o.push(function(){return n(t,r,i)})};return a(t.unit,e,n,r,i),o},e.prototype.unfoldStructure=function(){return this.root=this._expandUnitsStructure(this.config.unit),this.config},e.prototype.getDrawScenarioQueue=function(t){var e=this,n=this.grammarRules,r=this._flattenDrawScenario(t,function(t,r,i){var o=!1===r.expression.inherit?null:i,a=e._createFrameScalesFactoryMethod(o),s=e.unitSet.create(r.type,Object.assign({},r,{options:t.allocateRect(i.key)})),u=new Ge(s.defineGrammarModel(a)),c=Ye(s.getGrammarRules(),u,n);return Ye(s.getAdjustScalesRules(),c,n),s.node().screenModel=s.createScreenModel(c),s}),i=this._flattenDrawScenario(t,function(t,r,i){var o=!1===r.expression.inherit?null:i,a=e._createFrameScalesFactoryMethod(o),s=e.unitSet.create(r.type,Object.assign({},r,{options:t.allocateRect(i.key)})),u=new Ge(s.defineGrammarModel(a)),c=Ye(s.getGrammarRules(),u,n);return s.node().screenModel=s.createScreenModel(c),s.parentUnit=t,s.addInteraction(),s});return r.concat(function(){Object.keys(e.scales).forEach(function(t){return e.scalesHub.createScaleInfo(e.scales[t]).commit()})}).concat(i)},e.prototype._flattenDrawScenario=function(t,n){var r={},i=[],o=[t],a=e.traverseSpec({unit:this.root},function(t,e,a){var s;t.uid=(s=Nt((e?e.uid+"/":"")+JSON.stringify(Object.keys(t).filter(function(e){return"string"==typeof t[e]}).reduce(function(e,n){return e[n]=t[n],e},{}))+"-"+JSON.stringify(a.pipe)),r.hasOwnProperty(s)?s+="-"+ ++r[s]:r[s]=0,s),t.guide=Tt(t.guide);var u,c=n(o[0],t,a);i.push(c),0===t.type.indexOf("COORDS.")&&(u=c,o.unshift(u))},function(t){0===t.type.indexOf("COORDS.")&&o.shift()},null,this._datify({source:this.root.expression.source,pipe:[]}));return a.push(function(){return i}),a},e.prototype._expandUnitsStructure=function(t,e){var n=this;void 0===e&&(e=[]);var r=this;if(!1===t.expression.operator)t.frames=t.frames.map(function(t){return r._datify(t)});else{var i=this._parseExpression(t.expression,e,t.guide);t.transformation=t.transformation||[],t.frames=i.exec().map(function(n){var o=(i.inherit?e:[]).concat([{type:"where",args:n}]).concat(t.transformation);return r._datify({key:n,pipe:o,source:i.source,units:t.units?t.units.map(function(t){var e=Tt(t);return e.guide=t.guide,e}):[]})})}return t.frames.forEach(function(t){return t.units.forEach(function(e){return n._expandUnitsStructure(e,t.pipe)})}),t},e.prototype._createFrameScalesFactoryMethod=function(t){var e=this;return function(n,r,i){var o=r||n+":default";return e.scalesHub.createScaleInfo(e.scales[o],t).create("function"==typeof i?i(e.scales[o]):i)}},e.prototype._datify=function(t){return new Ue(t,this.sources[t.source].data,this.transformations)},e.prototype._parseExpression=function(t,e,n){var r=this,i=t.operator||"none",o=t.source,a=!1!==t.inherit,s=t.params,u={source:o,pipe:a?e:[]},c=He[i];if(!c)throw new Error(i+" operator is not supported");return{source:o,inherit:a,func:c,args:s,exec:function(){return c.apply(void 0,[function(){return r._datify(u).part()}].concat(s||[],[n]))}}},e}(h),Xe="tau-chart__",Qe=n(10),Ze=(Be=function(t,e){return(Be=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}Be(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)});Qe.Tooltip.defaults.baseClass=Xe+"tooltip";var $e,Je,tn=["top","bottom"],en=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Ze(e,t),e.prototype._pickPlace=function(t){if(!this.options.auto)return this.options.place;var e,n,r,i,o,a,s,u,c=0|this.options.winBound,l=(e=c,r=window,i=document.documentElement,o=(r.pageYOffset||i.scrollTop)-i.clientTop,a=(r.pageXOffset||i.scrollLeft)-i.clientTop,s=r.innerWidth||i.clientWidth,u=r.innerHeight||i.clientHeight,{top:(n={top:o,right:a+s,bottom:o+u,left:a,width:s,height:u}).top+e,right:n.right-e,bottom:n.bottom-e,left:n.left+e,width:n.width-2*e,height:n.height-2*e}),d=this.options.place.split("-"),p=this.spacing;if(-1!==tn.indexOf(d[0]))switch(t.top-this.height-p<=l.top?d[0]="bottom":t.bottom+this.height+p>=l.bottom&&(d[0]="top"),d[1]){case"left":t.right-this.width<=l.left&&(d[1]="right");break;case"right":t.left+this.width>=l.right&&(d[1]="left");break;default:t.left+t.width/2+this.width/2>=l.right?d[1]="left":t.right-t.width/2-this.width/2<=l.left&&(d[1]="right")}else switch(t.left-this.width-p<=l.left?d[0]="right":t.right+this.width+p>=l.right&&(d[0]="left"),d[1]){case"top":t.bottom-this.height<=l.top&&(d[1]="bottom");break;case"bottom":t.top+this.height>=l.bottom&&(d[1]="top");break;default:t.top+t.height/2+this.height/2>=l.bottom?d[1]="top":t.bottom-t.height/2-this.height/2<=l.top&&(d[1]="bottom")}return d.join("-")},e}(Qe.Tooltip),nn=function(){function t(t,e){this.chart=e,this.handlers=new Map,this.plugins=t.map(this.initPlugin,this)}return t.prototype.initPlugin=function(t){var e=this;t.init&&t.init(this.chart);var n=[];this.handlers.set(t,n);var r=function(r,i){n.push(e.chart.on(r,i,t))};return t.destroy&&r("destroy",t.destroy.bind(t)),Object.keys(t).forEach(function(e){if(0===e.indexOf("on")){var n=e.substr(2).toLowerCase();r(n,t[e].bind(t))}}),t},t.prototype.destroyPlugin=function(t){var e=this;t.destroy&&t.destroy(),this.handlers.get(t).forEach(function(n){e.chart.removeHandler(n,t)})},t.prototype.destroy=function(){var t=this;this.plugins.forEach(function(e){return t.destroyPlugin(e)})},t}(),rn={},on={},an={reg:function(t,e,n){return n?(on[t]=n,rn[t]=function(t,e){this.___tauchartsseed___=new e(this.init(t))},rn[t].prototype=Object.assign({init:function(t){return t},defineGrammarModel:function(t){return this.node().defineGrammarModel(t)},getGrammarRules:function(t){return this.node().getGrammarRules(t)},getAdjustScalesRules:function(t){return this.node().getAdjustScalesRules(t)},createScreenModel:function(t){return this.node().createScreenModel(t)},addInteraction:function(){this.node().addInteraction()},node:function(){return this.___tauchartsseed___},draw:function(){this.node().draw()}},e)):rn[t]=e,this},get:function(t){if(!rn.hasOwnProperty(t))throw new j("Unknown unit type: "+t,O.UNKNOWN_UNIT_TYPE);return rn[t]},create:function(t,e){var n=this.get(t);return on[t]?new n(e,this.get(on[t])):new n(e)}},sn={},un={},cn={reg:function(t,e,n){return void 0===n&&(n=function(t){return t}),sn[t]=e,un[t]=n,cn},get:function(t){return sn[t]},instance:function(t){return void 0===t&&(t={}),{create:function(e,n,r){return new(cn.get(e))(n,(0,un[e])(r,t))}}}},ln=function(){function t(t,e,n){this.registry=t,this.sources=e,this.scales=n}return t.prototype.createScaleInfo=function(t,e){void 0===e&&(e=null);var n=t.dim,r=t.source,i=(this.sources[r].dims[n]||{}).type,o=this.sources[r].data,a=e||new Ue({source:r},o);return t.dimType=i,this.registry.create(t.type,a,t)},t.prototype.createScaleInfoByName=function(t,e){return void 0===e&&(e=null),this.createScaleInfo(this.scales[t],e)},t}(),dn=function(t,e,n){var r=!0,i=null;try{t.reduce(function(t,r){var o=function(t,e){var n=r[e],i=At(n)?JSON.stringify(n):n;return t.push(i),t},a=e.reduce(o,[]).join("/"),s=n.reduce(o,[]).join("/");if(t.hasOwnProperty(a)){var u=t[a];if(u!==s)throw i={type:"RelationIsNotAFunction",keyX:e.join("/"),keyY:n.join("/"),valX:a,errY:[u,s]},new Error("RelationIsNotAFunction")}else t[a]=s;return t},{})}catch(t){if("RelationIsNotAFunction"!==t.message)throw t;r=!1}return{result:r,error:i}},pn=function(t,e,n){var r=t,i=["period","time"].indexOf(n.scale)>=0?function(t){return new Date(t)}:function(t){return t},o=t.reduce(function(t,e,n){return t.set(e,n),t},new Map);if("measure"===n.type||"period"===n.scale)r=t.slice().sort(ne(function(t,n){return i(t[e])-i(n[e])},function(t,e){return o.get(t)-o.get(e)}));else if(n.order){var a=n.order.reduce(function(t,e,n){return t[e]=n,t},{}),s=n.order.length,u="(___"+e+"___)";r=t.map(function(t){var n=a[t[e]];return n=n>=0?n:s,t[u]=n,t}).sort(ne(function(t,e){return t[u]-e[u]},function(t,e){return o.get(t)-o.get(e)})).map(function(t){return delete t[u],t})}return r},fn=function(t,e){var n=document.createElement("div");return n.classList.add(Xe+t),e&&e.appendChild(n),n},hn=function(){var t=fn("layout"),e=fn("layout__header",t),n=fn("layout__container",t),r=fn("layout__sidebar",n),i=fn("layout__content",n),o=fn("layout__content__wrap",i),a=fn("layout__sidebar-right",n);return{layout:t,header:e,content:o,contentContainer:i,leftSidebar:r,rightSidebar:fn("layout__sidebar-right__wrap",a),rightSidebarContainer:a,footer:fn("layout__footer",t)}},mn=function(){function t(t){this.spec=t,this.dist={sources:{"?":{dims:{},data:[{}]},"/":{dims:{},data:[]}},scales:{x_null:{type:"ordinal",source:"?"},y_null:{type:"ordinal",source:"?"},size_null:{type:"size",source:"?"},color_null:{type:"color",source:"?"},split_null:{type:"value",source:"?"},"pos:default":{type:"ordinal",source:"?"},"size:default":{type:"size",source:"?"},"label:default":{type:"value",source:"?"},"color:default":{type:"color",source:"?"},"split:default":{type:"value",source:"?"}},settings:t.settings}}return t.prototype.convert=function(){var t=this.spec,e=this.dist;return this.ruleAssignSourceDims(t,e),this.ruleAssignStructure(t,e),this.ruleAssignSourceData(t,e),this.ruleApplyDefaults(e),e},t.prototype.ruleApplyDefaults=function(t){var e=t.settings||{},n=function(t,e,r){e(t,r),(t.units||[]).map(function(r){return n(r,e,t)})};n(t.unit,function(t,n){if(t.namespace="chart",t.guide=Jt(t.guide||{},{animationSpeed:e.animationSpeed||0,utcTime:e.utcTime||!1}),n&&!t.hasOwnProperty("units")){t=Jt(t,{x:n.x,y:n.y});var r=Tt(n.guide)||{};t.guide.x=Jt(t.guide.x||{},r.x),t.guide.y=Jt(t.guide.y||{},r.y),t.expression.inherit=n.expression.inherit}return!n||t.guide&&t.guide.hasOwnProperty("obsoleteVerticalStackOrder")||(t.guide=Object.assign(t.guide||{},{obsoleteVerticalStackOrder:(n.guide||{}).obsoleteVerticalStackOrder})),t},null)},t.prototype.ruleAssignSourceData=function(t,e){var n=t.spec.dimensions||{},r=e.sources["/"].dims,i=function(t,e){var n=t[e];return At(n)&&!kt(n)&&Object.keys(n).forEach(function(r){return t[e+"."+r]=n[r]}),t};e.sources["/"].data=t.data.map(function(t){var e=Object.keys(t).reduce(i,t);return Object.keys(r).reduce(function(t,e){return t.hasOwnProperty(e)||(t[e]=null),null!==t[e]&&n[e]&&["period","time"].indexOf(n[e].scale)>=0&&(t[e]=new Date(t[e])),t},e)})},t.prototype.ruleAssignSourceDims=function(t,e){var n=t.spec.dimensions;e.sources["/"].dims=Object.keys(n).reduce(function(t,e){return t[e]={type:n[e].type},t},{})},t.prototype.ruleAssignStructure=function(t,e){var n=this,r=function(t){var i=Tt(te(t,"unit"));return n.ruleCreateScales(t,i,e.settings),i.expression=n.ruleInferExpression(t),t.unit&&(i.units=t.unit.map(r)),i},i=r(t.spec.unit);i.expression.inherit=!1,e.unit=i},t.prototype.ruleCreateScales=function(t,e,n){var r=this,i=t.guide||{};["identity","color","size","label","x","y","split"].forEach(function(o){t.hasOwnProperty(o)&&(e[o]=r.scalesPool(o,t[o],i[o]||{},n))})},t.prototype.ruleInferDim=function(t,e){var n=t,r=this.spec.spec.dimensions;if(!r.hasOwnProperty(n))return n;e.hasOwnProperty("tickLabel")?n=t+"."+e.tickLabel:r[t].value&&(n=t+"."+r[t].value);var i=this.dist.sources["/"].dims;return i.hasOwnProperty(n)||(i[n]={type:i[t].type},delete i[t]),n},t.prototype.scalesPool=function(t,e,n,r){var i=t+"_"+e;if(this.dist.scales.hasOwnProperty(i))return i;var o=this.spec.spec.dimensions,a={};if("color"===t&&null!==e&&(a={type:"color",source:"/",dim:this.ruleInferDim(e,n)},n.hasOwnProperty("brewer")&&(a.brewer=n.brewer),o[e]&&o[e].hasOwnProperty("order")&&(a.order=o[e].order),n.hasOwnProperty("min")&&(a.min=n.min),n.hasOwnProperty("max")&&(a.max=n.max),n.hasOwnProperty("nice")&&(a.nice=n.nice)),"size"===t&&null!==e&&(a={type:"size",source:"/",dim:this.ruleInferDim(e,n)},n.hasOwnProperty("func")&&(a.func=n.func),n.hasOwnProperty("min")&&(a.min=n.min),n.hasOwnProperty("max")&&(a.max=n.max),n.hasOwnProperty("minSize")&&(a.minSize=n.minSize),n.hasOwnProperty("maxSize")&&(a.maxSize=n.maxSize)),"label"===t&&null!==e&&(a={type:"value",source:"/",dim:this.ruleInferDim(e,n)}),"split"===t&&null!==e&&(a={type:"value",source:"/",dim:this.ruleInferDim(e,n)}),"identity"===t&&null!==e&&(a={type:"identity",source:"/",dim:this.ruleInferDim(e,n)}),o.hasOwnProperty(e)&&("x"===t||"y"===t)){if(a={type:o[e].scale,source:"/",dim:this.ruleInferDim(e,n)},o[e].hasOwnProperty("order")&&(a.order=o[e].order),n.hasOwnProperty("min")&&(a.min=n.min),n.hasOwnProperty("max")&&(a.max=n.max),n.hasOwnProperty("autoScale")?a.autoScale=n.autoScale:a.autoScale=!0,n.hasOwnProperty("nice")?a.nice=n.nice:a.nice=a.autoScale,n.hasOwnProperty("niceInterval")?a.niceInterval=n.niceInterval:a.niceInterval=null,n.hasOwnProperty("tickPeriod")&&(a.period=n.tickPeriod,a.type="period"),n.hasOwnProperty("tickPeriod")&&n.hasOwnProperty("timeInterval"))throw new Error('Use "tickPeriod" for period scale, "timeInterval" for time scale, but not both');if(n.hasOwnProperty("timeInterval")){a.period=n.timeInterval,a.type="time";var s=x.get(a.period,{utc:r.utcTime});n.hasOwnProperty("min")&&(a.min=s.cast(new Date(n.min))),n.hasOwnProperty("max")&&(a.max=s.cast(new Date(n.max)))}a.fitToFrameByDims=n.fitToFrameByDims,a.ratio=n.ratio}return this.dist.scales[i]=a,i},t.prototype.getScaleConfig=function(t,e){var n=t+"_"+e;return this.dist.scales[n]},t.prototype.ruleInferExpression=function(t){var e={operator:"none",params:[]},n=t.guide||{},r=n.x||{},i=n.y||{},o=this.getScaleConfig("x",t.x),a=this.getScaleConfig("y",t.y);return 0===t.type.indexOf("ELEMENT.")?t.color&&(e={operator:"groupBy",params:[this.ruleInferDim(t.color,n.color||{})]}):"COORDS.RECT"===t.type&&1===t.unit.length&&"COORDS.RECT"===t.unit[0].type&&(e=o.period||a.period?{operator:"cross_period",params:[this.ruleInferDim(t.x,r),this.ruleInferDim(t.y,i),o.period,a.period]}:{operator:"cross",params:[this.ruleInferDim(t.x,r),this.ruleInferDim(t.y,i)]}),Object.assign({inherit:!0,source:"/"},e)},t}(),gn=n(6),yn=n(7),_n=function(){return(_n=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)},vn=_n({},gn,yn),bn=vn.format(".4s"),xn=vn.format(".2r"),wn=vn.format(".1e"),jn=($e=/\.0+([^\d].*)?$/,Je=/(\.\d+?)0+([^\d].*)?$/,function(t){return t.replace($e,"$1").replace(Je,"$1$2")}),On={_identity:function(t,e){return String(null==t?e:t)},"x-num-auto":function(t){if(isNaN(t))return"NaN";var e=Math.abs(t);return jn(e<1?0===e?"0":e<1e-6?wn(t):xn(t):bn(t))},percent:function(t){return parseFloat((100*t).toFixed(2)).toString()+"%"},day:vn.timeFormat("%d-%b-%Y"),"day-utc":vn.utcFormat("%d-%b-%Y"),"day-short":vn.timeFormat("%d-%b"),"day-short-utc":vn.utcFormat("%d-%b"),week:vn.timeFormat("%d-%b-%Y"),"week-utc":vn.utcFormat("%d-%b-%Y"),"week-short":vn.timeFormat("%d-%b"),"week-short-utc":vn.utcFormat("%d-%b"),month:function(t){var e=0===new Date(t).getMonth()?"%B, %Y":"%B";return vn.timeFormat(e)(t)},"month-utc":function(t){var e=0===new Date(t).getUTCMonth()?"%B, %Y":"%B";return vn.utcFormat(e)(t)},"month-short":function(t){var e=0===new Date(t).getMonth()?"%b '%y":"%b";return vn.timeFormat(e)(t)},"month-short-utc":function(t){var e=0===new Date(t).getUTCMonth()?"%b '%y":"%b";return vn.utcFormat(e)(t)},"month-year":vn.timeFormat("%B, %Y"),"month-year-utc":vn.utcFormat("%B, %Y"),quarter:function(t){var e=new Date(t),n=e.getMonth();return"Q"+((n-n%3)/3+1)+" "+e.getFullYear()},"quarter-utc":function(t){var e=new Date(t),n=e.getUTCMonth();return"Q"+((n-n%3)/3+1)+" "+e.getUTCFullYear()},year:vn.timeFormat("%Y"),"year-utc":vn.utcFormat("%Y"),"x-time-auto":null},Cn={get:function(t,e){var n=On._identity,r=On.hasOwnProperty(t),i=r?On[t]:n;return r&&(i=On[t]),!r&&t&&(i=function(e){return(kt(e)?vn.timeFormat(t):vn.format(t))(e)}),r||t||(i=n),null!==i?function(t){return i(t,e||"")}:null},add:function(t,e){On[t]=e}},En=function(t){return t.reduce(function(t,e){return t+e},0)},Sn=function(t,e,n){return t[e]=Jt(t[e]||{},{label:""}),t[e].label=At(t[e].label)?t[e].label:{text:t[e].label},t[e].label=Jt(t[e].label,n||{},{padding:32,rotate:0,textAnchor:"middle",cssClass:"label",dock:null}),t[e]},Mn=function(t,e,n){return t[e]=Jt(t[e],n||{},{padding:0,density:30,rotate:0,tickPeriod:null,tickFormat:null,autoScale:!0}),t[e].tickFormat=t[e].tickFormat||t[e].tickPeriod,t[e].nice=t[e].hasOwnProperty("nice")?t[e].nice:t[e].autoScale,t[e]},Tn=function(t){var e=t.units||[],n=!t.hasOwnProperty("units"),r=!e.some(function(t){return t.hasOwnProperty("units")});return{type:t.type,isLeaf:n,isLeafParent:!n&&r}},kn=function(t,e,n,r){if(0===t.length)return{width:0,height:0};if(null===e){var i=n("TauChart Library");return i.width=.625*r,i}t.every(function(t){return"number"==typeof t})&&(t=u.scaleLinear().domain(t).ticks());var o=t.reduce(function(t,n){var r=e(n).toString().length;return!t.computed||r>t.computed?{value:n,computed:r}:t},{}).value;return n(e(o))},An=function(t,e){var n=t.dimType,r=t.scaleType,i=[n,r,"*"].join(":"),o=[n,r].join(":");return e[i]||e[o]||e[n]||null},Dn=function(t,e,n){return t.hasOwnProperty(e+":"+n)?t[e+":"+n]:t[""+e]},In=function(t,e){return["day","week","month"].indexOf(t)>=0&&(t+="-short"+(e?"-utc":"")),t},Pn=function(t,e){var n=t.width,r=t.height,i=Math.abs(Ut(e));return{width:Math.max(Math.cos(i)*n,r),height:Math.max(Math.sin(i)*n,r)}},Rn=function(t,e){void 0===e&&(e="x");var n=Wt(t),r="x"===e?[[0,45,"middle"],[45,135,"start"],[135,225,"middle"],[225,315,"end"],[315,360,"middle"]]:[[0,90,"end"],[90,135,"middle"],[135,225,"start"],[225,315,"middle"],[315,360,"end"]],i=r.findIndex(function(t){return n>=t[0]&&n<t[1]});return r[i][2]},Nn=function(t,e,n){var r=Math.ceil(t.width/e);return{height:Math.min(r,n)*t.height,width:e}};function Fn(t,e,n,r,i,o){var a=n.values,s=r.values,u=n.isEmpty||t.x.hideTicks,c=r.isEmpty||t.y.hideTicks,l=kn(a,Cn.get(t.x.tickFormat,t.x.tickFormatNullAlias),e.getAxisTickLabelSize,e.xAxisTickLabelLimit),d=kn(s,Cn.get(t.y.tickFormat,t.y.tickFormatNullAlias),e.getAxisTickLabelSize,e.yAxisTickLabelLimit),p=l,f=d;l.width>e.xAxisTickLabelLimit&&(t.x.tickFormatWordWrap=!0,t.x.tickFormatWordWrapLines=e.xTickWordWrapLinesLimit,p=Nn(l,e.xAxisTickLabelLimit,e.xTickWordWrapLinesLimit)),d.width>e.yAxisTickLabelLimit&&(t.y.tickFormatWordWrap=!0,t.y.tickFormatWordWrapLines=e.yTickWordWrapLinesLimit,f=Nn(d,e.yAxisTickLabelLimit,e.yTickWordWrapLinesLimit)),o&&(t.y.tickFormatWordWrap=!1,t.y.tickFormatWordWrapLines=1,(f=Nn(d,2*e.yAxisTickLabelLimit,1)).width=20);var h=u?0:1,m=c?0:1,g=t.x.label,y=t.y.label,_=g.text&&!g.hide?1:0,v=y.text&&!y.hide?1:0,b=Pn(p,t.x.rotate),x=Pn(f,t.y.rotate);i?(g.padding=(-e.xAxisPadding-e.xFontLabelHeight)/2+e.xFontLabelHeight,g.paddingNoTicks=g.padding,y.padding=(-e.yAxisPadding-e.yFontLabelHeight)/2,y.paddingNoTicks=y.padding,_=0,v=0):(g.padding=En([h*(e.xTickWidth+b.height),_*(e.distToXAxisLabel+e.xFontLabelHeight)]),g.paddingNoTicks=_*(e.distToXAxisLabel+e.xFontLabelHeight),y.padding=En([m*(e.yTickWidth+x.width),v*e.distToYAxisLabel]),y.paddingNoTicks=v*e.distToYAxisLabel),o&&(y.padding=0,y.paddingNoTicks=0);var w=e.xFontLabelDescenderLineHeight;return t.padding=Object.assign(t.padding,{b:t.x.hide?0:En([t.x.padding,h*(e.xTickWidth+b.height),_*(e.distToXAxisLabel+e.xFontLabelHeight+w)]),l:t.y.hide?0:En([t.y.padding,o?0:m*(e.yTickWidth+x.width),v*(e.distToYAxisLabel+e.yFontLabelHeight)])}),t.paddingNoTicks=Object.assign({},t.paddingNoTicks,{b:t.x.hide?0:En([t.x.padding,_*(e.distToXAxisLabel+e.xFontLabelHeight+w)]),l:t.y.hide?0:En([t.y.padding,v*(e.distToYAxisLabel+e.yFontLabelHeight)])}),t.x=Object.assign(t.x,{density:b.width+2*Dn(e,"xDensityPadding",n.dimType),tickFontHeight:l.height,$minimalDomain:a.length,$maxTickTextW:p.width,$maxTickTextH:p.height,tickFormatWordWrapLimit:e.xAxisTickLabelLimit}),t.y=Object.assign(t.y,{density:x.height+2*Dn(e,"yDensityPadding",r.dimType),tickFontHeight:d.height,$minimalDomain:s.length,$maxTickTextW:f.width,$maxTickTextH:f.height,tickFormatWordWrapLimit:e.yAxisTickLabelLimit}),t}var Ln=function(t){var e=t.unit,n=t.meta,r=t.settings,i=t.allowXVertical,o=t.allowYVertical,a=t.inlineLabels,s=n.dimension(e.x),u=n.dimension(e.y),c=n.scaleMeta(e.x,e.guide.x),l=n.scaleMeta(e.y,e.guide.y),d=c.isEmpty,p=l.isEmpty;e.guide.x.tickFormat=In(e.guide.x.tickFormat||An(s,r.defaultFormats),r.utcTime),e.guide.y.tickFormat=In(e.guide.y.tickFormat||An(u,r.defaultFormats),r.utcTime);var f=!(!i||"measure"===s.dimType),h=!(!o||"measure"===u.dimType);if(e.guide.x.padding=d?0:r.xAxisPadding,e.guide.x.paddingNoTicks=e.guide.x.padding,e.guide.y.padding=p?0:r.yAxisPadding,e.guide.y.paddingNoTicks=e.guide.y.padding,e.guide.x.rotate=f?-90:0,e.guide.x.textAnchor=Rn(e.guide.x.rotate,"x"),e.guide.y.rotate=h?-90:0,e.guide.y.textAnchor=Rn(e.guide.y.rotate,"y"),e.guide=Fn(e.guide,r,c,l,a,he(e)),a){var m=e.guide.x.label,g=e.guide.y.label;m.cssClass+=" inline",m.dock="right",m.textAnchor="end",g.cssClass+=" inline",g.dock="right",g.textAnchor="end"}return e},Bn={NONE:function(t,e,n){var r=Tt(t);return Vn(Tt(r.unit),r.unit,function(t,e){return e.guide.x.tickFontHeight=n.getAxisTickLabelSize("X").height,e.guide.y.tickFontHeight=n.getAxisTickLabelSize("Y").height,e.guide.x.tickFormatWordWrapLimit=n.xAxisTickLabelLimit,e.guide.y.tickFormatWordWrapLimit=n.yAxisTickLabelLimit,e}),r},"BUILD-LABELS":function(t,e){var n=Tt(t),r=[],i=[],o=null,a=null;return yt(n.unit,"units",Tn,function(t,n){if(t.isLeaf)return n;!o&&n.x&&(o=n),!a&&n.y&&(a=n),n.guide=n.guide||{},n.guide.x=n.guide.x||{label:{text:""}},n.guide.y=n.guide.y||{label:{text:""}},n.guide.x.label=At(n.guide.x.label)?n.guide.x.label:{text:n.guide.x.label&&n.guide.x.label.text?n.guide.x.label.text:""},n.guide.y.label=At(n.guide.y.label)?n.guide.y.label:{text:n.guide.y.label&&n.guide.y.label.text?n.guide.y.label.text:""},n.x&&(n.guide.x.label.text=n.guide.x.label.text||e.dimension(n.x).dimName),n.y&&(n.guide.y.label.text=n.guide.y.label.text||e.dimension(n.y).dimName);var s=n.guide.x.label.text;s&&(r.push(s),n.guide.x.tickFormatNullAlias=n.guide.x.hasOwnProperty("tickFormatNullAlias")?n.guide.x.tickFormatNullAlias:"No "+s,n.guide.x.label.text="",n.guide.x.label._original_text=s);var u=n.guide.y.label.text;return u&&(i.push(u),n.guide.y.tickFormatNullAlias=n.guide.y.hasOwnProperty("tickFormatNullAlias")?n.guide.y.tickFormatNullAlias:"No "+u,n.guide.y.label.text="",n.guide.y.label._original_text=u),n}),o&&(o.guide.x.label.text=o.guide.x.label.hide?"":r.join(" → ")),a&&(a.guide.y.label.text=a.guide.y.label.hide?"":i.join(" → ")),n},"BUILD-GUIDE":function(t,e,n){var r=Tt(t);return Vn(Tt(r.unit),r.unit,function(t,r){if(t.isLeaf)return r;var i=!t.isLeaf&&!t.isLeafParent,o=e.scaleMeta(r.x,r.guide.x),a=e.scaleMeta(r.y,r.guide.y),s=!i&&Boolean(o.dimType)&&"measure"!==o.dimType;return r.guide.x.rotate=r.guide.x.rotate||(s?-90:0),r.guide.x.textAnchor=Rn(r.guide.x.rotate),r.guide.x.tickFormat=r.guide.x.tickFormat||An(o,n.defaultFormats),r.guide.y.tickFormat=r.guide.y.tickFormat||An(a,n.defaultFormats),r.guide.x.padding=i?0:n.xAxisPadding,r.guide.x.paddingNoTicks=r.guide.x.padding,r.guide.y.padding=i?0:n.yAxisPadding,r.guide.y.paddingNoTicks=r.guide.y.padding,r.guide=Fn(r.guide,Jt({distToXAxisLabel:o.isEmpty?n.xTickWidth:n.distToXAxisLabel,distToYAxisLabel:a.isEmpty?n.yTickWidth:n.distToYAxisLabel},n),o,a,null,he(r)),r.guide.x=Object.assign(r.guide.x,{cssClass:i?r.guide.x.cssClass+" facet-axis":r.guide.x.cssClass,avoidCollisions:!!i||r.guide.x.avoidCollisions}),r.guide.y=Object.assign(r.guide.y,{cssClass:i?r.guide.y.cssClass+" facet-axis":r.guide.y.cssClass,avoidCollisions:!i&&r.guide.y.avoidCollisions}),r.guide=Object.assign(r.guide,{showGridLines:r.guide.hasOwnProperty("showGridLines")?r.guide.showGridLines:t.isLeafParent?"xy":""}),r}),r},"BUILD-COMPACT":function(t,e,n){var r=Tt(t);return Vn(Tt(r.unit),r.unit,function(t,r){return t.isLeaf?r:(r.guide.hasOwnProperty("showGridLines")||(r.guide.showGridLines=t.isLeafParent?"xy":""),t.isLeafParent?Ln({unit:r,meta:e,settings:Jt({xTickWordWrapLinesLimit:1,yTickWordWrapLinesLimit:1},n),allowXVertical:!0,allowYVertical:!1,inlineLabels:!0}):(r.guide.x.cssClass+=" facet-axis compact",r.guide.x.avoidCollisions=!0,r.guide.y.cssClass+=" facet-axis compact",r.guide.y.avoidCollisions=!0,Ln({unit:r,meta:e,settings:Jt({xAxisPadding:0,yAxisPadding:0,distToXAxisLabel:0,distToYAxisLabel:0,xTickWordWrapLinesLimit:1,yTickWordWrapLinesLimit:1},n),allowXVertical:!1,allowYVertical:!0,inlineLabels:!1})))}),r},AUTO:function(t,e,n){return["BUILD-LABELS","BUILD-GUIDE"].reduce(function(t,r){return Bn[r](t,e,n)},t)},COMPACT:function(t,e,n){return["BUILD-LABELS","BUILD-COMPACT"].reduce(function(t,r){return Bn[r](t,e,n)},t)}},Vn=function(t,e,n){var r,i=((r=e).options=r.options||{},r.guide=r.guide||{},r.guide.padding=Jt(r.guide.padding||{},{l:0,b:0,r:0,t:0}),r.guide.x=Sn(r.guide,"x"),r.guide.x=Mn(r.guide,"x",{cssClass:"x axis",scaleOrient:"bottom",textAnchor:"middle"}),r.guide.y=Sn(r.guide,"y",{rotate:-90}),r.guide.y=Mn(r.guide,"y",{cssClass:"y axis",scaleOrient:"left",textAnchor:"end"}),r.guide.size=Sn(r.guide,"size"),r.guide.color=Sn(r.guide,"color"),r);i=n(Tn(i),i);var o,a,s,u=te((o=i,a=t.guide||{},s={x:["label"],y:["label"],size:["label"],color:["label"],padding:[]},Object.keys(s).forEach(function(t){var e=s[t];!function(t,e,n,r){var i=t.hasOwnProperty(n)?t[n]:{};i=i||{},r.forEach(function(t){Object.assign(e.guide[n][t],i[t])})}(a,o,t,e)}),Object.assign(o.guide,Object.keys(a).reduce(function(t,e){return s.hasOwnProperty(e)||(t[e]=a[e]),t},{})),i=o),"units");return(i.units||[]).forEach(function(t){return Vn(Tt(t),(r=u,(e=t).guide=e.guide||{},e.guide.padding=e.guide.padding||{l:0,t:0,r:0,b:0},e.hasOwnProperty("units")||((e=Jt(e,r)).guide=Jt(e.guide,Tt(r.guide)),e.guide.x=Jt(e.guide.x,Tt(r.guide.x)),e.guide.y=Jt(e.guide.y,Tt(r.guide.y))),e),n);var e,r}),i},zn=function(){function t(t){this.spec=t,this.isApplicable=Lt(t.unit)}return t.prototype.transform=function(t){var e=this.spec;if(!this.isApplicable)return e;var n,r,i,o,a,s,u=e.settings.size,c=e.settings.specEngine.find(function(t){return u.width<=t.width||u.height<=t.height});return n=c.name,r=e.settings,i=e,o=function(e,n){return t.getScaleInfo(n||e+":default")},a={dimension:function(t){var e=i.scales[t],n=i.sources[e.source].dims[e.dim]||{};return{dimName:e.dim,dimType:n.type,scaleType:e.type}},scaleMeta:function(t){var e=o("pos",t).domain(),n=i.scales[t],r=i.sources[n.source].dims[n.dim]||{};return{dimName:n.dim,dimType:r.type,scaleType:n.type,values:e,isEmpty:null==r.type}}},s=(Bn[n]||Bn.NONE)({unit:Tt(i.unit)},a,r),i.unit=s.unit,i},t}(),Hn=function(){function t(){}return t.optimizeXAxisLabel=function(t,e){var n=e.xAxisTickLabelLimit,r=function(t){if(!t.guide.x.hide&&!t.guide.x.hideTicks&&0!==t.guide.x.rotate){t.guide.x.rotate=0,t.guide.x.textAnchor="middle";var e=0-Math.min(n,t.guide.x.$maxTickTextW)+t.guide.x.$maxTickTextH;i(t,e)}(t.units||[]).filter(function(t){return"COORDS.RECT"===t.type}).forEach(function(t){return r(t)})},i=function(e,n){t!==e&&"extract-axes"===e.guide.autoLayout?(t.guide.x.padding+=n,t.guide.padding.b+=n):(e.guide.x.label.padding+=e.guide.x.label.padding>0?n:0,e.guide.padding.b+=e.guide.padding.b>0?n:0)};r(t)},t.hideAxisTicks=function(t,e,n){var r=function(t){var e="x"===n?"b":"l",i=t.guide;if(!i[n].hide&&!i[n].hideTicks){he(t)&&"y"===n||(i[n].hideTicks=!0);var o=i[n].label.text&&!i[n].label.hide;i.padding[e]=i.paddingNoTicks?i.paddingNoTicks[e]:0,i[n].padding=i[n].paddingNoTicks||0,i[n].label.padding=o?i[n].label.paddingNoTicks:0}(t.units||[]).filter(function(t){return"COORDS.RECT"===t.type}).forEach(function(t){return r(t)})};r(t)},t.facetsLabelsAtTop=function(t,e){var n=function(t){var e=t.units||[];if(he(t)){var r=t.guide;r.y.facetAxis=!0,r.y.rotate=0,r.y.textAnchor="start",e.forEach(function(t){t.guide.padding.t=20})}e.filter(function(t){return"COORDS.RECT"===t.type}).forEach(function(t){return n(t)})};n(t)},t}(),Un=function(t){return t.$maxTickTextW},Wn=function(t){return 0==t.rotate?t.$maxTickTextW:t.$maxTickTextH},qn=function(t){return t.density},Gn=function(t){var e=[],n=[],r=function(t){if("COORDS.RECT"===t.type&&t.units&&t.units[0]&&"COORDS.RECT"===t.units[0].type){var i=t.x.replace(/^x_/,""),o=t.y.replace(/^y_/,"");"null"!==i&&e.push(i),"null"!==o&&n.push(o),t.units.forEach(r)}};r(t.unit);var i={},o={},a=function(t){"COORDS.RECT"===t.type&&t.frames.forEach(function(t){t.key&&(Object.keys(t.key).forEach(function(r){e.indexOf(r)>=0&&(i.hasOwnProperty(r)||(i[r]=[]),i[r].indexOf(t.key[r])<0&&i[r].push(t.key[r])),n.indexOf(r)>=0&&(o.hasOwnProperty(r)||(o[r]=[]),o[r].indexOf(t.key[r])<0&&o[r].push(t.key[r]))}),t.units&&t.units.forEach(a))})};return a(t.unit),{xFacetCount:Object.keys(i).reduce(function(t,e){return t*i[e].length},1),yFacetCount:Object.keys(o).reduce(function(t,e){return t*o[e].length},1)}},Yn={"entire-view":function(t,e,n,r){var i=n.unit.guide,o=Gn(n),a=o.xFacetCount,s=o.yFacetCount;s>0&&Hn.facetsLabelsAtTop(n.unit,n.settings);var u=i.paddingNoTicks?i.padding.l-i.paddingNoTicks.l:0,c=i.paddingNoTicks?i.padding.b-i.paddingNoTicks.b:0,l=i.paddingNoTicks&&t.height-c<n.settings.minChartHeight||s*n.settings.minFacetHeight+c>t.height||a*n.settings.minFacetWidth+u>t.width,d=i.paddingNoTicks&&t.width-u<n.settings.minChartWidth||s*n.settings.minFacetHeight+c>t.height||a*n.settings.minFacetWidth+u>t.width;l&&Hn.hideAxisTicks(n.unit,n.settings,"x"),d&&Hn.hideAxisTicks(n.unit,n.settings,"y");var p=t.width;if(e("x",n.unit,Un)<=t.width)r(n.unit,n.settings);else{var f=e("x",n.unit,Wn);if(f>t.width){var h=Math.max(t.width,e("x",n.unit,qn));p=Math.min(f,h)}}var m=Math.max(t.height,e("y",n.unit,qn));return!l&&p>t.width&&Hn.hideAxisTicks(n.unit,n.settings,"x"),!d&&m>t.height&&Hn.hideAxisTicks(n.unit,n.settings,"y"),{newW:t.width,newH:t.height}},minimal:function(t,e,n){return{newW:e("x",n.unit,qn),newH:e("y",n.unit,qn)}},normal:function(t,e,n,r){Gn(n).yFacetCount>0&&Hn.facetsLabelsAtTop(n.unit,n.settings);var i=n.unit.guide;i.paddingNoTicks&&(t.width-i.padding.l+i.paddingNoTicks.l<n.settings.minChartWidth&&Hn.hideAxisTicks(n.unit,n.settings,"y"),t.height-i.padding.b+i.paddingNoTicks.b<n.settings.minChartHeight&&Hn.hideAxisTicks(n.unit,n.settings,"x"));var o=t.width;if(e("x",n.unit,Un)<=t.width)r(n.unit,n.settings);else{var a=e("x",n.unit,Wn);if(a>t.width){var s=Math.max(t.width,e("x",n.unit,qn));o=Math.min(a,s)}}return{newW:o,newH:Math.max(t.height,e("y",n.unit,qn))}},"fit-width":function(t,e,n,r){var i=n.unit.guide,o=i.paddingNoTicks?i.padding.l-i.paddingNoTicks.l:0;return(i.paddingNoTicks&&t.width-o<n.settings.minChartWidth||Gn(n).xFacetCount*n.settings.minFacetWidth+o>t.width)&&Hn.hideAxisTicks(n.unit,n.settings,"y"),e("x",n.unit,Un)<=t.width&&r(n.unit,n.settings),{newW:t.width,newH:e("y",n.unit,qn)}},"fit-height":function(t,e,n){var r=n.unit.guide,i=r.paddingNoTicks?r.padding.b-r.paddingNoTicks.b:0;return(r.paddingNoTicks&&t.height-i<n.settings.minChartHeight||Gn(n).yFacetCount*n.settings.minFacetHeight+i>t.height)&&Hn.hideAxisTicks(n.unit,n.settings,"x"),{newW:e("x",n.unit,qn),newH:t.height}}},Kn=function(){function t(t){this.spec=t,this.isApplicable=Lt(t.unit)}return t.prototype.transform=function(t){var e=this.spec;if(!this.isApplicable)return e;var n=e.settings.fitModel;if(!n)return e;var r,i,o,a,s,u,c,l,d=e.scales,p=function(t,e){return t.discrete?e*t.domain().length:4*e},f=function(e,n,r,i){void 0===i&&(i=null);var o="x"===e?n.x:n.y,a="x"===e?n.y:n.x,s=n.guide,u=r("x"===e?s.x:s.y),c=n.units[0],l="x"===e?s.padding.l+s.padding.r:s.padding.b+s.padding.t;if("ELEMENT.INTERVAL"===c.type&&"y"===e===Boolean(c.flip)&&c.label&&!t.getScaleInfo(c.label,i).isEmpty()){var h,m=2*(s.label&&s.label.fontSize?s.label.fontSize:10),g=t.getScaleInfo(o,i);if(g.discrete&&(null==c.guide.enableColorToBarPosition?!c.stack:c.guide.enableColorToBarPosition)){var y=c.color;if(y){var _=t.getScaleInfo(y,i);_.discrete&&(_.domain(),h=_)}}var v=n.frames.reduce(function(t,e){return t+function(t){var e=t.part(),n=Xt(e.map(function(t){return t[g.dim]})).length,r=1;if(h){var i=e.reduce(function(t,e){var n=e[g.dim],r=e[h.dim];return t[n]=t[n]||{},t[n][r]||(t[n][r]=!0),t},{}),o=Object.keys(i).map(function(t){return Object.keys(i[t]).length});r=Math.max.apply(Math,o)}return n*r*m}(e)},0),b=p(g,u);return l+Math.max(v,b)}if("COORDS.RECT"!==c.type){var x=t.getScaleInfo(o,i);return l+p(x,u)}var w,j,O=(w=n.frames,j=d[a].dim,w.reduce(function(t,e){var n=(e.key||{})[j];return t[n]=t[n]||[],t[n].push(e),t},{})),C=Object.keys(O).map(function(t){return O[t].map(function(t){return f(e,t.units[0],r,t)}).reduce(function(t,e){return t+e},0)});return l+Math.max.apply(Math,C)},h=e.settings.size,m=h.width,g=h.height,y=Yn[n];if(y){var _=y(h,f,e,Hn.optimizeXAxisLabel);m=_.newW,g=_.newH}return e.settings.size=(r=h,i={width:m,height:g},o=e.settings.avoidScrollAtRatio,a=e.settings.getScrollbarSize(t.getLayout().contentContainer),s=i.width>r.width&&i.width<=r.width*o?r.width:i.width,u=i.height>r.height&&i.height<=r.height*o?r.height:i.height,c=r.width-s,l=r.height-u>=0?0:a.width,{height:u-(c>=0?0:a.height),width:s-l}),e},t}(),Xn=function(){function t(t){this.spec=t,this.isApplicable=t.settings.autoRatio&&Lt(t.unit)}return t.prototype.transform=function(t){var e=this.spec;if(!this.isApplicable)return e;try{this.ruleApplyRatio(e,t)}catch(t){if("Not applicable"!==t.message)throw t}return e},t.prototype.ruleApplyRatio=function(t,e){var n=function(t){return"COORDS.RECT"===t.type||"RECT"===t.type},r=function(t,e,n,i){void 0===i&&(i=0),e(t,i)&&(t.units||[]).map(function(t){return r(t,e,n,i+1)}),n(t,i)},i=[],o=[];r(t.unit,function(t,e){if(e>1||!n(t))throw new Error("Not applicable");return i.push(t.x),o.push(t.y),1===(t.units||[]).map(function(t){if(!n(t)&&0!==t.type.indexOf("ELEMENT."))throw new Error("Not applicable");return t}).filter(n).length},function(){return 0});var a=function(e){return t.scales[e]},s=function(t){return"/"===t.source&&!t.ratio&&!t.fitToFrameByDims},u=function(t){return"ordinal"===t.type||"period"===t.type&&!t.period},c=i.map(a).filter(s),l=o.map(a).filter(s);if([c.length,l.length].some(function(t){return 2===t})){var d=function(t){return t.dim},p=function(t,e,n){t.fitToFrameByDims=n.slice(0,e).map(d)},f=function(t,n){2===n.filter(u).length&&(n.forEach(p),n[0].ratio=Ft(t,n.map(d),e))};f("x",c),f("y",l)}},t}(),Qn=function(){function t(t){this.spec=t,this.isApplicable="EXTRACT"===t.settings.layoutEngine&&Lt(t.unit)}return t.prototype.transform=function(){var t=this.spec;if(!this.isApplicable)return t;try{this.ruleExtractAxes(t)}catch(t){if("Not applicable"!==t.message)throw t;console.log("[TauCharts]: can't extract axes for the given chart specification")}return t},t.prototype.ruleExtractAxes=function(t){var e=function(t){return"COORDS.RECT"===t.type||"RECT"===t.type},n=function(t){return t?10:0},r={l:0,r:10,t:10,b:0},i={l:0,b:0},o=[],a=[];_t(t.unit,function(t,n){if(n>1||!e(t))throw new Error("Not applicable");t.guide=t.guide||{};var s=t.guide,u=s.padding||{l:0,r:0,t:0,b:0},c=s.paddingNoTicks||{l:0,b:0};return r.l+=u.l,r.r+=u.r,r.t+=u.t,r.b+=u.b,i.l+=c.l,i.b+=c.b,o.push(Object.assign({},r)),a.push(Object.assign({},i)),1===(t.units||[]).map(function(t){if(!e(t)&&0!==t.type.indexOf("ELEMENT."))throw new Error("Not applicable");return t}).filter(e).length},function(t){var e=o.pop(),s=a.pop(),u=t.guide||{};u.x=u.x||{},u.x.padding=u.x.padding||0,u.x.paddingNoTicks=u.x.paddingNoTicks||0,u.y=u.y||{},u.y.padding=u.y.padding||0,u.y.paddingNoTicks=u.y.paddingNoTicks||0,u.padding={l:n(t.y),r:n(1),t:n(1),b:n(t.x)},u.paddingNoTicks={l:0,b:0},u.autoLayout="extract-axes",u.x.padding+=r.b-e.b,u.y.padding+=r.l-e.l,u.x.paddingNoTicks+=i.b-s.b,u.y.paddingNoTicks+=i.l-s.l}),t.unit.guide.padding=r,t.unit.guide.paddingNoTicks=i},t}(),Zn=function(){function t(t){var e=void 0===t?{}:t,n=e.src,r=void 0===n?null:n,i=e.timeout,o=void 0===i?Number.MAX_SAFE_INTEGER:i,a=e.syncInterval,s=void 0===a?Number.MAX_SAFE_INTEGER:a,u=e.callbacks,c=void 0===u?{}:u,l=this;this.setTimeoutDuration(o),this.setSyncInterval(s),this.setCallbacks(c),this._running=!1,this._queue=[],this._result=r,this._syncDuration=0,this._asyncDuration=0,this._requestedFrameId=null,this._visibilityChangeHandler=function(){if(l._running&&l._requestedFrameId){var t=l._getCancelFrameFunction(),e=l._getRequestFrameFunction();t(l._requestedFrameId),l._requestedFrameId=e(l._requestedFrameCallback)}},this._tasksCount=0,this._finishedTasksCount=0}return t.prototype.setTimeoutDuration=function(e){t.checkType(e,"number","timeout"),this._timeout=e},t.prototype.setSyncInterval=function(e){t.checkType(e,"number","syncInterval"),this._syncInterval=e},t.prototype.setCallbacks=function(e){t.checkType(e,"object","callbacks"),this._callbacks=Object.assign(this._callbacks||{},e)},t.prototype.addTask=function(t){return this._queue.push(t),this._tasksCount++,this},t.prototype.run=function(){if(this._running)throw new Error("Task Runner is already running");this._running=!0,t.runnersInProgress++,document.addEventListener("visibilitychange",this._visibilityChangeHandler),this._loopTasks()},t.prototype.isRunning=function(){return this._running},t.prototype._loopTasks=function(){for(var e,n,r,i,o=0,a=this._syncInterval/t.runnersInProgress;this._running&&!(r=this._asyncDuration>this._timeout)&&!(i=o>a)&&(e=this._queue.shift());){if(null===(n=this._runTask(e)))return;this._syncDuration+=n,this._asyncDuration+=n,o+=n}r&&this._queue.length>0&&(this.stop(),this._callbacks.timeout&&this._callbacks.timeout.call(null,this._asyncDuration,this)),!r&&i&&this._queue.length>0&&this._requestFrame(),0===this._queue.length&&(this.stop(),this._callbacks.done&&this._callbacks.done.call(null,this._result,this))},t.prototype._runTask=function(t){var e=performance.now();if(this._callbacks.error)try{this._result=t.call(null,this._result,this)}catch(t){return this.stop(),this._callbacks.error.call(null,t,this),null}else this._result=t.call(null,this._result,this);var n=performance.now()-e;return this._finishedTasksCount++,this._callbacks.progress&&this._callbacks.progress.call(null,this._finishedTasksCount/this._tasksCount,this),n},t.prototype._requestFrame=function(){var t=this,e=performance.now(),n=function(){t._requestedFrameId=null;var n=performance.now();t._asyncDuration+=n-e,t._loopTasks()},r=this._getRequestFrameFunction();this._requestedFrameCallback=n,this._requestedFrameId=r(n)},t.prototype._getRequestFrameFunction=function(){var t=this;return document.hidden?function(e){return t._requestedFrameType="idle",window.requestIdleCallback(e,{timeout:17})}:function(e){return t._requestedFrameType="animation",requestAnimationFrame(e)}},t.prototype._getCancelFrameFunction=function(){switch(this._requestedFrameType){case"animation":return function(t){return cancelAnimationFrame(t)};case"idle":return function(t){return window.cancelIdleCallback(t)}}},t.prototype.stop=function(){if(!this._running)throw new Error("Task Runner is already stopped");this._running=!1,t.runnersInProgress--,document.removeEventListener("visibilitychange",this._visibilityChangeHandler),this._requestedFrameId&&(this._getCancelFrameFunction()(this._requestedFrameId),this._requestedFrameId=null)},t.checkType=function(t,e,n){if(typeof t!==e)throw new Error('Task Runner "'+n+'" property is not "'+e+'"')},t}();Zn.runnersInProgress=0;var $n,Jn,tr=Zn,er=(Jn=function(t,e){return(Jn=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}Jn(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),nr=function(){return(nr=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)},rr=nr({},a),ir=Ee,or=Se,ar=function(t){function e(e){var n=t.call(this)||this;return n._nodes=[],n._svg=null,n._filtersStore={filters:{},tick:0},n._layout=hn(),n.transformers=[Xn,zn],n.onUnitsStructureExpandedTransformers=[Qn,Kn],n._chartDataModel=function(t){return t},n._reportProgress=null,n._taskRunner=null,n._renderingPhase=null,n.applyConfig(e),n}return er(e,t),e.prototype.on=function(e,n,r){return t.prototype.on.call(this,e,n,r)},e.prototype.updateConfig=function(t){this.applyConfig(t),this.refresh()},e.prototype.applyConfig=function(t){this._filtersStore.filters={},t=this.setupConfigSettings(t),this.configGPL=this.createGPLConfig(t),this._originData=Object.assign({},this.configGPL.sources),this._liveSpec=this.configGPL,this._emptyContainer=t.emptyContainer||"",this.setupPlugins(t)},e.prototype.createGPLConfig=function(t){var n;return this.isGPLConfig(t)?n=t:(t=this.setupConfig(t),n=new mn(t).convert()),n=e.setupPeriodData(n)},e.prototype.isGPLConfig=function(t){return 2===["sources","scales"].filter(function(e){return t.hasOwnProperty(e)}).length},e.prototype.setupPlugins=function(t){var e=t.plugins||[];this._plugins&&this._plugins.destroy(),this._plugins=new nn(e,this)},e.prototype.setupConfigSettings=function(t){var n;return this._dataRefs=this._dataRefs||(n=0,{references:new WeakMap,refCounter:function(){return++n}}),t.settings=e.setupSettings(Jt(t.settings||{},this._dataRefs)),t},e.prototype.destroy=function(){this.destroyNodes(),rr.select(this._svg).remove(),rr.select(this._layout.layout).remove(),this._cancelRendering(),t.prototype.destroy.call(this)},e.prototype.setupChartSourceModel=function(t){this._chartDataModel=t},e.prototype.setupConfig=function(t){if(!t.spec||!t.spec.unit)throw new Error("Provide spec for plot");var n=Jt(t,{spec:{},data:[],plugins:[],settings:{}});n.spec.dimensions=e.setupMetaInfo(n.spec.dimensions,n.data);var r,i,o,a=n.settings.log;return n.settings.excludeNull&&this.addFilter({tag:"default",src:"/",predicate:(r=n.spec.dimensions,i=function(t){return a([t,"point was excluded, because it has undefined values."],"WARN")},o=Object.keys(r).reduce(function(t,e){var n=r[e];return n.hasOwnProperty("hasNull")&&!n.hasNull||"measure"!==n.type&&"period"!==n.scale||t.push(e),t},[]),function(t){var e=!o.some(function(e){return!t.hasOwnProperty(e)||null===t[e]});return e||i(t),e})}),n},e.setupPeriodData=function(t){var n=e.__api__.tickPeriod,r=t.settings.log;return Object.keys(t.scales).map(function(e){return t.scales[e]}).filter(function(t){return"period"===t.type}).forEach(function(e){n.get(e.period,{utc:t.settings.utcTime})||(r(['Unknown period "'+e.period+'".',"Docs: https://api.taucharts.com/plugins/customticks.html#how-to-add-custom-tick-period"],"WARN"),e.period=null)}),t},e.setupMetaInfo=function(t,e){var n,r,i,o,a=t||(n={type:"category",scale:"ordinal"},e.reduce(function(t,e){return Object.keys(e).forEach(function(r){var i=e.hasOwnProperty(r)?e[r]:null;if(t[r]=t[r]||{type:null,hasNull:!1},null===i)t[r].hasNull=!0;else{var o=function(t,e){var n=e;return kt(t)?(n.type="measure",n.scale="time"):At(t)?(n.type="order",n.scale="ordinal"):Number.isFinite(t)&&(n.type="measure",n.scale="linear"),n}(i,Tt(n)),a=o.type,s=o.scale,u=null!==t[r].type&&t[r].type!==a;t[r].type=u?n.type:a,t[r].scale=u?n.scale:s}}),t},{}));return r=a,i={category:"ordinal",order:"ordinal",measure:"linear"},o={},Object.keys(r).forEach(function(t){var e=r[t],n=(e.type||"category").toLowerCase();o[t]=Object.assign({},e,{type:n,scale:e.scale||i[n],value:e.value})}),o},e.setupSettings=function(t){var n=e.globalSettings,r=Jt(t||{},Object.keys(n).reduce(function(t,e){return t[e]="function"==typeof n[e]?n[e]:Tt(n[e]),t},{}));return Array.isArray(r.specEngine)||(r.specEngine=[{width:Number.MAX_VALUE,name:r.specEngine}]),r},e.prototype.insertToLeftSidebar=function(t){return ye(t,this._layout.leftSidebar)},e.prototype.insertToRightSidebar=function(t){return ye(t,this._layout.rightSidebar)},e.prototype.insertToFooter=function(t){return ye(t,this._layout.footer)},e.prototype.insertToHeader=function(t){return ye(t,this._layout.header)},e.prototype.addBalloon=function(t){return new en("",t||{})},e.prototype.destroyNodes=function(){this._nodes.forEach(function(t){return t.destroy()}),this._nodes=[],this._renderedItems=[]},e.prototype.onUnitDraw=function(t){var e=this;this._nodes.push(t),this.fire("unitdraw",t),["click","mouseover","mouseout"].forEach(function(n){return t.on(n,function(t,r){e.fire("element"+n,{element:t,data:r.data,event:r.event})})})},e.prototype.onUnitsStructureExpanded=function(t){var e=this;this.onUnitsStructureExpandedTransformers.forEach(function(n){return new n(t).transform(e)}),this.fire("unitsstructureexpanded",t)},e.prototype._getClosestElementPerUnit=function(t,e){return this._renderedItems.filter(function(t){return t.getClosestElement}).map(function(n){var r=n.getClosestElement(t,e);return{unit:n.node(),closest:r}})},e.prototype.disablePointerEvents=function(){this._layout.layout.style.pointerEvents="none"},e.prototype.enablePointerEvents=function(){this._layout.layout.style.pointerEvents=""},e.prototype._handlePointerEvent=function(t){var e=this._svg.getBoundingClientRect(),n=t.clientX-e.left,r=t.clientY-e.top,i="click"===t.type?"data-click":"data-hover",o=null,a=null,s=null,u=this._getClosestElementPerUnit(n,r),c=u.filter(function(t){return t.closest}).sort(function(t,e){return t.closest.distance===e.closest.distance?t.closest.secondaryDistance-e.closest.secondaryDistance:t.closest.distance-e.closest.distance});if(c.length>0){var l=c.findIndex(function(t){return t.closest.distance!==c[0].closest.distance||t.closest.secondaryDistance!==c[0].closest.secondaryDistance}),d=l<0?c:c.slice(0,l);if(1===d.length)o=d[0].closest.data,a=d[0].closest.node,s=d[0].unit;else{var p=d.reduce(function(t,e){return t+e.closest.x},0)/d.length,f=d.reduce(function(t,e){return t+e.closest.y},0)/d.length,h=Math.atan2(f-r,p-n)+Math.PI,m=Math.round((d.length-1)*h/2/Math.PI),g=d[m].closest;o=g.data,a=g.node,s=d[m].unit}}u.forEach(function(e){return e.unit.fire(i,{event:t,data:o,node:a,unit:s})})},e.prototype._initPointerEvents=function(){var t=this;this._liveSpec.settings.syncPointerEvents||(this._pointerAnimationFrameId=null);var e=rr.select(this._svg),n=this._liveSpec.settings.syncPointerEvents?function(t){return function(){return t(a.event)}}:function(e){return function(){var n=a.event;t._pointerAnimationFrameId&&"mousemove"!==n.type&&t._cancelPointerAnimationFrame(),t._pointerAnimationFrameId||(t._pointerAnimationFrameId=requestAnimationFrame(function(){t._pointerAnimationFrameId=null,e(n)}))}},r=function(e){return t._handlePointerEvent(e)};e.on("mousemove",n(r)),e.on("click",n(r)),e.on("mouseleave",n(function(e){"none"!==window.getComputedStyle(t._svg).pointerEvents&&t.select(function(){return!0}).forEach(function(t){return t.fire("data-hover",{event:e,data:null,node:null,unit:null})})}))},e.prototype._cancelPointerAnimationFrame=function(){cancelAnimationFrame(this._pointerAnimationFrameId),this._pointerAnimationFrameId=null},e.prototype._setupTaskRunner=function(t){var e=this;return this._resetTaskRunner(),this._taskRunner=new tr({timeout:t.settings.renderingTimeout||Number.MAX_SAFE_INTEGER,syncInterval:t.settings.asyncRendering?t.settings.syncRenderingInterval:Number.MAX_SAFE_INTEGER,callbacks:{done:function(){e._completeRendering(),e._renderingPhase=null},timeout:function(t,n){e._displayTimeoutWarning({timeout:t,proceed:function(){e.disablePointerEvents(),n.setTimeoutDuration(Number.MAX_SAFE_INTEGER),n.run()},cancel:function(){e._cancelRendering()}}),e.enablePointerEvents(),e.fire("renderingtimeout",t)},progress:function(t){var n={spec:0,draw:1}[e._renderingPhase]/2+t/2;e._reportProgress(n)},error:t.settings.handleRenderingErrors?function(n){e._cancelRendering(),e._displayRenderingError(n),e.fire("renderingerror",n),t.settings.log(["An error occured during chart rendering.",'Set "handleRenderingErrors: false" in chart settings to debug.',"Error message: "+n.message].join(" "),"ERROR")}:null}}),this._taskRunner},e.prototype._resetTaskRunner=function(){this._taskRunner&&this._taskRunner.isRunning()&&(this._taskRunner.stop(),this._taskRunner=null)},e.prototype.renderTo=function(t,e){this._resetProgressLayout(),this.disablePointerEvents(),this._insertLayout(t,e);var n=this._createLiveSpec();if(!n)return this._svg=null,this._layout.content.innerHTML=this._emptyContainer,void this.enablePointerEvents();var r=this._createGPL(n),i=this._setupTaskRunner(n);this._scheduleDrawScenario(i,r),this._scheduleDrawing(i,r),i.run()},e.prototype._insertLayout=function(t,e){this._target=t,this._defaultSize=Object.assign({},e);var n=rr.select(t).node();if(null===n)throw new Error("Target element not found");this._layout.layout.parentNode!==n&&n.appendChild(this._layout.layout);var r=this._layout.content,i=_e(this._layout.contentContainer);this._layout.contentContainer.style.padding="0 "+i.width+"px "+i.height+"px 0",ve(this._layout.rightSidebarContainer,"vertical");var o=Object.assign({},e)||{};if(!o.width||!o.height){var a=r.parentElement,s=a.scrollLeft,u=a.scrollTop;r.style.display="none",o=Jt(o,we(r.parentNode)),r.style.display="",r.parentElement.scrollLeft=s,r.parentElement.scrollTop=u,o.height||(o.height=we(this._layout.layout).height)}this.configGPL.settings.size=o},e.prototype._createLiveSpec=function(){var t=this;return this._liveSpec=Tt(te(this.configGPL,"plugins")),this._liveSpec.sources=this.getDataSources(),this._liveSpec.settings=this.configGPL.settings,this._experimentalSetupAnimationSpeed(this._liveSpec),this.isEmptySources(this._liveSpec.sources)?null:(this._liveSpec=this.transformers.reduce(function(e,n){return new n(e).transform(t)},this._liveSpec),this.destroyNodes(),this.fire("specready",this._liveSpec),this._liveSpec)},e.prototype._experimentalSetupAnimationSpeed=function(t){t.settings.initialAnimationSpeed=t.settings.initialAnimationSpeed||t.settings.animationSpeed;var e=t.settings.experimentalShouldAnimate(t)?t.settings.initialAnimationSpeed:0;t.settings.animationSpeed=e;var n=function(t){t.guide=t.guide||{},t.guide.animationSpeed=e,t.units&&t.units.forEach(n)};n(t.unit)},e.prototype._createGPL=function(t){var e=new Ke(t,this.getScaleFactory(),an,E),n=e.unfoldStructure();return this.onUnitsStructureExpanded(n),e},e.prototype._scheduleDrawScenario=function(t,e){var n=this,r=rr.select(this._layout.content),i=e.config.settings.size;t.addTask(function(){return n._renderingPhase="spec"}),e.getDrawScenarioQueue({allocateRect:function(){return{slot:function(t){return r.selectAll(".uid_"+t)},frameId:"root",left:0,top:0,width:i.width,containerWidth:i.width,height:i.height,containerHeight:i.height}}}).forEach(function(e){return t.addTask(e)})},e.prototype._scheduleDrawing=function(t,e){var n=this,r=e.config.settings.size;t.addTask(function(t){n._renderingPhase="draw",n._renderRoot({scenario:t,newSize:r}),n._cancelPointerAnimationFrame(),n._scheduleRenderScenario(t)})},e.prototype._resetProgressLayout=function(){this._createProgressBar(),this._clearRenderingError(),this._clearTimeoutWarning()},e.prototype._renderRoot=function(t){var e=this,n=t.scenario,r=t.newSize,i=rr.select(this._layout.content),o=n[0].config.uid,a=ir(i,"svg").attr("width",Math.floor(r.width)).attr("height",Math.floor(r.height));a.attr("class")||a.attr("class",Xe+"svg"),this._svg=a.node(),this._initPointerEvents(),this.fire("beforerender",this._svg);var s=a.selectAll("g.frame-root").data([o],function(t){return t});s.enter().append("g").classed(Xe+"cell cell frame-root uid_"+o,!0).merge(s).call(function(t){t.classed("tau-active",!0),rt(t,e.configGPL.settings.animationSpeed,"frameRootToggle").attr("opacity",1)}),s.exit().call(function(t){t.classed("tau-active",!1),rt(t,e.configGPL.settings.animationSpeed,"frameRootToggle").attr("opacity",1e-6).remove()})},e.prototype._scheduleRenderScenario=function(t){var e=this;t.forEach(function(t){e._taskRunner.addTask(function(){t.draw(),e.onUnitDraw(t.node()),e._renderedItems.push(t)})})},e.prototype._completeRendering=function(){ve(this._layout.contentContainer),this._layout.rightSidebar.style.maxHeight=this._liveSpec.settings.size.height+"px",this.enablePointerEvents(),this._svg&&this.fire("render",this._svg),ve(this._layout.rightSidebarContainer,"vertical")},e.prototype._cancelRendering=function(){this.enablePointerEvents(),this._resetTaskRunner(),this._cancelPointerAnimationFrame()},e.prototype._createProgressBar=function(){var t=rr.select(this._layout.header),e=ir(t,"div."+Xe+"progress");e.select("div."+Xe+"progress__value").remove();var n=e.append("div").classed(Xe+"progress__value",!0).style("width",0);this._reportProgress=function(t){requestAnimationFrame(function(){e.classed(Xe+"progress_active",t<1),n.style("width",100*t+"%")})}},e.prototype._displayRenderingError=function(t){this._layout.layout.classList.add(Xe+"layout_rendering-error")},e.prototype._clearRenderingError=function(){this._layout.layout.classList.remove(Xe+"layout_rendering-error")},e.prototype.getScaleFactory=function(t){return void 0===t&&(t=null),new ln(cn.instance(this._liveSpec.settings),t||this._liveSpec.sources,this._liveSpec.scales)},e.prototype.getScaleInfo=function(t,e){return void 0===e&&(e=null),this.getScaleFactory().createScaleInfoByName(t,e)},e.prototype.getSourceFiltersIterator=function(t){var e=this,n=Kt(Object.keys(this._filtersStore.filters).map(function(t){return e._filtersStore.filters[t]})).filter(function(e){return!t(e)}).map(function(t){return t.predicate});return function(t){return n.reduce(function(e,n){return e&&n(t)},!0)}},e.prototype.getDataSources=function(t){var e=this;void 0===t&&(t={});var n=this._chartDataModel(this._originData);return Object.keys(n).filter(function(t){return"?"!==t}).reduce(function(r,i){var o,a=n[i],s=e.getSourceFiltersIterator((o=i,function(e){return t.excludeFilter&&-1!==t.excludeFilter.indexOf(e.tag)||e.src!==o}));return r[i]={dims:a.dims,data:a.data.filter(s)},r},{"?":n["?"]})},e.prototype.isEmptySources=function(t){return!Object.keys(t).filter(function(t){return"?"!==t}).filter(function(e){return t[e].data.length>0}).length},e.prototype.getChartModelData=function(t,e){return void 0===t&&(t={}),void 0===e&&(e="/"),this.getDataSources(t)[e].data},e.prototype.getDataDims=function(t){return void 0===t&&(t="/"),this._originData[t].dims},e.prototype.getData=function(t){return void 0===t&&(t="/"),this._originData[t].data},e.prototype.setData=function(t,e){void 0===e&&(e="/"),this._originData[e].data=t,this.refresh()},e.prototype.getSVG=function(){return this._svg},e.prototype.addFilter=function(t){t.src=t.src||"/";var e=t.tag,n=this._filtersStore.filters[e]=this._filtersStore.filters[e]||[],r=this._filtersStore.tick++;return t.id=r,n.push(t),r},e.prototype.removeFilter=function(t){var e=this;return Object.keys(this._filtersStore.filters).map(function(n){e._filtersStore.filters[n]=e._filtersStore.filters[n].filter(function(e){return e.id!==t})}),this},e.prototype.refresh=function(){this._target&&this.renderTo(this._target,this._defaultSize)},e.prototype.resize=function(t){void 0===t&&(t={}),this.renderTo(this._target,t)},e.prototype.select=function(t){return this._nodes.filter(t)},e.prototype.traverseSpec=function(t,e){var n=function(t,e,r,i){e(t,r,i),t.frames?t.frames.forEach(function(r){(r.units||[]).map(function(i){return n(i,e,t,r)})}):(t.units||[]).map(function(r){return n(r,e,t,null)})};n(t.unit,e,null,null)},e.prototype.getSpec=function(){return this._liveSpec},e.prototype.getLayout=function(){return this._layout},e.prototype._displayTimeoutWarning=function(t){var e=this,n=t.proceed,r=t.cancel,i=t.timeout,o=Math.round(100/3/1.5),a=function(t){return Math.round(100/3/1.5*t)};this._layout.content.style.height="100%",this._layout.content.insertAdjacentHTML("beforeend",'\n            <div class="'+Xe+'rendering-timeout-warning">\n            <svg\n                viewBox="0 0 200 100">\n                <text\n                    text-anchor="middle"\n                    font-size="'+o+'">\n                    <tspan x="100" y="'+a(1)+'">Rendering took more than '+Math.round(i)/1e3+'s</tspan>\n                    <tspan x="100" y="'+a(2)+'">Would you like to continue?</tspan>\n                </text>\n                <text\n                    class="'+Xe+'rendering-timeout-continue-btn"\n                    text-anchor="end"\n                    font-size="'+o+'"\n                    cursor="pointer"\n                    text-decoration="underline"\n                    x="'+(100-o/3)+'"\n                    y="'+a(3)+'">\n                    Continue\n                </text>\n                <text\n                    class="'+Xe+'rendering-timeout-cancel-btn"\n                    text-anchor="start"\n                    font-size="'+o+'"\n                    cursor="pointer"\n                    text-decoration="underline"\n                    x="'+(100+o/3)+'"\n                    y="'+a(3)+'">\n                    Cancel\n                </text>\n            </svg>\n            </div>\n        '),this._layout.content.querySelector("."+Xe+"rendering-timeout-continue-btn").addEventListener("click",function(){e._clearTimeoutWarning(),n.call(e)}),this._layout.content.querySelector("."+Xe+"rendering-timeout-cancel-btn").addEventListener("click",function(){e._clearTimeoutWarning(),r.call(e)})},e.prototype._clearTimeoutWarning=function(){var t=or(this._layout.content,"."+Xe+"rendering-timeout-warning");t&&(this._layout.content.removeChild(t),this._layout.content.style.height="")},e}(h),sr={},ur={},cr=function(t){var e="Chart type "+t+" is not supported.";throw console.log(e),console.log("Use one of "+Object.keys(sr).join(", ")+"."),new j(e,O.NOT_SUPPORTED_TYPE_CHART)},lr={validate:function(t,e){return ur.hasOwnProperty(t)||cr(t),ur[t].reduce(function(t,n){return t.concat(n(e)||[])},[])},get:function(t){var e=sr[t];return"function"!=typeof e&&cr(t),e},add:function(t,e,n){return void 0===n&&(n=[]),sr[t]=e,ur[t]=n,lr},getAllRegisteredTypes:function(){return sr}},dr=($n=function(t,e){return($n=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}$n(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),pr=function(t){function e(n){var r=t.call(this,n)||this;return n.autoResize&&e.winAware.push(r),r}return dr(e,t),e.prototype.applyConfig=function(e){var n=lr.validate(e.type,e);if(n.length>0)throw new Error(n[0]);var r=lr.get(e.type);(e=Jt(e,{autoResize:!0})).settings=ar.setupSettings(e.settings),e.dimensions=ar.setupMetaInfo(e.dimensions,e.data),t.prototype.applyConfig.call(this,r(e))},e.prototype.destroy=function(){var n=e.winAware.indexOf(this);-1!==n&&e.winAware.splice(n,1),t.prototype.destroy.call(this)},e}(ar);function fr(t){return t}pr.winAware=[],pr.resizeOnWindowEvent=function(){var t;function e(){t=0;for(var e=0,n=pr.winAware.length;e<n;e++)pr.winAware[e].resize()}return function(){!t&&pr.winAware.length&&(t=window.requestAnimationFrame(e))}}(),window.addEventListener("resize",pr.resizeOnWindowEvent);var hr=1e-6;function mr(t){return"translate("+(t+.5)+",0)"}function gr(t){return"translate(0,"+(t+.5)+")"}var yr={top:1,right:2,bottom:3,left:4};function _r(t){var e=yr[t.scaleGuide.scaleOrient],n=t.scale,r=t.scaleGuide,i=r.label,o=Jt(t,{tickSize:6,tickPadding:3,gridOnly:!1}),a=o.ticksCount,s=o.tickFormat,u=o.tickSize,c=o.tickPadding,l=o.gridOnly,d="linear"===n.scaleType,p="ordinal"===n.scaleType||"period"===n.scaleType,f=e===yr.top||e===yr.bottom,h=e===yr.top||e===yr.left?-1:1,m=f?"x":"y",g=f?"y":"x",y=f?mr:gr,_=f?1:-1;return function(o){var v;if(n.ticks){v=n.ticks(a);for(var b=Math.floor(1.25*a);v.length>b&&b>2&&v.length>2;)v=n.ticks(--b)}else v=n.domain();r.hideTicks&&(v=l?v.filter(function(t){return 0==t}):[]);var x=null==s?n.tickFormat?n.tickFormat(a):fr:s,w=Math.max(u,0)+c,j=n.range(),O=j[0]+.5,C=j[j.length-1]+.5,E=(n.bandwidth?function(t){var e=Math.max(0,t.bandwidth()-1)/2;return t.round()&&(e=Math.round(e)),function(n){return t(n)+e}}:fr)(n);if(r.facetAxis){var S=E;E=function(t){return S(t)-n.stepSize(t)/2}}var M,T=o.selection?o:null,k=T?T.selection():o,A=function(t){for(var e=t.node();e&&"svg"!==e.tagName;)e=e.parentNode;return e}(k).getBoundingClientRect();function D(t){var n=Wt(r.rotate);if(t.attr("transform",Ie(n)),Math.abs(n/90)%2>0){var i=n<180?1:-1,o=f?.5:-2,a=e===yr.top||e===yr.left?-1:1,s=o*(e===yr.top||e===yr.bottom?a<0?0:.71:.32);t.attr("x",9*i).attr("y",0).attr("dx",f?null:s+"em").attr("dy",s+"em")}}function I(t,e){void 0===e&&(e=0);var i=function(t){return Math.max(n.stepSize(t),r.tickFormatWordWrapLimit,e)};r.tickFormatWordWrap?nt(t,i,r.tickFormatWordWrapLines,r.tickFontHeight,!f):et(t,i)}function P(t){!function(t,e){var n=e?-10:20,r=e?0:1,i=e?1:-1,o=[];t.each(function(){var t=J.select(this),e=t.attr("transform").replace("translate(","").replace(" ",",").split(",")[r],n=i*parseFloat(e),a=t.select("text"),s=a.node().getBBox().width/2,u=n-s,c=n+s;o.push({c:n,s:u,e:c,l:0,textRef:a,tickRef:t})});var a,s=o.sort(function(t,e){return t.c-e.c});a=function(t,r,i){var o,a,s,u=t.e>r.s,c=i.s<r.e;if(u||c){r.l=(o=t.l,a={"[T][1]":-1,"[T][-1]":0,"[T][0]":1,"[F][0]":-1},s="["+u.toString().toUpperCase().charAt(0)+"]["+o+"]",a.hasOwnProperty(s)?a[s]:0);var l=r.textRef.size(),d=r.textRef.text();l>1&&(d=d.replace(/([\.]*$)/gi,"")+"...");var p=11*r.l,f=e?parseFloat(r.textRef.attr("y"))+p:0,h=e?0:p,m=e?p:0,g=function(t){var e=0;if(!t)return e;var n=t.indexOf("rotate(");if(n>=0){var r=t.indexOf(")",n+7),i=t.substring(n+7,r);e=parseFloat(i.trim())}return e}(r.textRef.attr("transform"));r.textRef.text(function(t,e){return 0===e?d:""}).attr("transform","translate("+h+","+m+") rotate("+g+")");var y={x1:0,x2:0,y1:f+(e?-1:5),y2:n};e||(y.transform="rotate(-90)"),Ee(r.tickRef,"line.label-ref").call(dt(y))}else r.tickRef.selectAll("line.label-ref").remove();return r},s.map(function(t,e,n){return a(n[e-1]||{e:-1/0,s:-1/0,l:0},t,n[e+1]||{e:1/0,s:1/0,l:0})})}(t,f)}k.attr("fill","none").attr("font-size",10).attr("font-family","sans-serif").attr("text-anchor",e===yr.right?"start":e===yr.left?"end":"middle"),l||(M=r.hideTicks||r.hide?[]:[null],de(k.selectAll(".domain").data(M)).then(function(t){return T&&t.exit().transition(T).attr("opacity",0).remove(),t.merge(t.enter().insert("path",".tick").attr("class","domain").attr("opacity",1).attr("stroke","#000"))}).then(function(t){return T?t.transition(T):t}).then(function(t){t.attr("d",e===yr.left||e==yr.right?"M"+h*u+","+O+"H0.5V"+C+"H"+h*u:"M"+O+","+h*u+"V0.5H"+C+"V"+h*u)}));var R,N,F=de(k.selectAll(".tick").data(v,function(t){return String(n(t))}).order()).then(function(t){var e=t.exit(),n=t.enter().append("g").attr("class","tick");return{tickExit:e,tickEnter:n,tick:t.merge(n)}}).then(function(t){if(d){var e=n.ticks(),r=n.domain(),i=v.length-1,o=e.length>1&&r[0]*r[1]<0&&-r[0]>(e[1]-e[0])/2&&r[1]>(e[i]-e[i-1])/2;t.tick.classed("zero-tick",function(t){return 0==t&&o})}return t}).result();de(F).then(function(t){var e=t.tickEnter,n=t.tickExit,r=t.tick;return T?(e.attr("opacity",hr).attr("transform",function(t){var e=E(t);return y(e)}),{tick:r.transition(T),tickExit:n.transition(T).attr("opacity",hr).attr("transform",function(t){var e=E(t);return isFinite(e)?y(e):this.getAttribute("transform")})}):{tick:r,tickExit:n}}).then(function(t){var e=t.tick;t.tickExit.remove(),e.attr("opacity",1).attr("transform",function(t){return y(E(t))})}),r.facetAxis||(R=h*u,N=p?function(t){return _*n.stepSize(t)/2}:null,de(F).then(function(t){var e=t.tick,n=t.tickEnter,r=e.select("line"),i=n.append("line").attr("stroke","#000").attr(g+"2",R);return p&&i.attr(m+"1",N).attr(m+"2",N),r.merge(i)}).then(function(t){return T?t.transition(T):t}).then(function(t){t.attr(g+"2",R),p&&t.attr(m+"1",N).attr(m+"2",N)})),p&&l&&p&&v&&v.length&&de(k.selectAll(".extra-tick-line").data([null])).then(function(t){return t.merge(t.enter().insert("line",".tick").attr("class","extra-tick-line").attr("stroke","#000"))}).then(function(t){return T?t.transition(T):t}).then(function(t){t.attr(m+"1",O).attr(m+"2",O).attr(g+"1",0).attr(g+"2",h*u)}),l||(function(i){var o=r.textAnchor,a=h*w,s=e===yr.top?"0em":e===yr.bottom?"0.71em":"0.32em";function u(e){if(r.facetAxis)return e.attr("dx",18-t.position[0]).attr("dy",16)}de(i).then(function(t){var e=t.tick,n=t.tickEnter,r=e.select("text"),i=n.append("text").attr("fill","#000").attr(g,a).attr("dy",s);return D(i),u(i),r.merge(i)}).then(function(t){if(t.text(x).attr("text-anchor",o),!1===f&&!0===r.facetAxis){var e=Re(k.node().parentNode.getAttribute("transform"));I(t,A.width-Math.abs(e.x))}else I(t);return f&&"time"===n.scaleType&&function(t){if(!(v.length<2)){var e=0,n=-1,r=t.nodes();r.forEach(function(t,r){var i=t.textContent||"",o=i.length;o>e&&(e=o,n=r)});var i=E(v[1])-E(v[0]),o=!1;if(n>=0){var a=r[n].getBoundingClientRect();o=i-a.width<8}k.classed(Xe+"time-axis-overflow",o)}}(t),!f||"time"!==n.scaleType&&"linear"!==n.scaleType||function(t){if(0!==v.length){var e=v[0],n=v[v.length-1],r=k.append("line").attr("x1",E(e)).attr("x2",E(e)).attr("y1",0).attr("y2",1),i=k.append("line").attr("x1",E(n)).attr("x2",E(n)).attr("y1",0).attr("y2",1),o={left:r.node().getBoundingClientRect().left-A.left,right:A.right-i.node().getBoundingClientRect().right};r.remove(),i.remove();var a=function(t,e,n){var r=t.getBoundingClientRect(),i=e>0?"right":"left",a=(E(n),o[i]),s=Math.ceil(r.width/2-a+1);t.setAttribute("dx",String(s>0?-e*s:0))},s=t.filter(function(t){return t===e}).node(),u=t.filter(function(t){return t===n}).node();t.attr("dx",null),a(s,-1,e),a(u,1,n)}}(t),t}).then(function(t){return T?t.transition(T):t}).then(function(t){t.attr(g,a),D(t),u(t),p&&r.avoidCollisions&&!r.facetAxis&&(T?T.on("end.fixTickTextCollision",function(){return P(i.tick)}):P(i.tick))})}(F),i.hide||function(){var t=i,e=Ee(k,"text.label").attr("class",ke("label",t.cssClass)).attr("transform",Ie(t.rotate)).attr("text-anchor",t.textAnchor);de(e).then(function(t){return T?t.transition(T):t}).then(function(e){var n=_*t.padding,r=Math.abs(C-O),i=f?r:0;e.attr("x",i).attr("y",n).attr("text-anchor","end")});for(var n=t.text.split(" → "),r=n.length-1;r>0;r--)n.splice(r,0," → ");e.selectAll("tspan").data(n).enter().append("tspan").attr("class",function(t,e){return e%2?"label-token-delimiter label-token-delimiter-"+e:"label-token label-token-"+e}).text(function(t){return t}).exit().remove()}())}}function vr(t){return _r({scale:t.scale,scaleGuide:t.scaleGuide,ticksCount:t.ticksCount,tickSize:t.tickSize,gridOnly:!0,position:t.position})}var br,xr,wr,jr=(xr=function(t,e){return(xr=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}xr(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),Or=Ee,Cr=function(t){var e=t<=20?1:.75;return Math.max(2,Math.round(t*e))},Er=function(t){function e(e){var n=t.call(this,e)||this;n.config=e,n.config.guide=Jt(n.config.guide||{},{showGridLines:"xy",padding:{l:50,r:0,t:0,b:50}}),n.config.guide.x=n.config.guide.x||{},n.config.guide.x=Jt(n.config.guide.x,{cssClass:"x axis",textAnchor:"middle",padding:10,hide:!1,scaleOrient:"bottom",rotate:0,density:20,label:{},tickFormatWordWrapLimit:100}),"string"==typeof n.config.guide.x.label&&(n.config.guide.x.label={text:n.config.guide.x.label}),n.config.guide.x.label=Jt(n.config.guide.x.label,{text:"X",rotate:0,padding:40,textAnchor:"middle"}),n.config.guide.y=n.config.guide.y||{},n.config.guide.y=Jt(n.config.guide.y,{cssClass:"y axis",textAnchor:"start",padding:10,hide:!1,scaleOrient:"left",rotate:0,density:20,label:{},tickFormatWordWrapLimit:100}),"string"==typeof n.config.guide.y.label&&(n.config.guide.y.label={text:n.config.guide.y.label}),n.config.guide.y.label=Jt(n.config.guide.y.label,{text:"Y",rotate:-90,padding:20,textAnchor:"middle"});var r=n.config,i=r.guide;if("extract-axes"===i.autoLayout){var o=r.options.containerHeight-(r.options.top+r.options.height);i.x.hide=i.x.hide||Math.floor(o)>0,i.y.hide=i.y.hide||Math.floor(r.options.left)>0}var a=n.config.options,s=n.config.guide.padding;return n.L=a.left+s.l,n.T=a.top+s.t,n.W=a.width-(s.l+s.r),n.H=a.height-(s.t+s.b),n}return jr(e,t),e.prototype.defineGrammarModel=function(t){var e=this.W,n=this.H;return this.xScale=t("pos",this.config.x,[0,e]),this.yScale=t("pos",this.config.y,function(t){return["ordinal","period"].indexOf(t.type)>=0?[0,n]:[n,0]}),this.regScale("x",this.xScale).regScale("y",this.yScale),{scaleX:this.xScale,scaleY:this.yScale,xi:function(){return e/2},yi:function(){return n/2},sizeX:function(){return e},sizeY:function(){return n}}},e.prototype.getGrammarRules=function(){return[function(t){var e=t.scaleX,n=t.scaleY;return{xi:function(n){return n?e(n[e.dim]):t.xi(n)},yi:function(e){return e?n(e[n.dim]):t.yi(e)},sizeX:function(n){return n?e.stepSize(n[e.dim]):t.sizeX(n)},sizeY:function(e){return e?n.stepSize(e[n.dim]):t.sizeY(e)}}}]},e.prototype.createScreenModel=function(t){return t},e.prototype.allocateRect=function(t){var e=this,n=this.screenModel;return{slot:function(t){return e.config.options.container.selectAll(".uid_"+t)},left:n.xi(t)-n.sizeX(t)/2,top:n.yi(t)-n.sizeY(t)/2,width:n.sizeX(t),height:n.sizeY(t),containerWidth:this.W,containerHeight:this.H}},e.prototype.drawFrames=function(t){var e=Object.assign({},this.config),n=e.options,r=this.W,i=this.H;if(e.x=this.xScale,e.y=this.yScale,e.x.scaleObj=this.xScale,e.y.scaleObj=this.yScale,e.x.guide=e.guide.x,e.y.guide=e.guide.y,e.x.guide.label.size=r,e.y.guide.label.size=i,(n.container.attr("transform")?rt(n.container,this.config.guide.animationSpeed,"cartesianContainerTransform"):n.container).attr("transform",De(this.L,this.T)),e.x.guide.hide)this._removeDimAxis(n.container,e.x);else{var o="top"===e.x.guide.scaleOrient?[0,0-e.guide.x.padding]:[0,i+e.guide.x.padding];this._drawDimAxis(n.container,e.x,o,r)}if(e.y.guide.hide)this._removeDimAxis(n.container,e.y);else{var a="right"===e.y.guide.scaleOrient?[r+e.guide.y.padding,0]:[0-e.guide.y.padding,0];this._drawDimAxis(n.container,e.y,a,i)}var s=t.reduce(function(t,e){return t.concat((e.units||[]).map(function(t){return t.uid}))},[]),u=this._drawGrid(n.container,e,r,i,n),c=ct(u,".cell").data(s,function(t){return t});c.enter().append("g").attr("class",function(t){return Xe+"cell cell uid_"+t}).merge(c).classed("tau-active",!0),rt(c,this.config.guide.animationSpeed).attr("opacity",1),rt(c.exit().classed("tau-active",!1),this.config.guide.animationSpeed).attr("opacity",1e-6).remove()},e.prototype._drawDimAxis=function(t,e,n,r){var i=Cn.get(e.guide.tickFormat,e.guide.tickFormatNullAlias),a=_r({scale:e.scaleObj,scaleGuide:e.guide,ticksCount:i?Cr(r/e.guide.density):null,tickFormat:i||null,position:n}),s=this.config.guide.animationSpeed;Or(t,this._getAxisSelector(e)).classed("tau-active",!0).classed(e.guide.cssClass,!0).call(function(t){var e=rt(t,s,"axisTransition"),r=t.attr("transform");De.apply(o,n)!==r&&(r?e:t).attr("transform",De.apply(o,n)),e.call(a),e.attr("opacity",1)})},e.prototype._removeDimAxis=function(t,e){var n=ct(t,this._getAxisSelector(e)).classed("tau-active",!1);rt(n,this.config.guide.animationSpeed,"axisTransition").attr("opacity",1e-6).remove()},e.prototype._getAxisSelector=function(t){return"g."+("h"===Pe(t.guide.scaleOrient)?"x":"y")+".axis"},e.prototype._drawGrid=function(t,e,n,r){var i=this;return Or(t,"g.grid").attr("transform",De(0,0)).call(function(t){var o=t,a=i.config.guide.animationSpeed,s=(e.guide.showGridLines||"").toLowerCase();if(s.length>0){var u=Or(o,"g.grid-lines");if(s.indexOf("x")>-1){var c=e.x,l=Cn.get(c.guide.tickFormat),d=vr({scale:c.scaleObj,scaleGuide:c.guide,tickSize:r,ticksCount:l?Cr(n/c.guide.density):null}),p=Or(u,"g.grid-lines-x");rt(p,a).call(d)}if(s.indexOf("y")>-1){var f=e.y,h=(l=Cn.get(f.guide.tickFormat),vr({scale:f.scaleObj,scaleGuide:f.guide,tickSize:-n,ticksCount:l?Cr(r/f.guide.density):null})),m=Or(u,"g.grid-lines-y");rt(m,a).call(h)}}})},e}(y),Sr=n(11),Mr=n(8),Tr=(br=function(t,e){return(br=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}br(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),kr=function(){return(kr=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)},Ar=kr({},Sr,Mr,a),Dr=function(t){function e(e){var n=t.call(this,e)||this;n.config=e,n.config.guide=Jt(n.config.guide||{},{padding:{l:50,r:50,t:50,b:50},enableBrushing:!1}),n.columnsBrushes={},n.columnsSelections={},n.on("force-brush",function(t,e){return n._forceBrushing(e)});var r=n.config.options,i=n.config.guide.padding;return n.L=r.left+i.l,n.T=r.top+i.t,n.W=r.width-(i.l+i.r),n.H=r.height-(i.t+i.b),n}return Tr(e,t),e.prototype.defineGrammarModel=function(t){var e=this.config,n=this.W,r=this.H;this.columnsScalesMap=e.columns.reduce(function(e,n){return e[n]=t("pos",n,[0,r]),e},{});var i=n/(e.columns.length-1),o=e.columns.reduce(function(t,e,n){return t[e]=n*i,t},{});return this.xBase=function(t){return o[t]},this.regScale("columns",this.columnsScalesMap),{}},e.prototype.allocateRect=function(){var t=this;return{slot:function(e){return t.config.options.container.selectAll(".uid_"+e)},left:0,top:0,width:this.W,height:this.H,containerWidth:this.W,containerHeight:this.H}},e.prototype.drawFrames=function(t){var e=this,n=Object.assign({},this.config),r=n.options,i=this._fnDrawGrid(r.container,n,r.frameId,Object.keys(this.columnsScalesMap).reduce(function(t,n){return t.concat([e.columnsScalesMap[n].getHash()])},[]).join("_")),o=i.selectAll(".parent-frame-"+r.frameId).data(t,function(t){return t.hash()});o.exit().remove(),o.enter().append("g").attr("class",function(t){return Xe+"cell cell parent-frame-"+r.frameId+" frame-"+t.hash()}).merge(o).each(function(t){var e,n,i,o;e=r.frameId,n=Ar.select(this),i=t,(o=n.selectAll(".layer_"+e).data(i.units,function(t){return t.uid})).exit().remove(),o.enter().append("g").attr("class",function(t){return"layer_"+e+" uid_"+t.uid})});var a=this._fnDrawColumns(i,n);n.guide.enableBrushing&&this._enableBrushing(a)},e.prototype._fnDrawGrid=function(t,e,n,r){var i=t.selectAll(".grid_"+n).data([r],function(t){return t});return i.exit().remove(),i.enter().append("g").attr("class","grid grid_"+n).attr("transform",De(this.L,this.T)).merge(i)},e.prototype._fnDrawColumns=function(t,e){var n=e.guide.columns||{},r=this.xBase,i=this.columnsScalesMap,o=Ar.axisLeft(),a=t.selectAll(".column").data(e.columns,function(t){return t});return a.exit().remove(),a.enter().append("g").attr("class","column").attr("transform",function(t){return De(r(t),0)}).call(function(t){t.append("g").attr("class","y axis").each(function(t){var e=i[t].dim,r=o.scale(i[t]),a=n[e]||{},s=Cn.get(a.tickFormat,a.tickFormatNullAlias);null!==s&&r.tickFormat(s),Ar.select(this).call(r)}).append("text").attr("class","label").attr("text-anchor","middle").attr("y",-9).text(function(t){return((n[t]||{}).label||{}).text||i[t].dim})}).merge(a)},e.prototype._enableBrushing=function(t){var e=this,n=this.columnsSelections,r=this.columnsScalesMap,i=this.columnsBrushes,o=!0,s=function(t){return t},u=function(t){return t},c=function(){var t=Object.keys(i).find(function(t){return i[t]===a.event.target});if(n[t]=a.event.selection,o){var s=Object.keys(i).filter(function(t){return n[t]}).map(function(t){var e=[];if(r[t].discrete){var i=n[t];e=r[t].domain().filter(function(e){var n=r[t](e);return i[0]<=n&&i[1]>=n})}else{var o=n[t].map(r[t].invert);e=[o[0],o[1]]}return{dim:r[t].dim,func:r[t].discrete?"inset":"between",args:e}});e.fire("brush",s)}};return t.selectAll(".brush").remove(),t.append("g").attr("class","brush").each(function(t){var e=r[t].range();i[t]=Ar.brushY().extent([[0,e[0]],[16,e[1]]]).on("start",s).on("brush",c).on("end",u),Ar.select(this).classed("brush-"+Nt(t),!0).call(i[t]),o=!1,i[t].move(Ar.select(this),e),o=!0}).selectAll("rect").attr("transform","translate(-8,0)").attr("width",16),t},e.prototype._forceBrushing=function(t){void 0===t&&(t={});var e=this.columnsBrushes,n=this.columnsScalesMap;this.columnsSelections,Object.keys(t).filter(function(r){return e[r]&&n[r]&&t[r]}).forEach(function(r){var i=t[r],o=[];if(n[r].discrete){var a=i.map(n[r]).filter(function(t){return t>=0}),s=n[r].stepSize()/2;o=[Math.min.apply(Math,a)-s,Math.max.apply(Math,a)+s]}else o=[i[0],i[1]];var u=Nt(r);e[r](Ar.select(".brush-"+u)),e[r].move(Ar.select(".brush-"+u),o.map(n[r]))})},e}(y),Ir=n(12),Pr=n(4),Rr=function(){var t,e=[],n=[],r=1,i=1,o={},a=!1,s=function(t){var r=e.length,i=0,o=e[t].x-n[t].x,a=n[t].y-e[t].y,s=Math.sqrt(o*o+a*a);s>0&&(i+=.2*s),a/=s,i+=(o/=s)>0&&a>0?0:o<0&&a>0?3:o<0&&a<0?6:9;for(var u,c,d,p,f=e[t].x,h=e[t].y-e[t].height+2,m=e[t].x+e[t].width,g=e[t].y+2,y=0;y<r;y++)y!=t&&(l(n[t].x,e[t].x,n[y].x,e[y].x,n[t].y,e[t].y,n[y].y,e[y].y)&&(i+=1),u=e[y].x,d=e[y].y-e[y].height+2,c=e[y].x+e[y].width,p=e[y].y+2,i+=Math.max(0,Math.min(c,m)-Math.max(u,f))*Math.max(0,Math.min(p,g)-Math.max(d,h))*30),u=n[y].x-n[y].r,d=n[y].y-n[y].r,c=n[y].x+n[y].r,p=n[y].y+n[y].r,i+=Math.max(0,Math.min(c,m)-Math.max(u,f))*Math.max(0,Math.min(p,g)-Math.max(d,h))*30;return i},u=function(o){var u,c=Math.floor(Math.random()*e.length),l=e[c].x,d=e[c].y;u=a?t(c,e,n):s(c),e[c].x+=5*(Math.random()-.5),e[c].y+=5*(Math.random()-.5),e[c].x>r&&(e[c].x=l),e[c].x<0&&(e[c].x=l),e[c].y>i&&(e[c].y=d),e[c].y<0&&(e[c].y=d);var p=(a?t(c,e,n):s(c))-u;Math.random()<Math.exp(-p/o)||(e[c].x=l,e[c].y=d)},c=function(o){var u,c=Math.floor(Math.random()*e.length),l=e[c].x,d=e[c].y;u=a?t(c,e,n):s(c);var p=.5*(Math.random()-.5),f=Math.sin(p),h=Math.cos(p);e[c].x-=n[c].x,e[c].y-=n[c].y;var m=e[c].x*h-e[c].y*f,g=e[c].x*f+e[c].y*h;e[c].x=m+n[c].x,e[c].y=g+n[c].y,e[c].x>r&&(e[c].x=l),e[c].x<0&&(e[c].x=l),e[c].y>i&&(e[c].y=d),e[c].y<0&&(e[c].y=d);var y=(a?t(c,e,n):s(c))-u;Math.random()<Math.exp(-y/o)||(e[c].x=l,e[c].y=d)},l=function(t,e,n,r,i,o,a,s){var u,c,l;return c=((e-t)*(i-a)-(o-i)*(t-n))/(l=(s-a)*(e-t)-(r-n)*(o-i)),!((u=((r-n)*(i-a)-(s-a)*(t-n))/l)<0||u>1||c<0||c>1)},d=function(t,e,n){return t-e/n};return o.start=function(t){for(var n=e.length,r=1,i=0;i<t;i++){for(var o=0;o<n;o++)Math.random()<.5?u(r):c(r);r=d(r,1,t)}},o.width=function(t){return arguments.length?(r=t,o):r},o.height=function(t){return arguments.length?(i=t,o):i},o.label=function(t){return arguments.length?(e=t,o):e},o.anchor=function(t){return arguments.length?(n=t,o):n},o.alt_energy=function(e){return arguments.length?(t=e,a=!0,o):s},o.alt_schedule=function(){return arguments.length?o:d},o},Nr=(wr=function(t,e){return(wr=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}wr(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),Fr=function(){return(Fr=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)},Lr=Fr({},s,Ir,a),Br=["land","continents","georegions","countries","regions","subunits","states","counties"],Vr=function(t){function e(e){var n=t.call(this,e)||this;return n.config=e,n.config.guide=Jt(n.config.guide||{},{defaultFill:"rgba(128,128,128,0.25)",padding:{l:0,r:0,t:0,b:0},showNames:!0}),n.contourToFill=null,n.on("highlight-area",function(t,e){return n._highlightArea(e)}),n.on("highlight-point",function(t,e){return n._highlightPoint(e)}),n.on("highlight",function(t,e){return n._highlightPoint(e)}),n}return Nr(e,t),e.prototype.defineGrammarModel=function(t){var e=this.config,n=e.options,r=e.guide.padding,i=n.width-(r.l+r.r),o=n.height-(r.t+r.b);return this.latScale=t("pos",e.latitude,[0,o]),this.lonScale=t("pos",e.longitude,[i,0]),this.sizeScale=t("size",e.size),this.colorScale=t("color",e.color),this.codeScale=t("value",e.code),this.fillScale=t("fill",e.fill),this.W=i,this.H=o,this.regScale("latitude",this.latScale).regScale("longitude",this.lonScale).regScale("size",this.sizeScale).regScale("color",this.colorScale).regScale("code",this.codeScale).regScale("fill",this.fillScale),{}},e.prototype.drawFrames=function(t){var e,n,r,i=this,o=this.config.guide;"string"==typeof o.sourcemap?(e=o.sourcemap,n=function(e,n){if(e)throw e;i._drawMap(t,n)},(r=new XMLHttpRequest).onload=function(){if(r.status>=200&&r.status<300){var t=r.responseText;try{n(null,JSON.parse(t))}catch(t){n(t,null)}}else n(new Error(r.status+": "+r.statusText),null)},r.onerror=function(t){n(t.error,null)},r.open("GET",e,!0),r.send(null)):this._drawMap(t,o.sourcemap)},e.prototype._calcLabels=function(t,e,n){var r=this.W,i=this.H,o={};return e.forEach(function(e){var a=(Pr.feature(t,t.objects[e]).features||[]).map(function(t){var i=t.properties||{},o=n.centroid(t),a=n.bounds(t),s=o[0],u=o[1],c=a[1][0]-a[0][0],l=i.name||"",d=i.abbr||l,p=c<5.5*l.length?d:l,f=c<13.75,h=f?r-s-16.5:0;return{id:e+"-"+t.id,sx:s,sy:u,x:s+h,y:u,width:5.5*p.length,height:10,name:p,r:h,isRef:f}}).filter(function(t){return!Number.isNaN(t.x)&&!Number.isNaN(t.y)}),s=a.map(function(t){return{x:t.sx,y:t.sy,r:t.r}});Rr().label(a).anchor(s).width(r).height(i).start(10),a.filter(function(t){return!t.isRef}).map(function(t){return t.x=t.sx,t.y=t.sy,t}).reduce(function(t,e){return t[e.id]=e,t},o);var u=a.filter(function(t){return t.isRef});u.length<6&&u.reduce(function(t,e){return t[e.id]=e,t},o)}),o},e.prototype._drawMap=function(t,e){var n,r,i=this,o=this,s=this.config.guide,u=this.config.options,c=this.config.options.container,l=this.latScale,d=this.lonScale,p=this.sizeScale,f=this.colorScale,h=this.codeScale,m=this.fillScale,g=this.W,y=this.H,_=Br.filter(function(t){return(e.objects||{}).hasOwnProperty(t)});if(0===_.length)throw new Error("Invalid map: should contain some contours");if(m.dim){if(!h.georole)throw console.log("Specify [georole] for code scale"),new Error("[georole] is missing");if(-1===_.indexOf(h.georole))throw console.log('There is no contour for georole "'+h.georole+'"'),console.log("Available contours are: "+_.join(" | ")),new Error("Invalid [georole]");n=h.georole}else n=_[_.length-1];if(this.contourToFill=n,l.dim&&d.dim){var v=Lr.extent(l.domain()),b=Lr.extent(d.domain());r=[(b[1]+b[0])/2,(v[1]+v[0])/2]}var x=this._createProjection(e,_[0],r),w=Lr.geoPath().projection(x),j=c.selectAll(".map-container").data([""+g+y+r+_.join("-")],function(t){return t});j.exit().remove();var O=j.enter().append("g").call(function(t){var n=t;n.attr("class","map-container");var r={},i=_.reduceRight(function(t,e){return t.concat(e)},[]);if(s.showNames&&(r=o._calcLabels(e,i,w)),i.forEach(function(t,i){var o=function(e){return r[t+"-"+e.id]};n.selectAll(".map-contour-"+t).data(Pr.feature(e,e.objects[t]).features||[]).enter().append("g").call(function(e){var n=e;n.attr("class","map-contour-"+t+" map-contour-level map-contour-level-"+i).attr("fill","none"),n.append("title").text(function(t){return(t.properties||{}).name}),n.append("path").attr("d",w),n.append("text").attr("class","place-label-"+t).attr("transform",function(t){var e=o(t);return e?"translate("+[e.x,e.y]+")":""}).text(function(t){var e=o(t);return e?e.name:""}),n.append("line").attr("class","place-label-link-"+t).attr("stroke","gray").attr("stroke-width",.25).attr("x1",function(t){var e=o(t);return e&&e.isRef?e.sx:0}).attr("y1",function(t){var e=o(t);return e&&e.isRef?e.sy:0}).attr("x2",function(t){var e=o(t);return e&&e.isRef?e.x-.6*e.name.length*5.5:0}).attr("y2",function(t){var e=o(t);return e&&e.isRef?e.y-3.5:0})})}),e.objects.hasOwnProperty("places")){var a=Pr.feature(e,e.objects.places),u=a.features.map(function(t){var e=x(t.geometry.coordinates);return{x:e[0]+3.5,y:e[1]+3.5,width:5.5*t.properties.name.length,height:12,name:t.properties.name}}),c=a.features.map(function(t){var e=x(t.geometry.coordinates);return{x:e[0],y:e[1],r:2.5}});Rr().label(u).anchor(c).width(g).height(y).start(100),n.selectAll(".place").data(c).enter().append("circle").attr("class","place").attr("transform",function(t){return"translate("+t.x+","+t.y+")"}).attr("r",function(t){return t.r+"px"}),n.selectAll(".place-label").data(u).enter().append("text").attr("class","place-label").attr("transform",function(t){return"translate("+t.x+","+t.y+")"}).text(function(t){return t.name})}}).merge(j);this.groupByCode=t.reduce(function(t,e){return e.part().reduce(function(t,e){return t[(e[h.dim]||"").toLowerCase()]=e,t},t)},{});var C=this._resolveFeature.bind(this);if(O.selectAll(".map-contour-"+n).data(Pr.feature(e,e.objects[n]).features).call(function(t){t.classed("map-contour",!0).attr("fill",function(t){var e=C(t);return null===e?s.defaultFill:m(e[m.dim])})}).on("mouseover",function(t){return i.fire("area-mouseover",{data:C(t),event:a.event})}).on("mouseout",function(t){return i.fire("area-mouseout",{data:C(t),event:a.event})}).on("click",function(t){return i.fire("area-click",{data:C(t),event:a.event})}),!l.dim||!d.dim)return[];var E=function(t){return t.attr("r",function(t){return p(t[p.dim])}).attr("transform",function(t){var e=t.data;return"translate("+x([e[d.dim],e[l.dim]])+")"}).attr("class",function(t){var e=t.data;return f(e[f.dim])}).attr("opacity",.5).on("mouseover",function(t){var e=t.data;return o.fire("point-mouseover",{data:e,event:a.event})}).on("mouseout",function(t){var e=t.data;return o.fire("point-mouseout",{data:e,event:a.event})}).on("click",function(t){var e=t.data;return o.fire("point-click",{data:e,event:a.event})})},S=O.selectAll(".frame").data(t.map(function(t){return{tags:t.key||{},hash:t.hash(),data:t.part()}}),function(t){return t.hash});return S.exit().remove(),S.enter().append("g").merge(S).call(function(t){t.attr("class",function(t){return"frame frame-"+t.hash}).call(function(t){var e=t.selectAll("circle").data(function(t){return t.data.map(function(t){return{data:t,uid:u.uid}})});e.exit().remove(),e.call(E),e.enter().append("circle").call(E)})}),[]},e.prototype._resolveFeature=function(t){var e,n=this.groupByCode,r=t.properties,i=["c1","c2","c3","abbr","name"].filter(function(t){return r.hasOwnProperty(t)&&r[t]&&n.hasOwnProperty(r[t].toLowerCase())});if(0===i.length)e=null;else if(i.length>0){var o=r[i[0]].toLowerCase();e=n[o]}return e},e.prototype._highlightArea=function(t){var e=this,n=this.config.options.container,r=this.contourToFill;n.selectAll(".map-contour-"+r).classed("map-contour-highlighted",function(n){return t(e._resolveFeature(n))})},e.prototype._highlightPoint=function(t){this.config.options.container.selectAll("circle").classed("map-point-highlighted",function(e){var n=e.data;return t(n)}).attr("opacity",function(e){var n=e.data;return t(n)?.5:.1})},e.prototype._createProjection=function(t,e,n){var r=this.W,i=this.H,o=this.config.guide,a=100,s=[r/2,i/2],u=n||t.center,c=o.projection||t.projection||"mercator",l=this._createD3Projection(c,u,a,s),d=Lr.geoPath().projection(l).bounds(Pr.feature(t,t.objects[e])),p=a*r/(d[1][0]-d[0][0]),f=a*i/(d[1][1]-d[0][1]);return a=p<f?p:f,s=[r-(d[0][0]+d[1][0])/2,i-(d[0][1]+d[1][1])/2],this._createD3Projection(c,u,a,s)},e.prototype._createD3Projection=function(t,e,n,r){var i="geo"+t.substring(0,1).toUpperCase()+t.substring(1),o=Lr[i];if(!o)throw console.log('Unknown projection "'+t+'"'),console.log("See available projection types here: https://github.com/mbostock/d3/wiki/Geo-Projections"),new Error('Invalid map: unknown projection "'+t+'"');var a=o();return[{method:"scale",args:n},{method:"center",args:e},{method:"translate",args:r}].filter(function(t){return t.args}).reduce(function(t,e){return t[e.method]&&(t=t[e.method](e.args)),t},a)},e}(y),zr=function(t){return function(){return t}},Hr=function(){function t(t){this.model=t.model,this.x=t.x||zr(0),this.y=t.y||zr(0),this.dx=t.dx||zr(0),this.dy=t.dy||zr(0),this.w=t.w||zr(0),this.h=t.h||zr(0),this.hide=t.hide||zr(!1),this.label=t.label||zr(""),this.color=t.color||zr(""),this.angle=t.angle||zr(0),this.labelLinesAndSeparator=t.labelLinesAndSeparator||zr({lines:[],linesWidths:[],separator:""})}return t.seed=function(e,n){var r=n.fontColor,i=n.flip,o=n.formatter,a=n.labelRectSize,s=n.paddingKoeff,u=void 0===s?.5:s,c=n.lineBreakAvailable,l=n.lineBreakSeparator,d=i?e.yi:e.xi,p=i?e.xi:e.yi,f=function(t){return o(e.label(t))},h=function(t){var e=c?f(t).split(l):[f(t)];return{lines:e,linesWidths:e.map(function(t){return a([t]).width}),separator:l}};return new t({model:e,x:function(t){return d(t)},y:function(t){return p(t)},dy:function(t){return a(h(t).lines).height*u},w:function(t){return a(h(t).lines).width},h:function(t){return a(h(t).lines).height},label:f,labelLinesAndSeparator:h,color:function(){return r},angle:function(){return 0}})},t.compose=function(e,n){return void 0===n&&(n={}),Object.keys(n).reduce(function(t,e){return t[e]=n[e],t},new t(e))},t}(),Ur={},Wr=function(){function t(){}return t.regRule=function(t,e){return Ur[t]=e,this},t.getRule=function(t){return Ur[t]},t}(),qr=function(t,e){var n=t.lines,r=t.linesWidths,i=t.separator;return n.map(function(t,n){var i,o,a,s,u=(i=t,o=r[n],(a=e)<o?Math.max(1,Math.floor(a*i.length/o))-1:i.length);return u<t.length?0===(s=u)?"":t.slice(0,s).replace(/\.+$/g,"")+"…":t}).join(i)},Gr=function(t,e){var n=5-(t.labelLinesAndSeparator(e).lines.length-1);return n<1?1:n},Yr=function(t,e){return t.discrete||!t.discrete&&e[t.dim]>=0},Kr=function(t,e){return!t.discrete&&e[t.dim]<0},Xr=function(t,e){return t.w(e)/2+Math.floor(t.model.size(e)/Gr(t,e))},Qr=function(t,e){return t.h(e)/2+Math.floor(t.model.size(e)/Gr(t,e))},Zr=function(t){return function(e){return{dx:function(n){var r=e.model.scaleY;if("+"===t[2]&&!Yr(r,n))return e.dx(n);if("-"===t[2]&&!Kr(r,n))return e.dx(n);var i=t[1],o=t[0]===t[0].toUpperCase()?1:0;return e.dx(n)+i*o*e.model.size(n)/2+i*Xr(e,n)}}}},$r=function(t){return function(e){return{dy:function(n){var r=e.model.scaleY;if("+"===t[2]&&!Yr(r,n))return e.dy(n);if("-"===t[2]&&!Kr(r,n))return e.dy(n);var i=t[1],o=t[0]===t[0].toUpperCase()?1:0;return e.dy(n)+i*o*e.model.size(n)/2+i*Qr(e,n)}}}};Wr.regRule("l",Zr(["l",-1,null])).regRule("L",Zr(["L",-1,null])).regRule("l+",Zr(["l",-1,"+"])).regRule("l-",Zr(["l",-1,"-"])).regRule("L+",Zr(["L",-1,"+"])).regRule("L-",Zr(["L",-1,"-"])).regRule("r",Zr(["r",1,null])).regRule("R",Zr(["R",1,null])).regRule("r+",Zr(["r",1,"+"])).regRule("r-",Zr(["r",1,"-"])).regRule("R+",Zr(["R",1,"+"])).regRule("R-",Zr(["R",1,"-"])).regRule("t",$r(["t",-1,null])).regRule("T",$r(["T",-1,null])).regRule("t+",$r(["t",-1,"+"])).regRule("t-",$r(["t",-1,"-"])).regRule("T+",$r(["T",-1,"+"])).regRule("T-",$r(["T",-1,"-"])).regRule("b",$r(["b",1,null])).regRule("B",$r(["B",1,null])).regRule("b+",$r(["b",1,"+"])).regRule("b-",$r(["b",1,"-"])).regRule("B+",$r(["B",1,"+"])).regRule("B-",$r(["B",1,"-"])).regRule("rotate-on-size-overflow",function(t,e){var n=e.data,r=e.lineBreakAvailable,i={};if(n.reduce(function(e,n){return e+(r=n,t.model.size(r)<t.w(r)?1:0);var r},0)/n.length>.5){var o=r?-.5:.5;i={angle:function(){return-90},w:function(e){return t.h(e)},h:function(e){return t.w(e)},dx:function(e){return t.h(e)*o-2},dy:function(){return 0}}}return i}).regRule("hide-by-label-height-vertical",function(t){return{hide:function(e){var n,r;return 0===t.angle(e)?(r=t.h(e),n=Math.abs(t.model.y0(e)-t.model.yi(e))):(r=t.w(e),n=t.model.size(e)),r>n||t.hide(e)}}}).regRule("cut-label-vertical",function(t){return{h:function(e){var n=t.h(e);if(Math.abs(t.angle(e))>0){var r=Math.abs(t.model.y0(e)-t.model.yi(e));return r<n?r:n}return n},w:function(e){var n=t.w(e);if(0===t.angle(e)){var r=t.model.size(e);return r<n?r:n}return n},label:function(e){var n;return n=0===t.angle(e)?t.model.size(e):Math.abs(t.model.y0(e)-t.model.yi(e)),qr(t.labelLinesAndSeparator(e),n)},dy:function(e){var n=t.dy(e);if(0!==t.angle(e)){var r=t.h(e),i=Math.abs(t.model.y0(e)-t.model.yi(e));return i<r?i*n/r:n}return n}}}).regRule("cut-outer-label-vertical",function(t){return{h:function(e,n){var r=t.h(e);if(Math.abs(t.angle(e))>0){var i=t.model.y0(e)<t.model.yi(e)?n.maxHeight-t.model.yi(e):t.model.yi(e);return i<r?i:r}return r},w:function(e){var n=t.w(e);if(0===t.angle(e)){var r=t.model.size(e);return r<n?r:n}return n},label:function(e,n){var r;return r=0===t.angle(e)?t.model.size(e):t.model.y0(e)<t.model.yi(e)?n.maxHeight-t.model.yi(e):t.model.yi(e),qr(t.labelLinesAndSeparator(e),r)},dy:function(e,n){var r=t.dy(e);if(0!==t.angle(e)){var i=t.h(e),o=t.model.y0(e)<t.model.yi(e)?n.maxHeight-t.model.yi(e):t.model.yi(e);return o<i?o*r/i:r}return r}}}).regRule("from-beginning",function(t){var e=function(e){return t.model.y0(e)};return t.model.flip?{x:e}:{y:e}}).regRule("to-end",function(t){var e=function(e){return t.model.yi(e)};return t.model.flip?{x:e}:{y:e}}).regRule("towards",function(t){var e=t.model.flip?Xr:Qr,n=function(n){return i=n,((r=t).model.yi(i)-r.model.y0(i)>=0?1:-1)*e(t,n);var r,i};return t.model.flip?{dx:n}:{dy:n}}).regRule("inside-start-then-outside-end-horizontal",function(t,e){var n=[Wr.getRule("from-beginning"),Wr.getRule("towards"),Wr.getRule("cut-label-horizontal")].reduce(function(t,n){return Hr.compose(t,n(t,e))},t),r=[Wr.getRule("to-end"),Wr.getRule("towards"),Wr.getRule("cut-outer-label-horizontal")].reduce(function(t,n){return Hr.compose(t,n(t,e))},t);return Object.assign({},n,["x","dx","hide","label"].reduce(function(t,e){return t[e]=function(t){return(i=t,n.label(i).length>=r.label(i).length?n:r)[e](t);var i},t},{}))}).regRule("inside-start-then-outside-end-vertical",function(t,e){var n=[Wr.getRule("from-beginning"),Wr.getRule("towards"),Wr.getRule("cut-label-vertical")].reduce(function(t,n){return Hr.compose(t,n(t,e))},t),r=[Wr.getRule("to-end"),Wr.getRule("towards"),Wr.getRule("cut-outer-label-vertical")].reduce(function(t,n){return Hr.compose(t,n(t,e))},t);return Object.assign({},n,["y","dy","hide","label"].reduce(function(t,e){return t[e]=function(t){return(i=t,n.label(i).length>=r.label(i).length?n:r)[e](t);var i},t},{}))}).regRule("outside-then-inside-horizontal",function(t,e){var n=["r+","l-","cut-outer-label-horizontal"].map(Wr.getRule).reduce(function(t,n){return Hr.compose(t,n(t,e))},t),r=["r-","l+","hide-by-label-height-horizontal","cut-label-horizontal"].map(Wr.getRule).reduce(function(t,n){return Hr.compose(t,n(t,e))},t);return Object.assign({},n,["x","dx","hide","label"].reduce(function(t,e){return t[e]=function(t){return(i=t,r.label(i).length>n.label(i).length?r:n)[e](t);var i},t},{}))}).regRule("outside-then-inside-vertical",function(t,e){var n=["t+","b-","cut-outer-label-vertical"].map(Wr.getRule).reduce(function(t,n){return Hr.compose(t,n(t,e))},t),r=["t-","b+","hide-by-label-height-vertical","cut-label-vertical"].map(Wr.getRule).reduce(function(t,n){return Hr.compose(t,n(t,e))},t);return Object.assign({},n,["y","dy","hide","label"].reduce(function(t,i){return t[i]=function(t){return(o=t,a=n.y(o,e)+n.dy(o,e),a<=0||a>=e.maxHeight?r:n)[i](t,e);var o,a},t},{}))}).regRule("hide-by-label-height-horizontal",function(t){return{hide:function(e){return t.model.size(e)<t.h(e)||t.hide(e)}}}).regRule("cut-label-horizontal",function(t){return{dx:function(e){var n=t.w(e),r=Math.abs(t.model.y0(e)-t.model.yi(e)),i=t.dx(e);return r<n?r*i/n:i},w:function(e){var n=t.w(e),r=Math.abs(t.model.y0(e)-t.model.yi(e));return r<n?r:n},label:function(e){var n=Math.abs(t.model.y0(e)-t.model.yi(e));return qr(t.labelLinesAndSeparator(e),n)}}}).regRule("cut-outer-label-horizontal",function(t,e){return{dx:function(n){var r=t.w(n),i=t.model.y0(n)<t.model.yi(n)?e.maxWidth-t.model.yi(n):t.model.yi(n),o=t.dx(n);return i<r?i*o/r:o},w:function(n){var r=t.w(n),i=t.model.y0(n)<t.model.yi(n)?e.maxWidth-t.model.yi(n):t.model.yi(n);return i<r?i:r},label:function(n){var r=t.model.y0(n)<t.model.yi(n)?e.maxWidth-t.model.yi(n):t.model.yi(n);return qr(t.labelLinesAndSeparator(n),r)}}}).regRule("keep-within-diameter-or-top",function(t){return{dy:function(e){return t.model.size(e)/t.w(e)<1?t.dy(e)-t.h(e)/2-t.model.size(e)/2:t.dy(e)}}}).regRule("keep-in-box",function(t,e){var n=e.maxWidth,r=e.maxHeight;return{dx:function(e){var r=t.dx(e),i=t.x(e)+r,o=t.w(e),a=0-(i-o/2);if(a>0)return r+a;var s=i+o/2-n;return s>0?r-s:r},dy:function(e){var n=t.dy(e),i=t.y(e)+n,o=t.h(e);if(0-(i-o/2)>0)return 0;var a=i+o/2-r;return a>0?n-a:n}}}).regRule("multiline-label-left-align",function(t){return{dy:function(e){var n=t.dy(e);return-90===t.angle(e)?n+t.h(e)/2:n}}}).regRule("multiline-label-vertical-center-align",function(t){return{dy:function(e){return t.dy(e)-t.h(e)/2}}}).regRule("multiline-hide-on-container-overflow",function(t,e){var n=e.maxWidth,r=e.maxHeight;return{hide:function(e){var i=t.angle(e),o=t.x(e)+t.dx(e),a=t.y(e)+t.dy(e);return!(!Gt(o,t.w(e),i,n)&&!Yt(a,t.h(e),i,r))||t.hide(e)}}});var Jr=function(){function t(t){this.minError=Number.MAX_VALUE,this.items=t.items,this.revision=this.items.map(function(t){return{i:t.i,x:t.x,y:t.y}}),this.penalties=t.penalties,this.transactor=t.transactor,this.cooling_schedule=t.cooling_schedule||function(t,e,n){return t-e/n}}return t.prototype.energy=function(t){return this.penalties.reduce(function(e,n){return e+n(t)},0)},t.prototype.move=function(t){var e=Math.floor(Math.random()*this.items.length),n=this.transactor(this.items[e]),r=this.energy(e);this.items[e]=n.modify();var i=this.energy(e),o=i-r,a=o<0?1:Math.exp(-o/t);Math.random()>=a?this.items[e]=n.revert():i<this.minError&&(this.minError=i,this.revision=this.items.map(function(t){return{i:t.i,x:t.x,y:t.y}}))},t.prototype.start=function(t){var e=1,n=this.items.length;t:for(var r=0;r<t;r++){for(var i=0;i<n;i++)if(this.move(e),this.minError<=10)break t;e=this.cooling_schedule(e,1,t)}return this.revision},t}(),ti=function(t,e,n,r,i,o,a,s){return Ne(t,i,e,o,n,a,r,s)},ei={},ni=function(){function t(){}return t.reg=function(t,e){return ei[t]=e,this},t.get=function(t){return ei[t]},t}();ni.reg("auto:avoid-label-label-overlap",function(t,e,n){return void 0===n&&(n=1),function(e){var r=t[e].x,i=t[e].y-t[e].h+2,o=t[e].x+t[e].w,a=t[e].y+2;return t.reduce(function(t,s,u){var c=Number(u!==e),l=s.x,d=s.y-s.h+2,p=s.x+s.w,f=s.y+2;return t+c*(Math.max(0,Math.min(p,o)-Math.max(l,r))*Math.max(0,Math.min(f,a)-Math.max(d,i))*n)},0)}}).reg("auto:avoid-label-anchor-overlap",function(t,e,n){return void 0===n&&(n=1),function(e){var r=t[e],i=r.x-r.w/2,o=r.x+r.w/2,a=r.y-r.h/2+2,s=r.y+r.h/2+2;return t.reduce(function(t,e){var r=e.x0-e.size/2,u=e.x0+e.size/2,c=e.y0-e.size/2,l=e.y0+e.size/2;return t+Math.max(0,Math.min(u,o)-Math.max(r,i))*Math.max(0,Math.min(l,s)-Math.max(c,a))*n},0)}}).reg("auto:avoid-label-edges-overlap",function(t,e,n){return void 0===n&&(n=1),function(r){var i=t[r],o=i.x-i.w/2,a=i.x+i.w/2,s=i.y-i.h/2,u=i.y+i.h/2;return e.reduce(function(t,e){var r=ti(o,a,e.x0,e.x1,s,u,e.y0,e.y1),i=ti(o,a,e.x0,e.x1,u,s,e.y0,e.y1);return t+(Number(r)+Number(i))*n},0)}});var ri=function(t,e,n,r,i,o,a,s){return Ne(t,i,e,o,n,a,r,s)},ii=function(){function t(t,e,n,r){var i=r.width,o=r.height,a=r.container;this.container=a,this.model=t,this.flip=e,this.w=i,this.h=o,this.guide=Jt(n||{},{fontFamily:"Helvetica Neue, Segoe UI, Open Sans, Ubuntu, sans-serif",fontWeight:"normal",fontSize:10,fontColor:"#000",hideEqualLabels:!1,lineBreak:!1,lineBreakSeparator:"",position:[],tickFormat:null,tickFormatNullAlias:""})}return t.prototype.draw=function(t){var e=this,n=this.model,r=this.guide,i=r.lineBreak,o=r.lineBreakSeparator,s=Hr.seed(n,{fontColor:r.fontColor,flip:e.flip,formatter:Cn.get(r.tickFormat,r.tickFormatNullAlias),labelRectSize:function(t){return Oe(t,r)},lineBreakAvailable:i,lineBreakSeparator:o,paddingKoeff:i?0:.5}),u={lineBreakAvailable:i,maxWidth:e.w,maxHeight:e.h,data:t.reduce(function(t,e){return t.concat(e)},[])},c=this.applyFixedPositionRules(r,u,s,i,this.flip),l=t.reduce(function(t,e){var n=e.map(function(t){return{data:t,x:c.x(t)+c.dx(t),y:c.y(t)+c.dy(t),w:c.w(t),h:c.h(t,u),hide:c.hide(t),extr:null,size:c.model.size(t),angle:c.angle(t),label:c.label(t),labelLinesAndSeparator:c.labelLinesAndSeparator(t),color:c.color(t)}});return t.text=t.text.concat(n),t.edges=t.edges.concat(function(t,e){for(var n=t.length-1,r=[],i=0;i<=n;i++){var o=0===i?i:i-1,a=i,s=i===n?i:i+1;r.push((u=t[o],c=t[a],l=t[s],c.y===Math.max(c.y,u.y,l.y)?c.extr="min":c.y===Math.min(c.y,u.y,l.y)?c.extr="max":c.extr="norm",{x0:u.x,x1:c.x,y0:u.y,y1:c.y}))}var u,c,l;return r}(n)),t},{text:[],edges:[]});l.text=l.text.filter(function(t){return t.label}).map(function(t,e){return Object.assign(t,{i:e})});var d=i?["auto:hide-on-label-label-overlap","auto:adjust-on-multiline-label-overflow"]:this.guide.position,p=d.filter(function(t){return 0===t.indexOf("auto:avoid")});l=l.text.length>0&&p.length>0?this.autoPosition(l,p):l;var f=d.reduce(function(t,e){var n;return Object.assign(t,((n={})[e]=!0,n))},{});l.text=l.text=f["auto:adjust-on-label-overflow"]?this.adjustOnOverflow(l.text,u):l.text,l.text=l.text=f["auto:adjust-on-multiline-label-overflow"]?this.adjustOnMultilineOverflow(l.text,u):l.text,l.text=f["auto:hide-on-label-edges-overlap"]?this.hideOnLabelEdgesOverlap(l.text,l.edges):l.text,l.text=f["auto:hide-on-label-label-overlap"]?this.hideOnLabelLabelOverlap(l.text):l.text,l.text=f["auto:hide-on-label-anchor-overlap"]?this.hideOnLabelAnchorOverlap(l.text):l.text;var h=l.text,m=function(t){return function(e,n){return h[n][t]}},g=m("x"),y=m("y"),_=m("angle"),v=m("color"),b=m("label"),x=function(t){t.style("fill",v).style("font-size",e.guide.fontSize+"px").style("display",function(t,e){return h[e].hide?"none":null}).attr("class","i-role-label").attr("text-anchor","middle").attr("transform",function(t,e){return"translate("+g(t,e)+","+y(t,e)+") rotate("+_(t,e)+")"}),i?t.each(function(t,e){var n=a.select(this),r=_(t,e);n.text(null),b(t,e).split(o).forEach(function(t,e){n.append("tspan").attr("text-anchor",0!==r?"start":"middle").attr("x",0).attr("y",0).attr("dy",1.2*(e+1)+"em").text(t)})}):t.text(b)};r.hideEqualLabels&&h.filter(function(t){return!t.hide}).filter(function(t,e,n){return e<n.length-1&&t.label===n[e+1].label}).forEach(function(t){return t.hide=!0});var w=this.container.selectAll(".i-role-label").data(h.map(function(t){return t.data}));return w.exit().remove(),w.call(x),w.enter().append("text").call(x),w},t.prototype.applyFixedPositionRules=function(t,e,n,r,i){var o=t.position.filter(function(t){return-1===t.indexOf("auto:")});return r&&(i&&o.push("multiline-label-vertical-center-align"),o.push("multiline-label-left-align","multiline-hide-on-container-overflow")),o.map(Wr.getRule).reduce(function(t,n){return Hr.compose(t,n(t,e))},n)},t.prototype.autoPosition=function(t,e){var n=function(t,e){var n=4+(t.size+t.w)/2,r=2+(t.size+t.h)/2;return{x:n*Math.cos(e),y:r*Math.sin(e)}},r=t.edges,i=t.text.map(function(t){var e={max:-Math.PI/2,min:Math.PI/2,norm:Math.random()*Math.PI*2},r=n(t,e[t.extr]);return{i:t.i,x0:t.x,y0:t.y,x:t.x+r.x,y:t.y+r.y,w:t.w,h:t.h,size:t.size,hide:t.hide,extr:t.extr}}).filter(function(t){return!t.hide}),o=new Jr({items:i,transactor:function(t){var e=t.x,r=t.y;return{modify:function(){var e={max:-Math.PI,min:Math.PI,norm:2*Math.PI}[t.extr],r=e/4+Math.random()*(2*e)/4,i=n(t,r);return t.x=t.x0+i.x,t.y=t.y0+i.y,t},revert:function(){return t.x=e,t.y=r,t}}},penalties:e.map(function(t){return ni.get(t)}).filter(function(t){return t}).map(function(t){return t(i,r)})}).start(5);return t.text=o.reduce(function(t,e){var n=t[e.i];return n.x=e.x,n.y=e.y,t},t.text),t},t.prototype.hideOnLabelEdgesOverlap=function(t,e){var n=this;return t.filter(function(t){return!t.hide}).forEach(function(t){var r,i,o;r=t,i=e,o=n.getLabelRect(r),i.reduce(function(t,e){var n=ri(o.x0,o.x1,e.x0,e.x1,o.y0,o.y1,e.y0,e.y1),r=ri(o.x0,o.x1,e.x0,e.x1,o.y1,o.y0,e.y0,e.y1);return t+2*(Number(n)+Number(r))},0)>0&&(t.hide=!0)}),t},t.prototype.hideOnLabelLabelOverlap=function(t){var e=this,n={min:0,max:1,norm:2},r={"min/min":function(t,e){return e.y-t.y},"max/max":function(t,e){return t.y-e.y},"min/max":function(){return-1},"min/norm":function(){return-1},"max/norm":function(){return-1},"norm/norm":function(t,e){return t.y-e.y}};return t.filter(function(t){return!t.hide}).sort(function(t,e){return n[t.extr]-n[e.extr]}).forEach(function(i){t.forEach(function(t){i.i!==t.i&&function(t,i){var o=e.getLabelRect(t),a=e.getLabelRect(i),s=Number(!t.hide&&!i.hide);if(s*Math.max(0,Math.min(a.x1,o.x1)-Math.max(o.x0,a.x0))*(s*Math.max(0,Math.min(a.y1,o.y1)-Math.max(o.y0,a.y0)))>0){var u=[t,i];u.sort(function(t,e){return n[t.extr]-n[e.extr]}),(r[u[0].extr+"/"+u[1].extr](u[0],u[1])<0?u[0]:u[1]).hide=!0}}(i,t)})}),t},t.prototype.getLabelRect=function(t,e){return void 0===e&&(e=0),{x0:t.x-t.w/2-e,x1:t.x+t.w/2+e,y0:t.y-t.h/2-e,y1:t.y+t.h/2+e}},t.prototype.getPointRect=function(t,e){return void 0===e&&(e=0),{x0:t.x-t.size/2-e,x1:t.x+t.size/2+e,y0:t.y-t.size/2-e,y1:t.y+t.size/2+e}},t.prototype.hideOnLabelAnchorOverlap=function(t){var e=this;return t.filter(function(t){return!t.hide}).forEach(function(n){for(var r=t.length,i=0;i<r;i++){var o=t[i];if(n.i!==o.i&&(a=n,s=o,u=e.getLabelRect(a,2),c=e.getPointRect(s,2),Math.max(0,Math.min(c.x1,u.x1)-Math.max(c.x0,u.x0))*Math.max(0,Math.min(c.y1,u.y1)-Math.max(c.y0,u.y0))>.001)){n.hide=!0;break}}var a,s,u,c}),t},t.prototype.adjustOnOverflow=function(t,e){var n=e.maxWidth,r=e.maxHeight;return t.map(function(t){return t.hide||(t.x=Math.min(Math.max(t.x,t.w/2),n-t.w/2),t.y=Math.max(Math.min(t.y,r-t.h/2),t.h/2)),t})},t.prototype.adjustOnMultilineOverflow=function(t,e){var n=e.maxWidth;return t.map(function(t){return t.hide||0!==t.angle||(t.x=Math.min(Math.max(t.x,t.w/2),n-t.w/2)),t})},t}(),oi=n(13),ai=function(){return(ai=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)},si=ai({},oi,a),ui={init:function(t){var e=Object.assign({},t);e.guide=Jt(e.guide||{},{animationSpeed:0,avoidScalesOverflow:!0,enableColorToBarPosition:!1,maxHighlightDistance:32}),e.guide.size=e.guide.size||{},e.guide.label=Jt(e.guide.label||{},{position:["auto:avoid-label-label-overlap","auto:avoid-label-anchor-overlap","auto:adjust-on-label-overflow","auto:hide-on-label-label-overlap","auto:hide-on-label-anchor-overlap"]});var n=e.guide.avoidScalesOverflow,r=e.guide.enableColorToBarPosition;return e.transformRules=[function(t){var e=[t.scaleX,t.scaleY].sort(function(t,e){var n=t.discrete?1:0;return(e.discrete?1:0)*e.domain().length-n*t.domain().length})[0];return t.scaleY===e?E.get("flip")(t):E.get("identity")(t)},e.stack&&E.get("stack"),r&&E.get("positioningByColor")].filter(function(t){return t}),e.adjustRules=[e.stack&&E.get("adjustYScale"),function(t,n){var r=t.scaleSize.isEmptyScale(),i=Jt(e.guide.size,{defMinSize:10,defMaxSize:r?10:40,enableDistributeEvenly:!r}),o=Object.assign({},n,{defMin:i.defMinSize,defMax:i.defMaxSize,minLimit:i.minSize,maxLimit:i.maxSize});return(i.enableDistributeEvenly?E.get("adjustSigmaSizeScale"):E.get("adjustStaticSizeScale"))(t,o)},n&&function(t,e){var n=Object.assign({},e,{sizeDirection:"xy"});return E.get("avoidScalesOverflow")(t,n)}].filter(function(t){return t}),e},addInteraction:function(){var t=this,e=this.node();e.on("highlight",function(e,n){return t.highlight(n)}),e.on("data-hover",function(e,n){return t.highlight((r=n.data,function(t){return t===r||null}));var r})},draw:function(){var t=this.node(),e=t.config,n=e.options;n.container=n.slot(e.uid);var r,i=function(t){return rt(t,e.guide.animationSpeed)},o=Xe+"dot dot i-role-element i-role-datum",a=t.screenModel,s={fill:function(t){return a.color(t)},class:function(t){return o+" "+a.class(t)}},u={r:function(t){return Math.round(1e4*a.size(t)/2)/1e4},cx:function(t){return a.x(t)},cy:function(t){return a.y(t)}},c=[],l=a.toFibers();this._getGroupOrder=(r=l.reduce(function(t,e,n){return t.set(e,n),t},new Map),function(t){return r.get(t)});var d=n.container.selectAll(".frame").data(l,function(t){return a.group(t[0])});d.enter().append("g").attr("opacity",0).merge(d).call(function(e){e.attr("class","frame").call(function(e){var n=e.selectAll("circle").data(function(t){return t},a.id),r=n.enter().append("circle").call(dt(u)).merge(n).call(dt(s));i(r).call(dt(u)),i(n.exit()).attr("r",0).remove(),c.push.apply(c,r.nodes()),t.subscribe(r)}),i(e).attr("opacity",1)}),this._boundsInfo=this._getBoundsInfo(c),i(d.exit()).attr("opacity",0).remove().selectAll("circle").attr("r",0),t.subscribe(new ii(a.model,a.flip,e.guide.label,n).draw(l))},_getBoundsInfo:function(t){if(0===t.length)return null;var e=this.node().screenModel,n=t.map(function(t){var n=si.select(t).data()[0];return{node:t,data:n,x:e.x(n),y:e.y(n),r:e.size(n)/2}}).filter(function(t){return!isNaN(t.x)&&!isNaN(t.y)}),r=n.reduce(function(t,e){var n=e.x,r=e.y;return t.left=Math.min(n,t.left),t.right=Math.max(n,t.right),t.top=Math.min(r,t.top),t.bottom=Math.max(r,t.bottom),t},{left:Number.MAX_VALUE,right:Number.MIN_VALUE,top:Number.MAX_VALUE,bottom:Number.MIN_VALUE}),i=n.reduce(function(t,e){var n=e.x+","+e.y;return t[n]||(t[n]=[]),t[n].push(e),t},{});return{bounds:r,tree:si.quadtree().x(function(t){return t[0].x}).y(function(t){return t[0].y}).addAll(Object.keys(i).map(function(t){return i[t]}))}},getClosestElement:function(t,e){if(!this._boundsInfo)return null;var n=this._boundsInfo,r=n.bounds,i=n.tree,o=Fe(this.node().config.options.container.node()),a=t-o.x,s=e-o.y,u=this.node().config.guide.maxHighlightDistance;if(a<r.left-u||a>r.right+u||s<r.top-u||s>r.bottom+u)return null;var c=(i.find(a,s)||[]).map(function(t){var e=Math.sqrt(Math.pow(a-t.x,2)+Math.pow(s-t.y,2));if(e>u)return null;var n=e<t.r?t.r-e:e;return{node:t.node,data:t.data,x:t.x,y:t.y,distance:e,secondaryDistance:n}}).filter(function(t){return t}).sort(function(t,e){return t.secondaryDistance-e.secondaryDistance}),l=c.findIndex(function(t){return t.distance!==c[0].distance||t.secondaryDistance!==c[0].secondaryDistance}),d=l<0?c:c.slice(0,l);if(1===d.length)return d[0];var p=d.reduce(function(t,e){return t+e.x},0)/d.length,f=d.reduce(function(t,e){return t+e.y},0)/d.length,h=Math.atan2(f-s,p-a)+Math.PI;return d[Math.round((d.length-1)*h/2/Math.PI)]},highlight:function(t){var e,n=this.node().config.options.container,r=((e={})["tau-chart__highlighted"]=function(e){return!0===t(e)},e["tau-chart__dimmed"]=function(e){return!1===t(e)},e);n.selectAll(".dot").call(pt(r)),n.selectAll(".i-role-label").call(pt(r)),this._sortElements(t)},_sortElements:function(t){var e=this,n=this.node().config.options.container,r=new Map,i=new Map;n.selectAll(".frame").each(function(e){r.set(this,e.some(t)),i.set(this,e)});var o=ne(function(t,e){return r.get(t)-r.get(e)},function(t,n){return e._getGroupOrder(i.get(t))-e._getGroupOrder(i.get(n))});Te(n.node(),function(t,e){return"g"===t.tagName&&"g"===e.tagName?o(t,e):t.tagName.localeCompare(e.tagName)}),Le(n,".dot",t)}},ci=n(9),li="taucharts_synthetic_record";function di(t){return!0!==t[li]}function pi(t){return function(e){var n=t.stack,r=e.scaleX.period,i="linear"===e.scaleY.scaleType,o=!t.guide.x||null==t.guide.x.fillGaps;return!o&&t.guide.x.fillGaps||o&&(n||r&&i)?E.get("fillGaps")(e,{isStack:n,xPeriod:r,utc:t.guide.utcTime}):{}}}function fi(t){if(0===t.length)return"";if(1===t.length)return mi(t[0]);for(var e=[],n=1;n<t.length;n++)e.push(gi(t[n-1],t[n]));return e.join(" ")}function hi(t){if(0===t.length)return"";if(1===t.length)return mi(t[0]);for(var e=[],n=3;n<t.length;n+=3)e.push(yi(t[n-3],t[n-2],t[n-1],t[n]));return e.join(" ")}function mi(t){var e=t.size/2;return["M"+t.x+","+(t.y-e),"A"+e+","+e+" 0 0 1",t.x+","+(t.y+e),"A"+e+","+e+" 0 0 1",t.x+","+(t.y-e),"Z"].join(" ")}function gi(t,e){var n=Oi(t,e);return n?["M"+n.left[0].x+","+n.left[0].y,"L"+n.left[1].x+","+n.left[1].y,"A"+e.size/2+","+e.size/2+" 0 "+Number(t.size<e.size)+" 1",n.right[1].x+","+n.right[1].y,"L"+n.right[0].x+","+n.right[0].y,"A"+t.size/2+","+t.size/2+" 0 "+Number(t.size>e.size)+" 1",n.left[0].x+","+n.left[0].y,"Z"].join(" "):mi(t.size>e.size?t:e)}function yi(t,e,n,r){var i=function(t,e,n,r){var i=bi(t,r);if(0===i||i+t.size/2<=r.size/2||i+r.size/2<=t.size/2)return null;var o=function(i){var o=i?[r,n,e,t]:[t,e,n,r],a=wi.apply(void 0,[1/12*2].concat(o)),s=wi.apply(void 0,[.5].concat(a.slice(0,4))),u=s[3],c=s[6],l=Oi(o[0],u),d=Oi(u,c),p=[l.left[0],k(.5,l.left[1],d.left[0]),d.left[1]],f=[l.right[0],k(.5,l.right[1],d.right[0]),d.right[1]],h=ji.apply(void 0,p)[1],m=ji.apply(void 0,f)[1],g=k(4,l.left[0],h),y=k(4,l.right[0],m);return{left:i?[y,f[0]]:[p[0],g],right:i?[g,p[0]]:[f[0],y]}},a=o(!1),s=o(!0);return{left:a.left.concat(s.left),right:a.right.concat(s.right)}}(t,e,n,r);if(!i)return gi(t,r);var o=vi(_i(t,i.right[0]),_i(t,i.left[0])),a=vi(_i(r,i.right[1]),_i(r,i.left[1]));return["M"+i.left[0].x+","+i.left[0].y,"C"+i.left[1].x+","+i.left[1].y,i.left[2].x+","+i.left[2].y,i.left[3].x+","+i.left[3].y,"A"+r.size/2+","+r.size/2+" 0 "+Number(o>Math.PI)+" 1",i.right[3].x+","+i.right[3].y,"C"+i.right[2].x+","+i.right[2].y,i.right[1].x+","+i.right[1].y,i.right[0].x+","+i.right[0].y,"A"+t.size/2+","+t.size/2+" 0 "+Number(a>Math.PI)+" 1",i.left[0].x+","+i.left[0].y,"Z"].join(" ")}function _i(t,e){return Math.atan2(e.y-t.y,e.x-t.x)}function vi(t,e){return e<t&&(e+=2*Math.PI),e-t}function bi(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];for(var n=0,r=1;r<t.length;r++)n+=Math.sqrt((t[r].x-t[r-1].x)*(t[r].x-t[r-1].x)+(t[r].y-t[r-1].y)*(t[r].y-t[r-1].y));return n}function xi(t,e,n){return{x:t.x+e*Math.cos(n),y:t.y+e*Math.sin(n)}}function wi(t,e,n,r,i){var o=A(t,e,n,r,i),a=1/(1+bi(o[3],o[4],o[5],o[6],o[3])/bi(o[0],o[1],o[2],o[3],o[0]));return o[3].size=e.size*(1-a)+i.size*a,o}function ji(t,e,n){var r=k(bi(t,e)/bi(t,e,n),t,n);return[t,k(2,r,e),n]}function Oi(t,e){var n=bi(t,e);if(0===n||n+t.size/2<=e.size/2||n+e.size/2<=t.size/2)return null;var r=_i(t,e),i=Math.asin((t.size-e.size)/n/2),o=r-Math.PI/2+i,a=r+Math.PI/2-i;return{left:[xi(t,t.size/2,o),xi(e,e.size/2,o)],right:[xi(t,t.size/2,a),xi(e,e.size/2,a)]}}var Ci=function(){return(Ci=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)},Ei={circle:{element:"circle",getInitialAttrs:function(t,e){var n=t.config.guide,r=t.screenModel;return{r:"hover"===n.showAnchors?0:function(t){return r.size(t)/2},cx:function(t){return e.x(t)},cy:function(t){return e.y(t)}}},getHighlightAttrs:function(t,e,n){t.config.guide;var r=t.screenModel,i="hover"===t.config.guide.showAnchors;return{r:i?function(t){return n(t)?Math.max(4,r.size(t)/2):0}:function(t){var e=r.size(t)/2;return n(t)&&(e=Math.max(4,Math.ceil(1.25*e))),e}}}},"vertical-stick":{element:"path",getInitialAttrs:function(t,e){var n=t.config.guide,r=t.screenModel;return{"shape-rendering":"crispEdges",d:function(t){var i=e.x(t),o=e.y(t),a=e.x0(t),s=e.y0(t),u="hover"===n.showAnchors?0:r.size(t)/2;return fi([{x:i,y:o,size:u},{x:a,y:s,size:u}])}}},getHighlightAttrs:function(t,e,n){t.config.guide;var r=t.screenModel,i="hover"===t.config.guide.showAnchors;return{d:function(t){var o=e.x(t),a=e.y(t),s=e.x0(t),u=e.y0(t),c=i?n(t)?Math.max(4,r.size(t)/2):0:n(t)?Math.max(4,Math.ceil(r.size(t)/2*1.25)):r.size(t)/2;return fi([{x:o,y:a,size:c},{x:s,y:u,size:c}])}}}}};function Si(t,e,n){var r=n.sort(function(t,e){return t.distance===e.distance?t.secondaryDistance-e.secondaryDistance:t.distance-e.distance}),i=r.findIndex(function(t){return t.distance!==r[0].distance||t.secondaryDistance!==r[0].secondaryDistance}),o=i<0?r:r.slice(0,i);if(1===o.length)return o[0];var a=o.reduce(function(t,e){return t+e.x},0)/o.length,s=o.reduce(function(t,e){return t+e.y},0)/o.length,u=Math.atan2(s-e,a-t)+Math.PI;return o[Math.round((o.length-1)*u/2/Math.PI)]}var Mi=Xe+"dot-line dot-line i-role-dot i-role-datum "+Xe+"dot",Ti={init:function(t){var e=t;return e.guide=Jt(e.guide||{},{animationSpeed:0,cssClass:"",maxHighlightDistance:32,widthCssClass:"",color:{},label:{}}),e.guide.label=Jt(e.guide.label,{fontSize:11,hideEqualLabels:!0,position:["auto:avoid-label-label-overlap","auto:avoid-label-anchor-overlap","auto:avoid-label-edges-overlap","auto:adjust-on-label-overflow","auto:hide-on-label-label-overlap","auto:hide-on-label-edges-overlap"]}),e.guide.color=Jt(e.guide.color||{},{fill:null}),["never","hover","always"].indexOf(e.guide.showAnchors)<0&&(e.guide.showAnchors="hover"),e.transformRules=[],e.adjustRules=[],e},baseModel:function(t){var e={gog:t.model,x:t.x,y:t.y,x0:t.x0,y0:t.y0,size:t.size,group:t.group,order:t.order,color:t.color,class:t.class,groupAttributes:{},pathAttributesUpdateInit:{},pathAttributesUpdateDone:{},pathAttributesEnterInit:{},pathAttributesEnterDone:{},pathElement:null,dotAttributes:{r:function(t){return Math.round(1e4*e.size(t)/2)/1e4},cx:function(t){return e.x(t)},cy:function(t){return e.y(t)},fill:function(t){return e.color(t)},class:function(t){return Mi+" "+e.class(t)}},dotAttributesDefault:{r:0,cy:function(t){return e.y0(t)}}};return e},addInteraction:function(){var t=this,e=this.node(),n=this.node().config;e.on("highlight",function(e,n){return t.highlight(n)}),e.on("highlight-data-points",function(e,n){return t.highlightDataPoints(n)}),"never"!==n.guide.showAnchors&&e.on("data-hover",function(e,n){return t.highlightDataPoints((r=n.data,function(t){return t===r||null}));var r})},draw:function(){var t=this.node(),e=t.config,n=e.guide,r=e.options;r.container=r.slot(e.uid);var i=t.screenModel,o=this.buildModel(i);this.domElementModel=o;var s=ut,u=function(r){r.call(dt(o.groupAttributes));var u,c=r.selectAll((u=Mi,u.split(/\s+/g).map(function(t){return"."+t}).join(""))).data(function(t){return t.length<=1?t:[]},i.id);c.exit().call(s(n.animationSpeed,null,{r:0},function(t){return a.select(t).remove()})),c.call(s(n.animationSpeed,null,o.dotAttributes));var l=c.enter().append("circle").call(s(n.animationSpeed,o.dotAttributesDefault,o.dotAttributes)).merge(c);t.subscribe(l);var d,f,h,m,g,y,_,v,b=r.selectAll(o.pathElement+":not(.i-data-anchor)").data(function(t){return t.length>1?[t]:[]},p);if(b.exit().remove(),b.call(s(n.animationSpeed,o.pathAttributesUpdateInit,o.pathAttributesUpdateDone,o.afterPathUpdate)),b.enter().append(o.pathElement).call(s(n.animationSpeed,o.pathAttributesEnterInit,o.pathAttributesEnterDone,o.afterPathUpdate)).merge(b).call(function(t){e.guide.animationSpeed>0&&!document.hidden?(t.attr(o.pathTween.attr,function(t){return o.pathTween.fn.call(this,t)(0)}),rt(t,e.guide.animationSpeed,"pathTransition").attrTween(o.pathTween.attr,o.pathTween.fn)):t.attr(o.pathTween.attr,function(t){return o.pathTween.fn.call(this,t)(1)})}),t.subscribe(l),"never"!==n.showAnchors){var x=(d=t,h=r,m=(f=o).anchorShape,g=d.config.guide,y=d.screenModel,_=Ci({},Ei[m].getInitialAttrs(d,f),{opacity:"hover"===g.showAnchors?0:1,fill:function(t){return y.color(t)},class:"i-data-anchor"}),(v=h.selectAll(".i-data-anchor").data(function(t){return t.filter(di)},y.id)).exit().remove(),v.call(ut(g.animationSpeed,null,_)),v.enter().append(Ei[f.anchorShape].element).call(ut(g.animationSpeed,{r:0},_)).merge(v));t.subscribe(x)}},c=i.toFibers(),l=c.map(function(t){return t.filter(di)}),d=r.container.selectAll(".frame"),p=function(){var t=d.empty()?[]:d.data(),e=new Map;d.each(function(t){e.set(t,Number(this.getAttribute("data-id")))});var n=t.reduce(function(t,e){return t.set(e,e.map(i.id)),t},new Map),r=new Map,o=Math.max.apply(Math,[0].concat(Array.from(e.values())));return function(t){if(r.has(t))return r.get(t);var a,s=t.map(function(t){return i.id(t)}),u=(Array.from(n.entries()).find(function(t){var e=t[1];return s.some(function(t){return e.some(function(e){return e===t})})})||[null])[0];return a=u?e.get(u):++o,r.set(t,a),a}}();this._getDataSetId=p;var f=d.data(c,p);f.exit().remove(),f.call(u),f.enter().append("g").attr("data-id",p).call(u),f.order(),this._boundsInfo=this._getBoundsInfo(r.container.selectAll(".i-data-anchor").nodes()),t.subscribe(new ii(i.model,e.flip,e.guide.label,r).draw(l))},_getBoundsInfo:function(t){if(0===t.length)return null;var e=this.node().screenModel,n=this.node().config.flip,r=t.map(function(t){var n=a.select(t).data()[0];return{node:t,data:n,x:e.x(n),y:e.y(n)}}).filter(function(t){return!isNaN(t.x)&&!isNaN(t.y)}),i=r.reduce(function(t,e){var n=e.x,r=e.y;return t.left=Math.min(n,t.left),t.right=Math.max(n,t.right),t.top=Math.min(r,t.top),t.bottom=Math.max(r,t.bottom),t},{left:Number.MAX_VALUE,right:Number.MIN_VALUE,top:Number.MAX_VALUE,bottom:Number.MIN_VALUE}),o=Xt(r.map(n?function(t){return t.y}:function(t){return t.x})).sort(function(t,e){return t-e}),s=o.reduce(function(t,e){return t[e]=[],t},{});r.forEach(function(t){var e=o.find(n?function(e){return t.y===e}:function(e){return t.x===e});s[e].push(t)});var u=function(t){if(1===t.length)return s[t[0]];var e=Math.ceil(t.length/2);return{middle:(t[e-1]+t[e])/2,lower:u(t.slice(0,e)),greater:u(t.slice(e))}};return{bounds:i,tree:u(o)}},getClosestElement:function(t,e){if(!this._boundsInfo)return null;var n=this._boundsInfo,r=n.bounds,i=n.tree,o=this.node().config.options.container,a=this.node().config.flip,s=Fe(o.node()),u=this.node().config.guide.maxHighlightDistance;if(t<r.left+s.x-u||t>r.right+s.x+u||e<r.top+s.y-u||e>r.bottom+s.y+u)return null;var c=a?e-s.y:t-s.x,l=function t(e){return Array.isArray(e)?e:t(c>e.middle?e.greater:e.lower)}(i).map(function(n){var r=n.x+s.x,i=n.y+s.y,o=Math.abs(a?e-i:t-r),u=Math.abs(a?t-r:e-i);return{node:n.node,data:n.data,distance:o,secondaryDistance:u,x:r,y:i}});return Si(t,e,l)},highlight:function(t){var e,n,r=this.node().config.options.container,i="tau-chart__highlighted",o=r.selectAll(".i-role-path"),a=o.data().filter(function(e){return e.filter(di).some(t)}),s=a.length>0;o.call(pt(((e={})[i]=function(t){return s&&a.indexOf(t)>=0},e["tau-chart__dimmed"]=function(t){return s&&a.indexOf(t)<0},e)));var u=((n={})[i]=function(e){return!0===t(e)},n["tau-chart__dimmed"]=function(e){return!1===t(e)},n);r.selectAll(".i-role-dot").call(pt(u)),r.selectAll(".i-role-label").call(pt(u)),this._sortElements(t)},highlightDataPoints:function(t){var e,n,r,i,o,a,s=this.node(),u=(e=s,n=this.domElementModel,r=t,i=n.anchorShape,o=e.screenModel,a="hover"===e.config.guide.showAnchors,e.config.options.container.selectAll(".i-data-anchor").call(dt(Ei[i].getHighlightAttrs(e,n,r))).attr("opacity",a?function(t){return r(t)?1:0}:function(){return 1}).attr("fill",function(t){return o.color(t)}).attr("class",function(t){return ke("i-data-anchor",o.class(t))}).classed(Xe+"highlighted",r)),c=s.config.options.container,l=s.config.flip,d=u.filter(t),p=c.select(".cursor-line");if(d.empty())p.remove();else{p.empty()&&(p=c.append("line"));var f=s.screenModel.model,h=f.xi(d.data()[0]),m=f.xi(d.data()[0]),g=f.scaleY.domain(),y=f.scaleY(g[0]),_=f.scaleY(g[1]);p.attr("class","cursor-line").attr("x1",l?y:h).attr("y1",l?h:y).attr("x2",l?_:m).attr("y2",l?m:_)}this._sortElements(t)},_sortElements:function(t){var e=this.node().config.options.container,n=new Map,r=new Map,i=this._getDataSetId;e.selectAll(".i-role-path").each(function(e){n.set(this,i(e)),r.set(this,e.filter(di).some(t))});var o=ne(function(t,e){return r.get(t)-r.get(e)},function(t,e){return n.get(t)-n.get(e)}),a={line:0,g:1,text:2};Te(e.node(),function(t,e){return"g"===t.tagName&&"g"===e.tagName?o(t,e):a[t.tagName]-a[e.tagName]})}},ki=[1,2,3,4,5],Ai=ki.map(function(t){return Xe+"line-opacity-"+t}),Di=ki.map(function(t){return Xe+"line-width-"+t});function Ii(t){return Ai[t-1]||Ai[4]}function Pi(t,e){return t.length<2?"":String.prototype.concat.apply("",t.concat(e.slice().reverse()).map(function(t,e){return(0===e?"":" ")+t.x+","+t.y}))}function Ri(t,e){if(t.length<2)return"";var n=function(t){var e=t.map(function(t,e){return((e-1)%3==0?"C":"")+t.x+","+t.y+" "});return String.prototype.concat.apply("",e)};return"M"+n(t)+"L"+n(e.slice().reverse())+"Z"}var Ni=function(){return(Ni=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)},Fi=Ni({},ci,a),Li={draw:Ti.draw,highlight:Ti.highlight,highlightDataPoints:Ti.highlightDataPoints,addInteraction:Ti.addInteraction,_sortElements:Ti._sortElements,init:function(t){var e=Ti.init(t),n=e.stack;return e.transformRules=[e.flip&&E.get("flip"),e.guide.obsoleteVerticalStackOrder&&E.get("obsoleteVerticalStackOrder"),!n&&E.get("groupOrderByAvg"),pi(e),n&&E.get("stack")],e.adjustRules=[function(t,n){var r=t.scaleSize.isEmptyScale(),i=Jt(e.guide.size||{},{defMinSize:2,defMaxSize:r?6:40}),o=Object.assign({},n,{defMin:i.defMinSize,defMax:i.defMaxSize,minLimit:i.minSize,maxLimit:i.maxSize});return E.get("adjustStaticSizeScale")(t,o)}],e},buildModel:function(t){var e=Ti.baseModel(t),n=this.node().config.guide,r=Ii(t.model.scaleColor.domain().length),i=Xe+"area area i-role-path "+r+" "+n.cssClass+" ";e.groupAttributes={class:function(t){return i+" "+e.class(t[0])+" frame"}};var o={fill:function(t){return e.color(t[0])},stroke:function(t){var n=e.color(t[0]);return n.length>0&&(n=Fi.rgb(n).darker(1)),n}};e.pathAttributesEnterInit=o,e.pathAttributesUpdateDone=o;var a="polyline"===Q(n.interpolate);return e.pathElement=a?"polygon":"path",e.anchorShape="vertical-stick",e.pathTween={attr:a?"points":"d",fn:lt(0,a?Pi:Ri,[function(n){return{id:t.id(n),x:e.x(n),y:e.y(n)}},function(n){return{id:t.id(n),x:e.x0(n),y:e.y0(n)}}],t.id,n.interpolate)},e},_getBoundsInfo:function(t){if(0===t.length)return null;var e,n,r=this.node().screenModel,i=this.node().config.flip,o=t.map(function(t){var e=Fi.select(t).data()[0];return{node:t,data:e,x:r.x(e),y:r.y(e),y0:r.y0(e),group:r.group(e)}}),a=o.reduce(function(t,e){var n=e.x,r=e.y,i=e.y0;return t.left=Math.min(n,t.left),t.right=Math.max(n,t.right),t.top=Math.min(r,i,t.top),t.bottom=Math.max(r,i,t.bottom),t},{left:Number.MAX_VALUE,right:Number.MIN_VALUE,top:Number.MAX_VALUE,bottom:Number.MIN_VALUE}),s=Xt(o.map(i?function(t){return t.y}:function(t){return t.x})).sort(function(t,e){return t-e}),u=s.reduce(function(t,e){return t[e]=[],t},{});if(o.forEach(function(t){var e=s.find(i?function(e){return t.y===e}:function(e){return t.x===e});u[e].push(t)}),e=Object.keys(o.reduce(function(t,e){return t[e.group]=!0,t},{})),n=e.reduce(function(t,e,n){return t[e]=n,t},{}),s.forEach(function(t){var r=u[t];if(r.sort(function(t,e){return n[t.group]-n[e.group]}),r.length<e.length)for(var i=0;i<e.length;i++){var o=!1,a=void 0,s=void 0;if(i===r.length?(a=s=0===i?0:r[i-1].y,o=!0):r[i].group!==e[i]&&(a=s=r[i].y0,o=!0),o){var c={x:t,y:a,y0:s,data:null,node:null,group:e[i]};r.splice(i,0,c)}}}),1===s.length)return{bounds:a,tree:{start:s[0],end:s[0],isLeaf:!0,items:{start:u[s[0]],end:u[s[0]]}}};var c=function(t){if(2===t.length){var e=t[0],n=t[1];return{start:e,end:n,isLeaf:!0,items:{start:u[e],end:u[n]}}}var r=t.length%2==0?t.length/2:(t.length-1)/2;return t[r],{start:t[0],end:t[t.length-1],isLeaf:!1,left:c(t.slice(0,r+1)),right:c(t.slice(r))}};return{bounds:a,tree:c(s)}},getClosestElement:function(t,e){if(!this._boundsInfo)return null;var n=this._boundsInfo,r=n.bounds,i=n.tree,o=this.node().config.options.container,a=this.node().config.flip,s=Fe(o.node()),u=this.node().config.guide.maxHighlightDistance;if(t<r.left+s.x-u||t>r.right+s.x+u||e<r.top+s.y-u||e>r.bottom+s.y+u)return null;var c=a?e-s.y:t-s.x,l=function t(e){if(e.isLeaf)return e;var n=e.left.end;return t(c<n?e.left:e.right)}(i),d=l.end===l.start?0:(c-l.start)/(l.end-l.start);d<0&&(d=0),d>1&&(d=1);var p,f=(p=l.items.start.reduce(function(t,e){return t[e.group]={start:e,end:null,y:null,y0:null},t},{}),l.items.end.forEach(function(t){void 0!==p[t.group]?p[t.group].end=t:delete p[t.group]}),Object.keys(p).forEach(function(t){var e=p[t];e.end?(e.y=e.start.y+d*(e.end.y-e.start.y),e.y0=e.start.y0+d*(e.end.y0-e.start.y0)):delete p[t]}),Object.keys(p).map(function(t){return p[t]}).map(function(t){return{y:t.y,y0:t.y0,el:d<.5?t.start:t.end}}).filter(function(t){return null!=t.el.data})),h=e-s.y,m=f.filter(function(t){return h>=t.y&&h<=t.y0}),g=(m.length>0?m:f).map(function(t){return t.el}).map(function(n){var r=n.x+s.x,i=n.y+s.y,o=Math.abs(a?e-i:t-r),u=Math.abs(a?t-r:e-i);return{node:n.node,data:n.data,distance:o,secondaryDistance:u,x:r,y:i}});return Si(t,e,g)}},Bi={draw:Ti.draw,getClosestElement:Ti.getClosestElement,highlight:Ti.highlight,highlightDataPoints:Ti.highlightDataPoints,addInteraction:Ti.addInteraction,_getBoundsInfo:Ti._getBoundsInfo,_sortElements:Ti._sortElements,init:function(t){var e=Ti.init(t);return e.transformRules=[e.flip&&E.get("flip")],e.adjustRules=[function(t,n){var r=t.scaleSize.isEmptyScale(),i=Jt(e.guide.size||{},{defMinSize:2,defMaxSize:r?6:40}),o=Object.assign({},n,{defMin:i.defMinSize,defMax:i.defMaxSize,minLimit:i.minSize,maxLimit:i.maxSize});return E.get("adjustStaticSizeScale")(t,o)}],e},buildModel:function(t){var e=Ti.baseModel(t),n=this.node().config.guide,r=Ii(t.model.scaleColor.domain().length),i=Xe+"area area i-role-path "+r+" "+n.cssClass+" ";e.groupAttributes={class:function(t){return i+" "+e.class(t[0])+" frame"}};var o,a,s={fill:function(t){return e.color(t[0])},stroke:function(t){return e.color(t[0])}};return e.pathAttributesEnterInit=s,e.pathAttributesUpdateDone=s,e.pathElement="polygon",e.anchorShape="circle",e.pathTween={attr:"points",fn:lt(0,(o=function(t){return t.x},a=function(t){return t.y},function(t){return t.map(function(t){return[o(t),a(t)].join(",")}).join(" ")}),[function(n){return{id:t.id(n),x:e.x(n),y:e.y(n)}}],t.id)},e}};function Vi(t){if(t.length<2)return"";for(var e="",n=0;n<t.length;n++)e+=(0===n?"M":" L")+t[n].x+","+t[n].y;return e}function zi(t){if(t.length<4)return"";for(var e="M"+t[0].x+","+t[0].y,n=3;n<t.length;n+=3)e+=" C"+t[n-2].x+","+t[n-2].y+" "+t[n-1].x+","+t[n-1].y+" "+t[n].x+","+t[n].y;return e}var Hi,Ui,Wi,qi,Gi,Yi,Ki,Xi={draw:Ti.draw,getClosestElement:Ti.getClosestElement,highlight:Ti.highlight,highlightDataPoints:Ti.highlightDataPoints,addInteraction:Ti.addInteraction,_getBoundsInfo:Ti._getBoundsInfo,_sortElements:Ti._sortElements,init:function(t){var e=Ti.init(t),n=e.stack;e.guide=Jt(e.guide||{},{avoidScalesOverflow:!0,interpolate:"linear"}),e.transformRules=[e.flip&&E.get("flip"),!n&&E.get("groupOrderByAvg"),pi(e),n&&E.get("stack")];var r=e.guide.avoidScalesOverflow,i=function(t){return t.scaleSize.isEmptyScale()};return e.adjustRules=[function(t,n){var r=Jt(e.guide.size||{},{defMinSize:2,defMaxSize:i(t)?6:40}),o=Object.assign({},n,{defMin:r.defMinSize,defMax:r.defMaxSize,minLimit:r.minSize,maxLimit:r.maxSize});return E.get("adjustStaticSizeScale")(t,o)},r&&function(t,e){if(i(t))return function(){return{}};var n=Object.assign({},e,{sizeDirection:"xy"});return E.get("avoidScalesOverflow")(t,n)}].filter(function(t){return t}),e},buildModel:function(t){var e,n,r=this.node().config,i=r.guide,o=r.options,a=!t.model.scaleSize.dim,s=a?i.widthCssClass||(e=o.width,n=0,e>=160&&e<320?n=1:e>=320&&e<480?n=2:e>=480&&e<640?n=3:e>=640&&(n=4),Di[n]):"",u=Ii(t.model.scaleColor.domain().length),c=a?"line":"area",l=""+Xe+c+" "+c+" i-role-path "+s+" "+u+" "+i.cssClass+" ",d=a?{stroke:function(t){return f.color(t[0])},class:"i-role-datum"}:{fill:function(t){return f.color(t[0])}},p="cubic"===Q(i.interpolate)?a?zi:hi:a?Vi:fi,f=Ti.baseModel(t),h=a?function(e){return{id:t.id(e),x:f.x(e),y:f.y(e)}}:function(e){return{id:t.id(e),x:f.x(e),y:f.y(e),size:f.size(e)}};return f.groupAttributes={class:function(t){return l+" "+f.class(t[0])+" frame"}},f.pathElement="path",f.anchorShape="circle",f.pathAttributesEnterInit=d,f.pathAttributesUpdateDone=d,f.pathTween={attr:"d",fn:lt(0,p,[h],t.id,i.interpolate)},f}},Qi=function(t){return a.select(t).data()[0]},Zi={init:function(t){var e=Object.assign({},t);e.guide=e.guide||{},e.guide=Jt(e.guide,{animationSpeed:0,avoidScalesOverflow:!0,maxHighlightDistance:32,prettify:!0,sortByBarHeight:!0,enableColorToBarPosition:null!=e.guide.enableColorToBarPosition?e.guide.enableColorToBarPosition:!e.stack}),e.guide.size=Jt(e.guide.size||{},{enableDistributeEvenly:!0}),e.guide.label=Jt(e.guide.label||{},{position:e.flip?e.stack?["r-","l+","hide-by-label-height-horizontal","cut-label-horizontal"]:["outside-then-inside-horizontal","auto:hide-on-label-label-overlap"]:e.stack?["rotate-on-size-overflow","t-","b+","hide-by-label-height-vertical","cut-label-vertical","auto:hide-on-label-label-overlap"]:["rotate-on-size-overflow","outside-then-inside-vertical","auto:hide-on-label-label-overlap"]});var n=e.guide.avoidScalesOverflow,r=e.guide.enableColorToBarPosition,i=e.guide.size.enableDistributeEvenly;return e.transformRules=[e.flip&&E.get("flip"),e.guide.obsoleteVerticalStackOrder&&E.get("obsoleteVerticalStackOrder"),e.stack&&E.get("stack"),r&&E.get("positioningByColor")].filter(function(t){return t}),e.adjustRules=[i&&function(t,n){var r=Jt(e.guide.size||{},{defMinSize:e.guide.prettify?3:0,defMaxSize:e.guide.prettify?40:Number.MAX_VALUE}),i=Object.assign({},n,{defMin:r.defMinSize,defMax:r.defMaxSize,minLimit:r.minSize,maxLimit:r.maxSize});return E.get("size_distribute_evenly")(t,i)},n&&i&&function(t,e){var n=Object.assign({},e,{sizeDirection:"x"});return E.get("avoidScalesOverflow")(t,n)},e.stack&&E.get("adjustYScale")].filter(function(t){return t}),e},addInteraction:function(){var t=this,e=this.node();e.on("highlight",function(e,n){return t.highlight(n)}),e.on("data-hover",function(e,n){return t.highlight((r=n.data,function(t){return t===r||null}));var r})},draw:function(){var t,e,n=this.node(),r=n.config,i=r.options;i.container=i.slot(r.uid);var o=r.guide.prettify,s="i-role-element i-role-datum bar "+Xe+"bar",u=n.screenModel,c=this.buildModel(u,{prettify:o,minBarH:1,minBarW:1,baseCssClass:s}),l=ut,d=r.flip?"y":"x",p=r.flip?"x":"y",f=r.flip?"width":"height",h=r.flip?"height":"width",m=u.toFibers(),g=m.reduce(function(t,e){return t.concat(e)},[]),y=c.class,_=te(c,"class"),v=i.container.selectAll(".bar").data(g,u.id);v.exit().classed("tau-removing",!0).call(l(r.guide.animationSpeed,null,((t={})[d]=function(){var t=a.select(this);return t.attr(d)-0+(t.attr(h)-0)/2},t[p]=function(){return this.getAttribute("data-zero")},t[h]=0,t[f]=0,t),function(t){var e=a.select(t);e.classed("tau-removing")&&e.remove()})),v.call(l(r.guide.animationSpeed,null,_));var b=v.enter().append("rect").call(l(r.guide.animationSpeed,(e={},e[p]=u[p+"0"],e[f]=0,e),_)).merge(v).attr("class",y).attr("data-zero",u[p+"0"]);n.subscribe(new ii(u.model,u.model.flip,r.guide.label,i).draw(m));var x,w=(x=g.reduce(function(t,e,n){return t.set(e,n+1),t},new Map),function(t,e){return(x.get(Qi(t))||-1)-(x.get(Qi(e))||-1)});this._barsSorter=r.guide.sortByBarHeight?r.flip?function(t,e){var n=Qi(t),r=Qi(e),i=c.width(n),o=c.width(r);if(i===o){var a=c.y(n),s=c.y(r);return a===s?w(t,e):a-s}return o-i}:function(t,e){var n=Qi(t),r=Qi(e),i=c.height(n),o=c.height(r);if(i===o){var a=c.x(n),s=c.x(r);return a===s?w(t,e):a-s}return o-i}:w;var j={rect:0,text:1};this._typeSorter=function(t,e){return j[t.tagName]-j[e.tagName]},this._sortElements(this._typeSorter,this._barsSorter),n.subscribe(b),this._boundsInfo=this._getBoundsInfo(b.nodes())},buildModel:function(t,e){var n,r=e.prettify,i=e.minBarH,o=e.minBarW,a=e.baseCssClass,s=function(e){var n=t.size(e);return r&&(n=Math.max(o,n)),n},u=function(e){return e[t.model.scaleY.dim]};if(t.flip){var c=function(e){return Math.abs(t.x(e)-t.x0(e))};n={y:function(e){return t.y(e)-.5*s(e)},x:function(e){var n=Math.min(t.x0(e),t.x(e));if(r){var o=c(e),a=u(e),s=0;return 0===a&&(s=0),a>0&&(s=o),a<0&&(s=0-i),o<i?n+s:n}return n},height:function(t){return s(t)},width:function(t){var e=c(t);return r?0===u(t)?e:Math.max(i,e):e}}}else{var l=function(e){return Math.abs(t.y(e)-t.y0(e))};n={x:function(e){return t.x(e)-.5*s(e)},y:function(e){var n=Math.min(t.y0(e),t.y(e));return r&&(n=l(e)<i&&u(e)>0?n-i:n),n},width:function(t){return s(t)},height:function(t){var e=l(t);return r&&(e=0===u(t)?e:Math.max(i,e)),e}}}return Object.assign(n,{class:function(e){return a+" "+t.class(e)},fill:function(e){return t.color(e)}})},_sortElements:function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];Te(this.node().config.options.container.node(),ne.apply(r,t))},_getBoundsInfo:function(t){if(0===t.length)return null;var e=this.node().screenModel,n=this.node().config.flip,r=t.map(function(t){var n=a.select(t).data()[0],r=e.x(n),i=e.x0(n),o=e.y(n),s=e.y0(n),u=Math.abs(r-i),c=Math.abs(o-s),l=(r+i)/2,d=(o+s)/2;return{node:t,data:n,cx:l,cy:d,box:{top:d-c/2,right:l+u/2,bottom:d+c/2,left:l-u/2},invert:o>s}}),i=r.reduce(function(t,e){var n=e.box;return t.left=Math.min(n.left,t.left),t.right=Math.max(n.right,t.right),t.top=Math.min(n.top,t.top),t.bottom=Math.max(n.bottom,t.bottom),t},{left:Number.MAX_VALUE,right:Number.MIN_VALUE,top:Number.MAX_VALUE,bottom:Number.MIN_VALUE}),o=Xt(r.map(n?function(t){return t.cy}:function(t){return t.cx})).sort(function(t,e){return t-e}),s=o.reduce(function(t,e){return t[e]=[],t},{});r.forEach(function(t){var e=o.find(n?function(e){return t.cy===e}:function(e){return t.cx===e});s[e].push(t)});var u=function(t){if(1===t.length)return s[t];var e=Math.ceil(t.length/2);return{middle:(t[e-1]+t[e])/2,lower:u(t.slice(0,e)),greater:u(t.slice(e))}};return{bounds:i,tree:u(o)}},getClosestElement:function(t,e){if(!this._boundsInfo)return null;var n=this._boundsInfo,r=n.bounds,i=n.tree,o=this.node().config.options.container,a=this.node().config.flip,s=Fe(o.node()),u=t-s.x,c=e-s.y,l=this.node().config.guide.maxHighlightDistance;if(u<r.left-l||u>r.right+l||c<r.top-l||c>r.bottom+l)return null;var d=a?c:u,p=a?u:c;return function t(e){return Array.isArray(e)?e:t(d>e.middle?e.greater:e.lower)}(i).map(function(t){var e,n=a?t.box.left:t.box.top,r=a?t.box.right:t.box.bottom,i=(e=p)>=n&&e<=r;if(!i&&Math.abs(p-n)>l&&Math.abs(p-r)>l)return null;var o=Math.abs(p-(t.invert!==a?r:n));return Object.assign(t,{distToValue:o,cursorInside:i})}).filter(function(t){return t}).sort(function(t,e){return t.cursorInside!==e.cursorInside?e.cursorInside-t.cursorInside:Math.abs(t.distToValue)-Math.abs(e.distToValue)}).map(function(t){var e=t.cx,n=t.cy,r=Math.abs(a?c-n:u-e),i=Math.abs(a?u-e:c-n);return{node:t.node,data:t.data,distance:r,secondaryDistance:i,x:e,y:n}})[0]||null},highlight:function(t){var e,n=this.node().config.options.container,r=((e={})["tau-chart__highlighted"]=function(e){return!0===t(e)},e["tau-chart__dimmed"]=function(e){return!1===t(e)},e);n.selectAll(".bar").call(pt(r)),n.selectAll(".i-role-label").call(pt(r)),this._sortElements(function(e,n){return t(Qi(e))-t(Qi(n))},this._typeSorter,this._barsSorter)}},$i=n(14),Ji=(Ki=function(t,e){return(Ki=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}Ki(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),to=function(t){function e(e){var n=t.call(this,e)||this;return n.config=e,n.config.guide=Jt(n.config.guide||{},{}),n.on("highlight",function(t,e){return n.highlight(e)}),n}return Ji(e,t),e.prototype.defineGrammarModel=function(t){var e=this.config,n=e.options;this.color=t("color",e.color,{}),this.scalesMap=e.columns.reduce(function(e,r){return e[r]=t("pos",r,[0,n.height]),e},{});var r=n.width/(e.columns.length-1),i=e.columns.reduce(function(t,e,n){return t[e]=n*r,t},{});return this.xBase=function(t){return i[t]},this.regScale("columns",this.scalesMap).regScale("color",this.color),{}},e.prototype.drawFrames=function(t){var e=this.config,n=this.config.options,r=this.scalesMap,i=this.xBase,o=this.color,a=$i.line(),s=function(t){t.attr("d",function(t){return a(e.columns.map(function(e){return[i(e),r[e](t[r[e].dim])]}))})},u=function(t){t.attr("stroke",function(t){return o.toColor(o(t[o.dim]))}),t.attr("class",function(t){return Xe+"__line line "+o.toClass(o(t[o.dim]))+" foreground"})},c=function(t){var e=t.selectAll(".background").data(function(t){return t.part()});e.exit().remove(),e.call(s),e.enter().append("path").attr("class","background line").call(s);var n=t.selectAll(".foreground").data(function(t){return t.part()});n.exit().remove(),n.call(function(t){s(t),u(t)}),n.enter().append("path").call(function(t){s(t),u(t)})},l=n.container.selectAll(".lines-frame").data(t,function(t){return t.hash()});l.exit().remove(),l.call(c),l.enter().append("g").attr("class","lines-frame").call(c),this.subscribe(n.container.selectAll(".lines-frame .foreground"))},e.prototype.highlight=function(t){this.config.options.container.selectAll(".lines-frame .foreground").style("visibility",function(e){return t(e)?"":"hidden"})},e}(y),eo=function(){function t(t,e){var n,r=this;this._fields={},n=Array.isArray(e.fitToFrameByDims)&&e.fitToFrameByDims.length?t.part(function(t){var n={};return"where"===t.type&&t.args?(n.type=t.type,n.args=e.fitToFrameByDims.reduce(function(e,n){return t.args.hasOwnProperty(n)&&(e[n]=t.args[n]),e},{})):n=t,n}):t.full();var i=this.getVarSet(n,e);e.order&&(i=Zt($t(e.order,i),i)),this.vars=i;var o=i.map(function(t){return t});this.scaleConfig=e,this.scaleConfig.nice=this.scaleConfig.hasOwnProperty("nice")?this.scaleConfig.nice:this.scaleConfig.autoScale,this.addField("dim",this.scaleConfig.dim).addField("scaleDim",this.scaleConfig.dim).addField("scaleType",this.scaleConfig.type).addField("source",this.scaleConfig.source).addField("domain",function(){return r.vars}).addField("isInteger",o.every(Number.isInteger)).addField("originalSeries",function(){return o}).addField("isContains",function(t){return r.isInDomain(t)}).addField("isEmptyScale",function(){return r.isEmpty()}).addField("fixup",function(t){var e=r.scaleConfig;e.__fixup__=e.__fixup__||{},e.__fixup__=Object.assign(e.__fixup__,t(Object.assign({},e,e.__fixup__)))}).addField("commit",function(){r.scaleConfig=Object.assign(r.scaleConfig,r.scaleConfig.__fixup__),delete r.scaleConfig.__fixup__})}return t.prototype.isInDomain=function(t){return this.domain().indexOf(t)>=0},t.prototype.addField=function(t,e){return this._fields[t]=e,this[t]=e,this},t.prototype.getField=function(t){return this._fields[t]},t.prototype.isEmpty=function(){return!Boolean(this._fields.dim)},t.prototype.toBaseScale=function(t,e){var n=this;void 0===e&&(e=null);var r=Object.keys(this._fields).reduce(function(t,e){return t[e]=n._fields[e],t},t);return r.getHash=function(){return Nt([n.vars,e].map(function(t){return JSON.stringify}).join(""))},r.value=r,r},t.prototype.getVarSet=function(t,e){return Xt(e.hasOwnProperty("series")?e.series:t.map(function(t){return t[e.dim]}),"date"===e.dimType?function(t){return new Date(t).getTime()}:function(t){return t})},t}(),no=(Yi=function(t,e){return(Yi=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}Yi(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),ro=function(t){function e(e,n){var r=t.call(this,e,n)||this;return r._references=n.references,r._refCounter=n.refCounter,r.addField("scaleType","identity"),r}return no(e,t),e.prototype.create=function(){var t=this._references,e=this._refCounter;return this.toBaseScale(function(n,r){if(null==n){var i=t.get(r);null==i&&(i=e(),t.set(r,i))}else i=n;return i})},e}(eo),io=(Gi=function(t,e){return(Gi=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}Gi(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),oo=function(){return(oo=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)},ao=oo({},s,u),so=function(t){function e(e,n){var r=t.call(this,e,n)||this,i="measure"!==n.dimType,o=r.scaleConfig.brewer||(i?qt(20).map(function(t){return"color20-"+(1+t)}):["#eee","#000"]),a=r.scaleConfig;if(!i){var s=ao.extent(r.vars),u=function(t){return Number.isFinite(t)||kt(t)},c=u(a.min)?a.min:s[0],l=u(a.max)?a.max:s[1],d=[c,s[0]].filter(u),p=[l,s[1]].filter(u);if(s=[d.sort(function(t,e){return t-e})[0],p.sort(function(t,e){return e-t})[0]],a.nice&&s[0]<0&&s[1]>0){var f=Math.max.apply(Math,s.map(Math.abs));s=[-f,f]}r.vars=s}return r.addField("scaleType","color").addField("discrete",i).addField("brewer",o).addField("toColor",zt).addField("toClass",Ht),r}return io(e,t),e.prototype.create=function(){var t=this.discrete,e=this.vars,n=this.getField("brewer"),r=t?this.createDiscreteScale(e,n):this.createContinuesScale(e,n);return this.toBaseScale(r)},e.prototype.createDiscreteScale=function(t,e){var n,r,i,o,a,s=function(t,e){var n=t.map(function(t){return String(t).toString()});return ao.scaleOrdinal().range(e).domain(n)},u=function(t){return function(e){return t(String(e).toString())}};if(Array.isArray(e))n=u(s(t,e));else if("function"==typeof e)n=function(n){return e(n,u(s(t,qt(20).map(function(t){return"color20-"+(1+t)}))))};else{if(!At(e))throw new Error("This brewer is not supported");r=e,i=Object.keys(r),o=i.map(function(t){return r[t]}),a=ao.scaleOrdinal().range(o).domain(i),n=function(t){return r.hasOwnProperty(t)?a(String(t)):"color-default"}}return n},e.prototype.createContinuesScale=function(t,e){if(!Array.isArray(e))throw new Error("This brewer is not supported");return ao.scaleLinear().domain(Vt(t.map(function(t){return t-0}),e.length)).range(e)},e}(eo),uo=(qi=function(t,e){return(qi=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}qi(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),co={sqrt:function(t){return Math.sqrt(t)},linear:function(t){return t}},lo=function(t){function e(e,n){var r=t.call(this,e,n)||this,i=r.scaleConfig,o=s.extent(r.vars),a=Number.isFinite(i.min)?i.min:o[0],u=Number.isFinite(i.max)?i.max:o[1];return r.vars=[Math.min.apply(Math,[a,o[0]].filter(Number.isFinite)),Math.max.apply(Math,[u,o[1]].filter(Number.isFinite))],r.addField("scaleType","size"),r.addField("funcType",n.func||"sqrt"),r}return uo(e,t),e.prototype.isInDomain=function(t){var e=this.domain().sort(),n=e[0],r=e[e.length-1];return!Number.isNaN(n)&&!Number.isNaN(r)&&t<=r&&t>=n},e.prototype.create=function(){var t,e=this.scaleConfig,n=this.vars,r=Jt({},e,{func:"sqrt",minSize:0,maxSize:1}),i=r.func,o=r.minSize,a=r.maxSize,s=co[i],u=n.filter(function(t){return Number.isFinite(Number(t))});if(0===u.length)t=function(){return a};else{var c,l,d=Math.min.apply(Math,u),p=Math.max.apply(Math,u),f=s(Math.max(Math.abs(d),Math.abs(p),p-d));l=d<0?d:0,c=0===f?1:(a-o)/f,t=function(t){var e=null!==t?parseFloat(t):0;return Number.isFinite(e)?o+s(e-l)*c:a}}return this.toBaseScale(t)},e}(eo),po=(Wi=function(t,e){return(Wi=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}Wi(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),fo=function(t){function e(e,n){var r=t.call(this,e,n)||this;return r.addField("scaleType","ordinal").addField("discrete",!0),r}return po(e,t),e.prototype.create=function(t){var e=this.scaleConfig,n=this.vars,r=u.scalePoint().domain(n).range(t).padding(.5),i=Math.max.apply(Math,t),o=function(t){return"function"==typeof e.ratio?e.ratio(t,i,n):"object"==typeof e.ratio?e.ratio[t]:1/n.length},a=function(t){return e.ratio?i-n.slice(n.indexOf(t)+1).reduce(function(t,e){return t+i*o(e)},i*o(t)*.5):r(t)};return Object.keys(r).forEach(function(t){return a[t]=r[t]}),a.stepSize=function(t){return o(t)*i},this.toBaseScale(a,t)},e}(eo),ho=(Ui=function(t,e){return(Ui=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}Ui(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),mo=function(){return(mo=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)},go=mo({},s,u),yo=function(t){function e(e,n){var r=t.call(this,e,n)||this,i=r.scaleConfig,o=r.vars,a=go.extent(o),s=null===i.min||void 0===i.min?a[0]:new Date(i.min).getTime(),u=null===i.max||void 0===i.max?a[1]:new Date(i.max).getTime(),c=[new Date(Math.min(s,a[0])),new Date(Math.max(u,a[1]))],l=x.get(i.period,{utc:i.utcTime});return i.fitToFrameByDims||null===l?r.vars=Xt(o.map(function(t){return new Date(t)}),function(t){return t.getTime()}).sort(function(t,e){return Number(e)-Number(t)}):r.vars=x.generate(c[0],c[1],i.period,{utc:i.utcTime}),r.periodGenerator=l,r.addField("scaleType","period").addField("utcTime",r.scaleConfig.utcTime).addField("period",r.scaleConfig.period).addField("discrete",!0),r}return ho(e,t),e.prototype.isInDomain=function(t){var e=this.periodGenerator,n=new Date(t),r=(e?e.cast(n):n).getTime();return this.domain().map(function(t){return t.getTime()}).indexOf(r)>=0},e.prototype.create=function(t){var e=this.periodGenerator,n=this.vars,r=this.vars.map(function(t){return t.getTime()}),i=this.scaleConfig,o=go.scalePoint().domain(n).range(t).padding(.5),a=go.scalePoint().domain(r.map(String)).range(t).padding(.5),s=Math.max.apply(Math,t),u=function(t){var e=new Date(t).getTime();return"function"==typeof i.ratio?i.ratio(e,s,r):"object"==typeof i.ratio?i.ratio[e]:1/n.length},c=function(t){var n=new Date(t),o=(e?e.cast(n):n).getTime();return i.ratio?s-r.slice(r.indexOf(o)+1).reduce(function(t,e){return t+s*u(e)},s*u(t)*.5):a(String(o))};return Object.keys(o).forEach(function(t){return c[t]=o[t]}),c.stepSize=function(t){return u(t)*s},this.toBaseScale(c,t)},e}(eo),_o=n(15),vo=(Hi=function(t,e){return(Hi=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}Hi(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),bo=function(){return(bo=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)},xo=bo({},s,u,_o),wo=function(t){function e(e,n){var r,i=t.call(this,e,n)||this,o=i.scaleConfig,a=i.vars,s=o.period?x.get(i.scaleConfig.period,{utc:o.utcTime}):null,u=xo.extent(a).map(s?function(t){return s.cast(new Date(t))}:function(t){return new Date(t)}),c=null==o.min?u[0]:new Date(o.min).getTime(),l=null==o.max?u[1]:new Date(o.max).getTime();if(a=[new Date(Math.min(c,Number(u[0]))),new Date(Math.max(l,Number(u[1])))],i.niceIntervalFn=null,o.nice&&!s){var d=o.niceInterval,p=d&&jo(d)?o.utcTime?xo["utc"+(r=d)[0].toUpperCase()+r.slice(1)]:jo(d):null;i.niceIntervalFn=p||null,i.vars=It(a,i.niceIntervalFn,{utc:o.utcTime})}else i.vars=a;if(s&&Number(i.vars[0])===Number(i.vars[1])){var f=i.vars[0];i.vars[0]=s.cast(new Date(Number(f)-1)),i.vars[1]=s.next(f)}return i.periodGenerator=s,i.addField("scaleType","time").addField("utcTime",i.scaleConfig.utcTime).addField("period",i.scaleConfig.period),i}return vo(e,t),e.prototype.isInDomain=function(t){var e=new Date(t);this.scaleConfig.period&&(e=this.periodGenerator.cast(e));var n=this.domain(),r=n[0],i=n[n.length-1];return!Number.isNaN(r)&&!Number.isNaN(i)&&e<=i&&e>=r},e.prototype.create=function(t){var e=this,n=this.vars,r=this.scaleConfig.utcTime,i=this.periodGenerator,o=(r?xo.scaleUtc:xo.scaleTime)().domain(n).range(t),a=function(t){var e=n[0],r=n[1];return t>r&&(t=r),t<e&&(t=e),o(new Date(t))};if(this.scaleConfig.period){var s=n[0],u=n[1];o.ticks,o.ticks=function(t){return"number"!=typeof t&&(t=10),function(t,e,n,r){void 0===r&&(r=10);var i=t[0],o=t[1],a=x.get(e,{utc:n}),s=Number(i),u=Number(o),c=Number(a.next(a.cast(i)))-Number(a.cast(i));return(u-s)/c<=r?x.generate(i,o,e,{utc:n}).filter(function(t){return t>=i&&t<=o}):function(t,e,n){void 0===n&&(n=10);var r,i,o=Number(t[0]),a=Number(t[1]),s=Math.abs(a-o)/n,u=xo.bisector(function(t){return t.duration}).right(Mo,s);if(u===Mo.length)r=e?xo.utcYear:xo.timeYear,i=xo.tickStep(o/So.year.duration,a/So.year.duration,n);else if(u){var c=s/Mo[u-1].duration,l=Mo[u].duration/s,d=Mo[c<l?u-1:u];r=e?d.time.utc:d.time.interval,i=d.step}else r=e?xo.utcMillisecond:xo.timeMillisecond,i=xo.tickStep(o,a,n);return r.every(i).range(new Date(o),new Date(a+1))}(t,n,r)}([s,u],e.scaleConfig.period,r,t)};for(var c=i.cast(s);c<s;)c=i.next(c);var l=i.cast(u);a=function(t){var e=i.cast(t);return e<c&&(e=c),e>l&&(e=l),o(e)}}return Object.keys(o).forEach(function(t){return a[t]=o[t]}),a.stepSize=function(){return 0},this.toBaseScale(a,t)},e}(eo);function jo(t){return xo["time"+t[0].toUpperCase()+t.slice(1)]}var Oo,Co,Eo,So={second:{duration:1e3,interval:xo.timeSecond,utc:xo.utcSecond},minute:{duration:6e4,interval:xo.timeMinute,utc:xo.utcMinute},hour:{duration:36e5,interval:xo.timeHour,utc:xo.utcHour},day:{duration:864e5,interval:xo.timeDay,utc:xo.utcDay},week:{duration:6048e5,interval:xo.timeWeek,utc:xo.utcWeek},month:{duration:2592e6,interval:xo.timeMonth,utc:xo.utcMonth},year:{duration:31536e6,interval:xo.timeYear,utc:xo.utcYear}},Mo=[(Oo=function(t,e){return{time:t,step:e,duration:e*t.duration}})(So.second,1),Oo(So.second,5),Oo(So.second,15),Oo(So.second,30),Oo(So.minute,1),Oo(So.minute,5),Oo(So.minute,15),Oo(So.minute,30),Oo(So.hour,1),Oo(So.hour,3),Oo(So.hour,6),Oo(So.hour,12),Oo(So.day,1),Oo(So.day,2),Oo(So.week,1),Oo(So.month,1),Oo(So.month,3),Oo(So.year,1)],To=(Eo=function(t,e){return(Eo=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}Eo(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),ko=function(){return(ko=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)},Ao=ko({},s,u),Do=function(t){function e(e,n){var r=t.call(this,e,n)||this,i=r.scaleConfig,o=Ao.extent(r.vars),a=Number.isFinite(i.min)?i.min:o[0],s=Number.isFinite(i.max)?i.max:o[1];if(o=[Math.min.apply(Math,[a,o[0]].filter(Number.isFinite)),Math.max.apply(Math,[s,o[1]].filter(Number.isFinite))],r.vars=i.nice?Dt(o):Ao.extent(o),r.vars[0]===r.vars[1]){var u=Math.pow(10,Math.floor(Math.log(r.vars[0])/Math.LN10));r.vars[0]-=u,r.vars[1]+=u||10}return r.addField("scaleType","linear").addField("discrete",!1),r}return To(e,t),e.prototype.isInDomain=function(t){var e=this.domain(),n=e[0],r=e[e.length-1];return!Number.isNaN(n)&&!Number.isNaN(r)&&t<=r&&t>=n},e.prototype.create=function(t){var e=this.vars,n=this.extendScale(Ao.scaleLinear());return n.domain(e).range(t).clamp(!0),this.toBaseScale(n,t)},e.prototype.extendScale=function(t){var e=this,n=t.copy,r=t.ticks;return Object.assign(t,{stepSize:function(){return 0},copy:function(){return e.extendScale(n.call(t))},ticks:this.getField("isInteger")?function(e){return r.call(t,e).filter(Number.isInteger)}:t.ticks}),t},e}(eo),Io=(Co=function(t,e){return(Co=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}Co(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),Po=function(){return(Po=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)},Ro=Po({},s,u),No=function(t){function e(e,n){var r,i,o,a,s,u,c,l,d=t.call(this,e,n)||this,p=d.scaleConfig,f=Ro.extent(d.vars),h=Number.isFinite(p.min)?p.min:f[0],m=Number.isFinite(p.max)?p.max:f[1];return Lo(f=[Math.min.apply(Math,[h,f[0]].filter(Number.isFinite)),Math.max.apply(Math,[m,f[1]].filter(Number.isFinite))]),p.nice&&(i=(r=f)[0]>0,o=r.map(function(t){return Math.abs(t)}),a=Math.max.apply(Math,o),s=Math.min.apply(Math,o).toExponential().split("e"),u=a.toExponential().split("e"),c=parseFloat(Math.floor(Number(s[0]))+"e"+s[1]),l=parseFloat(Math.ceil(Number(u[0]))+"e"+u[1]),f=i?[c,l]:[-l,-c]),d.vars=f,d.addField("scaleType","logarithmic").addField("discrete",!1),d}return Io(e,t),e.prototype.isInDomain=function(t){var e=this.domain(),n=e[0],r=e[e.length-1];return!Number.isNaN(n)&&!Number.isNaN(r)&&t<=r&&t>=n},e.prototype.create=function(t){var e=this.vars;Lo(e);var n=function t(e){var n=e.copy;return e.ticks=function(t){for(var n=[],r=Ro.extent(e.domain()),i=Math.floor(Fo(r[0])),o=Math.ceil(Fo(r[1])),a=Math.ceil(10*(o-i)/(10*Math.ceil(t/10))),s=i;s<=o;s+=a)for(var u=1;u<=10;u++){var c=Math.pow(u,a)*Math.pow(10,s);(c=parseFloat(c.toExponential(0)))>=r[0]&&c<=r[1]&&n.push(c)}return n},e.copy=function(){var r=n.call(e);return t(r),r},e}(Ro.scaleLog()).domain(e).range(t);return n.stepSize=function(){return 0},this.toBaseScale(n,t)},e}(eo);function Fo(t){return Math.log(t)/Math.LN10}function Lo(t){if(t[0]*t[1]<=0)throw new j("Logarithmic scale domain cannot cross zero.",O.INVALID_LOG_DOMAIN)}var Bo,Vo,zo,Ho=(zo=function(t,e){return(zo=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}zo(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),Uo=function(t){function e(e,n){var r=t.call(this,e,n)||this;return r.addField("scaleType","value").addField("georole",n.georole),r}return Ho(e,t),e.prototype.create=function(){return this.toBaseScale(function(t){return t})},e}(eo),Wo=(Vo=function(t,e){return(Vo=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)},function(t,e){function n(){this.constructor=t}Vo(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}),qo=function(){return(qo=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)},Go=qo({},s,u),Yo=function(t){function e(e,n){var r=t.call(this,e,n)||this,i=r.scaleConfig,o=Go.extent(r.vars),a=Number.isFinite(i.min)?i.min:o[0],s=Number.isFinite(i.max)?i.max:o[1];o=[Math.min(a,o[0]),Math.max(s,o[1])],r.vars=i.nice?Dt(o):Go.extent(o);var u=qt(10).map(function(t){return"rgba(90,180,90,"+(.2+t*(.8/9)).toFixed(2)+")"}),c=i.brewer||u;return r.addField("scaleType","fill").addField("brewer",c),r}return Wo(e,t),e.prototype.isInDomain=function(t){var e=this.domain(),n=e[0],r=e[e.length-1];return!Number.isNaN(n)&&!Number.isNaN(r)&&t<=r&&t>=n},e.prototype.create=function(){var t=this.vars,e=this.getField("brewer");if(!Array.isArray(e))throw new Error("This brewer is not supported");var n=e.length,r=(t[1]-t[0])/n,i=qt(n-1).map(function(t){return t+1}).reduce(function(e,n){return e.concat([t[0]+n*r])},[]),o=Go.scaleThreshold().domain(i).range(e);return this.toBaseScale(o)},e}(eo),Ko=function(t){return t||null},Xo=function(t,e){return void 0===e&&(e=null),Array.isArray(t)?0===t.length?[e]:t:[t]},Qo=function(t,e){return{type:t,x:e.x,y:e.y,identity:e.identity,size:e.size,color:e.color,split:e.split,label:e.label,guide:{color:e.colorGuide,obsoleteVerticalStackOrder:e.obsoleteVerticalStackOrder,size:e.sizeGuide},flip:e.flip,stack:e.stack}},Zo={SUCCESS:"SUCCESS",WARNING:"WARNING",FAIL:"FAIL"},$o=((Bo={})[Zo.SUCCESS]=function(t){return t},Bo[Zo.FAIL]=function(t,e){throw new Error((e.messages||[]).join("\n")||["This configuration is not supported,","See https://api.taucharts.com/basic/facet.html#easy-approach-for-creating-facet-chart"].join(" "))},Bo[Zo.WARNING]=function(t,e,n){var r=e.axis,i=e.indexMeasureAxis[0],o=t[i],a=t.filter(function(t){return t!==o});a.push(o);var s=n[i][r]||{},u=n[n.length-1][r]||{};return n[n.length-1][r]=s,n[i][r]=u,a},Bo);function Jo(t,e,n){return e.reduce(function(e,r,i){var o=t[r];return o?e.status!=Zo.FAIL&&("measure"===o.type&&(e.countMeasureAxis++,e.indexMeasureAxis.push(i)),"measure"!==o.type&&1===e.countMeasureAxis?e.status=Zo.WARNING:e.countMeasureAxis>1&&(e.status=Zo.FAIL,e.messages.push('There is more than one measure dimension for "'+n+'" axis'))):(e.status=Zo.FAIL,r?e.messages.push('"'+r+'" dimension is undefined for "'+n+'" axis'):e.messages.push('"'+n+'" axis should be specified')),e},{status:Zo.SUCCESS,countMeasureAxis:0,indexMeasureAxis:[],messages:[],axis:n})}function ta(t){for(var e=Xo(t.x),n=Xo(t.y),r=Math.max(e.length,n.length),i=Xo(t.guide||{},{}),o=r-i.length,a=0;a<o;a++)i.push({});i=i.slice(0,r);var s=Jo(t.dimensions,e,"x"),u=Jo(t.dimensions,n,"y");e=$o[s.status](e,s,i),n=$o[u.status](n,u,i);var c=["identity","color","size","label","split"].reduce(function(e,n){var r,i=t[n],o=(t.dimensions,null==(r=i)?r:String(r));return null!=o&&(e[n]=o),e},{});return Object.assign({},t,{x:e,y:n,guide:i},c)}function ea(t,e){for(var n=e.x,r=e.y,i=e.guide,o=Math.max(n.length,r.length),a={type:"COORDS.RECT",unit:[]},s=[].concat(n),u=[].concat(r),c=[].concat(i),l=o;l>0;l--){var d=s.pop(),p=u.pop(),f=c.pop()||{};l===o?(a.x=d,a.y=p,a.unit.push(Qo(t,{x:Ko(d),y:Ko(p),identity:e.identity,split:e.split,color:e.color,label:e.label,size:e.size,flip:e.flip,stack:e.stack,colorGuide:f.color,obsoleteVerticalStackOrder:f.obsoleteVerticalStackOrder,sizeGuide:f.size})),a.guide=Jt(f,{x:{label:d},y:{label:p}})):a={type:"COORDS.RECT",x:Ko(d),y:Ko(p),unit:[a],guide:Jt(f,{x:{label:d},y:{label:p}})}}return e.spec={dimensions:e.dimensions,unit:a},e}var na=function(t){var e,n,r,i=ta(t);return n=((e=i).flip?e.y:e.x).indexOf(e.color)>=0,r=e.guide[e.guide.length-1],n&&!r.hasOwnProperty("enableColorToBarPosition")&&(r.enableColorToBarPosition=!1),ea("ELEMENT.INTERVAL",i=e)},ra=function(t){var e=ta(t),n=e.data,r=e.settings.log,i=(0,{horizontal:function(t){return{prop:t.x[t.x.length-1],flip:!1}},vertical:function(t){return{prop:t.y[t.y.length-1],flip:!0}},auto:function(t){var e,i=t.x,o=t.y,a=i[i.length-1],s=i.slice(0,i.length-1),u=o[o.length-1],c=o.slice(0,o.length-1),l=t.color,d=s.concat(c).concat([l]).filter(function(t){return null!==t}),p=-1,f=[[[a].concat(d),u],[[u].concat(d),a]],h=null;return f.some(function(t,e){var i=t[0],o=t[1],a=dn(n,i,[o]);return a.result?p=e:r(["Attempt to find a functional relation between",t[0]+" and "+t[1]+" is failed.","There are several "+a.error.keyY+" values (e.g. "+a.error.errY.join(",")+")","for ("+a.error.keyX+" = "+a.error.valX+")."].join(" ")),a.result})?(e=f[p][0][0],h=0!==p):(r("All attempts are failed. Gonna transform AREA to general PATH."),e=null),{prop:e,flip:h}}}["boolean"!=typeof e.flip?"auto":e.flip?"vertical":"horizontal"])(e);return null!==i.prop&&(e.data=pn(n,i.prop,e.dimensions[i.prop]),e.flip=i.flip),ea("ELEMENT.AREA",e)},ia=function(){function t(t){this.unitRef=t}return t.prototype.value=function(){return this.unitRef},t.prototype.clone=function(){return JSON.parse(JSON.stringify(this.unitRef))},t.prototype.traverse=function(t){var e=function(t,n,r){n(t,r),(t.units||[]).map(function(r){return e(r,n,t)})};return e(this.unitRef,t,null),this},t.prototype.reduce=function(t,e){var n=e;return this.traverse(function(e,r){return n=t(n,e,r)}),n},t.prototype.addFrame=function(t){return this.unitRef.frames=this.unitRef.frames||[],t.key.__layerid__=["L",(new Date).getTime(),this.unitRef.frames.length].join(""),t.source=t.hasOwnProperty("source")?t.source:this.unitRef.expression.source,t.pipe=t.pipe||[],this.unitRef.frames.push(t),this},t.prototype.addTransformation=function(t,e){return this.unitRef.transformation=this.unitRef.transformation||[],this.unitRef.transformation.push({type:t,args:e}),this},t.prototype.isCoordinates=function(){return 0===(this.unitRef.type||"").toUpperCase().indexOf("COORDS.")},t.prototype.isElementOf=function(t){if(this.isCoordinates())return!1;var e=(this.unitRef.type||"").split("/");return 1===e.length&&e.unshift("RECT"),e[0].toUpperCase()===t.toUpperCase()},t}(),oa=function(){function t(t){this.specRef=t}return t.prototype.value=function(){return this.specRef},t.prototype.unit=function(t){return t&&(this.specRef.unit=t),new ia(this.specRef.unit)},t.prototype.addTransformation=function(t,e){return this.specRef.transformations=this.specRef.transformations||{},this.specRef.transformations[t]=e,this},t.prototype.getSettings=function(t){return this.specRef.settings[t]},t.prototype.setSettings=function(t,e){return this.specRef.settings=this.specRef.settings||{},this.specRef.settings[t]=e,this},t.prototype.getScale=function(t){return this.specRef.scales[t]},t.prototype.addScale=function(t,e){return this.specRef.scales[t]=e,this},t.prototype.regSource=function(t,e){return this.specRef.sources[t]=e,this},t.prototype.getSourceData=function(t){return(this.specRef.sources[t]||{data:[]}).data},t.prototype.getSourceDim=function(t,e){return(this.specRef.sources[t]||{dims:{}}).dims[e]||{}},t}(),aa={},sa=function(){function t(){}return t.unit=function(t){return new ia(t)},t.spec=function(t){return new oa(t)},t.cloneObject=function(t){return JSON.parse(JSON.stringify(t))},t.depthFirstSearch=function(e,n){if(n(e))return e;for(var r=e.hasOwnProperty("frames")?e.frames:[{units:e.units}],i=0;i<r.length;i++)for(var o=r[i].units||[],a=0;a<o.length;a++){var s=t.depthFirstSearch(o[a],n);if(s)return s}},t.traverseSpec=function(t,e){var n=function(t,e,r){e(t,r),(t.units||[]).map(function(r){return n(r,e,t)})};n(t.unit,e,null)},t.extractFieldsFormatInfo=function(e){var n=e.scales,r=function(t){return!n[t].dim},i=function(t,e,r){var i=e.guide||{},o=n[e[r]],a=i[r]||{};t[o.dim]=t[o.dim]||{label:[],format:[],nullAlias:[],tickLabel:[]};var s=a.label,u=a.label||{};t[o.dim].label.push("string"==typeof s?s:u._original_text||u.text);var c=a.tickFormat||a.tickPeriod;t[o.dim].format.push(c),t[o.dim].nullAlias.push(a.tickFormatNullAlias),t[o.dim].tickLabel.push(a.tickLabel)},o=[];t.traverseSpec(e,function(t){o.push(t)});var a=o.reduce(function(t,e){return"COORDS.RECT"===e.type&&e.hasOwnProperty("x")&&!r(e.x)&&i(t,e,"x"),"COORDS.RECT"===e.type&&e.hasOwnProperty("y")&&!r(e.y)&&i(t,e,"y"),e.hasOwnProperty("color")&&!r(e.color)&&i(t,e,"color"),e.hasOwnProperty("size")&&!r(e.size)&&i(t,e,"size"),e.hasOwnProperty("label")&&!r(e.label)&&i(t,e,"label"),t},{}),s=function(t,e){return t.filter(function(t){return t})[0]||e};return Object.keys(a).reduce(function(t,n){t[n]={},t[n].label=s(a[n].label,n);var r=s(a[n].format,null);t[n].nullAlias=s(a[n].nullAlias,"No "+t[n].label),t[n].tickLabel=s(a[n].tickLabel,null);var i="x-time-auto"===r?e.settings.utcTime?"day-utc":"day":r,o=t[n].nullAlias,u=i?Cn.get(i,o):function(t){return null==t?o:String(t)};if(t[n].format=u,t[n].tickLabel){var c=n.replace("."+t[n].tickLabel,"");t[c]={label:t[n].label,nullAlias:t[n].nullAlias,tickLabel:t[n].tickLabel,format:function(e){return u(e&&e[t[c].tickLabel])},isComplexField:!0},t[n].parentField=c}return t},{})},t.getFieldFormatters=function(e,n){var r=t.extractFieldsFormatInfo(e);return Object.keys(r).forEach(function(t){r[t].parentField&&delete r[t]}),Object.keys(n).forEach(function(t){var e,i,o=(e=n[t],i={},"function"==typeof e||"string"==typeof e?i={format:e}:At(e)&&(i=re(e,"label","format","nullAlias")),i);r[t]=Object.assign({label:t,nullAlias:"No "+t},r[t]||{},re(o,"label","nullAlias")),o.hasOwnProperty("format")?r[t].format="function"==typeof o.format?o.format:Cn.get(o.format,r[t].nullAlias):r[t].format=r[t].hasOwnProperty("format")?r[t].format:Cn.get(null,r[t].nullAlias)}),Object.keys(r).reduce(function(t,e){var n=r[e];return t[e]={label:n.label,format:n.format},t},{})},t.tokens=function(){return{reg:function(t,e){return aa[t]=e,this},get:function(t){return aa[t]||t}}},t.getParentUnit=function(t,e){var n=null,r=function(t,i){return t.uid===e.uid?(n=i,!0):(t.frames?t.frames.some(function(e){return(e.units||[]).some(function(e){return r(e,t)})}):(t.units||[]).some(function(e){return r(e,t)}),!1)};return r(t.unit,null),n},t}();n(16),n.d(e,"api",function(){return la}),n.d(e,"version",function(){return pa}),n.d(e,"GPL",function(){return Ke}),n.d(e,"Plot",function(){return ar}),n.d(e,"Chart",function(){return pr});var ua={},ca={},la={errorCodes:O,unitsRegistry:an,scalesRegistry:cn,grammarRegistry:E,tickFormat:Cn,isChartElement:fe,utils:r,svgUtils:o,domUtils:i,tickPeriod:x,colorBrewers:{add:function(t,e){ua.hasOwnProperty(t)||(ua[t]=e)},get:function(t){return ua[t]}},d3_animationInterceptor:ut,pluginsSDK:sa,plugins:{add:function(t,e){if(ca.hasOwnProperty(t))throw new Error("Plugin is already registered.");ca[t]=e},get:function(t){return ca[t]||function(e){throw new Error('"'+t+'" plugin is not defined')}}},chartTypesRegistry:lr,globalSettings:{animationSpeed:750,renderingTimeout:1e4,asyncRendering:!0,syncRenderingInterval:50,syncPointerEvents:!1,handleRenderingErrors:!1,experimentalShouldAnimate:function(t){var e=function(t,e){var n=document.createElementNS("http://www.w3.org/2000/svg",t);return Object.keys(e).forEach(function(t){return n.setAttribute(t,String(e[t]))}),n},n=document.createElement("div");n.style.position="absolute",n.style.visibility="hidden",document.body.appendChild(n);var r=e("svg",{width:100,height:100});n.appendChild(r);var i,o,a,s=performance.now();for(i=0;i<10;i++)for(o=0;o<10;o++)a=e("circle",{fill:"black",r:5,cx:10*i,cy:10*o}),r.appendChild(a);var u=performance.now()-s;return document.body.removeChild(n),t.sources["/"].data.length*u<500},defaultNiceColor:!0,defaultColorBrewer:["#fde725","#fbe723","#f8e621","#f6e620","#f4e61e","#f1e51d","#efe51c","#ece51b","#eae51a","#e7e419","#e5e419","#e2e418","#dfe318","#dde318","#dae319","#d8e219","#d5e21a","#d2e21b","#d0e11c","#cde11d","#cae11f","#c8e020","#c5e021","#c2df23","#c0df25","#bddf26","#bade28","#b8de29","#b5de2b","#b2dd2d","#b0dd2f","#addc30","#aadc32","#a8db34","#a5db36","#a2da37","#a0da39","#9dd93b","#9bd93c","#98d83e","#95d840","#93d741","#90d743","#8ed645","#8bd646","#89d548","#86d549","#84d44b","#81d34d","#7fd34e","#7cd250","#7ad151","#77d153","#75d054","#73d056","#70cf57","#6ece58","#6ccd5a","#69cd5b","#67cc5c","#65cb5e","#63cb5f","#60ca60","#5ec962","#5cc863","#5ac864","#58c765","#56c667","#54c568","#52c569","#50c46a","#4ec36b","#4cc26c","#4ac16d","#48c16e","#46c06f","#44bf70","#42be71","#40bd72","#3fbc73","#3dbc74","#3bbb75","#3aba76","#38b977","#37b878","#35b779","#34b679","#32b67a","#31b57b","#2fb47c","#2eb37c","#2db27d","#2cb17e","#2ab07f","#29af7f","#28ae80","#27ad81","#26ad81","#25ac82","#25ab82","#24aa83","#23a983","#22a884","#22a785","#21a685","#21a585","#20a486","#20a386","#1fa287","#1fa187","#1fa188","#1fa088","#1f9f88","#1f9e89","#1e9d89","#1e9c89","#1e9b8a","#1f9a8a","#1f998a","#1f988b","#1f978b","#1f968b","#1f958b","#1f948c","#20938c","#20928c","#20928c","#21918c","#21908d","#218f8d","#218e8d","#228d8d","#228c8d","#228b8d","#238a8d","#23898e","#23888e","#24878e","#24868e","#25858e","#25848e","#25838e","#26828e","#26828e","#26818e","#27808e","#277f8e","#277e8e","#287d8e","#287c8e","#297b8e","#297a8e","#29798e","#2a788e","#2a778e","#2a768e","#2b758e","#2b748e","#2c738e","#2c728e","#2c718e","#2d718e","#2d708e","#2e6f8e","#2e6e8e","#2e6d8e","#2f6c8e","#2f6b8e","#306a8e","#30698e","#31688e","#31678e","#31668e","#32658e","#32648e","#33638d","#33628d","#34618d","#34608d","#355f8d","#355e8d","#365d8d","#365c8d","#375b8d","#375a8c","#38598c","#38588c","#39568c","#39558c","#3a548c","#3a538b","#3b528b","#3b518b","#3c508b","#3c4f8a","#3d4e8a","#3d4d8a","#3e4c8a","#3e4a89","#3e4989","#3f4889","#3f4788","#404688","#404588","#414487","#414287","#424186","#424086","#423f85","#433e85","#433d84","#443b84","#443a83","#443983","#453882","#453781","#453581","#463480","#46337f","#46327e","#46307e","#472f7d","#472e7c","#472d7b","#472c7a","#472a7a","#482979","#482878","#482677","#482576","#482475","#482374","#482173","#482071","#481f70","#481d6f","#481c6e","#481b6d","#481a6c","#48186a","#481769","#481668","#481467","#471365","#471164","#471063","#470e61","#470d60","#460b5e","#460a5d","#46085c","#46075a","#450559","#450457","#440256","#440154"],defaultClassBrewer:qt(20).map(function(t){return"color20-"+(1+t)}),log:function(t,e){e=e||"INFO",Array.isArray(t)||(t=[t]),console[e.toLowerCase()].apply(console,t)},facetLabelDelimiter:" → ",excludeNull:!0,minChartWidth:300,minChartHeight:200,minFacetWidth:150,minFacetHeight:100,specEngine:[{name:"COMPACT",width:600,height:400},{name:"AUTO",width:Number.MAX_VALUE,height:Number.MAX_VALUE}],fitModel:"normal",layoutEngine:"EXTRACT",autoRatio:!0,defaultSourceMap:["https://raw.githubusercontent.com","TargetProcess/tauCharts/master/src/addons","world-countries.json"].join("/"),getAxisTickLabelSize:ee(je,function(t){return String(String(t).length)}),getScrollbarSize:_e,avoidScrollAtRatio:1.5,xAxisTickLabelLimit:150,yAxisTickLabelLimit:150,xTickWordWrapLinesLimit:2,yTickWordWrapLinesLimit:2,xTickWidth:9,yTickWidth:9,distToXAxisLabel:10,distToYAxisLabel:10,xAxisPadding:20,yAxisPadding:20,xFontLabelDescenderLineHeight:4,xFontLabelHeight:10,yFontLabelHeight:10,xDensityPadding:2,yDensityPadding:2,"xDensityPadding:measure":8,"yDensityPadding:measure":8,utcTime:!1,defaultFormats:{measure:"x-num-auto","measure:time":"x-time-auto"}}};ar.__api__=la,ar.globalSettings=la.globalSettings,la.unitsRegistry.reg("COORDS.RECT",Er).reg("COORDS.MAP",Vr).reg("COORDS.PARALLEL",Dr).reg("ELEMENT.GENERIC.CARTESIAN",ht).reg("ELEMENT.POINT",ui,"ELEMENT.GENERIC.CARTESIAN").reg("ELEMENT.LINE",Xi,"ELEMENT.GENERIC.CARTESIAN").reg("ELEMENT.PATH",Bi,"ELEMENT.GENERIC.CARTESIAN").reg("ELEMENT.AREA",Li,"ELEMENT.GENERIC.CARTESIAN").reg("ELEMENT.INTERVAL",Zi,"ELEMENT.GENERIC.CARTESIAN").reg("ELEMENT.INTERVAL.STACKED",Zi,"ELEMENT.GENERIC.CARTESIAN").reg("ELEMENT.INTERVAL.STACKED",Zi,"ELEMENT.GENERIC.CARTESIAN").reg("PARALLEL/ELEMENT.LINE",to),la.scalesRegistry.reg("identity",ro,function(t,e){return Jt(t,{references:e.references,refCounter:e.refCounter})}).reg("color",so,function(t,e){return Jt(t,{nice:e.defaultNiceColor,brewer:"measure"===t.dimType?e.defaultColorBrewer:e.defaultClassBrewer})}).reg("fill",Yo).reg("size",lo).reg("ordinal",fo).reg("period",yo,function(t,e){return Jt(t,{utcTime:e.utcTime})}).reg("time",wo,function(t,e){return Jt(t,{utcTime:e.utcTime})}).reg("linear",Do).reg("logarithmic",No).reg("value",Uo);var da=[function(t){return t.data?[]:["[data] must be specified"]}];la.chartTypesRegistry.add("scatterplot",function(t){return ea("ELEMENT.POINT",ta(t))},da).add("line",function(t){var e=ta(t),n=e.data,r=e.settings.log,i={none:function(){return null},horizontal:function(t){return t.x[t.x.length-1]},vertical:function(t){return t.y[t.y.length-1]},auto:function(t){var e,i=t.x,o=t.y,a=i[i.length-1],s=i.slice(0,i.length-1),u=o[o.length-1],c=o.slice(0,o.length-1),l=t.color,d=s.concat(c).concat([l]).filter(function(t){return null!==t}),p=-1,f=[[[a].concat(d),u],[[u].concat(d),a]];return f.some(function(t,e){var i=t[0],o=t[1],a=dn(n,i,[o]);return a.result?p=e:r(["Attempt to find a functional relation between",t[0]+" and "+t[1]+" is failed.","There are several "+a.error.keyY+" values (e.g. "+a.error.errY.join(",")+")","for ("+a.error.keyX+" = "+a.error.valX+")."].join(" ")),a.result})?e=f[p][0][0]:(r(["All attempts are failed.","Will orient line horizontally by default.","NOTE: the [scatterplot] chart is more convenient for that data."].join(" ")),e=a),e}},o=(e.lineOrientation||"").toLowerCase(),a=(i.hasOwnProperty(o)?i[o]:i.auto)(e);return null!==a&&(e.data=pn(n,a,e.dimensions[a])),ea("ELEMENT.LINE",e)},da).add("area",ra,da).add("stacked-area",function(t){return ra(Jt(t,{stack:!0}))},da).add("bar",function(t){return na(Jt(t,{flip:!1}))},da).add("horizontalBar",function(t){return na(Jt({flip:!0},t))},da).add("horizontal-bar",function(t){return na(Jt({flip:!0},t))},da).add("stacked-bar",function(t){return na(Jt({flip:!1,stack:!0},t))},da).add("horizontal-stacked-bar",function(t){return na(Jt({flip:!0,stack:!0},t))},da).add("map",function(t){var e=Object.assign({sourcemap:t.settings.defaultSourceMap},t.guide||{});e.size=Jt(e.size||{},{min:1,max:10}),e.code=Jt(e.code||{},{georole:"countries"});var n={},r=function(t,e,r){var i;void 0===r&&(r={});var o,a=e;return e?(i=t+"_"+e,o="/"):(i=t+":default",o="?"),n.hasOwnProperty(i)||(n[i]=Object.assign({type:t,source:o,dim:a},r)),i};return{sources:{"?":{dims:{},data:[{}]},"/":{dims:Object.keys(t.dimensions).reduce(function(e,n){return e[n]={type:t.dimensions[n].type},e},{}),data:t.data}},scales:n,unit:{type:"COORDS.MAP",expression:{operator:"none",source:"/"},code:r("value",t.code,e.code),fill:r("fill",t.fill,e.fill),size:r("size",t.size,e.size),color:r("color",t.color,e.color),latitude:r("linear",t.latitude,{nice:!1}),longitude:r("linear",t.longitude,{nice:!1}),guide:e},plugins:t.plugins||[]}},da.concat([function(t){var e=t.fill&&t.code;if(t.fill&&!e)return"[code] must be specified when using [fill]"},function(t){var e=t.latitude&&t.longitude;if((t.latitude||t.longitude)&&!e)return"[latitude] and [longitude] both must be specified"}])).add("parallel",function(t){var e=Object.assign({columns:{}},t.guide||{}),n={},r=function(t,e,r){var i;void 0===r&&(r={});var o,a=e;return e?(i=t+"_"+e,o="/"):(i=t+":default",o="?"),n.hasOwnProperty(i)||(n[i]=Object.assign({type:t,source:o,dim:a},r)),i},i=t.columns.map(function(n){return r(t.dimensions[n].scale,n,e.columns[n])});return{sources:{"?":{dims:{},data:[{}]},"/":{dims:Object.keys(t.dimensions).reduce(function(e,n){return e[n]={type:t.dimensions[n].type},e},{}),data:t.data}},scales:n,unit:{type:"COORDS.PARALLEL",expression:{operator:"none",source:"/"},columns:i,guide:e,units:[{type:"PARALLEL/ELEMENT.LINE",color:r("color",t.color,e.color),columns:i,expression:{operator:"none",source:"/"}}]},plugins:t.plugins||[]}},da.concat([function(t){if(!(t.columns&&t.columns.length>1))return"[columns] property must contain at least 2 dimensions"}]));var pa="2.7.1";e.default={GPL:Ke,Plot:ar,Chart:pr,api:la,version:"2.7.1"}},function(t,e){t.exports=i},function(t,e){t.exports=o},function(t,e){t.exports=a},function(t,e){t.exports=s},function(t,e){t.exports=u},function(t,e){t.exports=c},function(t,e){t.exports=l},function(t,e){t.exports=d},function(t,e,n){var r,i;void 0===(i="function"==typeof(r=function(){return function(t){var e={};function n(r){if(e[r])return e[r].exports;var i=e[r]={exports:{},id:r,loaded:!1};return t[r].call(i.exports,i,i.exports,n),i.loaded=!0,i.exports}return n.m=t,n.c=e,n.p="",n(0)}([function(t,e){"use strict";Object.defineProperty(e,"__esModule",{value:!0});var n=function(t,e){return t.indexOf(e)},r=window,i=r.document,o=i.documentElement,a=["top","bottom"];function s(t,e){for(var n in e)t[n]=e[n];return t}function u(t){var e,n=r.pageYOffset||o.scrollTop,i=r.pageXOffset||o.scrollLeft,a={left:0,right:0,top:0,bottom:0,width:0,height:0};if((e=t)&&null!=e.setInterval)a.width=r.innerWidth||o.clientWidth,a.height=r.innerHeight||o.clientHeight;else{if(!o.contains(t)||null==t.getBoundingClientRect)return a;s(a,t.getBoundingClientRect()),a.width=a.right-a.left,a.height=a.bottom-a.top}return a.top=a.top+n-o.clientTop,a.left=a.left+i-o.clientLeft,a.right=a.left+a.width,a.bottom=a.top+a.height,a}var c=function(t,e){return r.getComputedStyle(t)[e]};function l(t){var e=String(c(t,l.propName)),n=e.match(/([0-9.]+)([ms]{1,2})/);return n&&(e=Number(n[1]),"s"===n[2]&&(e*=1e3)),0|e}l.propName=function(){for(var t=i.createElement("div"),e=["transitionDuration","webkitTransitionDuration"],n=0;n<e.length;n++)if(t.style[e[n]]="1s","1s"===t.style[e[n]])return e[n]}();var d=Object.create;function p(t,e){if(!(this instanceof p))return new p(t,e);this.hidden=1,this.options=s(d(p.defaults),e),this._createElement(),t&&this.content(t)}p.prototype._createElement=function(){var t,e;this.element=i.createElement("div"),this.classes=(t=this.element,{add:function(e){t.classList.add(e)},remove:function(e){t.classList.remove(e)}}),this.classes.add(this.options.baseClass);for(var n=0;n<p.classTypes.length;n++)e=p.classTypes[n]+"Class",this.options[e]&&this.classes.add(this.options[e])},p.prototype.type=function(t){return this.changeClassType("type",t)},p.prototype.effect=function(t){return this.changeClassType("effect",t)},p.prototype.changeClassType=function(t,e){return t+="Class",this.options[t]&&this.classes.remove(this.options[t]),this.options[t]=e,e&&this.classes.add(e),this},p.prototype.updateSize=function(){var t;return this.hidden&&(this.element.style.visibility="hidden",i.body.appendChild(this.element)),this.width=this.element.offsetWidth,this.height=this.element.offsetHeight,null==this.spacing&&(this.spacing=null!=this.options.spacing?this.options.spacing:(t=c(this.element,"top"),0|Math.round(String(t).replace(/[^\-0-9.]/g,"")))),this.hidden?(i.body.removeChild(this.element),this.element.style.visibility=""):this.position(),this},p.prototype.content=function(t){return"object"==typeof t?(this.element.innerHTML="",this.element.appendChild(t)):this.element.innerHTML=t,this.updateSize(),this},p.prototype.place=function(t){return this.options.place=t,this.hidden||this.position(),this},p.prototype.attach=function(t){return this.attachedTo=t,this.hidden||this.position(),this},p.prototype.detach=function(){return this.hide(),this.attachedTo=null,this},p.prototype._pickPlace=function(t){if(!this.options.auto)return this.options.place;var e=u(r),i=this.options.place.split("-"),o=this.spacing;if(-1!==n(a,i[0]))switch(t.top-this.height-o<=e.top?i[0]="bottom":t.bottom+this.height+o>=e.bottom&&(i[0]="top"),i[1]){case"left":t.right-this.width<=e.left&&(i[1]="right");break;case"right":t.left+this.width>=e.right&&(i[1]="left");break;default:t.left+t.width/2+this.width/2>=e.right?i[1]="left":t.right-t.width/2-this.width/2<=e.left&&(i[1]="right")}else switch(t.left-this.width-o<=e.left?i[0]="right":t.right+this.width+o>=e.right&&(i[0]="left"),i[1]){case"top":t.bottom-this.height<=e.top&&(i[1]="bottom");break;case"bottom":t.top+this.height>=e.bottom&&(i[1]="top");break;default:t.top+t.height/2+this.height/2>=e.bottom?i[1]="top":t.bottom-t.height/2-this.height/2<=e.top&&(i[1]="bottom")}return i.join("-")},p.prototype.position=function(t,e){this.attachedTo&&(t=this.attachedTo),null==t&&this._p?(t=this._p[0],e=this._p[1]):this._p=arguments;var n,r,i="number"==typeof t?{left:0|t,right:0|t,top:0|e,bottom:0|e,width:0,height:0}:u(t),o=this.spacing,a=this._pickPlace(i);switch(a!==this.curPlace&&(this.curPlace&&this.classes.remove(this.curPlace),this.classes.add(a),this.curPlace=a),this.curPlace){case"top":n=i.top-this.height-o,r=i.left+i.width/2-this.width/2;break;case"top-left":n=i.top-this.height-o,r=i.right-this.width;break;case"top-right":n=i.top-this.height-o,r=i.left;break;case"bottom":n=i.bottom+o,r=i.left+i.width/2-this.width/2;break;case"bottom-left":n=i.bottom+o,r=i.right-this.width;break;case"bottom-right":n=i.bottom+o,r=i.left;break;case"left":n=i.top+i.height/2-this.height/2,r=i.left-this.width-o;break;case"left-top":n=i.bottom-this.height,r=i.left-this.width-o;break;case"left-bottom":n=i.top,r=i.left-this.width-o;break;case"right":n=i.top+i.height/2-this.height/2,r=i.right+o;break;case"right-top":n=i.bottom-this.height,r=i.right+o;break;case"right-bottom":n=i.top,r=i.right+o}return this.element.style.top=Math.round(n)+"px",this.element.style.left=Math.round(r)+"px",this},p.prototype.show=function(t,e){return t=this.attachedTo?this.attachedTo:t,clearTimeout(this.aIndex),null!=t&&this.position(t,e),this.hidden&&(this.hidden=0,i.body.appendChild(this.element)),this.attachedTo&&this._aware(),this.options.inClass&&(this.options.effectClass&&this.element.clientHeight,this.classes.add(this.options.inClass)),this},p.prototype.getElement=function(){return this.element},p.prototype.hide=function(){if(!this.hidden){var t=this,e=0;return this.options.inClass&&(this.classes.remove(this.options.inClass),this.options.effectClass&&(e=l(this.element))),this.attachedTo&&this._unaware(),clearTimeout(this.aIndex),this.aIndex=setTimeout(function(){t.aIndex=0,i.body.removeChild(t.element),t.hidden=1},e),this}},p.prototype.toggle=function(t,e){return this[this.hidden?"show":"hide"](t,e)},p.prototype.destroy=function(){clearTimeout(this.aIndex),this._unaware(),this.hidden||i.body.removeChild(this.element),this.element=this.options=null},p.prototype._aware=function(){-1===n(p.winAware,this)&&p.winAware.push(this)},p.prototype._unaware=function(){var t=n(p.winAware,this);-1!==t&&p.winAware.splice(t,1)},p.reposition=function(){var t,e=window.requestAnimationFrame||window.webkitRequestAnimationFrame||function(t){return setTimeout(t,17)};function n(){t=0;for(var e=0,n=p.winAware.length;e<n;e++)p.winAware[e].position()}return function(){!t&&p.winAware.length&&(t=e(n))}}(),p.winAware=[],window.addEventListener("resize",p.reposition),window.addEventListener("scroll",p.reposition),p.classTypes=["type","effect"],p.defaults={baseClass:"tooltip",typeClass:null,effectClass:null,inClass:"in",place:"top",spacing:null,auto:0},e.Tooltip=p}])})?r.apply(e,[]):r)||(t.exports=i)},function(t,e){t.exports=p},function(t,e){t.exports=f},function(t,e){t.exports=h},function(t,e){t.exports=m},function(t,e){t.exports=g},function(t,e,n){"use strict";var r,i,o,a,s;if(window.requestAnimationFrame||(r=0,window.requestAnimationFrame=function(t){var e=Date.now(),n=Math.max(0,16-e+r);return r=e+n,setTimeout(function(){t.call(null,e+n)},n)},window.cancelAnimationFrame=function(t){clearTimeout(t)}),window.requestIdleCallback||(window.requestIdleCallback=function(t,e){return setTimeout(t,e&&e.timeout?e.timeout:0)},window.cancelIdleCallback=function(t){clearTimeout(t)}),Number.isFinite||Object.defineProperty(Number,"isFinite",{value:function(t){return"number"==typeof t&&isFinite(t)},configurable:!0,enumerable:!1,writable:!0}),Number.isNaN||Object.defineProperty(Number,"isNaN",{value:function(t){return"number"==typeof t&&isNaN(t)},configurable:!0,enumerable:!1,writable:!0}),Number.isInteger||Object.defineProperty(Number,"isInteger",{value:function(t){return"number"==typeof t&&isFinite(t)&&Math.floor(t)===t},configurable:!0,enumerable:!1,writable:!0}),Number.MAX_SAFE_INTEGER||Object.defineProperty(Number,"MAX_SAFE_INTEGER",{value:9007199254740991,configurable:!1,enumerable:!1,writable:!1}),Math.sign||(Math.sign=function(t){return(t>0)-(t<0)||Number(t)}),Array.prototype.find||Object.defineProperty(Array.prototype,"find",{value:function(t){if(null==this)throw new TypeError("Array.prototype.find called on null or undefined");if("function"!=typeof t)throw new TypeError("predicate must be a function");for(var e,n=Object(this),r=n.length>>>0,i=arguments[1],o=0;o<r;o++)if(e=n[o],t.call(i,e,o,n))return e},configurable:!0,enumerable:!1,writable:!0}),Array.prototype.findIndex||Object.defineProperty(Array.prototype,"findIndex",{value:function(t){if(null==this)throw new TypeError("Array.prototype.findIndex called on null or undefined");if("function"!=typeof t)throw new TypeError("predicate must be a function");for(var e,n=Object(this),r=n.length>>>0,i=arguments[1],o=0;o<r;o++)if(e=n[o],t.call(i,e,o,n))return o;return-1},configurable:!0,enumerable:!1,writable:!0}),!Array.from){Object.defineProperty(Array,"from",{value:(i=Object.prototype.toString,o=function(t){return"function"==typeof t||"[object Function]"===i.call(t)},a=Math.pow(2,53)-1,s=function(t){var e,n=(e=Number(t),isNaN(e)?0:0!==e&&isFinite(e)?(e>0?1:-1)*Math.floor(Math.abs(e)):e);return Math.min(Math.max(n,0),a)},function(t){var e=Object(t);if(null==t)throw new TypeError("Array.from requires an array-like object - not null or undefined");var n,r=arguments.length>1?arguments[1]:void 0;if(void 0!==r){if(!o(r))throw new TypeError("Array.from: when provided, the second argument must be a function");arguments.length>2&&(n=arguments[2])}for(var i,a=s(e.length),u=o(this)?Object(new this(a)):new Array(a),c=0;c<a;)i=e[c],u[c]=r?void 0===n?r(i,c):r.call(n,i,c):i,c+=1;return u.length=a,u}),configurable:!0,enumerable:!1,writable:!0});var u=Map.prototype.set;Object.defineProperty(Map.prototype,"set",{value:function(){return u.apply(this,arguments),this},configurable:!0,enumerable:!1,writable:!0}),Object.defineProperty(Map.prototype,"values",{value:function(){var t={},e=0;return this.forEach(function(n){return t[String(e++)]=n}),t.length=e,t},configurable:!0,enumerable:!1,writable:!0}),Object.defineProperty(Map.prototype,"entries",{value:function(){var t={},e=0;return this.forEach(function(n,r){return t[String(e++)]=[r,n]}),t.length=e,t},configurable:!0,enumerable:!1,writable:!0})}Object.assign||Object.defineProperty(Object,"assign",{value:function(t){if(null==t)throw new TypeError("Cannot convert undefined or null to object");for(var e=Object(t),n=1;n<arguments.length;n++){var r=arguments[n];if(null!=r)for(var i in r)r.hasOwnProperty(i)&&(e[i]=r[i])}return e},configurable:!0,enumerable:!1,writable:!0}),Element.prototype.matches||Object.defineProperty(Element.prototype,"matches",{value:Element.prototype.msMatchesSelector||Element.prototype.webkitMatchesSelector,configurable:!0,enumerable:!0,writable:!0})},,,,function(t,e,n){t.exports=n(1)}]))},"./node_modules/taucharts/node_modules/d3-color/src/color.js":function(t,e,n){"use strict";n.r(e),n.d(e,"Color",function(){return i}),n.d(e,"darker",function(){return o}),n.d(e,"brighter",function(){return a}),n.d(e,"default",function(){return v}),n.d(e,"rgbConvert",function(){return w}),n.d(e,"rgb",function(){return j}),n.d(e,"Rgb",function(){return O}),n.d(e,"hslConvert",function(){return S}),n.d(e,"hsl",function(){return M});var r=n("./node_modules/taucharts/node_modules/d3-color/src/define.js");function i(){}var o=.7,a=1/o,s="\\s*([+-]?\\d+)\\s*",u="\\s*([+-]?\\d*\\.?\\d+(?:[eE][+-]?\\d+)?)\\s*",c="\\s*([+-]?\\d*\\.?\\d+(?:[eE][+-]?\\d+)?)%\\s*",l=/^#([0-9a-f]{3})$/,d=/^#([0-9a-f]{6})$/,p=new RegExp("^rgb\\("+[s,s,s]+"\\)$"),f=new RegExp("^rgb\\("+[c,c,c]+"\\)$"),h=new RegExp("^rgba\\("+[s,s,s,u]+"\\)$"),m=new RegExp("^rgba\\("+[c,c,c,u]+"\\)$"),g=new RegExp("^hsl\\("+[u,c,c]+"\\)$"),y=new RegExp("^hsla\\("+[u,c,c,u]+"\\)$"),_={aliceblue:15792383,antiquewhite:16444375,aqua:65535,aquamarine:8388564,azure:15794175,beige:16119260,bisque:16770244,black:0,blanchedalmond:16772045,blue:255,blueviolet:9055202,brown:10824234,burlywood:14596231,cadetblue:6266528,chartreuse:8388352,chocolate:13789470,coral:16744272,cornflowerblue:6591981,cornsilk:16775388,crimson:14423100,cyan:65535,darkblue:139,darkcyan:35723,darkgoldenrod:12092939,darkgray:11119017,darkgreen:25600,darkgrey:11119017,darkkhaki:12433259,darkmagenta:9109643,darkolivegreen:5597999,darkorange:16747520,darkorchid:10040012,darkred:9109504,darksalmon:15308410,darkseagreen:9419919,darkslateblue:4734347,darkslategray:3100495,darkslategrey:3100495,darkturquoise:52945,darkviolet:9699539,deeppink:16716947,deepskyblue:49151,dimgray:6908265,dimgrey:6908265,dodgerblue:2003199,firebrick:11674146,floralwhite:16775920,forestgreen:2263842,fuchsia:16711935,gainsboro:14474460,ghostwhite:16316671,gold:16766720,goldenrod:14329120,gray:8421504,green:32768,greenyellow:11403055,grey:8421504,honeydew:15794160,hotpink:16738740,indianred:13458524,indigo:4915330,ivory:16777200,khaki:15787660,lavender:15132410,lavenderblush:16773365,lawngreen:8190976,lemonchiffon:16775885,lightblue:11393254,lightcoral:15761536,lightcyan:14745599,lightgoldenrodyellow:16448210,lightgray:13882323,lightgreen:9498256,lightgrey:13882323,lightpink:16758465,lightsalmon:16752762,lightseagreen:2142890,lightskyblue:8900346,lightslategray:7833753,lightslategrey:7833753,lightsteelblue:11584734,lightyellow:16777184,lime:65280,limegreen:3329330,linen:16445670,magenta:16711935,maroon:8388608,mediumaquamarine:6737322,mediumblue:205,mediumorchid:12211667,mediumpurple:9662683,mediumseagreen:3978097,mediumslateblue:8087790,mediumspringgreen:64154,mediumturquoise:4772300,mediumvioletred:13047173,midnightblue:1644912,mintcream:16121850,mistyrose:16770273,moccasin:16770229,navajowhite:16768685,navy:128,oldlace:16643558,olive:8421376,olivedrab:7048739,orange:16753920,orangered:16729344,orchid:14315734,palegoldenrod:15657130,palegreen:10025880,paleturquoise:11529966,palevioletred:14381203,papayawhip:16773077,peachpuff:16767673,peru:13468991,pink:16761035,plum:14524637,powderblue:11591910,purple:8388736,rebeccapurple:6697881,red:16711680,rosybrown:12357519,royalblue:4286945,saddlebrown:9127187,salmon:16416882,sandybrown:16032864,seagreen:3050327,seashell:16774638,sienna:10506797,silver:12632256,skyblue:8900331,slateblue:6970061,slategray:7372944,slategrey:7372944,snow:16775930,springgreen:65407,steelblue:4620980,tan:13808780,teal:32896,thistle:14204888,tomato:16737095,turquoise:4251856,violet:15631086,wheat:16113331,white:16777215,whitesmoke:16119285,yellow:16776960,yellowgreen:10145074};function v(t){var e;return t=(t+"").trim().toLowerCase(),(e=l.exec(t))?new O((e=parseInt(e[1],16))>>8&15|e>>4&240,e>>4&15|240&e,(15&e)<<4|15&e,1):(e=d.exec(t))?b(parseInt(e[1],16)):(e=p.exec(t))?new O(e[1],e[2],e[3],1):(e=f.exec(t))?new O(255*e[1]/100,255*e[2]/100,255*e[3]/100,1):(e=h.exec(t))?x(e[1],e[2],e[3],e[4]):(e=m.exec(t))?x(255*e[1]/100,255*e[2]/100,255*e[3]/100,e[4]):(e=g.exec(t))?E(e[1],e[2]/100,e[3]/100,1):(e=y.exec(t))?E(e[1],e[2]/100,e[3]/100,e[4]):_.hasOwnProperty(t)?b(_[t]):"transparent"===t?new O(NaN,NaN,NaN,0):null}function b(t){return new O(t>>16&255,t>>8&255,255&t,1)}function x(t,e,n,r){return r<=0&&(t=e=n=NaN),new O(t,e,n,r)}function w(t){return t instanceof i||(t=v(t)),t?new O((t=t.rgb()).r,t.g,t.b,t.opacity):new O}function j(t,e,n,r){return 1===arguments.length?w(t):new O(t,e,n,null==r?1:r)}function O(t,e,n,r){this.r=+t,this.g=+e,this.b=+n,this.opacity=+r}function C(t){return((t=Math.max(0,Math.min(255,Math.round(t)||0)))<16?"0":"")+t.toString(16)}function E(t,e,n,r){return r<=0?t=e=n=NaN:n<=0||n>=1?t=e=NaN:e<=0&&(t=NaN),new T(t,e,n,r)}function S(t){if(t instanceof T)return new T(t.h,t.s,t.l,t.opacity);if(t instanceof i||(t=v(t)),!t)return new T;if(t instanceof T)return t;var e=(t=t.rgb()).r/255,n=t.g/255,r=t.b/255,o=Math.min(e,n,r),a=Math.max(e,n,r),s=NaN,u=a-o,c=(a+o)/2;return u?(s=e===a?(n-r)/u+6*(n<r):n===a?(r-e)/u+2:(e-n)/u+4,u/=c<.5?a+o:2-a-o,s*=60):u=c>0&&c<1?0:s,new T(s,u,c,t.opacity)}function M(t,e,n,r){return 1===arguments.length?S(t):new T(t,e,n,null==r?1:r)}function T(t,e,n,r){this.h=+t,this.s=+e,this.l=+n,this.opacity=+r}function k(t,e,n){return 255*(t<60?e+(n-e)*t/60:t<180?n:t<240?e+(n-e)*(240-t)/60:e)}Object(r.default)(i,v,{displayable:function(){return this.rgb().displayable()},hex:function(){return this.rgb().hex()},toString:function(){return this.rgb()+""}}),Object(r.default)(O,j,Object(r.extend)(i,{brighter:function(t){return t=null==t?a:Math.pow(a,t),new O(this.r*t,this.g*t,this.b*t,this.opacity)},darker:function(t){return t=null==t?o:Math.pow(o,t),new O(this.r*t,this.g*t,this.b*t,this.opacity)},rgb:function(){return this},displayable:function(){return 0<=this.r&&this.r<=255&&0<=this.g&&this.g<=255&&0<=this.b&&this.b<=255&&0<=this.opacity&&this.opacity<=1},hex:function(){return"#"+C(this.r)+C(this.g)+C(this.b)},toString:function(){var t=this.opacity;return(1===(t=isNaN(t)?1:Math.max(0,Math.min(1,t)))?"rgb(":"rgba(")+Math.max(0,Math.min(255,Math.round(this.r)||0))+", "+Math.max(0,Math.min(255,Math.round(this.g)||0))+", "+Math.max(0,Math.min(255,Math.round(this.b)||0))+(1===t?")":", "+t+")")}})),Object(r.default)(T,M,Object(r.extend)(i,{brighter:function(t){return t=null==t?a:Math.pow(a,t),new T(this.h,this.s,this.l*t,this.opacity)},darker:function(t){return t=null==t?o:Math.pow(o,t),new T(this.h,this.s,this.l*t,this.opacity)},rgb:function(){var t=this.h%360+360*(this.h<0),e=isNaN(t)||isNaN(this.s)?0:this.s,n=this.l,r=n+(n<.5?n:1-n)*e,i=2*n-r;return new O(k(t>=240?t-240:t+120,i,r),k(t,i,r),k(t<120?t+240:t-120,i,r),this.opacity)},displayable:function(){return(0<=this.s&&this.s<=1||isNaN(this.s))&&0<=this.l&&this.l<=1&&0<=this.opacity&&this.opacity<=1}}))},"./node_modules/taucharts/node_modules/d3-color/src/cubehelix.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return p}),n.d(e,"Cubehelix",function(){return f});var r=n("./node_modules/taucharts/node_modules/d3-color/src/define.js"),i=n("./node_modules/taucharts/node_modules/d3-color/src/color.js"),o=n("./node_modules/taucharts/node_modules/d3-color/src/math.js"),a=-.29227,s=-.90649,u=1.97294,c=u*s,l=1.78277*u,d=1.78277*a- -.14861*s;function p(t,e,n,r){return 1===arguments.length?function(t){if(t instanceof f)return new f(t.h,t.s,t.l,t.opacity);t instanceof i.Rgb||(t=Object(i.rgbConvert)(t));var e=t.r/255,n=t.g/255,r=t.b/255,p=(d*r+c*e-l*n)/(d+c-l),h=r-p,m=(u*(n-p)-a*h)/s,g=Math.sqrt(m*m+h*h)/(u*p*(1-p)),y=g?Math.atan2(m,h)*o.rad2deg-120:NaN;return new f(y<0?y+360:y,g,p,t.opacity)}(t):new f(t,e,n,null==r?1:r)}function f(t,e,n,r){this.h=+t,this.s=+e,this.l=+n,this.opacity=+r}Object(r.default)(f,p,Object(r.extend)(i.Color,{brighter:function(t){return t=null==t?i.brighter:Math.pow(i.brighter,t),new f(this.h,this.s,this.l*t,this.opacity)},darker:function(t){return t=null==t?i.darker:Math.pow(i.darker,t),new f(this.h,this.s,this.l*t,this.opacity)},rgb:function(){var t=isNaN(this.h)?0:(this.h+120)*o.deg2rad,e=+this.l,n=isNaN(this.s)?0:this.s*e*(1-e),r=Math.cos(t),c=Math.sin(t);return new i.Rgb(255*(e+n*(-.14861*r+1.78277*c)),255*(e+n*(a*r+s*c)),255*(e+n*(u*r)),this.opacity)}}))},"./node_modules/taucharts/node_modules/d3-color/src/define.js":function(t,e,n){"use strict";function r(t,e){var n=Object.create(t.prototype);for(var r in e)n[r]=e[r];return n}n.r(e),n.d(e,"extend",function(){return r}),e.default=function(t,e,n){t.prototype=e.prototype=n,n.constructor=t}},"./node_modules/taucharts/node_modules/d3-color/src/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-color/src/color.js");n.d(e,"color",function(){return r.default}),n.d(e,"rgb",function(){return r.rgb}),n.d(e,"hsl",function(){return r.hsl});var i=n("./node_modules/taucharts/node_modules/d3-color/src/lab.js");n.d(e,"lab",function(){return i.default}),n.d(e,"hcl",function(){return i.hcl}),n.d(e,"lch",function(){return i.lch}),n.d(e,"gray",function(){return i.gray});var o=n("./node_modules/taucharts/node_modules/d3-color/src/cubehelix.js");n.d(e,"cubehelix",function(){return o.default})},"./node_modules/taucharts/node_modules/d3-color/src/lab.js":function(t,e,n){"use strict";n.r(e),n.d(e,"gray",function(){return h}),n.d(e,"default",function(){return m}),n.d(e,"Lab",function(){return g}),n.d(e,"lch",function(){return w}),n.d(e,"hcl",function(){return j}),n.d(e,"Hcl",function(){return O});var r=n("./node_modules/taucharts/node_modules/d3-color/src/define.js"),i=n("./node_modules/taucharts/node_modules/d3-color/src/color.js"),o=n("./node_modules/taucharts/node_modules/d3-color/src/math.js"),a=.96422,s=1,u=.82521,c=4/29,l=6/29,d=3*l*l,p=l*l*l;function f(t){if(t instanceof g)return new g(t.l,t.a,t.b,t.opacity);if(t instanceof O){if(isNaN(t.h))return new g(t.l,0,0,t.opacity);var e=t.h*o.deg2rad;return new g(t.l,Math.cos(e)*t.c,Math.sin(e)*t.c,t.opacity)}t instanceof i.Rgb||(t=Object(i.rgbConvert)(t));var n,r,c=b(t.r),l=b(t.g),d=b(t.b),p=y((.2225045*c+.7168786*l+.0606169*d)/s);return c===l&&l===d?n=r=p:(n=y((.4360747*c+.3850649*l+.1430804*d)/a),r=y((.0139322*c+.0971045*l+.7141733*d)/u)),new g(116*p-16,500*(n-p),200*(p-r),t.opacity)}function h(t,e){return new g(t,0,0,null==e?1:e)}function m(t,e,n,r){return 1===arguments.length?f(t):new g(t,e,n,null==r?1:r)}function g(t,e,n,r){this.l=+t,this.a=+e,this.b=+n,this.opacity=+r}function y(t){return t>p?Math.pow(t,1/3):t/d+c}function _(t){return t>l?t*t*t:d*(t-c)}function v(t){return 255*(t<=.0031308?12.92*t:1.055*Math.pow(t,1/2.4)-.055)}function b(t){return(t/=255)<=.04045?t/12.92:Math.pow((t+.055)/1.055,2.4)}function x(t){if(t instanceof O)return new O(t.h,t.c,t.l,t.opacity);if(t instanceof g||(t=f(t)),0===t.a&&0===t.b)return new O(NaN,0,t.l,t.opacity);var e=Math.atan2(t.b,t.a)*o.rad2deg;return new O(e<0?e+360:e,Math.sqrt(t.a*t.a+t.b*t.b),t.l,t.opacity)}function w(t,e,n,r){return 1===arguments.length?x(t):new O(n,e,t,null==r?1:r)}function j(t,e,n,r){return 1===arguments.length?x(t):new O(t,e,n,null==r?1:r)}function O(t,e,n,r){this.h=+t,this.c=+e,this.l=+n,this.opacity=+r}Object(r.default)(g,m,Object(r.extend)(i.Color,{brighter:function(t){return new g(this.l+18*(null==t?1:t),this.a,this.b,this.opacity)},darker:function(t){return new g(this.l-18*(null==t?1:t),this.a,this.b,this.opacity)},rgb:function(){var t=(this.l+16)/116,e=isNaN(this.a)?t:t+this.a/500,n=isNaN(this.b)?t:t-this.b/200;return e=a*_(e),t=s*_(t),n=u*_(n),new i.Rgb(v(3.1338561*e-1.6168667*t-.4906146*n),v(-.9787684*e+1.9161415*t+.033454*n),v(.0719453*e-.2289914*t+1.4052427*n),this.opacity)}})),Object(r.default)(O,j,Object(r.extend)(i.Color,{brighter:function(t){return new O(this.h,this.c,this.l+18*(null==t?1:t),this.opacity)},darker:function(t){return new O(this.h,this.c,this.l-18*(null==t?1:t),this.opacity)},rgb:function(){return f(this).rgb()}}))},"./node_modules/taucharts/node_modules/d3-color/src/math.js":function(t,e,n){"use strict";n.r(e),n.d(e,"deg2rad",function(){return r}),n.d(e,"rad2deg",function(){return i});var r=Math.PI/180,i=180/Math.PI},"./node_modules/taucharts/node_modules/d3-format/src/defaultLocale.js":function(t,e,n){"use strict";n.r(e),n.d(e,"format",function(){return i}),n.d(e,"formatPrefix",function(){return o}),n.d(e,"default",function(){return s});var r,i,o,a=n("./node_modules/taucharts/node_modules/d3-format/src/locale.js");function s(t){return r=Object(a.default)(t),i=r.format,o=r.formatPrefix,r}s({decimal:".",thousands:",",grouping:[3],currency:["$",""]})},"./node_modules/taucharts/node_modules/d3-format/src/exponent.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-format/src/formatDecimal.js");e.default=function(t){return(t=Object(r.default)(Math.abs(t)))?t[1]:NaN}},"./node_modules/taucharts/node_modules/d3-format/src/formatDecimal.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){if((n=(t=e?t.toExponential(e-1):t.toExponential()).indexOf("e"))<0)return null;var n,r=t.slice(0,n);return[r.length>1?r[0]+r.slice(2):r,+t.slice(n+1)]}},"./node_modules/taucharts/node_modules/d3-format/src/formatGroup.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){return function(n,r){for(var i=n.length,o=[],a=0,s=t[0],u=0;i>0&&s>0&&(u+s+1>r&&(s=Math.max(1,r-u)),o.push(n.substring(i-=s,i+s)),!((u+=s+1)>r));)s=t[a=(a+1)%t.length];return o.reverse().join(e)}}},"./node_modules/taucharts/node_modules/d3-format/src/formatNumerals.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return function(e){return e.replace(/[0-9]/g,function(e){return t[+e]})}}},"./node_modules/taucharts/node_modules/d3-format/src/formatPrefixAuto.js":function(t,e,n){"use strict";n.r(e),n.d(e,"prefixExponent",function(){return r});var r,i=n("./node_modules/taucharts/node_modules/d3-format/src/formatDecimal.js");e.default=function(t,e){var n=Object(i.default)(t,e);if(!n)return t+"";var o=n[0],a=n[1],s=a-(r=3*Math.max(-8,Math.min(8,Math.floor(a/3))))+1,u=o.length;return s===u?o:s>u?o+new Array(s-u+1).join("0"):s>0?o.slice(0,s)+"."+o.slice(s):"0."+new Array(1-s).join("0")+Object(i.default)(t,Math.max(0,e+s-1))[0]}},"./node_modules/taucharts/node_modules/d3-format/src/formatRounded.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-format/src/formatDecimal.js");e.default=function(t,e){var n=Object(r.default)(t,e);if(!n)return t+"";var i=n[0],o=n[1];return o<0?"0."+new Array(-o).join("0")+i:i.length>o+1?i.slice(0,o+1)+"."+i.slice(o+1):i+new Array(o-i.length+2).join("0")}},"./node_modules/taucharts/node_modules/d3-format/src/formatSpecifier.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return i});var r=/^(?:(.)?([<>=^]))?([+\-( ])?([$#])?(0)?(\d+)?(,)?(\.\d+)?(~)?([a-z%])?$/i;function i(t){return new o(t)}function o(t){if(!(e=r.exec(t)))throw new Error("invalid format: "+t);var e;this.fill=e[1]||" ",this.align=e[2]||">",this.sign=e[3]||"-",this.symbol=e[4]||"",this.zero=!!e[5],this.width=e[6]&&+e[6],this.comma=!!e[7],this.precision=e[8]&&+e[8].slice(1),this.trim=!!e[9],this.type=e[10]||""}i.prototype=o.prototype,o.prototype.toString=function(){return this.fill+this.align+this.sign+this.symbol+(this.zero?"0":"")+(null==this.width?"":Math.max(1,0|this.width))+(this.comma?",":"")+(null==this.precision?"":"."+Math.max(0,0|this.precision))+(this.trim?"~":"")+this.type}},"./node_modules/taucharts/node_modules/d3-format/src/formatTrim.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){t:for(var e,n=t.length,r=1,i=-1;r<n;++r)switch(t[r]){case".":i=e=r;break;case"0":0===i&&(i=r),e=r;break;default:if(i>0){if(!+t[r])break t;i=0}}return i>0?t.slice(0,i)+t.slice(e+1):t}},"./node_modules/taucharts/node_modules/d3-format/src/formatTypes.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-format/src/formatPrefixAuto.js"),i=n("./node_modules/taucharts/node_modules/d3-format/src/formatRounded.js");e.default={"%":function(t,e){return(100*t).toFixed(e)},b:function(t){return Math.round(t).toString(2)},c:function(t){return t+""},d:function(t){return Math.round(t).toString(10)},e:function(t,e){return t.toExponential(e)},f:function(t,e){return t.toFixed(e)},g:function(t,e){return t.toPrecision(e)},o:function(t){return Math.round(t).toString(8)},p:function(t,e){return Object(i.default)(100*t,e)},r:i.default,s:r.default,X:function(t){return Math.round(t).toString(16).toUpperCase()},x:function(t){return Math.round(t).toString(16)}}},"./node_modules/taucharts/node_modules/d3-format/src/identity.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return t}},"./node_modules/taucharts/node_modules/d3-format/src/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-format/src/defaultLocale.js");n.d(e,"formatDefaultLocale",function(){return r.default}),n.d(e,"format",function(){return r.format}),n.d(e,"formatPrefix",function(){return r.formatPrefix});var i=n("./node_modules/taucharts/node_modules/d3-format/src/locale.js");n.d(e,"formatLocale",function(){return i.default});var o=n("./node_modules/taucharts/node_modules/d3-format/src/formatSpecifier.js");n.d(e,"formatSpecifier",function(){return o.default});var a=n("./node_modules/taucharts/node_modules/d3-format/src/precisionFixed.js");n.d(e,"precisionFixed",function(){return a.default});var s=n("./node_modules/taucharts/node_modules/d3-format/src/precisionPrefix.js");n.d(e,"precisionPrefix",function(){return s.default});var u=n("./node_modules/taucharts/node_modules/d3-format/src/precisionRound.js");n.d(e,"precisionRound",function(){return u.default})},"./node_modules/taucharts/node_modules/d3-format/src/locale.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-format/src/exponent.js"),i=n("./node_modules/taucharts/node_modules/d3-format/src/formatGroup.js"),o=n("./node_modules/taucharts/node_modules/d3-format/src/formatNumerals.js"),a=n("./node_modules/taucharts/node_modules/d3-format/src/formatSpecifier.js"),s=n("./node_modules/taucharts/node_modules/d3-format/src/formatTrim.js"),u=n("./node_modules/taucharts/node_modules/d3-format/src/formatTypes.js"),c=n("./node_modules/taucharts/node_modules/d3-format/src/formatPrefixAuto.js"),l=n("./node_modules/taucharts/node_modules/d3-format/src/identity.js"),d=["y","z","a","f","p","n","µ","m","","k","M","G","T","P","E","Z","Y"];e.default=function(t){var e=t.grouping&&t.thousands?Object(i.default)(t.grouping,t.thousands):l.default,n=t.currency,p=t.decimal,f=t.numerals?Object(o.default)(t.numerals):l.default,h=t.percent||"%";function m(t){var r=(t=Object(a.default)(t)).fill,i=t.align,o=t.sign,l=t.symbol,m=t.zero,g=t.width,y=t.comma,_=t.precision,v=t.trim,b=t.type;"n"===b?(y=!0,b="g"):u.default[b]||(null==_&&(_=12),v=!0,b="g"),(m||"0"===r&&"="===i)&&(m=!0,r="0",i="=");var x="$"===l?n[0]:"#"===l&&/[boxX]/.test(b)?"0"+b.toLowerCase():"",w="$"===l?n[1]:/[%p]/.test(b)?h:"",j=u.default[b],O=/[defgprs%]/.test(b);function C(t){var n,a,u,l=x,h=w;if("c"===b)h=j(t)+h,t="";else{var C=(t=+t)<0;if(t=j(Math.abs(t),_),v&&(t=Object(s.default)(t)),C&&0==+t&&(C=!1),l=(C?"("===o?o:"-":"-"===o||"("===o?"":o)+l,h=("s"===b?d[8+c.prefixExponent/3]:"")+h+(C&&"("===o?")":""),O)for(n=-1,a=t.length;++n<a;)if(48>(u=t.charCodeAt(n))||u>57){h=(46===u?p+t.slice(n+1):t.slice(n))+h,t=t.slice(0,n);break}}y&&!m&&(t=e(t,1/0));var E=l.length+t.length+h.length,S=E<g?new Array(g-E+1).join(r):"";switch(y&&m&&(t=e(S+t,S.length?g-h.length:1/0),S=""),i){case"<":t=l+t+h+S;break;case"=":t=l+S+t+h;break;case"^":t=S.slice(0,E=S.length>>1)+l+t+h+S.slice(E);break;default:t=S+l+t+h}return f(t)}return _=null==_?6:/[gprs]/.test(b)?Math.max(1,Math.min(21,_)):Math.max(0,Math.min(20,_)),C.toString=function(){return t+""},C}return{format:m,formatPrefix:function(t,e){var n=m(((t=Object(a.default)(t)).type="f",t)),i=3*Math.max(-8,Math.min(8,Math.floor(Object(r.default)(e)/3))),o=Math.pow(10,-i),s=d[8+i/3];return function(t){return n(o*t)+s}}}}},"./node_modules/taucharts/node_modules/d3-format/src/precisionFixed.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-format/src/exponent.js");e.default=function(t){return Math.max(0,-Object(r.default)(Math.abs(t)))}},"./node_modules/taucharts/node_modules/d3-format/src/precisionPrefix.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-format/src/exponent.js");e.default=function(t,e){return Math.max(0,3*Math.max(-8,Math.min(8,Math.floor(Object(r.default)(e)/3)))-Object(r.default)(Math.abs(t)))}},"./node_modules/taucharts/node_modules/d3-format/src/precisionRound.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-format/src/exponent.js");e.default=function(t,e){return t=Math.abs(t),e=Math.abs(e)-t,Math.max(0,Object(r.default)(e)-Object(r.default)(t))+1}},"./node_modules/taucharts/node_modules/d3-geo/src/adder.js":function(t,e,n){"use strict";function r(){this.reset()}n.r(e),e.default=function(){return new r},r.prototype={constructor:r,reset:function(){this.s=this.t=0},add:function(t){o(i,t,this.t),o(this,i.s,this.s),this.s?this.t+=i.t:this.s=i.t},valueOf:function(){return this.s}};var i=new r;function o(t,e,n){var r=t.s=e+n,i=r-e,o=r-i;t.t=e-o+(n-i)}},"./node_modules/taucharts/node_modules/d3-geo/src/area.js":function(t,e,n){"use strict";n.r(e),n.d(e,"areaRingSum",function(){return p}),n.d(e,"areaStream",function(){return h});var r,i,o,a,s,u=n("./node_modules/taucharts/node_modules/d3-geo/src/adder.js"),c=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),l=n("./node_modules/taucharts/node_modules/d3-geo/src/noop.js"),d=n("./node_modules/taucharts/node_modules/d3-geo/src/stream.js"),p=Object(u.default)(),f=Object(u.default)(),h={point:l.default,lineStart:l.default,lineEnd:l.default,polygonStart:function(){p.reset(),h.lineStart=m,h.lineEnd=g},polygonEnd:function(){var t=+p;f.add(t<0?c.tau+t:t),this.lineStart=this.lineEnd=this.point=l.default},sphere:function(){f.add(c.tau)}};function m(){h.point=y}function g(){_(r,i)}function y(t,e){h.point=_,r=t,i=e,t*=c.radians,e*=c.radians,o=t,a=Object(c.cos)(e=e/2+c.quarterPi),s=Object(c.sin)(e)}function _(t,e){t*=c.radians,e=(e*=c.radians)/2+c.quarterPi;var n=t-o,r=n>=0?1:-1,i=r*n,u=Object(c.cos)(e),l=Object(c.sin)(e),d=s*l,f=a*u+d*Object(c.cos)(i),h=d*r*Object(c.sin)(i);p.add(Object(c.atan2)(h,f)),o=t,a=u,s=l}e.default=function(t){return f.reset(),Object(d.default)(t,h),2*f}},"./node_modules/taucharts/node_modules/d3-geo/src/bounds.js":function(t,e,n){"use strict";n.r(e);var r,i,o,a,s,u,c,l,d,p,f=n("./node_modules/taucharts/node_modules/d3-geo/src/adder.js"),h=n("./node_modules/taucharts/node_modules/d3-geo/src/area.js"),m=n("./node_modules/taucharts/node_modules/d3-geo/src/cartesian.js"),g=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),y=n("./node_modules/taucharts/node_modules/d3-geo/src/stream.js"),_=Object(f.default)(),v={point:b,lineStart:w,lineEnd:j,polygonStart:function(){v.point=O,v.lineStart=C,v.lineEnd=E,_.reset(),h.areaStream.polygonStart()},polygonEnd:function(){h.areaStream.polygonEnd(),v.point=b,v.lineStart=w,v.lineEnd=j,h.areaRingSum<0?(r=-(o=180),i=-(a=90)):_>g.epsilon?a=90:_<-g.epsilon&&(i=-90),p[0]=r,p[1]=o}};function b(t,e){d.push(p=[r=t,o=t]),e<i&&(i=e),e>a&&(a=e)}function x(t,e){var n=Object(m.cartesian)([t*g.radians,e*g.radians]);if(l){var u=Object(m.cartesianCross)(l,n),c=[u[1],-u[0],0],f=Object(m.cartesianCross)(c,u);Object(m.cartesianNormalizeInPlace)(f),f=Object(m.spherical)(f);var h,y=t-s,_=y>0?1:-1,v=f[0]*g.degrees*_,b=Object(g.abs)(y)>180;b^(_*s<v&&v<_*t)?(h=f[1]*g.degrees)>a&&(a=h):b^(_*s<(v=(v+360)%360-180)&&v<_*t)?(h=-f[1]*g.degrees)<i&&(i=h):(e<i&&(i=e),e>a&&(a=e)),b?t<s?S(r,t)>S(r,o)&&(o=t):S(t,o)>S(r,o)&&(r=t):o>=r?(t<r&&(r=t),t>o&&(o=t)):t>s?S(r,t)>S(r,o)&&(o=t):S(t,o)>S(r,o)&&(r=t)}else d.push(p=[r=t,o=t]);e<i&&(i=e),e>a&&(a=e),l=n,s=t}function w(){v.point=x}function j(){p[0]=r,p[1]=o,v.point=b,l=null}function O(t,e){if(l){var n=t-s;_.add(Object(g.abs)(n)>180?n+(n>0?360:-360):n)}else u=t,c=e;h.areaStream.point(t,e),x(t,e)}function C(){h.areaStream.lineStart()}function E(){O(u,c),h.areaStream.lineEnd(),Object(g.abs)(_)>g.epsilon&&(r=-(o=180)),p[0]=r,p[1]=o,l=null}function S(t,e){return(e-=t)<0?e+360:e}function M(t,e){return t[0]-e[0]}function T(t,e){return t[0]<=t[1]?t[0]<=e&&e<=t[1]:e<t[0]||t[1]<e}e.default=function(t){var e,n,s,u,c,l,f;if(a=o=-(r=i=1/0),d=[],Object(y.default)(t,v),n=d.length){for(d.sort(M),e=1,c=[s=d[0]];e<n;++e)T(s,(u=d[e])[0])||T(s,u[1])?(S(s[0],u[1])>S(s[0],s[1])&&(s[1]=u[1]),S(u[0],s[1])>S(s[0],s[1])&&(s[0]=u[0])):c.push(s=u);for(l=-1/0,e=0,s=c[n=c.length-1];e<=n;s=u,++e)u=c[e],(f=S(s[1],u[0]))>l&&(l=f,r=u[0],o=s[1])}return d=p=null,r===1/0||i===1/0?[[NaN,NaN],[NaN,NaN]]:[[r,i],[o,a]]}},"./node_modules/taucharts/node_modules/d3-geo/src/cartesian.js":function(t,e,n){"use strict";n.r(e),n.d(e,"spherical",function(){return i}),n.d(e,"cartesian",function(){return o}),n.d(e,"cartesianDot",function(){return a}),n.d(e,"cartesianCross",function(){return s}),n.d(e,"cartesianAddInPlace",function(){return u}),n.d(e,"cartesianScale",function(){return c}),n.d(e,"cartesianNormalizeInPlace",function(){return l});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js");function i(t){return[Object(r.atan2)(t[1],t[0]),Object(r.asin)(t[2])]}function o(t){var e=t[0],n=t[1],i=Object(r.cos)(n);return[i*Object(r.cos)(e),i*Object(r.sin)(e),Object(r.sin)(n)]}function a(t,e){return t[0]*e[0]+t[1]*e[1]+t[2]*e[2]}function s(t,e){return[t[1]*e[2]-t[2]*e[1],t[2]*e[0]-t[0]*e[2],t[0]*e[1]-t[1]*e[0]]}function u(t,e){t[0]+=e[0],t[1]+=e[1],t[2]+=e[2]}function c(t,e){return[t[0]*e,t[1]*e,t[2]*e]}function l(t){var e=Object(r.sqrt)(t[0]*t[0]+t[1]*t[1]+t[2]*t[2]);t[0]/=e,t[1]/=e,t[2]/=e}},"./node_modules/taucharts/node_modules/d3-geo/src/centroid.js":function(t,e,n){"use strict";n.r(e);var r,i,o,a,s,u,c,l,d,p,f,h,m,g,y,_,v=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),b=n("./node_modules/taucharts/node_modules/d3-geo/src/noop.js"),x=n("./node_modules/taucharts/node_modules/d3-geo/src/stream.js"),w={sphere:b.default,point:j,lineStart:C,lineEnd:M,polygonStart:function(){w.lineStart=T,w.lineEnd=k},polygonEnd:function(){w.lineStart=C,w.lineEnd=M}};function j(t,e){t*=v.radians,e*=v.radians;var n=Object(v.cos)(e);O(n*Object(v.cos)(t),n*Object(v.sin)(t),Object(v.sin)(e))}function O(t,e,n){o+=(t-o)/++r,a+=(e-a)/r,s+=(n-s)/r}function C(){w.point=E}function E(t,e){t*=v.radians,e*=v.radians;var n=Object(v.cos)(e);g=n*Object(v.cos)(t),y=n*Object(v.sin)(t),_=Object(v.sin)(e),w.point=S,O(g,y,_)}function S(t,e){t*=v.radians,e*=v.radians;var n=Object(v.cos)(e),r=n*Object(v.cos)(t),o=n*Object(v.sin)(t),a=Object(v.sin)(e),s=Object(v.atan2)(Object(v.sqrt)((s=y*a-_*o)*s+(s=_*r-g*a)*s+(s=g*o-y*r)*s),g*r+y*o+_*a);i+=s,u+=s*(g+(g=r)),c+=s*(y+(y=o)),l+=s*(_+(_=a)),O(g,y,_)}function M(){w.point=j}function T(){w.point=A}function k(){D(h,m),w.point=j}function A(t,e){h=t,m=e,t*=v.radians,e*=v.radians,w.point=D;var n=Object(v.cos)(e);g=n*Object(v.cos)(t),y=n*Object(v.sin)(t),_=Object(v.sin)(e),O(g,y,_)}function D(t,e){t*=v.radians,e*=v.radians;var n=Object(v.cos)(e),r=n*Object(v.cos)(t),o=n*Object(v.sin)(t),a=Object(v.sin)(e),s=y*a-_*o,h=_*r-g*a,m=g*o-y*r,b=Object(v.sqrt)(s*s+h*h+m*m),x=Object(v.asin)(b),w=b&&-x/b;d+=w*s,p+=w*h,f+=w*m,i+=x,u+=x*(g+(g=r)),c+=x*(y+(y=o)),l+=x*(_+(_=a)),O(g,y,_)}e.default=function(t){r=i=o=a=s=u=c=l=d=p=f=0,Object(x.default)(t,w);var e=d,n=p,h=f,m=e*e+n*n+h*h;return m<v.epsilon2&&(e=u,n=c,h=l,i<v.epsilon&&(e=o,n=a,h=s),(m=e*e+n*n+h*h)<v.epsilon2)?[NaN,NaN]:[Object(v.atan2)(n,e)*v.degrees,Object(v.asin)(h/Object(v.sqrt)(m))*v.degrees]}},"./node_modules/taucharts/node_modules/d3-geo/src/circle.js":function(t,e,n){"use strict";n.r(e),n.d(e,"circleStream",function(){return s});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/cartesian.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/constant.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),a=n("./node_modules/taucharts/node_modules/d3-geo/src/rotation.js");function s(t,e,n,i,a,s){if(n){var c=Object(o.cos)(e),l=Object(o.sin)(e),d=i*n;null==a?(a=e+i*o.tau,s=e-d/2):(a=u(c,a),s=u(c,s),(i>0?a<s:a>s)&&(a+=i*o.tau));for(var p,f=a;i>0?f>s:f<s;f-=d)p=Object(r.spherical)([c,-l*Object(o.cos)(f),-l*Object(o.sin)(f)]),t.point(p[0],p[1])}}function u(t,e){(e=Object(r.cartesian)(e))[0]-=t,Object(r.cartesianNormalizeInPlace)(e);var n=Object(o.acos)(-e[1]);return((-e[2]<0?-n:n)+o.tau-o.epsilon)%o.tau}e.default=function(){var t,e,n=Object(i.default)([0,0]),r=Object(i.default)(90),u=Object(i.default)(6),c={point:function(n,r){t.push(n=e(n,r)),n[0]*=o.degrees,n[1]*=o.degrees}};function l(){var i=n.apply(this,arguments),l=r.apply(this,arguments)*o.radians,d=u.apply(this,arguments)*o.radians;return t=[],e=Object(a.rotateRadians)(-i[0]*o.radians,-i[1]*o.radians,0).invert,s(c,l,d,1),i={type:"Polygon",coordinates:[t]},t=e=null,i}return l.center=function(t){return arguments.length?(n="function"==typeof t?t:Object(i.default)([+t[0],+t[1]]),l):n},l.radius=function(t){return arguments.length?(r="function"==typeof t?t:Object(i.default)(+t),l):r},l.precision=function(t){return arguments.length?(u="function"==typeof t?t:Object(i.default)(+t),l):u},l}},"./node_modules/taucharts/node_modules/d3-geo/src/clip/antimeridian.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/clip/index.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js");e.default=Object(r.default)(function(){return!0},function(t){var e,n=NaN,r=NaN,o=NaN;return{lineStart:function(){t.lineStart(),e=1},point:function(a,s){var u,c,l,d,p,f,h,m=a>0?i.pi:-i.pi,g=Object(i.abs)(a-n);Object(i.abs)(g-i.pi)<i.epsilon?(t.point(n,r=(r+s)/2>0?i.halfPi:-i.halfPi),t.point(o,r),t.lineEnd(),t.lineStart(),t.point(m,r),t.point(a,r),e=0):o!==m&&g>=i.pi&&(Object(i.abs)(n-o)<i.epsilon&&(n-=o*i.epsilon),Object(i.abs)(a-m)<i.epsilon&&(a-=m*i.epsilon),u=n,c=r,l=a,d=s,h=Object(i.sin)(u-l),r=Object(i.abs)(h)>i.epsilon?Object(i.atan)((Object(i.sin)(c)*(f=Object(i.cos)(d))*Object(i.sin)(l)-Object(i.sin)(d)*(p=Object(i.cos)(c))*Object(i.sin)(u))/(p*f*h)):(c+d)/2,t.point(o,r),t.lineEnd(),t.lineStart(),t.point(m,r),e=0),t.point(n=a,r=s),o=m},lineEnd:function(){t.lineEnd(),n=r=NaN},clean:function(){return 2-e}}},function(t,e,n,r){var o;if(null==t)o=n*i.halfPi,r.point(-i.pi,o),r.point(0,o),r.point(i.pi,o),r.point(i.pi,0),r.point(i.pi,-o),r.point(0,-o),r.point(-i.pi,-o),r.point(-i.pi,0),r.point(-i.pi,o);else if(Object(i.abs)(t[0]-e[0])>i.epsilon){var a=t[0]<e[0]?i.pi:-i.pi;o=n*a/2,r.point(-a,o),r.point(0,o),r.point(a,o)}else r.point(e[0],e[1])},[-i.pi,-i.halfPi])},"./node_modules/taucharts/node_modules/d3-geo/src/clip/buffer.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/noop.js");e.default=function(){var t,e=[];return{point:function(e,n){t.push([e,n])},lineStart:function(){e.push(t=[])},lineEnd:r.default,rejoin:function(){e.length>1&&e.push(e.pop().concat(e.shift()))},result:function(){var n=e;return e=[],t=null,n}}}},"./node_modules/taucharts/node_modules/d3-geo/src/clip/circle.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/cartesian.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/circle.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),a=n("./node_modules/taucharts/node_modules/d3-geo/src/pointEqual.js"),s=n("./node_modules/taucharts/node_modules/d3-geo/src/clip/index.js");e.default=function(t){var e=Object(o.cos)(t),n=6*o.radians,u=e>0,c=Object(o.abs)(e)>o.epsilon;function l(t,n){return Object(o.cos)(t)*Object(o.cos)(n)>e}function d(t,n,i){var a=Object(r.cartesian)(t),s=Object(r.cartesian)(n),u=[1,0,0],c=Object(r.cartesianCross)(a,s),l=Object(r.cartesianDot)(c,c),d=c[0],p=l-d*d;if(!p)return!i&&t;var f=e*l/p,h=-e*d/p,m=Object(r.cartesianCross)(u,c),g=Object(r.cartesianScale)(u,f),y=Object(r.cartesianScale)(c,h);Object(r.cartesianAddInPlace)(g,y);var _=m,v=Object(r.cartesianDot)(g,_),b=Object(r.cartesianDot)(_,_),x=v*v-b*(Object(r.cartesianDot)(g,g)-1);if(!(x<0)){var w=Object(o.sqrt)(x),j=Object(r.cartesianScale)(_,(-v-w)/b);if(Object(r.cartesianAddInPlace)(j,g),j=Object(r.spherical)(j),!i)return j;var O,C=t[0],E=n[0],S=t[1],M=n[1];E<C&&(O=C,C=E,E=O);var T=E-C,k=Object(o.abs)(T-o.pi)<o.epsilon,A=k||T<o.epsilon;if(!k&&M<S&&(O=S,S=M,M=O),A?k?S+M>0^j[1]<(Object(o.abs)(j[0]-C)<o.epsilon?S:M):S<=j[1]&&j[1]<=M:T>o.pi^(C<=j[0]&&j[0]<=E)){var D=Object(r.cartesianScale)(_,(-v+w)/b);return Object(r.cartesianAddInPlace)(D,g),[j,Object(r.spherical)(D)]}}}function p(e,n){var r=u?t:o.pi-t,i=0;return e<-r?i|=1:e>r&&(i|=2),n<-r?i|=4:n>r&&(i|=8),i}return Object(s.default)(l,function(t){var e,n,r,i,s;return{lineStart:function(){i=r=!1,s=1},point:function(f,h){var m,g=[f,h],y=l(f,h),_=u?y?0:p(f,h):y?p(f+(f<0?o.pi:-o.pi),h):0;if(!e&&(i=r=y)&&t.lineStart(),y!==r&&(!(m=d(e,g))||Object(a.default)(e,m)||Object(a.default)(g,m))&&(g[0]+=o.epsilon,g[1]+=o.epsilon,y=l(g[0],g[1])),y!==r)s=0,y?(t.lineStart(),m=d(g,e),t.point(m[0],m[1])):(m=d(e,g),t.point(m[0],m[1]),t.lineEnd()),e=m;else if(c&&e&&u^y){var v;_&n||!(v=d(g,e,!0))||(s=0,u?(t.lineStart(),t.point(v[0][0],v[0][1]),t.point(v[1][0],v[1][1]),t.lineEnd()):(t.point(v[1][0],v[1][1]),t.lineEnd(),t.lineStart(),t.point(v[0][0],v[0][1])))}!y||e&&Object(a.default)(e,g)||t.point(g[0],g[1]),e=g,r=y,n=_},lineEnd:function(){r&&t.lineEnd(),e=null},clean:function(){return s|(i&&r)<<1}}},function(e,r,o,a){Object(i.circleStream)(a,t,n,o,e,r)},u?[0,-t]:[-o.pi,t-o.pi])}},"./node_modules/taucharts/node_modules/d3-geo/src/clip/extent.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/clip/rectangle.js");e.default=function(){var t,e,n,i=0,o=0,a=960,s=500;return n={stream:function(n){return t&&e===n?t:t=Object(r.default)(i,o,a,s)(e=n)},extent:function(r){return arguments.length?(i=+r[0][0],o=+r[0][1],a=+r[1][0],s=+r[1][1],t=e=null,n):[[i,o],[a,s]]}}}},"./node_modules/taucharts/node_modules/d3-geo/src/clip/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/clip/buffer.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/clip/rejoin.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),a=n("./node_modules/taucharts/node_modules/d3-geo/src/polygonContains.js"),s=n("./node_modules/d3-array/index.js");function u(t){return t.length>1}function c(t,e){return((t=t.x)[0]<0?t[1]-o.halfPi-o.epsilon:o.halfPi-t[1])-((e=e.x)[0]<0?e[1]-o.halfPi-o.epsilon:o.halfPi-e[1])}e.default=function(t,e,n,o){return function(l){var d,p,f,h=e(l),m=Object(r.default)(),g=e(m),y=!1,_={point:v,lineStart:x,lineEnd:w,polygonStart:function(){_.point=j,_.lineStart=O,_.lineEnd=C,p=[],d=[]},polygonEnd:function(){_.point=v,_.lineStart=x,_.lineEnd=w,p=Object(s.merge)(p);var t=Object(a.default)(d,o);p.length?(y||(l.polygonStart(),y=!0),Object(i.default)(p,c,t,n,l)):t&&(y||(l.polygonStart(),y=!0),l.lineStart(),n(null,null,1,l),l.lineEnd()),y&&(l.polygonEnd(),y=!1),p=d=null},sphere:function(){l.polygonStart(),l.lineStart(),n(null,null,1,l),l.lineEnd(),l.polygonEnd()}};function v(e,n){t(e,n)&&l.point(e,n)}function b(t,e){h.point(t,e)}function x(){_.point=b,h.lineStart()}function w(){_.point=v,h.lineEnd()}function j(t,e){f.push([t,e]),g.point(t,e)}function O(){g.lineStart(),f=[]}function C(){j(f[0][0],f[0][1]),g.lineEnd();var t,e,n,r,i=g.clean(),o=m.result(),a=o.length;if(f.pop(),d.push(f),f=null,a)if(1&i){if((e=(n=o[0]).length-1)>0){for(y||(l.polygonStart(),y=!0),l.lineStart(),t=0;t<e;++t)l.point((r=n[t])[0],r[1]);l.lineEnd()}}else a>1&&2&i&&o.push(o.pop().concat(o.shift())),p.push(o.filter(u))}return _}}},"./node_modules/taucharts/node_modules/d3-geo/src/clip/line.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e,n,r,i,o){var a,s=t[0],u=t[1],c=0,l=1,d=e[0]-s,p=e[1]-u;if(a=n-s,d||!(a>0)){if(a/=d,d<0){if(a<c)return;a<l&&(l=a)}else if(d>0){if(a>l)return;a>c&&(c=a)}if(a=i-s,d||!(a<0)){if(a/=d,d<0){if(a>l)return;a>c&&(c=a)}else if(d>0){if(a<c)return;a<l&&(l=a)}if(a=r-u,p||!(a>0)){if(a/=p,p<0){if(a<c)return;a<l&&(l=a)}else if(p>0){if(a>l)return;a>c&&(c=a)}if(a=o-u,p||!(a<0)){if(a/=p,p<0){if(a>l)return;a>c&&(c=a)}else if(p>0){if(a<c)return;a<l&&(l=a)}return c>0&&(t[0]=s+c*d,t[1]=u+c*p),l<1&&(e[0]=s+l*d,e[1]=u+l*p),!0}}}}}},"./node_modules/taucharts/node_modules/d3-geo/src/clip/rectangle.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return l});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/clip/buffer.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/clip/line.js"),a=n("./node_modules/taucharts/node_modules/d3-geo/src/clip/rejoin.js"),s=n("./node_modules/d3-array/index.js"),u=1e9,c=-u;function l(t,e,n,l){function d(r,i){return t<=r&&r<=n&&e<=i&&i<=l}function p(r,i,o,a){var s=0,u=0;if(null==r||(s=f(r,o))!==(u=f(i,o))||m(r,i)<0^o>0)do{a.point(0===s||3===s?t:n,s>1?l:e)}while((s=(s+o+4)%4)!==u);else a.point(i[0],i[1])}function f(i,o){return Object(r.abs)(i[0]-t)<r.epsilon?o>0?0:3:Object(r.abs)(i[0]-n)<r.epsilon?o>0?2:1:Object(r.abs)(i[1]-e)<r.epsilon?o>0?1:0:o>0?3:2}function h(t,e){return m(t.x,e.x)}function m(t,e){var n=f(t,1),r=f(e,1);return n!==r?n-r:0===n?e[1]-t[1]:1===n?t[0]-e[0]:2===n?t[1]-e[1]:e[0]-t[0]}return function(r){var f,m,g,y,_,v,b,x,w,j,O,C=r,E=Object(i.default)(),S={point:M,lineStart:function(){S.point=T,m&&m.push(g=[]);j=!0,w=!1,b=x=NaN},lineEnd:function(){f&&(T(y,_),v&&w&&E.rejoin(),f.push(E.result()));S.point=M,w&&C.lineEnd()},polygonStart:function(){C=E,f=[],m=[],O=!0},polygonEnd:function(){var e=function(){for(var e=0,n=0,r=m.length;n<r;++n)for(var i,o,a=m[n],s=1,u=a.length,c=a[0],d=c[0],p=c[1];s<u;++s)i=d,o=p,c=a[s],d=c[0],p=c[1],o<=l?p>l&&(d-i)*(l-o)>(p-o)*(t-i)&&++e:p<=l&&(d-i)*(l-o)<(p-o)*(t-i)&&--e;return e}(),n=O&&e,i=(f=Object(s.merge)(f)).length;(n||i)&&(r.polygonStart(),n&&(r.lineStart(),p(null,null,1,r),r.lineEnd()),i&&Object(a.default)(f,h,e,p,r),r.polygonEnd());C=r,f=m=g=null}};function M(t,e){d(t,e)&&C.point(t,e)}function T(r,i){var a=d(r,i);if(m&&g.push([r,i]),j)y=r,_=i,v=a,j=!1,a&&(C.lineStart(),C.point(r,i));else if(a&&w)C.point(r,i);else{var s=[b=Math.max(c,Math.min(u,b)),x=Math.max(c,Math.min(u,x))],p=[r=Math.max(c,Math.min(u,r)),i=Math.max(c,Math.min(u,i))];Object(o.default)(s,p,t,e,n,l)?(w||(C.lineStart(),C.point(s[0],s[1])),C.point(p[0],p[1]),a||C.lineEnd(),O=!1):a&&(C.lineStart(),C.point(r,i),O=!1)}b=r,x=i,w=a}return S}}},"./node_modules/taucharts/node_modules/d3-geo/src/clip/rejoin.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/pointEqual.js");function i(t,e,n,r){this.x=t,this.z=e,this.o=n,this.e=r,this.v=!1,this.n=this.p=null}function o(t){if(e=t.length){for(var e,n,r=0,i=t[0];++r<e;)i.n=n=t[r],n.p=i,i=n;i.n=n=t[0],n.p=i}}e.default=function(t,e,n,a,s){var u,c,l=[],d=[];if(t.forEach(function(t){if(!((e=t.length-1)<=0)){var e,n,o=t[0],a=t[e];if(Object(r.default)(o,a)){for(s.lineStart(),u=0;u<e;++u)s.point((o=t[u])[0],o[1]);s.lineEnd()}else l.push(n=new i(o,t,null,!0)),d.push(n.o=new i(o,null,n,!1)),l.push(n=new i(a,t,null,!1)),d.push(n.o=new i(a,null,n,!0))}}),l.length){for(d.sort(e),o(l),o(d),u=0,c=d.length;u<c;++u)d[u].e=n=!n;for(var p,f,h=l[0];;){for(var m=h,g=!0;m.v;)if((m=m.n)===h)return;p=m.z,s.lineStart();do{if(m.v=m.o.v=!0,m.e){if(g)for(u=0,c=p.length;u<c;++u)s.point((f=p[u])[0],f[1]);else a(m.x,m.n.x,1,s);m=m.n}else{if(g)for(p=m.p.z,u=p.length-1;u>=0;--u)s.point((f=p[u])[0],f[1]);else a(m.x,m.p.x,-1,s);m=m.p}p=(m=m.o).z,g=!g}while(!m.v);s.lineEnd()}}}},"./node_modules/taucharts/node_modules/d3-geo/src/compose.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){function n(n,r){return n=t(n,r),e(n[0],n[1])}return t.invert&&e.invert&&(n.invert=function(n,r){return(n=e.invert(n,r))&&t.invert(n[0],n[1])}),n}},"./node_modules/taucharts/node_modules/d3-geo/src/constant.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return function(){return t}}},"./node_modules/taucharts/node_modules/d3-geo/src/contains.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/polygonContains.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/distance.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),a={Feature:function(t,e){return u(t.geometry,e)},FeatureCollection:function(t,e){for(var n=t.features,r=-1,i=n.length;++r<i;)if(u(n[r].geometry,e))return!0;return!1}},s={Sphere:function(){return!0},Point:function(t,e){return c(t.coordinates,e)},MultiPoint:function(t,e){for(var n=t.coordinates,r=-1,i=n.length;++r<i;)if(c(n[r],e))return!0;return!1},LineString:function(t,e){return l(t.coordinates,e)},MultiLineString:function(t,e){for(var n=t.coordinates,r=-1,i=n.length;++r<i;)if(l(n[r],e))return!0;return!1},Polygon:function(t,e){return d(t.coordinates,e)},MultiPolygon:function(t,e){for(var n=t.coordinates,r=-1,i=n.length;++r<i;)if(d(n[r],e))return!0;return!1},GeometryCollection:function(t,e){for(var n=t.geometries,r=-1,i=n.length;++r<i;)if(u(n[r],e))return!0;return!1}};function u(t,e){return!(!t||!s.hasOwnProperty(t.type))&&s[t.type](t,e)}function c(t,e){return 0===Object(i.default)(t,e)}function l(t,e){var n=Object(i.default)(t[0],t[1]);return Object(i.default)(t[0],e)+Object(i.default)(e,t[1])<=n+o.epsilon}function d(t,e){return!!Object(r.default)(t.map(p),f(e))}function p(t){return(t=t.map(f)).pop(),t}function f(t){return[t[0]*o.radians,t[1]*o.radians]}e.default=function(t,e){return(t&&a.hasOwnProperty(t.type)?a[t.type]:u)(t,e)}},"./node_modules/taucharts/node_modules/d3-geo/src/distance.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/length.js"),i=[null,null],o={type:"LineString",coordinates:i};e.default=function(t,e){return i[0]=t,i[1]=e,Object(r.default)(o)}},"./node_modules/taucharts/node_modules/d3-geo/src/graticule.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return s}),n.d(e,"graticule10",function(){return u});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js");function o(t,e,n){var o=Object(r.range)(t,e-i.epsilon,n).concat(e);return function(t){return o.map(function(e){return[t,e]})}}function a(t,e,n){var o=Object(r.range)(t,e-i.epsilon,n).concat(e);return function(t){return o.map(function(e){return[e,t]})}}function s(){var t,e,n,s,u,c,l,d,p,f,h,m,g=10,y=g,_=90,v=360,b=2.5;function x(){return{type:"MultiLineString",coordinates:w()}}function w(){return Object(r.range)(Object(i.ceil)(s/_)*_,n,_).map(h).concat(Object(r.range)(Object(i.ceil)(d/v)*v,l,v).map(m)).concat(Object(r.range)(Object(i.ceil)(e/g)*g,t,g).filter(function(t){return Object(i.abs)(t%_)>i.epsilon}).map(p)).concat(Object(r.range)(Object(i.ceil)(c/y)*y,u,y).filter(function(t){return Object(i.abs)(t%v)>i.epsilon}).map(f))}return x.lines=function(){return w().map(function(t){return{type:"LineString",coordinates:t}})},x.outline=function(){return{type:"Polygon",coordinates:[h(s).concat(m(l).slice(1),h(n).reverse().slice(1),m(d).reverse().slice(1))]}},x.extent=function(t){return arguments.length?x.extentMajor(t).extentMinor(t):x.extentMinor()},x.extentMajor=function(t){return arguments.length?(s=+t[0][0],n=+t[1][0],d=+t[0][1],l=+t[1][1],s>n&&(t=s,s=n,n=t),d>l&&(t=d,d=l,l=t),x.precision(b)):[[s,d],[n,l]]},x.extentMinor=function(n){return arguments.length?(e=+n[0][0],t=+n[1][0],c=+n[0][1],u=+n[1][1],e>t&&(n=e,e=t,t=n),c>u&&(n=c,c=u,u=n),x.precision(b)):[[e,c],[t,u]]},x.step=function(t){return arguments.length?x.stepMajor(t).stepMinor(t):x.stepMinor()},x.stepMajor=function(t){return arguments.length?(_=+t[0],v=+t[1],x):[_,v]},x.stepMinor=function(t){return arguments.length?(g=+t[0],y=+t[1],x):[g,y]},x.precision=function(r){return arguments.length?(b=+r,p=o(c,u,90),f=a(e,t,b),h=o(d,l,90),m=a(s,n,b),x):b},x.extentMajor([[-180,-90+i.epsilon],[180,90-i.epsilon]]).extentMinor([[-180,-80-i.epsilon],[180,80+i.epsilon]])}function u(){return s()()}},"./node_modules/taucharts/node_modules/d3-geo/src/identity.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return t}},"./node_modules/taucharts/node_modules/d3-geo/src/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/area.js");n.d(e,"geoArea",function(){return r.default});var i=n("./node_modules/taucharts/node_modules/d3-geo/src/bounds.js");n.d(e,"geoBounds",function(){return i.default});var o=n("./node_modules/taucharts/node_modules/d3-geo/src/centroid.js");n.d(e,"geoCentroid",function(){return o.default});var a=n("./node_modules/taucharts/node_modules/d3-geo/src/circle.js");n.d(e,"geoCircle",function(){return a.default});var s=n("./node_modules/taucharts/node_modules/d3-geo/src/clip/antimeridian.js");n.d(e,"geoClipAntimeridian",function(){return s.default});var u=n("./node_modules/taucharts/node_modules/d3-geo/src/clip/circle.js");n.d(e,"geoClipCircle",function(){return u.default});var c=n("./node_modules/taucharts/node_modules/d3-geo/src/clip/extent.js");n.d(e,"geoClipExtent",function(){return c.default});var l=n("./node_modules/taucharts/node_modules/d3-geo/src/clip/rectangle.js");n.d(e,"geoClipRectangle",function(){return l.default});var d=n("./node_modules/taucharts/node_modules/d3-geo/src/contains.js");n.d(e,"geoContains",function(){return d.default});var p=n("./node_modules/taucharts/node_modules/d3-geo/src/distance.js");n.d(e,"geoDistance",function(){return p.default});var f=n("./node_modules/taucharts/node_modules/d3-geo/src/graticule.js");n.d(e,"geoGraticule",function(){return f.default}),n.d(e,"geoGraticule10",function(){return f.graticule10});var h=n("./node_modules/taucharts/node_modules/d3-geo/src/interpolate.js");n.d(e,"geoInterpolate",function(){return h.default});var m=n("./node_modules/taucharts/node_modules/d3-geo/src/length.js");n.d(e,"geoLength",function(){return m.default});var g=n("./node_modules/taucharts/node_modules/d3-geo/src/path/index.js");n.d(e,"geoPath",function(){return g.default});var y=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/albers.js");n.d(e,"geoAlbers",function(){return y.default});var _=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/albersUsa.js");n.d(e,"geoAlbersUsa",function(){return _.default});var v=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/azimuthalEqualArea.js");n.d(e,"geoAzimuthalEqualArea",function(){return v.default}),n.d(e,"geoAzimuthalEqualAreaRaw",function(){return v.azimuthalEqualAreaRaw});var b=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/azimuthalEquidistant.js");n.d(e,"geoAzimuthalEquidistant",function(){return b.default}),n.d(e,"geoAzimuthalEquidistantRaw",function(){return b.azimuthalEquidistantRaw});var x=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/conicConformal.js");n.d(e,"geoConicConformal",function(){return x.default}),n.d(e,"geoConicConformalRaw",function(){return x.conicConformalRaw});var w=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/conicEqualArea.js");n.d(e,"geoConicEqualArea",function(){return w.default}),n.d(e,"geoConicEqualAreaRaw",function(){return w.conicEqualAreaRaw});var j=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/conicEquidistant.js");n.d(e,"geoConicEquidistant",function(){return j.default}),n.d(e,"geoConicEquidistantRaw",function(){return j.conicEquidistantRaw});var O=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/equalEarth.js");n.d(e,"geoEqualEarth",function(){return O.default}),n.d(e,"geoEqualEarthRaw",function(){return O.equalEarthRaw});var C=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/equirectangular.js");n.d(e,"geoEquirectangular",function(){return C.default}),n.d(e,"geoEquirectangularRaw",function(){return C.equirectangularRaw});var E=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/gnomonic.js");n.d(e,"geoGnomonic",function(){return E.default}),n.d(e,"geoGnomonicRaw",function(){return E.gnomonicRaw});var S=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/identity.js");n.d(e,"geoIdentity",function(){return S.default});var M=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/index.js");n.d(e,"geoProjection",function(){return M.default}),n.d(e,"geoProjectionMutator",function(){return M.projectionMutator});var T=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/mercator.js");n.d(e,"geoMercator",function(){return T.default}),n.d(e,"geoMercatorRaw",function(){return T.mercatorRaw});var k=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/naturalEarth1.js");n.d(e,"geoNaturalEarth1",function(){return k.default}),n.d(e,"geoNaturalEarth1Raw",function(){return k.naturalEarth1Raw});var A=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/orthographic.js");n.d(e,"geoOrthographic",function(){return A.default}),n.d(e,"geoOrthographicRaw",function(){return A.orthographicRaw});var D=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/stereographic.js");n.d(e,"geoStereographic",function(){return D.default}),n.d(e,"geoStereographicRaw",function(){return D.stereographicRaw});var I=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/transverseMercator.js");n.d(e,"geoTransverseMercator",function(){return I.default}),n.d(e,"geoTransverseMercatorRaw",function(){return I.transverseMercatorRaw});var P=n("./node_modules/taucharts/node_modules/d3-geo/src/rotation.js");n.d(e,"geoRotation",function(){return P.default});var R=n("./node_modules/taucharts/node_modules/d3-geo/src/stream.js");n.d(e,"geoStream",function(){return R.default});var N=n("./node_modules/taucharts/node_modules/d3-geo/src/transform.js");n.d(e,"geoTransform",function(){return N.default})},"./node_modules/taucharts/node_modules/d3-geo/src/interpolate.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js");e.default=function(t,e){var n=t[0]*r.radians,i=t[1]*r.radians,o=e[0]*r.radians,a=e[1]*r.radians,s=Object(r.cos)(i),u=Object(r.sin)(i),c=Object(r.cos)(a),l=Object(r.sin)(a),d=s*Object(r.cos)(n),p=s*Object(r.sin)(n),f=c*Object(r.cos)(o),h=c*Object(r.sin)(o),m=2*Object(r.asin)(Object(r.sqrt)(Object(r.haversin)(a-i)+s*c*Object(r.haversin)(o-n))),g=Object(r.sin)(m),y=m?function(t){var e=Object(r.sin)(t*=m)/g,n=Object(r.sin)(m-t)/g,i=n*d+e*f,o=n*p+e*h,a=n*u+e*l;return[Object(r.atan2)(o,i)*r.degrees,Object(r.atan2)(a,Object(r.sqrt)(i*i+o*o))*r.degrees]}:function(){return[n*r.degrees,i*r.degrees]};return y.distance=m,y}},"./node_modules/taucharts/node_modules/d3-geo/src/length.js":function(t,e,n){"use strict";n.r(e);var r,i,o,a=n("./node_modules/taucharts/node_modules/d3-geo/src/adder.js"),s=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),u=n("./node_modules/taucharts/node_modules/d3-geo/src/noop.js"),c=n("./node_modules/taucharts/node_modules/d3-geo/src/stream.js"),l=Object(a.default)(),d={sphere:u.default,point:u.default,lineStart:function(){d.point=f,d.lineEnd=p},lineEnd:u.default,polygonStart:u.default,polygonEnd:u.default};function p(){d.point=d.lineEnd=u.default}function f(t,e){t*=s.radians,e*=s.radians,r=t,i=Object(s.sin)(e),o=Object(s.cos)(e),d.point=h}function h(t,e){t*=s.radians,e*=s.radians;var n=Object(s.sin)(e),a=Object(s.cos)(e),u=Object(s.abs)(t-r),c=Object(s.cos)(u),d=a*Object(s.sin)(u),p=o*n-i*a*c,f=i*n+o*a*c;l.add(Object(s.atan2)(Object(s.sqrt)(d*d+p*p),f)),r=t,i=n,o=a}e.default=function(t){return l.reset(),Object(c.default)(t,d),+l}},"./node_modules/taucharts/node_modules/d3-geo/src/math.js":function(t,e,n){"use strict";n.r(e),n.d(e,"epsilon",function(){return r}),n.d(e,"epsilon2",function(){return i}),n.d(e,"pi",function(){return o}),n.d(e,"halfPi",function(){return a}),n.d(e,"quarterPi",function(){return s}),n.d(e,"tau",function(){return u}),n.d(e,"degrees",function(){return c}),n.d(e,"radians",function(){return l}),n.d(e,"abs",function(){return d}),n.d(e,"atan",function(){return p}),n.d(e,"atan2",function(){return f}),n.d(e,"cos",function(){return h}),n.d(e,"ceil",function(){return m}),n.d(e,"exp",function(){return g}),n.d(e,"floor",function(){return y}),n.d(e,"log",function(){return _}),n.d(e,"pow",function(){return v}),n.d(e,"sin",function(){return b}),n.d(e,"sign",function(){return x}),n.d(e,"sqrt",function(){return w}),n.d(e,"tan",function(){return j}),n.d(e,"acos",function(){return O}),n.d(e,"asin",function(){return C}),n.d(e,"haversin",function(){return E});var r=1e-6,i=1e-12,o=Math.PI,a=o/2,s=o/4,u=2*o,c=180/o,l=o/180,d=Math.abs,p=Math.atan,f=Math.atan2,h=Math.cos,m=Math.ceil,g=Math.exp,y=Math.floor,_=Math.log,v=Math.pow,b=Math.sin,x=Math.sign||function(t){return t>0?1:t<0?-1:0},w=Math.sqrt,j=Math.tan;function O(t){return t>1?0:t<-1?o:Math.acos(t)}function C(t){return t>1?a:t<-1?-a:Math.asin(t)}function E(t){return(t=b(t/2))*t}},"./node_modules/taucharts/node_modules/d3-geo/src/noop.js":function(t,e,n){"use strict";function r(){}n.r(e),n.d(e,"default",function(){return r})},"./node_modules/taucharts/node_modules/d3-geo/src/path/area.js":function(t,e,n){"use strict";n.r(e);var r,i,o,a,s=n("./node_modules/taucharts/node_modules/d3-geo/src/adder.js"),u=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),c=n("./node_modules/taucharts/node_modules/d3-geo/src/noop.js"),l=Object(s.default)(),d=Object(s.default)(),p={point:c.default,lineStart:c.default,lineEnd:c.default,polygonStart:function(){p.lineStart=f,p.lineEnd=g},polygonEnd:function(){p.lineStart=p.lineEnd=p.point=c.default,l.add(Object(u.abs)(d)),d.reset()},result:function(){var t=l/2;return l.reset(),t}};function f(){p.point=h}function h(t,e){p.point=m,r=o=t,i=a=e}function m(t,e){d.add(a*t-o*e),o=t,a=e}function g(){m(r,i)}e.default=p},"./node_modules/taucharts/node_modules/d3-geo/src/path/bounds.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/noop.js"),i=1/0,o=i,a=-i,s=a,u={point:function(t,e){t<i&&(i=t);t>a&&(a=t);e<o&&(o=e);e>s&&(s=e)},lineStart:r.default,lineEnd:r.default,polygonStart:r.default,polygonEnd:r.default,result:function(){var t=[[i,o],[a,s]];return a=s=-(o=i=1/0),t}};e.default=u},"./node_modules/taucharts/node_modules/d3-geo/src/path/centroid.js":function(t,e,n){"use strict";n.r(e);var r,i,o,a,s=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),u=0,c=0,l=0,d=0,p=0,f=0,h=0,m=0,g=0,y={point:_,lineStart:v,lineEnd:w,polygonStart:function(){y.lineStart=j,y.lineEnd=O},polygonEnd:function(){y.point=_,y.lineStart=v,y.lineEnd=w},result:function(){var t=g?[h/g,m/g]:f?[d/f,p/f]:l?[u/l,c/l]:[NaN,NaN];return u=c=l=d=p=f=h=m=g=0,t}};function _(t,e){u+=t,c+=e,++l}function v(){y.point=b}function b(t,e){y.point=x,_(o=t,a=e)}function x(t,e){var n=t-o,r=e-a,i=Object(s.sqrt)(n*n+r*r);d+=i*(o+t)/2,p+=i*(a+e)/2,f+=i,_(o=t,a=e)}function w(){y.point=_}function j(){y.point=C}function O(){E(r,i)}function C(t,e){y.point=E,_(r=o=t,i=a=e)}function E(t,e){var n=t-o,r=e-a,i=Object(s.sqrt)(n*n+r*r);d+=i*(o+t)/2,p+=i*(a+e)/2,f+=i,h+=(i=a*t-o*e)*(o+t),m+=i*(a+e),g+=3*i,_(o=t,a=e)}e.default=y},"./node_modules/taucharts/node_modules/d3-geo/src/path/context.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return o});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/noop.js");function o(t){this._context=t}o.prototype={_radius:4.5,pointRadius:function(t){return this._radius=t,this},polygonStart:function(){this._line=0},polygonEnd:function(){this._line=NaN},lineStart:function(){this._point=0},lineEnd:function(){0===this._line&&this._context.closePath(),this._point=NaN},point:function(t,e){switch(this._point){case 0:this._context.moveTo(t,e),this._point=1;break;case 1:this._context.lineTo(t,e);break;default:this._context.moveTo(t+this._radius,e),this._context.arc(t,e,this._radius,0,r.tau)}},result:i.default}},"./node_modules/taucharts/node_modules/d3-geo/src/path/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/identity.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/stream.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/path/area.js"),a=n("./node_modules/taucharts/node_modules/d3-geo/src/path/bounds.js"),s=n("./node_modules/taucharts/node_modules/d3-geo/src/path/centroid.js"),u=n("./node_modules/taucharts/node_modules/d3-geo/src/path/context.js"),c=n("./node_modules/taucharts/node_modules/d3-geo/src/path/measure.js"),l=n("./node_modules/taucharts/node_modules/d3-geo/src/path/string.js");e.default=function(t,e){var n,d,p=4.5;function f(t){return t&&("function"==typeof p&&d.pointRadius(+p.apply(this,arguments)),Object(i.default)(t,n(d))),d.result()}return f.area=function(t){return Object(i.default)(t,n(o.default)),o.default.result()},f.measure=function(t){return Object(i.default)(t,n(c.default)),c.default.result()},f.bounds=function(t){return Object(i.default)(t,n(a.default)),a.default.result()},f.centroid=function(t){return Object(i.default)(t,n(s.default)),s.default.result()},f.projection=function(e){return arguments.length?(n=null==e?(t=null,r.default):(t=e).stream,f):t},f.context=function(t){return arguments.length?(d=null==t?(e=null,new l.default):new u.default(e=t),"function"!=typeof p&&d.pointRadius(p),f):e},f.pointRadius=function(t){return arguments.length?(p="function"==typeof t?t:(d.pointRadius(+t),+t),f):p},f.projection(t).context(e)}},"./node_modules/taucharts/node_modules/d3-geo/src/path/measure.js":function(t,e,n){"use strict";n.r(e);var r,i,o,a,s,u=n("./node_modules/taucharts/node_modules/d3-geo/src/adder.js"),c=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),l=n("./node_modules/taucharts/node_modules/d3-geo/src/noop.js"),d=Object(u.default)(),p={point:l.default,lineStart:function(){p.point=f},lineEnd:function(){r&&h(i,o),p.point=l.default},polygonStart:function(){r=!0},polygonEnd:function(){r=null},result:function(){var t=+d;return d.reset(),t}};function f(t,e){p.point=h,i=a=t,o=s=e}function h(t,e){a-=t,s-=e,d.add(Object(c.sqrt)(a*a+s*s)),a=t,s=e}e.default=p},"./node_modules/taucharts/node_modules/d3-geo/src/path/string.js":function(t,e,n){"use strict";function r(){this._string=[]}function i(t){return"m0,"+t+"a"+t+","+t+" 0 1,1 0,"+-2*t+"a"+t+","+t+" 0 1,1 0,"+2*t+"z"}n.r(e),n.d(e,"default",function(){return r}),r.prototype={_radius:4.5,_circle:i(4.5),pointRadius:function(t){return(t=+t)!==this._radius&&(this._radius=t,this._circle=null),this},polygonStart:function(){this._line=0},polygonEnd:function(){this._line=NaN},lineStart:function(){this._point=0},lineEnd:function(){0===this._line&&this._string.push("Z"),this._point=NaN},point:function(t,e){switch(this._point){case 0:this._string.push("M",t,",",e),this._point=1;break;case 1:this._string.push("L",t,",",e);break;default:null==this._circle&&(this._circle=i(this._radius)),this._string.push("M",t,",",e,this._circle)}},result:function(){if(this._string.length){var t=this._string.join("");return this._string=[],t}return null}}},"./node_modules/taucharts/node_modules/d3-geo/src/pointEqual.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js");e.default=function(t,e){return Object(r.abs)(t[0]-e[0])<r.epsilon&&Object(r.abs)(t[1]-e[1])<r.epsilon}},"./node_modules/taucharts/node_modules/d3-geo/src/polygonContains.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/adder.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/cartesian.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),a=Object(r.default)();e.default=function(t,e){var n=e[0],r=e[1],s=Object(o.sin)(r),u=[Object(o.sin)(n),-Object(o.cos)(n),0],c=0,l=0;a.reset(),1===s?r=o.halfPi+o.epsilon:-1===s&&(r=-o.halfPi-o.epsilon);for(var d=0,p=t.length;d<p;++d)if(h=(f=t[d]).length)for(var f,h,m=f[h-1],g=m[0],y=m[1]/2+o.quarterPi,_=Object(o.sin)(y),v=Object(o.cos)(y),b=0;b<h;++b,g=w,_=O,v=C,m=x){var x=f[b],w=x[0],j=x[1]/2+o.quarterPi,O=Object(o.sin)(j),C=Object(o.cos)(j),E=w-g,S=E>=0?1:-1,M=S*E,T=M>o.pi,k=_*O;if(a.add(Object(o.atan2)(k*S*Object(o.sin)(M),v*C+k*Object(o.cos)(M))),c+=T?E+S*o.tau:E,T^g>=n^w>=n){var A=Object(i.cartesianCross)(Object(i.cartesian)(m),Object(i.cartesian)(x));Object(i.cartesianNormalizeInPlace)(A);var D=Object(i.cartesianCross)(u,A);Object(i.cartesianNormalizeInPlace)(D);var I=(T^E>=0?-1:1)*Object(o.asin)(D[2]);(r>I||r===I&&(A[0]||A[1]))&&(l+=T^E>=0?1:-1)}}return(c<-o.epsilon||c<o.epsilon&&a<-o.epsilon)^1&l}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/albers.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/conicEqualArea.js");e.default=function(){return Object(r.default)().parallels([29.5,45.5]).scale(1070).translate([480,250]).rotate([96,0]).center([-.6,38.7])}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/albersUsa.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/albers.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/conicEqualArea.js"),a=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/fit.js");e.default=function(){var t,e,n,s,u,c,l=Object(i.default)(),d=Object(o.default)().rotate([154,0]).center([-2,58.5]).parallels([55,65]),p=Object(o.default)().rotate([157,0]).center([-3,19.9]).parallels([8,18]),f={point:function(t,e){c=[t,e]}};function h(t){var e=t[0],r=t[1];return c=null,n.point(e,r),c||(s.point(e,r),c)||(u.point(e,r),c)}function m(){return t=e=null,h}return h.invert=function(t){var e=l.scale(),n=l.translate(),r=(t[0]-n[0])/e,i=(t[1]-n[1])/e;return(i>=.12&&i<.234&&r>=-.425&&r<-.214?d:i>=.166&&i<.234&&r>=-.214&&r<-.115?p:l).invert(t)},h.stream=function(n){return t&&e===n?t:(r=[l.stream(e=n),d.stream(n),p.stream(n)],i=r.length,t={point:function(t,e){for(var n=-1;++n<i;)r[n].point(t,e)},sphere:function(){for(var t=-1;++t<i;)r[t].sphere()},lineStart:function(){for(var t=-1;++t<i;)r[t].lineStart()},lineEnd:function(){for(var t=-1;++t<i;)r[t].lineEnd()},polygonStart:function(){for(var t=-1;++t<i;)r[t].polygonStart()},polygonEnd:function(){for(var t=-1;++t<i;)r[t].polygonEnd()}});var r,i},h.precision=function(t){return arguments.length?(l.precision(t),d.precision(t),p.precision(t),m()):l.precision()},h.scale=function(t){return arguments.length?(l.scale(t),d.scale(.35*t),p.scale(t),h.translate(l.translate())):l.scale()},h.translate=function(t){if(!arguments.length)return l.translate();var e=l.scale(),i=+t[0],o=+t[1];return n=l.translate(t).clipExtent([[i-.455*e,o-.238*e],[i+.455*e,o+.238*e]]).stream(f),s=d.translate([i-.307*e,o+.201*e]).clipExtent([[i-.425*e+r.epsilon,o+.12*e+r.epsilon],[i-.214*e-r.epsilon,o+.234*e-r.epsilon]]).stream(f),u=p.translate([i-.205*e,o+.212*e]).clipExtent([[i-.214*e+r.epsilon,o+.166*e+r.epsilon],[i-.115*e-r.epsilon,o+.234*e-r.epsilon]]).stream(f),m()},h.fitExtent=function(t,e){return Object(a.fitExtent)(h,t,e)},h.fitSize=function(t,e){return Object(a.fitSize)(h,t,e)},h.fitWidth=function(t,e){return Object(a.fitWidth)(h,t,e)},h.fitHeight=function(t,e){return Object(a.fitHeight)(h,t,e)},h.scale(1070)}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/azimuthal.js":function(t,e,n){"use strict";n.r(e),n.d(e,"azimuthalRaw",function(){return i}),n.d(e,"azimuthalInvert",function(){return o});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js");function i(t){return function(e,n){var i=Object(r.cos)(e),o=Object(r.cos)(n),a=t(i*o);return[a*o*Object(r.sin)(e),a*Object(r.sin)(n)]}}function o(t){return function(e,n){var i=Object(r.sqrt)(e*e+n*n),o=t(i),a=Object(r.sin)(o),s=Object(r.cos)(o);return[Object(r.atan2)(e*a,i*s),Object(r.asin)(i&&n*a/i)]}}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/azimuthalEqualArea.js":function(t,e,n){"use strict";n.r(e),n.d(e,"azimuthalEqualAreaRaw",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/azimuthal.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/index.js"),a=Object(i.azimuthalRaw)(function(t){return Object(r.sqrt)(2/(1+t))});a.invert=Object(i.azimuthalInvert)(function(t){return 2*Object(r.asin)(t/2)}),e.default=function(){return Object(o.default)(a).scale(124.75).clipAngle(179.999)}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/azimuthalEquidistant.js":function(t,e,n){"use strict";n.r(e),n.d(e,"azimuthalEquidistantRaw",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/azimuthal.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/index.js"),a=Object(i.azimuthalRaw)(function(t){return(t=Object(r.acos)(t))&&t/Object(r.sin)(t)});a.invert=Object(i.azimuthalInvert)(function(t){return t}),e.default=function(){return Object(o.default)(a).scale(79.4188).clipAngle(179.999)}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/conic.js":function(t,e,n){"use strict";n.r(e),n.d(e,"conicProjection",function(){return o});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/index.js");function o(t){var e=0,n=r.pi/3,o=Object(i.projectionMutator)(t),a=o(e,n);return a.parallels=function(t){return arguments.length?o(e=t[0]*r.radians,n=t[1]*r.radians):[e*r.degrees,n*r.degrees]},a}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/conicConformal.js":function(t,e,n){"use strict";n.r(e),n.d(e,"conicConformalRaw",function(){return s});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/conic.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/mercator.js");function a(t){return Object(r.tan)((r.halfPi+t)/2)}function s(t,e){var n=Object(r.cos)(t),i=t===e?Object(r.sin)(t):Object(r.log)(n/Object(r.cos)(e))/Object(r.log)(a(e)/a(t)),s=n*Object(r.pow)(a(t),i)/i;if(!i)return o.mercatorRaw;function u(t,e){s>0?e<-r.halfPi+r.epsilon&&(e=-r.halfPi+r.epsilon):e>r.halfPi-r.epsilon&&(e=r.halfPi-r.epsilon);var n=s/Object(r.pow)(a(e),i);return[n*Object(r.sin)(i*t),s-n*Object(r.cos)(i*t)]}return u.invert=function(t,e){var n=s-e,o=Object(r.sign)(i)*Object(r.sqrt)(t*t+n*n);return[Object(r.atan2)(t,Object(r.abs)(n))/i*Object(r.sign)(n),2*Object(r.atan)(Object(r.pow)(s/o,1/i))-r.halfPi]},u}e.default=function(){return Object(i.conicProjection)(s).scale(109.5).parallels([30,30])}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/conicEqualArea.js":function(t,e,n){"use strict";n.r(e),n.d(e,"conicEqualAreaRaw",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/conic.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/cylindricalEqualArea.js");function a(t,e){var n=Object(r.sin)(t),i=(n+Object(r.sin)(e))/2;if(Object(r.abs)(i)<r.epsilon)return Object(o.cylindricalEqualAreaRaw)(t);var a=1+n*(2*i-n),s=Object(r.sqrt)(a)/i;function u(t,e){var n=Object(r.sqrt)(a-2*i*Object(r.sin)(e))/i;return[n*Object(r.sin)(t*=i),s-n*Object(r.cos)(t)]}return u.invert=function(t,e){var n=s-e;return[Object(r.atan2)(t,Object(r.abs)(n))/i*Object(r.sign)(n),Object(r.asin)((a-(t*t+n*n)*i*i)/(2*i))]},u}e.default=function(){return Object(i.conicProjection)(a).scale(155.424).center([0,33.6442])}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/conicEquidistant.js":function(t,e,n){"use strict";n.r(e),n.d(e,"conicEquidistantRaw",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/conic.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/equirectangular.js");function a(t,e){var n=Object(r.cos)(t),i=t===e?Object(r.sin)(t):(n-Object(r.cos)(e))/(e-t),a=n/i+t;if(Object(r.abs)(i)<r.epsilon)return o.equirectangularRaw;function s(t,e){var n=a-e,o=i*t;return[n*Object(r.sin)(o),a-n*Object(r.cos)(o)]}return s.invert=function(t,e){var n=a-e;return[Object(r.atan2)(t,Object(r.abs)(n))/i*Object(r.sign)(n),a-Object(r.sign)(i)*Object(r.sqrt)(t*t+n*n)]},s}e.default=function(){return Object(i.conicProjection)(a).scale(131.154).center([0,13.9389])}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/cylindricalEqualArea.js":function(t,e,n){"use strict";n.r(e),n.d(e,"cylindricalEqualAreaRaw",function(){return i});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js");function i(t){var e=Object(r.cos)(t);function n(t,n){return[t*e,Object(r.sin)(n)/e]}return n.invert=function(t,n){return[t/e,Object(r.asin)(n*e)]},n}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/equalEarth.js":function(t,e,n){"use strict";n.r(e),n.d(e,"equalEarthRaw",function(){return l});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/index.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),o=1.340264,a=-.081106,s=893e-6,u=.003796,c=Object(i.sqrt)(3)/2;function l(t,e){var n=Object(i.asin)(c*Object(i.sin)(e)),r=n*n,l=r*r*r;return[t*Object(i.cos)(n)/(c*(o+3*a*r+l*(7*s+9*u*r))),n*(o+a*r+l*(s+u*r))]}l.invert=function(t,e){for(var n,r=e,l=r*r,d=l*l*l,p=0;p<12&&(d=(l=(r-=n=(r*(o+a*l+d*(s+u*l))-e)/(o+3*a*l+d*(7*s+9*u*l)))*r)*l*l,!(Object(i.abs)(n)<i.epsilon2));++p);return[c*t*(o+3*a*l+d*(7*s+9*u*l))/Object(i.cos)(r),Object(i.asin)(Object(i.sin)(r)/c)]},e.default=function(){return Object(r.default)(l).scale(177.158)}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/equirectangular.js":function(t,e,n){"use strict";n.r(e),n.d(e,"equirectangularRaw",function(){return i});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/index.js");function i(t,e){return[t,e]}i.invert=i,e.default=function(){return Object(r.default)(i).scale(152.63)}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/fit.js":function(t,e,n){"use strict";n.r(e),n.d(e,"fitExtent",function(){return a}),n.d(e,"fitSize",function(){return s}),n.d(e,"fitWidth",function(){return u}),n.d(e,"fitHeight",function(){return c});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/stream.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/path/bounds.js");function o(t,e,n){var o=t.clipExtent&&t.clipExtent();return t.scale(150).translate([0,0]),null!=o&&t.clipExtent(null),Object(r.default)(n,t.stream(i.default)),e(i.default.result()),null!=o&&t.clipExtent(o),t}function a(t,e,n){return o(t,function(n){var r=e[1][0]-e[0][0],i=e[1][1]-e[0][1],o=Math.min(r/(n[1][0]-n[0][0]),i/(n[1][1]-n[0][1])),a=+e[0][0]+(r-o*(n[1][0]+n[0][0]))/2,s=+e[0][1]+(i-o*(n[1][1]+n[0][1]))/2;t.scale(150*o).translate([a,s])},n)}function s(t,e,n){return a(t,[[0,0],e],n)}function u(t,e,n){return o(t,function(n){var r=+e,i=r/(n[1][0]-n[0][0]),o=(r-i*(n[1][0]+n[0][0]))/2,a=-i*n[0][1];t.scale(150*i).translate([o,a])},n)}function c(t,e,n){return o(t,function(n){var r=+e,i=r/(n[1][1]-n[0][1]),o=-i*n[0][0],a=(r-i*(n[1][1]+n[0][1]))/2;t.scale(150*i).translate([o,a])},n)}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/gnomonic.js":function(t,e,n){"use strict";n.r(e),n.d(e,"gnomonicRaw",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/azimuthal.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/index.js");function a(t,e){var n=Object(r.cos)(e),i=Object(r.cos)(t)*n;return[n*Object(r.sin)(t)/i,Object(r.sin)(e)/i]}a.invert=Object(i.azimuthalInvert)(r.atan),e.default=function(){return Object(o.default)(a).scale(144.049).clipAngle(60)}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/identity.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/clip/rectangle.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/identity.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/transform.js"),a=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/fit.js");function s(t,e,n,r){return 1===t&&1===e&&0===n&&0===r?i.default:Object(o.transformer)({point:function(i,o){this.stream.point(i*t+n,o*e+r)}})}e.default=function(){var t,e,n,o,u,c,l=1,d=0,p=0,f=1,h=1,m=i.default,g=null,y=i.default;function _(){return o=u=null,c}return c={stream:function(t){return o&&u===t?o:o=m(y(u=t))},postclip:function(r){return arguments.length?(y=r,g=t=e=n=null,_()):y},clipExtent:function(o){return arguments.length?(y=null==o?(g=t=e=n=null,i.default):Object(r.default)(g=+o[0][0],t=+o[0][1],e=+o[1][0],n=+o[1][1]),_()):null==g?null:[[g,t],[e,n]]},scale:function(t){return arguments.length?(m=s((l=+t)*f,l*h,d,p),_()):l},translate:function(t){return arguments.length?(m=s(l*f,l*h,d=+t[0],p=+t[1]),_()):[d,p]},reflectX:function(t){return arguments.length?(m=s(l*(f=t?-1:1),l*h,d,p),_()):f<0},reflectY:function(t){return arguments.length?(m=s(l*f,l*(h=t?-1:1),d,p),_()):h<0},fitExtent:function(t,e){return Object(a.fitExtent)(c,t,e)},fitSize:function(t,e){return Object(a.fitSize)(c,t,e)},fitWidth:function(t,e){return Object(a.fitWidth)(c,t,e)},fitHeight:function(t,e){return Object(a.fitHeight)(c,t,e)}}}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/index.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return m}),n.d(e,"projectionMutator",function(){return g});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/clip/antimeridian.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/clip/circle.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/clip/rectangle.js"),a=n("./node_modules/taucharts/node_modules/d3-geo/src/compose.js"),s=n("./node_modules/taucharts/node_modules/d3-geo/src/identity.js"),u=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),c=n("./node_modules/taucharts/node_modules/d3-geo/src/rotation.js"),l=n("./node_modules/taucharts/node_modules/d3-geo/src/transform.js"),d=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/fit.js"),p=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/resample.js"),f=Object(l.transformer)({point:function(t,e){this.stream.point(t*u.radians,e*u.radians)}});function h(t,e,n,r){var i=Object(u.cos)(r),o=Object(u.sin)(r),a=i*t,s=o*t,c=i/t,l=o/t,d=(o*n-i*e)/t,p=(o*e+i*n)/t;function f(t,r){return[a*t-s*r+e,n-s*t-a*r]}return f.invert=function(t,e){return[c*t-l*e+d,p-l*t-c*e]},f}function m(t){return g(function(){return t})()}function g(t){var e,n,m,g,y,_,v,b,x,w,j=150,O=480,C=250,E=0,S=0,M=0,T=0,k=0,A=0,D=null,I=r.default,P=null,R=s.default,N=.5;function F(t){return b(t[0]*u.radians,t[1]*u.radians)}function L(t){return(t=b.invert(t[0],t[1]))&&[t[0]*u.degrees,t[1]*u.degrees]}function B(){var t=h(j,0,0,A).apply(null,e(E,S)),r=(A?h:function(t,e,n){function r(r,i){return[e+t*r,n-t*i]}return r.invert=function(r,i){return[(r-e)/t,(n-i)/t]},r})(j,O-t[0],C-t[1],A);return n=Object(c.rotateRadians)(M,T,k),v=Object(a.default)(e,r),b=Object(a.default)(n,v),_=Object(p.default)(v,N),V()}function V(){return x=w=null,F}return F.stream=function(t){return x&&w===t?x:x=f((e=n,Object(l.transformer)({point:function(t,n){var r=e(t,n);return this.stream.point(r[0],r[1])}}))(I(_(R(w=t)))));var e},F.preclip=function(t){return arguments.length?(I=t,D=void 0,V()):I},F.postclip=function(t){return arguments.length?(R=t,P=m=g=y=null,V()):R},F.clipAngle=function(t){return arguments.length?(I=+t?Object(i.default)(D=t*u.radians):(D=null,r.default),V()):D*u.degrees},F.clipExtent=function(t){return arguments.length?(R=null==t?(P=m=g=y=null,s.default):Object(o.default)(P=+t[0][0],m=+t[0][1],g=+t[1][0],y=+t[1][1]),V()):null==P?null:[[P,m],[g,y]]},F.scale=function(t){return arguments.length?(j=+t,B()):j},F.translate=function(t){return arguments.length?(O=+t[0],C=+t[1],B()):[O,C]},F.center=function(t){return arguments.length?(E=t[0]%360*u.radians,S=t[1]%360*u.radians,B()):[E*u.degrees,S*u.degrees]},F.rotate=function(t){return arguments.length?(M=t[0]%360*u.radians,T=t[1]%360*u.radians,k=t.length>2?t[2]%360*u.radians:0,B()):[M*u.degrees,T*u.degrees,k*u.degrees]},F.angle=function(t){return arguments.length?(A=t%360*u.radians,B()):A*u.degrees},F.precision=function(t){return arguments.length?(_=Object(p.default)(v,N=t*t),V()):Object(u.sqrt)(N)},F.fitExtent=function(t,e){return Object(d.fitExtent)(F,t,e)},F.fitSize=function(t,e){return Object(d.fitSize)(F,t,e)},F.fitWidth=function(t,e){return Object(d.fitWidth)(F,t,e)},F.fitHeight=function(t,e){return Object(d.fitHeight)(F,t,e)},function(){return e=t.apply(this,arguments),F.invert=e.invert&&L,B()}}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/mercator.js":function(t,e,n){"use strict";n.r(e),n.d(e,"mercatorRaw",function(){return a}),n.d(e,"mercatorProjection",function(){return s});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/rotation.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/index.js");function a(t,e){return[t,Object(r.log)(Object(r.tan)((r.halfPi+e)/2))]}function s(t){var e,n,s,u=Object(o.default)(t),c=u.center,l=u.scale,d=u.translate,p=u.clipExtent,f=null;function h(){var o=r.pi*l(),c=u(Object(i.default)(u.rotate()).invert([0,0]));return p(null==f?[[c[0]-o,c[1]-o],[c[0]+o,c[1]+o]]:t===a?[[Math.max(c[0]-o,f),e],[Math.min(c[0]+o,n),s]]:[[f,Math.max(c[1]-o,e)],[n,Math.min(c[1]+o,s)]])}return u.scale=function(t){return arguments.length?(l(t),h()):l()},u.translate=function(t){return arguments.length?(d(t),h()):d()},u.center=function(t){return arguments.length?(c(t),h()):c()},u.clipExtent=function(t){return arguments.length?(null==t?f=e=n=s=null:(f=+t[0][0],e=+t[0][1],n=+t[1][0],s=+t[1][1]),h()):null==f?null:[[f,e],[n,s]]},h()}a.invert=function(t,e){return[t,2*Object(r.atan)(Object(r.exp)(e))-r.halfPi]},e.default=function(){return s(a).scale(961/r.tau)}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/naturalEarth1.js":function(t,e,n){"use strict";n.r(e),n.d(e,"naturalEarth1Raw",function(){return o});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/index.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js");function o(t,e){var n=e*e,r=n*n;return[t*(.8707-.131979*n+r*(r*(.003971*n-.001529*r)-.013791)),e*(1.007226+n*(.015085+r*(.028874*n-.044475-.005916*r)))]}o.invert=function(t,e){var n,r=e,o=25;do{var a=r*r,s=a*a;r-=n=(r*(1.007226+a*(.015085+s*(.028874*a-.044475-.005916*s)))-e)/(1.007226+a*(.045255+s*(.259866*a-.311325-.005916*11*s)))}while(Object(i.abs)(n)>i.epsilon&&--o>0);return[t/(.8707+(a=r*r)*(a*(a*a*a*(.003971-.001529*a)-.013791)-.131979)),r]},e.default=function(){return Object(r.default)(o).scale(175.295)}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/orthographic.js":function(t,e,n){"use strict";n.r(e),n.d(e,"orthographicRaw",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/azimuthal.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/index.js");function a(t,e){return[Object(r.cos)(e)*Object(r.sin)(t),Object(r.sin)(e)]}a.invert=Object(i.azimuthalInvert)(r.asin),e.default=function(){return Object(o.default)(a).scale(249.5).clipAngle(90+r.epsilon)}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/resample.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-geo/src/cartesian.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/transform.js"),a=16,s=Object(i.cos)(30*i.radians);e.default=function(t,e){return+e?function(t,e){function n(r,o,a,u,c,l,d,p,f,h,m,g,y,_){var v=d-r,b=p-o,x=v*v+b*b;if(x>4*e&&y--){var w=u+h,j=c+m,O=l+g,C=Object(i.sqrt)(w*w+j*j+O*O),E=Object(i.asin)(O/=C),S=Object(i.abs)(Object(i.abs)(O)-1)<i.epsilon||Object(i.abs)(a-f)<i.epsilon?(a+f)/2:Object(i.atan2)(j,w),M=t(S,E),T=M[0],k=M[1],A=T-r,D=k-o,I=b*A-v*D;(I*I/x>e||Object(i.abs)((v*A+b*D)/x-.5)>.3||u*h+c*m+l*g<s)&&(n(r,o,a,u,c,l,T,k,S,w/=C,j/=C,O,y,_),_.point(T,k),n(T,k,S,w,j,O,d,p,f,h,m,g,y,_))}}return function(e){var i,o,s,u,c,l,d,p,f,h,m,g,y={point:_,lineStart:v,lineEnd:x,polygonStart:function(){e.polygonStart(),y.lineStart=w},polygonEnd:function(){e.polygonEnd(),y.lineStart=v}};function _(n,r){n=t(n,r),e.point(n[0],n[1])}function v(){p=NaN,y.point=b,e.lineStart()}function b(i,o){var s=Object(r.cartesian)([i,o]),u=t(i,o);n(p,f,d,h,m,g,p=u[0],f=u[1],d=i,h=s[0],m=s[1],g=s[2],a,e),e.point(p,f)}function x(){y.point=_,e.lineEnd()}function w(){v(),y.point=j,y.lineEnd=O}function j(t,e){b(i=t,e),o=p,s=f,u=h,c=m,l=g,y.point=b}function O(){n(p,f,d,h,m,g,o,s,i,u,c,l,a,e),y.lineEnd=x,x()}return y}}(t,e):(n=t,Object(o.transformer)({point:function(t,e){t=n(t,e),this.stream.point(t[0],t[1])}}));var n}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/stereographic.js":function(t,e,n){"use strict";n.r(e),n.d(e,"stereographicRaw",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/azimuthal.js"),o=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/index.js");function a(t,e){var n=Object(r.cos)(e),i=1+Object(r.cos)(t)*n;return[n*Object(r.sin)(t)/i,Object(r.sin)(e)/i]}a.invert=Object(i.azimuthalInvert)(function(t){return 2*Object(r.atan)(t)}),e.default=function(){return Object(o.default)(a).scale(250).clipAngle(142)}},"./node_modules/taucharts/node_modules/d3-geo/src/projection/transverseMercator.js":function(t,e,n){"use strict";n.r(e),n.d(e,"transverseMercatorRaw",function(){return o});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/projection/mercator.js");function o(t,e){return[Object(r.log)(Object(r.tan)((r.halfPi+e)/2)),-t]}o.invert=function(t,e){return[-e,2*Object(r.atan)(Object(r.exp)(t))-r.halfPi]},e.default=function(){var t=Object(i.mercatorProjection)(o),e=t.center,n=t.rotate;return t.center=function(t){return arguments.length?e([-t[1],t[0]]):[(t=e())[1],-t[0]]},t.rotate=function(t){return arguments.length?n([t[0],t[1],t.length>2?t[2]+90:90]):[(t=n())[0],t[1],t[2]-90]},n([0,0,90]).scale(159.155)}},"./node_modules/taucharts/node_modules/d3-geo/src/rotation.js":function(t,e,n){"use strict";n.r(e),n.d(e,"rotateRadians",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-geo/src/compose.js"),i=n("./node_modules/taucharts/node_modules/d3-geo/src/math.js");function o(t,e){return[Object(i.abs)(t)>i.pi?t+Math.round(-t/i.tau)*i.tau:t,e]}function a(t,e,n){return(t%=i.tau)?e||n?Object(r.default)(u(t),c(e,n)):u(t):e||n?c(e,n):o}function s(t){return function(e,n){return[(e+=t)>i.pi?e-i.tau:e<-i.pi?e+i.tau:e,n]}}function u(t){var e=s(t);return e.invert=s(-t),e}function c(t,e){var n=Object(i.cos)(t),r=Object(i.sin)(t),o=Object(i.cos)(e),a=Object(i.sin)(e);function s(t,e){var s=Object(i.cos)(e),u=Object(i.cos)(t)*s,c=Object(i.sin)(t)*s,l=Object(i.sin)(e),d=l*n+u*r;return[Object(i.atan2)(c*o-d*a,u*n-l*r),Object(i.asin)(d*o+c*a)]}return s.invert=function(t,e){var s=Object(i.cos)(e),u=Object(i.cos)(t)*s,c=Object(i.sin)(t)*s,l=Object(i.sin)(e),d=l*o-c*a;return[Object(i.atan2)(c*o+l*a,u*n+d*r),Object(i.asin)(d*n-u*r)]},s}o.invert=o,e.default=function(t){function e(e){return(e=t(e[0]*i.radians,e[1]*i.radians))[0]*=i.degrees,e[1]*=i.degrees,e}return t=a(t[0]*i.radians,t[1]*i.radians,t.length>2?t[2]*i.radians:0),e.invert=function(e){return(e=t.invert(e[0]*i.radians,e[1]*i.radians))[0]*=i.degrees,e[1]*=i.degrees,e},e}},"./node_modules/taucharts/node_modules/d3-geo/src/stream.js":function(t,e,n){"use strict";function r(t,e){t&&o.hasOwnProperty(t.type)&&o[t.type](t,e)}n.r(e);var i={Feature:function(t,e){r(t.geometry,e)},FeatureCollection:function(t,e){for(var n=t.features,i=-1,o=n.length;++i<o;)r(n[i].geometry,e)}},o={Sphere:function(t,e){e.sphere()},Point:function(t,e){t=t.coordinates,e.point(t[0],t[1],t[2])},MultiPoint:function(t,e){for(var n=t.coordinates,r=-1,i=n.length;++r<i;)t=n[r],e.point(t[0],t[1],t[2])},LineString:function(t,e){a(t.coordinates,e,0)},MultiLineString:function(t,e){for(var n=t.coordinates,r=-1,i=n.length;++r<i;)a(n[r],e,0)},Polygon:function(t,e){s(t.coordinates,e)},MultiPolygon:function(t,e){for(var n=t.coordinates,r=-1,i=n.length;++r<i;)s(n[r],e)},GeometryCollection:function(t,e){for(var n=t.geometries,i=-1,o=n.length;++i<o;)r(n[i],e)}};function a(t,e,n){var r,i=-1,o=t.length-n;for(e.lineStart();++i<o;)r=t[i],e.point(r[0],r[1],r[2]);e.lineEnd()}function s(t,e){var n=-1,r=t.length;for(e.polygonStart();++n<r;)a(t[n],e,1);e.polygonEnd()}e.default=function(t,e){t&&i.hasOwnProperty(t.type)?i[t.type](t,e):r(t,e)}},"./node_modules/taucharts/node_modules/d3-geo/src/transform.js":function(t,e,n){"use strict";function r(t){return function(e){var n=new i;for(var r in t)n[r]=t[r];return n.stream=e,n}}function i(){}n.r(e),n.d(e,"transformer",function(){return r}),e.default=function(t){return{stream:r(t)}},i.prototype={constructor:i,point:function(t,e){this.stream.point(t,e)},sphere:function(){this.stream.sphere()},lineStart:function(){this.stream.lineStart()},lineEnd:function(){this.stream.lineEnd()},polygonStart:function(){this.stream.polygonStart()},polygonEnd:function(){this.stream.polygonEnd()}}},"./node_modules/taucharts/node_modules/d3-quadtree/src/add.js":function(t,e,n){"use strict";function r(t,e,n,r){if(isNaN(e)||isNaN(n))return t;var i,o,a,s,u,c,l,d,p,f=t._root,h={data:r},m=t._x0,g=t._y0,y=t._x1,_=t._y1;if(!f)return t._root=h,t;for(;f.length;)if((c=e>=(o=(m+y)/2))?m=o:y=o,(l=n>=(a=(g+_)/2))?g=a:_=a,i=f,!(f=f[d=l<<1|c]))return i[d]=h,t;if(s=+t._x.call(null,f.data),u=+t._y.call(null,f.data),e===s&&n===u)return h.next=f,i?i[d]=h:t._root=h,t;do{i=i?i[d]=new Array(4):t._root=new Array(4),(c=e>=(o=(m+y)/2))?m=o:y=o,(l=n>=(a=(g+_)/2))?g=a:_=a}while((d=l<<1|c)==(p=(u>=a)<<1|s>=o));return i[p]=f,i[d]=h,t}function i(t){var e,n,i,o,a=t.length,s=new Array(a),u=new Array(a),c=1/0,l=1/0,d=-1/0,p=-1/0;for(n=0;n<a;++n)isNaN(i=+this._x.call(null,e=t[n]))||isNaN(o=+this._y.call(null,e))||(s[n]=i,u[n]=o,i<c&&(c=i),i>d&&(d=i),o<l&&(l=o),o>p&&(p=o));if(c>d||l>p)return this;for(this.cover(c,l).cover(d,p),n=0;n<a;++n)r(this,s[n],u[n],t[n]);return this}n.r(e),n.d(e,"addAll",function(){return i}),e.default=function(t){var e=+this._x.call(null,t),n=+this._y.call(null,t);return r(this.cover(e,n),e,n,t)}},"./node_modules/taucharts/node_modules/d3-quadtree/src/cover.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){if(isNaN(t=+t)||isNaN(e=+e))return this;var n=this._x0,r=this._y0,i=this._x1,o=this._y1;if(isNaN(n))i=(n=Math.floor(t))+1,o=(r=Math.floor(e))+1;else{for(var a,s,u=i-n,c=this._root;n>t||t>=i||r>e||e>=o;)switch(s=(e<r)<<1|t<n,(a=new Array(4))[s]=c,c=a,u*=2,s){case 0:i=n+u,o=r+u;break;case 1:n=i-u,o=r+u;break;case 2:i=n+u,r=o-u;break;case 3:n=i-u,r=o-u}this._root&&this._root.length&&(this._root=c)}return this._x0=n,this._y0=r,this._x1=i,this._y1=o,this}},"./node_modules/taucharts/node_modules/d3-quadtree/src/data.js":function(t,e,n){"use strict";n.r(e),e.default=function(){var t=[];return this.visit(function(e){if(!e.length)do{t.push(e.data)}while(e=e.next)}),t}},"./node_modules/taucharts/node_modules/d3-quadtree/src/extent.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return arguments.length?this.cover(+t[0][0],+t[0][1]).cover(+t[1][0],+t[1][1]):isNaN(this._x0)?void 0:[[this._x0,this._y0],[this._x1,this._y1]]}},"./node_modules/taucharts/node_modules/d3-quadtree/src/find.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-quadtree/src/quad.js");e.default=function(t,e,n){var i,o,a,s,u,c,l,d=this._x0,p=this._y0,f=this._x1,h=this._y1,m=[],g=this._root;for(g&&m.push(new r.default(g,d,p,f,h)),null==n?n=1/0:(d=t-n,p=e-n,f=t+n,h=e+n,n*=n);c=m.pop();)if(!(!(g=c.node)||(o=c.x0)>f||(a=c.y0)>h||(s=c.x1)<d||(u=c.y1)<p))if(g.length){var y=(o+s)/2,_=(a+u)/2;m.push(new r.default(g[3],y,_,s,u),new r.default(g[2],o,_,y,u),new r.default(g[1],y,a,s,_),new r.default(g[0],o,a,y,_)),(l=(e>=_)<<1|t>=y)&&(c=m[m.length-1],m[m.length-1]=m[m.length-1-l],m[m.length-1-l]=c)}else{var v=t-+this._x.call(null,g.data),b=e-+this._y.call(null,g.data),x=v*v+b*b;if(x<n){var w=Math.sqrt(n=x);d=t-w,p=e-w,f=t+w,h=e+w,i=g.data}}return i}},"./node_modules/taucharts/node_modules/d3-quadtree/src/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-quadtree/src/quadtree.js");n.d(e,"quadtree",function(){return r.default})},"./node_modules/taucharts/node_modules/d3-quadtree/src/quad.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e,n,r,i){this.node=t,this.x0=e,this.y0=n,this.x1=r,this.y1=i}},"./node_modules/taucharts/node_modules/d3-quadtree/src/quadtree.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return m});var r=n("./node_modules/taucharts/node_modules/d3-quadtree/src/add.js"),i=n("./node_modules/taucharts/node_modules/d3-quadtree/src/cover.js"),o=n("./node_modules/taucharts/node_modules/d3-quadtree/src/data.js"),a=n("./node_modules/taucharts/node_modules/d3-quadtree/src/extent.js"),s=n("./node_modules/taucharts/node_modules/d3-quadtree/src/find.js"),u=n("./node_modules/taucharts/node_modules/d3-quadtree/src/remove.js"),c=n("./node_modules/taucharts/node_modules/d3-quadtree/src/root.js"),l=n("./node_modules/taucharts/node_modules/d3-quadtree/src/size.js"),d=n("./node_modules/taucharts/node_modules/d3-quadtree/src/visit.js"),p=n("./node_modules/taucharts/node_modules/d3-quadtree/src/visitAfter.js"),f=n("./node_modules/taucharts/node_modules/d3-quadtree/src/x.js"),h=n("./node_modules/taucharts/node_modules/d3-quadtree/src/y.js");function m(t,e,n){var r=new g(null==e?f.defaultX:e,null==n?h.defaultY:n,NaN,NaN,NaN,NaN);return null==t?r:r.addAll(t)}function g(t,e,n,r,i,o){this._x=t,this._y=e,this._x0=n,this._y0=r,this._x1=i,this._y1=o,this._root=void 0}function y(t){for(var e={data:t.data},n=e;t=t.next;)n=n.next={data:t.data};return e}var _=m.prototype=g.prototype;_.copy=function(){var t,e,n=new g(this._x,this._y,this._x0,this._y0,this._x1,this._y1),r=this._root;if(!r)return n;if(!r.length)return n._root=y(r),n;for(t=[{source:r,target:n._root=new Array(4)}];r=t.pop();)for(var i=0;i<4;++i)(e=r.source[i])&&(e.length?t.push({source:e,target:r.target[i]=new Array(4)}):r.target[i]=y(e));return n},_.add=r.default,_.addAll=r.addAll,_.cover=i.default,_.data=o.default,_.extent=a.default,_.find=s.default,_.remove=u.default,_.removeAll=u.removeAll,_.root=c.default,_.size=l.default,_.visit=d.default,_.visitAfter=p.default,_.x=f.default,_.y=h.default},"./node_modules/taucharts/node_modules/d3-quadtree/src/remove.js":function(t,e,n){"use strict";function r(t){for(var e=0,n=t.length;e<n;++e)this.remove(t[e]);return this}n.r(e),n.d(e,"removeAll",function(){return r}),e.default=function(t){if(isNaN(o=+this._x.call(null,t))||isNaN(a=+this._y.call(null,t)))return this;var e,n,r,i,o,a,s,u,c,l,d,p,f=this._root,h=this._x0,m=this._y0,g=this._x1,y=this._y1;if(!f)return this;if(f.length)for(;;){if((c=o>=(s=(h+g)/2))?h=s:g=s,(l=a>=(u=(m+y)/2))?m=u:y=u,e=f,!(f=f[d=l<<1|c]))return this;if(!f.length)break;(e[d+1&3]||e[d+2&3]||e[d+3&3])&&(n=e,p=d)}for(;f.data!==t;)if(r=f,!(f=f.next))return this;return(i=f.next)&&delete f.next,r?(i?r.next=i:delete r.next,this):e?(i?e[d]=i:delete e[d],(f=e[0]||e[1]||e[2]||e[3])&&f===(e[3]||e[2]||e[1]||e[0])&&!f.length&&(n?n[p]=f:this._root=f),this):(this._root=i,this)}},"./node_modules/taucharts/node_modules/d3-quadtree/src/root.js":function(t,e,n){"use strict";n.r(e),e.default=function(){return this._root}},"./node_modules/taucharts/node_modules/d3-quadtree/src/size.js":function(t,e,n){"use strict";n.r(e),e.default=function(){var t=0;return this.visit(function(e){if(!e.length)do{++t}while(e=e.next)}),t}},"./node_modules/taucharts/node_modules/d3-quadtree/src/visit.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-quadtree/src/quad.js");e.default=function(t){var e,n,i,o,a,s,u=[],c=this._root;for(c&&u.push(new r.default(c,this._x0,this._y0,this._x1,this._y1));e=u.pop();)if(!t(c=e.node,i=e.x0,o=e.y0,a=e.x1,s=e.y1)&&c.length){var l=(i+a)/2,d=(o+s)/2;(n=c[3])&&u.push(new r.default(n,l,d,a,s)),(n=c[2])&&u.push(new r.default(n,i,d,l,s)),(n=c[1])&&u.push(new r.default(n,l,o,a,d)),(n=c[0])&&u.push(new r.default(n,i,o,l,d))}return this}},"./node_modules/taucharts/node_modules/d3-quadtree/src/visitAfter.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-quadtree/src/quad.js");e.default=function(t){var e,n=[],i=[];for(this._root&&n.push(new r.default(this._root,this._x0,this._y0,this._x1,this._y1));e=n.pop();){var o=e.node;if(o.length){var a,s=e.x0,u=e.y0,c=e.x1,l=e.y1,d=(s+c)/2,p=(u+l)/2;(a=o[0])&&n.push(new r.default(a,s,u,d,p)),(a=o[1])&&n.push(new r.default(a,d,u,c,p)),(a=o[2])&&n.push(new r.default(a,s,p,d,l)),(a=o[3])&&n.push(new r.default(a,d,p,c,l))}i.push(e)}for(;e=i.pop();)t(e.node,e.x0,e.y0,e.x1,e.y1);return this}},"./node_modules/taucharts/node_modules/d3-quadtree/src/x.js":function(t,e,n){"use strict";function r(t){return t[0]}n.r(e),n.d(e,"defaultX",function(){return r}),e.default=function(t){return arguments.length?(this._x=t,this):this._x}},"./node_modules/taucharts/node_modules/d3-quadtree/src/y.js":function(t,e,n){"use strict";function r(t){return t[1]}n.r(e),n.d(e,"defaultY",function(){return r}),e.default=function(t){return arguments.length?(this._y=t,this):this._y}},"./node_modules/taucharts/node_modules/d3-scale/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-scale/src/band.js");n.d(e,"scaleBand",function(){return r.default}),n.d(e,"scalePoint",function(){return r.point});var i=n("./node_modules/taucharts/node_modules/d3-scale/src/identity.js");n.d(e,"scaleIdentity",function(){return i.default});var o=n("./node_modules/taucharts/node_modules/d3-scale/src/linear.js");n.d(e,"scaleLinear",function(){return o.default});var a=n("./node_modules/taucharts/node_modules/d3-scale/src/log.js");n.d(e,"scaleLog",function(){return a.default});var s=n("./node_modules/taucharts/node_modules/d3-scale/src/ordinal.js");n.d(e,"scaleOrdinal",function(){return s.default}),n.d(e,"scaleImplicit",function(){return s.implicit});var u=n("./node_modules/taucharts/node_modules/d3-scale/src/pow.js");n.d(e,"scalePow",function(){return u.default}),n.d(e,"scaleSqrt",function(){return u.sqrt});var c=n("./node_modules/taucharts/node_modules/d3-scale/src/quantile.js");n.d(e,"scaleQuantile",function(){return c.default});var l=n("./node_modules/taucharts/node_modules/d3-scale/src/quantize.js");n.d(e,"scaleQuantize",function(){return l.default});var d=n("./node_modules/taucharts/node_modules/d3-scale/src/threshold.js");n.d(e,"scaleThreshold",function(){return d.default});var p=n("./node_modules/taucharts/node_modules/d3-scale/src/time.js");n.d(e,"scaleTime",function(){return p.default});var f=n("./node_modules/taucharts/node_modules/d3-scale/src/utcTime.js");n.d(e,"scaleUtc",function(){return f.default});var h=n("./node_modules/taucharts/node_modules/d3-scale/src/category10.js");n.d(e,"schemeCategory10",function(){return h.default});var m=n("./node_modules/taucharts/node_modules/d3-scale/src/category20b.js");n.d(e,"schemeCategory20b",function(){return m.default});var g=n("./node_modules/taucharts/node_modules/d3-scale/src/category20c.js");n.d(e,"schemeCategory20c",function(){return g.default});var y=n("./node_modules/taucharts/node_modules/d3-scale/src/category20.js");n.d(e,"schemeCategory20",function(){return y.default});var _=n("./node_modules/taucharts/node_modules/d3-scale/src/cubehelix.js");n.d(e,"interpolateCubehelixDefault",function(){return _.default});var v=n("./node_modules/taucharts/node_modules/d3-scale/src/rainbow.js");n.d(e,"interpolateRainbow",function(){return v.default}),n.d(e,"interpolateWarm",function(){return v.warm}),n.d(e,"interpolateCool",function(){return v.cool});var b=n("./node_modules/taucharts/node_modules/d3-scale/src/viridis.js");n.d(e,"interpolateViridis",function(){return b.default}),n.d(e,"interpolateMagma",function(){return b.magma}),n.d(e,"interpolateInferno",function(){return b.inferno}),n.d(e,"interpolatePlasma",function(){return b.plasma});var x=n("./node_modules/taucharts/node_modules/d3-scale/src/sequential.js");n.d(e,"scaleSequential",function(){return x.default})},"./node_modules/taucharts/node_modules/d3-scale/src/array.js":function(t,e,n){"use strict";n.r(e),n.d(e,"map",function(){return i}),n.d(e,"slice",function(){return o});var r=Array.prototype,i=r.map,o=r.slice},"./node_modules/taucharts/node_modules/d3-scale/src/band.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return o}),n.d(e,"point",function(){return a});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/taucharts/node_modules/d3-scale/src/ordinal.js");function o(){var t,e,n=Object(i.default)().unknown(void 0),a=n.domain,s=n.range,u=[0,1],c=!1,l=0,d=0,p=.5;function f(){var n=a().length,i=u[1]<u[0],o=u[i-0],f=u[1-i];t=(f-o)/Math.max(1,n-l+2*d),c&&(t=Math.floor(t)),o+=(f-o-t*(n-l))*p,e=t*(1-l),c&&(o=Math.round(o),e=Math.round(e));var h=Object(r.range)(n).map(function(e){return o+t*e});return s(i?h.reverse():h)}return delete n.unknown,n.domain=function(t){return arguments.length?(a(t),f()):a()},n.range=function(t){return arguments.length?(u=[+t[0],+t[1]],f()):u.slice()},n.rangeRound=function(t){return u=[+t[0],+t[1]],c=!0,f()},n.bandwidth=function(){return e},n.step=function(){return t},n.round=function(t){return arguments.length?(c=!!t,f()):c},n.padding=function(t){return arguments.length?(l=d=Math.max(0,Math.min(1,t)),f()):l},n.paddingInner=function(t){return arguments.length?(l=Math.max(0,Math.min(1,t)),f()):l},n.paddingOuter=function(t){return arguments.length?(d=Math.max(0,Math.min(1,t)),f()):d},n.align=function(t){return arguments.length?(p=Math.max(0,Math.min(1,t)),f()):p},n.copy=function(){return o().domain(a()).range(u).round(c).paddingInner(l).paddingOuter(d).align(p)},f()}function a(){return function t(e){var n=e.copy;return e.padding=e.paddingOuter,delete e.paddingInner,delete e.paddingOuter,e.copy=function(){return t(n())},e}(o().paddingInner(1))}},"./node_modules/taucharts/node_modules/d3-scale/src/category10.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-scale/src/colors.js");e.default=Object(r.default)("1f77b4ff7f0e2ca02cd627289467bd8c564be377c27f7f7fbcbd2217becf")},"./node_modules/taucharts/node_modules/d3-scale/src/category20.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-scale/src/colors.js");e.default=Object(r.default)("1f77b4aec7e8ff7f0effbb782ca02c98df8ad62728ff98969467bdc5b0d58c564bc49c94e377c2f7b6d27f7f7fc7c7c7bcbd22dbdb8d17becf9edae5")},"./node_modules/taucharts/node_modules/d3-scale/src/category20b.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-scale/src/colors.js");e.default=Object(r.default)("393b795254a36b6ecf9c9ede6379398ca252b5cf6bcedb9c8c6d31bd9e39e7ba52e7cb94843c39ad494ad6616be7969c7b4173a55194ce6dbdde9ed6")},"./node_modules/taucharts/node_modules/d3-scale/src/category20c.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-scale/src/colors.js");e.default=Object(r.default)("3182bd6baed69ecae1c6dbefe6550dfd8d3cfdae6bfdd0a231a35474c476a1d99bc7e9c0756bb19e9ac8bcbddcdadaeb636363969696bdbdbdd9d9d9")},"./node_modules/taucharts/node_modules/d3-scale/src/colors.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return t.match(/.{6}/g).map(function(t){return"#"+t})}},"./node_modules/taucharts/node_modules/d3-scale/src/constant.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return function(){return t}}},"./node_modules/taucharts/node_modules/d3-scale/src/continuous.js":function(t,e,n){"use strict";n.r(e),n.d(e,"deinterpolateLinear",function(){return c}),n.d(e,"copy",function(){return p}),n.d(e,"default",function(){return f});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/d3-interpolate/index.js"),o=n("./node_modules/taucharts/node_modules/d3-scale/src/array.js"),a=n("./node_modules/taucharts/node_modules/d3-scale/src/constant.js"),s=n("./node_modules/taucharts/node_modules/d3-scale/src/number.js"),u=[0,1];function c(t,e){return(e-=t=+t)?function(n){return(n-t)/e}:Object(a.default)(e)}function l(t,e,n,r){var i=t[0],o=t[1],a=e[0],s=e[1];return o<i?(i=n(o,i),a=r(s,a)):(i=n(i,o),a=r(a,s)),function(t){return a(i(t))}}function d(t,e,n,i){var o=Math.min(t.length,e.length)-1,a=new Array(o),s=new Array(o),u=-1;for(t[o]<t[0]&&(t=t.slice().reverse(),e=e.slice().reverse());++u<o;)a[u]=n(t[u],t[u+1]),s[u]=i(e[u],e[u+1]);return function(e){var n=Object(r.bisect)(t,e,1,o)-1;return s[n](a[n](e))}}function p(t,e){return e.domain(t.domain()).range(t.range()).interpolate(t.interpolate()).clamp(t.clamp())}function f(t,e){var n,r,a,p=u,f=u,h=i.interpolate,m=!1;function g(){return n=Math.min(p.length,f.length)>2?d:l,r=a=null,y}function y(e){return(r||(r=n(p,f,m?(i=t,function(t,e){var n=i(t=+t,e=+e);return function(r){return r<=t?0:r>=e?1:n(r)}}):t,h)))(+e);var i}return y.invert=function(t){return(a||(a=n(f,p,c,m?(r=e,function(t,e){var n=r(t=+t,e=+e);return function(r){return r<=0?t:r>=1?e:n(r)}}):e)))(+t);var r},y.domain=function(t){return arguments.length?(p=o.map.call(t,s.default),g()):p.slice()},y.range=function(t){return arguments.length?(f=o.slice.call(t),g()):f.slice()},y.rangeRound=function(t){return f=o.slice.call(t),h=i.interpolateRound,g()},y.clamp=function(t){return arguments.length?(m=!!t,g()):m},y.interpolate=function(t){return arguments.length?(h=t,g()):h},g()}},"./node_modules/taucharts/node_modules/d3-scale/src/cubehelix.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-color/src/index.js"),i=n("./node_modules/d3-interpolate/index.js");e.default=Object(i.interpolateCubehelixLong)(Object(r.cubehelix)(300,.5,0),Object(r.cubehelix)(-240,.5,1))},"./node_modules/taucharts/node_modules/d3-scale/src/identity.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-scale/src/array.js"),i=n("./node_modules/taucharts/node_modules/d3-scale/src/linear.js"),o=n("./node_modules/taucharts/node_modules/d3-scale/src/number.js");function a(){var t=[0,1];function e(t){return+t}return e.invert=e,e.domain=e.range=function(n){return arguments.length?(t=r.map.call(n,o.default),e):t.slice()},e.copy=function(){return a().domain(t)},Object(i.linearish)(e)}},"./node_modules/taucharts/node_modules/d3-scale/src/linear.js":function(t,e,n){"use strict";n.r(e),n.d(e,"linearish",function(){return s}),n.d(e,"default",function(){return u});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/d3-interpolate/index.js"),o=n("./node_modules/taucharts/node_modules/d3-scale/src/continuous.js"),a=n("./node_modules/taucharts/node_modules/d3-scale/src/tickFormat.js");function s(t){var e=t.domain;return t.ticks=function(t){var n=e();return Object(r.ticks)(n[0],n[n.length-1],null==t?10:t)},t.tickFormat=function(t,n){return Object(a.default)(e(),t,n)},t.nice=function(n){null==n&&(n=10);var i,o=e(),a=0,s=o.length-1,u=o[a],c=o[s];return c<u&&(i=u,u=c,c=i,i=a,a=s,s=i),(i=Object(r.tickIncrement)(u,c,n))>0?(u=Math.floor(u/i)*i,c=Math.ceil(c/i)*i,i=Object(r.tickIncrement)(u,c,n)):i<0&&(u=Math.ceil(u*i)/i,c=Math.floor(c*i)/i,i=Object(r.tickIncrement)(u,c,n)),i>0?(o[a]=Math.floor(u/i)*i,o[s]=Math.ceil(c/i)*i,e(o)):i<0&&(o[a]=Math.ceil(u*i)/i,o[s]=Math.floor(c*i)/i,e(o)),t},t}function u(){var t=Object(o.default)(o.deinterpolateLinear,i.interpolateNumber);return t.copy=function(){return Object(o.copy)(t,u())},s(t)}},"./node_modules/taucharts/node_modules/d3-scale/src/log.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return h});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/taucharts/node_modules/d3-format/src/index.js"),o=n("./node_modules/taucharts/node_modules/d3-scale/src/constant.js"),a=n("./node_modules/taucharts/node_modules/d3-scale/src/nice.js"),s=n("./node_modules/taucharts/node_modules/d3-scale/src/continuous.js");function u(t,e){return(e=Math.log(e/t))?function(n){return Math.log(n/t)/e}:Object(o.default)(e)}function c(t,e){return t<0?function(n){return-Math.pow(-e,n)*Math.pow(-t,1-n)}:function(n){return Math.pow(e,n)*Math.pow(t,1-n)}}function l(t){return isFinite(t)?+("1e"+t):t<0?0:t}function d(t){return 10===t?l:t===Math.E?Math.exp:function(e){return Math.pow(t,e)}}function p(t){return t===Math.E?Math.log:10===t&&Math.log10||2===t&&Math.log2||(t=Math.log(t),function(e){return Math.log(e)/t})}function f(t){return function(e){return-t(-e)}}function h(){var t=Object(s.default)(u,c).domain([1,10]),e=t.domain,n=10,o=p(10),l=d(10);function m(){return o=p(n),l=d(n),e()[0]<0&&(o=f(o),l=f(l)),t}return t.base=function(t){return arguments.length?(n=+t,m()):n},t.domain=function(t){return arguments.length?(e(t),m()):e()},t.ticks=function(t){var i,a=e(),s=a[0],u=a[a.length-1];(i=u<s)&&(f=s,s=u,u=f);var c,d,p,f=o(s),h=o(u),m=null==t?10:+t,g=[];if(!(n%1)&&h-f<m){if(f=Math.round(f)-1,h=Math.round(h)+1,s>0){for(;f<h;++f)for(d=1,c=l(f);d<n;++d)if(!((p=c*d)<s)){if(p>u)break;g.push(p)}}else for(;f<h;++f)for(d=n-1,c=l(f);d>=1;--d)if(!((p=c*d)<s)){if(p>u)break;g.push(p)}}else g=Object(r.ticks)(f,h,Math.min(h-f,m)).map(l);return i?g.reverse():g},t.tickFormat=function(e,r){if(null==r&&(r=10===n?".0e":","),"function"!=typeof r&&(r=Object(i.format)(r)),e===1/0)return r;null==e&&(e=10);var a=Math.max(1,n*e/t.ticks().length);return function(t){var e=t/l(Math.round(o(t)));return e*n<n-.5&&(e*=n),e<=a?r(t):""}},t.nice=function(){return e(Object(a.default)(e(),{floor:function(t){return l(Math.floor(o(t)))},ceil:function(t){return l(Math.ceil(o(t)))}}))},t.copy=function(){return Object(s.copy)(t,h().base(n))},t}},"./node_modules/taucharts/node_modules/d3-scale/src/nice.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){var n,r=0,i=(t=t.slice()).length-1,o=t[r],a=t[i];return a<o&&(n=r,r=i,i=n,n=o,o=a,a=n),t[r]=e.floor(o),t[i]=e.ceil(a),t}},"./node_modules/taucharts/node_modules/d3-scale/src/number.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return+t}},"./node_modules/taucharts/node_modules/d3-scale/src/ordinal.js":function(t,e,n){"use strict";n.r(e),n.d(e,"implicit",function(){return o}),n.d(e,"default",function(){return a});var r=n("./node_modules/d3-collection/index.js"),i=n("./node_modules/taucharts/node_modules/d3-scale/src/array.js"),o={name:"implicit"};function a(t){var e=Object(r.map)(),n=[],s=o;function u(r){var i=r+"",a=e.get(i);if(!a){if(s!==o)return s;e.set(i,a=n.push(r))}return t[(a-1)%t.length]}return t=null==t?[]:i.slice.call(t),u.domain=function(t){if(!arguments.length)return n.slice();n=[],e=Object(r.map)();for(var i,o,a=-1,s=t.length;++a<s;)e.has(o=(i=t[a])+"")||e.set(o,n.push(i));return u},u.range=function(e){return arguments.length?(t=i.slice.call(e),u):t.slice()},u.unknown=function(t){return arguments.length?(s=t,u):s},u.copy=function(){return a().domain(n).range(t).unknown(s)},u}},"./node_modules/taucharts/node_modules/d3-scale/src/pow.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return s}),n.d(e,"sqrt",function(){return u});var r=n("./node_modules/taucharts/node_modules/d3-scale/src/constant.js"),i=n("./node_modules/taucharts/node_modules/d3-scale/src/linear.js"),o=n("./node_modules/taucharts/node_modules/d3-scale/src/continuous.js");function a(t,e){return t<0?-Math.pow(-t,e):Math.pow(t,e)}function s(){var t=1,e=Object(o.default)(function(e,n){return(n=a(n,t)-(e=a(e,t)))?function(r){return(a(r,t)-e)/n}:Object(r.default)(n)},function(e,n){return n=a(n,t)-(e=a(e,t)),function(r){return a(e+n*r,1/t)}}),n=e.domain;return e.exponent=function(e){return arguments.length?(t=+e,n(n())):t},e.copy=function(){return Object(o.copy)(e,s().exponent(t))},Object(i.linearish)(e)}function u(){return s().exponent(.5)}},"./node_modules/taucharts/node_modules/d3-scale/src/quantile.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return o});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/taucharts/node_modules/d3-scale/src/array.js");function o(){var t=[],e=[],n=[];function a(){var i=0,o=Math.max(1,e.length);for(n=new Array(o-1);++i<o;)n[i-1]=Object(r.quantile)(t,i/o);return s}function s(t){if(!isNaN(t=+t))return e[Object(r.bisect)(n,t)]}return s.invertExtent=function(r){var i=e.indexOf(r);return i<0?[NaN,NaN]:[i>0?n[i-1]:t[0],i<n.length?n[i]:t[t.length-1]]},s.domain=function(e){if(!arguments.length)return t.slice();t=[];for(var n,i=0,o=e.length;i<o;++i)n=e[i],null==n||isNaN(n=+n)||t.push(n);return t.sort(r.ascending),a()},s.range=function(t){return arguments.length?(e=i.slice.call(t),a()):e.slice()},s.quantiles=function(){return n.slice()},s.copy=function(){return o().domain(t).range(e)},s}},"./node_modules/taucharts/node_modules/d3-scale/src/quantize.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return a});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/taucharts/node_modules/d3-scale/src/array.js"),o=n("./node_modules/taucharts/node_modules/d3-scale/src/linear.js");function a(){var t=0,e=1,n=1,s=[.5],u=[0,1];function c(t){if(t<=t)return u[Object(r.bisect)(s,t,0,n)]}function l(){var r=-1;for(s=new Array(n);++r<n;)s[r]=((r+1)*e-(r-n)*t)/(n+1);return c}return c.domain=function(n){return arguments.length?(t=+n[0],e=+n[1],l()):[t,e]},c.range=function(t){return arguments.length?(n=(u=i.slice.call(t)).length-1,l()):u.slice()},c.invertExtent=function(r){var i=u.indexOf(r);return i<0?[NaN,NaN]:i<1?[t,s[0]]:i>=n?[s[n-1],e]:[s[i-1],s[i]]},c.copy=function(){return a().domain([t,e]).range(u)},Object(o.linearish)(c)}},"./node_modules/taucharts/node_modules/d3-scale/src/rainbow.js":function(t,e,n){"use strict";n.r(e),n.d(e,"warm",function(){return o}),n.d(e,"cool",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-color/src/index.js"),i=n("./node_modules/d3-interpolate/index.js"),o=Object(i.interpolateCubehelixLong)(Object(r.cubehelix)(-100,.75,.35),Object(r.cubehelix)(80,1.5,.8)),a=Object(i.interpolateCubehelixLong)(Object(r.cubehelix)(260,.75,.35),Object(r.cubehelix)(80,1.5,.8)),s=Object(r.cubehelix)();e.default=function(t){(t<0||t>1)&&(t-=Math.floor(t));var e=Math.abs(t-.5);return s.h=360*t-100,s.s=1.5-1.5*e,s.l=.8-.9*e,s+""}},"./node_modules/taucharts/node_modules/d3-scale/src/sequential.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return i});var r=n("./node_modules/taucharts/node_modules/d3-scale/src/linear.js");function i(t){var e=0,n=1,o=!1;function a(r){var i=(r-e)/(n-e);return t(o?Math.max(0,Math.min(1,i)):i)}return a.domain=function(t){return arguments.length?(e=+t[0],n=+t[1],a):[e,n]},a.clamp=function(t){return arguments.length?(o=!!t,a):o},a.interpolator=function(e){return arguments.length?(t=e,a):t},a.copy=function(){return i(t).domain([e,n]).clamp(o)},Object(r.linearish)(a)}},"./node_modules/taucharts/node_modules/d3-scale/src/threshold.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return o});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/taucharts/node_modules/d3-scale/src/array.js");function o(){var t=[.5],e=[0,1],n=1;function a(i){if(i<=i)return e[Object(r.bisect)(t,i,0,n)]}return a.domain=function(r){return arguments.length?(t=i.slice.call(r),n=Math.min(t.length,e.length-1),a):t.slice()},a.range=function(r){return arguments.length?(e=i.slice.call(r),n=Math.min(t.length,e.length-1),a):e.slice()},a.invertExtent=function(n){var r=e.indexOf(n);return[t[r-1],t[r]]},a.copy=function(){return o().domain(t).range(e)},a}},"./node_modules/taucharts/node_modules/d3-scale/src/tickFormat.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/taucharts/node_modules/d3-format/src/index.js");e.default=function(t,e,n){var o,a=t[0],s=t[t.length-1],u=Object(r.tickStep)(a,s,null==e?10:e);switch((n=Object(i.formatSpecifier)(null==n?",f":n)).type){case"s":var c=Math.max(Math.abs(a),Math.abs(s));return null!=n.precision||isNaN(o=Object(i.precisionPrefix)(u,c))||(n.precision=o),Object(i.formatPrefix)(n,c);case"":case"e":case"g":case"p":case"r":null!=n.precision||isNaN(o=Object(i.precisionRound)(u,Math.max(Math.abs(a),Math.abs(s))))||(n.precision=o-("e"===n.type));break;case"f":case"%":null!=n.precision||isNaN(o=Object(i.precisionFixed)(u))||(n.precision=o-2*("%"===n.type))}return Object(i.format)(n)}},"./node_modules/taucharts/node_modules/d3-scale/src/time.js":function(t,e,n){"use strict";n.r(e),n.d(e,"calendar",function(){return v});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/d3-interpolate/index.js"),o=n("./node_modules/taucharts/node_modules/d3-time/src/index.js"),a=n("./node_modules/taucharts/node_modules/d3-time-format/src/index.js"),s=n("./node_modules/taucharts/node_modules/d3-scale/src/array.js"),u=n("./node_modules/taucharts/node_modules/d3-scale/src/continuous.js"),c=n("./node_modules/taucharts/node_modules/d3-scale/src/nice.js"),l=1e3,d=60*l,p=60*d,f=24*p,h=7*f,m=30*f,g=365*f;function y(t){return new Date(t)}function _(t){return t instanceof Date?+t:+new Date(+t)}function v(t,e,n,o,a,b,x,w,j){var O=Object(u.default)(u.deinterpolateLinear,i.interpolateNumber),C=O.invert,E=O.domain,S=j(".%L"),M=j(":%S"),T=j("%I:%M"),k=j("%I %p"),A=j("%a %d"),D=j("%b %d"),I=j("%B"),P=j("%Y"),R=[[x,1,l],[x,5,5*l],[x,15,15*l],[x,30,30*l],[b,1,d],[b,5,5*d],[b,15,15*d],[b,30,30*d],[a,1,p],[a,3,3*p],[a,6,6*p],[a,12,12*p],[o,1,f],[o,2,2*f],[n,1,h],[e,1,m],[e,3,3*m],[t,1,g]];function N(r){return(x(r)<r?S:b(r)<r?M:a(r)<r?T:o(r)<r?k:e(r)<r?n(r)<r?A:D:t(r)<r?I:P)(r)}function F(e,n,i,o){if(null==e&&(e=10),"number"==typeof e){var a=Math.abs(i-n)/e,s=Object(r.bisector)(function(t){return t[2]}).right(R,a);s===R.length?(o=Object(r.tickStep)(n/g,i/g,e),e=t):s?(o=(s=R[a/R[s-1][2]<R[s][2]/a?s-1:s])[1],e=s[0]):(o=Math.max(Object(r.tickStep)(n,i,e),1),e=w)}return null==o?e:e.every(o)}return O.invert=function(t){return new Date(C(t))},O.domain=function(t){return arguments.length?E(s.map.call(t,_)):E().map(y)},O.ticks=function(t,e){var n,r=E(),i=r[0],o=r[r.length-1],a=o<i;return a&&(n=i,i=o,o=n),n=(n=F(t,i,o,e))?n.range(i,o+1):[],a?n.reverse():n},O.tickFormat=function(t,e){return null==e?N:j(e)},O.nice=function(t,e){var n=E();return(t=F(t,n[0],n[n.length-1],e))?E(Object(c.default)(n,t)):O},O.copy=function(){return Object(u.copy)(O,v(t,e,n,o,a,b,x,w,j))},O}e.default=function(){return v(o.timeYear,o.timeMonth,o.timeWeek,o.timeDay,o.timeHour,o.timeMinute,o.timeSecond,o.timeMillisecond,a.timeFormat).domain([new Date(2e3,0,1),new Date(2e3,0,2)])}},"./node_modules/taucharts/node_modules/d3-scale/src/utcTime.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-scale/src/time.js"),i=n("./node_modules/taucharts/node_modules/d3-time-format/src/index.js"),o=n("./node_modules/taucharts/node_modules/d3-time/src/index.js");e.default=function(){return Object(r.calendar)(o.utcYear,o.utcMonth,o.utcWeek,o.utcDay,o.utcHour,o.utcMinute,o.utcSecond,o.utcMillisecond,i.utcFormat).domain([Date.UTC(2e3,0,1),Date.UTC(2e3,0,2)])}},"./node_modules/taucharts/node_modules/d3-scale/src/viridis.js":function(t,e,n){"use strict";n.r(e),n.d(e,"magma",function(){return o}),n.d(e,"inferno",function(){return a}),n.d(e,"plasma",function(){return s});var r=n("./node_modules/taucharts/node_modules/d3-scale/src/colors.js");function i(t){var e=t.length;return function(n){return t[Math.max(0,Math.min(e-1,Math.floor(n*e)))]}}e.default=i(Object(r.default)("44015444025645045745055946075a46085c460a5d460b5e470d60470e6147106347116447136548146748166848176948186a481a6c481b6d481c6e481d6f481f70482071482173482374482475482576482677482878482979472a7a472c7a472d7b472e7c472f7d46307e46327e46337f463480453581453781453882443983443a83443b84433d84433e85423f854240864241864142874144874045884046883f47883f48893e49893e4a893e4c8a3d4d8a3d4e8a3c4f8a3c508b3b518b3b528b3a538b3a548c39558c39568c38588c38598c375a8c375b8d365c8d365d8d355e8d355f8d34608d34618d33628d33638d32648e32658e31668e31678e31688e30698e306a8e2f6b8e2f6c8e2e6d8e2e6e8e2e6f8e2d708e2d718e2c718e2c728e2c738e2b748e2b758e2a768e2a778e2a788e29798e297a8e297b8e287c8e287d8e277e8e277f8e27808e26818e26828e26828e25838e25848e25858e24868e24878e23888e23898e238a8d228b8d228c8d228d8d218e8d218f8d21908d21918c20928c20928c20938c1f948c1f958b1f968b1f978b1f988b1f998a1f9a8a1e9b8a1e9c891e9d891f9e891f9f881fa0881fa1881fa1871fa28720a38620a48621a58521a68522a78522a88423a98324aa8325ab8225ac8226ad8127ad8128ae8029af7f2ab07f2cb17e2db27d2eb37c2fb47c31b57b32b67a34b67935b77937b87838b9773aba763bbb753dbc743fbc7340bd7242be7144bf7046c06f48c16e4ac16d4cc26c4ec36b50c46a52c56954c56856c66758c7655ac8645cc8635ec96260ca6063cb5f65cb5e67cc5c69cd5b6ccd5a6ece5870cf5773d05675d05477d1537ad1517cd2507fd34e81d34d84d44b86d54989d5488bd6468ed64590d74393d74195d84098d83e9bd93c9dd93ba0da39a2da37a5db36a8db34aadc32addc30b0dd2fb2dd2db5de2bb8de29bade28bddf26c0df25c2df23c5e021c8e020cae11fcde11dd0e11cd2e21bd5e21ad8e219dae319dde318dfe318e2e418e5e419e7e419eae51aece51befe51cf1e51df4e61ef6e620f8e621fbe723fde725"));var o=i(Object(r.default)("00000401000501010601010802010902020b02020d03030f03031204041405041606051806051a07061c08071e0907200a08220b09240c09260d0a290e0b2b100b2d110c2f120d31130d34140e36150e38160f3b180f3d19103f1a10421c10441d11471e114920114b21114e22115024125325125527125829115a2a115c2c115f2d11612f116331116533106734106936106b38106c390f6e3b0f703d0f713f0f72400f74420f75440f764510774710784910784a10794c117a4e117b4f127b51127c52137c54137d56147d57157e59157e5a167e5c167f5d177f5f187f601880621980641a80651a80671b80681c816a1c816b1d816d1d816e1e81701f81721f817320817521817621817822817922827b23827c23827e24828025828125818326818426818627818827818928818b29818c29818e2a81902a81912b81932b80942c80962c80982d80992d809b2e7f9c2e7f9e2f7fa02f7fa1307ea3307ea5317ea6317da8327daa337dab337cad347cae347bb0357bb2357bb3367ab5367ab73779b83779ba3878bc3978bd3977bf3a77c03a76c23b75c43c75c53c74c73d73c83e73ca3e72cc3f71cd4071cf4070d0416fd2426fd3436ed5446dd6456cd8456cd9466bdb476adc4869de4968df4a68e04c67e24d66e34e65e44f64e55064e75263e85362e95462ea5661eb5760ec5860ed5a5fee5b5eef5d5ef05f5ef1605df2625df2645cf3655cf4675cf4695cf56b5cf66c5cf66e5cf7705cf7725cf8745cf8765cf9785df9795df97b5dfa7d5efa7f5efa815ffb835ffb8560fb8761fc8961fc8a62fc8c63fc8e64fc9065fd9266fd9467fd9668fd9869fd9a6afd9b6bfe9d6cfe9f6dfea16efea36ffea571fea772fea973feaa74feac76feae77feb078feb27afeb47bfeb67cfeb77efeb97ffebb81febd82febf84fec185fec287fec488fec68afec88cfeca8dfecc8ffecd90fecf92fed194fed395fed597fed799fed89afdda9cfddc9efddea0fde0a1fde2a3fde3a5fde5a7fde7a9fde9aafdebacfcecaefceeb0fcf0b2fcf2b4fcf4b6fcf6b8fcf7b9fcf9bbfcfbbdfcfdbf")),a=i(Object(r.default)("00000401000501010601010802010a02020c02020e03021004031204031405041706041907051b08051d09061f0a07220b07240c08260d08290e092b10092d110a30120a32140b34150b37160b39180c3c190c3e1b0c411c0c431e0c451f0c48210c4a230c4c240c4f260c51280b53290b552b0b572d0b592f0a5b310a5c320a5e340a5f3609613809623909633b09643d09653e0966400a67420a68440a68450a69470b6a490b6a4a0c6b4c0c6b4d0d6c4f0d6c510e6c520e6d540f6d550f6d57106e59106e5a116e5c126e5d126e5f136e61136e62146e64156e65156e67166e69166e6a176e6c186e6d186e6f196e71196e721a6e741a6e751b6e771c6d781c6d7a1d6d7c1d6d7d1e6d7f1e6c801f6c82206c84206b85216b87216b88226a8a226a8c23698d23698f24699025689225689326679526679727669827669a28659b29649d29649f2a63a02a63a22b62a32c61a52c60a62d60a82e5fa92e5eab2f5ead305dae305cb0315bb1325ab3325ab43359b63458b73557b93556ba3655bc3754bd3853bf3952c03a51c13a50c33b4fc43c4ec63d4dc73e4cc83f4bca404acb4149cc4248ce4347cf4446d04545d24644d34743d44842d54a41d74b3fd84c3ed94d3dda4e3cdb503bdd513ade5238df5337e05536e15635e25734e35933e45a31e55c30e65d2fe75e2ee8602de9612bea632aeb6429eb6628ec6726ed6925ee6a24ef6c23ef6e21f06f20f1711ff1731df2741cf3761bf37819f47918f57b17f57d15f67e14f68013f78212f78410f8850ff8870ef8890cf98b0bf98c0af98e09fa9008fa9207fa9407fb9606fb9706fb9906fb9b06fb9d07fc9f07fca108fca309fca50afca60cfca80dfcaa0ffcac11fcae12fcb014fcb216fcb418fbb61afbb81dfbba1ffbbc21fbbe23fac026fac228fac42afac62df9c72ff9c932f9cb35f8cd37f8cf3af7d13df7d340f6d543f6d746f5d949f5db4cf4dd4ff4df53f4e156f3e35af3e55df2e661f2e865f2ea69f1ec6df1ed71f1ef75f1f179f2f27df2f482f3f586f3f68af4f88ef5f992f6fa96f8fb9af9fc9dfafda1fcffa4")),s=i(Object(r.default)("0d088710078813078916078a19068c1b068d1d068e20068f2206902406912605912805922a05932c05942e05952f059631059733059735049837049938049a3a049a3c049b3e049c3f049c41049d43039e44039e46039f48039f4903a04b03a14c02a14e02a25002a25102a35302a35502a45601a45801a45901a55b01a55c01a65e01a66001a66100a76300a76400a76600a76700a86900a86a00a86c00a86e00a86f00a87100a87201a87401a87501a87701a87801a87a02a87b02a87d03a87e03a88004a88104a78305a78405a78606a68707a68808a68a09a58b0aa58d0ba58e0ca48f0da4910ea3920fa39410a29511a19613a19814a099159f9a169f9c179e9d189d9e199da01a9ca11b9ba21d9aa31e9aa51f99a62098a72197a82296aa2395ab2494ac2694ad2793ae2892b02991b12a90b22b8fb32c8eb42e8db52f8cb6308bb7318ab83289ba3388bb3488bc3587bd3786be3885bf3984c03a83c13b82c23c81c33d80c43e7fc5407ec6417dc7427cc8437bc9447aca457acb4679cc4778cc4977cd4a76ce4b75cf4c74d04d73d14e72d24f71d35171d45270d5536fd5546ed6556dd7566cd8576bd9586ada5a6ada5b69db5c68dc5d67dd5e66de5f65de6164df6263e06363e16462e26561e26660e3685fe4695ee56a5de56b5de66c5ce76e5be76f5ae87059e97158e97257ea7457eb7556eb7655ec7754ed7953ed7a52ee7b51ef7c51ef7e50f07f4ff0804ef1814df1834cf2844bf3854bf3874af48849f48948f58b47f58c46f68d45f68f44f79044f79143f79342f89441f89540f9973ff9983ef99a3efa9b3dfa9c3cfa9e3bfb9f3afba139fba238fca338fca537fca636fca835fca934fdab33fdac33fdae32fdaf31fdb130fdb22ffdb42ffdb52efeb72dfeb82cfeba2cfebb2bfebd2afebe2afec029fdc229fdc328fdc527fdc627fdc827fdca26fdcb26fccd25fcce25fcd025fcd225fbd324fbd524fbd724fad824fada24f9dc24f9dd25f8df25f8e125f7e225f7e425f6e626f6e826f5e926f5eb27f4ed27f3ee27f3f027f2f227f1f426f1f525f0f724f0f921"))},"./node_modules/taucharts/node_modules/d3-shape/src/arc.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-path/index.js"),i=n("./node_modules/taucharts/node_modules/d3-shape/src/constant.js"),o=n("./node_modules/taucharts/node_modules/d3-shape/src/math.js");function a(t){return t.innerRadius}function s(t){return t.outerRadius}function u(t){return t.startAngle}function c(t){return t.endAngle}function l(t){return t&&t.padAngle}function d(t,e,n,r,i,a,s){var u=t-n,c=e-r,l=(s?a:-a)/Object(o.sqrt)(u*u+c*c),d=l*c,p=-l*u,f=t+d,h=e+p,m=n+d,g=r+p,y=(f+m)/2,_=(h+g)/2,v=m-f,b=g-h,x=v*v+b*b,w=i-a,j=f*g-m*h,O=(b<0?-1:1)*Object(o.sqrt)(Object(o.max)(0,w*w*x-j*j)),C=(j*b-v*O)/x,E=(-j*v-b*O)/x,S=(j*b+v*O)/x,M=(-j*v+b*O)/x,T=C-y,k=E-_,A=S-y,D=M-_;return T*T+k*k>A*A+D*D&&(C=S,E=M),{cx:C,cy:E,x01:-d,y01:-p,x11:C*(i/w-1),y11:E*(i/w-1)}}e.default=function(){var t=a,e=s,n=Object(i.default)(0),p=null,f=u,h=c,m=l,g=null;function y(){var i,a,s=+t.apply(this,arguments),u=+e.apply(this,arguments),c=f.apply(this,arguments)-o.halfPi,l=h.apply(this,arguments)-o.halfPi,y=Object(o.abs)(l-c),_=l>c;if(g||(g=i=Object(r.path)()),u<s&&(a=u,u=s,s=a),u>o.epsilon)if(y>o.tau-o.epsilon)g.moveTo(u*Object(o.cos)(c),u*Object(o.sin)(c)),g.arc(0,0,u,c,l,!_),s>o.epsilon&&(g.moveTo(s*Object(o.cos)(l),s*Object(o.sin)(l)),g.arc(0,0,s,l,c,_));else{var v,b,x=c,w=l,j=c,O=l,C=y,E=y,S=m.apply(this,arguments)/2,M=S>o.epsilon&&(p?+p.apply(this,arguments):Object(o.sqrt)(s*s+u*u)),T=Object(o.min)(Object(o.abs)(u-s)/2,+n.apply(this,arguments)),k=T,A=T;if(M>o.epsilon){var D=Object(o.asin)(M/s*Object(o.sin)(S)),I=Object(o.asin)(M/u*Object(o.sin)(S));(C-=2*D)>o.epsilon?(j+=D*=_?1:-1,O-=D):(C=0,j=O=(c+l)/2),(E-=2*I)>o.epsilon?(x+=I*=_?1:-1,w-=I):(E=0,x=w=(c+l)/2)}var P=u*Object(o.cos)(x),R=u*Object(o.sin)(x),N=s*Object(o.cos)(O),F=s*Object(o.sin)(O);if(T>o.epsilon){var L,B=u*Object(o.cos)(w),V=u*Object(o.sin)(w),z=s*Object(o.cos)(j),H=s*Object(o.sin)(j);if(y<o.pi&&(L=function(t,e,n,r,i,a,s,u){var c=n-t,l=r-e,d=s-i,p=u-a,f=p*c-d*l;if(!(f*f<o.epsilon))return[t+(f=(d*(e-a)-p*(t-i))/f)*c,e+f*l]}(P,R,z,H,B,V,N,F))){var U=P-L[0],W=R-L[1],q=B-L[0],G=V-L[1],Y=1/Object(o.sin)(Object(o.acos)((U*q+W*G)/(Object(o.sqrt)(U*U+W*W)*Object(o.sqrt)(q*q+G*G)))/2),K=Object(o.sqrt)(L[0]*L[0]+L[1]*L[1]);k=Object(o.min)(T,(s-K)/(Y-1)),A=Object(o.min)(T,(u-K)/(Y+1))}}E>o.epsilon?A>o.epsilon?(v=d(z,H,P,R,u,A,_),b=d(B,V,N,F,u,A,_),g.moveTo(v.cx+v.x01,v.cy+v.y01),A<T?g.arc(v.cx,v.cy,A,Object(o.atan2)(v.y01,v.x01),Object(o.atan2)(b.y01,b.x01),!_):(g.arc(v.cx,v.cy,A,Object(o.atan2)(v.y01,v.x01),Object(o.atan2)(v.y11,v.x11),!_),g.arc(0,0,u,Object(o.atan2)(v.cy+v.y11,v.cx+v.x11),Object(o.atan2)(b.cy+b.y11,b.cx+b.x11),!_),g.arc(b.cx,b.cy,A,Object(o.atan2)(b.y11,b.x11),Object(o.atan2)(b.y01,b.x01),!_))):(g.moveTo(P,R),g.arc(0,0,u,x,w,!_)):g.moveTo(P,R),s>o.epsilon&&C>o.epsilon?k>o.epsilon?(v=d(N,F,B,V,s,-k,_),b=d(P,R,z,H,s,-k,_),g.lineTo(v.cx+v.x01,v.cy+v.y01),k<T?g.arc(v.cx,v.cy,k,Object(o.atan2)(v.y01,v.x01),Object(o.atan2)(b.y01,b.x01),!_):(g.arc(v.cx,v.cy,k,Object(o.atan2)(v.y01,v.x01),Object(o.atan2)(v.y11,v.x11),!_),g.arc(0,0,s,Object(o.atan2)(v.cy+v.y11,v.cx+v.x11),Object(o.atan2)(b.cy+b.y11,b.cx+b.x11),_),g.arc(b.cx,b.cy,k,Object(o.atan2)(b.y11,b.x11),Object(o.atan2)(b.y01,b.x01),!_))):g.arc(0,0,s,O,j,_):g.lineTo(N,F)}else g.moveTo(0,0);if(g.closePath(),i)return g=null,i+""||null}return y.centroid=function(){var n=(+t.apply(this,arguments)+ +e.apply(this,arguments))/2,r=(+f.apply(this,arguments)+ +h.apply(this,arguments))/2-o.pi/2;return[Object(o.cos)(r)*n,Object(o.sin)(r)*n]},y.innerRadius=function(e){return arguments.length?(t="function"==typeof e?e:Object(i.default)(+e),y):t},y.outerRadius=function(t){return arguments.length?(e="function"==typeof t?t:Object(i.default)(+t),y):e},y.cornerRadius=function(t){return arguments.length?(n="function"==typeof t?t:Object(i.default)(+t),y):n},y.padRadius=function(t){return arguments.length?(p=null==t?null:"function"==typeof t?t:Object(i.default)(+t),y):p},y.startAngle=function(t){return arguments.length?(f="function"==typeof t?t:Object(i.default)(+t),y):f},y.endAngle=function(t){return arguments.length?(h="function"==typeof t?t:Object(i.default)(+t),y):h},y.padAngle=function(t){return arguments.length?(m="function"==typeof t?t:Object(i.default)(+t),y):m},y.context=function(t){return arguments.length?(g=null==t?null:t,y):g},y}},"./node_modules/taucharts/node_modules/d3-shape/src/area.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-path/index.js"),i=n("./node_modules/taucharts/node_modules/d3-shape/src/constant.js"),o=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/linear.js"),a=n("./node_modules/taucharts/node_modules/d3-shape/src/line.js"),s=n("./node_modules/taucharts/node_modules/d3-shape/src/point.js");e.default=function(){var t=s.x,e=null,n=Object(i.default)(0),u=s.y,c=Object(i.default)(!0),l=null,d=o.default,p=null;function f(i){var o,a,s,f,h,m=i.length,g=!1,y=new Array(m),_=new Array(m);for(null==l&&(p=d(h=Object(r.path)())),o=0;o<=m;++o){if(!(o<m&&c(f=i[o],o,i))===g)if(g=!g)a=o,p.areaStart(),p.lineStart();else{for(p.lineEnd(),p.lineStart(),s=o-1;s>=a;--s)p.point(y[s],_[s]);p.lineEnd(),p.areaEnd()}g&&(y[o]=+t(f,o,i),_[o]=+n(f,o,i),p.point(e?+e(f,o,i):y[o],u?+u(f,o,i):_[o]))}if(h)return p=null,h+""||null}function h(){return Object(a.default)().defined(c).curve(d).context(l)}return f.x=function(n){return arguments.length?(t="function"==typeof n?n:Object(i.default)(+n),e=null,f):t},f.x0=function(e){return arguments.length?(t="function"==typeof e?e:Object(i.default)(+e),f):t},f.x1=function(t){return arguments.length?(e=null==t?null:"function"==typeof t?t:Object(i.default)(+t),f):e},f.y=function(t){return arguments.length?(n="function"==typeof t?t:Object(i.default)(+t),u=null,f):n},f.y0=function(t){return arguments.length?(n="function"==typeof t?t:Object(i.default)(+t),f):n},f.y1=function(t){return arguments.length?(u=null==t?null:"function"==typeof t?t:Object(i.default)(+t),f):u},f.lineX0=f.lineY0=function(){return h().x(t).y(n)},f.lineY1=function(){return h().x(t).y(u)},f.lineX1=function(){return h().x(e).y(n)},f.defined=function(t){return arguments.length?(c="function"==typeof t?t:Object(i.default)(!!t),f):c},f.curve=function(t){return arguments.length?(d=t,null!=l&&(p=d(l)),f):d},f.context=function(t){return arguments.length?(null==t?l=p=null:p=d(l=t),f):l},f}},"./node_modules/taucharts/node_modules/d3-shape/src/areaRadial.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/radial.js"),i=n("./node_modules/taucharts/node_modules/d3-shape/src/area.js"),o=n("./node_modules/taucharts/node_modules/d3-shape/src/lineRadial.js");e.default=function(){var t=Object(i.default)().curve(r.curveRadialLinear),e=t.curve,n=t.lineX0,a=t.lineX1,s=t.lineY0,u=t.lineY1;return t.angle=t.x,delete t.x,t.startAngle=t.x0,delete t.x0,t.endAngle=t.x1,delete t.x1,t.radius=t.y,delete t.y,t.innerRadius=t.y0,delete t.y0,t.outerRadius=t.y1,delete t.y1,t.lineStartAngle=function(){return Object(o.lineRadial)(n())},delete t.lineX0,t.lineEndAngle=function(){return Object(o.lineRadial)(a())},delete t.lineX1,t.lineInnerRadius=function(){return Object(o.lineRadial)(s())},delete t.lineY0,t.lineOuterRadius=function(){return Object(o.lineRadial)(u())},delete t.lineY1,t.curve=function(t){return arguments.length?e(Object(r.default)(t)):e()._curve},t}},"./node_modules/taucharts/node_modules/d3-shape/src/array.js":function(t,e,n){"use strict";n.r(e),n.d(e,"slice",function(){return r});var r=Array.prototype.slice},"./node_modules/taucharts/node_modules/d3-shape/src/constant.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return function(){return t}}},"./node_modules/taucharts/node_modules/d3-shape/src/curve/basis.js":function(t,e,n){"use strict";function r(t,e,n){t._context.bezierCurveTo((2*t._x0+t._x1)/3,(2*t._y0+t._y1)/3,(t._x0+2*t._x1)/3,(t._y0+2*t._y1)/3,(t._x0+4*t._x1+e)/6,(t._y0+4*t._y1+n)/6)}function i(t){this._context=t}n.r(e),n.d(e,"point",function(){return r}),n.d(e,"Basis",function(){return i}),i.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=NaN,this._point=0},lineEnd:function(){switch(this._point){case 3:r(this,this._x1,this._y1);case 2:this._context.lineTo(this._x1,this._y1)}(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2;break;case 2:this._point=3,this._context.lineTo((5*this._x0+this._x1)/6,(5*this._y0+this._y1)/6);default:r(this,t,e)}this._x0=this._x1,this._x1=t,this._y0=this._y1,this._y1=e}},e.default=function(t){return new i(t)}},"./node_modules/taucharts/node_modules/d3-shape/src/curve/basisClosed.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/noop.js"),i=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/basis.js");function o(t){this._context=t}o.prototype={areaStart:r.default,areaEnd:r.default,lineStart:function(){this._x0=this._x1=this._x2=this._x3=this._x4=this._y0=this._y1=this._y2=this._y3=this._y4=NaN,this._point=0},lineEnd:function(){switch(this._point){case 1:this._context.moveTo(this._x2,this._y2),this._context.closePath();break;case 2:this._context.moveTo((this._x2+2*this._x3)/3,(this._y2+2*this._y3)/3),this._context.lineTo((this._x3+2*this._x2)/3,(this._y3+2*this._y2)/3),this._context.closePath();break;case 3:this.point(this._x2,this._y2),this.point(this._x3,this._y3),this.point(this._x4,this._y4)}},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._x2=t,this._y2=e;break;case 1:this._point=2,this._x3=t,this._y3=e;break;case 2:this._point=3,this._x4=t,this._y4=e,this._context.moveTo((this._x0+4*this._x1+t)/6,(this._y0+4*this._y1+e)/6);break;default:Object(i.point)(this,t,e)}this._x0=this._x1,this._x1=t,this._y0=this._y1,this._y1=e}},e.default=function(t){return new o(t)}},"./node_modules/taucharts/node_modules/d3-shape/src/curve/basisOpen.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/basis.js");function i(t){this._context=t}i.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=NaN,this._point=0},lineEnd:function(){(this._line||0!==this._line&&3===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1;break;case 1:this._point=2;break;case 2:this._point=3;var n=(this._x0+4*this._x1+t)/6,i=(this._y0+4*this._y1+e)/6;this._line?this._context.lineTo(n,i):this._context.moveTo(n,i);break;case 3:this._point=4;default:Object(r.point)(this,t,e)}this._x0=this._x1,this._x1=t,this._y0=this._y1,this._y1=e}},e.default=function(t){return new i(t)}},"./node_modules/taucharts/node_modules/d3-shape/src/curve/bundle.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/basis.js");function i(t,e){this._basis=new r.Basis(t),this._beta=e}i.prototype={lineStart:function(){this._x=[],this._y=[],this._basis.lineStart()},lineEnd:function(){var t=this._x,e=this._y,n=t.length-1;if(n>0)for(var r,i=t[0],o=e[0],a=t[n]-i,s=e[n]-o,u=-1;++u<=n;)r=u/n,this._basis.point(this._beta*t[u]+(1-this._beta)*(i+r*a),this._beta*e[u]+(1-this._beta)*(o+r*s));this._x=this._y=null,this._basis.lineEnd()},point:function(t,e){this._x.push(+t),this._y.push(+e)}},e.default=function t(e){function n(t){return 1===e?new r.Basis(t):new i(t,e)}return n.beta=function(e){return t(+e)},n}(.85)},"./node_modules/taucharts/node_modules/d3-shape/src/curve/cardinal.js":function(t,e,n){"use strict";function r(t,e,n){t._context.bezierCurveTo(t._x1+t._k*(t._x2-t._x0),t._y1+t._k*(t._y2-t._y0),t._x2+t._k*(t._x1-e),t._y2+t._k*(t._y1-n),t._x2,t._y2)}function i(t,e){this._context=t,this._k=(1-e)/6}n.r(e),n.d(e,"point",function(){return r}),n.d(e,"Cardinal",function(){return i}),i.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._x2=this._y0=this._y1=this._y2=NaN,this._point=0},lineEnd:function(){switch(this._point){case 2:this._context.lineTo(this._x2,this._y2);break;case 3:r(this,this._x1,this._y1)}(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2,this._x1=t,this._y1=e;break;case 2:this._point=3;default:r(this,t,e)}this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}},e.default=function t(e){function n(t){return new i(t,e)}return n.tension=function(e){return t(+e)},n}(0)},"./node_modules/taucharts/node_modules/d3-shape/src/curve/cardinalClosed.js":function(t,e,n){"use strict";n.r(e),n.d(e,"CardinalClosed",function(){return o});var r=n("./node_modules/taucharts/node_modules/d3-shape/src/noop.js"),i=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/cardinal.js");function o(t,e){this._context=t,this._k=(1-e)/6}o.prototype={areaStart:r.default,areaEnd:r.default,lineStart:function(){this._x0=this._x1=this._x2=this._x3=this._x4=this._x5=this._y0=this._y1=this._y2=this._y3=this._y4=this._y5=NaN,this._point=0},lineEnd:function(){switch(this._point){case 1:this._context.moveTo(this._x3,this._y3),this._context.closePath();break;case 2:this._context.lineTo(this._x3,this._y3),this._context.closePath();break;case 3:this.point(this._x3,this._y3),this.point(this._x4,this._y4),this.point(this._x5,this._y5)}},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._x3=t,this._y3=e;break;case 1:this._point=2,this._context.moveTo(this._x4=t,this._y4=e);break;case 2:this._point=3,this._x5=t,this._y5=e;break;default:Object(i.point)(this,t,e)}this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}},e.default=function t(e){function n(t){return new o(t,e)}return n.tension=function(e){return t(+e)},n}(0)},"./node_modules/taucharts/node_modules/d3-shape/src/curve/cardinalOpen.js":function(t,e,n){"use strict";n.r(e),n.d(e,"CardinalOpen",function(){return i});var r=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/cardinal.js");function i(t,e){this._context=t,this._k=(1-e)/6}i.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._x2=this._y0=this._y1=this._y2=NaN,this._point=0},lineEnd:function(){(this._line||0!==this._line&&3===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1;break;case 1:this._point=2;break;case 2:this._point=3,this._line?this._context.lineTo(this._x2,this._y2):this._context.moveTo(this._x2,this._y2);break;case 3:this._point=4;default:Object(r.point)(this,t,e)}this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}},e.default=function t(e){function n(t){return new i(t,e)}return n.tension=function(e){return t(+e)},n}(0)},"./node_modules/taucharts/node_modules/d3-shape/src/curve/catmullRom.js":function(t,e,n){"use strict";n.r(e),n.d(e,"point",function(){return o});var r=n("./node_modules/taucharts/node_modules/d3-shape/src/math.js"),i=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/cardinal.js");function o(t,e,n){var i=t._x1,o=t._y1,a=t._x2,s=t._y2;if(t._l01_a>r.epsilon){var u=2*t._l01_2a+3*t._l01_a*t._l12_a+t._l12_2a,c=3*t._l01_a*(t._l01_a+t._l12_a);i=(i*u-t._x0*t._l12_2a+t._x2*t._l01_2a)/c,o=(o*u-t._y0*t._l12_2a+t._y2*t._l01_2a)/c}if(t._l23_a>r.epsilon){var l=2*t._l23_2a+3*t._l23_a*t._l12_a+t._l12_2a,d=3*t._l23_a*(t._l23_a+t._l12_a);a=(a*l+t._x1*t._l23_2a-e*t._l12_2a)/d,s=(s*l+t._y1*t._l23_2a-n*t._l12_2a)/d}t._context.bezierCurveTo(i,o,a,s,t._x2,t._y2)}function a(t,e){this._context=t,this._alpha=e}a.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._x2=this._y0=this._y1=this._y2=NaN,this._l01_a=this._l12_a=this._l23_a=this._l01_2a=this._l12_2a=this._l23_2a=this._point=0},lineEnd:function(){switch(this._point){case 2:this._context.lineTo(this._x2,this._y2);break;case 3:this.point(this._x2,this._y2)}(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){if(t=+t,e=+e,this._point){var n=this._x2-t,r=this._y2-e;this._l23_a=Math.sqrt(this._l23_2a=Math.pow(n*n+r*r,this._alpha))}switch(this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2;break;case 2:this._point=3;default:o(this,t,e)}this._l01_a=this._l12_a,this._l12_a=this._l23_a,this._l01_2a=this._l12_2a,this._l12_2a=this._l23_2a,this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}},e.default=function t(e){function n(t){return e?new a(t,e):new i.Cardinal(t,0)}return n.alpha=function(e){return t(+e)},n}(.5)},"./node_modules/taucharts/node_modules/d3-shape/src/curve/catmullRomClosed.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/cardinalClosed.js"),i=n("./node_modules/taucharts/node_modules/d3-shape/src/noop.js"),o=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/catmullRom.js");function a(t,e){this._context=t,this._alpha=e}a.prototype={areaStart:i.default,areaEnd:i.default,lineStart:function(){this._x0=this._x1=this._x2=this._x3=this._x4=this._x5=this._y0=this._y1=this._y2=this._y3=this._y4=this._y5=NaN,this._l01_a=this._l12_a=this._l23_a=this._l01_2a=this._l12_2a=this._l23_2a=this._point=0},lineEnd:function(){switch(this._point){case 1:this._context.moveTo(this._x3,this._y3),this._context.closePath();break;case 2:this._context.lineTo(this._x3,this._y3),this._context.closePath();break;case 3:this.point(this._x3,this._y3),this.point(this._x4,this._y4),this.point(this._x5,this._y5)}},point:function(t,e){if(t=+t,e=+e,this._point){var n=this._x2-t,r=this._y2-e;this._l23_a=Math.sqrt(this._l23_2a=Math.pow(n*n+r*r,this._alpha))}switch(this._point){case 0:this._point=1,this._x3=t,this._y3=e;break;case 1:this._point=2,this._context.moveTo(this._x4=t,this._y4=e);break;case 2:this._point=3,this._x5=t,this._y5=e;break;default:Object(o.point)(this,t,e)}this._l01_a=this._l12_a,this._l12_a=this._l23_a,this._l01_2a=this._l12_2a,this._l12_2a=this._l23_2a,this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}},e.default=function t(e){function n(t){return e?new a(t,e):new r.CardinalClosed(t,0)}return n.alpha=function(e){return t(+e)},n}(.5)},"./node_modules/taucharts/node_modules/d3-shape/src/curve/catmullRomOpen.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/cardinalOpen.js"),i=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/catmullRom.js");function o(t,e){this._context=t,this._alpha=e}o.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._x2=this._y0=this._y1=this._y2=NaN,this._l01_a=this._l12_a=this._l23_a=this._l01_2a=this._l12_2a=this._l23_2a=this._point=0},lineEnd:function(){(this._line||0!==this._line&&3===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){if(t=+t,e=+e,this._point){var n=this._x2-t,r=this._y2-e;this._l23_a=Math.sqrt(this._l23_2a=Math.pow(n*n+r*r,this._alpha))}switch(this._point){case 0:this._point=1;break;case 1:this._point=2;break;case 2:this._point=3,this._line?this._context.lineTo(this._x2,this._y2):this._context.moveTo(this._x2,this._y2);break;case 3:this._point=4;default:Object(i.point)(this,t,e)}this._l01_a=this._l12_a,this._l12_a=this._l23_a,this._l01_2a=this._l12_2a,this._l12_2a=this._l23_2a,this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}},e.default=function t(e){function n(t){return e?new o(t,e):new r.CardinalOpen(t,0)}return n.alpha=function(e){return t(+e)},n}(.5)},"./node_modules/taucharts/node_modules/d3-shape/src/curve/linear.js":function(t,e,n){"use strict";function r(t){this._context=t}n.r(e),r.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._point=0},lineEnd:function(){(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2;default:this._context.lineTo(t,e)}}},e.default=function(t){return new r(t)}},"./node_modules/taucharts/node_modules/d3-shape/src/curve/linearClosed.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/noop.js");function i(t){this._context=t}i.prototype={areaStart:r.default,areaEnd:r.default,lineStart:function(){this._point=0},lineEnd:function(){this._point&&this._context.closePath()},point:function(t,e){t=+t,e=+e,this._point?this._context.lineTo(t,e):(this._point=1,this._context.moveTo(t,e))}},e.default=function(t){return new i(t)}},"./node_modules/taucharts/node_modules/d3-shape/src/curve/monotone.js":function(t,e,n){"use strict";function r(t){return t<0?-1:1}function i(t,e,n){var i=t._x1-t._x0,o=e-t._x1,a=(t._y1-t._y0)/(i||o<0&&-0),s=(n-t._y1)/(o||i<0&&-0),u=(a*o+s*i)/(i+o);return(r(a)+r(s))*Math.min(Math.abs(a),Math.abs(s),.5*Math.abs(u))||0}function o(t,e){var n=t._x1-t._x0;return n?(3*(t._y1-t._y0)/n-e)/2:e}function a(t,e,n){var r=t._x0,i=t._y0,o=t._x1,a=t._y1,s=(o-r)/3;t._context.bezierCurveTo(r+s,i+s*e,o-s,a-s*n,o,a)}function s(t){this._context=t}function u(t){this._context=new c(t)}function c(t){this._context=t}function l(t){return new s(t)}function d(t){return new u(t)}n.r(e),n.d(e,"monotoneX",function(){return l}),n.d(e,"monotoneY",function(){return d}),s.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=this._t0=NaN,this._point=0},lineEnd:function(){switch(this._point){case 2:this._context.lineTo(this._x1,this._y1);break;case 3:a(this,this._t0,o(this,this._t0))}(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){var n=NaN;if(e=+e,(t=+t)!==this._x1||e!==this._y1){switch(this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2;break;case 2:this._point=3,a(this,o(this,n=i(this,t,e)),n);break;default:a(this,this._t0,n=i(this,t,e))}this._x0=this._x1,this._x1=t,this._y0=this._y1,this._y1=e,this._t0=n}}},(u.prototype=Object.create(s.prototype)).point=function(t,e){s.prototype.point.call(this,e,t)},c.prototype={moveTo:function(t,e){this._context.moveTo(e,t)},closePath:function(){this._context.closePath()},lineTo:function(t,e){this._context.lineTo(e,t)},bezierCurveTo:function(t,e,n,r,i,o){this._context.bezierCurveTo(e,t,r,n,o,i)}}},"./node_modules/taucharts/node_modules/d3-shape/src/curve/natural.js":function(t,e,n){"use strict";function r(t){this._context=t}function i(t){var e,n,r=t.length-1,i=new Array(r),o=new Array(r),a=new Array(r);for(i[0]=0,o[0]=2,a[0]=t[0]+2*t[1],e=1;e<r-1;++e)i[e]=1,o[e]=4,a[e]=4*t[e]+2*t[e+1];for(i[r-1]=2,o[r-1]=7,a[r-1]=8*t[r-1]+t[r],e=1;e<r;++e)n=i[e]/o[e-1],o[e]-=n,a[e]-=n*a[e-1];for(i[r-1]=a[r-1]/o[r-1],e=r-2;e>=0;--e)i[e]=(a[e]-i[e+1])/o[e];for(o[r-1]=(t[r]+i[r-1])/2,e=0;e<r-1;++e)o[e]=2*t[e+1]-i[e+1];return[i,o]}n.r(e),r.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x=[],this._y=[]},lineEnd:function(){var t=this._x,e=this._y,n=t.length;if(n)if(this._line?this._context.lineTo(t[0],e[0]):this._context.moveTo(t[0],e[0]),2===n)this._context.lineTo(t[1],e[1]);else for(var r=i(t),o=i(e),a=0,s=1;s<n;++a,++s)this._context.bezierCurveTo(r[0][a],o[0][a],r[1][a],o[1][a],t[s],e[s]);(this._line||0!==this._line&&1===n)&&this._context.closePath(),this._line=1-this._line,this._x=this._y=null},point:function(t,e){this._x.push(+t),this._y.push(+e)}},e.default=function(t){return new r(t)}},"./node_modules/taucharts/node_modules/d3-shape/src/curve/radial.js":function(t,e,n){"use strict";n.r(e),n.d(e,"curveRadialLinear",function(){return r}),n.d(e,"default",function(){return o});var r=o(n("./node_modules/taucharts/node_modules/d3-shape/src/curve/linear.js").default);function i(t){this._curve=t}function o(t){function e(e){return new i(t(e))}return e._curve=t,e}i.prototype={areaStart:function(){this._curve.areaStart()},areaEnd:function(){this._curve.areaEnd()},lineStart:function(){this._curve.lineStart()},lineEnd:function(){this._curve.lineEnd()},point:function(t,e){this._curve.point(e*Math.sin(t),e*-Math.cos(t))}}},"./node_modules/taucharts/node_modules/d3-shape/src/curve/step.js":function(t,e,n){"use strict";function r(t,e){this._context=t,this._t=e}function i(t){return new r(t,0)}function o(t){return new r(t,1)}n.r(e),n.d(e,"stepBefore",function(){return i}),n.d(e,"stepAfter",function(){return o}),r.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x=this._y=NaN,this._point=0},lineEnd:function(){0<this._t&&this._t<1&&2===this._point&&this._context.lineTo(this._x,this._y),(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line>=0&&(this._t=1-this._t,this._line=1-this._line)},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2;default:if(this._t<=0)this._context.lineTo(this._x,e),this._context.lineTo(t,e);else{var n=this._x*(1-this._t)+t*this._t;this._context.lineTo(n,this._y),this._context.lineTo(n,e)}}this._x=t,this._y=e}},e.default=function(t){return new r(t,.5)}},"./node_modules/taucharts/node_modules/d3-shape/src/descending.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){return e<t?-1:e>t?1:e>=t?0:NaN}},"./node_modules/taucharts/node_modules/d3-shape/src/identity.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return t}},"./node_modules/taucharts/node_modules/d3-shape/src/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/arc.js");n.d(e,"arc",function(){return r.default});var i=n("./node_modules/taucharts/node_modules/d3-shape/src/area.js");n.d(e,"area",function(){return i.default});var o=n("./node_modules/taucharts/node_modules/d3-shape/src/line.js");n.d(e,"line",function(){return o.default});var a=n("./node_modules/taucharts/node_modules/d3-shape/src/pie.js");n.d(e,"pie",function(){return a.default});var s=n("./node_modules/taucharts/node_modules/d3-shape/src/areaRadial.js");n.d(e,"areaRadial",function(){return s.default}),n.d(e,"radialArea",function(){return s.default});var u=n("./node_modules/taucharts/node_modules/d3-shape/src/lineRadial.js");n.d(e,"lineRadial",function(){return u.default}),n.d(e,"radialLine",function(){return u.default});var c=n("./node_modules/taucharts/node_modules/d3-shape/src/pointRadial.js");n.d(e,"pointRadial",function(){return c.default});var l=n("./node_modules/taucharts/node_modules/d3-shape/src/link/index.js");n.d(e,"linkHorizontal",function(){return l.linkHorizontal}),n.d(e,"linkVertical",function(){return l.linkVertical}),n.d(e,"linkRadial",function(){return l.linkRadial});var d=n("./node_modules/taucharts/node_modules/d3-shape/src/symbol.js");n.d(e,"symbol",function(){return d.default}),n.d(e,"symbols",function(){return d.symbols});var p=n("./node_modules/taucharts/node_modules/d3-shape/src/symbol/circle.js");n.d(e,"symbolCircle",function(){return p.default});var f=n("./node_modules/taucharts/node_modules/d3-shape/src/symbol/cross.js");n.d(e,"symbolCross",function(){return f.default});var h=n("./node_modules/taucharts/node_modules/d3-shape/src/symbol/diamond.js");n.d(e,"symbolDiamond",function(){return h.default});var m=n("./node_modules/taucharts/node_modules/d3-shape/src/symbol/square.js");n.d(e,"symbolSquare",function(){return m.default});var g=n("./node_modules/taucharts/node_modules/d3-shape/src/symbol/star.js");n.d(e,"symbolStar",function(){return g.default});var y=n("./node_modules/taucharts/node_modules/d3-shape/src/symbol/triangle.js");n.d(e,"symbolTriangle",function(){return y.default});var _=n("./node_modules/taucharts/node_modules/d3-shape/src/symbol/wye.js");n.d(e,"symbolWye",function(){return _.default});var v=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/basisClosed.js");n.d(e,"curveBasisClosed",function(){return v.default});var b=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/basisOpen.js");n.d(e,"curveBasisOpen",function(){return b.default});var x=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/basis.js");n.d(e,"curveBasis",function(){return x.default});var w=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/bundle.js");n.d(e,"curveBundle",function(){return w.default});var j=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/cardinalClosed.js");n.d(e,"curveCardinalClosed",function(){return j.default});var O=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/cardinalOpen.js");n.d(e,"curveCardinalOpen",function(){return O.default});var C=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/cardinal.js");n.d(e,"curveCardinal",function(){return C.default});var E=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/catmullRomClosed.js");n.d(e,"curveCatmullRomClosed",function(){return E.default});var S=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/catmullRomOpen.js");n.d(e,"curveCatmullRomOpen",function(){return S.default});var M=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/catmullRom.js");n.d(e,"curveCatmullRom",function(){return M.default});var T=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/linearClosed.js");n.d(e,"curveLinearClosed",function(){return T.default});var k=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/linear.js");n.d(e,"curveLinear",function(){return k.default});var A=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/monotone.js");n.d(e,"curveMonotoneX",function(){return A.monotoneX}),n.d(e,"curveMonotoneY",function(){return A.monotoneY});var D=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/natural.js");n.d(e,"curveNatural",function(){return D.default});var I=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/step.js");n.d(e,"curveStep",function(){return I.default}),n.d(e,"curveStepAfter",function(){return I.stepAfter}),n.d(e,"curveStepBefore",function(){return I.stepBefore});var P=n("./node_modules/taucharts/node_modules/d3-shape/src/stack.js");n.d(e,"stack",function(){return P.default});var R=n("./node_modules/taucharts/node_modules/d3-shape/src/offset/expand.js");n.d(e,"stackOffsetExpand",function(){return R.default});var N=n("./node_modules/taucharts/node_modules/d3-shape/src/offset/diverging.js");n.d(e,"stackOffsetDiverging",function(){return N.default});var F=n("./node_modules/taucharts/node_modules/d3-shape/src/offset/none.js");n.d(e,"stackOffsetNone",function(){return F.default});var L=n("./node_modules/taucharts/node_modules/d3-shape/src/offset/silhouette.js");n.d(e,"stackOffsetSilhouette",function(){return L.default});var B=n("./node_modules/taucharts/node_modules/d3-shape/src/offset/wiggle.js");n.d(e,"stackOffsetWiggle",function(){return B.default});var V=n("./node_modules/taucharts/node_modules/d3-shape/src/order/appearance.js");n.d(e,"stackOrderAppearance",function(){return V.default});var z=n("./node_modules/taucharts/node_modules/d3-shape/src/order/ascending.js");n.d(e,"stackOrderAscending",function(){return z.default});var H=n("./node_modules/taucharts/node_modules/d3-shape/src/order/descending.js");n.d(e,"stackOrderDescending",function(){return H.default});var U=n("./node_modules/taucharts/node_modules/d3-shape/src/order/insideOut.js");n.d(e,"stackOrderInsideOut",function(){return U.default});var W=n("./node_modules/taucharts/node_modules/d3-shape/src/order/none.js");n.d(e,"stackOrderNone",function(){return W.default});var q=n("./node_modules/taucharts/node_modules/d3-shape/src/order/reverse.js");n.d(e,"stackOrderReverse",function(){return q.default})},"./node_modules/taucharts/node_modules/d3-shape/src/line.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/d3-path/index.js"),i=n("./node_modules/taucharts/node_modules/d3-shape/src/constant.js"),o=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/linear.js"),a=n("./node_modules/taucharts/node_modules/d3-shape/src/point.js");e.default=function(){var t=a.x,e=a.y,n=Object(i.default)(!0),s=null,u=o.default,c=null;function l(i){var o,a,l,d=i.length,p=!1;for(null==s&&(c=u(l=Object(r.path)())),o=0;o<=d;++o)!(o<d&&n(a=i[o],o,i))===p&&((p=!p)?c.lineStart():c.lineEnd()),p&&c.point(+t(a,o,i),+e(a,o,i));if(l)return c=null,l+""||null}return l.x=function(e){return arguments.length?(t="function"==typeof e?e:Object(i.default)(+e),l):t},l.y=function(t){return arguments.length?(e="function"==typeof t?t:Object(i.default)(+t),l):e},l.defined=function(t){return arguments.length?(n="function"==typeof t?t:Object(i.default)(!!t),l):n},l.curve=function(t){return arguments.length?(u=t,null!=s&&(c=u(s)),l):u},l.context=function(t){return arguments.length?(null==t?s=c=null:c=u(s=t),l):s},l}},"./node_modules/taucharts/node_modules/d3-shape/src/lineRadial.js":function(t,e,n){"use strict";n.r(e),n.d(e,"lineRadial",function(){return o});var r=n("./node_modules/taucharts/node_modules/d3-shape/src/curve/radial.js"),i=n("./node_modules/taucharts/node_modules/d3-shape/src/line.js");function o(t){var e=t.curve;return t.angle=t.x,delete t.x,t.radius=t.y,delete t.y,t.curve=function(t){return arguments.length?e(Object(r.default)(t)):e()._curve},t}e.default=function(){return o(Object(i.default)().curve(r.curveRadialLinear))}},"./node_modules/taucharts/node_modules/d3-shape/src/link/index.js":function(t,e,n){"use strict";n.r(e),n.d(e,"linkHorizontal",function(){return h}),n.d(e,"linkVertical",function(){return m}),n.d(e,"linkRadial",function(){return g});var r=n("./node_modules/d3-path/index.js"),i=n("./node_modules/taucharts/node_modules/d3-shape/src/array.js"),o=n("./node_modules/taucharts/node_modules/d3-shape/src/constant.js"),a=n("./node_modules/taucharts/node_modules/d3-shape/src/point.js"),s=n("./node_modules/taucharts/node_modules/d3-shape/src/pointRadial.js");function u(t){return t.source}function c(t){return t.target}function l(t){var e=u,n=c,s=a.x,l=a.y,d=null;function p(){var o,a=i.slice.call(arguments),u=e.apply(this,a),c=n.apply(this,a);if(d||(d=o=Object(r.path)()),t(d,+s.apply(this,(a[0]=u,a)),+l.apply(this,a),+s.apply(this,(a[0]=c,a)),+l.apply(this,a)),o)return d=null,o+""||null}return p.source=function(t){return arguments.length?(e=t,p):e},p.target=function(t){return arguments.length?(n=t,p):n},p.x=function(t){return arguments.length?(s="function"==typeof t?t:Object(o.default)(+t),p):s},p.y=function(t){return arguments.length?(l="function"==typeof t?t:Object(o.default)(+t),p):l},p.context=function(t){return arguments.length?(d=null==t?null:t,p):d},p}function d(t,e,n,r,i){t.moveTo(e,n),t.bezierCurveTo(e=(e+r)/2,n,e,i,r,i)}function p(t,e,n,r,i){t.moveTo(e,n),t.bezierCurveTo(e,n=(n+i)/2,r,n,r,i)}function f(t,e,n,r,i){var o=Object(s.default)(e,n),a=Object(s.default)(e,n=(n+i)/2),u=Object(s.default)(r,n),c=Object(s.default)(r,i);t.moveTo(o[0],o[1]),t.bezierCurveTo(a[0],a[1],u[0],u[1],c[0],c[1])}function h(){return l(d)}function m(){return l(p)}function g(){var t=l(f);return t.angle=t.x,delete t.x,t.radius=t.y,delete t.y,t}},"./node_modules/taucharts/node_modules/d3-shape/src/math.js":function(t,e,n){"use strict";n.r(e),n.d(e,"abs",function(){return r}),n.d(e,"atan2",function(){return i}),n.d(e,"cos",function(){return o}),n.d(e,"max",function(){return a}),n.d(e,"min",function(){return s}),n.d(e,"sin",function(){return u}),n.d(e,"sqrt",function(){return c}),n.d(e,"epsilon",function(){return l}),n.d(e,"pi",function(){return d}),n.d(e,"halfPi",function(){return p}),n.d(e,"tau",function(){return f}),n.d(e,"acos",function(){return h}),n.d(e,"asin",function(){return m});var r=Math.abs,i=Math.atan2,o=Math.cos,a=Math.max,s=Math.min,u=Math.sin,c=Math.sqrt,l=1e-12,d=Math.PI,p=d/2,f=2*d;function h(t){return t>1?0:t<-1?d:Math.acos(t)}function m(t){return t>=1?p:t<=-1?-p:Math.asin(t)}},"./node_modules/taucharts/node_modules/d3-shape/src/noop.js":function(t,e,n){"use strict";n.r(e),e.default=function(){}},"./node_modules/taucharts/node_modules/d3-shape/src/offset/diverging.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){if((s=t.length)>0)for(var n,r,i,o,a,s,u=0,c=t[e[0]].length;u<c;++u)for(o=a=0,n=0;n<s;++n)(i=(r=t[e[n]][u])[1]-r[0])>=0?(r[0]=o,r[1]=o+=i):i<0?(r[1]=a,r[0]=a+=i):r[0]=o}},"./node_modules/taucharts/node_modules/d3-shape/src/offset/expand.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/offset/none.js");e.default=function(t,e){if((i=t.length)>0){for(var n,i,o,a=0,s=t[0].length;a<s;++a){for(o=n=0;n<i;++n)o+=t[n][a][1]||0;if(o)for(n=0;n<i;++n)t[n][a][1]/=o}Object(r.default)(t,e)}}},"./node_modules/taucharts/node_modules/d3-shape/src/offset/none.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){if((i=t.length)>1)for(var n,r,i,o=1,a=t[e[0]],s=a.length;o<i;++o)for(r=a,a=t[e[o]],n=0;n<s;++n)a[n][1]+=a[n][0]=isNaN(r[n][1])?r[n][0]:r[n][1]}},"./node_modules/taucharts/node_modules/d3-shape/src/offset/silhouette.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/offset/none.js");e.default=function(t,e){if((n=t.length)>0){for(var n,i=0,o=t[e[0]],a=o.length;i<a;++i){for(var s=0,u=0;s<n;++s)u+=t[s][i][1]||0;o[i][1]+=o[i][0]=-u/2}Object(r.default)(t,e)}}},"./node_modules/taucharts/node_modules/d3-shape/src/offset/wiggle.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/offset/none.js");e.default=function(t,e){if((o=t.length)>0&&(i=(n=t[e[0]]).length)>0){for(var n,i,o,a=0,s=1;s<i;++s){for(var u=0,c=0,l=0;u<o;++u){for(var d=t[e[u]],p=d[s][1]||0,f=(p-(d[s-1][1]||0))/2,h=0;h<u;++h){var m=t[e[h]];f+=(m[s][1]||0)-(m[s-1][1]||0)}c+=p,l+=f*p}n[s-1][1]+=n[s-1][0]=a,c&&(a-=l/c)}n[s-1][1]+=n[s-1][0]=a,Object(r.default)(t,e)}}},"./node_modules/taucharts/node_modules/d3-shape/src/order/appearance.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/order/none.js");function i(t){for(var e,n=-1,r=0,i=t.length,o=-1/0;++n<i;)(e=+t[n][1])>o&&(o=e,r=n);return r}e.default=function(t){var e=t.map(i);return Object(r.default)(t).sort(function(t,n){return e[t]-e[n]})}},"./node_modules/taucharts/node_modules/d3-shape/src/order/ascending.js":function(t,e,n){"use strict";n.r(e),n.d(e,"sum",function(){return i});var r=n("./node_modules/taucharts/node_modules/d3-shape/src/order/none.js");function i(t){for(var e,n=0,r=-1,i=t.length;++r<i;)(e=+t[r][1])&&(n+=e);return n}e.default=function(t){var e=t.map(i);return Object(r.default)(t).sort(function(t,n){return e[t]-e[n]})}},"./node_modules/taucharts/node_modules/d3-shape/src/order/descending.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/order/ascending.js");e.default=function(t){return Object(r.default)(t).reverse()}},"./node_modules/taucharts/node_modules/d3-shape/src/order/insideOut.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/order/appearance.js"),i=n("./node_modules/taucharts/node_modules/d3-shape/src/order/ascending.js");e.default=function(t){var e,n,o=t.length,a=t.map(i.sum),s=Object(r.default)(t),u=0,c=0,l=[],d=[];for(e=0;e<o;++e)n=s[e],u<c?(u+=a[n],l.push(n)):(c+=a[n],d.push(n));return d.reverse().concat(l)}},"./node_modules/taucharts/node_modules/d3-shape/src/order/none.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){for(var e=t.length,n=new Array(e);--e>=0;)n[e]=e;return n}},"./node_modules/taucharts/node_modules/d3-shape/src/order/reverse.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/order/none.js");e.default=function(t){return Object(r.default)(t).reverse()}},"./node_modules/taucharts/node_modules/d3-shape/src/pie.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/constant.js"),i=n("./node_modules/taucharts/node_modules/d3-shape/src/descending.js"),o=n("./node_modules/taucharts/node_modules/d3-shape/src/identity.js"),a=n("./node_modules/taucharts/node_modules/d3-shape/src/math.js");e.default=function(){var t=o.default,e=i.default,n=null,s=Object(r.default)(0),u=Object(r.default)(a.tau),c=Object(r.default)(0);function l(r){var i,o,l,d,p,f=r.length,h=0,m=new Array(f),g=new Array(f),y=+s.apply(this,arguments),_=Math.min(a.tau,Math.max(-a.tau,u.apply(this,arguments)-y)),v=Math.min(Math.abs(_)/f,c.apply(this,arguments)),b=v*(_<0?-1:1);for(i=0;i<f;++i)(p=g[m[i]=i]=+t(r[i],i,r))>0&&(h+=p);for(null!=e?m.sort(function(t,n){return e(g[t],g[n])}):null!=n&&m.sort(function(t,e){return n(r[t],r[e])}),i=0,l=h?(_-f*b)/h:0;i<f;++i,y=d)o=m[i],d=y+((p=g[o])>0?p*l:0)+b,g[o]={data:r[o],index:i,value:p,startAngle:y,endAngle:d,padAngle:v};return g}return l.value=function(e){return arguments.length?(t="function"==typeof e?e:Object(r.default)(+e),l):t},l.sortValues=function(t){return arguments.length?(e=t,n=null,l):e},l.sort=function(t){return arguments.length?(n=t,e=null,l):n},l.startAngle=function(t){return arguments.length?(s="function"==typeof t?t:Object(r.default)(+t),l):s},l.endAngle=function(t){return arguments.length?(u="function"==typeof t?t:Object(r.default)(+t),l):u},l.padAngle=function(t){return arguments.length?(c="function"==typeof t?t:Object(r.default)(+t),l):c},l}},"./node_modules/taucharts/node_modules/d3-shape/src/point.js":function(t,e,n){"use strict";function r(t){return t[0]}function i(t){return t[1]}n.r(e),n.d(e,"x",function(){return r}),n.d(e,"y",function(){return i})},"./node_modules/taucharts/node_modules/d3-shape/src/pointRadial.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){return[(e=+e)*Math.cos(t-=Math.PI/2),e*Math.sin(t)]}},"./node_modules/taucharts/node_modules/d3-shape/src/stack.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/array.js"),i=n("./node_modules/taucharts/node_modules/d3-shape/src/constant.js"),o=n("./node_modules/taucharts/node_modules/d3-shape/src/offset/none.js"),a=n("./node_modules/taucharts/node_modules/d3-shape/src/order/none.js");function s(t,e){return t[e]}e.default=function(){var t=Object(i.default)([]),e=a.default,n=o.default,u=s;function c(r){var i,o,a=t.apply(this,arguments),s=r.length,c=a.length,l=new Array(c);for(i=0;i<c;++i){for(var d,p=a[i],f=l[i]=new Array(s),h=0;h<s;++h)f[h]=d=[0,+u(r[h],p,h,r)],d.data=r[h];f.key=p}for(i=0,o=e(l);i<c;++i)l[o[i]].index=i;return n(l,o),l}return c.keys=function(e){return arguments.length?(t="function"==typeof e?e:Object(i.default)(r.slice.call(e)),c):t},c.value=function(t){return arguments.length?(u="function"==typeof t?t:Object(i.default)(+t),c):u},c.order=function(t){return arguments.length?(e=null==t?a.default:"function"==typeof t?t:Object(i.default)(r.slice.call(t)),c):e},c.offset=function(t){return arguments.length?(n=null==t?o.default:t,c):n},c}},"./node_modules/taucharts/node_modules/d3-shape/src/symbol.js":function(t,e,n){"use strict";n.r(e),n.d(e,"symbols",function(){return p});var r=n("./node_modules/d3-path/index.js"),i=n("./node_modules/taucharts/node_modules/d3-shape/src/symbol/circle.js"),o=n("./node_modules/taucharts/node_modules/d3-shape/src/symbol/cross.js"),a=n("./node_modules/taucharts/node_modules/d3-shape/src/symbol/diamond.js"),s=n("./node_modules/taucharts/node_modules/d3-shape/src/symbol/star.js"),u=n("./node_modules/taucharts/node_modules/d3-shape/src/symbol/square.js"),c=n("./node_modules/taucharts/node_modules/d3-shape/src/symbol/triangle.js"),l=n("./node_modules/taucharts/node_modules/d3-shape/src/symbol/wye.js"),d=n("./node_modules/taucharts/node_modules/d3-shape/src/constant.js"),p=[i.default,o.default,a.default,u.default,s.default,c.default,l.default];e.default=function(){var t=Object(d.default)(i.default),e=Object(d.default)(64),n=null;function o(){var i;if(n||(n=i=Object(r.path)()),t.apply(this,arguments).draw(n,+e.apply(this,arguments)),i)return n=null,i+""||null}return o.type=function(e){return arguments.length?(t="function"==typeof e?e:Object(d.default)(e),o):t},o.size=function(t){return arguments.length?(e="function"==typeof t?t:Object(d.default)(+t),o):e},o.context=function(t){return arguments.length?(n=null==t?null:t,o):n},o}},"./node_modules/taucharts/node_modules/d3-shape/src/symbol/circle.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/math.js");e.default={draw:function(t,e){var n=Math.sqrt(e/r.pi);t.moveTo(n,0),t.arc(0,0,n,0,r.tau)}}},"./node_modules/taucharts/node_modules/d3-shape/src/symbol/cross.js":function(t,e,n){"use strict";n.r(e),e.default={draw:function(t,e){var n=Math.sqrt(e/5)/2;t.moveTo(-3*n,-n),t.lineTo(-n,-n),t.lineTo(-n,-3*n),t.lineTo(n,-3*n),t.lineTo(n,-n),t.lineTo(3*n,-n),t.lineTo(3*n,n),t.lineTo(n,n),t.lineTo(n,3*n),t.lineTo(-n,3*n),t.lineTo(-n,n),t.lineTo(-3*n,n),t.closePath()}}},"./node_modules/taucharts/node_modules/d3-shape/src/symbol/diamond.js":function(t,e,n){"use strict";n.r(e);var r=Math.sqrt(1/3),i=2*r;e.default={draw:function(t,e){var n=Math.sqrt(e/i),o=n*r;t.moveTo(0,-n),t.lineTo(o,0),t.lineTo(0,n),t.lineTo(-o,0),t.closePath()}}},"./node_modules/taucharts/node_modules/d3-shape/src/symbol/square.js":function(t,e,n){"use strict";n.r(e),e.default={draw:function(t,e){var n=Math.sqrt(e),r=-n/2;t.rect(r,r,n,n)}}},"./node_modules/taucharts/node_modules/d3-shape/src/symbol/star.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-shape/src/math.js"),i=Math.sin(r.pi/10)/Math.sin(7*r.pi/10),o=Math.sin(r.tau/10)*i,a=-Math.cos(r.tau/10)*i;e.default={draw:function(t,e){var n=Math.sqrt(.8908130915292852*e),i=o*n,s=a*n;t.moveTo(0,-n),t.lineTo(i,s);for(var u=1;u<5;++u){var c=r.tau*u/5,l=Math.cos(c),d=Math.sin(c);t.lineTo(d*n,-l*n),t.lineTo(l*i-d*s,d*i+l*s)}t.closePath()}}},"./node_modules/taucharts/node_modules/d3-shape/src/symbol/triangle.js":function(t,e,n){"use strict";n.r(e);var r=Math.sqrt(3);e.default={draw:function(t,e){var n=-Math.sqrt(e/(3*r));t.moveTo(0,2*n),t.lineTo(-r*n,-n),t.lineTo(r*n,-n),t.closePath()}}},"./node_modules/taucharts/node_modules/d3-shape/src/symbol/wye.js":function(t,e,n){"use strict";n.r(e);var r=Math.sqrt(3)/2,i=1/Math.sqrt(12),o=3*(i/2+1);e.default={draw:function(t,e){var n=Math.sqrt(e/o),a=n/2,s=n*i,u=a,c=n*i+n,l=-u,d=c;t.moveTo(a,s),t.lineTo(u,c),t.lineTo(l,d),t.lineTo(-.5*a-r*s,r*a+-.5*s),t.lineTo(-.5*u-r*c,r*u+-.5*c),t.lineTo(-.5*l-r*d,r*l+-.5*d),t.lineTo(-.5*a+r*s,-.5*s-r*a),t.lineTo(-.5*u+r*c,-.5*c-r*u),t.lineTo(-.5*l+r*d,-.5*d-r*l),t.closePath()}}},"./node_modules/taucharts/node_modules/d3-time-format/src/defaultLocale.js":function(t,e,n){"use strict";n.r(e),n.d(e,"timeFormat",function(){return i}),n.d(e,"timeParse",function(){return o}),n.d(e,"utcFormat",function(){return a}),n.d(e,"utcParse",function(){return s}),n.d(e,"default",function(){return c});var r,i,o,a,s,u=n("./node_modules/taucharts/node_modules/d3-time-format/src/locale.js");function c(t){return r=Object(u.default)(t),i=r.format,o=r.parse,a=r.utcFormat,s=r.utcParse,r}c({dateTime:"%x, %X",date:"%-m/%-d/%Y",time:"%-I:%M:%S %p",periods:["AM","PM"],days:["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"],shortDays:["Sun","Mon","Tue","Wed","Thu","Fri","Sat"],months:["January","February","March","April","May","June","July","August","September","October","November","December"],shortMonths:["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"]})},"./node_modules/taucharts/node_modules/d3-time-format/src/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-time-format/src/defaultLocale.js");n.d(e,"timeFormatDefaultLocale",function(){return r.default}),n.d(e,"timeFormat",function(){return r.timeFormat}),n.d(e,"timeParse",function(){return r.timeParse}),n.d(e,"utcFormat",function(){return r.utcFormat}),n.d(e,"utcParse",function(){return r.utcParse});var i=n("./node_modules/taucharts/node_modules/d3-time-format/src/locale.js");n.d(e,"timeFormatLocale",function(){return i.default});var o=n("./node_modules/taucharts/node_modules/d3-time-format/src/isoFormat.js");n.d(e,"isoFormat",function(){return o.default});var a=n("./node_modules/taucharts/node_modules/d3-time-format/src/isoParse.js");n.d(e,"isoParse",function(){return a.default})},"./node_modules/taucharts/node_modules/d3-time-format/src/isoFormat.js":function(t,e,n){"use strict";n.r(e),n.d(e,"isoSpecifier",function(){return i});var r=n("./node_modules/taucharts/node_modules/d3-time-format/src/defaultLocale.js"),i="%Y-%m-%dT%H:%M:%S.%LZ";var o=Date.prototype.toISOString?function(t){return t.toISOString()}:Object(r.utcFormat)(i);e.default=o},"./node_modules/taucharts/node_modules/d3-time-format/src/isoParse.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-time-format/src/isoFormat.js"),i=n("./node_modules/taucharts/node_modules/d3-time-format/src/defaultLocale.js");var o=+new Date("2000-01-01T00:00:00.000Z")?function(t){var e=new Date(t);return isNaN(e)?null:e}:Object(i.utcParse)(r.isoSpecifier);e.default=o},"./node_modules/taucharts/node_modules/d3-time-format/src/locale.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return s});var r=n("./node_modules/taucharts/node_modules/d3-time/src/index.js");function i(t){if(0<=t.y&&t.y<100){var e=new Date(-1,t.m,t.d,t.H,t.M,t.S,t.L);return e.setFullYear(t.y),e}return new Date(t.y,t.m,t.d,t.H,t.M,t.S,t.L)}function o(t){if(0<=t.y&&t.y<100){var e=new Date(Date.UTC(-1,t.m,t.d,t.H,t.M,t.S,t.L));return e.setUTCFullYear(t.y),e}return new Date(Date.UTC(t.y,t.m,t.d,t.H,t.M,t.S,t.L))}function a(t){return{y:t,m:0,d:1,H:0,M:0,S:0,L:0}}function s(t){var e=t.dateTime,n=t.date,s=t.time,c=t.periods,l=t.days,d=t.shortDays,p=t.months,f=t.shortMonths,_t=h(c),vt=m(c),bt=h(l),xt=m(l),wt=h(d),jt=m(d),Ot=h(p),Ct=m(p),Et=h(f),St=m(f),Mt={a:function(t){return d[t.getDay()]},A:function(t){return l[t.getDay()]},b:function(t){return f[t.getMonth()]},B:function(t){return p[t.getMonth()]},c:null,d:R,e:R,f:V,H:N,I:F,j:L,L:B,m:z,M:H,p:function(t){return c[+(t.getHours()>=12)]},Q:gt,s:yt,S:U,u:W,U:q,V:G,w:Y,W:K,x:null,X:null,y:X,Y:Q,Z:Z,"%":mt},Tt={a:function(t){return d[t.getUTCDay()]},A:function(t){return l[t.getUTCDay()]},b:function(t){return f[t.getUTCMonth()]},B:function(t){return p[t.getUTCMonth()]},c:null,d:$,e:$,f:rt,H:J,I:tt,j:et,L:nt,m:it,M:ot,p:function(t){return c[+(t.getUTCHours()>=12)]},Q:gt,s:yt,S:at,u:st,U:ut,V:ct,w:lt,W:dt,x:null,X:null,y:pt,Y:ft,Z:ht,"%":mt},kt={a:function(t,e,n){var r=wt.exec(e.slice(n));return r?(t.w=jt[r[0].toLowerCase()],n+r[0].length):-1},A:function(t,e,n){var r=bt.exec(e.slice(n));return r?(t.w=xt[r[0].toLowerCase()],n+r[0].length):-1},b:function(t,e,n){var r=Et.exec(e.slice(n));return r?(t.m=St[r[0].toLowerCase()],n+r[0].length):-1},B:function(t,e,n){var r=Ot.exec(e.slice(n));return r?(t.m=Ct[r[0].toLowerCase()],n+r[0].length):-1},c:function(t,n,r){return It(t,e,n,r)},d:C,e:C,f:A,H:S,I:S,j:E,L:k,m:O,M:M,p:function(t,e,n){var r=_t.exec(e.slice(n));return r?(t.p=vt[r[0].toLowerCase()],n+r[0].length):-1},Q:I,s:P,S:T,u:y,U:_,V:v,w:g,W:b,x:function(t,e,r){return It(t,n,e,r)},X:function(t,e,n){return It(t,s,e,n)},y:w,Y:x,Z:j,"%":D};function At(t,e){return function(n){var r,i,o,a=[],s=-1,c=0,l=t.length;for(n instanceof Date||(n=new Date(+n));++s<l;)37===t.charCodeAt(s)&&(a.push(t.slice(c,s)),null!=(i=u[r=t.charAt(++s)])?r=t.charAt(++s):i="e"===r?" ":"0",(o=e[r])&&(r=o(n,i)),a.push(r),c=s+1);return a.push(t.slice(c,s)),a.join("")}}function Dt(t,e){return function(n){var i,s,u=a(1900);if(It(u,t,n+="",0)!=n.length)return null;if("Q"in u)return new Date(u.Q);if("p"in u&&(u.H=u.H%12+12*u.p),"V"in u){if(u.V<1||u.V>53)return null;"w"in u||(u.w=1),"Z"in u?(i=(s=(i=o(a(u.y))).getUTCDay())>4||0===s?r.utcMonday.ceil(i):Object(r.utcMonday)(i),i=r.utcDay.offset(i,7*(u.V-1)),u.y=i.getUTCFullYear(),u.m=i.getUTCMonth(),u.d=i.getUTCDate()+(u.w+6)%7):(i=(s=(i=e(a(u.y))).getDay())>4||0===s?r.timeMonday.ceil(i):Object(r.timeMonday)(i),i=r.timeDay.offset(i,7*(u.V-1)),u.y=i.getFullYear(),u.m=i.getMonth(),u.d=i.getDate()+(u.w+6)%7)}else("W"in u||"U"in u)&&("w"in u||(u.w="u"in u?u.u%7:"W"in u?1:0),s="Z"in u?o(a(u.y)).getUTCDay():e(a(u.y)).getDay(),u.m=0,u.d="W"in u?(u.w+6)%7+7*u.W-(s+5)%7:u.w+7*u.U-(s+6)%7);return"Z"in u?(u.H+=u.Z/100|0,u.M+=u.Z%100,o(u)):e(u)}}function It(t,e,n,r){for(var i,o,a=0,s=e.length,c=n.length;a<s;){if(r>=c)return-1;if(37===(i=e.charCodeAt(a++))){if(i=e.charAt(a++),!(o=kt[i in u?e.charAt(a++):i])||(r=o(t,n,r))<0)return-1}else if(i!=n.charCodeAt(r++))return-1}return r}return Mt.x=At(n,Mt),Mt.X=At(s,Mt),Mt.c=At(e,Mt),Tt.x=At(n,Tt),Tt.X=At(s,Tt),Tt.c=At(e,Tt),{format:function(t){var e=At(t+="",Mt);return e.toString=function(){return t},e},parse:function(t){var e=Dt(t+="",i);return e.toString=function(){return t},e},utcFormat:function(t){var e=At(t+="",Tt);return e.toString=function(){return t},e},utcParse:function(t){var e=Dt(t,o);return e.toString=function(){return t},e}}}var u={"-":"",_:" ",0:"0"},c=/^\s*\d+/,l=/^%/,d=/[\\^$*+?|[\]().{}]/g;function p(t,e,n){var r=t<0?"-":"",i=(r?-t:t)+"",o=i.length;return r+(o<n?new Array(n-o+1).join(e)+i:i)}function f(t){return t.replace(d,"\\$&")}function h(t){return new RegExp("^(?:"+t.map(f).join("|")+")","i")}function m(t){for(var e={},n=-1,r=t.length;++n<r;)e[t[n].toLowerCase()]=n;return e}function g(t,e,n){var r=c.exec(e.slice(n,n+1));return r?(t.w=+r[0],n+r[0].length):-1}function y(t,e,n){var r=c.exec(e.slice(n,n+1));return r?(t.u=+r[0],n+r[0].length):-1}function _(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.U=+r[0],n+r[0].length):-1}function v(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.V=+r[0],n+r[0].length):-1}function b(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.W=+r[0],n+r[0].length):-1}function x(t,e,n){var r=c.exec(e.slice(n,n+4));return r?(t.y=+r[0],n+r[0].length):-1}function w(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.y=+r[0]+(+r[0]>68?1900:2e3),n+r[0].length):-1}function j(t,e,n){var r=/^(Z)|([+-]\d\d)(?::?(\d\d))?/.exec(e.slice(n,n+6));return r?(t.Z=r[1]?0:-(r[2]+(r[3]||"00")),n+r[0].length):-1}function O(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.m=r[0]-1,n+r[0].length):-1}function C(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.d=+r[0],n+r[0].length):-1}function E(t,e,n){var r=c.exec(e.slice(n,n+3));return r?(t.m=0,t.d=+r[0],n+r[0].length):-1}function S(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.H=+r[0],n+r[0].length):-1}function M(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.M=+r[0],n+r[0].length):-1}function T(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.S=+r[0],n+r[0].length):-1}function k(t,e,n){var r=c.exec(e.slice(n,n+3));return r?(t.L=+r[0],n+r[0].length):-1}function A(t,e,n){var r=c.exec(e.slice(n,n+6));return r?(t.L=Math.floor(r[0]/1e3),n+r[0].length):-1}function D(t,e,n){var r=l.exec(e.slice(n,n+1));return r?n+r[0].length:-1}function I(t,e,n){var r=c.exec(e.slice(n));return r?(t.Q=+r[0],n+r[0].length):-1}function P(t,e,n){var r=c.exec(e.slice(n));return r?(t.Q=1e3*+r[0],n+r[0].length):-1}function R(t,e){return p(t.getDate(),e,2)}function N(t,e){return p(t.getHours(),e,2)}function F(t,e){return p(t.getHours()%12||12,e,2)}function L(t,e){return p(1+r.timeDay.count(Object(r.timeYear)(t),t),e,3)}function B(t,e){return p(t.getMilliseconds(),e,3)}function V(t,e){return B(t,e)+"000"}function z(t,e){return p(t.getMonth()+1,e,2)}function H(t,e){return p(t.getMinutes(),e,2)}function U(t,e){return p(t.getSeconds(),e,2)}function W(t){var e=t.getDay();return 0===e?7:e}function q(t,e){return p(r.timeSunday.count(Object(r.timeYear)(t),t),e,2)}function G(t,e){var n=t.getDay();return t=n>=4||0===n?Object(r.timeThursday)(t):r.timeThursday.ceil(t),p(r.timeThursday.count(Object(r.timeYear)(t),t)+(4===Object(r.timeYear)(t).getDay()),e,2)}function Y(t){return t.getDay()}function K(t,e){return p(r.timeMonday.count(Object(r.timeYear)(t),t),e,2)}function X(t,e){return p(t.getFullYear()%100,e,2)}function Q(t,e){return p(t.getFullYear()%1e4,e,4)}function Z(t){var e=t.getTimezoneOffset();return(e>0?"-":(e*=-1,"+"))+p(e/60|0,"0",2)+p(e%60,"0",2)}function $(t,e){return p(t.getUTCDate(),e,2)}function J(t,e){return p(t.getUTCHours(),e,2)}function tt(t,e){return p(t.getUTCHours()%12||12,e,2)}function et(t,e){return p(1+r.utcDay.count(Object(r.utcYear)(t),t),e,3)}function nt(t,e){return p(t.getUTCMilliseconds(),e,3)}function rt(t,e){return nt(t,e)+"000"}function it(t,e){return p(t.getUTCMonth()+1,e,2)}function ot(t,e){return p(t.getUTCMinutes(),e,2)}function at(t,e){return p(t.getUTCSeconds(),e,2)}function st(t){var e=t.getUTCDay();return 0===e?7:e}function ut(t,e){return p(r.utcSunday.count(Object(r.utcYear)(t),t),e,2)}function ct(t,e){var n=t.getUTCDay();return t=n>=4||0===n?Object(r.utcThursday)(t):r.utcThursday.ceil(t),p(r.utcThursday.count(Object(r.utcYear)(t),t)+(4===Object(r.utcYear)(t).getUTCDay()),e,2)}function lt(t){return t.getUTCDay()}function dt(t,e){return p(r.utcMonday.count(Object(r.utcYear)(t),t),e,2)}function pt(t,e){return p(t.getUTCFullYear()%100,e,2)}function ft(t,e){return p(t.getUTCFullYear()%1e4,e,4)}function ht(){return"+0000"}function mt(){return"%"}function gt(t){return+t}function yt(t){return Math.floor(+t/1e3)}},"./node_modules/taucharts/node_modules/d3-time/src/day.js":function(t,e,n){"use strict";n.r(e),n.d(e,"days",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-time/src/interval.js"),i=n("./node_modules/taucharts/node_modules/d3-time/src/duration.js"),o=Object(r.default)(function(t){t.setHours(0,0,0,0)},function(t,e){t.setDate(t.getDate()+e)},function(t,e){return(e-t-(e.getTimezoneOffset()-t.getTimezoneOffset())*i.durationMinute)/i.durationDay},function(t){return t.getDate()-1});e.default=o;var a=o.range},"./node_modules/taucharts/node_modules/d3-time/src/duration.js":function(t,e,n){"use strict";n.r(e),n.d(e,"durationSecond",function(){return r}),n.d(e,"durationMinute",function(){return i}),n.d(e,"durationHour",function(){return o}),n.d(e,"durationDay",function(){return a}),n.d(e,"durationWeek",function(){return s});var r=1e3,i=6e4,o=36e5,a=864e5,s=6048e5},"./node_modules/taucharts/node_modules/d3-time/src/hour.js":function(t,e,n){"use strict";n.r(e),n.d(e,"hours",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-time/src/interval.js"),i=n("./node_modules/taucharts/node_modules/d3-time/src/duration.js"),o=Object(r.default)(function(t){t.setTime(t-t.getMilliseconds()-t.getSeconds()*i.durationSecond-t.getMinutes()*i.durationMinute)},function(t,e){t.setTime(+t+e*i.durationHour)},function(t,e){return(e-t)/i.durationHour},function(t){return t.getHours()});e.default=o;var a=o.range},"./node_modules/taucharts/node_modules/d3-time/src/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/d3-time/src/interval.js");n.d(e,"timeInterval",function(){return r.default});var i=n("./node_modules/taucharts/node_modules/d3-time/src/millisecond.js");n.d(e,"timeMillisecond",function(){return i.default}),n.d(e,"timeMilliseconds",function(){return i.milliseconds}),n.d(e,"utcMillisecond",function(){return i.default}),n.d(e,"utcMilliseconds",function(){return i.milliseconds});var o=n("./node_modules/taucharts/node_modules/d3-time/src/second.js");n.d(e,"timeSecond",function(){return o.default}),n.d(e,"timeSeconds",function(){return o.seconds}),n.d(e,"utcSecond",function(){return o.default}),n.d(e,"utcSeconds",function(){return o.seconds});var a=n("./node_modules/taucharts/node_modules/d3-time/src/minute.js");n.d(e,"timeMinute",function(){return a.default}),n.d(e,"timeMinutes",function(){return a.minutes});var s=n("./node_modules/taucharts/node_modules/d3-time/src/hour.js");n.d(e,"timeHour",function(){return s.default}),n.d(e,"timeHours",function(){return s.hours});var u=n("./node_modules/taucharts/node_modules/d3-time/src/day.js");n.d(e,"timeDay",function(){return u.default}),n.d(e,"timeDays",function(){return u.days});var c=n("./node_modules/taucharts/node_modules/d3-time/src/week.js");n.d(e,"timeWeek",function(){return c.sunday}),n.d(e,"timeWeeks",function(){return c.sundays}),n.d(e,"timeSunday",function(){return c.sunday}),n.d(e,"timeSundays",function(){return c.sundays}),n.d(e,"timeMonday",function(){return c.monday}),n.d(e,"timeMondays",function(){return c.mondays}),n.d(e,"timeTuesday",function(){return c.tuesday}),n.d(e,"timeTuesdays",function(){return c.tuesdays}),n.d(e,"timeWednesday",function(){return c.wednesday}),n.d(e,"timeWednesdays",function(){return c.wednesdays}),n.d(e,"timeThursday",function(){return c.thursday}),n.d(e,"timeThursdays",function(){return c.thursdays}),n.d(e,"timeFriday",function(){return c.friday}),n.d(e,"timeFridays",function(){return c.fridays}),n.d(e,"timeSaturday",function(){return c.saturday}),n.d(e,"timeSaturdays",function(){return c.saturdays});var l=n("./node_modules/taucharts/node_modules/d3-time/src/month.js");n.d(e,"timeMonth",function(){return l.default}),n.d(e,"timeMonths",function(){return l.months});var d=n("./node_modules/taucharts/node_modules/d3-time/src/year.js");n.d(e,"timeYear",function(){return d.default}),n.d(e,"timeYears",function(){return d.years});var p=n("./node_modules/taucharts/node_modules/d3-time/src/utcMinute.js");n.d(e,"utcMinute",function(){return p.default}),n.d(e,"utcMinutes",function(){return p.utcMinutes});var f=n("./node_modules/taucharts/node_modules/d3-time/src/utcHour.js");n.d(e,"utcHour",function(){return f.default}),n.d(e,"utcHours",function(){return f.utcHours});var h=n("./node_modules/taucharts/node_modules/d3-time/src/utcDay.js");n.d(e,"utcDay",function(){return h.default}),n.d(e,"utcDays",function(){return h.utcDays});var m=n("./node_modules/taucharts/node_modules/d3-time/src/utcWeek.js");n.d(e,"utcWeek",function(){return m.utcSunday}),n.d(e,"utcWeeks",function(){return m.utcSundays}),n.d(e,"utcSunday",function(){return m.utcSunday}),n.d(e,"utcSundays",function(){return m.utcSundays}),n.d(e,"utcMonday",function(){return m.utcMonday}),n.d(e,"utcMondays",function(){return m.utcMondays}),n.d(e,"utcTuesday",function(){return m.utcTuesday}),n.d(e,"utcTuesdays",function(){return m.utcTuesdays}),n.d(e,"utcWednesday",function(){return m.utcWednesday}),n.d(e,"utcWednesdays",function(){return m.utcWednesdays}),n.d(e,"utcThursday",function(){return m.utcThursday}),n.d(e,"utcThursdays",function(){return m.utcThursdays}),n.d(e,"utcFriday",function(){return m.utcFriday}),n.d(e,"utcFridays",function(){return m.utcFridays}),n.d(e,"utcSaturday",function(){return m.utcSaturday}),n.d(e,"utcSaturdays",function(){return m.utcSaturdays});var g=n("./node_modules/taucharts/node_modules/d3-time/src/utcMonth.js");n.d(e,"utcMonth",function(){return g.default}),n.d(e,"utcMonths",function(){return g.utcMonths});var y=n("./node_modules/taucharts/node_modules/d3-time/src/utcYear.js");n.d(e,"utcYear",function(){return y.default}),n.d(e,"utcYears",function(){return y.utcYears})},"./node_modules/taucharts/node_modules/d3-time/src/interval.js":function(t,e,n){"use strict";n.r(e),n.d(e,"default",function(){return o});var r=new Date,i=new Date;function o(t,e,n,a){function s(e){return t(e=new Date(+e)),e}return s.floor=s,s.ceil=function(n){return t(n=new Date(n-1)),e(n,1),t(n),n},s.round=function(t){var e=s(t),n=s.ceil(t);return t-e<n-t?e:n},s.offset=function(t,n){return e(t=new Date(+t),null==n?1:Math.floor(n)),t},s.range=function(n,r,i){var o,a=[];if(n=s.ceil(n),i=null==i?1:Math.floor(i),!(n<r&&i>0))return a;do{a.push(o=new Date(+n)),e(n,i),t(n)}while(o<n&&n<r);return a},s.filter=function(n){return o(function(e){if(e>=e)for(;t(e),!n(e);)e.setTime(e-1)},function(t,r){if(t>=t)if(r<0)for(;++r<=0;)for(;e(t,-1),!n(t););else for(;--r>=0;)for(;e(t,1),!n(t););})},n&&(s.count=function(e,o){return r.setTime(+e),i.setTime(+o),t(r),t(i),Math.floor(n(r,i))},s.every=function(t){return t=Math.floor(t),isFinite(t)&&t>0?t>1?s.filter(a?function(e){return a(e)%t==0}:function(e){return s.count(0,e)%t==0}):s:null}),s}},"./node_modules/taucharts/node_modules/d3-time/src/millisecond.js":function(t,e,n){"use strict";n.r(e),n.d(e,"milliseconds",function(){return o});var r=n("./node_modules/taucharts/node_modules/d3-time/src/interval.js"),i=Object(r.default)(function(){},function(t,e){t.setTime(+t+e)},function(t,e){return e-t});i.every=function(t){return t=Math.floor(t),isFinite(t)&&t>0?t>1?Object(r.default)(function(e){e.setTime(Math.floor(e/t)*t)},function(e,n){e.setTime(+e+n*t)},function(e,n){return(n-e)/t}):i:null},e.default=i;var o=i.range},"./node_modules/taucharts/node_modules/d3-time/src/minute.js":function(t,e,n){"use strict";n.r(e),n.d(e,"minutes",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-time/src/interval.js"),i=n("./node_modules/taucharts/node_modules/d3-time/src/duration.js"),o=Object(r.default)(function(t){t.setTime(t-t.getMilliseconds()-t.getSeconds()*i.durationSecond)},function(t,e){t.setTime(+t+e*i.durationMinute)},function(t,e){return(e-t)/i.durationMinute},function(t){return t.getMinutes()});e.default=o;var a=o.range},"./node_modules/taucharts/node_modules/d3-time/src/month.js":function(t,e,n){"use strict";n.r(e),n.d(e,"months",function(){return o});var r=n("./node_modules/taucharts/node_modules/d3-time/src/interval.js"),i=Object(r.default)(function(t){t.setDate(1),t.setHours(0,0,0,0)},function(t,e){t.setMonth(t.getMonth()+e)},function(t,e){return e.getMonth()-t.getMonth()+12*(e.getFullYear()-t.getFullYear())},function(t){return t.getMonth()});e.default=i;var o=i.range},"./node_modules/taucharts/node_modules/d3-time/src/second.js":function(t,e,n){"use strict";n.r(e),n.d(e,"seconds",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-time/src/interval.js"),i=n("./node_modules/taucharts/node_modules/d3-time/src/duration.js"),o=Object(r.default)(function(t){t.setTime(t-t.getMilliseconds())},function(t,e){t.setTime(+t+e*i.durationSecond)},function(t,e){return(e-t)/i.durationSecond},function(t){return t.getUTCSeconds()});e.default=o;var a=o.range},"./node_modules/taucharts/node_modules/d3-time/src/utcDay.js":function(t,e,n){"use strict";n.r(e),n.d(e,"utcDays",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-time/src/interval.js"),i=n("./node_modules/taucharts/node_modules/d3-time/src/duration.js"),o=Object(r.default)(function(t){t.setUTCHours(0,0,0,0)},function(t,e){t.setUTCDate(t.getUTCDate()+e)},function(t,e){return(e-t)/i.durationDay},function(t){return t.getUTCDate()-1});e.default=o;var a=o.range},"./node_modules/taucharts/node_modules/d3-time/src/utcHour.js":function(t,e,n){"use strict";n.r(e),n.d(e,"utcHours",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-time/src/interval.js"),i=n("./node_modules/taucharts/node_modules/d3-time/src/duration.js"),o=Object(r.default)(function(t){t.setUTCMinutes(0,0,0)},function(t,e){t.setTime(+t+e*i.durationHour)},function(t,e){return(e-t)/i.durationHour},function(t){return t.getUTCHours()});e.default=o;var a=o.range},"./node_modules/taucharts/node_modules/d3-time/src/utcMinute.js":function(t,e,n){"use strict";n.r(e),n.d(e,"utcMinutes",function(){return a});var r=n("./node_modules/taucharts/node_modules/d3-time/src/interval.js"),i=n("./node_modules/taucharts/node_modules/d3-time/src/duration.js"),o=Object(r.default)(function(t){t.setUTCSeconds(0,0)},function(t,e){t.setTime(+t+e*i.durationMinute)},function(t,e){return(e-t)/i.durationMinute},function(t){return t.getUTCMinutes()});e.default=o;var a=o.range},"./node_modules/taucharts/node_modules/d3-time/src/utcMonth.js":function(t,e,n){"use strict";n.r(e),n.d(e,"utcMonths",function(){return o});var r=n("./node_modules/taucharts/node_modules/d3-time/src/interval.js"),i=Object(r.default)(function(t){t.setUTCDate(1),t.setUTCHours(0,0,0,0)},function(t,e){t.setUTCMonth(t.getUTCMonth()+e)},function(t,e){return e.getUTCMonth()-t.getUTCMonth()+12*(e.getUTCFullYear()-t.getUTCFullYear())},function(t){return t.getUTCMonth()});e.default=i;var o=i.range},"./node_modules/taucharts/node_modules/d3-time/src/utcWeek.js":function(t,e,n){"use strict";n.r(e),n.d(e,"utcSunday",function(){return a}),n.d(e,"utcMonday",function(){return s}),n.d(e,"utcTuesday",function(){return u}),n.d(e,"utcWednesday",function(){return c}),n.d(e,"utcThursday",function(){return l}),n.d(e,"utcFriday",function(){return d}),n.d(e,"utcSaturday",function(){return p}),n.d(e,"utcSundays",function(){return f}),n.d(e,"utcMondays",function(){return h}),n.d(e,"utcTuesdays",function(){return m}),n.d(e,"utcWednesdays",function(){return g}),n.d(e,"utcThursdays",function(){return y}),n.d(e,"utcFridays",function(){return _}),n.d(e,"utcSaturdays",function(){return v});var r=n("./node_modules/taucharts/node_modules/d3-time/src/interval.js"),i=n("./node_modules/taucharts/node_modules/d3-time/src/duration.js");function o(t){return Object(r.default)(function(e){e.setUTCDate(e.getUTCDate()-(e.getUTCDay()+7-t)%7),e.setUTCHours(0,0,0,0)},function(t,e){t.setUTCDate(t.getUTCDate()+7*e)},function(t,e){return(e-t)/i.durationWeek})}var a=o(0),s=o(1),u=o(2),c=o(3),l=o(4),d=o(5),p=o(6),f=a.range,h=s.range,m=u.range,g=c.range,y=l.range,_=d.range,v=p.range},"./node_modules/taucharts/node_modules/d3-time/src/utcYear.js":function(t,e,n){"use strict";n.r(e),n.d(e,"utcYears",function(){return o});var r=n("./node_modules/taucharts/node_modules/d3-time/src/interval.js"),i=Object(r.default)(function(t){t.setUTCMonth(0,1),t.setUTCHours(0,0,0,0)},function(t,e){t.setUTCFullYear(t.getUTCFullYear()+e)},function(t,e){return e.getUTCFullYear()-t.getUTCFullYear()},function(t){return t.getUTCFullYear()});i.every=function(t){return isFinite(t=Math.floor(t))&&t>0?Object(r.default)(function(e){e.setUTCFullYear(Math.floor(e.getUTCFullYear()/t)*t),e.setUTCMonth(0,1),e.setUTCHours(0,0,0,0)},function(e,n){e.setUTCFullYear(e.getUTCFullYear()+n*t)}):null},e.default=i;var o=i.range},"./node_modules/taucharts/node_modules/d3-time/src/week.js":function(t,e,n){"use strict";n.r(e),n.d(e,"sunday",function(){return a}),n.d(e,"monday",function(){return s}),n.d(e,"tuesday",function(){return u}),n.d(e,"wednesday",function(){return c}),n.d(e,"thursday",function(){return l}),n.d(e,"friday",function(){return d}),n.d(e,"saturday",function(){return p}),n.d(e,"sundays",function(){return f}),n.d(e,"mondays",function(){return h}),n.d(e,"tuesdays",function(){return m}),n.d(e,"wednesdays",function(){return g}),n.d(e,"thursdays",function(){return y}),n.d(e,"fridays",function(){return _}),n.d(e,"saturdays",function(){return v});var r=n("./node_modules/taucharts/node_modules/d3-time/src/interval.js"),i=n("./node_modules/taucharts/node_modules/d3-time/src/duration.js");function o(t){return Object(r.default)(function(e){e.setDate(e.getDate()-(e.getDay()+7-t)%7),e.setHours(0,0,0,0)},function(t,e){t.setDate(t.getDate()+7*e)},function(t,e){return(e-t-(e.getTimezoneOffset()-t.getTimezoneOffset())*i.durationMinute)/i.durationWeek})}var a=o(0),s=o(1),u=o(2),c=o(3),l=o(4),d=o(5),p=o(6),f=a.range,h=s.range,m=u.range,g=c.range,y=l.range,_=d.range,v=p.range},"./node_modules/taucharts/node_modules/d3-time/src/year.js":function(t,e,n){"use strict";n.r(e),n.d(e,"years",function(){return o});var r=n("./node_modules/taucharts/node_modules/d3-time/src/interval.js"),i=Object(r.default)(function(t){t.setMonth(0,1),t.setHours(0,0,0,0)},function(t,e){t.setFullYear(t.getFullYear()+e)},function(t,e){return e.getFullYear()-t.getFullYear()},function(t){return t.getFullYear()});i.every=function(t){return isFinite(t=Math.floor(t))&&t>0?Object(r.default)(function(e){e.setFullYear(Math.floor(e.getFullYear()/t)*t),e.setMonth(0,1),e.setHours(0,0,0,0)},function(e,n){e.setFullYear(e.getFullYear()+n*t)}):null},e.default=i;var o=i.range},"./node_modules/taucharts/node_modules/topojson-client/index.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/topojson-client/src/bbox.js");n.d(e,"bbox",function(){return r.default});var i=n("./node_modules/taucharts/node_modules/topojson-client/src/feature.js");n.d(e,"feature",function(){return i.default});var o=n("./node_modules/taucharts/node_modules/topojson-client/src/mesh.js");n.d(e,"mesh",function(){return o.default}),n.d(e,"meshArcs",function(){return o.meshArcs});var a=n("./node_modules/taucharts/node_modules/topojson-client/src/merge.js");n.d(e,"merge",function(){return a.default}),n.d(e,"mergeArcs",function(){return a.mergeArcs});var s=n("./node_modules/taucharts/node_modules/topojson-client/src/neighbors.js");n.d(e,"neighbors",function(){return s.default});var u=n("./node_modules/taucharts/node_modules/topojson-client/src/quantize.js");n.d(e,"quantize",function(){return u.default});var c=n("./node_modules/taucharts/node_modules/topojson-client/src/transform.js");n.d(e,"transform",function(){return c.default});var l=n("./node_modules/taucharts/node_modules/topojson-client/src/untransform.js");n.d(e,"untransform",function(){return l.default})},"./node_modules/taucharts/node_modules/topojson-client/src/bbox.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/topojson-client/src/transform.js");e.default=function(t){var e,n=Object(r.default)(t.transform),i=1/0,o=i,a=-i,s=-i;function u(t){(t=n(t))[0]<i&&(i=t[0]),t[0]>a&&(a=t[0]),t[1]<o&&(o=t[1]),t[1]>s&&(s=t[1])}function c(t){switch(t.type){case"GeometryCollection":t.geometries.forEach(c);break;case"Point":u(t.coordinates);break;case"MultiPoint":t.coordinates.forEach(u)}}t.arcs.forEach(function(t){for(var e,r=-1,u=t.length;++r<u;)(e=n(t[r],r))[0]<i&&(i=e[0]),e[0]>a&&(a=e[0]),e[1]<o&&(o=e[1]),e[1]>s&&(s=e[1])});for(e in t.objects)c(t.objects[e]);return[i,o,a,s]}},"./node_modules/taucharts/node_modules/topojson-client/src/bisect.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){for(var n=0,r=t.length;n<r;){var i=n+r>>>1;t[i]<e?n=i+1:r=i}return n}},"./node_modules/taucharts/node_modules/topojson-client/src/feature.js":function(t,e,n){"use strict";n.r(e),n.d(e,"feature",function(){return o}),n.d(e,"object",function(){return a});var r=n("./node_modules/taucharts/node_modules/topojson-client/src/reverse.js"),i=n("./node_modules/taucharts/node_modules/topojson-client/src/transform.js");function o(t,e){var n=e.id,r=e.bbox,i=null==e.properties?{}:e.properties,o=a(t,e);return null==n&&null==r?{type:"Feature",properties:i,geometry:o}:null==r?{type:"Feature",id:n,properties:i,geometry:o}:{type:"Feature",id:n,bbox:r,properties:i,geometry:o}}function a(t,e){var n=Object(i.default)(t.transform),o=t.arcs;function a(t,e){e.length&&e.pop();for(var i=o[t<0?~t:t],a=0,s=i.length;a<s;++a)e.push(n(i[a],a));t<0&&Object(r.default)(e,s)}function s(t){return n(t)}function u(t){for(var e=[],n=0,r=t.length;n<r;++n)a(t[n],e);return e.length<2&&e.push(e[0]),e}function c(t){for(var e=u(t);e.length<4;)e.push(e[0]);return e}function l(t){return t.map(c)}return function t(e){var n,r=e.type;switch(r){case"GeometryCollection":return{type:r,geometries:e.geometries.map(t)};case"Point":n=s(e.coordinates);break;case"MultiPoint":n=e.coordinates.map(s);break;case"LineString":n=u(e.arcs);break;case"MultiLineString":n=e.arcs.map(u);break;case"Polygon":n=l(e.arcs);break;case"MultiPolygon":n=e.arcs.map(l);break;default:return null}return{type:r,coordinates:n}}(e)}e.default=function(t,e){return"GeometryCollection"===e.type?{type:"FeatureCollection",features:e.geometries.map(function(e){return o(t,e)})}:o(t,e)}},"./node_modules/taucharts/node_modules/topojson-client/src/identity.js":function(t,e,n){"use strict";n.r(e),e.default=function(t){return t}},"./node_modules/taucharts/node_modules/topojson-client/src/merge.js":function(t,e,n){"use strict";n.r(e),n.d(e,"mergeArcs",function(){return o});var r=n("./node_modules/taucharts/node_modules/topojson-client/src/feature.js"),i=n("./node_modules/taucharts/node_modules/topojson-client/src/stitch.js");function o(t,e){var n={},o=[],a=[];function s(t){t.forEach(function(e){e.forEach(function(e){(n[e=e<0?~e:e]||(n[e]=[])).push(t)})}),o.push(t)}function u(e){return function(t){for(var e,n=-1,r=t.length,i=t[r-1],o=0;++n<r;)e=i,i=t[n],o+=e[0]*i[1]-e[1]*i[0];return Math.abs(o)}(Object(r.object)(t,{type:"Polygon",arcs:[e]}).coordinates[0])}return e.forEach(function t(e){switch(e.type){case"GeometryCollection":e.geometries.forEach(t);break;case"Polygon":s(e.arcs);break;case"MultiPolygon":e.arcs.forEach(s)}}),o.forEach(function(t){if(!t._){var e=[],r=[t];for(t._=1,a.push(e);t=r.pop();)e.push(t),t.forEach(function(t){t.forEach(function(t){n[t<0?~t:t].forEach(function(t){t._||(t._=1,r.push(t))})})})}}),o.forEach(function(t){delete t._}),{type:"MultiPolygon",arcs:a.map(function(e){var r,o=[];if(e.forEach(function(t){t.forEach(function(t){t.forEach(function(t){n[t<0?~t:t].length<2&&o.push(t)})})}),(r=(o=Object(i.default)(t,o)).length)>1)for(var a,s,c=1,l=u(o[0]);c<r;++c)(a=u(o[c]))>l&&(s=o[0],o[0]=o[c],o[c]=s,l=a);return o})}}e.default=function(t){return Object(r.object)(t,o.apply(this,arguments))}},"./node_modules/taucharts/node_modules/topojson-client/src/mesh.js":function(t,e,n){"use strict";n.r(e),n.d(e,"meshArcs",function(){return o});var r=n("./node_modules/taucharts/node_modules/topojson-client/src/feature.js"),i=n("./node_modules/taucharts/node_modules/topojson-client/src/stitch.js");function o(t,e,n){var r,o,a;if(arguments.length>1)r=function(t,e,n){var r,i=[],o=[];function a(t){var e=t<0?~t:t;(o[e]||(o[e]=[])).push({i:t,g:r})}function s(t){t.forEach(a)}function u(t){t.forEach(s)}return function t(e){switch(r=e,e.type){case"GeometryCollection":e.geometries.forEach(t);break;case"LineString":s(e.arcs);break;case"MultiLineString":case"Polygon":u(e.arcs);break;case"MultiPolygon":e.arcs.forEach(u)}}(e),o.forEach(null==n?function(t){i.push(t[0].i)}:function(t){n(t[0].g,t[t.length-1].g)&&i.push(t[0].i)}),i}(0,e,n);else for(o=0,r=new Array(a=t.arcs.length);o<a;++o)r[o]=o;return{type:"MultiLineString",arcs:Object(i.default)(t,r)}}e.default=function(t){return Object(r.object)(t,o.apply(this,arguments))}},"./node_modules/taucharts/node_modules/topojson-client/src/neighbors.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/topojson-client/src/bisect.js");e.default=function(t){var e={},n=t.map(function(){return[]});function i(t,n){t.forEach(function(t){t<0&&(t=~t);var r=e[t];r?r.push(n):e[t]=[n]})}function o(t,e){t.forEach(function(t){i(t,e)})}var a={LineString:i,MultiLineString:o,Polygon:o,MultiPolygon:function(t,e){t.forEach(function(t){o(t,e)})}};t.forEach(function t(e,n){"GeometryCollection"===e.type?e.geometries.forEach(function(e){t(e,n)}):e.type in a&&a[e.type](e.arcs,n)});for(var s in e)for(var u=e[s],c=u.length,l=0;l<c;++l)for(var d=l+1;d<c;++d){var p,f=u[l],h=u[d];(p=n[f])[s=Object(r.default)(p,h)]!==h&&p.splice(s,0,h),(p=n[h])[s=Object(r.default)(p,f)]!==f&&p.splice(s,0,f)}return n}},"./node_modules/taucharts/node_modules/topojson-client/src/quantize.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/topojson-client/src/bbox.js"),i=n("./node_modules/taucharts/node_modules/topojson-client/src/untransform.js");e.default=function(t,e){if(t.transform)throw new Error("already quantized");if(e&&e.scale)c=t.bbox;else{if(!((n=Math.floor(e))>=2))throw new Error("n must be ≥2");var n,o=(c=t.bbox||Object(r.default)(t))[0],a=c[1],s=c[2],u=c[3];e={scale:[s-o?(s-o)/(n-1):1,u-a?(u-a)/(n-1):1],translate:[o,a]}}var c,l,d=Object(i.default)(e),p=t.objects,f={};function h(t){return d(t)}function m(t){var e;switch(t.type){case"GeometryCollection":e={type:"GeometryCollection",geometries:t.geometries.map(m)};break;case"Point":e={type:"Point",coordinates:h(t.coordinates)};break;case"MultiPoint":e={type:"MultiPoint",coordinates:t.coordinates.map(h)};break;default:return t}return null!=t.id&&(e.id=t.id),null!=t.bbox&&(e.bbox=t.bbox),null!=t.properties&&(e.properties=t.properties),e}for(l in p)f[l]=m(p[l]);return{type:"Topology",bbox:c,transform:e,objects:f,arcs:t.arcs.map(function(t){var e,n=0,r=1,i=t.length,o=new Array(i);for(o[0]=d(t[0],0);++n<i;)((e=d(t[n],n))[0]||e[1])&&(o[r++]=e);return 1===r&&(o[r++]=[0,0]),o.length=r,o})}}},"./node_modules/taucharts/node_modules/topojson-client/src/reverse.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){for(var n,r=t.length,i=r-e;i<--r;)n=t[i],t[i++]=t[r],t[r]=n}},"./node_modules/taucharts/node_modules/topojson-client/src/stitch.js":function(t,e,n){"use strict";n.r(e),e.default=function(t,e){var n={},r={},i={},o=[],a=-1;function s(t,e){for(var r in t){var i=t[r];delete e[i.start],delete i.start,delete i.end,i.forEach(function(t){n[t<0?~t:t]=1}),o.push(i)}}return e.forEach(function(n,r){var i,o=t.arcs[n<0?~n:n];o.length<3&&!o[1][0]&&!o[1][1]&&(i=e[++a],e[a]=n,e[r]=i)}),e.forEach(function(e){var n,o,a=function(e){var n,r=t.arcs[e<0?~e:e],i=r[0];t.transform?(n=[0,0],r.forEach(function(t){n[0]+=t[0],n[1]+=t[1]})):n=r[r.length-1];return e<0?[n,i]:[i,n]}(e),s=a[0],u=a[1];if(n=i[s])if(delete i[n.end],n.push(e),n.end=u,o=r[u]){delete r[o.start];var c=o===n?n:n.concat(o);r[c.start=n.start]=i[c.end=o.end]=c}else r[n.start]=i[n.end]=n;else if(n=r[u])if(delete r[n.start],n.unshift(e),n.start=s,o=i[s]){delete i[o.end];var l=o===n?n:o.concat(n);r[l.start=o.start]=i[l.end=n.end]=l}else r[n.start]=i[n.end]=n;else r[(n=[e]).start=s]=i[n.end=u]=n}),s(i,r),s(r,i),e.forEach(function(t){n[t<0?~t:t]||o.push([t])}),o}},"./node_modules/taucharts/node_modules/topojson-client/src/transform.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/topojson-client/src/identity.js");e.default=function(t){if(null==t)return r.default;var e,n,i=t.scale[0],o=t.scale[1],a=t.translate[0],s=t.translate[1];return function(t,r){r||(e=n=0);var u=2,c=t.length,l=new Array(c);for(l[0]=(e+=t[0])*i+a,l[1]=(n+=t[1])*o+s;u<c;)l[u]=t[u],++u;return l}}},"./node_modules/taucharts/node_modules/topojson-client/src/untransform.js":function(t,e,n){"use strict";n.r(e);var r=n("./node_modules/taucharts/node_modules/topojson-client/src/identity.js");e.default=function(t){if(null==t)return r.default;var e,n,i=t.scale[0],o=t.scale[1],a=t.translate[0],s=t.translate[1];return function(t,r){r||(e=n=0);var u=2,c=t.length,l=new Array(c),d=Math.round((t[0]-a)/i),p=Math.round((t[1]-s)/o);for(l[0]=d-e,e=d,l[1]=p-n,n=p;u<c;)l[u]=t[u],++u;return l}}},"./node_modules/tslib/tslib.es6.js":function(t,e,n){"use strict";n.r(e),n.d(e,"__extends",function(){return i}),n.d(e,"__assign",function(){return o}),n.d(e,"__rest",function(){return a}),n.d(e,"__decorate",function(){return s}),n.d(e,"__param",function(){return u}),n.d(e,"__metadata",function(){return c}),n.d(e,"__awaiter",function(){return l}),n.d(e,"__generator",function(){return d}),n.d(e,"__exportStar",function(){return p}),n.d(e,"__values",function(){return f}),n.d(e,"__read",function(){return h}),n.d(e,"__spread",function(){return m}),n.d(e,"__await",function(){return g}),n.d(e,"__asyncGenerator",function(){return y}),n.d(e,"__asyncDelegator",function(){return _}),n.d(e,"__asyncValues",function(){return v}),n.d(e,"__makeTemplateObject",function(){return b}),n.d(e,"__importStar",function(){return x}),n.d(e,"__importDefault",function(){return w});var r=function(t,e){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)};function i(t,e){function n(){this.constructor=t}r(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}var o=function(){return(o=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++){e=arguments[n];for(var i in e)Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i])}return t}).apply(this,arguments)};function a(t,e){var n={};for(var r in t)Object.prototype.hasOwnProperty.call(t,r)&&e.indexOf(r)<0&&(n[r]=t[r]);if(null!=t&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(t);i<r.length;i++)e.indexOf(r[i])<0&&(n[r[i]]=t[r[i]])}return n}function s(t,e,n,r){var i,o=arguments.length,a=o<3?e:null===r?r=Object.getOwnPropertyDescriptor(e,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(t,e,n,r);else for(var s=t.length-1;s>=0;s--)(i=t[s])&&(a=(o<3?i(a):o>3?i(e,n,a):i(e,n))||a);return o>3&&a&&Object.defineProperty(e,n,a),a}function u(t,e){return function(n,r){e(n,r,t)}}function c(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)}function l(t,e,n,r){return new(n||(n=Promise))(function(i,o){function a(t){try{u(r.next(t))}catch(t){o(t)}}function s(t){try{u(r.throw(t))}catch(t){o(t)}}function u(t){t.done?i(t.value):new n(function(e){e(t.value)}).then(a,s)}u((r=r.apply(t,e||[])).next())})}function d(t,e){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=e.call(t,a)}catch(t){o=[6,t],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(t,e){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}function f(t){var e="function"==typeof Symbol&&t[Symbol.iterator],n=0;return e?e.call(t):{next:function(){return t&&n>=t.length&&(t=void 0),{value:t&&t[n++],done:!t}}}}function h(t,e){var n="function"==typeof Symbol&&t[Symbol.iterator];if(!n)return t;var r,i,o=n.call(t),a=[];try{for(;(void 0===e||e-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(t){i={error:t}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var t=[],e=0;e<arguments.length;e++)t=t.concat(h(arguments[e]));return t}function g(t){return this instanceof g?(this.v=t,this):new g(t)}function y(t,e,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(t,e||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(t){i[t]&&(r[t]=function(e){return new Promise(function(n,r){o.push([t,e,n,r])>1||s(t,e)})})}function s(t,e){try{(n=i[t](e)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(t){l(o[0][3],t)}var n}function u(t){s("next",t)}function c(t){s("throw",t)}function l(t,e){t(e),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(t){var e,n;return e={},r("next"),r("throw",function(t){throw t}),r("return"),e[Symbol.iterator]=function(){return this},e;function r(r,i){e[r]=t[r]?function(e){return(n=!n)?{value:g(t[r](e)),done:"return"===r}:i?i(e):e}:i}}function v(t){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var e,n=t[Symbol.asyncIterator];return n?n.call(t):(t=f(t),e={},r("next"),r("throw"),r("return"),e[Symbol.asyncIterator]=function(){return this},e);function r(n){e[n]=t[n]&&function(e){return new Promise(function(r,i){var o,a,s,u;e=t[n](e),o=r,a=i,s=e.done,u=e.value,Promise.resolve(u).then(function(t){o({value:t,done:s})},a)})}}}function b(t,e){return Object.defineProperty?Object.defineProperty(t,"raw",{value:e}):t.raw=e,t}function x(t){if(t&&t.__esModule)return t;var e={};if(null!=t)for(var n in t)Object.hasOwnProperty.call(t,n)&&(e[n]=t[n]);return e.default=t,e}function w(t){return t&&t.__esModule?t:{default:t}}}}]);
</script><script type="text/javascript">
(window.webpackJsonp=window.webpackJsonp||[]).push([["main"],{"./src/$$_lazy_route_resource lazy recursive":function(t,e){function n(t){return Promise.resolve().then(function(){var e=new Error("Cannot find module '"+t+"'");throw e.code="MODULE_NOT_FOUND",e})}n.keys=function(){return[]},n.resolve=n,t.exports=n,n.id="./src/$$_lazy_route_resource lazy recursive"},"./src/app/accordion/accordion.component.html":function(t,e){t.exports='<div *ngIf="data.byReg(case).length > 1">\n  <mat-expansion-panel\n  [expanded]="data.allExpandState"\n  #expPanel\n  class="mat-elevation-z0">\n     <mat-expansion-panel-header\n     class="mat-expansion-panel-header"\n     collapsedHeight="auto"\n     expandedHeight="auto"\n     style="padding:0.05vh">\n       <mat-panel-title fxLayoutAlign=\'space-between center\'>\n         <td class="test-title">\n           {{ data.byID(data.byReg(case)[0])?.title }}\n         </td>\n         <td>\n           <button mat-button style="color:black"> Info </button>\n         </td>\n       </mat-panel-title>\n     </mat-expansion-panel-header>\n     <ng-template matExpansionPanelContent>\n       <div>\n         <p> {{ data.byID(data.byReg(case)[0])?.summary }} </p>\n       </div>\n     </ng-template>\n </mat-expansion-panel>\n</div>\n  <mat-accordion\n  multi="true"\n  displayMode="flat">\n       <mat-expansion-panel\n       *ngFor="let testId of data.byReg(case); let i = index"\n       #expPanel\n       [expanded]="data.allExpandState"\n       class="mat-elevation-z0">\n          <mat-expansion-panel-header\n          class="mat-expansion-panel-header"\n          collapsedHeight="auto"\n          expandedHeight="auto"\n          style="padding:0.05vh">\n            <mat-panel-title\n            *ngIf="data.byReg(case).length === 1; else noTitle"\n            fxLayoutAlign=\'space-between center\'>\n              <td class="test-title">\n                {{ data.byID(testId)?.title }}\n              </td>\n                <div [ngSwitch]="data.reportType">\n                <td *ngSwitchCase="\'snapshot\'">\n                  <app-result-button [testId]="testId"></app-result-button>\n                </td>\n                  <td *ngSwitchCase="\'diff\'">\n                    <app-diff-button [testId]="testId" ></app-diff-button>\n                  </td>\n              </div>\n            </mat-panel-title>\n            <ng-template #noTitle>\n              <mat-panel-title fxLayoutAlign=\'space-between center\'>\n                <td style="padding-left:1vh"> {{ data.getParam(testId, 1) }} </td>\n                <div [ngSwitch]="data.reportType">\n                <td *ngSwitchCase="\'snapshot\'">\n                  <app-result-button [testId]="testId"></app-result-button>\n                </td>\n                <td *ngSwitchCase="\'diff\'">\n                    <app-diff-button [testId]="testId"></app-diff-button>\n                  </td>\n                </div>\n              </mat-panel-title>\n            </ng-template>\n          <div *ngIf=\'isWeighted(testId) ; else whiteBadge\'> \n            <div class=\'badge\'> <span class="vert-align"> X{{ weight }} </span> </div> \n          </div>\n          <ng-template #whiteBadge>\n              <div class=\'badge\' style="background-color: white"></div> \n          </ng-template>\n          </mat-expansion-panel-header>\n          <ng-template matExpansionPanelContent>\n          <div fxLayout=\'column\' flexLayoutAlign=\'space-between center\'>\n            <div *ngIf="data.byReg(case).length === 1; else justMessages">\n              <p>{{ data.byID(testId)?.summary }}</p>\n              <p>{{ data.byID(testId)?.message }}</p>\n            </div>\n            <ng-template #justMessages>\n              <div>\n                <tr> {{ data.byID(testId)?.message }} </tr>\n              </div>\n            </ng-template>\n            <div [ngSwitch]="data.reportType">\n              <div *ngSwitchCase="\'snapshot\'">\n              <br>\n              <textarea rows="4" style="overflow-wrap: break-word; width:100%">\n                {{ data.getString(data.byID(testId)?.data) }}\n              </textarea>\n              </div>\n              <div *ngSwitchCase="\'diff\'">\n                  <div *ngFor=\'let results of data.byID(testId)?.diff; let i = index\'>\n                    <p> For model "{{ results.model }}" : {{ results.message }} </p>\n                  </div>\n                  <br>\n                  <textarea rows="4" style="overflow-wrap: break-word; width:100%">\n                      {{ data.extractData(testId) }}\n                  </textarea>\n              </div>\n              <div *ngSwitchCase="\'history\'">\n                  <br>\n                    <div>\n                      <app-tauchart-history [testObject]="data.byID(testId)" [testId]="testId" ></app-tauchart-history>\n                    </div>\n              </div>\n             </div>\n          </div>\n        </ng-template>\n      </mat-expansion-panel>\n  </mat-accordion>\n'},"./src/app/accordion/accordion.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL2FjY29yZGlvbi9hY2NvcmRpb24uY29tcG9uZW50LnNjc3MifQ== */"},"./src/app/accordion/accordion.component.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"AccordionComponent",function(){return i});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./src/app/report-data.service.ts"),i=function(){function t(t){this.data=t,this.weight=1}return t.prototype.ngOnInit=function(){},t.prototype.isWeighted=function(t){return t in this.data.testWeightsExpanded&&1!==this.data.testWeightsExpanded[t]&&null!==this.data.testWeightsExpanded[t]&&(this.weight=this.data.testWeightsExpanded[t],!0)},s.__decorate([Object(a.Input)(),s.__metadata("design:type",String)],t.prototype,"case",void 0),t=s.__decorate([Object(a.Component)({selector:"app-accordion",template:n("./src/app/accordion/accordion.component.html"),encapsulation:a.ViewEncapsulation.None,styles:[n("./src/app/accordion/accordion.component.scss")]}),s.__metadata("design:paramtypes",[o.ReportDataService])],t)}()},"./src/app/app.component.html":function(t,e){t.exports="<app-header></app-header>\n<app-dashboard></app-dashboard>\n"},"./src/app/app.component.scss":function(t,e){t.exports=".example-container {\n  display: flex;\n  flex-direction: column;\n  min-width: 300px; }\n\n.mat-table {\n  overflow: auto;\n  max-height: 500px; }\n\n.mat-header-cell.mat-sort-header-sorted {\n  color: black; }\n\n.mat-expansion-panel-content:not(.mat-expanded) {\n  height: 0px;\n  visibility: hidden; }\n\n.mat-expansion-panel-header {\n  height: 48px; }\n\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbIi9ob21lL21vcml0ei9Db2RlYmFzZS9vcGVuY29icmEvbWVtb3RlL21lbW90ZS1yZXBvcnQtYXBwL3NyYy9hcHAvYXBwLmNvbXBvbmVudC5zY3NzIl0sIm5hbWVzIjpbXSwibWFwcGluZ3MiOiJBQUFBO0VBQ0UsYUFBYTtFQUNiLHNCQUFzQjtFQUN0QixnQkFBZ0IsRUFBQTs7QUFHbEI7RUFDRSxjQUFjO0VBQ2QsaUJBQWlCLEVBQUE7O0FBR25CO0VBQ0UsWUFBWSxFQUFBOztBQUdkO0VBQ0ksV0FBVztFQUNYLGtCQUFrQixFQUFBOztBQUd0QjtFQUNJLFlBQVksRUFBQSIsImZpbGUiOiJzcmMvYXBwL2FwcC5jb21wb25lbnQuc2NzcyIsInNvdXJjZXNDb250ZW50IjpbIi5leGFtcGxlLWNvbnRhaW5lciB7XG4gIGRpc3BsYXk6IGZsZXg7XG4gIGZsZXgtZGlyZWN0aW9uOiBjb2x1bW47XG4gIG1pbi13aWR0aDogMzAwcHg7XG59XG5cbi5tYXQtdGFibGUge1xuICBvdmVyZmxvdzogYXV0bztcbiAgbWF4LWhlaWdodDogNTAwcHg7XG59XG5cbi5tYXQtaGVhZGVyLWNlbGwubWF0LXNvcnQtaGVhZGVyLXNvcnRlZCB7XG4gIGNvbG9yOiBibGFjaztcbn1cblxuLm1hdC1leHBhbnNpb24tcGFuZWwtY29udGVudDpub3QoLm1hdC1leHBhbmRlZCkge1xuICAgIGhlaWdodDogMHB4O1xuICAgIHZpc2liaWxpdHk6IGhpZGRlbjtcbn1cblxuLm1hdC1leHBhbnNpb24tcGFuZWwtaGVhZGVyICB7XG4gICAgaGVpZ2h0OiA0OHB4O1xufVxuIl19 */"},"./src/app/app.component.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"AppComponent",function(){return o});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),o=function(){function t(){this.title="app"}return t=s.__decorate([Object(a.Component)({selector:"app-root",template:n("./src/app/app.component.html"),styles:[n("./src/app/app.component.scss")]})],t)}()},"./src/app/app.material.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"AppMaterialModule",function(){return r});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),i=n("./node_modules/@angular/material/esm5/material.es5.js"),r=function(){function t(){}return t=s.__decorate([Object(a.NgModule)({imports:[o.BrowserAnimationsModule,i.MatBadgeModule,i.MatButtonModule,i.MatCheckboxModule,i.MatListModule,i.MatToolbarModule,i.MatIconModule,i.MatFormFieldModule,i.MatInputModule,i.MatCardModule,i.MatSidenavModule,i.MatMenuModule,i.MatAutocompleteModule,i.MatDialogModule,i.MatTableModule,i.MatSortModule,i.MatPaginatorModule,i.MatTooltipModule,i.MatProgressBarModule,i.MatGridListModule,i.MatSelectModule,i.MatExpansionModule],exports:[o.BrowserAnimationsModule,i.MatBadgeModule,i.MatButtonModule,i.MatCheckboxModule,i.MatListModule,i.MatToolbarModule,i.MatIconModule,i.MatFormFieldModule,i.MatInputModule,i.MatCardModule,i.MatSidenavModule,i.MatMenuModule,i.MatAutocompleteModule,i.MatDialogModule,i.MatTableModule,i.MatSortModule,i.MatPaginatorModule,i.MatTooltipModule,i.MatProgressBarModule,i.MatGridListModule,i.MatSelectModule,i.MatExpansionModule]})],t)}()},"./src/app/app.module.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"AppModule",function(){return v});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./src/app/app.material.ts"),r=n("./node_modules/@angular/flex-layout/esm5/flex-layout.es5.js"),c=n("./node_modules/@angular/common/fesm5/http.js"),l=n("./src/app/app.component.ts"),d=n("./src/app/header/header.component.ts"),p=n("./src/app/dashboard/dashboard.component.ts"),u=n("./src/app/dashboard/universal-core/universal-core.component.ts"),m=n("./src/app/dashboard/statistics/statistics.component.ts"),h=n("./src/app/report-data.service.ts"),M=n("./src/app/keys.pipe.ts"),y=n("./src/app/dashboard/system-information/system-information.component.ts"),g=n("./src/app/accordion/accordion.component.ts"),f=n("./src/app/result-button/result-button.component.ts"),b=n("./src/app/diff-button/diff-button.component.ts"),L=n("./src/app/tauchart-history/tauchart-history.component.ts"),x=n("./src/app/tauchart-bar/tauchart-bar.component.ts"),j=n("./src/app/help-dialog/help-dialog.component.ts"),w=n("./src/app/score-formula/score-formula.component.ts"),v=function(){function t(t){this.reportDataService=t,this.reportDataService.loadResults()}return t.prototype.ngOnInit=function(){},t=s.__decorate([Object(o.NgModule)({declarations:[l.AppComponent,d.HeaderComponent,p.DashboardComponent,u.UniversalCoreComponent,m.StatisticsComponent,M.KeysPipe,y.SystemInformationComponent,g.AccordionComponent,f.ResultButtonComponent,b.DiffButtonComponent,L.TauChartHistoryComponent,x.TauChartBarComponent,j.HelpDialogComponent,j.HelpDialogTextComponent,w.ScoreFormulaComponent],imports:[a.BrowserModule,i.AppMaterialModule,r.FlexLayoutModule,c.HttpClientModule],exports:[a.BrowserModule,i.AppMaterialModule],entryComponents:[j.HelpDialogComponent,j.HelpDialogTextComponent],providers:[h.ReportDataService],bootstrap:[l.AppComponent]}),s.__metadata("design:paramtypes",[h.ReportDataService])],t)}()},"./src/app/dashboard/dashboard.component.html":function(t,e){t.exports="<div\n  fxLayout='row wrap'\n  fxLayoutAlign='center stretch'\n  fxLayout.xs='column'>\n  <div\n    fxLayout='column'\n    fxFlex=\"45%\"\n    fxFlexAlign='stretch'\n    style=\"margin:0.5vmax\">\n    <mat-card\n      class=\"mat-elevation-z2\"\n      style=\"background-color:whitesmoke; height: 5vmax\"> \n      <mat-card-title> Independent Section </mat-card-title>\n      <mat-card-subtitle> Contains tests that are independent of the class of modeled organism, a model's complexity or types of identifiers\n          that are used to describe its components. Parameterization or initialization of the network is not required. See readme for more details. </mat-card-subtitle>\n      </mat-card>\n      <mat-card fxFlex *ngIf='data.scoredCard'>\n        <app-universal-core [scoreData]=\"data.scoredCard\"></app-universal-core>\n      </mat-card>\n  </div>\n  <div\n    fxLayout='column'\n    fxFlex=\"45%\"\n    fxFlexAlign='stretch'\n    fxLayoutAlign='start stretch'\n    style=\"margin:0.5vmax\"\n    *ngIf='data'>\n      <mat-card\n      class=\"mat-elevation-z2\"\n      style=\"background-color:whitesmoke; height: 5vmax\"> \n      <mat-card-title> Specific Section </mat-card-title>\n      <mat-card-subtitle> Covers general statistics and specific aspects of a metabolic network that are not universally applicable. See readme for more details. </mat-card-subtitle>\n      </mat-card>\n      <div\n        *ngFor=\"let card of data.statisticsCards\">\n          <app-statistics [currentCard]=\"card\"></app-statistics>\n      </div>\n      <div\n        *ngIf='data.metaData'>\n          <app-system-information></app-system-information>\n      </div>\n   </div>\n</div>\n"},"./src/app/dashboard/dashboard.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL2Rhc2hib2FyZC9kYXNoYm9hcmQuY29tcG9uZW50LnNjc3MifQ== */"},"./src/app/dashboard/dashboard.component.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"DashboardComponent",function(){return i});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./src/app/report-data.service.ts"),i=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},t=s.__decorate([Object(a.Component)({selector:"app-dashboard",template:n("./src/app/dashboard/dashboard.component.html"),encapsulation:a.ViewEncapsulation.None,styles:[n("./src/app/dashboard/dashboard.component.scss")]}),s.__metadata("design:paramtypes",[o.ReportDataService])],t)}()},"./src/app/dashboard/statistics/statistics.component.html":function(t,e){t.exports='<mat-card fxFlex>\n  <mat-card-title>{{ currentCard.cardName }}</mat-card-title>\n    <mat-card-content>\n      <div *ngFor="let case of currentCard.associatedTests; let first = first">\n            <mat-expansion-panel\n            *ngIf="first && data.reportType === \'diff\'"\n            disabled\n            #expPanel\n            class="mat-elevation-z0">\n               <mat-expansion-panel-header\n               class="mat-expansion-panel-header"\n               collapsedHeight="auto"\n               style="padding:0.05vh">\n                 <mat-panel-title fxLayoutAlign=\'space-between center\'>\n                    <td class="test-title">\n                        Compared Models\n                    </td>\n                    <div style="margin-right: 10px">\n                    <td *ngFor="let results of data.byID(case)?.diff">\n                      <div class="MatButtonEmulation" style="background: white; color:black"> {{ results.model }} </div>\n                    </td>\n                    </div>\n                 </mat-panel-title>\n               </mat-expansion-panel-header>\n              </mat-expansion-panel>\n         <app-accordion [case]="case"></app-accordion>\n      </div>\n    </mat-card-content>\n  </mat-card>\n'},"./src/app/dashboard/statistics/statistics.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL2Rhc2hib2FyZC9zdGF0aXN0aWNzL3N0YXRpc3RpY3MuY29tcG9uZW50LnNjc3MifQ== */"},"./src/app/dashboard/statistics/statistics.component.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"StatisticsComponent",function(){return r});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./src/app/resultcard.model.ts"),i=n("./src/app/report-data.service.ts"),r=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},s.__decorate([Object(a.Input)(),s.__metadata("design:type",o.ResultCard)],t.prototype,"currentCard",void 0),t=s.__decorate([Object(a.Component)({selector:"app-statistics",template:n("./src/app/dashboard/statistics/statistics.component.html"),encapsulation:a.ViewEncapsulation.None,styles:[n("./src/app/dashboard/statistics/statistics.component.scss")]}),s.__metadata("design:paramtypes",[i.ReportDataService])],t)}()},"./src/app/dashboard/system-information/system-information.component.html":function(t,e){t.exports='  <mat-card fxFlex>\n    <mat-card-title>Environment</mat-card-title>\n    <mat-card-content>\n      <mat-accordion\n      multi="true"\n      displayMode="flat" >\n         <mat-expansion-panel\n         class="mat-elevation-z0"\n         hideToggle="true">\n          <mat-expansion-panel-header\n          class="mat-expansion-panel-header"\n          collapsedHeight="auto"\n          expandedHeight="auto"\n          style="padding:0.05vh">\n            <mat-panel-description fxLayoutAlign="space-between start">\n              <span> Python Version </span>\n              <span>{{data.metaData.python}}</span>\n            </mat-panel-description>\n          </mat-expansion-panel-header>\n        </mat-expansion-panel>\n        <mat-expansion-panel\n        class="mat-elevation-z0"\n        hideToggle="true">\n         <mat-expansion-panel-header\n         class="mat-expansion-panel-header"\n         collapsedHeight="auto"\n         expandedHeight="auto"\n         style="padding:0.05vh">\n           <mat-panel-description fxLayoutAlign="space-between start">\n             <span> Platform </span>\n             <span>{{data.metaData.platform}}</span>\n           </mat-panel-description>\n         </mat-expansion-panel-header>\n       </mat-expansion-panel>\n       <mat-expansion-panel\n       class="mat-elevation-z0"\n       *ngIf=\'data.metaData.packages\'\n       hideToggle="true">\n        <mat-expansion-panel-header\n        class="mat-expansion-panel-header"\n        collapsedHeight="auto"\n        expandedHeight="auto"\n        style="padding:0.05vh">\n          <mat-panel-description fxLayoutAlign="space-between start">\n            <span> Memote Version </span>\n            <span>{{data.metaData.packages.memote}}</span>\n          </mat-panel-description>\n        </mat-expansion-panel-header>\n        <div fxLayout=\'column\' flexLayoutAlign=\'space-between center\'>\n        <div> Package Versions </div>\n        <div>\n          <p></p>\n          <textarea rows="4" style="overflow-wrap: break-word; width:100%">\n          {{data.getString(data.metaData.packages)}}\n          </textarea>\n        </div>\n        </div>\n      </mat-expansion-panel>\n      </mat-accordion>\n    </mat-card-content>\n  </mat-card>\n'},"./src/app/dashboard/system-information/system-information.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL2Rhc2hib2FyZC9zeXN0ZW0taW5mb3JtYXRpb24vc3lzdGVtLWluZm9ybWF0aW9uLmNvbXBvbmVudC5zY3NzIn0= */"},"./src/app/dashboard/system-information/system-information.component.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"SystemInformationComponent",function(){return i});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./src/app/report-data.service.ts"),i=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},t=s.__decorate([Object(a.Component)({selector:"app-system-information",template:n("./src/app/dashboard/system-information/system-information.component.html"),encapsulation:a.ViewEncapsulation.None,styles:[n("./src/app/dashboard/system-information/system-information.component.scss")]}),s.__metadata("design:paramtypes",[o.ReportDataService])],t)}()},"./src/app/dashboard/universal-core/universal-core.component.html":function(t,e){t.exports='<mat-card-title></mat-card-title>\n    <mat-card-content>\n      <div>\n      <div fxFlex>\n        \x3c!-- <h2>Scored Tests</h2> --\x3e\n        <div *ngFor="let section of scoreData.sections | keys">\n          <h2> {{ scoreData.sections[section].title }}</h2>\n          <div *ngFor="let case of scoreData.sections[section].cases; let first = first">\n              <mat-expansion-panel\n              *ngIf="first && data.reportType === \'diff\'"\n              disabled\n              #expPanel\n              class="mat-elevation-z0">\n                 <mat-expansion-panel-header\n                 class="mat-expansion-panel-header"\n                 collapsedHeight="auto"\n                 style="padding:0.05vh">\n                 <mat-panel-title fxLayoutAlign=\'space-between center\'>\n                    <td class="test-title">\n                        Compared Models\n                    </td>\n                    <div style="margin-right: 10px">\n                    <td *ngFor="let results of data.byID(case)?.diff">\n                      <div class="MatButtonEmulation" style="background: white; color:black"> {{ results.model }} </div>\n                    </td>\n                    </div>\n                   </mat-panel-title>\n                 </mat-expansion-panel-header>\n                </mat-expansion-panel>\n            <app-accordion [case]="case"></app-accordion>\n          </div>\n          <div *ngIf="data.reportType != \'history\'">\n          <hr>\n          <app-score-formula [position]="section"></app-score-formula>\n          </div>\n        </div>\n      </div>\n      </div>\n      <div *ngIf="data.reportType != \'history\'">\n          <hr>\n          <hr>\n          <app-score-formula [position]="\'Total Score\'"></app-score-formula>\n          <hr>\n          </div>\n      <div\n        *ngIf="data.reportType != \'history\'; else vegaHistoryScore">\n        <div > Total Score\n          <div *ngIf="data.reportType === \'snapshot\'; else diffScore">\n            <div style="font-size:10vw">{{ data.score.total_score | percent:\'.\'  }}</div>\n          </div>\n          <ng-template #diffScore>\n              <div> \n                <app-tauchart-bar [plotType] = "\'diff-total\'"> </app-tauchart-bar>\n              </div>\n          </ng-template>\n        </div>\n        <div>\n        <div \n          *ngIf="data.reportType === \'snapshot\'; else diffSectionScore"> Score per Category\n              <app-tauchart-bar [plotType] = "\'snapshot-sections\'"> </app-tauchart-bar>\n        </div>\n        <ng-template #diffSectionScore>\n            <div> Score per Category\n            <br>\n                <app-tauchart-bar [plotType] = "\'diff-sections\'"> </app-tauchart-bar>\n          </div>\n        </ng-template>\n      </div>\n      </div>\n      <ng-template #vegaHistoryScore>\n          <div>\n            Total Score\n            <br>\n            <app-tauchart-history [testObject]="data.score.total_score"> </app-tauchart-history>\n          </div>\n      </ng-template>\n</mat-card-content>\n'},"./src/app/dashboard/universal-core/universal-core.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL2Rhc2hib2FyZC91bml2ZXJzYWwtY29yZS91bml2ZXJzYWwtY29yZS5jb21wb25lbnQuc2NzcyJ9 */"},"./src/app/dashboard/universal-core/universal-core.component.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"UniversalCoreComponent",function(){return i});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./src/app/report-data.service.ts"),i=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},s.__decorate([Object(a.Input)(),s.__metadata("design:type",Object)],t.prototype,"scoreData",void 0),t=s.__decorate([Object(a.Component)({selector:"app-universal-core",template:n("./src/app/dashboard/universal-core/universal-core.component.html"),encapsulation:a.ViewEncapsulation.None,styles:[n("./src/app/dashboard/universal-core/universal-core.component.scss")]}),s.__metadata("design:paramtypes",[o.ReportDataService])],t)}()},"./src/app/diff-button/diff-button.component.html":function(t,e){t.exports='<div *ngIf="data.isScored(data.getParam(testId, 0)); else notScored">\n    \x3c!-- If they are from scored tests use a percentage formatting for the results --\x3e\n    <div fxLayoutAlign=\'center stretch\' fxLayout=\'row wrap\'>\n    <div *ngFor="let results of data.byID(testId)?.diff">\n        <div\n          *ngIf="results.result === \'passed\' || results.result === \'failed\'"\n          class="MatButtonEmulation"\n          [ngStyle]="{\'background-color\':highlightColor}">\n            {{ 1 - results.metric | percent : \'1.1\' }}\n        </div>\n          \x3c!-- Skipped tests are marked with a gray button with beige text --\x3e\n        <div\n          *ngIf="results.result === \'skipped\'"\n          class="MatButtonEmulation"\n          style="background:gray; color:beige">\n            Skipped\n        </div>\n          \x3c!-- Tests with errors are marked with a black button with red text --\x3e\n        <div\n          class="MatButtonEmulation"\n          style="color:red;background-color:black"\n          *ngIf="results.result === \'error\'">\n            Errored\n        </div>\n    </div>\n  </div>\n</div>\n<ng-template #notScored>\n  \x3c!-- If they are from unscored tests the user has some control over how --\x3e\n  \x3c!-- they are displayed using the \'format_type\' attribute of each test. --\x3e\n  <div  fxLayoutAlign=\'center stretch\' fxLayout=\'row wrap\'>\n      <div *ngFor="let results of data.byID(testId)?.diff">\n        \x3c!-- Skipped tests are marked with a gray button with beige text --\x3e\n        <div\n        *ngIf="results.result === \'skipped\'"\n        class="MatButtonEmulation"\n        style="background:gray; color:beige">\n          Skipped\n        </div>\n          \x3c!-- Tests with errors are marked with a black button with red text --\x3e\n        <div\n          class="MatButtonEmulation"\n          style="color:red;background-color:black"\n          *ngIf="results.result === \'error\'">\n            Errored\n        </div>\n    <div\n    *ngIf="results.result === \'passed\' || results.result === \'failed\'"\n    class="MatButtonEmulation"\n    [ngStyle]="{\'background-color\':highlightColor}">\n      \x3c!-- Hence for each test, determine now how the result ought to be formatted --\x3e\n      <div *ngIf="data.byID(testId)?.format_type === \'count\'">\n        {{ results.data?.length | number : \'1.\'}}\n      </div>\n      <div *ngIf="data.byID(testId)?.format_type === \'number\'">\n        {{ results.data | number : \'1.2-2\'}}\n      </div>\n      <div *ngIf="data.byID(testId)?.format_type === \'percent\'">\n        {{ results.metric | number : \'1.2-2\'}}\n      </div>\n      <div *ngIf="data.byID(testId)?.format_type === \'raw\'">\n        {{ results.data }}\n      </div>\n    </div>\n    </div>\n  </div>\n</ng-template>\n'},"./src/app/diff-button/diff-button.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL2RpZmYtYnV0dG9uL2RpZmYtYnV0dG9uLmNvbXBvbmVudC5zY3NzIn0= */"},"./src/app/diff-button/diff-button.component.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"DiffButtonComponent",function(){return i});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./src/app/report-data.service.ts"),i=function(){function t(t){this.data=t,this.metricList=[]}return t.prototype.ngOnInit=function(){this.getDiffColor()},t.prototype.getDiffColor=function(){if(this.data.isScored(this.data.getParam(this.testId,0))){for(var t=0,e=this.data.byID(this.testId).diff;t<e.length;t++){var n=e[t];this.metricList.push(n.metric)}this.spectrum(this.ratioOrVariance(this.metricList))}else if("raw"!==this.data.byID(this.testId).format_type){switch(this.data.byID(this.testId).format_type){case"count":for(var s=0,a=this.data.byID(this.testId).diff;s<a.length;s++){n=a[s];this.metricList.push(n.data.length)}break;case"number":for(var o=0,i=this.data.byID(this.testId).diff;o<i.length;o++){n=i[o];this.metricList.push(n.data)}break;case"percent":for(var r=0,c=this.data.byID(this.testId).diff;r<c.length;r++){n=c[r];this.metricList.push(n.metric)}}this.spectrum(this.ratioOrVariance(this.metricList))}else{for(var l=0,d=this.data.byID(this.testId).diff;l<d.length;l++){n=d[l];this.metricList.push(n.data)}this.spectrum(this.determineStringEquality(this.metricList))}},t.prototype.ratioOrVariance=function(t){var e=Math.min.apply(Math,t),n=Math.max.apply(Math,t);return 2===t.length&&t[0]===t[1]?0:0===e?100*(1-(e+1)/(n+1)):100*(1-e/n)},t.prototype.determineStringEquality=function(t){return!!t.length&&!!t.reduce(function(t,e){return t===e?t:NaN})?0:100},t.prototype.spectrum=function(t){var e=(1.19*t+42).toString(10),n=(123-1.05*t).toString(10),s=(184-1.66*t).toString(10);this.highlightColor=["rgb(",e,",",n,",",s,")"].join("")},s.__decorate([Object(a.Input)(),s.__metadata("design:type",String)],t.prototype,"testId",void 0),t=s.__decorate([Object(a.Component)({selector:"app-diff-button",template:n("./src/app/diff-button/diff-button.component.html"),encapsulation:a.ViewEncapsulation.None,styles:[n("./src/app/diff-button/diff-button.component.scss")]}),s.__metadata("design:paramtypes",[o.ReportDataService])],t)}()},"./src/app/header/header.component.html":function(t,e){t.exports='<mat-toolbar\n  class="mat-elevation-z9 fixed-header"\n  color=\'primary\'>\n      <img aria-label="logo" style="width: 5vw; padding-right:1vw"\n       width="48px"\n       height="48px"\n       src="data:image/svg+xml;base64, "/>\n      <span style="text-align:left; padding-right:1vw" *ngIf="data.byID(\'test_model_id_presence\')"> {{ data.byID("test_model_id_presence").data }} </span>\n      <button mat-button (click)="data.togglePanel()"> {{ !data.allExpandState ? "Expand" : "Collapse" }} All</button>\n      <app-help-dialog></app-help-dialog>\n      <span style="flex:1 1 auto"></span>\n      <span style="text-align:right; padding-right:1vw" *ngIf="data.metaData">{{ data.metaData.timestamp | date:\'yyyy-MM-dd HH:mm\' }}</span>\n      <a\n        aria-label="GitHub Repository"\n        href="https://github.com/opencobra/memote"\n        mat-button>\n        <span class="mat-button-wrapper">\n          <img\n          width="40px"\n          height="40px"\n          src="data:image/svg+xml;base64,PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0idXRmLTgiPz48IURPQ1RZUEUgc3ZnIFBVQkxJQyAiLS8vVzNDLy9EVEQgU1ZHIDEuMS8vRU4iICJodHRwOi8vd3d3LnczLm9yZy9HcmFwaGljcy9TVkcvMS4xL0RURC9zdmcxMS5kdGQiPjxzdmcgdmVyc2lvbj0iMS4xIiBpZD0iTGF5ZXJfMSIgeG1sbnM9Imh0dHA6Ly93d3cudzMub3JnLzIwMDAvc3ZnIiB4bWxuczp4bGluaz0iaHR0cDovL3d3dy53My5vcmcvMTk5OS94bGluayIgeD0iMHB4IiB5PSIwcHgiIHdpZHRoPSIxNnB4IiBoZWlnaHQ9IjE2cHgiIHZpZXdCb3g9IjAgMCAxNiAxNiIgZW5hYmxlLWJhY2tncm91bmQ9Im5ldyAwIDAgMTYgMTYiIHhtbDpzcGFjZT0icHJlc2VydmUiPjx0aXRsZT5tYXJrLWdpdGh1YjwvdGl0bGU+PGRlc2M+Q3JlYXRlZCB3aXRoIFNrZXRjaC48L2Rlc2M+PGcgaWQ9Ik9jdGljb25zIj48ZyBpZD0ibWFyay1naXRodWIiPjxwYXRoIGlkPSJTaGFwZSIgZmlsbD0iI0ZGRkZGRiIgZD0iTTgsMEMzLjU4LDAsMCwzLjU4LDAsOGMwLDMuNTQsMi4yOSw2LjUzLDUuNDcsNy41OWMwLjQsMC4wNywwLjU1LTAuMTcsMC41NS0wLjM4YzAtMC4xOS0wLjAxLTAuODE5LTAuMDEtMS40OUM0LDE0LjA5LDMuNDgsMTMuMjMsMy4zMiwxMi43OGMtMC4wOS0wLjIzLTAuNDgtMC45NC0wLjgyLTEuMTNjLTAuMjgtMC4xNS0wLjY4LTAuNTIxLTAuMDEtMC41M2MwLjYzLTAuMDExLDEuMDgsMC41OCwxLjIzLDAuODE5YzAuNzIsMS4yMTEsMS44NywwLjg3MSwyLjMzLDAuNjZjMC4wNy0wLjUyLDAuMjgtMC44NjksMC41MS0xLjA2OWMtMS43OC0wLjItMy42NC0wLjg5LTMuNjQtMy45NWMwLTAuODcsMC4zMS0xLjU5LDAuODItMi4xNWMtMC4wOC0wLjItMC4zNi0xLjAyLDAuMDgtMi4xMmMwLDAsMC42Ny0wLjIxLDIuMiwwLjgyYzAuNjQtMC4xOCwxLjMyLTAuMjcsMi0wLjI3czEuMzYsMC4wOSwyLDAuMjdjMS41My0xLjA0LDIuMjAxLTAuODIsMi4yMDEtMC44MmMwLjQzOSwxLjEsMC4xNTksMS45MiwwLjA3OSwyLjEyYzAuNTExLDAuNTYsMC44MiwxLjI3LDAuODIsMi4xNWMwLDMuMDctMS44NywzLjc1LTMuNjQ5LDMuOTVjMC4yODksMC4yNSwwLjUzOSwwLjcyOSwwLjUzOSwxLjQ3OWMwLDEuMDctMC4wMSwxLjkzLTAuMDEsMi4yYzAsMC4yMSwwLjE1LDAuNDYsMC41NSwwLjM4QzEzLjcxLDE0LjUzLDE2LDExLjUzLDE2LDhDMTYsMy41OCwxMi40MiwwLDgsMEw4LDB6Ii8+PC9nPjwvZz48L3N2Zz4="/>\n        </span>\n        </a>\n</mat-toolbar>'},"./src/app/header/header.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL2hlYWRlci9oZWFkZXIuY29tcG9uZW50LnNjc3MifQ== */"},"./src/app/header/header.component.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"HeaderComponent",function(){return i});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./src/app/report-data.service.ts"),i=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},t=s.__decorate([Object(a.Component)({selector:"app-header",template:n("./src/app/header/header.component.html"),encapsulation:a.ViewEncapsulation.None,styles:[n("./src/app/header/header.component.scss")]}),s.__metadata("design:paramtypes",[o.ReportDataService])],t)}()},"./src/app/help-dialog/help-dialog-text.component.html":function(t,e){t.exports='<h2 mat-dialog-title>Readme</h2>\n<mat-dialog-content>\n<mat-card>\n<mat-card-title> Understanding the reports </mat-card-title>\n<mat-card-content>\n<p>Memote will return one of four possible outputs.\nIf your preferred workflow is to benchmark one or several genome-scale\nmetabolic models (GSM) memote generates either a snapshot or a diff report,\nrespectively. For the reconstruction workflow the primary output is a history\nreport. This will only work if the provided input models are formatted\ncorrectly in the systems biology markup language (<a href="http://sbml.org/Main_Page"> SBML </a>). However, if a\nprovided model is not a valid SBML file, memote composes a report\nenumerating errors and warnings from the SBML validator in\norder of appearance. To better understand the output of the error report we\nrefer the reader to this section of the <a href="http://sbml.org/Facilities/Documentation/Error_Categories"> SBML documentation </a>. In this section\nwe will focus on how to understand the snapshot, diff and history reports.</p>\n\n</mat-card-content>\n</mat-card>\n\n<mat-card>\n<mat-card-title> Orientation </mat-card-title>\n<mat-card-content>\n\n<h3>Toolbar</h3>\n\n\n<p>In all three reports, the blue toolbar at the top shows (from left to right)\nthe memote logo, a button which expand/ collapses all test results, a button\nwhich displays the readme and the github icon which links to memote\'s github\npage. On the snapshot report, the toolbar will also display the identifier of\nthe tested GSM and a timestamp showing when the test run was initiated.</p>\n\n<h3>Main Body</h3>\n\n<p>The main body of the reports is divided into an independent section to the left\nand a specific section to the right.</p>\n\n<p> The tests in the independent section are\nagnostic of the type of modeled organism, preferred modeling paradigms,\nthe complexity of a genome-scale metabolic model (GSM) or the types of\nidentifiers that are used to describe its components. The tests in this section\nfocus on testing adherence to fundamental principles of\nconstraint-based modeling: mass, charge and stoichiometric balance as well as\nthe presence of annotations. The results in this section can be normalized, and\nthus enable a comparison of GSMs. The <b>Score</b> at the bottom\nof the page summarises the results to further simplify comparison.</p>\n\n<p>The specific section on the right provides model specific statistics\nand covers aspects of a metabolic network that can not be normalized\nwithout introducing bias. For instance, dedicated quality control of the biomass\nequation only applies to GSMs which are used to investigate cell growth, i.e.,\nthose for which a biomass equation has been generated. Some tests in this\nsection are also influenced by whether the tested GSM represents a prokaryote or\na eukaryote. Therefore the results cannot be generalized and direct comparisons\nought to take bias into account.</p>\n\n<h3>Test Results</h3>\n\n<p>Test results are arranged in rows with the title visible to the left and the\nresult on the right. The result is displayed as white text in a coloured\nrectangle detailed in <b>Color</b>.</p>\n\n<p>By default only the minimum information is visible as indicated by an arrow pointing\ndown right of the result. Clicking anywhere in the row will expand the result\nrevealing a description of the concept behind the test, its implementation\nand a brief summary of the result.\nIn addition, there is a text field which contains plaintext representations of\nPython objects which can be copy/pasted into Python code for follow up\nprocedures.</p>\n\n<p>Some tests carry out one operation on several parameters and therefore deviate\nslightly from the descriptions above. Expanding the title row reveals only the\ndescription, while rows of the individual parameters reveal the text fields.</p>\n\n<p>In the history report, instead of text fields scatterplots show how the\nrespective metrics developed over the commit history for each branch of a\nrepository. By clicking an entry in the legend, it is possible to toggle\nits visibility in the plot.</p>\n\n</mat-card-content>\n</mat-card>\n\n<mat-card>\n<mat-card-title> Interpretation </mat-card-title>\n<mat-card-content>\n\n<p>The variety of constraints-based modeling approaches and differences between\nvarious organisms compound the assessment of GSMs. While memote facilitates\nmodel assessment it can only do so within limitations. Please bear in mind the\ndiversity of Paradigms_ that challenge some of memote\'s results. </p>\n\n    <h3> Color </h3>\n\n\n<h4> Snapshot Report </h4>\n\n<p>Results without highlights are kept in the main <span style="color: rgb(42, 123, 184)"> <b> blue </b> </span>  color of the memote\ncolor scheme. Scored results will be marked with a gradient ranging from <span style="color: rgb(161, 18, 18)"> <b> red </b> </span>\nto <span style="color: rgb(18, 161, 46)"> <b> green </b> </span> denoting a low or a high score respectively: </p>\n\n\n    <div fxLayoutAlign=\'center center\'>\n    <span> <b>0%</b></span>\n    <div style="display: inline-block; vertical-align: middle">\n    <svg width="120" height="60">\n    <defs>\n    <linearGradient id="grad1">\n    <stop offset="0%" style="stop-color:rgb(161, 18, 18);stop-opacity:1" />\n    <stop offset="100%" style="stop-color:rgb(18, 161, 46);stop-opacity:1" />\n    </linearGradient>\n    </defs>\n    <rect fill="url(#grad1)" x="10" y="10" width="100" height="100"/>\n    </svg>\n    </div>\n    <span> <b>100% </b></span>\n    </div>\n    <br>\n\n\n<h4> Diff Report</h4>\n\n<p>The colour in the Diff Report depends on the ratio of the sample minimum to\nthe sample maxium. Result sets where the sample minimum and the sample\nmaximum are identical will be coloured in the main <span style="color: rgb(42, 123, 184)"> <b> blue </b> </span> colour of the\nmemote colour scheme. Result sets where the sample minimum is very small\nrelative to the sample maximum will appear red . This ratio is calculated\nwith as <b> 1 - (Min / Max)) * 100</b>. </p>\n\n<p>This is then mapped to the following gradient: </p>\n\n<div fxLayoutAlign=\'center center\'>\n            <span> <b>Identical</b></span>\n            <div style="display: inline-block; vertical-align: middle">\n                <svg width="120" height="60">\n                    <defs>\n                        <linearGradient id="grad2">\n                        <stop offset="0%" style="stop-color:rgb(42, 123, 184);stop-opacity:1" />\n                        <stop offset="100%" style="stop-color:rgb(161, 18, 18);stop-opacity:1" />\n                        </linearGradient>\n                    </defs>\n                    <rect fill="url(#grad2)"\n                        x="10" y="10" width="100" height="100"/>\n                </svg>\n            </div>\n                <span> <b>Different</b></span>\n    </div>\n\n<h3> Score</h3>\n\n\n<p>Each test in the independent section provides a relative measure of\ncompleteness with regard to the tested property. The final score is the\nweighted sum of all individual test results normalized by the maximally\nachievable score, i.e., all individual results at 100%. Individual tests can\nbe weighted, but it is also possible to apply weighting to entire test\ncategories. Hence the final score is calculated:</p>\n  <br>\n         \x3c!-- Generated with this latex formula \\[ TotalScore = \\frac{\\sum_{Sections} weight_{section} \\times ( \\sum_{Tests} weight_{test} \\times score_{test})}{MaxScore}\\] at http://www.tlhiv.org/ltxpreview/ --\x3e\n         <svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" viewBox="0 0 310 23">\n            <defs>\n              <symbol id="a" overflow="visible">\n                <path d="M4.25-6.05c.08-.28.1-.34.23-.37.1-.02.43-.02.63-.02 1.01 0 1.45.03 1.45.82 0 .15-.03.54-.08.8l-.03.18c0 .06.03.14.13.14s.12-.08.15-.23L7-6.47c.02-.05.02-.14.02-.17 0-.11-.1-.11-.27-.11H1.22c-.24 0-.25.02-.33.2L.3-4.8c0 .02-.06.16-.06.2 0 .04.07.1.13.1.1 0 .1-.06.17-.22.53-1.55.8-1.72 2.27-1.72h.39c.28 0 .28.03.28.13 0 .04-.03.17-.05.2L2.1-.78C2-.42 1.97-.31.91-.31c-.36 0-.43 0-.43.18C.48 0 .6 0 .66 0l.8-.02.87-.01.83.01.86.02c.09 0 .21 0 .21-.2 0-.11-.07-.11-.34-.11-.23 0-.37 0-.62-.02-.3-.03-.38-.06-.38-.22 0-.01 0-.06.05-.2zm0 0"/>\n              </symbol>\n              <symbol id="b" overflow="visible">\n                <path d="M4.67-2.72c0-1.05-.69-1.69-1.6-1.69C1.76-4.4.42-2.98.42-1.58.4-.59 1.08.11 2 .11c1.33 0 2.67-1.38 2.67-2.83zM2.02-.1c-.44 0-.88-.31-.88-1.1 0-.48.27-1.56.58-2.06.5-.76 1.08-.92 1.36-.92.58 0 .87.49.87 1.08 0 .4-.2 1.44-.58 2.08-.35.58-.9.92-1.35.92zm0 0"/>\n              </symbol>\n              <symbol id="c" overflow="visible">\n                <path d="M2.05-3.98h.93c.2 0 .32 0 .32-.2 0-.12-.11-.12-.28-.12h-.88c.36-1.42.4-1.6.4-1.67 0-.17-.12-.26-.29-.26-.03 0-.31 0-.4.36L1.48-4.3H.53c-.2 0-.3 0-.3.2 0 .12.08.12.29.12h.87C.67-1.16.63-.98.63-.81c0 .54.37.92.92.92 1.01 0 1.57-1.45 1.57-1.53 0-.11-.07-.11-.1-.11-.1 0-.11.03-.16.14C2.44-.34 1.9-.11 1.56-.11c-.2 0-.31-.12-.31-.45 0-.25.03-.31.06-.49zm0 0"/>\n              </symbol>\n              <symbol id="d" overflow="visible">\n                <path d="M3.72-3.77c-.19-.37-.47-.64-.92-.64C1.64-4.4.4-2.94.4-1.48.4-.55.96.1 1.73.1c.2 0 .7-.05 1.3-.75.07.42.43.75.9.75.36 0 .58-.23.75-.55.16-.36.3-.97.3-.98 0-.11-.1-.11-.13-.11-.09 0-.1.05-.14.19-.17.64-.34 1.23-.75 1.23-.28 0-.3-.27-.3-.45 0-.22.02-.31.13-.75.11-.4.13-.52.22-.9l.36-1.38c.06-.29.06-.3.06-.35 0-.17-.1-.26-.28-.26-.25 0-.39.22-.42.43zm-.64 2.58c-.06.19-.06.2-.2.38-.44.54-.85.7-1.13.7-.5 0-.64-.55-.64-.94 0-.5.31-1.72.55-2.18.3-.58.75-.96 1.15-.96.64 0 .78.82.78.88 0 .06-.01.12-.03.17zm0 0"/>\n              </symbol>\n              <symbol id="e" overflow="visible">\n                <path d="M2.58-6.81s0-.11-.14-.11c-.22 0-.96.08-1.22.1-.08 0-.19.02-.19.21 0 .11.11.11.25.11.49 0 .5.1.5.17l-.03.2L.48-1.13a.97.97 0 0 0-.04.33c0 .58.43.92.9.92.33 0 .58-.2.75-.56.18-.38.3-.96.3-.97 0-.11-.1-.11-.12-.11-.1 0-.11.05-.13.19-.17.64-.36 1.23-.77 1.23-.3 0-.3-.31-.3-.45 0-.25.02-.3.07-.49zm0 0"/>\n              </symbol>\n              <symbol id="f" overflow="visible">\n                <path d="M6.44-6.92c0-.03-.03-.11-.13-.11-.04 0-.04.01-.17.16l-.48.56c-.25-.47-.77-.72-1.43-.72-1.26 0-2.45 1.16-2.45 2.36 0 .8.52 1.26 1.03 1.4l1.06.29c.38.09.93.25.93 1.06C4.8-1.03 3.98-.1 3-.1c-.64 0-1.75-.22-1.75-1.46 0-.23.06-.48.06-.53l.02-.08c0-.09-.06-.1-.11-.1a.14.14 0 0 0-.11.04C1.08-2.19.5.1.5.13c0 .04.05.09.11.09C.68.22.7.2.82.06L1.3-.5c.42.58 1.1.72 1.68.72 1.36 0 2.54-1.33 2.54-2.56 0-.7-.35-1.04-.49-1.18-.23-.21-.39-.26-1.26-.48l-.68-.19a.93.93 0 0 1-.6-.9c0-.82.8-1.66 1.74-1.66.82 0 1.43.44 1.43 1.55 0 .3-.04.5-.04.56 0 0 0 .1.11.1s.13-.04.16-.21zm0 0"/>\n              </symbol>\n              <symbol id="g" overflow="visible">\n                <path d="M3.95-3.78c-.17 0-.3 0-.43.12a.53.53 0 0 0-.2.4c0 .24.2.35.38.35.28 0 .55-.25.55-.64 0-.48-.47-.86-1.17-.86C1.73-4.4.4-2.98.4-1.58.4-.68.98.11 2.03.11 3.45.1 4.28-.95 4.28-1.06c0-.06-.05-.14-.1-.14-.07 0-.09.03-.15.1-.78 1-1.87 1-1.98 1-.63 0-.9-.5-.9-1.1 0-.4.2-1.38.54-1.99.31-.58.86-1 1.4-1 .33 0 .72.13.86.4zm0 0"/>\n              </symbol>\n              <symbol id="h" overflow="visible">\n                <path d="M.88-.6l-.1.44c0 .18.14.27.3.27.12 0 .3-.08.37-.28.02-.03.35-1.4.4-1.58.07-.33.26-1.02.3-1.3.05-.12.33-.6.57-.8.08-.07.37-.34.8-.34.26 0 .4.13.42.13-.3.04-.52.28-.52.54 0 .16.11.35.38.35.26 0 .54-.24.54-.6 0-.34-.3-.64-.82-.64-.65 0-1.08.49-1.27.77a.92.92 0 0 0-.92-.77c-.46 0-.64.4-.74.57-.17.34-.3.93-.3.96 0 .11.1.11.12.11.1 0 .1-.01.17-.23.17-.7.37-1.19.73-1.19.16 0 .3.08.3.46 0 .21-.03.32-.16.84zm0 0"/>\n              </symbol>\n              <symbol id="i" overflow="visible">\n                <path d="M1.86-2.3c.3 0 1.03-.03 1.53-.23.7-.3.75-.9.75-1.03 0-.44-.37-.85-1.06-.85-1.11 0-2.63.97-2.63 2.72 0 1.02.6 1.8 1.58 1.8C3.45.1 4.28-.95 4.28-1.06c0-.06-.05-.14-.1-.14-.07 0-.09.03-.15.1-.78 1-1.87 1-1.98 1-.78 0-.88-.85-.88-1.17 0-.12.02-.42.16-1.03zm-.47-.22c.4-1.51 1.42-1.67 1.69-1.67.45 0 .73.3.73.63 0 1.04-1.6 1.04-2.01 1.04zm0 0"/>\n              </symbol>\n              <symbol id="t" overflow="visible">\n                <path d="M4.6-3.38c.06-.21.15-.59.15-.65 0-.17-.14-.27-.28-.27-.13 0-.3.08-.38.28-.03.07-.5 1.97-.56 2.24a2.34 2.34 0 0 0-.08.83c-.23.53-.53.84-.92.84-.8 0-.8-.73-.8-.9 0-.32.05-.7.52-1.94.1-.3.17-.44.17-.64a.8.8 0 0 0-.81-.82C.66-4.4.3-2.95.3-2.87c0 .1.1.1.1.1.12 0 .12-.03.16-.18.28-.92.66-1.24 1.02-1.24.1 0 .25.02.25.33 0 .25-.11.53-.19.7a6.08 6.08 0 0 0-.55 2.02C1.1-.24 1.75.1 2.5.1c.17 0 .64 0 1.03-.7.27.64.95.7 1.25.7.75 0 1.19-.63 1.45-1.22.33-.78.66-2.12.66-2.6 0-.54-.26-.7-.44-.7-.25 0-.5.27-.5.49 0 .12.07.19.14.26.11.11.36.36.36.85a5.9 5.9 0 0 1-.54 1.83c-.25.53-.61.87-1.1.87-.47 0-.73-.3-.73-.87 0-.27.06-.58.1-.72zm0 0"/>\n              </symbol>\n              <symbol id="u" overflow="visible">\n                <path d="M2.83-6.23c0-.2-.14-.36-.36-.36-.28 0-.55.26-.55.53 0 .18.14.36.38.36.23 0 .53-.24.53-.53zm-.75 3.75c.1-.29.1-.32.22-.58.08-.2.12-.35.12-.53a.79.79 0 0 0-.81-.82C.67-4.4.3-2.95.3-2.87c0 .1.1.1.1.1.12 0 .12-.03.16-.18.28-.94.67-1.24 1.02-1.24.08 0 .25 0 .25.32 0 .21-.08.42-.11.53-.08.25-.53 1.4-.69 1.84-.1.25-.23.58-.23.8 0 .47.34.8.8.8.95 0 1.32-1.43 1.32-1.52 0-.11-.1-.11-.12-.11-.1 0-.1.03-.14.19-.2.62-.52 1.23-1.02 1.23-.17 0-.25-.1-.25-.33 0-.25.06-.39.3-1zm0 0"/>\n              </symbol>\n              <symbol id="v" overflow="visible">\n                <path d="M4.69-3.77c.01-.04.03-.1.03-.17 0-.17-.11-.26-.28-.26-.1 0-.38.06-.4.42-.2-.36-.54-.63-.95-.63C1.97-4.4.73-3 .73-1.58.73-.59 1.33 0 2.05 0c.6 0 1.06-.47 1.15-.58l.02.02C3.02.3 2.89.74 2.89.75c-.05.1-.37 1.08-1.44 1.08-.18 0-.51-.02-.8-.11.3-.08.41-.34.41-.52 0-.15-.1-.34-.37-.34a.55.55 0 0 0-.53.58c0 .4.36.6 1.3.6 1.26 0 1.98-.77 2.13-1.37zM3.4-1.28c-.07.26-.3.51-.52.7-.2.17-.52.36-.81.36-.5 0-.64-.51-.64-.92 0-.47.28-1.66.56-2.16.27-.48.69-.89 1.1-.89.67 0 .8.82.8.86 0 .05 0 .11-.02.14zm0 0"/>\n              </symbol>\n              <symbol id="w" overflow="visible">\n                <path d="M2.86-6.81s0-.11-.13-.11c-.23 0-.95.08-1.21.1-.08 0-.2.02-.2.2 0 .12.1.12.24.12.49 0 .5.06.5.17l-.03.2L.6-.38c-.04.14-.04.16-.04.22 0 .23.2.28.3.28a.4.4 0 0 0 .35-.27l.2-.75.2-.89.18-.67c.02-.06.11-.39.11-.45.03-.1.34-.64.69-.92.22-.16.51-.35.95-.35.42 0 .53.35.53.7 0 .54-.37 1.63-.6 2.24-.08.22-.15.34-.15.55 0 .47.36.8.83.8.94 0 1.3-1.44 1.3-1.52 0-.11-.08-.11-.11-.11-.11 0-.11.03-.16.19-.14.53-.47 1.23-1.01 1.23-.18 0-.24-.1-.24-.33 0-.25.08-.48.17-.7.16-.44.61-1.63.61-2.2 0-.64-.39-1.07-1.14-1.07-.62 0-1.1.32-1.48.77zm0 0"/>\n              </symbol>\n              <symbol id="A" overflow="visible">\n                <path d="M3.9-3.73c-.27.01-.48.23-.48.45 0 .14.1.3.31.3.22 0 .46-.18.46-.57C4.19-4 3.77-4.4 3-4.4c-1.31 0-1.69 1.02-1.69 1.46 0 .78.74.92 1.03.98.52.11 1.04.22 1.04.77 0 .25-.22 1.1-1.43 1.1-.14 0-.9 0-1.14-.54.4.05.64-.25.64-.53 0-.22-.17-.35-.37-.35-.27 0-.56.2-.56.66 0 .56.57.97 1.42.97 1.62 0 2.01-1.2 2.01-1.66 0-.36-.18-.6-.3-.72-.27-.28-.57-.34-1-.42-.37-.08-.76-.15-.76-.6 0-.29.24-.9 1.11-.9.25 0 .75.08.9.46zm0 0"/>\n              </symbol>\n              <symbol id="C" overflow="visible">\n                <path d="M9.23-6.05c.1-.36.11-.45.86-.45.22 0 .32 0 .32-.2 0-.11-.1-.11-.27-.11H8.83c-.27 0-.28 0-.4.18L4.8-.92l-.78-5.65c-.04-.23-.05-.23-.32-.23H2.34c-.18 0-.3 0-.3.18 0 .13.1.13.29.13l.44.02c.15.03.21.04.21.17l-.04.18-1.27 5.07c-.1.4-.26.72-1.08.75-.04 0-.17.01-.17.18 0 .1.06.13.14.13.31 0 .66-.03 1-.03L2.58 0c.04 0 .19 0 .19-.2 0-.11-.11-.11-.2-.11-.55 0-.66-.2-.66-.44 0-.06 0-.13.03-.23L3.3-6.41h.01l.86 6.18c.02.12.03.23.14.23.11 0 .17-.1.22-.17l4.03-6.31h.02L7.14-.78c-.1.39-.1.47-.9.47-.16 0-.27 0-.27.18 0 .13.1.13.14.13l1.23-.03L8.6 0c.07 0 .2 0 .2-.2 0-.11-.1-.11-.29-.11-.36 0-.64 0-.64-.17 0-.05 0-.07.05-.25zm0 0"/>\n              </symbol>\n              <symbol id="D" overflow="visible">\n                <path d="M3.33-3.02c.06-.25.3-1.17.98-1.17a1 1 0 0 1 .5.13.58.58 0 0 0-.47.54c0 .16.11.35.38.35.22 0 .53-.17.53-.58 0-.52-.58-.66-.92-.66-.58 0-.92.54-1.05.75A1.1 1.1 0 0 0 2.2-4.4C1.17-4.4.6-3.12.6-2.87c0 .1.1.1.12.1.08 0 .1-.03.12-.1.35-1.07 1-1.32 1.35-1.32.19 0 .53.1.53.67 0 .32-.17.97-.53 2.38-.16.6-.52 1.03-.96 1.03-.06 0-.28 0-.5-.12.25-.07.47-.27.47-.55 0-.27-.22-.34-.36-.34-.3 0-.54.25-.54.57 0 .46.48.66.92.66.67 0 1.03-.7 1.05-.75.12.36.48.75 1.07.75 1.04 0 1.6-1.28 1.6-1.53 0-.11-.08-.11-.11-.11-.1 0-.11.05-.14.1C4.36-.33 3.69-.1 3.37-.1c-.39 0-.54-.31-.54-.66 0-.21.04-.43.15-.87zm0 0"/>\n              </symbol>\n              <symbol id="j" overflow="visible">\n                <path d="M6.84-3.27c.16 0 .35 0 .35-.18 0-.2-.19-.2-.33-.2H.89c-.14 0-.33 0-.33.2 0 .18.19.18.33.18zm.02 1.94c.14 0 .33 0 .33-.2 0-.19-.19-.19-.35-.19H.9c-.14 0-.33 0-.33.19 0 .2.19.2.33.2zm0 0"/>\n              </symbol>\n              <symbol id="y" overflow="visible">\n                <path d="M3.3 2.4c0-.04 0-.06-.17-.23C1.89.92 1.55-.97 1.55-2.5c0-1.73.38-3.47 1.61-4.7.13-.13.13-.14.13-.17 0-.08-.03-.11-.1-.11-.1 0-1 .68-1.6 1.95-.5 1.1-.62 2.2-.62 3.03 0 .78.11 1.98.66 3.13A4.37 4.37 0 0 0 3.2 2.5c.07 0 .1-.03.1-.1zm0 0"/>\n              </symbol>\n              <symbol id="B" overflow="visible">\n                <path d="M2.88-2.5c0-.77-.11-1.97-.66-3.1A4.39 4.39 0 0 0 .67-7.49a.1.1 0 0 0-.1.1c0 .04 0 .05.18.24.98.98 1.55 2.56 1.55 4.64C2.3-.78 1.94.97.7 2.22c-.14.12-.14.14-.14.17 0 .06.05.11.11.11.1 0 1-.69 1.58-1.95.52-1.1.63-2.2.63-3.05zm0 0"/>\n              </symbol>\n              <symbol id="k" overflow="visible">\n                <path d="M4.2 5.33L.66 9.7c-.08.1-.1.11-.1.16 0 .1.1.1.28.1H9.1l.86-2.48H9.7a2.7 2.7 0 0 1-1.76 1.64c-.16.07-.85.3-2.32.3H1.4l3.47-4.28c.06-.1.08-.1.08-.16 0-.04 0-.04-.07-.14L1.65.41h3.94C6.72.4 9.02.47 9.7 2.33h.25L9.1 0H.84C.56 0 .56.02.56.31zm0 0"/>\n              </symbol>\n              <symbol id="l" overflow="visible">\n                <path d="M4.95-4.81c0-.03-.03-.1-.1-.1-.04 0-.04.02-.13.11l-.34.4c-.27-.38-.7-.5-1.13-.5-.98 0-1.86.8-1.86 1.6 0 .1.03.38.24.64.23.27.5.33.96.46.14.03.49.1.6.14.22.04.64.2.64.72 0 .56-.6 1.23-1.38 1.23-.62 0-1.36-.22-1.36-.98 0-.08.02-.24.05-.36v-.03c0-.1-.08-.1-.1-.1-.1 0-.12.02-.13.14L.55-.04C.53-.03.52.01.52.05c0 .03.03.08.1.08.04 0 .05-.02.15-.1l.32-.38c.38.4.91.48 1.33.48 1.06 0 1.9-.9 1.9-1.73 0-.3-.12-.58-.27-.74-.24-.25-.35-.28-1.25-.5l-.44-.1c-.17-.07-.47-.24-.47-.63 0-.56.63-1.11 1.34-1.11.77 0 1.13.42 1.13 1.08l-.03.3c0 .12.08.12.12.12.1 0 .11-.05.13-.16zm0 0"/>\n              </symbol>\n              <symbol id="m" overflow="visible">\n                <path d="M1.56-1.6c.19 0 .75 0 1.13-.13.51-.2.6-.52.6-.72 0-.4-.38-.63-.85-.63-.85 0-1.97.64-1.97 1.85 0 .7.44 1.3 1.25 1.3 1.19 0 1.73-.7 1.73-.79 0-.05-.06-.12-.12-.12-.03 0-.05.01-.11.07-.55.65-1.36.65-1.49.65-.42 0-.7-.29-.7-.85 0-.1 0-.23.1-.62zm-.39-.18c.3-1.02 1.06-1.1 1.27-1.1.3 0 .56.16.56.43 0 .67-1.19.67-1.48.67zm0 0"/>\n              </symbol>\n              <symbol id="n" overflow="visible">\n                <path d="M3.05-2.67c-.25.04-.35.23-.35.39 0 .19.14.26.27.26.15 0 .39-.1.39-.45 0-.47-.53-.6-.9-.6-1.05 0-2.02.96-2.02 1.93 0 .6.4 1.2 1.28 1.2 1.19 0 1.73-.69 1.73-.78 0-.05-.06-.12-.12-.12-.03 0-.05.01-.11.07-.55.65-1.36.65-1.49.65-.5 0-.71-.35-.71-.79 0-.18.09-.95.45-1.43.26-.35.62-.54.98-.54.1 0 .42.02.6.2zm0 0"/>\n              </symbol>\n              <symbol id="o" overflow="visible">\n                <path d="M1.72-2.75h.7c.14 0 .22 0 .22-.16 0-.09-.08-.09-.2-.09h-.66l.25-1.03a.32.32 0 0 0 .03-.1c0-.14-.1-.23-.25-.23-.17 0-.26.13-.33.3-.04.18.05-.16-.26 1.06h-.7C.39-3 .3-3 .3-2.84c0 .09.08.09.2.09h.66L.75-1.11c-.05.17-.1.42-.1.52 0 .4.35.65.74.65C2.17.06 2.61-.9 2.61-1c0-.1-.1-.1-.11-.1-.1 0-.1.02-.16.15-.18.43-.54.82-.92.82-.15 0-.25-.09-.25-.34 0-.06.03-.22.05-.28zm0 0"/>\n              </symbol>\n              <symbol id="p" overflow="visible">\n                <path d="M2.27-4.36c0-.1-.1-.26-.29-.26-.18 0-.39.18-.39.39 0 .1.08.26.28.26s.4-.2.4-.39zM.84-.8c-.03.1-.06.17-.06.3 0 .32.27.57.66.57.69 0 1-.95 1-1.06 0-.1-.1-.1-.11-.1-.1 0-.11.05-.14.13-.16.56-.46.84-.74.84-.14 0-.17-.09-.17-.24 0-.16.05-.29.11-.44l.22-.56c.06-.18.33-.8.34-.9A.5.5 0 0 0 2-2.48c0-.33-.28-.6-.66-.6C.64-3.08.33-2.12.33-2c0 .08.1.08.12.08.1 0 .1-.03.13-.11.17-.6.48-.85.73-.85.11 0 .17.05.17.24 0 .17-.03.27-.2.7zm0 0"/>\n              </symbol>\n              <symbol id="q" overflow="visible">\n                <path d="M3.7-1.86c0-.76-.58-1.22-1.26-1.22-1.03 0-2 .97-2 1.92C.44-.46.94.06 1.7.06c1 0 2-.9 2-1.92zm-2 1.74c-.34 0-.68-.22-.68-.79 0-.28.12-.96.4-1.36.3-.43.7-.6 1.02-.6.37 0 .69.25.69.76 0 .17-.08.86-.4 1.34-.26.43-.67.65-1.03.65zm0 0"/>\n              </symbol>\n              <symbol id="r" overflow="visible">\n                <path d="M.84-.44c-.01.1-.06.27-.06.28 0 .16.13.22.24.22.12 0 .23-.08.28-.14.03-.06.07-.3.12-.44l.14-.62c.05-.16.1-.31.13-.47.08-.28.1-.34.3-.62.18-.29.51-.65 1.04-.65.4 0 .4.36.4.49 0 .42-.29 1.19-.4 1.48-.08.2-.1.27-.1.38 0 .37.29.6.65.6.7 0 1-.96 1-1.07 0-.1-.08-.1-.11-.1-.1 0-.1.05-.13.13-.15.56-.46.84-.73.84-.16 0-.19-.09-.19-.24 0-.16.05-.26.17-.57.08-.22.36-.95.36-1.34 0-.67-.53-.8-.9-.8-.58 0-.97.36-1.17.64-.05-.48-.46-.64-.75-.64-.3 0-.46.22-.55.38-.16.26-.25.65-.25.7 0 .08.1.08.12.08.1 0 .1-.02.14-.2.11-.41.25-.75.52-.75.19 0 .23.15.23.34 0 .12-.06.39-.12.58l-.14.62zm0 0"/>\n              </symbol>\n              <symbol id="s" overflow="visible">\n                <path d="M3-2.63c-.17.05-.3.2-.3.33 0 .17.14.24.24.24.08 0 .34-.05.34-.4 0-.46-.5-.62-.94-.62-1.07 0-1.28.81-1.28 1.02 0 .26.16.44.25.51.17.14.3.18.78.25.16.03.6.11.6.46 0 .12-.08.39-.36.56-.28.16-.63.16-.7.16-.29 0-.68-.07-.85-.3.24-.03.4-.2.4-.4 0-.16-.13-.26-.27-.26C.7-1.08.5-.9.5-.6c0 .42.44.67 1.1.67 1.3 0 1.54-.87 1.54-1.14 0-.64-.7-.76-.97-.81l-.28-.06c-.25-.05-.37-.2-.37-.35 0-.15.12-.34.28-.45.18-.1.42-.13.54-.13.14 0 .5.02.66.25zm0 0"/>\n              </symbol>\n              <symbol id="z" overflow="visible">\n                <path d="M3.4-4.22c.04-.2.05-.22.2-.23h.42c.43 0 .6 0 .78.04.3.1.33.3.33.55 0 .11 0 .2-.05.56l-.02.08c0 .08.05.11.13.11.1 0 .1-.06.12-.17l.19-1.33c0-.1-.08-.1-.2-.1H1.02c-.18 0-.2 0-.24.15L.33-3.33.3-3.2c0 .03.01.1.12.1.1 0 .1-.04.14-.18.4-1.11.64-1.17 1.7-1.17h.29c.22 0 .22 0 .22.06 0 0 0 .05-.04.14L1.81-.58c-.06.25-.08.33-.81.33-.25 0-.31 0-.31.16C.69-.08.7 0 .8 0l.6-.02L2-.03l.64.01.58.02c.06 0 .15 0 .15-.16 0-.09-.06-.09-.28-.09l-.4-.02c-.24-.01-.25-.04-.25-.12 0-.06 0-.06.03-.17zm0 0"/>\n              </symbol>\n              <symbol id="x" overflow="visible">\n                <path d="M3.88-2.77L1.89-4.75c-.12-.13-.14-.14-.22-.14a.2.2 0 0 0-.2.2c0 .07.01.08.12.19l2 2-2 2.02c-.1.1-.12.12-.12.18 0 .13.1.2.2.2.08 0 .1 0 .22-.13l1.99-1.99L5.94-.16c.01.02.08.07.14.07.12 0 .2-.08.2-.2 0-.02 0-.05-.03-.12-.02-.01-1.6-1.57-2.1-2.09l1.83-1.81c.05-.07.2-.19.25-.25 0-.02.05-.07.05-.13 0-.12-.08-.2-.2-.2-.08 0-.11.03-.22.14zm0 0"/>\n              </symbol>\n            </defs>\n            <use x=".11" y="15.07" xlink:href="#a"/>\n            <use x="7.32" y="15.07" xlink:href="#b"/>\n            <use x="12.15" y="15.07" xlink:href="#c"/>\n            <use x="15.75" y="15.07" xlink:href="#d"/>\n            <use x="21.02" y="15.07" xlink:href="#e"/>\n            <use x="24.19" y="15.07" xlink:href="#f"/>\n            <use x="30.87" y="15.07" xlink:href="#g"/>\n            <use x="35.18" y="15.07" xlink:href="#b"/>\n            <use x="40.02" y="15.07" xlink:href="#h"/>\n            <use x="44.79" y="15.07" xlink:href="#i"/>\n            <use x="52.18" y="15.07" xlink:href="#j"/>\n            <use x="63.89" y=".73" xlink:href="#k"/>\n            <use x="74.41" y="11.19" xlink:href="#l"/>\n            <use x="79.68" y="11.19" xlink:href="#m"/>\n            <use x="83.46" y="11.19" xlink:href="#n"/>\n            <use x="87.02" y="11.19" xlink:href="#o"/>\n            <use x="90.03" y="11.19" xlink:href="#p"/>\n            <use x="92.85" y="11.19" xlink:href="#q"/>\n            <use x="96.78" y="11.19" xlink:href="#r"/>\n            <use x="101.7" y="11.19" xlink:href="#s"/>\n            <use x="107.62" y="8.2" xlink:href="#t"/>\n            <use x="115.02" y="8.2" xlink:href="#i"/>\n            <use x="119.66" y="8.2" xlink:href="#u"/>\n            <use x="123.09" y="8.2" xlink:href="#v"/>\n            <use x="128.2" y="8.2" xlink:href="#w"/>\n            <use x="133.94" y="8.2" xlink:href="#c"/>\n            <use x="137.54" y="9.69" xlink:href="#s"/>\n            <use x="141.3" y="9.69" xlink:href="#m"/>\n            <use x="145.08" y="9.69" xlink:href="#n"/>\n            <use x="148.65" y="9.69" xlink:href="#o"/>\n            <use x="151.65" y="9.69" xlink:href="#p"/>\n            <use x="154.47" y="9.69" xlink:href="#q"/>\n            <use x="158.4" y="9.69" xlink:href="#r"/>\n            <use x="166.04" y="8.2" xlink:href="#x"/>\n            <use x="176" y="8.2" xlink:href="#y"/>\n            <use x="179.87" y=".73" xlink:href="#k"/>\n            <use x="190.39" y="11.19" xlink:href="#z"/>\n            <use x="196.17" y="11.19" xlink:href="#m"/>\n            <use x="199.95" y="11.19" xlink:href="#s"/>\n            <use x="203.71" y="11.19" xlink:href="#o"/>\n            <use x="206.71" y="11.19" xlink:href="#s"/>\n            <use x="212.64" y="8.2" xlink:href="#t"/>\n            <use x="220.03" y="8.2" xlink:href="#i"/>\n            <use x="224.67" y="8.2" xlink:href="#u"/>\n            <use x="228.1" y="8.2" xlink:href="#v"/>\n            <use x="233.21" y="8.2" xlink:href="#w"/>\n            <use x="238.95" y="8.2" xlink:href="#c"/>\n            <use x="242.55" y="9.69" xlink:href="#o"/>\n            <use x="245.56" y="9.69" xlink:href="#m"/>\n            <use x="249.34" y="9.69" xlink:href="#s"/>\n            <use x="253.1" y="9.69" xlink:href="#o"/>\n            <use x="258.82" y="8.2" xlink:href="#x"/>\n            <use x="268.79" y="8.2" xlink:href="#A"/>\n            <use x="273.46" y="8.2" xlink:href="#g"/>\n            <use x="277.77" y="8.2" xlink:href="#b"/>\n            <use x="282.6" y="8.2" xlink:href="#h"/>\n            <use x="287.37" y="8.2" xlink:href="#i"/>\n            <use x="292.01" y="9.69" xlink:href="#o"/>\n            <use x="295.01" y="9.69" xlink:href="#m"/>\n            <use x="298.79" y="9.69" xlink:href="#s"/>\n            <g>\n              <use x="302.56" y="9.69" xlink:href="#o"/>\n            </g>\n            <g>\n              <use x="306.07" y="8.2" xlink:href="#B"/>\n            </g>\n            <path fill="none" stroke="#000" stroke-miterlimit="10" stroke-width=".405" d="M63.89 12.58h246.06"/>\n            <g>\n              <use x="163.44" y="21.91" xlink:href="#C"/>\n            </g>\n            <g>\n              <use x="174.2" y="21.91" xlink:href="#d"/>\n            </g>\n            <g>\n              <use x="179.46" y="21.91" xlink:href="#D"/>\n            </g>\n            <g>\n              <use x="185.16" y="21.91" xlink:href="#f"/>\n            </g>\n            <g>\n              <use x="191.85" y="21.91" xlink:href="#g"/>\n            </g>\n            <g>\n              <use x="196.16" y="21.91" xlink:href="#b"/>\n            </g>\n            <g>\n              <use x="200.99" y="21.91" xlink:href="#h"/>\n            </g>\n            <g>\n              <use x="205.76" y="21.91" xlink:href="#i"/>\n            </g>\n         </svg>\n         <br>\n\n<p>Weights for sections and individual tests are indicated by a white number\ninside a magenta badge. No badge means that the weight is defaults to 1.</p>\n\n            <mat-expansion-panel>\n            <mat-expansion-panel-header>\n                <mat-panel-title fxLayoutAlign=\'space-between center\'>\n                    <td class="test-title"> Test Title </td>\n                    <td class="MatButtonEmulation"> 100% </td>\n                </mat-panel-title>\n                <div class=\'badge\'> <span class=\'vert-align\'> X3 </span> </div>\n            </mat-expansion-panel-header>\n          </mat-expansion-panel>\n\n</mat-card-content>\n</mat-card>\n\n\n<mat-card>\n<mat-card-title>Paradigms</mat-card-title>\n<mat-card-content>\n\n<h3>"Reconstructions" and "Models"</h3>\n\n<p>Some authors may publish metabolic networks which are parameterized,\nready to run flux balance analysis (FBA), these are referred to simply as\n\'models\'. Alternatively, others may publish unconstrained metabolic knowledgebases\n(referred to as \'reconstructions\'), from which several models can be derived\nby applying different constraints. Both can be encoded in SBML. With having\nan independent test section, we attempt to make both \'models\' and\n\'reconstructions\' comparable, although a user should be aware that this\ndifference exists and is subject to <a href="https://github.com/opencobra/memote/issues/228"> some discussion </a>. Please note, that some\ntests in the specific section may error for a reconstruction as they\nrequire initialization.</p>\n\n<h3>"Lumped" and "Split" Biomass Reaction</h3>\n\n<p>There are two basic ways of specifying the biomass composition. The most\ncommon is a single lumped reaction containing all biomass precursors.\nAlternatively, the biomass equation can be split into several reactions\neach focusing on a different macromolecular component for instance\na (1 gDW ash) + b (1 gDW phospholipids) + c (free fatty acids)+\nd (1 gDW carbs) + e (1 gDW protein) + f (1 gDW RNA) + g (1 gDW DNA) +\nh (vitamins/cofactors) + xATP + xH2O-> 1 gDCW biomass + xADP + xH + xPi. The\nbenefit of either approach depends very much on the use cases which are\n<a href="https://github.com/opencobra/memote/issues/243"> discussed by the community </a>. Memote employs heuristics to identify the type\nof biomass which may fail to distinguish edge cases.</p>\n\n\n<h3>"Average" and "Unique" Metabolites</h3>\n\n<p>A metabolite consisting of a fixed core with variable branches such as a\nmembrane lipid are sometimes implemented by averaging over the distribution of\nindividual lipid species. The resulting pseudo-metabolite is assigned an\naverage chemical formula, which requires scaling of stoichiometries of\nassociated reactions to avoid floating point numbers in the chemical formulae.\nAn alternative approach is to implement each species as a distinct\nmetabolite in the model, which increases the total count of reactions. Memote\ncannot yet distinguish between these paradigms, which means that results\nin the specific sections that rely on the total number of reactions or scaling\nof stochiometric parameters may be biased.</p>\n\n</mat-card-content>\n</mat-card>\n\n</mat-dialog-content>\n<mat-dialog-actions>\n        <button mat-raised-button mat-dialog-close color="primary">Close</button>\n        <a mat-raised-button color="primary" href="https://memote.readthedocs.io/en/latest/index.html" target="_blank" rel="noopener noreferrer">Open Online Documentation</a>\n</mat-dialog-actions>'},"./src/app/help-dialog/help-dialog.component.html":function(t,e){t.exports='<button mat-button (click)="openDialog()">Readme</button>'},"./src/app/help-dialog/help-dialog.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL2hlbHAtZGlhbG9nL2hlbHAtZGlhbG9nLmNvbXBvbmVudC5zY3NzIn0= */"},"./src/app/help-dialog/help-dialog.component.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"HelpDialogComponent",function(){return i}),n.d(e,"HelpDialogTextComponent",function(){return r});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/material/esm5/material.es5.js"),i=function(){function t(t){this.dialog=t}return t.prototype.openDialog=function(){this.dialog.open(r,{height:"90%",width:"70%"}).afterClosed().subscribe(function(t){console.log("Dialog result: "+t)})},t=s.__decorate([Object(a.Component)({selector:"app-help-dialog",template:n("./src/app/help-dialog/help-dialog.component.html"),styles:[n("./src/app/help-dialog/help-dialog.component.scss")]}),s.__metadata("design:paramtypes",[o.MatDialog])],t)}(),r=function(){function t(){}return t=s.__decorate([Object(a.Component)({selector:"app-help-dialog-text",template:n("./src/app/help-dialog/help-dialog-text.component.html")})],t)}()},"./src/app/keys.pipe.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"KeysPipe",function(){return o});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),o=function(){function t(){}return t.prototype.transform=function(t,e){return void 0===e&&(e=null),Object.keys(t)},t=s.__decorate([Object(a.Pipe)({name:"keys",pure:!1})],t)}()},"./src/app/report-data.service.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"ReportDataService",function(){return d});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/common/fesm5/http.js"),i=n("./src/app/resultcard.model.ts"),r=n("./src/app/test-result.model.ts"),c=n("./src/app/test-history.model.ts"),l=n("./src/app/test-diff.model.ts"),d=function(){function t(t){this.http=t,this.allTests=[],this.statisticsCards=[],this.scoredTests=[],this.allExpandState=!1,this.scorePerSection={},this.testWeights={},this.testWeightsExpanded={}}return t.prototype.loadResults=function(){var t=this;switch(this.reportType=window.reportType,this.reportType){case"history":this.convertHistoryResults(window.data);break;case"snapshot":this.convertResults(window.data);break;case"diff":this.convertDiffResults(window.data);break;default:this.http.get("/data/testHistory.json").subscribe(function(e){t.convertHistoryResults(e)}),this.reportType="history"}},t.prototype.togglePanel=function(){this.allExpandState=!this.allExpandState},t.prototype.byID=function(t){return this.allTests.find(function(e){return e.id===t})},t.prototype.byReg=function(t){for(var e=new RegExp(t),n=[],s=0,a=this.allTests;s<a.length;s++){var o=a[s];e.test(o.id)&&n.push(o.id)}return n},t.prototype.isScored=function(t){return this.scoredTests.includes(t)},t.prototype.getParam=function(t,e){return t.split(":")[e]},t.prototype.getString=function(t){return JSON.stringify(t)},t.prototype.convertResults=function(t){for(var e=0,n=Object.keys(t.tests);e<n.length;e++){var s=n[e];if(t.tests[s].message instanceof Object)for(var a=0,o=Object.keys(t.tests[s].result);a<o.length;a++){var i=o[a],c=s+":"+i;this.allTests.push(new r.TestResult(c,{data:t.tests[s].data[i],duration:t.tests[s].duration[i],message:t.tests[s].message[i],metric:t.tests[s].metric[i],result:t.tests[s].result[i],summary:t.tests[s].summary,title:t.tests[s].title,format_type:t.tests[s].format_type}))}else this.allTests.push(new r.TestResult(s,t.tests[s]))}this.extractMetadata(t),this.extractScoring(t),this.extractTestWeights(t),this.distributeCardsToSections(t),this.determineScoredTests(),this.determineScorePerSection()},t.prototype.convertHistoryResults=function(t){for(var e=0,n=Object.keys(t.tests);e<n.length;e++){var s=n[e];if(t.tests[s].history instanceof Array)this.allTests.push(new c.TestHistory(s,t.tests[s]));else for(var a=0,o=Object.keys(t.tests[s].history);a<o.length;a++){var i=o[a],r=s+":"+i;this.allTests.push(new c.TestHistory(r,{history:t.tests[s].history[i],summary:t.tests[s].summary,title:t.tests[s].title,format_type:t.tests[s].format_type}))}}this.distributeCardsToSections(t),this.determineScoredTests(),this.extractScoring(t)},t.prototype.convertDiffResults=function(t){for(var e=0,n=Object.keys(t.tests);e<n.length;e++){var s=n[e];if(t.tests[s].diff instanceof Array)this.allTests.push(new l.TestDiff(s,t.tests[s]));else for(var a=0,o=Object.keys(t.tests[s].diff);a<o.length;a++){var i=o[a],r=s+":"+i;this.allTests.push(new l.TestDiff(r,{diff:t.tests[s].diff[i],summary:t.tests[s].summary,title:t.tests[s].title,format_type:t.tests[s].format_type}))}}this.distributeCardsToSections(t),this.determineScoredTests(),this.extractScoring(t),this.extractTestWeights(t),this.determineScorePerSection()},t.prototype.extractMetadata=function(t){this.metaData=t.meta},t.prototype.extractScoring=function(t){this.score=t.score},t.prototype.extractTestWeights=function(t){this.testWeights=t.weights;for(var e=0,n=Object.keys(this.testWeights);e<n.length;e++)for(var s=n[e],a=0,o=this.byReg(s);a<o.length;a++){var i=o[a];this.testWeightsExpanded[i]=this.testWeights[s]}},t.prototype.distributeCardsToSections=function(t){for(var e=0,n=Object.keys(t.cards);e<n.length;e++){var s=n[e];"scored"===s?this.scoredCard=t.cards.scored:this.statisticsCards.push(new i.ResultCard(s,t.cards[s]))}},t.prototype.determineScoredTests=function(){for(var t=0,e=Object.keys(this.scoredCard.sections);t<e.length;t++){var n=e[t];if(this.scoredCard.sections[n].cases instanceof Array)for(var s=0,a=this.scoredCard.sections[n].cases;s<a.length;s++){var o=a[s];this.scoredTests.push(o)}}},t.prototype.extractData=function(t){for(var e={},n=0,s=this.byID(t).diff;n<s.length;n++){var a=s[n];e[a.model]=a.data}return this.getString(e)},t.prototype.determineScorePerSection=function(){var t=[],e=[];if("snapshot"===this.reportType)e=this.score.sections,t=[this.score.total_score];else{e=this.score.sections.diff;for(var n=0,s=this.score.total_score.diff;n<s.length;n++){var a=s[n];t.push(a.total_score)}}for(var o=[],i=[],r=0,c=e;r<c.length;r++){var l=c[r],d=l.section;this.scorePerSection.hasOwnProperty(d)?(this.scorePerSection[d].weights.push(this.scoredCard.sections[d].weight),this.scorePerSection[d].scores.push(l.score)):(o.push(this.scoredCard.sections[d].weight),i.push(l.score),this.scorePerSection[d]={weights:o,scores:i}),o=[],i=[]}this.scorePerSection["Total Score"]={scores:t}},t=s.__decorate([Object(a.Injectable)(),s.__metadata("design:paramtypes",[o.HttpClient])],t)}()},"./src/app/result-button/result-button.component.html":function(t,e){t.exports="\x3c!-- First distinguish wether the tests have passed/failed or skipped or errored --\x3e\n<div *ngIf=\"data.byID(testId)?.result === 'passed' || data.byID(testId)?.result === 'failed'\">\n  \x3c!-- Then determine if the test results are from scored or from unscored tests --\x3e\n  <div *ngIf=\"data.isScored(data.getParam(testId, 0)); else notScored\">\n    \x3c!-- If they are from scored tests colorize the result button --\x3e\n    <div\n    class=\"MatButtonEmulation\"\n    [ngStyle]=\"{'background-color':getColor(1 - data.byID(testId)?.metric | number : '1.2-2')}\">\n      {{ 1 - data.byID(testId)?.metric | percent : '1.1' }}\n      \x3c!-- {{ 1 - data.byID(testId)?.metric | percent : '1.2-2' }} --\x3e\n  </div>\n  </div>\n  <ng-template #notScored>\n    \x3c!-- If they are from unscored tests the user has some control over how --\x3e\n    \x3c!-- they are displayed using the 'format_type' attribute of each test. --\x3e\n    <div\n    class=\"MatButtonEmulation\">\n      \x3c!-- Hence for each test, determine now how the result ought to be formatted --\x3e\n      <div *ngIf=\"data.byID(testId)?.format_type === 'count'\">\n        {{ data.byID(testId)?.data.length | number : '1.'}}\n      </div>\n      <div *ngIf=\"data.byID(testId)?.format_type === 'number'\">\n        {{ data.byID(testId)?.data | number : '1.2-2'}}\n      </div>\n      <div *ngIf=\"data.byID(testId)?.format_type === 'percent'\">\n        {{ data.byID(testId)?.metric | number : '1.2-2'}}\n      </div>\n      <div *ngIf=\"data.byID(testId)?.format_type === 'raw'\">\n        {{ data.byID(testId)?.data }}\n      </div>\n    </div>\n  </ng-template>\n</div>\n<div *ngIf=\"data.byID(testId)?.result === 'skipped'\">\n\x3c!-- Skipped tests are marked with a gray button with beige text --\x3e\n    <div\n     class=\"MatButtonEmulation\"\n     style=\"background:gray; color:beige\">\n       Skipped\n     </div>\n</div>\n<div *ngIf=\"data.byID(testId)?.result === 'error'\">\n\x3c!-- Tests with errors are marked with a black button with red text --\x3e\n    <div\n     class=\"MatButtonEmulation\"\n     style=\"color:red;background-color:black\">\n       Errored\n    </div>\n</div>\n"},"./src/app/result-button/result-button.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL3Jlc3VsdC1idXR0b24vcmVzdWx0LWJ1dHRvbi5jb21wb25lbnQuc2NzcyJ9 */"},"./src/app/result-button/result-button.component.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"ResultButtonComponent",function(){return i});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./src/app/report-data.service.ts"),i=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},t.prototype.getColor=function(t){return["hsl(",(132*t).toString(10),",80%,35%)"].join("")},s.__decorate([Object(a.Input)(),s.__metadata("design:type",String)],t.prototype,"testId",void 0),t=s.__decorate([Object(a.Component)({selector:"app-result-button",template:n("./src/app/result-button/result-button.component.html"),encapsulation:a.ViewEncapsulation.None,styles:[n("./src/app/result-button/result-button.component.scss")]}),s.__metadata("design:paramtypes",[o.ReportDataService])],t)}()},"./src/app/resultcard.model.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"ResultCard",function(){return s});var s=function(){return function(t,e){var n=e.title,s=e.cases;this.cardID=t,this.cardName=n,this.associatedTests=s}}()},"./src/app/score-formula/score-formula.component.html":function(t,e){t.exports='<div>\n    <mat-expansion-panel\n    #expPanel\n    class="mat-elevation-z0">\n       <mat-expansion-panel-header\n       class="mat-expansion-panel-header"\n       collapsedHeight="auto"\n       style="padding:0.05vh">\n       <mat-panel-title fxLayoutAlign=\'space-between center\'>\n          <td class="test-title">\n              {{ scoreType }}\n          </td>\n          <div style="margin-right: 10px">\n          <td *ngFor=\'let score of modelScore\'>\n            <div class="MatButtonEmulation" style="background: white; color:black"> {{ score | percent : \'1.0\' }} </div>\n          </td>\n          </div>\n         </mat-panel-title>\n         <div *ngIf=\'getSectionWeight() ; else whiteBadge\'> \n            <div class=\'badge\'> <span class="vert-align"> X{{ sectionWeight }} </span> </div> \n          </div>\n          <ng-template #whiteBadge>\n              <div class=\'badge\' style="background-color: white"></div> \n          </ng-template>\n       </mat-expansion-panel-header>\n       <div>\n         <p>The {{ scoreType}} is the result of the following calculation. For more information please click on "Readme" in the top left of the report.</p>\n       </div>\n       <div *ngFor=\'let formula of formulae\'>\n          <p style="font-size: 36px; font-size: 0.6vmax"> {{ formula }} </p>\n        </div>\n      </mat-expansion-panel>  \n</div>\n'},"./src/app/score-formula/score-formula.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL3Njb3JlLWZvcm11bGEvc2NvcmUtZm9ybXVsYS5jb21wb25lbnQuc2NzcyJ9 */"},"./src/app/score-formula/score-formula.component.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"ScoreFormulaComponent",function(){return i});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./src/app/report-data.service.ts"),i=function(){function t(t){this.data=t,this.scoreType="Total Score",this.modelScore=[],this.formulae=[],this.sections=[],this.sectionWeight=1}return t.prototype.ngOnInit=function(){this.getScore(),this.constructFormulae()},t.prototype.getSectionWeight=function(){if("Total Score"!==this.position){var t=this.data.scorePerSection[this.position].weights[0];return 1!==t&&void 0!==t&&(this.sectionWeight=t,!0)}},t.prototype.getScore=function(){this.modelScore=this.data.scorePerSection[this.position].scores,"Total Score"!==this.position&&(this.scoreType="Sub Total")},t.prototype.constructFormulae=function(){var t;for(t=0;t<this.data.scorePerSection["Total Score"].scores.length;t++)if("Total Score"===this.scoreType){var e=this.constructTotalFormula(t);this.formulae.push(e)}else{e=this.constructSubTotalFormula(t);this.formulae.push(e)}},t.prototype.constructTotalFormula=function(t){for(var e="",n="",s=!0,a="",o=0,i=Object.keys(this.data.scorePerSection);o<i.length;o++){var r=i[o];if("Total Score"!==r){var c=this.data.scorePerSection[r].weights[t];e+=a+"("+c+" * "+(100*this.data.scorePerSection[r].scores[t]).toFixed(2)+")",n+=a+"("+c+" * 100)"}s&&(a=" + ",s=!1)}return"`("+e+")/("+n+") = "+(100*this.data.scorePerSection["Total Score"].scores[t]).toFixed(2)+"`"},t.prototype.constructSubTotalFormula=function(t){for(var e="",n="",s=0,a=0,o=0,i=0,r=0,c=this.data.scoredCard.sections[this.position].cases;r<c.length;r++){var l=c[r];if(l in Object.keys(this.data.testWeights))this.data.testWeights[l];for(var d=0,p=this.data.byReg(l);d<p.length;d++){var u=p[d];0!==(i="snapshot"===this.data.reportType?100*(1-this.data.byID(u).metric):100*(1-this.data.byID(u).diff[t].metric))?100!==i?(e+=i.toFixed(2)+" + ",s+=1):(a+=1,s+=1):(s+=1,o+=1)}}return a>0&&(e+="("+a+" * 100) + "),o>0&&(e+="("+o+" * 0)"),"`("+e+")/("+(n+="("+s+" * 100)")+") = "+(100*this.data.scorePerSection[this.position].scores[t]).toFixed(2)+"`"},s.__decorate([Object(a.Input)(),s.__metadata("design:type",String)],t.prototype,"position",void 0),t=s.__decorate([Object(a.Component)({selector:"app-score-formula",template:n("./src/app/score-formula/score-formula.component.html"),encapsulation:a.ViewEncapsulation.None,styles:[n("./src/app/score-formula/score-formula.component.scss")]}),s.__metadata("design:paramtypes",[o.ReportDataService])],t)}()},"./src/app/tauchart-bar/tauchart-bar.component.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"TauChartBarComponent",function(){return r});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./src/app/report-data.service.ts"),i=n("./node_modules/taucharts/dist/taucharts.js"),r=(n("./node_modules/taucharts/dist/plugins/tooltip.js"),n("./node_modules/taucharts/dist/plugins/export-to.js"),function(){function t(t){this.data=t,this.unique_id="total"}return t.prototype.initialize=function(){this.chart=new i.Chart(this.chart_settings),this.chart.renderTo("#tau-horizbar-"+this.unique_id)},t.prototype.composeChartSettings=function(){switch(this.plotType){case"snapshot-sections":this.chart_settings.y="section",this.chart_settings.x="score",this.chart_settings.color="score",this.chart_settings.data=this.data.score.sections,this.chart_settings.guide.x={nice:!1,min:0,max:1,tickFormat:"percent"},this.chart_settings.guide.color={brewer:["rgb(161, 18, 18)","rgb(18, 161, 46)"]},this.unique_id="sections";break;case"diff-sections":this.chart_settings.y="model",this.chart_settings.x="score",this.chart_settings.color="score",this.chart_settings.data=this.data.score.sections.diff,this.chart_settings.guide.x={nice:!1,min:0,max:1,tickFormat:"percent"},this.chart_settings.guide.color={brewer:["rgb(161, 18, 18)","rgb(18, 161, 46)"]},this.unique_id="sections";break;case"diff-total":this.chart_settings.y="model",this.chart_settings.x="total_score",this.chart_settings.data=this.data.score.total_score.diff,this.chart_settings.guide.x={nice:!1,min:0,max:1,tickFormat:"percent"}}},t.prototype.ngOnInit=function(){this.chart_settings={type:"horizontal-bar",y:NaN,x:NaN,color:NaN,data:NaN,plugins:[i.api.plugins.get("tooltip")(),i.api.plugins.get("export-to")()],guide:{showGridLines:"xy"},settings:{fitModel:"normal"}},this.composeChartSettings()},t.prototype.ngAfterViewInit=function(){this.initialize()},s.__decorate([Object(a.Input)(),s.__metadata("design:type",String)],t.prototype,"plotType",void 0),t=s.__decorate([Object(a.Component)({selector:"app-tauchart-bar",template:'<div id="tau-horizbar-{{unique_id}}" style="height: 50vh; width: 40vw"> </div>',encapsulation:a.ViewEncapsulation.None}),s.__metadata("design:paramtypes",[o.ReportDataService])],t)}())},"./src/app/tauchart-history/tauchart-history.component.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"TauChartHistoryComponent",function(){return r});var s=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./src/app/report-data.service.ts"),i=n("./node_modules/taucharts/dist/taucharts.js"),r=(n("./node_modules/taucharts/dist/plugins/tooltip.js"),n("./node_modules/taucharts/dist/plugins/legend.js"),n("./node_modules/taucharts/dist/plugins/export-to.js"),function(){function t(t){this.data=t,this.unique_id="total"}return t.prototype.initialize=function(){this.chart.renderTo("#tau-history-"+this.unique_id)},t.prototype.invertScoredData=function(t){for(var e=0,n=t;e<n.length;e++){var s=n[e];s.metric=1-s.metric}},t.prototype.ngOnInit=function(){this.format_type=this.testObject.format_type;var t={showAnchors:"always",interpolate:"linear",showGridLines:"xy",x:{nice:!1}};this.testId?(this.unique_id=this.testId.replace(/:|\./g,"_"),this.data.isScored(this.data.getParam(this.testId,0))?(this.format_type="metric",this.invertScoredData(this.testObject.history),t.y={min:0,max:1,nice:!1,tickFormat:"percent",label:{text:this.testObject.title}}):(this.format_type="data",t.y={label:{text:this.testObject.title}})):(this.format_type="metric",t.y={min:0,max:1,nice:!1,tickFormat:"percent"}),this.chart=new i.Chart({data:this.testObject.history,type:"line",x:"commit",y:this.format_type,color:"branch",settings:{asyncRendering:!0,renderingTimeout:1e3},plugins:[i.api.plugins.get("legend")(),i.api.plugins.get("tooltip")(),i.api.plugins.get("export-to")()],guide:t})},t.prototype.ngAfterViewInit=function(){this.initialize()},s.__decorate([Object(a.Input)(),s.__metadata("design:type",Object)],t.prototype,"testObject",void 0),s.__decorate([Object(a.Input)(),s.__metadata("design:type",String)],t.prototype,"testId",void 0),t=s.__decorate([Object(a.Component)({selector:"app-tauchart-history",template:'<div id="tau-history-{{unique_id}}" style="height: 50vh; width: 40vw"> </div>',encapsulation:a.ViewEncapsulation.None}),s.__metadata("design:paramtypes",[o.ReportDataService])],t)}())},"./src/app/test-diff.model.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"TestDiff",function(){return s});var s=function(){return function(t,e){var n=e.diff,s=e.summary,a=e.title,o=e.format_type;this.id=t,this.diff=n,this.summary=s,this.title=a,this.format_type=o}}()},"./src/app/test-history.model.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"TestHistory",function(){return s});var s=function(){return function(t,e){var n=e.history,s=e.summary,a=e.title,o=e.format_type;this.id=t,this.history=n,this.summary=s,this.title=a,this.format_type=o}}()},"./src/app/test-result.model.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"TestResult",function(){return s});var s=function(){function t(t,e){var n=e.data,s=e.duration,a=e.message,o=e.metric,i=e.result,r=e.summary,c=e.title,l=e.format_type;this.id=t,this.data=n,this.duration=s,this.message=a,this.metric=o,this.errorFailsafe(i),this.summary=r,this.title=c,this.format_type=l}return t.prototype.errorFailsafe=function(t){"skipped"===t||null!==this.data&&void 0!==this.data?this.result=t:this.result="error"},t}()},"./src/environments/environment.ts":function(t,e,n){"use strict";n.r(e),n.d(e,"environment",function(){return s});var s={production:!1}},"./src/main.ts":function(t,e,n){"use strict";n.r(e);var s=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/platform-browser-dynamic/fesm5/platform-browser-dynamic.js"),o=n("./src/app/app.module.ts");n("./src/environments/environment.ts").environment.production&&Object(s.enableProdMode)(),Object(a.platformBrowserDynamic)().bootstrapModule(o.AppModule).catch(function(t){return console.log(t)})},0:function(t,e,n){t.exports=n("./src/main.ts")}},[[0,"runtime","vendor"]]]);
</script></body>
</html>
